/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
 
const IOUtils = Java.type('org.apache.commons.io.IOUtils');
const StringUtils = Java.type("org.apache.commons.lang.StringUtils");
const outgoingReferencesExtractor = Ametys.serviceManager.lookup('org.ametys.cms.content.references.OutgoingReferencesExtractor');
const XPATH_PREFIX = "//element(*, ametys:content)[@ametys-internal:contentType = '";
const XPATH_SUFFIX = "']";

let handledContentsIds = {}; // List of modified contents (to avoid multiple modifications)
let handledContentTypes = 0;
let contentTypeExtensionPoint = Ametys.serviceManager.lookup('org.ametys.cms.contenttype.ContentTypeExtensionPoint');
let contentTypeIds = contentTypeExtensionPoint.getExtensionsIds();
let totalContentTypes = contentTypeIds.size();
let modifiedContents = 0;

/* Iterate on content types */
for (let contentTypeId of contentTypeIds)
{
    let contentType = contentTypeExtensionPoint.getExtension(contentTypeId);

    if (!contentType.isAbstract())
    {
        let richTexts = __getRichTexts(contentType);
        if (richTexts.size > 0)
        {
            logger.info(`Handling contents of type '${contentTypeId}'`);
            __migrateContentsOfContentType(contentTypeId, richTexts);
        }
    }
    else
    {
        logger.info(`Ignoring abstract content type ${contentTypeId}`);
    }
    
    handledContentTypes++;
    if (handledContentTypes % 10 == 0)
    {
        logger.info(`[${handledContentTypes}/${totalContentTypes}] handled content types`);
    }
}
logger.info(`Finished with ${modifiedContents} contents modified on default workspace`);

/* Methods */
function __getRichTexts(modelItemContainer)
{
    let paths = new Map();
    for (let modelItem of modelItemContainer.getModelItems())
    {
        let typeId = modelItem.getType().getId();
        if (typeId == 'repeater' || typeId == 'composite')
        {
            let richTextPaths = __getRichTexts(modelItem);
            if (richTextPaths.size > 0)
            {
                paths.set(modelItem.getName(), { "type": typeId, "dataPaths": richTextPaths });
            }
        }
        else if (typeId == 'rich-text' && modelItem.isEditable())
        {
            paths.set(modelItem.getName(), { "type": typeId });
        }
    }
    return paths;
}

function __migrateContentsOfContentType(contentTypeId, richTexts)
{
    let contents = Repository.query(XPATH_PREFIX + contentTypeId + XPATH_SUFFIX);
    let totalContents = contents.getSize();
    
    let handledContents = 0;
        
    contents.forEach(function(content) {
        try
        {
            __redoOutgoingReferences(content, richTexts);
        }
        catch (e)
        {
            logger.error(`Content ${content.getId()} migration failure`, e);
            throw e;
        }
        
        handledContents++;
        if (handledContents % 500 == 0)
        {
            logger.info(`[${handledContents}/${totalContents}] handled contents of type ${contentTypeId}`);
        }    
    });
    logger.info(`${handledContents} handled contents of type ${contentTypeId}`);
}

function __redoOutgoingReferences(content, richTexts)
{
    if (!handledContentsIds[content.getId()] && content instanceof org.ametys.cms.repository.ModifiableContent)
    {
        if (__hasToMigrateDataHolder(richTexts, content, "", "'" + content.getTitle() + "' (" + content.getId() + ")"))
        {
            logger.debug(`Modifying content ${content.getId()} - ${content.getTitle()}`);
            
            if (content.isLocked())
            {
                content.unlock();
            }
            let outgoingReferencesByPath = outgoingReferencesExtractor.getOutgoingReferences(content);
            content.setOutgoingReferences(outgoingReferencesByPath);
            content.saveChanges();
        
            // Remember that this content was modified to avoid working on it again if it has multiple contnet types
            handledContentsIds[content.getId()] = true;
            
            modifiedContents++;
        }
    }
}

function __hasToMigrateDataHolder(richTexts, dataHolder, completePath, contentDisplay)
{
    for (let dataPath of richTexts.keys())
    {
        let value = dataHolder.getValue(dataPath, false, null)
        if (value != null)
        {
            let newCompletePath = completePath + dataPath;
            let entryValue = richTexts.get(dataPath);
            let type = entryValue["type"];
            if (type == "repeater")
            {
                for (let repeaterEntry of value.getEntries())
                {
                    let hasToMigrateRepeaterEntry = __hasToMigrateDataHolder(entryValue["dataPaths"], repeaterEntry, newCompletePath + "[" + repeaterEntry.getPosition() + "]/", contentDisplay);
                    if (hasToMigrateRepeaterEntry)
                    {
                        return true;
                    }
                }
            }
            else if (type == "composite")
            {
                let hasToMigrateComposite = __hasToMigrateDataHolder(entryValue["dataPaths"], value, newCompletePath + "/", contentDisplay);
                if (hasToMigrateComposite)
                {
                    return true;
                }
            }
            else if (type == "rich-text")
            {
                if (dataHolder.isMultiple(dataPath))
                {
                    for (let richText of value)
                    {
                        if (__doesRichTextRequireMigration(richText))
                        {
                            return true;
                        }
                    }
                }
                else
                {
                    if (__doesRichTextRequireMigration(value))
                    {
                        return true;
                    }
                }
            }
            else
            {
                logger.warn("Unknown type '" + typeId + "' for data " + newCompletePath + " on content " + contentDisplay);
            }
        }
    }
    
    // No rich text to migrate has been found
    return false;
}

function __doesRichTextRequireMigration(richText)
{
    let stream = null;
    try
    {
        stream = richText.getInputStream();
        let contentByteArray = IOUtils.toByteArray(stream);
        return StringUtils.contains(new java.lang.String(contentByteArray, "UTF-8"), "<mediaobject")
            && StringUtils.contains(new java.lang.String(contentByteArray, "UTF-8"), "<link");
    }
    finally
    {
        if (stream != null)
        {
            IOUtils.closeQuietly(stream);
        }
    }
}