/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/* Constants */
const CREDENTIALS = new javax.jcr.SimpleCredentials('ametys', []);
const ARCHIVES_SESSION = repository.login(CREDENTIALS, 'archives');
const XPATH_PREFIX = "//element(*, ametys:content)[@ametys-internal:contentType = '";
const XPATH_SUFFIX = "']";
const RICH_TEXT_MIGRATION_XSL_CONTENT =
"<xsl:stylesheet version=\"1.0\" xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\" xmlns=\"http://docbook.org/ns/docbook\" xmlns:docbook=\"http://docbook.org/ns/docbook\">" +
    "<xsl:param name=\"dataPath\"/>" +    
    "<xsl:template match=\"docbook:imagedata[@type='local']\">" +
        "<imagedata>" +
            "<xsl:copy-of select=\"@*[name() != 'fileref']\"/>" +
            "<xsl:attribute name=\"fileref\">" +
                "<xsl:value-of select=\"concat(substring-before(@fileref, '@'), '@', $dataPath, ';', substring-after(@fileref, ';'))\"/>" +
            "</xsl:attribute>" +
        "</imagedata>" +
    "</xsl:template>" +
    "<xsl:template match=\"*\">" + 
        "<xsl:copy>" +
            "<xsl:copy-of select=\"@*\"/>" +
            "<xsl:apply-templates/>" +
        "</xsl:copy>" +
    "</xsl:template>" +
"</xsl:stylesheet>";

/* Imports */
const IOUtils = Java.type('org.apache.commons.io.IOUtils');

/* Methods */
function __getRichTextsInRepeater(modelItemContainer, inRepeater)
{
    let paths = new Map();
    for (let modelItem of modelItemContainer.getModelItems())
    {
        let typeId = modelItem.getType().getId();
        if (typeId == 'repeater' || typeId == 'composite')
        {
            let isCurrentlyInRepeater = inRepeater || typeId == 'repeater';
            let richTextPaths = __getRichTextsInRepeater(modelItem, isCurrentlyInRepeater);
            if (richTextPaths.size > 0)
            {
                paths.set(modelItem.getName(), { "type": typeId, "dataPaths": richTextPaths });
            }
        }
        else if (typeId == 'rich-text' && inRepeater && modelItem.isEditable())
        {
            paths.set(modelItem.getName(), { "type": typeId });
        }
    }
    return paths;
}

function __migrateRichTextsDataInAllContentsByContentType(session, richTextsInRepeater, contentTypeId)
{
    let query = XPATH_PREFIX + contentTypeId + XPATH_SUFFIX;
    let contents = Repository.query(query, true, session);
    let totalContents = contents.getSize();
    let handledContents = 0;
    for (let content of contents)
    {
        Content.migrate(
            content,
            [__migrateRichTextsDataInContent.bind(null, richTextsInRepeater)],
            false,
            null,
            false,
            session == null
        );
        
        handledContents++;
        if (handledContents % 500 == 0)
        {
            logger.info("[" + handledContents + "/" + totalContents + "] Handled contents of type '" + contentTypeId + "'");
        }
    }
    
    logger.info(handledContents + " handled contents of type '" + contentTypeId + "'");
}

function __migrateRichTextsDataInContent(richTextsInRepeater, content)
{
	try
	{
		__migrateRichTextsDataInDataHolder(richTextsInRepeater, content, "", "'" + content.getTitle() + "' (" + content.getId() + ")");
	}
	catch (e)
	{
		logger.error("An error occurred on content " + content.getId(), e);
		throw e;
	}
}

function __migrateRichTextsDataInDataHolder(richTextsInRepeater, dataHolder, completePath, contentDisplay)
{
    for (let dataPath of richTextsInRepeater.keys())
    {
        let value = dataHolder.getValue(dataPath, false, null)
        if (value != null)
        {
            let newCompletePath = completePath + dataPath;
            let entryValue = richTextsInRepeater.get(dataPath);
            let type = entryValue["type"];
            if (type == "repeater")
            {
                for (let repeaterEntry of value.getEntries())
                {
                    __migrateRichTextsDataInDataHolder(entryValue["dataPaths"], repeaterEntry, newCompletePath + "[" + repeaterEntry.getPosition() + "]/", contentDisplay);
                }
            }
            else if (type == "composite")
            {
                __migrateRichTextsDataInDataHolder(entryValue["dataPaths"], value, newCompletePath + "/", contentDisplay);
            }
            else if (type == "rich-text")
            {
                if (dataHolder.isMultiple(dataPath))
                {
                    for (let richText of value)
                    {
                        __migrateRichTextData(newCompletePath, richText);
                    }
                }
                else
                {
                    __migrateRichTextData(newCompletePath, value);
                }
                dataHolder.setValue(dataPath, value)
                logger.info("Rich-text has been migrated for content " + contentDisplay + " on " + newCompletePath);
            }
            else
            {
                logger.warn("Unknown type '" + typeId + "' for data " + newCompletePath + " on content " + contentDisplay);
            }
        }
    }
}

function __migrateRichTextData(dataPath, richText)
{
    let inputStream = null;
    let outputStream = null;
    try
    {
        let transformer = templates.newTransformer();
        
        inputStream = richText.getInputStream();
        outputStream = richText.getOutputStream();
        
        let source = new javax.xml.transform.stream.StreamSource(inputStream);
        let result = new javax.xml.transform.stream.StreamResult(outputStream);
        
        transformer.setParameter("dataPath", dataPath);
        transformer.transform(source, result);
    }
    finally
    {
        IOUtils.closeQuietly(inputStream);
        IOUtils.closeQuietly(outputStream);
    }
}

/* Execute the migration */
let handledContentTypes = 0;
let contentTypeExtensionPoint = Ametys.serviceManager.lookup('org.ametys.cms.contenttype.ContentTypeExtensionPoint');
let contentTypeIds = contentTypeExtensionPoint.getExtensionsIds();
let totalContentTypes = contentTypeIds.size();
let saxFactory = javax.xml.transform.TransformerFactory.newInstance();
let reader = null;
let templates = null;
try
{
    /* Prepare XSL once */; 
    reader = new java.io.StringReader(RICH_TEXT_MIGRATION_XSL_CONTENT);
    let xsl = new javax.xml.transform.stream.StreamSource(reader);
    templates = saxFactory.newTemplates(xsl);
    
    /* Iterate on content types */
    for (let contentTypeId of contentTypeIds)
    {
        let contentType = contentTypeExtensionPoint.getExtension(contentTypeId);
    
        if (!contentType.isAbstract())
        {
            let richTextsInRepeater = __getRichTextsInRepeater(contentType, false);
            if (richTextsInRepeater.size > 0)
            {
                logger.info("Handling contents of type '" + contentTypeId + "'");
                logger.debug("Default workspace");
                __migrateRichTextsDataInAllContentsByContentType(null, richTextsInRepeater, contentTypeId);
                logger.debug("Archives workspace");
                __migrateRichTextsDataInAllContentsByContentType(ARCHIVES_SESSION, richTextsInRepeater, contentTypeId);
            }
        }
        
        handledContentTypes++;
        if (handledContentTypes % 10 == 0)
        {
            logger.info("[" + handledContentTypes + "/" + totalContentTypes + "] Handled content types");
        }
    }
    
    logger.info(handledContentTypes + " handled content types");
}
finally
{
    IOUtils.closeQuietly(reader);
}
