/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
 
const DataHolderHelper = Java.type("org.ametys.plugins.repository.data.holder.impl.DataHolderHelper");

const EMPTY_LONG = java.lang.Long.MIN_VALUE + 1;
const EMPTY_DOUBLE = java.lang.Double.NEGATIVE_INFINITY;
const EMPTY_CALENDAR_TIME_IN_MILLIS = -1855872000000;

/* --------------------------------------------------------------------- */
/*                     MODEL AWARE OBJECTS                               */
/* --------------------------------------------------------------------- */

const LONG_ELEMENT_TYPE_ID = org.ametys.runtime.model.type.ModelItemTypeConstants.LONG_TYPE_ID;
const DOUBLE_ELEMENT_TYPE_ID = org.ametys.runtime.model.type.ModelItemTypeConstants.DOUBLE_TYPE_ID;
const DATE_ELEMENT_TYPE_ID = org.ametys.runtime.model.type.ModelItemTypeConstants.DATE_TYPE_ID;
const DATETIME_ELEMENT_TYPE_ID = org.ametys.runtime.model.type.ModelItemTypeConstants.DATETIME_TYPE_ID;
const GEOCODE_ELEMENT_TYPE_ID = org.ametys.cms.data.type.ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID;

// For model aware data holders, visit the model to find long, double, date and geocode data
// And use dataHolder#setValue to set the values as empty
function __migrateAllEmtpyDataOfModelAwareDataHolder(dataHolder, handle)
{
    try
    {
        let hasChanged = false;

        // Longs
        DataHolderHelper.findEditableItemsByType(dataHolder, LONG_ELEMENT_TYPE_ID).forEach(function(dataPath, value) {
            hasChanged = __migrateEmtpyDataOfModelAwareDataHolder(dataHolder, dataPath, value, EMPTY_LONG, handle) || hasChanged;
        });

        // Doubles
        DataHolderHelper.findEditableItemsByType(dataHolder, DOUBLE_ELEMENT_TYPE_ID).forEach(function(dataPath, value) {
            hasChanged = __migrateEmtpyDataOfModelAwareDataHolder(dataHolder, dataPath, value, EMPTY_DOUBLE, handle) || hasChanged;
        });

        // Dates
        DataHolderHelper.findEditableItemsByType(dataHolder, DATE_ELEMENT_TYPE_ID).forEach(function(dataPath, value) {
            hasChanged = __migrateEmtpyDateDataOfModelAwareDataHolder(dataHolder, dataPath, value, handle) || hasChanged;
        });

        // Date times
        DataHolderHelper.findEditableItemsByType(dataHolder, DATETIME_ELEMENT_TYPE_ID).forEach(function(dataPath, value) {
            hasChanged = __migrateEmtpyDateDataOfModelAwareDataHolder(dataHolder, dataPath, value, handle) || hasChanged;
        });

        // Geocodes
        DataHolderHelper.findEditableItemsByType(dataHolder, GEOCODE_ELEMENT_TYPE_ID).forEach(function(dataPath, value) {
            hasChanged = __migrateEmtpyGeocodeDataOfModelAwareDataHolder(dataHolder, dataPath, value, handle) || hasChanged;
        });

        return hasChanged;
    }
    catch (e)
    {
        logger.error("Unable to migrate empty data of object " + dataHolder, e);
        throw e;
    }
}

function __migrateEmtpyDateDataOfModelAwareDataHolder(dataHolder, dataPath, value, handle)
{
    if (value)
    {
        let dateValue = org.ametys.core.util.DateUtils.asCalendar(value).getTimeInMillis();
        return __migrateEmtpyDataOfModelAwareDataHolder(dataHolder, dataPath, dateValue, EMPTY_CALENDAR_TIME_IN_MILLIS, handle);
    }
    else
    {
        return false;
    }
}

function __migrateEmtpyDataOfModelAwareDataHolder(dataHolder, dataPath, value, emptyValue, handle)
{
    if (value == emptyValue)
    {
        if (handle)
        {
            dataHolder.setValue(dataPath, null);
            if (logger.isDebugEnabled())
            {
                logger.debug("The empty " + dataHolder.getType(dataPath).getId() + " data at path '" + dataPath + "' on object '" + dataHolder + "' has been migrated");
            }
        }
        return true;
    }
    else
    {
        return false;
    }
}

function __migrateEmtpyGeocodeDataOfModelAwareDataHolder(dataHolder, dataPath, value, handle)
{
    if (value && !Array.isArray(value) && (value.getLatitude() == EMPTY_DOUBLE || value.getLongitude() == EMPTY_DOUBLE))
    {
        if (handle)
        {
            dataHolder.setValue(dataPath, null);
            if (logger.isDebugEnabled())
            {
                logger.debug("The empty geocode data at path '" + dataPath + "' on object '" + dataHolder + "' has been migrated");
            }
        }
        return true;
    }
    else
    {
        return false;
    }
}


/* --------------------------------------------------------------------- */
/*                         MODEL LESS OBJECTS                            */
/* --------------------------------------------------------------------- */

const NAMESPACE_PREFIX = org.ametys.plugins.repository.RepositoryConstants.NAMESPACE_PREFIX;
const NAMESPACE_PREFIX_INTERNAL = org.ametys.plugins.repository.RepositoryConstants.NAMESPACE_PREFIX_INTERNAL;

const COMPOSITE_NODETYPE = org.ametys.plugins.repository.RepositoryConstants.COMPOSITE_NODETYPE;
const LONG_REPOSITORY_DATA_TYPE = org.ametys.plugins.repository.data.repositorydata.RepositoryData.LONG_REPOSITORY_DATA_TYPE;
const CALENDAR_REPOSITORY_DATA_TYPE = org.ametys.plugins.repository.data.repositorydata.RepositoryData.CALENDAR_REPOSITORY_DATA_TYPE;

const EMPTY_METADATA_SUFFIX = org.ametys.plugins.repository.data.type.RepositoryModelItemType.EMPTY_METADATA_SUFFIX;

// For model less data holders (and view parameters because of multiple view nodes), use the Reposiory data to look for empty data
// And also use the RepositoryData API to set the values as empty by setting the boolean internal property
function __migrateAllEmptyDataOfRepositoryData(repositoryData)
{
    try
    {
        let hasChanged = false;
        repositoryData.getDataNames().forEach(
            dataName => {
                let dataType = repositoryData.getType(dataName);
                if (dataType == LONG_REPOSITORY_DATA_TYPE)
                {
                    hasChanged = __migrateEmptyDataOfRepositoryData(repositoryData, dataName, "long", repositoryData.getLong(dataName), EMPTY_LONG) || hasChanged;
                }
                else if (dataType == CALENDAR_REPOSITORY_DATA_TYPE)
                {
                    hasChanged = __migrateEmptyDataOfRepositoryData(repositoryData, dataName, "datetime", repositoryData.getDate(dataName).getTimeInMillis(), EMPTY_CALENDAR_TIME_IN_MILLIS) || hasChanged;
                }
                else if (dataType == COMPOSITE_NODETYPE)
                {
                    let groupData = repositoryData.getRepositoryData(dataName);
                    hasChanged = __migrateAllEmptyDataOfRepositoryData(groupData) || hasChanged;
                }
            }
        );
        return hasChanged;
    }
    catch (e)
    {
        logger.error("Unable to migrate empty data of " + repositoryData, e);
        throw e;
    }
}

function __migrateEmptyDataOfRepositoryData(repositoryData, dataName, typeId, value, emptyValue)
{
    if (value == emptyValue)
    {
        repositoryData.removeValue(dataName);
        repositoryData.setValue(dataName + EMPTY_METADATA_SUFFIX, true, NAMESPACE_PREFIX_INTERNAL);
        if (logger.isDebugEnabled())
        {
            logger.debug("The empty " + typeId + " data named '" + dataName + "' on node '" + repositoryData + "' has been migrated");
        }
        return true;
    }
    else
    {
        return false;
    }
}

// CONTENTS AND UNVERSIONED (on default workspace)
logger.info("Start migration of empty values for contents and unversioned (default workspace)");
let contents = Repository.query("//element(*, ametys:content)");
let totalContents = contents.getSize();
let handledContents = 0;
contents.forEach(
    content => {
        
        // Check if there are some empty data to migrate in default workspace
        let hasToMigrate = __migrateAllEmtpyDataOfModelAwareDataHolder(content, false);
        if (!hasToMigrate)
        {
            // If there is no modification in default workspace, check the live one (only if the current version is not the live one)
            let hasLiveVersion = Java.from(content.getAllLabels()).indexOf('Live') >= 0;
            let currentVersionIsLive = Java.from(content.getLabels()).indexOf('Live') >= 0;
            
            if (hasLiveVersion && !currentVersionIsLive)
            {
                content.switchToLabel('Live'); // Swith to live workspace if it exists
                hasToMigrate = __migrateAllEmtpyDataOfModelAwareDataHolder(content, false);
                content.switchToRevision(null); // Reset the content to the current revision before calling the content migration function
            }
        }

        if (hasToMigrate)
        {
            // Launch content migration only if there are empty data to migrate in default or live workspace
            Content.migrate(content, [__migrateAllEmtpyDataOfContent], true, null, false, true);
        }
        
        // UNVERSIONED (on default workspace)
        __migrationAllEmptyDataOfUnversioned(content);
        
        handledContents++;
        if (handledContents % 500 == 0)
        {
            logger.info("[" + handledContents + "/" + totalContents + "] handled contents");
        }
    }
);
logger.info(handledContents + " handled contents and unversioned (default workspace)");

function __migrateAllEmtpyDataOfContent(content)
{
    if (__migrateAllEmtpyDataOfModelAwareDataHolder(content, true))
    {
        logger.info("The empty data of content " + content + " have been migrated");
    }
}

// UNVERSIONED (on live workspace)
if (Repository.liveSession)
{
    logger.info("Start migration of empty values for unversioned (live workspace)");
    contents = Repository.query("//element(*, ametys:content)", true, Repository.liveSession);
    totalContents = contents.getSize();
    handledContents = 0;
    contents.forEach(
        content => {
            __migrationAllEmptyDataOfUnversioned(content);
            
            handledContents++;
            if (handledContents % 500 == 0)
            {
                logger.info("[" + handledContents + "/" + totalContents + "] handled contents");
            }
        }
    );
    logger.info(handledContents + " handled unversioned (live workspace)");
}

function __migrationAllEmptyDataOfUnversioned(content)
{
    let unversionedRepositoryData = new org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData(content.getNode().getNode("ametys-internal:unversioned"));
    if(__migrateAllEmptyDataOfRepositoryData(unversionedRepositoryData))
    {
        logger.info("The empty unversioned data of content " + content + " have been migrated");
        content.saveChanges();
    }
}

logger.info("Migration of empty values in CMS is successfully terminated")
