/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/* Constants */
const CREDENTIALS = new javax.jcr.SimpleCredentials('ametys', []);
const ARCHIVES_SESSION = repository.login(CREDENTIALS, 'archives');
const XPATH_PREFIX = "//element(*, ametys:content)[@ametys-internal:contentType = '";
const XPATH_SUFFIX = "']";
const RICH_TEXT_MIGRATION_XSL_CONTENT =
"<xsl:stylesheet version=\"1.0\" xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\" xmlns=\"http://docbook.org/ns/docbook\" xmlns:docbook=\"http://docbook.org/ns/docbook\">" +
    "<xsl:template match=\"docbook:imagedata[@type='local']\">" +
        "<xsl:call-template name=\"migrate-mediadata\"/>" +
    "</xsl:template>" +
    "<xsl:template match=\"docbook:audiodata[@type='local']\">" +
        "<xsl:call-template name=\"migrate-mediadata\"/>" +
    "</xsl:template>" +
    "<xsl:template match=\"docbook:videodata[@type='local']\">" +
        "<xsl:call-template name=\"migrate-mediadata\"/>" +
    "</xsl:template>" +
    "<xsl:template name=\"migrate-mediadata\">" +
        "<xsl:copy>" +
            "<xsl:copy-of select=\"@*[name() != 'fileref']\"/>" +
            "<xsl:attribute name=\"fileref\">" +
                "<xsl:choose>" +    
                    "<xsl:when test=\"contains(@fileref, ';')\">" +
                        "<xsl:value-of select=\"substring-after(@fileref, ';')\"/>" +
                    "</xsl:when>" +
                    "<xsl:otherwise>" +
                        "<xsl:value-of select=\"@fileref\"/>" +
                    "</xsl:otherwise>" +
                "</xsl:choose>" +
            "</xsl:attribute>" +
        "</xsl:copy>" +
    "</xsl:template>" +    
    "<xsl:template match=\"*\">" + 
        "<xsl:copy>" +
            "<xsl:copy-of select=\"@*\"/>" +
            "<xsl:apply-templates/>" +
        "</xsl:copy>" +
    "</xsl:template>" +
"</xsl:stylesheet>";

/* Imports */
const IOUtils = Java.type('org.apache.commons.io.IOUtils');
const StringUtils = Java.type("org.apache.commons.lang.StringUtils");
const outgoingReferencesExtractor = Ametys.serviceManager.lookup('org.ametys.cms.content.references.OutgoingReferencesExtractor');

/* Methods */
function __getRichTexts(modelItemContainer)
{
    let paths = new Map();
    for (let modelItem of modelItemContainer.getModelItems())
    {
        let typeId = modelItem.getType().getId();
        if (typeId == 'repeater' || typeId == 'composite')
        {
            let richTextPaths = __getRichTexts(modelItem);
            if (richTextPaths.size > 0)
            {
                paths.set(modelItem.getName(), { "type": typeId, "dataPaths": richTextPaths });
            }
        }
        else if (typeId == 'rich-text' && modelItem.isEditable())
        {
            paths.set(modelItem.getName(), { "type": typeId });
        }
    }
    return paths;
}

function __migrateRichTextsDataInAllContentsByContentType(session, richTexts, contentTypeId)
{
    let query = XPATH_PREFIX + contentTypeId + XPATH_SUFFIX;
    let contents = Repository.query(query, true, session);
    let totalContents = contents.getSize();
    let handledContents = 0;
    for (let content of contents)
    {
        if (content instanceof org.ametys.cms.repository.ModifiableContent)
        {
            if (__hasToMigrateContent(content, richTexts))
            {
                Content.migrate(
                    content,
                    [__migrateRichTextsDataInContent.bind(null, richTexts)],
                    true,
                    null,
                    false,
                    session == null
                );
            }
        }
        else
        {
            logger.warn("Unable to migrate rich-texts of the content " + content + " because it is not modifiable. This content is ignored");
        }
        
        handledContents++;
        if (handledContents % 500 == 0)
        {
            logger.info("[" + handledContents + "/" + totalContents + "] handled contents of type '" + contentTypeId + "'");
        }
    }
    
    logger.info(handledContents + " handled contents of type '" + contentTypeId + "'");
}

function __hasToMigrateContent(content, richTexts)
{
    let hasToMigrate = __hasToMigrateDataHolder(richTexts, content, "", "'" + content.getTitle() + "' (" + content.getId() + ")");
    if (!hasToMigrate)
    {
        // If there is no modification in default workspace, check the live one (only if the current version is not the live one)
        let hasLiveVersion = Java.from(content.getAllLabels()).indexOf('Live') >= 0;
        let currentVersionIsLive = Java.from(content.getLabels()).indexOf('Live') >= 0;
        
        if (hasLiveVersion && !currentVersionIsLive)
        {
            content.switchToLabel('Live'); // Swith to live workspace if it exists
            hasToMigrate = __hasToMigrateDataHolder(richTexts, content, "", "'" + content.getTitle() + "' (" + content.getId() + ")");
            content.switchToRevision(null); // Reset the content to the current revision before calling the content migration function
        }
    }
    return hasToMigrate;
}

function __hasToMigrateDataHolder(richTexts, dataHolder, completePath, contentDisplay)
{
    for (let dataPath of richTexts.keys())
    {
        let value = dataHolder.getValue(dataPath, false, null)
        if (value != null)
        {
            let newCompletePath = completePath + dataPath;
            let entryValue = richTexts.get(dataPath);
            let type = entryValue["type"];
            if (type == "repeater")
            {
                for (let repeaterEntry of value.getEntries())
                {
                    let hasToMigrateRepeaterEntry = __hasToMigrateDataHolder(entryValue["dataPaths"], repeaterEntry, newCompletePath + "[" + repeaterEntry.getPosition() + "]/", contentDisplay);
                    if (hasToMigrateRepeaterEntry)
                    {
                        return true;
                    }
                }
            }
            else if (type == "composite")
            {
                let hasToMigrateComposite = __hasToMigrateDataHolder(entryValue["dataPaths"], value, newCompletePath + "/", contentDisplay);
                if (hasToMigrateComposite)
                {
                    return true;
                }
            }
            else if (type == "rich-text")
            {
                if (dataHolder.isMultiple(dataPath))
                {
                    for (let richText of value)
                    {
                        hasToMigrateRichText = __doesContentRichTextContainsLocalMedia(richText);
                        if (hasToMigrateRichText)
                        {
                            return true;
                        }
                    }
                }
                else
                {
                    hasToMigrateRichText = __doesContentRichTextContainsLocalMedia(value);
                    if (hasToMigrateRichText)
                    {
                        return true;
                    }
                }
            }
            else
            {
                logger.warn("Unknown type '" + typeId + "' for data " + newCompletePath + " on content " + contentDisplay);
            }
        }
    }
    
    // No rich text to migrate has been found
    return false;
}

function __doesContentRichTextContainsLocalMedia(richText)
{
    let stream = null;
    try
    {
        stream = richText.getInputStream();
        let contentByteArray = IOUtils.toByteArray(stream);
        return StringUtils.contains(new java.lang.String(contentByteArray, "UTF-8"), "mediaobject")
            && StringUtils.contains(new java.lang.String(contentByteArray, "UTF-8"), "type=\"local\"");
    }
    finally
    {
        if (stream != null)
        {
            IOUtils.closeQuietly(stream);
        }
    }
}

function __migrateRichTextsDataInContent(richTexts, content)
{
    try
    {
        let attachmentsHaveChanged = __copyAttachmentsFromOtherContents(content);
        __migrateRichTextsDataInDataHolder(richTexts, content, "", "'" + content.getTitle() + "' (" + content.getId() + ")");
        
        if (attachmentsHaveChanged)
        {
            // if attachments have changed, it is necessary to update the outgoing references
            let outgoingReferencesByPath = outgoingReferencesExtractor.getOutgoingReferences(content);
            content.setOutgoingReferences(outgoingReferencesByPath);
        }
    }
    catch (e)
    {
        logger.error("An error occurred on content " + content.getId(), e);
        throw e;
    }
}

function __copyAttachmentsFromOtherContents(content)
{
    let hasChanges = false;
    let outgoingReferences = content.getOutgoingReferences();
    for (let dataPath of outgoingReferences.keySet())
    {
        let dataOutgoingReferences = outgoingReferences.get(dataPath);
        if (dataOutgoingReferences.containsKey("local"))
        {
            let dataLocalOutgoingReferences = dataOutgoingReferences.get("local");
            let destRichTextValue = null;
            for (let reference of dataLocalOutgoingReferences)
            {
                if (!reference.includes(content.getId() + "@"))
                {
                    logger.debug("The attribute " + dataPath + " of content '" + content.getTitle() + "' (" + content.getId() + ") has an attachment that references another content."
                                + " The script will first copy the attachment before migrate its content.");
    
                    let sourceAttachment = __getAttachmentFromURI(reference, content, dataPath);
                    
                    if (sourceAttachment != null)
                    {
                        if (destRichTextValue == null)
                        {
                            destRichTextValue = content.getValue(dataPath);
                        }
                        if (content.isMultiple(dataPath))
                        {
                            // If the attribute is mulitple, we should read each value to know wich rich text contains the references
                            // This case should not happen as we don't have multiple rich texts in 4.4 and this script is used for 4.4 -> 4.5 migration
                            logger.warn("The attribute " + dataPath + " of content '" + content.getTitle() + "' (" + content.getId() + ") is multiple. Unable to copy its remote attachments");
                        }
                        else
                        {
                            destRichTextValue.addAttachment(sourceAttachment);
                        }
                    }
                    
                    hasChanges = true;
                }
            }
            
            if (destRichTextValue != null)
            {
                // The #getValue method for the rich text has been called only if it has to change
                content.setValue(dataPath, destRichTextValue);
            }
        }
    }
    
    return hasChanges;
}

function __getAttachmentFromURI(uri, referencingContent, referencingDataPath)
{
    // uri are like <contentId>@<attribute>;<file.ext>
    let i = uri.indexOf('@');
    let j = uri.indexOf(';', i);

    if (i > 0 && j > 0)
    {
        let contentId = uri.substring(0, i);
        let dataPath = uri.substring(i + 1, j);
        let filename = uri.substring(j + 1);
        
        try
        {
            let content = Repository.resolver.resolveById(contentId);
            let value = content.getValue(dataPath);

            if (content.isMultiple(dataPath))
            {
                for (let richText of value)
                {
                    if (richText.hasAttachment(filename))
                    {
                        return richText.getAttachment(filename);
                    }
                }
            }
            else
            {
                if (value.hasAttachment(filename))
                {
                    return value.getAttachment(filename);
                }
            }
        }
        catch (e)
        {
            logger.warn("The attribute " + referencingDataPath + " of content '" + referencingContent.getTitle() + "' (" + referencingContent.getId() + ") references an attachment from the nonexisting content '" + contentId + "'. Unable to copy this remote attachment");
            return null;
        }
    }
    
    // If no attachment has been found
    return null;
}

function __migrateRichTextsDataInDataHolder(richTexts, dataHolder, completePath, contentDisplay)
{
    for (let dataPath of richTexts.keys())
    {
        let value = dataHolder.getValue(dataPath, false, null)
        if (value != null)
        {
            let newCompletePath = completePath + dataPath;
            let entryValue = richTexts.get(dataPath);
            let type = entryValue["type"];
            if (type == "repeater")
            {
                for (let repeaterEntry of value.getEntries())
                {
                    __migrateRichTextsDataInDataHolder(entryValue["dataPaths"], repeaterEntry, newCompletePath + "[" + repeaterEntry.getPosition() + "]/", contentDisplay);
                }
            }
            else if (type == "composite")
            {
                __migrateRichTextsDataInDataHolder(entryValue["dataPaths"], value, newCompletePath + "/", contentDisplay);
            }
            else if (type == "rich-text")
            {
                if (dataHolder.isMultiple(dataPath))
                {
                    for (let richText of value)
                    {
                        __migrateRichTextData(richText);
                    }
                }
                else
                {
                    __migrateRichTextData(value);
                }
                dataHolder.setValue(dataPath, value)
                logger.debug("Rich-text has been migrated for content " + contentDisplay + " on " + newCompletePath);
            }
            else
            {
                logger.warn("Unknown type '" + typeId + "' for data " + newCompletePath + " on content " + contentDisplay);
            }
        }
    }
}

function __migrateRichTextData(richText)
{
    let inputStream = null;
    let outputStream = null;
    try
    {
        let transformer = templates.newTransformer();
        
        inputStream = richText.getInputStream();
        outputStream = richText.getOutputStream();
        
        let source = new javax.xml.transform.stream.StreamSource(inputStream);
        let result = new javax.xml.transform.stream.StreamResult(outputStream);
        
        transformer.transform(source, result);
    }
    finally
    {
        IOUtils.closeQuietly(inputStream);
        IOUtils.closeQuietly(outputStream);
    }
}

/* Execute the migration */
let handledContentTypes = 0;
let contentTypeExtensionPoint = Ametys.serviceManager.lookup('org.ametys.cms.contenttype.ContentTypeExtensionPoint');
let contentTypeIds = contentTypeExtensionPoint.getExtensionsIds();
let totalContentTypes = contentTypeIds.size();
let saxFactory = javax.xml.transform.TransformerFactory.newInstance();
let reader = null;
let templates = null;
try
{
    /* Prepare XSL once */; 
    reader = new java.io.StringReader(RICH_TEXT_MIGRATION_XSL_CONTENT);
    let xsl = new javax.xml.transform.stream.StreamSource(reader);
    templates = saxFactory.newTemplates(xsl);
    
    /* Iterate on content types */
    for (let contentTypeId of contentTypeIds)
    {
        let contentType = contentTypeExtensionPoint.getExtension(contentTypeId);
    
        if (!contentType.isAbstract())
        {
            let richTexts = __getRichTexts(contentType);
            if (richTexts.size > 0)
            {
                logger.info("Handling contents of type '" + contentTypeId + "'");
                logger.debug("Default workspace");
                __migrateRichTextsDataInAllContentsByContentType(null, richTexts, contentTypeId);
                logger.debug("Archives workspace");
                __migrateRichTextsDataInAllContentsByContentType(ARCHIVES_SESSION, richTexts, contentTypeId);
            }
        }
        
        handledContentTypes++;
        if (handledContentTypes % 10 == 0)
        {
            logger.info("[" + handledContentTypes + "/" + totalContentTypes + "] Handled content types");
        }
    }
    
    logger.info(handledContentTypes + " handled content types");
}
finally
{
    IOUtils.closeQuietly(reader);
}
