/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */


/* Execute the migration */
let saveChanges = true;

const IOUtils = Java.type('org.apache.commons.io.IOUtils');
const StringUtils = Java.type("org.apache.commons.lang3.StringUtils");
const contentTypeExtensionPoint = Ametys.serviceManager.lookup('org.ametys.cms.contenttype.ContentTypeExtensionPoint');
const outgoingManager = Ametys.serviceManager.lookup(org.ametys.cms.content.references.OutgoingReferencesExtractor.ROLE);

let attachmentCount = 0;
let attachmentNotFindCount = 0;
let contentCount = 0;
let impactContents = 0;

let contentsToMigrateWithWrongAttachments = _getContentsToMigrateWithWrongAttachments();
let ctypes = _getRichTextByCtype();

for (let contentId in contentsToMigrateWithWrongAttachments)
{
    impactContents++;
    try
    {
        let content = Repository.resolver.resolveById("content://" + contentId);
        let wrongAttachmentsInfo = contentsToMigrateWithWrongAttachments[contentId];
        let mappingOldToNewAttachments = _getMappingOldToNewAttachments(content, wrongAttachmentsInfo, true);
        if (Object.keys(mappingOldToNewAttachments).length > 0)
        {
            if (saveChanges)
            {
                Content.migrate(             
                    content,             
                    [_migrateContent.bind(null, wrongAttachmentsInfo, ctypes)],             
                    true,             
                    null,             
                    false,
                    true
                );   
            }
            else
            {
                _migrateContent(wrongAttachmentsInfo, ctypes, content);            
            }
                    
            contentCount++;            
        }
    }
    catch(e)
    {
        logger.error("An error occurred migrating content '" + contentId + "' : " + e)
    }   
}

logger.info(impactContents + " content(s) with invalid attachments.")
logger.info(attachmentCount + " attachment(s) migrated.")
logger.info(attachmentNotFindCount + " attachment(s) not migrated because we don't find equivalent in history. Probably some broken links.")
logger.info("End of the migration with " + contentCount + " content(s) migrated.")

function _getRichTextByCtype()
{
    let ctypes = {};
    let contentTypeIds = contentTypeExtensionPoint.getExtensionsIds();
    for (let contentTypeId of contentTypeIds)
    {
        let contentType = contentTypeExtensionPoint.getExtension(contentTypeId);
        if (!contentType.isAbstract())
        {
            ctypes[contentTypeId] = _getRichTexts(contentType);
        }
    }
    
    return ctypes;
}

function _getRichTexts(modelItemContainer)
{
    let paths = new Map();
    for (let modelItem of modelItemContainer.getModelItems())
    {
        let typeId = modelItem.getType().getId();
        if (typeId == 'repeater' || typeId == 'composite')
        {
            let richTextPaths = _getRichTexts(modelItem);
            if (richTextPaths.size > 0)
            {
                paths.set(modelItem.getName(), { "type": typeId, "dataPaths": richTextPaths });
            }
        }
        else if (typeId == 'rich-text' && modelItem.isEditable())
        {
            paths.set(modelItem.getName(), { "type": typeId });
        }
    }
    return paths;
}

function _getContentsToMigrateWithWrongAttachments()
{
    let contentsToMigrate = {};
    let nodes = Repository.query("//element(attachment-content, ametys:reference)", false);
    while (nodes.hasNext())
    {
        let attachmentNode = nodes.next();
        let wrongAttachments = _getWrongAttachments(attachmentNode);
    
        if (wrongAttachments.length > 0)
        {
            let contentNode = attachmentNode.getParent().getParent().getParent();
            let history = _getContentAttachementsInHistory(contentNode);

            contentsToMigrate[contentNode.getId()] = {
                "wrongAttachments": wrongAttachments,
                "history": history
            }
       }
    }
    
    return contentsToMigrate;
}

function _getWrongAttachments(attachmentsNode)
{
    let wrongAttachments = [];
    if (attachmentsNode.hasProperty("ametys:reference"))
    {
        for (let val of attachmentsNode.getProperty("ametys:reference").getValues())
        {
            try
            {
                Repository.resolver.resolveById(val.getString());
            }
            catch(e)
            {
                wrongAttachments.push(val.getString());
            }   
        }
    }
    
    return wrongAttachments;
}

function _getContentAttachments(contentNode)
{
    let attachments = [];
    let attachmentNodes = contentNode.getNode("ametys-internal:attachments").getNodes();
    while (attachmentNodes.hasNext())
    {
        let attachmentNode = attachmentNodes.next();
        let jsonAttch = {};
        jsonAttch.title = attachmentNode.getName();
        jsonAttch.id = "resource://" + attachmentNode.getId();
            
        attachments.push(jsonAttch);
    }
    
    return attachments;
}

function _getContentAttachementsInHistory(contentNode)
{
    let oldAttachments = [];
    let oldAttachmentIds = [];
    let frozenNodes = Repository.query("//element(*, nt:frozenNode)[@jcr:frozenUuid = '" + contentNode.getId() + "']/ametys-internal:attachments/*", false);
    while (frozenNodes.hasNext())
    {
        let frozenNode = frozenNodes.next();
        let uuid = "resource://" + frozenNode.getProperty("jcr:frozenUuid").getString();
          
        oldAttachmentIds.push(uuid);
        if (!oldAttachments[uuid])
        {
             oldAttachments[uuid] = [];
        }
        oldAttachments[uuid].push(frozenNode.getName());
    }
    
    return {
        "ids": oldAttachmentIds,
        "objects": oldAttachments
    }
}

function _migrateContent(wrongAttachmentsInfo, ctypes, content)
{
    let richTexts = new Map();
    for (let contentTypeId of content.getTypes())
    {
        ctypes[contentTypeId].forEach(function(value, key) {
          richTexts.set(key, value);
        });
    }
    
    let mappingOldToNewAttachments = _getMappingOldToNewAttachments(content, wrongAttachmentsInfo, false);
        
    _migrateRichTextsDataInDataHolder(richTexts, content, mappingOldToNewAttachments, "", "'" + content.getTitle() + "' (" + content.getId() + ")");
    
    let outgoingReferencesByPath = outgoingManager.getOutgoingReferences(content);
    content.setOutgoingReferences(outgoingReferencesByPath);
}

function _getMappingOldToNewAttachments(content, wrongAttachmentsInfo, increment)
{
    let contentAttachments = _getContentAttachments(content.getNode());
    let wrongAttachments = wrongAttachmentsInfo.wrongAttachments;
    let history = wrongAttachmentsInfo.history;
    let oldAttachments = history.objects;
    let oldAttachmentIds = history.ids;
    
    let mappingOldToNewAttachments = {};
    for (let wrongAttachementId of wrongAttachments)
    {
        if (oldAttachmentIds.indexOf(wrongAttachementId) != -1)
        {
            let oldAttachementNames = oldAttachments[wrongAttachementId];
            for (let attchInfo of contentAttachments)
            {                    
                if (oldAttachementNames.indexOf(attchInfo.title) != -1)
                {
                    mappingOldToNewAttachments[wrongAttachementId] = attchInfo.id;
                }
            }
            if (increment) {attachmentCount++;}
        }
        else
        {
            if (increment) {attachmentNotFindCount++;}
            logger.warn("Can find the equivalent attachment id for old attachment id '" + wrongAttachementId + "' for content id '" + content.getId() + "'");
        }
    }
    
    return mappingOldToNewAttachments;
}

function _migrateRichTextsDataInDataHolder(richTexts, dataHolder, mappingOldToNewAttachments, completePath, contentDisplay)
{
    for (let dataPath of richTexts.keys())
    {
        let value = dataHolder.getValue(dataPath, false, null)
        if (value != null)
        {
            let newCompletePath = completePath + dataPath;
            let entryValue = richTexts.get(dataPath);
            let type = entryValue["type"];
            if (type == "repeater")
            {
                for (let repeaterEntry of value.getEntries())
                {
                    _migrateRichTextsDataInDataHolder(entryValue["dataPaths"], repeaterEntry, mappingOldToNewAttachments, newCompletePath + "[" + repeaterEntry.getPosition() + "]/", contentDisplay);
                }
            }
            else if (type == "composite")
            {
                _migrateRichTextsDataInDataHolder(entryValue["dataPaths"], value, mappingOldToNewAttachments, newCompletePath + "/", contentDisplay);
            }
            else if (type == "rich-text")
            {
                if (dataHolder.isMultiple(dataPath))
                {
                    for (let richText of value)
                    {
                        _migrateRichTextData(richText, mappingOldToNewAttachments);
                    }
                }
                else
                {
                    _migrateRichTextData(value, mappingOldToNewAttachments);
                }
                dataHolder.setValue(dataPath, value)
                logger.debug("Rich-text has been migrated for content " + contentDisplay + " on " + newCompletePath);
            }
            else
            {
                logger.warn("Unknown type '" + type + "' for data " + newCompletePath + " on content " + contentDisplay);
            }
        }
    }
}

function _migrateRichTextData(richText, mappingOldToNewAttachments)
{
    let inputStream = null;
    let outputStream = null;
    try
    {
        inputStream = richText.getInputStream();
        outputStream = richText.getOutputStream();
        
        let contentAsString = IOUtils.toString(inputStream, "UTF-8");
        for (let wrongAttachmentId in mappingOldToNewAttachments)
        {
            contentAsString = contentAsString.replaceAll(wrongAttachmentId, mappingOldToNewAttachments[wrongAttachmentId]);
        }
        
        outputStream.write(StringUtils.getBytes(contentAsString, "UTF-8"));
    }
    finally
    {
        if (inputStream != null)
        {
            IOUtils.closeQuietly(inputStream);
        }
        
        if (outputStream != null)
        {
            IOUtils.closeQuietly(outputStream);
        }
    }
}
