/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

const previousVersionIncompatible = false;
const labelOnNewVersion = null;
const verbose = false;
const synchronizeLive = true;

// Load services
let workflowProvider = Ametys.serviceManager.lookup('org.ametys.plugins.workflow.support.WorkflowProvider');

// Retrieve all the content that have been validated at least once
const query = "//element(*, ametys:content)[@ametys:lastValidationDate]";

let contents = Repository.query(query);
let totalContents = contents.getSize();
let handledContents = 0;

logger.info(`Starting migration of ${totalContents} contents to add validator metadata…`);

let error = 0;

for (let content of contents)
{
    try
    {
        // ignore content that have not been set with a validator to avoid remigrating a node
        // this filtering can not be done with the query as it targets a subnode
        if (!content.getNode().hasNode('ametys:lastValidator'))
        {
            /* Execute the migration */
            let validators = _getValidators(content);
            
            Content.migrate(
                content,
                [_setValidators.bind(null, validators)],
                previousVersionIncompatible,
                labelOnNewVersion,
                verbose,
                synchronizeLive
            );
        }
        
        handledContents++;
        if (handledContents % 500 == 0)
        {
            logger.info(`[${handledContents}/${totalContents}] Handled contents.`);
        }
    }
    catch (e)
    {
        logger.error("An error occurred on content " + content.getId(), e);
        error++;
    }
}
logger.info(`End of migration of ${totalContents} contents to add validator metadata.`);

if (error > 0)
{
    throw new Error(`${error} error${error > 1 ? "s": ""} occurred during content migration. See previous log for more informations`);
}


function _getValidators(content)
{
    let firstValidator = null;
    let lastValidator = null;
    let lastMajorValidator = null;
    
    if (content instanceof org.ametys.cms.repository.WorkflowAwareContent && content instanceof org.ametys.cms.repository.ModifiableContent)
    {
        // get the workflow step from newest to oldest
        let steps = _getWorkflowSteps(content);
        
        let size = steps.size();
        
        // We start by iterating from start of list to retrieve the most recent validation which is the last validator
        // We keep iterating to find the last major validator (first valid step followed by invalid step)
        // Once it is found, we stop the iteration to restart from the other end of the list
        for (let i=0; i < size && lastMajorValidator == null; i++)
        {
            let valid = _isValid(steps.get(i));
            // first case : we haven't found a valid state yet so the first valid step hold the lastValidator step
            if (valid && lastValidator == null)
            {
                lastValidator = _getValidator(content, steps, i);
            }
            // second case : we have found a valid step, now we look for the next invalid step and we know the previous step is valid and hold the last major
            else if (!valid && lastValidator != null)
            {
                lastMajorValidator = _getValidator(content, steps, i-1);
            }
        }
        // If we didn't find a lastValidator, there wont be a first…
        if (lastValidator != null)
        {
            // Now we iterate from the end to retrieve the first
            for (let i=size-1; i>=0 && firstValidator == null; i--)
            {
                if (_isValid(steps.get(i)))
                {
                    firstValidator = _getValidator(content, steps, i);
                }
            }
        }
        
        // The loop to determine the last Major will not get it if it's the first step so we catch it back here
        if (lastMajorValidator == null)
        {
            // even if there is no first, we don't care as there is no last either
            lastMajorValidator = firstValidator;
        }
    }
    return {"first": firstValidator, "last": lastValidator, "lastMajor": lastMajorValidator};
}

function _isValid(step)
{
    let validation = step.getProperty("validation");

    return validation == null ? false : validation;
}

function _getValidator(content, steps, i)
{
    // The validator is stored in the previous step, or is the creator if it's the first in history
    let size = steps.size();
    if (i < size-1)
    {
        return org.ametys.core.user.UserIdentity.stringToUserIdentity(steps.get(i+1).getCaller());
    }
    else
    {
        return content.getCreator();
    }
}

function _getWorkflowSteps(workflowAwareContent)
{
    let workflow = workflowProvider.getAmetysObjectWorkflow(workflowAwareContent);
    let workflowId = workflowAwareContent.getWorkflowId();
    let steps = workflow.getCurrentSteps(workflowId);
    steps.addAll(workflow.getHistorySteps(workflowId));
    return steps;
}

function _setValidators(validators, content)
{
    if (validators.first)
    {
        content.setFirstValidator(validators.first);
    }
    if (validators.last)
    {
        content.setLastValidator(validators.last);
    }
    if (validators.lastMajor)
    {
        content.setLastMajorValidator(validators.lastMajor);
    }
}
