/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/**
 * This class provides a widget to set coordinates (latitude, longitude) by pointing a marker on a GoogleMaps popup window.<br>
 * 
 * This widget overrides the runtime {@link Ametys.form.widget.GeoCode} to initialize the address to center the map using fields on distant ametys objects (ex: referenced content).
 * 
 */
Ext.define('Ametys.cms.form.widget.GeoCode', {
    override : 'Ametys.form.widget.GeoCode',
    
    _getInitialAddressFieldInfo: function (fieldName)
    {
        // Do not log error message if there is no relative field with the given name,
        // the field may be on a referenced ametys object
        var field = this.form.getRelativeField(fieldName, this, true);
        if (field)
        {
            return {field: field};
        }
        else
        {
            // No relative field found, seach for a field on referenced ametys object

            var prefix = this.name.substring(0, this.name.lastIndexOf(this.form.defaultPathSeparator));
            // Handling '..' in field name.
            var me = this;
            Ext.Array.forEach(fieldName.split('/'), function(pathPart) {
                if (pathPart == '..')
                {
                    prefix = prefix.substring(0, prefix.lastIndexOf(me.form.defaultPathSeparator));
                    fieldName = fieldName.substring(3);
                }
            });
            
            var remotePath = '';
            var field = null;
            var relativeFieldPath;
            while (field == null && fieldName.length > 0)
            {
                // The address field is computed from an attribnute of a referenced ametys object
                var index = fieldName.lastIndexOf(this.form.defaultPathSeparator);
                if (index != -1)
                {
                    remotePath = fieldName.substring(index + 1) + (remotePath.length > 0 ? this.form.defaultPathSeparator : '') + remotePath;
                    fieldName = fieldName.substring(0, index);
                    relativeFieldPath = prefix == '' ? this.form.getFieldNamePrefix() + fieldName : (prefix + this.form.defaultPathSeparator + fieldName);
                    field = this.form.getField(relativeFieldPath);
                }
                else
                {
                    fieldName = '';
                }
            }
            
            if (!field)
            {
                var message = "{{i18n plugin.core-ui:PLUGINS_CORE_UI_WIDGET_GEOCODE_UNKNOWN_FIELD}}" + relativeFieldPath;
                this.getLogger().error(message);
                return null;
            }
            else
            {
                return {field: field, remotePath: remotePath};
            }
        }
    },
    
    _computeInitialAddressValues: function (defaultLatLng, defaultZoomLevel)
    {
        var needServerCall = false;
        
        Ametys.data.ServerComm.suspend();
        
        var position = 0;
        for (var i in this._addressFields)
        {
            var field = this._addressFields[i].field;
            
            if (field.getValue())
            {
                var remotePath = this._addressFields[i].remotePath;
                if (!Ext.isEmpty(remotePath))
                {
                    var ametysObjectId = field.getValue();
                    if (Ext.isArray(ametysObjectId))
                    {
                        ametysObjectId = ametysObjectId[0];
                    }
                    
                    this._addressFieldValues[position] = null;
                    
                    Ametys.data.ServerComm.callMethod({
                        role: "org.ametys.plugins.repository.data.holder.impl.DataHolderHelper", 
                        methodName: 'getValue', 
                        parameters: [ametysObjectId, remotePath],
                        callback: {
                            handler: this._computeInitialAddressValuesCb,
                            scope: this,
                            arguments: {
                                position: position,
                                defaultLatLng: defaultLatLng,
                                defaultZoomLevel: defaultZoomLevel
                            }
                        }
                    });
                    
                    needServerCall = true;
                }
                else
                {
                    var value = field.getValue();
                    if (Ext.isArray(value) && value.length > 0)
                    {
                        value = value[0];
                    }
        
                    this._addressFieldValues[position] = value;
                }
                
                position++;
            }
        }
        
        Ametys.data.ServerComm.restart();
        
        if (!needServerCall)
        {
            this._openGoogleMap(defaultLatLng, defaultZoomLevel);
        }
    },

    /**
     * @private
     * Callback function after retrieving remote field value
     * @param {Object} result The JSON result
     * @param {Object} args The callback arguments
     */
    _computeInitialAddressValuesCb: function (result, args)
    {
        var value = '';
        if (result)
        {
            if (Ext.isArray(result) && result.length > 0)
            {
                value = result[0];
            }
            else
            {
                value = result;
            }
        }
        this._addressFieldValues[args.position] = value;
        
        var addressComplete = true;
        for (var i in this._addressFieldValues)
        {
            if (this._addressFieldValues[i] == null)
            {
                addressComplete = false;
                break;
            }
        }
        
        if (addressComplete)
        {
            this._openGoogleMap(args.defaultLatLng, args.defaultZoomLevel);
        }
    }
});