/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/**
 * This helper provides a dialog box to search then select contents
 */
Ext.define('Ametys.cms.uihelper.SelectContentBySearch', {
	singleton: true,
	
	/**
	 * @property {Number} PAGE_SIZE The number of records to display by 'page'
	 * @readonly
	 */
	PAGE_SIZE: 50,
	
	/**
	 * @property {Number} MIN_HEIGHT The minimum height of the dialog box
	 * @readonly
	 */
	MIN_HEIGHT: 300,
	
	/**
	 * @property {Number} MAX_HEIGHT The maximum height of the dialog box
	 * @readonly
	 */
	MAX_HEIGHT: 600,
	
	/**
	 * @property {Boolean} [_multiple=false] True to allow multiple selections. Defaults to false.
	 * @private
	 */
    /**
     * @property {Boolean} [_validated=false] True when the user validate the dialog box
     * @private
     */
	/**
	 * @property {Function} _cbFn The call back function to call after choosing contents or canceling the contents search
	 * @private
	 */
	/**
	 * @property {Boolean} _fullContents True to returns {@link Ametys.cms.content.Content}. If false only the content identifiers will be returned
	 * @private
	 */
	/**
	 * @property {String} _modelId The id of search model to base search on.
	 * @private
	 */
	/**
	 * @property {String} _excludeSubContents True to exclude sub-contents from search
	 * @private
	 */
	/**
	 * @property {Ametys.window.DialogBox} _box The dialog box
	 * @private
	 */
	/**
	 * @property {Object} _contextualParameters The contextual parameters used to get search model and search results.
	 */
	
	/**
	 * Configure and open the search contents dialog box.
     * @param {Object} config The configuration options.
     * @param {Object} config.modelId The id of search model to base search on.
     * @param {Object} [config.searchModelParameters] A map of parameters passed to the search model.
     * @param {String} [config.contentType] The id of content type to restrict the search on. Can be null.
     * @param {String} [config.icon] The full path to icon (16x16 pixels) for the dialog box. Can be null if iconCls is used
     * @param {String} [config.iconCls] One or more space separated CSS classes to be applied to the icon dialog box
     * @param {String} config.title The title of the dialog box.
     * @param {Function} config.callback The callback function invoked when contents are selected. The callback function will received the following parameters:
     * @param {String[]/Ametys.cms.content.Content[]} config.callback.contents The selected contents as a Array of ids or an Array of Ametys.cms.content.Content[]. See #config.fullContents
     * @param {String} config.helpmessage1 The message displayed at the top of the first card
     * @param {String} config.helpmessage2 The message displayed at the top of the second card
     * @param {Boolean} [config.fullContents=false] If `true`, the callback function receives Content objects as argument. If `false`, only content ids are provided.
     * @param {Boolean} [config.multiple=false] `true` to allow selecting multiple contents, `false` otherwise.
     * @param {Boolean} [config.excludeSubContents=false] `true` to exclude sub-contents from search
     * @param {Function} [config.initSearchFn] Optional function to invoked for doing some additional stuff on initialization. Parameters are :
     * @param {Ametys.form.ConfigurableFormPanel} config.initSearchFn.form The search form
     * @param {Object} config.initSearchFn.searchModel The search model configuration
     * @param {String} [config.boxSize] The dialog box size: 'large' for the window width with a 0.8 ratio or 'default' for a width of 600 pixels
	 */
	open: function (config)
	{
		config = config || {};
	    
		this._modelId = config.modelId;
		this._contentType = config.contentType;
		this._multiple = config.multiple || false;
		this._excludeSubContents = config.excludeSubContents ? config.excludeSubContents : false;
		
        this._validated = false;
		this._cbFn = config.callback;
		this._fullContents = config.fullContents || false;
		
		this._createDialogBox(config.title, config.icon, config.iconCls, config.helpmessage1 || '', config.helpmessage2 || '', config.boxSize || 'default');
		
		var restrictedContentTypes = this._contentType != null ? [this._contentType] : null;
		this._contextualParameters = Ext.applyIf(config.searchModelParameters || {}, this.getContextualParameters());
		
		Ametys.data.ServerComm.callMethod({
			role: "org.ametys.cms.search.model.SearchModelHelper",
			methodName: "getSearchModelConfiguration",
			parameters: [this._modelId, restrictedContentTypes, this._contextualParameters],
			callback: {
				handler: this._getSearchModelCb,
			 	scope: this,
			 	arguments: {
					config: config
			 	}
			},
			errorMessage: {
				msg: "{{i18n UITOOL_SEARCH_ERROR}}",
				category: Ext.getClassName(this)
			},
			waitMessage: "{{i18n plugin.core-ui:PLUGINS_CORE_UI_LOADMASK_DEFAULT_MESSAGE}}"
		});
	},
	
    /**
     * @template
     * Get the contextual parameters used to get search model and search results.
     * @return {Object} the contextual parameters
     */
    getContextualParameters: function ()
    {
        return {
            language: this._getCurrentLanguage()
        };
    },
    
    /**
     * Provides the current language to be used.
     * @return {String} The current language
     */
    _getCurrentLanguage: function()
    {
        return Ametys.cms.language.LanguageDAO.getCurrentLanguage();
    },
    
	/**
	 * @private
	 * Callback function after getting model.
	 * Updates the model and grid columns
	 * @param {Object} result The server response as JSON object
	 * @param {Object} params The callback arguments
	 */
	_getSearchModelCb: function (result, params)
	{
        if (this._form instanceof Ametys.form.ConfigurableFormPanel)
        {
            this._form.configure(result['simple-criteria']);
        }
        
        var fields = Ametys.plugins.cms.search.SearchGridHelper.getFieldsFromJson(result.columns);
        Ametys.plugins.cms.search.ContentSearchTool.addDefaultContentFields(fields);
        var columns = Ametys.plugins.cms.search.SearchGridHelper.getColumnsFromJson(result.columns, false);
        
		this.store.setProxy({
			type: 'ametys',
			plugin: result.searchUrlPlugin,
			url: result.searchUrl,
			reader: {
                type: 'json',
                rootProperty: 'contents'
			}
		});
		
        if (result.pageSize != null && result.pageSize > 0)
        {
            this.store.setPageSize(result.pageSize);
        }
		
		// Update model
		this.store.model.replaceFields(fields, true);
		this._grid.reconfigure (this.store, columns);
        this.store.setSorters([]); // sort by score by default
		
		this._form.getForm().reset();
		this._navHandler(0);
		
        if (this._form instanceof Ametys.form.ConfigurableFormPanel)
        {
            this._form.setValues(); // setValues must always be called for configurable form panel in order to complete its initialization
        }
        
        if  (Ext.isFunction(params.config.initSearchFn))
        {
            params.config.initSearchFn(this._form, result);
        }
		
		this._box.show();
	},
	
	/**
	 * Creates the dialog box
	 * @param {String} title The title of the dialog box.
	 * @param {String} icon The full path to icon (16x16 pixels) for the dialog box. Can be null if iconCls is used
     * @param {String} iconCls CSS classes to be applied to the icon dialog box. Can be null ig icon is used.
	 * @param {String} helpMsg1 The message displayed at the top of the first card
	 * @param {String} helpMsg2 The message displayed at the top of the second card
	 * @param {String} boxSize The dialog box size: 'large' for the window width with a 0.8 ratio or 'default' for a width of 600 pixels
	 * @private
	 */
	_createDialogBox: function (title, icon, iconCls, helpMsg1, helpMsg2, boxSize)
	{
		this._form = Ext.create ('Ametys.form.ConfigurableFormPanel', {
			border: false,
			bodyPadding: 10,
			
			flex: 1,
			scrollable: true,
			defaultFieldConfig: {
				labelWidth: 130,
				style: 'margin-right: 30px'
			},
			
            additionalWidgetsConfFromParams: {
                contentType: 'contentType' // some widgets requires the contentType configuration  
            },
              
            showAmetysComments: false
		});
		
		this.store = this._getStore();
		this._grid = Ext.create ('Ext.grid.Panel', {
			flex: 1,
      	  	store: this.store,
				
      	  	scrollable: true,
			    
      	  	columns: [],
			    
      	  	selModel : {
			    mode: this._multiple ? 'MULTI' : 'SINGLE'
			},
			
			viewConfig: {
				loadingText: "{{i18n plugin.cms:UITOOL_SEARCH_WAITING_MESSAGE}}"
			},
			
			dockedItems: [{
					xtype: 'pagingtoolbar',
					store: this.store,
					dock: 'bottom',
					displayInfo: true,
					displayMsg: "{{i18n plugin.cms:UITOOL_CONTENTEDITIONGRID_RESULT_1}}{0}{{i18n plugin.cms:UITOOL_CONTENTEDITIONGRID_RESULT_2}}{1}{{i18n plugin.cms:UITOOL_CONTENTEDITIONGRID_RESULT_3}}{2}",
					emptyMsg: "{{i18n plugin.cms:UITOOL_CONTENTEDITIONGRID_NO_RESULT}}"
			}],
				
			listeners: { 'selectionchange': Ext.bind(this._onSelectionChange, this)}
		});
		
		this._box = Ext.create('Ametys.window.DialogBox', {
			title: title,
			icon: icon,
            iconCls: icon ? null : iconCls,
            
			width: boxSize === "large" ? window.innerWidth * 0.8 : 600,
			scrollable: false,
			layout: 'card',
			activeItem: 0,
			minHeight: this.MIN_HEIGHT,
			maxHeight: this.MAX_HEIGHT,
			freezeHeight: true,
			
			items: [{
			        	 border: false,
                         layout: {
                             type: 'vbox',
                             align: 'stretch' 
                         },
			        	 items : [
			        	          {
			        	        	  xtype: 'component',
			        	        	  cls: 'a-text',
			        	        	  html: helpMsg1
			        	          },
			        	          this._form
			        	 ]
			         },
			         {
			        	 border: false,
			        	 layout: {
                             type: 'vbox',
                             align: 'stretch' 
                         },
			        	 items : [
			        	          {
			        	        	  xtype: 'component',
			        	        	  cls: 'a-text',
			        	        	  html: helpMsg2
			        	          },
			        	          this._grid
			        	 ]
			         }
			],
			
			defaultFocus: this._form,
			closeAction: 'destroy',
			
			referenceHolder: true,
			defaultButton: 'validate',
			
			buttons : [{
	            text: "{{i18n PLUGINS_CMS_HELPER_SELECTCONTENTBYSEARCH_BUTTON_PRECEDE}}", 
	            handler: Ext.bind(this._navHandler, this, [0]),
	            disabled: true
			}, {
				reference: 'next',
				text: "{{i18n PLUGINS_CMS_HELPER_SELECTCONTENTBYSEARCH_BUTTON_NEXT}}", 
	            handler: Ext.bind(this._navHandler, this, [1]),
	            disabled: false
	        }, {
	        	reference: 'validate',
				text: "{{i18n PLUGINS_CMS_HELPER_SELECTCONTENTBYSEARCH_BUTTON_OK}}",
				disabled: true,
				handler: Ext.bind(this._validate, this)
		    }, {
		    	text: "{{i18n PLUGINS_CMS_HELPER_SELECTCONTENTBYSEARCH_BUTTON_CANCEL}}",
		    	handler: Ext.bind(function() {this._box.close();}, this) 
		    }],
            
            listeners: {
                close: Ext.bind(this._onClose, this)
            }
		});
	},
	
	/**
	 * This function is called when navigating into the card layout by 'Next' or 'Precede' buttons
	 * @param {Number} index the index of card layout to show
	 * @private
	 */
	_navHandler: function (index)
	{
		if (index == 1)
		{
			if (!this._form.isValid())
			{
				return;
			}
			
			this._grid.getSelectionModel().clearSelections();
			this.store.loadPage(1, {
				scope: this,
				callback: function(records, operation, success) {
					if (success && records.length > 1)
					{
						this._grid.getSelectionModel().select(records[0]);
						this._grid.getView().focusRow(0);
					}
				}
			});
			
			this._box.defaultButton = 'validate';
		}
		
		this._box.getLayout().setActiveItem(index);
		
		if (index == 0)
		{
			this._form.focus();
			this._box.defaultButton = 'next'; 
		}
		
		this._box.getDockedItems('toolbar[dock="bottom"]')[0].items.get(0).setDisabled(index==0);
		this._box.getDockedItems('toolbar[dock="bottom"]')[0].items.get(1).setDisabled(index==1);
        this._box.getDockedItems('toolbar[dock="bottom"]')[0].items.get(2).setDisabled(index==0);
	},
	
	/**
	 * Function called before loading the store
	 * @param {Ext.data.Store} store The store
	 * @param {Ext.data.operation.Operation} operation The object that will be passed to the Proxy to load the store
	 * @private
	 */
	_onBeforeLoad: function (store, operation)
	{
		operation.setParams(Ext.apply(operation.getParams() || {}, {
			values: this._form.getJsonValues(),
			excludeSubContents: this._excludeSubContents,
			restrictedContentTypes: this._contentType != null ? [this._contentType] : null,
			model: this._modelId,
			contextualParameters: this._contextualParameters
		}));
	},
	
	/**
	 * Listener when the selection has changed in the grid
	 * @param {Ext.selection.Model} sm The selection model
	 * @param { Ext.data.Model[]} selected The selected records
	 * @private
	 */
	_onSelectionChange: function (sm, selected)
	{
		this._box.getDockedItems('toolbar[dock="bottom"]')[0].items.get(2).setDisabled(selected.length == 0);
	},
	
	/**
	 * This function is called when validating the dialog box
	 * @private
	 */
	_validate: function ()
	{
		var contentIds = [];
		
		var selection = this._grid.getSelectionModel().getSelection();
		if (selection.length == 0)
		{
			return;
		}
		
		for (var i=0; i < selection.length; i++)
		{
			contentIds.push(selection[i].get('id'));
		}
		
		if (this._cbFn)
		{
		    if (this._fullContents)
	        {
		        Ametys.cms.content.ContentDAO.getContents(contentIds, Ext.bind(this._callCbFullContents, this));
	        }
		    else
	        {
		        this._multiple ? Ext.Function.defer (this._cbFn, 0, null, [contentIds]) : Ext.Function.defer (this._cbFn, 0, null, [[contentIds[0]]]);
	        }
		}
		
        this._validated = true;
		this._box.close();
	},
	
	/**
	 * @private
	 * Invoke the callback function, providing content objects.
     * @param {Array} contents The content objects.
	 */
	_callCbFullContents: function(contents)
	{
	    this._multiple ? Ext.Function.defer(this._cbFn, 0, null, [contents]) : Ext.Function.defer(this._cbFn, 0, null, [[contents[0]]]);
	},
    
    /**
     * @private
     * Listener when closing the dialog box
     * @param {Boolean} validate true the dialog box was closing after cliking on 'Ok' button
     */
    _onClose: function(validate)
    {
        if (!this._validated && this._cbFn)
        {
            Ext.Function.defer (this._cbFn, 0, null, [null]);
        }
    },
	
	/**
	 * Get the store the result grid should use as its data source
	 * @param {String} storeId The store ID to use, or null to use the default store. 
	 * @return {Ext.data.Store} The store for contents
	 * @private
	 */
	_getStore: function (storeId)
	{
		if (!this._storeModelName)
		{
			this._storeModelName = 'Ametys.cms.uihelper.SelectContentBySearch.Content.Model';
			
			// Copying model, to avoid that the same model huge side effects with model being shared by different stores.
			Ext.define(this._storeModelName, {
			    extend: 'Ext.data.Model'
			});
		}
		
		return Ext.create('Ext.data.Store', {
            model: this._storeModelName,
            
            remoteSort: true,
            proxy: {
                type: 'ametys',
                plugin: 'cms',
                url: 'search/list.json',
                reader: {
                    type: 'json',
                    rootProperty: 'contents'
                }
            },
            
            pageSize: this.PAGE_SIZE,
            sortOnLoad: true,
            
			listeners: {
				'beforeload': {fn: this._onBeforeLoad, scope: this}
			}
        });
	}
});

