/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/**
 * Singleton class to save content modifications
 * @private
 */
Ext.define('Ametys.plugins.cms.content.actions.SaveAction', {
	singleton: true,
	
	/**
	 * Action function to be called by the controller.
	 * Quit the edition without saving
	 * @param {Ametys.ribbon.element.ui.ButtonController} controller The controller calling this function
	 */
	unsave: function (controller)
	{
        var target = controller.getMatchingTargets()[0];
        var contentId = target.getParameters().id;
        var tool = Ametys.tool.ToolsManager.getFocusedTool();
        
        if (tool.isDirty())
        {
    		Ametys.Msg.confirm("{{i18n CONTENT_EDITION_UNSAVE_LABEL}}", 
    				"{{i18n CONTENT_EDITION_UNSAVE_CONFIRM}}", 
    				Ext.bind(this._unsaveConfirm, this, [controller], 1),
    				this
    		);
        }
        else
        {
            this._unsaveConfirm('yes', controller);
        }
	},
	
	/**
	 * Callback function invoked after the 'unsave' confirm box is closed
	 * @param {String} answer Id of the button that was clicked
	 * @param {Ametys.ribbon.element.ui.ButtonController} controller The controller calling the #unsave function
	 * @private
	 */
	_unsaveConfirm: function (answer, controller)
	{
		if (answer == 'yes')
		{
			var target = controller.getMatchingTargets()[0];
			var contentId = target.getParameters().id;
            var tool = Ametys.tool.ToolsManager.getFocusedTool();
			
			// Delete the automatic save.
			Ametys.data.ServerComm.callMethod({
				role: "org.ametys.cms.content.autosave.ContentBackupClientInteraction",
				methodName: "deleteContentBackup",
				parameters: [contentId],
				callback: {
	 				scope: this,
					handler: this._deleteAutoBackupCb
				},
				errorMessage: false
			});
            
            tool.quitEditionMode();
            tool.fireMessagesOnSuccessEdition();

            // Unlock content
            Ametys.cms.content.ContentDAO.unlockContent(contentId, target.getId());
		}
	},
   
	
	/**
	 * Action function to be called by the controller.
	 * Will save the content modifications
	 * @param {Ametys.ribbon.element.ui.ButtonController} controller The controller calling this function
	 */
	save: function (controller)
	{
		this._startSave(controller, false);
	},
	
	/**
	 * Action function to be called by the controller.
	 * Will save the content modifications and quit edition form
	 * @param {Ametys.ribbon.element.ui.ButtonController} controller The controller calling this function
	 */
	saveAndQuit: function (controller)
	{
		this._startSave(controller, true);
	},
	
	/**
	 * Action function to save the content modifications
	 * @param {Ametys.ribbon.element.ui.ButtonController} controller The controller calling this function
	 * @param quit true to quit edition form after saving modifications
	 * @private
	 */
	_startSave: function (controller, quit)
	{
		var form = Ametys.form.SaveHelper.getForm(controller.getMatchingTargets());
		if (form != null)
		{
			var target = controller.getMatchingTargets()[0];
			var contentId = target.getParameters().id;
			
			var tool = Ametys.tool.ToolsManager.getFocusedTool();
			
			// Validate form
			var invalidFields = Ext.Array.merge(form.getInvalidFields(), form.getInvalidRepeaters());
			if (invalidFields.length > 0)
			{
				Ametys.Msg.show({
					   title: "{{i18n PLUGINS_CMS_SAVE_ACTION_SAVE}}", //{{i18n CONTENT_EDITION_SAVE_LABEL}}",
					   msg: "{{i18n PLUGINS_CMS_SAVE_ACTION_INVALIDFIELDS}}" + invalidFields.join(', '), //"{{i18n CONTENT_EDITION_SAVE_INVALID_FIELD}}",
					   buttons: Ext.Msg.OK,
					   icon: Ext.MessageBox.ERROR
				});
				return;
			}
			
			var warnedFields = form.getWarnedFields();
			if (!Ext.Object.isEmpty(warnedFields))
			{
			    Ametys.Msg.show({
			        title: "{{i18n PLUGINS_CMS_SAVE_ACTION_SAVE}}",
			        msg: "{{i18n PLUGINS_CMS_SAVE_ACTION_WARNED_FIELDS_START}}" + Ext.Object.getKeys(warnedFields).join(', ') + '.' + "{{i18n PLUGINS_CMS_SAVE_ACTION_WARNED_FIELDS_END}}",
			        buttons: Ext.Msg.YESNO,
			        icon: Ext.Msg.QUESTION,
			        modal: true,
			        fn: Ext.bind (this._bypassWarnedFields, this, [tool, form, contentId, quit], 1),
			        scope: this
			    });
			    return;
			}
			
			this._doSave (tool, form, contentId, quit);
		}
	},
	
	/**
	 * @private
	 * Save the content
	 * @param {Ametys.tool.Tool} tool The content tool
	 * @param {Ametys.form.ConfigurableFormPanel} form The form panel
	 * @param {String} contentId The content id
	 * @param {Boolean} quit True to quit edition after saving
	 */
	_doSave: function (tool, form, contentId, quit)
	{
        // Disarm the autosave timer function.
        tool.clearAutosaveTimeout();
        
		var params = {};
		
		params.values = form.getJsonValues();
		params.comments = form.getComments();
		params.contentId = contentId;
		params.unlock = quit;
        params.editWorkflowActionId = tool._workflowAction;
		params.viewName = tool._viewName;
        params.fallbackViewName = tool._fallbackViewName;
		
        Ametys.cms.content.ContentDAO.editContent(params, Ext.bind(this._saveCb, this, [quit, form, tool], true), this, false, null, null, tool.getContentPanel());
		
		tool.fireMessageBeforeSaving();
	},
	
	/**
	 * @private
	 * This function is called after validated confirm dialog box
	 * @param {String} answer The button answer ('yes' or 'no')
	 * @param {Ametys.tool.Tool} tool The content tool
	 * @param {Ametys.form.ConfigurableFormPanel} form The form panel
	 * @param {String} contentId The content id
	 * @param {Boolean} quit True to quit edition after saving
	 */
	_bypassWarnedFields : function (answer, tool, form, contentId, quit)
	{
		if (answer == 'yes')
		{
			this._doSave (tool, form, contentId, quit);
		}
	},
	
	/**
	 * Callback function called after #_doSave is processed.
	 * Fires {@link Ametys.message.Message#MODIFIED} message for edited contents
     * @param {boolean} success If the save was successful
     * @param {String} contentId The content id
     * @param {String[]} workflowErrors Errors messages related to the workflow
     * @param {Object} fieldsInError The list of fields in error associated with their error message
     * @param {Boolean} quit True to quit edition after saving
     * @param {Ametys.form.ConfigurableFormPanel} form The form panel
     * @param {Ametys.tool.Tool} tool The content tool
	 * @private
	 */
	_saveCb: function (success, contentId, workflowErrors, fieldsInError, quit, form, tool)
	{
        // Re-arm the autosave timer function.
	    tool.armAutosaveTimeout();
	    
        if (!success && workflowErrors == null && fieldsInError == null)
        {
            tool.fireMessageOnFailEdition();
            return;
        }
        
        if (!success && !Ext.Object.isEmpty(fieldsInError))
        {
            form.markFieldsInvalid (fieldsInError);
            return;
        }

        if (!success)
        {
            return;
        }

        // If the current content has indexing references, send a message.
        if (tool.hasIndexingReferences())
        {
            Ametys.Msg.show({
                title: "{{i18n PLUGINS_CMS_SAVE_ACTION_CONTENT_MODIFICATION}}",
                msg: "{{i18n PLUGINS_CMS_SAVE_ACTION_CONTENT_HAS_INDEXING_REFERENCES}}",
                buttons: Ext.Msg.OK,
                icon: Ext.MessageBox.INFO
            });
        }
		
		// Delete the automatic save.
		Ametys.data.ServerComm.callMethod({
			role: "org.ametys.cms.content.autosave.ContentBackupClientInteraction",
			methodName: "deleteContentBackup",
			parameters: [contentId],
			callback: {
 				scope: this,
				handler: this._deleteAutoBackupCb
			},
			errorMessage: false
		});
		
		if (quit)
		{
			tool.quitEditionMode ();
		}
		
		tool.setDirty(false);
		tool.fireMessagesOnSuccessEdition();
	},
	
	/**
	 * Callback invoked when an automatic save is deleted.
     * @param {Object} response The XML response provided by the {@link Ametys.data.ServerComm}
     * @param {Object} params The callback parameters passed to the {@link Ametys.data.ServerComm#send} method
     * @private
	 */
    _deleteAutoBackupCb: function(response, params)
    {
        // Ignore the callback.
    }
});