/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/**
 * This class controls a ribbon button representing the alerts of a content
 * @private
 */
Ext.define('Ametys.plugins.cms.content.controller.AlertsController', {
	extend: 'Ametys.ribbon.element.ui.ButtonController',
	
	/**
	 * @property {String[]} [_currentAlertEnabledContentsId=[]] List of identifiers of contents with active alert
	 * @private
	 */
	/**
	 * @property {String[]} [_currentAlertDisaabledContentsId=[]] List of identifiers of contents with no alert
	 * @private
	 */
	/**
	 * @property {String[]} [_allRightContentsId=[]] List of identifiers of all right contents
	 * @private
	 */
	
	/**
	 * @property {Number} _validationAlertDelay Delay to alert for content awaiting validation
	 * @private
	 */
	/**
	 * @property {Number} _unmodifiedAlertDelay Delay to alert for unmodified content
	 * @private
	 */
	
	constructor: function(config)
	{
		this.callParent(arguments);
		
		this._currentAlertEnabledContentsId = [];
		this._currentAlertDisaabledContentsId = [];
		this._allRightContentsId = [];
		
		this._validationAlertDelay = 0;
		this._unmodifiedAlertDelay = 0;
		
		Ametys.message.MessageBus.on(Ametys.message.Message.MODIFIED, this._onModified, this);
	},

	/**
	 * Listener when the content has been modified
	 * Will update the state of the buttons effectively upon the current selection.
	 * @param {Ametys.message.Message} message The modified message.
	 * @protected
	 */
	_onModified: function (message)
	{
        if (this.updateTargetsInCurrentSelectionTargets(message))
        {
            this.refresh();
        }
	},
	
	updateState: function()
	{
		this._getAlertsInformations(this.getMatchingTargets());
	},
	
	/**
	 * Get the alert state of given targets
	 * @param targets The content targets
	 * @private
	 */
	_getAlertsInformations: function (targets)
	{
		var contentIds = [];
		for (var i=0; i < targets.length; i++)
		{
			contentIds.push(targets[i].getParameters().id);
		}

		this.disable();
		this.serverCall ('getAlertsInformations', [contentIds], this._getAlertsInformationsCb, { errorMessage: true, refreshing: true });
	},
	
	/**
	 * @private
	 * Callback function called after retrieving the 'alert' state of content targets
	 * @param params The JSON result 
	 */
	_getAlertsInformationsCb: function (params)
	{
		this._validationAlertDelay = params['validation-alert-delay'];
	    this._unmodifiedAlertDelay = params['unmodified-alert-delay'];
	    
		this._currentAlertEnabledContentsId = [];
		this._currentAlertDisabledContentsId = [];
		this._currentNotUnlockableContentsId = [];
		this._allRightContentsId = [];
		
		var allRightContents = params["all-right-contents"];
		if (allRightContents.length > 0)
		{
			for (var i=0; i < allRightContents.length; i++)
			{
				this._allRightContentsId.push(allRightContents[i].id);
			}
		}
		
		// Update tooltip description
		this._updateTooltipDescription("", params);
		
		// Toggle element if at least one content has an active alert
		this.toggle (this._currentAlertEnabledContentsId.length > 0);
		
		this._allRightContentsId.length == 0 ? this.disable() : this.enable();
	},
	
	/**
	 * @private
	 * Update the tooltip description according to 'alert' state of the current selection
	 * @param {String} description The initial description
	 * @param {Object} params The JSON result object
	 */
	_updateTooltipDescription: function (description, params)
	{
		var alertEnabledContents = params["alert-enabled-contents"];
		if (alertEnabledContents.length > 0)
		{
			if (description != "")
			{
				description += "<br/><br/>";
			}
			
			description += this.getInitialConfig("alerts-enabled-start-description");
			for (var i=0; i < alertEnabledContents.length; i++)
			{
				if (i != 0) 
				{
					description += ", ";
				}
				description += alertEnabledContents[i].description;
				this._currentAlertEnabledContentsId.push(alertEnabledContents[i].id);
			}
			description += this.getInitialConfig("alerts-enabled-end-description");
		}
		
		var alertDisabledContents = params["alert-disabled-contents"];
		if (alertDisabledContents.length > 0)
		{
			if (description != "")
			{
				description += "<br/><br/>";
			}
			
			description += this.getInitialConfig("alerts-disabled-start-description");
			for (var i=0; i < alertDisabledContents.length; i++)
			{
				if (i != 0) 
				{
					description += ", ";
				}
				description += alertDisabledContents[i].description;
				this._currentAlertDisabledContentsId.push(alertDisabledContents[i].id);
			}
			description += this.getInitialConfig("alerts-disabled-end-description");
		}
		
        var notUnlockableContents = params["not-unlockable-contents"];
        if (notUnlockableContents.length > 0)
        {
            if (description != "")
            {
                description += "<br/><br/>";
            }
            
            description += this.getInitialConfig("not-unlockable-start-description");
            for (var i=0; i < notUnlockableContents.length; i++)
            {
                if (i != 0) 
                {
                    description += ", ";
                }
                description += notUnlockableContents[i].description;
                this._currentNotUnlockableContentsId.push(notUnlockableContents[i].id);
            }
            description += this.getInitialConfig("not-unlockable-end-description");
        }
		
		this.setDescription (description);
	}
	
});
