/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/**
 * This tool does display the properties of the current selected content
 * @private
 */
Ext.define('Ametys.plugins.cms.content.tool.PropertiesTool', {
	extend: "Ametys.tool.SelectionTool",
	
	/**
	 * @cfg {String} [metadata-set-name="details"] The name of metadata set to use.
	 */
	/**
	 * @property {String} _viewName The view to display. See #cfg-metadata-set-name.
	 */
	
	/**
	 * @property {String} [_contentPluginName=cms] The name of the plugin used to get the content properties.
	 */
	
	/**
	 * @property {String} [_contentUrl=content/properties/view.xml] The url used to get the content properties. 
	 */
	
	statics: {
		
		/**
		 * Show/Hide types hierarchy
		 */
		showHideSubTypes: function ()
		{
			Ext.select('.details-view-card ul.subtypes').each (function (ul) {
				Ext.get(ul).toggleCls('hidden');
			});
		},

		/**
		 * Show/Hide mixins hierarchy
		 */
		showHideMixins: function ()
		{
			Ext.select('.details-view-card ul.mixins').each (function (ul) {
				Ext.get(ul).toggleCls('hidden');
			});
		},
		
		/**
         * Copy value in clipboard
         * @param {Object} button The copy button.
         * @param {String} value The value to copy.
         */
        copyToClipboard: function (button, value)
        {
            var copyText = Ext.getElementById("input-text-clipboard");
            copyText.value = value;
            copyText.select();
            
            var spanIcon = Ext.get(button).child("span");
            spanIcon.removeCls("ametysicon-document28");
            if (document.execCommand("copy"))
            {
                spanIcon.addCls("ametysicon-sign-raw-check");
            }
            else
            {
                spanIcon.addCls("ametysicon-sign-raw-cross");
            }
            
            setTimeout(function() {
                spanIcon.removeCls("ametysicon-sign-raw-check ametysicon-sign-raw-cross");
                spanIcon.addCls("ametysicon-document28");
            }, 1000); 
        }
	},
	
	constructor: function(config)
	{
		this.callParent(arguments);
		this._viewName = this.getInitialConfig('view-name') || 'details';
		
		Ametys.message.MessageBus.on(Ametys.message.Message.MODIFIED, this._onContentChanged, this);
		Ametys.message.MessageBus.on(Ametys.message.Message.WORKFLOW_CHANGED, this._onContentChanged, this);
		Ametys.message.MessageBus.on(Ametys.message.Message.DELETED, this._onContentDeleted, this);
	},
	
	createPanel: function ()
	{
        return Ext.create('Ext.panel.Panel', {
            scrollable: false, 
            border: false,
            layout: 'card',
            activeItem: 0,
            
            items: [{
                    xtype: 'component',
                    cls: 'a-panel-text-empty',
                    border: false,
                    html: ''
                }, 
                {
		            scrollable: true, 
		            border: false,
		            html : '',
		            id : 'content-details-view',
		            cls : 'details-view'
		        }
            ]
        });
	},
	
	setParams: function (params)
	{
		this._viewName = params['viewName'] || 'details';
		this._contentPluginName = params['contentPluginName'] || 'cms';
		this._contentUrl = params['contentUrl'] || 'content/properties/view.xml';
	    
		this.callParent(arguments);
	},
	
	refresh: function (manual)
	{
		var contentTarget = this.getCurrentSelectionTargets()[0];
		if (contentTarget != undefined)
		{
    		this.showRefreshing();
    		
    		this._contentId = contentTarget.getParameters().id;
    		
    		var params = {contentId: this._contentId, viewName: this._viewName};
    		
    		Ametys.data.ServerComm.send({
    			plugin: this._contentPluginName,
    			url: this._contentUrl,
    			parameters: params, 
    			priority: Ametys.data.ServerComm.PRIORITY_MAJOR, 
    			callback: {
    				handler: this._updateContentProperties,
    				scope: this,
    				arguments: [this._contentId]
    			},
    			responseType: 'xml2text'
    		});
        }
	},
	
	/**
	 * @private
	 * Callback function called after #refresh to update the content's properties
	 * @param {HTMLElement} response The XML document.
	 * @param {Array} args The callback arguments.
	 */
	_updateContentProperties: function (response, args)
	{
		var contentId = args[0];
		
		if (this.getCurrentSelectionTargets() == null || this.getCurrentSelectionTargets().length == 0 ||this.getCurrentSelectionTargets()[0].getParameters().id != contentId)
		{
			// too late => discard (another content has been selected)
			return;
		}
		
		if (Ametys.data.ServerComm.handleBadResponse("{{i18n UITOOL_DETAILS_ERROR}} '" + contentId + "'", response, 'Ametys.plugins.cms.content.tool.PropertiesTool'))
        {
            return;
        }
		
        
		this.getContentPanel().items.get(1).update(Ext.dom.Query.selectValue("", response));
        this.getContentPanel().getLayout().setActiveItem(1);
		this.showRefreshed();
	},
    
	setNoSelectionMatchState: function (message)
	{
        this.callParent(arguments);
        
        var panel = this.getContentPanel().items.get(0);
        panel.update(message);
        this.getContentPanel().getLayout().setActiveItem(0);
        this._contentId = null;
	},
	
	/**
	 * Listener on {@link Ametys.message.Message#WORKFLOW_CHANGED}, {@link Ametys.message.Message#MODIFIED} and {@link Ametys.message.Message#DELETED} messages. If the current
	 * content is concerned, the tool will be out-of-date.
	 * 
	 * @param {Ametys.message.Message} message The bus message.
	 * @protected
	 */
	_onContentChanged: function (message)
	{
		if (this.getTargetsInCurrentSelectionTargets(message).length > 0)
		{
			this.showOutOfDate();
		}
	},
    
    _onContentDeleted: function (message)
    {
        if (this.getTargetsInCurrentSelectionTargets(message).length > 0)
        {
            this.setNoSelectionMatchState();
        }
    }
	
});