/*
 *  Copyright 2024 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/**
 * This tool displays the thread of an Ametys object
 * @private
 */
Ext.define('Ametys.plugins.cms.threads.AmetysObjectThreadTool', {
    extend: "Ametys.tool.SelectionTool",
    
    /**
     * @private
     * @property {Ametys.timeline.Timeline} _timeline the timeline of comments
     */
    
    /**
     * @private
     * @property {Ametys.timeline.Timeline} _objectId the id of the selected object
     */
    _objectId: null,
    
    /**
     * @private
     * @property {Ext.Template} _contentHintTpl The template used for hint description
     */
    _contentHintTpl: new Ext.Template("{{i18n PLUGINS_CMS_UITOOL_AMETYS_OBJECT_THREAD_CONTENT_HINT}}"),
    
    /**
     * @private
     * @property {Ext.Template} _mentionedUserTpl The template used for mentiond users in comment content
     */
    _mentionedUserTpl: new Ext.Template("<span class='mention {cls}' data-qtip='{userTooltip}'>{username}</span>"),
    
    constructor: function()
    {
        this.callParent(arguments);
        
        Ametys.message.MessageBus.on(Ametys.message.Message.DELETED, this._onObjectDeleted, this);
    },
    
    createPanel: function()
    {
        var me = this;
        this._timeline = Ext.create('Ametys.timeline.Timeline', {
            scrollable: true,
            
            dockedItems: [{
                xtype: 'component',
                dock: 'top',
                ui: 'tool-hintmessage',
                itemId: 'object-info',
                html: ''
            },
            {
                xtype: 'toolbar',
                dock: 'top',
                layout: {
                    type: 'hbox',
                    align: 'stretch'
                },
                items: [
                    {
                        // Comment input
                        xtype: 'ametystextarea.withmentions',
                        cls: 'ametys',
                        flex: 1,
                        itemId: 'comment-input',
                        emptyText: "{{i18n PLUGINS_CMS_UITOOL_AMETYS_OBJECT_THREAD_INPUT_EMPTY}}",
                        msgTarget: 'qtip',
                        listeners: {
                            change: this._onFieldChange,
                            focus: this._onFieldFocus,
                            focusleave: this._onFieldFocusleave,
                            scope: this
                        }
                    },
                    {
                        xtype: 'container',
                        layout: 'vbox',
                        items: [
                            {
                                // Help
                                xtype: 'component',
                                html: '',
                                cls: 'ametys-description',
                                flex: 1,
                                listeners: {
                                   afterrender: function(me)
                                   {
                                       // Register the new tip with an element's ID
                                       Ext.tip.QuickTipManager.register({
                                           target: me.getId(), // Target button's ID
                                           text  : "{{i18n PLUGINS_CMS_UITOOL_AMETYS_OBJECT_THREAD_INPUT_DESC}}" // Tip content  
                                       });
                     
                                   },
                                   destroy: function(me)
                                   {
                                       Ext.tip.QuickTipManager.unregister(me.getId());
                                   }
                               }
                            },
                            {
                                // Add button
                                xtype: 'button',
                                itemId: 'comment-send-button',
                                tooltip: "{{i18n PLUGINS_CMS_UITOOL_AMETYS_OBJECT_THREAD_SEND_COMMENT_BTN}}",
                                handler: Ext.bind (this._sendComment, me),
                                iconCls: 'a-btn-glyph ametysicon-desktop-paper-plane-light size-16',
                                cls: 'a-btn-light',
                                disabled: true,
                                flex: 1,
                                margin: '0 0 0 0'
                            }
                        ]
                    }
                ]
            }
            ],
            
            timelineItemHTML: ['<div class="timeline-item {additionalCls}">',
                                   '<div class="profile-img-wrap">',
                                       '<img src="{profileImg}" alt="">',
                                       '<div>{hour}</div>',
                                   '</div>',
                                   '<div class="contents-wrap">',
                                       '<span class="vertical-line"></span>',
                                       '<div class="text">',
                                           '<strong data-qtip="{userTooltip}">{username}</strong></br>',
                                           '{text}',
                                       '</div>',
                                   '</div>',
                               '</div>'],
        });
        
        return Ext.create('Ext.panel.Panel', {
            cls: 'ao-thread-tool',
            scrollable: false, 
            border: false,
            layout: 'card',
            activeItem: 0,
            
            items: [{
                    xtype: 'component',
                    cls: 'a-panel-text-empty',
                    border: false,
                    html: ''
                }, 
                this._timeline
            ]
        });
    },
    
    refresh: function ()
    {
        this.showRefreshing();
        
        var objectTarget = this.getCurrentSelectionTargets()[0];
        this._objectId = objectTarget.getParameters().id;
                
        this._timeline.getComponent("object-info").update(this._contentHintTpl.applyTemplate([objectTarget.getParameters().title]));
        
        var comment = this._timeline.queryById("comment-input");
        comment.reset();
        
        this.serverCall ('getThread', [this._objectId], this._getCommentsCb, { errorMessage: true });
    },
    
    /**
     * Clear the filter search
     * @param {Ext.Button} btn The button
     * @private
     */
    _sendComment: function()
    {
        var comment = this._timeline.queryById("comment-input");
        var commentValue = this._escapeCommentText(comment.getValue());

        this.serverCall ('addComment', [this._objectId, commentValue], this._addCommentCb, { errorMessage: true, refreshing: true });
    },
    
    _addCommentCb: function()
    {
        var comment = this._timeline.queryById("comment-input");
        comment.reset();
        
        this.showOutOfDate();
    },
    
    /**
     * @private
     * Callback function called after retrieving the 'alert' state of content targets
     * @param params The JSON result 
     */
    _getCommentsCb: function (comments)
    {
        var data = [];
        
        for (var i = 0; i < comments.length; i++)
        {
            var d = this._convertComment2Timeline(comments[i]);
            data.push(d);
        }
        
        this._timeline.getStore().loadData(data);
        this.getContentPanel().getLayout().setActiveItem(1);
        this.showRefreshed();
    },
    
    _convertComment2Timeline: function (comment)
    {
        var additionalCls;
        if (this._isCurrentUser(comment.author))
        {
            additionalCls = 'current-user';
        }
        
        var username = comment.author.fullname || "{{i18n plugin.core:PLUGINS_CORE_USERS_UNKNOWN_USER}}";
        var userTooltip = this._getUserTooltip(comment.author);
        
        var text = comment.mentions ? this._convertCommentContentMentions(comment) : comment.content;
        
        return {
            date: comment["creation-date"],
            additionalCls: additionalCls, 
            username: username,
            userTooltip: userTooltip,
            profileImg: Ametys.getPluginDirectPrefix('core-ui') + '/user/' + comment.author.populationId + '/' + comment.author.login + '/image_46',
            text: text
        }
    },
    
    _convertCommentContentMentions: function (comment)
    {
        var text = comment.content;
        
        for (var i = 0; i < comment.mentions.length; i++)
        {
            var mentionedUser = comment.mentions[i];
            
            var username = mentionedUser.fullname ? mentionedUser.fullname : "{{i18n plugin.core:PLUGINS_CORE_USERS_UNKNOWN_USER}}";
            var userTooltip = this._getUserTooltip(mentionedUser);
            var cls = this._isCurrentUser(mentionedUser) ? 'current-user' : undefined;
            
            var replaceMentionBy = this._mentionedUserTpl.applyTemplate( {
                userTooltip: userTooltip,
                username: username,
                cls: cls
            });
            
            var login = mentionedUser.login;
            var populationId = mentionedUser.populationId;
            var mentionValue = "@(" + login + "#" + populationId + ")";
            
            text = text.replaceAll(mentionValue, replaceMentionBy);
        }
        
        return text;
    },
    
    _isCurrentUser: function (user)
    {
        var currentUser = Ametys.getAppParameter('user');
        return currentUser.login === user.login && currentUser.population === user.populationId;
    },
    
    _getUserTooltip: function(user)
    {
        var login = user.login;
        var populationId = user.populationId;
        var username = user.fullname || "{{i18n plugin.core:PLUGINS_CORE_USERS_UNKNOWN_USER}}";
        
        var populationLabel = user.populationLabel || populationId;
        return Ametys.plugins.core.users.UsersDAO.renderUser(login, populationLabel, username);
    },
    
    _onFieldChange: function(field, newValue)
    {
        var sendCommentButton = this._timeline.queryById("comment-send-button");
        var disable = Ext.isEmpty(newValue);
        sendCommentButton.setDisabled(disable);
    },
    
    _onFieldFocus: function(field)
    {
        field.addCls("a-field-focus");
    },
    
    _onFieldFocusleave: function(field)
    {
        field.removeCls("a-field-focus");
    },
    
    _escapeCommentText: function(text)
    {
        return text.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/\n/g, '<br/>');
    },
    
    setNoSelectionMatchState: function (message)
    {
        this.callParent(arguments);
        
        var panel = this.getContentPanel().items.get(0);
        panel.update(message);
        this.getContentPanel().getLayout().setActiveItem(0);
        
        this._objectId = null;
    },

    _onObjectDeleted: function (message)
    {
        if (this.getTargetsInCurrentSelectionTargets(message).length > 0)
        {
            this.setNoSelectionMatchState();
        }
    }
});