/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
 
 AmetysFront.Comments = {
    
    /** Regexp for comments' data such as %%author-name%% */
    COMMENT_VALUE_REGEXP: /%%(\S+?)%%/,
    
    /**
     * Load and display the comments
     * @param {String} uniqueId the unique id
     * @param {String} contentId the unique id
     * @param {Boolean} sortCommentsAsc set to true to sort comments in ascending order
     * @param {Boolean} sortSubCommentsAsc set to true to sort subcomments in ascending order
     * @param {Function} [callback] the callback function to invoked after loading comments
     */
    loadComments: function (uniqueId, contentId, sortCommentsAsc, sortSubCommentsAsc, callback)
    {
        AmetysFront.Event.fire("comments-loading", {contentId: contentId});
        
        AmetysFront.ServerComm.callMethod({
            role: 'org.ametys.cms.repository.comment.CommentsDAO',
            methodName: 'getComments',
            parameters: [contentId, AmetysFront.getAppParameters()],
            callback: {
                handler: this._loadCommentsCb,
                scope: this,
                arguments: {
                    uniqueId: uniqueId,
                    contentId: contentId,
                    sortCommentsAsc: sortCommentsAsc,
                    sortSubCommentsAsc: sortSubCommentsAsc,
                    callback: callback
                }
            }
        });
    },
    
    _loadCommentsCb: function (comments, args)
    {
        let uniqueId = args.uniqueId,
            contentId = args.contentId;
            
        let templateForComment = $j('#content-comment-template-' + uniqueId);
        let templateForSubcomment = $j('#content-subcomment-template-' + uniqueId);
        
        const compareComments = (c1, c2, asc) => {
          let date1 = new Date(c1['creation-date']).getTime();
          let date2 = new Date(c2['creation-date']).getTime();
        
          if (date1 < date2) {
            return -1 * (asc ? 1 : -1);
          } else if (date1 > date2) {
            return 1 * (asc ? 1 : -1);
          } else {
            return 0;
          }
        };
        
        var me = this;
        if (templateForComment.length)
        {
            // Sort comments
            comments = $j(comments).sort((c1, c2) => { return compareComments(c1, c2, args.sortCommentsAsc)} );

            $j(comments).each(function (i, comment){
            
                let $commentEl = me._insertComment(templateForComment, comment, $j('[data-ametys-comments="' + contentId + '"]'), true);
                
                var subcomments = comment['sub-comments'] || [];
               
                // Sort sub comments
                subcomments = $j(subcomments).sort((c1, c2) => { return compareComments(c1, c2, args.sortSubCommentsAsc)} );
               
                $j(subcomments).each(function (j, subcomment){
                    me._insertComment(templateForSubcomment, subcomment, $commentEl.find("*[data-ametys-comment-subcomments]"), true);
                });
            });
            
            // Update nb comments
            this._updateNbComments(contentId, comments.length);
            
            if (typeof(args.callback) == 'function')
            {
                args.callback(comments, contentId);
            }
        }
        
	    AmetysFront.Event.fire("comments-loaded", {contentId: contentId});
    },
    
    /**
     * Insert a new comment
     * @param {String} uniqueId the unique id
     * @param {String} contentId The content id
     * @param {Object} comment the comment or subcomment to insert
     * @param {Boolean} [ascendingOrder] 'false' to insert new comment at first position, otherwise it will be inserted at last position
     */
    insertComment: function(uniqueId, contentId, commentId, comment, ascendingOrder)
    {
        let $htmlTpl = comment.level == 0 ? $j('#content-comment-template-' + uniqueId) : $j('#content-subcomment-template-' + uniqueId)
        let $rootComments = $j('[data-ametys-comments="' + contentId + '"]');
        let $ulComments = comment.level == 0 ? $rootComments : $rootComments.find('[data-ametys-comment-id=' + commentId + '][data-ametys-is-comment] [data-ametys-comment-subcomments]');
        
        let $newComment = this._insertComment($htmlTpl, comment, $ulComments, ascendingOrder)
        
        // show and focus new comment 
        $newComment.show();
        $newComment.attr("tabindex", "-1");
        $newComment.focus();
        
	    AmetysFront.Event.fire("comments-comment-added", {contentId: contentId, commentEl: $newComment, level: comment.level});
    },
    
    _insertComment: function($htmlTpl, data, $ulComments, ascendingOrder)
    {
        let $commentTpl = $htmlTpl.clone();
      
        let $commentEl = $j(document.createElement("div"));
        $commentEl.hide();
        $j('body').append($commentEl);
        
        let htmlComment = AmetysFront.Comments.fillComment($commentTpl.html(), data);
        
        $commentEl.html(htmlComment);
      
        if (!ametysUser)
        {    
             // Remove like, report and delete buttons if there is no connected user
             $commentEl.find("[data-ametys-comment-like]").remove();
             $commentEl.find("[data-ametys-comment-report]").remove();
             $commentEl.find("[data-ametys-comment-delete]").remove();
        }
        else if (!data.canDelete)
        {
             // Remove delete button if no right
             $commentEl.find("[data-ametys-comment-delete]").remove();
        }
        
        let $newComment;
        if (ascendingOrder === false)
        {
            // insert in first position
            $ulComments.prepend($commentEl.children());
            $newComment = $ulComments.find('[data-ametys-is-comment]').first();
        }
        else
        {
            // insert in last position
            $ulComments.append($commentEl.children());
            $newComment = $ulComments.find('[data-ametys-is-comment]').last();
        }
        
        if (data.level != 0)
        {
            $ulComments.show();
        }
        
        $commentEl.remove();
        
        return $newComment;
    },
    
    /**
     * Fill comment template from the JSON comment
     * @param {String} commentHtml The HTML of the comment
     * @param {Object} comment the comment as JSON
     * @return the HTML comment
     */
    fillComment: function (commentHtml, comment)
    {
        let me = this;
        
        AmetysFront.Comments._transformComment(comment);
        
        while ((result =  AmetysFront.Comments.COMMENT_VALUE_REGEXP.exec(commentHtml)) !== null) 
        {
            // replace string matched
            let value = me._getValue(comment, result[1]);
            if (Array.isArray(value) || value !== null && typeof value === 'object')
            {
                value = JSON.stringify(value).replace(/"/g, '&quot;');
            }
            commentHtml = commentHtml.replace(result[0], value);
        }
        
        return commentHtml;
    },
    
    _transformComment: function(comment)
    {
        // Replace breakline and sanitize untrusted HTML
        comment.content = comment.content.replace(/&/g, "&amp;").replace(/</g, "&lt;").replace(/\r?\n/g, "<br/>");
        
        comment.likers = [];
        let likers = [];
        if (comment.reactions)
        {
            $j(comment.reactions).each(function (i, reaction){
                if ("like" == reaction.type.toLowerCase())
                {
                    likers = reaction.actors;
                }
            })
        }
        comment.likers = likers;
        comment['nb-likes'] = likers.length;
        
        // Create uniqueid that can be used in a function name
        comment['uniqueId'] = comment.id.replaceAll("-", "_") + "_" + (''+Math.random()).replace(/\./g, '');
        
        if (ametysUser)
        {
            $j.each(comment.likers, function(i, liker) {
                if (liker.login == ametysUser.login && liker.populationId == ametysUser.populationId)
                {
                    comment["own-liker"] = true;
                }
            });
        }
        
        return comment;
    },
    
    _getValue: function (comment, dataName)
    {
        // dateName format can be : name, path.to.data
        let paths = dataName.split('.');
        
        var data = comment;
        for (var i=0; i < paths.length; i++)
        {
            var name = paths[i];
            if (i == paths.length -1)
            {
                return data[name] || '';
            }
            
            if (data[name])
            {
                data = data[name];
            }
            else
            {
                return '';
            }
        }
        
        return '';
    },
    
    /**
     * Delete a comment
     * @param {Element} btn the button calling this function
     * @param {String} contentId the content id
     * @param {String} commentId the comment id to delete
     * @param {Function} [callback] optional callback
     */
    deleteComment: function(btn, contentId, commentId, callback)
    {
        var me = this;
        AmetysFront.Utils.confirm("{{i18n plugin.cms:PLUGINS_CMS_CONTENT_COMMENTS_DELETE_TITLE}}", "{{i18n plugin.cms:PLUGINS_CMS_CONTENT_COMMENTS_DELETE_MSG}}", function () {
            AmetysFront.ServerComm.callMethod({
	            role: 'org.ametys.cms.repository.comment.CommentsDAO',
	            methodName: 'deleteComment',
	            parameters: [contentId, commentId],
	            callback: {
	                handler: me._deleteCommentCb,
	                scope: me,
	                arguments: {
	                    contentId: contentId, 
	                    commentId: commentId,
	                    btn: btn,
                        callback: callback
	                }
	            }
	        });
        });
    },
    
    _deleteCommentCb: function (result, args)
    {
       if (result && result.deleted)
       {
            if (args.commentId.indexOf("_") == -1)
            {
	            // Decrease nb comments in whole content
                this._updateNbComments(args.contentId, -1);
            }
                
            // remove comment
            $j(args.btn).closest('[data-ametys-is-comment]').remove();
            
	        AmetysFront.Event.fire("comments-comment-deleted", {contentId: args.contentId});
            
            if (typeof args.callback == 'function')
            {
                args.callback(result, args);
            }
       }
	   else
	   {
           AmetysFront.Utils.error("{{i18n plugin.cms:PLUGINS_CMS_CONTENT_COMMENTS_DELETE_ERROR_MSG}}");
	   }
    },
    
    _updateNbComments: function(contentId, gap)
    {
        // Update nb comments in whole content
        let $commentWrapper = $j('[data-ametys-comments-wrapper="' + contentId + '"]');
          
        let $nbComments;
        if ($commentWrapper.closest('.ametys-cms-content').length)
        {
            $nbComments = $commentWrapper.closest('.ametys-cms-content').find('[data-ametys-nb-comments]');
        }
        else
        {
            $nbComments = $j('[data-ametys-comments-wrapper="' + args.contentId + '"] [data-ametys-nb-comments]');
        }
        
        let count = Number($nbComments.html());
        count += gap;
        $nbComments.html(count);
    },
    
    /**
     * Report a comment
     * @param {Element} btn the button calling this function
     * @param {String} contentId the content id
     * @param {String} commentId the comment id to report
     * @param {Function} [callback] optional callback
     */
    reportComment: function(btn, contentId, commentId, callback)
    {
        var me = this;
        AmetysFront.Utils.confirm("{{i18n plugin.cms:PLUGINS_CMS_CONTENT_COMMENTS_REPORTED_CONFIRM_TITLE}}", "{{i18n plugin.cms:PLUGINS_CMS_CONTENT_COMMENTS_REPORTED_CONFIRM_MSG}}", function () {
            AmetysFront.ServerComm.callMethod({
                role: 'org.ametys.cms.repository.comment.CommentsDAO',
                methodName: 'reportComment',
                parameters: [contentId, commentId],
                callback: {
                    handler: callback || me._reportCommentCb,
                    scope: me,
                    arguments: {
                        contentId: contentId, 
                        commentId: commentId,
                        btn: btn
                    }
                }
            });
        });
    },
    
    _reportCommentCb: function(success, args)
     {
        if (success)
        {
            AmetysFront.Utils.info("{{i18n plugin.cms:PLUGINS_CMS_CONTENT_COMMENTS_REPORTED_TITLE}}", "{{i18n plugin.cms:PLUGINS_CMS_CONTENT_COMMENTS_REPORTED_MSG}}");
        }
        else
        {
            AmetysFront.Utils.error("{{i18n plugin.cms:PLUGINS_CMS_CONTENT_COMMENTS_REPORT_ERROR_MSG}}");
        }
     },
    
    /**
     * Add a comment
     * @param {String} contentId the content id
     * @param {Boolean} [ascendingOrder] Set to 'false' to insert new comment at first position, otherwise it will be inserted at last position
     * @param {Function} [callback] Optionnal callback function
     */
    addComment: function(uniqueId, contentId, ascendingOrder, callback)
    {
        var form = $j('[data-ametys-comments-form="' + contentId + '"]');
        this._addComment(form, uniqueId, contentId, null, ascendingOrder, callback);
    },
    
    /**
     * Add a subcomment
     * @param {String} contentId the content id
     * @param {Boolean} [ascendingOrder] Set to 'false' to insert new comment at first position, otherwise it will be inserted at last position
     * @param {String} [commentId] the parent comment id. Can be null.
     * @param {Function} [callback] Optionnal callback function
     */
    addSubComment: function(btn, uniqueId, contentId, ascendingOrder, commentId, callback)
    {
        var form = $j(btn).parents('form');
        var commentid = commentId || form.find('input[name=comment-id]').val();
        
        this._addComment(form, uniqueId, contentId, commentid, ascendingOrder, callback);
    },
    
    _addComment: function(form, uniqueId, contentId, commentId, ascendingOrder, callback)
    {
        if (!this._validateForm(form))
        {
            if (typeof callback == 'function')
            {
                callback(false);
            }
            
            return;
        }
        
        var values = form.serializeArray();
        
        var data = {};
        for(var i=0;i < values.length; i++)
        {
            if (values[i].name == 'hide-email')
            {
                data[values[i].name] = values[i].value == 'true' || values[i].value == 'on';
            }
            else
            {
                data[values[i].name] = values[i].value;
            }
        }
        
        AmetysFront.ServerComm.callMethod({
            role: 'org.ametys.cms.repository.comment.CommentsDAO',
            methodName: 'addComment',
            parameters: [contentId, commentId, data, AmetysFront.getAppParameters()],
            callback: {
                handler: this._addCommentCb,
                scope: this,
                arguments: {
                    uniqueId: uniqueId, 
                    contentId: contentId, 
                    commentId: commentId,
                    values: data,
                    form: form,
                    ascendingOrder: ascendingOrder,
                    callback: callback
                }
            }
        });
    },
    
    _validateForm: function(form)
    {
        var nameInput = form.find('input[name=name]'),
            emailInput = form.find('input[name=email]'),
            urlInput = form.find('input[name=url]'),
            contentInput = form.find('textarea[name=text]');
        
        // reset errors 
        this._resetErrors(form);
            
        var errors = false;
        if (/^\s*$/.test(nameInput.val()))
        {
            errors = true;
            this._markInvalid(nameInput, "{{i18n plugin.core-ui:PLUGINS_CORE_UI_DEFAULT_VALIDATOR_MANDATORY}}");
        }
        
        if (!/^.+@.+$/.test(emailInput.val()))
        {
            errors = true;
            this._markInvalid(emailInput, "{{i18n plugin.core-ui:PLUGINS_CORE_UI_DEFAULT_VALIDATOR_PATTERN_FAILED}}");
        }
        
        if (!/^(https?:\/\/.+)?$/.test(urlInput.val()))
        {
            errors = true;
            this._markInvalid(urlInput, "{{i18n plugin.core-ui:PLUGINS_CORE_UI_DEFAULT_VALIDATOR_PATTERN_FAILED}}");
        }

        if (/^\s*$/.test(contentInput.val()))
        {
            errors = true;
            this._markInvalid(contentInput, "{{i18n plugin.core-ui:PLUGINS_CORE_UI_DEFAULT_VALIDATOR_MANDATORY}}");
        }
        
        return !errors;
    },
    
    _resetForm: function(form)
    {
        form.find('input:not([type=hidden])').val('');
        form.find('textarea:not([type=hidden])').val('');
    },
    
    _resetErrors: function(form)
    {
        form.find('.field.error').removeClass('error');
        form.find("[aria-invalid='true']").each(function(i, input) {
            $j(input).removeAttr("aria-invalid");
            $j(input).removeAttr("aria-describedby");
            $j(input).next(".error").remove();
        });
    },
    
    _markInvalid: function(input, errorMsg)
    {
        input.attr("aria-invalid", true);
        $( `<span class="error" id="error-${input.attr('id')}">${errorMsg}</span>`).insertAfter(input);
        input.attr("aria-describedby", "error-" + input.attr('id'));
        input.closest(".field").addClass('error');
    },
    
    _addCommentCb: function (results, args)
    {
        var me = this;
        if (results.errors)
        {
            var form = args.form;
            var globalErrorMsg;
            $j(results.errors).each(function(i, error) {
                
                if (error.name)
                {
                    let input = form.find('[name=' + error.name + ']');
                    me._markInvalid(input, error.error);
                }
                else
                {
                    globalErrorMsg = error.error;
                }
            });
            
            if (globalErrorMsg)
            {
	            AmetysFront.Utils.error("{{i18n plugin.cms:PLUGINS_CMS_CONTENT_COMMENTS_FORM_ERROR_AJAX}}<br/>" + globalErrorMsg); // 
            }
        }
        else if (results.published)
        {
            // Merge comment's properties with form values
            let comment = results.comment;
            
            // Insert new comment
            AmetysFront.Comments.insertComment(args.uniqueId, args.contentId, args.commentId, comment, args.ascendingOrder);
            
            if (comment.level == 0)
            {
                // Increase nb comments
                this._updateNbComments(args.contentId, 1);
            }
            
	        // Reset form and errors
            this._resetForm(args.form);
            this._resetErrors(args.form);
            
            if (typeof args.callback == 'function')
	        {
	            args.callback(true, results, args);
	        }
        }
        else
        {
            AmetysFront.Utils.info("{{i18n plugin.cms:PLUGINS_CMS_CONTENT_COMMENTS_ADD_SENT_TITLE}}", "{{i18n plugin.cms:PLUGINS_CMS_CONTENT_COMMENTS_ADD_SENT_DESCRIPTION}}");
        }
    },
    
    
    /**
     * Add 'like' reaction on a comment
     * @param {String} contentId the content id
     * @param {String} commentId the comment id
     * @param {Boolean} remove true to remove the like, false to add like
     * @param {Function} [callback] the callback function to invoked after reacting to comment
     */
    likeOrUnlikeComment: function(btn, contentId, commentId, remove, callback)
    {
        AmetysFront.ServerComm.callMethod({
            role: 'org.ametys.cms.repository.comment.CommentsDAO',
            methodName: 'likeOrUnlikeComment',
            parameters: [contentId, commentId, remove],
            callback: {
                handler: this._likeOrUnlikeCommentCb,
                scope: this,
                arguments: {
                    callback: callback, 
                    contentId: contentId, 
                    commentId: commentId,
                    btn: btn,
                    callback: callback
                }
            }
        });
    },
    
    _likeOrUnlikeCommentCb: function(result, args)
    {
        var liked = result.liked,
            nbLikes = result.nbLikes;
            btn = args.btn;
        
        // Update nb likes
        $j(btn).closest('[data-ametys-is-comment]').find('[data-ametys-comment-nb-likes]').first().html(nbLikes != 0 ? nbLikes : '');
        
        // Update likers
        let likers = $j(btn).attr('data-ametys-comment-likers') || '[]';
        likers = JSON.parse(likers.replace(/&quot;/g, '"'));
        
        if (liked)
        {
            $j(btn).attr("data-ametys-comment-own-liker", true);
            likers.push(ametysUser);
        }
        else
        {
            $j(btn).attr("data-ametys-comment-own-liker", false);
            let index = -1;
            $j.each(likers, function(i, liker) {
                if (ametysUser && liker.login == ametysUser.login && liker.population == ametysUser.populationId)
                {
                    index = i;
                    return;
                }
            });
            likers.splice(index, 1);
        }
        
        $j(btn).attr('data-ametys-comment-likers', JSON.stringify(likers).replace(/"/g, '&quot;'));
        
        if (typeof args.callback == 'function')
        {
            args.callback(result, args);
        }
    }
}
