/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

function ping() 
{
    window.setTimeout(function() { 
        AmetysFront.ServerComm.send({
            plugin: 'core-ui',
            url: 'system-startuptime.xml',
            responseType: 'xml',
     
            callback: {
                handler: function(response) {
                    ping();
                },
                errorHandler: function() {}
            }
        }); 
    }, 60000);
}
ping();
 
AmetysFront.ServerComm = {
    
    SERVERCOMM_URL: undefined,
    
    callMethod: function(options)
    {
        return this.send({
            plugin: 'core-ui',
            url: 'client-call',
            responseType: 'json',
            parameters: {
                role: options.role,
                id: options.id,
                methodName: options.methodName,
                parameters: options.parameters || []
            },
            callback: options.callback
        });
    },
    
    send: function(options)
    {
        if (!AmetysFront.ServerComm.SERVERCOMM_URL)
        {
            AmetysFront.ServerComm.SERVERCOMM_URL = AmetysFront.getPluginDirectPrefix('cms') + '/servercomm/fo-messages.xml';
        }
        
        var defaultOpts = {
            pluginOrWorkspace: options.plugin,
            responseType: 'json',
            parameters: {},
            showSuccess: false,
            showError: true,
            showWait: false,
            upload: false
        };
        
        // Copies default options if they don't already exist in options.
        for (var property in defaultOpts) {
            if (options[property] === undefined) {
                options[property] = defaultOpts[property];
            }
        }
        
        var contentJson = JSON.stringify({
            pluginOrWorkspace: options.pluginOrWorkspace,
            url: options.url,
            parameters: options.parameters,
            responseType: options.responseType
        });
        
        var contextParamsJson = JSON.stringify(AmetysFront.getAppParameters());
        
        var data = new FormData();
        data.append("content", contentJson);
        data.append("context.parameters", contextParamsJson);

        // add uploaded files
        if (options.parameters.parameters && options.parameters.parameters.forEach)
        {
            var totalSize = 1000; // Let's keep some size for request overhead
                
            options.parameters.parameters.forEach(function(param, index) {
                if (param instanceof File)
                {
                    data.append("file-" + index, param);
                    options.parameters.parameters[index] = null;
                    totalSize += param.size;
                }
                else if (param instanceof Array)
                {
                    for (let i in param)
                    {
                        let subParam = param[i];
                        if (subParam instanceof File)
                        {
                            data.append("file-" + index + "-" + i, subParam);
                            options.parameters.parameters[index][i] = null;
                            totalSize += subParam.size;
                        }
                    }
                }
            });
            
            if (totalSize > AmetysFront.MAX_UPLOAD_SIZE)
            {
                let errorText = "{{i18n plugin.cms:PLUGINS_CMS_FRONT_SERVERCOMM_UPLOAD_ERROR_TOOBIG_TEXT}}"
                                    .replace(/\{0\}/g, AmetysFront.Utils.getReadableSize(totalSize))
                                    .replace(/\{1\}/g, AmetysFront.Utils.getReadableSize(AmetysFront.MAX_UPLOAD_SIZE));
                if (AmetysFront.Event)
                {
                    AmetysFront.Event.fire('loaderFail', {
                        title: "{{i18n plugin.cms:PLUGINS_CMS_FRONT_SERVERCOMM_UPLOAD_ERROR_TOOBIG_TEXT}}",
                        text : errorText
                    });
                }
                if (options.callback)
                {
                    let callback = options.callback;
                    (callback.errorHandler || callback.handler).apply(callback.scope, [callback.errorHandler ? errorText : null, callback.arguments]);
                }
                throw new Error("Cannot send files of " + totalSize + " bytes. The max size is " + AmetysFront.MAX_UPLOAD_SIZE + " bytes.");
            }
        }
        
        $j.ajax({ 
            url: AmetysFront.ServerComm.SERVERCOMM_URL, 
            dataType: "xml",
            data: data,
            contentType: false,
            processData: false,
            method: 'POST',
            type: 'POST',
            
            success: function(response, status, xhr) 
            { 
                var responseElement = response.documentElement;
                
                var responseCode = responseElement ? responseElement.getAttribute("code") : null,
                    isBadResponse = !responseCode || responseCode != '200';
                
                var callback = options.callback;
                
                if (responseElement && responseElement.tagName == "NotConnected")
                {
                    if (AmetysFront.Event)
                    {
                        AmetysFront.Event.fire('loaderFail', {
                                title: "{i18n plugin.cms:PLUGINS_CMS_FRONT_SERVERCOMM_ERROR_FATAL_SESSION_TITLE}}",
                                text: "{i18n plugin.cms:PLUGINS_CMS_FRONT_SERVERCOMM_ERROR_FATAL_SESSION_DESCRIPTION}}",
                                details: null,
                                fatal: 'session'
                        });
                        return;
                    }
                }
                else if (isBadResponse)
                {
                    if (callback)
                    {
                        (callback.errorHandler || callback.handler).apply(callback.scope, [callback.errorHandler ? $j("> message", responseElement).text() : null, callback.arguments]);
                    }
                    else if (AmetysFront.Event)
                    {
                        AmetysFront.Event.fire('loaderFail', {
                                title: null,
                                text: null,
                                details: null
                        });
                    }
                    return;
                }
                
                var responseData = responseElement; // xml response type
                if (options.responseType == 'json')
	            {
	                responseData = JSON.parse(responseElement.textContent);
	            }
	            
                if (callback)
                {
                    callback.handler.apply(callback.scope, [responseData, callback.arguments]);
                }
            },
            
            error: function(xhr, status, error)
            {
                window.console.error("Exception in AmetysFront.ServerComm " + error);
                
                if (AmetysFront.Event)
                {
                    if (xhr.status == 503)
                    {
                        AmetysFront.Event.fire('loaderFail', {
                                title: "{{i18n plugin.cms:PLUGINS_CMS_FRONT_SERVERCOMM_ERROR_FATAL_MAINTENANCE_TITLE}}",
                                text: "{{i18n plugin.cms:PLUGINS_CMS_FRONT_SERVERCOMM_ERROR_FATAL_MAINTENANCE_DESCRIPTION}}",
                                details: null,
                                fatal: 'maintenance'
                        });
                    }
                    else
                    {
                        AmetysFront.Event.fire('loaderFail', {
                                title: "{{i18n PLUGINS_CMS_FRONT_SERVERCOMM_ERROR_FATAL_SERVER_TITLE}}",
                                text: "{{i18n PLUGINS_CMS_FRONT_SERVERCOMM_ERROR_FATAL_SERVER_DESCRIPTION}}",
                                details: null,
                                fatal: 'server'
                        });
                    }
                }
                
                var callback = options.callback;
                if (callback)
                {
                    (callback.errorHandler || callback.handler).apply(callback.scope, [null, callback.arguments]);
                }
            }
        });
    }
}