/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
 
AmetysFront.Utils = {
    
    /**
     * Get a readable date, with relative time (eg: '4 days ago') for duration less that 7 days by default.
     * @param {String} date The raw date
     * @param {Number} [relativeLimit=7] limit in days for relative date 
     * @param {Number} [withoutRelativeSuffix=false] true to get the relative time without the suffix (eg: '4 days')
     * @param {Boolean} [withHours=false] true to format date with hours
     * @return {String} the readable date
     */
    toMomentDate: function (date, relativeLimit, withoutRelativeSuffix, withHours)
    {
        relativeLimit = relativeLimit || 7;
        withoutRelativeSuffix = withoutRelativeSuffix === true;
        
	    let now = moment();
	    let momentDate = moment(date);
	    
	    let diff = now.diff(momentDate);
	    let duration = moment.duration(diff);
	    let days = duration.asDays();
	    let isSameYear = now.isSame(momentDate, 'year');
	    
        if (days > 1 && days < 2)
        {
            // eg: hier, 16h45
            return moment(date).format("{{i18n plugin.cms:PLUGINS_CMS_MOMENT_DATE_FORMAT_YESTERDAY}}"); 
        }
	    else if (days < relativeLimit)
	    {
            // eg: il y a 26 min, il y a 2h, il y a 5 jours
	        return momentDate.fromNow(withoutRelativeSuffix); 
	    }
	    else if (isSameYear)
	    {
            // eg: le 26 juin, le 10 juillet à 11h30
            return withHours ? momentDate.format("{{i18n plugin.cms:PLUGINS_CMS_MOMENT_DATE_TIME_FORMAT}}") : momentDate.format("{{i18n plugin.cms:PLUGINS_CMS_MOMENT_DATE_FORMAT}}");
	    }
	    else
	    {
            // eg: le 26 juin 2020, le 10 juillet 2020 à 11h30
            return withHours ? momentDate.format("{{i18n plugin.cms:PLUGINS_CMS_MOMENT_DATE_TIME_FORMAT_WITH_YEAR}}") : momentDate.format("{{i18n plugin.cms:PLUGINS_CMS_MOMENT_DATE_FORMAT_WITH_YEAR}}");
	    }
    },
    
    /**
     * Format file size in readable size with units
     * @param {Number} bytes the file length in bytes
     */
    getReadableSize: function(bytes) 
    {
        let limit = 1024;
        let i = 0;
        
        let units = [' ' + "{{i18n plugin.core-ui:PLUGINS_CORE_UI_FORMAT_FILE_SIZE_UNIT_BYTE}}",
        ' ' + "{{i18n plugin.core-ui:PLUGINS_CORE_UI_FORMAT_FILE_SIZE_UNIT_MB}}",
        ' ' + "{{i18n plugin.core-ui:PLUGINS_CORE_UI_FORMAT_FILE_SIZE_UNIT_GB}}",
        ' ' + "{{i18n plugin.core-ui:PLUGINS_CORE_UI_FORMAT_FILE_SIZE_UNIT_TB}}"];
        
        let size = bytes;
    
        if (Math.abs(bytes) < limit) {
            return bytes + ' ' + "{{i18n plugin.core-ui:PLUGINS_CORE_UI_FORMAT_FILE_SIZE_UNIT_BYTE}}";
        }
        else {
            while (size > limit) {
                size = size / limit;
                i++;
            }
        }
    
        return Math.max(size, 0.1).toFixed(1) + units[i-1];
    },

    /**
     * Remove special languages characters to allow easy comparison
     * @param {String} s The string to deemphasize
     * @return {String} The modified string
     */
    deemphasize: function(s)
    {
        if (!s) return s;
          
        s = s.replace(new RegExp(/[ÀÁÂÃÄÅ]/g),"A");
        s = s.replace(new RegExp(/[àáâãäå]/g),"a");
        s = s.replace(new RegExp(/Æ/g),"AE");
        s = s.replace(new RegExp(/æ/g),"ae");
        s = s.replace(new RegExp(/Ç/g),"C");
        s = s.replace(new RegExp(/ç/g),"c");
        s = s.replace(new RegExp(/[ÈÉÊË]/g),"E");
        s = s.replace(new RegExp(/[èéêë]/g),"e");
        s = s.replace(new RegExp(/[ÌÍÎÏ]/g),"I");
        s = s.replace(new RegExp(/[ìíîï]/g),"i");
        s = s.replace(new RegExp(/Ñ/g),"N");
        s = s.replace(new RegExp(/ñ/g),"n");
        s = s.replace(new RegExp(/[ÒÓÔÕÖ]/g),"O");
        s = s.replace(new RegExp(/[òóôõö]/g),"o");
        s = s.replace(new RegExp(/Œ/g),"OE");
        s = s.replace(new RegExp(/œ/g),"oe");
        s = s.replace(new RegExp(/[ÙÚÛÜ]/g),"U");
        s = s.replace(new RegExp(/[ùúûü]/g),"u");
        s = s.replace(new RegExp(/[ÝŸ]/g),"y");
        s = s.replace(new RegExp(/[ýÿ]/g),"y");
    
        return s;
    },
    
    confirm: function(title, message, okFn)
    {
         $j('<div class="ui-confirm-box"></div>').html( message ).dialog({
            // Remove the closing 'X' from the dialog
            open: function(event, ui) { $j(".ui-dialog-titlebar-close").hide(); }, 
            title: title,
            resizable: true,
            minWidth: 500,
            maxWidth: 600,
            modal: true,
            zIndex: 10000,
            buttons: {
                "{{i18n plugin.cms:PLUGINS_CMS_AMETYS_FRONT_DIALOG_YES_BTN}}": function()  {
                    $j( this ).dialog( 'close' );
                    okFn();
                },
                "{{i18n plugin.cms:PLUGINS_CMS_AMETYS_FRONT_DIALOG_NO_BTN}}": function()  {
                    $j( this ).dialog( 'close' );
                }
            }
        });
    },
    
    info: function(title, message)
    {
        if ( !title )
            title = "{{i18n plugin.cms:PLUGINS_CMS_AMETYS_FRONT_DIALOG_INFO_TITLE}}";
            
        if ( !message )
            message = 'No Message to Display.';
            
        $j('<div class="ui-state-info ui-corner-all"></div>').html( message ).dialog({
            // Remove the closing 'X' from the dialog
            open: function(event, ui) { $j(".ui-dialog-titlebar-close").hide(); }, 
            title: title,
            resizable: true,
            minWidth: 500,
            maxWidth: 600,
            modal: true,
            zIndex: 10000,
            buttons: {
                "{{i18n plugin.cms:PLUGINS_CMS_AMETYS_FRONT_DIALOG_ERROR_OK_BTN}}": function()  {
                    $j( this ).dialog( 'close' );
                }
            }
        });
    },
    
    error: function(message, title)
    {
        if ( !title )
            title = "{{i18n plugin.cms:PLUGINS_CMS_AMETYS_FRONT_DIALOG_ERROR_TITLE}}";

        if ( !message )
            message = 'No Message to Display.';

        $j('<div class="ui-state-error ui-corner-all"></div>').html( message ).dialog({
            // Remove the closing 'X' from the dialog
            open: function(event, ui) { $j(".ui-dialog-titlebar-close").hide(); }, 
            classes: {
              "ui-dialog": "ametys-ui-dialog-error"
            },
            title: title,
            resizable: true,
            minWidth: 500,
            maxWidth: 600,
            modal: true,
            zIndex: 10000,
            buttons: {
                "{{i18n plugin.cms:PLUGINS_CMS_AMETYS_FRONT_DIALOG_ERROR_OK_BTN}}": function()  {
                    $j( this ).dialog( 'close' );
                }
            }
        });
    }
}
