/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
(function(mod) {
    if (typeof exports == "object" && typeof module == "object") // CommonJS
        mod(require("../../lib/codemirror"));
    else if (typeof define == "function" && define.amd) // AMD
        define(["../../lib/codemirror"], mod);
    else // Plain browser env
        mod(CodeMirror);
})(function(CodeMirror) {
    var Pos = CodeMirror.Pos;

    function scriptHint(editor, getToken, options) {
        // Find the token at the cursor
        var cur = editor.getCursor(), token = getToken(editor, cur), tprop = token;
        var hintOpt = editor.options.hintOptions || {};

        Ext.applyIf(hintOpt, { allowCommonOption: true })
        
        token.state = CodeMirror.innerMode(editor.getMode(), token.state).state;
        
        if (SolrHint.getLogger().isDebugEnabled())
        {
            SolrHint.getLogger().debug('Token string=[' + token.string + '], type=[' + token.type + '], state=' + JSON.stringify(token.state));
        }
        
        if (token.string.trim() == '')
        {
            token = tprop = {start: cur.ch, end: cur.ch, string: '', state: token.state, type: token.type};
        }
        else if (token.string == '.' || token.string == ',')
        {
            token = tprop = {start: cur.ch, end: cur.ch, string: '', state: token.state, type: token.type};
        }
        
        // If it is a property, find out what it is a property of.
        var context = null;
        
        while (tprop.type == 'string' && tprop.string != '' || tprop.type == 'dot')
        {
            tprop = getToken(editor, Pos(cur.line, tprop.start));
            if (tprop.type == 'string')
            {
                if (context == null) context = [];
                context.push(tprop);
            }
        }
        
        if (SolrHint.getLogger().isDebugEnabled())
        {
            var c = '';
            if (context)
            {
                for (var i = 0; i < context.length; i++)
                {
                    if (i > 0) c += ', ';
                    c += '[' + context[i].string + ':' + context[i].type + ']';
                }
            }
            SolrHint.getLogger().debug('Contexts= ' + c);
        }
        
        return {list: getCompletions(token, context, options, hintOpt),
                from: Pos(cur.line, token.start),
                to: Pos(cur.line, token.end)};
    }

    function solrColHint(editor, options) {
        return scriptHint(editor,
                          function (e, cur) {return e.getTokenAt(cur)},
                          options);
    };
    CodeMirror.registerHelper('hint', 'solr-ametys-columns', solrColHint);

    function getCompletions(token, context, options, hintOpt) {
        var found = [],
            completions = [],
            start = token.string,
            global = options && options.globalScope || window,
            fieldName = hintOpt.fieldName,
            commonFields = hintOpt.allowCommonOption ? SolrHint.getCommonFields() : {},
            cTypes = hintOpt.ctypes;
        function maybeAdd(str) {
            if (SolrHint.startsWithIgnoreCase(str, start) && !Ext.Array.contains(found, str))
            {
                found.push(str);
                completions.push(str);
            }
        }
        function maybeAddField(name, field) {
            var str = name;
            if (SolrHint.startsWithIgnoreCase(str, start) && !Ext.Array.contains(found, str))
            {
                var text = str;
                var details = field.type;
                
                if (field.type == 'composite'
                    || field.type == 'repeater' && fieldName == 'facets')
                {
                    text = str + '.';
                }
                else if (fieldName == 'facets' && !field.isFacetable)
                {
                    return;
                }
                else if (fieldName != 'facets' && !field.isDisplayable)
                {
                    return;
                }
                
                found.push(str);
                
                completions.push({
                    text: text,
                    displayText: str,
                    className: 'field',
                    details: details,
                    render: SolrHint.renderSolr
                });
            }
        }
        function maybeAddAllOption(hintOpt)
        {
            if (start === '' && hintOpt.allowAllOption)
            {
                completions.push({
                    text: '*',
                    displayText: '*',
                    className: 'field',
                    details: 'all',
                    render: SolrHint.renderSolr
                });
            }
        }
        function gatherCTypeMetas(cType) {
            var cType = SolrHint.getCTypeMetadatas(cType);
            if (cType == null)
            {
                return;
            }
            
            gatherMetas(cType.fields);
            
            if (cType.superTypes)
            {
                for (var i = 0; i < cType.superTypes.length; i++)
                {
                    gatherCTypeMetas(cType.superTypes[i]);
                }
            }
        }
        function gatherMetas(metas) {
            for (var name in metas)
            {
                maybeAddField(name, metas[name]);
            }
        }
        function getHolder(cTypes, context)
        {
            var element = SolrHint.getMeta(cTypes, context[0].string, null);
            for (var i = 1; i < context.length && element != null; i++)
            {
                if (element.type == 'composite' || element.type == 'repeater')
                {
                    element = SolrHint.getMetaInHolder(element, context[i].string, null);
                }
                else if (element.type == 'content')
                {
                    element = SolrHint.getMeta([element.cType], context[i].string, null);
                }
            }
            
            return element;
        }
        
        if (context && context.length && context.length > 0)
        {
            // We have a context: we're trying to complete a sub-column (a.b.)
            context.reverse(); // The context array is provided backwards, we have to reverse it.
            var holder = getHolder(cTypes, context);
            if (holder != null && (holder.type == 'composite' || holder.type == 'repeater'))
            {
                maybeAddAllOption(hintOpt);
                // Composite: gather sub-metadatas.
                gatherMetas(holder.fields);
            }
            else if (holder != null && holder.type == 'content')
            {
                maybeAddAllOption(hintOpt);
                

                if(hintOpt.allowCommonOption)
                {
                    // Add the common fields (system properties)
                    for (var name in commonFields)
                    {
                        maybeAddField(name, commonFields[name]);
                    }
                }
            
                // Content type: gather content type metadatas.
                gatherCTypeMetas(holder.cType);
            }
        }
        else
        {
            maybeAddAllOption(hintOpt);

            if(hintOpt.allowCommonOption)
            {
                // Add the common fields (system properties) and root metadata.
                for (var name in commonFields)
                {
                    maybeAddField(name, commonFields[name], false);
                }
            }
            
            for (var i = 0; i < cTypes.length; i++)
            {
                gatherCTypeMetas(cTypes[i]);
            }
        }
        
        return completions.sort(SolrHint.sortCompletions);
    }
    
});
