/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
Ext.define('SolrHint',
{
    singleton: true,
    
    /**
     * @private
     * @property {Boolean} _initialized true if the data have been initialized
     */
    _initialized: false,
    
    init: function()
    {
        if (!this._initialized)
        {
            var language = Ametys.cms.language.LanguageDAO.getCurrentLanguage();
            this._initCommonFields(language);
            this._initContentTypeMetadatas(language);
            this._initialized = true;
        }
    },
    
    startsWithIgnoreCase: function(str, start)
    {
        return str != null && start != null && str.toLowerCase().lastIndexOf(start.toLowerCase(), 0) == 0;
    },
    
    getMetaInHolder: function(holder, name, type)
    {
        if (holder.fields[name] && (type == null || holder.fields[name].type == type))
        {
            return holder.fields[name];
        }
        return null;
    },
    
    getMeta: function(cTypes, name, type)
    {
        for (var i = 0; i < cTypes.length; i++)
        {
            var cType = this.getCTypeMetadatas(cTypes[i]);
            var meta = this.getMetaInHolder(cType, name, type);
            if (meta != null)
            {
                return meta;
            }
            else if (cType.superTypes)
            {
                // If not found, search super types.
                meta = this.getMeta(cType.superTypes, name, type);
                if (meta != null)
                {
                    return meta;
                }
            }
        }
        return null;
    },
    
    getMetaByFieldInHolder: function(holder, type, fieldName, value)
    {
        for (var mName in holder.fields)
        {
            if (holder.fields[mName][fieldName] == value && (type == null || holder.fields[mName].type == type))
            {
                return holder.fields[mName];
            }
        }
        return null;
    },
    
    getMetaByField: function(cTypes, type, fieldName, value)
    {
        for (var i = 0; i < cTypes.length; i++)
        {
            var cType = this.getCTypeMetadatas(cTypes[i]);
            var meta = this.getMetaByFieldInHolder(cType, type, fieldName, value);
            
            if (meta != null)
            {
                return meta;
            }
            else if (cType.superTypes)
            {
                // If not found, search super types.
                meta = this.getMetaByField(cType.superTypes, type, fieldName, value);
                if (meta != null)
                {
                    return meta;
                }
            }
        }
        return null;
    },
    
    getComposite: function(cTypes, name) {
        for (var i = 0; i < cTypes.length; i++)
        {
            var cType = this.getCTypeMetadatas(cTypes[i]);
            if (cType.fields[name] && cType.fields[name].type == 'composite')
            {
                return cType.fields[name];
            }
            else if (cType.superTypes)
            {
                // If not found, search super types.
                var composite = this.getComposite(cType.superTypes, name);
                if (composite != null)
                {
                    return composite;
                }
            }
        }
        return null;
    },
    
    getPathElementInHolder: function(holder, token)
    {
        var element = null;
        if (token.type == 'composite')
        {
            element = this.getMetaInHolder(holder, token.string, 'composite');
        }
        else if (token.type == 'content')
        {
            // Can be either a content or a repeater.
            element = this.getMetaByFieldInHolder(holder, null, 'joinFl', token.string);
        }
        return element;
    },
    getPathElement: function(cTypes, token)
    {
        var element = null;
        if (token.type == 'composite')
        {
            element = this.getMeta(cTypes, token.string, 'composite');
        }
        else if (token.type == 'content')
        {
            // The token is the join field
            // Can be either a content or a repeater.
            element = this.getMetaByField(cTypes, null, 'joinFl', token.string);
        }
        return element;
    },
    
    sortCompletions: function(e1, e2)
    {
        var t1 = e1.displayText || e1;
        var t2 = e2.displayText || e2;
        return t1.toLowerCase() < t2.toLowerCase() ? -1 : 1;
    },
    
    renderSolr: function(elt, self, data)
    {
        // Default
//        elt.appendChild(document.createTextNode(data.displayText || getText(data)));
        
        var text = data.displayText || this.getText(data);
        
        var span1 = document.createElement('span');
        span1.setAttribute('class', 'hint-main');
        span1.appendChild(document.createTextNode(text));
        elt.appendChild(span1);
        
        if (data.details)
        {
            var span2 = document.createElement('span');
            span2.setAttribute('class', 'hint-details');
            span2.appendChild(document.createTextNode(' - ' + data.details));
            elt.appendChild(span2);
        }
    },
  
    getText: function(completion)
    {
        if (typeof completion == 'string') return completion;
        else return completion.text;
    },
    
    /**
     * 
     * Initialize the common fields
     */
    _initCommonFields: function(language)
    {
        Ametys.data.ServerComm.callMethod({
            role: 'org.ametys.cms.search.solr.SolrQueryHelper',
            methodName: 'getCommonFields',
            parameters: [language],
            callback: {
                scope: this,
                handler: this._initCommonFieldsCb,
                arguments: {}
            },
            waitMessage: true,
            errorMessage: {
                msg: "{{i18n PLUGINS_CMS_UITOOL_SOLR_ERROR_GET_COMMON_FIELDS}}",
                category: Ext.getClassName(this) + '.initCommonFields'
            }
        });
    },
    
    /**
     * @private
     * Callback function invoked after the common fields retrieving is complete
     * @param {Object} response The server's response
     * @param {Object} response.fields the common fields
     * @param {Object} args the callback arguments
     */
    _initCommonFieldsCb: function(response, args)
    {
        this._commonFields = response.fields;
    },
    
    /**
     * Get the common fields
     * @return {Object} The common fields
     */
    getCommonFields: function()
    {
        return this._commonFields;
    },
    
    /**
     * @private
     * Initialize an object containing information of all of the content types' metadata
     */
    _initContentTypeMetadatas: function(language)
    {
        Ametys.data.ServerComm.callMethod({
            role: 'org.ametys.cms.search.solr.SolrQueryHelper',
            methodName: 'getAllContentTypeIndexingFields',
            parameters: [language],
            callback: {
                scope: this,
                handler: this._initContentTypeMetadatasCb,
                arguments: {}
            },
            waitMessage: true,
            errorMessage: {
                msg: "{{i18n PLUGINS_CMS_UITOOL_SOLR_ERROR_GET_CTYPES}}",
                category: Ext.getClassName(this) + '.initContentTypeMetadatas'
            }
        });
    },
    
    /**
     * @private
     * Callback function invoked when the content types metadata have been retrieved
     * @param {Object} response the server's response
     * @param {Object} args the callback arguments
     */
    _initContentTypeMetadatasCb: function(response, args)
    {
        this._cTypeMetadatas = response.contentTypes;
    },
    
    /**
     * Get the chosen content type's metadata
     * @param {String} cType the content type
     * @return {Object} The content type's metadata
     */
    getCTypeMetadatas: function(cType)
    {
        return this._cTypeMetadatas[cType];
    }
});