/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
CodeMirror.defineMode("solr-ametys", function() {
  "use strict";

  var isStringChar = /[^\s\|\!\+\*\?\~\^\&\:\=\(\)\[\]\{\}\>\^\"\/\\]/;
  var isOperatorChar = /[\|\!\+\-\*\?\~\^\&]/;
  var isOperatorString = /^(OR|AND|NOT|TO)$/;

  function isNumber(word) {
    return parseFloat(word, 10).toString() === word;
  }

  function tokenString(quote) {
    return function(stream, state) {
      var escaped = false, next;
      while ((next = stream.next()) != null) {
        if (next == quote && !escaped) break;
        escaped = !escaped && next == "\\";
      }

      if (!escaped) state.tokenize = tokenBase;
      return "string";
    };
  }
  
  function tokenSubquery(openingBrace) {
      return function(stream, state) {
          var ch;
          var word = '';
          while ((ch = stream.peek()) && ch.match(isStringChar) != null) {
            word += stream.next();
          }
          
          state.solrInSubquery = true;
          
          state.tokenize = tokenBase;
          
          return 'sqname';
      };
  }

  function tokenOperator(operator) {
    return function(stream, state) {
      var style = "operator";
      if (operator == "+")
        style += " positive";
      else if (operator == "-")
        style += " negative";
      else if (operator == "|")
        stream.eat(/\|/);
      else if (operator == "&")
        stream.eat(/\&/);
      else if (operator == "^")
        style += " boost";

      state.tokenize = tokenBase;
      return style;
    };
  }
  
  function addJoin(state, field, type)
  {
      if (state.solrInSqJoin)
      {
          if (!state.sqJoin)
          {
              state.sqJoin = [];
          }
          state.sqJoin.push({field: field, type: type});
      }
  }
  
  function tokenWord(ch) {
    return function(stream, state) {
      var word = ch;
      while ((ch = stream.peek()) && ch.match(isStringChar) != null) {
          word += stream.next();
      }
      
      state.tokenize = tokenBase;
      if (isOperatorString.test(word))
          return "operator";
      else if (isNumber(word))
          return "number";
      else if (stream.peek() == '/')
      {
          if (state.solrInSubquery) addJoin(state, word, 'composite');
          return 'composite';
      }
      else if (stream.peek() == '>' && word.endsWith("-") && state.solrInSubquery && state.solrInSqJoin)
      {
          stream.backUp(1);
          word = word.substring(0, word.length - 1); // Take off the ending '-'
          if (state.solrInNestedQuery)
          {
            return 'string';
          }
          else
          {
            addJoin(state, word, 'content');
            return 'content';
          }
      }
      else if (stream.peek() == ':')
          return "field";
      else if (stream.peek() == '=' && state.solrInSubquery)
      {
          if (word == 'join') state.solrInSqJoin = true;
          return "sqparam";
      }
      else if (state.solrInSubquery && !state.solrInNestedQuery)
      {
          addJoin(state, word, 'content');
          return 'content';
      }
      else
          return "string";
    };
  }

  function tokenBase(stream, state) {
    if (state.afterNestedQuery)
    {
      delete state.afterNestedQuery;
      delete state.solrInNestedQuery;
    }
    
    var prev = stream.string.charAt(stream.pos - 1);
    var ch = stream.next();
    if (ch == '"')
      state.tokenize = tokenString(ch);
    else if (ch == '{' && stream.peek() == '!')
    {
      var next = stream.next(); // consume '!'
      state.tokenize = tokenSubquery(ch);
      return null;
    }
    else if (ch == '}')
    {
      delete state.solrInSubquery;
      delete state.sqJoin;
    }
    else if (state.solrInSubquery && ch == '-' && stream.peek() == '>')
    {
      var next = stream.next(); // consume '>'
    }
    else if (ch == '[' && state.solrInSubquery && state.solrInSqJoin)
    {
      state.solrInNestedQuery = true;
    }
    else if (ch == ']' && prev != '\\' && state.solrInSubquery && state.solrInSqJoin && state.solrInNestedQuery)
    {
      state.afterNestedQuery = true;
    }
    else if (isOperatorChar.test(ch))
      state.tokenize = tokenOperator(ch);
    else if (isStringChar.test(ch))
    {
      state.tokenize = tokenWord(ch);
    }

    return (state.tokenize != tokenBase) ? state.tokenize(stream, state) : null;
  }

  return {
    startState: function() {
      return {
        tokenize: tokenBase
      };
    },

    token: function(stream, state) {
      if (stream.eatSpace())
      {
          delete state.solrInSqJoin;
          return null;
      }
      return state.tokenize(stream, state);
    }
  };
});

CodeMirror.defineMIME("text/x-solr-ametys", "solr-ametys");
