/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

Ametys.namespace('Content', { migrate: function (content, migrateFunctions, previousVersionsAreIncompatible, labelOnNewVersions, verbose)
    {
        const _verbosePrint = verbose ? print : function() {};
      
        function _migrateContent(content, text)
        {
            _verbosePrint(text);
            _verbosePrint("<div style='padding-left: 10px; border-left: 1px solid #000; white-space: pre;'>");
            for (let i = 0; i < migrateFunctions.length; i++)
            {
                migrateFunctions[i](content);
            }
            if (content.needsSave())
            {
                if (previousVersionsAreIncompatible)
                {
                    _verbosePrint("setting previous version as not compatible<br/>");
                    content.addLabel("NotCompatible", true);
                    
                    previousVersionsAreIncompatible = null; // not to do twice
                }
              
                content.saveChanges();
                content.checkpoint();
            }
            _verbosePrint("\n</div>");
        }
      
        // Makes migrateFunctions an array of function
        const isArray = ('isArray' in Array) ? Array.isArray : function(value) {
                return toString.call(value) === '[object Array]';
        }
        migrateFunctions = isArray(migrateFunctions) ? migrateFunctions : [migrateFunctions];
      
    
        _verbosePrint("<div style='margin-left: 5px; border: 1px solid #000; white-space: nowrap;'>");
        _verbosePrint("<div style='padding: 10px; background-color: #000; color: #fff;'>");
        _verbosePrint({ toString: function() {"<b>Content</b>: " + content.getLanguage() + "/" + content.getName() + " (" + content.getId() + ")<br/>"
                        + "         " + content.getTitle()
                     }});
        _verbosePrint("</div>");
    
        _verbosePrint("<div style='padding: 10px;'>");
    
        // Unlock
        const LockAwareAmetysObject = Java.type('org.ametys.plugins.repository.lock.LockAwareAmetysObject');
        if (content instanceof LockAwareAmetysObject && content.isLocked())
        {
            _verbosePrint("unlocking...<br/>");
            content.unlock();
        }
      
        let hasLiveVersion = false;
        let currentVersionIsLive = false;
      
        const VersionableAmetysObject = Java.type('org.ametys.plugins.repository.version.VersionableAmetysObject');
        if (content instanceof VersionableAmetysObject)
        {
            hasLiveVersion = Java.from(content.getAllLabels()).indexOf('Live') >= 0;
            currentVersionIsLive = Java.from(content.getLabels()).indexOf('Live') >= 0;
          
            if (hasLiveVersion && !currentVersionIsLive)
            {
                const currentVersion = content.getNode().getBaseVersion().getName();
              
                _verbosePrint("switching to old live version<br/>");
                content.restoreFromLabel('Live');
                _migrateContent(content, "migrating old live version...");
                content.addLabel('Live', true); 
                
                if (labelOnNewVersions)
                {
                    _verbosePrint(`setting label ${labelOnNewVersions}<br/>`);
                    content.addLabel(labelOnNewVersions, true);
                    labelOnNewVersions = null; // Not to be set after 
                }    
              
                _verbosePrint("switching back to current revision<br/>");
                content.restoreFromRevision(currentVersion);
            }
        }  
    
        // Applying migration on head version
        _migrateContent(content, "migrating current version...<br/>");
      
        if (currentVersionIsLive)
        {
            // moving "live" label
            _verbosePrint("moving live label<br/>");
            content.addLabel('Live', true);     
        }
    
        if (labelOnNewVersions)
        {
            _verbosePrint(`setting label ${labelOnNewVersions}<br/>`);
            content.addLabel(labelOnNewVersions, true);
            labelOnNewVersions = null; // Not to be set after 
        }    
        
      
        _verbosePrint("</div></div>");
    }
});

// Obsolete
function migrateContent(content, migrateFunctions, previousVersionsAreIncompatible, labelOnNewVersions, verbose) {
    Ametys.console.warn("migrateContent is obsolete. Use Content.migrate instead.");
    return Content.migrate(content, migrateFunctions, previousVersionsAreIncompatible, labelOnNewVersions, verbose);
}
