/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

Ametys.namespace('Report', (function() {
	
	const Arrays = Java.type('java.util.Arrays');
 	const Collections = Java.type('java.util.Collections');
 	const Paths = Java.type('java.nio.file.Paths');
 	const Files = Java.type('java.nio.file.Files');
 	const StandardCharsets = Java.type('java.nio.charset.StandardCharsets');
  	const PrintWriter = Java.type('java.io.PrintWriter');
  	const OutputStreamWriter = Java.type('java.io.OutputStreamWriter');

			
	const cmsScriptHandler = Ametys.serviceManager.lookup("org.ametys.cms.scripts.CmsScriptHandler");
  	const valuesExtractorFactory = Ametys.serviceManager.lookup("org.ametys.cms.search.content.ContentValuesExtractorFactory");
 	const searchModelHelper = Ametys.serviceManager.lookup("org.ametys.cms.search.model.SearchModelHelper");
 	const i18nUtils = Ametys.serviceManager.lookup("org.ametys.core.util.I18nUtils");
 	const currentUserProvider = Ametys.serviceManager.lookup("org.ametys.core.user.CurrentUserProvider");

  
 	const dirName = "tempDirectoryForReports"

    function create(title, description, content) {
        return new Report(title, description, content);
    }

    class Report {
        constructor(title, description, content) {
            this.title = title;
            this.description = description;
			this.headers = [];
			this.rows = [];
			this.filePath = [];
			if(content != undefined)
			{
				var self = this;
				
                const columnsAsList = Arrays.asList(content.columns)
  				var model = cmsScriptHandler.getOrCreateModel(columnsAsList, content.results, "search-ui.default")
  
				var modelConfiguration = searchModelHelper.getSearchModelInfo(model, Collections.EMPTY_MAP);
				var columns = modelConfiguration.get("columns")
				var columnsTitle = []
				columns.forEach(function (column) {
					columnsTitle.push(i18nUtils.translate(column.get("label")))
					});
					
				var header = this.addHeader(columnsTitle)
				var values = model.getResultFields(Collections.EMPTY_MAP).values();
				var extractor = valuesExtractorFactory.create(model);
				
				content.results.forEach(function (result) {
					var row = self.addRow()
  					var jsonifiedContent = cmsScriptHandler.content2Json(result, values, extractor)
					columns.forEach(function (column) {
						var value;
						if(column.get("multiple"))
						{
							var valueArray = [];
							jsonifiedContent.get(column.get("id")).forEach(function (element) {
								valueArray.push(self._parseCellValue(column, element))
							});
							value = valueArray.join(", ")
						}
						else
						{
							value = self._parseCellValue(column, jsonifiedContent.get(column.get("id")))
						}
						
						row.addColumn(value != null ? value : "")
					});
				});
			}
        }

		addHeader(...args) {
			var header = new Header();
			args.forEach(function (arg) {
				if(typeof arg == "object" && Array.isArray(arg))
				{
					arg.forEach(function (elt) {
						header.addColumn(elt)
					});
				}
				else
				{
					header.addColumn(arg)
				}
			});
			this.headers.push(header);
			return header;
		}
		
		addRow(...args) {
			var row = new Header();
            args.forEach(function (arg) {
                if(typeof arg == "object" && Array.isArray(arg))
                {
                    arg.forEach(function (elt) {
                        row.addColumn(elt)
                    });
                }
                else
                {
                    row.addColumn(arg)
                }
            });
			this.rows.push(row);
			return row;
		}
		
		print() {
			print(this._exportAsHTML());
		}
		
		printLink() {
			if(this.lastFilePath == undefined)
			{
				print ("No file was saved for this report");
			}
			else
			{
				var fileName = this.lastFilePath.getFileName()
				var login = this.lastFilePath.getParent().getFileName()
				var populationId = this.lastFilePath.getParent().getParent().getFileName()
				print ("<a target='_blank' href='" + org.ametys.runtime.config.Config.getInstance().getValue("cms.url") + "/plugins/cms/script-report/" + populationId  + "/"  + login + "/" + fileName + "'>" + fileName + "</a>");
			}
			
		}
		
		saveAsCSV(addBOM = true) {
			var csvString = this._exportAsCSV(addBOM)
			return this._saveFile(csvString, "csv")
		}
		
		saveAsXLS() {
			var xlsString = this._exportAsHTML()
			return this._saveFile(xlsString, "xls")
		}
		
		saveAsHTML() {
			var htmlString = this._exportAsHTML()
			return this._saveFile(htmlString, "html")
		}
		
		_exportAsCSV(addBOM) {
			var lineArray = [];
			var headersAndRows = this.headers.concat(this.rows)
			headersAndRows.forEach(function (header) {
				var columnArray = []
				
				header.columns.forEach(function (column) {
					columnArray.push(column.valueAsCSV())
				});
			    var line = columnArray.join(";");
			    lineArray.push(line);
			});
			var csvContent = lineArray.join("\n");
			return (addBOM ? '\ufeff' : '') + csvContent;
		}
		
		_exportAsHTML() {
            var max = this._getMaxColumns();
            
			var htmlTable = "<html><head>"
            htmlTable += "<meta charset=\"UTF-8\">";
            htmlTable += "<title>" + this.title + "</title>";
			htmlTable += "</head><body><table border='1' cellspacing='0' cellpadding='3'>"
            htmlTable += "<tr><th colspan='" + max + "'><h1>" + this.title + "</h1></th></tr>";
            if (this.description)
            {
                htmlTable += "<tr><th colspan='" + max + "'>" + this.description + "</th></tr>";
            }
            htmlTable += "<tr><th colspan='" + max + "' align='right'>" + new Date() + "</th></tr>";
            
			this.headers.forEach(function (header) {
				htmlTable += "<tr>"
				header.columns.forEach(function (column) {
					htmlTable += this._createCell(column, true)
				}, this);
				htmlTable += "</tr>";
			}, this);
			
			this.rows.forEach(function (row) {
				htmlTable += "<tr>";
				row.columns.forEach(function (column) {
					htmlTable += this._createCell(column, false)
				}, this);
				htmlTable += "</tr>";
			}, this);
			htmlTable+="</table></body></html>"
			return htmlTable;
		}
		
		_saveFile(lines, extension)
		{
			if(this.filePath[extension] == undefined)
			{
				
				var user = currentUserProvider.getUser();
		        var populationId = user.getPopulationId()
		        var login = user.getLogin()
	
				var scriptReportDirPath = org.ametys.cms.scripts.ReportLocationAction.getScriptReportDirectory();	
	            if (!scriptReportDirPath.toFile().exists())
	            {
					Files.createDirectory(scriptReportDirPath);
				}
				
				var popIdDirPath = scriptReportDirPath.resolve(populationId)	
	            if (!popIdDirPath.toFile().exists())
	            {
					Files.createDirectory(popIdDirPath);
				}
				
				var userLoginPath = popIdDirPath.resolve(login)	
	            if (!userLoginPath.toFile().exists())
	            {
					Files.createDirectory(userLoginPath);
				}
				
				var fileToCreatePath = userLoginPath.resolve("report-" + org.ametys.core.util.StringUtils.generateKey() + "." + extension);
	        	this.filePath[extension] = Files.createFile(fileToCreatePath);
				this.lastFilePath = this.filePath[extension];
			}

  			return Files.writeString(this.filePath[extension], lines, java.nio.charset.StandardCharsets.UTF_8).toFile();
			
		}
		
		_createCell(column, isHeader)
		{
			var tagName = isHeader ? "th" : "td"
			var colspan = column.colspan != undefined ? ' colspan="' +column.colspan + '"' : ""
			var rowspan = column.rowspan != undefined ? ' rowspan="' +column.rowspan + '"' : ""
			var cell ="<" + tagName + colspan + rowspan + ">"
			cell += column.value
			cell += "</" + tagName + ">"
			return cell;
		}

		_parseCellValue(column, jsonifiedValue)
		{
			var id = column.get("id")
			var value;
			var type = column.get("type")
			if(type == "STRING")
			{
				if(typeof jsonifiedValue == "string")
				{
					value = jsonifiedValue
				}
				else if(id == "site")
				{
					value = jsonifiedValue.get("name")
				}
				else if(id == "contentLanguage")
				{
					value = jsonifiedValue.get("code")
				}
				else if(id == "tags")
				{
					value =  i18nUtils.translate(jsonifiedValue.get("label"))
				}
				else
				{
					value = i18nUtils.translate(jsonifiedValue)
				}
			}
			else if(type == "USER")
			{
				value = jsonifiedValue.get("fullname")
			}
			else if(type == "DATETIME")
			{
				value = jsonifiedValue
			}
			else if(type == "BOOLEAN")
			{
				if(id == "shared")
				{
					value = jsonifiedValue.get("isShared")
				}
				else
				{
					value = jsonifiedValue
				}
			}
			else if(type == "LONG")
			{
				if(id == "workflowStep")
				{
					value = i18nUtils.translate(jsonifiedValue.get("name"))
				}
				else
				{
					value = jsonifiedValue
				}
			}
			else
			{
				value = jsonifiedValue
			}
			return value
		}
		
		_getMaxColumns()
		{
			var headersAndRows = this.headers.concat(this.rows)
			var max = 0;
			headersAndRows.forEach(function (header) {
                var colMax = 0;
                for (var i = 0; i < header.columns.length; i++)
                {
                    colMax += header.columns[i].colspan || 1;
                }
                
				if (max < colMax)
				{
					max = colMax;
				}
			});
            return max;
		}
    }

    class Header {
        constructor() {
			this.columns = []
        }

		addColumn(text, colspan, rowspan) {
			var column = new Column(text, colspan, rowspan);
			this.columns.push(column);
			return column;
		}
	}

    class Column {
        constructor(text, colspan, rowspan) {
			this.value = text;
			this.colspan = colspan;
			this.rowspan = rowspan;
        }

		valueAsCSV()
		{
			if (typeof this.value == "string")
			{
				return '"' + this.value.replace(/"/g, '""') + '"';
			}
			else
			{
				return this.value;
			}
		}
	}

    return {
    	create
    }
})());