/*
 *  Copyright 2011 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms;

import javax.jcr.NoSuchWorkspaceException;
import javax.jcr.RepositoryException;
import javax.jcr.Session;

import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.cms.content.archive.ArchiveConstants;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.ModifiableTraversableAmetysObject;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.jcr.DefaultTraversableAmetysObject;
import org.ametys.plugins.repository.provider.AbstractRepository;
import org.ametys.plugins.repository.provider.JackrabbitRepository;

/**
 * CMS plugin init class.
 */
public class Init extends AbstractLogEnabled implements org.ametys.runtime.plugin.Init, Serviceable
{
    /** The contents root node name */
    private static final String __CONTENTS_ROOT_NODE = RepositoryConstants.NAMESPACE_PREFIX + ":contents";
    
    /** The resources root node name */
    private static final String __RESOURCES_ROOT_NODE = RepositoryConstants.NAMESPACE_PREFIX + ":resources";
    
    /** The plugins root node name */
    private static final String __PLUGINS_ROOT_NODE = RepositoryConstants.NAMESPACE_PREFIX + ":plugins";
    
    /** The repository */
    protected JackrabbitRepository _repository;
    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _repository = (JackrabbitRepository) manager.lookup(AbstractRepository.ROLE);
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    public void init() throws Exception
    {
        DefaultTraversableAmetysObject root = _resolver.resolveByPath("/");
        
        createPluginsRootNode(root);
        createArchiveWorkspace();
        
        // FIXME To move in a Init class for the project himself ?
        createContentsRootNode(root);
        
        // FIXME To move in a Init class for the project himself ?
        createResourcesRootNode(root);
        
        root.getNode().getSession().logout();
    }
    
    /**
     * Create the JCR root node for plugins storage
     * @param rootNode The JCR root node
     * @throws RepositoryException Thrown if the node cannot be created
     */
    protected void createPluginsRootNode (ModifiableTraversableAmetysObject rootNode) throws RepositoryException
    {
        if (!rootNode.hasChild(__PLUGINS_ROOT_NODE))
        {
            rootNode.createChild(__PLUGINS_ROOT_NODE, "ametys:unstructured");
            rootNode.saveChanges();
        }
    }
    
    /**
     * Create the archive workspace.
     * @throws RepositoryException if a repository error occurred.
     */
    protected void createArchiveWorkspace() throws RepositoryException
    {
        Session session = null;
        Session archiveSession = null;
        
        try
        {
            // Try to login 
            archiveSession = _repository.login(ArchiveConstants.ARCHIVE_WORKSPACE);
        }
        catch (NoSuchWorkspaceException e)
        {
            getLogger().info("Archives workspace does not exist, creating it...");
            
            session = _repository.login();
            session.getWorkspace().createWorkspace(ArchiveConstants.ARCHIVE_WORKSPACE);
            archiveSession = _repository.login(ArchiveConstants.ARCHIVE_WORKSPACE);

            archiveSession.getRootNode().addNode(AmetysObjectResolver.ROOT_REPO, AmetysObjectResolver.ROOT_TYPE);
            archiveSession.save();
        }
        finally
        {
            if (session != null)
            {
                session.logout();
            }
            if (archiveSession != null)
            {
                archiveSession.logout();
            }
        }
    }
    
    /**
     * Create the JCR root node for contents
     * @param rootNode The JCR root node
     * @throws RepositoryException if a repository error occurred.
     */
    protected void createContentsRootNode (ModifiableTraversableAmetysObject rootNode) throws RepositoryException
    {
        if (!rootNode.hasChild(__CONTENTS_ROOT_NODE))
        {
            rootNode.createChild(__CONTENTS_ROOT_NODE, "ametys:collection");
            rootNode.saveChanges();
        }
    }
    
    
    /**
     * Create the JCR root node for the resources
     * @param rootNode The JCR root node
     * @throws RepositoryException if a repository error occurred.
     */
    protected void createResourcesRootNode(ModifiableTraversableAmetysObject rootNode) throws RepositoryException
    {
        if (!rootNode.hasChild(__RESOURCES_ROOT_NODE))
        {
            rootNode.createChild(__RESOURCES_ROOT_NODE, "ametys:resources-collection");
            rootNode.saveChanges();
        }
    }
    
    
    
    
}
