/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.commons.lang.StringUtils;

import org.ametys.cms.transformation.URIResolver;
import org.ametys.runtime.workspace.WorkspaceMatcher;

/**
 * Component providing base paths for computing URIs.
 */
public class URIPrefixHandler implements Contextualizable, Component
{
    /** Avalon role */
    public static final String ROLE = URIPrefixHandler.class.getName();
    
    private Context _context;

    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    /**
     * Computes an URI prefix. Uses by various {@link URIResolver}.
     * @param absolute if the resulting prefix should be absolute.
     * @param internal if the resulting prefix should be internal.
     * @return the computes prefix.
     */
    public String computeUriPrefix(boolean absolute, boolean internal)
    {
        if (internal)
        {
            return "cocoon://";
        }
        else if (absolute)
        {
            return getAbsoluteUriPrefix();
        }
        else
        {
            return getUriPrefix();
        }
    }
    
    /**
     * Get the application root. Can be empty if the application resides in the root context. <br>
     * Use it to create a link beginning with the application root.
     * @return The application root uri.
     */
    public String getUriPrefix()
    {
        Request request = ContextHelper.getRequest(_context);
        String prefix = (String) request.getAttribute(CmsConstants.PATH_PREFIX_ATTRIBUTE);
        String workspaceURI = (String) request.getAttribute(WorkspaceMatcher.WORKSPACE_URI);
        
        return request.getContextPath() + workspaceURI + StringUtils.trimToEmpty(prefix);
    }
    
    /**
     * Get the absolutized version of the context path.<br>
     * Use it to create an absolute link beginning with the application root, 
     * for instance when sending a mail linking to the application.
     * @return The absolute context path.
     */
    public String getAbsoluteUriPrefix()
    {
        Request request = ContextHelper.getRequest(_context);
        
        String scheme = request.getScheme();
        
        int normalPort = scheme.equals("http") ? 80 : scheme.equals("https") ? 443 : -1;
        int requestPort = request.getServerPort();
        
        String uriPrefix = getUriPrefix();
        uriPrefix = request.getScheme() + "://" + request.getServerName() + (requestPort != normalPort ? ":" + requestPort : "") + uriPrefix;
        
        return uriPrefix;
    }
}
