/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.alerts;

import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.quartz.SchedulerException;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.lock.LockContentManager;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.core.schedule.Runnable;
import org.ametys.core.schedule.Runnable.FireProcess;
import org.ametys.core.schedule.Runnable.MisfirePolicy;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.core.impl.schedule.DefaultRunnable;
import org.ametys.plugins.core.schedule.Scheduler;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.data.holder.ModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.ModifiableModelLessDataHolder;
import org.ametys.plugins.repository.lock.LockAwareAmetysObject;
import org.ametys.plugins.repository.version.DataAndVersionAwareAmetysObject;
import org.ametys.plugins.repository.version.ModifiableDataAwareVersionableAmetysObject;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * This element creates a toggle button representing the reminders state.
 */
public class ContentAlertsClientSideElement extends StaticClientSideElement
{
    /** Repository content. */
    protected AmetysObjectResolver _resolver;

    /** The lock manager. */
    protected LockContentManager _lockManager;

    /** The ametys scheduler */
    protected Scheduler _scheduler;
    
    /** The service manager */
    protected ServiceManager _smanager;
    
    /** The content helper */
    protected ContentHelper _contentHelper;

    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _scheduler = (Scheduler) serviceManager.lookup(Scheduler.ROLE);
        _smanager = serviceManager;
        _resolver = (AmetysObjectResolver) serviceManager.lookup(AmetysObjectResolver.ROLE);
        _lockManager = (LockContentManager) serviceManager.lookup(LockContentManager.ROLE);
        _contentHelper = (ContentHelper) serviceManager.lookup(ContentHelper.ROLE);
    }

    enum AlertsStatus 
    {
        /** All the alerts are enabled */
        ENABLED,

        /** All the alerts are disabled */
        DISABLED,

        /** Some alerts are enabled, others are disabled */
        MIXED
    }

    /**
     * Get information on reminders state.
     * 
     * @param contentsId The list of contents' ids
     * @return informations on reminders state.
     */
    @Callable
    public Map<String, Object> getAlertsInformations(List<String> contentsId)
    {
        Map<String, Object> results = new HashMap<>();

        Long validationAlertDelay = Config.getInstance().getValue("remind.content.validation.delay");
        Long unmodifiedAlertDelay = Config.getInstance().getValue("remind.unmodified.content.delay");

        results.put("alert-enabled-contents", new ArrayList<>());
        results.put("alert-disabled-contents", new ArrayList<>());
        results.put("all-right-contents", new ArrayList<>());
        results.put("not-unlockable-contents", new ArrayList<>());

        AlertsStatus unmodifiedAlertStatus = null;
        AlertsStatus reminderAlertStatus = null;

        for (String contentId : contentsId)
        {
            Content content = _resolver.resolveById(contentId);

            if (content instanceof DataAndVersionAwareAmetysObject && content instanceof ModifiableContent)
            {
                Map<String, Object> contentParams = getContentDefaultParameters(content);

                // Test if the content is locked and cannot be unlocked by the
                // current user.
                boolean isValid = !isUnlockable(content, results);

                ModelLessDataHolder dataHolder = ((DataAndVersionAwareAmetysObject) content).getUnversionedDataHolder();
                
                boolean unmodifiedAlertEnabled = dataHolder.getValue(AlertsConstants.UNMODIFIED_ALERT_ENABLED, false);
                if (unmodifiedAlertEnabled)
                {
                    String unmodifiedAlertText = dataHolder.getValue(AlertsConstants.UNMODIFIED_ALERT_TEXT, StringUtils.EMPTY);
                    results.put("unmodifiedAlertText", unmodifiedAlertText);
                }

                unmodifiedAlertStatus = _getUnmodifiedStatusAlertsInformations(unmodifiedAlertStatus, unmodifiedAlertEnabled);

                boolean reminderEnabled = dataHolder.getValue(AlertsConstants.REMINDER_ENABLED, false);
                if (reminderEnabled)
                {
                    ZonedDateTime reminderDate = dataHolder.getValue(AlertsConstants.REMINDER_DATE, null);
                    String reminderText = dataHolder.getValue(AlertsConstants.REMINDER_TEXT, StringUtils.EMPTY);

                    results.put("reminderDate", reminderDate != null ? DateUtils.zonedDateTimeToString(reminderDate) : StringUtils.EMPTY);
                    results.put("reminderText", reminderText);
                }

                reminderAlertStatus = _getUnmodifiedStatusAlertsInformations(reminderAlertStatus, reminderEnabled);

                // Alerts enabled/disabled
                _getEnabledDisabledAlertsInformations(results, unmodifiedAlertDelay, content, contentParams, unmodifiedAlertEnabled, reminderEnabled);

                // All right
                if (isValid)
                {
                    @SuppressWarnings("unchecked")
                    List<Map<String, Object>> validContents = (List<Map<String, Object>>) results.get("all-right-contents");
                    validContents.add(contentParams);
                }
            }
        }

        results.put("reminderEnabled", reminderAlertStatus == AlertsStatus.MIXED ? null : reminderAlertStatus == AlertsStatus.ENABLED);
        results.put("unmodifiedAlertEnabled", unmodifiedAlertStatus == AlertsStatus.MIXED ? null : unmodifiedAlertStatus == AlertsStatus.ENABLED);

        results.put("validation-alert-delay", validationAlertDelay);
        results.put("unmodified-alert-delay", unmodifiedAlertDelay);

        return results;
    }

    private AlertsStatus _getUnmodifiedStatusAlertsInformations(AlertsStatus unmodifiedAlertStatus, boolean unmodifiedAlertEnabled)
    {
        AlertsStatus localUnmodifiedAlertStatus = unmodifiedAlertStatus;
        if (localUnmodifiedAlertStatus == null)
        {
            // Initialize the alert status with first content
            localUnmodifiedAlertStatus = unmodifiedAlertEnabled ? AlertsStatus.ENABLED : AlertsStatus.DISABLED;
        }
        else if (localUnmodifiedAlertStatus == AlertsStatus.ENABLED && !unmodifiedAlertEnabled
                || localUnmodifiedAlertStatus == AlertsStatus.DISABLED && unmodifiedAlertEnabled)
        {
            // Alert status is different for at least one content
            localUnmodifiedAlertStatus = AlertsStatus.MIXED;
        }
        return localUnmodifiedAlertStatus;
    }

    private void _getEnabledDisabledAlertsInformations(Map<String, Object> results, Long unmodifiedAlertDelay, Content content, Map<String, Object> contentParams,
            boolean unmodifiedAlertEnabled, boolean reminderEnabled)
    {
        if (unmodifiedAlertEnabled && unmodifiedAlertDelay > 0 || reminderEnabled)
        {
            I18nizableText ed = (I18nizableText) this._script.getParameters().get("alerts-enabled-description");
            I18nizableText msg = new I18nizableText(ed.getCatalogue(), ed.getKey(), Collections.singletonList(_contentHelper.getTitle(content)));

            contentParams.put("description", msg);

            @SuppressWarnings("unchecked")
            List<Map<String, Object>> enabledAlerts = (List<Map<String, Object>>) results.get("alert-enabled-contents");
            enabledAlerts.add(contentParams);
        }
        else
        {
            I18nizableText ed = (I18nizableText) this._script.getParameters().get("alerts-disabled-description");
            I18nizableText msg = new I18nizableText(ed.getCatalogue(), ed.getKey(), Collections.singletonList(_contentHelper.getTitle(content)));

            contentParams.put("description", msg);

            @SuppressWarnings("unchecked")
            List<Map<String, Object>> disabledAlerts = (List<Map<String, Object>>) results.get("alert-disabled-contents");
            disabledAlerts.add(contentParams);
        }
    }

    /**
     * Test if the content is locked and cannot be unlocked by the current user.
     * <br>
     * If this is so, fill the "not-unlockable" list in the result map.
     * 
     * @param content the content to test.
     * @param results the result map.
     * @return true if the content is locked and cannot be unlocked by the
     *         current user, false otherwise.
     */
    @SuppressWarnings("unchecked")
    protected boolean isUnlockable(Content content, Map<String, Object> results)
    {
        boolean isUnlockable = false;

        if (content instanceof LockAwareAmetysObject)
        {
            LockAwareAmetysObject lockAwareContent = (LockAwareAmetysObject) content;

            if (lockAwareContent.isLocked() && !_lockManager.canUnlock(lockAwareContent))
            {
                Map<String, Object> contentParams = getContentDefaultParameters(content);

                I18nizableText ed = (I18nizableText) this._script.getParameters().get("not-unlockable-description");
                I18nizableText msg = new I18nizableText(ed.getCatalogue(), ed.getKey(), Collections.singletonList(_contentHelper.getTitle(content)));

                contentParams.put("description", msg);

                List<Map<String, Object>> notUnlockable = (List<Map<String, Object>>) results.get("not-unlockable-contents");
                notUnlockable.add(contentParams);

                isUnlockable = true;
            }
        }

        return isUnlockable;
    }

    /**
     * Get the default content's parameters
     * 
     * @param content The content
     * @return The default parameters
     */
    protected Map<String, Object> getContentDefaultParameters(Content content)
    {
        Map<String, Object> contentParams = new HashMap<>();
        contentParams.put("id", content.getId());
        contentParams.put("title", _contentHelper.getTitle(content));

        return contentParams;
    }

    /**
     * Set alerts on content
     * 
     * @param contentIds the content's id
     * @param params the alerts' parameters
     * @throws ServiceException if an error occurred while retrieving the alert
     *             scheduler component
     * @throws SchedulerException if an error occurred while computing the runnable id 
     */
    @Callable
    public void setAlertsOnContent(List<String> contentIds, Map<String, Object> params) throws ServiceException, SchedulerException 
    {
        for (String contentId : contentIds)
        {
            Content content = _resolver.resolveById(contentId);
            if (content instanceof ModifiableDataAwareVersionableAmetysObject)
            {
                _setAlerts((ModifiableDataAwareVersionableAmetysObject) content, params);
            }
        }

        boolean instantAlertEnabled = (Boolean) params.get("instantAlertEnabled");
        if (instantAlertEnabled)
        {
            // get the schedulable id provided by the caller. Or the CMS alertSchedulable if none.
            String schedulableId = params.containsKey("schedulable-id") ? (String) params.get("schedulable-id") : AlertSchedulable.SCHEDULABLE_ID;

            String instantAlertText = (String) params.get("instantAlertText");
            Map<String, Object> jobParams = new HashMap<>();
            jobParams.put(AlertSchedulable.JOBDATAMAP_INSTANT_MODE_KEY, true);
            jobParams.put(AlertSchedulable.JOBDATAMAP_CONTENT_IDS_KEY, contentIds);
            jobParams.put(AlertSchedulable.JOBDATAMAP_MESSAGE_KEY, instantAlertText);
            
            List<String> i18nParams = new ArrayList<>();
            i18nParams.add(StringUtils.join(contentIds, ", "));
            
            Runnable runnable = new DefaultRunnable(_generateRunnableId(), 
                    new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENTS_INSTANT_ALERTS_RUNNABLE_LABEL", i18nParams),
                    new I18nizableText("plugin.cms", contentIds.size() <= 1 ? "PLUGINS_CMS_CONTENTS_INSTANT_ALERTS_RUNNABLE_DESCRIPTION" : "PLUGINS_CMS_CONTENTS_INSTANT_ALERTS_RUNNABLE_DESCRIPTION_MULTIPLE" , i18nParams),
                    FireProcess.NOW,
                    null /* cron*/, 
                    schedulableId, 
                    true /* removable */, 
                    false /* modifiable */, 
                    false /* deactivatable */, 
                    MisfirePolicy.FIRE_ONCE, 
                    true /* isVolatile */, 
                    _currentUserProvider.getUser(), 
                    jobParams
                );
            
            try
            {
                _scheduler.scheduleJob(runnable);
            }
            catch (SchedulerException e)
            {
                getLogger().error("An error occured while trying to schedule the sending of alerts", e);
            }
        }
    }

    private String _generateRunnableId()
    {
        return "org.ametys.cms.alerts.InstantAlertRunnable" + "$" + UUID.randomUUID();
    }

    /**
     * Sets the alerts on the specified content.
     * 
     * @param content the content to set the alerts on.
     * @param params the alerts' parameters
     * @throws AmetysRepositoryException if a repository error occurs.
     */
    protected void _setAlerts(ModifiableDataAwareVersionableAmetysObject content, Map<String, Object> params) throws AmetysRepositoryException
    {
        ModifiableModelLessDataHolder dataHolder = content.getUnversionedDataHolder();

        // Set alert for unmodified contents
        if (params.get("unmodifiedAlertEnabled") != null)
        {
            boolean unmodifiedAlertEnabled = (Boolean) params.get("unmodifiedAlertEnabled");
            String unmodifiedAlertText = (String) params.get("unmodifiedAlertText");

            dataHolder.setValue(AlertsConstants.UNMODIFIED_ALERT_ENABLED, unmodifiedAlertEnabled);
            dataHolder.setValue(AlertsConstants.UNMODIFIED_ALERT_TEXT, StringUtils.trimToEmpty(unmodifiedAlertText));
        }

        // Set reminders
        if (params.get("reminderEnabled") != null)
        {
            boolean reminderEnabled = (Boolean) params.get("reminderEnabled");
            dataHolder.setValue(AlertsConstants.REMINDER_ENABLED, reminderEnabled);

            String reminderDateStr = (String) params.get("reminderDate");
            String reminderText = (String) params.get("reminderText");
            dataHolder.setValue(AlertsConstants.REMINDER_TEXT, StringUtils.trimToEmpty(reminderText));

            // Parse the reminder date.
            if (StringUtils.isNotBlank(reminderDateStr))
            {
                LocalDate localDate = DateUtils.parseLocalDate(reminderDateStr);
                if (localDate != null)
                {
                    ZonedDateTime zonedDateTime = DateUtils.asZonedDateTime(localDate, null);
                    dataHolder.setValue(AlertsConstants.REMINDER_DATE, zonedDateTime);
                }
                else
                {
                    getLogger().error("Unable to parse reminder date " + reminderDateStr);
                }
            }
        }

        content.saveChanges();
    }

    @Override
    public List<Script> getScripts(boolean ignoreRights, Map<String, Object> contextParameters)
    {
        boolean enabled = Config.getInstance().getValue("remind.content.enabled");
        if (enabled)
        {
            return super.getScripts(ignoreRights, contextParameters);
        }

        return new ArrayList<>();
    }
}
