/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang.StringUtils;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.repository.Content;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.plugins.repository.lock.LockHelper;
import org.ametys.plugins.repository.lock.LockableAmetysObject;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Abstract client side element for contents.
 */
public abstract class AbstractContentClientSideElement extends StaticClientSideElement
{
    /** Runtime users manager */
    protected UserManager _userManager;
    /** The content helper */
    protected ContentHelper _contentHelper;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _userManager = (UserManager) smanager.lookup(UserManager.ROLE);
        _contentHelper = (ContentHelper) smanager.lookup(ContentHelper.ROLE);
    }
    
    /**
     * Get the default content's parameters
     * @param content The content
     * @return The default parameters
     */
    protected Map<String, Object> getContentDefaultParameters (Content content)
    {
        Map<String, Object> contentParams = new HashMap<>();
        contentParams.put("id", content.getId());
        contentParams.put("title", _contentHelper.getTitle(content));
        
        return contentParams;
    }
    
    /**
     * Get the description when user has no right on content
     * @param content the content
     * @return the description
     */
    protected I18nizableText getNoRightContentDescription (Content content)
    {
        List<String> i18nParameters = new ArrayList<>();
        i18nParameters.add(_contentHelper.getTitle(content));
        
        I18nizableText ed = (I18nizableText) this._script.getParameters().get("noright-content-description");
        return new I18nizableText(ed.getCatalogue(), ed.getKey(), i18nParameters);
    }
    
    /**
     * Get the description when content is not modifiable
     * @param content the content
     * @return the description
     */
    protected I18nizableText getNoModifiableContentDescription (Content content)
    {
        List<String> i18nParameters = new ArrayList<>();
        i18nParameters.add(_contentHelper.getTitle(content));
        
        I18nizableText ed = (I18nizableText) this._script.getParameters().get("nomodifiable-content-description");
        return new I18nizableText(ed.getCatalogue(), ed.getKey(), i18nParameters);
    }
    
    /**
     * Get the description when content is locked
     * @param content the content
     * @return the description
     */
    protected I18nizableText getLockedContentDescription (Content content)
    {
        UserIdentity currentLockerIdentity = ((LockableAmetysObject) content).getLockOwner();
        User currentLocker = currentLockerIdentity != null ? _userManager.getUser(currentLockerIdentity.getPopulationId(), currentLockerIdentity.getLogin()) : null;
        
        List<String> lockI18nParameters = new ArrayList<>();
        lockI18nParameters.add(_contentHelper.getTitle(content));
        lockI18nParameters.add(currentLocker != null ? currentLocker.getFullName() : "");
        lockI18nParameters.add(currentLockerIdentity != null ? currentLockerIdentity.getLogin() : "Anonymous");
        
        I18nizableText ed = (I18nizableText) this._script.getParameters().get("locked-content-description");
        return new I18nizableText(ed.getCatalogue(), ed.getKey(), lockI18nParameters);
    }
    
    /**
     * Get i18n description for all right content
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    protected I18nizableText getAllRightContentDescription (Content content)
    {
        List<String> i18nParameters = new ArrayList<>();
        i18nParameters.add(_contentHelper.getTitle(content));
        
        I18nizableText ed = (I18nizableText) this._script.getParameters().get("allright-content-description");
        return new I18nizableText(ed.getCatalogue(), ed.getKey(), i18nParameters);
    }
    
    /**
     * Determines if a content is locked
     * @param content the content
     * @return true if the content is locked.
     */
    protected boolean isLocked(Content content)
    {
        if (content instanceof LockableAmetysObject)
        {
            LockableAmetysObject lockableContent = (LockableAmetysObject) content;
            if (lockableContent.isLocked() && !LockHelper.isLockOwner(lockableContent, _currentUserProvider.getUser()))
            {
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * Determines if user has convenient right on content
     * @param content The content
     * @return true if the user has convenient right
     */
    protected boolean hasRight (Content content)
    {
        if (_rights.isEmpty())
        {
            return true;
        }
        
        Set<String> rightsToCheck = _rights.keySet();
        for (String rightToCheck : rightsToCheck)
        {
            if (StringUtils.isNotEmpty(rightToCheck))
            {
                if (_rightManager.hasRight(_currentUserProvider.getUser(), rightToCheck, content) == RightResult.RIGHT_ALLOW)
                {
                    return true;
                }
            }
        }
        
        return false;
    }
}
