/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.content.references.OutgoingReferences;
import org.ametys.cms.repository.Content;
import org.ametys.cms.transformation.ConsistencyChecker;
import org.ametys.cms.transformation.ConsistencyChecker.CHECK;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * This element creates a toggle button representing the consistency state.
 */
public class ContentConsistencyTestClientSideElement extends StaticClientSideElement
{
    /** Repository content */
    protected AmetysObjectResolver _resolver;
    /** The consistency checker */
    protected ConsistencyChecker _consistencyChecker;
    /** The content helper */
    protected ContentHelper _contentHelper;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _consistencyChecker = (ConsistencyChecker) smanager.lookup(ConsistencyChecker.ROLE);
        _contentHelper = (ContentHelper) smanager.lookup(ContentHelper.ROLE);
    }
    
    /**
     * Check the consistency of contents
     * @param contentsId The ids of contents to test
     * @param shortTest true to make a short test, that means that long tests will return UNKNOWN immediately
     * @return the consistency test results
     */
    @Callable
    public Map<String, Object> checkConsistency(List<String> contentsId, boolean shortTest)
    {
        Map<String, Object> results = new HashMap<>();
        
        results.put("fully-ok-contents", new ArrayList<>());
        results.put("mainly-ok-contents", new ArrayList<>());
        results.put("not-ok-contents", new ArrayList<>());
        
        for (String contentId : contentsId)
        {
            Content content = _resolver.resolveById(contentId);
            
            Map<String, OutgoingReferences> referencesByPath = content.getOutgoingReferences();
            
            if (referencesByPath.isEmpty())
            {
                Map<String, Object> contentParams = getContentDefaultParameters(content);
                contentParams.put("description", _getOkDescription(content));
                
                @SuppressWarnings("unchecked")
                List<Map<String, Object>> okContents = (List<Map<String, Object>>) results.get("fully-ok-contents");
                okContents.add(contentParams);
            }
            else
            {
                int failure = 0;
                int unknown = 0;
                
                for (String dataPath : referencesByPath.keySet())
                {
                    OutgoingReferences references = referencesByPath.get(dataPath);
                    for (String referenceType : references.keySet())
                    {
                        for (String referenceValue : references.get(referenceType))
                        {
                            CHECK check = CHECK.SERVER_ERROR;
                            try
                            {
                                check = _consistencyChecker.checkConsistency(referenceType, referenceValue, content.getId(), dataPath, shortTest).status();
                            }
                            catch (Exception e)
                            {
                                // Ignore, consider it a failure.
                            }
                            
                            switch (check)
                            {
                                case SUCCESS:
                                    break;
                                case UNKNOWN:
                                    unknown++;
                                    break;
                                case NOT_FOUND:
                                case UNAUTHORIZED:
                                case SERVER_ERROR:
                                default:
                                    failure++;
                                    break;
                            }
                        }
                    }
                }
                
                if (failure > 0)
                {
                    Map<String, Object> contentParams = getContentDefaultParameters(content);
                    contentParams.put("description", _getNotOkDescription(content));
                    
                    @SuppressWarnings("unchecked")
                    List<Map<String, Object>> notOkContents = (List<Map<String, Object>>) results.get("not-ok-contents");
                    notOkContents.add(contentParams);
                }
                else if (unknown > 0)
                {
                    Map<String, Object> contentParams = getContentDefaultParameters(content);
                    contentParams.put("description", _getMainlyOkDescription(content));
                    
                    @SuppressWarnings("unchecked")
                    List<Map<String, Object>> mainlyOkContents = (List<Map<String, Object>>) results.get("mainly-ok-contents");
                    mainlyOkContents.add(contentParams);
                }
                else 
                {
                    Map<String, Object> contentParams = getContentDefaultParameters(content);
                    contentParams.put("description", _getOkDescription(content));
                    
                    @SuppressWarnings("unchecked")
                    List<Map<String, Object>> okContents = (List<Map<String, Object>>) results.get("fully-ok-contents");
                    okContents.add(contentParams);
                }
            }
        }

        return results;
    }
    
    /**
     * Get the default content's parameters
     * @param content The content
     * @return The default parameters
     */
    protected Map<String, Object> getContentDefaultParameters(Content content)
    {
        Map<String, Object> contentParams = new HashMap<>();
        contentParams.put("id", content.getId());
        contentParams.put("title", _contentHelper.getTitle(content));
        
        return contentParams;
    }
    
    /**
     * Get content i18n description when the consistency check has succeed
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    protected I18nizableText _getOkDescription(Content content)
    {
        List<String> i18nParameters = new ArrayList<>();
        i18nParameters.add(_contentHelper.getTitle(content));
        
        I18nizableText ed = (I18nizableText) this._script.getParameters().get("consistency-ok-content-description");
        return new I18nizableText(ed.getCatalogue(), ed.getKey(), i18nParameters);
    }
    
    /**
     * Get content i18n description when the consistency check has failed for at least one link
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    protected I18nizableText _getNotOkDescription(Content content)
    {
        List<String> i18nParameters = new ArrayList<>();
        i18nParameters.add(_contentHelper.getTitle(content));
        
        I18nizableText ed = (I18nizableText) this._script.getParameters().get("consistency-error-content-description");
        return new I18nizableText(ed.getCatalogue(), ed.getKey(), i18nParameters);
    }
    
    /**
     * Get content i18n description when the consistency check returned at least one unknown link
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    protected I18nizableText _getMainlyOkDescription(Content content)
    {
        List<String> i18nParameters = new ArrayList<>();
        i18nParameters.add(_contentHelper.getTitle(content));
        
        I18nizableText ed = (I18nizableText) this._script.getParameters().get("consistency-okbut-content-description");
        return new I18nizableText(ed.getCatalogue(), ed.getKey(), i18nParameters);
    }
}
