/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.TreeMap;
import java.util.stream.Collectors;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.content.RootContentHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.languages.LanguagesManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.ui.ClientSideElement;
import org.ametys.core.ui.SimpleMenu;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.I18nUtils;
import org.ametys.core.util.I18nizableTextKeyComparator;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * This element creates a menu with one gallery item per content type classified by category (default mode).
 *  
 * This element supports an alternative display, to display content types into sub menu items (root menu items will be the categories),
 * if 'show-in-menu' parameters is set to 'true'. This mode should be privileged if the number of content types is important (&gt; 30)
 * 
 * The user rights are checked. 
 */
public class ContentTypesGallery extends SimpleMenu
{
    /** The list of content types */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    /** The language manager */
    protected LanguagesManager _languagesManager;
    
    /** The i18n utils */
    protected I18nUtils _i18nUtils;
    
    /** Helper for root content */
    protected RootContentHelper _rootContentHelper;
    
    private boolean _contentTypesInitialized;
    
    private Set<String> _addedComponents;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
        _i18nUtils = (I18nUtils) smanager.lookup(I18nUtils.ROLE);
        _languagesManager = (LanguagesManager) smanager.lookup(LanguagesManager.ROLE);
        _rootContentHelper = (RootContentHelper) smanager.lookup(RootContentHelper.ROLE);
    }
    
    @Override
    protected Script _configureScript(Configuration configuration) throws ConfigurationException
    {
        Script script = super._configureScript(configuration);
        
        // Add the CSS brought by content types
        for (String id: _contentTypeExtensionPoint.getExtensionsIds())
        {
            ContentType contentType = _contentTypeExtensionPoint.getExtension(id);
            script.getCSSFiles().addAll(contentType.getCSSFiles());
        }
        
        return script;
    }
    
    @Override
    protected void _getGalleryItems(Map<String, Object> parameters, Map<String, Object> contextualParameters)
    {
        if (!_showInMenu())
        {
            // In this mode (default mode), the available content types will be displayed into gallery items, classified by category

            try
            {
                _lazyInitializeContentTypeGallery();
            }
            catch (Exception e)
            {
                throw new IllegalStateException("Unable to lookup client side element local components", e);
            }
            
            if (_galleryItems.size() > 0)
            {
                parameters.put("gallery-item", new LinkedHashMap<>());
                
                for (GalleryItem galleryItem : _galleryItems)
                {
                    @SuppressWarnings("unchecked")
                    Map<String, Object> galleryItems = (Map<String, Object>) parameters.get("gallery-item");
                    galleryItems.put("gallery-groups", new ArrayList<>());
                    
                    for (GalleryGroup galleryGp : galleryItem.getGroups())
                    {
                        @SuppressWarnings("unchecked")
                        List<Object> galleryGroups = (List<Object>) galleryItems.get("gallery-groups");
                        
                        Map<String, Object> groupParams = new LinkedHashMap<>();
                        
                        I18nizableText label = galleryGp.getLabel();
                        groupParams.put("label", label);
                        
                        // Group items sorted by alphabetical order
                        List<ClientSideElement> gpItems = new ArrayList<>();
                        for (ClientSideElement element : galleryGp.getItems())
                        {
                            String cTypeId = element.getId().substring(this.getId().length() + 1);
                            ContentType cType = _contentTypeExtensionPoint.getExtension(cTypeId);
                            
                            if (hasRight(cType))
                            {
                                gpItems.add(element);
                            }
                        }
                        
                        if (gpItems.size() > 0)
                        {
                            List<String> gpItemIds = gpItems.stream().map(e -> e.getId()).collect(Collectors.toList());
                            groupParams.put("items", gpItemIds);
                            galleryGroups.add(groupParams);
                        }
                    }
                }
            }
        }
    }
    
    private boolean _showInMenu()
    {
        return _script.getParameters().containsKey("show-in-menu") && _script.getParameters().get("show-in-menu").equals("true");
    }
    
    private synchronized void _lazyInitializeContentTypeGallery() throws ConfigurationException
    {
        if (!_contentTypesInitialized)
        {
            _addedComponents = new HashSet<>();
            
            Map<I18nizableText, Set<ContentType>> cTypesByGroup = _getContentTypesByGroup();
            
            if (cTypesByGroup.size() > 0)
            {
                GalleryItem galleryItem = new GalleryItem();
                
                for (I18nizableText groupLabel : cTypesByGroup.keySet())
                {
                    GalleryGroup galleryGroup = new GalleryGroup(groupLabel);
                    galleryItem.addGroup(galleryGroup);
                    
                    Set<ContentType> cTypes = cTypesByGroup.get(groupLabel);
                    for (ContentType cType : cTypes)
                    {
                        String id = this.getId() + "-" + cType.getId();
                        
                        DefaultConfiguration conf = new DefaultConfiguration("extension");
                        conf.setAttribute("id", id);
                        _addContentTypeConfiguration (conf, cType);
                        
                        if (!_addedComponents.contains(id))
                        {
                            _getGalleryItemManager().addComponent(_pluginName, null, id, StaticClientSideElement.class, conf);
                            galleryGroup.addItem(new UnresolvedItem(id, true));
                            _addedComponents.add(id);
                        }
                    }
                }
                
                _galleryItems.add(galleryItem);
            }
            
            if (_galleryItems.size() > 0)
            {
                try
                {
                    _getGalleryItemManager().initialize();
                }
                catch (Exception e)
                {
                    throw new ConfigurationException("Unable to lookup parameter local components", e);
                }
            }
        }
        
        _contentTypesInitialized = true;
    }
    
    private synchronized void _lazyInitializeContentTypeMenu()
    {
        if (!_contentTypesInitialized)
        {
            _addedComponents = new HashSet<>();
            
            Map<I18nizableText, Set<ContentType>> cTypesByGroup = _getContentTypesByGroup();
            
            if (cTypesByGroup.size() > 0)
            {
                for (I18nizableText groupLabel : cTypesByGroup.keySet())
                {
                    String id = this.getId() + "-" + org.ametys.core.util.StringUtils.generateKey();
                    DefaultConfiguration conf = new DefaultConfiguration("extension");
                    conf.setAttribute("id", id);
                    
                    Set<ContentType> cTypes = cTypesByGroup.get(groupLabel);
                    _addGroupContentTypeConfiguration (conf, groupLabel, cTypes);
                    
                    if (!_addedComponents.contains(id))
                    {
                        _menuItemManager.addComponent(_pluginName, null, id, SimpleMenu.class, conf);
                        _unresolvedMenuItems.add(new UnresolvedItem(id, true));
                        _addedComponents.add(id);
                    }
                }
            }
        }
        
        _contentTypesInitialized = true;
    }
    
    /**
     * Get the configuration of the content type item
     * @param rootConf The root configuration
     * @param groupLabel The group's label
     * @param cTypes The content types of the group
     */
    protected void _addGroupContentTypeConfiguration (DefaultConfiguration rootConf, I18nizableText groupLabel, Set<ContentType> cTypes)
    {
        DefaultConfiguration classConf = new DefaultConfiguration("class");
        classConf.setAttribute("name", "Ametys.plugins.cms.content.controller.ContentTypeMenuItemController");
        
        // Label and description
        DefaultConfiguration labelConf = new DefaultConfiguration("label");
        DefaultConfiguration descConf = new DefaultConfiguration("description");
        if (groupLabel.isI18n())
        {
            labelConf.setAttribute("i18n", "true");
            labelConf.setValue(groupLabel.getCatalogue() + ":" + groupLabel.getKey());
            descConf.setAttribute("i18n", "true");
            descConf.setValue(groupLabel.getCatalogue() + ":" + groupLabel.getKey());
        }
        else
        {
            labelConf.setValue(groupLabel.getLabel());
            descConf.setValue(groupLabel.getLabel());
        }
        classConf.addChild(labelConf);
        classConf.addChild(descConf);
        
        // Icons or glyph (use the first content type of the list)
        ContentType firstCtype = cTypes.iterator().next();
        if (firstCtype.getIconGlyph() != null)
        {
            DefaultConfiguration iconGlyphConf = new DefaultConfiguration("icon-glyph");
            iconGlyphConf.setValue(firstCtype.getIconGlyph());
            classConf.addChild(iconGlyphConf);
        }
        if (firstCtype.getIconDecorator() != null)
        {
            DefaultConfiguration iconDecoratorConf = new DefaultConfiguration("icon-decorator");
            iconDecoratorConf.setValue(firstCtype.getIconDecorator());
            classConf.addChild(iconDecoratorConf);
        }
        if (firstCtype.getSmallIcon() != null)
        {
            DefaultConfiguration iconSmallConf = new DefaultConfiguration("icon-small");
            iconSmallConf.setValue(firstCtype.getSmallIcon());
            classConf.addChild(iconSmallConf);
            DefaultConfiguration iconMediumConf = new DefaultConfiguration("icon-medium");
            iconMediumConf.setValue(firstCtype.getMediumIcon());
            classConf.addChild(iconMediumConf);
            DefaultConfiguration iconLargeConf = new DefaultConfiguration("icon-large");
            iconLargeConf.setValue(firstCtype.getLargeIcon());
            classConf.addChild(iconLargeConf);
        }
        
        // Common configuration
        @SuppressWarnings("unchecked")
        Map<String, Object> commonConfig = (Map<String, Object>) this._script.getParameters().get("group-items-config");
        for (String tagName : commonConfig.keySet())
        {
            DefaultConfiguration c = new DefaultConfiguration(tagName);
            c.setValue(String.valueOf(commonConfig.get(tagName)));
            classConf.addChild(c);
        }
        
        rootConf.addChild(classConf);
        
        // Menu items
        DefaultConfiguration menuItemsConf = new DefaultConfiguration("menu-items");
        
        for (ContentType contentType : cTypes)
        {
            DefaultConfiguration menuItemConf = new DefaultConfiguration("item");
            menuItemConf.setAttribute("id", this.getId() + "-" + contentType.getId());
            _addContentTypeConfiguration (menuItemConf, contentType);
            
            menuItemsConf.addChild(menuItemConf);
        }
        
        rootConf.addChild(menuItemsConf);
    }
    
    /**
     * Get the configuration of the content type item
     * @param rootConf The root configuration
     * @param cType The content type
     */
    protected void _addContentTypeConfiguration (DefaultConfiguration rootConf, ContentType cType)
    {
        DefaultConfiguration classConf = new DefaultConfiguration("class");
        classConf.setAttribute("name", "Ametys.ribbon.element.ui.ButtonController");
        
        // enable toggle
        DefaultConfiguration enableToggleConf = new DefaultConfiguration("toggle-enabled");
        enableToggleConf.setValue("true");
        classConf.addChild(enableToggleConf);

        // Parent controller id
        DefaultConfiguration parentIdConf = new DefaultConfiguration("controllerParentId");
        parentIdConf.setValue(this.getId());
        classConf.addChild(parentIdConf);
        
        // Label
        DefaultConfiguration labelConf = _getI18nizableTextConfiguration("label", cType.getLabel());
        classConf.addChild(labelConf);
        
        // Description
        DefaultConfiguration descConf = _getI18nizableTextConfiguration("description", cType.getDescription());
        classConf.addChild(descConf);
        
        // Default content title
        DefaultConfiguration defaultTitleConf = _getI18nizableTextConfiguration("defaultContentTitle", cType.getDefaultTitle());
        classConf.addChild(defaultTitleConf);
        
        // Content type
        DefaultConfiguration typeConf = new DefaultConfiguration("contentTypes");
        typeConf.setValue(cType.getId());
        classConf.addChild(typeConf);
        
        // Icons or glyph
        _addContentTypeIconsConfiguration(classConf, cType);
        
        // Common configuration
        @SuppressWarnings("unchecked")
        Map<String, Object> commonConfig = (Map<String, Object>) this._script.getParameters().get("items-config");
        for (String tagName : commonConfig.keySet())
        {
            DefaultConfiguration c = new DefaultConfiguration(tagName);
            c.setValue(String.valueOf(commonConfig.get(tagName)));
            classConf.addChild(c);
        }
        
        rootConf.addChild(classConf);
        
        _addRightsOnContentTypeConfiguration(rootConf);
    }
    
    /**
     * Get the configuration for an i18nizable text
     * @param tagName the tag name
     * @param i18nText the i18n text
     * @return the configuration for i18nizable text
     */
    protected DefaultConfiguration _getI18nizableTextConfiguration(String tagName, I18nizableText i18nText)
    {
        DefaultConfiguration i18nConf = new DefaultConfiguration(tagName);
        if (i18nText.isI18n())
        {
            i18nConf.setAttribute("i18n", "true");
            i18nConf.setValue(i18nText.getCatalogue() + ":" + i18nText.getKey());
        }
        else
        {
            i18nConf.setValue(i18nText.getLabel());
        }
        return i18nConf;
    }
    
    /**
     * Get the 'rights' configuration of the content type item
     * @param rootConf The root configuration
     */
    protected void _addRightsOnContentTypeConfiguration(DefaultConfiguration rootConf)
    {
        DefaultConfiguration rightsConf = new DefaultConfiguration("rights");
        rightsConf.setAttribute("mode", _rightsMode);
        for (String rightId : _rights.keySet())
        {
            DefaultConfiguration rightConf = new DefaultConfiguration("right");
            Optional.ofNullable(_rights.get(rightId)).ifPresent(contextPrefix -> rightConf.setAttribute("context-prefix", contextPrefix));
            rightConf.setValue(rightId);
            rightsConf.addChild(rightConf);
        }
        rootConf.addChild(rightsConf);
    }
    
    /**
     * Add configuration for content type's icons and/or glyphes
     * @param classConf the class configuration
     * @param cType the content type
     */
    protected void _addContentTypeIconsConfiguration(DefaultConfiguration classConf, ContentType cType)
    {
        if (cType.getIconGlyph() != null)
        {
            DefaultConfiguration iconGlyphConf = new DefaultConfiguration("icon-glyph");
            iconGlyphConf.setValue(cType.getIconGlyph());
            classConf.addChild(iconGlyphConf);
        }
        if (cType.getIconDecorator() != null)
        {
            DefaultConfiguration iconDecoratorConf = new DefaultConfiguration("icon-decorator");
            iconDecoratorConf.setValue(cType.getIconDecorator());
            classConf.addChild(iconDecoratorConf);
        }
        if (cType.getSmallIcon() != null)
        {
            DefaultConfiguration iconSmallConf = new DefaultConfiguration("icon-small");
            iconSmallConf.setValue(cType.getSmallIcon());
            classConf.addChild(iconSmallConf);
            DefaultConfiguration iconMediumConf = new DefaultConfiguration("icon-medium");
            iconMediumConf.setValue(cType.getMediumIcon());
            classConf.addChild(iconMediumConf);
            DefaultConfiguration iconLargeConf = new DefaultConfiguration("icon-large");
            iconLargeConf.setValue(cType.getLargeIcon());
            classConf.addChild(iconLargeConf);
        }
    }
    
    /**
     * Get the list of content types classified by groups
     * @return The content types
     */
    protected Map<I18nizableText, Set<ContentType>> _getContentTypesByGroup ()
    {
        Map<I18nizableText, Set<ContentType>> groups = new TreeMap<>(new I18nizableTextKeyComparator());
        
        if (this._script.getParameters().get("contentTypes") != null)
        {
            String[] contentTypesIds = ((String) this._script.getParameters().get("contentTypes")).split(",");
            
            boolean allowInherit = "true".equals(this._script.getParameters().get("allowInherit"));
            
            for (String contentTypeId : contentTypesIds)
            {
                ContentType contentType = _contentTypeExtensionPoint.getExtension(contentTypeId);
                
                if (isValidContentType(contentType))
                {
                    addContentType (contentType, groups);
                }
                
                if (allowInherit)
                {
                    for (String subTypeId : _contentTypeExtensionPoint.getSubTypes(contentTypeId))
                    {
                        ContentType subContentType = _contentTypeExtensionPoint.getExtension(subTypeId);
                        if (isValidContentType(subContentType))
                        {
                            addContentType (subContentType, groups);
                        }
                    }
                }
            }
        }
        else
        {
            Set<String> contentTypesIds = _contentTypeExtensionPoint.getExtensionsIds();
            
            for (String contentTypeId : contentTypesIds)
            {
                ContentType contentType = _contentTypeExtensionPoint.getExtension(contentTypeId);
                
                if (isValidContentType(contentType))
                {
                    addContentType (contentType, groups);
                }
            }
        }
        
        return groups;
    }
    
    /**
     * Add content to groups
     * @param contentType The content type
     * @param groups The groups
     */
    protected void addContentType (ContentType contentType, Map<I18nizableText, Set<ContentType>> groups)
    {
        I18nizableText group = contentType.getCategory();
        if (group.isI18n() && group.getKey().isEmpty()
            || !group.isI18n() && group.getLabel().isEmpty())
        {
            group = new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_CREATECONTENTMENU_GROUP_10_CONTENT");
        }
        
        if (!groups.containsKey(group))
        {
            groups.put(group, new HashSet<>());
        }
        Set<ContentType> cTypes = groups.get(group);
        cTypes.add(contentType);
    }
    
    @Override
    public List<Script> getScripts(boolean ignoreRights, Map<String, Object> contextParameters)
    {
        if (_showInMenu())
        {
            // In this mode, the available content types will be displayed into menu items, classified by category (sub menu items)
            try
            {
                _lazyInitializeContentTypeMenu();
            }
            catch (Exception e)
            {
                throw new IllegalStateException("Unable to lookup client side element local components", e);
            }
        }
        
        for (String id: _contentTypeExtensionPoint.getExtensionsIds())
        {
            ContentType contentType = _contentTypeExtensionPoint.getExtension(id);
            if (isValidContentType(contentType))
            {
                return super.getScripts(ignoreRights, contextParameters);
            }
        }
        return new ArrayList<>();
    }
    
    /**
     * Determines if the content type is a valid content type for the gallery
     * @param contentType The coentent
     * @return true if it is a valid content type
     */
    protected boolean isValidContentType (ContentType contentType)
    {
        return !contentType.isAbstract() && !contentType.isPrivate() && !contentType.isMixin() && !contentType.isReferenceTable();
    }
    
    /**
     * Test if the current user has the right needed by the content type to create a content.
     * @param cType the content type
     * @return true if the user has the right needed, false otherwise.
     */
    protected boolean hasRight(ContentType cType)
    {
        String right = cType.getRight();
        
        if (right == null)
        {
            return true;
        }
        else
        {
            UserIdentity user = _currentUserProvider.getUser();
            return _rightManager.hasRight(user, right, "/cms") == RightResult.RIGHT_ALLOW || _rightManager.hasRight(user, right, _rootContentHelper.getRootContent()) == RightResult.RIGHT_ALLOW;
        }
    }
    
    /**
     * Comparator used to order the content types in the categories
     * We use the translated labels to make it easier to find a content type
     * But the order will be different for each language
     */
    class ContentTypeClientSideElementComparator implements Comparator<ClientSideElement>
    {
        String _parentMenuId;
        
        public ContentTypeClientSideElementComparator(String parentMenuId)
        {
            _parentMenuId = parentMenuId;
        }
        
        @Override
        public int compare(ClientSideElement c1, ClientSideElement c2)
        {
            if (c1 == c2)
            {
                return 0;
            }
            
            String cTypeId1 = c1.getId().substring(_parentMenuId.length() + 1);
            ContentType cType1 = _contentTypeExtensionPoint.getExtension(cTypeId1);
            I18nizableText t1 = cType1.getLabel();
            
            String cTypeId2 = c2.getId().substring(_parentMenuId.length() + 1);
            ContentType cType2 = _contentTypeExtensionPoint.getExtension(cTypeId2);
            I18nizableText t2 = cType2.getLabel();
            
            String str1 = _i18nUtils.translate(t1);
            if (str1 == null)
            {
                str1 = t1.isI18n() ? t1.getKey() : t1.getLabel();
            }
            String str2 = _i18nUtils.translate(t2);
            if (str2 == null)
            {
                str2 = t2.isI18n() ? t2.getKey() : t2.getLabel();
            }
            
            int compareTo = str1.toLowerCase().compareTo(str2.toLowerCase());
            if (compareTo == 0)
            {
                // Content types could have same labels but there are not equals, so do not return 0 to add it in TreeSet
                // Indeed, in a TreeSet implementation two elements that are equal by the method compareTo are, from the standpoint of the set, equal 
                return 1;
            }
            return compareTo;
        }
    }
}
