/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.content.RootContentHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.ui.SimpleMenu;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.I18nUtils;
import org.ametys.core.util.I18nizableTextKeyComparator;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * This element creates a menu item with one menu item per content type. 
 * The user rights are checked. 
 */
public class ContentTypesMenuItem extends SimpleMenu
{
    /** The list of content types */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    /** The i18n utils */
    protected I18nUtils _i18nUtils;
    /** Helper for root content */
    protected RootContentHelper _rootContentHelper;
    
    private boolean _contentTypesInitialized;
    
    private Set<String> _addedComponents;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
        _i18nUtils = (I18nUtils) smanager.lookup(I18nUtils.ROLE);
        _rootContentHelper = (RootContentHelper) smanager.lookup(RootContentHelper.ROLE);
    }
    
    private synchronized void _lazyInitializeContentTypeMenu()
    {
        if (!_contentTypesInitialized)
        {
            _addedComponents = new HashSet<>();
            
            Map<I18nizableText, Set<ContentType>> cTypesByGroup = _getContentTypesByGroup();
            
            if (cTypesByGroup.size() > 0)
            {
                for (I18nizableText groupLabel : cTypesByGroup.keySet())
                {
                    Set<ContentType> cTypes = cTypesByGroup.get(groupLabel);
                    for (ContentType cType : cTypes)
                    {
                        String id = this.getId() + "-" + cType.getId();
                        
                        Configuration conf = _getContentTypeConfiguration (id, cType);
                        if (!_addedComponents.contains(id))
                        {
                            _menuItemManager.addComponent(_pluginName, null, id, StaticClientSideElement.class, conf);
                            _unresolvedMenuItems.add(new UnresolvedItem(id, true));
                            _addedComponents.add(id);
                        }
                    }
                }
            }
        }
        
        _contentTypesInitialized = true;
    }
    
    /**
     * Get the configuration of the content type item
     * @param id The id of item
     * @param cType The content type
     * @return The configuration
     */
    protected Configuration _getContentTypeConfiguration (String id, ContentType cType)
    {
        DefaultConfiguration conf = new DefaultConfiguration("extension");
        conf.setAttribute("id", id);
        
        DefaultConfiguration classConf = new DefaultConfiguration("class");
        classConf.setAttribute("name", "Ametys.ribbon.element.ui.ButtonController");
        
        // Label
        DefaultConfiguration labelConf = new DefaultConfiguration("label");
        labelConf.setValue(_i18nUtils.translate(cType.getLabel()));
        classConf.addChild(labelConf);
        
        // Description
        DefaultConfiguration descConf = new DefaultConfiguration("description");
        descConf.setValue(_i18nUtils.translate(cType.getDescription()));
        classConf.addChild(descConf);
        
        // Content type
        DefaultConfiguration typeConf = new DefaultConfiguration("contentTypes");
        typeConf.setValue(cType.getId());
        classConf.addChild(typeConf);
        
        // Icons or glyph
        if (cType.getIconGlyph() != null)
        {
            DefaultConfiguration iconGlyphConf = new DefaultConfiguration("icon-glyph");
            iconGlyphConf.setValue(cType.getIconGlyph());
            classConf.addChild(iconGlyphConf);
        }
        if (cType.getIconDecorator() != null)
        {
            DefaultConfiguration iconDecoratorConf = new DefaultConfiguration("icon-decorator");
            iconDecoratorConf.setValue(cType.getIconDecorator());
            classConf.addChild(iconDecoratorConf);
        }
        if (cType.getSmallIcon() != null)
        {
            DefaultConfiguration iconSmallConf = new DefaultConfiguration("icon-small");
            iconSmallConf.setValue(cType.getSmallIcon());
            classConf.addChild(iconSmallConf);
            DefaultConfiguration iconMediumConf = new DefaultConfiguration("icon-medium");
            iconMediumConf.setValue(cType.getMediumIcon());
            classConf.addChild(iconMediumConf);
            DefaultConfiguration iconLargeConf = new DefaultConfiguration("icon-large");
            iconLargeConf.setValue(cType.getLargeIcon());
            classConf.addChild(iconLargeConf);
        }
        
        // Common configuration
        @SuppressWarnings("unchecked")
        Map<String, Object> commonConfig = (Map<String, Object>) this._script.getParameters().get("items-config");
        for (String tagName : commonConfig.keySet())
        {
            DefaultConfiguration c = new DefaultConfiguration(tagName);
            c.setValue(String.valueOf(commonConfig.get(tagName)));
            classConf.addChild(c);
        }
        
        conf.addChild(classConf);
        return conf;
    }
    
    /**
     * Get the list of content types classified by groups
     * @return The content types
     */
    protected Map<I18nizableText, Set<ContentType>> _getContentTypesByGroup ()
    {
        Map<I18nizableText, Set<ContentType>> groups = new TreeMap<>(new I18nizableTextKeyComparator());
        
        if (this._script.getParameters().get("contentTypes") != null)
        {
            String[] contentTypesIds = ((String) this._script.getParameters().get("contentTypes")).split(",");
            
            boolean allowInherit = "true".equals(this._script.getParameters().get("allowInherit"));
            
            for (String contentTypeId : contentTypesIds)
            {
                ContentType contentType = _contentTypeExtensionPoint.getExtension(contentTypeId);
                
                if (isValidContentType(contentType))
                {
                    addContentType (contentType, groups);
                }
                
                if (allowInherit)
                {
                    for (String subTypeId : _contentTypeExtensionPoint.getSubTypes(contentTypeId))
                    {
                        ContentType subContentType = _contentTypeExtensionPoint.getExtension(subTypeId);
                        if (isValidContentType(subContentType))
                        {
                            addContentType (subContentType, groups);
                        }
                    }
                }
            }
        }
        else
        {
            Set<String> contentTypesIds = _contentTypeExtensionPoint.getExtensionsIds();
            
            for (String contentTypeId : contentTypesIds)
            {
                ContentType contentType = _contentTypeExtensionPoint.getExtension(contentTypeId);
                
                if (isValidContentType(contentType))
                {
                    addContentType (contentType, groups);
                }
            }
        }
        
        return groups;
    }
    
    /**
     * Add content to groups
     * @param contentType The content type
     * @param groups The groups
     */
    protected void addContentType (ContentType contentType, Map<I18nizableText, Set<ContentType>> groups)
    {
        I18nizableText group = contentType.getCategory();
        if (group.isI18n() && group.getKey().isEmpty()
            || !group.isI18n() && group.getLabel().isEmpty())
        {
            group = new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_CREATECONTENTMENU_GROUP_10_CONTENT");
        }
        
        if (!groups.containsKey(group))
        {
            groups.put(group, new TreeSet<>(new ContentTypeComparator()));
        }
        Set<ContentType> cTypes = groups.get(group);
        cTypes.add(contentType);
    }
    
    @Override
    public List<Script> getScripts(boolean ignoreRights, Map<String, Object> contextParameters)
    {
        try
        {
            _lazyInitializeContentTypeMenu();
        }
        catch (Exception e)
        {
            throw new IllegalStateException("Unable to lookup client side element local components", e);
        }
        
        for (String id: _contentTypeExtensionPoint.getExtensionsIds())
        {
            ContentType contentType = _contentTypeExtensionPoint.getExtension(id);
            if (isValidContentType(contentType))
            {
                return super.getScripts(ignoreRights, contextParameters);
            }
        }
        return new ArrayList<>();
    }
    
    /**
     * Determines if the content type is a valid content type for the gallery
     * @param contentType The coentent
     * @return true if it is a valid content type
     */
    protected boolean isValidContentType (ContentType contentType)
    {
        return !contentType.isAbstract() && !contentType.isPrivate() && !contentType.isMixin();
    }
    
    /**
     * Test if the current user has the right needed by the content type to create a content.
     * @param cType the content type
     * @return true if the user has the right needed, false otherwise.
     */
    protected boolean hasRight(ContentType cType)
    {
        String right = cType.getRight();
        
        if (right == null)
        {
            return true;
        }
        else
        {
            UserIdentity user = _currentUserProvider.getUser();
            return _rightManager.hasRight(user, right, "/cms") == RightResult.RIGHT_ALLOW || _rightManager.hasRight(user, right, _rootContentHelper.getRootContent()) == RightResult.RIGHT_ALLOW;
        }
    }
    
    class ContentTypeComparator implements Comparator<ContentType>
    {
        @Override
        public int compare(ContentType c1, ContentType c2)
        {
            I18nizableText t1 = c1.getLabel();
            I18nizableText t2 = c2.getLabel();
            
            String str1 = t1.isI18n() ? t1.getKey() : t1.getLabel();
            String str2 = t2.isI18n() ? t2.getKey() : t2.getLabel();
            
            int compareTo = str1.toString().compareTo(str2.toString());
            if (compareTo == 0)
            {
                // Content types have same keys but there are not equals, so do not return 0 to add it in TreeSet
                // Indeed, in a TreeSet implementation two elements that are equal by the method compareTo are, from the standpoint of the set, equal 
                return 1;
            }
            return compareTo;
        }
    }
}
