/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang.StringUtils;

import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.SimpleMenu;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.ViewItemGroup;

/**
 * This element creates a menu for the tag policy gallery.
 * This gallery must be disabled is the current edited content does not contain any tab.
 */
public class FormEditionModeMenu extends SimpleMenu
{
    /** The ametys resolver */
    protected AmetysObjectResolver _resolver;
    
    /** Content type extension point. */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    /** Helper for content types */
    protected ContentTypesHelper _contentTypesHelper;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
        _contentTypesHelper = (ContentTypesHelper) smanager.lookup(ContentTypesHelper.ROLE);
    }
    
    /**
     * Return the current status of this menu.
     * It determines if the current content contain at least one tab in edition.
     * @param contentId The id of the content
     * @param viewName The name of the view in which the search must be done.
     * @param fallbackViewName The name of the fallback view if the initial was not found. Can be null.
     * @return A Map with ("has-tab", true/false).
     */
    @Callable
    public Map<String, Object> getStatus(String contentId, String viewName, String fallbackViewName)
    {
        Map<String, Object> results = new HashMap<>();
        
        Content content = _resolver.resolveById(contentId);
        View view = _contentTypesHelper.getViewWithFallback(viewName, fallbackViewName, content.getTypes(), content.getMixinTypes());
        
        if (view == null)
        {
            String errorMsg = String.format("Unknown metadata set '%s' of type 'edition' for content type(s) '%s'", viewName, StringUtils.join(content.getTypes(), ','));
            getLogger().error(errorMsg);
            throw new IllegalArgumentException(errorMsg);
        }
        
        boolean found = _searchForTabFieldSet(view.getViewItems());
        results.put("has-tab", found);
        
        return results;
    }

    /**
     * Determines if at least one tab is present in this list of {@link ViewItem}
     * @param viewItems The list of {@link ViewItem}
     * @return true if one tab has been found.
     */
    private boolean _searchForTabFieldSet(List<ViewItem> viewItems)
    {
        // First search in items.
        for (ViewItem viewItem : viewItems)
        {
            if (viewItem instanceof ViewItemGroup)
            {
                ViewItemGroup fieldSet = (ViewItemGroup) viewItem;
                
                if (ViewItemGroup.TAB_ROLE.equals(fieldSet.getRole().toLowerCase()))
                {
                    return true;
                }
            }
        }
        
        // Recursive search.
        for (ViewItem metadataSetElement : viewItems)
        {
            if (metadataSetElement instanceof ViewItemAccessor)
            {
                if (_searchForTabFieldSet(((ViewItemAccessor) metadataSetElement).getViewItems()))
                {
                    return true;
                }
            }
        }
        
        return false;
    }
}
