/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.languages.Language;
import org.ametys.cms.languages.LanguagesManager;
import org.ametys.core.ui.SimpleMenu;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.core.util.I18nUtils;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Menu that lists the available languages for ODF contents
 */
public class LanguageMenu extends SimpleMenu
{
    /** The plugin in which are the language resources (messages and icons). */
    protected static final String _RESOURCES_PLUGIN = "cms";
    
    /** The language manager */
    protected LanguagesManager _languagesManager;
    
    /** The i18n utils */
    protected I18nUtils _i18nUtils;
    
    private boolean _languagesInitialized;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _languagesManager = (LanguagesManager) smanager.lookup(LanguagesManager.ROLE);
        _i18nUtils = (I18nUtils) smanager.lookup(I18nUtils.ROLE);
    }
    
    @Override
    protected void _getGalleryItems(Map<String, Object> parameters, Map<String, Object> contextualParameters)
    {
        try
        {
            _lazyInitializeLanguageGallery(contextualParameters);
        }
        catch (Exception e)
        {
            throw new IllegalStateException("Unable to lookup client side element local components", e);
        }
        
        super._getGalleryItems(parameters, contextualParameters);
    }
    
    private synchronized void _lazyInitializeLanguageGallery(Map<String, Object> contextualParameters) throws ConfigurationException
    {
        if (!_languagesInitialized)
        {
            Map<String, Language> languages = getAvailableLanguages (contextualParameters);
            
            if (languages.size() > 0)
            {
                GalleryItem galleryItem = new GalleryItem();
                
                GalleryGroup galleryGroup = new GalleryGroup(new I18nizableText("plugin." + _RESOURCES_PLUGIN, "PLUGINS_CMS_LANGUAGEMENU_GROUP_LABEL"));
                galleryItem.addGroup(galleryGroup);
                    
                for (Language language : languages.values())
                {
                    String id = this.getId() + "." + language.getCode();
                    
                    Configuration conf = _getLanguageItemConfiguration (id, language);
                    _getGalleryItemManager().addComponent(_pluginName, null, id, StaticClientSideElement.class, conf);
                    galleryGroup.addItem(new UnresolvedItem(id, true));
                }
                
                _galleryItems.add(galleryItem);
            }
            
            if (_galleryItems.size() > 0)
            {
                try
                {
                    _getGalleryItemManager().initialize();
                }
                catch (Exception e)
                {
                    throw new ConfigurationException("Unable to lookup parameter local components", e);
                }
            }
        }
        
        _languagesInitialized = true;
    }
    
    /**
     * Get the available languages 
     * @param contextualParameters The contextual parameters
     * @return the available languages 
     */
    protected Map<String, Language> getAvailableLanguages (Map<String, Object> contextualParameters)
    {
        return  _languagesManager.getAvailableLanguages();
    }
    
    
    @Override
    public List<Script> getScripts(boolean ignoreRights, Map<String, Object> contextParameters)
    {
        Map<String, Language> availableLanguages = getAvailableLanguages(contextParameters);
        if (availableLanguages == null || availableLanguages.isEmpty())
        {
            // Hide menu if there is no language
            return new ArrayList<>();
        }
        
        return super.getScripts(ignoreRights, contextParameters);
    }
    
    /**
     * Get the configuration for a language item
     * @param itemId The item id
     * @param language The language
     * @return The configuration
     */
    protected Configuration _getLanguageItemConfiguration (String itemId, Language language)
    {
        DefaultConfiguration conf = new DefaultConfiguration("extension");
        conf.setAttribute("id", itemId);
        
        DefaultConfiguration classConf = new DefaultConfiguration("class");
        classConf.setAttribute("name", "Ametys.ribbon.element.ui.ButtonController");
        
        // Label
        DefaultConfiguration labelConf = new DefaultConfiguration("label");
        labelConf.setValue(_i18nUtils.translate(language.getLabel()));
        classConf.addChild(labelConf);
        
        // Description
        DefaultConfiguration descConf = new DefaultConfiguration("description");
        descConf.setValue(_i18nUtils.translate(language.getLabel()));
        classConf.addChild(descConf);
        
        // Code
        DefaultConfiguration codeConf = new DefaultConfiguration("code");
        codeConf.setValue(language.getCode());
        classConf.addChild(codeConf);
        
        // Icons
        DefaultConfiguration iconSmallConf = new DefaultConfiguration("icon-small");
        iconSmallConf.setValue(language.getSmallIcon());
        classConf.addChild(iconSmallConf);
        DefaultConfiguration iconMediumConf = new DefaultConfiguration("icon-medium");
        iconMediumConf.setValue(language.getMediumIcon());
        classConf.addChild(iconMediumConf);
        DefaultConfiguration iconLargeConf = new DefaultConfiguration("icon-large");
        iconLargeConf.setValue(language.getLargeIcon());
        classConf.addChild(iconLargeConf);
        
        // Common configuration
        @SuppressWarnings("unchecked")
        Map<String, Object> commonConfig = (Map<String, Object>) this._script.getParameters().get("items-config");
        _addChildrenToConf(classConf, commonConfig);
        
        conf.addChild(classConf);
        return conf;
    }
    
    @SuppressWarnings("unchecked")
    private void _addChildrenToConf (DefaultConfiguration conf, Map<String, Object> childrenAsMap)
    {
        for (String tagName : childrenAsMap.keySet())
        {
            DefaultConfiguration c = new DefaultConfiguration(tagName);
            Object value = childrenAsMap.get(tagName);
            if (value instanceof Map)
            {
                _addChildrenToConf(c, (Map<String, Object>) value);
            }
            else
            {
                c.setValue(String.valueOf(value));
            }
            conf.addChild(c);
        }
    }
}
