/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.content.referencetable.HierarchicalReferenceTablesHelper;
import org.ametys.cms.contenttype.ContentType;

/**
 * This element creates a menu with one gallery item per reference table content type classified by category. 
 */
public class ReferenceTablesGallery extends ContentTypesGallery
{
    /** The helper component for hierarchical simple contents */
    protected HierarchicalReferenceTablesHelper _hierarchicalSimpleContentsHelper;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _hierarchicalSimpleContentsHelper = (HierarchicalReferenceTablesHelper) smanager.lookup(HierarchicalReferenceTablesHelper.ROLE);
    }
    
    @Override
    protected boolean isValidContentType(ContentType contentType)
    {
        return contentType != null && contentType.isReferenceTable() && !contentType.isAbstract()  && !contentType.isPrivate()
               && (!_hierarchicalSimpleContentsHelper.isHierarchical(contentType) || _hierarchicalSimpleContentsHelper.isLeaf(contentType));
    }
    
    @Override
    public List<Script> getScripts(boolean ignoreRights, Map<String, Object> contextParameters)
    {
        List<ContentType> cTypes = getReferenceTables();
        if (cTypes.isEmpty())
        {
            // Hide menu if there is no simple contents
            return new ArrayList<>();
        }
        
        return super.getScripts(ignoreRights, contextParameters);
    }
    
    /**
     * Get the list of available simple content types
     * @return The simple content types
     */
    protected List<ContentType> getReferenceTables()
    {
        List<ContentType> refTables = new ArrayList<>();
        
        Set<String> cTypeIds = _contentTypeExtensionPoint.getExtensionsIds();
        for (String cTypeId : cTypeIds)
        {
            ContentType cType = _contentTypeExtensionPoint.getExtension(cTypeId);
            if (isValidContentType(cType))
            {
                refTables.add(cType);
            }
        }
        
        return refTables;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    protected void _addContentTypeConfiguration (DefaultConfiguration rootConf, ContentType cType)
    {
        boolean hasParentAttribute = !cType.getParentAttributeDefinition().isEmpty();
        
        DefaultConfiguration classConf = new DefaultConfiguration("class");
        if (hasParentAttribute)
        {
            classConf.setAttribute("name", "Ametys.plugins.cms.content.controller.OpenReferenceTableToolsButtonController");

        }
        else
        {
            classConf.setAttribute("name", "Ametys.ribbon.element.ui.button.OpenToolButtonController");
        }
        
        // Label
        DefaultConfiguration labelConf = _getI18nizableTextConfiguration("label", cType.getLabel());
        classConf.addChild(labelConf);
        
        // Description
        DefaultConfiguration descConf = _getI18nizableTextConfiguration("description", cType.getDescription());
        classConf.addChild(descConf);
        
        // Icons or glyph
        _addContentTypeIconsConfiguration(classConf, cType);
        
        // Open tool params
        DefaultConfiguration toolParams = new DefaultConfiguration("opentool-params");
        // Tool id
        DefaultConfiguration idConf = new DefaultConfiguration("id");
        idConf.setValue("reference-table-search-ui." + cType.getId());
        toolParams.addChild(idConf);
        // Content Type
        DefaultConfiguration typeConf = new DefaultConfiguration("contentType");
        typeConf.setValue(cType.getId());
        toolParams.addChild(typeConf);
        // Content default title
        DefaultConfiguration defaultTitleConf = _getI18nizableTextConfiguration("contentDefaultTitle", cType.getDefaultTitle());
        toolParams.addChild(defaultTitleConf);
        // Content language
        DefaultConfiguration langConf = null;
        if (this._script.getParameters().containsKey("contentLanguage"))
        {
            langConf = new DefaultConfiguration("contentLanguage");
            langConf.setValue((String) this._script.getParameters().get("contentLanguage"));
            toolParams.addChild(langConf);
        }
        // Start search at opening
        DefaultConfiguration startSearchConf = new DefaultConfiguration("startSearchAtOpening");
        startSearchConf.setValue(true);
        toolParams.addChild(startSearchConf);
        
        classConf.addChild(toolParams);
        
        // Common configuration
        Map<String, Object> commonConfig = (Map<String, Object>) this._script.getParameters().get("items-config");
        for (String tagName : commonConfig.keySet())
        {
            DefaultConfiguration c = new DefaultConfiguration(tagName);
            c.setValue(String.valueOf(commonConfig.get(tagName)));
            classConf.addChild(c);
        }
        
        // Potential configuration if the content type defines a parent
        if (hasParentAttribute)
        {
            // HierarchicalSimpleContentsTool id
            DefaultConfiguration treeIdConf = new DefaultConfiguration("open-tree-tool-id");
            treeIdConf.setValue("uitool-hierarchical-reference-tables");
            classConf.addChild(treeIdConf);
            
            // HierarchicalSimpleContentsTool params
            DefaultConfiguration treeToolParams = new DefaultConfiguration("open-tree-tool-params");
            
            DefaultConfiguration treeParamIdConf = new DefaultConfiguration("id");
            treeParamIdConf.setValue(cType.getId());
            treeToolParams.addChild(treeParamIdConf);
            
            DefaultConfiguration treeWorkflowEditActionIdConf = new DefaultConfiguration("workflowEditActionId");
            treeWorkflowEditActionIdConf.setValue("2");
            treeToolParams.addChild(treeWorkflowEditActionIdConf);
            
            treeToolParams.addChild(typeConf);
            treeToolParams.addChild(defaultTitleConf);
            if (langConf != null)
            {
                treeToolParams.addChild(langConf);
            }
            treeToolParams.addChild(startSearchConf);
            
            classConf.addChild(treeToolParams);
            
            // Action
            DefaultConfiguration action = new DefaultConfiguration("action");
            action.setValue("Ametys.plugins.cms.content.controller.OpenReferenceTableToolsButtonController._act");
            classConf.addChild(action);
        }
        
        rootConf.addChild(classConf);
        
        if (hasParentAttribute)
        {
            // When using the class Ametys.plugins.cms.content.controller.OpenReferenceTableToolsButtonController
            // we need the associated files
            
            DefaultConfiguration scriptsConfiguration = new DefaultConfiguration("scripts");
            
            DefaultConfiguration script1Configuration = new DefaultConfiguration("file");
            script1Configuration.setValue("js/Ametys/plugins/cms/content/controller/OpenReferenceTableToolsButtonController.js");
            scriptsConfiguration.addChild(script1Configuration);
            
            DefaultConfiguration script2Configuration = new DefaultConfiguration("file");
            script2Configuration.setValue("js/Ametys/plugins/cms/content/actions/ReferenceTableActions.js");
            scriptsConfiguration.addChild(script2Configuration);
    
            rootConf.addChild(scriptsConfiguration);
        }

        _addRightsOnContentTypeConfiguration(rootConf);
    }
    
    @Override
    protected boolean hasRight(ContentType cType)
    {
        return true;
    }
    
    @Override
    protected Map<String, List<String>> _configureDependencies(Configuration configuration) throws ConfigurationException
    {
        Map<String, List<String>> dependencies = super._configureDependencies(configuration);
        
        boolean needHierarchicalTool = _hierarchicalSimpleContentsHelper.hasAtLeastOneHierarchy();
        
        if (needHierarchicalTool)
        {
            List<String> uitoolsDependencies = dependencies.get("org.ametys.core.ui.UIToolsFactoriesManager");
            if (uitoolsDependencies == null)
            {
                uitoolsDependencies = new ArrayList<>();
            }
            uitoolsDependencies.add("uitool-hierarchical-reference-tables");
        }
        
        return dependencies;
    }
}
