/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement;

import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.alerts.AlertsConstants;
import org.ametys.cms.content.archive.ArchiveConstants;
import org.ametys.cms.repository.Content;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.holder.ModifiableModelLessDataHolder;
import org.ametys.plugins.repository.version.ModifiableDataAwareVersionableAmetysObject;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * This element creates a ribbon button to schedule the archiving of a content
 */
public class ScheduleArchivingClientSideElement extends StaticClientSideElement
{
    /** I18n error key : content type error */
    private static final String __I18N_KEY_CONTENT_TYPE_ERROR = "PLUGINS_CMS_ARCHIVE_SCHEDULE_CONTENT_TYPE_ERROR";

    /** I18n error key : unexpected error */
    private static final String __I18N_KEY_UNEXPECTED_ERROR = "PLUGINS_CMS_ARCHIVE_SCHEDULE_UNEXPECTED_ERROR";
    
    private ObservationManager _observationManager;
    private AmetysObjectResolver _resolver;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _observationManager = (ObservationManager) manager.lookup(ObservationManager.ROLE);
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    /**
     * Set or remove the <code>ArchiveConstants.META_ARCHIVE_SCHEDULED_DATE</code> on a content.
     * @param contentIds The ids of contents
     * @param scheduledDateAsStr The schedule date as string
     * @return the successful and failed contents
     */
    @Callable
    public Map<String, Object> setScheduledArchivingDate (List<String> contentIds, String scheduledDateAsStr)
    {
        Map<String, Object> results = new HashMap<>();
        
        // Error list
        List<I18nizableText> failedContents = new ArrayList<>();

        // Success list, containing the contentId that have been successfully scheduled
        List<String> successContents = new ArrayList<>();

        ZonedDateTime scheduledDate = DateUtils.parseZonedDateTime(scheduledDateAsStr);
        
        if (StringUtils.isNotEmpty(scheduledDateAsStr))
        {
            if (scheduledDate == null)
            {
                getLogger().error("Cannot cast value '" + scheduledDateAsStr + "' into type 'date'");
                throw new IllegalArgumentException("Invalid format for date value '" + scheduledDateAsStr + "'");
            }

            // Date must be set at least after tomorrow
            /*if (scheduledDate.compareTo(new DateMidnight().plusDays(1).toDate()) < 0)
            {
                getLogger().error("The scheduled date '" + scheduledDate + "' cannot be set before the current date");
                throw new IllegalArgumentException("The scheduled date '" + scheduledDate + "' cannot be set before the current date");
            }*/
        }
        
        // For each content
        for (String contentId : contentIds)
        {
            try
            {
                Content content = _resolver.resolveById(contentId);

                if (content instanceof ModifiableDataAwareVersionableAmetysObject)
                {
                    ModifiableDataAwareVersionableAmetysObject uContent = (ModifiableDataAwareVersionableAmetysObject) content;
                    ModifiableModelLessDataHolder dataHolder = uContent.getUnversionedDataHolder();

                    // Remove META_ARCHIVE_SCHEDULED_DATE and
                    // SCHEDULED_ARCHIVING_REMINDER_LAST_DATE if date is empty.
                    if (StringUtils.isEmpty(scheduledDateAsStr))
                    {
                        dataHolder.removeValue(ArchiveConstants.META_ARCHIVE_SCHEDULED_DATE);
                        if (getLogger().isInfoEnabled())
                        {
                            getLogger().info("Content with id : '" + uContent.getId() + "' does not have a scheduled archiving date anymore.");
                        }

                        dataHolder.removeValue(AlertsConstants.SCHEDULED_ARCHIVING_REMINDER_LAST_DATE);
                    }
                    else
                    {
                        dataHolder.setValue(ArchiveConstants.META_ARCHIVE_SCHEDULED_DATE, scheduledDate);
                    }

                    if (uContent.needsSave())
                    {
                        uContent.saveChanges();

                        // report success
                        successContents.add(uContent.getId());

                        // Notify observer
                        Map<String, Object> eventParams = new HashMap<>();
                        eventParams.put(ObservationConstants.ARGS_CONTENT, content);
                        eventParams.put(ObservationConstants.ARGS_CONTENT_ID, contentId);

                        _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_MODIFIED, _currentUserProvider.getUser(), eventParams));
                    }
                }
                else
                {
                    getLogger().error("Unable to set a scheduled archiving for the content : '" + content.getId() + "'. It is not a ModifiableMetadataAwareVersionableAmetysObject.");

                    // report error
                    List<String> i18nParams = new ArrayList<>();
                    i18nParams.add(content.getTitle(null));
                    failedContents.add(new I18nizableText("plugin.cms", __I18N_KEY_CONTENT_TYPE_ERROR, i18nParams));

                }
            }
            catch (Exception e)
            {
                getLogger().error("Unexpected exception while trying to schedule an archiving for the content : '" + contentId + "'.");

                // report error
                List<String> i18nParams = new ArrayList<>();
                i18nParams.add(contentId);
                failedContents.add(new I18nizableText("plugin.cms", __I18N_KEY_UNEXPECTED_ERROR, i18nParams));
            }
        }

        results.put("error", failedContents);
        results.put("success", successContents);
        
        return results;
    }

    /**
     * Gets the scheduled archiving dates of one or several contents
     * @param contentIds the ids of the contents
     * @return result the server's response in JSON
     */
    @Callable
    public Map<String, Object> getScheduledArchivingDate(List<String> contentIds)
    {
        Map<String, Object> results = new HashMap<> ();
        Set<ZonedDateTime> dates = new HashSet<>();
    
        // Retrieves the date for each content
        for (String contentId : contentIds)
        {
            Content content = _resolver.resolveById(contentId);
    
            if (content instanceof ModifiableDataAwareVersionableAmetysObject)
            {
                ZonedDateTime date = ((ModifiableDataAwareVersionableAmetysObject) content).getUnversionedDataHolder().getValue(ArchiveConstants.META_ARCHIVE_SCHEDULED_DATE);
                dates.add(date);
            }
        }
    
        // Retrieves the closest date from current time.
        dates.remove(null);
        ZonedDateTime scheduledDate = getDateNearestToCurrent(dates);
    
        // Generate the JSON
        Map<String, Object> content = new HashMap<> ();
        if (scheduledDate != null)
        {
            content.put("date", DateUtils.zonedDateTimeToString(scheduledDate));
        }
        results.put("content", content);
        
        return results;
    }
    
    private ZonedDateTime getDateNearestToCurrent(Set<ZonedDateTime> dates)
    {
        ZonedDateTime returnDate = null;
        ZonedDateTime currentDate = ZonedDateTime.now();

        for (ZonedDateTime date : dates)
        {
            if (date.compareTo(currentDate) >= 0 && (returnDate == null || date.compareTo(returnDate) < 0))
            {
                returnDate = date;
            }
        }
        return returnDate;
    }
    
    @Override
    public List<Script> getScripts(boolean ignoreRights, Map<String, Object> contextParameters)
    {
        boolean enabled = Config.getInstance().getValue("archive.scheduler.enabled");
        if (enabled)
        {
            return super.getScripts(ignoreRights, contextParameters);
        }
        
        return new ArrayList<>();
    }
}
