/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.component.ComponentException;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.parameter.Enumerator;
import org.ametys.runtime.parameter.StaticEnumerator;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * Search tool HMI item
 */
public class SearchClientSideElement extends StaticClientSideElement implements Contextualizable, Disposable
{
    /** Manager component for enumerators */
    protected ThreadSafeComponentManager<Enumerator> _enumeratorManager;
    /** The service manager */
    protected ServiceManager _manager;
    /** The context. */
    protected Context _context;
    
    private int _criteriaIndex;
    private int _columnIndex;
    private int _nbColumns;
    private int _propertyIndex;
    
    private final Map<String, String> _enumeratorsToLookup = new HashMap<>();
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _manager = smanager;
    }
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        _enumeratorManager = new ThreadSafeComponentManager<>();
        _enumeratorManager.setLogger(getLogger());
        _enumeratorManager.contextualize(_context);
        _enumeratorManager.service(_manager);
        
        super.configure(configuration);
        
        _configureSearchCriteria(configuration.getChild("search-criteria"));
        _configureColumns(configuration.getChild("columns"));
        _configureProperties(configuration.getChild("property-mapping"));
        _configurePageSize(configuration);
        _configureSearchUrl(configuration);
        _configureExportUrl(configuration);
        
        try
        {
            _enumeratorManager.initialize();
        }
        catch (Exception e)
        {
            throw new ConfigurationException("Unable to lookup parameter local components", configuration, e);
        }
    }
    
    @Override
    public List<Script> getScripts(boolean ignoreRights, Map<String, Object> contextParameters)
    {
        List<Script> scripts = super.getScripts(ignoreRights, contextParameters);
        
        Map<String, Object> parameters = new HashMap<>();
        List<ScriptFile> cssFiles = new ArrayList<>();
        List<ScriptFile> scriptFiles = new ArrayList<>();
        for (Script script : scripts)
        {
            cssFiles.addAll(script.getCSSFiles());
            scriptFiles.addAll(script.getScriptFiles());
            parameters.putAll(script.getParameters());
        }
        Script script = new Script(this.getId(), _script.getScriptClassname(), scriptFiles, cssFiles, parameters);

        try
        {
            for (String prefix : _enumeratorsToLookup.keySet())
            {
                String className = _enumeratorsToLookup.get(prefix);
                try
                {
                    Enumerator enumerator = _enumeratorManager.lookup(className);
                    
                    setEnumeratorValues(script, enumerator, prefix);
                }
                catch (ComponentException e)
                {
                    getLogger().error("Unable to instantiate enumerator for class: " + className, e);
                }
            }
        }
        catch (Exception e)
        {
            getLogger().error("Unable to lookup parameter local components", e);
        }
        
        List<Script> result = new ArrayList<>();
        result.add(script);
        return result;
    }
    
    /**
     * Configure the search criteria
     * @param configuration The configuration
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected void _configureSearchCriteria (Configuration configuration) throws ConfigurationException
    {
        for (Configuration criteriaConfiguration : configuration.getChildren("criteria"))
        {
            _configureCriteria (criteriaConfiguration, "criteria-" + _criteriaIndex);
            _criteriaIndex++;
        }
        this._script.getParameters().put("criteria-size", _criteriaIndex);
        this._script.getParameters().put("nb-columns", _nbColumns);
    }
    
    /**
     * Configures the criteria parameters
     * @param configuration the item configuration
     * @param prefix the criteria prefix
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected void _configureCriteria (Configuration configuration, String prefix) throws ConfigurationException
    {
        Map<String, Object> initialParameters = _script.getParameters();
        String name = configuration.getAttribute("name");
        String type = configuration.getAttribute("type", "string");
        int column = configuration.getAttributeAsInteger("column", 1);
        I18nizableText label = _configureLabel(configuration);
        String testOperator = configuration.getChild("test-operator", true).getValue("wd");
        String defaultValue = configuration.getChild("default-value", true).getValue("");
        boolean metadata = configuration.getAttributeAsBoolean("metadata", true);
        boolean multiple = configuration.getAttributeAsBoolean("multiple", false);
        boolean mandatory = configuration.getAttributeAsBoolean("mandatory", false);
        boolean hidden = configuration.getAttributeAsBoolean("hidden", false);
        Enumerator staticEnum = _configureEnumerator(configuration, prefix + "-enumerator");
        String widget = _configureWidget (configuration);
        Map<String, Object> widgetParams = _configureWidgetParams(configuration);
        String width = configuration.getChild("width", true).getValue(null);
        
        initialParameters.put(prefix + "-name", name);
        initialParameters.put(prefix + "-type", type);
        initialParameters.put(prefix + "-column", column);
        initialParameters.put(prefix + "-label", label);
        initialParameters.put(prefix + "-test-operator", testOperator);
        initialParameters.put(prefix + "-default-value", defaultValue);
        initialParameters.put(prefix + "-metadata", metadata);
        initialParameters.put(prefix + "-hidden", hidden);
        initialParameters.put(prefix + "-multiple", multiple);
        initialParameters.put(prefix + "-mandatory", mandatory);
        
        if (width != null)
        {
            initialParameters.put(prefix + "-width", width);
        }
        
        if (widget != null)
        {
            initialParameters.put(prefix + "-widget", widget);
        }
        
        initialParameters.put(prefix + "-widget-param-size", widgetParams.size());
        
        if (widgetParams.size() > 0)
        {
            int i = 0;
            for (String paramName : widgetParams.keySet())
            {
                initialParameters.put(prefix + "-widget-param-" + i, paramName);
                initialParameters.put(prefix + "-widget-param-" + i + "-value", widgetParams.get(paramName));
                i++;
            }
        }
        
        if (staticEnum != null)
        {
            setEnumeratorValues(_script, staticEnum, prefix + "-enumerator");
        }
        
        if (column > _nbColumns)
        {
            _nbColumns = column;
        }
    }
    
    /**
     * Set the enumerator values in script parameters
     * @param script The script into which set the parameters
     * @param enumerator The enumerator
     * @param prefix The parameters prefix
     */
    protected void setEnumeratorValues (Script script, Enumerator enumerator, String prefix)
    {
        try
        {
            script.getParameters().put(prefix, true);
            Map<Object, I18nizableText> entries = enumerator.getEntries();
            int index = 0;
            for (Object value : entries.keySet())
            {
                script.getParameters().put(prefix + "-" + index + "-value", value);
                if (entries.get(value) != null)
                {
                    script.getParameters().put(prefix + "-" + index + "-label", entries.get(value));
                }
                else
                {
                    script.getParameters().put(prefix + "-" + "-label", value);
                }
                index++;
            }
            
            script.getParameters().put(prefix + "-size", index);
        }
        catch (Exception e)
        {
            getLogger().error("Unable to set values for enumerator " + enumerator.getClass().getName(), e);
        }
    }
    /**
     * Configure the columns
     * @param configuration The configuration
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected void _configureColumns (Configuration configuration) throws ConfigurationException
    {
        for (Configuration colConfiguration : configuration.getChildren("column"))
        {
            _configureColumn (colConfiguration, "column-" + _columnIndex);
            _columnIndex++;
        }
        
        Configuration groupBy = configuration.getChild("group-by", false);
        if (groupBy != null)
        {
            this._script.getParameters().put("group-by", groupBy.getValue());
        }
        
        Configuration sortBy = configuration.getChild("sort-by", false);
        if (sortBy != null)
        {
            this._script.getParameters().put("sort-by", sortBy.getValue());
        }
        
        this._script.getParameters().put("column-size", _columnIndex);
    }
    
    /**
     * Configures the criteria parameters
     * @param configuration the item configuration
     * @param prefix the criteria prefix
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected void _configureColumn (Configuration configuration, String prefix) throws ConfigurationException
    {
        Map<String, Object> initialParameters = _script.getParameters();
        String id = configuration.getAttribute("id");
        I18nizableText label = _configureLabel(configuration);
        String width = configuration.getChild("width", true).getValue("");
        
        initialParameters.put(prefix + "-id", id);
        initialParameters.put(prefix + "-width", width);
        initialParameters.put(prefix + "-label", label);
        
        String renderFn = configuration.getChild("renderer", true).getValue(null);
        if (renderFn != null)
        {
            initialParameters.put(prefix + "-renderer", renderFn);
        }
        
        boolean sortable = configuration.getAttributeAsBoolean("sortable", true);
        boolean hideable = configuration.getAttributeAsBoolean("hideable", true);
        boolean hidden = configuration.getAttributeAsBoolean("hidden", false);
        
        initialParameters.put(prefix + "-sortable", sortable);
        initialParameters.put(prefix + "-hideable", hideable);
        initialParameters.put(prefix + "-hidden", hidden);
    }
    
    /**
     * Configure the columns
     * @param configuration The configuration
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected void _configureProperties (Configuration configuration) throws ConfigurationException
    {
        for (Configuration propConfiguration : configuration.getChildren("property"))
        {
            _configureProperty (propConfiguration, "property-" + _propertyIndex);
            _propertyIndex++;
        }
        this._script.getParameters().put("property-size", _propertyIndex);
    }
    
    /**
     * Configures the criteria parameters
     * @param configuration the item configuration
     * @param prefix the criteria prefix
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected void _configureProperty (Configuration configuration, String prefix) throws ConfigurationException
    {
        String id = configuration.getAttribute("id");
        String path = configuration.getAttribute("path");
        String type = configuration.getAttribute("type", "string");
        
        this._script.getParameters().put(prefix + "-id", id);
        this._script.getParameters().put(prefix + "-path", path);
        this._script.getParameters().put(prefix + "-type", type);
    }
    
    /**
     * Configures the label.
     * @param configuration the configuration to use.
     * @return the i18n text.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected I18nizableText _configureLabel (Configuration configuration) throws ConfigurationException
    {
        Configuration labelConfig = configuration.getChild("label");
        boolean i18nSupported = labelConfig.getAttributeAsBoolean("i18n", false);
        
        if (i18nSupported)
        {
            String catalogue = labelConfig.getAttribute("catalogue", null);
            if (catalogue == null)
            {
                catalogue = "plugin." + _pluginName;
            }
            
            return new I18nizableText(catalogue, labelConfig.getValue());
        }
        else
        {
            return new I18nizableText(labelConfig.getValue(""));
        }
    }
    
    /**
     * Configures the enumerator
     * @param configuration The configuration to use
     * @param prefix The prefix
     * @return The static enumerator or null.
     * @throws ConfigurationException If the configuraiton has an issue
     */
    @SuppressWarnings("unchecked")
    protected Enumerator _configureEnumerator (Configuration configuration, String prefix) throws ConfigurationException
    {
        Configuration enumeratorConfig = configuration.getChild("enumeration", false);
        if (enumeratorConfig != null)
        {
            Configuration customEnumerator = enumeratorConfig.getChild("custom-enumerator", false);
            if (customEnumerator != null)
            {
                String criterionName = configuration.getAttribute("name");
                String enumeratorClassName = customEnumerator.getAttribute("class");
                String role = criterionName + "$" + enumeratorClassName;
                
                try
                {
                    Class enumeratorClass = Class.forName(enumeratorClassName);
                    _enumeratorManager.addComponent(_pluginName, null, role, enumeratorClass, configuration);
                }
                catch (Exception e)
                {
                    throw new ConfigurationException("Unable to instantiate enumerator for class: " + enumeratorClassName, e);
                }

                _enumeratorsToLookup.put(prefix, role);
            }
            else
            {
                StaticEnumerator staticEnumerator = new StaticEnumerator();
                for (Configuration entryConfig : enumeratorConfig.getChildren("entry"))
                {
                    String value = entryConfig.getChild("value").getValue("");
                    I18nizableText label = null;
                    
                    if (entryConfig.getChild("label", false) != null)
                    {
                        label = _configureLabel(entryConfig);
                    }
                    staticEnumerator.add(label, value);
                }
                return staticEnumerator;
            }
        }
        
        return null;
    }
    
    /**
     * Configures the widget
     * @param configuration The configuration to use
     * @return The widget id
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected String _configureWidget (Configuration configuration) throws ConfigurationException
    {
        Configuration widgetConfig = configuration.getChild("widget", false);
        if (widgetConfig != null)
        {
            return widgetConfig.getValue("*");
        }
        
        return null;
    }
    
    /**
     * Configures the widget
     * @param configuration The configuration to use
     * @return The widget id
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected Map<String, Object> _configureWidgetParams (Configuration configuration) throws ConfigurationException
    {
        Map<String, Object> widgetParams = new HashMap<>();
        
        Configuration widgetConfig = configuration.getChild("widget-params", false);
        if (widgetConfig != null)
        {
            Configuration[] params = widgetConfig.getChildren("param");
            for (Configuration paramConfig : params)
            {
                boolean i18nSupported = paramConfig.getAttributeAsBoolean("i18n", false);
                if (i18nSupported)
                {
                    String catalogue = paramConfig.getAttribute("catalogue", null);
                    if (catalogue == null)
                    {
                        catalogue = "plugin." + _pluginName;
                    }
                    
                    widgetParams.put(paramConfig.getAttribute("name"), new I18nizableText(catalogue, paramConfig.getValue()));
                }
                else
                {
                    widgetParams.put(paramConfig.getAttribute("name"), paramConfig.getValue(""));
                }
            }
        }
        
        return widgetParams;
    }
    
    /**
     * Configures the result page size.
     * @param configuration the configuration to use.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected void _configurePageSize (Configuration configuration) throws ConfigurationException
    {
        Configuration pageSizeConfig = configuration.getChild("nb-result-per-page", false);
        if (pageSizeConfig != null)
        {
            this._script.getParameters().put("page-size", pageSizeConfig.getValue());
        }
    }
    
    /**
     * Configure the URL used for search
     * @param configuration The configuration to use.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected void _configureSearchUrl  (Configuration configuration) throws ConfigurationException
    {
        Configuration urlConfig = configuration.getChild("search-url", false);
        if (urlConfig != null)
        {
            String pluginName = urlConfig.getAttribute("plugin", _pluginName);
            this._script.getParameters().put("search-url", "/plugins/" + pluginName + "/" + urlConfig.getValue());
        }
    }
    
    /**
     * Configure the URL used for XSL export
     * @param configuration The configuration to use.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected void _configureExportUrl  (Configuration configuration) throws ConfigurationException
    {
        Configuration urlConfig = configuration.getChild("export-url", false);
        if (urlConfig != null)
        {
            String pluginName = urlConfig.getAttribute("plugin", _pluginName);
            this._script.getParameters().put("export-url", "/plugins/" + pluginName + "/" + urlConfig.getValue());
        }
    }
    
    @Override
    public void dispose()
    {
        _enumeratorManager.dispose();
        _enumeratorManager = null;
    }
}
