/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.component.ComponentException;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.configuration.MutableConfiguration;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.parameter.Enumerator;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * Simple Search tool HMI item.
 * Only SAX content type enumerator values.
 * TODO To remove
 */
public class SimpleSearchClientSideElement extends StaticClientSideElement implements Contextualizable, Disposable
{
    /** Manager component for enumerators */
    protected ThreadSafeComponentManager<Enumerator> _enumeratorManager;
    /** The service manager */
    protected ServiceManager _manager;
    /** The context. */
    protected Context _context;
    
    private String _enumeratorToLookup;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _manager = smanager;
    }
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        _enumeratorManager = new ThreadSafeComponentManager<>();
        _enumeratorManager.setLogger(getLogger());
        _enumeratorManager.contextualize(_context);
        _enumeratorManager.service(_manager);
        
        super.configure(configuration);
        
        _enumeratorToLookup = _configureContentTypeEnumerator();
        
        try
        {
            _enumeratorManager.initialize();
        }
        catch (Exception e)
        {
            throw new ConfigurationException("Unable to lookup parameter local components", configuration, e);
        }
    }
    
    @Override
    public List<Script> getScripts(boolean ignoreRights, Map<String, Object> contextParameters)
    {
        List<Script> scripts = super.getScripts(ignoreRights, contextParameters);
        Map<String, Object> parameters = new HashMap<>();
        List<ScriptFile> cssFiles = new ArrayList<>();
        List<ScriptFile> scriptFiles = new ArrayList<>();
        for (Script script : scripts)
        {
            cssFiles.addAll(script.getCSSFiles());
            scriptFiles.addAll(script.getScriptFiles());
            parameters.putAll(script.getParameters());
        }
        Script script = new Script(this.getId(), _script.getScriptClassname(), scriptFiles, cssFiles, parameters);
        
        try
        {
            String className = _enumeratorToLookup;
            try
            {
                Enumerator enumerator = _enumeratorManager.lookup(className);
                setEnumeratorValues(script, enumerator, "content-type");
            }
            catch (ComponentException e)
            {
                getLogger().error("Unable to instantiate enumerator for class: " + className, e);
            }
        }
        catch (Exception e)
        {
            getLogger().error("Unable to lookup parameter local components", e);
        }
        
        List<Script> result = new ArrayList<>();
        result.add(script);
        return result;
    }
    
    /**
     * Configures the content type enumerator
     * @return The content type enumerator role
     * @throws ConfigurationException If the configuration has an issue
     */
    @SuppressWarnings("unchecked")
    protected String _configureContentTypeEnumerator() throws ConfigurationException
    {
        String criterionName = "content-type";
        String enumeratorClassName = "org.ametys.cms.contenttype.ContentTypeEnumerator";
        String role = criterionName + "$" + enumeratorClassName;
        
        try
        {
            MutableConfiguration configuration = new DefaultConfiguration("custom-enumerator");
            
            MutableConfiguration allOption = new DefaultConfiguration("all-option");
            allOption.setValue("concat");
            MutableConfiguration excludePrivate = new DefaultConfiguration("exclude-private");
            excludePrivate.setValue("true");
            
            configuration.addChild(allOption);
            configuration.addChild(excludePrivate);
            
            Class enumeratorClass = Class.forName(enumeratorClassName);
            _enumeratorManager.addComponent(_pluginName, null, role, enumeratorClass, configuration);
        }
        catch (Exception e)
        {
            throw new ConfigurationException("Unable to instantiate enumerator for class: " + enumeratorClassName, e);
        }
        
        return role;
    }
    
    /**
     * Set the enumerator values in script parameters
     * @param script The script into which set the parameters
     * @param enumerator The enumerator
     * @param prefix The parameters prefix
     * @throws Exception If the enumerator has an issue 
     */
    protected void setEnumeratorValues(Script script, Enumerator enumerator, String prefix) throws Exception
    {
        Map<Object, I18nizableText> entries = enumerator.getEntries();
        int index = 0;
        
        for (Object value : entries.keySet())
        {
            script.getParameters().put(prefix + "-" + index + "-value", value);
            if (entries.get(value) != null)
            {
                script.getParameters().put(prefix + "-" + index + "-label", entries.get(value));
            }
            else
            {
                script.getParameters().put(prefix + "-" + "-label", value);
            }
            index++;
        }
    }

    @Override
    public void dispose()
    {
        _enumeratorManager.dispose();
        _enumeratorManager = null;
    }
}
