/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.clientsideelement.content.SmartContentClientSideElementHelper;
import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Edit HMI item
 */
public class SmartContentClientSideElement extends StaticClientSideElement
{
    /** Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /** The content helper */
    protected ContentHelper _contentHelper;
    /** Helper for smart content client elements */
    protected SmartContentClientSideElementHelper _smartHelper;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentHelper = (ContentHelper) manager.lookup(ContentHelper.ROLE);
        _smartHelper = (SmartContentClientSideElementHelper) manager.lookup(SmartContentClientSideElementHelper.ROLE);
    }
    
    /**
     * Get informations on contents' state
     * @param contentsId the ids of contents
     * @return informations on contents' state
     */
    @Callable
    public Map<String, Object> getStatus(List<String> contentsId)
    {
        Map<String, Object> results = new HashMap<>();
        
        results.put("unmodifiable-contents", new ArrayList<>());
        results.put("locked-contents", new ArrayList<>());
        results.put("noright-contents", new ArrayList<>());
        results.put("invalidworkflowaction-contents", new ArrayList<>());
        results.put("invalidworkflowstep-contents", new ArrayList<>());
        results.put("allright-contents", new ArrayList<>());
        
        for (String contentId : contentsId)
        {
            Content content = _resolver.resolveById(contentId);
            
            boolean error = false;
            
            if (content instanceof WorkflowAwareContent)
            {
                // Is modifiable
                String enabledOnModifiableOnly = (String) this._script.getParameters().get("enabled-on-modifiable-only"); 
                if ("true".equals(enabledOnModifiableOnly) && !_isModifiable(content))
                {
                    Map<String, Object> contentParams = getContentDefaultParameters (content);
                    contentParams.put("description", _getNoModifiableDescription(content));
                    
                    @SuppressWarnings("unchecked")
                    List<Map<String, Object>> unModifiableContents = (List<Map<String, Object>>) results.get("unmodifiable-contents");
                    unModifiableContents.add(contentParams);

                    error = true;
                }
                
                // Is locked
                String enabledOnUnlockOnly = (String) this._script.getParameters().get("enabled-on-unlock-only"); 
                if ("true".equals(enabledOnUnlockOnly) && _isLocked(content))
                {
                    Map<String, Object> contentParams = getContentDefaultParameters (content);
                    contentParams.put("description", _getLockedDescription(content));
                    
                    @SuppressWarnings("unchecked")
                    List<Map<String, Object>> lockedContents = (List<Map<String, Object>>) results.get("locked-contents");
                    lockedContents.add(contentParams);
                    
                    error = true;
                }
                
                // Has right correct
                String enabledOnRightOnly = (String) this._script.getParameters().get("enabled-on-right-only"); 
                if ("true".equals(enabledOnRightOnly) && !_hasRight(content))
                {
                    Map<String, Object> contentParams = getContentDefaultParameters (content);
                    contentParams.put("description", _getNoRightDescription (content));
                    
                    @SuppressWarnings("unchecked")
                    List<Map<String, Object>> norightContents = (List<Map<String, Object>>) results.get("noright-contents");
                    norightContents.add(contentParams);
                    
                    error = true;
                }
                
                // Is workflow action correct
                String enabledOnWorkflowActionOnly = (String) this._script.getParameters().get("enabled-on-workflow-action-only"); 
                if (enabledOnWorkflowActionOnly != null)
                {
                    int actionId = _workflowAction(content);
                    if (actionId == -1)
                    {
                        Map<String, Object> contentParams = getContentDefaultParameters (content);
                        contentParams.put("description", _getWorkflowActionUnvailableDescription(content));
                        
                        @SuppressWarnings("unchecked")
                        List<Map<String, Object>> invalidActionContents = (List<Map<String, Object>>) results.get("invalidworkflowaction-contents");
                        invalidActionContents.add(contentParams);
                        
                        error = true;
                    }
                    else
                    {
                        results.put("workflowaction-content-actionId", actionId);
                    }
                }
                
                // Is workflow step correct
                String enabledOnWorkflowStepOnly = (String) this._script.getParameters().get("enabled-on-workflow-step-only"); 
                if (enabledOnWorkflowStepOnly != null && !_isWorkflowStepCorrect(content))
                {
                    Map<String, Object> contentParams = getContentDefaultParameters (content);
                    contentParams.put("description", _getIncorrectWorkflowStepDescription(content));
                    
                    @SuppressWarnings("unchecked")
                    List<Map<String, Object>> invalidStepContents = (List<Map<String, Object>>) results.get("invalidworkflowstep-contents");
                    invalidStepContents.add(contentParams);
                    
                    error = true;
                }
                
                if (_isAllRight (content, error, results))
                {
                    Map<String, Object> contentParams = getContentDefaultParameters (content);
                    contentParams.put("description", _getAllRightDescription(content));
                    
                    @SuppressWarnings("unchecked")
                    List<Map<String, Object>> allrightContents = (List<Map<String, Object>>) results.get("allright-contents");
                    allrightContents.add(contentParams);
                }
            }
        }
        
        return results;
    }
    
    /**
     * Determines if the user can finally do action on content
     * @param content The content
     * @param hasError true if a error has already occurs
     * @param results The result parameters to be passed to client side
     * @return true if the user can finally do action on content
     */
    protected boolean _isAllRight (Content content, boolean hasError, Map<String, Object> results)
    {
        return !hasError;
    }
    
    /**
     * Get the default content's parameters
     * @param content The content
     * @return The default parameters
     */
    protected Map<String, Object> getContentDefaultParameters (Content content)
    {
        return _smartHelper.getContentDefaultParameters(content);
    }
    
    /**
     * Determines if the content is locked
     * @param content the content
     * @return true if the content is locked
     */
    protected boolean _isLocked(Content content)
    {
        return _smartHelper.isLocked(content);
    }
    
    /**
     * Determines if the content is modifiable
     * @param content the content
     * @return true if the content is modifiable
     */
    protected boolean _isModifiable(Content content)
    {
        return _smartHelper.isModifiable(content);
    }
    
    /**
     * Determines if the user has sufficient right for the given content
     * @param content the content
     * @return true if user has sufficient right
     */
    protected boolean _hasRight(Content content)
    {
        return _smartHelper.hasRight(_rights, content);
    }
    
    /**
     * Determines if the workflow action is correct
     * @param content the content
     * @return true if the workflow action is incorrect
     */
    protected int _workflowAction(Content content)
    {
        return _smartHelper.workflowAction(_script.getParameters(), content);
    }
    
    /**
     * Determines if the workflow step is correct
     * @param content the content
     * @return true if the workflow step is incorrect
     */
    protected boolean _isWorkflowStepCorrect(Content content)
    {
        return _smartHelper.isWorkflowStepCorrect(_script.getParameters(), content);
    }
    
    /**
     * Get i18n description when user can not do action because he has no right on content
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    protected I18nizableText _getNoRightDescription (Content content)
    {
        return _smartHelper.getNoRightDescription(_script.getParameters(), content);
    }
    
    /**
     * Get i18n description when user can not do action because the content is locked
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    protected I18nizableText _getLockedDescription (Content content)
    {
        return _smartHelper.getLockedDescription(_script.getParameters(), content);
    }
    
    /**
     * Get i18n description when user can not do action because there is no workflow action unvailable
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    protected I18nizableText _getWorkflowActionUnvailableDescription (Content content)
    {
        return _smartHelper.getWorkflowActionUnvailableDescription(_script.getParameters(), content);
    }
    
    /**
     * Get i18n description when user can not do action because the workflow step is incorrect
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    protected I18nizableText _getIncorrectWorkflowStepDescription (Content content)
    {
        return _smartHelper.getIncorrectWorkflowStepDescription(_script.getParameters(), content);
    }
    
    /**
     * Get i18n description when user can not do action because the content is not modifiable
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    protected I18nizableText _getNoModifiableDescription (Content content)
    {
        return _smartHelper.getNoModifiableDescription(_script.getParameters(), content);
    }
    
    /**
     * Get i18n description when user can do action
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    protected I18nizableText _getAllRightDescription (Content content)
    {
        return _smartHelper.getAllRightDescription(_script.getParameters(), content);
    }
}
