/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement.content;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang.StringUtils;

import org.ametys.cms.clientsideelement.SmartContentClientSideElement;
import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.cms.workflow.AbstractContentWorkflowComponent;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.plugins.repository.lock.LockHelper;
import org.ametys.plugins.repository.lock.LockableAmetysObject;
import org.ametys.plugins.workflow.support.WorkflowProvider;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;
import org.ametys.runtime.i18n.I18nizableText;

import com.opensymphony.workflow.spi.Step;

/**
 * Component helper for {@link SmartContentClientSideElement}
 *
 */
public class SmartContentClientSideElementHelper implements Serviceable, Component
{
    /** The Avalon Role */
    public static final String ROLE = SmartContentClientSideElementHelper.class.getName();
    
    private UserManager _userManager;
    private WorkflowProvider _workflowProvider;
    private CurrentUserProvider _userProvider;
    private RightManager _rightManager;
    private ContentHelper _contentHelper;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
        _workflowProvider = (WorkflowProvider) manager.lookup(WorkflowProvider.ROLE);
        _userProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
        _rightManager = (RightManager) manager.lookup(RightManager.ROLE);
        _contentHelper = (ContentHelper) manager.lookup(ContentHelper.ROLE);
    }
    
    /**
     * Determines if the workflow step is correct
     * @param parameters The client side element 's parameters
     * @param content the content
     * @return true if the workflow step is incorrect
     */
    public boolean isWorkflowStepCorrect(Map<String, Object> parameters, Content content)
    {
        WorkflowAwareContent waContent = (WorkflowAwareContent) content;
        AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(waContent);
        long wId = waContent.getWorkflowId();
        List<Step> steps = workflow.getCurrentSteps(wId);
        
        List<Integer> stepIds = new ArrayList<>();
        for (Step step : steps)
        {
            stepIds.add(step.getStepId());
        }

        String correctStepsAsString = parameters.get("enabled-on-workflow-step-only").toString();
        String[] correctStepsAsStringArray = correctStepsAsString.split(", ?");
        
        for (String correctStepAsString : correctStepsAsStringArray)
        {
            if (!StringUtils.isEmpty(correctStepAsString) && stepIds.contains(Integer.parseInt(correctStepAsString)))
            {
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * Get i18n description when user can not do action because he has no right on content
     * @param parameters The client side element 's parameters
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    public I18nizableText getNoRightDescription (Map<String, Object> parameters, Content content)
    {
        List<String> rightI18nParameters = new ArrayList<>();
        rightI18nParameters.add(_contentHelper.getTitle(content));
        
        I18nizableText ed = (I18nizableText) parameters.get("noright-content-description");
        return  new I18nizableText(ed.getCatalogue(), ed.getKey(), rightI18nParameters);
    }
    
    /**
     * Get i18n description when user can not do action because the content is locked
     * @param parameters The client side element 's parameters
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    public I18nizableText getLockedDescription (Map<String, Object> parameters, Content content)
    {
        UserIdentity currentLockerIdentity = ((LockableAmetysObject) content).getLockOwner();
        User currentLocker = currentLockerIdentity != null ? _userManager.getUser(currentLockerIdentity.getPopulationId(), currentLockerIdentity.getLogin()) : null;
        
        List<String> lockI18nParameters = new ArrayList<>();
        lockI18nParameters.add(_contentHelper.getTitle(content));
        lockI18nParameters.add(currentLocker != null ? currentLocker.getFullName() : "");
        lockI18nParameters.add(currentLockerIdentity != null ? currentLockerIdentity.getLogin() : "Anonymous");
        
        I18nizableText ed = (I18nizableText) parameters.get("locked-content-description");
        return new I18nizableText(ed.getCatalogue(), ed.getKey(), lockI18nParameters);
    }
    
    /**
     * Get i18n description when user can not do action because there is no workflow action unvailable
     * @param parameters The client side element 's parameters
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    public I18nizableText getWorkflowActionUnvailableDescription (Map<String, Object> parameters, Content content)
    {
        List<String> workflowI18nParameters = new ArrayList<>();
        workflowI18nParameters.add(_contentHelper.getTitle(content));
        
        I18nizableText ed = (I18nizableText) parameters.get("workflowaction-content-description");
        return new I18nizableText(ed.getCatalogue(), ed.getKey(), workflowI18nParameters);
    }
    
    /**
     * Get i18n description when user can not do action because the workflow step is incorrect
     * @param parameters The client side element 's parameters
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    public I18nizableText getIncorrectWorkflowStepDescription (Map<String, Object> parameters, Content content)
    {
        List<String> workflowI18nParameters = new ArrayList<>();
        workflowI18nParameters.add(_contentHelper.getTitle(content));
        
        I18nizableText ed = (I18nizableText) parameters.get("workflowstep-content-description");
        return new I18nizableText(ed.getCatalogue(), ed.getKey(), workflowI18nParameters);
    }
    
    /**
     * Get i18n description when user can not do action because the content is not modifiable
     * @param parameters The client side element 's parameters
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    public I18nizableText getNoModifiableDescription (Map<String, Object> parameters, Content content)
    {
        List<String> modifiableI18nParameters = new ArrayList<>();
        modifiableI18nParameters.add(_contentHelper.getTitle(content));
        
        I18nizableText ed = (I18nizableText) parameters.get("nomodifiable-content-description");
        return new I18nizableText(ed.getCatalogue(), ed.getKey(), modifiableI18nParameters);
    }
    
    /**
     * Get i18n description when user can do action
     * @param parameters The client side element 's parameters
     * @param content The content
     * @return The {@link I18nizableText} description
     */
    public I18nizableText getAllRightDescription (Map<String, Object> parameters, Content content)
    {
        List<String> allrightI18nParameters = new ArrayList<>();
        allrightI18nParameters.add(_contentHelper.getTitle(content));
        
        I18nizableText ed = (I18nizableText) parameters.get("allright-content-description");
        return new I18nizableText(ed.getCatalogue(), ed.getKey(), allrightI18nParameters);
    }
    
    /**
     * Determines if the content is locked
     * @param content the content
     * @return true if the content is locked
     */
    public boolean isLocked(Content content)
    {
        if (content instanceof LockableAmetysObject)
        {
            LockableAmetysObject lockableContent = (LockableAmetysObject) content;
            if (lockableContent.isLocked() && !LockHelper.isLockOwner(lockableContent, _userProvider.getUser()))
            {
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * Determines if the content is modifiable
     * @param content the content
     * @return true if the content is modifiable
     */
    public boolean isModifiable(Content content)
    {
        return content instanceof ModifiableContent;
    }
    
    /**
     * Determines if the user has sufficient right for the given content
     * @param rights The client side element 's rights
     * @param content the content
     * @return true if user has sufficient right
     */
    public boolean hasRight (Map<String, String> rights, Content content)
    {
        if (rights.isEmpty())
        {
            return true;
        }
        
        Set<String> rightsToCheck = rights.keySet();
        for (String rightToCheck : rightsToCheck)
        {
            if (StringUtils.isNotEmpty(rightToCheck))
            {
                if (_rightManager.hasRight(_userProvider.getUser(), rightToCheck, content) == RightResult.RIGHT_ALLOW)
                {
                    return true;
                }
            }
        }
        
        return false;
    }
    
    /**
     * Determines if the workflow action is correct
     * @param parameters The client side element 's parameters
     * @param content the content
     * @return true if the workflow action is incorrect
     */
    public int workflowAction(Map<String, Object> parameters, Content content)
    {
        WorkflowAwareContent waContent = (WorkflowAwareContent) content;
        AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(waContent);
        long wId = waContent.getWorkflowId();
        
        Map<String, Object> vars = new HashMap<>();
        vars.put(AbstractContentWorkflowComponent.CONTENT_KEY, content);
        int[] actionIds = workflow.getAvailableActions(wId, vars);
        Arrays.sort(actionIds);
        
        String correctActionsAsString = parameters.get("enabled-on-workflow-action-only").toString();
        String[] correctActionsAsStringArray = correctActionsAsString.split(", ?");
        
        int correctActionId = -1;
        for (String correctActionAsString : correctActionsAsStringArray)
        {
            int actionId = Integer.parseInt(correctActionAsString);
            if (!StringUtils.isEmpty(correctActionAsString) && Arrays.binarySearch(actionIds, actionId) >= 0)
            {
                correctActionId = actionId;
                break;
            }
        }
        
        return correctActionId;
    }

    /**
     * Get the default content's parameters
     * @param content The content
     * @return The default parameters
     */
    public Map<String, Object> getContentDefaultParameters(Content content)
    {
        Map<String, Object> contentParams = new HashMap<>();
        contentParams.put("id", content.getId());
        contentParams.put("title", _contentHelper.getTitle(content));
        
        return contentParams;
    }
}
