/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement.htmlexpert;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.components.ContextHelper;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.ui.ClientSideElement;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.plugins.core.ui.help.HelpManager;
import org.ametys.plugins.core.ui.util.ConfigurationHelper;

/**
 * This client side element completes its configuration by loading HTMLExpertToolbarExtensionPoint buttons 
 */
public class HTMLExpertEditClientSideElement extends StaticClientSideElement implements Contextualizable
{
    /** The HTMLExpertToolbarExtensionPoint instance */
    protected HTMLExpertToolbarExtensionPoint _htmlExpertToolbarExtensionPoint;
    
    /** The helper manager */
    protected HelpManager _helpManager;
    
    /** The context */
    protected Context _context;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _htmlExpertToolbarExtensionPoint = (HTMLExpertToolbarExtensionPoint) smanager.lookup(HTMLExpertToolbarExtensionPoint.ROLE);
        _helpManager = (HelpManager) smanager.lookup(HelpManager.ROLE);
    }
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public List<Script> getScripts(boolean ignoreRights, Map<String, Object> contextParameters)
    {
        List<ScriptFile> cssFiles = new ArrayList<>();
        List<ScriptFile> scriptFiles = new ArrayList<>();
        Map<String, Object> parameters = ConfigurationHelper.clonePluginParameters(_script.getParameters());
        
        _addHelpURL(parameters);
        
        parameters.put("toolbar-buttons", new ArrayList<>());
        
        for (String extensionId : _htmlExpertToolbarExtensionPoint.getExtensionsIds())
        {
            ClientSideElement button = _htmlExpertToolbarExtensionPoint.getExtension(extensionId);

            Map<String, Object> btnParams = new HashMap<>();
            
            btnParams.put("id", extensionId);
            btnParams.put("pluginName", button.getPluginName());
            
            Map<String, Object> btnConfig = new HashMap<>();
            for (Script script : button.getScripts(ignoreRights, contextParameters))
            {
                if (script != null)
                {
                    btnParams.put("classname", script.getScriptClassname());
                    for (String parameterKey : script.getParameters().keySet())
                    {
                        Object parameter = script.getParameters().get(parameterKey);
                        btnConfig.put(parameterKey, parameter);
                    }
                }
            }
            
            btnParams.put("config", btnConfig);
            
            ((ArrayList<Map<String, Object>>) parameters.get("toolbar-buttons")).add(btnParams);
        }
        
        
        List<Script> scripts = super.getScripts(ignoreRights, contextParameters);
        
        if (scripts.size() == 1)
        {
            Script script = scripts.get(0);
            
            cssFiles.addAll(script.getCSSFiles());
            scriptFiles.addAll(script.getScriptFiles());
            ConfigurationHelper.clonePluginParameters(script.getParameters());
            
            for (String extensionId : _htmlExpertToolbarExtensionPoint.getExtensionsIds())
            {
                ClientSideElement button = _htmlExpertToolbarExtensionPoint.getExtension(extensionId);
                for (Script btnScript : button.getScripts(ignoreRights, contextParameters))
                {
                    if (btnScript != null)
                    {
                        cssFiles.addAll(btnScript.getCSSFiles());
                        scriptFiles.addAll(btnScript.getScriptFiles());
                    }
                }
            }
            
            List<Script> scriptList = new ArrayList<>();
            scriptList.add(new Script(this.getId(), _script.getScriptClassname(), scriptFiles, cssFiles, parameters));
            return scriptList;
        }
        
        return new ArrayList<>();
    }

    /**
     * Add the help url in parameters
     * @param parameters the parameters
     */
    protected void _addHelpURL(Map<String, Object> parameters)
    {
        Map objectModel = ContextHelper.getObjectModel(_context);
        
        Locale locale = org.apache.cocoon.i18n.I18nUtils.findLocale(objectModel, "locale", null, Locale.getDefault(), true);
        String langCode = locale.getLanguage();
        
        String familyId = "org.ametys.core.ui.RibbonControlsManager";
        String id = "org.ametys.cms.edition.htmlexpert.Edit";
        try
        {
            String helpURL = _helpManager.getHelp(familyId, id, langCode);
            if (StringUtils.isNotBlank(helpURL))
            {
                parameters.put("helpURL", helpURL);
            }
        }
        catch (Exception e)
        {
            getLogger().error("An error occurred getting help url for family '{}' and id '{}'", familyId, id, e);
        }
    }
}
