/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement.styles;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.core.ui.SimpleMenu;
import org.ametys.core.util.I18nUtils;
import org.ametys.plugins.core.ui.util.ConfigurationHelper;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * This StaticMenu also search for special tags to transform them into parameters thanks to HTMLEditorStyle
 * Special tags are
 * <ul>
 *  <li>table</li>
 * </ul> 
 */
public abstract class AbstractEditorStyleMenu extends SimpleMenu
{
    /** The HTMLEditorStyle instance */
    protected HTMLEditorStyle _htmlEditorStyle;
    /** I18n helper */
    protected I18nUtils _i18nUtils;
    
    /** The default value */
    protected Map<String, Object> _defaultValues = new HashMap<>();
    /** The label of auto group */
    protected I18nizableText _autoGroupLabel;
    /** JS class name for auto items */
    protected String _autoItemClassName;
    
    /**
     * Return the {@link StyleCategory}s to use
     * @param contextParameters Contextual parameters transmitted by the environment.
     * @return A map with style category by category
     */
    protected abstract Map<String, StyleCategory> _getStyleCategories(Map<String, Object> contextParameters);
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _htmlEditorStyle = (HTMLEditorStyle) smanager.lookup(HTMLEditorStyle.ROLE);
        _i18nUtils = (I18nUtils) smanager.lookup(I18nUtils.ROLE);
    }

    @Override
    public List<Script> getScripts(boolean ignoreRights, Map<String, Object> contextParameters)
    {
        List<Script> scripts = super.getScripts(ignoreRights, contextParameters);
        
        if (scripts.size() == 0)
        {
            return scripts;
        }
        
        List<ScriptFile> cssFiles = new ArrayList<>();
        List<ScriptFile> jsFiles = new ArrayList<>();
        Map<String, Object> parameters = new HashMap<>();
        for (Script script : scripts)
        {
            cssFiles.addAll(script.getCSSFiles());
            jsFiles.addAll(script.getScriptFiles());
            parameters.putAll(getParameters(ConfigurationHelper.clonePluginParameters(script.getParameters()), contextParameters));
        }
        
        Map<String, StyleCategory> styleCategories = _getStyleCategories(contextParameters);
        for (StyleCategory styleCategory : styleCategories.values())
        {
            for (String cssFile : styleCategory.getBackOfficeCSSFiles())
            {
                if (cssFile != null)
                {
                    cssFiles.add(new ScriptFile(cssFile));
                }
            }
        }
        
        List<Script> scriptList = new ArrayList<>();
        scriptList.add(new Script(this.getId(), _script.getScriptClassname(), jsFiles, cssFiles, parameters));
        return scriptList;
    }

    private Map<String, Object> getParameters(Map<String, Object> parameters, Map<String, Object> contextualParameters)
    {
        if (_galleryItems.size() > 0 && _autoGroupLabel != null)
        {
            @SuppressWarnings("unchecked")
            Map<String, Object> galleryItems = (Map<String, Object>) parameters.get("gallery-item");
            @SuppressWarnings("unchecked")
            List<Map<String, Object>> galleryGroups = (List<Map<String, Object>>) galleryItems.get("gallery-groups");
            
            for (Map<String, Object> groupParams : galleryGroups)
            {
                if (groupParams.get("label").equals(_autoGroupLabel))
                {
                    List<Map<String, Object>> gpItems = new ArrayList<>();
                    
                    Map<String, StyleCategory> styleCategories = _getStyleCategories(contextualParameters);
                    for (Entry<String, StyleCategory> entry : styleCategories.entrySet())
                    {
                        for (Style style : entry.getValue().getStyles())
                        {
                            Map<String, Object> itemCfg = new HashMap<>();
                            
                            itemCfg.put("className", _autoItemClassName);
                            
                            Map<String, Object> styleConfig = new HashMap<>();
                            styleConfig.put("label", style.getButtonLabel());
                            styleConfig.put("description", style.getButtonDescription());
                            styleConfig.put("category", entry.getKey());
                            
                            String smallIcon = style.getButtonSmallIcon();
                            if (smallIcon != null)
                            {
                                styleConfig.put("icon-small", smallIcon);
                            }
                            
                            String mediumIcon = style.getButtonMediumIcon();
                            if (mediumIcon != null)
                            {
                                styleConfig.put("icon-medium", mediumIcon);
                            }
                            
                            String largeIcon = style.getButtonLargeIcon();
                            if (largeIcon != null)
                            {
                                styleConfig.put("icon-large", largeIcon);
                            }
                            
                            String buttonCSSClass = style.getButtonCSSClass();
                            if (buttonCSSClass != null)
                            {
                                styleConfig.put("cls", buttonCSSClass);
                            }
                            
                            styleConfig.put("css-class", style.getInlineEditorRender());
                            styleConfig.putAll(_defaultValues);
                            
                            itemCfg.put("config", styleConfig);
                            gpItems.add(itemCfg);
                        }
                    }
                    
                    groupParams.put("items", gpItems);
                }
            }
        }
        
        return parameters;
    }
    
    @Override
    protected GalleryGroup _configureGroupGallery(Configuration configuration) throws ConfigurationException
    {
        GalleryGroup group = super._configureGroupGallery(configuration);
        
        if (configuration.getChild("auto", false) != null)
        {
            // Register default values from auto configuration
            Configuration autoConfig = configuration.getChild("auto");
            _defaultValues = _configureParameters (autoConfig);
            
            _autoGroupLabel = group.getLabel();
            _autoItemClassName = autoConfig.getAttribute("className");
        }
        
        return group;
    }
}
