/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement.styles;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ametys.core.ui.ClientSideElement.ScriptFile;

/**
 * This extension point allows to determine which styles does exist for html edition 
 */
public interface HTMLEditorStyle
{
    /** Avalon role name */
    public static final String ROLE = HTMLEditorStyle.class.getName();
    
    /**
     * Get the categories supported by this component
     * @return The categories supported. Cannot be null, but may be empty and of course contains the "" key for the default category.
     */
    public Set<String> getCategories();
    
    /**
     * Get the list of styles for paragraphs
     * @param category The richtext category
     * @param contextualParameters Contextuals parameters transmitted by the environment.
     * @return A category. Can be null.
     */
    public StyleCategory getPara(String category, Map<String, Object> contextualParameters);
    
    /**
     * Get the list of styles for tables
     * @param category The richtext category
     * @param contextualParameters Contextuals parameters transmitted by the environment.
     * @return A category. Can be null.
     */
    public StyleCategory getTable(String category, Map<String, Object> contextualParameters);
    
    /**
     * Get the list of styles for link
     * @param category The richtext category
     * @param contextualParameters Contextuals parameters transmitted by the environment.
     * @return A category. Can be null.
     */
    public StyleCategory getLink(String category, Map<String, Object> contextualParameters);
    
    /**
     * Get the list of styles for images
     * @param category The richtext category
     * @param contextualParameters Contextuals parameters transmitted by the environment.
     * @return A category. Can be null.
     */
    public StyleCategory getImage(String category, Map<String, Object> contextualParameters);
    
    /**
     * Get the list of styles for unordered lists
     * @param category The richtext category
     * @param contextualParameters Contextuals parameters transmitted by the environment.
     * @return A category. Can be null.
     */
    public StyleCategory getUnorderedList(String category, Map<String, Object> contextualParameters);
    
    /**
     * Get the list of styles for ordered lists
     * @param category The richtext category
     * @param contextualParameters Contextuals parameters transmitted by the environment.
     * @return A category. Can be null.
     */
    public StyleCategory getOrderedList(String category, Map<String, Object> contextualParameters);
    
    /**
     * Get the aggregation of all backoffice css files
     * @param category The richtext category
     * @param contextParameters The contextual parameters
     * @return A non null list of ScriptFile
     */
    public default List<ScriptFile> getBackOfficeCSSFiles(String category, Map<String, Object> contextParameters)
    {
        List<ScriptFile> cssFiles = new ArrayList<>();
        
        List<StyleCategory> styleCategories = new ArrayList<>();
        styleCategories.add(getImage(category, contextParameters));
        styleCategories.add(getLink(category, contextParameters));
        styleCategories.add(getOrderedList(category, contextParameters));
        styleCategories.add(getPara(category, contextParameters));
        styleCategories.add(getTable(category, contextParameters));
        styleCategories.add(getUnorderedList(category, contextParameters));
        
        Set<String> filesAsString = new LinkedHashSet<>();
        for (StyleCategory styleCategory : styleCategories)
        {
            if (styleCategory != null)
            {
                filesAsString.addAll(styleCategory.getBackOfficeCSSFiles());
            }
        }
        filesAsString.remove(null);
        
        for (String fileAsString : filesAsString)
        {
            cssFiles.add(new ScriptFile(fileAsString));
        }
        
        return cssFiles;
    }
    
    /**
     * Get the aggregation of all inline editor css files
     * @param category The richtext category
     * @param contextParameters The contextual parameters
     * @return A non null list of ScriptFile
     */
    public default List<ScriptFile> getInlineEditorCSSFiles(String category, Map<String, Object> contextParameters)
    {
        List<ScriptFile> cssFiles = new ArrayList<>();
        
        List<StyleCategory> styleCategories = new ArrayList<>();
        styleCategories.add(getImage(category, contextParameters));
        styleCategories.add(getLink(category, contextParameters));
        styleCategories.add(getOrderedList(category, contextParameters));
        styleCategories.add(getPara(category, contextParameters));
        styleCategories.add(getTable(category, contextParameters));
        styleCategories.add(getUnorderedList(category, contextParameters));
        
        Set<String> filesAsString = new LinkedHashSet<>();
        for (StyleCategory styleCategory : styleCategories)
        {
            if (styleCategory != null)
            {
                filesAsString.addAll(styleCategory.getInlineEditorCSSFiles());
            }
        }
        filesAsString.remove(null);
        
        for (String fileAsString : filesAsString)
        {
            cssFiles.add(new ScriptFile(fileAsString));
        }
        
        return cssFiles;
    }
}
