/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement.styles;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.core.ui.ClientSideElement;
import org.ametys.core.ui.ClientSideElement.ScriptFile;
import org.ametys.core.ui.widgets.richtext.RichTextConfiguration;
import org.ametys.core.ui.widgets.richtext.RichTextConfigurationAttribute;
import org.ametys.core.ui.widgets.richtext.RichTextConfigurationStyle;
import org.ametys.core.ui.widgets.richtext.RichTextConfigurationStyleGroup;
import org.ametys.core.ui.widgets.richtext.RichTextConfigurationTag;
import org.ametys.core.ui.widgets.richtext.RichTextConfigurationTag.EMPTY_TAG;
import org.ametys.core.ui.widgets.richtext.StaticRichTextConfiguration.StaticRichTextConfigurationAttribute;
import org.ametys.core.ui.widgets.richtext.StaticRichTextConfiguration.StaticRichTextConfigurationStyle;
import org.ametys.core.ui.widgets.richtext.StaticRichTextConfiguration.StaticRichTextConfigurationStyleGroup;
import org.ametys.core.ui.widgets.richtext.StaticRichTextConfiguration.StaticRichTextConfigurationTag;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Declares all rich text configurations relative the {@link HTMLEditorStyle}
 */
public class HTMLEditorStyleRichTextConfiguration implements RichTextConfiguration, Serviceable
{
    private static final Map<String, I18nizableText> STYLE_GROUP_LABELS = new HashMap<>();
    private HTMLEditorStyle _htmlEditorStyle;

    static 
    {
        STYLE_GROUP_LABELS.put("image", new I18nizableText("plugin.cms", "CONTENT_EDITION_IMAGE_STYLES_GROUP_STYLED"));
        STYLE_GROUP_LABELS.put("link", new I18nizableText("plugin.cms", "CONTENT_EDITION_LINK_STYLES_GROUP_SPECIAL"));
        STYLE_GROUP_LABELS.put("ol", new I18nizableText("plugin.cms", "CONTENT_EDITION_CHARACTER_ORDEREDLIST_STYLES_GROUP_STYLED"));
        STYLE_GROUP_LABELS.put("paragraph", new I18nizableText("TODO HTMLEditorStyleRichTextConfiguration")); // TODO
        STYLE_GROUP_LABELS.put("table", new I18nizableText("plugin.cms", "CONTENT_EDITION_TABLE_STYLES_GROUP_STYLED"));
        STYLE_GROUP_LABELS.put("ul", new I18nizableText("plugin.cms", "CONTENT_EDITION_CHARACTER_UNORDEREDLIST_STYLES_GROUP_STYLED"));
    }

    public void service(ServiceManager manager) throws ServiceException
    {
        _htmlEditorStyle = (HTMLEditorStyle) manager.lookup(HTMLEditorStyle.ROLE);
    }
    
    public Set<String> getCategories()
    {
        return _htmlEditorStyle.getCategories();
    }
    
    public Collection<RichTextConfigurationTag> getHandledTags(String category, Map<String, Object> contextParameters)
    {
        List<RichTextConfigurationTag> tags = new ArrayList<>();
        
        if (getCategories().contains(category))
        {
            Map<String, StyleCategory> styleCategories = new HashMap<>();
            styleCategories.put("img", _htmlEditorStyle.getImage(category, contextParameters));
            styleCategories.put("a", _htmlEditorStyle.getLink(category, contextParameters));
            styleCategories.put("ol", _htmlEditorStyle.getOrderedList(category, contextParameters));
            styleCategories.put("table", _htmlEditorStyle.getTable(category, contextParameters));
            styleCategories.put("ul", _htmlEditorStyle.getUnorderedList(category, contextParameters));
            
            for (Entry<String, StyleCategory> entry : styleCategories.entrySet())
            {
                if (entry.getValue() != null)
                {
                    Set<RichTextConfigurationAttribute> attributes = new HashSet<>();
                    Set<String> authorizedValues = new LinkedHashSet<>();
                    attributes.add(new StaticRichTextConfigurationAttribute("class", null, authorizedValues, null));
                    RichTextConfigurationTag tag = new StaticRichTextConfigurationTag(entry.getKey(), null, null, attributes);
                    tags.add(tag);
                    
                    for (Style style : entry.getValue().getStyles())
                    {
                        authorizedValues.add(style.getInlineEditorRender());
                    }
                }
            }
            
            StyleCategory paraStyleCategory = _htmlEditorStyle.getPara(category, contextParameters);
            if (paraStyleCategory != null)
            {
                for (Style style : paraStyleCategory.getStyles())
                {
                    String tagName = style.getInlineEditorRender();
                    Set<RichTextConfigurationAttribute> attributes = new HashSet<>();
                    Set<String> authorizedValues = new HashSet<>();
                    authorizedValues.add("text-align");
                    attributes.add(new StaticRichTextConfigurationAttribute("style", null, authorizedValues, null));
                    
                    Set<String> authorizedClassValues = new HashSet<>();
                    
                    int dotPosition = tagName.indexOf('.');
                    if (dotPosition >= 0)
                    {
                        String className = tagName.substring(dotPosition + 1);
                        tagName = tagName.substring(0, dotPosition);
                        
                        authorizedClassValues.add(className);
                    }
                    
                    attributes.add(new StaticRichTextConfigurationAttribute("class", null, authorizedClassValues, null));
                    
                    RichTextConfigurationTag tag = new StaticRichTextConfigurationTag(tagName, "p".equals(tagName) ? null : EMPTY_TAG.REMOVE_EMPTY_CONTENT, null, attributes);
                    tags.add(tag);
                }                
            }
        }
        
        return tags;
    }
    
    public List<ScriptFile> getCSSFiles(String category, Map<String, Object> contextParameters)
    {
        return _htmlEditorStyle.getInlineEditorCSSFiles(category, contextParameters);
    }
    
    public Set<ClientSideElement> getConvertors(String category, Map<String, Object> contextParameters)
    {
        return Collections.EMPTY_SET;
    }
    
    public Set<ClientSideElement> getValidators(String category, Map<String, Object> contextParameters)
    {
        return Collections.EMPTY_SET;
    }
    
    public Map<String, Map<RichTextConfigurationStyleGroup, List<RichTextConfigurationStyle>>> getAvailableStyles(String category, Map<String, Object> contextualParameters)
    {
        Map<String, Map<RichTextConfigurationStyleGroup, List<RichTextConfigurationStyle>>> lists = new HashMap<>();
        
        if (getCategories().contains(category))
        {
            Map<String, StyleCategory> styleCategories = new HashMap<>();
            
            styleCategories.put("image", _htmlEditorStyle.getImage(category, contextualParameters));
            styleCategories.put("link", _htmlEditorStyle.getLink(category, contextualParameters));
            styleCategories.put("ol", _htmlEditorStyle.getOrderedList(category, contextualParameters));
            styleCategories.put("paragraph", _htmlEditorStyle.getPara(category, contextualParameters));
            styleCategories.put("table", _htmlEditorStyle.getTable(category, contextualParameters));
            styleCategories.put("ul", _htmlEditorStyle.getUnorderedList(category, contextualParameters));
            
            for (Entry<String, StyleCategory> entry : styleCategories.entrySet())
            {
                if (entry.getValue() != null)
                {
                    Map<RichTextConfigurationStyleGroup, List<RichTextConfigurationStyle>> rtcStylesGroup = new HashMap<>();
                    lists.put(entry.getKey(), rtcStylesGroup);
                    
                    List<RichTextConfigurationStyle> rtcStyles = new ArrayList<>();
                    rtcStylesGroup.put(new StaticRichTextConfigurationStyleGroup(STYLE_GROUP_LABELS.get(entry.getKey()), 1000), rtcStyles);
                    
                    for (Style style : entry.getValue().getStyles())
                    {
                        rtcStyles.add(new StaticRichTextConfigurationStyle(style.getInlineEditorRender(), style.getButtonLabel(), style.getButtonDescription(), style.getButtonCSSClass(), style.getButtonSmallIcon(), style.getButtonMediumIcon(), style.getButtonLargeIcon(), !"paragraph".equals(entry.getKey())));
                    }
                }
            }
        }
        
        return lists;
    }
}
