/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement.styles;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.SourceResolver;

import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.PluginAware;

/**
 * This implementation is based upon a static configuration.
 * FIXME For now, this implementation only supports styles for contents ('content' category). It cannot provided styles for other richtext 
 */
public class StaticHTMLEditorContentStyle extends AbstractLogEnabled implements HTMLEditorStyle, Configurable, Serviceable, PluginAware
{
    /** Default icons style for table */
    protected static final String[] TABLE_DEFAULT_ICONS = new String[] {"/plugins/cms/resources/img/content/edition/table/Table_Border_16.png", "/plugins/cms/resources/img/content/edition/table/Table_Border_32.png", "/plugins/cms/resources/img/content/edition/table/Table_Border_32.png"};
    /** Default icons style for link */
    protected static final String[] LINK_DEFAULT_ICONS = new String[] {"/plugins/cms/resources/img/content/edition/link/Style_Normal_16.png", "/plugins/cms/resources/img/content/edition/link/Style_Normal_32.png", "/plugins/cms/resources/img/content/edition/link/Style_Normal_32.png"};
    /** Default icons style for image */
    protected static final String[] IMAGE_DEFAULT_ICONS = new String[] {"/plugins/cms/resources/img/content/edition/image/Image_Style_Border_16.png", "/plugins/cms/resources/img/content/edition/link/Style_Normal_32.png", "/plugins/cms/resources/img/content/edition/link/Style_Normal_32.png"};
    /** Default icons style for unordered list */
    protected static final String[] UL_DEFAULT_ICONS = new String[] {"/plugins/cms/resources/img/content/edition/list/disc_16.png", "/plugins/cms/resources/img/content/edition/link/Style_Normal_32.png", "/plugins/cms/resources/img/content/edition/link/Style_Normal_32.png"};
    /** Default icons style for ordered list */
    protected static final String[] OL_DEFAULT_ICONS = new String[] {"/plugins/cms/resources/img/content/edition/list/decimal_16.png", "/plugins/cms/resources/img/content/edition/link/Style_Normal_32.png", "/plugins/cms/resources/img/content/edition/link/Style_Normal_32.png"};
    
    /** Excalibur source resolver */
    protected SourceResolver _resolver;
    /** The values of the file */
    protected StyleCategory _paraStyleCategory;
    /** The values of the file */
    protected StyleCategory _tableStyleCategory;
    /** The values of the file */
    protected StyleCategory _linkStyleCategory;
    /** The values of the file */
    protected StyleCategory _imageStyleCategory;
    /** The values of the file */
    protected StyleCategory _ulStyleCategory;
    /** The values of the file */
    protected StyleCategory _olStyleCategory;
    /** The plugin that declared the instance */
    protected String _pluginName;
    /** The feature that declared the instance */
    protected String _featureName;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    @Override
    public void setPluginInfo(String pluginName, String featureName, String id)
    {
        _pluginName = pluginName;
        _featureName = featureName;
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        _configure(configuration, "param.edition-styles", new HashMap<>());
    }
    
    /**
     * Configure this instance upon a configuration
     * @param configuration The configuration
     * @param defaultCatalogue The default i18n catalogue to use
     * @param contextParameters Contextuals parameters transmitted by the environment.
     * @throws ConfigurationException If the configuration is incorrect
     */
    protected void _configure(Configuration configuration, String defaultCatalogue, Map<String, Object> contextParameters) throws ConfigurationException
    {
        _paraStyleCategory = _configureStyleCategory(configuration.getChild("para"), defaultCatalogue, null, contextParameters);
        _tableStyleCategory = _configureStyleCategory(configuration.getChild("table"), defaultCatalogue, TABLE_DEFAULT_ICONS, contextParameters);
        _linkStyleCategory = _configureStyleCategory(configuration.getChild("link"), defaultCatalogue, LINK_DEFAULT_ICONS, contextParameters);
        _imageStyleCategory = _configureStyleCategory(configuration.getChild("image"), defaultCatalogue, IMAGE_DEFAULT_ICONS, contextParameters);
        _ulStyleCategory = _configureStyleCategory(configuration.getChild("unorderedlist"), defaultCatalogue, UL_DEFAULT_ICONS, contextParameters);
        _olStyleCategory = _configureStyleCategory(configuration.getChild("orderedlist"), defaultCatalogue, OL_DEFAULT_ICONS, contextParameters);
    }
    
    /**
     * Configure this instance upon the style part of a configuration
     * @param styleConfiguration The style part of the configuration
     * @param defaultCatalogue The default i18n catalogue to use
     * @param defaultIcons the path to the button's default icons in small, medium and large size
     * @param contextParameters Contextuals parameters transmitted by the environment.
     * @return The style that represents the values read
     * @throws ConfigurationException If the configuration is incorrect
     */
    protected StyleCategory _configureStyleCategory(Configuration styleConfiguration, String defaultCatalogue, String[] defaultIcons, Map<String, Object> contextParameters) throws ConfigurationException
    {
        Set<String> boCSSFiles = new LinkedHashSet<>();
        Set<String> inlineEditorCSSFiles = new LinkedHashSet<>();
        
        for (Configuration configuration : styleConfiguration.getChild("import").getChildren("button"))
        {
            boCSSFiles.add(_getImport(configuration, contextParameters));
            // adding null will be ignored later
        }
        for (Configuration configuration : styleConfiguration.getChild("import").getChildren("inline-editor"))
        {
            inlineEditorCSSFiles.add(_getImport(configuration, contextParameters));
            // adding null will be ignored later
        }
        
        List<Style> styleList = new ArrayList<>();
        for (Configuration styleConf : styleConfiguration.getChildren("style"))
        {
            Style style = _configureStyle(styleConf, defaultCatalogue, defaultIcons, contextParameters);
            styleList.add(style);
        }
        
        return new StyleCategory(boCSSFiles, inlineEditorCSSFiles, styleList);
    }
    
    /**
     * Determine an url from a configuration node.
     * @param configuration Read the optional plugin attribute and concatenate it to the value of the node. Can be null
     * @param contextParameters Contextuals parameters transmitted by the environment.
     * @return The non-null path to the plugin resource
     */
    protected String _getImport(Configuration configuration, Map<String, Object> contextParameters)
    {
        String boPlugin = configuration.getAttribute("plugin", null);
        
        String boFileLocation;
        if (StringUtils.isNotBlank(boPlugin))
        {
            boFileLocation = "/plugins/" + boPlugin + "/resources/";
        }
        else
        {
            boFileLocation = "/param_resources/edition-styles/";
        }
        
        String uri = configuration.getValue(null);
        return StringUtils.isBlank(uri) ? null : boFileLocation + uri;
    }
    
    /**
     * Configure a style
     * @param configuration The style configuration
     * @param defaultCatalogue The default i18n catalogue to use
     * @param defaultIcons the path to the button's default icons in small, medium and large size
     * @param contextParameters Contextual parameters transmitted by the environment.
     * @return The style configured
     * @throws ConfigurationException If the configuration is incorrect
     */
    protected Style _configureStyle(Configuration configuration, String defaultCatalogue, String[] defaultIcons, Map<String, Object> contextParameters) throws ConfigurationException
    {
        String buttonSmallIcon = defaultIcons != null ? defaultIcons[0] : null;
        String buttonMediumIcon = defaultIcons != null ? defaultIcons[1] : null;
        String buttonLargeIcon = defaultIcons != null ? defaultIcons[2] : null;
        I18nizableText buttonLabel;
        I18nizableText buttonDescription;
        String buttonCSSClass;

        Configuration iconSmallConf = configuration.getChild("button").getChild("icon-small", false);
        if (iconSmallConf != null)
        {
            buttonSmallIcon = _getImport(iconSmallConf, contextParameters);
        }
        
        Configuration iconMediumConf = configuration.getChild("button").getChild("icon-medium", false);
        if (iconMediumConf == null)
        {
            // Try to look for "icon" if no "icon-medium".
            iconMediumConf = configuration.getChild("button").getChild("icon", false);
        }
        
        if (iconMediumConf != null)
        {
            buttonMediumIcon = _getImport(iconMediumConf, contextParameters);
        }
        
        Configuration iconLargeConf = configuration.getChild("button").getChild("icon-large", false);
        if (iconLargeConf == null)
        {
            // Try to look for "icon" if no "icon-large".
            iconLargeConf = configuration.getChild("button").getChild("icon", false);
        }
        
        if (iconLargeConf != null)
        {
            buttonLargeIcon = _getImport(iconLargeConf, contextParameters);
        }
        
        Configuration labelConf = configuration.getChild("button").getChild("label");
        if (labelConf.getAttributeAsBoolean("i18n", false))
        {
            buttonLabel = new I18nizableText(defaultCatalogue, labelConf.getValue());
        }
        else 
        {
            buttonLabel = new I18nizableText(labelConf.getValue());
        }

        Configuration descriptionConf = configuration.getChild("button").getChild("description");
        if (descriptionConf.getAttributeAsBoolean("i18n", false))
        {
            buttonDescription = new I18nizableText(defaultCatalogue, descriptionConf.getValue());
        }
        else 
        {
            buttonDescription = new I18nizableText(descriptionConf.getValue());
        }
        
        buttonCSSClass = configuration.getChild("button").getChild("cssclass").getValue("");
        
        String inlineEditorRender = configuration.getChild("inline-editor").getValue();
        
        return new Style(buttonLabel, buttonDescription, buttonSmallIcon, buttonMediumIcon, buttonLargeIcon, buttonCSSClass, inlineEditorRender);
    }
    
    @Override
    public Set<String> getCategories()
    {
        return Set.of("content");
    }

    @Override
    public StyleCategory getPara(String category, Map<String, Object> contextualParameters)
    {
        return "content".equals(category) ? _paraStyleCategory : null;
    }
    
    @Override
    public StyleCategory getTable(String category, Map<String, Object> contextualParameters)
    {
        return  "content".equals(category) ? _tableStyleCategory : null;
    }
    
    @Override
    public StyleCategory getLink(String category, Map<String, Object> contextualParameters)
    {
        return "content".equals(category) ? _linkStyleCategory : null;
    }
    
    @Override
    public StyleCategory getImage(String category, Map<String, Object> contextualParameters)
    {
        return  "content".equals(category) ? _imageStyleCategory : null;
    }
    
    @Override
    public StyleCategory getUnorderedList(String category, Map<String, Object> contextualParameters)
    {
        return  "content".equals(category) ? _ulStyleCategory : null;
    }
    
    @Override
    public StyleCategory getOrderedList(String category, Map<String, Object> contextualParameters)
    {
        return  "content".equals(category) ? _olStyleCategory : null;
    }
}
