/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.clientsideelement.styles;

import java.util.HashMap;
import java.util.Map;

import org.apache.cocoon.xml.XMLUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * This class handle style for paragraphs 
 */
public class Style
{
    /** See constructor */
    protected String _buttonSmallIcon;
    /** See constructor */
    protected String _buttonMediumIcon;
    /** See constructor */
    protected String _buttonLargeIcon;
    /** See constructor */
    protected I18nizableText _buttonLabel;
    /** See constructor */
    protected I18nizableText _buttonDescription;
    /** See constructor */
    protected String _buttonCSSClass;
    /** See constructor */
    protected String _inlineEditorRender;
    
    /**
     * Create a style for paragraphs
     * @param buttonLabel Label of the backoffice button to apply this style
     * @param buttonDescription Description of the backoffice button to apply this style
     * @param buttonSmallIcon Image in 16x16 pixels (can be null) of the backoffice button to apply this style
     * @param buttonMediumIcon Image in 32x32 pixels (can be null) of the backoffice button to apply this style
     * @param buttonLargeIcon Image in 48x48 pixels (can be null) of the backoffice button to apply this style
     * @param buttonCSSClass CSS class of the backoffice button to apply this style (used to give a preview appearance to the button) The CSS class has to be defined in the imported files (See HTMLEditorStyle)
     * @param inlineEditorRender Information necessary in the inline editor to "apply" the style. Can be a "tagname.classname" or just a "classname" depending on the element to style
     */
    public Style (I18nizableText buttonLabel, I18nizableText buttonDescription, String buttonSmallIcon, String buttonMediumIcon, String buttonLargeIcon, String buttonCSSClass, String inlineEditorRender)
    {
        _buttonLabel = buttonLabel;
        _buttonDescription = buttonDescription;
        _buttonCSSClass = buttonCSSClass;
        _buttonSmallIcon = buttonSmallIcon;
        _buttonMediumIcon = buttonMediumIcon;
        _buttonLargeIcon = buttonLargeIcon;
        _inlineEditorRender = inlineEditorRender;
    }
    
    /**
     * Transform the current para style as a map of parameters. Parameters are :
     * <ul>
     * <li>icon-small (optional)</li>
     * <li>icon-medium (optional)</li>
     * <li>icon-large (optional)</li>
     * <li>label</li>
     * <li>description</li>
     * <li>cssclass (optional)</li>
     * <li>inline-editor-render</li>
     * </ul>
     * @return The non null map with parameters.
     */
    public Map<String, Object> getAsParameters()
    {
        Map<String, Object> params = new HashMap<>();
        
        params.put("label", _buttonLabel);
        params.put("description", _buttonDescription);
        if (_buttonCSSClass != null)
        {
            params.put("cls", _buttonCSSClass);
        }
        
        if (_buttonSmallIcon != null)
        {
            params.put("icon-small", _buttonSmallIcon);
        }
        if (_buttonMediumIcon != null)
        {
            params.put("icon-medium", _buttonMediumIcon);
        }
        if (_buttonLargeIcon != null)
        {
            params.put("icon-large", _buttonLargeIcon);
        }
        
        String inlineEditorTagName = null;
        String inlineEditorCSSClass = null;
        
        int index = _inlineEditorRender.indexOf(".");
        if (index == -1)
        {
            inlineEditorTagName = _inlineEditorRender;
        }
        else
        {
            inlineEditorTagName = _inlineEditorRender.substring(0, index);
            inlineEditorCSSClass = _inlineEditorRender.substring(index + 1);
        }
        
        params.put("inlineEditorRender", _inlineEditorRender);
        params.put("tagName", inlineEditorTagName);
        if (inlineEditorCSSClass != null)
        {
            params.put("style", inlineEditorCSSClass);
        }
        
        return params;
    }
    
    /**
     * SAX the current para style
     * @param contentHandler Where to SAX
     * @throws SAXException If an error occured
     */
    public void toSAX(ContentHandler contentHandler) throws SAXException
    {
        XMLUtils.startElement(contentHandler, "style");
        
        XMLUtils.startElement(contentHandler, "button");
        
        _buttonLabel.toSAX(contentHandler, "label");
        _buttonDescription.toSAX(contentHandler, "description");
        
        if (_buttonSmallIcon != null)
        {
            XMLUtils.createElement(contentHandler, "icon-small", _buttonSmallIcon);
        }
        if (_buttonMediumIcon != null)
        {
            XMLUtils.createElement(contentHandler, "icon-medium", _buttonMediumIcon);
        }
        if (_buttonLargeIcon != null)
        {
            XMLUtils.createElement(contentHandler, "icon-large", _buttonLargeIcon);
        }
       
        if (_buttonCSSClass != null)
        {
            XMLUtils.createElement(contentHandler, "cssclass", _buttonCSSClass);
        }
        
        XMLUtils.endElement(contentHandler, "button");
        
        XMLUtils.createElement(contentHandler, "inline-editor", _inlineEditorRender);
        
        XMLUtils.endElement(contentHandler, "style");
    }
    
    /**
     * See constructor
     * @return the buttonCssClass.
     */
    public String getButtonCSSClass()
    {
        return _buttonCSSClass;
    }
    
    /**
     * See constructor
     * @return the buttonDescription
     */
    public I18nizableText getButtonDescription()
    {
        return _buttonDescription;
    }
    
    /**
     * See constructor
     * @return the buttonSmallIcon
     */
    public String getButtonSmallIcon()
    {
        return _buttonSmallIcon;
    }
    
    /**
     * See constructor
     * @return the buttonSmallIcon
     */
    public String getButtonMediumIcon()
    {
        return _buttonMediumIcon;
    }
    
    /**
     * See constructor
     * @return the buttonSmallIcon
     */
    public String getButtonLargeIcon()
    {
        return _buttonLargeIcon;
    }
    
    /**
     * See constructor
     * @return the buttonLabel
     */
    public I18nizableText getButtonLabel()
    {
        return _buttonLabel;
    }
    
    /**
     * See constructor
     * @return the inlineEditorRender
     */
    public String getInlineEditorRender()
    {
        return _inlineEditorRender;
    }
}
