/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.color;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.ui.Callable;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Component listing the available colors
 */
public abstract class AbstractColorsComponent extends AbstractLogEnabled implements Component, Configurable
{
    private Map<String, Map<String, String>> _colors;
    private String _defaultKey;
    private String _colorCSSClass;
    
    public void configure(Configuration configuration) throws ConfigurationException
    {
        Configuration newConfiguration = configuration;
        if (!"colors".equals(configuration.getName()))
        {
            newConfiguration = configuration.getChild("colors");
        }
        
        ColorsDefinition colorsConf = parseColors(newConfiguration);
        
        _defaultKey = colorsConf.defaultKey();
        _colors = colorsConf.colors();
        _colorCSSClass = colorsConf.colorClass();
    }
    
    /**
     * Get the key to use as default in the colors
     * @return A non null key.
     */
    public String getDefaultKey()
    {
        return _defaultKey;
    }
    
    /**
     * Get the colors by index
     * A 'color' is an association key-hexacolor. Where key can be 'main', 'text'... 
     * @return The map
     */
    @Callable
    public Map<String, Map<String, String>> getColors()
    {
        return _colors;
    }
    
    /**
     * Get the color class
     * @return the color class
     */
    public String getColorCSSClassPrefix()
    {
        return _colorCSSClass;
    }
    
    /**
     * Parse colors from configuration
     * @param configuration the configuration
     * @return the colors configuration
     * @throws ConfigurationException if a configuration error occurred
     */
    public ColorsDefinition parseColors(Configuration configuration) throws ConfigurationException
    {
        Map<String, Map<String, String>> colors = new LinkedHashMap<>();
        String defaultKey = configuration.getAttribute("default", "");
        for (Configuration colorConfiguration : configuration.getChildren("color"))
        {
            Map<String, String> color = new HashMap<>();
            
            for (Configuration child : colorConfiguration.getChildren())
            {
                String name = child.getName();
                String value = child.getValue();
                
                color.put(name, value);
            }

            String id = colorConfiguration.getAttribute("id");
            colors.put(id, color);
            
            if (StringUtils.isEmpty(defaultKey))
            {
                defaultKey = id;
            }
        }
        
        String colorClass = configuration.getAttribute("css-class", "color");
        return new ColorsDefinition(defaultKey, colors, colorClass);
    }
    
    /**
     * Class representing a colors definition
     * @param defaultKey the default color key
     * @param colors the colors
     * @param colorClass the color class
     */
    protected record ColorsDefinition(String defaultKey, Map<String, Map<String, String>> colors, String colorClass) { /* empty */ }
}
