/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.color;

import java.io.IOException;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.generation.ServiceableGenerator;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.SAXException;


/**
 * Generates the colors from the component
 */
public class ColorsGenerator extends ServiceableGenerator implements Configurable
{
    /** The colors component */
    protected AbstractColorsComponent _colorsComponent;
    
    private ServiceManager _smanager;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _smanager = smanager;
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        String componentRole = configuration.getChild("role").getValue();
        if (StringUtils.isBlank(componentRole))
        {
            throw new IllegalArgumentException("The child role is mandatory");
        }
        
        try
        {
            _colorsComponent = (AbstractColorsComponent) _smanager.lookup(componentRole);
        }
        catch (ServiceException e)
        {
            throw new IllegalArgumentException("Can't get component with role " + componentRole, e);
        }
    }

    public void generate() throws IOException, SAXException, ProcessingException
    {
        contentHandler.startDocument();
        
        AttributesImpl attrs = new AttributesImpl();
        attrs.addCDATAAttribute("default", _getDefaultKey());
        XMLUtils.startElement(contentHandler, "colors", attrs);

        Map<String, Map<String, String>> colors = _getColors();
        for (String id : colors.keySet())
        {
            AttributesImpl attrs2 = new AttributesImpl();
            attrs2.addCDATAAttribute("id", id);
            XMLUtils.startElement(contentHandler, "color", attrs2);
            
            Map<String, String> map = colors.get(id);
            for (String key : map.keySet())
            {
                XMLUtils.createElement(contentHandler, key, map.get(key));
            }
            
            XMLUtils.endElement(contentHandler, "color");
        }
        
        XMLUtils.endElement(contentHandler, "colors");
        contentHandler.endDocument();
    }
    
    /**
     * Get colors
     * @return the colors
     */
    protected Map<String, Map<String, String>> _getColors()
    {
        return _colorsComponent.getColors();
    }
    
    /**
     * Get default key
     * @return the default key
     */
    protected String _getDefaultKey()
    {
        return _colorsComponent.getDefaultKey();
    }
}
