/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.io.IOException;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.generation.ServiceableGenerator;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.Content.ReferencingContentsSearch;
import org.ametys.cms.repository.DefaultContent;
import org.ametys.cms.tag.Tag;
import org.ametys.cms.tag.TagProviderExtensionPoint;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.version.VersionableAmetysObject;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Generates addition information on content
 */
public class AdditionalContentPropertiesGenerator extends ServiceableGenerator
{
    /** The AmetysObject resolver. */
    protected AmetysObjectResolver _resolver;
    /** the Ametys rights manager */
    protected RightManager _rightManager;
    /** the tag provider extension point */
    protected TagProviderExtensionPoint _tagProviderEP;
    private ContentTypeExtensionPoint _contentTypeEP;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _resolver = (AmetysObjectResolver) serviceManager.lookup(AmetysObjectResolver.ROLE);
        _contentTypeEP = (ContentTypeExtensionPoint) serviceManager.lookup(ContentTypeExtensionPoint.ROLE);
        _tagProviderEP = (TagProviderExtensionPoint) serviceManager.lookup(TagProviderExtensionPoint.ROLE);
        _rightManager = (RightManager) serviceManager.lookup(RightManager.ROLE);
    }
    
    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        String contentId = parameters.getParameter("contentId", request.getParameter("contentId"));
        
        Content content = null;
        if (StringUtils.isNotEmpty(contentId))
        {
            content = _resolver.resolveById(contentId);
        }
        else
        {
            content = (Content) request.getAttribute(Content.class.getName());
        }
        
        contentHandler.startDocument();
        
        AttributesImpl attrs = new AttributesImpl();
        attrs.addCDATAAttribute("id", content.getId());
        
        Boolean live = content.getValue("hasLive");
        if (live)
        {
            attrs.addCDATAAttribute("has-live", "true");
        }
        
        if (content instanceof DefaultContent)
        {
            try
            {
                DefaultContent dc = (DefaultContent) content;
                attrs.addCDATAAttribute("path", dc.getNode().getPath());
            }
            catch (Exception e)
            {
                getLogger().error("Cannot determine JCR path for content " + content.getId(), e);
            }
        }
        XMLUtils.startElement(contentHandler, "content", attrs);
        
        // Take the HEAD revision
        String revision = null;
        if (content instanceof VersionableAmetysObject)
        {
            revision = ((VersionableAmetysObject) content).getRevision();
            ((VersionableAmetysObject) content).switchToRevision(null);
        }
        
        _saxContentTypesHierarchy (content);
        _saxMixinsHierarchy(content);
        _saxReferencingContents(content);
        _saxTags(content);
        _saxOtherProperties(content);
        _saxRepositoryRights();
        
        if (content instanceof VersionableAmetysObject && revision != null)
        {
            ((VersionableAmetysObject) content).switchToRevision(revision);
        }
        
        XMLUtils.endElement(contentHandler, "content");
        contentHandler.endDocument();
    }
    
    private void _saxRepositoryRights() throws SAXException
    {
        XMLUtils.startElement(contentHandler, "repository");
        XMLUtils.startElement(contentHandler, "rights");
        XMLUtils.createElement(contentHandler, "access", Boolean.toString(_rightManager.currentUserHasRight("REPOSITORY_Rights_Access", "/repository") == RightResult.RIGHT_ALLOW));
        XMLUtils.endElement(contentHandler, "rights");
        XMLUtils.endElement(contentHandler, "repository");
        
    }

    /**
     * SAX other additional properties
     * @param content the content 
     * @throws SAXException if an error occurred while saxing
     */
    protected void _saxOtherProperties (Content content) throws SAXException
    {
        // Nothing to do
    }
    
    /**
     * SAX content's tags
     * @param content The content
     * @throws SAXException if an error occurred
     */
    protected void _saxTags (Content content) throws SAXException
    {
        Set<String> tags = content.getTags();
        
        XMLUtils.startElement(contentHandler, "tags");
        
        for (String tagName : tags)
        {
            Tag tag = _tagProviderEP.getTag(tagName, getContextualParameters(content));
            if (tag != null)
            {
                AttributesImpl attrs = new AttributesImpl();
                attrs.addCDATAAttribute("name", tag.getName());
                XMLUtils.startElement(contentHandler, "tag", attrs);
                tag.getTitle().toSAX(contentHandler);
                XMLUtils.endElement(contentHandler, "tag");
            }
        }
        
        XMLUtils.endElement(contentHandler, "tags");
    }
    
    /**
     * Get the contextual parameters for the given object
     * @param object the object
     * @return the object's contextual parameters
     */
    protected Map<String, Object> getContextualParameters (AmetysObject object)
    {
        return new HashMap<>();
    }
    
    /**
     * Sax the referencing contents
     * @param content The content
     * @throws SAXException if an error occurred while saxing
     */
    protected void _saxReferencingContents (Content content) throws SAXException
    {
        ReferencingContentsSearch referencingContents = content.searchReferencingContents(100);
        
        AttributesImpl referersAttrs = new AttributesImpl();
        referersAttrs.addCDATAAttribute("count", String.valueOf(referencingContents.total()));
        referersAttrs.addCDATAAttribute("hasOther", String.valueOf(referencingContents.hasOtherReferences()));
        XMLUtils.startElement(contentHandler, "referencing-contents", referersAttrs);
        
        Locale locale = content.getLanguage() != null ? new Locale(content.getLanguage()) : null;
        for (Content referrerContent : referencingContents.referencingContents())
        {
            AttributesImpl refererAttrs = new AttributesImpl();
            refererAttrs.addCDATAAttribute("id", referrerContent.getId());
            refererAttrs.addCDATAAttribute("name", referrerContent.getName());
            refererAttrs.addCDATAAttribute("title", referrerContent.getTitle(locale));
            
            XMLUtils.createElement(contentHandler, "referencing-content", refererAttrs);
        }
        
        XMLUtils.endElement(contentHandler, "referencing-contents");
    }
    
    /**
     * SAX the content types hierarchy
     * @param content The content type
     * @throws SAXException if an error occurred while saxing
     */
    protected void _saxContentTypesHierarchy (Content content) throws SAXException
    {
        // Sort content types
        TreeSet<ContentType> treeSet = new TreeSet<>(new ContentTypeComparator());
        for (String id : content.getTypes())
        {
            ContentType contentType = _contentTypeEP.getExtension(id);
            treeSet.add(contentType);
        }
        
        XMLUtils.startElement(contentHandler, "content-types");
        for (ContentType cType : treeSet)
        {
            _saxContentType (cType);
        }
        XMLUtils.endElement(contentHandler, "content-types");
    }
    
    /**
     * SAX the content types hierarchy
     * @param content The content type
     * @throws SAXException if an error occurred while saxing
     */
    protected void _saxMixinsHierarchy (Content content) throws SAXException
    {
        XMLUtils.startElement(contentHandler, "mixins");
        
        String[] contentTypes = content.getMixinTypes();
        for (String id : contentTypes)
        {
            ContentType cType = _contentTypeEP.getExtension(id);
            _saxMixin(cType);
        }
        
        XMLUtils.endElement(contentHandler, "mixins");
    }
    
    /**
     * SAX the content type hierarchy
     * @param cType The content type
     * @throws SAXException if an error occurred while saxing
     */
    protected void _saxContentType (ContentType cType) throws SAXException
    {
        AttributesImpl attrs = new AttributesImpl();
        
        attrs.addCDATAAttribute("id", cType.getId());
        
        if (cType.getSmallIcon() != null)
        {
            attrs.addCDATAAttribute("icon", cType.getSmallIcon());
        }
        if (cType.getIconGlyph() != null)
        {
            attrs.addCDATAAttribute("icon-glyph", cType.getIconGlyph());
        }
        if (cType.getIconDecorator() != null)
        {
            attrs.addCDATAAttribute("icon-decorator", cType.getIconDecorator());
        }
        
        XMLUtils.startElement(contentHandler, "content-type", attrs);
        cType.getLabel().toSAX(contentHandler, "label");
        
        for (String id : cType.getSupertypeIds())
        {
            ContentType superType = _contentTypeEP.getExtension(id);
            _saxContentType(superType);
        }
        XMLUtils.endElement(contentHandler, "content-type");
    }
    
    /**
     * SAX the mixin hierarchy
     * @param mixin The mixin type
     * @throws SAXException if an error occurred while saxing
     */
    protected void _saxMixin (ContentType mixin) throws SAXException
    {
        AttributesImpl attrs = new AttributesImpl();
        
        attrs.addCDATAAttribute("id", mixin.getId());
        if (mixin.getSmallIcon() != null)
        {
            attrs.addCDATAAttribute("icon", mixin.getSmallIcon());
        }
        if (mixin.getIconGlyph() != null)
        {
            attrs.addCDATAAttribute("icon-glyph", mixin.getIconGlyph());
        }
        if (mixin.getIconDecorator() != null)
        {
            attrs.addCDATAAttribute("icon-decorator", mixin.getIconDecorator());
        }
        
        XMLUtils.startElement(contentHandler, "mixin", attrs);
        mixin.getLabel().toSAX(contentHandler, "label");
        for (String id : mixin.getSupertypeIds())
        {
            ContentType superType = _contentTypeEP.getExtension(id);
            _saxMixin(superType);
        }
        XMLUtils.endElement(contentHandler, "mixin");
    }

    class ContentTypeComparator implements Comparator<ContentType>
    {
        @Override
        public int compare(ContentType c1, ContentType c2)
        {
            I18nizableText t1 = c1.getLabel();
            I18nizableText t2 = c2.getLabel();
            
            String str1 = t1.isI18n() ? t1.getKey() : t1.getLabel();
            String str2 = t2.isI18n() ? t2.getKey() : t2.getLabel();
            
            int compareTo = str1.toString().compareTo(str2.toString());
            if (compareTo == 0)
            {
                // Content types have same keys but there are not equals, so do not return 0 to add it in TreeSet
                // Indeed, in a TreeSet implementation two elements that are equal by the method compareTo are, from the standpoint of the set, equal 
                return 1;
            }
            return compareTo;
        }
    }
}
