/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Map;
import java.util.stream.Stream;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.avalon.framework.thread.ThreadSafe;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceFactory;

import org.ametys.cms.data.Binary;
import org.ametys.cms.data.BinarySource;
import org.ametys.cms.repository.Content;
import org.ametys.core.util.FilenameUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.ametysobject.ModelAwareDataAwareAmetysObject;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * {@link SourceFactory} for content attributes.
 */
public class AttributeSourceFactory extends AbstractLogEnabled implements SourceFactory, ThreadSafe, Serviceable, Contextualizable
{
    /** source scheme */
    public static final String ATTRIBUTE_SCHEME = "attribute";
    
    /** The context */
    protected Context _context;
    
    /** The resolver */
    protected AmetysObjectResolver _resolver;
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    public Source getSource(String location, Map parameters) throws IOException, MalformedURLException
    {
        String uri = _getURI(location);
        
        // uri are like <contentPath>@<attributePath>;<filename>
        int i = uri.indexOf('@');
        int j = uri.indexOf(';', i);
        String path = uri.substring(0, i);
        String attribute = uri.substring(i + 1, j);
        String filename = uri.substring(j + 1);
        
        ModelAwareDataAwareAmetysObject ametysObject = _getModelAwareDataAwareAmetysObject();
        
        try
        {
            if (ametysObject == null) 
            {
                ametysObject = _resolver.resolveByPath(FilenameUtils.decode(path));
            }
            
            Binary binary = _getBinary(ametysObject, attribute, filename);
    
            return new BinarySource(binary, FilenameUtils.decode(filename), location, ATTRIBUTE_SCHEME);
        }
        catch (Exception e)
        {
            getLogger().error("Unable to resolve binary attribute for uri " + location, e);
            
            // returns an unexisting Source
            return new BinarySource(null, FilenameUtils.decode(filename), location, ATTRIBUTE_SCHEME);
        }
    }

    /**
     * Get binary
     * @param object the ametys object
     * @param attribute the attribut
     * @param filename the filename
     * @return the binary
     */
    protected Binary _getBinary(ModelAwareDataAwareAmetysObject object, String attribute, String filename)
    {
        if (object.isMultiple(attribute))
        {
            // if multiple binary, take the first file with name equals to the filename
            Binary[] binaries = object.getValue(attribute, false, new Binary[0]);
            return Stream.of(binaries)
                .filter(b -> b.getFilename().equals(filename))
                .findFirst()
                .orElse(null);
        }
        
        return object.getValue(attribute);
    }
    
    /**
     * Get URI
     * @param location the location
     * @return the uri
     */
    protected String _getURI(String location)
    {
        return location.substring((ATTRIBUTE_SCHEME + "://").length());
    }
    
    /**
     * Get model aware data aware ametys object
     * @return model aware data aware ametys object
     */
    protected ModelAwareDataAwareAmetysObject _getModelAwareDataAwareAmetysObject()
    {
        Request request = ContextHelper.getRequest(_context);
        return (Content) request.getAttribute(Content.class.getName());
    }
    
    public void release(Source source)
    {
        // nothing to do
    }
}
