/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.content;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.generation.ServiceableGenerator;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.cms.content.references.OutgoingReferences;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.transformation.ConsistencyChecker;
import org.ametys.cms.transformation.ConsistencyChecker.CHECK;
import org.ametys.cms.transformation.ConsistencyChecker.CheckReport;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.plugins.repository.data.type.ModelItemTypeConstants;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.exception.UndefinedItemPathException;

/**
 * Generates the consistency report for the given content
 */
public class ConsistencyGenerator extends ServiceableGenerator
{
    /** Repository content */
    protected AmetysObjectResolver _resolver;
    /** The consistency checker */
    protected ConsistencyChecker _consistencyChecker;
    /** The content types helper */
    protected ContentTypesHelper _contentTypesHelper;
    /** The content helper */
    protected ContentHelper _contentHelper;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _consistencyChecker = (ConsistencyChecker) smanager.lookup(ConsistencyChecker.ROLE);
        _contentTypesHelper = (ContentTypesHelper) smanager.lookup(ContentTypesHelper.ROLE);
        _contentHelper = (ContentHelper) smanager.lookup(ContentHelper.ROLE);
    }

    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        contentHandler.startDocument();
        XMLUtils.startElement(contentHandler, "contents");
        
        @SuppressWarnings("unchecked")
        Map<String, Object> jsParameters = (Map<String, Object>) objectModel.get(ObjectModelHelper.PARENT_CONTEXT);
        @SuppressWarnings("unchecked")
        List<String> contentsId = (List<String>) jsParameters.get("contentsId");
        for (String contentId : contentsId)
        {
            try
            {
                Content content = _resolver.resolveById(contentId);
                
                AttributesImpl attrs = new AttributesImpl();
                attrs.addCDATAAttribute("id", content.getId());
                attrs.addCDATAAttribute("title", _contentHelper.getTitle(content));
                attrs.addCDATAAttribute("name", content.getName());
                attrs.addCDATAAttribute("path", content.getPath());
                attrs.addCDATAAttribute("type", StringUtils.join(content.getTypes(), ','));
                if (content.getLanguage() != null)
                {
                    attrs.addCDATAAttribute("lang", content.getLanguage());
                }

                XMLUtils.startElement(contentHandler, "content", attrs);
                
                List<ModelItem> modelItemsToSax = new ArrayList<>();
                Map<String, OutgoingReferences> referencesByPath = content.getOutgoingReferences();
                
                for (String dataPath : referencesByPath.keySet())
                {
                    try
                    {
                        List<ModelItem> modelItemPath = _contentTypesHelper.getModelItemPath(dataPath, content);
                        
                        // Add these model items to the set of model items to SAX.
                        modelItemsToSax.addAll(modelItemPath);
                        
                        saxReferencesConsistency(content, referencesByPath.get(dataPath), dataPath);
                    }
                    catch (UndefinedItemPathException e)
                    {
                        getLogger().warn("Trying to check consistency for a item path that doesn't exist in the model. Reference will be ignored", e);
                    }
                }
                
                // SAX model items used for this content
                for (ModelItem modelItem : modelItemsToSax)
                {
                    attrs.clear();
                    attrs.addCDATAAttribute("path", modelItem.getPath());
                    attrs.addCDATAAttribute("is-repeater", Boolean.toString(ModelItemTypeConstants.REPEATER_TYPE_ID.equals(modelItem.getType().getId())));
                    XMLUtils.startElement(contentHandler, "metadata-definition", attrs);
                    modelItem.getLabel().toSAX(contentHandler);
                    XMLUtils.endElement(contentHandler, "metadata-definition");
                }
                
                XMLUtils.endElement(contentHandler, "content");
            }
            catch (UnknownAmetysObjectException e)
            {
                getLogger().warn("Can not check consistency of non existing content '" + contentId + "'");
            }
        }
        
        XMLUtils.endElement(contentHandler, "contents");
        contentHandler.endDocument();
    }

    /**
     * Check consistency for each outgoing references and generate SAX event for the result.
     * @param content the content containing the references
     * @param references the outgoing references
     * @param dataPath the path of the data containing the outgoing references
     * @throws SAXException if an error occurred
     */
    protected void saxReferencesConsistency(Content content, OutgoingReferences references, String dataPath) throws SAXException
    {
        // SAX'ing consistency info
        String definitionPath = ModelHelper.getDefinitionPathFromDataPath(dataPath);
        for (String referenceType : references.keySet())
        {
            for (String referenceValue : references.get(referenceType))
            {
                AttributesImpl attrs = new AttributesImpl();
                attrs.addCDATAAttribute("type", referenceType);
                attrs.addCDATAAttribute("element", referenceValue);
                attrs.addCDATAAttribute("path", definitionPath);
                
                CheckReport report;
                try
                {
                    report = _consistencyChecker.checkConsistency(referenceType, referenceValue, content.getId(), dataPath, false);
                }
                catch (Exception e)
                {
                    // Consider it a failure.
                    report = new CheckReport(CHECK.SERVER_ERROR, Optional.ofNullable(e.getMessage()));
                }
                
                String elementName;
                switch (report.status())
                {
                    case SUCCESS:
                        elementName = "success";
                        break;
                    case UNKNOWN:
                        elementName = "unknown";
                        break;
                    case UNAUTHORIZED:
                        elementName = "unauthorized";
                        break;
                    case NOT_FOUND:
                        elementName = "not-found";
                        break;
                    case SERVER_ERROR:
                    default:
                        elementName = "server-error";
                        break;
                }
                XMLUtils.startElement(contentHandler, elementName, attrs);
                I18nizableText label = _consistencyChecker.getLabel(referenceType, referenceValue, content.getId(), dataPath);
                label.toSAX(contentHandler, "label");
                if (report.message().isPresent())
                {
                    XMLUtils.createElement(contentHandler, "message", report.message().get());
                }
                XMLUtils.endElement(contentHandler, elementName);
            }
        }
    }
}
