/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Collection;
import java.util.Date;
import java.util.Map;
import java.util.function.BiFunction;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.stream.IntStream;

import javax.xml.transform.TransformerException;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;
import org.apache.xpath.XPathAPI;
import org.apache.xpath.objects.XObject;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.repository.ReactionableObject;
import org.ametys.cms.repository.ReportableObject;
import org.ametys.cms.repository.ReactionableObject.ReactionType;
import org.ametys.cms.repository.comment.Comment;
import org.ametys.cms.repository.comment.CommentableContent;
import org.ametys.cms.repository.comment.contributor.ContributorCommentableContent;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.core.util.LambdaUtils.ThrowingFunction;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.plugins.repository.data.extractor.xml.ModelAwareXMLValuesExtractor;
import org.ametys.plugins.repository.data.extractor.xml.XMLValuesExtractorAdditionalDataGetter;
import org.ametys.plugins.repository.dublincore.ModifiableDublinCoreAwareAmetysObject;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewHelper;

/**
 * Component responsible to extract a {@link Content} from an XML document.
 */
public class ContentExtractor implements Component, Serviceable
{
    /** Avalon role. */
    public static final String CMS_CONTENT_EXTACTOR_ROLE = ContentExtractor.class.getName();
    
    private UserHelper _userHelper;

    public void service(ServiceManager manager) throws ServiceException
    {
        _userHelper = (UserHelper) manager.lookup(UserHelper.ROLE);
    }
    
    /**
     * Fills the given content with the values from the provided {@link org.w3c.dom.Node}.
     * <br>This is the anti-operation of {@link ContentSaxer#saxContent}, as the org.w3c.dom.Node should be a Node previously generated with SAX events from this method.
     * @param content The content to fill
     * @param node The node to read for retrieving values to fill
     * @param additionalDataGetter The object that will retrieve potential additional data for the content's attributes
     * @throws Exception if an exception occurs
     */
    public void fillContent(ModifiableContent content, org.w3c.dom.Node node, XMLValuesExtractorAdditionalDataGetter additionalDataGetter) throws Exception
    {
        org.w3c.dom.Node contentNode = XPathAPI.selectSingleNode(node, "content");
        
        fillAttributes(content, contentNode, additionalDataGetter);
        fillDublinCore(content, contentNode);
        
        if (content instanceof CommentableContent commentableContent)
        {
            fillContentComments(contentNode, "comments", commentableContent::createComment);
        }
        
        if (content instanceof ContributorCommentableContent contributorCommentableContent)
        {
            fillContentComments(contentNode, "contributor-comments", contributorCommentableContent::createContributorComment);
        }
        
        if (content instanceof ReactionableObject reactionableObject)
        {
            fillReactions(reactionableObject, contentNode);
        }
        
        if (content instanceof ReportableObject reportableObject)
        {
            fillReports(reportableObject, contentNode);
        }
    }
    
    /**
     * Fills the given object with the dublin core values from the provided {@link org.w3c.dom.Node}.
     * @param dcObject The object to fill
     * @param node The node to read to get the values to fill
     * @throws Exception if an exception occurs
     */
    protected void fillDublinCore(ModifiableDublinCoreAwareAmetysObject dcObject, org.w3c.dom.Node node) throws Exception
    {
        org.w3c.dom.Node dcNode = XPathAPI.selectSingleNode(node, "dublin-core-metadata");
        if (dcNode != null)
        {
            setIfNotNull(dcNode, "title", XObject::str, dcObject::setDCTitle);
            setIfNotNull(dcNode, "creator", XObject::str, dcObject::setDCCreator);
            setIfNotNull(dcNode, "subject", this::values, dcObject::setDCSubject);
            setIfNotNull(dcNode, "description", XObject::str, dcObject::setDCDescription);
            setIfNotNull(dcNode, "publisher", XObject::str, dcObject::setDCPublisher);
            setIfNotNull(dcNode, "contributor", XObject::str, dcObject::setDCContributor);
            setIfNotNull(dcNode, "date", this::dateValue, dcObject::setDCDate);
            setIfNotNull(dcNode, "type", XObject::str, dcObject::setDCType);
            setIfNotNull(dcNode, "format", XObject::str, dcObject::setDCFormat);
            setIfNotNull(dcNode, "identifier", XObject::str, dcObject::setDCIdentifier);
            setIfNotNull(dcNode, "source", XObject::str, dcObject::setDCSource);
            setIfNotNull(dcNode, "language", XObject::str, dcObject::setDCLanguage);
            setIfNotNull(dcNode, "relation", XObject::str, dcObject::setDCRelation);
            setIfNotNull(dcNode, "coverage", XObject::str, dcObject::setDCCoverage);
            setIfNotNull(dcNode, "rights", XObject::str, dcObject::setDCRights);
        }
    }
    
    /**
     * Fills the content with the comments from the provided {@link org.w3c.dom.Node}
     * @param contentNode the node to read to get the comments' values
     * @param commentsNodesName the name of comments nodes
     * @param createCommentFunction the function to use to create comments on the content
     * @throws Exception if an error occurs
     */
    protected void fillContentComments(org.w3c.dom.Node contentNode, String commentsNodesName, BiFunction<String, ZonedDateTime, Comment> createCommentFunction) throws Exception
    {
        NodeList commnentsNodes = XPathAPI.selectNodeList(contentNode, commentsNodesName + "/comment");
        for (int i = 0; i < commnentsNodes.getLength(); i++)
        {
            org.w3c.dom.Node commentNode = commnentsNodes.item(i);
            
            String commentId = XPathAPI.eval(commentNode, "@id").str();
            String creationDateAsString = XPathAPI.eval(commentNode, "@creation-date").str();
            ZonedDateTime creationDate = DateUtils.parseZonedDateTime(creationDateAsString);
            
            Comment comment = createCommentFunction.apply(commentId, creationDate);
            fillComment(comment, commentNode);
        }
    }
    
    /**
     * Fills the given comment with the values from the provided {@link org.w3c.dom.Node}
     * @param comment The comment to fill
     * @param commentNode the node to read to get the comment's values
     * @throws Exception if an error occurs
     */
    protected void fillComment(Comment comment, org.w3c.dom.Node commentNode) throws Exception
    {
        setIfNotNull(commentNode, "@is-validated", this::booleanValue, comment::setValidated);
        setIfNotNull(commentNode, "@is-email-hidden", this::booleanValue, comment::setEmailHiddenStatus);
        setIfNotNull(commentNode, "@author-name", XObject::str, comment::setAuthorName);
        setIfNotNull(commentNode, "@author-email", XObject::str, comment::setAuthorEmail);
        setIfNotNull(commentNode, "@author-url", XObject::str, comment::setAuthorURL);
        
        org.w3c.dom.Node authorNode = XPathAPI.selectSingleNode(commentNode, "author");
        if (authorNode != null)
        {
            UserIdentity author = _userHelper.xml2userIdentity(authorNode);
            comment.setAuthor(author);
        }
        
        StringBuilder content = new StringBuilder();
        NodeList paragraphs = XPathAPI.selectNodeList(commentNode, "p");
        for (int i = 0; i < paragraphs.getLength(); i++)
        {
            org.w3c.dom.Node paragraph = paragraphs.item(i);
            content.append(paragraph.getTextContent())
                .append("\r\n");
        }
        comment.setContent(content.toString());
        
        fillReactions(comment, commentNode);
        
        fillReports(comment, commentNode);
        
        NodeList subCommnentsNodes = XPathAPI.selectNodeList(commentNode, "sub-comments/comment");
        for (int i = 0; i < subCommnentsNodes.getLength(); i++)
        {
            org.w3c.dom.Node subCommentNode = subCommnentsNodes.item(i);
            
            String subCommentId = XPathAPI.eval(subCommentNode, "@id").str();
            String creationDateAsString = XPathAPI.eval(subCommentNode, "@creation-date").str();
            ZonedDateTime creationDate = DateUtils.parseZonedDateTime(creationDateAsString);
            
            Comment subComment = comment.createSubComment(subCommentId, creationDate);
            fillComment(subComment, subCommentNode);
        }
    }
    
    /**
     * Fills the given {@link ReactionableObject} with the reactions from the provided {@link org.w3c.dom.Node}
     * @param reactionable The {@link ReactionableObject} to fill
     * @param node the node to read to get the reactions
     * @throws Exception if an error occurs
     */
    protected void fillReactions(ReactionableObject reactionable, org.w3c.dom.Node node) throws Exception
    {
        NodeList reactions = XPathAPI.selectNodeList(node, "reactions/reaction");
        for (int i = 0; i < reactions.getLength(); i++)
        {
            org.w3c.dom.Node reactionNode = reactions.item(i);
            XObject reactionTypeAttr = XPathAPI.eval(reactionNode, "@type");
            ReactionType reactionType = ReactionType.valueOf(reactionTypeAttr.str());
            NodeList actors = XPathAPI.selectNodeList(reactionNode, "actor");
            for (int j = 0; j < actors.getLength(); j++)
            {
                org.w3c.dom.Node actorNode = actors.item(j);
                UserIdentity actor = _userHelper.xml2userIdentity(actorNode);
                reactionable.addReaction(actor, reactionType);
            }
        }
    }
    
    /**
     * Fills the given {@link ReportableObject} with the reports from the provided {@link org.w3c.dom.Node}
     * @param reportable The {@link ReportableObject} to fill
     * @param node the node to read to get the reports
     * @throws Exception if an error occurs
     */
    protected void fillReports(ReportableObject reportable, org.w3c.dom.Node node) throws Exception
    {
        org.w3c.dom.Node reportsNode = XPathAPI.selectSingleNode(node, "reports");
        if (reportsNode != null)
        {
            XObject reportsCountAttr = XPathAPI.eval(reportsNode, "@count");
            long reportsCount = (long) reportsCountAttr.num();
            
            if (reportsCount > 0)
            {
                reportable.setReportsCount(reportsCount);
            }
        }
    }
    
    /**
     * Fills the given content with the attributes from the provided {@link org.w3c.dom.Node}
     * @param content The content to fill
     * @param contentNode the node to read to get the attributes
     * @param additionalDataGetter The object that will retrieve potential additional data for the content's attributes
     * @throws Exception if an error occurs
     */
    protected void fillAttributes(ModifiableContent content, org.w3c.dom.Node contentNode, XMLValuesExtractorAdditionalDataGetter additionalDataGetter) throws Exception
    {
        Element attributesElement = (Element) XPathAPI.selectSingleNode(contentNode, "attributes");
        @SuppressWarnings("unchecked")
        Collection<Model> contentModels = (Collection<Model>) content.getModel();
        View view = View.of(contentModels);
        View truncatedView = ViewHelper.getTruncatedView(view);
        Map<String, Object> values = new ModelAwareXMLValuesExtractor(attributesElement, additionalDataGetter, contentModels)
                .extractValues(truncatedView);
        
        content.synchronizeValues(truncatedView, values);
    }

    /**
     * Sets a value through the given setter if the value is not null
     * @param <T> The type of the value to set
     * @param node The node to read to get the value
     * @param expression The expression to apply on the node to get the value
     * @param retriever The {@link Function} that will retrieve the typed value from the {@link XObject} evaluated from the expression
     * @param setter The {@link Consumer} that will be used to set the value 
     * @throws Exception if an error occurs
     */
    protected <T> void setIfNotNull(org.w3c.dom.Node node, String expression, ThrowingFunction<XObject, T> retriever, Consumer<T> setter) throws Exception
    {
        XObject xObject = XPathAPI.eval(node, expression);
        if (xObject.getType() != XObject.CLASS_NULL)
        {
            T value = retriever.apply(xObject);
            if (value != null)
            {
                setter.accept(value);
            }
        }
    }
    
    /**
     * Consumes an {@link XObject} to retrieve the value as a {@link Date}
     * @param xObject The consumed {@link XObject}
     * @return The {@link Date} value
     */
    protected Date dateValue(XObject xObject)
    {
        String dateAsString = xObject.str();
        if (StringUtils.isBlank(dateAsString))
        {
            return null;
        }
        LocalDate localDate = DateTimeFormatter.ISO_LOCAL_DATE.parse(dateAsString, LocalDate::from);
        return DateUtils.asDate(localDate);
    }
    
    /**
     * Consumes an {@link XObject} to retrieve the value as a {@link Boolean}
     * @param xObject The consumed {@link XObject}
     * @return The {@link Boolean} value
     */
    protected Boolean booleanValue(XObject xObject)
    {
        String booleanAsString = xObject.str();
        return Boolean.valueOf(booleanAsString);
    }
    
    /**
     * Consumes an {@link XObject} to retrieve the value as a String array
     * @param xObject The consumed {@link XObject}
     * @return The String array
     * @throws TransformerException if an error occurs
     */
    protected String[] values(XObject xObject) throws TransformerException
    {
        NodeList nodeList = xObject.nodelist();
        return IntStream.range(0, nodeList.getLength())
                .mapToObj(nodeList::item)
                .map(org.w3c.dom.Node::getTextContent)
                .toArray(String[]::new);
    }
}
