/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.io.IOException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Locale;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.generation.Generator;
import org.apache.cocoon.generation.ServiceableGenerator;
import org.apache.cocoon.i18n.I18nUtils;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewHelper;

/**
 * {@link Generator} for rendering raw content data.
 */
public class ContentGenerator extends ServiceableGenerator
{
    /** The display date format. */
    protected static final DateFormat _DC_DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd");
    
    /** Content type extension point. */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    /** Helper for content types */
    protected ContentTypesHelper _cTypesHelper; 
    /** The content saxer */
    protected ContentSaxer _contentSaxer;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) serviceManager.lookup(ContentTypeExtensionPoint.ROLE);
        _cTypesHelper = (ContentTypesHelper) serviceManager.lookup(ContentTypesHelper.ROLE);
        _contentSaxer = (ContentSaxer) serviceManager.lookup(ContentSaxer.CMS_CONTENT_SAXER_ROLE);
    }
    
    public void generate() throws IOException, SAXException, ProcessingException
    {
        contentHandler.startDocument();
        _generateContent();
        contentHandler.endDocument();
    }
    
    /**
     * Generate the content (with the start/end document)
     * @throws SAXException if an error occurs while SAXing
     * @throws IOException if an error occurs
     * @throws ProcessingException if an error occurs
     */
    protected void _generateContent() throws SAXException, IOException, ProcessingException
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        Content content = (Content) request.getAttribute(Content.class.getName());
        
        // SAX the content
        _saxContent(content, getDefaultLocale(request));
    }
    
    /**
     * Get the default locale to use to sax localized values
     * @param request the request
     * @return the default locale
     */
    protected Locale getDefaultLocale(Request request)
    {
        String lang = parameters.getParameter("lang", request.getParameter("lang"));
        return StringUtils.isNotEmpty(lang) ? new Locale(lang) : I18nUtils.findLocale(objectModel, "locale", null, Locale.getDefault(), true);
    }
    
    /**
     * SAX the content 
     * @param content The content to SAX
     * @param defaultLocale The default locale to use to sax localized values if the content's language is null.
     * @throws SAXException if an error occurs while SAXing
     * @throws IOException if an error occurs
     * @throws ProcessingException if an error occurs
     */
    protected void _saxContent (Content content, Locale defaultLocale) throws SAXException, IOException, ProcessingException
    {
        _contentSaxer.saxRootTag(content, contentHandler, defaultLocale, "content");
        
        boolean isEdition = parameters.getParameterAsBoolean("isEdition", false);

        View view = _getView(content, isEdition);
        Locale locale = content.getLanguage() != null ? new Locale(content.getLanguage()) : defaultLocale;
        
        // FIXME CMS-3057
        Request request = ObjectModelHelper.getRequest(objectModel);
        boolean displayWorkflow = !"true".equals(request.getParameter("ignore-workflow"));
        
        _contentSaxer.saxContent(content, contentHandler, locale, view, null, displayWorkflow, displayWorkflow, true, "metadata", isEdition);
        
        if (isEdition)
        {
            XMLUtils.startElement(contentHandler, "comments");
            _saxAttributesComments(content, view);
            XMLUtils.endElement(contentHandler, "comments");
        }
        
        String[] cTypes = (String[]) ArrayUtils.addAll(content.getTypes(), content.getMixinTypes());
        for (String cTypeId : cTypes)
        {
            ContentType cType = _contentTypeExtensionPoint.getExtension(cTypeId);
            cType.saxContentTypeAdditionalData(contentHandler, content);
        }
        
        _saxOtherData(content, defaultLocale);
        
        XMLUtils.endElement(contentHandler, "content");
    }
    
    /**
     * SAX content attributes comments.
     * @param content the content.
     * @param view the view containing the attributes.
     * @throws SAXException if an error occurs while SAXing.
     */
    protected void _saxAttributesComments(Content content, View view) throws SAXException
    {
        content.commentsToSAX(contentHandler, view);
    }
    
    /**
     * SAX any other data needed by the view.<p>
     * Default implementation does nothing.
     * @param content the content.
     * @throws SAXException if an error occurs while SAXing.
     * @throws ProcessingException if an error occurs.
     * @deprecated Use and/or override {@link #_saxOtherData(Content, Locale)} instead
     */
    @Deprecated
    protected void _saxOtherData(Content content) throws SAXException, ProcessingException
    {
        // No other data to SAX
    }
    
    /**
     * SAX any other data needed by the view.<p>
     * Default implementation does nothing.
     * @param content the content.
     * @param defaultLocale The default locale
     * @throws SAXException if an error occurs while SAXing.
     * @throws ProcessingException if an error occurs.
     * @throws IOException if an error occurs.
     */
    protected void _saxOtherData(Content content, Locale defaultLocale) throws SAXException, ProcessingException, IOException
    {
        // For legacy purposes
        _saxOtherData(content);
    }
    
    /**
     * Retrieves the view to be used when SAX'ing attributes and attribute comments.
     * @param content The content to consider. Cannot be null.
     * @param isEdition <code>true</code> if the view is to use for edition, <code>false</code> otherwise
     * @return The retrieved view
     * @throws ProcessingException If the view could not be retrieved
     */
    protected View _getView(Content content, boolean isEdition) throws ProcessingException
    {
        String fallbackViewName = parameters.getParameter("fallbackViewName", StringUtils.EMPTY);
        String viewName = parameters.getParameter("viewName", StringUtils.EMPTY);
        
        View view = _cTypesHelper.getViewWithFallback(viewName, fallbackViewName, content);
        if (view == null)
        {
            String errorMsg = String.format("Unknown view '%s' nor fallback view '%s' for content '%s'", viewName, fallbackViewName, content.getId());
            getLogger().error(errorMsg);
            throw new IllegalArgumentException(errorMsg);
        }
        
        return isEdition ? ViewHelper.getTruncatedView(view) : ViewHelper.mergeDuplicatedItems(view); 
    }
}
