/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.generation.ServiceableGenerator;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.cms.lock.LockContentManager;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.cms.workflow.ContentWorkflowHelper;
import org.ametys.core.right.RightManager;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.plugins.repository.lock.LockAwareAmetysObject;
import org.ametys.plugins.workflow.support.WorkflowProvider;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;

import com.opensymphony.workflow.spi.Step;

/**
 * SAX the contents information
 */
public class ContentInformationGenerator extends ServiceableGenerator
{
    /** Ametys object resovler. */
    protected AmetysObjectResolver _resolver;
    /** The rights manager */
    protected RightManager _rightManager;
    /** The user provider */
    protected CurrentUserProvider _userProvider;
    /** The workflow provider */
    protected WorkflowProvider _workflowProvider;
    /** The lock manager. */
    protected LockContentManager _lockManager;
    /** The content workflow helper */
    protected ContentWorkflowHelper _contentWorkflowHelper;
    /** The content helper */
    protected ContentHelper _contentHelper;

    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _resolver = (AmetysObjectResolver) serviceManager.lookup(AmetysObjectResolver.ROLE);
        _userProvider = (CurrentUserProvider) serviceManager.lookup(CurrentUserProvider.ROLE);
        _rightManager = (RightManager) serviceManager.lookup(RightManager.ROLE);
        _workflowProvider = (WorkflowProvider) serviceManager.lookup(WorkflowProvider.ROLE);
        _lockManager = (LockContentManager) serviceManager.lookup(LockContentManager.ROLE);
        _contentWorkflowHelper = (ContentWorkflowHelper) serviceManager.lookup(ContentWorkflowHelper.ROLE);
        _contentHelper = (ContentHelper) serviceManager.lookup(ContentHelper.ROLE);
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        Map<String, Object> jsParameters = (Map<String, Object>) objectModel.get(ObjectModelHelper.PARENT_CONTEXT);
        List<String> ids = null;
        Object idParam = jsParameters.get("id");
        if (idParam instanceof String)
        {
            ids = new ArrayList<>();
            ids.add((String) idParam);
        }
        else
        {
            ids = (List<String>) idParam;
        }
        
        contentHandler.startDocument();
        XMLUtils.startElement(contentHandler, "contents");
        
        for (String contentID : ids)
        {
            AttributesImpl attrs = new AttributesImpl();
            
            try
            {
                Content content = _resolver.resolveById(contentID);
                
                attrs.addCDATAAttribute("id", content.getId());
                attrs.addCDATAAttribute("title", _contentHelper.getTitle(content));
                attrs.addCDATAAttribute("name", content.getName());
                attrs.addCDATAAttribute("path", content.getPath());
                attrs.addCDATAAttribute("type", StringUtils.join(content.getTypes(), ','));
                if (content.getLanguage() != null)
                {
                    attrs.addCDATAAttribute("lang", content.getLanguage());
                }
                attrs.addCDATAAttribute("creator", content.getCreator().getLogin());
                attrs.addCDATAAttribute("lastContributor", content.getLastContributor().getLogin());
                attrs.addCDATAAttribute("creationDate", DateUtils.zonedDateTimeToString(content.getCreationDate()));
                attrs.addCDATAAttribute("lastModified", DateUtils.zonedDateTimeToString(content.getLastModified()));
                
                if (content instanceof WorkflowAwareContent)
                {
                    WorkflowAwareContent waContent = (WorkflowAwareContent) content;
                    AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(waContent);
                    
                    attrs.addCDATAAttribute("workflowName", workflow.getWorkflowName(waContent.getWorkflowId()));
                    attrs.addCDATAAttribute("workflowStep", this.__joinSteps(workflow.getCurrentSteps(waContent.getWorkflowId())));
                    
                    int[] availableActions = _contentWorkflowHelper.getAvailableActions(waContent);
                    String workflowAvailableAction = StringUtils.join(ArrayUtils.toObject(availableActions), ", ");
                    attrs.addCDATAAttribute("workflowAvailableActions", workflowAvailableAction);
                }
                
                if (content instanceof ModifiableContent)
                {
                    attrs.addCDATAAttribute("isModifiable", "true");
                }
                
                if (content instanceof LockAwareAmetysObject)
                {
                    LockAwareAmetysObject lockAwareContent = (LockAwareAmetysObject) content;
                    if (lockAwareContent.isLocked())
                    {
                        attrs.addCDATAAttribute("locked", String.valueOf(lockAwareContent.isLocked()));
                        attrs.addCDATAAttribute("canUnlock", String.valueOf(_lockManager.canUnlock(lockAwareContent)));
                    }
                }
                
                saxAdditionalAtttributes (content, attrs);
                
                XMLUtils.startElement(contentHandler, "content", attrs);
                
                saxUserRights(content);
                saxAdditionalInformation(content);
                XMLUtils.endElement(contentHandler, "content");
            }
            catch (UnknownAmetysObjectException e)
            {
                getLogger().warn("Content of id '" + contentID + "' does not exist.", e);
                
                attrs.addCDATAAttribute("id", contentID);
                XMLUtils.createElement(contentHandler, "content-not-found", attrs);
            }
            
        }
        
        XMLUtils.endElement(contentHandler, "contents");
        contentHandler.endDocument();
    }

    private String __joinSteps(List<Step> steps)
    {
        StringBuffer a = new StringBuffer();
        
        for (Step s : steps)
        {
            if (a.length() > 0)
            {
                a.append(", ");
            }
            a.append(s.getStepId());
        }
        
        return a.toString();
    }
    
    
    /**
     * SAX the user rights on content
     * @param content The content to check rights
     * @throws SAXException if an error occurred while SAXing
     */
    protected void saxUserRights (Content content) throws SAXException
    {
        UserIdentity user = _userProvider.getUser();
        Set<String> rights = _rightManager.getUserRights(user, content);
        
        XMLUtils.startElement(contentHandler, "user-rights");
        
        for (String rightID : rights)
        {
            AttributesImpl rightAttr = new AttributesImpl();
            rightAttr.addCDATAAttribute("id", rightID);
            XMLUtils.createElement(contentHandler, "right", rightAttr);
        }
        XMLUtils.endElement(contentHandler, "user-rights");
    }
    
    /**
     * SAX the additional attributes
     * @param content The content
     * @param attrs The attributes
     * @throws SAXException if an error occurred while SAXing
     */
    protected void saxAdditionalAtttributes (Content content, AttributesImpl attrs) throws SAXException
    {
        // Nothing
    }
    
    /**
     * SAX additional information on content
     * @param content The content
     * @throws SAXException if an error occurred while SAXing
     */
    protected void saxAdditionalInformation (Content content) throws SAXException
    {
        // Nothing
    }
}
