/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Set;

import javax.jcr.Node;
import javax.jcr.RepositoryException;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.languages.Language;
import org.ametys.cms.languages.LanguagesManager;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ReactionableObject;
import org.ametys.cms.repository.ReactionableObjectHelper;
import org.ametys.cms.repository.ReportableObject;
import org.ametys.cms.repository.ReportableObjectHelper;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.cms.repository.comment.Comment;
import org.ametys.cms.repository.comment.CommentableContent;
import org.ametys.cms.repository.comment.CommentsDAO;
import org.ametys.cms.repository.comment.contributor.ContributorCommentableContent;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.data.external.ExternalizableDataProviderExtensionPoint;
import org.ametys.plugins.repository.dublincore.DublinCoreAwareAmetysObject;
import org.ametys.plugins.repository.jcr.JCRAmetysObject;
import org.ametys.plugins.repository.model.RepositoryDataContext;
import org.ametys.plugins.repository.version.VersionAwareAmetysObject;
import org.ametys.plugins.workflow.store.AmetysStep;
import org.ametys.plugins.workflow.support.WorkflowHelper;
import org.ametys.plugins.workflow.support.WorkflowProvider;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.type.DataContext;

import com.opensymphony.workflow.WorkflowException;
import com.opensymphony.workflow.spi.Step;

/**
 * Component responsible for generating SAX events representing a {@link Content}.
 */
public class ContentSaxer implements Serviceable, Component 
{
    /** Avalon role. */
    public static final String CMS_CONTENT_SAXER_ROLE = ContentSaxer.class.getName();
    
    private WorkflowProvider _workflowProvider;
    private WorkflowHelper _worklflowHelper;
    private ContentTypesHelper _contentTypesHelper; 
    private LanguagesManager _languageManager;
    private ExternalizableDataProviderExtensionPoint _externalizableDataProviderEP;
    private CommentsDAO _commentsDAO;
    private ReactionableObjectHelper _reactionableHelper;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _workflowProvider = (WorkflowProvider) manager.lookup(WorkflowProvider.ROLE);
        _worklflowHelper = (WorkflowHelper) manager.lookup(WorkflowHelper.ROLE);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
        _languageManager = (LanguagesManager) manager.lookup(LanguagesManager.ROLE);
        _externalizableDataProviderEP = (ExternalizableDataProviderExtensionPoint) manager.lookup(ExternalizableDataProviderExtensionPoint.ROLE);
        _commentsDAO = (CommentsDAO) manager.lookup(CommentsDAO.ROLE);
        _reactionableHelper = (ReactionableObjectHelper) manager.lookup(ReactionableObjectHelper.ROLE);
    }
    
    /**
     * Generates SAX events representing a {@link Content}. 
     * <br>When called with a non null tag name, a surrounding element will be generated, 
     * along with XML attributes representing the content's metadata (creation/modification/validation dates and authors, ...).
     * @param content the {@link Content}.
     * @param contentHandler the ContentHandler receving SAX events.
     * @param locale the {@link Locale} to use for eg. multilingual attributes.
     * @param view the View or null to select all attributes.
     * @param tagName the surrounding tag name or null to SAX events without root tag.
     * @param saxWorkflowStep if true, also produces SAX events for the current workflow step.
     * @param saxWorkflowInfo if true, also produces SAX events for detailed information about the current workflow step.
     * @param saxLanguageInfo if true, also produces SAX events for detailed information about the content language.
     * @param attributesTagName the name of the tag surrounding attributes. Used for legacy purposes.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    public void saxContent(Content content, ContentHandler contentHandler, Locale locale, View view, String tagName, boolean saxWorkflowStep, boolean saxWorkflowInfo, boolean saxLanguageInfo, String attributesTagName) throws SAXException
    {
        saxContent(content, contentHandler, locale, view, tagName, saxWorkflowStep, saxWorkflowInfo, saxLanguageInfo, attributesTagName, false);
    }
    
    /**
     * Generates SAX events representing a {@link Content}. 
     * <br>When called with a non null tag name, a surrounding element will be generated, 
     * along with XML attributes representing the content's metadata (creation/modification/validation dates and authors, ...).
     * @param content the {@link Content}.
     * @param contentHandler the ContentHandler receving SAX events.
     * @param locale the {@link Locale} to use for eg. multilingual attributes.
     * @param view the View or null to select all attributes.
     * @param tagName the surrounding tag name or null to SAX events without root tag.
     * @param saxWorkflowStep if true, also produces SAX events for the current workflow step.
     * @param saxWorkflowInfo if true, also produces SAX events for detailed information about the current workflow step.
     * @param saxLanguageInfo if true, also produces SAX events for detailed information about the content language.
     * @param attributesTagName the name of the tag surrounding attributes. Used for legacy purposes.
     * @param isEdition <code>true</code> if SAX events are generated in edition mode, <code>false</code> otherwise
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    public void saxContent(Content content, ContentHandler contentHandler, Locale locale, View view, String tagName, boolean saxWorkflowStep, boolean saxWorkflowInfo, boolean saxLanguageInfo, String attributesTagName, boolean isEdition) throws SAXException
    {
        if (StringUtils.isNotEmpty(tagName))
        {
            saxRootTag(content, contentHandler, locale, tagName);
        }
        
        saxBody(content, contentHandler, locale, view, tagName, saxWorkflowStep, saxWorkflowInfo, saxLanguageInfo, attributesTagName, isEdition);
        
        if (StringUtils.isNotEmpty(tagName))
        {
            XMLUtils.endElement(contentHandler, tagName);
        }
    }
    
    /**
     * Generates SAX events for the content data.
     * @param content the {@link Content}.
     * @param contentHandler the ContentHandler receving SAX events.
     * @param locale the {@link Locale} to use for eg. multilingual attributes.
     * @param view the View or null to select all attributes.
     * @param tagName the surrounding tag name or null to SAX events without root tag.
     * @param saxWorkflowStep if true, also produces SAX events for the current workflow step.
     * @param saxWorkflowInfo if true, also produces SAX events for detailed information about the current workflow step.
     * @param saxLanguageInfo if true, also produces SAX events for detailed information about the content language.
     * @param attributesTagName the name of the tag surrounding attributes. Used for legacy purposes.
     * @param isEdition <code>true</code> if SAX events are generated in edition mode, <code>false</code> otherwise
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    protected void saxBody(Content content, ContentHandler contentHandler, Locale locale, View view, String tagName, boolean saxWorkflowStep, boolean saxWorkflowInfo, boolean saxLanguageInfo, String attributesTagName, boolean isEdition) throws SAXException
    {
        saxContentTypes(content, contentHandler, true);
        saxAttributes(content, contentHandler, locale, view, tagName, attributesTagName, isEdition);
        
        if (saxWorkflowStep || saxWorkflowInfo)
        {
            saxWorkflowStep(content, contentHandler, saxWorkflowInfo);
        }
        
        if (saxLanguageInfo)
        {
            saxLanguage(content, contentHandler);
        }
        
        saxDublinCoreMetadata(content, contentHandler);
        
        if (content instanceof CommentableContent commentableContent)
        {
            saxContentComments(commentableContent, contentHandler);
        }
        
        if (content instanceof ContributorCommentableContent contributorCommentableContent)
        {
            saxContentContributorComments(contributorCommentableContent, contentHandler);
        }

        if (content instanceof ReactionableObject reactionableObject)
        {
            _reactionableHelper.saxReactions(reactionableObject, contentHandler);
        }
        
        if (content instanceof ReportableObject reportableObject)
        {
            ReportableObjectHelper.saxReports(reportableObject, contentHandler);
        }
    }
    
    /**
     * Generates a surrounding tag, with content metadata.
     * @param content the {@link Content}.
     * @param contentHandler the ContentHandler receving SAX events.
     * @param locale the {@link Locale} to use for eg. multilingual attributes.
     * @param tagName the surrounding tag name or null to SAX events without root tag.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    protected void saxRootTag(Content content, ContentHandler contentHandler, Locale locale, String tagName) throws SAXException
    {
        AttributesImpl attrs = new AttributesImpl();
        
        if (content instanceof JCRAmetysObject)
        {
            _addJcrAttributes((JCRAmetysObject) content, attrs);
        }
        
        attrs.addCDATAAttribute("id", content.getId());
        attrs.addCDATAAttribute("name", content.getName());
        attrs.addCDATAAttribute("title", content.getTitle(locale));
        if (content.getLanguage() != null)
        {
            attrs.addCDATAAttribute("language", content.getLanguage());
        }
        attrs.addCDATAAttribute("createdAt", DateUtils.zonedDateTimeToString(content.getCreationDate()));
        attrs.addCDATAAttribute("creator", UserIdentity.userIdentityToString(content.getCreator()));
        attrs.addCDATAAttribute("lastModifiedAt", DateUtils.zonedDateTimeToString(content.getLastModified()));
        
        ZonedDateTime lastValidatedAt = content.getLastValidationDate();
        if (lastValidatedAt != null)
        {
            attrs.addCDATAAttribute("lastValidatedAt", DateUtils.zonedDateTimeToString(lastValidatedAt));
        }
        
        attrs.addCDATAAttribute("lastContributor", UserIdentity.userIdentityToString(content.getLastContributor()));
        attrs.addCDATAAttribute("commentable", Boolean.toString(content instanceof CommentableContent));
        
        addAttributeIfNotNull (attrs, "iconGlyph", _contentTypesHelper.getIconGlyph(content));
        addAttributeIfNotNull (attrs, "iconDecorator", _contentTypesHelper.getIconDecorator(content));
        
        addAttributeIfNotNull (attrs, "smallIcon", _contentTypesHelper.getSmallIcon(content));
        addAttributeIfNotNull (attrs, "mediumIcon", _contentTypesHelper.getMediumIcon(content));
        addAttributeIfNotNull (attrs, "largeIcon", _contentTypesHelper.getLargeIcon(content));
        
        XMLUtils.startElement(contentHandler, tagName, attrs);
    }
    
    private void _addJcrAttributes(JCRAmetysObject content, AttributesImpl attrs)
    {
        Node node = content.getNode();
        try
        {
            attrs.addCDATAAttribute("uuid", node.getIdentifier());
        }
        catch (RepositoryException e)
        {
            throw new IllegalArgumentException("Unable to get jcr UUID for content '" + content.getId() + "'", e);
        }
        
        try
        {
            attrs.addCDATAAttribute("primaryType", node.getPrimaryNodeType().getName());
        }
        catch (RepositoryException e)
        {
            throw new IllegalArgumentException("Unable to get jcr Primary Type for content '" + content.getId() + "'", e);
        }
    }
    
    /**
     * Generates SAX events for {@link Content#getTypes content types}, and possibly {@link Content#getMixinTypes mixin types}
     * @param content the {@link Content}.
     * @param contentHandler the ContentHandler receving SAX events.
     * @param saxMixins if true, also produces SAX events for {@link Content#getMixinTypes mixin types}.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    protected void saxContentTypes(Content content, ContentHandler contentHandler, boolean saxMixins) throws SAXException
    {
        _saxContentTypes(content, contentHandler);
        if (saxMixins)
        {
            _saxMixins(content, contentHandler);
        }
    }
    
    private void _saxContentTypes(Content content, ContentHandler contentHandler) throws SAXException
    {
        String contentTypesTagName = "contentTypes";
        String singleContentTypeTagName = "contentType";
        XMLUtils.startElement(contentHandler, contentTypesTagName);
        for (String contentType : content.getTypes())
        {
            XMLUtils.createElement(contentHandler, singleContentTypeTagName, contentType);
        }
        XMLUtils.endElement(contentHandler, contentTypesTagName);
    }
    
    private void _saxMixins(Content content, ContentHandler contentHandler) throws SAXException
    {
        String mixinsTagName = "mixins";
        String singleMixinTagName = "mixin";
        XMLUtils.startElement(contentHandler, mixinsTagName);
        for (String mixinType : content.getMixinTypes())
        {
            XMLUtils.createElement(contentHandler, singleMixinTagName, mixinType);
        }
        XMLUtils.endElement(contentHandler, mixinsTagName);
    }
    
    /**
     * Generates SAX events for actual content's data.
     * @param content the {@link Content}.
     * @param contentHandler the ContentHandler receving SAX events.
     * @param locale the {@link Locale} to use for eg. multilingual attributes.
     * @param view the View or null to select all attributes.
     * @param tagName the surrounding tag name or null to SAX events without root tag.
     * @param attributesTagName the name of the tag surrounding attributes. Used for legacy purposes.
     * @param isEdition <code>true</code> if SAX events are generated in edition mode, <code>false</code> otherwise
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    protected void saxAttributes(Content content, ContentHandler contentHandler, Locale locale, View view, String tagName, String attributesTagName, boolean isEdition) throws SAXException
    {
        XMLUtils.startElement(contentHandler, attributesTagName);

        RepositoryDataContext context = RepositoryDataContext.newInstance()
                                                   .withLocale(locale);

        if (view == null)
        {
            content.dataToSAX(contentHandler, context.<DataContext>withEmptyValues(false));
        }
        else
        {
            if (isEdition)
            {
                Set<String> externalizableData = _externalizableDataProviderEP.getExternalizableDataPaths(content);
                content.dataToSAXForEdition(contentHandler, view, context.withExternalizableData(externalizableData));
            }
            else
            {
                content.dataToSAX(contentHandler, view, context.withEmptyValues(false));
            }
        }

        XMLUtils.endElement(contentHandler, attributesTagName);
    }
    
    /**
     * Generates SAX events representing the current workflow step.
     * @param content the {@link Content}.
     * @param contentHandler the ContentHandler receiving SAX events.
     * @param saxWorkflowInfo if true, also produces SAX events for detailed information about the current workflow step.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    protected void saxWorkflowStep(Content content, ContentHandler contentHandler, boolean saxWorkflowInfo) throws SAXException
    {
        if (content instanceof WorkflowAwareContent)
        {
            WorkflowAwareContent waContent = (WorkflowAwareContent) content;
            AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(waContent);
            
            try
            {
                long workflowId = waContent.getWorkflowId();
                String workflowName = workflow.getWorkflowName(workflowId);
                
                Step currentStep = getCurrentStep(waContent, workflow);
                
                int currentStepId = currentStep.getStepId();
                
                I18nizableText workflowStepName = new I18nizableText("application",  _worklflowHelper.getStepName(workflowName, currentStepId));
                
                AttributesImpl atts = new AttributesImpl();
                atts.addAttribute("", "id", "id", "CDATA", String.valueOf(currentStepId));
                atts.addAttribute("", "workflowName", "workflowName", "CDATA", String.valueOf(workflowName));
                
                if (saxWorkflowInfo)
                {
                    if ("application".equals(workflowStepName.getCatalogue()))
                    {
                        atts.addAttribute("", "icon-small", "icon-small", "CDATA", "/plugins/cms/resources_workflow/" + workflowStepName.getKey() + "-small.png");
                        atts.addAttribute("", "icon-medium", "icon-medium", "CDATA", "/plugins/cms/resources_workflow/" + workflowStepName.getKey() + "-medium.png");
                        atts.addAttribute("", "icon-large", "icon-large", "CDATA", "/plugins/cms/resources_workflow/" + workflowStepName.getKey() + "-large.png");
                    }
                    else
                    {
                        String pluginName = workflowStepName.getCatalogue().substring("plugin.".length());
                        atts.addAttribute("", "icon-small", "icon-small", "CDATA", "/plugins/" + pluginName + "/resources/img/workflow/" + workflowStepName.getKey() + "-small.png");
                        atts.addAttribute("", "icon-medium", "icon-medium", "CDATA", "/plugins/" + pluginName + "/resources/img/workflow/" + workflowStepName.getKey() + "-medium.png");
                        atts.addAttribute("", "icon-large", "icon-large", "CDATA", "/plugins/" + pluginName + "/resources/img/workflow/" + workflowStepName.getKey() + "-large.png");
                    }
                }
                
                XMLUtils.startElement(contentHandler, "workflow-step", atts);
                
                if (saxWorkflowInfo)
                {
                    workflowStepName.toSAX(contentHandler);
                }
                
                XMLUtils.endElement(contentHandler, "workflow-step");
            }
            catch (AmetysRepositoryException e)
            {
                // Current step id was not positioned
            }
            catch (WorkflowException e)
            {
                // Ignore, just don't SAX the workflow step.
            }
        }
    }

    /**
     * Get the current workflow step of the content.
     * @param content the {@link Content}.
     * @param workflow the associated workflow.
     * @return the current step
     * @throws WorkflowException if somethng got wrong processing workflow data.
     */
    protected Step getCurrentStep(WorkflowAwareContent content, AmetysObjectWorkflow workflow) throws WorkflowException
    {
        long workflowId = content.getWorkflowId();
        
        Step currentStep = (Step) workflow.getCurrentSteps(workflowId).get(0);
        
        if (content instanceof VersionAwareAmetysObject)
        {
            VersionAwareAmetysObject vaContent = (VersionAwareAmetysObject) content;
            String currentRevision = vaContent.getRevision();
            
            if (currentRevision != null)
            {
                
                String[] allRevisions = vaContent.getAllRevisions();
                int currentRevIndex = ArrayUtils.indexOf(allRevisions, currentRevision);
                
                if (currentRevIndex > -1 && currentRevIndex < (allRevisions.length - 1))
                {
                    String nextRevision = allRevisions[currentRevIndex + 1];
                    
                    Date currentRevTimestamp = vaContent.getRevisionTimestamp();
                    Date nextRevTimestamp = vaContent.getRevisionTimestamp(nextRevision);
                    
                    // Get all steps between the two revisions. 
                    List<Step> steps = _worklflowHelper.getStepsBetween(workflow, workflowId, currentRevTimestamp, nextRevTimestamp);
                    
                    // In the old workflow structure
                    // We take the second, which is current revision's last step.
                    if (steps.size() > 0 && steps.get(0) instanceof AmetysStep)
                    {
                        AmetysStep amStep = (AmetysStep) steps.get(0);
                        if (amStep.getProperty("actionFinishDate") != null)
                        {
                            // New workflow structure detected: cut the first workflow step
                            // in the list, as it belongs to the next version.
                            steps = steps.subList(1, steps.size());
                        }
                    }
                    
                    // Order by step descendant.
                    Collections.sort(steps, new Comparator<Step>()
                    {
                        public int compare(Step step1, Step step2)
                        {
                            return -Long.valueOf(step1.getId()).compareTo(step2.getId());
                        }
                    });
                    
                    // The first step in the list is the current version's last workflow step.
                    if (steps.size() > 0)
                    {
                        currentStep = steps.get(0);
                    }
                }
            }
        }
        
        return currentStep;
    }
    
    /**
     * Generates SAX events for the content's language.
     * @param content the {@link Content}.
     * @param contentHandler the ContentHandler receving SAX events.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    protected void saxLanguage(Content content, ContentHandler contentHandler) throws SAXException
    {
        String code = content.getLanguage();
        if (code != null)
        {
            Language language = _languageManager.getLanguage(code);
            
            AttributesImpl atts = new AttributesImpl();
            atts.addCDATAAttribute("code", code);
            
            if (language != null)
            {
                atts.addCDATAAttribute("icon-small", language.getSmallIcon());
                atts.addCDATAAttribute("icon-medium", language.getMediumIcon());
                atts.addCDATAAttribute("icon-large", language.getLargeIcon());
            }
            
            XMLUtils.startElement(contentHandler, "content-language", atts);
            if (language != null)
            {
                language.getLabel().toSAX(contentHandler);
            }
            XMLUtils.endElement(contentHandler, "content-language");
        }
    }
    
    /**
     * Generates SAX events for the DC metadata.
     * @param dcObject the {@link Content}.
     * @param contentHandler the ContentHandler receving SAX events.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    protected void saxDublinCoreMetadata(DublinCoreAwareAmetysObject dcObject, ContentHandler contentHandler) throws SAXException
    {
        XMLUtils.startElement(contentHandler, "dublin-core-metadata");
        saxIfNotNull("title", dcObject.getDCTitle(), contentHandler);
        saxIfNotNull("creator", dcObject.getDCCreator(), contentHandler);
        saxIfNotNull("subject", dcObject.getDCSubject(), contentHandler);
        saxIfNotNull("description", dcObject.getDCDescription(), contentHandler);
        saxIfNotNull("publisher", dcObject.getDCPublisher(), contentHandler);
        saxIfNotNull("contributor", dcObject.getDCContributor(), contentHandler);
        saxIfNotNull("date", dcObject.getDCDate(), contentHandler);
        saxIfNotNull("type", dcObject.getDCType(), contentHandler);
        saxIfNotNull("format", dcObject.getDCFormat(), contentHandler);
        saxIfNotNull("identifier", dcObject.getDCIdentifier(), contentHandler);
        saxIfNotNull("source", dcObject.getDCSource(), contentHandler);
        saxIfNotNull("language", dcObject.getDCLanguage(), contentHandler);
        saxIfNotNull("relation", dcObject.getDCRelation(), contentHandler);
        saxIfNotNull("coverage", dcObject.getDCCoverage(), contentHandler);
        saxIfNotNull("rights", dcObject.getDCRights(), contentHandler);
        XMLUtils.endElement(contentHandler, "dublin-core-metadata");
    }

    /**
     * Send a value if not null.
     * @param name the tag name.
     * @param value the value.
     * @param contentHandler the ContentHandler receving SAX events.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    protected void saxIfNotNull(String name, String value, ContentHandler contentHandler) throws SAXException
    {
        if (value != null)
        {
            XMLUtils.createElement(contentHandler, name, value);
        }
    }
    
    /**
     * Send values if not null.
     * @param name the tag name.
     * @param values the values.
     * @param contentHandler the ContentHandler receving SAX events.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    protected void saxIfNotNull(String name, String[] values, ContentHandler contentHandler) throws SAXException
    {
        if (values != null)
        {
            for (String value : values)
            {
                XMLUtils.createElement(contentHandler, name, value);
            }
        }
    }
    
    /**
     * Send a value if not null.
     * @param name the tag name.
     * @param value the value.
     * @param contentHandler the ContentHandler receving SAX events.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    protected void saxIfNotNull(String name, Date value, ContentHandler contentHandler) throws SAXException
    {
        if (value != null)
        {
            LocalDate ld = DateUtils.asLocalDate(value);
            XMLUtils.createElement(contentHandler, name, ld.format(DateTimeFormatter.ISO_LOCAL_DATE));
        }
    }
    
    /**
     * Generates SAX events for content's comments.
     * @param content the {@link Content}.
     * @param contentHandler the ContentHandler receiving SAX events.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    protected void saxContentComments(CommentableContent content, ContentHandler contentHandler) throws SAXException
    {
        List<Comment> comments = content.getComments(false, true);
        saxComments(comments, "comments", contentHandler);
    }
    
    /**
     * Generates SAX events for content's contributor comments.
     * @param content the {@link Content}.
     * @param contentHandler the ContentHandler receiving SAX events.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    protected void saxContentContributorComments(ContributorCommentableContent content, ContentHandler contentHandler) throws SAXException
    {
        List<Comment> comments = content.getContributorComments();
        saxComments(comments, "contributor-comments", contentHandler);
    }
    
    /**
     * Generates SAX events for content's comments.
     * @param comments the content's comments
     * @param commentsTagName the tag name of the SAX event to generate for comments
     * @param contentHandler the ContentHandler receiving SAX events.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    protected void saxComments(List<Comment> comments, String commentsTagName, ContentHandler contentHandler) throws SAXException
    {
        if (comments.size() > 0)
        {
            XMLUtils.startElement(contentHandler, commentsTagName);
            for (Comment comment : comments)
            {
                _commentsDAO.saxComment(contentHandler, comment, 0);
            }
            XMLUtils.endElement(contentHandler, commentsTagName);
        }
    }
    
    /**
     * Add attribute if value is not null
     * @param attrs The attributes
     * @param name The name of attribute
     * @param value The value
     */
    protected void addAttributeIfNotNull (AttributesImpl attrs, String name, String value)
    {
        if (value != null)
        {
            attrs.addCDATAAttribute(name, value);
        }
    }
}
