/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * The report object, used internally to provide useful informations once
 * the whole duplication process has ended.
 */
public final class CopyReport
{
    private String _viewName;
    private String _fallbackViewName;
    private String _baseContentId;
    private String _baseContentTitle;
    private Boolean _baseContentIsReferenceTable;
    
    private CopyMode _mode;
    
    private String _targetContentId;
    private String _targetContentTitle;
    private boolean _targetContentIsReferenceTable;
    
    private List<CopyReport> _innerReports;
    
    private List<String> _copiedAttachments;
    
    private CopyState _state;
    private I18nizableText _errorMsg;
    
    /** Available copy mode */
    public enum CopyMode
    {
        /** Creation : a new content is created to perform the copy */
        CREATION,
        /** Edition : attributes are duplicated into an already existing target content */
        EDITION
    }
    
    /** The possible state of the copy */
    public enum CopyState
    {
        /** Running : when the copy is still running */
        RUNNING,
        /** Success : when the copy has ended well */
        SUCCESS,
        /** Error : when an error was encountered during the copy (except for attributes copy errors). */
        ERROR
    }
    
    /**
     * Constructor.
     * If the title of the base content if available, should be used.
     * @param baseContentId The content id of the base content
     * @param baseContentIsRefTable Indicates if the base content is a reference table's entry (can be null if unavailable)
     * @param viewName The view name
     * @param fallbackViewName The fallback view name. Use for legacy purpose.
     * @param copyMode The mode of copy
     */
    public CopyReport(String baseContentId, Boolean baseContentIsRefTable, String viewName, String fallbackViewName, CopyMode copyMode)
    {
        this(baseContentId, null, baseContentIsRefTable, viewName, fallbackViewName, copyMode);
    }
    
    /**
     * Constructor.
     * @param baseContentId The content id of the base content
     * @param baseContentTitle The title of the base content
     * @param baseContentIsRefTable Indicates if the base content is a reference table's entry
     * @param viewName The view name
     * @param fallbackViewName The fallback view name. Use for legacy purpose.
     * @param copyMode The mode of copy
     */
    public CopyReport(String baseContentId, String baseContentTitle, Boolean baseContentIsRefTable, String viewName, String fallbackViewName, CopyMode copyMode)
    {
        _baseContentId = baseContentId;
        _baseContentTitle = baseContentTitle;
        _baseContentIsReferenceTable = baseContentIsRefTable;
        _viewName = viewName;
        _fallbackViewName = fallbackViewName;
        _mode = copyMode;
        
        _innerReports = new LinkedList<>();
        _copiedAttachments = new LinkedList<>();
        
        _state = CopyState.RUNNING;
        
        _errorMsg = null;
    }
    
    /**
     * Set whether the base content is a reference table's entry.
     * @param isRefTable true for reference table's entry.
     */
    public void setReferenceTable(boolean isRefTable)
    {
        _baseContentIsReferenceTable = isRefTable;
    }
    
    /**
     * Set the target content title.
     * @param title The target content title.
     */
    public void setTargetContentTitle(String title)
    {
        _targetContentTitle = title;
    }
    
    /**
     * Notify information about the target content to the report.
     * @param id The content id
     * @param title The content title
     * @param isRefTable Indicates if the target content is a reference table's entry.
     */
    public void notifyContentCreation(String id, String title, boolean isRefTable)
    {
        _targetContentId = id;
        _targetContentTitle = title;
        _targetContentIsReferenceTable = isRefTable;
    }
    
    /**
     * Notify a copy success
     */
    public void notifyContentCopySuccess()
    {
        _state = CopyState.SUCCESS;
    }
    
    /**
     * Notify that the copy ended with an error
     */
    public void notifyContentCopyError()
    {
        _state = CopyState.ERROR;
    }
    
    /**
     * Set the error message
     * @param errorMsg the error message
     */
    public void setErrorMessage(I18nizableText errorMsg)
    {
        _errorMsg = errorMsg;
    }
    
    /**
     * Get the error message
     * @return the error message. Can be null.
     */
    public I18nizableText getErrorMessage()
    {
        return _errorMsg;
    }
    
    /**
     * Add an attachments to the copied attachments report list
     * @param relPath The relative path where to copy
     */
    public void addAttachment(String relPath)
    {
        _copiedAttachments.add(relPath);
    }
    
    /**
     * Get the status of the copy
     * @return A copy state enum value
     */
    public CopyState getStatus()
    {
        return _state;
    }
    
    /**
     * Get the mode of the copy
     * @return A cope mode enum value
     */
    public CopyMode getMode()
    {
        return _mode;
    }
    
    /**
     * Get the name of the view used for the copy.
     * @return the view name
     */
    public String getViewName()
    {
        return _viewName;
    }
    
    /**
     * Get the name of the fallback view used for the copy.
     * @return the fallback view name
     */
    public String getFallbackViewName()
    {
        return _fallbackViewName;
    }
    
    /**
    /**
     * Get the base content identifier
     * @return The identifier
     */
    public String getBaseContentId()
    {
        return _baseContentId;
    }
    
    /**
     * Get the base content title.
     * @return The title or null (if unavailable)
     */
    public String getBaseContentTitle()
    {
        return _baseContentTitle;
    }
    
    /**
     * Is the base content a reference table entry ?
     * @return true if it is, null if unknown.
     */
    public Boolean getBaseContentIsReferenceTable()
    {
        return _baseContentIsReferenceTable;
    }
    
    /**
     * Get the target content identifier
     * @return The identifier
     */
    public String getTargetContentId()
    {
        return _targetContentId;
    }
    
    /**
     * Get the target content title
     * @return The title
     */
    public String getTargetContentTitle()
    {
        return _targetContentTitle;
    }
    
    /**
     * Is the target content a reference table entry?
     * @return true if it is.
     */
    public boolean getTargetContentIsReferenceTable()
    {
        return _targetContentIsReferenceTable;
    }
    
    /**
     * Get the list of child copy reports.
     * Each child report denotes an inner copy
     * @return The list of copy reports.
     */
    public List<CopyReport> getChildReports()
    {
        return Collections.unmodifiableList(_innerReports);
    }
    
    /**
     * Get the list of the copied attachments.
     * @return List of paths (relative to the root attachment node of the target content)
     */
    public List<String> getCopiedAttachments()
    {
        return Collections.unmodifiableList(_copiedAttachments);
    }
    
    /**
     * Add a child report to the report. This must be done when an copy is requested 
     * @param report The report of the copy to add
     */
    public void addReport(CopyReport report)
    {
        _innerReports.add(report);
    }
}
