/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.acting.AbstractAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;

import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;

/**
 * This action export the wrapping XSLT URI.<br>
 * Content negociation will be performed with the given URI: if content.xsl is asked, content-[contentType].xsl will be looked up first.
 * 
 * This action needs the parameter <code>path</code> pointing the default path to the XSL. 
 */
public class GetWrapperContextAction extends AbstractAction implements Serviceable
{
    /** The source resolver */
    protected SourceResolver _srcResolver;
    /** The content types helper */
    protected ContentTypesHelper _contentTypesHelper;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _srcResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
    }
    
    @Override
    public Map<String, String> act(Redirector redirector, org.apache.cocoon.environment.SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        HashMap<String, String> result = new HashMap<>();
        String uri = parameters.getParameter("path");
        String wrapperURI = getWrapperURI(request, uri);
        
        result.put("wrapper-uri", wrapperURI);
        return result;
    }
    
    /**
     * Compute the wrapping URI for the current Content.
     * @param request the current Request.
     * @param uri the default uri.
     * @return the wrapper XSLT URI.
     * @throws IOException if an error occurs whuile resolving Sources.
     */
    protected String getWrapperURI(Request request, String uri) throws IOException
    {
        for (String wrapperURI : getSourceURIs(request, uri))
        {
            Source src = _srcResolver.resolveURI(wrapperURI);
            if (!src.exists())
            {
                if (getLogger().isDebugEnabled())
                {
                    getLogger().debug("Failed to find a stylesheet at '" + wrapperURI + "'.");
                }
            }
            else
            {
                return wrapperURI;
            }
        }
        
        return null;
    }
    
    /**
     * Get the source URI for wrapped content
     * @param request The request
     * @param requestedURI The requested URI
     * @return the wrapped URI
     */
    protected List<String> getSourceURIs (Request request, String requestedURI)
    {
        Content content = (Content) request.getAttribute(Content.class.getName());
        
        List<String> sourceURIs = new ArrayList<>();
        
        // Ex: context://WEB-INF/stylesheets/content/_wrapper/content-[cType].xsl 
        int index = requestedURI.indexOf(".xsl");
        
        String cTypeId = _contentTypesHelper.getContentTypeIdForRendering(content);
        sourceURIs.add("context://WEB-INF/" + requestedURI.substring(0, index) + "-" + cTypeId + ".xsl");
        
        cTypeId = _contentTypesHelper.getFirstContentType(content).getId();
        sourceURIs.add("context://WEB-INF/" + requestedURI.substring(0, index) + "-" + cTypeId + ".xsl");
        
        sourceURIs.add("context://WEB-INF/" + requestedURI);
        
        sourceURIs.add("workspace:" + request.getAttribute("workspaceName") + "://" + requestedURI);
        
        return sourceURIs;
        
    }
}
