/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.io.IOException;
import java.io.InputStream;

import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceNotFoundException;
import org.apache.excalibur.source.SourceValidity;
import org.apache.excalibur.source.impl.validity.TimeStampValidity;

import org.ametys.cms.data.Resource;
import org.ametys.cms.data.RichText;
import org.ametys.cms.repository.Content;
import org.ametys.core.cocoon.source.NamedSource;

/**
 * {@link Source} representing a {@link RichText}'s local file.
 */
public class RichTextFileSource implements NamedSource
{
    private Content _content;
    private RichText _richText;
    private String _filename;
    private Resource _resource;
    
    private String _uri;
    
    /**
     * Constructor.
     * @param content the {@link Content}.
     * @param richText the {@link RichText}.
     * @param filename the local file name.
     * @param uri this source's uri.
     */
    public RichTextFileSource(Content content, RichText richText, String filename, String uri)
    {
        _content = content;
        _richText = richText;
        _filename = filename;
        _uri = uri;
    }
    
    private Resource _getResource() throws SourceNotFoundException
    {
        if (_resource == null)
        {
            if (!exists())
            {
                throw new SourceNotFoundException("No rich text resource found at " + _uri);
            }
            
            _resource = _richText.getAttachment(_filename);
        }
        
        return _resource;
    }

    public boolean exists()
    {
        return _content != null && _richText != null && _richText.hasAttachment(_filename);
    }

    public InputStream getInputStream() throws IOException, SourceNotFoundException
    {
        return _getResource().getInputStream();
    }

    public String getURI()
    {
        return _uri;
    }

    public String getScheme()
    {
        return RichTextFileSourceFactory.RICHTEXT_FILE_SCHEME;
    }

    public SourceValidity getValidity()
    {
        return new TimeStampValidity(getLastModified());
    }
    
    public String getName()
    {
        return _filename;
    }

    public void refresh()
    {
        // does nothing
    }

    public String getMimeType()
    {
        try
        {
            return _getResource().getMimeType();
        }
        catch (SourceNotFoundException e)
        {
            return null;
        }
    }

    public long getContentLength()
    {
        try
        {
            return _getResource().getLength();
        }
        catch (SourceNotFoundException e)
        {
            return -1;
        }
    }

    public long getLastModified()
    {
        return _content != null ? _content.getLastModified().toInstant().toEpochMilli() : 0;
    }
}
