/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Map;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.avalon.framework.thread.ThreadSafe;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceFactory;

import org.ametys.cms.data.RichText;
import org.ametys.cms.repository.Content;
import org.ametys.core.util.FilenameUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * {@link SourceFactory} for {@link RichText}'s local files.
 */
public class RichTextFileSourceFactory extends AbstractLogEnabled implements SourceFactory, ThreadSafe, Serviceable, Contextualizable
{
    /** source scheme */
    public static final String RICHTEXT_FILE_SCHEME = "richtext";
    
    private Context _context;
    private AmetysObjectResolver _resolver;
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    public Source getSource(String location, Map parameters) throws IOException, MalformedURLException
    {
        String uri = location.substring((RICHTEXT_FILE_SCHEME + "://").length());
        
        // uri are like <contentPath>@<attributePath>;<filename>
        int i = uri.indexOf('@');
        int j = uri.indexOf(';', i);
        String path = uri.substring(0, i);
        String attribute = uri.substring(i + 1, j);
        String filename = uri.substring(j + 1);
        
        Request request = ContextHelper.getRequest(_context);
        Content content = (Content) request.getAttribute(Content.class.getName());
        
        try
        {
            if (content == null) 
            {
                content = _resolver.resolveByPath(FilenameUtils.decode(path));
            }
            
            RichText richText = content.getValue(attribute);
    
            return new RichTextFileSource(content, richText, FilenameUtils.decode(filename), location);
        }
        catch (Exception e)
        {
            getLogger().error("Unable to resolve richText data for uri " + location, e);
            
            // returns an unexisting Source
            return new RichTextFileSource(content, null, FilenameUtils.decode(filename), location);
        }
    }

    public void release(Source source)
    {
        // nothing to do
    }
}
