/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import org.apache.commons.lang.ArrayUtils;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * This handle parse a rich text to get only text nodes.
 * HTML expert is ignored.
 * Call {@link #getValue()} to get parse string.
 */
public class RichTextHandler extends DefaultHandler
{
    /** List of elements to be ignored during the body text retrieval processing*/
    protected static final String[] __IGNORED_ELEMENTS = new String[]{"htmlexpert"};
    
    private static final String __SEPARATOR = " ";
    
    private StringBuilder _bodyText = new StringBuilder();
    private boolean _ignoreElmt;
    private boolean _newElmt;
    private int _excerptLength;
    
    /**
     * Creates a rich text handler 
     */
    public RichTextHandler()
    {
        super();
        _excerptLength = 0;
    }
    
    /**
     * Creates a rich text handler to parse a rich text with a limit a characters 
     * @param excerptLength The length for content excerpt
     */
    public RichTextHandler(int excerptLength)
    {
        super();
        _excerptLength = excerptLength;
    }
    
    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        if (_newElmt && _bodyText.length() > 0)
        {
            _bodyText.append(__SEPARATOR);
            _newElmt = false;
        }
        
        if (!_ignoreElmt)
        {
            String text = new String(ch, start, length);
            _bodyText.append(text);
        }
    }
    
    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        if (ArrayUtils.contains(__IGNORED_ELEMENTS, localName))
        {
            _ignoreElmt = true;
        }
        else
        {
            _newElmt = true;
        }
    }
    
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if (ArrayUtils.contains(__IGNORED_ELEMENTS, localName))
        {
            _ignoreElmt = false;
        }
    }
    
    /**
     * Gets the value of the parsed data
     * @return the value of the parsed data
     */
    public String getValue()
    {
        String bodyText = _bodyText.toString();
        bodyText = bodyText.replaceAll("\\t|\\r?\\n", " ").replaceAll("\\s+", " ").trim();
        
        if (_excerptLength > 0 && bodyText.length() > _excerptLength)
        {
            int summaryEndIndex = bodyText.lastIndexOf(' ', _excerptLength);
            if (summaryEndIndex == -1)
            {
                summaryEndIndex = bodyText.length();
            }
            
            return bodyText.substring(0, summaryEndIndex) + "…";
        }

        return bodyText;
    }
}
