/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.jcr.RepositoryException;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.avalon.framework.thread.ThreadSafe;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang.StringUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.languages.Language;
import org.ametys.cms.languages.LanguagesManager;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.cms.repository.comment.CommentableContent;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.core.util.I18nUtils;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.plugins.workflow.support.WorkflowProvider;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;
import org.ametys.runtime.i18n.I18nizableText;

import com.opensymphony.workflow.loader.StepDescriptor;
import com.opensymphony.workflow.loader.WorkflowDescriptor;
import com.opensymphony.workflow.spi.Step;

/**
 * Component helper for saxing content's properties
 *
 */
public class SAXContentHelper implements Component, ThreadSafe, Serviceable
{
    /** The avalon ROLE */
    public static final String ROLE = SAXContentHelper.class.getName();
    
    private UserHelper _userHelper;
    private WorkflowProvider _workflowProvider;
    private LanguagesManager _languagesManager;
    private ContentTypeExtensionPoint _contentTypeExtensionPoint;
    private I18nUtils _i18nUtils;
    
    public void service(ServiceManager smanager) throws ServiceException
    {
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
        _languagesManager = (LanguagesManager) smanager.lookup(LanguagesManager.ROLE);
        _workflowProvider = (WorkflowProvider) smanager.lookup(WorkflowProvider.ROLE);
        _userHelper = (UserHelper) smanager.lookup(UserHelper.ROLE);
        _i18nUtils = (I18nUtils) smanager.lookup(I18nUtils.ROLE);
    }

    /**
     * SAX the content's content types
     * @param contentHandler The content handler where to sax
     * @param content the content The content concerned
     * @throws SAXException if an error occurred while saxing
     */
    public void saxContentTypes (ContentHandler contentHandler, Content content) throws SAXException
    {
        List<String> labels = new ArrayList<>();
        
        for (String cTypeId : content.getTypes())
        {
            ContentType cType = _contentTypeExtensionPoint.getExtension(cTypeId);
            labels.add(_i18nUtils.translate(cType.getLabel(), content.getLanguage()));
        }
        XMLUtils.createElement(contentHandler, "contentType", StringUtils.join(labels, ", "));
    }
    
    /**
     * SAX the content's mixins
     * @param contentHandler The content handler where to sax
     * @param content the content The content concerned
     * @throws SAXException if an error occurred while saxing
     */
    public void saxMixinTypes (ContentHandler contentHandler, Content content) throws SAXException
    {
        List<String> labels = new ArrayList<>();
        
        for (String cTypeId : content.getMixinTypes())
        {
            ContentType cType = _contentTypeExtensionPoint.getExtension(cTypeId);
            labels.add(_i18nUtils.translate(cType.getLabel(), content.getLanguage()));
        }
        XMLUtils.createElement(contentHandler, "mixin", StringUtils.join(labels, ", "));
    }
    
    /**
     * SAX the content's mixins and content types
     * @param contentHandler The content handler where to sax
     * @param content the content The content concerned
     * @throws SAXException if an error occurred while saxing
     */
    public void saxContentTypesAndMixins (ContentHandler contentHandler, Content content) throws SAXException
    {
        List<String> labels = new ArrayList<>();
        
        for (String cTypeId : content.getTypes())
        {
            ContentType cType = _contentTypeExtensionPoint.getExtension(cTypeId);
            labels.add(_i18nUtils.translate(cType.getLabel(), content.getLanguage()));
        }
        
        for (String cTypeId : content.getMixinTypes())
        {
            ContentType cType = _contentTypeExtensionPoint.getExtension(cTypeId);
            labels.add(_i18nUtils.translate(cType.getLabel(), content.getLanguage()));
        }
        
        XMLUtils.createElement(contentHandler, "contentTypeOrMixin", StringUtils.join(labels, ", "));
    }
    
    /**
     * SAX the creator
     * @param contentHandler The content handler where to sax
     * @param content the content The content concerned
     * @throws SAXException if an error occurred while saxing
     */
    public void saxCreator (ContentHandler contentHandler, Content content) throws SAXException
    {
        UserIdentity user = content.getCreator();
        XMLUtils.createElement(contentHandler, "creator", user.getLogin());
        
        String fullName = _userHelper.getUserFullName(user);
        if (fullName != null)
        {
            XMLUtils.createElement(contentHandler, "creatorDisplay", fullName);
        }
    }
    
    /**
     * SAX the contributor
     * @param contentHandler The content handler where to sax
     * @param content the content The content concerned
     * @throws SAXException if an error occurred while saxing
     */
    public void saxContributor (ContentHandler contentHandler, Content content) throws SAXException
    {
        UserIdentity user = content.getLastContributor();
        XMLUtils.createElement(contentHandler, "contributor", user.getLogin());
        
        String fullName = _userHelper.getUserFullName(user);
        if (fullName != null)
        {
            XMLUtils.createElement(contentHandler, "contributorDisplay", fullName);
        }
    }
    
    /**
     * SAX the last modified date
     * @param contentHandler The content handler where to sax
     * @param content the content The content concerned
     * @throws SAXException if an error occurred while saxing
     */
    public void saxLastModified (ContentHandler contentHandler, Content content) throws SAXException
    {
        XMLUtils.createElement(contentHandler, "lastModified", DateUtils.zonedDateTimeToString(content.getLastModified()));
    }
    
    /**
     * SAX the date of creation
     * @param contentHandler The content handler where to sax
     * @param content the content The content concerned
     * @throws SAXException if an error occurred while saxing
     */
    public void saxCreationDate (ContentHandler contentHandler, Content content) throws SAXException
    {
        XMLUtils.createElement(contentHandler, "creationDate", DateUtils.zonedDateTimeToString(content.getCreationDate()));
    }
    
    /**
     * SAX the content's comments
     * @param contentHandler The content handler where to sax
     * @param content the content The content concerned
     * @throws SAXException if an error occurred while saxing
     */
    public void saxComments (ContentHandler contentHandler, Content content) throws SAXException
    {
        if (content instanceof CommentableContent)
        {
            boolean hasComment = ((CommentableContent) content).getComments(true, true).size() != 0;
            XMLUtils.createElement(contentHandler, "comments", String.valueOf(hasComment));
        }
    }
    
    /**
     * SAX the content language
     * @param contentHandler The content handler where to sax
     * @param content the content The content concerned
     * @throws SAXException if an error occurred while saxing
     */
    public void saxContentLanguage (ContentHandler contentHandler, Content content) throws SAXException
    {
        String contentLanguage = content.getLanguage();
        if (contentLanguage != null)
        {
            Language language = _languagesManager.getAvailableLanguages().get(contentLanguage);
            if (language != null)
            {
                XMLUtils.createElement(contentHandler, "contentLanguageIcon", language.getSmallIcon());
            }
            
            XMLUtils.createElement(contentHandler, "contentLanguage", contentLanguage);
        }
    }
    
    /**
     * SAX workflow current step of a {@link Content}
     * @param contentHandler The content handler where to sax
     * @param content the content The content concerned
     * @throws SAXException if an error occurs while SAXing
     * @throws RepositoryException if an error occurs while retrieving current step
     */
    public void saxContentCurrentState(ContentHandler contentHandler, Content content) throws SAXException, RepositoryException
    {
        if (content instanceof WorkflowAwareContent)
        {
            WorkflowAwareContent waContent = (WorkflowAwareContent) content;
            int currentStepId = 0;
            
            AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(waContent);
            long workflowId = waContent.getWorkflowId();
            
            Iterator iterator = workflow.getCurrentSteps(workflowId).iterator();
            
            while (iterator.hasNext())
            {
                Step step = (Step) iterator.next();
                currentStepId = step.getStepId();
            }
            
            String workflowName = workflow.getWorkflowName(workflowId);
            WorkflowDescriptor workflowDescriptor = workflow.getWorkflowDescriptor(workflowName);
            StepDescriptor stepDescriptor = workflowDescriptor.getStep(currentStepId);
            
            if (stepDescriptor != null)
            {
                I18nizableText workflowStepName = new I18nizableText("application", stepDescriptor.getName());
                
                AttributesImpl attr = new AttributesImpl();
                attr.addAttribute("", "id", "id", "CDATA", String.valueOf(currentStepId));
                
                XMLUtils.startElement(contentHandler, "workflowStep", attr);
                workflowStepName.toSAX(contentHandler);
                XMLUtils.endElement(contentHandler, "workflowStep");
                
                String[] icons = new String[]{"-small", "-medium", "-large"};
                for (String icon : icons)
                {
                    if ("application".equals(workflowStepName.getCatalogue()))
                    {
                        XMLUtils.createElement(contentHandler, "workflow-icon" + icon, "/plugins/cms/resources_workflow/" + workflowStepName.getKey() + icon + ".png");
                    }
                    else
                    {
                        String pluginName = workflowStepName.getCatalogue().substring("plugin.".length());
                        XMLUtils.createElement(contentHandler, "workflow-icon" + icon, "/plugins/" + pluginName + "/resources/img/workflow/" + workflowStepName.getKey() + icon + ".png");
                    }
                }
            }
        }
    }
}
