/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.autosave;

import javax.jcr.Node;

import org.apache.avalon.framework.service.ServiceException;

import org.ametys.cms.data.holder.ModifiableIndexableDataHolder;
import org.ametys.cms.data.holder.impl.DefaultModifiableModelAwareDataHolder;
import org.ametys.cms.repository.ModelAwareJCRAmetysObject;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.type.ModelItemTypeExtensionPoint;
import org.ametys.plugins.repository.jcr.DefaultTraversableAmetysObjectFactory;
import org.ametys.plugins.repository.jcr.SimpleAmetysObject;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.runtime.model.DefaultElementDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.UnknownTypeException;
import org.ametys.runtime.model.type.ModelItemTypeConstants;

/**
 * Implementation of a {@link ModelAwareJCRAmetysObject}, for content backup.<br>
 * This implementation heavily relies on its {@link ContentBackupAmetysObjectFactory} counterpart.
 */
public class ContentBackupAmetysObject extends SimpleAmetysObject<ContentBackupAmetysObjectFactory> implements ModelAwareJCRAmetysObject
{
    /** Content identifier data name */
    public static final String AUTOSAVE_CONTENT_ID = "contentId";
    /** Creator data name */
    public static final String AUTOSAVE_CREATOR = "creator";
    /** Temporary date data name */
    public static final String AUTOSAVE_TEMP_DATE = "tempContentDate";
    /** Comments data name */
    public static final String AUTOSAVE_COMMENTS = "jsonComments";
    /** Valid values data name */
    public static final String AUTOSAVE_VALUES = "values";
    /** Invalid values data name */
    public static final String AUTOSAVE_INVALID_VALUES = "invalid";
    /** Repeaters data name */
    public static final String AUTOSAVE_REPEATERS = "repeaters";
    
    /** Attribute name data name for values (valid and invalid) data */
    public static final String ATTRIBUTE_NAME = "name";
    /** Attribute value data name for values (valid and invalid) data */
    public static final String ATTRIBUTE_VALUE = "value";
    
    /** Repeater name data name */
    public static final String REPEATER_NAME = "name";
    /** Repeater count data name */
    public static final String REPEATER_COUNT = "count";
    /** Repeater prefix data name */
    public static final String REPEATER_PREFIX = "prefix";
    
    private ModifiableIndexableDataHolder _dataHolder;

    /**
     * Creates a DefaultTraversableAmetysObject.
     * @param node the node backing this {@link AmetysObject}.
     * @param parentPath the parent path in the Ametys hierarchy.
     * @param factory the {@link DefaultTraversableAmetysObjectFactory} which creates the AmetysObject.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public ContentBackupAmetysObject(Node node, String parentPath, ContentBackupAmetysObjectFactory factory) throws AmetysRepositoryException
    {
        super(node, parentPath, factory);

        // Initialize Model and data holder
        ModifiableRepositoryData repositoryData = new JCRRepositoryData(getNode());
        _dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _createContentBackupModel());
    }
    
    /**
     * Creates the content backup model
     * @return the created model
     * @throws AmetysRepositoryException if an error occurs.
     */
    protected Model _createContentBackupModel() throws AmetysRepositoryException
    {
        try
        {
            String role = ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC;
            return Model.of(ContentBackupClientInteraction.class.getName(), ContentBackupClientInteraction.class.getName(),
                    DefaultElementDefinition.of(AUTOSAVE_CONTENT_ID, false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                    DefaultElementDefinition.of(AUTOSAVE_CREATOR, false, org.ametys.cms.data.type.ModelItemTypeConstants.USER_ELEMENT_TYPE_ID, role),
                    DefaultElementDefinition.of(AUTOSAVE_TEMP_DATE, false, ModelItemTypeConstants.DATETIME_TYPE_ID, role),
                    DefaultElementDefinition.of(AUTOSAVE_COMMENTS, false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                    RepeaterDefinition.of(AUTOSAVE_REPEATERS, role,
                            DefaultElementDefinition.of(REPEATER_NAME, false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                            DefaultElementDefinition.of(REPEATER_PREFIX, false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                            DefaultElementDefinition.of(REPEATER_COUNT, false, ModelItemTypeConstants.STRING_TYPE_ID, role)),
                    RepeaterDefinition.of(AUTOSAVE_VALUES, role,
                            DefaultElementDefinition.of(ATTRIBUTE_NAME, false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                            DefaultElementDefinition.of(ATTRIBUTE_VALUE, false, ModelItemTypeConstants.STRING_TYPE_ID, role)),
                    RepeaterDefinition.of(AUTOSAVE_INVALID_VALUES, role,
                            DefaultElementDefinition.of(ATTRIBUTE_NAME, false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                            DefaultElementDefinition.of(ATTRIBUTE_VALUE, false, ModelItemTypeConstants.STRING_TYPE_ID, role)));
        }
        catch (UnknownTypeException | BadItemTypeException | ServiceException e)
        {
            throw new AmetysRepositoryException("An error occurred while create the content backup model", e);
        }
    }

    public ModifiableIndexableDataHolder getDataHolder()
    {
        return _dataHolder;
    }
}
