/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.autosave;

import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.ui.Callable;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.core.util.JSONUtils;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.ModifiableAmetysObject;
import org.ametys.plugins.repository.ModifiableTraversableAmetysObject;
import org.ametys.plugins.repository.RemovableAmetysObject;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.ModifiableModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeaterEntry;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeaterEntry;
import org.ametys.plugins.repository.jcr.ModelAwareJCRAmetysObject;
import org.ametys.plugins.repository.jcr.NameHelper;
import org.ametys.plugins.repository.query.QueryHelper;
import org.ametys.plugins.repository.query.expression.Expression;
import org.ametys.plugins.repository.query.expression.Expression.Operator;
import org.ametys.plugins.repository.query.expression.StringExpression;
import org.ametys.runtime.config.Config;

/**
 * Component for manipulating auto-backup on contents
 *
 */
public class ContentBackupClientInteraction extends AbstractLogEnabled implements Serviceable, Component
{
    private static final String __AUTOSAVE_NODETYPE = RepositoryConstants.NAMESPACE_PREFIX + ":autoSave";
    
    private AmetysObjectResolver _resolver;
    private CurrentUserProvider _currentUserProvider;
    private JSONUtils _jsonUtils;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _currentUserProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
        _jsonUtils = (JSONUtils) manager.lookup(JSONUtils.ROLE);
    }
    
    /**
     * Get the content backup information
     * @param contentId The content id
     * @return The saved data
     */
    @Callable
    public Map<String, Object> getContentBackup (String contentId)
    {
        if (getLogger().isDebugEnabled())
        {
            getLogger().debug(String.format("Get automatic data backup for content '%s'", contentId));
        }
        
        Map<String, Object> result = new HashMap<>();
        
        boolean autoSaveEnabled = Config.getInstance().getValue("automatic.save.enabled");
        Long autoSaveFrequency = Config.getInstance().getValue("automatic.save.frequency");
        
        result.put("enabled", autoSaveEnabled);
        result.put("frequency", autoSaveFrequency);
        
        ModelAwareDataHolder contentNode = getContentNode(contentId, false);
        
        if (contentNode != null)
        {
            Map<String, Object> autoBackup = new HashMap<>();
            
            UserIdentity creator = contentNode.getValue(ContentBackupAmetysObject.AUTOSAVE_CREATOR);
            autoBackup.put("creator", creator != null ? creator.toString() : StringUtils.EMPTY);
            autoBackup.put("contentId", contentId);
            
            ZonedDateTime backupDate = contentNode.getValue(ContentBackupAmetysObject.AUTOSAVE_TEMP_DATE);
            if (backupDate != null)
            {
                autoBackup.put("date", DateUtils.zonedDateTimeToString(backupDate));
            }
            
            autoBackup.put("data", getBackupData(contentNode));
            
            result.put("auto-backup", autoBackup);
        }
        
        return result;
    }
    
    /**
     * Retrieves the content backup information
     * @param dataHolder the content backup data holder
     * @return the content backup information
     */
    protected Map<String, Object> getBackupData (ModelAwareDataHolder dataHolder)
    {
        Map<String, Object> data = new HashMap<>();
        
        // Get and generate the valid attributes values.
        if (dataHolder.hasValue(ContentBackupAmetysObject.AUTOSAVE_VALUES))
        {
            List<Map<String, Object>> values = getValuesBackupData(dataHolder, ContentBackupAmetysObject.AUTOSAVE_VALUES);
            data.put("attributes", values);
        }
        
        // Get and generate the invalid attributes values.
        if (dataHolder.hasValue(ContentBackupAmetysObject.AUTOSAVE_INVALID_VALUES))
        {
            List<Map<String, Object>> invalidValues = getValuesBackupData(dataHolder, ContentBackupAmetysObject.AUTOSAVE_INVALID_VALUES);
            data.put("invalid-attributes", invalidValues);
        }
        
        // Get the field comments
        data.put("comments", dataHolder.getValue(ContentBackupAmetysObject.AUTOSAVE_COMMENTS, false, "{}"));
        
        // Get and generate the repeater item counts.
        if (dataHolder.hasValue(ContentBackupAmetysObject.AUTOSAVE_REPEATERS))
        {
            List<Map<String, Object>> repeaters = getRepeatersBackupData(dataHolder);
            data.put("repeaters", repeaters);
        }
        
        return data;
    }

    /**
     * Retrieves the values backup information
     * @param dataHolder the content backup data holder
     * @param valuesRepeaterPath the path of the values attribute
     * @return the values backup information
     */
    protected List<Map<String, Object>> getValuesBackupData(ModelAwareDataHolder dataHolder, String valuesRepeaterPath)
    {
        List<Map<String, Object>> attributes = new ArrayList<>();
        
        ModelAwareRepeater values = dataHolder.getRepeater(valuesRepeaterPath);
        for (ModelAwareRepeaterEntry entry : values.getEntries())
        {
            Map<String, Object> attribute = new HashMap<>();
            
            attribute.put("name", entry.getValue(ContentBackupAmetysObject.ATTRIBUTE_NAME, false, StringUtils.EMPTY));
            
            String encodedValue = entry.getValue(ContentBackupAmetysObject.ATTRIBUTE_VALUE, false, StringUtils.EMPTY);
            attribute.put("value", _decodeValue(encodedValue));
            
            attributes.add(attribute);
        }
        
        return attributes;
    }
    
    private Object _decodeValue (String value)
    {
        if (StringUtils.isBlank(value))
        {
            return StringUtils.EMPTY;
        }
        
        try
        {
            return _jsonUtils.convertJsonToMap(value);
        }
        catch (IllegalArgumentException e)
        {
            // Failed to convert into map, continue
        }
        
        try
        {
            return _jsonUtils.convertJsonToList(value);
        }
        catch (IllegalArgumentException e)
        {
            // Failed to convert into list, continue
        }
        
        try
        {
            return _jsonUtils.convertJsonToArray(value);
        }
        catch (IllegalArgumentException e)
        {
            // Failed to convert into array, continue
        }
        
        return value;
    }

    /**
     * Retrieves the repeaters backup information
     * @param dataHolder the content backup data holder
     * @return the repeaters backup information
     */
    protected List<Map<String, Object>> getRepeatersBackupData(ModelAwareDataHolder dataHolder)
    {
        List<Map<String, Object>> repeaters = new ArrayList<>();
        
        ModelAwareRepeater repeatersData = dataHolder.getRepeater(ContentBackupAmetysObject.AUTOSAVE_REPEATERS);
        for (ModelAwareRepeaterEntry entry : repeatersData.getEntries())
        {
            Map<String, Object> repeater = new HashMap<>();
            
            repeater.put("name", entry.getValue(ContentBackupAmetysObject.REPEATER_NAME, false, StringUtils.EMPTY));
            repeater.put("prefix", entry.getValue(ContentBackupAmetysObject.REPEATER_PREFIX, false, StringUtils.EMPTY));
            repeater.put("count", entry.getValue(ContentBackupAmetysObject.REPEATER_COUNT, false, "0"));
            
            repeaters.add(repeater);
        }
        
        return repeaters;
    }
    
    /**
     * Delete an automatic backup for a content.
     * @param contentId The content id
     * @return A empty map
     */
    @Callable
    public Map<String, Object> deleteContentBackup (String contentId)
    {
        if (StringUtils.isNotEmpty(contentId))
        {
            if (getLogger().isDebugEnabled())
            {
                getLogger().debug(String.format("Delete automatic backup for content '%s'", contentId));
            }
            
            RemovableAmetysObject contentNode = getContentNode(contentId, false);
            
            if (contentNode != null)
            {
                ModifiableAmetysObject parent = (ModifiableAmetysObject) contentNode.getParent();
                
                contentNode.remove();
                
                parent.saveChanges();
            }
        }
        
        return java.util.Collections.EMPTY_MAP; 
    }
    
    /**
     * Store an automatic backup for a content.
     * @param contentId The content id
     * @param values The valid values to store
     * @param invalidValues The invalid values to store
     * @param comments The comments as JSON string
     * @param repeaters The repeaters to store
     * @return A empty Map
     */
    @Callable
    public Map<String, Object> setContentBackup (String contentId, Map<String, Object> values, Map<String, Object> invalidValues, String comments, Collection<Map<String, String>> repeaters)
    {
        if (getLogger().isDebugEnabled())
        {
            getLogger().debug(String.format("Start automatic backup for content '%s'", contentId));
        }
        
        UserIdentity currentUser = _currentUserProvider.getUser();
        
        ModelAwareJCRAmetysObject contentNode = getContentNode(contentId, true);
        
        // Remove all existing metadata.
        removeAllRepeaterEntries(contentNode);
        
        contentNode.setValue(ContentBackupAmetysObject.AUTOSAVE_TEMP_DATE, ZonedDateTime.now());
        contentNode.setValue(ContentBackupAmetysObject.AUTOSAVE_CREATOR, currentUser);
        
        // Store the valid values.
        storeValues(contentNode, ContentBackupAmetysObject.AUTOSAVE_VALUES, values);
        
        // Store the invalid values.
        storeValues(contentNode, ContentBackupAmetysObject.AUTOSAVE_INVALID_VALUES, invalidValues);
        
        // Store the comments.
        contentNode.setValue(ContentBackupAmetysObject.AUTOSAVE_COMMENTS, comments);
        
        // Store the repeater item counts.
        storeRepeaters(contentNode, repeaters);
        
        contentNode.saveChanges();
        
        return java.util.Collections.EMPTY_MAP; 
    }
    
    /**
     * Store the attribute values
     * @param dataHolder the content backup data holder
     * @param valuesRepeaterPath the path of the values attribute
     * @param values the values to store, as a Map of values, indexed by name.
     */
    protected void storeValues(ModifiableModelAwareDataHolder dataHolder, String valuesRepeaterPath, Map<String, Object> values)
    {
        ModifiableModelAwareRepeater valuesData = dataHolder.getRepeater(valuesRepeaterPath, true);
        for (String name : values.keySet())
        {
            ModifiableModelAwareRepeaterEntry entry = valuesData.addEntry();
            
            entry.setValue(ContentBackupAmetysObject.ATTRIBUTE_NAME, name);

            // Store value as JSON string
            Object value = values.get(name);
            entry.setValue(ContentBackupAmetysObject.ATTRIBUTE_VALUE, _encodeValue(value));
        }
    }
    
    private String _encodeValue (Object value)
    {
        if (value instanceof Map || value instanceof List)
        {
            return _jsonUtils.convertObjectToJson(value);
        }
        else
        {
            return value != null ? value.toString() : "";
        }
    }
    
    /**
     * Store the repeater item counts to be able to re-initialize them
     * @param dataHolder the content backup data holder
     * @param repeaters the repeaters data
     */
    protected void storeRepeaters(ModifiableModelAwareDataHolder dataHolder, Collection<Map<String, String>> repeaters)
    {
        ModifiableModelAwareRepeater repeatersData = dataHolder.getRepeater(ContentBackupAmetysObject.AUTOSAVE_REPEATERS, true);
        for (Map<String, String> repeaterData : repeaters)
        {
            ModifiableModelAwareRepeaterEntry entry = repeatersData.addEntry();
            entry.setValue(ContentBackupAmetysObject.REPEATER_NAME, repeaterData.get("name"));
            entry.setValue(ContentBackupAmetysObject.REPEATER_PREFIX, repeaterData.get("prefix"));
            entry.setValue(ContentBackupAmetysObject.REPEATER_COUNT, repeaterData.get("count"));
        }
    }
    
    /**
     * Remove all the repeater entries of a given data holder.
     * @param dataHolder the data holder to clear.
     */
    protected void removeAllRepeaterEntries(ModifiableModelAwareDataHolder dataHolder)
    {
        for (String dataName : dataHolder.getDataNames())
        {
            if (org.ametys.plugins.repository.data.type.ModelItemTypeConstants.REPEATER_TYPE_ID.equals(dataHolder.getType(dataName).getId()))
            {
                dataHolder.removeValue(dataName);
            }
        }
    }

    /**
     * Get the storage node for a content in the automatic backup space, or create it if it doesn't exist.
     * @param contentId the content ID.
     * @param createNew <code>true</code> to create automatically the node when missing.
     * @return the content backup storage node.
     */
    protected ModelAwareJCRAmetysObject getContentNode(String contentId, boolean createNew)
    {
        ModelAwareJCRAmetysObject contentNode = null;
        
        // Query the content node by path and contentId property.
        Expression expression = new StringExpression(ContentBackupAmetysObject.AUTOSAVE_CONTENT_ID, Operator.EQ, contentId);
        String query = QueryHelper.getXPathQuery(null, __AUTOSAVE_NODETYPE, expression);
        
        AmetysObjectIterable<ModelAwareJCRAmetysObject> autoSaveObjects = _resolver.query(query);
        Iterator<ModelAwareJCRAmetysObject> it = autoSaveObjects.iterator();
        
        // Get or create the content node.
        if (it.hasNext())
        {
            contentNode = it.next();
        }
        else if (createNew)
        {
            ModifiableTraversableAmetysObject tempRoot = getOrCreateTempRoot();
            
            String objectName = NameHelper.filterName(contentId);
            contentNode = (ModelAwareJCRAmetysObject) tempRoot.createChild(objectName, __AUTOSAVE_NODETYPE);
            contentNode.setValue(ContentBackupAmetysObject.AUTOSAVE_CONTENT_ID, contentId);
            
            tempRoot.saveChanges();
        }
        
        return contentNode;
    }
    
    /**
     * Get the temporary backup root, or create it if it doesn't exist.
     * @return the temporary backup root.
     */
    protected ModifiableTraversableAmetysObject getOrCreateTempRoot()
    {
        ModifiableTraversableAmetysObject pluginsRoot = _resolver.resolveByPath("/ametys:plugins");
        
        ModifiableTraversableAmetysObject cmsNode = null;
        if (pluginsRoot.hasChild("cms"))
        {
            cmsNode = (ModifiableTraversableAmetysObject) pluginsRoot.getChild("cms");
        }
        else
        {
            cmsNode = (ModifiableTraversableAmetysObject) pluginsRoot.createChild("cms", "ametys:unstructured");
        }
        
        ModifiableTraversableAmetysObject editionNode = null;
        if (cmsNode.hasChild("edition"))
        {
            editionNode = (ModifiableTraversableAmetysObject) cmsNode.getChild("edition");
        }
        else
        {
            editionNode = (ModifiableTraversableAmetysObject) cmsNode.createChild("edition", "ametys:unstructured");
        }
        
        ModifiableTraversableAmetysObject tempNode = null;
        if (editionNode.hasChild("temp"))
        {
            tempNode = (ModifiableTraversableAmetysObject) editionNode.getChild("temp");
        }
        else
        {
            tempNode = (ModifiableTraversableAmetysObject) editionNode.createChild("temp", "ametys:unstructured");
        }
        
        if (pluginsRoot.needsSave())
        {
            pluginsRoot.saveChanges();
        }
        
        return tempNode;
    }
}
