/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.compare;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeater;
import org.ametys.plugins.repository.data.type.ModelItemTypeConstants;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.SimpleViewItemGroup;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.compare.DataChangeType;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Object used to compare two contents
 *
 */
public class ContentComparator extends AbstractLogEnabled implements Component, Serviceable
{
    /** The Avalon role */
    public static final String ROLE = ContentComparator.class.getName();
    
    private ContentTypesHelper _contentTypesHelper;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        this._contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
    }
    
    /**
     * Compare 2 contents, for all their common attributes (contents with different contentTypes are rejected)
     * @param content1 1st content
     * @param content2 new content
     * @return {@link ContentComparatorResult}
     * @throws AmetysRepositoryException repository exception
     * @throws IOException IO exception
     */
    public ContentComparatorResult compare(Content content1, Content content2) throws AmetysRepositoryException, IOException
    {
        return compare(content1, content2, null);
    }
    
    /**
     * Compare 2 contents, for all their common attributes
     * @param content1 1st content
     * @param content2 new content
     * @param strictCompare true to reject contents with different content types
     * @return {@link ContentComparatorResult}
     * @throws AmetysRepositoryException repository exception
     * @throws IOException IO exception
     */
    public ContentComparatorResult compare(Content content1, Content content2, boolean strictCompare) throws AmetysRepositoryException, IOException
    {
        return compare(content1, content2, null, strictCompare);
    }
    
    /**
     * Compare 2 contents, filtering with a view (contents with different contentTypes are rejected)
     * @param content1 1st content
     * @param content2 new content
     * @param viewName name of the view
     * @return {@link ContentComparatorResult}
     * @throws AmetysRepositoryException repository exception
     * @throws IOException IO exception
     */
    public ContentComparatorResult compare(Content content1, Content content2, String viewName) throws AmetysRepositoryException, IOException
    {
        return compare(content1, content2, viewName, true);
    }
    
    /**
     * Compare 2 contents, filtering with a view
     * @param content1 1st content
     * @param content2 new content
     * @param viewName name of the view
     * @param strictCompare true to reject contents with different content types
     * @return {@link ContentComparatorResult}
     * @throws AmetysRepositoryException repository exception
     * @throws IOException IO exception
     */
    public ContentComparatorResult compare(Content content1, Content content2, String viewName, boolean strictCompare) throws AmetysRepositoryException, IOException
    {
        return _compare(content1, content2, viewName, strictCompare);
    }
    
    private ContentComparatorResult _compare(Content content1, Content content2, String viewName, boolean strictCompare) throws AmetysRepositoryException, IOException
    {
        //if strictCompare, we check if the types are strictly equals
        List<String> types1 = Arrays.asList(content1.getTypes());
        List<String> types2 = Arrays.asList(content2.getTypes());
        if (strictCompare && (types1.size() != types2.size() || !types1.containsAll(types2) || !types2.containsAll(types1)))
        {
            ContentComparatorResult result = new ContentComparatorResult(content1, content2);
            result.setNotEquals();
            return result;
        }
        else
        {
            View view = _generateView(viewName, content1);
            ContentComparatorResult result = new ContentComparatorResult(content1, content2, viewName, view);
            if (view != null)
            {
                for (ViewItem viewItem : view.getViewItems())
                {
                    _compareModelItems(result, content1, content2, viewItem, StringUtils.EMPTY);
                }
            }
            else
            {
                result.setNotEquals();
            }
            
            return result;
        }
    }
    
    /**
     * Generate a view using a name (can be null) and a content
     * @param viewName name of the view. If null, a __generated__ view will be created using all content's attributes
     * @param content content where the view should be retrieved
     * @return {@link View} from the name, or generated
     */
    private View _generateView(String viewName, Content content)
    {
        if (viewName != null)
        {
            return _contentTypesHelper.getView(viewName, content.getTypes(), content.getMixinTypes());
        }
        else
        {
            View  view = View.of(content.getModel());
            view.setName("__generated__");
            view.setLabel(new I18nizableText("Live edition metadataset"));
            view.setDescription(new I18nizableText("Live edition metadataset"));
            view.setSmallIcon(null);
            view.setMediumIcon(null);
            view.setLargeIcon(null);
            view.setInternal(true);
            
            return view;
        }
    }

    /**
     * Compare 2 model items and add the differences in result
     * @param result global result where differences will be stored
     * @param content1 1st content
     * @param content2 2nd content
     * @param viewItem view item that will be compared
     * @param parentDataPath absolute data path of the container of the current view item 
     * @throws AmetysRepositoryException repository exception
     * @throws IOException IO Exception (comparison of InputStream mainly)
     */
    private void _compareModelItems(ContentComparatorResult result, Content content1, Content content2, ViewItem viewItem, String parentDataPath) throws AmetysRepositoryException, IOException
    {
        // If the view item is a reference to a model item, compare the model items and their values
        if (viewItem instanceof ModelViewItem modelViewItem)
        {
            ModelItem modelItemFromView = modelViewItem.getDefinition();
            String modelItemPath = modelItemFromView.getPath();
            
            ModelItem modelItem1 = null;
            if (content1.hasDefinition(modelItemPath))
            {
                modelItem1 = content1.getDefinition(modelItemPath);
            }
            
            ModelItem modelItem2 = null;
            if (content2.hasDefinition(modelItemPath))
            {
                modelItem2 = content2.getDefinition(modelItemPath);
            }
            
            if (modelItem1 != null && modelItem2 != null && modelItem1.getModel().equals(modelItem2.getModel()))
            {
                String modelItemName = modelViewItem.getDefinition().getName();
                String currentDataPath = StringUtils.isNotEmpty(parentDataPath) ? parentDataPath + ModelItem.ITEM_PATH_SEPARATOR + modelItemName : modelItemName;
                if (_checkModelItemPresence(result, content1, content2, currentDataPath, true))
                {
                    if (ModelItemTypeConstants.COMPOSITE_TYPE_ID.equals(modelItemFromView.getType().getId())) 
                    {
                        _compareComposite(result, (ModelViewItemGroup) modelViewItem, content1, content2, modelItemFromView, currentDataPath);
                    }
                    else if (ModelItemTypeConstants.REPEATER_TYPE_ID.equals(modelItemFromView.getType().getId()))
                    {
                        _compareRepeater(result, (ModelViewItemGroup) modelViewItem, content1, content2, modelItemFromView, currentDataPath);
                    }
                    else
                    {
                        // If the model item is not a group, it has to be an ElementDefinition
                        _compareElement(result, content1, content2, (ElementDefinition) modelItemFromView, currentDataPath);
                    }
                }
            }
            else
            {
                // TODO CMS-9978 Ignore or add the difference? What happens if an item is defined by a content but not by the other one?
                // TODO CMS-9978 Shouldn't we compare the definitions? (cardinality, ...)?
            }
        }

        // If the view item contains other view items, compare the children
        if (viewItem instanceof SimpleViewItemGroup simpleViewItemGroup)
        {
            for (ViewItem child : simpleViewItemGroup.getViewItems())
            {
                _compareModelItems(result, content1, content2, child, parentDataPath);
            }
        }
    }
    
    /**
     * Check if a model item is present in both contents and add a change if requested
     * @param result The result
     * @param content1 1st content
     * @param content2 2nd content
     * @param dataPath absolute path of the data to check (also used to log)
     * @param logChange true to add a change in the list (if one has it and the other doesn't)
     * @return true if both contents have the node
     */
    private boolean _checkModelItemPresence(ContentComparatorResult result, Content content1, Content content2, String dataPath, boolean logChange)
    {
        boolean hasContent1Value = content1.hasValue(dataPath);
        boolean hasContent2Value = content2.hasValue(dataPath);
        
        if (hasContent1Value && hasContent2Value)
        {
            return true;
        }
        else if (logChange && hasContent1Value && !hasContent2Value)
        {
            ModelItem definition = content1.getDefinition(dataPath);
            result.addChange(new ContentComparatorChange(dataPath, StringUtils.EMPTY, definition, DataChangeType.REMOVED));
        }
        else if (logChange && !hasContent1Value && hasContent2Value)
        {
            ModelItem definition = content2.getDefinition(dataPath);
            result.addChange(new ContentComparatorChange(dataPath, StringUtils.EMPTY, definition, DataChangeType.ADDED));
        }
        
        return false;
    }
    
    /**
     * Compare 2 composites
     * @param result global result where differences will be stored
     * @param viewItemGroup view item of the current composite
     * @param content1 1st content
     * @param content2 2nd content
     * @param definition definition of the composite
     * @param compositeDataPath absolute data path of the current composite
     * @throws AmetysRepositoryException repository exception
     * @throws IOException IO Exception
     */
    private void _compareComposite(ContentComparatorResult result, ModelViewItemGroup viewItemGroup, Content content1, Content content2, ModelItem definition, String compositeDataPath) throws AmetysRepositoryException, IOException
    {
        for (ViewItem viewItem : viewItemGroup.getViewItems())
        {
            _compareModelItems(result, content1, content2, viewItem, compositeDataPath);
        }
    }
    
    /**
     * Compare 2 repeaters
     * @param result global result where differences will be stored
     * @param viewItemGroup view item of the current repeater
     * @param content1 1st content
     * @param content2 2nd content
     * @param definition definition of the repeater
     * @param repeaterDataPath absolute data path of the current repeater
     * @throws AmetysRepositoryException repository exception
     * @throws IOException IO Exception
     */
    private void _compareRepeater(ContentComparatorResult result, ModelViewItemGroup viewItemGroup, Content content1, Content content2, ModelItem definition, String repeaterDataPath) throws AmetysRepositoryException, IOException
    {
        ModelAwareRepeater repeater1 = content1.getRepeater(repeaterDataPath);
        ModelAwareRepeater repeater2 = content2.getRepeater(repeaterDataPath);

        if (repeater1.getSize() > repeater2.getSize())
        {
            for (int i = repeater2.getSize() + 1; i <= repeater1.getSize(); i++)
            {
                result.addChange(new ContentComparatorChange(repeaterDataPath + "[" + Integer.toString(i) + "]", StringUtils.EMPTY, definition, DataChangeType.REMOVED));
            }
        }
        else if (repeater1.getSize() < repeater2.getSize())
        {
            for (int i = repeater1.getSize() + 1; i <= repeater2.getSize(); i++)
            {
                result.addChange(new ContentComparatorChange(repeaterDataPath + "[" + Integer.toString(i) + "]", StringUtils.EMPTY, definition, DataChangeType.ADDED));
            }
        }
        
        // Compare common entries
        int numberOfCommonEntries = Math.min(repeater1.getSize(), repeater2.getSize());
        for (int i = 1; i <= numberOfCommonEntries; i++)
        {
            String repeaterEntryDataPath = repeaterDataPath + "[" + Integer.toString(i) + "]";
            if (_checkModelItemPresence(result, content1, content2, repeaterEntryDataPath, true))
            {
                // Compare repeater's children for each entry
                for (ViewItem viewItem : viewItemGroup.getViewItems())
                {
                    _compareModelItems(result, content1, content2, viewItem, repeaterEntryDataPath);
                }
            }
        }
    }
    
    /**
     * Compare 2 elements
     * @param result global result where differences will be stored
     * @param content1 1st content
     * @param content2 2nd content
     * @param definition definition of the attribute
     * @param elementPath absolute data path of the current attribute
     * @throws AmetysRepositoryException repository exception
     * @throws IOException IO Exception
     */
    private void _compareElement(ContentComparatorResult result, Content content1, Content content2, ElementDefinition<?> definition, String elementPath) throws AmetysRepositoryException, IOException
    {
        definition.getType()
                  .compareValues(content1.getValue(elementPath), content2.getValue(elementPath))
                  .map(change ->
                  {
                      String detailPath = change.getRight();
                      return new ContentComparatorChange(elementPath, detailPath, definition, change.getLeft(), change.getMiddle());
                  })
                  .forEach(result::addChange);
    }
}
