/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.compare;

import org.apache.commons.lang3.StringUtils;

import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.compare.DataChangeType;
import org.ametys.runtime.model.compare.DataChangeTypeDetail;

/**
 * One of the changes between 2 contents
 */
public class ContentComparatorChange
{
    private final String _attributeDataPath;
    private final String _detailDataPath;
    private final ModelItem _modelItem;
    private final DataChangeType _changeType;
    private final DataChangeTypeDetail _changeTypeDetail;
    
    /**
     * Constructor with no details for this change
     * @param attributeDataPath path of the attribute data (can be rep[2]/someComposite/someRichText for instance)
     * @param detailDataPath relative path of the detail (can be lastModified for a RichText for instance)
     * @param modeItem the model item
     * @param type type of change
     */
    public ContentComparatorChange(String attributeDataPath, String detailDataPath, ModelItem modeItem, DataChangeType type)
    {
        this(attributeDataPath, detailDataPath, modeItem, type, DataChangeTypeDetail.NONE);
    }
    
    /**
     * Constructor with change detail
     * @param attributeDataPath path of the attribute data (can be rep[2]/someComposite/someRichText for instance)
     * @param detailDataPath relative path of the detail (can be lastModified for a RichText for instance)
     * @param modeItem the model item
     * @param type type of change
     * @param detail detail of the change
     */
    public ContentComparatorChange(String attributeDataPath, String detailDataPath, ModelItem modeItem, DataChangeType type, DataChangeTypeDetail detail)
    {
        _attributeDataPath = attributeDataPath;
        _detailDataPath = detailDataPath;
        _modelItem = modeItem;
        _changeType = type;
        _changeTypeDetail = detail;
    }
    
    /**
     * path of the change
     * @return path of the change
     */
    public String getPath()
    {
        if (StringUtils.isEmpty(_detailDataPath))
        {
            return _attributeDataPath;
        }
        else
        {
            return _attributeDataPath + ModelItem.ITEM_PATH_SEPARATOR + _detailDataPath;
        }
    }
    
    /**
     * path of the attribute holding the change (is less specific than {@link #getPath})
     * @return path of the attribute holding the change
     */
    public String getAttributeDataPath()
    {
        return _attributeDataPath;
    }
    
    /**
     * relative path of the detail of the change
     * @return relative path of the detail of the change
     */
    public String getDetailDataPath()
    {
        return _detailDataPath;
    }
    
    /**
     * the model item of the change
     * @return model item of the change
     */
    public ModelItem getModelItem()
    {
        return _modelItem;
    }
    
    /**
     * Type of change
     * @return Type of change
     */
    public DataChangeType getDataChangeType()
    {
        return _changeType;
    }
    
    /**
     * Detail of change
     * @return Detail of change
     */
    public DataChangeTypeDetail getDataChangeTypeDetail()
    {
        return _changeTypeDetail;
    }

    @Override
    public String toString()
    {
        return "ContentComparatorChange [path=" + getPath() + ", changeType=" + _changeType + ", changeTypeDetail=" + _changeTypeDetail + "]";
    }
    
    
}
