/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.compare;

import java.util.ArrayList;
import java.util.List;

import org.ametys.cms.repository.Content;
import org.ametys.runtime.model.View;

/**
 * Result of a comparison between 2 contents
 */
public class ContentComparatorResult
{
    private Content _content1;
    private Content _content2;
    private String _viewName;
    private View _view;
    
    private boolean _areEquals;
    private List<ContentComparatorChange> _changes;
    
    /**
     * Create a new result from 2 contents
     * @param content1 1st content
     * @param content2 2nd content
     */
    public ContentComparatorResult(Content content1, Content content2)
    {
        this(content1, content2, null, null);
    }
    
    /**
     * Create a new result from 2 contents and the name of a metadataSet
     * @param content1 1st content
     * @param content2 2nd content
     * @param viewName name of the {@link View}
     * @param view {@link View}
     */
    public ContentComparatorResult(Content content1, Content content2, String viewName, View view)
    {
        this._content1 = content1;
        this._content2 = content2;
        this._viewName = viewName;
        this._view = view;
        if (content1 != null && content2 != null)
        {
            this._areEquals = true;
        }
        else
        {
            this._areEquals = false;
        }
        this._changes = new ArrayList<>();
    }
    
    /**
     * are the contents equals ?
     * @return true if equals
     */
    public boolean areEquals()
    {
        return this._areEquals;
    }
    
    /**
     * If contents are not equals, a list of {@link ContentComparatorChange} is generated
     * @return list of {@link ContentComparatorChange}
     */
    public List<ContentComparatorChange> getChanges()
    {
        return this._changes;
    }
    
    /**
     * get the 1st content of the comparison
     * @return 1st content
     */
    public Content getContent1()
    {
        return this._content1;
    }
    
    /**
     * get the 1st content of the comparison
     * @return 2nd content
     */
    public Content getContent2()
    {
        return this._content2;
    }
    
    /**
     * get the name of the {@link View}
     * @return View's name (can be null)
     */
    public String getViewName()
    {
        return this._viewName;
    }
    
    /**
     * get the {@link View} (generated from the view name, of from all attributes)
     * @return {@link View}
     */
    public View getView()
    {
        return this._view;
    }
    
    /**
     * add a change to the list
     * @param change a new change
     */
    protected void addChange(ContentComparatorChange change)
    {
        this._changes.add(change);
        setNotEquals();
    }
    
    /**
     * declare that the 2 contents are not equals
     */
    protected void setNotEquals()
    {
        this._areEquals = false;
    }

    @Override
    public String toString()
    {
        return "ContentComparatorResult [content1=" + _content1 + ", content2=" + _content2 + ", viewName=" + _viewName + ", view=" + _view + ", areEquals=" + _areEquals
                + ", changes=" + _changes + "]";
    }
    
    
}
