/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.consistency;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.quartz.JobExecutionContext;

import org.ametys.cms.content.consistency.ContentConsistencyManager.ConsistenciesReport;
import org.ametys.core.right.RightManager;
import org.ametys.core.schedule.progression.ContainerProgressionTracker;
import org.ametys.core.ui.mail.StandardMailBodyHelper;
import org.ametys.core.ui.mail.StandardMailBodyHelper.MailBodyBuilder;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.util.I18nUtils;
import org.ametys.core.util.mail.SendMailHelper;
import org.ametys.core.util.mail.SendMailHelper.NamedStream;
import org.ametys.plugins.core.impl.schedule.AbstractStaticSchedulable;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.servlet.RuntimeConfig;

import jakarta.mail.MessagingException;

/**
 * Content consistency schedulable: generate consistency information for all contents.
 * Sends a report e-mail if there are inconsistencies.
 */
public class CheckContentConsistencySchedulable extends AbstractStaticSchedulable implements Initializable
{
    /** The report e-mail will be sent to users who possess this right on the application context. */
    protected static final String __MAIL_RIGHT = "CMS_Rights_ReceiveConsistencyReport";
    
    /** The server base URL. */
    protected String _baseUrl;
    
    /** The report directory. */
    protected File _reportDirectory;
    
    /** The ametys object resolver. */
    protected AmetysObjectResolver _ametysResolver;
    
    /** The avalon source resolver. */
    protected SourceResolver _sourceResolver;
    
    /** The rights manager. */
    protected RightManager _rightManager;
    
    /** The i18n utils. */
    protected I18nUtils _i18nUtils;
    
    /** The content of "from" field in emails. */
    protected String _mailFrom;
    
    /** The content consistency manager */
    protected ContentConsistencyManager _contentConsistencyManager;

    private ContentConsistencySearchModel _contentConsistencySearchModel;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        
        _ametysResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        
        _rightManager = (RightManager) manager.lookup(RightManager.ROLE);
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
        
        _contentConsistencySearchModel = (ContentConsistencySearchModel) manager.lookup(ContentConsistencySearchModel.ROLE);
        _contentConsistencyManager = (ContentConsistencyManager) manager.lookup(ContentConsistencyManager.ROLE);
    }
    
    public void initialize() throws Exception
    {
        _baseUrl = StringUtils.stripEnd(StringUtils.removeEndIgnoreCase(Config.getInstance().getValue("cms.url"), "index.html"), "/");
        _mailFrom = Config.getInstance().getValue("smtp.mail.from");
        _reportDirectory = new File(RuntimeConfig.getInstance().getAmetysHome(), "consistency");
    }
    
    @Override
    public void execute(JobExecutionContext context, ContainerProgressionTracker progressionTracker) throws Exception
    {
        ConsistenciesReport report = _contentConsistencyManager.checkAllContents();
        
        Set<UserIdentity> users = getUsersToNotify();
        if (!report.isEmpty() && !users.isEmpty())
        {
            _sendEmail(report, users);
        }
    }

    /**
     * Send a reminder e-mail to all the users who have the right to edit.
     * @param report the consistency report
     * @param users the users to notify
     * @throws IOException if an error occurs building or sending the mail.
     * @throws ProcessingException if an error occurs 
     */
    protected void _sendEmail(ConsistenciesReport report, Set<UserIdentity> users) throws IOException, ProcessingException
    {
        I18nizableText i18nSubject = _getMailSubject(report);
        String subject = _i18nUtils.translate(i18nSubject);
        
        String htmlBody = _getMailBody(report);
        
        if (StringUtils.isEmpty(htmlBody))
        {
            return;
        }
        try (InputStream attachment = _getReport(report.results()))
        {
            _sendMails(subject, htmlBody, users, _mailFrom, attachment);
        }
    }

    /**
     * Compute the list of user to notify
     * @return a list of user identity
     */
    protected Set<UserIdentity> getUsersToNotify()
    {
        Set<UserIdentity> users = _rightManager.getAllowedUsers(__MAIL_RIGHT, "/cms").resolveAllowedUsers(Config.getInstance().getValue("runtime.mail.massive.sending"));
        return users;
    }
    
    private InputStream _getReport(List<String> resultsId) throws ProcessingException
    {
        Request request = ContextHelper.getRequest(_context);
        Source attachement = null;
        try
        {
            
            List<Map<String, Object>> results = new ArrayList<>();
            for (String resultId : resultsId)
            {
                ContentConsistencyResult result = _ametysResolver.resolveById(resultId);
                List<Map<String, Object>> columns = _contentConsistencySearchModel.getColumns();
                results.add(_contentConsistencyManager.resultToJSON(result, columns));
            }
            request.setAttribute(ContentConsistencyResultGenerator.RESULTS_REQUEST_ATTRIBUTE_NAME, results);
            
            attachement = _sourceResolver.resolveURI("cocoon://_plugins/cms/consistency/report.csv");
            InputStream attachmentCopy = null;
            // Get the attachment and store it in a reusable inputStream.
            // If the attachment size exceed 1M we won't attach it
            try (InputStream is = attachement.getInputStream())
            {
                // Override the byteArray to avoid duplication of buffer during copy
                ByteArrayOutputStream tmp = new ByteArrayOutputStream()
                {
                    @Override
                    public synchronized byte[] toByteArray()
                    {
                        return buf;
                    }
                };
                IOUtils.copyLarge(is, tmp, 0, 1_000_000);
                if (is.read() == -1)
                {
                    // We could read all the stream so the size is acceptable and we keep the attachment
                    attachmentCopy = new ByteArrayInputStream(tmp.toByteArray());
                }
                
            }
            catch (IOException e)
            {
                getLogger().error("Failed to retrieve consistency report", e);
            }
            return attachmentCopy;
        }
        catch (IOException e)
        {
            getLogger().error("Failed to retrieve report", e);
            return null;
        }
        finally
        {
            _sourceResolver.release(attachement);
            request.removeAttribute(ContentConsistencyResultGenerator.RESULTS_REQUEST_ATTRIBUTE_NAME);
        }
    }

    /**
     * Retrieves the mail's subject
     * @param report the consistency report
     * @return the mail's subject
     */
    protected I18nizableText _getMailSubject(ConsistenciesReport report)
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_GLOBAL_CONTENT_CONSISTENCY_MAIL_SUBJECT");
    }
    
    /**
     * Retrieves the mail's body
     * @param report the consistency report
     * @return the mail's body
     */
    protected String _getMailBody(ConsistenciesReport report)
    {
        try
        {
            MailBodyBuilder mailBuilder = StandardMailBodyHelper.newHTMLBody()
                    .withTitle(_getMailSubject(report));
                
            int nbFailure = report.results() != null ? report.results().size() : 0;
            
            if (nbFailure == 1)
            {
                mailBuilder.addMessage(new I18nizableText("plugin.cms", "PLUGINS_CMS_GLOBAL_CONTENT_CONSISTENCY_MAIL_BODY_FAILURE"));
            }
            else if (nbFailure > 1)
            {
                mailBuilder.addMessage(new I18nizableText("plugin.cms", "PLUGINS_CMS_GLOBAL_CONTENT_CONSISTENCY_MAIL_BODY_FAILURES", List.of(String.valueOf(nbFailure))));
            }
            
            if (!(report.unchecked() == null || report.unchecked().isEmpty()))
            {
                mailBuilder.addMessage(new I18nizableText("plugin.cms", "PLUGINS_CMS_GLOBAL_CONTENT_CONSISTENCY_MAIL_BODY_UNCHECKED", List.of(StringUtils.join(report.unchecked(), ", "))));
            }
            
            mailBuilder.addMessage(new I18nizableText("plugin.cms", "PLUGINS_CMS_GLOBAL_CONTENT_CONSISTENCY_MAIL_BODY_REPORT", List.of(_getReportUri())));
            
            mailBuilder.withLink(_getReportUri(), new I18nizableText("plugin.cms", "PLUGINS_CMS_GLOBAL_CONTENT_CONSISTENCY_MAIL_BODY_REPORT_LINK_TITLE"));

            return mailBuilder.build();
        }
        catch (IOException e)
        {
            getLogger().error("Fail to build HTML content consistency mail report", e);
            return null;
        }
    }
    
    /**
     * Get the report URI
     * @return the report uri
     */
    protected String _getReportUri()
    {
        StringBuilder url = new StringBuilder(_baseUrl);
        url.append("/index.html?uitool=uitool-global-consistency");
        return url.toString();
    }
    
    /**
     * Send the alert e-mails.
     * @param subject the e-mail subject.
     * @param body the e-mail body.
     * @param users users to send the mail to.
     * @param from the address sending the e-mail.
     * @param attachment the e-mail attachment
     */
    protected void _sendMails(String subject, String body, Set<UserIdentity> users, String from, InputStream attachment)
    {
        for (UserIdentity userIdentity : users)
        {
            User user = _userManager.getUser(userIdentity);
            
            if (user != null && StringUtils.isNotBlank(user.getEmail()))
            {
                String mail = user.getEmail();
                
                Collection<NamedStream> attachments = null;
                if (attachment != null)
                {
                    try
                    {
                        attachment.reset();
                    }
                    catch (IOException e)
                    {
                        // in case reset is not implemented.
                    }
                    attachments = List.of(new NamedStream(attachment, "report.csv", "text/csv"));
                }
                
                try
                {
                    SendMailHelper.newMail()
                                  .withSubject(subject)
                                  .withHTMLBody(body)
                                  .withSender(from)
                                  .withRecipient(mail)
                                  .withAttachmentsAsStream(attachments)
                                  .sendMail();
                }
                catch (MessagingException | IOException e)
                {
                    if (getLogger().isWarnEnabled())
                    {
                        getLogger().warn("Could not send an alert e-mail to " + mail, e);
                    }
                }
            }
        }
    }
}
