/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.consistency;

import java.util.Collection;
import java.util.List;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;

import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.data.type.ModelItemTypeExtensionPoint;
import org.ametys.runtime.model.DefaultElementDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.UnknownTypeException;
import org.ametys.runtime.model.type.ModelItemTypeConstants;

/** 
 * Model of a consistency check.
 */
public class ContentConsistencyModel implements Model, Component
{
    /** The avalon role */
    public static final String ROLE = ContentConsistencyModel.class.getName();
    
    /** data path for the server error*/
    public static final String SERVER_ERROR = "serverError";
    /** data path for the not found error */
    public static final String NOT_FOUND = "notFound";
    /** data path for the unauthorized error */
    public static final String UNAUTHORIZED = "unauthorized";
    /** data path for the unknown error */
    public static final String UNKNOWN = "unknown";
    /** data path for the success checks */
    public static final String SUCCESS = "success";
    /** data path for the date of the check */
    public static final String DATE = "date";
    /** data path for the workflow step */
    public static final String WORKFLOW_STEP = "workflowStep";
    /** data path for the content types */
    public static final String CONTENT_TYPES = "contentTypes";
    /** data path for the title */
    public static final String TITLE = "title";
    /** data path for the content id */
    public static final String CONTENT_ID = "contentId";
    /** data path for the creator */
    public static final String CREATOR = "creator";
    /** data path for the last major validator */
    public static final String LAST_MAJOR_VALIDATOR = "lastMajorValidator";
    /** data path for the last validator */
    public static final String LAST_VALIDATOR = "lastValidator";
    /** data path for the last contributor */
    public static final String LAST_CONTRIBUTOR = "contributor";
    /** data path for the context */
    public static final String CONTEXT = "context";
    /** data path for the creation date */
    public static final String CREATION_DATE = "creationDate";
    /** data path for the content last major validation date */
    public static final String LAST_MAJOR_VALIDATION_DATE = "lastMajorValidationDate";
    /** data path for the content last validation date */
    public static final String LAST_VALIDATION_DATE = "lastValidationDate";
    /** data path for the content last modification date */
    public static final String LAST_MODIFICATION_DATE = "lastModified";
    
    private Collection<? extends ModelItem> _modelItems;

    public String getId()
    {
        return "ContentConsistency";
    }

    public String getFamilyId()
    {
        return "ContentConsistency";
    }

    public Collection< ? extends ModelItem> getModelItems()
    {
        if (_modelItems == null)
        {
            try
            {
                _modelItems = List.of(
                        DefaultElementDefinition.of(CONTENT_ID, false, ModelItemTypeConstants.STRING_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(CONTEXT, false, ModelItemTypeConstants.STRING_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(TITLE, false, ModelItemTypeConstants.STRING_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(CONTENT_TYPES, true, ModelItemTypeConstants.STRING_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(CREATOR, false, org.ametys.cms.data.type.ModelItemTypeConstants.USER_ELEMENT_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(CREATION_DATE, false, ModelItemTypeConstants.DATETIME_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(LAST_MAJOR_VALIDATOR, false, org.ametys.cms.data.type.ModelItemTypeConstants.USER_ELEMENT_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(LAST_MAJOR_VALIDATION_DATE, false, ModelItemTypeConstants.DATETIME_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(LAST_VALIDATOR, false, org.ametys.cms.data.type.ModelItemTypeConstants.USER_ELEMENT_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(LAST_VALIDATION_DATE, false, ModelItemTypeConstants.DATETIME_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(LAST_CONTRIBUTOR, false, org.ametys.cms.data.type.ModelItemTypeConstants.USER_ELEMENT_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(LAST_MODIFICATION_DATE, false, ModelItemTypeConstants.DATETIME_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(WORKFLOW_STEP, false, ModelItemTypeConstants.LONG_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(DATE, false, ModelItemTypeConstants.DATETIME_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(SUCCESS, false, ModelItemTypeConstants.LONG_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(UNKNOWN, false, ModelItemTypeConstants.LONG_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(UNAUTHORIZED, false, ModelItemTypeConstants.LONG_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(NOT_FOUND, false, ModelItemTypeConstants.LONG_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC),
                        DefaultElementDefinition.of(SERVER_ERROR, false, ModelItemTypeConstants.LONG_TYPE_ID, ModelItemTypeExtensionPoint.ROLE_MODEL_AWARE_BASIC)
                );
            }
            catch (UnknownTypeException | BadItemTypeException | ServiceException e)
            {
                throw new AmetysRepositoryException("An error occurred while create the content consistency result model", e);
            }
        }
        return _modelItems;
    }
}
