/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.consistency;

import java.time.ZonedDateTime;
import java.util.Optional;

import javax.jcr.Node;

import org.ametys.cms.data.holder.impl.DefaultModifiableModelAwareDataHolder;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.ametysobject.ModifiableModelAwareDataAwareAmetysObject;
import org.ametys.plugins.repository.data.holder.ModifiableModelAwareDataHolder;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.jcr.SimpleAmetysObject;

/** 
 * Simple object representing the result of a consistency check.
 */
public class ContentConsistencyResult extends SimpleAmetysObject<ContentConsistencyResultFactory> implements ModifiableModelAwareDataAwareAmetysObject
{
    /** JCR node type for content consistency result */
    public static final String CONTENT_CONSISTENCY_RESULT_NODETYPE = RepositoryConstants.NAMESPACE_PREFIX + ":consistencyResult";
    
    /**
     * Creates an {@link ContentConsistencyResult}.
     * @param node the node backing this result
     * @param parentPath the parentPath in the Ametys hierarchy
     * @param factory the factory which created the activity
     */
    public ContentConsistencyResult(Node node, String parentPath, ContentConsistencyResultFactory factory)
    {
        super(node, parentPath, factory);
    }
    
    /**
     * Get the content id
     * @return the content Id
     */
    public String getContentId()
    {
        return getValue(ContentConsistencyModel.CONTENT_ID);
    }

    /**
     * Get the result date
     * @return the date
     */
    public ZonedDateTime getDate()
    {
        return getValue(ContentConsistencyModel.DATE);
    }

    /**
     * Get the unauthorized count
     * @return the count
     */
    public long getUnauthorizedCount()
    {
        return getValue(ContentConsistencyModel.UNAUTHORIZED);
    }

    /**
     * Get the unknown count
     * @return the unknown count
     */
    public long getUnknownCount()
    {
        return getValue(ContentConsistencyModel.UNKNOWN);
    }

    /**
     * Get the success count
     * @return the success count
     */
    public long getSuccessCount()
    {
        return getValue(ContentConsistencyModel.SUCCESS);
    }

    /**
     * Get the not found count
     * @return the count
     */
    public long getNotFoundCount()
    {
        return getValue(ContentConsistencyModel.NOT_FOUND);
    }

    /**
     * Get the server error count
     * @return the count
     */
    public long getServerErrorCount()
    {
        return getValue(ContentConsistencyModel.SERVER_ERROR);
    }

    /**
     * Get the title
     * @return the title
     */
    public String getTitle()
    {
        return getValue(ContentConsistencyModel.TITLE);
    }

    /**
     * Get the content types
     * @return the content types
     */
    public String[] getContentTypes()
    {
        return getValue(ContentConsistencyModel.CONTENT_TYPES);
    }

    /**
     * Get the creator
     * @return the creator
     */
    public UserIdentity getCreator()
    {
        return getValue(ContentConsistencyModel.CREATOR);
    }
    
    /**
     * Get the creation date
     * @return the content creation date
     */
    public ZonedDateTime getCreationDate()
    {
        return getValue(ContentConsistencyModel.CREATION_DATE);
    }

    /**
     * Get the last major validator
     * @return the last major validator
     */
    public Optional<UserIdentity> getLastMajorValidator()
    {
        return Optional.ofNullable(getValue(ContentConsistencyModel.LAST_MAJOR_VALIDATOR));
    }
    
    /**
     * Get the last major validation date
     * @return the last content major validation date
     */
    public ZonedDateTime getLastMajorValidationDate()
    {
        return getValue(ContentConsistencyModel.LAST_MAJOR_VALIDATION_DATE);
    }

    /**
     * Get the last validator
     * @return the last validator
     */
    public Optional<UserIdentity> getLastValidator()
    {
        return Optional.ofNullable(getValue(ContentConsistencyModel.LAST_VALIDATOR));
    }
    
    /**
     * Get the last validation date
     * @return the last content validation date
     */
    public ZonedDateTime getLastValidationDate()
    {
        return getValue(ContentConsistencyModel.LAST_VALIDATION_DATE);
    }
    
    /**
     * Get the last contributor
     * @return the last contributor
     */
    public UserIdentity getLastContributor()
    {
        return getValue(ContentConsistencyModel.LAST_CONTRIBUTOR);
    }
    
    /**
     * Get the last modification date
     * @return the content last modification date
     */
    public ZonedDateTime getLastModificationDate()
    {
        return getValue(ContentConsistencyModel.LAST_MODIFICATION_DATE);
    }
    
    /**
     * Get the workflow step
     * @return the workflow step
     */
    public long getWorkflowStep()
    {
        return getValue(ContentConsistencyModel.WORKFLOW_STEP);
    }

    public ModifiableModelAwareDataHolder getDataHolder()
    {
        ModifiableRepositoryData repositoryData = new JCRRepositoryData(getNode());
        return new DefaultModifiableModelAwareDataHolder(repositoryData, Optional.empty(), Optional.empty(), _getFactory().getContentConsistencyModel());
    }
}
