/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.consistency;

import java.io.IOException;
import java.time.ZonedDateTime;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.generation.AbstractGenerator;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.cms.repository.Content;
import org.ametys.core.util.DateUtils;
import org.ametys.core.util.ServerCommHelper;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Retrieve all content consistency result stored and sax them
 */
public class ContentConsistencyResultGenerator extends AbstractGenerator implements Serviceable
{
    /** the name of the attribute use to recover the list of result */
    public static final String RESULTS_REQUEST_ATTRIBUTE_NAME = "contentConsistencyResult";
    
    /** The ametys object resolver */
    protected AmetysObjectResolver _resolver;
    private ServerCommHelper _serverCommHelper;
    private ContentConsistencySearchModel _searchModel;

    public void service(ServiceManager manager) throws ServiceException
    {
        _searchModel = (ContentConsistencySearchModel) manager.lookup(ContentConsistencySearchModel.ROLE);
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _serverCommHelper = (ServerCommHelper) manager.lookup(ServerCommHelper.ROLE);
    }
    
    public void generate() throws IOException, SAXException, ProcessingException
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        Map<String, Object> jsParameters = _serverCommHelper.getJsParameters();
        
        
        contentHandler.startDocument();
        XMLUtils.startElement(contentHandler, "export");
        
        XMLUtils.startElement(contentHandler, "columns");
        
        Map<String, Object> model = _searchModel.getModel();
        @SuppressWarnings("unchecked")
        List<Map<String, Object>> modelColumns = (List<Map<String, Object>>) model.get("columns");
        @SuppressWarnings("unchecked")
        List<String> columns = jsParameters != null ? (List<String>) jsParameters.get("columns") : List.of();
        for (Map<String, Object> column : modelColumns)
        {
            if (columns.isEmpty() || columns.contains(column.get("path")))
            {
                AttributesImpl attrs = new AttributesImpl();
                attrs.addCDATAAttribute("id", (String) column.get("path"));
                attrs.addCDATAAttribute("type", (String) column.get("type"));
                XMLUtils.startElement(contentHandler, "column", attrs);
                Object label = column.get("label");
                if (label instanceof I18nizableText i18n)
                {
                    i18n.toSAX(contentHandler);
                }
                else
                {
                    XMLUtils.data(contentHandler, label.toString());
                    
                }
                XMLUtils.endElement(contentHandler, "column");
            }
        }
        XMLUtils.endElement(contentHandler, "columns");
        
        XMLUtils.startElement(contentHandler, "contents");
        
        @SuppressWarnings("unchecked")
        List<Map<String, Object>> results = (List<Map<String, Object>>) request.getAttribute(RESULTS_REQUEST_ATTRIBUTE_NAME);
        if (results != null)
        {
            for (Map<String, Object> result : results)
            {
                _saxContentConsistency(result, columns);
            }
        }
        
        XMLUtils.endElement(contentHandler, "contents");
        XMLUtils.endElement(contentHandler, "export");
        contentHandler.endDocument();
    }

    /**
     * Generate information on content consistency. 
     * @param result the content consistency checker result
     * @param columns the names of the columns to SAX
     * @throws SAXException if an errors occurs generating the data.
     */
    protected void _saxContentConsistency(Map<String, Object> result, List<String> columns) throws SAXException
    {
        XMLUtils.startElement(contentHandler, "content");
        // if there is no columns, we use all the keys
        for (String column : !columns.isEmpty() ? columns : result.keySet())
        {
            AttributesImpl attrs = new AttributesImpl();
            attrs.addCDATAAttribute("fullModelPath", column);
            Object value = result.get(column);
            if (value instanceof String str)
            {
                XMLUtils.createElement(contentHandler, column, attrs, str);
            }
            else if (value instanceof Long count)
            {
                XMLUtils.createElement(contentHandler, column, attrs, count.toString());
            }
            else if (value instanceof ZonedDateTime date)
            {
                XMLUtils.createElement(contentHandler, column, attrs, DateUtils.zonedDateTimeToString(date));
            }
            else if (StringUtils.equals(column, "contributor"))
            {
                @SuppressWarnings("unchecked")
                Map<String, Object> mapValue = (Map<String, Object>) value;
                String fullName = (String) mapValue.get("fullname");
                if (StringUtils.isNotBlank(fullName))
                {
                    XMLUtils.createElement(contentHandler, column, attrs, fullName);
                }
                else
                {
                    String login = (String) mapValue.get("login");
                    String population = (String) mapValue.get("population");
                    if (StringUtils.isNoneBlank(login, population))
                    {
                        XMLUtils.createElement(contentHandler, column, attrs, login + "#" + population);
                    }
                    else
                    {
                        XMLUtils.createElement(contentHandler, column, attrs);
                    }
                }
            }
            else if (StringUtils.equals(column, "workflowStep"))
            {
                @SuppressWarnings("unchecked")
                Map<String, Object> mapValue = (Map<String, Object>) value;
                XMLUtils.startElement(contentHandler, column, attrs);
                I18nizableText label = (I18nizableText) mapValue.get("name");
                if (label != null)
                {
                    label.toSAX(contentHandler);
                }
                XMLUtils.endElement(contentHandler, column);
            }
            else if (StringUtils.equals(column, "contentTypes"))
            {
                @SuppressWarnings("unchecked")
                List<Map<String, Object>> jsonValues = (List<Map<String, Object>>) value;
                List<I18nizableText> labels = jsonValues.stream()
                    .map(map -> (I18nizableText) map.get("label"))
                    .filter(Objects::nonNull)
                    .toList();
                for (I18nizableText label : labels)
                {
                    XMLUtils.startElement(contentHandler, column, attrs);
                    label.toSAX(contentHandler);
                    XMLUtils.endElement(contentHandler, column);
                }
            }
        }
        XMLUtils.endElement(contentHandler, "content");
    }
    
    /**
     * Sax additional data on the content 
     * @param content the content.
     * @param atts the attributes the will be saxed
     */
    protected void _saxAdditionalContentAttributes (Content content, AttributesImpl atts)
    {
        // Nothing to do
    }
}
