/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.consistency;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.ProcessingException;

import org.ametys.cms.search.ui.model.SearchUIColumn;
import org.ametys.cms.search.ui.model.SearchUIColumnHelper;
import org.ametys.core.ui.Callable;
import org.ametys.plugins.workflow.support.WorkflowHelper;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.DefinitionContext;
import org.ametys.runtime.model.ModelItem;

import com.opensymphony.workflow.loader.StepDescriptor;

/**
 * Provide the search model for Content consistency search
 */
public class ContentConsistencySearchModel implements Component, Serviceable
{
    /** The avalon role */
    public static final String ROLE = ContentConsistencySearchModel.class.getName();
    
    private WorkflowHelper _workflowHelper;
    private ContentConsistencyModel _contentConsistencyModel;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _workflowHelper = (WorkflowHelper) manager.lookup(WorkflowHelper.ROLE);
        _contentConsistencyModel = (ContentConsistencyModel) manager.lookup(ContentConsistencyModel.ROLE);
    }
    
    /**
     * Get the model for a content consistency search
     * @return the model
     * @throws ProcessingException if an error occurs processing the model
     */
    @Callable
    public Map<String, Object> getModel() throws ProcessingException
    {
        Map<String, Object> model = new HashMap<>();
        
        // Proxy configuration
        model.put("searchMethodName", "searchResults");
        model.put("searchRole", "org.ametys.cms.content.consistency.ContentConsistencySearcher");
        
        model.put("columns", getColumns());
        
        model.put("criteria", getCriteria());
        
        model.put("facets", getFacets());
        
        return model;
    }
    
    /**
     * Provide the list of column.
     * Any column additional column must not be sortable as there info won't be present in the search result.
     * @return a json description of the column
     * @throws ProcessingException if an error occurs processing the columns
     */
    protected List<Map<String, Object>> getColumns() throws ProcessingException
    {
        List<SearchUIColumn> columns = new ArrayList<>();

        columns.add(getSearchUIColumn(ContentConsistencyModel.TITLE,
                                        new I18nizableText("plugin.cms", "PLUGINS_CMS_METADATA_TITLE_LABEL"),
                                        Optional.empty(),
                                        Optional.empty(),
                                        false));
        
        columns.add(getSearchUIColumn(ContentConsistencyModel.CONTENT_TYPES,
                                        new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_ALL_CONTENT_TYPE"),
                                        Optional.of("Ametys.plugins.cms.content.tool.GlobalContentConsistencyTool.renderContentType"),
                                        Optional.empty(),
                                        false));
        
        columns.add(getSearchUIColumn(ContentConsistencyModel.CREATOR,
                                        new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_CREATOR"),
                                        Optional.of("Ametys.plugins.cms.search.SearchGridHelper.renderUser"),
                                        Optional.empty(),
                                        true));
        
        columns.add(getSearchUIColumn(ContentConsistencyModel.CREATION_DATE,
                                        new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_CREATIONDATE"),
                                        Optional.empty(),
                                        Optional.empty(),
                                        true));
        
        columns.add(getSearchUIColumn(ContentConsistencyModel.LAST_CONTRIBUTOR,
                                        new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_AUTHOR"),
                                        Optional.of("Ametys.plugins.cms.search.SearchGridHelper.renderUser"),
                                        Optional.empty(),
                                        false));
        
        columns.add(getSearchUIColumn(ContentConsistencyModel.LAST_MODIFICATION_DATE,
                                        new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_LASTMODIFIED"),
                                        Optional.empty(),
                                        Optional.empty(),
                                        true));
        
        columns.add(getSearchUIColumn(ContentConsistencyModel.LAST_VALIDATOR,
                                        new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_LAST_VALIDATOR"),
                                        Optional.of("Ametys.plugins.cms.search.SearchGridHelper.renderUser"),
                                        Optional.empty(),
                                        true));
        
        columns.add(getSearchUIColumn(ContentConsistencyModel.LAST_VALIDATION_DATE,
                                        new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_LASTVALIDATION"),
                                        Optional.empty(),
                                        Optional.empty(),
                                        true));
        
        columns.add(getSearchUIColumn(ContentConsistencyModel.WORKFLOW_STEP,
                                        new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_WORKFLOW_STEP"),
                                        Optional.of("Ametys.plugins.cms.search.SearchGridHelper.renderWorkflowStep"),
                                        Optional.of("Ametys.plugins.cms.search.SearchGridHelper.convertWorkflowStep"),
                                        false));
        
        columns.add(getSearchUIColumn(ContentConsistencyModel.DATE,
                                        new I18nizableText("plugin.cms", "PLUGINS_CMS_GLOBAL_CONTENT_CONSISTENCY_CHECK_DATE"),
                                        Optional.empty(),
                                        Optional.empty(),
                                        false));
        
        columns.add(getSearchUIColumn(ContentConsistencyModel.SUCCESS,
                                        new I18nizableText("plugin.cms", "PLUGINS_CMS_GLOBAL_CONTENT_CONSISTENCY_SUCCESS_COUNT"),
                                        Optional.empty(),
                                        Optional.empty(),
                                        true));
        
        columns.add(getSearchUIColumn(ContentConsistencyModel.UNAUTHORIZED,
                                        new I18nizableText("plugin.cms", "PLUGINS_CMS_GLOBAL_CONTENT_CONSISTENCY_UNAUTHORIZED_COUNT"),
                                        Optional.empty(),
                                        Optional.empty(),
                                        false));
        
        columns.add(getSearchUIColumn(ContentConsistencyModel.UNKNOWN,
                                        new I18nizableText("plugin.cms", "PLUGINS_CMS_GLOBAL_CONTENT_CONSISTENCY_UNKNOWN_COUNT"),
                                        Optional.empty(),
                                        Optional.empty(),
                                        false));
        
        columns.add(getSearchUIColumn(ContentConsistencyModel.NOT_FOUND,
                                        new I18nizableText("plugin.cms", "PLUGINS_CMS_GLOBAL_CONTENT_CONSISTENCY_NOT_FOUND_COUNT"),
                                        Optional.empty(),
                                        Optional.empty(),
                                        false));
        
        columns.add(getSearchUIColumn(ContentConsistencyModel.SERVER_ERROR,
                                        new I18nizableText("plugin.cms", "PLUGINS_CMS_GLOBAL_CONTENT_CONSISTENCY_SERVER_ERROR_COUNT"),
                                        Optional.empty(),
                                        Optional.empty(),
                                        false));

        List<Map<String, Object>>  columnsInfo = new ArrayList<>();
        DefinitionContext definitionContext = DefinitionContext.newInstance();
        for (SearchUIColumn column : columns)
        {
            columnsInfo.add(column.toJSON(definitionContext));
        }
        return columnsInfo;
    }
    
    /**
     * Retrieves the column corresponding to the given item name
     * @param name the name of the column's item
     * @param label the column's label
     * @param renderer the column's renderer
     * @param converter  the column's converter
     * @param hidden <code>true</code> if the column should be hidden, <code>false</code> otherwise
     * @return the search column's info
     */
    protected SearchUIColumn getSearchUIColumn(String name, I18nizableText label, Optional<String> renderer, Optional<String> converter, boolean hidden)
    {
        ModelItem modelItem = _contentConsistencyModel.getModelItem(name);
        SearchUIColumn<? extends ModelItem> column = SearchUIColumnHelper.createModelItemColumn(modelItem);

        column.setLabel(label);
        column.setRenderer(renderer);
        column.setConverter(converter);
        column.setHidden(hidden);
        
        return column;
    }
    
    /**
     * Get the criteria for the search
     * @return the JSON description of the criteria
     */
    protected Map<String, Object> getCriteria()
    {
        Map<String, Object> criteria = new LinkedHashMap<>(); // order is important here
        criteria.put("title", Map.of(
                "label", new I18nizableText("plugin.cms", "PLUGINS_CMS_METADATA_TITLE_LABEL"),
                "description", new I18nizableText("plugin.cms", "PLUGINS_CMS_METADATA_TITLE_DESCRIPTION"),
                "type", "STRING"
                ));
        
        List<Map<String, Object>> workflowStep = new ArrayList<>();
        workflowStep.add(Map.of("label", new I18nizableText("plugin.cms", "WIDGET_COMBOBOX_ALL_OPTIONS"), "value", 0));
        for (Iterator iterator = _workflowHelper.getWorkflowDescriptor("content").getSteps().iterator(); iterator.hasNext();)
        {
            StepDescriptor step = (StepDescriptor) iterator.next();
            workflowStep.add(Map.of("label", new I18nizableText(null, step.getName()), "value", step.getId()));
        }

        criteria.put("workflowStep", Map.of(
                "label", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_WORKFLOW_STEP"),
                "description", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_WORKFLOW_STEP"),
                "type", "LONG",
                "enumeration", workflowStep,
                "default-value", 0 // - Tous -
                ));
        
        Map<String, Object> ctype = new HashMap<>();
        ctype.put("label", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_ALL_CONTENT_TYPE"));
        ctype.put("description", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_ALL_CONTENT_TYPE"));
        ctype.put("type", "STRING");
        ctype.put("multiple", false);
        ctype.put("enumeration", null);
        ctype.put("widget", "edition.select-content-types");
        ctype.put("widget-params", Map.of("emptyText", new I18nizableText("plugin.cms", "WIDGET_COMBOBOX_ALL_OPTIONS")));
        criteria.put("contentTypes", ctype);
        
        criteria.put(ContentConsistencyModel.CREATOR, Map.of(
                "label", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_CREATOR"),
                "description", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_CREATOR"),
                "type", "USER"
                ));
        criteria.put("contributor", Map.of(
                "label", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_AUTHOR"),
                "description", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_AUTHOR"),
                "type", "USER"
                ));
        criteria.put(ContentConsistencyModel.LAST_VALIDATOR, Map.of(
                "label", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_LAST_VALIDATOR"),
                "description", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_LAST_VALIDATOR"),
                "type", "USER"
                ));
        return criteria;
    }
    
    /**
     * Get the facets for the search
     * @return the JSON description of the facets
     */
    protected List<Map<String, Object>> getFacets()
    {
        List<Map<String, Object>> facets = new ArrayList<>();
        facets.add(Map.of(
                "name", ContentConsistencyModel.CREATOR,
                "label", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_CREATOR"),
                "type", "criterion",
                "children", new ArrayList()));
        facets.add(Map.of(
                "name", ContentConsistencyModel.LAST_CONTRIBUTOR,
                "label", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_AUTHOR"),
                "type", "criterion",
                "children", new ArrayList())); // needs to be mutable to add the value later
        facets.add(Map.of(
                "name", ContentConsistencyModel.LAST_VALIDATOR,
                "label", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_LAST_VALIDATOR"),
                "type", "criterion",
                "children", new ArrayList())); // needs to be mutable to add the value later
        facets.add(Map.of(
                "name", ContentConsistencyModel.WORKFLOW_STEP,
                "label", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_WORKFLOW_STEP"),
                "type", "criterion",
                "children", new ArrayList())); // needs to be mutable to add the value later
        facets.add(Map.of(
                "name", ContentConsistencyModel.CONTENT_TYPES,
                "label", new I18nizableText("plugin.cms", "UITOOL_SEARCH_CONTENT_ALL_CONTENT_TYPE"),
                "type", "criterion",
                "children", new ArrayList())); // needs to be mutable to add the value later
        return facets;
    }
}
