/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.consistency;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.ProcessingException;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.core.ui.Callable;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.JSONUtils;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.plugins.repository.query.QueryHelper;
import org.ametys.plugins.repository.query.SortCriteria;
import org.ametys.plugins.repository.query.expression.AndExpression;
import org.ametys.plugins.repository.query.expression.Expression;
import org.ametys.plugins.repository.query.expression.Expression.Operator;
import org.ametys.plugins.repository.query.expression.LongExpression;
import org.ametys.plugins.repository.query.expression.OrExpression;
import org.ametys.plugins.repository.query.expression.StringExpression;
import org.ametys.plugins.repository.query.expression.UserExpression;
import org.ametys.plugins.workflow.support.WorkflowHelper;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

import com.opensymphony.workflow.loader.StepDescriptor;
import com.opensymphony.workflow.loader.WorkflowDescriptor;

/**
 * Execute JCR query to search for content consistency result
 */
public class ContentConsistencySearcher extends AbstractLogEnabled implements Serviceable, Component
{
    /** right id to access global consistency tool */
    public static final String CMS_RIGHTS_TOOLS_GLOBAL_CONSISTENCY = "CMS_Rights_Tools_GlobalConsistency";

    /** the avalon role */
    public static final String ROLE = ContentConsistencySearcher.class.getName();
    
    private ContentTypeExtensionPoint _cTypeEP;
    private JSONUtils _jsonUtils;
    private AmetysObjectResolver _resolver;
    private ContentConsistencySearchModel _searchModel;
    private UserHelper _userHelper;
    private WorkflowHelper _workflowHelper;

    private ContentConsistencyManager _contentConsistencyManager;

    public void service(ServiceManager manager) throws ServiceException
    {
        _cTypeEP = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _jsonUtils = (JSONUtils) manager.lookup(JSONUtils.ROLE);
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _searchModel = (ContentConsistencySearchModel) manager.lookup(ContentConsistencySearchModel.ROLE);
        _userHelper = (UserHelper) manager.lookup(UserHelper.ROLE);
        _workflowHelper = (WorkflowHelper) manager.lookup(WorkflowHelper.ROLE);
        _contentConsistencyManager = (ContentConsistencyManager) manager.lookup(ContentConsistencyManager.ROLE);
    }
    
    /**
     * Execute a search based on the provided parameters.
     * Only results with failure will be searched
     * 
     * Parameters must include :
     * <ul>
     * <li><code>start</code> and <code>limit</code> for pagination</li>
     * <li><code>sort</code> for sort criteria definition</li>
     * <li><code>values</code> for the criteria definition</li>
     * </ul>
     * 
     * @param jsonParams the json params
     * @return json representation of the results based on {@link ContentConsistencySearchModel}
     * @throws ProcessingException if an error occurs while processing search model
     */
    @Callable(rights = CMS_RIGHTS_TOOLS_GLOBAL_CONSISTENCY)
    public Map<String, Object> searchResults(Map<String, Object> jsonParams) throws ProcessingException
    {
        int offset = (int) jsonParams.getOrDefault("start", 0);
        int limit = (int) jsonParams.getOrDefault("limit", Integer.MAX_VALUE);
        List<Object> sorters = _jsonUtils.convertJsonToList((String) jsonParams.getOrDefault("sort", "[]"));
        SortCriteria sortCriteria = _getSortCriteria(sorters);
        
        @SuppressWarnings("unchecked")
        Map<String, Object> criteria = (Map<String, Object>) jsonParams.getOrDefault("values", Map.of());
        List<Expression> criteriaExpressions = _getCriteriaExpressions(criteria);
        
        @SuppressWarnings("unchecked")
        Map<String, Object> facetValues = (Map<String, Object>) jsonParams.getOrDefault("facetValues", Map.of());
        criteriaExpressions.addAll(_getCriteriaExpressions(facetValues));
        
        Expression expression = getExpression(criteriaExpressions);
        String xPathQuery = QueryHelper.getXPathQuery(null, "ametys:consistencyResult", expression, sortCriteria);
        try (AmetysObjectIterable<ContentConsistencyResult> results = _resolver.query(xPathQuery))
        {
            return buildSearchResults(results, offset, limit);
        }
    }
    
    /**
     * Get the final expression based on the list of criteria expression
     * @param criteriaExpressions a list of expressions
     * @return an expression or null if the list is empty
     */
    protected Expression getExpression(List<Expression> criteriaExpressions)
    {
        return criteriaExpressions.isEmpty() ? null : new AndExpression(criteriaExpressions.toArray(new Expression[0]));
    }

    private List<Expression> _getCriteriaExpressions(Map<String, Object> criteria)
    {
        List<Expression> expressions = new ArrayList<>();
        for (Entry<String, Object> criterion : criteria.entrySet())
        {
            String dataPath = criterion.getKey();
            Object value = criterion.getValue();
            if (value instanceof List list)
            {
                List<Expression> orExpressions = list.stream()
                    .map(v -> _getCriterionExpression(dataPath, v))
                    .filter(Objects::nonNull)
                    .toList();
                if (!orExpressions.isEmpty())
                {
                    expressions.add(new OrExpression(orExpressions.toArray(new Expression[0])));
                }
            }
            else
            {
                Expression criterionExpression = _getCriterionExpression(dataPath, value);
                if (criterionExpression != null)
                {
                    expressions.add(criterionExpression);
                }
            }
        }
        return expressions;
    }

    private Expression _getCriterionExpression(String dataPath, Object value)
    {
        Expression expression = null;
        switch (dataPath)
        {
            case "workflowStep":
                Integer step = (Integer) value;
                if (step != null && step != 0)
                {
                    expression = new LongExpression(dataPath, Operator.EQ, step);
                }
                break;
            case "contentTypes":
                String str = (String) value;
                if (StringUtils.isNotBlank(str))
                {
                    expression = new StringExpression(dataPath, Operator.EQ, str);
                }
                break;
            case "contributor":
            case ContentConsistencyModel.CREATOR:
            case ContentConsistencyModel.LAST_VALIDATOR:
            case ContentConsistencyModel.LAST_MAJOR_VALIDATOR:
                if (value instanceof String strValue)
                {
                    expression = new UserExpression(dataPath, Operator.EQ, UserIdentity.stringToUserIdentity(strValue));
                }
                else
                {
                    @SuppressWarnings("unchecked") Map<String, Object> json = (Map<String, Object>) value;
                    if (json != null)
                    {
                        expression = new UserExpression(dataPath, Operator.EQ, _userHelper.json2userIdentity(json));
                    }
                }
                break;
            case "title":
                str = (String) value;
                if (StringUtils.isNotBlank(str))
                {
                    expression = new StringExpression(dataPath, Operator.WD, str);
                }
                break;
            default :
                throw new UnsupportedOperationException("datapath " + dataPath + " is not a supported criterion.");
        }
        return expression;
    }

    private SortCriteria _getSortCriteria(List<Object> sorters)
    {
        SortCriteria sortCriteria = new SortCriteria();
        for (Object sorter : sorters)
        {
            if (sorter instanceof Map sorterMap)
            {
                sortCriteria.addCriterion((String) sorterMap.get("property"), StringUtils.equals("ASC", (String) sorterMap.get("direction")), false);
            }
        }
        return sortCriteria;
    }
    
    private Map<String, Object> buildSearchResults(AmetysObjectIterable<ContentConsistencyResult> results, int offset, int limit) throws ProcessingException
    {
        ArrayList<Map<String, Object>> searchResults = new ArrayList<>((int) results.getSize());
        Map<String, Object> model = _searchModel.getModel();
        @SuppressWarnings("unchecked")
        List<Map<String, Object>> columns = (List<Map<String, Object>>) model.get("columns");
        @SuppressWarnings("unchecked")
        List<Map<String, Object>> facets = (List<Map<String, Object>>) model.get("facets");
        Map<String, Map<Object, Map<String, Object>>> computedFacets = new HashMap<>();
        
        int resultIdx = -1;
        for (ContentConsistencyResult result : results)
        {
            try
            {
                // add the results if they are inside the pagination interval
                if (offset <= ++resultIdx && resultIdx < offset + limit) // increment index before anything to ensure that the value is consistent inside the loop
                {
                    searchResults.add(_contentConsistencyManager.resultToJSON(result, columns));
                }
                
                // Compute facets
                for (Map<String, Object> facet : facets)
                {
                    String facetId = (String) facet.get("name");
                    Map<Object, Map<String, Object>> computedFacet = computedFacets.computeIfAbsent(facetId, __ -> new HashMap<>());
                    switch (facetId)
                    {
                        case ContentConsistencyModel.CONTENT_TYPES:
                            _updateContentTypesFacet(facetId, computedFacet, result);
                            break;
                        case ContentConsistencyModel.CREATOR :
                        case ContentConsistencyModel.LAST_VALIDATOR :
                        case ContentConsistencyModel.LAST_MAJOR_VALIDATOR :
                        case ContentConsistencyModel.LAST_CONTRIBUTOR :
                            _updateUserFacet(facetId, computedFacet, result);
                            break;
                        case ContentConsistencyModel.WORKFLOW_STEP :
                            _updateWorkflowStepFacet(facetId, computedFacet, result);
                            break;
                        default :
                            throw new UnsupportedOperationException("facet '" + facetId + "' is not a supported facets");
                    }
                }
            }
            catch (UnknownAmetysObjectException e)
            {
                getLogger().info("A consistency result was describing the result of the inexisting content '{}' and was ignored", result.getContentId(), e);
                // decrease the number of included content as the result was not actually included
                resultIdx--;
            }
        }
        
        // inject computed value in facets
        for (Map<String, Object> facet : facets)
        {
            @SuppressWarnings("unchecked")
            List<Map<String, Object>> facetValues = (List<Map<String, Object>>) facet.get("children");
            Map<Object, Map<String, Object>> facetValue = computedFacets.get(facet.get("name"));
            // will be the case when there is no results
            if (facetValue != null)
            {
                facetValues.addAll(facetValue.values());
            }
        }
        
        return  Map.of(
                "consistencyResults", searchResults,
                "facets", facets,
                "total", resultIdx + 1); // use result index here to take into account ignored results
    }
    
    private void _updateWorkflowStepFacet(String facetId, Map<Object, Map<String, Object>> computedFacet, ContentConsistencyResult result)
    {
        Long value = result.getValue(facetId);
        if (value != null)
        {
            computedFacet.compute(value, this::_incrementWorkflowStepFacetValue);
        }
    }
    
    private Map<String, Object> _incrementWorkflowStepFacetValue(Object value, Map<String, Object> existingFacetValue)
    {
        if (existingFacetValue == null)
        {
            // Create a new facet value and return it
            Map<String, Object> newValue = new HashMap<>();
            newValue.put("value", value);
            Long stepId = (Long) value;
            
            WorkflowDescriptor defaultWorkflow = _workflowHelper.getWorkflowDescriptor("content");
            // Use the default 'content' workflow to retrieve the label.
            // We can only have one label for a step id. So we try to take it from content.
            // Even if the content might actually use a different workflow.
            // If the step is not available in the 'content' workflow, use the value as a fallback
            StepDescriptor step = defaultWorkflow != null ? defaultWorkflow.getStep(stepId.intValue()) : null;
            if (step != null)
            {
                newValue.put("label", new I18nizableText(null, step.getName()));
            }
            else
            {
                newValue.put("label", value);
            }

            newValue.put("count", 1L);
            newValue.put("type", "facet");
            return newValue;
        }
        else
        {
            existingFacetValue.compute("count", (k, v) -> ((Long) v) + 1);
            return existingFacetValue;
        }
    }
    
    private void _updateContentTypesFacet(String facetId, Map<Object, Map<String, Object>> computedFacet, ContentConsistencyResult result)
    {
        String[] typeIds = result.getValue(facetId);
        for (String typeId : typeIds)
        {
            computedFacet.compute(typeId, this::_incrementContentTypeFacetValue);
        }
    }
    
    private Map<String, Object> _incrementContentTypeFacetValue(Object value, Map<String, Object> existingFacetValue)
    {
        if (existingFacetValue == null)
        {
            // Create a new facet value and return it
            Map<String, Object> newValue = new HashMap<>();
            newValue.put("value", value);
            ContentType contentType = _cTypeEP.getExtension((String) value);
            newValue.put("label", contentType != null ? contentType.getLabel() : value);
            newValue.put("count", 1L);
            newValue.put("type", "facet");
            return newValue;
        }
        else
        {
            existingFacetValue.compute("count", (k, v) -> ((Long) v) + 1);
            return existingFacetValue;
        }
    }
    
    private void _updateUserFacet(String facetId, Map<Object, Map<String, Object>> computedFacet, ContentConsistencyResult result)
    {
        UserIdentity user = result.getValue(facetId);
        if (user != null)
        {
            computedFacet.compute(user, this::_incrementUserFacetValue);
        }
    }
    
    private Map<String, Object> _incrementUserFacetValue(Object value, Map<String, Object> existingFacetValue)
    {
        if (existingFacetValue == null)
        {
            // Create a new facet value and return it
            Map<String, Object> newValue = new HashMap<>();
            newValue.put("value", UserIdentity.userIdentityToString((UserIdentity) value));
            newValue.put("label", _userHelper.getUserFullName((UserIdentity) value));
            newValue.put("count", 1L);
            newValue.put("type", "facet");
            return newValue;
        }
        else
        {
            existingFacetValue.compute("count", (k, v) -> ((Long) v) + 1);
            return existingFacetValue;
        }
    }
}
