/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.indexing.solr;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;
import org.apache.solr.common.SolrInputDocument;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.content.indexing.solr.content.attachment.ContentVisibleAttachmentIndexerExtensionPoint;
import org.ametys.cms.content.references.OutgoingReferences;
import org.ametys.cms.content.references.OutgoingReferencesExtractor;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.indexing.IndexableDataContext;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Component for {@link Content} indexing into a Solr server.
 */
public class SolrContentIndexer extends AbstractLogEnabled implements Component, Serviceable, SolrFieldNames
{
    /** The component role. */
    public static final String ROLE = SolrContentIndexer.class.getName();
    
    /** The resource indexer */
    protected SolrResourceIndexer _resourceIndexer;
    /** The system property extension point. */
    protected SystemPropertyExtensionPoint _systemPropEP;
    /** The content helper */
    protected ContentHelper _contentHelper;
    /** The outgoing references extractor */
    protected OutgoingReferencesExtractor _outgoingReferencesExtractor;
    /** The extension point for ContentVisibleAttachmentIndexers */
    protected ContentVisibleAttachmentIndexerExtensionPoint _contentVisibleAttachmentIndexerEP;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _resourceIndexer = (SolrResourceIndexer) manager.lookup(SolrResourceIndexer.ROLE);
        _contentHelper = (ContentHelper) manager.lookup(ContentHelper.ROLE);
        _systemPropEP = (SystemPropertyExtensionPoint) manager.lookup(SystemPropertyExtensionPoint.ROLE);
        _outgoingReferencesExtractor = (OutgoingReferencesExtractor) manager.lookup(OutgoingReferencesExtractor.ROLE);
        _contentVisibleAttachmentIndexerEP = (ContentVisibleAttachmentIndexerExtensionPoint) manager.lookup(ContentVisibleAttachmentIndexerExtensionPoint.ROLE);
    }
    
    /**
     * Populate a solr input document by adding fields to index into it.
     * @param content The content to index
     * @param document The main solr document to index into
     * @return Additional documents for repeater instances
     * @throws Exception if an error occurred while indexing
     */
    public List<SolrInputDocument> indexContent(Content content, SolrInputDocument document) throws Exception
    {
        // Properties specific to a stand-alone indexation.
        String contentId = content.getId();
        document.addField(ID, contentId);
        document.addField(DOCUMENT_TYPE, TYPE_CONTENT);
        
        indexContentTitle(content, document); 
       
        document.addField(CONTENT_NAME, SolrIndexer.truncateUtf8StringValue(content.getName(), getLogger(), contentId, CONTENT_NAME));
        _indexOutgoingReferences(content, document);
        _indexVisibleAttachments(content, document);
        
        document.addField(WORKFLOW_REF_DV, contentId + "#workflow");
        
        // Index content properties and attributes
        IndexableDataContext context = IndexableDataContext.newInstance();
        Optional.ofNullable(content.getLanguage())
                .filter(StringUtils::isNotBlank)
                .map(Locale::new)
                .ifPresent(context::withLocale);
        return content.indexData(document, context);
    }
    
    private void _indexOutgoingReferences(Content content, SolrInputDocument document)
    {
        // Found by the extractor (resource references found in all data of the content)
        _outgoingReferencesExtractor.getOutgoingReferences(content).values() // key is the data path, we do not care what data it comes from
                .parallelStream()
                .map(OutgoingReferences::entrySet)
                .flatMap(Set::parallelStream)
                .filter(outgoingRefs -> outgoingRefs.getKey().equals("explorer")) // only references of the resource explorer
                .map(Entry::getValue)
                .flatMap(List::parallelStream) // flat the resource ids
                .forEach(resourceId -> document.addField(CONTENT_OUTGOING_REFEERENCES_RESOURCE_IDS, resourceId));
        
        // Attachments of the content (just the root folder)
        Optional.ofNullable(content.getRootAttachments())
                .map(AmetysObject::getId)
                .ifPresent(id -> document.addField(CONTENT_OUTGOING_REFEERENCES_RESOURCE_IDS, id));
    }
    
    private void _indexVisibleAttachments(Content content, SolrInputDocument document)
    {
        Collection<String> values = _contentVisibleAttachmentIndexerEP.getExtensionsIds()
                .stream()
                .map(_contentVisibleAttachmentIndexerEP::getExtension)
                .map(attachmentIndexer -> attachmentIndexer.getVisibleAttachmentIds(content))
                .flatMap(Collection::stream)
                .collect(Collectors.toList());
        document.addField(CONTENT_VISIBLE_ATTACHMENT_RESOURCE_IDS, values);
    }
    
    /**
     * Index the content title
     * @param content The title
     * @param document The main solr document to index into
     */
    protected void indexContentTitle(Content content, SolrInputDocument document)
    {
        if (!ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID.equals(content.getType(Content.ATTRIBUTE_TITLE).getId()))
        {
            String title = _contentHelper.getTitle(content);
            document.addField(TITLE, SolrIndexer.truncateUtf8StringValue(title, getLogger(), content.getId(), TITLE));
            document.addField(TITLE_SORT, title);
        }
    }
    
    /**
     * Populate a Solr input document by adding fields for a single system property.
     * @param content The content to index
     * @param propertyId The system property ID.
     * @param document The solr document
     * @return true if there are partial update to apply
     * @throws Exception if an error occurred
     */
    @SuppressWarnings("unchecked")
    public boolean indexPartialSystemProperty(Content content, String propertyId, SolrInputDocument document) throws Exception
    {
        if (!_systemPropEP.hasExtension(propertyId))
        {
            throw new IllegalStateException("The property '" + propertyId + "' can't be indexed as it does not exist.");
        }
        
        SolrInputDocument tempDocument = new SolrInputDocument();
        
        SystemProperty property = _systemPropEP.getExtension(propertyId);
        property.indexValue(tempDocument, content, IndexableDataContext.newInstance());
        
        if (tempDocument.isEmpty())
        {
            // Does not have any partial update to apply, avoid to erase all the existing fields on the Solr document corresponding to this content (it would be lost)
            return false;
        }
        
        // Copy the indexed values as partial updates.
        for (String fieldName : tempDocument.getFieldNames())
        {
            Collection<Object> fieldValues = tempDocument.getFieldValues(fieldName);
            
            Map<String, Object> partialUpdate = new HashMap<>();
            partialUpdate.put("set", fieldValues);
            document.addField(fieldName, partialUpdate);
        }
        
        document.addField("id", content.getId());
        
        return true;
    }
}
