/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.indexing.solr;

import java.util.HashSet;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.runtime.plugin.Init;

/**
 * Creates one or several Solr cores at initialization.
 */
public class SolrCoreInit extends AbstractLogEnabled implements Init, Serviceable, Configurable
{
    
    /** The Solr indexer. */
    protected SolrIndexer _solr;
    
    /** The name of the cores to create. */
    protected Set<String> _coreNames;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _solr = (SolrIndexer) manager.lookup(SolrIndexer.ROLE);
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        _coreNames = new HashSet<>();
        for (Configuration coreConf : configuration.getChildren("core"))
        {
            String name = coreConf.getAttribute("name");
            if (StringUtils.isBlank(name))
            {
                throw new ConfigurationException("The core name can't be blank.");
            }
            _coreNames.add(name.trim());
        }
    }
    
    @Override
    public void init() throws Exception
    {
        boolean error = createCores(_coreNames);
        if (!error)
        {
            updateAmetysUrlCoreProperty();
        }
    }
    
    /**
     * Create the cores.
     * @param coreNames the names of the cores to create.
     * @return true if an exception occured. The exception was caught and not re-thrown (just logged)
     */
    protected boolean createCores(Set<String> coreNames)
    {
        try
        {
            Set<String> cores = _solr.getCoreNames();
            
            for (String coreName : coreNames)
            {
                if (!cores.contains(coreName))
                {
                    _solr.createCore(coreName);
                }
                else
                {
                    if (getLogger().isDebugEnabled())
                    {
                        getLogger().debug("Core '" + coreName + "' already exists, skipping it.");
                    }
                }
            }
        }
        catch (Exception e)
        {
            // Don't re-throw the exception: we don't want to prevent the server from starting.
            getLogger().warn("Error checking if some cores exist.", e);
            return true;
        }
        return false;
    }
    
    /**
     * Updates the Ametys URL on Solr server
     * @return true if an exception occured. The exception was caught and not re-thrown (just logged)
     */
    protected boolean updateAmetysUrlCoreProperty()
    {
        try
        {
            _solr.updateAmetysUrlCoreProperty();
        }
        catch (Exception e)
        {
            // Don't re-throw the exception: we don't want to prevent the server from starting.
            getLogger().warn("Error when trying to update Ametys URL core property on Solr server.", e);
            return true;
        }
        return false;
    }
    
}
