/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.indexing.solr;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;

/**
 * Utility for indexing a "group" of MIME-Type in Solr
 */
public final class SolrResourceGroupedMimeTypes
{
    /** Group for pdf */
    public static final String PDF_GROUP = "pdf";
    /** Group for image */
    public static final String IMAGE_GROUP = "image";
    /** Group for text documents */
    public static final String TEXT_GROUP = "text";
    /** Group for presentation */
    public static final String PRES_GROUP = "pres";
    /** Group for spreadsheet */
    public static final String SPREADSHEET_GROUP = "spreadsheet";
    /** Group for video */
    public static final String VIDEO_GROUP = "video";
    /** Group for audio */
    public static final String AUDIO_GROUP = "audio";
    
    /** MIME types for pdf */
    protected static final String[] PDF_MIMETYPE = new String []
    {
        "application/pdf"
    };
    /** MIME types for images */
    protected static final String[] IMAGE_MIMETYPE = new String []
    {
        "image/gif", 
        "image/png", 
        "image/jpeg",
        "image/tiff",
        "image/webp"
    };
    /** MIME types for text documents */
    protected static final String[] TEXT_MIMETYPE = new String []
    {
        "application/msword",
        "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
        "application/vnd.oasis.opendocument.text", 
        "application/rtf", 
        "text/rtf",
        "application/x-abiword",
        "text/plain"
    };
    /** MIME types for presentation */
    protected static final String[] PRES_MIMETYPE = new String []
    {
        "application/vnd.ms-powerpoint",
        "application/vnd.openxmlformats-officedocument.presentationml.presentation",
        "application/vnd.oasis.opendocument.presentation"
    };
    /** MIME types for spreadsheets */
    protected static final String[] SPREADSHEET_MIMETYPE = new String []
    {
        "application/vnd.ms-excel",
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        "application/vnd.oasis.opendocument.spreadsheet"
    };
    /** MIME types for videos */
    protected static final String[] VIDEO_MIMETYPE = new String []
    {
        "application/x-shockwave-flash",
        "video/mpeg",
        "video/mp4",
        "video/quicktime",
        "video/x-flv",
        "video/msvideo",
        "video/x-msvideo",
        "video/x-ms-wmv",
        "video/ogg",
        "video/webm",
        "video/3gpp",
        "video/3gpp2"
        
    };
    /** MIME types for audio files */
    protected static final String[] AUDIO_MIMETYPE = new String []
    {
        "audio/mpeg",
        "audio/x-mpeg",
        "audio/x-ms-wma",
        "audio/vnd.rn-realaudio",
        "audio/x-wav",
        "audio/ogg",
        "audio/webm",
        "audio/3gpp",
        "audio/3gpp2",
        "audio/aac",
        "audio/midi"
    };
    
    private static final Map<String, List<String>> __GROUPED_MIME_TYPES = Map.of(
        PDF_GROUP, Arrays.asList(PDF_MIMETYPE),
        IMAGE_GROUP, Arrays.asList(IMAGE_MIMETYPE),
        TEXT_GROUP, Arrays.asList(TEXT_MIMETYPE),
        PRES_GROUP, Arrays.asList(PRES_MIMETYPE),
        SPREADSHEET_GROUP, Arrays.asList(SPREADSHEET_MIMETYPE),
        VIDEO_GROUP, Arrays.asList(VIDEO_MIMETYPE),
        AUDIO_GROUP, Arrays.asList(AUDIO_MIMETYPE)
    );
    
    private SolrResourceGroupedMimeTypes()
    {  }

    /**
     * Gets a view of the map of list of MIME-Types by group
     * @return a view of the map of list of MIME-Types by group
     */
    public static Map<String, List<String>> get()
    {
        return Collections.unmodifiableMap(__GROUPED_MIME_TYPES);
    }
    
    /**
     * Gets the group for the given MIME-Type
     * @param mimeType The MIME-Type
     * @return the group for the given MIME-Type
     */
    public static Optional<String> getGroup(String mimeType)
    {
        if (StringUtils.isBlank(mimeType))
        {
            return Optional.empty();
        }
        
        for (Map.Entry<String, List<String>> group : __GROUPED_MIME_TYPES.entrySet())
        {
            List<String> mimeTypesForGroup = group.getValue();
            if (mimeTypesForGroup.contains(mimeType))
            {
                return Optional.of(group.getKey());
            }
        }
        return Optional.empty();
    }
    
    /**
     * Tests if the given group value is valid
     * @param groupMimeType The group to test
     * @return <code>true</code> if the given group value is valid
     */
    public static boolean isValidGroup(String groupMimeType)
    {
        return __GROUPED_MIME_TYPES.containsKey(groupMimeType);
    }
}
