/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.indexing.solr;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.solr.client.solrj.SolrClient;
import org.apache.solr.client.solrj.SolrServerException;
import org.apache.solr.client.solrj.response.UpdateResponse;
import org.apache.solr.client.solrj.util.ClientUtils;
import org.apache.solr.common.SolrInputDocument;

import org.ametys.cms.search.solr.SolrClientProvider;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.workflow.repository.WorkflowAwareAmetysObject;
import org.ametys.plugins.workflow.support.WorkflowProvider;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

import com.opensymphony.workflow.spi.Step;

/**
 * Component indexing the whole workflow of {@link AmetysObject}s.
 */
public class SolrWorkflowIndexer extends AbstractLogEnabled implements Component, Serviceable, SolrFieldNames, Initializable
{
    /** The component role. */
    public static final String ROLE = SolrWorkflowIndexer.class.getName();
    
    /** The Solr client provider */
    protected SolrClientProvider _solrClientProvider;
    
    /** The workflow provider. */
    protected WorkflowProvider _workflowProvider;

    private boolean _enabled;

    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        _solrClientProvider = (SolrClientProvider) serviceManager.lookup(SolrClientProvider.ROLE);
        _workflowProvider = (WorkflowProvider) serviceManager.lookup(WorkflowProvider.ROLE);
    }
    
    public void initialize() throws Exception
    {
        _enabled = Config.getInstance().getValue("cms.solr.index.workflow");
    }
    
    /**
     * Index the workflow history of an Ametys object.
     * @param object The ametys object.
     * @param workspaceName The workspace where to work in
     * @throws SolrServerException If a solr error occurs.
     * @throws IOException If an I/O error occurs while indexing.
     */
    public void indexAmetysObjectWorkflow(WorkflowAwareAmetysObject object, String workspaceName) throws SolrServerException, IOException
    {
        if (_enabled)
        {
            SolrClient solrClient = _solrClientProvider.getUpdateClient(workspaceName, true);
            indexAmetysObjectWorkflow(object, workspaceName, solrClient);
        }
    }
    
    /**
     * Index the workflow history of an Ametys object.
     * @param object The ametys object.
     * @param workspaceName The workspace where to work in
     * @param solrClient The solr client to use
     * @throws SolrServerException If a solr error occurs.
     * @throws IOException If an I/O error occurs while indexing.
     */
    public void indexAmetysObjectWorkflow(WorkflowAwareAmetysObject object, String workspaceName, SolrClient solrClient) throws SolrServerException, IOException
    {
        if (_enabled)
        {
            long time_0 = System.currentTimeMillis();
            getLogger().debug("Indexing the workflow of {}", object);
            
            AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(object);
            
            long workflowId = object.getWorkflowId();
            List<Step> currentSteps = workflow.getCurrentSteps(workflowId);
            List<Step> historySteps = workflow.getHistorySteps(workflowId);
            
            String objectId = object.getId();
            
            List<SolrInputDocument> documents = new ArrayList<>();
            
            List<String> historyStepIds = indexSteps(historySteps, objectId, documents);
            List<String> currentStepIds = indexSteps(currentSteps, objectId, documents);
            
            indexWorkflowEntry(workflow, workflowId, objectId, historyStepIds, currentStepIds, documents);
            
            String collection = _solrClientProvider.getCollectionName(workspaceName);
            UpdateResponse solrResponse = solrClient.add(collection, documents);
            int status = solrResponse.getStatus();
            
            if (status != 0)
            {
                throw new IOException("Workflow history indexation error (status code=" + status + ")");
            }
            
            getLogger().debug("Successfully indexed workflow of {} in {} ms", object, System.currentTimeMillis() - time_0);
        }
    }
    
    private void indexWorkflowEntry(AmetysObjectWorkflow workflow, long workflowId, String objectId, List<String> historyStepIds, List<String> currentStepIds, List<SolrInputDocument> documents)
    {
        SolrInputDocument entryDoc = new SolrInputDocument();
        
        String id = objectId + "#workflow";
        
        String name = workflow.getWorkflowName(workflowId);
        int state = workflow.getEntryState(workflowId);
        
        entryDoc.addField(ID, id);
        entryDoc.addField(DOCUMENT_TYPE, TYPE_WF_ENTRY);
        entryDoc.addField(WORKFLOW_NAME, name);
        entryDoc.addField(WORKFLOW_ENTRY_STATE, state);
        entryDoc.addField(WORKFLOW_HISTORY_STEPS_DV, historyStepIds);
        entryDoc.addField(WORKFLOW_CURRENT_STEPS_DV, currentStepIds);
        
        documents.add(entryDoc);
    }
    
    private List<String> indexSteps(List<Step> steps, String objectId, List<SolrInputDocument> documents)
    {
        List<String> ids = new ArrayList<>();
        
        for (Step step : steps)
        {
            String id = indexStep(step, objectId, documents);
            ids.add(id);
        }
        
        return ids;
    }

    private String indexStep(Step step, String objectId, List<SolrInputDocument> documents)
    {
        SolrInputDocument entryDoc = new SolrInputDocument();
        
        String id = objectId + "#step" + step.getId();
        int stepId = step.getStepId();
        int actionId = step.getActionId();
        String owner = step.getOwner();
        String caller = step.getCaller();
        Date startDate = step.getStartDate();
        Date dueDate = step.getDueDate();
        Date finishDate = step.getFinishDate();
        String status = step.getStatus();
        
        entryDoc.addField(ID, id);
        entryDoc.addField(DOCUMENT_TYPE, TYPE_WF_STEP);
        
        entryDoc.addField(WORKFLOW_STEP_ID, stepId);
        entryDoc.addField(WORKFLOW_STEP_ACTIONID, actionId);
        if (owner != null)
        {
            entryDoc.addField(WORKFLOW_STEP_OWNER, owner);
        }
        if (caller != null)
        {
            entryDoc.addField(WORKFLOW_STEP_CALLER, caller);
        }
        if (startDate != null)
        {
            entryDoc.addField(WORKFLOW_STEP_STARTDATE, startDate);
        }
        if (dueDate != null)
        {
            entryDoc.addField(WORKFLOW_STEP_DUEDATE, dueDate);
        }
        if (finishDate != null)
        {
            entryDoc.addField(WORKFLOW_STEP_FINISHDATE, finishDate);
        }
        if (status != null)
        {
            entryDoc.addField(WORKFLOW_STEP_STATUS, status);
        }
        
        documents.add(entryDoc);
        
        return id;
    }
    
    /**
     * Unindex the workflow history of an Ametys object.
     * @param objectId The Ametys object ID.
     * @param workspaceName The workspace where to work in 
     * @param solrClient The solr client to use
     * @throws SolrServerException If a solr error occurs.
     * @throws IOException If an I/O error occurs while indexing.
     */
    public void unindexAmetysObjectWorkflow(String objectId, String workspaceName, SolrClient solrClient) throws SolrServerException, IOException
    {
        if (_enabled)
        {
            // _documentType:(workflowEntry OR workflowStep) AND id:content\://xxx#*
            StringBuilder query = new StringBuilder();
            query.append(DOCUMENT_TYPE).append(":(")
            .append(TYPE_WF_ENTRY).append(" OR ").append(TYPE_WF_STEP)
            .append(") AND id:").append(ClientUtils.escapeQueryChars(objectId + "#")).append("*");
            
            solrClient.deleteByQuery(_solrClientProvider.getCollectionName(workspaceName), query.toString());
        }
    }
    
}
