/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.indexing.solr.observation;

import java.util.Map;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.solr.client.solrj.SolrClient;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.content.indexing.solr.SolrIndexer;
import org.ametys.cms.content.indexing.solr.SolrWorkflowIndexer;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.cms.search.solr.SolrClientProvider;
import org.ametys.cms.search.solr.field.WorkflowStepSearchField;
import org.ametys.core.observation.AsyncObserver;
import org.ametys.core.observation.Event;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;

/**
 * Observer in charge for indexing contents when created or modified.
 */
public class IndexContentObserver extends AbstractLogEnabled implements AsyncObserver, Contextualizable, Serviceable
{
    /** The provider of solr clients */
    protected SolrClientProvider _solrClientProvider;
    
    /** The Solr indexer. */
    protected SolrIndexer _solrIndexer;
    
    /** The solr workflow indexer. */
    protected SolrWorkflowIndexer _solrWfIndexer;
    
    /** The component context. */
    protected Context _context;

    private AmetysObjectResolver _resolver;

    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        _solrClientProvider = (SolrClientProvider) serviceManager.lookup(SolrClientProvider.ROLE);
        _solrIndexer = (SolrIndexer) serviceManager.lookup(SolrIndexer.ROLE);
        _solrWfIndexer = (SolrWorkflowIndexer) serviceManager.lookup(SolrWorkflowIndexer.ROLE);
        _resolver = (AmetysObjectResolver) serviceManager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    public boolean supports(Event event)
    {
        return event.getId().equals(ObservationConstants.EVENT_CONTENT_MODIFIED)
            || event.getId().equals(ObservationConstants.EVENT_CONTENT_ADDED)
            || event.getId().equals(ObservationConstants.EVENT_CONTENT_WORKFLOW_CHANGED)
            || event.getId().equals(ObservationConstants.EVENT_CONTENT_TAGGED);
    }
    
    @Override
    public int getPriority(Event event)
    {
        return MAX_PRIORITY + 3000;
    }
    
    @Override
    public void observe(Event event, Map<String, Object> transientVars) throws Exception
    {
        if (ObserverHelper.isNotSuspendedObservationForIndexation())
        {
            Map<String, Object> args = event.getArguments();
            String contentId = (String) args.get(ObservationConstants.ARGS_CONTENT_ID);
            Boolean commit = (Boolean) args.get(ObservationConstants.ARGS_CONTENT_COMMIT);
            if (commit == null)
            {
                // If not specified, let commit argument be 'true' by default, in order to be sure to commit changes.
                commit = true;
            }
            String workspaceName = RepositoryConstants.DEFAULT_WORKSPACE;
            SolrClient solrClient = _solrClientProvider.getUpdateClient(workspaceName, commit);
            
            Content content = _resolver.resolveById(contentId);
            
            if (content != null && index(content))
            {
                if (event.getId().equals(ObservationConstants.EVENT_CONTENT_WORKFLOW_CHANGED))
                {
                    _solrIndexer.updateSystemProperty(content, WorkflowStepSearchField.NAME, workspaceName, solrClient);
                    if (content instanceof WorkflowAwareContent)
                    {
                        _solrWfIndexer.indexAmetysObjectWorkflow((WorkflowAwareContent) content, workspaceName, solrClient);
                    }
                }
                else
                {
                    _solrIndexer.indexContent(content.getId(), workspaceName, false, solrClient);
                }
            }
        }
    }
    
    /**
     * Test whether to index the given content or not.
     * @param content the content to test.
     * @return true to index the content, false otherwise.
     */
    protected boolean index(Content content)
    {
        return true;
    }
    
}
