/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.indexing.solr.observation;

import java.util.Collection;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.cms.content.indexing.solr.SolrIndexer;
import org.ametys.cms.indexing.solr.SolrAclCacheInfluence;
import org.ametys.cms.indexing.solr.SolrAclCacheUninfluentialObject;
import org.ametys.core.ObservationConstants;
import org.ametys.core.observation.AsyncObserver;
import org.ametys.core.observation.Event;
import org.ametys.core.right.RightManager;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Reload the ACL Solr cache when a right assignment on READER profile changed.
 * 
 * <br>The {@link ObservationConstants#ARGS_ACL_CONTEXT context object} can be an uninfluential one
 * on the Solr Acl cache, and avoiding this reloading is possible by marking it with the {@link SolrAclCacheUninfluentialObject} interface.
 */
public class ReloadSolrCacheForAclUpdatedObserver extends AbstractLogEnabled implements AsyncObserver, Serviceable
{
    /** The solr indexer */
    protected SolrIndexer _solrIndexer;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _solrIndexer = (SolrIndexer) manager.lookup(SolrIndexer.ROLE);
    }
    
    @Override
    public boolean supports(Event event)
    {
        Map<String, Object> arguments = event.getArguments();
        @SuppressWarnings("unchecked")
        Collection<String> profileIds = (Collection<String>) arguments.get(ObservationConstants.ARGS_ACL_PROFILES);
        
        return ObservationConstants.EVENT_ACL_UPDATED.equals(event.getId()) 
               && profileIds.contains(RightManager.READER_PROFILE_ID);
    }
    
    @Override
    public int getPriority(Event event)
    {
        return MAX_PRIORITY;
    }
    
    @Override
    public void observe(Event event, Map<String, Object> transientVars) throws Exception
    {
        Map<String, Object> eventArgs = event.getArguments();
        Object aclContext = eventArgs.get(ObservationConstants.ARGS_ACL_CONTEXT);
        Boolean solrAclCacheUninfluential = (Boolean) eventArgs.get(org.ametys.cms.ObservationConstants.ARGS_ACL_SOLR_CACHE_UNINFLUENTIAL);
        if (solrAclCacheUninfluential == null)
        {
            // If not specified, let be 'false' by default
            solrAclCacheUninfluential = false;
        }
        
        getLogger().info("ReloadSolrCacheForAclUpdatedObserver called for context \"{}\" and profile \"READER\"", aclContext);
        
        if (solrAclCacheUninfluential)
        {
            getLogger().info("Acl cache will not be reloaded for \"{}\" because it was notified as uninfluential", aclContext);
        }
        else
        {
            if (SolrAclCacheInfluence.isUninfluent(aclContext))
            {
                getLogger().info("Acl cache will not be reloaded because \"{}\" is uninfluent on it", aclContext);
            }
            else
            {
                _solrIndexer.reloadAclCache();
            }
        }
    }
}
