/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.content.references;

import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.repository.Content;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeaterEntry;
import org.ametys.plugins.repository.data.type.ModelItemTypeConstants;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.type.ModelItemType;

/**
 * The outgoing references extractor is a class that analyzes a content to extract the list of outgoing references in the metadata.
 * Theses references should be stored on the content to allow research or post-processing later (example: consistency analyzer).
 */
public class OutgoingReferencesExtractor implements Component
{
    /** Avalon role */
    public static final String ROLE = OutgoingReferencesExtractor.class.getName();

    /**
     * This method analyzes the content to return the list of outgoing references grouped by data path.
     * @param content The content to analyze
     * @return A Map where keys are data path and values are outgoing references for this data. Can not be null.
     */
    public Map<String, OutgoingReferences> getOutgoingReferences(Content content)
    {
        return _getOutgoingReferences(content, StringUtils.EMPTY);
    }

    /**
     * Analyze a data holder to extract the outgoing references
     * @param dataHolder the data holder to analyze
     * @param dataPathPrefix the data path of the current data holder
     * @return the outgoing references, indexed by data path
     */
    protected Map<String, OutgoingReferences> _getOutgoingReferences(ModelAwareDataHolder dataHolder, String dataPathPrefix)
    {
        Map<String, OutgoingReferences> outgoingReferencesByPath = new HashMap<>();
        
        for (String dataName : dataHolder.getDataNames())
        {
            ModelItem definition = dataHolder.getDefinition(dataName);
            ModelItemType type = definition.getType();

            String dataPath = StringUtils.isNotEmpty(dataPathPrefix) ? dataPathPrefix + ModelItem.ITEM_PATH_SEPARATOR + dataName : dataName;

            if (type instanceof Referencer)
            {
                Object value = dataHolder.getValue(dataName);
                OutgoingReferences outgoingReferences = ((Referencer) type).getOutgoingReferences(value);
                outgoingReferencesByPath.put(dataPath, outgoingReferences);
            }
            else if (ModelItemTypeConstants.REPEATER_TYPE_ID.equals(type.getId()))
            {
                ModelAwareRepeater repeater = dataHolder.getRepeater(dataName);
                for (ModelAwareRepeaterEntry entry : repeater.getEntries())
                {
                    Map<String, OutgoingReferences> entryOutgoingReferences = _getOutgoingReferences(entry, dataPath + '[' + entry.getPosition() + ']');
                    outgoingReferencesByPath.putAll(entryOutgoingReferences);
                }
            }
            else if (ModelItemTypeConstants.COMPOSITE_TYPE_ID.equals(type.getId()))
            {
                Map<String, OutgoingReferences> compOutgoingReferences = _getOutgoingReferences(dataHolder.getComposite(dataName), dataPath);
                outgoingReferencesByPath.putAll(compOutgoingReferences);
            }
        }

        return outgoingReferencesByPath;
    }
}
