/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.referencetable;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.data.ContentValue;
import org.ametys.cms.repository.Content;
import org.ametys.cms.workflow.NeedAllValues;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.parameter.DefaultValidator;
import org.ametys.runtime.parameter.Errors;

/**
 * This validator does not allow a set of values only containing candidates. At
 * least one content is expected
 */
@NeedAllValues
public class CandidatesValidator extends DefaultValidator
{
    /** Ametys resolver */
    protected AmetysObjectResolver _resolver;
    
    /** The helper component for hierarchical reference tables */
    protected HierarchicalReferenceTablesHelper _hierarchicalReferenceTablesHelper;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _hierarchicalReferenceTablesHelper = (HierarchicalReferenceTablesHelper) manager.lookup(HierarchicalReferenceTablesHelper.ROLE);
    }
    
    @Override
    protected void validateSingleValue(Object value, Errors errors)
    {
        super.validateSingleValue(value, errors);
        
        ContentValue contentValue = (ContentValue) value;
        Content content = contentValue != null ? contentValue.getContentIfExists().orElse(null) : null;
        
        if (content != null && _hierarchicalReferenceTablesHelper.isCandidate(content))
        {
            if (getLogger().isDebugEnabled())
            {
                getLogger().debug("The validator refused a value because it contained only candidates");
            }
            
            errors.addError(new I18nizableText("plugin.cms", "PLUGINS_CMS_HIERARCHICAL_REFERENCE_TABLES_VALIDATOR_SINGLE_VALUE_REJECTED_MSG"));
        }
    }
    
    @Override
    protected void validateArrayValues(Object[] values, Errors errors)
    {
        super.validateArrayValues(values, errors);
        
        if (values != null && values.length > 0 && _hasOnlyCandidates((ContentValue[]) values))
        {
            if (getLogger().isDebugEnabled())
            {
                getLogger().debug("The validator refused a value because it contained only candidates");
            }
            
            errors.addError(new I18nizableText("plugin.cms", "PLUGINS_CMS_HIERARCHICAL_REFERENCE_TABLES_VALIDATOR_ARRAY_VALUES_REJECTED_MSG"));
        }
    }
    
    /**
     * Indicates if the array of contents contains only candidates
     * @param contents the list of content
     * @return true if it contains only candidates
     */
    protected boolean _hasOnlyCandidates(ContentValue[] contents)
    {
        for (ContentValue contentValue : contents)
        {
            Content content = contentValue.getContentIfExists().orElse(null);
            
            if (content != null && !_hierarchicalReferenceTablesHelper.isCandidate(content))
            {
                return false;
            }
        }
        
        return true;
    }
}
