/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.referencetable;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.excalibur.source.Source;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;
import org.ametys.core.cocoon.JSonReader;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Action for getting information about the hierarchy of a reference table
 */
public class GetHierarchicalReferenceTablesAction extends ServiceableAction
{
    /** The extension point for content types */
    protected ContentTypeExtensionPoint _contentTypeEP;
    /** The resolver for Ametys Objects */
    protected AmetysObjectResolver _ametysResolver;
    /** The helper component for hierarchical reference tables */
    protected HierarchicalReferenceTablesHelper _hierarchicalReferenceTablesHelper;
    /** Content types helper */
    protected ContentTypesHelper _cTypeHelper;
    /** The content helper */
    protected ContentHelper _contentHelper;
    /** The source resolver */
    protected org.apache.excalibur.source.SourceResolver _srcResolver;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _srcResolver = (org.apache.excalibur.source.SourceResolver) smanager.lookup(org.apache.excalibur.source.SourceResolver.ROLE);
        _contentTypeEP = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
        _contentHelper = (ContentHelper) smanager.lookup(ContentHelper.ROLE);
        _ametysResolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _hierarchicalReferenceTablesHelper = (HierarchicalReferenceTablesHelper) smanager.lookup(HierarchicalReferenceTablesHelper.ROLE);
        _cTypeHelper = (ContentTypesHelper) smanager.lookup(ContentTypesHelper.ROLE);
    }
    
    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Map<String, Object> result = new HashMap<>();
        Request request = ObjectModelHelper.getRequest(objectModel);
        String contextPath = request.getContextPath();
        
        @SuppressWarnings("unchecked")
        Map<String, Object> jsParameters = (Map<String, Object>) objectModel.get(ObjectModelHelper.PARENT_CONTEXT);
        
        String contentId = jsParameters != null ? (String) jsParameters.get("contentId") : request.getParameter("contentId");
        String leafContentTypeId = jsParameters != null ? (String) jsParameters.get("leafContentType") : request.getParameter("leafContentType");
        ContentType leafContentType = _contentTypeEP.getExtension(leafContentTypeId);

        if (leafContentType == null)
        {
            getLogger().warn(String.format("Unknown content type id '%s'", leafContentTypeId));
            return EMPTY_MAP;
        }
        
        boolean checkSelection = jsParameters != null ? jsParameters.get("checkSelection") != null && (Boolean) jsParameters.get("checkSelection") : false;
        ContentType topLevelContentType = _hierarchicalReferenceTablesHelper.getTopLevelType(leafContentType);
        List<Map<String, Object>> children = new ArrayList<>();
        
        try (AmetysObjectIterable<Content> contents = _getChildren(contentId, topLevelContentType, leafContentType))
        {
            for (Content child : contents)
            {
                children.add(_contentToJson(child, false, topLevelContentType, leafContentType, checkSelection, contextPath));
            }
        }
        
        if ("root".equals(contentId) && _hierarchicalReferenceTablesHelper.supportCandidates(leafContentType))
        {
            // Add root for candidates
            children.add(_rootCandidateToJson(leafContentType, checkSelection));
        }

        result.put("children", children);
        
        request.setAttribute(JSonReader.OBJECT_TO_READ, result);
        return EMPTY_MAP;
    }
    
    private AmetysObjectIterable<Content> _getChildren(String parentId, ContentType topLevelContentType, ContentType leafContentType)
    {
        if ("root".equals(parentId))
        {
            return _hierarchicalReferenceTablesHelper.getRootChildren(topLevelContentType, true);
        }
        else if ("rootCandidate".equals(parentId))
        {
            return _hierarchicalReferenceTablesHelper.getCandidates(leafContentType.getId());
        }
        else
        {
            return _hierarchicalReferenceTablesHelper.getDirectChildren((Content) _ametysResolver.resolveById(parentId));
        }
    }
    
    private Map<String, Object> _contentToJson(Content content, boolean full, ContentType topContentType, ContentType leafContentType, boolean checkSelection, String contextPath)
    {
        Map<String, Object> map = new HashMap<>();
        
        map.put("contentId", content.getId());
        map.put("contenttypesIds", content.getTypes());
        map.put("name", content.getName()); 
        map.put("title", _contentHelper.getTitle(content));
        map.put("lang", content.getLanguage()); 

        map.put("iconGlyph", _cTypeHelper.getIconGlyph(content));
        map.put("iconDecorator", _cTypeHelper.getIconDecorator(content));
        map.put("iconSmall", _cTypeHelper.getSmallIcon(content));
        map.put("iconMedium", _cTypeHelper.getMediumIcon(content));
        map.put("iconLarge", _cTypeHelper.getLargeIcon(content));
        map.put("tooltip", _getTooltip(content));
        
        if (checkSelection)
        {
            // Check if the node if selectable
            map.put("disabled", !_cTypeHelper.isInstanceOf(content, leafContentType.getId()));
        }
        
        map.put("leaf", _isLeaf(content, topContentType.getId(), leafContentType.getId()));
        map.put("isSimple", leafContentType.isSimple());
        map.put("isCandidate", _hierarchicalReferenceTablesHelper.isCandidate(content));
        map.put("areCandidatesAllowed", _hierarchicalReferenceTablesHelper.supportCandidates(leafContentType));
        map.put("mixin", content.getMixinTypes());
        map.put("type", "content");
        
        AmetysObjectIterable<Content> children = _hierarchicalReferenceTablesHelper.getDirectChildren(content);
        if (children.getSize() == 0)
        {
            // No child
            map.put("children", new ArrayList());
        }
        else if (full)
        {
            List<Map<String, Object>> childrenInfos = new ArrayList<>();
            for (Content child : children)
            {
                childrenInfos.add(_contentToJson(child, false, topContentType, leafContentType, checkSelection, contextPath));
            }
            map.put("children", childrenInfos);
        }
        
        return map;
    }
    
    private String _getTooltip(Content content)
    {
        if (_cTypeHelper.getView("tooltip", content.getTypes(), content.getMixinTypes()) != null)
        {
            String uri = _contentHelper.getContentHtmlViewUrl(content, "tooltip");
            Source source = null;
            try
            {
                source = _srcResolver.resolveURI(uri);
                try (InputStream is = source.getInputStream())
                {
                    String tooltip = IOUtils.toString(is, "utf-8");
                    
                    // Lets's keep only inside <body>
                    tooltip = StringUtils.substringAfter(tooltip, "<body");
                    tooltip = StringUtils.substringAfter(tooltip, ">");
                    tooltip = StringUtils.substringBefore(tooltip, "</body>");
                    
                    return tooltip;
                }
            }
            catch (IOException e)
            {
                getLogger().error("Failed to get tooltip view for content " + content.getId(), e);
            }
            finally 
            {
                _srcResolver.release(source);
            }
        }
        
        return StringUtils.EMPTY;
    }
    
    private Map<String, Object> _rootCandidateToJson(ContentType leafContentType, boolean checkSelection)
    {
        ContentType candidateType = _contentTypeEP.getExtension(HierarchicalReferenceTablesHelper.CANDIDATE_CONTENT_TYPE);
        Map<String, Object> infos = _cTypeHelper.getContentTypeProperties(candidateType);
        
        infos.put("contentId", "rootCandidate");
        infos.put("contenttypesIds", new String[] {leafContentType.getId()});
        infos.put("name", "rootCandidate"); 
        infos.put("title", new I18nizableText("plugin.cms", "PLUGINS_CMS_HIERARCHICAL_REFERENCE_TABLES_CANDIDATES_LABEL"));
        infos.put("leaf", false);
        infos.put("type", "root-candidate");
        
        if (checkSelection)
        {
            // Check if the node if selectable
            infos.put("disabled", true);
        }
        
        AmetysObjectIterable<Content> candidates = _hierarchicalReferenceTablesHelper.getCandidates(leafContentType.getId());
        if (candidates.getSize() == 0)
        {
            // No child
            infos.put("children", new ArrayList());
        }
        
        return infos;
    }
    
    private boolean _isLeaf(Content content, String topContentTypeId, String leafContentTypeId)
    {
        if (_cTypeHelper.isInstanceOf(content, leafContentTypeId))
        {
            // a child can be a leaf, or it can be a hierarchy with only one content type, referencing itself as parent
            // hierarchy with only one content type, referencing itself as parent
            return !topContentTypeId.equals(leafContentTypeId);
        }
        else
        {
            return false;
        }
    }
}
