/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.referencetable;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.contenttype.ContentAttributeDefinition;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.repository.Content;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.runtime.model.ModelItem;

import com.opensymphony.workflow.WorkflowException;

/**
 * This element creates a button enabling to do operation on reference table contents which can have a child.
 */
public class HierarchicalReferenceTableClientSideElement extends StaticClientSideElement
{
    /** The extension point for content types */
    protected ContentTypeExtensionPoint _contentTypeEP;
    /** The helper component for hierarchical reference tables */
    protected HierarchicalReferenceTablesHelper _hierarchicalReferenceTableContentsHelper;
    /** The Ametys Object resolver */
    protected AmetysObjectResolver _resolver;
    /** The content types helper */
    protected ContentTypesHelper _cTypeHelper;
    /** The content helper */
    protected ContentHelper _contentHelper;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _contentTypeEP = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
        _hierarchicalReferenceTableContentsHelper = (HierarchicalReferenceTablesHelper) smanager.lookup(HierarchicalReferenceTablesHelper.ROLE);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _cTypeHelper = (ContentTypesHelper) smanager.lookup(ContentTypesHelper.ROLE);
        _contentHelper = (ContentHelper) smanager.lookup(ContentHelper.ROLE);
    }
    
    /**
     * Returns true if the given content can have a child, according to its hierarchy. See {@link ContentType#getParentAttributeDefinition()}
     * @param contentId The id of the content
     * @return true if the given content can have a child, according to its hierarchy
     */
    @Callable
    public boolean canHaveChild(String contentId)
    {
        try
        {
            Content content = _resolver.resolveById(contentId);
            for (String cTypeId : content.getTypes())
            {
                ContentType cType = _contentTypeEP.getExtension(cTypeId);
                if (cType != null && _hierarchicalReferenceTableContentsHelper.hasChildContentType(cType))
                {
                    return true;
                }
            }
        }
        catch (AmetysRepositoryException e)
        {
            // Ignore
        }
        
        return false;
    }
    
    /**
     * Get the hierarchy properties of the content types tree
     * @param leafContentTypeId The id of the leaf content type, which defines the hierarchy.
     * @return the hierarchy properties
     */
    @Callable
    public Map<String, Object> getReferenceTableProperties(String leafContentTypeId)
    {
        Map<String, Object> result = new HashMap<>();
        
        result.put("rootContentType", getContentTypeForRoot(leafContentTypeId));
        result.put("supportedContentTypes", _hierarchicalReferenceTableContentsHelper.getHierarchicalContentTypes(leafContentTypeId));
        ContentType leafContentType = _contentTypeEP.getExtension(leafContentTypeId);
        result.put("allowCandidates", _hierarchicalReferenceTableContentsHelper.supportCandidates(leafContentType));
        return result;
    }
    
    /**
     * Gets the content types ids for a child of the given content (must be non-leaf in a hierarchy of reference tables contents)
     * @param parentContentId The id of the parent content
     * @return the content types ids
     */
    @Callable
    public List<String> getContentTypesForChild(String parentContentId)
    {
        Content parentContent = _resolver.resolveById(parentContentId);
        List<ContentType> childContentTypes = _hierarchicalReferenceTableContentsHelper.getChildContentTypes(parentContent);
        
        return childContentTypes.stream().map(ContentType::getId).collect(Collectors.toList());
    }
    
    /**
     * Gets the content type id and the default title for a child of the root node of the hierarchy of reference table contents, defined by the given leaf content type
     * @param leafContentTypeId The id of the leaf content type, which defines the hierarchy.
     * @return the content type id for the root node of the hierarchy of reference table contents
     */
    @Callable
    public String getContentTypeForRoot(String leafContentTypeId)
    {
        ContentType leafContentType = _contentTypeEP.getExtension(leafContentTypeId);
        ContentType rootContentType = _hierarchicalReferenceTableContentsHelper.getTopLevelType(leafContentType);
        
        if (rootContentType != null)
        {
            return rootContentType.getId();
        }
        
        return null;
    }
    
    /**
     * Return the leafContentType
     * @param contentTypeId the content type id
     * @param contentId the content id to get the leaf content type from
     * @return the map containing the properties of the leaf content type
     */
    @Callable
    public Map<String, Object> getLeafContentType(String contentTypeId, String contentId)
    {
        ContentType cType = _contentTypeEP.getExtension(contentTypeId);
        if (_hierarchicalReferenceTableContentsHelper.isLeaf(cType))
        {
            return _cTypeHelper.getContentTypeProperties(cType);
        }
        else
        {
            AmetysObjectIterable<Content> contents = _hierarchicalReferenceTableContentsHelper.getDirectChildren(_resolver.resolveById(contentId));
            for (Content content: contents)
            {
                return getLeafContentType(contentTypeId, content.getId());
            }
        }
        return null;
    }
    
    /**
     * Get the path of attribute holding the parent
     * @param contentId The id of content
     * @return the path of parent attribute or null if not found
     */
    @Callable
    public String getParentAttributePath(String contentId)
    {
        Content content = _resolver.resolveById(contentId);
        List<ContentType> childrenContentTypes = _hierarchicalReferenceTableContentsHelper.getChildContentTypes(content);
        for (ContentType contentType : childrenContentTypes)
        {
            Optional<ContentAttributeDefinition> parentAttribute = contentType.getParentAttributeDefinition();
            if (!parentAttribute.isEmpty())
            {
                return parentAttribute.get().getPath();
            }
        }
        return null;
    }
    
    /**
     * Get the path of attribute when root is selected
     * @param leafContentTypeId The id of content
     * @return the path of parent attribute or null if not found
     */
    @Callable
    public String getParentAttributePathForRoot(String leafContentTypeId)
    {
        ContentType leafContentType = _contentTypeEP.getExtension(leafContentTypeId);
        return leafContentType.getParentAttributeDefinition()
                .map(ModelItem::getPath)
                .filter(StringUtils::isNotEmpty)
                .orElse(null);
    }
    
    /** Validate the candidate by removing the mixin from the content
     * @param contentId The id of the content
     * @param actionId the id of the workflow action
     * @return The result in a Map
     * @throws WorkflowException The Workflow Exception
     * @throws AmetysRepositoryException The Ametys Repository Exception
     */
    @Callable
    public Map<String, Object> validateCandidate(String contentId, int actionId) throws AmetysRepositoryException, WorkflowException
    {
        return _contentHelper.removeMixinType(contentId, HierarchicalReferenceTablesHelper.CANDIDATE_CONTENT_TYPE, actionId);
    }
    
    /**
     * Get the title's properties for edition
     * @param contentId The id of content being renamed
     * @return the title's properties
     */
    @Callable
    public Map<String, Object> getTitleForEdition(String contentId)
    {
        Content content = _resolver.resolveById(contentId);
        
        Map<String, Object> titleInfos = new HashMap<>();
        
        ModelItem titleDef = content.getDefinition(Content.ATTRIBUTE_TITLE);
        titleInfos.put("type", titleDef.getType().getId());
        titleInfos.put("label", titleDef.getLabel());
        
        if (ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID.equals(titleDef.getType().getId()))
        {
            titleInfos.put("value", _contentHelper.getTitleVariants(content));
        }
        else
        {
            titleInfos.put("value", _contentHelper.getTitle(content));
        }
        
        return titleInfos;
    }
}
