/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.referencetable;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.collections.CollectionUtils;

import org.ametys.cms.clientsideelement.relations.SetContentAttributeClientSideElement;
import org.ametys.cms.contenttype.ContentAttributeDefinition;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.repository.Content;
import org.ametys.core.ui.Callable;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.runtime.parameter.Validator;

/**
 * Set the "parent" attribute of a simple content
 */
public class SetParentContentClientSideElement extends SetContentAttributeClientSideElement
{
    /** The helper component for hierarchical simple contents */
    protected HierarchicalReferenceTablesHelper _hierarchicalSimpleContentsHelper;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _hierarchicalSimpleContentsHelper = (HierarchicalReferenceTablesHelper) manager.lookup(HierarchicalReferenceTablesHelper.ROLE);
    }
    
    /**
     * Tells if the given simple contents can have their parent attribute changed for the given new simple content parent.
     * @param contentIdsToMove The identifiers of the simple contents willing to change their parent attribute
     * @param newParentContentId The identifier of the new parent simple content. Can be "root" if it is the root of simple contents
     * @param leafContentTypeId If newParentContentId is equal to "root", is the identifier of the content type of the leaves of the tree. Otherwise, must be null
     * @return A map with key 'validContents' and key 'invalidContents'
     */
    @Callable
    public Map<String, Object> areValidMoves(List<String> contentIdsToMove, String newParentContentId, String leafContentTypeId)
    {
        Map<String, Object> result = new HashMap<>();
        List<Map<String, String>> validContents = new ArrayList<>();
        result.put("validContents", validContents);
        List<String> invalidContents = new ArrayList<>(); 
        result.put("invalidContents", invalidContents);
        
        if ("root".equals(newParentContentId))
        {
            ContentType leafContentType = _contentTypeExtensionPoint.getExtension(leafContentTypeId);
            String topLevelType = _hierarchicalSimpleContentsHelper.getTopLevelType(leafContentType).getId();
            
            for (String contentIdToMove : contentIdsToMove)
            {
                try
                {
                    Content contentToMove = _resolver.resolveById(contentIdToMove);
                    ContentAttributeDefinition parentAttributeDefinition = _getParentAttributeDefinition(contentToMove);
                    if (parentAttributeDefinition == null)
                    {
                        invalidContents.add(contentIdToMove);
                        break;
                    }
                    
                    // 1) As move towards the root node means removing the "parent" attribute, check it is not mandatory
                    Validator validator = parentAttributeDefinition.getValidator();
                    if (validator != null && (Boolean) validator.getConfiguration().get("mandatory"))
                    {
                        invalidContents.add(contentIdToMove);
                        break;
                    }
                    
                    // 2) Check the content is a potential child of root node, i.e. has good content type
                    if (Arrays.asList(contentToMove.getTypes()).contains(topLevelType))
                    {
                        validContents.add(_validContentToMap(contentToMove, Collections.singletonList(topLevelType)));
                    }
                    else
                    {
                        invalidContents.add(contentIdToMove);
                    }
                }
                catch (AmetysRepositoryException e)
                {
                    getLogger().error("An error occured while retrieving the content to move.", e);
                    invalidContents.add(contentIdToMove);
                }
            }
        }
        else
        {
            Content newParent;
            try
            {
                newParent = _resolver.resolveById(newParentContentId);
            }
            catch (AmetysRepositoryException e)
            {
                getLogger().error("An error occured while retrieving the new parent content.", e);
                result.put("error", "unknown-parent");
                return result;
            }
            List<String> childContentTypes = _hierarchicalSimpleContentsHelper.getChildContentTypes(newParent).stream()
                                                                                                              .map(ContentType::getId)
                                                                                                              .collect(Collectors.toList());
            for (String contentIdToMove : contentIdsToMove)
            {
                try
                {
                    Content contentToMove = _resolver.resolveById(contentIdToMove);
                    if (CollectionUtils.containsAny(Arrays.asList(contentToMove.getTypes()), childContentTypes))
                    {
                        // The content types of the content is valid with the new parent
                        Map<String, String> validContent = _validContentToMap(contentToMove, childContentTypes);
                        validContents.add(validContent);
                    }
                    else
                    {
                        invalidContents.add(contentIdToMove);
                    }
                }
                catch (AmetysRepositoryException e)
                {
                    getLogger().error("An error occured while retrieving the content to move.", e);
                    invalidContents.add(contentIdToMove);
                }
            }
        }
        
        return result;
    }
    
    private ContentAttributeDefinition _getParentAttributeDefinition(Content content)
    {
        for (String cTypeId : content.getTypes())
        {
            ContentType cType = _contentTypeExtensionPoint.getExtension(cTypeId);
            Optional<ContentAttributeDefinition> parentAttribute = cType.getParentAttributeDefinition();
            if (!parentAttribute.isEmpty())
            {
                return parentAttribute.get();
            }
        }
        
        return null;
    }
    
    private Map<String, String> _validContentToMap(Content validContent, List<String> childContentTypes)
    {
        Map<String, String> contentMap = new HashMap<>();
        contentMap.put("id", validContent.getId());
        
        List<String> types = Arrays.asList(validContent.getTypes());
        for (String childContentTypeId : childContentTypes)
        {
            if (types.contains(childContentTypeId))
            {
                ContentType childContentType = _contentTypeExtensionPoint.getExtension(childContentTypeId);
                childContentType.getParentAttributeDefinition()
                    .ifPresent(parentAttribute -> contentMap.put("parentAttributeName", parentAttribute.getName()));
            }
        }
        
        contentMap.put("parentAttributeValue", _hierarchicalSimpleContentsHelper.getParent(validContent));
        
        return contentMap;
    }
}
