/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.referencetable.search;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.languages.Language;
import org.ametys.cms.languages.LanguagesManager;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.cocoon.JSonReader;
import org.ametys.core.util.DateUtils;
import org.ametys.core.util.JSONUtils;
import org.ametys.core.util.ServerCommHelper;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.workflow.support.WorkflowProvider;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;
import org.ametys.runtime.i18n.I18nizableText;

import com.opensymphony.workflow.loader.StepDescriptor;
import com.opensymphony.workflow.loader.WorkflowDescriptor;

/**
 * Action for getting information about the referencing contents of a content
 *
 */
public class GetReferencingContentsAction extends ServiceableAction
{
    /** The extension point for content types */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    /** Content types helper */
    protected ContentTypesHelper _contentTypesHelper;
    /** The content helper */
    protected ContentHelper _contentHelper;
    /** Server comm helper */
    protected ServerCommHelper _serverCommHelper;
    /** Language Manager */
    protected LanguagesManager _languagesManager;
    /** Workflow provider */
    protected WorkflowProvider _workflowProvider;
    /** User helper */
    protected UserHelper _userHelper;
    /** The ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /** JSON Utils */
    protected JSONUtils _jsonUtils;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
        _contentTypesHelper = (ContentTypesHelper) smanager.lookup(ContentTypesHelper.ROLE);
        _serverCommHelper = (ServerCommHelper) smanager.lookup(ServerCommHelper.ROLE);
        _contentHelper = (ContentHelper) smanager.lookup(ContentHelper.ROLE);
        _languagesManager = (LanguagesManager) smanager.lookup(LanguagesManager.ROLE);
        _workflowProvider = (WorkflowProvider) smanager.lookup(WorkflowProvider.ROLE);
        _userHelper = (UserHelper) smanager.lookup(UserHelper.ROLE);
        _jsonUtils = (JSONUtils) smanager.lookup(JSONUtils.ROLE);
    }
    
    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Map<String, Object> result = new HashMap<>();
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        Map<String, Object> jsParameters = _serverCommHelper.getJsParameters();
        
        String contentId = (String) jsParameters.get("contentId");

        int begin = jsParameters.containsKey("start") ? (Integer) jsParameters.get("start") : 0; // Index of search
        int offset = jsParameters.containsKey("limit") ? (Integer) jsParameters.get("limit") : Integer.MAX_VALUE; // Number of results to SAX
        
        Set<String> filteredCTypes = _getFilteredContentTypes(jsParameters);
                   
        String lang = (String) jsParameters.get("lang");
        Locale defaultLocale = StringUtils.isNotEmpty(lang) ? new Locale(lang) : null;
        
        Set<Content> referencingContents = new LinkedHashSet<>();
        List<Map<String, Object>> referencingContentsJson = new ArrayList<>();
        
        int index = 0;
        referencingContents.addAll(getReferencingContents(contentId, true, filteredCTypes));
        
        for (Content content : referencingContents)
        {
            if (index >= begin && index < begin + offset)
            {
                referencingContentsJson.add(getContentData(content, defaultLocale));
            }
            
            index++;
        }
            
        result.put("contents", referencingContentsJson);
        result.put("total", referencingContents.size());
        request.setAttribute(JSonReader.OBJECT_TO_READ, result);
        return EMPTY_MAP;
    }
    
    @SuppressWarnings("unchecked")
    private Set<String> _getFilteredContentTypes(Map<String, Object> jsParameters)
    {
        String jsonFilter = (String) jsParameters.get("filter");
        if (!StringUtils.isEmpty(jsonFilter))
        {
            List<Object> filters = _jsonUtils.convertJsonToList(jsonFilter);
            
            return (Set<String>) filters.stream()
                .filter(Map.class::isInstance)
                .map(Map.class::cast)
                .filter(filter -> "contentTypes".equals(filter.get("property")))
                .map(filter -> filter.get("value"))
                .filter(List.class::isInstance)
                .map(List.class::cast)
                .flatMap(List::stream)
                .collect(Collectors.toSet());
        }
        else
        {
            return Set.of();
        }
    }
    
    /**
     * Get the contents referencing the entry of a reference table
     * @param contentId The id of the entry
     * @param excludeContentOfSameType true to exclude the contents of the same reference table
     * @param filteredCTypes The filtered content types. If empty no filter on content types should be applied
     * @return The referencing contents
     */
    protected Collection<Content> getReferencingContents (String contentId, boolean excludeContentOfSameType, Set<String> filteredCTypes)
    {
        Content content = _resolver.resolveById(contentId);
        
        Collection<Content> referencingContents = content.getReferencingContents();
        
        if (excludeContentOfSameType || !filteredCTypes.isEmpty())
        {
            return referencingContents.stream()
                .filter(refContent -> !excludeContentOfSameType || _isNotSameType(content, refContent))
                .filter(refContent -> filteredCTypes.isEmpty() || _filterContentType(refContent, filteredCTypes))
                .collect(Collectors.toCollection(LinkedHashSet::new));
        }
        else
        {
            return referencingContents;
        }
    }
    
    private boolean _filterContentType(Content refContent, Set<String> filteredCTypes)
    {
        for (String cType : filteredCTypes)
        {
            if (_contentTypesHelper.isInstanceOf(refContent, cType))
            {
                return true;
            }
        }
        
        return false;
    }
    
    private boolean _isNotSameType(Content content, Content refContent)
    {
        String[] ctypes = content.getTypes();
        String[] refCTypes = refContent.getTypes();
        
        for (String ctype : ctypes)
        {
            if (ArrayUtils.contains(refCTypes, ctype))
            {
                return false;
            }
        }
        
        return true;
    }
    
    /**
     * Get the JSON representation of content
     * @param content The content
     * @param defaultLocale The default locale
     * @return The content data
     */
    protected Map<String, Object> getContentData(Content content, Locale defaultLocale)
    {
        Map<String, Object> contentData = new HashMap<>();
        
        contentData.put("id", content.getId());
        contentData.put("name", content.getName());
        
        if (_contentHelper.isMultilingual(content) && ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID.equals(content.getType(Content.ATTRIBUTE_TITLE).getId()))
        {
            contentData.put("title", _contentHelper.getTitleVariants(content));
        }
        else
        {
            contentData.put("title", _contentHelper.getTitle(content));
        }
        
        contentData.put("language", _language2json(content));
        contentData.put("contentTypes", _contentTypes2json(content));
        contentData.put("iconGlyph", _contentTypesHelper.getIconGlyph(content));
        contentData.put("iconDecorator", _contentTypesHelper.getIconDecorator(content));
        contentData.put("smallIcon", _contentTypesHelper.getSmallIcon(content));
        contentData.put("mediumIcon", _contentTypesHelper.getMediumIcon(content));
        contentData.put("largeIcon", _contentTypesHelper.getLargeIcon(content));
        contentData.put("isSimple", _contentHelper.isSimple(content));
        contentData.put("lastModified", DateUtils.zonedDateTimeToString(content.getLastModified())); 
        contentData.put("creationDate", DateUtils.zonedDateTimeToString(content.getCreationDate())); 
        contentData.put("creator", _userHelper.user2json(content.getCreator(), true));
        contentData.put("contributor", _userHelper.user2json(content.getLastContributor(), true));
        contentData.put("workflowStep", _workflowStep2json(content));
        
        return contentData;
    }
    
    private List<I18nizableText> _contentTypes2json(Content content)
    {
        List<I18nizableText> cTypes = new ArrayList<>();
        for (String cTypeId : content.getTypes())
        {
            ContentType cType = _contentTypeExtensionPoint.getExtension(cTypeId);
            if (cType != null)
            {
                cTypes.add(cType.getLabel());
            }
            else if (getLogger().isWarnEnabled())
            {
                getLogger().warn(String.format("Trying to get the label for an unknown content type : '%s'.", cTypeId));
            }
        }
        
        return cTypes;
    }
    
    private Map<String, Object> _language2json(Content content)
    {
        Map<String, Object> infos = new LinkedHashMap<>();
        
        String languageCode = content.getLanguage();
        infos.put("code", languageCode);
        
        if (languageCode != null)
        {
            Language language = _languagesManager.getLanguage(languageCode);
            if (language != null)
            {
                infos.put("icon", language.getSmallIcon());
                infos.put("label", language.getLabel());
            }
        }
        
        return infos;
    }
    
    private Map<String, Object> _workflowStep2json(Content content)
    {
        Map<String, Object> workflowInfos = new LinkedHashMap<>();
        
        if (content instanceof WorkflowAwareContent)
        {
            WorkflowAwareContent waContent = (WorkflowAwareContent) content;
            int currentStepId = Math.toIntExact(waContent.getCurrentStepId());
            
            StepDescriptor stepDescriptor = _getStepDescriptor(waContent, currentStepId);
            
            if (stepDescriptor != null)
            {
                I18nizableText workflowStepName = new I18nizableText("application", stepDescriptor.getName());
                
                workflowInfos.put("stepId", currentStepId);
                workflowInfos.put("name", workflowStepName);
                
                String[] icons = new String[] {"small", "medium", "large"};
                for (String icon : icons)
                {
                    if ("application".equals(workflowStepName.getCatalogue()))
                    {
                        workflowInfos.put(icon + "Icon", "/plugins/cms/resources_workflow/" + workflowStepName.getKey() + "-" + icon + ".png");
                    }
                    else
                    {
                        String pluginName = workflowStepName.getCatalogue().substring("plugin.".length());
                        workflowInfos.put(icon + "Icon", "/plugins/" + pluginName + "/resources/img/workflow/" + workflowStepName.getKey() + "-" + icon + ".png");
                    }
                }
            }
        }
        
        return workflowInfos;
    }
    
    private StepDescriptor _getStepDescriptor(WorkflowAwareContent content, int stepId)
    {
        long workflowId = content.getWorkflowId();
        
        AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(content);
        
        String workflowName = workflow.getWorkflowName(workflowId);
        WorkflowDescriptor workflowDescriptor = workflow.getWorkflowDescriptor(workflowName);
        
        if (workflowDescriptor != null)
        {
            StepDescriptor stepDescriptor = workflowDescriptor.getStep(stepId);
            if (stepDescriptor != null)
            {
                return stepDescriptor;
            }
            else if (getLogger().isWarnEnabled())
            {
                getLogger().warn("Unknown step id '" + stepId + "' for workflow for name : " + workflowName);
            }
        }
        else if (getLogger().isWarnEnabled())
        {
            getLogger().warn("Unknown workflow for name : " + workflowName);
        }
        
        return null;
    }
}
