/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.referencetable.search;

import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.solr.client.solrj.util.ClientUtils;

import org.ametys.cms.content.referencetable.HierarchicalReferenceTablesHelper;
import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.model.ContentElementDefinition;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.systemprop.AbstractSystemProperty;
import org.ametys.runtime.model.ModelItem;

/**
 * {@link SystemProperty} which represents all parents (parent, parent of the parent, etc.) of a simple content.
 */
public class ParentContentSystemProperty extends AbstractSystemProperty<ContentValue, Content> implements ContentElementDefinition
{
    /** The helper component for hierarchical simple contents */
    protected HierarchicalReferenceTablesHelper _hierarchicalSimpleContentsHelper;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _hierarchicalSimpleContentsHelper = (HierarchicalReferenceTablesHelper) manager.lookup(HierarchicalReferenceTablesHelper.ROLE);
    }
    
    @Override
    public boolean isMultiple()
    {
        return true;
    }
    
    @Override
    public boolean isSortable()
    {
        return false;
    }
    
    @Override
    public String getConverter()
    {
        return "Ametys.plugins.cms.search.SearchGridHelper.convertContent";
    }
    
    @Override
    public String getRenderer()
    {
        return "Ametys.plugins.cms.search.SearchGridHelper.renderContent";
    }
    
    @Override
    public String getCriterionWidget()
    {
        return "edition.hidden";
        // return "edition.select-referencetable-content";
    }
    
    @Override
    public Query getQuery(Object value, Operator operator, String language, Map<String, Object> contextualParameters)
    {
        if (value instanceof String)
        {
            return () -> ParentContentSearchField.NAME + ":" + ClientUtils.escapeQueryChars((String) value);
        }
        else
        {
            return null;
        }
    }

    @Override
    public SearchField getSearchField()
    {
        return new ParentContentSearchField();
    }

    @Override
    public Object getValue(Content content)
    {
        List<String> parentsAndMe = _hierarchicalSimpleContentsHelper.getAllParents(content);
        parentsAndMe.add(content.getId());
        
        return parentsAndMe.stream()
                           .map(id -> new ContentValue(_resolver, id))
                           .collect(Collectors.toList())
                           .toArray(new ContentValue[parentsAndMe.size()]);
    }
    
    @Override
    protected String _getTypeId()
    {
        return ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID;
    }
    
    @Override
    public String getContentTypeId()
    {
        return null;
    }
    
    @Override
    public void setContentTypeId(String contentTypeId)
    {
        // Do nothing by default
    }
    
    @Override
    public Collection< ? extends ModelItem> getModelItems()
    {
        return Collections.EMPTY_LIST;
    }
}
