/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.referencetable.search;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.model.SearchCriterionHelper;
import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.ui.model.AbstractSearchUIModel;
import org.ametys.cms.search.ui.model.SearchUIColumn;
import org.ametys.cms.search.ui.model.SearchUIColumnHelper;
import org.ametys.cms.search.ui.model.SearchUICriterion;
import org.ametys.cms.search.ui.model.SearchUIModel;
import org.ametys.cms.search.ui.model.impl.IndexingFieldSearchUICriterion;
import org.ametys.cms.search.ui.model.impl.SystemSearchUICriterion;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewHelper;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.ViewItemContainer;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * Implementation of {@link SearchUIModel} for contents referencing the same content's values for a given attribute.
 */
public class ReferencingContentsWithSameValuesSearchUIModel extends AbstractSearchUIModel implements Serviceable, Contextualizable
{
    private static final String __CONTENT_ID_CONTEXTUAL_PARAM_NAME = "contentId";
    private static final String __REFERENCE_FIELD_CONTEXTUAL_PARAM_NAME = "referenceField";
    
    /** ComponentManager for {@link SearchUICriterion}s. */
    protected ThreadSafeComponentManager<SearchUICriterion> _searchCriterionManager;
    
    private Context _context;
    private ServiceManager _manager;
    private AmetysObjectResolver _resolver;
    private ContentHelper _contentHelper;
    private SystemPropertyExtensionPoint _systemPropertyExtensionPoint;
    private SearchCriterionHelper _searchCriterionHelper;
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }

    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
        
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentHelper = (ContentHelper) manager.lookup(ContentHelper.ROLE);
        _systemPropertyExtensionPoint = (SystemPropertyExtensionPoint) manager.lookup(SystemPropertyExtensionPoint.ROLE);
        _searchCriterionHelper = (SearchCriterionHelper) manager.lookup(SearchCriterionHelper.ROLE);
    }
    
    @Override
    public Set<String> getContentTypes(Map<String, Object> contextualParameters)
    {
        if (contextualParameters != null && contextualParameters.containsKey(__CONTENT_ID_CONTEXTUAL_PARAM_NAME)  && contextualParameters.containsKey(__REFERENCE_FIELD_CONTEXTUAL_PARAM_NAME))
        {
            // Get the content type which has declared the attribute
            String contentId = (String) contextualParameters.get(__CONTENT_ID_CONTEXTUAL_PARAM_NAME);
            Content content = _resolver.resolveById(contentId);
            
            String referenceAttributePath = (String) contextualParameters.get(__REFERENCE_FIELD_CONTEXTUAL_PARAM_NAME);
            if (content.hasDefinition(referenceAttributePath))
            {
                ModelItem modelItem = content.getDefinition(referenceAttributePath);
                return Set.of(modelItem.getModel().getId());
            }
        }
        
        return Set.of();
    }
    
    @Override
    public Set<String> getExcludedContentTypes(Map<String, Object> contextualParameters)
    {
        return Collections.emptySet();
    }
    
    @Override
    public Map<String, SearchUICriterion> getCriteria(Map<String, Object> contextualParameters)
    {
        List<SearchUICriterion> criteria = new ArrayList<>();
        
        if (contextualParameters.containsKey(__CONTENT_ID_CONTEXTUAL_PARAM_NAME))
        {
            try
            {
                _searchCriterionManager = new ThreadSafeComponentManager<>();
                _searchCriterionManager.setLogger(getLogger());
                _searchCriterionManager.contextualize(_context);
                _searchCriterionManager.service(_manager);
                
                List<String> roles = new ArrayList<>();
                
                String contentId = (String) contextualParameters.get(__CONTENT_ID_CONTEXTUAL_PARAM_NAME);
                
                Content content = _resolver.resolveById(contentId);
                
                Set<String> setWithOnlyFirstCTypeId = Set.of(content.getTypes()[0]);
                
                Configuration conf = _searchCriterionHelper.getIndexingFieldCriteriaConfiguration(this, Optional.empty(), setWithOnlyFirstCTypeId, Content.ATTRIBUTE_TITLE, Optional.of(Operator.SEARCH), Optional.empty());
                _searchCriterionManager.addComponent("cms", null, Content.ATTRIBUTE_TITLE, IndexingFieldSearchUICriterion.class, conf);
                roles.add("title");
                
                conf = _searchCriterionHelper.getSystemCriteriaConfiguration(this, Optional.empty(), setWithOnlyFirstCTypeId, "full", Optional.empty());
                _searchCriterionManager.addComponent("cms", null, "full", SystemSearchUICriterion.class, conf);
                roles.add("full");
                
                conf = _searchCriterionHelper.getSystemCriteriaConfiguration(this, Optional.empty(), setWithOnlyFirstCTypeId, "workflowStep", Optional.empty());
                _searchCriterionManager.addComponent("cms", null, "workflowStep", SystemSearchUICriterion.class, conf);
                roles.add("workflowStep");
                
                if (contextualParameters.containsKey(__REFERENCE_FIELD_CONTEXTUAL_PARAM_NAME) && contextualParameters.containsKey("referenceValues"))
                {
                    String referenceMetadataPath = ((String) contextualParameters.get(__REFERENCE_FIELD_CONTEXTUAL_PARAM_NAME)).replaceAll("\\.", "/");
                    
                    if (content.hasDefinition(referenceMetadataPath))
                    {
                        ModelItem modelItem = content.getDefinition(referenceMetadataPath);
                        @SuppressWarnings("unchecked")
                        List<String> referenceAttributeValues = (List<String>) contextualParameters.get("referenceValues");
                        
                        DefaultConfiguration referenceFieldConf = new DefaultConfiguration("criteria");
                        // Default value(s)
                        for (String referenceAttributeValue : referenceAttributeValues)
                        {
                            DefaultConfiguration defaultValueConf = new DefaultConfiguration("default-value");
                            defaultValueConf.setValue(referenceAttributeValue);
                            referenceFieldConf.addChild(defaultValueConf);
                        }
                        // AND query on multiple entries
                        DefaultConfiguration multipleOperandConf = new DefaultConfiguration("multiple-operand");
                        multipleOperandConf.setValue("and");
                        referenceFieldConf.addChild(multipleOperandConf);
                        // Hidden
                        referenceFieldConf.setAttribute("hidden", "true");
                        
                        conf = _searchCriterionHelper.getIndexingFieldCriteriaConfiguration(this, Optional.of(referenceFieldConf), Set.of(modelItem.getModel().getId()), referenceMetadataPath, Optional.empty(), Optional.empty());
                        _searchCriterionManager.addComponent("cms", null, "referenceField", IndexingFieldSearchUICriterion.class, conf);
                        roles.add("referenceField");
                    }
                }
                
                _searchCriterionManager.initialize();
                
                // Lookup.
                for (String role : roles)
                {
                    criteria.add(_searchCriterionManager.lookup(role));
                }
                
                setCriteria(criteria);
                
                // Dispose
                _searchCriterionManager.dispose();
                _searchCriterionManager = null;
            }
            catch (Exception e)
            {
                getLogger().error("Error getting the search criteria.", e);
                throw new IllegalStateException("Error getting the search criteria.", e);
            }
        }
        else
        {
            setCriteria(criteria);
        }
        
        return super.getCriteria(contextualParameters);
    }

    @Override
    public Map<String, SearchUICriterion> getAdvancedCriteria(Map<String, Object> contextualParameters)
    {
        return Collections.emptyMap();
    }
    
    @Override
    public Map<String, SearchUICriterion> getFacetedCriteria(Map<String, Object> contextualParameters)
    {
        return Collections.emptyMap();
    }
    
    @Override
    public ViewItemContainer getResultItems(Map<String, Object> contextualParameters)
    {
        ViewItemContainer resultItems = new View();
        
        if (contextualParameters.containsKey(__CONTENT_ID_CONTEXTUAL_PARAM_NAME))
        {
            String contentId = (String) contextualParameters.get(__CONTENT_ID_CONTEXTUAL_PARAM_NAME);
            Content content = _resolver.resolveById(contentId);
            
            // Add a column for the title
            ModelItem titleDefinition = content.getDefinition(Content.ATTRIBUTE_TITLE);
            SearchUIColumn titleColumn = SearchUIColumnHelper.createModelItemColumn(titleDefinition);
            titleColumn.setAllowSortOnMultipleJoin(allowSortOnMultipleJoin());
            resultItems.addViewItem(titleColumn);
            
            // Add a column for the referenced field
            if (contextualParameters.containsKey(__REFERENCE_FIELD_CONTEXTUAL_PARAM_NAME))
            {
                Collection<ContentType> contentTypes = _contentHelper.getContentTypes(content);

                String referenceModelItemPath = ((String) contextualParameters.get(__REFERENCE_FIELD_CONTEXTUAL_PARAM_NAME)).replaceAll("\\.", "/");
                int lastIndexOfItemPathSeparator = referenceModelItemPath.lastIndexOf(ModelItem.ITEM_PATH_SEPARATOR);
                String referencedModelItemName = lastIndexOfItemPathSeparator > -1 ? referenceModelItemPath.substring(lastIndexOfItemPathSeparator + ModelItem.ITEM_PATH_SEPARATOR.length()) : referenceModelItemPath;

                // Get or create parent of the referenced field
                ViewItemAccessor parent = resultItems;
                if (lastIndexOfItemPathSeparator > -1)
                {
                    String parentPath = referenceModelItemPath.substring(0, lastIndexOfItemPathSeparator);
                    parent = (ViewItemAccessor) ViewHelper.addViewItem(parentPath, resultItems, contentTypes.toArray(new ContentType[contentTypes.size()]));
                }
                
                // Create the column for the referenced model item
                ModelItem referencedModelItem = _systemPropertyExtensionPoint.hasExtension(referencedModelItemName)
                        ? _systemPropertyExtensionPoint.getExtension(referencedModelItemName)
                        : ModelHelper.getModelItem(referencedModelItemName, contentTypes);
                SearchUIColumn referencedItemColumn = SearchUIColumnHelper.createModelItemColumn(referencedModelItem);
                referencedItemColumn.setAllowSortOnMultipleJoin(allowSortOnMultipleJoin());

                // Add the column to its parent
                parent.addViewItem(referencedItemColumn);
            }
            
            resultItems.addViewItem(SearchUIColumnHelper.createModelItemColumn(_systemPropertyExtensionPoint.getExtension("contributor")));
            resultItems.addViewItem(SearchUIColumnHelper.createModelItemColumn(_systemPropertyExtensionPoint.getExtension("lastModified")));
            resultItems.addViewItem(SearchUIColumnHelper.createModelItemColumn(_systemPropertyExtensionPoint.getExtension("contentTypes")));
            resultItems.addViewItem(SearchUIColumnHelper.createModelItemColumn(_systemPropertyExtensionPoint.getExtension("contentLanguage")));
            resultItems.addViewItem(SearchUIColumnHelper.createModelItemColumn(_systemPropertyExtensionPoint.getExtension("workflowStep")));
        }

        setResultItems(resultItems);
        return super.getResultItems(contextualParameters);
    }
}
