/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.referencetable.search;

import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.contenttype.ContentAttributeDefinition;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.repository.Content;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.ClientSideElement;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.runtime.model.ModelItem;

/**
 * {@link ClientSideElement} for search tool on contents with the same referencing values for a given attribute.
 */
public class SearchReferencingContentsWithSameValuesClientSideElement extends StaticClientSideElement
{
    private AmetysObjectResolver _resolver;
    private ContentTypeExtensionPoint _cTypeEP;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _cTypeEP = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    /**
     * Get the search context
     * @param initialContentId The id of initial content (from which the search was launched)
     * @param attributePath The path of the attribute
     * @param refIds The current values for this content and attribute
     * @return The search context
     */
    @Callable
    public Map<String, Object> getSearchContext(String initialContentId, String attributePath, List<String> refIds)
    {
        Map<String, Object> searchCtx = new HashMap<>();
        
        Content initialContent = _resolver.resolveById(initialContentId);
        
        Map<String, Object> initialContentInfos = new HashMap<>();
        initialContentInfos.put("contentId", initialContentId);
        initialContentInfos.put("contentTitle", initialContent.getTitle());
        searchCtx.put("initialContent", initialContentInfos);
        
        if (initialContent.hasDefinition(attributePath))
        {
            ModelItem modelItem = initialContent.getDefinition(attributePath);
            Map<String, Object> attributeInfos = new HashMap<>();
            
            attributeInfos.put("name", modelItem.getName());
            attributeInfos.put("path", modelItem.getPath());
            attributeInfos.put("label", modelItem.getLabel());
            
            ContentType refContentType = (ContentType) modelItem.getModel();
            attributeInfos.put("refContentType", refContentType.getLabel());
            
            searchCtx.put("attribute", attributeInfos);
            
            if (modelItem instanceof ContentAttributeDefinition)
            {
                String refTableId = ((ContentAttributeDefinition) modelItem).getContentTypeId();
                if (refTableId != null)
                {
                    Map<String, Object> refTableInfos = new HashMap<>();
                    
                    ContentType refTable = _cTypeEP.getExtension(refTableId);
                    refTableInfos.put("id", refTableId);
                    refTableInfos.put("label", refTable.getLabel());
                    
                    searchCtx.put("referenceTable", refTableInfos);
                }
            }
        }
        
        Locale defaultLocale = initialContent.getLanguage() != null ? new Locale(initialContent.getLanguage()) : null;

        Map<String, String> refValues = new HashMap<>();
        for (String refId : refIds)
        {
            Content refContent = _resolver.resolveById(refId);
            refValues.put(refId, refContent.getTitle(defaultLocale));
        }
        
        searchCtx.put("refValues", refValues);
        
        return searchCtx;
    }
}
