/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.version;

import java.util.Collections;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.repository.Content;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.ClientSideElement;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.version.VersionAwareAmetysObject;
import org.ametys.plugins.repository.version.VersionableAmetysObject;

/**
 * {@link ClientSideElement} for the button for comparing a content between two version
 */
public class CompareContentVersionClientSideElement extends StaticClientSideElement
{
    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /** The compare version helper */
    protected CompareVersionHelper _compareVersionHelper;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _compareVersionHelper = (CompareVersionHelper) smanager.lookup(CompareVersionHelper.ROLE);
    }
    
    /**
     * Gets comparison-information about the given content
     * <br>Information is if the content is comparable with its last validation version and the concerned versions.
     * @param contentId The content id
     * @param version The current version
     * @return The information
     */
    @Callable
    public Map<String, Object> isComparable(String contentId, String version)
    {
        Content content = _resolver.resolveById(contentId);
        if (!(content instanceof VersionableAmetysObject))
        {
            // is not versionable
            return Collections.EMPTY_MAP;
        }
        VersionableAmetysObject versionableContent = (VersionableAmetysObject) content;
        
        if (version != null)
        {
            versionableContent.switchToRevision(version);
        }
        
        String currentVersion = getVersionToCompare(versionableContent);
        String oldVersion = getBaseVersion(versionableContent);
        if (oldVersion == null)
        {
            return Map.of("noBaseVersion", true);
        }
        
        if (oldVersion.equals(currentVersion))
        {
            return Map.of("isSameVersion", true);
        }
        
        return Map.of(
                "isComparable", true,
                "version", currentVersion,
                "baseVersion", oldVersion
        );
    }
    
    /**
     * Gets the version to compare. By default, it is the current one.
     * @param versionable The {@link VersionAwareAmetysObject}
     * @return the source version
     */
    protected String getVersionToCompare(VersionAwareAmetysObject versionable)
    {
        return _compareVersionHelper.getCurrentVersion(versionable);
    }
    
    /**
     * Gets the base version of comparison. By default the previous one
     * @param versionable The {@link VersionAwareAmetysObject}
     * @return the base version
     */
    protected String getBaseVersion(VersionAwareAmetysObject versionable)
    {
        return _compareVersionHelper.getPreviousVersion(versionable);
    }
}
