/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.version;

import java.io.IOException;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.commons.io.IOUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.apache.http.client.utils.URIBuilder;

import org.ametys.cms.content.compare.ContentComparatorChange;
import org.ametys.cms.content.compare.ContentComparatorResult;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.ClientSideElement;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.data.external.ExternalizableDataProviderExtensionPoint;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewHelper;

/**
 * {@link ClientSideElement} for the tool for comparing a content between a base version and a target version. 
 */
public class CompareContentVersionToolClientSideElement extends StaticClientSideElement
{
    private CompareVersionHelper _compareVersionHelper;
    private SourceResolver _sourceResolver;
    private ContentTypesHelper _cTypesHelper;
    private ExternalizableDataProviderExtensionPoint _externalizableDataProviderEP;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _compareVersionHelper = (CompareVersionHelper) smanager.lookup(CompareVersionHelper.ROLE);
        _sourceResolver = (SourceResolver) smanager.lookup(SourceResolver.ROLE);
        _cTypesHelper = (ContentTypesHelper) smanager.lookup(ContentTypesHelper.ROLE);
        _externalizableDataProviderEP = (ExternalizableDataProviderExtensionPoint) smanager.lookup(ExternalizableDataProviderExtensionPoint.ROLE);
    }
    
    /**
     * Gets the JSON information about the changes for the given content id between the base and target provided versions.
     * @param contentId The {@link Content} id
     * @param viewName The view to use
     * @param showAllData true to display all data, false to display only data with diff
     * @param targetVersion The content version to be compared
     * @param baseVersion The base content version
     * @return The information
     * @throws IOException If an I/O exception occurred during the comparison between the two versions, or when getting content values
     * @throws ProcessingException If an exception occurred when converting the "change" view
     */
    @Callable
    public Map<String, Object> getDiffValues(String contentId, String viewName, boolean showAllData, String targetVersion, String baseVersion) throws IOException, ProcessingException
    {
        ContentVersionComparator comparator = new ContentVersionComparator(contentId, viewName, showAllData, targetVersion, baseVersion);
        comparator.compare();
        return comparator.resultToJson();
    }
    
    private class ContentVersionComparator
    {
        private final String _contentId;
        private final String _baseVersion;
        private final String _targetVersion;
        private Content _targetContent;
        private ContentComparatorResult _comparatorResult;
        private CompareView _resultCompareView;
        private String _viewName;
        private boolean _showAllData;
        
        ContentVersionComparator(String contentId, String viewName, boolean showAllData, String version, String baseVersion)
        {
            _contentId = contentId;
            _viewName = viewName;
            _showAllData = showAllData;
            _baseVersion = baseVersion;
            _targetVersion = version;
        }
        
        void compare() throws IOException
        {
            _targetContent = _compareVersionHelper.getContentVersion(_contentId, _targetVersion);
            
            _comparatorResult = _compareVersionHelper.compareVersions(_contentId, _targetVersion, _baseVersion);
            
            View wrappedEditionView = ViewHelper.getTruncatedView(_retrieveEditionView());
            Set<String> externalizableDataPaths = _externalizableDataProviderEP.getExternalizableDataPaths(_targetContent);
            _resultCompareView = new CompareView(wrappedEditionView, externalizableDataPaths, _showAllData, _comparatorResult);
        }
        
        Map<String, Object> resultToJson() throws IOException, ProcessingException
        {
            return Map.of(
                    "view", _resultCompareViewToJson(), 
                    "changedAttributeDataPaths", _changedAttributeDataPaths(),
                    // FIXME REPOSITORY-454 Use "baseValues", _contentToJson(_baseContent), 
                    "baseValues", _saxedContent(_baseVersion), 
                    // FIXME REPOSITORY-454 Use "values", _contentToJson(_targetContent));
                    "values", _saxedContent(_targetVersion));
        }
        
        private Map<String, Object> _resultCompareViewToJson() throws ProcessingException
        {
            return _resultCompareView.toJSON();
            
        }
        
        private Collection<String> _changedAttributeDataPaths() throws AmetysRepositoryException
        {
            List<ContentComparatorChange> contentComparatorChanges = _comparatorResult.getChanges();
            return _compareVersionHelper.filterChanges(contentComparatorChanges)
                    .map(ContentComparatorChange::getAttributeDataPath)
                    .collect(Collectors.toList());
        }
        
//        private Map<String, Object> _contentToJson(Content content)
//        {
//            return content.toJson(_changesView);
//        }
        
        private String _saxedContent(String contentVersion) throws IOException
        {
            // FIXME REPOSITORY-454 use Content#toJson instead, to manipulate JSON and to be able to provide the view through a simple API call, instead of calling '/_content.xml' URI
            // * when done, content#toJson will be available and the line for _targetContent can be uncommented, as well as declaration and calls of _contentToJson
            // * when done, _changesView can be passed to this method
            String uri;
            try
            {
                // cocoon://_content.xml?contentId=...
                uri = new URIBuilder()
                        .setScheme("cocoon")
                        .setHost("_content.xml")
                        .addParameter("contentId", _contentId)
                        .addParameter("contentVersion", contentVersion)
                        .addParameter("viewName", _viewName)
                        .addParameter("fallbackViewName", "main")
                        .addParameter("isEdition", "true")
                        .build()
                        .toString();
            }
            catch (URISyntaxException e)
            {
                throw new IOException(e);
            }
            
            Source saxedContent = _sourceResolver.resolveURI(uri);
            return IOUtils.toString(saxedContent.getInputStream(), StandardCharsets.UTF_8);
        }
        
        private View _retrieveEditionView()
        {
            String[] contentTypeIds = _targetContent.getTypes();
            String[] mixinIds = _targetContent.getMixinTypes();
            View view = _cTypesHelper.getView(_viewName, contentTypeIds, mixinIds);
            if (view == null)
            {
                view = _cTypesHelper.getView("main", contentTypeIds, mixinIds);
            }
            return Objects.requireNonNull(view);
        }
    }
}
