/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.version;

import java.io.IOException;
import java.util.List;
import java.util.Objects;
import java.util.function.Predicate;
import java.util.stream.Stream;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.content.compare.ContentComparator;
import org.ametys.cms.content.compare.ContentComparatorChange;
import org.ametys.cms.content.compare.ContentComparatorResult;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ResourceElementTypeHelper;
import org.ametys.cms.repository.Content;
import org.ametys.core.util.LambdaUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.version.VersionAwareAmetysObject;
import org.ametys.runtime.model.ModelItem;

import com.google.common.base.Predicates;

/**
 * Helper for retrieving some useful version informations about {@link VersionAwareAmetysObject}s, and compare some versions.
 */
public class CompareVersionHelper implements Component, Serviceable
{
    /** Avalon Role */
    public static final String ROLE = CompareVersionHelper.class.getName();
    
    /** The ametys object resolver */
    protected AmetysObjectResolver _ametysObjectResolver;
    /** The content comparator */
    protected ContentComparator _contentComparator;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentComparator = (ContentComparator) manager.lookup(ContentComparator.ROLE);
    }
    
    /**
     * Gets the current version of the given {@link VersionAwareAmetysObject}
     * @param versionable The {@link VersionAwareAmetysObject}
     * @return The current version
     */
    public String getNewestVersion(VersionAwareAmetysObject versionable)
    {
        String[] allRevisions = versionable.getAllRevisions();
        if (allRevisions.length < 1)
        {
            throw new IllegalStateException(String.format("An unexpected error occured, the versionable content '%s' has no revision.", versionable));
        }
        String revision = allRevisions[allRevisions.length - 1];
        return revision;
    }
    
    /**
     * Gets the current version of the given {@link VersionAwareAmetysObject}
     * @param versionable The {@link VersionAwareAmetysObject}
     * @return The current version
     */
    public String getCurrentVersion(VersionAwareAmetysObject versionable)
    {
        String currentRevision = versionable.getRevision();
        if (currentRevision != null)
        {
            return currentRevision;
        }
        
        return getNewestVersion(versionable);
    }
    
    /**
     * Gets the previous version of the given {@link VersionAwareAmetysObject}
     * @param versionable The {@link VersionAwareAmetysObject}
     * @return The previous version. Can be null.
     */
    public String getPreviousVersion(VersionAwareAmetysObject versionable)
    {
        String[] allRevisions = versionable.getAllRevisions();
        
        String currentRevision = versionable.getRevision();
        if (currentRevision != null)
        {
            for (int i = 0; i < allRevisions.length - 1; i++)
            {
                if (StringUtils.equals(allRevisions[i], currentRevision))
                {
                    if (i > 0)
                    {
                        return allRevisions[i - 1];
                    }
                    else
                    {
                        return null;
                    }
                }
            }
            return null;
        }
        else if (allRevisions.length > 2)
        {
            return allRevisions[allRevisions.length - 2];
        }
        else
        {
            return null;
        }
    }
    
    /**
     * Gets the last version of the given {@link VersionAwareAmetysObject} with the given label
     * @param versionable The {@link VersionAwareAmetysObject}
     * @param label The label
     * @return the last version
     */
    public String getLastVersionWithLabel(VersionAwareAmetysObject versionable, String label)
    {
        String[] allRevisions = versionable.getAllRevisions();
        String stopAtRevision = versionable.getRevision();
        for (int i = allRevisions.length - 1; i >= 0; i--)
        {
            String revision = allRevisions[i];
            String[] revisionLabels = versionable.getLabels(revision);
            if (ArrayUtils.contains(revisionLabels, label))
            {
                return revision;
            }
            if (StringUtils.equals(revision, stopAtRevision))
            {
                return null;
            }
        }
        
        return null;
    }
    
    /**
     * Compare two versions of the same content
     * @param <C> The type of the {@link VersionAwareAmetysObject} {@link Content}
     * @param versionableContentId The {@link Content} id to compare, which is {@link VersionAwareAmetysObject}
     * @param baseVersion The base version
     * @param version The version to be compared
     * @return The {@link ContentComparatorResult}
     * @throws AmetysRepositoryException repository exception
     * @throws IOException IO exception
     */
    public <C extends Content & VersionAwareAmetysObject> ContentComparatorResult compareVersions(String versionableContentId, String baseVersion, String version) throws AmetysRepositoryException, IOException
    {
        C baseContent = getContentVersion(versionableContentId, baseVersion);
        C targetContent = getContentVersion(versionableContentId, version);
        ContentComparatorResult comparatorResult = _contentComparator.compare(targetContent/* the 1st content is the target */, baseContent/* the 2nd is the base */);
        return comparatorResult;
    }
    
    /**
     * Gets the given version of the given {@link Content} od
     * @param <C> The type of the {@link VersionAwareAmetysObject} {@link Content}
     * @param contentId The {@link Content} id
     * @param version The version. Can be null for current version
     * @return the given version of the given {@link Content} od
     */
    public <C extends Content & VersionAwareAmetysObject> C getContentVersion(String contentId, String version)
    {
        C versionableContent = _ametysObjectResolver.resolveById(contentId);
        versionableContent.switchToRevision(version);
        return versionableContent;
    }
    
    /**
     * Gets the attribute changes (as {@link ModelItem}s) from the given {@link ContentComparatorChange}s (as a {@link ContentComparatorChange} path can be more detailed than a {@link ModelItem#getPath model item path})
     * <br>The false positive changes are also {@link #filterChanges filtered}
     * @param changes The {@link ContentComparatorChange}s
     * @return The changed {@link ModelItem}s
     */
    public Stream<ModelItem> getChangedModelItems(List<ContentComparatorChange> changes)
    {
        return _filterChanges(changes)
                .distinct() // some changes are different but are about the same ModelItem => remove those duplicates
                .map(attributeChange -> attributeChange._contentComparatorChange)
                .map(ContentComparatorChange::getModelItem);
    }
    
    /**
     * Gets the {@link ContentComparatorChange}s, with some of them filtered (such as {@link ResourceElementTypeHelper#LAST_MODIFICATION_DATE_IDENTIFIER someRichText/lastModified}),
     * because they can be considered as "false positive", from the given {@link ContentComparatorChange}s
     * @param changes The {@link ContentComparatorChange}s
     * @return The filtered changed {@link ContentComparatorChange}s
     */
    public Stream<ContentComparatorChange> filterChanges(List<ContentComparatorChange> changes)
    {
        return _filterChanges(changes)
                .map(attributeChange -> attributeChange._contentComparatorChange);
    }
    
    private Stream<AttributeModelChange> _filterChanges(List<ContentComparatorChange> changes)
    {
        return changes
                .stream()
                .map(LambdaUtils.wrap(contentComparatorChange -> _getAttributeModelChange(contentComparatorChange)))
                .filter(_attributeModelChangeFilters()); // filter out some changes, such as "someRichTextAttribute/lastModified"
    }

    private AttributeModelChange _getAttributeModelChange(ContentComparatorChange contentComparatorChange)
    {
        return new AttributeModelChange(contentComparatorChange);
    }
    
    private static Predicate<AttributeModelChange> _attributeModelChangeFilters()
    {
        return Predicates.compose(CompareVersionHelper::_filterContentComparatorChange, attributeModelChange -> attributeModelChange._contentComparatorChange);
    }
    
    private static boolean _filterContentComparatorChange(ContentComparatorChange contentComparatorChange)
    {
        return !_isLastModifiedOnRichText(contentComparatorChange); // hardcoded filter on "someRichTextAttribute/lastModified"
        // here we can add some other static filters...
    }
    
    private static boolean _isLastModifiedOnRichText(ContentComparatorChange contentComparatorChange)
    {
        ModelItem modelItem = contentComparatorChange.getModelItem();
        return ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID.equals(modelItem.getType().getId())
                && contentComparatorChange.getDetailDataPath().equals(ResourceElementTypeHelper.LAST_MODIFICATION_DATE_IDENTIFIER);
    }
    
    private static class AttributeModelChange
    {
        /** The change (is more detailed than on a specific attribute) given by {@link ContentComparator} */
        private final ContentComparatorChange _contentComparatorChange;
        /** The path of the involved model item */
        private final String _modelItemPath;

        AttributeModelChange(ContentComparatorChange change)
        {
            _contentComparatorChange = change;
            _modelItemPath = _contentComparatorChange.getModelItem().getPath();
        }
        
        @Override
        public int hashCode()
        {
            return Objects.hash(_modelItemPath);
        }

        // consider equals if it concerns the same modelItem (i.e. for a change about a/b/c and a/b/d, if the modelItem is 'a/b', it is the same as we are just interested in attribute change)
        @Override
        public boolean equals(Object obj)
        {
            if (this == obj)
            {
                return true;
            }
            if (obj == null)
            {
                return false;
            }
            if (getClass() != obj.getClass())
            {
                return false;
            }
            AttributeModelChange other = (AttributeModelChange) obj;
            return Objects.equals(_modelItemPath, other._modelItemPath);
        }
    }
}
