/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.version;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.cocoon.ProcessingException;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.content.compare.ContentComparatorResult;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.AbstractViewItemGroup;
import org.ametys.runtime.model.DefaultElementDefinition;
import org.ametys.runtime.model.DefinitionContext;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.SimpleViewItemGroup;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElement;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.ViewItemContainer;
import org.ametys.runtime.model.ViewItemGroup;

// Wrap another view, to be able to wrap its own ViewElement to modify the ViewElement#toJson
class CompareView implements ViewItemContainer
{
    private final View _wrapped;
    private final Set<String> _externalizableDataPaths;
    private boolean _showAllData;
    private List<String> _changesPaths;
    
    CompareView(View fromView, Set<String> externalizableDataPaths, boolean showAllData, ContentComparatorResult comparatorResult)
    {
        _wrapped = new View();
        _showAllData = showAllData;
        if (!_showAllData)
        {
            _changesPaths = comparatorResult.getChanges().stream().map(c -> c.getPath().replaceAll("\\[.*?\\]", "")).distinct().toList();
        }
        _populateViewItemContainerWithWrappedElements(fromView, _wrapped);
        _externalizableDataPaths = externalizableDataPaths;
    }
    
    private void _populateViewItemContainerWithWrappedElements(ViewItemContainer sourceOfCopy, ViewItemContainer toPopulate)
    {
        for (ViewItem viewItem : sourceOfCopy.getViewItems())
        {
            if (_showAllData || _containsChanges(viewItem))
            {
                toPopulate.addViewItem(_wrapViewElement(viewItem));
            }
        }
    }
    
    private ViewItem _wrapViewElement(ViewItem item)
    {
        if (item instanceof ViewElement)
        {
            return new CompareViewElement((ViewElement) item);
        }
        else if (item instanceof ModelViewItemGroup)
        {
            return new CompareModelViewItemGroup((ModelViewItemGroup) item);
        }
        else if (item instanceof SimpleViewItemGroup)
        {
            return new CompareSimpleViewItemGroup((SimpleViewItemGroup) item);
        }
        else
        {
            throw new IllegalArgumentException("Unexpected and unhandled ViewItem class '" + item.getClass() + "' to wrap it into a CompareView. Please contact your administrator.");
        }
    }

    private boolean _containsChanges(ViewItem viewItem)
    {
        if (viewItem instanceof ViewElement viewElement)
        {
            return _changesPaths.contains(viewElement.getDefinition().getPath())
                    || _changesPaths.stream().anyMatch(e -> e.startsWith(viewElement.getDefinition().getPath() + "/"))
                    || _changesPaths.stream().anyMatch(e -> viewElement.getDefinition().getPath().startsWith(e + "/"));
        }
        else if (viewItem instanceof ModelViewItemGroup modelViewItemGroup)
        {
            return _changesPaths.contains(modelViewItemGroup.getDefinition().getPath())
                    || _changesPaths.stream().anyMatch(e -> e.startsWith(modelViewItemGroup.getDefinition().getPath() + "/"))
                    || _changesPaths.stream().anyMatch(e -> modelViewItemGroup.getDefinition().getPath().startsWith(e + "/")); 
        }
        else if (viewItem instanceof SimpleViewItemGroup simpleViewItemGroup)
        {
            for (ViewItem subViewItem : simpleViewItemGroup.getViewItems())
            {
                if (_containsChanges(subViewItem))
                {
                    return true;
                }
            }
            return false;
        }
        else
        {
            throw new IllegalArgumentException("Unexpected and unhandled ViewItem class '" + viewItem.getClass() + "' to wrap it into a CompareView. Please contact your administrator.");
        }
    }
    
    @Override
    public List<ViewItem> getViewItems()
    {
        return _wrapped.getViewItems();
    }

    @Override
    public void addViewItem(ViewItem item)
    {
        throw new UnsupportedOperationException("Should not be called");
    }

    @Override
    public void insertViewItem(ViewItem item, int index)
    {
        throw new UnsupportedOperationException("Should not be called");
    }
    
    public boolean removeViewItem(ViewItem item)
    {
        throw new UnsupportedOperationException("Should not be called");
    }
    
    public void clear()
    {
        throw new UnsupportedOperationException("Should not be called");
    }
    
    Map<String, Object> toJSON() throws ProcessingException
    {
        return _wrapped.toJSON(DefinitionContext.newInstance());
    }
    
    // Wrap another group, to be able to wrap its own ViewElement to modify the ViewElement#toJson
    private abstract static class AbstractCompareViewItemGroup implements ViewItemGroup
    {
        protected AbstractViewItemGroup _wrapped;
        
        public Map<String, Object> toJSON(DefinitionContext context) throws ProcessingException
        {
            return _wrapped.toJSON(context);
        }
        
        public void toSAX(ContentHandler contentHandler, DefinitionContext context) throws SAXException
        {
            _wrapped.toSAX(contentHandler, context);
        }

        @Override
        public String getName()
        {
            return _wrapped.getName();
        }
        
        public I18nizableText getLabel()
        {
            return _wrapped.getLabel();
        }
        
        public I18nizableText getDescription()
        {
            return _wrapped.getDescription();
        }

        @Override
        public List<ViewItem> getViewItems()
        {
            return _wrapped.getViewItems();
        }
        
        public void setName(String name)
        {
            throw new UnsupportedOperationException("Should not be called");
        }
        
        public void setLabel(I18nizableText label)
        {
            throw new UnsupportedOperationException("Should not be called");
        }
        
        public void setDescription(I18nizableText description)
        {
            throw new UnsupportedOperationException("Should not be called");
        }
        
        public ViewItemAccessor getParent()
        {
            return _wrapped.getParent();
        }
        
        public void setParent(ViewItemAccessor parent)
        {
            // Do nothing, ignore parent setting
        }

        @Override
        public void addViewItem(ViewItem item)
        {
            throw new UnsupportedOperationException("Should not be called");
        }

        @Override
        public void insertViewItem(ViewItem item, int index)
        {
            throw new UnsupportedOperationException("Should not be called");
        }
        
        public boolean removeViewItem(ViewItem item)
        {
            throw new UnsupportedOperationException("Should not be called");
        }
        
        public void clear()
        {
            throw new UnsupportedOperationException("Should not be called");
        }
        
        @Override
        public void copyTo(ViewItem item)
        {
            throw new UnsupportedOperationException("Should not be called");
        }
        
        @Override
        public ViewItem createInstance()
        {
            throw new UnsupportedOperationException("Should not be called");
        }
        
        public boolean equals(Object obj, boolean checkDetails)
        {
            throw new UnsupportedOperationException("Should not be called");
        }

        @Override
        public String getRole()
        {
            return _wrapped.getRole();
        }

        @Override
        public void setRole(String role)
        {
            _wrapped.setRole(role);
        }
    }
    
    private class CompareModelViewItemGroup extends AbstractCompareViewItemGroup
    {
        CompareModelViewItemGroup(ModelViewItemGroup fromGroup)
        {
            var wrapped = new ModelViewItemGroup<>(); // empty constructor and not ModelViewItemGroup.of(fromGroup.getDefinition()), in order to avoid duplication of view items
            wrapped.setDefinition(fromGroup.getDefinition());
            _wrapped = wrapped;
            _populateViewItemContainerWithWrappedElements(fromGroup, _wrapped);
        }
    }
    
    private class CompareSimpleViewItemGroup extends AbstractCompareViewItemGroup
    {
        CompareSimpleViewItemGroup(SimpleViewItemGroup fromGroup)
        {
            var wrapped = new SimpleViewItemGroup();
            wrapped.setName(fromGroup.getName());
            wrapped.setLabel(fromGroup.getLabel());
            wrapped.setDescription(fromGroup.getDescription());
            wrapped.setRole(fromGroup.getRole());
            _wrapped = wrapped;
            _populateViewItemContainerWithWrappedElements(fromGroup, _wrapped);
        }
    }
    
    // Extends ViewElement, so as to modify the ElementDefinition#toJson
    private class CompareViewElement extends ViewElement
    {
        CompareViewElement(ViewElement wrapped)
        {
            @SuppressWarnings("unchecked")
            CompareElementDefinition compareElementDefinition = new CompareElementDefinition(wrapped.getDefinition());
            setDefinition(compareElementDefinition);
        }
    }
    
    // here we are able to modify the ElementDefinition#toJson as we want
    private class CompareElementDefinition<T> extends DefaultElementDefinition<T>
    {
        CompareElementDefinition(ElementDefinition<T> definitionToCopy)
        {
            super(definitionToCopy);
        }
        
        /*
         * Here is the business logic for what we created all those wrappers
         */
        @Override
        protected Map<String, Object> _toJSON(DefinitionContext context) throws ProcessingException
        {
            Map<String, Object> elementDefinitionJson = new HashMap<>(super._toJSON(context));
            
            elementDefinitionJson.remove("disableCondition");
            
            @SuppressWarnings("unchecked")
            Map<String, I18nizableText> widgetParams = (Map<String, I18nizableText>) elementDefinitionJson.computeIfAbsent("widget-params", __ -> new HashMap<>());
            String widgetToWrap = (String) elementDefinitionJson.get("widget");
            widgetParams.put("comparator-wrapped-widget", new I18nizableText(widgetToWrap));
            elementDefinitionJson.put("widget", "edition.comparison");
            _handleExternal(widgetToWrap, widgetParams);
            
            return elementDefinitionJson;
        }
        
        private void _handleExternal(String widgetToWrap, Map<String, I18nizableText> widgetParams)
        {
            if (_externalizableDataPaths.contains(getPath()))
            {
                // Wrap the widget as in org.ametys.cms.contenttype.AttributeDefinition#toJSON
                // Should be more generic
                widgetParams.put("comparator-wrapped-widget", new I18nizableText("edition.externalizable"));
                widgetParams.put("wrapped-widget", new I18nizableText(widgetToWrap)); // wrapped widget for externalizable (so, it is wrapped twice)
            }
        }
    }
}
