/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.ArrayList;
import java.util.List;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.SourceResolver;

import org.ametys.core.ui.ClientSideElement.ScriptFile;
import org.ametys.plugins.core.ui.util.ConfigurationHelper;
import org.ametys.plugins.explorer.dublincore.DublinCoreMetadataProvider;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.ModelHelper.ConfigurationAndPluginName;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;
import org.ametys.runtime.plugin.component.PluginAware;

/**
 * This abstract class represents a content type descriptor
 */
public abstract class AbstractContentTypeDescriptor extends AbstractLogEnabled implements ContentTypeDescriptor, Configurable, PluginAware, Serviceable
{
    /** Plugin name. */
    protected String _pluginName;
    
    /** Content type id. */
    protected String _id;
    /** Label. */
    protected I18nizableText _label;
    /** Description. */
    protected I18nizableText _description;
    /** Default title. */
    protected I18nizableText _defaultTitle;
    /** Category. */
    protected I18nizableText _category;
    /** Glyph icon */
    protected String _iconGlyph;
    /** Icon decorator */
    protected String _iconDecorator;
    /** Small icon URI 16x16. */
    protected String _smallIcon;
    /** Medium icon URI 32x32. */
    protected String _mediumIcon;
    /** Large icon URI 48x48. */
    protected String _largeIcon;
    
    /** The CSS files */
    protected List<ScriptFile> _cssFiles;
    
    /** This content-type's supertype ids (can be empty). */
    protected String[] _superTypeIds;
    
    /** The content type extension point. */
    protected ContentTypeExtensionPoint _cTypeEP;
    /** The content types helper */
    protected ContentTypesHelper _contentTypesHelper;
    /** The content types parser helper */
    protected ContentTypesParserHelper _contentTypesParserHelper;
    /** The source resolver */
    protected SourceResolver _srcResolver;
    /** The DublinCore metadata provider */
    protected DublinCoreMetadataProvider _dcProvider;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _contentTypesHelper = (ContentTypesHelper) smanager.lookup(ContentTypesHelper.ROLE);
        _contentTypesParserHelper = (ContentTypesParserHelper) smanager.lookup(ContentTypesParserHelper.ROLE);
        _srcResolver = (SourceResolver) smanager.lookup(SourceResolver.ROLE);
        _dcProvider = (DublinCoreMetadataProvider) smanager.lookup(DublinCoreMetadataProvider.ROLE);
        _cTypeEP = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    @Override
    public void setPluginInfo(String pluginName, String featureName, String id)
    {
        _pluginName = pluginName;
        _id = id;
    }
    
    /**
     * Get the root configuration
     * @param configuration The configuration 
     * @return The main configuration
     */
    protected abstract Configuration getRootConfiguration (Configuration configuration);
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        Configuration rootConfiguration = getRootConfiguration(configuration);
        _configureSuperTypes(rootConfiguration);
        
        _configureLabels(rootConfiguration);
        _configureIcons(rootConfiguration);
        
        _configureCSSFiles(rootConfiguration);
    }
    
    /**
     * Configures the super types
     * @param mainConfig The main configuration
     */
    protected void _configureSuperTypes (Configuration mainConfig)
    {
        String extendedCTypes = mainConfig.getAttribute("extends", null);
        List<String> superTypeIds = new ArrayList<>();
        if (extendedCTypes != null)
        {
            String[] superIds = extendedCTypes.split(",");
            for (String id : superIds)
            {
                superTypeIds.add(StringUtils.trim(id));
            }
        }
        
        _superTypeIds = superTypeIds.toArray(new String[superTypeIds.size()]);
    }
    
    /**
     * Configures the labels
     * @param mainConfig The main configuration
     * @throws ConfigurationException if an error occurred
     */
    protected void _configureLabels (Configuration mainConfig)  throws ConfigurationException
    {
        ConfigurationAndPluginName mainConfigurationAndPluginName = new ConfigurationAndPluginName(mainConfig, getPluginName());
        _label = ModelHelper.parseI18nizableText(mainConfigurationAndPluginName, "label");
        _description = ModelHelper.parseI18nizableText(mainConfigurationAndPluginName, "description");
        _defaultTitle = ModelHelper.parseI18nizableText(mainConfigurationAndPluginName, "default-title");
        _category = ModelHelper.parseI18nizableText(mainConfigurationAndPluginName, "category");
    }
    
    /**
     * Configures the icons
     * @param mainConfig The main configuration
     * @throws ConfigurationException if an error occurred
     */
    protected void _configureIcons (Configuration mainConfig)  throws ConfigurationException
    {
        _iconGlyph = mainConfig.getChild("icons").getChild("glyph").getValue(null);
        _iconDecorator = mainConfig.getChild("icons").getChild("decorator").getValue(null);
        _smallIcon = _contentTypesParserHelper.parseIcon(this, mainConfig.getChild("icons"), "small");
        _mediumIcon = _contentTypesParserHelper.parseIcon(this, mainConfig.getChild("icons"), "medium");
        _largeIcon = _contentTypesParserHelper.parseIcon(this, mainConfig.getChild("icons"), "large");
    }
    
    
    /**
     * Configure the CSS to import
     * @param configuration The imports configuration
     * @throws ConfigurationException If an error occurs
     */
    protected void _configureCSSFiles(Configuration configuration) throws ConfigurationException
    {
        _cssFiles = ConfigurationHelper.parsePluginResourceList(configuration.getChild("css"), getPluginName(), getLogger());
    }
    
    @Override
    public String getPluginName()
    {
        return _pluginName;
    }

    @Override
    public String getId()
    {
        return _id;
    }
    
    @Override
    public I18nizableText getLabel()
    {
        return _label;
    }
    
    @Override
    public I18nizableText getDescription()
    {
        return _description;
    }
    
    @Override
    public I18nizableText getCategory()
    {
        return _category;
    }

    @Override
    public I18nizableText getDefaultTitle()
    {
        return _defaultTitle;
    }
    
    @Override
    public String getIconGlyph()
    {
        return _iconGlyph;
    }
    
    @Override
    public String getIconDecorator()
    {
        return _iconDecorator;
    }
    
    @Override
    public String getSmallIcon()
    {
        return _smallIcon;
    }
    
    @Override
    public String getMediumIcon()
    {
        return _mediumIcon;
    }
    
    @Override
    public String getLargeIcon()
    {
        return _largeIcon;
    }
    
    @Override
    public String[] getSupertypeIds()
    {
        return _superTypeIds;
    }
    
    public List<ScriptFile> getCSSFiles()
    {
        return _cssFiles;
    }
}

