/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.Collection;
import java.util.List;
import java.util.Optional;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.runtime.model.AbstractViewParser;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.ModelHelper.ConfigurationAndPluginName;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemAccessor;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElement;
import org.ametys.runtime.model.ViewElementAccessor;
import org.ametys.runtime.model.ViewHelper;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;

/**
 * Abstract component that parses view's configuration for content types
 */
public abstract class AbstractContentTypeViewParser extends AbstractViewParser implements Component, Serviceable
{
    /** The content type extension point */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    /** The content types helper */
    protected ContentTypesHelper _contentTypesHelper;
    /** The system property extension point */
    protected SystemPropertyExtensionPoint _systemPropertyExtensionPoint;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
        _systemPropertyExtensionPoint = (SystemPropertyExtensionPoint) manager.lookup(SystemPropertyExtensionPoint.ROLE);
    }
    
    @Override
    protected ModelItem _getModelItem(ConfigurationAndPluginName itemConfiguration, String modelItemName, Collection< ? extends ModelItemAccessor> parents) throws ConfigurationException
    {
        return _systemPropertyExtensionPoint.hasExtension(modelItemName)
                ? _systemPropertyExtensionPoint.getExtension(modelItemName)
                : super._getModelItem(itemConfiguration, modelItemName, parents);
    }
    
    @Override
    protected ViewElement _parseViewElement(ConfigurationAndPluginName itemConfiguration, ElementDefinition definition, View referenceView, boolean override) throws ConfigurationException
    {
        return definition instanceof ContentAttributeDefinition contentAttributeDefinition
                ? _parseContentAttributeViewItem(itemConfiguration, contentAttributeDefinition, referenceView, override)
                : super._parseViewElement(itemConfiguration, definition, referenceView, override);
    }
    
    /**
     * Parses the view item referencing an attribute of type content 
     * @param itemConfiguration configuration of the view item
     * @param definition definition of the attribute
     * @param referenceView view that references the item
     * @param override <code>true</code> if the configuration is an override, <code>false</code> otherwise
     * @return the view item
     * @throws ConfigurationException if the configuration is not valid
     */
    protected ViewElement _parseContentAttributeViewItem(ConfigurationAndPluginName itemConfiguration, ContentAttributeDefinition definition, View referenceView, boolean override) throws ConfigurationException
    {
        ViewElementAccessor viewItem = (ViewElementAccessor) _createModelViewItemInstance(definition);

        // Parse children
        for (Configuration childConfiguration : itemConfiguration.configuration().getChildren())
        {
            String childConfigurationName = childConfiguration.getName();
            
            if ("view".equals(childConfigurationName))
            {
                String viewName = childConfiguration.getAttribute("name");
                _parseViewReference(viewName, viewItem, definition);
            }
            else
            {
                _parseViewItemAccessorChild(new ConfigurationAndPluginName(childConfiguration, itemConfiguration.pluginName()), viewItem, definition, referenceView, override);
            }
        }

        return viewItem;
    }
    
    @Override
    protected void _parseViewReference(String viewName, ViewItemAccessor viewItemAccessor, ModelItemAccessor modelItemAccessor) throws ConfigurationException
    {
        ContentType contentType = Optional.of(modelItemAccessor)
                                          .filter(ContentAttributeDefinition.class::isInstance)
                                          .map(ContentAttributeDefinition.class::cast)
                                          .map(ContentAttributeDefinition::getContentTypeId)
                                          .map(_contentTypeExtensionPoint::getExtension)
                                          .orElse(null);

        if (contentType != null)
        {
            View referencedView = contentType.getView(viewName);
        
            if (referencedView != null)
            {
                List<ViewItem> referencedViewItems = ViewHelper.copyViewItems(referencedView.getViewItems(), this::_createViewItemCopyInstanceForReferencedView);
                viewItemAccessor.addViewItems(referencedViewItems);
            }
            else
            {
                String modelItemAccessorName = modelItemAccessor instanceof ModelItem modelItem ? modelItem.getName() : ((Model) modelItemAccessor).getId();
                throw new ConfigurationException("The view '" + viewName + "' does not exist in content type '" + contentType.getId() + "' referenced by the attribute named '" + modelItemAccessorName + "'.");
            }
        }
        else
        {
            String modelItemAccessorName = modelItemAccessor instanceof ModelItem modelItem ? modelItem.getName() : ((Model) modelItemAccessor).getId();
            throw new ConfigurationException("The attribute '" + modelItemAccessorName + "' doesn't reference any content type. It is not possible to add the items of the view '" + viewName + "'.");
        }
    }
    
    /**
     * Creates a {@link ViewItem} instance, that will be the copy of an existing view item in a referenced view
     * @param viewItem the view item that will be copied
     * @return the created {@link ViewItem} instance
     */
    protected ViewItem _createViewItemCopyInstanceForReferencedView (ViewItem viewItem)
    {
        return ViewHelper.createViewItemInstance(viewItem);
    }
}
