/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.xml.XMLUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.model.restrictions.ContentRestrictedModelItemHelper;
import org.ametys.cms.model.restrictions.RestrictedModelItem;
import org.ametys.cms.model.restrictions.Restrictions;
import org.ametys.cms.repository.Content;
import org.ametys.plugins.repository.data.external.ExternalizableDataProviderExtensionPoint;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.DefaultElementDefinition;
import org.ametys.runtime.model.DefinitionContext;
import org.ametys.runtime.model.ModelItem;

/**
 * The definition of a content type attribute
 * @param <T> Type of the attribute value
 */
public class AttributeDefinition<T> extends DefaultElementDefinition<T> implements RestrictedModelItem<Content>
{
    private ContentRestrictedModelItemHelper _restrictedModelItemHelper;
    private ExternalizableDataProviderExtensionPoint _externalizableDataProviderExtensionPoint;

    private Restrictions _restrictions;
    
    /**
     * Set the restrictions for the attribute
     * @param restrictions the restrictions to set
     */
    public void setRestrictions(Restrictions restrictions)
    {
        _restrictions = restrictions;
    }

    public boolean canRead(Content content)
    {
        ContentRestrictedModelItemHelper restrictedModelItemHelper = _getRestrictedModelItemHelper();
        ContentType contentType = (ContentType) getModel();
        return restrictedModelItemHelper.canRead(content, this, _restrictions) && contentType != null && contentType.canRead(content, this);
    }

    public boolean canWrite(Content content)
    {
        ContentRestrictedModelItemHelper restrictedModelItemHelper = _getRestrictedModelItemHelper();
        ContentType contentType = (ContentType) getModel();
        return restrictedModelItemHelper.canWrite(content, this, _restrictions) && contentType != null && contentType.canWrite(content, this);
    }
    
    /**
     * Retrieves the helper for {@link ModelItem}s with restrictions on contents 
     * @return the {@link ContentRestrictedModelItemHelper}
     */
    protected ContentRestrictedModelItemHelper _getRestrictedModelItemHelper()
    {
        if (_restrictedModelItemHelper == null)
        {
            try
            {
                _restrictedModelItemHelper = (ContentRestrictedModelItemHelper) __serviceManager.lookup(ContentRestrictedModelItemHelper.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException("Unable to lookup after the restricted model item helper component", e);
            }
        }
        
        return _restrictedModelItemHelper;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    protected Map<String, Object> _toJSON(DefinitionContext context) throws ProcessingException
    {
        Map<String, Object> result = super._toJSON(context);
        
        Content content = context.getObject()
                .filter(Content.class::isInstance)
                .map(Content.class::cast)
                .orElse(null);
        
        if (content != null)
        {
            if (context.isEdition() && _getExternalizableDataProviderExtensionPoint().isDataExternalizable(content, this))
            {
                // Wrap the widget
                result.put("widget", "edition.externalizable");

                if (getWidget() != null)
                {
                    Map<String, I18nizableText> widgetParameters = new HashMap<>();
                    if (result.containsKey("widget-params"))
                    {
                        widgetParameters = (Map<String, I18nizableText>) result.get("widget-params");
                    }

                    widgetParameters.put("wrapped-widget", new I18nizableText(getWidget()));

                    if (!widgetParameters.isEmpty())
                    {
                        result.put("widget-params", widgetParameters);
                    }
                }
            }
        }
        
        if (!canWrite(content))
        {
            result.put("can-not-write", true);
        }
        
        return result;
    }
    
    @Override
    protected boolean _shouldJSONBeEmpty(DefinitionContext context)
    {
        return context.getObject()
                .filter(Content.class::isInstance)
                .map(Content.class::cast)
                .map(content -> !canRead(content))
                .orElse(false);
    }
    
    @Override
    public void toSAX(ContentHandler contentHandler, DefinitionContext context) throws SAXException
    {
        super.toSAX(contentHandler, context);

        Content content = context.getObject()
                .filter(Content.class::isInstance)
                .map(Content.class::cast)
                .orElse(null);

        if (!canWrite(content))
        {
            XMLUtils.createElement(contentHandler, "can-not-write", String.valueOf(true));
        }
    }
    
    @Override
    protected void _widgetToSAX(ContentHandler contentHandler, DefinitionContext context) throws SAXException
    {
        Content content = context.getObject()
                .filter(Content.class::isInstance)
                .map(Content.class::cast)
                .orElse(null);
        
        if (content != null && context.isEdition() && _getExternalizableDataProviderExtensionPoint().isDataExternalizable(content, this))
        {
            // Wrap the widget
            XMLUtils.createElement(contentHandler, "widget", "edition.externalizable");

            if (getWidget() != null)
            {
                XMLUtils.startElement(contentHandler, "widget-params");

                Map<String, I18nizableText> widgetParameters = getWidgetParameters();
                if (widgetParameters != null && !widgetParameters.isEmpty())
                {
                    for (Map.Entry<String, I18nizableText> param : widgetParameters.entrySet())
                    {
                        _widgetParameterToSAX(contentHandler, param.getKey(), param.getValue(), context);
                    }
                }

                _widgetParameterToSAX(contentHandler, "wrapped-widget", new I18nizableText(getWidget()), context);

                XMLUtils.endElement(contentHandler, "widget-params");
            }
        }
        else
        {
            super._widgetToSAX(contentHandler, context);
        }
    }
    
    /**
     * Retrieves the {@link ExternalizableDataProviderExtensionPoint} 
     * @return the {@link ExternalizableDataProviderExtensionPoint}
     */
    protected ExternalizableDataProviderExtensionPoint _getExternalizableDataProviderExtensionPoint()
    {
        if (_externalizableDataProviderExtensionPoint == null)
        {
            try
            {
                _externalizableDataProviderExtensionPoint = (ExternalizableDataProviderExtensionPoint) __serviceManager.lookup(ExternalizableDataProviderExtensionPoint.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException("Unable to lookup after the externalizable data provider extension point", e);
            }
        }
        
        return _externalizableDataProviderExtensionPoint;
    }
}
