/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.cocoon.ProcessingException;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.data.ContentValue;
import org.ametys.cms.model.ContentElementDefinition;
import org.ametys.cms.repository.ContentQueryHelper;
import org.ametys.cms.repository.ContentTypeExpression;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.core.util.XMLUtils;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectIterator;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.query.expression.AndExpression;
import org.ametys.plugins.repository.query.expression.DoubleExpression;
import org.ametys.plugins.repository.query.expression.Expression;
import org.ametys.plugins.repository.query.expression.Expression.Operator;
import org.ametys.plugins.repository.query.expression.LongExpression;
import org.ametys.plugins.repository.query.expression.StringExpression;
import org.ametys.runtime.model.DefinitionContext;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.type.ElementType;
import org.ametys.runtime.model.type.ModelItemTypeConstants;

/**
 * Definition of content types attributes of type Content
 */
public class ContentAttributeDefinition extends AttributeDefinition<ContentValue> implements ContentElementDefinition
{
    /** attribute type for default values */
    public static final String ATTRIBUTE_DEFAULT_VALUE_TYPE = "attribute";
    
    private ContentTypesHelper _contentTypesHelper;
    private ContentTypeExtensionPoint _contentTypeExtensionPoint;
    
    private String _contentTypeId;
    private String _invertRelationPath;
    private boolean _forceInvert;
    
    private boolean _isDefaultValueAlreadyChecked;
    
    /**
     * Definition's constructor
     * @param contentTypeExtensionPoint the content type extension point
     * @param contentAttributeTypeExtensionPoint the content attribute type extension point
     */
    public ContentAttributeDefinition(ContentTypeExtensionPoint contentTypeExtensionPoint, ContentTypesHelper contentAttributeTypeExtensionPoint)
    {
        _contentTypeExtensionPoint = contentTypeExtensionPoint;
        _contentTypesHelper = contentAttributeTypeExtensionPoint;
    }
    
    public String getContentTypeId()
    {
        return _contentTypeId;
    }
    
    public void setContentTypeId(String contentTypeId)
    {
        _contentTypeId = contentTypeId;
    }
    
    /**
     * Checks the default value of this definition
     * @throws ConfigurationException if there is a problem with default value declaration
     */
    @SuppressWarnings("unchecked")
    public void checkDefaultValue() throws ConfigurationException
    {
        if (!_isDefaultValueAlreadyChecked)
        {
            if (_getParsedDefaultValues() != null)
            {
                List<Pair<String, Object>> parsedDefaultValues = new ArrayList<>();
                for (Pair<String, Object> parsedDefaultValue : _getParsedDefaultValues())
                {
                    String defaultValueType = parsedDefaultValue.getLeft();
                    if (ATTRIBUTE_DEFAULT_VALUE_TYPE.equals(defaultValueType))
                    {
                        assert parsedDefaultValue.getRight() instanceof Pair;
                        Pair<String, Configuration> defaultValue = (Pair<String, Configuration>) parsedDefaultValue.getRight();
                        String attributeName = defaultValue.getLeft();
                        Configuration defaultValueConfiguration = defaultValue.getRight();
        
                        Object parsedAttributeDefaultValue = _checkAndParseAttributeDefaultValue(attributeName, defaultValueConfiguration);
                        parsedDefaultValues.add(new ImmutablePair<>(defaultValueType, new ImmutablePair<>(attributeName, parsedAttributeDefaultValue)));
                    }
                    else
                    {
                        parsedDefaultValues.add(parsedDefaultValue);
                    }
                }
                
                setParsedDefaultValues(parsedDefaultValues);
            }
            
            _isDefaultValueAlreadyChecked = true;
        }
    }
    
    private Object _checkAndParseAttributeDefaultValue(String attributeName, Configuration defaultValueConfiguration) throws ConfigurationException
    {
        if (StringUtils.isEmpty(_contentTypeId))
        {
            throw new ConfigurationException("The type '" + ATTRIBUTE_DEFAULT_VALUE_TYPE + "' is not available for the default value of item '" + getPath() + " (content)': this item does not specify a content type.", defaultValueConfiguration);
        }
        else
        {
            ContentType contentType = _contentTypeExtensionPoint.getExtension(_contentTypeId);
            if (contentType.hasModelItem(attributeName)  && contentType.getModelItem(attributeName) instanceof ElementDefinition)
            {
                ElementDefinition attribute = (ElementDefinition) contentType.getModelItem(attributeName);
                ElementType attributeType = attribute.getType();
                if (ModelItemTypeConstants.STRING_TYPE_ID.equals(attributeType.getId()) || ModelItemTypeConstants.LONG_TYPE_ID.equals(attributeType.getId()) || ModelItemTypeConstants.DOUBLE_TYPE_ID.equals(attributeType.getId()))
                {
                    // Parse the value of the attribute that must 
                    return attributeType.parseConfiguration(defaultValueConfiguration);
                }
                else
                {
                    throw new ConfigurationException("The type '" + ATTRIBUTE_DEFAULT_VALUE_TYPE + "' supports attributes of type " + ModelItemTypeConstants.STRING_TYPE_ID + ", " + ModelItemTypeConstants.LONG_TYPE_ID + " and " + ModelItemTypeConstants.DOUBLE_TYPE_ID + ". The attribute named '" + attributeName + "' is a " + attributeType.getId() + ", so it is not supported.", defaultValueConfiguration);
                }
            }
            else
            {
                throw new ConfigurationException("The content type '" + _contentTypeId + "' has no attribute with named '" + attributeName + "'", defaultValueConfiguration);
            }
        }
    }
    
    @Override
    protected List<ContentValue> _getDefaultValues(String defaultValueType, Object parsedDefaultValue)
    {
        if (ATTRIBUTE_DEFAULT_VALUE_TYPE.equals(defaultValueType))
        {
            // Launch a repository query to get the corresponding content 

            // Create the expression on the content type 
            Expression contentTypeExpression = new ContentTypeExpression(Operator.EQ, _contentTypeId);

            // Create the expression on the attribute, according to its type
            @SuppressWarnings("unchecked")
            Pair<String, Configuration> defaultValue = (Pair<String, Configuration>) parsedDefaultValue;
            String attributeName = defaultValue.getLeft();
            Object value = defaultValue.getRight();

            ContentType contentType = _contentTypeExtensionPoint.getExtension(_contentTypeId);
            ElementType attributeType = ((ElementDefinition) contentType.getModelItem(attributeName)).getType();

            Expression attributeExpression;
            if (ModelItemTypeConstants.LONG_TYPE_ID.equals(attributeType.getId()))
            {
                attributeExpression = new LongExpression(attributeName, Operator.EQ, (Long) value);
            }
            else if (ModelItemTypeConstants.DOUBLE_TYPE_ID.equals(attributeType.getId()))
            {
                attributeExpression = new DoubleExpression(attributeName, Operator.EQ, (Double) value);
            }
            else
            {
                attributeExpression = new StringExpression(attributeName, Operator.EQ, (String) value);
            }
            
            // Launch the repository query
            Expression expression = new AndExpression(contentTypeExpression, attributeExpression);
            String xPathQuery = ContentQueryHelper.getContentXPathQuery(expression);
            
            try
            {
                AmetysObjectResolver resolver = (AmetysObjectResolver) __serviceManager.lookup(AmetysObjectResolver.ROLE);
                AmetysObjectIterable<ModifiableContent> contents = resolver.query(xPathQuery);
                AmetysObjectIterator<ModifiableContent> iterator = contents.iterator();
                
                return iterator.hasNext() ? List.of(new ContentValue(iterator.next())) : List.of();
            }
            catch (ServiceException e)
            {
                throw new RuntimeException("Unable to lookup after the ametys object resolver", e);
            }
        }
        else
        {
            return super._getDefaultValues(defaultValueType, parsedDefaultValue);
        }
    }
    
    /**
     * Get the element's mutual relationship path.
     * @return the element's mutual relationship path.
     */
    public String getInvertRelationPath()
    {
        return _invertRelationPath;
    }
    
    /**
     * Set the element's mutual relationship path.
     * @param invertRelationPath the element's mutual relationship path, separated by slashes.
     */
    public void setInvertRelationPath(String invertRelationPath)
    {
        _invertRelationPath = invertRelationPath;
    }
    
    /**
     * Force mutual relationship regardless of user's rights (only applicable for an element with invert relation path).
     * @param force true to force mutual relationship regardless of user's rights
     */
    public void setForceInvert (boolean force)
    {
        _forceInvert = force;
    }
    
    /**
     * Returns true if mutual relationship should be set regardless of user's rights
     * @return true if mutual relationship should be set regardless of user's rights
     */
    public boolean getForceInvert ()
    {
        return _forceInvert;
    }
    
    public Collection< ? extends ModelItem> getModelItems()
    {
        if (_contentTypeId != null && _contentTypeExtensionPoint.hasExtension(_contentTypeId))
        {
            ContentType contentType = _contentTypeExtensionPoint.getExtension(_contentTypeId);
            return contentType.getModelItems();
        }
        else
        {
            return Collections.singleton(_contentTypesHelper.getTitleAttributeDefinition());
        }
    }
    
    @Override
    protected Map<String, Object> _toJSON(DefinitionContext context) throws ProcessingException
    {
        Map<String, Object> result = super._toJSON(context);
        result.put("contentType", _contentTypeId);
        result.put("canCreate", _contentTypesHelper.hasRight(_contentTypeId));
        return result;
    }
    
    @Override
    public void toSAX(ContentHandler contentHandler, DefinitionContext context) throws SAXException
    {
        super.toSAX(contentHandler, context);
        XMLUtils.createElementIfNotNull(contentHandler, "contentType", _contentTypeId);
    }
}
