/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.ImmutablePair;

import org.ametys.cms.contenttype.DefaultContentType.AnnotableDefinition;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.model.ElementDefinitionParser;
import org.ametys.cms.model.restrictions.ContentRestrictedModelItemHelper;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Enumerator;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemGroup;
import org.ametys.runtime.model.type.ModelItemType;
import org.ametys.runtime.parameter.Validator;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * This class parses the content attributes definition
 */
public class ContentAttributeDefinitionParser extends ElementDefinitionParser
{
    /** Pattern for annotation names */
    protected static Pattern __annotationNamePattern;
    
    /** the content type extension point */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    
    /** the content types helper */
    protected ContentTypesHelper _contentTypesHelper;
    
    /**
     * Creates a content attribute definition parser.
     * @param contentAttributeTypeExtensionPoint the extension point to use to get available element types
     * @param enumeratorManager the enumerator component manager.
     * @param validatorManager the validator component manager.
     */
    public ContentAttributeDefinitionParser(ModelItemTypeExtensionPoint contentAttributeTypeExtensionPoint, ThreadSafeComponentManager<Enumerator> enumeratorManager, ThreadSafeComponentManager<Validator> validatorManager)
    {
        super(contentAttributeTypeExtensionPoint, enumeratorManager, validatorManager);
    }

    @Override
    @SuppressWarnings("unchecked")
    public <T extends ModelItem> T parse(ServiceManager serviceManager, String pluginName, String catalog, Configuration definitionConfig, Model model, ModelItemGroup parent) throws ConfigurationException
    {
        try
        {
            _contentTypesHelper = (ContentTypesHelper) serviceManager.lookup(ContentTypesHelper.ROLE);
            _contentTypeExtensionPoint = (ContentTypeExtensionPoint) serviceManager.lookup(ContentTypeExtensionPoint.ROLE);
        }
        catch (ServiceException e)
        {
            throw new ConfigurationException("Unable to parse the attribute '" + _parseName(definitionConfig) + "' in model'" + model.getId() + "'.", definitionConfig, e);
        }
        
        AttributeDefinition definition = (AttributeDefinition) super.parse(serviceManager, pluginName, catalog, definitionConfig, model, parent);
        
        try
        {
            ContentRestrictedModelItemHelper restrictedModelItemHelper = (ContentRestrictedModelItemHelper) serviceManager.lookup(ContentRestrictedModelItemHelper.ROLE);
            definition.setRestrictions(restrictedModelItemHelper._parseRestrictions(definitionConfig));
        }
        catch (ServiceException e)
        {
            throw new ConfigurationException("Unable to resolve restrictions on the element '" + definition.getName() + "'.", e);
        }
        
        if (definition instanceof AnnotableDefinition annotableDefinition)
        {
            annotableDefinition.setSemanticAnnotations(_parseDefinitionWithAnnotations(catalog, definitionConfig));
        }
        
        if (definition instanceof ContentAttributeDefinition contentAttributeDefinition)
        {
            String invertRelationPath = _parseInvertRelationPath(definitionConfig);
            if (StringUtils.isNotEmpty(invertRelationPath))
            {
                contentAttributeDefinition.setInvertRelationPath(invertRelationPath);
                contentAttributeDefinition.setForceInvert(_parseForceInvert(definitionConfig));
            }
        }

        
        return (T) definition;
    }

    @Override
    protected AttributeDefinition _createModelItem(Configuration definitionConfig) throws ConfigurationException
    {
        ModelItemType type = _parseType(definitionConfig);
        if (ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID.equals(type.getId()))
        {
            return new ContentAttributeDefinition(_contentTypeExtensionPoint, _contentTypesHelper);
        }
        else if (ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID.equals(type.getId()))
        {
            return new RichTextAttributeDefinition();
        }
        else
        {
            return new AttributeDefinition<>();            
        }
    }
    
    @Override
    protected String _parseName(Configuration itemConfig) throws ConfigurationException
    {
        String itemName = itemConfig.getAttribute(_getNameConfigurationAttribute());
        
        if (!itemName.matches("^[a-zA-Z]((?!__)[a-zA-Z0-9_-])*$"))
        {
            throw new ConfigurationException("Invalid model item name: " + itemName + ". The item name must start with a letter and must contain only letters, digits, underscore or dash characters.", itemConfig);
        }
        
        return itemName;
    }
    
    @Override
    protected Object _parseDefaultValue(Configuration defaultValueConfig, ElementDefinition definition, String defaultValueType) throws ConfigurationException
    {
        if (definition instanceof ContentAttributeDefinition && ContentAttributeDefinition.ATTRIBUTE_DEFAULT_VALUE_TYPE.equals(defaultValueType))
        {
            String attributeName = defaultValueConfig.getAttribute("name", null);
            if (attributeName == null)
            {
                throw new ConfigurationException("The type '" + defaultValueType + " needs to specify the name of the attribute as an XML attribute of the default-value tag", defaultValueConfig);
            }
            else
            {
                if (attributeName.contains(ModelItem.ITEM_PATH_SEPARATOR))
                {
                    throw new ConfigurationException("The type '" + defaultValueType + " can't accept the attribute '" + attributeName + "' because it is in a composite or a repeater", defaultValueConfig);
                }
                else
                {
                    return new ImmutablePair<>(attributeName, defaultValueConfig);
                }
            }
        }
        
        return super._parseDefaultValue(defaultValueConfig, definition, defaultValueType);
    }
    
    /**
     * Parses the semantic annotations of the model item
     * @param catalog the catalog.
     * @param itemConfig the model item configuration to use.
     * @return the list of the declared annotations.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected List<SemanticAnnotation> _parseDefinitionWithAnnotations(String catalog, Configuration itemConfig) throws ConfigurationException
    {            
        Configuration annotationsConfiguration = itemConfig.getChild("annotations");
        List<SemanticAnnotation> annotations = new ArrayList<>();
        
        for (Configuration annotationConfig : annotationsConfiguration.getChildren("annotation"))
        {
            String id = annotationConfig.getAttribute("name");
            
            if (!_getAnnotationNamePattern().matcher(id).matches())
            {
                throw new ConfigurationException("Invalid annonation name '" + id + "'. This value is not permitted: only [a-zA-Z][a-zA-Z0-9-_]* are allowed.");
            }
            
            I18nizableText label = _parseI18nizableText(annotationConfig, catalog, "label");
            I18nizableText description = _parseI18nizableText(annotationConfig, catalog, "description");
            annotations.add(new SemanticAnnotation(id, label, description));
        }
        
        return annotations;
    }
    
    /**
     * Get the annotation name pattern to test validity.
     * @return The annotation name pattern.
     */
    protected Pattern _getAnnotationNamePattern()
    {
        if (__annotationNamePattern == null)
        {
            // [a-zA-Z][a-zA-Z0-9_]*
            __annotationNamePattern = Pattern.compile("[a-z][a-z0-9-_]*", Pattern.CASE_INSENSITIVE);
        }

        return __annotationNamePattern;
    }
    
    /**
     * Parses the invert relation path attribute.
     * @param itemConfig the item configuration to use.
     * @return the invert relation path or <code>null</code> if none defined.
     */
    protected String _parseInvertRelationPath(Configuration itemConfig)
    {
        return itemConfig.getAttribute("invert", null);
    }
    
    /**
     * Parses the force invert attribute.
     * @param itemConfig the item configuration to use.
     * @return <code>true</code> if mutual relationship of the item should be set regardless of user's rights, <code>false</code> otherwise.
     */
    protected boolean _parseForceInvert(Configuration itemConfig)
    {
        return itemConfig.getAttributeAsBoolean("forceInvert", false);
    }
}
