/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.ContentTypesHelper.ViewConfigurations;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.model.restrictions.RestrictedModelItem;
import org.ametys.cms.repository.Content;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.View;

/**
 * This class represents a type of content.
 */
public interface ContentType extends ContentTypeDescriptor, Model
{
    /** Tag for private content type */
    public static final String TAG_PRIVATE = "private";
    
    /** Tag for reference table */
    public static final String TAG_REFERENCE_TABLE = "reference-table";
    
    /** Tag for renderable reference table */
    public static final String TAG_RENDERABLE_FERENCE_TABLE = "renderable-reference-table";
    
    /** Tag for simple content type */
    public static final String TAG_MIXIN = "mixin";
    
    /** Tag name for a view declaration */
    public static final String VIEW_TAG_NAME = "view";
    /** Tag name for a view declaration using the legacy syntax */
    public static final String VIEW_TAG_NAME_WITH_LEGACY_SYNTAX = "metadata-set";
    /** Tag name for a group declaration in a view using legacy syntax */
    public static final String GROUP_TAG_NAME_WITH_LEGACY_SYNTAX = "fieldset";
    /** Tag name for a model item reference declaration in a view using the legacy syntax */
    public static final String ITEM_TAG_NAME_WITH_LEGACY_SYNTAX = "metadata-ref";
    /** Attribute name for an item reference in a view using the legacy syntax */
    public static final String ITEM_REF_ATTRIBUTE_NAME_WITH_LEGACY_SYNTAX = "name";
    
    
    /**
     * Called by the extension point when the model items of all content types have been initialized.
     * Can be used to do more initialization, checks, ... needing model items (attributes / properties) of distant contents
     * @throws Exception if an error occurs of if an additional check fails.
     */
    void initializeAfterModelItemsInitialization() throws Exception;
    
    /**
     * Called by the extension point when the views of all content types have been initialized.
     * Can be used to do more initialization, checks, ... needing views of distant contents
     * @throws Exception if an error occurs of if an additional check fails.
     */
    void initializeAfterViewsInitialization() throws Exception;
    
    /**
     * Retrieves the potential global validators.
     * @return the global validators or an empty List if none.
     */
    List<ContentValidator> getGlobalValidators();
    
    /**
     * Retrieves the RichText updater
     * @return the RichText updater or <code>null</code> if none.
     */
    RichTextUpdater getRichTextUpdater();
    
    /**
     * Retrieves the properties references
     * Properties can reference items on distant content type.
     * Those references are like: Impacted ContentType -> local IndexingField name -> path to impacted content.
     * @return this content type's properties references
     */
    Map<String, Map<String, List<String>>> getPropertiesReferences();
    
    /**
     * Retrieves the names of all the content type's view
     * @return the views' names
     */
    default Set<String> getViewNames()
    {
        return getViewNames(true);
    }

    /**
     * Retrieves the names of all the content type's view
     * @param includeInternals if the result should include internal views.
     * @return the views' names
     */
    Set<String> getViewNames(boolean includeInternals);
    
    /**
     * Retrieves the view with the given name
     * @param viewName the name of the view to retrieve
     * @return the view
     */
    View getView(String viewName);
    
    /**
     * Retrieves all the views' configurations declared by this content type
     * @return the views' configurations
     */
    Map<String, ViewConfigurations> getViewConfigurations();
    
    /**
     * Retrieves the optional view's configuration declared by this content type
     * @param viewName the name of the view to retrieve
     * @return the view's configuration
     */
    Optional<ViewConfigurations> getViewConfigurations(String viewName);
    
    /**
     * Determine whether an attribute can be read at this time.
     * Method called by {@link AttributeDefinition#canRead(Content)} to do some other checks, depending on the content type
     * @param definition the attribute definition
     * @param content The content where attribute is to be read on. Can be null, on content creation. 
     * @return <code>true</code> if the current user is allowed to read the attribute of this content.
     * @throws AmetysRepositoryException if an error occurs while accessing the content.
     */
    public default boolean canRead(Content content, RestrictedModelItem<Content> definition) throws AmetysRepositoryException
    {
        return true;
    }
    
    /**
     * Determine whether an attribute can be written at this time.
     * Method called by {@link AttributeDefinition#canWrite(Content)} to do some other checks, depending on the content type
     * @param definition the attribute definition
     * @param content The content where attribute is to be written on. Can be null, on content creation. 
     * @return <code>true</code> if the current user is allowed to write the attribute of this content.
     * @throws AmetysRepositoryException if an error occurs while accessing the content.
     */
    public default boolean canWrite(Content content, RestrictedModelItem<Content> definition) throws AmetysRepositoryException
    {
        return true;
    }
    
    /**
     * Get the registered tags for this content type
     * @return the tags
     */
    public Set<String> getTags ();
    
    /**
     * Get the registered and inheritable tags for this content type
     * @return the inheritable tags
     */
    public Set<String> getInheritableTags ();
    
    /**
     * Determines if the content type has the given tag
     * @param tagName The tag name
     * @return true if the content type is tagged with the given tag name
     */
    public boolean hasTag (String tagName);
    
    /**
     * Get whether the content type is private, i.e. should not be created by
     * the regular content interface.
     * @return true if the content is private, false if it is public.
     */
    public boolean isPrivate();
    
    /**
     * Get whether the content type is simple, i.e. should contains only simple metadata type to be edited in a grid.
     * @return true if the content is simple, false otherwise
     */
    public boolean isSimple();
    
    /**
     * Get whether the content type is a reference table
     * @return true if the content type is a reference table
     */
    public boolean isReferenceTable();
    
    /**
     * Get whether the content type is multilingual
     * @return true if the content is multilingual, false otherwise
     */
    public boolean isMultilingual();
    
    /**
     * Get whether the content type is abstract, i.e. should not be created by the regular content interface.
     * @return true if the content is abstract, false otherwise.
     */
    public boolean isAbstract();
    
    /**
     * Get whether the content type is a mixin, i.e. should used to add metadata to a existing content.
     * @return true if the content is mixin, false otherwise.
     */
    public boolean isMixin();
    
    /**
     * Get the workflow names configured in the content type definition or in its supertypes.
     * @return A {@link Set} of workflow names
     */
    public Set<String> getConfiguredDefaultWorkflowNames();
    
    /**
     * Get the default workflow name.
     * @return The default workflow name
     */
    public Optional<String> getDefaultWorkflowName();
    
    /**
     * Get the right needed to create a content of this type.
     * @return the right needed to create a content of this type. If null is returned, no right is needed.
     */
    public String getRight();
    
    /**
     * SAX the additional content data linked to its {@link ContentType}
     * These additional data will be available in content view
     * @param contentHandler The handler to SAX into
     * @param content The content
     * @throws AmetysRepositoryException if an error occurs while accessing the content.
     * @throws SAXException if an error occurs while SAXing
     */
    public void saxContentTypeAdditionalData (ContentHandler contentHandler, Content content) throws AmetysRepositoryException, SAXException;
    
    /**
     * Get the additional data relative to its {@link ContentType}.
     * These additional data are delivered to client side.
     * @param content The content
     * @return the additional properties in a Map
     */
    public Map<String, Object> getAdditionalData (Content content);
    
//    /**
//     * Gets the "parent" content type of this content type. Must be a private and simple content type.
//     * @return the "parent" content type of this content type
//     */
//    public ContentType getParentContentType();
    
    /**
     * Gets the attribute holding the "parent" of the contents of this content type. Must be of type {@link ModelItemTypeConstants#CONTENT_ELEMENT_TYPE_ID} and must reference a private and simple content type.
     * @return  the attribute holding the "parent" of the contents of this content type 
     */
    public Optional<ContentAttributeDefinition> getParentAttributeDefinition();
    
    /**
     *  Get the overridden model items list
     *  @return the overridden model items list
     */
    public List<String> getOverriddenModelItems();
    
    /**
     *  Get the overridden views list
     *  @return the overridden views list
     */
    public List<String> getOverriddenViews();
    
}
