/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.List;
import java.util.Set;

import org.ametys.core.ui.ClientSideElement.ScriptFile;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.View;

/**
 * This class represents a type of content
 */
public class ContentTypeDefinition implements ContentTypeDescriptor
{
    private String _id;
    private String _pluginName;
    private I18nizableText _label;
    private I18nizableText _description;
    private I18nizableText _defaultTitle;
    private I18nizableText _category;
    private String[] _supertypeIds;
    private boolean _isAbstract;
    private String _iconGlyph;
    private String _iconDecorator;
    private String _largeIcon;
    private String _mediumIcon;
    private String _right;
    private Set<String> _tags;
    private List<ScriptFile> _cssFiles;
    private List<ModelItem> _modelItems;
    private List<View> _views;
    private String _parentRef;

    /**
     * Constructs a new representative of a content type
     * @param id id of content type
     */
    public ContentTypeDefinition(String id)
    {
        this._id = id;
    }

    public String getId()
    {
        return this._id;
    }

    public String getPluginName()
    {
        return this._pluginName;
    }

    public I18nizableText getLabel()
    {
        return this._label;
    }

    public I18nizableText getDescription()
    {
        return this._description;
    }

    public I18nizableText getDefaultTitle()
    {
        return this._defaultTitle;
    }

    public I18nizableText getCategory()
    {
        return this._category;
    }

    public String[] getSupertypeIds()
    {
        return this._supertypeIds;
    }

    /**
     * Indicates if the content type is abstract
     * 
     * @return True if the content type is abstract, false otherwise.
     */
    public boolean isAbstract()
    {
        return this._isAbstract;
    }

    public String getIconGlyph()
    {
        return this._iconGlyph;
    }

    public String getIconDecorator()
    {
        return this._iconDecorator;
    }

    public String getSmallIcon()
    {
        return this.getSmallIcon();
    }

    public String getMediumIcon()
    {
        return this._mediumIcon;
    }

    public String getLargeIcon()
    {
        return this._largeIcon;
    }

    public List<ScriptFile> getCSSFiles()
    {
        return this._cssFiles;
    }

    /**
     * Get the right of the content type
     * 
     * @return the right of the content type
     */
    public String getRight()
    {
        return this._right;
    }

    /**
     * Get tags of content type
     * 
     * @return tags of content type
     */
    public Set<String> getTags()
    {
        return this._tags;
    }

    /**
     * Get metadataSet of content type
     * 
     * @return metadataset of content type
     */
    public List<View> getViews()
    {
        return this._views;
    }

    /**
     * Get metadata of content type
     * 
     * @return metadata of content type
     */
    public List<ModelItem> getModelItems()
    {
        return this._modelItems;
    }

    /**
     * Get the parent reference of content type
     * 
     * @return the parent reference of content type
     */
    public String getParentRef()
    {
        return this._parentRef;
    }

    /**
     * Set the right of content type
     * 
     * @param right the right to set
     */
    public void setRight(String right)
    {
        this._right = right;
    }

    /**
     * Set tags of content type
     * 
     * @param tags tags to set
     */
    public void setTags(Set<String> tags)
    {
        this._tags = tags;
    }

    /**
     * Set model items of content type
     * 
     * @param modelItems model items to set
     */
    public void setModelItems(List<ModelItem> modelItems)
    {
        this._modelItems = modelItems;
    }

    /**
     * Set views of content type
     * 
     * @param views views to set
     */
    public void setViews(List<View> views)
    {
        this._views = views;
    }

    /**
     * Set id of content type
     * 
     * @param id the id to set
     */
    public void setId(String id)
    {
        this._id = id;
    }

    /**
     * Set the plugin name of content type
     * 
     * @param pluginName the plugin name to set
     */
    public void setPluginName(String pluginName)
    {
        this._pluginName = pluginName;
    }

    /**
     * Set the label of content type
     * 
     * @param label the label to set
     */
    public void setLabel(I18nizableText label)
    {
        this._label = label;
    }

    /**
     * Set the description of content type
     * 
     * @param description the description to set
     */
    public void setDescription(I18nizableText description)
    {
        this._description = description;
    }

    /**
     * Set the default title of content type
     * 
     * @param defaultTitle the default title to set
     */
    public void setDefaultTitle(I18nizableText defaultTitle)
    {
        this._defaultTitle = defaultTitle;
    }

    /**
     * Set the category of content type
     * 
     * @param category the category of content type
     */
    public void setCategory(I18nizableText category)
    {
        this._category = category;
    }

    /**
     * Set supertypes of content type
     * 
     * @param supertypeIds supertypes to set
     */
    public void setSupertypeIds(String[] supertypeIds)
    {
        this._supertypeIds = supertypeIds;
    }

    /**
     * Set icon glyph of content type
     * 
     * @param iconGlyph icon glyph to set
     */
    public void setIconGlyph(String iconGlyph)
    {
        this._iconGlyph = iconGlyph;
    }

    /**
     * Set icon decorator of content type
     * 
     * @param iconDecorator icon decorator to set
     */
    public void setIconDecorator(String iconDecorator)
    {
        this._iconDecorator = iconDecorator;
    }

    /**
     * Set large icon of content type
     * 
     * @param largeIcon large icon to set
     */
    public void setLargeIcon(String largeIcon)
    {
        this._largeIcon = largeIcon;
    }

    /**
     * Set medium icon of content type
     * 
     * @param mediumIcon medium icon to set
     */
    public void setMediumIcon(String mediumIcon)
    {
        this._mediumIcon = mediumIcon;
    }

    /**
     * Set CSS files of content type
     * 
     * @param cSSFiles css files to set
     */
    public void setCSSFiles(List<ScriptFile> cSSFiles)
    {
        this._cssFiles = cSSFiles;
    }

    /**
     * Set if content type is abstract
     * 
     * @param isAbstract the boolean to set
     * 
     */
    public void setIsAbstract(boolean isAbstract)
    {
        this._isAbstract = isAbstract;
    }

    /**
     * Set the parent reference of content type
     * 
     * @param parentRef the parent reference to set
     */
    public void setParentRef(String parentRef)
    {
        this._parentRef = parentRef;
    }

}
