/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.List;

import org.ametys.core.ui.ClientSideElement.ScriptFile;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * This class represents a descriptor for content type
 */
public interface ContentTypeDescriptor
{
    /**
     * Retrieves the id of the content type.
     * @return the id.
     */
    String getId();
    
    /**
     * Retrieves the name of the plugin declaring this content type.
     * @return the name of the plugin.
     */
    String getPluginName();
    
    /**
     * Retrieves the label of the content type.
     * @return the label.
     */
    I18nizableText getLabel();
    
    /**
     * Retrieves the description of the content type.
     * @return the description.
     */
    I18nizableText getDescription();
    
    /**
     * Retrieves the default title of the content type.
     * @return the default title.
     */
    I18nizableText getDefaultTitle();
    
    /**
     * Retrieves the category of the content type.
     * @return the category.
     */
    I18nizableText getCategory();

    /**
     * Retrieves the super type's ids.
     * @return the super type's ids, or empty if this content type doesn't extend a specific content type.
     */
    String[] getSupertypeIds();
    
    /**
     * Retrieves the CSS class to use for glyph icon
     * @return the glyph name.
     */
    String getIconGlyph();
    
    /**
     * Retrieves the CSS class to use for decorator above the main icon
     * @return the glyph name.
     */
    String getIconDecorator();
    
    /**
     * Retrieves the URL of the icon without the context path.
     * @return the icon URL for the small image 16x16.
     */
    String getSmallIcon();

    /**
     * Retrieves the URL of the icon without the context path.
     * @return the icon URL for the medium sized image 32x32.
     */
    String getMediumIcon();

    /**
     * Retrieves the URL of the icon without the context path.
     * @return the icon URL for the large image 48x48.
     */
    String getLargeIcon();
    
    /**
     * Returns the list of CSS files needed to correctly display the content's in BO, such as the content's icon
     * @return The list of CSS files needed. Must not be null.
     */
    List<ScriptFile> getCSSFiles();
    
    /**
     * Returns the default i18n catalog for this content type.
     * @return the default i18n catalog for this content type.
     */
    public default String getDefaultCatalog()
    {
        return "plugin." + getPluginName();
    }
    
    /**
     * Returns the path for icons
     * @param pluginName the configured plugin
     * @return the path for icons
     */
    public default String getIconPath(String pluginName)
    {
        return "/plugins/" + pluginName + "/resources/";
    }
}
