/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang.StringUtils;

import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.Enumerator;

/**
 * Enumerator for {@link ContentType}
 *
 */
public class ContentTypeEnumerator implements Enumerator<String>, org.ametys.runtime.parameter.Enumerator, Serviceable, Configurable
{
    /** The content type extension point */
    protected ContentTypeExtensionPoint _cTypeExtPt;
    
    /** All option : 'disabled' to disable the 'all' option, 
     * 'blank' to use a blank value for 'all' option,
     * or 'concat' to concat all content types' ids for 'all' option
     **/
    protected String _allOption;
    /** True to exclude private content types */
    protected boolean _excludePrivate;
    /** True to exclude reference table content types */
    protected boolean _excludeReferenceTable;
    /** True to exclude abstract content types */
    protected boolean _excludeAbstract;
    /** True to exclude mixin content types */
    protected boolean _excludeMixin;
    /** True to include only mixin content types */
    protected boolean _mixinOnly;
    /** True to include only reference table content types */
    protected boolean _referenceTableOnly;
    /** The parent content types */
    protected String[] _contentTypes;
    /** The strict content types */
    protected String[] _strictContentTypes;
    
    private ContentTypesTreeComponent _cTypeTreeComponent;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _cTypeExtPt = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
        _cTypeTreeComponent = (ContentTypesTreeComponent) smanager.lookup(ContentTypesTreeComponent.ROLE);
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        Configuration allOptConf = configuration.getChild("enumeration").getChild("custom-enumerator").getChild("allOption", false);
        _allOption = allOptConf != null ? allOptConf.getValue("disabled") : "disabled";
        
        Configuration privateConf = configuration.getChild("enumeration").getChild("custom-enumerator").getChild("excludePrivate", false);
        _excludePrivate = privateConf != null ? Boolean.valueOf(privateConf.getValue("false")) : false;
        
        Configuration refTableConf = configuration.getChild("enumeration").getChild("custom-enumerator").getChild("excludeReferenceTable", false);
        _excludeReferenceTable = refTableConf != null ? Boolean.valueOf(refTableConf.getValue("false")) : true;
        
        Configuration abstractConf = configuration.getChild("enumeration").getChild("custom-enumerator").getChild("excludeAbstract", false);
        _excludeAbstract = abstractConf != null ? Boolean.valueOf(abstractConf.getValue("false")) : false;
        
        Configuration mixinConf = configuration.getChild("enumeration").getChild("custom-enumerator").getChild("excludeMixin", false);
        _excludeMixin = mixinConf != null ? Boolean.valueOf(mixinConf.getValue("false")) : false;
        
        Configuration mixinOnlyConf = configuration.getChild("enumeration").getChild("custom-enumerator").getChild("includeMixinOnly", false);
        _mixinOnly = mixinOnlyConf != null ? Boolean.valueOf(mixinOnlyConf.getValue("false")) : false;
        
        Configuration refTableOnlyConf = configuration.getChild("enumeration").getChild("custom-enumerator").getChild("includeReferenceTableOnly", false);
        _referenceTableOnly = refTableOnlyConf != null ? Boolean.valueOf(refTableOnlyConf.getValue("false")) : false;
        
        String strictCTypes = configuration.getChild("enumeration").getChild("custom-enumerator").getChild("strictContentTypes").getValue(null);
        _strictContentTypes = strictCTypes != null ? strictCTypes.split(",") : null;
        
        String cTypes = configuration.getChild("enumeration").getChild("custom-enumerator").getChild("contentTypes").getValue(null);
        _contentTypes = cTypes != null ? cTypes.split(",") : null;
    }
    
    @Override
    public Map<String, Object> getConfiguration()
    {
        Map<String, Object> params = new HashMap<>();
        
        params.put("allOption", _allOption);
        params.put("excludePrivate", _excludePrivate);
        params.put("excludeReferenceTable", _excludeReferenceTable);
        params.put("excludeAbstract", _excludeAbstract);
        params.put("excludeMixin", _excludeMixin);
        params.put("mixinOnly", _mixinOnly);
        if (_strictContentTypes != null)
        {
            params.put("strictContentTypes", _strictContentTypes);
        }
        if (_contentTypes != null)
        {
            params.put("contentTypes", _contentTypes);
        }
        
        return params;
    }
    
    @Override
    // TODO NEWATTRIBUTEAPI_CONTENT: remove this method when org.ametys.runtime.parameter.Enumerator will be removed
    public Map<Object, I18nizableText> getEntries() throws Exception
    {
        Map<Object, I18nizableText> result = new HashMap<>();
        for (Map.Entry<String, I18nizableText> entry : getTypedEntries().entrySet())
        {
            result.put(entry.getKey(), entry.getValue());
        }
        return Collections.unmodifiableMap(result);
    }
    
    public Map<String, I18nizableText> getTypedEntries() throws Exception
    {
        Map<String, I18nizableText> entries = new HashMap<>();
        
        Set<ContentType> cTypes = getMatchingContentTypes();
        for (ContentType cType : cTypes)
        {
            entries.put(cType.getId(), cType.getLabel());
        }

        // All contents
        _handleAllOptionEntry(entries);
        
        return entries;
    }
    
    /**
     * Get the matching content types
     * @return the matching content types
     */
    protected Set<ContentType> getMatchingContentTypes()
    {
        if (_strictContentTypes != null && _strictContentTypes.length != 0)
        {
            return _cTypeTreeComponent.getMatchingContentTypes(_strictContentTypes, false, _excludeReferenceTable, _excludePrivate, _excludeAbstract, _excludeMixin, _mixinOnly, _referenceTableOnly);
        }
        else if (_contentTypes != null && _contentTypes.length != 0)
        {
            return _cTypeTreeComponent.getMatchingContentTypes(_contentTypes, true, _excludeReferenceTable, _excludePrivate, _excludeAbstract, _excludeMixin, _mixinOnly, _referenceTableOnly);
        }
        else
        {
            return _cTypeTreeComponent.getMatchingContentTypes(_excludeReferenceTable, _excludePrivate, _excludeAbstract, _excludeMixin, _mixinOnly, _referenceTableOnly);
        }
    }

    /**
     * Add the all option entry to the entry list if necessary
     * @param entries The enumerator entries
     */
    protected void _handleAllOptionEntry(Map<String, I18nizableText> entries)
    {
        if (!_allOption.equals("disabled"))
        {
            String value = _allOption.equals("concat") ? StringUtils.join(entries.keySet(), ',') : "";
            if (!entries.containsKey(value))
            {
                entries.put(value, new I18nizableText("plugin.cms", "WIDGET_COMBOBOX_ALL_OPTIONS"));
            }
        }
    }

    @Override
    public I18nizableText getEntry(String value) throws Exception
    {
        ContentType cType = _cTypeExtPt.getExtension(value);
        return cType.getLabel();
    }
}
