/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.configuration.DefaultConfigurationBuilder;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.Constants;
import org.apache.cocoon.environment.Context;

import org.ametys.cms.repository.ContentTypeExpression;
import org.ametys.cms.repository.ContentTypeOrMixinExpression;
import org.ametys.cms.repository.MixinTypeExpression;
import org.ametys.plugins.repository.query.expression.Expression.Operator;
import org.ametys.runtime.plugin.Feature;
import org.ametys.runtime.plugin.PluginsManager;
import org.ametys.runtime.plugin.component.AbstractThreadSafeComponentExtensionPoint;

import com.google.common.collect.HashMultimap;

/**
 * This class is in charge of handling content type extension point.<br>
 * This point handles the pool of available content types.
 */
public class ContentTypeExtensionPoint extends AbstractThreadSafeComponentExtensionPoint<ContentType>
{
    /** Avalon Role */
    public static final String ROLE = ContentTypeExtensionPoint.class.getName();
    
    /** A registry of all the content type's sub types. */
    protected HashMultimap<String, String> _subTypes;
    
    /** The indexing references, as a Map of Referenced ContentType -&gt; referencing ContentType
     *  -&gt; local IndexingField name -&gt; path to impacted content. */
    protected Map<String, Map<String, Map<String, List<String>>>> _indexingReferences;
    
    private ContentTypeOverridesExtensionPoint _contentTypeOverridesExtensionPoint;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        
        _contentTypeOverridesExtensionPoint = (ContentTypeOverridesExtensionPoint) manager.lookup(ContentTypeOverridesExtensionPoint.ROLE);
    }

    @Override
    public void initializeExtensions() throws Exception
    {
        // add content types defined in WEB-INF/param/content-types/<pluginName>/<id>.xml
        Context ctx = (Context) _context.get(Constants.CONTEXT_ENVIRONMENT_CONTEXT);
        
        final Set<String> pluginNames = PluginsManager.getInstance().getPluginNames();
        
        File rootCT = new File(ctx.getRealPath("/WEB-INF/param/content-types"));
        File[] plugins = rootCT.listFiles((file, name) -> pluginNames.contains(name));
        
        if (plugins != null)
        {
            for (File pluginDir : plugins)
            {
                String pluginName = pluginDir.getName();
                
                File[] ctFiles = pluginDir.listFiles((file, name) -> name.endsWith(".xml"));
                
                for (File ctFile : ctFiles)
                {
                    String fileName = ctFile.getName();
                    
                    try (InputStream is = new FileInputStream(ctFile))
                    {
                        String id = "content-type." + fileName.substring(0, fileName.lastIndexOf('.'));
                        
                        if (hasExtension(id))
                        {
                            throw new IllegalArgumentException("The content type of id " + id + " at " + ctFile.getAbsolutePath() + " is already declared.");
                        }
                        
                        DefaultConfiguration conf = new DefaultConfiguration("extension");
                        
                        Configuration c = new DefaultConfigurationBuilder(true).build(is, ctFile.getAbsolutePath());
                        
                        conf.addChild(c);
                        
                        addComponent(pluginName, Feature.DEFAULT_NAME, id, AutomaticContentType.class, conf);
                    }
                    catch (Exception e)
                    {
                        getLogger().error("Error parsing content type at '" + ctFile.getAbsolutePath() + "'.", e);
                        throw e;
                    }
                }
            }
        }
        
        super.initializeExtensions();
        
        // Compute all the content types hierarchy.
        initializeContentTypeHierarchy();
        
        // The model items of all content types have been initialized, we can do more stuff
        initializeAfterModelItemsInitialization();
        
        // The views of all content types have been initialized, we can do more stuff
        initializeAfterViewsInitialization();
        
        // Initialize indexing references.
        initializeContentTypeIndexingReferences();
        
        // Checks that the overriding extensions are not trying to override non existing content types
        checkOverridesOfUnexistingContentTypes();
    }
    
    /**
     * Create a hierarchical {@link ContentTypeExpression} which returns
     * all contents of the specified type and all its descendant types.
     * @param contentTypeIds the content type id.
     * @return the hierarchical content type expression.
     */
    public ContentTypeExpression createHierarchicalCTExpression(String... contentTypeIds)
    {
        Set<String> typesToSearch = new HashSet<>();
        
        for (String contentTypeId : contentTypeIds)
        {
            typesToSearch.add(contentTypeId);
            typesToSearch.addAll(getSubTypes(contentTypeId));
        }
        
        return new ContentTypeExpression(Operator.EQ, typesToSearch.toArray(new String[typesToSearch.size()]));
    }
    
    /**
     * Create a hierarchical {@link ContentTypeExpression} which returns
     * all contents of the specified mixin and all its descendant types.
     * @param mixinIds the mixin ids
     * @return the hierarchical content type expression.
     */
    public MixinTypeExpression createHierarchicalMixinExpression(String... mixinIds)
    {
        Set<String> typesToSearch = new HashSet<>();
        
        for (String mixinId : mixinIds)
        {
            typesToSearch.add(mixinId);
            typesToSearch.addAll(getSubTypes(mixinId));
        }
        
        return new MixinTypeExpression(Operator.EQ, typesToSearch.toArray(new String[typesToSearch.size()]));
    }
    
    /**
     * Create a hierarchical {@link ContentTypeExpression} which returns
     * all contents of the specified type or mixin and all its descendant types.
     * @param contentIds the mixin ids
     * @return the hierarchical content type expression.
     */
    public ContentTypeOrMixinExpression createHierarchicalContentTypeOrMixinExpression(String... contentIds)
    {
        Set<String> typesToSearch = new HashSet<>();
        
        for (String contentTypeId : contentIds)
        {
            typesToSearch.add(contentTypeId);
            typesToSearch.addAll(getSubTypes(contentTypeId));
        }
        
        return new ContentTypeOrMixinExpression(Operator.EQ, typesToSearch.toArray(new String[typesToSearch.size()]));
    }
    
    /**
     * Get all the descendant types of a given {@link ContentType}.
     * @param contentTypeId the content type id.
     * @return a collection of the ids of all the ContentType's descendants.
     */
    public Set<String> getSubTypes(String contentTypeId)
    {
        Set<String> allSubtypes = new HashSet<>();
        
        Set<String> subtypeIds = getDirectSubTypes(contentTypeId);
        allSubtypes.addAll(subtypeIds);
        
        for (String subtypeId : subtypeIds)
        {
            allSubtypes.addAll(getSubTypes(subtypeId));
        }
        
        return allSubtypes;
    }
    
    /**
     * Get all the descendant types of a given {@link ContentType}.
     * @param contentTypeId the content type id to test
     * @param parentTypeId the content type that should be the parent content type
     * @return a collection of the ids of all the ContentType's descendants.
     */
    public boolean isDescendant(String contentTypeId, String parentTypeId)
    {
        Set<String> subtypeIds = getDirectSubTypes(parentTypeId);
        
        if (subtypeIds.contains(contentTypeId))
        {
            return true;
        }
        else
        {
            for (String subtypeId : subtypeIds)
            {
                if (isDescendant(subtypeId, parentTypeId))
                {
                    return true;
                }
            }
        }
        
        return false;
    }
    
    /**
     * Get all the descendant types of a given {@link ContentType}.
     * @param contentTypeId the content type id to test
     * @param parentTypeId the content type that should be the parent content type
     * @return a collection of the ids of all the ContentType's descendants.
     */
    public boolean isSameOrDescendant(String contentTypeId, String parentTypeId)
    {
        return contentTypeId.equals(parentTypeId) || isDescendant(contentTypeId, parentTypeId);
    }
    
    /**
     * Get the immediate children types of a given {@link ContentType}.
     * @param contentTypeId the content type id.
     * @return a collection of the ids of the immediate ContentType's children.
     */
    public Set<String> getDirectSubTypes(String contentTypeId)
    {
        if (_subTypes.containsKey(contentTypeId))
        {
            return Collections.unmodifiableSet(_subTypes.get(contentTypeId));
        }
        else
        {
            return Collections.emptySet();
        }
    }
    
    /**
     * Get the content types referencing this content type as an indexing field.
     * @param referencedContentType the referenced content type.
     * @return the {@link ContentType}s referencing the given {@link ContentType} as an indexing field,
     * as a Map of referencing ContentType -&gt; local IndexingField name -&gt; path to impacted content.<br>
     * <code>null</code> is returned if the ContentType isn't referenced in any indexing field.
     */
    public Map<String, Map<String, List<String>>> getIndexingReferences(String referencedContentType)
    {
        if (_indexingReferences.containsKey(referencedContentType))
        {
            return Collections.unmodifiableMap(_indexingReferences.get(referencedContentType));
        }
        else
        {
            return Collections.emptyMap();
        }
    }
    
    /**
     * Test if the given content type has indexing references, i.e. if other content types
     * reference this one in an indexing field chain.
     * @param contentTypeId the referenced content type ID.
     * @return <code>true</code> if the given content type has indexing references, <code>false</code> otherwise.
     */
    public boolean hasIndexingReferences(String contentTypeId)
    {
        return _indexingReferences.containsKey(contentTypeId);
    }
    
    /**
     * Initialize the content type hierarchy.
     */
    protected void initializeContentTypeHierarchy()
    {
        _subTypes = HashMultimap.create();
        
        for (String id : getExtensionsIds())
        {
            ContentType contentType = getExtension(id);
            
            String[] supertypeIds = contentType.getSupertypeIds();
            for (String supertypeId : supertypeIds)
            {
                _subTypes.put(supertypeId, id);
            }
        }
    }
    
    /**
     * Model items of all content types have been initialized: call the corresponding post initialization method on each ContentType.
     * @throws Exception if an error occurs.
     */
    protected void initializeAfterModelItemsInitialization() throws Exception
    {
        for (String id : getExtensionsIds())
        {
            ContentType contentType = getExtension(id);
            contentType.initializeAfterModelItemsInitialization();
        }
    }
    
    /**
     * Views of all content types have been initialized: call the corresponding post initialization method on each ContentType.
     * @throws Exception if an error occurs.
     */
    protected void initializeAfterViewsInitialization() throws Exception
    {
        for (String id : getExtensionsIds())
        {
            ContentType contentType = getExtension(id);
            contentType.initializeAfterViewsInitialization();
        }
    }
    
    /**
     * Initialize the indexing references on all content types.
     */
    protected void initializeContentTypeIndexingReferences()
    {
        // Referenced ContentType -> referencing ContentType -> local IndexingField name -> path to impacted content.
        _indexingReferences = new HashMap<>();
        
        for (String cTypeId : getExtensionsIds())
        {
            ContentType cType = getExtension(cTypeId);
            
            // Get the references in the content type indexing model.
            Map<String, Map<String, List<String>>> references = cType.getPropertiesReferences();
            
            for (String refCTypeId : references.keySet())
            {
                Map<String, Map<String, List<String>>> allCTypeBackReferences;
                if (_indexingReferences.containsKey(refCTypeId))
                {
                    allCTypeBackReferences = _indexingReferences.get(refCTypeId);
                }
                else
                {
                    allCTypeBackReferences = new HashMap<>();
                    _indexingReferences.put(refCTypeId, allCTypeBackReferences);
                }
                
                Map<String, List<String>> cTypeBackReferences;
                if (allCTypeBackReferences.containsKey(cTypeId))
                {
                    cTypeBackReferences = allCTypeBackReferences.get(cTypeId);
                }
                else
                {
                    cTypeBackReferences = new HashMap<>();
                    allCTypeBackReferences.put(cTypeId, cTypeBackReferences);
                }
                
                cTypeBackReferences.putAll(references.get(refCTypeId));
            }
        }
    }
    
    /**
     * Check that there is no override of unexisting content types from an ContentTypeOverrides extension
     * @throws ConfigurationException If an error occurs
     */
    private void checkOverridesOfUnexistingContentTypes() throws ConfigurationException 
    {
        Set<String> contentTypesOverridden = _contentTypeOverridesExtensionPoint.getOverriddenContentTypes();
        
        for (String contentType : contentTypesOverridden)
        {
            if (!hasExtension(contentType))
            {
                throw new ConfigurationException("The content type '" + contentType + "' was not found and therefore cannot be overridden");
            }
        }
    }
}
