/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.io.IOException;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.generation.AbstractGenerator;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.cms.repository.Content;
import org.ametys.plugins.repository.model.CompositeDefinition;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.ViewItemGroup;
import org.ametys.runtime.util.Labelable;

/**
 * Generates content type's definition.
 */
public class ContentTypeGenerator extends AbstractGenerator implements Serviceable
{
    private ContentTypesHelper _cTypeHelper;
    private ContentTypeExtensionPoint _cTypeEP;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _cTypeHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
        _cTypeEP = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    public void generate() throws IOException, SAXException, ProcessingException
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        String cTypeId = parameters.getParameter("contentTypeId", null);
        Content content = (Content) request.getAttribute(Content.class.getName());
        
        if (StringUtils.isBlank(cTypeId) && content == null)
        {
            throw new IllegalArgumentException("Missing content in request or content type id parameter. Cannot compute rendering from model.");
        }
        
        // If an empty param is provided, the default value would be ignored, so we have to create the default value mecanism by hand
        String fallbackViewName = parameters.getParameter("fallbackViewName", StringUtils.EMPTY);
        String viewName = parameters.getParameter("viewName", StringUtils.EMPTY);
        
        contentHandler.startDocument();
        if (StringUtils.isNotBlank(cTypeId))
        {
            _saxViewFromContentType(cTypeId, viewName, fallbackViewName);
        }
        else
        {
            _saxViewFromContent(content, viewName, fallbackViewName);
        }
        contentHandler.endDocument();
    }
    
    private void _saxViewFromContent(Content content, String viewName, String fallbackViewName) throws SAXException
    {
        View view = _cTypeHelper.getViewWithFallback(viewName, fallbackViewName, content);
        
        AttributesImpl atts = new AttributesImpl();
        atts.addCDATAAttribute("id", _cTypeHelper.getContentTypeIdForRendering(content));
        atts.addCDATAAttribute("view", viewName);
        XMLUtils.startElement(contentHandler, "contentType", atts);
        
        _saxNonNullI18nizableText(_cTypeHelper.getContentTypeLabel(content), "label");
        _saxNonNullI18nizableText(_cTypeHelper.getContentTypeDescription(content), "description");
        
        _saxViewItemAccessor(view, null);
        
        XMLUtils.endElement(contentHandler, "contentType");
    }
    
    private void _saxViewFromContentType(String cTypeId, String viewName, String fallbackViewName) throws SAXException
    {
        View view = _cTypeHelper.getViewWithFallback(viewName, fallbackViewName, new String[] {cTypeId}, ArrayUtils.EMPTY_STRING_ARRAY);
        
        ContentType cType = _cTypeEP.getExtension(cTypeId);
        AttributesImpl atts = new AttributesImpl();
        atts.addCDATAAttribute("id", cTypeId);
        atts.addCDATAAttribute("view", viewName);
        XMLUtils.startElement(contentHandler, "contentType", atts);
        
        _saxNonNullI18nizableText(cType.getLabel(), "label");
        _saxNonNullI18nizableText(cType.getDescription(), "description");
        
        _saxViewItemAccessor(view, null);
        
        XMLUtils.endElement(contentHandler, "contentType");
    }
    
    private void _saxViewItemAccessor(ViewItemAccessor viewItemAccessor, String prefix) throws SAXException
    {
        for (ViewItem item : viewItemAccessor.getViewItems())
        {
            if (item instanceof ModelViewItem)
            {
                _saxModelViewItem((ModelViewItem) item, prefix);
            }
            else if (item instanceof ViewItemGroup)
            {
                _saxViewItemGroup((ViewItemGroup) item, prefix);
            }
        }
    }
    
    private void _saxModelViewItem(ModelViewItem modelViewItem, String prefix) throws SAXException
    {
        ModelItem modelItem = modelViewItem.getDefinition();
        String name = modelItem.getName();
        
        AttributesImpl atts = new AttributesImpl();
        atts.addCDATAAttribute("name", name);
        atts.addCDATAAttribute("path", prefix == null ? name : prefix + name);

        if (modelItem instanceof ElementDefinition)
        {
            ElementDefinition definition = (ElementDefinition) modelItem;
            
            atts.addCDATAAttribute("type", definition.getType().getId());
            atts.addCDATAAttribute("multiple", String.valueOf(definition.isMultiple()));
        }
        else if (modelItem instanceof CompositeDefinition)
        {
            atts.addCDATAAttribute("type", "composite");
        }
        else if (modelItem instanceof org.ametys.plugins.repository.model.RepeaterDefinition)
        {
            atts.addCDATAAttribute("type", "repeater");
        }
        
        XMLUtils.startElement(contentHandler, "metadata", atts);

        _saxNonNullI18nizableText(modelViewItem.getLabel(), "label");
        _saxNonNullI18nizableText(modelViewItem.getDescription(), "description");
        
        if (modelViewItem instanceof ViewItemAccessor)
        {
            _saxViewItemAccessor((ViewItemAccessor) modelViewItem, prefix == null ? name  + "/" : prefix + name + "/");
        }

        XMLUtils.endElement(contentHandler, "metadata");
    }
    
    private void _saxViewItemGroup(ViewItemGroup group, String prefix) throws SAXException
    {
        AttributesImpl atts = new AttributesImpl();
        
        if (group.getRole() != null)
        {
            atts.addCDATAAttribute("role", group.getRole());
        }
        if (group.getName() != null)
        {
            atts.addCDATAAttribute("name", group.getName());
        }
        
        XMLUtils.startElement(contentHandler, "fieldset", atts);
        
        _saxNonNullI18nizableText(((Labelable) group).getLabel(), "label");
        
        _saxViewItemAccessor(group, prefix);
        
        XMLUtils.endElement(contentHandler, "fieldset");
    }
    
    private void _saxNonNullI18nizableText(I18nizableText text, String tagName) throws SAXException
    {
        if (text != null)
        {
            text.toSAX(contentHandler, tagName);
        }
    }
}
