/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.ametys.runtime.model.ModelHelper.ConfigurationAndPluginName;
import org.ametys.runtime.plugin.component.AbstractThreadSafeComponentExtensionPoint;

/**
 * This class is in charge of handling content type override extension point.
 * This point handles the pool of content types overrides.
 */
public class ContentTypeOverridesExtensionPoint extends AbstractThreadSafeComponentExtensionPoint<ContentTypeOverridesExtension>
{
    /** The avalon role */
    public static final String ROLE = ContentTypeOverridesExtensionPoint.class.getName();
    
    /**
     * Get override configurations for a content type.
     * @param contentTypeId the content type id
     * @return The list of override configurations
     */
    public List<ConfigurationAndPluginName> getOverrideConfigurations(String contentTypeId)
    {
        List<ConfigurationAndPluginName> configurations = new ArrayList<>();
        for (String extensionId : getExtensionsIds())
        {
            ContentTypeOverridesExtension overrideExtension = getExtension(extensionId);
            if (overrideExtension.getOverriddenContentTypes().contains(contentTypeId))
            {
                configurations.add(new ConfigurationAndPluginName(overrideExtension.getOverrideConfiguration(), overrideExtension.getPluginName()));
            }
        }
        
        return configurations;
    }

    /**
     * Retrieves a Set of content types that have been overridden from this point
     * @return the set of overridden content types identifiers
     */
    public Set<String> getOverriddenContentTypes() 
    {
        Set<String> contentTypes = new HashSet<>();
        for (String extensionId : getExtensionsIds())
        {
            ContentTypeOverridesExtension overrideExtension = getExtension(extensionId);
            contentTypes.addAll(overrideExtension.getOverriddenContentTypes());
        }
        
        return contentTypes;
    }
}
