/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.List;
import java.util.Optional;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfigurationBuilder;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.ContentTypesHelper.ViewConfigurationsByType;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.ModelHelper.ConfigurationAndPluginName;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemAccessor;
import org.ametys.runtime.model.ModelItemGroup;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.SimpleViewItemGroup;
import org.ametys.runtime.model.TemporaryViewReference;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElement;
import org.ametys.runtime.model.ViewHelper;
import org.ametys.runtime.model.ViewItemAccessor;

/**
 * Component that parses the configuration of a content type's view
 */
public class ContentTypeViewParser extends AbstractContentTypeViewParser
{
    /** The content types parser helper */
    protected ContentTypesParserHelper _contentTypesParserHelper;
    /** The source resolver */
    protected SourceResolver _srcResolver;

    /** The content type containing the model items to reference in the view */
    protected ContentType _contentType;
    
    /** The id of the content type containing the configuration to parse */
    protected ContentType _originalContentType;
    
    /** The tag name for simple groups */
    protected Optional<String> _addGroupTagName;
    /** The tag name to add a view item */
    protected Optional<String> _addItemTagName;
    /** Teh attribute name to reference an item */
    protected Optional<String> _itemReferenceAttributeName;
    /** Determine if this parser supports attribute nested in an attribute of type content */
    protected boolean _supportsNestedAttributes;
    
    /**
     * Creates a view parser for content types
     * @param contentType the content type containing the model items to reference in the view
     * @param originalContentType the content type containing the configuration to parse
     */
    public ContentTypeViewParser(ContentType contentType, ContentType originalContentType)
    {
        this(contentType, originalContentType, Optional.empty(), Optional.empty(), Optional.empty(), true);
    }
    
    /**
     * Creates a view parser for content types
     * @param contentType the content type containing the view
     * @param originalContentType the content type containing the configuration to parse
     * @param addGroupTagName the tag name for added simple groups
     * @param addItemTagName the tag name for added items
     * @param itemReferenceAttributeName the attribute name to reference an item
     * @param supportsNestedAttributes <code>false</code> to avoid this parser to support nested attributes
     */
    public ContentTypeViewParser(ContentType contentType, ContentType originalContentType, Optional<String> addGroupTagName, Optional<String> addItemTagName, Optional<String> itemReferenceAttributeName, boolean supportsNestedAttributes)
    {
        _contentType = contentType;
        _originalContentType = originalContentType;
        _addGroupTagName = addGroupTagName;
        _addItemTagName = addItemTagName;
        _itemReferenceAttributeName = itemReferenceAttributeName;
        _supportsNestedAttributes = supportsNestedAttributes;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        
        _contentTypesParserHelper = (ContentTypesParserHelper) manager.lookup(ContentTypesParserHelper.ROLE);
        _srcResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    @Override
    protected Collection<? extends Model> _getModel()
    {
        return List.of(_contentType);
    }
    
    @Override
    protected void _fillViewGeneralInformation(ConfigurationAndPluginName viewConfiguration, View view, View existingView, Collection< ? extends Model> model) throws ConfigurationException
    {
        super._fillViewGeneralInformation(viewConfiguration, view, existingView, model);
        
        // Icons
        Configuration iconConf = viewConfiguration.configuration().getChild("icons");
        
        view.setSmallIcon(_contentTypesParserHelper.parseIcon(_contentType, iconConf, "small", existingView.getSmallIcon()));
        view.setMediumIcon(_contentTypesParserHelper.parseIcon(_contentType, iconConf, "medium", existingView.getMediumIcon()));
        view.setLargeIcon(_contentTypesParserHelper.parseIcon(_contentType, iconConf, "large", existingView.getLargeIcon()));
        
        // Icons - glyph
        String defaultIconGlyph = StringUtils.defaultString(existingView.getIconGlyph(), _contentTypesParserHelper.getIconGlyphDefaultValue(view.getName(), view.getMediumIcon()));
        view.setIconGlyph(iconConf.getChild("glyph").getValue(defaultIconGlyph));
        
        // Icons - icon decorator
        view.setIconDecorator(iconConf.getChild("decorator").getValue(existingView.getIconDecorator()));
    }
    
    @Override
    protected void _parseViewChild(ConfigurationAndPluginName itemConfiguration, View view, Collection<? extends Model> model, boolean override) throws ConfigurationException
    {
        String itemConfigurationName = itemConfiguration.configuration().getName();
        if ("dublin-core".equals(itemConfigurationName))
        {
            _parseDublinCoreViewItems(view);
        }
        else if ("include".equals(itemConfigurationName))
        {
            View viewToInclude = _getViewToInclude(itemConfiguration.configuration(), view.getName());
            view.includeView(viewToInclude);
        }
        else
        {
            super._parseViewChild(itemConfiguration, view, model, override);
        }
    }
    
    @Override
    protected boolean _isAddingItemConfiguration(Configuration itemConfiguration)
    {
        String itemConfigurationName = itemConfiguration.getName();
        return _addItemTagName.map(name -> name.equals(itemConfigurationName))
                              .orElseGet(() -> super._isAddingItemConfiguration(itemConfiguration));
    }
    
    @Override
    protected boolean _isAddingGroupConfiguration(Configuration itemConfiguration)
    {
        String itemConfigurationName = itemConfiguration.getName();
        return _addGroupTagName.map(name -> name.equals(itemConfigurationName))
                              .orElseGet(() -> super._isAddingGroupConfiguration(itemConfiguration));
    }
    
    @Override
    protected String _getModelItemReference(Configuration itemConfiguration) throws ConfigurationException
    {
        return _itemReferenceAttributeName.isPresent()
                ? itemConfiguration.getAttribute(_itemReferenceAttributeName.get())
                : super._getModelItemReference(itemConfiguration);
    }
    
    @Override
    protected Configuration _getAddItemChildConfiguration(Configuration accessorConfiguration)
    {
        return _addItemTagName.map(name -> accessorConfiguration.getChild(name, false))
                              .orElseGet(() -> super._getAddItemChildConfiguration(accessorConfiguration));
    }
    
    @Override
    protected ViewElement _parseViewElement(ConfigurationAndPluginName itemConfiguration, ElementDefinition definition, View referenceView, boolean override) throws ConfigurationException
    {
        return definition instanceof ContentAttributeDefinition contentAttributeDefinition && _supportsNestedAttributes
                ? _parseContentAttributeViewItem(itemConfiguration, contentAttributeDefinition, referenceView, override)
                : super._parseViewElement(itemConfiguration, definition, referenceView, override);
    }
    
    @Override
    protected void _parseViewReference(String viewName, ViewItemAccessor viewItemAccessor, ModelItemAccessor modelItemAccessor) throws ConfigurationException
    {
        // Insert a temporary view reference, that will be resolved when all view of all content types will be parsed
        TemporaryViewReference viewReference = new TemporaryViewReference();
        viewReference.setName(viewName);
        viewItemAccessor.addViewItem(viewReference);
    }
    
    @Override
    protected void _parseSimpleViewItemGroupChild(ConfigurationAndPluginName itemConfiguration, SimpleViewItemGroup group, Collection< ? extends ModelItemAccessor> modelItemAccessors, View referenceView, boolean override) throws ConfigurationException
    {
        String itemConfigurationName = itemConfiguration.configuration().getName();
        if ("dublin-core".equals(itemConfigurationName))
        {
            _parseDublinCoreViewItems(group);
        }
        else if ("include".equals(itemConfigurationName))
        {
            View viewToInclude = _getViewToInclude(itemConfiguration.configuration(), referenceView.getName());
            group.includeView(viewToInclude, referenceView);
        }
        else
        {
            super._parseSimpleViewItemGroupChild(itemConfiguration, group, modelItemAccessors, referenceView, override);
        }
    }
    
    /**
     * Retrieves the view included by the given configuration
     * @param itemConfiguration configuration of the item that includes the view
     * @param viewName name of the view containing the item. (the item can be the view itself or a simple view item group inside the view)
     * @return the view included by the given configuration
     * @throws ConfigurationException if the configuration is not valid
     */
    protected View _getViewToInclude(Configuration itemConfiguration, String viewName) throws ConfigurationException
    {
        String superTypeId = itemConfiguration.getAttribute("from-supertype");
        if ("true".equals(superTypeId))
        {
            Optional<ViewConfigurationsByType> superTypesViewConfigurations = _contentTypesHelper.getViewConfigurationsByType(viewName, _originalContentType.getSupertypeIds(), new String[0]);
            if (superTypesViewConfigurations.isPresent())
            {
                return _contentTypesParserHelper.parseView(_contentType, superTypesViewConfigurations.get());
            }
            else
            {
                throw new ConfigurationException("The view '" + viewName + "' in content type '" + _contentType.getId() + "' includes a super-view not found in its super-types.", itemConfiguration);
            }
        }
        else
        {
            if (!_contentTypesHelper.getAncestors(_contentType.getId()).contains(superTypeId))
            {
                throw new ConfigurationException("The view '" + viewName + "' in content type '" + _contentType.getId() + "' includes the super-view of the type '" + superTypeId + "' that is not a super-type.", itemConfiguration);
            }
            
            ContentType superType = _contentTypeExtensionPoint.getExtension(superTypeId);
            if (superType == null)
            {
                throw new ConfigurationException("The view '" + viewName + "' in content type '" + _contentType.getId() + "' includes the super-view of an unknown super-type '" + superTypeId + "'.", itemConfiguration);
            }
            
            Optional<ViewConfigurationsByType> superTypeViewConfigurations = _contentTypesHelper.getViewConfigurationsByType(viewName, new String[] {superTypeId}, new String[0]);
            if (superTypeViewConfigurations.isPresent())
            {
                return _contentTypesParserHelper.parseView(_contentType, superTypeViewConfigurations.get());
            }
            else
            {
                throw new ConfigurationException("The view '" + viewName + "' in content type '" + _contentType.getId() + "' includes the super-view not found in its super-type '" + superTypeId + "'.", itemConfiguration);
            }
        }
    }

    /**
     * Parses the DublinCore view items and add it to its parent
     * @param parent the parent view item accessor
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected void _parseDublinCoreViewItems(ViewItemAccessor parent) throws ConfigurationException
    {
        ModelItemGroup dublinCoreRootDef = null;
        if (_contentType.hasModelItem("/dc"))
        {
            ModelItem modelItem = _contentType.getModelItem("/dc");
            if (modelItem instanceof ModelItemGroup)
            {
                dublinCoreRootDef = (ModelItemGroup) modelItem;
            }
        }
        
        if (dublinCoreRootDef == null)
        {
            throw new ConfigurationException("Unable to include Dublin Core view in a view of the content type '" + _contentType.getId() + "'. The Dublin Core attributes have not been defined.");
        }
        
        Source src = null;
        try
        {
            src = _srcResolver.resolveURI("resource://org/ametys/cms/dublincore/dublincore.xml");
            
            if (src.exists())
            {
                ModelViewItemGroup<ModelItemGroup> dublinCoreGroup = new ModelViewItemGroup<>();
                dublinCoreGroup.setDefinition(dublinCoreRootDef);
                try (InputStream is = src.getInputStream())
                {
                    Configuration configuration = new DefaultConfigurationBuilder(true).build(is);
                    Configuration viewConfiguration = configuration.getChild("view");
                    
                    for (Configuration childConfiguration : viewConfiguration.getChildren(ADD_ITEM_TAG_NAME))
                    {
                        String childAttributeName = "/dc" + ModelItem.ITEM_PATH_SEPARATOR + childConfiguration.getAttribute(ITEM_REFERENCE_ATTRIBUTE_NAME);
                        if (_contentType.hasModelItem(childAttributeName))
                        {
                            ModelItem definition = _contentType.getModelItem(childAttributeName);
                            ModelViewItem<ModelItem> viewItem = ViewHelper.createModelViewItemInstance(definition);
                            viewItem.setDefinition(definition);
                            
                            dublinCoreGroup.addViewItem(viewItem);
                        }
                        else
                        {
                            throw new ConfigurationException("Unable to include Dublin Core view in a view of the content type '" + _contentType.getId() + "'. The Dublin Core attribute '" + childAttributeName + "' has not been defined.");
                        }
                    }
                }
                parent.addViewItem(dublinCoreGroup);
            }
        }
        catch (IOException | SAXException e)
        {
            throw new ConfigurationException("Unable to parse Dublin Core view", e);
        }
        finally
        {
            if (src != null)
            {
                _srcResolver.release(src);
            }
        }
        
    }
}
