/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.LifecycleHelper;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.contenttype.ContentTypesHelper.ViewConfigurationsByType;
import org.ametys.runtime.model.ModelHelper.ConfigurationAndPluginName;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewParser;

/**
 * Helper for content types parsing
 */
public class ContentTypesParserHelper extends AbstractLogEnabled implements Component, Serviceable, Contextualizable
{
    /** The Avalon role */
    public static final String ROLE = ContentTypesParserHelper.class.getName();
    
    /** The content types helper */
    protected ContentTypesHelper _contentTypesHelper;

    private Context _context;
    private ServiceManager _manager;
    

    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
        _contentTypesHelper = (ContentTypesHelper) _manager.lookup(ContentTypesHelper.ROLE);
    }
    
    /**
     * Parses the view of the given content type from the configurations
     * @param contentType the content type
     * @param viewConfigurationsByType the view's configurations, indexed by the content types defining this view
     * @return the parsed view
     * @throws ConfigurationException if the configuration is invalid
     */
    public View parseView(ContentType contentType, ViewConfigurationsByType viewConfigurationsByType) throws ConfigurationException
    {
        // Check that there is at least one main configuration
        if (viewConfigurationsByType.mainConfigurations().isEmpty())
        {
            throw new ConfigurationException ("There is no main configuration but only overrides for view '" + viewConfigurationsByType.viewName() + "' in content type '" + contentType.getId() + "'");
        }
        
        // Parse View from main configurations
        List<View> views = new ArrayList<>();
        for (ContentType originalContentType : viewConfigurationsByType.mainConfigurations().keySet())
        {
            View view = _parseView(contentType, originalContentType, viewConfigurationsByType.mainConfigurations().get(originalContentType));
            views.add(view);
        }
        View view = _contentTypesHelper.joinViews(views);
        
        // Parse overrides
        for (ContentType originalContentType : viewConfigurationsByType.overrides().keySet())
        {
            for (ConfigurationAndPluginName configuration : viewConfigurationsByType.overrides().get(originalContentType))
            {
                view = _overrideView(contentType, originalContentType, configuration, view);
            }
        }
        
        return view;
    }
    
    private View _parseView(ContentType contentType, ContentType originalContentType, ConfigurationAndPluginName viewConfiguration) throws ConfigurationException
    {
        ViewParser parser = _getContentTypeViewParser(contentType, originalContentType, viewConfiguration.configuration());
        try
        {
            LifecycleHelper.setupComponent(parser, getLogger(), _context, _manager, null);
            return parser.parseView(viewConfiguration);
        }
        catch (Exception e)
        {
            throw new ConfigurationException("Unable to initialize the content view parser", viewConfiguration.configuration(), e);
        }
        finally
        {
            LifecycleHelper.dispose(parser);
        }
    }
    
    private View _overrideView(ContentType contentType, ContentType originalContentType, ConfigurationAndPluginName viewConfiguration, View existingView) throws ConfigurationException
    {
        ViewParser parser = _getContentTypeViewParser(contentType, originalContentType, viewConfiguration.configuration());
        try
        {
            LifecycleHelper.setupComponent(parser, getLogger(), _context, _manager, null);
            return parser.overrideView(viewConfiguration, existingView);
        }
        catch (Exception e)
        {
            throw new ConfigurationException("Unable to initialize the content view parser", viewConfiguration.configuration(), e);
        }
        finally
        {
            LifecycleHelper.dispose(parser);
        }
    }
    
    private ViewParser _getContentTypeViewParser(ContentType contentType, ContentType originalContentType, Configuration viewConfiguration)
    {
        return ContentType.VIEW_TAG_NAME.equals(viewConfiguration.getName())
                ? new ContentTypeViewParser(contentType, originalContentType)
                : new ContentTypeViewParser(contentType, originalContentType, Optional.of(ContentType.GROUP_TAG_NAME_WITH_LEGACY_SYNTAX), Optional.of(ContentType.ITEM_TAG_NAME_WITH_LEGACY_SYNTAX), Optional.of(ContentType.ITEM_REF_ATTRIBUTE_NAME_WITH_LEGACY_SYNTAX), false);
    }
    
    /**
     * Parse an icon path
     * @param contentType the content type
     * @param configuration the configuration to use
     * @param name the child name.
     * @return The icon path
     * @throws ConfigurationException if the configuration is not valid.
     */
    public String parseIcon(ContentTypeDescriptor contentType, Configuration configuration, String name) throws ConfigurationException
    {
        return parseIcon(contentType, configuration, name, "/plugins/cms/resources/img/contenttype/unknown-" + name + ".png");
    }
    
    /**
     * Parse an icon path
     * @param contentType the content type
     * @param configuration the configuration to use
     * @param name the child name.
     * @param defaultValue the default value.
     * @return The icon path
     * @throws ConfigurationException if the configuration is not valid.
     */
    public String parseIcon(ContentTypeDescriptor contentType, Configuration configuration, String name, String defaultValue) throws ConfigurationException
    {
        Configuration iconConfig = configuration.getChild(name, false);
        if (iconConfig != null)
        {
            String pluginName = iconConfig.getAttribute("plugin", contentType.getPluginName());
            return contentType.getIconPath(pluginName) + iconConfig.getValue();
        }
        
        return defaultValue;
    }
    
    /**
     * Retrieves the default value for the view's icon glyph.
     * For all views, use medium icon or a default one.
     * For some common views (main, abstract, link), a specific default icon is used
     * @param viewName the name of the view
     * @param mediumIcon the medium icon of the view
     * @return the default value for the view's icon glyph. 
     */
    public String getIconGlyphDefaultValue(String viewName, String mediumIcon)
    {
        String defaultValue = StringUtils.defaultString(mediumIcon, "ametysicon-column3");
        switch (viewName)
        {
            case "main":
                defaultValue = "ametysicon-document112";
                break;
            case "abstract":
                defaultValue = "ametysicon-document77";
                break;
            case "link":
                defaultValue = "ametysicon-internet58";
                break;
            default:
                break;
        }
        
        return defaultValue;
    }
}
