/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.core.ui.Callable;
import org.ametys.runtime.i18n.I18nizableTextParameter;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * This component get content types according to a lot of parameters (hide mixin content type, hide non reference content type...)
 */
public class ContentTypesTreeComponent extends AbstractLogEnabled implements Component, Serviceable
{
    /** Avalon Role. */
    public static final String ROLE = ContentTypesTreeComponent.class.getName();
    
    private ContentTypeExtensionPoint _contentTypeEP;
    
    public void service(ServiceManager smanager) throws ServiceException
    {
        _contentTypeEP = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    enum ExcludeMode
    {
        DISABLED,
        HIDE
    }
    
    /**
     * Get content types to hierarchical or flat representation according to given parameters
     * @param params The parameters :<ul>
     *  <li>{Boolean} excludeReferenceTable : to exclude reference table. Defaults to true.</li>
     *  <li>{Boolean} excludePrivate : to exclude private content types. Defaults to false.</li>
     *  <li>{Boolean} excludeMixin : to exclude mixins. Defaults to false.</li>
     *  <li>{Boolean} excludeAbstract : to exclude asbtract content types. Defaults to false.</li>
     *  <li>{String}  excludeMode : 'hide' or 'disabled'. The mode of exclusion. Defaults to 'disabled'.</li>
     *  <li>{Boolean} hierarchicalView: true to get a hierarchical representation of content types, or false to get a flat representation. Defaults to true.</li>
     *  <li>{List&lt;String&gt;} contentTypes: the ids of content types to restrict the list. The child content types will be also returned</li>
     *  <li>{List&lt;String&gt;} strictContentTypes: the ids of content types to restrict the list. Only these content types will be returned</li></ul>
     * @return The content types according to received parameters
     */
    @SuppressWarnings("unchecked")
    @Callable
    public Map<String, Object> getContentTypes(Map<String, Object> params)
    {
        Map<String, Object> result = new HashMap<>();
        
        boolean excludeReferenceTable = params.containsKey("excludeReferenceTable") ? (boolean) params.get("excludeReferenceTable") : true;
        boolean excludePrivate = params.containsKey("excludePrivate") ? (boolean) params.get("excludePrivate") : false;
        boolean excludeMixin = params.containsKey("excludeMixin") ? (boolean) params.get("excludeMixin") : false;
        boolean excludeAbstract = params.containsKey("excludeAbstract") ? (boolean) params.get("excludeAbstract") : false;
        boolean includeMixinOnly = params.containsKey("includeMixinOnly") ? (boolean) params.get("includeMixinOnly") : false;
        boolean includeReferenceTableOnly = params.containsKey("includeReferenceTableOnly") ? (boolean) params.get("includeReferenceTableOnly") : false;
        
        String excludeModeAsStr = params.containsKey("excludeMode") ? (String) params.get("excludeMode") : "disabled";
        ExcludeMode excludeMode = ExcludeMode.valueOf(excludeModeAsStr.toUpperCase());
        
        boolean hierarchicalView = params.containsKey("hierarchicalView") ? (boolean) params.get("hierarchicalView") : true;
        List<String> strictContentTypeIds = params.containsKey("strictContentTypes") ? (List<String>) params.get("strictContentTypes") : null;
        List<String> contentTypeIds = params.containsKey("contentTypes") ? (List<String>) params.get("contentTypes") : null;
        
        if ((strictContentTypeIds == null || strictContentTypeIds.isEmpty()) && contentTypeIds != null && !contentTypeIds.isEmpty())
        {
            strictContentTypeIds = _getRestrictedContentTypes(contentTypeIds);
        }
        
        List<Map<String, Object>> cTypes = _getContentTypes(excludeReferenceTable, excludePrivate, excludeMixin, excludeAbstract, includeMixinOnly, includeReferenceTableOnly, excludeMode, hierarchicalView, strictContentTypeIds);
        
        result.put("children", cTypes);
        return result;
    }
    
    private List<Map<String, Object>> _getContentTypes(boolean excludeReferenceTable, boolean excludePrivate, boolean excludeMixin, boolean excludeAbstract, boolean includeMixinOnly, boolean includeReferenceTableOnly, ExcludeMode excludeMode, boolean hierarchicalView, List<String> strictContentTypeIds)
    {
        List<Map<String, Object>> cTypes = new ArrayList<>();
        
        if (hierarchicalView)
        {
            Set<ContentType> rootCTypes = _getRootContentTypes();
            for (ContentType rootCType : rootCTypes)
            {
                cTypes.addAll(_getHierarchicalContentType(rootCType, strictContentTypeIds, excludeReferenceTable, excludePrivate, excludeMixin, excludeAbstract, includeMixinOnly, includeReferenceTableOnly, excludeMode));
            }
        }
        else
        {
            Set<ContentType> rootCTypes = _getContentTypes();
            for (ContentType rootCType : rootCTypes)
            {
                cTypes.addAll(_getNonHierarchicalContentType(rootCType, strictContentTypeIds, excludeReferenceTable, excludePrivate, excludeMixin, excludeAbstract, includeMixinOnly, includeReferenceTableOnly));
            }
        }
        
        return cTypes;
    }
    
    private List<String> _getRestrictedContentTypes(Collection<String> contentTypeIds)
    {
        Set<String> restrictedContentTypeIds = new HashSet<>();
        restrictedContentTypeIds.addAll(contentTypeIds);
        
        for (String contentTypeId : contentTypeIds)
        {
            restrictedContentTypeIds.addAll(_contentTypeEP.getSubTypes(contentTypeId));
        }
        
        return new ArrayList<>(restrictedContentTypeIds);
    }
    
    private List<Map<String, Object>> _getHierarchicalContentType(ContentType contentType, Collection<String> restrictedContentTypes, boolean excludeReferenceTable, boolean excludePrivate, boolean excludeMixin, boolean excludeAbstract, boolean mixinOnly, boolean refTableOnly, ExcludeMode excludeMode)
    {
        List<Map<String, Object>> cTypes = new ArrayList<>();
        
        boolean matchCondition = _matchCondition(contentType, restrictedContentTypes, excludeReferenceTable, excludePrivate, excludeMixin, excludeAbstract, mixinOnly, refTableOnly);
        if (matchCondition || excludeMode == ExcludeMode.DISABLED)
        {
            Map<String, Object> params = _getContentTypeParams(contentType, !matchCondition);
            
            List<Map<String, Object>> children = new ArrayList<>();
            Collection<String> childContentTypes = _contentTypeEP.getDirectSubTypes(contentType.getId());
            for (String id : childContentTypes)
            {
                ContentType childContentType = _contentTypeEP.getExtension(id);
                children.addAll(_getHierarchicalContentType(childContentType, restrictedContentTypes, excludeReferenceTable, excludePrivate, excludeMixin, excludeAbstract, mixinOnly, refTableOnly, excludeMode));
            }
            
            if (!children.isEmpty() || matchCondition)
            {
                params.put("leaf", children.isEmpty());
                cTypes.add(params); 
                
                if (!children.isEmpty())
                {
                    params.put("children", children);
                }
            }
          
        }
        
        return cTypes;
    }
    
    private List<Map<String, Object>> _getNonHierarchicalContentType(ContentType contentType, Collection<String> restrictedContentTypes, boolean excludeReferenceTable, boolean excludePrivate, boolean excludeMixin, boolean excludeAbstract, boolean mixinOnly, boolean refTableOnly)
    {
        List<Map<String, Object>> cTypes = new ArrayList<>();
        boolean matchCondition = _matchCondition(contentType, restrictedContentTypes, excludeReferenceTable, excludePrivate, excludeMixin, excludeAbstract, mixinOnly, refTableOnly);
        if (matchCondition)
        {
            Map<String, Object> params = _getContentTypeParams(contentType, !matchCondition);
            params.put("leaf", true);
            cTypes.add(params); 
        }
        return cTypes;
    }
    
    private Set<ContentType> _getRootContentTypes()
    {
        Set<ContentType> rootContentTypes = new HashSet<>();
        
        Set<String> ids = _contentTypeEP.getExtensionsIds();
        for (String id : ids)
        {
            ContentType contentType = _contentTypeEP.getExtension(id);
            if (contentType.getSupertypeIds().length == 0)
            {
                rootContentTypes.add(contentType);
            }
        }
        return rootContentTypes;
    }
    
    private Set<ContentType> _getContentTypes()
    {
        Set<ContentType> contentTypes = new HashSet<>();
        
        Set<String> ids = _contentTypeEP.getExtensionsIds();
        for (String id : ids)
        {
            ContentType contentType = _contentTypeEP.getExtension(id);
            contentTypes.add(contentType);
        }
        return contentTypes;
    }
    
    private boolean _matchCondition(ContentType cType, Collection<String> restrictedContentTypes, boolean excludeReferenceTable, boolean excludePrivate, boolean excludeMixin, boolean excludeAbstract, boolean mixinOnly, boolean refTableOnly)
    {
        if (cType.isReferenceTable() && excludeReferenceTable
                || cType.isPrivate() && excludePrivate
                || cType.isMixin() && excludeMixin
                || cType.isAbstract() && excludeAbstract
                || !cType.isMixin() && mixinOnly
                || !cType.isReferenceTable() && refTableOnly
                || _isContentTypeRestricted(cType, restrictedContentTypes))
        {
            return false;
        }
        
        return true;
    }
    
    private boolean _isContentTypeRestricted(ContentType cType, Collection<String> restrictedContentTypes)
    {
        return restrictedContentTypes != null && !restrictedContentTypes.isEmpty() && !restrictedContentTypes.contains(cType.getId());
    }
    
    /**
     * Get the matching content types
     * @param excludeReferenceTable true to exclude reference table
     * @param excludePrivate true to exclude the private content types
     * @param excludeAbstract true to exclude the abstract content types
     * @param excludeMixin true to exclude the mixins
     * @param mixinOnly true to include the mixins only
     * @param referenceTableOnly true to include the reference table only
     * @return matching content types
     */
    public Set<ContentType> getMatchingContentTypes(boolean excludeReferenceTable, boolean excludePrivate, boolean excludeAbstract, boolean excludeMixin, boolean mixinOnly, boolean referenceTableOnly)
    {
        return getMatchingContentTypes(null, true, excludeReferenceTable, excludePrivate, excludeAbstract, excludeMixin, mixinOnly, referenceTableOnly);
    }
    
    /**
     * Get the matching content types among the given list
     * @param contentTypesIds The list of content types to browse. Can be null to browse all available content types
     * @param browseChildren to browse the child content types
     * @param excludeReferenceTable true to exclude reference table
     * @param excludePrivate true to exclude the private content types
     * @param excludeAbstract true to exclude the abstract content types
     * @param excludeMixin true to exclude the mixins
     * @param mixinOnly true to include the mixins only
     * @param referenceTableOnly true to include the reference table only
     * @return matching content types
     */
    public Set<ContentType> getMatchingContentTypes(String[] contentTypesIds, boolean browseChildren, boolean excludeReferenceTable, boolean excludePrivate, boolean excludeAbstract, boolean excludeMixin, boolean mixinOnly, boolean referenceTableOnly)
    {
        Set<ContentType> matchingContentTypes = new HashSet<>();
        Set<String> ids;
        if (contentTypesIds != null && contentTypesIds.length != 0)
        {
            ids = new HashSet<>(Arrays.asList(contentTypesIds));
        }
        else
        {
            ids = _contentTypeEP.getExtensionsIds();
        }
        
        for (String id : ids)
        {
            ContentType contentType = _contentTypeEP.getExtension(id);
            
            if (_matchCondition(contentType, null, excludeReferenceTable, excludePrivate, excludeMixin, excludeAbstract, mixinOnly, referenceTableOnly))
            {
                matchingContentTypes.add(contentType);
            }
            
            if (browseChildren)
            {
                Collection<String> childIds = _contentTypeEP.getSubTypes(id);
                for (String childId : childIds)
                {
                    matchingContentTypes.addAll(getMatchingContentTypes(new String[]{childId}, true, excludeReferenceTable, excludePrivate, excludeAbstract, excludeMixin, mixinOnly, referenceTableOnly));
                }
            }
        }
        
        return matchingContentTypes;
    }
    
    private Map<String, Object> _getContentTypeParams(ContentType contentType, boolean disabled) 
    {
        Map<String, Object> params = new HashMap<>();
        params.put("contentTypeId", contentType.getId());
        params.put("text", contentType.getLabel());
        params.put("label", contentType.getLabel());
        params.put("description", contentType.getDescription());
        params.put("abstract", contentType.isAbstract());
        params.put("private", contentType.isPrivate());
        params.put("reference-table", contentType.isReferenceTable());
        params.put("mixin", contentType.isMixin());
        params.put("simple", contentType.isSimple());
        params.put("multilingual", contentType.isMultilingual());
        params.put("iconGlyph", contentType.getIconGlyph());
        params.put("iconDecorator", contentType.getIconDecorator());
        params.put("iconSmall", contentType.getSmallIcon());
        params.put("iconMedium", contentType.getMediumIcon());
        params.put("iconLarge", contentType.getLargeIcon());
        if (contentType instanceof AutomaticContentType)
        {
            params.put("origin", new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENTTYPE_TREE_CONF_ORIGIN"));
            params.put("glyphOrigin", "ametysicon-xml6");
        }
        else
        {
            Map<String, I18nizableTextParameter> pluginName = Collections.singletonMap("pluginName", new I18nizableText(contentType.getPluginName()));
            I18nizableText origin = new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENTTYPE_TREE_PLUGIN_ORIGIN", pluginName);
            params.put("origin", origin);
            params.put("glyphOrigin", "ametysicon-puzzle-piece1");
        }
        if (disabled)
        {
            params.put("disabled", true);
        }
        
        return params;
    }
}
