/*
 *  Copyright 2011 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.xml.AttributesImpl;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.TraversableAmetysObject;
import org.ametys.plugins.repository.UnknownAmetysObjectException;

/**
 * This handler look for links, images or video url which make reference to a AmetysObject (content, resources, ...) in a RichText.
 * It must be used after a copy of a Content to changes theses references if necessary.
 */
public class DefaultDocbookUpdateHandler extends DockbookUpdateHandler implements Serviceable
{
    private static Logger _logger = LoggerFactory.getLogger(DefaultDocbookUpdateHandler.class.getName());
    
    /** The ametys object resolver */
    protected AmetysObjectResolver _resolver;
        
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
    }
   
    @Override
    public void startElement(String uri, String loc, String raw, Attributes attrs) throws SAXException
    {
        if ("link".equals(loc))
        {
            Attributes newAttrs = _getAttributesForLink(attrs);
            super.startElement(uri, loc, raw, newAttrs);
            return;
        }
        else if ("imagedata".equals(loc) || "videodata".equals(loc))
        {
            Attributes newAttrs = _getAttributesForMediaObject(attrs);
            super.startElement(uri, loc, raw, newAttrs);
            return;
        }
        
        super.startElement(uri, loc, raw, attrs);
    }
    
    /**
     * Get attributes for link elements
     * @param attrs the attributes
     * @return the new attributes
     */
    protected Attributes _getAttributesForLink(Attributes attrs)
    {
        String href = attrs.getValue("xlink:href");
        
        String updateHref = _getUpdatedAmetysObjectId (href);
        
        AttributesImpl newAttrs = new AttributesImpl();
        _copyAttributes(attrs, newAttrs);
        newAttrs.addAttribute("", "href", "xlink:href", "CDATA", updateHref);
     
        return newAttrs;
    }
    
    /**
     * Get attributes for media objects such as image, flash
     * @param attrs the attributes
     * @return the new attributes
     */
    protected Attributes _getAttributesForMediaObject (Attributes attrs)
    {
        String href = attrs.getValue("fileref");
        String type = attrs.getValue("type");
        
        String updateHref = href;
        if (!"local".equals(type))
        {
            updateHref = _getUpdatedAmetysObjectId (href);
        }
        
        AttributesImpl newAttrs = new AttributesImpl();
        _copyAttributes(attrs, newAttrs);
        newAttrs.addAttribute("", "fileref", "fileref", "CDATA", updateHref);
     
        return newAttrs;
    }
    
    /**
     * Copy attributes
     * @param attrs the attributes to copy
     * @param newAttrs the new attributes
     */
    protected void _copyAttributes(Attributes attrs, AttributesImpl newAttrs)
    {
        for (int i = 0; i < attrs.getLength(); i++)
        {
            String name = attrs.getQName(i);
            
            if (!"xlink:href".equals(name) && !"fileref".equals(name))
            {
                newAttrs.addAttribute(attrs.getURI(i), attrs.getLocalName(i), name, attrs.getType(i), attrs.getValue(i));
            }
        }
    }
    
    /**
     * Return the updated ametys object id
     * @param id the id to update
     * @return the updated id
     */
    protected String _getUpdatedAmetysObjectId (String id)
    {
        try
        {
            if (id.equals(_initialContent.getId()))
            {
                return _createdContent.getId();
            }
            else if (_resolver.hasAmetysObjectForId(id))
            {
                AmetysObject ametysObject = _resolver.resolveById(id);
                String path = ametysObject.getPath();
                
                String initialContentPath = _initialContent.getPath();
                
                if (path.startsWith(initialContentPath))
                {
                    // Is in path
                    String relPath = path.equals(initialContentPath) ? "" : path.substring(initialContentPath.length() + 1);
                    try
                    {
                        // Find symmetric object on copied sub-tree
                        if (_createdContent instanceof TraversableAmetysObject)
                        {
                            AmetysObject child = "".equals(relPath) ? _createdContent : ((TraversableAmetysObject) _createdContent).getChild(relPath);
                            return child.getId();
                        }
                    }
                    catch (UnknownAmetysObjectException e)
                    {
                        _logger.warn("Object of path " + relPath + " was not found on copied sub-tree " + _createdContent.getPath(), e);
                    }
                }
                else if (path.equals(_initialContentPath))
                {
                    // Return id of the new content
                    return _createdContent.getId();
                }
            }
            
            // Return same id
            return id;
        }
        catch (UnknownAmetysObjectException e)
        {
            // Return same id
            return id;
        }
    }
}
