/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Set;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.sax.TransformerHandler;
import javax.xml.transform.stream.StreamResult;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfigurationSerializer;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.source.impl.FileSource;
import org.apache.xml.serializer.OutputPropertiesFactory;
import org.xml.sax.SAXException;

import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ContentDAO;
import org.ametys.cms.repository.ContentQueryHelper;
import org.ametys.cms.repository.ContentTypeOrMixinExpression;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.model.CompositeDefinition;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.plugins.repository.query.expression.Expression;
import org.ametys.plugins.repository.query.expression.Expression.Operator;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Enumerator;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemContainer;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.SimpleViewItemGroup;
import org.ametys.runtime.model.StaticEnumerator;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemContainer;
import org.ametys.runtime.parameter.DefaultValidator;
import org.ametys.runtime.parameter.Validator;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Helper for create, edit and remove a content type
 */
public class EditContentTypeHelper extends AbstractLogEnabled implements Component, Serviceable
{
    /** The Avalon role name */
    public static final String ROLE = EditContentTypeHelper.class.getName();

    /** The directory path of content types */
    private static final String __CONTENT_TYPES_DIR = "context://WEB-INF/param/content-types/";
    /** The prefix for id of automatic content types */
    private static final String __CONTENT_TYPE_PREFIX = "content-type.";

    /** The namespace uri */
    private static final String __NAMESPACE_URI = "http://www.ametys.org/schema/cms";

    /** The source resolver */
    protected SourceResolver _sourceResolver;

    /** The ametys object resolver instance */
    protected AmetysObjectResolver _ametysObjectResolver;

    /** The content type extension point instance */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;

    /** DAO for contents */
    protected ContentDAO _contentDAO;

    public void service(ServiceManager manager) throws ServiceException
    {
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentDAO = (ContentDAO) manager.lookup(ContentDAO.ROLE);
    }

    /**
     * Get the root directory for all custom content types. The directory is created if does not exist.
     * @return the root directory for all custom content types.
     * @throws MalformedURLException if failed to get the root directory
     * @throws IOException if failed to get the root directory
     */
    public File getContentTypeRootDirectory() throws MalformedURLException, IOException
    {
        FileSource rootSource = null;
        try
        {
            rootSource = (FileSource) _sourceResolver.resolveURI(__CONTENT_TYPES_DIR);
            File rootDir = rootSource.getFile();
            if (!rootDir.exists())
            {
                rootDir.mkdir();
            }
            return rootDir;
        }
        finally
        {
            _sourceResolver.release(rootSource);
        }
    }
    
    /**
     * Get the root directory for all archived content types. The directory is created if does not exist.
     * @return the root directory for archived content types.
     * @throws MalformedURLException if failed to get the root directory
     * @throws IOException if failed to get the root directory
     */
    public File getArchiveRootDirectory() throws MalformedURLException, IOException
    {
        File rootDir = getContentTypeRootDirectory();
        
        File archiveDir = new File(rootDir, "_archives");
        if (!archiveDir.exists())
        {
            archiveDir.mkdir();
        }
        
        return archiveDir;
    }
    
    /**
     * Get the root archive directory for a given content type. The directory is created if does not exist.
     * @param contentType The content type
     * @return the root archive directory for this content type.
     * @throws MalformedURLException if failed to get the archive directory
     * @throws IOException if failed to get the archive directory
     */
    public File getArchiveRootDirectory(ContentType contentType) throws MalformedURLException, IOException
    {
        File archiveDir = getArchiveRootDirectory();
        
        File cTypeDir = new File(archiveDir, contentType.getPluginName());
        if (!cTypeDir.exists())
        {
            cTypeDir.mkdir();
        }
        
        return cTypeDir;
    }
    
    /**
     * Get the root directory for a given content type. The directory is created if does not exist.
     * @param contentType The content type definition
     * @param override true if this is a overriden content type.
     * @return the root directory for this content type.
     * @throws MalformedURLException if failed to get the archive directory
     * @throws IOException if failed to get the archive directory
     */
    public File getContentTypeDirectory(ContentTypeDefinition contentType, boolean override) throws MalformedURLException, IOException
    {
        File rootDir = getContentTypeRootDirectory();
        
        File cTypeDir = new File(rootDir, override ? "_override" : contentType.getPluginName());
        if (!cTypeDir.exists())
        {
            cTypeDir.mkdir();
        }
        
        return cTypeDir;
    }
    
    /**
     * Get the root directory for a given content type. The directory is created if does not exist.
     * @param contentType The content type
     * @return the root directory for this content type.
     * @throws MalformedURLException if failed to get the archive directory
     * @throws IOException if failed to get the archive directory
     */
    public File getContentTypeDirectory(ContentType contentType) throws MalformedURLException, IOException
    {
        File rootDir = getContentTypeRootDirectory();
        
        File cTypeDir = new File(rootDir, contentType.getPluginName());
        if (!cTypeDir.exists())
        {
            cTypeDir.mkdir();
        }
        
        return cTypeDir;
    }
    
    /**
     * Get the XML definition file for a given content type.
     * @param contentType The content type definition
     * @param override true if this is a overriden content type.
     * @return the XML definition file for this content type.
     * @throws MalformedURLException if failed to get the archive directory
     * @throws IOException if failed to get the archive directory
     */
    public File getContentTypeFile(ContentTypeDefinition contentType, boolean override) throws MalformedURLException, IOException
    {
        File rootCTypeDir = getContentTypeDirectory(contentType, override);
        
        String id = contentType.getId().replaceFirst(__CONTENT_TYPE_PREFIX, "");
        return new File(rootCTypeDir, id + ".xml");
    }
    
    /**
     * Get the XML definition file for a given content type.
     * @param contentType The content type
     * @return the XML definition file for this content type.
     * @throws MalformedURLException if failed to get the archive directory
     * @throws IOException if failed to get the archive directory
     */
    public File getContentTypeFile(ContentType contentType) throws MalformedURLException, IOException
    {
        File rootCTypeDir = getContentTypeDirectory(contentType);
        
        String id = contentType.getId().replaceFirst(__CONTENT_TYPE_PREFIX, "");
        return new File(rootCTypeDir, id + ".xml");
    }
    
    /**
     * Create a new content type. 
     * This method creates a new xml file into WEB-INF/param/content-types folder.
     * @param contentTypeDef The content type definition
     * @throws EditContentTypeException if failed to create the content type
     */
    public void createContentType(ContentTypeDefinition contentTypeDef) throws EditContentTypeException
    {
        String id = contentTypeDef.getId().replaceFirst(__CONTENT_TYPE_PREFIX, "");
        String realId = __CONTENT_TYPE_PREFIX + id;

        Set<String> extensionsIds = _contentTypeExtensionPoint.getExtensionsIds();
        if (!extensionsIds.contains(realId))
        {
            if (StringUtils.isNotBlank(contentTypeDef.getPluginName()))
            {
                _saxContentType(contentTypeDef, false);
            }
            else
            {
                throw new EditContentTypeException("The content type '" + contentTypeDef.getId() + "' can't be created without plugin name.");
            }
        }
        else
        {
            throw new EditContentTypeException("The content type '" + contentTypeDef.getId() + "' can't be created because it already exists.");
        }
    }

    /**
     * Modify a existing content type. 
     * This method edit the corresponding XML file of the content type in WEB-INF/param/content-types folder.
     * @param contentTypeDef The content type definition
     * @throws EditContentTypeException if failed to edit this content type
     */
    public void editContentType(ContentTypeDefinition contentTypeDef) throws EditContentTypeException
    {
        // For the content type id we consider that it is the real id, for
        // example for content type in web-inf the id is prefixed by
        // "content-type."
        Set<String> extensionsIds = _contentTypeExtensionPoint.getExtensionsIds();
        if (extensionsIds.contains(contentTypeDef.getId()))
        {
            boolean override = !_isAutomaticContentType(contentTypeDef.getId());
            _saxContentType(contentTypeDef, override);
        }
        else
        {
            throw new EditContentTypeException("The content type '" + contentTypeDef.getId() + "' can't be edited because this content type doesn't exist.");
        }
    }

    private void _saxContentType(ContentTypeDefinition contentTypeDef, boolean override) throws EditContentTypeException
    {
        try 
        {
            File cTypeFile = getContentTypeFile(contentTypeDef, override);

            // Sax the content type in file
            try (OutputStream os = Files.newOutputStream(Paths.get(cTypeFile.getAbsolutePath())))
            {
                // Create a transformer for saving sax into a file
                TransformerHandler handler = ((SAXTransformerFactory) TransformerFactory.newInstance()).newTransformerHandler();

                StreamResult result = new StreamResult(os);
                handler.setResult(result);

                // Create the format of result
                Properties format = new Properties();
                format.put(OutputKeys.METHOD, "xml");
                format.put(OutputKeys.INDENT, "yes");
                format.put(OutputKeys.ENCODING, "UTF-8");
                format.put(OutputPropertiesFactory.S_KEY_INDENT_AMOUNT, "4");
                handler.getTransformer().setOutputProperties(format);

                try
                {
                    _saxContentType(contentTypeDef, override, handler);
                }
                catch (SAXException | ConfigurationException e)
                {
                    throw new EditContentTypeException("Error when trying to sax the content type '" + contentTypeDef.getId() + "'", e);
                }
            }
            catch (IOException | TransformerConfigurationException | TransformerFactoryConfigurationError e)
            {
                throw new EditContentTypeException(e);
            }
        }
        catch (IOException e)
        {
            throw new EditContentTypeException(e.getMessage(), e);
        }
    }

    private void _saxContentType(ContentTypeDefinition contentTypeDef, boolean override, TransformerHandler handler)
            throws SAXException, EditContentTypeException, ConfigurationException
    {
        handler.startDocument();
        handler.startPrefixMapping("cms", __NAMESPACE_URI);

        if (override)
        {
            XMLUtils.startElement(handler, "cms:content-type");
        }
        else
        {
            AttributesImpl attributes = new AttributesImpl();

            if (contentTypeDef.isAbstract())
            {
                attributes.addCDATAAttribute("abstract", "true");
            }
            if (contentTypeDef.getSupertypeIds() != null && contentTypeDef.getSupertypeIds().length != 0)
            {
                attributes.addCDATAAttribute("extends", String.join(",", contentTypeDef.getSupertypeIds()));
            }
            XMLUtils.startElement(handler, "cms:content-type", attributes);

            _saxGeneralInformation(contentTypeDef, handler);
            _saxIcons(contentTypeDef.getIconGlyph(), handler);
            _saxTags(contentTypeDef.getTags(), handler);
            _saxParentRef(contentTypeDef.getParentRef(), handler);
            _saxRight(contentTypeDef.getRight(), handler);
        }

        _saxModelItems(contentTypeDef, handler);
        _saxViews(contentTypeDef, handler);

        XMLUtils.endElement(handler, "cms:content-type");
        handler.endDocument();
    }

    private void _saxGeneralInformation(ContentTypeDefinition contentTypeDef, TransformerHandler handler) throws SAXException, EditContentTypeException
    {
        I18nizableText label = contentTypeDef.getLabel();
        if (label == null)
        {
            throw new EditContentTypeException("The content type '" + contentTypeDef.getId() + "' can't be created without label");
        }
        else
        {
            _saxI18nizableText(handler, "cms:label", label);
        }

        _saxI18nizableText(handler, "cms:description", contentTypeDef.getDescription());
        
        I18nizableText defaultTitle = contentTypeDef.getDefaultTitle();
        if (defaultTitle == null)
        {
            throw new EditContentTypeException("The content type '" + contentTypeDef.getId() + "' can't be created without default title");
        }
        else
        {
            _saxI18nizableText(handler, "cms:default-title", defaultTitle);
        }
        
        _saxI18nizableText(handler, "cms:category", contentTypeDef.getCategory());
    }

    private void _saxIcons(String iconGlyph, TransformerHandler handler) throws SAXException
    {
        if (StringUtils.isNotBlank(iconGlyph))
        {
            XMLUtils.startElement(handler, "cms:icons");
            XMLUtils.createElement(handler, "cms:glyph", iconGlyph);
            XMLUtils.endElement(handler, "cms:icons");
        }
    }

    private void _saxTags(Set<String> tags, TransformerHandler handler) throws SAXException
    {
        if (tags != null && !tags.isEmpty())
        {
            XMLUtils.startElement(handler, "cms:tags");

            for (String tag : tags)
            {
                XMLUtils.createElement(handler, "cms:tag", tag);
            }

            XMLUtils.endElement(handler, "cms:tags");
        }
    }

    private void _saxParentRef(String parentRef, TransformerHandler handler) throws SAXException
    {
        if (StringUtils.isNotBlank(parentRef))
        {
            AttributesImpl attributes = new AttributesImpl();
            attributes.addCDATAAttribute("name", parentRef);
            XMLUtils.createElement(handler, "cms:parent-ref", attributes);
        }
    }

    private void _saxRight(String right, TransformerHandler handler) throws SAXException
    {
        if (StringUtils.isNotBlank(right))
        {
            XMLUtils.createElement(handler, "cms:right", right);
        }
    }

    private void _saxModelItems(ContentTypeDefinition contentTypeDef, TransformerHandler handler) throws SAXException, ConfigurationException
    {
        if (contentTypeDef.getModelItems() != null)
        {
            for (ModelItem modelItem : contentTypeDef.getModelItems())
            {
                _saxModelItem(modelItem, handler);
            }
        }
    }

    @SuppressWarnings("unchecked")
    private void _saxModelItem(ModelItem modelItem, TransformerHandler handler) throws SAXException, ConfigurationException
    {
        if (modelItem != null)
        {
            if (modelItem.getType() != null)
            {
                if (modelItem instanceof RepeaterDefinition)
                {
                    _saxRepeater((RepeaterDefinition) modelItem, handler);
                }
                else if (modelItem instanceof CompositeDefinition)
                {
                    if ("dc".equals(modelItem.getName())) // Dublin core exception
                    {
                        XMLUtils.createElement(handler, "cms:dublin-core");
                    }
                    else
                    {
                        _saxComposite((CompositeDefinition) modelItem, handler);
                    }
                }
                else if (modelItem instanceof ElementDefinition)
                {
                    _saxElementDefinition((ElementDefinition) modelItem, handler);
                }
            }
        }
    }

    private void _saxRepeater(RepeaterDefinition repeater, TransformerHandler handler) throws SAXException, ConfigurationException
    {
        AttributesImpl attributes = new AttributesImpl();
        attributes.addCDATAAttribute("name", repeater.getName());

        String initialSize = Integer.toString(repeater.getInitialSize());
        String minSize = Integer.toString(repeater.getMinSize());
        String maxSize = Integer.toString(repeater.getMaxSize());
        if (StringUtils.isNotBlank(initialSize))
        {
            attributes.addCDATAAttribute("initial-size", initialSize);
        }
        if (StringUtils.isNotBlank(minSize))
        {
            attributes.addCDATAAttribute("min-size", minSize);
        }
        if (StringUtils.isNotBlank(maxSize))
        {
            attributes.addCDATAAttribute("max-size", maxSize);
        }

        XMLUtils.startElement(handler, "cms:repeater", attributes);

        _saxI18nizableText(handler, "cms:label", repeater.getLabel());
        _saxI18nizableText(handler, "cms:description", repeater.getDescription());
        _saxI18nizableText(handler, "cms:add-label", repeater.getAddLabel());
        _saxI18nizableText(handler, "cms:del-label", repeater.getDeleteLabel());
        _saxChildrenModelItems(repeater, handler);

        XMLUtils.endElement(handler, "cms:repeater");
    }

    private void _saxComposite(CompositeDefinition composite, TransformerHandler handler) throws SAXException, ConfigurationException
    {
        AttributesImpl attributes = new AttributesImpl();
        attributes.addCDATAAttribute("name", composite.getName());
        attributes.addCDATAAttribute("type", composite.getType().getId());

        XMLUtils.startElement(handler, "cms:metadata", attributes);

        _saxI18nizableText(handler, "cms:label", composite.getLabel());
        _saxI18nizableText(handler, "cms:description", composite.getDescription());
        _saxChildrenModelItems(composite, handler);

        XMLUtils.endElement(handler, "cms:metadata");
    }

    private void _saxChildrenModelItems(ModelItemContainer modelItemContainer, TransformerHandler handler) throws ConfigurationException, SAXException
    {
        if (modelItemContainer.getModelItems() != null)
        {
            for (ModelItem modelItem : modelItemContainer.getModelItems())
            {
                _saxModelItem(modelItem, handler);
            }
        }
    }
    
    private void _saxElementDefinition(ElementDefinition<? extends Object> elementDefinition, TransformerHandler handler) throws SAXException, ConfigurationException
    {
        AttributesImpl attributes = new AttributesImpl();
        attributes.addCDATAAttribute("name", elementDefinition.getName());
        attributes.addCDATAAttribute("multiple", Boolean.toString(elementDefinition.isMultiple()));
        attributes.addCDATAAttribute("type", elementDefinition.getType().getId());

        if (elementDefinition instanceof ContentAttributeDefinition)
        {
            ContentAttributeDefinition contentAttributeDefinition = (ContentAttributeDefinition) elementDefinition;
            String linkedContentTypeId = contentAttributeDefinition.getContentTypeId();
            if (linkedContentTypeId != null)
            {
                attributes.addCDATAAttribute("contentType", linkedContentTypeId);
                if (contentAttributeDefinition.getInvertRelationPath() != null)
                {
                    attributes.addCDATAAttribute("invert", contentAttributeDefinition.getInvertRelationPath());
                }
            }
        }
        
        XMLUtils.startElement(handler, "cms:metadata", attributes);

        _saxI18nizableText(handler, "cms:label", elementDefinition.getLabel());
        _saxI18nizableText(handler, "cms:description", elementDefinition.getDescription());
        _saxWidget(elementDefinition, handler);
        _saxEnumeration(elementDefinition, handler);
        _saxValidator(elementDefinition, handler);

        XMLUtils.endElement(handler, "cms:metadata");
    }

    private void _saxWidget(ElementDefinition<? extends Object> elementDefinition, TransformerHandler handler) throws SAXException
    {
        if (StringUtils.isNotBlank(elementDefinition.getWidget()))
        {
            XMLUtils.createElement(handler, "widget", elementDefinition.getWidget());
        }

        if (elementDefinition.getWidgetParameters() != null)
        {
            XMLUtils.startElement(handler, "widget-params");

            for (Entry<String, I18nizableText> widgetParam : elementDefinition.getWidgetParameters().entrySet())
            {
                if (widgetParam.getValue() != null)
                {
                    AttributesImpl attributes = new AttributesImpl();
                    attributes.addCDATAAttribute("name", widgetParam.getKey());
                    _saxI18nizableText(handler, "param", widgetParam.getValue(), attributes);
                }
            }

            XMLUtils.endElement(handler, "widget-params");
        }
    }

    private void _saxEnumeration(ElementDefinition elementDefinition, TransformerHandler handler) throws SAXException, ConfigurationException
    {
        Enumerator enumerator = elementDefinition.getEnumerator();
        if (enumerator != null && enumerator instanceof StaticEnumerator)
        {
            _saxDefaultEnumerator(elementDefinition, handler);
        }
        else if (StringUtils.isNotBlank(elementDefinition.getCustomEnumerator()))
        {
            _saxCustomEnumerator(elementDefinition, handler);
        }
    }

    private void _saxDefaultEnumerator(ElementDefinition elementDefinition, TransformerHandler handler)
    {
        try
        {
            Map<Object, I18nizableText> entries = elementDefinition.getEnumerator().getTypedEntries();

            if (entries != null)
            {
                // Enumerated values
                XMLUtils.startElement(handler, "enumeration");
                for (Entry<Object, I18nizableText> entry : entries.entrySet())
                {
                    if (entry != null)
                    {
                        if (entry.getValue() != null)
                        {
                            XMLUtils.startElement(handler, "entry");
                            XMLUtils.createElement(handler, "value", entry.getKey().toString());
                            _saxI18nizableText(handler, "label", entry.getValue());
                            XMLUtils.endElement(handler, "entry");
                        }
                    }
                }
                XMLUtils.endElement(handler, "enumeration");
            }
        }
        catch (Exception e)
        {
            // Silently ignore
        }
    }

    private void _saxCustomEnumerator(ElementDefinition elementDefinition, TransformerHandler handler) throws SAXException, ConfigurationException
    {
        XMLUtils.startElement(handler, "enumeration");

        AttributesImpl attributes = new AttributesImpl();
        attributes.addCDATAAttribute("class", elementDefinition.getCustomEnumerator());
        XMLUtils.startElement(handler, "custom-enumerator", attributes);

        if (elementDefinition.getEnumeratorConfiguration() != null)
        {
            DefaultConfigurationSerializer configurationSerializer = new DefaultConfigurationSerializer();
            configurationSerializer.serialize(handler, elementDefinition.getEnumeratorConfiguration());
        }

        XMLUtils.endElement(handler, "custom-enumerator");

        XMLUtils.endElement(handler, "enumeration");
    }

    private void _saxValidator(ElementDefinition elementDefinition, TransformerHandler handler) throws SAXException, ConfigurationException
    {
        Validator validator = elementDefinition.getValidator();
        if (validator != null && validator.getClass().equals(DefaultValidator.class))
        {
            _saxDefaultValidator(elementDefinition, handler);
        }
        else if (StringUtils.isNotBlank(elementDefinition.getCustomValidator()))
        {
            _saxCustomValidator(elementDefinition, handler);
        }
    }

    private void _saxDefaultValidator(ElementDefinition elementDefinition, TransformerHandler handler) throws SAXException
    {
        Validator validator = elementDefinition.getValidator();

        XMLUtils.startElement(handler, "validation");

        Map<String, Object> configuration = validator.getConfiguration();

        // Mandatory traitment
        Object mandatory = configuration.get("mandatory");
        if (mandatory != null && mandatory.toString().equals("true"))
        {
            XMLUtils.createElement(handler, "mandatory");
        }
        // Regexp traitment
        Object regexp = configuration.get("regexp");
        if (regexp != null)
        {
            XMLUtils.createElement(handler, "regexp", regexp.toString());
        }
        // Invalid text traitment
        Object invalidTextAsObject = configuration.get("invalidText");
        if (invalidTextAsObject != null)
        {
            I18nizableText invalidTextAsI18nizableText = (I18nizableText) invalidTextAsObject;
            _saxI18nizableText(handler, "invalidText", invalidTextAsI18nizableText);
        }

        XMLUtils.endElement(handler, "validation");
    }

    private void _saxCustomValidator(ElementDefinition elementDefinition, TransformerHandler handler) throws SAXException, ConfigurationException
    {
        XMLUtils.startElement(handler, "validation");

        AttributesImpl attributes = new AttributesImpl();
        attributes.addCDATAAttribute("class", elementDefinition.getCustomValidator());
        XMLUtils.startElement(handler, "custom-validator", attributes);

        if (elementDefinition.getValidatorConfiguration() != null)
        {
            DefaultConfigurationSerializer configurationSerializer = new DefaultConfigurationSerializer();
            configurationSerializer.serialize(handler, elementDefinition.getValidatorConfiguration());
        }

        XMLUtils.endElement(handler, "custom-validator");

        XMLUtils.endElement(handler, "validation");
    }

    private void _saxViews(ContentTypeDefinition contentType, TransformerHandler handler) throws SAXException
    {
        List<View> views = contentType.getViews();
        if (views != null)
        {
            for (View view : views)
            {
                _saxView(view, handler);
            }
        }
        else
        {
            _saxDefaultViews(handler);
        }
    }

    private void _saxDefaultViews(TransformerHandler handler) throws SAXException
    {
        _saxDefaultView("main", "view", handler);
        _saxDefaultView("main", "edition", handler);
        _saxDefaultView("abstract", "view", handler);
        _saxDefaultView("link", "view", handler);
        _saxDefaultView("details", "view", handler);
        _saxDefaultView("index", "view", handler);
    }

    private void _saxDefaultView(String name, String view, TransformerHandler handler) throws SAXException
    {
        AttributesImpl attributes = new AttributesImpl();
        attributes.addCDATAAttribute("name", name);
        attributes.addCDATAAttribute("type", view);
        XMLUtils.createElement(handler, "cms:metadata-set", attributes);
    }
    
    private void _saxView(View view, TransformerHandler handler) throws SAXException
    {
        AttributesImpl attributes = new AttributesImpl();
        attributes.addCDATAAttribute("name", view.getName());
        attributes.addCDATAAttribute("type", "view");
        attributes.addCDATAAttribute("internal", Boolean.toString(view.isInternal()));
        
        XMLUtils.startElement(handler, "cms:metadata-set", attributes);
        
        _saxI18nizableText(handler, "cms:label", view.getLabel());
        _saxI18nizableText(handler, "cms:description", view.getDescription());
        for (ViewItem viewItem : view.getViewItems())
        {
            _saxViewItem(viewItem, handler);
        }
        
        XMLUtils.endElement(handler, "cms:metadata-set");
    }

    private void _saxViewItem(ViewItem viewItem, TransformerHandler handler) throws SAXException
    {
        if (viewItem instanceof ModelViewItem)
        {
            ModelViewItem modelViewItem = (ModelViewItem) viewItem;
            
            if ("dc".equals(modelViewItem.getDefinition().getName())) // dublin core exception
            {
                XMLUtils.createElement(handler, "cms:dublin-core");
            }
            else
            {
                AttributesImpl attributes = new AttributesImpl();
                attributes.addCDATAAttribute("name", modelViewItem.getDefinition().getName());
                
                XMLUtils.startElement(handler, "cms:metadata-ref", attributes);
                if (modelViewItem instanceof ViewItemContainer)
                {
                    for (ViewItem child : ((ViewItemContainer) modelViewItem).getViewItems())
                    {
                        _saxViewItem(child, handler);
                    }
                }
                XMLUtils.endElement(handler, "cms:metadata-ref");
            }
        }
        else if (viewItem instanceof SimpleViewItemGroup)
        {
            SimpleViewItemGroup simpleViewItemGroup = (SimpleViewItemGroup) viewItem;
            AttributesImpl attributes = new AttributesImpl();
            attributes.addCDATAAttribute("role", simpleViewItemGroup.getRole());
            
            XMLUtils.startElement(handler, "cms:fieldset", attributes);
            
            _saxI18nizableText(handler, "cms:label", simpleViewItemGroup.getLabel());
            for (ViewItem child : simpleViewItemGroup.getViewItems())
            {
                _saxViewItem(child, handler);
            }
            
            XMLUtils.endElement(handler, "cms:fieldset");
        }
    }

    private void _saxI18nizableText(TransformerHandler handler, String tagName, I18nizableText text) throws SAXException
    {
        _saxI18nizableText(handler, tagName, text, new AttributesImpl());
    }
    
    private void _saxI18nizableText(TransformerHandler handler, String tagName, I18nizableText text, AttributesImpl attrs) throws SAXException
    {
        if (text != null)
        {
            attrs.addCDATAAttribute("i18n", Boolean.toString(text.isI18n()));
            
            String value = "";
            
            if (text.isI18n())
            {
                if (StringUtils.isNotBlank(text.getCatalogue()))
                {
                    value = text.getCatalogue() + ":";
                }
                value = value + text.getKey();
            }
            else
            {
                value = text.getLabel();
            }
            
            XMLUtils.createElement(handler, tagName, attrs, value);
        }
    }

    /**
     * Determines if the content type can be removed
     * @param contentTypeId The id of content type
     * @return true if the content type can be removed
     */
    public boolean canRemoveContentType(String contentTypeId)
    {
        return _isAutomaticContentType(contentTypeId)
                && _isLeafContentType(contentTypeId)
                && !_isInUseContentType(contentTypeId);
    }
    
    /**
     * Remove a existing content type defined in WEB-INF/param/content-types directory.<br>
     * This method archived the content type into the WEB-INF/param/content-types/_archives/pluginName folder.
     * @param contentTypeId the id of content type to remove
     * @throws RemoveContentTypeException if the content type could not be removed
     */
    public void removeContentType(String contentTypeId) throws RemoveContentTypeException
    {
        ContentType contentType = _contentTypeExtensionPoint.getExtension(contentTypeId);
        
        if (contentType == null)
        {
            throw new RemoveContentTypeException("The content type with id '" + contentTypeId + "' can't be removed because this content type doesn't exist.");
        }
        
        if (!_isAutomaticContentType(contentTypeId))
        {
            throw new RemoveContentTypeException("The content type with id '" + contentTypeId + "' can't be removed because it is not a modifiable content type");
        }
        
        if (!_isLeafContentType(contentTypeId))
        {
            throw new RemoveContentTypeException("The content type with id '" + contentTypeId + "' can't be removed because it has one or more sub-types");
        }
        
        if (_isInUseContentType(contentTypeId))
        {
            throw new RemoveContentTypeException("The content type with id '" + contentTypeId + "' can't be removed because it was currently in use");
        }
        
        // Archive the content type
        _archiveContentType(contentType);
    }
    
    private void _archiveContentType(ContentType contentType) throws RemoveContentTypeException
    {
        try
        {
            File archiveRootDir = getArchiveRootDirectory(contentType);
            
            String realId = contentType.getId().replaceFirst(__CONTENT_TYPE_PREFIX, "");
            File cTypeFile = getContentTypeFile(contentType);
            
            File destFile = new File(archiveRootDir, realId + ".xml");
            
            int index = 2;
            while (destFile.exists())
            {
                destFile = new File(archiveRootDir, realId + "-" + index + ".xml");
                index++;
            }
            
            FileUtils.moveFile(cTypeFile, destFile);
        }
        catch (IOException e)
        {
            throw new RemoveContentTypeException("Fail to archive content type " + contentType.getId(), e);
        }
    }
    
    private boolean _isInUseContentType(String contentTypeId)
    {
        Expression cTypeExpr = new ContentTypeOrMixinExpression(Operator.EQ, contentTypeId);
        String xPathQuery = ContentQueryHelper.getContentXPathQuery(cTypeExpr);
        AmetysObjectIterable<Content> contents = _ametysObjectResolver.query(xPathQuery);
        return contents.iterator().hasNext();
    }

    private boolean _isAutomaticContentType(String contentTypeId)
    {
        ContentType contentType = _contentTypeExtensionPoint.getExtension(contentTypeId);
        return contentType instanceof AutomaticContentType;
    }

    private boolean _isLeafContentType(String contentTypeId)
    {
        return _contentTypeExtensionPoint.getDirectSubTypes(contentTypeId).size() == 0;
    }

}
