/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.commons.lang.BooleanUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.ObjectUtils;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.repository.DefaultContent;
import org.ametys.core.cocoon.JSonReader;
import org.ametys.core.util.I18nUtils;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelItem;

/**
 * Get the common attributes between given content types and/or among given contents
 */
public class GetCommonAttributesAction extends ServiceableAction
{
    
    /** Not sortable attribute types */
    @SuppressWarnings("static-access")
    protected static final String[] __NOT_SORTABLE_TYPE_IDS =
    {
        ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID,
        ModelItemTypeConstants.FILE_ELEMENT_TYPE_ID,
        ModelItemTypeConstants.REFERENCE_ELEMENT_TYPE_ID,
        ModelItemTypeConstants.COMPOSITE_TYPE_ID
    };
    
    /** Allow enumerated attributes */
    protected static final String __ALLOW_ENUMERATED = "ENUMERATED";
    
    /** Allow not enumerated attributes */
    protected static final String __ALLOW_NOT_ENUMERATED = "NOT-ENUMERATED";

    /** The content type EP */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    
    /** The content types Helper */
    protected ContentTypesHelper _contentTypesHelper;
    
    /** I18n utils */
    protected I18nUtils _i18nUtils;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
        _contentTypesHelper = (ContentTypesHelper) smanager.lookup(ContentTypesHelper.ROLE);
        _i18nUtils = (I18nUtils) smanager.lookup(I18nUtils.ROLE);
    }
    
    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        String viewName = request.getParameter("viewName");
        List<String> acceptedTypes = Arrays.asList(ObjectUtils.defaultIfNull(request.getParameterValues("acceptedTypes"), ArrayUtils.EMPTY_STRING_ARRAY));
        boolean onlySortable = BooleanUtils.toBoolean(request.getParameter("onlySortable")); // false by default
        boolean includeComposite = BooleanUtils.toBoolean(request.getParameter("includeComposite")); // false by default
        boolean includeSubRepeaters = !"false".equals(request.getParameter("includeSubRepeaters")); // true by default
        boolean withFullLabel = BooleanUtils.toBoolean(request.getParameter("withFullLabel")); // false by default
        boolean withLastModified = BooleanUtils.toBoolean(request.getParameter("withLastModified")); // false by default
        boolean withLastValidation = BooleanUtils.toBoolean(request.getParameter("withLastValidation")); // false by default
        boolean withCreationDate = BooleanUtils.toBoolean(request.getParameter("withCreationDate")); // false by default
        boolean withResources = BooleanUtils.toBoolean(request.getParameter("withResources")); // false by default
        
        Set<String> contentTypeIds = getContentTypes(request, withResources);
        boolean hasResources = contentTypeIds.contains("resource");
        
        Map<String, Object> result = new HashMap<>();
        
        List<Map<String, Object>> commonModelItemsInfo = getCommonModelItemsInfo(contentTypeIds, viewName, acceptedTypes, onlySortable, includeComposite, withFullLabel, hasResources, includeSubRepeaters);
        handleSpecificInfo(commonModelItemsInfo, withLastModified, withLastValidation, withCreationDate, hasResources, withFullLabel);
        result.put("attributes", commonModelItemsInfo);
        
        request.setAttribute(JSonReader.OBJECT_TO_READ, result);
        return EMPTY_MAP;
    }
    
    /**
     * Retrieve model items information for a list of content types
     * @param contentTypeIds the identifiers of the content types
     * @param viewName The view name to list model items
     * @param acceptedTypes The types to accept. All types accepted if null or empty.
     * @param onlySortable true to only accept sortable model items.
     * @param includeCompositesAndRepeaters true to include the composites and repeaters as well
     * @param withFullLabel true to use the full label of the model items, false otherwise
     * @param hasResources true if there is at least one content type that is a resource, false otherwise.
     * @param includeSubRepeaters true to include the children of repeaters as well
     * @return The list of common model items properties (each set of properties is a map representing the model item)
     */
    protected List<Map<String, Object>> getCommonModelItemsInfo(Collection<String> contentTypeIds, String viewName, List<String> acceptedTypes, boolean onlySortable, boolean includeCompositesAndRepeaters, boolean withFullLabel, boolean hasResources, boolean includeSubRepeaters)
    {
        List<Map<String, Object>> commonModelItemsInfos = new LinkedList<>();
        
        if (hasResources)
        {
            // no common if resource selected
            return commonModelItemsInfos;
        }
        
        Map<String, ModelItem> modelItems = _contentTypesHelper.getCommonModelItems(contentTypeIds, viewName);

        for (ModelItem modelItem : modelItems.values())
        {
            if (_modelItemPassesConditions(modelItem, acceptedTypes, onlySortable, includeCompositesAndRepeaters, includeSubRepeaters))
            {
                Map<String, Object> metadataInfo = getModelItemInfo(modelItem, withFullLabel);
                commonModelItemsInfos.add(metadataInfo);
            }
        }
        
        return commonModelItemsInfos;
    }
    
    /**
     * Indicates if the attribute should be retrieved or filtered out given the options
     * @param modelItem the attribute definition
     * @param acceptedTypes The types to accept. All types accepted if null or empty.
     * @param onlySortable true to only accept sortable attributes.
     * @param includeCompositeAndRepeaters true to include the composites and repeaters as well
     * @param includeSubRepeaters true to include the children of repeaters as well
     * @return true if the model item passes the conditions
     */
    @SuppressWarnings("static-access")
    protected boolean _modelItemPassesConditions(ModelItem modelItem, List<String> acceptedTypes, boolean onlySortable, boolean includeCompositeAndRepeaters, boolean includeSubRepeaters)
    {
        String typeId = modelItem.getType().getId();
        boolean passesCondition = true;
        
        if (acceptedTypes != null && !acceptedTypes.isEmpty() && !(acceptedTypes.size() == 1 && StringUtils.isEmpty(acceptedTypes.get(0))))
        {
            passesCondition = _passesAcceptedTypesCondition(acceptedTypes, modelItem);
        }
        
        if (passesCondition && onlySortable)
        {
            passesCondition = !ArrayUtils.contains(__NOT_SORTABLE_TYPE_IDS, typeId);
        }
        
        if (passesCondition && !includeCompositeAndRepeaters)
        {
            passesCondition = !ModelItemTypeConstants.COMPOSITE_TYPE_ID.equals(typeId) && !ModelItemTypeConstants.REPEATER_TYPE_ID.equals(typeId);
        }
        
        if (passesCondition && !includeSubRepeaters)
        {
            passesCondition = !_isInARepeater(modelItem);
        }
        
        return passesCondition;
    }
    
    /**
     * Checks if the model item passes conditions from accepted types
     * @param acceptedTypes the accepted types
     * @param modelItem the model item to check
     * @return true if the model item passes conditions
     */
    protected boolean _passesAcceptedTypesCondition(List<String> acceptedTypes, ModelItem modelItem)
    {
        boolean passesCondition = false;
        String typeId = modelItem.getType().getId();
        
        Map<String, List<String>> acceptedTypesWithProperties = _getAcceptedTypesWithProperties(acceptedTypes);
        if (acceptedTypesWithProperties.containsKey(typeId))
        {
            List<String> properties = acceptedTypesWithProperties.get(typeId);
            passesCondition = properties.isEmpty() 
                || properties.contains(__ALLOW_ENUMERATED) && modelItem instanceof ElementDefinition && ((ElementDefinition) modelItem).getEnumerator() != null
                || properties.contains(__ALLOW_NOT_ENUMERATED) && (!(modelItem instanceof ElementDefinition) || ((ElementDefinition) modelItem).getEnumerator() == null);
        }
        
        return passesCondition;
    }

    /**
     * Get the map of accepted types with its properties
     * @param acceptedTypes the accepted types configuration
     * @return the map of accepted type with its properties
     */
    protected Map<String, List<String>> _getAcceptedTypesWithProperties(List<String> acceptedTypes)
    {
        Map<String, List<String>> acceptedTypesWithProperties = new HashMap<>();
        for (String acceptedType : acceptedTypes)
        {
            String propertiesAsString = StringUtils.substringBetween(acceptedType, "[", "]");
            if (StringUtils.isNotBlank(propertiesAsString))
            {
                List<String> properties = Arrays.asList(StringUtils.split(propertiesAsString, ","));
                acceptedTypesWithProperties.put(StringUtils.substringBefore(acceptedType, "[").toLowerCase(), properties);
            }
            else
            {
                acceptedTypesWithProperties.put(acceptedType.toLowerCase(), Collections.EMPTY_LIST);
            }
        }
        return acceptedTypesWithProperties;
    }
    
    /**
     * Checks if the given model item has a parent that is a repeater
     * @param modelItem the model item to check
     * @return <code>true</code> if the given model item has a parent that is a repeater, <code>false</code> otherwise
     */
    @SuppressWarnings("static-access")
    protected boolean _isInARepeater(ModelItem modelItem)
    {
        ModelItem parent = modelItem.getParent();
        if (parent != null)
        {
            if (ModelItemTypeConstants.REPEATER_TYPE_ID.equals(parent.getType().getId()))
            {
                return true;
            }
            else
            {
                return _isInARepeater(parent);
            }
        }
        else
        {
            return false;
        }
    }
    
    /**
     * Get the model item's information
     * @param modelItem the model item
     * @param withFullLabel true to use the full label of the model item, false otherwise
     * @return Model item's information as a map of properties
     */
    protected Map<String, Object> getModelItemInfo(ModelItem modelItem, boolean withFullLabel)
    {
        Map<String, Object> modelItemInfo = new HashMap<>();
        
        modelItemInfo.put("name", modelItem.getPath());
        modelItemInfo.put("type", modelItem.getType().getId());
        modelItemInfo.put("label", _i18nUtils.translate(modelItem.getLabel()));
        
        if (withFullLabel)
        {
            modelItemInfo.put("fullLabel", getModelItemFullLabel(modelItem));
        }
        
        return modelItemInfo;
    }
    
    /**
     * Get the user friendly full label for the given model item
     * @param modelItem the model items
     * @return The full label of the model item
     */
    protected String getModelItemFullLabel(ModelItem modelItem)
    {
        StringBuilder sb = new StringBuilder();
        ModelItem parent = modelItem.getParent();
        if (parent != null)
        {
            sb.append(getModelItemFullLabel(parent));
            sb.append(" > ");
        }
        
        sb.append(_i18nUtils.translate(modelItem.getLabel()));
        return sb.toString();
    }
    
    /**
     * Handle specific information, such as adding system properties into the model  list
     * @param commonModelItemsInfo the mapping of model items' properties
     * @param withLastModified true to add the last modification date to the model item's info
     * @param withLastValidation true to add the last validation date to the model item's info
     * @param withCreationDate true to add the creation date to the model item's info
     * @param hasResources true if there is at least one content type that is a resource, false otherwise.
     * @param withFullLabel true to use the full label of the model items, false otherwise
     */
    @SuppressWarnings("static-access")
    protected void handleSpecificInfo(List<Map<String, Object>> commonModelItemsInfo, boolean withLastModified, boolean withLastValidation, boolean withCreationDate, boolean hasResources, boolean withFullLabel)
    {
        if (withLastModified)
        {
            Map<String, Object> metadataInfo = new HashMap<>();
            
            // FIXME web...?
            metadataInfo.put("name", DefaultContent.METADATA_MODIFIED);
            metadataInfo.put("type", ModelItemTypeConstants.DATE_TYPE_ID);
            
            String label = _i18nUtils.translate(new I18nizableText("plugin.cms", "PLUGINS_CMS_LAST_MODIFICATION_DATE"));
            metadataInfo.put("label", label);
            
            if (withFullLabel)
            {
                metadataInfo.put("fullLabel", label);
            }
            
            commonModelItemsInfo.add(metadataInfo);
        }
        
        if (!hasResources)
        {
            if (withLastValidation)
            {
                Map<String, Object> metadataInfo = new HashMap<>();
                
                // FIXME web...?
                metadataInfo.put("name", DefaultContent.METADATA_LAST_VALIDATION);
                metadataInfo.put("type", ModelItemTypeConstants.DATE_TYPE_ID);
                
                String label = _i18nUtils.translate(new I18nizableText("plugin.cms", "PLUGINS_CMS_LAST_VALIDATION_DATE"));
                metadataInfo.put("label", label);
                
                if (withFullLabel)
                {
                    metadataInfo.put("fullLabel", label);
                }
                
                commonModelItemsInfo.add(metadataInfo);
            }
            
            if (withCreationDate)
            {
                Map<String, Object> metadataInfo = new HashMap<>();
                
                // FIXME web...?
                metadataInfo.put("name", DefaultContent.METADATA_CREATION);
                metadataInfo.put("type", ModelItemTypeConstants.DATE_TYPE_ID);
                
                String label = _i18nUtils.translate(new I18nizableText("plugin.cms", "PLUGINS_CMS_CREATION_DATE"));
                metadataInfo.put("label", label);
                
                if (withFullLabel)
                {
                    metadataInfo.put("fullLabel", label);
                }
                
                commonModelItemsInfo.add(metadataInfo);
            }
        }
        
    }
    
    /**
     * Get the content types id to search for
     * @param request the request
     * @param withResources is there at least one resource ?
     * @return the content types
     */
    protected Set<String> getContentTypes(Request request, boolean withResources)
    {
        Set<String> cTypeIds = new HashSet<>();
        String[] ids = request.getParameterValues("ids");
        
        if (ids != null)
        {
            for (String id : ids)
            {
                if (StringUtils.isNotEmpty(id))
                {
                    // id can contains comma (when allOption is selected for example).
                    for (String idPart : StringUtils.split(id, ','))
                    {
                        cTypeIds.add(idPart);
                    }
                }
            }
        }
        
        if (cTypeIds.isEmpty())
        {
            cTypeIds.addAll(getAllAvailablesContentTypes(request, true, withResources));
        }
        
        return cTypeIds;
    }
    
    /**
     * Get all the available content types 
     * @param request the request
     * @param publicOnly Only the non private content types will be returned
     * @param withResources True to add the resources to the list (which is not a content type)
     * @return all the available content types 
     */
    protected Set<String> getAllAvailablesContentTypes (Request request, boolean publicOnly, boolean withResources)
    {
        Set<String> types = new HashSet<>();
        
        for (String id : _contentTypeExtensionPoint.getExtensionsIds())
        {
            if (!publicOnly || !_contentTypeExtensionPoint.getExtension(id).isPrivate())
            {
                types.add(id);
            }
        }
        
        if (withResources)
        {
            types.add("resource");
        }
        
        return types;
    }
}
