/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.commons.lang.BooleanUtils;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.repository.Content;
import org.ametys.core.cocoon.JSonReader;
import org.ametys.plugins.repository.AmetysObjectResolver;

/**
 * Get the common views between given content types and/or among given contents
 */
public class GetCommonViewsAction extends ServiceableAction
{
    /** The content type EP */
    protected ContentTypeExtensionPoint _contentTypeEP;
    
    /** The content types Helper */
    protected ContentTypesHelper _contentTypesHelper;
    
    /** The ametys object resolver */
    protected AmetysObjectResolver _resolver;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _contentTypeEP = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
        _contentTypesHelper = (ContentTypesHelper) smanager.lookup(ContentTypesHelper.ROLE);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        boolean includeInternals = BooleanUtils.toBoolean(request.getParameter("includeInternals")); // false by default
        
        Set<String> requestedContentTypeIds = getContentTypes(request);
        Map<String, Set<String>> contentTypesByContent = getContentTypesFromContents(request);
        if (requestedContentTypeIds.isEmpty() && contentTypesByContent.keySet().isEmpty())
        {
            requestedContentTypeIds = getAllAvailablesContentTypes(request, true);
        }
        
        Map<String, Object> result = new HashMap<>();
        
        Collection<Map<String, Object>> commonViewsInfo = getCommonViewsInfo(contentTypesByContent, requestedContentTypeIds, includeInternals);
        result.put("views", commonViewsInfo);
        
        request.setAttribute(JSonReader.OBJECT_TO_READ, result);
        return EMPTY_MAP;
    }
    
    /**
     * Retrieves the common views info by name
     * @param contentTypesByContent the content type ids by content
     * @param contentTypeIds the content type ids (not coming from contents)
     * @param includeInternals <code>true</code> to include internal views, <code>false</code> otherwise
     * @return The map of views' info
     */
    protected Collection<Map<String, Object>> getCommonViewsInfo(Map<String, Set<String>> contentTypesByContent, Collection<String> contentTypeIds, boolean includeInternals)
    {
        Map<String, Map<String, Object>> commonViewsInfo = null;
        
        for (Set<String> contentTypeIdsForCurrentContent : contentTypesByContent.values())
        {
            List<Map<String, Object>> viewsInfoForCurrentContent = _viewsUnion(contentTypeIdsForCurrentContent, includeInternals);
            commonViewsInfo = _viewsIntersection(commonViewsInfo, viewsInfoForCurrentContent);
        }
        
        for (String contentTypeId : contentTypeIds)
        {
            List<Map<String, Object>> viewsInfoForCurrentContentType = _getViewsInfo(contentTypeId, includeInternals);
            commonViewsInfo = _viewsIntersection(commonViewsInfo, viewsInfoForCurrentContentType);
        }
        
        return commonViewsInfo != null ? commonViewsInfo.values() : Collections.EMPTY_LIST;
    }
    
    private List<Map<String, Object>> _viewsUnion(Set<String> contentTypes, boolean includeInternals)
    {
        List<Map<String, Object>> viewsInfo = new ArrayList<>();
        for (String contentType : contentTypes)
        {
            viewsInfo.addAll(_getViewsInfo(contentType, includeInternals));
        }
        return viewsInfo;
    }
    
    /**
     * Get the available views for a given content type
     * @param contentTypeId the id of content type
     * @param includeInternals true to include internal view
     * @return the available views information
     */
    protected List<Map<String, Object>> _getViewsInfo(String contentTypeId, boolean includeInternals)
    {
        return _contentTypesHelper.getViewsInfo(contentTypeId, includeInternals);
    }
    
    private Map<String, Map<String, Object>> _viewsIntersection(Map<String, Map<String, Object>> commonViewsInfo, List<Map<String, Object>> viewsInfo)
    {
        Map<String, Map<String, Object>> result = commonViewsInfo;
        if (result == null)
        {
            result = new HashMap<>();
            for (Map<String, Object> entry : viewsInfo)
            {
                String viewName = (String) entry.get("name");
                result.put(viewName, entry);
            }
        }
        else
        {
            Set<String> viewNames = new HashSet<>();
            for (Map<String, Object> entry : viewsInfo)
            {
                String viewName = (String) entry.get("name");
                viewNames.add(viewName);
            }
            
            // only retains common metadata (performs a set intersection)
            result.keySet().retainAll(viewNames);
        }
        
        return result;
    }
    
    /**
     * Get the content types id to search for
     * @param request the request
     * @return the content types
     */
    protected Set<String> getContentTypes(Request request)
    {
        Set<String> contentTypeIds = new HashSet<>();
        String[] ids = request.getParameterValues("ids");
        
        if (ids != null)
        {
            for (String id : ids)
            {
                if (StringUtils.isNotEmpty(id))
                {
                    // id can contains comma (when allOption is selected for example).
                    for (String idPart : StringUtils.split(id, ','))
                    {
                        contentTypeIds.add(idPart);
                    }
                }
            }
        }
        
        return contentTypeIds;
    }
    
    /**
     * Get the content types id to search for (by content id)
     * @param request the request
     * @return the content types by content
     */
    protected Map<String, Set<String>> getContentTypesFromContents(Request request)
    {
        Map<String, Set<String>> result = new HashMap<>();
        
        String[] contentIds = request.getParameterValues("contentIds");
        
        if (contentIds != null)
        {
            for (String contentId : contentIds)
            {
                if (StringUtils.isNotEmpty(contentId))
                {
                    Content content = _resolver.resolveById(contentId);
                    
                    Set<String> allContentTypes = Stream.concat(Stream.of(content.getTypes()), Stream.of(content.getMixinTypes()))
                            .collect(Collectors.toSet());
                    result.put(contentId, allContentTypes);
                }
            }
        }
        
        return result;
    }
    
    /**
     * Get all the available content types 
     * @param request the request
     * @param publicOnly Only the non private content types will be returned
     * @return all the available content types 
     */
    protected Set<String> getAllAvailablesContentTypes (Request request, boolean publicOnly)
    {
        Set<String> types = new HashSet<>();
        
        for (String id : _contentTypeEP.getExtensionsIds())
        {
            if (!publicOnly || !_contentTypeEP.getExtension(id).isPrivate())
            {
                types.add(id);
            }
        }
        
        return types;
    }
}
