/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;

import org.ametys.cms.content.RootContentHelper;
import org.ametys.core.cocoon.JSonReader;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;

/**
 * Get information on content types.<br>
 * If "inherited" request parameter is set to true, the sub-types will be also returned.<br>
 * If "checkRights" request parameter is set to true, only content types allowed for creation will be returned<br>
 */
public class GetContentTypesListAction extends ServiceableAction
{
    /** Content type extension point. */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    /** The current user provider */
    protected CurrentUserProvider _userProvider;
    /** The rights manager */
    protected RightManager _rightManager;
    /** Helper for root content */
    protected RootContentHelper _rootContentHelper;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
        _userProvider = (CurrentUserProvider) smanager.lookup(CurrentUserProvider.ROLE);
        _rightManager = (RightManager) smanager.lookup(RightManager.ROLE);
        _rootContentHelper = (RootContentHelper) smanager.lookup(RootContentHelper.ROLE);
    }
    
    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Map<String, Object> result = new HashMap<>();
        
        Request request = ObjectModelHelper.getRequest(objectModel);
        List<String> contentTypesIds = Arrays.asList(request.getParameterValues("ids"));
        
        boolean inherited = request.getParameter("inherited") != null && "true".equalsIgnoreCase(request.getParameter("inherited"));
        boolean checkRights = request.getParameter("checkRights") != null && "true".equalsIgnoreCase(request.getParameter("checkRights"));
        boolean includePrivate = request.getParameter("includePrivate") != null && "true".equalsIgnoreCase(request.getParameter("includePrivate"));
        
        List<Map<String, Object>> contentTypes = new ArrayList<>();
        List<String> unknownContentTypes = new ArrayList<>();
        List<String> noRightContentTypes = new ArrayList<>();
        List<String> privateContentTypes = new ArrayList<>();
        
        // Collect content types.
        Collection<String> allContentTypesIds = new ArrayList<>();
        for (String id : contentTypesIds)
        {
            _addIfNotPresent(allContentTypesIds, id);
            
            if (inherited)
            {
                for (String subTypeId : _contentTypeExtensionPoint.getSubTypes(id))
                {
                    _addIfNotPresent(allContentTypesIds, subTypeId);
                }
            }
        }
        
        // Resolve and organize content types
        for (String id : allContentTypesIds)
        {
            ContentType cType = _contentTypeExtensionPoint.getExtension(id);
            
            if (cType != null && !cType.isAbstract())
            {
                if (cType.isPrivate() && !includePrivate)
                {
                    privateContentTypes.add(id);
                }
                else if (checkRights && _hasRight(cType)
                         || !checkRights)
                {
                    contentTypes.add(getContentTypeProperties(cType));
                }
                else
                {
                    noRightContentTypes.add(id);
                }
            }
            else
            {
                unknownContentTypes.add(id);
            }
        }
        
        result.put("contentTypes", contentTypes);
        result.put("noRightContentTypes", noRightContentTypes);
        result.put("unknownContentTypes", unknownContentTypes);
        result.put("privateContentTypes", noRightContentTypes);
        
        request.setAttribute(JSonReader.OBJECT_TO_READ, result);
        
        return EMPTY_MAP;
    }
    
    /**
     * Get the content type properties
     * @param contentType The content type
     * @return The content type properties
     */
    protected Map<String, Object> getContentTypeProperties (ContentType contentType)
    {
        Map<String, Object> infos = new HashMap<>();
        
        infos.put("id", contentType.getId());
        infos.put("label", contentType.getLabel());
        infos.put("description", contentType.getDescription());
        infos.put("defaultTitle", contentType.getDefaultTitle());
        infos.put("iconGlyph", contentType.getIconGlyph());
        infos.put("iconDecorator", contentType.getIconDecorator());
        infos.put("iconSmall", contentType.getSmallIcon());
        infos.put("iconMedium", contentType.getMediumIcon());
        infos.put("iconLarge", contentType.getLargeIcon());
        
        return infos;
    }
    
    /**
     * Test if the current user has the right needed by the content type to create a content.
     * @param contentType The content type
     * @return true if the user has the right needed, false otherwise.
     */
    protected boolean _hasRight(ContentType contentType)
    {
        boolean hasRight = false;
        
        String right = contentType.getRight();
        
        if (right == null)
        {
            hasRight = true;
        }
        else
        {
            UserIdentity user = _userProvider.getUser();
            hasRight = _rightManager.hasRight(user, right, "/cms") == RightResult.RIGHT_ALLOW || _rightManager.hasRight(user, right, _rootContentHelper.getRootContent()) == RightResult.RIGHT_ALLOW;
        }
        
        return hasRight;
    }
    
    private void _addIfNotPresent(Collection<String> collection, String value)
    {
        if (!collection.contains(value))
        {
            collection.add(value);
        }
    }
}
