/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.repository.Content;
import org.ametys.core.cocoon.JSonReader;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.runtime.model.DefinitionContext;
import org.ametys.runtime.model.View;

/**
 * Get information about the definition of the structure a {@link View}.<br>
 */
public class GetViewDefinitionAction extends ServiceableAction
{
    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /** Content type extension point. */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    /** Helper for content types */
    protected ContentTypesHelper _contentTypesHelper;
    /** Helper for contents */
    protected ContentHelper _contentHelper;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
        _contentTypesHelper = (ContentTypesHelper) smanager.lookup(ContentTypesHelper.ROLE);
        _contentHelper = (ContentHelper) smanager.lookup(ContentHelper.ROLE);
    }
    
    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        String contentTypeId = StringUtils.defaultString(request.getParameter("contentType"));
        String contentId = StringUtils.defaultString(request.getParameter("contentId"));
        String viewName = StringUtils.defaultIfEmpty(request.getParameter("viewName"), "default-edition");
        String fallbackViewName = StringUtils.defaultIfEmpty(request.getParameter("fallbackViewName"), "main");
        boolean isEdition = StringUtils.defaultIfEmpty(request.getParameter("viewMode"), "edition").equals("edition");
        
        String[] contentTypes;
        String[] mixins = new String[0];
        
        Content content = null;
        if (StringUtils.isNotEmpty(contentTypeId))
        {
            contentTypes = contentTypeId.split(",");
        }
        else if (StringUtils.isNotEmpty(contentId))
        {
            content = _resolver.resolveById(contentId);
            contentTypes = content.getTypes();
            mixins = content.getMixinTypes();
        }
        else
        {
            String errorMsg = String.format("Unable to get view definition: missing parameter 'contentId' or 'contentTypeId'");
            getLogger().error(errorMsg);
            throw new IllegalArgumentException(errorMsg);
        }
        
        for (String id : contentTypes)
        {
            ContentType cType = _contentTypeExtensionPoint.getExtension(id);
            
            if (cType == null)
            {
                String errorMsg = String.format("Unable to get view definition: unknown content type '%s'", id);
                getLogger().error(errorMsg);
                throw new IllegalArgumentException(errorMsg);
            }
        }

        Object objectToRead;
        View view = _contentTypesHelper.getViewWithFallback(viewName, fallbackViewName, contentTypes, mixins);
        
        // If view is unknown
        if (view == null)
        {
            objectToRead = error2JsonObject(contentTypes, viewName);
        }
        // If the content id is not blank
        else if (StringUtils.isNotBlank(contentId))
        {
            objectToRead = _contentHelper.getContentViewAsJSON(contentId, viewName, fallbackViewName, isEdition).get("view");
        }
        // Other cases based on content types
        else
        {
            objectToRead = view.toJSON(DefinitionContext.newInstance().withEdition(isEdition));
        }
        
        request.setAttribute(JSonReader.OBJECT_TO_READ, objectToRead);
        
        return EMPTY_MAP;
    }
    
    /**
     * Get some error properties when the requested view is not found. 
     * @param contentTypes The content types 
     * @param viewName The view name
     * @return A map containing the properties.
     */
    protected Map<String, Object> error2JsonObject(String[] contentTypes, String viewName)
    {
        getLogger().error(String.format("Unknown view '%s' for content types '%s'", viewName, StringUtils.join(contentTypes, ',')));
        
        Map<String, Object> jsonObject = new HashMap<>();
        
        jsonObject.put("type", "error");
        jsonObject.put("contentType", StringUtils.join(contentTypes, ','));
        jsonObject.put("viewName", viewName);
        
        return jsonObject;
    }
}
