/*
 *  Copyright 2011 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.ext.LexicalHandler;

import org.ametys.cms.repository.Content;
import org.ametys.plugins.repository.TraversableAmetysObject;

/**
 * This class represent a handler from HTML RichText
 */
public class HTMLUpdateHandler extends AbstractLogEnabled implements ContentHandler, LexicalHandler
{
    /** The lexical handler */
    protected LexicalHandler _lexicalHandler;
    /** The content handler where to write */
    protected ContentHandler _contentHandler;
    
    /** The original content */
    protected Content _initialContent;
    /** The created content after copy */
    protected Content _createdContent;
    
    /** The path of the original content */
    protected String _initialContentPath;
    
    /** The initial object which implied the content copy */
    protected TraversableAmetysObject _initialObject;
    /** The created object which implied the content copy */
    protected TraversableAmetysObject _createdObject;
    /** The path of the initial object */
    protected String _initialAOPath;
    
    /**
     * Set the target content handler
     * @param contentHandler the target content handler
     */
    public void setContentHandler(ContentHandler contentHandler)
    {
        _contentHandler = contentHandler;
    }
    
    /**
     * Set the target lexical handler
     * @param lexicalHandler the target lexical handler
     */
    public void setLexicalHandler(LexicalHandler lexicalHandler)
    {
        _lexicalHandler = lexicalHandler;
    }
    
    /**
     * Set the initial content copied
     * @param initialContent the initial content
     */
    public void setInitialContent (Content initialContent)
    {
        _initialContent = initialContent;
        _initialContentPath = initialContent.getPath();
    }
    
    /**
     * Set the created content
     * @param createdContent the created content after copy
     */
    public void setCreatedContent (Content createdContent)
    {
        _createdContent = createdContent;
    }
    
    /**
     * Set the initial object responsible for copy
     * @param initialObject the initial object responsible for copy
     */
    public void setInitialObject (TraversableAmetysObject initialObject)
    {
        _initialObject = initialObject;
        _initialAOPath = initialObject.getPath();
    }
    
    /**
     * Set created object after copy
     * @param createdObject the created object after copy
     */
    public void setCreatedObject (TraversableAmetysObject createdObject)
    {
        _createdObject = createdObject;
    }
    
    @Override
    public void startDocument() throws SAXException
    {
        _contentHandler.startDocument();
    }
    @Override
    public void endDocument() throws SAXException
    {
        _contentHandler.endDocument();
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        _contentHandler.startElement(uri, localName, qName, atts);
    }
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        _contentHandler.endElement(uri, localName, qName);
    }
    
    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        _contentHandler.characters(ch, start, length);
    }
    @Override
    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
    {
        _contentHandler.ignorableWhitespace(ch, start, length);
    }
    
    @Override
    public void processingInstruction(String target, String data) throws SAXException
    {
        _contentHandler.processingInstruction(target, data);
    }
    @Override
    public void setDocumentLocator(Locator locator)
    {
        _contentHandler.setDocumentLocator(locator);
    }
    @Override
    public void skippedEntity(String name) throws SAXException
    {
        _contentHandler.skippedEntity(name);
    }
    
    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {
        _contentHandler.startPrefixMapping(prefix, uri);
    }
    @Override
    public void endPrefixMapping(String prefix) throws SAXException
    {
        _contentHandler.endPrefixMapping(prefix);
    }
    
    @Override
    public void startEntity(String name) throws SAXException
    {
        _lexicalHandler.startEntity(name);
    }
    @Override
    public void endEntity(String name) throws SAXException
    {
        _lexicalHandler.endEntity(name);
    }

    @Override
    public void startCDATA() throws SAXException
    {
        _lexicalHandler.startCDATA();
    }
    @Override
    public void endCDATA() throws SAXException
    {
        _lexicalHandler.endCDATA();
    }

    @Override
    public void startDTD(String name, String publicId, String systemId) throws SAXException
    {
        _lexicalHandler.startDTD(name, publicId, systemId);
    }
    @Override
    public void endDTD() throws SAXException
    {
        _lexicalHandler.endCDATA();
    }

    @Override
    public void comment(char[] ch, int start, int length) throws SAXException
    {
        _lexicalHandler.comment(ch, start, length);
    }
}
