/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.xml.AttributesImpl;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.DefaultContentType.AnnotableDefinition;
import org.ametys.cms.data.RichText;
import org.ametys.cms.repository.Content;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.util.XMLUtils;
import org.ametys.runtime.model.DefinitionContext;

/**
 * RichText definition with its list of allowed semantic annotations
 */
public class RichTextAttributeDefinition extends AttributeDefinition<RichText> implements AnnotableDefinition
{
    /** Rights manager */
    private RightManager _rightManager;
    /** The current user provider */
    private CurrentUserProvider _currentUserProvider;
    
    private List<SemanticAnnotation> _semanticAnnotations;

    public List<SemanticAnnotation> getSemanticAnnotations()
    {
        return _semanticAnnotations;
    }

    public void setSemanticAnnotations(List<SemanticAnnotation> semAnnotations)
    {
        this._semanticAnnotations = semAnnotations;
    }
    
    @Override
    protected Map<String, Object> _toJSON(DefinitionContext context) throws ProcessingException
    {
        Map<String, Object> result = super._toJSON(context);
        
        if (context.getObject().isPresent())
        {
            Object object = context.getObject().get();
            if (object instanceof Content)
            {
                Content content = (Content) object;
                result.put("editableSource", _getRightManager().hasRight(_getCurrentUserProvider().getUser(), "CORE_Rights_SourceEdit", content) == RightResult.RIGHT_ALLOW);
            }
        }
        
        if (_semanticAnnotations != null && !_semanticAnnotations.isEmpty())
        {
            List<Map<String, Object>> annotationsObject = new ArrayList<>();

            for (SemanticAnnotation annotation : _semanticAnnotations)
            {
                Map<String, Object> annotationObject = new LinkedHashMap<>();

                annotationObject.put("name", annotation.getId());
                annotationObject.put("label", annotation.getLabel());
                annotationObject.put("description", annotation.getDescription());

                annotationsObject.add(annotationObject);
            }

            result.put("annotations", annotationsObject);
        }
        
        return result;
    }
    
    @SuppressWarnings("static-access")
    @Override
    public void toSAX(ContentHandler contentHandler, DefinitionContext context) throws SAXException
    {
        super.toSAX(contentHandler, context);
        
        if (context.getObject().isPresent())
        {
            Object object = context.getObject().get();
            if (object instanceof Content)
            {
                Content content = (Content) object;
                boolean isEditableSource = _getRightManager().hasRight(_getCurrentUserProvider().getUser(), "CORE_Rights_SourceEdit", content) == RightResult.RIGHT_ALLOW;
                XMLUtils.createElement(contentHandler, "editableSource", String.valueOf(isEditableSource));
            }
        }
        
        if (_semanticAnnotations != null && !_semanticAnnotations.isEmpty())
        {
            XMLUtils.startElement(contentHandler, "annotations");

            for (SemanticAnnotation annotation : _semanticAnnotations)
            {
                AttributesImpl attributes = new AttributesImpl();
                attributes.addCDATAAttribute("name", annotation.getId());
                XMLUtils.startElement(contentHandler, "annotation", attributes);
                
                XMLUtils.createI18nElementIfNotNull(contentHandler, "label", annotation.getLabel());
                XMLUtils.createI18nElementIfNotNull(contentHandler, "description", annotation.getDescription());

                XMLUtils.endElement(contentHandler, "annotation");
            }

            XMLUtils.endElement(contentHandler, "annotations");
        }
    }
    
    /**
     * Retrieves the {@link RightManager} 
     * @return the {@link RightManager}
     */
    protected RightManager _getRightManager()
    {
        if (_rightManager == null)
        {
            try
            {
                _rightManager = (RightManager) __serviceManager.lookup(RightManager.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException("Unable to lookup after the right manager", e);
            }
        }
        
        return _rightManager;
    }
    
    /**
     * Retrieves the {@link CurrentUserProvider} 
     * @return the {@link CurrentUserProvider}
     */
    protected CurrentUserProvider _getCurrentUserProvider()
    {
        if (_currentUserProvider == null)
        {
            try
            {
                _currentUserProvider = (CurrentUserProvider) __serviceManager.lookup(CurrentUserProvider.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException("Unable to lookup after the current user provider", e);
            }
        }
        
        return _currentUserProvider;
    }
}
