/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype.validation;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.contenttype.AttributeDefinition;
import org.ametys.cms.contenttype.ContentValidator;
import org.ametys.cms.repository.Content;
import org.ametys.plugins.repository.data.holder.impl.DataHolderHelper;
import org.ametys.plugins.repository.data.holder.values.SynchronizableRepeater;
import org.ametys.plugins.repository.data.holder.values.SynchronizationContext;
import org.ametys.plugins.repository.model.CompositeDefinition;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemGroup;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.type.ElementType;
import org.ametys.runtime.parameter.Errors;

/**
 * Base {@link ContentValidator} validating the content by comparing one or more couples of values.
 * For each couple of values, the identified max value should be greater than or equal to the identified min value.
 * 
 * @param <T> The type tested by the validator.
 */
public abstract class AbstractIntervalValidator<T> extends AbstractContentValidator implements Configurable
{
    private String _minDefinitionPath;
    private String _maxDefinitionPath;
    private boolean _mandatory;
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        _minDefinitionPath = configureMinPath(configuration);
        _maxDefinitionPath = configureMaxPath(configuration);
        _mandatory = configuration.getChild("mandatory", true).getValueAsBoolean(false);
    }
    
    @Override
    public void validate(Content content, Errors errors)
    {
        ElementDefinition minDefinition = (ElementDefinition) content.getDefinition(_minDefinitionPath);
        ElementDefinition maxDefinition = (ElementDefinition) content.getDefinition(_maxDefinitionPath);
        
        _checkAttributeDefinitions(minDefinition, maxDefinition);
        
        Object minValues = content.getValue(_minDefinitionPath, true);
        Object maxValues = content.getValue(_maxDefinitionPath, true);
        
        _validateInterval(content, errors, minDefinition, maxDefinition, minValues, maxValues);
    }
    
    @SuppressWarnings("unchecked")
    private List<T> _getValuesAsList(Object value)
    {
        if (value == null)
        {
            List<T> result = new ArrayList<>();
            result.add(null);
            return result;
        }
        
        return value instanceof List ? (List<T>) value : value.getClass().isArray() ? Arrays.asList((T[]) value) : List.of((T) value);
    }
    
    /**
     * Validate the interval values of a content
     * @param content The content
     * @param errors The list of errors to fill
     * @param minDefinition The definition for the min value
     * @param maxDefinition The definition for the max value
     * @param minValues The list of min values
     * @param maxValues The list of max values
     */
    protected void _validateInterval(Content content, Errors errors, ElementDefinition minDefinition, ElementDefinition maxDefinition, Object minValues, Object maxValues)
    {
        List<T> listMinValues = _getValuesAsList(minValues);
        List<T> listMaxValues = _getValuesAsList(maxValues);
                
        if (listMinValues.size() != listMaxValues.size())
        {
            if (getLogger().isDebugEnabled())
            {
                getLogger().debug("Could not process with the content validation: did not found a min/max couples for each values.");
            }
            
            errors.addError(new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_VALIDATOR_NUMBER_INTERVAL_SIZEERROR"));
            return;
        }
        
        for (int i = 0; i < listMinValues.size(); i++)
        {
            T min = listMinValues.get(i);
            T max = listMaxValues.get(i);
            
            if (_mandatory && (min != null ^ max != null))
            {
                I18nizableText emptyLabel = min == null ? minDefinition.getLabel() : maxDefinition.getLabel();
                if (getLogger().isDebugEnabled())
                {
                    getLogger().debug("Content '" + content.getId() + "' is invalid : " + emptyLabel + " is empty but is part of a mandatory interval");
                }
                
                addErrorEmpty(errors, minDefinition, maxDefinition, emptyLabel);
            }
            else if (min != null && max != null && isLessThan(max, min))
            {
                if (getLogger().isDebugEnabled())
                {
                    getLogger().debug("Content '" + content.getId() + "' is invalid : " + min + " [" + _minDefinitionPath + "] is greater than " + max + " [" + _maxDefinitionPath + "]");
                }
                
                addIntervalError(errors, minDefinition, maxDefinition, min, max);
            }
        }
    }

    @Override
    public void validate(Content content, Map<String, Object> values, View view, Errors errors)
    {
        AttributeDefinition minDefinition = (AttributeDefinition) content.getDefinition(_minDefinitionPath);
        AttributeDefinition maxDefinition = (AttributeDefinition) content.getDefinition(_maxDefinitionPath);
        
        _checkAttributeDefinitions(minDefinition, maxDefinition);
        
        Object minValues;
        Object maxValues;
        
        if (minDefinition.canWrite(content))
        {
            minValues = _getValues(content, values, _minDefinitionPath, Optional.of(StringUtils.EMPTY), null);
        }
        else
        {
            minValues = content.getValue(_minDefinitionPath, true);
        }

        if (maxDefinition.canWrite(content))
        {
            maxValues = _getValues(content, values, _maxDefinitionPath, Optional.of(StringUtils.EMPTY), null);
        }
        else
        {
            maxValues = content.getValue(_maxDefinitionPath , true);
        }

        _validateInterval(content, errors, minDefinition, maxDefinition, minValues, maxValues);
    }
    
    private void _checkAttributeDefinitions (ElementDefinition minDefinition, ElementDefinition maxDefinition)
    {
        if (minDefinition == null)
        {
            throw new IllegalArgumentException("Unable to validate content: the path '" + _minDefinitionPath + "' refers to a non existing attribute for content type " + getContentType().getId());
        }
        if (maxDefinition == null)
        {
            throw new IllegalArgumentException("Unable to validate content: the path '" + _maxDefinitionPath + "' refers to a non existing attribute for content type " + getContentType().getId());
        }
        
        if (!isSupportedType(minDefinition.getType()))
        {
            throw new IllegalArgumentException("Unable to validate content: the attribute to path '" + _minDefinitionPath + "' (" + minDefinition.getType() + ") is not the expected type");
        }
        
        if (!isSupportedType(maxDefinition.getType()))
        {
            throw new IllegalArgumentException("Unable to validate content: the attribute to path '" + _maxDefinitionPath + "' (" + maxDefinition.getType() + ") is not the expected type");
        }
    }

    @SuppressWarnings("unchecked")
    private List<T> _getValues(Content content, Map<String, Object> values, String definitionPath, Optional<String> oldPrefix, ModelItemGroup parentDefinition)
    {
        String[] pathSegments = StringUtils.split(definitionPath, ModelItem.ITEM_PATH_SEPARATOR);
        
        if (pathSegments.length == 0)
        {
            return null;
        }
        
        String fieldName = pathSegments[0];

        ModelItem modelItem = parentDefinition == null ? content.getDefinition(fieldName) : parentDefinition.getChild(fieldName);
        
        Object valueFromMap = values.get(fieldName);
        Optional<String> oldDataPath = oldPrefix.map(p -> p + fieldName);
        Object value = DataHolderHelper.getValueFromSynchronizableValue(valueFromMap, content, modelItem, oldDataPath, SynchronizationContext.newInstance());
        if (value == null)
        {
            return null;
        }
            
        if (pathSegments.length == 1)
        {
            if (!(modelItem instanceof ElementDefinition))
            {
                throw new IllegalArgumentException("Last path segment should correspond to an ElementType");
            }
            
            ElementDefinition definition = (ElementDefinition) modelItem;
            
            if (isSupportedType(definition.getType()))
            {
                return value.getClass().isArray() ? Arrays.asList((T[]) value) : List.of((T) value);
            }
            else
            {
                // Should never happen (this was checked before)
                throw new IllegalArgumentException("Attribute '" + fieldName + "' (" + definition.getType() + ") is not the expected type");
            }
        }
        else
        {
            String remainingPath = StringUtils.join(pathSegments, ModelItem.ITEM_PATH_SEPARATOR, 1, pathSegments.length);
            
            if (modelItem instanceof CompositeDefinition)
            {
                Optional<String> updatedOldPrefix = oldPrefix.map(p -> p + fieldName + ModelItem.ITEM_PATH_SEPARATOR);
                return _getValues(content, (Map<String, Object>) value, remainingPath, updatedOldPrefix, (CompositeDefinition) modelItem);
            }
            else if (modelItem instanceof RepeaterDefinition)
            {
                List<Map<String, Object>> entries = value instanceof List ? (List<Map<String, Object>>) value : ((SynchronizableRepeater) value).getEntries();
                
                List<T> result = new ArrayList<>();
                
                for (int i = 0; i < entries.size(); i++)
                {
                    Optional<String> updatedOldPrefix = _getRepeaterUpdatedOldPrefix(fieldName, oldPrefix, value, i);
                    List<T> entryValues = _getValues(content, entries.get(i), remainingPath, updatedOldPrefix, (RepeaterDefinition) modelItem);

                    if (entryValues != null)
                    {
                        result.addAll(entryValues);
                    }
                }
                
                return result;
            }
            else
            {
                throw new IllegalArgumentException("Inner path segments should correspond to an ModelItemGroup");
            }
        }
    }

    private Optional<String> _getRepeaterUpdatedOldPrefix(String fieldName, Optional<String> oldPrefix, Object value, int i)
    {
        if (value instanceof SynchronizableRepeater)
        {
            SynchronizableRepeater syncRepeater = (SynchronizableRepeater) value;
            Optional<Integer> previousPosition = syncRepeater.getPreviousPosition(i);
            if (previousPosition.isPresent())
            {
                return oldPrefix.map(p -> p + fieldName + "[" + previousPosition.get() + "]" + ModelItem.ITEM_PATH_SEPARATOR);
            }
        }
        
        return Optional.empty();
    }
    
    /**
     * Get the metadata min path from the configuration
     * @param configuration The configuration
     * @return The metadata min path
     * @throws ConfigurationException If an error occurs
     */
    protected abstract String configureMinPath(Configuration configuration) throws ConfigurationException;
    
    /**
     * Get the metadata max path from the configuration
     * @param configuration The configuration
     * @return The metadata max path
     * @throws ConfigurationException If an error occurs
     */
    protected abstract String configureMaxPath(Configuration configuration) throws ConfigurationException;

    /**
     * Test if the first number is less than the second one.
     * @param n1 The first number to compare.
     * @param n2 The second number to compare.
     * @return true if the first number is less than the first, false otherwise.
     */
    protected abstract boolean isLessThan(T n1, T n2);
    
    /**
     * Test if the attribute type is one expected by the validator
     * @param type The attribute type
     * @return True if the type is supported
     */
    protected abstract boolean isSupportedType(ElementType type);
    

    /**
     * Add an error when the max value is less than the min value
     * @param errors The list of errors
     * @param minDefinition The min definition
     * @param maxDefinition The max definition
     * @param min The min value
     * @param max The max value
     */
    protected abstract void addIntervalError(Errors errors, ElementDefinition minDefinition, ElementDefinition maxDefinition, T min, T max);

    /**
     * Add an error when the max or the min value is empty but the interval is mandatory
     * @param errors The list of errors
     * @param minDefinition The min definition
     * @param maxDefinition The max definition
     * @param emptyLabel The label of the empty field
     */
    protected abstract void addErrorEmpty(Errors errors, ElementDefinition minDefinition, ElementDefinition maxDefinition, I18nizableText emptyLabel);
}
