/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype.validation;

import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;

import org.ametys.cms.contenttype.ContentValidator;
import org.ametys.runtime.i18n.I18nizableTextParameter;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.type.ElementType;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.parameter.Errors;

/**
 * Base {@link ContentValidator} validating the content by comparing one or more couples of numerical values.
 * For each couple of values, the identified max value should be greater than or equal to the identified min value.
 * 
 * The max and min values to compare are identified in validator's configuration by the metadata path as follows:
 * &lt;min path="path/to/attribute/min"/&gt;
 * &lt;max path="path/to/attribute/max"/&gt;
 * 
 * The path of attribute can contain one or more repeaters.
 * 
 * @param <T> The numeric type tested by the validator.
 */
public abstract class AbstractNumberIntervalValidator<T extends Number> extends AbstractIntervalValidator<T>
{
    @Override
    protected String configureMinPath(Configuration configuration) throws ConfigurationException
    {
        return configuration.getChild("min").getAttribute("path");
    }
    
    @Override
    protected String configureMaxPath(Configuration configuration) throws ConfigurationException
    {
        return configuration.getChild("max").getAttribute("path");
    }
    
    @Override
    protected boolean isSupportedType(ElementType type)
    {
        return type.getId().equals(ModelItemTypeConstants.DOUBLE_TYPE_ID) || type.getId().equals(ModelItemTypeConstants.LONG_TYPE_ID);
    }
    
    @Override
    protected void addIntervalError(Errors errors, ElementDefinition minDefinition, ElementDefinition maxDefinition, T min, T max)
    {
        Map<String, I18nizableTextParameter> i18nParams = new HashMap<>();
        i18nParams.put("minValue", new I18nizableText(min.toString()));
        i18nParams.put("maxValue", new I18nizableText(max.toString()));
        i18nParams.put("minLabel", minDefinition.getLabel());
        i18nParams.put("maxLabel", maxDefinition.getLabel());
        errors.addError(new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_VALIDATOR_NUMBER_INTERVAL_ERROR", i18nParams));
    }

    @Override
    protected void addErrorEmpty(Errors errors, ElementDefinition minDefinition, ElementDefinition maxDefinition, I18nizableText emptyLabel)
    {
        Map<String, I18nizableTextParameter> i18nParams = new HashMap<>();
        i18nParams.put("emptyLabel", emptyLabel);
        i18nParams.put("minLabel", minDefinition.getLabel());
        i18nParams.put("maxLabel", maxDefinition.getLabel());
        errors.addError(new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_VALIDATOR_NUMBER_INTERVAL_NULL_ERROR", i18nParams));
    }
}
