/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype.validation;

import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.FormatStyle;
import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;

import org.ametys.cms.contenttype.ContentValidator;
import org.ametys.runtime.i18n.I18nizableDateTime;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.i18n.I18nizableTextParameter;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.type.ElementType;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.parameter.Errors;

/**
 * This implementation of {@link ContentValidator} validates the content by comparing one or more couple of date values.
 * For each couple of values, the identified end date should be greater than or equal to the identified start date.
 * 
 * The start and end dates to compare are identified in validator's configuration by the attribute path as follows:
 * &lt;startDate path="path/to/attribute/start"/&gt;
 * &lt;endDate path="path/to/attribute/end"/&gt;
 * 
 * If the path of metadata can contain one or more repeaters.
 */
public class DateTimeIntervalValidator extends AbstractIntervalValidator<ZonedDateTime>
{
    @Override
    protected String configureMinPath(Configuration configuration) throws ConfigurationException
    {
        return configuration.getChild("startDate").getAttribute("path");
    }
    
    @Override
    protected String configureMaxPath(Configuration configuration) throws ConfigurationException
    {
        return configuration.getChild("endDate").getAttribute("path");
    }

    @Override
    protected boolean isLessThan(ZonedDateTime date1, ZonedDateTime date2)
    {
        return date1.isBefore(date2);
    }
    
    @Override
    protected boolean isSupportedType(ElementType type)
    {
        return type.getId().equals(ModelItemTypeConstants.DATETIME_TYPE_ID);
    }
    
    @Override
    protected void addIntervalError(Errors errors, ElementDefinition minDefinition, ElementDefinition maxDefinition, ZonedDateTime startDate, ZonedDateTime endDate)
    {
        Map<String, I18nizableTextParameter> i18nParams = new HashMap<>();
        i18nParams.put("startDate", new I18nizableDateTime(startDate, ZoneId.systemDefault(), FormatStyle.SHORT));
        i18nParams.put("endDate", new I18nizableDateTime(endDate, ZoneId.systemDefault(), FormatStyle.SHORT));
        i18nParams.put("startLabel", minDefinition.getLabel());
        i18nParams.put("endLabel", maxDefinition.getLabel());
        errors.addError(new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_VALIDATOR_DATE_INTERVAL_ERROR", i18nParams));
    }

    @Override
    protected void addErrorEmpty(Errors errors, ElementDefinition minDefinition, ElementDefinition maxDefinition, I18nizableText emptyLabel)
    {
        Map<String, I18nizableTextParameter> i18nParams = new HashMap<>();
        i18nParams.put("emptyLabel", emptyLabel);
        i18nParams.put("startLabel", minDefinition.getLabel());
        i18nParams.put("endLabel", maxDefinition.getLabel());
        errors.addError(new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_VALIDATOR_DATE_INTERVAL_NULL_ERROR", i18nParams));
    }
}
