/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data;

import java.util.Collection;
import java.util.Optional;

import org.ametys.plugins.repository.data.external.ExternalizableDataProvider.ExternalizableDataStatus;
import org.ametys.plugins.repository.data.holder.values.SynchronizationContext;

/**
 * Object that gives some context for content attributes synchronization
 */
public class ContentSynchronizationContext extends SynchronizationContext
{
    private boolean _enableInvertRelations = true;
    private Optional<Collection<ReferencedContents>> _referencedContents = Optional.empty();
    
    /**
     * Creates a new instance of a {@link ContentSynchronizationContext}
     */
    protected ContentSynchronizationContext()
    {
        // Empty constructor
    }
    
    /**
     * Creates a new instance of a {@link ContentSynchronizationContext}
     * @return the created instance
     */
    public static ContentSynchronizationContext newInstance()
    {
        return new ContentSynchronizationContext();
    }
    
    /**
     * Determines if the invert relations of the content attributes have to be synchronized
     * @return <code>true</code> if the invert relations of the content attributes have to be synchronized, <code>false</code> otherwise.
     */
    public boolean invertRelationsEnabled()
    {
        return _enableInvertRelations;
    }
    
    /**
     * Set to <code>false</code> to avoid to synchronize the invert relations of the content attributes (default to <code>true</code>)
     * @param enableInvertRelations <code>true</code> to synchronized the invert relations of the content attributes, <code>false</code> otherwise.
     * @return the current {@link ContentSynchronizationContext}
     */
    public ContentSynchronizationContext withInvertRelations(boolean enableInvertRelations)
    {
        _enableInvertRelations = enableInvertRelations;
        return this;
    }
    
    /**
     * Retrieves the referenced contents to modify for invert relations
     * @return the referenced contents to modify for invert relations
     */
    public Optional<Collection<ReferencedContents>> getReferencedContents()
    {
        return _referencedContents;
    }
    
    /**
     * Sets the referenced contents to modify for invert relations
     * @param referencedContents the {@link ReferencedContents}s to set
     * @return the current {@link ContentSynchronizationContext}
     */
    public ContentSynchronizationContext withReferencedContents(Collection<ReferencedContents> referencedContents)
    {
        _referencedContents = Optional.ofNullable(referencedContents);
        return this;
    }
    
    @Override
    public ContentSynchronizationContext withDefaultFromModel(boolean useDefaultFromModel)
    {
        return (ContentSynchronizationContext) super.withDefaultFromModel(useDefaultFromModel);
    }
    
    @Override
    public ContentSynchronizationContext withStatus(ExternalizableDataStatus externalizableDataStatus)
    {
        return (ContentSynchronizationContext) super.withStatus(externalizableDataStatus);
    }
}
