/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data;

import java.io.InputStream;
import java.time.ZoneId;
import java.time.ZonedDateTime;

import javax.jcr.RepositoryException;
import javax.jcr.Session;

import org.ametys.core.util.DateUtils;
import org.ametys.plugins.explorer.resources.ModifiableResource;
import org.ametys.plugins.explorer.resources.Resource;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.UnknownAmetysObjectException;

/**
 * Class representing a file from explorer
 */
public class ExplorerFile implements File
{
    private AmetysObjectResolver _resolver;
    private String _resourceId;
    private Resource _resource;
    private Session _session;
    
    /**
     * Constructor of the explorer file
     * @param resolver resolver used to get the resource from its identifier
     * @param resourceId resource's identifier
     */
    public ExplorerFile(AmetysObjectResolver resolver, String resourceId)
    {
        this(resolver, resourceId, null);
    }
    
    /**
     * Constructor of the explorer file
     * @param resolver resolver used to get the resource from its identifier
     * @param resourceId resource's identifier
     * @param session the current session. If <code>null</code>, a new session will be used to retrieve the resource's data
     */
    public ExplorerFile(AmetysObjectResolver resolver, String resourceId, Session session)
    {
        _resolver = resolver;
        _resourceId = resourceId;
        _session = session;
    }
    
    /**
     * Retrieves the resource's identifier
     * @return the resource's identifier
     */
    public String getResourceId()
    {
        return _resourceId;
    }
    
    public InputStream getInputStream()
    {
        return getResource().getInputStream();
    }
    
    public ZonedDateTime getLastModificationDate()
    {
        return DateUtils.asZonedDateTime(getResource().getLastModified(), ZoneId.systemDefault());
    }

    public void setLastModificationDate(ZonedDateTime lastModificationDate)
    {
        if (getResource() instanceof ModifiableResource resource)
        {
            resource.setLastModified(DateUtils.asDate(lastModificationDate));
        }
        else
        {
            throw new UnsupportedOperationException("Impossible to set the last modification date on resource '" + getResource().getId() + "'. The resource is not modifiable.");
        }
    }

    public long getLength()
    {
        return getResource().getLength();
    }

    public String getMimeType()
    {
        return getResource().getMimeType();
    }

    public void setMimeType(String mimeType)
    {
        if (getResource() instanceof ModifiableResource resource)
        {
            resource.setMimeType(mimeType);
        }
        else
        {
            throw new UnsupportedOperationException("Impossible to set the MimeType on resource '" + getResource().getId() + "'. The resource is not modifiable.");
        }
    }

    public String getName()
    {
        return getResource().getName();
    }

    /**
     * Retrieves the file resource
     * @return the file resource
     * @throws AmetysRepositoryException if an error occurs while resolving the resource by its id 
     * @throws UnknownAmetysObjectException if the resource does not exist
     */
    public Resource getResource() throws AmetysRepositoryException, UnknownAmetysObjectException
    {
        if (_resource == null)
        {
            if (_session != null)
            {
                try
                {
                    _resource = _resolver.resolveById(_resourceId, _session);
                }
                catch (RepositoryException e)
                {
                    throw new AmetysRepositoryException("Unable to retrieve the resource with the id '" + _resourceId + "'.", e);
                }
            }
            else
            {
                _resource = _resolver.resolveById(_resourceId);
            }
        }
        
        return _resource; 
    }
}
