/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.commons.collections4.ListUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.text.StringTokenizer;

import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.i18n.I18nizableTextParameter;
import org.ametys.runtime.parameter.DefaultValidator;
import org.ametys.runtime.parameter.Errors;

/**
 * {@link DefaultValidator} extension to enforce the file 'type' or extension based on the widget params.
 */
public class FileValidator extends DefaultValidator
{
    /** The list of file extension that are consider as image */
    protected static final List<String> IMAGE_FILTER = List.of("jpg", "jpeg", "gif", "png", "svg");
    /** The list of file extension that are considered as image */
    protected static final List<String> VIDEO_FILTER = List.of("ogv", "mp4", "webm", "avi", "mkv", "mov", "mpeg");
    /** The list of file extension that are consider as sound */
    protected static final List<String> SOUND_FILTER = List.of("mp3", "oga", "wav");
    /** The list of file extension that are consider as multimedia */
    protected static final List<String> MULTIMEDIA_FILTER = ListUtils.union(SOUND_FILTER, VIDEO_FILTER);
    /** The list of file extension that are consider as pdf */
    protected static final List<String> PDF_FILTER = List.of("pdf");
    
    private List<String> _allowedExtensions;

    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        super.configure(configuration);
        
        // Check if filtering are enabled in widget-params
        Configuration[] params = configuration.getChild("widget-params").getChildren("param");
        for (Configuration param : params)
        {
            String name = param.getAttribute("name");
            if (StringUtils.equals("allowExtensions", name))
            {
                this._allowedExtensions = StringTokenizer.getCSVInstance(param.getValue()).getTokenList();
            }
            else if (StringUtils.equals("filter", name))
            {
                switch (param.getValue())
                {
                    case "image":
                        this._allowedExtensions = IMAGE_FILTER;
                        break;
                    case "video":
                        this._allowedExtensions = VIDEO_FILTER;
                        break;
                    case "multimedia":
                        this._allowedExtensions = MULTIMEDIA_FILTER;
                        break;
                    case "audio":
                        this._allowedExtensions = SOUND_FILTER;
                        break;
                    case "pdf":
                        this._allowedExtensions = PDF_FILTER;
                        break;
                    default:
                        throw new ConfigurationException("Unsupported file filter value '" + param.getValue() + "'.", configuration);
                }
            }
        }
    }
    
    @Override
    protected void validateSingleValue(Object value, Errors errors)
    {
        super.validateSingleValue(value, errors);
        if (this._allowedExtensions != null && value instanceof File file)
        {
            _checkFile(file, errors);
        }
    }

    private void _checkFile(File file, Errors errors)
    {
        String extension = StringUtils.substringAfterLast(file.getName(), ".");
        extension = StringUtils.lowerCase(extension);
        if (!this._allowedExtensions.contains(extension))
        {
            Map<String, I18nizableTextParameter> params = new HashMap<>();
            params.put("filename", new I18nizableText(file.getName()));
            params.put("extension", new I18nizableText(extension));
            params.put("allowedExtensions", new I18nizableText(this._allowedExtensions.stream().collect(Collectors.joining(", "))));
            errors.addError(new I18nizableText("plugin.cms", "PLUGINS_CMS_FILE_VALIDATOR_INVALID_EXTENSION", params));
        }
    }
    
    @Override
    protected void validateArrayValues(Object[] values, Errors errors)
    {
        super.validateArrayValues(values, errors);
        if (this._allowedExtensions != null && values instanceof File[] files)
        {
            for (File file : files)
            {
                _checkFile(file, errors);
            }
        }
    }
}
