/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.data;

import java.util.Optional;

import org.apache.cocoon.xml.AttributesImpl;
import org.apache.excalibur.xml.sax.ContentHandlerProxy;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.plugins.repository.data.ametysobject.DataAwareAmetysObject;
import org.ametys.plugins.repository.model.RepositoryDataContext;
import org.ametys.plugins.repository.version.VersionAwareAmetysObject;
import org.ametys.runtime.model.type.DataContext;

/**
 * Proxy handler to enhance the rich text attachments.
 * Add the content id and the data path to the local attachment's fileref attribute 
 */
public class LocalMediaObjectHandler extends ContentHandlerProxy
{
    private static final String __ATTACHMENT_IMAGE_TAG_NAME = "imagedata";
    private static final String __ATTACHMENT_VIDEO_TAG_NAME = "videodata";
    private static final String __ATTACHMENT_AUDIO_TAG_NAME = "audiodata";
    private static final String __ATTACHMENT_TYPE_ATTRIBUTE_NAME = "type";
    private static final String __ATTACHMENT_TYPE_ATTRIBUTE_LOCAL_VALUE = "local";

    private DataContext _context;
    
    /**
     * Constructor
     * @param contentHandler the contentHandler to pass SAX events to
     * @param context the context of the rich text to SAX
     */
    public LocalMediaObjectHandler(ContentHandler contentHandler, DataContext context)
    {
        super(contentHandler);
        _context = context;
    }

    @Override
    public void startElement(String uri, String loc, String raw, Attributes attrs) throws SAXException
    {
        // A new attachment starts being saxed
        boolean isAttachment = _isAttachment(loc);
        String type = attrs.getValue(__ATTACHMENT_TYPE_ATTRIBUTE_NAME);
        if (isAttachment && __ATTACHMENT_TYPE_ATTRIBUTE_LOCAL_VALUE.equals(type))
        {
            Attributes newAttrs = _processAttachment(attrs);
            super.startElement(uri, loc, raw, newAttrs);
        }
        else
        {
            super.startElement(uri, loc, raw, attrs);
        }
    }

    private boolean _isAttachment(String loc)
    {
        return __ATTACHMENT_IMAGE_TAG_NAME.equals(loc) || __ATTACHMENT_VIDEO_TAG_NAME.equals(loc) || __ATTACHMENT_AUDIO_TAG_NAME.equals(loc);
    }

    private Attributes _processAttachment(Attributes attrs) throws SAXException
    {
        RepositoryDataContext repoContext = _context instanceof RepositoryDataContext rc ? rc : RepositoryDataContext.newInstance(_context);
        
        DataAwareAmetysObject content = repoContext.getObject()
                                                   .orElseThrow(() -> new SAXException("The object is required in the data context in order to set the fileref"));

        String contentId = content.getId();
        
        Optional<String> contentVersion = Optional.of(content)
                                                  .filter(VersionAwareAmetysObject.class::isInstance)
                                                  .map(VersionAwareAmetysObject.class::cast)
                                                  .map(VersionAwareAmetysObject::getRevision);
        
        String dataPath = _context.getDataPath();
        String fileName = attrs.getValue("fileref");

        AttributesImpl newAttrs = new AttributesImpl();
        _copyAttributes(attrs, newAttrs);

        StringBuilder newFilerefAttribute = new StringBuilder();
        newFilerefAttribute.append(contentId);
        contentVersion.ifPresent(v -> newFilerefAttribute.append("|").append(v));
        newFilerefAttribute.append("@").append(dataPath)
                           .append(";").append(fileName);
        
        newAttrs.addCDATAAttribute("fileref", newFilerefAttribute.toString());

        return newAttrs;
    }

    /**
     * Copy the attributes except the fileref attribute
     * @param attrs the attributes to copy.
     * @param newAttrs the attributes to copy to.
     */
    private void _copyAttributes(Attributes attrs, AttributesImpl newAttrs)
    {
        for (int i = 0; i < attrs.getLength(); i++)
        {
            String name = attrs.getQName(i);

            if (!"fileref".equals(name))
            {
                newAttrs.addAttribute(attrs.getURI(i), attrs.getLocalName(i), name, attrs.getType(i), attrs.getValue(i));
            }
        }
    }
}
