/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data;

import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ametys.cms.contenttype.ContentAttributeDefinition;
import org.ametys.cms.repository.Content;

/**
 * Collection of {@link Content} referenced by an invert relation path
 */
public class ReferencedContents
{
    private ContentAttributeDefinition _definition;
    private Map<ContentValue, List<String>> _addedContents;
    private Map<ContentValue, List<String>> _removedContents;
    private Map<ContentValue, ContentValue> _thirdPartyContents;
    
    /**
     * Creates an instance of {@link ReferencedContents}
     * @param definition the definition of the attribute holding content references
     * @param addedContents the content where to add the invert relation
     * @param removedContents the content where to remove the invert relation
     * @param thirdPartyContents other contents involded indirectly due to the write operation
     */
    public ReferencedContents(ContentAttributeDefinition definition, Map<ContentValue, List<String>> addedContents, Map<ContentValue, List<String>> removedContents, Map<ContentValue, ContentValue> thirdPartyContents)
    {
        _definition = definition;
        _addedContents = addedContents;
        _removedContents = removedContents;
        _thirdPartyContents = thirdPartyContents;
    }
    
    /**
     * Returns the attribute definition.
     * @return the attribute definition.
     */
    public ContentAttributeDefinition getDefinition()
    {
        return _definition;
    }
    
    /**
     * Retrieve the content where to add the invert relation
     * @return the content where to add the invert relation
     */
    public Set<ContentValue> getAddedContents()
    {
        return _addedContents.keySet();
    }
    
    /**
     * Retrieve the content where to add the invert relation, along with associated data paths
     * @return the content where to add the invert relation
     */
    public Map<ContentValue, List<String>> getAddedContentsWithPaths()
    {
        return _addedContents;
    }
    
    /**
     * Retrieve the content where to remove the invert relation
     * @return the content where to remove the invert relation
     */
    public Set<ContentValue> getRemovedContents()
    {
        return _removedContents.keySet();
    }
    
    /**
     * Retrieve the content where to remove the invert relation, along with associated data paths
     * @return the content where to remove the invert relation
     */
    public Map<ContentValue, List<String>> getRemovedContentsWithPaths()
    {
        return _removedContents;
    }
    
    /**
     * Returns the contents which values will change due the current write operation.<br>
     * They are typically contents involved in an single-valued invert relation. The key is the source content, the value is the referenced Content.
     * @return the involved contents
     */
    public Map<ContentValue, ContentValue> getThirdPartyContents()
    {
        return _thirdPartyContents;
    }
    
    @Override
    public String toString()
    {
        return _definition.toString();
    }
}
