/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data;

import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.util.Map;

import javax.jcr.Node;
import javax.jcr.Repository;
import javax.jcr.RepositoryException;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.caching.CacheableProcessingComponent;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.cocoon.reading.AbstractReader;
import org.apache.commons.io.IOUtils;
import org.apache.excalibur.source.SourceValidity;
import org.apache.excalibur.source.impl.validity.TimeStampValidity;
import org.xml.sax.SAXException;

import org.ametys.cms.data.type.ResourceElementTypeHelper;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.provider.AbstractRepository;

/**
 * Read a resource stored in the repository by its UUID
 */
public class ResourceByUUIDReader extends AbstractReader implements Serviceable, CacheableProcessingComponent
{
    private String _uuid;
    private Repository _repository;
    private Resource _resource;
    
    public void service(ServiceManager smanager) throws ServiceException
    {
        _repository = (Repository) smanager.lookup(AbstractRepository.ROLE);
    }
    
    @Override
    public void setup(SourceResolver sourceResolver, Map objectModelMap, String src, Parameters par) throws ProcessingException, SAXException, IOException
    {
        super.setup(sourceResolver, objectModelMap, src, par);
        _uuid = src;
        
        javax.jcr.Session session = null;
        try
        {
            session = _repository.login(); 
            Node node = session.getNodeByIdentifier(_uuid);
            JCRRepositoryData resourceData = new JCRRepositoryData(node);
            _resource = new Resource(resourceData);
            ResourceElementTypeHelper.readResourceData(resourceData, _resource);
        }
        catch (RepositoryException e)
        {
            if (session != null)
            {
                session.logout();
            }

            throw new AmetysRepositoryException("Unable to get resource for uuid: " + _uuid, e);
        }
    }

    public void generate() throws IOException, SAXException, ProcessingException
    {
        try (InputStream is = _resource.getInputStream())
        {
            IOUtils.copy(is, out);
        }
    }

    public Serializable getKey()
    {
        return this.getClass().getName() + "_" + _uuid;
    }

    public SourceValidity getValidity()
    {
        return new TimeStampValidity(getLastModified());
    }
    
    @Override
    public long getLastModified()
    {
        return _resource.getLastModificationDate().toInstant().toEpochMilli();
    }
    
    @Override
    public void recycle()
    {
        super.recycle();
        _resource = null;
    }
}
