/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data;

import java.io.IOException;
import java.io.InputStream;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.excalibur.xml.sax.SAXParser;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import org.ametys.cms.content.RichTextHandler;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Rich-text helper
 */
public class RichTextHelper extends AbstractLogEnabled implements Serviceable, Component
{
    /** The Avalon role */
    public static final String ROLE = RichTextHelper.class.getName();
    
    /** The service manager */
    protected ServiceManager _manager;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
    }
    
    /**
     * Transform the rich text to string
     * @param richText the rich-text to transform
     * @return the transformed rich-text
     */
    public String richTextToString(RichText richText)
    {
        return richTextToString(richText, 0);
    }
    
    /**
     * Transform the rich text to string
     * @param richText the rich-text to transform
     * @param maxSize the max size of the rich-text. 0 if no max size
     * @return the transformed rich-text
     */
    public String richTextToString(RichText richText, int maxSize)
    {
        SAXParser saxParser = null;
        try (InputStream inputStream = richText.getInputStream())
        {
            RichTextHandler txtHandler = new RichTextHandler(maxSize);
            saxParser = (SAXParser) _manager.lookup(SAXParser.ROLE);
            saxParser.parse(new InputSource(inputStream), txtHandler);
            return txtHandler.getValue().trim();
        }
        catch (ServiceException e)
        {
            getLogger().error("Unable to get a SAX parser", e);
            return null;
        }
        catch (IOException | SAXException e)
        {
            getLogger().error("Cannot parse inputstream", e);
            return null;
        }
        finally
        {
            _manager.release(saxParser);
        }
    }
}
