/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.data;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Map;
import java.util.Optional;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.cocoon.Constants;
import org.apache.cocoon.environment.Context;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.xml.sax.ContentHandlerProxy;
import org.apache.http.HttpEntity;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.entity.ContentType;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.slf4j.Logger;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.repository.comment.CommentsDAO;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Factory for the transformer that imports a rich text from docbook.
 */
public class RichTextImportHandlerFactory extends AbstractLogEnabled implements Component, Contextualizable
{
    /** Avalon role. */
    public static final String ROLE = RichTextImportHandlerFactory.class.getName();
    private Context _cocoonContext;
    
    public void contextualize(org.apache.avalon.framework.context.Context context) throws ContextException
    {
        _cocoonContext = (Context) context.get(Constants.CONTEXT_ENVIRONMENT_CONTEXT);
    }

    /**
     * Creates a handler proxy to import the rich text
     * @param contentHandler the contentHandler to pass SAX events to
     * @param richText the rich text
     * @param files the attachments of this rich text
     * @return the created handler
     */
    public RichTextImportHandler createHandlerProxy(ContentHandler contentHandler, RichText richText, Map<String, InputStream> files)
    {
        return new RichTextImportHandler(contentHandler, richText, files, getLogger());
    }
    
    /**
     * This transformer imports the rich text from docbook.
     */
    public class RichTextImportHandler extends ContentHandlerProxy
    {
        private static final String __ATTACHMENT_IMAGE_TAG_NAME = "imagedata";
        private static final String __ATTACHMENT_VIDEO_TAG_NAME = "videodata";
        private static final String __ATTACHMENT_AUDIO_TAG_NAME = "audiodata";
        private static final String __ATTACHMENT_TYPE_ATTRIBUTE_NAME = "type";
        private static final String __ATTACHMENT_TYPE_ATTRIBUTE_LOCAL_VALUE = "local";

        private static final String __ANNOTATION_TAG_NAME = "phrase";
        private static final String __ANNOTATION_NAME_ATTRIBUTE_NAME = "role";
        private static final String __ANNOTATION_CLASS_ATTRIBUTE_NAME = "class";
        private static final String __ANNOTATION_CLASS_ATTRIBUTE_VALUE = "semantic";


        private RichText _richText;
        private Map<String, InputStream> _files;
        private Logger _logger;

        private boolean _isCurrentlyInAnnotation;
        private String _currentAnnotationName;
        private StringBuilder _currentAnnotationValue;
        private int _cptrElementsInsideCurrentAnnotation;
        
        /**
         * Creates a handler proxy to import a rich text
         * @param contentHandler the contentHandler to pass SAX events to
         * @param richText the rich text
         * @param files the attachments of this rich text
         * @param logger the logger
         */
        public RichTextImportHandler(ContentHandler contentHandler, RichText richText, Map<String, InputStream> files, Logger logger)
        {
            super(contentHandler);
            _richText = richText;
            _files = files;
            _logger = logger;
        } 

        @Override
        public void startDocument() throws SAXException
        {
            // Remove all existing attachments from the rich text.
            _richText.removeAttachments();

            // Remove all existing annotations from the rich text.
            _richText.removeAllAnnotations();

            super.startDocument();
        }

        @Override
        public void startElement(String uri, String loc, String raw, Attributes attrs) throws SAXException
        {
            // A new attachment starts being saxed
            boolean isAttachment = _isAttachment(loc);
            String type = attrs.getValue(__ATTACHMENT_TYPE_ATTRIBUTE_NAME);
            Attributes newAttrs = attrs;
            if (isAttachment && __ATTACHMENT_TYPE_ATTRIBUTE_LOCAL_VALUE.equals(type))
            {
                newAttrs = _processAttachment(attrs);
            }

            // A new semantic annotation starts being saxed
            String clazz = attrs.getValue(__ANNOTATION_CLASS_ATTRIBUTE_NAME);
            String annotationName = attrs.getValue(__ANNOTATION_NAME_ATTRIBUTE_NAME);
            if (__ANNOTATION_TAG_NAME.equals(loc) && __ANNOTATION_CLASS_ATTRIBUTE_VALUE.equals(clazz) && annotationName != null) 
            {
                _processAnnotation(attrs);
            }
            else if (_isCurrentlyInAnnotation)
            {
                // A new element is being SAXed inside the current annotation
                _cptrElementsInsideCurrentAnnotation++;
            }

            super.startElement(uri, loc, raw, newAttrs);
        }
        
        private boolean _isAttachment(String loc)
        {
            return __ATTACHMENT_IMAGE_TAG_NAME.equals(loc) || __ATTACHMENT_VIDEO_TAG_NAME.equals(loc) || __ATTACHMENT_AUDIO_TAG_NAME.equals(loc);
        }

        private Attributes _processAttachment(Attributes attrs) throws SAXException
        {
            String fileRefAttribute = attrs.getValue("fileref");
            String filename;
            
            if (CommentsDAO.URL_VALIDATOR.matcher(fileRefAttribute).matches())
            {
                try
                {
                    NamedResource attachment = new NamedResource();
    
                    RequestConfig requestConfig = RequestConfig.custom()
                            .setConnectTimeout(2000)
                            .setSocketTimeout(2000)
                            .build();
                    
                    URL url = new URL(fileRefAttribute);
                    String path = url.getPath();
                    filename = path.substring(path.lastIndexOf("/") + 1);
                    
                    try (CloseableHttpClient httpclient = HttpClientBuilder.create()
                                                                           .setDefaultRequestConfig(requestConfig)
                                                                           .useSystemProperties()
                                                                           .build())
                    {
                        HttpGet httpGet = new HttpGet(fileRefAttribute);
                        try (CloseableHttpResponse httpResponse = httpclient.execute(httpGet))
                        {
                            if (httpResponse.getStatusLine().getStatusCode() != 200)
                            {
                                _logger.warn("Can't import file with url '" + fileRefAttribute + "' in the imported rich text");
                            }
                            else
                            {
                                HttpEntity entity = httpResponse.getEntity();
                                try (InputStream is = entity.getContent())
                                {
                                    if (is == null)
                                    {
                                        _logger.warn("The attachment named '" + filename + "' of the imported rich text is empty");
                                    }
                                    else
                                    {
                                        attachment.setInputStream(is);
                                    }
                                    
                                    String mimeType = Optional.ofNullable(ContentType.get(entity))
                                            .map(ContentType::getMimeType)
                                            .filter(StringUtils::isNotEmpty)
                                            .orElse(_cocoonContext.getMimeType(filename.toLowerCase()));

                                    attachment.setMimeType(mimeType);
                                    attachment.setFilename(filename);
                                    
                                    _richText.addAttachment(attachment);
                                }
                            }
                        }
                    }
                }
                catch (IOException e)
                {
                    throw new SAXException("Unable to process the attachment '" + fileRefAttribute + "'. An error occured while setting its content", e);
                }
            }
            else
            {
                // file reference is of the form ownerId@dataName;fileName
                int indexOfFilenameSeparator = fileRefAttribute.lastIndexOf(';');
                filename = fileRefAttribute.substring(indexOfFilenameSeparator + 1);
    
                if (indexOfFilenameSeparator == -1)
                {
                    throw new IllegalArgumentException("A local image should have a file reference of the form <protocol>://<protocol-specific-part>;<filename> : " + fileRefAttribute);
                }
    
                if (_files.containsKey(filename))
                {
                    try
                    {
                        NamedResource attachment = new NamedResource();
                        String mimeType = _cocoonContext.getMimeType(filename.toLowerCase());
                        attachment.setMimeType(mimeType);
                        attachment.setFilename(filename);
                        attachment.setInputStream(_files.get(filename));
                        _richText.addAttachment(attachment);
                    }
                    catch (IOException e)
                    {
                        throw new SAXException("Unable to process the attachment '" + filename + "'. An error occured while setting its content", e);
                    }
                }
            }
            
            AttributesImpl newAttrs = new AttributesImpl();
            _copyAttributes(attrs, newAttrs);
            newAttrs.addCDATAAttribute("fileref", filename);
            return newAttrs;
        }
        
        /**
         * Copy the attributes except the fileref attribute
         * @param attrs the attributes to copy.
         * @param newAttrs the attributes to copy to.
         */
        private void _copyAttributes(Attributes attrs, AttributesImpl newAttrs)
        {
            for (int i = 0; i < attrs.getLength(); i++)
            {
                String name = attrs.getQName(i);

                if (!"fileref".equals(name))
                {
                    newAttrs.addAttribute(attrs.getURI(i), attrs.getLocalName(i), name, attrs.getType(i), attrs.getValue(i));
                }
            }
        }

        private void _processAnnotation(Attributes attrs)
        {
            _isCurrentlyInAnnotation = true;
            _currentAnnotationName = attrs.getValue(__ANNOTATION_NAME_ATTRIBUTE_NAME);
            _currentAnnotationValue = new StringBuilder();
            _cptrElementsInsideCurrentAnnotation = 0;
        }

        @Override
        public void characters(char[] ch, int start, int length) throws SAXException
        {
            if (_isCurrentlyInAnnotation)
            {
                _currentAnnotationValue.append(ch, start, length);
            }
            
            super.characters(ch, start, length);
        }

        @Override
        public void endElement(String uri, String loc, String raw) throws SAXException
        {    
            if (_isCurrentlyInAnnotation)
            {
                if (_cptrElementsInsideCurrentAnnotation == 0)
                {                
                    // When the semantic annotation is fully saxed, add it to the rich text
                    _richText.addAnnotations(_currentAnnotationName, _currentAnnotationValue.toString());
                    _isCurrentlyInAnnotation = false;
                }
                else 
                {
                    _cptrElementsInsideCurrentAnnotation--;                
                }
            }
            
            super.endElement(uri, loc, raw);
        }
    }
}
