/*
 *  Copyright 2024 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.ametysobject;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;

import org.ametys.cms.data.ContentValue;
import org.ametys.cms.model.ContentElementDefinition;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.impl.DataHolderHelper;
import org.ametys.plugins.repository.data.holder.values.SynchronizableRepeater;
import org.ametys.plugins.repository.data.holder.values.SynchronizableValue;
import org.ametys.plugins.repository.data.holder.values.SynchronizableValue.Mode;
import org.ametys.plugins.repository.data.holder.values.SynchronizationContext;
import org.ametys.plugins.repository.data.holder.values.UntouchedValue;
import org.ametys.plugins.repository.model.CompositeDefinition;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemAccessor;
import org.ametys.runtime.model.disableconditions.AbstractDisableConditionsEvaluator;
import org.ametys.runtime.model.disableconditions.DisableConditions;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.type.ElementType;

/**
 * Evaluator for {@link DisableConditions} on model aware {@link AmetysObject}
 */
public class AmetysObjectDisableConditionsEvaluator extends AbstractDisableConditionsEvaluator<ModelAwareDataAwareAmetysObject>
{
    /** The Avalon role */
    public static final String ROLE = AmetysObjectDisableConditionsEvaluator.class.getName();
    
    /** The contextual parameter key for synchronization context */
    public static final String SYNCHRONIZATION_CONTEXT_PARAMETER_KEY = "synchronizationContext";
    
    @Override
    protected Optional<ElementDefinition> getConditionDefinition(String conditionDataPath, ModelAwareDataAwareAmetysObject ametysObject)
    {
        return Optional.ofNullable(conditionDataPath)
                       .map(ametysObject::getDefinition)
                       .filter(ElementDefinition.class::isInstance)
                       .map(ElementDefinition.class::cast);
    }
    
    
    /**
     * {@inheritDoc}
     * @throws IllegalArgumentException if the given condition data path is empty
     * @throws BadItemTypeException if a value does not correspond to the model of given object
     */
    @Override
    protected Optional<Object> getValueFromMap(String conditionDataPath, Optional<String> oldConditionDataPath, Map<String, Object> values, ModelAwareDataAwareAmetysObject ametysObject, Map<String, Object> contextualParameters) throws IllegalArgumentException, BadItemTypeException
    {
        return _getValueFromMap(conditionDataPath, oldConditionDataPath, values, ametysObject, ametysObject.getModel(), contextualParameters);
    }
    
    /**
     * Retrieves the condition value from the values {@link Map}
     * @param conditionDataPath the absolute data path of the condition
     * @param oldConditionDataPath the absolute old data path of the condition
     * @param values the values {@link Map}
     * @param ametysObject the ametys object holding the data to evaluate
     * @param modelItemAccessors the relative accessor to the given condition path
     * @param contextualParameters the contextual parameters
     * @return the condition value if found in the values {@link Map}
     * @throws IllegalArgumentException if the given condition data path is empty
     * @throws BadItemTypeException if a value does not correspond to the given model item accessors 
     */
    protected Optional<Object> _getValueFromMap(String conditionDataPath, Optional<String> oldConditionDataPath, Map<String, Object> values, ModelAwareDataAwareAmetysObject ametysObject, Collection<? extends ModelItemAccessor> modelItemAccessors, Map<String, Object> contextualParameters) throws IllegalArgumentException, BadItemTypeException
    {
        String[] conditionDataPathSegments = StringUtils.split(conditionDataPath, ModelItem.ITEM_PATH_SEPARATOR);
        
        if (conditionDataPathSegments == null || conditionDataPathSegments.length < 1)
        {
            throw new IllegalArgumentException("Unable to retrieve the value at the given condition path. This path is empty.");
        }
        
        ModelItem modelItem = ModelHelper.getModelItem(conditionDataPathSegments[0], modelItemAccessors);
        if (conditionDataPathSegments.length == 1)
        {
            return _getElementValueFromMap(conditionDataPathSegments[0], oldConditionDataPath, values, ametysObject, modelItem, contextualParameters);
        }
        else
        {
            String subConditionDataPath = StringUtils.join(conditionDataPathSegments, ModelItem.ITEM_PATH_SEPARATOR, 1, conditionDataPathSegments.length);
            if (modelItem instanceof CompositeDefinition compositeDefinition)
            {
                return _getValueFromComposite(conditionDataPathSegments[0], oldConditionDataPath, subConditionDataPath, values, ametysObject, compositeDefinition, contextualParameters);
            }
            else if (modelItem instanceof RepeaterDefinition repeaterDefinition)
            {
                return _getValueFromRepeater(conditionDataPathSegments[0], oldConditionDataPath, subConditionDataPath, values, ametysObject, repeaterDefinition, contextualParameters);
            }
            else if (modelItem instanceof ContentElementDefinition contentElementDefinition)
            {
                return _getValueFromContentValue(conditionDataPathSegments[0], subConditionDataPath, values, contentElementDefinition, contextualParameters);
            }
            else
            {
                throw new BadItemTypeException("Unable to retrieve the value at the given path. the the segment '" + conditionDataPathSegments[0] + "' is not a model item accessor");
            }
        }
    }
    
    /**
     * Retrieves the condition value from the values {@link Map}
     * @param dataName the name of the condition element
     * @param oldConditionDataPath the absolute old data path of the condition
     * @param values the values {@link Map}
     * @param ametysObject the ametys object holding the data to evaluate
     * @param definition the definition of the condition element
     * @param contextualParameters the contextual parameters
     * @return the condition value if found in the values {@link Map}
     * @throws IllegalArgumentException if the given condition data path is empty
     * @throws BadItemTypeException if a value does not correspond to the given model item accessors 
     */
    protected Optional<Object> _getElementValueFromMap(String dataName, Optional<String> oldConditionDataPath, Map<String, Object> values, ModelAwareDataAwareAmetysObject ametysObject, ModelItem definition, Map<String, Object> contextualParameters)
    {
        Object value = values.get(dataName);
        
        if (value == null || value instanceof UntouchedValue)
        {
            return Optional.empty();
        }
        else
        {
            SynchronizationContext synchronizationContext = (SynchronizationContext) contextualParameters.get(SYNCHRONIZATION_CONTEXT_PARAMETER_KEY);
            if (synchronizationContext != null)
            {
                Object valueFromSyncValue = DataHolderHelper.getValueFromSynchronizableValue(value, ametysObject, definition, oldConditionDataPath, synchronizationContext);
                if (valueFromSyncValue == null
                    || valueFromSyncValue instanceof UntouchedValue
                    || value instanceof SynchronizableValue syncValue && syncValue.getMode() == Mode.REMOVE)
                {
                    return Optional.empty();
                }
                else
                {
                    return Optional.of(valueFromSyncValue);
                }
            }
            else
            {
                return Optional.of(value);
            }
        }
    }
    
    /**
     * Retrieves the condition value in the given composite, from the values {@link Map}
     * @param compositeName the name of the composite
     * @param oldConditionDataPath the absolute old data path of the condition
     * @param subConditionDataPath the data path of the condition, relative to the composite
     * @param values the values {@link Map}
     * @param ametysObject the ametys object holding the data to evaluate
     * @param compositeDefinition the composite definition 
     * @param contextualParameters the contextual parameters
     * @return the condition value if found in the values {@link Map}
     * @throws IllegalArgumentException if the given condition data path is empty
     * @throws BadItemTypeException if a value does not correspond to the given model item accessors 
     */
    @SuppressWarnings("unchecked")
    protected Optional<Object> _getValueFromComposite(String compositeName, Optional<String> oldConditionDataPath, String subConditionDataPath, Map<String, Object> values, ModelAwareDataAwareAmetysObject ametysObject, CompositeDefinition compositeDefinition, Map<String, Object> contextualParameters) throws IllegalArgumentException, BadItemTypeException
    {
        Object value = values.get(compositeName);
        if (value == null || value instanceof UntouchedValue)
        {
            return Optional.empty();
        }
        else if (value instanceof Map)
        {
            return _getValueFromMap(subConditionDataPath, oldConditionDataPath, (Map<String, Object>) value, ametysObject, List.of(compositeDefinition), contextualParameters);
        }
        else
        {
            throw new BadItemTypeException("Unable to retrieve the value at the given path. the value of the segment '" + compositeName + "' for composite is not a Map");
        }
    }
    
    /**
     * Retrieves the condition value in the given repeater, from the values {@link Map}
     * @param repeaterNameAndPositionSegment the segment of the repeater, containing repeater name and entry position
     * @param oldConditionDataPath the absolute old data path of the condition
     * @param subConditionDataPath the data path of the condition, relative to the composite
     * @param values the values {@link Map}
     * @param ametysObject the ametys object holding the data to evaluate
     * @param repeaterDefinition the repeater definition 
     * @param contextualParameters the contextual parameters
     * @return the condition value if found in the values {@link Map}
     * @throws IllegalArgumentException if the given condition data path is empty
     * @throws BadItemTypeException if a value does not correspond to the given model item accessors 
     */
    @SuppressWarnings("unchecked")
    protected Optional<Object> _getValueFromRepeater(String repeaterNameAndPositionSegment, Optional<String> oldConditionDataPath, String subConditionDataPath, Map<String, Object> values, ModelAwareDataAwareAmetysObject ametysObject, RepeaterDefinition repeaterDefinition, Map<String, Object> contextualParameters) throws IllegalArgumentException, BadItemTypeException
    {
        Pair<String, Integer> repeaterNameAndEntryPosition = DataHolderHelper.getRepeaterNameAndEntryPosition(repeaterNameAndPositionSegment);
        String repeaterName = repeaterNameAndEntryPosition.getLeft();
        
        Object value = values.get(repeaterName);
        if (value == null || value instanceof UntouchedValue)
        {
            return Optional.empty();
        }
        else if (value instanceof List || value instanceof SynchronizableRepeater)
        {
            Integer entryPosition = repeaterNameAndEntryPosition.getRight();
            List<Map<String, Object>> entries = value instanceof List ? (List<Map<String, Object>>) value : ((SynchronizableRepeater) value).getEntries();
            SynchronizableRepeater.Mode mode = value instanceof SynchronizableRepeater ? ((SynchronizableRepeater) value).getMode() : SynchronizableRepeater.Mode.REPLACE_ALL;
            
            if (mode == SynchronizableRepeater.Mode.REPLACE_ALL)
            {
                Map<String, Object> entry = entries.get(entryPosition - 1);
                return _getValueFromMap(subConditionDataPath, oldConditionDataPath, entry, ametysObject, List.of(repeaterDefinition), contextualParameters);
            }
            else if (mode == SynchronizableRepeater.Mode.REPLACE)
            {
                int indexOfEntryInReplaceEntries = ((SynchronizableRepeater) value).getReplacePositions().indexOf(entryPosition);
                if (indexOfEntryInReplaceEntries >= 0)
                {
                    Map<String, Object> entry = entries.get(indexOfEntryInReplaceEntries);
                    return _getValueFromMap(subConditionDataPath, oldConditionDataPath, entry, ametysObject, List.of(repeaterDefinition), contextualParameters);
                }
                else
                {
                    return Optional.empty();
                }
            }
            else // mode == SynchronizableRepeater.Mode.APPEND
            {
                Set<Integer> removedEntries = ((SynchronizableRepeater) value).getRemovedEntries();
                if (removedEntries.contains(entryPosition))
                {
                    // Non sense, we should not evaluate disable condiitons on a data in a repeater entry that will be removed
                    return Optional.empty();
                }
                
                String repeaterPath = repeaterDefinition.getPath();
                ModelAwareRepeater repeater = ametysObject.getRepeater(repeaterPath);
                int actualRepeaterSize = repeater != null ? repeater.getSize() : 0;
                if (removedEntries.size() > actualRepeaterSize)
                {
                    throw new IllegalArgumentException("Try to remove more entries than exist in repeater at path '" + repeaterPath + "'.");
                }
                
                int repeaterSizeAfterRemoving = actualRepeaterSize - removedEntries.size();
                if (entryPosition <= repeaterSizeAfterRemoving)
                {
                    // The entry is not part of the appended entries
                    return Optional.empty();
                }
                else
                {
                    int positionInAppendedEntries = entryPosition - repeaterSizeAfterRemoving;
                    int indexInAppendedEntries = positionInAppendedEntries - 1;
                    Map<String, Object> entry = entries.get(indexInAppendedEntries);
                    return _getValueFromMap(subConditionDataPath, oldConditionDataPath, entry, ametysObject, List.of(repeaterDefinition), contextualParameters);
                }
            }
        }
        else
        {
            throw new BadItemTypeException("Unable to retrieve the value at the given path. the value of the segment '" + repeaterNameAndPositionSegment + "' for repeater is not a List<Map>");
        }
    }
    
    /**
     * Retrieves the condition value in the given content value, from the values {@link Map}
     * @param dataName the name of the content data
     * @param subConditionDataPath the data path of the condition, relative to the composite
     * @param values the values {@link Map}
     * @param contentElementDefinition the content element definition 
     * @param contextualParameters the contextual parameters
     * @return the condition value if found in the values {@link Map}
     * @throws IllegalArgumentException if the given condition data path is empty
     * @throws BadItemTypeException if a value does not correspond to the given model item accessors 
     */
    protected Optional<Object> _getValueFromContentValue(String dataName, String subConditionDataPath, Map<String, Object> values, ContentElementDefinition contentElementDefinition, Map<String, Object> contextualParameters) throws IllegalArgumentException, BadItemTypeException
    {
        Object value = values.get(dataName);

        if (value == null || value instanceof UntouchedValue)
        {
            return Optional.empty();
        }
        else
        {
            ElementType<ContentValue> type = contentElementDefinition.getType();
            ContentValue contentValue = type.castValue(value);
            return contentValue.getContentIfExists()
                               .map(content -> content.getValue(subConditionDataPath));
        }
    }

    @Override
    protected Object getStoredValue(String conditionDataPath, Optional<String> oldConditionDataPath, ModelAwareDataAwareAmetysObject ametysObject, Map<String, Object> contextualParameters)
    {
        String dataPath = oldConditionDataPath.orElse(conditionDataPath);
        return ametysObject.getValue(dataPath);
    }
}
