/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.ametysobject;

import java.util.List;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.apache.solr.common.SolrInputDocument;

import org.ametys.cms.data.holder.IndexableDataHolder;
import org.ametys.cms.data.holder.group.IndexableComposite;
import org.ametys.cms.data.holder.group.IndexableRepeater;
import org.ametys.cms.data.holder.impl.IndexableDataHolderHelper;
import org.ametys.cms.data.type.indexing.IndexableDataContext;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.runtime.model.ViewHelper;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.exception.BadDataPathCardinalityException;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.UndefinedItemPathException;

/**
 * Model aware {@link AmetysObject} that can handle indexable data.
 */
public interface ModelAwareDataAwareAmetysObject extends org.ametys.plugins.repository.data.ametysobject.ModelAwareDataAwareAmetysObject, IndexableDataHolder
{
    @Override
    public IndexableDataHolder getDataHolder();
    
    /**
     * Retrieves the system properties extension point of this {@link AmetysObject}, or an empty {@link Optional} if this object does not support system properties 
     * @return the system properties extension point of this {@link AmetysObject}
     */
    public default Optional<SystemPropertyExtensionPoint> getSystemPropertyExtensionPoint()
    {
        return Optional.empty();
    }
    
    /**
     * Indexes all data and properties of this {@link AmetysObject}
     * @param document the solr document representing this {@link AmetysObject}
     * @return additional solr documents that may have been created (ex: repeater entries)
     * @throws BadItemTypeException if the saxed value's type does not matches the stored data
     */
    public default List<SolrInputDocument> indexData(SolrInputDocument document) throws BadItemTypeException
    {
        return indexData(document, IndexableDataContext.newInstance());
    }
    
    /**
     * Indexes all data and properties of this {@link AmetysObject}
     * @param document the solr document representing this {@link AmetysObject}
     * @param context The context of the data to index
     * @return additional solr documents that may have been created (ex: repeater entries)
     * @throws BadItemTypeException if the saxed value's type does not matches the stored data
     */
    @SuppressWarnings("unchecked")
    public default List<SolrInputDocument> indexData(SolrInputDocument document, IndexableDataContext context) throws BadItemTypeException
    {
        IndexableDataContext newContext = context.cloneContext()
                                                 .withObject(this);
        
        // Index system properties
        if (getSystemPropertyExtensionPoint().isPresent())
        {
            SystemPropertyExtensionPoint systemPropertyExtensionPoint = getSystemPropertyExtensionPoint().get();
            for (String systemPropertyId : systemPropertyExtensionPoint.getExtensionsIds())
            {
                SystemProperty systemProperty = systemPropertyExtensionPoint.getExtension(systemPropertyId);
                IndexableDataContext systemPropertyContext = newContext.cloneContext()
                                                           .addSegmentToDataPath(systemPropertyId)
                                                           .withModelItem(systemProperty);
                systemProperty.indexValue(document, this, systemPropertyContext);
            }
        }
        
        // Index data
        ViewItemAccessor viewItemAccessor = ViewHelper.createViewItemAccessor(getModel());
        return IndexableDataHolderHelper.indexData(getDataHolder(), viewItemAccessor, document, document, StringUtils.EMPTY, newContext);
    }
    
    @Override
    public default IndexableComposite getComposite(String compositePath) throws IllegalArgumentException, UndefinedItemPathException, BadItemTypeException, BadDataPathCardinalityException
    {
        return getDataHolder().getComposite(compositePath);
    }

    @Override
    public default IndexableComposite getLocalComposite(String compositePath) throws IllegalArgumentException, UndefinedItemPathException, BadItemTypeException, BadDataPathCardinalityException
    {
        return getDataHolder().getLocalComposite(compositePath);
    }

    @Override
    public default IndexableComposite getExternalComposite(String compositePath) throws IllegalArgumentException, UndefinedItemPathException, BadItemTypeException, BadDataPathCardinalityException
    {
        return getDataHolder().getExternalComposite(compositePath);
    }

    @Override
    public default IndexableRepeater getRepeater(String repeaterPath) throws IllegalArgumentException, UndefinedItemPathException, BadItemTypeException, BadDataPathCardinalityException
    {
        return getDataHolder().getRepeater(repeaterPath);
    }

    @Override
    public default IndexableRepeater getLocalRepeater(String repeaterPath) throws IllegalArgumentException, UndefinedItemPathException, BadItemTypeException, BadDataPathCardinalityException
    {
        return getDataHolder().getLocalRepeater(repeaterPath);
    }
    
    @Override
    public default IndexableRepeater getExternalRepeater(String repeaterPath) throws IllegalArgumentException, UndefinedItemPathException, BadItemTypeException, BadDataPathCardinalityException
    {
        return getDataHolder().getExternalRepeater(repeaterPath);
    }
    
    @Override
    public default Optional<? extends IndexableDataHolder> getParentDataHolder()
    {
        return getDataHolder().getParentDataHolder();
    }
    
    @Override
    public default IndexableDataHolder getRootDataHolder()
    {
        return getDataHolder().getRootDataHolder();
    }
}
