/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder.group.impl;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.solr.common.SolrInputDocument;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.content.indexing.solr.SolrFieldNames;
import org.ametys.cms.data.holder.IndexableDataHolder;
import org.ametys.cms.data.holder.group.IndexableRepeater;
import org.ametys.cms.data.holder.group.IndexableRepeaterEntry;
import org.ametys.cms.data.holder.impl.IndexableDataHolderHelper;
import org.ametys.cms.data.type.indexing.IndexableDataContext;
import org.ametys.plugins.repository.data.holder.DataHolder;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeaterEntry;
import org.ametys.plugins.repository.data.holder.group.ModifiableRepeater;
import org.ametys.plugins.repository.data.holder.group.ModifiableRepeaterEntry;
import org.ametys.plugins.repository.data.holder.group.Repeater;
import org.ametys.plugins.repository.data.holder.group.RepeaterEntry;
import org.ametys.plugins.repository.data.holder.values.SynchronizableRepeater;
import org.ametys.plugins.repository.data.holder.values.SynchronizationContext;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.ViewHelper;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.NotUniqueTypeException;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.model.exception.UnknownTypeException;
import org.ametys.runtime.model.type.DataContext;

/**
 * Class for model aware repeaters
 */
public class DefaultModelAwareRepeater implements IndexableRepeater
{
    private static final Logger __LOGGER = LoggerFactory.getLogger(ModelAwareDataHolder.class);

    /** Definition of this repeater */
    protected RepeaterDefinition _definition;
    
    /** Parent of the current {@link Repeater} */
    protected IndexableDataHolder _parent;
    
    /** Root {@link DataHolder} */
    protected IndexableDataHolder _root;
    
    /** Repository data to use to store entries in the repository */
    protected RepositoryData _repositoryData;
    
    /**
     * Creates a model aware repeater
     * @param repositoryData the repository data of the repeater
     * @param parent the parent of the created {@link Repeater}
     * @param root the root {@link DataHolder}
     * @param definition the definition of the repeater
     */
    public DefaultModelAwareRepeater(RepositoryData repositoryData, IndexableDataHolder parent, IndexableDataHolder root, RepeaterDefinition definition)
    {
        _repositoryData = repositoryData;
        _definition = definition;
        _parent = parent;
        _root = root;
    }
    
    @Override
    public List<? extends IndexableRepeaterEntry> getEntries()
    {
        SortedSet<IndexableRepeaterEntry> entries = new TreeSet<>(new Comparator<ModelAwareRepeaterEntry>()
        {
            public int compare(ModelAwareRepeaterEntry entry1, ModelAwareRepeaterEntry entry2)
            {
                return Integer.compare(entry1.getPosition(), entry2.getPosition());
            }
        });
        
        for (String entryName : _repositoryData.getDataNames())
        {
            IndexableRepeaterEntry entry = getEntry(Integer.parseInt(entryName));
            entries.add(entry);
        }
        
        return Collections.unmodifiableList(new ArrayList<>(entries));
    }
    
    public IndexableRepeaterEntry getEntry(int position)
    {
        if (1 <= position && position <= getSize())
        {
            RepositoryData entryRepositoryData = _repositoryData.getRepositoryData(String.valueOf(position));
            return new DefaultModelAwareRepeaterEntry(entryRepositoryData, this, _definition);
        }
        else if (-getSize() < position && position <= 0)
        {
            // Find the positive equivalent position and call the getEntry method with this position
            return getEntry(getSize() + position);
        }
        else
        {
            return null;
        }
    }

    public int getSize()
    {
        return _repositoryData.getDataNames().size();
    }
    
    public boolean hasEntry(int position)
    {
        if (1 <= position)
        {
            return _repositoryData.hasValue(String.valueOf(position));
        }
        else
        {
            return _repositoryData.hasValue(String.valueOf(getSize() + position));
        }
    }
    
    /**
     * Retrieves the repeater's model
     * @return the repeater's model
     */
    public RepeaterDefinition getModel()
    {
        return _definition;
    }
    
    public void dataToSAX(ContentHandler contentHandler, String dataPath, DataContext context) throws SAXException
    {
        for (ModelAwareRepeaterEntry entry : getEntries())
        {
            XMLUtils.startElement(contentHandler, "entry", _getEntryAttributes(entry));
            entry.dataToSAX(contentHandler, dataPath, context);
            XMLUtils.endElement(contentHandler, "entry");
        }
    }
    
    public void dataToSAX(ContentHandler contentHandler, DataContext context) throws SAXException, BadItemTypeException
    {
        ModelViewItemGroup viewItemGroup = ModelViewItemGroup.of(_definition);
        dataToSAX(contentHandler, viewItemGroup, context);
    }
    
    public void dataToSAX(ContentHandler contentHandler, ViewItemAccessor viewItemAccessor, DataContext context) throws SAXException, BadItemTypeException
    {
        for (ModelAwareRepeaterEntry entry : getEntries())
        {
            XMLUtils.startElement(contentHandler, "entry", _getEntryAttributes(entry));
            entry.dataToSAX(contentHandler, viewItemAccessor, context);
            XMLUtils.endElement(contentHandler, "entry");
        }
    }
    
    public void dataToSAXForEdition(ContentHandler contentHandler, ViewItemAccessor viewItemAccessor, DataContext context) throws SAXException, BadItemTypeException
    {
        for (ModelAwareRepeaterEntry entry : getEntries())
        {
            XMLUtils.startElement(contentHandler, "entry", _getEntryAttributes(entry));
            entry.dataToSAXForEdition(contentHandler, viewItemAccessor, context);
            XMLUtils.endElement(contentHandler, "entry");
        }
    }
    
    private Attributes _getEntryAttributes(ModelAwareRepeaterEntry entry)
    {
        AttributesImpl entryAttrs = new AttributesImpl();
        String entryName = Integer.toString(entry.getPosition());
        entryAttrs.addCDATAAttribute("name", entryName);
        return entryAttrs;
    }
    
    public Map<String, Object> dataToJSON(String dataPath, DataContext context) throws IOException
    {
        return _dataToJSON(Optional.of(dataPath), Optional.empty(), context, false);
    }
    
    public Map<String, Object> dataToJSON(DataContext context) throws BadItemTypeException
    {
        ModelViewItemGroup viewItemGroup = ModelViewItemGroup.of(_definition);
        return dataToJSON(viewItemGroup, context);
    }
    
    public Map<String, Object> dataToJSON(ViewItemAccessor viewItemAccessor, DataContext context) throws BadItemTypeException
    {
        return _dataToJSON(Optional.empty(), Optional.of(viewItemAccessor), context, false);
    }
    
    public Map<String, Object> dataToJSONForEdition(ViewItemAccessor viewItemAccessor, DataContext context) throws BadItemTypeException
    {
        return _dataToJSON(Optional.empty(), Optional.of(viewItemAccessor), context, true);
    }
    
    @SuppressWarnings("unchecked")
    private Map<String, Object> _dataToJSON(Optional<String> dataPath, Optional<ViewItemAccessor> viewItemAccessor, DataContext context, boolean isEdition) throws BadItemTypeException
    {
        List<Map<String, Object>> entriesValues = new ArrayList<>();
        for (ModelAwareRepeaterEntry entry : getEntries())
        {
            DataContext entryContext = context.cloneContext();
            if (StringUtils.isNotEmpty(context.getDataPath()))
            {
                entryContext.addSuffixToLastSegment("[" + entry.getPosition() + "]");
            }
            
            Map<String, Object> entryValues = null;
            if (dataPath.isPresent())
            {
                entryValues = (Map<String, Object>) entry.dataToJSON(dataPath.get(), entryContext);
            }
            else if (viewItemAccessor.isPresent())
            {
                entryValues = isEdition
                        ? entry.dataToJSONForEdition(viewItemAccessor.get(), entryContext)
                        : entry.dataToJSON(viewItemAccessor.get(), entryContext);
            }
            
            entriesValues.add(entryValues);
        }
        
        Map<String, Object> result = new HashMap<>();
        result.put("entryCount", getSize());
        result.put("entries", entriesValues);
        result.put("label", _definition.getLabel());
        
        Optional.ofNullable(_definition.getHeaderLabel())
                .ifPresent(headerLabel -> result.put("header-label", headerLabel));
        
        return result;
    }
    
    /**
     * Generates SAX events for the comments of the data in the given view in the current {@link DataHolder}
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param viewItemAccessor the {@link ViewItemAccessor} referencing the items for which generate SAX events
     * @throws SAXException if an error occurs during the SAX events generation
     */
    public void commentsToSAX(ContentHandler contentHandler, ViewItemAccessor viewItemAccessor) throws SAXException
    {
        for (ModelAwareRepeaterEntry entry : getEntries())
        {
            entry.commentsToSAX(contentHandler, viewItemAccessor);
        }
    }
    
    public void copyTo(ModifiableRepeater repeater) throws UndefinedItemPathException, BadItemTypeException, UnknownTypeException, NotUniqueTypeException
    {
        for (RepeaterEntry entry : getEntries())
        {
            ModifiableRepeaterEntry entryDestination = repeater.addEntry(entry.getPosition());
            entry.copyTo(entryDestination);
        }
    }
    
    public List<SolrInputDocument> indexData(SolrInputDocument document, SolrInputDocument rootDocument, String solrFieldPrefix, IndexableDataContext context) throws BadItemTypeException
    {
        List<SolrInputDocument> additionalDocuments = new ArrayList<>();
        String solrFieldName = solrFieldPrefix + context.getDataPathLastSegment();
        
        for (IndexableRepeaterEntry entry : getEntries())
        {
            // Update the context with entry position
            IndexableDataContext newContext = context.cloneContext()
                                            .addSuffixToLastSegment("[" + entry.getPosition() + "]");
            
            SolrInputDocument repeaterEntryDoc = new SolrInputDocument();
            
            if (!context.indexForFullTextField())
            {
                // Creates a new Solr document for each entry
                String repeaterEntryDocId = document.getField("id").getFirstValue().toString() + "/" + solrFieldName + "/" + entry.getPosition();
                repeaterEntryDoc.addField("id", repeaterEntryDocId);
                repeaterEntryDoc.addField(SolrFieldNames.DOCUMENT_TYPE, SolrFieldNames.TYPE_REPEATER);
                repeaterEntryDoc.addField(SolrFieldNames.REPEATER_ENTRY_POSITION, entry.getPosition());
                
                document.addField(solrFieldName + "_s_dv", repeaterEntryDocId);
            }
            
            // Add the created document to additional documents
            additionalDocuments.add(repeaterEntryDoc);
            
            ViewItemAccessor viewItemAccessor = context.getViewItem()
                .map(ViewItemAccessor.class::cast)
                .orElse(ViewHelper.createViewItemAccessor(entry.getModel()));
            additionalDocuments.addAll(IndexableDataHolderHelper.indexData(entry, viewItemAccessor, repeaterEntryDoc, rootDocument, StringUtils.EMPTY, newContext));
        }
        
        return additionalDocuments;
    }
    
    public boolean hasDifferences(ViewItemAccessor viewItemAccessor, SynchronizableRepeater repeaterValues, SynchronizationContext context) throws UndefinedItemPathException, BadItemTypeException
    {
        return SynchronizableRepeater.Mode.APPEND.equals(repeaterValues.getMode())
                ? _hasDifferencesInAppendMode(viewItemAccessor, repeaterValues, context)
                : SynchronizableRepeater.Mode.REPLACE.equals(repeaterValues.getMode())
                        ? _hasDifferencesInReplaceMode(viewItemAccessor, repeaterValues, context)
                        : _hasDifferencesInReplaceAllMode(viewItemAccessor, repeaterValues, context);
    }
    
    /**
     * Check if there are differences between the given values and the repeater's entries if {@link SynchronizableRepeater#getMode()} is APPEND
     * @param viewItemAccessor The {@link ViewItemAccessor} for all items to check
     * @param repeaterValues the values of the repeater to check
     * @param context the context of the synchronization
     * @return <code>true</code> if there are differences, <code>false</code> otherwise
     * @throws UndefinedItemPathException if an entry's key refers to a data that is not defined by the model
     * @throws BadItemTypeException if the type defined by the model of one of the entry's key doesn't match the corresponding value
     */
    protected boolean _hasDifferencesInAppendMode(ViewItemAccessor viewItemAccessor, SynchronizableRepeater repeaterValues, SynchronizationContext context) throws UndefinedItemPathException, BadItemTypeException
    {
        boolean hasEntriesToAppend = !repeaterValues.getRemovedEntries().isEmpty() || !repeaterValues.getEntries().isEmpty();
        
        if (__LOGGER.isDebugEnabled())
        {
            String viewItemPath = viewItemAccessor instanceof ModelViewItem modelViewItem 
                                    ? ViewHelper.getModelViewItemPath(modelViewItem)
                                    : StringUtils.EMPTY;
            if (hasEntriesToAppend)
            {
                __LOGGER.debug("#hasDifferences[{}] differences detected: some entries will be appended", viewItemPath);
            }
            else
            {
                __LOGGER.debug("#hasDifferences[{}] no difference detected.", viewItemPath);
            }
        }
        
        return hasEntriesToAppend;
    }
    
    /**
     * Check if there are differences between the given values and the repeater's entries if {@link SynchronizableRepeater#getMode()} is REPLACE
     * @param viewItemAccessor The {@link ViewItemAccessor} for all items to check
     * @param repeaterValues the values of the repeater to check
     * @param context the context of the synchronization
     * @return <code>true</code> if there are differences, <code>false</code> otherwise
     * @throws UndefinedItemPathException if an entry's key refers to a data that is not defined by the model
     * @throws BadItemTypeException if the type defined by the model of one of the entry's key doesn't match the corresponding value
     */
    protected boolean _hasDifferencesInReplaceMode(ViewItemAccessor viewItemAccessor, SynchronizableRepeater repeaterValues, SynchronizationContext context) throws UndefinedItemPathException, BadItemTypeException
    {
        List<Integer> positions = repeaterValues.getReplacePositions();
        List<Map<String, Object>> entriesValues = repeaterValues.getEntries();
        
        for (int i = 0; i < positions.size(); i++)
        {
            int position = positions.get(i);
            ModelAwareRepeaterEntry repeaterEntry = getEntry(position);
            if (repeaterEntry.hasDifferences(viewItemAccessor, entriesValues.get(i), context))
            {
                return true;
            }
        }
        
        // No differences has been found in entries
        if (__LOGGER.isDebugEnabled())
        {
            String viewItemPath = viewItemAccessor instanceof ModelViewItem modelViewItem 
                                        ? ViewHelper.getModelViewItemPath(modelViewItem)
                                        : StringUtils.EMPTY;
            __LOGGER.debug("#hasDifferences[{}] no difference detected.", viewItemPath);
        }
        return false;
    }
    
    /**
     * Check if there are differences between the given values and the repeater's entries if {@link SynchronizableRepeater#getMode()} is REPLACE_ALL
     * @param viewItemAccessor The {@link ViewItemAccessor} for all items to check
     * @param repeaterValues the values of the repeater to check
     * @param context the context of the synchronization
     * @return <code>true</code> if there are differences, <code>false</code> otherwise
     * @throws UndefinedItemPathException if an entry's key refers to a data that is not defined by the model
     * @throws BadItemTypeException if the type defined by the model of one of the entry's key doesn't match the corresponding value
     */
    protected boolean _hasDifferencesInReplaceAllMode(ViewItemAccessor viewItemAccessor, SynchronizableRepeater repeaterValues, SynchronizationContext context) throws UndefinedItemPathException, BadItemTypeException
    {
        List<Map<String, Object>> entriesValues = repeaterValues.getEntries();
        if (hasToMoveEntries(repeaterValues.getPositionsMapping(), entriesValues.size()))
        {
            if (__LOGGER.isDebugEnabled())
            {
                String viewItemPath = viewItemAccessor instanceof ModelViewItem modelViewItem 
                                            ? ViewHelper.getModelViewItemPath(modelViewItem)
                                            : StringUtils.EMPTY;
                __LOGGER.debug("#hasDifferences[{}] differences detected: some entries will be moved", viewItemPath);
            }
            return true;
        }
        
        for (ModelAwareRepeaterEntry repeaterEntry : getEntries())
        {
            int entryIndex = repeaterEntry.getPosition() - 1;
            Map<String, Object> entryValues = entriesValues.get(entryIndex);
            if (repeaterEntry.hasDifferences(viewItemAccessor, entryValues, context))
            {
                return true;
            }
        }
        
        // No differences has been found in entries
        if (__LOGGER.isDebugEnabled())
        {
            String viewItemPath = viewItemAccessor instanceof ModelViewItem modelViewItem 
                                        ? ViewHelper.getModelViewItemPath(modelViewItem)
                                        : StringUtils.EMPTY;
            __LOGGER.debug("#hasDifferences[{}] no difference detected.", viewItemPath);
        }
        return false;
    }
    
    public boolean hasToMoveEntries(Map<Integer, Integer> positionsMapping, int targetSize)
    {
        int initialSize = getSize();

        if (targetSize != initialSize)
        {
            return true;
        }
        
        for (Map.Entry<Integer, Integer> mapping : positionsMapping.entrySet())
        {
            if (!mapping.getKey().equals(mapping.getValue()))
            {
                return true;
            }
        }
        
        return false;
    }
    
    public RepositoryData getRepositoryData()
    {
        return _repositoryData;
    }
    
    public IndexableDataHolder getParentDataHolder()
    {
        return _parent;
    }
    
    public IndexableDataHolder getRootDataHolder()
    {
        return _root;
    }
}
