/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder.group.impl;

import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import org.ametys.cms.data.holder.IndexableDataHolder;
import org.ametys.cms.data.holder.group.IndexableRepeater;
import org.ametys.cms.data.holder.group.IndexableRepeaterEntry;
import org.ametys.cms.data.holder.impl.DefaultModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.runtime.model.ModelItemGroup;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.type.DataContext;

/**
 * Class for model aware repeater entries
 */
public class DefaultModelAwareRepeaterEntry implements IndexableRepeaterEntry
{
    /** The repository data of the repeater entry */
    protected RepositoryData _repositoryData;
    
    /** the default implementation of a {@link ModelAwareDataHolder} to use */
    protected IndexableDataHolder _defaultDataHolder;
    
    /** the repeater containing this entry */
    protected IndexableRepeater _repeater;

    /** the definition of this repeater entry */
    protected ModelItemGroup _definition;
    
    /**
     * Creates a model aware repeater entry
     * @param repositoryData the repository data of the repeater entry
     * @param repeater the repeater containing this entry
     * @param definition the definition of the repeater
     */
    public DefaultModelAwareRepeaterEntry(RepositoryData repositoryData, IndexableRepeater repeater, ModelItemGroup definition)
    {
        _defaultDataHolder = new DefaultModelAwareDataHolder(repositoryData, Optional.of(repeater.getParentDataHolder()), Optional.of(repeater.getRootDataHolder()), definition);
        _repositoryData = repositoryData;
        _repeater = repeater;
        _definition = definition;
    }
    
    public int getPosition()
    {
        return Integer.parseInt(_repositoryData.getName());
    }
    
    public Object dataToJSON(String dataPath, DataContext context)
    {
        return _dataToJSON(Optional.of(dataPath), Optional.empty(), context, false);
    }
    
    public Map<String, Object> dataToJSON(ViewItemAccessor viewItemAccessor, DataContext context) throws BadItemTypeException
    {
        return _dataToJSON(Optional.empty(), Optional.of(viewItemAccessor), context, false);
    }
    
    public Map<String, Object> dataToJSONForEdition(ViewItemAccessor viewItemAccessor, DataContext context) throws BadItemTypeException
    {
        return _dataToJSON(Optional.empty(), Optional.of(viewItemAccessor), context, true);
    }
    
    private Map<String, Object> _dataToJSON(Optional<String> dataPath, Optional<ViewItemAccessor> viewItemAccessor, DataContext context, boolean isEdition) throws BadItemTypeException
    {
        Map<String, Object> entryValues = new HashMap<>();
        entryValues.put("position", getPosition());
        
        if (dataPath.isPresent())
        {
            entryValues.put("value", IndexableRepeaterEntry.super.dataToJSON(dataPath.get(), context));
        }
        else if (viewItemAccessor.isPresent())
        {
            if (isEdition)
            {
                entryValues.put("values", IndexableRepeaterEntry.super.dataToJSONForEdition(viewItemAccessor.get(), context));
            }
            else
            {
                entryValues.put("values", IndexableRepeaterEntry.super.dataToJSON(viewItemAccessor.get(), context));
            }
        }
        
        return entryValues;
    }
    
    @Override
    public IndexableDataHolder getDefaultDataHolder()
    {
        return _defaultDataHolder;
    }
    
    @Override
    public RepositoryData getRepositoryData()
    {
        return _repositoryData;
    }
    
    public IndexableRepeater getHoldingRepeater()
    {
        return _repeater;
    }
}
