/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder.group.impl;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.ametys.cms.data.holder.ModifiableIndexableDataHolder;
import org.ametys.cms.data.holder.group.ModifiableIndexableRepeater;
import org.ametys.cms.data.holder.group.ModifiableIndexableRepeaterEntry;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.holder.DataHolder;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeaterEntry;
import org.ametys.plugins.repository.data.holder.group.Repeater;
import org.ametys.plugins.repository.data.holder.values.SynchronizableRepeater;
import org.ametys.plugins.repository.data.holder.values.SynchronizationContext;
import org.ametys.plugins.repository.data.holder.values.SynchronizationResult;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.UndefinedItemPathException;

/**
 * CLass for modifiable model aware repeaters
 */
public class DefaultModifiableModelAwareRepeater extends DefaultModelAwareRepeater implements ModifiableIndexableRepeater
{
    private static final String __ENTRIES_TEMPORARY_NAME_PREFIX = "temp_";
    
    /** Modifiable repository data to use to store entries in the repository */
    protected ModifiableRepositoryData _modifiableRepositoryData;
    
    /** Parent of the current {@link Repeater} */
    protected ModifiableIndexableDataHolder _modifiableParent;
    
    /** Root {@link DataHolder} */
    protected ModifiableIndexableDataHolder _modifiableRoot;
    
    /**
     * Creates a modifiable model aware repeater
     * @param repositoryData the repository data of the repeater
     * @param parent the parent of the created {@link Repeater}
     * @param root the root {@link DataHolder}
     * @param definition the definition of the repeater
     */
    public DefaultModifiableModelAwareRepeater(ModifiableRepositoryData repositoryData, ModifiableIndexableDataHolder parent, ModifiableIndexableDataHolder root, RepeaterDefinition definition)
    {
        super(repositoryData, parent, root, definition);
        _modifiableRepositoryData = repositoryData;
        _modifiableParent = parent;
        _modifiableRoot = root;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public List<? extends ModifiableIndexableRepeaterEntry> getEntries()
    {
        return (List< ? extends ModifiableIndexableRepeaterEntry>) super.getEntries();
    }
    
    @Override
    public ModifiableIndexableRepeaterEntry getEntry(int position)
    {
        if (1 <= position && position <= getSize())
        {
            ModifiableRepositoryData entryRepositoryData = _modifiableRepositoryData.getRepositoryData(String.valueOf(position));
            return new DefaultModifiableModelAwareRepeaterEntry(entryRepositoryData, this, _definition);
        }
        else if (-getSize() < position && position <= 0)
        {
            // Find the positive equivalent position and call the getEntry method with this position
            return getEntry(getSize() + position);
        }
        else
        {
            return null;
        }
    }
    
    public ModifiableIndexableRepeaterEntry addEntry()
    {
        return addEntry(getSize() + 1);
    }
    
    public ModifiableIndexableRepeaterEntry addEntry(int position) throws IllegalArgumentException
    {
        if (1 <= position && position <= getSize() + 1)
        {
            // rename all entries that will be after the one that will be inserted
            for (int currentEntryPosition = getSize(); currentEntryPosition >= position; currentEntryPosition--)
            {
                ModifiableRepositoryData entryRepositoryData = _modifiableRepositoryData.getRepositoryData(String.valueOf(currentEntryPosition));
                entryRepositoryData.rename(String.valueOf(currentEntryPosition + 1));
            }
            
            // Add the entry at the right position
            ModifiableRepositoryData createdRepositoryData = _modifiableRepositoryData.addRepositoryData(String.valueOf(position), RepositoryConstants.COMPOSITE_NODETYPE);
            return new DefaultModifiableModelAwareRepeaterEntry(createdRepositoryData, this, _definition);
        }
        else if (-getSize() <= position && position <= 0)
        {
            // Find the positive equivalent position and call the addEntry method with this position
            return addEntry(getSize() + position + 1);
        }
        else
        {
            throw new IllegalArgumentException("The repeater named '" + _modifiableRepositoryData.getName() + "' has '" + getSize() + "' entries. You can not create an entry at position '" + position + "'.");
        }
    }
    
    @SuppressWarnings("unchecked")
    public <T extends SynchronizationResult> T synchronizeValues(ViewItemAccessor viewItemAccessor, SynchronizableRepeater repeaterValues, SynchronizationContext context) throws UndefinedItemPathException, BadItemTypeException
    {
        SynchronizationResult result = new SynchronizationResult();
        
        if (SynchronizableRepeater.Mode.APPEND.equals(repeaterValues.getMode()))
        {
            if (!repeaterValues.getRemovedEntries().isEmpty())
            {
                removeEntries(repeaterValues.getRemovedEntries());
                result.setHasChanged(true);
            }
            
            for (Map<String, Object> entryValues : repeaterValues.getEntries())
            {
                ModifiableModelAwareRepeaterEntry entry = addEntry();
                SynchronizationResult entryResult = entry.synchronizeValues(viewItemAccessor, entryValues, context);
                result.aggregateResult(entryResult);
            }
        }
        else if (SynchronizableRepeater.Mode.REPLACE.equals(repeaterValues.getMode()))
        {
            List<Integer> positions = repeaterValues.getReplacePositions();
            List<Map<String, Object>> entriesValues = repeaterValues.getEntries();
            
            for (int i = 0; i < positions.size(); i++)
            {
                int position = positions.get(i);
                ModifiableModelAwareRepeaterEntry repeaterEntry = getEntry(position);
                SynchronizationResult entryResult = repeaterEntry.synchronizeValues(viewItemAccessor, entriesValues.get(i), context);
                result.aggregateResult(entryResult);
            }
        }
        else  // REPLACE_ALL mode
        {
            List<Map<String, Object>> entriesValues = repeaterValues.getEntries();
            boolean moved = moveEntries(repeaterValues.getPositionsMapping(), entriesValues.size());
            result.setHasChanged(moved);
            
            for (ModifiableModelAwareRepeaterEntry repeaterEntry : getEntries())
            {
                int entryIndex = repeaterEntry.getPosition() - 1;
                Map<String, Object> entryValues = entriesValues.get(entryIndex);
                SynchronizationResult entryResult = repeaterEntry.synchronizeValues(viewItemAccessor, entryValues, context);
                result.aggregateResult(entryResult);
            }
        }
        
        return (T) result;
    }
    
    public boolean moveEntries(Map<Integer, Integer> positionsMapping, int targetSize)
    {
        // For each entry, remove it if it does not appear in the mapping, or rename it with a temporary name
        Map<Integer, String> temporaryNamesIndexedByPreviousPosition = new HashMap<>();
        int initialSize = getSize();
        boolean hasChanged = false;
        for (int position = 1; position <= initialSize; position++)
        {
            if (positionsMapping.containsKey(position))
            {
                if (!positionsMapping.get(position).equals(position))
                {
                    // Give a temporary name to the entry
                    String entryNewName = __ENTRIES_TEMPORARY_NAME_PREFIX + position;
                    ModifiableRepositoryData entryData = _modifiableRepositoryData.getRepositoryData(String.valueOf(position));
                    entryData.rename(entryNewName);
                    temporaryNamesIndexedByPreviousPosition.put(position, entryNewName);
                    hasChanged = true;
                }
            }
            else
            {
                // If the entry does not appear in the mapping, remove it
                _modifiableRepositoryData.removeValue(String.valueOf(position));
                hasChanged = true;
            }
        }
        
        // Rename all existent entries with the new position
        for (Integer previousPosition : temporaryNamesIndexedByPreviousPosition.keySet())
        {
            String temporaryName = temporaryNamesIndexedByPreviousPosition.get(previousPosition);
            ModifiableRepositoryData entryData = _modifiableRepositoryData.getRepositoryData(temporaryName);

            Integer newPosition = positionsMapping.get(previousPosition);
            entryData.rename(String.valueOf(newPosition));
        }
        
        // Add the new entries to fill holes until the target size is reached
        for (int position = 1; position <= targetSize; position++)
        {
            if (!_modifiableRepositoryData.hasValue(String.valueOf(position)))
            {
                _modifiableRepositoryData.addRepositoryData(String.valueOf(position), RepositoryConstants.COMPOSITE_NODETYPE);
                hasChanged = true;
            }
        }
        
        return hasChanged;
    }
    
    public void removeEntries(Set<Integer> positions) throws UnknownDataException
    {
        // Create a list to sort the entries positions and remove from the last one to the first one
        Set<Integer> sortedPositions = new TreeSet<>(Collections.reverseOrder());
        for (Integer position : positions)
        {
            if (1 <= position && position <= getSize())
            {
                sortedPositions.add(position);
            }
            else if (-getSize() < position && position <= 0)
            {
                // Find the positive equivalent position
                Integer equivalentPosition = getSize() + position;
                sortedPositions.add(equivalentPosition);
            }
            else
            {
                throw new UnknownDataException("Unable to remove the entry at position '" + position + "' because there is no entry at this position.");
            }
        }
        
        sortedPositions.stream()
                       .forEach(position -> this.removeEntry(position));
    }
    
    public void removeEntry(int position) throws UnknownDataException
    {
        if (1 <= position && position <= getSize())
        {
            // remove the entry
            _modifiableRepositoryData.removeValue(String.valueOf(position));
        
            // rename all entries after the removed one
            for (int currentEntryPosition = position + 1; currentEntryPosition <= getSize() + 1; currentEntryPosition++)
            {
                ModifiableRepositoryData entryRepositoryData = _modifiableRepositoryData.getRepositoryData(String.valueOf(currentEntryPosition));
                entryRepositoryData.rename(String.valueOf(currentEntryPosition - 1));
            }
        }
        else if (-getSize() < position && position <= 0)
        {
            // Find the positive equivalent position and call the removeEntry method with this position
            removeEntry(getSize() + position);
        }
        else
        {
            throw new UnknownDataException("Unable to remove the entry at position '" + position + "' because there is no entry at this position.");
        }
    }

    @Override
    public ModifiableRepositoryData getRepositoryData()
    {
        return _modifiableRepositoryData;
    }
    
    @Override
    public ModifiableIndexableDataHolder getParentDataHolder()
    {
        return _modifiableParent;
    }
    
    @Override
    public ModifiableIndexableDataHolder getRootDataHolder()
    {
        return _modifiableRoot;
    }
}
