/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder.group.impl;

import java.util.Optional;

import org.ametys.cms.data.holder.ModifiableIndexableDataHolder;
import org.ametys.cms.data.holder.group.ModifiableIndexableRepeater;
import org.ametys.cms.data.holder.group.ModifiableIndexableRepeaterEntry;
import org.ametys.cms.data.holder.impl.DefaultModifiableModelAwareDataHolder;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.model.RepeaterDefinition;

/**
 * Class for modifiable model aware repeater entries
 */
public class DefaultModifiableModelAwareRepeaterEntry extends DefaultModelAwareRepeaterEntry implements ModifiableIndexableRepeaterEntry
{
    /** The modifiable repository data if the repeater entry is modifiable */
    protected ModifiableRepositoryData _modifiableRepositoryData;
    
    /** the repeater containing this entry */
    protected ModifiableIndexableRepeater _modifiableRepeater;
    
    /** the default implementation of a {@link ModifiableIndexableDataHolder} to use */
    protected ModifiableIndexableDataHolder _modifiableDefaultDataHolder;
    
    /**
     * Creates a modifiable model aware repeater entry
     * @param repositoryData the repository data of the repeater entry
     * @param repeater the repeater containing this entry
     * @param definition the definition of the repeater
     */
    public DefaultModifiableModelAwareRepeaterEntry(ModifiableRepositoryData repositoryData, ModifiableIndexableRepeater repeater, RepeaterDefinition definition)
    {
        super(repositoryData, repeater, definition);
        _modifiableDefaultDataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, Optional.of(repeater.getParentDataHolder()), Optional.of(repeater.getRootDataHolder()), definition);
        _modifiableRepositoryData = repositoryData;
        _modifiableRepeater = repeater;
    }
    
    public void setPosition(int position) throws IllegalArgumentException
    {
        if (1 <= position && position <= _modifiableRepeater.getSize())
        {
            if (position != getPosition())
            {
                int oldPosition = getPosition();
                _rename(String.valueOf(position) + "_tmp");
                
                // rename entries between the old and new position
                if (position > oldPosition)
                {
                    for (int currentEntryPosition = oldPosition + 1; currentEntryPosition <= position; currentEntryPosition++)
                    {
                        DefaultModifiableModelAwareRepeaterEntry entry = (DefaultModifiableModelAwareRepeaterEntry) _modifiableRepeater.getEntry(currentEntryPosition);
                        entry._rename(String.valueOf(currentEntryPosition - 1));
                    }
                }
                else if (position < oldPosition)
                {
                    for (int currentEntryPosition = oldPosition - 1; currentEntryPosition >= position; currentEntryPosition--)
                    {
                        DefaultModifiableModelAwareRepeaterEntry entry = (DefaultModifiableModelAwareRepeaterEntry) _modifiableRepeater.getEntry(currentEntryPosition);
                        entry._rename(String.valueOf(currentEntryPosition + 1));
                    }
                }
                
                _rename(String.valueOf(position));
            }
        
        }
        else if (-_modifiableRepeater.getSize() < position && position <= 0)
        {
            // Find the positive equivalent position and call the setPosition method with this position
            setPosition(_modifiableRepeater.getSize() + position);
        }
        else
        {
            throw new IllegalArgumentException("The repeater named '" + _modifiableRepositoryData.getName() + "' has '" + _modifiableRepeater.getSize() + "' entries. You can not set an entry at position '" + position + "'.");
        }
    }
    
    private void _rename(String newName)
    {
        _modifiableRepositoryData.rename(newName);
    }
    
    @Override
    public ModifiableIndexableDataHolder getDefaultDataHolder()
    {
        return _modifiableDefaultDataHolder;
    }
    
    @Override
    public ModifiableRepositoryData getRepositoryData()
    {
        return _modifiableRepositoryData;
    }

    @Override
    public ModifiableIndexableRepeater getHoldingRepeater()
    {
        return _modifiableRepeater;
    }
}
