/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.solr.common.SolrInputDocument;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.data.ametysobject.ModelAwareDataAwareAmetysObject;
import org.ametys.cms.data.holder.IndexableDataHolder;
import org.ametys.cms.data.holder.group.IndexableComposite;
import org.ametys.cms.data.holder.group.IndexableRepeater;
import org.ametys.cms.data.type.indexing.IndexableDataContext;
import org.ametys.cms.data.type.indexing.IndexableElementType;
import org.ametys.cms.model.properties.Property;
import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.core.util.JSONUtils;
import org.ametys.plugins.repository.data.external.ExternalizableDataProvider.ExternalizableDataStatus;
import org.ametys.plugins.repository.data.holder.DataHolder;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeaterEntry;
import org.ametys.plugins.repository.data.holder.values.SynchronizableValue;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.plugins.repository.model.CompositeDefinition;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.plugins.repository.model.RepositoryDataContext;
import org.ametys.plugins.repository.model.ViewHelper;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ElementType;
import org.ametys.runtime.model.type.ModelItemType;

/**
 * Helper for implementations of indexable data holder
 */
public final class IndexableDataHolderHelper implements Component, Serviceable, Disposable
{
    private static JSONUtils _jsonUtils;
    private static SystemPropertyExtensionPoint _contentSystemPropertyExtentionPoint;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _jsonUtils = (JSONUtils) manager.lookup(JSONUtils.ROLE);
        _contentSystemPropertyExtentionPoint = (SystemPropertyExtensionPoint) manager.lookup(SystemPropertyExtensionPoint.ROLE);
    }
    
    public void dispose()
    {
        _jsonUtils = null;
        _contentSystemPropertyExtentionPoint = null;
    }
    
    /**
     * Indexes data of the given {@link IndexableDataHolder}
     * @param dataHolder the {@link IndexableDataHolder} to index
     * @param viewItemAccessor the view item accessor to explore
     * @param document the solr document representing this {@link IndexableDataHolder}
     * @param rootDocument the solr document of the root object.
     * @param solrFieldPrefix the prefix of the solr field
     * @param context The context of the data to index
     * @return additional solr documents that may have been created (ex: repeater entries)
     * @throws BadItemTypeException if the saxed value's type does not matches the stored data
     */
    public static List<SolrInputDocument> indexData(IndexableDataHolder dataHolder, ViewItemAccessor viewItemAccessor, SolrInputDocument document, SolrInputDocument rootDocument, String solrFieldPrefix, IndexableDataContext context) throws BadItemTypeException
    {
        List<SolrInputDocument> additionalDocuments = new ArrayList<>();
        
        for (ViewItem viewItem : viewItemAccessor.getViewItems())
        {
            additionalDocuments.addAll(_indexDataFromViewItem(dataHolder, viewItem, document, rootDocument, solrFieldPrefix, context));
        }
        
        return additionalDocuments;
    }

    @SuppressWarnings("unchecked")
    private static List<SolrInputDocument> _indexDataFromViewItem(IndexableDataHolder dataHolder, ViewItem viewItem, SolrInputDocument document, SolrInputDocument rootDocument, String solrFieldPrefix, IndexableDataContext context)
    {
        if (viewItem instanceof ModelViewItem modelViewItem)
        {
            ModelItem modelItem = modelViewItem.getDefinition();
            String dataName = modelItem.getName();
            IndexableDataContext newContext = context.cloneContext()
                    .addSegmentToDataPath(dataName)
                    .withModelItem(modelItem)
                    .withViewItem(viewItem);

            if (modelItem instanceof Property property)
            {
                property.indexValue(document, getAmetysObjectFromContext(newContext), newContext);
                return Collections.EMPTY_LIST;
            }
            else if (dataHolder.hasValue(dataName))
            {
                if (modelItem instanceof ElementDefinition definition)
                {
                    ElementType type = definition.getType();
                    
                    if (type instanceof IndexableElementType indexingType)
                    {
                        Object value = dataHolder.getValue(dataName);
                        
                        String solrFieldName = solrFieldPrefix + dataName;
                        indexingType.indexValue(document, rootDocument, solrFieldName, value, newContext);
                    }
                    
                    return Collections.EMPTY_LIST;
                }
                else if (modelItem instanceof CompositeDefinition definition)
                {
                    IndexableComposite composite = dataHolder.getValue(dataName);
                    String newSolrFieldPrefix = solrFieldPrefix + dataName + ModelItem.ITEM_PATH_SEPARATOR;

                    return indexData(composite, (ViewItemAccessor) viewItem, document, rootDocument, newSolrFieldPrefix, newContext);
                }
                else if (modelItem instanceof RepeaterDefinition definition)
                {
                    IndexableRepeater repeater = dataHolder.getValue(dataName);
                    return repeater.indexData(document, rootDocument, solrFieldPrefix, newContext);
                }
            }
        }
        else if (viewItem instanceof ViewItemAccessor accessor)
        {
            return indexData(dataHolder, accessor, document, rootDocument, solrFieldPrefix, context);
        }
        return Collections.EMPTY_LIST;
    }
    
    /**
     * Generates SAX events for the data in the given view in the given {@link DataHolder}
     * @param dataHolder the {@link ModelAwareDataHolder} to SAX
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param viewItemAccessor the {@link ViewItemAccessor} referencing the items for which generate SAX events
     * @param context The context of the data to SAX
     * @param isEdition <code>true</code> if SAX events are generated in edition mode, <code>false</code> otherwise
     * @throws SAXException if an error occurs during the SAX events generation
     * @throws BadItemTypeException if the saxed value's type does not matches the stored data
     * @throws UndefinedItemPathException if an item in the view is not part of the model
     */
    public static void dataToSAX(ModelAwareDataHolder dataHolder, ContentHandler contentHandler, ViewItemAccessor viewItemAccessor, DataContext context, boolean isEdition) throws SAXException, BadItemTypeException, UndefinedItemPathException
    {
        for (ViewItem viewItem : viewItemAccessor.getViewItems())
        {
            if (viewItem instanceof ModelViewItem modelViewItem)
            {
                ModelItem modelItem = modelViewItem.getDefinition();
                String dataName = modelItem.getName();
                DataContext newContext = context.cloneContext()
                                                .addSegmentToDataPath(dataName)
                                                .withViewItem(viewItem)
                                                .withModelItem(modelItem);
                
                if (modelItem instanceof Property property)
                {
                    _propertyToSAX(property, contentHandler, newContext, isEdition);
                }
                else
                {
                    ModelItemType type = modelItem.getType();
    
                    if (isEdition)
                    {
                        if (newContext instanceof RepositoryDataContext repositoryDataContext && repositoryDataContext.isDataExternalizable())
                        {
                            _saxExternalizableValuesAsJson(dataHolder, contentHandler, dataName, newContext);
                        }
                        else
                        {
                            if (_hasValue(dataHolder, dataName, newContext))
                            {
                                Object value = dataHolder.getValue(dataName);
                                type.valueToSAXForEdition(contentHandler, modelItem.getName(), value, newContext);
                            }
                        }
                    }
                    else
                    {
                        if (_hasValue(dataHolder, dataName, newContext))
                        {
                            Object value = dataHolder.getValue(dataName);
                            type.valueToSAX(contentHandler, modelItem.getName(), value, newContext);
                        }
                    }
                }
            }
            else if (viewItem instanceof ViewItemAccessor accessor)
            {
                dataToSAX(dataHolder, contentHandler, accessor, context, isEdition);
            }
        }
    }
    
    @SuppressWarnings("unchecked")
    private static void _propertyToSAX(Property property, ContentHandler contentHandler, DataContext context, boolean isEdition) throws SAXException
    {
        ModelAwareDataAwareAmetysObject ametysObject = getAmetysObjectFromContext(context);
        if (isEdition)
        {
            throw new SAXException("Unable to generate SAX events for property '" + context.getDataPath() + "', for object '" + ametysObject + "' in edition mode: properties are not modifiables.");
        }
        else
        {
            property.valueToSAX(contentHandler, ametysObject, context);
        }
    }
    
    private static void _saxExternalizableValuesAsJson (ModelAwareDataHolder dataHolder, ContentHandler contentHandler, String dataName, DataContext context) throws SAXException
    {
        Map<String, Object> values = externalizableValuesAsJson(dataHolder, dataName, context);
        
        if (!values.isEmpty())
        {
            String jsonString = _jsonUtils.convertObjectToJson(values);
    
            AttributesImpl attrs = new AttributesImpl();
            attrs.addCDATAAttribute("json", "true");
            XMLUtils.createElement(contentHandler, dataName, attrs, jsonString);
        }
    }
    
    /**
     * Convert the data in the given view of the given {@link DataHolder}
     * @param dataHolder the {@link ModelAwareDataHolder} containing the data to convert
     * @param viewItemAccessor the {@link ViewItemAccessor} referencing the items to convert
     * @param context The context of the data to convert
     * @param isEdition <code>true</code> to convert in edition mode, <code>false</code> otherwise
     * @return The data of the given view as JSON
     * @throws BadItemTypeException if the value's type does not matches the stored data
     * @throws UndefinedItemPathException if an item in the view is not part of the model
     */
    public static Map<String, Object> dataToJSON(ModelAwareDataHolder dataHolder, ViewItemAccessor viewItemAccessor, DataContext context, boolean isEdition) throws BadItemTypeException, UndefinedItemPathException
    {
        Map<String, Object> result = new HashMap<>();
        
        for (ViewItem viewItem : viewItemAccessor.getViewItems())
        {
            if (viewItem instanceof ModelViewItem modelViewItem)
            {
                ModelItem modelItem = modelViewItem.getDefinition();
                String dataName = modelItem.getName();
                DataContext newContext = context.cloneContext()
                                                .addSegmentToDataPath(dataName)
                                                .withViewItem(viewItem)
                                                .withModelItem(modelItem);
                
                if (modelItem instanceof Property property)
                {
                    Object json = _propertyToJSON(property, newContext, isEdition);
                    result.put(dataName, json);
                }
                else
                {
                    ModelItemType type = modelItem.getType();
    
                    if (isEdition)
                    {
                        if (newContext instanceof RepositoryDataContext repositoryDataContext && repositoryDataContext.isDataExternalizable())
                        {
                            Map<String, Object> json = externalizableValuesAsJson(dataHolder, dataName, newContext);
                            if (!json.isEmpty())
                            {
                                result.put(dataName, json);
                            }
                        }
                        else
                        {
                            if (_hasValue(dataHolder, dataName, newContext))
                            {
                                Object value = dataHolder.getValue(dataName);
                                Object json = type.valueToJSONForEdition(value, newContext);
                                result.put(dataName, json);
                            }
                        }
                    }
                    else
                    {
                        if (_hasValue(dataHolder, dataName, newContext))
                        {
                            Object value = dataHolder.getValue(dataName);
                            Object json = type.valueToJSONForClient(value, newContext);
                            result.put(dataName, json);
                        }
                    }
                }
            }
            else if (viewItem instanceof ViewItemAccessor accessor)
            {
                result.putAll(dataToJSON(dataHolder, accessor, context, isEdition));
            }
        }
        
        return result;
    }
    
    @SuppressWarnings("unchecked")
    private static Object _propertyToJSON(Property property, DataContext context, boolean isEdition) throws UndefinedItemPathException
    {
        ModelAwareDataAwareAmetysObject ametysObject = getAmetysObjectFromContext(context);
        if (isEdition)
        {
            throw new UndefinedItemPathException("Unable to convert property '" + context.getDataPath() + "', for object '" + ametysObject + "' in edition mode: properties are not modifiables.");
        }
        else
        {
            return property.valueToJSON(ametysObject, context);
        }
    }
    
    /**
     * Convert the externalizable data with the given name
     * @param dataHolder the {@link ModelAwareDataHolder} containing the data to convert
     * @param dataName the name of the data to convert
     * @param context The context of the data to convert
     * @return The data with the given name as JSON
     */
    public static Map<String, Object> externalizableValuesAsJson(ModelAwareDataHolder dataHolder, String dataName, DataContext context)
    {
        Map<String, Object> result = new LinkedHashMap<>();

        RepositoryElementType type = dataHolder.getType(dataName);
        if (dataHolder.hasLocalValue(dataName)
            || context.renderEmptyValues() && dataHolder.hasLocalValueOrEmpty(dataName))
        {
            Object localValue = dataHolder.getLocalValue(dataName);
            Object localValueAsJSON = type.externalizableValueToJSON(localValue, context);
            result.put("local", localValueAsJSON);
        }

        if (dataHolder.hasExternalValue(dataName)
            || context.renderEmptyValues() && dataHolder.hasExternalValueOrEmpty(dataName))
        {
            Object externalValue = dataHolder.getExternalValue(dataName);
            Object externalValueAsJSON = type.externalizableValueToJSON(externalValue, context);
            result.put("external", externalValueAsJSON);
        }
        
        if (!result.isEmpty())
        {
            ExternalizableDataStatus status = dataHolder.getStatus(dataName);
            result.put("status", status.name().toLowerCase());
        }
        
        return result;
    }
    
    /**
     * Returns all data of the given DataHolder as a typed-values Map.
     * @param dataHolder the DataHolder to export
     * @param viewItemAccessor the {@link ViewItemAccessor} referencing the items to include in the resulting Map
     * @param context The context of the data
     * @return a Map containing all data.
     */
    @SuppressWarnings("unchecked")
    public static Map<String, Object> dataToMap(ModelAwareDataHolder dataHolder, ViewItemAccessor viewItemAccessor, DataContext context)
    {
        Map<String, Object> result = new HashMap<>();
        
        ViewHelper.visitView(viewItemAccessor, 
            (element, definition) -> {
                // simple element
                String name = definition.getName();
                DataContext newContext = context.cloneContext().addSegmentToDataPath(name);
                
                if (definition instanceof Property property)
                {
                    ModelAwareDataAwareAmetysObject ametysObject = getAmetysObjectFromContext(newContext);
                    Object value = property.getValue(ametysObject);
                    if (value != null) // there is no empty values for properties, so a null value means an inexistent data
                    {
                        result.put(name, value);
                    }
                }
                else
                {
                    if (newContext instanceof RepositoryDataContext repositoryDataContext && repositoryDataContext.isDataExternalizable())
                    {
                        if (_hasExternalizableValue(dataHolder, name, newContext))
                        {
                            SynchronizableValue value = new SynchronizableValue(dataHolder.getLocalValue(name));
                            value.setExternalValue(dataHolder.getExternalValue(name));
                            value.setExternalizableStatus(dataHolder.getStatus(name));
                            result.put(name, value);
                        }
                    }
                    else if (_hasValue(dataHolder, name, newContext))
                    {
                        Object value = dataHolder.getValue(name);
                        result.put(name, value);
                    }
                }
            }, 
            (group, definition) -> {
                // composite
                String name = definition.getName();
                DataContext newContext = context.cloneContext().addSegmentToDataPath(name);
                if (_hasValue(dataHolder, name, newContext))
                {
                    ModelAwareComposite value = dataHolder.getValue(name);
                    result.put(name, value == null ? null : value.dataToMap(group, newContext));
                }
            }, 
            (group, definition) -> {
                // repeater
                String name = definition.getName();
                if (_hasValue(dataHolder, name, context))
                {
                    ModelAwareRepeater repeater = dataHolder.getValue(name);
                    List<Map<String, Object>> entries = null;
                    if (repeater != null)
                    {
                        entries = new ArrayList<>();
                        for (ModelAwareRepeaterEntry entry : repeater.getEntries())
                        {
                            DataContext newContext = context.cloneContext().addSegmentToDataPath(name + "[" + entry.getPosition() + "]");
                            entries.add(entry.dataToMap(group, newContext));
                        }
                    }
                    result.put(name, entries);
                }
            }, 
            group -> result.putAll(dataToMap(dataHolder, group, context)));
        
        return result;
    }
    
    /**
     * Retrieves the {@link ModelAwareDataAwareAmetysObject} from the given {@link DataContext}
     * @param context the context containing the ametys object
     * @return the ametys object, or <code>null</code> if there is no object id in the context
     */
    public static ModelAwareDataAwareAmetysObject getAmetysObjectFromContext(DataContext context)
    {
        RepositoryDataContext repoContext = context instanceof RepositoryDataContext rc ? rc : RepositoryDataContext.newInstance(context);
        
        return repoContext.getObject()
                          .filter(ModelAwareDataAwareAmetysObject.class::isInstance)
                          .map(ModelAwareDataAwareAmetysObject.class::cast)
                          .orElse(null);
    }
    
    private static boolean _hasValue(ModelAwareDataHolder dataHolder, String dataName, DataContext context)
    {
        return dataHolder.hasValue(dataName)
               || context.renderEmptyValues() && dataHolder.hasValueOrEmpty(dataName);
    }
    
    private static boolean _hasExternalizableValue(ModelAwareDataHolder dataHolder, String dataName, DataContext context)
    {
        return dataHolder.hasLocalValue(dataName)
                || dataHolder.hasExternalValue(dataName)
                || context.renderEmptyValues()
                    && (dataHolder.hasLocalValueOrEmpty(dataName) || dataHolder.hasExternalValueOrEmpty(dataName));
    }
    
    /**
     * Retrieves the {@link SystemPropertyExtensionPoint} for content system properties
     * @return the {@link SystemPropertyExtensionPoint} for content system properties
     */
    public static SystemPropertyExtensionPoint getContentSystemPropertyExtensionPoint()
    {
        return _contentSystemPropertyExtentionPoint;
    }
}
