/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type;

import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.Constants;
import org.apache.cocoon.environment.Context;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Triple;
import org.w3c.dom.Element;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.data.Binary;
import org.ametys.core.model.type.AbstractElementType;
import org.ametys.core.model.type.ModelItemTypeHelper;
import org.ametys.core.upload.Upload;
import org.ametys.core.upload.UploadManager;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.plugins.repository.data.holder.values.UntouchedValue;
import org.ametys.runtime.model.compare.DataChangeType;
import org.ametys.runtime.model.compare.DataChangeTypeDetail;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.type.DataContext;

/**
 * Abstract class for binary type of elements
 */
public abstract class AbstractBinaryElementType extends AbstractElementType<Binary> implements Configurable, Contextualizable
{
    /** Constant for non-modified binary during edition */
    public static final String UNTOUCHED = "untouched";
    
    /** Type of the binary */
    protected String _binaryType;
    
    /** {@link UploadManager} */
    protected UploadManager _uploadManager;
    
    /** {@link CurrentUserProvider} */
    protected CurrentUserProvider _userProvider;
    
    /** Coocon's context */
    protected Context _context;
    
    public void contextualize(org.apache.avalon.framework.context.Context context) throws ContextException
    {
        _context = (Context) context.get(Constants.CONTEXT_ENVIRONMENT_CONTEXT);
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _uploadManager = (UploadManager) manager.lookup(UploadManager.ROLE);
        _userProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
    }
    
    @Override
    public Binary convertValue(Object value) throws BadItemTypeException
    {
        if (value instanceof String)
        {
            return ResourceElementTypeHelper.convertStringToBinary((String) value);
        }
        else if (value instanceof byte[])
        {
            return ResourceElementTypeHelper.convertByteArrayToBinary((byte[]) value);
        }

        return super.convertValue(value);
    }

    public void configure(Configuration configuration) throws ConfigurationException
    {
        Configuration typeNode = configuration.getChild("type", false);
        if (typeNode != null)
        {
            _binaryType = typeNode.getValue(null);
        }
        
        if (StringUtils.isBlank(_binaryType))
        {
            throw new ConfigurationException("The configuration of binary type should contain a node 'binary-type' with the type of the binary ('metadata', 'explorer', ...)", configuration);
        }
    }

    @Override
    public String toString(Binary value)
    {
        throw new UnsupportedOperationException("Unable to convert a binary to a string value");
    }
    
    @Override
    public Object fromJSONForClient(Object json, DataContext context) throws BadItemTypeException
    {
        if (json == null)
        {
            return null;
        }
        else if (json instanceof List jsonList)
        {
            return jsonList.stream()
                           .map(singleJSONValue -> fromJSONForClient(singleJSONValue, context))
                           .toArray(__ -> new Object[jsonList.size()]);
        }
        else if (_isEnumeratedJSONObject(json))
        {
            Object value = ((Map) json).get("value");
            return _singleBinaryFromJSON(value);
        }
        else
        {
            return _singleBinaryFromJSON(json);
        }
    }
    
    private Object _singleBinaryFromJSON(Object json) throws BadItemTypeException
    {
        if (json instanceof Map jsonMap)
        {
            if (jsonMap.isEmpty())
            {
                return null;
            }
            
            String id = (String) jsonMap.get("id");
            if (id.equals(UNTOUCHED))
            {
                return new UntouchedValue();
            }
            
            Upload upload = _uploadManager.getUpload(_userProvider.getUser(), id);
            return ResourceElementTypeHelper.binaryFromUpload(upload);
        }
        else
        {
            throw new BadItemTypeException("Unable to convert the given json value '" + json + "' into a " + getManagedClass().getName());
        }
    }

    @Override
    public boolean isCompatible(Object value)
    {
        return super.isCompatible(value) || value instanceof UntouchedValue;
    }
    
    @Override
    protected Object _singleTypedValueToJSON(Binary value, DataContext context)
    {
        return ResourceElementTypeHelper.singleBinaryToJSON(value, _binaryType, context);
    }
    
    @Override
    protected Binary _singleValueFromXML(Element element, Optional<Object> additionalData)
    {
        Binary binary = new Binary();
        ResourceElementTypeHelper.resourceFromXML(binary, element, additionalData, _context);
        return binary;
    }
    
    @Override
    protected void _singleTypedNotEnumeratedValueToSAX(ContentHandler contentHandler, String tagName, Binary binary, DataContext context, AttributesImpl attributes) throws SAXException
    {
        ResourceElementTypeHelper.singleBinaryToSAX(contentHandler, tagName, binary, _binaryType, context, attributes);
    }
    
    @Override
    protected Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> _compareMultipleValues(Binary[] value1, Binary[] value2)
    {
        throw new UnsupportedOperationException("Unable to compare multiple values of type '" + getId() + "'");
    }
    
    @Override
    protected Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> _compareSingleValues(Binary value1, Binary value2)
    {
        if (ModelItemTypeHelper.areSingleObjectsBothNotNullAndDifferents(value1, value2))
        {
            return ResourceElementTypeHelper.compareSingleBinaries(value1, value2);
        }
        else
        {
            return Stream.of(ModelItemTypeHelper.compareSingleObjects(value1, value2, StringUtils.EMPTY))
                    .filter(Optional::isPresent)
                    .map(Optional::get);
        }
    }

    public boolean isSimple()
    {
        return false;
    }
    
    @Override
    protected boolean _useJSONForEdition()
    {
        return true;
    }
}
