/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.w3c.dom.Element;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.holder.impl.IndexableDataHolderHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.core.model.type.AbstractElementType;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.model.RepositoryDataContext;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.type.DataContext;

/**
 * Abstract class for content type of elements
 */
public abstract class AbstractContentElementType extends AbstractElementType<ContentValue>
{
    /** Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    
    /** Helper for content types */
    protected ContentTypesHelper _contentTypesHelper;
    
    /** The content helper. */
    protected ContentHelper _contentHelper;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
        _contentHelper = (ContentHelper) manager.lookup(ContentHelper.ROLE);
    }

    @Override
    public ContentValue convertValue(Object value) throws BadItemTypeException
    {
        if (value == null)
        {
            return null;
        }
        if (value instanceof ContentValue)
        {
            return (ContentValue) value;
        }
        else if (value instanceof String)
        {
            return new ContentValue(_resolver, (String) value);
        }
        else if (value instanceof ModifiableContent)
        {
            return new ContentValue((ModifiableContent) value);
        }
        
        return super.convertValue(value);
    }
    
    @Override
    public String toString(ContentValue value)
    {
        return value.getContentId();
    }

    @Override
    protected Object _singleValueToJSON(Object value, DataContext context) throws BadItemTypeException
    {
        if (value instanceof ContentValue singleContentValue)
        {
            return _content2JsonForClient(singleContentValue, context);
        }
        else if (value instanceof String singleString)
        {
            return _content2JsonForClient(singleString, context);
        }
        else if (value instanceof Content singleContent)
        {
            return _content2JsonForClient(singleContent, context);
        }
        else
        {
            throw new BadItemTypeException("Try to convert the non " + getId() + " JSON object '" + value + "' into a " + getManagedClass().getName());
        }
    }
    
    private Object _content2JsonForClient(ContentValue contentValue, DataContext context)
    {
        return _content2JsonForClient(contentValue.getContentId(), context);
    }

    private Object _content2JsonForClient(String contentId, DataContext context)
    {
        try
        {
            Content content = _resolver.resolveById(contentId);
            return _content2JsonForClient(content, context);
        }
        catch (AmetysRepositoryException e)
        {
            if (getLogger().isWarnEnabled())
            {
                getLogger().warn(_getLogForNonExistingContent(contentId, "Unable to convert it as JSON.", context), e);
            }
            return null;
        }
    }
    
    private Object _content2JsonForClient(Content content, DataContext context)
    {
        Map<String, Object> contentProperties = new HashMap<>();

        contentProperties.put("id", content.getId());
        contentProperties.put("title", content.getTitle(context.getLocale()));
        contentProperties.put("iconGlyph", StringUtils.defaultString(_contentTypesHelper.getIconGlyph(content)));
        contentProperties.put("iconDecorator", StringUtils.defaultString(_contentTypesHelper.getIconDecorator(content)));
        contentProperties.put("smallIcon", StringUtils.defaultString(_contentTypesHelper.getSmallIcon(content)));
        contentProperties.put("mediumIcon", StringUtils.defaultString(_contentTypesHelper.getMediumIcon(content)));
        contentProperties.put("largeIcon", StringUtils.defaultString(_contentTypesHelper.getLargeIcon(content)));
        contentProperties.put("contentTypes", content.getTypes());
        contentProperties.put("mixins", content.getMixinTypes());
        contentProperties.put("isSimple", _contentHelper.isSimple(content));
        
        if (context.getViewItem().isPresent())
        {
            ViewItem contentViewItem = context.getViewItem().get();
            if (contentViewItem instanceof ViewItemAccessor contentViewItemAccessor)
            {
                RepositoryDataContext newContext = RepositoryDataContext.newInstance(context)
                                                                        .withObject(content)
                                                                        .withDataPath(StringUtils.EMPTY);
                
                Map<String, Object> values = IndexableDataHolderHelper.dataToJSON(content, contentViewItemAccessor, newContext, false);
                contentProperties.putAll(values);
            }
        }
        
        return contentProperties;
    }
    
    @Override
    protected Object _singleValueToJSONForEdition(Object value, DataContext context) throws BadItemTypeException
    {
        if (value instanceof ContentValue singleContentValue)
        {
            return singleContentValue.getContentId();
        }
        else if (value instanceof String singleString)
        {
            return singleString;
        }
        else if (value instanceof Content singleContent)
        {
            return singleContent.getId();
        }
        else
        {
            throw new BadItemTypeException("Try to convert the non " + getId() + " JSON object '" + value + "' into a " + getManagedClass().getName());
        }
    }
    
    @Override
    protected ContentValue _singleValueFromXML(Element element, Optional<Object> additionalData)
    {
        String contentId = element.getAttribute("id");
        if (StringUtils.isNotEmpty(contentId))
        {
            return new ContentValue(_resolver, contentId);
        }
        
        return null;
    }
    
    @Override
    protected void _singleNotEnumeratedValueToSAX(ContentHandler contentHandler, String tagName, Object value, DataContext context, AttributesImpl attributes)
            throws SAXException, BadItemTypeException
    {
        if (value instanceof String)
        {
            _singleContentAsStringToSAX(contentHandler, tagName, (String) value, attributes, context);
        }
        else if (value instanceof Content)
        {
            _singleContentToSAX(contentHandler, tagName, (Content) value, attributes, context);
        }
        else if (value instanceof ContentValue)
        {
            _singleContentValueToSAX(contentHandler, tagName, (ContentValue) value, attributes, context);
        }
    }
    
    private void _singleContentValueToSAX(ContentHandler contentHandler, String tagName, ContentValue contentValue, AttributesImpl attributes, DataContext context) throws SAXException
    {
        _singleContentAsStringToSAX(contentHandler, tagName, contentValue.getContentId(), attributes, context);
    }
    
    private void _singleContentAsStringToSAX(ContentHandler contentHandler, String tagName, String contentId, AttributesImpl attributes, DataContext context) throws SAXException
    {
        try
        {
            Content content = _resolver.resolveById(contentId);
            _singleContentToSAX(contentHandler, tagName, content, attributes, context);
        }
        catch (AmetysRepositoryException e)
        {
            if (getLogger().isWarnEnabled())
            {
                String error = String.format("Unable to generate SAX event in the tag '%s'.", tagName);
                getLogger().warn(_getLogForNonExistingContent(contentId, error, context), e);
            }
        }
    }
    
    private String _getLogForNonExistingContent(String contentId, String error, DataContext context)
    {
        List<String> sentences = new ArrayList<>();
        
        sentences.add(String.format("The content with id '%s' does not exist.", contentId));
        sentences.add(error);
        
        RepositoryDataContext.newInstance(context)
                             .getObjectId()
                             .ifPresent(id -> sentences.add(String.format("This content is referenced by the object with id '%s' on data at path '%s'.", id, context.getDataPath())));
        
        return StringUtils.join(sentences, " ");
    }
    
    private void _singleContentToSAX(ContentHandler contentHandler, String tagName, Content content, AttributesImpl attributes, DataContext context) throws SAXException
    {
        AttributesImpl localAttributes = new AttributesImpl(attributes);
        localAttributes.addCDATAAttribute("id", content.getId());
        localAttributes.addCDATAAttribute("name", content.getName());
        localAttributes.addCDATAAttribute("title", content.getTitle(context.getLocale()));
        if (content.getLanguage() != null)
        {
            localAttributes.addCDATAAttribute("language", content.getLanguage());
        }
        localAttributes.addCDATAAttribute("createdAt", DateUtils.zonedDateTimeToString(content.getCreationDate()));
        localAttributes.addCDATAAttribute("creator", content.getCreator().getLogin());
        localAttributes.addCDATAAttribute("lastModifiedAt", DateUtils.zonedDateTimeToString(content.getLastModified()));
        
        XMLUtils.startElement(contentHandler, tagName, localAttributes);
        
        if (context.getViewItem().isPresent())
        {
            ViewItem contentViewItem = context.getViewItem().get();
            if (contentViewItem instanceof ViewItemAccessor contentViewItemAccessor)
            {
                RepositoryDataContext newContext = RepositoryDataContext.newInstance(context)
                                                                        .withObject(content)
                                                                        .withDataPath(StringUtils.EMPTY);
                
                IndexableDataHolderHelper.dataToSAX(content, contentHandler, contentViewItemAccessor, newContext, false);
            }
        }
        
        XMLUtils.endElement(contentHandler, tagName);
    }
    
    @Override
    public boolean isCompatible(Object value)
    {
        return super.isCompatible(value)
            || value instanceof String || value instanceof String[]
            || value instanceof Content || value instanceof Content[];
    }
    
    public boolean isSimple()
    {
        return true;
    }
    
    @Override
    protected boolean _useJSONForEdition()
    {
        return true;
    }
    
    @Override
    protected boolean _isMultiple(Object value)
    {
        return super._isMultiple(value) || value instanceof String[] || value instanceof Content[];
    }
    
    @Override
    protected boolean _isSingle(Object value)
    {
        return super._isSingle(value) || value instanceof String || value instanceof Content;
    }
}
