/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type;

import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.Constants;
import org.apache.cocoon.environment.Context;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.ImmutableTriple;
import org.apache.commons.lang3.tuple.Triple;
import org.w3c.dom.Element;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.data.Binary;
import org.ametys.cms.data.ExplorerFile;
import org.ametys.cms.data.File;
import org.ametys.cms.repository.comment.CommentsDAO;
import org.ametys.core.model.type.AbstractElementType;
import org.ametys.core.model.type.ModelItemTypeHelper;
import org.ametys.core.upload.Upload;
import org.ametys.core.upload.UploadManager;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.plugins.repository.data.holder.values.UntouchedValue;
import org.ametys.runtime.model.compare.DataChangeType;
import org.ametys.runtime.model.compare.DataChangeTypeDetail;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.type.DataContext;

/**
 * Abstract class for file type of elements
 */
public abstract class AbstractFileElementType extends AbstractElementType<File> implements Configurable, Contextualizable
{
    /** Type of he file from explorer */
    protected static final String __EXPLORER_FILE_TYPE = "explorer";
    
    /** Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    
    /** Type of the binary */
    protected String _binaryType;
    
    /** {@link UploadManager} */
    protected UploadManager _uploadManager;
    
    /** {@link CurrentUserProvider} */
    protected CurrentUserProvider _userProvider;
    
    /** Coocon's context */
    protected Context _context;
    
    public void contextualize(org.apache.avalon.framework.context.Context context) throws ContextException
    {
        _context = (Context) context.get(Constants.CONTEXT_ENVIRONMENT_CONTEXT);
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _uploadManager = (UploadManager) manager.lookup(UploadManager.ROLE);
        _userProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
    }
    
    public void configure(Configuration configuration) throws ConfigurationException
    {
        Configuration typeNode = configuration.getChild("binary-type", false);
        if (typeNode != null)
        {
            _binaryType = typeNode.getValue(null);
        }
        
        if (StringUtils.isBlank(_binaryType))
        {
            throw new ConfigurationException("The configuration of binary type should contain a node 'binary-type' with the type of the binary ('attribute', 'explorer', ...)", configuration);
        }
    }
    
    @Override
    public File convertValue(Object value)
    {
        if (value instanceof String)
        {
            String strValue = (String) value;
            if (_isAmetysObject(strValue))
            {
                return new ExplorerFile(_resolver, strValue);
            }
            else
            {
                return ResourceElementTypeHelper.convertStringToBinary(strValue);
            }
        }
        else if (value instanceof byte[])
        {
            return ResourceElementTypeHelper.convertByteArrayToBinary((byte[]) value);
        }
        
        return super.convertValue(value);
    }
    
    private boolean _isAmetysObject (String value)
    {
        try
        {
            return _resolver.hasAmetysObjectForId(value);
        }
        catch (AmetysRepositoryException e)
        {
            return false;
        }
    }

    @Override
    public String toString(File value)
    {
        if (value instanceof ExplorerFile)
        {
            return ((ExplorerFile) value).getResourceId();
        }
        else
        {
            throw new UnsupportedOperationException("Unable to convert a binary to a string value");
        }
    }
    
    @Override
    public Object fromJSONForClient(Object json, DataContext context) throws BadItemTypeException
    {
        if (json == null)
        {
            return null;
        }
        else if (json instanceof List jsonList)
        {
            return jsonList.stream()
                           .map(singleJSONValue -> fromJSONForClient(singleJSONValue, context))
                           .toArray(__ -> new Object[jsonList.size()]);
        }
        else if (_isEnumeratedJSONObject(json))
        {
            Object value = ((Map) json).get("value");
            return _singleFileFromJSON(value);
        }
        else
        {
            return _singleFileFromJSON(json);
        }
    }
    
    private Object _singleFileFromJSON(Object json)
    {
        if (json == null)
        {
            return json;
        }
        else if (json instanceof String)
        {
            return convertValue(json);
        }
        else if (json instanceof Map)
        {
            Map data = (Map) json;
            if (data.isEmpty())
            {
                return null;
            }
            
            String type = (String) ((Map) json).get("type");
            String id = (String) ((Map) json).get("id");
            if ("explorer".equals(type))
            {
                return new ExplorerFile(_resolver, id);
            }
            else if ("attribute".equals(type))
            {
                if (id.equals(AbstractBinaryElementType.UNTOUCHED))
                {
                    return new UntouchedValue();
                }
                
                Upload upload = _uploadManager.getUpload(_userProvider.getUser(), id);
                return ResourceElementTypeHelper.binaryFromUpload(upload);
            }
        }

        throw new BadItemTypeException("Unable to convert the given json value '" + json + "' into a " + getManagedClass().getName());
    }
    
    @Override
    protected Object _singleValueToJSONForEdition(Object value, DataContext context) throws BadItemTypeException
    {
        return _singleValueToJSON(value, context);
    }
    
    @Override
    protected Object _singleValueToJSON(Object value, DataContext context) throws BadItemTypeException
    {
        if (value == null)
        {
            return value;
        }
        else if (value instanceof String)
        {
            ExplorerFile file = new ExplorerFile(_resolver, (String) value);
            return _singleExplorerFileToJSON(file);
        }
        else if (value instanceof ExplorerFile)
        {
            return _singleExplorerFileToJSON((ExplorerFile) value);
        }
        else if (value instanceof Binary)
        {
            return ResourceElementTypeHelper.singleBinaryToJSON((Binary) value, _binaryType, context);
        }
        else
        {
            throw new BadItemTypeException("Try to convert the non " + getId() + " JSON object '" + value + "' into a " + getManagedClass().getName());
        }
    }
    
    @Override
    public boolean isCompatible(Object value)
    {
        return super.isCompatible(value) || value instanceof UntouchedValue;
    }
    
    private Object _singleExplorerFileToJSON(ExplorerFile explorerFile)
    {
        String explorerFileId = explorerFile.getResourceId();
        try
        {
            Map<String, Object> singleFileToJSON = ResourceElementTypeHelper.singleFileToJSON(explorerFile, __EXPLORER_FILE_TYPE, explorerFileId, explorerFileId);
            singleFileToJSON.put("id", explorerFileId);
            
            return singleFileToJSON;
        }
        catch (UnknownAmetysObjectException e)
        {
            getLogger().warn("The resource of id '{}' does not exist", explorerFileId, e);
            return null;
        }
    }
    
    @Override
    protected File _singleValueFromXML(Element element, Optional<Object> additionalData)
    {
        String type = element.getAttribute("type");
        String value = element.getTextContent().trim();

        if (__EXPLORER_FILE_TYPE.equals(type))
        {
            String resourceId = element.getAttribute("path");
            return new ExplorerFile(_resolver, resourceId);
        }
        else if (StringUtils.isEmpty(type) && !CommentsDAO.URL_VALIDATOR.matcher(value).matches())
        {
            return new ExplorerFile(_resolver, value);
        }
        else
        {
            Binary binary = new Binary();
            ResourceElementTypeHelper.resourceFromXML(binary, element, additionalData, _context);
            return binary;
        }
    }
    
    @Override
    protected void _singleNotEnumeratedValueToSAX(ContentHandler contentHandler, String tagName, Object value, DataContext context, AttributesImpl attributes) throws SAXException, BadItemTypeException
    {
        AttributesImpl localAttributes = new AttributesImpl(attributes);
        
        if (value instanceof String resourceId)
        {
            ExplorerFile file = new ExplorerFile(_resolver, resourceId);
            _singleExplorerFileToSAX(contentHandler, tagName, file, attributes);
        }
        else if (value instanceof ExplorerFile file)
        {
            _singleExplorerFileToSAX(contentHandler, tagName, file, attributes);
        }
        else if (value instanceof Binary binary)
        {
            ResourceElementTypeHelper.singleBinaryToSAX(contentHandler, tagName, (Binary) value, _binaryType, context, localAttributes);
        }
    }
    
    private void _singleExplorerFileToSAX(ContentHandler contentHandler, String tagName, ExplorerFile explorerFile, AttributesImpl attributes) throws SAXException
    {
        String explorerFileId = explorerFile.getResourceId();

        try
        {
            
            AttributesImpl localAttributes = new AttributesImpl(attributes);
            localAttributes.addCDATAAttribute("id", explorerFileId);
            
            ResourceElementTypeHelper.singleFileToSAX(contentHandler, tagName, explorerFile, __EXPLORER_FILE_TYPE, explorerFileId, explorerFileId, localAttributes);
        }
        catch (UnknownAmetysObjectException e)
        {
            getLogger().warn("The resource of id '{}' does not exist", explorerFileId, e);
        }
    }
    
    @Override
    protected Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> _compareMultipleValues(File[] value1, File[] value2)
    {
        throw new UnsupportedOperationException("Unable to compare multiple values of type '" + getId() + "'");
    }
    
    @Override
    protected Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> _compareSingleValues(File value1, File value2)
    {
        if (ModelItemTypeHelper.areSingleObjectsBothNotNullAndDifferents(value1, value2))
        {
            if (_havValuesSameFileType(value1, value2))
            {
                if (value1 instanceof Binary)
                {
                    return ResourceElementTypeHelper.compareSingleBinaries((Binary) value1, (Binary) value2);
                }
                else if (value1 instanceof ExplorerFile)
                {
                    return Stream.of(ModelItemTypeHelper.compareSingleObjects(((ExplorerFile) value1).getResourceId(), ((ExplorerFile) value2).getResourceId(), StringUtils.EMPTY))
                                 .filter(Optional::isPresent)
                                 .map(Optional::get);
                }
                else
                {
                    return super._compareSingleValues(value1, value2);
                }
            }
            else
            {
                return Stream.of(new ImmutableTriple<>(DataChangeType.MODIFIED, DataChangeTypeDetail.TYPE, StringUtils.EMPTY));
            }
        }
        else
        {
            return Stream.of(ModelItemTypeHelper.compareSingleObjects(value1, value2, StringUtils.EMPTY))
                         .filter(Optional::isPresent)
                         .map(Optional::get);
        }
    }
    
    private boolean _havValuesSameFileType(File value1, File value2)
    {
        return value1 instanceof ExplorerFile && value2 instanceof ExplorerFile
               || value1 instanceof Binary && value2 instanceof Binary;
    }

    public boolean isSimple()
    {
        return false;
    }
    
    @Override
    protected boolean _useJSONForEdition()
    {
        return true;
    }
    
    @Override
    protected boolean _isMultiple(Object value)
    {
        return super._isMultiple(value) || value instanceof String[];
    }
    
    @Override
    protected boolean _isSingle(Object value)
    {
        return super._isSingle(value) || value instanceof String;
    }
}
