/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type;

import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Triple;
import org.w3c.dom.Element;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.data.Geocode;
import org.ametys.core.model.type.AbstractElementType;
import org.ametys.core.model.type.ModelItemTypeHelper;
import org.ametys.runtime.model.compare.DataChangeType;
import org.ametys.runtime.model.compare.DataChangeTypeDetail;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.type.DataContext;

/**
 * Abstract class for geocode type of elements
 */
public abstract class AbstractGeocodeElementType extends AbstractElementType<Geocode>
{
    /** The separator between the latitude and the longitude for the string representation of a geocode */
    protected static final String __SEPARATOR = ",";
    
    @Override
    public Geocode convertValue(Object value) throws BadItemTypeException
    {
        if (value instanceof String)
        {
            String strValue = (String) value;
            if (StringUtils.isEmpty(strValue))
            {
                return null;
            }
            
            if (StringUtils.countMatches(strValue, __SEPARATOR) == 1)
            {
                try
                {
                    Double latitude = Double.valueOf(StringUtils.substringBeforeLast(strValue, __SEPARATOR));
                    Double longitude = Double.valueOf(StringUtils.substringAfterLast(strValue, __SEPARATOR));
                    return new Geocode(latitude, longitude);
                }
                catch (NumberFormatException e)
                {
                    throw new BadItemTypeException("Unable to cast '" + value + "' to a geocode", e);
                }
            }
        }
        
        return super.convertValue(value);
    }
    
    @Override
    public String toString(Geocode value)
    {
        if (value != null)
        {
            return String.valueOf(value.getLatitude()) + __SEPARATOR + String.valueOf(value.getLongitude());
        }
        else
        {
            return null;
        }
    }
    
    @Override
    protected Geocode _singleValueFromJSON(Object json, DataContext context) throws BadItemTypeException
    {
        if (json == null)
        {
            return null;
        }
        else if (json instanceof Map jsonMap)
        {
            return Optional.ofNullable(jsonMap)
                    .filter(geocode -> !geocode.isEmpty()
                         && geocode.get(Geocode.LATITUDE_IDENTIFIER) != null && geocode.get(Geocode.LATITUDE_IDENTIFIER) instanceof Number
                         && geocode.get(Geocode.LONGITUDE_IDENTIFIER) != null && geocode.get(Geocode.LONGITUDE_IDENTIFIER) instanceof Number)
                    .map(geocode -> new Geocode(((Number) geocode.get(Geocode.LATITUDE_IDENTIFIER)).doubleValue(), ((Number) geocode.get(Geocode.LONGITUDE_IDENTIFIER)).doubleValue()))
                    .orElse(null);
        }
        else
        {
            throw new BadItemTypeException("Unable to convert the given json value '" + json + "' into a " + getManagedClass().getName());
        }
    }
    
    @Override
    protected Object _singleTypedValueToJSON(Geocode value, DataContext context)
    {
        Map<String, Object> geocodeInfos = new HashMap<>();
        
        geocodeInfos.put(Geocode.LATITUDE_IDENTIFIER, value.getLatitude());
        geocodeInfos.put(Geocode.LONGITUDE_IDENTIFIER, value.getLongitude());
        
        return geocodeInfos;
    }
    
    @Override
    protected Geocode _singleValueFromXML(Element element, Optional<Object> additionalData)
    {
        Optional<Double> latitude = _coordinateFromXML(element, Geocode.LATITUDE_IDENTIFIER);
        Optional<Double> longitude = _coordinateFromXML(element, Geocode.LONGITUDE_IDENTIFIER);
        
        if (latitude.isPresent() && longitude.isPresent())
        {
            return new Geocode(latitude.get(), longitude.get());
        }
        else
        {
            return null;
        }
    }
    
    private Optional<Double> _coordinateFromXML(Element element, String coordinateIdentifier)
    {
        String coordinateAsString = element.getAttribute(coordinateIdentifier);
        return Optional.ofNullable(coordinateAsString)
                .filter(StringUtils::isNotEmpty)
                .map(Double::valueOf);
    }

    @Override
    protected void _singleTypedNotEnumeratedValueToSAX(ContentHandler contentHandler, String tagName, Geocode geocode, DataContext context, AttributesImpl attributes) throws SAXException
    {
        AttributesImpl localAttributes = new AttributesImpl(attributes);
        localAttributes.addCDATAAttribute(Geocode.LATITUDE_IDENTIFIER, geocode.getLatitude().toString());
        localAttributes.addCDATAAttribute(Geocode.LONGITUDE_IDENTIFIER, geocode.getLongitude().toString());
        XMLUtils.createElement(contentHandler, tagName, localAttributes);
    }
    
    @Override
    protected Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> _compareMultipleValues(Geocode[] value1, Geocode[] value2)
    {
        throw new UnsupportedOperationException("Unable to compare multiple values of type '" + getId() + "'");
    }
    
    @Override
    protected Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> _compareSingleValues(Geocode value1, Geocode value2)
    {
        if (ModelItemTypeHelper.areSingleObjectsBothNotNullAndDifferents(value1, value2))
        {
            return Stream.of
                         (
                             ModelItemTypeHelper.compareSingleNumbers(value1.getLatitude(), value2.getLatitude(), Geocode.LATITUDE_IDENTIFIER),
                             ModelItemTypeHelper.compareSingleNumbers(value1.getLongitude(), value2.getLongitude(), Geocode.LONGITUDE_IDENTIFIER)
                         )
                         .filter(Optional::isPresent)
                         .map(Optional::get);
        }
        else
        {
            return Stream.of(ModelItemTypeHelper.compareSingleObjects(value1, value2, StringUtils.EMPTY))
                         .filter(Optional::isPresent)
                         .map(Optional::get);
        }
    }
    
    public boolean isSimple()
    {
        return true;
    }
    
    @Override
    protected boolean _useJSONForEdition()
    {
        return true;
    }
}
