/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type;

import java.util.Map;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Stream;

import org.apache.cocoon.xml.AttributesImpl;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.ImmutableTriple;
import org.apache.commons.lang3.tuple.Triple;
import org.w3c.dom.Element;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.core.model.type.AbstractElementType;
import org.ametys.core.model.type.ModelItemTypeHelper;
import org.ametys.plugins.repository.metadata.MultilingualString;
import org.ametys.plugins.repository.metadata.MultilingualStringHelper;
import org.ametys.runtime.model.compare.DataChangeType;
import org.ametys.runtime.model.compare.DataChangeTypeDetail;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.type.DataContext;

/**
 * Abstract class for multilingual string type of elements
 */
public abstract class AbstractMultilingualStringElementType extends AbstractElementType<MultilingualString>
{
    @Override
    public MultilingualString convertValue(Object value) throws BadItemTypeException
    {
        if (value instanceof String)
        {
            try
            {
                return MultilingualStringHelper.fromString((String) value);
            }
            catch (IllegalArgumentException e)
            {
                throw new BadItemTypeException("Unable to cast '" + value + "' to a multilingual string");
            }
        }
        
        return super.convertValue(value);
    }
    
    @Override
    public String toString(MultilingualString value)
    {
        return MultilingualStringHelper.toString(value);
    }
    
    @SuppressWarnings("unchecked")
    @Override
    protected MultilingualString _singleValueFromJSON(Object json, DataContext context) throws BadItemTypeException
    {
        if (json instanceof Map)
        {
            return MultilingualStringHelper.fromJSON((Map<String, Object>) json);
        }
        else
        {
            throw new BadItemTypeException("Unable to convert the given json value '" + json + "' into a " + getManagedClass().getName());
        }
    }
    
    @Override
    protected Object _singleTypedValueToJSON(MultilingualString value, DataContext context)
    {
        return MultilingualStringHelper.toJson(value);
    }
    
    @Override
    protected MultilingualString _singleValueFromXML(Element element, Optional<Object> additionalData)
    {
        return MultilingualStringHelper.fromXML(element);
    }
    
    @Override
    protected void _singleTypedNotEnumeratedValueToSAX(ContentHandler contentHandler, String tagName, MultilingualString value, DataContext context, AttributesImpl attributes) throws SAXException
    {
        MultilingualStringHelper.sax(contentHandler, tagName, value, attributes, context.getLocale());
    }
    
    @Override
    protected Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> _compareMultipleValues(MultilingualString[] value1, MultilingualString[] value2)
    {
        throw new UnsupportedOperationException("Unable to compare multiple values of type '" + getId() + "'");
    }
    
    @Override
    protected Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> _compareSingleValues(MultilingualString value1, MultilingualString value2)
    {
        if (ModelItemTypeHelper.areSingleObjectsBothNotNullAndDifferents(value1, value2))
        {
            return Stream.of
                    (
                        // Manages modified and removed strings
                        value1.getLocales()
                            .stream()
                            .map
                            (
                                locale -> ModelItemTypeHelper.compareSingleStrings
                                (
                                    value1.getValue(locale),
                                    value2.hasLocale(locale) ? value2.getValue(locale) : StringUtils.EMPTY,
                                    locale.toString()
                                )
                            )
                            .filter(Optional::isPresent)
                            .map(Optional::get),

                        // Manages added strings
                        value2.getLocales()
                            .stream()
                            .filter(locale -> !value1.hasLocale(locale))
                            .map(locale -> new ImmutableTriple<>(DataChangeType.ADDED, DataChangeTypeDetail.NONE, locale.toString()))
                    )
                    .flatMap(Function.identity());
        }
        else
        {
            return Stream.of(ModelItemTypeHelper.compareSingleObjects(value1, value2, StringUtils.EMPTY))
                         .filter(Optional::isPresent)
                         .map(Optional::get);
        }
    }
    
    public boolean isSimple()
    {
        return true;
    }
    
    @Override
    protected boolean _useJSONForEdition()
    {
        return true;
    }
}
