/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type;

import java.util.Map;
import java.util.Optional;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.w3c.dom.Element;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.core.model.type.AbstractElementType;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.type.DataContext;

/**
 * Abstract class for user type of elements
 */
public abstract class AbstractUserElementType extends AbstractElementType<UserIdentity>
{
    /** The user helper */
    protected UserHelper _helper;
    /** The user manager */
    protected UserManager _userManager;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _helper = (UserHelper) manager.lookup(UserHelper.ROLE);
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
    }

    @Override
    public UserIdentity convertValue(Object value)
    {
        if (value instanceof String)
        {
            if (StringUtils.isEmpty((String) value))
            {
                return null;
            }
            
            UserIdentity user = UserIdentity.stringToUserIdentity((String) value);
            if (user != null)
            {
                return user;
            }
        }
        
        return super.convertValue(value);
    }
    
    @Override
    public String toString(UserIdentity value)
    {
        return UserIdentity.userIdentityToString(value);
    }
    
    @SuppressWarnings("unchecked")
    @Override
    protected UserIdentity _singleValueFromJSON(Object json, DataContext context) throws BadItemTypeException
    {
        if (json instanceof Map)
        {
            return _helper.json2userIdentity((Map<String, Object>) json);
        }
        else
        {
            throw new BadItemTypeException("Unable to convert the given json value '" + json + "' into a " + getManagedClass().getName());
        }
    }
    
    @Override
    protected Object _singleTypedValueToJSON(UserIdentity value, DataContext context)
    {
        return _helper.user2json(value, true);
    }
    
    @Override
    protected UserIdentity _singleValueFromXML(Element element, Optional<Object> additionalData)
    {
        String login = element.getAttribute("login");
        String populationId = element.getAttribute("populationId");

        if (StringUtils.isNoneEmpty(login, populationId))
        {
            return new UserIdentity(login, populationId);
        }
        
        return null;
    }
    
    @Override
    protected void _singleTypedNotEnumeratedValueToSAX(ContentHandler contentHandler, String tagName, UserIdentity value, DataContext context, AttributesImpl attributes) throws SAXException
    {
        User user = _userManager.getUser(value);
        if (user != null)
        {
            AttributesImpl localAttributes = new AttributesImpl(attributes);
            localAttributes.addCDATAAttribute("login", value.getLogin());
            localAttributes.addCDATAAttribute("populationId", value.getPopulationId());
            localAttributes.addCDATAAttribute("email", user.getEmail());
            XMLUtils.createElement(contentHandler, tagName, localAttributes, user.getFullName());
        }
    }
    
    public boolean isSimple()
    {
        return true;
    }
    
    @Override
    protected boolean _useJSONForEdition()
    {
        return true;
    }
}
