/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Stream;

import org.apache.cocoon.environment.Context;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.ImmutableTriple;
import org.apache.commons.lang3.tuple.Triple;
import org.apache.tika.Tika;
import org.apache.tika.mime.MimeType;
import org.apache.tika.mime.MimeTypeException;
import org.apache.tika.mime.MimeTypes;
import org.w3c.dom.Element;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.data.Binary;
import org.ametys.cms.data.File;
import org.ametys.cms.data.NamedResource;
import org.ametys.cms.data.Resource;
import org.ametys.cms.data.RichText;
import org.ametys.cms.transformation.xslt.ResolveURIComponent;
import org.ametys.core.model.type.ModelItemTypeHelper;
import org.ametys.core.upload.Upload;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.model.RepositoryDataContext;
import org.ametys.plugins.repository.version.VersionAwareAmetysObject;
import org.ametys.runtime.model.compare.DataChangeType;
import org.ametys.runtime.model.compare.DataChangeTypeDetail;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.type.DataContext;

/**
 * Helper for resource type of elements stored in the repository
 */
public final class ResourceElementTypeHelper
{
    /** Prefix to use for resource's metadata */
    public static final String METADATA_PREFIX = "jcr";
    
    /** Mime type metadata identifier */
    public static final String MIME_TYPE_IDENTIFIER = "mimeType";
    
    /** Encoding metadata identifier */
    public static final String ENCODING_IDENTIFIER = "encoding";
    
    /** Last modification date metadata identifier */
    public static final String LAST_MODIFICATION_DATE_IDENTIFIER = "lastModified";
    
    /** Identifier of the data containing the resource's data */
    public static final String DATA_IDENTIFIER = "data";
    
    /** hash metadata identifier */ 
    public static final String HASH_IDENTIFIER = "hash";
    
    /** file name metadata identifier */
    public static final String FILENAME_IDENTIFIER = "filename";
    
    /** Identifier of the data to check to know if the resource is empty */
    public static final String EMPTY_RESOURCE_IDENTIFIER = "isEmpty";
    
    private ResourceElementTypeHelper()
    {
        // Empty constructor
    }
    
    /**
     * Converts the given string to a {@link Binary}
     * @param value the string to convert
     * @return the converted {@link Binary}
     */
    public static Binary convertStringToBinary(String value)
    {
        try
        {
            byte [] valueAsByteArray = IOUtils.toByteArray(new StringReader(value), StandardCharsets.UTF_8);
            Binary binary = new Binary();
            binary.setInputStream(new ByteArrayInputStream(valueAsByteArray));
            binary.setMimeType("text/plain");
            binary.setLastModificationDate(ZonedDateTime.now());
            binary.setFilename("no-name.txt");

            return binary;
        }
        catch (IOException e)
        {
            throw new IllegalArgumentException("Unable to convert the given value to a Binary", e);
        }
    }
    
    /**
     * Converts the given byte array to a {@link Binary}
     * @param value the byte array to convert
     * @return the converted {@link Binary}
     */
    public static Binary convertByteArrayToBinary(byte[] value)
    {
        try
        {
            ByteArrayInputStream byteArrayInputStreamForTika = new ByteArrayInputStream(value);
            Tika tika = new Tika();
            String mediaTypeAsString = tika.detect(byteArrayInputStreamForTika);
            MimeType mimeType = MimeTypes.getDefaultMimeTypes().forName(mediaTypeAsString);

            Binary binary = new Binary();
            binary.setInputStream(new ByteArrayInputStream(value));
            binary.setMimeType(mediaTypeAsString);
            binary.setLastModificationDate(ZonedDateTime.now());
            binary.setFilename("no-name" + mimeType.getExtension());

            return binary;
        }
        catch (MimeTypeException | IOException e)
        {
            throw new IllegalArgumentException("Unable to convert the given value to a Binary", e);
        }
    }
    
    /**
     * Convert the single binary into a JSON object
     * @param binary the binary to convert
     * @param binaryType the type of the binary
     * @param context The context of the binary to convert
     * @return The file as JSON
     */
    public static Map<String, Object> singleBinaryToJSON(Binary binary, String binaryType, DataContext context)
    {
        return ResourceElementTypeHelper.singleFileToJSON(binary, binaryType, context.getDataPath(), _getBinaryURI(context));
    }

    /**
     * Convert the single file into a JSON object
     * @param file the file to convert
     * @param fileType the type of the file
     * @param path the path of file (ex: data path for a binary or resource id for an explorer file)
     * @param fileURI uri of the file to give to the {@link ResolveURIComponent}
     * @return The file as JSON
     */
    // TODO CMS-11714: use contants for all those json data
    public static Map<String, Object> singleFileToJSON(File file, String fileType, String path, String fileURI)
    {
        Map<String, Object> fileInfos = new LinkedHashMap<>();
        
        String mimeType = file.getMimeType();
        boolean isMimeTypeImage = false;
        if (StringUtils.isNotEmpty(mimeType))
        {
            fileInfos.put("mime-type", mimeType);
            isMimeTypeImage = mimeType.startsWith("image/");
        }

        Optional.ofNullable(file.getLastModificationDate())
                .map(DateUtils::zonedDateTimeToString)
                .ifPresent(lastModificationDate -> fileInfos.put("lastModified", lastModificationDate));

        fileInfos.put("size", String.valueOf(file.getLength()));
        if (path != null)
        {
            fileInfos.put("path", path);
        }
        fileInfos.put("type", fileType);
        Optional.ofNullable(file.getName()).ifPresent(filename -> fileInfos.put("filename", filename));
        
        String viewUrl = isMimeTypeImage
                ? ResolveURIComponent.resolveBoundedImage(fileType, fileURI, 100, 100)
                : ResolveURIComponent.resolve(fileType, fileURI);
        fileInfos.put("viewUrl", viewUrl);

        String downloadUrl = ResolveURIComponent.resolve(fileType, fileURI, true);
        fileInfos.put("downloadUrl", downloadUrl);
        
        return fileInfos;
    }
    
    /**
     * Sets the given resource's data using the given DOM element and additional data 
     * @param resource the resource
     * @param element the DOM element
     * @param additionalData additional data containing the input stream
     * @param context the Cocoon's context, used to retrieve mime types from file names if needed
     */
    public static void resourceFromXML(NamedResource resource, Element element, Optional<Object> additionalData, Context context) 
    {
        String urlValue = element.getTextContent().trim();
        String filename = null;
        if (StringUtils.isNotEmpty(urlValue))
        {
            try
            {
                URL url = new URL(urlValue);
                
                try (InputStream is = url.openStream())
                {
                    resource.setInputStream(is);
                }
                
                String path = url.getPath();
                filename = path.substring(path.lastIndexOf("/") + 1);
            }
            catch (IOException e)
            {
                throw new IllegalArgumentException("Unable to get the resource value from the given XML", e);
            }
        }
        else
        {
            filename = element.getAttribute("filename");

            @SuppressWarnings("unchecked")
            Map<String, InputStream> files = additionalData.isPresent() ? (Map<String, InputStream>) additionalData.get() : Map.of();
            if (files.containsKey(filename))
            {
                try
                {
                    InputStream inputStream = files.get(filename);
                    resource.setInputStream(inputStream);
                }
                catch (IOException e)
                {
                    throw new IllegalArgumentException("Unable to get the resource value from the given XML", e);
                }
            }
        }
        
        resource.setFilename(filename);

        String lastModified = element.getAttribute("lastModified");
        ZonedDateTime time = StringUtils.isEmpty(lastModified) ? ZonedDateTime.now() : ZonedDateTime.parse(lastModified, DateTimeFormatter.ISO_OFFSET_DATE_TIME);
        resource.setLastModificationDate(time);
        
        String mimeType = element.getAttribute("mime-type");
        if (StringUtils.isEmpty(mimeType))
        {
            mimeType = context.getMimeType(filename.toLowerCase());
            
            if (mimeType == null)
            {
                mimeType = "application/unknown";
            }
        }
        resource.setMimeType(mimeType);
    }
    
    /**
     * Converts an uploaded file to a {@link Binary}.
     * @param upload the uploaded file.
     * @return the newly created {@link Binary}.
     */
    public static Binary binaryFromUpload(Upload upload)
    {
        Binary binary = new Binary();
        
        binary.setFilename(upload.getFilename());
        binary.setMimeType(upload.getMimeType());
        binary.setLastModificationDate(upload.getUploadedDate());
        
        try
        {
            binary.setInputStream(upload.getInputStream());
        }
        catch (IOException e)
        {
            throw new RuntimeException("Unable to set binary data", e);
        }
        
        return binary;
    }
    
    /**
     * Generates SAX events for the given single binary
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param tagName the tag name of the SAX event to generate.
     * @param binary the single file to SAX
     * @param binaryType the type of the file
     * @param context The context of the binary to convert
     * @param attributes the attributes for the SAX event to generate
     * @throws SAXException if an error occurs during the SAX events generation
     */
    public static void singleBinaryToSAX(ContentHandler contentHandler, String tagName, Binary binary, String binaryType, DataContext context, AttributesImpl attributes) throws SAXException
    {
        singleFileToSAX(contentHandler, tagName, binary, binaryType, context.getDataPath(), _getBinaryURI(context), attributes);
    }
    
    private static String _getBinaryURI(DataContext context)
    {
        RepositoryDataContext repoContext = context instanceof RepositoryDataContext rc ? rc : RepositoryDataContext.newInstance(context);
        String dataPath = context.getDataPath();
        Optional<? extends AmetysObject> object = repoContext.getObject();
        
        // Add data path
        String uri = dataPath;
        
        // Add object id
        uri += object.map(AmetysObject::getId)
                     .map(id -> "?objectId=" + id)
                     .orElse(StringUtils.EMPTY);
        
        // Add object version.
        uri += object.filter(VersionAwareAmetysObject.class::isInstance)
                     .map(VersionAwareAmetysObject.class::cast)
                     .map(VersionAwareAmetysObject::getRevision)
                     .map(revision -> "&contentVersion=" + revision)  // Version cannot be present if object id is not
                     .orElse(StringUtils.EMPTY);

        return uri;
    }
    
    /**
     * Generates SAX events for the given single file
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param tagName the tag name of the SAX event to generate.
     * @param file the single file to SAX
     * @param fileType the type of the file
     * @param path the path of file (ex: data path for a binary or resource id for an explorer file)
     * @param fileURI uri of the file
     * @param attributes the attributes for the SAX event to generate
     * @throws SAXException if an error occurs during the SAX events generation
     */
    // TODO CMS-11714: use contants for all those data in SAX events
    public static void singleFileToSAX(ContentHandler contentHandler, String tagName, File file, String fileType, String path, String fileURI, AttributesImpl attributes) throws SAXException
    {
        AttributesImpl localAttributes = new AttributesImpl(attributes);
        
        Optional.ofNullable(file.getMimeType())
            .ifPresent(mimeType -> localAttributes.addCDATAAttribute("mime-type", mimeType));
        
        Optional.ofNullable(file.getLastModificationDate())
            .map(DateUtils::zonedDateTimeToString)
            .ifPresent(lastModified -> localAttributes.addCDATAAttribute("lastModified", lastModified));
        
        Optional.ofNullable(file.getName())
            .ifPresent(filename -> localAttributes.addCDATAAttribute("filename", filename));

        localAttributes.addCDATAAttribute("type", fileType);
        localAttributes.addCDATAAttribute("size", String.valueOf(file.getLength()));
        localAttributes.addCDATAAttribute("path", path);
        localAttributes.addCDATAAttribute("uri", fileURI);
        
        XMLUtils.createElement(contentHandler, tagName, localAttributes);
    }
    
    /**
     * Checks if the given resource data is empty. A resource data is considered as empty if it has no stream data
     * @param resourceData the resource data to check
     * @return <code>true</code> if the resource has is empty, <code>false</code> otherwise
     */
    public static boolean isResourceDataEmpty(RepositoryData resourceData)
    {
        if (resourceData.hasValue(EMPTY_RESOURCE_IDENTIFIER, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL))
        {
            return resourceData.getBoolean(EMPTY_RESOURCE_IDENTIFIER, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        }
        else
        {
            return false;
        }
    }
    
    /**
     * Read the binary from the given repository data
     * @param binaryData the repository data containing the binary's data
     * @return the read binary
     */
    public static Binary readBinaryData(RepositoryData binaryData)
    {
        String hash = getStringValue(binaryData, HASH_IDENTIFIER, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        Binary binary = new Binary(binaryData, hash);
        
        readResourceData(binaryData, binary);
        Optional.ofNullable(getStringValue(binaryData, FILENAME_IDENTIFIER, RepositoryConstants.NAMESPACE_PREFIX)).ifPresent(binary::setFilename);
        
        return binary;
    }
    
    /**
     * Read the resource from the given repository data
     * @param resourceData the repository data containing the resource's data
     * @param resource the resource to read
     */
    public static void readResourceData(RepositoryData resourceData, Resource resource)
    {
        Optional.ofNullable(getStringValue(resourceData, MIME_TYPE_IDENTIFIER, METADATA_PREFIX)).ifPresent(resource::setMimeType);
        Optional.ofNullable(getStringValue(resourceData, ENCODING_IDENTIFIER, METADATA_PREFIX)).ifPresent(resource::setEncoding);
        Optional.ofNullable(getDateValue(resourceData, LAST_MODIFICATION_DATE_IDENTIFIER, METADATA_PREFIX)).ifPresent(resource::setLastModificationDate);
    }
    
    /**
     * Empties the resource data with the given name. A resource data is considered as empty if it has no stream data
     * @param parentResourceData the parent of the resource data to empty
     * @param name the name of the resource data
     * @param nodeType the node type of the resource data
     */
    public static void emptyResourceData(ModifiableRepositoryData parentResourceData, String name, String nodeType)
    {
        ModifiableRepositoryData resourceData = parentResourceData.hasValue(name) ? parentResourceData.getRepositoryData(name) : parentResourceData.addRepositoryData(name, nodeType);
        
        try (ByteArrayInputStream is = new ByteArrayInputStream(StringUtils.EMPTY.getBytes()))
        {
            resourceData.setValue(DATA_IDENTIFIER, is, METADATA_PREFIX);
        }
        catch (IOException e)
        {
            throw new AmetysRepositoryException(e);
        }
        
        resourceData.setValue(EMPTY_RESOURCE_IDENTIFIER, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
    }
    
    /**
     * Write the resource in the given repository data
     * @param parentData the repository data where to write the binary
     * @param name the name of the element to write
     * @param value the binary to write
     */
    public static void writeSingleBinaryValue(ModifiableRepositoryData parentData, String name, Binary value)
    {
        ModifiableRepositoryData binaryData;
        if (parentData.hasValue(name))
        {
            binaryData = getRepositoryData(parentData, RepositoryConstants.BINARY_NODETYPE, name, RepositoryConstants.NAMESPACE_PREFIX);
        }
        else
        {
            binaryData = parentData.addRepositoryData(name, RepositoryConstants.BINARY_NODETYPE);
        }
        
        if (value != null)
        {
            binaryData.setValue(EMPTY_RESOURCE_IDENTIFIER, false, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);

            writeResourceData(binaryData, value);
            Optional.ofNullable(value.getFilename()).ifPresent(filename -> binaryData.setValue(FILENAME_IDENTIFIER, filename, RepositoryConstants.NAMESPACE_PREFIX));
            
            if (binaryData.hasValue(HASH_IDENTIFIER, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL))
            {
                String currentHash = getStringValue(binaryData, HASH_IDENTIFIER, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
                if (!currentHash.equals(value.getHash()))
                {
                    Optional.ofNullable(value.getHash()).ifPresent(hash -> binaryData.setValue(HASH_IDENTIFIER, hash, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
                    Optional.ofNullable(value.getInputStream()).ifPresent(stream -> binaryData.setValue(DATA_IDENTIFIER, stream, METADATA_PREFIX));
                }
            }
            else
            {
                Optional.ofNullable(value.getHash()).ifPresent(hash -> binaryData.setValue(HASH_IDENTIFIER, hash, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
                Optional.ofNullable(value.getInputStream()).ifPresent(stream -> binaryData.setValue(DATA_IDENTIFIER, stream, METADATA_PREFIX));
            }
        }
    }

    /**
     * Write the resource in the given repository data
     * @param resourceData the repository data where to write the resource's data
     * @param value the resource to write
     */
    public static void writeResourceData(ModifiableRepositoryData resourceData, Resource value)
    {
        Optional.ofNullable(value.getMimeType()).ifPresent(mimeType -> resourceData.setValue(MIME_TYPE_IDENTIFIER, mimeType, METADATA_PREFIX));
        Optional.ofNullable(value.getEncoding()).ifPresent(encoding -> resourceData.setValue(ENCODING_IDENTIFIER, encoding, METADATA_PREFIX));
        Optional.ofNullable(value.getLastModificationDate()).ifPresent(lastModificationDate -> resourceData.setValue(LAST_MODIFICATION_DATE_IDENTIFIER, _getCalendarFromZonedDateTime(lastModificationDate), METADATA_PREFIX));
    }
    
    /**
     * Compare the given single binaries and retrieves the changes as a stream of {@link Triple}s. The {@link Triple} contains:
     * <ul>
     * <li>the general type of the change (added, modified or removed) as a {@link DataChangeType},</li>
     * <li>some details about this change if possible (after or before for a date, more or less for a number, ...) as a {@link DataChangeTypeDetail}</li>
     * <li>The data concerned by this change if not the element itself (or an empty String)</li>
     * </ul>
     * @param binary1 the 1st single binary
     * @param binary2 the 2nd single binary
     * @return the changes between the two given single binaries as a stream of {@link Triple}s. Retrieves an empty stream if there is no change
     */
    public static Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> compareSingleBinaries(Binary binary1, Binary binary2)
    {
        try
        {
            return Stream.of
                         (
                             // Get changes about the binary's metadata
                             _compareSingleResourcesMetadata(binary1, binary2),
                             
                             // Get changes about the binary's file name
                             ModelItemTypeHelper.compareSingleObjects(binary1.getFilename(), binary2.getFilename(), "fileName")
                                 .map(Stream::of)
                                 .orElseGet(Stream::empty),
                             
                             //  Get changes about the binary's content
                             _compareSingleResourcesContent(binary1, binary2)
                         )
                         .flatMap(Function.identity());
        }
        catch (IOException e)
        {
            throw new IllegalArgumentException("Unable to compare the two given binaries", e);
        }
    }
    
    /**
     * Compare the given single rich texts and retrieves the changes as a stream of {@link Triple}s. The {@link Triple} contains:
     * <ul>
     * <li>the general type of the change (added, modified or removed) as a {@link DataChangeType},</li>
     * <li>some details about this change if possible (after or before for a date, more or less for a number, ...) as a {@link DataChangeTypeDetail}</li>
     * <li>The data concerned by this change if not the element itself (or an empty String)</li>
     * </ul>
     * @param richText1 the 1st single rich text
     * @param richText2 the 2nd single rich text
     * @return the changes between the two given single rich texts as a stream of {@link Triple}s. Retrieves an empty stream if there is no change
     */
    public static Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> compareSingleRichTexts(RichText richText1, RichText richText2)
    {
        try
        {
            // Note: folders are not compared, they are used for pictures and id of the pictures are in the InputStream, so if something changed, the InputStream changed.
            return Stream.of
                         (
                             _compareSingleResourcesMetadata(richText1, richText2),
                             _compareSingleResourcesContent(richText1, richText2)
                         )
                         .flatMap(Function.identity());
        }
        catch (IOException e)
        {
            throw new IllegalArgumentException("Unable to compare the two given rich texts", e);
        }
    }
    
    /**
     * Compare the metadata of the given single resources and retrieves the changes as a stream of {@link Triple}s. The {@link Triple}s contain:
     * <ul>
     * <li>the general type of the change (added, modified or removed) as a {@link DataChangeType},</li>
     * <li>some details about this change if possible (after or before for a date, more or less for a number, ...) as a {@link DataChangeTypeDetail}</li>
     * <li>The data concerned by this change if not the element itself (or an empty String)</li>
     * </ul>
     * @param resource1 the 1st single resource
     * @param resource2 the 2nd single resource
     * @return the changes between the metadata of the two given single resources as a stream of {@link Triple}s. Retrieves an empty stream if there is no change
     * @throws IOException if an error occurs while reading the resources' data
     */
    protected static Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> _compareSingleResourcesMetadata(Resource resource1, Resource resource2) throws IOException
    {
        return Stream.of
                     (
                         ModelItemTypeHelper.compareSingleObjects(resource1.getEncoding(), resource2.getEncoding(), ENCODING_IDENTIFIER),
                         ModelItemTypeHelper.compareSingleObjects(resource1.getMimeType(), resource2.getMimeType(), MIME_TYPE_IDENTIFIER)
                     )
                     .filter(Optional::isPresent)
                     .map(Optional::get);
    }
    
    /**
     * Compare the content of the given single resources and retrieves the changes as a stream of {@link Triple}. The {@link Triple}s contain:
     * <ul>
     * <li>the general type of the change (added, modified or removed) as a {@link DataChangeType},</li>
     * <li>some details about this change if possible (after or before for a date, more or less for a number, ...) as a {@link DataChangeTypeDetail}</li>
     * <li>The data concerned by this change if not the element itself (or an empty String)</li>
     * </ul>
     * @param resource1 the 1st single resource
     * @param resource2 the 2nd single resource
     * @return the changes between the content of the two given single resources as a stream of {@link Triple}s. Retrieves an empty stream if there is no change
     * @throws IOException if an error occurs while reading the resources' data
     */
    protected static Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> _compareSingleResourcesContent(Resource resource1, Resource resource2) throws IOException
    {
        try (
                InputStream is1 = resource1.getInputStream();
                InputStream is2 = resource2.getInputStream();
            )
        {
            if (!IOUtils.contentEquals(is1, is2))
            {
                return Stream.of(new ImmutableTriple<>(DataChangeType.MODIFIED, DataChangeTypeDetail.NONE, "inputStream"));
            }
            else
            {
                return Stream.empty();
            }
        }
    }
    
    /**
     * Retrieves the string value from the given repository data
     * @param repositoryData the repository data containing the data to retrieve
     * @param name the name of the data to retrieve
     * @param prefix the prefix of the data to retrieve
     * @return the string value
     */
    public static String getStringValue(RepositoryData repositoryData, String name, String prefix)
    {
        if (!repositoryData.hasValue(name, prefix))
        {
            return null;
        }
        else if (RepositoryData.STRING_REPOSITORY_DATA_TYPE.equals(repositoryData.getType(name, prefix)))
        {
            return repositoryData.getString(name, prefix);
        }
        else
        {
            throw new BadItemTypeException("Try to get string value from the non string data '" + prefix + ":" + name + "' on '" + repositoryData + "'");
        }
    }
    
    /**
     * Retrieves the string values from the given repository data
     * @param repositoryData the repository data containing the data to retrieve
     * @param name the name of the data to retrieve
     * @param prefix the prefix of the data to retrieve
     * @return the string values
     */
    public static String[] getStringValues(RepositoryData repositoryData, String name, String prefix)
    {
        if (!repositoryData.hasValue(name, prefix))
        {
            return null;
        }
        else if (RepositoryData.STRING_REPOSITORY_DATA_TYPE.equals(repositoryData.getType(name, prefix)))
        {
            return repositoryData.getStrings(name, prefix);
        }
        else
        {
            throw new BadItemTypeException("Try to get string value from the non string data '" + prefix + ":" + name + "' on '" + repositoryData + "'");
        }
    }
    
    /**
     * Retrieves the child repository data from the given repository data
     * @param <T> Type of the repository data (modifiable or not)
     * @param parentData the repository data containing the data to retrieve
     * @param dataTypeName the type of the data to retrieve
     * @param name the name of the data to retrieve
     * @param prefix the prefix of the data to retrieve
     * @return the child repository data
     */
    @SuppressWarnings("unchecked")
    public static <T extends RepositoryData> T getRepositoryData(T parentData, String dataTypeName, String name, String prefix)
    {
        if (!parentData.hasValue(name, prefix))
        {
            return null;
        }
        else if (dataTypeName.equals(parentData.getType(name, prefix)))
        {
            return (T) parentData.getRepositoryData(name, prefix);
        }
        else
        {
            throw new BadItemTypeException("Try to get data of type '" + dataTypeName + "' from the non data '" + prefix + ":" + name + "' on '" + parentData + "'");
        }
    }
    
    /**
     * Retrieves the date value from the given repository data
     * @param repositoryData the repository data containing data to retrieve
     * @param prefix the prefix of the data to retrieve
     * @param name the name of the data to retrieve
     * @return the date value
     */
    public static ZonedDateTime getDateValue(RepositoryData repositoryData, String name, String prefix)
    {
        if (!repositoryData.hasValue(name, prefix))
        {
            return null;
        }
        else if (RepositoryData.CALENDAR_REPOSITORY_DATA_TYPE.equals(repositoryData.getType(name, prefix)))
        {
            return DateUtils.asZonedDateTime(repositoryData.getDate(name, prefix));
        }
        else
        {
            throw new BadItemTypeException("Try to get date value from the non date data '" + prefix + ":" + name + "' on '" + repositoryData + "'");
        }
    }
    
    private static Calendar _getCalendarFromZonedDateTime(ZonedDateTime zonedDateTime)
    {
        ZonedDateTime dateTimeOnUTC = zonedDateTime.withZoneSameInstant(ZoneId.of("UTC"));
        return GregorianCalendar.from(dateTimeOnUTC);
    }
}
