/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.util.Arrays;
import java.util.Optional;

import org.apache.commons.lang3.ArrayUtils;

import org.ametys.cms.data.type.indexing.impl.BooleanIndexableElementType;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.runtime.model.exception.BadItemTypeException;

/**
 * Class for boolean type of elements stored in the repository
 */
public class BooleanRepositoryElementType extends BooleanIndexableElementType implements RepositoryElementType<Boolean>
{

    public Object read(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return null;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to get boolean value from the non boolean data '" + name + "' on '" + parentData + "'");
        }
        
        if (parentData.isMultiple(name))
        {
            return parentData.getBooleans(name);
        }
        else
        {
            return parentData.getBoolean(name);
        }
    }
    
    @Override
    public boolean hasValue(RepositoryData parentData, String name) throws BadItemTypeException
    {
        return RepositoryElementType.super.hasValue(parentData, name) || parentData.hasValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
    }
    
    public boolean hasNonEmptyValue(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return false;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to check boolean value from the non boolean data '" + name + "' on '" + parentData + "'");
        }
        
        if (parentData.isMultiple(name))
        {
            return parentData.getBooleans(name).length > 0;
        }
        else
        {
            return true;
        }
    }

    public void write(ModifiableRepositoryData parentData, String name, Object value) throws BadItemTypeException
    {
        if (value == null)
        {
            if (parentData.hasValue(name) && parentData.isMultiple(name))
            {
                parentData.setValues(name, new Boolean[0]);
            }
            else
            {
                parentData.setValue(name + EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);

                if (parentData.hasValue(name))
                {
                    parentData.removeValue(name);
                }
            }
        }
        else if (value instanceof Boolean)
        {
            parentData.setValue(name, (Boolean) value);
            
            if (parentData.hasValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL))
            {
                parentData.removeValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
            }
        }
        else if (value instanceof boolean[])
        {
            parentData.setValues(name, ArrayUtils.toObject((boolean[]) value));
        }
        else if (value instanceof Boolean[])
        {
            Arrays.stream((Boolean[]) value)
                  .forEach(v -> Optional.ofNullable(v)
                                        .orElseThrow(() -> new IllegalArgumentException("Try to set a null value into the multiple " + getId() + " data '" + name + "' on '" + parentData + "'")));
            
            parentData.setValues(name, (Boolean[]) value);
        }
        else
        {
            throw new BadItemTypeException("Try to set the non boolean value '" + value + "' to the boolean data '" + name + "' on '" + parentData + "'");
        }
    }
    
    public String getRepositoryDataType()
    {
        return RepositoryData.BOOLEAN_REPOSITORY_DATA_TYPE;
    }
}
