/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.util.Collection;

import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.data.holder.impl.IndexableDataHolderHelper;
import org.ametys.core.model.type.AbstractModelItemType;
import org.ametys.core.model.type.ModelItemTypeHelper;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.holder.group.ModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModelLessComposite;
import org.ametys.plugins.repository.data.holder.impl.DataHolderHelper;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryModelItemGroupType;
import org.ametys.plugins.repository.model.CompositeDefinition;
import org.ametys.runtime.model.ModelItemContainer;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.type.DataContext;

/**
 * Class for composite type
 */
public class CompositeRepositoryModelItemType extends AbstractModelItemType implements RepositoryModelItemGroupType
{
    public void valueToSAXForEdition(ContentHandler contentHandler, String tagName, Object value, DataContext context) throws SAXException
    {
        _valueToSAX(contentHandler, tagName, value, context, true);
    }
    
    public void valueToSAX(ContentHandler contentHandler, String tagName, Object value, DataContext context) throws SAXException
    {
        _valueToSAX(contentHandler, tagName, value, context, false);
    }
    
    private void _valueToSAX(ContentHandler contentHandler, String tagName, Object value, DataContext context, boolean isEdition) throws SAXException
    {
        AttributesImpl attributes = ModelItemTypeHelper.getXMLAttributesFromDataContext(context);
        
        if (value == null)
        {
            XMLUtils.createElement(contentHandler, tagName, attributes);
        }
        else if (value instanceof ModelAwareComposite composite)
        {
            if (context.getViewItem().isPresent())
            {
                ViewItem compositeViewItem = context.getViewItem().get();
                assert compositeViewItem instanceof ViewItemAccessor;
                
                XMLUtils.startElement(contentHandler, tagName, attributes);
                IndexableDataHolderHelper.dataToSAX(composite, contentHandler, (ViewItemAccessor) compositeViewItem, context, isEdition);
                XMLUtils.endElement(contentHandler, tagName);
            }
            else
            {
                XMLUtils.startElement(contentHandler, tagName, attributes);
                
                CompositeDefinition compositeDefinition = _getCompositeDefinition(composite);
                ModelViewItemGroup viewItemAccessor = ModelViewItemGroup.of(compositeDefinition);
                IndexableDataHolderHelper.dataToSAX(composite, contentHandler, viewItemAccessor, context, isEdition);
                
                XMLUtils.endElement(contentHandler, tagName);
            }
        }
        else if (value instanceof ModelLessComposite composite)
        {
            XMLUtils.startElement(contentHandler, tagName, attributes);
            DataHolderHelper.dataToSAX(composite, contentHandler, context);
            XMLUtils.endElement(contentHandler, tagName);
        }
        else
        {
            throw new IllegalArgumentException("Try to sax the non composite value '" + value + "' in tag name '" + tagName + "'");
        }
    }
    
    public Object valueToJSONForClient(Object value, DataContext context)
    {
        return _valueToJSON(value, context, false);
    }
    
    public Object valueToJSONForEdition(Object value, DataContext context)
    {
        return _valueToJSON(value, context, true);
    }
    
    private Object _valueToJSON(Object value, DataContext context, boolean isEdition)
    {
        if (value == null)
        {
            return null;
        }
        else if (value instanceof ModelAwareComposite composite)
        {
            if (context.getViewItem().isPresent())
            {
                ViewItem compositeViewItem = context.getViewItem().get();
                assert compositeViewItem instanceof ViewItemAccessor;
                return IndexableDataHolderHelper.dataToJSON(composite, (ViewItemAccessor) compositeViewItem, context, isEdition);
            }
            else
            {
                CompositeDefinition compositeDefinition = _getCompositeDefinition(composite);
                ModelViewItemGroup viewItemAccessor = ModelViewItemGroup.of(compositeDefinition);
                return IndexableDataHolderHelper.dataToJSON(composite, viewItemAccessor, context, isEdition);
            }
        }
        else if (value instanceof ModelLessComposite composite)
        {
            return DataHolderHelper.dataToJSON(composite, context);
        }
        else
        {
            throw new IllegalArgumentException("Try to convert the non composite value '" + value + "'");
        }
    }
    
    private CompositeDefinition _getCompositeDefinition(ModelAwareComposite composite)
    {
        Collection<? extends ModelItemContainer> modelItemContainers = composite.getModel();
        assert modelItemContainers.size() == 1;
        
        ModelItemContainer modelItemContainer = modelItemContainers.iterator().next();
        assert modelItemContainer instanceof CompositeDefinition;
        
        return (CompositeDefinition) modelItemContainer;
    }
    
    public boolean isCompatible(RepositoryData parentData, String name) throws UnknownDataException
    {
        // TODO NEWATTRIBUTEAPI: The second part of the test is here just to be compatible with the old compositMetadata type. When all composites are migrated, remove this part of the test
        return RepositoryModelItemGroupType.super.isCompatible(parentData, name) || RepositoryConstants.COMPOSITE_METADTA_NODETYPE.equals(parentData.getType(name));
    }
    
    public String getRepositoryDataType()
    {
        return RepositoryConstants.COMPOSITE_NODETYPE;
    }
}
