/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import javax.jcr.Session;

import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.type.indexing.impl.ContentIndexableElementType;
import org.ametys.cms.repository.Content;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.runtime.model.exception.BadItemTypeException;

/**
 * Class for content type of elements stored in the repository
 */
public class ContentRepositoryElementType extends ContentIndexableElementType implements RepositoryElementType<ContentValue>
{

    public Object read(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return null;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to get content value from the non content data '" + name + "' on '" + parentData + "'");
        }
        else
        {
            final Session session =  parentData instanceof JCRRepositoryData ? ((JCRRepositoryData) parentData).getSession() : null;
            
            if (parentData.isMultiple(name))
            {
                String[] values = parentData.getStrings(name);
                return Arrays.stream(values)
                             .map(contentId -> new ContentValue(_resolver, contentId, session))
                             .collect(Collectors.toList())
                             .toArray(new ContentValue[values.length]);
            }
            else
            {
                String contentId = parentData.getString(name);
                return StringUtils.isNotEmpty(contentId) ? new ContentValue(_resolver, contentId, session) : null;
            }
        }
    }
    
    public boolean hasNonEmptyValue(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return false;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to check content value from the non content data '" + name + "' on '" + parentData + "'");
        }
        
        if (parentData.isMultiple(name))
        {
            return parentData.getStrings(name).length > 0;
        }
        else
        {
            String contentId = parentData.getString(name);
            return StringUtils.isNotEmpty(contentId);
        }
    }

    public void write(ModifiableRepositoryData parentData, String name, Object value) throws BadItemTypeException
    {
        if (value == null)
        {
            if (parentData.hasValue(name) && parentData.isMultiple(name))
            {
                parentData.setValues(name, new String[0]);
            }
            else
            {
                parentData.setValue(name, StringUtils.EMPTY);
            }
        }
        else if (value instanceof String)
        {
            parentData.setValue(name, (String) value);
        }
        else if (value instanceof String[])
        {
            List<String> values = Arrays.stream((String[]) value)
                  .distinct()
                  .map(v -> Optional.ofNullable(v)
                                        .orElseThrow(() -> new IllegalArgumentException("Try to set a null value into the multiple " + getId() + " data '" + name + "' on '" + parentData + "'")))
                  .collect(Collectors.toList());
            
            parentData.setValues(name, values.toArray(new String[values.size()]));
        }
        else if (value instanceof Content)
        {
            parentData.setValue(name, ((Content) value).getId());
        }
        else if (value instanceof Content[])
        {
            List<String> values = Arrays.stream((Content[]) value)
                                    .distinct()
                                    .map(content -> Optional.ofNullable(content)
                                                            .map(Content::getId)
                                                            .orElseThrow(() -> new IllegalArgumentException("Try to set a null value into the multiple " + getId() + " data '" + name + "' on '" + parentData + "'")))
                                    .collect(Collectors.toList());
            
            parentData.setValues(name, values.toArray(new String[values.size()]));
        }
        else if (value instanceof ContentValue)
        {
            parentData.setValue(name, ((ContentValue) value).getContentId());
        }
        else if (value instanceof ContentValue[])
        {
            List<String> values = Arrays.stream((ContentValue[]) value)
                                     .distinct()
                                     .map(content -> Optional.ofNullable(content)
                                                             .map(ContentValue::getContentId)
                                                             .orElseThrow(() -> new IllegalArgumentException("Try to set a null value into the multiple " + getId() + " data '" + name + "' on '" + parentData + "'")))
                                     .collect(Collectors.toList());
            
            parentData.setValues(name, values.toArray(new String[values.size()]));
        }
        else
        {
            throw new BadItemTypeException("Try to set the non string value '" + value + "' to the string data '" + name + "' on '" + parentData + "'");
        }
    }
    
    public String getRepositoryDataType()
    {
        return RepositoryData.STRING_REPOSITORY_DATA_TYPE;
    }
}
