/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.time.LocalDate;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Optional;

import org.ametys.cms.data.type.indexing.impl.DateIndexableElementType;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.runtime.model.exception.BadItemTypeException;

/**
 * Class for date type of elements stored in the repository
 */
public class DateRepositoryElementType extends DateIndexableElementType implements RepositoryElementType<LocalDate>
{
    public Object read(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return null;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to get date value from the non date data '" + name + "' on '" + parentData + "'");
        }
        
        if (parentData.isMultiple(name))
        {
            return Arrays.stream(parentData.getDates(name))
                         .map(DateUtils::asLocalDate)
                         .toArray(LocalDate[]::new);
        }
        else
        {
            Calendar date = parentData.getDate(name);
            return DateUtils.asLocalDate(date);
        }
    }
    
    @Override
    public boolean hasValue(RepositoryData parentData, String name) throws BadItemTypeException
    {
        return RepositoryElementType.super.hasValue(parentData, name) || parentData.hasValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
    }
    
    public boolean hasNonEmptyValue(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return false;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to check date value from the non date data '" + name + "' on '" + parentData + "'");
        }
        
        if (parentData.isMultiple(name))
        {
            return parentData.getDates(name).length > 0;
        }
        else
        {
            return true;
        }
    }
    
    public void write(ModifiableRepositoryData parentData, String name, Object value) throws BadItemTypeException
    {
        if (value == null)
        {
            if (parentData.hasValue(name) && parentData.isMultiple(name))
            {
                parentData.setValues(name, new Calendar[0]);
            }
            else
            {
                parentData.setValue(name + EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);

                if (parentData.hasValue(name))
                {
                    parentData.removeValue(name);
                }
            }
        }
        else if (value instanceof LocalDate)
        {
            Calendar date = DateUtils.asCalendar((LocalDate) value);
            parentData.setValue(name, date);
            
            if (parentData.hasValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL))
            {
                parentData.removeValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
            }
        }
        else if (value instanceof LocalDate[])
        {
            Calendar[] calendars = Arrays.stream((LocalDate[]) value)
                                         .map(v -> Optional.ofNullable(v)
                                                           .orElseThrow(() -> new IllegalArgumentException("Try to set a null value into the multiple " + getId() + " data '" + name + "' on '" + parentData + "'")))
                                         .map(localDate -> DateUtils.asCalendar(localDate))
                                         .toArray(Calendar[]::new);
            
            parentData.setValues(name, calendars);
        }
        else
        {
            throw new BadItemTypeException("Try to set the non date value '" + value + "' to the date data '" + name + "' on '" + parentData + "'");
        }
    }
    
    public boolean isCompatible(RepositoryData parentData, String name) throws UnknownDataException
    {
        if (parentData.hasValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL))
        {
            return true;
        }
        
        if (!parentData.hasValue(name))
        {
            throw new UnknownDataException("No data found for '" + name + "' in repository data named '" + parentData.getName() + "'.");
        }
        
        if (getRepositoryDataType().equals(parentData.getType(name)))
        {
            if (parentData.isMultiple(name))
            {
                Calendar[] calendars = parentData.getDates(name);
                for (Calendar calendar : calendars)
                {
                    if (!_isDate(calendar))
                    {
                        // If there is at least one datetime, it could not be a data of type date
                        return false;
                    }
                }
                
                // No date is a datetime, so it could be a data of type date
                return true;
            }
            else
            {
                return _isDate(parentData.getDate(name));
            }
        }
        else
        {
            return false;
        }
    }
    
    private boolean _isDate(Calendar calendar)
    {
        return calendar.get(Calendar.HOUR_OF_DAY) == 0
            && calendar.get(Calendar.MINUTE) == 0
            && calendar.get(Calendar.SECOND) == 0
            && calendar.get(Calendar.MILLISECOND) == 0;
    }
    
    public String getRepositoryDataType()
    {
        return RepositoryData.CALENDAR_REPOSITORY_DATA_TYPE;
    }
}
