/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.time.ZonedDateTime;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.Optional;
import java.util.TimeZone;

import org.ametys.cms.data.type.indexing.impl.DateTimeIndexableElementType;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.runtime.model.exception.BadItemTypeException;

/**
 * Class for date time type of elements stored in the repository
 */
public class DateTimeRepositoryElementType extends DateTimeIndexableElementType implements RepositoryElementType<ZonedDateTime>
{
    /**
     * {@inheritDoc}
     * The DateTime type retrieves a single or an array of {@link ZonedDateTime}.
     * To convert a {@link ZonedDateTime} in {@link Date} and keep the {@link TimeZone} information, use the {@link ZonedDateTime#toEpochSecond()} method. 
     */
    public Object read(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return null;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to get date time value from the non date time data '" + name + "' on '" + parentData + "'");
        }
        
        if (parentData.isMultiple(name))
        {
            return Arrays.stream(parentData.getDates(name))
                         .map(DateUtils::asZonedDateTime)
                         .toArray(ZonedDateTime[]::new);
        }
        else
        {
            Calendar date = parentData.getDate(name);
            return DateUtils.asZonedDateTime(date);
        }
    }
    
    @Override
    public boolean hasValue(RepositoryData parentData, String name) throws BadItemTypeException
    {
        return RepositoryElementType.super.hasValue(parentData, name) || parentData.hasValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
    }
    
    public boolean hasNonEmptyValue(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return false;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to check date time value from the non date time data '" + name + "' on '" + parentData + "'");
        }
        
        if (parentData.isMultiple(name))
        {
            return parentData.getDates(name).length > 0;
        }
        else
        {
            return true;
        }
    }
    
    public void write(ModifiableRepositoryData parentData, String name, Object value) throws BadItemTypeException
    {
        if (value == null)
        {
            if (parentData.hasValue(name) && parentData.isMultiple(name))
            {
                parentData.setValues(name, new Calendar[0]);
            }
            else
            {
                parentData.setValue(name + EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);

                if (parentData.hasValue(name))
                {
                    parentData.removeValue(name);
                }
            }
        }
        else if (value instanceof ZonedDateTime)
        {
            Calendar date = DateUtils.asCalendar((ZonedDateTime) value);
            parentData.setValue(name, date);
            
            if (parentData.hasValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL))
            {
                parentData.removeValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
            }
        }
        else if (value instanceof ZonedDateTime[])
        {
            Calendar[] calendars = Arrays.stream((ZonedDateTime[]) value)
                                         .map(v -> Optional.ofNullable(v)
                                                           .orElseThrow(() -> new IllegalArgumentException("Try to set a null value into the multiple " + getId() + " data '" + name + "' on '" + parentData + "'")))
                                         .map(zonedDateTime -> DateUtils.asCalendar(zonedDateTime))
                                         .toArray(Calendar[]::new);

            parentData.setValues(name, calendars);
        }
        else
        {
            throw new BadItemTypeException("Try to set the non date time value '" + value + "' to the date time data '" + name + "' on '" + parentData + "'");
        }
    }
    
    public String getRepositoryDataType()
    {
        return RepositoryData.CALENDAR_REPOSITORY_DATA_TYPE;
    }
}
