/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.util.Arrays;
import java.util.Optional;

import org.apache.commons.lang3.ArrayUtils;

import org.ametys.cms.data.type.indexing.impl.DoubleIndexableElementType;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.runtime.model.exception.BadItemTypeException;

/**
 * Class for double type of elements stored in the repository
 */
public class DoubleRepositoryElementType extends DoubleIndexableElementType implements RepositoryElementType<Double>
{
    public Object read(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return null;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to get double value from the non double data '" + name + "' on '" + parentData + "'");
        }
        
        if (parentData.isMultiple(name))
        {
            return parentData.getDoubles(name);
        }
        else
        {
            return parentData.getDouble(name);
        }
    }
    
    @Override
    public boolean hasValue(RepositoryData parentData, String name) throws BadItemTypeException
    {
        return RepositoryElementType.super.hasValue(parentData, name) || parentData.hasValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
    }
    
    public boolean hasNonEmptyValue(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return false;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to check double value from the non double data '" + name + "' on '" + parentData + "'");
        }
        
        if (parentData.isMultiple(name))
        {
            return parentData.getDoubles(name).length > 0;
        }
        else
        {
            return true;
        }
    }

    public void write(ModifiableRepositoryData parentData, String name, Object value) throws BadItemTypeException
    {
        if (value == null)
        {
            if (parentData.hasValue(name) && parentData.isMultiple(name))
            {
                parentData.setValues(name, new Double[0]);
            }
            else
            {
                parentData.setValue(name + EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);

                if (parentData.hasValue(name))
                {
                    parentData.removeValue(name);
                }
            }
        }
        else if (value instanceof Number)
        {
            parentData.setValue(name, ((Number) value).doubleValue());
            
            if (parentData.hasValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL))
            {
                parentData.removeValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
            }
        }
        else if (value instanceof double[])
        {
            parentData.setValues(name, ArrayUtils.toObject((double[]) value));
        }
        else if (value instanceof float[])
        {
            Double[] valueAsDoubleArray = Arrays.stream(ArrayUtils.toObject((float[]) value))
                    .map(floatValue -> floatValue.doubleValue())
                    .toArray(Double[]::new);
            
            parentData.setValues(name, valueAsDoubleArray);
        }
        else if (value instanceof Number[])
        {
            Double[] valueAsDoubleArray = Arrays.stream((Number[]) value)
                                                .map(v -> Optional.ofNullable(v)
                                                                  .orElseThrow(() -> new IllegalArgumentException("Try to set a null value into the multiple " + getId() + " data '" + name + "' on '" + parentData + "'")))
                                                .map(number -> number.doubleValue())
                                                .toArray(Double[]::new);
            
            parentData.setValues(name, valueAsDoubleArray);
        }
        else
        {
            throw new BadItemTypeException("Try to set the non double value '" + value + "' to the double data '" + name + "' on '" + parentData + "'");
        }
    }
    
    public String getRepositoryDataType()
    {
        return RepositoryData.DOUBLE_REPOSITORY_DATA_TYPE;
    }
}
