/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.util.Arrays;
import java.util.Optional;
import java.util.stream.Collectors;

import javax.jcr.Session;

import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.data.Binary;
import org.ametys.cms.data.ExplorerFile;
import org.ametys.cms.data.File;
import org.ametys.cms.data.type.ResourceElementTypeHelper;
import org.ametys.cms.data.type.indexing.impl.FileIndexableElementType;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.type.ComplexRepositoryElementType;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.runtime.model.exception.BadItemTypeException;

/**
 * Class for file type of elements stored in the repository
 */
public class FileRepositoryElementType extends FileIndexableElementType implements RepositoryElementType<File>
{
    public Object read(RepositoryData parentData, String name) throws BadItemTypeException, AmetysRepositoryException
    {
        if (!parentData.hasValue(name))
        {
            return null;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to get file value from the non file data '" + name + "' on '" + parentData + "'");
        }
        
        if (RepositoryData.STRING_REPOSITORY_DATA_TYPE.equals(parentData.getType(name)))
        {
            return _readFileFromStringData(parentData, name);
        }
        else
        {
            return _readFileFromBinaryData(parentData, name);
        }
    }

    private Object _readFileFromStringData(RepositoryData parentData, String name)
    {
        final Session session =  parentData instanceof JCRRepositoryData ? ((JCRRepositoryData) parentData).getSession() : null;
        
        if (parentData.isMultiple(name))
        {
            String[] values = parentData.getStrings(name);
            return Arrays.stream(values)
                         .map(resourceId -> new ExplorerFile(_resolver, resourceId, session))
                         .collect(Collectors.toList())
                         .toArray(new ExplorerFile[values.length]);
        }
        else
        {
            String value = parentData.getString(name);
            return StringUtils.isNotEmpty(value) ? new ExplorerFile(_resolver, value, session) : null;
        }
    }

    private Object _readFileFromBinaryData(RepositoryData parentData, String name)
    {
        if (parentData.isMultiple(name))
        {
            RepositoryData multipleParentData = parentData.getRepositoryData(name);
            
            return multipleParentData.getDataNames()
                                     .stream()
                                     .map(singleDataName -> multipleParentData.getRepositoryData(singleDataName))
                                     .map(singleData -> ResourceElementTypeHelper.readBinaryData(singleData))
                                     .filter(singleValue -> singleValue != null)
                                     .toArray(Binary[]::new);
        }
        else
        {
            RepositoryData data = parentData.getRepositoryData(name);
            if (!ResourceElementTypeHelper.isResourceDataEmpty(data))
            {
                return ResourceElementTypeHelper.readBinaryData(data);
            }
            else
            {
                return null;
            }
        }
    }
    
    public boolean hasNonEmptyValue(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return false;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to check file value from the non file data '" + name + "' on '" + parentData + "'");
        }
        
        if (RepositoryData.STRING_REPOSITORY_DATA_TYPE.equals(parentData.getType(name)))
        {
            if (parentData.isMultiple(name))
            {
                return parentData.getStrings(name).length > 0;
            }
            else
            {
                return StringUtils.isNotEmpty(parentData.getString(name));
            }
        }
        else
        {
            // There is a non string value, it cannot be empty
            return true;
        }
    }

    public void write(ModifiableRepositoryData parentData, String name, Object value) throws BadItemTypeException
    {
        if (value == null)
        {
            if (parentData.hasValue(name) && parentData.isMultiple(name))
            {
                parentData.setValues(name, new String[0]);
            }
            else
            {
                parentData.setValue(name, StringUtils.EMPTY);
            }
        }
        else if (value instanceof String)
        {
            parentData.setValue(name, (String) value);
        }
        else if (value instanceof String[])
        {
            String[] values = Arrays.stream((String[]) value)
                                    .map(v -> Optional.ofNullable(v)
                                                      .orElseThrow(() -> new IllegalArgumentException("Try to set a null value into the multiple " + getId() + " data '" + name + "' on '" + parentData + "'")))
                                    .collect(Collectors.toList())
                                    .toArray(new String[((String[]) value).length]);
            
            parentData.setValues(name, values);
        }
        else if (value instanceof ExplorerFile)
        {
            writeExplorerFile(parentData, name, (ExplorerFile) value);
        }
        else if (value instanceof Binary)
        {
            _writeSingleBinaryValue(parentData, name, (Binary) value);
        }
        else if (value instanceof File[])
        {
            if (((File[]) value).length == 0)
            {
                parentData.setValues(name, new String[0]);
            }
            else if (value instanceof ExplorerFile[])
            {
                // FIXME : maybe handle the untouched value
                String[] values = Arrays.stream((ExplorerFile[]) value)
                                        .map(file -> Optional.ofNullable(file)
                                                             .map(ExplorerFile::getResourceId)
                                                             .orElseThrow(() -> new IllegalArgumentException("Try to set a null value into the multiple " + getId() + " data '" + name + "' on '" + parentData + "'")))
                                        .collect(Collectors.toList())
                                        .toArray(new String[((ExplorerFile[]) value).length]);
                
                parentData.setValues(name, values);
            }
            else if (value instanceof Binary[])
            {
                _writeMultipleBinaryValue(parentData, name, (Binary[]) value);
            }
        }
        else
        {
            throw new BadItemTypeException("Try to set the non file value '" + value + "' to the file data '" + name + "' on '" + parentData + "'");
        }
    }
    
    private void writeExplorerFile(ModifiableRepositoryData parentData, String name, ExplorerFile explorerFile)
    {
        // FIXME : handle better way the untouched value
        String resourceId = explorerFile.getResourceId();
        if (!"untouched".equals(resourceId))
        {
            parentData.setValue(name, resourceId);
        }
    }

    private void _writeMultipleBinaryValue(ModifiableRepositoryData parentData, String name, Binary[] value)
    {
        ModifiableRepositoryData multipleParentData = null;
        if (parentData.hasValue(name))
        {
            if (parentData.getType(name).equals(RepositoryConstants.MULTIPLE_ITEM_NODETYPE))
            {
                multipleParentData = parentData.getRepositoryData(name);
            }
            else
            {
                // The value was stored as String, remove the old property
                parentData.removeValue(name);
            }
        }
        
        if (multipleParentData == null)
        {
            multipleParentData = parentData.addRepositoryData(name, RepositoryConstants.MULTIPLE_ITEM_NODETYPE);
            // Add the type identifier in a property to know what the type of the data even if there is no value
            multipleParentData.setValue(ComplexRepositoryElementType.TYPE_ID_DATA_NAME, getId(), RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        }
        
        // Loop on all given values to replace the existing ones
        for (int i = 0; i < value.length; i++)
        {
            if (value[i] != null)
            {
                _writeSingleBinaryValue(multipleParentData, String.valueOf(i + 1), value[i]);
            }
            else
            {
                throw new IllegalArgumentException("Try to set a null value into the multiple " + getId() + " data '" + name + "' on '" + parentData + "'");
            }
        }
        
        // Remove existing entries from the last one to the last of given values
        for (int i = multipleParentData.getDataNames().size(); i > value.length; i--)
        {
            multipleParentData.removeValue(String.valueOf(i));
        }
    }
    
    private void _writeSingleBinaryValue(ModifiableRepositoryData parentData, String name, Binary binary)
    {
        if (parentData.hasValue(name))
        {
            String dataType = parentData.getType(name);
            if (dataType.equals(RepositoryData.STRING_REPOSITORY_DATA_TYPE))
            {
                // In this case, there are a explorer data (type String) in the repo, so delete it before adding the binary data
                parentData.removeValue(name);
            }
        }
        
        ResourceElementTypeHelper.writeSingleBinaryValue(parentData, name, binary);
    }
    
    public String getRepositoryDataType()
    {
        return RepositoryData.STRING_REPOSITORY_DATA_TYPE;
    }
    
    public boolean isCompatible(RepositoryData parentData, String name) throws UnknownDataException
    {
        if (parentData.hasValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL))
        {
            return true;
        }
        
        String dataType = parentData.getType(name);
        switch (dataType)
        {
            case RepositoryData.STRING_REPOSITORY_DATA_TYPE:
            case RepositoryConstants.BINARY_NODETYPE:
                return true;
            case RepositoryConstants.MULTIPLE_ITEM_NODETYPE:
                RepositoryData multipleItemNode = parentData.getRepositoryData(name);
                return getId().equals(multipleItemNode.getString(ComplexRepositoryElementType.TYPE_ID_DATA_NAME, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
            default:
                return false;
        }
    }
}
