/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.util.Arrays;
import java.util.Optional;

import org.apache.commons.lang3.ArrayUtils;

import org.ametys.cms.data.type.indexing.impl.LongIndexableElementType;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.runtime.model.exception.BadItemTypeException;

/**
 * Class for long type of elements stored in the repository
 */
public class LongRepositoryElementType extends LongIndexableElementType implements RepositoryElementType<Long>
{
    public Object read(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return null;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to get long value from the non long data '" + name + "' on '" + parentData + "'");
        }
        
        if (parentData.isMultiple(name))
        {
            return parentData.getLongs(name);
        }
        else
        {
            return parentData.getLong(name);
        }
    }
    
    @Override
    public boolean hasValue(RepositoryData parentData, String name) throws BadItemTypeException
    {
        return RepositoryElementType.super.hasValue(parentData, name) || parentData.hasValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
    }
    
    public boolean hasNonEmptyValue(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return false;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to check long value from the non long data '" + name + "' on '" + parentData + "'");
        }
        
        if (parentData.isMultiple(name))
        {
            return parentData.getLongs(name).length > 0;
        }
        else
        {
            return true;
        }
    }

    public void write(ModifiableRepositoryData parentData, String name, Object value) throws BadItemTypeException
    {
        if (value == null)
        {
            if (parentData.hasValue(name) && parentData.isMultiple(name))
            {
                parentData.setValues(name, new Long[0]);
            }
            else
            {
                parentData.setValue(name + EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);

                if (parentData.hasValue(name))
                {
                    parentData.removeValue(name);
                }
            }
        }
        else if (value instanceof Number)
        {
            parentData.setValue(name, ((Number) value).longValue());
            
            if (parentData.hasValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL))
            {
                parentData.removeValue(name + EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
            }
        }
        else if (value instanceof long[])
        {
            parentData.setValues(name, ArrayUtils.toObject((long[]) value));
        }
        else if (value instanceof int[])
        {
            Long[] valueAsLongArray = Arrays.stream(ArrayUtils.toObject((int[]) value))
                    .map(intValue -> intValue.longValue())
                    .toArray(Long[]::new);

            parentData.setValues(name, valueAsLongArray);
        }
        else if (value instanceof Number[])
        {
            Long[] valueAsLongArray = Arrays.stream((Number[]) value)
                                            .map(v -> Optional.ofNullable(v)
                                                              .orElseThrow(() -> new IllegalArgumentException("Try to set a null value into the multiple " + getId() + " data '" + name + "' on '" + parentData + "'")))
                                            .map(number -> number.longValue())
                                            .toArray(Long[]::new);
            
            parentData.setValues(name, valueAsLongArray);
        }
        else
        {
            throw new BadItemTypeException("Try to set the non long value '" + value + "' to the long data '" + name + "' on '" + parentData + "'");
        }
    }
    
    public String getRepositoryDataType()
    {
        return RepositoryData.LONG_REPOSITORY_DATA_TYPE;
    }
}
