/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.util.Collection;
import java.util.Locale;

import org.apache.commons.lang3.LocaleUtils;

import org.ametys.cms.data.type.indexing.impl.MultilingualStringIndexableElementType;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.type.ComplexRepositoryElementType;
import org.ametys.plugins.repository.metadata.MultilingualString;

/**
 * Class for multilingual string type of elements stored in the repository
 */
public class MultilingualStringRepositoryElementType extends MultilingualStringIndexableElementType implements ComplexRepositoryElementType<MultilingualString>
{
    public boolean isSingleValueEmpty(RepositoryData multilingualStringData)
    {
        return multilingualStringData.getDataNames()
                                     .stream()
                                     .noneMatch(this::_isAValidLocale);
    }
    
    public MultilingualString readSingleValue(RepositoryData multilingualStringData)
    {
        Collection<String> localesAsString = multilingualStringData.getDataNames();
        if (!localesAsString.isEmpty())
        {
            MultilingualString multilingualString = new MultilingualString();
            
            for (String localeAsString : localesAsString)
            {
                Locale locale = LocaleUtils.toLocale(localeAsString);
                multilingualString.add(locale, multilingualStringData.getString(localeAsString));
            }
            
            return multilingualString;
        }
        else
        {
            return null;
        }
    }
    
    public void writeSingleValue(ModifiableRepositoryData parentData, String name, MultilingualString value)
    {
        ModifiableRepositoryData multilingualStringData = parentData.addRepositoryData(name, getRepositoryDataType());
        
        if (value != null)
        {
            for (Locale locale : value.getLocales())
            {
                multilingualStringData.setValue(locale.toString(), value.getValue(locale));
            }
        }
    }
    
    public String getRepositoryDataType()
    {
        return RepositoryConstants.MULTILINGUAL_STRING_METADATA_NODETYPE;
    }
    
    /**
     * Checks if the given string is an available {@link Locale}
     * @param candidate the string to check
     * @return <code>true</code> if the given string is an available locale, <code>false</code> otherwise
     */
    protected boolean _isAValidLocale(String candidate)
    {
        try
        {
            LocaleUtils.toLocale(candidate);
            return true;
        }
        catch (IllegalArgumentException e)
        {
            // the given candidate is not a locale
            return false;
        }
    }
}
