/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.data.Reference;
import org.ametys.cms.data.type.indexing.impl.ReferenceIndexableElementType;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.type.ComplexRepositoryElementType;
import org.ametys.runtime.model.exception.BadItemTypeException;

/**
 * Class for reference type of elements stored in the repository
 */
public class ReferenceRepositoryElementType extends ReferenceIndexableElementType implements ComplexRepositoryElementType<Reference>
{
    public boolean isSingleValueEmpty(RepositoryData referenceData)
    {
        return _isStringValueEmpty(referenceData, Reference.VALUE_IDENTIFIER) || _isStringValueEmpty(referenceData, Reference.TYPE_IDENTIFIER);
    }
    
    private boolean _isStringValueEmpty(RepositoryData referenceData, String name)
    {
        return !referenceData.hasValue(name) || StringUtils.isEmpty(referenceData.getString(name));
    }

    public Reference readSingleValue(RepositoryData singleValueData)
    {
        String value = _getStringValue(singleValueData, Reference.VALUE_IDENTIFIER);
        String type = _getStringValue(singleValueData, Reference.TYPE_IDENTIFIER);
        return (StringUtils.isNoneEmpty(value, type)) ? new Reference(value, type) : null;
    }
    
    private String _getStringValue(RepositoryData repositoryData, String name)
    {
        if (!repositoryData.hasValue(name))
        {
            return null;
        }
        else if (RepositoryData.STRING_REPOSITORY_DATA_TYPE.equals(repositoryData.getType(name)))
        {
            return repositoryData.getString(name);
        }
        else
        {
            throw new BadItemTypeException("Try to get string value from the non string data '" + name + "' on '" + repositoryData + "'");
        }
    }
    
    public void emptySingleValue(ModifiableRepositoryData parentData, String name)
    {
        ComplexRepositoryElementType.super.emptySingleValue(parentData, name);

        ModifiableRepositoryData referenceData = parentData.getRepositoryData(name);
        referenceData.setValue(Reference.VALUE_IDENTIFIER, StringUtils.EMPTY);
        referenceData.setValue(Reference.TYPE_IDENTIFIER, StringUtils.EMPTY);
    }

    public void writeSingleValue(ModifiableRepositoryData parentData, String name, Reference value)
    {
        ModifiableRepositoryData referenceData = parentData.addRepositoryData(name, getRepositoryDataType());
        
        if (value != null)
        {
            referenceData.setValue(Reference.VALUE_IDENTIFIER, value.getValue());
            referenceData.setValue(Reference.TYPE_IDENTIFIER, value.getType());
        }
    }
    
    @Override
    public boolean isCompatible(RepositoryData parentData, String name) throws UnknownDataException
    {
        if (ComplexRepositoryElementType.super.isCompatible(parentData, name))
        {
            return true;
        }
        // TODO NEWATTRIBUTEAPI: This part of the test is here just to be compatible with the old compositMetadata type. When all references are migrated, remove this part of the test
        else if (RepositoryConstants.COMPOSITE_METADTA_NODETYPE.equals(parentData.getType(name)))
        {
            RepositoryData possibleReference = parentData.getRepositoryData(name);
            return possibleReference.getDataNames().size() == 2
                && possibleReference.hasValue(Reference.VALUE_IDENTIFIER)
                && possibleReference.hasValue(Reference.TYPE_IDENTIFIER);
        }
        else
        {
            return false;
        }
    }

    public String getRepositoryDataType()
    {
        return RepositoryConstants.REFERENCE_NODETYPE;
    }

}
