/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.util.Collection;
import java.util.Optional;

import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.data.holder.impl.IndexableDataHolderHelper;
import org.ametys.core.model.type.AbstractModelItemType;
import org.ametys.core.model.type.ModelItemTypeHelper;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeaterEntry;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryModelItemGroupType;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.runtime.model.ModelItemContainer;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.type.DataContext;

/**
 * Class for repeater type
 */
public class RepeaterRepositoryModelItemType extends AbstractModelItemType implements RepositoryModelItemGroupType
{
    public void valueToSAXForEdition(ContentHandler contentHandler, String tagName, Object value, DataContext context) throws SAXException
    {
        _valueToSAX(contentHandler, tagName, value, context, true);
    }
    
    public void valueToSAX(ContentHandler contentHandler, String tagName, Object value, DataContext context) throws SAXException
    {
        _valueToSAX(contentHandler, tagName, value, context, false);
    }
    
    private void _valueToSAX(ContentHandler contentHandler, String tagName, Object value, DataContext context, boolean isEdition) throws SAXException
    {
        AttributesImpl attributes = ModelItemTypeHelper.getXMLAttributesFromDataContext(context);
        
        if (value == null)
        {
            XMLUtils.createElement(contentHandler, tagName, attributes);
        }
        else if (value instanceof ModelAwareRepeater repeater)
        {
            attributes.addCDATAAttribute("entryCount", String.valueOf(repeater.getSize()));
            
            ViewItemAccessor viewItemAccessor = context.getViewItem()
                                                       .filter(ViewItemAccessor.class::isInstance)
                                                       .map(ViewItemAccessor.class::cast)
                                                       .orElse(ModelViewItemGroup.of(repeater.getModel()));

            XMLUtils.startElement(contentHandler, tagName, attributes);
            for (ModelAwareRepeaterEntry entry : repeater.getEntries())
            {
                _repeaterEntryToSAX(contentHandler, entry, viewItemAccessor, context, isEdition);
            }
            XMLUtils.endElement(contentHandler, tagName);
        }
        else if (value instanceof ModelAwareRepeaterEntry entry)
        {
            ViewItemAccessor viewItemAccessor = context.getViewItem()
                                                       .filter(ViewItemAccessor.class::isInstance)
                                                       .map(ViewItemAccessor.class::cast)
                                                       .orElse(ModelViewItemGroup.of(_getRepeaterEntryDefinition(entry)));
            
            _repeaterEntryToSAX(contentHandler, entry, viewItemAccessor, context, isEdition);
        }
        else
        {
            throw new IllegalArgumentException("Try to sax the non repeater value '" + value + "' in tag name '" + tagName + "'");
        }
    }
    
    private RepeaterDefinition _getRepeaterEntryDefinition(ModelAwareRepeaterEntry entry)
    {
        Collection<? extends ModelItemContainer> modelItemContainers = entry.getModel();
        assert modelItemContainers.size() == 1;
        
        ModelItemContainer modelItemContainer = modelItemContainers.iterator().next();
        assert modelItemContainer instanceof RepeaterDefinition;
        
        return (RepeaterDefinition) modelItemContainer;
    }

    private void _repeaterEntryToSAX(ContentHandler contentHandler, ModelAwareRepeaterEntry entry, ViewItemAccessor viewItemAccessor, DataContext context, boolean isEdition) throws SAXException
    {
        DataContext newContext = context.cloneContext()
                                        .addSuffixToLastSegment("[" + entry.getPosition() + "]");
        
        // Don't sax model path on entry, to avoid the same value as the repeater level
        XMLUtils.startElement(contentHandler, "entry", _getEntryAttributes(entry, newContext));
        IndexableDataHolderHelper.dataToSAX(entry, contentHandler, viewItemAccessor, newContext, isEdition);
        XMLUtils.endElement(contentHandler, "entry");
    }
    
    private Attributes _getEntryAttributes(ModelAwareRepeaterEntry entry, DataContext context)
    {
        AttributesImpl entryAttrs = ModelItemTypeHelper.getXMLAttributesFromDataContext(context, false);
        String entryName = Integer.toString(entry.getPosition());
        entryAttrs.addCDATAAttribute("name", entryName);
        return entryAttrs;
    }
    
    public Object valueToJSONForClient(Object value, DataContext context)
    {
        return _valueToJSON(value, context, false);
    }
    
    public Object valueToJSONForEdition(Object value, DataContext context)
    {
        return _valueToJSON(value, context, true);
    }
    
    private Object _valueToJSON(Object value, DataContext context, boolean isEdition)
    {
        if (value == null)
        {
            return null;
        }
        
        Optional<ViewItemAccessor> optViewItemAccessor = context.getViewItem()
                                                                .filter(ViewItemAccessor.class::isInstance)
                                                                .map(ViewItemAccessor.class::cast);
        
        if (value instanceof ModelAwareRepeater repeater)
        {
            return optViewItemAccessor.isEmpty() ? repeater.dataToJSON(context)
                                                 : isEdition
                                                     ? repeater.dataToJSONForEdition(optViewItemAccessor.get(), context)
                                                     : repeater.dataToJSON(optViewItemAccessor.get(), context);
        }
        else if (value instanceof ModelAwareRepeaterEntry entry)
        {
            DataContext entryContext = context.cloneContext().addSuffixToLastSegment("[" + entry.getPosition() + "]");
            
            return optViewItemAccessor.isEmpty() ? entry.dataToJSON(entryContext)
                                                 : isEdition
                                                     ? entry.dataToJSONForEdition(optViewItemAccessor.get(), entryContext)
                                                     : entry.dataToJSON(optViewItemAccessor.get(), entryContext);
        }
        else
        {
            throw new IllegalArgumentException("Try to convert the non repeater value '" + value + "'");
        }
    }

    public boolean isCompatible(RepositoryData parentData, String name) throws UnknownDataException
    {
        if (RepositoryModelItemGroupType.super.isCompatible(parentData, name))
        {
            return true;
        }
        // TODO NEWATTRIBUTEAPI: This part of the test is here just to be compatible with the old compositMetadata type. When all repeaters are migrated, remove this part of the test
        else if (RepositoryConstants.COMPOSITE_METADTA_NODETYPE.equals(parentData.getType(name)))
        {
            RepositoryData possibleRepeater = parentData.getRepositoryData(name);
            for (String dataName : possibleRepeater.getDataNames())
            {
                if (!NumberUtils.isParsable(dataName))
                {
                    // If there is at least one data that is not a number (entry position), this data can not be a repeater
                    return false;
                }
            }
            
            // All data are numbers and could be entry positions, the data could be a repeater
            return true;
        }
        else
        {
            return false;
        }
    }
    
    public String getRepositoryDataType()
    {
        return RepositoryConstants.REPEATER_NODETYPE;
    }
}
