/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.apache.jackrabbit.util.Text;

import org.ametys.cms.data.NamedResource;
import org.ametys.cms.data.RichText;
import org.ametys.cms.data.type.ResourceElementTypeHelper;
import org.ametys.cms.data.type.indexing.impl.RichTextIndexableElementType;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.type.ComplexRepositoryElementType;
import org.ametys.runtime.model.type.DataContext;

/**
 * Class for rich text type of elements stored in the repository
 */
public class RichTextRepositoryElementType extends RichTextIndexableElementType implements ComplexRepositoryElementType<RichText>
{
    private static final String __ANNOTATIONS_IDENTIFIER = "annotations";
    private static final String __FOLDER_IDENTIFIER = "data";
    private static final String __FILE_CONTENT_IDENTIFIER = "content";
    
    public boolean isSingleValueEmpty(RepositoryData singleValueData)
    {
        return ResourceElementTypeHelper.isResourceDataEmpty(singleValueData);
    }

    public RichText readSingleValue(RepositoryData singleValueData)
    {
        RepositoryData folderData = ResourceElementTypeHelper.getRepositoryData(singleValueData, RepositoryConstants.FOLDER_NODETYPE, __FOLDER_IDENTIFIER, StringUtils.EMPTY);
        
        RichText richText = new RichText(singleValueData, folderData);
        ResourceElementTypeHelper.readResourceData(singleValueData, richText);

        // Read annotations
        if (singleValueData.hasValue(__ANNOTATIONS_IDENTIFIER, StringUtils.EMPTY))
        {
            RepositoryData annotationsData = ResourceElementTypeHelper.getRepositoryData(singleValueData, RepositoryConstants.COMPOSITE_METADTA_NODETYPE, __ANNOTATIONS_IDENTIFIER, StringUtils.EMPTY);
            for (String annotationName : annotationsData.getDataNames())
            {
                String[] annotationValues = ResourceElementTypeHelper.getStringValues(annotationsData, annotationName, RepositoryConstants.NAMESPACE_PREFIX);
                if (annotationValues != null && annotationValues.length > 0)
                {
                    richText.addAnnotations(annotationName, annotationValues);
                }
            }
        }

        return richText;
    }
    
    public void emptySingleValue(ModifiableRepositoryData parentData, String name)
    {
        ResourceElementTypeHelper.emptyResourceData(parentData, name, getRepositoryDataType());

        // Remove attachments (because it could be big)
        ModifiableRepositoryData richTextData = parentData.hasValue(name) ? parentData.getRepositoryData(name) : parentData.addRepositoryData(name, getRepositoryDataType());
        if (richTextData.hasValue(__FOLDER_IDENTIFIER, StringUtils.EMPTY))
        {
            ModifiableRepositoryData folderData = ResourceElementTypeHelper.getRepositoryData(richTextData, RepositoryConstants.FOLDER_NODETYPE, __FOLDER_IDENTIFIER, StringUtils.EMPTY);
            for (String attachmentNodeName : folderData.getDataNames(StringUtils.EMPTY))
            {
                folderData.removeValue(attachmentNodeName, StringUtils.EMPTY);
            }
        }
    }

    public void writeSingleValue(ModifiableRepositoryData parentData, String name, RichText value)
    {
        InputStream stream = value.getInputStream();
        if (stream == null)
        {
            emptySingleValue(parentData, name);
        }
        else
        {
            ModifiableRepositoryData richTextData;
            if (parentData.hasValue(name))
            {
                richTextData = ResourceElementTypeHelper.getRepositoryData(parentData, getRepositoryDataType(), name, RepositoryConstants.NAMESPACE_PREFIX);
            }
            else
            {
                richTextData = parentData.addRepositoryData(name, getRepositoryDataType());
            }
            
            richTextData.setValue(ResourceElementTypeHelper.EMPTY_RESOURCE_IDENTIFIER, false, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);

            ResourceElementTypeHelper.writeResourceData(richTextData, value);
            richTextData.setValue(ResourceElementTypeHelper.DATA_IDENTIFIER, stream, ResourceElementTypeHelper.METADATA_PREFIX);

            _writeAnnotations(richTextData, value.getAllAnnotations());
            _writeAttachments(richTextData, value);
        }
    }
    
    private void _writeAnnotations(ModifiableRepositoryData richTextData, Map<String, List<String>> annotations)
    {
        if (richTextData.hasValue(__ANNOTATIONS_IDENTIFIER, StringUtils.EMPTY))
        {
            richTextData.removeValue(__ANNOTATIONS_IDENTIFIER, StringUtils.EMPTY);
        }
        
        ModifiableRepositoryData annotationsData = richTextData.addRepositoryData(__ANNOTATIONS_IDENTIFIER, RepositoryConstants.COMPOSITE_METADTA_NODETYPE, StringUtils.EMPTY);
        for (Map.Entry<String, List<String>> annotationEntry : annotations.entrySet())
        {
            List<String> annotationValues = annotationEntry.getValue();
            annotationsData.setValues(annotationEntry.getKey(), annotationValues.toArray(new String[annotationValues.size()]));
        }
    }
    
    private void _writeAttachments(ModifiableRepositoryData richTextData, RichText richText)
    {
        ModifiableRepositoryData folderData;
        if (richTextData.hasValue(__FOLDER_IDENTIFIER, StringUtils.EMPTY))
        {
            folderData = ResourceElementTypeHelper.getRepositoryData(richTextData, RepositoryConstants.FOLDER_NODETYPE, __FOLDER_IDENTIFIER, StringUtils.EMPTY);
        }
        else
        {
            folderData = richTextData.addRepositoryData(__FOLDER_IDENTIFIER, RepositoryConstants.FOLDER_NODETYPE, StringUtils.EMPTY);
        }
        
        // Fetched attachments
        for (NamedResource attachment : richText.getFetchedAttachments())
        {
            String escapedFilename = Text.escapeIllegalJcrChars(attachment.getFilename());
            ModifiableRepositoryData resourceData = _getFileResourceRepositoryData(folderData, escapedFilename);
            
            ResourceElementTypeHelper.writeResourceData(resourceData, attachment);

            InputStream stream = attachment.getInputStream();
            if (stream != null)
            {
                resourceData.setValue(ResourceElementTypeHelper.DATA_IDENTIFIER, stream, ResourceElementTypeHelper.METADATA_PREFIX);
            }
            else
            {
                try (ByteArrayInputStream is = new ByteArrayInputStream(StringUtils.EMPTY.getBytes()))
                {
                    resourceData.setValue(ResourceElementTypeHelper.DATA_IDENTIFIER, is, ResourceElementTypeHelper.METADATA_PREFIX);
                }
                catch (IOException e)
                {
                    throw new AmetysRepositoryException(e);
                }
            }
        }
        
        // Removed attachments
        for (String filename : richText.getRemovedAttachments())
        {
            String escapedFilename = Text.escapeIllegalJcrChars(filename);
            if (folderData.hasValue(escapedFilename, StringUtils.EMPTY))
            {
                folderData.removeValue(escapedFilename, StringUtils.EMPTY);
            }
        }
        
        // Add unfetched attachments
        // Ex: when copying a rich text from a data holder to another, the untouched attachments are not yet in the destination repository data 
        for (String filename : richText.getAttachmentNames())
        {
            String escapedFilename = Text.escapeIllegalJcrChars(filename);
            if (!folderData.hasValue(escapedFilename, StringUtils.EMPTY))
            {
                NamedResource attachment = richText.getAttachment(filename);
                ModifiableRepositoryData resourceData = _getFileResourceRepositoryData(folderData, escapedFilename);
                
                ResourceElementTypeHelper.writeResourceData(resourceData, attachment);
                Optional.ofNullable(attachment.getInputStream()).ifPresent(stream -> resourceData.setValue(ResourceElementTypeHelper.DATA_IDENTIFIER, stream, ResourceElementTypeHelper.METADATA_PREFIX));
            }
        }
    }
    
    private ModifiableRepositoryData _getFileResourceRepositoryData(ModifiableRepositoryData folderData, String filename)
    {
        if (folderData.hasValue(filename, StringUtils.EMPTY))
        {
            ModifiableRepositoryData fileData = ResourceElementTypeHelper.getRepositoryData(folderData, RepositoryConstants.FILE_NODETYPE, filename, StringUtils.EMPTY);
            if (fileData.hasValue(__FILE_CONTENT_IDENTIFIER, ResourceElementTypeHelper.METADATA_PREFIX))
            {
                return ResourceElementTypeHelper.getRepositoryData(fileData, RepositoryConstants.RESOURCE_NODETYPE, __FILE_CONTENT_IDENTIFIER, ResourceElementTypeHelper.METADATA_PREFIX);
            }
            else
            {
                return fileData.addRepositoryData(__FILE_CONTENT_IDENTIFIER, RepositoryConstants.RESOURCE_NODETYPE, ResourceElementTypeHelper.METADATA_PREFIX);
            }
        }
        else
        {
            ModifiableRepositoryData fileData = folderData.addRepositoryData(filename, RepositoryConstants.FILE_NODETYPE, StringUtils.EMPTY);
            return fileData.addRepositoryData(__FILE_CONTENT_IDENTIFIER, RepositoryConstants.RESOURCE_NODETYPE, ResourceElementTypeHelper.METADATA_PREFIX);
        }
    }
    
    public Object externalizableValueToJSON(Object value, DataContext context)
    {
        if (value == null)
        {
            return null;
        }
        else if (value instanceof RichText)
        {
            return _richTextAsString((RichText) value, context);
        }
        else if (value instanceof RichText[])
        {
            List<Object> jsonList = new ArrayList<>();
            for (RichText singleValue : (RichText[]) value)
            {
                jsonList.add(_richTextAsString(singleValue, context));
            }
            return jsonList;
        }
        else
        {
            throw new IllegalArgumentException("Try to convert the non " + getManagedClass().getName() + " value '" + value + "' to JSON");
        }
    }
    
    private String _richTextAsString (RichText richText, DataContext context)
    {
        StringBuilder result = new StringBuilder(2048);
        
        try
        {
            _richTextTransformer.transformForEditing(richText, context, result);
        }
        catch (IOException e)
        {
            throw new AmetysRepositoryException("Unable to transform a rich text into a string", e);
        }
        
        return result.toString();
    }
    
    @Override
    public boolean removeValueBeforeWritingIt()
    {
        return false;
    }

    public String getRepositoryDataType()
    {
        return RepositoryConstants.RICH_TEXT_NODETYPE;
    }
}
