/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.util.Arrays;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.data.type.indexing.impl.StringIndexableElementType;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.runtime.model.exception.BadItemTypeException;

/**
 * Class for string type of elements stored in the repository
 */
public class StringRepositoryElementType extends StringIndexableElementType implements RepositoryElementType<String>
{
    public Object read(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return null;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to get string value from the non string data '" + name + "' on '" + parentData + "'");
        }
        
        if (parentData.isMultiple(name))
        {
            return parentData.getStrings(name);
        }
        else
        {
            return parentData.getString(name);
        }
    }
    
    public boolean hasNonEmptyValue(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (!parentData.hasValue(name))
        {
            return false;
        }
        
        if (!isCompatible(parentData, name))
        {
            throw new BadItemTypeException("Try to check string value from the non string data '" + name + "' on '" + parentData + "'");
        }
        
        if (parentData.isMultiple(name))
        {
            return parentData.getStrings(name).length > 0;
        }
        else
        {
            return StringUtils.isNotEmpty(parentData.getString(name));
        }
    }

    public void write(ModifiableRepositoryData parentData, String name, Object value) throws BadItemTypeException
    {
        if (value == null)
        {
            if (parentData.hasValue(name) && parentData.isMultiple(name))
            {
                parentData.setValues(name, new String[0]);
            }
            else
            {
                parentData.setValue(name, StringUtils.EMPTY);
            }
        }
        else if (value instanceof String)
        {
            parentData.setValue(name, (String) value);
        }
        else if (value instanceof String[])
        {
            Arrays.stream((String[]) value)
                  .forEach(v -> Optional.ofNullable(v)
                                        .orElseThrow(() -> new IllegalArgumentException("Try to set a null value into the multiple " + getId() + " data '" + name + "' on '" + parentData + "'")));
            
            parentData.setValues(name, (String[]) value);
        }
        else
        {
            throw new BadItemTypeException("Try to set the non string value '" + value + "' to the string data '" + name + "' on '" + parentData + "'");
        }
    }
    
    public String getRepositoryDataType()
    {
        return RepositoryData.STRING_REPOSITORY_DATA_TYPE;
    }
}
