/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.util.Arrays;

import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.data.type.indexing.impl.UserIndexableElementType;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.type.ComplexRepositoryElementType;
import org.ametys.runtime.model.exception.BadItemTypeException;

/**
 * Class for user type of elements stored in the repository
 */
public class UserRepositoryElementType extends UserIndexableElementType implements ComplexRepositoryElementType<UserIdentity>
{
    @Override
    // FIXME REPOSITORY-428: This read method overriding exists just to be compatible with the old API. When all users are migrated, remove this read overriding
    public Object read(RepositoryData parentData, String name) throws BadItemTypeException
    {
        if (_isMultipleButSetViaOldAPI(parentData, name))
        {
            RepositoryData[] data = parentData.getAllRepositoryData(name);
            
            return Arrays.stream(data)
                         .map(singleData -> !isSingleValueEmpty(singleData) ? readSingleValue(singleData) : null)
                         .toArray(UserIdentity[]::new);
        }
        else
        {
            return ComplexRepositoryElementType.super.read(parentData, name);
        }
    }
    
    public boolean isSingleValueEmpty(RepositoryData userData)
    {
        return _isStringValueEmpty(userData, "login") || _isStringValueEmpty(userData, "population");
    }
    
    private boolean _isStringValueEmpty(RepositoryData userData, String name)
    {
        return !userData.hasValue(name) || StringUtils.isEmpty(userData.getString(name));
    }
    
    public UserIdentity readSingleValue(RepositoryData userData)
    {
        String login = _getStringValue(userData, "login");
        String populationId = _getStringValue(userData, "population");
        return StringUtils.isNoneEmpty(login, populationId) ? new UserIdentity(login, populationId) : null;
    }

    private String _getStringValue(RepositoryData userData, String name)
    {
        if (!userData.hasValue(name))
        {
            return null;
        }
        else if (RepositoryData.STRING_REPOSITORY_DATA_TYPE.equals(userData.getType(name)))
        {
            return userData.getString(name);
        }
        else
        {
            throw new BadItemTypeException("Try to get string value from the non string data '" + name + "' on '" + userData + "'");
        }
    }
    
    public void emptySingleValue(ModifiableRepositoryData parentData, String name)
    {
        ComplexRepositoryElementType.super.emptySingleValue(parentData, name);
        
        ModifiableRepositoryData userData = parentData.getRepositoryData(name);
        userData.setValue("login", StringUtils.EMPTY);
        userData.setValue("population", StringUtils.EMPTY);
    }

    public void writeSingleValue(ModifiableRepositoryData parentData, String name, UserIdentity value)
    {
        ModifiableRepositoryData userData = parentData.addRepositoryData(name, getRepositoryDataType());
        
        if (value != null)
        {
            userData.setValue("login", value.getLogin());
            userData.setValue("population", value.getPopulationId());
        }
    }
    
    public boolean isMultiple(RepositoryData parentData, String name) throws UnknownDataException
    {
        return _isMultipleButSetViaOldAPI(parentData, name) || ComplexRepositoryElementType.super.isMultiple(parentData, name);
    }
    
    private boolean _isMultipleButSetViaOldAPI(RepositoryData parentData, String name)
    {
        if (parentData.hasValue(name))
        {
            // With the old API, multiple users were set as many nodes with the same name.
            // With the new API, the node with the given name is of type multipleItem and users node are below with numbers as names
            // So if there is many nodes with the given name, we can conclude that the value has been set with the old API
            RepositoryData[] data = parentData.getAllRepositoryData(name);
            return data.length > 1;
        }
        else
        {
            return false;
        }
    }
    
    public String getRepositoryDataType()
    {
        return RepositoryConstants.USER_NODETYPE;
    }
}
