/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.indexing;

import org.ametys.cms.content.indexing.solr.SolrFieldNames;
import org.ametys.cms.model.CMSDataContext;
import org.ametys.runtime.model.type.DataContext;

/**
 * Object that gives some context for data indexation manipulation
 */
public class IndexableDataContext extends CMSDataContext
{
    private boolean _indexForFullTextField;
    
    private String _fullTextFieldName = SolrFieldNames.SYSTEM_FULL;
    
    /**
     * Creates a new instance of a {@link IndexableDataContext}
     */
    protected IndexableDataContext()
    {
        // Empty constructor
    }
    
    /**
     * Creates a new instance of a {@link IndexableDataContext} from another {@link DataContext}
     * @param context the data context to copy
     */
    protected IndexableDataContext(DataContext context)
    {
        super(context);
        if (context instanceof IndexableDataContext indexableContext)
        {
            withIndexForFullTextField(indexableContext.indexForFullTextField());
            withFullTextFieldName(indexableContext.getFullTextFieldName());
        }
    }
    
    /**
     * Creates a new instance of a {@link IndexableDataContext}
     * @return the created instance
     */
    public static IndexableDataContext newInstance()
    {
        return new IndexableDataContext();
    }
    
    /**
     * Creates a new instance of a {@link IndexableDataContext} from another {@link DataContext}.
     * It can be the same implementation or another one, but it will be casted to the current implementation.
     * @param context the data context to copy
     * @return the created instance
     */
    public static IndexableDataContext newInstance(DataContext context)
    {
        return new IndexableDataContext(context);
    }
    
    /**
     * Creates a new instance of a {@link IndexableDataContext}, with the current context values
     * @return the created instance
     */
    @Override
    public IndexableDataContext cloneContext()
    {
        return newInstance(this);
    }
    
    /**
     * Determines if the value should be indexed for a fulltext field. The fulltext field is given by this {@link IndexableDataContext}
     * @return <code>true</code> to index value for a fulltext field, <code>false</code> otherwise.
     */
    public boolean indexForFullTextField()
    {
        return _indexForFullTextField;
    }
    
    /**
     * Set to <code>true</code> to index the value for a fulltext field. The fulltext field is given by this {@link IndexableDataContext}
     * @param <T> the type of the retrieved {@link DataContext}
     * @param indexForFullTextField <code>true</code> to index the value for a fulltext field, <code>false</code> otherwise
     * @return the current {@link IndexableDataContext}
     */
    @SuppressWarnings("unchecked")
    public <T extends IndexableDataContext> T withIndexForFullTextField(boolean indexForFullTextField)
    {
        _indexForFullTextField = indexForFullTextField;
        return (T) this;
    }
    
    /**
     * Get the field name of full-text field. Use if this {@link IndexableDataContext} is instanciated for full-text field
     * @return the field name of full-text field. Default to SolrFieldNames.SYSTEM_FULL
     */
    public String getFullTextFieldName()
    {
        return _fullTextFieldName;
    }

    /**
     * Set the field name of full-text field. Use if this {@link IndexableDataContext} is instanciated for full-text field
     * @param <T> the type of the retrieved {@link DataContext}
     * @param fulltextFieldName the full-text field name
     * @return the current {@link IndexableDataContext}
     */
    @SuppressWarnings("unchecked")
    public <T extends IndexableDataContext> T withFullTextFieldName(String fulltextFieldName)
    {
        _fullTextFieldName = fulltextFieldName;
        return (T) this;
    }
}
