/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.indexing;

import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.apache.solr.common.SolrInputDocument;

import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ElementType;

/**
 * Interface for element types that can be indexed
 * @param <T> Type of the element value
 */
public interface IndexableElementType<T> extends IndexableModelItemType, ElementType<T>
{
    @SuppressWarnings("unchecked")
    public default void indexValue(SolrInputDocument document, SolrInputDocument rootObjectDocument, String fieldName, Object value, IndexableDataContext context)
    {
        if (value != null)
        {
            if (getManagedClass().isInstance(value))
            {
                if (context.indexForFullTextField())
                {
                    indexSingleValueForFullTextField(document, rootObjectDocument, (T) value, context);
                }
                else
                {
                    indexSingleValue(document, rootObjectDocument, fieldName, (T) value, context);
                }
            }
            else if (getManagedClassArray().isInstance(value))
            {
                for (T singleValue : (T[]) value)
                {
                    if (context.indexForFullTextField())
                    {
                        indexSingleValueForFullTextField(document, rootObjectDocument, singleValue, context);
                    }
                    else
                    {
                        indexSingleValue(document, rootObjectDocument, fieldName, singleValue, context);
                    }
                }
            }
            else
            {
                throw new IllegalArgumentException("Try to convert the non " + getId() + " value '" + value + "' to JSON");
            }
        }
    }
    
    /**
     * Index the given single value in a solr document.
     * @param document the solr document to index into.
     * @param rootObjectDocument the solr document of the root object.
     * @param fieldName the name of the indexing field
     * @param value the single value to index
     * @param context the context of the data to index. Can not be null.
     */
    public void indexSingleValue(SolrInputDocument document, SolrInputDocument rootObjectDocument, String fieldName, T value, IndexableDataContext context);
    
    /**
     * Index the given single value in a full-text field of solr document. The solr full-text field name is given by the data context.
     * @param document the solr document to index into.
     * @param rootObjectDocument the solr document of the root object.
     * @param value the single value to index in fulltext field.
     * @param context the context of the data to index. Can not be null.
     */
    public void indexSingleValueForFullTextField(SolrInputDocument document, SolrInputDocument rootObjectDocument, T value, IndexableDataContext context);
    
    /**
     * Retrieves the formatted value to index
     * @param <X> Type of the value to retrieve
     * @param value the value to index
     * @return the formatted value
     */
    @SuppressWarnings("unchecked")
    public default <X> X getSingleValueToIndex(T value)
    {
        return (X) value;
    }
    
    /**
     * Get the solr schema type from the type.
     * @return The solr schema type.
     */
    public String getSchemaType();
    
    /**
     * Retrieves the indexing solr field suffix for this type
     * @param context the indexable data context
     * @return the indexing solr field suffix
     */
    public default String getIndexingFieldSuffix(DataContext context)
    {
        return StringUtils.EMPTY;
    }
    
    /**
     * Retrieves the wildcard solr field suffix for this type
     * @param context the indexable data context
     * @return the wildcard solr field suffix
     */
    public default Optional<String> getWildcardFieldSuffix(DataContext context)
    {
        return Optional.empty();
    }
    
    /**
     * Retrieves the text solr field suffix for this type
     * @param context the indexable data context
     * @return the text solr field suffix
     */
    public default Optional<String> getTextFieldSuffix(DataContext context)
    {
        return Optional.empty();
    }
    
    /**
     * Checks if data of this type is facetable
     * @param context the indexable data context
     * @return <code>true</code> if the type is facetable, <code>false</code> otherwise
     */
    public default boolean isFacetable(DataContext context)
    {
        return IndexableElementTypeHelper.isEnumerated(context);
    }
}
