/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.indexing;

import java.io.InputStream;
import java.util.Optional;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;
import org.apache.solr.common.SolrInputDocument;
import org.apache.tika.Tika;
import org.slf4j.Logger;

import org.ametys.cms.content.indexing.solr.SolrFieldNames;
import org.ametys.cms.content.indexing.solr.SolrIndexer;
import org.ametys.cms.data.File;
import org.ametys.cms.languages.LanguagesManager;
import org.ametys.plugins.repository.data.ametysobject.ModelAwareDataAwareAmetysObject;
import org.ametys.plugins.repository.model.RepositoryDataContext;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.type.DataContext;

/**
 * Helper for indexable element types
 */
public class IndexableElementTypeHelper implements Component, Serviceable
{
    private static LanguagesManager _languagesManager;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _languagesManager = (LanguagesManager) manager.lookup(LanguagesManager.ROLE);
    }
    
    /**
     * Index the given string value in a solr document.
     * @param document the solr document to index into.
     * @param rootObjectDocument the solr document of the root object.
     * @param fieldName the name of the indexing field
     * @param valueToIndex the value to index
     * @param context the context of the data to index. Can not be null
     * @param logger the logger to log in WARN level in case of the given string is too long and will be truncated
     */
    public static void indexStringValue(SolrInputDocument document, SolrInputDocument rootObjectDocument, String fieldName, String valueToIndex, IndexableDataContext context, Logger logger)
    {
        if (!context.indexForFullTextField())
        {
            if (!isStringDataFacetable(context))
            {
                if (context.getLocale() != null) // Locale can be null for multilingual content
                {
                    // No enumerator: index as full-text.
                    String language = context.getLocale().getLanguage();
                    document.addField(fieldName + getStringTextFieldSuffix(context), valueToIndex);
                    document.addField(fieldName + "_txt_stemmed_" + language, valueToIndex);
                    document.addField(fieldName + "_txt_ws_" + language, valueToIndex);
                }
            }
            else
            {
                // Facets (enumeration only)
                document.addField(fieldName + "_s_dv", valueToIndex);
            }
            
            // Index without analyzing.
            String possiblyTruncatedValue = SolrIndexer.truncateUtf8StringValue(valueToIndex, logger, context.getObjectId().orElse(null), fieldName);
            document.addField(fieldName + getStringIndexingFieldSuffix(context), possiblyTruncatedValue);
            
            // Index without analyzing but lower-case (for wildcard queries).
            document.addField(fieldName + getStringWildcardFieldSuffix(context), possiblyTruncatedValue.toLowerCase());
            
            // Exact words tokenized by whitespace.
            document.addField(fieldName + "_s_ws", valueToIndex.toLowerCase());
            
            // Index with analyze (full-text search).
            document.addField(fieldName + "_txt", valueToIndex);
        }
        
        indexFulltextValue(document, rootObjectDocument, valueToIndex, context);
    }
    
    /**
     * Retrieves the indexing solr field suffix for type string
     * @param context the indexable data context
     * @return the indexing solr field suffix for type string
     */
    public static String getStringIndexingFieldSuffix(DataContext context)
    {
        return "_s";
    }
    
    /**
     * Retrieves the wildcard solr field suffix for type string
     * @param context the indexable data context
     * @return the wildcard solr field suffix for type string
     */
    public static String getStringWildcardFieldSuffix(DataContext context)
    {
        return "_s_lower";
    }
    
    /**
     * Retrieves the text solr field suffix for type string
     * @param context the indexable data context
     * @return the text solr field suffix for type string
     */
    public static String getStringTextFieldSuffix(DataContext context)
    {
        String language = context.getLocale().getLanguage();
        return "_txt_" + language;
    }
    
    /**
     * Checks if data of type string is facetable
     * @param context the data context
     * @return <code>true</code> if the type string is facetable, <code>false</code> otherwise
     */
    public static boolean isStringDataFacetable(DataContext context)
    {
        return isEnumerated(context);
    }
    
    /**
     * Checks if the data is enumerated (=> if its definition has an enumerator)
     * @param context the data context
     * @return <code>true</code> if the data is enumerated, <code>false</code> otherwise
     */
    public static boolean isEnumerated(DataContext context)
    {
        RepositoryDataContext repoContext = context instanceof RepositoryDataContext rc ? rc : RepositoryDataContext.newInstance(context);
        
        return Optional.ofNullable(context.getModelItem()
                                          .orElse(Optional.ofNullable(repoContext.getObject())
                                                          .filter(ModelAwareDataAwareAmetysObject.class::isInstance)
                                                          .map(ModelAwareDataAwareAmetysObject.class::cast)
                                                          .map(ametysObject -> ametysObject.getDefinition(context.getDataPath()))
                                                          .orElse(null)))
                       .filter(ElementDefinition.class::isInstance)
                       .map(ElementDefinition.class::cast)
                       .map(ElementDefinition::getEnumerator)
                       .isPresent();
    }
    
    /**
     * Index the given file value in a solr document.
     * @param document the solr document to index into.
     * @param rootObjectDocument the solr document of the root object.
     * @param fieldName the name of the indexing field
     * @param valueToIndex the value to index
     * @param context the context of the data to index. Can not be null
     * @param logger the logger to log in WARN level in case of an error occurs while indexing the given binary
     */
    public static void indexFileValue(SolrInputDocument document, SolrInputDocument rootObjectDocument, String fieldName, File valueToIndex, IndexableDataContext context, Logger logger)
    {
        if (!context.indexForFullTextField())
        {
            document.addField(fieldName + "_txt", valueToIndex.getName());
        }
        
        // Full text
        indexFulltextFileValue(document, rootObjectDocument, valueToIndex, context, logger);
    }
    
    /**
     * Index the given file value as full-text in a Solr document. The full-text field name is given by the context
     * @param document the solr document to index into.
     * @param rootObjectDocument the solr document of the root object.
     * @param valueToIndex the value to index
     * @param context the context of the data to index. Can not be null
     * @param logger the logger to log in WARN level in case of an error occurs while indexing the given binary
     */
    public static void indexFulltextFileValue(SolrInputDocument document, SolrInputDocument rootObjectDocument, File valueToIndex, IndexableDataContext context, Logger logger)
    {
        try (InputStream is = valueToIndex.getInputStream())
        {
            Tika tika = new Tika();
            String text = tika.parseToString(is);
            indexFulltextValue(document, rootObjectDocument, text, context);
        }
        catch (Exception e)
        {
            logger.warn("Failed to index file value in full-text field", e);
        }
    }
    
    /**
     * Index a full-text value.
     * @param currentDocument The current document, can be either the ametys object document itself or a repeater document.
     * @param ametysObjectDocument The parent ametys object document.
     * @param text The text to index.
     * @param context The context of the data to index.
     */
    public static void indexFulltextValue(SolrInputDocument currentDocument, SolrInputDocument ametysObjectDocument, String text, IndexableDataContext context)
    {
        _indexFulltextValue(currentDocument, ametysObjectDocument, text, context);
    }
    
    /**
     * Index a full-text value.
     * @param currentDocument The current document, can be either the ametys object document itself or a repeater document.
     * @param ametysObjectDocument The parent ametys object document.
     * @param text The text to index.
     * @param context The context of the data to index.
     */
    private static void _indexFulltextValue(SolrInputDocument currentDocument, SolrInputDocument ametysObjectDocument, String text, IndexableDataContext context)
    {
        // Index full text in the current document
        _indexFulltextValue(currentDocument, text, context);
        
        // If the current document is not the ametys object's one, index also in the ametys object document
        if (currentDocument != ametysObjectDocument)
        {
            _indexFulltextValue(ametysObjectDocument, text, context);
        }
    }
    
    /**
     * Index a full-text value.
     * @param document The document to index into.
     * @param text The text to index.
     * @param context The context of the data to index.
     */
    public static void indexFulltextValue(SolrInputDocument document, String text, IndexableDataContext context)
    {
        _indexFulltextValue(document, text, context);
    }
    
    /**
     * Index a full-text value.
     * @param document The document to index into.
     * @param text The text to index.
     * @param context The context of the data to index.
     */
    private static void _indexFulltextValue(SolrInputDocument document, String text, IndexableDataContext context)
    {
        if (StringUtils.isNotBlank(text))
        {
            String fullFieldName = context.getFullTextFieldName();
            
            document.addField(fullFieldName, text);
            document.addField(fullFieldName + SolrFieldNames.EXACT_WS_OPERATOR, text);
            
            if (context.getLocale() != null)
            {
                String language = context.getLocale().getLanguage();
                _indexLanguageFulltextValue(document, fullFieldName, text, language);
            }
            else
            {
                _languagesManager.getAvailableLanguages()
                    .keySet()
                    .stream()
                    .forEach(language -> _indexLanguageFulltextValue(document, fullFieldName, text, language));
            }
        }
    }
    
    private static void _indexLanguageFulltextValue(SolrInputDocument document, String fullFieldName, String text, String language)
    {
        document.addField(fullFieldName + SolrFieldNames.LANGUAGE_SEPARATOR + language, text);
        document.addField(fullFieldName + SolrFieldNames.STEMMED_OPERATOR + SolrFieldNames.LANGUAGE_SEPARATOR + language, text);
    }
}
