/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.indexing;

import java.util.Arrays;

import org.apache.solr.common.SolrInputDocument;

import org.ametys.runtime.model.type.DataContext;

/**
 * Interface for element types that can be indexed and with sort value
 * @param <T> Type of the element value
 */
public interface SortableIndexableElementType<T> extends IndexableElementType<T>
{
    @SuppressWarnings("unchecked")
    public default void indexValue(SolrInputDocument document, SolrInputDocument rootObjectDocument, String fieldName, Object value, IndexableDataContext context)
    {
        IndexableElementType.super.indexValue(document, rootObjectDocument, fieldName, value, context);

        if (value != null && !context.indexForFullTextField())
        {
            if (getManagedClass().isInstance(value))
            {
                indexSingleSortValue(document, fieldName, (T) value, context);
            }
            else if (getManagedClassArray().isInstance(value))
            {
                // index sort value only for the first occurrence
                Arrays.stream((T[]) value)
                      .findFirst()
                      .ifPresent(firstValue -> indexSingleSortValue(document, fieldName, firstValue, context));
            }
            else
            {
                throw new IllegalArgumentException("Try to convert the non " + getId() + " value '" + value + "' to JSON");
            }
        }
    }
    
    /**
     * Index the given single value in sort field in a solr document.
     * @param document the solr document to index into.
     * @param fieldName the name of the indexing field
     * @param value the single value to index
     * @param context the context of the data to index. Can not be null.
     */
    public default void indexSingleSortValue(SolrInputDocument document, String fieldName, T value, DataContext context)
    {
        String sortField = fieldName + getSortFieldSuffix();
        Object sortValue = getSingleSortValueToIndex(value);
        if (sortValue != null && !document.containsKey(sortField))
        {
            // FIXME If it is an enumerated date, shouldn't we index the label? In which language?
            document.addField(sortField, sortValue);
        }
    }
    
    /**
     * Get the suffix for sorting
     * @return The sort field suffix.
     */
    public String getSortFieldSuffix();
    
    /**
     * Retrieves the formatted value to index
     * @param <X> Type of the value to retrieve
     * @param value the value to index
     * @return the formatted value
     */
    public default <X> X getSingleSortValueToIndex(T value)
    {
        return getSingleValueToIndex(value);
    }
}
