/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.indexing.impl;

import java.util.Map;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.apache.solr.common.SolrInputDocument;

import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.holder.impl.IndexableDataHolderHelper;
import org.ametys.cms.data.type.AbstractContentElementType;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.indexing.IndexableDataContext;
import org.ametys.cms.data.type.indexing.SortableIndexableElementType;
import org.ametys.cms.repository.Content;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.type.DataContext;

/**
 * Class for content type of elements that can be indexed
 */
public class ContentIndexableElementType extends AbstractContentElementType implements SortableIndexableElementType<ContentValue>
{
    public void indexSingleValue(SolrInputDocument document, SolrInputDocument rootObjectDocument, String fieldName, ContentValue value, IndexableDataContext context)
    {
        String valueToIndex = getSingleValueToIndex(value);
        
        document.addField(fieldName + getIndexingFieldSuffix(context), valueToIndex);

        // Facets
        document.addField(fieldName + "_s_dv", valueToIndex);
    }
    
    public void indexSingleValueForFullTextField(SolrInputDocument document, SolrInputDocument rootObjectDocument, ContentValue value, IndexableDataContext context)
    {
        value.getContentIfExists().ifPresent(
            content ->
            {
                ViewItemAccessor viewItemAccessor = context.getViewItem()
                        .filter(ViewItemAccessor.class::isInstance)
                        .map(ViewItemAccessor.class::cast)
                        .filter(via -> !via.getViewItems().isEmpty())
                        .orElse(View.of(content.getModel(), "title"));
                
                IndexableDataHolderHelper.indexData(content, viewItemAccessor, document, rootObjectDocument, null, context);
            }
        );
    }
    
    @SuppressWarnings("unchecked")
    public <X> X getSingleValueToIndex(ContentValue value)
    {
        return (X) value.getContentId();
    }

    public void indexSingleSortValue(SolrInputDocument document, String fieldName, ContentValue value, DataContext context)
    {
        Optional<? extends Content> optContent = value.getContentIfExists();
        
        if (optContent.isPresent())
        {
            String sortField = fieldName + getSortFieldSuffix(); 
            Content content = optContent.get();
            if (_contentHelper.isMultilingual(content) && ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID.equals(content.getType(Content.ATTRIBUTE_TITLE).getId()))
            {
                Map<String, String> titleVariants = _contentHelper.getTitleVariants(content);
                for (String language : titleVariants.keySet())
                {
                    String langSortField = sortField + "_" + language;
                    String langSortValue = titleVariants.get(language);
                    if (StringUtils.isNotEmpty(langSortValue) && !document.containsKey(langSortField))
                    {
                        document.addField(langSortField, langSortValue);
                    }
                }
            }
            else
            {
                String sortValue = _contentHelper.getTitle(content);
                if (StringUtils.isNotEmpty(sortValue) && !document.containsKey(sortField))
                {
                    document.addField(sortField, sortValue);
                }
            }
        }
    }
    
    public String getSortFieldSuffix()
    {
        return "_s_sort";
    }
    
    public String getSchemaType()
    {
        return "string";
    }
    
    public String getIndexingFieldSuffix(DataContext context)
    {
        return "_s";
    }
    
    public boolean isFacetable(DataContext context)
    {
        return true;
    }
}
