/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.indexing.impl;

import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.apache.solr.common.SolrInputDocument;

import org.ametys.cms.content.indexing.solr.SolrIndexer;
import org.ametys.cms.data.type.AbstractMultilingualStringElementType;
import org.ametys.cms.data.type.indexing.IndexableDataContext;
import org.ametys.cms.data.type.indexing.IndexableElementTypeHelper;
import org.ametys.cms.data.type.indexing.SortableIndexableElementType;
import org.ametys.cms.languages.Language;
import org.ametys.cms.languages.LanguagesManager;
import org.ametys.plugins.repository.metadata.MultilingualString;
import org.ametys.plugins.repository.metadata.MultilingualStringHelper;
import org.ametys.runtime.model.type.DataContext;

/**
 * Class for multilingual string type of elements that can be indexed
 */
public class MultilingualStringIndexableElementType extends AbstractMultilingualStringElementType implements SortableIndexableElementType<MultilingualString>
{
    /** The manager for languages */
    protected LanguagesManager _languagesManager;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _languagesManager = (LanguagesManager) manager.lookup(LanguagesManager.ROLE);
    }
    
    public void indexSingleValue(SolrInputDocument document, SolrInputDocument rootObjectDocument, String fieldName, MultilingualString multilingualString, IndexableDataContext context)
    {
        Set<Locale> itemLocales = multilingualString.getLocales();
        List<String> appLanguages = _languagesManager.getAvailableLanguages()
                                                     .values()
                                                     .stream()
                                                     .map(Language::getCode)
                                                     .collect(Collectors.toList());
        
        for (String appLanguageCode : appLanguages)
        {
            Locale appLocale = new Locale(appLanguageCode);
            if (itemLocales.contains(appLocale))
            {
                String value = multilingualString.getValue(appLocale);
                
                String language = appLocale.getLanguage();
                
                document.addField(fieldName + "_txt_" + language, value);
                document.addField(fieldName + "_txt_stemmed_" + language, value);
                document.addField(fieldName + "_txt_ws_" + language, value);
                
                // Index without analyzing.
                String possiblyTruncatedValue = SolrIndexer.truncateUtf8StringValue(value, getLogger(), context.getObjectId().orElse(null), fieldName);
                document.addField(fieldName + "_" + language + getIndexingFieldSuffix(context), possiblyTruncatedValue);
                
                // Index without analyzing but lower-case (for wildcard queries).
                document.addField(fieldName + "_" + language + "_s_lower", possiblyTruncatedValue.toLowerCase());
                
                // Exact words tokenized by whitespace.
                document.addField(fieldName + "_" + language + "_s_ws", value.toLowerCase());
                
                // Index with analyze (full-text search).
                document.addField(fieldName + "_" + language + "_txt", value);
            
                IndexableElementTypeHelper.indexFulltextValue(document, rootObjectDocument, value, context);
            }
        }
    }
    
    public void indexSingleValueForFullTextField(SolrInputDocument document, SolrInputDocument rootObjectDocument, MultilingualString multilingualString, IndexableDataContext context)
    {
        Set<Locale> itemLocales = multilingualString.getLocales();
        List<String> appLanguages = _languagesManager.getAvailableLanguages()
                                                     .values()
                                                     .stream()
                                                     .map(Language::getCode)
                                                     .collect(Collectors.toList());
        
        for (String appLanguageCode : appLanguages)
        {
            Locale appLocale = new Locale(appLanguageCode);
            if (itemLocales.contains(appLocale))
            {
                String value = multilingualString.getValue(appLocale);
                IndexableElementTypeHelper.indexFulltextValue(document, rootObjectDocument, value, context);
            }
        }
        
    }
    
    public void indexSingleSortValue(SolrInputDocument document, String fieldName, MultilingualString multilingualString, DataContext context)
    {
        List<String> appLanguages = _languagesManager.getAvailableLanguages()
                                                     .values()
                                                     .stream()
                                                     .map(Language::getCode)
                                                     .collect(Collectors.toList());
        
        for (String appLanguageCode : appLanguages)
        {
            Locale appLocale = new Locale(appLanguageCode);
            
            // Need to index sort field for every language of application, even if metadata does not have value for the given language
            String value = MultilingualStringHelper.getValue(multilingualString, appLocale);
            String language = appLocale.getLanguage();
            
            String sortField = fieldName + "_" + language + getSortFieldSuffix();
            if (StringUtils.isNotEmpty(value) && !document.containsKey(sortField))
            {
                document.addField(sortField, value);
            }
        }
    }
    
    public String getSortFieldSuffix()
    {
        return "_s_sort";
    }
    
    public String getSchemaType()
    {
        return "string";
    }
    
    public String getIndexingFieldSuffix(DataContext context)
    {
        return "_s";
    }
}
