/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.indexing.impl;

import java.util.Optional;

import org.apache.solr.common.SolrInputDocument;

import org.ametys.cms.data.RichText;
import org.ametys.cms.data.type.AbstractRichTextElementType;
import org.ametys.cms.data.type.indexing.IndexableDataContext;
import org.ametys.cms.data.type.indexing.IndexableElementType;
import org.ametys.cms.data.type.indexing.IndexableElementTypeHelper;
import org.ametys.runtime.model.type.DataContext;

/**
 * Class for rich-text type of elements that can be indexed
 */
public class RichTextIndexableElementType extends AbstractRichTextElementType implements IndexableElementType<RichText>
{
    public void indexSingleValue(SolrInputDocument document, SolrInputDocument rootObjectDocument, String fieldName, RichText value, IndexableDataContext context)
    {
        String valueToIndex = getSingleValueToIndex(value);
        if (valueToIndex != null)
        {
            // Locale can be null for multilingual content
            if (context.getLocale() != null) 
            {
                // No enumerator: index as full-text.
                String language = context.getLocale().getLanguage();
                document.addField(fieldName + getTextFieldSuffix(context).get(), valueToIndex);
                document.addField(fieldName + getIndexingFieldSuffix(context), valueToIndex);
                document.addField(fieldName + "_txt_ws_" + language, valueToIndex);
            }
            
            // Index in the full-text value.
            IndexableElementTypeHelper.indexFulltextValue(document, rootObjectDocument, valueToIndex, context);
        }
        else
        {
            getLogger().warn("Failed to index rich-text '{}'", fieldName);
        }
    }
    
    public void indexSingleValueForFullTextField(SolrInputDocument document, SolrInputDocument rootObjectDocument, RichText value, IndexableDataContext context)
    {
        String valueToIndex = getSingleValueToIndex(value);
        if (valueToIndex != null)
        {
            IndexableElementTypeHelper.indexFulltextValue(document, rootObjectDocument, valueToIndex, context);
        }
        else
        {
            getLogger().warn("Failed to index rich-text for fulltext field '{}'", context.getFullTextFieldName());
        }
        
    }
    
    @SuppressWarnings("unchecked")
    public <X> X getSingleValueToIndex(RichText value)
    {
        return (X) toString(value);
    }

    public String getSchemaType()
    {
        return null;
    }
    
    public String getIndexingFieldSuffix(DataContext context)
    {
        String language = context.getLocale().getLanguage();
        return "_txt_stemmed_" + language;
    }
    
    public Optional<String> getTextFieldSuffix(DataContext context)
    {
        String language = context.getLocale().getLanguage();
        return Optional.of("_txt_" + language);
    }
}
