/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.dublincore;

import java.io.IOException;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.Date;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.generation.ServiceableGenerator;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.xml.sax.SAXException;

import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.dublincore.DublinCoreAwareAmetysObject;

/**
 * Generates Dublin Core metadata.
 */
public class DublinCoreMetadataGenerator extends ServiceableGenerator
{
    /** The Ametys object resolver. */
    protected AmetysObjectResolver _resolver;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _resolver = (AmetysObjectResolver) serviceManager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        String id = parameters.getParameter("id", "");
        try
        {
            AmetysObject object = _resolver.resolveById(id);
            if (object instanceof DublinCoreAwareAmetysObject)
            {
                DublinCoreAwareAmetysObject dcObject = (DublinCoreAwareAmetysObject) object;
                
                contentHandler.startDocument();
                
                AttributesImpl atts = new AttributesImpl();
                atts.addCDATAAttribute("id", id);
                
                XMLUtils.startElement(contentHandler, "object", atts);
                _saxDublinCoreMetadata(dcObject);
                XMLUtils.endElement(contentHandler, "object");
                
                contentHandler.endDocument();
            }
            else
            {
                throw new IllegalArgumentException("Object of id " + id + " is not Dublin Core aware.");
            }
        }
        catch (AmetysRepositoryException e)
        {
            getLogger().error("Error trying to get Dublin Core metadata of content " + id, e);
            throw new ProcessingException("Error trying to get Dublin Core metadata of content " + id, e);
        }
    }
    
    /**
     * SAX content Dublin Core metadata.
     * @param dcObject the Dublin Core object.
     * @throws SAXException if an error occurs while SAXing.
     */
    protected void _saxDublinCoreMetadata(DublinCoreAwareAmetysObject dcObject) throws SAXException
    {
        XMLUtils.startElement(contentHandler, "dublin-core-metadata");
        _saxIfNotNull("title", dcObject.getDCTitle());
        _saxIfNotNull("creator", dcObject.getDCCreator());
        _saxIfNotNull("subject", dcObject.getDCSubject());
        _saxIfNotNull("description", dcObject.getDCDescription());
        _saxIfNotNull("publisher", dcObject.getDCPublisher());
        _saxIfNotNull("contributor", dcObject.getDCContributor());
        _saxIfNotNull("date", dcObject.getDCDate());
        _saxIfNotNull("type", dcObject.getDCType());
        _saxIfNotNull("format", dcObject.getDCFormat());
        _saxIfNotNull("identifier", dcObject.getDCIdentifier());
        _saxIfNotNull("source", dcObject.getDCSource());
        _saxIfNotNull("language", dcObject.getDCLanguage());
        _saxIfNotNull("relation", dcObject.getDCRelation());
        _saxIfNotNull("coverage", dcObject.getDCCoverage());
        _saxIfNotNull("rights", dcObject.getDCRights());
        XMLUtils.endElement(contentHandler, "dublin-core-metadata");
    }
    
    /**
     * SAX string Dublin Core metadata.
     * @param name the metadata name.
     * @param value the metadata value.
     * @throws SAXException if an error occurs while SAXing.
     */
    protected void _saxIfNotNull(String name, String value) throws SAXException
    {
        if (value != null)
        {
            XMLUtils.createElement(contentHandler, name, value);
        }
    }
    
    /**
     * SAX string Dublin Core metadata.
     * @param name the metadata name.
     * @param values the metadata values.
     * @throws SAXException if an error occurs while SAXing.
     */
    protected void _saxIfNotNull(String name, String[] values) throws SAXException
    {
        if (values != null)
        {
            for (String value : values)
            {
                XMLUtils.createElement(contentHandler, name, value);
            }
        }
    }
    
    /**
     * SAX date Dublin Core metadata.
     * @param name the metadata name.
     * @param value the metadata value.
     * @throws SAXException if an error occurs while SAXing.
     */
    protected void _saxIfNotNull(String name, Date value) throws SAXException
    {
        if (value != null)
        {
            LocalDate ld = DateUtils.asLocalDate(value);
            XMLUtils.createElement(contentHandler, name, ld.format(DateTimeFormatter.ISO_LOCAL_DATE));
        }
    }

}
