/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.duplicate.contents;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.commons.lang3.tuple.Pair;

/**
 * Represents the global configuration of content duplicate detection configuration
 */
public class DuplicateContentConfiguration
{
    private Map<String, DuplicateContentTypeConfiguration> _contentTypes;
    private List<Pair<String, List<Object>>> _errorList = new ArrayList<>();
    private List<Pair<String, List<Object>>> _warnList = new ArrayList<>();

    /**
     * Create a new DuplicateContentConfiguration
     * @param configuration the configuration used to create the DuplicateContentConfiguration
     * @param duplicateContentsManager the duplicate contents manager, only use this parameter to get linked components
     * @throws ConfigurationException if the configuration is not valid.
     */
    public DuplicateContentConfiguration(Configuration configuration, DuplicateContentsManager duplicateContentsManager) throws ConfigurationException
    {
        _contentTypes = new HashMap<>();
        Configuration[] contentTypes = configuration.getChildren("content-type");
        if (contentTypes.length == 0)
        {
            _warnList.add(Pair.of("Missing duplicate-contents.xml file inside WEB-INF/param folder", List.of()));
        }
        for (Configuration config : contentTypes)
        {
            DuplicateContentTypeConfiguration duplicateContentTypeConfiguration = new DuplicateContentTypeConfiguration(config, duplicateContentsManager);
            if (!duplicateContentTypeConfiguration.getAttributeList().isEmpty())
            {
                _contentTypes.put(duplicateContentTypeConfiguration.getContentTypeId(), duplicateContentTypeConfiguration);
            }
            _errorList.addAll(duplicateContentTypeConfiguration.getErrors());
            _warnList.addAll(duplicateContentTypeConfiguration.getWarns());
        }
    }

    /**
     * Get the content type configuration map
     * @return the content type configuration map
     */
    public Map<String, DuplicateContentTypeConfiguration> getContentTypes()
    {
        return _contentTypes;
    }

    /**
     * Set the content type configuration map
     * @param contentTypes the content type configuration map
     */
    public void setContentTypes(Map<String, DuplicateContentTypeConfiguration> contentTypes)
    {
        this._contentTypes = contentTypes;
    }

    /**
     * Get the list of duplicate content types
     * @return the list of duplicate content types
     */
    public Set<String> getDuplicatesContentTypes()
    {
        return _contentTypes.keySet();
    }

    /**
     * Get the DuplicateContentTypeConfiguration related to a given content type
     * @param duplicateContentTypeId the id of the content type
     * @return the DuplicateContentTypeConfiguration 
     */
    public DuplicateContentTypeConfiguration get(String duplicateContentTypeId)
    {
        return _contentTypes.get(duplicateContentTypeId);
    }

    /**
     * get the list of errors
     * @return the error list
     */
    public List<Pair<String, List<Object>>> getErrors()
    {
        return _errorList;
    }

    /**
     * get the list of warns
     * @return the warn list
     */
    public List<Pair<String, List<Object>>> getWarns()
    {
        return _warnList;
    }
}
