/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.duplicate.contents;

import java.util.ArrayList;
import java.util.List;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.commons.lang3.tuple.Pair;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.duplicate.contents.attr.AbstractDuplicateAttributeConfiguration;
import org.ametys.cms.duplicate.contents.attr.DuplicateAttributeConfiguration;
import org.ametys.cms.duplicate.contents.attr.impl.BooleanDuplicateAttributeConfiguration;
import org.ametys.cms.duplicate.contents.attr.impl.ContentDuplicateAttributeConfiguration;
import org.ametys.cms.duplicate.contents.attr.impl.DateDuplicateAttributeConfiguration;
import org.ametys.cms.duplicate.contents.attr.impl.DateTimeDuplicateAttributeConfiguration;
import org.ametys.cms.duplicate.contents.attr.impl.DoubleDuplicateAttributeConfiguration;
import org.ametys.cms.duplicate.contents.attr.impl.LongDuplicateAttributeConfiguration;
import org.ametys.cms.duplicate.contents.attr.impl.StringDuplicateAttributeConfiguration;
import org.ametys.runtime.model.ModelItemGroup;
import org.ametys.runtime.model.type.ModelItemTypeConstants;

/**
 * Represents the configuration for a content type of content duplicate detection configuration
 */
public class DuplicateContentTypeConfiguration
{
        
    private String _contentTypeId;
    private List<DuplicateAttributeConfiguration> _attributeList = new ArrayList<>();
    private List<Pair<String, List<Object>>> _errorList = new ArrayList<>();
    private List<Pair<String, List<Object>>> _warnList = new ArrayList<>();
    
    /**
     * Create a new DuplicateContentTypeConfiguration
     * @param configuration the configuration used to create the DuplicateContentTypeConfiguration
     * @param duplicateContentsManager the duplicate contents manager, only use this parameter to get linked components
     * @throws ConfigurationException if the configuration is not valid.
     */
    public DuplicateContentTypeConfiguration(Configuration configuration, DuplicateContentsManager duplicateContentsManager) throws ConfigurationException
    {
        String contentTypeId;
        contentTypeId = configuration.getAttribute("id");
        
        if (duplicateContentsManager._cTypeEP.hasExtension(contentTypeId))
        {
            _contentTypeId = contentTypeId;
            
            ContentType contentType = duplicateContentsManager._cTypeEP.getExtension(contentTypeId);

            for (Configuration attributeConf : configuration.getChildren("attribute"))
            {
                DuplicateAttributeConfiguration duplicateAttributeConfiguration = _createDuplicateAttributeConfiguration(contentType, attributeConf);

                if (duplicateAttributeConfiguration != null)
                {
                    _attributeList.add(duplicateAttributeConfiguration);
                }
            }
        }
        else
        {
            _errorList.add(Pair.of("Content type not found for identifier '{}'. @{}", List.of(contentTypeId, configuration.getLocation())));
        }
    }

    private AbstractDuplicateAttributeConfiguration _createDuplicateAttributeConfiguration(ContentType contentType, Configuration attributeConfiguration) throws ConfigurationException
    {

        String path = attributeConfiguration.getAttribute("path");
        if (contentType.hasModelItem(path))
        {
            //ModelItemTypeConstants.REPEATER_TYPE_ID.equals
            boolean insideOfRepeater = false;

            ModelItemGroup parent = contentType.getModelItem(path).getParent();
            while (parent != null)
            {
                insideOfRepeater = insideOfRepeater || org.ametys.plugins.repository.data.type.ModelItemTypeConstants.REPEATER_TYPE_ID.equals(parent.getType().getId());
                parent = parent.getParent();
            }
            
            if (insideOfRepeater)
            {
                _warnList.add(Pair.of("Ignored attribute query for attribute with path: {}. Attributes inside of a repeater are not handled.", 
                        List.of(path)));
            }
            else
            {
                String attributeType = contentType.getModelItem(path).getType().getId();
                switch (attributeType)
                {
                    case ModelItemTypeConstants.STRING_TYPE_ID:
                        return new StringDuplicateAttributeConfiguration(attributeConfiguration, attributeType);
                    case org.ametys.cms.data.type.ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID:
                        return new ContentDuplicateAttributeConfiguration(attributeConfiguration, attributeType);
                    case ModelItemTypeConstants.DATE_TYPE_ID:
                        return new DateDuplicateAttributeConfiguration(attributeConfiguration, attributeType);
                    case ModelItemTypeConstants.DATETIME_TYPE_ID:
                        return new DateTimeDuplicateAttributeConfiguration(attributeConfiguration, attributeType);
                    case ModelItemTypeConstants.LONG_TYPE_ID:
                        return new LongDuplicateAttributeConfiguration(attributeConfiguration, attributeType);
                    case ModelItemTypeConstants.DOUBLE_TYPE_ID:
                        return new DoubleDuplicateAttributeConfiguration(attributeConfiguration, attributeType);
                    case ModelItemTypeConstants.BOOLEAN_TYPE_ID:
                        return new BooleanDuplicateAttributeConfiguration(attributeConfiguration, attributeType);
                    default:
                        _warnList.add(Pair.of("Ignored attribute query for attribute with path: {}. This attribute type is not handled for duplicate attributes: {}.", 
                                List.of(path, 
                                        contentType.getModelItem(path).getType())));
                        break;
                }
            }
        }
        else
        {
            _errorList.add(Pair.of("Attribute '{}' not found for content type '{}'. @{}"
                    , List.of(path
                    , contentType.getId()
                    , attributeConfiguration.getLocation())));
        }
        return null;
    }
    
    /**
     * Check if the content type has any attributes that require near duplicate check
     * @return true if the content type has any attributes that require near duplicate check
     */
    public boolean hasAnyNearDuplicateAttributes()
    {
        return _attributeList.stream()
            .anyMatch(DuplicateAttributeConfiguration::checkNearDuplicate);
    }
    
    /**
     * Get the content type id
     * @return the content type id
     */
    public String getContentTypeId()
    {
        return _contentTypeId;
    }
    
    /**
     * Get the attribute list
     * @return the attribute list
     */
    public List<DuplicateAttributeConfiguration> getAttributeList()
    {
        return _attributeList;
    }

    /**
     * get the list of errors
     * @return the error list
     */
    public List<Pair<String, List<Object>>> getErrors()
    {
        return _errorList;
    }

    /**
     * get the list of warns
     * @return the warn list
     */
    public List<Pair<String, List<Object>>> getWarns()
    {
        return _warnList;
    }

}
