/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.duplicate.contents;

import java.io.IOException;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.quartz.JobExecutionContext;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.duplicate.contents.DuplicateContentsManager.Status;
import org.ametys.cms.repository.Content;
import org.ametys.cms.schedule.AbstractSendingMailSchedulable;
import org.ametys.core.schedule.Schedulable;
import org.ametys.core.schedule.progression.ContainerProgressionTracker;
import org.ametys.core.ui.mail.StandardMailBodyHelper;
import org.ametys.core.ui.mail.StandardMailBodyHelper.MailBodyBuilder;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.i18n.I18nizableTextParameter;

/**
 * A {@link Schedulable} job for detecting duplicates.
 */
public class DuplicateContentsDetectionSchedulable extends AbstractSendingMailSchedulable
{

    private static final Object _HAS_CONFIGURATION_ERRORS = "hasConfigurationErrors";
    
    /** The duplicates manager */
    protected DuplicateContentsManager _duplicatesManager;

    /** The server base URL. */
    protected String _baseUrl;

    /** The content types helper */
    protected ContentTypesHelper _contentTypesHelper;
    
    /** The content type extension point */
    protected ContentTypeExtensionPoint _cTypeEP;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
        _duplicatesManager = (DuplicateContentsManager) manager.lookup(DuplicateContentsManager.ROLE);
        _cTypeEP = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
    }

    @Override
    public void initialize() throws Exception
    {
        super.initialize();
        _baseUrl = StringUtils.stripEnd(StringUtils.removeEndIgnoreCase(Config.getInstance().getValue("cms.url"), "index.html"), "/");
    }
    
    @Override
    protected void _doExecute(JobExecutionContext context, ContainerProgressionTracker progressionTracker) throws Exception
    {
        Map<String, Object> duplicatesInfo = _duplicatesManager.searchDuplicates();
        _duplicatesManager.logConfigurationErrors(getLogger());

        context.put(DuplicateContentsManager.NO_DUPLICATE_CONTENTS_CONTENT_TYPE_KEY, duplicatesInfo.get(DuplicateContentsManager.NO_DUPLICATE_CONTENTS_CONTENT_TYPE_KEY));
        context.put(DuplicateContentsManager.DUPLICATE_CONTENTS_KEY, duplicatesInfo.get(DuplicateContentsManager.DUPLICATE_CONTENTS_KEY));
        context.put(DuplicateContentsManager.NEAR_DUPLICATE_CONTENTS_KEY, duplicatesInfo.get(DuplicateContentsManager.NEAR_DUPLICATE_CONTENTS_KEY));
        context.put(DuplicateContentsManager.STATUS_KEY, duplicatesInfo.get(DuplicateContentsManager.STATUS_KEY));
        context.put(_HAS_CONFIGURATION_ERRORS, _duplicatesManager.hasConfigurationErrors());

        if ((boolean) duplicatesInfo.get(DuplicateContentsManager.NO_DUPLICATE_CONTENTS_CONTENT_TYPE_KEY))
        {
            throw new IllegalArgumentException("Missing or badly configured duplicate-contents.xml file inside WEB-INF/param folder");
        }
    }
    
    @Override
    protected I18nizableText _getSuccessMailSubject(JobExecutionContext context) throws Exception
    {
        @SuppressWarnings("unchecked")
        Map<Content, List<Content>> duplicatesMap = (Map<Content, List<Content>>) context.get(DuplicateContentsManager.DUPLICATE_CONTENTS_KEY);
        @SuppressWarnings("unchecked")
        Map<Content, List<Content>> nearDuplicatesMap = (Map<Content, List<Content>>) context.get(DuplicateContentsManager.NEAR_DUPLICATE_CONTENTS_KEY);
        int resultsCount = duplicatesMap.size() + nearDuplicatesMap.size();
        Map<String, I18nizableTextParameter> i18nParameters = new HashMap<>();
        i18nParameters.put("resultsCount", new I18nizableText(String.valueOf(resultsCount)));

        return new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_GLOBAL_DETECTION_MAIL_SUBJECT", i18nParameters);
    }
    
    @Override
    protected boolean _isMailBodyInHTML(JobExecutionContext context)
    {
        return true;
    }
    
    @Override
    protected String _getSuccessMailBody(JobExecutionContext context) throws Exception
    {
        @SuppressWarnings("unchecked")
        Map<Content, List<Content>> duplicatesMap = (Map<Content, List<Content>>) context.get(DuplicateContentsManager.DUPLICATE_CONTENTS_KEY);
        @SuppressWarnings("unchecked")
        Map<Content, List<Content>> nearDuplicatesMap = (Map<Content, List<Content>>) context.get(DuplicateContentsManager.NEAR_DUPLICATE_CONTENTS_KEY);
        @SuppressWarnings("unchecked")
        Map<Content, DuplicateContentsManager.Status> tooComplexMap = (Map<Content, DuplicateContentsManager.Status>) context.get(DuplicateContentsManager.STATUS_KEY);

        boolean hasConfigurationErrors = (boolean) context.get(_HAS_CONFIGURATION_ERRORS);
        
        return _createBody(duplicatesMap, nearDuplicatesMap, tooComplexMap, hasConfigurationErrors);        
        
    }

    /**
     * Create the body of the email
     * @param duplicatesMap map linking a content to his duplicates
     * @param nearDuplicatesMap map linking a content to his near duplicates
     * @param tooComplexMap map linking a content to his search status (used to know if a query failed because the fuzzy query was too complex for example)
     * @param hasErrors true if the configuration have some errors
     * @return the body
     * @throws IOException if failed to build HTML email body for report
     */
    protected String _createBody(Map<Content, List<Content>> duplicatesMap, Map<Content, List<Content>> nearDuplicatesMap, Map<Content, DuplicateContentsManager.Status> tooComplexMap, boolean hasErrors) throws IOException
    {
        MailBodyBuilder bodyBuilder = StandardMailBodyHelper.newHTMLBody()
            .withTitle(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_GLOBAL_DETECTION_MAIL_BODY_TITLE"));
        
        // Intro
        bodyBuilder.addMessage(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_GLOBAL_DETECTION_MAIL_BODY_INTRO"));
        
        // Duplicate and near duplicate lists
        if (duplicatesMap.isEmpty() && nearDuplicatesMap.isEmpty() && tooComplexMap.isEmpty())
        {
            bodyBuilder.addMessage(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_GLOBAL_DETECTION_MAIL_BODY_NO_DUPLICATE_CONTENTS"));
        }
        else
        {
            Map<String, I18nizableTextParameter> introI18nParameters = new HashMap<>();
            introI18nParameters.put("duplicateCount", new I18nizableText(String.valueOf(duplicatesMap.size())));
            introI18nParameters.put("nearDuplicateCount", new I18nizableText(String.valueOf(nearDuplicatesMap.size())));
            
            bodyBuilder.addMessage(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_GLOBAL_DETECTION_MAIL_BODY_DETECTED", introI18nParameters));
            
            bodyBuilder.addMessage(_displayContents(duplicatesMap, nearDuplicatesMap, tooComplexMap));
            
        }
        if (hasErrors)
        {
            bodyBuilder.addMessage(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_GLOBAL_DETECTION_MAIL_FOOTER_ERROR_OR_WARN"));
        }
        
        return bodyBuilder.build();
    }

    /**
     * Create StringBuilder used to display contents
     * @param duplicatesMap map linking a content to his duplicates
     * @param nearDuplicatesMap map linking a content to his near duplicates
     * @param tooComplexMap map linking a content to his search status (used to know if a query failed because the fuzzy query was too complex for example)
     * @return the StringBuilder used to display contents
     */
    protected String _displayContents(Map<Content, List<Content>> duplicatesMap, Map<Content, List<Content>> nearDuplicatesMap,
            Map<Content, DuplicateContentsManager.Status> tooComplexMap)
    {
        Set<String> contentTypes = _duplicatesManager.getContentTypeIds();

        StringBuilder body = new StringBuilder();
        body.append("<ul>");
        for (String contentTypeId : contentTypes)
        {
            Set<Content> tooComplexMapKeySet = tooComplexMap.entrySet()
                    .stream()
                    .filter(entry -> entry.getValue().equals(Status.TOO_COMPLEX))
                    .map(Entry::getKey)
                    .collect(Collectors.toSet());
            
            Set<Content> allMapKeySet = Stream.of(duplicatesMap.keySet(), nearDuplicatesMap.keySet(), tooComplexMapKeySet)
                    .flatMap(Collection::stream)
                    .filter(content -> _contentTypesHelper.isInstanceOf(content, contentTypeId))
                    .sorted(Comparator.comparing(Content::getTitle, String.CASE_INSENSITIVE_ORDER))
                    .collect(Collectors.toCollection(LinkedHashSet::new));
            if (!allMapKeySet.isEmpty())
            {
                body.append("<li>");
                ContentType contentType = _cTypeEP.getExtension(contentTypeId);
                body.append(_i18nUtils.translate(contentType.getLabel()));
                body.append("<ul>");
                // Duplicate list
                for (Content content : allMapKeySet)
                {
                    body.append(_displayDuplicateContent(duplicatesMap, nearDuplicatesMap, tooComplexMap, content));
                    
                    // Remove contents from maps, in case of contents with multiple content types appear twice
                    duplicatesMap.remove(content);
                    nearDuplicatesMap.remove(content);
                    tooComplexMap.remove(content);
                }
                body.append("</ul>");
                body.append("</li>");
            }
            
        }
        body.append("</ul>");
        return body.toString();
    }

    /**
     * Create StringBuilder used to display a content
     * @param duplicatesMap map linking a content to his duplicates
     * @param nearDuplicatesMap map linking a content to his near duplicates
     * @param tooComplexMap map linking a content to his search status (used to know if a query failed because the fuzzy query was too complex for example)
     * @param content the content
     * @return the StringBuilder used to display a content
     */
    protected StringBuilder _displayDuplicateContent(Map<Content, List<Content>> duplicatesMap, Map<Content, List<Content>> nearDuplicatesMap,
            Map<Content, DuplicateContentsManager.Status> tooComplexMap, Content content)
    {
        StringBuilder body = new StringBuilder();
        body.append("<li>");
        body.append(_renderContentTitle(content));
        
        body.append("<ul>");
        List<Content> duplicates = duplicatesMap.get(content);
        if (duplicates != null)
        {
            body.append("<li>");
            body.append(_i18nUtils.translate(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_GLOBAL_DETECTION_MAIL_BODY_LIST_DUPLICATE_CONTENTS")));
            body.append("<ul>");
            for (Content duplicate : duplicates)
            {
                body.append("<li>");
                body.append(_renderContentTitle(duplicate));
                body.append("</li>");
            }
            body.append("</ul>");
            body.append("</li>");
        }
        
        List<Content> nearDuplicates = nearDuplicatesMap.get(content);
        if (nearDuplicates != null)
        {
            body.append("<li>");
            body.append(_i18nUtils.translate(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_GLOBAL_DETECTION_MAIL_BODY_LIST_NEAR_DUPLICATE_CONTENTS")));
            body.append("<ul>");
            for (Content nearDuplicate : nearDuplicates)
            {
                body.append("<li>");
                body.append(_renderContentTitle(nearDuplicate));
                body.append("</li>");
            }
            body.append("</ul>");
            body.append("</li>");
        }
        
        if (tooComplexMap.get(content) == Status.TOO_COMPLEX)
        {
            body.append("<li>");
            body.append(_i18nUtils.translate(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_GLOBAL_DETECTION_MAIL_BODY_TOO_COMPLEX")));
            body.append("</li>");
        }
        
        body.append("</ul>");
        body.append("</li>");
        return body;
    }

    /**
     * Render the content title
     * @param content The content
     * @return the link
     */
    protected String _renderContentTitle(Content content)
    {
        return _renderCmsToolLink(content);
    }
    
    /**
     * Render the cms tool link
     * @param content The content
     * @return the link
     */
    protected String _renderCmsToolLink(Content content)
    {
        return "<a href=\"" + _baseUrl + "/index.html?uitool=uitool-content,id:%27" + content.getId() + "%27\">" + content.getTitle() + "</a>";
    }

    @Override
    protected I18nizableText _getErrorMailSubject(JobExecutionContext context) throws Exception
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_GLOBAL_DETECTION_ERROR_MAIL_SUBJECT");
    }
    
    @Override
    protected String _getErrorMailBody(JobExecutionContext context, Throwable throwable) throws Exception
    {
        MailBodyBuilder bodyBuilder = StandardMailBodyHelper.newHTMLBody()
                .withTitle(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_GLOBAL_DETECTION_ERROR_MAIL_BODY_TITLE"));
        
        if ((boolean) context.get(DuplicateContentsManager.NO_DUPLICATE_CONTENTS_CONTENT_TYPE_KEY))
        {
            bodyBuilder.withMessage(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_GLOBAL_DETECTION_MAIL_BODY_NO_CONTENT_TYPE"));
        }
        else
        {
            bodyBuilder.withMessage(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_GLOBAL_DETECTION_ERROR_MAIL_BODY"));
        }
        
        return bodyBuilder.build();
    }
}
