/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.duplicate.contents;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.collections.ListUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.solr.common.SolrException;
import org.slf4j.Logger;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.duplicate.contents.attr.DuplicateAttributeConfiguration;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.content.ContentSearcherFactory;
import org.ametys.cms.search.query.AndQuery;
import org.ametys.cms.search.query.ContentTypeQuery;
import org.ametys.cms.search.query.MatchAllQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.EmptyIterable;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Component able to detect duplicates (and near duplicates) for a given content.
 */
public class DuplicateContentsManager extends AbstractLogEnabled implements Component, Serviceable, Configurable
{
    
    /** The component role. */
    public static final String ROLE = DuplicateContentsManager.class.getName();

    /** key for duplicate contents list */
    public static final String DUPLICATE_CONTENTS_KEY = "duplicates";
    /** key for near duplicate contents list */
    public static final String NEAR_DUPLICATE_CONTENTS_KEY = "nearDuplicates";
    /** key for boolean to know whether there are some content types or not  */
    public static final String NO_DUPLICATE_CONTENTS_CONTENT_TYPE_KEY = "noDuplicatesContentType";
    /** key for boolean to know the query status  */
    public static final String STATUS_KEY = "status";
    
    /**
     * The status of the query
     */
    public enum Status
    {
        /** Successful query */
        SUCCESSFUL, 
        /** too complex query */
        TOO_COMPLEX, 
        /** empty query */
        EMPTY
    }

    /** The content searcher factory. */
    protected ContentSearcherFactory _contentSearcherFactory;

    /** The content helper */
    protected ContentHelper _contentHelper;
    
    /** The duplicate content description */
    protected DuplicateContentConfiguration _duplicateContentConfiguration;
    
    /** The content type extension point */
    ContentTypeExtensionPoint _cTypeEP;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _cTypeEP = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _contentSearcherFactory = (ContentSearcherFactory) manager.lookup(ContentSearcherFactory.ROLE);
        _contentHelper = (ContentHelper) manager.lookup(ContentHelper.ROLE);
    }
        
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        _duplicateContentConfiguration = new DuplicateContentConfiguration(configuration, this);
        logConfigurationErrors(getLogger());
    }

    /**
     * Get the data about duplicates and near duplicated for a given content
     * @param content The content
     * @return A map of data. key "duplicates" contains a list of the duplicates
     *         (id and label for each entry), and key "nearDuplicates" contains
     *         the near duplicates if requested (duplicates excluded).
     */
    @SuppressWarnings("unchecked")
    public Map<String, Object> searchDuplicates(Content content)
    {
        Map<String, Object> results = searchDuplicates(List.of(content));

        results.put(STATUS_KEY, ((Map<Content, Status>) results.get(STATUS_KEY)).get(content));
        results.put(DUPLICATE_CONTENTS_KEY, ((Map<Content, List<Content>>) results.get(DUPLICATE_CONTENTS_KEY)).getOrDefault(content, new ArrayList<>()));
        results.put(NEAR_DUPLICATE_CONTENTS_KEY, ((Map<Content, List<Content>>) results.get(NEAR_DUPLICATE_CONTENTS_KEY)).getOrDefault(content, new ArrayList<>()));
        return results;
    }

    /**
     * Get the data about duplicates and near duplicates for all contents that match the content types included in the configuration
     * @return the data about duplicates and near duplicates
     */
    public Map<String, Object> searchDuplicates()
    {
        Map<String, Object> results = new HashMap<>();
        // Get content given duplicate content types.
        Set<String> duplicatesContentTypes = _duplicateContentConfiguration.getDuplicatesContentTypes();
        
        if (duplicatesContentTypes.isEmpty())
        {
            results.put(NO_DUPLICATE_CONTENTS_CONTENT_TYPE_KEY, true);
            return results;
        }

        results.put(NO_DUPLICATE_CONTENTS_CONTENT_TYPE_KEY, false);
        AmetysObjectIterable<Content> contents = _getContents(duplicatesContentTypes);
        results.putAll(searchDuplicates(contents));
        return results;
    }
    
    /**
     * Get the data about duplicates and near duplicates for a list of contents
     * @param contents the contents to check
     * @return the data about duplicates and near duplicates
     */
    public Map<String, Object> searchDuplicates(Iterable<Content> contents)
    {
        Map<String, Object> results = new HashMap<>();

        Map<Content, List<Content>> duplicatesMap = new HashMap<>();
        Map<Content, List<Content>> nearDuplicatesMap = new HashMap<>();
        Map<Content, Status> statusMap = new HashMap<>();
        
        List<String> duplicatesFound = new LinkedList<>();
        
        for (Content content : contents)
        {
            if (!duplicatesFound.contains(content.getId()))
            {
                
                // Find the content types that will act as references to determine the duplicates attributes.
                String[] contentTypes = content.getTypes();
                Set<String> duplicateContentTypes = _addDuplicatesContentTypes(contentTypes);
                
                String[] mixinTypes = content.getMixinTypes();
                duplicateContentTypes.addAll(_addDuplicatesContentTypes(mixinTypes));
                

                // Duplicates attributes
                Set<DuplicateContentTypeConfiguration> duplicateContentTypeConfigurations = duplicateContentTypes.stream()
                    .map(duplicateCtype -> _duplicateContentConfiguration.get(duplicateCtype))
                    .collect(Collectors.toSet());
                
                // Search only if configuration is defined for the given content types
                if (!duplicateContentTypeConfigurations.isEmpty())
                {
                    try
                    {
                        // Search for duplicate contents
                        List<Content> duplicates = _getDuplicates(content, duplicateContentTypeConfigurations, false, contentTypes);
                        if (!duplicates.isEmpty())
                        {
                            duplicatesMap.put(content, duplicates);
                        }
                        duplicatesFound.addAll(duplicates.stream().map(Content::getId).collect(Collectors.toList()));
                        
                        // Search for near duplicate contents if needed (different query for near duplicates)
                        boolean checkNearDuplicates = duplicateContentTypeConfigurations.stream()
                                .anyMatch(DuplicateContentTypeConfiguration::hasAnyNearDuplicateAttributes);
                        if (checkNearDuplicates)
                        {
                            List<Content> nearDuplicates = _getDuplicates(content, duplicateContentTypeConfigurations, true, contentTypes);
                            statusMap.put(content, Status.SUCCESSFUL);
                            nearDuplicates = ListUtils.removeAll(nearDuplicates, duplicates);
                            if (!nearDuplicates.isEmpty())
                            {
                                nearDuplicatesMap.put(content, nearDuplicates);
                            }
                        }
                    }
                    catch (Exception e)
                    {
                        if (e instanceof SolrException && StringUtils.equals(((SolrException) e).getRootThrowable(), "org.apache.lucene.util.automaton.TooComplexToDeterminizeException"))
                        {
                            getLogger().warn("Fuzzy query too complex", e);
                            statusMap.put(content, Status.TOO_COMPLEX);
                        }
                        else
                        {
                            getLogger().error("Unable to query to the Solr server", e);
                        }
                    }
                }
            }
        }
        results.put(DUPLICATE_CONTENTS_KEY, duplicatesMap);
        results.put(NEAR_DUPLICATE_CONTENTS_KEY, nearDuplicatesMap);
        results.put(STATUS_KEY, statusMap);
        return results;
    }
    
    /**
     * Get the list of duplicates
     * @param content The content
     * @param duplicateContentTypeConfigurations the attribute list
     * @param nearDuplicates true to check for near duplicates
     * @param contentTypes the content types
     * @return list of duplicates 
     * @throws Exception if a problem occurs while searching for duplicates
     */
    protected List<Content> _getDuplicates(Content content, Set<DuplicateContentTypeConfiguration> duplicateContentTypeConfigurations, boolean nearDuplicates, String[] contentTypes) throws Exception
    {
        // Query
        List<Query> queries = _getDuplicatesQueries(content, duplicateContentTypeConfigurations, nearDuplicates, contentTypes);
        
        // Query building
        Query query = new AndQuery(queries);

        AmetysObjectIterable<Content> results;
        results = _contentSearcherFactory.create()
                .search(query);
        
        return results.stream()
                .filter(Predicate.not(content::equals))
                .sorted(Comparator.comparing(Content::getTitle, String.CASE_INSENSITIVE_ORDER))
                .collect(Collectors.toList());
    }

    /**
     * Get the list of queries to search for duplicates
     * @param content The content
     * @param duplicateContentTypeConfigurations the attribute list
     * @param nearDuplicates true to check for near duplicates
     * @param contentTypes the content types
     * @return list of duplicates 
     */
    protected List<Query> _getDuplicatesQueries(Content content, Set<DuplicateContentTypeConfiguration> duplicateContentTypeConfigurations, boolean nearDuplicates, String[] contentTypes)
    {
        // Query
        List<Query> queries = new LinkedList<>();

        // Content types (mixins are not added)
        for (String contentType : contentTypes)
        {
            queries.add(new ContentTypeQuery(contentType));
        }

        for (DuplicateContentTypeConfiguration duplicateContentTypeConfiguration : duplicateContentTypeConfigurations)
        {
            for (DuplicateAttributeConfiguration duplicateAttributeConfiguration : duplicateContentTypeConfiguration.getAttributeList())
            {
                String path = duplicateAttributeConfiguration.getPath();
                
                queries.add(duplicateAttributeConfiguration.getQuery(content.getValue(path), nearDuplicates));
            }
        }
        return queries;
    }
    
    /**
     * Get the configured duplicates content types
     * @param contentTypeIds The content type identifiers for which duplicates content types should be found
     * @return the duplicate content types
     */
    protected Set<String> _addDuplicatesContentTypes(String[] contentTypeIds)
    {
        Set<String> duplicateContentTypes = new HashSet<>();
        for (String contentTypeId : contentTypeIds)
        {
            if (_duplicateContentConfiguration.getDuplicatesContentTypes().contains(contentTypeId))
            {
                duplicateContentTypes.add(contentTypeId);
            }
            else
            {
                ContentType contentType = _cTypeEP.getExtension(contentTypeId);
                
                duplicateContentTypes.addAll(_addDuplicatesContentTypes(contentType.getSupertypeIds()));
            }
        }
        return duplicateContentTypes;
    }
            
    /**
     * Retrieves indexed contents that have at least one of the listed content types
     * @param contentTypes The desired content types
     * @return solr query results
     */
    protected AmetysObjectIterable<Content> _getContents(Collection<String> contentTypes)
    {
        try
        {
            return _contentSearcherFactory.create(contentTypes).search(new MatchAllQuery());
        }
        catch (Exception e) 
        {
            getLogger().error("Unable to query to the Solr server", e);
        }
        return new EmptyIterable<>();
    }
    
    /**
     * Get the list of configuration errors
     * @return the configuration error list
     */
    protected List<Pair<String, List<Object>>> _getConfigurationErrors()
    {
        return _duplicateContentConfiguration.getErrors();
    }

    /**
     * Get the list of configuration warns
     * @return the configuration warn list
     */
    protected List<Pair<String, List<Object>>> _getConfigurationWarns()
    {
        return _duplicateContentConfiguration.getWarns();
    }
    
    /**
     * Get the content types set
     * @return the content types set
     */
    public Set<String> getContentTypeIds()
    {
        return _duplicateContentConfiguration.getContentTypes()
                .keySet()
                .stream()
                .collect(Collectors.toSet());
    }
    
    /**
     * Log all errors of the configuration
     * @param logger the logger
     */
    public void logConfigurationErrors(Logger logger)
    {
        Pair<String, Object[]> configurationErrorsPair = _getConfigurationErrorsPair();
        if (!_getConfigurationErrors().isEmpty())
        {
            getLogger().error(configurationErrorsPair.getKey(), configurationErrorsPair.getValue());
        }
        else if (!_getConfigurationWarns().isEmpty())
        {
            getLogger().warn(configurationErrorsPair.getKey(), configurationErrorsPair.getValue());
        }
    }
    
    private Pair<String, Object[]> _getConfigurationErrorsPair()
    {
        List<Pair<String, List<Object>>> errorsMap = new ArrayList<>(_getConfigurationErrors());
        errorsMap.addAll(_getConfigurationWarns());
        StringBuilder errors = new StringBuilder();
        List<Object> parameters = new ArrayList<>();
        for (Pair<String, List<Object>> error : errorsMap)
        {
            errors.append(error.getKey()).append("\n");
            parameters.addAll(error.getValue());
        }
        return Pair.of(errors.toString(), parameters.toArray(Object[]::new));
    }
    
    /**
     * Returns <code>true</code> if there is at least one configuration error
     * @return <code>true</code> if there is at least one configuration error
     */
    public boolean hasConfigurationErrors()
    {
        return !_getConfigurationErrors().isEmpty() || !_getConfigurationWarns().isEmpty();
    }
}
