/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.duplicate.contents.attr;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;

import org.ametys.cms.search.query.AndQuery;
import org.ametys.cms.search.query.NotQuery;
import org.ametys.cms.search.query.Query;

/**
 * This abstract class represents a {@link DuplicateAttributeConfiguration}
 * @param <T> The type of the attribute
 */
public abstract class AbstractDuplicateAttributeConfiguration<T> implements DuplicateAttributeConfiguration<T>
{
    
    /** The component role. */
    public static final String ROLE = AbstractDuplicateAttributeConfiguration.class.getName();

    /** The type of the attribute. */
    protected String _type;

    /** The path of the attribute. */
    protected String _path;

    /** Defines this attribute should also check for near duplicates. */
    protected boolean _checkNearDuplicate;

    /**
     * Create a new DuplicateAttributeConfiguration
     * @param configuration the configuration used to create the DuplicateAttributeConfiguration
     * @param attributeType the attribute type id
     * @throws ConfigurationException if the configuration is not valid.
     */
    public AbstractDuplicateAttributeConfiguration(Configuration configuration, String attributeType) throws ConfigurationException
    {
        _path = configuration.getAttribute("path");
        _checkNearDuplicate = _computeCheckNearDuplicate(configuration);
        _type = attributeType;
    }

    /**
     * Check if this attribute should also check for near duplicates
     * @param configuration the configuration used to create the DuplicateAttributeConfiguration
     * @return true if this attribute should also check for near duplicates
     */
    protected boolean _computeCheckNearDuplicate(Configuration configuration)
    {
        return false;
    }

    @Override
    public String getType()
    {
        return _type;
    }

    @Override
    public void setType(String type)
    {
        this._type = type;
    }

    @Override
    public String getPath()
    {
        return _path;
    }

    @Override
    public void setPath(String path)
    {
        this._path = path;
    }

    @Override
    public boolean checkNearDuplicate()
    {
        return _checkNearDuplicate;
    }

    /**
     * Get the query to check if a field is filled
     * @return the query
     */
    protected abstract Query getExistQuery();

    /**
     * Get the query to check if other contents have the same value
     * @param value the value to check
     * @return the query
     */
    protected abstract Query getDuplicateQuery(T value);

    /**
     * Get the query to check if other contents have almost the same value
     * @param value the value to check
     * @return the query
     */
    protected Query getNearDuplicateQuery(T value)
    {
        return getDuplicateQuery(value);
    }

    /**
     * Get the query to find duplicate of this attributes
     * @param value the value of the attribute
     * @param nearDuplicateQuery true if the query should check near duplicates
     * @return the query to find duplicate of this attributes
     */
    protected Query getDuplicate(T value, boolean nearDuplicateQuery)
    {
        return nearDuplicateQuery ? getNearDuplicateQuery(value) : getDuplicateQuery(value);
    }

    @Override
    @SuppressWarnings("unchecked")
    public Query getQuery(Object value, boolean nearDuplicateQuery)
    {
        if (value == null || isEmpty(value))
        {
            return new NotQuery(getExistQuery());
        }
        
        Class<?> type = value.getClass();
        if (type.isArray())
        {
            if (((T[]) value).length == 0)
            {
                return new NotQuery(getExistQuery());
            }
            
            Set<Query> queries = new HashSet<>();
            for (T typedValue : (T[]) value)
            {
                queries.add(getDuplicate(typedValue, nearDuplicateQuery));
            }
            return new AndQuery(new ArrayList<>(queries));
        }
        
        return getDuplicate((T) value, nearDuplicateQuery);
    }

    /**
     * Check if the value is empty
     * @param value the value
     * @return true if the value is empty
     */
    protected boolean isEmpty(Object value)
    {
        return false;
    }
}
