/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.explorer;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.avalon.framework.thread.ThreadSafe;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceFactory;

import org.ametys.core.util.FilenameUtils;
import org.ametys.plugins.explorer.resources.Resource;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.version.VersionAwareAmetysObject;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * {@link SourceFactory} for {@link Resource}.
 */
public class AmetysResourceSourceFactory extends AbstractLogEnabled implements SourceFactory, ThreadSafe, Serviceable
{
    /** source scheme */
    public static final String RESOURCE_SCHEME = "ametys-resource";
    
    private AmetysObjectResolver _resolver;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    public Source getSource(String location, Map parameters) throws IOException, MalformedURLException
    {
        String uri = location.substring((RESOURCE_SCHEME + "://").length());
        
        // uri are like <version>@<resourcePath>
        int i = uri.indexOf('@');
        String version = i == -1 ? null : uri.substring(0, i);
        String path = uri.substring(i + 1);
        
        path = FilenameUtils.decode(path);
        String filename = org.apache.commons.io.FilenameUtils.getName(path);
        
        try
        {
            Resource resource = _resolver.resolveByPath(path);
            
            if (!StringUtils.isEmpty(version))
            {
                ((VersionAwareAmetysObject) resource).switchToRevision(version);
            }
    
            return new AmetysResourceSource(resource, filename, location);
        }
        catch (Exception e)
        {
            getLogger().debug("Unable to resolve resource for uri {}", location, e);
            
            // returns an unexisting Source
            return new AmetysResourceSource(null, filename, location);
        }
    }

    public void release(Source source)
    {
        // nothing to do
    }
}
