/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.explorer;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.cms.content.indexing.solr.SolrFieldNames;
import org.ametys.cms.search.Sort;
import org.ametys.cms.search.Sort.Order;
import org.ametys.cms.search.query.AndQuery;
import org.ametys.cms.search.query.DocumentTypeQuery;
import org.ametys.cms.search.query.DublinCoreQuery;
import org.ametys.cms.search.query.FilenameQuery;
import org.ametys.cms.search.query.FullTextQuery;
import org.ametys.cms.search.query.OrQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.query.ResourceLocationQuery;
import org.ametys.cms.search.query.UsersQuery;
import org.ametys.cms.search.solr.SearcherFactory;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.JSONUtils;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.plugins.explorer.ExplorerNode;
import org.ametys.plugins.explorer.resources.Resource;
import org.ametys.plugins.explorer.resources.generators.SearchGenerator;
import org.ametys.plugins.repository.AmetysObjectIterable;

/**
 * Search resources in a specific explorer node.
 */
public class SearchResourcesGenerator extends SearchGenerator
{
    /** The searcher factory. */
    protected SearcherFactory _searcherFactory;
    
    /** The JSON utils */
    protected JSONUtils _jsonUtils;
    
    /** The user helper */
    protected UserHelper _userHelper;
    
    @Override
    public void service(ServiceManager sManager) throws ServiceException
    {
        super.service(sManager);
        _searcherFactory = (SearcherFactory) sManager.lookup(SearcherFactory.ROLE);
        _jsonUtils = (JSONUtils) sManager.lookup(JSONUtils.ROLE);
        _userHelper = (UserHelper) manager.lookup(UserHelper.ROLE);
    }
    
    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        String id = request.getParameter("id");
        String rootId = request.getParameter("rootId");
        ExplorerNode node = _resolver.resolveById(id);
        
        Query query = getQuery(request, node, rootId);
        
        contentHandler.startDocument();
        XMLUtils.startElement(contentHandler, "resources");
        
        try
        {
            Sort sort = new Sort(SolrFieldNames.TITLE_SORT, Order.ASC);
            
            AmetysObjectIterable<Resource> resources = _searcherFactory.create()
                                                        .withQuery(query)
                                                        .withFilterQueries(new DocumentTypeQuery(SolrFieldNames.TYPE_RESOURCE))
                                                        .withSort(sort)
                                                        .search();
            
            for (Resource resource : resources)
            {
                saxResource(resource);
            }
        }
        catch (Exception e)
        {
            getLogger().error("Error searching resources in the explorer node " + id, e);
            throw new ProcessingException("Error searching resources in the explorer node " + id, e);
        }
        
        XMLUtils.endElement(contentHandler, "resources");
        contentHandler.endDocument();
    }
    
    /**
     * Build the query object corresponding to the user search.
     * @param request The user request.
     * @param explorerNode The explorer node to search into.
     * @param rootId The root ID, can be null to search all resources.
     * @return the query object.
     */
    protected Query getQuery(Request request, ExplorerNode explorerNode, String rootId)
    {
        List<Query> queries = new ArrayList<>();
        
        if (explorerNode != null || StringUtils.isNotEmpty(rootId))
        {
            String path = explorerNode == null ? "" : explorerNode.getExplorerPath();
            queries.add(new ResourceLocationQuery(rootId, path));
        }
        
        // File name
        String name = request.getParameter("name");
        if (StringUtils.isNotEmpty(name))
        {
            queries.add(new FilenameQuery(name));
        }
        
        // Full text
        String fulltext = request.getParameter("fulltext");
        if (StringUtils.isNotEmpty(fulltext))
        {
            queries.add(new FullTextQuery(fulltext, Operator.LIKE));
        }
        
        String keywords = request.getParameter("keywords");
        if (StringUtils.isNotEmpty(keywords))
        {
            Query subjectQuery = new DublinCoreQuery("subject", keywords);
            Query descQuery = new DublinCoreQuery("description", keywords);
            
            queries.add(new OrQuery(subjectQuery, descQuery));
        }
        
        String authorAsStr = request.getParameter("author");
        Map<String, Object> json = _jsonUtils.convertJsonToMap(authorAsStr);
        UserIdentity author = _userHelper.json2userIdentity(json);
        if (null != author)
        {
            queries.add(new UsersQuery(SolrFieldNames.RESOURCE_CREATOR, author));
        }
        
        return new AndQuery(queries);
    }
}
