/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.filter;

import java.util.List;
import java.util.Map;

import org.ametys.cms.repository.Content;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.query.SortCriteria;
import org.ametys.plugins.repository.query.expression.Expression;
import org.ametys.runtime.model.View;

/**
 * This interface represents a filter for {@link Content}
 */
public interface ContentFilter
{
    /**
     * The context language
     */
    public enum ContextLanguage 
    {
        /** Constant to search only in the current sitemap */
        CURRENT
        {
            @Override
            public String toString ()
            {
                return "current";
            }
        },
        /** Constant to search in all sitemap */
        ALL
        {
            @Override
            public String toString ()
            {
                return "all";
            }
        },
        /** Constant to search in other sitemaps */
        OTHERS
        {
            @Override
            public String toString ()
            {
                return "others";
            }
        }
    }
    
    /** The condition. */
    public enum Condition
    {
        /** Or condition */
        OR,
        /** And condition. */
        AND
    }
    
    /**
     * Get the filter id
     * @return The filter id
     */
    public String getId ();
    
    /**
     * Set the filter id.
     * @param id The id to set.
     */
    public void setId (String id);
    
    /**
     * Get the content types to match contents
     * @return The content types to match contents
     */
    public List<String> getContentTypes ();
    
    /**
     * Add a content type to the filter
     * @param cTypeId The content type id to add
     */
    public void addContentType (String cTypeId);
    
    /**
     * Get the content types to match contents
     * @return The content types to match contents
     */
    public Map<String, String> getMetadataValues ();
    
    /**
     * The metadata condition
     * @return The metadata condition
     */
    public Condition getMetadataCondition ();
    
    /**
     * Add a metadata filter
     * @param metadataId The metadata id.
     * @param value The value to test. Set null to test existence
     */
    public void addMetadata (String metadataId, String value);
    
    /**
     * Set the metadata condition
     * @param condition The condition to set
     */
    public void setMetadataCondition (Condition condition);
    
    /**
     * Get the current filter expression
     * @return The expression. Can be null.
     */
    public Expression getAdditionalFilterExpression();
    
    /**
     * Set the filter expression that should complete metadata filtering (by doing an AND between both)
     * @param expression The expression to set. Can be null to remove curren expression
     */
    public void setAdditionalFilterExpression (Expression expression);
    

    /**
     * Get the context language
     * @return The context language
     */
    public ContextLanguage getContextLanguage ();
    
    /**
     * Set the context language
     * @param context The context language to set
     */
    public void setContextLanguage (ContextLanguage context);
    
    /**
     * Get the name of {@link View} used for matching contents 
     * @return The name of {@link View} used for matching contents 
     */
    public String getView ();
    
    /**
     * Set the filter view
     * @param viewName The name of {@link View}
     */
    public void setView (String viewName);
    
    /**
     * Get the max number of results
     * @return The max number of results
     */
    public int getLength ();
    
    /**
     * Set the max number o results
     * @param length The max number of results
     */
    public void setLength (int length);
    
    /**
     * Get the sort criteria allows to sort query results.
     * @return the sort criteria
     */
    public SortCriteria getSortCriteria ();
    
    /**
     * Add a sort criteria
     * @param metadataId The metadata id to sort
     * @param ascending true to ascending sort
     * @param useLowerCase <code>true</code> for case insensitive order on a string property
     */
    public void addSortCriteria (String metadataId, boolean ascending, boolean useLowerCase);
    
    /**
     * Set the ametys object resolver to execute query
     * @param resolver The ametys object resolver to set
     */
    public void setResolver (AmetysObjectResolver resolver);
    
    /**
     * Get the contents matching this filter.
     * @return The matching contents.
     */
    public AmetysObjectIterable<Content> getMatchingContents ();
    
    /**
     * Get the contents matching this filter.
     * @param lang The current language.
     * @return The matching contents.
     */
    public AmetysObjectIterable<Content> getMatchingContents (String lang);
    
}
