/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.filter;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.excalibur.source.SourceResolver;

import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ContentTypeExpression;
import org.ametys.cms.repository.LanguageExpression;
import org.ametys.core.util.SizeUtils.ExcludeFromSizeCalculation;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.query.SortCriteria;
import org.ametys.plugins.repository.query.SortCriteria.SortCriterion;
import org.ametys.plugins.repository.query.expression.AndExpression;
import org.ametys.plugins.repository.query.expression.Expression;
import org.ametys.plugins.repository.query.expression.Expression.Operator;
import org.ametys.plugins.repository.query.expression.MetadataExpression;
import org.ametys.plugins.repository.query.expression.OrExpression;
import org.ametys.plugins.repository.query.expression.StringExpression;

/**
 *  This is the default implementation of a {@link ContentFilter}. The filter's property are set by setter function and constructor
 */
public class DefaultContentFilter implements ContentFilter
{
    /** The Ametys object resolver */
    @ExcludeFromSizeCalculation
    protected AmetysObjectResolver _resolver;
    /** The source resolver */
    @ExcludeFromSizeCalculation
    protected SourceResolver _srcResolver;
    
    /** The filter id */
    protected String _id;
    /** The list of content types to match */
    protected List<String> _contentTypes;
    /** The list of content languages to match */
    protected ContextLanguage _contextLang;
    /** The metadata to match */
    protected Map<String, String> _metadata;
    /** The metadata condition*/
    protected Condition _metadataCondition;
    /** The number max of results */
    protected int _length;
    /** The view */
    protected String _viewName;
    /** The sort criteria */
    protected SortCriteria _sortCriteria;
    /** The additional expression */
    protected Expression _additionalFilterExpression;
    /** The extension point for content types */
    @ExcludeFromSizeCalculation
    protected ContentTypeExtensionPoint _contentTypeEP;
    
    /**
     * Constructor
     */
    public DefaultContentFilter ()
    {
        // Empty
    }
    
    /**
     * Creates a new filter
     * @param id The filter unique identifier
     * @param resolver The ametys object resolver
     * @param contentTypeExtensionPoint The extension point for content types
     */
    public DefaultContentFilter(String id, AmetysObjectResolver resolver, ContentTypeExtensionPoint contentTypeExtensionPoint)
    {
        _id = id;
        _resolver = resolver;
        _contentTypeEP = contentTypeExtensionPoint;
    }
    
    /**
     * Creates a new filter from copy of another
     * @param id The filter unique identifier
     * @param originalFilter The original filter to be copied
     * @param resolver The ametys object resolver
     * @param contentTypeExtensionPoint The extension point for content types
     */
    public DefaultContentFilter(String id, DefaultContentFilter originalFilter, AmetysObjectResolver resolver, ContentTypeExtensionPoint contentTypeExtensionPoint)
    {
        _id = id;
        _resolver = resolver;
        _contentTypeEP = contentTypeExtensionPoint;
        _contentTypes = new ArrayList<>(originalFilter._contentTypes);
        _metadataCondition = originalFilter._metadataCondition;
        _contextLang = originalFilter._contextLang;
        _length = originalFilter._length;
        _viewName = originalFilter._viewName;
        _metadata = new HashMap<>(originalFilter._metadata);
        
        SortCriteria originalSC = originalFilter._sortCriteria;
        if (originalSC != null)
        {
            SortCriteria sortCriteria = new SortCriteria();
            for (SortCriterion criterion : originalFilter.getSortCriteria().getCriteria())
            {
                if (criterion.getMetadataPath() != null)
                {
                    sortCriteria.addCriterion(criterion.getMetadataPath(), criterion.isAscending(), criterion.isNormalizedSort());
                }
                else if (criterion.getJcrProperty() != null)
                {
                    sortCriteria.addCriterion(criterion.getJcrProperty(), criterion.isAscending(), criterion.isNormalizedSort());
                }
            }
            _sortCriteria = sortCriteria;
        }
        
        // FIXME The filter expressions should be copied here but it missing a Expression.clone() or Expression.copy() method to do it.
        // For now the filter expression is not used when this constructor is used, so we do nothing for now since filters will be refactored soon (for 4.3 ?)
    }
    
    @Override
    public List<String> getContentTypes()
    {
        return _contentTypes;
    }
    
    @Override
    public Map<String, String> getMetadataValues()
    {
        return _metadata;
    }
    
    @Override
    public Condition getMetadataCondition()
    {
        return _metadataCondition;
    }
    
    @Override
    public ContextLanguage getContextLanguage()
    {
        return _contextLang;
    }
    
    @Override
    public String getId()
    {
        return _id;
    }

    @Override
    public int getLength()
    {
        return _length;
    }

    @Override
    public String getView()
    {
        return _viewName;
    }
    
    @Override
    public SortCriteria getSortCriteria()
    {
        return _sortCriteria;
    }
    
    @Override
    public void addContentType(String cTypeId)
    {
        if (_contentTypes == null)
        {
            _contentTypes = new ArrayList<>();
        }
        _contentTypes.add(cTypeId);
    }

    @Override
    public void addMetadata(String metadataId, String value)
    {
        if (_metadata == null)
        {
            _metadata = new HashMap<>();
        }
        _metadata.put(metadataId, value);
    }
    
    @Override
    public void setMetadataCondition(Condition condition)
    {
        _metadataCondition = condition;
    }

    @Override
    public Expression getAdditionalFilterExpression()
    {
        return _additionalFilterExpression;
    }
    
    @Override
    public void setAdditionalFilterExpression(Expression expression)
    {
        _additionalFilterExpression = expression;
    }
    
    @Override
    public void setContextLanguage(ContextLanguage context)
    {
        _contextLang = context;
    }
    
    @Override
    public void setId(String id)
    {
        _id = id;
    }
    
    @Override
    public void setResolver (AmetysObjectResolver resolver)
    {
        _resolver = resolver;
    }


    @Override
    public void setLength(int length)
    {
        _length = length;
    }


    @Override
    public void setView(String viewName)
    {
        _viewName = viewName;
    }
    
    @Override
    public void addSortCriteria(String metadataId, boolean ascending, boolean useLowerCase)
    {
        if (_sortCriteria == null)
        {
            _sortCriteria = new SortCriteria();
        }
        _sortCriteria.addCriterion(metadataId, ascending, useLowerCase);
    }
    
    @Override
    public AmetysObjectIterable<Content> getMatchingContents()
    {
        String xpathQuery =  getXPathQuery();
        AmetysObjectIterable<Content> contents = _resolver.query(xpathQuery);
                
        return contents;
    }
    
    @Override
    public AmetysObjectIterable<Content> getMatchingContents(String lang)
    {
        String xpathQuery =  getXPathQuery(lang);
        AmetysObjectIterable<Content> contents = _resolver.query(xpathQuery);
        
        return contents;
    }
    
    /**
     * Creates the XPath query corresponding to this filter.
     * @return the created XPath query
     */
    public String getXPathQuery ()
    {
        Expression expr = getFilterExpression ();
        return org.ametys.plugins.repository.query.QueryHelper.getXPathQuery(null, "ametys:content", expr, _sortCriteria);
    }
    
    /**
     * Creates the XPath query corresponding to this filter.
     * @param lang The current language
     * @return the created XPath query
     */
    public String getXPathQuery (String lang)
    {
        Expression expr = getFilterExpression ();
        Expression langExpr = getContextLanguagesExpression(lang);
        expr = langExpr != null ? new AndExpression(expr, langExpr) : expr;
        
        return org.ametys.plugins.repository.query.QueryHelper.getXPathQuery(null, "ametys:content", expr, _sortCriteria);

    }
    
    /**
     * Get the expression corresponding to this filter
     * @return The created expression to match contents
     */
    protected Expression getFilterExpression ()
    {
        Expression expr = getContentTypesExpression ();
        Expression metadataExpr = getMetadataExpression();
        Expression additionalExpr = getAdditionalFilterExpression();
        
        return new AndExpression(expr, metadataExpr, additionalExpr);
    }
    
    /**
     * Get the expression corresponding to the filter's content types
     * @return The expression corresponding to the filter's content types
     */
    protected Expression getContentTypesExpression ()
    {
        Expression expr = null;
        
        List<Expression> cTypesExpr = new ArrayList<>();
        if (_contentTypes != null && _contentTypes.size() > 0)
        {
            for (String contentType : _contentTypes)
            {
                // Filter on content type and all its children
                Set<String> cTypeAndSubTypes = new LinkedHashSet<>();
                cTypeAndSubTypes.add(contentType);
                cTypeAndSubTypes.addAll(_contentTypeEP.getSubTypes(contentType));
                cTypesExpr.add(new ContentTypeExpression(Operator.EQ, cTypeAndSubTypes.toArray(new String[cTypeAndSubTypes.size()])));
            }
            
            expr = new OrExpression(cTypesExpr.toArray(new Expression[cTypesExpr.size()]));
        }
        
        return expr;
    }
    
    /**
     * Get the {@link Expression} associated with the given language context
     * @param lang The current language
     * @return a {@link Expression} associated with the given language context
     */
    protected Expression getContextLanguagesExpression (String lang)
    {
        if (lang == null)
        {
            return null;
        }
        
        Expression expr = null;
        
        if (ContextLanguage.CURRENT.equals(_contextLang))
        {
            expr = new LanguageExpression(Operator.EQ, lang);
        }
        else if (ContextLanguage.OTHERS.equals(_contextLang))
        {
            expr = new LanguageExpression(Operator.NE, lang);
        }
        
        return expr;
    }
    
    /**
     * Get the expression corresponding to the filter's tags
     * @return The expression corresponding to the filter's tags
     */
    protected Expression getMetadataExpression ()
    {
        Expression expr = null;
        
        List<Expression> metadataExpr = new ArrayList<>();
        
        if (_metadata != null && _metadata.size() > 0)
        {
            for (String metadataId : _metadata.keySet())
            {
                String value = _metadata.get(metadataId);
                metadataExpr.add(value != null ? new StringExpression(metadataId, Operator.EQ, value) : new MetadataExpression(metadataId));
            }
            
            if (_metadataCondition == Condition.OR)
            {
                expr = new OrExpression(metadataExpr.toArray(new Expression[metadataExpr.size()]));
            }
            else
            {
                expr = new AndExpression(metadataExpr.toArray(new Expression[metadataExpr.size()]));
            }
        }
            
        return expr;
    }
}


