/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.filter;

import java.time.LocalDate;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeParseException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.query.SortCriteria;
import org.ametys.plugins.repository.query.SortCriteria.SortCriterion;
import org.ametys.plugins.repository.query.expression.AndExpression;
import org.ametys.plugins.repository.query.expression.DateExpression;
import org.ametys.plugins.repository.query.expression.Expression;
import org.ametys.plugins.repository.query.expression.Expression.Operator;
import org.ametys.plugins.repository.query.expression.MetadataExpression;
import org.ametys.plugins.repository.query.expression.OrExpression;
import org.ametys.runtime.plugin.component.PluginAware;

/**
 * Class representing a content filter. <br>
 */
public class StaticContentFilter extends DefaultContentFilter implements Configurable, PluginAware, Serviceable
{
    /** The plugin name */
    protected String _pluginName;
    /** The feature name */
    protected String _featureName;
    
    /**
     * Constructor
     */
    public StaticContentFilter ()
    {
        // Empty construction needed for component
        super();
    }
    
    /**
     * Constructor
     * @param id The filter id
     * @param resolver The ametys object resolver
     * @param contentTypeExtensionPoint The extension point for content types
     */
    public StaticContentFilter(String id, AmetysObjectResolver resolver, ContentTypeExtensionPoint contentTypeExtensionPoint)
    {
        super(id, resolver, contentTypeExtensionPoint);
    }
    
    /**
     * Creates a new filter from copy of another
     * @param id The filter unique identifier
     * @param originalFilter The original filter to be copied
     * @param resolver The ametys object resolver
     * @param contentTypeExtensionPoint The extension point for content types
     */
    public StaticContentFilter(String id, StaticContentFilter originalFilter, AmetysObjectResolver resolver, ContentTypeExtensionPoint contentTypeExtensionPoint)
    {
        super(id, originalFilter, resolver, contentTypeExtensionPoint);
        
        _contentTypes = new ArrayList<>(originalFilter._contentTypes);
        _metadataCondition = originalFilter._metadataCondition;
        _contextLang = originalFilter._contextLang;
        _length = originalFilter._length;
        _viewName = originalFilter._viewName;
        _metadata = new HashMap<>(originalFilter._metadata);
        
        SortCriteria originalSC = originalFilter._sortCriteria;
        if (originalSC != null)
        {
            SortCriteria sortCriteria = new SortCriteria();
            for (SortCriterion criterion : originalFilter.getSortCriteria().getCriteria())
            {
                if (criterion.getMetadataPath() != null)
                {
                    sortCriteria.addCriterion(criterion.getMetadataPath(), criterion.isAscending(), criterion.isNormalizedSort());
                }
                else if (criterion.getJcrProperty() != null)
                {
                    sortCriteria.addCriterion(criterion.getJcrProperty(), criterion.isAscending(), criterion.isNormalizedSort());
                }
            }
            _sortCriteria = sortCriteria;
        }
        
        // FIXME The filter expressions should be copied here but it missing a Expression.clone() or Expression.copy() method to do it.
        // For now the filter expression is not used when this constructor is used, so we do nothing for now since filters will be refactored soon (for 4.3 ?)
    }
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _contentTypeEP = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    @Override
    public void setPluginInfo(String pluginName, String featureName, String id)
    {
        _pluginName = pluginName;
        _featureName = featureName;
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        _id = configuration.getAttribute("id");
        _contentTypes = _configureContentTypes(configuration.getChild("content-types", true));
        _viewName = configuration.getChild("view").getValue("main");
        _contextLang = _configureContextLanguage(configuration.getChild("context", true));
        _length = configuration.getChild("max-result", true).getValueAsInteger(Integer.MAX_VALUE);
        _metadata = _configureMetadata(configuration.getChild("metadata", true));
        _metadataCondition = Condition.valueOf(configuration.getChild("metadata", true).getAttribute("condition", "AND").toUpperCase());
        _additionalFilterExpression = _configureComplexMetadata(configuration.getChild("metadata"), _metadataCondition);
        _sortCriteria = _configureSortCriteria(configuration.getChild("sort-information"));
        
    }
    
    /**
     * Configure the content type ids
     * @param configuration The content types configuration
     * @return The set of content type ids
     * @throws ConfigurationException If an error occurs
     */
    protected List<String> _configureContentTypes(Configuration configuration) throws ConfigurationException
    {
        List<String> cTypes = new ArrayList<>();
        for (Configuration cType : configuration.getChildren("type"))
        {
            cTypes.add(cType.getAttribute("id"));
        }
        return cTypes;
    }
    
    /**
     * Configure simple metadata clauses (fixed string values).
     * @param configuration The metadata configuration
     * @return The metadata to filter by, as a Map of metadata name -&gt; value.
     * @throws ConfigurationException If an error occurs
     */
    protected Map<String, String> _configureMetadata(Configuration configuration) throws ConfigurationException
    {
        Map<String, String> metadata = new HashMap<>();
        for (Configuration elt : configuration.getChildren("metadata"))
        {
            String op = elt.getAttribute("operator", null);
            String type = elt.getAttribute("type", null);
            
            // Process only conditions without type and operator.
            if (op == null && type == null)
            {
                metadata.put(elt.getAttribute("id"), elt.getValue(""));
            }
        }
        return metadata;
    }
    
    /**
     * Configure complex metadata conditions (may filter on non-string metadata,
     * and not limited to equality.)
     * @param configuration The metadata conditions configuration.
     * @param metadataOperator the metadata operator.
     * @return An expression for complex metadata conditions.
     * @throws ConfigurationException If an error occurs
     */
    protected Expression _configureComplexMetadata(Configuration configuration, Condition metadataOperator) throws ConfigurationException
    {
        List<Expression> expressions = new ArrayList<>();
        
        for (Configuration elt : configuration.getChildren("metadata"))
        {
            String op = elt.getAttribute("operator", null);
            String type = elt.getAttribute("type", null);
            
            // Process only metadata with a type and operator.
            if (op != null && type != null)
            {
                String id = elt.getAttribute("id");
                String value = elt.getValue("");
                
                // Date expression.
                if (type.equals("date"))
                {
                    expressions.add(_getComplexDateExpression(id, op, value, elt));
                }
            }
            else if (op == null && type != null
                     || op != null && type == null)
            {
                throw new ConfigurationException("Both type and operator must be specified for complex metadata conditions.", elt);
            }
        }
        
        Expression[] exprArr = expressions.toArray(new Expression[expressions.size()]);
        
        Expression expr = null;
        if (metadataOperator == Condition.AND)
        {
            expr = new AndExpression(exprArr);
        }
        else if (metadataOperator == Condition.OR)
        {
            expr = new OrExpression(exprArr); 
        }
        
        return expr;
    }
    
    /**
     * Get a complex date metadata expression.
     * @param metadataName the metadata name.
     * @param operator the comparison operator.
     * @param value the compared value.
     * @param configuration the configuration being processed.
     * @return the date Expression.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected Expression _getComplexDateExpression(String metadataName, String operator, String value, Configuration configuration) throws ConfigurationException
    {
        int operatorOffset = 0;
        
        Operator op = Operator.EQ;
        if ("eq".equalsIgnoreCase(operator))
        {
            op = Operator.EQ;
        }
        else if ("gte".equalsIgnoreCase(operator))
        {
            op = Operator.GE;
        }
        else if ("gt".equalsIgnoreCase(operator))
        {
            op = Operator.GE;
            operatorOffset = 1;
        }
        else if ("lte".equalsIgnoreCase(operator))
        {
            op = Operator.LT;
            operatorOffset = 1;
        }
        else if ("lt".equalsIgnoreCase(operator))
        {
            op = Operator.LT;
        }
        else
        {
            throw new ConfigurationException("Invalid date operator: '" + operator + "'", configuration);
        }
        
        // Date value can be one of "now"/"today", "2013-09-07",
        // or "-7" to specify a date relative to the current date.
        Date date = null;
        Integer valueOffset = null;
        if ("now".equalsIgnoreCase(value) || "today".equalsIgnoreCase(value))
        {
            // "now" or "today".
            valueOffset = 0;
        }
        else
        {
            try
            {
                // Try to parse as a fixed date.
                date = Date.from(LocalDate.parse(value).atStartOfDay(ZoneId.systemDefault()).plusDays(operatorOffset).toInstant());
            }
            catch (DateTimeParseException | IllegalArgumentException e)
            {
                try
                {
                    // If not a fixed date, try to parse as an offset (signed integer).
                    valueOffset = Integer.parseInt(value);
                }
                catch (NumberFormatException nfe)
                {
                    // Ignore: leave date value null to throw an exception.
                }
            }
        }
        
        if (date != null)
        {
            // Fixed date: standard date expression.
            return new DateExpression(metadataName, op, date);
        }
        else if (valueOffset != null)
        {
            // Offset: dynamic date expression.
            int offset = operatorOffset + valueOffset;
            return new DynamicDateExpression(metadataName, op, offset);
        }
        else
        {
            throw new ConfigurationException("Invalid date value: '" + value + "'", configuration);
        }
    }
    
    /**
     * Configure the sort criteria
     * @param configuration The sort criteria configuration
     * @return The sort criteria
     * @throws ConfigurationException If an error occurs
     */
    protected SortCriteria _configureSortCriteria(Configuration configuration) throws ConfigurationException
    {
        SortCriteria sortCriteria = null;
        if (configuration != null)
        {
            sortCriteria = new SortCriteria();
            for (Configuration sort : configuration.getChildren("sort"))
            {
                sortCriteria.addCriterion(sort.getAttribute("metadataId"), sort.getAttributeAsBoolean("ascending", false), sort.getAttributeAsBoolean("lower-case", false));
            }
        }
        return sortCriteria;
    }
    
    /**
     * Configure the context language
     * @param configuration The context configuration
     * @return The context language
     * @throws ConfigurationException If an error occurs
     */
    protected ContextLanguage _configureContextLanguage (Configuration configuration) throws ConfigurationException
    {
        String context = configuration.getAttribute("lang", "current");
        if (context.equals(ContextLanguage.OTHERS.toString()))
        {
            return ContextLanguage.OTHERS;
        }
        else if (context.equals(ContextLanguage.ALL.toString()))
        {
            return ContextLanguage.ALL;
        }
        
        return ContextLanguage.CURRENT;
    }
    
    /**
     * Constructs an {@link Expression} representing a comparison between a metadata and the time the build() method is called.
     */
    public static class DynamicDateExpression implements Expression
    {
        
        private MetadataExpression _metadataExpr;
        private Operator _operator;
        /**
         * The offset value.
         */
        private int _offsetValue;
        
        /**
         * Create a comparison Expression relative to the current date, with an offset, normalizing the 
         * @param metadata the metadata name
         * @param operator the operator to make the comparison
         * @param offsetValue the offset value.
         */
        public DynamicDateExpression(String metadata, Operator operator, int offsetValue)
        {
            _operator = operator;
            _metadataExpr = new MetadataExpression(metadata);
            _offsetValue = offsetValue;
        }
        
        @Override
        public String build()
        {
            ZonedDateTime dateTime = LocalDate.now().atStartOfDay(ZoneId.systemDefault()).plusDays(_offsetValue);
            
            StringBuffer buff = new StringBuffer();
            
            buff.append(_metadataExpr.build()).append(' ').append(_operator).append(" xs:dateTime('");
            DateUtils.getISODateTimeFormatter().formatTo(dateTime, buff);
            buff.append("')");
            
            return buff.toString();
        }
        
    }
    
}
