/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.indexing;

import java.io.IOException;
import java.util.List;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.exception.ExceptionUtils;
import org.quartz.JobExecutionContext;

import org.ametys.cms.schedule.AbstractSendingMailSchedulable;
import org.ametys.core.schedule.Schedulable;
import org.ametys.core.schedule.progression.ContainerProgressionTracker;
import org.ametys.core.ui.mail.StandardMailBodyHelper;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * A {@link Schedulable} job for indexing all the workspaces.
 */
public class GlobalWorkspaceIndexerSchedulable extends AbstractSendingMailSchedulable
{
    /** The workspace indexer */
    protected WorkspaceIndexer _workspaceIndexer;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _workspaceIndexer = (WorkspaceIndexer) manager.lookup(WorkspaceIndexer.ROLE);
    }
    
    @Override
    protected void _doExecute(JobExecutionContext context, ContainerProgressionTracker progressionTracker) throws IndexingException
    {
        _workspaceIndexer.indexAllWorkspaces(progressionTracker);
    }

    @Override
    protected boolean _isMailBodyInHTML(JobExecutionContext context) throws Exception
    {
        return true;
    }
    
    @Override
    protected I18nizableText _getSuccessMailSubject(JobExecutionContext context)
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_SOLR_GLOBAL_INDEXATION_SUCCESS_MAIL_SUBJECT");
    }
    
    @Override
    protected String _getSuccessMailBody(JobExecutionContext context)
    {
        try
        {
            return StandardMailBodyHelper.newHTMLBody()
                    .withTitle(new I18nizableText("plugin.cms", "PLUGINS_CMS_SOLR_GLOBAL_INDEXATION_SUCCESS_MAIL_BODY_TITLE"))
                    .withMessage(new I18nizableText("plugin.cms", "PLUGINS_CMS_SOLR_GLOBAL_INDEXATION_SUCCESS_MAIL_BODY"))
                    .build();
        }
        catch (IOException e)
        {
            getLogger().warn("Failed to build HTML email body for indexation result. Fallback to no wrapped email", e);
            return _i18nUtils.translate(new I18nizableText("plugin.cms", "PLUGINS_CMS_SOLR_GLOBAL_INDEXATION_SUCCESS_MAIL_BODY"));
        }
        
    }
    
    @Override
    protected I18nizableText _getErrorMailSubject(JobExecutionContext context)
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_SOLR_GLOBAL_INDEXATION_ERROR_MAIL_SUBJECT");
    }
    
    @Override
    protected String _getErrorMailBody(JobExecutionContext context, Throwable throwable)
    {
        String cmsUrl = StringUtils.stripEnd(StringUtils.removeEndIgnoreCase(Config.getInstance().getValue("cms.url"), "index.html"), "/");
        String url = cmsUrl + "/_admin/";
        String error = ExceptionUtils.getStackTrace(throwable);
        
        try
        {
            return StandardMailBodyHelper.newHTMLBody()
                    .withTitle(new I18nizableText("plugin.cms", "PLUGINS_CMS_SOLR_GLOBAL_INDEXATION_ERROR_MAIL_BODY_TITLE"))
                    .withMessage(new I18nizableText("plugin.cms", "PLUGINS_CMS_SOLR_GLOBAL_INDEXATION_ERROR_MAIL_BODY", List.of(url)))
                    .withDetails(null, error, true)
                    .build();
        }
        catch (IOException e)
        {
            getLogger().warn("Failed to build HTML email body for indexation result. Fallback to no wrapped email", e);
            return _i18nUtils.translate(new I18nizableText("plugin.cms", "PLUGINS_CMS_SOLR_GLOBAL_INDEXATION_ERROR_MAIL_BODY"));
        }
    }
}
