/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.indexing.explorer;

import java.util.Map;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.commons.lang.BooleanUtils;

import org.ametys.cms.content.indexing.solr.SolrIndexer;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.indexing.IndexingException;
import org.ametys.core.observation.AsyncObserver;
import org.ametys.core.observation.Event;
import org.ametys.plugins.explorer.ObservationConstants;
import org.ametys.plugins.explorer.resources.Resource;
import org.ametys.plugins.explorer.resources.ResourceCollection;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.provider.RequestAttributeWorkspaceSelector;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Abstract observer in charge of indexing resources when created, modified, moved...
 */
public abstract class AbstractSolrIndexResourceObserver extends AbstractLogEnabled implements AsyncObserver, Contextualizable, Serviceable
{
    /** Request attribute disabling content indexation when set to <code>true</code>. */
    public static final String DISABLE_INDEXING_KEY = SolrIndexResourceObserver.class.getName() + "$disableIndexing";
    
    /** The Solr indexer. */
    protected SolrIndexer _solrIndexer;
    
    /** The AmetysObject resolver. */
    protected AmetysObjectResolver _resolver;
    
    /** The component context. */
    protected Context _context;
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) serviceManager.lookup(AmetysObjectResolver.ROLE);
        _solrIndexer = (SolrIndexer) serviceManager.lookup(SolrIndexer.ROLE);
    }
    
    @Override
    public boolean supports(Event event)
    {
        // If the resource is moved or renamed, the document stays the same, only its data changes
        // (there is no need to unindex a document and index a new one).
        return event.getId().equals(ObservationConstants.EVENT_RESOURCE_CREATED)
            || event.getId().equals(ObservationConstants.EVENT_RESOURCE_UPDATED)
            || event.getId().equals(ObservationConstants.EVENT_RESOURCE_RENAMED)
            || event.getId().equals(ObservationConstants.EVENT_RESOURCE_MOVED)
            || event.getId().equals(ObservationConstants.EVENT_COLLECTION_RENAMED)
            || event.getId().equals(ObservationConstants.EVENT_COLLECTION_MOVED);
    }
    
    @Override
    public int getPriority(Event event)
    {
        return MAX_PRIORITY + 3000;
    }
    
    @Override
    public void observe(Event event, Map<String, Object> transientVars) throws Exception
    {
        Request request = ContextHelper.getRequest(_context);
        Boolean disableIndexingAttr = (Boolean) request.getAttribute(DISABLE_INDEXING_KEY);

        if (ObserverHelper.isNotSuspendedObservationForIndexation() && BooleanUtils.isNotTrue(disableIndexingAttr))
        {
            String[] workspaces = getWorkspacesToIndex();
            for (String workspace : workspaces)
            {
                _index(event, workspace);
            }
        }
    }
    
    /**
     * Get the list of workspace to index
     * @return An array of workspace names
     */
    protected abstract String[] getWorkspacesToIndex();
    
    /**
     * Perform the indexation process
     * @param event The current event
     * @param workspaceName The workspace name for the indexation
     * @throws IndexingException if an error occurs during indexation.
     */
    protected void _index(Event event, String workspaceName) throws IndexingException
    {
        Request request = ContextHelper.getRequest(_context);
        String currentWsp = RequestAttributeWorkspaceSelector.getForcedWorkspace(request);
        
        try
        {
            // Force the workspace.
            RequestAttributeWorkspaceSelector.setForcedWorkspace(request, workspaceName);
            
            if (event.getId().equals(ObservationConstants.EVENT_RESOURCE_CREATED))
            {
                // Resources were created: index each one.
                @SuppressWarnings("unchecked")
                Map<String, Resource> resources = (Map<String, Resource>) event.getArguments().get(ObservationConstants.ARGS_RESOURCES);
                if (resources != null)
                {
                    for (String resourceId : resources.keySet())
                    {
                        Resource resource = _getHandledObject(resourceId);
                        if (resource != null)
                        {
                            onResourceCreated(resource, workspaceName);
                        }
                    }
                }
            }
            else if (event.getId().equals(ObservationConstants.EVENT_RESOURCE_UPDATED)
                    || event.getId().equals(ObservationConstants.EVENT_RESOURCE_RENAMED)
                    || event.getId().equals(ObservationConstants.EVENT_RESOURCE_MOVED))
            {
                // A resource was modified, renamed or moved: reindex the corresponding document.
                String resourceId = (String) event.getArguments().get(ObservationConstants.ARGS_ID);
                Resource resource = _getHandledObject(resourceId);
                if (resource != null)
                {
                    onResourceUpdated(resource, workspaceName);
                }
            }
            else if (event.getId().equals(ObservationConstants.EVENT_COLLECTION_RENAMED)
                    || event.getId().equals(ObservationConstants.EVENT_COLLECTION_MOVED))
            {
                // A collection was moved: recursively reindex all resources in it.
                String collectionId = (String) event.getArguments().get(ObservationConstants.ARGS_ID);
                
                ResourceCollection resourceCollection = _getHandledObject(collectionId);
                if (resourceCollection != null)
                {
                    onCollectionRenamedOrMoved(resourceCollection, workspaceName);
                }
            }
        }
        catch (Exception e)
        {
            String error = String.format("Failed to index some resources in workspace %s", workspaceName);
            getLogger().error(error, e);
            throw new IndexingException(error, e);
        }
        finally
        {
            // Restore context
            RequestAttributeWorkspaceSelector.setForcedWorkspace(request, currentWsp);
        }
    }
    
    private  <A extends AmetysObject> A _getHandledObject(String id)
    {
        if (_resolver.hasAmetysObjectForId(id))
        {
            A object = _resolver.resolveById(id);
            if (isHandledResource(object))
            {
                return object;
            }
        }
        
        return null;
    }

    /**
     * Returns true if it is a {@link Resource} or {@link ResourceCollection} handled by this observer
     * @param object The resource or collection
     * @return true if it is handled by this observer
     */
    protected abstract boolean isHandledResource(AmetysObject object);
    
    /**
     * Method called when a resource handled by this observer is created. Typically for indexing the created resource.
     * @param resource The created resource
     * @param workspaceName The workspace name
     * @throws Exception if an error occurs
     */
    protected abstract void onResourceCreated(Resource resource, String workspaceName) throws Exception;
    
    /**
     * Method called when a resource handled by this observer is updated, moved or renamed. Typically for indexing the updated resource.
     * @param resource The updated resource
     * @param workspaceName The workspace name
     * @throws Exception if an error occurs
     */
    protected abstract void onResourceUpdated(Resource resource, String workspaceName) throws Exception;
    
    /**
    * Method called when a resource collection handled by this observer is renamed or moved. Typically for indexing the children of the resource collection.
    * @param resourceCollection The resource collection
     * @param workspaceName The workspace name
    * @throws Exception if an error occurs
    */
    protected abstract void onCollectionRenamedOrMoved(ResourceCollection resourceCollection, String workspaceName) throws Exception;
}
