/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.indexing.solr;

import org.apache.commons.lang3.StringUtils;

import org.ametys.core.right.RightManager;
import org.ametys.runtime.util.Annotations;
import org.ametys.runtime.workspace.WorkspaceManager;

/**
 * Methods for handling influence of objects on Solr Acl cache.
 * <br>See {@link #isUninfluent}.
 */
public final class SolrAclCacheInfluence
{
    private SolrAclCacheInfluence()
    {
    }
    
    /**
     * Determines if any change of profile assignment on the given context object ensures that there is no indexed object
     * (neither the given object nor any other object in the whole application) that is affected in 
     * its own {@link RightManager#getReadAccessAllowedUsers read access} computing.
     * <br>i.e. the given object has no influence at all on the whole Solr Acl cache.
     * <br>
     * <br>See {@link SolrAclCacheUninfluentialObject}.
     * @param object The object
     * @return <code>true</code> if the given object is uninfluent on the Solr Acl cache.
     */
    public static boolean isUninfluent(Object object)
    {
        return _isWorkspaceContext(object)
                || Annotations.isAnnotationPresent(object.getClass(), SolrAclCacheUninfluentialObject.class);
    }
    
    private static boolean _isWorkspaceContext(Object object)
    {
        return object instanceof String
                && StringUtils.startsWithAny((String) object, _workspacesContexts());
    }
    
    private static String[] _workspacesContexts()
    {
        return WorkspaceManager.getInstance()
                .getWorkspaceNames()
                .stream()
                .map("/"::concat)
                .toArray(String[]::new);
    }
}

