/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.indexing.solr;

import java.io.IOException;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.solr.client.solrj.SolrServerException;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.content.indexing.solr.SolrIndexer;
import org.ametys.core.observation.ObservationManager;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Helper to control the Solr indexation
 */
public class SolrIndexHelper extends AbstractLogEnabled implements Serviceable, Component
{
    /** The Avalon Role */
    public static final String ROLE = SolrIndexHelper.class.getName();
    
    private ObservationManager _observationManager;
    private SolrIndexer _solrIndexer;

    public void service(ServiceManager manager) throws ServiceException
    {
        _observationManager = (ObservationManager) manager.lookup(ObservationManager.ROLE);
        _solrIndexer = (SolrIndexer) manager.lookup(SolrIndexer.ROLE);
    }
    
    /**
     * Add an argument to all events of the provided ids to disable the committing of content to Solr
     * @param handledEventIds the event ids that will not be committed
     */
    public void pauseSolrCommitForEvents(String[] handledEventIds)
    {
        _observationManager.addArgumentForEvents(handledEventIds, ObservationConstants.ARGS_CONTENT_COMMIT, false);
    }
    
    /**
     * Restart the committing of content to Solr for the provided event ids after committing all pending modification
     * @param handledEventIds the event ids to restart
     */
    public void restartSolrCommitForEvents(String[] handledEventIds)
    {
        _observationManager.removeArgumentForEvents(handledEventIds, ObservationConstants.ARGS_CONTENT_COMMIT);
        
        // Before trying to commit, be sure all the async observers of the current request are finished
        for (Future future : _observationManager.getFuturesForRequest())
        {
            try
            {
                future.get();
            }
            catch (ExecutionException | InterruptedException e)
            {
                getLogger().info("An exception occured when calling #get() on Future result of an observer." , e);
            }
        }
        
        // Commit all uncommitted changes
        try
        {
            _solrIndexer.commit();
            
            getLogger().debug("Copied contents are now committed into Solr.");
        }
        catch (IOException | SolrServerException e)
        {
            getLogger().error("Impossible to commit changes", e);
        }
    }
}
