/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.languages;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;

import org.ametys.core.ui.Callable;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Handle possible languages
 */
public class LanguagesManager implements Configurable, Component
{
    /** The component role */
    public static final String ROLE = LanguagesManager.class.getName();
    
    private Map<String, Language> _languages;

    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        _languages = new LinkedHashMap<>();
        
        for (Configuration lConf : configuration.getChildren("language"))
        {
            String code = lConf.getAttribute("code");
            Language language;
            
            if (lConf.getAttributeAsBoolean("i18n", true))
            {
                language = new Language(code, new I18nizableText("application", lConf.getValue(code)));
            }
            else
            {
                language = new Language(code, new I18nizableText(lConf.getValue(code)));
            }
            _languages.put(code, language);
        }
    }
    
    /**
     * Get the language by its code
     * @param code The language code
     * @return The language or <code>null</code> if not found
     */
    public Language getLanguage (String code)
    {
        return _languages.get(code);
    }
    
    /**
     * Get the list of possible languages
     * @return the association language-code Language object
     */
    public Map<String, Language> getAvailableLanguages()
    {
        return _languages;
    }
    
    /**
     * Get the list of available languages as a JSON object
     * @return The available languages
     */
    @Callable
    public List<Map<String, Object>> getAvailableLanguagesAsJsonObject()
    {
        List<Map<String, Object>> availableLanguages = new ArrayList<>();
        
        for (Language language : getAvailableLanguages().values())
        {
            availableLanguages.add(toJson(language));
        }
        
        return availableLanguages;
    }
    
    /**
     * Return the JSON representation of a language
     * @param language The language
     * @return The json object representing the language
     */
    protected Map<String, Object> toJson(Language language)
    {
        Map<String, Object> info = new HashMap<>();
        
        info.put("name", language.getCode());
        info.put("label", language.getLabel());
        
        return info;
    }
}
