/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.lock;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang.StringUtils;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.repository.Content;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.ui.Callable;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.lock.LockAwareAmetysObject;
import org.ametys.plugins.repository.lock.LockHelper;
import org.ametys.plugins.repository.lock.LockableAmetysObject;


/**
 * Unlock or lock one or several content(s) 
 */
public class LockContentManager extends AbstractLogEnabled implements Serviceable, Component
{
    
    /** The component role. */
    public static final String ROLE = LockContentManager.class.getName();
    
    /** The "unlock all" right ID. */
    public static final String UNLOCK_ALL_RIGHT = "CMS_Rights_UnlockAll";
    
    /** The rights manager */
    protected RightManager _rightManager;
    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /**  The current user provider.*/
    protected CurrentUserProvider _currentUserProvider;
    /** Helper for users */
    protected UserHelper _userHelper;
    /** The content helper */
    protected ContentHelper _contentHelper;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _currentUserProvider = (CurrentUserProvider) smanager.lookup(CurrentUserProvider.ROLE);
        _userHelper = (UserHelper) smanager.lookup(UserHelper.ROLE);
        _rightManager = (RightManager) smanager.lookup(RightManager.ROLE);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _contentHelper = (ContentHelper) smanager.lookup(ContentHelper.ROLE);
    }
    
    /**
     * Test if the current user can unlock the given locked object.<br>
     * A user can unlock a locked object if he is the lock owner, or if he has the right to unlock all objects.
     * @param object the locked object to test.
     * @return true if the current user can unlock the given locked object, false otherwise.
     */
    public boolean canUnlock(LockAwareAmetysObject object)
    {
        return canUnlock(object, _currentUserProvider.getUser());
    }
    
    /**
     * Test if a specific user can unlock the given locked object.<br>
     * A user can unlock a locked object if he is the lock owner, or if he has the right to unlock all objects.
     * @param object the locked object to test.
     * @param user the user.
     * @return true if the current user can unlock the given locked object, false otherwise.
     */
    public boolean canUnlock(LockAwareAmetysObject object, UserIdentity user)
    {
        return LockHelper.isLockOwner(object, user) || canUnlockAll(user);
    }
    
    /**
     * Test if the current user has the right to unlock all contents.
     * @return true if the current user has the right to unlock all contents, false otherwise.
     */
    public boolean canUnlockAll()
    {
        return canUnlockAll(_currentUserProvider.getUser());
    }
    
    /**
     * Test if a specific user has the right to unlock all contents.
     * @param user the user.
     * @return true if the user has the right to unlock all contents, false otherwise.
     */
    public boolean canUnlockAll(UserIdentity user)
    {
        return _rightManager.hasRight(_currentUserProvider.getUser(), UNLOCK_ALL_RIGHT, "/cms") == RightResult.RIGHT_ALLOW;
    }
    
    /**
     * Unlock or lock contents
     * @param contents the contents to unlock or lock
     * @param mode the mode ('lock' or 'unlock')
     * @return the result JSON map
     */
    @Callable
    public Map<String, Object> unlockOrLock(List<String> contents, String mode) 
    {
        Map<String, Object> result = new HashMap<>();
        
        if ("unlock".equals(mode))
        {
            result = unlock(contents);
        }
        else
        {
            result = lock(contents);
        }
        
        result.put("mode", mode);
        return result;
    }
    
    /**
     * Unlock contents
     * @param contents The ids of contents to unlock
     * @return the result JSON map
     */
    protected Map<String, Object> unlock(List<String> contents)
    {
        UserIdentity currentUser = _currentUserProvider.getUser();
        boolean canDelockAll = _rightManager.hasRight(currentUser, "CMS_Rights_UnlockAll", "/cms") == RightResult.RIGHT_ALLOW;
        
        List<String> unlockContentsId = new ArrayList<>();
        String unlockedContents = "";
        String stillLockedContents = "";
        String failLockedContents = "";
        
        for (String contentId : contents)
        {
            LockableAmetysObject lockableContent = (LockableAmetysObject) _resolver.resolveById(contentId);
            if (!lockableContent.isLocked())
            {
                getLogger().warn("The content '" + contentId + "' is not locked anymore.");
            }
            else if (!LockHelper.isLockOwner(lockableContent, currentUser) && !canDelockAll)
            {
                getLogger().warn("The user '" + currentUser + "' try to unlock content '" + contentId + "' but he is not the lock owner");
                
                if (stillLockedContents.length() > 0)
                {
                    stillLockedContents += ", ";
                }
                stillLockedContents += _contentHelper.getTitle((Content) lockableContent);
            }
            else
            {
                try
                {
                    lockableContent.unlock();
                    
                    unlockContentsId.add(contentId);
                    if (unlockedContents.length() > 0)
                    {
                        unlockedContents += ", ";
                    }
                    unlockedContents += _contentHelper.getTitle((Content) lockableContent);
                    if (getLogger().isDebugEnabled())
                    {
                        getLogger().debug("The user was unlocked content '" + contentId + "'");
                    }
                    
                }
                catch (AmetysRepositoryException e)
                {
                    getLogger().error("Unable to unlock content '" + contentId + "'", e);
                    
                    if (failLockedContents.length() > 0)
                    {
                        failLockedContents += ", ";
                    }
                    failLockedContents += _contentHelper.getTitle((Content) lockableContent);
                }
            }
        }
        
        Map<String, Object> result = new HashMap<>();
        
        result.put("unlocked-contents-id", StringUtils.join(unlockContentsId, ","));
        result.put("unlocked-contents", unlockedContents);
        result.put("still-locked-contents", stillLockedContents);
        result.put("fail-unlocked-contents", failLockedContents);
        
        return result;
    }
    
    /**
     * Lock contents
     * @param contents The ids of contents to lock
     * @return the result JSON map
     */
    protected Map<String, Object> lock (List<String> contents)
    {
        List<String> lockContentsId = new ArrayList<>();
        String lockedContents = "";
        String alreadyLockedContents = "";
        String failUnlockedContents = "";
        
        for (String contentId : contents)
        {
            LockableAmetysObject lockableContent = (LockableAmetysObject) _resolver.resolveById(contentId);
            if (lockableContent.isLocked() && !LockHelper.isLockOwner(lockableContent, _currentUserProvider.getUser()))
            {
                getLogger().warn("The content '" + contentId + "' is already locked.");
                if (alreadyLockedContents.length() > 0)
                {
                    alreadyLockedContents += ", ";
                }
                alreadyLockedContents += _contentHelper.getTitle((Content) lockableContent);
            }
            else if (!lockableContent.isLocked())
            {
                try
                {
                    lockableContent.lock();
                    lockContentsId.add(contentId);
                    if (lockedContents.length() > 0)
                    {
                        lockedContents += ", ";
                    }
                    lockedContents += _contentHelper.getTitle((Content) lockableContent);
                }
                catch (AmetysRepositoryException e)
                {
                    getLogger().error("Unable to lock content '" + contentId + "'", e);
                    if (failUnlockedContents.length() > 0)
                    {
                        failUnlockedContents += ", ";
                    }
                    failUnlockedContents += _contentHelper.getTitle((Content) lockableContent);
                }
            }
        }
        
        Map<String, Object> result = new HashMap<>();
        
        result.put("locked-contents-id", StringUtils.join(lockContentsId, ","));
        result.put("locked-contents", lockedContents);
        result.put("already-locked-contents", alreadyLockedContents);
        result.put("fail-locked-contents", failUnlockedContents);
        
        UserIdentity currentUser = _currentUserProvider.getUser();
        result.put("lockOwner", _userHelper.user2json(currentUser));
        
        return result;
    }
}
