/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.lock;

import java.io.IOException;
import java.time.Duration;
import java.time.LocalDateTime;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.xml.XMLUtils;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.repository.Content;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.util.cocoon.AbstractCurrentUserProviderServiceableGenerator;
import org.ametys.plugins.repository.lock.LockHelper;
import org.ametys.plugins.repository.lock.LockableAmetysObject;
import org.ametys.plugins.repository.lock.UnlockHelper;

/**
 * List contents that can be unlocked.
 */
public class UnlockGenerator extends AbstractCurrentUserProviderServiceableGenerator
{
    /** The unlock helper */
    protected UnlockHelper _unlockHelper;
    /** The rights manager */
    protected RightManager _rightManager;
    /** The users manager */
    protected UserManager _userManager;
    /** The content helper */
    protected ContentHelper _contentHelper;
    
    @Override
    public void service(ServiceManager sManager) throws ServiceException
    {
        super.service(sManager);

        _unlockHelper = (UnlockHelper) manager.lookup(UnlockHelper.ROLE);
        _rightManager = (RightManager) manager.lookup(RightManager.ROLE);
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
        _contentHelper = (ContentHelper) manager.lookup(ContentHelper.ROLE);
    }

    public void generate() throws IOException, SAXException, ProcessingException
    {
        contentHandler.startDocument();
        XMLUtils.startElement(contentHandler, "List");
        
        Map<LockableAmetysObject, LocalDateTime> lockedObjects = _getLockedObjects();
        
        _saxLockedContents(lockedObjects);
        
        XMLUtils.endElement(contentHandler, "List");
        contentHandler.endDocument();
    }
    
    /**
     * Get the locked objects.
     * @return the locked objects.
     */
    protected Map<LockableAmetysObject, LocalDateTime> _getLockedObjects()
    {
        return _unlockHelper.getLockedObjects();
    }
    
    /**
     * SAX the locked objects.
     * @param lockedObjects the locked objects.
     * @throws SAXException if an error occurs generating the XML data.
     */
    protected void _saxLockedContents(Map<LockableAmetysObject, LocalDateTime> lockedObjects) throws SAXException
    {
        boolean unlockingActivated = _unlockHelper.isUnlockingActivated();
    
        // Get the current user
        UserIdentity userIdentity = _getCurrentUser();
    
        boolean canSeeLocked = _rightManager.hasRight(userIdentity, "CMS_Rights_SeeLocked", "/cms") == RightResult.RIGHT_ALLOW;
        boolean canUnlockAll = _rightManager.hasRight(userIdentity, "CMS_Rights_UnlockAll", "/cms") == RightResult.RIGHT_ALLOW;
        long period = _unlockHelper.getTimeBeforeUnlock();
    
        for (LockableAmetysObject ametysObject : lockedObjects.keySet())
        {
            // Filter contents
            if (ametysObject instanceof Content)
            {
                UserIdentity lockOwner = ametysObject.getLockOwner();
                boolean isLockOwner = LockHelper.isLockOwner(ametysObject, userIdentity);
                
                if (isLockOwner || canSeeLocked)
                {
                    LocalDateTime lockTime = lockedObjects.get(ametysObject);
                    
                    AttributesImpl attrs = new AttributesImpl();
                    attrs.addAttribute("", "id", "id", "CDATA", ametysObject.getId());
                    attrs.addAttribute("", "title", "title", "CDATA", _contentHelper.getTitle((Content) ametysObject));
                    
                    if (unlockingActivated)
                    {
                        long unlockDuration = Duration.between(LocalDateTime.now(), lockTime.plusHours(period)).toMillis();
                        attrs.addAttribute("", "unlockTime", "unlockTime", "CDATA", Long.toString(unlockDuration));
                    }
                    Duration sinceDuration = Duration.between(lockTime, LocalDateTime.now());
                    attrs.addAttribute("", "since", "since", "CDATA", Long.toString(sinceDuration.toMillis()));
                    
                    User user = null;
                    if (lockOwner != null)
                    {
                        user = _userManager.getUser(lockOwner.getPopulationId(), lockOwner.getLogin());
                    }
                    String lockBy = user != null ? user.getFullName() + " (" + user.getIdentity().getLogin() + ")" : (lockOwner != null ? lockOwner.getLogin() : "Anonymous");
                    attrs.addAttribute("", "lockBy", "lockBy", "CDATA", lockBy);
                    
                    if (canUnlockAll || isLockOwner)
                    {
                        attrs.addAttribute("", "canUnlock", "canUnlock", "CDATA", "true");
                    }
                    
                    if (isLockOwner)
                    {
                        attrs.addAttribute("", "isLockOwner", "isLockOwner", "CDATA", "true");
                    }
                    
                    XMLUtils.createElement(contentHandler, "LockedContent", attrs);
                }
            }
        }
    }

}
