/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.model;

import org.ametys.plugins.repository.model.RepositoryDataContext;
import org.ametys.runtime.model.type.DataContext;

/**
 * Object that gives some context for cms data manipulation
 */
public class CMSDataContext extends RepositoryDataContext
{
    private int _richTextMaxLength;
    
    /**
     * Creates a new instance of a {@link CMSDataContext}
     */
    protected CMSDataContext()
    {
        // Empty constructor
    }

    /**
     * Creates a new instance of a {@link CMSDataContext} from another {@link DataContext}
     * @param context the data context to copy
     */
    protected CMSDataContext(DataContext context)
    {
        super(context);
        if (context instanceof CMSDataContext cmsContext)
        {
            withRichTextMaxLength(cmsContext.getRichTextMaxLength());
        }
    }
    
    /**
     * Creates a new instance of a {@link CMSDataContext}
     * @return the created instance
     */
    public static CMSDataContext newInstance()
    {
        return new CMSDataContext();
    }
    
    /**
     * Creates a new instance of a {@link CMSDataContext} from another {@link DataContext}.
     * It can be the same implementation or another one, but it will be casted to the current implementation.
     * @param context the data context to copy
     * @return the created instance
     */
    public static CMSDataContext newInstance(DataContext context)
    {
        return new CMSDataContext(context);
    }
    
    /**
     * Creates a new instance of a {@link CMSDataContext}, with the current context values
     * @return the created instance
     */
    @Override
    public CMSDataContext cloneContext()
    {
        return newInstance(this);
    }
    
    /**
     * Retrieves the rich text max length. If the max length is 0, the whole rich text content is kept
     * @return the rich text max length
     */
    public int getRichTextMaxLength()
    {
        return _richTextMaxLength;
    }
    
    /**
     * Sets the rich text max length
     * Set to 0 to keep the whole rich text content (default) 
     * @param <T> the type of the retrieved {@link DataContext}
     * @param richTextMaxLength the rich text max length to set
     * @return the current {@link CMSDataContext}
     */
    @SuppressWarnings("unchecked")
    public <T extends CMSDataContext> T withRichTextMaxLength(int richTextMaxLength)
    {
        _richTextMaxLength = richTextMaxLength;
        return (T) this;
    }
}
