/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.model;

import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.xml.XMLUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.model.restrictions.ContentRestrictedModelItemHelper;
import org.ametys.cms.model.restrictions.RestrictedModelItem;
import org.ametys.cms.model.restrictions.Restrictions;
import org.ametys.cms.repository.Content;
import org.ametys.plugins.repository.model.CompositeDefinition;
import org.ametys.runtime.model.DefinitionContext;

/**
 * Definition of a composite for content types attributes
 */
public class ContentRestrictedCompositeDefinition extends CompositeDefinition implements RestrictedModelItem<Content>
{
    private ContentRestrictedModelItemHelper _restrictedModelItemHelper;
    private Restrictions _restrictions;
    
    /**
     * Set the restrictions for the composite
     * @param restrictions the restrictions to set
     */
    public void setRestrictions(Restrictions restrictions)
    {
        _restrictions = restrictions;
    }

    public boolean canRead(Content content)
    {
        ContentRestrictedModelItemHelper restrictedModelItemHelper = _getRestrictedModelItemHelper();
        ContentType contentType = (ContentType) getModel();
        return restrictedModelItemHelper.canRead(content, this, _restrictions) && contentType.canRead(content, this);
    }

    public boolean canWrite(Content content)
    {
        ContentRestrictedModelItemHelper restrictedModelItemHelper = _getRestrictedModelItemHelper();
        ContentType contentType = (ContentType) getModel();
        return restrictedModelItemHelper.canWrite(content, this, _restrictions) && contentType.canWrite(content, this);
    }
    
    private ContentRestrictedModelItemHelper _getRestrictedModelItemHelper()
    {
        if (_restrictedModelItemHelper == null)
        {
            try
            {
                _restrictedModelItemHelper = (ContentRestrictedModelItemHelper) __serviceManager.lookup(ContentRestrictedModelItemHelper.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException("Unable to lookup after the restricted model item helper component", e);
            }
        }
        
        return _restrictedModelItemHelper;
    }
    
    @Override
    protected Map<String, Object> _toJSON(DefinitionContext context, boolean includeChildren) throws ProcessingException
    {
        Map<String, Object> result = super._toJSON(context, includeChildren);
        
        Content content = context.getObject()
                .filter(Content.class::isInstance)
                .map(Content.class::cast)
                .orElse(null);
        
        if (!canWrite(content))
        {
            result.put("can-not-write", true);
        }
        
        return result;
    }
    
    @Override
    public void toSAX(ContentHandler contentHandler, DefinitionContext context) throws SAXException
    {
        super.toSAX(contentHandler, context);
        
        Content content = context.getObject()
                .filter(Content.class::isInstance)
                .map(Content.class::cast)
                .orElse(null);
        
        if (!canWrite(content))
        {
            XMLUtils.createElement(contentHandler, "can-not-write", String.valueOf(true));
        }
    }
    
    @Override
    protected boolean _shouldJSONBeEmpty(DefinitionContext context)
    {
        return context.getObject()
                .filter(Content.class::isInstance)
                .map(Content.class::cast)
                .map(content -> !canRead(content))
                .orElse(false);
    }
}
