/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.model;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.cocoon.ProcessingException;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.data.ContentValue;
import org.ametys.core.util.XMLUtils;
import org.ametys.runtime.model.DefaultElementDefinition;
import org.ametys.runtime.model.DefinitionContext;
import org.ametys.runtime.model.ModelItem;

/**
 * Default implementation for element definition of type Content
 */
public class DefaultContentElementDefinition extends DefaultElementDefinition<ContentValue> implements ContentElementDefinition
{
    private ContentTypesHelper _contentTypesHelper;
    private ContentTypeExtensionPoint _contentTypeExtensionPoint;
    
    private String _contentTypeId;
    
    public String getContentTypeId()
    {
        return _contentTypeId;
    }
    
    public void setContentTypeId(String contentTypeId)
    {
        _contentTypeId = contentTypeId;
    }
    
    public Collection< ? extends ModelItem> getModelItems()
    {
        if (_contentTypeId != null && _getContentTypeExtensionPoint().hasExtension(_contentTypeId))
        {
            ContentType contentType = _getContentTypeExtensionPoint().getExtension(_contentTypeId);
            return contentType.getModelItems();
        }
        else
        {
            return Collections.singleton(_getContentTypesHelper().getTitleAttributeDefinition());
        }
    }
    
    @Override
    protected Map<String, Object> _toJSON(DefinitionContext context) throws ProcessingException
    {
        Map<String, Object> result = super._toJSON(context);
        result.put("contentType", _contentTypeId);
        result.put("canCreate", _getContentTypesHelper().hasRight(_contentTypeId));
        return result;
    }
    
    @Override
    public void toSAX(ContentHandler contentHandler, DefinitionContext context) throws SAXException
    {
        super.toSAX(contentHandler, context);
        XMLUtils.createElementIfNotNull(contentHandler, "contentType", _contentTypeId);
    }
    
    /**
     * Retrieves the {@link ContentTypesHelper}
     * @return the {@link ContentTypesHelper}
     */
    protected ContentTypesHelper _getContentTypesHelper()
    {
        if (_contentTypesHelper == null)
        {
            try
            {
                _contentTypesHelper = (ContentTypesHelper) __serviceManager.lookup(ContentTypesHelper.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException("Unable to lookup after the content types helper component", e);
            }
        }
        
        return _contentTypesHelper;
    }
    
    /**
     * Retrieves the {@link ContentTypeExtensionPoint} 
     * @return the {@link ContentTypeExtensionPoint}
     */
    protected ContentTypeExtensionPoint _getContentTypeExtensionPoint()
    {
        if (_contentTypeExtensionPoint == null)
        {
            try
            {
                _contentTypeExtensionPoint = (ContentTypeExtensionPoint) __serviceManager.lookup(ContentTypeExtensionPoint.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException("Unable to lookup after the content type extension point", e);
            }
        }
        
        return _contentTypeExtensionPoint;
    }
}
