/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.model.properties;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Stream;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.ProcessingException;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.ametysobject.ModelAwareDataAwareAmetysObject;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.model.ContentElementDefinition;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.core.util.XMLUtils;
import org.ametys.runtime.model.DefinitionContext;
import org.ametys.runtime.model.ModelItem;

/**
 * Abstract class for content property
 * @param <X> type of ametys object supported by this property
 */
public abstract class AbstractContentProperty<X extends ModelAwareDataAwareAmetysObject> extends AbstractProperty<ContentValue, X> implements ContentElementDefinition, Serviceable
{
    private ContentTypeExtensionPoint _contentTypeExtensionPoint;
    private ContentTypesHelper _contentTypesHelper;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    @Override
    protected String _getTypeId()
    {
        return ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID;
    }
    
    public String getCriterionWidget()
    {
        return "edition.select-content";
    }
    
    public void setContentTypeId(String contentTypeId)
    {
        throw new UnsupportedOperationException("Setting the content type is not supported on " + getClass() + ", it is defined by the property itself.");
    }
    
    public Collection< ? extends ModelItem> getModelItems()
    {
        if (_contentTypeExtensionPoint.hasExtension(getContentTypeId()))
        {
            ContentType contentType = _contentTypeExtensionPoint.getExtension(getContentTypeId());
            return contentType.getModelItems();
        }
        else
        {
            return Collections.singleton(_contentTypesHelper.getTitleAttributeDefinition());
        }
    }
    
    @Override
    protected Map<String, Object> _toJSON(DefinitionContext context) throws ProcessingException
    {
        Map<String, Object> result = super._toJSON(context);
        result.put("contentType", getContentTypeId());
        return result;
    }
    
    @Override
    public void toSAX(ContentHandler contentHandler, DefinitionContext context) throws SAXException
    {
        super.toSAX(contentHandler, context);
        XMLUtils.createElementIfNotNull(contentHandler, "contentType", getContentTypeId());
    }
    
    @Override
    public Object getValue(X dataHolder)
    {
        Set<? extends ModifiableContent> contents = _getLinkedContents(dataHolder);
        
        return isMultiple()
                ? _transformToContentValueArray(contents)
                : Optional.ofNullable(contents)
                    .map(Set::stream)
                    .orElseGet(Stream::empty)
                    .findFirst()
                    .orElse(null);
    }
    
    /**
     * Get the linked contents to index.
     * @param dataHolder The data holder
     * @return The linked contents
     */
    protected abstract Set<? extends ModifiableContent> _getLinkedContents(X dataHolder);
    
    private ContentValue[] _transformToContentValueArray(Set<? extends ModifiableContent> contents)
    {
        if (contents == null)
        {
            return null;
        }
        
        return contents.stream()
            .map(ContentValue::new)
            .toArray(ContentValue[]::new);
    }
}
