/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.model.properties;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;

import org.ametys.cms.data.ametysobject.ModelAwareDataAwareAmetysObject;
import org.ametys.cms.data.type.indexing.IndexableElementType;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.DefaultElementDefinition;
import org.ametys.runtime.model.type.ModelItemType;
import org.ametys.runtime.plugin.ExtensionPoint;
import org.ametys.runtime.plugin.component.PluginAware;

/**
 * Abstract class for single property
 * @param <T> type of the property values
 * @param <X> type of ametys object supported by this property
 */
public abstract class AbstractProperty<T, X extends ModelAwareDataAwareAmetysObject> extends DefaultElementDefinition<T> implements Property<T, X>, Configurable, PluginAware
{
    /** extension point containing the types available for this property */
    protected ExtensionPoint<ModelItemType> _availableTypesExtensionPoint;
    
    public void configure(Configuration configuration) throws ConfigurationException
    {
        setName(configuration.getAttribute(_getNameConfigurationAttribute()));
        setLabel(_parseI18nizableText(configuration, getPluginName(), "label"));
        setDescription(_parseI18nizableText(configuration, getPluginName(), "description"));
    }
    
    /**
     * Retrieves the name of the configuration attribute that contains the name of the property
     * @return the name of the configuration attribute that contains the name of the property
     */
    protected String _getNameConfigurationAttribute()
    {
        return "name";
    }

    /**
     * Parse an i18n text.
     * @param config the configuration to use.
     * @param pluginName the current plugin name.
     * @param name the child name.
     * @return the i18n text.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected I18nizableText _parseI18nizableText(Configuration config, String pluginName, String name) throws ConfigurationException
    {
        return I18nizableText.parseI18nizableText(config.getChild(name), "plugin." + pluginName, getName());
    }
    
    @Override
    public void setPluginInfo(String pluginName, String featureName, String id)
    {
        setPluginName(pluginName);
    }
    
    public boolean isEditable()
    {
        return false;
    }
    
    public void setAvailableTypeExtensionPoint(ExtensionPoint<ModelItemType> availableTypesExtensionPoint)
    {
        _availableTypesExtensionPoint = availableTypesExtensionPoint;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public IndexableElementType<T> getType()
    {
        return (IndexableElementType<T>) _availableTypesExtensionPoint.getExtension(_getTypeId());
    }
    
    /**
     * Retrieves the id of  the property's type
     * @return the id of  the property's type
     */
    protected abstract String _getTypeId();
}
