/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.model.properties;

import java.util.Map;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.solr.common.SolrInputDocument;

import org.ametys.cms.data.ametysobject.ModelAwareDataAwareAmetysObject;
import org.ametys.cms.data.type.indexing.IndexableDataContext;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Enumerator;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * Property referencing an {@link ElementDefinition}
 * @param <T> Type of the element value
 * @param <X> type of ametys object supported by this property
 */
public class ElementRefProperty<T, X extends ModelAwareDataAwareAmetysObject> extends AbstractProperty<T, X>
{
    private String _elementPath;
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        super.configure(configuration);
        configureReferencedElementPath(configuration);
    }
    
    @Override
    protected I18nizableText _parseI18nizableText(Configuration config, String pluginName, String name) throws ConfigurationException
    {
        // Do no use name as default value, to use the referenced element's one if the the concerned text is not filled
        return I18nizableText.parseI18nizableText(config.getChild(name), "plugin." + pluginName, (I18nizableText) null);
    }
    
    /**
     * Configures the path of the referenced element
     * @param configuration the property configuration
     * @throws ConfigurationException if an error occurred
     */
    protected void configureReferencedElementPath(Configuration configuration) throws ConfigurationException
    {
        _elementPath = configuration.getAttribute("path");
    }
    
    @Override
    public String getName()
    {
        String name = super.getName();
        if (name == null)
        {
            // If no name has been configured, use the referenced element's one
            name = getElementDefinition().getName();
        }
        
        return name;
    }
    
    @Override
    public I18nizableText getLabel()
    {
        I18nizableText label = super.getLabel();
        if (label == null)
        {
            label = getElementDefinition().getLabel();
        }
        
        return label;
    }
    
    @Override
    public I18nizableText getDescription()
    {
        I18nizableText description = super.getDescription();
        if (description == null)
        {
            description = getElementDefinition().getDescription();
        }
        
        return description;
    }
    
    @Override
    public Enumerator<T> getEnumerator()
    {
        Enumerator<T> enumerator = super .getEnumerator();
        if (enumerator == null)
        {
            enumerator = getElementDefinition().getEnumerator();
        }
        
        return enumerator;
    }
    
    /**
     * Retrieves the referenced element definition
     * @return the referenced element definition
     */
    @SuppressWarnings("unchecked")
    public ElementDefinition<T> getElementDefinition()
    {
        return (ElementDefinition<T>) getModel().getModelItem(getElementPath());
    }
    
    public Object getValue(X ametysObject)
    {
        return ametysObject.getValue(getElementPath(), true);
    }
    
    public void indexValue(SolrInputDocument document, X ametysObject, IndexableDataContext context)
    {
        super.indexValue(document, ametysObject, context.cloneContext().withModelItem(getElementDefinition()));
    }
    
    /**
     * Retrieves the path of the referenced element definition
     * @return the path of the referenced element definition
     */
    public String getElementPath()
    {
        return _elementPath;
    }
    
    /**
     * Set the path of the referenced element definition
     * @param elementPath the path to set
     */
    public void setElementPath(String elementPath)
    {
        _elementPath = elementPath;
    }
    
    @Override
    protected String _getTypeId()
    {
        return getElementDefinition().getType().getId();
    }
    
    public Enumerator getCriterionEnumerator(Configuration configuration, ThreadSafeComponentManager<Enumerator> enumeratorManager) throws ConfigurationException
    {
        return getElementDefinition().getCriterionEnumerator(configuration, enumeratorManager);
    }
    
    public String getCriterionWidget()
    {
        return getElementDefinition().getCriterionWidget();
    }
    
    public Map<String, I18nizableText> getCriterionWidgetParameters(Configuration configuration)
    {
        return getElementDefinition().getCriterionWidgetParameters(configuration);
    }
}
