/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.model.properties;

import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.solr.common.SolrInputDocument;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.data.ametysobject.ModelAwareDataAwareAmetysObject;
import org.ametys.cms.data.type.indexing.IndexableDataContext;
import org.ametys.cms.data.type.indexing.IndexableElementType;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ModelItemType;
import org.ametys.runtime.plugin.ExtensionPoint;

/**
 * Interface for a property
 * @param <T> type of the property values
 * @param <X> type of ametys object supported by this property
 */
public interface Property<T, X extends ModelAwareDataAwareAmetysObject> extends ElementDefinition<T>
{
    @Override
    public IndexableElementType<T> getType();
    
    /**
     * Set the extension point containing the types available for this property
     * @param availableTypesExtensionPoint the extension point to set
     */
    public void setAvailableTypeExtensionPoint(ExtensionPoint<ModelItemType> availableTypesExtensionPoint);
    
    /**
     * Get the property's value
     * @param ametysObject the ametysObject containing the property
     * @return the values to index.
     */
    public Object getValue(X ametysObject);
    
    /**
     * Generates SAX events for the property
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param ametysObject the ametysObject containing the property to SAX.
     * @param context the context of the data to SAX. Can not be null.
     * @throws SAXException if an error occurs during the SAX events generation
     */
    public default void valueToSAX(ContentHandler contentHandler, X ametysObject, DataContext context) throws SAXException
    {
        Object value = getValue(ametysObject);
        if (value != null)
        {
            getType().valueToSAX(contentHandler, getName(), value, context);
        }
    }
    
    /**
     * Convert the property into a JSON object
     * @param ametysObject the ametysObject containing the property to convert.
     * @param context the context of the data to convert. Can not be null.
     * @return The property as JSON
     */
    public default Object valueToJSON(X ametysObject, DataContext context)
    {
        Object value = getValue(ametysObject);
        return getType().valueToJSONForClient(value, context);
    }
    
    /**
     * Index the property in a solr document.
     * @param document the solr document to index into.
     * @param ametysObject the ametysObject containing the property to index.
     * @param context the context of the data to index. Can not be null.
     */
    public default void indexValue(SolrInputDocument document, X ametysObject, IndexableDataContext context)
    {
        Object value = getValue(ametysObject);
        getType().indexValue(document, document, getName(), value, context);
    }
    
    public default String getCriterionWidget()
    {
        return null;
    }
    
    public default Map<String, I18nizableText> getCriterionWidgetParameters(Configuration configuration)
    {
        return new HashMap<>();
    }
}
