/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.model.properties;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.solr.common.SolrInputDocument;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.RichTextAttributeDefinition;
import org.ametys.cms.data.RichText;
import org.ametys.cms.data.type.indexing.IndexableDataContext;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.systemprop.AbstractSystemProperty;
import org.ametys.core.util.JSONUtils;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ModelItemTypeConstants;

/**
 * Property for semantic annotations.
 * Do not use the property in search model. There is no search field nor query, the criteria would not work
 */
public class SemanticAnnotationSystemProperty extends AbstractSystemProperty<String, Content>
{
    /** Prefix for annotations */
    public static final String ANNOTATION_PREFIX = "annotation-";
    
    /** The JSON utils */
    protected JSONUtils _jsonUtils;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _jsonUtils = (JSONUtils) manager.lookup(JSONUtils.ROLE);
    }
    
    @Override
    public boolean isMultiple()
    {
        return true;
    }
    
    @Override
    public boolean isSortable()
    {
        return false;
    }
    
    @Override
    public boolean isDisplayable()
    {
        return false;
    }
    
    @Override
    public Object getValue(Content content)
    {
        List<String> values = new ArrayList<>();
        
        Map<String, List<String>> semanticAnnotations = _getSemanticAnnotations(content);
        for (String annotationName : semanticAnnotations.keySet())
        {
            String value = annotationName + ":" + _jsonUtils.convertObjectToJson(semanticAnnotations.get(annotationName));
            values.add(value);
        }
        
        return values.toArray(new String[values.size()]);
    }
    
    public Object valueToJSON(Content content, DataContext context)
    {
        return _getSemanticAnnotations(content);
    }
    
    public void valueToSAX(ContentHandler contentHandler, Content content, DataContext context) throws SAXException
    {
        Map<String, List<String>> semanticAnnotations = _getSemanticAnnotations(content);
        for (String annotationName : semanticAnnotations.keySet())
        {
            List<String> semanticAnnotation = semanticAnnotations.get(annotationName);
            String[] semanticAnnotationAsStringArray = semanticAnnotation.toArray(new String[semanticAnnotation.size()]);
            getType().valueToSAX(contentHandler, annotationName, semanticAnnotationAsStringArray, context);
        }
    }
    
    @Override
    public void indexValue(SolrInputDocument document, Content content, IndexableDataContext context)
    {
        Map<String, List<String>> semanticAnnotations = _getSemanticAnnotations(content);
        for (String annotationName : semanticAnnotations.keySet())
        {
            List<String> semanticAnnotation = semanticAnnotations.get(annotationName);
            String[] semanticAnnotationAsStringArray = semanticAnnotation.toArray(new String[semanticAnnotation.size()]);
            getType().indexValue(document, document, annotationName, semanticAnnotationAsStringArray, context);
        }
    }
    
    private Map<String, List<String>> _getSemanticAnnotations(Content content)
    {
        // Search for all rich text definitions
        Set<RichTextAttributeDefinition> richTextDefinitions = new HashSet<>();
        for (Model model : content.getModel())
        {
            ModelHelper.findModelItemsByType(model, org.ametys.cms.data.type.ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID)
                       .stream()
                       .filter(RichTextAttributeDefinition.class::isInstance)
                       .map(RichTextAttributeDefinition.class::cast)
                       .forEach(richTextDefinitions::add);
        }
        
        // For each richText definition, get the rich text values
        Map<String, List<String>> semanticAnnotations = new HashMap<>();
        for (RichTextAttributeDefinition richTextDefinition : richTextDefinitions)
        {
            Object value = content.getValue(richTextDefinition.getPath(), true);
            List<RichText> richTexts = null;
            if (value instanceof RichText richText)
            {
                richTexts = List.of(richText);
            }
            else if (value instanceof RichText[] richTextsArray)
            {
                richTexts = Arrays.asList(richTextsArray);
            }
            
            if (richTexts != null)
            {
                for (RichText richText : richTexts)
                {
                    Map<String, List<String>> richTextAnnotations = richText.getAllAnnotations();
                    for (String richTextAnnotationName : richTextAnnotations.keySet())
                    {
                        String prefixedAnnotationName = ANNOTATION_PREFIX + richTextAnnotationName;
                        List<String> semanticAnnotationValues = semanticAnnotations.computeIfAbsent(prefixedAnnotationName, __ -> new ArrayList<>());
                        semanticAnnotationValues.addAll(richTextAnnotations.get(richTextAnnotationName));
                    }
                }
            }
        }
        
        return semanticAnnotations;
    }
    
    @Override
    protected String _getTypeId()
    {
        return ModelItemTypeConstants.STRING_TYPE_ID;
    }

    public Query getQuery(Object value, Operator operator, String language, Map<String, Object> contextualParameters)
    {
        return null;
    }

    public SearchField getSearchField()
    {
        return null;
    }
}
