/*
 *  Copyright 2011 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository;

import javax.jcr.ItemExistsException;
import javax.jcr.ItemNotFoundException;
import javax.jcr.Node;
import javax.jcr.NodeIterator;
import javax.jcr.PathNotFoundException;
import javax.jcr.Property;
import javax.jcr.PropertyIterator;
import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Session;
import javax.jcr.Value;
import javax.jcr.version.VersionHistory;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.thread.ThreadSafe;
import org.apache.cocoon.util.HashUtil;
import org.apache.commons.collections.Predicate;
import org.apache.commons.collections.PredicateUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.jackrabbit.core.NodeImpl;

import org.ametys.cms.support.AmetysPredicateUtils;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.collection.AmetysObjectCollectionFactory;
import org.ametys.plugins.repository.jcr.NameHelper;

/**
 * Helper for common processing used while synchronizing.
 */
public class CloneComponent extends AbstractLogEnabled implements Component, ThreadSafe
{
    /** Avalon Role */
    public static final String ROLE = CloneComponent.class.getName();
    
    /**
     * Adds a node to a parent node using a source node for name, type
     * and potential UUID.
     * @param srcNode the source node.
     * @param parentNode the parent node for the newly created node.
     * @param nodeName the node name to use.
     * @return the created node.
     * @throws RepositoryException if an error occurs.
     */
    public Node addNodeWithUUID(Node srcNode, Node parentNode, String nodeName) throws RepositoryException
    {
        if (srcNode.isNodeType("nt:frozenNode"))
        {
            String nodeTypeName = srcNode.getProperty("jcr:frozenPrimaryType").getString();
            String uuid = null;
            
            if (srcNode.hasProperty("jcr:frozenUuid"))
            {
                uuid = srcNode.getProperty("jcr:frozenUuid").getString();
            }
            
            if (uuid == null)
            {
                return parentNode.addNode(nodeName, nodeTypeName);
            }
            else
            {
                return ((NodeImpl) parentNode).addNodeWithUuid(nodeName, nodeTypeName, uuid);
            }
        }
        else
        {
            if (!srcNode.isNodeType("mix:referenceable"))
            {
                return parentNode.addNode(nodeName, srcNode.getPrimaryNodeType().getName());
            }
            else
            {
                return ((NodeImpl) parentNode).addNodeWithUuid(nodeName, srcNode.getPrimaryNodeType().getName(), srcNode.getIdentifier());
            }
        }
    }
    
    /**
     * Clones all the properties of a node.
     * @param srcNode the source node.
     * @param clonedNode the cloned node.
     * @param propertyPredicate the property selector.
     * @throws RepositoryException if an error occurs.
     */
    public void cloneAllProperties(Node srcNode, Node clonedNode, Predicate propertyPredicate) throws RepositoryException
    {
        // First remove existing matching properties from cloned Node
        PropertyIterator clonedProperties = clonedNode.getProperties();
        while (clonedProperties.hasNext())
        {
            Property property = clonedProperties.nextProperty();
            if (AmetysPredicateUtils.ignoreProtectedProperties(propertyPredicate).evaluate(property))
            {
                property.remove();
            }
        }
        
        // Then copy properties
        PropertyIterator itProperties = srcNode.getProperties();
        
        while (itProperties.hasNext())
        {
            Property property =  itProperties.nextProperty();
            
            // Ignore protected properties
            if (AmetysPredicateUtils.ignoreProtectedProperties(propertyPredicate).evaluate(property))
            {
                cloneProperty(clonedNode, property);
            }
        }
    }
    
    /**
     * Clone a property.
     * @param clonedNode the node to copy the property to.
     * @param property the property to clone.
     * @throws RepositoryException if an error occurs.
     */
    protected void cloneProperty(Node clonedNode, Property property) throws RepositoryException
    {
        Session session = clonedNode.getSession();
        
        if (property.getType() == PropertyType.REFERENCE)
        {
            try
            {
                // Clone the referenced nodes.
                if (property.isMultiple())
                {
                    Value[] sourceValues = property.getValues();
                    Value[] clonedValues = new Value[sourceValues.length];
                    
                    for (int i = 0; i < sourceValues.length; i++)
                    {
                        Node referencedNode = session.getNodeByIdentifier(sourceValues[i].getString());
                        
                        Node clonedReferencedNode = null;
                        try
                        {
                            clonedReferencedNode = clonedNode.getSession().getNodeByIdentifier(referencedNode.getIdentifier());
                        }
                        catch (ItemNotFoundException e)
                        {
                            clonedReferencedNode = getOrCloneNode(clonedNode.getSession(), referencedNode);
                        }
                        
                        clonedValues[i] = session.getValueFactory().createValue(clonedReferencedNode);
                    }
                    
                    clonedNode.setProperty(property.getName(), clonedValues);
                }
                else
                {
                    Node referencedNode = property.getNode();
                    
                    Node clonedReferencedNode = null;
                    try
                    {
                        clonedReferencedNode = clonedNode.getSession().getNodeByIdentifier(referencedNode.getIdentifier());
                    }
                    catch (ItemNotFoundException e)
                    {
                        clonedReferencedNode = getOrCloneNode(clonedNode.getSession(), referencedNode);
                    }
                    
                    clonedNode.setProperty(property.getName(), clonedReferencedNode);
                }
            }
            catch (ItemNotFoundException e)
            {
                // the target node does not exist anymore, this could be due to workflow having been deleted
            }
        }
        else
        {
            if (property.getDefinition().isMultiple())
            {
                clonedNode.setProperty(property.getName(), property.getValues());
            }
            else
            {
                clonedNode.setProperty(property.getName(), property.getValue());
            }
        }
    }
    
    /**
     * Clones a node by preserving the source node UUID.
     * @param srcNode the source node.
     * @param clonedNode the cloned node.
     * @param propertyPredicate the property selector.
     * @param nodePredicate the node selector.
     * @throws RepositoryException if an error occurs.
     */
    public void cloneNodeAndPreserveUUID(Node srcNode, Node clonedNode, Predicate propertyPredicate, Predicate nodePredicate) throws RepositoryException
    {
        // Clone properties
        cloneAllProperties(srcNode, clonedNode, propertyPredicate);
        
        // Remove all matching subNodes before cloning, for better handling of same name siblings
        NodeIterator subNodes = clonedNode.getNodes();
        
        while (subNodes.hasNext())
        {
            Node subNode = subNodes.nextNode();
            
            if (nodePredicate.evaluate(subNode))
            {
                subNode.remove();
            }
        }
        
        // Then copy sub nodes
        NodeIterator itNodes = srcNode.getNodes();
        
        while (itNodes.hasNext())
        {
            Node subNode = itNodes.nextNode();
            
            if (nodePredicate.evaluate(subNode))
            {
                Node clonedSubNode = addNodeWithUUID(subNode, clonedNode, subNode.getName());
                cloneNodeAndPreserveUUID(subNode, clonedSubNode, propertyPredicate, nodePredicate);
            }
        }
    }
    
    /**
     * Get or clone a node.
     * @param destSession the destination session.
     * @param node the node in the source workspace.
     * @return the cloned Node in the destination session.
     * @throws RepositoryException if an error occurs.
     */
    public Node getOrCloneNode(Session destSession, Node node) throws RepositoryException
    {
        return getOrCloneNode(destSession, node, null);
    }
    
    /**
     * Get or clone a node in a specified version.
     * @param destSession the destination session.
     * @param node the node in the source workspace.
     * @param version the node version to clone, null to get the current version.
     * @return the cloned Node in the destination session.
     * @throws RepositoryException if an error occurs.
     */
    public Node getOrCloneNode(Session destSession, Node node, String version) throws RepositoryException
    {
        Node clonedParentNode = cloneAncestorsAndPreserveUUID(node, destSession);
        String nodeName = node.getName();
        
        if (clonedParentNode.hasNode(nodeName))
        {
            return clonedParentNode.getNode(nodeName);
        }
        else
        {
            Node clonedNode = null;
            
            if (StringUtils.isNotEmpty(version))
            {
                VersionHistory versionHistory = node.getSession().getWorkspace().getVersionManager().getVersionHistory(node.getPath());
                Node validatedNode = versionHistory.getVersionByLabel(version).getFrozenNode();
                clonedNode = addNodeWithUUID(validatedNode, clonedParentNode, nodeName);
                
                cloneNodeAndPreserveUUID(validatedNode, clonedNode, PredicateUtils.truePredicate(), PredicateUtils.truePredicate());
            }
            else
            {
                clonedNode = addNodeWithUUID(node, clonedParentNode, nodeName);
                
                cloneNodeAndPreserveUUID(node, clonedNode, PredicateUtils.truePredicate(), PredicateUtils.truePredicate());
            }
            
            return clonedNode;
        }
    }
    
    /**
     * Clones ancestors of a node by preserving the source node UUID.
     * @param srcNode the source node.
     * @param destSession the destination session.
     * @return the parent node the destination workspace.
     * @throws RepositoryException if an error occurs.
     */
    public Node cloneAncestorsAndPreserveUUID(Node srcNode, Session destSession) throws RepositoryException
    {
        if (srcNode.getName().length() == 0)
        {
            // We are on the root node which already exists
            return destSession.getRootNode();
        }
        else
        {
            Node destRootNode = destSession.getRootNode();
            Node parentNode = srcNode.getParent();
            String parentNodePath = parentNode.getPath().substring(1);
            
            if (parentNodePath.length() == 0)
            {
                return destSession.getRootNode();
            }
            else if (destRootNode.hasNode(parentNodePath))
            {
                // Found existing parent
                return destRootNode.getNode(parentNodePath);
            }
            else
            {
                Node clonedAncestorNode = cloneAncestorsAndPreserveUUID(parentNode, destSession);
                Node clonedParentNode = null;
                
                if (clonedAncestorNode.hasNode(parentNode.getName()))
                {
                    // Possible with autocreated children
                    clonedParentNode = clonedAncestorNode.getNode(parentNode.getName());
                }
                else
                {
                    clonedParentNode = addNodeWithUUID(parentNode, clonedAncestorNode, parentNode.getName());
                }
                
                // Copy only properties
                cloneNodeAndPreserveUUID(parentNode, clonedParentNode, PredicateUtils.truePredicate(), PredicateUtils.falsePredicate());
                
                return clonedParentNode;
            }
        }
    }
    
    /**
     * Clone a content node, cloning its workflow along.
     * @param destSession the destination session.
     * @param node the node to clone.
     * @return the cloned node.
     * @throws RepositoryException if an error occurs.
     */
    public Node cloneContentNodeWithWorkflow(Session destSession, Node node) throws RepositoryException
    {
        return cloneContentNodeWithWorkflow(destSession, node, PredicateUtils.truePredicate(), PredicateUtils.truePredicate(), null);
    }
    
    /**
     * Clone a content node, cloning its workflow along.
     * @param destSession the destination session.
     * @param node the node to clone.
     * @param propertyPredicate a test on the properties.
     * @param nodePredicate a test on the nodes.
     * @return the cloned node.
     * @throws RepositoryException if an error occurs.
     */
    public Node cloneContentNodeWithWorkflow(Session destSession, Node node, Predicate propertyPredicate, Predicate nodePredicate) throws RepositoryException
    {
        return cloneContentNodeWithWorkflow(destSession, node, propertyPredicate, nodePredicate, null);
    }
    
    /**
     * Clone a content node, cloning its workflow along.
     * @param destSession the destination session.
     * @param node the node to clone.
     * @param propertyPredicate a test on the properties.
     * @param nodePredicate a test on the nodes.
     * @param version the version of the node to clone.
     * @return the cloned node.
     * @throws RepositoryException if an error occurs.
     */
    public Node cloneContentNodeWithWorkflow(Session destSession, Node node, Predicate propertyPredicate, Predicate nodePredicate, String version) throws RepositoryException
    {
        Node clonedContentParentNode = null;
        
        String nodeName = node.getName();
        
        if (_inCollection(node))
        {
            // Clone the ancestors until the collection (hashed nodes will be processed separatel).
            Node destCollectionNode = cloneAncestorsAndPreserveUUID(node.getParent().getParent(), destSession);
            // Search for an available node name.
            nodeName = _getAvailableNodeName(destCollectionNode, node);
            String[] path = _getHashedPath(nodeName);
            
            // Create the two hashed nodes, the content will be created in the second one.
            Node destHash1 = _getOrAddNode(destCollectionNode, path[0], AmetysObjectCollectionFactory.COLLECTION_ELEMENT_NODETYPE);
            clonedContentParentNode = _getOrAddNode(destHash1, path[1], AmetysObjectCollectionFactory.COLLECTION_ELEMENT_NODETYPE);
        }
        else
        {
            clonedContentParentNode = cloneAncestorsAndPreserveUUID(node, destSession);
        }
        
        Node clonedNode = null;
        try
        {
            clonedNode = destSession.getNodeByIdentifier(node.getIdentifier());
        }
        catch (ItemNotFoundException e)
        {
            clonedNode = createNodeClone(node, clonedContentParentNode, nodeName);
        }
        
        cloneNodeAndPreserveUUID(node, clonedNode, propertyPredicate, nodePredicate);
        
        // Clone unversioned metadata
        String unversionedNodeName = RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ":unversioned";
        Node unversionedNode = null;
        
        if (clonedNode.hasNode(unversionedNodeName))
        {
            unversionedNode = clonedNode.getNode(unversionedNodeName);
        }
        else
        {
            unversionedNode = clonedNode.addNode(unversionedNodeName, "ametys:compositeMetadata");
        }
        
        cloneNodeAndPreserveUUID(node.getNode(unversionedNodeName), unversionedNode, propertyPredicate, nodePredicate);
        
        return clonedNode;
    }
    
    private boolean _inCollection(Node node) throws RepositoryException
    {
        boolean inCollection = false;
        
        try
        {
            inCollection = node.getParent().isNodeType(AmetysObjectCollectionFactory.COLLECTION_ELEMENT_NODETYPE)
                && node.getParent().getParent().isNodeType(AmetysObjectCollectionFactory.COLLECTION_ELEMENT_NODETYPE)
                && node.getParent().getParent().getParent().isNodeType(AmetysObjectCollectionFactory.COLLECTION_NODETYPE);
        }
        catch (ItemNotFoundException e)
        {
            // Ignore, one parent does not exist, just return false.
        }
        
        return inCollection;
    }
    
    private String _getAvailableNodeName(Node destCollectionNode, Node srcNode) throws RepositoryException
    {
        String baseName = srcNode.getName();
        String nodeName = NameHelper.filterName(baseName);
        
        int index = 2;
        while (_nodeExistsInCollection(destCollectionNode, nodeName))
        {
            baseName = srcNode.getName() + "-" + index;
            nodeName = NameHelper.filterName(baseName);
            index++;
        }
        
        return nodeName;
    }
    
    private boolean _nodeExistsInCollection(Node collectionNode, String nodeName) throws RepositoryException
    {
        try
        {
            String[] path = _getHashedPath(nodeName);
            return collectionNode.getNode(path[0]).getNode(path[1]).hasNode(nodeName);
        }
        catch (PathNotFoundException e)
        {
            // If a path element is not found, the node doesn't exist. 
            return false;
        }
    }
    
    private String[] _getHashedPath(String name)
    {
        long hash = Math.abs(HashUtil.hash(name));
        String hashStr = Long.toString(hash, 16);
        hashStr = StringUtils.leftPad(hashStr, 4, '0');
        
        return new String[]{hashStr.substring(0, 2), hashStr.substring(2, 4)};
    }
    
    private static Node _getOrAddNode(Node parent, String name, String type) throws RepositoryException
    {
        if (parent.hasNode(name))
        {
            return parent.getNode(name);
        }
        else
        {
            return parent.addNode(name, type);
        }
    }
    
    /**
     * Create a clone of the specified node.
     * @param srcNode the node to clone.
     * @param parentNode the node under which to create the clonde.
     * @param desiredNodeName the wanted node name.
     * @return the cloned node.
     * @throws RepositoryException if an error occurs.
     */
    protected Node createNodeClone(Node srcNode, Node parentNode, String desiredNodeName) throws RepositoryException
    {
        Node clonedNode = null;
        
        String nodeName = NameHelper.filterName(desiredNodeName);
        
        int errorCount = 0;
        do
        {
            try
            {
                clonedNode = addNodeWithUUID(srcNode, parentNode, nodeName);
            }
            catch (ItemExistsException e)
            {
                // Node name is already used.
                errorCount++;
                
                nodeName = NameHelper.filterName(desiredNodeName + " " + (errorCount + 1));
            }
        }
        while (clonedNode == null);
        
        return clonedNode;
    }
    
    /**
     * Reorder a node, mirroring the order in the default workspace.
     * @param parentNode the parent of the source Node in the default workspace.
     * @param nodeName the node name.
     * @param node the node in the destination workspace to be reordered.
     * @throws RepositoryException if an error occurs.
     */
    public void orderNode(Node parentNode, String nodeName, Node node) throws RepositoryException
    {
        Session destSession = node.getSession();
        
        // iterate over the siblings to find the following
        NodeIterator siblings = parentNode.getNodes();
        boolean iterate = true;
        
        while (siblings.hasNext() && iterate)
        {
            Node sibling = siblings.nextNode();
            iterate = !sibling.getName().equals(nodeName);
        }
        
        // iterator is currently on the pageNode
        
        Node nextSibling = null;
        while (siblings.hasNext() && nextSibling == null)
        {
            Node sibling = siblings.nextNode();
            String name = sibling.getName();
            String path = sibling.getPath();
            if (!name.startsWith(RepositoryConstants.NAMESPACE_PREFIX + ":") && !name.startsWith(RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ":") && destSession.itemExists(path))
            {
                nextSibling = sibling;
            }
        }
        
        // nextSibling is either null meaning that the Node must be ordered last or is equals to the following sibling
        if (nextSibling != null)
        {
            node.getParent().orderBefore(nodeName, nextSibling.getName());
        }
        else
        {
            node.getParent().orderBefore(nodeName, null);
        }
    }
    
}
