/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository;

import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;

import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.content.references.OutgoingReferences;
import org.ametys.cms.data.ametysobject.ModelAwareDataAwareAmetysObject;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.explorer.resources.ResourceCollection;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.ModifiableACLAmetysObject;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.holder.DataHolder;
import org.ametys.plugins.repository.data.holder.ModifiableModelLessDataHolder;
import org.ametys.plugins.repository.dublincore.DublinCoreAwareAmetysObject;
import org.ametys.plugins.repository.metadata.MetadataAwareAmetysObject;
import org.ametys.plugins.repository.tag.TagAwareAmetysObject;
import org.ametys.runtime.model.View;

/**
 * Content abstraction defined by the following properties:
 * <dl>
 *   <dt>type
 *   <dd>the content type (can only be set on creation)
 *   <dt>language
 *   <dd>the language (can only be set on creation)
 *   <dt>title
 *   <dd>the title
 *   <dt>creator
 *   <dd>the login of the creator
 *   <dt>creationDate
 *   <dd>the date when the content was created
 *   <dt>lastContributor
 *   <dd>the login of the last contributor
 *   <dt>lastModified
 *   <dd>the date when the last modification takes place
 * </dl>
 */
public interface Content extends ModelAwareDataAwareAmetysObject, MetadataAwareAmetysObject, DublinCoreAwareAmetysObject, TagAwareAmetysObject, ModifiableACLAmetysObject
{
    /** Constants for title attribute */
    public static final String ATTRIBUTE_TITLE = "title";
    
    /** 
     * Constants for title Metadata 
     * Use ATTRIBUTE_TITLE
     */
    @Deprecated
    public static final String METADATA_TITLE = "title";
    
    /**
     * Record to get referencing contents with additional informations, we can get a limited number of referencing contents, then we would
     * need the total number of refencing contents and if all references has been resolved.
     * One referencing content can have several references.
     * 
     * @param total The total number of references
     * @param referencingContents The resolved referencing contents (can be limited)
     * @param hasOtherReferences <code>true</code> if all references has not been resolved in referencingContents attribute.
     */
    public record ReferencingContentsSearch(long total, Collection<Content> referencingContents, boolean hasOtherReferences) { /* empty */ }
    
    /**
     * Retrieves the type identifiers of this content.
     * @return the type identifiers of this content.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public String[] getTypes() throws AmetysRepositoryException;
    
    /**
     * Set the type of this content
     * @param type the type to set
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setType(String type) throws AmetysRepositoryException;
    
    /**
     * Set the types of this Content.<br>
     * @param types the types of this content.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setTypes(String[] types) throws AmetysRepositoryException;
    
    /**
     * Retrieves the mixin type identifiers of this content.
     * @return the mixin type identifiers of this content.
     * @throws AmetysRepositoryException  if an error occurs.
     */
    public String[] getMixinTypes() throws AmetysRepositoryException;
    
    /**
     * Set the mixins of this Content.<br>
     * @param mixins the mixins of this content.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setMixinTypes(String[] mixins) throws AmetysRepositoryException;
    
    /**
     * Retrieves the language of this content.<br>
     * @return the language of this content or <code>null</code> if the content is a multilingual content
     * @throws AmetysRepositoryException if an error occurs.
     */
    public String getLanguage() throws AmetysRepositoryException;

    /**
     * Set the type of this Content.<br>
     * This method may only be called on a new Content, ie. before its first save.
     * @param language the language of this content.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setLanguage(String language) throws AmetysRepositoryException;
    
    /**
     * Retrieves the title for the given locale. If the locale is null or does not exist, the first locale will be used.
     * @param locale The locale. Can be null if the content is not a multilingual content or to get the title in the default locale.
     * @return the title.
     * @throws UnknownDataException if this property does not exist.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public String getTitle(Locale locale) throws UnknownDataException, AmetysRepositoryException;
    
    /**
     * Retrieves the title.
     * This method is same as {@link #getTitle(Locale)} with a null locale.
     * Use this method only if you are manipulating no-multilingual content. If not sure, use {@link #getTitle(Locale)} instead.
     * @return the title.
     * @throws UnknownDataException if this property does not exist.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public String getTitle() throws UnknownDataException, AmetysRepositoryException;

    /**
     * Retrieves the login of the creator.
     * @return the login of the creator.
     * @throws UnknownDataException if this property does not exist.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public UserIdentity getCreator() throws UnknownDataException, AmetysRepositoryException;
    
    /**
     * Retrieves the creation date.
     * @return the creation date.
     * @throws UnknownDataException if this property does not exist.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public ZonedDateTime getCreationDate() throws UnknownDataException, AmetysRepositoryException;
    
    /**
     * Retrieves the login of the last contributor.
     * @return the login of the last contributor.
     * @throws UnknownDataException if this property does not exist.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public UserIdentity getLastContributor() throws UnknownDataException, AmetysRepositoryException;
    
    /**
     * Retrieves the last modification date.
     * @return the last modification date.
     * @throws UnknownDataException if this property does not exist.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public ZonedDateTime getLastModified() throws UnknownDataException, AmetysRepositoryException;

    /**
     * Retrieves the identity of the first validator
     * An empty value should not be considered as a content never being validated.
     * Old content could have been validated before the introduction of this value and have not be updated.
     * {@link #getFirstValidationDate()} is more reliable for this usage.
     * 
     * @return the first validator
     * @throws AmetysRepositoryException if an error occurs.
     */
    public Optional<UserIdentity> getFirstValidator() throws AmetysRepositoryException;
    
    /**
     * Retrieves the first validation date
     * @return the first validation date
     * @throws UnknownDataException if this property does not exist.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public ZonedDateTime getFirstValidationDate() throws UnknownDataException, AmetysRepositoryException;
    
    /**
     * Retrieves the identity of the last validator.
     * An empty value should not be considered as a content never being validated.
     * Old content could have been validated before the introduction of this value and have not be updated.
     * {@link #getLastValidationDate()} is more reliable for this usage.
     * 
     * @return the last validator if the value exists
     * @throws AmetysRepositoryException if an error occurs.
     */
    public Optional<UserIdentity> getLastValidator() throws AmetysRepositoryException;
    
    /**
     * Retrieves the last validation date
     * @return the last validation date
     * @throws UnknownDataException if this property does not exist.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public ZonedDateTime getLastValidationDate() throws UnknownDataException, AmetysRepositoryException;
    
    /**
     * Retrieves the identity of the last major validator
     * An empty value should not be considered as a content never being validated.
     * Old content could have been validated before the introduction of this value and have not be updated.
     * {@link #getLastMajorValidationDate()} is more reliable for this usage.
     * 
     * @return the last major validator
     * @throws AmetysRepositoryException if an error occurs.
     */
    public Optional<UserIdentity> getLastMajorValidator() throws AmetysRepositoryException;
    
    /**
     * Retrieves the last validation date resulting from a major modification. At least this is the first validation date
     * @return the last validation date resulting from a major modification
     * @throws UnknownDataException if this property does not exist.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public ZonedDateTime getLastMajorValidationDate () throws UnknownDataException, AmetysRepositoryException;
    
    /**
     * Returns all Contents referencing this Content (as a metadata).
     * @return all Contents referencing this Content.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public Collection<Content> getReferencingContents() throws AmetysRepositoryException;
    
    /**
     * Returns an object for a limited search of referencing contents with the total of referencing contents, a collection of referencing contents limited to the given limit, and a boolean indicates if there are other references.
     * @param limit The limit for resolved contents
     * @return a record with total number of contents, resolved contents and a boolean indicates if there are more references.
     * @throws AmetysRepositoryException if an error occurs
     */
    public ReferencingContentsSearch searchReferencingContents(int limit) throws AmetysRepositoryException;
    
    /**
     * Returns <code>true</code> if there is at least one Content referencing this Content (as a metadata).
     * @return <code>true</code> if there is at least one Content referencing this Content.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public boolean hasReferencingContents() throws AmetysRepositoryException;
    
    /**
     * Get the stored outgoing references of the content. This references can be used for different purposes, such as testing link consistency for example.
     * @return A non null map of outgoing references grouped by metadata (key are metadata path)
     * @throws AmetysRepositoryException if an error occurs.
     */
    public Map<String, OutgoingReferences> getOutgoingReferences() throws AmetysRepositoryException;

    /**
     * Retrieves the attachments root node
     * @return The attachments root node, or null if the content is working on
     * an (unmodifiable) old version and the attachments root is missing.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public ResourceCollection getRootAttachments() throws AmetysRepositoryException;
    
    /**
     * Generates SAX events representing this Content.
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events.
     * @param locale the {@link Locale} to use for eg. multilingual attributes.
     * @param view the associated View, or null to generate SAX events for all attributes from the model.
     * @param saxWorkflowStep if true, also generates SAX events for the current workflow step.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    public void toSAX(ContentHandler contentHandler, Locale locale, View view, boolean saxWorkflowStep) throws SAXException;
    
    /**
     * Returns the {@link DataHolder} for internal data of this {@link Content}.
     * @return the {@link DataHolder} for internal data of this {@link Content}
     */
    public ModifiableModelLessDataHolder getInternalDataHolder();
}
