/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository;

import javax.jcr.Node;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.content.ContentSaxer;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.data.ContentDataHelper;
import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectFactory;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.plugins.repository.data.type.ModelItemTypeExtensionPoint;
import org.ametys.plugins.repository.jcr.DefaultAmetysObjectFactory;
import org.ametys.plugins.repository.jcr.DefaultTraversableAmetysObject;
import org.ametys.plugins.repository.jcr.TraversableAmetysObjectHelper;

/**
 * {@link AmetysObjectFactory} handling {@link DefaultContent}.
 */
public class ContentFactory extends DefaultAmetysObjectFactory
{
    private ServiceManager _serviceManager;
    private ContentDAO _contentDAO;
    private ContentSaxer _contentSaxer;
    private ContentHelper _contentHelper;
    private ModifiableContentHelper _modifiableContentHelper;
    private ContentDataHelper _contentDataHelper;
    private ModelItemTypeExtensionPoint _modelLessBasicTypesExtensionPoint;
    private SystemPropertyExtensionPoint _systemPropertyExtensionPoint;
    private ContentTypeExtensionPoint _contentTypeExtensionPoint;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _serviceManager = smanager;
        _contentHelper = (ContentHelper) smanager.lookup(ContentHelper.ROLE);
        _modifiableContentHelper = (ModifiableContentHelper) smanager.lookup(ModifiableContentHelper.ROLE);
        _contentSaxer = (ContentSaxer) smanager.lookup(ContentSaxer.CMS_CONTENT_SAXER_ROLE);
        _contentDataHelper = (ContentDataHelper) smanager.lookup(ContentDataHelper.ROLE);
        _modelLessBasicTypesExtensionPoint = (ModelItemTypeExtensionPoint) smanager.lookup(ModelItemTypeExtensionPoint.ROLE_MODEL_LESS_BASIC);
        _systemPropertyExtensionPoint = (SystemPropertyExtensionPoint) smanager.lookup(SystemPropertyExtensionPoint.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    synchronized ContentDAO _getContentDAO()
    {
        // lazy initialization to prevent chicken/egg scenario on startup
        if (_contentDAO == null)
        {
            try
            {
                _contentDAO = (ContentDAO) _serviceManager.lookup(ContentDAO.ROLE);
            }
            catch (ServiceException e)
            {
                throw new AmetysRepositoryException("Caught an exception while retrieving ContentDAO", e);
            }
        }
        return _contentDAO;
    }
    
    AmetysObjectResolver _getAOResolver()
    {
        return _resolver;
    }
    
    ModifiableContentHelper _getModifiableContentHelper()
    {
        return _modifiableContentHelper;
    }
    
    /**
     * Retrieves the content saxer
     * @return the content saxer
     */
    public ContentSaxer getContentSaxer()
    {
        return _contentSaxer;
    }
    
    /**
     * Retrieves the content helper
     * @return the content helper
     */
    public ContentHelper getContentHelper()
    {
        return _contentHelper;
    }
    
    /**
     * Retrieves the content data helper
     * @return the content data helper
     */
    public ContentDataHelper getContentDataHelper()
    {
        return _contentDataHelper;
    }
    
    /**
     * Retrieves the extension point with available data types for internal data
     * @return the extension point with available data types for internal data
     */
    public ModelItemTypeExtensionPoint getInternalDataTypesExtensionPoint()
    {
        return _modelLessBasicTypesExtensionPoint;
    }
    
    /**
     * Retrieves the {@link SystemPropertyExtensionPoint} for contents
     * @return the {@link SystemPropertyExtensionPoint} for contents
     */
    public SystemPropertyExtensionPoint getSystemPropertyExtensionPoint()
    {
        return _systemPropertyExtensionPoint;
    }
    
    /**
     * Retrieves the {@link ContentTypeExtensionPoint}
     * @return the {@link ContentTypeExtensionPoint}
     */
    public ContentTypeExtensionPoint getContentTypeExtensionPoint()
    {
        return _contentTypeExtensionPoint;
    }
    
    
    
    @Override
    public DefaultContent getAmetysObject(Node node, String parentPath) throws AmetysRepositoryException
    {
        return new DefaultContent<>(node, parentPath, this);
    }
    
    /**
     * Returns the {@link AmetysObject} at the given subPath,
     * relative to the given {@link DefaultTraversableAmetysObject}.
     * @param <A> the actual type of {@link AmetysObject}.
     * @param object the context {@link DefaultTraversableAmetysObject}.
     * @param path the sub path. Cannot be <code>null</code>, empty or absolute.
     * @return the {@link AmetysObject} at the given subPath,
     *         relative to the given {@link DefaultTraversableAmetysObject}.
     * @throws AmetysRepositoryException if an error occurs.
     * @throws UnknownAmetysObjectException if no such object exists.
     */
    public <A extends AmetysObject> A getChild(DefaultContent object, String path) throws AmetysRepositoryException, UnknownAmetysObjectException
    {
        return TraversableAmetysObjectHelper.<A>getChild(object, this, path, _resolver, getLogger());
    }
    
    /**
     * Returns all children of the given {@link DefaultTraversableAmetysObject}.
     * @param <A> the actual type of {@link AmetysObject}s
     * @param object a {@link DefaultTraversableAmetysObject}.
     * @return a List containing all children object in the Ametys hierarchy.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public <A extends AmetysObject> AmetysObjectIterable<A> getChildren(DefaultContent object) throws AmetysRepositoryException
    {
        return TraversableAmetysObjectHelper.getChildren(object, this, _resolver, getLogger());
    }
    
    /**
     * Tests if a given object has a child with a given name.
     * @param object the context object.
     * @param name the name to test.
     * @return <code>true</code> is the given object has a child with the given name,
     *         <code>false</code> otherwise.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public boolean hasChild(DefaultContent object, String name) throws AmetysRepositoryException
    {
        return TraversableAmetysObjectHelper.hasChild(object, name, _ametysFactoryExtensionPoint, getLogger());
    }
    
    /**
     * Creates a child to the given object.
     * @param <A> the actual type of {@link AmetysObject}.
     * @param object the parent {@link AmetysObject}.
     * @param name the new object's name.
     * @param type the new object's type.
     * @return the newly created {@link AmetysObject}.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public <A extends AmetysObject> A createChild(DefaultContent object, String name, String type) throws AmetysRepositoryException
    {
        return TraversableAmetysObjectHelper.<A>createChild(object, this, name, type, _ametysFactoryExtensionPoint, _resolver, getLogger());
    }
}
