/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository;

import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.query.QueryHelper;
import org.ametys.plugins.repository.query.SortCriteria;
import org.ametys.plugins.repository.query.expression.Expression;

/**
 * Helper for creating JCR XPath queries involving content predicates.<br>
 * Created XPath queries are like : <code>//element(*, ametys:content)[<i>&lt;content predicate&gt;</i>]</code>
 */
public final class ContentQueryHelper
{
    private ContentQueryHelper()
    {
        // empty constructor
    }
    
    /**
     * Creates the XPath query corresponding to specified {@link Expression}.
     * @param contentExpression the query predicates
     * @return the created XPath query
     */
    public static String getContentXPathQuery(Expression contentExpression)
    {
        return getContentXPathQuery(contentExpression, null);
    }
    
    /**
     * Creates the XPath query corresponding to specified {@link Expression}.
     * @param contentExpression the query predicates
     * @param sortCriteria criteria for sorting results
     * @return the created XPath query
     */
    public static String getContentXPathQuery(Expression contentExpression, SortCriteria sortCriteria)
    {
        return QueryHelper.getXPathQuery(null, "ametys:content", contentExpression, sortCriteria);
    }
    
    /**
     * Creates the XPath query joining criteria on workflow and criteria on content's metadata.
     * @param workflowExpression the workflow query predicates
     * @param contentExpression the content query predicates
     * @return the created XPath query
     */
    public static String getWorkflowXPathQuery(Expression workflowExpression, Expression contentExpression)
    {
        if (workflowExpression == null)
        {
            return getContentXPathQuery(contentExpression);
        }
        
        String contentPredicate = contentExpression == null ? "" : "[" + contentExpression.build() + "]"; 
        
        return "/jcr:root/" + AmetysObjectResolver.ROOT_REPO + "//element(*, oswf:entry)[" + workflowExpression.build() + "]/../.." + contentPredicate;
    }
}
