/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.repository;

import java.time.ZonedDateTime;
import java.util.Locale;
import java.util.Map;

import org.w3c.dom.Node;

import org.ametys.cms.content.references.OutgoingReferences;
import org.ametys.cms.data.ametysobject.ModifiableModelAwareDataAwareAmetysObject;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.RemovableAmetysObject;
import org.ametys.plugins.repository.data.extractor.xml.XMLValuesExtractorAdditionalDataGetter;
import org.ametys.plugins.repository.dublincore.ModifiableDublinCoreAwareAmetysObject;
import org.ametys.plugins.repository.metadata.ModifiableMetadataAwareAmetysObject;
import org.ametys.plugins.repository.tag.TaggableAmetysObject;

/**
 * Modifiable content abstraction 
 */
public interface ModifiableContent extends Content, ModifiableModelAwareDataAwareAmetysObject, ModifiableMetadataAwareAmetysObject, ModifiableDublinCoreAwareAmetysObject, RemovableAmetysObject, TaggableAmetysObject
{
    /**
     * Set the title from the given locale
     * @param title the title.
     * @param locale The locale
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setTitle(String title, Locale locale) throws AmetysRepositoryException;
    
    /**
     * Set the title.
     * Be careful ! Use only if content's title is not a multilingual string. If not sure use {@link #setTitle(String, Locale)} instead.
     * @param title the title.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setTitle(String title) throws AmetysRepositoryException;
    
    /**
     * Set the login of the creator.
     * @param user the creator.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setCreator(UserIdentity user) throws AmetysRepositoryException;

    /**
     * Set the creation date.
     * @param creationDate the creation date.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setCreationDate(ZonedDateTime creationDate) throws AmetysRepositoryException;

    /**
     * Set the login of the last contributor.
     * @param user the last contributor.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setLastContributor(UserIdentity user) throws AmetysRepositoryException;

    /**
     * Set the last modification date.
     * @param lastModified the last modification date.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setLastModified(ZonedDateTime lastModified) throws AmetysRepositoryException;
    
    /**
     * Set the identity of the first validator.
     * @param user the first validator.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setFirstValidator(UserIdentity user) throws AmetysRepositoryException;
    
    /**
     * Set the first validation date
     * @param validationDate the validation date.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setFirstValidationDate(ZonedDateTime validationDate) throws AmetysRepositoryException;
    
    /**
     * Set the identity of the last validator.
     * @param user the last validator.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setLastValidator(UserIdentity user) throws AmetysRepositoryException;
    
    /**
     * Set the last validation date
     * @param validationDate the validation date.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setLastValidationDate (ZonedDateTime validationDate) throws AmetysRepositoryException;
    
    /**
     * Set the identity of the last major validator.
     * @param user the last major validator.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setLastMajorValidator(UserIdentity user) throws AmetysRepositoryException;
    
    /**
     * Set the last validation date resulting from a major modification
     * @param validationDate the validation date.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setLastMajorValidationDate (ZonedDateTime validationDate) throws AmetysRepositoryException;
    
    /**
     * Store the outgoing references of the content.
     * @param references A non null map of outgoing references grouped by metadata (key are metadata path)
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setOutgoingReferences(Map<String, OutgoingReferences> references) throws AmetysRepositoryException;
    
    /**
     * Fills the current content with the values from the provided {@link Node}.
     * <br>This is the anti-operation of {@link Content#toSAX}, as the Node should be a Node previously generated with SAX events from this method.
     * @param node The node to read for retrieving values to fill
     * @param additionalDataGetter The object that will retrieve potential additional data for the content's attributes
     * @throws Exception if an exception occurs
     */
    public void fillContent(Node node, XMLValuesExtractorAdditionalDataGetter additionalDataGetter) throws Exception;
}
