/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import org.ametys.cms.data.ContentDataHelper;
import org.ametys.cms.data.ContentSynchronizationContext;
import org.ametys.cms.data.ContentSynchronizationResult;
import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.ReferencedContents;
import org.ametys.cms.data.holder.impl.DefaultModifiableModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.values.SynchronizationContext;
import org.ametys.plugins.repository.data.holder.values.SynchronizationResult;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.runtime.model.ModelItemContainer;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.UndefinedItemPathException;

/**
 * Implementation modifiable data holder for content
 */
public class ModifiableContentDataHolder extends DefaultModifiableModelAwareDataHolder
{
    /** the content */
    protected ModifiableContent _content;
    
    /** the content data helper */
    protected ContentDataHelper _contentDataHelper;
    
    /**
     * Creates a modifiable content data holder
     * @param content the content
     * @param contentDataHelper the content data helper
     * @param repositoryData the repository data to use
     * @param itemContainers the model containers to use to get information about definitions. Must match the given repository data. A repository data can have several item containers. For example, a content can have several content types.
     */
    public ModifiableContentDataHolder(ModifiableContent content, ContentDataHelper contentDataHelper, ModifiableRepositoryData repositoryData, Collection<? extends ModelItemContainer> itemContainers)
    {
        super(repositoryData, Optional.empty(), Optional.of(content), itemContainers);
        _content = content;
        _contentDataHelper = contentDataHelper;
    }
    
    @Override
    public <T extends SynchronizationResult> T synchronizeValues(ViewItemAccessor viewItemAccessor, Map<String, Object> values, SynchronizationContext context) throws UndefinedItemPathException, BadItemTypeException
    {
        Collection<ReferencedContents> allReferencedContents = new ArrayList<>();
        if (context instanceof ContentSynchronizationContext)
        {
            ContentSynchronizationContext contentContext = (ContentSynchronizationContext) context;
            if (contentContext.invertRelationsEnabled())
            {
                allReferencedContents.addAll(contentContext.getReferencedContents()
                        .orElseGet(() -> _contentDataHelper.collectReferencedContents(viewItemAccessor, _content, values, contentContext)));
            }
        }
        
        T result = super.synchronizeValues(viewItemAccessor, values, context);
        
        if (result instanceof ContentSynchronizationResult)
        {
            Map<String, ModifiableContent> modifiedContents = new HashMap<>();
            
            for (ReferencedContents referencedContents : allReferencedContents)
            {
                String invertRelationPath = referencedContents.getDefinition().getInvertRelationPath();
                modifiedContents.putAll(ContentDataHelper.manageInvertRelations(invertRelationPath, referencedContents.getAddedContents(), _content.getId(), ContentDataHelper::addInvertRelation, modifiedContents, context.getExternalizableDataContext()));
                modifiedContents.putAll(ContentDataHelper.manageInvertRelations(invertRelationPath, referencedContents.getRemovedContents(), _content.getId(), ContentDataHelper::removeInvertRelation, modifiedContents, context.getExternalizableDataContext()));
                
                Map<ContentValue, Collection<String>> thirdPartyContents = new HashMap<>();
                for (Map.Entry<ContentValue, ContentValue> thirdPartyContent : referencedContents.getThirdPartyContents().entrySet())
                {
                    thirdPartyContents.computeIfAbsent(thirdPartyContent.getValue(), __ -> new ArrayList<>()).add(thirdPartyContent.getKey().getContentId());
                }
                modifiedContents.putAll(ContentDataHelper.manageInvertRelations(referencedContents.getDefinition().getPath(), thirdPartyContents, ContentDataHelper::removeInvertRelation, modifiedContents, context.getExternalizableDataContext()));
            }
            
            ((ContentSynchronizationResult) result).addModifiedContents(modifiedContents.values());
        }
        
        return result;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    protected <T extends SynchronizationContext> T _createSynchronizationContextInstance()
    {
        return (T) ContentSynchronizationContext.newInstance();
    }
    
    @SuppressWarnings("unchecked")
    @Override
    protected <T extends SynchronizationResult> T _createSetValueResultInstance()
    {
        return (T) new ContentSynchronizationResult();
    }
}
