/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository;

import java.time.ZonedDateTime;
import java.util.Calendar;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import javax.jcr.Node;
import javax.jcr.RepositoryException;

import org.apache.avalon.framework.component.Component;

import org.ametys.cms.content.references.OutgoingReferences;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.holder.ModifiableModelLessDataHolder;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelItem;


/**
 * Provides helper methods to use the {@link ModifiableContent} API on {@link DefaultContent}s.
 */
public class ModifiableContentHelper implements Component
{
    /** The Avalon role */
    public static final String ROLE = ModifiableContentHelper.class.getName();
    
    /** Constants for the root of comments */
    public static final String METADATA_COMMENTS = "comments";
    /** Constants for the root of contributor comments */
    public static final String METADATA_CONTRIBUTOR_COMMENTS = "contributor-comments";
    
    /**
     * Set a {@link DefaultContent} title for the given locale. 
     * @param content the {@link DefaultContent} to set.
     * @param title the title to set.
     * @param locale The locale. Can be null if the content is not a multilingual content.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setTitle(DefaultContent content, String title, Locale locale) throws AmetysRepositoryException
    {
        ModelItem titleDefinition = content.getDefinition(Content.ATTRIBUTE_TITLE);
        ModifiableRepositoryData repositoryData = new JCRRepositoryData(content.getNode());
        if (titleDefinition instanceof ElementDefinition && ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID.equals(((ElementDefinition) titleDefinition).getType().getId()))
        {
            if (locale == null)
            {
                throw new IllegalArgumentException("Cannot set a title with null locale on a multilingual title");
            }
            
            ModifiableRepositoryData titleRepositoryData;
            if (repositoryData.hasValue(Content.ATTRIBUTE_TITLE))
            {
                titleRepositoryData = repositoryData.getRepositoryData(Content.ATTRIBUTE_TITLE);
            }
            else
            {
                titleRepositoryData = repositoryData.addRepositoryData(Content.ATTRIBUTE_TITLE, RepositoryConstants.MULTILINGUAL_STRING_METADATA_NODETYPE);
            }
            titleRepositoryData.setValue(locale.toString(), title);
        }
        else
        {
            repositoryData.setValue(Content.ATTRIBUTE_TITLE, title);
        }
    }
    
    /**
     * Set the title of non-multilingual {@link DefaultContent}.
     * Be careful! Use only if content's title is not a multilingual string. If not sure use {@link #setTitle(DefaultContent, String, Locale)} instead.
     * @param content the {@link DefaultContent} to set.
     * @param title the title to set.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setTitle(DefaultContent content, String title) throws AmetysRepositoryException
    {
        setTitle(content, title, null);
    }
    
    /**
     * Copy the title of the source content to the target content
     * @param srcContent The source content
     * @param targetContent The target content
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void copyTitle(Content srcContent, ModifiableContent targetContent) throws AmetysRepositoryException
    {
        targetContent.setValue(Content.ATTRIBUTE_TITLE, srcContent.getValue(Content.ATTRIBUTE_TITLE));
    }
    
    /**
     * Set a {@link DefaultContent} user.
     * @param content the {@link DefaultContent} to set.
     * @param user the user to set.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setCreator(DefaultContent content, UserIdentity user) throws AmetysRepositoryException
    {
        _storeUserMetadata(content, DefaultContent.METADATA_CREATOR, user);
    }

    /**
     * Set a {@link DefaultContent} creation date.
     * @param content the {@link DefaultContent} to set.
     * @param creationDate the creation date to set.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setCreationDate(DefaultContent content, ZonedDateTime creationDate) throws AmetysRepositoryException
    {
        ModifiableRepositoryData repositoryData = new JCRRepositoryData(content.getNode());
        Calendar calendar = DateUtils.asCalendar(creationDate);
        repositoryData.setValue(DefaultContent.METADATA_CREATION, calendar);
    }
    
    /**
     * Set a {@link DefaultContent} contributor.
     * @param content the {@link DefaultContent} to set.
     * @param user the contributor to set.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setLastContributor(DefaultContent content, UserIdentity user) throws AmetysRepositoryException
    {
        _storeUserMetadata(content, DefaultContent.METADATA_CONTRIBUTOR, user);
    }
    
    /**
     * Set a {@link DefaultContent} last modification date.
     * @param content the {@link DefaultContent} to set.
     * @param lastModified the last modification date to set.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setLastModified(DefaultContent content, ZonedDateTime lastModified) throws AmetysRepositoryException
    {
        ModifiableRepositoryData repositoryData = new JCRRepositoryData(content.getNode());
        Calendar calendar = DateUtils.asCalendar(lastModified);
        repositoryData.setValue(DefaultContent.METADATA_MODIFIED, calendar);
    }
    
    /**
     * Set a {@link DefaultContent} first validator.
     * @param content the {@link DefaultContent} to set.
     * @param user the validator to set.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setFirstValidator(DefaultContent content, UserIdentity user) throws AmetysRepositoryException
    {
        _storeUserMetadata(content, DefaultContent.METADATA_FIRST_VALIDATOR, user);
    }
    
    /**
     * Set a {@link DefaultContent} first validation date.
     * @param content the {@link DefaultContent} to set.
     * @param validationDate the first validation date.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setFirstValidationDate(DefaultContent content, ZonedDateTime validationDate) throws AmetysRepositoryException
    {
        ModifiableRepositoryData repositoryData = new JCRRepositoryData(content.getNode());
        Calendar calendar = DateUtils.asCalendar(validationDate);
        repositoryData.setValue(DefaultContent.METADATA_FIRST_VALIDATION, calendar);
    }
    
    /**
     * Set a {@link DefaultContent} last validator.
     * @param content the {@link DefaultContent} to set.
     * @param user the validator to set.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setLastValidator(DefaultContent content, UserIdentity user) throws AmetysRepositoryException
    {
        _storeUserMetadata(content, DefaultContent.METADATA_LAST_VALIDATOR, user);
    }
    
    /**
     * Set a {@link DefaultContent} last validation date.
     * @param content the {@link DefaultContent} to set.
     * @param validationDate the last validation date.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setLastValidationDate(DefaultContent content, ZonedDateTime validationDate) throws AmetysRepositoryException
    {
        ModifiableRepositoryData repositoryData = new JCRRepositoryData(content.getNode());
        Calendar calendar = DateUtils.asCalendar(validationDate);
        repositoryData.setValue(DefaultContent.METADATA_LAST_VALIDATION, calendar);
    }
    
    /**
     * Set a {@link DefaultContent} last major validator.
     * @param content the {@link DefaultContent} to set.
     * @param user the validator to set.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setLastMajorValidator(DefaultContent content, UserIdentity user) throws AmetysRepositoryException
    {
        _storeUserMetadata(content, DefaultContent.METADATA_LAST_MAJOR_VALIDATOR, user);
    }
    
    /**
     * Set a {@link DefaultContent} last major validation date.
     * @param content the {@link DefaultContent} to set.
     * @param validationDate the last major validation date.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setLastMajorValidationDate(DefaultContent content, ZonedDateTime validationDate) throws AmetysRepositoryException
    {
        ModifiableRepositoryData repositoryData = new JCRRepositoryData(content.getNode());
        Calendar calendar = DateUtils.asCalendar(validationDate);
        repositoryData.setValue(DefaultContent.METADATA_LAST_MAJORVALIDATION, calendar);
    }
    
    /**
     * Store the outgoing references on the content.
     * @param content The content concerned by these outgoing references.
     * @param references A non null map of outgoing references grouped by metadata (key are metadata path)
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setOutgoingReferences(DefaultContent content, Map<String, OutgoingReferences> references) throws AmetysRepositoryException
    {
        try
        {
            Node contentNode = content.getNode();
            if (contentNode.hasNode(RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ':' + DefaultContent.METADATA_ROOT_OUTGOING_REFERENCES))
            {
                contentNode.getNode(RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ':' + DefaultContent.METADATA_ROOT_OUTGOING_REFERENCES).remove();
            }
            
            if (references.size() != 0)
            {
                Node rootOutgoingRefsNode = contentNode.addNode(RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ':' + DefaultContent.METADATA_ROOT_OUTGOING_REFERENCES);
                for (String path : references.keySet())
                {
                    // Outgoing references node creation (for the given path)
                    Node outgoingRefsNode = rootOutgoingRefsNode.addNode(RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ':' + DefaultContent.METADATA_OUTGOING_REFERENCES);
                    outgoingRefsNode.setProperty(RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ':' + DefaultContent.METADATA_OUTGOING_REFERENCES_PATH_PROPERTY, path);
                    
                    // Reference nodes per type
                    OutgoingReferences outgoingReferences = references.get(path);
                    
                    for (String type : outgoingReferences.keySet())
                    {
                        List<String> referenceValues = outgoingReferences.get(type);
                        if (referenceValues != null && !referenceValues.isEmpty())
                        {
                            Node outgoingReferenceNode = outgoingRefsNode.addNode(type, RepositoryConstants.NAMESPACE_PREFIX + ':' + DefaultContent.METADATA_OUTGOING_REFERENCE_NODETYPE);
                            outgoingReferenceNode.setProperty(RepositoryConstants.NAMESPACE_PREFIX + ':' + DefaultContent.METADATA_OUTGOING_REFERENCE_PROPERTY, referenceValues.toArray(new String[referenceValues.size()]));
                        }
                    }
                }
            }
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException(e);
        }
    }

    /**
     * Store a metadata of type user in the content
     * @param content the content described by the metadata
     * @param metadataName the name of the metadata
     * @param user the value to set
     * @throws AmetysRepositoryException when an error occurred
     */
    protected void _storeUserMetadata(DefaultContent content, String metadataName, UserIdentity user) throws AmetysRepositoryException
    {
        try
        {
            // TODO CMS-9336 All the metatadata here should be stored using types
            ModifiableRepositoryData repositoryData = new JCRRepositoryData(content.getNode());
            ModifiableRepositoryData creatorRepositoryData;
            if (repositoryData.hasValue(metadataName))
            {
                creatorRepositoryData = repositoryData.getRepositoryData(metadataName);
            }
            else
            {
                creatorRepositoryData = repositoryData.addRepositoryData(metadataName, RepositoryConstants.USER_NODETYPE);
            }
            creatorRepositoryData.setValue("login", user.getLogin());
            creatorRepositoryData.setValue("population", user.getPopulationId());
        }
        catch (AmetysRepositoryException e)
        {
            throw new AmetysRepositoryException("Error setting the metadata '" + metadataName + "' for content '" + content.getId() + "' with value '" + UserIdentity.userIdentityToString(user) + "'.", e);
        }
    }
    
    /**
     * Retrieves the data holder for the given content's comments
     * @param content the content
     * @return the data holder for content's comments
     */
    public ModifiableModelLessDataHolder getCommentsDataHolder(DefaultContent content)
    {
        return content.getUnversionedDataHolder().getComposite(METADATA_COMMENTS, true);
    }
    
    /**
     * Retrieves the data holder for the given content's contributor comments
     * @param content the content
     * @return the data holder for content's contributor comments
     */
    public ModifiableModelLessDataHolder getContributorCommentsDataHolder(DefaultContent content)
    {
        return content.getUnversionedDataHolder().getComposite(METADATA_CONTRIBUTOR_COMMENTS, true);
    }
}
