/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository;

import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import javax.jcr.Node;

import org.ametys.cms.content.references.OutgoingReferences;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.repository.comment.Comment;
import org.ametys.cms.repository.comment.CommentableContent;
import org.ametys.cms.repository.comment.contributor.ContributorCommentableContent;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.RepositoryIntegrityViolationException;
import org.ametys.plugins.repository.data.extractor.xml.XMLValuesExtractorAdditionalDataGetter;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.jcr.DublinCoreHelper;
import org.ametys.plugins.repository.lock.LockableAmetysObject;
import org.ametys.plugins.repository.metadata.MetadataAwareAmetysObjectHelper;
import org.ametys.plugins.repository.tag.TaggableAmetysObjectHelper;

/**
 * Default implementation of a {@link Content}, also versionable, lockable and workflow-aware.
 * @param <F> the actual type of factory.
 */
public class ModifiableDefaultContent<F extends ModifiableContentFactory> extends DefaultContent<F> implements CommentableContent, ContributorCommentableContent, LockableAmetysObject, ModifiableWorkflowAwareContent/*, AnnotableContent*/ //, CopiableAmetysObject
{
    /**
     * Creates a JCR-based Content.
     * @param node the JCR Node backing this Content.
     * @param parentPath the parent path in the Ametys hierarchy.
     * @param factory the corresponding {@link ContentFactory}.
     */
    public ModifiableDefaultContent(Node node, String parentPath, F factory)
    {
        super(node, parentPath, factory);
    }

    public void setTitle(String title, Locale locale) throws AmetysRepositoryException
    {
        _getFactory()._getModifiableContentHelper().setTitle(this, title, locale);
    }
    
    public void setTitle(String title) throws AmetysRepositoryException
    {
        _getFactory()._getModifiableContentHelper().setTitle(this, title);
    }
    
    public void setCreator(UserIdentity user) throws AmetysRepositoryException
    {
        _getFactory()._getModifiableContentHelper().setCreator(this, user);
    }
    
    public void setCreationDate(ZonedDateTime creationDate) throws AmetysRepositoryException
    {
        _getFactory()._getModifiableContentHelper().setCreationDate(this, creationDate);
    }
    
    public void setLastContributor(UserIdentity user)
    {
        _getFactory()._getModifiableContentHelper().setLastContributor(this, user);
    }
    
    public void setLastModified(ZonedDateTime lastModified) throws AmetysRepositoryException
    {
        _getFactory()._getModifiableContentHelper().setLastModified(this, lastModified);
    }
    
    public void setFirstValidator(UserIdentity user) throws AmetysRepositoryException
    {
        _getFactory()._getModifiableContentHelper().setFirstValidator(this, user);
    }
    
    @Override
    public void setFirstValidationDate(ZonedDateTime validationDate) throws AmetysRepositoryException
    {
        _getFactory()._getModifiableContentHelper().setFirstValidationDate(this, validationDate);
    }
    
    public void setLastValidator(UserIdentity user) throws AmetysRepositoryException
    {
        _getFactory()._getModifiableContentHelper().setLastValidator(this, user);
    }
    
    @Override
    public void setLastValidationDate(ZonedDateTime validationDate) throws AmetysRepositoryException
    {
        _getFactory()._getModifiableContentHelper().setLastValidationDate(this, validationDate);
    }
    
    public void setLastMajorValidator(UserIdentity user) throws AmetysRepositoryException
    {
        _getFactory()._getModifiableContentHelper().setLastMajorValidator(this, user);
    }
    
    @Override
    public void setLastMajorValidationDate(ZonedDateTime validationDate) throws AmetysRepositoryException
    {
        _getFactory()._getModifiableContentHelper().setLastMajorValidationDate(this, validationDate);
    }
    
    @Override
    public void setOutgoingReferences(Map<String, OutgoingReferences> references) throws AmetysRepositoryException
    {
        _getFactory()._getModifiableContentHelper().setOutgoingReferences(this, references);
    }
    
    // ------------------- Tags management -------------------- //
    @Override
    public void tag(String tag) throws AmetysRepositoryException
    {
        TaggableAmetysObjectHelper.tag(this, tag);
    }
    
    @Override
    public void untag(String tag) throws AmetysRepositoryException
    {
        TaggableAmetysObjectHelper.untag(this, tag);
    }
    
    // ------------------- Workflow management -------------------- //
    @Override
    public long getWorkflowId() throws AmetysRepositoryException
    {
        return WorkflowAwareContentHelper.getWorkflowId(this);
    }
    
    @Override
    public void setWorkflowId(long workflowId) throws AmetysRepositoryException
    {
        WorkflowAwareContentHelper.setWorkflowId(this, workflowId);
    }
    
    @Override
    public long getCurrentStepId() throws AmetysRepositoryException
    {
        return WorkflowAwareContentHelper.getCurrentStepId(this);
    }
    
    @Override
    public void setCurrentStepId (long stepId) throws AmetysRepositoryException
    {
        WorkflowAwareContentHelper.setCurrentStepId(this, stepId);
    }
    
    @Override
    public ZonedDateTime getProposalDate() throws AmetysRepositoryException
    {
        return WorkflowAwareContentHelper.getProposalDate(this);
    }
    
    @Override
    public void setProposalDate(ZonedDateTime proposalDate) throws AmetysRepositoryException
    {
        WorkflowAwareContentHelper.setProposalDate(this, proposalDate);
    }
    
    @Override
    public void remove() throws AmetysRepositoryException, RepositoryIntegrityViolationException
    {
        MetadataAwareAmetysObjectHelper.removeSubObjects(this);
        
        super.remove();
    }
    
    // --------------------- Lock management ----------------------- //
    @Override
    public void lock() throws AmetysRepositoryException
    {
        _getFactory().getLockComponent().lock(this);
    }
    
    @Override
    public void unlock() throws AmetysRepositoryException
    {
        _getFactory().getLockComponent().unlock(this);
    }
    
    @Override
    public boolean isLocked() throws AmetysRepositoryException
    {
        return _getFactory().getLockComponent().isLocked(this);
    }
    
    @Override
    public UserIdentity getLockOwner() throws AmetysRepositoryException
    {
        return _getFactory().getLockComponent().getLockOwner(this);
    }
    
    // -------------------------- DUBLIN CORE ---------------------------- //
    @Override
    public void setDCTitle(String title) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCTitle(this, title);
    }

    @Override
    public void setDCCreator(String creator) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCCreator(this, creator);
    }

    @Override
    public void setDCSubject(String[] subject) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCSubject(this, subject);
    }

    @Override
    public void setDCDescription(String description) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCDescription(this, description);
    }

    @Override
    public void setDCPublisher(String publisher) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCPublisher(this, publisher);
    }

    @Override
    public void setDCContributor(String contributor) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCContributor(this, contributor);
    }

    @Override
    public void setDCDate(Date date) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCDate(this, date);
    }

    @Override
    public void setDCType(String type) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCType(this, type);
    }

    @Override
    public void setDCFormat(String format) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCFormat(this, format);
    }

    @Override
    public void setDCIdentifier(String identifier) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCIdentifier(this, identifier);
    }

    @Override
    public void setDCSource(String source) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCSource(this, source);
    }

    @Override
    public void setDCLanguage(String language) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCLanguage(this, language);
    }

    @Override
    public void setDCRelation(String relation) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCRelation(this, relation);
    }
    
    @Override
    public void setDCCoverage(String coverage) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCCoverage(this, coverage);
    }
    
    @Override
    public void setDCRights(String rights) throws AmetysRepositoryException
    {
        DublinCoreHelper.setDCRights(this, rights);
    }
    
    // ---------------------------- COMMENTS ------------------------------ //
    @Override
    public Comment createComment()
    {
        return new Comment(_getFactory()._getModifiableContentHelper().getCommentsDataHolder(this));
    }
    
    public Comment createComment(String commentId, ZonedDateTime creationDate)
    {
        return new Comment(_getFactory()._getModifiableContentHelper().getCommentsDataHolder(this), commentId, creationDate);
    }
    
    @Override
    public Comment getComment(String commentId) throws AmetysRepositoryException
    {
        return Comment.getComment(_getFactory()._getModifiableContentHelper().getCommentsDataHolder(this), commentId);
    }
    
    @Override
    public List<Comment> getComments(boolean includeNotValidatedComments, boolean includeValidatedComments) throws AmetysRepositoryException
    {
        return Comment.getComments(_getFactory()._getModifiableContentHelper().getCommentsDataHolder(this), includeNotValidatedComments, includeValidatedComments);
    }
    
    // ---------------------------- CONTRIBUTOR COMMENTS ------------------------------ //
    @Override
    public Comment createContributorComment()
    {
        return new Comment(_getFactory()._getModifiableContentHelper().getContributorCommentsDataHolder(this));
    }
    
    public Comment createContributorComment(String commentId, ZonedDateTime creationDate)
    {
        return new Comment(_getFactory()._getModifiableContentHelper().getContributorCommentsDataHolder(this), commentId, creationDate);
    }
    
    @Override
    public Comment getContributorComment(String commentId) throws AmetysRepositoryException
    {
        return Comment.getComment(_getFactory()._getModifiableContentHelper().getContributorCommentsDataHolder(this), commentId);
    }
    
    @Override
    public List<Comment> getContributorComments() throws AmetysRepositoryException
    {
        return Comment.getComments(_getFactory()._getModifiableContentHelper().getContributorCommentsDataHolder(this), false, true);
    }
    
    // ---------------------------- DATA HOLDER --------------------------- //
    @Override
    public ModifiableContentDataHolder getDataHolder()
    {
        // It is necessary to instantiate the repository data because of locks:
        // when the content is locked, the lock token is released to be taken by anyone
        // in repository data there is a lock checked boolean that need to be reset in order to take the token
        // this boolean is only reset at instantiation
        Collection<ContentType> contentTypes = _getFactory().getContentHelper().getContentTypes(this);
        return new ModifiableContentDataHolder(this, _getFactory().getContentDataHelper(), new JCRRepositoryData(getNode()), contentTypes);
    }
    
    public void fillContent(org.w3c.dom.Node node, XMLValuesExtractorAdditionalDataGetter additionalDataGetter) throws Exception
    {
        _getFactory().getContentExtractor().fillContent(this, node, additionalDataGetter);
    }
}
