/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.repository.ReactionableObject.ReactionType;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.plugins.repository.data.holder.ModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.ModifiableModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelLessComposite;

/**
 * Helper class which provides methods to manage reactions on a object
 */
public final class ReactionableObjectHelper implements Serviceable, Component
{
    /** The Avalon role */
    public static final String ROLE = ReactionableObjectHelper.class.getName();
    
    private UserHelper _userHelper;
    
    public void service(ServiceManager smanager) throws ServiceException
    {
        _userHelper = (UserHelper) smanager.lookup(UserHelper.ROLE);
    }
    
    /**
     * Add a user reaction
     * @param unversionedDataHolder the unversioned data holder
     * @param user the issuer of reaction
     * @param reactionType the reaction type
     */
    public static void addReaction(ModifiableModelLessDataHolder unversionedDataHolder, UserIdentity user, ReactionType reactionType)
    {
        ModifiableModelLessComposite reactionHolder = _getReactionsHolder(unversionedDataHolder, reactionType);
        
        UserIdentity[] userArray = _getUsers(reactionHolder);
        if (!ArrayUtils.contains(userArray, user))
        {
            UserIdentity[] newUserArray = ArrayUtils.add(userArray, user);
            reactionHolder.setValue("users", newUserArray);
        }
    }

    /**
     * Remove a user reaction
     * @param unversionedDataHolder the unversioned data holder
     * @param user the issuer of reaction to remove
     * @param reactionType the reaction type
     */
    public static void removeReaction(ModifiableModelLessDataHolder unversionedDataHolder, UserIdentity user, ReactionType reactionType)
    {
        ModifiableModelLessComposite reactionHolder = _getReactionsHolder(unversionedDataHolder, reactionType);
        
        UserIdentity[] userArray = _getUsers(reactionHolder);
        if (ArrayUtils.contains(userArray, user))
        {
            UserIdentity[] newUserArray = ArrayUtils.removeElement(userArray, user);
            reactionHolder.setValue("users", newUserArray);
        }
    }
    
    /**
     * Get the issuers of a reaction
     * @param unversionedDataHolder the unversioned data holder
     * @param reactionType the reaction type
     * @return the issuers of a reaction as a List
     */
    public static List<UserIdentity> getReactionUsers(ModifiableModelLessDataHolder unversionedDataHolder, ReactionType reactionType)
    {
        ModifiableModelLessComposite reactionHolder = _getReactionsHolder(unversionedDataHolder, reactionType);
        UserIdentity[] userArray = _getUsers(reactionHolder);
        return Arrays.asList(userArray);
    }
    
    /**
     * Generates SAX events for the given object's reactions.
     * @param reactionable the {@link ReactionableObject}.
     * @param contentHandler the ContentHandler receiving SAX events.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    public void saxReactions(ReactionableObject reactionable, ContentHandler contentHandler) throws SAXException
    {
        XMLUtils.startElement(contentHandler, "reactions");

        for (ReactionType reactionType : ReactionType.values())
        {
            List<UserIdentity> actors = reactionable.getReactionUsers(reactionType);
            if (!actors.isEmpty())
            {
                AttributesImpl attrs = new AttributesImpl();
                attrs.addCDATAAttribute("type", reactionType.name());
                XMLUtils.startElement(contentHandler, "reaction", attrs);
    
                for (UserIdentity actor : actors)
                {
                    _userHelper.saxUserIdentity(actor, contentHandler, "actor");
                }
    
                XMLUtils.endElement(contentHandler, "reaction");
            }
        }
        XMLUtils.endElement(contentHandler, "reactions");
    }
    
    /**
     * Get the reactions to JSON format
     * @param reactionable the reactionable object
     * @return the reactions as JSON
     */
    public List<Map<String, Object>> reactionsToJson(ReactionableObject reactionable)
    {
        List<Map<String, Object>> reactions2json = new ArrayList<>();
        for (ReactionType reactionType : ReactionType.values())
        {
            List<UserIdentity> actors = reactionable.getReactionUsers(reactionType);
            if (!actors.isEmpty())
            {
                Map<String, Object> reaction2json = new HashMap<>();
                
                reaction2json.put("type", reactionType.name());
                
                List<Map<String, Object>> actors2json = new ArrayList<>();
                for (UserIdentity actor : actors)
                {
                    actors2json.add(_userHelper.user2json(actor));
                }
                
                reaction2json.put("actors", actors2json);
                
                reactions2json.add(reaction2json);
            }
        }
        
        return reactions2json;
    }
    
    private static ModifiableModelLessComposite _getReactionsHolder(ModifiableModelLessDataHolder unversionedDataHolder, ReactionType reactionType)
    {
        return unversionedDataHolder.getComposite(reactionType.name().toLowerCase(), true);
    }
    
    private static UserIdentity[] _getUsers(ModelLessDataHolder dataHolder)
    {
        if (dataHolder.hasValue("users") && !dataHolder.isMultiple("users"))
        {
            UserIdentity user = dataHolder.getValue("users");
            return new UserIdentity[] {user};
        }
        else
        {
            return dataHolder.getValue("users", new UserIdentity[0]);
        }
    }
}
