/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository;

import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.plugins.repository.data.holder.ModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.ModifiableModelLessDataHolder;

/**
 * Helper class which provides methods to manage reports on a object
 */
public final class ReportableObjectHelper
{
    /** Constants for the reports count attribute */
    public static final String REPORTS_COUNT_ATTRIBUTE_NAME = "reportsCount";
    
    private ReportableObjectHelper()
    {
        // Hide the default constructor
    }
    
    /**
     * Add a report to the object
     * @param unversionedDataHolder the unversioned data holder
     */
    public static void addReport(ModifiableModelLessDataHolder unversionedDataHolder)
    {
        long reportsCount = getReportsCount(unversionedDataHolder);
        long increasedReportsCount = reportsCount + 1;
        unversionedDataHolder.setValue(REPORTS_COUNT_ATTRIBUTE_NAME, increasedReportsCount);
    }
    
    /**
     * Sets the number of reports on the object
     * @param unversionedDataHolder the unversioned data holder
     * @param reportsCount the number of reports to set
     */
    public static void setReportsCount(ModifiableModelLessDataHolder unversionedDataHolder, long reportsCount)
    {
        unversionedDataHolder.setValue(REPORTS_COUNT_ATTRIBUTE_NAME, reportsCount);
    }
    
    /**
     * Remove all reports on the object
     * @param unversionedMetadataHolder the unversioned data holder
     */
    public static void clearReports(ModifiableModelLessDataHolder unversionedMetadataHolder)
    {
        unversionedMetadataHolder.removeValue(REPORTS_COUNT_ATTRIBUTE_NAME);
    }
    
    /**
     * Retrieves the number of reports on the object
     * @param unversionedDataHolder the unversioned data holder
     * @return the number of reports
     */
    public static long getReportsCount(ModelLessDataHolder unversionedDataHolder)
    {
        return unversionedDataHolder.getValue(REPORTS_COUNT_ATTRIBUTE_NAME, 0L);
    }
    
    /**
     * Generates SAX events for the given object's reports.
     * @param reportable the {@link Content}.
     * @param contentHandler the ContentHandler receiving SAX events.
     * @throws SAXException if an error occurs during the SAX events generation.
     */
    public static void saxReports(ReportableObject reportable, ContentHandler contentHandler) throws SAXException
    {
        long reportsCount = reportable.getReportsCount();
        if (reportsCount > 0)
        {
            AttributesImpl attrs = new AttributesImpl();
            attrs.addCDATAAttribute("count", String.valueOf(reportsCount));
            XMLUtils.createElement(contentHandler, "reports", attrs);
        }
    }
}
