/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository;

import java.time.ZonedDateTime;

import javax.jcr.Node;
import javax.jcr.RepositoryException;

import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.jcr.JCRAmetysObject;
import org.ametys.plugins.repository.version.ModifiableDataAwareVersionableAmetysObject;
import org.ametys.runtime.model.type.ModelItemTypeConstants;

/**
 * Provides helper methods to use the {@link WorkflowAwareContent} API on {@link JCRAmetysObject}s.
 */
public final class WorkflowAwareContentHelper
{
    
    /** Constants for workflowId Metadata* */
    public static final String METADATA_WORKFLOW_ID = "workflowId";
    
    /** Constants for currentStepId Metadata* */
    public static final String METADATA_CURRENT_STEP_ID = "currentStepId";
    
    /** Name of the metadata storing the proposal date. */
    public static final String METADATA_LAST_PROPOSAL_DATE = "proposalDate";
    
    private WorkflowAwareContentHelper()
    {
        // Hide default constructor.
    }
    
    /**
     * Get an object's workflow ID.
     * @param object the object.
     * @return the object's workflow ID.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public static long getWorkflowId(JCRAmetysObject object) throws AmetysRepositoryException
    {
        try
        {
            return object.getNode().getProperty(RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ':' + METADATA_WORKFLOW_ID).getLong();
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to get workflowId property", e);
        }
    }
    
    /**
     * Set an object's workflow ID.
     * @param object the object.
     * @param workflowId the workflow ID to set.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public static void setWorkflowId(JCRAmetysObject object, long workflowId) throws AmetysRepositoryException
    {
        Node node = object.getNode();
        
        try
        {
            if (node.hasProperty(RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ':' + METADATA_WORKFLOW_ID))
            {
                throw new AmetysRepositoryException("Cannot call setWorkflowId on an already persisted Content");
            }
            
            node.setProperty(RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ':' + METADATA_WORKFLOW_ID, workflowId);
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to set workflowId property", e);
        }
    }
    
    /**
     * Remove the workflow ID property if exists on an object
     * @param object The Ametys object
     */
    public static void removeWorkflowId (JCRAmetysObject object)
    {
        Node node = object.getNode();
        
        try
        {
            if (node.hasProperty(RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ':' + METADATA_WORKFLOW_ID))
            {
                node.getProperty(RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ':' + METADATA_WORKFLOW_ID).remove();
            }
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to remove workflowId property", e);
        }
    }
    
    /**
     * Get an object's current step ID.
     * @param object the object.
     * @return the object's current step ID.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public static long getCurrentStepId(JCRAmetysObject object) throws AmetysRepositoryException
    {
        try
        {
            return object.getNode().getProperty(RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ':' + METADATA_CURRENT_STEP_ID).getLong();
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to get currentStepId property", e);
        }
    }
    
    /**
     * Set an object's current step ID.
     * @param object the object.
     * @param stepId the current step ID to set.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public static void setCurrentStepId(JCRAmetysObject object, long stepId) throws AmetysRepositoryException
    {
        Node node = object.getNode();
        
        try
        {
            node.setProperty(RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ':' + METADATA_CURRENT_STEP_ID, stepId);
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to set currentStepId property", e);
        }
    }
    
    /**
     * Get an object's proposal date.
     * @param object the object.
     * @return the object's proposal date.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public static ZonedDateTime getProposalDate(ModifiableDataAwareVersionableAmetysObject object) throws AmetysRepositoryException
    {
        return object.getUnversionedDataHolder().getValue(METADATA_LAST_PROPOSAL_DATE, null);
    }
    
    /**
     * Set an object's proposal date.
     * @param object the object.
     * @param proposalDate the proposal date to set.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public static void setProposalDate(ModifiableDataAwareVersionableAmetysObject object, ZonedDateTime proposalDate) throws AmetysRepositoryException
    {
        if (proposalDate != null)
        {
            object.getUnversionedDataHolder().setValue(METADATA_LAST_PROPOSAL_DATE, proposalDate, ModelItemTypeConstants.DATETIME_TYPE_ID);
        }
        else
        {
            object.getUnversionedDataHolder().removeValue(METADATA_LAST_PROPOSAL_DATE);
        }
    }
}
