/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository;

import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.query.expression.Expression;

/**
 * Constructs an {@link Expression} testing the current workflow step of a content.
 * This can be used only to search on the current step. To search on history steps, use the {@link CurrentStepExpression} instead
 */
public class WorkflowStepExpression implements Expression
{
    /** The logical operator to use in xpath query */
    public enum LogicalOperator 
    {
        /** Logical operator AND */
        AND
        {
            @Override
            public String toString()
            {
                return " and ";
            }
        },
        /** Logical operator OR */
        OR
        {
            @Override
            public String toString()
            {
                return " or ";
            }
        }
    }
    
    private Operator _operator;
    private int[] _values;
    private LogicalOperator _logicalOperator;

    /**
     * Creates the expression.
     * @param operator the operator to make the comparison (only Operator.EQ and Operator.NE allowed)
     * @param value the step value 
     */
    public WorkflowStepExpression(Operator operator, int value)
    {
        if (Operator.EQ != operator && Operator.NE != operator)
        {
            throw new AmetysRepositoryException("Test operator '" + "' is unknown for test's expression.");
        }

        _operator = operator;
        _values = new int[]{value};
    }
    
    /**
     * Creates the expression.
     * @param operator the operator to make the comparison (only Operator.EQ and Operator.NE allowed)
     * @param values the tags value in a array
     * @param logicalOperator the logical operator to use for given value
     */
    public WorkflowStepExpression(Operator operator, int[] values, LogicalOperator logicalOperator)
    {
        if (Operator.EQ != operator && Operator.NE != operator)
        {
            throw new AmetysRepositoryException("Test operator '" + "' is unknown for test's expression.");
        }

        _operator = operator;
        _values = values;
        _logicalOperator = logicalOperator;
    }
    
    @Override
    public String build()
    {
        StringBuffer sb = new StringBuffer();
        
        if (_values.length > 0)
        {
            sb.append('(');
            for (int i = 0; i < _values.length; i++)
            {
                if (i != 0)
                {
                    sb.append(_logicalOperator);
                }
                sb.append("@").append(RepositoryConstants.NAMESPACE_PREFIX_INTERNAL).append(':').append(WorkflowAwareContentHelper.METADATA_CURRENT_STEP_ID).append(' ').append(_operator).append(' ').append(_values[i]);
            }
            sb.append(')');
        }
        
        return sb.toString();
    }
}
