/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.repository.comment;

import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.repository.CommentableAmetysObject;
import org.ametys.cms.repository.ReactionableObject;
import org.ametys.cms.repository.ReactionableObjectHelper;
import org.ametys.cms.repository.ReportableObject;
import org.ametys.cms.repository.ReportableObjectHelper;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.data.holder.ModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.ModifiableModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelLessComposite;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.type.ModelItemTypeConstants;

/**
 * A comment on a {@link CommentableAmetysObject}
 */
public class Comment implements ReactionableObject, ReportableObject
{
    /** Constants for comments Metadata not validted */
    public static final String METADATA_COMMENTS_VALIDATED = "validated";
    /** Constants for comments Metadata validated */
    public static final String METADATA_COMMENTS_NOTVALIDATED = "not-validated";

    /** Constants for sub comments data name */
    public static final String SUB_COMMENTS_DATA_NAME = "comments";
    
    /** Constants for creation Metadata */
    public static final String METADATA_COMMENT_CREATIONDATE = "creation";
    /** Constants for author Metadata */
    public static final String METADATA_COMMENT_AUTHOR = "author";
    /** Constants for author name Metadata */
    public static final String METADATA_COMMENT_AUTHORNAME = "author-name";
    /** Constants for author email Metadata */
    public static final String METADATA_COMMENT_AUTHOREMAIL = "author-email";
    /** Constants for author email hidden Metadata */
    public static final String METADATA_COMMENT_AUTHOREMAIL_HIDDEN = "author-email-hidden";
    /** Constants for author url Metadata */
    public static final String METADATA_COMMENT_AUTHORURL = "author-url";
    /** Constants for the content Metadata */
    public static final String METADATA_COMMENT_CONTENT = "content";
    /** Constants for the validated status Metadata */
    public static final String METADATA_COMMENT_VALIDATED = "validated";
    /** Constants for the is edited status Metadata */
    public static final String METADATA_COMMENT_IS_EDITED = "is-edited";
    /** Constants for the is deleted status Metadata */
    public static final String METADATA_COMMENT_IS_DELETED = "is-deleted";
    /** Constants for the is marked status Metadata */
    public static final String METADATA_COMMENT_IS_ACCEPTED = "is-accepted";
    
    /** Constants for the separator */
    public static final String ID_SEPARATOR = "_";
    
    /** Regex used to extract mentions from a text. A mention is inside a @() */
    private static final String __EXTRACT_MENTIONS_REGEX = "@\\(([^()]+)\\)";
    
    /** The root data holder where to store the first level of comments */
    protected ModifiableModelLessDataHolder _rootDataHolder;
    /** The node of the comment */
    protected ModifiableModelLessComposite _commentComposite;
    /** The id of the comment (unique in the ametys object) */
    protected String _id;
    
    /**
     * Retrieves a comment by its id
     * @param rootDataHolder The root data holder hosting the first level of comments
     * @param commentId The id of the comment to retrieve
     * @throws AmetysRepositoryException if an error occurred
     */
    public Comment(ModifiableModelLessDataHolder rootDataHolder, String commentId)
    {
        _rootDataHolder = rootDataHolder;
        _id = commentId;
        
        String commentDataPath = getCommentDataPath(_id);
        _commentComposite = _rootDataHolder.getComposite(commentDataPath);
    }
    
    /**
     * Creates a new comment
     * @param dataHolder The data holder where to add the new comment 
     */
    public Comment(ModifiableModelLessDataHolder dataHolder)
    {
        this(dataHolder, Optional.empty(), Optional.empty());
    }
    
    /**
     * Creates a new comment, with the given id and creation date
     * This method allows to create a comment from existing data (ex: data import from archive) 
     * The id is not generated here, the source is trusted. Be careful using this method
     * @param dataHolder The data holder where to add the new comment
     * @param commentId the comment's id
     * @param creationDate the comment's creation date
     */
    public Comment(ModifiableModelLessDataHolder dataHolder, String commentId, ZonedDateTime creationDate)
    {
        this(dataHolder, Optional.ofNullable(commentId), Optional.ofNullable(creationDate));
    }
    
    private Comment(ModifiableModelLessDataHolder dataHolder, Optional<String> commentId, Optional<ZonedDateTime> creationDate)
    {
        _rootDataHolder = dataHolder;
        
        _id = commentId.orElseGet(() -> _getNextCommentName(_rootDataHolder));
        _commentComposite = _rootDataHolder.getComposite(_id, true);
        _commentComposite.setValue(METADATA_COMMENT_CREATIONDATE, creationDate.orElseGet(ZonedDateTime::now));
        
        update();
    }
    
    /**
     * Creates a new sub comment of the comment
     * @param comment The parent comment
     */
    public Comment(Comment comment)
    {
        this(comment, Optional.empty(), Optional.empty());
    }
    
    /**
     * Creates a new sub comment of the comment, with the given id and creation date
     * This method allow to create a sub comment from existing data (ex: data import from archive) 
     * The id is not generated here, the source is trusted. Be careful using this method
     * @param comment The parent comment
     * @param commentId the sub comment's id
     * @param creationDate the sub comment's creation date
     */
    public Comment(Comment comment, String commentId, ZonedDateTime creationDate)
    {
        this(comment, Optional.ofNullable(commentId), Optional.ofNullable(creationDate));
    }
    
    private Comment(Comment comment, Optional<String> commentId, Optional<ZonedDateTime> creationDate)
    {
        _rootDataHolder = comment._rootDataHolder;
        ModifiableModelLessComposite subCommentsDataHolder = comment._commentComposite.getComposite(SUB_COMMENTS_DATA_NAME, true);
        
        String commentName = commentId.map(id -> StringUtils.substringAfterLast(id, ID_SEPARATOR))
                .orElseGet(() -> _getNextCommentName(subCommentsDataHolder));
        _id = commentId.orElseGet(() -> comment.getId() + ID_SEPARATOR + commentName);
        
        _commentComposite = subCommentsDataHolder.getComposite(commentName, true);
        _commentComposite.setValue(METADATA_COMMENT_CREATIONDATE, creationDate.orElseGet(ZonedDateTime::now));
        
        update();
    }
    
    private static String _getNextCommentName(ModelLessDataHolder dataHolder)
    {
        String base = "comment-";
        int i = 0;
        while (dataHolder.hasValueOrEmpty(base + i))
        {
            i++;
        }
        
        return base + i;
    }
    
    /**
     * Retrieves the path of the comment with the given identifier
     * @param commentId the comment identifier
     * @return the path of the comment
     */
    protected String getCommentDataPath(String commentId)
    {
        String[] commentIdSegments = commentId.split(ID_SEPARATOR);
        return StringUtils.join(commentIdSegments, ModelItem.ITEM_PATH_SEPARATOR + SUB_COMMENTS_DATA_NAME + ModelItem.ITEM_PATH_SEPARATOR);
    }
    
    /**
     * Retrieves the repository data of the {@link Comment}
     * @return the repository data of the {@link Comment}
     */
    public ModifiableRepositoryData getRepositoryData()
    {
        return _commentComposite.getRepositoryData();
    }
    
    /**
     * The comment id (unique to the ametys object)
     * @return The id. Cannot be null.
     */
    public String getId()
    {
        return _id;
    }
    
    /**
     * Get the date and time the comment was created
     * @return The non null date of creation of the comment.
     */
    public ZonedDateTime getCreationDate()
    {
        return _commentComposite.getValue(METADATA_COMMENT_CREATIONDATE);
    }
    
    /**
     * Get the comment's author
     * @return the comment's author
     */
    public UserIdentity getAuthor()
    {
        return _commentComposite.getValueOfType(METADATA_COMMENT_AUTHOR, org.ametys.cms.data.type.ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
    }
    
    /**
     * Set the comment's author.
     * @param author the author. Can be null to remove the name.
     */
    public void setAuthor(UserIdentity author)
    {
        _commentComposite.setValue(METADATA_COMMENT_AUTHOR, author, org.ametys.cms.data.type.ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
    }
    
    /**
     * Get the readable name of the author.
     * @return The full name. Can be null.
     */
    public String getAuthorName()
    {
        return _commentComposite.getValueOfType(METADATA_COMMENT_AUTHORNAME, ModelItemTypeConstants.STRING_TYPE_ID);
    }
    
    /**
     * Set the readable name of the author.
     * @param name The full name. Can be null to remove the name.
     */
    public void setAuthorName(String name)
    {
        _commentComposite.setValue(METADATA_COMMENT_AUTHORNAME, name, ModelItemTypeConstants.STRING_TYPE_ID);
    }
    
    /**
     * Get the email of the author.
     * @return The ameil. Can be null.
     */
    public String getAuthorEmail()
    {
        return _commentComposite.getValueOfType(METADATA_COMMENT_AUTHOREMAIL, ModelItemTypeConstants.STRING_TYPE_ID);
    }
    
    /**
     * Set the email of the author.
     * @param email The email. Can be null to remove the email.
     */
    public void setAuthorEmail(String email)
    {
        _commentComposite.setValue(METADATA_COMMENT_AUTHOREMAIL, email, ModelItemTypeConstants.STRING_TYPE_ID);
    }
    
    /**
     * Get the url given by the author as its personal site url.
     * @return The url. Can be null.
     */
    public String getAuthorURL()
    {
        return _commentComposite.getValueOfType(METADATA_COMMENT_AUTHORURL, ModelItemTypeConstants.STRING_TYPE_ID);
    }
    
    /**
     * Set the personal site url of the author
     * @param url The url. Can be null to remove url.
     */
    public void setAuthorURL(String url)
    {
        _commentComposite.setValue(METADATA_COMMENT_AUTHORURL, url, ModelItemTypeConstants.STRING_TYPE_ID);
    }
    
    /**
     * Does the email of the authors have to be hidden ?
     * @return true (default value) if the email does not have to appears to others users. Can still be used for administration.
     */
    public boolean isEmailHidden()
    {
        return _commentComposite.getValueOfType(METADATA_COMMENT_AUTHOREMAIL_HIDDEN, ModelItemTypeConstants.BOOLEAN_TYPE_ID, true);
    }
    /**
     * Set the email hidden status.
     * @param hideEmail true to set the email as hidden.
     */
    public void setEmailHiddenStatus(boolean hideEmail)
    {
        _commentComposite.setValue(METADATA_COMMENT_AUTHOREMAIL_HIDDEN, hideEmail, ModelItemTypeConstants.BOOLEAN_TYPE_ID);
    }
    
    /**
     * Does the comment is edited
     * @return true the email is edited
     */
    public boolean isEdited()
    {
        return _commentComposite.getValueOfType(METADATA_COMMENT_IS_EDITED, ModelItemTypeConstants.BOOLEAN_TYPE_ID, false);
    }
    /**
     * Set the comment to edited.
     * @param isEdited true to set the comment to edited.
     */
    public void setEdited(boolean isEdited)
    {
        _commentComposite.setValue(METADATA_COMMENT_IS_EDITED, isEdited, ModelItemTypeConstants.BOOLEAN_TYPE_ID);
    }
    
    /**
     * Does the comment is deleted
     * @return true the comment is deleted
     */
    public boolean isDeleted()
    {
        return _commentComposite.getValueOfType(METADATA_COMMENT_IS_DELETED, ModelItemTypeConstants.BOOLEAN_TYPE_ID, false);
    }
    /**
     * Set the comment to deleted.
     * @param isEdited true to set the comment to deleted.
     */
    public void setDeleted(boolean isEdited)
    {
        _commentComposite.setValue(METADATA_COMMENT_IS_DELETED, isEdited, ModelItemTypeConstants.BOOLEAN_TYPE_ID);
    }
    
    /**
     * Get the content of the comment. A simple String (with \n or \t).
     * @return The content. Can be null.
     */
    public String getContent()
    {
        return _commentComposite.getValueOfType(METADATA_COMMENT_CONTENT, ModelItemTypeConstants.STRING_TYPE_ID);
    }
    /**
     * Set the content of the comment.
     * @param content The content to set. Can be null to remove the content. Have to be a simple String (with \n or \t).
     */
    public void setContent(String content)
    {
        _commentComposite.setValue(METADATA_COMMENT_CONTENT, content, ModelItemTypeConstants.STRING_TYPE_ID);
    }
    
    /**
     * Extract users mentioned in the comment
     * @return the mentioned users
     */
    public Collection<UserIdentity> extractMentions()
    {
        Set<UserIdentity> mentionendUsers = new HashSet<>();
        
        String content = getContent();
        if (content != null)
        {
            Pattern pattern = Pattern.compile(__EXTRACT_MENTIONS_REGEX);
            Matcher matcher = pattern.matcher(content);
            
            while (matcher.find())
            {
                String userIdentityAsString = matcher.group(1);
                UserIdentity userIdentity = UserIdentity.stringToUserIdentity(userIdentityAsString);
                mentionendUsers.add(userIdentity);
            }
        }
        
        return mentionendUsers;
    } 
    
    /**
     * Is the comment validated
     * @return the status of validation of the comment
     */
    public boolean isValidated()
    {
        return _commentComposite.getValueOfType(METADATA_COMMENT_VALIDATED, ModelItemTypeConstants.BOOLEAN_TYPE_ID, false);
    }
    
    /**
     * Set the validation status of the comment
     * @param validated true the comment is validated
     */
    public void setValidated(boolean validated)
    {
        _commentComposite.setValue(METADATA_COMMENT_VALIDATED, validated, ModelItemTypeConstants.BOOLEAN_TYPE_ID);
        update();
    }
    
    public void addReport()
    {
        ReportableObjectHelper.addReport(_commentComposite);
    }
    
    public void setReportsCount(long reportsCount)
    {
        ReportableObjectHelper.setReportsCount(_commentComposite, reportsCount);
    }

    public void clearReports()
    {
        ReportableObjectHelper.clearReports(_commentComposite);
    }

    public long getReportsCount()
    {
        return ReportableObjectHelper.getReportsCount(_commentComposite);
    }

    /**
     * Remove the comment.
     */
    public void remove()
    {
        String commentDataPath = getCommentDataPath(_id);
        _rootDataHolder.removeValue(commentDataPath);
        update();
    }
    
    /**
     * Get sub comments of the comment
     * @param includeNotValidatedComments True to include the comments that are not validated
     * @param includeValidatedComments True to include the comments that are validated
     * @return the list of comments
     */
    public List<Comment> getSubComment(boolean includeNotValidatedComments, boolean includeValidatedComments)
    {
        return getComments(this, includeNotValidatedComments, includeValidatedComments);
    }
    
    /**
     * Create sub comment from this comment
     * @return the sub comment
     */
    public Comment createSubComment()
    {
        return new Comment(this);
    }
    
    /**
     * Creates a sub comment from this comment, with the given id and creation date
     * This method allow to create a sub comment from existing data (ex: data import from archive) 
     * The id is not generated here, the source is trusted. Be careful using this method
     * @param commentId the comment's id
     * @param creationDate the comment's creation date
     * @return the new comment
     */
    public Comment createSubComment(String commentId, ZonedDateTime creationDate)
    {
        return new Comment(this, commentId, creationDate);
    }
    
    /**
     * Update the comment tag statistics
     */
    protected void update()
    {
        long validated = 0;
        long notValidated = 0;
        
        ModifiableModelLessDataHolder dataHolder = _rootDataHolder;
        if (isSubComment())
        {
            dataHolder = getCommentParent()._rootDataHolder;
        }
        
        List<Comment> comments = getComments(dataHolder, true, true, false);
        
        for (Comment comment : comments)
        {
            if (comment.isValidated())
            {
                validated++;
            }
            else
            {
                notValidated++;
            }
        }
        
        dataHolder.setValue(METADATA_COMMENTS_VALIDATED, validated);
        dataHolder.setValue(METADATA_COMMENTS_NOTVALIDATED, notValidated);
    }

    /**
     * Get a comment
     * @param rootDataHolder The root data holder of the first level of comments
     * @param commentId The comment identifier
     * @return The comment
     * @throws AmetysRepositoryException if the comment does not exist
     */
    public static Comment getComment(ModifiableModelLessDataHolder rootDataHolder, String commentId) throws AmetysRepositoryException
    {
        return new Comment(rootDataHolder, commentId);
    }
    
    /**
     * Get the sub comments of the given comment
     * @param parentComment The parent comment
     * @param includeNotValidatedComments True to include the comments that are not validated
     * @param includeValidatedComments True to include the comments that are validated
     * @return the list of sub comments
     * @throws AmetysRepositoryException If an error occurred
     */
    public static List<Comment> getComments(Comment parentComment, boolean includeNotValidatedComments, boolean includeValidatedComments) throws AmetysRepositoryException
    {
        return getComments(parentComment, includeNotValidatedComments, includeValidatedComments, false);
    }
    
    /**
     * Get the sub comments of the given comment
     * @param parentComment The parent comment
     * @param includeNotValidatedComments True to include the comments that are not validated
     * @param includeValidatedComments True to include the comments that are validated
     * @param withSubComment true if we want to get all child comments
     * @return the list of comments
     * @throws AmetysRepositoryException If an error occurred
     */
    public static List<Comment> getComments(Comment parentComment, boolean includeNotValidatedComments, boolean includeValidatedComments, boolean withSubComment) throws AmetysRepositoryException
    {
        ModifiableModelLessComposite parentComposite = parentComment._commentComposite;
        List<Comment> comments = new ArrayList<>();
        
        if (parentComposite.hasValue(SUB_COMMENTS_DATA_NAME, org.ametys.plugins.repository.data.type.ModelItemTypeConstants.COMPOSITE_TYPE_ID))
        {
            ModifiableModelLessComposite subCommentsComposite = parentComposite.getComposite(SUB_COMMENTS_DATA_NAME);
            return _getComments(parentComment._rootDataHolder, subCommentsComposite, parentComment.getId() + ID_SEPARATOR, includeNotValidatedComments, includeValidatedComments, withSubComment);
        }
        
        return comments;
    }
    
    /**
     * Get the comments in the given root data holder
     * @param rootDataHolder The root data holder
     * @param includeNotValidatedComments True to include the comments that are not validated
     * @param includeValidatedComments True to include the comments that are validated
     * @return the list of comments
     * @throws AmetysRepositoryException If an error occurred
     */
    public static List<Comment> getComments(ModifiableModelLessDataHolder rootDataHolder, boolean includeNotValidatedComments, boolean includeValidatedComments) throws AmetysRepositoryException
    {
        return getComments(rootDataHolder, includeNotValidatedComments, includeValidatedComments, false);
    }
    
    /**
     * Get the comments in the given root data holder
     * @param rootDataHolder The root data holder
     * @param includeNotValidatedComments True to include the comments that are not validated
     * @param includeValidatedComments True to include the comments that are validated
     * @param isRecursive true if we want to have sub comments
     * @return the list of comments
     * @throws AmetysRepositoryException If an error occurred
     */
    public static List<Comment> getComments(ModifiableModelLessDataHolder rootDataHolder, boolean includeNotValidatedComments, boolean includeValidatedComments, boolean isRecursive) throws AmetysRepositoryException
    {
        return _getComments(rootDataHolder, rootDataHolder, StringUtils.EMPTY, includeNotValidatedComments, includeValidatedComments, isRecursive);
    }
    
    private static List<Comment> _getComments(ModifiableModelLessDataHolder rootDataHolder, ModifiableModelLessDataHolder dataHolder, String commentIdPrefix, boolean includeNotValidatedComments, boolean includeValidatedComments, boolean isRecursive)
    {
        List<Comment> comments = new ArrayList<>();
        
        for (String name : dataHolder.getDataNames())
        {
            if (METADATA_COMMENTS_NOTVALIDATED.equals(name) || METADATA_COMMENTS_VALIDATED.equals(name))
            {
                continue;
            }

            String id = commentIdPrefix + name;
            Comment c = new Comment(rootDataHolder, id);
            if (includeNotValidatedComments && !c.isValidated() || includeValidatedComments && c.isValidated())
            {
                comments.add(c);
                if (isRecursive)
                {
                    comments.addAll(getComments(c, includeNotValidatedComments, includeValidatedComments, isRecursive));
                }
            }
        }
        
        return comments;
    }

    @Override
    public void addReaction(UserIdentity user, ReactionType reactionType)
    {
        ReactionableObjectHelper.addReaction(_commentComposite, user, reactionType);
    }

    @Override
    public void removeReaction(UserIdentity user, ReactionType reactionType)
    {
        ReactionableObjectHelper.removeReaction(_commentComposite, user, reactionType);
    }

    @Override
    public List<UserIdentity> getReactionUsers(ReactionType reactionType)
    {
        return ReactionableObjectHelper.getReactionUsers(_commentComposite, reactionType);
    }
    
    /**
     * Get parent of comment if exists
     * @return the comment parent. null if the comment is not a sub comment
     */
    public Comment getCommentParent()
    {
        if (isSubComment())
        {
            String parentId = StringUtils.substringBeforeLast(this.getId(), ID_SEPARATOR);
            return getComment(_rootDataHolder, parentId);
        }
        return null;
    }

    /**
     * Check if the comment is a sub-comment from an other comment
     * @return true if the comment is a sub-comment
     */
    public boolean isSubComment()
    {
        return this.getId().contains(ID_SEPARATOR);
    }
    
    /**
     * Check if the comment have any sub-comments
     * @return true if the comment have any sub-comments
     */
    public boolean hasSubComments()
    {
        return !this.getSubComment(true, true).isEmpty();
    }
    
    /**
     * Is the comment the accepted answer
     * @return true if the comment is the accepted answer
     */
    public boolean isAccepted()
    {
        return _commentComposite.getValueOfType(METADATA_COMMENT_IS_ACCEPTED, ModelItemTypeConstants.BOOLEAN_TYPE_ID, false);
    }
    
    /**
     * Set the comment as accepted answer.
     * @param isAccepted true to set the comment as accepted answer.
     */
    public void setAccepted(boolean isAccepted)
    {
        _commentComposite.setValue(METADATA_COMMENT_IS_ACCEPTED, isAccepted, ModelItemTypeConstants.BOOLEAN_TYPE_ID);
    }
    
}
