/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository.comment;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.regex.Pattern;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang.StringUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ReactionableObject.ReactionType;
import org.ametys.cms.repository.ReactionableObjectHelper;
import org.ametys.cms.repository.ReportableObjectHelper;
import org.ametys.cms.repository.comment.contributor.ContributorCommentableAmetysObject;
import org.ametys.cms.rights.ContentRightAssignmentContext;
import org.ametys.core.captcha.CaptchaHelper;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.ui.Callable;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.user.directory.NotUniqueUserException;
import org.ametys.core.user.population.PopulationContextHelper;
import org.ametys.core.user.population.UserPopulation;
import org.ametys.core.user.population.UserPopulationDAO;
import org.ametys.core.util.DateUtils;
import org.ametys.core.util.mail.SendMailHelper;
import org.ametys.plugins.core.ui.user.ProfileImageResolverHelper;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.provider.RequestAttributeWorkspaceSelector;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * DAO for content's comments
 *
 */
public class CommentsDAO extends AbstractLogEnabled implements Component, Serviceable, Contextualizable
{
    /** The Avalon role */
    public static final String ROLE = CommentsDAO.class.getName();
    
    /** The form input name for author name */
    public static final String FORM_AUTHOR_NAME = "name";
    /** The form input name for author email */
    public static final String FORM_AUTHOR_EMAIL = "email";
    /** The form input name for author email to hide */
    public static final String FORM_AUTHOR_HIDEEMAIL = "hide-email";
    /** The form input name for author url */
    public static final String FORM_AUTHOR_URL = "url";
    /** The form input name for content */
    public static final String FORM_CONTENTTEXT = "text";
    /** The form input name for captcha */
    public static final String FORM_CAPTCHA_KEY = "captcha-key";
    /** The form input name for captcha */
    public static final String FORM_CAPTCHA_VALUE = "captcha-value";
    
    /** The pattern to check url */
    public static final Pattern URL_VALIDATOR = Pattern.compile("^(https?:\\/\\/.+)?$");
    
    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /** The observation manager */
    protected ObservationManager _observationManager;
    /** The currenrt user provider */
    protected CurrentUserProvider _userProvider;
    /** The user helper */
    protected UserHelper _userHelper;
    /** The right manager */
    protected RightManager _rightManager;
    /** Helper for reactionable object */
    protected ReactionableObjectHelper _reactionableHelper;
    /** The user manager */
    protected UserManager _userManager;
    /** The population context helper */
    protected PopulationContextHelper _populationContextHelper;
    /** The user population DAO */
    protected UserPopulationDAO _userPopulationDAO;
    
    /** The avalon context */
    protected Context _context;

    public void service(ServiceManager smanager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _userProvider = (CurrentUserProvider) smanager.lookup(CurrentUserProvider.ROLE);
        _observationManager = (ObservationManager) smanager.lookup(ObservationManager.ROLE);
        _userHelper = (UserHelper) smanager.lookup(UserHelper.ROLE);
        _rightManager = (RightManager) smanager.lookup(RightManager.ROLE);
        _reactionableHelper = (ReactionableObjectHelper) smanager.lookup(ReactionableObjectHelper.ROLE);
        _populationContextHelper = (PopulationContextHelper) smanager.lookup(PopulationContextHelper.ROLE);
        _userManager = (UserManager) smanager.lookup(UserManager.ROLE);
        _userPopulationDAO = (UserPopulationDAO) smanager.lookup(UserPopulationDAO.ROLE);
    }
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    /**
     * Get the content's comments
     * @param contentId the content id
     * @param contextualParameters the contextual parameters
     * @return the comments
     * @throws IllegalAccessException if user is not allowed to get comments
     */
    @Callable (allowAnonymous = true, rights = Callable.READ_ACCESS, rightContext = ContentRightAssignmentContext.ID, paramIndex = 0)
    public List<Map<String, Object>> getComments(String contentId, Map<String, Object> contextualParameters) throws IllegalAccessException
    {
        Content content = _resolver.resolveById(contentId);
        
        List<Map<String, Object>> comments2json = new ArrayList<>();
        
        if (content instanceof CommentableContent commentableContent)
        {
            List<Comment> comments = commentableContent.getComments(false, true);
            for (Comment comment : comments)
            {
                comments2json.add(getComment(content, comment, 0, contextualParameters));
            }
        }   
        
        return comments2json;
    }
    
    /**
     * Get the content's contributor comments
     * @param contentId the content id
     * @param contextualParameters the contextual parameters
     * @return the comments
     */
    public List<Map<String, Object>> getContributorComments(String contentId, Map<String, Object> contextualParameters)
    {
        Content content = _resolver.resolveById(contentId);
        
        List<Map<String, Object>> comments2json = new ArrayList<>();
        
        if (content instanceof ContributorCommentableAmetysObject commentableContent)
        {
            List<Comment> comments = commentableContent.getContributorComments();
            for (Comment comment : comments)
            {
                comments2json.add(getComment(content, comment, 0, contextualParameters));
            }
        }   
        
        return comments2json;
    }
    
    /**
     * Add a new comment
     * @param contentId the content id
     * @param commentId the id of of parent comment. Can be null if it is not a subcomment
     * @param formValues the form's values
     * @param contextualParameters the contextual parameters
     * @return the results
     * @throws IllegalAccessException if anonymous or current user has no access to the content
     */
    @Callable (allowAnonymous = true, rights = Callable.READ_ACCESS, rightContext = ContentRightAssignmentContext.ID, paramIndex = 0)
    public Map<String, Object> addComment(String contentId, String commentId, Map<String, Object> formValues, Map<String, Object> contextualParameters) throws IllegalAccessException
    {
        CommentableContent cContent = getContent(contentId);
        
        List<Map<String, Object>> errors = getErrors(cContent, formValues);

        if (errors.isEmpty())
        {
            Map<String, Object> results = new HashMap<>();
            
            if (StringUtils.isNotBlank(commentId))
            {
                Comment comment = cContent.getComment(commentId);
                Comment subComment = comment.createSubComment();
                results = _setCommentAttributes(cContent, subComment, formValues);
                results.put("comment", getComment(cContent, subComment, 1, contextualParameters));
            }
            else
            {
                Comment comment = cContent.createComment();
                results = _setCommentAttributes(cContent, comment, formValues);
                results.put("comment", getComment(cContent, comment, 0, contextualParameters));
            }
            
            return results;
        }
        else
        {
            Map<String, Object> results = new HashMap<>();
            results.put("errors", errors);
            return results;
        }
    }
    
    /**
     * Delete a comment
     * @param contentId the content id
     * @param commentId the comment id to remove
     * @return the results
     * @throws IllegalAccessException if current user is null
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> deleteComment(String contentId, String commentId) throws IllegalAccessException
    {
        CommentableContent content = getContent(contentId);
        Comment comment = content.getComment(commentId);
        
        if (!canDeleteComment(content, comment))
        {
            throw new IllegalAccessException("User " + getCurrentUser() + " tried to delete comment on content " + content + " without sufficient rights");
        }
        
        Map<String, Object> results = new HashMap<>();
        
        Map<String, Object> eventParams = new HashMap<>();
        eventParams.put(ObservationConstants.ARGS_CONTENT, content);
        eventParams.put(ObservationConstants.ARGS_COMMENT_ID, comment.getId());
        eventParams.put(ObservationConstants.ARGS_COMMENT_AUTHOR, comment.getAuthorName());
        eventParams.put(ObservationConstants.ARGS_COMMENT_AUTHOR_EMAIL, comment.getAuthorEmail());
        eventParams.put(ObservationConstants.ARGS_COMMENT_VALIDATED, comment.isValidated());
        eventParams.put(ObservationConstants.ARGS_COMMENT_CREATION_DATE, comment.getCreationDate());
        
        eventParams.put("comment.content", comment.getContent());

        _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_COMMENT_DELETING, getCurrentUser(), eventParams));

        comment.remove();
        content.saveChanges();

        _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_COMMENT_DELETED, getCurrentUser(), eventParams));

        results.put("deleted", true);
        results.put("contentId", content.getId());
        results.put("commentId", comment.getId());
        
        return results;
    }
    
    /**
     * Determines if current user is allowed to delete a comment 
     * @param contentId the content id
     * @return true if the current user is allowed
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public boolean canDeleteComment(String contentId)
    {
        CommentableContent content = getContent(contentId);
        UserIdentity currentUser = getCurrentUser();
        return currentUser != null && _rightManager.hasRight(currentUser, "CMS_Rights_CommentModerate", content).equals(RightResult.RIGHT_ALLOW);
    }
    
    /**
     * Determines if current user is allowed to delete a comment 
     * @param content the content
     * @param comment the comment
     * @return true if the current user is allowed
     */
    public boolean canDeleteComment(Content content, Comment comment)
    {
        boolean isUserOwnComment = false;
        
        UserIdentity currentUser = getCurrentUser();
        if (currentUser != null)
        {
            User user = _userManager.getUser(currentUser);
            String authorEmail = comment.getAuthorEmail();
            isUserOwnComment = user != null  && StringUtils.equals(authorEmail, user.getEmail());
        }
        
        return isUserOwnComment || currentUser != null && _rightManager.hasRight(currentUser, "CMS_Rights_CommentModerate", content).equals(RightResult.RIGHT_ALLOW);
    }

    /**
     * React (like or unlike) to a comment
     * @param contentId the content id
     * @param commentId the comment id
     * @return the results
     * @throws IllegalAccessException if current user is null or user has no access to the content
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> likeOrUnlikeComment(String contentId, String commentId) throws IllegalAccessException
    {
        return likeOrUnlikeComment(contentId, commentId, null);
    }
    
    /**
     * React (like or unlike) to a comment
     * @param contentId the content id
     * @param commentId the comment id
     * @param remove true to remove the reaction, false to add reaction. If null, check if the current user has already like the comment.
     * @return the results
     * @throws IllegalAccessException if current user is null or user has no access to the content
     */
    @Callable  (rights = Callable.READ_ACCESS, rightContext = ContentRightAssignmentContext.ID, paramIndex = 0)
    public Map<String, Object> likeOrUnlikeComment(String contentId, String commentId, Boolean remove) throws IllegalAccessException
    {
        Map<String, Object> results = new HashMap<>();
        
        CommentableContent content = getContent(contentId);
        Comment comment = content.getComment(commentId);
        
        UserIdentity currentUser = getCurrentUser();
        if (currentUser == null)
        {
            throw new IllegalAccessException("Anonymous user is not allowed to react to a comment");
        }
        
        List<UserIdentity> likers = comment.getReactionUsers(ReactionType.LIKE);
        int nbLikes = likers.size();
        
        if (Boolean.TRUE.equals(remove)
            || remove == null && likers.contains(currentUser))
        {
            comment.removeReaction(currentUser, ReactionType.LIKE);
            results.put("liked", false);
            nbLikes--;
        }
        else
        {
            comment.addReaction(currentUser, ReactionType.LIKE);
            results.put("liked", true);
            nbLikes++;
        }
        
        results.put("nbLikes", nbLikes);
        
        content.saveChanges();
        
        Map<String, Object> eventParams = new HashMap<>();
        eventParams.put(ObservationConstants.ARGS_CONTENT, content);
        eventParams.put(ObservationConstants.ARGS_COMMENT, comment);
        eventParams.put(ObservationConstants.ARGS_REACTION_TYPE, ReactionType.LIKE);
        eventParams.put(ObservationConstants.ARGS_REACTION_ISSUER, currentUser);
        _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_COMMENT_REACTION_CHANGED, getCurrentUser(), eventParams));
        
        results.put("contentId", content.getId());
        results.put("commentId", comment.getId());
        return results;
    }

    /**
     * Get the commentable content
     * @param contentId The content id
     * @return The content
     */
    protected CommentableContent getContent (String contentId)
    {
        Request request = ContextHelper.getRequest(_context);
        String currentWorkspace = RequestAttributeWorkspaceSelector.getForcedWorkspace(request);

        try
        {
            RequestAttributeWorkspaceSelector.setForcedWorkspace(request, "default");
            return _resolver.resolveById(contentId);
        }
        finally
        {
            RequestAttributeWorkspaceSelector.setForcedWorkspace(request, currentWorkspace);
        }
    }
    
    /**
     * Set comment attributes
     * @param content the content
     * @param comment the comment
     * @param formValues the form's values
     * @return the result
     */
    protected Map<String, Object> _setCommentAttributes(CommentableContent content, Comment comment, Map<String, Object> formValues)
    {
        String authorName = (String) formValues.get(FORM_AUTHOR_NAME);
        String authorEmail = (String) formValues.get(FORM_AUTHOR_EMAIL);
        String authorUrl = (String) formValues.get(FORM_AUTHOR_URL);
        String text = (String) formValues.get(FORM_CONTENTTEXT);
        boolean hideEmail = formValues.containsKey(FORM_AUTHOR_HIDEEMAIL) && Boolean.TRUE.equals(formValues.get(FORM_AUTHOR_HIDEEMAIL));
        
        comment.setAuthorName(authorName);
        comment.setAuthorEmail(authorEmail);
        comment.setEmailHiddenStatus(hideEmail);
        comment.setAuthorURL(authorUrl);
        comment.setContent(text.replaceAll("\r", ""));
        
        boolean isValidated = isValidatedByDefault(content);
        comment.setValidated(isValidated);
        
        content.saveChanges();
        
        if (isValidated)
        {
            Map<String, Object> eventParams = new HashMap<>();
            eventParams.put(ObservationConstants.ARGS_CONTENT, content);
            eventParams.put(ObservationConstants.ARGS_COMMENT, comment);
            _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_COMMENT_VALIDATED, getCurrentUser(), eventParams));
        }
        else 
        {
            Map<String, Object> eventParams = new HashMap<>();
            eventParams.put(ObservationConstants.ARGS_CONTENT, content);
            eventParams.put(ObservationConstants.ARGS_COMMENT, comment);
            _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_COMMENT_ADDED, getCurrentUser(), eventParams));
        }
        
        Map<String, Object> results = new HashMap<>();
        
        results.put("published", comment.isValidated());
        results.put("contentId", content.getId());
        results.put("commentId", comment.getId());
        
        return results;
    }
    
    /**
     * Report a comment
     * @param contentId the content id
     * @param commentId the comment id
     * @return the results
     * @throws IllegalAccessException if anonymous or current user has no access to the content
     */
    @Callable (allowAnonymous = true, rights = Callable.READ_ACCESS, rightContext = ContentRightAssignmentContext.ID, paramIndex = 0)
    public Map<String, Object> reportComment(String contentId, String commentId) throws IllegalAccessException
    {
        CommentableContent content = getContent(contentId);
        Comment comment = content.getComment(commentId);
        
        comment.addReport();
        content.saveChanges();
        
        Map<String, Object> eventParams = new HashMap<>();
        eventParams.put(ObservationConstants.ARGS_CONTENT, content);
        eventParams.put(ObservationConstants.ARGS_COMMENT, comment);
        _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_COMMENT_REPORTED, getCurrentUser(), eventParams));
        
        Map<String, Object> results = new HashMap<>();
        results.put("reported", true);
        results.put("contentId", content.getId());
        results.put("commentId", comment.getId());
        return results;
    }
    
    /**
     * Get the validation flag default value for a content asking all listeners
     * @param content The content having a new comment
     * @return a positive value if the comments have to be validated by default or a negative value in the other case. The absolute value is the priority of your listener. E.G. If a listener set +1 and another -10: the sum is negative (so comments not validated be default).
     */
    public boolean isValidatedByDefault(Content content)
    {
        boolean postValidation = Config.getInstance().getValue("cms.contents.comments.postvalidation");
        return postValidation;
    }
    
    /**
     * Checks if a captcha have to be checked.
     * @param content The content to comment
     * @param formValues The form values
     * @return true if the comments have to be protected by a captcha or false otherwise
     */
    public boolean isCaptchaRequired(Content content, Map<String, Object> formValues)
    {
        return false;
    }   
    
    /**
     * Get errors when submitting comment
     * @param content The content to comment
     * @param formValues The form values to submit a comment
     * @return An list of error messages (empty if no errors)
     */
    public List<Map<String, Object>> getErrors(CommentableContent content, Map<String, Object> formValues)
    {
        List<Map<String, Object>> errors = new ArrayList<>();

        String name = (String) formValues.get(FORM_AUTHOR_NAME);
        if (StringUtils.isBlank(name))
        {
            Map<String, Object> error = new HashMap<>();
            error.put("name", FORM_AUTHOR_NAME);
            error.put("error", new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_COMMENTS_ADD_ERROR_NAME"));
            errors.add(error);
        }
        
        String email = (String) formValues.get(FORM_AUTHOR_EMAIL);
        if (!SendMailHelper.EMAIL_VALIDATION.matcher(StringUtils.trimToEmpty(email)).matches())
        {
            Map<String, Object> error = new HashMap<>();
            error.put("name", FORM_AUTHOR_EMAIL);
            error.put("error", new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_COMMENTS_ADD_ERROR_EMAIL"));
            errors.add(error);
        }

        String url = (String) formValues.get(FORM_AUTHOR_URL);
        if (!URL_VALIDATOR.matcher(StringUtils.trimToEmpty(url)).matches())
        {
            Map<String, Object> error = new HashMap<>();
            error.put("name", FORM_AUTHOR_URL);
            error.put("error", new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_COMMENTS_ADD_ERROR_URL"));
            errors.add(error);
        }
        
        String text = (String) formValues.get(FORM_CONTENTTEXT);
        if (StringUtils.isBlank(text))
        {
            Map<String, Object> error = new HashMap<>();
            error.put("name", FORM_CONTENTTEXT);
            error.put("error", new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_COMMENTS_ADD_ERROR_CONTENT"));
            errors.add(error);
        }
        
        if (isCaptchaRequired(content, formValues))
        {
            String captchaKey = (String) formValues.get(FORM_CAPTCHA_KEY);
            String captchaValue = (String) formValues.get(FORM_CAPTCHA_VALUE);
            if (!CaptchaHelper.checkAndInvalidate(captchaKey, captchaValue))
            {
                Map<String, Object> error = new HashMap<>();
                error.put("name", FORM_CAPTCHA_VALUE);
                error.put("error", new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_COMMENTS_ADD_ERROR_CAPTCHA"));
                errors.add(error);
            }
        }

        return errors;
    }
    /**
     * Get the current user
     * @return The current user
     */
    protected UserIdentity getCurrentUser()
    {
        return _userProvider.getUser();
    }
    
    /**
     * Get JSON representation of a comment
     * @param content The content
     * @param comment the comment
     * @param level the level of comment (0 for parent comment, 1 for sub-comment, etc ....)
     * @param contextualParameters the contextual parameters
     * @return the comment as JSON
     */
    public Map<String, Object> getComment(Content content, Comment comment, int level, Map<String, Object> contextualParameters)
    {
        Map<String, Object> comment2json = new HashMap<>();

        comment2json.put("id", comment.getId());
        comment2json.put("content-id", content.getId());
        comment2json.put("creation-date", comment.getCreationDate());
        comment2json.put("level", level);
        
        comment2json.putAll(getCommentAuthor(comment, contextualParameters));
        
        if (comment.getContent() != null)
        {
            String commentContent = comment.getContent();
            comment2json.put("content", commentContent);
            
            List<Map<String, Object>> mentionedUsers2json = comment.extractMentions()
                                                                   .stream()
                                                                   .map(userIdentity -> getUserPropertiesFromIdentity(userIdentity, contextualParameters))
                                                                   .toList();
            if (!mentionedUsers2json.isEmpty())
            {
                comment2json.put("mentions", mentionedUsers2json);
            }
        }
        
        List<Map<String, Object>> reactions2json = _reactionableHelper.reactionsToJson(comment);
        comment2json.put("reactions", reactions2json);
        
        comment2json.put("reports", comment.getReportsCount());

        List<Comment> subComments = comment.getSubComment(false, true);
        if (!subComments.isEmpty())
        {
            List<Map<String, Object>> subComments2json = new ArrayList<>();
            for (Comment subComment : subComments)
            {
                subComments2json.add(getComment(content, subComment, level + 1, contextualParameters));
            }
            comment2json.put("sub-comments", subComments2json); 
        }
        
        comment2json.put("canDelete", canDeleteComment(content, comment));
        
        return comment2json;
    }
    
    /**
     * Get JSON representation of a comment's author
     * @param comment the comment
     * @param contextualParameters the contextual parameters
     * @return the comment's author as JSON
     */
    protected Map<String, Object> getCommentAuthor(Comment comment, Map<String, Object> contextualParameters)
    {
        Map<String, Object> author2json = new HashMap<>();
        
        UserIdentity authorIdentity = comment.getAuthor();
        if (authorIdentity != null)
        {
            Map<String, Object> user2json = Optional.ofNullable(_userManager.getUser(authorIdentity))
                                                    .map(author -> getUserProperties(author, contextualParameters))
                                                    .orElseGet(() -> getUserIdentityProperties(authorIdentity, contextualParameters));
            author2json.put("author", user2json);
        }
        else
        {
            String authorName = comment.getAuthorName();
            if (StringUtils.isNotBlank(authorName))
            {
                author2json.put("author-name", authorName);
            }

            String authorEmail = comment.getAuthorEmail();
            if (StringUtils.isNotBlank(authorEmail))
            {
                if (!comment.isEmailHidden())
                {
                    author2json.put("author-email", authorEmail);
                }
                
                getUserByEmail(authorEmail, contextualParameters)
                    .map(author -> getUserProperties(author, contextualParameters))
                    .ifPresent(json -> author2json.put("author", json));
            }

            String authorURL = comment.getAuthorURL();
            if (StringUtils.isNotBlank(authorURL))
            {
                author2json.put("author-url", authorURL);
            }
        }
        
        return author2json;
    }
    
    /**
     * Get the author as a User if exists
     * @param userEmail the email of the user to retrieve
     * @param contextualParameters the contextual parameters
     * @return the user or null
     */
    protected Optional<User> getUserByEmail(String userEmail, Map<String, Object> contextualParameters)
    {
        Request request = ContextHelper.getRequest(_context);
        
        Set<String> userPopulationsOnSite = _populationContextHelper.getUserPopulationsOnContexts(getUserPopulationsContexts(request, contextualParameters), false, false);
        
        try
        {
            return Optional.ofNullable(_userManager.getUserByEmail(userPopulationsOnSite, userEmail));
        }
        catch (NotUniqueUserException e)
        {
            getLogger().error("Cannot find user because 2 or more users match", e);
            return Optional.empty();
        }
    }
    
    /**
     * Get JSON representation of a user from its identity. If the user does not exist, get all known identity properties
     * @param userIdentity the user identity
     * @param contextualParameters the contextual parameters
     * @return the user as JSON
     */
    protected Map<String, Object> getUserPropertiesFromIdentity(UserIdentity userIdentity, Map<String, Object> contextualParameters)
    {
        return Optional.ofNullable(_userManager.getUser(userIdentity))
                       .map(user -> getUserProperties(user, contextualParameters))
                       .orElseGet(() -> getUserIdentityProperties(userIdentity, contextualParameters));
    }
    
    /**
     * Get JSON representation of a user
     * @param user the user
     * @param contextualParameters the contextual parameters
     * @return the user as JSON
     */
    protected Map<String, Object> getUserProperties(User user, Map<String, Object> contextualParameters)
    {
        Map<String, Object> user2json = _userHelper.user2json(user, true);
        user2json.put("imgUrl", ProfileImageResolverHelper.resolve(user.getIdentity().getLogin(), user.getIdentity().getPopulationId(), 64, null));
        return user2json;
    }
    
    /**
     * Get JSON representation of a user identity. Put all known data in the json result
     * @param userIdentity the user identity
     * @param contextualParameters the contextual parameters
     * @return the user identity as JSON
     */
    protected Map<String, Object> getUserIdentityProperties(UserIdentity userIdentity, Map<String, Object> contextualParameters)
    {
        Map<String, Object> user2json = new HashMap<>();
        
        user2json.put("login", userIdentity.getLogin());
        user2json.put("populationId", userIdentity.getPopulationId());

        UserPopulation userPopulation = _userPopulationDAO.getUserPopulation(userIdentity.getPopulationId());
        if (userPopulation != null)
        {
            user2json.put("populationLabel", userPopulation.getLabel());
        }
        
        return user2json;
    }
    
    /**
     * Get the user population contexts
     * @param request The request
     * @param contextualParameters The contextual parameters
     * @return The contexts
     * @throws IllegalArgumentException If there is no context set
     */
    protected List<String> getUserPopulationsContexts(Request request, Map<String, Object> contextualParameters)
    {
        return List.of("/application");
    }
    
    /**
     * SAX a comment
     * @param contentHandler the content handler
     * @param comment the comment
     * @param level the level of comment
     * @throws SAXException if an error occurred hile saxing
     */
    public void saxComment(ContentHandler contentHandler, Comment comment, int level) throws SAXException
    {
        AttributesImpl attrs = new AttributesImpl();

        attrs.addCDATAAttribute("id", comment.getId());
        attrs.addCDATAAttribute("creation-date", DateUtils.zonedDateTimeToString(comment.getCreationDate()));
        attrs.addCDATAAttribute("level", String.valueOf(level));
        attrs.addCDATAAttribute("is-validated", String.valueOf(comment.isValidated()));
        attrs.addCDATAAttribute("is-email-hidden", String.valueOf(comment.isEmailHidden()));
        
        UserIdentity authorIdentity = comment.getAuthor();
        if (authorIdentity == null)
        {
            String authorName = comment.getAuthorName();
            if (StringUtils.isNotBlank(authorName))
            {
                attrs.addCDATAAttribute("author-name", authorName);
            }
            
            String authorEmail = comment.getAuthorEmail();
            if (!comment.isEmailHidden() && StringUtils.isNotBlank(authorEmail))
            {
                attrs.addCDATAAttribute("author-email", authorEmail);
            }
    
            String authorURL = comment.getAuthorURL();
            if (!StringUtils.isBlank(authorURL))
            {
                attrs.addCDATAAttribute("author-url", authorURL);
            }
        }

        XMLUtils.startElement(contentHandler, "comment", attrs);
        
        if (authorIdentity != null)
        {
            User author = _userManager.getUser(authorIdentity);
            if (author != null)
            {
                _userHelper.saxUser(author, contentHandler, "author");
            }
            else
            {
                // If author does not exist anymore, still generate sax event for its identity
                saxUserIdentity(contentHandler, authorIdentity, "author");
            }
        }
        else
        {
            String authorEmail = comment.getAuthorEmail();
            if (StringUtils.isNotBlank(authorEmail))
            {
                Optional<User> author = getUserByEmail(authorEmail, Map.of());
                if (author.isPresent())
                {
                    _userHelper.saxUser(author.get(), contentHandler, "author");
                }
            }
        }

        if (comment.getContent() != null)
        {
            String[] contents = comment.getContent().split("\r?\n");
            for (String c : contents)
            {
                XMLUtils.createElement(contentHandler, "p", c);
            }
        }
        
        // The generated SAXed events for the comments' reaction have changed.
        // In the SAXed events of the ContentGenerator (that should one day use this ContentSaxer):
        // - there is a "nb-like" attributes on the comment node that disappears here
        // - reactions are SAXed as a simple list of ("likers")
        _reactionableHelper.saxReactions(comment, contentHandler);
        
        ReportableObjectHelper.saxReports(comment, contentHandler);
        
        // Additional properties
        saxCommentAdditionalProperties(contentHandler, comment, level);
        
        List<Comment> subComments = comment.getSubComment(false, true);
        if (!subComments.isEmpty())
        {
            XMLUtils.startElement(contentHandler, "sub-comments");
            
            for (Comment subComment : subComments)
            {
                saxComment(contentHandler, subComment, level + 1);
            }
            
            XMLUtils.endElement(contentHandler, "sub-comments");
        }

        XMLUtils.endElement(contentHandler, "comment");
    }
    
    /**
     * Generate SAX events for all known data of the given user identity
     * @param contentHandler the content handler
     * @param userIdentity the user identity
     * @param tagName The XML tag for saxed user
     * @throws SAXException if an error occurred while generating SAX events
     */
    protected void saxUserIdentity(ContentHandler contentHandler, UserIdentity userIdentity, String tagName) throws SAXException
    {
        AttributesImpl attr = new AttributesImpl();
        attr.addCDATAAttribute("login", userIdentity.getLogin());
        attr.addCDATAAttribute("population", userIdentity.getPopulationId());
        
        XMLUtils.startElement(contentHandler, tagName, attr);
        
        UserPopulation userPopulation = _userPopulationDAO.getUserPopulation(userIdentity.getPopulationId());
        if (userPopulation != null)
        {
            userPopulation.getLabel().toSAX(contentHandler, "populationLabel");
        }
        
        XMLUtils.endElement(contentHandler, tagName);
    }
    
    /**
     * SAX additional comment properties
     * @param contentHandler the content handler
     * @param comment the comment
     * @param level the level of comment
     * @throws SAXException if an error occurred while saxing
     */
    protected void saxCommentAdditionalProperties(ContentHandler contentHandler, Comment comment, int level) throws SAXException
    {
        // Nothing
    }
}
