/*
 *  Copyright 2024 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository.comment.contributor;

import java.io.IOException;
import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang.StringUtils;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.repository.comment.Comment;
import org.ametys.core.observation.AsyncObserver;
import org.ametys.core.observation.Event;
import org.ametys.core.ui.mail.StandardMailBodyHelper;
import org.ametys.core.ui.mail.StandardMailBodyHelper.MailBodyBuilder;
import org.ametys.core.ui.mail.StandardMailBodyHelper.MailBodyBuilder.UnauthenticatedUserInput;
import org.ametys.core.ui.mail.StandardMailBodyHelper.MailBodyBuilder.UserInput;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.util.I18nUtils;
import org.ametys.core.util.mail.SendMailHelper;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

import jakarta.mail.MessagingException;

/**
 * Abstract observer to send mails to mentioned users in contributor comments of {@link AmetysObject}
 * @param <T> type of the {@link AmetysObject}
 */
public abstract class AbstractNotifyMentionsObserver<T extends AmetysObject> extends AbstractLogEnabled implements Component, Serviceable, AsyncObserver
{
    /** The ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /** The user manager */
    protected UserManager _userManager;
    /** The current user provider */
    protected CurrentUserProvider _currentUserProvider;
    /** The i18n utils. */
    protected I18nUtils _i18nUtils;
    
    /** Cache for resolved users */
    protected Map<UserIdentity, User> _resolvedUsers = new HashMap<>();
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
        _currentUserProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
    }

    public int getPriority(Event event)
    {
        return MIN_PRIORITY;
    }

    public void observe(Event event, Map<String, Object> transientVars) throws Exception
    {
        Map<String, Object> arguments = event.getArguments();
        
        String ametysObjectId = (String) arguments.get(ObservationConstants.ARGS_AMETYS_OBJECT_ID);
        T ametysObject = _resolver.resolveById(ametysObjectId);
        
        String commentId = (String) arguments.get(ObservationConstants.ARGS_COMMENT_ID);
        Comment comment = ((ContributorCommentableAmetysObject) ametysObject).getContributorComment(commentId);
        String commentContent = comment.getContent();
        

        String language = _getLanguage(ametysObject);
        I18nizableText i18nSubject = _getMailSubject(ametysObject);
        String subject = _i18nUtils.translate(i18nSubject, language);
        
        String ametysObjectTitle = _getAmetysObjectTitle(ametysObject);
        LinkToAmetysObject linkToAmetysObject = _getLinkToAmetysObject(ametysObject);
        
        UserIdentity authorIdentity = comment.getAuthor();
        Optional<User> author = _getUserFromIdentity(authorIdentity);
        ZonedDateTime creationDate = comment.getCreationDate();
        
        Collection<UserIdentity> mentionedUsers = comment.extractMentions();
        
        for (UserIdentity mentionedUserIdentity : mentionedUsers)
        {
            if (!mentionedUserIdentity.equals(authorIdentity))
            {
                // Get mentioned user
                Optional<User> mentionedUser = _getUserFromIdentity(mentionedUserIdentity);
                
                if (mentionedUser.isEmpty())
                {
                    getLogger().warn("Could not send a notification e-mail to user with login {}: there is no user with this login in population {}", mentionedUserIdentity.getLogin(), mentionedUserIdentity.getPopulationId());
                    continue;
                }
                
                if (StringUtils.isBlank(mentionedUser.get().getEmail()))
                {
                    getLogger().warn("Could not send a notification e-mail to user {}: this user has no email address", mentionedUser);
                    continue;
                }
                
                String mentionedUserEmail = mentionedUser.get().getEmail();
                
                try
                {
                    MailBodyBuilder mailBuilder =  StandardMailBodyHelper.newHTMLBody()
                                                                         .withLanguage(language)
                                                                         .withTitle(new I18nizableText("plugin.cms", "PLUGINS_CMS_AMETYS_OBJECT_THREAD_MENTION_MAIL_TITLE"))
                                                                         .addMessage(new I18nizableText("plugin.cms", "PLUGINS_CMS_AMETYS_OBJECT_THREAD_MENTION_MAIL_MESSAGE", List.of(ametysObjectTitle)))
                                                                         .withLink(linkToAmetysObject.linkUrl(), linkToAmetysObject.linkText());
                    
                    // Replace mentions in comment content to highlight mentioned users.
                    // The current mentioned user highlight might be different than other mentioned users, that's why this modifications are made for each mentioned users 
                    String commentContentWithReplacedMentions = _getCommentContentWithReplacedMentions(commentContent, mentionedUserIdentity, mentionedUsers);
                    if (author.isPresent())
                    {
                        UserInput userInput = new UserInput(author.get(), creationDate, commentContentWithReplacedMentions);
                        mailBuilder.withUserInput(userInput, new I18nizableText("plugin.cms", "PLUGINS_CMS_AMETYS_OBJECT_THREAD_MENTION_MAIL_COMMENT_TITLE"));
                    }
                    else
                    {
                        String unknownAuthorName = _i18nUtils.translate(new I18nizableText("plugin.core", "PLUGINS_CORE_USERS_UNKNOWN_USER"));
                        UnauthenticatedUserInput userInput = new UnauthenticatedUserInput(unknownAuthorName, creationDate, commentContentWithReplacedMentions);
                        mailBuilder.withUnauthenticatedUserInput(userInput, new I18nizableText("plugin.cms", "PLUGINS_CMS_AMETYS_OBJECT_THREAD_MENTION_MAIL_COMMENT_TITLE"));
                    }
                    
                    String body = mailBuilder.build();
                    
                    if (StringUtils.isBlank(body))
                    {
                        getLogger().warn("Could not send a notification e-mail to {}: the email body is empty", mentionedUser);
                        continue;
                    }
    
                    SendMailHelper.newMail()
                                  .withSubject(subject)
                                  .withHTMLBody(body)
                                  .withRecipient(mentionedUserEmail)
                                  .sendMail();
                }
                catch (MessagingException | IOException e)
                {
                    getLogger().warn("Could not send a notification e-mail to " + mentionedUser + ": an error occured while sending the email", e);
                }
            }
        }
    }
    
    /**
     * Retrieves the notification mail's subject
     * @param ametysObject the ametys object
     * @return the notification mail's subject
     */
    protected I18nizableText _getMailSubject(T ametysObject)
    {
        List<String> i18nParams = List.of(_getAmetysObjectTitle(ametysObject));
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_AMETYS_OBJECT_THREAD_MENTION_MAIL_SUBJECT", i18nParams);
    }
    
    /**
     * Retrieves the string to use to display the ametys object in the email
     * @param ametysObject the ametysObject
     * @return the ametys object's title
     */
    protected String _getAmetysObjectTitle(T ametysObject)
    {
        return ametysObject.toString();
    }
    
    /**
     * Retrieves the notification mail's body link to ametys object
     * @param ametysObject the ametys object
     * @return the notification mail's body link
     */
    protected abstract LinkToAmetysObject _getLinkToAmetysObject(T ametysObject);
    
    /**
     * Retrieves the given comment content with mentioned replaced by the version to display in notification 
     * @param commentContent the original comment content
     * @param recipient the recipient user identity
     * @param allMentionedUsers all the users mentioned in the comment
     * @return the content with replaced mentioned
     */
    protected String _getCommentContentWithReplacedMentions(String commentContent, UserIdentity recipient, Collection<UserIdentity> allMentionedUsers)
    {
        String commentContentWithReplacedMentions = commentContent;
        
        for (UserIdentity mentionedUserIdentity : allMentionedUsers)
        {
            String mentionedUserName = _getUserFromIdentity(mentionedUserIdentity)
                    .map(User::getFullName)
                    .orElseGet(() -> _i18nUtils.translate(new I18nizableText("plugin.core", "PLUGINS_CORE_USERS_UNKNOWN_USER")));
            
            String mentionBackgroundColor = "rgba(29, 155, 209, 0.1)";
            if (mentionedUserIdentity.equals(recipient))
            {
                mentionBackgroundColor = "rgba(255, 198, 0, 0.18)";                
            }
            
            StringBuilder replaceBy = new StringBuilder("<span style=\"background: ").append(mentionBackgroundColor).append("; color: #1264a3\">")
                    .append(mentionedUserName)
                    .append("</span>");
            
            String toReplace = "@\\(" + mentionedUserIdentity.getLogin() + "#" + mentionedUserIdentity.getPopulationId() + "\\)";
            
            commentContentWithReplacedMentions = commentContentWithReplacedMentions.replaceAll(toReplace, replaceBy.toString());
        }
        
        return commentContentWithReplacedMentions;
    }
    
    /**
     * Retrieves the {@link User} object from the given {@link UserIdentity}
     * @param userIdentity the user identity
     * @return the corresponding {@link User} object, or an empty optional if the user does not exists
     */
    protected Optional<User> _getUserFromIdentity(UserIdentity userIdentity)
    {
        if (_resolvedUsers.containsKey(userIdentity))
        {
            return Optional.ofNullable(_resolvedUsers.get(userIdentity));
        }
        else
        {
            User user = _userManager.getUser(userIdentity);
            _resolvedUsers.put(userIdentity, user);
            
            return Optional.ofNullable(user);
        }
    }

    /**
     * Retrieves the language to use in sent mails
     * @param ametysObject the ametys object
     * @return the language to use in sent mails
     */
    protected abstract String _getLanguage(T ametysObject);
    
    /**
     * Link to the ametys object
     * @param linkUrl the link to the ametys object
     * @param linkText the text of the link to the ametys object
     */
    public record LinkToAmetysObject(String linkUrl, I18nizableText linkText) { /* empty */ }
}
