/*
 *  Copyright 2024 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository.comment.contributor.ui;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.repository.comment.Comment;
import org.ametys.cms.repository.comment.CommentsDAO;
import org.ametys.cms.repository.comment.contributor.ContributorCommentableAmetysObject;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.ModifiableAmetysObject;

/**
 * Abstract client side element to manage thread of contributor comments of an {@link AmetysObject}
 * @param <T> type of the {@link AmetysObject}
 */
public abstract class AbstractAmetysObjectThreadClientSideElement<T extends AmetysObject> extends StaticClientSideElement
{
    /** The ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /** The observation manager */
    protected ObservationManager _observationManager;
    /** The comments DAO */
    protected CommentsDAO _commentsDAO;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        
        _resolver = (AmetysObjectResolver) serviceManager.lookup(AmetysObjectResolver.ROLE);
        _observationManager = (ObservationManager) serviceManager.lookup(ObservationManager.ROLE);
        _commentsDAO = (CommentsDAO) serviceManager.lookup(CommentsDAO.ROLE);
    }
    
    /**
     * Return the thread of comments
     * @param ametysObjectId the ametys object identifier 
     * @return The comments
     */
    public List<Map<String, Object>> getThread(String ametysObjectId)
    {
        return _commentsDAO.getContributorComments(ametysObjectId, new HashMap<>());
    }
    
    /**
     * Add a contributor comment in thread
     * @param ametysObjectId the ametys object identifier
     * @param commentContent the comment content
     * @return a result Map with created comment identifier
     * @throws IllegalStateException if the ametys object is not contributor commentable
     */
    public Map<String, Object> addComment(String ametysObjectId, String commentContent) throws IllegalStateException
    {
        T ametysObject = _resolver.resolveById(ametysObjectId);
        
        if (ametysObject instanceof ContributorCommentableAmetysObject commentableAmetysObject && ametysObject instanceof ModifiableAmetysObject modifiableAmetysObject)
        {
            // Create contributor comment
            Comment comment = commentableAmetysObject.createContributorComment();
            
            UserIdentity currentUser = _currentUserProvider.getUser();
            comment.setAuthor(currentUser);
            
            comment.setContent(commentContent);
            comment.setValidated(true);
            
            modifiableAmetysObject.saveChanges();
            
            // Notify observers
            Map<String, Object> eventParams = _getAddedCommentNotificationEventParams(ametysObject, comment);
            _observationManager.notify(new Event(_getAddedCommentNotificationEventId(), currentUser, eventParams));
            
            // Result
            Map<String, Object> results = new HashMap<>();
            results.put("id", comment.getId());
            return results;
        }
        else
        {
            String errorMessage = "Can not edit contributor comment for non-contributor-commentable ametys object of id '" + ametysObjectId + "'"; 
            getLogger().error(errorMessage);
            throw new IllegalStateException(errorMessage);
        }
    }
    
    /**
     * Retrieve the identifier of the event to send when a comment is added
     * @return the identifier of the event to send when a comment is added
     */
    protected abstract String _getAddedCommentNotificationEventId();
    
    /**
     * Retrieve the parameters of the event to send when a comment is added
     * @param ametysObject the ametys object holding the comments
     * @param comment the added comment 
     * @return the parameters of the event to send when a comment is added
     */
    protected Map<String, Object> _getAddedCommentNotificationEventParams(T ametysObject, Comment comment)
    {
        Map<String, Object> eventParams = new HashMap<>();
        
        eventParams.put(ObservationConstants.ARGS_AMETYS_OBJECT_ID, ametysObject.getId());
        eventParams.put(ObservationConstants.ARGS_COMMENT_ID, comment.getId());
        
        return eventParams;
    }
}
