/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.repository.comment.generators;

import java.io.IOException;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.generation.ServiceableGenerator;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.comment.Comment;
import org.ametys.cms.repository.comment.CommentableContent;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;

/**
 * Generates the comments of the given contents 
 */
public class CommentsGenerator extends ServiceableGenerator
{
    private AmetysObjectResolver _resolver;
    private RightManager _rightManager;
    private CurrentUserProvider _currentUserProvider;
    private ContentHelper _contentHelper;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _rightManager = (RightManager) smanager.lookup(RightManager.ROLE);
        _currentUserProvider = (CurrentUserProvider) smanager.lookup(CurrentUserProvider.ROLE);
        _contentHelper = (ContentHelper) smanager.lookup(ContentHelper.ROLE);
    }
    
    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        @SuppressWarnings("unchecked")
        Map<String, Object> jsParameters = (Map<String, Object>) objectModel.get(ObjectModelHelper.PARENT_CONTEXT);
        @SuppressWarnings("unchecked")
        List<String> contentIds = (List<String>) jsParameters.get("contentIds");
        
        String commentId = (String) jsParameters.get("commentId");
        
        UserIdentity user = _currentUserProvider.getUser();
        
        contentHandler.startDocument();
        XMLUtils.startElement(contentHandler, "comments");
        
        for (String contentId : contentIds)
        {
            Content content = _resolver.resolveById(contentId);

            if (content instanceof CommentableContent)
            {
                if (_rightManager.hasRight(user, "CMS_Rights_CommentModerate", content) == RightResult.RIGHT_ALLOW)
                {
                    CommentableContent commentableContent = (CommentableContent) content;
                    List<Comment> comments = commentableContent.getComments(true, true);
                    
                    for (Comment comment : comments)
                    {
                        if (StringUtils.isNotEmpty(commentId) && !commentId.equals(comment.getId()))
                        {
                            continue;
                        }
                        
                        _saxComment(comment, content);
                    }
                }
            }
        }
        
        XMLUtils.endElement(contentHandler, "comments");
        contentHandler.endDocument();
    }

    /**
     * SAX a comment
     * @param comment The comment
     * @param content The content
     * @throws SAXException if an error occurred while saxing
     */
    protected void _saxComment (Comment comment, Content content) throws SAXException
    {
        AttributesImpl attrs = new AttributesImpl();
        attrs.addCDATAAttribute("id", comment.getId());
        attrs.addCDATAAttribute("validated", Boolean.toString(comment.isValidated()));

        XMLUtils.startElement(contentHandler, "comment", attrs);
        
        _saxContent (content);
        
        XMLUtils.createElement(contentHandler, "creationDate", DateUtils.zonedDateTimeToString(comment.getCreationDate()));
        
        String authorName = comment.getAuthorName();
        if (authorName != null)
        {
            XMLUtils.createElement(contentHandler, "author-name", authorName);
        }
        
        String authorEmail = comment.getAuthorEmail();
        if (authorEmail != null)
        {
            AttributesImpl emailAttrs = new AttributesImpl();
            emailAttrs.addCDATAAttribute("hidden", Boolean.toString(comment.isEmailHidden()));
            XMLUtils.createElement(contentHandler, "author-email", emailAttrs, authorEmail);
        }
        
        String authorUrl = comment.getAuthorURL();
        if (authorUrl != null)
        {
            XMLUtils.createElement(contentHandler, "author-url", authorUrl);
        }
        
        String text = comment.getContent();
        if (text != null)
        {
            XMLUtils.createElement(contentHandler, "text", comment.getContent());
        }
        XMLUtils.endElement(contentHandler, "comment");
    }
    
    /**
     * SAX a content
     * @param content The content
     * @throws SAXException if an error occurred while saxing
     */
    protected void _saxContent (Content content) throws SAXException
    {
        AttributesImpl attrs = new AttributesImpl();
        attrs.addCDATAAttribute("id", content.getId());
        attrs.addCDATAAttribute("title", _contentHelper.getTitle(content));
        attrs.addCDATAAttribute("name", content.getName());
        
        XMLUtils.createElement(contentHandler, "content", attrs);
    }
}
