/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.repository.comment.ui;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.commons.lang.StringUtils;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.comment.Comment;
import org.ametys.cms.repository.comment.CommentableContent;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.UnknownAmetysObjectException;

/**
 * This client site elements creates a button representing the validation state of a content's comment
 */
public class CommentClientSideElement extends StaticClientSideElement
{
    /** The constant for the name of the author from parameters */
    public static final String PARAMETER_AUTHOR_NAME = "author-name";
    /** The constant for the email of the author from parameters */
    public static final String PARAMETER_AUTHOR_EMAIL = "author-email";
    /** The constant for the hidden status of the email of the author from parameters */
    public static final String PARAMETER_AUTHOR_EMAILHIDDEN = "author-emailhidden";
    /** The constant for the url of the author from parameters */
    public static final String PARAMETER_AUTHOR_URL = "author-url";
    /** The constant for the text from parameters */
    public static final String PARAMETER_TEXT = "text";
    /** The constant for the content id from parameters */
    public static final String PARAMETER_CONTENT_ID = "contentId";
    /** The constant for the comment id from parameters */
    public static final String PARAMETER_COMMENT_ID = "commentId";
    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /** The observation manager */
    protected ObservationManager _observationManager;
    /** The content helper */
    private ContentHelper _contentHelper;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _observationManager = (ObservationManager) smanager.lookup(ObservationManager.ROLE);
        _contentHelper = (ContentHelper) smanager.lookup(ContentHelper.ROLE);
    }
    
    /**
     * Return the list of comments with its validation state
     * @param parameters The contents and their comments
     * @return The comments and their validated status
     */
    @Callable
    public Map<String, Object> getComments(Map<String, Object> parameters)
    {
        Map<String, Object> results = new HashMap<>();
        results.put("comments", new ArrayList<>());
        
        for (String contentId : parameters.keySet())
        {
            try
            {
                Content content = _resolver.resolveById(contentId);
                if (content instanceof CommentableContent)
                {
                    CommentableContent cContent = (CommentableContent) content;
                    
                    @SuppressWarnings("unchecked")
                    List<String> commentIds = (List<String>) parameters.get(contentId);
                    
                    for (String commentId : commentIds)
                    {
                        Comment comment = cContent.getComment(commentId);
                        @SuppressWarnings("unchecked")
                        List<Map<String, Object>> comments = (List<Map<String, Object>>) results.get("comments");
                        comments.add(getCommentParameters(cContent, comment));
                    }   
                }
                else
                {
                    getLogger().warn("Ignoring request to get comments status on non commentable content " + contentId);
                }            
            }
            catch (UnknownAmetysObjectException e)
            {
                getLogger().warn("Ignoring request to get comments status on inexisting content " + contentId, e);
            }
        }
        
        return results;
    }
    
    /**
     * Get a comment properties
     * @param contentId the id of the content
     * @param commentId the id of the comment. Can be null
     * @return results the server's response in JSON.
     * @throws IOException If an error occurred
     * @throws ProcessingException If an error occurred
     */
    @Callable
    public Map<String, Object> getComment(String contentId, String commentId) throws IOException, ProcessingException
    {
        Map<String, Object> results = new HashMap<> ();
               
        UserIdentity user = _currentUserProvider.getUser();
        
        Map<String, Object> comments = new HashMap<> ();
        
        Content content = _resolver.resolveById(contentId);
        if (StringUtils.isNotEmpty(commentId))
        {
            if (content instanceof CommentableContent)
            {
                if (_rightManager.hasRight(user, "CMS_Rights_CommentModerate", content) == RightResult.RIGHT_ALLOW)
                {
                    CommentableContent commentableContent = (CommentableContent) content;
                    Comment comment = commentableContent.getComment(commentId);
                    comments = _jsonifyComment(comment, content);
                }
            }
        }
        
        results.put("comments", comments);
        return results;
    }

    /**
     * Edit a comment if connected user has sufficient rights
     * @param parameters the JS parameters. Necessarily contains the content and comment id, and the values to edit
     * @return An empty map
     */
    @Callable 
    public Map<String, Object> editComment(Map<String, Object> parameters)
    {
        String contentId = (String) parameters.get(PARAMETER_CONTENT_ID);
        String commentId = (String) parameters.get(PARAMETER_COMMENT_ID);
        
        try
        {
            Content content = _resolver.resolveById(contentId);

            if (!hasRight("CMS_Rights_CommentModerate", content))
            {
                String errorMessage = "User " + getCurrentUser() + " try to edit a comment on content of id '" + contentId + "' with no sufficient rights"; 
                getLogger().error(errorMessage);
                throw new IllegalStateException(errorMessage);
            }
            
            if (!(content instanceof CommentableContent))
            {
                String errorMessage = "Can not edit comment for non-commentable content of id '" + contentId + "'"; 
                getLogger().error(errorMessage);
                throw new IllegalStateException(errorMessage);
            }
            
            CommentableContent cContent = (CommentableContent) content;
            
            Comment comment = cContent.getComment(commentId);
            
            String oldAuthorName = comment.getAuthorName();
            String oldAuthorEmail = comment.getAuthorEmail();
            boolean oldAuthorEmailHidden = comment.isEmailHidden();
            String oldAuthorURL = comment.getAuthorURL();
            String oldContent = comment.getContent();
            
            boolean needSave = false;
            
            String authorName = (String) parameters.get(PARAMETER_AUTHOR_NAME);
            if (!authorName.equals(oldAuthorName))
            {
                comment.setAuthorName(authorName);
                needSave = true;
            }
            
            String authorEmail = (String) parameters.get(PARAMETER_AUTHOR_EMAIL);
            if (!authorEmail.equals(oldAuthorEmail))
            {
                comment.setAuthorEmail(authorEmail);
                needSave = true;
            }
            
            boolean authorEmailHidden = (Boolean) parameters.get(PARAMETER_AUTHOR_EMAILHIDDEN);
            if (authorEmailHidden != oldAuthorEmailHidden)
            {
                comment.setEmailHiddenStatus(authorEmailHidden);
                needSave = true;
            }
            
            String authorUrl = (String) parameters.get(PARAMETER_AUTHOR_URL);
            if (!authorUrl.equals(oldAuthorURL))
            {
                comment.setAuthorURL(authorUrl);
                needSave = true;
            }
            
            String text = (String) parameters.get(PARAMETER_TEXT);
            if (!text.equals(oldContent))
            {
                comment.setContent(text);
                needSave = true;
            }
            
            if (needSave)
            {
                cContent.saveChanges();
                
                Map<String, Object> eventParams = new HashMap<>();
                eventParams.put(ObservationConstants.ARGS_CONTENT, content);
                eventParams.put(ObservationConstants.ARGS_COMMENT, comment);
                eventParams.put("content.comment.old.author", oldAuthorName);
                eventParams.put("content.comment.old.author.email", oldAuthorEmail);
                eventParams.put("content.comment.old.author.email.hidden", oldAuthorEmailHidden);
                eventParams.put("content.comment.old.author.url", oldAuthorURL);
                eventParams.put("content.comment.old.content", oldContent);
                
                _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_COMMENT_MODIFYING, getCurrentUser(), eventParams));
            }
        
            return getCommentParameters(content, comment);
        }
        catch (UnknownAmetysObjectException e)
        {
            getLogger().error("Unknown content of id '" + contentId + "'", e);
            throw new IllegalArgumentException("Unknown content of id '" + contentId + "'", e);
        }
    }
    
    /**
     * Removes comments
     * @param contents the contents with comments to remove
     * @return the JSON result with deleted comments or not
     */
    @SuppressWarnings("unchecked")
    @Callable
    public Map<String, Object> deleteComments(Map<String, List<String>> contents)
    {
        Map<String, Object> results = new HashMap<> ();
        
        results.put("deleted-comments", new ArrayList<>());
        results.put("undeleted-comments", new ArrayList<>());
        results.put("uncommentable-contents", new ArrayList<>());
        results.put("noright-contents", new ArrayList<>());
        results.put("unknown-contents", new ArrayList<>());
        
        for (String contentId : contents.keySet())
        {
            try
            {
                Content content = _resolver.resolveById(contentId);
                
                if (hasRight("CMS_Rights_CommentModerate", content))
                {
                    List<String> deleteCommentIds = new ArrayList<>();
                    // For each associated comment
                    for (String commentId : contents.get(contentId))
                    {
                        if (!_isParentCommentAlreadyDelete(deleteCommentIds, commentId))
                        {
                            Map<String, Object> commentParams = new HashMap<>();
                            commentParams.put("id", commentId);
                            commentParams.put("contentId", contentId);
                            commentParams.put("contentTitle", _contentHelper.getTitle(content));
                            
                            if (content instanceof CommentableContent)
                            {
                                CommentableContent cContent = (CommentableContent) content;
                                Comment comment = cContent.getComment(commentId);
                                
                                Map<String, Object> eventParams = new HashMap<>();
                                eventParams.put(ObservationConstants.ARGS_CONTENT, content);
                                eventParams.put(ObservationConstants.ARGS_COMMENT_ID, comment.getId());
                                eventParams.put(ObservationConstants.ARGS_COMMENT_AUTHOR, comment.getAuthorName());
                                eventParams.put(ObservationConstants.ARGS_COMMENT_AUTHOR_EMAIL, comment.getAuthorEmail());
                                eventParams.put(ObservationConstants.ARGS_COMMENT_VALIDATED, comment.isValidated());
                                eventParams.put(ObservationConstants.ARGS_COMMENT_CREATION_DATE, comment.getCreationDate());
                                eventParams.put("comment.content", comment.getContent());
                                
                                _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_COMMENT_DELETING, getCurrentUser(), eventParams));
                                
                                comment.remove();
                                cContent.saveChanges();
        
                                List<Map<String, Object>> deletedComments = (List<Map<String, Object>>) results.get("deleted-comments");
                                deletedComments.add(commentParams);
                                deleteCommentIds.add(commentId);
                                
                                _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_COMMENT_DELETED, getCurrentUser(), eventParams));
                            }
                            else
                            {
                                getLogger().error("Can not remove a comment on a non commentable content");
                                
                                List<Map<String, Object>> uncommentableContents = (List<Map<String, Object>>) results.get("uncommentable-contents");
                                uncommentableContents.add(commentParams);
                            }
                        }
                    }
                }
                else
                {
                    // No right
                    getLogger().error("User '" + getCurrentUser() + "' does not have right to moderate comments on content '" + content.getId() + "'");
                    
                    Map<String, Object> contentParams = new HashMap<>();
                    contentParams.put("contentId", contentId);
                    contentParams.put("contentTitle", _contentHelper.getTitle(content));
                    
                    List<Map<String, Object>> norightContents = (List<Map<String, Object>>) results.get("noright-contents");
                    norightContents.add(contentParams);
                }
            }
            catch (UnknownAmetysObjectException e)
            {
                getLogger().error("Can not remove a comment on a non existing content", e);
                
                Map<String, Object> contentParams = new HashMap<>();
                contentParams.put("contentId", contentId);
                
                List<Map<String, Object>> unknownContents = (List<Map<String, Object>>) results.get("unknown-contents");
                unknownContents.add(contentParams);
            }
        }

        return results;
    }
    
    /**
     * True if the a parent comment of the comment id is already deleted
     * @param deleteCommentIds the delete comment ids
     * @param commentId the comment id
     * @return true if the a parent comment of the comment id is already deleted
     */
    protected boolean _isParentCommentAlreadyDelete(List<String> deleteCommentIds, String commentId)
    {
        for (String deleteCommentId : deleteCommentIds)
        {
            if (commentId.startsWith(deleteCommentId))
            {
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * Validates comments when it is possible.
     * @param contents the contents with comments to validate
     * @return the JSON result with validated comments or not
     */
    @SuppressWarnings("unchecked")
    @Callable
    public Map<String, Object> validateComments(Map<String, List<String>> contents)
    {
        Map<String, Object> results = new HashMap<>();

        results.put("validated-comments", new ArrayList<>());
        results.put("error-comments", new ArrayList<>());
        results.put("uncommentable-contents", new ArrayList<>());
        results.put("noright-contents", new ArrayList<>());
        results.put("unknown-contents", new ArrayList<>());
        
        for (String contentId : contents.keySet())
        {
            try
            {
                Content content = _resolver.resolveById(contentId);
                
                if (hasRight("CMS_Rights_CommentModerate", content))
                {
                    // For each associated comment
                    for (String commentId : contents.get(contentId))
                    {
                        Map<String, Object> commentParams = new HashMap<>();
                        commentParams.put("id", commentId);
                        commentParams.put("contentId", contentId);
                        commentParams.put("contentTitle", _contentHelper.getTitle(content));
                        
                        if (content instanceof CommentableContent)
                        {
                            CommentableContent cContent = (CommentableContent) content;
        
                            Comment comment = cContent.getComment(commentId);
                            commentParams.put("reportsCount", comment.getReportsCount());
                            
                            if (!comment.isValidated())
                            {
                                comment.setValidated(true);
                                cContent.saveChanges();
                                
                                List<Map<String, Object>> validatedComments = (List<Map<String, Object>>) results.get("validated-comments");
                                validatedComments.add(commentParams);
                                
                                Map<String, Object> eventParams = new HashMap<>();
                                eventParams.put(ObservationConstants.ARGS_CONTENT, content);
                                eventParams.put(ObservationConstants.ARGS_COMMENT, comment);
                                _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_COMMENT_VALIDATED, getCurrentUser(), eventParams));
                            }
                        }
                        else
                        {
                            getLogger().error("Can not validate a comment on a non commentable content");
                            
                            List<Map<String, Object>> errorComments = (List<Map<String, Object>>) results.get("error-comments");
                            errorComments.add(commentParams);
                        }
                    }
                }
                else
                {
                    // No right
                    getLogger().error("User '" + getCurrentUser() + "' does not have right to validate comments on content '" + content.getId() + "'");
                    
                    Map<String, Object> contentParams = new HashMap<>();
                    contentParams.put("contentId", contentId);
                    contentParams.put("contentTitle", _contentHelper.getTitle(content));
                }
            }
            catch (UnknownAmetysObjectException e)
            {
                getLogger().error("Can not validate a comment on a non existing content", e);
                
                Map<String, Object> contentParams = new HashMap<>();
                contentParams.put("contentId", contentId);
                
                List<Map<String, Object>> unknownContents = (List<Map<String, Object>>) results.get("unknown-contents");
                unknownContents.add(contentParams);
            }
        }

        return results;
    }
    
    /**
     * Invalidates comments when it is possible.
     * @param contents the contents with comments to invalidate
     * @return the JSON result with invalidated comments or not
     */
    @SuppressWarnings("unchecked")
    @Callable
    public Map<String, Object> invalidateComments(Map<String, List<String>> contents)
    {
        Map<String, Object> results = new HashMap<>();

        results.put("unvalidated-comments", new ArrayList<>());
        results.put("error-comments", new ArrayList<>());
        results.put("uncommentable-contents", new ArrayList<>());
        results.put("noright-contents", new ArrayList<>());
        results.put("unknown-contents", new ArrayList<>());
        
        for (String contentId : contents.keySet())
        {
            try
            {
                Content content = _resolver.resolveById(contentId);
                
                if (hasRight("CMS_Rights_CommentModerate", content))
                {
                    // For each associated comment
                    for (String commentId : contents.get(contentId))
                    {
                        Map<String, Object> commentParams = new HashMap<>();
                        commentParams.put("id", commentId);
                        commentParams.put("contentId", contentId);
                        commentParams.put("contentTitle", _contentHelper.getTitle(content));
                        
                        if (content instanceof CommentableContent)
                        {
                            CommentableContent cContent = (CommentableContent) content;
        
                            Comment comment = cContent.getComment(commentId);
                            commentParams.put("reportsCount", comment.getReportsCount());
                            if (comment.isValidated())
                            {
                                comment.setValidated(false);
                                cContent.saveChanges();

                                List<Map<String, Object>> validatedComments = (List<Map<String, Object>>) results.get("unvalidated-comments");
                                validatedComments.add(commentParams);

                                Map<String, Object> eventParams = new HashMap<>();
                                eventParams.put(ObservationConstants.ARGS_CONTENT, content);
                                eventParams.put(ObservationConstants.ARGS_COMMENT, comment);
                                _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_COMMENT_UNVALIDATED, getCurrentUser(), eventParams));
                            }
                        }
                        else
                        {
                            getLogger().error("Can not validate a comment on a non commentable content");
                            
                            List<Map<String, Object>> errorComments = (List<Map<String, Object>>) results.get("error-comments");
                            errorComments.add(commentParams);
                        }
                    }
                }
                else
                {
                    // No right
                    getLogger().error("User '" + getCurrentUser() + "' does not have right to validate comments on content '" + content.getId() + "'");
                    
                    Map<String, Object> contentParams = new HashMap<>();
                    contentParams.put("contentId", contentId);
                    contentParams.put("contentTitle", _contentHelper.getTitle(content));
                }
            }
            catch (UnknownAmetysObjectException e)
            {
                getLogger().error("Can not validate a comment on a non existing content", e);
                
                Map<String, Object> contentParams = new HashMap<>();
                contentParams.put("contentId", contentId);
                
                List<Map<String, Object>> unknownContents = (List<Map<String, Object>>) results.get("unknown-contents");
                unknownContents.add(contentParams);
            }
        }

        return results;
    }
    
    /**
     * Jsonify the comment.
     * @param comment The comment
     * @param content The content
     * @return commentMap the comment map
     */
    protected Map<String, Object> _jsonifyComment (Comment comment, Content content) 
    {
        Map<String, Object> result = new HashMap<> ();
        
        Map<String, Object> commentMap = new HashMap<> ();
            
        
        commentMap.put("validated", Boolean.toString(comment.isValidated()));
        commentMap.put("id", comment.getId());
        commentMap.put("creationDate", DateUtils.zonedDateTimeToString(comment.getCreationDate()));
        
        String authorName = comment.getAuthorName();
        if (authorName != null)
        {
            commentMap.put("author-name", authorName);
        }
        
        String authorEmail = comment.getAuthorEmail();
        if (authorEmail != null)
        {
            Map<String, Object> authorEmailMap = new HashMap<> ();
            authorEmailMap.put("hidden", Boolean.toString(comment.isEmailHidden()));
            authorEmailMap.put("value", authorEmail);
            commentMap.put("author-email", authorEmailMap);
        }
        
        String authorUrl = comment.getAuthorURL();
        if (authorUrl != null)
        {
            commentMap.put("author-url", authorUrl);
        }
        
        String text = comment.getContent();
        if (text != null)
        {
            commentMap.put("text", comment.getContent());
        }
        commentMap.put("content", _jsonifyContent(content));
        
        result.put("comment", commentMap);
        
        return result;
    }
    
    /**
     * Jsonify the content.
     * @param content The content
     * @return contentMap the content map
     */
    protected Map<String, Object> _jsonifyContent (Content content) 
    {
        Map<String, Object> contentMap = new HashMap<> ();
        
        contentMap.put("id", content.getId());
        contentMap.put("title", _contentHelper.getTitle(content));
        contentMap.put("name", content.getName());
        
        return contentMap;
    }

    /**
     * Get the parameters for a comment
     * @param content The content
     * @param comment The comment
     * @return The parameters
     */
    protected Map<String, Object> getCommentParameters (Content content, Comment comment)
    {
        Map<String, Object> params = new HashMap<>();
        
        params.put("contentId", content.getId());
        params.put("contentTitle", _contentHelper.getTitle(content));
        params.put("id", comment.getId());
        params.put("validated", comment.isValidated());
        params.put("reportsCount", comment.getReportsCount());
        
        return params;
    }
    
    /**
     * Get the current user
     * @return The current user
     */
    protected UserIdentity getCurrentUser ()
    {
        return _currentUserProvider.getUser();
    }

    /**
     * Determines if connected user has right on content
     * @param rightId The right id
     * @param content The content
     * @return true if user has right
     */
    protected boolean hasRight (String rightId, Content content)
    {
        UserIdentity user = _currentUserProvider.getUser();
        
        return _rightManager.hasRight(user, "CMS_Rights_CommentModerate", content) == RightResult.RIGHT_ALLOW;
    }
}
