/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository.comment.ui;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jcr.Node;
import javax.jcr.NodeIterator;
import javax.jcr.Repository;
import javax.jcr.RepositoryException;
import javax.jcr.Session;
import javax.jcr.query.Query;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.commons.lang.StringUtils;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ReportableObject;
import org.ametys.cms.repository.ReportableObjectHelper;
import org.ametys.cms.repository.comment.Comment;
import org.ametys.cms.repository.comment.CommentableContent;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.ui.Callable;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.provider.AbstractRepository;

/**
 * Action for getting the list of comments.
 */
public class CommentsAndReportsTreeComponent implements Component, Serviceable, Contextualizable
{
    /** The ametys object resolver */
    protected AmetysObjectResolver _resolver;
   
    /** The right manager */
    protected RightManager _rightManager;
    
    /** The current user provider */
    protected CurrentUserProvider _currentUserProvider;
    
    /** The content helper */
    protected ContentHelper _contentHelper;
    
    /** The user manager */
    protected UserManager _userManager;
    
    /** The context */
    protected Context _context;
    
    /** The JCR repository */
    protected Repository _repository;

    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _rightManager = (RightManager) smanager.lookup(RightManager.ROLE);
        _currentUserProvider = (CurrentUserProvider) smanager.lookup(CurrentUserProvider.ROLE);
        _contentHelper = (ContentHelper) smanager.lookup(ContentHelper.ROLE);
        _userManager = (UserManager) smanager.lookup(UserManager.ROLE);
        _repository = (Repository) smanager.lookup(AbstractRepository.ROLE);
    }
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    /**
     * Retrieves the comments and reports of the given selected contents
     * @param contentIds identifiers of the contents
     * @return the comments and reports of the contents
     */
    @Callable
    public Map getCommentsAndReportsFromSelectedContents(List<String> contentIds)
    {
        // Get current user to check rights
        UserIdentity user = _currentUserProvider.getUser();
        
        // Get comments and reports for each content
        List<Map<String, Object>> contentComments = new ArrayList<>();
        for (String contentId : contentIds)
        {
            Content content = _resolver.resolveById(contentId);
            contentComments.add(_content2json(content, user));
        }
        
        return Map.of("comments", contentComments);
    }
    
    /**
     * Retrieves the comments and reports of the contents of the given type
     * Manages only contents that have at least one report (on itself or on a comment)
     * @param contentTypeId the content type identifier
     * @return the comments and reports of the contents
     * @throws RepositoryException if an error occurs while retrieving contents from the repository
     */
    @Callable
    public Map getCommentsAndReportsFromContentTypeId(String contentTypeId) throws RepositoryException
    {
        // Get current user to check rights
        UserIdentity user = _currentUserProvider.getUser();
        
        // Get contents from the given content type
        Iterable<Content> contents = _getReportedContentsFromContentType(contentTypeId);
     
        // Get comments and reports for each content
        List<Map<String, Object>> contentComments = new ArrayList<>();
        for (Content content : contents)
        {
            contentComments.add(_content2json(content, user, false));
        }
        
        return Map.of("comments", contentComments);
    }
    
    /**
     * Retrieves the contents from the current site with the given content type that have at least a report or a reported comment 
     * @param contentTypeId the content type identifier
     * @return the contents
     * @throws RepositoryException if an error occurs while retrieving the contents from the repository
     */
    protected Iterable<Content> _getReportedContentsFromContentType(String contentTypeId) throws RepositoryException
    {
        Request request = ContextHelper.getRequest(_context);
        String currentSiteName = (String) request.getAttribute("siteName");
        
        String xPathQuery = "//element(*, ametys:content)[@ametys:site = '" + currentSiteName + "' and @ametys-internal:contentType = '" + contentTypeId + "']"
                + "//*[@ametys:" + ReportableObjectHelper.REPORTS_COUNT_ATTRIBUTE_NAME + " > 0]";
        
        Session session = _repository.login();
        @SuppressWarnings("deprecation")
        Query query = session.getWorkspace().getQueryManager().createQuery(xPathQuery, Query.XPATH);
        NodeIterator reportedNodesIterator = query.execute().getNodes();
        
        Set<Content> contents = new HashSet<>();
        
        while (reportedNodesIterator.hasNext())
        {
            Node node = reportedNodesIterator.nextNode();
            while (!node.isNodeType("ametys:content"))
            {
                node = node.getParent();
            }
            contents.add(_resolver.resolve(node, false));
        }
        
        return contents;
    }
    
    /**
     * Get the JSON representation of a content
     * @param content The content
     * @param user the current user
     * @return The content as JSON object
     */
    protected Map<String, Object> _content2json (Content content, UserIdentity user)
    {
        return _content2json(content, user, true);
    }
    
    /**
     * Get the JSON representation of a content
     * @param content The content
     * @param user the current user
     * @param includeNoReportedComments <code>true</code> to include the comment that have no report, <code>false</code> otherwise
     * @return The content as JSON object
     */
    protected Map<String, Object> _content2json (Content content, UserIdentity user, boolean includeNoReportedComments)
    {
        Map<String, Object> object = new HashMap<>();
        
        object.put("contentId", content.getId());
        object.put("commentId", "");
        object.put("contentTitle", StringUtils.defaultString(content.getTitle()));
        object.put("text", StringUtils.defaultString(content.getTitle()));
        object.put("creationDate", DateUtils.zonedDateTimeToString(content.getLastModified()));
        
        UserIdentity lastContributorIdentity = content.getLastContributor();
        User lastContributor = _userManager.getUser(lastContributorIdentity);
        if (lastContributor != null)
        {
            object.put("authorName", lastContributor.getFullName());
            object.put("authorEmail", lastContributor.getEmail());
        }
        
        if (content instanceof ReportableObject)
        {
            object.put("reportsCount", ((ReportableObject) content).getReportsCount());
        }
        
        List<Map<String, Object>> commentsList = new ArrayList<>();
        if (content instanceof CommentableContent)
        {
            if (_rightManager.hasRight(user, "CMS_Rights_CommentModerate", content) == RightResult.RIGHT_ALLOW)
            {
                CommentableContent commentableContent = (CommentableContent) content;
                List<Comment> comments = commentableContent.getComments(true, true);
                
                for (Comment comment : comments)
                {
                    Map<String, Object> commentAsJSON = _comment2json(comment, content, includeNoReportedComments);
                    if (!commentAsJSON.isEmpty())
                    {
                        commentsList.add(commentAsJSON);
                    }
                }
            }
        }
        
        object.put("comments", commentsList);
        
        return object;
    }
    
    /**
     * Get the JSON representation of a comment
     * @param comment the comment
     * @param content the content
     * @param includeNoReportedComments <code>true</code> to include the comment that have no report, <code>false</code> otherwise
     * @return The comment as JSON object
     */
    protected Map<String, Object> _comment2json (Comment comment, Content content, boolean includeNoReportedComments)
    {
        List<Map<String, Object>> commentsList = new ArrayList<>();
        for (Comment subcomment : comment.getSubComment(true, true))
        {
            Map<String, Object> subcommentAsJSON = _comment2json(subcomment, content, includeNoReportedComments);
            if (!subcommentAsJSON.isEmpty())
            {
                commentsList.add(_comment2json(subcomment, content, includeNoReportedComments));
            }
        }
        
        if (includeNoReportedComments || comment.getReportsCount() > 0 || !commentsList.isEmpty())
        {
            Map<String, Object> object = new HashMap<>();
            object.put("commentId", comment.getId());
            object.put("contentId", content.getId());
            object.put("contentTitle", StringUtils.defaultString(content.getTitle()));
            object.put("text", comment.getContent());
            object.put("authorName", comment.getAuthorName());
            object.put("authorEmail", comment.getAuthorEmail());
            object.put("authorUrl", comment.getAuthorURL());
            object.put("reportsCount", comment.getReportsCount());
            object.put("creationDate", DateUtils.zonedDateTimeToString(comment.getCreationDate()));
            object.put("validated", Boolean.toString(comment.isValidated()));
            object.put("authorIsEmailHidden", Boolean.toString(comment.isEmailHidden()));
            object.put("comments", commentsList);
            return object;
        }
        
        return Map.of();
    }
}
