/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.repository.reports.ui;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.repository.ReportableObject;
import org.ametys.cms.repository.comment.Comment;
import org.ametys.cms.repository.comment.CommentableContent;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.UnknownAmetysObjectException;

/**
 * This client site elements creates a button representing the validation state of a content's comment
 */
public class ReportsClientSideElement extends StaticClientSideElement
{
    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /** The current user provider */
    protected CurrentUserProvider _userProvider;
    /** The observation manager */
    protected ObservationManager _observationManager;
    /** The content helper */
    private ContentHelper _contentHelper;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _userProvider = (CurrentUserProvider) smanager.lookup(CurrentUserProvider.ROLE);
        _observationManager = (ObservationManager) smanager.lookup(ObservationManager.ROLE);
        _contentHelper = (ContentHelper) smanager.lookup(ContentHelper.ROLE);
    }
    
    /**
     * Ignore reports on objects (contents or comments) when it is possible.
     * @param reportedObjects the objects where to ignore reports
     * @return the JSON result with contents and comments with ignored reports or not
     */
    @Callable
    public Map<String, Object> ignoreReports(List<Map<String, String>> reportedObjects)
    {
        List<Map<String, Object>> contentsWithIgnoredReports = new ArrayList<>();
        List<Map<String, Object>> commentsWithIgnoredReports = new ArrayList<>();
        List<Map<String, Object>> unknownContents = new ArrayList<>();
        List<Map<String, Object>> unknownComments = new ArrayList<>();
        List<Map<String, Object>> uncommentableContents = new ArrayList<>();
        List<Map<String, Object>> unreportableContents = new ArrayList<>();
        List<Map<String, Object>> noRightContents = new ArrayList<>();
        
        Map<String, Content> resolvedContentsById = new HashMap<>();
        Set<ModifiableContent> modifiedContents = new HashSet<>();
        
        UserIdentity currentUser = _userProvider.getUser();
        
        for (Map<String, String> reportedObject : reportedObjects)
        {
            String contentId = reportedObject.get("contentId");
            
            try
            {
                Content content = resolvedContentsById.get(contentId);
                if (content == null)
                {
                    content = _resolver.resolveById(contentId);
                    resolvedContentsById.put(contentId, content);
                }
                
                Map<String, Object> contentParams = _getContentParameters(content);
                
                if (_rightManager.hasRight(currentUser, "CMS_Rights_CommentModerate", content) == RightResult.RIGHT_ALLOW)
                {
                    if (reportedObject.containsKey("commentId"))
                    {
                        String commentId = reportedObject.get("commentId");
                        Map<String, Object> commentParams = _getCommentParameters(commentId, content);
                        
                        if (content instanceof CommentableContent)
                        {
                            CommentableContent cContent = (CommentableContent) content;
                            boolean ignored = _ignoreReportsOnComment(commentId, cContent, currentUser);
                            if (ignored)
                            {
                                modifiedContents.add(cContent);
                                commentsWithIgnoredReports.add(commentParams);
                            }
                        }
                        else
                        {
                            getLogger().error("Can not ignore reports on a comment on a non commentable content");
                            uncommentableContents.add(contentParams);
                        }
                    }
                    else
                    {
                        if (content instanceof ReportableObject && content instanceof ModifiableContent)
                        {
                            boolean ignored = _ignoreReportsOnContent((ReportableObject) content, currentUser);
                            if (ignored)
                            {
                                modifiedContents.add((ModifiableContent) content);
                                contentsWithIgnoredReports.add(contentParams);
                            }
                        }
                        else
                        {
                            getLogger().error("Can not ignore reports on a non reportable content");
                            unreportableContents.add(contentParams);
                        }
                    }
                }
                else
                {
                    getLogger().error("User '" + currentUser + "' does not have right to ignore reports on content '" + content.getId() + "'");
                    noRightContents.add(contentParams);
                }
            }
            catch (UnknownAmetysObjectException e)
            {
                getLogger().error("Can not ignore reports on a comment on a non existing content", e);
                unknownContents.add(Map.of("id", contentId));
            }
        }

        for (ModifiableContent modifiedContent : modifiedContents)
        {
            modifiedContent.saveChanges();
        }
        
        return Map.of("contents-with-ignored-reports", contentsWithIgnoredReports,
            "comments-with-ignored-reports", commentsWithIgnoredReports,
            "unknown-contents", unknownContents,
            "unknown-comments", unknownComments,
            "uncommentable-contents", uncommentableContents,
            "unreportable-contents", unreportableContents,
            "noright-contents", noRightContents);
    }

    /**
     * Retrieves the given content's parameters
     * @param content the content
     * @return the content's parameters
     */
    protected Map<String, Object> _getContentParameters(Content content)
    {
        Map<String, Object> contentParams = new HashMap<>();
        contentParams.put("id", content.getId());
        contentParams.put("title", _contentHelper.getTitle(content));
        return contentParams;
    }

    /**
     * Ignores the reports on the given content
     * @param content the content
     * @param currentUser the current user
     * @return <code>true</code> if there are reports that have been ignored, <code>false</code> otherwise
     */
    protected boolean _ignoreReportsOnContent(ReportableObject content, UserIdentity currentUser)
    {
        if (content.getReportsCount() > 0)
        {
            content.clearReports();
            
            Map<String, Object> eventParams = new HashMap<>();
            eventParams.put(ObservationConstants.ARGS_CONTENT, content);
            _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_IGNORE_REPORTS, currentUser, eventParams));
            
            return true;
        }
        else
        {
            return false;
        }
    }

    /**
     * Retrieves the given comment's parameters
     * @param commentId the comment identifier
     * @param content the content of the comment
     * @return the comment's parameters
     */
    protected Map<String, Object> _getCommentParameters(String commentId, Content content)
    {
        Map<String, Object> commentParams = new HashMap<>();
        commentParams.put("id", commentId);
        commentParams.put("contentId", content.getId());
        commentParams.put("contentTitle", _contentHelper.getTitle(content));
        return commentParams;
    }

    /**
     * Ignores the reports on the given comment
     * @param commentId the identifier of the comment
     * @param content the content of the comment
     * @param currentUser the current user
     * @return <code>true</code> if there are reports that have been ignored, <code>false</code> otherwise
     */
    protected boolean _ignoreReportsOnComment(String commentId, CommentableContent content, UserIdentity currentUser)
    {
        Comment comment = content.getComment(commentId);
        
        if (comment.getReportsCount() > 0)
        {
            comment.clearReports();
            
            Map<String, Object> eventParams = new HashMap<>();
            eventParams.put(ObservationConstants.ARGS_CONTENT, content);
            eventParams.put(ObservationConstants.ARGS_COMMENT, comment);
            _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_COMMENT_IGNORE_REPORTS, currentUser, eventParams));
            
            return true;
        }
        else
        {
            return false;
        }
    }
}
