/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.rights;

import java.io.UnsupportedEncodingException;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ResourceNotFoundException;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.commons.lang.StringUtils;

import org.ametys.cms.repository.Content;
import org.ametys.core.right.RightManager;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.FilenameUtils;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.runtime.authentication.AccessDeniedException;
import org.ametys.runtime.authentication.AuthorizationRequiredException;

/**
 * Test if the accessed object is not restricted, i.e. an anonymous user has READ access
 * or if the connected user has READ access on it.
 * <ul>
 * <li>If the object is not restricted, returns EMPTY_MAP.</li>
 * <li>If the object is restricted and the current user is allowed, return null.</li>
 * <li>If the object is restricted but the current user is not allowed, an {@link AccessDeniedException} is thrown.</li>
 * <li>If the object is restricted but no one is logged in, an {@link AuthorizationRequiredException} is thrown.</li>
 * </ul>
 */
public class CheckReadAccessAction extends ServiceableAction
{
    /** The Ametys object resolver. */
    protected AmetysObjectResolver _resolver;
    
    /** The current user provider */
    protected CurrentUserProvider _currentUserProvider;

    /** The right manager */
    protected RightManager _rightManager;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _rightManager = (RightManager) manager.lookup(RightManager.ROLE);
        _resolver = (AmetysObjectResolver) serviceManager.lookup(AmetysObjectResolver.ROLE);
        _currentUserProvider = (CurrentUserProvider) serviceManager.lookup(CurrentUserProvider.ROLE);
    }
    
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
            
        UserIdentity user = _currentUserProvider.getUser();
        
        // Get the Ametys object from either the parameters or the request attributes.
        AmetysObject ao = getAmetysObject(parameters, request);
        if (ao == null)
        {
            throw new ResourceNotFoundException("CheckReadAccessAction: a valid ametys object must be provided.");
        }
        
        if (_rightManager.hasAnonymousReadAccess(ao))
        {
            // Anonymous access allowed
            return EMPTY_MAP;
        }
        else if (user == null)
        {
            // User not yet authenticated
            throw new AuthorizationRequiredException();
        }
        else if (_rightManager.hasReadAccess(user, ao))
        {
            // User has read access
            return null;
        }
        
        // User is not authorized
        throw new AccessDeniedException("Access to object " + ao.getId() + " is not allowed for user " + user);
    }
    
    /**
     * Get the Ametys object from either the parameters or the request attributes.
     * @param parameters the action parameters.
     * @param request the request.
     * @return the Ametys object or null if not found.
     * @throws UnsupportedEncodingException if failed to decode object path
     */
    protected AmetysObject getAmetysObject(Parameters parameters, Request request) throws UnsupportedEncodingException
    {
        try
        {
            String objectId = parameters.getParameter("objectId", "");
            String objectPath = parameters.getParameter("objectPath", "");
            if (StringUtils.isNotEmpty(objectId))
            {
                return _resolver.resolveById(objectId);
            }
            else if (StringUtils.isNotEmpty(objectPath))
            {
                String decodedPath = FilenameUtils.decode(objectPath);
                return _resolver.resolveByPath(decodedPath);
            }
            else
            {
                Content content = (Content) request.getAttribute(Content.class.getName());
                return content != null ? content : (AmetysObject) request.getAttribute(AmetysObject.class.getName());
            }
        }
        catch (UnknownAmetysObjectException e)
        {
            return null;
        }
    }
}
