/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.rights;

import java.util.Collections;
import java.util.Set;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.content.RootContentHelper;
import org.ametys.cms.content.archive.ArchiveConstants;
import org.ametys.cms.repository.Content;
import org.ametys.core.right.AccessController;
import org.ametys.plugins.core.impl.right.AbstractHierarchicalAccessController;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.collection.AmetysObjectCollection;
import org.ametys.plugins.repository.provider.RequestAttributeWorkspaceSelector;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * {@link AccessController} for a {@link Content}
 */
public class ContentAccessController extends AbstractHierarchicalAccessController<AmetysObject> implements Contextualizable
{
    /** The helper for root content */
    protected RootContentHelper _rootContentHelper;
    /** The helper for contents */
    protected ContentHelper _contentHelper;
    /** The avalon context */
    protected Context _context;
    /** Ametys Object Resolver */
    protected AmetysObjectResolver _resolver;


    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _rootContentHelper = (RootContentHelper) manager.lookup(RootContentHelper.ROLE);
        _contentHelper = (ContentHelper) manager.lookup(ContentHelper.ROLE);
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_ACCESS_CONTROLLER_LABEL");
    }
    
    @Override
    public boolean isSupported(Object object)
    {
        try
        {
            return object instanceof Content && _rootContentHelper.isChildOfRootContent((Content) object)
                   || object != null && object.equals(_rootContentHelper.getRootContent());
        }
        catch (AmetysRepositoryException e)
        {
            getLogger().info("The contents root cannot be retrieved.", e);
            return false;
        }
    }
    
    @Override
    protected Set<AmetysObject> _getParents(AmetysObject object)
    {
        if (object instanceof Content)
        {
            AmetysObject parent = object.getParent();
            if (parent instanceof AmetysObjectCollection && (RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ":contents").equals(((AmetysObjectCollection) parent).getName()))
            {
                Request request = ContextHelper.getRequest(_context);
                String originalWorkspace = RequestAttributeWorkspaceSelector.getForcedWorkspace(request);
                if (ArchiveConstants.ARCHIVE_WORKSPACE.equals(originalWorkspace))
                {
                    try
                    {
                        RequestAttributeWorkspaceSelector.setForcedWorkspace(request, RepositoryConstants.DEFAULT_WORKSPACE);
                        AmetysObject parentFromDefault = _resolver.resolveByPath(parent.getPath());
                        return Collections.singleton(parentFromDefault);
                        
                    }
                    finally
                    {
                        RequestAttributeWorkspaceSelector.setForcedWorkspace(request, originalWorkspace);
                    }
                }
            }
            

            return Collections.singleton(parent);
        }
        else
        {
            return null;
        }
    }
    
    @Override
    protected Set< ? extends Object> _convertWorkspaceToRootRightContexts(Set<Object> workspacesContexts)
    {
        if (workspacesContexts.contains("/cms"))
        {
            return Collections.singleton(_rootContentHelper.getRootContent());
        }
        return null;
    }

}
