/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.rights;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Stream;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.content.RootContentHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.repository.Content;
import org.ametys.core.right.AbstractStaticRightAssignmentContext;
import org.ametys.core.right.RightAssignmentContext;
import org.ametys.core.ui.ClientSideElementHelper;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.collection.AmetysObjectCollection;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * {@link RightAssignmentContext} for assign rights to a {@link Content} or a jcr node root holding the contents
 */
public class ContentRightAssignmentContext extends AbstractStaticRightAssignmentContext
{
    /** The id if this right context */
    public static final String ID = "right.assignment.context.content";
    
    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /** The root content helper */
    protected RootContentHelper _rootContentHelper;

    private ContentTypeExtensionPoint _cTypeEP;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _rootContentHelper = (RootContentHelper) smanager.lookup(RootContentHelper.ROLE);
        _cTypeEP = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    @Override
    public Object convertJSContext(Object context)
    {
        if (context instanceof String)
        {
            return _resolver.resolveById((String) context);
        }
        return null;
    }
    
    @Override
    public String getContextIdentifier(Object context)
    {
        if (context instanceof Content || context instanceof AmetysObjectCollection)
        {
            return ((AmetysObject) context).getId();
        }
        return null;
    }
    
    public boolean isSupportedContext(Object context, Map<String, Object> contextualParameters)
    {
        if (context instanceof Content content)
        {
            // check that any content types is neither private nor simple
            return isSupportedContent(content, contextualParameters);
        }
        else if (context instanceof AmetysObjectCollection collection)
        {
            return collection.equals(getRootContent());
        }
        return false;
    }

    /**
     * Indicate for a given content, if this RightAssignmentContext is able to handle it.
     * 
     * Note that the result of this method should be in coherence with the list of item displayed by in the assignment tool.
     * @param content the content to test
     * @param contextualParameters the application context
     * @return true if the context is supported
     */
    protected boolean isSupportedContent(Content content, Map<String, Object> contextualParameters)
    {
        return Stream.of(content.getTypes())
            .map(_cTypeEP::getExtension)
            .filter(Objects::nonNull)
            .filter(Predicate.not(ContentType::isPrivate))
            .filter(Predicate.not(ContentType::isSimple)) // FIXME this should be ref table instead of simple
            .findAny().isPresent();
    }
    
    public ContextInfo getContextInfo(Object context)
    {
        if (context instanceof Content content)
        {
            return new ContextInfo(
                    getId(),
                    _label,
                    content.getId(),
                    new I18nizableText(content.getTitle()),
                    false
                    );
        }
        else if (context instanceof AmetysObjectCollection collection)
        {
            return new ContextInfo(
                    getId(),
                    _label,
                    collection.getId(),
                    _getRootContextLablel(),
                    true
                    );
        }
        else
        {
            throw new IllegalStateException("Expected a context of type Content or AmetysObjectCollection but got " + context.getClass().getName() + " instead.");
        }
    }

    /**
     * Get the label to use for the root context
     * @return the label
     */
    protected I18nizableText _getRootContextLablel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_RIGHT_ASSIGNMENT_CONTEXT_CONTENTS_ROOT_CONTEXT_LABEL");
    }
    
    @Override
    public Set<Object> getParentContexts(Object context)
    {
        if (context instanceof Content)
        {
            return Collections.singleton(((Content) context).getParent());
        }
        return null;
    }
    
    @Override
    public List<Object> getRootContexts(Map<String, Object> contextParameters)
    {
        List<Object> rootContexts = new ArrayList<>();
        if (matchWorkspace(contextParameters))
        {
            rootContexts.add(getRootContent());
        }
        return rootContexts;
    }
    
    @Override
    public List<Script> getScripts(boolean ignoreRights, Map<String, Object> contextParameters)
    {
        List<Script> scripts = new ArrayList<>();
        
        List<Script> superScripts = super.getScripts(ignoreRights, contextParameters);
        if (superScripts != null && superScripts.size() == 1)
        {
            // First duplicate the script
            Script script = ClientSideElementHelper.cloneScript(superScripts.get(0));
            script.getParameters().put("root-context", ((AmetysObject) getRootContent()).getId());
            scripts.add(script);
        }
        
        return scripts;
    }

    /**
     * Gets the root of contents
     * @return the root of contents
     */
    protected Object getRootContent()
    {
        return _rootContentHelper.getRootContent();
    }
}
