/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.rights;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.core.right.AccessController;
import org.ametys.plugins.core.impl.right.AbstractHierarchicalAccessController;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * {@link AccessController} for a context objects (strings starting with "/content-types").
 */
public class ContentTypeAccessController extends AbstractHierarchicalAccessController<String>
{
    /** The content type extension point */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_TYPE_ACCESS_CONTROLLER_LABEL");
    }
    
    @Override
    public boolean isSupported(Object object)
    {
        return object instanceof String && (Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX.equals(object) || ((String) object).startsWith(Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX + '/'));
    }

    @Override
    protected Set<String> _getParents(String context)
    {
        if (context == null || Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX.equals(context))
        {
            return null;
        }
        
        String ctypeId = StringUtils.substringAfter(context, Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX + "/");
        ContentType contentType = _contentTypeExtensionPoint.getExtension(ctypeId);
        if (contentType == null)
        {
            getLogger().warn("Cannot get hierarchical rights of a content type that do not exists anymore '{}'. Silently goes to '{}'", ctypeId, Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX);
            return null;
        }
        else
        {
            Set<String> parents = new HashSet<>();
            String[] supertypeIds = contentType.getSupertypeIds();
            for (int i = 0; i < supertypeIds.length; i++)
            {
                parents.add(Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX + "/" + supertypeIds[i]);
            }
            
            if (parents.size() == 0 && !contentType.isReferenceTable()) // We do hierarchy for simple, but not until to the root...
            {
                parents.add(Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX);
            }
            
            return parents;
        }
    }
    
    @Override
    protected Set< ? extends Object> _convertWorkspaceToRootRightContexts(Set<Object> workspacesContexts)
    {
        if (workspacesContexts.contains("/cms"))
        {
            return Collections.singleton(Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX);
        }
        return null;
    }
}
