/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.rights;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.core.right.AbstractStaticRightAssignmentContext;
import org.ametys.core.right.RightAssignmentContext;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * {@link RightAssignmentContext} for assign rights to a reference table {@link ContentType} or all reference tables root
 */
public class ReferenceTableRightAssignmentContext extends AbstractStaticRightAssignmentContext
{
    /** The id if this right context */
    public static final String ID = "right.assignment.context.referencetable";
    
    /** The content type extension point */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    /** Once computed, lets remember it */
    protected Boolean _hasAnyReferenceTable;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    @Override
    public Object convertJSContext(Object context)
    {
        if (context instanceof String)
        {
            return context;
        }
        return null;
    }

    @Override
    public String getContextIdentifier(Object context)
    {
        if (context instanceof String)
        {
            return (String) context;
        }
        return null;
    }
    
    public boolean isSupportedContext(Object context, Map<String, Object> contextualParameters)
    {
        if (context instanceof String str)
        {
            if (ReferenceTableAccessController.CONTEXT.equals(str))
            {
                return true;
            }
            else if (StringUtils.startsWith(str, Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX))
            {
                String ctypeId = StringUtils.substringAfter((String) context, Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX + "/");
                ContentType contentType = _contentTypeExtensionPoint.getExtension(ctypeId);
                if (contentType != null && contentType.isReferenceTable())
                {
                    return true;
                }
            }
        }
        return false;
    }
    
    public ContextInfo getContextInfo(Object context)
    {
        I18nizableText contextLabel = getContextLabel((String) context);
        
        return new ContextInfo(getId(), _label, (String) context, contextLabel, ReferenceTableAccessController.CONTEXT.equals(context));
    }

    protected I18nizableText getContextLabel(String context)
    {
        I18nizableText contextLabel = null;
        if (ReferenceTableAccessController.CONTEXT.equals(context))
        {
            contextLabel = new I18nizableText("plugin.cms", "PLUGINS_CMS_RIGHT_ASSIGNMENT_CONTEXT_REFERENCETABLES_ROOT_CONTEXT_LABEL");
        }
        else
        {
            String ctypeId = StringUtils.substringAfter(context, Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX + "/");
            ContentType contentType = _contentTypeExtensionPoint.getExtension(ctypeId);
            if (contentType.isReferenceTable())
            {
                contextLabel = contentType.getLabel();
            }
        }
        
        if (contextLabel == null)
        {
            throw new IllegalStateException("This context is not supported by this right assignment context");
        }
        return contextLabel;
    }
    
    @Override
    public Set<Object> getParentContexts(Object context)
    {
        if (ReferenceTableAccessController.CONTEXT.equals(context))
        {
            return null;
        }
        else if (context instanceof String)
        {
            Set<Object> parents = new HashSet<>();
            
            String ctypeId = StringUtils.substringAfter((String) context, Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX + "/");
            ContentType contentType = _contentTypeExtensionPoint.getExtension(ctypeId);
            
            String[] supertypeIds = contentType.getSupertypeIds();
            for (int i = 0; i < supertypeIds.length; i++)
            {
                String supertypeId = supertypeIds[i];
                if (_contentTypeExtensionPoint.getExtension(supertypeId).isReferenceTable())
                {
                    parents.add(Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX + "/" + supertypeId);
                }
            }
            
            return parents;
        }
        
        return null;
    }
    
    @Override
    public List<Object> getRootContexts(Map<String, Object> contextParameters)
    {
        List<Object> rootContexts = new ArrayList<>();
        if (matchWorkspace(contextParameters))
        {
            rootContexts.add(ReferenceTableAccessController.CONTEXT);
        }
        return rootContexts;
    }
    
    @Override
    public final List<Script> getScripts(boolean ignoreRights, Map<String, Object> contextParameters)
    {
        if (!_hasAnyReferenceTable())
        {
            return Collections.EMPTY_LIST;
        }
        else
        {
            return super.getScripts(ignoreRights, contextParameters);
        }
    }
    
    private boolean _hasAnyReferenceTable()
    {
        if (_hasAnyReferenceTable != null)
        {
            return _hasAnyReferenceTable;
        }
        
        for (String extensionId : _contentTypeExtensionPoint.getExtensionsIds())
        {
            ContentType contentType = _contentTypeExtensionPoint.getExtension(extensionId);
            if (contentType.isReferenceTable())
            {
                _hasAnyReferenceTable = true;
                return true;
            }
        }
        
        _hasAnyReferenceTable = false;
        return false;
    }
}
