/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.schedule;

import java.io.IOException;
import java.util.Optional;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.quartz.JobExecutionContext;

import org.ametys.core.schedule.progression.ContainerProgressionTracker;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.User;
import org.ametys.core.util.I18nUtils;
import org.ametys.core.util.mail.SendMailHelper;
import org.ametys.core.util.mail.SendMailHelper.MailBuilder;
import org.ametys.plugins.core.impl.schedule.AbstractStaticSchedulable;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;

import jakarta.mail.MessagingException;

/**
 * Abstract schedulable that send an email at the end of the execution
 * By default, the email is sent to the user that launched the schedulable. This behavior can be overridden thanks to the {@link #_getRecipient(JobExecutionContext)} method
 * If this user has no email address and there is an error during the execution, an email is sent to the system administrator
 */
public abstract class AbstractSendingMailSchedulable extends AbstractStaticSchedulable implements Initializable
{
    /** The utils for i18n */
    protected I18nUtils _i18nUtils;
    /** Mail sender */
    protected String _mailSender;
    /** Sys admin mail */
    protected String _sysadminMail;
    /** Current user provider */
    protected CurrentUserProvider _currentUserProvider;
    /** User helper */
    protected UserHelper _userHelper;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager); 
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
        _currentUserProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
        _userHelper = (UserHelper) manager.lookup(UserHelper.ROLE);
    }

    public void initialize() throws Exception
    {
        _mailSender = Config.getInstance().getValue("smtp.mail.from");
        _sysadminMail = Config.getInstance().getValue("smtp.mail.sysadminto");
    }

    @Override
    public void execute(JobExecutionContext context, ContainerProgressionTracker progressionTracker) throws Exception
    {
        Optional<String> recipient = _getRecipient(context);
        
        I18nizableText mailSubject = null;
        String mailBody = null;
        try
        {
            _doExecute(context, progressionTracker);
            
            mailSubject = _getSuccessMailSubject(context);
            mailBody = _getSuccessMailBody(context);
        }
        catch (Exception e)
        {
            if (recipient.isEmpty())
            {
                recipient = Optional.ofNullable(_sysadminMail)
                        .filter(StringUtils::isNotEmpty);
            }
            
            mailSubject = _getErrorMailSubject(context);
            mailBody = _getErrorMailBody(context, e);

            throw e;
        }
        finally
        {
            if (recipient.isPresent() && StringUtils.isNotEmpty(mailBody))
            {
                _sendMail(mailSubject, mailBody, recipient.get(), context);
            }
        }
    }
    
    /**
     * Executes the schedulable.
     * @param context the context
     * @param progressionTracker The progression tracker
     * @throws Exception if an error occurred
     */
    protected abstract void _doExecute(JobExecutionContext context, ContainerProgressionTracker progressionTracker) throws Exception;
    
    /**
     * Retrieves the optional recipient of the mail
     * @param context the context
     * @return the optional recipient of the mail
     */
    protected Optional<String> _getRecipient(JobExecutionContext context)
    {
        return Optional.of(_currentUserProvider)
            .map(CurrentUserProvider::getUser)
            .map(_userManager::getUser)
            .map(User::getEmail)
            .filter(StringUtils::isNotEmpty);
    }
    
    /**
     * Determines if the mail body is in HTML
     * @param context the context
     * @return <code>true</code> if the mail body is in HTML, <code>false</code> otherwise
     * @throws Exception If an error occurs while retrieving if mail body should be HTML
     */
    protected boolean _isMailBodyInHTML(JobExecutionContext context) throws Exception
    {
        return false;
    }
    
    /**
     * Retrieves the subject of the success mail
     * @param context the context
     * @return the subject of the success mail
     * @throws Exception If an error occurs while building the mail subject
     */
    protected abstract I18nizableText _getSuccessMailSubject(JobExecutionContext context) throws Exception;
    
    /**
     * Retrieves the body of the success mail
     * @param context the context
     * @return the body of the success mail
     * @throws Exception If an error occurs while building the mail body
     */
    protected abstract String _getSuccessMailBody(JobExecutionContext context) throws Exception;
    
    /**
     * Retrieves the subject of the error mail
     * @param context the context
     * @return the subject of the error mail
     * @throws Exception If an error occurs while building the mail subject
     */
    protected abstract I18nizableText _getErrorMailSubject(JobExecutionContext context) throws Exception;
    
    /**
     * Retrieves the body of the error mail
     * @param context the context
     * @param throwable the error
     * @return the body of the error mail
     * @throws Exception If an error occurs while building the mail body
     */
    protected abstract String _getErrorMailBody(JobExecutionContext context, Throwable throwable) throws Exception;
    
    /**
     * Send an email
     * @param subject the email's subject
     * @param body the email's body (to HTML or text format depending on {@link #_isMailBodyInHTML(JobExecutionContext)}
     * @param recipient the recipient address
     * @param context the context
     */
    protected void _sendMail (I18nizableText subject, String body, String recipient, JobExecutionContext context)
    {
        try
        {
            MailBuilder mailBuilder = SendMailHelper.newMail()
                                                    .withSubject(_i18nUtils.translate(subject))
                                                    .withRecipient(recipient)
                                                    .withSender(_mailSender);
            
            if (_isMailBodyInHTML(context))
            {
                mailBuilder.withHTMLBody(body);
            }
            else
            {
                mailBuilder.withTextBody(body);
            }
            
            mailBuilder.sendMail();
        }
        catch (MessagingException | IOException e)
        {
            if (getLogger().isWarnEnabled())
            {
                getLogger().warn("Unable to send the e-mail '" + subject  + "' to '" + recipient + "'", e);
            }
        }
        catch (Exception e)
        {
            getLogger().error("An unknown error has occured while sending the mail.", e);
        }
    }
}
