/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.scripts;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.configuration.MutableConfiguration;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.content.ContentValuesExtractorFactory;
import org.ametys.cms.search.content.ContentValuesExtractorFactory.SearchModelContentValuesExtractor;
import org.ametys.cms.search.model.SearchModelHelper;
import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.cms.search.ui.model.SearchUIModel;
import org.ametys.cms.search.ui.model.SearchUIModelExtensionPoint;
import org.ametys.cms.search.ui.model.StaticSearchUIModel;
import org.ametys.plugins.core.ui.script.ScriptExecArguments;
import org.ametys.plugins.core.ui.script.ScriptHandler;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ViewItemContainer;
import org.ametys.runtime.model.ViewParser;
import org.ametys.runtime.model.type.ModelItemType;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * Content aware script handler using search model 
 */
public class CmsScriptHandler extends ScriptHandler
{
    private SearchModelHelper _searchUIModelHelper;
    private SearchUIModelExtensionPoint _searchUIModelEP;
    private ContentTypesHelper _contentTypesHelper;
    private ContentTypeExtensionPoint _contentTypeExtensionPoint;
    private ServiceManager _manager;
    private ContentValuesExtractorFactory _valuesExtractorFactory;
    private ContentHelper _contentHelper;
    private SystemPropertyExtensionPoint _systemPropertyExtensionPoint;

    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        _manager = serviceManager;
        super.service(serviceManager);
        _searchUIModelEP = (SearchUIModelExtensionPoint) serviceManager.lookup(SearchUIModelExtensionPoint.ROLE);
        _contentTypesHelper = (ContentTypesHelper) serviceManager.lookup(ContentTypesHelper.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) serviceManager.lookup(ContentTypeExtensionPoint.ROLE);
        _searchUIModelHelper = (SearchModelHelper) serviceManager.lookup(SearchModelHelper.ROLE);
        _valuesExtractorFactory = (ContentValuesExtractorFactory) serviceManager.lookup(ContentValuesExtractorFactory.ROLE);
        _contentHelper = (ContentHelper) serviceManager.lookup(ContentHelper.ROLE);
        _systemPropertyExtensionPoint = (SystemPropertyExtensionPoint) serviceManager.lookup(SystemPropertyExtensionPoint.ROLE);
    }
    
    @Override
    protected ScriptExecArguments buildExecArguments(Map<String, Object> arguments)
    {
        return new CmsScriptExecArguments.Impl(arguments);
    }
    
    @SuppressWarnings("unchecked")
    @Override
    protected Object processScriptResult(Map<String, Object> results, Object scriptResult, ScriptExecArguments execArgs)
    {
        Object processedScriptResult = super.processScriptResult(results, scriptResult, execArgs);
        
        Collection<String> columns = null;
        String defaultModelId = Optional.of(execArgs)
                .filter(CmsScriptExecArguments.class::isInstance)
                .map(CmsScriptExecArguments.class::cast)
                .flatMap(CmsScriptExecArguments::searchModelId)
                .orElse(null);
        List<Content> contents = null;
        SearchUIModel model = null;

        Object processedResults = Optional.ofNullable(results.get("results")).orElse(results.get("contents"));
        
        Object columnsObject = results.get("columns");
        if (columnsObject instanceof Collection)
        {
            columns = (Collection<String>) columnsObject;
        }
        
        if (processedResults instanceof Collection)
        {
            Collection< ? > proccessedResultsCollection = (Collection<?>) processedResults;
            contents = proccessedResultsCollection.stream()
                    .filter(Content.class::isInstance)
                    .map(Content.class::cast)
                    .collect(Collectors.toList());
        }
        
        try
        {
            model = getOrCreateModel(columns, contents, defaultModelId);
            if (columns != null)
            {
                Map<String, Object> modelConfiguration = _searchUIModelHelper.getSearchModelInfo(model, Collections.EMPTY_MAP);
                results.put("columns", modelConfiguration.get("columns"));
            }
            
            if (model != null && contents != null && !contents.isEmpty())
            {
                ViewItemContainer resultItems = model.getResultItems(Collections.EMPTY_MAP);
                SearchModelContentValuesExtractor extractor = _valuesExtractorFactory.create(model);
                List<Map<String, Object>> contentsJson = contents.stream()
                                                                 .map(content -> content2Json(content, resultItems, extractor))
                                                                 .collect(Collectors.toList());
                results.put("contents", contentsJson);
                results.put("size", contents.size());
            }
        }
        catch (Exception e)
        {
            throw new RuntimeException(e);
        }
        
        return processedScriptResult;
    }
    
    @Override
    protected ResultProcessor getProcessor()
    {
        return new CmsResultProcessor();
    }

    /**
     * Computes the model associated with the given contents and result columns.<br>
     * Used by scripts functions creating reports or search results.
     * @param columns the columns of the results
     * @param contents the contents for the results (one content per line)
     * @param defaultModelId the fallback model
     * @return the computed search model
     * @throws Exception if something went wrong.
     */
    public SearchUIModel getOrCreateModel(Collection<String> columns, List<Content> contents, String defaultModelId) throws Exception
    {
        if (columns != null
            || contents != null && !contents.isEmpty())
        {
            ThreadSafeComponentManager<SearchUIModel> localSearchModelManager = null;
            // Handling model
            try
            {
                localSearchModelManager = new ThreadSafeComponentManager<>();
                localSearchModelManager.setLogger(getLogger());
                localSearchModelManager.contextualize(_context);
                localSearchModelManager.service(_manager);
                
                return _createModel(localSearchModelManager, defaultModelId, columns, contents);
            }
            catch (Exception e)
            {
                getLogger().error("Error while retrieving the search model :" + e.getMessage(), e);
                throw new Exception("Error while retrieving the search model : " + e.getMessage(), e);
            }
            finally
            {
                if (localSearchModelManager != null)
                {
                    localSearchModelManager.dispose();
                }
            }
        }
        else
        {
            return defaultModelId != null ? _searchUIModelEP.getExtension(defaultModelId) : null;
        }
    }
    

    /**
     * Create and return a dynamic model based on desired columns or return a default model.
     * @param localSearchModelManager The local search manager
     * @param defaultModelId The default model id
     * @param columns The columns
     * @param contents The contents
     * @return The search model
     * @throws Exception If an error occurred 
     */
    protected SearchUIModel _createModel(ThreadSafeComponentManager<SearchUIModel> localSearchModelManager, String defaultModelId, Collection<String> columns, Collection<Content> contents) throws Exception
    {
        /*
         * Configuration will have the following structure :
         *  <SearchModel>
         *      <content-types>
         *          <content-type id="CTYPE_ID"/>
         *          <...>
         *      </content-types>
         *      <columns>
         *          <default>
         *              <item ref="COLUMN_ID|*" [specific attr might be needed depending on column]>
         *                  [specific value or child elements might be needed depending on column]
         *              </item>
         *              <...>
         *          </default>
         *      </columns>
         *  </SearchModel>
         */
        
        MutableConfiguration conf = new DefaultConfiguration((String) null);
        MutableConfiguration modelConf = new DefaultConfiguration("SearchModel");
        conf.addChild(modelConf);
        
        // content types
        MutableConfiguration contentTypesConf = new DefaultConfiguration("content-types");
        modelConf.addChild(contentTypesConf);
        
        Set<ContentType> cTypeCommonAncestors = new HashSet<>();
        if (contents != null)
        {
            Set<String> contentTypeIds = new HashSet<>();
            for (Content content : contents)
            {
                contentTypeIds.addAll(Arrays.asList(content.getTypes()));
            }
            
            cTypeCommonAncestors = _contentTypesHelper.getCommonAncestors(contentTypeIds).stream()
                    .map(_contentTypeExtensionPoint::getExtension)
                    .collect(Collectors.toSet());
        }
        
        for (ContentType ancestor : cTypeCommonAncestors)
        {
            MutableConfiguration cTypeConf = new DefaultConfiguration("content-type");
            cTypeConf.setAttribute("id", ancestor.getId());
            contentTypesConf.addChild(cTypeConf);
        }
        
        // columns
        MutableConfiguration columnsConf = new DefaultConfiguration("columns");
        modelConf.addChild(columnsConf);
        
        MutableConfiguration columnsDefaultConf = new DefaultConfiguration("default");
        columnsConf.addChild(columnsDefaultConf);
        
        if (columns != null && !columns.isEmpty())
        {
            // Add an item for each column
            for (String column : columns)
            {
                String modelItemPath = StringUtils.replace(column, ".", ModelItem.ITEM_PATH_SEPARATOR);
                MutableConfiguration columnConf = new DefaultConfiguration(ViewParser.ADD_ITEM_TAG_NAME);
                columnConf.setAttribute(ViewParser.ITEM_REFERENCE_ATTRIBUTE_NAME, modelItemPath);
                columnsDefaultConf.addChild(columnConf);
            }
        }
        else if (_hasNonAbstractAncestors(cTypeCommonAncestors))
        {
            // Add all attributes
            MutableConfiguration columnConf = new DefaultConfiguration(ViewParser.ADD_ITEM_TAG_NAME);
            columnConf.setAttribute(ViewParser.ITEM_REFERENCE_ATTRIBUTE_NAME, ViewParser.ALL_ITEMS_REFERENCE);
            columnsDefaultConf.addChild(columnConf);
            
            // Add all system properties
            for (String systemPropertyId : _systemPropertyExtensionPoint.getExtensionsIds())
            {
                if (_systemPropertyExtensionPoint.isDisplayable(systemPropertyId))
                {
                    columnConf = new DefaultConfiguration(ViewParser.ADD_ITEM_TAG_NAME);
                    columnConf.setAttribute(ViewParser.ITEM_REFERENCE_ATTRIBUTE_NAME, systemPropertyId);
                    columnsDefaultConf.addChild(columnConf);
                }
            }
        }
        
        if (columnsDefaultConf.getChildren().length == 0)
        {
            if (getLogger().isInfoEnabled())
            {
                getLogger().info("No columns found. The default model will be used");
            }
            
            return _searchUIModelEP.getExtension(defaultModelId);
        }
        else
        {
            localSearchModelManager.addComponent("script", null, "script-search-model", StaticSearchUIModel.class, conf);
            localSearchModelManager.initialize();
            return localSearchModelManager.lookup("script-search-model");
        }
    }
    
    private boolean _hasNonAbstractAncestors(Set<ContentType> commonAncestors)
    {
        for (ContentType ancestor : commonAncestors)
        {
            if (ancestor.getView("main") != null && ancestor.hasModelItem(Content.ATTRIBUTE_TITLE))
            {
                return true;
            }
        }
        
        // None of the content type has a main view and a title attribute
        return false;
    }
    
    /**
     * Convert content to json
     * @param content The content
     * @param resultItems The result items, to know which value to fill
     * @param extractor The properties extractor
     * @return The json data
     */
    public Map<String, Object> content2Json(Content content, ViewItemContainer resultItems, SearchModelContentValuesExtractor extractor)
    {
        Map<String, Object> contentData = new HashMap<>();
        
        contentData.put("id", content.getId());
        contentData.put("name", content.getName());
        
        if (_contentHelper.isMultilingual(content) && _isTitleMultilingual(content))
        {
            contentData.put(Content.ATTRIBUTE_TITLE, _contentHelper.getTitleVariants(content));
        }
        else
        {
            contentData.put(Content.ATTRIBUTE_TITLE, _contentHelper.getTitle(content));
        }
        
        contentData.put("language", content.getLanguage());
        contentData.put("contentTypes", content.getTypes());
        contentData.put("mixins", content.getMixinTypes());
        contentData.put("iconGlyph", _contentTypesHelper.getIconGlyph(content));
        contentData.put("iconDecorator", _contentTypesHelper.getIconDecorator(content));
        contentData.put("smallIcon", _contentTypesHelper.getSmallIcon(content));
        contentData.put("mediumIcon", _contentTypesHelper.getMediumIcon(content));
        contentData.put("largeIcon", _contentTypesHelper.getLargeIcon(content));
        contentData.put("isSimple", _contentHelper.isSimple(content));
        
        contentData.putAll(extractor.getValues(content, null, new HashMap<>()));
        
        return contentData;
    }
    
    private boolean _isTitleMultilingual(Content content)
    {
        return Optional.ofNullable(content)
                .map(c -> c.getDefinition(Content.ATTRIBUTE_TITLE))
                .map(ModelItem::getType)
                .map(ModelItemType::getId)
                .map(ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID::equals)
                .orElse(false);
    }
    
    static class CmsResultProcessor extends ResultProcessor
    {
        @SuppressWarnings("unchecked")
        @Override
        protected Object process(Map<String, Object> results, Object scriptResult)
        {
            if (scriptResult instanceof Content)
            {
                List<Object> contents = (List<Object>) results.computeIfAbsent("contents", __ -> new ArrayList<>());
                contents.add(scriptResult);
                
                return ((Content) scriptResult).toString();
            }
            else if (scriptResult instanceof Map)
            {
                Map<Object, Object> elements = new HashMap<>();
                Map scriptResultMap = (Map) scriptResult;
                List<String> contents = _processScriptResultContents(results, scriptResultMap);
                if (contents != null)
                {
                    elements.put("results", contents);
                }

                if (scriptResultMap.containsKey("columns"))
                {
                    results.put("columns", scriptResultMap.get("columns"));
                }
                
                // Map
                for (Object key : scriptResultMap.keySet())
                {
                    if (!"results".equals(key))
                    {
                        Object value = scriptResultMap.get(key);
                        elements.put(process(results, key), process(results, value));
                    }
                }
                
                return elements;
            }
            
            return super.process(results, scriptResult);
        }
        
        @SuppressWarnings("unchecked")
        private List<String> _processScriptResultContents(Map<String, Object> results, Map scriptResultMap)
        {
            if (scriptResultMap.containsKey("results"))
            {
                Object rawResults = scriptResultMap.get("results");
                Collection<Content> rawResultCollection = null;
                if (rawResults instanceof AmetysObjectIterable)
                {
                    try (AmetysObjectIterable<Content> rawResultIterable = (AmetysObjectIterable<Content>) rawResults)
                    {
                        rawResultCollection = rawResultIterable.stream()
                                .collect(Collectors.toList());
                    }
                }
                else if (rawResults instanceof Collection)
                {
                    rawResultCollection = new LinkedList<>((Collection<Content>) rawResults);
                }
                else if (rawResults instanceof Map)
                {
                    rawResultCollection = ((Map) rawResults).values();
                }
                
                if (rawResultCollection != null)
                {
                    results.put("contents", rawResultCollection);
                    return rawResultCollection.stream()
                            .map(content -> content.toString())
                            .collect(Collectors.toList());
                }
            }
            return null;
        }
    }
}
