/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.search.Sort.Order;
import org.ametys.cms.search.model.SearchModel;
import org.ametys.cms.search.model.SearchModelHelper;
import org.ametys.cms.search.query.AndQuery;
import org.ametys.cms.search.query.ContentTypeOrMixinTypeQuery;
import org.ametys.cms.search.query.ContentTypeQuery;
import org.ametys.cms.search.query.MixinTypeQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.query.QuerySyntaxException;
import org.ametys.cms.search.ui.model.SearchUIColumn;
import org.ametys.core.util.JSONUtils;
import org.ametys.runtime.model.ViewHelper;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.ViewItemContainer;

/**
 * Helper to get query infos from JSON object
 */
public class GetQueryFromJSONHelper extends AbstractLogEnabled implements Component, Serviceable
{
    /** The Avalon role name */
    public static final String ROLE = GetQueryFromJSONHelper.class.getName();
    
    private QueryBuilder _queryBuilder;
    private JSONUtils _jsonUtils;
    private SearchModelHelper _searchModelHelper;
    
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        _queryBuilder = (QueryBuilder) serviceManager.lookup(QueryBuilder.ROLE);
        _jsonUtils = (JSONUtils) serviceManager.lookup(JSONUtils.ROLE);
        _searchModelHelper = (SearchModelHelper) serviceManager.lookup(SearchModelHelper.ROLE);
    }

    /**
     * Retrieves the search UI model
     * @param parameters The JS parameters
     * @return the search UI model
     */
    public SearchModel getSearchModel(Map<String, Object> parameters)
    {
        return getSearchModel(parameters, "model");
    }
    
    /**
     * Retrieves the search UI model
     * @param parameters The JS parameters
     * @param modelKeyName key name of the model parameter
     * @return the search UI model
     */
    @SuppressWarnings("unchecked")
    public SearchModel getSearchModel(Map<String, Object> parameters, String modelKeyName)
    {
        String modelId = (String) parameters.get(modelKeyName);

        List<String> restrictedContentTypeIds = (List<String>) parameters.get("restrictedContentTypes");
        Map<String, Object> contextualParameters = Optional.ofNullable((Map<String, Object>) parameters.get("contextualParameters")).orElse(new HashMap<>());
        
        return _searchModelHelper.getSearchUIModel(modelId, restrictedContentTypeIds, contextualParameters);
    }
    
    /**
     * Retrieves a Query object from the SearchUIModel and the JSON parameters
     * @param model Model
     * @param parameters JSON parameters
     * @param contentTypes List of content types to fill
     * @return the created Query object
     * @throws QuerySyntaxException If an error occurs during the query parsing
     */
    @SuppressWarnings("unchecked")
    public Query getQueryFromModel(SearchModel model, Map<String, Object> parameters, List<String> contentTypes) throws QuerySyntaxException
    {
        Map<String, Object> contextualParameters = Optional.ofNullable((Map<String, Object>) parameters.get("contextualParameters")).orElseGet(HashMap::new);
        
        Map<String, Object> values = (Map<String, Object>) parameters.get("values");
        String searchMode = StringUtils.defaultString((String) parameters.get("searchMode"), "simple");

        Query query = _queryBuilder.build(model, searchMode, true , values, contextualParameters);
        if (query instanceof AndQuery)
        {
            Set<Query> subqueries = new LinkedHashSet<>(((AndQuery) query).getQueries());
            Predicate<Query> isCTypeOrMixinOrBothQuery = q -> ContentTypeQuery.class.isInstance(q) || MixinTypeQuery.class.isInstance(q) || ContentTypeOrMixinTypeQuery.class.isInstance(q);
            List<Query> matchingQueries = subqueries.stream().distinct().filter(isCTypeOrMixinOrBothQuery).collect(Collectors.toList());
            
            if (matchingQueries.size() > 1)
            {
                // Mutliple candidates, try to be more selective and only keep a `ContentTypeQuery` one
                Optional<Query> cTypeQuery = matchingQueries.stream().filter(ContentTypeQuery.class::isInstance).findFirst();
                if (cTypeQuery.isPresent())
                {
                    matchingQueries = Collections.singletonList(cTypeQuery.get());
                }
                else
                {
                    // Mutliple candidates, try to be more selective and only keep a `ContentTypeOrMixinTypeQuery` one
                    cTypeQuery = matchingQueries.stream().filter(ContentTypeOrMixinTypeQuery.class::isInstance).findFirst();
                    if (cTypeQuery.isPresent())
                    {
                        matchingQueries = Collections.singletonList(cTypeQuery.get());
                    }
                }
            }
            
            if (matchingQueries.size() == 1)
            {
                // be smart, fill the "contentTypes" field with this query
                Query matchingQuery = matchingQueries.iterator().next();
                boolean changeQuery = false;
                if (matchingQuery instanceof ContentTypeQuery)
                {
                    ContentTypeQuery cTypeQuery = (ContentTypeQuery) matchingQuery;
                    if (cTypeQuery.getOperator() == Operator.EQ)
                    {
                        contentTypes.addAll(cTypeQuery.getValue());
                        changeQuery = true;
                    }
                }
                else if (matchingQuery instanceof MixinTypeQuery)
                {
                    MixinTypeQuery mixinQuery = (MixinTypeQuery) matchingQuery;
                    if (mixinQuery.getOperator() == Operator.EQ)
                    {
                        contentTypes.addAll(mixinQuery.getValue());
                        changeQuery = true;
                    }
                }
                else if (matchingQuery instanceof ContentTypeOrMixinTypeQuery)
                {
                    ContentTypeOrMixinTypeQuery cTypeOrMixinQuery = (ContentTypeOrMixinTypeQuery) matchingQuery;
                    if (cTypeOrMixinQuery.getOperator() == Operator.EQ)
                    {
                        contentTypes.addAll(cTypeOrMixinQuery.getIds());
                        changeQuery = true;
                    }
                }
                
                if (changeQuery)
                {
                    // Change the query by putting the same AndQuery but without the part about contentTypes which are in the contentTypes field
                    subqueries.remove(matchingQuery);
                    query = new AndQuery(subqueries);
                }
            }
        }
        
        return query;
    }
    
    /**
     * Retrieves a list of column ids
     * @param model search model containing result fields
     * @param contextualParameters JSON object containing the contextual parameters
     * @return a list of columns ids
     */
    public List<String> getColumnsFromSearchModel(SearchModel model, Map<String, Object> contextualParameters)
    {
        ViewItemContainer resultItems = model.getResultItems(contextualParameters != null ? contextualParameters : Collections.EMPTY_MAP);
        return _getColumnIds(resultItems);
    }
    
    private List<String> _getColumnIds(ViewItemAccessor viewItemAccessor)
    {
        List<String> columnIds = new ArrayList<>();
        
        for (ViewItem viewItem : viewItemAccessor.getViewItems())
        {
            if (viewItem instanceof SearchUIColumn searchUIColumn)
            {
                String columnId = ViewHelper.getModelViewItemPath(searchUIColumn);
                columnIds.add(columnId);
            }
            else if (viewItem instanceof ViewItemAccessor itemAccessor)
            {
                columnIds.addAll(_getColumnIds(itemAccessor));
            }
        }
        
        return columnIds;
    }
    
    /**
     * Retrieves the sort criteria from a sort string.
     * @param parameters The parameters (containing sort and optionally group keys)
     * @return the sort criteria as a List of {@link Sort}.
     */
    public List<Sort> getSort(Map<String, Object> parameters)
    {
        String sortString = (String) parameters.get("sort");
        String groupString = (String) parameters.get("group");
        return getSort(sortString, groupString);
    }
    
    /**
     * Retrieves the sort criteria from a sort string.
     * @param sortString The sort criteria as a JSON-encoded string.
     * @param groupString The group criteria as a JSON-encoded string (for server-side grouping feature). Can be null.
     * @return the sort criteria as a List of {@link Sort}.
     */
    public List<Sort> getSort(String sortString, String groupString)
    {
        List<Sort> sort = new ArrayList<>();
        
        List<Object> sortList = new ArrayList<>(_jsonUtils.convertJsonToList(sortString));
        if (StringUtils.isNotEmpty(groupString))
        {
            // Grouping will be treated server side as a sort. It just needs to be before all the sorters
            sortList.add(0, _jsonUtils.convertJsonToMap(groupString));
        }
        
        for (Object object : sortList)
        {
            if (object instanceof Map)
            {
                Map map = (Map) object;
                String fieldId = (String) map.get("property");
                boolean ascending = "ASC".equals(map.get("direction"));
                
                sort.add(new Sort(fieldId, ascending ? Order.ASC : Order.DESC));
            }
        }
        
        return sort;
    }    
}

