/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.data.ContentValue;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.cocoon.GroupSearchContent;
import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.util.DateUtils;
import org.ametys.core.util.I18nUtils;
import org.ametys.plugins.repository.metadata.MultilingualString;
import org.ametys.plugins.repository.metadata.MultilingualStringHelper;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.ModelItem;

/**
 * Helper to group search contents
 */
public class GroupSearchContentHelper implements Component, Serviceable
{
    /** The Avalon role name */
    public static final String ROLE = GroupSearchContentHelper.class.getName();
    
    private SystemPropertyExtensionPoint _systemPropertyExtensionPoint;
    private ContentHelper _contentHelper;
    private UserManager _userManager;
    private I18nUtils _i18nUtils;

    public void service(ServiceManager serviceManager) throws ServiceException
    {
        _systemPropertyExtensionPoint = (SystemPropertyExtensionPoint) serviceManager.lookup(SystemPropertyExtensionPoint.ROLE);
        _contentHelper = (ContentHelper) serviceManager.lookup(ContentHelper.ROLE);
        _userManager = (UserManager) serviceManager.lookup(UserManager.ROLE);
        _i18nUtils = (I18nUtils) serviceManager.lookup(I18nUtils.ROLE);
    }
    
    /**
     * Create GroupSearchContent to organize the contents in a hierarchy according to a list of attributes
     * Example Continent/Country/City
     * @param contents contents that need to be sorted
     * @param groupingFields list of attributes to group with (ordered)
     * @param contentTypes the content type
     * @param locale The locale for groups. Can be null. Only useful if grouping by multilingual contents or multilingual string
     * @return the root group of contents
     */
    public GroupSearchContent organizeContentsInGroups(Iterable<Content> contents, List<String> groupingFields, Set<ContentType> contentTypes, Locale locale)
    {
        GroupSearchContent groupSearchContent = new GroupSearchContent();
        _organizeContents(groupSearchContent, contents, groupingFields, contentTypes, locale);
        return groupSearchContent;
    }
    
    /**
     * Recursive function that create sub GroupSearchContent to organize the contents in a hierarchy according to a list of attributes
     * Example Continent/Country/City
     * @param parent parent group in which contents will be added, or in which new groups will be created
     * @param contents contents that need to be sorted
     * @param groupingFields list of attributes to group with (ordered)
     * @param contentTypes the content types
     * @param locale The local to get i18n titles
     */
    private void _organizeContents(GroupSearchContent parent, Iterable<Content> contents, List<String> groupingFields, Set<ContentType> contentTypes, Locale locale)
    {
        if (groupingFields != null && groupingFields.size() > 0)
        {
            String fieldId = groupingFields.get(0);
            // Group contents by attribute value
            Map<Object, List<Content>> groups = _groupBy(contents, fieldId);

            // Get the rest of the attribute regroupment list
            List<String> subAttributeRegroupments = new ArrayList<>();
            if (groupingFields.size() > 1)
            {
                subAttributeRegroupments.addAll(groupingFields);
                subAttributeRegroupments.remove(0);
            }
            // Then : add sorted contents into groups, or re-organize if there are still some attributeRegroupments
            for (Map.Entry<Object, List<Content>> entry : groups.entrySet())
            {
                String keyStr = _getGroupValueAsString(fieldId, entry.getKey(), contentTypes, locale);
                
                GroupSearchContent groupSearchContent = new GroupSearchContent(fieldId, keyStr);
                if (groupingFields.size() == 1)
                {
                    // If no more groups-level, add all contents in the group
                    groupSearchContent.addContents(entry.getValue());
                }
                else
                {
                    // If more groups to do : let's do it !
                    _organizeContents(groupSearchContent, entry.getValue(), subAttributeRegroupments, contentTypes, locale);
                }
                // Add current group into parent
                parent.addToSubList(groupSearchContent);
            }
        }
        else
        {
            // If there are no group to do, just add all contents into the parent
            contents.forEach(parent::addContent);
        }
    }
    
    /**
     * Group contents by given grouping field
     * @param contents list of content
     * @param fieldId The id of grouping field. Can be the path of an attribute or the system property
     * @return a Map where the key will be the value of the attribute
     */
    private Map<Object, List<Content>> _groupBy(Iterable<Content> contents, String fieldId)
    {
        Map<Object, List<Content>> groups = new HashMap<>();
        for (Content content : contents)
        {
            Object value = _contentHelper.getValue(content, fieldId);
            if (value instanceof List)
            {
                // If the value is a list, the content is added into different groups, depending on the values in the list
                List valueList = (List) value;
                for (Object subValue : valueList)
                {
                    groups.computeIfAbsent(subValue, __ -> new ArrayList<>())
                          .add(content);
                }
            }
            else
            {
                groups.computeIfAbsent(value, __ -> new ArrayList<>())
                      .add(content);
            }
        }
        return groups;
    }
    
    private String _getGroupValueAsString(String fieldId, Object value, Set<ContentType> contentTypes, Locale locale)
    {
        if (value == null)
        {
            return StringUtils.EMPTY;
        }
        else if (value instanceof ContentValue)
        {
            return ((ContentValue) value).getContentIfExists()
                                         .map(content -> content.getTitle(locale))
                                         .orElse(StringUtils.EMPTY);
        }
        else if (value instanceof UserIdentity)
        {
            User user = _userManager.getUser((UserIdentity) value);
            return user != null ? user.getFullName() : UserIdentity.userIdentityToString((UserIdentity) value);
        }
        else if (value instanceof Boolean)
        {
            String language = Optional.ofNullable(locale)
                    .map(Locale::getLanguage)
                    .orElse(null);
            StringBuilder sb = new StringBuilder();
            if (ModelHelper.hasModelItem(fieldId, contentTypes))
            {
                ModelItem modelItem = ModelHelper.getModelItem(fieldId, contentTypes);
                sb.append(_i18nUtils.translate(modelItem.getLabel(), language));
                sb.append(_i18nUtils.translate(new I18nizableText("plugin.cms", "PLUGINS_CMS_SEARCH_EXPORT_SEPARATOR"), language));
            }
            sb.append(_i18nUtils.translate(new I18nizableText("plugin.cms", "PLUGINS_CMS_SEARCH_EXPORT_BOOLEAN_" + ((Boolean) value).toString().toUpperCase()), language));
            return sb.toString();
        }
        else if (value instanceof MultilingualString)
        {
            return MultilingualStringHelper.getValue((MultilingualString) value, locale);
        }
        else if (!_systemPropertyExtensionPoint.hasExtension(fieldId))
        {
            ModelItem modelItem = ModelHelper.getModelItem(fieldId, contentTypes);
            if (modelItem instanceof ElementDefinition)
            {
                ElementDefinition definition = (ElementDefinition) modelItem;
                if (definition.getEnumerator() != null)
                {
                    try
                    {
                        I18nizableText label = definition.getEnumerator().getEntry(value);
                        return Optional.ofNullable(label)
                                       .map(_i18nUtils::translate)
                                       .orElse(StringUtils.EMPTY);
                    }
                    catch (Exception e)
                    {
                        // An unexpected error occurs while getting the enumerator entry
                        return definition.getType().toString(value);
                    }
                }
                else
                {
                    return definition.getType().toString(value);
                }
            }
            else
            {
                throw new IllegalArgumentException("Attribute at path '" + modelItem.getPath() + "' is a composite or a repeater : can not group values");
            }
        }
        else if (value instanceof Date)
        {
            return DateUtils.dateToString((Date) value);
        }
        else
        {
            return value.toString();
        }

    }
}
