/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.advanced;

import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.ametys.cms.search.query.Query.LogicalOperator;

/**
 * This abstract class represents a node in a tree data structure. <br>
 * There are only two allowed sublcasses: {@link TreeLeaf} and {@link TreeInternalNode}
 * 
 * @param <T> the type of the values of the leaves of the tree.
 */
public abstract class AbstractTreeNode<T>
{
    /**
     * Generic helper method to walk a tree (i.e. traverse a tree) for visiting each node and doing
     * a recursive treatment
     * <br>This is a post-order walk (the children are traversed before their respective parents are traversed)
     * <br>See also <a href="https://en.wikipedia.org/wiki/Tree_traversal#Post-order">this link</a>
     * @param <T> the type of the values of the leaves of the given tree
     * @param <R> the type of the object returned by this method.
     * @param tree The tree to walk
     * @param leafFunction The operation to perform for a leaf
     * @param internalNodeFunction The operation to perform for an internal node.
     * <br>This function takes two arguments: <ul><li>the results for each child as a stream; </li><li>the logical operator of the internal node</li></ul>
     * @return a result
     */
    public static <T, R> R walk(AbstractTreeNode<T> tree, 
                        Function<TreeLeaf<T>, R> leafFunction, 
                        BiFunction<Stream<R>, LogicalOperator, R> internalNodeFunction)
    {
        if (tree instanceof TreeLeaf<?>)
        {
            return leafFunction.apply((TreeLeaf<T>) tree);
        }
        else if (tree instanceof TreeInternalNode<?>)
        {
            TreeInternalNode<T> internalNode = (TreeInternalNode<T>) tree;
            // localResult is the result of already walked children, in a stream
            Stream<R> localResult = internalNode.getChildren()
                    .stream()
                    .map(child -> walk(child, leafFunction, internalNodeFunction));
            // from localResult stream and the current logical operator, apply function to get final result
            return internalNodeFunction.apply(localResult, internalNode.getLogicalOperator());
        }
        else
        {
            throw new IllegalArgumentException("Wrong type of TreeNode. It must only be a TreeInternalNode or a TreeLeaf.");
        }
    }
    
    /**
     * Gets all leaves of this tree node, recursively flattened.
     * @return the flattened leaves
     */
    public Collection<TreeLeaf<T>> getFlatLeaves()
    {
        return walk(this,
            // leaf => just return it in a singleton
            Collections::singletonList,
            // internal node => collect in a list
            (singletons, operator) -> singletons.flatMap(List<TreeLeaf<T>>::stream)
                                                .collect(Collectors.toList()));
    }
}

