/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.advanced;

import java.util.List;
import java.util.function.Function;
import java.util.stream.Stream;

import org.apache.avalon.framework.component.Component;

import org.ametys.cms.search.query.AndQuery;
import org.ametys.cms.search.query.OrQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.LogicalOperator;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Builds a {@link Query} object from advanced search criteria (as a {@link AbstractTreeNode}).
 */
public class AdvancedQueryBuilder extends AbstractLogEnabled implements Component
{
    /** The component role. */
    public static final String ROLE = AdvancedQueryBuilder.class.getName();
    
    /**
     * Builds the {@link Query} object represented by the given tree of advanced search criteria.
     * @param <T> the type of the values of the leaves of the given tree
     * @param tree The tree
     * @param queryMapper The mapper to build a Query from a {@link TreeLeaf leaf}
     * @return the built query of the {@link AbstractTreeNode tree}
     */
    public <T> Query build(AbstractTreeNode<T> tree, Function<T, Query> queryMapper)
    {
        return AbstractTreeNode.walk(tree, 
            // leaf => get the value of the leaf, apply the mapper to have the leaf Query
            leaf -> queryMapper.apply(leaf.getValue()),
            // internalNode => build the logical (OR | AND) query of the internal node
            (queries, operator) -> _logicalQuery(queries, operator));
    }
    
    private Query _logicalQuery(Stream<Query> queries, LogicalOperator operator)
    {
        List<Query> queriesAsCol = queries.toList();
        return operator == LogicalOperator.AND ? new AndQuery(queriesAsCol) : new OrQuery(queriesAsCol);
    }
}
