/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.advanced;

import java.util.List;

import org.ametys.cms.search.query.Query.LogicalOperator;

/**
 * This class represents an internal node in a tree data structure, which have {@link AbstractTreeNode} {@link #getChildren children} (at least 1), 
 * linked by a {@link #getLogicalOperator() logical operator}.
 * @param <T> the type of the values of the leaves of the tree.
 */
public final class TreeInternalNode<T> extends AbstractTreeNode<T>
{
    private List<AbstractTreeNode<T>> _children;
    private LogicalOperator _operator;

    /**
     * Constructs a new internal node in a tree data structure with the given children 
     * and the given logical operator to link these children.
     * @param children the children of the internal node
     * @param operator the logical operator to link the children
     */
    public TreeInternalNode(List<AbstractTreeNode<T>> children, LogicalOperator operator)
    {
        if (children.size() < 1)
        {
            throw new IllegalArgumentException("There must be at least one child.");
        }
        _children = children;
        _operator = operator;
    }
    
    /**
     * Gets the children
     * @return the children
     */
    public List<AbstractTreeNode<T>> getChildren()
    {
        return _children;
    }
    
    /**
     * Gets the logical operator
     * @return the logical operator
     */
    public LogicalOperator getLogicalOperator()
    {
        return _operator;
    }
}

