/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.advanced.utils;

import java.util.List;
import java.util.function.Function;

import org.ametys.cms.search.advanced.AbstractTreeNode;
import org.ametys.cms.search.advanced.TreeInternalNode;
import org.ametys.cms.search.advanced.TreeLeaf;

/**
 * Helper to {@link #print(AbstractTreeNode) print} a tree for debug purposes.
 */
public final class TreePrinter
{
    private TreePrinter()
    {
        // Nothing
    }
    
    /**
     * Prints the given tree
     * <br>Use for debug purposes only.
     * @param <T> the type of the values of the leaves of the tree.
     * @param tree The tree to print
     * @return The string representing the tree
     */
    public static <T> String print(AbstractTreeNode<T> tree)
    {
        return print(tree, Object::toString);
    }
    
    /**
     * Prints the given tree
     * <br>Use for debug purposes only.
     * @param <T> the type of the values of the leaves of the tree.
     * @param tree The tree to print
     * @param leafStringifier The function to transform a leaf value to a readable String
     * @return The string representing the tree
     */
    public static <T> String print(AbstractTreeNode<T> tree, Function<T, String> leafStringifier)
    {
        StringBuilder sb = new StringBuilder();
        _print(sb, tree, leafStringifier, new StringBuffer(), true);
        return sb.toString();
    }
    
    private static <T> void _print(StringBuilder sb, AbstractTreeNode<T> tree, Function<T, String> leafStringifier, StringBuffer prefix, boolean isTail)
    {
        String name;
        if (tree instanceof TreeLeaf<?>)
        {
            name = leafStringifier.apply(((TreeLeaf<T>) tree).getValue());
        }
        else if (tree instanceof TreeInternalNode<?>)
        {
            name = ((TreeInternalNode<T>) tree).getLogicalOperator().toString();
        }
        else
        {
            throw new IllegalArgumentException("Wrong type of TreeNode. It must only be a TreeInternalNode or a TreeLeaf.");
        }
        sb.append(prefix).append(isTail ? "└── " : "├── ").append(name).append("\n");
        
        if (tree instanceof TreeInternalNode<?>)
        {
            List<AbstractTreeNode<T>> children = ((TreeInternalNode<T>) tree).getChildren();
            StringBuffer newPrefix = new StringBuffer(prefix).append(isTail ? "    " : "│   ");
            int i = 1;
            for (AbstractTreeNode<T> child : children)
            {
                boolean childIsTail = i == children.size();
                _print(sb, child, leafStringifier, newPrefix, childIsTail);
                i++;
            }
        }
    }
}
