/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.cocoon;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.components.ContextHelper;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.content.ContentValuesExtractorFactory;
import org.ametys.cms.search.content.ContentValuesExtractorFactory.ContentValuesExtractor;
import org.ametys.cms.search.content.ContentValuesExtractorFactory.SimpleContentValuesExtractor;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.ui.model.ColumnHelper;
import org.ametys.cms.search.ui.model.SearchUIColumn;
import org.ametys.cms.search.ui.model.SearchUIColumnHelper;
import org.ametys.cms.search.ui.model.impl.RepeaterSearchUIColumn;
import org.ametys.core.ui.Callable;
import org.ametys.core.util.ServerCommHelper;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.model.CompositeDefinition;
import org.ametys.runtime.model.DefinitionContext;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.SimpleViewItemGroup;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewHelper;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;

/**
 * Generates the columns information for the grid based upon a view of a contenttype
 */
public class ContentGridComponent implements Contextualizable, Serviceable, Component
{
    /** The avalon role */
    public static final String ROLE = ContentGridComponent.class.getName();
    
    /** The servercomm helper */
    protected ServerCommHelper _serverCommHelper;
    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /** The contenttypes extension point */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    /** Cocoon context */
    protected Context _context;
    /** The helper for columns */
    protected ColumnHelper _columnHelper;
    /** The service manager */
    protected ServiceManager _manager;
    /** The ContentValuesExtractorFactory */
    protected ContentValuesExtractorFactory _contentValuesExtractorFactory;
    /** The AmetysObjectResolver instance */
    protected AmetysObjectResolver _ametysObjectResolver;
    /** The ContentTypesHelper instance */
    protected ContentTypesHelper _contentTypesHelper;
    /** The ContentHelper instance */
    protected ContentHelper _contentHelper;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
        _contentHelper = (ContentHelper) manager.lookup(ContentHelper.ROLE);
        _serverCommHelper = (ServerCommHelper) manager.lookup(ServerCommHelper.ROLE);
        _columnHelper = (ColumnHelper) manager.lookup(ColumnHelper.ROLE);
        _contentValuesExtractorFactory = (ContentValuesExtractorFactory) manager.lookup(ContentValuesExtractorFactory.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    /**
     * Generates the columns informations for a View of a ContentType
     * @param contentIds The contents identifiers
     * @param contentTypeId The content type id. Mandatory.
     * @param viewName The view of the content type.
     * @return The columns informations
     * @throws IllegalArgumentException If one argument is not as expected
     * @throws IllegalStateException If an error occurred while processing the columns
     * @throws ProcessingException if an error occurs while generating columns infos 
     */
    @Callable
    public Map<String, Object> getColumnsAndValues(List<String> contentIds, String contentTypeId, String viewName) throws IllegalStateException, IllegalArgumentException, ProcessingException
    {
        ContentType contentType = _getContentType(contentTypeId);
        View editionView = _getEditionView(contentType, viewName);
        View copyWithColumns = new View();
        editionView.copyTo(copyWithColumns);
        copyWithColumns.addViewItems(_copyViewItemsAsColumns(editionView));
        
        List<Map<String, Object>> columnsInfo = _getColumnsInfo(copyWithColumns);
        
        Map<String, Object> results = new HashMap<>();
        results.put("columns", columnsInfo);
        results.put("contentType", contentType.getId());
        results.put("view", copyWithColumns.getName());

        SimpleContentValuesExtractor contentValuesExtractor = _contentValuesExtractorFactory.create(copyWithColumns);
        Map objectModel = ContextHelper.getObjectModel(_context);
        Locale defaultLocale = org.apache.cocoon.i18n.I18nUtils.findLocale(objectModel, "locale", null, Locale.getDefault(), true);
        
        List<Map<String, Object>> contents = new ArrayList<>();
        for (String contentId : contentIds)
        {
            Content content = _ametysObjectResolver.resolveById(contentId);
            Map<String, Object> properties = getContentData(content, contentValuesExtractor, defaultLocale, Collections.emptyMap());
            contents.add(properties);
        }
        results.put("contents", contents);
        
        return results;
    }
    
    private List<ViewItem> _copyViewItemsAsColumns(ViewItemAccessor original)
    {
        List<ViewItem> viewItems = new ArrayList<>();
        
        org.ametys.plugins.repository.model.ViewHelper.visitView(original, 
                (element, definition) -> {
                    // simple element
                    if (definition instanceof SystemProperty systemProperty && !systemProperty.isDisplayable())
                    {
                        throw new IllegalArgumentException("The system property '" + systemProperty.getName() + "' is not displayable.");
                    }
                    
                    SearchUIColumn searchUIColumn =  SearchUIColumnHelper.createModelItemColumn(definition);
                    element.copyTo(searchUIColumn);
                    
                    viewItems.add(searchUIColumn);
                },
                (group, definition) -> {
                    // composite
                    ModelViewItemGroup<CompositeDefinition> groupCopy = new ModelViewItemGroup<>();
                    group.copyTo(groupCopy);
                    
                    groupCopy.addViewItems(_copyViewItemsAsColumns(group));
                    
                    viewItems.add(groupCopy);
                },
                (group, definition) -> {
                    // repeater
                    RepeaterSearchUIColumn searchUIColumn =  new RepeaterSearchUIColumn();
                    group.copyTo(searchUIColumn);
                    
                    searchUIColumn.addViewItems(_copyViewItemsAsColumns(group));
                    
                    viewItems.add(searchUIColumn);
                },
                group -> {
                    // simple view item group
                    SimpleViewItemGroup groupCopy = new SimpleViewItemGroup();
                    group.copyTo(groupCopy);
                    
                    groupCopy.addViewItems(_copyViewItemsAsColumns(group));
                    
                    viewItems.add(groupCopy);
                });
        
        return viewItems;
    }
    
    private List<Map<String, Object>> _getColumnsInfo(ViewItemAccessor viewItemAccessor) throws ProcessingException
    {
        List<Map<String, Object>> jsonObject = new ArrayList<>();
        
        for (ViewItem viewItem : viewItemAccessor.getViewItems())
        {
            if (viewItem instanceof SearchUIColumn searchUIColumn)
            {
                Map<String, Object> columnInfo = searchUIColumn.toJSON(DefinitionContext.newInstance());
                columnInfo.put("field", searchUIColumn.getDefinition().getPath());
                jsonObject.add(columnInfo);
            }
            else if (viewItem instanceof ViewItemAccessor itemAccessor)
            {
                jsonObject.addAll(_getColumnsInfo(itemAccessor));
            }
        }
        
        return jsonObject;
    }
    
    private ContentType _getContentType(String contentTypeId) throws IllegalArgumentException
    {
        if (contentTypeId.isBlank())
        {
            throw new IllegalArgumentException("The contentType argument is mandatory");
        }
        
        ContentType contentType = _contentTypeExtensionPoint.getExtension(contentTypeId);
        if (contentType == null)
        {
            throw new IllegalArgumentException("The content type '" + contentTypeId +  "' specified does not exist");
        }
        
        return contentType;
    }
    
    private View _getEditionView(ContentType contentType, String viewName) throws IllegalArgumentException
    {
        View view;
        if (StringUtils.isBlank(viewName))
        {
            view = contentType.getView("default-edition");
            if (view == null)
            {
                view = contentType.getView("main");
                if (view == null)
                {
                    throw new IllegalArgumentException("The content type '" + contentType.getId() + "' has no 'default-edition' nor 'main' view. Specify a view to use");
                }
            }
        }
        else
        {
            view = contentType.getView(viewName);
            if (view == null)
            {
                throw new IllegalArgumentException("The content type '" + contentType.getId() + "' has no '" + viewName + "' view");
            }
        }
        
        return ViewHelper.getTruncatedView(view);
    }
    
    /**
     * Generate standard content data.
     * @param content The content.
     * @param extractor The content values extractor which generates.
     * @param defaultLocale the default locale for localized values if content's language is null. 
     * @param contextualParameters The search contextual parameters.
     * @return A Map containing the content data.
     */
    public Map<String, Object> getContentData(Content content, ContentValuesExtractor extractor, Locale defaultLocale, Map<String, Object> contextualParameters)
    {
        Map<String, Object> contentData = new HashMap<>();
        
        contentData.put("id", content.getId());
        contentData.put("name", content.getName());
        
        if (_contentHelper.isMultilingual(content) && ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID.equals(content.getDefinition("title").getType().getId()))
        {
            contentData.put("title", _contentHelper.getTitleVariants(content));
        }
        else
        {
            contentData.put("title", _contentHelper.getTitle(content));
        }
        
        contentData.put("language", content.getLanguage());
        contentData.put("contentTypes", content.getTypes());
        contentData.put("mixins", content.getMixinTypes());
        contentData.put("iconGlyph", _contentTypesHelper.getIconGlyph(content));
        contentData.put("iconDecorator", _contentTypesHelper.getIconDecorator(content));
        contentData.put("smallIcon", _contentTypesHelper.getSmallIcon(content));
        contentData.put("mediumIcon", _contentTypesHelper.getMediumIcon(content));
        contentData.put("largeIcon", _contentTypesHelper.getLargeIcon(content));
        contentData.put("isSimple", _contentHelper.isSimple(content));
        contentData.put("archived", _contentHelper.isArchivedContent(content));
        
        contentData.put("properties", extractor.getValues(content, defaultLocale, contextualParameters));
        
        return contentData;
    }
}
