/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.cocoon;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.container.ContainerUtil;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.components.source.SourceUtil;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.GroupSearchContentHelper;
import org.ametys.cms.search.SearchResults;
import org.ametys.cms.search.model.SearchModel;
import org.ametys.cms.search.solr.CriteriaSearchModelWrapper;
import org.ametys.cms.search.ui.model.ColumnHelper.Column;
import org.ametys.cms.search.ui.model.SearchUIModel;
import org.ametys.core.util.AvalonLoggerAdapter;
import org.ametys.core.util.IgnoreRootHandler;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.runtime.model.ViewItemContainer;

/**
 * Generate contents returned by the {@link SearchAction}, groups by selected fields.
 */
public class DocSearchGenerator extends SearchGenerator
{
    private GroupSearchContentHelper _groupSearchContentHelper;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _groupSearchContentHelper = (GroupSearchContentHelper) serviceManager.lookup(GroupSearchContentHelper.ROLE);
    }
    
    @Override
    protected void saxContents(SearchModel model, SearchResults<Content> results, String versionLabel, Map<String, Object> jsParameters, Locale locale, Map<String, Object> contextualParameters) throws SAXException, AmetysRepositoryException, IOException, ProcessingException
    {
        String modelId = (String) jsParameters.get("model");
        SearchUIModel uiModel = _searchModelManager.getExtension(modelId);
        
        Set<String> cTypeIds = getCommonContentTypeIds(jsParameters, uiModel, contextualParameters);
        
        List<String> groupingFields = _getGroupingFields(jsParameters);

        // Columns
        List<Column> columns = getColumnsFromParameters(jsParameters, cTypeIds);
        List<String> columnIds = columns.stream()
                                        .map(Column::getId)
                                        .collect(Collectors.toList());

        ViewItemContainer resultItems = model.getResultItems(contextualParameters);
        ViewItemContainer filteredResultItems = (ViewItemContainer) SearchGeneratorHelper.copyAndFilterViewItemAccessor(resultItems, columnIds);
        SearchGeneratorHelper.removeResultItems(filteredResultItems, groupingFields);
        
        CriteriaSearchModelWrapper modelWrapper = new CriteriaSearchModelWrapper(uiModel, manager, _context, new AvalonLoggerAdapter(getLogger()));
        
        try
        {
            ContainerUtil.service(modelWrapper, manager);
            modelWrapper.setResultColumns(cTypeIds, columns, contextualParameters);
        }
        catch (Exception e)
        {
            throw new AmetysRepositoryException(e);
        }
        ViewItemContainer uiResultItems = modelWrapper.getResultItems(contextualParameters);
        
        Set<ContentType> contentTypes = cTypeIds.stream()
                .map(_contentTypeExtensionPoint::getExtension)
                .collect(Collectors.toSet());
        GroupSearchContent groupSearchContent = _groupSearchContentHelper.organizeContentsInGroups(results.getObjects(), groupingFields, contentTypes, locale);
        
        // Set the request attributes from the contextual parameters
        Request request = ContextHelper.getRequest(_context);
        for (String name : contextualParameters.keySet())
        {
            request.setAttribute(name, contextualParameters.get(name));
        }
        
        XMLUtils.startElement(contentHandler, "contents");
        _saxGroup(groupSearchContent, 0, filteredResultItems, uiResultItems, locale);
        XMLUtils.endElement(contentHandler, "contents");
    }
    
    private List<String> _getGroupingFields(Map<String, Object> jsParameters)
    {
        @SuppressWarnings("unchecked")
        List<String> groupingFields = (List<String>) jsParameters.get("groupingFields");
        if (groupingFields == null || groupingFields.size() == 1 && StringUtils.isEmpty(groupingFields.get(0)))
        {
            groupingFields = new ArrayList<>();
        }
        
        return groupingFields;
    }
    
    /**
     * Sax a group of content recursively.
     * First level (level=0) will not create a xml group (usually, 1st group is just here to handle a list)
     * @param group group to sax
     * @param level current level (start with zero)
     * @param resultItems result items to sax
     * @param uiResultItems The columns to be exported
     * @param locale The locale for search. Can be null.
     * @throws SAXException if a error occurred during sax
     * @throws AmetysRepositoryException if a error occurred
     * @throws IOException if a error occurred
     * @throws ProcessingException if a error occurred
     */
    private void _saxGroup(GroupSearchContent group, int level, ViewItemContainer resultItems, ViewItemContainer uiResultItems, Locale locale) throws SAXException, AmetysRepositoryException, IOException, ProcessingException
    {
        if (level > 0)
        {
            AttributesImpl attrs = new AttributesImpl();
            attrs.addCDATAAttribute("level", String.valueOf(level));
            attrs.addCDATAAttribute("fullModelPath", group.getGroupFieldPath());
            attrs.addCDATAAttribute("value", group.getGroupName());
            XMLUtils.startElement(contentHandler, "group", attrs);
        }
        
        for (GroupSearchContent groupSearchContent : group.getSubList())
        {
            _saxGroup(groupSearchContent, level + 1, resultItems, uiResultItems, locale);
        }
        
        for (Content content : group.getContents())
        {
            _saxContentForDoc(content, resultItems, uiResultItems, locale);
        }
        
        if (level > 0)
        {
            XMLUtils.endElement(contentHandler, "group");
        }
    }
    /**
     * SAX the result content
     * @param content the result
     * @param resultItems the result items
     * @param uiResultItems The columns to be exported
     * @param locale The locale for search. Can be null.
     * @throws SAXException if a error occurred during sax
     * @throws AmetysRepositoryException if a error occurred
     * @throws IOException if a error occurred
     * @throws ProcessingException if a error occurred
     */
    private void _saxContentForDoc(Content content, ViewItemContainer resultItems, ViewItemContainer uiResultItems, Locale locale) throws SAXException, AmetysRepositoryException, IOException, ProcessingException
    {
        Request request = ContextHelper.getRequest(_context);
        
        try
        {
            request.setAttribute(Content.class.getName(), content);
            request.setAttribute("resultItems", resultItems);
            request.setAttribute("viewName", "export-word");
            request.setAttribute("forceRemoteUrl", true);
            
            request.setAttribute("uiResultItems", uiResultItems);

            AttributesImpl attrs = new AttributesImpl();
            attrs.addCDATAAttribute("id", content.getId());
            attrs.addCDATAAttribute("name", content.getName());
            attrs.addCDATAAttribute("title", content.getTitle(locale));
            if (content.getLanguage() != null)
            {
                attrs.addCDATAAttribute("language", content.getLanguage());
            }
                
            XMLUtils.startElement(contentHandler, "content", attrs);
            Source contentSource = resolver.resolveURI("cocoon:/export/content.doc" + (locale != null ? "?lang=" + locale.getLanguage() : ""));
            SourceUtil.toSAX(contentSource, new IgnoreRootHandler(contentHandler));
            
            XMLUtils.endElement(contentHandler, "content");
        }
        finally
        {
            request.setAttribute(Content.class.getName(), null);
            request.removeAttribute("forceRemoteUrl");
        }
    }
    
}
