/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.cocoon;

import java.io.IOException;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.generation.ServiceableGenerator;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.search.model.SearchModel;
import org.ametys.cms.search.solr.SolrContentQueryHelper;
import org.ametys.cms.search.ui.model.ColumnHelper;
import org.ametys.cms.search.ui.model.ColumnHelper.Column;
import org.ametys.cms.search.ui.model.SearchUIModelExtensionPoint;
import org.ametys.core.util.ServerCommHelper;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.ViewHelper;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.ViewItemContainer;

/**
 * SAX the columns of a search tool model.
 */
public class ModelColumnsGenerator extends ServiceableGenerator
{
    /** The extenstion point for search UI */
    protected SearchUIModelExtensionPoint _searchModelManager;
    /** The servercomm helper */
    protected ServerCommHelper _serverCommHelper;
    /** The content type helper. */
    protected ContentTypesHelper _cTypeHelper;
    /** The helper for columns */
    protected ColumnHelper _columnHelper;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _searchModelManager = (SearchUIModelExtensionPoint) smanager.lookup(SearchUIModelExtensionPoint.ROLE);
        _serverCommHelper = (ServerCommHelper) manager.lookup(ServerCommHelper.ROLE);
        _cTypeHelper = (ContentTypesHelper) smanager.lookup(ContentTypesHelper.ROLE);
        _columnHelper = (ColumnHelper) smanager.lookup(ColumnHelper.ROLE);
    }
    
    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        Map<String, Object> jsParameters = _serverCommHelper.getJsParameters();
        
        contentHandler.startDocument();
        saxColumns(jsParameters);
        contentHandler.endDocument();
    }
    
    /**
     * Get search model.
     * @param jsParameters JS parameters
     * @return the search model
     */
    protected SearchModel getSearchModel(Map<String, Object> jsParameters)
    {
        String modelId = (String) jsParameters.get("model");
        return _searchModelManager.getExtension(modelId);
    }
    
    /**
     * Sax search columns.
     * @param jsParameters JS parameters
     * @throws SAXException if an error occurs
     */
    protected void saxColumns(Map<String, Object> jsParameters) throws SAXException
    {
        @SuppressWarnings("unchecked")
        Map<String, Object> contextualParameters = (Map<String, Object>) jsParameters.get("contextualParameters");
        
        SearchModel model = getSearchModel(jsParameters);
        List<Column> columns = _getColumns(model, jsParameters, contextualParameters);
        ViewItemContainer resultItems = model.getResultItems(contextualParameters);

        XMLUtils.startElement(contentHandler, "columns");
        _saxColumns(resultItems, columns);
        XMLUtils.endElement(contentHandler, "columns");
    }
    
    private List<Column> _getColumns(SearchModel model, Map<String, Object> jsParameters, Map<String, Object> contextualParameters)
    {
        Set<String> cTypes;
        if ("search-ui.solr".equals(jsParameters.get("model")))
        {
            cTypes = SolrContentQueryHelper.getContentTypes(jsParameters);
        }
        else
        {
            cTypes = model.getContentTypes(contextualParameters);
        }
        Set<String> cTypeIds = _cTypeHelper.getCommonAncestors(cTypes);
        List<Column> columns = getColumnsFromParameters(jsParameters, cTypeIds);
        return columns;
    }
    
    private void _saxColumns(ViewItemAccessor viewItemAccessor, List<Column> columns) throws SAXException
    {
        for (ViewItem viewItem : viewItemAccessor.getViewItems())
        {
            if (viewItem instanceof ViewItemAccessor itemAccessor && !itemAccessor.getViewItems().isEmpty())
            {
                _saxColumns(itemAccessor, columns);
            }
            else if (viewItem instanceof ModelViewItem modelViewItem)
            {
                String modelViewItemPath = ViewHelper.getModelViewItemPath(modelViewItem);
                Optional<Column> column = columns.stream()
                                                 .filter(c -> c.getId().equals(modelViewItemPath))
                                                 .findFirst();
                
                if (columns.isEmpty() || column.isPresent())
                {
                    AttributesImpl attrs = new AttributesImpl();
                    attrs.addCDATAAttribute("id", ViewHelper.getModelViewItemPath(modelViewItem));
                    attrs.addCDATAAttribute("type", modelViewItem.getDefinition().getType().getId());
                    
                    XMLUtils.startElement(contentHandler, "column", attrs);
                    
                    column.map(Column::getLabel)
                          .filter(Optional::isPresent)
                          .map(Optional::get)
                          .map(I18nizableText::new)
                          .orElseGet(modelViewItem::getLabel)
                          .toSAX(contentHandler);
                    
                    XMLUtils.endElement(contentHandler, "column");
                }
            }
        }
    }
    
    /**
     * Get the columns from JS parameters
     * @param jsParameters The JS parameters
     * @param contentTypes The content type
     * @return the requested columns
     */
    @SuppressWarnings("unchecked")
    protected List<Column> getColumnsFromParameters(Map<String, Object> jsParameters, Set<String> contentTypes)
    {
        Map<String, Object> values = (Map<String, Object>) jsParameters.get("values");

        if (values != null && values.containsKey("columns"))
        {
            Object columnsAsObj = values.get("columns");
            if (columnsAsObj instanceof String)
            {
                return _columnHelper.getColumns((String) columnsAsObj, contentTypes);
            }
            else
            {
                return _columnHelper.getColumns((List<String>) columnsAsObj, contentTypes);
            }
        }
        return Collections.EMPTY_LIST;
    }
}
