/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.cocoon;

import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.GetQueryFromJSONHelper;
import org.ametys.cms.search.QueryBuilder;
import org.ametys.cms.search.SearchResults;
import org.ametys.cms.search.Sort;
import org.ametys.cms.search.content.ContentSearcherFactory;
import org.ametys.cms.search.model.SearchCriterion;
import org.ametys.cms.search.model.SearchModel;
import org.ametys.cms.search.query.QuerySyntaxException;
import org.ametys.cms.search.ui.model.SearchUIModel;
import org.ametys.cms.search.ui.model.SearchUIModelExtensionPoint;
import org.ametys.core.util.ServerCommHelper;
import org.ametys.plugins.repository.provider.RequestAttributeWorkspaceSelector;

/**
 * Search contents and put a result object in the request (to be serialized in JSON).
 */
public class SearchAction extends ServiceableAction implements Contextualizable
{
    /** Name of request attribute for storing contents' ids */
    public static final String SEARCH_CONTENTS = SearchAction.class.getName() + "$contentIds";
    
    /** Name of request attribute for storing search results */
    public static final String SEARCH_RESULTS = SearchAction.class.getName() + "$searchResults";
    
    /** Name of request attribute for storing the language of search */
    public static final String SEARCH_LOCALE = SearchAction.class.getName() + "$searchLocale";
    
    /** Name of request attribute for storing the search model */
    public static final String SEARCH_MODEL = SearchAction.class.getName() + "$searchModel";
    
    /** Name of request attribute for storing the query error, if any. */
    public static final String QUERY_ERROR = SearchAction.class.getName() + "$queryError";
    
    /** The search model manager */
    protected SearchUIModelExtensionPoint _searchModelManager;
    /** The ContentType Manager*/
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    /** The server comm helper */
    protected ServerCommHelper _serverCommHelper;
    /** The avalon context */
    protected Context _context;
    /** The searcher factory. */
    protected ContentSearcherFactory _searcherFactory;
    /** the helper to get query infos from JSON */
    protected GetQueryFromJSONHelper _getQueryFromJSONHelper;
    /** The query builder */
    protected QueryBuilder _queryBuilder;
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _searchModelManager = (SearchUIModelExtensionPoint) smanager.lookup(SearchUIModelExtensionPoint.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
        _serverCommHelper = (ServerCommHelper) smanager.lookup(ServerCommHelper.ROLE);
        _searcherFactory = (ContentSearcherFactory) smanager.lookup(ContentSearcherFactory.ROLE);
        _getQueryFromJSONHelper = (GetQueryFromJSONHelper) smanager.lookup(GetQueryFromJSONHelper.ROLE);
        _queryBuilder = (QueryBuilder) smanager.lookup(QueryBuilder.ROLE);
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        Map<String, Object> jsParameters = _serverCommHelper.getJsParameters();
        
        Map<String, Object> contextualParameters = (Map<String, Object>) jsParameters.get("contextualParameters");
        if (contextualParameters == null)
        {
            contextualParameters = Collections.emptyMap();
        }
        
        SearchModel model = getSearchModel(jsParameters, contextualParameters);
        String workspaceName = parameters.getParameter("workspace", model.getWorkspace(contextualParameters));
        
        int begin = getOffset(jsParameters);
        int maxResults = getMaxResults(model, jsParameters, contextualParameters);
        
        // TODO Remove or replace by a custom search criterion AND contextual parameter.
        // If true the contents marked as 'subContent' will be excluded
        // boolean excludeSubContents = jsParameters.containsKey("excludeSubContents") ? (Boolean) jsParameters.get("excludeSubContents") : false;
        
        String originalWorkspace = RequestAttributeWorkspaceSelector.getForcedWorkspace(request);
        
        try
        {
            if (StringUtils.isNotEmpty(workspaceName))
            {
                RequestAttributeWorkspaceSelector.setForcedWorkspace(request, workspaceName);
            }
            
            doSearch(request, model, begin, maxResults, jsParameters, contextualParameters);
            
            if (request.getAttribute(SEARCH_MODEL) == null)
            {
                request.setAttribute(SEARCH_MODEL, model);
            }
        }
        catch (QuerySyntaxException e)
        {
            // Query syntax error: catch the error without logging or rethrowing it.
            request.setAttribute(QUERY_ERROR, e.getI18nMessage());
        }
        catch (Exception e)
        {
            getLogger().error("Cannot search for contents : " + e.getMessage(), e);
            throw new ProcessingException("Cannot search for contents : " + e.getMessage(), e);
        }
        finally
        {
            if (StringUtils.isNotEmpty(workspaceName))
            {
                RequestAttributeWorkspaceSelector.setForcedWorkspace(request, originalWorkspace);
            }
        }
        
        return EMPTY_MAP;
    }
    
    /**
     * Do the search and set the results in request attributes.
     * @param request The request. The results or contents' id have to be set in request attributes
     * @param model The search model
     * @param offset The index of search
     * @param maxResults The max results
     * @param jsParameters The JS parameters
     * @param contextualParameters The contextual parameters
     * @throws Exception if the search failed 
     */
    @SuppressWarnings("unchecked")
    protected void doSearch (Request request, SearchModel model, int offset, int maxResults, Map<String, Object> jsParameters, Map<String, Object> contextualParameters) throws Exception
    {
        try
        {
            Map<String, Object> values = (Map<String, Object>) jsParameters.get("values");
            if (values == null)
            {
                values = Collections.emptyMap();
            }
            Map<String, List<String>> facetValues = (Map<String, List<String>>) jsParameters.get("facetValues");
            if (facetValues == null)
            {
                facetValues = Collections.emptyMap();
            }
            String searchMode = StringUtils.defaultString((String) jsParameters.get("searchMode"), "simple");
            
            String sortInfo = (String) jsParameters.get("sort");
            String groupInfo = (String) jsParameters.get("group");
            List<Sort> sort = getSort(sortInfo, groupInfo);
            
            Map<String, ? extends SearchCriterion> criteria = searchMode.equals("advanced") && model instanceof SearchUIModel uiModel ? uiModel.getAdvancedCriteria(contextualParameters) : model.getCriteria(contextualParameters);
            String lang = _queryBuilder.getCriteriaLanguage(criteria, searchMode, values, contextualParameters);
            if (StringUtils.isNotEmpty(lang))
            {
                request.setAttribute(SEARCH_LOCALE, new Locale(lang));
            }
            
            SearchResults<Content> results = _searcherFactory.create(model)
                                                             .withSearchMode(searchMode)
                                                             .withSort(sort)
                                                             .withLimits(offset, maxResults)
                                                             .searchWithFacets(values, facetValues, contextualParameters);
            
            
            request.setAttribute(SEARCH_RESULTS, results);
        }
        catch (QuerySyntaxException e)
        {
            // Query syntax error: catch the error without logging or rethrowing it.
            request.setAttribute(QUERY_ERROR, e.getI18nMessage());
        }
    }
    
    /**
     * Get the search UI model
     * @param jsParameters The JS parameters
     * @param contextualParameters The contextual parameters
     * @return the search UI model
     */
    protected SearchModel getSearchModel(Map<String, Object> jsParameters, Map<String, Object> contextualParameters)
    {
        return _getQueryFromJSONHelper.getSearchModel(jsParameters);
    }
    
    /**
     * Get the index of search
     * @param jsParameters The JS parameters
     * @return The offset
     */
    protected int getOffset(Map<String, Object> jsParameters)
    {
        return jsParameters.containsKey("start") ? (Integer) jsParameters.get("start") : 0; // Index of search
    }
    
    /**
     * Get the max number of results
     * @param model The search model
     * @param jsParameters The JS parameters
     * @param contextualParameters The contextual parameters
     * @return The max number of results
     */
    protected int getMaxResults(SearchModel model, Map<String, Object> jsParameters, Map<String, Object> contextualParameters)
    {
        int modelPageSize = model instanceof SearchUIModel uiModel ? uiModel.getPageSize(contextualParameters) : -1;
        int maxResults = Integer.MAX_VALUE; // Number of results to generate
        if (jsParameters.containsKey("limit"))
        {
            maxResults = (Integer) jsParameters.get("limit");
        }
        else if (modelPageSize >= 0)
        {
            maxResults = modelPageSize;
        }
        
        if (maxResults < 0)
        {
            maxResults = Integer.MAX_VALUE;
        }
        
        return maxResults;
    }
    
    /**
     * Get the sort criteria from a sort string.
     * @param sortString The sort criteria as a JSON-encoded string.
     * @param groupString The group criteria as a JSON-encoded string (for server-side grouping feature). Can be null.
     * @return the sort criteria as a List of {@link Sort}.
     */
    protected List<Sort> getSort(String sortString, String groupString)
    {
        return _getQueryFromJSONHelper.getSort(sortString, groupString);
    }
}
