/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.cocoon;

import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.generation.ServiceableGenerator;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.SearchResult;
import org.ametys.cms.search.SearchResults;
import org.ametys.cms.search.model.SearchModel;
import org.ametys.cms.search.ui.model.ColumnHelper;
import org.ametys.cms.search.ui.model.ColumnHelper.Column;
import org.ametys.cms.search.ui.model.SearchUIModelExtensionPoint;
import org.ametys.core.util.ServerCommHelper;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.version.VersionAwareAmetysObject;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewHelper;
import org.ametys.runtime.model.ViewItemContainer;
import org.ametys.runtime.model.type.DataContext;

/**
 * Generate contents returned by the {@link SearchAction}.
 */
public class SearchGenerator extends ServiceableGenerator implements Contextualizable
{
    /** Constant for getting content in specific version label */
    public static final String CONTENT_VERSION_LABEL = "versionLabel";
    
    /** The server comm helper */
    protected ServerCommHelper _serverCommHelper;
    /** Context */
    protected Context _context;
    /** The search model manager */
    protected SearchUIModelExtensionPoint _searchModelManager;
    /** The content type helper. */
    protected ContentTypesHelper _cTypeHelper;
    /** The helper for columns */
    protected ColumnHelper _columnHelper;
    /** The content type extension point */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _serverCommHelper = (ServerCommHelper) smanager.lookup(ServerCommHelper.ROLE);
        
        _searchModelManager = (SearchUIModelExtensionPoint) smanager.lookup(SearchUIModelExtensionPoint.ROLE);
        
        _cTypeHelper = (ContentTypesHelper) smanager.lookup(ContentTypesHelper.ROLE);
        _columnHelper = (ColumnHelper) smanager.lookup(ColumnHelper.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        SearchResults<Content> results = (SearchResults<Content>) request.getAttribute(SearchAction.SEARCH_RESULTS);
        SearchModel model = (SearchModel) request.getAttribute(SearchAction.SEARCH_MODEL);
        
        Map<String, Object> jsParameters = _serverCommHelper.getJsParameters();
        
        Map<String, Object> contextualParameters = (Map<String, Object>) jsParameters.get("contextualParameters");
        if (contextualParameters == null)
        {
            contextualParameters = Collections.emptyMap();
        }
        
        Locale defaultLocale = (Locale) request.getAttribute(SearchAction.SEARCH_LOCALE);
        String versionLabel = (String) jsParameters.get(CONTENT_VERSION_LABEL);

        contentHandler.startDocument();
        saxContents(model, results, versionLabel, jsParameters, defaultLocale, contextualParameters);
        contentHandler.endDocument();
    }
    
    /**
     * Sax a set of contents
     * @param model search model
     * @param results contents
     * @param versionLabel version label
     * @param jsParameters parameters of the search
     * @param defaultLocale The locale to use for localized values if content's language is null.
     * @param contextualParameters The contextual parameters
     * @throws SAXException if a error occurred during sax
     * @throws AmetysRepositoryException if a error occurred
     * @throws IOException if a error occurred
     * @throws ProcessingException if a error occurred
     */
    protected void saxContents(SearchModel model, SearchResults<Content> results, String versionLabel, Map<String, Object> jsParameters, Locale defaultLocale, Map<String, Object> contextualParameters) throws SAXException, AmetysRepositoryException, IOException, ProcessingException
    {
        ViewItemContainer viewItems = getViewItems(model, jsParameters, contextualParameters);
        
        XMLUtils.startElement(contentHandler, "contents");
        for (SearchResult<Content> result : results.getResults())
        {
            Content content = result.getObject();
            
            if (StringUtils.isBlank(versionLabel) || switchToLabel(content, versionLabel))
            {
                saxContent(content, viewItems, defaultLocale);
            }
        }
        XMLUtils.endElement(contentHandler, "contents");
    }
    
    /**
     * Get the filtered view items to sax the content results.
     * @param model search model
     * @param jsParameters parameters of the search
     * @param contextualParameters The contextual parameters
     * @return a {@link ViewItemContainer}
     */
    protected ViewItemContainer getViewItems(SearchModel model, Map<String, Object> jsParameters, Map<String, Object> contextualParameters)
    {
        // Common content type
        Set<String> commonContentTypeIds = getCommonContentTypeIds(jsParameters, model, contextualParameters);
        
        // Filter result items to keep only requested columns
        List<String> columns = getColumnsFromParameters(jsParameters, commonContentTypeIds).stream()
                                                                                           .map(Column::getId)
                                                                                           .collect(Collectors.toList());
        View resultItems = (View) model.getResultItems(contextualParameters);
        View filteredResultItems = (View) SearchGeneratorHelper.copyAndFilterViewItemAccessor(resultItems, columns);
        
        return filteredResultItems.getViewItems().isEmpty()
                ? ViewHelper.mergeDuplicatedItems(resultItems)
                : ViewHelper.mergeDuplicatedItems(filteredResultItems);
    }
    
    /**
     * Get the common content type
     * @param jsParameters The JS parameters
     * @param model The search model
     * @param contextualParameters The contextual parameters
     * @return The common content type
     */
    @SuppressWarnings("unchecked")
    protected Set<String> getCommonContentTypeIds(Map<String, Object> jsParameters, SearchModel model, Map<String, Object> contextualParameters)
    {
        // Get content types from model...
        Collection<String> contentTypes = model.getContentTypes(contextualParameters);
        
        // ...or from the JS parameters
        Map<String, Object> values = (Map<String, Object>) jsParameters.get("values");
        if (values != null && values.containsKey("contentTypes"))
        {
            contentTypes = (List<String>) values.get("contentTypes");
            
        }
        
        // Retrieves the common ancestors of the content types
        return _cTypeHelper.getCommonAncestors(contentTypes).stream()
                .filter(_contentTypeExtensionPoint::hasExtension)
                .collect(Collectors.toSet());
    }
    
    /**
     * Get the columns from JS parameters
     * @param jsParameters The JS parameters
     * @param contentTypeIds The content type identifiers
     * @return the requested columns
     */
    @SuppressWarnings("unchecked")
    protected List<Column> getColumnsFromParameters(Map<String, Object> jsParameters, Set<String> contentTypeIds)
    {
        Map<String, Object> values = (Map<String, Object>) jsParameters.get("values");
        
        if (values != null && values.containsKey("columns"))
        {
            Object columnsAsObj = values.get("columns");
            if (columnsAsObj instanceof String)
            {
                return _columnHelper.getColumns((String) columnsAsObj, contentTypeIds);
            }
            else
            {
                return _columnHelper.getColumns((List<String>) columnsAsObj, contentTypeIds);
            }
        }
        
        return List.of();
    }
    
    /**
     * Switch to the revision corresponding to the specified label.
     * @param content The content
     * @param label the label to switch to
     * @return <code>true</code> if a revision with this label exists and the content was switch, <code>false</code> otherwise.
     */
    protected boolean switchToLabel(Content content, String label)
    {
        if (content instanceof VersionAwareAmetysObject)
        {
            String[] allLabels = ((VersionAwareAmetysObject) content).getAllLabels();
            if (ArrayUtils.contains(allLabels, label))
            {
                ((VersionAwareAmetysObject) content).switchToLabel(label);
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * SAX the result content
     * @param content the result
     * @param resultItems the result fields
     * @param defaultLocale The locale to use for localized values if content's language is null.
     * @throws SAXException if a error occurred during sax
     * @throws AmetysRepositoryException if a error occurred
     * @throws IOException if a error occurred
     */
    protected void saxContent(Content content, ViewItemContainer resultItems, Locale defaultLocale) throws SAXException, AmetysRepositoryException, IOException
    {
        Request request = ContextHelper.getRequest(_context);
        
        try
        {
            request.setAttribute(Content.class.getName(), content);
            
            AttributesImpl attrs = new AttributesImpl();
            attrs.addCDATAAttribute("id", content.getId());
            attrs.addCDATAAttribute("name", content.getName());
            attrs.addCDATAAttribute("title", content.getTitle(defaultLocale));
            if (content.getLanguage() != null)
            {
                attrs.addCDATAAttribute("language", content.getLanguage());
            }
                
            XMLUtils.startElement(contentHandler, "content", attrs);
            
            DataContext context = DataContext.newInstance()
                                             .withLocale(defaultLocale)
                                             .withEmptyValues(false);
            content.dataToSAX(contentHandler, resultItems, context);
            
            XMLUtils.endElement(contentHandler, "content");
        }
        finally
        {
            request.setAttribute(Content.class.getName(), null);
        }
    }
}
