/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.cocoon;

import java.util.List;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;

import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.SimpleViewItemGroup;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewHelper;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;

/**
 * Helper for search generator
 */
public final class SearchGeneratorHelper
{
    private SearchGeneratorHelper()
    {
        // Empty constructor
    }
    
    /**
     * Copy the items of the given accessor, keeping only the items with the given paths
     * @param viewItemAccessor the view item accessor to copy
     * @param itemPathsToKeep the paths of the items to keep. If the list is empty, copy the accessor and keep all items
     * @return the view items accessor's copy
     */
    public static ViewItemAccessor copyAndFilterViewItemAccessor(ViewItemAccessor viewItemAccessor, List<String> itemPathsToKeep)
    {
        // Copy the given container, but not its children
        ViewItemAccessor copy;
        if (viewItemAccessor instanceof ViewItem viewItem)
        {
            copy = (ViewItemAccessor) viewItem.createInstance();
            viewItem.copyTo((ViewItem) copy);
        }
        else
        {
            copy = new View();
            ((View) viewItemAccessor).copyTo((View) copy);
        }
        
        for (ViewItem viewItem : viewItemAccessor.getViewItems())
        {
            Optional<ViewItem> childCopy = Optional.empty();
            if (viewItem instanceof ModelViewItem modelViewItem)
            {
                if (_shouldKeepModelViewItem(modelViewItem, itemPathsToKeep))
                {
                    if (viewItem instanceof ViewItemAccessor childAccessor)
                    {
                        childCopy = Optional.of((ViewItem) copyAndFilterViewItemAccessor(childAccessor, itemPathsToKeep));
                    }
                    else
                    {
                        childCopy = Optional.of(viewItem.createInstance());
                        viewItem.copyTo(childCopy.get());
                    }
                }
            }
            else if (viewItem instanceof SimpleViewItemGroup group)
            {
                childCopy = Optional.of((ViewItem) copyAndFilterViewItemAccessor(group, itemPathsToKeep));
            }
            
            childCopy.ifPresent(copy::addViewItem);
        }
        
        return copy;
    }
    
    private static boolean _shouldKeepModelViewItem(ModelViewItem modelViewItem, List<String> itemsToKeep)
    {
        if (itemsToKeep.isEmpty())
        {
            // If no items are specified, keep them all
            return true;
        }
        
        String modelViewItemPath = ViewHelper.getModelViewItemPath(modelViewItem);
        for (String itemToKeep : itemsToKeep)
        {
            if (StringUtils.startsWith(itemToKeep, modelViewItemPath))
            {
                if (modelViewItem instanceof ViewItemAccessor viewItemAccessor)
                {
                    String subItemPath = StringUtils.substring(itemToKeep, modelViewItemPath.length() + 1);
                    
                    if (subItemPath.isEmpty() && viewItemAccessor.getViewItems().isEmpty())
                    {
                        return true;
                    }
                    else if (ViewHelper.hasModelViewItem(viewItemAccessor, subItemPath))
                    {
                        ModelViewItem subItem = ViewHelper.getModelViewItem(viewItemAccessor, subItemPath);
                        if (subItem instanceof ViewItemAccessor subViewItemAccessor && subViewItemAccessor.getViewItems().isEmpty()
                            || !(subItem instanceof ViewItemAccessor))
                        {
                            return true;
                        }
                    }
                }
                else
                {
                    return true;
                }
            }
        }
        
        return false;
    }
    
    /**
     * Removes the items with the given paths from the given accessor
     * @param viewItemAccessor the view item accessor
     * @param itemPathsToRemove the paths of the items to remove
     */
    public static void removeResultItems(ViewItemAccessor viewItemAccessor, List<String> itemPathsToRemove)
    {
        for (String itemPathToRemove : itemPathsToRemove)
        {
            Optional<ModelViewItem> modelViewItem = _getModelViewItem(viewItemAccessor, itemPathToRemove);
            if (modelViewItem.isPresent())
            {
                ViewItem itemToRemove = modelViewItem.get();
                ViewItemAccessor parent = itemToRemove.getParent();
                while (parent != null)
                {
                    parent.removeViewItem(itemToRemove);
                    
                    if (parent instanceof ViewItem parentViewItem &&  parent.getViewItems().isEmpty())
                    {
                        parent = parentViewItem.getParent();
                        itemToRemove = parentViewItem;
                    }
                    else
                    {
                        parent = null;
                    }
                }
            }
        }
    }
    
    private static Optional<ModelViewItem> _getModelViewItem(ViewItemAccessor viewItemAccessor, String itemPath) throws IllegalArgumentException
    {
        String[] pathSegments = StringUtils.split(itemPath, ModelItem.ITEM_PATH_SEPARATOR);
        
        if (pathSegments == null || pathSegments.length < 1)
        {
            throw new IllegalArgumentException("Unable to retrieve the data at the given path. This path is empty.");
        }
        else if (pathSegments.length == 1)
        {
            for (ViewItem viewItem : viewItemAccessor.getViewItems())
            {
                if (viewItem instanceof ModelViewItem modelViewItem
                        && _isSearchedModelViewItem(modelViewItem, itemPath))
                {
                    return Optional.of(modelViewItem);
                }
                else if (viewItem instanceof SimpleViewItemGroup group)
                {
                    Optional<ModelViewItem> modelViewItem = _getModelViewItem(group, itemPath);
                    if (modelViewItem.isPresent())
                    {
                        return modelViewItem;
                    }
                }
            }
        }
        else
        {
            String fisrtSegment = pathSegments[0];
            String subPath = StringUtils.join(pathSegments, ModelItem.ITEM_PATH_SEPARATOR, 1, pathSegments.length);
            for (ViewItem viewItem : viewItemAccessor.getViewItems())
            {
                if (viewItem instanceof ViewItemAccessor childAccessor)
                {
                    Optional<ModelViewItem> modelViewItem = Optional.empty();
                    if (viewItem instanceof SimpleViewItemGroup)
                    {
                        modelViewItem = _getModelViewItem(childAccessor, itemPath);
                    }
                    else if (viewItem.getName().equals(fisrtSegment))
                    {
                        modelViewItem = _getModelViewItem(childAccessor, subPath);
                    }
                    
                    if (modelViewItem.isPresent())
                    {
                        return modelViewItem;
                    }
                }
            }
        }
        
        // no model view item has been found with this path
        return Optional.empty();
    }
    
    private static boolean _isSearchedModelViewItem(ModelViewItem modelViewItem, String itemName)
    {
        return modelViewItem.getName().equals(itemName)
                && (!(modelViewItem instanceof ViewItemAccessor)
                        || modelViewItem instanceof ViewItemAccessor childAccessor && childAccessor.getViewItems().isEmpty());
    }
}
