/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.cocoon;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.reading.ServiceableReader;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.SearchResult;
import org.ametys.cms.search.SearchResults;
import org.ametys.cms.search.content.ContentValuesExtractorFactory;
import org.ametys.cms.search.content.ContentValuesExtractorFactory.ContentValuesExtractor;
import org.ametys.cms.search.content.ContentValuesExtractorFactory.SearchModelContentValuesExtractor;
import org.ametys.cms.search.model.SearchCriterion;
import org.ametys.cms.search.model.SearchCriterionHelper;
import org.ametys.cms.search.model.SearchModel;
import org.ametys.core.util.JSONUtils;
import org.ametys.core.util.ServerCommHelper;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.plugins.repository.provider.RequestAttributeWorkspaceSelector;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * JSON reader for search for contents.
 */
public class SearchJsonReader extends ServiceableReader
{
    /** JSON utils */
    protected JSONUtils _jsonUtils;
    
    /** The serverComm helper */
    protected ServerCommHelper _serverCommHelper;
    
    /** The content types helper. */
    protected ContentTypesHelper _contentTypesHelper;
    
    /** The content helper */
    protected ContentHelper _contentHelper;
    
    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    
    /** The ContentGridComponent instance */
    protected ContentGridComponent _contentGridComponent;
    
    /** The search criterion helper */
    protected SearchCriterionHelper _searchCriterionHelper;
    
    private ContentValuesExtractorFactory _valuesExtractorFactory;


    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        
        _jsonUtils = (JSONUtils) serviceManager.lookup(JSONUtils.ROLE);
        _serverCommHelper = (ServerCommHelper) serviceManager.lookup(ServerCommHelper.ROLE);
        _contentTypesHelper = (ContentTypesHelper) serviceManager.lookup(ContentTypesHelper.ROLE);
        _contentHelper = (ContentHelper) serviceManager.lookup(ContentHelper.ROLE);
        _valuesExtractorFactory = (ContentValuesExtractorFactory) serviceManager.lookup(ContentValuesExtractorFactory.ROLE);
        _resolver = (AmetysObjectResolver) serviceManager.lookup(AmetysObjectResolver.ROLE);
        _contentGridComponent = (ContentGridComponent) serviceManager.lookup(ContentGridComponent.ROLE);
        _searchCriterionHelper = (SearchCriterionHelper) serviceManager.lookup(SearchCriterionHelper.ROLE);
    }
    
    @Override
    public String getMimeType()
    {
        return "application/json";
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        String workspaceName = parameters.getParameter("workspace", null);
        String originalWorkspace = RequestAttributeWorkspaceSelector.getForcedWorkspace(request);
        try
        {
            if (StringUtils.isNotEmpty(workspaceName))
            {
                RequestAttributeWorkspaceSelector.setForcedWorkspace(request, workspaceName);
            }
            
            SearchResults<Content> results = (SearchResults<Content>) request.getAttribute(SearchAction.SEARCH_RESULTS);
            SearchModel model = (SearchModel) request.getAttribute(SearchAction.SEARCH_MODEL);
            I18nizableText queryError = (I18nizableText) request.getAttribute(SearchAction.QUERY_ERROR);
            
            Locale defaultLocale = (Locale) request.getAttribute(SearchAction.SEARCH_LOCALE);
            
            Map<String, Object> jsonObject = new HashMap<>();
            
            if (queryError != null)
            {
                jsonObject.put("error", queryError);
            }
            else if (results != null)
            {
                convertResults2JsonObject(jsonObject, results, model, defaultLocale);
            }
            else
            {
                List<String> contentIds = (List<String>) request.getAttribute(SearchAction.SEARCH_CONTENTS);
                convertContents2JsonObject(jsonObject, contentIds, model, defaultLocale);
            }
            
            _jsonUtils.convertObjectToJson(out, jsonObject);
        }
        finally
        {
            if (StringUtils.isNotEmpty(workspaceName))
            {
                RequestAttributeWorkspaceSelector.setForcedWorkspace(request, originalWorkspace);
            }
        }
    }
    
    /**
     * Convert the query results to a JSON object
     * @param results2Json the JSON object to fill.
     * @param results the query results
     * @param model the search model
     * @param defaultLocale the default locale for localized values if content's language is null. 
     * @throws ProcessingException If an error occurs.
     */
    protected void convertResults2JsonObject(Map<String, Object> results2Json, SearchResults<Content> results, SearchModel model, Locale defaultLocale) throws ProcessingException
    {
        Map<String, Object> jsParameters = _serverCommHelper.getJsParameters();
        @SuppressWarnings("unchecked")
        Map<String, Object> values = (Map<String, Object>) jsParameters.get("values");
        @SuppressWarnings("unchecked")
        Map<String, Object> contextualParameters = (Map<String, Object>) jsParameters.get("contextualParameters");
        if (contextualParameters == null)
        {
            contextualParameters = Collections.emptyMap();
        }
        
        setContents(results2Json, results, model, defaultLocale, contextualParameters);
        
        setFacets(results2Json, results.getFacetResults(), model, defaultLocale, contextualParameters);
        
        results2Json.put("values", values);
        results2Json.put("total", results.getTotalCount());
    }
    
    /**
     * Convert the query results to a JSON object
     * @param results2Json the JSON object to fill.
     * @param contentIds the ids of contents
     * @param model the search model
     * @param defaultLocale the default locale for localized values if content's language is null. 
     * @throws ProcessingException If an error occurs.
     */
    protected void convertContents2JsonObject(Map<String, Object> results2Json, List<String> contentIds, SearchModel model, Locale defaultLocale) throws ProcessingException
    {
        Map<String, Object> jsParameters = _serverCommHelper.getJsParameters();
        @SuppressWarnings("unchecked")
        Map<String, Object> contextualParameters = (Map<String, Object>) jsParameters.get("contextualParameters");
        if (contextualParameters == null)
        {
            contextualParameters = Collections.emptyMap();
        }
        
        int size = setContents(results2Json, contentIds, model, defaultLocale, contextualParameters);
        results2Json.put("total", size);
    }
    
    /**
     * Extract the desired contents from the {@link SearchResults} object and set them in the search results.
     * @param searchResults the search results to populate.
     * @param results the {@link SearchResults}.
     * @param model the search model.
     * @param defaultLocale the default locale for localized values if content's language is null. 
     * @param contextualParameters the search contextual parameters.
     */
    protected void setContents(Map<String, Object> searchResults, SearchResults<Content> results, SearchModel model, Locale defaultLocale, Map<String, Object> contextualParameters)
    {
        List<Map<String, Object>> contents = new ArrayList<>();
        searchResults.put("contents", contents);
        
        SearchModelContentValuesExtractor extractor = _valuesExtractorFactory.create(model);
        
        for (SearchResult<Content> result : results.getResults())
        {
            contents.add(getContentData(result.getObject(), extractor, defaultLocale, contextualParameters));
        }
    }
    
    /**
     * Extract the desired contents and set them in the search results.
     * @param searchResults the search results to populate.
     * @param contentIds the id of desired contents
     * @param model the search model.
     * @param defaultLocale the default locale for localized values if content's language is null. 
     * @param contextualParameters the search contextual parameters.
     * @return the number of resolved contents without error 
     */
    protected int setContents(Map<String, Object> searchResults, List<String> contentIds, SearchModel model, Locale defaultLocale, Map<String, Object> contextualParameters)
    { 
        List<Map<String, Object>> contents = new ArrayList<>();
        searchResults.put("contents", contents);
        
        SearchModelContentValuesExtractor extractor = _valuesExtractorFactory.create(model);
        
        int count = 0;
        for (String contentId : contentIds)
        {
            try
            {
                Content content = _resolver.resolveById(contentId);
                contents.add(getContentData(content, extractor, defaultLocale, contextualParameters));
                count++;
            }
            catch (UnknownAmetysObjectException e)
            {
                getLogger().warn("The Ametys object with id '" + contentId + "' does not exist anymore.");
            }
        }
        return count;
    }
    
    /**
     * Generate standard content data.
     * @param content The content.
     * @param extractor The content values extractor which generates.
     * @param defaultLocale the default locale for localized values if content's language is null. 
     * @param contextualParameters The search contextual parameters.
     * @return A Map containing the content data.
     */
    protected Map<String, Object> getContentData(Content content, ContentValuesExtractor extractor, Locale defaultLocale, Map<String, Object> contextualParameters)
    {
        return _contentGridComponent.getContentData(content, extractor, defaultLocale, contextualParameters);
    }
    
    /**
     * Set the facet results in the search results.
     * @param searchResults the search results.
     * @param facetResults the facet results as a Map&lt;column name, Map&lt;value, result count&gt;&gt;.
     * @param model the search model.
     * @param locale the locale for localized values
     * @param contextualParameters the search contextual parameters.
     */
    protected void setFacets(Map<String, Object> searchResults, Map<String, Map<String, Integer>> facetResults, SearchModel model, Locale locale, Map<String, Object> contextualParameters)
    {
        List<Object> facets = new ArrayList<>();
        searchResults.put("facets", facets);
        
        // Index facets by search field name.
        Map<String, SearchCriterion> facetByName = new HashMap<>();
        for (SearchCriterion facet : model.getFacetedCriteria(contextualParameters).values())
        {
            SearchField searchField = facet.getSearchField();
            if (searchField != null)
            {
                facetByName.put(searchField.getName(), facet);
            }
        }
        
        for (String criterionName : facetResults.keySet())
        {
            SearchCriterion searchCriterion = facetByName.get(criterionName);
            
            if (searchCriterion != null)
            {
                Map<String, Integer> values = facetResults.get(criterionName);
                
                Map<String, Object> column = new HashMap<>();
                facets.add(column);
                
                List<Map<String, Object>> criterionFacets = new ArrayList<>();
                column.put("children", criterionFacets);
                column.put("name", criterionName);
                column.put("label", searchCriterion.getLabel());
                column.put("type", "criterion");
                
                for (String value : values.keySet())
                {
                    Map<String, Object> facet = new LinkedHashMap<>();
                    criterionFacets.add(facet);
                    
                    Integer count = values.get(value);
                    I18nizableText label = _searchCriterionHelper.getFacetLabel(searchCriterion, value, locale);
                    
                    facet.put("value", value);
                    facet.put("count", count);
                    facet.put("label", label);
                    facet.put("type", "facet");
                }
            }
        }
    }
}
