/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.content;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.contenttype.ContentAttributeDefinition;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.contenttype.MetadataType;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.model.ContentElementDefinition;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.cms.search.solr.field.BooleanSearchField;
import org.ametys.cms.search.solr.field.ContentSearchField;
import org.ametys.cms.search.solr.field.DateSearchField;
import org.ametys.cms.search.solr.field.DoubleSearchField;
import org.ametys.cms.search.solr.field.JoinedSystemSearchField;
import org.ametys.cms.search.solr.field.LongSearchField;
import org.ametys.cms.search.solr.field.MultilingualStringSearchField;
import org.ametys.cms.search.solr.field.StringSearchField;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemAccessor;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.model.type.ModelItemType;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Component which helps content searching by providing a simple way to access
 * content properties (either attributes or properties).
 */
public class ContentSearchHelper extends AbstractLogEnabled implements Component, Serviceable, Contextualizable
{
    /** The component role. */
    public static final String ROLE = ContentSearchHelper.class.getName();
    
    /** The content type extension point. */
    private ContentTypeExtensionPoint _cTypeEP;
    
    /** The content type helper. */
    private ContentTypesHelper _cTypeHelper;
    
    /** The system property extension point. */
    private SystemPropertyExtensionPoint _sysPropEP;
    
    /** Content Types helper */
    private ContentTypesHelper _contentTypesHelper;

    private Context _context;
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _cTypeEP = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _cTypeHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
        _sysPropEP = (SystemPropertyExtensionPoint) manager.lookup(SystemPropertyExtensionPoint.ROLE);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
    }
    
    /**
     * Get a {@link SearchField} corresponding to a metadata.
     * @param joinPaths The join paths
     * @param metadataPath The metadata path.
     * @param metadataType The metadata type.
     * @param isTypeContentWithMultilingualTitle <code>true</code> if the type is Content and the linked contents have multilingual titles
     * @return the search field.
     */
    public SearchField getMetadataSearchField(List<String> joinPaths, String metadataPath, MetadataType metadataType, boolean isTypeContentWithMultilingualTitle)
    {
        switch (metadataType)
        {
            case STRING:
            case USER:
            case REFERENCE:
                return new StringSearchField(joinPaths, metadataPath);
            case CONTENT:
            case SUB_CONTENT:
                return new ContentSearchField(joinPaths, metadataPath, isTypeContentWithMultilingualTitle, Optional.ofNullable(_context));
            case LONG:
                return new LongSearchField(joinPaths, metadataPath);
            case DOUBLE:
                return new DoubleSearchField(joinPaths, metadataPath);
            case BOOLEAN:
                return new BooleanSearchField(joinPaths, metadataPath);
            case DATE:
            case DATETIME:
                return new DateSearchField(joinPaths, metadataPath);
            case MULTILINGUAL_STRING:
                return new MultilingualStringSearchField(joinPaths, metadataPath, Optional.ofNullable(_context));
            case COMPOSITE:
            case BINARY:
            case FILE:
            case RICH_TEXT:
            default:
                return null;
        }
    }
    
    /**
     * Get a {@link SearchField} from a field name in a batch of content types.
     * @param contentTypes The content types, can be empty to search on any content type.
     * In that case, only the title metadata and system properties will be usable in sort and facets specs.
     * @param fieldPath The field path, can be either a system property ID or a indexing field name or path.
     * @return The {@link SearchField} corresponding to the field path, or an {@link Optional#empty() empty optional} if not found
     */
    public Optional<SearchField> getSearchField(Collection<String> contentTypes, String fieldPath)
    {
        // Root system property
        if (!fieldPath.contains(ModelItem.ITEM_PATH_SEPARATOR) && _sysPropEP.hasExtension(fieldPath))
        {
            return Optional.ofNullable(_sysPropEP.getExtension(fieldPath).getSearchField());
        }
        
        Set<String> commonContentTypeIds = _cTypeHelper.getCommonAncestors(contentTypes);
        if (!commonContentTypeIds.isEmpty())
        {
            try
            {
                List<String> joinPaths = computeJoinPaths(fieldPath, commonContentTypeIds, false);

                String fieldName = fieldPath.contains(ModelItem.ITEM_PATH_SEPARATOR) ? StringUtils.substringAfterLast(fieldPath, ModelItem.ITEM_PATH_SEPARATOR) : fieldPath;
                if (_sysPropEP.hasExtension(fieldName))
                {
                    return Optional.of(new JoinedSystemSearchField(joinPaths, _sysPropEP.getExtension(fieldName).getSearchField()));
                }
                else
                {
                    Collection<ContentType> commonContentTypes = commonContentTypeIds.stream()
                                                                                     .map(_cTypeEP::getExtension)
                                                                                     .collect(Collectors.toList());
                    ModelItem modelItem = ModelHelper.getModelItem(fieldPath, commonContentTypes);
                    return Optional.of(_getModelItemSearchField(joinPaths, modelItem.getPath(), modelItem.getType(), isTitleMultilingual(modelItem)));
                }
            }
            catch (UndefinedItemPathException e)
            {
                throw new IllegalArgumentException("Search field with path '" + fieldPath + "' refers to an unknown attribute or property", e);
            }
        }
        else if (fieldPath.equals(Content.ATTRIBUTE_TITLE))
        {
            // No specific content type: allow only title.
            ModelItem titleDefinition = _contentTypesHelper.getTitleAttributeDefinition();
            return Optional.of(_getModelItemSearchField(null, fieldPath, titleDefinition.getType(), false));
        }
        
        return Optional.empty();
    }
    
    /**
     * Retrieves the join paths from the given data path
     * @param dataPath the data path
     * @param contentTypeIds the ids of the content types containing the items of the given path
     * @param addLast <code>true</code> to add the last join path element to the list, <code>false</code> otherwise.
     * @return the join paths
     * @throws UndefinedItemPathException if there is no item defined at the given path in given item accessors
     */
    public List<String> computeJoinPaths(String dataPath, Set<String> contentTypeIds, boolean addLast) throws UndefinedItemPathException
    {
        Collection<ContentType> contentTypes = contentTypeIds.stream()
                .map(_cTypeEP::getExtension)
                .collect(Collectors.toList());
        
        return _computeJoinPaths(dataPath, StringUtils.EMPTY, contentTypes, addLast);
    }
    
    /**
     * Retrieves the join paths from the given data path
     * @param dataPath the data path
     * @param prefix the prefix of current join path
     * @param modelItemAccessor the model item to access the items of the given path
     * @param addLast <code>true</code> to add the last join path element to the list, <code>false</code> otherwise.
     * @return the join paths
     * @throws UndefinedItemPathException if there is no item defined at the given path in given item accessor
     */
    protected List<String> _computeJoinPaths(String dataPath, String prefix, ModelItemAccessor modelItemAccessor, boolean addLast) throws UndefinedItemPathException
    {
        return _computeJoinPaths(dataPath, prefix, List.of(modelItemAccessor), addLast);
    }
    
    /**
     * Retrieves the join paths from the given data path
     * @param dataPath the data path
     * @param prefix the prefix of current join path
     * @param modelItemAccessors the model items to access the items of the given path
     * @param addLast <code>true</code> to add the last join path element to the list, <code>false</code> otherwise.
     * @return the join paths
     * @throws UndefinedItemPathException if there is no item defined at the given path in given item accessors
     */
    protected List<String> _computeJoinPaths(String dataPath, String prefix, Collection<? extends ModelItemAccessor> modelItemAccessors, boolean addLast) throws UndefinedItemPathException
    {
        String[] pathSegments = StringUtils.split(dataPath, ModelItem.ITEM_PATH_SEPARATOR);
        if (pathSegments == null || pathSegments.length < 1)
        {
            throw new IllegalArgumentException("Unable to retrieve the data at the given path. This path is empty.");
        }
        else if (pathSegments.length == 1)
        {
            return addLast ? List.of(prefix + dataPath) : List.of();
        }
        else
        {
            List<String> joinPaths = new ArrayList<>();
            String subDataPath = StringUtils.join(pathSegments, ModelItem.ITEM_PATH_SEPARATOR, 1, pathSegments.length);
            ModelItem modelItem = ModelHelper.getModelItem(pathSegments[0], modelItemAccessors);
            if (modelItem instanceof ContentElementDefinition contentElementDefinition)
            {
                joinPaths.add(prefix + pathSegments[0]);
                joinPaths.addAll(_computeJoinPaths(subDataPath, StringUtils.EMPTY, contentElementDefinition, addLast));
            }
            else if (modelItem instanceof org.ametys.plugins.repository.model.RepeaterDefinition repeaterDefinition)
            {
                joinPaths.add(prefix + pathSegments[0]);
                joinPaths.addAll(_computeJoinPaths(subDataPath, StringUtils.EMPTY, repeaterDefinition, addLast));
            }
            else if (modelItem instanceof ModelItemAccessor modelItemAccessor)
            {
                joinPaths.addAll(_computeJoinPaths(subDataPath, pathSegments[0] + ModelItem.ITEM_PATH_SEPARATOR, modelItemAccessor, addLast));
            }
            
            return joinPaths;
        }
    }
    
    /**
     * Get a {@link SearchField} from a model view item.
     * @param modelViewItem The model view item
     * @return The {@link SearchField} corresponding to given model view item
     */
    public SearchField getSearchField(ModelViewItem modelViewItem)
    {
        List<String> joinPaths = _computeJoinPaths(modelViewItem);
        ModelItem modelItem = modelViewItem.getDefinition();
        
        if (modelItem instanceof SystemProperty systemProperty)
        {
            SearchField systemPropertySearchField = systemProperty.getSearchField();
            if (joinPaths.isEmpty())
            {
                return systemPropertySearchField;
            }
            else
            {
                return new JoinedSystemSearchField(joinPaths, systemPropertySearchField);
            }
        }
        else
        {
            return _getModelItemSearchField(joinPaths, modelItem.getPath(), modelItem.getType(), false);
        }
    }
    
    /**
     * Retrieves the join paths from the given view item
     * @param viewItem the view item
     * @return the join paths
     */
    public List<String> _computeJoinPaths(ViewItem viewItem)
    {
        List<String> joinPaths = new ArrayList<>();
        String currentJoinPath = StringUtils.EMPTY;

        ViewItemAccessor parent = viewItem.getParent();
        while (parent != null)
        {
            if (parent instanceof ModelViewItem parentModelViewItem)
            {
                ModelItem modelItem = parentModelViewItem.getDefinition();
                if (modelItem instanceof ContentElementDefinition || modelItem instanceof org.ametys.plugins.repository.model.RepeaterDefinition)
                {
                    if (StringUtils.isNotEmpty(currentJoinPath))
                    {
                        joinPaths.add(0, currentJoinPath);
                    }
                    currentJoinPath = modelItem.getName();
                }
                else if (StringUtils.isNotEmpty(currentJoinPath))
                {
                    currentJoinPath = modelItem.getName() + ModelItem.ITEM_PATH_SEPARATOR + currentJoinPath;
                }
            }
            
            parent = parent instanceof ViewItem parentViewItem ? parentViewItem.getParent() : null;
            if (parent == null && StringUtils.isNotEmpty(currentJoinPath))
            {
                joinPaths.add(0, currentJoinPath);
            }
        }
            
        return joinPaths;
    }
    
    /**
     * Determines if the given model item represents an attribute of type content with contents with multilingual titles
     * @param modelItem The model item
     * @return <code>true</code> if the given model item represents an attribute of type content with contents with multilingual titles
     */
    public boolean isTitleMultilingual(ModelItem modelItem)
    {
        return Optional.ofNullable(modelItem)
                .filter(ContentAttributeDefinition.class::isInstance)
                .map(ContentAttributeDefinition.class::cast)
                .map(ContentAttributeDefinition::getContentTypeId)
                .map(_cTypeEP::getExtension)
                .filter(cType -> cType.hasModelItem(Content.ATTRIBUTE_TITLE))
                .map(cType -> cType.getModelItem(Content.ATTRIBUTE_TITLE))
                .map(ModelItem::getType)
                .map(ModelItemType::getId)
                .map(ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID::equals)
                .orElse(false);
    }
    
    @SuppressWarnings("static-access")
    // FIXME CMS-11713: put the SearchField implementation in types?
    private SearchField _getModelItemSearchField(List<String> joinPaths, String modelItemPath, ModelItemType modelItemType, boolean isTypeContentWithMultilingualTitle)
    {
        switch (modelItemType.getId())
        {
            case ModelItemTypeConstants.STRING_TYPE_ID:
            case ModelItemTypeConstants.USER_ELEMENT_TYPE_ID:
            case ModelItemTypeConstants.REFERENCE_ELEMENT_TYPE_ID:
                return new StringSearchField(joinPaths, modelItemPath);
            case ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID:
                return new ContentSearchField(joinPaths, modelItemPath, isTypeContentWithMultilingualTitle, Optional.ofNullable(_context));
            case ModelItemTypeConstants.LONG_TYPE_ID:
                return new LongSearchField(joinPaths, modelItemPath);
            case ModelItemTypeConstants.DOUBLE_TYPE_ID:
                return new DoubleSearchField(joinPaths, modelItemPath);
            case ModelItemTypeConstants.BOOLEAN_TYPE_ID:
                return new BooleanSearchField(joinPaths, modelItemPath);
            case ModelItemTypeConstants.DATE_TYPE_ID:
            case ModelItemTypeConstants.DATETIME_TYPE_ID:
                return new DateSearchField(joinPaths, modelItemPath);
            case ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID:
                return new MultilingualStringSearchField(joinPaths, modelItemPath, Optional.ofNullable(_context));
            case ModelItemTypeConstants.COMPOSITE_TYPE_ID:
            case ModelItemTypeConstants.REPEATER_TYPE_ID:
            case ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID:
            case ModelItemTypeConstants.FILE_ELEMENT_TYPE_ID:
            case ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID:
            default:
                return null;
        }
    }
}
