/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.content;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.ametysobject.ModelAwareDataAwareAmetysObject;
import org.ametys.cms.data.holder.impl.IndexableDataHolderHelper;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.model.CMSDataContext;
import org.ametys.cms.model.properties.Property;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.model.SearchModel;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.ui.model.SearchUIColumn;
import org.ametys.cms.search.ui.model.SearchUIColumnHelper;
import org.ametys.cms.search.ui.model.impl.RepeaterSearchUIColumn;
import org.ametys.plugins.repository.data.external.ExternalizableDataProviderExtensionPoint;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeaterEntry;
import org.ametys.plugins.repository.model.CompositeDefinition;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.plugins.repository.model.ViewHelper;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemGroup;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElement;
import org.ametys.runtime.model.ViewElementAccessor;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.ViewItemContainer;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ModelItemType;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Component creating content values extractors from {@link SearchModel}s or {@link View}s.
 */
public class ContentValuesExtractorFactory extends AbstractLogEnabled implements Component, Serviceable
{
    /** The component role. */
    public static final String ROLE = ContentValuesExtractorFactory.class.getName();
    
    /** The content search helper. */
    protected ContentSearchHelper _searchHelper;
    
    /** To determine the externalizable status */
    protected ExternalizableDataProviderExtensionPoint _externalizableDataProviderEP;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _searchHelper = (ContentSearchHelper) manager.lookup(ContentSearchHelper.ROLE);
        _externalizableDataProviderEP = (ExternalizableDataProviderExtensionPoint) manager.lookup(ExternalizableDataProviderExtensionPoint.ROLE);
    }
    
    /**
     * Create a ContentValuesExtractor from a search model.
     * @param searchModel The reference search model.
     * @return a ContentValuesExtractor backed by the given search model.
     */
    public SearchModelContentValuesExtractor create(SearchModel searchModel)
    {
        return new SearchModelContentValuesExtractor(searchModel);
    }
    
    /**
     * Create a simple {@link ContentValuesExtractor} from a view
     * @param view The view.
     * @return The created {@link ContentValuesExtractor}
     */
    public SimpleContentValuesExtractor create(View view)
    {
        return new SimpleContentValuesExtractor(view);
    }
    
    /**
     * A ContentValuesExtractor 
     */
    public interface ContentValuesExtractor
    {
        /**
         * Get the values from the given content.
         * @param content The content.
         * @param defaultLocale The default locale for localized values if the content's language is null. Can be null.
         * @param contextualParameters The search contextual parameters.
         * @return the extracted values.
         */
        public Map<String, Object> getValues(Content content, Locale defaultLocale, Map<String, Object> contextualParameters);
    }
    
    /**
     * An abstract implementation of ContentValuesExtractor
     */
    public abstract class AbstractContentValuesExtractor implements ContentValuesExtractor
    {
        public Map<String, Object> getValues(Content content, Locale defaultLocale, Map<String, Object> contextualParameters)
        {
            CMSDataContext context = CMSDataContext.newInstance()
                                                   .withRichTextMaxLength(100)
                                                   .withObject(content)
                                                   .withLocale(defaultLocale)
                                                   .withEmptyValues(false);

            boolean handleExternalizable = (boolean) contextualParameters.getOrDefault("externalizable", true);
            if (handleExternalizable)
            {
                Set<String> externalizableData = _externalizableDataProviderEP.getExternalizableDataPaths(content);
                context.withExternalizableData(externalizableData);
            }

            ViewItemContainer viewItemContainer = _getResultItems(content.getModel(), contextualParameters);
            return _dataToJSON(content, viewItemContainer, context, StringUtils.EMPTY);
        }
        
        /**
         * Retrieves the view to use to extract values
         * @param model the model
         * @param contextualParameters The search contextual parameters.
         * @return the {@link View} to use to extract values
         */
        protected abstract ViewItemContainer _getResultItems(Collection<? extends Model> model, Map<String, Object> contextualParameters);
    }
    
    /**
     * A ContentValuesExtractor backed by a {@link SearchModel}.
     */
    public class SearchModelContentValuesExtractor extends AbstractContentValuesExtractor
    {
        private SearchModel _searchModel;
        
        /**
         * Build a ContentValuesExtractor referencing a {@link SearchModel}.
         * @param searchModel the {@link SearchModel}.
         */
        public SearchModelContentValuesExtractor(SearchModel searchModel)
        {
            _searchModel = searchModel;
        }
        
        @Override
        protected ViewItemContainer _getResultItems(Collection< ? extends Model> model, Map<String, Object> contextualParameters)
        {
            return _searchModel.getResultItems(contextualParameters);
        }
    }
    
    /**
     * A simple ContentValuesExtractor on a list of content types.
     */
    public class SimpleContentValuesExtractor extends AbstractContentValuesExtractor
    {
        private ViewItemContainer _resultItems;
        
        /**
         * Build a simple {@link ContentValuesExtractor} from the given result items
         * @param resultItems The result items
         * @throws IllegalArgumentException if the given result items contain a composite as leaf, or a {@link SystemProperty} that is not displayable
         */
        public SimpleContentValuesExtractor(ViewItemContainer resultItems) throws IllegalArgumentException
        {
            _resultItems = resultItems;
        }
        
        @Override
        protected ViewItemContainer _getResultItems(Collection< ? extends Model> model, Map<String, Object> contextualParameters)
        {
            return _resultItems;
        }
    }
    
    @SuppressWarnings("unchecked")
    private Map<String, Object> _dataToJSON(ModelAwareDataHolder dataHolder, ViewItemAccessor viewItemAccessor, DataContext context, String prefix) throws BadItemTypeException, UndefinedItemPathException
    {
        Map<String, Object> result = new HashMap<>();
        Map<String, Content> resolvedContents = new HashMap<>();
        
        ViewHelper.visitView(viewItemAccessor, 
                (element, definition) -> {
                    // simple element
                    String name = definition.getName();
                    String newDataPath = prefix + name;

                    DataContext newContext = CMSDataContext.newInstance(context)
                                                           .addSegmentToDataPath(name)
                                                           .withViewItem(element)
                                                           .withModelItem(definition);
                    
                    if (element instanceof ViewElementAccessor viewElementAccessor
                        && !viewElementAccessor.getViewItems().isEmpty())
                    {
                        // The view item is an accessor -> convert all its children to JSON
                        if (dataHolder.hasValue(name) && ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID.equals(definition.getType().getId()))
                        {
                            Map<String, Object> contentJSON = definition.isMultiple()
                                    ? _multipleContentToJSON(dataHolder, viewElementAccessor, name, newDataPath, newContext, resolvedContents)
                                    : _singleContentToJSON(dataHolder, viewElementAccessor, name, newDataPath, newContext, resolvedContents);
                            
                            result.putAll(contentJSON);
                        }
                    }
                    else if (definition instanceof Property property)
                    {
                        ModelAwareDataAwareAmetysObject ametysObject = IndexableDataHolderHelper.getAmetysObjectFromContext(newContext);
                        Object json = property.valueToJSON(ametysObject, newContext);
                        result.put(newDataPath, json);
                    }
                    else
                    {
                        if (((CMSDataContext) newContext).isDataExternalizable())
                        {
                            Map<String, Object> json = IndexableDataHolderHelper.externalizableValuesAsJson(dataHolder, name, newContext);
                            if (!json.isEmpty())
                            {
                                result.put(newDataPath, json);
                            }
                        }
                        else if (dataHolder.hasValue(name))
                        {
                            Object value = dataHolder.getValue(name);
                            if (ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID.equals(definition.getType().getId()))
                            {
                                // Remove view item from context because children must be processed by the current algorithm, not by the type's one
                                newContext.withViewItem(null);
                                
                                // Resolve contents here to use resolved contents cache
                                value = _resolveContentValues(definition, value, resolvedContents);
                            }
                            
                            ModelItemType type = definition.getType();
                            Object json = type.valueToJSONForClient(value, newContext);
                            result.put(newDataPath, json);
                        }
                    }
                }, 
                (group, definition) -> {
                    // composite
                    String name = definition.getName();
                    DataContext newContext = CMSDataContext.newInstance(context)
                                                           .addSegmentToDataPath(name)
                                                           .withViewItem(group)
                                                           .withModelItem(definition);
                    
                    if (group.getViewItems().isEmpty())
                    {
                        throw new IllegalArgumentException("Attribute at path '" + newContext.getDataPath() + "' is a composite metadata : can not invoked #getAttributeValue");
                    }
                    
                    if (dataHolder.hasValue(name))
                    {
                        ModelAwareComposite value = dataHolder.getValue(name);
                        Map<String, Object> json = _dataToJSON(value, group, newContext, prefix + name + ModelItem.ITEM_PATH_SEPARATOR);
                        result.putAll(json);
                    }
                }, 
                (group, definition) -> {
                    // repeater
                    String name = definition.getName();
                    DataContext repeaterContext = CMSDataContext.newInstance(context)
                                                                .addSegmentToDataPath(name)
                                                                .withViewItem(group)
                                                                .withModelItem(definition);
                    
                    if (dataHolder.hasValue(name))
                    {
                        ModelAwareRepeater repeater = dataHolder.getValue(name);
                        String repeaterPath = prefix + name;
                        
                        if (group instanceof SearchUIColumn || group.getViewItems().isEmpty())
                        {
                            ModelViewItemGroup<RepeaterDefinition> repeaterLeaf = _getRepeaterLeafViewItem(group, definition);
                            Map<String, Object> repeaterLeafJSON = _repeaterLeafToJSON(repeater, repeaterLeaf, repeaterContext);
                            result.put(repeaterPath, repeaterLeafJSON);
                        }
                        else
                        {
                            Map<String, Object> repeaterJSON = _repeaterToJSON(repeater, group, repeaterPath, repeaterContext);
                            result.putAll(repeaterJSON);
                        }
                    }
                }, 
                group -> result.putAll(_dataToJSON(dataHolder, group, context, prefix)));
        
        return result;
    }
    
    @SuppressWarnings("unchecked")
    private Map<String, Object> _multipleContentToJSON(ModelAwareDataHolder dataHolder, ViewElementAccessor viewElementAccessor, String modelItemName, String dataPath, DataContext context, Map<String, Content> resolvedContents)
    {
        Map<String, List<Object>> allDataJSON = new HashMap<>(); // JSON for all data except repeaters
        Map<String, Object> repeatersJSON = new HashMap<>();     // JSON for repeaters that are leaves

        ContentValue[] conntentValues = dataHolder.getValue(modelItemName);
        for (ContentValue contentValue : conntentValues)
        {
            Content linkedContent = _getResolvedContent(contentValue, resolvedContents);
            if (linkedContent != null)
            {
                DataContext contentContext = CMSDataContext.newInstance(context)
                                                           .withObject(linkedContent)
                                                           .withDataPath(StringUtils.EMPTY);
                
                Map<String, Object> contentJSON = _dataToJSON(linkedContent, viewElementAccessor, contentContext, dataPath + ModelItem.ITEM_PATH_SEPARATOR);
                for (String childDataPath : contentJSON.keySet())
                {
                    String definitionPath = ModelHelper.getDefinitionPathFromDataPath(childDataPath);
                    Object childJSON = contentJSON.get(childDataPath);
                    if (_isRepeaterJSON(childJSON))
                    {
                        // The child is a repeater -> merge the entries
                        _mergeRepeaterEntriesJSON(repeatersJSON, (Map<String, Object>) childJSON, definitionPath);
                    }
                    else
                    {
                        // Merge data of all contents
                        List<Object> dataJSON = allDataJSON.computeIfAbsent(definitionPath, __ -> new ArrayList<>());
                        if (childJSON instanceof List jsonList)
                        {
                            dataJSON.addAll(jsonList);
                        }
                        else
                        {
                            dataJSON.add(childJSON);
                        }
                    }
                }
            }
        }

        Map<String, Object> result = new HashMap<>();
        result.putAll(allDataJSON);
        result.putAll(repeatersJSON);
        
        return result;
    }
    
    private Map<String, Object> _singleContentToJSON(ModelAwareDataHolder dataHolder, ViewElementAccessor viewElementAccessor, String modelItemName, String dataPath, DataContext context, Map<String, Content> resolvedContents)
    {
        Map<String, Object> result = new HashMap<>();
        
        ContentValue contentValue = dataHolder.getValue(modelItemName);
        Content linkedContent = _getResolvedContent(contentValue, resolvedContents);
        
        if (linkedContent != null)
        {
            DataContext contentContext = CMSDataContext.newInstance(context)
                                                       .withObject(linkedContent)
                                                       .withDataPath(StringUtils.EMPTY);
            
            Map<String, Object> contentJSON = _dataToJSON(linkedContent, viewElementAccessor, contentContext, dataPath + ModelItem.ITEM_PATH_SEPARATOR);
            result.putAll(contentJSON);
        }
        
        return result;
    }
    
    private Object _resolveContentValues(ElementDefinition definition, Object value, Map<String, Content> resolvedContents)
    {
        Object result = value;
        
        if (definition.isMultiple())
        {
            ContentValue[] contentValues = (ContentValue[]) value;
            List<Content> contents = new ArrayList<>();
            for (ContentValue contentValue : contentValues)
            {
                Content content = _getResolvedContent(contentValue, resolvedContents);
                if (content != null)
                {
                    contents.add(content);
                }
            }
            
            result = contents.toArray(new Content[contents.size()]);
        }
        else
        {
            ContentValue contentValue = (ContentValue) value;
            Content content = _getResolvedContent(contentValue, resolvedContents);
            result = content != null ? content : result;
        }
        
        return result;
    }
    
    private Content _getResolvedContent(ContentValue contentValue, Map<String, Content> resolvedContents)
    {
        return resolvedContents.computeIfAbsent(contentValue.getContentId(),
                id -> contentValue.getContentIfExists().orElse(null));
    }
    
    private Map<String, Object> _repeaterLeafToJSON(ModelAwareRepeater repeater, ModelViewItemGroup<RepeaterDefinition> repeaterViewItem, DataContext context)
    {
        List<Map<String, Object>> entriesJSON = new ArrayList<>();
        for (ModelAwareRepeaterEntry entry : repeater.getEntries())
        {
            DataContext entryContext = CMSDataContext.newInstance(context)
                                                     .addSuffixToLastSegment("[" + entry.getPosition() + "]");
            
            Map<String, Object> entryValuesJson = _dataToJSON(entry, repeaterViewItem, entryContext, StringUtils.EMPTY);
            
            Map<String, Object> entryJSON = new HashMap<>();
            entryJSON.put("values", entryValuesJson);
            entryJSON.put("position", entry.getPosition());
            
            entriesJSON.add(entryJSON);
        }
        
        Map<String, Object> json = new HashMap<>();
        json.put("type", org.ametys.plugins.repository.data.type.ModelItemTypeConstants.REPEATER_TYPE_ID);
        json.put("entries", entriesJSON);
        json.put("label", repeaterViewItem.getDefinition().getLabel());
        
        Optional.ofNullable(repeaterViewItem.getDefinition().getHeaderLabel())
                .ifPresent(headerLabel -> json.put("header-label", headerLabel));
        
        return json;
    }
    
    private ModelViewItemGroup<RepeaterDefinition> _getRepeaterLeafViewItem(ModelViewItemGroup<RepeaterDefinition> group, RepeaterDefinition definition)
    {
        boolean useColumns = group instanceof SearchUIColumn;
        ModelViewItemGroup<RepeaterDefinition> viewItemGroup = useColumns ? new RepeaterSearchUIColumn() : new ModelViewItemGroup<>();
        viewItemGroup.setDefinition(definition);
        
        for (ModelItem child : definition.getChildren())
        {
            _addViewItemForRepeaterLeaf(child, viewItemGroup, useColumns);
        }
        
        return viewItemGroup;
    }
    
    @SuppressWarnings("unchecked")
    private void _addViewItemForRepeaterLeaf(ModelItem modelItem, ModelViewItemGroup group, boolean isColumns)
    {
        ModelViewItem viewItem;
        if (modelItem instanceof CompositeDefinition compositeDefinition)
        {
            viewItem = new ModelViewItemGroup<>();

            // Add children only for composites, children of type repeater do not need to have children
            for (ModelItem child : compositeDefinition.getChildren())
            {
                _addViewItemForRepeaterLeaf(child, (ModelViewItemGroup) viewItem, isColumns);
            }
        }
        else
        {
            viewItem = isColumns
                    ? SearchUIColumnHelper.createModelItemColumn(modelItem)
                    : modelItem instanceof ModelItemGroup
                        ? new ModelViewItemGroup<>()
                        : new ViewElement();
        }
        
        viewItem.setDefinition(modelItem);
        group.addViewItem(viewItem);
    }
    
    @SuppressWarnings("unchecked")
    private Map<String, Object> _repeaterToJSON(ModelAwareRepeater repeater, ModelViewItemGroup<RepeaterDefinition> group, String repeaterPath, DataContext repeaterContext)
    {
        Map<String, List<Object>> singleDataJSON = new HashMap<>();
        Map<String, Object> repeatersJSON = new HashMap<>();
        
        for (ModelAwareRepeaterEntry entry : repeater.getEntries())
        {
            DataContext entryContext = CMSDataContext.newInstance(repeaterContext)
                                                     .addSuffixToLastSegment("[" + entry.getPosition() + "]");
            
            Map<String, Object> entryJson = _dataToJSON(entry, group, entryContext, repeaterPath + "[" + entry.getPosition() + "]" + ModelItem.ITEM_PATH_SEPARATOR);
            
            for (String entryPath : entryJson.keySet())
            {
                String definitionPath = ModelHelper.getDefinitionPathFromDataPath(entryPath);
                Object dataJSON = entryJson.get(entryPath);
                if (_isRepeaterJSON(dataJSON))
                {
                    // The child is a repeater -> merge the entries
                    _mergeRepeaterEntriesJSON(repeatersJSON, (Map<String, Object>) dataJSON, definitionPath);
                }
                else
                {
                    List<Object> dataJsonList = singleDataJSON.computeIfAbsent(definitionPath, __ -> new ArrayList<>());
                    dataJsonList.add(dataJSON);
                }
            }
        }
        
        Map<String, Object> result = new HashMap<>();
        result.putAll(singleDataJSON);
        result.putAll(repeatersJSON);
        
        return result;
    }
    
    private boolean _isRepeaterJSON(Object json)
    {
        return json instanceof Map map && org.ametys.plugins.repository.data.type.ModelItemTypeConstants.REPEATER_TYPE_ID.equals(map.get("type"));
    }
    
    @SuppressWarnings("unchecked")
    private void _mergeRepeaterEntriesJSON(Map<String, Object> repeatersJSON, Map<String, Object> json, String definitionPath)
    {
        if (repeatersJSON.containsKey(definitionPath))
        {
            Map<String, Object> repeaterJSON = (Map<String, Object>) repeatersJSON.get(definitionPath);
            List<Object> allEntries = (List<Object>) repeaterJSON.get("entries");
            List<Object> entries = (List<Object>) json.get("entries");
            allEntries.addAll(entries);
        }
        else
        {
            repeatersJSON.put(definitionPath, json);
        }
    }
}
