/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewHelper;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemContainer;

/**
 * Default implementation of a {@link SearchModel}.
 */
public class DefaultSearchModel implements SearchModel
{
    /** The content types of this search model. */
    protected Set<String> _cTypes;
    /** The content types excluded from this search model. */
    protected Set<String> _excludedCTypes;
    
    /** The simple search criteria, indexed by ID. */
    protected Map<String, SearchCriterion> _criteria;
    /** The faceted search criteria, indexed by ID. */
    protected Map<String, SearchCriterion> _facetedCriteria;
    /** The result items */
    protected ViewItemContainer _resultItems;
    /** the workspace */
    protected String _workspace;
    
    /**
     * Default constructor.
     */
    public DefaultSearchModel()
    {
        _cTypes = new HashSet<>();
        _excludedCTypes = new HashSet<>();
        
        _criteria = new LinkedHashMap<>();
        _facetedCriteria = new LinkedHashMap<>();
    }
    
    /**
     * Constructor by copying an existing {@link SearchModel}.
     * @param searchModelToCopy The {@link SearchModel} to copy
     * @param contextualParameters The contextual parameters
     */
    public DefaultSearchModel(SearchModel searchModelToCopy, Map<String, Object> contextualParameters)
    {
        setContentTypes(new HashSet<>(searchModelToCopy.getContentTypes(contextualParameters)));
        setExcludedContentTypes(new HashSet<>(searchModelToCopy.getExcludedContentTypes(contextualParameters)));
        
        setCriteria(new ArrayList<>(searchModelToCopy.getCriteria(contextualParameters).values()));
        setFacetedCriteria(new ArrayList<>(searchModelToCopy.getFacetedCriteria(contextualParameters).values()));
        
        View resultItems = new View();
        ViewItemContainer resultItemsToCopy = searchModelToCopy.getResultItems(contextualParameters);
        List<ViewItem> viewItems = ViewHelper.copyViewItems(resultItemsToCopy.getViewItems());
        resultItems.addViewItems(viewItems);
        setResulItems(resultItems);
        
        setWorkspace(searchModelToCopy.getWorkspace(contextualParameters));
    }
    
    public Set<String> getContentTypes(Map<String, Object> contextualParameters)
    {
        return Collections.unmodifiableSet(_cTypes);
    }
    
    /**
     * Set the content types of the models
     * @param cTypes The content types.
     */
    public void setContentTypes(Set<String> cTypes)
    {
        _cTypes = new HashSet<>(cTypes);
    }
    
    public Set<String> getExcludedContentTypes(Map<String, Object> contextualParameters)
    {
        return Collections.unmodifiableSet(_excludedCTypes);
    }
    
    /**
     * Set the content types to exclude
     * @param cTypes The content types to exclude
     */
    public void setExcludedContentTypes(Set<String> cTypes)
    {
        _excludedCTypes = new HashSet<>(cTypes);
    }
    
    public Map<String, ? extends SearchCriterion> getCriteria(Map<String, Object> contextualParameters)
    {
        return Collections.unmodifiableMap(_criteria);
    }
    
    /**
     * Add a criterion
     * @param criterion the criterion to add
     */
    public void addCriterion(SearchCriterion criterion)
    {
        _criteria.put(criterion.getId(), criterion);
    }
    
    /**
     * Set the criteria
     * @param criteria The criteria list
     */
    public void setCriteria(Collection<? extends SearchCriterion> criteria)
    {
        _criteria = new LinkedHashMap<>();
        for (SearchCriterion criterion : criteria)
        {
            _criteria.put(criterion.getId(), criterion);
        }
    }
    
    public Map<String, ? extends SearchCriterion> getFacetedCriteria(Map<String, Object> contextualParameters)
    {
        return Collections.unmodifiableMap(_facetedCriteria);
    }

    /**
     * Add a faceted criterion
     * @param criterion the faceted criterion to add
     */
    public void addFacetedCriterion(SearchCriterion criterion)
    {
        _facetedCriteria.put(criterion.getId(), criterion);
    }
    
    /**
     * Set the faceted criteria
     * @param criteria The list of faceted criteria
     */
    public void setFacetedCriteria(Collection<? extends SearchCriterion> criteria)
    {
        _facetedCriteria = new LinkedHashMap<>();
        for (SearchCriterion criterion : criteria)
        {
            _facetedCriteria.put(criterion.getId(), criterion);
        }
    }
    
    public ViewItemContainer getResultItems(Map<String, Object> contextualParameters)
    {
        return _resultItems;
    }
    
    /**
     * Set the result items
     * @param resultItems The result items to set
     */
    public void setResulItems(ViewItemContainer resultItems)
    {
        _resultItems = resultItems;
    }
    
    public String getWorkspace(Map<String, Object> contextualParameters)
    {
        return _workspace;
    }
    
    /**
     * Set the workspace
     * @param workspace The workspace to set
     */
    public void setWorkspace(String workspace)
    {
        _workspace = workspace;
    }
}
