/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.model;

import java.util.Map;

import org.ametys.cms.contenttype.MetadataType;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Represents a Field that can be searched on.
 */
public interface SearchCriterion extends Field
{
    /**
     * Get the SearchCriterion ID.
     * @return the SearchCriterion ID.
     */
    String getId();
    
    /**
     * Retrieves the label.
     * @return the label.
     */
    public I18nizableText getLabel();
    
    /**
     * Retrieves the description.
     * @return the description.
     */
    public I18nizableText getDescription();
    
    /**
     * Test if the SearchCriterion is sortable.
     * @return <code>true</code> if the criterion is sortable, false otherwise.
     */
    default boolean isSortable()
    {
        return isSortable(this);
    }
    
    /**
     * Test if a field is sortable.
     * @param field the field to test.
     * @return true if the field is sortable, false otherwise.
     */
    static boolean isSortable(Field field)
    {
        switch (field.getType())
        { 
            case STRING:
            case MULTILINGUAL_STRING:
            case LONG:
            case DATE:
            case DATETIME:
            case BOOLEAN:
            case CONTENT:
            case SUB_CONTENT:
            case DOUBLE:
            case USER:
            case REFERENCE:
                return true;
            case COMPOSITE:
            case BINARY:
            case FILE:
            case RICH_TEXT:
                return false;
            default:
                return false;
        }
    }
    
    /**
     * Test if the criterion can be set as a facet (i.e. it has a defined list of discrete values).
     * @return true if the criterion can be set as a facet, false otherwise.
     */
    default boolean isFacetable()
    {
        return isFacetable(getType(), getEnumerator() != null);
    }
    
    /**
     * Test if a field is facetable.
     * @param type the type of the metadata
     * @param isEnumerated is the metadata enumerated ?
     * @return true if the field can be used as a facet, false otherwise.
     */
    static boolean isFacetable(MetadataType type, boolean isEnumerated)
    {
        return type == MetadataType.CONTENT || type == MetadataType.SUB_CONTENT || type == MetadataType.USER || type == MetadataType.BOOLEAN
                || isEnumerated;
    }
    
    /**
     * Get the criterion Operator.
     * @return the criterion Operator.
     */
    Operator getOperator();
    
    /**
     * Get the {@link Query} associated to the given value.
     * @param value The user-submitted value (or the default value if not set) for this criterion.
     * @param allValues All the user-submitted values.
     * @param language The current search language.
     * @param contextualParameters the search contextual parameters.
     * @return {@link Query} associated to the given value.
    */
    default Query getQuery(Object value, Map<String, Object> allValues, String language, Map<String, Object> contextualParameters)
    {
        return getQuery(value, null, allValues, language, contextualParameters);
    }
    
    /**
     * Get the {@link Query} associated to the given value.
     * @param value The user-submitted value (or the default value if not set) for this criterion.
     * @param customOperator In advanced search mode, the operator chosen by the user. <code>null</code> to use the criterion-defined operator (simple search mode).
     * @param allValues All the user-submitted values.
     * @param language The current search language.
     * @param contextualParameters the search contextual parameters.
     * @return {@link Query} associated to the given value.
     */
    Query getQuery(Object value, Operator customOperator, Map<String, Object> allValues, String language, Map<String, Object> contextualParameters);
    
    /**
     * Get the {@link SearchField} representing this search criterion.
     * @return the {@link SearchField} representing this search criterion.
     */
    SearchField getSearchField();
}
