/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.model;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.MetadataType;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.user.population.UserPopulationDAO;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.i18n.I18nizableTextParameter;
import org.ametys.runtime.parameter.Enumerator;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Helper for {@link SearchCriterion}
 */
public class SearchCriterionHelper extends AbstractLogEnabled implements Component, Serviceable
{
    /** The component role. */
    public static final String ROLE = SearchCriterionHelper.class.getName();
    
    /** The ametys object resolver. */
    protected AmetysObjectResolver _resolver;
    /** The content type extension point */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    /** The searchModelHelper */
    protected SearchModelHelper _searchModelHelper;
    /** The user manager. */
    protected UserManager _userManager;
    /** The user population DAO */
    protected UserPopulationDAO _userPopulationDAO;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _searchModelHelper = (SearchModelHelper) manager.lookup(SearchModelHelper.ROLE);
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
        _userPopulationDAO = (UserPopulationDAO) manager.lookup(UserPopulationDAO.ROLE);
    }
    
    /**
     * Get the configuration of a criteria component.
     * @param originalConf the optional original criteria configuration.
     * @param baseContentTypeIds the "base" content type identifiers.
     * @param path the field path.
     * @param operator the optional criteria operator
     * @param group the optional group
     * @return the configuration to provide to the criteria component.
     * @throws ConfigurationException if an error occurs.
     */
    public Configuration getIndexingFieldCriteriaConfiguration(Optional<Configuration> originalConf, Set<String> baseContentTypeIds, String path, Optional<Operator> operator, Optional<I18nizableText> group) throws ConfigurationException
    {
        return _getIndexingFieldCriteriaConfiguration(Optional.empty(), originalConf, baseContentTypeIds, path, operator, group);
    }
    
    /**
     * Get the configuration of a criteria component.
     * @param searchModel the search model.
     * @param originalConf the optional original criteria configuration.
     * @param baseContentTypeIds the "base" content type identifiers.
     * @param path the field path.
     * @param operator the optional criteria operator
     * @param group the optional group
     * @return the configuration to provide to the criteria component.
     * @throws ConfigurationException if an error occurs.
     */
    public Configuration getIndexingFieldCriteriaConfiguration(SearchModel searchModel, Optional<Configuration> originalConf, Set<String> baseContentTypeIds, String path, Optional<Operator> operator, Optional<I18nizableText> group) throws ConfigurationException
    {
        return _getIndexingFieldCriteriaConfiguration(Optional.ofNullable(searchModel), originalConf, baseContentTypeIds, path, operator, group);
    }
    
    private Configuration _getIndexingFieldCriteriaConfiguration(Optional<SearchModel> searchModel, Optional<Configuration> originalConf, Set<String> baseContentTypeIds, String path, Optional<Operator> operator, Optional<I18nizableText> group) throws ConfigurationException
    {
        DefaultConfiguration conf = originalConf.isPresent()
                ? new DefaultConfiguration(originalConf.get())
                : new DefaultConfiguration("criteria");
        
        DefaultConfiguration metaConf = new DefaultConfiguration("field");
        conf.addChild(metaConf);
        metaConf.setAttribute("path", path);
        
        if (operator.isPresent())
        {
            DefaultConfiguration operatorConf = new DefaultConfiguration("test-operator");
            conf.addChild(operatorConf);
            operatorConf.setValue(operator.get().getName());
        }
        
        Configuration contentTypesConfiguration = _getContentTypesConfiguration(searchModel, baseContentTypeIds);
        conf.addChild(contentTypesConfiguration);
        
        if (group.isPresent())
        {
            Configuration groupConfiguration = _getGroupConfiguration(group.get());
            conf.addChild(groupConfiguration);
        }
        
        return conf;
    }
    
    /**
     * Get the configuration of a system criteria component.
     * @param originalConf the optional original criteria configuration.
     * @param baseContentTypeIds the "base" content type identifiers.
     * @param propertyId the system property identifier.
     * @param group The optional group.
     * @return the configuration to provide to the system criterion component.
     * @throws ConfigurationException if an error occurs.
     */
    public Configuration getSystemCriteriaConfiguration(Optional<Configuration> originalConf, Set<String> baseContentTypeIds, String propertyId, Optional<I18nizableText> group) throws ConfigurationException
    {
        return _getSystemCriteriaConfiguration(Optional.empty(), originalConf, baseContentTypeIds, propertyId, group);
    }
    
    /**
     * Get the configuration of a system criteria component.
     * @param searchModel the search model.
     * @param originalConf the optional original criteria configuration.
     * @param baseContentTypeIds the "base" content type identifiers.
     * @param propertyId the system property identifier.
     * @param group The optional group.
     * @return the configuration to provide to the system criterion component.
     * @throws ConfigurationException if an error occurs.
     */
    public Configuration getSystemCriteriaConfiguration(SearchModel searchModel, Optional<Configuration> originalConf, Set<String> baseContentTypeIds, String propertyId, Optional<I18nizableText> group) throws ConfigurationException
    {
        return _getSystemCriteriaConfiguration(Optional.ofNullable(searchModel), originalConf, baseContentTypeIds, propertyId, group);
    }
    
    private Configuration _getSystemCriteriaConfiguration(Optional<SearchModel> searchModel, Optional<Configuration> originalConf, Set<String> baseContentTypeIds, String propertyId, Optional<I18nizableText> group) throws ConfigurationException
    {
        DefaultConfiguration conf = originalConf.isPresent()
                ? new DefaultConfiguration(originalConf.get())
                : new DefaultConfiguration("criteria");
        
        DefaultConfiguration propConf = new DefaultConfiguration("systemProperty");
        propConf.setAttribute("name", propertyId);
        conf.addChild(propConf);
        
        Configuration contentTypesConfiguration = _getContentTypesConfiguration(searchModel, baseContentTypeIds);
        conf.addChild(contentTypesConfiguration);
        
        if (group.isPresent())
        {
            Configuration groupConfiguration = _getGroupConfiguration(group.get());
            conf.addChild(groupConfiguration);
        }
        
        return conf;
    }
    
    /**
     * Get the configuration of a custom criteria component.
     * @param searchModel the search model.
     * @param originalConf the original criteria configuration.
     * @param baseContentTypeIds the "base" content type identifiers.
     * @param customCriterionId the custom criterion id
     * @param group The group. Can be null.
     * @return the configuration to provide to the custom criterion component.
     * @throws ConfigurationException if an error occurs.
     */
    public Configuration getCustomCriteriaConfiguration(SearchModel searchModel, Optional<Configuration> originalConf, Set<String> baseContentTypeIds, String customCriterionId, Optional<I18nizableText> group) throws ConfigurationException
    {
        DefaultConfiguration conf = originalConf.isPresent()
                ? new DefaultConfiguration(originalConf.get())
                : new DefaultConfiguration("criteria");
        
        DefaultConfiguration customCriterionConf = new DefaultConfiguration("customCriterion");
        customCriterionConf.setAttribute("id", customCriterionId);
        conf.addChild(customCriterionConf);
        
        Configuration contentTypesConfiguration = _getContentTypesConfiguration(Optional.ofNullable(searchModel), baseContentTypeIds);
        conf.addChild(contentTypesConfiguration);
        
        if (group.isPresent())
        {
            Configuration groupConfiguration = _getGroupConfiguration(group.get());
            conf.addChild(groupConfiguration);
        }
        
        return conf;
    }
    
    private Configuration _getContentTypesConfiguration(Optional<SearchModel> searchModel, Set<String> baseContentTypeIds)
    {
        DefaultConfiguration contentTypesConf = new DefaultConfiguration("contentTypes");

        if (searchModel.isPresent())
        {
            Set<String> contentTypes = getAllContentTypes(searchModel.get(), Collections.emptyMap());
            for (String contentType : contentTypes)
            {
                DefaultConfiguration cTypeConf = new DefaultConfiguration("type");
                cTypeConf.setAttribute("id", contentType);
                contentTypesConf.addChild(cTypeConf);
            }
        }
        
        if (baseContentTypeIds != null)
        {
            for (String baseContentTypeId : baseContentTypeIds)
            {
                DefaultConfiguration cTypeConf = new DefaultConfiguration("baseType");
                cTypeConf.setAttribute("id", baseContentTypeId);
                contentTypesConf.addChild(cTypeConf);
            }
        }
        
        return contentTypesConf;
    }
    
    private Configuration _getGroupConfiguration(I18nizableText group)
    {
        DefaultConfiguration groupConf = new DefaultConfiguration("group");
        groupConf.setAttribute("i18n", group.isI18n());
        groupConf.setValue(group.isI18n() ? group.getCatalogue() + ":" + group.getKey() : group.getLabel());
        return groupConf;
    }
    

    
    /**
     * Get all the real content types that a model works on (the included content types, minus the excluded types).
     * @param model the search model.
     * @param contextualParameters the contextual parameters.
     * @return a Set of the content type IDs.
     */
    public Set<String> getAllContentTypes(SearchModel model, Map<String, Object> contextualParameters)
    {
        Set<String> allContentTypes = new HashSet<>();
        
        Set<String> modelCTypes = model.getContentTypes(contextualParameters);
        Set<String> modelExcludedCTypes = model.getExcludedContentTypes(contextualParameters);
        
        if (modelCTypes.isEmpty())
        {
            // Empty "declared" content types: the model works on all content types.
            allContentTypes.addAll(_contentTypeExtensionPoint.getExtensionsIds());
        }
        else
        {
            // Otherwise, add all declared content types and their sub-types.
            for (String cTypeId : modelCTypes)
            {
                allContentTypes.add(cTypeId);
                allContentTypes.addAll(_contentTypeExtensionPoint.getSubTypes(cTypeId));
            }
        }
        
        // Remove all excluded types.
        allContentTypes.removeAll(modelExcludedCTypes);
        
        return allContentTypes;
    }

    /**
     * Get the label of a facet value for the given criterion.
     * @param criterion the criterion
     * @param value the facet value.
     * @param currentLocale the current locale
     * @return the label, or null if the value does not exist.
     */
    public I18nizableText getFacetLabel(SearchCriterion criterion, String value, Locale currentLocale)
    {
        I18nizableText label = null;
        
        try
        {
            MetadataType type = criterion.getType();
            Enumerator enumerator = criterion.getEnumerator();
            
            if (type == MetadataType.CONTENT)
            {
                Content content = _resolver.resolveById(value);
                label = new I18nizableText(content.getTitle(currentLocale));
            }
            else if (type == MetadataType.USER)
            {
                UserIdentity userIdentity = UserIdentity.stringToUserIdentity(value);
                String login = userIdentity.getLogin();
                String populationId = userIdentity.getPopulationId();
                User user = _userManager.getUser(populationId, login);
                if (user != null)
                {
                    // Default i18n key only use login, sortablename and population.
                    // But we provide more parameters if the user want to override it.
                    Map<String, I18nizableTextParameter> i18nParams = new HashMap<>();
                    i18nParams.put("login", new I18nizableText(login));
                    i18nParams.put("firstname", new I18nizableText(user.getFirstName()));
                    i18nParams.put("lastname", new I18nizableText(user.getLastName()));
                    i18nParams.put("fullname", new I18nizableText(user.getFullName()));
                    i18nParams.put("sortablename", new I18nizableText(user.getSortableName()));
                    i18nParams.put("population", _userPopulationDAO.getUserPopulation(populationId).getLabel());
                    
                    label = new I18nizableText("plugin.cms", "PLUGINS_CMS_UITOOL_SEARCH_FACET_USER_LABEL", i18nParams);
                }
                else
                {
                    Map<String, I18nizableTextParameter> i18nParams = new HashMap<>();
                    i18nParams.put("login", new I18nizableText(login));
                    i18nParams.put("population", new I18nizableText(populationId));
                    
                    label = new I18nizableText("plugin.cms", "PLUGINS_CMS_UITOOL_SEARCH_FACET_UNKNOWN_USER_LABEL", i18nParams);
                }
            }
            else if (enumerator != null)
            {
                label = enumerator.getEntry(value);
            }
            else if (type == MetadataType.BOOLEAN)
            {
                boolean boolValue = "1".equals(value) /* if joined facet, value will be "0" or "1" */ || Boolean.valueOf(value);
                label = new I18nizableText("plugin.cms", boolValue ? "PLUGINS_CMS_UITOOL_SEARCH_FACET_BOOLEAN_TRUE_LABEL" : "PLUGINS_CMS_UITOOL_SEARCH_FACET_BOOLEAN_FALSE_LABEL");
            }
        }
        catch (Exception e)
        {
            // Ignore, just return null.
            getLogger().warn("Unable to get facel label for value '{}'", value, e);
        }
        
        return label;
    }
}
