/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.model;

import java.util.Map;
import java.util.Set;

import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.ViewHelper;
import org.ametys.runtime.model.ViewItemContainer;

/**
 * This interface represents a search model.
 */
public interface SearchModel
{
    /**
     * Get the list of content types.
     * @param contextualParameters the contextual parameters.
     * @return The list of content types.
     */
    Set<String> getContentTypes(Map<String, Object> contextualParameters);
    
    /**
     * Get the list of excluded content types.
     * @param contextualParameters the contextual parameters
     * @return The list of excluded content types.
     */
    Set<String> getExcludedContentTypes(Map<String, Object> contextualParameters);
    
    /**
     * Get the list of search criteria in simple mode
     * @param contextualParameters the contextual parameters
     * @return the list of search criteria in simple mode
     */
    Map<String, ? extends SearchCriterion> getCriteria(Map<String, Object> contextualParameters);
    
    /**
     * Get a simple search criterion by its id
     * @param id The search criterion id
     * @param contextualParameters the contextual parameters
     * @return the criterion or <code>null</code> if not found
     */
    default SearchCriterion getCriterion(String id, Map<String, Object> contextualParameters)
    {
        return getCriteria(contextualParameters).get(id);
    }
    
    /**
     * Get the list of faceted search criteria.
     * @param contextualParameters the contextual parameters
     * @return the list of faceted search criteria.
     */
    Map<String, ? extends SearchCriterion> getFacetedCriteria(Map<String, Object> contextualParameters);
    
    /**
     * Get a faceted search criterion by its id.
     * @param id The faceted search criterion id.
     * @param contextualParameters the contextual parameters
     * @return the faceted criterion or <code>null</code> if not found
     */
    default SearchCriterion getFacetedCriterion(String id, Map<String, Object> contextualParameters)
    {
        return getFacetedCriteria(contextualParameters).get(id);
    }
    
    /**
     * Retrieves the search result items 
     * @param contextualParameters the contextual parameters
     * @return the search result items
     */
    ViewItemContainer getResultItems(Map<String, Object> contextualParameters);
    
    /**
     * Retrieves the search result item with the given path
     * @param itemPath the path of the result item to retrieve
     * @param contextualParameters the contextual parameters
     * @return the search result item with
     */
    default ModelViewItem getResultItem(String itemPath, Map<String, Object> contextualParameters)
    {
        ViewItemContainer resultItems = getResultItems(contextualParameters);
        return ViewHelper.getModelViewItem(resultItems, itemPath);
    }
    
    /**
     * Get the specific workspace to use.
     * @param contextualParameters the contextual parameters.
     * @return the workspace to use when searching, or null to use the default workspace.
     */
    String getWorkspace(Map<String, Object> contextualParameters);
}
