/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.ProcessingException;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.search.ui.model.DynamicWrappedSearchUIModel;
import org.ametys.cms.search.ui.model.SearchUIColumn;
import org.ametys.cms.search.ui.model.SearchUICriterion;
import org.ametys.cms.search.ui.model.SearchUIModel;
import org.ametys.cms.search.ui.model.SearchUIModelExtensionPoint;
import org.ametys.cms.search.ui.model.impl.DefaultSearchUIModel;
import org.ametys.cms.search.ui.model.impl.DefaultSolrFilterSearchUICriterion;
import org.ametys.cms.search.ui.model.impl.SystemSearchUICriterion;
import org.ametys.core.ui.Callable;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.DefinitionContext;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.parameter.Enumerator;
import org.ametys.runtime.parameter.ParameterHelper;
import org.ametys.runtime.parameter.Validator;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Helper for {@link SearchModel}.
 */
public class SearchModelHelper extends AbstractLogEnabled implements Component, Serviceable, Contextualizable
{
    /** The component role. */
    public static final String ROLE = SearchModelHelper.class.getName();
    
    private static final String __SOLR_REQUEST_PARAMETER_NAME = "solrRequest"; 
    private static final String __SOLR_REQUEST_CRITERION_ID = "solr-filter-criterion";
    
    private ServiceManager _serviceManager;
    private SearchUIModelExtensionPoint _searchUIModelExtensionPoint;
    private ContentTypeExtensionPoint _contentTypeExtensionPoint;
    private Context _context;
    
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _serviceManager = manager;
        _searchUIModelExtensionPoint = (SearchUIModelExtensionPoint) manager.lookup(SearchUIModelExtensionPoint.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    /**
     * Get the search model configuration as JSON object
     * @param modelId The id of search model
     * @param restrictedContentTypes The restricted content types. Can be null.
     * @param contextualParameters the contextual parameters
     * @return The search model configuration in a Map
     * @throws ProcessingException if an error occurred
     */
    @Callable
    public Map<String, Object> getSearchModelConfiguration(String modelId, List<String> restrictedContentTypes, Map<String, Object> contextualParameters) throws ProcessingException
    {
        SearchUIModel model = getSearchUIModel(modelId, restrictedContentTypes, contextualParameters);
        return getSearchModelInfo(model, contextualParameters);
    }
    
    /**
     * Get the column configurations of search model as JSON object
     * @param modelId The id of search model
     * @param restrictedContentTypes The restricted content types. Can be null.
     * @param contextualParameters the contextual parameters
     * @return The column configurations in a List
     * @throws ProcessingException if an error occurred
     */
    @Callable
    public List<Object> getColumnConfigurations(String modelId, List<String> restrictedContentTypes, Map<String, Object> contextualParameters) throws ProcessingException
    {
        SearchUIModel model = getSearchUIModel(modelId, restrictedContentTypes, contextualParameters);
        return getColumnsInfo(model.getResultItems(contextualParameters));
    }
    
    /**
     * Retrieves the {@link SearchUIModel} with the given model identifier, with restrictions on content types
     * @param modelId the model identifier
     * @param restrictedContentTypes the restricted content types
     * @param contextualParameters the contextual parameters
     * @return the {@link SearchUIModel}
     */
    public SearchUIModel getSearchUIModel(String modelId, List<String> restrictedContentTypes, Map<String, Object> contextualParameters)
    {
        SearchUIModel model = _searchUIModelExtensionPoint.getExtension(modelId);
        if (model == null)
        {
            throw new IllegalArgumentException("The search model '" + modelId + "' does not exist");
        }
        
        // TODO Replace DynamicWrappedSearchUIModel?
        if (restrictedContentTypes != null)
        {
            model = new DynamicWrappedSearchUIModel(model, restrictedContentTypes, _contentTypeExtensionPoint, getLogger(), _context, _serviceManager);
        }
        
        Optional<String> solrRequest = Optional.ofNullable(contextualParameters.get(__SOLR_REQUEST_PARAMETER_NAME))
                                               .filter(String.class::isInstance)
                                               .map(String.class::cast)
                                               .filter(StringUtils::isNotEmpty);
        if (solrRequest.isPresent())
        {
            DefaultSearchUIModel modelCopy = new DefaultSearchUIModel(model, contextualParameters);
            
            // Create a criterion with solr request
            DefaultSolrFilterSearchUICriterion criterion = new DefaultSolrFilterSearchUICriterion();
            criterion.setId(__SOLR_REQUEST_CRITERION_ID);
            criterion.setQuery(solrRequest.get());
            
            // Add the criterion to the search model
            modelCopy.addCriterion(criterion);
            
            model = modelCopy;
        }
        
        return model;
    }
    
    /**
     * Return information on a {@link SearchUIModel} object serialized in a Map.
     * @param model The search model.
     * @param contextualParameters The contextual parameters
     * @return the detailed information serialized in a Map.
     * @throws ProcessingException if an error occurs.
     */
    public Map<String, Object> getSearchModelInfo(SearchUIModel model, Map<String, Object> contextualParameters) throws ProcessingException
    {
        Map<String, Object> jsonObject = new HashMap<>();
        
        jsonObject.put("pageSize", model.getPageSize(contextualParameters));
        jsonObject.put("workspace", model.getWorkspace(contextualParameters));
        jsonObject.put("searchUrl", model.getSearchUrl(contextualParameters));
        jsonObject.put("searchUrlPlugin", model.getSearchUrlPlugin(contextualParameters));
        jsonObject.put("exportCSVUrl", model.getExportCSVUrl(contextualParameters));
        jsonObject.put("exportCSVUrlPlugin", model.getExportCSVUrlPlugin(contextualParameters));
        jsonObject.put("exportDOCUrl", model.getExportDOCUrl(contextualParameters));
        jsonObject.put("exportDOCUrlPlugin", model.getExportDOCUrlPlugin(contextualParameters));
        jsonObject.put("exportXMLUrl", model.getExportXMLUrl(contextualParameters));
        jsonObject.put("exportXMLUrlPlugin", model.getExportXMLUrlPlugin(contextualParameters));
        jsonObject.put("exportPDFUrl", model.getExportPDFUrl(contextualParameters));
        jsonObject.put("exportPDFUrlPlugin", model.getExportPDFUrlPlugin(contextualParameters));
        jsonObject.put("printUrl", model.getPrintUrl(contextualParameters));
        jsonObject.put("printUrlPlugin", model.getPrintUrlPlugin(contextualParameters));
        jsonObject.put("summaryView", model.getSummaryView());
        
        jsonObject.put("simple-criteria", getCriteriaListInfo(model.getCriteria(contextualParameters)));
        jsonObject.put("advanced-criteria", getAdvancedCriteriaListInfo(model.getAdvancedCriteria(contextualParameters)));
        
        
        jsonObject.put("columns", getColumnsInfo(model.getResultItems(contextualParameters)));
        
        jsonObject.put("hasFacets", !model.getFacetedCriteria(contextualParameters).isEmpty());
        
        return jsonObject;
    }
    
    /**
     * Return information of columns in the given view item accessor, serialized as a Map.
     * @param viewItemAccessor the view item accessor containing the columns.
     * @return the detailed information serialized in a Map.
     * @throws ProcessingException if an error occurs.
     */
    public List<Object> getColumnsInfo(ViewItemAccessor viewItemAccessor) throws ProcessingException
    {
        List<Object> jsonObject = new ArrayList<>();
        
        for (ViewItem viewItem : viewItemAccessor.getViewItems())
        {
            if (viewItem instanceof SearchUIColumn searchUIColumn)
            {
                jsonObject.add(searchUIColumn.toJSON(DefinitionContext.newInstance()));
            }
            else if (viewItem instanceof ViewItemAccessor itemAccessor)
            {
                jsonObject.addAll(getColumnsInfo(itemAccessor));
            }
        }
        
        return jsonObject;
    }
    
    /**
     * Return information on a list of {@link SearchUICriterion}, serialized as a Map.
     * @param criteria a map of search criteria.
     * @return the detailed information serialized in a Map.
     * @throws ProcessingException if an error occurs.
     */
    public Map<String, Object> getCriteriaListInfo(Map<String, ? extends SearchUICriterion> criteria) throws ProcessingException
    {
        Map<String, Object> jsonObject = new LinkedHashMap<>();
        
        Map<I18nizableText, List<SearchUICriterion>> criteriaByGroup = _getCriteriaByGroup(criteria);
        
        for (I18nizableText group : criteriaByGroup.keySet())
        {
            Map<String, Object> elements = new LinkedHashMap<>();
            
            for (SearchUICriterion sc : criteriaByGroup.get(group))
            {
                elements.put(sc.getId(), getCriterionInfo(sc));
            }
            
            Map<String, Object> groupAsJSON = new LinkedHashMap<>();
            if (group != null)
            {
                groupAsJSON.put("label", group);
            }
            groupAsJSON.put("role", "fieldset");
            groupAsJSON.put("elements", elements);
            
            String groupUUID = UUID.randomUUID().toString();
            jsonObject.put(groupUUID, groupAsJSON);
        }
        return jsonObject;
    }
    
    /**
     * Return information on a list of advanced {@link SearchUICriterion}, serialized as a Map.
     * @param criteria A map of advanced search criteria.
     * @return the detailed information serialized in a Map.
     * @throws ProcessingException if an error occurs.
     */
    public Map<String, Object> getAdvancedCriteriaListInfo(Map<String, ? extends SearchUICriterion> criteria) throws ProcessingException
    {
        // No groups for advanced search
        Map<String, Object> jsonObject = new LinkedHashMap<>();
        Map<String, Object> criteriaObject = new LinkedHashMap<>();
        
        for (SearchUICriterion sc : criteria.values())
        {
            if (sc instanceof SystemSearchCriterion)
            {
                SystemSearchUICriterion sysCrit = (SystemSearchUICriterion) sc;
                if (sysCrit.getSystemPropertyId().equals("contentLanguage"))
                {
                    // Separate the language from the other criteria since it is mandatory
                    jsonObject.put("language", getCriterionInfo(sc));
                }
                else
                {
                    criteriaObject.put(sc.getId(), getCriterionInfo(sc));
                }
            }
            else
            {
                criteriaObject.put(sc.getId(), getCriterionInfo(sc));
            }
        }
        
        if (MapUtils.isNotEmpty(criteriaObject))
        {
            jsonObject.put("criteria", criteriaObject);
        }
        
        return jsonObject;
    }
    
    /**
     * Return information on a {@link SearchUICriterion}, serialized as a Map.
     * @param criterion a search criterion.
     * @return the detailed information serialized in a Map.
     * @throws ProcessingException if an error occurs.
     */
    public Map<String, Object> getCriterionInfo(SearchUICriterion criterion) throws ProcessingException
    {
        Map<String, Object> jsonObject = new HashMap<>();
        
        // TODO Parameter interface
        _putCriterionParameter(jsonObject, criterion);
        
        jsonObject.put("multiple", criterion.isMultiple());
        jsonObject.put("hidden", criterion.isHidden());
        
        String contentTypeId = criterion.getContentTypeId();
        if (contentTypeId != null)
        {
            jsonObject.put("contentType", contentTypeId);
        }
        
        jsonObject.put("criterionProperty", criterion.getFieldId());
        // TODO use operator.getName()?
        if (criterion.getOperator() != null)
        {
            jsonObject.put("criterionOperator", criterion.getOperator().toString().toLowerCase());
        }
        
//        if (criterion.getValue() != null)
//        {
//            jsonObject.put("value", criterion.getValue());
//        }
        
        return jsonObject;
    }
    
    private Map<I18nizableText, List<SearchUICriterion>> _getCriteriaByGroup(Map<String, ? extends SearchUICriterion> criteria)
    {
        Map<I18nizableText, List<SearchUICriterion>> criteriaByGroup = new LinkedHashMap<>();
        
        for (SearchUICriterion sc : criteria.values())
        {
            I18nizableText group = sc.getGroup();
            
            if (!criteriaByGroup.containsKey(group))
            {
                criteriaByGroup.put(group, new ArrayList<>());
            }
            
            criteriaByGroup.get(group).add(sc);
        }
        
        return criteriaByGroup;
    }
    
    private void _putCriterionParameter(Map<String, Object> jsonObject, SearchUICriterion criterion) throws ProcessingException
    {
        jsonObject.put("id", criterion.getId());
        jsonObject.put("label", criterion.getLabel());
        jsonObject.put("description", criterion.getDescription());
        jsonObject.put("type", criterion.getType().getId());
        
        Object defaultValue = criterion.getDefaultValue();
        if (defaultValue != null)
        {
            jsonObject.put("default-value", defaultValue);
        }
        
        _putEnumerator(jsonObject, criterion.getEnumerator());
        _putValidator(jsonObject, criterion.getValidator());
        
        String widget = criterion.getWidget();
        if (widget != null)
        {
            jsonObject.put("widget", widget);
        }
        
        Map<String, I18nizableText> widgetParameters = criterion.getWidgetParameters();
        if (widgetParameters != null && !widgetParameters.isEmpty())
        {
            jsonObject.put("widget-params", criterion.getWidgetParameters());
        }
    }
    
    private void _putEnumerator(Map<String, Object> jsonObject, Enumerator enumerator) throws ProcessingException
    {
        if (enumerator != null)
        {
            try
            {
                List<Map<String, Object>> options = new ArrayList<>();
                
                for (Map.Entry<Object, I18nizableText> entry : enumerator.getEntries().entrySet())
                {
                    String valueAsString = ParameterHelper.valueToString(entry.getKey());
                    I18nizableText entryLabel = entry.getValue();
                    
                    Map<String, Object> option = new HashMap<>();
                    option.put("label", entryLabel != null ? entryLabel : valueAsString);
                    option.put("value", valueAsString);
                    options.add(option);
                }
                
                jsonObject.put("enumeration", options);
                jsonObject.put("enumerationConfig", enumerator.getConfiguration());
            }
            catch (Exception e)
            {
                throw new ProcessingException("Unable to enumerate entries with enumerator: " + enumerator, e);
            }
        }
    }
    
    private void _putValidator(Map<String, Object> jsonObject, Validator validator)
    {
        if (validator != null)
        {
            jsonObject.put("validation", validator.getConfiguration());
        }
    }
}
