/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.model;

import java.util.Collection;
import java.util.Map;

import org.ametys.cms.data.ametysobject.ModelAwareDataAwareAmetysObject;
import org.ametys.cms.model.properties.Property;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.solr.schema.SchemaDefinition;

/**
 * Represents a universal property of a supported {@link ModelAwareDataAwareAmetysObject}
 * Ex: content types, language or current workflow steps are some system properties supporting contents
 * @param <T> type of the property values 
 * @param <X> type of ametys object supported by this property
 */
public interface SystemProperty<T, X extends ModelAwareDataAwareAmetysObject> extends Property<T, X>
{
    /**
     * Get if the property can be searched on (i.e. used in a SearchCriterion).
     * @return <code>true</code> if the property can be searched on, <code>false</code> otherwise.
     */
    public default boolean isSearchable()
    {
        // Default to true: override when the property is not searchable.
        return true;
    }
    
    /**
     * Get if the property can be displayed (i.e. used in a ResultField).
     * @return <code>true</code> if the property can be displayed, <code>false</code> otherwise.
     */
    public default boolean isDisplayable()
    {
        // Default to true: override when the property is not displayable.
        return true;
    }
    
    /**
     * Get if the property can be sorted on.
     * @return <code>true</code> if the property can be sorted on, <code>false</code> otherwise.
     */
    public default boolean isSortable()
    {
        // Default to true: override when the property is not sortable.
        return true;
    }
    
    /**
     * Gets if the property can be facetable
     * @return <code>true</code> if the property can be facetable, <code>false</code> otherwise.
     */
    public default boolean isFacetable()
    {
        SearchField searchField = getSearchField();
        return searchField != null ? searchField.getFacetField() != null : false;
    }
    
    /**
     * Get the {@link Query} associated to the given value.
     * @param value the user-submitted value for this property.
     * @param operator In advanced search mode, the operator chosen by the user. <code>null</code> to use the criterion-defined operator (simple search mode).
     * @param language The current search language.
     * @param contextualParameters the search contextual parameters.
     * @return The {@link Query} associated to the given value.
     */
    public Query getQuery(Object value, Operator operator, String language, Map<String, Object> contextualParameters);
    
    /**
     * Get the renderer.
     * @return The column renderer.
     */
    public default String getRenderer()
    {
        return null;
    }
    
    /**
     * Get the property column converter.
     * @return The property column converter.
     */
    public default String getConverter()
    {
        return null;
    }
    
    /**
     * Get the column width.
     * @return the default column width, can be null.
     */
    public default Integer getColumnWidth()
    {
        return null;
    }
    
    /**
     * Get the {@link SearchField} representing this system property.
     * @return The search field representing this system property.
     */
    public SearchField getSearchField();
    
    /**
     * Get the sort value represented by this field in the given result ametys object.
     * @param ametysObject the result ametys object.
     * @return the content sort value, must be scalar (String, long, double, Date).
     */
    public default Object getSortValue(X ametysObject)
    {
        // Not sortable by default;
        return null;
    }
    
    /**
     * Get the schema definitions brought by this property.
     * @return The schema definitions used by this property.
     */
    public Collection<SchemaDefinition> getSchemaDefinitions();
}
