/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.avalon.framework.component.ComponentException;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.runtime.plugin.component.AbstractThreadSafeComponentExtensionPoint;

/**
 * Extension point handling {@link SystemProperty} objects.
 */
public class SystemPropertyExtensionPoint extends AbstractThreadSafeComponentExtensionPoint<SystemProperty>// implements Initializable
{
    /** The extension point role. */
    public static final String ROLE = SystemPropertyExtensionPoint.class.getName();
    
    /** Content properties types extension point */
    protected ModelItemTypeExtensionPoint _contentPropertyTypeExtensionPoint;
    
    /** Collection of the displayable properties. */
    protected List<String> _displayProps;
    
    /** Collection of the searchable properties. */
    protected List<String> _searchProps;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _contentPropertyTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.ROLE_CONTENT_PROPERTY);
    }
    
    @Override
    public void addExtension(String id, String pluginName, String featureName, Configuration configuration) throws ConfigurationException
    {
        if (getLogger().isDebugEnabled())
        {
            getLogger().debug("Adding system properties from feature " + pluginName + "/" + featureName);
        }
        
        try
        {
            for (Configuration propConf : configuration.getChildren("property"))
            {
                addSystemProperty(pluginName, featureName, propConf);
            }
        }
        catch (ConfigurationException e)
        {
            if (getLogger().isWarnEnabled())
            {
                getLogger().warn("The feature '" + pluginName + "/" + featureName + "' declares a system property extension with an invalid configuration.", e);
            }
        }
    }
    
    /**
     * Declare a new system property.
     * @param pluginName The name of the plugin declaring the extension
     * @param featureName The feature name.
     * @param configuration The configuration of the extension
     * @throws ConfigurationException if configuration if not complete
     */
    @SuppressWarnings("unchecked")
    protected void addSystemProperty(String pluginName, String featureName, Configuration configuration) throws ConfigurationException
    {
        String id = configuration.getAttribute("id", "");
        if (StringUtils.isBlank(id))
        {
            throw new ConfigurationException("SystemProperty declaration is incorrect since no 'id' attribute is specified (or may be empty)", configuration);
        }
        
        String className = configuration.getAttribute("class", "");
        if (StringUtils.isBlank(className))
        {
            throw new ConfigurationException("SystemProperty declaration is incorrect since no 'class' attribute is specified (or may be empty)", configuration);
        }
        
        try
        {
            Class<?> clazz = Class.forName(className);
            if (!SystemProperty.class.isAssignableFrom(clazz))
            {
                throw new ConfigurationException("The system property declaration must be declared with a class implementing SystemProperty.");
            }
            
            addComponent(pluginName, featureName, id, (Class<? extends SystemProperty>) clazz, configuration);
        }
        catch (ComponentException | ClassNotFoundException e)
        {
            throw new ConfigurationException("Unable to instanciate the SystemProperty of id '" + id + "'.", e);
        }
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public void initializeExtensions() throws Exception
    {
        super.initializeExtensions();
        
        _displayProps = new ArrayList<>();
        _searchProps = new ArrayList<>();
        
        for (String id : getExtensionsIds())
        {
            SystemProperty prop = getExtension(id);
            prop.setAvailableTypeExtensionPoint(_contentPropertyTypeExtensionPoint);
            
            if (prop.isDisplayable())
            {
                _displayProps.add(id);
            }
            if (prop.isSearchable())
            {
                _searchProps.add(id);
            }
        }
    }
    
    /**
     * Get the IDs of the properties that can be displayed.
     * @return The Collection of properties that can be displayed.
     */
    public List<String> getDisplayProperties()
    {
        return Collections.unmodifiableList(_displayProps);
    }
    
    /**
     * Get the IDs of the properties that can be searched on.
     * @return The Collection of properties that can be searched on.
     */
    public List<String> getSearchProperties()
    {
        return Collections.unmodifiableList(_searchProps);
    }
    
    /**
     * Tests if a property exists and is displayable.
     * @param name The property name.
     * @return <code>true</code> if the property exists and is displayable.
     */
    public boolean isDisplayable(String name)
    {
        return _displayProps.contains(name);
    }
    
    /**
     * Tests if a property exists and is searchable.
     * @param name The property name.
     * @return <code>true</code> if the property exists and is searchable.
     */
    public boolean isSearchable(String name)
    {
        return _searchProps.contains(name);
    }

}
