/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import org.apache.solr.client.solrj.util.ClientUtils;

/**
 * Base class for multivalued queries.
 * @param <T> the value type.
 */
public abstract class AbstractMultivaluedQuery<T> extends AbstractOperatorQuery<List<T>>
{
    private LogicalOperator _logicalOperator; 
    
    /**
     * Create a new Query
     * @param fieldName the Solr field name
     * @param operator the {@link org.ametys.cms.search.query.Query.Operator}
     * @param logicalOperator the logical operator, default is LogicalOperator.OR
     * @param values the values
     */
    public AbstractMultivaluedQuery(String fieldName, Operator operator, LogicalOperator logicalOperator, List<T> values)
    {
        super(fieldName, operator, values);
        _logicalOperator = Optional.ofNullable(logicalOperator).orElse(LogicalOperator.OR);
        
        if (Operator.EQ != operator && Operator.NE != operator && Operator.EXISTS != operator)
        {
            throw new IllegalArgumentException("Test operator '" + operator + "' is unknown for test's expression.");
        }
    }
    
    /**
     * Computes the value for Solr client
     * @param value the typed value
     * @return the value, adapted for Solr
     */
    public String singleValueForQuery(T value)
    {
        return ClientUtils.escapeQueryChars(value.toString());
    }
    
    @Override
    public String build() throws QuerySyntaxException
    {
        List<T> values = getValue();
        StringBuilder query = new StringBuilder();

        if (getOperator() == Operator.EXISTS)
        {
            query.append(getFieldName()).append(":").append(QueryHelper.EXISTS_VALUE);
            
            return query.toString();
        }
        
        if (getOperator() == Operator.NE)
        {
            NotQuery.appendNegation(query);
        }
        
        if (values.contains(""))
        {
            query.append(getFieldName()).append(":*");
        }
        else
        {
            int count = values.size();
            if (count > 1)
            {
                query.append('(');
            }

            boolean first = true;
            for (T v : values)
            {
                if (!first)
                {
                    query.append(" " + _logicalOperator + " ");
                }
                
                query.append(getFieldName()).append(':').append(singleValueForQuery(v));
                first = false;
            }
            
            if (count > 1)
            {
                query.append(')');
            }
        }
        
        return query.toString();
    }

    public Optional<Object> buildAsJson() throws QuerySyntaxException
    {
        if (getOperator() == Operator.EXISTS)
        {
            return Optional.of(getFieldName() + ":" + QueryHelper.EXISTS_VALUE);
        }
        
        List<String> clauses = getValue().stream().map(v -> getFieldName() + ":" + singleValueForQuery(v)).toList();
        
        if (clauses.isEmpty())
        {
            return new MatchAllQuery().buildAsJson();
        }
        else if (clauses.size() == 1)
        {
            return Optional.of(clauses.get(0));
        }
        
        return Optional.of(Map.of("bool", Map.of(getOperator() == Operator.EQ ? "should" : "must_not", clauses)));
    }
    
    @Override
    public int hashCode()
    {
        return 31 * super.hashCode() + Objects.hash(_logicalOperator);
    }

    @Override
    public boolean equals(Object obj)
    {
        return super.equals(obj) && Objects.equals(_logicalOperator, ((AbstractMultivaluedQuery) obj)._logicalOperator);
    }
}
