/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.Objects;
import java.util.Optional;

/**
 * Base class for all operator-based queries.
 * @param <T> the value type.
 */
public abstract class AbstractOperatorQuery<T> implements Query, Cloneable
{
    private Operator _operator;
    private String _fieldName;
    private T _value;
    
    /**
     * Create a new Query
     * @param fieldName the Solr field name
     * @param operator the {@link org.ametys.cms.search.query.Query.Operator}
     * @param value the value
     */
    public AbstractOperatorQuery(String fieldName, Operator operator, T value)
    {
        _fieldName = fieldName;
        _operator = operator;
        _value = value;
    }
    
    /**
     * Return the {@link org.ametys.cms.search.query.Query.Operator}
     * @return the operator
     */
    public Operator getOperator()
    {
        return _operator;
    }
    
    /**
     * Returns the typed value
     * @return the value
     */
    public T getValue()
    {
        return _value;
    }
    
    /**
     * Returns the Solr field name.
     * @return the field name
     */
    public String getFieldName()
    {
        return _fieldName;
    }
    
    /**
     * Computes the value for Solr client
     * @param value the typed value
     * @return the value, adapted for Solr
     */
    public String valueForQuery(T value)
    {
        return value != null ? value.toString() : null;
    }

    @Override
    public String build() throws QuerySyntaxException
    {
        return QueryHelper.getStandardQuery(getFieldName(), getOperator(), valueForQuery(getValue()));
    }
    
    public Optional<Query> rewrite()
    {
        if (_operator == Operator.NE)
        {
            try
            {
                @SuppressWarnings("unchecked")
                AbstractOperatorQuery<T> query = (AbstractOperatorQuery<T>) clone();
                
                query._operator = Operator.EQ;
                return Optional.of(new NotQuery(query));
            }
            catch (CloneNotSupportedException e)
            {
                throw new IllegalArgumentException("This query could not be cloned", e);
            }
        }
        
        return Optional.of(this);
    }
    
    @Override
    public int hashCode()
    {
        return Objects.hash(_fieldName, _operator, _value);
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        
        if (obj == null || getClass() != obj.getClass())
        {
            return false;
        }
        
        AbstractOperatorQuery other = (AbstractOperatorQuery) obj;
        return Objects.equals(_fieldName, other._fieldName)
            && Objects.equals(_value, other._value)
            && Objects.equals(_operator, other._operator);
    }
}
