/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.Objects;

/**
 * Base class for all range-based queries.
 * @param <T> the value type.
 */
public abstract class AbstractRangeQuery<T> implements Query
{
    private String _fieldName;
    private T _lower;
    private T _upper;
    private boolean _includeLower;
    private boolean _includeUpper;

    /**
     * Build a range query.
     * @param fieldName The field path.
     * @param lowerBound The lower end of the range.
     * @param upperBound The upper end of the range.
     * @param includeLower Whether to include the lower end or not.
     * @param includeUpper Whether to include the upper end or not.
     */
    public AbstractRangeQuery(String fieldName, T lowerBound, T upperBound, boolean includeLower, boolean includeUpper)
    {
        _fieldName = fieldName;
        _lower = lowerBound;
        _upper = upperBound;
        _includeLower = includeLower;
        _includeUpper = includeUpper;
    }
    
    /**
     * Returns the Solr field name.
     * @return the field name
     */
    public String getFieldName()
    {
        return _fieldName;
    }
    
    /**
     * Returns true if the query should include the lower bound.
     * @return true if the query should include the lower bound.
     */
    public boolean includeLowerBound()
    {
        return _includeLower;
    }
    
    /**
     * Returns true if the query should include the upper bound.
     * @return true if the query should include the upper bound.
     */
    public boolean includeUpperBound()
    {
        return _includeUpper;
    }
    
    /**
     * Returns the lower bound of the range.
     * @return the lower bound of the range.
     */
    public T getLowerBound()
    {
        return _lower;
    }
    
    /**
     * Returns the upper bound of the range.
     * @return the upper bound of the range.
     */
    public T getUpperBound()
    {
        return _upper;
    }
    
    /**
     * Computes the lower bound for Solr client
     * @param value the typed value
     * @return the value, adapted for Solr
     */
    public String lowerBoundForQuery(T value)
    {
        return value.toString();
    }
    
    /**
     * Computes the upper bound for Solr client
     * @param value the typed value
     * @return the value, adapted for Solr
     */
    public String upperBoundForQuery(T value)
    {
        return value.toString();
    }
    
    public String build() throws QuerySyntaxException
    {
        StringBuilder query = new StringBuilder();
        
        query.append(getFieldName())
             .append(includeLowerBound() ? '[' : '{')
             .append(lowerBoundForQuery(getLowerBound())).append(" TO ").append(upperBoundForQuery(getUpperBound()))
             .append(includeUpperBound() ? ']' : '}');
        
        return query.toString();
    }

    @Override
    public int hashCode()
    {
        return Objects.hash(_fieldName, _lower, _upper, _includeLower, _includeUpper);
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        
        if (obj == null || getClass() != obj.getClass())
        {
            return false;
        }
        
        AbstractRangeQuery other = (AbstractRangeQuery) obj;
        return Objects.equals(_fieldName, other._fieldName)
                && Objects.equals(_lower, other._lower)
                && Objects.equals(_upper, other._upper)
                && Objects.equals(_includeLower, other._includeLower)
                && Objects.equals(_includeUpper, other._includeUpper);
    }
}
