/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.Collections;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.apache.solr.client.solrj.util.ClientUtils;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * Represents a {@link Query} testing a text field.
 */
public abstract class AbstractTextQuery extends AbstractOperatorQuery<String>
{
    /** The language. */
    protected String _language;
    /** <code>true</code> if the value is already escaped and there is no need to escape again the value during {@link #build() the build of the query}. */
    protected boolean _valueAlreadyEscaped;
    
    /**
     * Build a AbstractTextQuery testing the existence of the field.
     * @param fieldPath the field path
     */
    public AbstractTextQuery(String fieldPath)
    {
        this(fieldPath, Operator.EXISTS, null, null);
    }
    
    /**
     * Build a text query.
     * @param fieldPath the field's path
     * @param value the value.
     */
    public AbstractTextQuery(String fieldPath, String value)
    {
        this(fieldPath, value, null);
    }
    
    /**
     * Build a text query.
     * @param fieldPath the field's path
     * @param value the value.
     * @param language the query language (can be null).
     */
    public AbstractTextQuery(String fieldPath, String value, String language)
    {
        this(fieldPath, Operator.EQ, value, language);
    }
    
    /**
     * Build a text query.
     * @param fieldPath the field's path
     * @param op the operator.
     * @param value the value.
     * @param language the query language (can be null).
     */
    public AbstractTextQuery(String fieldPath, Operator op, String value, String language)
    {
        this(fieldPath, op, value, language, false);
    }
    
    /**
     * Build a text query.
     * @param fieldPath the field's path
     * @param op the operator.
     * @param value the value.
     * @param language the query language (can be null).
     * @param alreadyEscaped true if the value is already escaped and there is no need to escape again the value during {@link #build() the build of the query}.
     */
    public AbstractTextQuery(String fieldPath, Operator op, String value, String language, boolean alreadyEscaped)
    {
        super(fieldPath, op, value);
        _language = language;
        _valueAlreadyEscaped = alreadyEscaped;
    }
    
    /**
     * Get the language.
     * @return the language.
     */
    public String getLanguage()
    {
        return _language;
    }
    
    @Override
    public String build() throws QuerySyntaxException
    {
        Operator operator = getOperator();
        String value = getValue();
        
        StringBuilder query = new StringBuilder();
                
        checkStringValue(value);
        
        String escapedValue = _valueAlreadyEscaped ? value : escapeStringValue(value, operator);
        
        if (operator == Operator.NE)
        {
            NotQuery.appendNegation(query);
        }
        
        query.append(getFieldName());
        
        if (operator == Operator.SEARCH)
        {
            // Test query, unstemmed.
            query.append("_txt");
            if (_language != null)
            {
                query.append("_").append(_language);
            }
            query.append(":(").append(escapedValue).append(')');
        }
        else if (operator == Operator.SEARCH_STEMMED)
        {
            // Full-text query
            if (_language == null)
            {
                throw new IllegalArgumentException("Cannot build a string query with stemming without language");
            }
            query.append("_txt_stemmed_").append(_language).append(":(").append(escapedValue).append(')');
        }
        else if (operator == Operator.LIKE)
        {
            if (_language != null)
            {
                query.append("_").append(_language);
            }
            // Wildcard query: run a lower-case search.
            query.append("_s_lower:").append('(').append(escapedValue.toLowerCase()).append(')');
        }
        else if (operator == Operator.FUZZY)
        {
            if (_language != null)
            {
                query.append("_").append(_language);
            }
            // Run a lower-case fuzzy search with a maximum edit distance of 2 characters.
            query.append("_s_lower:").append('(').append(escapedValue.toLowerCase()).append("~2").append(')');
        }
        else if (operator == Operator.PHONETIC)
        {
            // Full-text query
            if (_language == null)
            {
                throw new IllegalArgumentException("Cannot build a phonetic query without language");
            }
            query.append("_phonetic_").append(_language).append(":(").append(escapedValue).append(')');
        }
        else if (operator == Operator.NGRAM)
        {
            // Full-text query
            if (_language == null)
            {
                throw new IllegalArgumentException("Cannot build a ngram query without language");
            }
            query.append("_ngram_").append(_language).append(":(").append(escapedValue).append(')');
        }
        else
        {
            // Strict string comparison (enumerator value, ID, ...)
            query.append("_s:").append('"').append(escapedValue).append('"');
        }
        
        return query.toString();
    }
    
    @Override
    public int hashCode()
    {
        return 31 * super.hashCode() + Objects.hash(_language, _valueAlreadyEscaped);
    }

    @Override
    public boolean equals(Object obj)
    {
        if (!super.equals(obj))
        {
            return false;
        }

        AbstractTextQuery other = (AbstractTextQuery) obj;
        return Objects.equals(_language, other._language) && Objects.equals(_valueAlreadyEscaped, other._valueAlreadyEscaped);
    }

    /**
     * Ensure that the string value is valid (i.e. the parentheses are balanced),
     * and throw an exception if it isn't.
     * @param value the string value to check.
     * @throws QuerySyntaxException if the value is invalid.
     */
    public static void checkStringValue(String value) throws QuerySyntaxException
    {
        // The parenthesis nesting level.
        boolean invalid = false;
        int level = 0;
        
        for (int i = 0; i < value.length() && !invalid; i++)
        {
            char ch = value.charAt(i);
            // The current character is escaped if the previous character is a slash
            // that is not itself escaped (the previous-previous character must not be a slash).
            boolean escaped = i > 0 && value.charAt(i - 1) == '\\' && !(i > 1 && value.charAt(i - 2) == '\\');
            
            if (ch == '(' && !escaped)
            {
                level++;
            }
            else if (ch == ')' && !escaped)
            {
                level--;
                if (level < 0)
                {
                    // More closing than opening parentheses at this point: the value is invalid.
                    invalid = true;
                }
            }
        }
        
        // If the parentheses are balanced, the level is 0 at this point.
        if (level != 0 || invalid)
        {
            String message = "The string search " + value + " is illegal, check the parentheses.";
            I18nizableText details = new I18nizableText("plugin.cms", "UITOOL_SEARCH_ERROR_QUERY_LABEL", Collections.singletonMap("value", new I18nizableText(value)));
            
            throw new QuerySyntaxException(message, details);
        }
    }
    
    /**
     * Escape from a string value the characters that can modify the query field
     * @param value the string value.
     * @param operator the operator
     * @return the escaped value.
     */
    public static String escapeStringValue(String value, Operator operator)
    {
        switch (operator)
        {
            case LIKE:
                // '*' are allowed characters
                // So escape all characters except '*'
                return Stream.of(StringUtils.splitByWholeSeparatorPreserveAllTokens(value, "*"))
                             .map(ClientUtils::escapeQueryChars)
                             .collect(Collectors.joining("*"));
            case SEARCH:
            case SEARCH_STEMMED:
                // '*' are allowed characters
                // So escape all characters (except '*' and whitespaces)
                return QueryHelper.escapeQueryCharsExceptStarsAndWhitespaces(value);

            default:
                return ClientUtils.escapeQueryChars(value);
        }
    }
}
