/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collector;
import java.util.stream.Collectors;

import org.apache.commons.collections4.map.HashedMap;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.util.LambdaUtils;

/**
 * Represents a search {@link Query} corresponding to the logical "and" between several other queries.
 */
public class AndQuery implements Query
{
    private static final Predicate<Object> __NOT_EMPTY_QUERY = q -> !(q instanceof String s) || !StringUtils.isBlank(s);
    
    /** The list of queries. The queries on this list are distinct. */
    protected List<Query> _queries;
    
    /**
     * Build an AndQuery object.
     * @param queries the queries.
     */
    public AndQuery(Query... queries)
    {
        this(Arrays.asList(queries));
    }
    
    /**
     * Build an AndQuery object.
     * @param queries the queries as a Collection.
     */
    public AndQuery(Collection<Query> queries)
    {
        _queries = queries.stream().distinct().toList();
    }
    
    /**
     * Returns a {@link Collector} which collects {@link Query Queries} into an AND query
     * @return a {@link Collector} which collects {@link Query Queries} into an AND query
     */
    public static Collector<Query, ?, AndQuery> collector()
    {
        return LambdaUtils.Collectors.withListAccumulation(AndQuery::new);
    }
    
    /**
     * Get the list of queries in this "and".
     * @return the list of queries.
     */
    public List<Query> getQueries()
    {
        return Collections.unmodifiableList(_queries);
    }
    
    @Override
    public String build() throws QuerySyntaxException
    {
        Predicate<Query> isMatchAll = MatchAllQuery.class::isInstance;
        List<Query> queriesWithoutMatchAll = _queries
                .stream()
                .filter(isMatchAll.negate())
                .collect(Collectors.toList());
        if (!_queries.isEmpty() && queriesWithoutMatchAll.isEmpty())
        {
            // special case where all are MatchAll
            return new MatchAllQuery().build();
        }
        
        boolean isFirst = true;
        StringBuilder sb = new StringBuilder();
        
        for (Query subQuery : queriesWithoutMatchAll)
        {
            if (subQuery instanceof MatchNoneQuery)
            {
                // short-circuit => no doc will match
                return new MatchNoneQuery().build();
            }
            else if (subQuery != null)
            {
                String exprAsString = subQuery.build();
                if (StringUtils.isNotBlank(exprAsString))
                {
                    if (!isFirst)
                    {
                        sb.append(" AND ");
                    }
                    sb.append("(").append(exprAsString).append(")");
                    isFirst = false;
                }    
            }
        }
        
        if (isFirst)
        {
            return "";
        }
        else
        {
            return sb.toString();
        }
    }
    
    public Optional<Object> buildAsJson() throws QuerySyntaxException
    {
        Optional<Query> rewrittenQueryAsOptional = rewrite();
        if (rewrittenQueryAsOptional.isEmpty())
        {
            return Optional.empty();
        }
        
        Query rewrittenQuery = rewrittenQueryAsOptional.get();
        if (!(rewrittenQuery instanceof AndQuery andQuery))
        {
            return rewrittenQuery.buildAsJson();
        }
        
        List<Query> rewrittenQueries = andQuery.getQueries();
        
        Set<Query> mustQueries = new HashSet<>();
        Set<Query> filterQueries = new HashSet<>();
        Set<Query> mustNotQueries = new HashSet<>();
        
        _dispatchClauses(rewrittenQueries, mustQueries, filterQueries, mustNotQueries);
        
        boolean matchNone = mustQueries.stream().anyMatch(MatchNoneQuery.class::isInstance)
                         || filterQueries.stream().anyMatch(MatchNoneQuery.class::isInstance)
                         || mustNotQueries.stream().anyMatch(MatchAllQuery.class::isInstance);
        
        if (matchNone)
        {
            return new MatchNoneQuery().buildAsJson();
        }
        
        // filter out empty queries
        List<Object> builtMustQueries = mustQueries.stream()
                                                   .map(LambdaUtils.wrap(Query::buildAsJson))
                                                   .flatMap(Optional::stream)
                                                   .filter(__NOT_EMPTY_QUERY)
                                                   .toList();

        List<Object> builtFilterQueries = filterQueries.stream()
                                                       .map(LambdaUtils.wrap(Query::buildAsJson))
                                                       .flatMap(Optional::stream)
                                                       .filter(__NOT_EMPTY_QUERY)
                                                       .toList();

        List<Object> builtMustNotQueries = mustNotQueries.stream()
                                                         .map(LambdaUtils.wrap(Query::buildAsJson))
                                                         .flatMap(Optional::stream)
                                                         .filter(__NOT_EMPTY_QUERY)
                                                         .toList();
        
        boolean noPositiveQueries = builtMustQueries.isEmpty() && builtFilterQueries.isEmpty();

        if (noPositiveQueries && builtMustNotQueries.isEmpty())
        {
            return Optional.empty();
        }
        
        if (builtMustQueries.size() == 1 && builtFilterQueries.isEmpty() && builtMustNotQueries.isEmpty())
        {
            return Optional.of(builtMustQueries.get(0));
        }
        
        Map<String, Object> clauses = new HashedMap<>();
        if (!builtMustQueries.isEmpty())
        {
            clauses.put(Query.BOOL_MUST, builtMustQueries);
        }
        
        if (!builtFilterQueries.isEmpty())
        {
            clauses.put(Query.BOOL_FILTER, builtFilterQueries);
        }
               
        if (!builtMustNotQueries.isEmpty())
        {
            if (noPositiveQueries)
            {
                clauses.put(Query.BOOL_MUST, "*:*");
            }

            clauses.put(Query.BOOL_MUST_NOT, builtMustNotQueries);
        }
               
        return Optional.of(Map.of("bool", clauses));
    }
    
    private void _dispatchClauses(List<Query> queries, Set<Query> mustQueries, Set<Query> filterQueries, Set<Query> mustNotQueries)
    {
        // dispatch queries by nature, while discarding useless clauses
        for (Query q : queries)
        {
            if (q instanceof ConstantNilScoreQuery filterQuery)
            {
                Query filteredQuery = filterQuery.getSubQuery();
                if (!(filteredQuery instanceof MatchAllQuery))
                {
                    filterQueries.add(filteredQuery);
                }
            }
            else if (q instanceof NotQuery notQuery)
            {
                Query negatedQuery = notQuery.getSubQuery();
                if (negatedQuery instanceof OrQuery orQuery)
                {
                    mustNotQueries.addAll(orQuery.getQueries());
                }
                else if (!(negatedQuery instanceof MatchNoneQuery))
                {
                    mustNotQueries.add(negatedQuery);
                }
            }
            else if (!(q instanceof MatchAllQuery))
            {
                mustQueries.add(q);
            }
        }
    }
    
    public Optional<Query> rewrite()
    {
        List<Query> rewrittenQueries = _queries.stream()
                .filter(Objects::nonNull)
                .map(Query::rewrite)
                .flatMap(Optional::stream)
                .toList();
        
        Set<Query> queries = new HashSet<>();
        for (Query q : rewrittenQueries)
        {
            // do not use instanceof here, as we don't want to match AndQuery subclasses
            if (q.getClass().getName().equals(AndQuery.class.getName()))
            {
                queries.addAll(((AndQuery) q).getQueries());
            }
            else if (q instanceof MatchNoneQuery)
            {
                return Optional.of(new MatchNoneQuery());
            }
            else if (!(q instanceof MatchAllQuery))
            {
                queries.add(q);
            }
        }
        
        if (queries.isEmpty())
        {
            return Optional.empty();
        }
        
        if (queries.size() == 1)
        {
            return Optional.of(queries.iterator().next());
        }
        
        return Optional.of(new AndQuery(queries));
    }
    
    @Override
    public String toString(int indent)
    {
        final String andLineIndent = StringUtils.repeat(' ', indent);
        final int subIndent = indent + 2;
        final String subQueries = _queries
                .stream()
                .filter(Objects::nonNull)
                .map(sq -> sq.toString(subIndent))
                .collect(Collectors.joining("\n"));
        return andLineIndent + "[AND]\n" + subQueries + "\n" + andLineIndent + "[/AND]";
    }
    
    @Override
    public int hashCode()
    {
        return _queries.hashCode();
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        
        if (obj == null || getClass() != obj.getClass())
        {
            return false;
        }
        
        AndQuery other = (AndQuery) obj;
        return Objects.equals(_queries, other._queries);
    }
}
