/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.apache.solr.client.solrj.util.ClientUtils;

/**
 * Wraps another {@link Query}, but giving to each matching document a boosted score (scores are multiplied). 
 */
public class BoostedQuery implements Query
{
    private Query _query;
    private float _boost;

    /**
     * Build a BoostedQuery object.
     * @param query The wrapped query
     * @param boost The boost
     */
    public BoostedQuery(Query query, float boost)
    {
        _query = query;
        _boost = boost;
    }
    
    @Override
    public String build() throws QuerySyntaxException
    {
        StringBuilder sb = new StringBuilder()
                .append("{!boost b=")
                .append(_boost)
                .append(" v=\"")
                .append(ClientUtils.escapeQueryChars(_query.build()))
                .append("\"}");
        return sb.toString();
    }
    
    public Optional<Object> buildAsJson() throws QuerySyntaxException
    {
        return _query.buildAsJson()
                .map(q -> Map.of("boost", Map.of("query", q, "b", _boost)));
    }
    
    @Override
    public String toString(int indent)
    {
        final String thisLineIndent = StringUtils.repeat(' ', indent);
        final int subIndent = indent + 2;
        final String subLineIndent = StringUtils.repeat(' ', subIndent);
        final String boost = subLineIndent + "[BOOST]" + _boost + "[/BOOST]";
        final String q = subLineIndent + "[Q]\n" + _query.toString(subIndent + 2) + "\n" + subLineIndent + "[/Q]";
        return thisLineIndent + "[BOOSTED]\n" + boost + "\n" + q + "\n" + thisLineIndent + "[/BOOSTED]";
    }

    @Override
    public int hashCode()
    {
        return Objects.hash(_boost, _query);
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        
        if (obj == null || getClass() != obj.getClass())
        {
            return false;
        }
        
        BoostedQuery other = (BoostedQuery) obj;
        return Float.floatToIntBits(_boost) == Float.floatToIntBits(other._boost) && Objects.equals(_query, other._query);
    }
}
