/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.Objects;

import org.ametys.cms.content.indexing.solr.SolrFieldNames;
import org.ametys.cms.search.solr.field.CommentsSearchField;

/**
 * Represents a {@link Query} testing if contents have comments.
 */
public class CommentQuery implements Query
{
    private CommentInclusion _commentInclusion;
    private Operator _operator;
    
    /**
     * The type (validated, non validated or both) of comments to include for {@link CommentQuery}
     */
    public static enum CommentInclusion
    {
        /** Request only contents with both validated and non-validated comments */
        VALIDATED_AND_NON_VALIDATED,
        /** Request only contents with validated comments */
        VALIDATED,
        /** Request only contents with non validated-comments */
        NON_VALIDATED,
    }
    
    /**
     * Constructor
     * @param commentInclusion The type of comments to include in search
     * (request only contents with {@link CommentInclusion#VALIDATED validated} comments, 
     * or only contents with {@link CommentInclusion#NON_VALIDATED non validated} comments, 
     * or {@link CommentInclusion#VALIDATED_AND_NON_VALIDATED both})
     * @param operator The operator. Only {@link Query.Operator#EQ} and {@link Query.Operator#NE} are allowed.
     */
    public CommentQuery(CommentInclusion commentInclusion, Operator operator)
    {
        if (Operator.EQ != operator && Operator.NE != operator && Operator.EXISTS != operator)
        {
            throw new IllegalArgumentException("Test operator '" + operator + "' is unknown for comment query");
        }
        
        _commentInclusion = commentInclusion;
        _operator = operator;
    }
    
    @Override
    public String build() throws QuerySyntaxException
    {
        StringBuilder query = new StringBuilder();

        if (_operator == Operator.EXISTS)
        {   
            query.append(CommentsSearchField.NAME).append(":true");
            return query.toString();
        }
        
        boolean appendNegation = _operator == Operator.NE; 
        if (appendNegation)
        {
            NotQuery.appendNegation(query).append('(');
        }
        
        switch (_commentInclusion)  
        {
            case VALIDATED_AND_NON_VALIDATED:
                query.append(CommentsSearchField.NAME).append(":true");
                break;
            case VALIDATED:
                query.append(SolrFieldNames.CONTENT_COMMENTS_VALIDATED).append(":true");
                break;
            case NON_VALIDATED:
                query.append(SolrFieldNames.CONTENT_COMMENTS_NONVALIDATED).append(":true");
                break;
            default:
                throw new IllegalArgumentException("CommentInclusion '" + _commentInclusion + "' is not supported.");
        }
        
        if (appendNegation)
        {
            query.append(')');
        }
        
        return query.toString();
    }

    @Override
    public int hashCode()
    {
        return Objects.hash(_commentInclusion, _operator);
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        
        if (obj == null || getClass() != obj.getClass())
        {
            return false;
        }
        
        CommentQuery other = (CommentQuery) obj;
        return _commentInclusion == other._commentInclusion && _operator == other._operator;
    }
}
