/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.Optional;

/**
 * Represents a {@link Query} testing the content types or the mixins of contents.
 */
public class ContentTypeOrMixinTypeQuery implements Query
{
    private Operator _operator;
    private List<String> _ids;
    
    /**
     * Build a ContentTypeOrMixinTypeQuery.
     * @param ids the content type ids.
     */
    public ContentTypeOrMixinTypeQuery(String... ids)
    {
        this(Operator.EQ, ids);
    }
    
    /**
     * Build a ContentTypeOrMixinTypeQuery.
     * @param ids the content type ids.
     */
    public ContentTypeOrMixinTypeQuery(Collection<String> ids)
    {
        this(Operator.EQ, ids);
    }
    
    /**
     * Build a ContentTypeOrMixinTypeQuery.
     * @param operator the operator.
     * @param ids the content type ids.
     */
    public ContentTypeOrMixinTypeQuery(Operator operator, String... ids)
    {
        this(operator, Arrays.asList(ids));
    }
    
    /**
     * Build a ContentTypeOrMixinTypeQuery.
     * @param operator the operator.
     * @param ids the content type ids.
     */
    public ContentTypeOrMixinTypeQuery(Operator operator, Collection<String> ids)
    {
        if (Operator.EQ != operator && Operator.NE != operator && Operator.EXISTS != operator)
        {
            throw new IllegalArgumentException("Test operator '" + operator + "' is unknown for test's expression.");
        }
        
        _operator = operator;
        _ids = new ArrayList<>(ids);
    }
    
    /**
     * Get the operator.
     * @return the operator
     */
    public Operator getOperator()
    {
        return _operator;
    }
    
    /**
     * Get the content type or mixin ids.
     * @return the content type or mixin ids.
     */
    public List<String> getIds()
    {
        return Collections.unmodifiableList(_ids);
    }
    
    @Override
    public String build() throws QuerySyntaxException
    {
        return _buildQuery().build();
    }

    public Optional<Object> buildAsJson() throws QuerySyntaxException
    {
        return _buildQuery().buildAsJson();
    }
    
    private Query _buildQuery()
    {
        Query query = new OrQuery(new ContentTypeQuery(getIds()), new MixinTypeQuery(getIds()));
        
        if (_operator == Operator.NE)
        {
            return new NotQuery(query);
        }
        else
        {
            return query;
        }
    }
    
    public Optional<Query> rewrite()
    {
        return Optional.of(_buildQuery());
    }
    
    @Override
    public int hashCode()
    {
        return Objects.hash(_operator, _ids);
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        
        if (obj == null || getClass() != obj.getClass())
        {
            return false;
        }
        
        ContentTypeOrMixinTypeQuery other = (ContentTypeOrMixinTypeQuery) obj;
        return Objects.equals(_operator, other._operator)
            && Objects.equals(_ids, other._ids);
    }
}
