/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.time.LocalDate;
import java.time.LocalTime;

import org.ametys.core.util.date.AdaptableDate;

/**
 * Represents a {@link Query} testing a date range.
 */
public class DateRangeQuery extends AbstractRangeQuery<AdaptableDate>
{
    /**
     * Build a date range query.
     * @param fieldPath The field path.
     * @param from The lower end of the range.
     * @param to The upper end of the range.
     */
    public DateRangeQuery(String fieldPath, LocalDate from, LocalDate to)
    {
        this(fieldPath, AdaptableDate.fromDate(from), AdaptableDate.fromDate(to));
    }
    
    /**
     * Build a date range query.
     * @param fieldPath The field path.
     * @param from The lower end of the range.
     * @param to The upper end of the range.
     */
    public DateRangeQuery(String fieldPath, AdaptableDate from, AdaptableDate to)
    {
        this(fieldPath, from, to, true, true);
    }
    
    /**
     * Build a date range query.
     * @param fieldPath The field path.
     * @param from The lower end of the range.
     * @param to The upper end of the range.
     * @param includeFrom Whether to include the lower end or not.
     * @param includeTo Whether to include the upper end or not.
     */
    public DateRangeQuery(String fieldPath, LocalDate from, LocalDate to, boolean includeFrom, boolean includeTo)
    {
        this(fieldPath, AdaptableDate.fromDate(from), AdaptableDate.fromDate(to), includeFrom, includeTo);
    }
    
    /**
     * Build a date range query.
     * @param fieldPath The field path.
     * @param from The lower end of the range.
     * @param to The upper end of the range.
     * @param includeFrom Whether to include the lower end or not.
     * @param includeTo Whether to include the upper end or not.
     */
    public DateRangeQuery(String fieldPath, AdaptableDate from, AdaptableDate to, boolean includeFrom, boolean includeTo)
    {
        super(fieldPath + "_dt", from, to, includeFrom, includeTo);
    }
    
    @Override
    public String lowerBoundForQuery(AdaptableDate value)
    {
        return value.resolveDate()
                    .atTime(includeLowerBound() ? LocalTime.MIN : LocalTime.MAX)
                    .format(AbstractDateOperatorQuery.DATE_FORMATTER);
    }
    
    @Override
    public String upperBoundForQuery(AdaptableDate value)
    {
        return value.resolveDate()
                    .atTime(includeUpperBound() ? LocalTime.MIN : LocalTime.MAX)
                    .format(AbstractDateOperatorQuery.DATE_FORMATTER);
    }
}
