/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;

import org.apache.solr.client.solrj.util.ClientUtils;

import org.ametys.core.util.date.AdaptableDate;
import org.ametys.core.util.date.AdaptableDate.DateTimeInterval;

/**
 * Represents a {@link Query} testing a datetime field.
 */
public class DateTimeQuery extends AbstractDateOperatorQuery
{
    /**
     * Build a DateQuery testing the existence of the field.
     * @param fieldPath the field path.
     */
    public DateTimeQuery(String fieldPath)
    {
        this(fieldPath, Operator.EXISTS, (AdaptableDate) null);
    }
    
    /**
     * Build a DateQuery.
     * @param fieldPath the field's path
     * @param value the value.
     */
    public DateTimeQuery(String fieldPath, ZonedDateTime value)
    {
        this(fieldPath, AdaptableDate.fromDateTime(value.withZoneSameInstant(ZoneOffset.UTC).toLocalDateTime()));
    }
    
    /**
     * Build a DateQuery.
     * @param fieldPath the field's path
     * @param value the value.
     */
    public DateTimeQuery(String fieldPath, AdaptableDate value)
    {
        this(fieldPath, Operator.EQ, value);
    }
    
    /**
     * Build a DateQuery.
     * @param fieldPath the field's path
     * @param op the operator.
     * @param value the value.
     */
    public DateTimeQuery(String fieldPath, Operator op, ZonedDateTime value)
    {
        this(fieldPath, op, AdaptableDate.fromDateTime(value.withZoneSameInstant(ZoneOffset.UTC).toLocalDateTime()));
    }
    
    /**
     * Build a DateQuery.
     * @param fieldPath the field's path
     * @param op the operator.
     * @param value the value.
     */
    public DateTimeQuery(String fieldPath, Operator op, AdaptableDate value)
    {
        super(fieldPath + "_dt", op, value);
    }

    @Override
    protected void appendDateValue(StringBuilder query, Operator operator, AdaptableDate value)
    {
        DateTimeInterval interval = value.resolveDateTime();
        
        LocalDateTime min = interval.min();
        ZonedDateTime minDateTime = min.atZone(ZoneId.systemDefault()).withZoneSameInstant(ZoneOffset.UTC);
        String minStrValue = minDateTime.format(DATE_FORMATTER);
        
        LocalDateTime max = interval.max();
        ZonedDateTime maxDateTime = max.atZone(ZoneId.systemDefault()).withZoneSameInstant(ZoneOffset.UTC);
        String maxStrValue = maxDateTime.format(DATE_FORMATTER);
        
        if (operator == Operator.EQ || operator == Operator.NE)
        {
            if (min.equals(max))
            {
                query.append(ClientUtils.escapeQueryChars(minStrValue));
            }
            else
            {
                query.append('[').append(minStrValue).append(" TO ").append(maxStrValue).append(']');
            }
            
            return;
        }
        else if (operator == Operator.GT)
        {
            query.append('{').append(minStrValue).append(" TO *]");
        }
        else if (operator == Operator.GE)
        {
            query.append('[').append(minStrValue).append(" TO *]");
        }
        else if (operator == Operator.LT)
        {
            query.append("[* TO ").append(maxStrValue).append('}');
        }
        else if (operator == Operator.LE)
        {
            query.append("[* TO ").append(maxStrValue).append(']');
        }
    }
}
