/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.Objects;

import org.ametys.cms.content.indexing.solr.SolrFieldNames;
import org.ametys.cms.search.QueryBuilder;

/**
 * A {@link Query} realizing a full-text search on contents (search on all properties).
 * Available operators:<ul>
 * <li>SEARCH (default)</li>
 * <li>SEARCH_STEMMED</li>
 * <li>LIKE</li>
 * <li>EQ</li>
 * </ul>
 */
public class FullTextQuery implements Query
{
    /** The value to test. */
    protected String _value;
    /** The field name. */
    protected String _fieldName;
    /** The language. */
    protected String _language;
    /** The operator. */
    protected Operator _operator;
    /** <code>true</code> if the value is already escaped and there is no need to escape again the value during {@link #build() the build of the query}. */
    protected boolean _valueAlreadyEscaped;
    
    /**
     * Build a FullTextQuery.
     * @param value the value.
     */
    public FullTextQuery(String value)
    {
        this(value, Operator.SEARCH);
    }
    
    /**
     * Build a FullTextQuery.
     * @param value the value.
     * @param operator The operator to use.
     */
    public FullTextQuery(String value, Operator operator)
    {
        this(value, null, operator);
    }
    
    /**
     * Build a FullTextQuery.
     * @param value the value.
     * @param language the language.
     */
    public FullTextQuery(String value, String language)
    {
        this(value, language, Operator.SEARCH);
    }
    
    /**
     * Build a FullTextQuery.
     * @param value the value.
     * @param language the language.
     * @param operator The operator to use.
     */
    public FullTextQuery(String value, String language, Operator operator)
    {
        this(value, language, operator, false);
    }
    
    /**
     * Build a FullTextQuery.
     * @param value the value.
     * @param language the language.
     * @param operator The operator to use.
     * @param alreadyEscaped true if the value is already escaped and there is no need to escape again the value during {@link #build() the build of the query}.
     */
    public FullTextQuery(String value, String language, Operator operator, boolean alreadyEscaped)
    {
        this(value, SolrFieldNames.SYSTEM_FULL, language, operator, alreadyEscaped);
    }
    
    /**
     * Build a FullTextQuery.
     * @param value the value.
     * @param fieldName the field name.
     * @param language the language.
     * @param operator The operator to use.
     */
    public FullTextQuery(String value, String fieldName, String language, Operator operator)
    {
        this(value, fieldName, language, operator, false);
    }
    
    /**
     * Build a FullTextQuery.
     * @param value the value.
     * @param fieldName the field name.
     * @param language the language.
     * @param operator The operator to use.
     * @param alreadyEscaped true if the value is already escaped and there is no need to escape again the value during {@link #build() the build of the query}.
     */
    public FullTextQuery(String value, String fieldName, String language, Operator operator, boolean alreadyEscaped)
    {
        _value = value;
        _fieldName = fieldName;
        _language = language;
        _operator = operator;
        _valueAlreadyEscaped = alreadyEscaped;
    }
    
    /**
     * Get the value.
     * @return the value.
     */
    public String getValue()
    {
        return _value;
    }
    
    /**
     * Get the language.
     * @return the language.
     */
    public String getLanguage()
    {
        return _language;
    }
    
    /**
     * Get the operator to use.
     * @return the operator.
     */
    public Operator getOperator()
    {
        return _operator;
    }
    
    @Override
    public String build() throws QuerySyntaxException
    {
        String language = _language != null ? _language : QueryBuilder.DEFAULT_LANGUAGE;
        
        AbstractTextQuery.checkStringValue(_value);
        
        String escapedValue = _valueAlreadyEscaped ? _value : AbstractTextQuery.escapeStringValue(_value, _operator);
        
        StringBuilder solrQuery = new StringBuilder();
        
        switch (_operator)
        {
            case EQ:
                solrQuery.append(_fieldName).append(SolrFieldNames.EXACT_WS_OPERATOR)
                         .append(":\"").append(escapedValue).append('"');
                break;
            case LIKE:
                solrQuery.append(_fieldName)
                         .append(":(").append(escapedValue).append(')');
                break;
            case SEARCH_STEMMED:
                solrQuery.append(_fieldName).append(SolrFieldNames.STEMMED_OPERATOR).append(SolrFieldNames.LANGUAGE_SEPARATOR).append(language)
                         .append(":(").append(escapedValue).append(')');
                break;
            case SEARCH:
            default:
                solrQuery.append(_fieldName).append(SolrFieldNames.LANGUAGE_SEPARATOR).append(language)
                         .append(":(").append(escapedValue).append(')');
                break;
        }
        
        return solrQuery.toString();
    }

    @Override
    public int hashCode()
    {
        return 31 + Objects.hash(_language, _operator, _value);
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        
        if (obj == null || getClass() != obj.getClass())
        {
            return false;
        }
        
        FullTextQuery other = (FullTextQuery) obj;
        return Objects.equals(_language, other._language)
                && Objects.equals(_operator, other._operator)
                && Objects.equals(_value, other._value);
    }
}
