/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.time.LocalDateTime;
import java.util.List;

import org.ametys.cms.content.indexing.solr.SolrFieldNames;
import org.ametys.core.util.date.AdaptableDate;

/**
 * {@link Query} testing the existence of a particular step on the workflow history.
 */
public class HistoryStepQuery extends JoinQuery
{
    /**
     * Create a new HistoryStepQuery.
     * @param stepId the workflow step ID.
     */
    public HistoryStepQuery(int stepId)
    {
        this(stepId, (AdaptableDate) null, (AdaptableDate) null);
    }
    
    /**
     * Create a new HistoryStepQuery.
     * @param stepId the workflow step ID.
     * @param startAfter the beginning Date
     * @param startBefore the ending Date
     */
    public HistoryStepQuery(int stepId, LocalDateTime startAfter, LocalDateTime startBefore)
    {
        this(stepId, AdaptableDate.fromDateTime(startAfter), AdaptableDate.fromDateTime(startBefore));
    }
    
    /**
     * Create a new HistoryStepQuery.
     * @param stepId the workflow step ID.
     * @param startAfter the beginning Date
     * @param startBefore the ending Date
     */
    public HistoryStepQuery(int stepId, AdaptableDate startAfter, AdaptableDate startBefore)
    {
        super(_getSubQuery(stepId, startAfter, startBefore), List.of(SolrFieldNames.WORKFLOW_REF, SolrFieldNames.WORKFLOW_HISTORY_STEPS));
    }
    
    private static Query _getSubQuery(int stepId, AdaptableDate startAfter, AdaptableDate startBefore)
    {
        Query stepQuery = () -> SolrFieldNames.WORKFLOW_STEP_ID + ":" + stepId;
        
        if (startAfter != null || startBefore != null)
        {
            StringBuilder query = new StringBuilder();
            query.append(SolrFieldNames.WORKFLOW_STEP_STARTDATE).append(":[")
                 .append(startAfter == null ? "*" : startAfter.resolveDateTime().min().format(AbstractDateOperatorQuery.DATE_FORMATTER))
                 .append(" TO ")
                 .append(startBefore == null ? "*" : startBefore.resolveDateTime().max().format(AbstractDateOperatorQuery.DATE_FORMATTER))
                 .append("}");
            
            Query dateQuery = () -> query.toString();
            
            return new AndQuery(stepQuery, dateQuery);
        }
        
        return stepQuery;
    }
}
