/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.apache.solr.client.solrj.util.ClientUtils;

import org.ametys.core.util.LambdaUtils;

/**
 * Same as {@link OrQuery}, but the score for each result will be the max score from the clauses (instead of the sum).
 */
public class MaxScoreOrQuery extends OrQuery
{
    private static final float __DEFAULT_TIE_BREAKER = 0.01f;
    private float _tieBreaker;
    
    /**
     * Build a MaxScoreOrQuery object with default tiebreaker.
     * @param queries the queries.
     */
    public MaxScoreOrQuery(Query... queries)
    {
        this(__DEFAULT_TIE_BREAKER, queries);
    }
    
    /**
     * Build a MaxScoreOrQuery object with default tiebreaker.
     * @param queries the queries as a Collection.
     */
    public MaxScoreOrQuery(Collection<Query> queries)
    {
        this(__DEFAULT_TIE_BREAKER, queries);
    }
    
    /**
     * Build a MaxScoreOrQuery object.
     * @param tieBreaker The tiebreaker
     * @param queries the queries.
     */
    public MaxScoreOrQuery(float tieBreaker, Query... queries)
    {
        super(queries);
        _tieBreaker = tieBreaker;
    }
    
    /**
     * Build a MaxScoreOrQuery object.
     * @param tieBreaker The tiebreaker
     * @param queries the queries as a Collection.
     */
    public MaxScoreOrQuery(float tieBreaker, Collection<Query> queries)
    {
        super(queries);
        _tieBreaker = tieBreaker;
    }
    
    @Override
    public String build() throws QuerySyntaxException
    {
        StringBuilder sb = new StringBuilder()
                .append("{!maxscore tie=")
                .append(_tieBreaker)
                .append(" v=\"")
                .append(ClientUtils.escapeQueryChars(super.build()))
                .append("\"}");
        return sb.toString();
    }
    
    @Override
    public Optional<Object> buildAsJson() throws QuerySyntaxException
    {
        Optional<Query> rewrittenQueryAsOptional = rewrite();
        if (rewrittenQueryAsOptional.isEmpty())
        {
            return Optional.empty();
        }
        
        Query rewrittenQuery = rewrittenQueryAsOptional.get();
        if (!(rewrittenQuery instanceof MaxScoreOrQuery orQuery))
        {
            return rewrittenQuery.buildAsJson();
        }
        
        List<Query> rewrittenQueries = orQuery.getQueries();

        // filter out empty queries
        List<Object> builtQueries = rewrittenQueries.stream()
                                                    .map(LambdaUtils.wrap(Query::buildAsJson))
                                                    .flatMap(Optional::stream)
                                                    .filter(q -> !(q instanceof String s) || !StringUtils.isBlank(s))
                                                    .toList();

        if (builtQueries.isEmpty())
        {
            return Optional.empty();
        }
        
        if (builtQueries.size() == 1)
        {
            return Optional.of(builtQueries.get(0));
        }

        return Optional.of(Map.of("maxscore", Map.of("query", Map.of("bool", Map.of(Query.BOOL_SHOULD, builtQueries)),
                                                     "tie", _tieBreaker)));
    }
    
    @Override
    public Optional<Query> rewrite()
    {
        if (_queries.isEmpty())
        {
            return Optional.empty();
        }
        
        if (_queries.size() == 1)
        {
            return  _queries.get(0).rewrite();
        }
        
        List<Query> queries = _queries.stream()
            .map(Query::rewrite)
            .flatMap(Optional::stream)
            .toList();
        return Optional.of(new MaxScoreOrQuery(_tieBreaker, queries));
    }
    
    @Override
    protected String _tagNameForToString()
    {
        return "OR(MAXSCORE)";
    }

    @Override
    public int hashCode()
    {
        return Objects.hash(_tieBreaker, _queries);
    }

    @Override
    public boolean equals(Object obj)
    {
        if (!super.equals(obj))
        {
            return false;
        }
        
        MaxScoreOrQuery other = (MaxScoreOrQuery) obj;
        return Float.floatToIntBits(_tieBreaker) == Float.floatToIntBits(other._tieBreaker);
    }
}
