/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.Map;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;

/**
 * Represents a {@link Query} corresponding to the logical negation of another {@link Query}.
 */
public class NotQuery extends AbstractWrapperQuery
{
    /** The prefix for making a query negative. */
    public static final String NEGATION_QUERY_PREFIX = "*:* -";
    
    /**
     * Build a NotQuery object.
     * @param query the negated query.
     */
    public NotQuery(Query query)
    {
        super(query);
    }
    
    /**
     * Appends a negation to the query being built.
     * <br>This method just does <code>sb.append({@value #NEGATION_QUERY_PREFIX});</code>
     * @param query The query builder
     * @return The given query builder
     */
    public static StringBuilder appendNegation(StringBuilder query)
    {
        query.append(NEGATION_QUERY_PREFIX);
        return query;
    }
    
    @Override
    public String build() throws QuerySyntaxException
    {
        String queryString = getSubQuery().build();
        if (queryString.isEmpty())
        {
            return "";
        }
        else
        {
            return NEGATION_QUERY_PREFIX + "(" + queryString + ")";
        }
    }
    
    public Optional<Object> buildAsJson() throws QuerySyntaxException
    {
        Optional<Query> rewrittenQueryOptional = getSubQuery().rewrite();
        if (rewrittenQueryOptional.isEmpty())
        {
            return Optional.empty();
        }
        
        Query rewrittenQuery = rewrittenQueryOptional.get();
        if (rewrittenQuery instanceof NotQuery nq)
        {
            return nq.getSubQuery().buildAsJson();
        }
        else
        {
            return rewrittenQuery.buildAsJson()
                    .map(q -> Map.of("bool", Map.of("must", "*:*", "must_not", q)));
        }
    }
    
    public Optional<Query> rewrite()
    {
        Query subQuery = getSubQuery();
        Optional<Query> rewrittenQueryOptional = subQuery.rewrite();
        if (rewrittenQueryOptional.isEmpty())
        {
            return Optional.empty();
        }
        
        Query rewrittenQuery = rewrittenQueryOptional.get();
        if (rewrittenQuery == subQuery)
        {
            return Optional.of(this);
        }
        else if (rewrittenQuery instanceof NotQuery nq)
        {
            return Optional.of(nq.getSubQuery());
        }
        else
        {
            return Optional.of(new NotQuery(rewrittenQuery));
        }
    }
    
    @Override
    public String toString(int indent)
    {
        final String notLineIndent = StringUtils.repeat(' ', indent);
        final String subq = getSubQuery().toString(indent + 2);
        return notLineIndent + "[NOT]\n" +  subq + "\n" + notLineIndent + "[/NOT]";
    }
}
