/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collector;
import java.util.stream.Collectors;

import org.apache.commons.lang3.StringUtils;

import org.ametys.core.util.LambdaUtils;

/**
 * Represents a search {@link Query} corresponding to the logical "or" between several other queries.
 */
public class OrQuery implements Query
{
    /** The list of queries. The queries on this list are distinct. */
    protected List<Query> _queries;
    
    /**
     * Build an OrQuery object.
     * @param queries the queries.
     */
    public OrQuery(Query... queries)
    {
        this(Arrays.asList(queries));
    }
    
    /**
     * Build an OrQuery object.
     * @param queries the queries as a Collection.
     */
    public OrQuery(Collection<Query> queries)
    {
        _queries = queries.stream().distinct().toList();
    }
    
    /**
     * Returns a {@link Collector} which collects {@link Query Queries} into an OR query
     * @return a {@link Collector} which collects {@link Query Queries} into an OR query
     */
    public static Collector<Query, ?, OrQuery> collector()
    {
        return LambdaUtils.Collectors.withListAccumulation(OrQuery::new);
    }
    
    /**
     * Get the list of queries in this "or".
     * @return the list of queries.
     */
    public List<Query> getQueries()
    {
        return Collections.unmodifiableList(_queries);
    }
    
    @Override
    public String build() throws QuerySyntaxException
    {
        Predicate<Query> isMatchNone = MatchNoneQuery.class::isInstance;
        List<Query> queriesWithoutMatchNone = _queries
                .stream()
                .filter(isMatchNone.negate())
                .collect(Collectors.toList());
        if (!_queries.isEmpty() && queriesWithoutMatchNone.isEmpty())
        {
            // special case where all are MatchNone
            return new MatchNoneQuery().build();
        }
        
        boolean isFirst = true;
        StringBuilder sb = new StringBuilder();
        
        for (Query subQuery : queriesWithoutMatchNone)
        {
            if (subQuery instanceof MatchAllQuery)
            {
                // short-circuit => all docs will match
                return new MatchAllQuery().build();
            }
            else if (subQuery != null)
            {
                String exprAsString = subQuery.build();
                if (StringUtils.isNotBlank(exprAsString))
                {
                    if (!isFirst)
                    {
                        sb.append(" OR ");
                    }
                    sb.append("(").append(exprAsString).append(")");
                    isFirst = false;
                }
            }
        }
        
        if (isFirst)
        {
            return "";
        }
        else
        {
            return sb.toString();
        }
    }
    
    public Optional<Object> buildAsJson() throws QuerySyntaxException
    {
        Optional<Query> rewrittenQueryAsOptional = rewrite();
        if (rewrittenQueryAsOptional.isEmpty())
        {
            return Optional.empty();
        }
        
        Query rewrittenQuery = rewrittenQueryAsOptional.get();
        if (!(rewrittenQuery instanceof OrQuery orQuery))
        {
            return rewrittenQuery.buildAsJson();
        }
        
        List<Query> rewrittenQueries = orQuery.getQueries();
        
        // filter out empty queries
        List<Object> builtQueries = rewrittenQueries.stream()
                                                    .map(LambdaUtils.wrap(Query::buildAsJson))
                                                    .flatMap(Optional::stream)
                                                    .filter(q -> !(q instanceof String s) || !StringUtils.isBlank(s))
                                                    .toList();

        if (builtQueries.isEmpty())
        {
            return Optional.empty();
        }
        
        if (builtQueries.size() == 1)
        {
            return Optional.of(builtQueries.get(0));
        }

        return Optional.of(Map.of("bool", Map.of(Query.BOOL_SHOULD, builtQueries)));
    }
    
    public Optional<Query> rewrite()
    {
        List<Query> rewrittenQueries = _queries.stream()
                .filter(Objects::nonNull)
                .map(Query::rewrite)
                .flatMap(Optional::stream)
                .toList();
        
        Set<Query> queries = new HashSet<>();
        for (Query q : rewrittenQueries)
        {
            // do not use instanceof here, as we don't want to match OrQuery subclasses
            if (q.getClass().getName().equals(OrQuery.class.getName()))
            {
                queries.addAll(((OrQuery) q).getQueries());
            }
            else if (q instanceof MatchAllQuery)
            {
                return Optional.of(new MatchAllQuery());
            }
            else if (!(q instanceof MatchNoneQuery))
            {
                queries.add(q);
            }
        }

        if (queries.isEmpty())
        {
            return Optional.empty();
        }
        
        if (queries.size() == 1)
        {
            return Optional.of(queries.iterator().next());
        }
        
        return Optional.of(new OrQuery(queries));
    }
    
    @Override
    public String toString(int indent)
    {
        String tagName = _tagNameForToString();
        final String orLineIndent = StringUtils.repeat(' ', indent);
        final int subIndent = indent + 2;
        final String subQueries = _queries
                .stream()
                .filter(Objects::nonNull)
                .map(sq -> sq.toString(subIndent))
                .collect(Collectors.joining("\n"));
        return orLineIndent + "[" + tagName + "]\n" + subQueries + "\n" + orLineIndent + "[/" + tagName + "]";
    }
    
    /**
     * The tag name for {@link #toString(int)} debug method.
     * @return The tag name for {@link #toString(int)} debug method.
     */
    protected String _tagNameForToString()
    {
        return "OR";
    }

    @Override
    public int hashCode()
    {
        return _queries.hashCode();
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        
        if (obj == null || getClass() != obj.getClass())
        {
            return false;
        }
        
        OrQuery other = (OrQuery) obj;
        return Objects.equals(_queries, other._queries);
    }
}
