/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

/**
 * Represents a search predicate.
 */
@FunctionalInterface
public interface Query
{
    /** Constant for mandatory boolean clauses */
    public static final String BOOL_MUST = "must";
    
    /** Constant for prohibited boolean clauses */
    public static final String BOOL_MUST_NOT = "must_not";
    
    /** Constant for optional boolean clauses */
    public static final String BOOL_SHOULD = "should";
    
    /** Constant for mandatory, but not affecting score, boolean clauses */
    public static final String BOOL_FILTER = "filter";
    
    /** Enumeration of available operators in {@link Query} */
    public enum Operator
    {
        /** Operator testing the existence of a property. */
        EXISTS("exists"),
        /** Constant of test's operator for textual search, on unstemmed terms. Works only for type String */
        SEARCH("search"),
        /** Constant of test's operator for textual search, on stemmed terms. Works only for type String */
        SEARCH_STEMMED("searchStemmed"),
        /** Constant of test's operator for 'like' comparison. Works only for type String */
        LIKE("like"), 
        /** Constant of test's operator for fuzzy search. Works only for type String */
        FUZZY("fuzzy"),
        /** Constant of test's operator for phonetic comparison. Works only for type String */
        PHONETIC("phonetic"),
        /** Constant of test's operator for 'less than' comparison. Works only for type String */
        NGRAM("ngram"),
        /** Constant of test's operator for 'less than' comparison */
        LT("lt"),
        /** Constant of test's operator for 'less than or equals to' comparison */
        LE("le"),
        /** Constant of test's operator for 'greater than' comparison */
        GT("gt"),
        /** Constant of test's operator for 'greater than or equals to' comparison */
        GE("ge"),
        /** Constant of test's operator for 'equals to' comparison */
        EQ("eq"),
        /** Constant of test's operator for 'not equals to' comparison */
        NE("ne");
        
        private static Map<String, Operator> _OP_NAMES = new HashMap<>();
        static 
        {
            for (Operator op : values())
            {
                _OP_NAMES.put(op.getName(), op);
            }
        }
        
        private final String _name;
        
        Operator(String name)
        {
            _name = name;
        }
        
        /**
         * Get the operator name.
         * @return the operator name.
         */
        public String getName()
        {
            return _name;
        }
        
        /**
         * Get an Operator object from its name.
         * @param name the operator name.
         * @return the operator object, or null if not found.
         */
        public static Operator fromName(String name)
        {
            if (_OP_NAMES.containsKey(name))
            {
                return _OP_NAMES.get(name);
            }
            throw new IllegalArgumentException("No operator with code '" + name + "'");
        }
    }
    
    /** Enumeration of available logical operators in {@link Query} */
    public enum LogicalOperator 
    {
        /** Logical operator AND */
        AND,
        /** Logical operator OR */
        OR
    }
    
    /**
     * Build the solr query string representing the Query object.
     * @return the solr query string representing the Query object.
     * @throws QuerySyntaxException if the query can't be built because of a syntax error.
     */
    String build() throws QuerySyntaxException;
    
    /**
     * Build the solr query representing the Query object.<br>
     * The return type may be either String or Map&lt;String, Object> following the Solr JSON Query DSL.
     * @return the solr query representing the Query object. Can be empty in case of empty query.
     * @throws QuerySyntaxException if the query can't be built because of a syntax error.
     */
    default Optional<Object> buildAsJson() throws QuerySyntaxException
    {
        return Optional.of(build());
    }
    
    /**
     * Rewrite this Query to allow further optimisations.
     * @return a rewritten Query, sementically identical. Can be empty in case of empty query
     */
    default Optional<Query> rewrite()
    {
        return Optional.of(this);
    }
    
    /**
     * Gets a representation of this {@link Query}, for pretty-printing for logging and debugging purposes
     * @param indent The current indentation. Base indentation is 2 (for printing a sub-level)
     * @return a representation of this {@link Query}
     */
    default String toString(int indent)
    {
        return QueryDebugger.toDebugString(this, indent);
    }
}
