/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import org.ametys.cms.search.query.Query.Operator;

/**
 * Class providing helper methods to Query implementations.
 */
public final class QueryHelper
{
    /** The special value testing that the field exists. */
    public static final String EXISTS_VALUE = "*";
    
    private QueryHelper()
    {
        // Hides the default constructor.
    }
    
    /**
     * Get a standard query on a value, with all possible standard operators.
     * @param fieldName The field name.
     * @param operator The operator.
     * @param value The value, can be null if the operator is EXISTS.
     * @return The standard value query.
     */
    public static final String getStandardQuery(String fieldName, Operator operator, Object value)
    {
        StringBuilder query = new StringBuilder();
        
        if (operator == Operator.NE)
        {
            NotQuery.appendNegation(query);
        }
        
        query.append(fieldName).append(':');
        
        if (operator == Operator.EQ || operator == Operator.NE)
        {
            query.append(value);
        }
        else if (operator == Operator.GT)
        {
            query.append('{').append(value).append(" TO *]");
        }
        else if (operator == Operator.GE)
        {
            query.append('[').append(value).append(" TO *]");
        }
        else if (operator == Operator.LT)
        {
            query.append("[* TO ").append(value).append('}');
        }
        else if (operator == Operator.LE)
        {
            query.append("[* TO ").append(value).append(']');
        }
        else if (operator == Operator.EXISTS)
        {
            query.append(EXISTS_VALUE);
        }
        
        return query.toString();
    }
    
    /**
     * Escape the query except stars and whitespaces.
     * (Identical to ClientUtils#escapeQueryChars except for stars or whitespaces)
     * @param query the string query to escape
     * @return the escaped query
     */
    @SuppressWarnings("all")
    public static String escapeQueryCharsExceptStarsAndWhitespaces(String query)
    {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < query.length(); i++)
        {
            char c = query.charAt(i);
            // These characters are part of the query syntax and must be escaped (except '*' and whitespaces)
            if (c == '\\' || c == '+' || c == '-' || c == '!'  || c == '(' || c == ')' || c == ':'
                || c == '^' || c == '[' || c == ']' || c == '\"' || c == '{' || c == '}' || c == '~'
                || c == '?' || c == '|' || c == '&'  || c == ';' || c == '/')
            {
                sb.append('\\');
            }
            sb.append(c);
        }
        return sb.toString();
    }
}
