/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

/**
 * Represents a {@link Query} testing a rich text field.
 */
public class RichTextQuery extends AbstractTextQuery
{
    /**
     * Build a StringQuery testing the existence of the field.
     * @param fieldPath the field path
     */
    public RichTextQuery(String fieldPath)
    {
        this(fieldPath, Operator.EXISTS, null, null);
    }
    
    /**
     * Build a string query.
     * @param fieldPath the field's path
     * @param value the value.
     */
    public RichTextQuery(String fieldPath, String value)
    {
        this(fieldPath, value, null);
    }
    
    /**
     * Build a string query.
     * @param fieldPath the field's path
     * @param value the value.
     * @param language the query language (can be null).
     */
    public RichTextQuery(String fieldPath, String value, String language)
    {
        this(fieldPath, Operator.EQ, value, language);
    }
    
    /**
     * Build a string query.
     * @param fieldPath the field's path
     * @param op the operator.
     * @param value the value.
     * @param language the query language (can be null).
     */
    public RichTextQuery(String fieldPath, Operator op, String value, String language)
    {
        this(fieldPath, op, value, language, false);
    }
    
    /**
     * Build a string query.
     * @param fieldPath the field's path
     * @param op the operator.
     * @param value the value.
     * @param language the query language (can be null).
     * @param alreadyEscaped true if the value is already escaped and there is no need to escape again the value during {@link #build() the build of the query}.
     */
    public RichTextQuery(String fieldPath, Operator op, String value, String language, boolean alreadyEscaped)
    {
        super(fieldPath, op, value, language, alreadyEscaped);
    }
    
    @Override
    public String build() throws QuerySyntaxException
    {
        StringBuilder query = new StringBuilder();
        
        if (getOperator() == Operator.EXISTS)
        {
            query.append(getFieldName()).append("_txt");
            if (_language != null)
            {
                query.append("_").append(_language);
            }
            
            query.append(":").append(QueryHelper.EXISTS_VALUE);
            return query.toString();
        }
        
        return super.build();
    }
}
