/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.Arrays;

import org.ametys.cms.content.indexing.solr.SolrFieldNames;

/**
 * Represents a {@link Query} testing if an object possesses one or more tags.
 */
public class TagQuery extends AbstractMultivaluedQuery<String>
{
    /**
     * Build a TagQuery.
     * @param value the tag id to test.
     */
    public TagQuery(String value)
    {
        this(new String[] {value});
    }
    
    /**
     * Build a TagQuery.
     * @param values the tag IDs to test.
     */
    public TagQuery(String... values)
    {
        this(Operator.EQ, values);
    }
    
    /**
     * Build a TagQuery.
     * @param operator the operator.
     * @param value the tag id to test.
     */
    public TagQuery(Operator operator, String value)
    {
        this(operator, new String[] {value});
    }
    
    /**
     * Build a TagQuery.
     * @param operator the operator.
     * @param values the tag IDs to test.
     */
    public TagQuery(Operator operator, String... values)
    {
        this(operator, false, values);
    }
    
    /**
     * Build a TagQuery.
     * @param operator the operator.
     * @param descendantAutoposting true to enable descendant autoposting (parent tags are found), false otherwise.
     * @param value the tag id to test.
     */
    public TagQuery(Operator operator, boolean descendantAutoposting, String value)
    {
        this(operator, descendantAutoposting, new String[] {value});
    }
    
    /**
     * Build a TagQuery.
     * @param operator the operator.
     * @param descendantAutoposting true to enable descendant autoposting (parent tags are found), false otherwise.
     * @param values the tag IDs to test.
     */
    public TagQuery(Operator operator, boolean descendantAutoposting, String... values)
    {
        this(operator, descendantAutoposting, LogicalOperator.OR, values);
    }
    
    /**
     * Build a TagQuery.
     * @param operator the operator. The operator is applied to each value individually. 
     * For instance, if operator is {@link Query.Operator#NE NE} and logicalOperator is
     * {@link Query.LogicalOperator#OR OR}, then the Query represents: <code>(NOT A) OR (NOT B) ...</code>
     * @param descendantAutoposting true to enable descendant autoposting (parent tags are found), false otherwise.
     * @param logicalOperator the logical operator.
     * @param values the tag IDs to test.
     */
    public TagQuery(Operator operator, boolean descendantAutoposting, LogicalOperator logicalOperator, String... values)
    {
        super(descendantAutoposting ? SolrFieldNames.ALL_TAGS : SolrFieldNames.TAGS, operator, logicalOperator, values != null ? Arrays.asList(values) : null);
    }
}
