/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query.join;

import java.util.Objects;
import java.util.Optional;

import org.apache.solr.client.solrj.util.ClientUtils;

import org.ametys.cms.search.query.JoinQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.QuerySyntaxException;

/**
 * Class representing a join key and its (optional) nested query for creating a {@link JoinQuery}
 */
public class JoinKey
{
    private String _key;
    private Optional<String> _joinField;
    private Optional<Query> _nestedQuery;
    
    /**
     * Creates a JoinKey
     * @param key The key. Cannot be null
     */
    public JoinKey(String key)
    {
        this(key, null, null);
    }
    
    /**
     * Creates a JoinKey
     * @param key The key. Cannot be null
     * @param nestedQuery The nested query of the join key. Can be null
     */
    public JoinKey(String key, Query nestedQuery)
    {
        this(key, null, nestedQuery);
    }
    
    /**
     * Creates a JoinKey
     * @param key The key. Cannot be null
     * @param joinField the field to join on. Can be null in which case "id_dv" will be assumed on Solr side.
     * @param nestedQuery The nested query of the join key. Can be null
     */
    public JoinKey(String key, String joinField, Query nestedQuery)
    {
        if (key == null)
        {
            throw new IllegalArgumentException("The key of the JoinKey cannot be null");
        }
        
        _key = key;
        _joinField = Optional.ofNullable(joinField);
        _nestedQuery = Optional.ofNullable(nestedQuery);
    }
    
    /**
     * Gets the key of the JoinKey
     * @return the key of the JoinKey
     */
    public String getKey()
    {
        return _key;
    }
    
    /**
     * Returns the join field, if any.
     * @return the join field, if any.
     */
    public Optional<String> getJoinField()
    {
        return _joinField;
    }

    /**
     * Gets the optional nested query of the join key
     * @return the optional nested query of the join key
     */
    public Optional<Query> getNestedQuery()
    {
        return _nestedQuery;
    }
    
    /**
     * Build the Solr {@link JoinQuery} part representing this {@link JoinKey}.
     * @param escape if the nested query should be escaped
     * @return a Solr query part
     * @throws QuerySyntaxException if the query part can't be built because of a syntax error.
     */
    public String build(boolean escape) throws QuerySyntaxException
    {
        StringBuilder queryString = new StringBuilder();
        
        queryString.append(_key);
        
        if (_joinField.isPresent())
        {
            queryString.append('%').append(_joinField.get());
        }
       
        if (_nestedQuery.isPresent())
        {
            String query = _nestedQuery.get().build();
            queryString.append('[')
                       .append(escape ? ClientUtils.escapeQueryChars(query) : query)
                       .append(']');
        }
        
        return queryString.toString();
    }
    
    @Override
    public int hashCode()
    {
        return Objects.hash(_key, _joinField, _nestedQuery);
    }

    @Override
    public boolean equals(Object obj)
    {
        if (!(obj instanceof JoinKey other))
        {
            return false;
        }
        
        return Objects.equals(_key, other._key)
            && Objects.equals(_joinField, other._joinField)
            && Objects.equals(_nestedQuery, other._nestedQuery);
    }
    
    @Override
    public String toString()
    {
        return "JoinKey(key=" + _key + ",joinField=" + _joinField.orElse("<none>") + ",nestedQuery=" + _nestedQuery.map(Query::toString).orElse("<none>") + ")";
    }
}
