/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr;

import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.TimeUnit;

import org.apache.commons.lang.StringUtils;
import org.apache.solr.client.solrj.SolrServerException;
import org.apache.solr.client.solrj.impl.ConcurrentUpdateHttp2SolrClient;
import org.apache.solr.client.solrj.impl.Http2SolrClient;
import org.apache.solr.client.solrj.response.UpdateResponse;
import org.apache.solr.common.SolrException;
import org.apache.solr.common.SolrInputDocument;
import org.eclipse.jetty.client.api.Response;
import org.slf4j.Logger;

/**
 * Concurrent update solr client specific for Ametys.<br>
 * Allow operations on only one collection (specified by the constructor),
 * and forbid operations on all other collections.
 * Delegate the processing to the ConcurrentUpdateSolrClient (superclass) if the collection is the one in use
 */
public abstract class AbstractAmetysConcurrentUpdateClient extends ConcurrentUpdateHttp2SolrClient
{
    /** Name of the collection, on which operation are allowed */
    private String _collectionName;
    
    private Logger _logger;
    
    /**
     * Constructor
     * @param solrServerUrl The Solr server URL
     * @param solrSocketTimeout The Solr socket timeout (in millis)
     * @param collectionName The name of the collection, on which operations will be allowed. Operations on other collections will be forbidden.
     * @param queueSize The buffer size before the documents are sent to the server
     * @param threadCount The number of background threads used to empty the queue
     * @param logger internal logger
     */
    public AbstractAmetysConcurrentUpdateClient(String solrServerUrl, Optional<Integer> solrSocketTimeout, String collectionName, int queueSize, int threadCount, Logger logger)
    {
        super(_getBuilder(solrServerUrl, solrSocketTimeout, queueSize, threadCount));
        
        if (collectionName == null)
        {
            throw new IllegalArgumentException("The collection name cannot be null.");
        }
        else
        {
            _collectionName = collectionName;
        }
        
        _logger = logger;
    }
    
    private static Builder _getBuilder(String solrServerUrl, Optional<Integer> solrSocketTimeout, int queueSize, int threadCount)
    {
        Http2SolrClient.Builder solrReadClientBuilder = new Http2SolrClient.Builder(solrServerUrl);
        if (solrSocketTimeout.isPresent())
        {
            solrReadClientBuilder.withRequestTimeout(solrSocketTimeout.get(), TimeUnit.MILLISECONDS);
        }
        
        Http2SolrClient solrReadClient = solrReadClientBuilder.build();
        
        Builder builder = new Builder(solrServerUrl, solrReadClient)
            .withQueueSize(queueSize)
            .withThreadCount(threadCount)
            .setPollQueueTime(10, TimeUnit.MILLISECONDS);
        
        return builder;
    }
    
    @Override
    public void handleError(Throwable t)
    {
        StringBuilder msg = new StringBuilder("Solr client indexing error");
        if (t instanceof SolrException)
        {
            msg.append(" caused by the following errors: ").append(((SolrException) t).getMetadata());
        }
        
        msg.append("\nPlease check the Solr logs for more details.");
        _logger.error(msg.toString(), t);
    }
    
    @Override
    public void onSuccess(Response resp, InputStream respBody)
    {
        if (_logger.isDebugEnabled())
        {
            String reason = resp.getReason();
            _logger.debug("Successfully processed solr client request and obtained the response : " + StringUtils.defaultIfEmpty(reason, "unknown"));
        }
        
        super.onSuccess(resp, respBody);
    }
    
    private void _checkCollectionInUse(String collection) throws SolrServerException
    {
        if (collection == null)
        {
            throw new UnsupportedOperationException("Collection cannot be null");
        }
        
        if (!StringUtils.equals(_collectionName, collection))
        {
            String msg = String.format("Cannot process this update operation for collection '%s' because operations for this client can only be done in '%s'", collection, _collectionName);
            throw new SolrServerException(msg);
        }
    }
    
    @Override
    public UpdateResponse add(String collection, Collection<SolrInputDocument> docs) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.add(collection, docs);
    }
    
    @Override
    public UpdateResponse add(Collection<SolrInputDocument> docs) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.add(docs);
    }
    
    @Override
    public UpdateResponse add(String collection, Collection<SolrInputDocument> docs, int commitWithinMs) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.add(collection, docs, commitWithinMs);
    }
    
    @Override
    public UpdateResponse add(Collection<SolrInputDocument> docs, int commitWithinMs) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.add(docs, commitWithinMs);
    }
    
    @Override
    public UpdateResponse add(String collection, SolrInputDocument doc) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.add(collection, doc);
    }
    
    @Override 
    public UpdateResponse add(SolrInputDocument doc) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.add(doc);
    }
    
    @Override
    public UpdateResponse add(String collection, SolrInputDocument doc, int commitWithinMs) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.add(collection, doc, commitWithinMs);
    }
    
    @Override
    public UpdateResponse add(SolrInputDocument doc, int commitWithinMs) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.add(doc, commitWithinMs);
    }
    
    @Override
    public UpdateResponse add(String collection, Iterator<SolrInputDocument> docIterator) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.add(collection, docIterator);
    }
    
    @Override
    public UpdateResponse add(Iterator<SolrInputDocument> docIterator) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.add(docIterator);
    }
    
    @Override
    public UpdateResponse addBean(String collection, Object obj) throws IOException, SolrServerException
    {
        _checkCollectionInUse(collection);
        return super.addBean(collection, obj);
    }
    
    @Override
    public UpdateResponse addBean(Object obj) throws IOException, SolrServerException
    {
        _checkCollectionInUse(null);
        return super.addBean(obj);
    }
    
    @Override
    public UpdateResponse addBean(String collection, Object obj, int commitWithinMs) throws IOException, SolrServerException
    {
        _checkCollectionInUse(collection);
        return super.addBean(collection, obj, commitWithinMs);
    }
    
    @Override
    public UpdateResponse addBean(Object obj, int commitWithinMs) throws IOException, SolrServerException
    {
        _checkCollectionInUse(null);
        return super.addBean(obj, commitWithinMs);
    }
    
    @Override
    public UpdateResponse addBeans(String collection, Collection<?> beans) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.addBeans(collection, beans);
    }
    
    @Override
    public UpdateResponse addBeans(Collection<?> beans) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.addBeans(beans);
    }
    
    @Override
    public UpdateResponse addBeans(String collection, Collection<?> beans, int commitWithinMs) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.addBeans(collection, beans, commitWithinMs);
    }
    
    @Override
    public UpdateResponse addBeans(Collection<?> beans, int commitWithinMs) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.addBeans(beans, commitWithinMs);
    }
    
    @Override
    public UpdateResponse addBeans(String collection, final Iterator<?> beanIterator) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.addBeans(collection, beanIterator);
    }
    
    @Override
    public UpdateResponse addBeans(final Iterator<?> beanIterator) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.addBeans(beanIterator);
    }
    
    @Override
    public UpdateResponse commit(String collection) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.commit(collection);
    }
    
    @Override
    public UpdateResponse commit() throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.commit();
    }
    
    @Override
    public UpdateResponse commit(String collection, boolean waitFlush, boolean waitSearcher) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.commit(collection, waitFlush, waitSearcher);
    }
    
    @Override
    public UpdateResponse commit(boolean waitFlush, boolean waitSearcher) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.commit(waitFlush, waitSearcher);
    }
    
    @Override
    public UpdateResponse commit(String collection, boolean waitFlush, boolean waitSearcher, boolean softCommit) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.commit(collection, waitFlush, waitSearcher, softCommit);
    }
    
    @Override
    public UpdateResponse commit(boolean waitFlush, boolean waitSearcher, boolean softCommit) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.commit(waitFlush, waitSearcher, softCommit);
    }
    
    @Override
    public UpdateResponse optimize(String collection) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.optimize(collection);
    }
    
    @Override
    public UpdateResponse optimize() throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.optimize();
    }
    
    @Override
    public UpdateResponse optimize(String collection, boolean waitFlush, boolean waitSearcher) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.optimize(collection, waitFlush, waitSearcher);
    }
    
    @Override
    public UpdateResponse optimize(boolean waitFlush, boolean waitSearcher) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.optimize(waitFlush, waitSearcher);
    }
    
    @Override
    public UpdateResponse optimize(String collection, boolean waitFlush, boolean waitSearcher, int maxSegments) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.optimize(collection, waitFlush, waitSearcher, maxSegments);
    }
    
    @Override
    public UpdateResponse optimize(boolean waitFlush, boolean waitSearcher, int maxSegments) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.optimize(waitFlush, waitSearcher, maxSegments);
    }
    
    @Override
    public UpdateResponse rollback(String collection) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.rollback(collection);
    }
    
    @Override
    public UpdateResponse rollback() throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.rollback();
    }
    
    @Override
    public UpdateResponse deleteById(String collection, String id) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.deleteById(collection, id);
    }
    
    @Override
    public UpdateResponse deleteById(String id) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.deleteById(id);
    }
    
    @Override
    public UpdateResponse deleteById(String collection, String id, int commitWithinMs) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.deleteById(collection, id, commitWithinMs);
    }
    
    @Override
    public UpdateResponse deleteById(String id, int commitWithinMs) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.deleteById(id, commitWithinMs);
    }
    
    @Override
    public UpdateResponse deleteById(String collection, List<String> ids) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.deleteById(collection, ids);
    }
    
    @Override
    public UpdateResponse deleteById(List<String> ids) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.deleteById(ids);
    }
    
    @Override
    public UpdateResponse deleteById(String collection, List<String> ids, int commitWithinMs) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.deleteById(collection, ids, commitWithinMs);
    }
    
    @Override
    public UpdateResponse deleteById(List<String> ids, int commitWithinMs) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.deleteById(ids, commitWithinMs);
    }
    
    @Override
    public UpdateResponse deleteByQuery(String collection, String query) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.deleteByQuery(collection, query);
    }
    
    @Override
    public UpdateResponse deleteByQuery(String query) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.deleteByQuery(query);
    }
    
    @Override
    public UpdateResponse deleteByQuery(String collection, String query, int commitWithinMs) throws SolrServerException, IOException
    {
        _checkCollectionInUse(collection);
        return super.deleteByQuery(collection, query, commitWithinMs);
    }
    
    @Override
    public UpdateResponse deleteByQuery(String query, int commitWithinMs) throws SolrServerException, IOException
    {
        _checkCollectionInUse(null);
        return super.deleteByQuery(query, commitWithinMs);
    }
}
