/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr;

import java.util.Iterator;
import java.util.NoSuchElementException;

import org.apache.solr.common.SolrDocument;
import org.slf4j.Logger;

import org.ametys.cms.content.indexing.solr.SolrFieldNames;
import org.ametys.cms.search.SearchResultsIterator;
import org.ametys.plugins.repository.AmetysObjectResolver;

/**
 * This abstract iterator is here for providing common code to an iterator over
 * SearchResult and to an AmetysObjectIterator
 * @param <O> the type of elements returned by this iterator
 */
abstract class AbstractResponseIterator<O> implements SearchResultsIterator<O>
{
    protected AmetysObjectResolver _ametysObjectResolver;
    private Iterator<SolrDocument> _it;
    private O _nextObject;
    private int _position;
    private long _size;
    private long _errors;
    private Logger _logger;
    
    public AbstractResponseIterator(Iterator<SolrDocument> it, long size, AmetysObjectResolver resolver, Logger logger)
    {
        _it = it;
        _size = size;
        _errors = 0;
        _ametysObjectResolver = resolver;
        _logger = logger;
    }
    
    @Override
    public boolean hasNext()
    {
        // Prefetch the next object
        if (_nextObject == null)
        {
            while (_it.hasNext())
            {
                SolrDocument document = _it.next();
                String id = (String) document.getFieldValue(SolrFieldNames.ID);
                
                if (_ametysObjectResolver.hasAmetysObjectForId(id))
                {
                    _nextObject = _setNextObject(document, id);
                    return true;
                }
                else
                {
                    // Go to next element
                    _errors++;
                    _logger.warn("The Solr indexed Ametys object with id '{}' does not exist. Either autoCommit did not execute yet or a full reindexation is required.", id);
                }
            }
            return false;
        }
        return true;
    }
    
    protected abstract O _setNextObject(SolrDocument document, String id);
    
    @Override
    public O next()
    {
        if (!hasNext())
        {
            throw new NoSuchElementException();
        }
        
        try
        {
            _position++;
            return _nextObject;
        }
        finally
        {
            _nextObject = null;
        }
    }
    
    public long getPosition()
    {
        return _position;
    }
    
    public long getSize()
    {
        return _size - _errors;
    }
    
    @Override
    public void skip(long skipNum)
    {
        if (skipNum <= 0)
        {
            return;
        }
        
        long size = getSize();
        if (size != -1 && skipNum > size - getPosition())
        {
            throw new NoSuchElementException();
        }
        
        for (long i = skipNum; i > 0; i--)
        {
            next();
        }
    }
}
