/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr;

import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.quartz.JobKey;
import org.quartz.SchedulerException;

import org.ametys.core.schedule.Runnable;
import org.ametys.core.schedule.Runnable.FireProcess;
import org.ametys.core.ui.Callable;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.core.impl.schedule.DefaultRunnable;
import org.ametys.plugins.core.schedule.Scheduler;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Export a solr search
 */
public class AsyncExportLauncher extends AbstractLogEnabled implements Component, Serviceable
{
    /** The Avalon Role */
    public static final String ROLE = AsyncExportLauncher.class.getName();
    
    /** The provider of current user */
    protected CurrentUserProvider _currentUserProvider;
    
    /** Scheduler */
    protected Scheduler _scheduler;

    public void service(ServiceManager manager) throws ServiceException
    {
        _currentUserProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
        _scheduler = (Scheduler) manager.lookup(Scheduler.ROLE);
    }
    
    /**
     * Launch a search export
     * @param type csv, xml or word
     * @param recipient email of the recipient of the export
     * @param searchParams search parameters
     * @param lang language for the mail
     * @param exportUrl the internal url for export
     * @return A result map
     * @throws SchedulerException if an error occured
     */
    @Callable
    public Map<String, Object> add(String type, String recipient, String searchParams, String lang, String exportUrl)  throws SchedulerException
    {
        I18nizableText label = new I18nizableText("plugin.cms", "UITOOL_SEARCH_ASYNC_EXPORT_TASK_TITLE");
        I18nizableText description = new I18nizableText("plugin.cms", "UITOOL_SEARCH_ASYNC_EXPORT_TASK_DESCRIPTION");
        
        
        String id = _generateUniqueId();

        String cron = "";
        String schedulableId = "org.ametys.cms.search.schedule.Export";
        
        UserIdentity user = _currentUserProvider.getUser();
        
        Map<String, Object> typedParams = new HashMap<>();
        typedParams.put("type", type);
        typedParams.put("recipient", recipient);
        typedParams.put("searchParams", searchParams);
        typedParams.put("lang", lang);
        typedParams.put("exportUrl", exportUrl);
        
        Runnable runnable = new DefaultRunnable(id, label, description, FireProcess.NOW, cron, schedulableId, false, false, false, null, true, user, typedParams);
        
        Map<String, Object> result = new HashMap<>();
        try
        {
            _scheduler.scheduleJob(runnable);
        }
        catch (SchedulerException e)
        {
            getLogger().error("An error occured when trying to create the task " + id, e);
            result.put("error", "scheduler-error");
            return result;
        }
        
        result.put("id", id);
        return result;
    }
    
    /**
     * Generate an unique ID for a scheduler label
     * @return a unique id for this sceduler
     * @throws SchedulerException Something went wrong
     */
    protected String _generateUniqueId() throws SchedulerException
    {
        String value = "async-export-solr-search";
        int i = 2;
        String suffixedValue = value;
        while (_scheduler.getScheduler().checkExists(new JobKey(suffixedValue, Scheduler.JOB_GROUP)))
        {
            suffixedValue = value + i;
            i++;
        }
        
        return suffixedValue;
    }

}
