/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;

import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.model.SearchCriterion;
import org.ametys.cms.search.model.SearchCriterionHelper;
import org.ametys.cms.search.model.SearchModel;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.ui.model.ColumnHelper;
import org.ametys.cms.search.ui.model.ColumnHelper.Column;
import org.ametys.cms.search.ui.model.SearchUICriterion;
import org.ametys.cms.search.ui.model.SearchUIModel;
import org.ametys.cms.search.ui.model.SearchUIModelExtensionPoint;
import org.ametys.cms.search.ui.model.impl.IndexingFieldSearchUICriterion;
import org.ametys.cms.search.ui.model.impl.SystemSearchUICriterion;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.ViewItemContainer;
import org.ametys.runtime.plugin.component.LogEnabled;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * Search model wrapper which handles custom on-the-fly columns and facets.
 */
public class CriteriaSearchModelWrapper implements SearchModel, LogEnabled
{
    /** ComponentManager for {@link SearchUICriterion}s. */
    protected ThreadSafeComponentManager<SearchCriterion> _searchCriterionManager;
    
    /** The logger. */
    protected Logger _logger;
    
    private ServiceManager _serviceManager;
    
    private ContentTypeExtensionPoint _contentTypeExtensionPoint;
    private SystemPropertyExtensionPoint _systemPropertyExtensionPoint;
    private SearchUIModelExtensionPoint _searchUIModelExtensionPoint;
    private SearchCriterionHelper _searchCriterionHelper;
    private ColumnHelper _columnHelper;
    
    private SearchModel _wrappedModel;
    
    /** The search criteria used as facets, indexed by ID. */
    private Map<String, SearchCriterion> _facetedCriteria;
    /** The result items */
    private ViewItemContainer _resultItems;
    
    private int _criteriaIndex;
    
    /**
     * Build a model wrapper.
     * @param model the search model to wrap.
     * @param manager the service manager.
     * @param context the component context.
     * @param logger the logger.
     */
    public CriteriaSearchModelWrapper(SearchModel model, ServiceManager manager, Context context, Logger logger)
    {
        _wrappedModel = model;

        _serviceManager = manager;
        _logger = logger;
        
        try
        {
            _searchCriterionManager = new ThreadSafeComponentManager<>();
            _searchCriterionManager.setLogger(logger);
            _searchCriterionManager.contextualize(context);
            _searchCriterionManager.service(manager);
        }
        catch (Exception e)
        {
            _logger.error("Error initializing the SearchModel", e);
        }
    }
    
    public void setLogger(final Logger logger)
    {
        _logger = logger;
    }
    
    /**
     * Get the logger.
     * @return the logger.
     */
    protected final Logger getLogger()
    {
        return _logger;
    }
    
    /**
     * Set the custom faceted criteria.
     * @param baseContentTypeIds the reference content type identifiers.
     * @param criteriaIds the criteria IDs
     * @param contextualParameters the contextual parameters
     * @return the valid criteria IDs
     * @throws Exception if an error occurs initializing criteria.
     */
    public Collection<String> setFacetedCriteria(Set<String> baseContentTypeIds, Collection<String> criteriaIds, Map<String, Object> contextualParameters) throws Exception
    {
        Collection<String> resultCriteriaIds = null;
        
        if (criteriaIds != null)
        {
            _facetedCriteria = new LinkedHashMap<>(criteriaIds.size());
            
            List<Object> searchToolCriterionRoles = new ArrayList<>();
            
            resultCriteriaIds = configureFacets(searchToolCriterionRoles, baseContentTypeIds, criteriaIds, _wrappedModel, contextualParameters);
            
            _searchCriterionManager.initialize();
            
            for (Object critObj : searchToolCriterionRoles)
            {
                SearchCriterion criterion = null;
                if (critObj instanceof SearchCriterion)
                {
                    // Already existing SearchCriterion object (taken from the wrapped model).
                    criterion = (SearchCriterion) critObj;
                }
                else if (critObj instanceof String)
                {
                    // Criterion just added in the local component manager, we have to look it up.
                    criterion = _searchCriterionManager.lookup((String) critObj);
                }
                
                if (criterion != null && criterion.isFacetable())
                {
                    _facetedCriteria.put(criterion.getId(), criterion);
                }
            }
        }
        
        return resultCriteriaIds != null ? resultCriteriaIds : Collections.emptySet();
    }
    
    /**
     * Set the custom columns.
     * @param baseContentTypeIds the reference content type identifiers.
     * @param columns The columns
     * @param contextualParameters the contextual parameters
     * @throws Exception if an error occurs initializing columns.
     */
    public void setResultColumns(Set<String> baseContentTypeIds, Collection<Column> columns, Map<String, Object> contextualParameters) throws Exception
    {
        String wrappedModelId = (String) contextualParameters.get("wrappedModelId");
        if (StringUtils.isNotEmpty(wrappedModelId))
        {
            // Dashboard
            SearchUIModel model = _getSearchUIModelExtensionPoint().getExtension(wrappedModelId);
            _resultItems = model.getResultItems(contextualParameters);
        }
        else if (columns != null)
        {
            _resultItems = _getColumnHelper()._createViewFromColumns(baseContentTypeIds, columns, false);
        }
    }
    
    /**
     * Configure the list of faceted criteria.
     * @param criteriaRoles the roles of criteria to lookup (or the already existing SearchUICriterion objects).
     * @param baseContentTypeIds The reference content type identifiers.
     * @param criteriaIds the criteria IDs.
     * @param referenceModel the reference model.
     * @param contextualParameters the contextual parameters
     * @return the valid criteria IDs
     * @throws ConfigurationException if an error occurs creating a component configuration.
     */
    protected Collection<String> configureFacets(List<Object> criteriaRoles, Set<String> baseContentTypeIds, Collection<String> criteriaIds, SearchModel referenceModel, Map<String, Object> contextualParameters) throws ConfigurationException
    {
        Collection<String> resultCriteriaIds = new ArrayList<>();
        
        for (String criterionId : criteriaIds)
        {
            String[] facetPathSegments = StringUtils.split(criterionId, '/');
            String lastSegmentOfFacetPath = facetPathSegments[facetPathSegments.length - 1];
            
            SearchCriterion referenceCriterion = null;
            if (referenceModel != null)
            {
                referenceCriterion = getCriterion(referenceModel, criterionId, contextualParameters);
            }
            
            if (referenceCriterion != null)
            {
                criteriaRoles.add(referenceCriterion);
            }
            else if (_getSystemPropertyExtensionPoint().hasExtension(lastSegmentOfFacetPath))
            {
                SystemProperty systemProperty = _getSystemPropertyExtensionPoint().getExtension(lastSegmentOfFacetPath);
                if (systemProperty.isFacetable())
                {
                    addSystemCriteriaComponents(criteriaRoles, baseContentTypeIds, criterionId);
                }
                else
                {
                    getLogger().warn("The declared facet '{}' is a system property but is not facetable. Thus, it will not be added to the facets.", criterionId);
                    break;
                }
            }
            else
            {
                if (!baseContentTypeIds.isEmpty())
                {
                    // Metadata property.
                    addIndexingFieldCriteriaComponents(criteriaRoles, baseContentTypeIds, criterionId);
                }
                else if (Content.ATTRIBUTE_TITLE.equals(criterionId))
                {
                    // title property of a random ContentType.
                    String firstCTypeId = _getContentTypeExtensionPoint().getExtensionsIds().iterator().next();
                    addIndexingFieldCriteriaComponents(criteriaRoles, Set.of(firstCTypeId), criterionId);
                }
                else
                {
                    break;
                }
            }
            
            resultCriteriaIds.add(criterionId);
        }
        
        return resultCriteriaIds;
    }
    
    /**
     * Search a criterion in the reference model from its criterion identifier.
     * @param searchModel the reference search model.
     * @param criterionId the criterion identifier.
     * @param contextualParameters the contextual parameters
     * @return the criterion if found, null otherwise.
     */
    protected SearchCriterion getCriterion(SearchModel searchModel, String criterionId, Map<String, Object> contextualParameters)
    {
        Map<String, ? extends SearchCriterion> criteria = searchModel.getFacetedCriteria(contextualParameters);
        
        for (SearchCriterion criterion : criteria.values())
        {
            if (criterion instanceof IndexingFieldSearchUICriterion && ((IndexingFieldSearchUICriterion) criterion).getFieldPath().equals(criterionId))
            {
                return criterion;
            }
            else if (criterion instanceof SystemSearchUICriterion && ((SystemSearchUICriterion) criterion).getSystemPropertyId().equals(criterionId))
            {
                return criterion;
            }
            else if (criterion.getId().equals(criterionId))
            {
                return criterion;
            }
        }
        
        return null;
    }
    

    /**
     * Add a indexing field criteria component to the manager.
     * @param searchToolCriterionRoles the criteria role list to fill.
     * @param baseContentTypeIds the reference content type identifiers.
     * @param fieldRef the field path.
     * @throws ConfigurationException if an error occurs.
     */
    protected void addIndexingFieldCriteriaComponents(List<Object> searchToolCriterionRoles, Set<String> baseContentTypeIds, String fieldRef) throws ConfigurationException
    {
        try
        {
            String slashPath = fieldRef.replace('.', '/');
            
            String role = fieldRef + _criteriaIndex;
            _criteriaIndex++;
            Configuration criteriaConf = _getSearchCriterionHelper().getIndexingFieldCriteriaConfiguration(this, Optional.empty(), baseContentTypeIds, slashPath, Optional.of(Operator.EQ), Optional.empty());
            
            _searchCriterionManager.addComponent("search", null, role, IndexingFieldSearchUICriterion.class, criteriaConf);
            
            searchToolCriterionRoles.add(role);
        }
        catch (Exception e)
        {
            throw new ConfigurationException("Unable to instanciate IndexingFieldSearchUICriterion for field " + fieldRef, e);
        }        
    }
    
    /**
     * Add a system criteria component to the manager.
     * @param searchToolCriterionRoles the criteria role list to fill.
     * @param baseContentTypeIds the reference content type identifiers.
     * @param property the system property id.
     * @throws ConfigurationException if an error occurs.
     */
    protected void addSystemCriteriaComponents(List<Object> searchToolCriterionRoles, Set<String> baseContentTypeIds, String property) throws ConfigurationException
    {
        try
        {
            String role = property + _criteriaIndex;
            _criteriaIndex++;
            
            Configuration criteriaConf = _getSearchCriterionHelper().getSystemCriteriaConfiguration(this, Optional.empty(), baseContentTypeIds, property, Optional.empty());
            _searchCriterionManager.addComponent("search", null, role, SystemSearchUICriterion.class, criteriaConf);
            
            searchToolCriterionRoles.add(role);
        }
        catch (Exception e)
        {
            throw new ConfigurationException("Unable to instanciate SystemSearchUICriterion for property " + property, e);
        }
    }
    
    @Override
    public Map<String, ? extends SearchCriterion> getFacetedCriteria(Map<String, Object> contextualParameters)
    {
        if (_facetedCriteria != null && !_facetedCriteria.isEmpty())
        {
            return Collections.unmodifiableMap(_facetedCriteria);
        }
        else
        {
            return _wrappedModel.getFacetedCriteria(contextualParameters);
        }
    }
    
    @Override
    public ViewItemContainer getResultItems(Map<String, Object> contextualParameters)
    {
        return Optional.ofNullable(_resultItems)
                       .filter(viewItemContainer -> !(viewItemContainer.getViewItems().isEmpty()))
                       .orElseGet(() -> _wrappedModel.getResultItems(contextualParameters));
    }
    
    public ModelViewItem getResultItem(String itemPath, Map<String, Object> contextualParameters)
    {
        return Optional.ofNullable(_resultItems)
                       .map(ri -> SearchModel.super.getResultItem(itemPath, contextualParameters))
                       .orElseGet(() -> _wrappedModel.getResultItem(itemPath, contextualParameters));
    }
    
    //// PROXY METHODS ////
    
    @Override
    public Set<String> getContentTypes(Map<String, Object> contextualParameters)
    {
        return _wrappedModel.getContentTypes(contextualParameters);
    }
    
    @Override
    public Set<String> getExcludedContentTypes(Map<String, Object> contextualParameters)
    {
        return _wrappedModel.getExcludedContentTypes(contextualParameters);
    }
    
    @Override
    public Map<String, ? extends SearchCriterion> getCriteria(Map<String, Object> contextualParameters)
    {
        return _wrappedModel.getCriteria(contextualParameters);
    }
    
    @Override
    public SearchCriterion getCriterion(String id, Map<String, Object> contextualParameters)
    {
        return _wrappedModel.getCriterion(id, contextualParameters);
    }
    
    @Override
    public String getWorkspace(Map<String, Object> contextualParameters)
    {
        return _wrappedModel.getWorkspace(contextualParameters);
    }
    
    private ContentTypeExtensionPoint _getContentTypeExtensionPoint()
    {
        if (_contentTypeExtensionPoint == null)
        {
            try
            {
                _contentTypeExtensionPoint = (ContentTypeExtensionPoint) _serviceManager.lookup(ContentTypeExtensionPoint.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException("Unable to lookup after the content type extension point", e);
            }
        }
        
        return _contentTypeExtensionPoint;
    }
    
    private SystemPropertyExtensionPoint _getSystemPropertyExtensionPoint()
    {
        if (_systemPropertyExtensionPoint == null)
        {
            try
            {
                _systemPropertyExtensionPoint = (SystemPropertyExtensionPoint) _serviceManager.lookup(SystemPropertyExtensionPoint.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException("Unable to lookup after the system property extension point", e);
            }
        }
        
        return _systemPropertyExtensionPoint;
    }
    
    private SearchUIModelExtensionPoint _getSearchUIModelExtensionPoint()
    {
        if (_searchUIModelExtensionPoint == null)
        {
            try
            {
                _searchUIModelExtensionPoint = (SearchUIModelExtensionPoint) _serviceManager.lookup(SearchUIModelExtensionPoint.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException("Unable to lookup after the search ui model extension point", e);
            }
        }
        
        return _searchUIModelExtensionPoint;
    }
    
    private SearchCriterionHelper _getSearchCriterionHelper()
    {
        if (_searchCriterionHelper == null)
        {
            try
            {
                _searchCriterionHelper = (SearchCriterionHelper) _serviceManager.lookup(SearchCriterionHelper.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException("Unable to lookup after the search criterion helper", e);
            }
        }
        
        return _searchCriterionHelper;
    }
    
    private ColumnHelper _getColumnHelper()
    {
        if (_columnHelper == null)
        {
            try
            {
                _columnHelper = (ColumnHelper) _serviceManager.lookup(ColumnHelper.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException("Unable to lookup after the column helper", e);
            }
        }
        
        return _columnHelper;
    }
}
