/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr;

import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.TimeUnit;

import javax.jcr.RepositoryException;

import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.solr.client.solrj.SolrClient;
import org.apache.solr.client.solrj.impl.Http2SolrClient;

import org.ametys.plugins.repository.provider.AbstractRepository;
import org.ametys.plugins.repository.provider.JackrabbitRepository;
import org.ametys.plugins.repository.provider.WorkspaceSelector;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Component acting as a single entry point to get access to Solr clients.
 */
public class DefaultSolrClientProvider extends AbstractLogEnabled implements SolrClientProvider, Component, Serviceable, Initializable, Disposable
{
    private static final String __SOLR_URL_CONFIG = "cms.solr.core.url"; 
    private static final String __SOLR_SOCKET_TIMEOUT_CONFIG = "cms.solr.socket.timeout";
    private static final String __SOLR_CORE_PREFIX_CONFIG = "cms.solr.core.prefix";  
    
    /** The workspace selector. */
    protected WorkspaceSelector _workspaceSelector;
    
    /** The JCR repository */
    protected JackrabbitRepository _repository;

    /** The Solr read client. */
    protected SolrClient _solrReadClient;
    
    /** The Solr "default" update clients, per workspace */
    protected Map<String, SolrClient> _solrDefaultUpdateClients;
    
    /** The Solr "no auto commit" update clients, per workspace */
    protected Map<String, SolrClient> _solrNoAutoCommitUpdateClients;

    /** The solr URL. */
    protected String _solrUrl;
    
    /** The solr socket timeout (in millis). */
    protected Optional<Integer> _solrSocketTimeout;
    
    /** The solr core prefix. */
    protected String _solrCorePrefix;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        _workspaceSelector = (WorkspaceSelector) serviceManager.lookup(WorkspaceSelector.ROLE);
        _repository = (JackrabbitRepository) serviceManager.lookup(AbstractRepository.ROLE);
    }
    
    @Override
    public void initialize() throws Exception
    {
        _solrUrl = Config.getInstance().getValue(__SOLR_URL_CONFIG);
        _solrSocketTimeout = Optional.of(__SOLR_SOCKET_TIMEOUT_CONFIG)
                .map(Config.getInstance()::<Long>getValue)
                .map(Long::intValue)
                .map(s -> s * 1000);
        _solrCorePrefix = Config.getInstance().getValue(__SOLR_CORE_PREFIX_CONFIG);
        
        Http2SolrClient.Builder solrReadClientBuilder = new Http2SolrClient.Builder(_solrUrl);
        if (_solrSocketTimeout.isPresent())
        {
            solrReadClientBuilder.withRequestTimeout(_solrSocketTimeout.get(), TimeUnit.MILLISECONDS);
        }
        _solrReadClient = solrReadClientBuilder.build();
        
        String[] workspaces = _repository.getWorkspaces();
        _solrDefaultUpdateClients = new HashMap<>();
        _solrNoAutoCommitUpdateClients = new HashMap<>();
        for (String workspaceName : workspaces)
        {
            _solrDefaultUpdateClients.put(workspaceName, _createDefaultUpdateClient(workspaceName));
            _solrNoAutoCommitUpdateClients.put(workspaceName, _createNoAutoCommitUpdateClient(workspaceName));
        }
    }
    
    private SolrClient _createDefaultUpdateClient(String workspaceName)
    {
        return new DefaultUpdateClient(_solrUrl, _solrSocketTimeout, getCollectionName(workspaceName), 10, 4, getLogger());
    }
    
    private SolrClient _createNoAutoCommitUpdateClient(String workspaceName)
    {
        return new NoAutoCommitUpdateClient(_solrUrl, _solrSocketTimeout, getCollectionName(workspaceName), 10, 4, getLogger());
    }
    
    @Override
    public void dispose()
    {
        // Release the solr clients (as a Closeable).
        IOUtils.closeQuietly(_solrReadClient);
        _solrReadClient = null;
        
        for (SolrClient solrDefaultUpdateClient : _solrDefaultUpdateClients.values())
        {
            IOUtils.closeQuietly(solrDefaultUpdateClient);
        }
        _solrDefaultUpdateClients.clear();
        _solrDefaultUpdateClients = null;
        
        for (SolrClient solrNoAutoCommitUpdateClient : _solrNoAutoCommitUpdateClients.values())
        {
            IOUtils.closeQuietly(solrNoAutoCommitUpdateClient);
        }
        _solrNoAutoCommitUpdateClients.clear();
        _solrNoAutoCommitUpdateClients = null;
    }
    
    @Override
    public SolrClient getReadClient()
    {
        return _solrReadClient;
    }
    
    @Override
    public SolrClient getUpdateClient(String workspaceName, boolean autoCommit)
    {
        Map<String, SolrClient> updateClients = autoCommit ? _solrDefaultUpdateClients : _solrNoAutoCommitUpdateClients;
        
        SolrClient updateClient = updateClients.get(_nonNullWorkspaceName(workspaceName));
        if (updateClient == null)
        {
            // Perhaps the workspace was created after initializing this component, try to check if JCR workspace exist
            try
            {
                if (ArrayUtils.contains(_repository.getWorkspaces(), workspaceName))
                {
                    updateClient = autoCommit ? _createDefaultUpdateClient(workspaceName) : _createNoAutoCommitUpdateClient(workspaceName);
                    updateClients.put(workspaceName, updateClient);
                }
            }
            catch (RepositoryException e)
            {
                getLogger().error("An error occurs while trying to return all JCR workspaces", e);
            }
        }
        
        return updateClient;
    }
    
    @Override
    public String getCollectionName()
    {
        return getCollectionName(_workspaceSelector.getWorkspace());
    }
    
    @Override
    public String getCollectionName(String workspaceName)
    {
        return _solrCorePrefix + _nonNullWorkspaceName(workspaceName);
    }
    
    private String _nonNullWorkspaceName(String workspaceName)
    {
        if (workspaceName == null)
        {
            getLogger().debug("Passing null workspace name. Switching to current workspace.");
            return _workspaceSelector.getWorkspace();
        }
        else
        {
            return workspaceName;
        }
    }
}
