/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;

import org.ametys.cms.search.GetQueryFromJSONHelper;
import org.ametys.cms.search.model.SearchCriterion;
import org.ametys.cms.search.model.SearchModel;
import org.ametys.cms.search.query.Query;
import org.ametys.core.cocoon.JSonReader;
import org.ametys.core.util.ServerCommHelper;

/**
 * Get a solr query string from a search tool.
 */
public class GetSolrQueryAction extends ServiceableAction
{
    /** The server comm helper */
    protected ServerCommHelper _serverCommHelper;
    
    /** the helper to get query infos from JSON */
    protected GetQueryFromJSONHelper _getQueryFromJSONHelper;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _serverCommHelper = (ServerCommHelper) smanager.lookup(ServerCommHelper.ROLE);
        _getQueryFromJSONHelper = (GetQueryFromJSONHelper) smanager.lookup(GetQueryFromJSONHelper.ROLE);
    }
    
    @Override
    @SuppressWarnings ("unchecked")
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        Map<String, Object> queryResults = new HashMap<>();
        request.setAttribute(JSonReader.OBJECT_TO_READ, queryResults);
        
        Map<String, Object> jsParameters = _serverCommHelper.getJsParameters();
        
        
        Map<String, Object> contextualParameters = (Map<String, Object>) jsParameters.get("contextualParameters");
        
        if (contextualParameters == null)
        {
            contextualParameters = new HashMap<> ();
        }
        
        SearchModel model = _getQueryFromJSONHelper.getSearchModel(jsParameters, "id");
        
        try
        {
            List<String> cTypesToFill = new ArrayList<>();
            Query query = _getQueryFromJSONHelper.getQueryFromModel(model, jsParameters, cTypesToFill);
            
            // Set the query string into the results.
            queryResults.put("query", query.build());
            queryResults.put("contentTypes", cTypesToFill);
            
            // Get the model facets and set them into the results.
            Map<String, ? extends SearchCriterion> facetedCriteria = model.getFacetedCriteria(contextualParameters);
            setFacets(queryResults, facetedCriteria.values());
            
            // Get the model columns and set them into the results.
            List<String> columnIds = _getQueryFromJSONHelper.getColumnsFromSearchModel(model, contextualParameters);
            queryResults.put("columns", columnIds);
        }
        catch (Exception e)
        {
            getLogger().error("Cannot get solr query string: " + e.getMessage(), e);
            throw new ProcessingException("Cannot get solr query string: " + e.getMessage(), e);
        }
        
        return EMPTY_MAP;
    }

    /**
     * Set the facet list into the query results.
     * @param queryResults the query results.
     * @param facetedCriteria the facets to put in the results.
     */
    protected void setFacets(Map<String, Object> queryResults, Collection<? extends SearchCriterion> facetedCriteria)
    {
        List<String> facets = new ArrayList<>();
        queryResults.put("facets", facets);
        
        for (SearchCriterion criterion : facetedCriteria)
        {
            facets.add(criterion.getSearchField().getName());
        }
    }
}
