/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.container.ContainerUtil;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.search.cocoon.SearchAction;
import org.ametys.cms.search.model.SearchModelHelper;
import org.ametys.cms.search.ui.model.ColumnHelper;
import org.ametys.cms.search.ui.model.ColumnHelper.Column;
import org.ametys.cms.search.ui.model.SearchUIModel;
import org.ametys.core.cocoon.JSonReader;
import org.ametys.core.util.AvalonLoggerAdapter;
import org.ametys.runtime.model.ViewItemContainer;

/**
 * Create a dynamic model with the given columns and facets and return the model columns if valid.
 */
public class GetSolrSearchModelAction extends SearchAction
{
    /** The content types helper */
    protected ContentTypesHelper _contentTypesHelper;
    
    /** The search model helper. */
    protected SearchModelHelper _searchModelHelper;
    
    /** The helper for columns */
    protected ColumnHelper _columnHelper;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        
        _contentTypesHelper = (ContentTypesHelper) serviceManager.lookup(ContentTypesHelper.ROLE);
        _searchModelHelper = (SearchModelHelper) serviceManager.lookup(SearchModelHelper.ROLE);
        _columnHelper = (ColumnHelper) serviceManager.lookup(ColumnHelper.ROLE);
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        Map<String, Object> results = new HashMap<>();
        request.setAttribute(JSonReader.OBJECT_TO_READ, results);
        
        Map<String, Object> jsParameters = _serverCommHelper.getJsParameters();
        
        String modelId = (String) jsParameters.get("model");
        SearchUIModel model = _searchModelManager.getExtension(modelId);
        
        Map<String, Object> values = (Map<String, Object>) jsParameters.get("values");
        
        String columnsStr = StringUtils.defaultString((String) values.get("columns"));
        
        Object facetObj = values.get("facets");
        List<String> facets = null;
        if (facetObj != null && facetObj instanceof List<?>)
        {
            facets = (List<String>) facetObj;
        }
        
        Object cTypesObj = values.get("contentTypes");
        Set<String> commonContentTypes = null;
        if (cTypesObj != null && cTypesObj instanceof List<?>)
        {
            Set<String> contentTypes = new HashSet<>((List<String>) cTypesObj);
            commonContentTypes = _contentTypesHelper.getCommonAncestors(contentTypes);
        }
        
        CriteriaSearchModelWrapper modelWrapper = new CriteriaSearchModelWrapper(model, manager, _context, new AvalonLoggerAdapter(getLogger()));
        ContainerUtil.service(modelWrapper, manager);
        
        Map<String, Object> contextualParameters = (Map<String, Object>) jsParameters.get("contextualParameters");
        
        boolean success = true;
        try
        {
            modelWrapper.setFacetedCriteria(commonContentTypes, facets, contextualParameters);
        }
        catch (Exception e)
        {
            getLogger().warn("Error setting facets for solr query search.", e);
            
            success = false;
            results.put("error", "facet-error");
            results.put("message", e.getMessage());
        }
        
        try
        {
            modelWrapper.setResultColumns(commonContentTypes, _getColumns(columnsStr, commonContentTypes), contextualParameters);
        }
        catch (Exception e)
        {
            getLogger().warn("Error setting columns for solr query search.", e);
            
            success = false;
            results.put("error", "column-error");
            results.put("message", e.getMessage());
        }
        
        results.put("success", success);
        
        if (success)
        {
            ViewItemContainer resultItems = modelWrapper.getResultItems(contextualParameters);
            results.put("columns", _searchModelHelper.getColumnsInfo(resultItems));
        }
        
        return EMPTY_MAP;
    }
    
    private Collection<Column> _getColumns(String columnsStr, Set<String> baseContentTypes)
    {
        return _columnHelper.getColumns(columnsStr, baseContentTypes);
    }
    
}
