/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr;

import org.apache.solr.client.solrj.SolrClient;

/**
 * Component acting as a single entry point to get access to Solr clients.
 */
public interface SolrClientProvider
{
    
    /** The component role. */
    public static final String ROLE = SolrClientProvider.class.getName();
    
    /**
     * Get the "read" mode solr client.
     * @return the "read" mode solr client.
     */
    SolrClient getReadClient();
    
    /**
     * Get a (default) Solr client suited to update.
     * @param workspaceName The name of the workspace
     * @return a Solr client suited to update.
     */
    default SolrClient getUpdateClient(String workspaceName)
    {
        return getUpdateClient(workspaceName, true);
    }
    
    /**
     * Get a Solr client suited to update.
     * <br>You must call this method with <code>false</code> value for argument `autocommit` so as to get a client that will not trigger any autocommit on Solr side.
     * <br>Otherwise, calling this method with <code>true</code> value for argument `autocommit` is equivalent to calling {@link #getUpdateClient(String)}
     * @param workspaceName The name of the workspace
     * @param autoCommit <code>false</code> so as to get a client that will not trigger any autocommit on Solr side.
     * @return a Solr client suited to update.
     */
    SolrClient getUpdateClient(String workspaceName, boolean autoCommit);
    
    /**
     * Get the name of the collection to use.
     * @return The name of the collection to search in.
     */
    String getCollectionName();
    
    /**
     * Get the collection to use.
     * @param workspaceName The workspace name.
     * @return The name of the collection to search in.
     */
    String getCollectionName(String workspaceName);
}
