/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ametys.cms.search.QueryBuilder;
import org.ametys.cms.search.query.AndQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.QuerySyntaxException;
import org.ametys.cms.search.query.WorkflowStepQuery;

import com.google.common.primitives.Ints;

/**
 * Helper to build Solr content queries.
 */
public final class SolrContentQueryHelper
{
    private SolrContentQueryHelper()
    {
        // Nothing
    }
    
    /**
     * Gets the content types from JS parameters when using 'search-ui.solr' model
     * @param jsParameters The JS parameters
     * @return The content types
     */
    @SuppressWarnings("unchecked")
    public static Set<String> getContentTypes(Map<String, Object> jsParameters)
    {
        Map<String, Object> values = (Map<String, Object>) jsParameters.get("values");
        Object cTypesObj = values.get("contentTypes");
        if (cTypesObj != null && cTypesObj instanceof List<?>)
        {
            return new HashSet<>((List<String>) cTypesObj);
        }
        return Set.of();
    }
    
    /**
     * Build a Solr query string from inputs coming from a SolrQuerySearch (query, content types, workflow steps)
     * @param queryBuilder The {@link QueryBuilder} component
     * @param baseQuery The main query
     * @param contentTypesOrMixins The content types -or mixins- (can be empty)
     * @param wfSteps The workflow steps (can be empty)
     * @return The built query
     * @throws QuerySyntaxException If a query syntax is invalid
     */
    public static String buildQuery(QueryBuilder queryBuilder, String baseQuery, Set<String> contentTypesOrMixins, Set<Integer> wfSteps) throws QuerySyntaxException
    {
        // Base query
        List<Query> queriesInFinalQuery = new ArrayList<>();
        queriesInFinalQuery.add(() -> baseQuery);
        
        // Content types
        if (!contentTypesOrMixins.isEmpty())
        {
            Query cTypeQuery = queryBuilder.createContentTypeOrMixinQuery(null, contentTypesOrMixins, true);
            queriesInFinalQuery.add(0, cTypeQuery);
        }
        
        // Workflow steps
        if (!wfSteps.isEmpty())
        {
            WorkflowStepQuery wfStepQuery = new WorkflowStepQuery(Ints.toArray(wfSteps));
            queriesInFinalQuery.add(0, wfStepQuery);
        }
        
        Query query = new AndQuery(queriesInFinalQuery);
        return query.build();
    }
}
