/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Path;
import java.time.Duration;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.FormatStyle;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;
import java.util.Random;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.components.source.impl.SitemapSource;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.time.DurationFormatUtils;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.source.SourceUtil;
import org.quartz.JobDataMap;
import org.quartz.JobExecutionContext;

import org.ametys.cms.scripts.ReportLocationAction;
import org.ametys.core.schedule.Schedulable;
import org.ametys.core.schedule.progression.ContainerProgressionTracker;
import org.ametys.core.ui.mail.StandardMailBodyHelper;
import org.ametys.core.util.I18nUtils;
import org.ametys.core.util.URIUtils;
import org.ametys.core.util.mail.SendMailHelper;
import org.ametys.core.util.mail.SendMailHelper.MailBuilder;
import org.ametys.plugins.core.impl.schedule.AbstractStaticSchedulable;
import org.ametys.plugins.core.schedule.Scheduler;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.i18n.I18nizableTextParameter;

/**
 * A {@link Schedulable} job for executing scripts.
 */
public class SolrExportSchedulable extends AbstractStaticSchedulable
{
    /** The key for the export type */
    public static final String TYPE_KEY = "type";
    /** The key for the recipient of the report mail */
    public static final String RECIPIENT_KEY = "recipient";
    /** The key for the search parameters */
    public static final String SEARCHPARAMS_KEY = "searchParams";
    /** The key for the language */
    public static final String SEARCHPARAMS_LANGUAGE = "lang";
    /** The key for the export URL */
    public static final String EXPORT_URL = "exportUrl";
    
    private static final String __JOBDATAMAP_TYPE_KEY = Scheduler.PARAM_VALUES_PREFIX + TYPE_KEY;
    private static final String __JOBDATAMAP_RECIPIENT_KEY = Scheduler.PARAM_VALUES_PREFIX + RECIPIENT_KEY;
    private static final String __JOBDATAMAP_SEARCHPARAMS_KEY = Scheduler.PARAM_VALUES_PREFIX + SEARCHPARAMS_KEY;
    private static final String __JOBDATAMAP_LANGUAGE_KEY = Scheduler.PARAM_VALUES_PREFIX + SEARCHPARAMS_LANGUAGE;
    private static final String __JOBDATAMAP_EXPORT_URL_KEY = Scheduler.PARAM_VALUES_PREFIX + EXPORT_URL;

    /** The avalon source resolver. */
    protected SourceResolver _sourceResolver;
    /** I18n Utils */
    protected I18nUtils _i18nUtils;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
    }
    
    @Override
    public void execute(JobExecutionContext context, ContainerProgressionTracker progressionTracker) throws Exception
    {
        ZonedDateTime startDate = ZonedDateTime.now();
        Path reportDirectory = ReportLocationAction.getScriptReportDirectory();
        File folder = reportDirectory.toFile();
        
        JobDataMap jobDataMap = context.getJobDetail().getJobDataMap();
        String jsonString = jobDataMap.getString(__JOBDATAMAP_SEARCHPARAMS_KEY);
        String type = jobDataMap.getString(__JOBDATAMAP_TYPE_KEY);
        String recipient = jobDataMap.getString(__JOBDATAMAP_RECIPIENT_KEY);
        String lang = jobDataMap.getString(__JOBDATAMAP_LANGUAGE_KEY);
        String exportUrl = jobDataMap.getString(__JOBDATAMAP_EXPORT_URL_KEY);
        String sender = Config.getInstance().getValue("smtp.mail.from");
        
        String extension = _getExtension(type);
        String uri = _getUri(exportUrl);
        
        String subject;
        String html;
        String text;
        
        Locale locale = StringUtils.isNotBlank(lang) ? Locale.forLanguageTag(lang) : Locale.getDefault();
        I18nizableText undefinedText = new I18nizableText("plugin.cms", "UITOOL_SEARCH_ASYNC_EXPORT_UNDEFINED_TEXT");
        I18nizableText start = Optional.ofNullable(startDate)
                .map(DateTimeFormatter
                        .ofLocalizedDateTime(FormatStyle.LONG)
                        .withLocale(locale)
                        .withZone(ZoneId.systemDefault())
                        ::format
                        )
                .map(I18nizableText::new)
                .orElse(undefinedText);
        
        try
        {
            String generatedFile = _generateFile(folder, uri, jsonString, extension);
            
            ZonedDateTime endDate = ZonedDateTime.now();
    
            I18nizableText duration = Optional.ofNullable(endDate)
                    .map(end -> Duration.between(startDate, endDate))
                    .map(Duration::toMillis)
                    .map(DurationFormatUtils::formatDurationHMS)
                    .map(I18nizableText::new)
                    .orElse(undefinedText);
                
            
            subject = _getSubject(type, lang);
            html = _getHtmlBody(type, generatedFile, extension, start, duration, lang);
            text = _getTextBody(type, generatedFile, extension, start, duration, lang);
        }
        catch (Exception e)
        {
            subject = _getErrorSubject(type, lang);
            html = _getErrorHtmlBody(type, start, lang);
            text = _getErrorTextBody(type, start, lang);
        }
        
        MailBuilder mailBuilder = SendMailHelper.newMail()
                      .withSubject(subject)
                      .withTextBody(text)
                      .withSender(sender)
                      .withRecipient(recipient);
        
        if (html != null)
        {
            mailBuilder.withHTMLBody(html);
        }
        
        mailBuilder.sendMail();
    }
    
    private String _getSubject(String type, String lang)
    {
        I18nizableText i18nSubject = new I18nizableText("plugin.cms", "UITOOL_SEARCH_ASYNC_EXPORT_" + type.toUpperCase() + "_MAIL_SUBJECT");
        return _i18nUtils.translate(i18nSubject, lang);
    }
    
    private String _getURL(String fileName, String extension)
    {
        String baseURL = Config.getInstance().getValue("cms.url");
        return baseURL + "/plugins/cms/async-export/" + fileName + "/export." + extension;
    }
    
    private String _getHtmlBody(String type, String fileName, String extension, I18nizableText start, I18nizableText duration, String lang)
    {
        String url = _getURL(fileName, extension);
        
        Map<String, I18nizableTextParameter> params = Map.of("url", new I18nizableText(url),
                "filename", new I18nizableText("export." + extension),
                "start", start,
                "duration", duration);
        
        try
        {
            return StandardMailBodyHelper.newHTMLBody()
                    .withLanguage(lang)
                    .withTitle(_getSubject(type, lang))
                    .withMessage(new I18nizableText("plugin.cms", "UITOOL_SEARCH_ASYNC_EXPORT_" + type.toUpperCase() + "_MAIL_HTMLBODY", params))
                    .build();
        }
        catch (IOException e)
        {
            return null;
        }
    }
    
    private String _getTextBody(String type, String fileName, String extension, I18nizableText start, I18nizableText duration, String lang)
    {
        String url = _getURL(fileName, extension);
        
        Map<String, I18nizableTextParameter> params = Map.of("url", new I18nizableText(url),
                "filename", new I18nizableText("export." + extension),
                "start", start,
                "duration", duration);
        
        I18nizableText i18nBody = new I18nizableText("plugin.cms", "UITOOL_SEARCH_ASYNC_EXPORT_" + type.toUpperCase() + "_MAIL_TEXTBODY", params);
        return _i18nUtils.translate(i18nBody, lang);
    }

    private String _getErrorSubject(String type, String lang)
    {
        I18nizableText i18nSubject = new I18nizableText("plugin.cms", "UITOOL_SEARCH_ASYNC_EXPORT_" + type.toUpperCase() + "_MAIL_SUBJECT_ERROR");
        return _i18nUtils.translate(i18nSubject, lang);
    }

    private String _getErrorHtmlBody(String type, I18nizableText start, String lang)
    {
        Map<String, I18nizableTextParameter> params = Map.of("start", start);
        
        try
        {
            return StandardMailBodyHelper.newHTMLBody()
                    .withLanguage(lang)
                    .withTitle(_getErrorSubject(type, lang))
                    .withMessage(new I18nizableText("plugin.cms", "UITOOL_SEARCH_ASYNC_EXPORT_" + type.toUpperCase() + "_MAIL_HTMLBODY_ERROR", params))
                    .build();
        }
        catch (IOException e)
        {
            return null;
        }
    }

    private String _getErrorTextBody(String type, I18nizableText start, String lang)
    {
        Map<String, I18nizableTextParameter> params = Map.of("start", start);
        
        I18nizableText i18nBody = new I18nizableText("plugin.cms", "UITOOL_SEARCH_ASYNC_EXPORT_" + type.toUpperCase() + "_MAIL_TEXTBODY_ERROR", params);
        return _i18nUtils.translate(i18nBody, lang);
    }
    
    private String _getExtension(String type)
    {
        return type;
    }
    
    private String _getUri(String exportUrl)
    {
        return "cocoon://_plugins/" + exportUrl;
    }
    
    /**
     * Generate a file from the uri
     * @param sourceFolder the directory where the file are created
     * @param uri the uri
     * @param parameters the parameters of the uri
     * @param extension the extension of the file
     * @return output file name (name.extension)
     * @throws IOException if an error occured with files
     */
    protected String _generateFile(File sourceFolder, String uri, String parameters, String extension) throws IOException
    {
        SitemapSource source = null;
        
        Random rand = new Random();
        String fileName = String.valueOf(Math.abs(rand.nextInt()));
        File outputFile = new File(sourceFolder, fileName + "." + extension);
        while (outputFile.exists())
        {
            fileName = String.valueOf(Math.abs(rand.nextInt()));
            outputFile = new File(sourceFolder, fileName + "." + extension);
        }
        
        try
        {
            String fullUri = URIUtils.buildURI(uri, Map.of("parameters", parameters));
            // Resolve the export to the appropriate pdf url.
            source = (SitemapSource) _sourceResolver.resolveURI(fullUri, null, null);
            
            
            try (OutputStream pdfTmpOs = new FileOutputStream(outputFile); InputStream sourceIs = source.getInputStream())
            {
                SourceUtil.copy(sourceIs, pdfTmpOs);
            }
        }
        finally
        {
            if (source != null)
            {
                _sourceResolver.release(source);
            }
        }
        return fileName;
    }
}
