/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr;

import java.util.Collection;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.data.type.indexing.IndexableElementType;
import org.ametys.cms.model.ContentElementDefinition;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.cms.search.ui.model.SearchUIModelExtensionPoint;
import org.ametys.core.ui.Callable;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemContainer;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ModelItemType;

/**
 * Helper for solr query editor.
 */
public class SolrQueryHelper extends AbstractLogEnabled implements Component, Serviceable
{
    
    /** The component role. */
    public static final String ROLE = SolrQueryHelper.class.getName();
    
    /** The content type extension point. */
    protected ContentTypeExtensionPoint _cTypeEP;
    
    /** The content types helper. */
    protected ContentTypesHelper _contentTypesHelper;
    
    /** The search model helper. */
    protected SearchUIModelExtensionPoint _searchModelManager;
    
    /** The extension point for system properties */
    protected SystemPropertyExtensionPoint _systemPropertyEP;   
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        _cTypeEP = (ContentTypeExtensionPoint) serviceManager.lookup(ContentTypeExtensionPoint.ROLE);
        _contentTypesHelper = (ContentTypesHelper) serviceManager.lookup(ContentTypesHelper.ROLE);
        _searchModelManager = (SearchUIModelExtensionPoint) serviceManager.lookup(SearchUIModelExtensionPoint.ROLE);
        _systemPropertyEP = (SystemPropertyExtensionPoint) serviceManager.lookup(SystemPropertyExtensionPoint.ROLE);
    }
    
    /**
     * Get all the content types and their indexing fields.
     * @param language the query language (because string field names are language-dependent).
     * @return the content types and their metadatas.
     */
    @Callable
    public Map<String, Object> getAllContentTypeIndexingFields(String language)
    {
        Map<String, Object> results = new HashMap<>();
        Map<String, Object> contentTypes = new HashMap<>();
        results.put("contentTypes", contentTypes);
        
        for (String cTypeId : _cTypeEP.getExtensionsIds())
        {
            ContentType contentType = _cTypeEP.getExtension(cTypeId);
            Map<String, Object> properties = _getContentTypeProperties(contentType, language);
            contentTypes.put(cTypeId, properties);
        }
        
        return results;
    }
    
    private Map<String, Object> _getContentTypeProperties(ContentType contentType, String language)
    {
        Map<String, Object> properties = new HashMap<>();
        properties.put("type", "contentType");
        properties.put("superTypes", contentType.getSupertypeIds());
        
        Map<String, Object> fields = new HashMap<>();
        properties.put("fields", fields);
        
        // Model items and properties
        for (ModelItem modelItem : contentType.getModelItems())
        {
            Map<String, Object> modelItemProperties = _getModelItemProperties(modelItem, language);
            fields.put(modelItem.getName(), modelItemProperties);
        }
        
        return properties;
    }
    
    private Map<String, Object> _getModelItemProperties(ModelItem modelItem, String language)
    {
        Map<String, Object> properties = new HashMap<>();
        
        String name = modelItem.getName();
        ModelItemType type = modelItem.getType();
        properties.put("type", type.getId());
        
        if (modelItem instanceof ElementDefinition elementDefinition)
        {
            if (type instanceof IndexableElementType elementType)
            {
                DataContext context = DataContext.newInstance()
                                                 .withLocale(new Locale(language))
                                                 .withModelItem(modelItem);
                
                properties.put("fl", name + elementType.getIndexingFieldSuffix(context));
                elementType.getTextFieldSuffix(context)
                           .ifPresent(suffix -> properties.put("ftFl", name + suffix));
                elementType.getWildcardFieldSuffix(context)
                           .ifPresent(suffix -> properties.put("wcFl", name + suffix));
                
                if (elementDefinition instanceof ContentElementDefinition contentElementDefinition)
                {
                    properties.put("cType", contentElementDefinition.getContentTypeId());
                    properties.put("joinFl", name);
                }
                
                properties.put("isDisplayable", true);
                properties.put("isFacetable", elementType.isFacetable(context));
                properties.put("isMultiple", elementDefinition.isMultiple());
            }
        }
        else if (modelItem instanceof ModelItemContainer modelItemContainer)
        {
            properties.put("fl", name);
            
            if (modelItem instanceof RepeaterDefinition)
            {
                properties.put("joinFl", name);
                properties.put("isDisplayable", true);
            }
            
            Map<String, Object> childrenProperties = new HashMap<>();
            properties.put("fields", childrenProperties);
            for (ModelItem child : modelItemContainer.getModelItems())
            {
                Map<String, Object> childProperties = _getModelItemProperties(child, language);
                childrenProperties.put(child.getName(), childProperties);
            }
        }
        
        return properties;
    }
    
    /**
     * Get content types' common ancestors.
     * @param contentTypes The content types.
     * @return a Map with the common ancestors
     */
    @Callable
    public Map<String, Object> getCommonAncestors(Collection<String> contentTypes)
    {
        Set<String> commonAncestors = _contentTypesHelper.getCommonAncestors(contentTypes);
        return Map.of("commonAncestors", commonAncestors);
    }
    
    /**
     * Get the common fields.
     * @param language the query language (because string field names are language-dependent).
     * @return the common fields.
     */
    @Callable
    public Map<String, Object> getCommonFields(String language)
    {
        Map<String, Object> results = new HashMap<>();
        
        Map<String, Object> systemProps = new HashMap<>();
        results.put("fields", systemProps);
        
        Map<String, Object> titleProperties = _getTitleProperties(language);
        systemProps.put(Content.ATTRIBUTE_TITLE, titleProperties);
        
        Set<String> extensionsIds = _systemPropertyEP.getExtensionsIds();
        for (String extensionId : extensionsIds)
        {
            SystemProperty property = _systemPropertyEP.getExtension(extensionId); 
            SearchField searchField = property.getSearchField();
            if (searchField != null)
            {
                Map<String, Object> properties = new HashMap<>();
                systemProps.put(extensionId, properties);
                
                // type
                properties.put("type", property.getType().getId());
                
                // solr field
                properties.put("fl", searchField.getName());
                
                // is displayable
                properties.put("isDisplayable", _systemPropertyEP.isDisplayable(extensionId));
                
                // is facetable
                properties.put("isFacetable", property.isFacetable());
                
                // is multiple
                properties.put("isMultiple", property.isMultiple());
            }
        }
        
        return results;
    }
    
    private Map<String, Object> _getTitleProperties(String language)
    {
        ModelItem titleDefinition = _contentTypesHelper.getTitleAttributeDefinition();
        return _getModelItemProperties(titleDefinition, language);
    }
}
