/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.container.ContainerUtil;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.environment.Request;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.SearchResults;
import org.ametys.cms.search.Sort;
import org.ametys.cms.search.cocoon.SearchAction;
import org.ametys.cms.search.content.ContentSearcherFactory.SimpleContentSearcher;
import org.ametys.cms.search.model.SearchCriterion;
import org.ametys.cms.search.model.SearchModel;
import org.ametys.cms.search.query.QuerySyntaxException;
import org.ametys.cms.search.ui.model.ColumnHelper;
import org.ametys.cms.search.ui.model.ColumnHelper.Column;
import org.ametys.core.util.AvalonLoggerAdapter;

/**
 * Execute a solr query with custom columns and facets.
 */
public class SolrQuerySearchAction extends SearchAction
{
    /** The content types helper */
    protected ContentTypesHelper _contentTypesHelper;
    
    /** The helper for columns */
    protected ColumnHelper _columnHelper;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        
        _contentTypesHelper = (ContentTypesHelper) serviceManager.lookup(ContentTypesHelper.ROLE);
        _columnHelper = (ColumnHelper) serviceManager.lookup(ColumnHelper.ROLE);
    }
    
    @Override
    protected void doSearch(Request request, SearchModel model, int offset, int maxResults, Map<String, Object> jsParameters, Map<String, Object> contextualParameters) throws Exception
    {
        SearchValues searchValues = getSearchValues(jsParameters);
        
        CriteriaSearchModelWrapper modelWrapper = new CriteriaSearchModelWrapper(model, manager, _context, new AvalonLoggerAdapter(getLogger()));
        ContainerUtil.service(modelWrapper, manager);

        modelWrapper.setResultColumns(searchValues.getBaseContentTypes(), searchValues.getColumns(), contextualParameters);
        
        Map<String, ? extends SearchCriterion> criteria = model.getCriteria(contextualParameters);
        String lang = _queryBuilder.getCriteriaLanguage(criteria, null, searchValues.getValues(), contextualParameters);
        if (StringUtils.isNotEmpty(lang))
        {
            request.setAttribute(SEARCH_LOCALE, new Locale(lang));
        }
        
        Collection<String> transformedFacets = modelWrapper.setFacetedCriteria(searchValues.getBaseContentTypes(), searchValues.getFacets(), contextualParameters);
        List<Sort> sorts = getSort(searchValues.getSortInfo(), searchValues.getGroupInfo());
        
        SearchResults<Content> results = getResults(searchValues, offset, maxResults, transformedFacets, sorts);
        
        request.setAttribute(SEARCH_RESULTS, results);
        request.setAttribute(SEARCH_MODEL, modelWrapper);
    }
    
    /**
     * Get the object representing search values from the JS parameters.
     * @param jsParameters The JS parameters
     * @return an object representing the search values
     */
    protected SearchValues getSearchValues(Map<String, Object> jsParameters)
    {
        return new SearchValues(jsParameters);
    }
    
    /**
     * Get the query string from the search values.
     * @param searchValues The search values
     * @return the query string
     * @throws QuerySyntaxException if an error occurs
     */
    protected String getQueryString(SearchValues searchValues) throws QuerySyntaxException
    {
        return SolrContentQueryHelper.buildQuery(_queryBuilder, searchValues.getBaseQuery(), Collections.EMPTY_SET/*content types or mixin query will be handled by the SimpleContentSearcher*/, searchValues.getWorkflowSteps());
    }
    
    /**
     * Create the searcher and execute it from the search values.
     * @param searchValues The search values
     * @param offset The offset
     * @param maxResults The max number of results
     * @param facets Facets renamed for Solr
     * @param sorts The sorts
     * @return the search results
     * @throws Exception if an error occurs
     */
    protected SearchResults<Content> getResults(SearchValues searchValues, int offset, int maxResults, Collection<String> facets, List<Sort> sorts) throws Exception
    {
        return getContentSearcher(searchValues, offset, maxResults, facets, sorts)
                .searchWithFacets(getQueryString(searchValues), searchValues.getFacetValues());
    }
    
    /**
     * Get the content search from the search values
     * @param searchValues The search values
     * @param offset The offset
     * @param maxResults The max number of results
     * @param facets Facets renamed for Solr
     * @param sorts The sorts
     * @return the content searcher
     */
    protected SimpleContentSearcher getContentSearcher(SearchValues searchValues, int offset, int maxResults, Collection<String> facets, List<Sort> sorts)
    {
        return _searcherFactory.create(searchValues.getContentTypes())
                .withSort(sorts)
                .withFacets(facets)
                .withLimits(offset, maxResults);
    }
    
    /**
     * Object representing search values.
     */
    @SuppressWarnings("unchecked")
    protected class SearchValues
    {
        /** The JS parameters */
        protected Map<String, Object> _jsParameters;
        /** The values from JS parameters */
        protected Map<String, Object> _values;
        /** The base query */
        protected String _baseQuery;
        /** The content types */
        protected Set<String> _contentTypeIds;
        /** The base content types (common content types) */
        protected Set<String> _baseContentTypeIds;
        /** The facet fields */
        protected Collection<String> _facets;
        /** The facet values */
        protected Map<String, List<String>> _facetValues;
        /** The columns */
        protected Collection<Column> _columns;
        /** The sorts */
        protected String _sortInfo;
        /** The groups */
        protected String _groupInfo;
        /** The workflow steps */
        protected Set<Integer> _wfSteps;
        
        /**
         * Constructor to build the object from JS parameters.
         * @param jsParameters The JS parameters
         */
        protected SearchValues(Map<String, Object> jsParameters)
        {
            _jsParameters = jsParameters;
            _parseValues();
            _parseContentTypes();
            _parseQuery();
            _parseFacets();
            _parseFacetValues();
            _parseColumns();
            _parseSortInfo();
            _parseGroupInfo();
            _parseWorkflowSteps();
        }
        
        private void _parseValues()
        {
            _values = (Map<String, Object>) _jsParameters.get("values");
        }
        
        private void _parseContentTypes()
        {
            _contentTypeIds = SolrContentQueryHelper.getContentTypes(_jsParameters);
            _baseContentTypeIds = _contentTypesHelper.getCommonAncestors(_contentTypeIds);
        }
        
        private void _parseQuery()
        {
            _baseQuery = (String) _values.get("query");
        }
        
        private void _parseFacets()
        {
            String facetObj = StringUtils.defaultString((String) _values.get("facets"));
            _facets = Arrays.asList(StringUtils.split(facetObj, ", ")).stream().map(s -> s.replaceAll("\\.", "/")).collect(Collectors.toList());
        }
        
        private void _parseFacetValues()
        {
            _facetValues = (Map<String, List<String>>) _jsParameters.get("facetValues");
            if (_facetValues == null)
            {
                _facetValues = Collections.emptyMap();
            }
        }
        
        private void _parseColumns()
        {
            Object columnsObject = _values.get("columns");
            if (columnsObject == null)
            {
                // Empty list, but not immutable
                _columns = new ArrayList();
            }
            else if (columnsObject instanceof String)
            {
                _columns = _columnHelper.getColumns((String) columnsObject, _baseContentTypeIds);
            }
            else if (columnsObject instanceof List)
            {
                _columns = _columnHelper.getColumns((List) columnsObject, _baseContentTypeIds);
            }
        }
        
        private void _parseSortInfo()
        {
            _sortInfo = (String) _jsParameters.get("sort");
        }
        
        private void _parseGroupInfo()
        {
            _groupInfo = (String) _jsParameters.get("group");
        }
        
        private void _parseWorkflowSteps()
        {
            Object wfStepsObj = _values.get("workflowSteps");
            _wfSteps = new HashSet<>();
            if (wfStepsObj != null && wfStepsObj instanceof List<?>)
            {
                for (String wfStepObj : (List<String>) wfStepsObj)
                {
                    if (StringUtils.isNotEmpty(wfStepObj))
                    {
                        _wfSteps.add(Integer.parseInt(wfStepObj));
                    }
                }
            }
        }
        
        /**
         * Get the base query.
         * @return the base query
         */
        protected String getBaseQuery()
        {
            return _baseQuery;
        }
        
        /**
         * Get the columns.
         * @return the columns
         */
        protected Collection<Column> getColumns()
        {
            return _columns;
        }
        
        /**
         * Get the base content types (extract from content types, it's the common ancestors).
         * @return the base content types
         */
        protected Set<String> getBaseContentTypes()
        {
            return _baseContentTypeIds;
        }
        
        /**
         * Get the content types.
         * @return the content types
         */
        protected Set<String> getContentTypes()
        {
            return _contentTypeIds;
        }
        
        /**
         * Get the workflow steps.
         * @return the workflow steps
         */
        protected Set<Integer> getWorkflowSteps()
        {
            return _wfSteps;
        }
        
        /**
         * Get the sort info.
         * @return the sort info
         */
        protected String getSortInfo()
        {
            return _sortInfo;
        }
        
        /**
         * Get the group info.
         * @return the group info
         */
        protected String getGroupInfo()
        {
            return _groupInfo;
        }

        /**
         * Get the facet fields.
         * @return the facet fields
         */
        protected Collection<String> getFacets()
        {
            return _facets;
        }
        
        /**
         * Get the facet values.
         * @return the facet values
         */
        protected Map<String, List<String>> getFacetValues()
        {
            return _facetValues;
        }
        
        /**
         * Get the values from the JS parameters.
         * @return the values
         */
        protected Map<String, Object> getValues()
        {
            return _values;
        }
    }
}
