/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr;

import java.util.Collections;
import java.util.Map;
import java.util.Optional;

import org.apache.solr.client.solrj.response.QueryResponse;
import org.apache.solr.common.SolrDocumentList;
import org.slf4j.LoggerFactory;

import org.ametys.cms.search.SearchResult;
import org.ametys.cms.search.SearchResults;
import org.ametys.cms.search.SearchResultsIterable;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectIterator;
import org.ametys.plugins.repository.AmetysObjectResolver;

/**
 * Solr implementation of a {@link SearchResults}.
 * @param <A> the actual type of {@link AmetysObject}s. 
 */
public class SolrSearchResults<A extends AmetysObject> implements SearchResults<A>
{
    
    /** The solr query response */
    protected QueryResponse _response;
    
    /** The Solr documents of the search response */
    protected SolrDocumentList _docList;
    
    /** The AmetysObject resolver. */
    protected AmetysObjectResolver _resolver;
    
    /** The facet results. */
    protected Map<String, Map<String, Integer>> _facetResults;
    
    /**
     * Build a SolrSearchResults object.
     * @param response the solr search response.
     * @param resolver The Ametys object resolver.
     * @param facetResults The facet results.
     */
    public SolrSearchResults(QueryResponse response, AmetysObjectResolver resolver, Map<String, Map<String, Integer>> facetResults)
    {
        _response = response;
        _docList = response.getResults();
        _resolver = resolver;
        _facetResults = facetResults;
    }
    
    @Override
    public SearchResultsIterable<SearchResult<A>> getResults()
    {
        return new SolrResponseIterable<>(_docList, _resolver);
    }
    
    @Override
    public AmetysObjectIterable<A> getObjects()
    {
        return new AmetysObjectIterable<>()
        {
            @Override
            public long getSize()
            {
                return _docList.size();
            }

            @Override
            public AmetysObjectIterator<A> iterator()
            {
                return new SolrResponseAmetysObjectIterator<>(_docList.iterator(), _docList.size(), _resolver, LoggerFactory.getLogger(SolrResponseAmetysObjectIterator.class));
            }

            @Override
            public void close()
            {
                // Nothing to do.
            }
        };
    }
    
    @Override
    public Iterable<String> getObjectIds()
    {
        return new SolrResponseIdIterable(_docList);
    }
    
    @Override
    public Map<String, Map<String, Integer>> getFacetResults()
    {
        return Collections.unmodifiableMap(_facetResults);
    }
    
    @Override
    public long getTotalCount()
    {
        return _docList.getNumFound();
    }
    
    @Override
    public float getMaxScore()
    {
        return _docList.getMaxScore();
    }
    
    @Override
    public Optional<Map<String, Object>> getDebugMap()
    {
        return Optional.ofNullable(_response)
                .map(QueryResponse::getDebugMap);
    }
}
