/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr.field;

import java.util.Collections;
import java.util.List;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.search.SearchField;

/**
 * Metadata (abstract) generic search field.
 */
public abstract class AbstractModelItemSearchField implements SearchField
{
    /** The join paths */
    protected List<String> _joinPaths;
    /** The field path. */
    protected String _path;
    
    /**
     * Build a Metadata SearchField
     * @param path The field path
     */
    protected AbstractModelItemSearchField(String path)
    {
        this(null, path);
    }
    
    /**
     * Build a Metadata SearchField
     * @param joinPaths The join field paths
     * @param finalPath The final field path
     */
    protected AbstractModelItemSearchField(List<String> joinPaths, String finalPath)
    {
        _joinPaths = joinPaths;
        _path = finalPath;
    }
    
    @Override
    public String getSortField()
    {
        String sortFieldSuffix = _getSortFieldSuffix();
        if (sortFieldSuffix == null)
        {
            return null; // not sortable
        }
        
        if (isJoined())
        {
            String join = StringUtils.join(_joinPaths, "->");
            return "ametys(" + join + "," + _path + sortFieldSuffix + ")";
        }
        return _path + sortFieldSuffix;
    }
    
    /**
     * Get the sort field suffix corresponding to this field. Return null if not sortable.
     * @return the sort field suffix. Null if not sortable.
     */
    protected abstract String _getSortFieldSuffix();
    
    @Override
    public boolean isJoined()
    {
        return _joinPaths != null && _joinPaths.size() > 0;
    }
    
    @Override
    public List<String> getJoinedPaths()
    {
        return Optional.ofNullable(_joinPaths).orElse(Collections.emptyList());
    }
    
    @Override
    public String getFacetField()
    {
        return _path + _getFacetFieldSuffix();
    }
    
    /**
     * Get the facet field suffix corresponding to this field. Return null if not facetable.
     * @return the facet field suffix. Null if not facetable.
     */
    protected abstract String _getFacetFieldSuffix();
    
    @Override
    public String getName()
    {
        if (isJoined())
        {
            return StringUtils.join(_joinPaths, '/') + '/' + _path;
        }
        return _path;
    }
}
