/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr.field;

import java.util.Collections;
import java.util.List;
import java.util.Optional;

import org.apache.avalon.framework.context.Context;
import org.apache.commons.collections4.ListUtils;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.content.indexing.solr.SolrFieldNames;

/**
 * Content generic search field.
 */
public class ContentSearchField extends StringSearchField
{
    private boolean _isMultilingual;
    private Optional<Context> _context;
    
    /**
     * Build a content search field.
     * @param path The field path.
     */
    public ContentSearchField(String path)
    {
        this(path, false, Optional.empty());
    }
    /**
     * Build a content search field.
     * @param joinPaths The join field paths
     * @param finalPath The final field path.
     */
    public ContentSearchField(List<String> joinPaths, String finalPath)
    {
        this(joinPaths, finalPath, false, Optional.empty());
    }
    
    /**
     * Build a content search field.
     * @param path The field path.
     * @param isMultilingual <code>true</code> if the Content has a multlilingual title
     * @param context The context
     */
    public ContentSearchField(String path, boolean isMultilingual, Optional<Context> context)
    {
        super(path);
        _isMultilingual = isMultilingual;
        _context = context;
    }
    /**
     * Build a content search field.
     * @param joinPaths The join field paths
     * @param finalPath The final field path.
     * @param isMultilingual <code>true</code> if the Content has a multilingual title
     * @param context The context
     */
    public ContentSearchField(List<String> joinPaths, String finalPath, boolean isMultilingual, Optional<Context> context)
    {
        super(joinPaths, finalPath);
        _isMultilingual = isMultilingual;
        _context = context;
    }
    
    @Override
    public List<String> getJoinedPaths()
    {
        // As the join must be done until the id field, '_path' is part of the join
        return _joinPaths != null ? ListUtils.union(_joinPaths, Collections.singletonList(_path)) : Collections.emptyList();
    }
    
    @Override
    protected String _getSortFieldSuffix()
    {
        return _isMultilingual
                ? _getMultilingualSortFieldSuffix()
                : super._getSortFieldSuffix();
    }
    
    private String _getMultilingualSortFieldSuffix()
    {
        String languageSuffix = MultilingualStringSearchField.getCurrentLanguage(_context)
            .map("_"::concat)
            .orElse(StringUtils.EMPTY);
        return super._getSortFieldSuffix() + languageSuffix;
    }
    
    @Override
    public String getFacetField()
    {
        if (isJoined())
        {
            // The join must be done until the id field !
            return SolrFieldNames.ID + "_dv";
        }
        else
        {
            return super.getFacetField();
        }
    }
}
