/*
 *  Copyright 2024 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr.schema;

import java.util.Map;

import org.apache.solr.client.solrj.request.schema.SchemaRequest;

/**
 * Represents a dynamic field definition in a solr schema.
 */
public class DynamicFieldDefinition extends FieldDefinition
{
    /**
     * Build a dynamic field definition, indexed and stored.
     * @param name The field name, can contain a star character to represent a dynamic field (i.e. *_type).
     * @param type The field type.
     * @param multiValued <code>true</code> if the field is multi-valued.
     * @param docValues <code>true</code> if the field is stored as docValues.
     */
    public DynamicFieldDefinition(String name, String type, boolean multiValued, boolean docValues)
    {
        super(name, type, multiValued, docValues);
    }
    
    /**
     * Build a dynamic field definition.
     * @param name The field name, can contain a star character to represent a dynamic field (i.e. *_type).
     * @param type The field type.
     * @param multiValued <code>true</code> if the field is multi-valued.
     * @param docValues <code>true</code> if the field is stored as docValues.
     * @param indexed <code>true</code> if the field is indexed.
     * @param stored <code>true</code> if the field is stored.
     */
    public DynamicFieldDefinition(String name, String type, boolean multiValued, boolean docValues, boolean indexed, boolean stored)
    {
        super(name, type, multiValued, docValues, indexed, stored);
    }
    
    /**
     * Build a dynamic field definition from a map of attributes.
     * @param attributes the Map of attributes.
     */
    public DynamicFieldDefinition(Map<String, Object> attributes)
    {
        super(attributes);
    }
    
    @Override
    public SchemaRequest.Update getSchemaUpdate()
    {
        return new SchemaRequest.AddDynamicField(_getAttributes());
    }

    @Override
    public boolean exists(SchemaFields schemaFields)
    {
        return schemaFields.hasDynamicField(getName());
    }
}
