/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr.schema;

import java.util.HashMap;
import java.util.Map;

import org.apache.solr.client.solrj.request.schema.SchemaRequest;

/**
 * Represents a field definition in a solr schema.
 */
public class FieldDefinition implements SchemaDefinition
{
    /** The field name, can contain a star character to represent a dynamic field (i.e. *_type). */
    protected String _name;
    
    /** The field type. */
    protected String _type;
    
    /** If the field is multi-valued. */
    protected boolean _multiValued;
    
    /** If the field is stored as docValues. */
    protected boolean _docValues;
    
    /** Whether the field is indexed. */
    protected boolean _indexed;
    
    /** Whether the field is stored. */
    protected boolean _stored;
    
    /**
     * Build a field definition, indexed and stored.
     * @param name The field name, can contain a star character to represent a dynamic field (i.e. *_type).
     * @param type The field type.
     * @param multiValued <code>true</code> if the field is multi-valued.
     * @param docValues <code>true</code> if the field is stored as docValues.
     */
    public FieldDefinition(String name, String type, boolean multiValued, boolean docValues)
    {
        this(name, type, multiValued, docValues, true, true);
    }
    
    /**
     * Build a field definition.
     * @param name The field name, can contain a star character to represent a dynamic field (i.e. *_type).
     * @param type The field type.
     * @param multiValued <code>true</code> if the field is multi-valued.
     * @param docValues <code>true</code> if the field is stored as docValues.
     * @param indexed <code>true</code> if the field is indexed.
     * @param stored <code>true</code> if the field is stored.
     */
    public FieldDefinition(String name, String type, boolean multiValued, boolean docValues, boolean indexed, boolean stored)
    {
        this._name = name;
        this._type = type;
        this._multiValued = multiValued;
        this._docValues = docValues;
        this._indexed = indexed;
        this._stored = stored;
    }
    
    /**
     * Build a field definition from a map of attributes.
     * @param attributes the Map of attributes.
     */
    public FieldDefinition(Map<String, Object> attributes)
    {
        this._name = (String) attributes.get("name");
        this._type = (String) attributes.get("type");
        this._multiValued = getBoolean(attributes, "multiValued", false);
        this._docValues = getBoolean(attributes, "docValues", false);
        this._indexed = getBoolean(attributes, "indexed", false);
        this._stored = getBoolean(attributes, "stored", false);
    }
    
    private boolean getBoolean(Map<String, Object> attributes, String name, boolean defaultValue)
    {
        String value = (String) attributes.get(name);
        return value != null ? Boolean.parseBoolean(value) : defaultValue;
    }
    
    /**
     * Get the field name.
     * @return The field name.
     */
    public String getName()
    {
        return _name;
    }
    
    /**
     * Set the field name.
     * @param name the field name.
     */
    public void setName(String name)
    {
        this._name = name;
    }
    
    /**
     * Get the field type.
     * @return The field type.
     */
    public String getType()
    {
        return _type;
    }
    
    /**
     * Set the field type.
     * @param type the field type.
     */
    public void setType(String type)
    {
        this._type = type;
    }
    
    /**
     * If the field is multi-valued.
     * @return true if the field is multi-valued, false otherwise.
     */
    public boolean isMultiValued()
    {
        return _multiValued;
    }
    
    /**
     * Set if the field is multi-valued.
     * @param multiValued <code>true</code> if the field is multi-valued.
     */
    public void setMultiValued(boolean multiValued)
    {
        this._multiValued = multiValued;
    }
    
    /**
     * True to store the field as docValues.
     * @return true to store the field as docValues, false otherwise.
     */
    public boolean isDocValues()
    {
        return _docValues;
    }
    
    /**
     * Set if the field is stored as docValues.
     * @param docValues <code>true</code> if the field is stored as docValues.
     */
    public void setDocValues(boolean docValues)
    {
        this._docValues = docValues;
    }
    
    /**
     * True to store the field as indexed.
     * @return true to store the field as indexed, false otherwise.
     */
    public boolean isIndexed()
    {
        return _indexed;
    }
    
    /**
     * Set if the field is stored as indexed.
     * @param indexed <code>true</code> if the field is stored as indexed.
     */
    public void setIndexed(boolean indexed)
    {
        this._indexed = indexed;
    }
    
    /**
     * True to store the field as stored.
     * @return true to store the field as stored, false otherwise.
     */
    public boolean isStored()
    {
        return _stored;
    }
    
    /**
     * Set if the field is stored as stored.
     * @param stored <code>true</code> if the field is stored as stored.
     */
    public void setStored(boolean stored)
    {
        this._stored = stored;
    }
    
    /**
     * Get the attributes to build the field.
     * @return the attributes as a map
     */
    protected Map<String, Object> _getAttributes()
    {
        // Force indexed and stored attributes to true.
        Map<String, Object> attributes = new HashMap<>();
        attributes.put("name", getName());
        attributes.put("type", getType());
        attributes.put("multiValued", isMultiValued());
        attributes.put("docValues", isDocValues());
        attributes.put("indexed", Boolean.TRUE);
        attributes.put("stored", Boolean.TRUE);
        return attributes;
    }
    
    public SchemaRequest.Update getSchemaUpdate()
    {
        return new SchemaRequest.AddField(_getAttributes());
    }
    
    public boolean exists(SchemaFields schemaFields)
    {
        return schemaFields.hasField(getName());
    }
}
