/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.solr.schema;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.xml.dom.DOMParser;
import org.apache.excalibur.xml.xpath.XPathProcessor;
import org.apache.solr.client.solrj.request.schema.AnalyzerDefinition;
import org.apache.solr.client.solrj.request.schema.FieldTypeDefinition;
import org.apache.solr.client.solrj.response.schema.SchemaRepresentation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Component providing helper methods to work with search schema and fields.
 */
public class SchemaHelper extends AbstractLogEnabled implements Component, Serviceable
{
    
    /** The component role. */
    public static final String ROLE = SchemaHelper.class.getName();
    
    /** The source resolver. */
    protected SourceResolver _sourceResolver;
    
    /** A DOM parser. */
    protected DOMParser _domParser;
    
    /** A XPath processor. */
    protected XPathProcessor _xPathProcessor;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _domParser = (DOMParser) manager.lookup(DOMParser.ROLE);
        _xPathProcessor = (XPathProcessor) manager.lookup(XPathProcessor.ROLE);
    }
    
    /**
     * Chekcs that the passed Solr field name is valid
     * @param fieldName the Solr field name
     * @return true if name is valid, false otherwise
     */
    public static boolean isNameValid(String fieldName)
    {
        return fieldName.matches("^[a-zA-Z_][a-zA-Z0-9_\\/-]*$");
    }
    
    /**
     * Get the schema at the corresponding location (source URI).
     * @param location The location, as a source URI.
     * @return The schema representation.
     */
    public SchemaRepresentation getSchema(String location)
    {
        Source source = null;
        
        try
        {
            source = _sourceResolver.resolveURI(location);
            
            if (source.exists())
            {
                try (InputStream is = source.getInputStream())
                {
                    return readSchema(is);
                }
            }
        }
        catch (IOException | SAXException e)
        {
            getLogger().error("Error reading the schema from location '" + location + "'", e);
        }
        finally
        {
            if (source != null)
            {
                _sourceResolver.release(source);
            }
        }
        
        return null;
    }
    
    /**
     * Read the static schema.
     * @param is An input stream on the schema XML.
     * @return The representation of the schema.
     * @throws IOException If an error occurs reading the stream.
     * @throws SAXException If an error occurs parsing the XML.
     */
    public SchemaRepresentation readSchema(InputStream is) throws IOException, SAXException
    {
        SchemaRepresentation schema = new SchemaRepresentation();
        
        InputSource source = new InputSource(is);
        Document document = _domParser.parseDocument(source);
        
        Element root = document.getDocumentElement();
        
        String name = _xPathProcessor.evaluateAsString(root, "/schema/@name");
        float version = _xPathProcessor.evaluateAsNumber(root, "/schema/@version").floatValue();
        String uniqueKey = _xPathProcessor.evaluateAsString(root, "/schema/uniqueKey");
        
        schema.setName(name);
        schema.setVersion(version);
        schema.setUniqueKey(uniqueKey);
        
        NodeList fieldTypeNodes = _xPathProcessor.selectNodeList(root, "/schema/fieldType | /schema/types/fieldType");
        schema.setFieldTypes(getFieldTypes(fieldTypeNodes));
        
        NodeList fieldNodes = _xPathProcessor.selectNodeList(root, "/schema/field | /schema/fields/field");
        schema.setFields(filterListNode(getList(fieldNodes)));
        
        NodeList dynFieldNodes = _xPathProcessor.selectNodeList(root, "/schema/dynamicField | /schema/fields/dynamicField");
        schema.setDynamicFields(getList(dynFieldNodes));
        
        NodeList copyFieldNodes = _xPathProcessor.selectNodeList(root, "/schema/copyField");
        schema.setCopyFields(getList(copyFieldNodes));
        
        Node similarityNode = _xPathProcessor.selectSingleNode(root, "/schema/similarity");
        if (similarityNode != null)
        {
            schema.setSimilarity(getAttributes(similarityNode));
        }
        
        return schema;
    }
    
    /**
     * Get the field type definitions from the corresponding DOM nodes.
     * @param fieldTypeNodes The field type nodes.
     * @return The list of field type definitions.
     */
    protected List<FieldTypeDefinition> getFieldTypes(NodeList fieldTypeNodes)
    {
        List<FieldTypeDefinition> definitions = new ArrayList<>();
        
        for (int i = 0; i < fieldTypeNodes.getLength(); i++)
        {
            Node fieldTypeNode = fieldTypeNodes.item(i);
            definitions.add(getFieldType(fieldTypeNode));
        }
        
        return definitions;
    }
    
    /**
     * Get a field type definition from the corresponding DOM node.
     * @param fieldTypeNode The field type DOM node.
     * @return The field type definition.
     */
    protected FieldTypeDefinition getFieldType(Node fieldTypeNode)
    {
        FieldTypeDefinition fieldType = new FieldTypeDefinition();
        
        fieldType.setAttributes(getAttributes(fieldTypeNode));
        
        NodeList analyzerNodes = _xPathProcessor.selectNodeList(fieldTypeNode, "analyzer");
        
        for (int i = 0; i < analyzerNodes.getLength(); i++)
        {
            Node analyzerNode = analyzerNodes.item(i);
            Node typeNode = analyzerNode.getAttributes().getNamedItem("type");
            String type = typeNode != null ? typeNode.getNodeValue() : "";
            
            AnalyzerDefinition analyzer = getAnalyzer(analyzerNode);
            
            switch (type)
            {
                case "index":
                    fieldType.setIndexAnalyzer(analyzer);
                    break;
                case "query":
                    fieldType.setQueryAnalyzer(analyzer);
                    break;
                case "multiterm":
                    fieldType.setMultiTermAnalyzer(analyzer);
                    break;
                default:
                    fieldType.setAnalyzer(analyzer);
                    break;
            }
        }
        
        Node similarityNode = _xPathProcessor.selectSingleNode(fieldTypeNode, "similarity");
        if (similarityNode != null)
        {
            fieldType.setSimilarity(getAttributes(similarityNode));
        }
        
        return fieldType;
    }
    
    /**
     * Get an analyzer definition from the corresponding DOM node.
     * @param analyzerNode The analyzer node.
     * @return The analyzer definition.
     */
    protected AnalyzerDefinition getAnalyzer(Node analyzerNode)
    {
        AnalyzerDefinition analyzer = new AnalyzerDefinition();
        
        analyzer.setAttributes(getAttributes(analyzerNode));
        
        NodeList charFilterNodes = _xPathProcessor.selectNodeList(analyzerNode, "charFilter");
        analyzer.setCharFilters(getList(charFilterNodes));
        
        Node tokenizerNode = _xPathProcessor.selectSingleNode(analyzerNode, "tokenizer");
        if (tokenizerNode != null)
        {
            analyzer.setTokenizer(getAttributes(tokenizerNode));
        }
        
        NodeList filterNodes = _xPathProcessor.selectNodeList(analyzerNode, "filter");
        analyzer.setFilters(getList(filterNodes));
        
        return analyzer;
    }
    
    /**
     * Extract a DOM node attributes as a Map.
     * @param node The node.
     * @return The attributes as a Map.
     */
    protected Map<String, Object> getAttributes(Node node)
    {
        Map<String, Object> map = new HashMap<>();
        
        NamedNodeMap attributes = node.getAttributes();
        for (int i = 0; i < attributes.getLength(); i++)
        {
            Node attribute = attributes.item(i);
            
            String name = attribute.getNodeName();
            String value = attribute.getNodeValue();
            
            // Filter out ametys-specific attributes.
            if (!name.startsWith("ametys"))
            {
                map.put(name, value);
            }
        }
        
        return map;
    }
    
    /**
     * Extract the list of DOM node attributes.
     * @param nodeList The node list.
     * @return A List of the attribute values as Maps.
     */
    protected List<Map<String, Object>> getList(NodeList nodeList)
    {
        List<Map<String, Object>> list = new ArrayList<>();
        
        for (int i = 0; i < nodeList.getLength(); i++)
        {
            Node node = nodeList.item(i);
            list.add(getAttributes(node));
        }
        
        return list;
    }
    
    /**
     * Filters the list of field declarations and only keep valid ones.
     * @param fieldList The list of fields to filter
     * @return The filtered list
     */
    protected List<Map<String, Object>> filterListNode(List<Map<String, Object>> fieldList)
    {
        return fieldList.stream().filter(fieldMap -> 
        {
            String fieldName = (String) fieldMap.get("name");
            if (fieldName == null)
            {
                getLogger().warn("'name' attribute for field node cannot be null. Field will be ignored.");
                return false;
            }
            else if (!isNameValid(fieldName))
            {
                // https://lucene.apache.org/solr/guide/6_6/defining-fields.html#DefiningFields-FieldProperties
                getLogger().warn("Invalid field name: '{}'. Field names should consist of alphanumeric or underscore characters only and not start with a digit (Ametys also supports dash character). Field will be ignored.", fieldName);
                return false;
            }
            return true;
        }).collect(Collectors.toList());
    }
    
}
