/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.systemprop;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.solr.common.SolrInputDocument;

import org.ametys.cms.data.ametysobject.ModelAwareDataAwareAmetysObject;
import org.ametys.cms.data.type.indexing.IndexableDataContext;
import org.ametys.cms.data.type.indexing.IndexableElementType;
import org.ametys.cms.model.properties.AbstractProperty;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.solr.schema.CopyFieldDefinition;
import org.ametys.cms.search.solr.schema.FieldDefinition;
import org.ametys.cms.search.solr.schema.SchemaDefinition;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.util.I18nUtils;
import org.ametys.core.util.date.AdaptableDate;
import org.ametys.core.util.date.AdaptableDateParser;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.plugins.repository.AmetysObjectResolver;

/**
 * Abstract class providing standard behavior and helper methods for a {@link SystemProperty}.
 * @param <T> type of the property values
 * @param <X> type of ametys object supported by this property
 */
public abstract class AbstractSystemProperty<T, X extends ModelAwareDataAwareAmetysObject> extends AbstractProperty<T, X> implements SystemProperty<T, X>, Serviceable
{
    /** The i18n utils. */
    protected I18nUtils _i18nUtils;
    /** The user helper */
    protected UserHelper _userHelper;
    /** The ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /** The user manager */
    protected UserManager _userManager;
    /** The service manager */
    protected ServiceManager _manager;

    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
        _userHelper = (UserHelper) manager.lookup(UserHelper.ROLE);
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
    }
    
    @Override
    protected String _getNameConfigurationAttribute()
    {
        return "id";
    }

    @Override
    public void indexValue(SolrInputDocument document, X ametysObject, IndexableDataContext context)
    {
        Object value = getValue(ametysObject);
        
        SearchField searchField = getSearchField();
        
        if (value == null || searchField == null || searchField.getName() == null)
        {
            // Nothing to index
            return;
        }
        
        IndexableElementType<T> type = getType();
        if (type.getManagedClass().isInstance(value))
        {
            Object valueToIndex = type.getSingleValueToIndex(type.getManagedClass().cast(value));
            document.addField(searchField.getName(), valueToIndex);
        }
        else if (type.getManagedClassArray().isInstance(value))
        {
            T[] values = type.getManagedClassArray().cast(value);
            for (T singleValue : values)
            {
                Object valueToIndex = type.getSingleValueToIndex(singleValue);
                document.addField(searchField.getName(), valueToIndex);
            }
        }
        
        // Index sort field
        Object sortValue = getSortValue(ametysObject);
        if (isSortable() && sortValue != null && searchField.getSortField() != null && !searchField.getName().equals(searchField.getSortField()))
        {
            document.setField(searchField.getSortField(), sortValue);
        }
    }

    @Override
    public Object getSortValue(X ametysObject)
    {
        // Default to getValue(), override to provide a specific sort value.
        Object value = getValue(ametysObject);
        
        if (value == null)
        {
            return null;
        }

        if (isMultiple() && value instanceof Object[] && ((Object[]) value).length > 0)
        {
            return ((Object[]) value)[0]; // return the first value
        }
        else
        {
            return value;
        }
    }
    
    @Override
    public Collection<SchemaDefinition> getSchemaDefinitions()
    {
        List<SchemaDefinition> definitions = new ArrayList<>();

        SearchField searchField = getSearchField();
        if (searchField != null)
        {
            String name = searchField.getName();
            String sortFieldName = searchField.getSortField();
            String facetFieldName = searchField.getFacetField();
            boolean multiple = isMultiple();
            String type = getType().getSchemaType();

            if (type != null)
            {
                definitions.add(new FieldDefinition(name, type, multiple, false));

                if (sortFieldName != null && !sortFieldName.equals(name))
                {
                    definitions.add(new FieldDefinition(sortFieldName, type, false, false));
                }

                if (facetFieldName != null && !facetFieldName.equals(name))
                {
                    // By default the index value in field name will be automatically copy in facet field
                    // So we do not need the index facet field manually
                    definitions.add(new FieldDefinition(facetFieldName, type, multiple, true));
                    definitions.add(new CopyFieldDefinition(name, facetFieldName));
                }
            }
        }

        return definitions;
    }
    
    /**
     * Get the value as a long.
     * 
     * @param value The value as an object.
     * @return The value as a long.
     */
    protected String parseString(Object value)
    {
        if (value instanceof String)
        {
            return (String) value;
        }
        else
        {
            throw new IllegalArgumentException("The value " + value + " for criterion " + getName() + " is not a valid String value.");
        }
    }

    /**
     * Get the value as a long.
     * 
     * @param value The value as an object.
     * @return The value as a long.
     */
    protected long parseLong(Object value)
    {
        if (value instanceof Number)
        {
            return ((Number) value).longValue();
        }
        else if (value instanceof String)
        {
            return Long.parseLong((String) value);
        }
        else
        {
            throw new IllegalArgumentException("The value " + value + " for criterion " + getName() + " is not a valid long value.");
        }
    }

    /**
     * Get the value as a double.
     * 
     * @param value The value as an object.
     * @return The value as a double.
     */
    protected double parseDouble(Object value)
    {
        if (value instanceof Number)
        {
            return ((Number) value).doubleValue();
        }
        else if (value instanceof String)
        {
            return Double.parseDouble((String) value);
        }
        else
        {
            throw new IllegalArgumentException("The value " + value + " for criterion " + getName() + " is not a valid double value.");
        }
    }

    /**
     * Get the value as a boolean.
     * 
     * @param value The value as an object, can be a Boolean or a String.
     * @return The value as a boolean.
     */
    protected boolean parseBoolean(Object value)
    {
        if (value instanceof Boolean)
        {
            return (Boolean) value;
        }
        else if (value instanceof String)
        {
            return Boolean.parseBoolean((String) value);
        }
        else
        {
            throw new IllegalArgumentException("The value " + value + " for criterion " + getName() + " is not a valid boolean value.");
        }
    }

    /**
     * Get the value as a date.
     * 
     * @param value The value as an object, can be a Date or a String.
     * @return The value as a Date object.
     */
    protected AdaptableDate parseDate(Object value)
    {
        if (value instanceof AdaptableDate)
        {
            return (AdaptableDate) value;
        }
        else if (value instanceof String)
        {
            return AdaptableDateParser.parse((String) value);
        }
        else
        {
            throw new IllegalArgumentException("The value " + value + " for criterion " + getName() + " is not a valid date value.");
        }
    }
    
    /**
     * Get the value as array of {@link UserIdentity}
     * @param value The value to parse
     * @return A array of {@link UserIdentity}
     */
    @SuppressWarnings("unchecked")
    protected UserIdentity[] parseUserArray (Object value)
    {
        if (value instanceof UserIdentity)
        {
            return new UserIdentity[] {(UserIdentity) value};
        }
        else if (value instanceof UserIdentity[])
        {
            return (UserIdentity[]) value;
        }
        else if (value instanceof List<?>)
        {
            return (UserIdentity[]) ((List<?>) value).stream().map(v -> _userHelper.json2userIdentity((Map<String, String>) v)).toArray();
        }
        else if (value instanceof Map)
        {
            UserIdentity userIdentity = _userHelper.json2userIdentity((Map<String, String>) value);
            if (userIdentity != null)
            {
                return new UserIdentity[] {userIdentity};
            }
            return new UserIdentity[0];
        }
        else
        {
            throw new IllegalArgumentException("The value " + value + " for criterion " + getName() + " is not a valid UserIdentity values.");
        }
    }

    /**
     * Get the value as a array of String
     * @param value The value as an object to parse
     * @return The values as a String array
     */
    protected String[] parseStringArray(Object value)
    {
        if (value instanceof String)
        {
            return new String[] {(String) value};
        }
        else if (value instanceof String[])
        {
            return (String[]) value;
        }
        else if (value instanceof List<?>)
        {
            return ((List<?>) value).stream()
                    .map(v -> String.valueOf(v))
                    .toArray(String[]::new);
        }
        else
        {
            throw new IllegalArgumentException("The value " + value + " for criterion " + getName() + " is not a valid String[] values.");
        }
    }
}
