/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.systemprop;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.solr.common.SolrInputDocument;

import org.ametys.cms.content.indexing.solr.SolrFieldNames;
import org.ametys.cms.data.type.indexing.IndexableDataContext;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.comment.Comment;
import org.ametys.cms.repository.comment.CommentableContent;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.query.CommentQuery;
import org.ametys.cms.search.query.CommentQuery.CommentInclusion;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.solr.field.CommentsSearchField;
import org.ametys.cms.search.solr.schema.FieldDefinition;
import org.ametys.cms.search.solr.schema.SchemaDefinition;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.Enumerator;
import org.ametys.runtime.model.StaticEnumerator;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * {@link SystemProperty} which represents the comments of a content.
 */
public class CommentsSystemProperty extends AbstractSystemProperty<Boolean, Content>
{
    @Override
    public boolean isMultiple()
    {
        return false;
    }
    
    @Override
    public boolean isSortable()
    {
        return true;
    }
    
    @Override
    public Query getQuery(Object value, Operator operator, String language, Map<String, Object> contextualParameters)
    {
        if ("with-comments".equals(value))
        {
            return new CommentQuery(CommentInclusion.VALIDATED_AND_NON_VALIDATED, operator);
        }
        else if ("with-validated-comments".equals(value))
        {
            return new CommentQuery(CommentInclusion.VALIDATED, operator);
        }
        else // if ("with-nonvalidated-comments".equals(value))
        {
            return new CommentQuery(CommentInclusion.NON_VALIDATED, operator);
        }
    }
    
    @Override
    public String getCriterionWidget()
    {
        // Force standard combobox as the default widget, as the property type is boolean.
        return "edition.combobox";
    }
    
    @Override
    public String getRenderer()
    {
        return "Ametys.plugins.cms.search.SearchGridHelper.renderBooleanIcon";
    }
    
    @Override
    public Integer getColumnWidth()
    {
        return 40;
    }
    
    @Override
    public SearchField getSearchField()
    {
        return new CommentsSearchField();
    }
    
    @Override
    public Collection<SchemaDefinition> getSchemaDefinitions()
    {
        Collection<SchemaDefinition> definitions = super.getSchemaDefinitions();
        
        definitions.add(new FieldDefinition(SolrFieldNames.CONTENT_COMMENTS_VALIDATED, "boolean", false, false));
        definitions.add(new FieldDefinition(SolrFieldNames.CONTENT_COMMENTS_NONVALIDATED, "boolean", false, false));
        
        return definitions;
    }
    
    @Override
    public void indexValue(SolrInputDocument document, Content content, IndexableDataContext context)
    {
        super.indexValue(document, content, context);
        
        boolean hasComment = (boolean) getValue(content);
        if (hasComment)
        {
            List<Comment> comments = ((CommentableContent) content).getComments(true, true);
            document.addField(SolrFieldNames.CONTENT_COMMENTS_VALIDATED, comments.stream().anyMatch(c -> c.isValidated()));
            document.addField(SolrFieldNames.CONTENT_COMMENTS_NONVALIDATED, comments.stream().anyMatch(c -> !c.isValidated()));
        }
    }
    
    @Override
    public Object getValue(Content content)
    {
        if (content instanceof CommentableContent commentableContent)
        {
            List<Comment> comments = commentableContent.getComments(true, true);
            return Boolean.valueOf(!comments.isEmpty());
        }
        
        return Boolean.valueOf(false);
    }
    
    @Override
    public Object valueToJSON(Content content, DataContext context)
    {
        // FIXME pourquoi ne pas retourner getRawValue directement ?
        Map<String, Object> infos = new LinkedHashMap<>();
        infos.put("comments", getValue(content));
        return infos;
    }
    
    public Enumerator<String> getCriterionEnumerator(Configuration configuration, ThreadSafeComponentManager<Enumerator> enumeratorManager) throws ConfigurationException
    {
        StaticEnumerator<String> enumerator = new StaticEnumerator<>();
        enumerator.add(new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_COMMENTS_SEARCH_ALL"), "with-comments");
        enumerator.add(new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_COMMENTS_SEARCH_VALIDATED"), "with-validated-comments");
        enumerator.add(new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_COMMENTS_SEARCH_NOTVALIDATED"), "with-nonvalidated-comments");
        return enumerator;
    }
    
    @Override
    protected String _getTypeId()
    {
        return ModelItemTypeConstants.BOOLEAN_TYPE_ID;
    }
}
