/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.systemprop;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeEnumerator;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.query.ContentTypeOrMixinTypeQuery;
import org.ametys.cms.search.query.ContentTypeQuery;
import org.ametys.cms.search.query.MixinTypeQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.solr.field.ContentTypeSearchField;
import org.ametys.cms.search.solr.field.MixinTypeSearchField;
import org.ametys.core.model.type.ModelItemTypeHelper;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.Enumerator;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * {@link SystemProperty} which represents the content types (and optionally mixins) of a content.
 */
public class ContentTypeSystemProperty extends AbstractSystemProperty<String, Content>
{
    /** The content type extension point. */
    protected ContentTypeExtensionPoint _cTypeEP;
    
    /** The content types helper. */
    protected ContentTypesHelper _cTypeHelper;
    
    /** True to include content types. */
    protected boolean _includeCTypes;
    
    /** True to include mixins. */
    protected boolean _includeMixins;
    
    /** True to recursively include supertypes. */
    protected boolean _includeSupertypes;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _cTypeEP = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _cTypeHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        super.configure(configuration);
        
        _includeCTypes = configuration.getChild("contentTypes").getValueAsBoolean(true);
        _includeMixins = configuration.getChild("mixins").getValueAsBoolean(false);
        _includeSupertypes = configuration.getChild("includeSupertypes").getValueAsBoolean(false);
    }
    
    @Override
    public boolean isMultiple()
    {
        return true;
    }
    
    @Override
    public boolean isSortable()
    {
        return true;
    }
    
    @Override
    public Query getQuery(Object value, Operator operator, String language, Map<String, Object> contextualParameters)
    {
        String[] contentTypes = value != null ? parseStringArray(value) : new String[0];
        
        if (_includeCTypes && _includeMixins)
        {
            return new ContentTypeOrMixinTypeQuery(operator, contentTypes);
        }
        else if (_includeCTypes)
        {
            return new ContentTypeQuery(operator, contentTypes);
        }
        else if (_includeMixins)
        {
            return new MixinTypeQuery(operator, contentTypes);
        }
        
        return null;
    }
    
    @Override
    public String getRenderer()
    {
        return "Ametys.plugins.cms.search.SearchGridHelper.renderMultipleString";
    }
    
    @Override
    public SearchField getSearchField()
    {
        if (_includeCTypes && _includeMixins)
        {
            return null;
        }
        else if (_includeCTypes)
        {
            return new ContentTypeSearchField(_includeSupertypes);
        }
        else if (_includeMixins)
        {
            return new MixinTypeSearchField(_includeSupertypes);
        }
        
        return null;
    }
    
    @Override
    public Object getValue(Content content)
    {
        Set<String> types = new LinkedHashSet<>();
        
        if (_includeCTypes)
        {
            _addAll(types, content.getTypes());
        }
        if (_includeMixins)
        {
            _addAll(types, content.getMixinTypes());
        }
        
        return types.toArray(new String[types.size()]);
    }
    
    public Object valueToJSON(Content content, DataContext context)
    {
        String[] cTypeIds = (String[]) getValue(content);
        
        List<I18nizableText> cTypes = new ArrayList<>();
        for (String cTypeId : cTypeIds)
        {
            ContentType cType = _cTypeEP.getExtension(cTypeId);
            if (cType != null)
            {
                cTypes.add(cType.getLabel());
            }
            else if (_logger.isWarnEnabled())
            {
                _logger.warn(String.format("Trying to get the label for an unknown content type : '%s'.", cTypeId));
            }
        }
        
        return cTypes;
    }
    
    @Override
    public Object getSortValue(Content content)
    {
        String[] cTypeIds = (String[]) getValue(content);
        if (cTypeIds.length > 0)
        {
            ContentType cType = _cTypeEP.getExtension(cTypeIds[0]);
            if (cType != null)
            {
                return _i18nUtils.translate(cType.getLabel(), content.getLanguage());
            }
        }
        
        return null;
    }
    
    private void _addAll(Set<String> allTypes, String[] types)
    {
        for (String cType : types)
        {
            allTypes.add(cType);
            if (_includeSupertypes && _cTypeEP.hasExtension(cType))
            {
                allTypes.addAll(_cTypeHelper.getAncestors(cType));
            }
        }
    }
    
    public void valueToSAX(ContentHandler contentHandler, Content content, DataContext context) throws SAXException
    {
        String[] cTypeIds = (String[]) getValue(content);
        
        XMLUtils.startElement(contentHandler, getName());
        
        for (String cTypeId : cTypeIds)
        {
            AttributesImpl attr = ModelItemTypeHelper.getXMLAttributesFromDataContext(context);
            attr.addCDATAAttribute("id", cTypeId);
            
            XMLUtils.startElement(contentHandler, "contentType", attr);
            ContentType cType = _cTypeEP.getExtension(cTypeId);
            cType.getLabel().toSAX(contentHandler);
            XMLUtils.endElement(contentHandler, "contentType");
        }
        
        XMLUtils.endElement(contentHandler, getName());
    }
    
    @Override
    public Enumerator<String> getCriterionEnumerator(Configuration configuration, ThreadSafeComponentManager<Enumerator> enumeratorManager) throws ConfigurationException
    {
        DefaultConfiguration conf = new DefaultConfiguration("criteria");
        
        DefaultConfiguration enumConf = new DefaultConfiguration("enumeration");
        
        DefaultConfiguration customEnumerator = new DefaultConfiguration("custom-enumerator");
        customEnumerator.setAttribute("class", ContentTypeEnumerator.class.getName());
        
        try
        {
            boolean containsRefTable = false;
            Set<String> strictContentTypes = new HashSet<>();
            for (Configuration cTypeConf : configuration.getChild("contentTypes").getChildren("type"))
            {
                String cTypeId = cTypeConf.getAttribute("id");
                ContentType cType = _cTypeEP.getExtension(cTypeId);
                if (cType == null)
                {   
                    _logger.error("A search model references a non-existing content type with id '{}', it will be ignored for system property {} ", cTypeId, getName());
                }
                else
                {
                    strictContentTypes.add(cTypeId);
                    containsRefTable = cType.isReferenceTable() || containsRefTable;
                }
            }
            if (!strictContentTypes.isEmpty() && _includeCTypes)
            {
                DefaultConfiguration cTypeConf = new DefaultConfiguration("strictContentTypes");
                cTypeConf.setValue(StringUtils.join(strictContentTypes, ","));
                customEnumerator.addChild(cTypeConf);
                
                if (containsRefTable)
                {
                    // Search model focuses on at least one reference table, do not exclude reference table for content type enumerator
                    DefaultConfiguration excludeRefTableConf = new DefaultConfiguration("excludeReferenceTable");
                    excludeRefTableConf.setValue(false);
                    customEnumerator.addChild(excludeRefTableConf);
                }
            }
        }
        catch (ConfigurationException e)
        {
            _logger.error("Failed to configure strict content types for system property {}", getName(), e);
        }
        
        if (!_includeMixins)
        {
            DefaultConfiguration excludeConf = new DefaultConfiguration("excludeMixin");
            excludeConf.setValue(true);
            customEnumerator.addChild(excludeConf);
        }
        else if (!_includeCTypes) // Mixins and not CTypes
        {
            DefaultConfiguration includeConf = new DefaultConfiguration("includeMixinOnly");
            includeConf.setValue(true);
            customEnumerator.addChild(includeConf);
        }
        
        DefaultConfiguration allOptionConf = new DefaultConfiguration("all-option");
        allOptionConf.setValue("disabled");
        customEnumerator.addChild(allOptionConf);
        
        enumConf.addChild(customEnumerator);
        conf.addChild(enumConf);
        
        String role = "enumerator";
        enumeratorManager.addComponent(getPluginName(), null, role, ContentTypeEnumerator.class, conf);
        
        try
        {
            enumeratorManager.initialize();
            return enumeratorManager.lookup(role);
        }
        catch (Exception e)
        {
            throw new ConfigurationException("Unable to initialize the content type enumerator for system property '" + getName() + "'.", conf, e);
        }
    } 

    /**
     * Get the default widget to use when rendering this property as a criterion.
     * @return The default widget to use.
     */
    @Override
    public String getCriterionWidget()
    {
        return "edition.select-content-types";
    }
    
    @Override
    public Map<String, I18nizableText> getCriterionWidgetParameters(Configuration configuration)
    {
        Map<String, I18nizableText> parameters = new HashMap<>();
        
        try
        {
            Set<String> strictContentTypes = new HashSet<>();
            for (Configuration cTypeConf : configuration.getChild("contentTypes").getChildren("type"))
            {
                strictContentTypes.add(cTypeConf.getAttribute("id"));
            }
            
            if (!strictContentTypes.isEmpty() && _includeCTypes)
            {
                parameters.put("strictContentTypes", new I18nizableText(StringUtils.join(strictContentTypes, ",")));
            }
        }
        catch (ConfigurationException e)
        {
            _logger.error("Failed to configure strict content types for system property {}", getName(), e);
        }
        
        parameters.put("excludeMixin", new I18nizableText(_includeMixins ? "false" :  "true"));
        parameters.put("includeSupertype", new I18nizableText(_includeSupertypes ? "true" : "false"));
        parameters.put("includeContentTypes", new I18nizableText(_includeCTypes ? "true" : "false"));
        parameters.put("excludeReferenceTable", new I18nizableText("true"));
        parameters.put("excludePrivate", new I18nizableText("true"));
        parameters.put("emptyText", new I18nizableText("plugin.cms", "WIDGET_COMBOBOX_ALL_OPTIONS"));
        
        return parameters;
    }
    
    @Override
    protected String _getTypeId()
    {
        return ModelItemTypeConstants.STRING_TYPE_ID;
    }
}
