/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.systemprop;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Optional;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.languages.Language;
import org.ametys.cms.languages.LanguageEnumerator;
import org.ametys.cms.languages.LanguagesManager;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.query.ContentLanguageQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.solr.field.LanguageSearchField;
import org.ametys.core.model.type.ModelItemTypeHelper;
import org.ametys.runtime.model.Enumerator;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * {@link SystemProperty} which represents the language of a Content.
 */
public class LanguageSystemProperty extends AbstractSystemProperty<String, Content>
{
    /** The language manager. */
    protected LanguagesManager _languageManager;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _languageManager = (LanguagesManager) manager.lookup(LanguagesManager.ROLE);
    }
    
    @Override
    public boolean isMultiple()
    {
        return false;
    }
    
    @Override
    public boolean isSortable()
    {
        return true;
    }
    
    @Override
    public Query getQuery(Object value, Operator operator, String language, Map<String, Object> contextualParameters)
    {
        String langValue = null;
        if ("CURRENT".equals(value))
        {
            langValue = language;
        }
        else
        {
            langValue = (String) value;
        }
        
        return new ContentLanguageQuery(operator, langValue);
    }
    
    @Override
    public String getRenderer()
    {
        return "Ametys.plugins.cms.search.SearchGridHelper.renderLanguage";
    }
    
    @Override
    public String getConverter()
    {
        return "Ametys.plugins.cms.search.SearchGridHelper.convertLanguage";
    }
    
    @Override
    public Integer getColumnWidth()
    {
        return 70;
    }
    
    @Override
    public SearchField getSearchField()
    {
        return new LanguageSearchField();
    }
    
    @Override
    public Object getValue(Content content)
    {
        return content.getLanguage();
    }
    
    public Object valueToJSON(Content content, DataContext context)
    {
        String languageCode = (String) getValue(content);
        Map<String, Object> infos = new LinkedHashMap<>();
        
        infos.put("code", languageCode);
        
        if (languageCode != null)
        {
            Language language = _languageManager.getLanguage(languageCode);
            if (language != null)
            {
                infos.put("icon", language.getSmallIcon());
                infos.put("label", language.getLabel());
            }
        }
        
        return infos;
    }
    
    public void valueToSAX(ContentHandler contentHandler, Content content, DataContext context) throws SAXException
    {
        Language language = Optional.of(content)
                .map(this::getValue)
                .map(String.class::cast)
                .map(_languageManager::getLanguage)
                .orElse(null);
        
        if (language != null)
        {
            AttributesImpl attr = ModelItemTypeHelper.getXMLAttributesFromDataContext(context);
            attr.addCDATAAttribute("code", language.getCode());
            XMLUtils.startElement(contentHandler, getName(),  attr);
            language.getLabel().toSAX(contentHandler);
            XMLUtils.endElement(contentHandler, getName());
        }
    }
    
    @Override
    public Object getSortValue(Content content)
    {
        Language language = Optional.of(content)
                .map(this::getValue)
                .map(String.class::cast)
                .map(_languageManager::getLanguage)
                .orElse(null);
        
        if (language != null)
        {
            return _i18nUtils.translate(language.getLabel(), language.getCode());
        }
        
        return null;
    }
    
    public Enumerator<String> getCriterionEnumerator(Configuration configuration, ThreadSafeComponentManager<Enumerator> enumeratorManager) throws ConfigurationException
    {
        String role = "enumerator";
        enumeratorManager.addComponent(getPluginName(), null, role, LanguageEnumerator.class, configuration);
        
        try
        {
            enumeratorManager.initialize();
            return enumeratorManager.lookup(role);
        }
        catch (Exception e)
        {
            throw new ConfigurationException("Unable to initialize the language enumerator for system property '" + getName() + "'.", configuration, e);
        }
    }
    
    @Override
    protected String _getTypeId()
    {
        return ModelItemTypeConstants.STRING_TYPE_ID;
    }
}
