/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.systemprop;

import java.util.Map;
import java.util.Optional;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.apache.solr.common.SolrInputDocument;

import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.data.holder.impl.IndexableDataHolderHelper;
import org.ametys.cms.data.type.indexing.IndexableDataContext;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.query.FullTextQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.type.ModelItemTypeConstants;

/**
 * {@link SystemProperty} which represents the full textual content of a Content based on attributes of a view.
 * If the view does not exist for content, nothing will be indexed.
 */
public class ViewBasedFullTextSystemProperty extends AbstractSystemProperty<String, Content>
{
    /** The contents types helper */
    protected ContentTypesHelper _contentTypesHelper;
    
    /** The view name to fill the property */
    protected String _viewName;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        super.configure(configuration);
        _viewName = configuration.getChild("view").getValue();
        if (StringUtils.isBlank(_viewName))
        {
            throw new ConfigurationException("'view' configuration is missing for indexing property '" + this.getName() + "'");
        }
    }
    
    @Override
    public void indexValue(SolrInputDocument document, Content content, IndexableDataContext context)
    {
        Optional<View> view = Optional.of(_viewName)
                .map(v -> _contentTypesHelper.getView(v, content));
        
        if (view.isPresent())
        {
            IndexableDataContext clonedContext = IndexableDataContext.newInstance(context)
                    .withIndexForFullTextField(true)
                    .withFullTextFieldName(getName());
            
            IndexableDataHolderHelper.indexData(content, view.get(), document, document, StringUtils.EMPTY, clonedContext);
        }
    }
    
    @Override
    public boolean isMultiple()
    {
        return false;
    }
    
    @Override
    public boolean isDisplayable()
    {
        return false;
    }
    
    @Override
    public boolean isSortable()
    {
        return false;
    }
    
    @Override
    public Query getQuery(Object value, Operator operator, String language, Map<String, Object> contextualParameters)
    {
        return new FullTextQuery((String) value, getName(), language, operator);
    }
    
    @Override
    public SearchField getSearchField()
    {
        // No value to index, not sortable: the field won't be used.
        return null;
    }
    
    @Override
    public Object getValue(Content content)
    {
        // No real "value" to return here.
        // The values are indexed separately, while walking the content attribute tree.
        return null;
    }

    @Override
    protected String _getTypeId()
    {
        return ModelItemTypeConstants.STRING_TYPE_ID;
    }
}
