/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.systemprop;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.query.OrQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.query.WorkflowStepQuery;
import org.ametys.cms.search.solr.field.WorkflowStepSearchField;
import org.ametys.cms.workflow.DefaultWorkflowStepEnumerator;
import org.ametys.core.model.type.ModelItemTypeHelper;
import org.ametys.plugins.workflow.support.WorkflowHelper;
import org.ametys.plugins.workflow.support.WorkflowProvider;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.Enumerator;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

import com.opensymphony.workflow.loader.StepDescriptor;
import com.opensymphony.workflow.loader.WorkflowDescriptor;
import com.opensymphony.workflow.spi.Step;

/**
 * {@link SystemProperty} which represents the current workflow step ID of a content.
 */
public class WorkflowStepSystemProperty extends AbstractSystemProperty<Long, Content>
{
    /** The workflow provider */
    protected WorkflowProvider _workflowProvider;
    /** The workflow helper */
    protected WorkflowHelper _workflowHelper;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _workflowProvider = (WorkflowProvider) manager.lookup(WorkflowProvider.ROLE);
        _workflowHelper = (WorkflowHelper) manager.lookup(WorkflowHelper.ROLE);
    }
    
    @Override
    public boolean isMultiple()
    {
        return false;
    }
    
    @Override
    public boolean isSortable()
    {
        return true;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public Query getQuery(Object value, Operator operator, String language, Map<String, Object> contextualParameters)
    {
        if (value == null)
        {
            return new WorkflowStepQuery(Operator.EXISTS);
        }
        
        List<Object> values;
        if (value instanceof List)
        {
            values = (List<Object>) value;
        }
        else
        {
            values = List.of(value);
        }
        
        List<Query> queries = new ArrayList<>();
        for (Object val : values)
        {
            if (!"".equals(val)) // empty value when selecting "All" entry in the widget
            {
                long stepId = parseLong(val);
                if (stepId != 0)
                {
                    queries.add(new WorkflowStepQuery(operator, (int) stepId));
                }
            }
        }
        
        switch (queries.size())
        {
            case 0: return null;
            case 1: return queries.get(0);
            default: return new OrQuery(queries);
        }
    }
    
    @Override
    public SearchField getSearchField()
    {
        return new WorkflowStepSearchField();
    }
    
    @Override
    public String getRenderer()
    {
        return "Ametys.plugins.cms.search.SearchGridHelper.renderWorkflowStep";
    }
    
    @Override
    public String getConverter()
    {
        return "Ametys.plugins.cms.search.SearchGridHelper.convertWorkflowStep";
    }
    
    @Override
    public Integer getColumnWidth()
    {
        return 60;
    }
    
    @Override
    public Object getValue(Content content)
    {
        if (content instanceof WorkflowAwareContent waContent)
        {
            AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(waContent);
            
            long workflowId = waContent.getWorkflowId();
            List<Step> currentSteps = workflow.getCurrentSteps(workflowId);
            
            if (!currentSteps.isEmpty())
            {
                return Long.valueOf(currentSteps.iterator().next().getStepId());
            }
        }
        
        return null;
    }
    
    public Object valueToJSON(Content content, DataContext context)
    {
        Long value = (Long) getValue(content);
        if (value != null)
        {
            Map<String, Object> workflowInfos = new LinkedHashMap<>();
            
            if (content instanceof WorkflowAwareContent)
            {
                WorkflowAwareContent waContent = (WorkflowAwareContent) content;
                int currentStepId = Math.toIntExact(value);
                
                StepDescriptor stepDescriptor = _workflowHelper.getStepDescriptor(waContent, currentStepId);
                
                if (stepDescriptor != null)
                {
                    return _workflowHelper.workflowStep2JSON(stepDescriptor);
                }
            }
            
            return workflowInfos;
        }
        return value;
    }
    
    public void valueToSAX(ContentHandler contentHandler, Content content, DataContext context) throws SAXException
    {
        Long value = (Long) getValue(content);
        if (value != null && content instanceof WorkflowAwareContent waContent)
        {
            int currentStepId = Math.toIntExact(value);
            StepDescriptor stepDescriptor = _getStepDescriptor(waContent, currentStepId);
            
            if (stepDescriptor != null)
            {
                I18nizableText workflowStepName = new I18nizableText("application", stepDescriptor.getName());
                
                AttributesImpl attr = ModelItemTypeHelper.getXMLAttributesFromDataContext(context);
                attr.addCDATAAttribute("id", String.valueOf(currentStepId));
                XMLUtils.startElement(contentHandler, getName(), attr);
                workflowStepName.toSAX(contentHandler);
                XMLUtils.endElement(contentHandler, getName());
            }
        }
    }
    
    private StepDescriptor _getStepDescriptor(WorkflowAwareContent content, int stepId)
    {
        long workflowId = content.getWorkflowId();
        
        AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(content);
        
        String workflowName = workflow.getWorkflowName(workflowId);
        WorkflowDescriptor workflowDescriptor = workflow.getWorkflowDescriptor(workflowName);
        
        if (workflowDescriptor != null)
        {
            StepDescriptor stepDescriptor = workflowDescriptor.getStep(stepId);
            if (stepDescriptor != null)
            {
                return stepDescriptor;
            }
            else if (_logger.isWarnEnabled())
            {
                _logger.warn("Unknown step id '" + stepId + "' for workflow for name : " + workflowName);
            }
        }
        else if (_logger.isWarnEnabled())
        {
            _logger.warn("Unknown workflow for name : " + workflowName);
        }
        
        return null;
    }
    
    public Enumerator<String> getCriterionEnumerator(Configuration configuration, ThreadSafeComponentManager<Enumerator> enumeratorManager) throws ConfigurationException
    {
        DefaultConfiguration conf = new DefaultConfiguration("criteria");
        
        String workflowName = configuration.getChild("workflow").getAttribute("name", "content");
        
        DefaultConfiguration enumConf = new DefaultConfiguration("enumeration");
        
        DefaultConfiguration customEnumerator = new DefaultConfiguration("custom-enumerator");
        customEnumerator.setAttribute("class", DefaultWorkflowStepEnumerator.class.getName());
        enumConf.addChild(customEnumerator);
        
        DefaultConfiguration wfNameConf = new DefaultConfiguration("workflow-name");
        wfNameConf.setValue(workflowName);
        customEnumerator.addChild(wfNameConf);
        
        DefaultConfiguration excludeConf = new DefaultConfiguration("exclude-workflow-steps");
        
        // Exclude workflow steps greater than 9000
        List<StepDescriptor> steps = _workflowHelper.getWorkflowDescriptor(workflowName).getSteps();
        for (StepDescriptor stepDescriptor : steps)
        {
            if (stepDescriptor.getId() >= 9000)
            {
                DefaultConfiguration stepId = new DefaultConfiguration("id");
                stepId.setValue(stepDescriptor.getId());
                excludeConf.addChild(stepId);
            }
        }
        
        customEnumerator.addChild(excludeConf);
        
        conf.addChild(enumConf);
        
        String role = "enumerator";
        enumeratorManager.addComponent(getPluginName(), null, role, DefaultWorkflowStepEnumerator.class, conf);
        
        try
        {
            enumeratorManager.initialize();
            return enumeratorManager.lookup(role);
        }
        catch (Exception e)
        {
            throw new ConfigurationException("Unable to initialize the workflow step enumerator for system property '" + getName() + "'.", conf, e);
        }
    }

    @Override
    protected String _getTypeId()
    {
        return ModelItemTypeConstants.LONG_TYPE_ID;
    }
}
