/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.ui.model;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;

import org.ametys.cms.search.model.SearchCriterion;
import org.ametys.runtime.model.ViewItemContainer;
import org.ametys.runtime.plugin.component.LogEnabled;

/**
 * Abstract class for SearchUIModel.
 */
public abstract class AbstractSearchUIModel implements SearchUIModel, LogEnabled
{
    /** The default plugin name for URLs */
    protected static final String DEFAULT_URL_PLUGIN = "cms";
    /** The default URL for search */
    protected static final String DEFAULT_SEARCH_URL = "search/list.json";
    /** The default URL for CSV export */
    protected static final String DEFAULT_EXPORT_CSV_URL = "search/export.csv";
    /** The default URL for DOC export */
    protected static final String DEFAULT_EXPORT_DOC_URL = "search/export.doc";
    /** The default URL for XML export */
    protected static final String DEFAULT_EXPORT_XML_URL = "search/export.xml";
    /** The default URL for PDF export */
    protected static final String DEFAULT_EXPORT_PDF_URL = null;
    /** The URL for print results */
    protected static final String DEFAULT_PRINT_URL = "search/print.html";
    
    /** The logger. */
    protected Logger _logger;
    
    /** The content types of this search model. */
    protected Set<String> _cTypes;
    /** The content types excluded from this search model. */
    protected Set<String> _excludedCTypes;
    
    /** The search criteria in simple mode, indexed by ID. */
    protected Map<String, SearchUICriterion> _searchCriteria;
    /** The search criteria in advanced mode, indexed by ID. */
    protected Map<String, SearchUICriterion> _advancedSearchCriteria;
    /** The search criteria used as facets, indexed by ID. */
    protected Map<String, SearchUICriterion> _facetedCriteria;
    /** The result columns */
    protected ViewItemContainer _resultItems;
    
    public void setLogger(final Logger logger)
    {
        _logger = logger;
    }
    
    /**
     * Get the logger.
     * @return the logger.
     */
    protected final Logger getLogger()
    {
        return _logger;
    }
    
    @Override
    public Set<String> getContentTypes(Map<String, Object> contextualParameters)
    {
        return Collections.unmodifiableSet(_cTypes);
    }
    
    /**
     * Set the content types.
     * @param cTypes The content types.
     */
    public void setContentTypes(Set<String> cTypes)
    {
        _cTypes = new HashSet<>(cTypes);
    }
    
    @Override
    public Set<String> getExcludedContentTypes(Map<String, Object> contextualParameters)
    {
        return Collections.unmodifiableSet(_excludedCTypes);
    }
    
    /**
     * Set the excluded content types.
     * @param cTypes The excluded content types.
     */
    public void setExcludedContentTypes(Set<String> cTypes)
    {
        _excludedCTypes = new HashSet<>(cTypes);
    }
    
    @Override
    public Map<String, SearchUICriterion> getCriteria(Map<String, Object> contextualParameters)
    {
        return Collections.unmodifiableMap(_searchCriteria);
    }
    
    /**
     * Set the criteria in simple mode.
     * @param criteria A collection of search criteria.
     */
    public void setCriteria(Collection<SearchUICriterion> criteria)
    {
        _searchCriteria = new LinkedHashMap<>();
        for (SearchCriterion criterion : criteria)
        {
            _searchCriteria.put(criterion.getId(), (SearchUICriterion) criterion);
        }
    }
    
    @Override
    public Map<String, SearchUICriterion> getAdvancedCriteria(Map<String, Object> contextualParameters)
    {
        return Collections.unmodifiableMap(_advancedSearchCriteria);
    }
    
    /**
     * Set the criteria in advanced mode.
     * @param criteria A collection of search criteria.
     */
    public void setAdvancedCriteria(Collection<SearchUICriterion> criteria)
    {
        _advancedSearchCriteria = new LinkedHashMap<>();
        for (SearchUICriterion criterion : criteria)
        {
            _advancedSearchCriteria.put(criterion.getId(), criterion);
        }
    }
    
    @Override
    public Map<String, SearchUICriterion> getFacetedCriteria(Map<String, Object> contextualParameters)
    {
        return Collections.unmodifiableMap(_facetedCriteria);
    }
    
    /**
     * Set the criteria to use as facets.
     * @param criteria A collection of search criteria.
     */
    public void setFacetedCriteria(Collection<SearchCriterion> criteria)
    {
        _facetedCriteria = new LinkedHashMap<>();
        for (SearchCriterion criterion : criteria)
        {
            _facetedCriteria.put(criterion.getId(), (SearchUICriterion) criterion);
        }
    }
    
    public ViewItemContainer getResultItems(Map<String, Object> contextualParameters)
    {
        return _resultItems;
    }
    
    /**
     * Set the result items
     * @param resultItems The result items to set
     */
    public void setResultItems(ViewItemContainer resultItems)
    {
        _resultItems = resultItems;
    }
    
    @Override
    public int getPageSize(Map<String, Object> contextualParameters)
    {
        // Use the default value or unlimited.
        return -1;
    }
    
    @Override
    public String getWorkspace(Map<String, Object> contextualParameters)
    {
        // Use the default workspace.
        return null;
    }
    
    @Override
    public String getSearchUrl(Map<String, Object> contextualParameters)
    {
        return DEFAULT_SEARCH_URL;
    }
    
    @Override
    public String getSearchUrlPlugin(Map<String, Object> contextualParameters)
    {
        return DEFAULT_URL_PLUGIN;
    }
    
    @Override
    public String getExportCSVUrl(Map<String, Object> contextualParameters)
    {
        return DEFAULT_EXPORT_CSV_URL;
    }
    
    @Override
    public String getExportCSVUrlPlugin(Map<String, Object> contextualParameters)
    {
        return DEFAULT_URL_PLUGIN;
    }

    @Override
    public String getExportDOCUrl(Map<String, Object> contextualParameters)
    {
        return DEFAULT_EXPORT_DOC_URL;
    }

    @Override
    public String getExportDOCUrlPlugin(Map<String, Object> contextualParameters)
    {
        return DEFAULT_URL_PLUGIN;
    }

    @Override
    public String getExportXMLUrl(Map<String, Object> contextualParameters)
    {
        return DEFAULT_EXPORT_XML_URL;
    }

    @Override
    public String getExportXMLUrlPlugin(Map<String, Object> contextualParameters)
    {
        return DEFAULT_URL_PLUGIN;
    }
    
    @Override
    public String getExportPDFUrl(Map<String, Object> contextualParameters)
    {
        return DEFAULT_EXPORT_PDF_URL;
    }
    
    @Override
    public String getExportPDFUrlPlugin(Map<String, Object> contextualParameters)
    {
        return DEFAULT_URL_PLUGIN;
    }

    @Override
    public String getPrintUrl(Map<String, Object> contextualParameters)
    {
        return DEFAULT_PRINT_URL;
    }

    @Override
    public String getPrintUrlPlugin(Map<String, Object> contextualParameters)
    {
        return DEFAULT_URL_PLUGIN;
    }
    
    @Override
    public String getSummaryView()
    {
        return null;
    }
}
