/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.ui.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import org.apache.avalon.framework.component.ComponentException;
import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.cms.content.referencetable.HierarchicalReferenceTablesHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.model.SearchCriterionHelper;
import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.ui.model.impl.IndexingFieldSearchUICriterion;
import org.ametys.cms.search.ui.model.impl.SystemSearchUICriterion;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElement;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.ViewItemContainer;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * Generic implementation of {@link SearchUIModel} for reference tables
 * The search tool model automatically declares simple first level attributes as criteria and columns.
 */
public class ReferenceTableSearchUIModel extends AbstractSearchUIModel implements Serviceable, Contextualizable, Configurable
{
    /** ComponentManager for {@link SearchUICriterion}s. */
    protected ThreadSafeComponentManager<SearchUICriterion> _searchCriteriaManager;
    
    /** The search criteria roles. */
    protected List<String> _searchCriteriaRoles;
    
    /** The search column roles. */
    protected List<String> _searchColumnRoles;
    
    /** The context. */
    protected Context _context;
    
    /** The service manager */
    protected ServiceManager _manager;
    
    /** The helper component for hierarchical reference tables */
    protected HierarchicalReferenceTablesHelper _hierarchicalReferenceTableContentsHelper;
    
    /** The content type extension point */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    
    /** The systemPropertyExtension point */
    protected SystemPropertyExtensionPoint _systemPropertyExtensionPoint;
    
    /** The search criterion helper */
    protected SearchCriterionHelper _searchCriterionHelper;
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
        
        _hierarchicalReferenceTableContentsHelper = (HierarchicalReferenceTablesHelper) manager.lookup(HierarchicalReferenceTablesHelper.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _systemPropertyExtensionPoint = (SystemPropertyExtensionPoint) manager.lookup(SystemPropertyExtensionPoint.ROLE);
        _searchCriterionHelper = (SearchCriterionHelper) manager.lookup(SearchCriterionHelper.ROLE);
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        try
        {
            String cTypeId = configuration.getChild("contentType").getValue(null);
            if (cTypeId != null)
            {
                setContentTypes(Collections.singleton(cTypeId));
            }
            
            _searchCriteriaManager = new ThreadSafeComponentManager<>();
            _searchCriteriaManager.setLogger(getLogger());
            _searchCriteriaManager.contextualize(_context);
            _searchCriteriaManager.service(_manager);
        }
        catch (Exception e)
        {
            throw new ConfigurationException("Unable to create local component managers.", configuration, e);
        }
    }
    
    @Override
    public Set<String> getExcludedContentTypes(Map<String, Object> contextualParameters)
    {
        return Collections.emptySet();
    }
    
    @Override
    public Map<String, SearchUICriterion> getCriteria(Map<String, Object> contextualParameters)
    {
        try
        {
            if (_searchCriteria == null)
            {
                String cTypeId = getContentTypes(contextualParameters).iterator().next();
                ContentType cType = _contentTypeExtensionPoint.getExtension(cTypeId);
                
                _searchCriteriaRoles = new ArrayList<>();
                
                addCriteriaComponents(cType);
                _searchCriteriaManager.initialize();
                setCriteria(getSearchUICriteria(cType));
            }
        }
        catch (Exception e)
        {
            throw new RuntimeException("Impossible to initialize criteria components.", e);
        }
        
        return _searchCriteria;
    }
    
    @Override
    public Map<String, SearchUICriterion> getFacetedCriteria(Map<String, Object> contextualParameters)
    {
        return Collections.emptyMap();
    }
    
    @Override
    public Map<String, SearchUICriterion> getAdvancedCriteria(Map<String, Object> contextualParameters)
    {
        return Collections.emptyMap();
    }
    
    @Override
    public ViewItemContainer getResultItems(Map<String, Object> contextualParameters)
    {
        if (_resultItems == null)
        {
            String contentTypeId = getContentTypes(contextualParameters).iterator().next();
            ContentType contentType = _contentTypeExtensionPoint.getExtension(contentTypeId);
            
            ViewItemContainer resultItems = _getResultItems(contentType);
            setResultItems(resultItems);
        }
        
        return _resultItems;
    }
    
    /**
     * Add criteria components to the manager.
     * @param cType the simple content type.
     * @throws ConfigurationException if a configuration error occurs.
     * @throws ComponentException if a component cannot be initialized.
     */
    protected void addCriteriaComponents(ContentType cType) throws ConfigurationException, ComponentException
    {
        View view = cType.getView("criteria");
        if (view == null)
        {
            view = cType.getView("main");
        }

        addCriteriaComponents(cType, view);
        
        if (_hierarchicalReferenceTableContentsHelper.isHierarchical(cType) && _hierarchicalReferenceTableContentsHelper.supportCandidates(cType))
        {
            addExcludeCandidateSystemCriterionComponent(cType);
        }

        addSystemCriterionComponent(cType, "contributor");
        
        if (!cType.isMultilingual())
        {
            addSystemCriterionComponent(cType, "contentLanguage");
        }
    }

    /**
     * Add criteria components to the manager.
     * @param cType the simple content type.
     * @param viewContainer the view item container
     * @throws ConfigurationException if a configuration error occurs.
     * @throws ComponentException if a component cannot be initialized.
     */
    protected void addCriteriaComponents(ContentType cType, ViewItemContainer viewContainer) throws ConfigurationException, ComponentException
    {
        for (ViewItem viewItem : viewContainer.getViewItems())
        {
            if (viewItem instanceof ViewItemContainer)
            {
                addCriteriaComponents(cType, (ViewItemContainer) viewItem);
            }
            else if (viewItem instanceof ModelViewItem)
            {
                ModelItem modelItem = ((ModelViewItem) viewItem).getDefinition();
                if (_filterModelItemForCriteria(modelItem))
                {
                    String modelItemPath = modelItem.getPath();
                    
                    Operator operator = org.ametys.runtime.model.type.ModelItemTypeConstants.STRING_TYPE_ID.equals(modelItem.getType().getId()) || ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID.equals(modelItem.getType().getId()) ? Operator.SEARCH : Operator.EQ;
                    addAttributeCriterionComponent(cType, modelItemPath, operator);
                    
                    if (Content.ATTRIBUTE_TITLE.equals(modelItem.getName()))
                    {
                        addLikeTitleCriterionComponent(cType, modelItemPath);
                    }
                }
            }
        }
    }
    
    /** Add criteria component to exclude the candidates
     * @param cType the simple content type
     * @throws ConfigurationException if an error occurs while creating the criteria configuration
     */
    protected void addExcludeCandidateSystemCriterionComponent(ContentType cType) throws ConfigurationException
    {
        DefaultConfiguration conf = (DefaultConfiguration) _searchCriterionHelper.getSystemCriteriaConfiguration(this, Optional.empty(), Set.of(cType.getId()), "mixins", Optional.empty());
        
        DefaultConfiguration defaultValueConf = new DefaultConfiguration("default-value");
        defaultValueConf.setValue("org.ametys.cms.referencetable.mixin.Candidate");
        conf.addChild(defaultValueConf);
        
        DefaultConfiguration opConf = new DefaultConfiguration("test-operator");
        opConf.setValue(Operator.NE.getName());
        conf.addChild(opConf);
        
        DefaultConfiguration widgetConf = new DefaultConfiguration("widget");
        widgetConf.setValue("edition.hidden");
        conf.addChild(widgetConf);
        
        _searchCriteriaManager.addComponent("cms", null, "mixins", SystemSearchUICriterion.class, conf);
        _searchCriteriaRoles.add("mixins");
    }
    
    /**
     * Returns <code>true</code> if the model item can be used as criteria
     * @param modelItem the model item
     * @return <code>true</code> if the model item can be used as criteria
     */
    @SuppressWarnings("static-access")
    protected boolean _filterModelItemForCriteria(ModelItem modelItem)
    {
        String typeId = modelItem.getType().getId();
        switch (typeId)
        {
            case ModelItemTypeConstants.STRING_TYPE_ID:
            case ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID:
            case ModelItemTypeConstants.DATE_TYPE_ID:
            case ModelItemTypeConstants.DATETIME_TYPE_ID:
            case ModelItemTypeConstants.LONG_TYPE_ID:
            case ModelItemTypeConstants.DOUBLE_TYPE_ID:
            case ModelItemTypeConstants.BOOLEAN_TYPE_ID:
            case ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID:
            case ModelItemTypeConstants.REFERENCE_ELEMENT_TYPE_ID:
                return true;
            case ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID:
            case ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID:
            case ModelItemTypeConstants.FILE_ELEMENT_TYPE_ID:
            case ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID:
            case ModelItemTypeConstants.COMPOSITE_TYPE_ID:
            case ModelItemTypeConstants.REPEATER_TYPE_ID:
            default:
                return false;
        }
    }
    
    /**
     * Add the title attribute criterion component to the manager, with a 'LIKE' operator and an hidden widget
     * @param contentType the simple content type.
     * @param attributePath the attribute path.
     * @throws ConfigurationException if a configuration error occurs.
     * @throws ComponentException if a component cannot be initialized.
     */
    protected void addLikeTitleCriterionComponent(ContentType contentType, String attributePath) throws ConfigurationException, ComponentException
    {
        DefaultConfiguration originalConf = new DefaultConfiguration("criteria");
        DefaultConfiguration widgetConf = new DefaultConfiguration("widget");
        widgetConf.setValue("edition.hidden");
        originalConf.addChild(widgetConf);
        Configuration conf = _searchCriterionHelper.getIndexingFieldCriteriaConfiguration(this, Optional.of(originalConf), Set.of(contentType.getId()), attributePath, Optional.of(Operator.LIKE), Optional.empty());
        
        String role = attributePath + "1";
        _searchCriteriaManager.addComponent("cms", null, role, IndexingFieldSearchUICriterion.class, conf);
        _searchCriteriaRoles.add(role);
    }
    
    /**
     * Add an attribute criterion component to the manager. 
     * @param contentType the simple content type.
     * @param attributePath the attribute path.
     * @param operator the criterion operator.
     * @throws ConfigurationException if a configuration error occurs.
     * @throws ComponentException if a component cannot be initialized.
     */
    protected void addAttributeCriterionComponent(ContentType contentType, String attributePath, Operator operator) throws ConfigurationException, ComponentException
    {
        DefaultConfiguration conf = (DefaultConfiguration) _searchCriterionHelper.getIndexingFieldCriteriaConfiguration(this, Optional.empty(), Set.of(contentType.getId()), attributePath, Optional.ofNullable(operator), Optional.empty());
        ModelItem metadataDefinition = contentType.getModelItem(attributePath);
        
        if (
                ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID.equals(metadataDefinition.getType().getId()) 
                && contentType.isReferenceTable() 
                && contentType.getParentAttributeDefinition()
                    .map(ModelItem::getPath)
                    .map(parent -> parent.equals(attributePath))
                    .orElse(false)
            )
        {
            DefaultConfiguration widgetConfig = (DefaultConfiguration) conf.getChild("widget");
            widgetConfig.setValue("edition.select-referencetable-content");
            
            DefaultConfiguration widgetParamsConfig = (DefaultConfiguration) conf.getChild("widget-params");
            
            DefaultConfiguration allowAutopostingParamsConfig = (DefaultConfiguration) widgetParamsConfig.getChild("param");
            allowAutopostingParamsConfig.setAttribute("name", "allowToggleAutoposting");
            allowAutopostingParamsConfig.setValue(true);
            widgetParamsConfig.addChild(allowAutopostingParamsConfig);
            
            conf.addChild(widgetConfig);
            conf.addChild(widgetParamsConfig);
        }
        _searchCriteriaManager.addComponent("cms", null, attributePath, IndexingFieldSearchUICriterion.class, conf);
        _searchCriteriaRoles.add(attributePath);
    }
    
    /**
     * Add a system criterion component to the manager.
     * @param contentType the simple content type.
     * @param property the system property.
     * @throws ConfigurationException if a configuration error occurs.
     * @throws ComponentException if a component cannot be initialized.
     */
    protected void addSystemCriterionComponent(ContentType contentType, String property) throws ConfigurationException, ComponentException
    {
        DefaultConfiguration conf = (DefaultConfiguration) _searchCriterionHelper.getSystemCriteriaConfiguration(this, Optional.empty(), Set.of(contentType.getId()), property, Optional.empty());
        
        if (property.equals("contentLanguage"))
        {
            // FIXME Is this configuration should be provided by Language system property itself ?
            // FIXME For now the simple contents are only created for language 'fr'
            DefaultConfiguration widgetConf = new DefaultConfiguration("widget");
            widgetConf.setValue("edition.select-language");
            conf.addChild(widgetConf);
            
            DefaultConfiguration defaultConf = new DefaultConfiguration("default-value");
            defaultConf.setValue("CURRENT");
            conf.addChild(defaultConf);
            
            DefaultConfiguration validConf = new DefaultConfiguration("validation");
            DefaultConfiguration mandatoryConf = new DefaultConfiguration("mandatory");
            mandatoryConf.setValue(true);
            validConf.addChild(mandatoryConf);
            conf.addChild(validConf);
        }
        
        _searchCriteriaManager.addComponent("cms", null, property, SystemSearchUICriterion.class, conf);
        _searchCriteriaRoles.add(property);
    }
    
    /**
     * Lookup all the criteria.
     * @param cType the simple content type.
     * @return the search criteria list.
     * @throws ComponentException if a component cannot be looked up.
     */
    protected List<SearchUICriterion> getSearchUICriteria(ContentType cType) throws ComponentException
    {
        List<SearchUICriterion> criteria = new ArrayList<>();
        
        for (String role : _searchCriteriaRoles)
        {
            SearchUICriterion criterion = _searchCriteriaManager.lookup(role);
            criteria.add(criterion);
        }
        
        return criteria;
    }
    
    /**
     * Retrieves the result items for the given content type
     * @param contentType The content type
     * @return the result items
     */
    protected ViewItemContainer _getResultItems(ContentType contentType)
    {
        View view = Optional.ofNullable(contentType.getView("columns"))
                            .orElse(contentType.getView("main"));
        
        View resultItems = new View();
        resultItems.addViewItems(_copyAndFilterViewItemsForColumns(view.getViewItems()));
        
        resultItems.addViewItem(SearchUIColumnHelper.createModelItemColumn(_systemPropertyExtensionPoint.getExtension("contributor")));
        resultItems.addViewItem(SearchUIColumnHelper.createModelItemColumn(_systemPropertyExtensionPoint.getExtension("lastModified")));
        
        if (!contentType.isMultilingual())
        {
            resultItems.addViewItem(SearchUIColumnHelper.createModelItemColumn(_systemPropertyExtensionPoint.getExtension("contentLanguage")));
        }
        
        return resultItems;
    }
    
    /**
     * Copy the given view items and filter to keep only items that can be used in {@link SearchUIColumn}s.
     * Also copy the children of view item accessors
     * @param viewItems the view items to copy
     * @return the view items copies
     */
    protected List<ViewItem> _copyAndFilterViewItemsForColumns(List<ViewItem> viewItems)
    {
        List<ViewItem> copies = new ArrayList<>();
        
        for (ViewItem viewItem : viewItems)
        {
            if (!(viewItem instanceof ViewElement) || _filterModelItemForColumn(((ViewElement) viewItem).getDefinition()))
            {
                ViewItem copy = viewItem.createInstance();
                if (viewItem instanceof ViewItemAccessor viewItemAccessor && !viewItemAccessor.getViewItems().isEmpty())
                {
                    assert copy instanceof ViewItemAccessor;
                    ((ViewItemAccessor) copy).addViewItems(_copyAndFilterViewItemsForColumns(viewItemAccessor.getViewItems()));
                }
                else if (viewItem instanceof ModelViewItem modelViewItem)
                {
                    // If the view item is a leaf, create a column
                    ModelItem modelItem = modelViewItem.getDefinition();
                    copy = SearchUIColumnHelper.createModelItemColumn(modelItem);
                }
                
                viewItem.copyTo(copy);
                copies.add(copy);
            }
        }
        
        return copies;
    }
    
    /**
     * Returns <code>true</code> if model item can be used as column search UI
     * @param modelItem the model item
     * @return <code>true</code> if model item can be used as column search UI
     */
    @SuppressWarnings("static-access")
    protected boolean _filterModelItemForColumn(ModelItem modelItem)
    {
        String typeId = modelItem.getType().getId();
        switch (typeId)
        {
            case ModelItemTypeConstants.STRING_TYPE_ID:
            case ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID:
            case ModelItemTypeConstants.DATE_TYPE_ID:
            case ModelItemTypeConstants.DATETIME_TYPE_ID:
            case ModelItemTypeConstants.LONG_TYPE_ID:
            case ModelItemTypeConstants.DOUBLE_TYPE_ID:
            case ModelItemTypeConstants.BOOLEAN_TYPE_ID:
            case ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID:
            case ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID:
            case ModelItemTypeConstants.REFERENCE_ELEMENT_TYPE_ID:
                return true;
            case ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID:
            case ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID:
            case ModelItemTypeConstants.FILE_ELEMENT_TYPE_ID:
            default:
                return false;
        }
    }
}
