/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.ui.model;

import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.ui.model.impl.RepeaterSearchUIColumn;
import org.ametys.cms.search.ui.model.impl.ViewElementAccessorSearchUIColumn;
import org.ametys.cms.search.ui.model.impl.ViewElementSearchUIColumn;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemAccessor;
import org.ametys.runtime.model.ModelItemContainer;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;

/**
 * Helper class for search UI columns
 */
public final class SearchUIColumnHelper
{
    /** The default column width */
    public static final int DEFAULT_COLUMN_WIDTH = 200;
    /** The default column width for system properties */
    public static final int DEFAULT_SYSTEM_PROPERTY_COLUMN_WIDTH = 150;
    
    private SearchUIColumnHelper()
    {
        // Empty constructor
    }
    
    /**
     * Creates a column for the given model item
     * @param modelItem the model item
     * @return the created column
     * @throws IllegalArgumentException if the given model item is not an element or a repeater
     */
    public static SearchUIColumn createModelItemColumn(ModelItem modelItem) throws IllegalArgumentException
    {
        SearchUIColumn column;
        if (modelItem instanceof RepeaterDefinition repeaterDefinition)
        {
            column = new RepeaterSearchUIColumn();
            ((RepeaterSearchUIColumn) column).setDefinition(repeaterDefinition);
        }
        else if (modelItem instanceof ElementDefinition elementDefinition)
        {
            if (modelItem instanceof ModelItemAccessor)
            {
                column = new ViewElementAccessorSearchUIColumn();
                ((ViewElementAccessorSearchUIColumn) column).setDefinition(elementDefinition);
            }
            else
            {
                column = new ViewElementSearchUIColumn();
                ((ViewElementSearchUIColumn) column).setDefinition(elementDefinition);
            }
        }
        else
        {
            throw new IllegalArgumentException("Unable to create a column from the given model item '" + modelItem.getPath() + "'. This model item is not a repeater or an element.");
        }
        
        return column;
    }
    
    /**
     * Retrieves the default column width, corresponding to the referenced model item
     * @param column the column
     * @return the default width
     */
    public static int getDefaultColumnWidth(SearchUIColumn column)
    {
        if (column.getDefinition() instanceof SystemProperty systemProperty)
        {
            return Optional.ofNullable(systemProperty.getColumnWidth())
                    .orElse(DEFAULT_SYSTEM_PROPERTY_COLUMN_WIDTH);
        }
        else
        {
            return DEFAULT_COLUMN_WIDTH;
        }
    }
    
    /**
     * Determines if the inline edition is allowed for the given column
     * @param column the column
     * @return <code>true</code> if the column edition is allowed, <code>false</code> otherwise
     */
    public static boolean isEditionAllowed(SearchUIColumn column)
    {
        if (_isMultiLevelMultiple(column))
        {
            // column is not editable if it references a model item with a multiple parent  
            return false;
        }
        
        ModelItem modelItem = column.getDefinition();
        if (modelItem instanceof ElementDefinition definition && !definition.isEditable())
        {
            return false;
        }
        
        if (_isJoinedModelItem(column))
        {
            // column is not editable if it is on a distant content
            return false;
        }
        
        if (ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID.equals(modelItem.getType().getId()))
        {
            // richtext are never editable inline
            return false;
        }
        
        if (org.ametys.plugins.repository.data.type.ModelItemTypeConstants.REPEATER_TYPE_ID.equals(modelItem.getType().getId()))
        {
            // disallow edition for repeaters containing richtexts
            return !ModelHelper.hasModelItemOfType((ModelItemContainer) modelItem, ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID);
        }
        
        return true;
    }
    
    /**
     * Check if the column references a model item of a distant content type
     * @param column the column
     * @return <code>true</code> if the column references a model item of a distant content type, <code>false</code> otherwise
     */
    private static boolean _isJoinedModelItem(SearchUIColumn column)
    {
        ViewItemAccessor parent = column.getParent();
        while (parent != null)
        {
            if (parent instanceof ModelViewItem parentModelViewItem && ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID.equals(parentModelViewItem.getDefinition().getType().getId()))
            {
                // One of the parent level is a join (link on a distant content)
                return true;
            }
            
            parent = parent instanceof ViewItem parentViewItem ? parentViewItem.getParent() : null;
        }
        
        // No parent is a join
        return false;
    }
    

    
    /**
     * Determines if the sort is allowed for the given column
     * @param column the column
     * @return <code>true</code> if the column edition is allowed, <code>false</code> otherwise
     */
    public static boolean isSortAllowed(SearchUIColumn column)
    {
        return (column.allowSortOnMultipleJoin() || !_isMultiLevelMultiple(column))
                && _isModelItemSortable(column.getDefinition());
    }
    
    /**
     * Check if the column references a model item with a multiple parent
     * @param column the column
     * @return <code>true</code> if the column references a model item with a multiple parent, <code>false</code> otherwise
     */
    private static boolean _isMultiLevelMultiple(SearchUIColumn column)
    {
        ViewItemAccessor parent = column.getParent();
        while (parent != null && !(parent instanceof SearchUIColumn))
        {
            if (parent instanceof ModelViewItem parentModelViewItem && _isModelItemMultiple(parentModelViewItem.getDefinition()))
            {
                // One of the parent level is multiple
                return true;
            }
            
            parent = parent instanceof ViewItem parentViewItem ? parentViewItem.getParent() : null;
        }
        
        // No parent is multiple
        return false;
    }
    
    /**
     * Check if the given model item is multiple
     * @param modelItem the model item
     * @return <code>true</code> if the model item is multiple, <code>false</code> otherwise
     */
    private static boolean _isModelItemMultiple(ModelItem modelItem)
    {
        return modelItem instanceof ElementDefinition && ((ElementDefinition) modelItem).isMultiple()
                || modelItem instanceof RepeaterDefinition;
    }
    
    /**
     * Determines if the column is sortable according its model item
     * @param modelItem the model item
     * @return <code>true</code> if model item is sortable, <code>false</code> otherwise
     */
    @SuppressWarnings("static-access")
    private static boolean _isModelItemSortable(ModelItem modelItem)
    {
        if (modelItem instanceof SystemProperty systemProperty)
        {
            return systemProperty.isSortable();
        }
        else
        {
            switch (modelItem.getType().getId())
            { 
                case ModelItemTypeConstants.STRING_TYPE_ID:
                case ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID:
                case ModelItemTypeConstants.LONG_TYPE_ID:
                case ModelItemTypeConstants.DATE_TYPE_ID:
                case ModelItemTypeConstants.DATETIME_TYPE_ID:
                case ModelItemTypeConstants.BOOLEAN_TYPE_ID:
                case ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID:
                case ModelItemTypeConstants.DOUBLE_TYPE_ID:
                case ModelItemTypeConstants.USER_ELEMENT_TYPE_ID:
                case ModelItemTypeConstants.REFERENCE_ELEMENT_TYPE_ID:
                    return true;
                case ModelItemTypeConstants.REPEATER_TYPE_ID:
                case ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID:
                case ModelItemTypeConstants.FILE_ELEMENT_TYPE_ID:
                case ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID:
                default:
                    return false;
            }
        }
    }
    
    /**
     * Retrieves the default converter for the given definition
     * @param definition the definition
     * @return the default converter
     */
    public static Optional<String> getElementDefinitionDefaultConverter(ElementDefinition definition)
    {
        String defaultConverter = null;
        if (definition instanceof SystemProperty systemProperty)
        {
            defaultConverter = systemProperty.getConverter();
        }
        else if (ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID.equals(definition.getType().getId()))
        {
            defaultConverter = "Ametys.plugins.cms.search.SearchGridHelper.convertContent";
        }
        
        return Optional.ofNullable(defaultConverter)
                       .filter(StringUtils::isNotEmpty);
    }
    
    /**
     * Converts the given column's properties in a JSON map
     * @param column the column
     * @return The column's properties as a JSON map
     */
    public static Map<String, Object> columnPropertiesToJSON(SearchUIColumn column)
    {
        Map<String, Object> json = new HashMap<>();
        
        json.put("path", _getRelativePathToFirstParentColumn(column));
        json.put("width", column.getWidth());
        json.put("hidden", column.isHidden());
        json.put("renderer", column.getRenderer());
        json.put("converter", column.getConverter());
        json.put("editable", column.isEditable());
        json.put("sortable", column.isSortable());
        json.put("defaultSorter", column.getDefaultSorter());
        
        json.put("multiple", _isModelItemMultiple(column.getDefinition()) || _isMultiLevelMultiple(column));
        
        return json;
    }
    
    private static String _getRelativePathToFirstParentColumn(SearchUIColumn column)
    {
        String path = column.getName();
        ViewItemAccessor parent = column.getParent();
        while (parent != null && !(parent instanceof SearchUIColumn))
        {
            if (parent instanceof ModelViewItem parentModelViewItem)
            {
                path = parentModelViewItem.getName() + ModelItem.ITEM_PATH_SEPARATOR + path;
            }
            
            parent = parent instanceof ViewItem parentViewItem ? parentViewItem.getParent() : null;
        }
        
        return path;
    }
}
