/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.ui.model;

import java.util.Map;

import org.ametys.cms.search.model.SearchModel;
import org.ametys.runtime.model.ElementDefinition;

/**
 * Search tool model.
 */
public interface SearchUIModel extends SearchModel
{
    @Override
    Map<String, SearchUICriterion> getCriteria(Map<String, Object> contextualParameters);
    
    @Override
    default SearchUICriterion getCriterion(String id, Map<String, Object> contextualParameters)
    {
        return getCriteria(contextualParameters).get(id);
    }
    
    @Override
    Map<String, SearchUICriterion> getFacetedCriteria(Map<String, Object> contextualParameters);
    
    @Override
    default SearchUICriterion getFacetedCriterion(String id, Map<String, Object> contextualParameters)
    {
        return getFacetedCriteria(contextualParameters).get(id);
    }
    
    /**
     * Get the list of search criteria in advanced mode
     * @param contextualParameters the contextual parameters
     * @return the list of search criteria in advanced mode
     */
    Map<String, SearchUICriterion> getAdvancedCriteria(Map<String, Object> contextualParameters);
    
    /**
     * Get an advanced search criterion by its id.
     * @param id The advanced search criterion id.
     * @param contextualParameters the contextual parameters
     * @return the advanced criterion or <code>null</code> if not found
     */
    default SearchUICriterion getAdvancedCriterion(String id, Map<String, Object> contextualParameters)
    {
        return getAdvancedCriteria(contextualParameters).get(id);
    }
    
    /**
     * Get the page size.
     * @param contextualParameters the contextual parameters.
     * @return The page size, unlimited or default used when negative or 0.
     */
    int getPageSize(Map<String, Object> contextualParameters);
    
    /**
     * Get the URL for search
     * @param contextualParameters the contextual parameters
     * @return the URL for search
     */
    String getSearchUrl(Map<String, Object> contextualParameters);
    
    /**
     * Get the plugin name for search
     * @param contextualParameters the contextual parameters
     * @return the plugin name for search
     */
    String getSearchUrlPlugin(Map<String, Object> contextualParameters);
    
    /**
     * Get the URL for CSV export of results
     * @param contextualParameters the contextual parameters
     * @return the URL for CSV export
     */
    String getExportCSVUrl(Map<String, Object> contextualParameters);
    
    /**
     * Get the plugin name for CSV export of results
     * @param contextualParameters the contextual parameters
     * @return the plugin name for CSV export
     */
    String getExportCSVUrlPlugin(Map<String, Object> contextualParameters);
    
    /**
     * Get the URL for DOC export of results
     * @param contextualParameters the contextual parameters
     * @return the URL for DOC export
     */
    String getExportDOCUrl(Map<String, Object> contextualParameters);

    /**
     * Get the plugin name for DOC export of results
     * @param contextualParameters the contextual parameters
     * @return the plugin name for DOC export
     */
    String getExportDOCUrlPlugin(Map<String, Object> contextualParameters);

    /**
     * Get the URL for XML export of results
     * @param contextualParameters the contextual parameters
     * @return the URL for XML export
     */
    String getExportXMLUrl(Map<String, Object> contextualParameters);
    
    /**
     * Get the plugin name for XML export of results
     * @param contextualParameters the contextual parameters
     * @return the plugin name for XML export
     */
    String getExportXMLUrlPlugin(Map<String, Object> contextualParameters);
    
    /**
     * Get the URL for PDF export of results
     * @param contextualParameters the contextual parameters
     * @return the URL for PDF export
     */
    String getExportPDFUrl(Map<String, Object> contextualParameters);
    
    /**
     * Get the plugin name for PDF export of results
     * @param contextualParameters the contextual parameters
     * @return the plugin name for PDF export
     */
    String getExportPDFUrlPlugin(Map<String, Object> contextualParameters);
    
    /**
     * Get the URL for print results
     * @param contextualParameters the contextual parameters
     * @return the URL for print results
     */
    String getPrintUrl(Map<String, Object> contextualParameters);
    
    /**
     * Get the plugin name for print results
     * @param contextualParameters the contextual parameters    
     * @return the plugin name for print results
     */
    String getPrintUrlPlugin(Map<String, Object> contextualParameters);
    
    /**
     * Get the name of the view to use for summary of the content.
     * @return the name of the view to use for summary of the content. Can be null.
     */
    String getSummaryView();
    
    /**
     * Indicates if sorting on join columns which contain at least one {@link ElementDefinition#isMultiple() multiple} {@link ElementDefinition definition} in its path (intermediate ones only) must be allowed.
     * <br>If the final metadata of the path is multiple, the column will not be sortable though.
     * @return true if sorting on join columns which contains at least one multiple metadata in its path (intermediate ones only) must be allowed.
     */
    default boolean allowSortOnMultipleJoin()
    {
        return false;
    }
}
