/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.ui.model;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.HashSet;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.configuration.DefaultConfigurationBuilder;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.Constants;
import org.apache.cocoon.environment.Context;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.runtime.plugin.Feature;
import org.ametys.runtime.plugin.Plugin;
import org.ametys.runtime.plugin.component.AbstractThreadSafeComponentExtensionPoint;

/**
 * Extension point for {@link SearchUIModel}s.
 */
public class SearchUIModelExtensionPoint extends AbstractThreadSafeComponentExtensionPoint<SearchUIModel>
{
    /** The component role */
    public static final String ROLE = SearchUIModelExtensionPoint.class.getName();
    
    private ContentTypeExtensionPoint _cTypeEP;
    
    private Set<String> _filesLoaded;
    
    /**
     * Create the instance
     */
    public SearchUIModelExtensionPoint()
    {
        super();
        _filesLoaded = new HashSet<>();
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _cTypeEP = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    @Override
    public void addExtension(String id, String pluginName, String featureName, Configuration configuration) throws ConfigurationException
    {
        if (_filesLoaded.contains(id))
        {
            getLogger().warn("The search model extension '" + id + "' from feature '" + pluginName + "/" + featureName + "' will be ignored. It has already been loaded from the file in WEB-INF/param/search/" + id.substring("search-ui.".length()) + ".xml");
        }
        else
        {
            super.addExtension(id, pluginName, featureName, configuration);
        }
    }
    
    @Override
    public void initialize() throws Exception
    {
        super.initialize();
        
        try
        {
            Context ctx = (Context) _context.get(Constants.CONTEXT_ENVIRONMENT_CONTEXT);
            
            File root = new File(ctx.getRealPath("/WEB-INF/param/search"));
            
            if (root.exists())
            {
                File[] files = root.listFiles(file -> !file.isDirectory() && file.getName().endsWith(".xml"));
                
                for (File modelFile : files)
                {
                    String fileName = modelFile.getName();
                    String id = "search-ui." + fileName.substring(0, fileName.lastIndexOf('.'));
                    _filesLoaded.add(id);
                    
                    if (super.hasExtension(id))
                    {
                        throw new IllegalArgumentException("The search tool model of id " + id + " at " + modelFile.getAbsolutePath() + " is already declared.");
                    }
                    
                    try (InputStream is = new FileInputStream(modelFile))
                    {
                        
                        DefaultConfiguration conf = new DefaultConfiguration("extension");
                        
                        Configuration c = new DefaultConfigurationBuilder(true).build(is, modelFile.toURI().toString());
                        
                        conf.setAttribute("id", id);
                        conf.addChild(c);
                        
                        addComponent(Plugin.DEFAULT_NAME, Feature.DEFAULT_NAME, id, StaticSearchUIModel.class, conf);
                    }
                    catch (Exception e)
                    {
                        getLogger().error("Error initializing search model '" + id + "' from file '" + modelFile.getAbsolutePath() + "'.", e);
                    }
                }
            }
            
            // TODO Provide a component to generate SearchModel corresponding to contents on the fly?
            Set<String> cTypeIds = _cTypeEP.getExtensionsIds();
            for (String cTypeId : cTypeIds)
            {
                ContentType cType = _cTypeEP.getExtension(cTypeId);
                if (cType != null && cType.isReferenceTable())
                {
                    String id = "reference-table-search-ui." + cTypeId;
                    DefaultConfiguration conf = new DefaultConfiguration("extension");
                    conf.setAttribute("id", id);
                    
                    DefaultConfiguration cTypeConf = new DefaultConfiguration("contentType");
                    cTypeConf.setValue(cTypeId);
                    conf.addChild(cTypeConf);
                    
                    addComponent(Plugin.DEFAULT_NAME, Feature.DEFAULT_NAME, id, ReferenceTableSearchUIModel.class, conf);
                }
            }
        }
        catch (Exception e)
        {
            getLogger().error("Unable to initialize search tool models", e);
        }
    }
}
