/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.ui.model;

import java.util.Collection;
import java.util.Optional;

import org.apache.avalon.framework.configuration.ConfigurationException;

import org.ametys.cms.contenttype.AbstractContentTypeViewParser;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.ModelHelper.ConfigurationAndPluginName;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemAccessor;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItem;

/**
 * Component that parses the columns of a {@link StaticSearchUIModel}
 */
public class StaticSearchUIModelColumnsParser extends AbstractContentTypeViewParser
{
    private static final String __VIEW_NAME = "StaticSearchUIModel-Columns";
    
    /** Content types containing the item available in the search model's columns */
    protected Collection<ContentType> _contentTypes;

    /**
     * Creates a {@link StaticSearchUIModel}'s columns parser
     * @param contentTypes The content types containing the item available in the search model's columns
     */
    public StaticSearchUIModelColumnsParser(Collection<ContentType> contentTypes)
    {
        _contentTypes = contentTypes;
    }
    
    @Override
    protected String _parseViewName(ConfigurationAndPluginName viewConfiguration) throws ConfigurationException
    {
        return __VIEW_NAME;
    }

    @Override
    public View overrideView(ConfigurationAndPluginName viewConfiguration, View existingView) throws ConfigurationException
    {
        throw new UnsupportedOperationException("Unable to override columns of a search model");
    }
    
    @Override
    protected void _fillViewGeneralInformation(ConfigurationAndPluginName viewConfiguration, View view, View existingView, Collection< ? extends Model> model) throws ConfigurationException
    {
        // Do nothing - there is no general information in search model's columns
    }
    
    @Override
    protected Collection<? extends Model> _getModel()
    {
        return _contentTypes;
    }
    
    @Override
    protected ModelItem _getModelItem(ConfigurationAndPluginName itemConfiguration, String modelItemName, Collection< ? extends ModelItemAccessor> parents) throws ConfigurationException
    {
        ModelItem modelItem = parents.isEmpty() && Content.ATTRIBUTE_TITLE.equals(modelItemName)
                ? _contentTypesHelper.getTitleAttributeDefinition()
                : super._getModelItem(itemConfiguration, modelItemName, parents);
        
        if (modelItem instanceof SystemProperty systemProperty && !systemProperty.isDisplayable())
        {
            throw new ConfigurationException("The property '" + systemProperty.getName() + "' is not displayable.");
        }
        
        return modelItem;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    protected ModelViewItem createModelViewItem(ConfigurationAndPluginName itemConfiguration, ModelItem modelItem, View referenceView, boolean override) throws ConfigurationException
    {
        ModelViewItem<ModelItem> modelViewItem = super.createModelViewItem(itemConfiguration, modelItem, referenceView, override);
        
        if (modelViewItem instanceof SearchUIColumn searchColumn)
        {
            searchColumn.setWidth(itemConfiguration.configuration().getChild("width").getValueAsInteger(-1));
            searchColumn.setHidden(itemConfiguration.configuration().getChild("hidden").getValueAsBoolean(false));
            
            searchColumn.setEditable(itemConfiguration.configuration().getChild("editable").getValueAsBoolean(true));
            searchColumn.setSortable(itemConfiguration.configuration().getChild("sortable").getValueAsBoolean(true));
            searchColumn.setAllowSortOnMultipleJoin(itemConfiguration.configuration().getChild("allow-sort-on-multiple-join").getValueAsBoolean(false));
            searchColumn.setDefaultSorter(Optional.ofNullable(itemConfiguration.configuration().getChild("default-sorter").getValue(null)));
            
            searchColumn.setRenderer(Optional.ofNullable(itemConfiguration.configuration().getChild("renderer").getValue(null)));
            searchColumn.setConverter(Optional.ofNullable(itemConfiguration.configuration().getChild("converter").getValue(null)));
            
            return searchColumn;
        }
        else
        {
            return modelViewItem;
        }
    }
    
    @Override
    protected ModelViewItem _createModelViewItemInstance(ModelItem modelItem)
    {
        return SearchUIColumnHelper.createModelItemColumn(modelItem);
    }
    
    @Override
    protected ViewItem _createViewItemCopyInstanceForReferencedView(ViewItem viewItem)
    {
        return Optional.of(viewItem)
                       .filter(ModelViewItem.class::isInstance)
                       .map(ModelViewItem.class::cast)
                       .map(ModelViewItem::getDefinition)
                       // Keep only items that can be represented in columns
                       .filter(modelViewItem -> modelViewItem instanceof RepeaterDefinition || modelViewItem instanceof ElementDefinition)
                       .map(this::_createModelViewItemInstance)
                       .map(ViewItem.class::cast)
                       .orElseGet(() -> super._createViewItemCopyInstanceForReferencedView(viewItem));
    }
}
