/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.ui.model.impl;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.MetadataType;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.model.ContentElementDefinition;
import org.ametys.cms.model.properties.ElementRefProperty;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.ui.model.SearchUICriterion;
import org.ametys.cms.search.ui.model.SearchUIModel;
import org.ametys.plugins.core.ui.util.ConfigurationHelper;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Enumerator;
import org.ametys.runtime.model.StaticEnumerator;
import org.ametys.runtime.parameter.DefaultValidator;
import org.ametys.runtime.parameter.Parameter;
import org.ametys.runtime.parameter.Validator;
import org.ametys.runtime.plugin.component.LogEnabled;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * This class represents a search criterion of a {@link SearchUIModel}.
 */
public abstract class AbstractSearchUICriterion extends Parameter<MetadataType> implements SearchUICriterion, Contextualizable, Serviceable, Configurable, Disposable, LogEnabled
{
    /** The content type extension point. */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    
    /** The content type extension point. */
    protected ContentHelper _contentHelper;
    
    /** ComponentManager for {@link Enumerator}s. */
    protected ThreadSafeComponentManager<Enumerator> _enumeratorManager;
    
    /** The service manager */
    protected ServiceManager _manager;
    
    /** the logger */
    protected Logger _logger;
    
    /** The avalon context */
    protected Context _context;

    private String _onInitClassName;
    private String _onSubmitClassName;
    private String _onChangeClassName;
    private boolean _hidden;
    private boolean _multiple;
    private I18nizableText _group;
    
    private String _contentTypeId;
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _contentHelper = (ContentHelper) manager.lookup(ContentHelper.ROLE);
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        configureId(configuration);
        
        configureLabelsAndGroup(configuration);
        configureUIProperties(configuration);
        configureValues(configuration);
        
        _enumeratorManager = new ThreadSafeComponentManager<>();
        _enumeratorManager.setLogger(_logger);
        _enumeratorManager.contextualize(_context);
        _enumeratorManager.service(_manager);
    }
    
    public void dispose()
    {
        _enumeratorManager.dispose();
        _enumeratorManager = null;
    }
    
    @Override
    public void setLogger(Logger logger)
    {
        _logger = logger;
    }
    
    /**
     * Configure the criterion ID.
     * @param configuration The search criterion configuration.
     * @throws ConfigurationException If an error occurs.
     */
    protected void configureId(Configuration configuration) throws ConfigurationException
    {
        setId(configuration.getAttribute("id", null));
    }
    
    /**
     * Configure the labels and group.
     * @param configuration The search criterion configuration.
     * @throws ConfigurationException If an error occurs.
     */
    protected void configureLabelsAndGroup(Configuration configuration) throws ConfigurationException
    {
        setGroup(_configureI18nizableText(configuration.getChild("group", false), null));
        setLabel(_configureI18nizableText(configuration.getChild("label", false), new I18nizableText("")));
        setDescription(_configureI18nizableText(configuration.getChild("description", false), new I18nizableText("")));
    }
    
    /**
     * Configure the default value.
     * @param configuration The search criterion configuration.
     * @throws ConfigurationException If an error occurs.
     */
    protected void configureValues(Configuration configuration) throws ConfigurationException
    {
        // The default value can be empty.
        Configuration[] defaultValueConfs = configuration.getChildren("default-value");
        if (defaultValueConfs.length == 1)
        {
            setDefaultValue(ConfigurationHelper.parseObject(defaultValueConfs[0], ""));
        }
        else if (defaultValueConfs.length > 1)
        {
            // Make the default value a list
            List<Object> collection = Arrays.asList(defaultValueConfs).stream()
                    .map(conf -> ConfigurationHelper.parseObject(conf, ""))
                    .collect(Collectors.toList());
            setDefaultValue(collection);
        }
    }
    
    /**
     * Configure the standard UI properties (hidden, init class, change class, submit class).
     * @param configuration The search criterion configuration.
     * @throws ConfigurationException If an error occurs.
     */
    protected void configureUIProperties(Configuration configuration) throws ConfigurationException
    {
        setHidden(configuration.getAttributeAsBoolean("hidden", false));
        setInitClassName(configuration.getChild("oninit").getValue(null));
        setChangeClassName(configuration.getChild("onchange").getValue(null));
        setSubmitClassName(configuration.getChild("onsubmit").getValue(null));
    }
    
    /**
     * Configure enumerator of the criterion
     * @param configuration The search criterion configuration.
     * @param definition The definition of the criterion
     * @return the enumerator
     * @throws ConfigurationException if an error occurs
     */
    protected org.ametys.runtime.parameter.Enumerator configureEnumerator(Configuration configuration, ElementDefinition definition) throws ConfigurationException
    {
        return configureEnumerator(configuration, definition, configuration);
    }
    
    /**
     * Configure enumerator of the criterion
     * @param configuration The search criterion configuration.
     * @param definition The definition of the criterion
     * @param defaultEnumeratorConfig The configuration for property's widget params
     * @return the enumerator
     * @throws ConfigurationException if an error occurs
     */
    @SuppressWarnings("unchecked")
    protected org.ametys.runtime.parameter.Enumerator configureEnumerator(Configuration configuration, ElementDefinition definition, Configuration defaultEnumeratorConfig) throws ConfigurationException
    {
        Enumerator enumerator = definition.getCriterionEnumerator(defaultEnumeratorConfig, _enumeratorManager);
        org.ametys.runtime.parameter.Enumerator oldAPIEnumerator = Optional.ofNullable(enumerator)
                       .filter(org.ametys.runtime.parameter.Enumerator.class::isInstance)
                       .map(org.ametys.runtime.parameter.Enumerator.class::cast)
                       .orElse(null);
        
        if (enumerator instanceof StaticEnumerator)
        {
            oldAPIEnumerator = new org.ametys.runtime.parameter.StaticEnumerator();
            
            try
            {
                for (Map.Entry<? extends Object, I18nizableText> entry : ((Map<Object, I18nizableText>) (Object) enumerator.getTypedEntries()).entrySet())
                {
                    ((org.ametys.runtime.parameter.StaticEnumerator) oldAPIEnumerator).add(entry.getValue(), entry.getKey().toString());
                }
            }
            catch (Exception e)
            {
                throw new ConfigurationException("Unable to get entries of the static enumerator of criterion '" + definition.getPath() + "'", configuration, e);
            }
        }
        
        return oldAPIEnumerator;
    }
    
    /**
     * Configure widget of the criterion
     * @param configuration The search criterion configuration.
     * @param definition The definition of the criterion
     * @return the widget
     */
    protected String configureWidget(Configuration configuration, ElementDefinition definition)
    {
        String defaultWidget = definition.getCriterionWidget();
        return configureWidget(configuration, defaultWidget, MetadataType.fromModelItemType(definition.getType()));
    }
    
    /**
     * Configure widget of the criterion
     * @param configuration The search criterion configuration.
     * @param defaultWidget The default widget to use if not present in configuration. If {@link MetadataType#RICH_TEXT} or {@link MetadataType#MULTILINGUAL_STRING}, it will be forced.
     * @param type the type which must be supported by the widget
     * @return the widget
     */
    protected String configureWidget(Configuration configuration, String defaultWidget, MetadataType type)
    {
        String realDefaultWidget = defaultWidget;
        if ("edition.textarea".equals(realDefaultWidget))
        {
            realDefaultWidget = null;
        }
        else if (type == MetadataType.RICH_TEXT || type == MetadataType.MULTILINGUAL_STRING)
        {
            realDefaultWidget = "edition.textfield"; // Force simple text field
        }
        else if (realDefaultWidget == null && type == MetadataType.BOOLEAN)
        {
            realDefaultWidget = "edition.boolean-combobox";
        }
        
        return configuration.getChild("widget").getValue(realDefaultWidget);
    }
    
    /**
     * Configure widget parameters of the criterion
     * @param configuration The search criterion configuration.
     * @param defaultParams The default widget parameters to override with configuration
     * @param type the type which must be supported by the widget. If {@link MetadataType#CONTENT} or {@link MetadataType#SUB_CONTENT}, some parameters will be forced.
     * @param contentTypeId For {@link MetadataType#CONTENT} or {@link MetadataType#SUB_CONTENT} types only. The id of the content type.
     * @return the widget parameters
     * @throws ConfigurationException If an error occurs.
     */
    protected Map<String, I18nizableText> configureWidgetParameters(Configuration configuration, Map<String, I18nizableText> defaultParams, MetadataType type, String contentTypeId) throws ConfigurationException
    {
        Configuration widgetParamsConfig = configuration.getChild("widget-params");
        Map<String, I18nizableText> widgetParams = new HashMap<>(Optional.ofNullable(defaultParams).orElseGet(Collections::emptyMap));
        
        if (widgetParamsConfig != null)
        {
            // Overriden params
            Configuration[] params = widgetParamsConfig.getChildren();
            for (Configuration paramConfig : params)
            {
                String name = "param".equals(paramConfig.getName()) ? paramConfig.getAttribute("name") : paramConfig.getName();
                widgetParams.put(name, I18nizableText.parseI18nizableText(paramConfig, null));
            }
        }
        
        // Force some param for types CONTENT and SUB_CONTENT
        if (type == MetadataType.CONTENT || type == MetadataType.SUB_CONTENT)
        {
            setContentTypeId(contentTypeId);
            if (contentTypeId != null)
            {
                widgetParams.put("contentType", new I18nizableText(contentTypeId));
            }
            
            // Override the widget parameters to disable search and creation
            widgetParams.put("allowCreation", new I18nizableText("false"));
            widgetParams.put("allowSearch", new I18nizableText("false"));
        }

        return widgetParams;
    }
    
    /**
     * Configure widget parameters of the criterion
     * @param configuration The search criterion configuration.
     * @param definition The definition of the criterion's element
     * @return the widget parameters
     * @throws ConfigurationException If an error occurs.
     */
    protected Map<String, I18nizableText> configureWidgetParameters(Configuration configuration, ElementDefinition definition) throws ConfigurationException
    {
        return configureWidgetParameters(configuration, definition, configuration);
    }
    
    /**
     * Configure widget parameters of the criterion
     * @param configuration The search criterion configuration.
     * @param definition The definition of the criterion's element
     * @param defaultWidgetParamsConfig The configuration for property's widget params
     * @return the widget parameters
     * @throws ConfigurationException If an error occurs.
     */
    protected Map<String, I18nizableText> configureWidgetParameters(Configuration configuration, ElementDefinition definition, Configuration defaultWidgetParamsConfig) throws ConfigurationException
    {
        Map<String, I18nizableText> widgetParams = new HashMap<>();
        Map<String, I18nizableText> criterionWidgetParameters = definition.getCriterionWidgetParameters(defaultWidgetParamsConfig);
        if (criterionWidgetParameters != null)
        {
            widgetParams.putAll(criterionWidgetParameters);
        }
        
        // Force some params for type CONTENT
        if (ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID.equals(definition.getType().getId()))
        {
            Optional<ContentElementDefinition> contentElementDefinition = Optional.empty();
            if (definition instanceof ContentElementDefinition contentElementDef)
            {
                contentElementDefinition = Optional.of(contentElementDef);
            }
            else if (definition instanceof ElementRefProperty elementRefProperty)
            {
                contentElementDefinition = Optional.of(elementRefProperty.getElementDefinition())
                        .filter(ContentElementDefinition.class::isInstance)
                        .map(ContentElementDefinition.class::cast);
            }
            
            String contentTypeId = contentElementDefinition.map(ContentElementDefinition::getContentTypeId)
                    .filter(StringUtils::isNotEmpty)
                    .orElse(null);
            if (contentTypeId != null)
            {
                setContentTypeId(contentTypeId);
                widgetParams.put("contentType", new I18nizableText(contentTypeId));
            }
            
            // Override the widget parameters to disable search and creation
            widgetParams.put("allowCreation", new I18nizableText("false"));
            widgetParams.put("allowSearch", new I18nizableText("false"));
        }
        
        Configuration widgetParamsConfig = configuration.getChild("widget-params");
        if (widgetParamsConfig != null)
        {
            // Overridden params
            Configuration[] params = widgetParamsConfig.getChildren();
            for (Configuration paramConfig : params)
            {
                String name = "param".equals(paramConfig.getName()) ? paramConfig.getAttribute("name") : paramConfig.getName();
                widgetParams.put(name, I18nizableText.parseI18nizableText(paramConfig, null));
            }
        }
        
        return widgetParams;
    }
    
    /**
     * Get the JS class name to execute on 'init' event
     * @return the JS class name to execute on 'init' event
     */
    public String getInitClassName()
    {
        return _onInitClassName;
    }

    /**
     * Set the JS class name to execute on 'init' event
     * @param className the JS class name 
     */
    public void setInitClassName(String className)
    {
        this._onInitClassName = className;
    }

    /**
     * Get the JS class name to execute on 'submit' event
     * @return the JS class name to execute on 'submit' event
     */
    public String getSubmitClassName()
    {
        return _onSubmitClassName;
    }

    /**
     * Set the JS class name to execute on 'submit' event
     * @param className the JS class name 
     */
    public void setSubmitClassName(String className)
    {
        this._onSubmitClassName = className;
    }

    /**
     * Get the JS class name to execute on 'change' event
     * @return the JS class name to execute on 'change' event
     */
    public String getChangeClassName()
    {
        return _onChangeClassName;
    }

    /**
     * Set the JS class name to execute on 'change' event
     * @param className the JS class name 
     */
    public void setChangeClassName(String className)
    {
        this._onChangeClassName = className;
    }
    
    /**
     * Get the group of the search criteria
     * @return <code>null</code> if the search criteria does not belong to any group, the name of the group otherwise
     */
    public I18nizableText getGroup()
    {
        return _group;
    }
    
    /**
     * Set the group of the search criteria
     * @param group the group this search criteria will be added to
     */
    public void setGroup(I18nizableText group)
    {
        _group = group;
    }
    
    /**
     * Determines if the criteria is hidden
     * @return <code>true</code> if the criteria is hidden
     */
    public boolean isHidden()
    {
        return _hidden;
    }
    
    /**
     * Set the hidden property of the criteria
     * @param hidden true to hide the search criteria
     */
    public void setHidden (boolean hidden)
    {
        this._hidden = hidden;
    }
    
    /**
     * Set the multiple property
     * @param multiple the multiple property
     */
    public void setMultiple (boolean multiple)
    {
        this._multiple = multiple;
    }
    
    /**
     * Determines if the column value is multiple
     * @return <code>true</code> if the value is multiple
     */
    @Override
    public boolean isMultiple ()
    {
        return this._multiple;
    }
    
    /**
     * Get the content type ID (only when the search criteria is of type CONTENT).
     * @return the content type ID.
     */
    public String getContentTypeId()
    {
        return this._contentTypeId;
    }
    
    /**
     * Set the content type ID (only when the search criteria is of type CONTENT).
     * @param contentTypeId the content type ID. 
     */
    public void setContentTypeId(String contentTypeId)
    {
        this._contentTypeId = contentTypeId;
    }
    
    @Override
    public SearchField getSearchField()
    {
        // Override to provide a specific implementation.
        return null;
    }
    
    /**
     * Initialize the validator.
     * @param validatorManager The validator manager.
     * @param pluginName The plugin name.
     * @param role The validator role.
     * @param config The validator configuration.
     * @return true if the validator was successfully added, false otherwise.
     * @throws ConfigurationException If an error occurs.
     */
    @SuppressWarnings("unchecked")
    protected boolean _initializeValidator(ThreadSafeComponentManager<Validator> validatorManager, String pluginName, String role, Configuration config) throws ConfigurationException
    {
        Configuration validatorConfig = config.getChild("validation", false);
        
        if (validatorConfig != null)
        {
            String validatorClassName = StringUtils.defaultIfBlank(validatorConfig.getChild("custom-validator").getAttribute("class", ""), DefaultValidator.class.getName());
            
            try
            {
                Class validatorClass = Class.forName(validatorClassName);
                validatorManager.addComponent(pluginName, null, role, validatorClass, config);
                return true;
            }
            catch (Exception e)
            {
                throw new ConfigurationException("Unable to instantiate validator for class: " + validatorClassName, e);
            }
        }
        
        return false;
    }
    
    /**
     * Configure an i18nizable text
     * @param config The Configuration.
     * @param defaultValue The default value as an I18nizableText.
     * @return The i18nizable text
     */
    protected I18nizableText _configureI18nizableText(Configuration config, I18nizableText defaultValue)
    {
        if (config != null)
        {
            return I18nizableText.parseI18nizableText(config, null, "");
        }
        else
        {
            return defaultValue;
        }
    }
}
