/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.ui.model.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

import org.ametys.cms.search.model.DefaultSearchModel;
import org.ametys.cms.search.model.SearchCriterion;
import org.ametys.cms.search.ui.model.SearchUICriterion;
import org.ametys.cms.search.ui.model.SearchUIModel;

/**
 * Default implementation of a {@link SearchUIModel}.
 */
public class DefaultSearchUIModel extends DefaultSearchModel implements SearchUIModel
{
    /** The simple search ui criteria, indexed by ID. */
    protected Map<String, SearchUICriterion> _uiCriteria;
    /** The faceted search ui criteria, indexed by ID. */
    protected Map<String, SearchUICriterion> _uiFacetedCriteria;
    /** The advanced search ui criteria, indexed by ID. */
    protected Map<String, SearchUICriterion> _uiAdvancedCriteria;
    
    /** The page size **/
    protected int _pageSize;
    
    /** The URL for search */
    protected String _searchUrl;
    /** The plugin name for search */
    protected String _searchUrlPlugin;
    
    /** The URL for CSV export of results */
    protected String _exportCSVUrl;
    /** The plugin name for CSV export of results */
    protected String _exportCSVUrlPlugin;
    /** The URL for doc export of results */
    protected String _exportDOCUrl;
    /** The plugin name for doc export of results */
    protected String _exportDOCUrlPlugin;
    /** The URL for XML export of results */
    protected String _exportXMLUrl;
    /** The plugin name for XML export of results */
    protected String _exportXMLUrlPlugin;
    /** The URL for PDF export of results */
    protected String _exportPDFUrl;
    /** The plugin name for PDF export of results */
    protected String _exportPDFUrlPlugin;
    
    /** The URL for print results */    
    protected String _printUrl;
    /** The plugin name for print results */ 
    protected String _printUrlPlugin;
    
    /** The name of the view to use for summary of the content */
    protected String _summaryView;
    
    /**
     * Default constructor.
     */
    public DefaultSearchUIModel()
    {
        super();
        
        _uiCriteria = new LinkedHashMap<>();
        _uiFacetedCriteria = new LinkedHashMap<>();
        _uiAdvancedCriteria = new LinkedHashMap<>();
    }
    
    /**
     * Constructor by copying an existing {@link SearchUIModel}.
     * @param searchUIModelToCopy The {@link SearchUIModel} to copy
     * @param contextualParameters The contextual parameters
     */
    public DefaultSearchUIModel(SearchUIModel searchUIModelToCopy, Map<String, Object> contextualParameters)
    {
        super(searchUIModelToCopy, contextualParameters);
        
        setAdvancedCriteria(new ArrayList<>(searchUIModelToCopy.getAdvancedCriteria(contextualParameters).values()));
        
        setPageSize(searchUIModelToCopy.getPageSize(contextualParameters));
        
        setSearchUrl(searchUIModelToCopy.getSearchUrl(contextualParameters));
        setSearchUrlPlugin(searchUIModelToCopy.getSearchUrlPlugin(contextualParameters));
        
        setExportCSVUrl(searchUIModelToCopy.getExportCSVUrl(contextualParameters));
        setExportCSVUrlPlugin(searchUIModelToCopy.getExportCSVUrlPlugin(contextualParameters));
        setExportDOCUrl(searchUIModelToCopy.getExportDOCUrl(contextualParameters));
        setExportDOCUrlPlugin(searchUIModelToCopy.getExportDOCUrlPlugin(contextualParameters));
        setExportXMLUrl(searchUIModelToCopy.getExportXMLUrl(contextualParameters));
        setExportXMLUrlPlugin(searchUIModelToCopy.getExportXMLUrlPlugin(contextualParameters));
        setExportPDFUrl(searchUIModelToCopy.getExportPDFUrl(contextualParameters));
        setExportPDFUrlPlugin(searchUIModelToCopy.getExportPDFUrlPlugin(contextualParameters));
        
        setPrintUrl(searchUIModelToCopy.getPrintUrl(contextualParameters));
        setPrintUrlPlugin(searchUIModelToCopy.getPrintUrlPlugin(contextualParameters));
    }

    @Override
    public Map<String, SearchUICriterion> getCriteria(Map<String, Object> contextualParameters)
    {
        return Collections.unmodifiableMap(_uiCriteria);
    }
    
    @Override
    public void addCriterion(SearchCriterion criterion)
    {
        super.addCriterion(criterion);
        
        if (criterion instanceof SearchUICriterion uiCriterion)
        {
            _uiCriteria.put(uiCriterion.getId(), uiCriterion);
        }
    }
    
    @Override
    public void setCriteria(Collection< ? extends SearchCriterion> criteria)
    {
        super.setCriteria(criteria);
        
        _uiCriteria = new LinkedHashMap<>();
        for (SearchCriterion criterion : criteria)
        {
            if (criterion instanceof SearchUICriterion uiCriterion)
            {
                _uiCriteria.put(uiCriterion.getId(), uiCriterion);
            }
        }
    }
    
    @Override
    public Map<String, SearchUICriterion> getFacetedCriteria(Map<String, Object> contextualParameters)
    {
        return Collections.unmodifiableMap(_uiFacetedCriteria);
    }
    
    @Override
    public void addFacetedCriterion(SearchCriterion criterion)
    {
        super.addFacetedCriterion(criterion);
        
        if (criterion instanceof SearchUICriterion uiCriterion)
        {
            _uiFacetedCriteria.put(uiCriterion.getId(), uiCriterion);
        }
    }
    
    @Override
    public void setFacetedCriteria(Collection< ? extends SearchCriterion> criteria)
    {
        super.setFacetedCriteria(criteria);
        
        _uiFacetedCriteria = new LinkedHashMap<>();
        for (SearchCriterion criterion : criteria)
        {
            if (criterion instanceof SearchUICriterion uiCriterion)
            {
                _uiFacetedCriteria.put(uiCriterion.getId(), uiCriterion);
            }
        }
    }
    
    public Map<String, SearchUICriterion> getAdvancedCriteria(Map<String, Object> contextualParameters)
    {
        return Collections.unmodifiableMap(_uiAdvancedCriteria);
    }
    
    /**
     * Add an advanced criterion
     * @param criterion the advanced criterion to add
     */
    public void addAdvancedCriterion(SearchUICriterion criterion)
    {
        _uiAdvancedCriteria.put(criterion.getId(), criterion);
    }
    
    /**
     * Set the advanced criteria
     * @param criteria The list of advanced criterion
     */
    public void setAdvancedCriteria(Collection< ? extends SearchUICriterion> criteria)
    {
        _uiAdvancedCriteria = new LinkedHashMap<>();
        for (SearchUICriterion criterion : criteria)
        {
            _uiAdvancedCriteria.put(criterion.getId(), criterion);
        }
    }

    public int getPageSize(Map<String, Object> contextualParameters)
    {
        return _pageSize;
    }
    
    /**
     * Set the page size
     * @param pageSize The page size to set. Set to negative or 0 for unlimited or default
     */
    public void setPageSize(int pageSize)
    {
        _pageSize = pageSize;
    }

    public String getSearchUrl(Map<String, Object> contextualParameters)
    {
        return _searchUrl;
    }
    
    /**
     * Set the URL for search
     * @param searchUrl The URL to set
     */
    public void setSearchUrl(String searchUrl)
    {
        _searchUrl = searchUrl;
    }

    public String getSearchUrlPlugin(Map<String, Object> contextualParameters)
    {
        return _searchUrlPlugin;
    }
    
    /**
     * Set the plugin name for search
     * @param searchUrlPlugin the plugin name to set
     */
    public void setSearchUrlPlugin(String searchUrlPlugin)
    {
        _searchUrlPlugin = searchUrlPlugin;
    }

    public String getExportCSVUrl(Map<String, Object> contextualParameters)
    {
        return _exportCSVUrl;
    }
    
    /**
     * Set the URL for CSV export of results
     * @param exportCSVUrl The URL to set
     */
    public void setExportCSVUrl(String exportCSVUrl)
    {
        _exportCSVUrl = exportCSVUrl;
    }

    public String getExportCSVUrlPlugin(Map<String, Object> contextualParameters)
    {
        return _exportCSVUrlPlugin;
    }
    
    /**
     * Set the plugin name for CSV export of results
     * @param exportCSVUrlPlugin the plugin name to set
     */
    public void setExportCSVUrlPlugin(String exportCSVUrlPlugin)
    {
        _exportCSVUrlPlugin = exportCSVUrlPlugin;
    }

    public String getExportDOCUrl(Map<String, Object> contextualParameters)
    {
        return _exportDOCUrl;
    }

    /**
     * Set the URL for Doc export of results
     * @param exportDOCUrl The URL to set
     */
    public void setExportDOCUrl(String exportDOCUrl)
    {
        _exportDOCUrl = exportDOCUrl;
    }

    public String getExportDOCUrlPlugin(Map<String, Object> contextualParameters)
    {
        return _exportDOCUrlPlugin;
    }
    
    /**
     * Set the plugin name for Doc export of results
     * @param exportDOCUrlPlugin the plugin name to set
     */
    public void setExportDOCUrlPlugin(String exportDOCUrlPlugin)
    {
        _exportDOCUrlPlugin = exportDOCUrlPlugin;
    }

    public String getExportXMLUrl(Map<String, Object> contextualParameters)
    {
        return _exportXMLUrl;
    }
    
    /**
     * Set the URL for XML export of results
     * @param exportXMLUrl The URL to set
     */
    public void setExportXMLUrl(String exportXMLUrl)
    {
        _exportXMLUrl = exportXMLUrl;
    }

    public String getExportXMLUrlPlugin(Map<String, Object> contextualParameters)
    {
        return _exportXMLUrlPlugin;
    }
    
    /**
     * Set the plugin name for XML export of results
     * @param exportXMLUrlPlugin the plugin name to set
     */
    public void setExportXMLUrlPlugin(String exportXMLUrlPlugin)
    {
        _exportXMLUrlPlugin = exportXMLUrlPlugin;
    }

    public String getExportPDFUrl(Map<String, Object> contextualParameters)
    {
        return _exportPDFUrl;
    }
    
    /**
     * Set the URL for PDF export of results
     * @param exportPDFUrl The URL to set
     */
    public void setExportPDFUrl(String exportPDFUrl)
    {
        _exportPDFUrl = exportPDFUrl;
    }

    public String getExportPDFUrlPlugin(Map<String, Object> contextualParameters)
    {
        return _exportPDFUrlPlugin;
    }
    
    /**
     * Set the plugin name for PDF export of results
     * @param exportPDFUrlPlugin the plugin name to set
     */
    public void setExportPDFUrlPlugin(String exportPDFUrlPlugin)
    {
        _exportPDFUrlPlugin = exportPDFUrlPlugin;
    }

    public String getPrintUrl(Map<String, Object> contextualParameters)
    {
        return _printUrl;
    }
    
    /**
     * Set the URL for print results
     * @param printUrl The URL to set
     */
    public void setPrintUrl(String printUrl)
    {
        _printUrl = printUrl;
    }

    public String getPrintUrlPlugin(Map<String, Object> contextualParameters)
    {
        return _printUrlPlugin;
    }
    
    /**
     * Set the plugin name for print results
     * @param printUrlPlugin the plugin name to set
     */
    public void setPrintUrlPlugin(String printUrlPlugin)
    {
        _printUrlPlugin = printUrlPlugin;
    }

    public String getSummaryView()
    {
        return _summaryView;
    }

    /**
     * Set The name of the view to use for summary of the content
     * @param summaryView the name to set
     */
    public void setSummaryView(String summaryView)
    {
        _summaryView = summaryView;
    }
}
