/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.ui.model.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.BooleanUtils;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.content.properties.AbstractMultiValuesProperty;
import org.ametys.cms.content.referencetable.HierarchicalReferenceTablesHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.contenttype.MetadataType;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.QueryBuilder;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.content.ContentSearchHelper;
import org.ametys.cms.search.model.IndexingFieldSearchCriterion;
import org.ametys.cms.search.query.AndQuery;
import org.ametys.cms.search.query.BooleanQuery;
import org.ametys.cms.search.query.ContentQuery;
import org.ametys.cms.search.query.DateQuery;
import org.ametys.cms.search.query.DateTimeQuery;
import org.ametys.cms.search.query.DoubleQuery;
import org.ametys.cms.search.query.GeocodeQuery;
import org.ametys.cms.search.query.JoinQuery;
import org.ametys.cms.search.query.LongQuery;
import org.ametys.cms.search.query.OrQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.query.RichTextQuery;
import org.ametys.cms.search.query.StringQuery;
import org.ametys.cms.search.query.UsersQuery;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.date.AdaptableDateParser;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemGroup;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.parameter.Validator;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * This class is a search criteria on a metadata of a content
 */
public class IndexingFieldSearchUICriterion extends AbstractSearchUICriterion implements IndexingFieldSearchCriterion
{
    /** Prefix for id of metadata search criteria */
    public static final String SEARCH_CRITERIA_METADATA_PREFIX = "metadata-";
    
    /** The content search helper. */
    protected ContentSearchHelper _searchHelper;
    /** The hierarchical reference tables helper. */
    protected HierarchicalReferenceTablesHelper _hierarchicalReferenceTablesHelper;
    /** The user helper */
    protected UserHelper _userHelper;
    /** The helper for convenient methods on content types */
    protected ContentTypesHelper _contentTypesHelper;
    /** The ametys object resolver. */
    protected AmetysObjectResolver _resolver;
    
    /** The criteria operator */
    protected Operator _operator;
    
    /** The definition used for this criterion */
    protected ElementDefinition _definition;
    
    /** The field full path */
    protected String _fullPath;
    
    /** The field path */
    protected String _fieldPath;
    
    /** The join paths */
    protected List<String> _joinPaths;
    
    /** Is it AND or OR for multiple metadata */
    protected boolean _isMultipleOperandAnd;
    
    /** ComponentManager for {@link Validator}s. */
    protected ThreadSafeComponentManager<Validator> _validatorManager;
    
    private boolean _isTypeContentWithMultilingualTitle;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _searchHelper = (ContentSearchHelper) manager.lookup(ContentSearchHelper.ROLE);
        _hierarchicalReferenceTablesHelper = (HierarchicalReferenceTablesHelper) manager.lookup(HierarchicalReferenceTablesHelper.ROLE);
        _userHelper = (UserHelper) manager.lookup(UserHelper.ROLE);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    public void dispose()
    {
        super.dispose();
        
        _validatorManager.dispose();
        _validatorManager = null;
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        try
        {
            _validatorManager = new ThreadSafeComponentManager<>();
            _validatorManager.setLogger(_logger);
            _validatorManager.contextualize(_context);
            _validatorManager.service(_manager);
            
            _enumeratorManager = new ThreadSafeComponentManager<>();
            _enumeratorManager.setLogger(_logger);
            _enumeratorManager.contextualize(_context);
            _enumeratorManager.service(_manager);
            
            _fullPath = configuration.getChild("field").getAttribute("path");
            Set<String> baseContentTypeIds = new HashSet<>();
            for (Configuration cTypeConf : configuration.getChild("contentTypes").getChildren("baseType"))
            {
                baseContentTypeIds.add(cTypeConf.getAttribute("id"));
            }
            
            _operator = Operator.fromName(configuration.getChild("test-operator").getValue("eq"));
            _isMultipleOperandAnd = !StringUtils.equalsIgnoreCase(configuration.getChild("multiple-operand").getValue("or"), "or");
            _joinPaths = new ArrayList<>();
            
            String id = SEARCH_CRITERIA_METADATA_PREFIX + _fullPath + "-" + _operator.getName();
            
            if (!baseContentTypeIds.isEmpty())
            {
                try
                {
                    Collection<ContentType> baseContentTypes = baseContentTypeIds.stream()
                            .map(_contentTypeExtensionPoint::getExtension)
                            .collect(Collectors.toList());
                    ModelItem modelItem = ModelHelper.getModelItem(_fullPath, baseContentTypes);
                    if (modelItem instanceof ElementDefinition def)
                    {
                        _definition = def;
                    }
                    else
                    {
                        throw new ConfigurationException("Indexing field search criteria with path '" + _fullPath + "' refers to a group");
                    }

                    _joinPaths = _searchHelper.computeJoinPaths(_fullPath, baseContentTypeIds, false);
                }
                catch (UndefinedItemPathException e)
                {
                    throw new ConfigurationException("Indexing field search criteria with path '" + _fullPath + "' refers to an unknown model item", e);
                }
            }
            else if (Content.ATTRIBUTE_TITLE.equals(_fullPath))
            {
                // Only the title indexing field is allowed if there is no base content types.
                _definition = _contentTypesHelper.getTitleAttributeDefinition();
            }
            else
            {
                throw new ConfigurationException("The indexing field '" + _fullPath + "' is forbidden when no content type is specified: only title can be used.");
            }
            
            _fieldPath = _computeFieldPath(_definition);
            
            setId(id);
            setGroup(_configureI18nizableText(configuration.getChild("group", false), null));
            _configureLabel(configuration, _definition);
            _configureDescription(configuration, _definition);
            
            MetadataType type = MetadataType.fromModelItemType(_definition.getType());
            setType(type);
            
//            setValidator(finalDefinition.getValidator());
            String validatorRole = "validator";
            if (!_initializeValidator(_validatorManager, "cms", validatorRole, configuration))
            {
                validatorRole = null;
            }
            
            setEnumerator(configureEnumerator(configuration, _definition));
            setWidget(configureWidget(configuration, _definition));
            setWidgetParameters(configureWidgetParameters(configuration, _definition));

            setMultiple(_definition.isMultiple());
            
            _configureTypeContentWithMultilingualTitle(_definition);
            
            _configureOperator(configuration, _definition);
            
            configureUIProperties(configuration);
            configureValues(configuration);
        }
        catch (Exception e)
        {
            throw new ConfigurationException("Error configuring the indexing field search criterion.", configuration, e);
        }
    }
    
    /**
     * Get the operator.
     * @return the operator.
     */
    public Operator getOperator()
    {
        return _operator;
    }
    
    public String getFieldId()
    {
        return SEARCH_CRITERIA_METADATA_PREFIX + _fullPath;
    }
    
    private String _computeFieldPath(ModelItem definition)
    {
        StringBuilder path = new StringBuilder();
        
        ModelItemGroup parent = definition.getParent();
        if (parent != null && !(parent instanceof RepeaterDefinition))
        {
            path.append(_computeFieldPath(parent)).append(ModelItem.ITEM_PATH_SEPARATOR);
        }
        
        path.append(definition.getName());
        return path.toString();
    }
    
    /**
     * Get the path of field (separated by '/')
     * @return the path of the field.
     */
    public String getFieldPath()
    {
        return _fieldPath;
    }
    
    /**
     * Get the join paths, separated with slashes.
     * @return the join paths.
     */
    public List<String> getJoinPaths()
    {
        return Collections.unmodifiableList(_joinPaths);
    }
    
    @Override
    public Query getQuery(Object value, Operator customOperator, Map<String, Object> allValues, String language, Map<String, Object> contextualParameters)
    {
        String fieldPath = getFieldPath();
        List<String> joinPaths = getJoinPaths();
        Operator operator = customOperator != null ? customOperator : getOperator();
        
        if (operator != Operator.EXISTS && isEmpty(value))
        {
            return null;
        }
        boolean isValueEscaped = BooleanUtils.isTrue((Boolean) contextualParameters.get(QueryBuilder.VALUE_IS_ESCAPED));
        
        Query query = null;
        switch (getType())
        {
            case DATE:
                query = getDateQuery(value, fieldPath, operator);
                break;
            case DATETIME:
                query = getDateTimeQuery(value, fieldPath, operator);
                break;
            case LONG:
                query = getLongQuery(value, fieldPath, operator);
                break;
            case DOUBLE:
                query = getDoubleQuery(value, fieldPath, operator);
                break;
            case BOOLEAN:
                query = getBooleanQuery(value, fieldPath, operator);
                break;
            case STRING:
                if (_definition instanceof AbstractMultiValuesProperty property && ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID.equals(property.getElementDefinition().getType().getId()))
                {
                    query = getContentQuery(value, language, fieldPath, operator);
                }
                else
                {
                    if (contextualParameters.containsKey(QueryBuilder.MULTILINGUAL_SEARCH))
                    {
                        // Force language to null if the search is concern a string field on multilingual contents
                        query = getStringQuery(value, null, fieldPath, operator, isValueEscaped);
                    }
                    else
                    {
                        query = getStringQuery(value, language, fieldPath, operator, isValueEscaped);
                    }
                }
                break;
            case RICH_TEXT:
                if (contextualParameters.containsKey(QueryBuilder.MULTILINGUAL_SEARCH))
                {
                    // Force language to null if the search is concern a string field on multilingual contents
                    query = getRichTextQuery(value, null, fieldPath, operator, isValueEscaped);
                }
                else
                {
                    query = getRichTextQuery(value, language, fieldPath, operator, isValueEscaped);
                }
                break;
            case MULTILINGUAL_STRING:
                query = getStringQuery(value, language, fieldPath, operator, isValueEscaped);
                break;
            case USER:
                query = getUserQuery(value, fieldPath, operator);
                break;
            case CONTENT:
                query = getContentQuery(value, language, fieldPath, operator);
                break;
            case REFERENCE:
                query = getStringQuery(value, language, fieldPath, operator, isValueEscaped);
                break;
            case GEOCODE:
                query = new GeocodeQuery(fieldPath, operator, (Map<String, Integer>) value);
                break;
            default:
                return null;
        }
        
        if (query != null && !joinPaths.isEmpty())
        {
            query = new JoinQuery(query, joinPaths);
        }
        
        return query;
    }

    private static boolean isEmpty(Object value)
    {
        return value == null
            || value instanceof String && ((String) value).length() == 0
            || value instanceof List && ((List) value).isEmpty();
    }

    /**
     * Get a string query.
     * @param value The value to use for this criterion.
     * @param language The search language.
     * @param fieldPath The field path.
     * @param operator The query operator to use.
     * @param isValueEscaped <code>true</code> if the given value is already escaped
     * @return The query.
     */
    protected Query getStringQuery(Object value, String language, String fieldPath, Operator operator, boolean isValueEscaped)
    {
        if (operator == Operator.EXISTS)
        {
            return new StringQuery(fieldPath, operator, null, language);
        }
        else
        {
            return getTextQuery(value, language, fieldPath, operator, isValueEscaped);
        }
    }
    
    /**
     * Get a boolean query.
     * @param value The value to use for this criterion.
     * @param fieldPath The field path.
     * @param operator The query operator to use.
     * @return The query.
     */
    protected Query getBooleanQuery(Object value, String fieldPath, Operator operator)
    {
        if (operator == Operator.EXISTS)
        {
            return new BooleanQuery(fieldPath);
        }
        else if (value instanceof Collection<?>)
        {
            @SuppressWarnings("unchecked")
            Collection<Object> values = (Collection<Object>) value;
            
            List<Query> queries = new ArrayList<>();
            for (Object val : values)
            {
                queries.add(getBooleanQuery(val, fieldPath, operator));
            }
            
            return _isMultipleOperandAnd ? new AndQuery(queries) : new OrQuery(queries);
        }
        else if (value instanceof String)
        {
            Boolean boolValue = Boolean.parseBoolean((String) value);
            return new BooleanQuery(fieldPath, boolValue);
        }
        else
        {
            return new BooleanQuery(fieldPath, (Boolean) value);
        }
    }
    
    /**
     * Get a double query.
     * @param value The value to use for this criterion.
     * @param fieldPath The field path.
     * @param operator The query operator to use.
     * @return The query.
     */
    protected Query getDoubleQuery(Object value, String fieldPath, Operator operator)
    {
        if (operator == Operator.EXISTS)
        {
            return new DoubleQuery(fieldPath);
        }
        else if (value instanceof Collection<?>)
        {
            @SuppressWarnings("unchecked")
            Collection<Object> values = (Collection<Object>) value;
            
            List<Query> queries = new ArrayList<>();
            for (Object val : values)
            {
                queries.add(getDoubleQuery(val, fieldPath, operator));
            }
            
            return _isMultipleOperandAnd ? new AndQuery(queries) : new OrQuery(queries);
        }
        else if (value instanceof String)
        {
            Double doubleValue = Double.parseDouble((String) value);
            return new DoubleQuery(fieldPath, operator, doubleValue);
        }
        else if (value instanceof Integer)
        {
            return new DoubleQuery(fieldPath, operator, ((Integer) value).doubleValue());
        }
        else if (value instanceof Number)
        {
            Number valueAsNumber = (Number) value;
            Double valueAsDouble = valueAsNumber.doubleValue();
            // We check that the cast to long was safe
            if (valueAsNumber.toString().equals(valueAsDouble.toString()))
            {
                return new DoubleQuery(fieldPath, operator, valueAsDouble);
            }
            else
            {
                throw new NumberFormatException("Failed to convert the value " + value + " to Double. The value is out of bound.");
            }
        }
        else
        {
            throw new NumberFormatException("Failed to convert the value " + value + " to Double.");
        }
    }
    
    /**
     * Get a long query.
     * @param value The value to use for this criterion.
     * @param fieldPath The field path.
     * @param operator The query operator to use.
     * @return The query.
     */
    protected Query getLongQuery(Object value, String fieldPath, Operator operator)
    {
        if (operator == Operator.EXISTS)
        {
            return new LongQuery(fieldPath);
        }
        else if (value instanceof Collection<?>)
        {
            @SuppressWarnings("unchecked")
            Collection<Object> values = (Collection<Object>) value;
            
            List<Query> queries = new ArrayList<>();
            for (Object val : values)
            {
                queries.add(getLongQuery(val, fieldPath, operator));
            }
            
            return _isMultipleOperandAnd ? new AndQuery(queries) : new OrQuery(queries);
        }
        else if (value instanceof String)
        {
            return new LongQuery(fieldPath, operator, Long.parseLong((String) value));
        }
        else if (value instanceof Number)
        {   
            Number valueAsNumber = (Number) value;
            Long valueAsLong = valueAsNumber.longValue();
            // We check that the cast to long was safe
            if (valueAsNumber.toString().equals(valueAsLong.toString()))
            {
                return new LongQuery(fieldPath, operator, valueAsLong);
            }
            else
            {
                throw new NumberFormatException("Failed to convert the value " + value + " to Long. The value is out of bound.");
            }
        }
        else
        {
            throw new NumberFormatException("Failed to convert the value " + value + " to Long.");
        }
    }
    
    /**
     * Get a date query.
     * @param value The value to use for this criterion.
     * @param fieldPath The field path.
     * @param operator The query operator to use.
     * @return The query.
     */
    protected Query getDateQuery(Object value, String fieldPath, Operator operator)
    {
        if (operator == Operator.EXISTS)
        {
            return new DateQuery(fieldPath);
        }
        else if (value instanceof Collection<?>)
        {
            @SuppressWarnings("unchecked")
            Collection<Object> values = (Collection<Object>) value;
            
            List<Query> queries = new ArrayList<>();
            for (Object val : values)
            {
                queries.add(getDateQuery(val, fieldPath, operator));
            }
            
            return _isMultipleOperandAnd ? new AndQuery(queries) : new OrQuery(queries);
        }
        else 
        {
            return _getSingleDateQuery((String) value, fieldPath, operator);
        }
    }
    
    private Query _getSingleDateQuery(String value, String fieldPath, Operator operator)
    {
        return new DateQuery(fieldPath, operator, AdaptableDateParser.parse(value));
    }
    
    /**
     * Get a dateTime query.
     * @param value The value to use for this criterion.
     * @param fieldPath The field path.
     * @param operator The query operator to use.
     * @return The query.
     */
    protected Query getDateTimeQuery(Object value, String fieldPath, Operator operator)
    {
        if (operator == Operator.EXISTS)
        {
            return new DateTimeQuery(fieldPath);
        }
        else if (value instanceof Collection<?>)
        {
            @SuppressWarnings("unchecked")
            Collection<Object> values = (Collection<Object>) value;
            
            List<Query> queries = new ArrayList<>();
            for (Object val : values)
            {
                queries.add(getDateTimeQuery(val, fieldPath, operator));
            }
            
            return _isMultipleOperandAnd ? new AndQuery(queries) : new OrQuery(queries);
        }
        else 
        {
            return _getSingleDateTimeQuery((String) value, fieldPath, operator);
        }
    }
    
    private Query _getSingleDateTimeQuery(String value, String fieldPath, Operator operator)
    {
        return new DateTimeQuery(fieldPath, operator, AdaptableDateParser.parse(value));
    }
    
    /**
     * Get a content query.
     * @param value The value to use for this criterion.
     * @param language The search language.
     * @param fieldPath The field path.
     * @param operator The query operator to use.
     * @return The query.
     */
    protected Query getContentQuery(Object value, String language, String fieldPath, Operator operator)
    {        
        boolean isMultipleOperandAnd = operator == Operator.NE ? !_isMultipleOperandAnd : _isMultipleOperandAnd; // reverse it if NE
        return new ContentQuery(fieldPath, operator, value, _resolver, _contentHelper, _hierarchicalReferenceTablesHelper, false, isMultipleOperandAnd);
    }
    
    /**
     * Get a user query
     * @param value The value to use for this criterion.
     * @param fieldPath The field path.
     * @param operator The query operator to use.
     * @return The query.
     */
    @SuppressWarnings("unchecked")
    protected Query getUserQuery(Object value, String fieldPath, Operator operator)
    {
        List<UserIdentity> users;
        if (value instanceof Map<?, ?>)
        {
            users = List.of(_userHelper.json2userIdentity((Map<String, Object>) value));
        }
        else if (value instanceof Collection< ? >)
        {
            users = ((Collection<Map<String, Object>>) value).stream()
                                                             .map(_userHelper::json2userIdentity).toList();
        }
        else
        {
            throw new IllegalArgumentException("The value for the field '" + fieldPath + "' of type USER cannot be cast to a Map or a Collection");
        }
        return new UsersQuery(fieldPath + "_s", operator, users);
    }

    /**
     * Get a geocode query.
     * @param value The value to use for this criterion.
     * @param fieldPath The field path.
     * @param operator The query operator to use.
     * @return The query.
     */
    protected Query getGeocodeQuery(Object value, String fieldPath, Operator operator)
    {
        if (operator == Operator.EXISTS)
        {
            return new GeocodeQuery(fieldPath);
        }
        else if (value instanceof Collection<?>)
        {
            @SuppressWarnings("unchecked")
            Collection<Map<String, Integer>> values = (Collection<Map<String, Integer>>) value;
            
            List<Query> queries = new ArrayList<>();
            for (Object val : values)
            {
                queries.add(getGeocodeQuery(val, fieldPath, operator));
            }
            
            return _isMultipleOperandAnd ? new AndQuery(queries) : new OrQuery(queries);
        }
        else if (value instanceof Map<?, ?>)
        {
            return new GeocodeQuery(fieldPath, operator, (Map<String, Integer>) value);
        }
        else
        {
            throw new IllegalArgumentException("The value for the field '" + fieldPath + "' of type GEOCODE cannot be cast to a Map or a Collection");
        }
    }
    
    /**
     * Get a rich text query.
     * @param value The value to use for this criterion.
     * @param language The search language.
     * @param fieldPath The field path.
     * @param operator The query operator to use.
     * @param isValueEscaped <code>true</code> if the given value is already escaped
     * @return The query.
     */
    protected Query getRichTextQuery(Object value, String language, String fieldPath, Operator operator, boolean isValueEscaped)
    {
        if (operator == Operator.EXISTS)
        {
            return new RichTextQuery(fieldPath, operator, null, language);
        }
        else
        {
            return getTextQuery(value, language, fieldPath, operator, isValueEscaped);
        }
    }
    
    /**
     * Get a text query.
     * @param value The value to use for this criterion.
     * @param language The search language.
     * @param fieldPath The field path.
     * @param operator The query operator to use.
     * @param isValueEscaped <code>true</code> if the given value is already escaped
     * @return The query.
     */
    protected Query getTextQuery(Object value, String language, String fieldPath, Operator operator, boolean isValueEscaped)
    {
        if (value instanceof Collection<?>)
        {
            @SuppressWarnings("unchecked")
            Collection<Object> values = (Collection<Object>) value;
            
            List<Query> queries = new ArrayList<>();
            for (Object val : values)
            {
                queries.add(getStringQuery(val, language, fieldPath, operator, isValueEscaped));
            }
            
            return _isMultipleOperandAnd ? new AndQuery(queries) : new OrQuery(queries);
        }
        else if (operator.equals(Operator.LIKE))
        {
            String stringValue = (String) value;
            
            if (StringUtils.isNotEmpty(stringValue))
            {
                boolean doNotUseLanguage = getType() != MetadataType.MULTILINGUAL_STRING;
                return new StringQuery(fieldPath, operator, stringValue, doNotUseLanguage ? null : language, isValueEscaped);
            }
            
            return null;
        }
        else
        {
            return new StringQuery(fieldPath, operator, (String) value, language, isValueEscaped);
        }
    }
    
    @Override
    public SearchField getSearchField()
    {
        return _searchHelper.getMetadataSearchField(getJoinPaths(), getFieldPath(), getType(), _isTypeContentWithMultilingualTitle);
    }
    
    private void _configureTypeContentWithMultilingualTitle(ModelItem modelItem)
    {
        _isTypeContentWithMultilingualTitle = _searchHelper.isTitleMultilingual(modelItem);
    }
    
    /**
     * Configure the criterion operator.
     * @param configuration the global criterion configuration.
     * @param definition the metadata definition.
     * @throws ConfigurationException if an error occurs.
     */
    @SuppressWarnings("unchecked")
    private void _configureOperator(Configuration configuration, ElementDefinition definition) throws ConfigurationException
    {
        try
        {
            String definitionTypeId = definition.getType().getId();
            if (definition instanceof AbstractMultiValuesProperty property)
            {
                definitionTypeId = property.getElementDefinition().getType().getId();
            }
            
            String op = configuration.getChild("test-operator").getValue("");
            if (StringUtils.isNotBlank(op))
            {
                _operator = Operator.fromName(op);
            }
            else
            {
                if (definition.getCriterionEnumerator(configuration, _enumeratorManager) == null
                    && (org.ametys.runtime.model.type.ModelItemTypeConstants.STRING_TYPE_ID.equals(definitionTypeId)
                        || ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID.equals(definitionTypeId)
                        || ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID.equals(definitionTypeId)
                        || ModelItemTypeConstants.REFERENCE_ELEMENT_TYPE_ID.equals(definitionTypeId)))
                {
                    _operator = Operator.SEARCH;
                }
                else
                {
                    _operator = Operator.EQ;
                }
            }
        }
        catch (IllegalArgumentException e)
        {
            throw new ConfigurationException("Invalid operator", configuration, e);
        }
    }
    
    private void _configureLabel(Configuration configuration, ModelItem modelItem)
    {
        setLabel(_configureI18nizableText(configuration.getChild("label", false), modelItem.getLabel()));
    }
    
    private void _configureDescription(Configuration configuration, ModelItem modelItem)
    {
        setDescription(_configureI18nizableText(configuration.getChild("description", false), modelItem.getDescription()));
    }
}
