/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.ui.model.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.MetadataType;
import org.ametys.cms.model.ContentElementDefinition;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.content.ContentSearchHelper;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.cms.search.model.SystemSearchCriterion;
import org.ametys.cms.search.query.JoinQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.solr.field.JoinedSystemSearchField;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.parameter.Validator;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * This class is a search criteria on a system property (author, lastModified, with-comments, ...)
 */
public class SystemSearchUICriterion extends AbstractSearchUICriterion implements SystemSearchCriterion
{
    /** Prefix for id of system property search criteria */
    public static final String SEARCH_CRITERION_SYSTEM_PREFIX = "property-";
    
    /** ComponentManager for {@link Validator}s. */
    protected ThreadSafeComponentManager<Validator> _validatorManager;
    
    /** The system property extension point. */
    protected SystemPropertyExtensionPoint _systemPropEP;
    
    /** The content search helper. */
    protected ContentSearchHelper _searchHelper;
    
    /** The join paths */
    protected List<String> _joinPaths;
    
    private Operator _operator;
    private SystemProperty _systemProperty;
    private Set<String> _contentTypes;
    private String _fullPath;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _systemPropEP = (SystemPropertyExtensionPoint) manager.lookup(SystemPropertyExtensionPoint.ROLE);
        _searchHelper = (ContentSearchHelper) manager.lookup(ContentSearchHelper.ROLE);
    }
    
    @Override
    public void dispose()
    {
        super.dispose();
        _validatorManager.dispose();
        _validatorManager = null;
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        try
        {
            _validatorManager = new ThreadSafeComponentManager<>();
            _validatorManager.setLogger(_logger);
            _validatorManager.contextualize(_context);
            _validatorManager.service(_manager);
            
            _enumeratorManager = new ThreadSafeComponentManager<>();
            _enumeratorManager.setLogger(_logger);
            _enumeratorManager.contextualize(_context);
            _enumeratorManager.service(_manager);
            
            _fullPath = configuration.getChild("systemProperty").getAttribute("name");
            int pos = _fullPath.lastIndexOf(ModelItem.ITEM_PATH_SEPARATOR);
            
            String systemPropertyId = pos > -1 ? _fullPath.substring(pos + ModelItem.ITEM_PATH_SEPARATOR.length()) : _fullPath;
            _operator = Operator.fromName(configuration.getChild("test-operator").getValue("eq"));
            
            if (!_systemPropEP.isSearchable(systemPropertyId))
            {
                throw new ConfigurationException("The property '" + systemPropertyId + "' doesn't exist or is not searchable.");
            }
            
            Set<String> baseContentTypeIds = new HashSet<>();
            for (Configuration cTypeConf : configuration.getChild("contentTypes").getChildren("baseType"))
            {
                baseContentTypeIds.add(cTypeConf.getAttribute("id"));
            }
            
            _contentTypes = new HashSet<>();
            String joinPath = pos > -1 ? _fullPath.substring(0, pos) : "";
            _joinPaths = _configureJoinPaths(joinPath, baseContentTypeIds, configuration);
            _systemProperty = _systemPropEP.getExtension(systemPropertyId);
            
            setId(SEARCH_CRITERION_SYSTEM_PREFIX + _fullPath + "-" + _operator.getName());
            setGroup(_configureI18nizableText(configuration.getChild("group", false), null));
            
            String validatorRole = "validator";
            if (!_initializeValidator(_validatorManager, "cms", validatorRole, configuration))
            {
                validatorRole = null;
            }
            
            setLabel(_systemProperty.getLabel());
            setDescription(_systemProperty.getDescription());
            MetadataType type = MetadataType.fromModelItemType(_systemProperty.getType());
            setType(type);
            // Multiple defaults to false even for a multiple property.
            setMultiple(configuration.getAttributeAsBoolean("multiple", false));
            
            // TODO Add the current criterion configuration.
            Configuration enumeratorAndWidgetParamConf = _getEnumeratorAndWidgetParamConf(configuration);
            setEnumerator(configureEnumerator(configuration, _systemProperty, enumeratorAndWidgetParamConf));
            setWidget(configureWidget(configuration, _systemProperty));
            setWidgetParameters(configureWidgetParameters(configuration, _systemProperty, enumeratorAndWidgetParamConf));
            
            // Potentially replace the standard label and description by the custom ones.
            I18nizableText userLabel = _configureI18nizableText(configuration.getChild("label", false), null);
            if (userLabel != null)
            {
                setLabel(userLabel);
            }
            I18nizableText userDescription = _configureI18nizableText(configuration.getChild("description", false), null);
            if (userDescription != null)
            {
                setDescription(userDescription);
            }
            
            configureUIProperties(configuration);
            configureValues(configuration);
            
            if (validatorRole != null)
            {
                _validatorManager.initialize();
                setValidator(_validatorManager.lookup(validatorRole));
            }
        }
        catch (Exception e)
        {
            throw new ConfigurationException("Error configuring the system search criterion.", configuration, e);
        }
    }
    
    private Configuration _getEnumeratorAndWidgetParamConf(Configuration critConf) throws ConfigurationException
    {
        if (isJoined())
        {
            DefaultConfiguration widgetParamConf = new DefaultConfiguration(critConf);
            widgetParamConf.removeChild(widgetParamConf.getChild("contentTypes"));
            if (!_contentTypes.isEmpty())
            {
                DefaultConfiguration cTypesConf = new DefaultConfiguration("contentTypes");
                widgetParamConf.addChild(cTypesConf);
                for (String contentType : _contentTypes)
                {
                    DefaultConfiguration cTypeConf = new DefaultConfiguration("type");
                    cTypeConf.setAttribute("id", contentType);
                    cTypesConf.addChild(cTypeConf);
                }
            }
            
            return widgetParamConf;
        }
        
        // in case non join, the global conf already contains the real content types
        return critConf;
        
    }
    
    @Override
    public boolean isJoined()
    {
        return CollectionUtils.isNotEmpty(_joinPaths);
    }
    
    @Override
    public boolean isFacetable()
    {
        return _systemProperty.isFacetable();
    }
    
    /**
     * Get the operator.
     * @return the operator.
     */
    public Operator getOperator()
    {
        return _operator;
    }
    
    public String getFieldId()
    {
        return SEARCH_CRITERION_SYSTEM_PREFIX + _systemProperty.getName();
    }
    
    /**
     * Get id of this system property
     * @return The system property's id
     */
    public String getSystemPropertyId()
    {
        return _systemProperty.getName();
    }
    
    @Override
    public Query getQuery(Object value, Operator customOperator, Map<String, Object> allValues, String language, Map<String, Object> contextualParameters)
    {
        if (customOperator != Operator.EXISTS
            && (value == null
                || value instanceof String && ((String) value).length() == 0
                || value instanceof List && ((List) value).isEmpty()))
        {
            return null;
        }
        
        Operator operator = customOperator != null ? customOperator : getOperator();
        
        Query query = _systemProperty.getQuery(value, operator, language, contextualParameters);
        
        if (query != null && !_joinPaths.isEmpty())
        {
            query = new JoinQuery(query, _joinPaths);
        }
        
        return query;
    }
    
    @Override
    public SearchField getSearchField()
    {
        SearchField sysSearchField = _systemProperty.getSearchField();
        if (_joinPaths.isEmpty())
        {
            return sysSearchField;
        }
        else
        {
            return new JoinedSystemSearchField(_joinPaths, sysSearchField);
        }
    }
    
    /**
     * Configure the join paths.
     * @param joinPath The full join path.
     * @param contentTypeIds The base content type identifiers.
     * @param configuration The configuration of the criterion
     * @return The list of join paths.
     * @throws ConfigurationException If the join paths are missconfigured
     */
    private List<String> _configureJoinPaths(String joinPath, Set<String> contentTypeIds, Configuration configuration) throws ConfigurationException
    {
        List<String> joinPaths = new ArrayList<>();
        
        if (StringUtils.isNotBlank(joinPath))
        {
            if (contentTypeIds.isEmpty())
            {
                throw new ConfigurationException("System search criterion with path '" + _fullPath + "': impossible to configure a joint system property without base content types.");
            }
            
            try
            {
                Collection<ContentType> contentTypes = contentTypeIds.stream()
                                                                     .map(_contentTypeExtensionPoint::getExtension)
                                                                     .collect(Collectors.toList());
                ModelItem modelItem = ModelHelper.getModelItem(joinPath, contentTypes);
                // The definition must be a content (to be able to extract a system property from it).
                if (modelItem instanceof ContentElementDefinition contentElementDefinition)
                {
                    joinPaths.addAll(_searchHelper.computeJoinPaths(joinPath, contentTypeIds, true));
                    // Add in _contentTypes the good types
                    Optional.ofNullable(contentElementDefinition.getContentTypeId())
                            .map(Collections::singletonList)
                            .map(this::_typeAndSubTypes)
                            .ifPresent(_contentTypes::addAll);
                }
                else
                {
                    throw new ConfigurationException("'" + _fullPath + "' is not a valid system search criterion path as '" + joinPath + "' does not represent a content.");
                }
            }
            catch (UndefinedItemPathException e)
            {
                throw new ConfigurationException("System search criterion with path '" + _fullPath + "' refers to an unknown model item: " + joinPath, e);
            }
        }
        else
        {
            for (Configuration cTypeConf : configuration.getChild("contentTypes").getChildren("type"))
            {
                _contentTypes.add(cTypeConf.getAttribute("id"));
            }
        }
        
        return joinPaths;
    }
    
    private Collection<String> _typeAndSubTypes(List<String> singletonType)
    {
        String typeInSingleton = singletonType.iterator().next();
        return CollectionUtils.union(singletonType, _contentTypeExtensionPoint.getSubTypes(typeInSingleton));
    }
}
