/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.ui.model.impl;

import java.util.Map;
import java.util.Optional;

import org.apache.cocoon.ProcessingException;

import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.ui.model.SearchUIColumn;
import org.ametys.cms.search.ui.model.SearchUIColumnHelper;
import org.ametys.runtime.model.DefinitionContext;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ViewElementAccessor;
import org.ametys.runtime.model.ViewItem;

/**
 * This class represents a result column
 */
public class ViewElementAccessorSearchUIColumn extends ViewElementAccessor implements SearchUIColumn<ElementDefinition>
{
    private int _width = -1;
    private boolean _hidden;
    private boolean _editable = true;
    private boolean _sortable = true;
    private boolean _allowSortOnMultipleJoin;
    private Optional<String> _defaultSorter = Optional.empty();
    private Optional<String> _renderer = Optional.empty();
    private Optional<String> _converter = Optional.empty();
    
    public int getWidth()
    {
        return _width > 0
                ? _width
                : SearchUIColumnHelper.getDefaultColumnWidth(this);
    }
    
    public void setWidth (int width)
    {
        _width = width;
    }
    
    public boolean isHidden()
    {
        return _hidden;
    }
    
    public void setHidden (boolean hidden)
    {
        _hidden = hidden;
    }
    
    public boolean isEditable()
    {
        return _editable && SearchUIColumnHelper.isEditionAllowed(this);
    }
    
    public void setEditable (boolean editable)
    {
        _editable = editable;
    }
    
    public boolean isSortable()
    {
        return _sortable && SearchUIColumnHelper.isSortAllowed(this);
    }
    
    public void setSortable (boolean sortable)
    {
        _sortable = sortable;
    }
    
    public boolean allowSortOnMultipleJoin()
    {
        return _allowSortOnMultipleJoin;
    }
    
    public void setAllowSortOnMultipleJoin(boolean allowSortOnMultipleJoin)
    {
        _allowSortOnMultipleJoin = allowSortOnMultipleJoin;
    }
    
    public String getDefaultSorter()
    {
        return isSortable()
                ? _defaultSorter.orElse(null)
                : null;
    }
    
    public void setDefaultSorter(Optional<String> defaultSorter)
    {
        _defaultSorter = defaultSorter;
    }
    
    public String getRenderer()
    {
        return _renderer
                .or(this::_getDefaultRenderer)
                .orElse(null);
    }
    
    private Optional<String> _getDefaultRenderer()
    {
        return getDefinition() instanceof SystemProperty systemProperty
                ? Optional.ofNullable(systemProperty.getRenderer())
                : Optional.empty();
    }
    
    public void setRenderer(Optional<String> renderer)
    {
        this._renderer = renderer;
    }
    
    public String getConverter()
    {
        return _converter
                .or(() -> SearchUIColumnHelper.getElementDefinitionDefaultConverter(getDefinition()))
                .orElse(null);
    }
    
    public void setConverter(Optional<String> converter)
    {
        this._converter = converter;
    }
    
    @Override
    public ViewElementAccessorSearchUIColumn createInstance()
    {
        return new ViewElementAccessorSearchUIColumn();
    }
    
    @Override
    public Map<String, Object> toJSON(DefinitionContext context) throws ProcessingException
    {
        Map<String, Object> json = super.toJSON(context);

        if (!json.isEmpty())
        {
            json.putAll(SearchUIColumnHelper.columnPropertiesToJSON(this));
        }
        return json;
    }
    
    @Override
    public void copyTo(ViewItem item)
    {
        super.copyTo(item);
        
        assert item instanceof ViewElementAccessorSearchUIColumn;
        ViewElementAccessorSearchUIColumn column = (ViewElementAccessorSearchUIColumn) item;
        
        column.setWidth(_width);
        column.setHidden(_hidden);
        
        column.setEditable(_editable);
        column.setSortable(_sortable);
        column.setAllowSortOnMultipleJoin(_allowSortOnMultipleJoin);
        column.setDefaultSorter(_defaultSorter);
        
        column.setRenderer(_renderer);
        column.setConverter(_converter);
    }
}
