/*
 *  Copyright 2024 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.solr;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.exception.ExceptionUtils;
import org.apache.solr.client.solrj.SolrClient;
import org.apache.solr.client.solrj.impl.BaseHttpSolrClient.RemoteSolrException;
import org.apache.solr.client.solrj.request.json.DirectJsonQueryRequest;
import org.apache.solr.client.solrj.response.QueryResponse;

import org.ametys.cms.content.indexing.solr.SolrIndexer;
import org.ametys.cms.search.solr.SearcherFactory;
import org.ametys.cms.search.solr.SolrClientProvider;
import org.ametys.core.ui.Callable;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Handler to execute Solr JSON queries
 */
public class SolrQueryHandler extends AbstractLogEnabled  implements Component, Serviceable, Initializable
{
    /** Avalon role. */
    public static final String ROLE = SolrQueryHandler.class.getName();
    
    /** The searcher factory. */
    protected SearcherFactory _searcherFactory;
    /** The solr client */
    protected SolrClient _solrClient;
    /** The Solr indexer */
    protected SolrIndexer _solrIndexer;
    /** The solr client provider */
    protected SolrClientProvider _solrClientProvider;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _searcherFactory = (SearcherFactory) manager.lookup(SearcherFactory.ROLE);
        _solrClientProvider = (SolrClientProvider) manager.lookup(SolrClientProvider.ROLE);
        _solrIndexer = (SolrIndexer) manager.lookup(SolrIndexer.ROLE);
    }
    
    public void initialize() throws Exception
    {
        _solrClient = _solrClientProvider.getReadClient();
    }
    
    /**
     * Execute a solr query.
     * @param query The query as a String representing a JSON.
     * @param coreName The solr core name
     * @return The result as a JSON String
     * @throws Exception If an error occurs
     */
    @Callable(rights = "CMS_Rights_Solr_Query", context = "/admin")
    public Map<String, Object> executeSolrQuery(String query, String coreName) throws Exception
    {
        DirectJsonQueryRequest queryRequest = new DirectJsonQueryRequest(query);
        
        try
        {
            QueryResponse response = queryRequest.process(_solrClient, _solrClientProvider.getCollectionName(coreName));
            return Map.of("result", response.jsonStr());
        }
        catch (RemoteSolrException e)
        {
            return Map.of("errorMessage", e.getMessage(),
                          "errorStackTrace", ExceptionUtils.getStackTrace(e));
        }
    }
    
    /**
     * Get the solr core names
     * @return The solr core names
     * @throws Exception If an error occurs
     */
    @Callable(rights = "CMS_Rights_Solr_Query", context = "/admin")
    public List<Map<String, String>> getSolrCoreNames() throws Exception
    {
        List<Map<String, String>> results = new ArrayList<>();
        Set<String> coreNames = _solrIndexer.getCoreNames();
        for (String coreName : coreNames)
        {
            results.add(Map.of("name", coreName));
        }
        return results;
    }
}
