/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.source;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceFactory;

/**
 * This factory looks for files in the current skin and fallback in the current plugin dir.
 * Use: content-view:[view]://path_to_file
 * Will first look in the skin of the current site in the sub-directory stylesheet/{contenttype} =&gt; skins/{skin}/stylesheets/{contenttype}/path_to_file
 * And if the file does not exist will search in plugin:{currentPluginName}://path_to_file 
 */
public class ContentViewSourceFactory extends AbstractLogEnabled implements SourceFactory, Contextualizable, Serviceable
{
    // regexp for location : content-view://<ctype>/<view>/<format>.<extension>
    private static final Pattern __SOURCE_PATTERN = Pattern.compile("^[^:]*+://([^/]+)/view-([^/]*)/([^/]+)\\.([^/]+)$");

    private Context _context;
    private ContentView _contentView;
    private ServiceManager _manager;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
    }
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }

    @Override
    public Source getSource(String location, Map parameters) throws IOException, MalformedURLException
    {
        // lazy initialization to prevent chicken/egg scenario on startup
        if (_contentView == null)
        {
            try
            {
                _contentView = (ContentView) _manager.lookup(ContentView.ROLE);
            }
            catch (ServiceException e)
            {
                throw new IllegalStateException("Exception while getting ContentView", e);
            }
        }
        
        Matcher m = __SOURCE_PATTERN.matcher(location);
        if (!m.matches())
        {
            throw new MalformedURLException("URI must be like protocol://<ctype>/view-<view>/<format>.<extension>. Location was '" + location + "'");
        }
        
        Request request = ContextHelper.getRequest(_context);
        String pluginName = request.getParameter("pluginName");
        if (pluginName == null)
        {
            pluginName = (String) request.getAttribute("pluginName");
        }
        
        String contentType = m.group(1);
        String view = m.group(2);
        String format = m.group(3);
        String extension = m.group(4);
        
        if (view == null || "".equals(view))
        {
            view = "main";
        }
        
        return _contentView.getSource(location, contentType, view, format, pluginName, extension);
    }

    @Override
    public void release(Source source)
    {
        // empty method
    }
}
