/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.source;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceNotFoundException;
import org.apache.excalibur.source.SourceResolver;

import org.ametys.cms.content.GetContentAction;
import org.ametys.cms.contenttype.ContentTypeDescriptor;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.contenttype.DynamicContentTypeDescriptorExtentionPoint;

/**
 * Default implementation of a {@link ContentView}
 * Will first look in directory context://WEB-INF/stylesheets/content/article/article-[view].[extension]
 * And if the file does not exist will search in plugin:[currentPluginName]://stylesheets/content/article/article-[view].[extension]
 */
public class DefaultContentView extends AbstractLogEnabled implements ContentView, Serviceable, Contextualizable
{
    /** The source resolver */
    protected SourceResolver _resolver;
    /** Helper for content types */
    protected ContentTypesHelper _contentTypesHelper;
    /** The content types extension point */
    protected ContentTypeExtensionPoint _contentTypeEP;
    /** The dynamic content type descriptior extension point */
    protected DynamicContentTypeDescriptorExtentionPoint _dynamicContentTypeDescriptorEP;
    
    
    /** Context to get full content type ID from request */
    protected Context _context;
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
        _contentTypeEP = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _dynamicContentTypeDescriptorEP = (DynamicContentTypeDescriptorExtentionPoint) manager.lookup(DynamicContentTypeDescriptorExtentionPoint.ROLE);
    }
    
    @Override
    public Source getSource(String location, String contentType, String view, String format, String pluginName, String extension) throws IOException
    {
        Request request = ContextHelper.getRequest(_context);
        String fullContentTypeId = (String) request.getAttribute(GetContentAction.RESULT_CONTENTTYPE);
        
        for (String s : _getSourceURIs(contentType, fullContentTypeId, view, format, pluginName, extension))
        {
            try
            {
                Source src = _resolver.resolveURI(s);
                if (!src.exists())
                {
                    if (getLogger().isDebugEnabled())
                    {
                        getLogger().debug("Failed to find a stylesheet at '" + s + "'.");
                    }
                }
                else
                {
                    return src;
                }
            }
            catch (IOException e)
            {
                if (getLogger().isDebugEnabled())
                {
                    getLogger().debug("Resolving protocol failed for resolving '" + s + "'.");
                }
            }
        }
        
        // should never occur because of the default stylesheet
        throw new SourceNotFoundException("Can't find a stylesheet for: " + location);
    }
    
    /**
     * Returns the ordered list of URI to be tested to find a stylesheet to render the Content.
     * @param contentTypeAlias the content type alias for rendering
     * @param contentTypeId the content type full ID
     * @param view the content view
     * @param format the format of the output (html, pdf, ...)
     * @param pluginName the current plugin name
     * @param extension the extension (xsl, xml, ...)
     * @return a list of URIs
     */
    protected List<String> _getSourceURIs(String contentTypeAlias, String contentTypeId, String view, String format, String pluginName, String extension)
    {
        String formatSuffix = "html".equals(format) ? "" : "2" + format;
        
        ArrayList<String> result = new ArrayList<>();

        ContentTypeDescriptor contentType = _contentTypeEP.getExtension(contentTypeId);
        if (contentType != null)
        {
            result.addAll(_getSourceURIsForContentType(contentTypeAlias, contentTypeId, view, formatSuffix, extension));
            result.addAll(_getSourceURIsForSuperTypes(contentTypeAlias, contentTypeId, view, formatSuffix, extension));
        }
        else
        {
            contentType = _dynamicContentTypeDescriptorEP.getExtension(contentTypeId);
            if (contentType != null)
            {
                result.addAll(_getSourceURIsForDynamicContentType(contentTypeAlias, view, formatSuffix, contentType.getPluginName(), extension));
            }
        }
        
        // then a default stylesheet
        result.addAll(_getDefaultSourceURIsForContentType(contentType, pluginName, formatSuffix, extension));
        
        if ("xsl".equals(extension))
        {
            // and finally a dynamically constructed XSL, based on the model
            result.add("cocoon://_plugins/cms/default-content/" + format + "." + extension + "?viewName=" + view);
        }
        
        return result;
    }

    /**
     * Get the URIs of default stylesheets
     * @param contentType The content type
     * @param pluginName The current plugin name
     * @param formatSuffix the format of the output (html, pdf, ...)
     * @param extension the extension (xsl, xml, ...)
     * @return a list of URIs
     */
    protected List<String> _getDefaultSourceURIsForContentType(ContentTypeDescriptor contentType, String pluginName, String formatSuffix, String extension)
    {
        List<String> result = new ArrayList<>();
        
        if (contentType != null && !contentType.getPluginName().equals(pluginName))
        {
            result.add("plugin:" + contentType.getPluginName() + "://stylesheets/default-content" + formatSuffix + "." + extension);
        }
        
        if (pluginName != null)
        {
            result.add("plugin:" + pluginName + "://stylesheets/default-content" + formatSuffix + "." + extension);
        }
        
        result.addAll(_getDefaultSourceURIs(pluginName, formatSuffix, extension));
        
        return result;
    }
    
    /**
     * Get the URIs of default stylesheets
     * @param pluginName The current plugin name
     * @param formatSuffix the format of the output (html, pdf, ...)
     * @param extension the extension (xsl, xml, ...)
     * @return a list of URIs
     */
    protected List<String> _getDefaultSourceURIs(String pluginName, String formatSuffix, String extension)
    {
        List<String> result = new ArrayList<>();
        
        if (!"cms".equals(pluginName))
        {
            result.add("plugin:cms://stylesheets/default-content" + formatSuffix + "." + extension);
        }
        return result;
    }
    
    /**
     * Get source for a given content type
     * @param contentTypeAlias the content type alias for rendering
     * @param contentTypeId the content type full ID
     * @param view the content view
     * @param formatSuffix the format of the output (html, pdf, ...)
     * @param extension the extension (xsl, xml, ...)
     * @return the list
     */
    protected List<String> _getSourceURIsForSuperTypes(String contentTypeAlias, String contentTypeId, String view, String formatSuffix, String extension)
    {
        ArrayList<String> result = new ArrayList<>();
        
        Pair<String[], String[]> supertypeIds = _contentTypesHelper.getSupertypeIds(contentTypeId);
        
        final String dynamicContentTypeId = _contentTypesHelper.getDynamicContentTypeId(supertypeIds.getLeft(), supertypeIds.getRight());
        if (dynamicContentTypeId != null)
        {
            final String dynamicContentTypePluginName = _contentTypesHelper.getDynamicContentTypePlugin(supertypeIds.getLeft(), supertypeIds.getRight());
            result.addAll(_getSourceURIsForDynamicContentType(dynamicContentTypeId, view, formatSuffix, dynamicContentTypePluginName, extension));
        }
        
        for (String superTypeId : supertypeIds.getLeft())
        {
            result.addAll(_getSourceURIsForContentType(superTypeId, superTypeId, view, formatSuffix, extension));
        }
        
        for (String superTypeId : supertypeIds.getLeft())
        {
            result.addAll(_getSourceURIsForSuperTypes(superTypeId, superTypeId, view, formatSuffix, extension));
        }
        
        return result;
    }
    
    /**
     * get source for a dynamic content type
     * @param contentTypeAlias the content type alias for rendering
     * @param view the content view
     * @param formatSuffix the format of the output (html, pdf, ...)
     * @param pluginName the plugin name
     * @param extension the extension (xsl, xml, ...)
     * @return the list
     */
    protected List<String> _getSourceURIsForDynamicContentType(String contentTypeAlias, String view, String formatSuffix, String pluginName, String extension)
    {
        ArrayList<String> result = new ArrayList<>();
        
        result.add("context://WEB-INF/param/content-types/_dynamic/" + pluginName + "/stylesheets/" + contentTypeAlias + "/" + contentTypeAlias + formatSuffix + "-" + view + "." + extension);
        result.add("context://WEB-INF/param/content-types/_dynamic/" + pluginName + "/stylesheets/" + contentTypeAlias + "/" + contentTypeAlias + formatSuffix + "." + extension);
        
        return result;
    }

    /**
     * get source for a non dynamic content type
     * @param contentTypeAlias the content type alias for rendering
     * @param contentTypeId the content type full ID
     * @param view the content view
     * @param formatSuffix the format of the output (html, pdf, ...)
     * @param extension the extension (xsl, xml, ...)
     * @return the list
     */
    protected List<String> _getSourceURIsForContentType(String contentTypeAlias, String contentTypeId, String view, String formatSuffix, String extension)
    {
        ArrayList<String> result = new ArrayList<>();
        
        String pluginName = _contentTypesHelper.getPluginName(contentTypeId);
        
        result.addAll(_getSourceURIsForContentTypeWithPlugin(pluginName, contentTypeAlias, contentTypeId, view, formatSuffix, extension));
        if (!contentTypeAlias.equals(contentTypeId))
        {
            result.addAll(_getSourceURIsForContentTypeWithPlugin(pluginName, contentTypeId, contentTypeId, view, formatSuffix, extension));
        }
        
        return result;
    }
    
    /**
     * Get source for a non dynamic content type, call _getSourceURIsForContentType() which call this method twice if necessary (with the alias and the real content type id).
     * @param pluginName the plugin name of the content type
     * @param contentTypeAlias the content type alias for rendering
     * @param contentTypeId the content type id used to build the URIs
     * @param view the content view
     * @param formatSuffix the format of the output (html, pdf, ...)
     * @param extension the extension (xsl, xml, ...)
     * @return the list
     */
    protected List<String> _getSourceURIsForContentTypeWithPlugin(String pluginName, String contentTypeAlias, String contentTypeId, String view, String formatSuffix, String extension)
    {
        ArrayList<String> result = new ArrayList<>();

        // first look in the filesystem (case of automatic content types)
        result.add("context://WEB-INF/param/content-types/" + pluginName + "/stylesheets/" + contentTypeAlias + "/" + contentTypeAlias + formatSuffix + "-" + view + "." + extension);
        result.add("context://WEB-INF/param/content-types/" + pluginName + "/stylesheets/" + contentTypeAlias + "/" + contentTypeAlias + formatSuffix + "." + extension);
        
        // then look in the plugin
        result.add("plugin:" + pluginName + "://stylesheets/content/" + contentTypeAlias + "/" + contentTypeAlias + formatSuffix + "-" + view + "." + extension);
        result.add("plugin:" + pluginName + "://stylesheets/content/" + contentTypeAlias + "/" + contentTypeAlias + formatSuffix + "." + extension);
        
        return result;
    }
}
