/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.statistics;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ContentQueryHelper;
import org.ametys.cms.repository.ContentTypeExpression;
import org.ametys.core.util.I18nUtils;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.query.expression.AndExpression;
import org.ametys.plugins.repository.query.expression.Expression;
import org.ametys.plugins.repository.query.expression.Expression.Operator;
import org.ametys.plugins.repository.query.expression.NotExpression;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.PluginAware;
import org.ametys.runtime.plugins.admin.statistics.Statistics;
import org.ametys.runtime.plugins.admin.statistics.StatisticsNode;
import org.ametys.runtime.plugins.admin.statistics.StatisticsProvider;
import org.ametys.runtime.plugins.admin.statistics.StatisticsValue;

/**
 * CMS statistics
 */
public class CMSStatisticsProvider implements StatisticsProvider, Serviceable, PluginAware
{
    private String _id;
    private AmetysObjectResolver _ametysResolver;
    private ContentTypeExtensionPoint _contentTypeExtensionPoint;
    private I18nUtils _i18nUtils;

    public void service(ServiceManager manager) throws ServiceException
    {
        _ametysResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
    }
    
    public void setPluginInfo(String pluginName, String featureName, String id)
    {
        _id = id;
    }
    
    public Statistics getStatistics()
    {
        return new StatisticsNode(
            _id,
            new I18nizableText("plugin.cms", "PLUGINS_CMS_STATISTICS_LABEL"),
            "ametysicon-desktop-archive-black",
            null,
            List.of(
                new StatisticsNode(
                    "contents",
                    new I18nizableText("plugin.cms", "PLUGINS_CMS_STATISTICS_CONTENTS_LABEL"),
                    "ametysicon-text70",
                    _countAllContents(null),
                    _getContentTypes(),
                    false
                )
            ),
            true
        );
    }
    
    private List<Statistics> _getContentTypes()
    {
        List<Statistics> stats = new ArrayList<>();
        
        List<Expression> notContentType = new ArrayList<>();
        
        for (String contentTypeId : _contentTypeExtensionPoint.getExtensionsIds())
        {
            ContentType contentType = _contentTypeExtensionPoint.getExtension(contentTypeId);
            
            if (!contentType.isAbstract())
            {
                stats.add(new StatisticsValue(
                    contentTypeId,
                    new I18nizableText(StringUtils.defaultString(_i18nUtils.translate(contentType.getLabel(), contentType.getId()))), // Translate now for sorting
                    StringUtils.defaultIfBlank(contentType.getIconGlyph(), "ametysicon-text61"),
                    _countAllContents(contentTypeId)
                ));
                
                notContentType.add(new NotExpression(new ContentTypeExpression(Operator.EQ, contentTypeId)));
            }
        }
        
        stats.sort(new Comparator<>() {
            public int compare(Statistics o1, Statistics o2)
            {
                return o1.getLabel().getLabel().compareTo(o2.getLabel().getLabel());
            }
        });
        
        String query = ContentQueryHelper.getContentXPathQuery(new AndExpression(notContentType.toArray(new NotExpression[notContentType.size()])));
        try (AmetysObjectIterable<Content> contents = _ametysResolver.query(query))
        {
            stats.add(0, new StatisticsValue(
                "unknown",
                new I18nizableText("plugin.cms", "PLUGINS_CMS_STATISTICS_CONTENTS_UNKNOWN_LABEL"),
                "ametysicon-file-extension-generic-unknown",
                contents.getSize()
            ));
        }
        
        return stats;
    }

    private long _countAllContents(String contentTypeId)
    {
        String query = ContentQueryHelper.getContentXPathQuery(contentTypeId != null ? new ContentTypeExpression(Operator.EQ, contentTypeId) : null);
        try (AmetysObjectIterable<Content> contents = _ametysResolver.query(query))
        {
            return contents.getSize();
        }
    }
}
