/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jcr.RepositoryException;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.commons.lang.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.cms.tag.jcr.AbstractJCRTagProvider;
import org.ametys.core.cocoon.JSonReader;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;

/**
 * SAX events for tags
 *
 */
public abstract class AbstractGetTagsAction extends ServiceableAction
{
    /** The tag provider extension point */
    protected AbstractTagProviderExtensionPoint<? extends Tag> _tagProviderExtPt;

    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    
    @SuppressWarnings("unchecked")
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _resolver = (AmetysObjectResolver) serviceManager.lookup(AmetysObjectResolver.ROLE);
        _tagProviderExtPt = (AbstractTagProviderExtensionPoint< ? extends Tag>) serviceManager.lookup(getTagProviderEPRole());
    }

    /**
     * Get the tag provider extension point role
     * @return the tag provider extension point role
     */
    public abstract String getTagProviderEPRole();
    
    @SuppressWarnings("unchecked")
    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws IOException, SAXException, ProcessingException
    {
        Map<String, Object> jsParameters = (Map<String, Object>) objectModel.get(ObjectModelHelper.PARENT_CONTEXT);
        
        String parentId = (String) jsParameters.get("node");
        String nodeName = (String) jsParameters.get("nodeName");
        boolean onlyCustomTags = (Boolean) jsParameters.get("onlyCustomTags");
        
        List<String> objectTargetIds = (List<String>) jsParameters.get("objectTargetIds");
        
        Map<String, Object> contextualParameters = (Map<String, Object>) jsParameters.get("contextualParameters");
        
        List<Map<String, Object>> tags = new ArrayList<>();

        if (StringUtils.isEmpty(parentId) || parentId.equals("tag-root"))
        {
            Set<String> extensionsIds = getProvidersIds(onlyCustomTags);

            for (String id : extensionsIds)
            {
                TagProvider<? extends Tag> tagProvider = _tagProviderExtPt.getExtension(id);
                tags.add(providerToJSON(tagProvider, contextualParameters, jsParameters));
            }
        }
        else if (nodeName.startsWith("provider_"))
        {
            String providerId = nodeName.substring("provider_".length());
            TagProvider<? extends Tag> tagProvider = _tagProviderExtPt.getExtension(providerId);
            
            if (tagProvider != null)
            {
                for (Tag tag : _getFilteredTags(tagProvider.getTags(contextualParameters), jsParameters).values())
                {
                    tags.add(tagToJSON(tag, tagProvider, objectTargetIds, jsParameters));
                }
            }
        }
        else
        {
            Set<String> extensionsIds = getProvidersIds(onlyCustomTags);
            for (String id : extensionsIds)
            {
                TagProvider<? extends Tag> tagProvider = _tagProviderExtPt.getExtension(id);
                if (tagProvider.hasTag(nodeName, contextualParameters))
                {
                    Tag parent = tagProvider.getTag(nodeName, contextualParameters);
                    for (Tag tag : _getFilteredTags(parent.getTags(), jsParameters).values())
                    {
                        tags.add(tagToJSON(tag, tagProvider, objectTargetIds, jsParameters));
                    }
                }
            }
        }

        Map<String, Object> result = new HashMap<>();
        result.put("children", tags);

        Request request = ObjectModelHelper.getRequest(objectModel);
        request.setAttribute(JSonReader.OBJECT_TO_READ, result);

        return EMPTY_MAP;
    }

    /**
     * Get filtered tags
     * @param tags the tags to filter
     * @param jsParameters the js parameters
     * @return the map of tags
     */
    protected Map<String, ? extends Tag> _getFilteredTags(Map<String, ? extends Tag> tags, Map<String, Object> jsParameters)
    {
        return tags;
    }
    
    /**
     * Get all providers IDs
     * @param onlyCustomTags If true, return only JCR providers IDs.
     * @return a set of providers IDs
     */
    protected Set<String> getProvidersIds(boolean onlyCustomTags)
    {
        if (!onlyCustomTags)
        {
            return _tagProviderExtPt.getExtensionsIds();
        }
        
        return new HashSet<>();
    }
    
    /**
     * Returns the tag providers' id
     * @return The id of tag providers
     */
    protected Set<String> getProvidersId()
    {
        return getProvidersIds(false);
    }

    /**
     * Get tag's properties to JSON format
     * @param tag The tag
     * @param tagProvider the tag provider
     * @param objectTargetIds the list of Ametys object ids. Can be empty
     * @param jsParameters the js parameters
     * @return The tag properties
     */
    protected Map<String, Object> tagToJSON(Tag tag, TagProvider<? extends Tag> tagProvider, List<String> objectTargetIds, Map<String, Object> jsParameters)
    {
        Map<String, Object> infos = new HashMap<>();

        infos.put("id", tag.getId());
        infos.put("name", tag.getName());
        infos.put("title", tag.getTitle());
        infos.put("description", tag.getDescription());
        
        if (tag instanceof ColorableTag colorableTag)
        {
            infos.put("isColorable", true);
            infos.put("color", colorableTag.getColor(true));
            infos.put("class",  colorableTag.getColorCSSClassPrefix());
        }
        
        boolean authorized = isUserAuthorized(tag, objectTargetIds, jsParameters);
        infos.put("authorized", authorized);
        infos.put("creationAllowed", isCreationAllowed(tagProvider, tag));
        
        if (tag.getTags().size() == 0)
        {
            infos.put("leaf", true);
        }
        
        return infos;
    }

    /**
     * Get tag provider's properties to JSON format
     * @param tagProvider the tag provider
     * @param contextualParameters The contextual parameters
     * @param jsParameters The js parameters
     * @return The tag providers properties
     * @throws AmetysRepositoryException If an error occurred
     */
    protected Map<String, Object> providerToJSON(TagProvider<? extends Tag> tagProvider, Map<String, Object> contextualParameters, Map<String, Object> jsParameters)
    {
        Map<String, Object> infos = new HashMap<>();

        if (tagProvider instanceof AbstractJCRTagProvider)
        {
            AbstractJCRTagProvider<? extends Tag> provider = (AbstractJCRTagProvider<? extends Tag>) tagProvider;
            try
            {
                infos.put("id", provider.getRootNode(contextualParameters).getId());
            }
            catch (AmetysRepositoryException e)
            {
                getLogger().error("Unable to get JCR tag provider root node", e);
            }
            catch (RepositoryException e)
            {
                getLogger().error("Unable to get JCR tag provider root node", e);
            }
        }
        else
        {
            infos.put("id", tagProvider.getId());
        }
        
        infos.put("type", "provider");
        infos.put("name", "provider_" + tagProvider.getId());
        infos.put("title", tagProvider.getLabel());
        infos.put("description", tagProvider.getDescription());
        infos.put("authorized", true); 
        infos.put("creationAllowed", isCreationAllowed(tagProvider, null));
        infos.put("iconSmall", "/plugins/cms/resources/img/tag/tags_16.png");
        infos.put("iconMedium", "/plugins/cms/resources/img/tag/tags_32.png");
        infos.put("iconLarge", "/plugins/cms/resources/img/tag/tags_50.png");

        if (_getFilteredTags(tagProvider.getTags(contextualParameters), jsParameters).size() == 0)
        {
            infos.put("leaf", true);
        }
        return infos;
    }
    
    /**
     * Test if a tag is visible to the current user.
     * @param tag the Tag object.
     * @param objectTargetIds the ids of current Ametys objects
     * @param jsParameters The js parameters
     * @return true if the user has access to the tag, false otherwise.
     */
    protected boolean isUserAuthorized(Tag tag, List<String> objectTargetIds, Map<String, Object> jsParameters)
    {
        return true;
    }
    
    /**
     * Determines if the current user can created tag
     * @param tagProvider The tag provider
     * @param tag The tag. Can be null.
     * @return true if creation is allowed
     */
    protected boolean isCreationAllowed(TagProvider<? extends Tag> tagProvider, Tag tag)
    {
        return true;
    }
}
