/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.logger.AbstractLogEnabled;

import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.PluginAware;


/**
 * Class representing a static tag provider. <br>
 * @param <T> the tag class
 */
public abstract class AbstractTagProvider<T extends Tag> extends AbstractLogEnabled implements TagProvider<T>, Configurable, PluginAware
{
    /** The plugin name */
    protected String _pluginName;
    /** The feature name */
    protected String _featureName;
    /** The id */
    protected String _id;
    /** The label */
    protected I18nizableText _label;
    /** The description */
    protected I18nizableText _description;
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        _label = configureLabel(configuration, "plugin." + _pluginName);
        _description = configureDescription(configuration, "plugin." + _pluginName);
    }
    
    @Override
    public String getId()
    {
        return _id;
    }

    @Override
    public I18nizableText getLabel()
    {
        return _label;
    }
    
    @Override
    public I18nizableText getDescription()
    {
        return _description;
    }

    /**
     * Configure label from the passed configuration
     * @param configuration The configuration
     * @param defaultCatalogue The default catalogue
     * @return The label
     * @throws ConfigurationException If an error occurred
     */
    protected I18nizableText configureLabel (Configuration configuration, String defaultCatalogue) throws ConfigurationException
    {
        Configuration labelConfiguration = configuration.getChild("label");
        
        if (labelConfiguration.getAttributeAsBoolean("i18n", false))
        {
            return new I18nizableText(defaultCatalogue, labelConfiguration.getValue(""));
        }
        else
        {
            return new I18nizableText(labelConfiguration.getValue(""));
        }
    }
    
    /**
     * Configure description from the passed configuration
     * @param configuration The configuration
     * @param defaultCatalogue The default catalogue
     * @return The description
     * @throws ConfigurationException If an error occurred
     */
    protected I18nizableText configureDescription (Configuration configuration, String defaultCatalogue) throws ConfigurationException
    {
        Configuration descConfiguration = configuration.getChild("description");
        
        if (descConfiguration.getAttributeAsBoolean("i18n", false))
        {
            return new I18nizableText(defaultCatalogue, descConfiguration.getValue(""));
        }
        else
        {
            return new I18nizableText(descConfiguration.getValue(""));
        }
    }
    
    @Override
    public void setPluginInfo(String pluginName, String featureName, String id)
    {
        _pluginName = pluginName;
        _featureName = featureName;
        _id = id;
    }
    
    /**
     * Get the plugin name
     * @return the plugin name
     */
    public String getPluginName()
    {
        return _pluginName;
    }
    
    public List<String> getCSSUrls(Map<String, Object> contextualParameters)
    {
        return new ArrayList<>();
    }
}
