/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag;

import java.util.Map;
import java.util.Set;

import org.ametys.runtime.plugin.component.AbstractThreadSafeComponentExtensionPoint;

/**
 * This class is in charge to load and initialize tag providers.
 * @param <T> The tag class
 */
public abstract class AbstractTagProviderExtensionPoint<T extends Tag> extends AbstractThreadSafeComponentExtensionPoint<TagProvider<T>>
{
    /**
     * Get the tag
     * @param name The tag's name
     * @param contextualParameters Contextual parameters
     * @return the tag or null if tag is not found
     */
    public T getTag(String name, Map<String, Object> contextualParameters)
    {
        Set<String> extensionsIds = getExtensionsIds();
        
        for (String id : extensionsIds)
        {
            TagProvider<T> tagProvider = getExtension(id);
            if (tagProvider.hasTag(name, contextualParameters))
            {
                return tagProvider.getTag(name, contextualParameters);
            }
        }
        
        return null;
    }
    
    /**
     * Check if the tag exists
     * @param name The tag's name
     * @param contextualParameters Contextual parameters
     * @return <code>true</code> if the tag is found, <code>false</code> otherwise
     */
    public boolean hasTag(String name, Map<String, Object> contextualParameters)
    {
        Set<String> extensionsIds = getExtensionsIds();
        
        for (String id : extensionsIds)
        {
            TagProvider<T> tagProvider = getExtension(id);
            if (tagProvider.hasTag(name, contextualParameters))
            {
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * Get the tags node name
     * @return the tags node name
     */
    public abstract String getTagsNodeName();
    
    /**
     * Get the tag node type
     * @return the tag node type
     */
    public abstract String getTagsNodeType();
}
