/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag;

import java.text.Normalizer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jcr.RepositoryException;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.tag.jcr.AbstractJCRTagProvider;
import org.ametys.cms.tag.jcr.AbstractJCRTagsDAO;
import org.ametys.cms.tag.jcr.JCRTag;
import org.ametys.core.ui.Callable;
import org.ametys.core.util.I18nUtils;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * DAO for manipulating tags
 */
public abstract class AbstractTagsDAO extends AbstractLogEnabled implements Serviceable, Component
{
    /** The tag provider extension point */
    protected AbstractTagProviderExtensionPoint<? extends Tag> _tagProviderExtPt;
    /** The I18n utils */
    protected I18nUtils _i18nUtils;
    
    @SuppressWarnings("unchecked")
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
        _tagProviderExtPt = (AbstractTagProviderExtensionPoint< ? extends Tag>) manager.lookup(getTagProviderEPRole());
    }
    
    /**
     * Get the tag provider extension point role
     * @return the tag provider extension point role
     */
    public abstract String getTagProviderEPRole();
    
    /**
     * Get the paths of given tags
     * @param tagNames The name of tags
     * @param contextualParameters Contextual parameters
     * @return {String} path
     */
    @Callable
    public List<String> getTagPaths(List<String> tagNames, Map<String, Object> contextualParameters)
    {
        List<String> paths = new ArrayList<>();
        if (tagNames != null)
        {
            for (String tagName : tagNames)
            {
                if (tagName.startsWith("provider_"))
                {
                    paths.add(tagName);
                }
                else
                {
                    String tagPath = getFullPath(tagName, contextualParameters);
                    if (tagPath != null)
                    {
                        paths.add(tagPath);
                    }
                    
                }
            }
        }
        
        return paths;
    }

    /**
     * Get the full path of a tag, with its provider
     * @param tagName The tag name
     * @param contextualParameters Contextual parameters
     * @return the full path of tag or null if not found
     */
    protected String getFullPath (String tagName, Map<String, Object> contextualParameters)
    {
        Set<String> tagProviders = _tagProviderExtPt.getExtensionsIds();
        
        for (String id : tagProviders)
        {
            TagProvider<? extends Tag> tagProvider = _tagProviderExtPt.getExtension(id);
            if (tagProvider.hasTag(tagName, contextualParameters))
            {
                Tag tag = tagProvider.getTag(tagName, contextualParameters);
                return "provider_" + tagProvider.getId() + getPath(tag);
            }
        }
        
        // Not found
        return null;
    }
    
    /**
     * Get a tag by its name
     * @param tagName The tag name
     * @param contextualParameters Contextual parameters
     * @return The tag or null if not found
     */
    public Tag getTag (String tagName, Map<String, Object> contextualParameters)
    {
        Set<String> tagProviders = _tagProviderExtPt.getExtensionsIds();
        
        for (String id : tagProviders)
        {
            TagProvider<? extends Tag> tagProvider = _tagProviderExtPt.getExtension(id);
            if (tagProvider.hasTag(tagName, contextualParameters))
            {
                return tagProvider.getTag(tagName, contextualParameters);
            }
        }
        return null;
    }
    
    /**
     * Get the path of a tag inside its provider.
     * The path is composed of tags name and '/' as a separator
     * @param tag The tag
     * @return The path
     */
    protected String getPath (Tag tag)
    {
        return tag == null ? "" : getPath(tag.getParent()) + "/" + tag.getName();
    }
    
    /**
     * Get the title of given tags
     * @param tagNames The name of tags
     * @param contextualParameters Contextual parameters
     * @return the labels
     */
    @Callable
    public Map<String, Object> getTagsTitle(List<String> tagNames, Map<String, Object> contextualParameters)
    {
        Map<String, Object> result = new HashMap<>();
        
        Set<String> tagProviders = _tagProviderExtPt.getExtensionsIds();
        
        List<I18nizableText> labels = new ArrayList<>();
        
        for (String tagName : tagNames)
        {
            if (tagName.startsWith("provider_"))
            {
                TagProvider<? extends Tag> tagProvider = _tagProviderExtPt.getExtension(tagName.substring("provider_".length()));
                labels.add(tagProvider.getLabel());
            }
            else
            {
                for (String id : tagProviders)
                {
                    TagProvider<? extends Tag> tagProvider = _tagProviderExtPt.getExtension(id);
                    if (tagProvider.hasTag(tagName, contextualParameters))
                    {
                        Tag tag = tagProvider.getTag(tagName, contextualParameters);
                        labels.add(tag.getTitle());
                    }
                }
            }
        }
        
        result.put("titles", labels);
        return result;
    }

    /**
     * Get the path of node which match filter regexp
     * @param value the value to match
     * @param contextualParameters Contextual parameters
     * @return the matching paths
     */
    @Callable
    public List<String> filterTagsByRegExp(String value, Map<String, Object> contextualParameters)
    {
        List<String> matchingPaths = new ArrayList<>();

        String toMatch = Normalizer.normalize(value.toLowerCase(), Normalizer.Form.NFD).replaceAll("[\\p{InCombiningDiacriticalMarks}]", "").trim();
        
        Set<String> extensionsIds = _tagProviderExtPt.getExtensionsIds();
        for (String id : extensionsIds)
        {
            TagProvider<? extends Tag> tagProvider = _tagProviderExtPt.getExtension(id);
            for (Tag tag : tagProvider.getTags(contextualParameters).values())
            {
                _getMatchingTag(toMatch, tagProvider, tag, matchingPaths);
            }
        }
        
        return matchingPaths;
    }

    /**
     * Get the path of node which match filter regexp
     * @param filter the value to match
     * @param tagNames the list of tag's name to search inside
     * @param contextualParameters Contextual parameters
     * @return the matching paths
     */
    @Callable
    public List<String> filterTagsFromListByRegExp(String filter, List<String> tagNames, Map<String, Object> contextualParameters)
    {
        if (StringUtils.isEmpty(filter))
        {
            return getTagPaths(tagNames, contextualParameters);
        }
        
        if (tagNames != null && tagNames.size() > 0)
        {
            List<String> matchingPaths = new ArrayList<>();
            
            String toMatch = Normalizer.normalize(filter.toLowerCase(), Normalizer.Form.NFD).replaceAll("[\\p{InCombiningDiacriticalMarks}]", "").trim();
            
            for (String tagName : tagNames)
            {
                Tag tag = getTag(tagName, contextualParameters);
                if (tag != null)
                {
                    String title = _i18nUtils.translate(tag.getTitle());
                    String normalizedName = Normalizer.normalize(title.toLowerCase(), Normalizer.Form.NFD).replaceAll("[\\p{InCombiningDiacriticalMarks}]", "");
                    if (normalizedName.contains(toMatch))
                    {
                        matchingPaths.add(getFullPath(tagName, contextualParameters));
                    }
                }
            }
            
            return matchingPaths;
        }
        else
        {
            return filterTagsByRegExp(filter, contextualParameters);
        }
    }

    /**
     * Get paths of tag which match filter regexp
     * @param value the value to match
     * @param tagProvider the tag provider
     * @param tag the current tag
     * @param matchingPaths the matching paths
     */
    private void _getMatchingTag(String value, TagProvider<? extends Tag> tagProvider, Tag tag, List<String> matchingPaths)
    {
        String title = _i18nUtils.translate(tag.getTitle());
        
        if (title != null)
        {
            String normalizedName = Normalizer.normalize(title.toLowerCase(), Normalizer.Form.NFD).replaceAll("[\\p{InCombiningDiacriticalMarks}]", "");
            
            if (normalizedName.contains(value))
            {
                matchingPaths.add("provider_" + tagProvider.getId() + getPath(tag));
            }
        }
        
        for (Tag child : tag.getTags().values())
        {
            _getMatchingTag(value, tagProvider, child, matchingPaths);
        }
    }
    
    /**
     * Test if tags exists
     * @param tagNames The tag names to test
     * @param onlyCustomTags If true, return only custom tags
     * @param otherParameters the other parameters
     * @param contextualParameters Contextual parameters
     * @return The list of tag names without the invalid values. 
     */
    @Callable
    public List<String> checkTags(List<String> tagNames, boolean onlyCustomTags, Map<String, Object> otherParameters, Map<String, Object> contextualParameters)
    {
        List<String> existingTagNames = new ArrayList<>();
        List<TagProvider<? extends Tag>> tagProviders = new ArrayList<>();
        
        if (onlyCustomTags)
        {
            tagProviders.addAll(getCustomTagProvider());
        }
        else
        {
            Set<String> extensionsIds = _tagProviderExtPt.getExtensionsIds();
            for (String id : extensionsIds)
            {
                tagProviders.add(_tagProviderExtPt.getExtension(id));
            }
        }
        
        for (String tagName : tagNames)
        {
            if (tagName.startsWith("provider_"))
            {
                String providerId = tagName.substring("provider_".length());
                TagProvider<? extends Tag> tagProvider = _tagProviderExtPt.getExtension(providerId);
                
                if (tagProvider != null)
                {
                    existingTagNames.add(tagName);
                }
            }
            else
            {
                for (TagProvider<? extends Tag> tagProvider : tagProviders)
                {
                    if (tagProvider.hasTag(tagName, contextualParameters))
                    {
                        String filteredTagName = _getFilteredTagName(tagProvider, tagName, otherParameters, contextualParameters);
                        if (StringUtils.isNotBlank(filteredTagName))
                        {
                            existingTagNames.add(filteredTagName);
                            break;
                        }
                    }
                }
            }
        }
        
        return existingTagNames;
    }
    
    /**
     * Get filtered tag name
     * @param tagProvider the tag provider
     * @param tagName the tag name
     * @param otherParameters the other parameters
     * @param contextualParameters the contextual parameters
     * @return the tag name if it match
     */
    protected String _getFilteredTagName(TagProvider<? extends Tag> tagProvider, String tagName, Map<String, Object> otherParameters, Map<String, Object> contextualParameters)
    {
        return tagName;
    }
    
    /**
     * Get the list of custom tag provider
     * @return the list of custom tag provider
     */
    protected abstract List<TagProvider<? extends Tag>> getCustomTagProvider();
    
    /**
     * Add tags from FO
     * @param tagNames The name of tags to add
     * @return The new tags descriptions
     */
    public List<Map<String, Object>> addTags(String[] tagNames)
    {
        try
        {
            List<Map<String, Object>> tags = new ArrayList<>();
    
            for (String tagName : tagNames)
            {
                String filteredName = Normalizer.normalize(tagName.toLowerCase(), Normalizer.Form.NFD).replaceAll("[\\p{InCombiningDiacriticalMarks}]", "").trim(); 
                filteredName = filteredName.replaceAll("œ", "oe").replaceAll("æ", "ae").replaceAll("[^a-z0-9]", "_").replaceAll("_+", "_");
    
                JCRTag tag = _getTagJCRDAO().addTag(null, filteredName.toUpperCase(), tagName, "", Collections.emptyMap(), Collections.emptyMap());
                tags.add(_tagToJSON(tag));
            }
            
            // Clear the request cache after tag added
            for (String id : _tagProviderExtPt.getExtensionsIds())
            {
                TagProvider tagProvider = _tagProviderExtPt.getExtension(id);
                if (tagProvider instanceof AbstractJCRTagProvider)
                {
                    ((AbstractJCRTagProvider) tagProvider).clearCache();
                }
            }
            
            return tags;
        }
        catch (RepositoryException e)
        {
            throw new IllegalStateException("Cannot create tags", e);
        }
    }
    
    /**
     * Get the tag JCR DAO
     * @return the tag JCR DAO
     */
    protected abstract AbstractJCRTagsDAO _getTagJCRDAO();
    
    /**
     * Get tags to json
     * @param tags the tags
     * @return the tags to json
     */
    protected List<Map<String, Object>> _tagsToJSON(Collection< ? extends Tag> tags)
    {
        List<Map<String, Object>> result = new ArrayList<>();

        for (Tag tag : tags)
        {
            result.add(tagToJSON(tag));
            
            Map<String, ? extends Tag> subTags = tag.getTags();
            if (subTags != null)
            {
                result.addAll(_tagsToJSON(subTags.values()));
            }
        }
        
        return result;
    }
    
    /**
     * JCR tag to json
     * @param tag the jcr tag
     * @return the jcr tag to json
     */
    protected Map<String, Object> _tagToJSON(JCRTag tag)
    {
        Map<String, Object> tagInfo = new HashMap<>();
        tagInfo.put("id", tag.getId());
        tagInfo.put("name", tag.getName());
        tagInfo.put("text", tag.getTitle());
        return tagInfo;
    }
    
    /**
     * Tag to json
     * @param tag the tag
     * @return the tag to json
     */
    public Map<String, Object> tagToJSON(Tag tag)
    {
        Map<String, Object> tagInfo = new HashMap<>();
        tagInfo.put("id", tag.getId());
        tagInfo.put("name", tag.getName());
        tagInfo.put("text", tag.getTitle());
        return tagInfo;
    }
}
