/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.color.AbstractColorsComponent;
import org.ametys.cms.repository.Content;
import org.ametys.cms.tag.CMSTag.TagVisibility;
import org.ametys.cms.tag.jcr.CMSJCRTagProvider;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObject;

/**
 * SAX events for tags
 */
public class GetCMSTagsAction extends AbstractGetTagsAction
{
    /** The rights manager. */
    protected RightManager _rightManager;
    
    /** The current user provider. */
    protected CurrentUserProvider _currentUserProvider;

    /** The tags colors component */
    protected AbstractColorsComponent _colorComponent;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _rightManager = (RightManager) serviceManager.lookup(RightManager.ROLE);
        _currentUserProvider = (CurrentUserProvider) serviceManager.lookup(CurrentUserProvider.ROLE);
        _colorComponent = (AbstractColorsComponent) serviceManager.lookup(TagColorsComponent.ROLE);
    }
    
    @Override
    public String getTagProviderEPRole()
    {
        return TagProviderExtensionPoint.ROLE;
    }

    @Override
    protected Map<String, ? extends Tag> _getFilteredTags(Map<String, ? extends Tag> tags, Map<String, Object> jsParameters)
    {
        Map<String, CMSTag> filteredTags = new HashMap<>();
        if (tags == null)
        {
            return filteredTags;
        }
        
        String targetType = (String) jsParameters.get("targetType");
        for (Entry<String, ? extends Tag> entry : tags.entrySet())
        {
            CMSTag tag = (CMSTag) entry.getValue();
            if (targetType == null || tag.getTarget().getName().equals(targetType))
            {
                filteredTags.put(entry.getKey(), tag);
            }
        }
        
        return filteredTags;
    }
    
    @Override
    @SuppressWarnings("unchecked")
    protected Set<String> getProvidersIds(boolean onlyCustomTags)
    {
        if (onlyCustomTags)
        {
            Set<String> extensionsIds = new HashSet<>();
            TagProvider<CMSTag> tagProvider = (TagProvider<CMSTag>) _tagProviderExtPt.getExtension(CMSJCRTagProvider.class.getName());
            extensionsIds.add(tagProvider.getId());
            return extensionsIds;
        }
        
        return _tagProviderExtPt.getExtensionsIds();
    }
    
    @Override
    protected Map<String, Object> tagToJSON(Tag tag, TagProvider<? extends Tag> tagProvider, List<String> objectTargetIds, Map<String, Object> jsParameters)
    {
        Map<String, Object> infos = super.tagToJSON(tag, tagProvider, objectTargetIds, jsParameters);

        CMSTag cmsTag = (CMSTag) tag;
        TagVisibility visibility = cmsTag.getVisibility();
        infos.put("visibility", visibility.toString());
        
        if (cmsTag instanceof ColorableTag colorableTag)
        {
            String color = colorableTag.getColor(true);
            infos.put("isColorable", true);
            infos.put("color", color);
            infos.put("class",  colorableTag.getColorCSSClassPrefix());
        }
        
        TagTargetType tagTargetType = cmsTag.getTarget();
        
        if (tagTargetType != null)
        {
            Map<String, Object> target2json = new HashMap<>();
            target2json.put("name", tagTargetType.getName());
            target2json.put("label", tagTargetType.getLabel());
            target2json.put("description", tagTargetType.getDescription());
            
            infos.put("target", target2json);
        }
        
        if (TagVisibility.PRIVATE.equals(visibility))
        {
            if (isUserAuthorized(tag, objectTargetIds, jsParameters))
            {
                infos.put("iconSmall", "/plugins/cms/resources/img/tag/tag_private_16.png");
                infos.put("iconMedium", "/plugins/cms/resources/img/tag/tag_private_32.png");
                infos.put("iconLarge", "/plugins/cms/resources/img/tag/tag_private_50.png");
            }
            else
            {
                infos.put("iconSmall", "/plugins/cms/resources/img/tag/tag_unauthorized_16.png");
                infos.put("iconMedium", "/plugins/cms/resources/img/tag/tag_unauthorized_32.png");
                infos.put("iconLarge", "/plugins/cms/resources/img/tag/tag_unauthorized_50.png");
            }    
        }
        else
        {
            infos.put("iconSmall", "/plugins/cms/resources/img/tag/tag_16.png");
            infos.put("iconMedium", "/plugins/cms/resources/img/tag/tag_32.png");
            infos.put("iconLarge", "/plugins/cms/resources/img/tag/tag_50.png");
        }
        
        return infos;
    }

    /**
     * Test if a tag is visible to the current user.
     * @param tag the Tag object.
     * @param content the current content (can be null).
     * @return true if the user has access to the tag, false otherwise.
     */
    protected boolean isUserAuthorized(Tag tag, Content content)
    {
        if (content != null)
        {
            if (isPrivate(tag))
            {
                return _rightManager.hasRight(_getCurrentUser(), "CMS_Rights_Content_Private_Tag", content) == RightResult.RIGHT_ALLOW;
            }
            else
            {
                return _rightManager.hasRight(_getCurrentUser(), "CMS_Rights_Content_Tag", content) == RightResult.RIGHT_ALLOW
                        || _rightManager.hasRight(_getCurrentUser(), "CMS_Rights_Content_Private_Tag", content) == RightResult.RIGHT_ALLOW;
            }
        }
        
        return true;
    }
    
    @Override
    protected boolean isUserAuthorized(Tag tag, List<String> objectTargetIds, Map<String, Object> jsParameters)
    {
        if (tag instanceof CMSTag)
        {
            if (objectTargetIds != null && objectTargetIds.size() > 0)
            {
                boolean hasRight = false;
                for (String objectId : objectTargetIds)
                {
                    AmetysObject ao = _resolver.resolveById(objectId);
                    
                    if (ao instanceof Content && ((CMSTag) tag).getTarget().getName().equals("CONTENT"))
                    {
                        hasRight = isUserAuthorized (tag, (Content) ao) || hasRight;
                    }
                }
                return hasRight;
            }
        }
        
        return true;
    }
    
    @Override
    protected boolean isCreationAllowed(TagProvider<? extends Tag> tagProvider, Tag tag)
    {
        if (!(tagProvider instanceof CMSJCRTagProvider))
        {
            return false;
        }
        return _rightManager.hasRight(_getCurrentUser(), "CMS_Rights_Tags_HandleTag", "/cms") == RightResult.RIGHT_ALLOW;
    }

    /**
     * Determines if a tag is private
     * @param tag the tag to test
     * @return true if a tag is private
     */
    protected boolean isPrivate (Tag tag)
    {
        return ((CMSTag) tag).getVisibility() == TagVisibility.PRIVATE;
    }
    
    /**
     * Provides the current user.
     * @return the current user which cannot be <code>null</code>.
     */
    protected UserIdentity _getCurrentUser()
    {
        return _currentUserProvider.getUser();
    }
}
