/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag;

import java.util.Collection;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.group.Group;
import org.ametys.core.group.GroupDirectoryContextHelper;
import org.ametys.core.group.GroupDirectoryDAO;
import org.ametys.core.group.GroupIdentity;
import org.ametys.core.group.GroupManager;
import org.ametys.core.group.directory.GroupDirectory;
import org.ametys.runtime.i18n.I18nizableText;


/**
 * Class representing a group tag provider. <br>
 * @param <T> the tag class
 */
public class GroupTagProvider<T extends CMSTag> extends AbstractTagProvider<CMSTag> implements Serviceable
{
    /** The Avalon role */
    public static final String ROLE = GroupTagProvider.class.getName();
    
    private static final String _GROUP_TAG_PREFIX = "group_";
    private static final String _GROUP_ID_SEPERATOR = "#";
    
    /** The group manager */
    protected GroupManager _groupManager;
    
    /** The group directory DAO */
    protected GroupDirectoryDAO _groupDirectoryDAO;
    
    /** The tag target type extension point */
    protected TagTargetTypeExtensionPoint _targetTypeEP;
    
    /** The group directory context helper */
    protected GroupDirectoryContextHelper _groupDirectoryContextHelper;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _groupManager = (GroupManager) manager.lookup(GroupManager.ROLE);
        _groupDirectoryDAO = (GroupDirectoryDAO) manager.lookup(GroupDirectoryDAO.ROLE);
        _targetTypeEP = (TagTargetTypeExtensionPoint) manager.lookup(TagTargetTypeExtensionPoint.ROLE);
        _groupDirectoryContextHelper = (GroupDirectoryContextHelper) manager.lookup(GroupDirectoryContextHelper.ROLE);
    }
    
    public Map<String, CMSTag> getTags(Map<String, Object> contextualParameters)
    {
        Set<String> contexts = _getContexts();
        return _groupDirectoryContextHelper.getGroupDirectoriesOnContexts(contexts)
                    .stream()
                    .map(id -> _groupDirectoryDAO.getGroupDirectory(id))
                    .map(this::_getCMSTag)
                    .collect(Collectors.toMap(t -> t.getId(), t -> t));
    }
    
    public CMSTag getTag(String tagName, Map<String, Object> contextualParameters)
    {
        if (!tagName.contains(_GROUP_TAG_PREFIX))
        {
            return null;
        }
        
        String tagNameWithoutPrefix = StringUtils.substringAfter(tagName, _GROUP_TAG_PREFIX);
        if (tagNameWithoutPrefix.contains(_GROUP_ID_SEPERATOR))
        {
            String groupDirectoryId = StringUtils.substringBefore(tagNameWithoutPrefix, _GROUP_ID_SEPERATOR);
            String groupId = StringUtils.substringAfter(tagNameWithoutPrefix, _GROUP_ID_SEPERATOR);
            Group group = _groupManager.getGroup(groupDirectoryId, groupId);
            return _getCMSTag(groupDirectoryId, group);
        }
        else
        {
            GroupDirectory groupDirectory = _groupDirectoryDAO.getGroupDirectory(tagNameWithoutPrefix);
            return _getCMSTag(groupDirectory);
        }
    }

    public Collection<CMSTag> getTags(String tagName, Map<String, Object> contextualParameters)
    {
        return _groupManager.getGroups(tagName)
                .stream()
                .map(g -> _getCMSTag(tagName, g))
                .collect(Collectors.toList());
    }

    public boolean hasTag(String tagName, Map<String, Object> contextualParameters)
    {
        if (!tagName.contains(_GROUP_TAG_PREFIX))
        {
            return false;
        }
        
        String tagNameWithoutPrefix = StringUtils.substringAfter(tagName, _GROUP_TAG_PREFIX);
        if (tagNameWithoutPrefix.contains(_GROUP_ID_SEPERATOR))
        {
            String groupDirectoryId = StringUtils.substringBefore(tagNameWithoutPrefix, _GROUP_ID_SEPERATOR);
            String groupId = StringUtils.substringAfter(tagNameWithoutPrefix, _GROUP_ID_SEPERATOR);
            return _groupManager.getGroup(groupDirectoryId, groupId) != null;
        }
        else
        {
            return _groupDirectoryDAO.getGroupDirectory(tagNameWithoutPrefix) != null;
        }
    }
    
    private CMSTag _getCMSTag(GroupDirectory groupDirectory)
    {
        String tagId = getGroupDirectoryTagId(groupDirectory.getId());
        CMSTag groupDirectoryTag =  new CMSTag(
                                        tagId, 
                                        tagId, 
                                        null, 
                                        groupDirectory.getLabel(),
                                        null,
                                        _targetTypeEP.getTagTargetType("CONTENT")
                                    );
        
        groupDirectoryTag.setTags(_groupManager.getGroups(groupDirectory)
                                        .stream()
                                        .map(g -> _getCMSTag(groupDirectoryTag, g))
                                        .collect(Collectors.toMap(t -> t.getId(), t -> t))
                                  );
        
        return groupDirectoryTag;
    }
    
    private CMSTag _getCMSTag(String groupDirectoryId, Group group)
    {
        GroupDirectory groupDirectory = _groupDirectoryDAO.getGroupDirectory(groupDirectoryId);
        CMSTag groupDirectoryTag = _getCMSTag(groupDirectory);
        return _getCMSTag(groupDirectoryTag, group);
    }
    
    private CMSTag _getCMSTag(CMSTag parentTag, Group group)
    {
        String tagId = getGroupTagId(parentTag.getId(), group.getIdentity());
        return new CMSTag(
                    tagId, 
                    tagId, 
                    parentTag, 
                    new I18nizableText(group.getLabel()),
                    null,
                    _targetTypeEP.getTagTargetType("CONTENT")
                );
    }
    
    /**
     * Get the set of context to get group directories
     * @return the set of context
     */
    protected Set<String> _getContexts()
    {
        return Set.of("/application");
    }
    
    /**
     * Get tag id if from group directory
     * @param groupDirectoryTagId the group directory id
     * @return the tag id
     */
    public String getGroupDirectoryTagId(String groupDirectoryTagId)
    {
        return _GROUP_TAG_PREFIX + groupDirectoryTagId;
    }
    
    /**
     * Get tag id if from the group and its group directory
     * @param groupDirectoryTagId the group directory tag id
     * @param groupIdentity the group identity
     * @return the tag id
     */
    public String getGroupTagId(String groupDirectoryTagId, GroupIdentity groupIdentity)
    {
        return groupDirectoryTagId + _GROUP_ID_SEPERATOR + groupIdentity.getId();
    }
}
