/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.cms.color.AbstractColorsComponent;
import org.ametys.cms.tag.CMSTag.TagVisibility;
import org.ametys.runtime.i18n.I18nizableText;


/**
 * Class representing a static tag provider.
 */
public class StaticTagProvider extends AbstractTagProvider<CMSTag> implements Serviceable
{
    /** List of statically defined ids */
    protected List<String> _localIds;

    /** The tags */
    protected Map<String, CMSTag> _tags;
    
    /** Target types */
    protected TagTargetTypeExtensionPoint _targetTypeEP;
    
    /** The colors component */
    protected AbstractColorsComponent _colorsComponent;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _targetTypeEP = (TagTargetTypeExtensionPoint) smanager.lookup(TagTargetTypeExtensionPoint.ROLE);
        _colorsComponent = (AbstractColorsComponent) smanager.lookup(TagColorsComponent.ROLE);
    }

    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        super.configure(configuration);

        _localIds = new ArrayList<>();
        _tags = configureTags(configuration, null, "plugin." + _pluginName);
    }
    
    /**
     * Returns the list of statically defined ids
     * @return the list of statically defined ids
     */
    public List<String> getLocalIds()
    {
        return _localIds;
    }
    
    @Override
    public Map<String, CMSTag> getTags(Map<String, Object> contextualParameters)
    {
        return _tags;
    }
    
    @Override
    public boolean hasTag(String tagID, Map<String, Object> contextualParameters)
    {
        return getTag(tagID, contextualParameters) != null;
    }
    
    @Override
    public CMSTag getTag(String tagID, Map<String, Object> contextualParameters)
    {
        Map<String, CMSTag> tags = getTags(contextualParameters);
        return tags != null ? _recursiveSearchTags(tags, tagID) : null;
    }
    
    private CMSTag _recursiveSearchTags(Map<String, CMSTag> tags, String tagID)
    {
        if (tags.containsKey(tagID))
        {
            return tags.get(tagID);
        }
        
        for (CMSTag child : tags.values())
        {
            CMSTag tag = _recursiveSearchTags(child.getTags(), tagID);
            if (tag != null)
            {
                return tag;
            }
        }
        
        return null;
    }

    @Override
    public Collection<CMSTag> getTags(String tagID, Map<String, Object> contextualParameters)
    {
        CMSTag tag = getTag(tagID, contextualParameters);
        return tag != null ? tag.getTags().values() : null;
    }
    
    /**
     * Configure tag from the passed configuration
     * @param configuration The configuration
     * @param parent The parent tag if any
     * @param defaultCatalogue The default catalogue for i18n
     * @return a Set of {@link Tag}
     * @throws ConfigurationException If an error occurred
     */
    protected Map<String, CMSTag> configureTags (Configuration configuration, CMSTag parent, String defaultCatalogue)  throws ConfigurationException
    {
        Map<String, CMSTag> tags = new HashMap<>();
        
        Configuration[] tagsConfiguration = configuration.getChildren("tag");
        for (Configuration tagConfiguration : tagsConfiguration)
        {
            String id = tagConfiguration.getAttribute("id");
            if (!Tag.NAME_PATTERN.matcher(id).matches())
            {
                throw new ConfigurationException("Invalid tag ID '" + id + "': it must match the pattern " + Tag.NAME_PATTERN.toString(), configuration);
            }
            if (_localIds.contains(id))
            {
                throw new ConfigurationException("A tag with the id '" + id + "' already exists", configuration);
            }
            _localIds.add(id);
            
            TagVisibility visibility = TagVisibility.PUBLIC;
            if (tagConfiguration.getAttribute("private", "").equals("true"))
            {
                visibility = TagVisibility.PRIVATE;
            }
            
            Configuration child = tagConfiguration.getChild("color");
            String color = child != null ? child.getValue(null) : null;
            
            String typeName = tagConfiguration.getAttribute("target", "CONTENT");
            TagTargetType targetType = _targetTypeEP.getTagTargetType(typeName);
            
            I18nizableText label = configureLabel (tagConfiguration, defaultCatalogue);
            I18nizableText description = configureDescription (tagConfiguration, defaultCatalogue);
            CMSTag cmsTag = new CMSTag(id, id, parent, label, description, visibility, targetType);
            CMSTag tag = new ColorableCMSTag(cmsTag, color, _colorsComponent);
            tags.put(id, tag);
            
            // Recursive configuration
            Map<String, CMSTag> childTags = configureTags(tagConfiguration, tag, defaultCatalogue);
            tag.setTags(childTags);
        }
        
        return tags;
    }
    
    @Override
    public List<String> getCSSUrls(Map<String, Object> contextualParameters)
    {
        List<String> cssUrls = super.getCSSUrls(contextualParameters);
        cssUrls.add("/plugins/cms/cms-tags.min.css");
        return cssUrls;
    }
}
