/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag;

import java.util.Map;
import java.util.regex.Pattern;

import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * This class represent a Tag
 */
public interface Tag
{
    /** The pattern a tag name has to respect. */
    public static final Pattern NAME_PATTERN = Pattern.compile("^[a-zA-Z][a-zA-Z0-9_]*$");
    
    /**
     * Get the tag unique id. 
     * @return The unique id
     */
    public String getId ();
    
    /**
     * Get the tag name. 
     * @return The tag name
     */
    public String getName ();
    
    /**
     * Get the parent name.
     * @return The parent name. 
     */
    public String getParentName();
    
    /**
     * Get the parent tag
     * @param <T> the tag class
     * @return The parent tag or null if the tag has no parent.
     */
    public <T extends Tag> T getParent();
    
    /**
     * Retrieves the title.
     * @return the title.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public I18nizableText getTitle();
    
    /**
     * Retrieves the description.
     * @return the description.
     */
    public I18nizableText getDescription();
    
    /**
     * Add a new child tag.
     * @param tag The tag to add
     */
    public void addTag(Tag tag);
    
    /**
     * Set the child tags
     * @param tags The Map of Tag to set
     */
    public void setTags(Map<String, ? extends Tag> tags);
    
    /**
     * Retrieves the Set of child tags
     * @return The child tags in a Map
     */
    public Map<String, ? extends Tag> getTags();
    
    /**
     * Try to get a child tag by its id
     * @param <T> the tag class
     * @param tagId The id of the child tag
     * @return The child tag or null if not existing
     */
    public <T extends Tag> T getTag(String tagId);
    
    /**
     * Determines if a tag exists
     * @param tagId The tag id
     * @return true if the tag exists
     */
    public boolean hasTag(String tagId);
}
