/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag;

import java.util.HashSet;
import java.util.Set;

/**
 * Helper class which provides useful methods to operate on {@link Tag} objects.
 */
public final class TagHelper
{
    private TagHelper()
    {
        // Hide the default constructor.
    }
    
    /**
     * Retrieves all descendants (direct and non-direct children) of the current tag.
     * @param tag the current tag
     * @param includeCurrent true to include the current tag itself
     * @return the set of descendant tags.
     */
    public static Set<Tag> getDescendants(Tag tag, boolean includeCurrent)
    {
        return _getDescendants(tag, includeCurrent, null);
    }
    
    private static Set<Tag> _getDescendants(Tag tag, boolean includeCurrent, Set<Tag> accumulator)
    {
        Set<Tag> tags = accumulator != null ? accumulator : new HashSet<>();
        
        if (tag != null)
        {
            if (includeCurrent)
            {
                tags.add(tag);
            }
            
            for (Tag child : tag.getTags().values())
            {
                _getDescendants(child, true, tags);
            }
        }
        
        return tags;
    }
    
    /**
     * Retrieves all name of the descendants (direct and non-direct children) of the current tag.
     * @param tag the current tag
     * @param includeCurrent true to include the current tag itself
     * @return the set of descendant tag names
     */
    public static Set<String> getDescendantNames(Tag tag, boolean includeCurrent)
    {
        Set<Tag> descendants = _getDescendants(tag, includeCurrent, null);
        Set<String> descendantNames = new HashSet<>();
        
        // Tag to tag names
        for (Tag descendant : descendants)
        {
            descendantNames.add(descendant.getName());
        }
        
        return descendantNames;
    }
    
    /**
     * Retrieves all ancestors of the current tag.
     * @param tag the current tag
     * @param includeCurrent true to include the current tag itself
     * @return the set of ancestor tags.
     */
    public static Set<Tag> getAncestors(Tag tag, boolean includeCurrent)
    {
        return _getAncestors(tag, includeCurrent, null);
    }
    
    private static Set<Tag> _getAncestors(Tag tag, boolean includeCurrent, Set<Tag> accumulator)
    {
        Set<Tag> tags = accumulator != null ? accumulator : new HashSet<>();
        
        if (tag != null)
        {
            if (includeCurrent)
            {
                tags.add(tag);
            }
            
            Tag parent = tag.getParent();
            if (parent != null)
            {
                _getAncestors(parent, true, tags);
            }
        }
        
        return tags;
    }
    
    /**
     * Retrieves all name of the ancestors of the current tag.
     * @param tag the current tag
     * @param includeCurrent true to include the current tag itself
     * @return the set of ancestor tag names
     */
    public static Set<String> getAncestorNames(Tag tag, boolean includeCurrent)
    {
        Set<Tag> ancestors = _getAncestors(tag, includeCurrent, null);
        Set<String> ancestorNames = new HashSet<>();
        
        // Tag to tag names
        for (Tag ancestor : ancestors)
        {
            ancestorNames.add(ancestor.getName());
        }
        
        return ancestorNames;
    }
}
