/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag.jcr;

import java.util.Map;

import javax.jcr.Node;
import javax.jcr.PathNotFoundException;
import javax.jcr.RepositoryException;

import org.apache.commons.lang.StringUtils;

import org.ametys.cms.color.AbstractColorsComponent;
import org.ametys.cms.tag.ColorableTag;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.RepositoryConstants;

/**
 * {@link JCRTag} for storing color tag informations.
 * @param <T> the tag factory
 */
public abstract class AbstractColorableJCRTag<T extends AbstractColorableCMSTagFactory> extends JCRTag
{
    /** Constants for color metadata. */
    private static final String __PROPERTY_COLOR = RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ":color";
    
    /**
     * Creates a {@link ColorableTag}.
     * @param node the node backing this {@link AmetysObject}.
     * @param parentPath the parent path in the Ametys hierarchy.
     * @param factory the factory which creates the AmetysObject.
     */
    public AbstractColorableJCRTag (Node node, String parentPath, T factory)
    {
        super(node, parentPath, factory);
    }
    
    /**
     * Retrieves the color type.
     * @return the color type.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public String getColor() throws AmetysRepositoryException
    {
        try
        {
            return getNode().getProperty(__PROPERTY_COLOR).getString();
        }
        catch (PathNotFoundException e)
        {
            return null;
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to get color property", e);
        }
    }
    
    /**
     * Set the color type.
     * @param color the color type.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setColor(String color) throws AmetysRepositoryException
    {
        try
        {
            getNode().setProperty(__PROPERTY_COLOR, color);
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to set color property", e);
        }
    }
    
    @Override
    public Map<String, Object> toJSON()
    {
        Map<String, Object> infos = super.toJSON();
        AbstractColorableCMSTagFactory factory = (AbstractColorableCMSTagFactory) _getFactory();
        infos.put("color", _getColor(factory));
        infos.put("isColorable", true);
        infos.put("class", factory.getColorsComponent().getColorCSSClassPrefix());
        
        return infos;
    }
    
    private String _getColor(AbstractColorableCMSTagFactory factory)
    {
        AbstractColorsComponent colorsComponent = factory.getColorsComponent();
        String color = getColor();
        Map<String, Map<String, String>> colors = colorsComponent.getColors();
        if (StringUtils.isNotEmpty(color) && colors.containsKey(color))
        {
            return color;
        }
        else
        {
            color = _getParentColor(getParent());
            return StringUtils.isNotBlank(color) ? color : colorsComponent.getDefaultKey();
        }
    }
    
    private String _getParentColor(AmetysObject parentTag)
    {
        if (parentTag == null || !(parentTag instanceof AbstractColorableJCRTag))
        {
            return null;
        }
        
        String color = ((AbstractColorableJCRTag) parentTag).getColor();
        return StringUtils.isNotBlank(color) ? color : _getParentColor(parentTag.getParent());
    }
}
