/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag.jcr;

import java.util.HashMap;
import java.util.Map;

import javax.jcr.RepositoryException;
import javax.jcr.Session;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;

import org.ametys.cms.tag.AbstractTagProvider;
import org.ametys.cms.tag.Tag;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.ModifiableTraversableAmetysObject;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.TraversableAmetysObject;
import org.ametys.plugins.repository.provider.AbstractRepository;
import org.ametys.plugins.repository.provider.JackrabbitRepository;

/**
 * Abstract class representing a jcr tag provider. <br>
 * @param <T> the tag class
 */
public abstract class AbstractJCRTagProvider<T extends Tag> extends AbstractTagProvider<T> implements Contextualizable, Serviceable
{
    /** The request attribute name with cache information */
    protected static final String CACHE_REQUEST_ATTRIBUTE = AbstractJCRTagProvider.class.getName() + "$cache";

    /** The context */
    protected Context _context;
    
    private JackrabbitRepository _repository;
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        super.configure(configuration);
        _id = configuration.getAttribute("id");
    }

    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _repository = (JackrabbitRepository) manager.lookup(AbstractRepository.ROLE);
    }
    
    /**
     * Get the Map of tags by their unique name
     * @param contextualParameters The contextual parameters
     * @return The non null cache
     * @throws RepositoryException If the cache cannot be filled
     */
    protected Map<String, T> _getCache(Map<String, Object> contextualParameters) throws RepositoryException
    {
        Request request = ContextHelper.getRequest(_context);
        if (request == null)
        {
            return new HashMap<>();
        }
        
        @SuppressWarnings("unchecked")
        Map<String, T> cache = (Map<String, T>) request.getAttribute(CACHE_REQUEST_ATTRIBUTE + "$" + getId());
        if (cache == null)
        {
            TraversableAmetysObject rootNode = getRootNode(contextualParameters);

            cache = new HashMap<>();
            _fillCache(rootNode, null, cache);

            request.setAttribute(CACHE_REQUEST_ATTRIBUTE + "$" + getId(), cache);
        }
        
        return cache;
    }
    
    /**
     * Clear the request cache
     */
    public void clearCache()
    {
        Request request = ContextHelper.getRequest(_context);
        if (request != null)
        {
            request.removeAttribute(CACHE_REQUEST_ATTRIBUTE + "$" + getId());
        }
    }
    
    @Override
    public Map<String, T> getTags(Map<String, Object> contextualParameters)
    {
        Map<String, T> tags = new HashMap<>();
        
        try
        {
            Map<String, T> cache = _getCache(contextualParameters);
            
            TraversableAmetysObject rootNode = getRootNode(contextualParameters);
            
            try (AmetysObjectIterable<AmetysObject> it = rootNode.getChildren())
            {
                for (AmetysObject object : it)
                {
                    if (object instanceof JCRTag)
                    {
                        T tag = cache.get(object.getName());
                        tags.put(object.getId(), tag);
                    }
                }
            }
        }
        catch (RepositoryException e)
        {
            getLogger().error("Unable to get JCR tags", e);
        }
        
        return tags;
    }
    
    @Override
    public T getTag(String tagName, Map<String, Object> contextualParameters)
    {
        try
        {
            Map<String, T> cache = _getCache(contextualParameters);
            return cache.get(tagName);
        }
        catch (RepositoryException e) 
        {
            getLogger().error("Unable to get JCR tags", e);
            return null;
        }
    }
    
    @Override
    public boolean hasTag(String tagName, Map<String, Object> contextualParameters)
    {
        try
        {
            Map<String, T> cache = _getCache(contextualParameters);
            return cache.containsKey(tagName);
        }
        catch (RepositoryException e) 
        {
            getLogger().error("Unable to get JCR tags", e);
            return false;
        }
    }
    
    /**
     * Fill cache 
     * @param parentTagNode The parent tag node
     * @param parentTag The parent tag
     * @param cache The cache
     * @throws RepositoryException If an error occurred
     */
    protected abstract void _fillCache(TraversableAmetysObject parentTagNode, T parentTag, Map<String, T> cache) throws RepositoryException;
    
    
    /**
     * Get the root node for tags
     * @param contextualParameters The contextual parameters
     * @return The root node
     * @throws RepositoryException if an error occurred
     */
    public ModifiableTraversableAmetysObject getRootNode (Map<String, Object> contextualParameters) throws RepositoryException
    {
        Session defaultSession = null;
        try
        {
            // Tags are not stored in archives workspace. And are the same in default and live ones
            // So we force to search in default workspace
            defaultSession = _repository.login(RepositoryConstants.DEFAULT_WORKSPACE);
            
            return _getRootNode(contextualParameters, defaultSession);
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException(e);
        }
        finally
        {
            if (defaultSession != null)
            {
                defaultSession.logout();
            }
        }
    }
    
    /**
     * Get the root node for tags
     * @param contextualParameters The contextual parameters
     * @param session the JCR Session to use to retrieve the root node.
     * @return The root node
     * @throws RepositoryException if an error occurred
     */
    protected abstract ModifiableTraversableAmetysObject _getRootNode (Map<String, Object> contextualParameters, Session session) throws RepositoryException;
    
    /**
     * Get a node or create it if not exist.
     * @param parentNode the parent node
     * @param nodeName the node name
     * @param nodeType the node type
     * @return the node
     * @throws AmetysRepositoryException if an error occurred
     */
    protected ModifiableTraversableAmetysObject _getOrCreateNode(ModifiableTraversableAmetysObject parentNode, String nodeName, String nodeType) throws AmetysRepositoryException
    {
        ModifiableTraversableAmetysObject definitionsNode;
        if (parentNode.hasChild(nodeName))
        {
            definitionsNode = parentNode.getChild(nodeName);
        }
        else
        {
            definitionsNode = parentNode.createChild(nodeName, nodeType);
            parentNode.saveChanges();
        }
        return definitionsNode;
    }
    
}
