/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag.jcr;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jcr.RepositoryException;
import javax.jcr.Session;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.ProcessingException;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.tag.Tag;
import org.ametys.cms.tag.TagHelper;
import org.ametys.cms.tag.TagProvider;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.ui.Callable;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.ModifiableTraversableAmetysObject;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.plugins.repository.jcr.JCRAmetysObject;
import org.ametys.plugins.repository.jcr.SimpleAmetysObject;

/**
 * Component for operations on JCR tags
 */
public abstract class AbstractJCRTagsDAO extends AbstractLogEnabled implements Serviceable, Component
{
    /** The ametys object resolver */
    protected AmetysObjectResolver _resolver;
    
    /** Observer manager. */
    protected ObservationManager _observationManager;
    
    /** The current user provider */
    protected CurrentUserProvider _currentUserProvider;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _observationManager = (ObservationManager) manager.lookup(ObservationManager.ROLE);
        _currentUserProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
    }
    
    /**
     * Create a new tag from client-side. User rights are checked.
     * @param parentId The id of parent tag
     * @param originalName The original name
     * @param title The tag's title
     * @param description The tag's description
     * @param otherParameters the other parameters
     * @param contextualParameters Contextual parameters transmitted by the environment.
     * @return The result map
     */
    @Callable
    public Map<String, Object> createTag (String parentId, String originalName, String title, String description, Map<String, Object> otherParameters, Map<String, Object> contextualParameters)
    {
        _checkUserRight();
        
        Map<String, Object> result = new HashMap<>();
        
        try
        {
            JCRTag jcrTag = addTag(parentId, originalName, title, description, otherParameters, contextualParameters);
            
            result.put("tagId", jcrTag.getId());
            result.put("tagName", jcrTag.getName());
            result.put("tagParentId", parentId);
        }
        catch (RepositoryException e)
        {
            getLogger().error("Unable to get JCR tag root node", e);
            result.put("message", "unknown-parent-tag");
        }
        catch (UnknownAmetysObjectException e)
        {
            getLogger().error("The tag '" + parentId + "' does not exist anymore", e);
            result.put("message", "unknown-parent-tag");
        }
        
        return result;
    }
    
    /**
     * Create a new tag regardless of user rights.
     * @param parentId The id of parent tag
     * @param originalName The original name
     * @param title The tag's title
     * @param description The tag's description
     * @param otherParameters the other parameters
     * @param contextualParameters Contextual parameters transmitted by the environment.
     * @return The created JCR tag
     * @throws UnknownAmetysObjectException if the parent does not exist
     * @throws RepositoryException if failed to create tag
     */
    public JCRTag addTag(String parentId, String originalName, String title, String description, Map<String, Object> otherParameters, Map<String, Object> contextualParameters) throws UnknownAmetysObjectException, RepositoryException
    {
        // Find an unique name
        String name = _findUniqueName (originalName, contextualParameters);
        
        JCRTag jcrTag = _createJCRTag(parentId, name, title, description, otherParameters, contextualParameters);
        
        // Notify observers that the tag has been added.
        Map<String, Object> eventParams = new HashMap<>();
        eventParams.put(ObservationConstants.ARGS_TAG_ID, jcrTag.getId());
        eventParams.put(ObservationConstants.ARGS_TAG_NAME, jcrTag.getName());
        _observationManager.notify(new Event(ObservationConstants.EVENT_TAG_ADDED, _currentUserProvider.getUser(), eventParams));
        
        return jcrTag;
    }
    
    /**
     * Deletes a JCR tag
     * @param tagId The tag's id
     * @param contextualParameters Contextual parameters transmitted by the environment.
     * @return the result map
     */
    @Callable
    public Map<String, Object> deleteTag (String tagId, Map<String, Object> contextualParameters)
    {
        _checkUserRight();
        
        Map<String, Object> result = new HashMap<>();
        
        try
        {
            JCRTag jcrTag = _resolver.resolveById(tagId);
            SimpleAmetysObject<?> parent = jcrTag.getParent();
            
            Map<String, Object> eventParams = new HashMap<>();
            eventParams.put(ObservationConstants.ARGS_TAG_ID, jcrTag.getId());
            eventParams.put(ObservationConstants.ARGS_TAG_NAME, jcrTag.getName());
            
            // descendant names for the event arguments
            Tag tag = _getTagFromName(jcrTag.getName(), contextualParameters);
            Collection<String> descendantNames = TagHelper.getDescendantNames(tag, true);
            eventParams.put("tag.descendantnames", descendantNames);
            
            result.put("id", jcrTag.getId());
            result.put("name", jcrTag.getName());
            result.put("title", jcrTag.getTitle());
            
            jcrTag.remove();
            parent.saveChanges();
            
            // Notify observers that the tag has been deleted.
            _observationManager.notify(new Event(ObservationConstants.EVENT_TAG_DELETED, _currentUserProvider.getUser(), eventParams));
        }
        catch (UnknownAmetysObjectException e)
        {
            getLogger().error("Unable to delete tag : the tag '" + tagId + "' does not exist anymore", e);
            result.put("message", "unknown-tag");
        }
        
        return result;
    }
    
    /**
     * Updates a JCR tag
     * @param tagId The tag's id
     * @param title The tag's title
     * @param description The tag's description
     * @param otherParameters the other parameters
     * @return The result map
     */
    @Callable
    public Map<String, Object> updateTag (String tagId, String title, String description, Map<String, Object> otherParameters)
    {
        _checkUserRight();
        
        Map<String, Object> result = new HashMap<>();
        
        try
        {
            JCRTag jcrTag = _updateJCRTag(tagId, title, description, otherParameters);
            
            result.put("id", jcrTag.getId());
            result.put("name", jcrTag.getName());
            result.put("title", title);
            
            // Notify observers that the tag has been modified.
            
            Map<String, Object> eventParams = new HashMap<>();
            eventParams.put(ObservationConstants.ARGS_TAG_ID, jcrTag.getId());
            eventParams.put(ObservationConstants.ARGS_TAG_NAME, jcrTag.getName());
            _observationManager.notify(new Event(ObservationConstants.EVENT_TAG_UPDATED, _currentUserProvider.getUser(), eventParams));
        }
        catch (UnknownAmetysObjectException e)
        {
            getLogger().error("Unable to update tag : the tag '" + tagId + "' does not exist anymore", e);
            result.put("message", "unknown-tag");
        }
               
        return result;
    }
    
    /**
     * Move a JCR tag
     * @param targetId The tag where to move to
     * @param ids The ids of tag to move
     * @return the result map
     * @throws ProcessingException If an error occurred
     */
    @Callable
    public Map<String, Object> moveTags(String targetId, List<String> ids) throws ProcessingException
    {
        _checkUserRight();
        
        Map<String, Object> result = new HashMap<>();
        List<Map<String, Object>> movedTags = new ArrayList<>();

        JCRAmetysObject target = _resolver.resolveById(targetId);
        String targetPath = "";
        try
        {
            targetPath = target.getNode().getPath();

            for (int i = 0; i < ids.size(); i++)
            {
                JCRAmetysObject tag = (JCRAmetysObject) _resolver.resolveById(ids.get(i));
                if (!tag.getNode().getPath().equals(targetPath + "/" + tag.getNode().getName()))
                {
                    Map<String, Object> eventParams = new HashMap<>();
                    eventParams.put("id", tag.getId());
                    eventParams.put("oldPath", tag.getNode().getPath());
                    eventParams.put("path", targetPath + "/" + tag.getNode().getName());
                    
                    Session session = tag.getNode().getSession();
                    session.move(tag.getNode().getPath(), targetPath + "/" + tag.getNode().getName());
                    
                    session.save();
                    
                    Map<String, Object> tagProperties = new HashMap<>();
                    tagProperties.put("id", tag.getId());
                    tagProperties.put("name", tag.getName());
                    movedTags.add(tagProperties);
                    
                    _observationManager.notify(new Event(ObservationConstants.EVENT_TAG_MOVED, _currentUserProvider.getUser(), eventParams));
                }
            }
        }
        catch (RepositoryException e)
        {
            getLogger().error("Unable to get JCR tag node path", e);
            throw new ProcessingException("Unable to get JCR tag node path", e);
        }
        
        if (movedTags.size() > 0)
        {
            result.put("movedTags", movedTags);
        }
        
        result.put("target", targetId);
    
        return result;
    }
    
    /**
     * Get the root node for tags
     * @param tagProviderId The tag provider id
     * @param contextualParameters Contextual parameters transmitted by the environment.
     * @return The root node in key "id"
     * @throws ProcessingException If an error occurred in the repository
     */
    @Callable
    public Map<String, Object> getTagRootNode (String tagProviderId, Map<String, Object> contextualParameters) throws ProcessingException
    {
        Map<String, Object> result = new HashMap<>();
        
        try
        {
            result.put("id", _getTagRootObject(tagProviderId, contextualParameters).getId());
        }
        catch (RepositoryException e)
        {
            getLogger().error("Unable to get JCR tag root node", e);
            throw new ProcessingException("Unable to get JCR tag root node", e);
        }
        
        return result;
    }
    
    /**
     * Find a JCR tag and return its data
     * @param tagId The tag's id
     * @return The result map
     */
    @Callable
    public Map<String, Object> getTag(String tagId)
    {
        Map<String, Object> result = new HashMap<>();
        
        try
        {
            JCRTag jcrTag = _resolver.resolveById(tagId);
            result = jcrTag.toJSON();
            
            if (jcrTag.getChildren().iterator().hasNext())
            {
                result.put("leaf", true);
            }
        }
        catch (UnknownAmetysObjectException e)
        {
            getLogger().error("Unable to get tag : the tag '" + tagId + "' does not exist anymore", e);
            result.put("message", "unknown-tag");
        }
        
        return result;
    }

    /**
     * Get the tag root node object
     * @param tagProviderId The tag provider id
     * @param contextualParameters Contextual parameters transmitted by the environment.
     * @return The tag root node object
     * @throws RepositoryException If an error occurred in the repository
     */
    public abstract ModifiableTraversableAmetysObject _getTagRootObject (String tagProviderId, Map<String, Object> contextualParameters) throws RepositoryException;
    
    /**
     * Check if the user right to access the feature
     * @throws IllegalStateException if the user has no right
     */
    protected abstract void _checkUserRight () throws IllegalStateException;
    
    /**
     * Get the tag from the name
     * @param name the name
     * @param contextualParameters the contextual parameters
     * @return the tag
     */
    protected abstract Tag _getTagFromName(String name, Map<String, Object> contextualParameters);
    
    /**
     * Get all tag's providers
     * @return the providers
     */
    protected abstract Set<TagProvider<? extends Tag>> _getTagProviders ();
    
    /**
     * Create a JCR tag under his parent
     * @param parentId the parent id
     * @param name the name
     * @param title the title
     * @param description the description
     * @param otherParameters the other parameters
     * @param contextualParameters Contextual parameters transmitted by the environment.
     * @return the created JCR tag
     * @throws RepositoryException if an error occurred
     */
    protected abstract JCRTag _createJCRTag(String parentId, String name, String title, String description, Map<String, Object> otherParameters, Map<String, Object> contextualParameters) throws RepositoryException;
    
    /**
     * Update a JCR tag
     * @param tagId the tag id to update
     * @param title the title
     * @param description the description
     * @param otherParameters the other parameters
     * @return return the updated JCR tag
     * @throws UnknownAmetysObjectException if an error occurred
     */
    protected abstract JCRTag _updateJCRTag(String tagId, String title, String description, Map<String, Object> otherParameters) throws UnknownAmetysObjectException;
    
    
    /**
     * Find a unique name for the tag
     * @param originalName The requested name
     * @param contextualParameters Contextual parameters transmitted by the environment.
     * @return A unique name
     */
    protected String _findUniqueName(String originalName, Map<String, Object> contextualParameters)
    {
        Set<TagProvider<? extends Tag>> providers = _getTagProviders();
        
        // Find an unique name
        int index = 2;
        String name = originalName;
        while (_hasTag(providers, name, contextualParameters))
        {
            name = originalName + "_" + (index++);
        }
        
        return name;
    }
    
    /**
     * Determines if a tag with given name already exists
     * @param providers The tag providers
     * @param name The name of the tag
     * @param contextualParameters Contextual parameters transmitted by the environment.
     * @return true if the tag exists
     */
    protected boolean _hasTag(Set<TagProvider<? extends Tag>> providers, String name, Map<String, Object> contextualParameters)
    {
        for (TagProvider<? extends Tag> provider : providers)
        {
            if (provider.hasTag(name, contextualParameters))
            {
                return true;
            }
        }
        return false;
    }
}
