/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag.jcr;

import java.util.HashMap;
import java.util.Map;

import javax.jcr.Node;
import javax.jcr.PathNotFoundException;
import javax.jcr.RepositoryException;

import org.ametys.cms.tag.CMSTag;
import org.ametys.cms.tag.CMSTag.TagVisibility;
import org.ametys.cms.tag.TagTargetType;
import org.ametys.cms.tag.TagTargetTypeExtensionPoint;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.RepositoryConstants;

/**
 * {@link JCRTag} for storing tag informations.
 */
public class CMSJCRTag extends AbstractColorableJCRTag<CMSTagFactory>
{
    /** Constants for visibility metadata. */
    private static final String __METADATA_VISIBILITY = RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ":visibility";
    /** Constants for type metadata. */
    private static final String __METADATA_TARGET = RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ":target";
    
    /**
     * Creates a {@link CMSTag}.
     * @param node the node backing this {@link AmetysObject}.
     * @param parentPath the parent path in the Ametys hierarchy.
     * @param factory the {@link TagFactory} which creates the AmetysObject.
     */
    public CMSJCRTag (Node node, String parentPath, CMSTagFactory factory)
    {
        super(node, parentPath, factory);
    }
    
    /**
     * Retrieves the visibility type.
     * @return the visibility type.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public TagVisibility getVisibility() throws AmetysRepositoryException
    {
        try
        {
            return TagVisibility.valueOf(getNode().getProperty(__METADATA_VISIBILITY).getString());
        }
        catch (PathNotFoundException e)
        {
            return TagVisibility.PUBLIC;
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to get visibility property", e);
        }
    }
    
    /**
     * Set the visibility type.
     * @param visibility the visibility type.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setVisibility(TagVisibility visibility) throws AmetysRepositoryException
    {
        try
        {
            getNode().setProperty(__METADATA_VISIBILITY, visibility.name());
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to set visibility property", e);
        }
    }
    
    /**
     * Retrieves the target type.
     * @return the target type.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public String getTargetType() throws AmetysRepositoryException
    {
        try
        {
            return getNode().getProperty(__METADATA_TARGET).getString();
        }
        catch (PathNotFoundException e)
        {
            return null;
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to get type property", e);
        }
    }
    
    /**
     * Set the target type.
     * @param target the target type.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setTargetType(TagTargetType target) throws AmetysRepositoryException
    {
        try
        {
            getNode().setProperty(__METADATA_TARGET, target.getName());
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to set type property", e);
        }
    }
    
    @Override
    public Map<String, Object> toJSON()
    {
        Map<String, Object> infos = super.toJSON();
        
        CMSTagFactory factory = (CMSTagFactory) _getFactory();
        TagTargetTypeExtensionPoint targetTypeEP = factory.getTagProviderEP();
        
        String targetType = getTargetType();
        if (targetType != null)
        {
            Map<String, Object> target2json = new HashMap<>();
            TagTargetType tagTargetType = targetTypeEP.getTagTargetType(targetType);
            target2json.put("name", tagTargetType.getName());
            target2json.put("label", tagTargetType.getLabel());
            target2json.put("description", tagTargetType.getDescription());
            
            infos.put("target-info", target2json);
            infos.put("target", tagTargetType.getName());
        }
        
        TagVisibility visibility = getVisibility();
        infos.put("visibility", visibility.toString());
        if (TagVisibility.PRIVATE.equals(visibility))
        {
            infos.put("iconSmall", "/plugins/cms/resources/img/tag/tag_private_16.png");
            infos.put("iconMedium", "/plugins/cms/resources/img/tag/tag_private_32.png");
            infos.put("iconLarge", "/plugins/cms/resources/img/tag/tag_private_50.png");
        }
        else
        {
            infos.put("iconSmall", "/plugins/cms/resources/img/tag/tag_16.png");
            infos.put("iconMedium", "/plugins/cms/resources/img/tag/tag_32.png");
            infos.put("iconLarge", "/plugins/cms/resources/img/tag/tag_50.png");
        }
        
        return infos;
    }
}
