/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag.jcr;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import javax.jcr.RepositoryException;
import javax.jcr.Session;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.color.AbstractColorsComponent;
import org.ametys.cms.tag.CMSTag;
import org.ametys.cms.tag.ColorableCMSTag;
import org.ametys.cms.tag.TagColorsComponent;
import org.ametys.cms.tag.TagProviderExtensionPoint;
import org.ametys.cms.tag.TagTargetType;
import org.ametys.cms.tag.TagTargetTypeExtensionPoint;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.ModifiableTraversableAmetysObject;
import org.ametys.plugins.repository.TraversableAmetysObject;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Class representing a jcr tag provider. <br>
 */
public class CMSJCRTagProvider extends AbstractJCRTagProvider<CMSTag>
{
    /** Constant for plugin node name */ 
    public static final String PLUGIN_NODE_NAME = "cms";
    
    /** The tag target type extension point */
    protected TagTargetTypeExtensionPoint _targetTypeEP;
    
    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    
    /** The tag provider extension point */
    protected TagProviderExtensionPoint _tagProviderEP;
    
    /** The colors component */
    protected AbstractColorsComponent _colorsComponent;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _targetTypeEP = (TagTargetTypeExtensionPoint) smanager.lookup(TagTargetTypeExtensionPoint.ROLE);
        _tagProviderEP = (TagProviderExtensionPoint) smanager.lookup(TagProviderExtensionPoint.ROLE);
        _colorsComponent = (AbstractColorsComponent) smanager.lookup(TagColorsComponent.ROLE);
    }
    
    @Override
    protected void _fillCache(TraversableAmetysObject parentTagNode, CMSTag parentTag, Map<String, CMSTag> cache) throws RepositoryException
    {
        for (AmetysObject child : parentTagNode.getChildren())
        {
            if (child instanceof CMSJCRTag)
            {
                CMSJCRTag jcrTag = (CMSJCRTag) child;
                TagTargetType targeType = _targetTypeEP.getTagTargetType(jcrTag.getTargetType());
                CMSTag cmsTag = new CMSTag(jcrTag.getId(), jcrTag.getName(), parentTag, new I18nizableText(jcrTag.getTitle()), new I18nizableText(jcrTag.getDescription()), jcrTag.getVisibility(), targeType);
                CMSTag tag = new ColorableCMSTag(cmsTag, jcrTag.getColor(), _colorsComponent);
                cache.put(child.getName(), tag);
                
                if (parentTag != null)
                {
                    parentTag.addTag(tag);
                }
                
                _fillCache(jcrTag, tag, cache);
            }
        }
    }
    
    @Override
    protected ModifiableTraversableAmetysObject _getRootNode (Map<String, Object> contextualParameters, Session session) throws RepositoryException
    {
        try
        {
            ModifiableTraversableAmetysObject pluginsNode = _resolver.resolveByPath("/ametys:plugins", session);
            
            ModifiableTraversableAmetysObject pluginNode = _getOrCreateNode(pluginsNode, PLUGIN_NODE_NAME, "ametys:unstructured");
            
            return _getOrCreateNode(pluginNode, _tagProviderEP.getTagsNodeName(), _tagProviderEP.getTagsNodeType());
        }
        catch (AmetysRepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to get the JCR tags root node", e);
        }
    }
    
    @Override
    public Collection<CMSTag> getTags(String tagID, Map<String, Object> contextualParameters)
    {
        CMSTag tag = getTag(tagID, contextualParameters);
        return tag != null ? tag.getTags().values() : null;
    }
    
    @Override
    public List<String> getCSSUrls(Map<String, Object> contextualParameters)
    {
        List<String> cssUrls = super.getCSSUrls(contextualParameters);
        cssUrls.add("/plugins/cms/cms-tags.min.css");
        return cssUrls;
    }
}
