/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag.jcr;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jcr.RepositoryException;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.commons.lang.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.cms.tag.AbstractTagProviderExtensionPoint;
import org.ametys.cms.tag.Tag;
import org.ametys.core.cocoon.JSonReader;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.TraversableAmetysObject;

/**
 * This action returns the JCR sub-tags of a given JCR node.
 */
public class GetJCRTagsAction extends ServiceableAction
{
    /** The tag provider extension point */
    protected AbstractTagProviderExtensionPoint<? extends Tag> _tagProviderExtPt;

    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;

    /** The service manager */
    protected ServiceManager _serviceManager;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _serviceManager = serviceManager;
        _resolver = (AmetysObjectResolver) serviceManager.lookup(AmetysObjectResolver.ROLE);
    }

    @SuppressWarnings("unchecked")
    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws IOException, SAXException, ProcessingException
    {
        Map<String, Object> jsParameters = (Map<String, Object>) objectModel.get(ObjectModelHelper.PARENT_CONTEXT);
        
        String parentId = (String) jsParameters.get("node");
        String tagProviderId = (String) jsParameters.get("tagProviderId");
        String tagProviderEPId = (String) jsParameters.get("tagProviderEPId");
        
        Map<String, Object> contextualParameters = (Map<String, Object>) jsParameters.get("contextualParameters");

        try
        {
            _tagProviderExtPt = (AbstractTagProviderExtensionPoint<? extends Tag>) _serviceManager.lookup(tagProviderEPId);
        }
        catch (ServiceException e1)
        {
            throw new IllegalArgumentException("Can't have a tag provider extension point with id " + tagProviderEPId, e1);
        }
        
        List<Map<String, Object>> tags = new ArrayList<>();

        try
        {
            if (StringUtils.isEmpty(parentId) || parentId.startsWith(_tagProviderExtPt.getTagsNodeName() + "://"))
            {
                AbstractJCRTagProvider<? extends Tag> provider = (AbstractJCRTagProvider< ? extends Tag>) _tagProviderExtPt.getExtension(tagProviderId); 

                TraversableAmetysObject rootNode = provider.getRootNode(contextualParameters);

                try (AmetysObjectIterable<AmetysObject> it = rootNode.getChildren())
                {
                    for (AmetysObject object : it)
                    {
                        if (object instanceof JCRTag)
                        {
                            tags.add(((JCRTag) object).toJSON());
                        }
                    }
                }
            }
            else
            {
                JCRTag jcrTag = _resolver.resolveById(parentId);

                try (AmetysObjectIterable< ? > objects = jcrTag.getChildren())
                {
                    for (AmetysObject object : objects)
                    {
                        if (object instanceof JCRTag)
                        {
                            tags.add(((JCRTag) object).toJSON());
                        }
                    }
                }
            }
        }
        catch (RepositoryException e)
        {
            getLogger().error("Unable to get JCR tag root node", e);
        }

        Map<String, Object> result = new HashMap<>();
        result.put("children", tags);

        Request request = ObjectModelHelper.getRequest(objectModel);
        request.setAttribute(JSonReader.OBJECT_TO_READ, result);

        return EMPTY_MAP;
    }
}
