/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag.jcr;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.xml.parsers.SAXParserFactory;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfigurationBuilder;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.cocoon.servlet.multipart.Part;
import org.apache.cocoon.servlet.multipart.PartOnDisk;
import org.apache.cocoon.servlet.multipart.RejectedPart;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.XMLReader;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.color.AbstractColorsComponent;
import org.ametys.cms.tag.CMSTag;
import org.ametys.cms.tag.ColorableCMSTag;
import org.ametys.cms.tag.CMSTag.TagVisibility;
import org.ametys.cms.tag.Tag;
import org.ametys.cms.tag.TagColorsComponent;
import org.ametys.cms.tag.TagProvider;
import org.ametys.cms.tag.TagProviderExtensionPoint;
import org.ametys.cms.tag.TagTargetType;
import org.ametys.cms.tag.TagTargetTypeExtensionPoint;
import org.ametys.core.cocoon.JSonReader;
import org.ametys.core.observation.AbstractNotifierAction;
import org.ametys.core.observation.Event;
import org.ametys.core.util.JSONUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.TraversableAmetysObject;
import org.ametys.plugins.repository.jcr.DefaultTraversableAmetysObject;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Import subscribers from a CSV or text file.
 */
public class ImportTagsAction extends AbstractNotifierAction
{
    private static final String[] _ALLOWED_EXTENSIONS = new String[] {"xml"};
    
    /** The tag provider extension point */
    protected TagProviderExtensionPoint _tagProviderExtPt;
    /** The Ametys object resolver */
    protected AmetysObjectResolver _resolver;
    /** Target types */
    protected TagTargetTypeExtensionPoint _targetTypeEP;
    /** The JSon Utils Component */
    protected JSONUtils _jsonUtils;
    /** The colors component */
    protected AbstractColorsComponent _colorsComponent;
    
    /** The count of new tags **/
    private int _createdTagsCount;
    /** The count of updated tags **/
    private int _updatedTagsCount;
    /** The count of error **/
    private int _errorCount;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _tagProviderExtPt = (TagProviderExtensionPoint) smanager.lookup(TagProviderExtensionPoint.ROLE);
        _resolver = (AmetysObjectResolver) smanager.lookup(AmetysObjectResolver.ROLE);
        _targetTypeEP = (TagTargetTypeExtensionPoint) smanager.lookup(TagTargetTypeExtensionPoint.ROLE);
        _jsonUtils = (JSONUtils) smanager.lookup(JSONUtils.ROLE);
        _colorsComponent = (AbstractColorsComponent) smanager.lookup(TagColorsComponent.ROLE);
    }
    
    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Map<String, Object> result = new HashMap<>();

        Request request = ObjectModelHelper.getRequest(objectModel);

        String parentId = request.getParameter("tagId");
        String forceProvider = request.getParameter("tagProvider");
        if (StringUtils.isEmpty(forceProvider))
        {
            forceProvider = CMSJCRTagProvider.class.getName();
        }
        
        Part part = (Part) request.get("importFile");
        if (part instanceof RejectedPart)
        {
            return Collections.singletonMap("error", "rejected-file");
        }
        
        PartOnDisk uploadedFilePart = (PartOnDisk) part;
        File uploadedFile = (uploadedFilePart != null) ? uploadedFilePart.getFile() : null;
        String filename = (uploadedFilePart != null) ? uploadedFilePart.getFileName().toLowerCase() : null;
        
        if (!FilenameUtils.isExtension(filename, _ALLOWED_EXTENSIONS))
        {
            result.put("success", false);
            result.put("error", "invalid-extension");
        }
        else if (uploadedFile == null)
        {
            result.put("success", false);
            result.put("error", "no-file");
        }
        else
        {
            String contextualParametersAsStr = request.getParameter("contextualParameters");
            Map<String, Object> contextualParameters = _jsonUtils.convertJsonToMap(contextualParametersAsStr);
            
            Configuration configuration = initializeTags(uploadedFile.getPath());
            Map<String, CMSTag> tags = configureTags(configuration, null, "plugin.cms");
            
            _createdTagsCount = 0;
            _updatedTagsCount = 0;
            _errorCount = 0;
            
            DefaultTraversableAmetysObject<?> parent = _resolver.resolveById(parentId);
            createOrUpdateTags(parent, tags, contextualParameters);
            
            CMSJCRTagProvider provider = (CMSJCRTagProvider) _tagProviderExtPt.getExtension(forceProvider);
            TraversableAmetysObject rootNode = provider.getRootNode(contextualParameters);

            result.put("success", true);
            result.put("rootId", rootNode.getId());
            result.put("createdCount", _createdTagsCount);
            result.put("updatedCount", _updatedTagsCount);
            result.put("errorCount", _errorCount);
        }
        
        request.setAttribute(JSonReader.OBJECT_TO_READ, result);
        return EMPTY_MAP;
    }
    
    /**
     * Save or update a tag
     * @param parent the JCR parent tag or provider
     * @param tags the list of tags to add or update
     * @param contextualParameters Contextual parameters transmitted by the environment.
     */
    protected void createOrUpdateTags(DefaultTraversableAmetysObject<?> parent, Map<String, CMSTag> tags, Map<String, Object> contextualParameters)
    {
        boolean update = false;
        for (CMSTag cmsTag : tags.values())
        {
            CMSJCRTag jcrTag = null;
            
            try
            {
                if (parent.hasChild(cmsTag.getName()))
                {
                    // Update existing tag
                    jcrTag = parent.getChild(cmsTag.getName());
                    update = true;
                }
                else
                {
                    // Create new tag
                    String name = _findUniqueName (cmsTag.getId(), contextualParameters);
                    jcrTag = parent.createChild(name, CMSTagFactory.TAG_NODETYPE);
                }
                
                jcrTag.setTitle(cmsTag.getTitle().getLabel());
                jcrTag.setDescription(cmsTag.getDescription().getLabel());
                jcrTag.setVisibility(cmsTag.getVisibility());
                jcrTag.setTargetType(cmsTag.getTarget());
                
                // recursive call to save/update
                createOrUpdateTags(jcrTag, cmsTag.getTags(), contextualParameters);
                
                jcrTag.saveChanges();
                
                Map<String, Object> eventParams = new HashMap<>();
                eventParams.put(ObservationConstants.ARGS_TAG_ID, jcrTag.getId());
                eventParams.put(ObservationConstants.ARGS_TAG_NAME, jcrTag.getName());
                
                if (update)
                {
                    _observationManager.notify(new Event(ObservationConstants.EVENT_TAG_UPDATED, _getCurrentUser(), eventParams));
                    _updatedTagsCount++;
                }
                else
                {
                    // Notify observers that the tag has been added.
                    _observationManager.notify(new Event(ObservationConstants.EVENT_TAG_ADDED, _getCurrentUser(), eventParams));
                    _createdTagsCount++;
                }
            }
            catch  (Exception e)
            {
                getLogger().error("Unable to add tag " + cmsTag.getName() + " to JCR tag category of id '" + parent.getId() + "'", e);
                _errorCount++;
            }
        }
    }
    
    /**
     * Initialize a configuration tags from the tags file.
     * @param path the path of the file to initialize tags.
     * @return The configuration
     * @throws Exception if an error occurs.
     */
    protected Configuration initializeTags(String path) throws Exception
    {
        try (InputStream is = new FileInputStream(path))
        {
            SAXParserFactory factory = SAXParserFactory.newInstance();
            
            XMLReader reader = factory.newSAXParser().getXMLReader();

            DefaultConfigurationBuilder confBuilder = new DefaultConfigurationBuilder(reader);

            return confBuilder.build(is);
        }
        catch (IOException e)
        {
            return null;
        }
    }
    
    /**
     * Configure tag from the passed configuration
     * @param configuration The configuration
     * @param parent The parent tag if any
     * @param defaultCatalogue The default catalogue for i18n
     * @return a Set of {@link CMSTag}
     * @throws ConfigurationException If an error occurred
     */
    protected Map<String, CMSTag> configureTags (Configuration configuration, CMSTag parent, String defaultCatalogue)  throws ConfigurationException
    {
        Map<String, CMSTag> tags = new HashMap<>();
        
        Configuration[] tagsConfiguration = configuration.getChildren("tag");
        for (Configuration tagConfiguration : tagsConfiguration)
        {
            String id = tagConfiguration.getAttribute("id");
            if (!Tag.NAME_PATTERN.matcher(id).matches())
            {
                throw new ConfigurationException("Invalid tag ID '" + id + "': it must match the pattern " + Tag.NAME_PATTERN.toString(), configuration);
            }
            
            TagVisibility visibility = TagVisibility.PUBLIC;
            if (tagConfiguration.getAttribute("private", "").equals("true"))
            {
                visibility = TagVisibility.PRIVATE;
            }
            
            Configuration child = tagConfiguration.getChild("color");
            String color = child != null ? child.getValue(null) : null;
            
            String typeName = tagConfiguration.getAttribute("target", "CONTENT");
            TagTargetType targetType = _targetTypeEP.getTagTargetType(typeName);
            
            I18nizableText label = configureLabel (tagConfiguration, defaultCatalogue);
            I18nizableText description = configureDescription (tagConfiguration, defaultCatalogue);
            CMSTag cmsTag = new CMSTag(id, id, parent, label, description, visibility, targetType);
            ColorableCMSTag tag = new ColorableCMSTag(cmsTag, color, _colorsComponent);
            
            tags.put(id, tag);
            
            // Recursive configuration
            Map<String, CMSTag> childTags = configureTags(tagConfiguration, tag, defaultCatalogue);
            tag.setTags(childTags);
        }
        
        return tags;
    }
    
    /**
     * Configure label from the passed configuration
     * @param configuration The configuration
     * @param defaultCatalogue The default catalogue
     * @return The label
     * @throws ConfigurationException If an error occurred
     */
    protected I18nizableText configureLabel (Configuration configuration, String defaultCatalogue) throws ConfigurationException
    {
        Configuration labelConfiguration = configuration.getChild("label");
        
        if (labelConfiguration.getAttributeAsBoolean("i18n", false))
        {
            return new I18nizableText(defaultCatalogue, labelConfiguration.getValue(""));
        }
        else
        {
            return new I18nizableText(labelConfiguration.getValue(""));
        }
    }
    
    /**
     * Configure description from the passed configuration
     * @param configuration The configuration
     * @param defaultCatalogue The default catalogue
     * @return The description
     * @throws ConfigurationException If an error occurred
     */
    protected I18nizableText configureDescription (Configuration configuration, String defaultCatalogue) throws ConfigurationException
    {
        Configuration descConfiguration = configuration.getChild("description");
        
        if (descConfiguration.getAttributeAsBoolean("i18n", false))
        {
            return new I18nizableText(defaultCatalogue, descConfiguration.getValue(""));
        }
        else
        {
            return new I18nizableText(descConfiguration.getValue(""));
        }
    }

    private String _findUniqueName(String originalName, Map<String, Object> contextualParameters)
    {
        Set<TagProvider<CMSTag>> providers = _getTagProviders();
        
        // Find an unique name
        int index = 2;
        String name = originalName;
        while (_hasTag(providers, name, contextualParameters))
        {
            name = originalName + "_" + (index++);
        }
        
        return name;
    }
    
    private boolean _hasTag(Set<TagProvider<CMSTag>> providers, String name, Map<String, Object> contextualParameters)
    {
        for (TagProvider<CMSTag> provider : providers)
        {
            if (provider.hasTag(name, contextualParameters))
            {
                return true;
            }
        }
        return false;
    }
    
    private Set<TagProvider<CMSTag>> _getTagProviders ()
    {
        Set<TagProvider<CMSTag>> providers = new HashSet<>();
        
        Set<String> ids = _tagProviderExtPt.getExtensionsIds();
        for (String id : ids)
        {
            TagProvider<CMSTag> provider = _tagProviderExtPt.getExtension(id);
            providers.add(provider);
        }
        
        return providers;
    }
}
