/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag.jcr;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jcr.Node;
import javax.jcr.PathNotFoundException;
import javax.jcr.RepositoryException;

import org.ametys.cms.tag.CMSTag;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.jcr.DefaultTraversableAmetysObject;

/**
 * {@link AmetysObject} for storing tag informations.
 */
public class JCRTag extends DefaultTraversableAmetysObject<TagFactory>
{
    /** Constants for title metadata. */
    private static final String __METADATA_TITLE = RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ":title";
    /** Constants for description metadata. */
    private static final String __METADATA_DESC = RepositoryConstants.NAMESPACE_PREFIX_INTERNAL + ":description";
    
    /**
     * Creates a {@link CMSTag}.
     * @param node the node backing this {@link AmetysObject}.
     * @param parentPath the parent path in the Ametys hierarchy.
     * @param factory the {@link TagFactory} which creates the AmetysObject.
     */
    public JCRTag (Node node, String parentPath, TagFactory factory)
    {
        super(node, parentPath, factory);
    }
    
    /**
     * Retrieves the title.
     * @return the title.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public String getTitle() throws AmetysRepositoryException
    {
        try
        {
            return getNode().getProperty(__METADATA_TITLE).getString();
        }
        catch (PathNotFoundException e)
        {
            return null;
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to get title property", e);
        }
    }
    
    /**
     * Set the title.
     * @param title the title.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setTitle(String title) throws AmetysRepositoryException
    {
        try
        {
            getNode().setProperty(__METADATA_TITLE, title);
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to set title property", e);
        }
    }
    
    /**
     * Retrieves the description.
     * @return the description.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public String getDescription() throws AmetysRepositoryException
    {
        try
        {
            return getNode().getProperty(__METADATA_DESC).getString();
        }
        catch (PathNotFoundException e)
        {
            return null;
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to get description property", e);
        }
    }
    
    /**
     * Set the description.
     * @param description the description.
     * @throws AmetysRepositoryException if an error occurs.
     */
    public void setDescription(String description) throws AmetysRepositoryException
    {
        try
        {
            getNode().setProperty(__METADATA_DESC, description);
        }
        catch (RepositoryException e)
        {
            throw new AmetysRepositoryException("Unable to set description property", e);
        }
    }
    
    /**
     * Get the json tag
     * @return the tag to json
     */
    public Map<String, Object> toJSON()
    {
        Map<String, Object> infos = new HashMap<>();

        infos.put("id", getId());
        infos.put("name", getName());
        infos.put("title", getTitle());
        infos.put("description", getDescription());
        
        infos.put("iconSmall", "/plugins/cms/resources/img/tag/tag_16.png");
        infos.put("iconMedium", "/plugins/cms/resources/img/tag/tag_32.png");
        infos.put("iconLarge", "/plugins/cms/resources/img/tag/tag_50.png");
        
        if (!getChildren().iterator().hasNext())
        {
            List<Map<String, Object>> qNodes = new ArrayList<>();
            infos.put("children", qNodes);
        }

        return infos;
    }
}
