/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.tag.jcr;

import java.util.HashSet;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import javax.jcr.RepositoryException;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang.StringUtils;

import org.ametys.cms.color.AbstractColorsComponent;
import org.ametys.cms.tag.CMSTag.TagVisibility;
import org.ametys.cms.tag.Tag;
import org.ametys.cms.tag.TagColorsComponent;
import org.ametys.cms.tag.TagProvider;
import org.ametys.cms.tag.TagProviderExtensionPoint;
import org.ametys.cms.tag.TagTargetTypeExtensionPoint;
import org.ametys.cms.tag.TagsDAO;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.ui.Callable;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.ModifiableTraversableAmetysObject;
import org.ametys.plugins.repository.UnknownAmetysObjectException;

/**
 * Component for operations on JCR tags
 */
public class JCRTagsDAO extends AbstractJCRTagsDAO
{
    /** The Avalon role */
    public static final String ROLE = JCRTagsDAO.class.getName();
    
    /** The attribute name for the visibility */
    public static final String VISIBILITY_ATTRIBUTE_NAME = "visibility";
    
    /** The attribute name for the target type */
    public static final String TARGET_TYPE_ATTRIBUTE_NAME = "target";
    
    /** The attribute name for the color */
    public static final String COLOR_ATTRIBUTE_NAME = "color";
    
    /** The tag provider extension point */
    protected TagProviderExtensionPoint _tagProviderExtPt;
    /** The rights manager */
    protected RightManager _rightManager;
    /** Target types */
    protected TagTargetTypeExtensionPoint _targetTypeEP;
    /** Tags DAO */
    protected TagsDAO _tagsDAO;
    /** The tags colors component */
    protected AbstractColorsComponent _colorComponent;
    
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _tagProviderExtPt = (TagProviderExtensionPoint) serviceManager.lookup(TagProviderExtensionPoint.ROLE);
        _rightManager = (RightManager) serviceManager.lookup(RightManager.ROLE);
        _targetTypeEP = (TagTargetTypeExtensionPoint) serviceManager.lookup(TagTargetTypeExtensionPoint.ROLE);
        _tagsDAO = (TagsDAO) serviceManager.lookup(TagsDAO.ROLE);
        _colorComponent = (AbstractColorsComponent) serviceManager.lookup(TagColorsComponent.ROLE);
    }
    
    @Override
    public ModifiableTraversableAmetysObject _getTagRootObject (String tagProviderId, Map<String, Object> contextualParameters) throws RepositoryException
    {
        CMSJCRTagProvider provider = (CMSJCRTagProvider) _tagProviderExtPt.getExtension(tagProviderId);
        return provider.getRootNode(contextualParameters);
    }
    
    @Override
    protected Tag _getTagFromName(String name, Map<String, Object> contextualParameters)
    {
        return _tagProviderExtPt.getTag(name, contextualParameters);
    }
    
    @Override
    protected void _checkUserRight () throws IllegalStateException
    {
        String rightId = "CMS_Rights_Tags_HandleTag";
        String context = "/cms";
        
        UserIdentity user = _currentUserProvider.getUser();
        if (_rightManager.hasRight(user, rightId, context) != RightResult.RIGHT_ALLOW)
        {
            getLogger().error("User '" + user + "' tried to access a privileged feature without convenient right. Should have right '" + rightId + "' on context '" + context + "'");
            throw new IllegalStateException("You have no right to access this feature.");
        }
    }
    
    @Override
    protected JCRTag _createJCRTag(String parentId, String name, String title, String description, Map<String, Object> otherParameters, Map<String, Object> contextualParameters) throws RepositoryException
    {
        ModifiableTraversableAmetysObject parent = null;
        if (StringUtils.isEmpty(parentId))
        {
            parent = _getTagRootObject(CMSJCRTagProvider.class.getName(), contextualParameters);
        }
        else
        {
            parent = _resolver.resolveById(parentId);
        }
        
        CMSJCRTag jcrTag = parent.createChild(name, CMSTagFactory.TAG_NODETYPE);
        jcrTag.setTitle(title);
        jcrTag.setDescription(description);
        
        // Visibility
        Optional.ofNullable(otherParameters.get(VISIBILITY_ATTRIBUTE_NAME))
                .filter(String.class::isInstance)
                .map(String.class::cast)
                .map(TagVisibility::valueOf)
                .ifPresent(jcrTag::setVisibility);

        // Target type
        Optional.ofNullable(otherParameters.get(TARGET_TYPE_ATTRIBUTE_NAME))
                .filter(String.class::isInstance)
                .map(String.class::cast)
                .map(_targetTypeEP::getTagTargetType)
                .ifPresent(jcrTag::setTargetType);
        
        // Color
        Optional.ofNullable(otherParameters.get(COLOR_ATTRIBUTE_NAME))
                .filter(String.class::isInstance)
                .map(String.class::cast)
                .ifPresent(jcrTag::setColor);
        
        parent.saveChanges();
        
        return jcrTag;
    }
    
    @Override
    protected JCRTag _updateJCRTag(String tagId, String title, String description, Map<String, Object> otherParameters) throws UnknownAmetysObjectException
    {
        String visibility = (String) otherParameters.get(VISIBILITY_ATTRIBUTE_NAME);
        String targetType = (String) otherParameters.get(TARGET_TYPE_ATTRIBUTE_NAME);
        String color = (String) otherParameters.get(COLOR_ATTRIBUTE_NAME);
        
        CMSJCRTag jcrTag = _resolver.resolveById(tagId);
        jcrTag.setTitle(title);
        jcrTag.setDescription(description);
        jcrTag.setVisibility(TagVisibility.valueOf(visibility));
        jcrTag.setTargetType(_targetTypeEP.getTagTargetType(targetType));
        jcrTag.setColor(color);
        
        jcrTag.saveChanges();
        
        return jcrTag;
    }
    
    @Override
    protected Set<TagProvider<? extends Tag>> _getTagProviders ()
    {
        Set<TagProvider<? extends Tag>> providers = new HashSet<>();
        
        Set<String> ids = _tagProviderExtPt.getExtensionsIds();
        for (String id : ids)
        {
            TagProvider<? extends Tag> provider = _tagProviderExtPt.getExtension(id);
            providers.add(provider);
        }
        
        return providers;
    }
    
    @Override
    @Callable
    public Map<String, Object> getTag(String tagId)
    {
        Map<String, Object> tag = super.getTag(tagId);
        String color = (String) tag.get(COLOR_ATTRIBUTE_NAME);
        if (StringUtils.isBlank(color))
        {
            color = _colorComponent.getDefaultKey();
            
        }
        tag.put(COLOR_ATTRIBUTE_NAME, color);
        
        return tag;
    }
}
