/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.transformation;

import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.ext.LexicalHandler;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * This interface is for component that wants to be in the docbook pipeline.
 * As a pipeline you receive events (from the ContentHandler and LexicalHandler interfaces) 
 * you may transform it and write it to the given contentHandler (setContentHandler) and 
 * given lexicalHandler (setLexicalHandler) 
 */
public abstract class AbstractEnhancementHandler extends AbstractLogEnabled implements ContentHandler, LexicalHandler
{
    /** Processing instruction for unmodifiable content */
    public static final String PROCESSING_INSTRUCTION_UNMODIFIABLE = "ametys-unmodifiable";
    
    /** The content handler where to write */
    protected ContentHandler _contentHandler;
    /** The lexical handler where to write */
    protected LexicalHandler _lexicalHandler;
    
    /** 
     * True if the current HTML should not be modified
     * Use to prevent to apply transformation on sax events (eg: HTML expert)
     */
    protected boolean _inUnmodifiableContent;
    
    /**
     * Set the target content handler
     * @param contentHandler the target content handler
     */
    public void setContentHandler(ContentHandler contentHandler)
    {
        _contentHandler = contentHandler;
    }
    
    /**
     * Set the target lexical handler
     * @param lexicalHandler the target lexical handler
     */
    public void setLexicalHandler(LexicalHandler lexicalHandler)
    {
        _lexicalHandler = lexicalHandler;
    }
    
    /**
     * SAX an error message in content.
     * @param title the error title. Can be null.
     * @param message the error message. Can be null.
     * @param absoluteIconPath the absolute path of the icon. Can be null.
     * @throws SAXException if an error occurred while SAXing
     */
    protected void saxError (I18nizableText title, I18nizableText message, String absoluteIconPath) throws SAXException
    {
        AttributesImpl atts = new AttributesImpl();
        
        atts.addCDATAAttribute("class", "ametys-cms-content-inline-error");
        XMLUtils.startElement(_contentHandler, "div", atts);
        
        if (absoluteIconPath != null)
        {
            atts = new AttributesImpl();
            atts.addCDATAAttribute("src", absoluteIconPath);
            XMLUtils.createElement(_contentHandler, "img", atts);
        }
        
        if (title != null)
        {
            XMLUtils.startElement(_contentHandler, "strong");
            title.toSAX(_contentHandler);
            XMLUtils.endElement(_contentHandler, "strong");
            XMLUtils.createElement(_contentHandler, "br");
        }
        
        if (message != null)
        {
            message.toSAX(_contentHandler);
        }
        
        XMLUtils.endElement(_contentHandler, "div");
    }
    
    @Override
    public void startDocument() throws SAXException
    {
        _contentHandler.startDocument();
    }
    @Override
    public void endDocument() throws SAXException
    {
        _contentHandler.endDocument();
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        _contentHandler.startElement(uri, localName, qName, atts);
    }
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        _contentHandler.endElement(uri, localName, qName);
    }
    
    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        _contentHandler.characters(ch, start, length);
    }
    @Override
    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
    {
        _contentHandler.ignorableWhitespace(ch, start, length);
    }
    
    @Override
    public void processingInstruction(String target, String data) throws SAXException
    {
        if (PROCESSING_INSTRUCTION_UNMODIFIABLE.equals(target))
        {
            if ("start".equals(data))
            {
                // Following characters or elements should not be transformed or modified by other content or lexical handlers
                _inUnmodifiableContent = true;
            }
            else
            {
                // Following characters or elements could be transformed or modified  by other content or lexical handlers
                _inUnmodifiableContent = false;
            }
        }
        
        _contentHandler.processingInstruction(target, data);
    }
    
    @Override
    public void setDocumentLocator(Locator locator)
    {
        _contentHandler.setDocumentLocator(locator);
    }
    @Override
    public void skippedEntity(String name) throws SAXException
    {
        _contentHandler.skippedEntity(name);
    }
    
    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {
        _contentHandler.startPrefixMapping(prefix, uri);
    }
    @Override
    public void endPrefixMapping(String prefix) throws SAXException
    {
        _contentHandler.endPrefixMapping(prefix);
    }
    
    @Override
    public void startEntity(String name) throws SAXException
    {
        _lexicalHandler.startEntity(name);
    }
    @Override
    public void endEntity(String name) throws SAXException
    {
        _lexicalHandler.endEntity(name);
    }

    @Override
    public void startCDATA() throws SAXException
    {
        _lexicalHandler.startCDATA();
    }
    @Override
    public void endCDATA() throws SAXException
    {
        _lexicalHandler.endCDATA();
    }

    @Override
    public void startDTD(String name, String publicId, String systemId) throws SAXException
    {
        _lexicalHandler.startDTD(name, publicId, systemId);
    }
    
    @Override
    public void endDTD() throws SAXException
    {
        _lexicalHandler.endCDATA();
    }

    @Override
    public void comment(char[] ch, int start, int length) throws SAXException
    {
        _lexicalHandler.comment(ch, start, length);
    }
}
