/*
 *  Copyright 2011 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.transformation;

import java.io.IOException;
import java.io.InputStream;
import java.time.ZonedDateTime;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.io.IOUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.xml.sax.SAXParser;
import org.xml.sax.ContentHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import org.ametys.cms.data.RichText;
import org.ametys.core.util.IgnoreRootHandler;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.metadata.ModifiableRichText;
import org.ametys.runtime.model.type.DataContext;

/**
 * Abstract class for {@link RichTextTransformer}s relying on Cocoon pipelines to actually transform HTML to RichText back and forth.
 */
public abstract class AbstractRichTextTransformer implements RichTextTransformer, Serviceable
{
    private SourceResolver _sourceResolver;
    private ServiceManager _manager;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    @SuppressWarnings("unchecked")
    @Deprecated
    public void transform(String source, ModifiableRichText dest) throws AmetysRepositoryException, IOException
    {
        Source xmlSource = null;
        
        try 
        {
            Map parameters = new HashMap();
            parameters.put("source", source);
            parameters.put("richText", dest);

            xmlSource = _sourceResolver.resolveURI(_getSourceUriForHTML2RichText(), null, parameters);
            
            try (InputStream is = xmlSource.getInputStream())
            {
                dest.setMimeType("application/xml");
                dest.setLastModified(new Date());
                dest.setInputStream(is);
            }
        }
        finally
        {
            _sourceResolver.release(xmlSource);
        }
    }
    
    public void transform(String source, RichText dest) throws AmetysRepositoryException, IOException
    {
        Source xmlSource = null;
        
        try 
        {
            Map<String, Object> parameters = new HashMap<>();
            parameters.put("source", source);
            parameters.put("richText", dest);

            xmlSource = _sourceResolver.resolveURI(_getSourceUriForHTML2RichText(), null, parameters);
            
            try (InputStream is = xmlSource.getInputStream())
            {
                dest.setMimeType("application/xml");
                dest.setLastModificationDate(ZonedDateTime.now());
                dest.setInputStream(is);
            }
        }
        finally
        {
            _sourceResolver.release(xmlSource);
        }
    }
    
    /**
     * Returns the internal {@link Source} URI used to process the HTML editor content. 
     * @return the internal {@link Source} URI used to process the HTML editor content. 
     */
    protected abstract String _getSourceUriForHTML2RichText();
    
    @Override
    @Deprecated
    public void transformForEditing(ModifiableRichText source, StringBuilder dest) throws AmetysRepositoryException, IOException
    {
        Source textSource = null;
        
        try (InputStream is = source.getInputStream())
        {
            Map<String, Object> parameters = new HashMap<>();
            parameters.put("source", is);
            
            textSource = _sourceResolver.resolveURI(_getSourceUriForRichText2HTML(), null, parameters);
            
            try (InputStream textSourceIs = textSource.getInputStream())
            {
                dest.append(IOUtils.toString(textSourceIs, "UTF-8"));
            }
        }
        finally
        {
            _sourceResolver.release(textSource);
        }
    }
    
    public void transformForEditing(RichText source, DataContext dataContext, StringBuilder dest) throws AmetysRepositoryException, IOException
    {
        Source textSource = null;
        
        try (InputStream is = source.getInputStream())
        {
            Map<String, Object> parameters = new HashMap<>();
            parameters.put("source", is);
            parameters.put("dataContext", dataContext);
            
            textSource = _sourceResolver.resolveURI(_getSourceUriForRichText2HTML(), null, parameters);
            
            try (InputStream textSourceIs = textSource.getInputStream())
            {
                dest.append(IOUtils.toString(textSourceIs, "UTF-8"));
            }
        }
        finally
        {
            _sourceResolver.release(textSource);
        }
    }
    
    /**
     * Returns the internal {@link Source} URI used to process the RichText's InputStream. 
     * @return the internal {@link Source} URI used to process the RichText's InputStream. 
     */
    protected abstract String _getSourceUriForRichText2HTML();

    @Override
    @Deprecated
    public void transformForRendering(org.ametys.plugins.repository.metadata.RichText source, ContentHandler handler) throws AmetysRepositoryException, SAXException, IOException
    {
        SAXParser saxParser = null;
        
        try
        {
            saxParser = (SAXParser) _manager.lookup(SAXParser.ROLE);
            
            try (InputStream is = source.getInputStream())
            {
                saxParser.parse(new InputSource(is), new IgnoreRootHandler(handler));
            }
        }
        catch (ServiceException e)
        {
            throw new AmetysRepositoryException("Unable to get SAX parser", e);
        }
        finally
        {
            _manager.release(saxParser);
        }
    }
    
    public void transformForRendering(RichText source, ContentHandler handler) throws AmetysRepositoryException, SAXException, IOException
    {
        SAXParser saxParser = null;
        
        try
        {
            saxParser = (SAXParser) _manager.lookup(SAXParser.ROLE);
            
            try (InputStream is = source.getInputStream())
            {
                saxParser.parse(new InputSource(is), new IgnoreRootHandler(handler));
            }
        }
        catch (ServiceException e)
        {
            throw new AmetysRepositoryException("Unable to get SAX parser", e);
        }
        finally
        {
            _manager.release(saxParser);
        }
    }
}
