/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.transformation;

import org.apache.avalon.framework.logger.AbstractLogEnabled;

/**
 * Abstract superclass for simple {@link URIResolver}.
 */
public abstract class AbstractURIResolver extends AbstractLogEnabled implements URIResolver
{
    @Override
    public String resolve(String uri, boolean download, boolean absolute, boolean internal)
    {
        return _resolve(uri, "", download, absolute, internal);
    }
    
    @Override
    public String resolveImage(String uri, int height, int width, boolean download, boolean absolute, boolean internal)
    {
        if (height == 0 && width == 0)
        {
            return resolve(uri, download, absolute, internal);
        }
        
        StringBuilder uriArgument = new StringBuilder();
        uriArgument.append("_").append(height).append("x").append(width);
        return _resolve(uri, uriArgument.toString(), download, absolute, internal);
    }
    
    @Override
    public String resolveBoundedImage(String uri, int maxHeight, int maxWidth, boolean download, boolean absolute, boolean internal)
    {
        if (maxHeight == 0 && maxWidth == 0)
        {
            return resolve(uri, download, absolute, internal);
        }
        
        StringBuilder uriArgument = new StringBuilder();
        uriArgument.append("_max").append(maxHeight).append("x").append(maxWidth);
        return _resolve(uri, uriArgument.toString(), download, absolute, internal);
    }
    
    
    @Override
    public String resolveCroppedImage(String uri, int cropHeight, int cropWidth, boolean download, boolean absolute, boolean internal)
    {
        if (cropHeight == 0 && cropWidth == 0)
        {
            return resolve(uri, download, absolute, internal);
        }
        
        StringBuilder uriArgument = new StringBuilder();
        uriArgument.append("_crop").append(cropHeight).append("x").append(cropWidth);
        return _resolve(uri, uriArgument.toString(), download, absolute, internal);
    }
    
    @Override
    public String resolveImageAsBase64(String uri, int height, int width)
    {
        return resolveImageAsBase64(uri, height, width, 0, 0, 0, 0);
    }
    
    @Override
    public String resolveBoundedImageAsBase64(String uri, int maxHeight, int maxWidth)
    {
        return resolveImageAsBase64(uri, 0, 0, maxHeight, maxWidth, 0, 0);
    }
    
    @Override
    public String resolveCroppedImageAsBase64(String uri, int cropHeight, int cropWidth)
    {
        return resolveImageAsBase64(uri, 0, 0, 0, 0, cropHeight, cropWidth);
    }

    /**
     * Resolves a link URI for rendering.<br>
     * The output must be a properly encoded path, relative to the webapp context, accessible from a browser.
     * @param uri the link URI.
     * @param uriArgument the argument to append to the uri
     * @param download true if the pointed resource is to be downloaded.
     * @param absolute true if the url must be absolute
     * @param internal true to get an internal URI.
     * @return the path to the image.
     */
    protected abstract String _resolve(String uri, String uriArgument, boolean download, boolean absolute, boolean internal);
    
    /**
     * Get an image's bytes encoded as base64, optionally resized. 
     * @param uri the image URI.
     * @param height the specified height. Ignored if negative.
     * @param width the specified width. Ignored if negative.
     * @param maxHeight the maximum image height. Ignored if height or width is specified.
     * @param maxWidth the maximum image width. Ignored if height or width is specified.
     * @param cropHeight The cropping height. Ignored if negative.
     * @param cropWidth The cropping width. Ignored if negative.
     * @return the image bytes encoded as base64.
     */
    protected abstract String resolveImageAsBase64(String uri, int height, int width, int maxHeight, int maxWidth, int cropHeight, int cropWidth);
}
