/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.transformation;

import java.io.InputStream;
import java.util.Collections;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.cms.URIPrefixHandler;
import org.ametys.cms.repository.Content;
import org.ametys.cms.transformation.ConsistencyChecker.CHECK;
import org.ametys.core.util.FilenameUtils;
import org.ametys.core.util.ImageResolverHelper;
import org.ametys.core.util.URIUtils;
import org.ametys.plugins.explorer.resources.Resource;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * {@link URIResolver} for type "attachment".<br>
 * These links point to a file from the attachments of the current Content.
 */
public class AttachmentURIResolver extends AbstractURIResolver implements Serviceable, Contextualizable
{
    /** Regular expression for contents stored under plugins */
    protected static final Pattern __PLUGIN_CONTENT_PATTERN = Pattern.compile("^/ametys:plugins/([^/]+)/ametys:contents/(.*)$");
    /** The ametys resolver */
    protected AmetysObjectResolver _resolver;
    /** The avalon context */
    protected Context _context;
    /** The URI prefix handler */
    protected URIPrefixHandler _prefixHandler;
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _prefixHandler = (URIPrefixHandler) manager.lookup(URIPrefixHandler.ROLE);
    }
    
    @Override
    public String getType()
    {
        return "attachment-content";
    }
    
    @Override
    public String _resolve(String uri, String uriArgument, boolean download, boolean absolute, boolean internal)
    {
        String path;
        Content content = null;
        String contentName = null;
        
        try
        {
            Resource resource = _resolver.resolveById(uri);
            path = resource.getResourcePath();
            
            AmetysObject ao = resource;
            do
            {
                ao = ao.getParent();
            }
            while (ao != null && !(ao instanceof Content));
            
            if (ao == null)
            {
                throw new IllegalArgumentException("The given attachment URI should be a content attachment.");
            }
            
            content = (Content) ao;
            contentName = content.getName();
        }
        catch (UnknownAmetysObjectException e)
        {
            getLogger().warn("Link to unexisting resource " + uri);
            return "";
        }
        catch (Exception e)
        {
            throw new IllegalStateException(e);
        }
        
        String filename = FilenameUtils.encodePath(path);
        String basePath = org.apache.commons.io.FilenameUtils.removeExtension(filename);
        String extension = org.apache.commons.io.FilenameUtils.getExtension(filename);
        
        StringBuilder resultPath = new StringBuilder();
        resultPath.append(getUriPrefix(content, download, absolute, internal));
        
        Matcher m = __PLUGIN_CONTENT_PATTERN.matcher(content.getPath());
        if (m.matches())
        {
            String pluginName = m.group(1);
            resultPath.append("/")
                .append(pluginName)
                .append("/_plugin-attachment/");
        }
        else
        {
            resultPath.append("/_attachment/");
        }
        
        resultPath.append(contentName)
                  .append(basePath)
                  .append(uriArgument)
                  .append(extension.isEmpty() ? "" : "." + extension);
        
        return URIUtils.encodeURI(resultPath.toString(), download ? Collections.singletonMap("download", "true") : null);
    }
    
    /**
     * Get the URI prefix
     * @param object The object
     * @param download true if the pointed resource is to be downloaded.
     * @param absolute true if the url must be absolute
     * @param internal true to get an internal URI.
     * @return the URI prefix
     */
    protected String getUriPrefix (AmetysObject object, boolean download, boolean internal, boolean absolute)
    {
        return _prefixHandler.computeUriPrefix(absolute, internal);
    }
    
    
    @Override
    protected String resolveImageAsBase64(String uri, int height, int width, int maxHeight, int maxWidth, int cropHeight, int cropWidth)
    {
        try
        {
            Resource resource = _resolver.resolveById(uri);
            
            try (InputStream dataIs = resource.getInputStream())
            {
                return ImageResolverHelper.resolveImageAsBase64(dataIs, resource.getMimeType(), height, width, maxHeight, maxWidth, cropHeight, cropWidth);
            }
        }
        catch (UnknownAmetysObjectException e)
        {
            getLogger().warn("Link to unexisting resource " + uri);
            return "";
        }
        catch (Exception e)
        {
            throw new IllegalStateException(e);
        }
    }
    
    @Override
    public CHECK checkLink(String uri, boolean shortTest)
    {
        try
        {
            _resolver.resolveById(uri);
            return CHECK.SUCCESS;
        }
        catch (UnknownAmetysObjectException e)
        {
            return CHECK.NOT_FOUND;
        }
    }
    
    @Override
    public I18nizableText getLabel(String uri)
    {
        try
        {
            Resource resource = _resolver.resolveById(uri);
            return new I18nizableText("plugin.cms", "PLUGINS_CMS_LINK_ATTACHMENT_LABEL", Collections.singletonList(resource.getResourcePath()));
        }
        catch (UnknownAmetysObjectException e)
        {
            return new I18nizableText("plugin.cms", "PLUGINS_CMS_LINK_ATTACHMENT_UNKNOWN");
        }
    }

    public String getMimeType(String uri)
    {
        return _resolver.<Resource>resolveById(uri).getMimeType();
    }
}
