/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.transformation;

import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.commons.lang3.StringUtils;
import org.apache.http.NameValuePair;
import org.apache.http.client.utils.URLEncodedUtils;

import org.ametys.cms.data.Binary;
import org.ametys.cms.repository.Content;
import org.ametys.cms.transformation.ConsistencyChecker.CHECK;
import org.ametys.core.util.FilenameUtils;
import org.ametys.core.util.ImageResolverHelper;
import org.ametys.core.util.URIUtils;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.plugins.repository.data.ametysobject.ModelAwareDataAwareAmetysObject;
import org.ametys.plugins.repository.version.VersionAwareAmetysObject;
import org.ametys.plugins.repository.version.VersionableAmetysObject;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.workspace.WorkspaceMatcher;

/**
 * {@link URIResolver} for type "attribute".<br>
 * These links or images point to a file from the attribute of the current ametys object.
 */
public class AttributeURIResolver extends AbstractURIResolver implements Serviceable, Contextualizable
{
    private static final String __OBJECT_ID_URI_QUERY_PARAM_NAME = "objectId";
    private static final String __CONTENT_ID_URI_QUERY_PARAM_NAME = "contentId";
    
    /** The ametys object resolver */
    protected AmetysObjectResolver _resolver;
    
    /** The context */
    protected Context _context;
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    public String getType()
    {
        return "attribute";
    }

    private record AmetysObjectAndRevision (ModelAwareDataAwareAmetysObject ametysObject, boolean hasSwitched, Optional<String> revision) { /* empty */ }
    
    @Override
    protected String _resolve(String uri, String uriArgument, boolean download, boolean absolute, boolean internal)
    {
        Request request = ContextHelper.getRequest(_context);
        AttributeInfo info = _getAttributeInfo(uri, request);
        
        AmetysObjectAndRevision ametysObjectAndRevision = _getAmetysObjectAndRevisionFromAttributeInfo(info);
        ModelAwareDataAwareAmetysObject ametysObject = ametysObjectAndRevision.ametysObject();

        try
        {
            String ametysObjectVersion = StringUtils.EMPTY;
            if (ametysObject instanceof VersionableAmetysObject)
            {
                ametysObjectVersion = ((VersionableAmetysObject) ametysObject).getRevision();
            }
            
            String path = info.getPath();
            Binary binary = ametysObject.getValue(path);
            
            String filename = FilenameUtils.encodeName(binary.getFilename());
            
            String baseName = org.apache.commons.io.FilenameUtils.getBaseName(filename);
            String extension = org.apache.commons.io.FilenameUtils.getExtension(filename);
            
            StringBuilder resultPath = new StringBuilder();
            
            resultPath.append("/_object")
                .append(FilenameUtils.encodePath(ametysObject.getPath()))
                .append("/_attribute/")
                .append(path)
                .append("/")
                .append(baseName)
                .append(uriArgument)
                .append(extension.isEmpty() ? "" : "." + extension);
      
            String resultUri = getUri(resultPath.toString(), ametysObject, download, absolute, internal);
    
            Map<String, String> params = new HashMap<>();
            params.put("objectId", ametysObject.getId());
            
            if (download)
            {
                params.put("download", "true");
            }
            
            if (ametysObjectVersion != null)
            {
                params.put("contentVersion", ametysObjectVersion);
            }
            
            return internal ? URIUtils.buildURI(resultUri, params) : URIUtils.encodeURI(resultUri, params);
        }
        finally
        {
            _resetAmetysObjectRevisionIfNeeded(ametysObjectAndRevision);
        }
        
    }
    
    @Override
    protected String resolveImageAsBase64(String uri, int height, int width, int maxHeight, int maxWidth, int cropHeight, int cropWidth)
    {
        Request request = ContextHelper.getRequest(_context);
        AttributeInfo info = _getAttributeInfo(uri, request);
        
        AmetysObjectAndRevision ametysObjectAndRevision = _getAmetysObjectAndRevisionFromAttributeInfo(info);
        ModelAwareDataAwareAmetysObject ametysObject = ametysObjectAndRevision.ametysObject();
        
        try
        {
            String path = info.getPath();
            Binary binary = ametysObject.getValue(path);
            
            try (InputStream dataIs = binary.getInputStream())
            {
                return ImageResolverHelper.resolveImageAsBase64(dataIs, binary.getMimeType(), height, width, maxHeight, maxWidth, cropHeight, cropWidth);
            }
        }
        catch (Exception e)
        {
            throw new IllegalStateException(e);
        }
        finally
        {
            _resetAmetysObjectRevisionIfNeeded(ametysObjectAndRevision);
        }
    }

    public String getMimeType(String uri)
    {
        Request request = ContextHelper.getRequest(_context);
        AttributeInfo info = _getAttributeInfo(uri, request);
        
        AmetysObjectAndRevision ametysObjectAndRevision = _getAmetysObjectAndRevisionFromAttributeInfo(info);
        ModelAwareDataAwareAmetysObject ametysObject = ametysObjectAndRevision.ametysObject();
        
        try
        {
            String path = info.getPath();
            Binary binary = ametysObject.getValue(path);
            return binary.getMimeType();
        }
        catch (Exception e)
        {
            throw new IllegalStateException(e);
        }
        finally
        {
            _resetAmetysObjectRevisionIfNeeded(ametysObjectAndRevision);
        }
    }
    
    private AmetysObjectAndRevision _getAmetysObjectAndRevisionFromAttributeInfo(AttributeInfo info)
    {
        ModelAwareDataAwareAmetysObject ametysObject = info.getAmetysObject();
        
        if (ametysObject == null)
        {
            Request request = ContextHelper.getRequest(_context);
            throw new IllegalStateException("Cannot resolve a local link to an unknown ametys object for uri " + request.getRequestURI());
        }
        
        boolean hasToSwitch = _hasToSwitchRevision(info);
        Optional<String> currentRevision = Optional.empty();
        if (hasToSwitch && ametysObject instanceof VersionAwareAmetysObject versionAwareAO)
        {
            // Keep current revision to switch again later
            currentRevision = Optional.ofNullable(versionAwareAO.getRevision());
            
            // Switch to targeted revision
            versionAwareAO.switchToRevision(info.getAmetysObjectVersion().orElse(null));
        }
        
        return new AmetysObjectAndRevision(ametysObject, hasToSwitch, currentRevision);
    }
    
    /**
     * Checks in {@link AttributeInfo} if the ametys object has the same revision than the targeted one 
     * @param infos the {@link AttributeInfo} containing the ametys object and the targeted revision (=ametysObjectVersion) 
     * @return <code>true</code> if a switch is needed, <code>false</code> otherwise
     */
    private boolean _hasToSwitchRevision(AttributeInfo infos)
    {
        ModelAwareDataAwareAmetysObject ametysObject = infos.getAmetysObject();
        
        if (ametysObject instanceof VersionAwareAmetysObject versionAwareAO)
        {
            String currentRevision = versionAwareAO.getRevision();
            Optional<String> targetRevision = infos.getAmetysObjectVersion();
            
            return currentRevision == null && targetRevision.isPresent() // ametys object is at latest revision but another one has been targeted 
                || targetRevision.isEmpty()                              // ametys object is at specific revision but latest has been targeted
                || !targetRevision.get().equals(currentRevision);        // ametys object is at specific revision but another one has been targeted
        }
        
        return false;
    }
    
    private void _resetAmetysObjectRevisionIfNeeded(AmetysObjectAndRevision ametysObjectAndRevision)
    {
        if (ametysObjectAndRevision.hasSwitched() && ametysObjectAndRevision.ametysObject() instanceof VersionAwareAmetysObject versionAwareAO)
        {
            // Switch again the content revision
            Optional<String> oldRevision = ametysObjectAndRevision.revision();
            versionAwareAO.switchToRevision(oldRevision.orElse(null));
        }
    }
    
    /**
     * Get the URI prefix
     * @param path the resource path
     * @param object The object
     * @param download true if the pointed resource is to be downloaded.
     * @param absolute true if the url must be absolute
     * @param internal true to get an internal URI.
     * @return the URI prefix
     */
    protected String getUri(String path, ModelAwareDataAwareAmetysObject object, boolean download, boolean absolute, boolean internal)
    {
        if (internal)
        {
            return "cocoon://" + path;
        }
        else 
        {
            Request request = ContextHelper.getRequest(_context);
            String workspaceURI = (String) request.getAttribute(WorkspaceMatcher.WORKSPACE_URI);
            String uriPrefix = request.getContextPath() + workspaceURI;
            
            if (absolute && !uriPrefix.startsWith(request.getScheme()))
            {
                uriPrefix = request.getScheme() + "://" + request.getServerName() + (request.getServerPort() != 80 ? ":" + request.getServerPort() : "") + uriPrefix;
            }
            
            return uriPrefix + path;
        }
    }
    
    @Override
    public CHECK checkLink(String uri, boolean shortTest)
    {
        return CHECK.SUCCESS;
    }
    
    @Override
    public I18nizableText getLabel(String uri)
    {
        try
        {
            Request request = ContextHelper.getRequest(_context);
            
            AttributeInfo info = _getAttributeInfo(uri, request);
            
            return new I18nizableText("plugin.cms", "PLUGINS_CMS_LINK_METADATA_LABEL", Collections.singletonList(info.getPath()));
        }
        catch (UnknownAmetysObjectException e)
        {
            return new I18nizableText("plugin.cms", "PLUGINS_CMS_LINK_METADATA_UNKNOWN");
        }
    }
    
    /**
     * Get attribute path and ametys object.
     * @param uri the attribute URI.
     * @param request the request.
     * @return the attribute info.
     */
    protected AttributeInfo _getAttributeInfo(String uri, Request request)
    {
        AttributeInfo info = new AttributeInfo();
        
        int i = uri.indexOf('?');

        String attributePath = i < 0 ? uri : uri.substring(0, i);
        info.setPath(attributePath);

        List<NameValuePair> queryParams = Optional.of(i)
                                                  .filter(index -> index >= 0)
                                                  .map(index -> uri.substring(index + 1))
                                                  .map(queryParamsAsString -> URLEncodedUtils.parse(queryParamsAsString, StandardCharsets.UTF_8))
                                                  .orElseGet(ArrayList::new);
        
        ModelAwareDataAwareAmetysObject ametysObject = (ModelAwareDataAwareAmetysObject) request.getAttribute(Content.class.getName());

        Optional<String> objectId = _getQueryParamValue(queryParams, __OBJECT_ID_URI_QUERY_PARAM_NAME)
                          .or(() -> _getQueryParamValue(queryParams, __CONTENT_ID_URI_QUERY_PARAM_NAME));
        if (objectId.isPresent() && (ametysObject == null || !objectId.get().equals(ametysObject.getId())))
        {
            ametysObject = _resolver.resolveById(objectId.get());
        }

        info.setAmetysObject(ametysObject);

        _getQueryParamValue(queryParams, "contentVersion").ifPresent(info::setAmetysObjectVersion);
        
        return info;
    }
    
    private Optional<String> _getQueryParamValue(List<NameValuePair> queryParams, String queryParamName)
    {
        return queryParams.stream()
                          .filter(pair -> queryParamName.equals(pair.getName()))
                          .map(NameValuePair::getValue)
                          .findFirst();
    }

    /**
     * Attribute information.
     */
    protected class AttributeInfo
    {
        private String _path;
        private ModelAwareDataAwareAmetysObject _ametysObject;
        private Optional<String> _ametysObjectVersion = Optional.empty();
        
        /**
         * Get the attribute path.
         * @return the attribute path
         */
        public String getPath()
        {
            return _path;
        }
        
        /**
         * Set the attribute path.
         * @param path the attribute path to set
         */
        public void setPath(String path)
        {
            _path = path;
        }
        
        /**
         * Get the ametys object.
         * @return the ametys object
         */
        public ModelAwareDataAwareAmetysObject getAmetysObject()
        {
            return _ametysObject;
        }
        
        /**
         * Set the ametys object.
         * @param ametysObject the {@link AmetysObject} to set
         */
        public void setAmetysObject(ModelAwareDataAwareAmetysObject ametysObject)
        {
            _ametysObject = ametysObject;
        }
        
        /**
         * Get the ametys object's version.
         * @return the ametys object's version
         */
        public Optional<String> getAmetysObjectVersion()
        {
            return _ametysObjectVersion;
        }
        
        /**
         * Set the ametys object's version.
         * @param ametysObjectVersion the version to set
         */
        public void setAmetysObjectVersion(String ametysObjectVersion)
        {
            _ametysObjectVersion = Optional.ofNullable(ametysObjectVersion);
        }
    }
}
