/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.transformation;

import java.util.HashMap;
import java.util.Map;

import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;

import org.ametys.core.util.I18nUtils;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ElementType;

/**
 * ContentHandler able to interpret configuration and transform it to XML value.
 * 
 * For example it will transform an element with an attribute type="i18n" to the translated message.
 * 
 * Currently interpretation are made for i18n and config.
 * 
 * In the case of i18n, the translated message will be processed by a parser, allowing to provide rich text value through configuration.
 */
public class Configuration2XMLValuesTransformer implements ContentHandler
{
    private DataContext _context;
    private I18nUtils _i18nUtils;
    private String _configType;
    private ContentHandler _contentHandler;
    
    private StringBuilder _charBuffer;
    private Map<String, Object> _configTypeParams;
    private boolean _transformToDocbook;

    /**
     * Construct a content handler able to interpret configuration element.
     * 
     * For example, the text content of an element with attribute type="i18n" will be replace by the corresponding translated message
     * 
     * @param contentHandler sax event will be forwarded to this content handler after possible interpretation
     * @param context the data context to use when interpreting the input
     * @param i18nUtils the I18nUtils component. Use for translation
     */
    public Configuration2XMLValuesTransformer(ContentHandler contentHandler, DataContext context, I18nUtils i18nUtils)
    {
        _contentHandler = contentHandler;
        _context = context;
        _i18nUtils = i18nUtils;
    }

    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        if ("true".equals(atts.getValue("transform-to-docbook")))
        {
            _transformToDocbook = true;
            
            _charBuffer = new StringBuilder();
        }
        
        // local reference in case we need to edit the attributes
        Attributes attrs = atts;
        if (_is18n(atts))
        {
            _configType = "i18n";
            _configTypeParams = new HashMap<>();
            
            _configTypeParams.put("catalogue", atts.getValue("catalogue"));
            
            _charBuffer = new StringBuilder();
            
            // clean attributes
            AttributesImpl attrsCopy = new AttributesImpl(atts);
            attrsCopy.removeAttribute("i18n");
            attrsCopy.removeAttribute("type");
            attrsCopy.removeAttribute("catalogue");
            
            attrs = attrsCopy;
            _contentHandler.startElement(uri, localName, qName, attrs);
        }
        else if (_isConfig(atts))
        {
            _configType = "config";
            
            _charBuffer = new StringBuilder();
            
            // clean attributes
            AttributesImpl attrsCopy = new AttributesImpl(atts);
            attrsCopy.removeAttribute("type");
            
            attrs = attrsCopy;
            // Do not start element, element will be created in the end element
        }
        else
        {
            _contentHandler.startElement(uri, localName, qName, attrs);
        }
    }

    private boolean _is18n(Attributes atts)
    {
        return "true".equals(atts.getValue("i18n")) || "i18n".equals(atts.getValue("type"));
    }

    private boolean _isConfig(Attributes atts)
    {
        return "config".equals(atts.getValue("type"));
    }
    
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        if (_configType != null || _transformToDocbook)
        {
            for (int i = start; i < start + length; i++)
            {
                _charBuffer.append(ch[i]);
            }
        }
        else
        {
            _contentHandler.characters(ch, start, length);
        }
    }

    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if (_configType != null)
        {
            switch (_configType)
            {
                case "i18n":
                    _configType = null;
                    String translation = _i18nUtils.translate(new I18nizableText((String) _configTypeParams.get("catalogue"), _charBuffer.toString()), _context.getLocale().getLanguage());
                    
                    if (translation != null)
                    {
                        if (_transformToDocbook)
                        {
                            saxStringAsDockbook(translation);
                        }
                        else
                        {
                            _contentHandler.characters(translation.toCharArray(), 0, translation.length());
                        }
                    }

                    _contentHandler.endElement(uri, localName, qName);
                    break;
                case "config":
                    _configType = null;
                    
                    Model configModel = Config.getModel();
                    if (configModel != null)
                    {
                        String configId = _charBuffer.toString();
                        ModelItem configDefinition = configModel.getChild(configId);
                        if (configDefinition != null && configDefinition instanceof ElementType elementType)
                        {
                            elementType.valueToSAX(_contentHandler, localName, Config.getInstance().getValue(configId), _context);
                        }
                        else
                        {
                            throw new SAXException("No config element is defined for id " + configId + ".");
                        }
                    }
                    break;
                default:
                    throw new SAXException("Configuration type " + _configType + " is not supported.");
            }
        }
        else
        {
            // in the case where we want to transform raw data
            if (_transformToDocbook)
            {
                saxStringAsDockbook(_charBuffer.toString());
            }
            _contentHandler.endElement(uri, localName, qName);
        }
    }

    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
    {
        if (_configType != null)
        {
            for (int i = start; i < start + length; i++)
            {
                _charBuffer.append(ch[i]);
            }
        }
        else
        {
            if (_transformToDocbook)
            {
                saxStringAsDockbook(_charBuffer.toString());
            }
            _contentHandler.ignorableWhitespace(ch, start, length);
        }
    }
    
    private void saxStringAsDockbook(String string) throws SAXException
    {
        _transformToDocbook = false;
        
        _contentHandler.startPrefixMapping("", "http://docbook.org/ns/docbook");
        _contentHandler.startPrefixMapping("xlink", "http://www.w3.org/1999/xlink");
        _contentHandler.startPrefixMapping("html", "http://www.w3.org/1999/xhtml");
        
        AttributesImpl atts = new AttributesImpl();
        atts.addCDATAAttribute("version", "5.0");
        XMLUtils.startElement(_contentHandler, "article", atts);
        XMLUtils.createElement(_contentHandler, "para", string);
        XMLUtils.endElement(_contentHandler, "article");
        
        _contentHandler.endPrefixMapping("");
        _contentHandler.endPrefixMapping("xlink");
        _contentHandler.endPrefixMapping("html");
    }

    public void setDocumentLocator(Locator locator)
    {
        // ignore
    }
    
    public void startDocument() throws SAXException
    {
        // ignore
    }
    
    public void endDocument() throws SAXException
    {
        // ignore
    }
    
    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {
        // ignore
    }
    
    public void endPrefixMapping(String prefix) throws SAXException
    {
        // ignore
    }

    public void processingInstruction(String target, String data) throws SAXException
    {
        // ignore
    }

    public void skippedEntity(String name) throws SAXException
    {
        // ignore
    }

}
