/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.transformation;

import java.net.CookieHandler;
import java.net.CookieManager;
import java.net.CookiePolicy;
import java.util.Collection;
import java.util.Collections;
import java.util.Optional;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.core.util.HttpUrlUtils;
import org.ametys.core.util.HttpUrlUtils.HttpCheck;
import org.ametys.core.util.HttpUrlUtils.HttpCheckReport;
import org.ametys.core.version.Version;
import org.ametys.core.version.VersionsHandler;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Consistency checker component.
 * Checks the state of the links extracted as outgoing references for a content.
 */
public class ConsistencyChecker extends AbstractLogEnabled implements Serviceable, Component
{
    /** Avalon role */
    public static final String ROLE = ConsistencyChecker.class.getName();
    
    /** Constants for the special external type for outgoing references */
    public static final String CONSISTENCY_EXTERNAL_REFERENCE_TYPE = "__external";
    
    private static final int __HTTP_TIMEOUT = 10_000;
    
    /** The state of the check */
    public enum CHECK
    {
        /** If the check was all right */
        SUCCESS,
        /** Server Error. */
        SERVER_ERROR,
        /** Not Found.*/
        NOT_FOUND,
        /** Unauthorized. */
        UNAUTHORIZED,
        /** If the check may be too long */
        UNKNOWN;
        
        /**
         * Convert a {@link HttpCheck} value to a {@link CHECK} value
         * @param httpCheck the HTTP check
         * @return the CHECK value
         */
        public static CHECK from(HttpCheck httpCheck)
        {
            switch (httpCheck)
            {
                case SUCCESS:
                    return CHECK.SUCCESS;
                    
                case TIMEOUT:
                case REDIRECT:
                    return CHECK.UNKNOWN;
                  
                case NOT_FOUND:
                    return CHECK.NOT_FOUND;
                    
                case SECURITY_LEVEL_ERROR:
                case SERVER_ERROR:
                default:
                    return CHECK.SERVER_ERROR;
            }
        }
    }
    
    /** The ametys uri resolver */
    protected URIResolverExtensionPoint _uriResolverEP;
    /** The version handler */
    protected VersionsHandler _versionsHandler;
    /** The URL utils */
    protected HttpUrlUtils _urlUtils;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _uriResolverEP = (URIResolverExtensionPoint) manager.lookup(URIResolverExtensionPoint.ROLE);
        _versionsHandler = (VersionsHandler) manager.lookup(VersionsHandler.ROLE);
        _urlUtils = (HttpUrlUtils) manager.lookup(HttpUrlUtils.ROLE);
    }
    
    /**
     * Check the consistency of a reference
     * @param referenceType The type of the reference to test (can be 'attachment', 'explorer', 'metadata', '__external', etc...)
     * @param referenceValue The value of the reference to test
     * @param contentId The id of the content containing the reference
     * @param dataPath The path of the data containing the reference
     * @param quickTest true to make a quick test, that means that long tests will return UNKNOWN immediately.
     * @return CHECK enum value
     */
    public CheckReport checkConsistency(String referenceType, String referenceValue, String contentId, String dataPath, boolean quickTest)
    {
        if (getLogger().isDebugEnabled())
        {
            getLogger().debug("Checking consistency for URI of type '" + referenceType + "' with uri '" + referenceValue + "'");
        }
        
        URIResolver resolver = _uriResolverEP.getResolverForType(referenceType);
        if (resolver == null)
        {
            // No resolver for external references
            if (CONSISTENCY_EXTERNAL_REFERENCE_TYPE.equals(referenceType))
            {
                if (referenceValue.startsWith("http:") || referenceValue.startsWith("https:"))
                {
                    return _checkHTTPLink(referenceValue, quickTest);
                }
                else if (referenceValue.startsWith("mailto:"))
                {
                    return new CheckReport(CHECK.SUCCESS);
                }
                else if (referenceValue.startsWith("tel:"))
                {
                    return new CheckReport(CHECK.SUCCESS);
                }
            }
            
            if (getLogger().isDebugEnabled())
            {
                getLogger().debug("Cannot test external link '" + referenceValue + "'");
            }
            
            return new CheckReport(CHECK.UNKNOWN);
        }
        else
        {
            String uri = _getURIFromAttachment(referenceType, referenceValue, contentId, dataPath);
            return new CheckReport(resolver.checkLink(uri, quickTest));
        }
    }
    
    /**
     * Get the label of a reference.
     * @param referenceType The type of the reference to test (can be 'attachment', 'explorer', 'metadata', '__external', etc...)
     * @param referenceValue The value of the reference to test
     * @param contentId The id of the content containing the reference
     * @param dataPath The path of the data containing the reference
     * @return the element label.
     */
    public I18nizableText getLabel(String referenceType, String referenceValue, String contentId, String dataPath)
    {
        URIResolver resolver = _uriResolverEP.getResolverForType(referenceType);
        if (resolver == null)
        {
            return new I18nizableText("plugin.cms", "PLUGINS_CMS_LINK_EXTERNAL_LABEL", Collections.singletonList(referenceValue));
        }
        else
        {
            String uri = _getURIFromAttachment(referenceType, referenceValue, contentId, dataPath);
            return resolver.getLabel(uri);
        }
    }
    
    private String _getURIFromAttachment(String referenceType, String referenceValue, String contentId, String dataPath)
    {
        return "local".equals(referenceType) ? contentId + "@" + dataPath + ";" + referenceValue : referenceValue;
    }
    
    /**
     * Test an http link
     * @param linkValue The http url to test
     * @param shortTest true to make a short test (with short timeout)
     * @return The state. UNKNOWN if test cannot be done fully
     */
    protected CheckReport _checkHTTPLink(String linkValue, boolean shortTest)
    {
        try
        {
            CookieHandler.setDefault(new CookieManager(null, CookiePolicy.ACCEPT_ALL));
            
            String httpUrl = linkValue.replaceAll(" ", "%20");
            String userAgent = "Ametys/" + _getCMSVersion();
            int timeout = shortTest ? 1000 : __HTTP_TIMEOUT;
            int readTimeout = shortTest ? 2000 : __HTTP_TIMEOUT;
            HttpCheckReport checkReport = HttpUrlUtils.checkHttpUrl(httpUrl, userAgent, null, timeout, readTimeout, true);
            
            return new CheckReport(CHECK.from(checkReport.status()), checkReport.message());
        }
        finally
        {
            ((CookieManager) CookieHandler.getDefault()).getCookieStore().removeAll();
        }
    }
    
    private String _getCMSVersion()
    {
        Collection<Version> versions = _versionsHandler.getVersions();
        for (Version version : versions)
        {
            if (version.getName().equals("CMS"))
            {
                return version.getVersion();
            }
        }
        return null;
    }
    
    /**
     * Represent the result of a consistency check
     * @param status the result status
     * @param message a potential message
     */
    public record CheckReport(CHECK status, Optional<String> message) {
        /**
         * Create a report with no message
         * @param result the result status
         */
        public CheckReport(CHECK result)
        {
            this(result, Optional.empty());
        }
    }
}
