/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.transformation;

import java.io.InputStream;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import javax.jcr.RepositoryException;
import javax.jcr.Session;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;

import org.ametys.cms.URIPrefixHandler;
import org.ametys.cms.data.Resource;
import org.ametys.cms.data.RichText;
import org.ametys.cms.repository.Content;
import org.ametys.cms.transformation.ConsistencyChecker.CHECK;
import org.ametys.core.util.FilenameUtils;
import org.ametys.core.util.ImageResolverHelper;
import org.ametys.core.util.URIUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.plugins.repository.version.VersionAwareAmetysObject;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * {@link URIResolver} for resources local to a rich text.
 */
public class LocalURIResolver extends AbstractLogEnabled implements URIResolver, Contextualizable, Serviceable
{
    /** The context */
    protected Context _context;
    /** The ametys object resolver */
    protected AmetysObjectResolver _ametysObjectResolver;
    
    private URIPrefixHandler _prefixHandler;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _prefixHandler = (URIPrefixHandler) manager.lookup(URIPrefixHandler.ROLE);
    }
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public String getType()
    {
        return "local";
    }

    @Override
    public String resolve(String uri, boolean download, boolean absolute, boolean internal)
    {
        URIInfo infos = getInfos(uri, false, null);
        
        StringBuilder resultPath = new StringBuilder();
        
        resultPath.append(_prefixHandler.computeUriPrefix(absolute, internal))
                  .append("/plugins/cms/richText-file/")
                  .append(FilenameUtils.encodeName(infos.getFilename()));
        
        Map<String, String> params = new HashMap<>();
        params.put("contentId", infos.getContentId());
        params.put("attribute", infos.getAttribute());
        
        if (download)
        {
            params.put("download", "true");
        }
        
        infos.getContentVersion()
             .ifPresent(version -> params.put("contentVersion", version));
        
        // Encode twice
        return URIUtils.encodeURI(resultPath.toString(), params);
    }
    
    /**
     * Parses the uri.
     * @param uri the incoming uri.
     * @param resolveContent true if the Content should be actually resolved if not found in the request.
     * @param session the JCR {@link Session} to use, or null to use the current Session.
     * @return an object containing all parsed infos.
     */
    protected URIInfo getInfos(String uri, boolean resolveContent, Session session)
    {
        // uri are like <contentId>[|<contentVersion>]@<attribute>;<file.ext>
        int i = Math.max(uri.indexOf('|'), 0);
        int j = uri.indexOf('@', i);
        int k = uri.indexOf(';', j);
        String id = uri.substring(0, i > 0 ? i : j);
        Optional<String> contentVersion = i > 0 ? Optional.of(uri.substring(i + 1, j)) : Optional.empty();
        String attribute = uri.substring(j + 1, k);
        String filename = uri.substring(k + 1);

        Content content = null;
        
        try
        {
            Request request  = ContextHelper.getRequest(_context);
            content = (Content) request.getAttribute(Content.class.getName());
        }
        catch (Exception e)
        {
            // there's no request, thus no "current" content 
        }

        if (content == null || !id.equals(content.getId()))
        {
            // Some time (such as frontoffice edition) the image is rendered with no content in attribute
            // The content should be resolved against the given id, but in that case, getRevision will be always the head on
            try
            {
                content = resolveContent ? session == null ? _ametysObjectResolver.resolveById(id) : _ametysObjectResolver.resolveById(id, session) : null;
            }
            catch (RepositoryException e)
            {
                throw new AmetysRepositoryException(e);
            }
        }
        
        URIInfo infos = new URIInfo();
        infos.setContentId(id);
        infos.setContentVersion(contentVersion);
        infos.setContent(content);
        infos.setAttribute(attribute);
        infos.setFilename(filename);
        
        return infos;
    }
    
    @Override
    public String resolveImage(String uri, int height, int width, boolean download, boolean absolute, boolean internal)
    {
        return resolve(uri, download, absolute, internal);
    }
    
    @Override
    public String resolveImageAsBase64(String uri, int height, int width)
    {
        return resolveImageAsBase64(uri, height, width, 0, 0, 0, 0);
    }
    
    @Override
    public String resolveBoundedImage(String uri, int maxHeight, int maxWidth, boolean download, boolean absolute, boolean internal)
    {
        return resolve(uri, download, absolute, internal);
    }
    
    @Override
    public String resolveBoundedImageAsBase64(String uri, int maxHeight, int maxWidth)
    {
        return resolveImageAsBase64(uri, 0, 0, maxHeight, maxWidth, 0, 0);
    }
    
    @Override
    public String resolveCroppedImage(String uri, int cropHeight, int cropWidth, boolean download, boolean absolute, boolean internal)
    {
        return resolve(uri, download, absolute, internal);
    }
    
    @Override
    public String resolveCroppedImageAsBase64(String uri, int cropHeight, int cropWidth)
    {
        return resolveImageAsBase64(uri, 0, 0, 0, 0, cropHeight, cropWidth);
    }

    public String getMimeType(String uri)
    {
        URIInfo infos = getInfos(uri, true, null);
        
        ContentAndRevision contentAndRevision = _getContentAndRevisionFromAttributeInfo(infos);
        Content content = contentAndRevision.content();
        
        try
        {
            RichText richText = content.getValue(infos.getAttribute());
            Resource resource = richText.getAttachment(infos.getFilename());
            return resource.getMimeType();
        }
        catch (Exception e)
        {
            throw new IllegalStateException(e);
        }
        finally
        {
            _resetContentRevisionIfNeeded(contentAndRevision);
        }
    }
    
    private record ContentAndRevision (Content content, boolean hasSwitched, Optional<String> revision) { /* empty */ }
    
    /**
     * Get an image's bytes encoded as base64, optionally resized. 
     * @param uri the image URI.
     * @param height the specified height. Ignored if negative.
     * @param width the specified width. Ignored if negative.
     * @param maxHeight the maximum image height. Ignored if height or width is specified.
     * @param maxWidth the maximum image width. Ignored if height or width is specified.
     * @param cropHeight The cropping height. Ignored if negative.
     * @param cropWidth The cropping width. Ignored if negative.
     * @return the image bytes encoded as base64.
     */
    protected String resolveImageAsBase64(String uri, int height, int width, int maxHeight, int maxWidth, int cropHeight, int cropWidth)
    {
        URIInfo infos = getInfos(uri, true, null);
        
        ContentAndRevision contentAndRevision = _getContentAndRevisionFromAttributeInfo(infos);
        Content content = contentAndRevision.content();
        
        try
        {
            RichText richText = content.getValue(infos.getAttribute());
            Resource resource = richText.getAttachment(infos.getFilename());
            
            try (InputStream dataIs = resource.getInputStream())
            {
                return ImageResolverHelper.resolveImageAsBase64(dataIs, resource.getMimeType(), height, width, maxHeight, maxWidth, cropHeight, cropWidth);
            }
        }
        catch (Exception e)
        {
            throw new IllegalStateException(e);
        }
        finally
        {
            _resetContentRevisionIfNeeded(contentAndRevision);
        }
    }
    
    @Override
    public CHECK checkLink(String uri, boolean shortTest)
    {
        URIInfo infos = getInfos(uri, true, null);
        
        ContentAndRevision contentAndRevision = _getContentAndRevisionFromAttributeInfo(infos);
        Content content = contentAndRevision.content();

        try
        {
            RichText richText = content.getValue(infos.getAttribute());
            return richText != null
                 ? richText.hasAttachment(infos.getFilename())
                         ? CHECK.SUCCESS
                         : CHECK.NOT_FOUND
                 : CHECK.NOT_FOUND;
        }
        catch (Exception e)
        {
            throw new RuntimeException("Cannot check the uri '" + uri + "'", e);
        }
        finally
        {
            _resetContentRevisionIfNeeded(contentAndRevision);
        }
    }
    
    private ContentAndRevision _getContentAndRevisionFromAttributeInfo(URIInfo info)
    {
        Content content = info.getContent();
        
        boolean hasToSwitch = _hasToSwitchRevision(info);
        Optional<String> currentRevision = Optional.empty();
        if (hasToSwitch && content instanceof VersionAwareAmetysObject versionAwareAO)
        {
            // Keep current revision to switch again later
            currentRevision = Optional.ofNullable(versionAwareAO.getRevision());
            
            // Switch to targeted revision
            versionAwareAO.switchToRevision(info.getContentVersion().orElse(null));
        }
        
        return new ContentAndRevision(content, hasToSwitch, currentRevision);
    }
    
    /**
     * Checks in {@link URIInfo} if the content has the same revision than the targeted one 
     * @param info the {@link URIInfo} containing the content and the targeted revision (=contentVersion) 
     * @return <code>true</code> if a switch is needed, <code>false</code> otherwise
     */
    private boolean _hasToSwitchRevision(URIInfo info)
    {
        Content content = info.getContent();
        
        if (content instanceof VersionAwareAmetysObject versionAwareAO)
        {
            String currentRevision = versionAwareAO.getRevision();
            Optional<String> targetRevision = info.getContentVersion();
            
            return currentRevision == null && targetRevision.isPresent() // content is at latest revision but another one has been targeted 
                || targetRevision.isEmpty()                              // content is at specific revision but latest has been targeted
                || !targetRevision.get().equals(currentRevision);        // content is at specific revision but another one has been targeted
        }
        
        return false;
    }
    
    private void _resetContentRevisionIfNeeded(ContentAndRevision contentAndRevision)
    {
        if (contentAndRevision.hasSwitched() && contentAndRevision.content() instanceof VersionAwareAmetysObject versionAwareAO)
        {
            // Switch again the content revision
            Optional<String> oldRevision = contentAndRevision.revision();
            versionAwareAO.switchToRevision(oldRevision.orElse(null));
        }
    }
    
    @Override
    public I18nizableText getLabel(String uri)
    {
        try
        {
            URIInfo infos = getInfos(uri, false, null);
            
            return new I18nizableText("plugin.cms", "PLUGINS_CMS_LINK_LOCAL_LABEL", Collections.singletonList(infos.getFilename()));
        }
        catch (UnknownAmetysObjectException e)
        {
            return new I18nizableText("plugin.cms", "PLUGINS_CMS_LINK_LOCAL_UNKNOWN");
        }
    }
    
    /**
     * Helper class containg all infos parsed from URI.
     */
    protected static class URIInfo
    {
        /** The content id. */
        private String _contentId;
        /** The relevant attribute */
        private String _attribute;
        /** The resource name */
        private String _filename;
        /** The content version, if any. */
        private Optional<String> _contentVersion = Optional.empty();
        /** The resolved content, if any. */
        private Content _content;
        
        /**
         * Returns the content id.
         * @return the content id.
         */
        public String getContentId()
        {
            return _contentId;
        }
        
        /**
         * Set the content id.
         * @param contentId the content id.
         */
        public void setContentId(String contentId)
        {
            _contentId = contentId;
        }
        
        /**
         * Returns the attribute.
         * @return the attribute.
         */
        public String getAttribute()
        {
            return _attribute;
        }
        
        /**
         * Set the attribute.
         * @param attribute the attribute.
         */
        public void setAttribute(String attribute)
        {
            _attribute = attribute;
        }
        
        /**
         * Returns the resource name.
         * @return the name
         */
        public String getFilename()
        {
            return _filename;
        }
        
        /**
         * Set the resource name.
         * @param name the name.
         */
        public void setFilename(String name)
        {
            _filename = name;
        }
        
        /**
         * Returns the content version, if any.
         * @return the content version.
         */
        public Optional<String> getContentVersion()
        {
            return _contentVersion;
        }
        
        /**
         * Set the content version.
         * @param contentVersion the content version.
         */
        public void setContentVersion(Optional<String> contentVersion)
        {
            _contentVersion = contentVersion;
        }
        
        /**
         * Returns the resolved content, if any.
         * @return the content.
         */
        public Content getContent()
        {
            return _content;
        }
        
        /**
         * Set the content.
         * @param content the content.
         */
        public void setContent(Content content)
        {
            _content = content;
        }
    }
}
