/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.transformation;

import java.io.InputStream;
import java.util.Collections;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang.StringUtils;

import org.ametys.cms.URIPrefixHandler;
import org.ametys.cms.transformation.ConsistencyChecker.CHECK;
import org.ametys.core.util.FilenameUtils;
import org.ametys.core.util.ImageResolverHelper;
import org.ametys.core.util.URIUtils;
import org.ametys.plugins.explorer.resources.Resource;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.plugins.repository.version.VersionableAmetysObject;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * {@link URIResolver} for type "explorer". <br>
 * These links point to a file from the resources explorer.
 */
public class ResourceURIResolver extends AbstractURIResolver implements Serviceable, Contextualizable
{
    /** The ametys object resolver. */
    protected AmetysObjectResolver _resolver;
    /** The avalon context. */
    protected Context _context;
    /** The URI prefix handler */
    protected URIPrefixHandler _prefixHandler;
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _prefixHandler = (URIPrefixHandler) manager.lookup(URIPrefixHandler.ROLE);
    }
    
    @Override
    public String getType()
    {
        return "explorer";
    }
    
    /**
     * Get the URI prefix
     * @param object The object
     * @param download true if the pointed resource is to be downloaded.
     * @param absolute true if the url must be absolute
     * @param internal true to get an internal URI.
     * @return the URI prefix
     */
    protected String getUriPrefix (AmetysObject object, boolean download, boolean internal, boolean absolute)
    {
        return _prefixHandler.computeUriPrefix(absolute, internal);
    }
    
    @Override
    protected String _resolve(String uri, String uriArgument, boolean download, boolean absolute, boolean internal)
    {
        String version = null;
        String path;
        Resource resource = null;
        try
        {
            String resourceId = uri;
            int i = uri.indexOf(";");
            if (i != -1)
            {
                resourceId = uri.substring(0, i);
                version = uri.substring(i + 1);
            }
            
            resource = _resolver.resolveById(resourceId);
            path = getResourcePath(resource);
        }
        catch (UnknownAmetysObjectException e)
        {
            getLogger().warn("Link to unexisting resource " + uri);
            return "";
        }
        
        String filename = FilenameUtils.encodePath(path);
        String baseName = org.apache.commons.io.FilenameUtils.removeExtension(filename);
        String extension = org.apache.commons.io.FilenameUtils.getExtension(filename);
        
        StringBuilder result = new StringBuilder();
        
        result.append(getUriPrefix(resource, download, absolute, internal));
        
        String realPrefix = getRealPrefix(resource, "resource");
        if (StringUtils.isNotEmpty(realPrefix))
        {
            result.append("/_").append(realPrefix);
        }
        
        if (StringUtils.isNotEmpty(version) && resource instanceof VersionableAmetysObject)
        {
            result.append("/__v_").append(version);
        }
        
        result.append(baseName)
              .append(uriArgument)
              .append(extension.isEmpty() ? "" : "." + extension);
        
        return URIUtils.encodeURI(result.toString(), download ? Collections.singletonMap("download", "true") : null);
    }
    
    /**
     * Get the resource path 
     * @param resource the resource
     * @return the path
     */
    protected String getResourcePath (Resource resource)
    {
        return resource.getResourcePath();
    }
    
    /**
     * Get the real prefix
     * @param resource the resource
     * @param prefix the initial prefix
     * @return the real prefix
     */
    protected String getRealPrefix (Resource resource, String prefix)
    {
        return prefix;
    }
    
    @Override
    protected String resolveImageAsBase64(String uri, int height, int width, int maxHeight, int maxWidth, int cropHeight, int cropWidth)
    {
        try
        {
            Resource resource = _resolver.resolveById(uri);
            try (InputStream dataIs = resource.getInputStream())
            {
                return ImageResolverHelper.resolveImageAsBase64(dataIs, resource.getMimeType(), height, width, maxHeight, maxWidth, cropHeight, cropWidth);
            }
        }
        catch (UnknownAmetysObjectException e)
        {
            getLogger().warn("Link to unexisting resource " + uri);
            return "";
        }
        catch (Exception e)
        {
            throw new IllegalStateException(e);
        }
    }

    public String getMimeType(String uri)
    {
        try
        {
            Resource resource = _resolver.resolveById(uri);
            return resource.getMimeType();
        }
        catch (UnknownAmetysObjectException e)
        {
            getLogger().warn("Link to unexisting resource " + uri);
            return "";
        }
        catch (Exception e)
        {
            throw new IllegalStateException(e);
        }
    }

    @Override
    public CHECK checkLink(String uri, boolean shortTest)
    {
        try
        {
            _resolver.resolveById(uri);
            return CHECK.SUCCESS;
        }
        catch (UnknownAmetysObjectException e)
        {
            return CHECK.NOT_FOUND;
        }
    }
    
    @Override
    public I18nizableText getLabel(String uri)
    {
        try
        {
            Resource resource = _resolver.resolveById(uri);
            return new I18nizableText("plugin.cms", "PLUGINS_CMS_LINK_RESOURCE_LABEL", Collections.singletonList(resource.getResourcePath()));
        }
        catch (UnknownAmetysObjectException e)
        {
            return new I18nizableText("plugin.cms", "PLUGINS_CMS_LINK_RESOURCE_UNKNOWN");
        }
    }
}
