/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.transformation;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * Resolver for link uris put by the inline editor in rich text fields.
 */
public interface URIResolver
{
    /**
     * Returns the type of links handled by this {@link URIResolver}.
     * @return the type of links handled by this {@link URIResolver}.
     */
    public String getType();
    
    /**
     * Resolves a link URI for rendering purposes.<br>
     * The output must be a properly encoded path, relative to the webapp context, accessible from a browser.
     * @param uri the link URI.
     * @param download true if the pointed resource is to be downloaded.
     * @param absolute true if the url must be absolute
     * @param internal true to get an internal URI.
     * @return the path to the resource.
     */
    public String resolve(String uri, boolean download, boolean absolute, boolean internal);
    
    /**
     * Resolves a link URI for rendering image.<br>
     * The output must be a properly encoded path, relative to the webapp context, accessible from a browser.
     * @param uri the link URI.
     * @param height the height of the image. 
     * @param width the width of the image
     * @param download true if the pointed resource is to be downloaded.
     * @param absolute true if the url must be absolute
     * @param internal true to get an internal URI.
     * @return the path to the image.
     */
    public String resolveImage(String uri, int height, int width, boolean download, boolean absolute, boolean internal);
    
    /**
     * Resolve an image and return it as a base64-encoded string.
     * @param uri the link URI.
     * @param height the height of the image. 
     * @param width the width of the image
     * @return a base64-encoded string representing the image.
     */
    public String resolveImageAsBase64(String uri, int height, int width);
    
    /**
     * Resolves a link URI for rendering image.<br>
     * The output must be a properly encoded path, relative to the webapp context, accessible from a browser.
     * @param uri the link URI.
     * @param maxHeight the max height
     * @param maxWidth the max width
     * @param download true if the pointed resource is to be downloaded.
     * @param absolute true if the url must be absolute
     * @param internal true to get an internal URI.
     * @return the path to the image.
     */
    public String resolveBoundedImage(String uri, int maxHeight, int maxWidth, boolean download, boolean absolute, boolean internal);
    
    /**
     * Resolve an image and return it as a base64-encoded string.
     * @param uri the link URI.
     * @param maxHeight the max height
     * @param maxWidth the max width
     * @return a base64-encoded string representing the image.
     */
    public String resolveBoundedImageAsBase64(String uri, int maxHeight, int maxWidth);
    
    /**
     * Resolves a link URI for rendering image.<br>
     * The output must be a properly encoded path, relative to the webapp context, accessible from a browser.
     * @param uri the link URI.
     * @param cropHeight the crop height
     * @param cropWidth the crop width
     * @param download true if the pointed resource is to be downloaded.
     * @param absolute true if the url must be absolute
     * @param internal true to get an internal URI.
     * @return the path to the image.
     */
    public String resolveCroppedImage(String uri, int cropHeight, int cropWidth, boolean download, boolean absolute, boolean internal);
    
    /**
     * Resolve an image and return it as a base64-encoded string.
     * @param uri the link URI.
     * @param cropHeight the crop height
     * @param cropWidth the crop width
     * @return a base64-encoded string representing the image.
     */
    public String resolveCroppedImageAsBase64(String uri, int cropHeight, int cropWidth);
    
    /**
     * Tests if the URI is not broken.
     * @param uri The uri to test
     * @param shortTest If shortTest is to true, only fast check should be done. If false, complete test (even if long) should be done.
     * @return SUCCESS if the test was done and ok, FAILURE if the test was done and ko, UNKNOWN if the test was not done (because too long). UNKNOWN can not be returned when shortTest is set to false
     */
    public ConsistencyChecker.CHECK checkLink(String uri, boolean shortTest);
    
    /**
     * Get an URI label.
     * @param uri the uri which label to get.
     * @return the label.
     */
    public I18nizableText getLabel(String uri);

    /**
     * Get the mime type for the targeted URI.
     * @param uri the link URI.
     * @return the mime type, can be empty or null.
     */
    public default String getMimeType(String uri)
    {
        return "application/octet-stream";
    }
}
