/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.transformation.docbook;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashSet;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.commons.io.IOUtils;

import org.ametys.runtime.plugin.AbstractExtensionPoint;

/**
 * This extension point handle the new tags you want to add to the default implementation of the docbook 
 */
public class DocbookEnhancementExtensionPoint extends AbstractExtensionPoint<DocbookEnhancer>
{
    /** The avalon role */
    public static final String ROLE = DocbookEnhancementExtensionPoint.class.getName();
    
    private String _docbook2htmleditorXSLT;
    private String _docbook2htmlXSLT;
    private String _htmleditor2docbookXSLT;
    private String _docbook2outgoingReferencesXSLT;
    private Set<String> _docbook2htmlHandlers;
    private Set<String> _htmleditor2docbookHandlers;
    private Set<String> _docbook2htmleditorHandlers;
    
    /**
     * Get the content of the XSLT file
     * @return A valid XSLT file content
     */
    public String getDocbook2htmleditorXSLT()
    {
        return _docbook2htmleditorXSLT;
    }
    
    /**
     * Get the content of the XSLT file
     * @return A valid XSLT file content
     */
    public String getDocbook2htmlXSLT()
    {
        return _docbook2htmlXSLT;
    }
    
    /**
     * Get the content of the XSLT file
     * @return A valid XSLT file content
     */
    public String getDocbook2outgoingReferencesXSLT()
    {
        return _docbook2outgoingReferencesXSLT;
    }
    
    /**
     * Get the content of the XSLT file
     * @return A valid XSLT file content
     */
    public String getHTMLEditor2docbookXSLT()
    {
        return _htmleditor2docbookXSLT;
    }
    
    /**
     * Get the list of available handlers
     * @return a non null set of roles
     */
    public Set<String> getDocbook2HTMLHandlers()
    {
        return _docbook2htmlHandlers;
    }
    
    /**
     * Get the list of available handlers
     * @return a non null set of roles
     */
    public Set<String> getHTMLEditor2DocbookHandlers()
    {
        return _htmleditor2docbookHandlers;
    }
    
    /**
     * Get the list of available handlers
     * @return a non null set of roles
     */
    public Set<String> getDocbook2HTMLEditorHandlers()
    {
        return _docbook2htmleditorHandlers;
    }

    @Override
    public void addExtension(String id, String pluginName, String featureName, Configuration configuration) throws ConfigurationException
    {
        String htmleditor2DocbookXSLT = configuration.getChild("htmleditor2docbook").getChild("xslt").getValue(null);
        String docbook2HTMLEditorXSLT = configuration.getChild("docbook2htmleditor").getChild("xslt").getValue(null);
        String docbook2HTMLXSLT = configuration.getChild("docbook2html").getChild("xslt").getValue(null);
        String docbook2outgoingReferencesXSLT = configuration.getChild("docbook2outgoingreferences").getChild("xslt").getValue(null);
        String docbook2HTMLHandler = configuration.getChild("docbook2html").getChild("handler").getValue(null);
        String htmlEditor2docbookHandler = configuration.getChild("htmleditor2docbook").getChild("handler").getValue(null);
        String docbook2htmlEditorHandler = configuration.getChild("docbook2htmleditor").getChild("handler").getValue(null);
        
        DocbookEnhancer de = new DocbookEnhancer(pluginName, htmleditor2DocbookXSLT, docbook2HTMLEditorXSLT, docbook2HTMLXSLT, docbook2outgoingReferencesXSLT, docbook2HTMLHandler, htmlEditor2docbookHandler, docbook2htmlEditorHandler);
        _extensions.put(id, de);
    }
    
    @Override
    public void initializeExtensions() throws Exception
    {
        String template = _readTemplate();
        
        Set<String> docbook2htmleditorXSLTs = _getDocbook2htmleditorXSLTs();
        String docbook2htmleditorXSLTsIncludes = _createIncludes(docbook2htmleditorXSLTs);
        _docbook2htmleditorXSLT = template.replace("@INCLUDE@", docbook2htmleditorXSLTsIncludes);

        Set<String> docbook2htmlXSLTs = _getDocbook2htmlXSLTs();
        String docbook2htmlXSLTsIncludes = _createIncludes(docbook2htmlXSLTs);
        _docbook2htmlXSLT = template.replace("@INCLUDE@", docbook2htmlXSLTsIncludes);

        Set<String> docbook2outgoingReferencesXSLTs = _getDocbook2outgoingReferencesXSLTs();
        String docbook2outgoingReferencesXSLTsIncludes = _createIncludes(docbook2outgoingReferencesXSLTs);
        _docbook2outgoingReferencesXSLT = template.replace("@INCLUDE@", docbook2outgoingReferencesXSLTsIncludes);

        Set<String> htmleditor2docbookXSLTs = _getHTMLEditor2docbookXSLTs();
        String htmleditor2docbookXSLTsIncludes = _createIncludes(htmleditor2docbookXSLTs);
        _htmleditor2docbookXSLT = template.replace("@INCLUDE@", htmleditor2docbookXSLTsIncludes);
        
        _docbook2htmlHandlers = new HashSet<>();
        for (String key : _extensions.keySet())
        {
            DocbookEnhancer de = _extensions.get(key);
         
            String handlerRole = de.getDocbook2HTMLHandler();
            if (handlerRole != null)
            {
                _docbook2htmlHandlers.add(handlerRole);
            }
        }
        
        _htmleditor2docbookHandlers = new HashSet<>();
        for (String key : _extensions.keySet())
        {
            DocbookEnhancer de = _extensions.get(key);
         
            String handlerRole = de.getHTMLEditor2DocbookHandler();
            if (handlerRole != null)
            {
                _htmleditor2docbookHandlers.add(handlerRole);
            }
        }
        
        _docbook2htmleditorHandlers = new HashSet<>();
        for (String key : _extensions.keySet())
        {
            DocbookEnhancer de = _extensions.get(key);
         
            String handlerRole = de.getDocbook2HTMLEditorHandler();
            if (handlerRole != null)
            {
                _docbook2htmleditorHandlers.add(handlerRole);
            }
        }
    }

    private String _createIncludes(Set<String> includes)
    {
        StringBuffer sb = new StringBuffer("");
        
        for (String file : includes)
        {
            sb.append("<xsl:include href=\"");
            sb.append(file);
            sb.append("\"/>");
        }
        
        return sb.toString();
    }
    
    private String _readTemplate() throws IOException
    {
        try (InputStream is = this.getClass().getResourceAsStream("template.xsl"))
        {
            return IOUtils.toString(is, "UTF-8");
        }
    }

    private Set<String> _getDocbook2htmlXSLTs()
    {
        Set<String> docbook2htmlXSLTs = new HashSet<>();

        for (String key : _extensions.keySet())
        {
            DocbookEnhancer de = _extensions.get(key);
            
            String docbook2html = de.getDocbook2HTMLXSLT();
            if (docbook2html != null)
            {
                docbook2htmlXSLTs.add(docbook2html);
            }
        }
        
        return docbook2htmlXSLTs;
    }
    
    private Set<String> _getDocbook2htmleditorXSLTs()
    {
        Set<String> docbook2htmleditorXSLTs = new HashSet<>();

        for (String key : _extensions.keySet())
        {
            DocbookEnhancer de = _extensions.get(key);
            
            String docbook2htmleditor = de.getDocbook2HTMLEditorXSLT();
            if (docbook2htmleditor != null)
            {
                docbook2htmleditorXSLTs.add(docbook2htmleditor);
            }
        }
        
        return docbook2htmleditorXSLTs;
    }
    
    private Set<String> _getDocbook2outgoingReferencesXSLTs()
    {
        Set<String> docbook2outgoingReferencesXSLTs = new HashSet<>();

        for (String key : _extensions.keySet())
        {
            DocbookEnhancer de = _extensions.get(key);
            
            String docbook2outgoingReferences = de.getDocbook2outgoingReferencesXSLT();
            if (docbook2outgoingReferences != null)
            {
                docbook2outgoingReferencesXSLTs.add(docbook2outgoingReferences);
            }
        }
        
        return docbook2outgoingReferencesXSLTs;
    }
    
    private Set<String> _getHTMLEditor2docbookXSLTs()
    {
        Set<String> htmleditor2docbookXSLTs = new HashSet<>();

        for (String key : _extensions.keySet())
        {
            DocbookEnhancer de = _extensions.get(key);
            
            String htmleditor2docbook = de.getHTMLEditor2DocbookXSLT();
            if (htmleditor2docbook != null)
            {
                htmleditor2docbookXSLTs.add(htmleditor2docbook);
            }
        }
        
        return htmleditor2docbookXSLTs;
    }
}
