/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.transformation.docbook;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.TransformerHandler;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.xml.sax.SAXParser;
import org.apache.excalibur.xml.xpath.XPathProcessor;
import org.apache.excalibur.xml.xslt.XSLTProcessor;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

import org.ametys.cms.content.references.OutgoingReferences;
import org.ametys.cms.content.references.RichTextOutgoingReferencesExtractor;
import org.ametys.cms.data.RichText;

/**
 * This impl extract outgoing references on docbook rich texts
 */
public class DocbookOutgoingReferencesExtractor implements RichTextOutgoingReferencesExtractor, Serviceable
{
    /** Avalon Role */
    public static final String ROLE = DocbookOutgoingReferencesExtractor.class.getName();

    /** Avalon service manager */
    protected ServiceManager _manager;
    
    private SourceResolver _sourceResolver;
    private XSLTProcessor _xsltProcessor;
    private XPathProcessor _xpathProcessor;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
        _xsltProcessor = (XSLTProcessor) manager.lookup(XSLTProcessor.ROLE + "/xalan");
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _xpathProcessor = (XPathProcessor) manager.lookup(XPathProcessor.ROLE);
    }

    /**
     * The uri to use for conversion to outgoing references format
     * @return The uri
     */
    protected String _getXSLTURI()
    {
        return "docbookenhancement://docbook2outgoingreferences.xsl";
    }
    
    /**
     * Take the metadata stream and transform it using the xslt configured (_getXSLTURI)
     * @param metadata The metadata to transform
     * @return The root node of the result XML after the transformation
     */
    protected Node _getNodeAfterXSLT(RichText metadata)
    {
        DOMResult result = null;
        Source xsltSource = null;
        SAXParser saxParser = null;
        
        try
        {
            xsltSource = _sourceResolver.resolveURI(_getXSLTURI());
            TransformerHandler th = _xsltProcessor.getTransformerHandler(xsltSource);
            
            result = new DOMResult();
            th.setResult(result);

            saxParser = (SAXParser) _manager.lookup(SAXParser.ROLE);
            
            try (InputStream is = metadata.getInputStream())
            {
                saxParser.parse(new InputSource(is), th);
            }
        }
        catch (Exception e)
        {
            throw new RuntimeException("Cannot find the xslt to extract rich text outgoing references", e);
        }
        finally
        {
            _sourceResolver.release(xsltSource);
            _manager.release(saxParser);
        }
        
        return result.getNode();
    }
    
    @Override
    public OutgoingReferences getOutgoingReferences(RichText richText)
    {
        OutgoingReferences references = new OutgoingReferences();
        
        Node resultNode = _getNodeAfterXSLT(richText);
        NodeList referenceNodes = _xpathProcessor.selectNodeList(resultNode, "/references/reference");
        
        for (int i = 0; i < referenceNodes.getLength(); i++)
        {
            Element referenceNode = (Element) referenceNodes.item(i);
            String type = referenceNode.getAttribute("type");
            String value = referenceNode.getTextContent();
            
            List<String> referenceValues = references.get(type);
            if (referenceValues == null)
            {
                referenceValues = new ArrayList<>();
                references.put(type, referenceValues);
            }
            referenceValues.add(value);
        }
        
        return references;
    }
}
