/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.transformation.html;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashSet;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.commons.io.IOUtils;

import org.ametys.runtime.plugin.AbstractExtensionPoint;

/**
 * This extension point handle the new tags you want to add to the default implementation of the docbook 
 */
public class HTMLEnhancementExtensionPoint extends AbstractExtensionPoint<HTMLEnhancer>
{
    /** The avalon role */
    public static final String ROLE = HTMLEnhancementExtensionPoint.class.getName();
    
    private String _html2htmleditorXSLT;
    private String _html2htmlXSLT;
    private String _html2outgoingReferencesXSLT;
    private String _htmleditor2htmlXSLT;
    private Set<String> _html2htmlHandlers;
    private Set<String> _htmleditor2htmlHandlers;
    private Set<String> _html2htmleditorHandlers;
    
    /**
     * Get the content of the XSLT file
     * @return A valid XSLT file content
     */
    public String getHtml2htmleditorXSLT()
    {
        return _html2htmleditorXSLT;
    }
    
    /**
     * Get the content of the XSLT file
     * @return A valid XSLT file content
     */
    public String getHtml2htmlXSLT()
    {
        return _html2htmlXSLT;
    }
    
    /**
     * Get the content of the XSLT file
     * @return A valid XSLT file content
     */
    public String getHtml2outgoingReferencesXSLT()
    {
        return _html2outgoingReferencesXSLT;
    }
    
    /**
     * Get the content of the XSLT file
     * @return A valid XSLT file content
     */
    public String getHtmlEditor2HTMLXSLT()
    {
        return _htmleditor2htmlXSLT;
    }
    
    /**
     * Get the list of available handlers
     * @return a non null set of roles
     */
    public Set<String> getHtml2HTMLHandlers()
    {
        return _html2htmlHandlers;
    }
    
    /**
     * Get the list of available handlers
     * @return a non null set of roles
     */
    public Set<String> getHtmleditor2HTMLHandlers()
    {
        return _htmleditor2htmlHandlers;
    }
    
    /**
     * Get the list of available handlers
     * @return a non null set of roles
     */
    public Set<String> getHtml2HTMLEditorHandlers()
    {
        return _html2htmleditorHandlers;
    }
    
    @Override
    public void addExtension(String id, String pluginName, String featureName, Configuration configuration) throws ConfigurationException
    {
        String htmleditor2HtmlXSLT = configuration.getChild("htmleditor2html").getChild("xslt").getValue(null);
        String html2HTMLEditorXSLT = configuration.getChild("html2htmleditor").getChild("xslt").getValue(null);
        String html2HTMLXSLT = configuration.getChild("html2html").getChild("xslt").getValue(null);
        String html2outgoingReferencesXSLT = configuration.getChild("html2outgoingreferences").getChild("xslt").getValue(null);
        String html2HTMLHandler = configuration.getChild("html2html").getChild("handler").getValue(null);
        String htmleditor2HTMLHandler = configuration.getChild("htmleditor2html").getChild("handler").getValue(null);
        String html2HTMLEditorHandler = configuration.getChild("html2htmleditor").getChild("handler").getValue(null);
        
        HTMLEnhancer de = new HTMLEnhancer(pluginName, htmleditor2HtmlXSLT, html2HTMLEditorXSLT, html2HTMLXSLT, html2outgoingReferencesXSLT, html2HTMLHandler, htmleditor2HTMLHandler, html2HTMLEditorHandler);
        _extensions.put(id, de);
    }
    
    @Override
    public void initializeExtensions() throws Exception
    {
        String template = _readTemplate();
        
        Set<String> html2htmleditorXSLTs = _getHtml2htmleditorXSLTs();
        String html2htmleditorXSLTsIncludes = _createIncludes(html2htmleditorXSLTs);
        _html2htmleditorXSLT = template.replace("@INCLUDE@", html2htmleditorXSLTsIncludes);

        Set<String> html2htmlXSLTs = _getHtml2htmlXSLTs();
        String html2htmlXSLTsIncludes = _createIncludes(html2htmlXSLTs);
        _html2htmlXSLT = template.replace("@INCLUDE@", html2htmlXSLTsIncludes);

        Set<String> htmleditor2htmlXSLTs = _getHTMLEditor2htmlXSLTs();
        String htmleditor2htmlXSLTsIncludes = _createIncludes(htmleditor2htmlXSLTs);
        _htmleditor2htmlXSLT = template.replace("@INCLUDE@", htmleditor2htmlXSLTsIncludes);
        
        Set<String> html2outgoingReferencesXSLTs = _getHtml2outgoingReferencesXSLTs();
        String html2outgoingReferencesXSLTsIncludes = _createIncludes(html2outgoingReferencesXSLTs);
        _html2outgoingReferencesXSLT = template.replace("@INCLUDE@", html2outgoingReferencesXSLTsIncludes);
        
        _html2htmlHandlers = new HashSet<>();
        for (String key : _extensions.keySet())
        {
            HTMLEnhancer de = _extensions.get(key);
         
            String handlerRole = de.getHtml2HTMLHandler();
            if (handlerRole != null)
            {
                _html2htmlHandlers.add(handlerRole);
            }
        }
        
        _htmleditor2htmlHandlers = new HashSet<>();
        for (String key : _extensions.keySet())
        {
            HTMLEnhancer de = _extensions.get(key);
         
            String handlerRole = de.getHtmleditor2HTMLHandler();
            if (handlerRole != null)
            {
                _htmleditor2htmlHandlers.add(handlerRole);
            }
        }
        
        _html2htmleditorHandlers = new HashSet<>();
        for (String key : _extensions.keySet())
        {
            HTMLEnhancer de = _extensions.get(key);
         
            String handlerRole = de.getHtml2HTMLEditorHandler();
            if (handlerRole != null)
            {
                _html2htmleditorHandlers.add(handlerRole);
            }
        }
    }

    private String _createIncludes(Set<String> includes)
    {
        StringBuffer sb = new StringBuffer("");
        
        for (String file : includes)
        {
            sb.append("<xsl:include href=\"");
            sb.append(file);
            sb.append("\"/>");
        }
        
        return sb.toString();
    }
    
    private String _readTemplate() throws IOException
    {
        try (InputStream is = this.getClass().getResourceAsStream("template.xsl"))
        {
            return IOUtils.toString(is, "UTF-8");
        }
    }

    private Set<String> _getHtml2htmlXSLTs()
    {
        Set<String> html2htmlXSLTs = new HashSet<>();

        for (String key : _extensions.keySet())
        {
            HTMLEnhancer de = _extensions.get(key);
            
            String html2html = de.getHtml2HTMLXSLT();
            if (html2html != null)
            {
                html2htmlXSLTs.add(html2html);
            }
        }
        
        return html2htmlXSLTs;
    }
    
    private Set<String> _getHtml2htmleditorXSLTs()
    {
        Set<String> html2htmleditorXSLTs = new HashSet<>();

        for (String key : _extensions.keySet())
        {
            HTMLEnhancer de = _extensions.get(key);
            
            String html2htmleditor = de.getHtml2HTMLEditorXSLT();
            if (html2htmleditor != null)
            {
                html2htmleditorXSLTs.add(html2htmleditor);
            }
        }
        
        return html2htmleditorXSLTs;
    }
    
    private Set<String> _getHtml2outgoingReferencesXSLTs()
    {
        Set<String> html2outgoingReferencesXSLTs = new HashSet<>();

        for (String key : _extensions.keySet())
        {
            HTMLEnhancer de = _extensions.get(key);
            
            String html2outgoingReferences = de.getHtml2outgoingReferencesXSLT();
            if (html2outgoingReferences != null)
            {
                html2outgoingReferencesXSLTs.add(html2outgoingReferences);
            }
        }
        
        return html2outgoingReferencesXSLTs;
    }
    
    private Set<String> _getHTMLEditor2htmlXSLTs()
    {
        Set<String> htmleditor2htmlXSLTs = new HashSet<>();

        for (String key : _extensions.keySet())
        {
            HTMLEnhancer de = _extensions.get(key);
            
            String htmleditor2html = de.getHTMLEditor2HtmlX();
            if (htmleditor2html != null)
            {
                htmleditor2htmlXSLTs.add(htmleditor2html);
            }
        }
        
        return htmleditor2htmlXSLTs;
    }
}
