/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.transformation.htmledition;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.ext.LexicalHandler;

/**
 * This abstract class is for component that wants to be in the saving pipeline.
 * As a pipeline you receive events (from the ContentHandler and LexicalHandler interfaces) 
 * you may transform it and write it to the given contentHandler (setContentHandler) and 
 * given lexicalHandler (setLexicalHandler)
 * You receive the html extracted from a htmleditor and you may operate some changes. 
 */
public abstract class AbstractHTMLEditionHandler extends AbstractLogEnabled implements ContentHandler, LexicalHandler, Serviceable, Contextualizable
{
    /** The content handler where to write */
    protected ContentHandler _contentHandler;
    /** The lexical handler where to write */
    protected LexicalHandler _lexicalHandler;
    /** The avalon service manager */
    protected ServiceManager _manager;
    /** The avalon context */
    protected Context _context;
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
    }
    
    /**
     * Set the target content handler
     * @param contentHandler the target content handler
     */
    public void setContentHandler(ContentHandler contentHandler)
    {
        _contentHandler = contentHandler;
    }
    
    /**
     * Set the target lexical handler
     * @param lexicalHandler the target lexical handler
     */
    public void setLexicalHandler(LexicalHandler lexicalHandler)
    {
        _lexicalHandler = lexicalHandler;
    }
    @Override
    public void startDocument() throws SAXException
    {
        _contentHandler.startDocument();
    }
    @Override
    public void endDocument() throws SAXException
    {
        _contentHandler.endDocument();
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        _contentHandler.startElement(uri, localName, qName, atts);
    }
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        _contentHandler.endElement(uri, localName, qName);
    }
    
    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        _contentHandler.characters(ch, start, length);
    }
    @Override
    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
    {
        _contentHandler.ignorableWhitespace(ch, start, length);
    }
    
    @Override
    public void processingInstruction(String target, String data) throws SAXException
    {
        _contentHandler.processingInstruction(target, data);
    }
    @Override
    public void setDocumentLocator(Locator locator)
    {
        _contentHandler.setDocumentLocator(locator);
    }
    @Override
    public void skippedEntity(String name) throws SAXException
    {
        _contentHandler.skippedEntity(name);
    }
    
    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {
        _contentHandler.startPrefixMapping(prefix, uri);
    }
    @Override
    public void endPrefixMapping(String prefix) throws SAXException
    {
        _contentHandler.endPrefixMapping(prefix);
    }
    
    @Override
    public void startEntity(String name) throws SAXException
    {
        _lexicalHandler.startEntity(name);
    }
    @Override
    public void endEntity(String name) throws SAXException
    {
        _lexicalHandler.endEntity(name);
    }

    @Override
    public void startCDATA() throws SAXException
    {
        _lexicalHandler.startCDATA();
    }
    @Override
    public void endCDATA() throws SAXException
    {
        _lexicalHandler.endCDATA();
    }

    @Override
    public void startDTD(String name, String publicId, String systemId) throws SAXException
    {
        _lexicalHandler.startDTD(name, publicId, systemId);
    }
    @Override
    public void endDTD() throws SAXException
    {
        _lexicalHandler.endCDATA();
    }

    @Override
    public void comment(char[] ch, int start, int length) throws SAXException
    {
        _lexicalHandler.comment(ch, start, length);
    }
}
