/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.transformation.htmledition;

import java.io.InputStream;
import java.util.Map;
import java.util.Optional;

import javax.jcr.Node;
import javax.jcr.RepositoryException;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

import org.ametys.cms.URIPrefixHandler;
import org.ametys.cms.data.Resource;
import org.ametys.cms.data.RichText;
import org.ametys.cms.repository.Content;
import org.ametys.core.util.ImageResolverHelper;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.model.RepositoryDataContext;
import org.ametys.runtime.model.type.DataContext;

/**
 * Handler to enhance the rich text attachments.
 * Add the content id and the data path to the local attachment's attribute
 */
public class DocbookLocalMediaObjectEditionHandler extends AbstractHTMLEditionHandler
{
    private static final String __ATTACHMENT_IMAGE_TAG_NAME = "imagedata";
    private static final String __ATTACHMENT_VIDEO_TAG_NAME = "videodata";
    private static final String __ATTACHMENT_AUDIO_TAG_NAME = "audiodata";
    private static final String __ATTACHMENT_TYPE_ATTRIBUTE_NAME = "type";
    private static final String __ATTACHMENT_TYPE_ATTRIBUTE_LOCAL_VALUE = "local";

    /** the context of the data being edited */
    protected DataContext _dataContext;
    
    private URIPrefixHandler _prefixHandler;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _prefixHandler = (URIPrefixHandler) manager.lookup(URIPrefixHandler.ROLE);
    }
    
    @Override
    public void startDocument() throws SAXException
    {
        Map objectModel = ContextHelper.getObjectModel(_context);
        Map parentContextParameters = (Map) objectModel.get(ObjectModelHelper.PARENT_CONTEXT);
        _dataContext = (DataContext) parentContextParameters.get("dataContext");
        
        super.startDocument();
    }

    @Override
    public void startElement(String uri, String loc, String raw, Attributes attrs) throws SAXException
    {
        // A new attachment starts being saxed
        boolean isAttachment = _isAttachment(loc);
        String type = attrs.getValue(__ATTACHMENT_TYPE_ATTRIBUTE_NAME);
        if (isAttachment && __ATTACHMENT_TYPE_ATTRIBUTE_LOCAL_VALUE.equals(type))
        {
            Attributes newAttrs = _processAttachment(attrs);
            super.startElement(uri, loc, raw, newAttrs);
        }
        else
        {
            super.startElement(uri, loc, raw, attrs);
        }
    }

    private boolean _isAttachment(String loc)
    {
        return __ATTACHMENT_IMAGE_TAG_NAME.equals(loc) || __ATTACHMENT_VIDEO_TAG_NAME.equals(loc) || __ATTACHMENT_AUDIO_TAG_NAME.equals(loc);
    }

    private Attributes _processAttachment(Attributes attrs) throws SAXException
    {
        RepositoryDataContext repoContext = _dataContext instanceof RepositoryDataContext rc ? rc : RepositoryDataContext.newInstance(_dataContext);
        Content content = (Content) repoContext.getObject()
                .orElseThrow(() -> new SAXException("The object id is required in the data context in order to set the fileref"));
            
        String dataPath = _dataContext.getDataPath();
        RichText richText = content.getValue(dataPath);
        
        String filename = attrs.getValue("fileref");
        Resource attachment = richText.getAttachment(filename);

        AttributesImpl newAttrs = new AttributesImpl(attrs);
        
        if (attachment != null)
        {
            _getAttachmentUUID(attachment)
                      .map(this::getRichTextByUUIDURL)
                      .ifPresentOrElse(url -> newAttrs.addCDATAAttribute("resourceByUUID", url),
                          () -> newAttrs.addCDATAAttribute("base64", _resolveAttachmentAsBase64(attrs, attachment)));
        }

        return newAttrs;
    }
    
    /**
     * Retrieves the URL to retrieve a rich text attachment by its UUID 
     * @param uuid the UUID of the rich text's attachment
     * @return the URL to retrieve a rich text attachment by its UUID
     */
    protected String getRichTextByUUIDURL(String uuid)
    {
        StringBuilder resultPath = new StringBuilder();
        
        resultPath.append(_prefixHandler.computeUriPrefix(false, false))
            .append("/plugins/cms/richText-file-by-uuid/")
            .append(uuid);
        
        return resultPath.toString();
    }
    
    private Optional<String> _getAttachmentUUID(Resource attachment)
    {
        return attachment.getRepositoryData()
                .filter(JCRRepositoryData.class::isInstance)
                .map(JCRRepositoryData.class::cast)
                .map(JCRRepositoryData::getNode)
                .map(this::_getNodeIdentifier);
    }
    
    private String _getNodeIdentifier(Node node)
    {
        try
        {
            return node.getIdentifier();
        }
        catch (RepositoryException e)
        {
            // Return null to use the fallback method (base64) if we are not able to get the node identifier 
            return null;
        }
    }
    
    private String _resolveAttachmentAsBase64(Attributes attrs, Resource attachment)
    {
        try (InputStream dataIs = attachment.getInputStream())
        {
            String heightAsString = StringUtils.substringBefore(attrs.getValue("depth"), "px");
            int height = heightAsString != null ? Integer.valueOf(heightAsString) : -1;
            
            String widthAsString = StringUtils.substringBefore(attrs.getValue("width"), "px");
            int width = widthAsString != null ? Integer.valueOf(widthAsString) : -1;
            
            return ImageResolverHelper.resolveImageAsBase64(dataIs, attachment.getMimeType(), height, width, 0, 0);
        }
        catch (Exception e)
        {
            throw new IllegalStateException(e);
        }
    }
}
