/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.transformation.htmledition;

import java.util.Map;

import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.xml.AttributesImpl;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

import org.ametys.plugins.repository.model.RepositoryDataContext;
import org.ametys.runtime.model.type.DataContext;

/**
 * Handler to enhance the rich text attachments.
 * Add the content id and the data path to the local attachment's attribute
 */
public class HTMLLocalMediaObjectEditionHandler extends AbstractHTMLEditionHandler
{
    private static final String __ATTACHMENT_IMAGE_TAG_NAME = "img";
    private static final String __ATTACHMENT_VIDEO_AUDIO_TAG_NAME = "media";
    private static final String __ATTACHMENT_TYPE_ATTRIBUTE_NAME = "data-ametys-type";
    private static final String __ATTACHMENT_TYPE_ATTRIBUTE_LOCAL_VALUE = "local";

    private DataContext _dataContext;
    
    @Override
    public void startDocument() throws SAXException
    {
        Map objectModel = ContextHelper.getObjectModel(_context);
        Map parentContextParameters = (Map) objectModel.get(ObjectModelHelper.PARENT_CONTEXT);
        _dataContext = (DataContext) parentContextParameters.get("dataContext");
        
        super.startDocument();
    }

    @Override
    public void startElement(String uri, String loc, String raw, Attributes attrs) throws SAXException
    {
        // A new attachment starts being saxed
        boolean isAttachment = _isAttachment(loc);
        String type = attrs.getValue(__ATTACHMENT_TYPE_ATTRIBUTE_NAME);
        if (isAttachment && __ATTACHMENT_TYPE_ATTRIBUTE_LOCAL_VALUE.equals(type))
        {
            Attributes newAttrs = _processAttachment(attrs);
            super.startElement(uri, loc, raw, newAttrs);
        }
        else
        {
            AttributesImpl newAttrs = new AttributesImpl(attrs);
            newAttrs.addCDATAAttribute("absolutehref", attrs.getValue("href"));
            super.startElement(uri, loc, raw, newAttrs);
        }
    }

    private boolean _isAttachment(String loc)
    {
        return __ATTACHMENT_IMAGE_TAG_NAME.equals(loc) || __ATTACHMENT_VIDEO_AUDIO_TAG_NAME.equals(loc);
    }

    private Attributes _processAttachment(Attributes attrs) throws SAXException
    {
        RepositoryDataContext repoContext = _dataContext instanceof RepositoryDataContext rc ? rc : RepositoryDataContext.newInstance(_dataContext);
        String contentId = repoContext.getObjectId()
                                      .orElseThrow(() -> new SAXException("The object id is required in the data context in order to set the href"));

        String dataPath = _dataContext.getDataPath();
        String fileName = attrs.getValue("href");

        AttributesImpl newAttrs = new AttributesImpl(attrs);
        newAttrs.addCDATAAttribute("absolutehref", contentId + "@" + dataPath + ";" + fileName);

        return newAttrs;
    }
}
