/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.transformation.xslt;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.URIResolver;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.sax.TransformerHandler;
import javax.xml.transform.stream.StreamSource;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.logger.LogEnabled;
import org.apache.avalon.framework.logger.Logger;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.components.source.SourceUtil;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.excalibur.source.Source;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.data.RichText;
import org.ametys.cms.data.RichTextHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.tag.CMSTag;
import org.ametys.cms.tag.ColorableTag;
import org.ametys.cms.tag.Tag;
import org.ametys.cms.tag.TagProviderExtensionPoint;
import org.ametys.cms.transformation.dom.TagElement;
import org.ametys.core.util.dom.AmetysNodeList;
import org.ametys.core.util.dom.EmptyElement;
import org.ametys.core.util.dom.MapElement;
import org.ametys.core.util.dom.StringElement;
import org.ametys.plugins.explorer.resources.Resource;
import org.ametys.plugins.explorer.resources.ResourceCollection;
import org.ametys.plugins.explorer.resources.dom.ResourceCollectionElement;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.metadata.CompositeMetadata;
import org.ametys.plugins.repository.metadata.CompositeMetadata.MetadataType;
import org.ametys.plugins.repository.metadata.UnknownMetadataException;
import org.ametys.plugins.repository.model.RepositoryDataContext;
import org.ametys.plugins.repository.version.VersionAwareAmetysObject;
import org.ametys.runtime.model.exception.BadDataPathCardinalityException;
import org.ametys.runtime.model.type.DataContext;

/**
 * Helper component to be used from XSL stylesheets.
 */
public class AmetysXSLTHelper extends org.ametys.core.util.AmetysXSLTHelper implements LogEnabled
{
    /** The Ametys object resolver */
    protected static AmetysObjectResolver _ametysObjectResolver;
    /** The content types extension point */
    protected static ContentTypeExtensionPoint _cTypeExtensionPoint;
    /** The tags provider */
    protected static TagProviderExtensionPoint _tagProviderExtPt;
    /** Helper for content */
    protected static ContentHelper _contentHelper;
    /** The avalon context */
    protected static Context _context;
    /** The logger */
    protected static Logger _logger;
    /** The sax parser */
    protected static RichTextHelper _richTextHelper;
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        super.contextualize(context);
        _context = context;
    }
    
    @Override
    public void enableLogging(Logger logger)
    {
        _logger = logger;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _cTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _tagProviderExtPt = (TagProviderExtensionPoint) manager.lookup(TagProviderExtensionPoint.ROLE);
        _contentHelper = (ContentHelper) manager.lookup(ContentHelper.ROLE);
        _richTextHelper = (RichTextHelper) manager.lookup(RichTextHelper.ROLE);
    }
    
    /* ------------------------ */
    /*      Content methods     */
    /* ------------------------ */
    
    /**
     * Get the content types of a content
     * @param contentId The content id
     * @return The content type or empty if the content does not exist
     */
    public static NodeList contentTypes(String contentId)
    {
        ArrayList<StringElement> contentTypes = new ArrayList<>();
        
        try
        {
            Content content = _ametysObjectResolver.resolveById(contentId);
            
            try
            {
                for (String id : content.getTypes())
                {
                    contentTypes.add(new StringElement("content-type", "id", id));
                }
            }
            catch (AmetysRepositoryException e)
            {
                _logger.error("Can not get type of content with id '" + contentId + "'", e);
            }
        }
        catch (UnknownAmetysObjectException e)
        {
            _logger.error("Can not get type of content with id '" + contentId + "'", e);
        }
        
        return new AmetysNodeList(contentTypes);
    }
    
    /**
     * Get the mixins of a content
     * @param contentId The content id
     * @return The content type or empty if the content does not exist
     */
    public static NodeList contentMixinTypes(String contentId)
    {
        ArrayList<StringElement> contentTypes = new ArrayList<>();
        
        try
        {
            Content content = _ametysObjectResolver.resolveById(contentId);
            
            try
            {
                for (String id : content.getMixinTypes())
                {
                    contentTypes.add(new StringElement("mixin", "id", id));
                }
            }
            catch (AmetysRepositoryException e)
            {
                _logger.error("Can not get type of content with id '" + contentId + "'", e);
            }
        }
        catch (UnknownAmetysObjectException e)
        {
            _logger.error("Can not get type of content with id '" + contentId + "'", e);
        }
        
        return new AmetysNodeList(contentTypes);
    }
    
    /**
     * Determines if the content of given id is a entry of reference table
     * @param contentId the content id
     * @return true if the content type is a reference table
     */
    public static boolean isReferenceTableContent(String contentId)
    {
        try
        {
            Content content = _ametysObjectResolver.resolveById(contentId);
            return _contentHelper.isReferenceTable(content);
        }
        catch (UnknownAmetysObjectException e)
        {
            _logger.error("Can not get type of unknown content with id '" + contentId + "'", e);
            return false;
        }
    }
    
    /**
     * Returns the current language for rendering.
     * @return the current language for rendering.
     */
    public static String lang()
    {
        Request request = ContextHelper.getRequest(_context);
        return (String) request.getAttribute("renderingLanguage");
    }
    
    /**
     * Determines if there is a value for the data at the given path
     * @param contentId The content id
     * @param dataPath the path of data
     * @return true if the data exists
     */
    public static boolean hasValue(String contentId, String dataPath)
    {
        try
        {
            Content content = _ametysObjectResolver.resolveById(contentId);
            return content.hasValue(dataPath);
        }
        catch (UnknownAmetysObjectException e)
        {
            if (_logger.isDebugEnabled())
            {
                _logger.debug("Can not check if attribute '" + dataPath + "' exists on unknown content with id '" + contentId + "'", e);
            }
            return false;
        }
    }
    
    /**
     * Get the attribute of a content at the given path
     * @param contentId The content id
     * @param dataPath The data path
     * @return The value into a "value" node or null if an error occurred
     */
    public static NodeList contentAttribute(String contentId, String dataPath)
    {
        return contentAttribute(contentId, dataPath, null);
    }
    
    /**
     * Get the attribute of a content at the given path
     * @param contentId The content id
     * @param dataPath The data path
     * @param lang The language for localized attribute. Can be null for non-localized attribute or to get the values for all existing locales. 
     * @return The value into a "value" node or null if an error occurred
     */
    public static NodeList contentAttribute(String contentId, String dataPath, String lang)
    {
        try
        {
            Content content = _ametysObjectResolver.resolveById(contentId);
            DataContext context = RepositoryDataContext.newInstance()
                                             .withObject(content);
            
            List<Node> values = _getNodeValues(content.getDataHolder(), dataPath, lang, context);
            if (values != null)
            {
                return new AmetysNodeList(values);
            }
        }
        catch (UnknownAmetysObjectException e)
        {
            _logger.error("Can not get attribute at path '" + dataPath + "' on unknown content with id '" + contentId + "'", e);
        }
        
        return null;
    }

    /**
     * Get values of an attribute of a model aware data holder at the given path
     * @param dataHolder the data holder
     * @param dataPath The data path
     * @param lang The language for localized attribute. Can be null for non-localized attribute or to get the values for all existing locales. 
     * @return A Node for each values or null if an error occurred
     */
    protected static List<Node> _getNodeValues(ModelAwareDataHolder dataHolder, String dataPath, String lang)
    {
        return _getNodeValues(dataHolder, dataPath, lang, DataContext.newInstance());
    }
    
    /**
     * Get values of an attribute of a model aware data holder at the given path
     * @param dataHolder the data holder
     * @param dataPath The data path
     * @param lang The language for localized attribute. Can be null for non-localized attribute or to get the values for all existing locales.
     * @param dataContext The data context 
     * @return A Node for each values or null if an error occurred
     */
    protected static List<Node> _getNodeValues(ModelAwareDataHolder dataHolder, String dataPath, String lang, DataContext dataContext)
    {
        
        if (dataHolder == null)
        {
            return null;
        }
        
        try
        {
            SAXTransformerFactory saxTransformerFactory = (SAXTransformerFactory) TransformerFactory.newInstance();
            TransformerHandler th = saxTransformerFactory.newTransformerHandler();
            
            DOMResult result = new DOMResult();
            th.setResult(result);
            
            th.startDocument();
            XMLUtils.startElement(th, "value");
            
            Locale locale = StringUtils.isEmpty(lang) ? null : new Locale(lang);
            dataHolder.dataToSAX(th, dataPath, dataContext.cloneContext().withLocale(locale).withEmptyValues(false));
            
            XMLUtils.endElement(th, "value");
            th.endDocument();
            
            List<Node> values = new ArrayList<>();
            
            // #getChildNodes() returns a NodeList that contains the value(s) saxed 
            // we cannot returns directly this NodeList because saxed values should be wrapped into a <value> tag.
            NodeList childNodes = result.getNode().getFirstChild().getChildNodes(); 
            for (int i = 0; i < childNodes.getLength(); i++)
            {
                Node n = childNodes.item(i);
                values.add(n);
            }
            
            return values;
        }
        catch (BadDataPathCardinalityException e)
        {
            _logger.error("Unable to get attribute at path '" + dataPath + "'. Path is invalid.", e);
        }
        catch (TransformerConfigurationException | SAXException e)
        {
            _logger.error("Fail to sax attribute at path '" + dataPath + "'", e);
        }
        catch (Exception e) 
        {
            _logger.error("An error occurred, impossible to get attribute at path '" + dataPath + "'", e);
        }
        
        return null;
    }
    
    /**
     * Get the metadata of a content
     * @param contentId The content id
     * @param metadataName The metadata name (/ for composite)
     * @param lang The language for localized metadata. Can be null to get the current language.
     * @return The name or empty if the metadata or the content does not exist
     */
    public static String contentMetadata(String contentId, String metadataName, String lang)
    {
        try
        {
            Content content = _ametysObjectResolver.resolveById(contentId);
            try
            {
                Locale locale = StringUtils.isEmpty(lang) ? null : new Locale(lang);
                return _getMetadata(content.getMetadataHolder(), metadataName, locale);
            }
            catch (UnknownMetadataException e)
            {
                _logger.debug("Can not get metadata '" + metadataName + "' on content with id '" + contentId + "'", e);
                return "";
            }
        }
        catch (UnknownAmetysObjectException e)
        {
            _logger.debug("Can not get metadata '" + metadataName + "' on unknown content with id '" + contentId + "'", e);
            return "";
        }
    }
    
    /**
     * Get the metadata of a content
     * @param contentId The content id
     * @param metadataName The metadata name (/ for composite)
     * @return The name or empty if the metadata or the content does not exist
     */
    public static String contentMetadata(String contentId, String metadataName)
    {
        return contentMetadata(contentId, metadataName, null);
    }
    
    private static String _getMetadata(CompositeMetadata cm, String metadataName, Locale locale)
    {
        int i = metadataName.indexOf("/");
        if (i == -1)
        {
            if (cm.getType(metadataName).equals(MetadataType.MULTILINGUAL_STRING))
            {
                if (locale == null)
                {
                    String currentLanguage = lang();
                    if (StringUtils.isEmpty(currentLanguage))
                    {
                        _logger.error("Can not get the value of a multilingual metadata " + metadataName + " without a defined locale");
                        return "";
                    }
                    return cm.getLocalizedString(metadataName, new Locale(currentLanguage));
                }
                else
                {
                    return cm.getLocalizedString(metadataName, locale);
                }
            }
            else
            {
                return cm.getString(metadataName);
            }
        }
        else
        {
            return _getMetadata(cm.getCompositeMetadata(metadataName.substring(0, i)), metadataName.substring(i + 1), locale);
        }
    }
    
    /**
     * Returns a DOM {@link Element} containing {@link Resource}s representing the attachments of the current content.
     * @return an Element containing the attachments of the current content as {@link Resource}s.
     */
    public static Node contentAttachments()
    {
        Request request = ContextHelper.getRequest(_context);
        
        Content content = (Content) request.getAttribute(Content.class.getName());
        
        return contentAttachments(content);
    }
    
    /**
     * Returns a DOM {@link Element} containing {@link Resource}s representing the attachments of a given content.
     * @param contentId the content ID.
     * @return an Element containing the attachments of the given content as {@link Resource}s.
     */
    public static Node contentAttachments(String contentId)
    {
        Content content = _ametysObjectResolver.resolveById(contentId);
        
        return contentAttachments(content);
    }
    
    /**
     * Returns a DOM {@link Element} containing {@link Resource}s representing the attachments of a given content.
     * @param content the content.
     * @return an Element containing the attachments of the given content as {@link Resource}s.
     */
    private static Node contentAttachments(Content content)
    {
        if (content == null)
        {
            return null;
        }
        
        ResourceCollection collection = content.getRootAttachments();
        
        return collection != null ? new ResourceCollectionElement(collection) : new EmptyElement("collection");
    }
    
    /**
     * Set the content of given id in request attribute
     * @param contentId the id of content
     */
    public static void setCurrentContent(String contentId)
    {
        setCurrentContent(contentId, null);
    }
    
    /**
     * Set the content of given id and version in request attribute
     * @param contentId the id of content
     * @param versionLabel The version label
     */
    public static void setCurrentContent(String contentId, String versionLabel)
    {
        Request request = ContextHelper.getRequest(_context);
        
        Content content = _ametysObjectResolver.resolveById(contentId);
        
        if (StringUtils.isNotEmpty(versionLabel) && content instanceof VersionAwareAmetysObject)
        {
            String[] allLabels = ((VersionAwareAmetysObject) content).getAllLabels();
            if (ArrayUtils.contains(allLabels, versionLabel))
            {
                ((VersionAwareAmetysObject) content).switchToLabel(versionLabel);
            }
        }
        
        request.setAttribute(Content.class.getName(), content);
        
    }
    
  //*************************
    // Tag methods
    //*************************
    
    /**
     * Returns all tags of the current content.
     * @return a list of tags.
     */
    public static NodeList contentTags()
    {
        Request request = ContextHelper.getRequest(_context);
        
        Content content = (Content) request.getAttribute(Content.class.getName());
        return _contentTags(content);
    }
    
    /**
     * Returns all tags of the given content
     * @param contentId The identifier of the content
     * @return a list of tags.
     */
    public static NodeList contentTags(String contentId)
    {
        try
        {
            Content content = _ametysObjectResolver.resolveById(contentId);
            return _contentTags(content);
        }
        catch (AmetysRepositoryException e)
        {
            _logger.warn("Cannot get tags for content '" + contentId + "'", e);
        }
        
        return null;
    }
    
    /**
     * Returns all tags of the given content
     * @param content The content
     * @return a list of tags.
     */
    protected static NodeList _contentTags(Content content)
    {
        if (content == null)
        {
            return null;
        }
        
        List<TagElement> list = new ArrayList<>();
        
        for (String tag : content.getTags())
        {
            list.add(new TagElement(tag));
        }
        
        return new AmetysNodeList(list);
    }
    
    /**
     * Get the name of the parent of a tag.
     * @param siteName the site name
     * @param tagName the tag's name
     * @return The id of parent or empty if not found
     */
    public static String tagParent(String siteName, String tagName)
    {
        Map<String, Object> contextParameters = new HashMap<>();
        contextParameters.put("siteName", siteName);
        
        Tag tag = _tagProviderExtPt.getTag(tagName, contextParameters);
        if (tag == null)
        {
            return StringUtils.EMPTY;
        }
        
        String parentName = tag.getParentName();
        return parentName != null ? parentName : StringUtils.EMPTY;
    }
    
    /**
     * Get the path of a tag. The path contains the tag's parents seprated by '/'.
     * @param siteName The site name
     * @param tagName The unique tag's name
     * @return The tag's path or empty string if tag does not exist
     */
    public static String tagPath (String siteName, String tagName)
    {
        Map<String, Object> contextParameters = new HashMap<>();
        contextParameters.put("siteName", siteName);
        
        Tag tag = _tagProviderExtPt.getTag(tagName, contextParameters);
        if (tag == null)
        {
            return StringUtils.EMPTY;
        }
        
        String path = tagName;
        
        Tag parentTag = tag.getParent();
        while (parentTag != null)
        {
            path = parentTag.getName() + "/" + path;
            parentTag = parentTag.getParent();
        }

        return path;
    }
    
    /**
     * Get the label of a tag
     * @param siteName the current site
     * @param tagName the name of the tag
     * @param lang the lang (if i18n tag)
     * @return the label of the tag or empty if it cannot be found
     */
    public static String tagLabel(String siteName, String tagName, String lang)
    {
        Map<String, Object> contextParameters = new HashMap<>();
        contextParameters.put("siteName", siteName);
        
        Tag tag = _tagProviderExtPt.getTag(tagName, contextParameters);
        return tag == null ? "" : _i18nUtils.translate(tag.getTitle(), lang);
    }
    
    /**
     * Get the description of a tag
     * @param siteName the current site
     * @param tagName the name of the tag
     * @param lang the lang (if i18n tag)
     * @return the label of the tag or empty if it cannot be found
     */
    public static String tagDescription(String siteName, String tagName, String lang)
    {
        Map<String, Object> contextParameters = new HashMap<>();
        contextParameters.put("siteName", siteName);
        
        Tag tag = _tagProviderExtPt.getTag(tagName, contextParameters);
        return tag == null ? "" : _i18nUtils.translate(tag.getDescription(), lang);
    }
    
    /**
     * Get the visibility of a tag
     * @param siteName the current site
     * @param tagName the name of the tag
     * @return the lower-cased visibility of the tag ("public" or "private")
     */
    public static String tagVisibility(String siteName, String tagName)
    {
        Map<String, Object> contextParameters = new HashMap<>();
        contextParameters.put("siteName", siteName);
        
        CMSTag tag = _tagProviderExtPt.getTag(tagName, contextParameters);
        return tag == null ? "" : tag.getVisibility().toString().toLowerCase();
    }
    
    /**
     * Get the color (main and text) of a tag
     * @param siteName the current site
     * @param tagName the name of the tag
     * @return the the color (main and text) of a tag
     */
    public static MapElement tagColor(String siteName, String tagName)
    {
        Map<String, Object> contextParameters = new HashMap<>();
        contextParameters.put("siteName", siteName);
        
        Tag tag = _tagProviderExtPt.getTag(tagName, contextParameters);
        if (tag instanceof ColorableTag colorTag)
        {
            String color = colorTag.getColor(true);
            Map<String, String> map = colorTag.getColorComponent().getColors().get(color);
            return map != null ? new MapElement("color", map) : null;
        }
        
        return null;
    }
    
    /* ----------------------------- */
    /*      Content type methods     */
    /* ----------------------------- */
    
    /**
     * Returns all tags of a content type
     * @param contentTypeId The id of the content type
     * @return a list of tags.
     */
    public static NodeList contentTypeTags(String contentTypeId)
    {
        ArrayList<TagElement> tags = new ArrayList<>();
        
        try
        {
            ContentType cType = _cTypeExtensionPoint.getExtension(contentTypeId);
            if (cType != null)
            {
                for (String tag : cType.getTags())
                {
                    tags.add(new TagElement(tag));
                }
            }
            else
            {
                _logger.error("Can not get tags of unknown content type of id '" + contentTypeId + "'");
            }
            
        }
        catch (AmetysRepositoryException e)
        {
            _logger.error("Can not get tags of content type of id '" + contentTypeId + "'", e);
        }
        
        return new AmetysNodeList(tags);
    }
    
    /**
     * Get the excerpt of content from the given richtext attribute
     * @param contentId the id of content
     * @param attributePath the attribute path of rich text attribute
     * @param limit the max length for content excerpt 
     * @return the excerpt
     */
    public static String contentExcerpt(String contentId, String attributePath, int limit)
    {
        Content content = _ametysObjectResolver.resolveById(contentId);
        
        if (content.hasValue(attributePath))
        {
            RichText richText = content.getValue(attributePath);
            return _richTextHelper.richTextToString(richText, limit);
        }
     
        return org.apache.commons.lang3.StringUtils.EMPTY;
    }
    
    /**
     * Get the HTML view of a content
     * @param contentId the id of content
     * @return the content html view wrapped into a &lt;content&gt; tag
     */
    public static Node getContentView(String contentId)
    {
        return getContentView(contentId, null, 1, null, false);
    }
    
    /**
     * Get the HTML view of a content with offset on headings
     * @param contentId the id of content
     * @param startHeadingsLevel The start level for headings (h1, h2, h3, ...). For example, set to 2 so that the highest level headings are &lt;h2&gt;. Set to 1 for no offset.
     * @return the content html view wrapped into a &lt;content&gt; tag
     */
    public static Node getContentView(String contentId, int startHeadingsLevel)
    {
        return getContentView(contentId, null, startHeadingsLevel, null, false);
    }
    
    /**
     * Get the HTML view of a content
     * @param contentId the id of content
     * @param viewName The content view name
     * @return the content html view wrapped into a &lt;content&gt; tag
     */
    public static Node getContentView(String contentId, String viewName)
    {
        return getContentView(contentId, viewName, 1, null, false);
    }
    
    /**
     * Get the HTML view of a content
     * @param contentId the id of content
     * @param viewName The content view name
     * @param startHeadingsLevel The start level for headings (h1, h2, h3, ...). For example, set to 2 so that the highest level headings are &lt;h2&gt;. Set to 1 for no offset.
     * @return the content html view wrapped into a &lt;content&gt; tag
     */
    public static Node getContentView(String contentId, String viewName, int startHeadingsLevel)
    {
        return getContentView(contentId, viewName, startHeadingsLevel, null, false);
    }
    
    /**
     * Get the HTML view of a content with offset on headings
     * @param contentId the id of content
     * @param viewName The content view name
     * @param lang the language. Can be null. Useful only if the content has a multilingual title.
     * @param startHeadingsLevel The start level for headings (h1, h2, h3, ...). For example, set to 2 so that the highest level headings are &lt;h2&gt;. Set to 1 for no offset.
     * @param checkReadAccess Set to <code>true</code> to check the read access on content. Be careful, do not use with <code>true</code> on a cacheable element.
     * @return the content html view wrapped into a &lt;content&gt; tag
     */
    public static Node getContentView(String contentId, String viewName, int startHeadingsLevel, String lang, boolean checkReadAccess)
    {
        Content content = _ametysObjectResolver.resolveById(contentId);
        
        if (checkReadAccess && !_rightManager.currentUserHasReadAccess(content))
        {
            _logger.warn("Current user is not authorized to see content of id '" + contentId + "'. AmetysXSLHelper#getContentView will return null element");
            return null;
        }
        
        Locale requestedLocale = StringUtils.isNotEmpty(lang) ? new Locale(lang) : null;
        
        DocumentBuilder builder = null;
        Source source = null;
        
        String uri = _contentHelper.getContentHtmlViewUrl(content, viewName);
        
        try
        {
            source = _sourceResolver.resolveURI(uri);
            
            Source src = null;
            try
            {
                // Wrap HTML view into a <content> tag and move titles hierarchy if needed
                src = _sourceResolver.resolveURI("plugin:cms://stylesheets/content/content2htmlview.xsl");
                try (InputStream is = src.getInputStream())
                {
                    SAXTransformerFactory tFactory = (SAXTransformerFactory) TransformerFactory.newInstance();
                    
                    // Set uri resolver to resolve import
                    tFactory.setURIResolver(new URIResolver()
                    {
                        public javax.xml.transform.Source resolve(String href, String base) throws TransformerException
                        {
                            try
                            {
                                Source resolvedSource = _sourceResolver.resolveURI(href);
                                return new StreamSource(resolvedSource.getInputStream());
                            }
                            catch (IOException e)
                            {
                                throw new TransformerException(e);
                            }
                        }
                    });
                    
                    TransformerHandler transformerHandler = tFactory.newTransformerHandler(new StreamSource(is));
                    Transformer transformer = transformerHandler.getTransformer();
                    
                    Properties format = new Properties();
                    format.put(OutputKeys.METHOD, "xml");
                    format.put(OutputKeys.ENCODING, "UTF-8");
                    
                    transformer.setOutputProperties(format);
                    
                    transformer.setParameter("contentId", content.getId());
                    transformer.setParameter("contentName", content.getName());
                    transformer.setParameter("contentTitle", content.getTitle(requestedLocale));
                    if (content.getLanguage() != null)
                    {
                        transformer.setParameter("contentLanguage", content.getLanguage());
                    }
                    transformer.setParameter("headingLevel", startHeadingsLevel);
                    
                    builder  = DocumentBuilderFactory.newInstance().newDocumentBuilder();
                    Document document = builder.newDocument();
                    DOMResult result = new DOMResult(document);

                    transformerHandler.setResult(result);
                    SourceUtil.toSAX(source, transformerHandler);
                    
                    return result.getNode();
                }
            }
            finally
            {
                _sourceResolver.release(src);
            }
        }
        catch (Exception e)
        {
            _logger.error("Fail to get HTML view of content " + contentId, e);
            return null;
        }
        finally
        {
            _sourceResolver.release(source);
        }
    }
}
