/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.transformation.xslt;

import java.util.Optional;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.transformation.URIResolver;
import org.ametys.cms.transformation.URIResolverExtensionPoint;

/**
 * This component resolve links and give a static hack access for xslt calls
 */
public class ResolveURIComponent extends AbstractLogEnabled implements Serviceable, Initializable, Contextualizable
{
    private static ResolveURIComponent _instance;
    
    private static Context _context;
    
    private URIResolverExtensionPoint _linkResolverExtensionPoint;
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _linkResolverExtensionPoint = (URIResolverExtensionPoint) manager.lookup(URIResolverExtensionPoint.ROLE);
    }
    
    @Override
    public void initialize() throws Exception
    {
        _instance = this;
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @return The uri resolved, or the uri if there is no resolver adapted
     */
    public static String resolve(String type, String uri)
    {
        return resolve(type, uri, false);
    }

    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param download Is this uri for download purposes.
     * @return The uri resolved, or the uri if there is no resolver adapted
     */
    public static String resolve(String type, String uri, boolean download)
    {
        // FIXME CMS-2611 Force absolute 
        Request request = ContextHelper.getRequest(_context);
        boolean absolute = request.getAttribute("forceAbsoluteUrl") != null ? (Boolean) request.getAttribute("forceAbsoluteUrl") : false;
        
        return resolve(type, uri, download, absolute);
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param download Is this uri for download purposes.
     * @param absolute true to generate absolute url
     * @return The uri resolved, the empty string if the uri could not be resolved, or the uri itself if there is no resolver adapted
     */
    public static String resolve(String type, String uri, boolean download, boolean absolute)
    {
        return resolve(type, uri, download, absolute, false);
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param download Is this uri for download purposes.
     * @param absolute true to generate absolute url
     * @param internal true to get an internal URI.
     * @return The uri resolved, the empty string if the uri could not be resolved, or the uri itself if there is no resolver adapted
     */
    public static String resolve(String type, String uri, boolean download, boolean absolute, boolean internal)
    {
        Request request = ContextHelper.getRequest(_context);
        Object forceRemoteUrl = request.getAttribute("forceRemoteUrl");
        
        boolean remote = forceRemoteUrl != null
                && (forceRemoteUrl instanceof Boolean && (Boolean) forceRemoteUrl 
                        || Boolean.valueOf((String) forceRemoteUrl));
        
        String proxiedUri = remote ? type + ";" + uri : uri;
        URIResolver uriResolver = _instance._linkResolverExtensionPoint.getResolverForType(remote ? "remote" : type);
        if (uriResolver != null)
        {
            try
            {
                request.removeAttribute("forceRemoteUrl");
                return uriResolver.resolve(proxiedUri, download, absolute, internal);
            }
            catch (Exception e)
            {
                _instance.getLogger().warn("Error resolving the uri '" + uri + "' with type " + type, e);
                return "";
            }
            finally
            {
                request.setAttribute("forceRemoteUrl", forceRemoteUrl);
            }
        }
        else
        {
            return uri;
        }
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param height the height
     * @param width the width
     * @return The uri resolved, or the uri if there is no resolver adapted
     */
    public static String resolveImage(String type, String uri, int height, int width)
    {
        return resolveImage (type, uri, height, width, false);
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param height the height
     * @param width the width
     * @param download Is this uri for download purposes.
     * @return The uri resolved, or the uri if there is no resolver adapted
     */
    public static String resolveImage(String type, String uri, int height, int width, boolean download)
    {
        // FIXME CMS-2611 Force absolute 
        Request request = ContextHelper.getRequest(_context);
        boolean absolute = request.getAttribute("forceAbsoluteUrl") != null ? (Boolean) request.getAttribute("forceAbsoluteUrl") : false;
        
        return resolveImage(type, uri, height, width, download, absolute);
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param height the height
     * @param width the width
     * @param download Is this uri for download purposes.
     * @param absolute true to generate absolute url
     * @return The uri resolved, the empty string if the uri could not be resolved, or the uri itself if there is no resolver adapted
     */
    public static String resolveImage(String type, String uri, int height, int width, boolean download, boolean absolute)
    {
        return resolveImage(type, uri, height, width, download, absolute, false);
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param height the height
     * @param width the width
     * @param download Is this uri for download purposes.
     * @param absolute true to generate absolute url
     * @param internal true to get an internal URI.
     * @return The uri resolved, the empty string if the uri could not be resolved, or the uri itself if there is no resolver adapted
     */
    public static String resolveImage(String type, String uri, int height, int width, boolean download, boolean absolute, boolean internal)
    {
        // FIXME CMS-4059 Force base64 encoding. 
        Request request = ContextHelper.getRequest(_context);
        boolean encodeBase64 = request.getAttribute("forceBase64Encoding") != null ? (Boolean) request.getAttribute("forceBase64Encoding") : false;
        
        Object forceRemoteUrl = request.getAttribute("forceRemoteUrl");
        boolean remote = forceRemoteUrl != null
                        && (forceRemoteUrl instanceof Boolean && (Boolean) forceRemoteUrl
                                || Boolean.valueOf((String) forceRemoteUrl));
        
        String proxiedUri = remote ? type + ";" + uri : uri;
        URIResolver uriResolver = _instance._linkResolverExtensionPoint.getResolverForType(remote ? "remote" : type);
        if (uriResolver != null)
        {
            try
            {
                request.removeAttribute("forceRemoteUrl");
                if (encodeBase64)
                {
                    return uriResolver.resolveImageAsBase64(proxiedUri, height, width);
                }
                else
                {
                    return uriResolver.resolveImage(proxiedUri, height, width, download, absolute, internal);
                }
            }
            catch (Exception e)
            {
                _instance.getLogger().warn("Error resolving the image of uri '" + uri + "' with type " + type, e);
                return "";
            }
            finally
            {
                request.setAttribute("forceRemoteUrl", forceRemoteUrl);
            }
        }
        else
        {
            return uri;
        }
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint return it as a base64-encoded string.
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param height the height
     * @param width the width
     * @return a base64-encoded string representing the image or empty string if the uri could not be resolved, or the uri itself if there is no resolver adapted
     */
    public static String resolveImageAsBase64(String type, String uri, int height, int width)
    {
        URIResolver uriResolver = _instance._linkResolverExtensionPoint.getResolverForType(type);
        if (uriResolver != null)
        {
            try
            {
                return uriResolver.resolveImageAsBase64(uri, height, width);
            }
            catch (Exception e)
            {
                _instance.getLogger().warn("Error resolving the image as base64 of uri '" + uri + "' with type " + type, e);
                return "";
            }
        }
        else
        {
            return uri;
        }
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param maxHeight the max height
     * @param maxWidth the max width
     * @return The uri resolved, or the uri if there is no resolver adapted
     */
    public static String resolveBoundedImage(String type, String uri, int maxHeight, int maxWidth)
    {
        return resolveBoundedImage (type, uri, maxHeight, maxWidth, false);
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param maxHeight the max height
     * @param maxWidth the max width
     * @param download Is this uri for download purposes.
     * @return The uri resolved, or the uri if there is no resolver adapted
     */
    public static String resolveBoundedImage(String type, String uri, int maxHeight, int maxWidth, boolean download)
    {
        // FIXME CMS-2611 Force absolute 
        Request request = ContextHelper.getRequest(_context);
        boolean absolute = request.getAttribute("forceAbsoluteUrl") != null ? (Boolean) request.getAttribute("forceAbsoluteUrl") : false;
        
        return resolveBoundedImage(type, uri, maxHeight, maxWidth, download, absolute);
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param maxHeight the max height
     * @param maxWidth the max width
     * @param download Is this uri for download purposes.
     * @param absolute true to generate absolute url
     * @return The uri resolved, the empty string if the uri could not be resolved, or the uri itself if there is no resolver adapted
     */
    public static String resolveBoundedImage(String type, String uri, int maxHeight, int maxWidth, boolean download, boolean absolute)
    {
        return resolveBoundedImage(type, uri, maxHeight, maxWidth, download, absolute, false);
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param maxHeight the max height
     * @param maxWidth the max width
     * @param download Is this uri for download purposes.
     * @param absolute true to generate absolute url
     * @param internal true to get an internal URI.
     * @return The uri resolved, the empty string if the uri could not be resolved, or the uri itself if there is no resolver adapted
     */
    public static String resolveBoundedImage(String type, String uri, int maxHeight, int maxWidth, boolean download, boolean absolute, boolean internal)
    {
        // FIXME CMS-4059 Force base64 encoding. 
        Request request = ContextHelper.getRequest(_context);
        boolean encodeBase64 = request.getAttribute("forceBase64Encoding") != null ? (Boolean) request.getAttribute("forceBase64Encoding") : false;
        
        Object forceRemoteUrl = request.getAttribute("forceRemoteUrl");
        boolean remote = forceRemoteUrl != null
                        && (forceRemoteUrl instanceof Boolean && (Boolean) forceRemoteUrl
                                || Boolean.valueOf((String) forceRemoteUrl));
        
        String proxiedUri = remote ? type + ";" + uri : uri;
        URIResolver uriResolver = _instance._linkResolverExtensionPoint.getResolverForType(remote ? "remote" : type);
        if (uriResolver != null)
        {
            try
            {
                request.removeAttribute("forceRemoteUrl");
                if (encodeBase64)
                {
                    return uriResolver.resolveBoundedImageAsBase64(proxiedUri, maxHeight, maxWidth);
                }
                else
                {
                    return uriResolver.resolveBoundedImage(proxiedUri, maxHeight, maxWidth, download, absolute, internal);
                }
            }
            catch (Exception e)
            {
                _instance.getLogger().warn("Error resolving the image of uri '" + uri + "' with type " + type, e);
                return "";
            }
            finally
            {
                request.setAttribute("forceRemoteUrl", forceRemoteUrl);
            }
        }
        else
        {
            return uri;
        }
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint return it as a base64-encoded string.
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param maxHeight the max height
     * @param maxWidth the max width
     * @return a base64-encoded string representing the image or empty string if the uri could not be resolved, or the uri itself if there is no resolver adapted
     */
    public static String resolveBoundedImageAsBase64(String type, String uri, int maxHeight, int maxWidth)
    {
        URIResolver uriResolver = _instance._linkResolverExtensionPoint.getResolverForType(type);
        if (uriResolver != null)
        {
            try
            {
                return uriResolver.resolveBoundedImageAsBase64(uri, maxHeight, maxWidth);
            }
            catch (Exception e)
            {
                _instance.getLogger().warn("Error resolving the bounded image as base64 of uri '" + uri + "' with type " + type, e);
                return "";
            }
        }
        else
        {
            return uri;
        }
    }

    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param cropHeight the crop height
     * @param cropWidth the crop width
     * @return The uri resolved, or the uri if there is no resolver adapted
     */
    public static String resolveCroppedImage(String type, String uri, int cropHeight, int cropWidth)
    {
        return resolveCroppedImage (type, uri, cropHeight, cropWidth, false);
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param cropHeight the crop height
     * @param cropWidth the crop width
     * @param download Is this uri for download purposes.
     * @return The uri resolved, or the uri if there is no resolver adapted
     */
    public static String resolveCroppedImage(String type, String uri, int cropHeight, int cropWidth, boolean download)
    {
        // FIXME CMS-2611 Force absolute 
        Request request = ContextHelper.getRequest(_context);
        boolean absolute = request.getAttribute("forceAbsoluteUrl") != null ? (Boolean) request.getAttribute("forceAbsoluteUrl") : false;
        
        return resolveCroppedImage(type, uri, cropHeight, cropWidth, download, absolute);
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param cropHeight the crop height
     * @param cropWidth the crop width
     * @param download Is this uri for download purposes.
     * @param absolute true to generate absolute url
     * @return The uri resolved, the empty string if the uri could not be resolved, or the uri itself if there is no resolver adapted
     */
    public static String resolveCroppedImage(String type, String uri, int cropHeight, int cropWidth, boolean download, boolean absolute)
    {
        return resolveCroppedImage(type, uri, cropHeight, cropWidth, download, absolute, false);
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param cropHeight the crop height
     * @param cropWidth the crop width
     * @param download Is this uri for download purposes.
     * @param absolute true to generate absolute url
     * @param internal true to get an internal URI.
     * @return The uri resolved, the empty string if the uri could not be resolved, or the uri itself if there is no resolver adapted
     */
    public static String resolveCroppedImage(String type, String uri, int cropHeight, int cropWidth, boolean download, boolean absolute, boolean internal)
    {
        // FIXME CMS-4059 Force base64 encoding. 
        Request request = ContextHelper.getRequest(_context);
        boolean encodeBase64 = request.getAttribute("forceBase64Encoding") != null ? (Boolean) request.getAttribute("forceBase64Encoding") : false;
        
        URIResolver uriResolver = _instance._linkResolverExtensionPoint.getResolverForType(type);
        if (uriResolver != null)
        {
            try
            {
                if (encodeBase64)
                {
                    return uriResolver.resolveCroppedImageAsBase64(uri, cropHeight, cropWidth);
                }
                else
                {
                    return uriResolver.resolveCroppedImage(uri, cropHeight, cropWidth, download, absolute, internal);
                }
            }
            catch (Exception e)
            {
                _instance.getLogger().warn("Error resolving the image of uri '" + uri + "' with type " + type, e);
                return "";
            }
        }
        else
        {
            return uri;
        }
    }
    
    /**
     * Resolve an uri upon the LinkResolverExtensionPoint return it as a base64-encoded string.
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @param cropHeight the crop height
     * @param cropWidth the crop width
     * @return a base64-encoded string representing the image or empty string if the uri could not be resolved, or the uri itself if there is no resolver adapted
     */
    public static String resolveCroppedImageAsBase64(String type, String uri, int cropHeight, int cropWidth)
    {
        URIResolver uriResolver = _instance._linkResolverExtensionPoint.getResolverForType(type);
        if (uriResolver != null)
        {
            try
            {
                return uriResolver.resolveCroppedImageAsBase64(uri, cropHeight, cropWidth);
            }
            catch (Exception e)
            {
                _instance.getLogger().warn("Error resolving the cropped image as base64 of uri '" + uri + "' with type " + type, e);
                return "";
            }
        }
        else
        {
            return uri;
        }
    }
    
    /**
     * Get the mime type.
     * @param type Type name (defined by the extension to use)
     * @param uri URI depending on the type
     * @return the mime type or application/octet-stream if not found
     */
    public static String getMimeType(String type, String uri)
    {
        URIResolver uriResolver = _instance._linkResolverExtensionPoint.getResolverForType(type);
        return Optional.ofNullable(uriResolver)
                .map(r -> _getMimeTypeSafe(r, uri))
                .filter(StringUtils::isNotEmpty)
                .orElse("application/octet-stream");
    }
    
    private static String _getMimeTypeSafe(URIResolver uriResolver, String uri)
    {
        try
        {
            return uriResolver.getMimeType(uri);
        }
        catch (Exception e)
        {
            _instance.getLogger().warn("Error resolving the image to get the mime type of uri '" + uri + "' with type " + uriResolver.getType(), e);
            return null;
        }
    }
}
