/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.HashMap;
import java.util.Map;

import javax.jcr.Node;
import javax.jcr.RepositoryException;
import javax.jcr.lock.Lock;
import javax.jcr.lock.LockManager;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.commons.lang.StringUtils;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.lock.LockAwareAmetysObject;
import org.ametys.plugins.repository.version.VersionableAmetysObject;
import org.ametys.plugins.workflow.EnhancedFunction;
import org.ametys.plugins.workflow.support.WorkflowProvider;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;

import com.opensymphony.workflow.WorkflowException;
import com.opensymphony.workflow.spi.Step;

/**
 * Abstract workflow function working on a Content, which provides common content methods.
 */
public abstract class AbstractContentFunction extends AbstractContentWorkflowComponent implements EnhancedFunction, Initializable
{
    /** The key in transiant var for the event to notify after the function */
    public static final String EVENT_TO_NOTIFY_KEY = "event-to-notify";
    
    /** The workflow provider. */
    protected WorkflowProvider _workflowProvider;
    /** The observation manager */
    protected ObservationManager _observationManager;

    @Override
    public void initialize() throws Exception
    {
        _workflowProvider = (WorkflowProvider) _manager.lookup(WorkflowProvider.ROLE);
        _observationManager = (ObservationManager) _manager.lookup(ObservationManager.ROLE);
    }
    
    /**
     * Add the content lock token to the current session if the content is locked.
     * @param content the content.
     * @throws WorkflowException if an error occurs.
     * @throws RepositoryException if a repository error occurs.
     */
    protected void _addLockToken(WorkflowAwareContent content) throws WorkflowException, RepositoryException
    {
        if (content != null && content instanceof LockAwareAmetysObject)
        {
            LockAwareAmetysObject lockAwareContent = (LockAwareAmetysObject) content;
            if (lockAwareContent.isLocked())
            {
                Node node = content.getNode();
                
                LockManager lockManager = node.getSession().getWorkspace().getLockManager();
                
                Lock lock = lockManager.getLock(node.getPath());
                Node lockHolder = lock.getNode();
                
                lockManager.addLockToken(lockHolder.getProperty(RepositoryConstants.METADATA_LOCKTOKEN).getString());
            }
        }
    }
    
    /**
     * Set the content's current step ID without notifying the observers that the content's workflow has changed.
     * @param content the content.
     * @param transientVars The workflow transient vars.
     * @throws WorkflowException if an error occurs.
     */
    protected void _setCurrentStepId(WorkflowAwareContent content, Map transientVars) throws WorkflowException
    {
        if (content != null)
        {
            // Save the current step.
            AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(content);
            Step currentStep = (Step) workflow.getCurrentSteps(content.getWorkflowId()).iterator().next();
            content.setCurrentStepId(currentStep.getStepId());
            content.saveChanges();
        }
    }
    
    /**
     * Set the content's current step ID and notify the observers that the content's workflow has changed.
     * @param content the content.
     * @param transientVars The workflow transient vars.
     * @throws WorkflowException if an error occurs.
     */
    protected void _setCurrentStepIdAndNotify(WorkflowAwareContent content, Map transientVars) throws WorkflowException
    {
        if (content != null)
        {
            // Save the current step.
            AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(content);
            Step currentStep = (Step) workflow.getCurrentSteps(content.getWorkflowId()).iterator().next();
            content.setCurrentStepId(currentStep.getStepId());
            content.saveChanges();
            
            _notify(content, transientVars);
        }
    }
    
    /**
     * Notify the event
     * @param content the content
     * @param transientVars the workflow transient vars
     * @throws WorkflowException if an error occurs
     */
    @SuppressWarnings("unchecked")
    protected void _notify(WorkflowAwareContent content, Map transientVars) throws WorkflowException
    {
        String eventId = (String) transientVars.getOrDefault(EVENT_TO_NOTIFY_KEY, _getDefaultEvent());
        if (StringUtils.isNotBlank(eventId))
        {
            Map<String, Object> eventParams = new HashMap<>();
            eventParams.put(ObservationConstants.ARGS_CONTENT, content);
            eventParams.put(ObservationConstants.ARGS_CONTENT_ID, content.getId());
            _observationManager.notify(new Event(eventId, getUser(transientVars), eventParams));
        }
    }
    
    /**
     * Get the default event to notify
     * @return the default event
     */
    protected String _getDefaultEvent()
    {
        return ObservationConstants.EVENT_CONTENT_WORKFLOW_CHANGED;
    }
    
    /**
     * Persists the current version and create a new one.
     * @param content the content.
     * @throws WorkflowException if an error occurs.
     */
    protected void _createVersion(WorkflowAwareContent content) throws WorkflowException
    {
        if (content != null)
        {
            // Create a new version
            if (content instanceof VersionableAmetysObject)
            {
                ((VersionableAmetysObject) content).checkpoint();
            }
        }
    }
    
    /**
     * Add a label on the content.
     * @param content the content.
     * @param label the label to add.
     * @throws WorkflowException if an error occurs.
     * @throws RepositoryException if an error occurs.
     */
    protected void _addLabel(WorkflowAwareContent content, String label) throws WorkflowException, RepositoryException
    {
        if (content != null)
        {
            if (content instanceof VersionableAmetysObject)
            {
                ((VersionableAmetysObject) content).addLabel(label, true);
            }
        }
    }
    
}
