/*
 *  Copyright 2011 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.HashMap;
import java.util.Map;

import javax.jcr.NoSuchWorkspaceException;
import javax.jcr.Node;
import javax.jcr.Repository;
import javax.jcr.RepositoryException;
import javax.jcr.Session;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.content.archive.ArchiveConstants;
import org.ametys.cms.repository.CloneComponent;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.cocoon.ActionResultGenerator;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.provider.AbstractRepository;
import org.ametys.plugins.repository.provider.RequestAttributeWorkspaceSelector;

import com.opensymphony.workflow.InvalidInputException;
import com.opensymphony.workflow.WorkflowException;

/**
 * Action for archiving a content.
 */
public class ArchiveContentAction extends ContentWorkflowAction
{
    
    /** The repository. */
    protected Repository _repository;
    
    /** The ametys object resolver. */
    protected AmetysObjectResolver _resolver;
    
    /** The clone component. */
    protected CloneComponent _cloneComponent;
    
    /** Observation manager available to subclasses. */
    protected ObservationManager _observationManager;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _repository = (Repository) serviceManager.lookup(AbstractRepository.ROLE);
        _resolver = (AmetysObjectResolver) serviceManager.lookup(AmetysObjectResolver.ROLE);
        _cloneComponent = (CloneComponent) serviceManager.lookup(CloneComponent.ROLE);
        _observationManager = (ObservationManager) serviceManager.lookup(ObservationManager.ROLE);
    }
    
    @Override
    protected Map<String, String> _act(Redirector redirector, Map objectModel, String source, Parameters parameters, int actionId, Map inputs) throws InvalidInputException, WorkflowException
    {
        WorkflowAwareContent content = _getContent(objectModel);
        
        UserIdentity currentUser = _getUser(objectModel);
        String contentId = content.getId();
        
        Event deletingEvent = _prepareDeletingEvent(currentUser, content);
        Event deletedEvent = _prepareDeletedEvent(currentUser, content);
        
        try
        {
            // Trigger the action.
            Map<String, String> result = super._act(redirector, objectModel, source, parameters, actionId, inputs);
            
            Map<String, Object> resultMap = new HashMap<>();
            
            _observationManager.notify(deletingEvent);
            
            // Move the content and its workflow to the "archives" workspace.
            archiveContent(content, objectModel, resultMap);
            
            Request request = ObjectModelHelper.getRequest(objectModel);
            request.setAttribute(ActionResultGenerator.MAP_REQUEST_ATTR, resultMap);
            
            // Notify that a content has been removed from the default workspace.
            _observationManager.notify(deletedEvent);
            
            String currentWsp = RequestAttributeWorkspaceSelector.getForcedWorkspace(request);
            try
            {
                // Resolve the content in the default workspace.
                RequestAttributeWorkspaceSelector.setForcedWorkspace(request, ArchiveConstants.ARCHIVE_WORKSPACE);
                WorkflowAwareContent archivedContent = _resolver.resolveById(contentId);
                
                if (archivedContent != null)
                {
                    // Notify that the archived content has been created from the archive workspace
                    Map<String, Object> eventParams = new HashMap<>();
                    eventParams.put(ObservationConstants.ARGS_CONTENT, archivedContent);
                    eventParams.put(ObservationConstants.ARGS_CONTENT_NAME, archivedContent.getName());
                    eventParams.put(ObservationConstants.ARGS_CONTENT_ID, archivedContent.getId());
                    _observationManager.notify(new Event(org.ametys.cms.ObservationConstants.EVENT_CONTENT_ARCHIVED, currentUser, eventParams));
                }
            }
            finally
            {
                // Restore context
                RequestAttributeWorkspaceSelector.setForcedWorkspace(request, currentWsp);
            }
            
            return result;
        }
        catch (RepositoryException e)
        {
            throw new WorkflowException("Unable to archive the content " + contentId, e);
        }
    }
    
    /**
     * Prepare event to notify observers
     * @param issuer the responsible for the action
     * @param content the archived content
     * @return the event
     */
    protected Event _prepareDeletingEvent (UserIdentity issuer, Content content)
    {
        Map<String, Object> eventParams = new HashMap<>();
        eventParams.put(ObservationConstants.ARGS_CONTENT, content);
        eventParams.put(ObservationConstants.ARGS_CONTENT_NAME, content.getName());
        eventParams.put(ObservationConstants.ARGS_CONTENT_ID, content.getId());
        
        return new Event(org.ametys.cms.ObservationConstants.EVENT_CONTENT_DELETING, issuer, eventParams);
    }
    
    /**
     * Prepare event to notify observers
     * @param issuer the responsible for the action
     * @param content the archived content
     * @return the event
     */
    protected Event _prepareDeletedEvent (UserIdentity issuer, Content content)
    {
        Map<String, Object> eventParams = new HashMap<>();
        eventParams.put(ObservationConstants.ARGS_CONTENT, content);
        eventParams.put(ObservationConstants.ARGS_CONTENT_NAME, content.getName());
        eventParams.put(ObservationConstants.ARGS_CONTENT_ID, content.getId());

        return new Event(org.ametys.cms.ObservationConstants.EVENT_CONTENT_DELETED, issuer, eventParams);
    }
    
    /**
     * Archive a content.
     * @param content the content to archive.
     * @param objectModel the current object model.
     * @param result the result map
     * @throws RepositoryException The repository exception
     */
    protected void archiveContent(WorkflowAwareContent content, Map objectModel, Map<String, Object> result) throws RepositoryException
    {
        Session session = null;
        Session archiveSession = null;
        
        try
        {
            archiveSession = getArchiveSession();
            
            Node contentNode = content.getNode();
            
            session = contentNode.getSession();
            
            _cloneComponent.cloneContentNodeWithWorkflow(archiveSession, contentNode);
            
            archiveSession.save();
            
            content.remove();
            session.save();
        }
        finally
        {
            if (archiveSession != null)
            {
                archiveSession.logout();
            }
        }
    }
    
    /**
     * Get a repository session on the archive workspace.
     * @return the archive session.
     * @throws RepositoryException The repository exception
     */
    protected Session getArchiveSession() throws RepositoryException
    {
        Session session = null;
        Session archiveSession = null;
        
        try
        {
            archiveSession = _repository.login(ArchiveConstants.ARCHIVE_WORKSPACE);
            
            return archiveSession;
        }
        catch (NoSuchWorkspaceException e)
        {
            session = _repository.login();
            session.getWorkspace().createWorkspace(ArchiveConstants.ARCHIVE_WORKSPACE);
            archiveSession = _repository.login(ArchiveConstants.ARCHIVE_WORKSPACE);

            archiveSession.getRootNode().addNode(AmetysObjectResolver.ROOT_REPO, AmetysObjectResolver.ROOT_TYPE);
            archiveSession.save();
            
            return archiveSession;
        }
        finally
        {
            if (session != null)
            {
                session.logout();
            }
        }
    }
    
}
