/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.workflow;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.duplicate.contents.DuplicateContentsManager;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.plugins.workflow.EnhancedFunction;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.i18n.I18nizableTextParameter;

import com.opensymphony.module.propertyset.PropertySet;
import com.opensymphony.workflow.WorkflowException;

/**
 * OS workflow function to send mail after an action is triggered.
 */
public class CheckDuplicateContentsFunction extends AbstractContentWorkflowComponent implements EnhancedFunction
{
    
    private DuplicateContentsManager _duplicateContentsManager;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _duplicateContentsManager = (DuplicateContentsManager) manager.lookup(DuplicateContentsManager.ROLE);
    }
    
    @Override
    public void execute(Map transientVars, Map args, PropertySet ps) throws WorkflowException
    {   
        WorkflowAwareContent content = getContent(transientVars);
        Map<String, Object> duplicateContentsInfo = _duplicateContentsManager.searchDuplicates(content);
        @SuppressWarnings("unchecked")
        List<Content> duplicateContents = (List<Content>) duplicateContentsInfo.get(DuplicateContentsManager.DUPLICATE_CONTENTS_KEY);
        @SuppressWarnings("unchecked")
        List<Content> nearDuplicateContents = (List<Content>) duplicateContentsInfo.get(DuplicateContentsManager.NEAR_DUPLICATE_CONTENTS_KEY);
        boolean isTooComplex = duplicateContentsInfo.get(DuplicateContentsManager.STATUS_KEY) == DuplicateContentsManager.Status.TOO_COMPLEX;
        
        if (duplicateContents.isEmpty() && nearDuplicateContents.isEmpty() && !isTooComplex)
        {
            return;
        }

        Map<String, Object> result = getResultsMap(transientVars);
        @SuppressWarnings("unchecked")
        List<Object> messages = (List<Object>) result.getOrDefault("messages", new ArrayList<>());
        
        if (!duplicateContents.isEmpty() || !nearDuplicateContents.isEmpty())
        {
            messages.add(_getDuplicateMessage(content, duplicateContents, nearDuplicateContents));
        }

        if (isTooComplex)
        {
            messages.add(_getTooComplexMessage());
        }
        
        result.put("messages", messages);
    }

    private Map<String, Object> _getTooComplexMessage()
    {
        Map<String, Object> message = new HashMap<>();
        message.put("type", "warning");
        List<Object> warningMessages = new ArrayList<>();
        warningMessages.add(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_SAVE_REPORT_TOO_COMPLEX"));
        message.put("message", warningMessages);
        return message;
    }

    private Map<String, Object> _getDuplicateMessage(WorkflowAwareContent content, List<Content> duplicateContents, List<Content> nearDuplicateContents)
    {
        Map<String, Object> message = new HashMap<>();
       
        message.put("type", "warning");
        
        List<Object> warningMessages = new ArrayList<>();
        
        message.put("message", warningMessages);
        
        Map<String, I18nizableTextParameter> introI18nParameters = new HashMap<>();
        introI18nParameters.put("contentTitle", new I18nizableText(content.getTitle()));
        
        warningMessages.add(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_SAVE_REPORT_INTRO", introI18nParameters));
        warningMessages.add("<br/><br/>");

        if (!duplicateContents.isEmpty())
        {

            Map<String, I18nizableTextParameter> duplicatesI18nParameters = new HashMap<>();
            duplicatesI18nParameters.put("duplicatesNumber", new I18nizableText(String.valueOf(duplicateContents.size())));
            
            warningMessages.add(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_SAVE_REPORT_DUPLICATE_CONTENTS", duplicatesI18nParameters));
            _fillToolLink(warningMessages, duplicateContents);
        }
        
        if (!nearDuplicateContents.isEmpty())
        {

            Map<String, I18nizableTextParameter> nearDuplicates18nParameters = new HashMap<>();
            nearDuplicates18nParameters.put("nearDuplicatesNumber", new I18nizableText(String.valueOf(nearDuplicateContents.size())));
            
            warningMessages.add(new I18nizableText("plugin.cms", "PLUGINS_CMS_DUPLICATE_CONTENTS_SAVE_REPORT_NEAR_DUPLICATE_CONTENTS", nearDuplicates18nParameters));
            _fillToolLink(warningMessages, nearDuplicateContents);
        }
        
        return message;
    }
    
    private void _fillToolLink(List<Object> warningMessages, List<Content> nearDuplicateContents)
    {
        warningMessages.add("<ul>");
        nearDuplicateContents.forEach(duplicate -> 
        {
            warningMessages.add("<li>");
            warningMessages.add(_getTitleTag(duplicate));
            warningMessages.add("</li>");
        });
        warningMessages.add("</ul>");
    }

    /**
     * Get the HTML tag to display the content
     * @param duplicate the duplicate content info
     * @return the HTML tag to display the content
     */
    protected String _getTitleTag(Content duplicate)
    {
        String title = duplicate.getTitle();
        if (_contentHelper.isSimple(duplicate))
        {
            return title;
        }
        else
        {
            String onclick = "Ametys.tool.ToolsManager.openTool('uitool-content', {id: '" + duplicate.getId() + "'})";
            return "<a href=\"#\" onclick=\"" + onclick + "\">" + title + "</a>";
        }
    }

    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_CHECK_DUPLICATES_FUNCTION_LABEL");
    }
    
}
