/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.List;
import java.util.Map;

import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.workflow.component.CheckRightsCondition;
import org.ametys.runtime.i18n.I18nizableText;

import com.opensymphony.workflow.WorkflowException;

/**
 * Condition for checking rights of an user for the current action
 * using the right content context.<p>
 * The following configuration can be used for checking rights:<br>
 * <pre>
 * &lt;condition type="avalon"&gt;
 * &nbsp;&nbsp;&lt;arg name="role"&gt;org.ametys.cms.workflow.ContentCheckRigthsCondition&lt;/arg&gt;
 * &nbsp;&nbsp;&lt;arg name="right"&gt;Right_Edition&lt;/arg&gt;
 * &lt;/condition&gt;
 * </pre>
 */
public class ContentCheckRightsCondition extends CheckRightsCondition
{
    @Override
    protected Object _computeContext(Map transientVars, Map args, UserIdentity user, String right) throws WorkflowException
    {
        try
        {
            return getContent(transientVars);
        }
        catch (AmetysRepositoryException e)
        {
            throw new WorkflowException("Unable to retrieve content name", e);
        }
    }
    
    /**
     * Retrieve the content associated with the workflow.
     * @param transientVars the parameters from the call.
     * @return the content.
     * @throws WorkflowException if the content is not found.
     */
    protected WorkflowAwareContent getContent(Map transientVars) throws WorkflowException
    {
        WorkflowAwareContent content = (WorkflowAwareContent) transientVars.get(AbstractContentWorkflowComponent.CONTENT_KEY);
        
        if (content == null)
        {
            throw new WorkflowException("Unable to retrieve content");
        }
        
        // Found in transient variables map
        return content;
    }
    
    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_CHECK_RIGHTS_CONDITION_LABEL");
    }
    
    @Override
    protected I18nizableText _getSingleConditionDescriptionKey(List<String> parameters)
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_CONTENT_CHECK_RIGHTS_CONDITION_DESCRIPTION", parameters);
    }
}
