/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jcr.Node;
import javax.jcr.RepositoryException;
import javax.jcr.Session;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;

import org.ametys.cms.content.ContentHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.cms.workflow.history.ElementWithWorkflow;
import org.ametys.cms.workflow.history.HistoryStep;
import org.ametys.cms.workflow.history.VersionInformation;
import org.ametys.cms.workflow.history.WorkflowHistory;
import org.ametys.cms.workflow.history.WorkflowHistoryHelper;
import org.ametys.core.cocoon.JSonReader;
import org.ametys.plugins.repository.version.VersionAwareAmetysObject;
import org.ametys.plugins.workflow.store.AbstractJackrabbitWorkflowStore;
import org.ametys.plugins.workflow.support.WorkflowProvider;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;

/**
 * This action returns the workflow history of a content
 *
 */
public class ContentHistoryAction extends ContentVersionHistoryAction
{
    /** The workflow provider */
    protected WorkflowProvider _workflowProvider;

    /** The content helper */
    protected ContentHelper _contentHelper;

    /** The workflow history helper */
    protected WorkflowHistoryHelper _workflowHistoryHelper;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _workflowProvider = (WorkflowProvider) serviceManager.lookup(WorkflowProvider.ROLE);
        _contentHelper = (ContentHelper) serviceManager.lookup(ContentHelper.ROLE);
        _workflowHistoryHelper = (WorkflowHistoryHelper) serviceManager.lookup(WorkflowHistoryHelper.ROLE);
    }

    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Request request = ObjectModelHelper.getRequest(objectModel);

        String id = request.getParameter("contentId");
        Content content = _resolver.resolveById(id);

        assert content instanceof VersionAwareAmetysObject;

        Map<String, Object> result = new HashMap<>();
        List<Map<String, Object>> workflowSteps = new ArrayList<>();
        
        Set<String> validatedVersions = new HashSet<>();
        
        try
        {
            if (content instanceof WorkflowAwareContent)
            {
                WorkflowAwareContent workflowAwareContent = (WorkflowAwareContent) content;
                
                AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(workflowAwareContent);
                long workflowId = workflowAwareContent.getWorkflowId();
                int initialActionId = (int) _getInitialActionId(workflow, workflowAwareContent);

                List<VersionInformation> versionsInformation = _resolveVersionInformations((VersionAwareAmetysObject) content);
                ElementWithWorkflow element = new ElementWithWorkflow(
                        _contentHelper.getTitle(content), 
                        workflowAwareContent.getCreationDate(), 
                        workflowAwareContent.getCreator(),
                        versionsInformation
                );
                
                WorkflowHistory workflowHistory = _workflowHistoryHelper.getWorkflowHistory(element, workflowId, workflow, initialActionId);
                for (HistoryStep step : workflowHistory.getSteps())
                {
                    workflowSteps.add(_workflowHistoryHelper.historyStep2Json(step));
                }
                validatedVersions = workflowHistory.getValidatedVersions();
            }
        }
        catch (RepositoryException e)
        {
            throw new ProcessingException("Unable to access version history", e);
        }
        
        for (Map<String, Object> stepInfo : workflowSteps)
        {
            @SuppressWarnings("unchecked")
            List<Map<String, Object>> versions = (List<Map<String, Object>>) stepInfo.get("versions");
            
            for (Map<String, Object> version : versions)
            {
                String versionName = (String) version.get("name");
                if (validatedVersions.contains(versionName))
                {
                    version.put("valid", true);
                }
            }
        }

        result.put("workflow", workflowSteps);
        request.setAttribute(JSonReader.OBJECT_TO_READ, result);
        return EMPTY_MAP;
    }

    private long _getInitialActionId(AmetysObjectWorkflow workflow, WorkflowAwareContent waContent)
    {
        try
        {
            Session session = waContent.getNode().getSession();
            AbstractJackrabbitWorkflowStore workflowStore = (AbstractJackrabbitWorkflowStore) workflow.getConfiguration().getWorkflowStore();
            Node workflowEntryNode = workflowStore.getEntryNode(session, waContent.getWorkflowId());
            return workflowEntryNode.getProperty("ametys-internal:initialActionId").getLong();
        }
        catch (Exception e)
        {
            getLogger().error("Unable to retrieves initial action id for workflow aware content : " + waContent.getId(), e);
            return 0;
        }
    }
}
