/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.cms.workflow;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;

import javax.jcr.RepositoryException;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;

import org.ametys.cms.repository.Content;
import org.ametys.cms.workflow.history.VersionInformation;
import org.ametys.core.cocoon.JSonReader;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.version.VersionAwareAmetysObject;

/**
 * This action returns the jcr history of a content
 */
public class ContentVersionHistoryAction extends ServiceableAction
{
    /** The ametys object resolver */
    protected AmetysObjectResolver _resolver;

    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _resolver = (AmetysObjectResolver) serviceManager.lookup(AmetysObjectResolver.ROLE);
    }

    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        List<Object> result = new ArrayList<>();

        Request request = ObjectModelHelper.getRequest(objectModel);

        String id = request.getParameter("contentId");
        Content content = _resolver.resolveById(id);

        if (content instanceof VersionAwareAmetysObject versionContent)
        {
            for (VersionInformation versionInformation : _resolveVersionInformations(versionContent))
            {
                if (versionInformation.getLabels().contains("NotCompatible"))
                {
                    break;
                }
                
                result.add(Map.of(
                    "label", versionInformation.getVersionName(),
                    "value", versionInformation.getVersionRawName(),
                    "date", DateUtils.dateToString(versionInformation.getCreatedAt())
                ));
            }
        }
        
        request.setAttribute(JSonReader.OBJECT_TO_READ, result);
        return EMPTY_MAP;
    }
    
    /**
     * Get the jcr info
     * @param content The content
     * @return The info
     * @throws RepositoryException if an error occurred
     */
    protected List<VersionInformation> _resolveVersionInformations(VersionAwareAmetysObject content) throws RepositoryException
    {
        List<VersionInformation> versionsInformation = new ArrayList<>();

        for (String revision : content.getAllRevisions())
        {
            VersionInformation versionInformation = new VersionInformation(revision, content.getRevisionTimestamp(revision));

            for (String label : content.getLabels(revision))
            {
                versionInformation.addLabel(label);
            }

            versionsInformation.add(versionInformation);
        }

        // Sort by date descendant
        Collections.sort(versionsInformation, new Comparator<VersionInformation>()
        {
            public int compare(VersionInformation o1, VersionInformation o2)
            {
                try
                {
                    return -o1.getCreatedAt().compareTo(o2.getCreatedAt());
                }
                catch (RepositoryException e)
                {
                    throw new RuntimeException("Unable to retrieve a creation date", e);
                }
            }
        });

        // Set the version name
        int count = versionsInformation.size();
        for (VersionInformation versionInformation : versionsInformation)
        {
            versionInformation.setVersionName(String.valueOf(count--));
        }

        return versionsInformation;
    }
}
