/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.Collections;
import java.util.Iterator;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;

import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.workflow.cocoon.WorkflowAction;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;

import com.opensymphony.workflow.InvalidInputException;
import com.opensymphony.workflow.WorkflowException;
import com.opensymphony.workflow.spi.Step;

/**
 * Action for firing a transition on a content's workflow.<p>
 * The following parameters are supported:
 * <dl>
 *  <dt>actionId
 *  <dd>the id of the action to fire
 * </dl>
 */
public class ContentWorkflowAction extends WorkflowAction
{
    /** Component to get the current user */
    protected CurrentUserProvider _userProvider;
    
    @Override
    public void initialize() throws Exception
    {
        super.initialize();
        _userProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
    }
    
    @Override
    protected Map _act(Redirector redirector, Map objectModel, String source, Parameters parameters, int actionId, Map inputs) throws InvalidInputException, WorkflowException
    {
        // Store the current step id
        WorkflowAwareContent content = _getContent(objectModel);
        
        if (getLogger().isInfoEnabled())
        {
            getLogger().info("User " + _getUser(objectModel) + " try to perform action " + actionId + " on content " + content.getId());
        }
        
        AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(content);
        workflow.doAction(_getWorkflowId(objectModel, source, parameters), actionId, inputs);
        return _getActionResult(redirector, objectModel, source, parameters);
    }
    
    @Override
    protected long _getWorkflowId(Map objectModel, String source, Parameters parameters)
    {
        return _getContent(objectModel).getWorkflowId();
    }
    
    @Override
    protected Map<String, Object> _getInputs(Redirector redirector, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        Map<String, Object> inputs = super._getInputs(redirector, objectModel, source, parameters);

        // Provide the content key
        inputs.put(AbstractContentWorkflowComponent.CONTENT_KEY, _getContent(objectModel));
        
        // Set the workflow comment into the inputs.
        inputs.put("comment", request.getParameter("comment"));
        
        return inputs;
    }
    
    /**
     * Retrieve the content to use.
     * @param objectModel the object model.
     * @return the content found.
     */
    protected WorkflowAwareContent _getContent(Map objectModel)
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        return (WorkflowAwareContent) request.getAttribute(AbstractContentWorkflowComponent.CONTENT_KEY);
    }
    
    /**
     * Retrieve the user responsible of the call.
     * @param objectModel the object model.
     * @return the user responsible of the call.
     */
    protected UserIdentity _getUser (Map objectModel)
    {
        return _userProvider.getUser();
    }
    
    @Override
    protected Map _processWorkflowException(Redirector redirector, Map objectModel, String source, Parameters parameters, long actionId, WorkflowException e) throws Exception
    {
        if (e instanceof InvalidInputWorkflowException)
        {
            // Validation error
            if (getLogger().isDebugEnabled())
            {
                getLogger().debug("Validation error while executing workflow action", e);
            }
            
            return Collections.EMPTY_MAP;
        }
        else
        {
            // Real error
            return super._processWorkflowException(redirector, objectModel, source, parameters, actionId, e);
        }
    }
    
    @Override
    protected String _getExceptionContext(Map objectModel, String source, Parameters parameters)
    {
        return String.format("the workflow of instance id: '%d' / content '%s' / current step '%d' / user '%s'",
                             _getWorkflowId(objectModel, source, parameters),
                             _getContent(objectModel),
                             _getCurrentStep(objectModel),
                             _getUser(objectModel));
    }
    
    /**
     * Get the content current step
     * @param objectModel The object model
     * @return The content current step as 
     */
    protected long _getCurrentStep (Map objectModel)
    {
        try
        {
            WorkflowAwareContent content = _getContent(objectModel);
            AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(content);
            
            Iterator<Step> steps = workflow.getCurrentSteps(content.getWorkflowId()).iterator();
            if (steps.hasNext())
            {
                return steps.next().getStepId();
            }
        }
        catch (Exception e)
        {
            // Log as INFO and return -1, as it's just debug information.
            getLogger().info("An error occurred retrieving a content's workflow steps.", e);
        }
        
        return -1;
    }
}
