/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang.StringUtils;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.repository.Content;
import org.ametys.cms.workflow.copy.CreateContentByCopyFunction;
import org.ametys.core.right.Right;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.runtime.i18n.I18nizableText;

import com.opensymphony.workflow.WorkflowException;

/**
 * Condition for checking if an user has the right to create a content of the requested type.
 */
public class CreateContentByCopyRightCondition extends CreateContentRightCondition
{
    private AmetysObjectResolver _resolver;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    protected Set<String> getRights(Map transientVars) throws WorkflowException
    {
        Set<String> rights = new HashSet<>();
        
        String[] contentTypeIds;
        
        if (transientVars.get(CreateContentFunction.CONTENT_TYPES_KEY) != null)
        {
            contentTypeIds = (String[]) transientVars.get(CreateContentFunction.CONTENT_TYPES_KEY);
        }
        else
        {
            Content baseContent = (Content) transientVars.get(CreateContentByCopyFunction.BASE_CONTENT_KEY);
            if (baseContent == null)
            {
                String baseContentId = (String) transientVars.get(CreateContentByCopyFunction.BASE_CONTENT_ID);
                
                try
                {
                    baseContent = _resolver.resolveById(baseContentId);
                }
                catch (UnknownAmetysObjectException e)
                {
                    throw new WorkflowException("Unable to retrieve the base content for the duplication.");
                }
            }
            
            contentTypeIds = baseContent.getTypes();
        }
        
        for (String contentTypeId : contentTypeIds)
        {
            ContentType contentType = _contentTypeExtensionPoint.getExtension(contentTypeId);
            
            if (contentType == null)
            {
                throw new WorkflowException("Unknown content type: " + contentType);
            }
            
            String right = contentType.getRight();
            if (StringUtils.isNotEmpty(right))
            {
                rights.add(right);
            }
        }
        
        
        return rights;
    }
    
    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_CREATE_CONTENT_BY_COPY_RIGHT_CONDITION_LABEL");
    }
    
    @Override
    public I18nizableText getFullLabel(Map<String, String> argumentsValues)
    {
        String rightNeeded = argumentsValues.get(__RIGHT_KEY);
        String translatedRight = ""; 
        if (StringUtils.isNotBlank(rightNeeded))
        {
            if (rightNeeded.contains("|"))
            {
                String[] rightsOr = StringUtils.split(rightNeeded, '|');
                translatedRight = "<strong>" + _i18nUtils.translate(_rightsExtensionPoint.getExtension(rightsOr[0]).getLabel()) + "</strong>";
                for (String rightId: rightsOr)
                {
                    Right right = _rightsExtensionPoint.getExtension(rightId);
                    translatedRight += _i18nUtils.translate(new I18nizableText("plugin.workflow", "PLUGINS_WORKFLOW_EDITOR_CONDITION_OR")) + "<strong>" + _i18nUtils.translate(right.getDescription()) + "</strong>";
                }
                return new I18nizableText("plugin.cms", "PLUGINS_CMS_CREATE_CONTENT_BY_COPY_RIGHT_MULTIPLES_CONDITION_DESCRIPTION", List.of(translatedRight));
            }
            else if (rightNeeded.contains("&"))
            {
                String[] rightsAnd = StringUtils.split(rightNeeded, '&');
                translatedRight = "<strong>" + _i18nUtils.translate(_rightsExtensionPoint.getExtension(rightsAnd[0]).getLabel()) + "</strong>";
                for (String rightId: rightsAnd)
                {
                    Right right = _rightsExtensionPoint.getExtension(rightId);
                    translatedRight += _i18nUtils.translate(new I18nizableText("plugin.workflow", "PLUGINS_WORKFLOW_EDITOR_CONDITION_AND")) + "<strong>" + _i18nUtils.translate(right.getDescription()) + "</strong>";
                }
                return new I18nizableText("plugin.cms", "PLUGINS_CMS_CREATE_CONTENT_BY_COPY_RIGHT_MULTIPLES_CONDITION_DESCRIPTION", List.of(translatedRight));
            }
            Right right = _rightsExtensionPoint.getExtension(rightNeeded);
            translatedRight = "<strong>" + _i18nUtils.translate(right.getLabel()) + "</strong>";
            return new I18nizableText("plugin.cms", "PLUGINS_CMS_CREATE_CONTENT_BY_COPY_RIGHT_CONDITION_DESCRIPTION_PARAM", List.of(translatedRight));
        }
        return getLabel();
    }
}
