/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang.StringUtils;

import org.ametys.cms.content.RootContentHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.workflow.component.CheckRightsCondition;
import org.ametys.runtime.i18n.I18nizableText;

import com.opensymphony.module.propertyset.PropertySet;
import com.opensymphony.workflow.WorkflowException;

/**
 * Condition for checking if an user has the right to create a content of the requested type.
 */
public class CreateContentRightCondition extends CheckRightsCondition
{
    /** Content type extension point. */
    protected ContentTypeExtensionPoint _contentTypeExtensionPoint;
    /** Helper for root content */
    protected RootContentHelper _rootContentHelper;

    @Override
    public void initialize() throws Exception
    {
        super.initialize();
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) _manager.lookup(ContentTypeExtensionPoint.ROLE);
        _rootContentHelper = (RootContentHelper) _manager.lookup(RootContentHelper.ROLE);
    }
    
    @Override
    public boolean passesCondition(Map transientVars, Map args, PropertySet ps) throws WorkflowException
    {
        Set<String> rights = getRights(transientVars);
        args.put(__RIGHT_KEY, StringUtils.join(rights, '&'));
        
        return !rights.isEmpty() ? super.passesCondition(transientVars, args, ps) : true;
    }
    
    @Override
    protected boolean hasRight(UserIdentity user, String right, Object context)
    {
        return _rightManager.hasRight(user, right, "/cms") == RightResult.RIGHT_ALLOW || _rightManager.hasRight(user, right, _rootContentHelper.getRootContent()) == RightResult.RIGHT_ALLOW;
    }
    
    /**
     * Get the needed right for creation
     * @param transientVars variables that will not be persisted.
     * @return The required right
     * @throws WorkflowException If an error occured
     */
    protected Set<String> getRights(Map transientVars) throws WorkflowException
    {
        Set<String> rights = new HashSet<>();
        
        String[] contentTypes = new String[0];
        if (transientVars.get(CreateContentFunction.CONTENT_TYPES_KEY) != null)
        {
            contentTypes = (String[]) transientVars.get(CreateContentFunction.CONTENT_TYPES_KEY);
        }
        
        for (String contentTypeId : contentTypes)
        {
            ContentType contentType = _contentTypeExtensionPoint.getExtension(contentTypeId);
            
            if (contentType == null)
            {
                throw new WorkflowException("Unknown content type: " + contentType);
            }
            
            String right = contentType.getRight();
            if (StringUtils.isNotEmpty(right))
            {
                rights.add(right);
            }
        }
        
        return rights;
    }
    
    @Override
    public void dispose()
    {
        _manager.release(_contentTypeExtensionPoint);
        _contentTypeExtensionPoint = null;
        super.dispose();
    }
    
    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_CREATE_CONTENT_RIGHT_CONDITION_LABEL");
    }
    
    @Override
    protected I18nizableText _getMultipleConditionsDescriptionKey(List<String> parameters)
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_CREATE_CONTENT_RIGHT_MULTIPLES_CONDITION_DESCRIPTION", parameters);
    }
    
    @Override
    protected I18nizableText _getSingleConditionDescriptionKey(List<String> parameters)
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_CREATE_CONTENT_RIGHT_CONDITION_DESCRIPTION_PARAM", parameters);
    }
    
    @Override
    public I18nizableText getFullLabel(Map<String, String> argumentsValues)
    {
        String rightNeeded = argumentsValues.get(__RIGHT_KEY);
        if (StringUtils.isNotBlank(rightNeeded))
        {
            super.getFullLabel(argumentsValues);
        }
        return getLabel();
    }
}
