/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.solr.client.solrj.SolrServerException;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.content.indexing.solr.SolrIndexer;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.repository.Content;
import org.ametys.plugins.repository.jcr.NameHelper.NameComputationMode;
import org.ametys.runtime.i18n.I18nizableText;

import com.opensymphony.workflow.WorkflowException;

/**
 * OSWorkflow function for creating a entry of reference table.
 */
public class CreateReferenceTableContentFunction extends CreateContentFunction
{
    /** The Solr indexer */
    protected SolrIndexer _solrIndexer;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _solrIndexer = (SolrIndexer) manager.lookup(SolrIndexer.ROLE);
    }
    
    @Override
    protected Map<String, Object> _eventParamsForContentAdded(Content content)
    {
        Map<String, Object> eventParams = super._eventParamsForContentAdded(content);
        eventParams.put(ObservationConstants.ARGS_CONTENT_COMMIT, false);
        return eventParams;
    }
    
    @Override
    protected List<Future> _notifyContentAdded(Content content, Map transientVars) throws WorkflowException
    {
        List<Future> futures = super._notifyContentAdded(content, transientVars);
        
        // Wait for asynchonous observers to finish
        for (Future future : futures)
        {
            try
            {
                future.get();
            }
            catch (ExecutionException | InterruptedException e)
            {
                _logger.error(String.format("Error while waiting for async observer to complete"));
            }
        }
        
        try
        {
            if (ObserverHelper.isNotSuspendedObservationForIndexation())
            {
                _solrIndexer.commit();
            }
        }
        catch (SolrServerException | IOException e)
        {
            _logger.error("An exception occured when trying to commit changes to Solr", e);
        }
        
        return futures;
    }
    
    @Override
    protected NameComputationMode _getDefaultNameComputationMode()
    {
        return NameComputationMode.GENERATED_KEY;
    }
    
    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_CREATE_REFERENCE_TABLE_CONTENT_FUNCTION_LABEL");
    }
}
