/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.plugins.workflow.support.WorkflowHelper;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.Enumerator;

import com.opensymphony.workflow.loader.StepDescriptor;
import com.opensymphony.workflow.loader.WorkflowDescriptor;

/**
 * {@link Enumerator} for the default workflow steps
 *
 */
public class DefaultWorkflowStepEnumerator implements Enumerator<String>, org.ametys.runtime.parameter.Enumerator, Serviceable, Configurable
{
    /** The workflow helper */
    protected WorkflowHelper _workflowHelper;
    
    /** The workflow name */
    protected String _workflowName;

    /** All option : 'disabled' to disable the 'all' option, 
     * 'blank' to use a blank value for 'all' option,
     * or 'concat' to concat all content types' ids for 'all' option
     **/
    protected String _allOption;
    
    /** The workflow step to exclude */
    protected List<Integer> _excludedSteps;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _workflowHelper = (WorkflowHelper) smanager.lookup(WorkflowHelper.ROLE);
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        Configuration customConf = configuration.getChild("enumeration").getChild("custom-enumerator");
        
        _workflowName = customConf.getChild("workflow-name").getValue("content");
        
        Configuration allOptConf = customConf.getChild("all-option", false);
        _allOption = allOptConf != null ? allOptConf.getValue("blank") : "blank";
        
        _excludedSteps = new ArrayList<>();
        
        Configuration excludeConf = customConf.getChild("exclude-workflow-steps", false);
        if (excludeConf != null)
        {
            Configuration[] children = excludeConf.getChildren("id");
            for (Configuration step : children)
            {
                _excludedSteps.add(step.getValueAsInteger());
            }
        }
    }
    
    @Override
    public Map<String, Object> getConfiguration()
    {
        Map<String, Object> config = new HashMap<>();
        
        config.put("workflowName", _workflowName);
        config.put("allOption", _allOption);
        config.put("excludedSteps", _excludedSteps);
        
        return config;
    }
    
    public Map<Object, I18nizableText> getEntries() throws Exception
    {
        return (Map<Object, I18nizableText>) (Object) getTypedEntries();
    }
    
    @Override
    public Map<String, I18nizableText> getTypedEntries() throws Exception
    {
        Map<String, I18nizableText> entries = new LinkedHashMap<>();
        
        // All steps
        if (!_allOption.equals("disabled"))
        {
            entries.put(StringUtils.EMPTY, new I18nizableText("plugin.cms", "WIDGET_COMBOBOX_ALL_OPTIONS"));
        }
        
        WorkflowDescriptor workflowDesc = _workflowHelper.getWorkflowDescriptor(_workflowName);
        List<StepDescriptor> steps = workflowDesc.getSteps();
        
        for (StepDescriptor stepDescriptor : steps)
        {
            int stepId = stepDescriptor.getId();
            if (!_excludedSteps.contains(stepId))
            {
                entries.put(String.valueOf(stepId), new I18nizableText("application", stepDescriptor.getName()));
            }
        }
        
        return entries;
    }
    
    @Override
    public I18nizableText getEntry(String value) throws Exception
    {
        WorkflowDescriptor workflowDesc = _workflowHelper.getWorkflowDescriptor(_workflowName);
        StepDescriptor stepDescriptor = workflowDesc.getStep(Integer.parseInt(value));
        return new I18nizableText("application", stepDescriptor.getName());
    }
}
