/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.content.references.OutgoingReferences;
import org.ametys.cms.content.references.OutgoingReferencesExtractor;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.lock.LockHelper;
import org.ametys.plugins.repository.lock.LockableAmetysObject;
import org.ametys.plugins.workflow.EnhancedFunction;
import org.ametys.runtime.i18n.I18nizableText;

import com.opensymphony.module.propertyset.PropertySet;
import com.opensymphony.workflow.WorkflowException;

/**
 * OSWorkflow function to extract outgoing references in a content.
 * 
 * The required transient variables:
 * - AbstractContentWorkflowComponent.RESULT_MAP_KEY - Map&lt;String, Object&gt; The map containing the results of the function.
 * - AbstractContentWorkflowComponent.RESULT_MAP_KEY.result - String "true" when everything goes fine. Missing in other case.
 * - AbstractContentWorkflowComponent.CONTENT_KEY - WorkflowAwareContent The content that will be edited. Should have the lock token.
 */
public class ExtractOutgoingReferencesFunction extends AbstractContentWorkflowComponent implements EnhancedFunction
{
    /** The outgoing references extractor */
    protected OutgoingReferencesExtractor _outgoingReferencesExtractor;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _outgoingReferencesExtractor = (OutgoingReferencesExtractor) smanager.lookup(OutgoingReferencesExtractor.ROLE);
    }
    
    @Override
    public void execute(Map transientVars, Map args, PropertySet ps) throws WorkflowException
    {
        _logger.info("Performing extract outgoing references workflow function");

        
        // Retrieve current content
        WorkflowAwareContent content = getContent(transientVars);
        UserIdentity user = getUser(transientVars);
        
        if (!(content instanceof ModifiableContent))
        {
            throw new IllegalArgumentException("The provided content " + content.getId() + " is not a ModifiableContent.");
        }
        
        ModifiableContent modifiableContent = (ModifiableContent) content;
        
        try
        {
            LockableAmetysObject lockableContent = null;
            if (content instanceof LockableAmetysObject)
            {
                lockableContent = (LockableAmetysObject) content;
                if (lockableContent.isLocked() && !LockHelper.isLockOwner(lockableContent, user))
                {
                    throw new WorkflowException("User '" + user + "' try to save content '" + modifiableContent.getName() + "' but it is locked by another user");
                }
            }
            
            _extractOutgoingReferences(modifiableContent);
            
            // Commit changes
            modifiableContent.saveChanges();
            
            getResultsMap(transientVars).put("result", "ok");
        }
        catch (AmetysRepositoryException e)
        {
            throw new WorkflowException("Unable to edit content " + modifiableContent + " from the repository", e);
        }
    }

    /**
     * Analyze the content to extract outgoing references and store them
     * @param content The content to analyze
     */
    protected void _extractOutgoingReferences(ModifiableContent content)
    {
        Map<String, OutgoingReferences> outgoingReferencesByPath = _outgoingReferencesExtractor.getOutgoingReferences(content);
        content.setOutgoingReferences(outgoingReferencesByPath);
    }

    @Override
    public FunctionType getFunctionExecType()
    {
        return FunctionType.POST;
    }
    
    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_EXTRACT_OUTGOING_REFERENCES_FUNCTION_LABEL");
    }
}
