/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.core.cocoon.ActionResultGenerator;
import org.ametys.core.util.JSONUtils;
import org.ametys.plugins.workflow.AbstractWorkflowComponent;
import org.ametys.plugins.workflow.cocoon.InitializeWorkflowAction;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;

import com.opensymphony.workflow.InvalidInputException;
import com.opensymphony.workflow.WorkflowException;

/**
 * Action for initializing a transition on a content's workflow.<p>
 * The following parameters are supported:
 * <dl>
 *  <dt>workflowName
 *  <dd>the name of the workflow to instantiate
 *  <dt>actionId
 *  <dd>the id of the action to fire
 * </dl>
 */
public class InitializeContentWorkflowAction extends InitializeWorkflowAction
{
    /** Constant for storing the content type to use into the transient variables map. */
    private static final String __CONTENT_TYPE_KEY = CreateContentFunction.class.getName() + "$contentType";
    /** Constant for storing the content type to use into the transient variables map. */
    private static final String _CONTENT_MIXINS_KEY = CreateContentFunction.class.getName() + "$mixins";
    /** Constant for storing the content title variant to use into the transient variables map. */
    private static final String __CONTENT_TITLE_VARIANTS_KEY = CreateContentFunction.class.getName() + "$contentTitleVariants";
    
    private static final String REQUEST_PARAMETER_PREFIX = "workflow-";
    
    private static final String __DEFAULT_WORKFLOW_NAME = "content";
    
    /** The JSON utils */
    protected JSONUtils _jsonUtils;
    /** The content type extension point */
    protected ContentTypeExtensionPoint _cTypeEP;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _jsonUtils = (JSONUtils) smanager.lookup(JSONUtils.ROLE);
        _cTypeEP = (ContentTypeExtensionPoint) smanager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    @Override
    protected Map _act(Redirector redirector, Map objectModel, String source, Parameters parameters, int actionId, Map inputs) throws InvalidInputException, WorkflowException
    {
        AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow();
        long workflowId = workflow.initialize(_getWorkflowName(objectModel, source, parameters), actionId, inputs);
        return _getActionResult(redirector, objectModel, source, parameters, workflowId);
    }
    
    @Override
    protected Map<String, Object> _getInputs(Redirector redirector, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Map<String, Object> inputs =  super._getInputs(redirector, objectModel, source, parameters);

        // Provide the parameters
        inputs.putAll(_getParameters(objectModel));
        
        // Provide a map for providing data to the generator
        Request request = ObjectModelHelper.getRequest(objectModel);
        Map<String, Object> result = new HashMap<>();
        request.setAttribute(ActionResultGenerator.MAP_REQUEST_ATTR, result);
        inputs.put(AbstractWorkflowComponent.RESULT_MAP_KEY, result);
        
        return inputs;
    }
    
    @Override
    protected String _getWorkflowName(Map objectModel, String source, Parameters parameters)
    {
        String workflowName = ObjectModelHelper.getRequest(objectModel).getParameter(REQUEST_PARAMETER_PREFIX + "workflowName");
        
        if (StringUtils.isEmpty(workflowName))
        {
            // Get default workflow from content type
            String cTypeId = ObjectModelHelper.getRequest(objectModel).getParameter(REQUEST_PARAMETER_PREFIX + __CONTENT_TYPE_KEY);
            ContentType contentType = _cTypeEP.getExtension(cTypeId);
            
            workflowName = contentType.getDefaultWorkflowName().orElse(__DEFAULT_WORKFLOW_NAME);
        }
        
        return workflowName;
    }
    
    /**
     * Return a map of parameters needed by the workflow that will create the content
     * @param objectModel The cocoon object model
     * @return the map with parameters
     */
    protected Map<String, Object> _getParameters(Map objectModel)
    {
        Map<String, Object> parameters = new HashMap<>();
        
        Request request = ObjectModelHelper.getRequest(objectModel);
        Enumeration<String> parameterNames = request.getParameterNames();
        
        while (parameterNames.hasMoreElements())
        {
            String parameterName = parameterNames.nextElement();
            if (parameterName.startsWith(REQUEST_PARAMETER_PREFIX))
            {
                String name = parameterName.substring(REQUEST_PARAMETER_PREFIX.length());
                String value = request.getParameter(parameterName);
                
                if (name.equals(__CONTENT_TYPE_KEY))
                {
                    parameters.put(CreateContentFunction.CONTENT_TYPES_KEY, new String[] {value});
                }
                else if (name.equals(_CONTENT_MIXINS_KEY))
                {
                    parameters.put(CreateContentFunction.CONTENT_MIXINS_KEY, new String[] {value});
                }
                else if (name.equals(__CONTENT_TITLE_VARIANTS_KEY))
                {
                    parameters.put(CreateContentFunction.CONTENT_TITLE_VARIANTS_KEY, _jsonUtils.convertJsonToMap(value));
                }
                else
                {
                    parameters.put(name, value);
                }
            }
        }
        
        return parameters;
    }
    
    @Override
    protected Map _processWorkflowException(Redirector redirector, Map objectModel, String source, Parameters parameters, long actionId, WorkflowException e) throws Exception
    {
        if (e instanceof InvalidInputWorkflowException)
        {
            // Validation error
            if (getLogger().isDebugEnabled())
            {
                getLogger().debug("Validation error while executing workflow action", e);
            }
            
            return Collections.EMPTY_MAP;
        }
        else
        {
            // Real error
            return super._processWorkflowException(redirector, objectModel, source, parameters, actionId, e);
        }
    }
}
