/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.lock.LockHelper;
import org.ametys.plugins.repository.lock.LockableAmetysObject;
import org.ametys.plugins.workflow.EnhancedCondition;
import org.ametys.runtime.i18n.I18nizableText;

import com.opensymphony.module.propertyset.PropertySet;
import com.opensymphony.workflow.WorkflowException;

/**
 * OSWorkflow condition for testing the lock state of a content.<p>
 * Check if the content is locked and if it is the case, the condition
 * passes only if the lock owner is the current caller.
 */
public class LockCondition extends AbstractContentWorkflowComponent implements EnhancedCondition
{
    /** The user manager */
    protected UserManager _userManager;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
    }
    
    public boolean passesCondition(Map transientVars, Map args, PropertySet ps) throws WorkflowException
    {
        WorkflowAwareContent content = getContent(transientVars);
        
        try
        {
            if (content instanceof LockableAmetysObject)
            {
                // Récupérer le content
                LockableAmetysObject lockableContent = (LockableAmetysObject) content;
        
                // Vérifier si le contenu n'est pas verrouillé
                if (lockableContent.isLocked())
                {
                    boolean sameUser = LockHelper.isLockOwner(lockableContent, getUser(transientVars));
                    
                    if (!sameUser)
                    {
                        List<String> conditionFailures = getConditionFailures(transientVars);
                        if (conditionFailures != null)
                        {
                            conditionFailures.add(String.format("Lock condition failed : content: %s is locked by user: %s", lockableContent, lockableContent.getLockOwner()));
                        }
                        
                        List<String> i18nParams = new ArrayList<>();
                        i18nParams.add(_contentHelper.getTitle(content));
                        UserIdentity lockOwnerIdentity = lockableContent.getLockOwner();
                        User lockOwner = _userManager.getUser(lockOwnerIdentity.getPopulationId(), lockOwnerIdentity.getLogin());
                        i18nParams.add(lockOwner != null ? lockOwner.getFullName() : "");
                        i18nParams.add(lockOwnerIdentity.getLogin());
                        
                        addWorkflowError(transientVars, new I18nizableText("plugin.cms", "WORKFLOW_CONTENT_LOCK_CONDITION_FAILED", i18nParams));
                        
                        if (_logger.isDebugEnabled())
                        {
                            _logger.debug(String.format("Content: %s is locked by user: %s, failing condition", lockableContent, lockableContent.getLockOwner()));
                        }
                    } 
                    else if (_logger.isDebugEnabled())
                    {
                        _logger.debug(String.format("Content: %s is locked by current user, passing condition", lockableContent));
                    }
                    
                    return sameUser;
                }
                else
                {
                    // Le contenu n'est pas verrouillé
                    // FIXME faut-il le vérrouiller ?, si oui il faut faire une pre-function
                    return true;
                }
            }
            
            // The content is not lockable: continue.
            return true;
        }
        catch (AmetysRepositoryException e)
        {
            _logger.error("Unable to check the lock state of the content from the repository", e);
            getConditionFailures(transientVars).add(String.format("Lock condition failed"));
            return false;
        }
    }

    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_LOCKABLE_CONDITION_LABEL");
    }
}
