/*
 *  Copyright 2011 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.Date;
import java.util.List;
import java.util.Map;

import org.ametys.cms.alerts.AlertsConstants;
import org.ametys.cms.content.archive.ArchiveConstants;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.plugins.repository.data.holder.ModifiableModelLessDataHolder;
import org.ametys.plugins.repository.version.ModifiableDataAwareVersionableAmetysObject;
import org.ametys.plugins.workflow.EnhancedFunction;
import org.ametys.plugins.workflow.component.WorkflowArgument;
import org.ametys.plugins.workflow.support.WorkflowElementDefinitionHelper;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.StaticEnumerator;

import com.opensymphony.module.propertyset.PropertySet;
import com.opensymphony.workflow.WorkflowException;

/**
 * Workflow function which marks a content archived.
 */
public class MarkContentArchivedFunction extends AbstractContentWorkflowComponent implements EnhancedFunction
{
    
    @Override
    public void execute(Map transientVars, Map args, PropertySet ps) throws WorkflowException
    {
        WorkflowAwareContent content = getContent(transientVars);
        boolean archive = !"true".equals(args.get("unarchive"));
        
        if (archive)
        {
            // TODO CMS-9336 this metadata should be stored as internal using data holders
            content.getMetadataHolder().setMetadata(ArchiveConstants.META_ARCHIVE_DATE, new Date());
            
            if (content instanceof ModifiableDataAwareVersionableAmetysObject)
            {
                ModifiableModelLessDataHolder unversionedMetadataHolder = ((ModifiableDataAwareVersionableAmetysObject) content).getUnversionedDataHolder();
                
                // Remove archive related metadata
                unversionedMetadataHolder.removeValue(ArchiveConstants.META_ARCHIVE_SCHEDULED_DATE);
                unversionedMetadataHolder.removeValue(AlertsConstants.SCHEDULED_ARCHIVING_REMINDER_LAST_DATE);
            }
        }
        else
        {
            // TODO CMS-9336 this metadata should be stored as internal using data holders
            content.getMetadataHolder().removeMetadata(ArchiveConstants.META_ARCHIVE_DATE);
        }
        
        content.saveChanges();
    }
    
    @Override
    public FunctionType getFunctionExecType()
    {
        return FunctionType.PRE;
    }

    @SuppressWarnings("unchecked")
    @Override
    public List<WorkflowArgument> getArguments()
    {
        WorkflowArgument unarchive = WorkflowElementDefinitionHelper.getElementDefinition(
            "unarchive", 
            new I18nizableText("plugin.cms", "PLUGINS_CMS_MARK_CONTENT_ARCHIVED_UNARCHIVE_ARGUMENT_UNARCHIVE_LABEL"),
            new I18nizableText("plugin.cms", "PLUGINS_CMS_MARK_CONTENT_ARCHIVED_UNARCHIVE_ARGUMENT_UNARCHIVE_DESC"),
            false,
            false
        );
        StaticEnumerator<String> booleanStaticEnumerator = new StaticEnumerator<>();
        booleanStaticEnumerator.add(new I18nizableText("plugin.cms", "PLUGINS_CMS_MARK_CONTENT_ARCHIVED_UNARCHIVE_ARGUMENT_TRUE_LABEL"), "true");
        booleanStaticEnumerator.add(new I18nizableText("plugin.cms", "PLUGINS_CMS_MARK_CONTENT_ARCHIVED_UNARCHIVE_ARGUMENT_FALSE_LABEL"), "false");
        unarchive.setEnumerator(booleanStaticEnumerator);
        unarchive.setDefaultValue("true");
        
        return List.of(unarchive);
    }
    
    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_MARK_CONTENT_ARCHIVED_FUNCTION_LABEL");
    }

    @Override
    public I18nizableText getFullLabel(Map<String, String> argumentsValues)
    {
        return "true".equals(argumentsValues.get("unarchive"))
                ? new I18nizableText("plugin.cms", "PLUGINS_CMS_MARK_CONTENT_ARCHIVED_UNARCHIVE_FUNCTION_DESCRIPTION")
                : new I18nizableText("plugin.cms", "PLUGINS_CMS_MARK_CONTENT_ARCHIVED_FUNCTION_DESCRIPTION");
    }
    
}
