/*
 *  Copyright 2011 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang.ArrayUtils;

import org.ametys.cms.CmsConstants;
import org.ametys.cms.ObservationConstants;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.version.VersionableAmetysObject;
import org.ametys.runtime.i18n.I18nizableText;

import com.opensymphony.module.propertyset.PropertySet;
import com.opensymphony.workflow.WorkflowException;

/**
 * OSWorkflow function for validating a content.
 */
public class RemoveLiveLabelFunction extends AbstractContentFunction
{
    /** Observer manager. */
    protected ObservationManager _observerManager;
    /** Current user provider */
    protected CurrentUserProvider _currentUserProvider;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _observerManager = (ObservationManager) manager.lookup(ObservationManager.ROLE);
        _currentUserProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
    }
    
    @Override
    public void execute(Map transientVars, Map args, PropertySet ps) throws WorkflowException
    {
        _logger.info("Performing content unpublish.");
        WorkflowAwareContent content = getContent(transientVars);
        
        try
        {
            unpublishContent(content);
        }
        catch (AmetysRepositoryException e)
        {
            throw new WorkflowException("Error unpublishing the content", e);
        }
        
        notifyObservers(content);
    }

    /**
     * Notify observers of content validation
     * @param content The content to unpublish
     */
    protected void notifyObservers(WorkflowAwareContent content)
    {
        _observerManager.notify(new Event(ObservationConstants.EVENT_CONTENT_UNTAG_LIVE, _currentUserProvider.getUser(), _getEventParams(content)));
    }
    
    /**
     * Get the event params for notifications.
     * @param content The content to unpublish
     * @return a {@link Map} of params
     */
    protected Map<String, Object> _getEventParams(WorkflowAwareContent content)
    {
        Map<String, Object> eventParams = new HashMap<>();
        eventParams.put(ObservationConstants.ARGS_CONTENT, content);
        return eventParams;
    }

    /**
     * Unpublish the content.
     * @param content the content to unpublish.
     * @throws WorkflowException if an error occurs.
     */
    protected void unpublishContent(WorkflowAwareContent content) throws WorkflowException
    {
        if (!(content instanceof VersionableAmetysObject))
        {
            throw new WorkflowException("Invalid content implementation: " + content);
        }
        
        VersionableAmetysObject versionableContent = (VersionableAmetysObject) content;
        
        if (ArrayUtils.contains(versionableContent.getAllLabels(), CmsConstants.LIVE_LABEL))
        {
            versionableContent.removeLabel(CmsConstants.LIVE_LABEL);
        }
        
        content.saveChanges();
    }
    
    @Override
    public FunctionType getFunctionExecType()
    {
        return FunctionType.PRE;
    }
    
    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_REMOVE_LIVE_LABEL_FUNCTION_LABEL");
    }
}
