/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.io.IOException;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.cms.content.compare.ContentComparator;
import org.ametys.cms.content.compare.ContentComparatorChange;
import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableWorkflowAwareContent;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.external.ExternalizableDataProviderExtensionPoint;
import org.ametys.plugins.repository.lock.LockHelper;
import org.ametys.plugins.repository.lock.LockableAmetysObject;
import org.ametys.plugins.repository.model.RepositoryDataContext;
import org.ametys.plugins.repository.version.VersionableAmetysObject;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItemContainer;
import org.ametys.runtime.model.type.DataContext;

import com.opensymphony.module.propertyset.PropertySet;
import com.opensymphony.workflow.WorkflowException;

/**
 * OSWorkflow function to restore an old revision of a content.
 * Builds a Map with the old content's attributes values, and passes it to the
 * {@link EditContentFunction}, which does the real job.
 */
public class RestoreRevisionFunction extends AbstractContentFunction
{
    private AmetysObjectResolver _resolver;
    private ExternalizableDataProviderExtensionPoint _externalizableDataProviderEP;
    private ContentComparator _contentComparator;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _resolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _externalizableDataProviderEP = (ExternalizableDataProviderExtensionPoint) manager.lookup(ExternalizableDataProviderExtensionPoint.ROLE);
        _contentComparator = (ContentComparator) manager.lookup(ContentComparator.ROLE);
    }
    
    @Override
    public void execute(Map transientVars, Map args, PropertySet ps) throws WorkflowException
    {
        WorkflowAwareContent content = getContent(transientVars);
        UserIdentity user = getUser(transientVars);
        
        if (!(content instanceof ModifiableWorkflowAwareContent))
        {
            throw new IllegalArgumentException("The provided content " + content.getId() + " is not a ModifiableWorkflowAwareContent.");
        }
        
        ModifiableWorkflowAwareContent modifiableContent = (ModifiableWorkflowAwareContent) content;
        
        if (content instanceof LockableAmetysObject)
        {
            LockableAmetysObject lockableContent = (LockableAmetysObject) content;
            if (lockableContent.isLocked() && !LockHelper.isLockOwner(lockableContent, user))
            {
                throw new WorkflowException ("The user '" + user + "' try to restore the content '" + content.getId() + "', but this content is locked by the user '" + user + "'");
            }
            else if (lockableContent.isLocked())
            {
                lockableContent.unlock();
            }
        }
        
        String contentVersion = (String) getContextParameters(transientVars).get("contentVersion");
        
        Content oldContent = _resolver.resolveById(content.getId());
        if (oldContent instanceof VersionableAmetysObject)
        {
            ((VersionableAmetysObject) oldContent).switchToRevision(contentVersion);
        }
        
        Map<String, Object> results = getResultsMap(transientVars);
        Map<String, Object> brokenReferences = new HashMap<>();
        results.put("brokenReferences", brokenReferences);
        
        Map<String, Object> parameters = getContextParameters(transientVars);
        Set<String> externalizableData = _externalizableDataProviderEP.getExternalizableDataPaths(content);
        
        // Get values from version to restore
        DataContext context = RepositoryDataContext.newInstance()
            .withExternalizableData(externalizableData)
            .withEmptyValues(true);
        Map<String, Object> newValues = oldContent.dataToMap(context);
        
        String[] changingAttributes;
        // CMS-11242 Only keep values with differences
        try
        {
            changingAttributes = _contentComparator.compare(content, oldContent)
                .getChanges()
                .stream()
                .map(ContentComparatorChange::getAttributeDataPath)
                .map(ModelHelper::getDefinitionPathFromDataPath)
                .distinct()
                .toArray(String[]::new);
        }
        catch (IOException e)
        {
            throw new WorkflowException("An exception occured during content version comparison", e);
        }
        
        View view = View.of(content.getModel(), changingAttributes);
        
        // Exclude invalid links
        newValues = _processContents(view, newValues, "", brokenReferences);
        
        parameters.put(EditContentFunction.VIEW, view);
        parameters.put(EditContentFunction.VALUES_KEY, newValues);
        parameters.put(EditContentFunction.GLOBAL_VALIDATION, false);
        parameters.put(EditContentFunction.QUIT, Boolean.TRUE);
        
        modifiableContent.setLastContributor(user);
        modifiableContent.setLastModified(ZonedDateTime.now());
        
        // Remove the proposal date.
        modifiableContent.setProposalDate(null);
        
        // Commit changes
        modifiableContent.saveChanges();
    }
    
    private ContentValue _processContentValue(ContentValue value, String dataPath, ElementDefinition definition, Map<String, Object> brokenReferences)
    {
        if (_resolver.hasAmetysObjectForId(value.getContentId()))
        {
            return value;
        }
        else if (!brokenReferences.containsKey(dataPath))
        {
            brokenReferences.put(dataPath, definition.getLabel());
        }
        
        return null;
    }
    
    @SuppressWarnings("unchecked")
    private Map<String, Object> _processContents(ViewItemContainer viewItemContainer, Map<String, Object> values, String dataPath, Map<String, Object> brokenReferences)
    {
        if (values == null)
        {
            return null;
        }
        
        Map<String, Object> result = new HashMap<>();
        
        org.ametys.plugins.repository.model.ViewHelper.visitView(viewItemContainer, 
            (element, definition) -> {
                // simple element
                String name = definition.getName();
                Object value = values.get(name);
                
                if (values.containsKey(name))
                {
                    if (definition.getType().getId().equals(ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID))
                    {
                        Object newValue = null;
                        if (value instanceof ContentValue[])
                        {
                            List<ContentValue> validContents = new ArrayList<>();
                            
                            for (ContentValue contentValue : (ContentValue[]) value)
                            {
                                ContentValue newContentValue = _processContentValue(contentValue, dataPath + name, definition, brokenReferences);
                                if (newContentValue != null)
                                {
                                    validContents.add(newContentValue);
                                }
                            }
                            
                            newValue = validContents;
                        }
                        else if (value instanceof ContentValue)
                        {
                            ContentValue newContentValue = _processContentValue((ContentValue) value, dataPath + name, definition, brokenReferences);
                            newValue = newContentValue;
                        }
                        
                        result.put(name, newValue);
                    }
                    else
                    {
                        result.put(name, value);
                    }
                }
            }, 
            (group, definition) -> {
                // composite
                String name = definition.getName();
                if (values.containsKey(name))
                {
                    result.put(name, _processContents(group, (Map<String, Object>) values.get(name), dataPath + name + "/" , brokenReferences));
                }
            }, 
            (group, definition) -> {
                // repeater
                String name = definition.getName();
                if (values.containsKey(name))
                {
                    List<Map<String, Object>> entries = (List<Map<String, Object>>) values.get(name);
                    
                    List<Map<String, Object>> newEntries = new ArrayList<>();
                    for (int i = 0; i < entries.size(); i++)
                    {
                        newEntries.add(_processContents(group, entries.get(i), dataPath + name + "[" + (i + 1) + "]/" , brokenReferences));
                    }
                
                    result.put(name, newEntries);
                }
            }, 
            group -> result.putAll(_processContents(group, values, dataPath, brokenReferences)));
        
        return result;
    }
    
    @Override
    public FunctionType getFunctionExecType()
    {
        return FunctionType.PRE;
    }
    
    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_RESTORE_REVISION_FUNCTION_LABEL");
    }
}
