/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.io.IOException;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.commons.lang.StringUtils;
import org.apache.excalibur.source.SourceResolver;

import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.right.Right;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightsExtensionPoint;
import org.ametys.core.ui.mail.StandardMailBodyHelper;
import org.ametys.core.ui.mail.StandardMailBodyHelper.MailBodyBuilder;
import org.ametys.core.ui.mail.StandardMailBodyHelper.MailBodyBuilder.UserInput;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.util.I18nUtils;
import org.ametys.core.util.mail.SendMailHelper;
import org.ametys.plugins.workflow.EnhancedFunction;
import org.ametys.plugins.workflow.component.WorkflowArgument;
import org.ametys.plugins.workflow.support.WorkflowElementDefinitionHelper;
import org.ametys.plugins.workflow.support.WorkflowProvider;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.i18n.I18nizableTextParameter;
import org.ametys.runtime.model.StaticEnumerator;
import org.ametys.runtime.plugin.component.PluginAware;

import com.opensymphony.module.propertyset.PropertySet;
import com.opensymphony.workflow.WorkflowException;

import jakarta.mail.MessagingException;

/**
 * OS workflow function to send mail after an action is triggered.
 */
public class SendMailFunction extends AbstractContentWorkflowComponent implements EnhancedFunction, Initializable, PluginAware, Contextualizable
{
    /**
     * Provide "false" to prevent the function sending the mail.
     * Useful when making large automatic workflow operations (for instance, when bulk importing and proposing in one action). 
     */
    public static final String SEND_MAIL = "send-mail";
    
    /** The rights key. */
    protected static final String RIGHTS_KEY = "rights";
    /** The mail subject key. */
    protected static final String SUBJECT_KEY = "subjectKey";
    /** The mail body key. */
    protected static final String BODY_KEY = "bodyKey";
    
    /** The rights manager. */
    protected RightManager _rightManager;
    
    /** The users manager. */
    protected UserManager _userManager;
    
    /** The source resolver. */
    protected SourceResolver _sourceResolver;
    
    /** The workflow. */
    protected WorkflowProvider _workflowProvider;
    
    /** The Avalon context. */
    protected Context _context;
    
    /** The plugin name. */
    protected String _pluginName;
    
    /** I18nUtils */
    protected I18nUtils _i18nUtils;
    
    /** The rights extension point */ 
    protected RightsExtensionPoint _rightsExtensionPoint;
    
    @Override
    public void initialize() throws Exception
    {
        _rightManager = (RightManager) _manager.lookup(RightManager.ROLE);
        _userManager = (UserManager) _manager.lookup(UserManager.ROLE);
        _sourceResolver = (SourceResolver) _manager.lookup(SourceResolver.ROLE);
        _workflowProvider = (WorkflowProvider) _manager.lookup(WorkflowProvider.ROLE);
        _i18nUtils = (I18nUtils) _manager.lookup(I18nUtils.ROLE);
        _rightsExtensionPoint = (RightsExtensionPoint) _manager.lookup(RightsExtensionPoint.ROLE);
    }
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void setPluginInfo(String pluginName, String featureName, String id)
    {
        _pluginName = pluginName;
    }
    
    @Override
    public void execute(Map transientVars, Map args, PropertySet ps) throws WorkflowException
    {
        String rightsParam = StringUtils.defaultString((String) args.get(RIGHTS_KEY));
        String subjectI18nKey = StringUtils.defaultString((String) args.get(SUBJECT_KEY));
        String bodyI18nKey = StringUtils.defaultString((String) args.get(BODY_KEY));
        
        Set<String> rights = _getRights(rightsParam);
        
        // If "send-mail" is set to true or is not present in the vars, send the mail.
        boolean dontSendMail = "false".equals(transientVars.get(SEND_MAIL));
        
        if (dontSendMail)
        {
            return;
        }
        
        try
        {
            WorkflowAwareContent content = getContent(transientVars);
            
            Set<String> recipients = getRecipients(transientVars, content, rights);
            
            if (recipients.size() > 0)
            {
                User caller = getCaller(transientVars, content);
                String sender = getSender(transientVars, content);
                String mailSubject = getMailSubject(subjectI18nKey, caller, content, transientVars);
                String mailBody = getMailBody(subjectI18nKey, bodyI18nKey, caller, content, transientVars);
                
                _sendMails(mailSubject, mailBody, recipients, sender);
            }
        }
        catch (Exception e)
        {
            _logger.error("An error occurred: unable to send mail to notify workflow change.", e);
        }
    }

    private Set<String> _getRights(String rightsParam)
    {
        Set<String> rights = new HashSet<>();
        for (String right : rightsParam.split(","))
        {
            if (StringUtils.isNotBlank(right))
            {
                rights.add(right.trim());
            }
        }
        return rights;
    }
    
    /**
     * Get the subject of mail
     * @param subjectI18nKey  the i18n key to use for subject
     * @param user the caller
     * @param content the content
     * @param transientVars the transient variables
     * @return the subject
     */
    protected String getMailSubject (String subjectI18nKey, User user, WorkflowAwareContent content, Map transientVars)
    {
        I18nizableText subjectKey = new I18nizableText(null, subjectI18nKey, getSubjectI18nParams(user, content));
        return _i18nUtils.translate(subjectKey, null); // FIXME Use user preference language
    }
    
    /**
     * Get the text body of mail
     * @param subjectI18nKey the i18n key to use for body's title
     * @param bodyI18nKey the i18n key to use for body
     * @param user the caller
     * @param content the content
     * @param transientVars the transient variables
     * @return the text body
     * @throws IOException if an error occurred while building HTML workflow email
     */
    protected String getMailBody (String subjectI18nKey, String bodyI18nKey, User user, WorkflowAwareContent content, Map transientVars) throws IOException
    {
        // FIXME Use user preference language
        
        MailBodyBuilder bodyBuilder = StandardMailBodyHelper.newHTMLBody()
            .withTitle(getMailSubject(subjectI18nKey, user, content, transientVars))
            .withMessage(new I18nizableText(null, bodyI18nKey, getBodyI18nParams(user, content)))
            .withLink(_getContentUri(content), new I18nizableText("plugin.cms", "WORKFLOW_MAIL_BODY_GO_TO_CONTENT"));
        
        // Get the workflow comment
        String comment = (String) transientVars.get("comment");
        if (StringUtils.isNotEmpty(comment))
        {
            bodyBuilder.withUserInput(new UserInput(user, ZonedDateTime.now(), comment), new I18nizableText("plugin.cms", "WORKFLOW_MAIL_BODY_USER_COMMENT"));
        }
        
        return bodyBuilder.build();
    }
    
    /**
     * Send the notification emails.
     * @param subject the e-mail subject.
     * @param body the e-mail body.
     * @param recipients the recipients emails address.
     * @param from the address sending the e-mail.
     */
    protected void _sendMails(String subject, String body, Set<String> recipients, String from)
    {
        for (String recipient : recipients)
        {
            try
            {
                SendMailHelper.newMail()
                              .withSubject(subject)
                              .withHTMLBody(body)
                              .withSender(from)
                              .withRecipient(recipient)
                              .withAsync(true)
                              .sendMail();
            }
            catch (MessagingException | IOException e)
            {
                _logger.warn("Could not send a workflow notification mail to " + recipient, e);
            }
        }
    }
    
    /**
     * Get the i18n parameters of mail subject
     * @param user the caller
     * @param content the content
     * @return the i18n parameters
     */
    protected List<String> getSubjectI18nParams (User user, WorkflowAwareContent content)
    {
        List<String> params = new ArrayList<>();
        params.add(_contentHelper.getTitle(content));
        return params;
    }
    
    /**
     * Get the i18n parameters of mail body text
     * @param user the caller
     * @param content the content
     * @return the i18n parameters
     */
    protected List<String> getBodyI18nParams (User user, WorkflowAwareContent content)
    {
        List<String> params = new ArrayList<>();
        
        params.add(user.getFullName()); // {0}
        params.add(content.getTitle()); // {1}
        params.add(_getContentUri(content)); // {2}
        
        return params;
    }
    
    /**
     * Get the content uri
     * @param content the content
     * @return the content uri
     */
    protected String _getContentUri(WorkflowAwareContent content)
    {
        String requestUri = _getRequestUri();
        return requestUri + "/index.html?uitool=uitool-content,id:%27" + content.getId() + "%27";
    }
    
    /**
     * Get the request URI.
     * @return the full request URI.
     */
    protected String _getRequestUri()
    {
        return StringUtils.stripEnd(StringUtils.removeEndIgnoreCase(Config.getInstance().getValue("cms.url"), "index.html"), "/");
    }
    
    /**
     * Retrieve the request from which this component is called.
     * @return the request or <code>null</code>.
     */
    public Request _getRequest()
    {
        try 
        {
            return (Request) _context.get(ContextHelper.CONTEXT_REQUEST_OBJECT);
        } 
        catch (ContextException ce)
        {
            _logger.info("Unable to get the request", ce);
            return null;
        }
    }
    
    /**
     * Get the caller of the workflow action 
     * @param transientVars the transient variables
     * @param content content the content
     * @return caller the caller if the workflow function
     * @throws WorkflowException if failed to get caller
     */
    public User getCaller(Map transientVars, WorkflowAwareContent content) throws WorkflowException
    {
        UserIdentity userIdentity = getUser(transientVars);
        return userIdentity != null ? _userManager.getUser(userIdentity) : null;
    }
    
    /**
     * Get the sender for mail
     * @param transientVars the transient variables
     * @param content the content
     * @return the sender email address
     * @throws WorkflowException if failed to get email for sender
     */
    protected String getSender(Map transientVars, WorkflowAwareContent content) throws WorkflowException
    {
        User user = getCaller(transientVars, content);
        return user != null ? user.getEmail() : null;
    }
    
    /**
     * Get the recipients
     * @param transientVars the transient variables
     * @param content the content.
     * @param rights the set of rights to check.
     * @return the recipients.
     * @throws WorkflowException If failed to get recipients
     */
    protected Set<String> getRecipients(Map transientVars, WorkflowAwareContent content, Set<String> rights) throws WorkflowException
    {
        Set<UserIdentity> users = _getUsers(content, rights);
        
        return users.stream()
            .map(_userManager::getUser)
            .filter(Objects::nonNull)
            .map(User::getEmail)
            .filter(StringUtils::isNotBlank)
            .collect(Collectors.toSet());
    }
    
    /**
     * Get the user logins.
     * @param content the content.
     * @param rights the set of rights to check.
     * @return the users.
     * @throws WorkflowException If an error occurred
     */
    protected Set<UserIdentity> _getUsers(WorkflowAwareContent content, Set<String> rights) throws WorkflowException
    {
        Set<UserIdentity> users = new HashSet<>();
        
        Iterator<String> rightIt = rights.iterator();
        
        // First right : add all the granted users.
        if (rightIt.hasNext())
        {
            users.addAll(_rightManager.getAllowedUsers(rightIt.next(), content).resolveAllowedUsers(Config.getInstance().getValue("runtime.mail.massive.sending")));
        }
        
        // Next rights : retain all the granted users.
        while (rightIt.hasNext())
        {
            users.retainAll(_rightManager.getAllowedUsers(rightIt.next(), content).resolveAllowedUsers(Config.getInstance().getValue("runtime.mail.massive.sending")));
        }
        
        return users;
    }
    
    @Override
    public FunctionType getFunctionExecType()
    {
        return FunctionType.POST;
    }

    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_SEND_MAIL_FUNCTION_LABEL");
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public List<WorkflowArgument> getArguments()
    {
        WorkflowArgument rights = WorkflowElementDefinitionHelper.getElementDefinition(
                RIGHTS_KEY,
                new I18nizableText("plugin.cms", "PLUGINS_CMS_SEND_MAIL_ARGUMENT_RIGHTS_KEY_LABEL"),
                new I18nizableText("plugin.cms", "PLUGINS_CMS_SEND_MAIL_ARGUMENT_RIGHTS_KEY_DESCRIPTION"),
                false,
                true
            );
        StaticEnumerator<String> rightsStaticEnumerator = new StaticEnumerator<>();
        for (String rightId : _rightsExtensionPoint.getExtensionsIds())
        {
            Right right = _rightsExtensionPoint.getExtension(rightId);
            Map<String, I18nizableTextParameter> params = Map.of("category", right.getCategory(), "label", right.getLabel());
            rightsStaticEnumerator.add(new I18nizableText("plugin.workflow", "PLUGINS_WORKFLOW_EDITOR_CHECK_RIGHTS_ARGUMENT_RIGHT_KEY_PARAMS_LABEL", params), right.getId());
        }
        rights.setEnumerator(rightsStaticEnumerator);
        
        return List.of(
            WorkflowElementDefinitionHelper.getElementDefinition(
                SUBJECT_KEY,
                new I18nizableText("plugin.cms", "PLUGINS_CMS_SEND_MAIL_ARGUMENT_SUBJECT_KEY_LABEL"),
                new I18nizableText("plugin.cms", "PLUGINS_CMS_SEND_MAIL_ARGUMENT_SUBJECT_KEY_DESCRIPTION"),
                true,
                false
            ), 
            WorkflowElementDefinitionHelper.getElementDefinition(
                BODY_KEY,
                new I18nizableText("plugin.cms", "PLUGINS_CMS_SEND_MAIL_ARGUMENT_BODY_KEY_LABEL"),
                new I18nizableText("plugin.cms", "PLUGINS_CMS_SEND_MAIL_ARGUMENT_BODY_KEY_DESCRIPTION"),
                true,
                false
            ), 
            rights
        );
    }

    @Override
    public I18nizableText getFullLabel(Map<String, String> argumentsValues)
    {
        String rightsParam = StringUtils.defaultString(argumentsValues.get(RIGHTS_KEY));
        if (!rightsParam.isBlank())
        {
            Object[] rightsIds = _getRights(rightsParam).toArray();
            Right right = _rightsExtensionPoint.getExtension((String) rightsIds[0]);
            String concatenatedRights = "<strong>" + _i18nUtils.translate(right.getLabel()) + "</strong>";
            String and = _i18nUtils.translate(new I18nizableText("plugin.cms", "PLUGINS_CMS_CONDITION_AND"));
            for (int i = 1; i < rightsIds.length; i++)
            {
                right = _rightsExtensionPoint.getExtension((String) rightsIds[i]);
                concatenatedRights += and + "<strong>" + _i18nUtils.translate(right.getLabel()) + "</strong>";
            }
            return new I18nizableText("plugin.cms", "PLUGINS_CMS_SEND_MAIL_FUNCTION_RIGHTS_DESCRIPTION", List.of(concatenatedRights));
        }
        return getLabel();
    }
}
