/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.time.ZonedDateTime;
import java.util.List;
import java.util.Map;

import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.workflow.EnhancedFunction;
import org.ametys.plugins.workflow.component.WorkflowArgument;
import org.ametys.plugins.workflow.support.WorkflowElementDefinitionHelper;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.StaticEnumerator;

import com.opensymphony.module.propertyset.PropertySet;
import com.opensymphony.workflow.WorkflowException;

/**
 * Function setting or removing the date at which the content was proposed for
 * validation, i.e. put in a workflow step in which it's awaiting validation.
 */
public class SetProposalDateContentFunction extends AbstractContentWorkflowComponent implements EnhancedFunction
{
    
    /** The "remove" parameter. */
    private static final String __REMOVE_PARAM = "remove";
    
    @Override
    public void execute(Map transientVars, Map args, PropertySet ps) throws WorkflowException
    {
        _logger.info("Performing content proposal for validation.");
        
        // Retrieve current content
        WorkflowAwareContent content = getContent(transientVars);
        boolean remove = "true".equals(args.get(__REMOVE_PARAM));
        
        try
        {
            content.setProposalDate(remove ? null : ZonedDateTime.now());
            content.saveChanges();
        }
        catch (AmetysRepositoryException e)
        {
            throw new WorkflowException("Unable to set the proposal date on the content", e);
        }
        
    }
    
    @Override
    public FunctionType getFunctionExecType()
    {
        return FunctionType.PRE;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public List<WorkflowArgument> getArguments()
    {
        WorkflowArgument remove = WorkflowElementDefinitionHelper.getElementDefinition(
            __REMOVE_PARAM,
            new I18nizableText("plugin.cms", "PLUGINS_CMS_SET_PROPOSAL_DATE_REMOVE_ARGUMENT_REMOVE_LABEL"),
            new I18nizableText("plugin.cms", "PLUGINS_CMS_SET_PROPOSAL_DATE_REMOVE_ARGUMENT_REMOVE_DESCRIPTION"),
            false,
            false
        );
        StaticEnumerator<String> booleanStaticEnumerator = new StaticEnumerator<>();
        booleanStaticEnumerator.add(new I18nizableText("plugin.cms", "PLUGINS_CMS_SET_PROPOSAL_DATE_REMOVE_ARGUMENT_REMOVE_TRUE_LABEL"), "true");
        booleanStaticEnumerator.add(new I18nizableText("plugin.cms", "PLUGINS_CMS_SET_PROPOSAL_DATE_REMOVE_ARGUMENT_REMOVE_FALSE_LABEL"), "false");
        remove.setEnumerator(booleanStaticEnumerator);
        remove.setDefaultValue("true");
        return List.of(remove);
    }

    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_SET_PROPOSAL_DATE_FUNCTION_LABEL");
    }
    
    @Override
    public I18nizableText getFullLabel(Map<String, String> args)
    {
        String toRemove = args.get(__REMOVE_PARAM);
        return ("true".equals(toRemove))
                ? new I18nizableText("plugin.cms", "PLUGINS_CMS_SET_PROPOSAL_DATE_REMOVE_FUNCTION_DESCRIPTION")
                : new I18nizableText("plugin.cms", "PLUGINS_CMS_SET_PROPOSAL_DATE_FUNCTION_DESCRIPTION");
    }
}
