/*
 *  Copyright 2011 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.HashMap;
import java.util.Map;

import javax.jcr.Node;
import javax.jcr.Repository;
import javax.jcr.RepositoryException;
import javax.jcr.Session;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.commons.lang.ArrayUtils;

import org.ametys.cms.ObservationConstants;
import org.ametys.cms.repository.CloneComponent;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.provider.AbstractRepository;
import org.ametys.plugins.repository.provider.RequestAttributeWorkspaceSelector;
import org.ametys.plugins.repository.version.VersionableAmetysObject;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;

import com.opensymphony.workflow.InvalidInputException;
import com.opensymphony.workflow.WorkflowException;

/**
 * Unarchive a content.
 */
public class UnarchiveContentAction extends ContentWorkflowAction
{
    /** The repository. */
    protected Repository _repository;
    
    /** The ametys object resolver. */
    protected AmetysObjectResolver _resolver;
    
    /** The clone component. */
    protected CloneComponent _cloneComponent;
    
    /** Observation manager available to subclasses. */
    protected ObservationManager _observationManager;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _repository = (Repository) serviceManager.lookup(AbstractRepository.ROLE);
        _resolver = (AmetysObjectResolver) serviceManager.lookup(AmetysObjectResolver.ROLE);
        _cloneComponent = (CloneComponent) serviceManager.lookup(CloneComponent.ROLE);
        _observationManager = (ObservationManager) serviceManager.lookup(ObservationManager.ROLE);
    }
    
    @Override
    protected Map<String, String> _act(Redirector redirector, Map objectModel, String source, Parameters parameters, int actionId, Map inputs) throws InvalidInputException, WorkflowException
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        WorkflowAwareContent archivedContent = _getContent(objectModel);
        UserIdentity user = _getUser(objectModel);
        
        String contentId = archivedContent.getId();
        
        try
        {
            Event unarchiveEvent = _prepareArchivedEvent(user, archivedContent);
            
            // Move the content and its workflow back to the default workspace.
            unarchiveContent(archivedContent);
            
            // Resolve the content in the default workspace.
            RequestAttributeWorkspaceSelector.setForcedWorkspace(request, "default");
            WorkflowAwareContent content = _resolver.resolveById(contentId);
            
            if (content instanceof VersionableAmetysObject)
            {
                ((VersionableAmetysObject) content).checkpoint();
            }
            
            request.setAttribute(Content.class.getName(), content);
            inputs.put(AbstractContentWorkflowComponent.CONTENT_KEY, content);
            
            // Trigger the action.
            Map<String, String> result = super._act(redirector, objectModel, source, parameters, actionId, inputs);
            
            // Notify that a content has been added to the default workspace.
            Map<String, Object> eventParams = new HashMap<>();
            eventParams.put(ObservationConstants.ARGS_CONTENT, content);
            eventParams.put(ObservationConstants.ARGS_CONTENT_ID, contentId);
            
            _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_ADDED, user, eventParams));
            
            // Notify that the archived content has been deleted from the archive workspace
            _observationManager.notify(unarchiveEvent);
            
            return result;
        }
        catch (RepositoryException e)
        {
            throw new WorkflowException("Error unarchiving the content " + contentId, e);
        }
    }
    
    /**
     * Prepare event to notify observers
     * @param issuer the responsible for the action
     * @param content the content after being archived
     * @return the event
     * @throws RepositoryException if an error occurred
     */
    protected Event _prepareArchivedEvent (UserIdentity issuer, Content content) throws RepositoryException
    {
        Map<String, Object> eventParams = new HashMap<>();
        eventParams.put(ObservationConstants.ARGS_CONTENT, content);
        eventParams.put(ObservationConstants.ARGS_CONTENT_NAME, content.getName());
        eventParams.put(ObservationConstants.ARGS_CONTENT_ID, content.getId());

        return new Event(org.ametys.cms.ObservationConstants.EVENT_CONTENT_UNARCHIVED, issuer, eventParams);
    }
    
    /**
     * Test if the action can be executed.
     * @param content the content
     * @param actionId the action ID.
     * @param inputs the workflow inputs.
     * @return true if the action is valid in the current environment, false otherwise.
     */
    protected boolean isActionValid(WorkflowAwareContent content, int actionId, Map inputs)
    {
        long workflowId = content.getWorkflowId();
        
        AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(content);
        int[] availableActions = workflow.getAvailableActions(workflowId, inputs);
        
        return ArrayUtils.contains(availableActions, actionId);
    }
    
    /**
     * Unarchive a content.
     * @param content the content (in the archives workspace).
     * @throws RepositoryException If an error occurred
     */
    protected void unarchiveContent(WorkflowAwareContent content) throws RepositoryException
    {
        Session defaultSession = null;
        Session archiveSession = null;
        
        try
        {
            // Open a session on the default workspace.
            defaultSession = _repository.login("default");
            
            // The content node in the archives workspace.
            Node contentNode = content.getNode();
            archiveSession = contentNode.getSession();
            
            // Clone the content in the default workspace.
            _cloneComponent.cloneContentNodeWithWorkflow(defaultSession, contentNode);
            
            // Remove the content in the archives workspace.
            content.remove();
            
            // Save everything.
            defaultSession.save();
            archiveSession.save();
        }
        finally
        {
            if (defaultSession != null)
            {
                defaultSession.logout();
            }
        }
    }
}
