/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.ArrayUtils;

import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentValidator;
import org.ametys.cms.repository.Content;
import org.ametys.core.util.I18nUtils;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeaterEntry;
import org.ametys.plugins.repository.model.CompositeDefinition;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.plugins.workflow.EnhancedCondition;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.parameter.Errors;

import com.opensymphony.module.propertyset.PropertySet;
import com.opensymphony.workflow.WorkflowException;

/**
 * OSWorkflow condition to check all content metadata are valid
 */
public class ValidateContentCondition extends AbstractContentWorkflowComponent implements EnhancedCondition, Initializable
{
    private ContentTypeExtensionPoint _cTypeEP;
    private I18nUtils _i18nUtils;
    
    @Override
    public void initialize() throws Exception
    {
        _cTypeEP = (ContentTypeExtensionPoint) _manager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _i18nUtils = (I18nUtils) smanager.lookup(I18nUtils.ROLE);
    }
    
    @Override
    public boolean passesCondition(Map transientVars, Map args, PropertySet ps) throws WorkflowException
    {
        List<String> conditionFailures = getConditionFailures(transientVars);
        try
        {
            Content content = getContent(transientVars);
            
            List<ModelItem> modelItems = ModelHelper.getModelItems(content.getModel())
                                                .stream()
                                                .filter(modelItem -> !(modelItem instanceof ElementDefinition) || ((ElementDefinition) modelItem).isEditable())
                                                .collect(Collectors.toList());
            
            if (!_validateValues(modelItems, Optional.of(content), "", content, conditionFailures))
            {
                return false;
            }
            
            String[] allContentTypes = ArrayUtils.addAll(content.getTypes(), content.getMixinTypes());
            for (String cTypeId : allContentTypes)
            {
                ContentType contentType = _cTypeEP.getExtension(cTypeId);
                
                for (ContentValidator contentValidator : contentType.getGlobalValidators())
                {
                    Errors errors = new Errors();
                    contentValidator.validate(content, errors);
                    if (errors.hasErrors())
                    {
                        List<String> translatedErrors = errors.getErrors().stream().map(error -> _i18nUtils.translate(error, "en")).collect(Collectors.toList());
                        
                        conditionFailures.add(String.format("Validate condition failed for content '%s' on global validator '%s' with errors %s", content.getId(), contentValidator.getClass().getName(), translatedErrors));
                        return false;
                    }
                }
            }
            
            return true;
        }
        catch (AmetysRepositoryException e)
        {
            conditionFailures.add("Validate metadata condition failed (" + e.getMessage() + ")");
            _logger.error("Cannot check condition ("  + e.getMessage() + "). Assuming it is false.", e);
            return false; 
        }
    }
    
    private boolean _validateValues(List<ModelItem> modelItems, Optional<? extends ModelAwareDataHolder> dataHolder, String dataPath, Content content, List<String> conditionFailures)
    {
        for (ModelItem modelItem : modelItems)
        {
            String name = modelItem.getName();
            if (modelItem instanceof ElementDefinition)
            {
                // simple element
                ElementDefinition definition = (ElementDefinition) modelItem;
                Object value = dataHolder.map(holder -> holder.getValue(name)).orElse(null);
                
                if (!ModelHelper.validateValue(definition, value).isEmpty())
                {
                    conditionFailures.add(String.format("Validate attribute condition failed for content %s on attribute %s with value %s", content.getId(), dataPath + name, String.valueOf(value)));
                    return false;
                }
            }
            else if (modelItem instanceof CompositeDefinition)
            {
                // composite
                Optional<ModelAwareComposite> composite = dataHolder.map(holder -> holder.getComposite(name));
                if (!_validateValues(((CompositeDefinition) modelItem).getChildren(), composite, dataPath + name + "/", content, conditionFailures))
                {
                    return false;
                }
            }
            else if (modelItem instanceof RepeaterDefinition)
            {
                // repeater
                RepeaterDefinition definition = (RepeaterDefinition) modelItem;
                Optional<ModelAwareRepeater> repeater = dataHolder.map(holder -> holder.getRepeater(name));
                
                Optional<List<? extends ModelAwareRepeaterEntry>> entries = repeater.map(ModelAwareRepeater::getEntries);
                
                int repeaterSize = repeater.map(ModelAwareRepeater::getSize).orElse(0);
                int minSize = definition.getMinSize();
                int maxSize = definition.getMaxSize();
                
                if (repeaterSize < minSize
                    || maxSize > 0 && repeaterSize > maxSize)
                {
                    conditionFailures.add(String.format("Validate repeater size condition failed for content %s on repeater %s with size %s", content.getId(), dataPath + name, repeaterSize));
                    return false;
                }
                
                if (entries.isPresent())
                {
                    for (ModelAwareRepeaterEntry entry : entries.get())
                    {
                        if (!_validateValues(definition.getChildren(), Optional.of(entry), dataPath + name + "[" + (entry.getPosition())  + "]/", content, conditionFailures))
                        {
                            return false;
                        }
                    }
                }
            }
        }
        
        return true;
    }

    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_VALIDATE_CONTENT_CONDITION_LABEL");
    }
}
