/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow;

import java.time.ZonedDateTime;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import javax.jcr.RepositoryException;

import org.ametys.cms.CmsConstants;
import org.ametys.cms.ObservationConstants;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableWorkflowAwareContent;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.observation.Event;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.runtime.i18n.I18nizableText;

import com.opensymphony.module.propertyset.PropertySet;
import com.opensymphony.workflow.WorkflowException;

/**
 * OSWorkflow function for validating a content.
 */
public class ValidateContentFunction extends AbstractContentFunction
{
    /** Context parameter for major validation */
    public static final String IS_MAJOR = "validation.major";
    
    @Override
    public void execute(Map transientVars, Map args, PropertySet ps) throws WorkflowException
    {
        _logger.info("Performing content validation");
        
        WorkflowAwareContent content = getContent(transientVars);
        
        if (!(content instanceof ModifiableWorkflowAwareContent))
        {
            throw new IllegalArgumentException("The provided content " + content.getId() + " is not a ModifiableWorkflowAwareContent.");
        }
        
        ModifiableWorkflowAwareContent modifiableContent = (ModifiableWorkflowAwareContent) content;
        
        try
        {
            // Set the validation metadata.
            _validateContent(modifiableContent, transientVars, args);
            // Set the current step ID.
            _setCurrentStepIdAndNotify(modifiableContent, transientVars);
            // Create a new version.
            _createVersion(modifiableContent);
            // Add the live label on the newly created version.
            _addLabel(modifiableContent, CmsConstants.LIVE_LABEL);
        }
        catch (RepositoryException e)
        {
            throw new WorkflowException("Unable to link the workflow to the content", e);
        }
        catch (AmetysRepositoryException e)
        {
            throw new WorkflowException("Unable to validate the content", e);
        }
        
        _notifyObservers(transientVars, modifiableContent);
    }
    
    /**
     * Notify observers of content validation
     * @param transientVars The transient variables
     * @param content The created content
     * @throws AmetysRepositoryException If an error occurred with the repository
     * @throws WorkflowException If an error occurred with the workflow
     */
    protected void _notifyObservers (Map transientVars, Content content) throws AmetysRepositoryException, WorkflowException
    {
        Map<String, Object> eventParams = new HashMap<>();
        eventParams.put(ObservationConstants.ARGS_CONTENT, content);
        eventParams.put(ObservationConstants.ARGS_CONTENT_ID, content.getId());
        _observationManager.notify(new Event(ObservationConstants.EVENT_CONTENT_VALIDATED, getUser(transientVars), eventParams));
    }

    /**
     * Validates the content: set the validation metadata.
     * @param content the content.
     * @param transientVars the transient vars
     * @param args the arguments
     * @throws WorkflowException if an error occurs.
     * @throws RepositoryException if an error occurs.
     */
    protected void _validateContent(ModifiableWorkflowAwareContent content, Map transientVars, Map args) throws WorkflowException, RepositoryException
    {
        UserIdentity user = getUser(transientVars);
        
        ZonedDateTime validationDate = ZonedDateTime.now();
        
        // Last major validation
        boolean isMajor = _isMajorValidation(getContextParameters(transientVars), args);
        if (isMajor)
        {
            if (user != null)
            {
                content.setLastMajorValidator(user);
            }
            content.setLastMajorValidationDate(validationDate);
        }
        
        // Last validation
        if (user != null)
        {
            content.setLastValidator(user);
        }
        content.setLastValidationDate(validationDate);
        
        // First validation
        if (content.getFirstValidationDate() == null)
        {
            if (user != null)
            {
                content.setFirstValidator(user);
            }
            content.setFirstValidationDate(validationDate);
        }
        
        // Remove the proposal date.
        content.setProposalDate(null);
        
        content.saveChanges();
    }
    
    /**
     * Search into context parameters and arguments if major information is filled, if not default value is <code>true</code>.
     * @param contextParameters the context parameters
     * @param args the arguments
     * @return <code>true</code> if the current validation is a major validation.
     */
    protected boolean _isMajorValidation(Map<String, Object> contextParameters, Map args)
    {
        return Optional.of(contextParameters)
            .map(params -> contextParameters.get(IS_MAJOR))
            .map(Boolean.class::cast)
            .orElseGet(() -> !"false".equals(args.get("major")));
    }
    
    @Override
    public FunctionType getFunctionExecType()
    {
        return FunctionType.POST;
    }
    
    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_VALIDATE_CONTENT_FUNCTION_LABEL");
    }
}
