/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow.copy;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang3.BooleanUtils;

import org.ametys.cms.content.CopyContentComponent;
import org.ametys.cms.content.CopyReport;
import org.ametys.cms.content.CopyReport.CopyMode;
import org.ametys.cms.content.CopyReport.CopyState;
import org.ametys.core.ui.Callable;
import org.ametys.core.util.JSONUtils;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Component for content copy
 */
public class CopyContentClientInteraction extends AbstractLogEnabled implements Component, Serviceable
{
    /** Helper for actually edit and copy contents */
    protected CopyContentComponent _copyContentComponent;
    
    private JSONUtils _jsonUtils;
    
    @Override
    public void service(ServiceManager sManager) throws ServiceException
    {
        _copyContentComponent = (CopyContentComponent) sManager.lookup(CopyContentComponent.ROLE);
        _jsonUtils = (JSONUtils) sManager.lookup(JSONUtils.ROLE);
    }
    
    /**
     * Creates a content by copy of another one.<br>
     * Also handle the possible inner duplication depending on the duplication mode for each attribute of type "content". 
     * @param baseContentId The id of content to copy
     * @param newContentTitle The title of content to create
     * @param jsonMap The attributes to copy as a JSON string
     * @param viewNameToCopy The view name to copy. Can be null
     * @param fallbackViewNameToCopy The fallback view name to use if 'viewNameToCopy' does not exist. Can be null
     * @param viewModeToCopy The view mode to copy. Can be null
     * @param initActionId The init workflow action id for copy
     * @param editActionId The workflow action for editing content
     * @return the copy result
     * @throws Exception if an error occurred during copy
     */
    @Callable
    public Map<String, Object> createContentByCopy(String baseContentId, String newContentTitle, String jsonMap, String viewNameToCopy, String fallbackViewNameToCopy, String viewModeToCopy, int initActionId, int editActionId) throws Exception
    {
        return createContentByCopy(baseContentId, newContentTitle, null, jsonMap, viewNameToCopy, fallbackViewNameToCopy, viewModeToCopy, initActionId, editActionId);
    }
    
    /**
     * Creates a content by copy of another one.<br>
     * Also handle the possible inner duplication depending on the duplication mode for each attribute of type "content". 
     * @param baseContentId The id of content to copy
     * @param newContentTitle The title of content to create
     * @param jsonMap The attributes to copy as a JSON string
     * @param viewNameToCopy The view name to copy. Can be null
     * @param fallbackViewNameToCopy The fallback view name to use if 'viewNameToCopy' does not exist. Can be null
     * @param viewModeToCopy The view mode to copy. Can be null
     * @param targetContentType The type of content to create. If null the type(s) of created content will be those of base content.
     * @param initActionId The workflow action id for copy to init the new content
     * @param editActionId The workflow action id for copy to edit the newly created content
     * @return the copy result
     * @throws Exception if an error occurred during copy
     */
    @Callable
    public Map<String, Object> createContentByCopy(String baseContentId, String newContentTitle, String targetContentType, String jsonMap, String viewNameToCopy, String fallbackViewNameToCopy, String viewModeToCopy, int initActionId, int editActionId) throws Exception
    {
        Map<String, Object> result = new HashMap<> ();
        
        if (StringUtils.isEmpty(baseContentId))
        {
            throw new IllegalArgumentException("The id of content to copy parameter is mandatory.");
        }
        
        String viewName = StringUtils.defaultIfEmpty(viewNameToCopy, "default-edition");
        String fallbackViewName = StringUtils.defaultIfEmpty(fallbackViewNameToCopy, "main");
        
        Map<String, Object> copyMap = jsonMap != null ? _jsonUtils.convertJsonToMap(jsonMap) : null;
        
        CopyReport copyReport = _copyContentComponent.copyContent(baseContentId, newContentTitle, copyMap, viewName, fallbackViewName, targetContentType, initActionId);
        
        // CMS-5137 Open directly the main created content after copied
        if (CopyState.SUCCESS.equals(copyReport.getStatus()))
        {
            result.put("mainContentId", copyReport.getTargetContentId());
        }
        
        result.put("contentIds", _getCreatedContentIds(copyReport));
        result.put("report", _buildReport(Arrays.asList(copyReport)));
        
        return result;
    }
    
    /**
     * Edits contents by copied of another one.<br>
     * @param baseContentId The id of content to copy
     * @param targetContentIds The ids of content to edit
     * @param jsonMap The attributes to copy as a JSON string
     * @param viewNameToCopy The view name to copy. Can be null.
     * @param fallbackViewNameToCopy The fallback view name if 'viewName' is not found. Can be null.
     * @param viewModeToCopy The view mode to copy. Can be null
     * @return the copy result
     * @throws Exception if an error occurred during copy
     */
    @Callable
    public Map<String, Object> editContentByCopy(String baseContentId, List<String> targetContentIds, String jsonMap, String viewNameToCopy, String fallbackViewNameToCopy, String viewModeToCopy) throws Exception
    {
        Map<String, Object> result = new HashMap<> ();
        
        if (StringUtils.isEmpty(baseContentId))
        {
            throw new IllegalArgumentException("The id of content to copy parameter is mandatory.");
        }
        
        String viewName = StringUtils.defaultIfEmpty(viewNameToCopy, "default-edition");
        String fallbackViewName = StringUtils.defaultIfEmpty(fallbackViewNameToCopy, "main");
        
        Map<String, Object> copyMap = jsonMap != null ? _jsonUtils.convertJsonToMap(jsonMap) : null;
        
        List<CopyReport> copyReports = new LinkedList<>();
        for (String targetContentId : targetContentIds)
        {
            CopyReport copyReport = _copyContentComponent.editContent(baseContentId, targetContentId, copyMap, viewName, fallbackViewName);
            copyReports.add(copyReport);
        }
        
        result.put("report", _buildReport(copyReports));
        result.put("createdContentIds", _getCreatedContentIds(copyReports));
        result.put("editedContentIds", _getEditedContentIds(copyReports));
        
        return result;
    }
    
    
    /**
     * Build the copy report
     * @param copyReports List of copy report
     * @return The builded report 
     */
    protected Map<String, Object> _buildReport(List<CopyReport> copyReports)
    {
        Map<String, Object > reportMap = new HashMap<>();
        
        Map<String, Object> mainContent = new HashMap<>();
        List<Map<String, Object>> createdContents = new LinkedList<>();
        List<Map<String, Object>> editedContents = new LinkedList<>();
        List<Map<String, Object>> failedContentCopies = new LinkedList<>();
        List<Map<String, Object>> attributesErrors = new LinkedList<>();
        
        reportMap.put("mainContent", mainContent);
        reportMap.put("createdContents", createdContents);
        reportMap.put("editedContents", editedContents);
        reportMap.put("failedContentCopies", failedContentCopies);
        reportMap.put("attributesErrors", attributesErrors);
        
        for (CopyReport copyReport : copyReports)
        {
            _buildReport(copyReport, mainContent, createdContents, editedContents, failedContentCopies, attributesErrors, true);
        }
        
        return reportMap;
    }
    
    /**
     * Helper method used to build the report (rootLevel is false)
     * @param copyReport The report of the copy
     * @param mainContent The map that will receive the copy report informations during a creation at rootLevel.
     * @param createdContents The list that will receive the copy report informations during a creation at another level.
     * @param editedContents The list that will receive the copy report informations during something else than a creation
     * @param failedContentCopies The list that will receive th copy report informations if an error occurred during the copy
     * @param attributesErrors The list of attributes errors reported
     */
    protected void _buildReport(CopyReport copyReport, Map<String, Object> mainContent, List<Map<String, Object>> createdContents, List<Map<String, Object>> editedContents, List<Map<String, Object>> failedContentCopies, List<Map<String, Object>> attributesErrors)
    {
        _buildReport(copyReport, mainContent, createdContents, editedContents, failedContentCopies, attributesErrors, false);
    }

    /**
     * Helper method used to build the report
     * @param copyReport The report of the copy
     * @param mainContent The map that will receive the copy report informations during a creation at rootLevel.
     * @param createdContents The list that will receive the copy report informations during a creation at another level.
     * @param editedContents The list that will receive the copy report informations during something else than a creation
     * @param failedContentCopies The list that will receive th copy report informations if an error occurred during the copy
     * @param attributesErrors The list of attributes errors reported
     * @param rootLevel True indicates the root level of the report.
     */
    protected void _buildReport(CopyReport copyReport, Map<String, Object> mainContent, List<Map<String, Object>> createdContents, List<Map<String, Object>> editedContents, List<Map<String, Object>> failedContentCopies, List<Map<String, Object>> attributesErrors, boolean rootLevel)
    {
        // success entry
        if (CopyState.SUCCESS.equals(copyReport.getStatus()))
        {
            Map<String, Object> contentEntry = new HashMap<>(3);
            contentEntry.put("id", copyReport.getTargetContentId());
            contentEntry.put("title", copyReport.getTargetContentTitle());
            contentEntry.put("isReferenceTable", copyReport.getTargetContentIsReferenceTable());
            contentEntry.put("copiedAttachments", !copyReport.getCopiedAttachments().isEmpty());
            
            if (CopyMode.CREATION.equals(copyReport.getMode()))
            {
                if (rootLevel)
                {
                    // CMS-5137 do not include main created content in the report.
                    mainContent.putAll(contentEntry);
                }
                else
                {
                    createdContents.add(contentEntry);
                }
            }
            else
            {
                editedContents.add(contentEntry);
            }
        }
        else
        {
            // error entry
            Map<String, Object> failedCopyEntry = new HashMap<>(3);
            failedContentCopies.add(failedCopyEntry);
            
            failedCopyEntry.put("id", copyReport.getBaseContentId());
            failedCopyEntry.put("title", copyReport.getBaseContentTitle());
            failedCopyEntry.put("isReferenceTable", BooleanUtils.isNotFalse(copyReport.getBaseContentIsReferenceTable()));
            I18nizableText errorMessage = copyReport.getErrorMessage();
            if (errorMessage != null)
            {
                failedCopyEntry.put("error", errorMessage);
            }
        }
        
        // Merge child reports.
        for (CopyReport childReport : copyReport.getChildReports())
        {
            _buildReport(childReport, mainContent, createdContents, editedContents, failedContentCopies, attributesErrors);
        }
    }
    
    /**
     * Retrieve the identifiers of the created contents through the copy report
     * @param copyReport The report of the copy
     * @return list of identifiers
     */
    protected List<String> _getCreatedContentIds(CopyReport copyReport)
    {
        List<String> contentIds = new ArrayList<>();
        
        if (CopyState.SUCCESS.equals(copyReport.getStatus()) && CopyMode.CREATION.equals(copyReport.getMode()))
        {
            contentIds.add(copyReport.getTargetContentId());
        }
        
        for (CopyReport childReport : copyReport.getChildReports())
        {
            contentIds.addAll(_getCreatedContentIds(childReport));
        }
        
        return contentIds;
    }
    
    /**
     * Retrieve the identifiers of the created contents through the copy reports
     * @param copyReports list of copy reports
     * @return list of identifiers
     */
    protected List<String> _getCreatedContentIds(List<CopyReport> copyReports)
    {
        List<String> contentIds = new ArrayList<>();
        
        for (CopyReport copyReport : copyReports)
        {
            contentIds.addAll(_getCreatedContentIds(copyReport));
        }
        
        return contentIds;
    }
    
    /**
     * Retrieve the identifiers of the edited contents through the copy reports
     * @param copyReports list of copy reports
     * @return list of identifiers
     */
    protected List<String> _getEditedContentIds(List<CopyReport> copyReports)
    {
        List<String> contentIds = new ArrayList<>();
        
        // Edited content are only at the first level.
        for (CopyReport copyReport : copyReports)
        {
            if (CopyState.SUCCESS.equals(copyReport.getStatus()) && CopyMode.EDITION.equals(copyReport.getMode()))
            {
                contentIds.add(copyReport.getTargetContentId());
            }
        }
        
        return contentIds;
    }
}
