/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow.copy;

import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;

import org.ametys.cms.content.CopyContentComponent;
import org.ametys.cms.content.CopyReport;
import org.ametys.cms.content.CopyReport.CopyMode;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.workflow.CreateContentFunction;
import org.ametys.cms.workflow.EditContentFunction;
import org.ametys.core.util.I18nUtils;
import org.ametys.plugins.repository.UnknownAmetysObjectException;
import org.ametys.plugins.workflow.AbstractWorkflowComponent;
import org.ametys.runtime.i18n.I18nizableText;

import com.opensymphony.module.propertyset.PropertySet;
import com.opensymphony.workflow.WorkflowException;

/**
 * OSWorkflow function to create a content by copy of another
 * 
 * The required transient variables:
 * - CreateContentByCopyFunction.BASE_CONTENT_KEY - Content The content that will be used for duplication.
 * - or CreateContentByCopyFunction.BASE_CONTENT_ID - String The id of content The content that will be used for duplication.
 * 
 * - CreateContentByCopyFunction.COPY_MAP_KEY - Map&lt;String, Object&gt; The map of properties to copy. Can be null if CreateContentByCopyFunction.COPY_VIEW_NAME is used
 */
public class CreateContentByCopyFunction extends CreateContentFunction
{
    /** Constant for storing the base content used for the duplication into the transient variables map. */
    public static final String BASE_CONTENT_KEY = CreateContentByCopyFunction.class.getName() + "$baseContent";
    
    /** Constant for storing the id of base content used for the duplication into the transient variables map. */
    public static final String BASE_CONTENT_ID = CreateContentByCopyFunction.class.getName() + "$baseContentId";
    
    /** Constant for storing the map containing the duplication info into the transient variables map. Can be null.*/
    public static final String COPY_MAP_KEY = CreateContentByCopyFunction.class.getName() + "$copyProperties";
    
    /** Constant for storing the copy report object into the transient variables map. Can be null. */
    public static final String COPY_REPORT_KEY = CreateContentByCopyFunction.class.getName() + "$copyReport";
    
    /** Constant for storing the name of view use for copy. */
    public static final String COPY_VIEW_NAME = CreateContentByCopyFunction.class.getName() + "$viewName";
    /** Constant for storing the name of fallback view to use for copy. */
    public static final String COPY_FALLBACK_VIEW_NAME = CreateContentByCopyFunction.class.getName() + "$fallbackViewName";

    /** The content copy component */
    protected CopyContentComponent _copyContentComponent;
    /** The content type helper */
    protected ContentTypesHelper _cTypesHelper;

    /** I18n Utils */
    protected I18nUtils _i18nUtils;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _copyContentComponent = (CopyContentComponent) manager.lookup(CopyContentComponent.ROLE);
        _cTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
    }
    
    @SuppressWarnings({"cast", "unchecked"})
    @Override
    public void execute(Map transientVars, Map args, PropertySet ps) throws WorkflowException
    {
        // Preparation of the transientVars to work with the super method (CreateContentFunction).
        Content baseContent = getBaseContentForCopy(transientVars);
        
        if (baseContent == null)
        {
            throw new WorkflowException("Unable to retrieve the base content for the duplication.");
        }
        
        if (transientVars.get(CreateContentFunction.CONTENT_TITLE_KEY) == null)
        {
            if (_contentHelper.isMultilingual(baseContent))
            {
                Map<String, String> titleVariants = new HashMap<>();
                Map<String, String> baseTitleVariants = _contentHelper.getTitleVariants(baseContent);
                
                for (Entry<String, String> entry : baseTitleVariants.entrySet())
                {
                    titleVariants.put(entry.getKey(), entry.getValue() + _i18nUtils.translate(new I18nizableText("plugin.cms", "CONTENT_COPY_CREATE_CONTENT_TITLE_SUFFIX")));
                }
                
                String defaultLang = (String) transientVars.get(CreateContentFunction.CONTENT_LANGUAGE_KEY);
                Locale defaultLocale = StringUtils.isNotEmpty(defaultLang) ? new Locale(defaultLang) : new Locale(titleVariants.keySet().iterator().next());
                transientVars.put(CreateContentFunction.CONTENT_NAME_KEY, baseContent.getTitle(defaultLocale));
                transientVars.put(CreateContentFunction.CONTENT_TITLE_VARIANTS_KEY, titleVariants);
            }
            else
            {
                transientVars.put(CreateContentFunction.CONTENT_NAME_KEY, baseContent.getTitle(null));
                // If title is not set, the base title is set with a copy suffix. 
                transientVars.put(CreateContentFunction.CONTENT_TITLE_KEY, baseContent.getTitle(null) + _i18nUtils.translate(new I18nizableText("plugin.cms", "CONTENT_COPY_CREATE_CONTENT_TITLE_SUFFIX")));
            }
        }
        else
        {
            transientVars.put(CreateContentFunction.CONTENT_NAME_KEY, (String) transientVars.get(CreateContentFunction.CONTENT_TITLE_KEY));
        }
        
        if (transientVars.get(CreateContentFunction.CONTENT_TYPES_KEY) == null)
        {
            transientVars.put(CreateContentFunction.CONTENT_TYPES_KEY, baseContent.getTypes());
        }
        
        transientVars.put(CreateContentFunction.CONTENT_MIXINS_KEY, baseContent.getMixinTypes());
        transientVars.put(CreateContentFunction.CONTENT_LANGUAGE_KEY, baseContent.getLanguage());
        
        // Super method call.
        super.execute(transientVars, args, ps);
        
        ModifiableContent targetContent = (ModifiableContent) getResultsMap(transientVars).get(CONTENT_KEY);
        
        // process copy map
        String viewName = (String) transientVars.getOrDefault(COPY_VIEW_NAME, "default-edition");
        String fallbackViewName = (String) transientVars.getOrDefault(COPY_FALLBACK_VIEW_NAME, "main");
        
        if (!transientVars.containsKey(COPY_REPORT_KEY))
        {
            transientVars.put(COPY_REPORT_KEY, new CopyReport(baseContent.getId(), _contentHelper.getTitle(baseContent), true, viewName, fallbackViewName, CopyMode.CREATION));
        }
        
        CopyReport copyReport = (CopyReport) transientVars.get(COPY_REPORT_KEY);
        
        Map<String, Object> copyMap = (Map<String, Object>) transientVars.get(COPY_MAP_KEY);
        Map<String, Object> additionalCopyMap = getAdditionalCopyMap(transientVars, baseContent, viewName, fallbackViewName);
        
        // get values from copy map
        Map<String, Object> values = _copyContentComponent.computeValues(baseContent, targetContent, copyMap, additionalCopyMap, viewName, fallbackViewName, copyReport);
        
        // Title attribute must never be copied, it has been set during the content creation.
        values.remove("title");
        
        // process input values for EditContentFunction
        processValues(transientVars, targetContent, values);
        
        Map<String, Object> parameters = new HashMap<>();
        parameters.put(EditContentFunction.VALUES_KEY, values);
        parameters.put(EditContentFunction.QUIT, true);
        transientVars.put(AbstractWorkflowComponent.CONTEXT_PARAMETERS_KEY, parameters);
    }
    
    /**
     * Called after creation and before execution of the {@link EditContentFunction} to allow subclasses to add/remove/modify the values.
     * @param transientVars The workflow transient vars
     * @param targetContent The newly created content
     * @param values the values computed from the copyMap and to be given to the EditContentFunction.
     * @throws WorkflowException if an error occurs
     */
    protected void processValues(Map transientVars, ModifiableContent targetContent, Map<String, Object> values) throws WorkflowException
    {
        // do nothing
    }
    
    /**
     * Get the content
     * @param transientVars The workflow transient vars with BASE_CONTENT_KEY (the content itself) or BASE_CONTENT_ID (the id of the content)
     * @return the content or null if the object cannot be found
     */
    protected Content getBaseContentForCopy (Map transientVars)
    {
        Content baseContent = (Content) transientVars.get(BASE_CONTENT_KEY);
        
        if (baseContent == null)
        {
            String baseContentId = (String) transientVars.get(BASE_CONTENT_ID);
            
            try
            {
                baseContent = _resolver.resolveById(baseContentId);
            }
            catch (UnknownAmetysObjectException e)
            {
                return null;
            }
        }
        
        return baseContent;
    }

    /**
     * Get an additional copy map, if any. It allows subclasses to provide additional data to copy.
     * This map is concatenated to the initial copy map. 
     * @param transientVars the workflow parameters
     * @param content the source content to be copied
     * @param viewName the view name
     * @param fallbackViewName the fallback view name, if required view does not exist
     * @return the additional copy map, or null if none.
     * @throws WorkflowException if an error occurs
     */
    protected Map<String, Object> getAdditionalCopyMap(Map transientVars, Content content, String viewName, String fallbackViewName) throws WorkflowException
    {
        // the default implementation does not provide any additional copy map
        return null;
    }
    
    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_CREATE_CONTENT_BY_COPY_FUNCTION_LABEL");
    }
}
