/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow.extensions;

import java.util.Collections;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.plugins.workflow.EnhancedFunction;
import org.ametys.plugins.workflow.EnhancedFunctionExtensionPoint;
import org.ametys.plugins.workflow.component.WorkflowArgument;
import org.ametys.plugins.workflow.support.WorkflowElementDefinitionHelper;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.StaticEnumerator;
import org.ametys.runtime.plugin.ExtensionPoint;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

import com.opensymphony.module.propertyset.PropertySet;
import com.opensymphony.workflow.FunctionProvider;
import com.opensymphony.workflow.WorkflowException;

/**
 * Workflow function that will call each function registered on the extension point given as an argument 
 */
public class ExtensibleFunction extends AbstractLogEnabled implements EnhancedFunction, Serviceable
{
    /** The rights key. */
    protected static final String EXTENSION_POINT_ROLE_ARGS_KEY = "extension-point";
    /** The service manager */
    protected ServiceManager _manager;
    /** The enhanced function extension point */
    protected EnhancedFunctionExtensionPoint _enhancedFunctionEP;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
        _enhancedFunctionEP  = (EnhancedFunctionExtensionPoint) manager.lookup(EnhancedFunctionExtensionPoint.ROLE);
    }
    
    
    public void execute(Map transientVars, Map args, PropertySet ps) throws WorkflowException
    {
        ExtensionPoint<String> functionsExtensionPoint = _getFunctionsExtensionPoint(args);
        for (String functionId : functionsExtensionPoint.getExtensionsIds())
        {
            getLogger().debug("{} is delegating to extension {} of point {}", ExtensibleFunction.class.getName(), functionId);
            String functionProviderRole = functionsExtensionPoint.getExtension(functionId);
            FunctionProvider functionProvider = _getFunctionProvider(functionId, functionProviderRole);
            functionProvider.execute(transientVars, Collections.emptyMap(), ps); // We do not transmit 'args', since those args are for this function, not for the extended functions that thus cannot use args
        }

    }
    
    /**
     * Get the function provider denoted
     * @param functionId The id of the extension that gave the role
     * @param functionProviderRole The role of the component to find
     * @return The workflow function
     * @throws WorkflowException If the role cannot be found
     */
    protected FunctionProvider _getFunctionProvider(String functionId, String functionProviderRole) throws WorkflowException
    {
        EnhancedFunction extension = _enhancedFunctionEP.getExtension(functionProviderRole);
        if (extension == null)
        {
            throw new WorkflowException("Cannot find workflow function with id '" + functionProviderRole + "' configured as an extension with id '" + functionId + "'");
        }
        return extension;
    }


    /**
     * Get the function extension point given in args
     * @param args The workflow arguments
     * @return The extension point
     * @throws WorkflowException If the extension point is not configured or does not exist
     */
    @SuppressWarnings("unchecked")
    protected ExtensionPoint<String> _getFunctionsExtensionPoint(Map args) throws WorkflowException
    {
        String extensionPoint = (String) args.get(EXTENSION_POINT_ROLE_ARGS_KEY);
        if (StringUtils.isBlank(extensionPoint))
        {
            throw new WorkflowException("Cannot use the workflow function '" + ExtensibleFunction.class.getName() + "' without the argument '" + EXTENSION_POINT_ROLE_ARGS_KEY + "' describing the role of the extension point of function components to use");
        }
        
        try
        {
            return (ExtensionPoint<String>) _manager.lookup(extensionPoint);
        }
        catch (ServiceException e)
        {
            throw new WorkflowException("The workflow function '" + ExtensibleFunction.class.getName() + "' has the argument '" + EXTENSION_POINT_ROLE_ARGS_KEY + "' valued to '" + extensionPoint + "', that does not denote a String extension point", e);
        }
    }
    
    @Override
    public FunctionType getFunctionExecType()
    {
        return FunctionType.POST;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public List<WorkflowArgument> getArguments()
    {
        WorkflowArgument extensionPoint = WorkflowElementDefinitionHelper.getElementDefinition(
            EXTENSION_POINT_ROLE_ARGS_KEY,
            new I18nizableText("plugin.cms", "PLUGINS_CMS_EXTENSIBLE_FUNCTION_ARGUMENT_EXTENSION_POINT_LABEL"),
            new I18nizableText("plugin.cms", "PLUGINS_CMS_EXTENSIBLE_FUNCTION_ARGUMENT_EXTENSION_POINT_DESC"),
            true,
            false
        );
        StaticEnumerator<String> extensionPointsStaticEnumerator = new StaticEnumerator<>();
        extensionPointsStaticEnumerator.add(new I18nizableText("plugin.cms", "PLUGINS_CMS_EXTENSIBLE_FUNCTION_POST_CONTENT_EDITION_LABEL"), "org.ametys.cms.workflow.extensions.PostContentEditionFunctionsExtensionPoint");
        extensionPointsStaticEnumerator.add(new I18nizableText("plugin.cms", "PLUGINS_CMS_EXTENSIBLE_FUNCTION_POST_CONTENT_VALIDATION_LABEL"), "org.ametys.cms.workflow.extensions.PostContentValidationFunctionsExtensionPoint");
        extensionPoint.setEnumerator(extensionPointsStaticEnumerator);
        extensionPoint.setDefaultValue("org.ametys.cms.workflow.extensions.PostContentEditionFunctionsExtensionPoint");
        return List.of(extensionPoint);
    }
    
    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.cms", "PLUGINS_CMS_EXTENSIBLE_FUNCTION_LABEL");
    }

    @Override
    public I18nizableText getFullLabel(Map<String, String> args)
    {
        return (PostContentEditionFunctionsExtensionPoint.class.getName().equals(args.get(EXTENSION_POINT_ROLE_ARGS_KEY)))
                ? new I18nizableText("plugin.cms", "PLUGINS_CMS_EXTENSIBLE_FUNCTION_EDITION_DESCRIPTION")
                : new I18nizableText("plugin.cms", "PLUGINS_CMS_EXTENSIBLE_FUNCTION_VALIDATION_DESCRIPTION");
    }
}
