/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow.history;

import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.List;

import org.ametys.core.user.UserIdentity;

/**
 * Object representing an element with a workflow
 */
public class ElementWithWorkflow
{
    private String _title;
    private ZonedDateTime _creationDate;
    private UserIdentity _creator;
    private boolean _isVersionable;
    private List<VersionInformation> _versions;
    
    /**
     * The constructor for a no versionable element
     * @param title the title
     * @param creationDate the creation date
     * @param creator the creator
     */
    public ElementWithWorkflow(String title, ZonedDateTime creationDate, UserIdentity creator)
    {
        this._title = title;
        this._creationDate = creationDate;
        this._creator = creator;
        this._isVersionable = false;
        this._versions = new ArrayList<>();
    }
    
    /**
     * The constructor for an versionable element
     * @param title the title
     * @param creationDate the creation date
     * @param creator the creator
     * @param versions the list of versions
     */
    public ElementWithWorkflow(String title, ZonedDateTime creationDate, UserIdentity creator, List<VersionInformation> versions)
    {
        this(title, creationDate, creator);
        this._isVersionable = true;
        this._versions = versions;
    }
    
    /**
     * Get the title of the element
     * @return the title
     */
    public String getTitle()
    {
        return _title;
    }
    
    /**
     * Get the creation date of the element
     * @return the creation date
     */
    public ZonedDateTime getCreationDate()
    {
        return _creationDate;
    }
    
    /**
     * Get the creator of the element
     * @return the creator
     */
    public UserIdentity getCreator()
    {
        return _creator;
    }
    
    /**
     * <code>true</code> if the element is versionable
     * @return <code>true</code> if the element is versionable
     */
    public boolean isVersionable()
    {
        return _isVersionable;
    }
    
    /**
     * The list of versions of the element
     * @return the list of versions
     */
    public List<VersionInformation> getVersions()
    {
        return _versions;
    }
}
