/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow.history;

import java.util.Date;
import java.util.List;
import java.util.Optional;

import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.workflow.store.AmetysStep;

import com.opensymphony.workflow.spi.Step;

/**
 * Object representing a history step
 */
public class HistoryStep 
{
    private int _id;
    private boolean _isCurrent;
    private boolean _isValid;
    private DescriptionStep _description;
    private WorkflowAction _workflowAction;
    private Optional<UserIdentity> _caller;
    private Optional<String> _comments;
    private Date _date;
    private Optional<List<VersionInformation>> _versions;
    
    /**
     * The constructor for a history step
     * @param step the step
     */
    public HistoryStep(Step step)
    {
        this._id = step.getStepId();
        this._isCurrent = step.getFinishDate() == null;
        this._isValid = _isValid(step);
        this._caller = Optional.empty();
        this._comments = Optional.empty();
        this._versions = Optional.empty();
    }
    
    /**
     * Get the id of the step
     * @return the id of the step
     */
    public int getId()
    {
        return this._id;
    }
    
    /**
     * <code>true</code> if the step is the current step
     * @return <code>true</code> if the step is the current step
     */
    public boolean isCurrent()
    {
        return this._isCurrent;
    }
    
    /**
     * <code>true</code> if the step is valid
     * @return <code>true</code> if the step is valid
     */
    public boolean isValid()
    {
        return this._isValid;
    }
    
    /**
     * Get the description of the step
     * @return the description of the step
     */
    public DescriptionStep getDescription()
    {
        return this._description;
    }
    
    /**
     * Set the description of the step
     * @param description the description of the step
     */
    public void setDescription(DescriptionStep description)
    {
        this._description = description;
    }
    
    /**
     * Get the workflow action responsible for being in the step
     * @return the workflow action
     */
    public WorkflowAction getWorkflowAction()
    {
        return this._workflowAction;
    }
    
    /**
     * Set the workflow action responsible for being in the step
     * @param workflowAction the workflow action
     */
    public void setWorkflowAction(WorkflowAction workflowAction)
    {
        this._workflowAction = workflowAction;
    }
    
    /**
     * Get the caller
     * @return the caller
     */
    public Optional<UserIdentity> getCaller()
    {
        return this._caller;
    }
    
    /**
     * Set the caller
     * @param caller the caller
     */
    public void setCaller(UserIdentity caller)
    {
        this._caller = Optional.ofNullable(caller);
    }
    
    /**
     * Get the comments
     * @return the comments
     */
    public Optional<String> getComments()
    {
        return this._comments;
    }
    
    /**
     * Set the comments
     * @param comments the comments
     */
    public void setComments(String comments)
    {
        this._comments = Optional.ofNullable(comments);
    }
    
    /**
     * Get the date
     * @return the date
     */
    public Date getDate()
    {
        return this._date;
    }
    
    /**
     * Set the date
     * @param date the date
     */
    public void setDate(Date date)
    {
        this._date = date;
    }

    /**
     * Get the versions of the step
     * @return the versions of the step
     */
    public Optional<List<VersionInformation>> getVersions()
    {
        return this._versions;
    }
    
    /**
     * Set the versions of the step
     * @param versions the versions of the step. Can be null;
     */
    public void setVersions(List<VersionInformation> versions)
    {
        this._versions = Optional.ofNullable(versions);
    }
    
    private boolean _isValid(Step step)
    {
        Boolean validation = false;
        if (step instanceof AmetysStep)
        {
            validation = (Boolean) ((AmetysStep) step).getProperty("validation");
        }

        return validation == null ? false : validation;
    }
}
