/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow.history;

import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import javax.jcr.RepositoryException;

/**
 * Object representing the version information
 */
public class VersionInformation
{
    private String _rawName;
    private String _name;
    private Date _creationDate;
    private Set<String> _labels = new HashSet<>();

    /**
     * Creates a {@link VersionInformation}.
     * 
     * @param rawName the revision name.
     * @param creationDate the revision creation date.
     * @throws RepositoryException if an error occurs.
     */
    public VersionInformation(String rawName, Date creationDate) throws RepositoryException
    {
        _creationDate = creationDate;
        _rawName = rawName;
    }

    /**
     * Creates a named {@link VersionInformation}.
     * 
     * @param name the name to use.
     */
    public VersionInformation(String name)
    {
        _name = name;
    }

    /**
     * Set the version name
     * 
     * @param name The name
     */
    public void setVersionName(String name)
    {
        _name = name;
    }

    /**
     * Retrieves the version name.
     * 
     * @return the version name.
     */
    public String getVersionName()
    {
        return _name;
    }

    /**
     * Retrieves the version raw name.
     * 
     * @return the version raw name.
     */
    public String getVersionRawName()
    {
        return _rawName;
    }

    /**
     * Retrieves the creation date.
     * 
     * @return the creation date.
     * @throws RepositoryException if an error occurs.
     */
    public Date getCreatedAt() throws RepositoryException
    {
        return _creationDate;
    }

    /**
     * Retrieves the labels associated with this version.
     * 
     * @return the labels.
     */
    public Set<String> getLabels()
    {
        return _labels;
    }

    /**
     * Add a label to this version.
     * 
     * @param label the label.
     */
    public void addLabel(String label)
    {
        _labels.add(label);
    }
}
