/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow.history;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import javax.jcr.RepositoryException;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.ProcessingException;
import org.apache.commons.lang.StringUtils;

import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.plugins.workflow.store.AmetysStep;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.i18n.I18nizableTextParameter;

import com.opensymphony.workflow.Workflow;
import com.opensymphony.workflow.loader.ActionDescriptor;
import com.opensymphony.workflow.loader.StepDescriptor;
import com.opensymphony.workflow.loader.WorkflowDescriptor;
import com.opensymphony.workflow.spi.SimpleStep;
import com.opensymphony.workflow.spi.Step;

/**
 * This action returns the workflow history of a element with workflow
 */
public class WorkflowHistoryHelper implements Component, Serviceable
{
    /** The component role. */
    public static final String ROLE = WorkflowHistoryHelper.class.getName();
    
    private static final I18nizableText __MESSAGE_NO_STEP = new I18nizableText("plugin.cms", "WORKFLOW_UNKNOWN_STEP");

    private static final I18nizableText __MESSAGE_NO_ACTION = new I18nizableText("plugin.cms", "WORKFLOW_UNKNOWN_ACTION");

    /** The user helper */
    protected UserHelper _userHelper;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _userHelper = (UserHelper) manager.lookup(UserHelper.ROLE);
    }

    /**
     * Get the workflow history
     * @param objectWithWorkflow the object with workflow
     * @param workflowId the workflow id
     * @param workflow the workflow
     * @param initialActionId the initial action id
     * @return a workflow history
     * @throws ProcessingException if a procession error occurred
     * @throws RepositoryException if a repository error occurred
     */
    public WorkflowHistory getWorkflowHistory(ElementWithWorkflow objectWithWorkflow, long workflowId, Workflow workflow, int initialActionId) throws ProcessingException, RepositoryException
    {
        WorkflowHistory workflowHistory = new WorkflowHistory();
        
        String workflowName = workflow.getWorkflowName(workflowId);

        if (workflowName == null)
        {
            throw new ProcessingException("Unknown workflow name for workflow instance id: " + workflowId);
        }

        WorkflowDescriptor workflowDescriptor = workflow.getWorkflowDescriptor(workflowName);

        if (workflowDescriptor == null)
        {
            throw new ProcessingException("No workflow description for workflow name: " + workflowName);
        }
        
        @SuppressWarnings("unchecked")
        List<Step> allSteps = new ArrayList(workflow.getCurrentSteps(workflowId));
        allSteps.addAll(workflow.getHistorySteps(workflowId));

        // Sort by start date descendant (only relevant when there is no
        // split and nor join!)
        Collections.sort(allSteps, new Comparator<Step>()
        {
            public int compare(Step s1, Step s2)
            {
                return -s1.getStartDate().compareTo(s2.getStartDate());
            }
        });

        Date dueDate = null;

        if (!allSteps.isEmpty())
        {
            // Use start date of the first step for the due date
            // of the unstored "real first step"
            dueDate = allSteps.get(allSteps.size() - 1).getStartDate();
        }

        // Use a custom step for the unstored "real first step"
        Date startDate = DateUtils.asDate(objectWithWorkflow.getCreationDate());
        allSteps.add(new SimpleStep(0, 0, 0, initialActionId, null, startDate, dueDate, null, "", new long[0], UserIdentity.userIdentityToString(objectWithWorkflow.getCreator())));

        Iterator<Step> itStep = allSteps.iterator();
        // We got at least one
        Step step = itStep.next();

        do
        {
            Step previousStep = null;

            if (itStep.hasNext())
            {
                previousStep = itStep.next();
            }

            _handleStep(objectWithWorkflow, workflowDescriptor, step, previousStep, workflowHistory);

            step = previousStep;
        }
        while (itStep.hasNext());
        
        
        return workflowHistory;
    }
    
    private void _handleStep(ElementWithWorkflow objectWithWorkflow, WorkflowDescriptor workflowDescriptor, Step step, Step previousStep, WorkflowHistory workflowHistory) throws RepositoryException
    {
        HistoryStep historyStep = new HistoryStep(step);
        
        // We want the caller of the action responsible for being in the step
        String caller = previousStep != null ? previousStep.getCaller() : null;
        UserIdentity callerIdentity = UserIdentity.stringToUserIdentity(caller);
        historyStep.setCaller(callerIdentity);
        
        historyStep.setDescription(_getDescription(workflowDescriptor, step.getStepId()));

        // We want the id of the action responsible for being in the step
        int actionId = previousStep != null ? previousStep.getActionId() : 0;
        historyStep.setWorkflowAction(getWorkfowAction(callerIdentity, objectWithWorkflow.getTitle(), workflowDescriptor, actionId));

        // Comment
        if (previousStep != null && previousStep instanceof AmetysStep)
        {
            String comments = (String) ((AmetysStep) previousStep).getProperty("comment");
            historyStep.setComments(comments);
        }

        Date actionFinishDate = null;
        if (step instanceof AmetysStep)
        {
            actionFinishDate = (Date) ((AmetysStep) step).getProperty("actionFinishDate");
        }

        Date startDate = null;
        Date finishDate = null;
        if (actionFinishDate != null)
        {
            // New format
            startDate = step.getStartDate() != null ? step.getStartDate() : null;
            finishDate = step.getFinishDate() != null ? step.getFinishDate() : null;
        }
        else
        {
            startDate = previousStep != null ? previousStep.getFinishDate() : null;
            finishDate = step.getStartDate();
        }

        historyStep.setDate(_getDate(startDate, finishDate, actionFinishDate));

        // Versions
        if (objectWithWorkflow.isVersionable())
        {
            List<VersionInformation> versions = _getVersionsBetween(objectWithWorkflow.getVersions(), startDate, finishDate);
            historyStep.setVersions(versions);
            if (historyStep.isValid())
            {
                for (VersionInformation version : versions)
                {
                    workflowHistory.addValidatedVersion(version.getVersionName());
                }
            }
        }

        workflowHistory.addStep(historyStep);
    }

    private DescriptionStep _getDescription(WorkflowDescriptor workflowDescriptor, int stepId)
    {
        DescriptionStep description = new DescriptionStep();

        if (stepId != 0)
        {
            StepDescriptor stepDescriptor = workflowDescriptor.getStep(stepId);
            if (stepDescriptor != null)
            {
                Map<String, Object> metaAttributes = stepDescriptor.getMetaAttributes();
                String stepStatus = (String) metaAttributes.getOrDefault("status", String.valueOf(stepId));
                description.setStatus(stepStatus);
            }
            else
            {
                description.setStatus(String.valueOf(stepId));
            }
            
            I18nizableText workflowStepName = stepDescriptor == null ? __MESSAGE_NO_STEP : new I18nizableText("application", stepDescriptor.getName());
            description.setDescription(workflowStepName);
            if ("application".equals(workflowStepName.getCatalogue()))
            {
                description.setSmallIcon("/plugins/cms/resources_workflow/" + workflowStepName.getKey() + "-small.png");
                description.setMediumIcon("/plugins/cms/resources_workflow/" + workflowStepName.getKey() + "-medium.png");
                description.setLargeIcon("/plugins/cms/resources_workflow/" + workflowStepName.getKey() + "-large.png");
            }
            else
            {
                String pluginName = workflowStepName.getCatalogue().substring("plugin.".length());
                description.setSmallIcon("/plugins/" + pluginName + "/resources/img/workflow/" + workflowStepName.getKey() + "-small.png");
                description.setMediumIcon("/plugins/" + pluginName + "/resources/img/workflow/" + workflowStepName.getKey() + "-medium.png");
                description.setLargeIcon("/plugins/" + pluginName + "/resources/img/workflow/" + workflowStepName.getKey() + "-large.png");
            }
        }
        else
        {
            description.setDescription(new I18nizableText("?"));
            description.setSmallIcon("/plugins/cms/resources/img/history/workflow/step_0_16.png");
            description.setMediumIcon("/plugins/cms/resources/img/history/workflow/step_0_32.png");
        }

        return description;
    }

    private WorkflowAction getWorkfowAction(UserIdentity caller, String title, WorkflowDescriptor workflowDescriptor, int actionId)
    {
        WorkflowAction action = new WorkflowAction(actionId);

        Map<String, I18nizableTextParameter> i18nParams = new HashMap<>();
        String userFullName = _userHelper.getUserFullName(caller);
        i18nParams.put("user", StringUtils.isNotEmpty(userFullName) ? new I18nizableText(_userHelper.getUserFullName(caller)) : new I18nizableText("plugin.cms", "PLUGINS_CMS_UITOOL_HISTORY_UNKNOWN_USER"));
        i18nParams.put("title", new I18nizableText(title));

        try
        {
            ActionDescriptor actionDescriptor = workflowDescriptor.getAction(actionId);
            
            I18nizableText workflowActionLabel = actionDescriptor == null ? __MESSAGE_NO_ACTION : new I18nizableText("application", actionDescriptor.getName() + "_ACTION_DESCRIPTION", i18nParams);
            action.setLabel(workflowActionLabel);
            
            I18nizableText workflowActionName = actionDescriptor == null ? __MESSAGE_NO_ACTION : new I18nizableText("application", actionDescriptor.getName());
            if ("application".equals(workflowActionName.getCatalogue()))
            {
                action.setSmallIcon("/plugins/cms/resources_workflow/" + workflowActionName.getKey() + "-small.png");
                action.setMediumIcon("/plugins/cms/resources_workflow/" + workflowActionName.getKey() + "-medium.png");
                action.setLargeIcon("/plugins/cms/resources_workflow/" + workflowActionName.getKey() + "-large.png");
            }
            else
            {
                String pluginName = workflowActionName.getCatalogue().substring("plugin.".length());
                action.setSmallIcon("/plugins/" + pluginName + "/resources/img/workflow/" + workflowActionName.getKey() + "-small.png");
                action.setMediumIcon("/plugins/" + pluginName + "/resources/img/workflow/" + workflowActionName.getKey() + "-medium.png");
                action.setLargeIcon("/plugins/" + pluginName + "/resources/img/workflow/" + workflowActionName.getKey() + "-large.png");
            }
        } 
        catch (Exception e) 
        {
            action.setLabel(__MESSAGE_NO_ACTION);
            action.setSmallIcon("/plugins/cms/resources/img/history/workflow/action_0_16.png");
            action.setMediumIcon("/plugins/cms/resources/img/history/workflow/action_0_32.png");
            action.setLargeIcon("/plugins/cms/resources/img/history/workflow/action_0_32.png");
        }

        return action;
    }

    private Date _getDate(Date startDate, Date finishDate, Date actionFinishDate)
    {
        if (actionFinishDate != null)
        {
            return startDate;
        }
        else
        {
            return finishDate;
        }
    }

    private List<VersionInformation> _getVersionsBetween(List<VersionInformation> versionsInformation, Date startDate, Date finishDate) throws RepositoryException
    {
        List<VersionInformation> versionsInsideStep = new ArrayList<>();

        for (VersionInformation versionInformation : versionsInformation)
        {
            Date versionCreationDate = versionInformation.getCreatedAt();

            if (startDate != null)
            {
                if (versionCreationDate.after(startDate))
                {
                    if (finishDate == null || versionCreationDate.before(finishDate))
                    {
                        versionsInsideStep.add(versionInformation);
                    }
                }
            }
            else
            {
                if (finishDate == null || versionCreationDate.before(finishDate))
                {
                    versionsInsideStep.add(versionInformation);
                }
            }
        }

        // If there is no version created inside the step,
        // retrieve the last one just before the step
        if (versionsInsideStep.isEmpty())
        {
            VersionInformation lastVersionBeforeStep = null;

            for (VersionInformation versionInformation : versionsInformation)
            {
                Date versionCreationDate = versionInformation.getCreatedAt();

                if (startDate != null && versionCreationDate.before(startDate))
                {
                    lastVersionBeforeStep = versionInformation;
                    break;
                }
                else if (startDate == null)
                {
                    // Use the first version
                    lastVersionBeforeStep = versionInformation;
                }
            }

            if (lastVersionBeforeStep != null)
            {
                versionsInsideStep.add(lastVersionBeforeStep);
            }
        }

        // if there is still no version for this step, then it is version 1
        // (case of migrations)
        if (versionsInsideStep.isEmpty())
        {
            versionsInsideStep.add(new VersionInformation("1"));
        }

        return versionsInsideStep;
    }

    /**
     * History step 2 JSON
     * @param historyStep the history step
     * @return the JSON map of the history step
     * @throws RepositoryException if a repository error occurred
     */
    public Map<String, Object> historyStep2Json(HistoryStep historyStep) throws RepositoryException
    {
        Map<String, Object> step2json = new HashMap<>();

        step2json.put("id", historyStep.getId());
        step2json.put("current", historyStep.isCurrent());

        step2json.putAll(_description2Json(historyStep.getDescription()));
        step2json.putAll(_workfowAction2Json(historyStep.getWorkflowAction()));
        
        Optional<UserIdentity> caller = historyStep.getCaller();
        if (caller.isPresent())
        {
            step2json.put("caller", _userHelper.user2json(caller.get()));
        }
        
        Optional<String> comments = historyStep.getComments();
        if (comments.isPresent())
        {
            step2json.put("comment", comments.get());
        }
        
        step2json.put("validation", historyStep.isValid());
        step2json.put("date", DateUtils.dateToString(historyStep.getDate()));
        
        Optional<List<VersionInformation>> versions = historyStep.getVersions();
        if (versions.isPresent())
        {
            step2json.put("versions", _versionsBetween2Json(versions.get()));
        }
        
        return step2json;
    }
    
    private Map<String, Object> _description2Json(DescriptionStep description)
    {
        Map<String, Object> desc2json = new HashMap<>();

        desc2json.put("description", description.getDescription());
        
        Optional<String> smallIcon = description.getSmallIcon();
        if (smallIcon.isPresent())
        {
            desc2json.put("iconSmall", smallIcon.get());
        }

        Optional<String> mediumIcon = description.getMediumIcon();
        if (mediumIcon.isPresent())
        {
            desc2json.put("iconMedium", mediumIcon.get());
        }
        
        Optional<String> largeIcon = description.getLargeIcon();
        if (largeIcon.isPresent())
        {
            desc2json.put("iconLarge", largeIcon.get());
        }

        return desc2json;
    }
    
    private Map<String, Object> _workfowAction2Json(WorkflowAction action)
    {
        Map<String, Object> action2json = new HashMap<>();

        action2json.put("actionId", action.getId());
        action2json.put("actionLabel", action.getLabel());
        
        Optional<String> smallIcon = action.getSmallIcon();
        if (smallIcon.isPresent())
        {
            action2json.put("actionIconSmall", smallIcon.get());
        }

        Optional<String> mediumIcon = action.getMediumIcon();
        if (mediumIcon.isPresent())
        {
            action2json.put("actionIconMedium", mediumIcon.get());
        }
        
        Optional<String> largeIcon = action.getLargeIcon();
        if (largeIcon.isPresent())
        {
            action2json.put("actionIconLarge", largeIcon.get());
        }

        return action2json;
    }
    
    private List<Map<String, Object>> _versionsBetween2Json(List<VersionInformation> versionsInformation) throws RepositoryException
    {
        List<Map<String, Object>> versions = new ArrayList<>();

        for (VersionInformation versionInformation : versionsInformation)
        {
            Map<String, Object> version = new HashMap<>();

            String versionName = versionInformation.getVersionName();
            version.put("label", versionInformation.getLabels());
            version.put("name", versionName);
            if (StringUtils.isNotEmpty(versionInformation.getVersionRawName()))
            {
                version.put("rawName", versionInformation.getVersionRawName());
            }

            version.put("createdAt", DateUtils.dateToString(versionInformation.getCreatedAt()));

            versions.add(version);
        }
        return versions;
    }
}
