/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.workflow.purge;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.time.ZoneId;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.source.SourceUtil;
import org.quartz.JobExecutionContext;

import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ContentQueryHelper;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.core.schedule.progression.ContainerProgressionTracker;
import org.ametys.core.ui.mail.StandardMailBodyHelper;
import org.ametys.core.ui.mail.StandardMailBodyHelper.MailBodyBuilder;
import org.ametys.core.util.DateUtils;
import org.ametys.core.util.I18nUtils;
import org.ametys.core.util.mail.SendMailHelper;
import org.ametys.plugins.core.impl.schedule.AbstractStaticSchedulable;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.workflow.support.WorkflowProvider;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;

import jakarta.mail.MessagingException;

/**
 * Schedulable that removes old versions of contents.
 */
public class PurgeContentsSchedulable extends AbstractStaticSchedulable implements Initializable
{
    /** The server base URL. */
    protected String _baseUrl;
    
    /** The ametys object resolver. */
    protected AmetysObjectResolver _ametysResolver;
    
    /** The avalon source resolver. */
    protected SourceResolver _sourceResolver;
    
    /** The version purger. */
    protected PurgeVersionsManager _versionPurger;
    
    /** The workflow provider */
    protected WorkflowProvider _workflowProvider;
    
    /** The i18n utils. */
    protected I18nUtils _i18nUtils;
    
    /** A Map of the validation step ID by workflow name. */
    protected Map<String, Long> _validationStepId;
    
    /** The count of oldest versions to keep. */
    protected int _firstVersionsToKeep;
    
    /** The content of "from" field in emails. */
    protected String _mailFrom;
    
    /** The sysadmin mail address, to which will be sent the report e-mail. */
    protected String _sysadminMail;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        // Lookup the needed components.
        _ametysResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _versionPurger = (PurgeVersionsManager) manager.lookup(PurgeVersionsManager.ROLE);
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _workflowProvider = (WorkflowProvider) manager.lookup(WorkflowProvider.ROLE);
        
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
    }
    
    public void initialize() throws Exception
    {
        
        _baseUrl = StringUtils.stripEnd(StringUtils.removeEndIgnoreCase(Config.getInstance().getValue("cms.url"), "index.html"), "/");
        _mailFrom = Config.getInstance().getValue("smtp.mail.from");
        _sysadminMail = Config.getInstance().getValue("smtp.mail.sysadminto");
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        super.configure(configuration);
        _validationStepId = configureValidationStepId(configuration);
        _firstVersionsToKeep = configuration.getChild("firstVersionsToKeep").getValueAsInteger();
    }
    
    /**
     * Get the validation step ID by workflow from the component configuration.
     * @param configuration the component configuration.
     * @return a Map of the validation step ID by workflow.
     * @throws ConfigurationException If an error occurred
     */
    protected Map<String, Long> configureValidationStepId(Configuration configuration) throws ConfigurationException
    {
        Map<String, Long> validationStepIds = new HashMap<>();
        
        for (Configuration workflowConf : configuration.getChildren("workflow"))
        {
            String workflowName = workflowConf.getAttribute("name");
            Long validationStepId = workflowConf.getChild("validationStepId").getValueAsLong();
            
            validationStepIds.put(workflowName, validationStepId);
        }
        
        return validationStepIds;
    }

    @Override
    public void execute(JobExecutionContext context, ContainerProgressionTracker progressionTracker) throws Exception
    {
        String query = ContentQueryHelper.getContentXPathQuery(null);
        
        Date startDate = new Date();
        
        int totalContentsPurged = 0;
        int totalVersionsPurged = 0;
        
        try
        {
            try (AmetysObjectIterable<Content> contents = _ametysResolver.query(query))
            {
                for (Content content : contents)
                {
                    if (content instanceof WorkflowAwareContent)
                    {
                        WorkflowAwareContent workflowContent = (WorkflowAwareContent) content;
                        long workflowId = workflowContent.getWorkflowId();
                        
                        AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(workflowContent);
                        String workflowName = workflow.getWorkflowName(workflowId);
                        
                        if (_validationStepId.containsKey(workflowName))
                        {
                            long validationStepId = _validationStepId.get(workflowName);
                            
                            int purged = _versionPurger.purgeContent(workflowContent, validationStepId, _firstVersionsToKeep);
                            if (purged > 0)
                            {
                                totalVersionsPurged += purged;
                                totalContentsPurged++;
                            }
                        }
                    }
                }
            }
            
            Date endDate = new Date();
            
            sendMail(startDate, endDate, totalContentsPurged, totalVersionsPurged);
        }
        catch (Exception e)
        {
            sendErrorMail(startDate, e);
        }
        
    }
    
    /**
     * Send the purge report e-mail.
     * @param startDate the purge start date.
     * @param endDate the purge end date.
     * @param totalContentsPurged the total count of contents of which versions were purged.
     * @param totalVersionsPurged the total count of content versions removed.
     */
    protected void sendMail(Date startDate, Date endDate, int totalContentsPurged, int totalVersionsPurged)
    {
        try
        {
            Map<String, String> params = getEmailParams(startDate, endDate, totalContentsPurged, totalVersionsPurged);
            
            I18nizableText i18nSubject = new I18nizableText("plugin.cms", "PLUGINS_CMS_PURGE_CONTENTS_REPORT_SUBJECT");
            
            String subject = _i18nUtils.translate(i18nSubject);
            String body = getMailBody(params);
            
            if (StringUtils.isNotBlank(_sysadminMail))
            {
                SendMailHelper.newMail()
                              .withSubject(subject)
                              .withHTMLBody(body)
                              .withSender(_mailFrom)
                              .withRecipient(_sysadminMail)
                              .sendMail();
            }
        }
        catch (MessagingException | IOException e)
        {
            getLogger().warn("Error sending the purge report e-mail.", e);
        }
    }
    
    /**
     * Send the error e-mail.
     * @param startDate the purge start date.
     * @param throwable the error.
     */
    protected void sendErrorMail(Date startDate, Throwable throwable)
    {
        try
        {
            Map<String, String> params = getErrorEmailParams(startDate, throwable);
            
            I18nizableText i18nSubject = new I18nizableText("plugin.cms", "PLUGINS_CMS_PURGE_CONTENTS_REPORT_ERROR_SUBJECT");
            
            String subject = _i18nUtils.translate(i18nSubject);
            String mailUri = getErrorMailUri(params);
            String body = getMailBody(mailUri, params, throwable);
            
            if (StringUtils.isNotBlank(_sysadminMail))
            {
                SendMailHelper.newMail()
                              .withSubject(subject)
                              .withHTMLBody(body)
                              .withSender(_mailFrom)
                              .withRecipient(_sysadminMail)
                              .sendMail();
            }
        }
        catch (MessagingException | IOException e)
        {
            getLogger().warn("Error sending the purge error e-mail.", e);
        }
    }
    
    /**
     * Get the report e-mail parameters.
     * @param startDate the purge start date.
     * @param endDate the purge end date.
     * @param totalContentsPurged the total count of contents of which versions were purged.
     * @param totalVersionsPurged the total count of content versions removed.
     * @return the e-mail parameters.
     */
    protected Map<String, String> getEmailParams(Date startDate, Date endDate, int totalContentsPurged, int totalVersionsPurged)
    {
        Map<String, String> params = new HashMap<>();
        
        params.put("startDate", DateUtils.getISODateTimeFormatter().format(startDate.toInstant().atZone(ZoneId.systemDefault())));
        params.put("endDate", DateUtils.getISODateTimeFormatter().format(endDate.toInstant().atZone(ZoneId.systemDefault())));
        params.put("totalContentsPurged", Integer.toString(totalContentsPurged));
        params.put("totalVersionsPurged", Integer.toString(totalVersionsPurged));
        params.put("url", _baseUrl);
        
        return params;
    }
    
    /**
     * Get the error e-mail parameters.
     * @param startDate the purge start date.
     * @param throwable the error.
     * @return the e-mail parameters.
     */
    protected Map<String, String> getErrorEmailParams(Date startDate, Throwable throwable)
    {
        Map<String, String> params = new HashMap<>();
        
        params.put("startDate", DateUtils.getISODateTimeFormatter().format(startDate.toInstant().atZone(ZoneId.systemDefault())));
        params.put("url", _baseUrl);
        
        return params;
    }
    
    /**
     * Get a mail part.
     * @param parameters the pipeline parameters.
     * @return the mail part.
     * @throws IOException If an error occurred
     */
    protected String getMailBody(Map<String, String> parameters) throws IOException
    {
        String uri = getMailUri(parameters);
        
        return getMailBody(uri, parameters, null);
    }
    
    /**
     * Get a mail part.
     * @param uri The url where to get the body of the mail
     * @param parameters the pipeline parameters.
     * @return the mail part.
     * @throws IOException If an error occurred
     */
    private String getMailBody(String uri, Map<String, String> parameters, Throwable throwable) throws IOException
    {
        Source source = null;
        InputStream is = null;
        try
        {
            source = _sourceResolver.resolveURI(uri, null, parameters);
            is = source.getInputStream();
            
            try (ByteArrayOutputStream bos = new ByteArrayOutputStream())
            {
                SourceUtil.copy(is, bos);
                
                String message = bos.toString("UTF-8");
                MailBodyBuilder bodyBuilder = StandardMailBodyHelper.newHTMLBody()
                    .withTitle(new I18nizableText("plugin.cms", throwable != null ? "PLUGINS_CMS_PURGE_CONTENTS_REPORT_ERROR_SUBJECT" : "PLUGINS_CMS_PURGE_CONTENTS_REPORT_SUBJECT"))
                    .withMessage(message)
                    .withLink(_baseUrl, new I18nizableText("plugin.cms", "PLUGINS_CMS_PURGE_CONTENTS_REPORT_BODY_LINK"));
                
                if (throwable != null)
                {
                    String stackTrace = ExceptionUtils.getStackTrace(throwable);
                    bodyBuilder.withDetails(new I18nizableText("plugin.cms", "PLUGINS_CMS_PURGE_CONTENTS_REPORT_BODY_ERROR_TITLE"), stackTrace, true);
                }
                
                return bodyBuilder.build();
            }
        }
        finally
        {
            if (is != null)
            {
                is.close();
            }
            
            if (source != null)
            {
                _sourceResolver.release(source);
            }
        }
    }
    
    /**
     * Get the pipeline uri for mail body
     * @param parameters the mail parameters
     * @return a pipeline uri 
     */
    protected String getMailUri(Map<String, String> parameters)
    {
        return "cocoon://_plugins/cms/purge/purge-mail.html";
    }
    
    /**
     * Get the pipeline uri for error mail body.
     * @param parameters the mail parameters
     * @return a pipeline uri 
     */
    protected String getErrorMailUri(Map<String, String> parameters)
    {
        return "cocoon://_plugins/cms/purge/purge-error-mail.html";
    }
    
}
