/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.test.framework;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.net.ServerSocket;
import java.time.LocalDate;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.SystemUtils;
import org.apache.solr.client.solrj.SolrClient;
import org.apache.solr.client.solrj.SolrServerException;
import org.apache.solr.common.SolrException;
import org.apache.solr.common.SolrInputDocument;
import org.junit.Assert;

import org.ametys.cms.content.archive.ArchiveConstants;
import org.ametys.cms.content.indexing.solr.SolrFieldNames;
import org.ametys.cms.content.indexing.solr.SolrIndexer;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.data.Binary;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.Reference;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableDefaultContent;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.cms.rights.Content2ContentTypeRightContextConvertor;
import org.ametys.cms.search.SearchResults;
import org.ametys.cms.search.solr.SolrClientProvider;
import org.ametys.cms.search.solr.field.WorkflowStepSearchField;
import org.ametys.cms.workflow.AbstractContentWorkflowComponent;
import org.ametys.cms.workflow.ContentWorkflowHelper;
import org.ametys.cms.workflow.EditContentFunction;
import org.ametys.core.right.Profile;
import org.ametys.core.right.ProfileAssignmentStorageExtensionPoint;
import org.ametys.core.right.RightProfilesDAO;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.MapUtils;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeater;
import org.ametys.plugins.repository.query.QueryHelper;
import org.ametys.plugins.workflow.AbstractWorkflowComponent;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.test.framework.AbstractGroupDirectoriesAwareTestCase;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;
import org.ametys.solr.test.SolrServerHelper;

import com.opensymphony.workflow.WorkflowException;
import com.sun.net.httpserver.HttpServer;

/**
 * Abstract test case for search.
 */
@NeedFeatures({
    "cms/org.ametys.cms.content.indexing.solr.init",
    "cms/org.ametys.cms.content.search.solr.client",
    "repository/resolver",
    "cms/workflow",
    "cms/workflow.reference.table",
    "cms/workflow.content",
    "cms/org.ametys.cms.content.indexing.solr",
    "cms/org.ametys.cms.data.holder.helpers",
    "core/model.helper",
    "core/runtime.rights.profiles.dao",
    "cms/right.contextconvertor.contenttype",
    "cms/cms.rights.accesscontroller.content-type",
    "cms/cms.rights.storage.assignments.impl.jdbc"
})
public abstract class AbstractSearcherTestCase extends AbstractGroupDirectoriesAwareTestCase implements WithCmsFeaturesTestCase
{
    /** The exhaustive content type. */
    public static final String EXHAUSTIVE_CONTENTTYPE = "org.ametys.test.Content.exhaustive";
    
    /** Test content type id */
    public static final String MULTILINGUAL_CONTENTTYPE = "org.ametys.test.Content.multilingual";
    
    /** The solr app directory. */
    protected static final String SOLR_APP_DIR = "test/environments/solr-app";
    
    /** The solr update client */
    protected static SolrClient _updateSolr;
    /** The ametys object resolver. */
    protected static AmetysObjectResolver _ametysObjectResolver;
    /** The content workflow helper. */
    protected static ContentWorkflowHelper _contentWorkflowHelper;
    /** The solr indexer. */
    protected static SolrIndexer _solrIndexer;
    /** The profile assignment */
    protected static ProfileAssignmentStorageExtensionPoint _profileAssignmentStorageEP;
    /** The profiles DAO */
    protected static RightProfilesDAO _profilesDAO;
    /** The Solr app directory */
    protected static File _solrAppDir;
    /** The Ametys mocker server */
    protected static HttpServer _httpServer;
    private int _solrPort;
    private int _httpServerPort;
    
    @Override
    public TestConfigSet getConfigSet()
    {
        return WithCmsFeaturesTestCase.super.getConfigSet();
    }
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        // Need to start Solr BEFORE Cocoon
        _startSolr();
        super.essentialsBefore();
    }
    
    private void _startSolr() throws Exception
    {
        _solrPort = _getFreePort();
        // Generates a unique solr app directory
        _solrAppDir = new File(SOLR_APP_DIR + "-" + _solrPort);
        _solrAppDir.mkdirs();
        _logger.info("Will start Solr server on port {}", _solrPort);
        SolrServerHelper.startupSolrServer(_solrAppDir, getSolrEnvFile(), _solrPort);
        ObserverHelper.suspendObservationForIndexation();
    }
    
    /**
     * The Solr environment file which will be copied. Override when running tests in local for instance
     * @return The Solr environment file.
     */
    protected File getSolrEnvFile()
    {
        File file = new File("test/environments/solr.env." + (SystemUtils.IS_OS_UNIX ? "sh" : "cmd"));
        return file.exists() ? file : null;
    }
    
    @Override
    public Map<String, String> variableValues()
    {
        _httpServerPort = _getFreePort();
        
        return MapUtils.merge(
                super.variableValues(),
                Map.of(
                    "cmsPort", String.valueOf(_httpServerPort), 
                    "solrPort", String.valueOf(_solrPort)
                )
        );
    }
    
    private static int _getFreePort()
    {
        try (ServerSocket s = new ServerSocket(0);)
        {
            return s.getLocalPort();
        }
        catch (IOException e)
        {
            throw new RuntimeException("Unable to find a free port", e);
        }
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        SolrClientProvider solrClientProvider = (SolrClientProvider) manager.lookup(SolrClientProvider.ROLE);
        _updateSolr = solrClientProvider.getUpdateClient(RepositoryConstants.DEFAULT_WORKSPACE, false);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentWorkflowHelper = (ContentWorkflowHelper) manager.lookup(ContentWorkflowHelper.ROLE);
        _solrIndexer = (SolrIndexer) manager.lookup(SolrIndexer.ROLE);
        _profileAssignmentStorageEP = (ProfileAssignmentStorageExtensionPoint) manager.lookup(ProfileAssignmentStorageExtensionPoint.ROLE);
        _profilesDAO = (RightProfilesDAO) manager.lookup(RightProfilesDAO.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        _logger.info("Will mock a HTTP server on port {}", _httpServerPort);
        _httpServer = mockServer(_httpServerPort);
        _sendSchema();
        _assignRightsToUser();
    }
    
    private static void _sendSchema() throws IOException, SolrServerException
    {
        if (_solrIndexer != null)
        {
            _solrIndexer.sendSchema();
        }
    }
    
    @Override
    protected void essentialsAfter() throws Exception
    {
        _deleteSolrCores();
        super.essentialsAfter();
        _stopHttpServer();
        _shutdownSolr();
    }
    
    private static void _deleteSolrCores() throws Exception
    {
        if (_solrIndexer != null)
        {
            // Remove solr cores
            try
            {
                _solrIndexer.deleteCore(RepositoryConstants.DEFAULT_WORKSPACE);
                _solrIndexer.deleteCore(ArchiveConstants.ARCHIVE_WORKSPACE);
            }
            catch (SolrException e)
            {
                _staticLogger.error("Cannot deleting cores.", e);
            }
        }
    }
    
    private static void _stopHttpServer()
    {
        if (_httpServer != null)
        {
            _httpServer.stop(0);
        }
    }
    
    private void _shutdownSolr() throws Exception
    {
        if (_solrAppDir != null)
        {
            SolrServerHelper.shutDownSolrServer(_solrAppDir, getSolrEnvFile(), _solrPort);
            FileUtils.deleteQuietly(_solrAppDir);
        }
    }

    /**
     * Assign content rights to the current user.
     */
    private static void _assignRightsToUser()
    {
        Profile profile = _profilesDAO.addProfile("profile-test");
        _profilesDAO.addRight(profile, "Workflow_Rights_Edition_Online");
        _profilesDAO.addRight(profile, "Workflow_Rights_Propose");
        _profilesDAO.addRight(profile, "Workflow_Rights_Validate");
        _profileAssignmentStorageEP.allowProfileToUser(CURRENT_USER_TEST_IDENTITY, profile.getId(), Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX);
    }
    
    /**
     * Populate the repository with test data and index all the contents.
     * @throws Exception If an error occurs.
     */
    protected static void populateRepositoryAndIndex() throws Exception
    {
        ModifiableDefaultContent content = createContent("Test content", "content", EXHAUSTIVE_CONTENTTYPE);
        
        LocalDate date = LocalDate.of(2007, 10, 9);
        ZonedDateTime dateTime = ZonedDateTime.of(2011, 11, 24, 10, 15, 30, 0, ZoneId.systemDefault());
        
        // Set content metadata.
        content.setValue("abstract", "Test abstract");
        content.setValue("string", "test string 2");
        content.setValue("stringmult", new String[] {"one", "two", "three"});
        content.setValue("long", 3L);
        content.setValue("longmult", new Long[] {5L, 6L, 12L});
        content.setValue("double", 1.234D);
        content.setValue("doublemult", new Double[] {3.14D, 6.12D});
        content.setValue("boolean", true);
        content.setValue("date", date);
        content.setValue("datetime", dateTime);
        ModifiableModelAwareComposite address = content.getComposite("address", true);
        address.setValue("zipcode", "31590");
        address.setValue("country", "france");
        
        Binary binaryMeta = new Binary();
        binaryMeta.setFilename("test.txt");
        binaryMeta.setMimeType("text/plain");
        binaryMeta.setEncoding("UTF-8");
        binaryMeta.setLastModificationDate(dateTime);
        String textContent = "Text content";
        binaryMeta.setInputStream(new ByteArrayInputStream(textContent.getBytes("UTF-8")));
        content.setValue("file", binaryMeta);
        
        content.setValue("user", AbstractPopulationAwareTestCase.ANONYMOUS_TEST_IDENTITY);
        content.setValue("usermult", new UserIdentity[] {AbstractPopulationAwareTestCase.USER1_TEST_IDENTITY, AbstractPopulationAwareTestCase.USER2_TEST_IDENTITY, AbstractPopulationAwareTestCase.USER3_TEST_IDENTITY});
        
        content.setValue("reference", new Reference("http://www.ametys.org"));
        content.setValue("gps", new Geocode(44.481, 3.208));
        
        // Contents to test references.
        ModifiableDefaultContent link1 = createContent("Linked content 1", "linked-content-1", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link2 = createContent("Linked content 2", "linked-content-2", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link3 = createContent("Linked content 3", "linked-content-3", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link4 = createContent("Linked content 4", "linked-content-4", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link11 = createContent("Linked content 11", "linked-content-11", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link21 = createContent("Linked content 21", "linked-content-21", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link22 = createContent("Linked content 22", "linked-content-22", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link31 = createContent("Linked content 31", "linked-content-31", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link221 = createContent("Linked content 221", "linked-content-221", EXHAUSTIVE_CONTENTTYPE);
        
        content.setValue("linkContent", link1);
        content.setValue("linkContents", new Content[] {link2, link3, link4});
        content.setValue("owns", new Content[] {link2, link3, link4});
        
        ModifiableModelAwareRepeater repeaterBase = content.getComposite("composite", true).getRepeater("repeater", true);
        repeaterBase.addEntry();
        repeaterBase.getEntry(1).setValue("comp-string", "entry 1");
        repeaterBase.getEntry(1).setValue("comp-owns", link1);
        repeaterBase.getEntry(1).getComposite("subcomposite", true).setValue("subcomp-str", "subcomposite in entry 1");
        repeaterBase.addEntry();
        repeaterBase.getEntry(2).setValue("comp-string", "entry 2");
        repeaterBase.getEntry(2).setValue("comp-owns", link2);
        repeaterBase.getEntry(2).getComposite("subcomposite", true).setValue("subcomp-str", "subcomposite in entry 2");
        
        ModifiableModelAwareRepeater repeaterThreeLevels = content.getRepeater("repeaterThreeLevels", true);
        
        repeaterThreeLevels.addEntry();
        repeaterThreeLevels.getEntry(1).setValue("rep-string", "repeater three levels entry 1");
        ModifiableModelAwareRepeater repeater2_1 = repeaterThreeLevels.getEntry(1).getRepeater("repeater2", true);
        ModifiableModelAwareRepeater repeater3_11 = repeater2_1.addEntry().getRepeater("repeater3", true);
        repeater3_11.addEntry().setValue("rep3-string", "111");
        repeater3_11.addEntry().setValue("rep3-string", "112");
        
        repeaterThreeLevels.addEntry();
        repeaterThreeLevels.getEntry(2).setValue("rep-string", "repeater three levels entry 2");
        ModifiableModelAwareRepeater repeater2_2 = repeaterThreeLevels.getEntry(2).getRepeater("repeater2", true);
        ModifiableModelAwareRepeater repeater3_21 = repeater2_2.addEntry().getRepeater("repeater3", true);
        repeater3_21.addEntry().setValue("rep3-string", "211");
        ModifiableModelAwareRepeater repeater3_22 = repeater2_2.addEntry().getRepeater("repeater3", true);
        repeater3_22.addEntry().setValue("rep3-string", "221");
        repeater3_22.addEntry().setValue("rep3-string", "222");
        repeater3_22.addEntry().setValue("rep3-string", "223");
        
        // Edit and validate the main content.
        content.saveChanges();
        editContent(content);
        doWorkflowAction(content, 4);
        
        // Contents to test sorting.
        ModifiableDefaultContent aaa = createContent("AAA", "aaa", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent bbb = createContent("BBB", "bbb", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent ccc = createContent("CCC", "ccc", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent ddd = createContent("DDD", "ddd", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent eee = createContent("EEE", "eee", EXHAUSTIVE_CONTENTTYPE);
        createContent("FFF", "fff", EXHAUSTIVE_CONTENTTYPE);
        createContent("GGG", "ggg", EXHAUSTIVE_CONTENTTYPE);
        createContent("XXX", "xxx", EXHAUSTIVE_CONTENTTYPE);
        createContent("YYY", "yyy", EXHAUSTIVE_CONTENTTYPE);
        createContent("ZZZ", "zzz", EXHAUSTIVE_CONTENTTYPE);
        
        ModifiableDefaultContent proposedA = createContent("Proposed content AAA", "proposed-aaa", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent proposedB = createContent("Proposed content BBB", "proposed-bbb", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent validatedC = createContent("Validated content CCC", "validated-ccc", EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent validatedD = createContent("Validated content DDD", "validated-ddd", EXHAUSTIVE_CONTENTTYPE);
        
        proposedA.getComposite("address", true).setValue("country", "spain");
        proposedB.getComposite("address", true).setValue("country", "france");
        validatedC.getComposite("address", true).setValue("country", "japan");
        validatedD.getComposite("address", true).setValue("country", "belgium");

        proposedA.setValue("string", "test string 1");
        proposedB.setValue("string", "test string 2");
        validatedC.setValue("string", "test string 3");
        validatedD.setValue("string", "test string 4");
        
        // Edit and trigger the workflow of the validated contents.
        editContent(proposedA);
        editContent(proposedB);
        editContent(validatedC);
        editContent(validatedD);
        doWorkflowAction(proposedA, 3);
        doWorkflowAction(proposedB, 3);
        doWorkflowAction(validatedC, 4);
        doWorkflowAction(validatedD, 4);
        
        // Set some metas on the linked contents.
        link1.getComposite("address", true).setValue("country", "france");
        link2.getComposite("address", true).setValue("country", "japan");
        link3.getComposite("address", true).setValue("country", "spain");
        link4.getComposite("address", true).setValue("country", "japan");
        
        link11.getComposite("address", true).setValue("country", "spain");
        link21.getComposite("address", true).setValue("country", "france");
        link22.getComposite("address", true).setValue("country", "japan");
        link31.getComposite("address", true).setValue("country", "italy");
        link221.getComposite("address", true).setValue("country", "belgium");
        
        // Use "abstract" metadata to sort on.
        link1.setValue("abstract", "aaaa");
        link2.setValue("abstract", "bbbb");
        link3.setValue("abstract", "cccc");
        link4.setValue("abstract", "dddd");
        link11.setValue("abstract", "eeee");
        link21.setValue("abstract", "ffff");
        link22.setValue("abstract", "gggg");
        link31.setValue("abstract", "hhhh");
        link221.setValue("abstract", "iiii");
        
        // link1 -> link11
        link1.setValue("linkContent", link11);
        // link2 -> link21, link22
        link2.setValue("linkContents", new Content[] {link21, link22});
        // link3 -> link31
        link3.setValue("linkContent", link31);
        // link3 -> link1
        link3.setValue("linkContents", new Content[] {link1});
        // link22 -> link221
        link22.setValue("linkContent", link221);
        // Same links with "owns" metadata.
        link1.setValue("owns", new Content[] {link11});
        link2.setValue("owns", new Content[] {link21, link22});
        link3.setValue("owns", new Content[] {link31});
        link22.setValue("owns", new Content[] {link221});
        
        
        ModifiableModelAwareRepeater repeater21 = link21.getComposite("composite", true).getRepeater("repeater", true);
        repeater21.addEntry();
        repeater21.getEntry(1).setValue("comp-string", "linked content 21 - entry 1");
        repeater21.getEntry(1).setValue("comp-owns", eee);
        repeater21.addEntry();
        repeater21.getEntry(2).setValue("comp-string", "linked content 21 - entry 2");
        repeater21.getEntry(2).setValue("comp-owns", ccc);

        ModifiableModelAwareRepeater repeater22 = link22.getComposite("composite", true).getRepeater("repeater", true);
        repeater22.addEntry();
        repeater22.getEntry(1).setValue("comp-string", "linked content 22 - entry 1");
        repeater22.getEntry(1).setValue("comp-owns", link221);
        repeater22.addEntry();
        repeater22.getEntry(2).setValue("comp-string", "linked content 22 - entry 2");
        repeater22.getEntry(2).setValue("comp-owns", aaa);

        ModifiableModelAwareRepeater repeater31 = link31.getComposite("composite", true).getRepeater("repeater", true);
        repeater31.addEntry();
        repeater31.getEntry(1).setValue("comp-string", "linked content 31 - entry 1");
        repeater31.getEntry(1).setValue("comp-owns", ddd);
        repeater31.addEntry();
        repeater31.getEntry(2).setValue("comp-string", "linked content 31 - entry 2");
        repeater31.getEntry(2).setValue("comp-owns", bbb);
        
        link1.saveChanges();
        link2.saveChanges();
        link3.saveChanges();
        link4.saveChanges();
        link11.saveChanges();
        link21.saveChanges();
        link22.saveChanges();
        link31.saveChanges();
        link221.saveChanges();

        link1.setValue("string", "test string 1");
        link2.setValue("string", "test string 2");
        link3.setValue("string", "test string 3");
        link4.setValue("string", "test string 4");
        
        // Edit and propose the four linked contents.
        editContent(link1);
        editContent(link2);
        editContent(link3);
        editContent(link4);
        doWorkflowAction(link1, 3);
        doWorkflowAction(link2, 3);
        doWorkflowAction(link3, 3);
        doWorkflowAction(link4, 3);
        
        Map<String, Object> indexResults = _solrIndexer.indexAllContents(RepositoryConstants.DEFAULT_WORKSPACE, false, _updateSolr);
        _solrIndexer.commit(RepositoryConstants.DEFAULT_WORKSPACE, _updateSolr);
        
        _staticLogger.info("Global indexing: success count={}, error count={}", indexResults.get("successCount"), indexResults.get("errorCount"));
    }
    
    /**
     * Add some fake documents to ensure they do not appear.
     * @throws Exception If an error occurs.
     */
    protected static void addFakeDocuments() throws Exception
    {
        SolrInputDocument fakeDoc1 = new SolrInputDocument();
        fakeDoc1.addField("id", "FAKE_DOC_1");
        fakeDoc1.addField(SolrFieldNames.DOCUMENT_TYPE, "FAKE_TYPE");
        fakeDoc1.addField("title", "Fake document number one");
        fakeDoc1.addField(WorkflowStepSearchField.NAME, 1);
        fakeDoc1.addField("abstract_s_sort", "ZZZ");
        fakeDoc1.addField("address/country_s", "japan");
        
        SolrInputDocument fakeDoc2 = new SolrInputDocument();
        fakeDoc2.addField("id", "FAKE_DOC_2");
        fakeDoc2.addField(SolrFieldNames.DOCUMENT_TYPE, "DUMMY_TYPE");
        fakeDoc2.addField("title", "Fake document number two");
        fakeDoc2.addField(WorkflowStepSearchField.NAME, 3);
        fakeDoc2.addField("abstract_s_sort", "XXX");
        fakeDoc2.addField("address/country_s", "france");
        
        String coreName = Config.getInstance().getValue("cms.solr.core.prefix") + RepositoryConstants.DEFAULT_WORKSPACE;
        _updateSolr.add(coreName, Arrays.asList(fakeDoc1, fakeDoc2));
        _updateSolr.commit(coreName);
    }
    
    /**
     * Create a content.
     * @param title The content title.
     * @param name The content name.
     * @param type The content type.
     * @return The created content.
     * @throws WorkflowException If an error occurs.
     */
    protected static ModifiableDefaultContent createContent(String title, String name, String type) throws WorkflowException
    {
        Map<String, Object> results = _contentWorkflowHelper.createContent("content", 1, name, title, new String[] {type}, new String[0], "en");
        
        return (ModifiableDefaultContent) results.get(Content.class.getName());
    }
    
    /**
     * Trigger a content edition (without data).
     * @param content The content to edit.
     * @return The worflow results.
     * @throws WorkflowException If an error occurs.
     */
    protected static Map<String, Object> editContent(WorkflowAwareContent content) throws WorkflowException
    {
        return editContent(content, Collections.emptyMap());
    }
    
    /**
     * Trigger a content edition.
     * @param content The content to edit.
     * @param values The content metadata values.
     * @return The worflow results.
     * @throws WorkflowException If an error occurs.
     */
    protected static Map<String, Object> editContent(WorkflowAwareContent content, Map<String, Object> values) throws WorkflowException
    {
        Map<String, Object> inputs = new HashMap<>();
        
        inputs.put(AbstractContentWorkflowComponent.CONTENT_KEY, content);
        inputs.put(AbstractWorkflowComponent.RESULT_MAP_KEY, new HashMap<>());
        inputs.put(AbstractWorkflowComponent.FAIL_CONDITIONS_KEY, new ArrayList<>());
        
        Map<String, Object> contextParameters = new HashMap<>();
        contextParameters.put(EditContentFunction.QUIT, Boolean.TRUE);
        contextParameters.put(EditContentFunction.FORM_RAW_VALUES, values);
        inputs.put(AbstractWorkflowComponent.CONTEXT_PARAMETERS_KEY, contextParameters);
        
        return _contentWorkflowHelper.doAction(content, 2, inputs);
    }
    
    /**
     * Trigger a workflow action on a content.
     * @param content The content.
     * @param actionId The action ID.
     * @return The workflow results.
     * @throws WorkflowException If an error occurs.
     */
    protected static Map<String, Object> doWorkflowAction(WorkflowAwareContent content, int actionId) throws WorkflowException
    {
        return _contentWorkflowHelper.doAction(content, actionId, new HashMap<>());
    }
    
    /**
     * Get a content by name.
     * @param name The content name.
     * @return The content.
     */
    protected static Content getContent(String name)
    {
        String query = QueryHelper.getXPathQuery(name, "ametys:content", null);
        AmetysObjectIterable<Content> contents = _ametysObjectResolver.query(query);
        return contents.iterator().next();
    }
    
    /**
     * Asserts that the facet value is the expected content.
     * @param expectedCount The expected content facet count.
     * @param results The search results.
     * @param facet The facet name
     * @param contentName The content name.
     */
    protected static void assertContentFacetValueCount(int expectedCount, SearchResults<?> results, String facet, String contentName)
    {
        Content content = getContent(contentName);
        assertFacetValueCount(expectedCount, results, facet, content.getId());
    }
    
    /**
     * Asserts the real size of an iterable: a returned iterator should return the exact number of objects.
     * @param expectedSize The expected size.
     * @param iterable The iterable to test.
     */
    protected static void assertIterableSize(int expectedSize, Iterable<?> iterable)
    {
        Iterator<?> iterator = iterable.iterator();
        
        for (int i = 0; i < expectedSize; i++)
        {
            if (iterator.hasNext())
            {
                iterator.next();
            }
            else
            {
                Assert.fail("The iterable has " + i + " items, " + expectedSize + " expected.");
            }
        }
        
        if (iterator.hasNext())
        {
            Assert.fail("The iterable has too much items, " + expectedSize + " expected.");
        }
        
        try
        {
            iterator.next();
            Assert.fail("The iterable has too much items, " + expectedSize + " expected.");
        }
        catch (NoSuchElementException e)
        {
            // Expected behavior.
        }
    }
    
    /**
     * Assert a facet value count.
     * @param expectedCount The expected count.
     * @param results The search results.
     * @param facet The facet name.
     * @param value The facet value.
     */
    protected static void assertFacetValueCount(int expectedCount, SearchResults<?> results, String facet, String value)
    {
        Map<String, Integer> counts = results.getFacetResults().get(facet);
        
        if (counts == null)
        {
            Assert.fail("The facet doesn't exist");
            return;
        }
        
        Assert.assertEquals(Integer.valueOf(expectedCount), counts.get(value));
    }
    
    /**
     * Assert that the given facet value doesn't exist.
     * @param results The search results.
     * @param facet The facet name.
     * @param value The facet value.
     */
    protected static void assertNullFacetValue(SearchResults<?> results, String facet, String value)
    {
        Map<String, Integer> counts = results.getFacetResults().get(facet);
        
        if (counts == null)
        {
            Assert.fail("The facet doesn't exist");
            return;
        }
        
        Assert.assertFalse(counts.containsKey(value));
    }
    
    /**
     * Assert that the given search results have the correct country facet count (for a blank search).
     * @param results The search results.
     */
    protected static void assertCountryFacetCounts(SearchResults<?> results)
    {
        assertFacetValueCount(4, results, "address/country", "france");
        assertFacetValueCount(4, results, "address/country", "japan");
        assertFacetValueCount(3, results, "address/country", "spain");
        assertNullFacetValue(results, "address/country", "china");
    }
    
    /**
     * Assert that the given search results have the correct workflow facet count (for a blank search).
     * @param results The search results.
     */
    protected static void assertWorkflowFacetCounts(SearchResults<?> results)
    {
        assertFacetValueCount(15, results, WorkflowStepSearchField.NAME, "1");
        assertFacetValueCount(6, results, WorkflowStepSearchField.NAME, "2");
        assertFacetValueCount(3, results, WorkflowStepSearchField.NAME, "3");
        assertNullFacetValue(results, WorkflowStepSearchField.NAME, "9999");
    }
    
    /**
     * Assert that the given search results have the correct workflow facet count
     * (for a blank search, when "france" is selected as a country in the facet values).
     * @param results The search results.
     */
    protected static void assertFranceWorkflowFacetCounts(SearchResults<?> results)
    {
        assertFacetValueCount(1, results, WorkflowStepSearchField.NAME, "1");
        assertFacetValueCount(2, results, WorkflowStepSearchField.NAME, "2");
        assertFacetValueCount(1, results, WorkflowStepSearchField.NAME, "3");
        assertNullFacetValue(results, WorkflowStepSearchField.NAME, "9999");
    }
}
