/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.compare.ContentComparator;
import org.ametys.cms.content.compare.ContentComparatorChange;
import org.ametys.cms.content.compare.ContentComparatorResult;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.data.Binary;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.Reference;
import org.ametys.cms.data.RichText;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableDefaultContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.cms.workflow.ContentWorkflowHelper;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.explorer.resources.ModifiableResource;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeater;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.compare.DataChangeType;
import org.ametys.runtime.model.compare.DataChangeTypeDetail;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

import com.opensymphony.workflow.WorkflowException;

/**
 * Test the Content comparator API
 */
@NeedFeatures({
    "cms/content.comparator",
    "cms/workflow",
    "cms/workflow.content",
    "repository/resolver",
    "explorer/factories",
    "exhaustif/content-type",
})
public class CompareContentTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String CONTENTTYPE_EXHAUSTIVE = "org.ametys.test.Content.exhaustive";
    private static final String CONTENTTYPE_CITY = "simple.City";
    private static final String CONTENTTYPE_ANIMAL = "simple.Animal";
    private static final String CONTENTTYPE_MAMMAL = "simple.Mammal";
    private static final String CONTENTTYPE_SNAKE = "simple.Snake";
    private static final String CONTENTTYPE_NICKNAME = "simple.Nickname";
    
    private static final String INITIAL_FILE_NAME = "bebeaabbaab";
    
    /** The date format. */
    private static final DateTimeFormatter DATE_FORMATTER = DateTimeFormatter.ofPattern("yyyy-MM-dd");
    
    /** The datetime format. */
    private static final DateTimeFormatter DATETIME_FORMATTER = DateTimeFormatter.ofPattern("yyyy-MM-dd'T'HH:mm:ssXXX");
    
    private static ContentComparator _contentComparator;
    private static ContentWorkflowHelper _contentWorkflowHelper;
    private static AmetysObjectResolver _ametysObjectResolver;
    
    private static String _fileId;
    private static String _firstAttachmentId;
    private static String _secondAttachmentId;
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _contentComparator = (ContentComparator) manager.lookup(ContentComparator.ROLE);
        _contentWorkflowHelper = (ContentWorkflowHelper) manager.lookup(ContentWorkflowHelper.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        _fileId = _createResource(INITIAL_FILE_NAME);
        _firstAttachmentId = _createResource("bebeaabbaabd");
        _secondAttachmentId = _createResource("aaabbbcccddd");
    }
    
    /**
     * Compare 2 equal contents
     * @throws Exception fail
     */
    @Test
    public void testEqual() throws Exception
    {
        String[] types = new String[]{CONTENTTYPE_CITY};

        ModifiableDefaultContent content1 = createContent("Test content", "content", types);
        content1.setValue("postCode", "31300");
        content1.saveChanges();
        
        ModifiableDefaultContent content2 = createContent("Test content", "content-2", types);
        content2.setValue("postCode", "31300");
        content2.saveChanges();
        
        // Tests content1 and content2 comparison
        ContentComparatorResult compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be equals", compare.areEquals());

        // Contents to test references.
        ModifiableDefaultContent content3 = createCompletedExhaustiveContent("Title", "complete1");
        ModifiableDefaultContent content4 = createCompletedExhaustiveContent("Title", "complete2");
        ModifiableDefaultContent link1 = createContent("Linked content 1", "linked-content-1", CONTENTTYPE_EXHAUSTIVE);
        ModifiableDefaultContent link2 = createContent("Linked content 2", "linked-content-2", CONTENTTYPE_EXHAUSTIVE);
        ModifiableDefaultContent link3 = createContent("Linked content 3", "linked-content-3", CONTENTTYPE_EXHAUSTIVE);
        ModifiableDefaultContent link4 = createContent("Linked content 4", "linked-content-4", CONTENTTYPE_EXHAUSTIVE);

        content3.setValue("linkContent", link1.getId());
        content3.setValue("linkContents", new String[] {link2.getId(), link3.getId(), link4.getId()});
        content3.setValue("owns", new String[] {link2.getId(), link3.getId(), link4.getId()});
        content4.setValue("linkContent", link1.getId());
        content4.setValue("linkContents", new String[] {link2.getId(), link3.getId(), link4.getId()});
        content4.setValue("owns", new String[] {link2.getId(), link3.getId(), link4.getId()});
        
        ModifiableModelAwareRepeater repeater3 = content3.getComposite("composite", true).getRepeater("repeater", true);
        repeater3.addEntry().setValue("comp-string", "entry 1");
        repeater3.getEntry(1).setValue("comp-owns", link1.getId());
        repeater3.addEntry().setValue("comp-string", "entry 2");
        repeater3.getEntry(2).setValue("comp-owns", link2.getId());
        
        ModifiableModelAwareRepeater repeater4 = content4.getComposite("composite", true).getRepeater("repeater", true);
        repeater4.addEntry().setValue("comp-string", "entry 1");
        repeater4.getEntry(1).setValue("comp-owns", link1.getId());
        repeater4.addEntry().setValue("comp-string", "entry 2");
        repeater4.getEntry(2).setValue("comp-owns", link2.getId());

        content3.saveChanges();
        content4.saveChanges();
        
        compare = _contentComparator.compare(content3, content4);
        Assert.assertTrue("Content should be equals", compare.areEquals());
    }
    
    /**
     * Compare 2 different Strings
     * @throws Exception something went wrong
     */
    @Test
    public void testDifferentStrings() throws Exception
    {
        String[] types = new String[]{CONTENTTYPE_CITY};
        ModifiableDefaultContent content1 = createContent("Test content", "content", types);
        ModifiableDefaultContent content2 = createContent("Test content", "content-2", types);
        
        //STRING
        content1.setValue("postCode", "31300");
        content1.saveChanges();

        content2.setValue("postCode", "313001");
        content2.saveChanges();
        ContentComparatorResult compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "postCode", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE_CONTENT_END);
        
        content2.setValue("postCode", "31400");
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "postCode", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        
        content2.setValue("postCode", "031300");
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "postCode", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE_CONTENT_START);
        
        content2.setValue("postCode", "1300");
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "postCode", DataChangeType.MODIFIED, DataChangeTypeDetail.LESS_CONTENT_START);
        
        content2.setValue("postCode", "3130");
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "postCode", DataChangeType.MODIFIED, DataChangeTypeDetail.LESS_CONTENT_END);
        
        content2.setValue("postCode", "130");
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "postCode", DataChangeType.MODIFIED, DataChangeTypeDetail.LESS_CONTENT);
        
        content2.setValue("postCode", "0313000");
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "postCode", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE_CONTENT);
        
        content2.removeValue("postCode");
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "postCode", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        
        content2.setValue("postCode", "31300");
        content2.saveChanges();
        content1.removeValue("postCode");
        content1.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "postCode", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        
        
        // Multiple string data
        content1 = createCompletedExhaustiveContent("Test content", "content-3");
        content2 = createCompletedExhaustiveContent("Test content", "content-4");
        //default = new String[] {"one", "two", "three"}
        content2.setValue("stringmult", new String[] {"one", "two", "forrrrr"});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        org.junit.Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "stringmult", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        
        content2.setValue("stringmult", new String[] {"one", "three", "two"});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "stringmult", DataChangeType.MODIFIED, DataChangeTypeDetail.ORDER);
        
        content2.setValue("stringmult", new String[] {"one", "three", "twelve", "two"});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "stringmult", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE);
        
        content2.setValue("stringmult", new String[] {"one", "three"});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "stringmult", DataChangeType.MODIFIED, DataChangeTypeDetail.LESS);
    }
    
    /**
     * Compare 2 different Contents attributes
     * @throws Exception something went wrong
     */
    @Test
    public void testDifferentContents() throws Exception
    {
        ModifiableDefaultContent content1 = createCompletedExhaustiveContent("Test content", "content-1");
        ModifiableDefaultContent content2 = createCompletedExhaustiveContent("Test content", "content-2");
        
        ModifiableDefaultContent link1 = createContent("Linked content 1", "linked-content-1", CONTENTTYPE_EXHAUSTIVE);
        ModifiableDefaultContent link2 = createContent("Linked content 2", "linked-content-2", CONTENTTYPE_EXHAUSTIVE);
        ModifiableDefaultContent link3 = createContent("Linked content 3", "linked-content-3", CONTENTTYPE_EXHAUSTIVE);
        ModifiableDefaultContent link4 = createContent("Linked content 4", "linked-content-4", CONTENTTYPE_EXHAUSTIVE);

        content1.setValue("linkContent", link1.getId());
        content1.setValue("linkContents", new String[] {link2.getId(), link3.getId(), link4.getId()});
        content1.setValue("owns", new String[] {link2.getId(), link3.getId(), link4.getId()});
        content1.saveChanges();
        ContentComparatorResult compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "linkContent", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "linkContents", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "owns", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        
        content2.setValue("linkContent", link2.getId());
        content2.setValue("linkContents", new String[] {link4.getId(), link3.getId(), link2.getId()});
        content2.setValue("owns", new String[] {link3.getId(), link4.getId(), link2.getId()});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "linkContent", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "linkContents", DataChangeType.MODIFIED, DataChangeTypeDetail.ORDER);
        assertContentChange(compare, "owns", DataChangeType.MODIFIED, DataChangeTypeDetail.ORDER);
        
        content2.setValue("linkContents", new String[] {link4.getId(), link1.getId(), link3.getId(), link2.getId()});
        content2.setValue("owns", new String[] {link3.getId(), link2.getId()});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "linkContents", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE);
        assertContentChange(compare, "owns", DataChangeType.MODIFIED, DataChangeTypeDetail.LESS);

        content2.setValue("linkContents", new String[] {link4.getId(), link1.getId()});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "linkContents", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        
        content1.removeValue("linkContent");
        content1.removeValue("linkContents");
        content1.removeValue("owns");
        content1.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "linkContent", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "linkContents", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "owns", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
    }
    
    /**
     * Compare 2 different Files/binary/richtext attributes
     * @throws Exception something went wrong
     */
    @Test
    public void testDifferentFiles() throws Exception
    {
        String file = "file";
        String binary = "binary";
        String illustration = "illustration";
        String image = "image";
        String richText = "content";

        ModifiableDefaultContent content1 = createCompletedExhaustiveContent("Test content", "content-1");
        ModifiableDefaultContent content2 = createCompletedExhaustiveContent("Test content", "content-2");
        
        ZonedDateTime dateTime = ZonedDateTime.parse("2017-06-29T14:33:57Z", DATETIME_FORMATTER);
        
        String textContent = "alt Text content Modified";

        content2.removeValue(file);
        content2.removeValue(binary);
        content2.removeValue(richText);
        content2.saveChanges();
        ContentComparatorResult compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, file, DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, binary, DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, richText, DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        
        content2.setValue(file, _createResource("bebeaabbaabd"));
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, file, DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        
        content2.removeValue(file);
        Binary fileBinaryAttribute = new Binary();
        fileBinaryAttribute.setFilename("test2.txt");
        fileBinaryAttribute.setMimeType("text/plain-alt");
        fileBinaryAttribute.setEncoding("UTF-16");
        fileBinaryAttribute.setLastModificationDate(dateTime);
        fileBinaryAttribute.setInputStream(new ByteArrayInputStream(textContent.getBytes("UTF-8")));
        content2.setValue(file, fileBinaryAttribute);
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, file, DataChangeType.MODIFIED, DataChangeTypeDetail.TYPE);
        
        content1.removeValue(file);
        content1.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, file, DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        
        
        ModifiableModelAwareComposite composite = content2.getComposite("illustration");
        
        composite.removeValue(image);
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, illustration + "/" + image, DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        
        composite.setValue("alt-text", "Quelquechose d'autre");
        
        Binary binaryAttribute = new Binary();
        binaryAttribute.setFilename("test2.txt");
        binaryAttribute.setMimeType("text/plain-alt");
        binaryAttribute.setEncoding("UTF-16");
        binaryAttribute.setLastModificationDate(dateTime);
        binaryAttribute.setInputStream(new ByteArrayInputStream(textContent.getBytes("UTF-8")));
        composite.setValue(image, binaryAttribute);
        
        Binary binaryAttribute2 = new Binary();
        binaryAttribute2.setFilename("test2.txt");
        binaryAttribute2.setMimeType("text/plain-alt");
        binaryAttribute2.setEncoding("UTF-16");
        binaryAttribute2.setLastModificationDate(dateTime);
        binaryAttribute2.setInputStream(new ByteArrayInputStream(textContent.getBytes("UTF-8")));
        content2.setValue(binary, binaryAttribute2);
        
        RichText richTextAttribute = new RichText();
        richTextAttribute.setEncoding("UTF-8");
        richTextAttribute.setMimeType("application/xml");
        richTextAttribute.setLastModificationDate(dateTime);
        richTextAttribute.setInputStream(new ByteArrayInputStream(textContent.getBytes("UTF-8")));
        content2.setValue(richText, richTextAttribute);
        
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, illustration + "/alt-text", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, illustration + "/" + image + "/fileName", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, illustration + "/" + image + "/mimeType", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, illustration + "/" + image + "/encoding", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, illustration + "/" + image + "/inputStream", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        
        assertContentChange(compare, binary + "/fileName", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, binary + "/mimeType", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, binary + "/encoding", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, binary + "/inputStream", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);

        assertContentChange(compare, richText + "/mimeType", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, richText + "/inputStream", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        

        content1.getComposite(illustration).removeValue("image");
        content1.removeValue(binary);
        content1.removeValue(richText);
        content1.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, illustration + "/" + image, DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, binary, DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, richText, DataChangeType.ADDED, DataChangeTypeDetail.NONE);
    }
    
    /**
     * Compare 2 different Boolean attributes
     * @throws Exception something went wrong
     */
    @Test
    public void testDifferentBoolean() throws Exception
    {
        ModifiableDefaultContent content1 = createCompletedExhaustiveContent("Test content", "content-1");
        ModifiableDefaultContent content2 = createCompletedExhaustiveContent("Test content", "content-2");

        content2.removeValue("boolean");
        content2.removeValue("booleanmult");
        content2.saveChanges();
        ContentComparatorResult compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "boolean", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "booleanmult", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        
        content2.setValue("boolean", false);
        //default : new boolean[]{true, true, false, true, false, true}
        content2.setValue("booleanmult", new Boolean[]{false, false, false});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "boolean", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "booleanmult", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        
        content2.setValue("booleanmult", new Boolean[]{true, true, true, true, false, false});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "booleanmult", DataChangeType.MODIFIED, DataChangeTypeDetail.ORDER);
        
        content2.setValue("booleanmult", new Boolean[]{true, true, true});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "booleanmult", DataChangeType.MODIFIED, DataChangeTypeDetail.LESS);
        
        content2.setValue("booleanmult", new Boolean[]{true, true, true, true, false, false, true, false, false});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "booleanmult", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE);

        content1.removeValue("boolean");
        content1.removeValue("booleanmult");
        content1.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "boolean", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "booleanmult", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
    }
    
    /**
     * Compare 2 different Reference attributes
     * @throws Exception something went wrong
     */
    @Test
    public void testDifferentReference() throws Exception
    {
        ModifiableDefaultContent content1 = createCompletedExhaustiveContent("Test content", "content-1");
        ModifiableDefaultContent content2 = createCompletedExhaustiveContent("Test content", "content-2");
        
        content2.removeValue("reference");
        content2.saveChanges();
        ContentComparatorResult compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "reference", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);

        content2.setValue("reference", new Reference("https://www.ametys.org", "__externall"));
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "reference/type", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "reference/value", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        
        content1.removeValue("reference");
        content1.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "reference", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
    }
    
    /**
     * Compare 2 different User attributes
     * @throws Exception something went wrong
     */
    @Test
    public void testDifferentUser() throws Exception
    {
        ModifiableDefaultContent content1 = createCompletedExhaustiveContent("Test content", "content-1");
        ModifiableDefaultContent content2 = createCompletedExhaustiveContent("Test content", "content-2");

        content2.removeValue("user");
        content2.removeValue("usermult");
        content2.saveChanges();
        ContentComparatorResult compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "user", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "usermult", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        
        content2.setValue("user", USER1_TEST_IDENTITY);
        content2.setValue("usermult", new UserIdentity[] {USER2_TEST_IDENTITY, USER1_TEST_IDENTITY, USER3_TEST_IDENTITY});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "user", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "usermult", DataChangeType.MODIFIED, DataChangeTypeDetail.ORDER);
        
        content2.setValue("usermult", new UserIdentity[] {USER2_TEST_IDENTITY, ANONYMOUS_TEST_IDENTITY, USER1_TEST_IDENTITY, USER3_TEST_IDENTITY});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "usermult", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE);
        
        content2.setValue("usermult", new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "usermult", DataChangeType.MODIFIED, DataChangeTypeDetail.LESS);
        
        content1.removeValue("user");
        content1.removeValue("usermult");
        content1.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "user", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "usermult", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
    }
    /**
     * Compare 2 different repeaters
     * @throws Exception something went wrong
     */
    @Test
    public void testDifferentRepeaters() throws Exception
    {
        ModifiableDefaultContent content1 = createCompletedExhaustiveContent("Test content", "content-1");
        ModifiableDefaultContent content2 = createCompletedExhaustiveContent("Test content", "content-2");
        
        content2.setValue("attachments[1]/attachment", _createResource("bebeaabbaabdd"));
        content2.setValue("attachments[1]/attachment-text", "Some Text 2");

        content2.removeValue("attachments[3]/attachment");
        content2.setValue("attachments[3]/attachment", _createResource("aeaeaeaeae"));
        content2.setValue("attachments[3]/attachment-text", "Some Text 3");
        content2.saveChanges();
        ContentComparatorResult compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "attachments[1]/attachment", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "attachments[1]/attachment-text", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE_CONTENT_END);
        assertContentChange(compare, "attachments[3]/attachment", DataChangeType.MODIFIED, DataChangeTypeDetail.TYPE);
        assertContentChange(compare, "attachments[3]/attachment-text", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
    }
    
    /**
     * Compare 2 different Date attributes
     * @throws Exception something went wrong
     */
    @Test
    public void testDifferentDate() throws Exception
    {
        ModifiableDefaultContent content1 = createCompletedExhaustiveContent("Test content", "content-1");
        ModifiableDefaultContent content2 = createCompletedExhaustiveContent("Test content", "content-2");

        LocalDate date1 = LocalDate.parse("2017-04-28", DATE_FORMATTER);
        ZonedDateTime dateTime1 = ZonedDateTime.parse("2017-04-29T14:33:57Z", DATETIME_FORMATTER);
        LocalDate date2 = LocalDate.parse("2017-04-29", DATE_FORMATTER);
        ZonedDateTime dateTime2 = ZonedDateTime.parse("2017-04-29T14:34:57Z", DATETIME_FORMATTER);
        LocalDate date3 = LocalDate.parse("2017-04-30", DATE_FORMATTER);
        ZonedDateTime dateTime3 = ZonedDateTime.parse("2017-04-29T14:35:57Z", DATETIME_FORMATTER);
        LocalDate date4 = LocalDate.parse("2018-04-30", DATE_FORMATTER);
        ZonedDateTime dateTime4 = ZonedDateTime.parse("2018-04-29T14:35:57Z", DATETIME_FORMATTER);

        content2.removeValue("date");
        content2.removeValue("datemult");
        content2.removeValue("datetime");
        content2.removeValue("datetimemult");
        content2.saveChanges();
        ContentComparatorResult compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "date", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "datemult", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "datetime", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "datetimemult", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);

        content1.setValue("date", date2);
        content1.setValue("datetime", dateTime2);
        content1.saveChanges();
        content2.setValue("date", date1);
        content2.setValue("datetime", dateTime1);
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "date", DataChangeType.MODIFIED, DataChangeTypeDetail.BEFORE);
        assertContentChange(compare, "datetime", DataChangeType.MODIFIED, DataChangeTypeDetail.BEFORE);
        content2.setValue("date", date3);
        content2.setValue("datetime", dateTime3);
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "date", DataChangeType.MODIFIED, DataChangeTypeDetail.AFTER);
        assertContentChange(compare, "datetime", DataChangeType.MODIFIED, DataChangeTypeDetail.AFTER);

        content2.setValue("datemult", new LocalDate[]{date2, date1, date3});
        content2.setValue("datetimemult", new ZonedDateTime[]{dateTime2, dateTime1, dateTime3});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "datemult", DataChangeType.MODIFIED, DataChangeTypeDetail.ORDER);
        assertContentChange(compare, "datetimemult", DataChangeType.MODIFIED, DataChangeTypeDetail.ORDER);

        content2.setValue("datemult", new LocalDate[]{date1, date3});
        content2.setValue("datetimemult", new ZonedDateTime[]{dateTime2, dateTime3});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "datemult", DataChangeType.MODIFIED, DataChangeTypeDetail.LESS);
        assertContentChange(compare, "datetimemult", DataChangeType.MODIFIED, DataChangeTypeDetail.LESS);

        content2.setValue("datemult", new LocalDate[]{date2, date1, date4, date3});
        content2.setValue("datetimemult", new ZonedDateTime[]{dateTime2, dateTime4, dateTime1, dateTime3});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "datemult", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE);
        assertContentChange(compare, "datetimemult", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE);

        content2.setValue("datemult", new LocalDate[]{date2, date1, date4});
        content2.setValue("datetimemult", new ZonedDateTime[]{dateTime2, dateTime4, dateTime3});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "datemult", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "datetimemult", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);

        content1.removeValue("date");
        content1.removeValue("datemult");
        content1.removeValue("datetime");
        content1.removeValue("datetimemult");
        content1.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "date", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "datemult", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "datetime", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "datetimemult", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
    }
    
    /**
     * Compare 2 different Geocode attributes
     * @throws Exception something went wrong
     */
    @Test
    public void testDifferentGeocode() throws Exception
    {
        String gps = "gps";
        String latitude = "latitude";
        String longitude = "longitude";

        ModifiableDefaultContent content1 = createCompletedExhaustiveContent("Test content", "content-1");
        ModifiableDefaultContent content2 = createCompletedExhaustiveContent("Test content", "content-2");

        content2.removeValue(gps);
        content2.saveChanges();
        ContentComparatorResult compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, gps, DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        
        content2.setValue(gps, null);
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, gps, DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        
        content2.setValue(gps, new Geocode(45.481, 32.208));
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, gps + "/" + latitude, DataChangeType.MODIFIED, DataChangeTypeDetail.MORE);
        assertContentChange(compare, gps + "/" + longitude, DataChangeType.MODIFIED, DataChangeTypeDetail.MORE);
        
        content2.setValue(gps, new Geocode(4.41, 2.28));
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, gps + "/" + latitude, DataChangeType.MODIFIED, DataChangeTypeDetail.LESS);
        assertContentChange(compare, gps + "/" + longitude, DataChangeType.MODIFIED, DataChangeTypeDetail.LESS);
        
        content1.removeValue(gps);
        content1.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, gps, DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        
        content1.setValue(gps, null);
        content1.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, gps, DataChangeType.ADDED, DataChangeTypeDetail.NONE);
    }
    
    /**
     * Compare 2 different Numbers attributes (long, double)
     * @throws Exception something went wrong
     */
    @Test
    public void testDifferentNumbers() throws Exception
    {
        ModifiableDefaultContent content1 = createCompletedExhaustiveContent("Test content", "content-1");
        ModifiableDefaultContent content2 = createCompletedExhaustiveContent("Test content", "content-2");

        content2.removeValue("long");
        content2.removeValue("longmult");
        content2.removeValue("double");
        content2.removeValue("doublemult");
        content2.saveChanges();
        ContentComparatorResult compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "long", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "longmult", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "double", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "doublemult", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);

        content2.setValue("long", 4L);
        content2.setValue("double", 1.434D);
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "long", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE);
        assertContentChange(compare, "double", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE);
        content2.setValue("long", 2L);
        content2.setValue("double", 1.04D);
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "long", DataChangeType.MODIFIED, DataChangeTypeDetail.LESS);
        assertContentChange(compare, "double", DataChangeType.MODIFIED, DataChangeTypeDetail.LESS);

        content2.setValue("longmult", new Long[] {6L, 5L, 12L});
        content2.setValue("doublemult", new Double[] {6.12D, 3.14D});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "longmult", DataChangeType.MODIFIED, DataChangeTypeDetail.ORDER);
        assertContentChange(compare, "doublemult", DataChangeType.MODIFIED, DataChangeTypeDetail.ORDER);

        content2.setValue("longmult", new Long[] {6L, 5L});
        content2.setValue("doublemult", new Double[] {6.12D});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "longmult", DataChangeType.MODIFIED, DataChangeTypeDetail.LESS);
        assertContentChange(compare, "doublemult", DataChangeType.MODIFIED, DataChangeTypeDetail.LESS);

        content2.setValue("longmult", new Long[] {6L, 12L, 5L, 12L});
        content2.setValue("doublemult", new Double[] {6.12D, 1.2D, 3.14D});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "longmult", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE);
        assertContentChange(compare, "doublemult", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE);

        content2.setValue("longmult", new Long[]{45L, 1L});
        content2.setValue("doublemult", new Double[]{3D, 45.2D});
        content2.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "longmult", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "doublemult", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);

        content1.removeValue("long");
        content1.removeValue("longmult");
        content1.removeValue("double");
        content1.removeValue("doublemult");
        content1.saveChanges();
        compare = _contentComparator.compare(content1, content2);
        Assert.assertTrue("Content should be different", !compare.areEquals());
        assertContentChange(compare, "long", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "longmult", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "double", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "doublemult", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
    }
    
    /**
     * Compare contents with filter using viewName
     * @throws Exception something went wrong
     */
    @Test
    public void testView() throws Exception
    {
        ModifiableDefaultContent content1 = createCompletedExhaustiveContent("Test content", "content-1");
        ModifiableDefaultContent content2 = createCompletedExhaustiveContent("Test content", "content-2");

        LocalDate date = LocalDate.parse("2018-04-30", DATE_FORMATTER);
        ZonedDateTime dateTime = ZonedDateTime.parse("2018-04-29T14:35:57Z", DATETIME_FORMATTER);
        
        content2.removeValue("long");
        content2.removeValue("longmult");
        content2.removeValue("double");
        content2.removeValue("doublemult");
        content2.setValue("date", date);
        content2.setValue("datetime", dateTime);
        content2.saveChanges();
        ContentComparatorResult compare = _contentComparator.compare(content1, content2, "abstract", true);
        Assert.assertTrue("Content should be equals with \"abstract\" view", compare.areEquals());
        
        compare = _contentComparator.compare(content1, content2, "link", true);
        Assert.assertTrue("Content should be equals with \"link\" view", compare.areEquals());
        
        compare = _contentComparator.compare(content1, content2, "main", true);
        Assert.assertTrue("Content should be different with \"main\" view", !compare.areEquals());

        ModifiableDefaultContent content3 = createCompletedExhaustiveContent("Test content", "content-3");
        ModifiableDefaultContent content4 = createCompletedExhaustiveContent("Test content", "content-4");
        
        content4.setValue("stringnotinmain", "one, two, three, four, five");
        content4.setValue("stringmultnotinmain", new String[] {"one", "two", "three", "four", "five"});
        content4.saveChanges();
        compare = _contentComparator.compare(content3, content4, "main", true);
        Assert.assertTrue("Content should be equals with \"main\" view", compare.areEquals());
        compare = _contentComparator.compare(content3, content4);
        Assert.assertTrue("Content should be different without view", !compare.areEquals());
        assertContentChange(compare, "stringnotinmain", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE_CONTENT_END);
        assertContentChange(compare, "stringmultnotinmain", DataChangeType.MODIFIED, DataChangeTypeDetail.MORE);

        content3.removeValue("stringnotinmain");
        content4.removeValue("stringmultnotinmain");
        content3.saveChanges();
        content4.saveChanges();
        compare = _contentComparator.compare(content3, content4, "main", true);
        Assert.assertTrue("Content should be equals with \"main\" view", compare.areEquals());
        compare = _contentComparator.compare(content3, content4);
        Assert.assertTrue("Content should be different without view", !compare.areEquals());
        assertContentChange(compare, "stringnotinmain", DataChangeType.ADDED, DataChangeTypeDetail.NONE);
        assertContentChange(compare, "stringmultnotinmain", DataChangeType.REMOVED, DataChangeTypeDetail.NONE);
    }
    
    /**
     * Compare contents, with content changes
     * @throws Exception something went wrong
     */
    @Test
    public void testGeneratedView() throws Exception
    {

        ModifiableDefaultContent dog = createContent("good boy", "dog-1", CONTENTTYPE_MAMMAL);
        ModifiableDefaultContent snake = createContent("bad boy", "snake-1", CONTENTTYPE_SNAKE);
        ContentComparatorResult compare;

        dog.setValue("color", "Brown");
        snake.setValue("length", 12L);

        dog.setValue("eyes", 2L);
        snake.setValue("eyes", 2L);
        
        dog.saveChanges();
        snake.saveChanges();
        
        compare = _contentComparator.compare(dog, snake, null, true);
        Assert.assertTrue("Content should be different because of content type", !compare.areEquals());
        
        compare = _contentComparator.compare(dog, snake, null, false);
        Assert.assertTrue("Content should be different because of title", !compare.areEquals());
        
        compare = _contentComparator.compare(snake, dog, null, false);
        Assert.assertTrue("Content should be different because of title", !compare.areEquals());
        
        compare = _contentComparator.compare(dog, snake, "main", false);
        Assert.assertTrue("Content should be equals with \"main\" view", compare.areEquals());
        
        compare = _contentComparator.compare(snake, dog, "main", false);
        Assert.assertTrue("Content should be equals with \"main\" view", compare.areEquals());
        
        dog.setValue("legs", 4L);
        snake.setValue("legs", 0L);
        
        dog.saveChanges();
        snake.saveChanges();

        compare = _contentComparator.compare(snake, dog, null, false);
        Assert.assertTrue("Content should be different without view", !compare.areEquals());
        compare = _contentComparator.compare(dog, snake, null, false);
        Assert.assertTrue("Content should be different without view", !compare.areEquals());
        compare = _contentComparator.compare(snake, dog, "details", false);
        Assert.assertTrue("Content should be equals with \"details\" without view", compare.areEquals());
        compare = _contentComparator.compare(dog, snake, "details", false);
        Assert.assertTrue("Content should be equals with \"details\" without view", compare.areEquals());

        ModifiableDefaultContent animal = createContent("Random animal", "dog-1", CONTENTTYPE_ANIMAL);
        ModifiableDefaultContent dog2 = createContent("good boy", "dog-1", new String[] {CONTENTTYPE_MAMMAL, CONTENTTYPE_NICKNAME});
        ModifiableDefaultContent snake2 = createContent("bad boy", "snake-1", new String[] {CONTENTTYPE_SNAKE, CONTENTTYPE_NICKNAME});

        animal.setValue("eyes", 2L);
        dog2.setValue("eyes", 2L);
        snake2.setValue("eyes", 2L);
        dog2.setValue("nickname", "Poppy");
        snake2.setValue("nickname", "Danger Noodle");
        animal.saveChanges();
        dog2.saveChanges();
        snake2.saveChanges();

        // A == B
        // B == C
        // A != C (because "nickname" is only on A and C, and are differents)
        compare = _contentComparator.compare(dog2, animal, null, false);
        Assert.assertTrue("Content should be different without view because of title", !compare.areEquals());
        compare = _contentComparator.compare(snake2, animal, null, false);
        Assert.assertTrue("Content should be different without view because of title", !compare.areEquals());
        compare = _contentComparator.compare(snake2, dog2, null, false);
        Assert.assertTrue("Content should be different without view", !compare.areEquals());
        assertContentChange(compare, "nickname", DataChangeType.MODIFIED, DataChangeTypeDetail.NONE);
        
        //change nickname, and now A == B == C
        snake2.setValue("nickname", "Poppy");
        snake2.saveChanges();
        compare = _contentComparator.compare(snake2, dog2, null, false);
        Assert.assertTrue("Content should be different without view because of title", !compare.areEquals());
    }
    
    private boolean assertContentChange(ContentComparatorResult compare, String path, DataChangeType type, DataChangeTypeDetail detail)
    {
        List<ContentComparatorChange> results = new ArrayList<>();
        for (ContentComparatorChange change : compare.getChanges())
        {
            if (path.equals(change.getPath()))
            {
                if (type.equals(change.getDataChangeType()) && detail.equals(change.getDataChangeTypeDetail()))
                {
                    return true;
                }
                else
                {
                    results.add(change);
                }
            }
        }
        if (results.size() == 0)
        {
            Assert.fail("No change found for " + path + " with type=" + type + " and detail=" + detail);
        }
        else
        {
            String resultString = "";
            for (ContentComparatorChange change : results)
            {
                resultString += "| type=" + change.getDataChangeType() + ";detail=" + change.getDataChangeTypeDetail() + " ";
            }
            
            Assert.fail("Not the good type of change found for " + path + ". Expected type=" + type + ";detail=" + detail + " found " + resultString);
        }
        
        return false;
    }
    
    private ModifiableDefaultContent createContent(String title, String name, String singleType) throws WorkflowException
    {
        return createContent(title, name, new String[]{singleType});
    }
    private ModifiableDefaultContent createContent(String title, String name, String[] type) throws WorkflowException
    {
        Map<String, Object> results = _contentWorkflowHelper.createContent("content", 1, name, title, type, new String[0], "en");
        
        return (ModifiableDefaultContent) results.get(Content.class.getName());
    }
    
    private ModifiableDefaultContent createCompletedExhaustiveContent(String title, String name) throws Exception
    {
        LocalDate date = LocalDate.parse("2017-04-28", DATE_FORMATTER);
        ZonedDateTime dateTime = ZonedDateTime.parse("2017-04-29T14:33:57Z", DATETIME_FORMATTER);
        
        LocalDate date2 = LocalDate.parse("2017-04-29", DATE_FORMATTER);
        ZonedDateTime dateTime2 = ZonedDateTime.parse("2017-04-29T14:34:57Z", DATETIME_FORMATTER);
        
        LocalDate date3 = LocalDate.parse("2017-04-30", DATE_FORMATTER);
        ZonedDateTime dateTime3 = ZonedDateTime.parse("2017-04-29T14:35:57Z", DATETIME_FORMATTER);
        
        ModifiableDefaultContent content = createContent(title, name, CONTENTTYPE_EXHAUSTIVE);
        
        content.setValue("title", "Super Titre");
        content.setValue("abstract", "Des données avec retour\nà la ligne");
        
        //illustration
        ModifiableModelAwareComposite illustration = content.getComposite("illustration", true);
        illustration.setValue("alt-text", "Alternative text");
        String textContent = "Text content";
        Binary image = new Binary();
        image.setFilename("test.txt");
        image.setMimeType("text/plain");
        image.setEncoding("UTF-8");
        image.setLastModificationDate(dateTime);
        image.setInputStream(new ByteArrayInputStream(textContent.getBytes("UTF-8")));
        illustration.setValue("image", image);
        
        //binary
        String textContent2 = "Text content Binaire";
        Binary binary = new Binary();
        binary.setFilename("test3.txt");
        binary.setMimeType("text/plain");
        binary.setEncoding("UTF-8");
        binary.setLastModificationDate(dateTime);
        binary.setInputStream(new ByteArrayInputStream(textContent2.getBytes("UTF-8")));
        content.setValue("binary", binary);

        content.setValue("string", "one, two, three");
        content.setValue("stringmult", new String[] {"one", "two", "three"});
        content.setValue("stringnotinmain", "one, two, three, four");
        content.setValue("stringmultnotinmain", new String[] {"one", "two", "three", "four"});
        content.setValue("long", 3L);
        content.setValue("longmult", new Long[] {5L, 6L, 12L});
        content.setValue("double", 1.234D);
        content.setValue("doublemult", new Double[] {3.14D, 6.12D});
        content.setValue("boolean", true);
        content.setValue("booleanmult", new Boolean[]{true, true, false, true, false, true});
        content.setValue("date", date);
        content.setValue("datemult", new LocalDate[]{date, date2, date3});
        content.setValue("datetime", dateTime);
        content.setValue("datetimemult", new ZonedDateTime[]{dateTime, dateTime2, dateTime3});
        content.getComposite("address", true).setValue("country", "france");
        
        //user, usermult
        content.setValue("user", ANONYMOUS_TEST_IDENTITY);
        content.setValue("usermult", new UserIdentity[] {USER1_TEST_IDENTITY, USER2_TEST_IDENTITY, USER3_TEST_IDENTITY});
        
        //reference externe
        content.setValue("reference", new Reference("http://www.ametys.org", "__external"));
        
        //gps
        content.setValue("gps", new Geocode(44.481, 3.208));
        
        //fichier (référence)
        content.setValue("file", _fileId);
        
        RichText richText = new RichText();
        String textContent3 = "Le contenu du richtext";
        richText.setEncoding("UTF-8");
        richText.setMimeType("text/xml");
        richText.setLastModificationDate(dateTime);
        richText.setInputStream(new ByteArrayInputStream(textContent3.getBytes("UTF-8")));
        content.setValue("content", richText);
        
        ModifiableModelAwareRepeater attachments = content.getRepeater("attachments", true);
        attachments.addEntry().setValue("attachment", _firstAttachmentId);
        attachments.getEntry(1).setValue("attachment-text", "Some Text");
        attachments.addEntry().setValue("attachment", _secondAttachmentId);
        attachments.getEntry(2).setValue("attachment-text", "Something interesting");
        
        Binary fileBinaryAttribute = new Binary();
        fileBinaryAttribute.setFilename("test2.txt");
        fileBinaryAttribute.setMimeType("text/plain-alt");
        fileBinaryAttribute.setEncoding("UTF-16");
        fileBinaryAttribute.setLastModificationDate(dateTime);
        fileBinaryAttribute.setInputStream(new ByteArrayInputStream(textContent.getBytes("UTF-8")));
        attachments.addEntry().setValue("attachment", fileBinaryAttribute);
        attachments.getEntry(3).setValue("attachment-text", "Something interesting");
        
        content.saveChanges();
        return content;
    }
    
    private String _createResource(String filename) throws IOException
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        ModifiableResource resource = root.createChild(filename, "ametys:resource");
        resource.setData(new ByteArrayInputStream("test".getBytes("UTF-8")), "text/plain", new Date(), USER1_TEST_IDENTITY);
        root.saveChanges();
        return resource.getId();
    }
}
