/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeaterEntry;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link Content} test cases for all dataToJSON methods
 */
@NeedFeatures({
    "repository/resolver",
    "workflow/workflow-provider",
    "content.properties/content-type-properties-tests",
    "cms/factories",
    "cms/workflow.content",
    "cms/org.ametys.cms.data.holder.helpers"
})
public class ContentDataToJSONTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    /** Default content node type name. */
    public static final String NT_DEFAULT_CONTENT = "ametys:defaultContent";
    
    /** Ametys object resolver to subclasses. */
    protected static AmetysObjectResolver _ametysObjectResolver;
    /** Content types helper */
    protected static ContentTypesHelper _contentTypesHelper;
    
    private static final String __CONTENT_TYPE_POINTING_TYPE_WITH_PROPERTIES = "tests.pointing.type.with.properties";
    private static final String __CONTENT_TYPE_WITH_PROPERTIES = "tests.with.properties";
    
    private static final ZonedDateTime __CREATION_DATE = ZonedDateTime.of(2022, 7, 1, 16, 11, 18, 0, ZoneOffset.UTC);
    private static final ZonedDateTime __MODIFICATION_DATE1 = ZonedDateTime.of(2022, 7, 30, 22, 30, 38, 0, ZoneOffset.UTC);
    private static final ZonedDateTime __MODIFICATION_DATE2 = ZonedDateTime.of(2022, 8, 6, 11, 10, 17, 0, ZoneOffset.UTC);
    private static final ZonedDateTime __VALIDATION_DATE = ZonedDateTime.of(2022, 8, 13, 12, 29, 59, 0, ZoneOffset.UTC);
    
    private static ModifiableContent _pointingContent;
    private static ModifiableContent _contentWithProperties1;
    private static ModifiableContent _contentWithProperties2;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();

        // CONTENTS WITH PROPERTIES
        _contentWithProperties1 = _createContent("contentWithProperties1", __CONTENT_TYPE_WITH_PROPERTIES);
        _contentWithProperties1.setLastContributor(new UserIdentity("user1", POPULATION_TEST_ID));
        _contentWithProperties1.setLastModified(__MODIFICATION_DATE2);
        _contentWithProperties1.setLastValidationDate(__VALIDATION_DATE);
        ModifiableModelAwareComposite contentWithProperties1Composite = _contentWithProperties1.getComposite("composite", true);
        contentWithProperties1Composite.setValue("long", 45L);
        contentWithProperties1Composite.setValue("longmult", new Long[] {45L, 18L, 97L}); // sum: 160
        ModifiableModelAwareRepeater contentWithProperties1Repeater = contentWithProperties1Composite.getRepeater("repeater", true);
        ModifiableModelAwareRepeaterEntry contentWithProperties1RepeaterEntry1 = contentWithProperties1Repeater.addEntry();
        contentWithProperties1RepeaterEntry1.setValue("string", "content 1 - entry 1");
        contentWithProperties1RepeaterEntry1.setValue("stringmult", new String[] {"content 1 - entry 1 - value 1", "content 1 - entry 1 - value 2", "content 1 - entry 1 - value 3"});
        ModifiableModelAwareRepeaterEntry contentWithProperties1RepeaterEntry2 = contentWithProperties1Repeater.addEntry();
        contentWithProperties1RepeaterEntry2.setValue("string", "content 1 - entry 2");
        contentWithProperties1RepeaterEntry2.setValue("stringmult", new String[] {"content 1 - entry 2 - value 1"});
        
        _contentWithProperties2 = _createContent("contentWithProperties2", __CONTENT_TYPE_WITH_PROPERTIES);
        ModifiableModelAwareComposite contentWithProperties2Composite = _contentWithProperties2.getComposite("composite", true);
        contentWithProperties2Composite.setValue("longmult", new Long[] {12L, 38L}); // sum: 50
        ModifiableModelAwareRepeater contentWithProperties2Repeater = contentWithProperties2Composite.getRepeater("repeater", true);
        ModifiableModelAwareRepeaterEntry contentWithProperties2RepeaterEntry1 = contentWithProperties2Repeater.addEntry();
        contentWithProperties2RepeaterEntry1.setValue("string", "content 2 - entry 1");
        contentWithProperties2RepeaterEntry1.setValue("stringmult", new String[] {"content 2 - entry 1 - value 1", "content 2 - entry 1 - value 2", "content 2 - entry 1 - value 3"});
        ModifiableModelAwareRepeaterEntry contentWithProperties2RepeaterEntry2 = contentWithProperties2Repeater.addEntry();
        contentWithProperties2RepeaterEntry2.setValue("string", "content 2 - entry 2");
        contentWithProperties2RepeaterEntry2.setValue("stringmult", new String[] {"content 2 - entry 2 - value 1"});
        ModifiableModelAwareRepeaterEntry contentWithProperties2RepeaterEntry3 = contentWithProperties2Repeater.addEntry();
        contentWithProperties2RepeaterEntry3.setValue("string", "content 2 - entry 3");
        contentWithProperties2RepeaterEntry3.setValue("stringmult", new String[] {"content 2 - entry 3 - value 1", "content 2 - entry 3 - value 2"});
        
        // POINTING CONTENT AND LINKS
        _pointingContent = _createContent("pointingContent", __CONTENT_TYPE_POINTING_TYPE_WITH_PROPERTIES);
        _pointingContent.setValue("linkContentWithProperties", _contentWithProperties2);
        _pointingContent.getComposite("composite", true).setValue("linkContentWithProperties", _contentWithProperties1);
    }
    
    private ModifiableContent _createContent(String title, String... contentTypeIds)
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        
        ModifiableContent content = root.createChild(title, NT_DEFAULT_CONTENT);

        content.setTypes(contentTypeIds);
        content.setTitle(title, Locale.FRANCE);
        content.setLanguage("fr");
        content.setCreationDate(__CREATION_DATE);
        content.setCreator(new UserIdentity("john", "anyPopulation"));
        content.setLastModified(__MODIFICATION_DATE1);
        content.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        
        return content;
    }
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }
    
    /**
     * Tests the {@link Content#dataToJSON(String)} method giving the path to an attribute
     * @throws Exception if an error occurs
     */
    @Test
    public void testAttributeToJSON() throws Exception
    {
        Object json = _contentWithProperties1.dataToJSON("composite/repeater[2]/string");
        Assert.assertEquals("content 1 - entry 2", json);
    }
    
    /**
     * Tests the {@link Content#dataToJSON(String)} method giving the path to a property
     * @throws Exception if an error occurs
     */
    @Test
    public void testSystemPropertyToJSON() throws Exception
    {
        @SuppressWarnings("unchecked")
        Map<String, Object> json = (Map<String, Object>) _contentWithProperties1.dataToJSON("contentLanguage");
        
        Assert.assertEquals(3, json.size());
        Assert.assertEquals("fr", json.get("code"));
        Assert.assertTrue(json.containsKey("icon"));
        Assert.assertTrue(json.containsKey("label"));
    }
    
    /**
     * Tests the {@link Content#dataToJSON(String)} method giving the path to a single property
     * Check that the multiple metadata is false
     * @throws Exception if an error occurs
     */
    @Test
    public void testSinglePropertyToJSON() throws Exception
    {
        Object json = _contentWithProperties1.dataToJSON("property-title");
        Assert.assertEquals("contentWithProperties1", json);
    }
    
    /**
     * Tests the {@link Content#dataToJSON(String)} method giving the path to a property
     * @throws Exception if an error occurs
     */
    @Test
    public void testPropertyToJSON() throws Exception
    {
        Assert.assertEquals(List.of("content 1 - entry 1", "content 1 - entry 2"), _contentWithProperties1.dataToJSON("property-string"));
    }
    
    /**
     * Tests the {@link Content#dataToJSON(String)} method giving the path to a property on a linked content
     * @throws Exception if an error occurs
     */
    @Test
    public void testPropertyToJSONFromLinkedContent() throws Exception
    {
        @SuppressWarnings("unchecked")
        List<Object> json = (List<Object>) _pointingContent.dataToJSON("linkContentWithProperties/property-string");

        Assert.assertEquals(3, json.size());
        Assert.assertEquals("content 2 - entry 1", json.get(0));
        Assert.assertEquals("content 2 - entry 2", json.get(1));
        Assert.assertEquals("content 2 - entry 3", json.get(2));
    }
    
    /**
     * Tests the {@link Content#dataToJSON()} method
     * SAX events should be generated for all attributes and properties 
     * @throws Exception if an error occurs
     */
    @Test
    @SuppressWarnings("unchecked")
    public void testToJSON() throws Exception
    {
        Map<String, Object> json = _contentWithProperties1.dataToJSON();
        Assert.assertEquals(8, json.size());
        
        // PROPERTIES
        
        // property-title
        Assert.assertEquals("contentWithProperties1", json.get("property-title"));
        
        // property-long
        Assert.assertEquals(45L, json.get("property-long"));
        
        // property-longmult
        List<Object> propertyLongmult = (List<Object>) json.get("property-longmult");
        Assert.assertEquals(3, propertyLongmult.size());
        Assert.assertEquals(45L, propertyLongmult.get(0));
        Assert.assertEquals(18L, propertyLongmult.get(1));
        Assert.assertEquals(97L, propertyLongmult.get(2));
        
        // property-string
        List<Object> propertyString = (List<Object>) json.get("property-string");
        Assert.assertEquals(2, propertyString.size());
        Assert.assertEquals("content 1 - entry 1", propertyString.get(0));
        Assert.assertEquals("content 1 - entry 2", propertyString.get(1));

        // property-stringmult
        List<Object> propertyStringmult = (List<Object>) json.get("property-stringmult");
        Assert.assertEquals(4, propertyStringmult.size());
        Assert.assertEquals("content 1 - entry 1 - value 1", propertyStringmult.get(0));
        Assert.assertEquals("content 1 - entry 1 - value 2", propertyStringmult.get(1));
        Assert.assertEquals("content 1 - entry 1 - value 3", propertyStringmult.get(2));
        Assert.assertEquals("content 1 - entry 2 - value 1", propertyStringmult.get(3));
        
        // property-longmult-sum
        Assert.assertEquals(160L, json.get("property-longmult-sum"));
        
        // ATTRIBUTES
        
        // title
        Assert.assertEquals("contentWithProperties1", json.get("title"));
        
        // composite
        Map<String, Object> composite = (Map<String, Object>) json.get("composite");
        Assert.assertEquals(3, composite.size());

        // composite > long
        Assert.assertEquals(45L, composite.get("long"));
        
        // composite > longmult
        List<Object> compositeLongmult = (List<Object>) composite.get("longmult");
        Assert.assertEquals(3, compositeLongmult.size());
        Assert.assertEquals(45L, compositeLongmult.get(0));
        Assert.assertEquals(18L, compositeLongmult.get(1));
        Assert.assertEquals(97L, compositeLongmult.get(2));
        
        // composite > repeater
        Map<String, Object> repeater = (Map<String, Object>) composite.get("repeater");
        Assert.assertEquals(3, repeater.size());
        Assert.assertEquals(new I18nizableText("Repeater"), repeater.get("label"));
        Assert.assertEquals(2, repeater.get("entryCount"));
        List<Map<String, Object>> entries = (List<Map<String, Object>>) repeater.get("entries");
        Assert.assertEquals(2, entries.size());
        
        // composite > repeater[1]
        Map<String, Object> entry1 = entries.get(0);
        Assert.assertEquals(2, entry1.size());
        Assert.assertEquals(1, entry1.get("position"));
        Map<String, Object> entry1Values = (Map<String, Object>) entry1.get("values");
        Assert.assertEquals(2, entry1Values.size());
        
        // composite > repeater[1] > string
        Assert.assertEquals("content 1 - entry 1", entry1Values.get("string"));

        // composite > repeater[1] > stringmult
        List<Object> entry1Stringmult = (List<Object>) entry1Values.get("stringmult");
        Assert.assertEquals(3, entry1Stringmult.size());
        Assert.assertEquals("content 1 - entry 1 - value 1", entry1Stringmult.get(0));
        Assert.assertEquals("content 1 - entry 1 - value 2", entry1Stringmult.get(1));
        Assert.assertEquals("content 1 - entry 1 - value 3", entry1Stringmult.get(2));
        
        // composite > repeater[2]
        Map<String, Object> entry2 = entries.get(1);
        Assert.assertEquals(2, entry2.size());
        Assert.assertEquals(2, entry2.get("position"));
        Map<String, Object> entry2Values = (Map<String, Object>) entry2.get("values");
        Assert.assertEquals(2, entry2Values.size());
        
        // composite > repeater[2] > string
        Assert.assertEquals("content 1 - entry 2", entry2Values.get("string"));
        
        // composite > repeater[2] > stringmult
        List<Object> entry2Stringmult = (List<Object>) entry2Values.get("stringmult");
        Assert.assertEquals(1, entry2Stringmult.size());
        Assert.assertEquals("content 1 - entry 2 - value 1", entry2Stringmult.get(0));
    }
    
    /**
     * Tests the {@link Content#dataToJSON(ViewItemAccessor)} method
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testViewToJSON() throws Exception
    {
        View view = _contentTypesHelper.getView("withBothAttributesAndProperties", _contentWithProperties1);
        Map<String, Object> json = _contentWithProperties1.dataToJSON(view);
        Assert.assertEquals(7, json.size());
        
        // PROPERTIES
        
        // property-long
        Assert.assertEquals(45L, json.get("property-long"));
        
        // property-string
        List<Object> propertyString = (List<Object>) json.get("property-string");
        Assert.assertEquals(2, propertyString.size());
        Assert.assertEquals("content 1 - entry 1", propertyString.get(0));
        Assert.assertEquals("content 1 - entry 2", propertyString.get(1));

        // property-longmult-sum
        Assert.assertEquals(160L, json.get("property-longmult-sum"));
        
        // contentLanguage
        Map<String, Object> language = (Map<String, Object>) json.get("contentLanguage");
        Assert.assertEquals(3, language.size());
        Assert.assertEquals("fr", language.get("code"));
        Assert.assertTrue(language.containsKey("icon"));
        Assert.assertTrue(language.containsKey("label"));
        
        // lastModified
        Assert.assertEquals("2022-08-06T11:10:17.000Z", json.get("lastModified"));
        
        // ATTRIBUTES
        
        // title
        Assert.assertEquals("contentWithProperties1", json.get("title"));
        
        // composite
        Map<String, Object> composite = (Map<String, Object>) json.get("composite");
        Assert.assertEquals(2, composite.size());

        // composite > long
        Assert.assertEquals(45L, composite.get("long"));
        
        // composite > repeater
        Map<String, Object> repeater = (Map<String, Object>) composite.get("repeater");
        Assert.assertEquals(3, repeater.size());
        Assert.assertEquals(new I18nizableText("Repeater"), repeater.get("label"));
        Assert.assertEquals(2, repeater.get("entryCount"));
        List<Map<String, Object>> entries = (List<Map<String, Object>>) repeater.get("entries");
        Assert.assertEquals(2, entries.size());
        
        // composite > repeater[1]
        Map<String, Object> entry1 = entries.get(0);
        Assert.assertEquals(2, entry1.size());
        Assert.assertEquals(1, entry1.get("position"));
        Map<String, Object> entry1Values = (Map<String, Object>) entry1.get("values");
        Assert.assertEquals(1, entry1Values.size());
        
        // composite > repeater[1] > stringmult
        List<Object> entry1Stringmult = (List<Object>) entry1Values.get("stringmult");
        Assert.assertEquals(3, entry1Stringmult.size());
        Assert.assertEquals("content 1 - entry 1 - value 1", entry1Stringmult.get(0));
        Assert.assertEquals("content 1 - entry 1 - value 2", entry1Stringmult.get(1));
        Assert.assertEquals("content 1 - entry 1 - value 3", entry1Stringmult.get(2));
        
        // composite > repeater[2]
        Map<String, Object> entry2 = entries.get(1);
        Assert.assertEquals(2, entry2.size());
        Assert.assertEquals(2, entry2.get("position"));
        Map<String, Object> entry2Values = (Map<String, Object>) entry2.get("values");
        Assert.assertEquals(1, entry2Values.size());
        
        // composite > repeater[2] > stringmult
        List<Object> entry2Stringmult = (List<Object>) entry2Values.get("stringmult");
        Assert.assertEquals(1, entry2Stringmult.size());
        Assert.assertEquals("content 1 - entry 2 - value 1", entry2Stringmult.get(0));
    }
    
    /**
     * Tests the {@link Content#dataToJSON(ViewItemAccessor)} method with properties pointed by linked content
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testViewToJSONWithLinkedContentsProperties() throws Exception
    {
        View view = _contentTypesHelper.getView("withBothAttributesAndProperties", _pointingContent);
        Map<String, Object> json = _pointingContent.dataToJSON(view);
        Assert.assertEquals(2, json.size());
        
        // title
        Assert.assertEquals("pointingContent", json.get("title"));
        
        // composite
        Map<String, Object> composite = (Map<String, Object>) json.get("composite");
        Assert.assertEquals(1, composite.size());
        
        // composite > linked content with properties
        Map<String, Object> linkedContent = (Map<String, Object>) composite.get("linkContentWithProperties");
        Assert.assertEquals(13, linkedContent.size());
        Assert.assertEquals(_contentWithProperties1.getId(), linkedContent.get("id"));
        Assert.assertEquals("contentWithProperties1", linkedContent.get("title"));
        
        // composite > linked content with properties > property-long
        Assert.assertEquals(45L, linkedContent.get("property-long"));
        
        // composite > linked content with properties > composite
        Map<String, Object> linkedContentComposite = (Map<String, Object>) linkedContent.get("composite");
        Assert.assertEquals(1, linkedContentComposite.size());
        
        // composite > linked content with properties > composite > longmult
        List<Object> linkedContentLongmult = (List<Object>) linkedContentComposite.get("longmult");
        Assert.assertEquals(3, linkedContentLongmult.size());
        Assert.assertEquals(45L, linkedContentLongmult.get(0));
        Assert.assertEquals(18L, linkedContentLongmult.get(1));
        Assert.assertEquals(97L, linkedContentLongmult.get(2));
        
        // composite > linked content with properties > contributor
        Map<String, Object> contributor = (Map<String, Object>) linkedContent.get("contributor");
        Assert.assertEquals("user1", contributor.get("login"));
        Assert.assertEquals(POPULATION_TEST_ID, contributor.get("populationId"));
        Assert.assertEquals("User One", contributor.get("fullname"));
    }
}
