/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.data.ContentValue;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeaterEntry;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link Content} test cases for all dataToMap methods
 */
@NeedFeatures({
    "repository/resolver",
    "workflow/workflow-provider",
    "content.properties/content-type-properties-tests",
    "cms/factories",
    "cms/workflow.content",
    "cms/org.ametys.cms.data.holder.helpers"
})
public class ContentDataToMapTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    /** Default content node type name. */
    public static final String NT_DEFAULT_CONTENT = "ametys:defaultContent";
    
    /** Ametys object resolver to subclasses. */
    protected static AmetysObjectResolver _ametysObjectResolver;
    /** Content types helper */
    protected static ContentTypesHelper _contentTypesHelper;
    
    private static final String __CONTENT_TYPE_POINTING_TYPE_WITH_PROPERTIES = "tests.pointing.type.with.properties";
    private static final String __CONTENT_TYPE_WITH_PROPERTIES = "tests.with.properties";
    
    private static final ZonedDateTime __CREATION_DATE = ZonedDateTime.of(2022, 7, 1, 16, 11, 18, 0, ZoneOffset.UTC);
    private static final ZonedDateTime __MODIFICATION_DATE1 = ZonedDateTime.of(2022, 7, 30, 22, 30, 38, 0, ZoneOffset.UTC);
    private static final ZonedDateTime __MODIFICATION_DATE2 = ZonedDateTime.of(2022, 8, 6, 11, 10, 17, 0, ZoneOffset.UTC);
    private static final ZonedDateTime __VALIDATION_DATE = ZonedDateTime.of(2022, 8, 13, 12, 29, 59, 0, ZoneOffset.UTC);
    
    private static ModifiableContent _pointingContent;
    private static ModifiableContent _contentWithProperties1;
    private static ModifiableContent _contentWithProperties2;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();

        // CONTENTS WITH PROPERTIES
        _contentWithProperties1 = _createContent("contentWithProperties1", __CONTENT_TYPE_WITH_PROPERTIES);
        _contentWithProperties1.setLastContributor(new UserIdentity("user1", POPULATION_TEST_ID));
        _contentWithProperties1.setLastModified(__MODIFICATION_DATE2);
        _contentWithProperties1.setLastValidationDate(__VALIDATION_DATE);
        ModifiableModelAwareComposite contentWithProperties1Composite = _contentWithProperties1.getComposite("composite", true);
        contentWithProperties1Composite.setValue("long", 45L);
        contentWithProperties1Composite.setValue("longmult", new Long[] {45L, 18L, 97L}); // sum: 160
        ModifiableModelAwareRepeater contentWithProperties1Repeater = contentWithProperties1Composite.getRepeater("repeater", true);
        ModifiableModelAwareRepeaterEntry contentWithProperties1RepeaterEntry1 = contentWithProperties1Repeater.addEntry();
        contentWithProperties1RepeaterEntry1.setValue("string", "content 1 - entry 1");
        contentWithProperties1RepeaterEntry1.setValue("stringmult", new String[] {"content 1 - entry 1 - value 1", "content 1 - entry 1 - value 2", "content 1 - entry 1 - value 3"});
        ModifiableModelAwareRepeaterEntry contentWithProperties1RepeaterEntry2 = contentWithProperties1Repeater.addEntry();
        contentWithProperties1RepeaterEntry2.setValue("string", "content 1 - entry 2");
        contentWithProperties1RepeaterEntry2.setValue("stringmult", new String[] {"content 1 - entry 2 - value 1"});
        
        _contentWithProperties2 = _createContent("contentWithProperties2", __CONTENT_TYPE_WITH_PROPERTIES);
        ModifiableModelAwareComposite contentWithProperties2Composite = _contentWithProperties2.getComposite("composite", true);
        contentWithProperties2Composite.setValue("longmult", new Long[] {12L, 38L}); // sum: 50
        ModifiableModelAwareRepeater contentWithProperties2Repeater = contentWithProperties2Composite.getRepeater("repeater", true);
        ModifiableModelAwareRepeaterEntry contentWithProperties2RepeaterEntry1 = contentWithProperties2Repeater.addEntry();
        contentWithProperties2RepeaterEntry1.setValue("string", "content 2 - entry 1");
        contentWithProperties2RepeaterEntry1.setValue("stringmult", new String[] {"content 2 - entry 1 - value 1", "content 2 - entry 1 - value 2", "content 2 - entry 1 - value 3"});
        ModifiableModelAwareRepeaterEntry contentWithProperties2RepeaterEntry2 = contentWithProperties2Repeater.addEntry();
        contentWithProperties2RepeaterEntry2.setValue("string", "content 2 - entry 2");
        contentWithProperties2RepeaterEntry2.setValue("stringmult", new String[] {"content 2 - entry 2 - value 1"});
        ModifiableModelAwareRepeaterEntry contentWithProperties2RepeaterEntry3 = contentWithProperties2Repeater.addEntry();
        contentWithProperties2RepeaterEntry3.setValue("string", "content 2 - entry 3");
        contentWithProperties2RepeaterEntry3.setValue("stringmult", new String[] {"content 2 - entry 3 - value 1", "content 2 - entry 3 - value 2"});
        
        // POINTING CONTENT AND LINKS
        _pointingContent = _createContent("pointingContent", __CONTENT_TYPE_POINTING_TYPE_WITH_PROPERTIES);
        _pointingContent.setValue("linkContentsWithProperties", new Content[] {_contentWithProperties1, _contentWithProperties2});
        _pointingContent.getComposite("composite", true).setValue("linkContentWithProperties", _contentWithProperties1);
    }
    
    private ModifiableContent _createContent(String title, String... contentTypeIds)
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        
        ModifiableContent content = root.createChild(title, NT_DEFAULT_CONTENT);

        content.setTypes(contentTypeIds);
        content.setTitle(title, Locale.FRANCE);
        content.setLanguage("fr");
        content.setCreationDate(__CREATION_DATE);
        content.setCreator(new UserIdentity("john", "anyPopulation"));
        content.setLastModified(__MODIFICATION_DATE1);
        content.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        
        return content;
    }
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }
    
    /**
     * Tests the {@link Content#dataToMap()} method
     * The {@link Map} should contain all attributes and properties 
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testToMap() throws Exception
    {
        Map<String, Object> map = _contentWithProperties1.dataToMap();
        Assert.assertEquals(8, map.size());
        
        // PROPERTIES
        
        // property-title
        Assert.assertTrue(map.containsKey("property-title"));
        Assert.assertEquals("contentWithProperties1", map.get("property-title"));
        
        // property-long
        Assert.assertTrue(map.containsKey("property-long"));
        Assert.assertEquals(45L, map.get("property-long"));
        
        // property-longmult
        Assert.assertTrue(map.containsKey("property-longmult"));
        Assert.assertArrayEquals(new Long[] {45L, 18L, 97L}, (Long[]) map.get("property-longmult"));
        
        // property-string
        Assert.assertTrue(map.containsKey("property-string"));
        Assert.assertArrayEquals(new String[] {"content 1 - entry 1", "content 1 - entry 2"}, (String[]) map.get("property-string"));

        // property-stringmult
        Assert.assertTrue(map.containsKey("property-stringmult"));
        Assert.assertArrayEquals(new String[] {"content 1 - entry 1 - value 1",
                                               "content 1 - entry 1 - value 2",
                                               "content 1 - entry 1 - value 3",
                                               "content 1 - entry 2 - value 1"}, (String[]) map.get("property-stringmult"));
        
        // property-longmult-sum
        Assert.assertTrue(map.containsKey("property-longmult-sum"));
        Assert.assertEquals(160L, map.get("property-longmult-sum"));
        
        // ATTRIBUTES

        // title
        Assert.assertTrue(map.containsKey("title"));
        Assert.assertEquals("contentWithProperties1", map.get("title"));
        
        // composite
        Assert.assertTrue(map.containsKey("composite"));
        Map<String, Object> composite = (Map<String, Object>) map.get("composite");
        Assert.assertEquals(3, composite.size());

        // composite > long
        Assert.assertTrue(composite.containsKey("long"));
        Assert.assertEquals(45L, composite.get("long"));
        
        // composite > longmult
        Assert.assertTrue(composite.containsKey("longmult"));
        Assert.assertArrayEquals(new Long[] {45L, 18L, 97L}, (Long[]) composite.get("longmult"));
        
        // composite > repeater
        Assert.assertTrue(composite.containsKey("repeater"));
        List<Map<String, Object>> repeater = (List<Map<String, Object>>) composite.get("repeater");
        Assert.assertEquals(2, repeater.size());
        
        // composite > repeater[1]
        Map<String, Object> entry1 = repeater.get(0);
        Assert.assertEquals(2, entry1.size());
        
        // composite > repeater[1] > string
        Assert.assertTrue(entry1.containsKey("string"));
        Assert.assertEquals("content 1 - entry 1", entry1.get("string"));

        // composite > repeater[1] > stringmult
        Assert.assertTrue(entry1.containsKey("stringmult"));
        Assert.assertArrayEquals(new String[] {"content 1 - entry 1 - value 1",
                                               "content 1 - entry 1 - value 2",
                                               "content 1 - entry 1 - value 3"}, (String[]) entry1.get("stringmult"));
        
        // composite > repeater[2]
        Map<String, Object> entry2 = repeater.get(1);
        Assert.assertEquals(2, entry2.size());
        
        // composite > repeater[2] > string
        Assert.assertTrue(entry2.containsKey("string"));
        Assert.assertEquals("content 1 - entry 2", entry2.get("string"));
        
        // composite > repeater[2] > stringmult
        Assert.assertTrue(entry2.containsKey("stringmult"));
        Assert.assertArrayEquals(new String[] {"content 1 - entry 2 - value 1"}, (String[]) entry2.get("stringmult"));
    }
    
    /**
     * Tests the {@link Content#dataToMap(ViewItemAccessor)} method
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testViewToMap() throws Exception
    {
        View view = _contentTypesHelper.getView("withBothAttributesAndProperties", _contentWithProperties1);
        Map<String, Object> map = _contentWithProperties1.dataToMap(view);
        Assert.assertEquals(7, map.size());
        
        // PROPERTIES
        
        // property-long
        Assert.assertTrue(map.containsKey("property-long"));
        Assert.assertEquals(45L, map.get("property-long"));
        
        // property-string
        Assert.assertTrue(map.containsKey("property-string"));
        Assert.assertArrayEquals(new String[] {"content 1 - entry 1", "content 1 - entry 2"}, (String[]) map.get("property-string"));

        // property-longmult-sum
        Assert.assertTrue(map.containsKey("property-longmult-sum"));
        Assert.assertEquals(160L, map.get("property-longmult-sum"));
        
        // contentLanguage
        Assert.assertTrue(map.containsKey("contentLanguage"));
        Assert.assertEquals("fr", map.get("contentLanguage"));
        
        // lastModified
        Assert.assertTrue(map.containsKey("lastModified"));
        Assert.assertEquals(__MODIFICATION_DATE2, map.get("lastModified"));
        
        // ATTRIBUTES

        // title
        Assert.assertTrue(map.containsKey("title"));
        Assert.assertEquals("contentWithProperties1", map.get("title"));
        
        // composite
        Assert.assertTrue(map.containsKey("composite"));
        Map<String, Object> composite = (Map<String, Object>) map.get("composite");
        Assert.assertEquals(2, composite.size());

        // composite > long
        Assert.assertTrue(composite.containsKey("long"));
        Assert.assertEquals(45L, composite.get("long"));
        
        // composite > repeater
        Assert.assertTrue(composite.containsKey("repeater"));
        List<Map<String, Object>> repeater = (List<Map<String, Object>>) composite.get("repeater");
        Assert.assertEquals(2, repeater.size());
        
        // composite > repeater[1]
        Map<String, Object> entry1 = repeater.get(0);
        Assert.assertEquals(1, entry1.size());
        
        // composite > repeater[1] > stringmult
        Assert.assertTrue(entry1.containsKey("stringmult"));
        Assert.assertArrayEquals(new String[] {"content 1 - entry 1 - value 1",
                                               "content 1 - entry 1 - value 2",
                                               "content 1 - entry 1 - value 3"}, (String[]) entry1.get("stringmult"));
        
        // composite > repeater[2]
        Map<String, Object> entry2 = repeater.get(1);
        Assert.assertEquals(1, entry2.size());
        
        // composite > repeater[2] > stringmult
        Assert.assertTrue(entry2.containsKey("stringmult"));
        Assert.assertArrayEquals(new String[] {"content 1 - entry 2 - value 1"}, (String[]) entry2.get("stringmult"));
    }
    
    /**
     * Tests the {@link Content#dataToMap(ViewItemAccessor)} method with properties pointed by linked content
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testViewToMapWithLinkedContentsProperties() throws Exception
    {
        View view = _contentTypesHelper.getView("withBothAttributesAndProperties", _pointingContent);
        Map<String, Object> map = _pointingContent.dataToMap(view);
        
        // title
        Assert.assertTrue(map.containsKey("title"));
        Assert.assertEquals("pointingContent", map.get("title"));
        
        // composite
        Assert.assertTrue(map.containsKey("composite"));
        Map<String, Object> composite = (Map<String, Object>) map.get("composite");
        Assert.assertEquals(1, composite.size());
        
        // composite > linked content with properties
        ContentValue linkContentWithProperties = (ContentValue) composite.get("linkContentWithProperties");
        Assert.assertEquals("contentWithProperties1", linkContentWithProperties.getValue("title"));
    }
}
