/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.Locale;
import java.util.Optional;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.TransformerHandler;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.xpath.XPathAPI;
import org.junit.Assert;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.ContentHandler;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeaterEntry;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link Content} test cases for all dataToSAX methods
 */
@NeedFeatures({
    "repository/resolver",
    "workflow/workflow-provider",
    "content.properties/content-type-properties-tests",
    "cms/factories",
    "cms/workflow.content",
    "cms/org.ametys.cms.data.holder.helpers"
})
public class ContentDataToSAXTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    /** Default content node type name. */
    public static final String NT_DEFAULT_CONTENT = "ametys:defaultContent";
    
    /** Ametys object resolver to subclasses. */
    protected static AmetysObjectResolver _ametysObjectResolver;
    /** Content types helper */
    protected static ContentTypesHelper _contentTypesHelper;
    
    private static final String __CONTENT_TYPE_POINTING_TYPE_WITH_PROPERTIES = "tests.pointing.type.with.properties";
    private static final String __CONTENT_TYPE_WITH_PROPERTIES = "tests.with.properties";
    
    private static final ZonedDateTime __CREATION_DATE = ZonedDateTime.of(2022, 7, 1, 16, 11, 18, 0, ZoneOffset.UTC);
    private static final ZonedDateTime __MODIFICATION_DATE1 = ZonedDateTime.of(2022, 7, 30, 22, 30, 38, 0, ZoneOffset.UTC);
    private static final ZonedDateTime __MODIFICATION_DATE2 = ZonedDateTime.of(2022, 8, 6, 11, 10, 17, 0, ZoneOffset.UTC);
    private static final ZonedDateTime __VALIDATION_DATE = ZonedDateTime.of(2022, 8, 13, 12, 29, 59, 0, ZoneOffset.UTC);
    
    private static ModifiableContent _pointingContent;
    private static ModifiableContent _contentWithProperties1;
    private static ModifiableContent _contentWithProperties2;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();

        // CONTENTS WITH PROPERTIES
        _contentWithProperties1 = _createContent("contentWithProperties1", __CONTENT_TYPE_WITH_PROPERTIES);
        _contentWithProperties1.setLastContributor(new UserIdentity("user1", POPULATION_TEST_ID));
        _contentWithProperties1.setLastModified(__MODIFICATION_DATE2);
        _contentWithProperties1.setLastValidationDate(__VALIDATION_DATE);
        ModifiableModelAwareComposite contentWithProperties1Composite = _contentWithProperties1.getComposite("composite", true);
        contentWithProperties1Composite.setValue("long", 45L);
        contentWithProperties1Composite.setValue("longmult", new Long[] {45L, 18L, 97L}); // sum: 160
        ModifiableModelAwareRepeater contentWithProperties1Repeater = contentWithProperties1Composite.getRepeater("repeater", true);
        ModifiableModelAwareRepeaterEntry contentWithProperties1RepeaterEntry1 = contentWithProperties1Repeater.addEntry();
        contentWithProperties1RepeaterEntry1.setValue("string", "content 1 - entry 1");
        contentWithProperties1RepeaterEntry1.setValue("stringmult", new String[] {"content 1 - entry 1 - value 1", "content 1 - entry 1 - value 2", "content 1 - entry 1 - value 3"});
        ModifiableModelAwareRepeaterEntry contentWithProperties1RepeaterEntry2 = contentWithProperties1Repeater.addEntry();
        contentWithProperties1RepeaterEntry2.setValue("string", "content 1 - entry 2");
        contentWithProperties1RepeaterEntry2.setValue("stringmult", new String[] {"content 1 - entry 2 - value 1"});
        
        _contentWithProperties2 = _createContent("contentWithProperties2", __CONTENT_TYPE_WITH_PROPERTIES);
        ModifiableModelAwareComposite contentWithProperties2Composite = _contentWithProperties2.getComposite("composite", true);
        contentWithProperties2Composite.setValue("longmult", new Long[] {12L, 38L}); // sum: 50
        ModifiableModelAwareRepeater contentWithProperties2Repeater = contentWithProperties2Composite.getRepeater("repeater", true);
        ModifiableModelAwareRepeaterEntry contentWithProperties2RepeaterEntry1 = contentWithProperties2Repeater.addEntry();
        contentWithProperties2RepeaterEntry1.setValue("string", "content 2 - entry 1");
        contentWithProperties2RepeaterEntry1.setValue("stringmult", new String[] {"content 2 - entry 1 - value 1", "content 2 - entry 1 - value 2", "content 2 - entry 1 - value 3"});
        ModifiableModelAwareRepeaterEntry contentWithProperties2RepeaterEntry2 = contentWithProperties2Repeater.addEntry();
        contentWithProperties2RepeaterEntry2.setValue("string", "content 2 - entry 2");
        contentWithProperties2RepeaterEntry2.setValue("stringmult", new String[] {"content 2 - entry 2 - value 1"});
        ModifiableModelAwareRepeaterEntry contentWithProperties2RepeaterEntry3 = contentWithProperties2Repeater.addEntry();
        contentWithProperties2RepeaterEntry3.setValue("string", "content 2 - entry 3");
        contentWithProperties2RepeaterEntry3.setValue("stringmult", new String[] {"content 2 - entry 3 - value 1", "content 2 - entry 3 - value 2"});
        
        // POINTING CONTENT AND LINKS
        _pointingContent = _createContent("pointingContent", __CONTENT_TYPE_POINTING_TYPE_WITH_PROPERTIES);
        _pointingContent.setValue("linkContentWithProperties", _contentWithProperties2);
        _pointingContent.getComposite("composite", true).setValue("linkContentWithProperties", _contentWithProperties1);
    }
    
    private ModifiableContent _createContent(String title, String... contentTypeIds)
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        
        ModifiableContent content = root.createChild(title, NT_DEFAULT_CONTENT);

        content.setTypes(contentTypeIds);
        content.setTitle(title, Locale.FRANCE);
        content.setLanguage("fr");
        content.setCreationDate(__CREATION_DATE);
        content.setCreator(new UserIdentity("john", "anyPopulation"));
        content.setLastModified(__MODIFICATION_DATE1);
        content.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        
        return content;
    }
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }
    
    /**
     * Tests the {@link Content#dataToSAX(ContentHandler, String)} method giving the path to an attribute
     * @throws Exception if an error occurs
     */
    @Test
    public void testAttributeToSAX() throws Exception
    {
        Document document = _generateSAXEvents(_contentWithProperties1, Optional.of("composite/repeater[2]/string"));
        
        _assertXpathEquals(document, "count(/result/string)", 1);
        _assertXpathEquals(document, "/result/string/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/string/@multiple", "false");
        _assertXpathEquals(document, "/result/string", "content 1 - entry 2");
    }
    
    /**
     * Tests the {@link Content#dataToSAX(ContentHandler, String)} method giving the path to a property
     * @throws Exception if an error occurs
     */
    @Test
    public void testSystemPropertyToSAX() throws Exception
    {
        Document document = _generateSAXEvents(_contentWithProperties1, Optional.of("contentLanguage"));
        
        _assertXpathEquals(document, "count(/result/contentLanguage)", 1);
        _assertXpathEquals(document, "/result/contentLanguage/@code", "fr");
    }
    
    /**
     * Tests the {@link Content#dataToSAX(ContentHandler, String)} method giving the path to a single property
     * Check that the multiple metadata is false
     * @throws Exception if an error occurs
     */
    @Test
    public void testSinglePropertyToSAX() throws Exception
    {
        Document document = _generateSAXEvents(_contentWithProperties1, Optional.of("property-title"));
        
        _assertXpathEquals(document, "count(/result/property-title)", 1);
        
        _assertXpathEquals(document, "/result/property-title/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-title/@multiple", "false");
        _assertXpathEquals(document, "/result/property-title", "contentWithProperties1");
    }
    
    /**
     * Tests the {@link Content#dataToSAX(ContentHandler, String)} method giving the path to a property
     * @throws Exception if an error occurs
     */
    @Test
    public void testPropertyToSAX() throws Exception
    {
        Document document = _generateSAXEvents(_contentWithProperties1, Optional.of("property-string"));
        
        _assertXpathEquals(document, "count(/result/property-string)", 2);
        
        _assertXpathEquals(document, "/result/property-string[1]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-string[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-string[1]", "content 1 - entry 1");
        
        _assertXpathEquals(document, "/result/property-string[2]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-string[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-string[2]", "content 1 - entry 2");
    }
    
    /**
     * Tests the {@link Content#dataToSAX(ContentHandler, String)} method giving the path to a property on a linked content
     * @throws Exception if an error occurs
     */
    @Test
    public void testPropertyToSAXFromLinkedContent() throws Exception
    {
        Document document = _generateSAXEvents(_pointingContent, Optional.of("linkContentWithProperties/property-string"));
        
        _assertXpathEquals(document, "count(/result/property-string)", 3);
        
        _assertXpathEquals(document, "/result/property-string[1]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-string[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-string[1]", "content 2 - entry 1");
        
        _assertXpathEquals(document, "/result/property-string[2]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-string[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-string[2]", "content 2 - entry 2");

        _assertXpathEquals(document, "/result/property-string[3]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-string[3]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-string[3]", "content 2 - entry 3");
    }
    
    /**
     * Tests the {@link Content#dataToSAX(ContentHandler)} method
     * SAX events should be generated for all attributes and properties 
     * @throws Exception if an error occurs
     */
    @Test
    public void testToSAX() throws Exception
    {
        Document document = _generateSAXEvents(_contentWithProperties1, Optional.empty());
        
        // PROPERTIES
        
        // property-title
        _assertXpathEquals(document, "count(/result/property-title)", 1);
        _assertXpathEquals(document, "/result/property-title/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-title/@multiple", "false");
        _assertXpathEquals(document, "/result/property-title", "contentWithProperties1");
        
        // property-long
        _assertXpathEquals(document, "count(/result/property-long)", 1);
        _assertXpathEquals(document, "/result/property-long/@typeId", ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/property-long/@multiple", "false");
        _assertXpathEquals(document, "/result/property-long", 45);
        
        // property-longmult
        _assertXpathEquals(document, "count(/result/property-longmult)", 3);
        _assertXpathEquals(document, "/result/property-longmult[1]/@typeId", ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/property-longmult[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-longmult[1]", 45);
        _assertXpathEquals(document, "/result/property-longmult[2]/@typeId", ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/property-longmult[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-longmult[2]", 18);
        _assertXpathEquals(document, "/result/property-longmult[3]/@typeId", ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/property-longmult[3]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-longmult[3]", 97);
        
        // property-string
        _assertXpathEquals(document, "count(/result/property-string)", 2);
        _assertXpathEquals(document, "/result/property-string[1]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-string[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-string[1]", "content 1 - entry 1");
        _assertXpathEquals(document, "/result/property-string[2]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-string[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-string[2]", "content 1 - entry 2");

        // property-stringmult
        _assertXpathEquals(document, "count(/result/property-stringmult)", 4);
        _assertXpathEquals(document, "/result/property-stringmult[1]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-stringmult[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-stringmult[1]", "content 1 - entry 1 - value 1");
        _assertXpathEquals(document, "/result/property-stringmult[2]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-stringmult[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-stringmult[2]", "content 1 - entry 1 - value 2");
        _assertXpathEquals(document, "/result/property-stringmult[3]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-stringmult[3]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-stringmult[3]", "content 1 - entry 1 - value 3");
        _assertXpathEquals(document, "/result/property-stringmult[4]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-stringmult[4]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-stringmult[4]", "content 1 - entry 2 - value 1");
        
        // property-longmult-sum
        _assertXpathEquals(document, "count(/result/property-longmult-sum)", 1);
        _assertXpathEquals(document, "/result/property-longmult-sum/@typeId", ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/property-longmult-sum/@multiple", "false");
        _assertXpathEquals(document, "/result/property-longmult-sum", 160);
        
        // ATTRIBUTES
        // title
        _assertXpathEquals(document, "count(/result/title)", 1);
        _assertXpathEquals(document, "/result/title/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/title/@multiple", "false");
        _assertXpathEquals(document, "/result/title", "contentWithProperties1");
        
        // composite
        _assertXpathEquals(document, "count(/result/composite)", 1);

        // composite > long
        _assertXpathEquals(document, "count(/result/composite/long)", 1);
        _assertXpathEquals(document, "/result/composite/long/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/long/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/long", 45);
        
        // composite > longmult
        _assertXpathEquals(document, "count(/result/composite/longmult)", 3);
        _assertXpathEquals(document, "/result/composite/longmult[1]/@typeId", ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/longmult[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/longmult[1]", 45);
        _assertXpathEquals(document, "/result/composite/longmult[2]/@typeId", ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/longmult[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/longmult[2]", 18);
        _assertXpathEquals(document, "/result/composite/longmult[3]/@typeId", ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/longmult[3]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/longmult[3]", 97);
        
        // composite > repeater
        _assertXpathEquals(document, "count(/result/composite/repeater)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/@entryCount", 2);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry)", 2);
        
        // composite > repeater[1] > string
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/string)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/string/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/string/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/string", "content 1 - entry 1");

        // composite > repeater[1] > stringmult
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/stringmult)", 3);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[1]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[1]", "content 1 - entry 1 - value 1");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[2]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[2]", "content 1 - entry 1 - value 2");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[3]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[3]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[3]", "content 1 - entry 1 - value 3");
        
        // composite > repeater[2] > string
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/string)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/string/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/string/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/string", "content 1 - entry 2");
        
        // composite > repeater[2] > stringmult
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/stringmult)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/stringmult/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/stringmult/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/stringmult", "content 1 - entry 2 - value 1");
    }
    
    /**
     * Tests the {@link Content#dataToSAX(ContentHandler, ViewItemAccessor)} method
     * @throws Exception if an error occurs
     */
    @Test
    public void testViewToSAX() throws Exception
    {
        View view = _contentTypesHelper.getView("withBothAttributesAndProperties", _contentWithProperties1);
        Document document = _generateSAXEvents(_contentWithProperties1, Optional.of(view));
        
        // PROPERTIES
        
        // property-long
        _assertXpathEquals(document, "count(/result/property-long)", 1);
        _assertXpathEquals(document, "/result/property-long/@typeId", ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/property-long/@multiple", "false");
        _assertXpathEquals(document, "/result/property-long", 45);
        
        // property-string
        _assertXpathEquals(document, "count(/result/property-string)", 2);
        _assertXpathEquals(document, "/result/property-string[1]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-string[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-string[1]", "content 1 - entry 1");
        _assertXpathEquals(document, "/result/property-string[2]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/property-string[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/property-string[2]", "content 1 - entry 2");

        // property-longmult-sum
        _assertXpathEquals(document, "count(/result/property-longmult-sum)", 1);
        _assertXpathEquals(document, "/result/property-longmult-sum/@typeId", ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/property-longmult-sum/@multiple", "false");
        _assertXpathEquals(document, "/result/property-longmult-sum", 160);
        
        // contentLanguage
        _assertXpathEquals(document, "count(/result/contentLanguage)", 1);
        _assertXpathEquals(document, "/result/contentLanguage/@code", "fr");
        
        // lastModified
        _assertXpathEquals(document, "count(/result/lastModified)", 1);
        _assertXpathEquals(document, "/result/lastModified/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.DATETIME_TYPE_ID);
        _assertXpathEquals(document, "/result/lastModified/@multiple", "false");
        _assertXpathEquals(document, "/result/lastModified", "2022-08-06T11:10:17.000Z");
        
        // ATTRIBUTES
        // title
        _assertXpathEquals(document, "count(/result/title)", 1);
        _assertXpathEquals(document, "/result/title/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/title/@multiple", "false");
        _assertXpathEquals(document, "/result/title", "contentWithProperties1");
        
        // composite
        _assertXpathEquals(document, "count(/result/composite)", 1);

        // composite > long
        _assertXpathEquals(document, "count(/result/composite/long)", 1);
        _assertXpathEquals(document, "/result/composite/long/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/long/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/long", 45);
        
        // composite > repeater
        _assertXpathEquals(document, "count(/result/composite/repeater)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/@entryCount", 2);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry)", 2);
        
        // composite > repeater[1] > stringmult
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/stringmult)", 3);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[1]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[1]", "content 1 - entry 1 - value 1");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[2]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[2]", "content 1 - entry 1 - value 2");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[3]/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[3]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/stringmult[3]", "content 1 - entry 1 - value 3");
        
        // composite > repeater[2] > stringmult
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/stringmult)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/stringmult/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/stringmult/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/stringmult", "content 1 - entry 2 - value 1");
    }
    
    /**
     * Tests the {@link Content#dataToSAX(ContentHandler, ViewItemAccessor)} method with properties pointed by linked content
     * @throws Exception if an error occurs
     */
    @Test
    public void testViewToSAXWithLinkedContentsProperties() throws Exception
    {
        View view = _contentTypesHelper.getView("withBothAttributesAndProperties", _pointingContent);
        Document document = _generateSAXEvents(_pointingContent, Optional.of(view));
        
        // title
        _assertXpathEquals(document, "count(/result/title)", 1);
        _assertXpathEquals(document, "/result/title/@typeId", ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/title/@multiple", "false");
        _assertXpathEquals(document, "/result/title", "pointingContent");
        
        // composite
        _assertXpathEquals(document, "count(/result/composite)", 1);
        
        // composite > linked content with properties
        _assertXpathEquals(document, "count(/result/composite/linkContentWithProperties)", 1);
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/@id", _contentWithProperties1.getId());
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/@title", "contentWithProperties1");

        // composite > linked content with properties > property-long
        _assertXpathEquals(document, "count(/result/composite/linkContentWithProperties/property-long)", 1);
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/property-long/@typeId", ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/property-long/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/property-long", 45);
        
        // composite > linked content with properties > composite
        _assertXpathEquals(document, "count(/result/composite/linkContentWithProperties/composite)", 1);
        
        // composite > linked content with properties > composite > longmult
        _assertXpathEquals(document, "count(/result/composite/linkContentWithProperties/composite/longmult)", 3);
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/composite/longmult[1]/@typeId", ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/composite/longmult[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/composite/longmult[1]", 45);
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/composite/longmult[2]/@typeId", ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/composite/longmult[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/composite/longmult[2]", 18);
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/composite/longmult[3]/@typeId", ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/composite/longmult[3]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/composite/longmult[3]", 97);
        
        // composite > linked content with properties > contributor
        _assertXpathEquals(document, "count(/result/composite/linkContentWithProperties/contributor)", 1);
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/contributor/@typeId", org.ametys.cms.data.type.ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/contributor/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/contributor/@login", "user1");
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/contributor/@populationId", POPULATION_TEST_ID);
        _assertXpathEquals(document, "/result/composite/linkContentWithProperties/contributor", "User One");
    }
    
    private Document _generateSAXEvents(ModelAwareDataHolder dataHolder, Optional<Object> dataToSAXOption) throws Exception
    {
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        if (dataToSAXOption.isEmpty())
        {
            dataHolder.dataToSAX(handler);
        }
        else
        {
            Object option = dataToSAXOption.get();
            if (option instanceof String dataPath)
            {
                dataHolder.dataToSAX(handler, dataPath);
            }
            else if (option instanceof View view)
            {
                dataHolder.dataToSAX(handler, view);
            }
        }
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        return document;
    }
    
    private void _assertXpathEquals(Node node, String expression, String value) throws Exception
    {
        Assert.assertEquals(value, XPathAPI.eval(node, expression).str());
    }
    
    private void _assertXpathEquals(Node node, String expression, long value) throws Exception
    {
        Assert.assertEquals(value, (long) XPathAPI.eval(node, expression).num());
    }
}
