/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.time.LocalDate;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.TransformerHandler;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.xpath.XPathAPI;
import org.junit.Assert;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.Reference;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableDefaultContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.cms.utils.ResourceElementTypeHelperForTests;
import org.ametys.cms.workflow.ContentWorkflowHelper;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeaterEntry;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.metadata.MultilingualString;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

import com.opensymphony.workflow.WorkflowException;

/**
 * Tests for the {@link ContentGenerator}
 */
@NeedFeatures({
    "cms/workflow",
    "cms/workflow.content",
    "cms/content-types.helper",
    "test/content-type-view-tests",
    "test/test.user.provider",
    "test/sitemap.component.dependencies", // for sitemap.xmap of plugin test
})
public class ContentGeneratorTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __EXHAUSTIVE_CONTENTTYPE = "tests.extends.exhaustive";

    private static final String __CONTENT_GENERATION_URI =  "_plugins/test/content-generation.xml";
    private static final Map<String, String> __EMPTY_REQUEST_HEADERS = new HashMap<>();
    
    private static final ZonedDateTime __COMMENT_DATE = ZonedDateTime.of(2020, 07, 07, 10, 37, 59, 120_000_000, ZoneId.systemDefault());
    private static final String __COMMENT_AUTHOR = "mbellamy";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ContentWorkflowHelper _contentWorkflowHelper;
    
    private static Content _exhaustiveContent;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentWorkflowHelper = (ContentWorkflowHelper) manager.lookup(ContentWorkflowHelper.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        _exhaustiveContent = _createExhaustiveContent();
    }
    
    /**
     * Test case for the {@link ContentGenerator}
     * @throws Exception if an error occurs 
     */
    @Test
    public void testContentGeneratorWithMainView() throws Exception
    {
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        Map<String, Object> requestAttributes = new HashMap<>();
        requestAttributes.put(Content.class.getName(), _exhaustiveContent);
        
        getCocoonWrapper().processURI(__CONTENT_GENERATION_URI, handler, Map.of(), requestAttributes, __EMPTY_REQUEST_HEADERS);
        
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _doAssertionsOnContentGeneralInfos(document);
        _doAssertionsOnMainView(document, false);
        _assertThereIsNoComment(document);
    }
    
    /**
     * Test case for the {@link ContentGenerator}
     * @throws Exception if an error occurs 
     */
    @Test
    public void testContentGeneratorWithView() throws Exception
    {
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        Map<String, String> requestParameters = new HashMap<>();
        requestParameters.put("viewName", "details");
        
        Map<String, Object> requestAttributes = new HashMap<>();
        requestAttributes.put(Content.class.getName(), _exhaustiveContent);
        
        getCocoonWrapper().processURI(__CONTENT_GENERATION_URI, handler, requestParameters, requestAttributes, __EMPTY_REQUEST_HEADERS);
        
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _doAssertionsOnContentGeneralInfos(document);
        _doAssertionsOnDetailsView(document);
        _assertThereIsNoComment(document);
    }
    
    /**
     * Test case for the {@link ContentGenerator}
     * @throws Exception if an error occurs 
     */
    @Test
    public void testContentGeneratorWithFallbackView() throws Exception
    {
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        Map<String, String> requestParameters = new HashMap<>();
        requestParameters.put("viewName", "inexistingView");
        requestParameters.put("fallbackViewName", "details");
        
        Map<String, Object> requestAttributes = new HashMap<>();
        requestAttributes.put(Content.class.getName(), _exhaustiveContent);
        
        getCocoonWrapper().processURI(__CONTENT_GENERATION_URI, handler, requestParameters, requestAttributes, __EMPTY_REQUEST_HEADERS);
        
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _doAssertionsOnContentGeneralInfos(document);
        _doAssertionsOnDetailsView(document);
        _assertThereIsNoComment(document);
    }
    
    /**
     * Test case for the {@link ContentGenerator}
     * @throws Exception if an error occurs 
     */
    @Test
    public void testContentGeneratorInEdition() throws Exception
    {
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        Map<String, String> requestParameters = new HashMap<>();
        requestParameters.put("isEdition", "true");
        
        Map<String, Object> requestAttributes = new HashMap<>();
        requestAttributes.put(Content.class.getName(), _exhaustiveContent);
        
        getCocoonWrapper().processURI(__CONTENT_GENERATION_URI, handler, requestParameters, requestAttributes, __EMPTY_REQUEST_HEADERS);
        
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _doAssertionsOnContentGeneralInfos(document);
        _doAssertionsOnMainView(document, true);
        _doAssertionsOnComments(document);
    }
    
    private void _doAssertionsOnContentGeneralInfos(Document document) throws Exception
    {
        _assertXpathEquals(document, "count(/content)", 1);
    }
    
    private void _doAssertionsOnMainView(Document document, boolean isEdition) throws Exception
    {
        _assertXpathEquals(document, "count(/content/metadata)", 1);
        // In edition mode, multiple values using JSON are not duplicated
        _assertXpathEquals(document, "count(/content/metadata/*)", isEdition ? 37 : 42);
        
        // Tests just on some attributes to ensure this is the main view that is used to SAX attributes
        _assertXpathEquals(document, "count(/content/metadata/title)", 1);
        _assertXpathEquals(document, "count(/content/metadata/abstract)", 1);
        _assertXpathEquals(document, "count(/content/metadata/file)", 1);
        _assertXpathEquals(document, "count(/content/metadata/attachments)", 1);
        _assertXpathEquals(document, "count(/content/metadata/stringmult)", 3);
        _assertXpathEquals(document, "count(/content/metadata/long)", 1);
    }
    
    private void _doAssertionsOnDetailsView(Document document) throws Exception
    {
        _assertXpathEquals(document, "count(/content/metadata)", 1);
        _assertXpathEquals(document, "count(/content/metadata/*)", 1);
        _assertXpathEquals(document, "count(/content/metadata/title)", 1);
    }
    
    private void _assertThereIsNoComment(Document document) throws Exception
    {
        _assertXpathEquals(document, "count(/content/comments)", 0);
    }
    
    private void _doAssertionsOnComments(Document document) throws Exception
    {
        _assertXpathEquals(document, "count(/content/comments)", 1);
        _assertXpathEquals(document, "count(/content/comments/*)", 3);
        _assertXpathEquals(document, "count(/content/comments/metadata)", 3);
        _assertXpathEquals(document, "/content/comments/metadata[1]/@path", "address");
        _assertXpathEquals(document, "/content/comments/metadata[2]/@path", "composite/repeater[2]/comp-owns");
        _assertXpathEquals(document, "/content/comments/metadata[3]/@path", "string");
    }
    
    /**
     * Create an exhaustive content with values for each type of attributes
     * @return the created content
     * @throws Exception If an error occurs.
     */
    private static Content _createExhaustiveContent() throws Exception
    {
        ModifiableDefaultContent content = _createContent("Test content", "content", __EXHAUSTIVE_CONTENTTYPE);
        
        // Simple attribute types
        content.setValue("abstract", "Test abstract");
        content.setValue("string", "string value");
        _addCommentsToData(content.getRepositoryData(), "string", "comment on string");
        content.setValue("stringmult", new String[] {"one", "two", "three"});
        content.setValue("long", 3L);
        content.setValue("longmult", new Long[] {5L, 6L, 12L});
        content.setValue("double", 1.234D);
        content.setValue("doublemult", new Double[] {3.14D, 6.12D});
        content.setValue("boolean", true);
        content.setValue("booleanmult", new Boolean[] {true, false, true, true});
        
        // Date attribute types
        content.setValue("date", LocalDate.of(2007, 10, 9));
        content.setValue("datemult", new LocalDate[] {LocalDate.of(2020, 10, 9), LocalDate.of(2020, 06, 18)});
        content.setValue("datetime", ZonedDateTime.of(2011, 11, 24, 10, 15, 30, 596, ZoneId.systemDefault()));
        content.setValue("datetimemult", new ZonedDateTime[] {ZonedDateTime.of(2020, 03, 16, 10, 15, 30, 154, ZoneId.systemDefault()), ZonedDateTime.of(2020, 05, 11, 10, 15, 30, 145, ZoneId.systemDefault())});
        
        // Complex attribute types
        content.setValue("user", USER1_TEST_IDENTITY);
        content.setValue("usermult", new UserIdentity[] {USER2_TEST_IDENTITY, USER3_TEST_IDENTITY});
        content.setValue("reference", new Reference("http://www.ametys.org"));
        content.setValue("referencemult", new Reference[] {new Reference("http://www.ametys.org")});
        content.setValue("gps", new Geocode(44.481, 3.208));
        content.setValue("gpsmult", new Geocode[] {new Geocode(44.481, 3.208), new Geocode(50.12, 3.458)});
        content.setValue("multilingualstring", new MultilingualString(Map.of(Locale.FRANCE, "valeur", Locale.UK, "value")));
        content.setValue("multilingualstringmult", new MultilingualString[] {new MultilingualString(Map.of(Locale.FRANCE, "valeur", Locale.UK, "value"))});
        
        // Resource attribute types
        content.setValue("binary", ResourceElementTypeHelperForTests.createBinary("my binary content"));
        content.setValue("file", ResourceElementTypeHelperForTests.createResource("myFile.txt", "my file content", _ametysObjectResolver));
        // CMS-9834 remove the rich text content because it is not well formatted for edition transformation
//        content.setValue("content", ResourceElementTypeHelperForTests.createRichText("my rich text content"));
        
        // Content attribute types
        ModifiableDefaultContent link1 = _createContent("Linked content 1", "linked-content-1", __EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link2 = _createContent("Linked content 2", "linked-content-2", __EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link3 = _createContent("Linked content 3", "linked-content-3", __EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link4 = _createContent("Linked content 4", "linked-content-4", __EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link11 = _createContent("Linked content 11", "linked-content-11", __EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link21 = _createContent("Linked content 21", "linked-content-21", __EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link22 = _createContent("Linked content 22", "linked-content-22", __EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link31 = _createContent("Linked content 31", "linked-content-31", __EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link221 = _createContent("Linked content 221", "linked-content-221", __EXHAUSTIVE_CONTENTTYPE);
        ModifiableDefaultContent link222 = _createContent("Linked content 222", "linked-content-222", __EXHAUSTIVE_CONTENTTYPE);
        
        content.setValue("linkContent", link1);
        content.setValue("linkContents", new Content[] {link2, link3, link4});
        content.setValue("owns", new Content[] {link2, link3, link4});
        
        // link1 -> link11
        link1.setValue("linkContent", link11);
        // link2 -> link21, link22
        link2.setValue("linkContents", new Content[] {link21, link22});
        // link3 -> link31
        link3.setValue("linkContent", link31);
        // link22 -> link221
        link22.setValue("linkContent", link221);
        // Same links with "owns" metadata.
        link1.setValue("owns", new Content[] {link11});
        link2.setValue("owns", new Content[] {link21, link22});
        link3.setValue("owns", new Content[] {link31});
        link22.setValue("owns", new Content[] {link221});
        
        // Composite and repeater attribute types
        content.getComposite("address", true).setValue("country", "france");
        _addCommentsToData(content.getRepositoryData(), "address", "comment on address composite");
        ModifiableModelAwareRepeater repeater1 = content.getComposite("composite", true).getRepeater("repeater", true);
        repeater1.addEntry();
        repeater1.getEntry(1).setValue("comp-string", "entry 1");
        repeater1.getEntry(1).setValue("comp-owns", link1);
        repeater1.addEntry();
        repeater1.getEntry(2).setValue("comp-string", "entry 2");
        repeater1.getEntry(2).setValue("comp-owns", link2);
        _addCommentsToData(repeater1.getEntry(2).getRepositoryData(), "comp-owns", "comment on a repeater entry data");
        
        link1.getComposite("address", true).setValue("country", "france");
        link2.getComposite("address", true).setValue("country", "japan");
        link3.getComposite("address", true).setValue("country", "spain");
        link4.getComposite("address", true).setValue("country", "japan");
        
        link11.getComposite("address", true).setValue("country", "spain");
        link21.getComposite("address", true).setValue("country", "france");
        link22.getComposite("address", true).setValue("country", "japan");
        link31.getComposite("address", true).setValue("country", "italy");
        link221.getComposite("address", true).setValue("country", "belgium");
        
        ModifiableModelAwareRepeater repeater2 = link22.getComposite("composite", true).getRepeater("repeater", true);
        repeater2.addEntry();
        repeater2.getEntry(1).setValue("comp-string", "entry 1");
        repeater2.getEntry(1).setValue("comp-owns", link221);
        repeater2.addEntry();
        repeater2.getEntry(2).setValue("comp-string", "entry 2");
        repeater2.getEntry(2).setValue("comp-owns", link222);
        
        // Resources attributes type in a repeater
        ModifiableModelAwareRepeater repeater = content.getRepeater("attachments", true);
        ModifiableModelAwareRepeaterEntry attachmentsEntry = repeater.addEntry();
        attachmentsEntry.setValue("attachment", ResourceElementTypeHelperForTests.createBinary("my binary in a repeater entry"));
        attachmentsEntry.setValue("attachment-text", "my attachment text");
        // CMS-9834 remove the rich text content because it is not well formatted for edition transformation
//        attachmentsEntry.setValue("attachment-desc", ResourceElementTypeHelperForTests.createRichText("my rich text attachment desc"));
        
        return content;
    }
    
    private static void _addCommentsToData(ModifiableRepositoryData repositoryData, String dataName, String... comments)
    {
        ModifiableRepositoryData commentsData = repositoryData.addRepositoryData(dataName + ModelAwareDataHolder.COMMENTS_SUFFIX, RepositoryConstants.COMPOSITE_METADTA_NODETYPE, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        
        int commentId = 1;
        for (String comment : comments)
        {
            ModifiableRepositoryData commentData = commentsData.addRepositoryData(String.valueOf(commentId++), RepositoryConstants.COMPOSITE_METADTA_NODETYPE, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
            
            commentData.setValue("comment", comment);
            commentData.setValue("date", DateUtils.asCalendar(__COMMENT_DATE));
            commentData.setValue("author", __COMMENT_AUTHOR);
        }
    }
    
    /**
     * Create a content.
     * @param title The content title.
     * @param name The content name.
     * @param type The content type.
     * @return The created content.
     * @throws WorkflowException If an error occurs.
     */
    private static ModifiableDefaultContent _createContent(String title, String name, String type) throws WorkflowException
    {
        Map<String, Object> results = _contentWorkflowHelper.createContent("content", 1, name, title, new String[] {type}, new String[0], "en");
        
        return (ModifiableDefaultContent) results.get(Content.class.getName());
    }
    
    private void _assertXpathEquals(Node node, String expression, String value) throws Exception
    {
        Assert.assertEquals(value, XPathAPI.eval(node, expression).str());
    }
    
    private void _assertXpathEquals(Node node, String expression, long value) throws Exception
    {
        Assert.assertEquals(value, (long) XPathAPI.eval(node, expression).num());
    }
}
