/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.io.OutputStream;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.data.RichText;
import org.ametys.cms.model.properties.SemanticAnnotationSystemProperty;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.DefaultContent;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeaterEntry;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link Content} test cases for properties
 */
@NeedFeatures({
    "repository/resolver",
    "workflow/workflow-provider",
    "content.properties/content-type-properties-tests",
    "content.properties/content-type-semantic-annotation-system-property-tests",
    "cms/factories",
    "cms/workflow.content",
    "cms/org.ametys.cms.data.holder.helpers"
})
public class ContentPropertiesTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    /** Default content node type name. */
    public static final String NT_DEFAULT_CONTENT = "ametys:defaultContent";
    
    /** Ametys object resolver to subclasses. */
    protected static AmetysObjectResolver _ametysObjectResolver;
    
    private static final String __CONTENT_TYPE_POINTING_TYPE_WITH_PROPERTIES = "tests.pointing.type.with.properties";
    private static final String __CONTENT_TYPE_WITH_PROPERTIES = "tests.with.properties";
    
    private static final ZonedDateTime __CREATION_DATE = ZonedDateTime.of(2022, 7, 1, 16, 11, 18, 0, ZoneOffset.UTC);
    private static final ZonedDateTime __MODIFICATION_DATE1 = ZonedDateTime.of(2022, 7, 30, 22, 30, 38, 0, ZoneOffset.UTC);
    private static final ZonedDateTime __MODIFICATION_DATE2 = ZonedDateTime.of(2022, 8, 6, 11, 10, 17, 0, ZoneOffset.UTC);
    private static final ZonedDateTime __MODIFICATION_DATE3 = ZonedDateTime.of(2022, 8, 6, 15, 30, 45, 0, ZoneOffset.UTC);
    private static final ZonedDateTime __VALIDATION_DATE = ZonedDateTime.of(2022, 8, 13, 12, 29, 59, 0, ZoneOffset.UTC);
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }
    
    /**
     * Tests the access to system properties from a {@link Content}
     * @throws Exception if an error occurs while creating the contents
     */
    @Test
    public void testSystemProperties() throws Exception
    {
        Content pointingContent = _createContentsWithProperties().get(0);
        
        // contributor
        Assert.assertTrue(pointingContent.hasValue(DefaultContent.METADATA_CONTRIBUTOR));
        UserIdentity contributor = pointingContent.getValue(DefaultContent.METADATA_CONTRIBUTOR);
        Assert.assertEquals("brian", contributor.getLogin());
        Assert.assertEquals("anyPopulation", contributor.getPopulationId());
        Assert.assertTrue(pointingContent.hasValue("linkContent/" + DefaultContent.METADATA_CONTRIBUTOR));
        contributor = pointingContent.getValue("linkContent/" + DefaultContent.METADATA_CONTRIBUTOR);
        Assert.assertEquals("enaour", contributor.getLogin());
        Assert.assertEquals("wallace", contributor.getPopulationId());
        
        // modification date
        Assert.assertTrue(pointingContent.hasValue(DefaultContent.METADATA_MODIFIED));
        Assert.assertEquals(__MODIFICATION_DATE1, pointingContent.getValue(DefaultContent.METADATA_MODIFIED));
        Assert.assertTrue(pointingContent.hasValue("linkContent/" + DefaultContent.METADATA_MODIFIED));
        Assert.assertEquals(__MODIFICATION_DATE2, pointingContent.getValue("linkContent/" + DefaultContent.METADATA_MODIFIED));
        Assert.assertArrayEquals(new ZonedDateTime[] {__MODIFICATION_DATE2, __MODIFICATION_DATE3}, pointingContent.getValue("linkContentsWithProperties/" + DefaultContent.METADATA_MODIFIED, true));
        
        // last validation date
        Assert.assertFalse(pointingContent.hasValue("lastValidation"));
        Assert.assertNull(pointingContent.getValue("lastValidation"));
        Assert.assertTrue(pointingContent.hasValue("linkContent/lastValidation"));
        Assert.assertEquals(__VALIDATION_DATE, pointingContent.getValue("linkContent/lastValidation"));
    }
    
    /**
     * Test the access of a property referencing a non multiple attribute, at the root of the content type
     * @throws Exception if an error occurs while creating the contents
     */
    @Test
    public void testSingleProperty() throws Exception
    {
        List<Content> contentsWithProperties = _createContentsWithProperties();
        Content pointingContent = contentsWithProperties.get(0);
        Content contentWithProperties1 = contentsWithProperties.get(1);
        
        Assert.assertTrue(contentWithProperties1.hasValue("property-title"));
        Assert.assertEquals("contentWithProperties1", contentWithProperties1.getValue("property-title"));
        
        try
        {
            Assert.assertFalse(pointingContent.hasValue("linkContent/property-title"));
            pointingContent.getValue("linkContent/property-title");
            Assert.fail("The attribute linkContent does not refer to the type with properties, so the property 'property-title'");
        }
        catch (UndefinedItemPathException e)
        {
            // Nothing to do
        }
        
        Assert.assertTrue(pointingContent.hasValue("linkContentWithProperties/property-title"));
        Assert.assertEquals("contentWithProperties1", pointingContent.getValue("linkContentWithProperties/property-title"));
        
        Assert.assertArrayEquals(new String[] {"contentWithProperties1", "contentWithProperties2"}, pointingContent.getValue("linkContentsWithProperties/property-title", true));
        
        Assert.assertTrue(pointingContent.hasValue("composite/linkContentWithProperties/property-title"));
        Assert.assertEquals("contentWithProperties2", pointingContent.getValue("composite/linkContentWithProperties/property-title"));
        
        Assert.assertArrayEquals(new String[] {"contentWithProperties1", "contentWithProperties3"}, pointingContent.getValue("composite/linkContentsWithProperties/property-title", true));
    }
    
    /**
     * Test the access of a property referencing a non multiple attribute, in a composite
     * @throws Exception if an error occurs while creating the contents
     */
    @Test
    public void testSinglePropertyInComposite() throws Exception
    {
        List<Content> contentsWithProperties = _createContentsWithProperties();
        Content pointingContent = contentsWithProperties.get(0);
        Content contentWithProperties1 = contentsWithProperties.get(1);
        
        Assert.assertTrue(contentWithProperties1.hasValue("property-long"));
        Assert.assertEquals(Long.valueOf(45), contentWithProperties1.getValue("property-long"));
        
        Assert.assertTrue(pointingContent.hasValue("linkContentWithProperties/property-long"));
        Assert.assertEquals(Long.valueOf(45), pointingContent.getValue("linkContentWithProperties/property-long"));
        
        Assert.assertArrayEquals(new Long[] {45L, 3L}, pointingContent.getValue("linkContentsWithProperties/property-long", true));
        
        Assert.assertTrue(pointingContent.hasValue("composite/linkContentWithProperties/property-long"));
        Assert.assertEquals(Long.valueOf(3), pointingContent.getValue("composite/linkContentWithProperties/property-long"));
        
        Assert.assertArrayEquals(new Long[] {45L, 32L}, pointingContent.getValue("composite/linkContentsWithProperties/property-long", true));
    }
    
    /**
     * Test the access of a property referencing a multiple attribute, in a composite
     * @throws Exception if an error occurs while creating the contents
     */
    @Test
    public void testMultiplePropertyInComposite() throws Exception
    {
        List<Content> contentsWithProperties = _createContentsWithProperties();
        Content pointingContent = contentsWithProperties.get(0);
        Content contentWithProperties1 = contentsWithProperties.get(1);
        
        Assert.assertTrue(contentWithProperties1.hasValue("property-longmult"));
        Assert.assertArrayEquals(new Long[] {45L, 18L, 97L}, contentWithProperties1.getValue("property-longmult"));
        
        Assert.assertTrue(pointingContent.hasValue("linkContentWithProperties/property-longmult"));
        Assert.assertArrayEquals(new Long[] {45L, 18L, 97L}, pointingContent.getValue("linkContentWithProperties/property-longmult"));
        
        Assert.assertArrayEquals(new Long[] {45L, 18L, 97L, 12L, 38L}, pointingContent.getValue("linkContentsWithProperties/property-longmult", true));
        
        Assert.assertTrue(pointingContent.hasValue("composite/linkContentWithProperties/property-longmult"));
        Assert.assertArrayEquals(new Long[] {12L, 38L}, pointingContent.getValue("composite/linkContentWithProperties/property-longmult"));
        
        Assert.assertArrayEquals(new Long[] {45L, 18L, 97L, 15L, 248L, 74L, 9874L}, pointingContent.getValue("composite/linkContentsWithProperties/property-longmult", true));
    }
    
    /**
     * Test the access of a property referencing a non multiple attribute, in a repeater
     * @throws Exception if an error occurs while creating the contents
     */
    @Test
    public void testSinglePropertyInRepeater() throws Exception
    {
        List<Content> contentsWithProperties = _createContentsWithProperties();
        Content pointingContent = contentsWithProperties.get(0);
        Content contentWithProperties1 = contentsWithProperties.get(1);
        
        Assert.assertTrue(contentWithProperties1.hasValue("property-string"));
        Assert.assertArrayEquals(new String[] {"content 1 - entry 1", "content 1 - entry 2"}, contentWithProperties1.getValue("property-string"));
        
        Assert.assertTrue(pointingContent.hasValue("linkContentWithProperties/property-string"));
        Assert.assertArrayEquals(new String[] {"content 1 - entry 1", "content 1 - entry 2"}, pointingContent.getValue("linkContentWithProperties/property-string"));
        
        Assert.assertArrayEquals(new String[] {"content 1 - entry 1", "content 1 - entry 2", "content 2 - entry 1", "content 2 - entry 2", "content 2 - entry 3"}, pointingContent.getValue("linkContentsWithProperties/property-string", true));
        
        Assert.assertTrue(pointingContent.hasValue("composite/linkContentWithProperties/property-string"));
        Assert.assertArrayEquals(new String[] {"content 2 - entry 1", "content 2 - entry 2", "content 2 - entry 3"}, pointingContent.getValue("composite/linkContentWithProperties/property-string"));
        
        Assert.assertArrayEquals(new String[] {"content 1 - entry 1", "content 1 - entry 2", "content 3 - entry 1"}, pointingContent.getValue("composite/linkContentsWithProperties/property-string", true));
    }
    
    /**
     * Test the access of a property referencing a multiple attribute, in a repeater
     * @throws Exception if an error occurs while creating the contents
     */
    @Test
    public void testMultiplePropertyInRepeater() throws Exception
    {
        List<Content> contentsWithProperties = _createContentsWithProperties();
        Content pointingContent = contentsWithProperties.get(0);
        Content contentWithProperties1 = contentsWithProperties.get(1);
        
        Assert.assertTrue(contentWithProperties1.hasValue("property-stringmult"));
        Assert.assertArrayEquals(new String[] {"content 1 - entry 1 - value 1", "content 1 - entry 1 - value 2", "content 1 - entry 1 - value 3",
                                               "content 1 - entry 2 - value 1"}, contentWithProperties1.getValue("property-stringmult"));
        
        Assert.assertTrue(pointingContent.hasValue("linkContentWithProperties/property-stringmult"));
        Assert.assertArrayEquals(new String[] {"content 1 - entry 1 - value 1", "content 1 - entry 1 - value 2", "content 1 - entry 1 - value 3",
                                               "content 1 - entry 2 - value 1"}, pointingContent.getValue("linkContentWithProperties/property-stringmult"));
        
        Assert.assertArrayEquals(new String[] {"content 1 - entry 1 - value 1", "content 1 - entry 1 - value 2", "content 1 - entry 1 - value 3",
                                               "content 1 - entry 2 - value 1",
                                               "content 2 - entry 1 - value 1", "content 2 - entry 1 - value 2", "content 2 - entry 1 - value 3",
                                               "content 2 - entry 2 - value 1",
                                               "content 2 - entry 3 - value 1", "content 2 - entry 3 - value 2"}, pointingContent.getValue("linkContentsWithProperties/property-stringmult", true));
        
        Assert.assertTrue(pointingContent.hasValue("composite/linkContentWithProperties/property-stringmult"));
        Assert.assertArrayEquals(new String[] {"content 2 - entry 1 - value 1", "content 2 - entry 1 - value 2", "content 2 - entry 1 - value 3",
                                               "content 2 - entry 2 - value 1",
                                               "content 2 - entry 3 - value 1", "content 2 - entry 3 - value 2"}, pointingContent.getValue("composite/linkContentWithProperties/property-stringmult"));
        
        Assert.assertArrayEquals(new String[] {"content 1 - entry 1 - value 1", "content 1 - entry 1 - value 2", "content 1 - entry 1 - value 3",
                                               "content 1 - entry 2 - value 1",
                                               "content 3 - entry 1 - value 1", "content 3 - entry 1 - value 2"}, pointingContent.getValue("composite/linkContentsWithProperties/property-stringmult", true));
    }
    
    /**
     * Test the access of a property represented by a specific class
     * @throws Exception if an error occurs while creating the contents
     */
    @Test
    public void testCustomProperty() throws Exception
    {
        List<Content> contentsWithProperties = _createContentsWithProperties();
        Content pointingContent = contentsWithProperties.get(0);
        Content contentWithProperties1 = contentsWithProperties.get(1);
        
        Assert.assertTrue(contentWithProperties1.hasValue("property-longmult-sum"));
        Assert.assertEquals(Long.valueOf(160), contentWithProperties1.getValue("property-longmult-sum"));
        
        Assert.assertTrue(pointingContent.hasValue("linkContentWithProperties/property-longmult-sum"));
        Assert.assertEquals(Long.valueOf(160), pointingContent.getValue("linkContentWithProperties/property-longmult-sum"));
        
        Assert.assertArrayEquals(new Long[] {160L, 50L}, pointingContent.getValue("linkContentsWithProperties/property-longmult-sum", true));
        
        Assert.assertTrue(pointingContent.hasValue("composite/linkContentWithProperties/property-longmult-sum"));
        Assert.assertEquals(Long.valueOf(50), pointingContent.getValue("composite/linkContentWithProperties/property-longmult-sum"));
        
        Assert.assertArrayEquals(new Long[] {160L, 10211L}, pointingContent.getValue("composite/linkContentsWithProperties/property-longmult-sum", true));
    }
    
    private List<Content> _createContentsWithProperties() throws Exception
    {
        ModifiableContent contentWithProperties1 = _createContent("contentWithProperties1", __CONTENT_TYPE_WITH_PROPERTIES);
        contentWithProperties1.setLastContributor(new UserIdentity("enaour", "wallace"));
        contentWithProperties1.setLastModified(__MODIFICATION_DATE2);
        contentWithProperties1.setLastValidationDate(__VALIDATION_DATE);
        ModifiableModelAwareComposite contentWithProperties1Composite = contentWithProperties1.getComposite("composite", true);
        contentWithProperties1Composite.setValue("long", 45L);
        contentWithProperties1Composite.setValue("longmult", new Long[] {45L, 18L, 97L}); // sum: 160
        ModifiableModelAwareRepeater contentWithProperties1Repeater = contentWithProperties1Composite.getRepeater("repeater", true);
        ModifiableModelAwareRepeaterEntry contentWithProperties1RepeaterEntry1 = contentWithProperties1Repeater.addEntry();
        contentWithProperties1RepeaterEntry1.setValue("string", "content 1 - entry 1");
        contentWithProperties1RepeaterEntry1.setValue("stringmult", new String[] {"content 1 - entry 1 - value 1", "content 1 - entry 1 - value 2", "content 1 - entry 1 - value 3"});
        ModifiableModelAwareRepeaterEntry contentWithProperties1RepeaterEntry2 = contentWithProperties1Repeater.addEntry();
        contentWithProperties1RepeaterEntry2.setValue("string", "content 1 - entry 2");
        contentWithProperties1RepeaterEntry2.setValue("stringmult", new String[] {"content 1 - entry 2 - value 1"});
        
        ModifiableContent contentWithProperties2 = _createContent("contentWithProperties2", __CONTENT_TYPE_WITH_PROPERTIES);
        contentWithProperties2.setLastModified(__MODIFICATION_DATE3);
        ModifiableModelAwareComposite contentWithProperties2Composite = contentWithProperties2.getComposite("composite", true);
        contentWithProperties2Composite.setValue("long", 3L);
        contentWithProperties2Composite.setValue("longmult", new Long[] {12L, 38L}); // sum: 50
        ModifiableModelAwareRepeater contentWithProperties2Repeater = contentWithProperties2Composite.getRepeater("repeater", true);
        ModifiableModelAwareRepeaterEntry contentWithProperties2RepeaterEntry1 = contentWithProperties2Repeater.addEntry();
        contentWithProperties2RepeaterEntry1.setValue("string", "content 2 - entry 1");
        contentWithProperties2RepeaterEntry1.setValue("stringmult", new String[] {"content 2 - entry 1 - value 1", "content 2 - entry 1 - value 2", "content 2 - entry 1 - value 3"});
        ModifiableModelAwareRepeaterEntry contentWithProperties2RepeaterEntry2 = contentWithProperties2Repeater.addEntry();
        contentWithProperties2RepeaterEntry2.setValue("string", "content 2 - entry 2");
        contentWithProperties2RepeaterEntry2.setValue("stringmult", new String[] {"content 2 - entry 2 - value 1"});
        ModifiableModelAwareRepeaterEntry contentWithProperties2RepeaterEntry3 = contentWithProperties2Repeater.addEntry();
        contentWithProperties2RepeaterEntry3.setValue("string", "content 2 - entry 3");
        contentWithProperties2RepeaterEntry3.setValue("stringmult", new String[] {"content 2 - entry 3 - value 1", "content 2 - entry 3 - value 2"});
        
        ModifiableContent contentWithProperties3 = _createContent("contentWithProperties3", __CONTENT_TYPE_WITH_PROPERTIES);
        ModifiableModelAwareComposite contentWithProperties3Composite = contentWithProperties3.getComposite("composite", true);
        contentWithProperties3Composite.setValue("long", 32L);
        contentWithProperties3Composite.setValue("longmult", new Long[] {15L, 248L, 74L, 9874L}); // sum: 10211
        ModifiableModelAwareRepeater contentWithProperties3Repeater = contentWithProperties3Composite.getRepeater("repeater", true);
        ModifiableModelAwareRepeaterEntry contentWithProperties3RepeaterEntry1 = contentWithProperties3Repeater.addEntry();
        contentWithProperties3RepeaterEntry1.setValue("string", "content 3 - entry 1");
        contentWithProperties3RepeaterEntry1.setValue("stringmult", new String[] {"content 3 - entry 1 - value 1", "content 3 - entry 1 - value 2"});
        
        // POINTING CONTENT AND LINKS
        ModifiableContent pointingContent = _createContent("pointingContent", __CONTENT_TYPE_POINTING_TYPE_WITH_PROPERTIES);
        
        pointingContent.setValue("linkContent", contentWithProperties1);
        pointingContent.setValue("linkContentWithProperties", contentWithProperties1);
        pointingContent.setValue("linkContentsWithProperties", new Content[] {contentWithProperties1, contentWithProperties2});
        
        ModifiableModelAwareComposite composite = pointingContent.getComposite("composite", true);
        composite.setValue("linkContentWithProperties", contentWithProperties2);
        composite.setValue("linkContentsWithProperties", new Content[] {contentWithProperties1, contentWithProperties3});
        
        return List.of(pointingContent, contentWithProperties1, contentWithProperties2, contentWithProperties3);
    }
    
    /**
     * Test the access of the semantic annotation's values for a content of a content type with several parent content types, sharing the same annotations
     * @throws Exception if an error occurs while creating the content
     */
    @Test
    public void testSemanticAnnotationSystemPropertyWithContentTypeMultiParent() throws Exception
    {
        Content contentWithSemanticAnnotations = _createContentWithSemanticAnnotations();
        
        @SuppressWarnings("unchecked")
        Map<String, List<String>> semanticAnnotations = (Map<String, List<String>>) contentWithSemanticAnnotations.dataToJSON("semanticAnnotations");
        Assert.assertEquals(2, semanticAnnotations.size());
        
        List<String> semanticAnnotation1 = semanticAnnotations.get(SemanticAnnotationSystemProperty.ANNOTATION_PREFIX + "annotation1");
        Assert.assertEquals(3, semanticAnnotation1.size());
        Assert.assertTrue(semanticAnnotation1.contains("a1-1"));
        Assert.assertTrue(semanticAnnotation1.contains("a1-2"));
        Assert.assertTrue(semanticAnnotation1.contains("a1-3"));
        
        List<String> semanticAnnotation2 = semanticAnnotations.get(SemanticAnnotationSystemProperty.ANNOTATION_PREFIX + "annotation2");
        Assert.assertEquals(2, semanticAnnotation2.size());
        Assert.assertTrue(semanticAnnotation2.contains("a2-1"));
        Assert.assertTrue(semanticAnnotation2.contains("a2-2"));
    }
    
    private Content _createContentWithSemanticAnnotations() throws Exception
    {
        ModifiableContent content = _createContent("contentWithSemanticAnnotations", "tests.semantic.annotation.system.property.child");
     
        RichText richText1 = _createBasicRichText();
        richText1.addAnnotations("annotation1", "a1-1", "a1-2", "a1-3");
        richText1.addAnnotations("annotation2", "a2-1", "a2-2");
        content.setValue("richText1", richText1);
        
        return content;
    }
    
    /**
     * Test the access of the semantic annotation's values for a content with several content types, sharing the same annotations
     * @throws Exception if an error occurs while creating the content
     */
    @Test
    public void testSemanticAnnotationSystemPropertyWithContentMultiType() throws Exception
    {
        Content contentWithSemanticAnnotations = _createContentOfMultiTypesWithSemanticAnnotations();
        
        @SuppressWarnings("unchecked")
        Map<String, List<String>> semanticAnnotations = (Map<String, List<String>>) contentWithSemanticAnnotations.dataToJSON("semanticAnnotations");
        Assert.assertEquals(3, semanticAnnotations.size());
        
        List<String> semanticAnnotation1 = semanticAnnotations.get(SemanticAnnotationSystemProperty.ANNOTATION_PREFIX + "annotation1");
        Assert.assertEquals(6, semanticAnnotation1.size());
        Assert.assertTrue(semanticAnnotation1.contains("a1-1"));
        Assert.assertTrue(semanticAnnotation1.contains("a1-2"));
        Assert.assertTrue(semanticAnnotation1.contains("a1-3"));
        Assert.assertTrue(semanticAnnotation1.contains("a1-4"));
        Assert.assertTrue(semanticAnnotation1.contains("a1-5"));
        Assert.assertTrue(semanticAnnotation1.contains("a1-6"));
        
        List<String> semanticAnnotation2 = semanticAnnotations.get(SemanticAnnotationSystemProperty.ANNOTATION_PREFIX + "annotation4");
        Assert.assertEquals(3, semanticAnnotation2.size());
        Assert.assertTrue(semanticAnnotation2.contains("a4-1"));
        Assert.assertTrue(semanticAnnotation2.contains("a4-2"));
        Assert.assertTrue(semanticAnnotation2.contains("a4-3"));
        
        List<String> semanticAnnotation5 = semanticAnnotations.get(SemanticAnnotationSystemProperty.ANNOTATION_PREFIX + "annotation5");
        Assert.assertEquals(6, semanticAnnotation5.size());
        Assert.assertTrue(semanticAnnotation5.contains("a5-1"));
        Assert.assertTrue(semanticAnnotation5.contains("a5-2"));
        Assert.assertTrue(semanticAnnotation5.contains("a5-3"));
        Assert.assertTrue(semanticAnnotation5.contains("a5-4"));
        Assert.assertTrue(semanticAnnotation5.contains("a5-5"));
        Assert.assertTrue(semanticAnnotation5.contains("a5-6"));
    }
    
    private Content _createContentOfMultiTypesWithSemanticAnnotations() throws Exception
    {
        ModifiableContent content = _createContent("contentMultiTypesWithSemanticAnnotations", "tests.semantic.annotation.system.property.parent1", "tests.semantic.annotation.system.property.parent2");
     
        ModifiableModelAwareRepeater repeater = content.getRepeater("repeater", true);
        
        ModifiableModelAwareRepeaterEntry entry1 = repeater.addEntry();
        RichText entry1RichText3 = _createBasicRichText();
        entry1RichText3.addAnnotations("annotation1", "a1-1", "a1-2", "a1-3");
        entry1RichText3.addAnnotations("annotation4", "a4-1");
        entry1RichText3.addAnnotations("annotation5", "a5-1", "a5-2");
        entry1.setValue("richText3", entry1RichText3);
        
        ModifiableModelAwareRepeaterEntry entry2 = repeater.addEntry();
        RichText entry2RichText3 = _createBasicRichText();
        entry2RichText3.addAnnotations("annotation1", "a1-4", "a1-5");
        entry2RichText3.addAnnotations("annotation4", "a4-2", "a4-3");
        entry2RichText3.addAnnotations("annotation5", "a5-3");
        entry2.setValue("richText3", entry2RichText3);
        
        RichText richText4 = _createBasicRichText();
        richText4.addAnnotations("annotation1", "a1-6");
        richText4.addAnnotations("annotation5", "a5-4", "a5-5", "a5-6");
        content.setValue("richText4", richText4);
        
        return content;
    }
    
    private RichText _createBasicRichText() throws Exception
    {
        RichText richText = new RichText();
        try (OutputStream os = richText.getOutputStream())
        {
            os.write("my rich text content".getBytes());
        }
        richText.setEncoding("UTF-8");
        richText.setMimeType("text/xml");
        richText.setLastModificationDate(ZonedDateTime.of(2019, 03, 20, 13, 34, 12, 0, ZoneId.systemDefault()));
        
        return richText;
    }
    
    private ModifiableContent _createContent(String title, String... contentTypeIds)
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        
        ModifiableContent content = root.createChild(title, NT_DEFAULT_CONTENT);

        content.setTypes(contentTypeIds);
        content.setTitle(title, Locale.FRANCE);
        content.setLanguage("fr");
        content.setCreationDate(__CREATION_DATE);
        content.setCreator(new UserIdentity("john", "anyPopulation"));
        content.setLastModified(__MODIFICATION_DATE1);
        content.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        
        return content;
    }
}
