/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content;

import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Date;
import java.util.Locale;

import javax.jcr.Node;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ContentQueryHelper;
import org.ametys.cms.repository.ContentTypeExpression;
import org.ametys.cms.repository.CurrentStepExpression;
import org.ametys.cms.repository.DefaultContent;
import org.ametys.cms.repository.HistoryStepExpression;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.repository.ModifiableWorkflowAwareContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.AmetysObjectIterator;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareComposite;
import org.ametys.plugins.repository.metadata.ModifiableCompositeMetadata;
import org.ametys.plugins.repository.metadata.UnknownMetadataException;
import org.ametys.plugins.repository.query.SortCriteria;
import org.ametys.plugins.repository.query.expression.AndExpression;
import org.ametys.plugins.repository.query.expression.Expression.Operator;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.plugins.workflow.store.AmetysObjectWorkflowStore;
import org.ametys.plugins.workflow.support.WorkflowProvider;
import org.ametys.plugins.workflow.support.WorkflowProvider.AmetysObjectWorkflow;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

import com.opensymphony.workflow.Workflow;

/**
 * {@link DefaultContent} implementation test case.
 */
@NeedFeatures({
    "repository/resolver",
    "workflow/workflow-provider",
    "exhaustif/content-type",
    "test/content-type",
    "cms/factories",
    "cms/workflow.content"
})
public class DefaultContentTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    /** Default content node type name. */
    public static final String NT_DEFAULT_CONTENT = "ametys:defaultContent";
    
    /** Ametys object resolver to subclasses. */
    protected static AmetysObjectResolver _ametysObjectResolver;
    
    private static final ZonedDateTime __DATE = ZonedDateTime.of(2009, 6, 17, 16, 11, 18, 356000, ZoneId.systemDefault());
    private static final ZonedDateTime __DATE2 = ZonedDateTime.of(2008, 5, 16, 15, 10, 17, 245000, ZoneId.systemDefault());
    
    /** Workflow provider */
    private static WorkflowProvider _workflowProvider;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _workflowProvider = (WorkflowProvider) manager.lookup(WorkflowProvider.ROLE);
    }
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }
    
    /**
     * Test {@link DefaultContent} implementation.
     * @throws Exception if an error occurs.
     */
    @Test
    public void testDefaultContent() throws Exception
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        ModifiableWorkflowAwareContent content = root.createChild("content", NT_DEFAULT_CONTENT);
        content.setTypes(new String[]{"type1"});
        
        Assert.assertNotNull(content.getId());
        Assert.assertTrue(content.getId().startsWith("content://"));
        Assert.assertEquals("content", content.getName());
        Assert.assertEquals("/content", content.getPath());
        Assert.assertEquals("", content.getParentPath());
        
        try
        {
            content.getWorkflowId();
            Assert.fail("DefaultContent#getWorkflowId() must throw AmetyRepositoryException");
        }
        catch (AmetysRepositoryException e)
        {
            // Expected
        }
        
        Assert.assertEquals(null, content.getLanguage());

        try
        {
            content.getTitle();
            Assert.fail("DefaultContent#getTitle() must throw UnknownDataException");
        }
        catch (UnknownDataException e)
        {
            // Expected
        }
        
        try
        {
            content.getCreationDate();
            Assert.fail("DefaultContent#getCreationDate() must throw UnknownDataException");
        }
        catch (UnknownDataException e)
        {
            // Expected
        }

        try
        {
            content.getCreator();
            Assert.fail("DefaultContent#getCreator() must throw AmetysRepositoryException");
        }
        catch (AmetysRepositoryException e)
        {
            // Expected
        }

        try
        {
            content.getLastContributor();
            Assert.fail("DefaultContent#getLastContributor() must throw AmetysRepositoryException");
        }
        catch (AmetysRepositoryException e)
        {
            // Expected
        }


        try
        {
            content.getLastModified();
            Assert.fail("DefaultContent#getLastModified() must throw AmetyRepositoryException");
        }
        catch (UnknownDataException e)
        {
            // Expected
        }
        
        // Test save changes
        content.setWorkflowId(2);
        content.setMixinTypes(new String[]{"mixin"});
        content.setLanguage("fr");
        content.setTitle("title");
        content.setCreationDate(__DATE);
        content.setCreator(new UserIdentity("john", "anyPopulation"));
        content.setLastContributor(new UserIdentity("john", "anyPopulation"));
        content.setLastModified(__DATE);
        content.saveChanges();
        
        Assert.assertEquals(2, content.getWorkflowId());
        Assert.assertEquals("type1", content.getTypes()[0]);
        Assert.assertEquals("mixin", content.getMixinTypes()[0]);
        Assert.assertEquals("fr", content.getLanguage());
        Assert.assertEquals("title", content.getTitle());
        Assert.assertEquals(__DATE.toInstant().toEpochMilli(), content.getCreationDate().toInstant().toEpochMilli());
        Assert.assertEquals(new UserIdentity("john", "anyPopulation"), content.getCreator());
        Assert.assertEquals(new UserIdentity("john", "anyPopulation"), content.getLastContributor());
        Assert.assertEquals(__DATE.toInstant().toEpochMilli(), content.getLastModified().toInstant().toEpochMilli());
        
        content.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        content.setLastModified(__DATE2);
        content.saveChanges();

        Assert.assertEquals(2, content.getWorkflowId());
        Assert.assertEquals("type1", content.getTypes()[0]);
        Assert.assertEquals("fr", content.getLanguage());
        Assert.assertEquals("title", content.getTitle());
        Assert.assertEquals(__DATE.toInstant().toEpochMilli(), content.getCreationDate().toInstant().toEpochMilli());
        Assert.assertEquals(new UserIdentity("john", "anyPopulation"), content.getCreator());
        Assert.assertEquals(new UserIdentity("brian", "anyPopulation"), content.getLastContributor());
        Assert.assertEquals(__DATE2.toInstant().toEpochMilli(), content.getLastModified().toInstant().toEpochMilli());
    }
    
    /**
    * Test {@link DefaultContent} implementation with a multilingual title.
    */
    @Test
    public void testMultilingualContent()
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        ModifiableContent content = root.createChild("content", NT_DEFAULT_CONTENT);
        
        Assert.assertNotNull(content.getId());
        Assert.assertTrue(content.getId().startsWith("content://"));
        
        content.setTypes(new String[]{"org.ametys.test.Content.multilingual"});
        content.setTitle("title_fr", Locale.FRENCH);
        Assert.assertEquals("title_fr", content.getTitle());
        
        content.setTitle("title_en", Locale.ENGLISH);
        Assert.assertEquals("title_en", content.getTitle());
        Assert.assertEquals("title_fr", content.getTitle(Locale.FRENCH));
        Assert.assertEquals("title_en", content.getTitle(Locale.ENGLISH));
        Assert.assertEquals("title_en", content.getTitle(Locale.GERMAN));
    }
    
    /**
     * Test {@link DefaultContent} implementation with workflow interaction.
     * @throws Exception if an error occurs.
     */
    @Test
    public void testDefaultContentWithWorkflow() throws Exception
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        
        // Populate
        _populateContents(root);
        
        // Generic sort criteria
        SortCriteria sortCriteria = new SortCriteria();
        sortCriteria.addCriterion(DefaultContent.METADATA_CREATION, true, true);
        sortCriteria.addCriterion(Content.ATTRIBUTE_TITLE, true, true);
        String sortQuery =  ' ' + sortCriteria.build();
        
        String typeQuery = ContentQueryHelper.getContentXPathQuery(new ContentTypeExpression(Operator.EQ, "type1"));
        typeQuery += sortQuery;
        AmetysObjectIterable<Content> itContent = _ametysObjectResolver.query(typeQuery);
        AmetysObjectIterator<Content> it = itContent.iterator();
        
        Assert.assertEquals(true, it.hasNext());
        Content content1 = it.next();
        Assert.assertEquals("type1", content1.getTypes()[0]);
        Assert.assertEquals("title1", content1.getTitle());
        Assert.assertEquals(true, it.hasNext());
        Content content2 = it.next();
        Assert.assertEquals("type1", content2.getTypes()[0]);
        Assert.assertEquals("title2", content2.getTitle());
        
        typeQuery = ContentQueryHelper.getContentXPathQuery(new ContentTypeExpression(Operator.EQ, "unknown"));
        typeQuery += sortQuery;
        itContent = _ametysObjectResolver.query(typeQuery);
        it = itContent.iterator();

        Assert.assertEquals(false, it.hasNext());
        
        String wfQuery = ContentQueryHelper.getWorkflowXPathQuery(new CurrentStepExpression(2), null);
        wfQuery += sortQuery;
        itContent = _ametysObjectResolver.query(wfQuery);
        it = itContent.iterator();

        Assert.assertEquals(true, it.hasNext());
        Assert.assertEquals("title1", it.next().getTitle());
        
        wfQuery = ContentQueryHelper.getWorkflowXPathQuery(new HistoryStepExpression(1), null);
        wfQuery += sortQuery;
        itContent = _ametysObjectResolver.query(wfQuery);
        it = itContent.iterator();

        Assert.assertEquals(true, it.hasNext());
        content1 = it.next();
        Assert.assertEquals("title1", content1.getTitle());
        Assert.assertEquals("title1", content1.getTitle(null));
        Assert.assertEquals("title1", content1.getTitle(Locale.FRENCH));
        Assert.assertEquals(true, it.hasNext());
        Content content3 = it.next();
        Assert.assertEquals("title3_fr", content3.getTitle(Locale.FRENCH));
        Assert.assertEquals("title3_en", content3.getTitle(Locale.ENGLISH));
        Assert.assertEquals("title3_en", content3.getTitle(Locale.GERMAN));
        
        wfQuery = ContentQueryHelper.getWorkflowXPathQuery(new AndExpression(new CurrentStepExpression(2), new HistoryStepExpression(1)), null);
        wfQuery += sortQuery;
        itContent = _ametysObjectResolver.query(wfQuery);
        it = itContent.iterator();

        Assert.assertEquals(true, it.hasNext());
        Assert.assertEquals("title1", it.next().getTitle());
        
        String hybridQuery = ContentQueryHelper.getWorkflowXPathQuery(new CurrentStepExpression(1), new ContentTypeExpression(Operator.EQ, "type1"));
        hybridQuery += sortQuery;
        itContent = _ametysObjectResolver.query(hybridQuery);
        it = itContent.iterator();

        Assert.assertEquals(true, it.hasNext());
        Assert.assertEquals("title2", it.next().getTitle());
    }

    private void _populateContents(RootAmetysObject root) throws Exception
    {
        ModifiableWorkflowAwareContent content1 = root.createChild("content1", NT_DEFAULT_CONTENT);

        content1.setTypes(new String[]{"type1"});
        content1.setLanguage("fr");
        content1.setValue(Content.ATTRIBUTE_TITLE, "title1");
        content1.setCreationDate(ZonedDateTime.of(2009, 6, 17, 16, 11, 18, 356000, ZoneId.systemDefault()));
        content1.setCreator(new UserIdentity("john", "anyPopulation"));
        content1.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        content1.setLastModified(__DATE2);
        Workflow content1Workflow = _createWorkflowForContent(content1);
        content1Workflow.doAction(content1.getWorkflowId(), 3, null);
        
        ModifiableWorkflowAwareContent content2 = root.createChild("content2", NT_DEFAULT_CONTENT);
        
        content2.setTypes(new String[]{"type1"});
        content2.setLanguage("fr");
        content2.setTitle("title2");
        content2.setCreationDate(ZonedDateTime.of(2009, 6, 17, 16, 11, 19, 356000, ZoneId.systemDefault()));
        content2.setCreator(new UserIdentity("john", "anyPopulation"));
        content2.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        content2.setLastModified(__DATE2);
        _createWorkflowForContent(content2);
        
        // Multilingual content
        ModifiableWorkflowAwareContent content3 = root.createChild("content3", NT_DEFAULT_CONTENT);
        
        content3.setTypes(new String[]{"type2"});
        content3.setLanguage(null);
        content3.setTitle("title3_fr", Locale.FRENCH);
        content3.setTitle("title3_en", Locale.ENGLISH);
        content3.setCreationDate(ZonedDateTime.of(2009, 6, 17, 16, 11, 20, 356000, ZoneId.systemDefault()));
        content3.setCreator(new UserIdentity("john", "anyPopulation"));
        content3.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        content3.setLastModified(__DATE2);
        Workflow content3Workflow = _createWorkflowForContent(content3);
        content3Workflow.doAction(content3.getWorkflowId(), 2, null);
    }

    private Workflow _createWorkflowForContent(ModifiableWorkflowAwareContent content) throws Exception
    {
        AmetysObjectWorkflow workflow = _workflowProvider.getAmetysObjectWorkflow(content);
        long workflowId = workflow.initialize("test", 1, null);
        
        content.setWorkflowId(workflowId);
        content.saveChanges();
        
        AmetysObjectWorkflowStore wfStore = (AmetysObjectWorkflowStore) workflow.getConfiguration().getWorkflowStore();
        Node node = wfStore.getEntryNode(content.getNode().getSession(), workflowId);
        node.setProperty("ametys-internal:initialActionId", 1);
        
        node.getSession().save();
        content.saveChanges();
        
        return workflow;
    }
    
    /**
     * Tests the {@link DefaultContent} behavior with attributes of type content
     */
    @SuppressWarnings("static-access")
    @Test
    public void testAccessOnLinkedContentAttributes()
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        ModifiableContent content = _createContent(root, "content", "org.ametys.test.Content.exhaustive");
        
        // Tests isMultiple method
        Assert.assertFalse(content.isMultiple("linkContent/title"));
        Assert.assertTrue(content.isMultiple("owns/longmult"));
        Assert.assertFalse(content.isMultiple("address/city/postCode"));
        
        // Tests getType
        Assert.assertEquals(ModelItemTypeConstants.STRING_TYPE_ID, content.getType("linkContent/title").getId());
        Assert.assertEquals(ModelItemTypeConstants.LONG_TYPE_ID, content.getType("owns/longmult").getId());
        Assert.assertEquals(ModelItemTypeConstants.STRING_TYPE_ID, content.getType("address/city/postCode").getId());
        
        // Tests hasValue method when the content is empty
        Assert.assertFalse(content.hasValue("linkContent/title"));
        Assert.assertFalse(content.hasValue("address/city/postCode"));
        
        // Create linked contents
        ModifiableContent linkContent = _createContent(root, "linkContent", "simple.Snake");
        content.setValue("linkContent", linkContent);
        ModifiableContent cityContent = _createContent(root, "cityContent", "simple.City");
        ModifiableModelAwareComposite contentAddress = content.getComposite("address", true);
        contentAddress.setValue("city", cityContent);
        
        // Tests to set and get the values
        content.setValue("linkContent/title", "Kaa");
        Assert.assertEquals("Kaa", linkContent.getTitle());
        Assert.assertEquals("Kaa", content.getValue("linkContent/title"));
        content.setValue("address/city/postCode", "31590");
        Assert.assertEquals("31590", cityContent.getValue("postCode"));
        Assert.assertEquals("31590", content.getValue("address/city/postCode"));
        
        // Tests hasValue method when the content is filled
        Assert.assertTrue(content.hasValue("linkContent/title"));
        Assert.assertTrue(content.hasValue("address/city/postCode"));
        
        // Tests removeValue method
        content.removeValue("linkContent/title");
        Assert.assertFalse(content.hasValue("linkContent/title"));
        content.removeValue("address/city/postCode");
        Assert.assertFalse(content.hasValue("address/city/postCode"));
    }
    
    /**
     * Tests that the old attribute API can read data set with the new one
     * @deprecated The old API is deprecated, so this test too. Remove this test when the old API is removed
     */
    @Test
    @Deprecated
    public void testSetDataWithNewAttributeAPIAndReadItWithTheOldOne()
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        ModifiableContent content = _createContent(root, "content", "org.ametys.test.Content.exhaustive");
        ModifiableCompositeMetadata metadataHolder = content.getMetadataHolder();
        
        // Empty long value
        try
        {
            metadataHolder.getLong("long");
            Assert.fail("Old API should throw an UnknownMetadataException while reading an empty value");
        }
        catch (UnknownMetadataException e)
        {
            Assert.assertTrue(e.getMessage().contains("long"));
        }
        
        // Empty double value
        content.setValue("double", null);
        try
        {
            metadataHolder.getDouble("double");
            Assert.fail("Old API should throw an UnknownMetadataException while reading an empty value");
        }
        catch (UnknownMetadataException e)
        {
            Assert.assertTrue(e.getMessage().contains("double"));
        }
        
        // Empty date value
        content.setValue("date", null);
        try
        {
            metadataHolder.getDate("date");
            Assert.fail("Old API should throw an UnknownMetadataException while reading an empty value");
        }
        catch (UnknownMetadataException e)
        {
            Assert.assertTrue(e.getMessage().contains("date"));
        }
        
        // Empty date time value
        content.setValue("datetime", null);
        try
        {
            metadataHolder.getDate("datetime");
            Assert.fail("Old API should throw an UnknownMetadataException while reading an empty value");
        }
        catch (UnknownMetadataException e)
        {
            Assert.assertTrue(e.getMessage().contains("datetime"));
        }
        
        // Multiple users
        content.setValue("usermult", new UserIdentity[] {new UserIdentity("jhetfield", "metallica"), new UserIdentity("lulrich", "metallica"), new UserIdentity("dmustain", "megadeth")});
        UserIdentity[] users = metadataHolder.getUserArray("usermult");
        Assert.assertEquals(3, users.length);
        Assert.assertEquals("jhetfield", users[0].getLogin());
        Assert.assertEquals("metallica", users[0].getPopulationId());
        Assert.assertEquals("lulrich", users[1].getLogin());
        Assert.assertEquals("metallica", users[1].getPopulationId());
        Assert.assertEquals("dmustain", users[2].getLogin());
        Assert.assertEquals("megadeth", users[2].getPopulationId());
    }
    
    /**
     * Tests that the new API can read data set with the old one
     * @deprecated The old API is deprecated, so this test too. Remove this test when the old API is removed
     */
    @Test
    @Deprecated
    public void testSetDataWithOldAttributeAPIAndReadItWithTheNewOne()
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        ModifiableContent content = _createContent(root, "content", "org.ametys.test.Content.exhaustive");
        ModifiableCompositeMetadata metadataHolder = content.getMetadataHolder();
        
        // Datetime
        Instant instant = Instant.ofEpochMilli(1572009837000L);
        Date dateToSet = Date.from(instant);
        metadataHolder.setMetadata("datetime", dateToSet);
        ZonedDateTime retrievedDate = content.getValue("datetime");
        Assert.assertEquals(DateUtils.dateToString(dateToSet), DateUtils.zonedDateTimeToString(retrievedDate));
        
        // Multiple users
        metadataHolder.setMetadata("usermult", new UserIdentity[] {new UserIdentity("jhetfield", "metallica"), new UserIdentity("lulrich", "metallica"), new UserIdentity("dmustain", "megadeth")});
        UserIdentity[] users = content.getValue("usermult");
        Assert.assertEquals(3, users.length);
        Assert.assertEquals("jhetfield", users[0].getLogin());
        Assert.assertEquals("metallica", users[0].getPopulationId());
        Assert.assertEquals("lulrich", users[1].getLogin());
        Assert.assertEquals("metallica", users[1].getPopulationId());
        Assert.assertEquals("dmustain", users[2].getLogin());
        Assert.assertEquals("megadeth", users[2].getPopulationId());
    }
    
    private ModifiableContent _createContent(RootAmetysObject root, String name, String contentTypeId)
    {
        ModifiableContent content = root.createChild(name, NT_DEFAULT_CONTENT);
        
        content.setTypes(new String[]{contentTypeId});
        content.setLanguage("fr");
        content.setCreationDate(ZonedDateTime.of(2009, 6, 17, 16, 11, 18, 356000, ZoneId.systemDefault()));
        content.setCreator(new UserIdentity("john", "anyPopulation"));
        content.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        content.setLastModified(__DATE2);
        
        content.saveChanges();
        return content;
    }
}
