/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.search;

import java.util.List;
import java.util.Optional;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.content.ContentSearchHelper;
import org.ametys.cms.search.model.SystemProperty;
import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.cms.search.solr.field.ContentSearchField;
import org.ametys.cms.search.solr.field.DoubleSearchField;
import org.ametys.cms.search.solr.field.JoinedSystemSearchField;
import org.ametys.cms.search.solr.field.LastValidationSearchField;
import org.ametys.cms.search.solr.field.LongSearchField;
import org.ametys.cms.search.solr.field.StringSearchField;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElement;
import org.ametys.runtime.model.ViewHelper;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Tests for {@link ContentSearchHelper}
 */
@NeedFeatures({
    "cms/org.ametys.cms.content.search.helper",
    "exhaustif/content-type"
})
public class ContentSearchHelperTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __EXHAUSTIVE_CONTENT_TYPE_ID = "org.ametys.test.Content.exhaustive"; 
    private static final Set<String> __EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET = Set.of(__EXHAUSTIVE_CONTENT_TYPE_ID);
    
    private static ContentTypeExtensionPoint _contentTypeExtensionPoint;
    private static SystemPropertyExtensionPoint _systemPropertyExtensionPoint; 
    private static ContentSearchHelper _contentSearchHelper;
    
    private static View _view;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _systemPropertyExtensionPoint = (SystemPropertyExtensionPoint) manager.lookup(SystemPropertyExtensionPoint.ROLE);
        _contentSearchHelper = (ContentSearchHelper) manager.lookup(ContentSearchHelper.ROLE);
    }
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }

    public void prepareComponents() throws Exception
    {
        WithCmsFeaturesTestCase.super.prepareComponents();
        
        ContentType exhaustive = _contentTypeExtensionPoint.getExtension(__EXHAUSTIVE_CONTENT_TYPE_ID);
        _view = View.of(List.of(exhaustive), "title",
                                             "owns/double",
                                             "composite/long-min",
                                             "composite/repeater/comp-user",
                                             "composite/repeater/comp-owns/sport",
                                             "composite/repeater/comp-owns/linkContent/title",
                                             "composite/repeater/comp-owns/owns/composite/repeater/comp-owns/title");
        
        _addSystemPropertyViewItem(StringUtils.EMPTY, "lastValidation");
        _addSystemPropertyViewItem("owns", "creator");
        _addSystemPropertyViewItem("composite/repeater/comp-owns/linkContent", "workflowStep");
        _addSystemPropertyViewItem("composite/repeater/comp-owns/owns/composite/repeater/comp-owns", "tags");
    }
    
    private void _addSystemPropertyViewItem(String prefix, String systemPropertyId)
    {
        ViewItemAccessor parent = _view;
        if (StringUtils.isNotEmpty(prefix))
        {
            parent = (ViewItemAccessor) ViewHelper.getModelViewItem(_view, prefix);
        }
        
        SystemProperty systemProperty = _systemPropertyExtensionPoint.getExtension(systemPropertyId);
        ViewElement viewElement = new ViewElement();
        viewElement.setDefinition(systemProperty);
        
        parent.addViewItem(viewElement);
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(java.util.Collection, String)} with an attribute at type's root
     */
    @Test
    public void testGetSearchFieldByDataPathOnRootAttribute()
    {
        Optional<SearchField> searchField = _contentSearchHelper.getSearchField(__EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, "title");
        
        Assert.assertTrue(searchField.isPresent());
        Assert.assertTrue(searchField.get() instanceof StringSearchField);
        Assert.assertEquals("title", searchField.get().getName());
        Assert.assertFalse(searchField.get().isJoined());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(ModelViewItem)} with an attribute at type's root
     */
    @Test
    public void testGetSearchFieldOnRootAttribute()
    {
        ModelViewItem modelViewItem = ViewHelper.getModelViewItem(_view, "title");
        SearchField searchField = _contentSearchHelper.getSearchField(modelViewItem);
        
        Assert.assertTrue(searchField instanceof StringSearchField);
        Assert.assertEquals("title", searchField.getName());
        Assert.assertFalse(searchField.isJoined());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(java.util.Collection, String)} with an attribute at type's root
     */
    @Test
    public void testGetSearchByDataPathFieldOnRootSystemProperty()
    {
        Optional<SearchField> searchField = _contentSearchHelper.getSearchField(__EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, "lastValidation");

        Assert.assertTrue(searchField.isPresent());
        Assert.assertTrue(searchField.get() instanceof LastValidationSearchField);
        Assert.assertEquals("lastValidation", searchField.get().getName());
        Assert.assertFalse(searchField.get().isJoined());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(ModelViewItem)} with an attribute at type's root
     */
    @Test
    public void testGetSearchFieldOnRootSystemProperty()
    {
        ModelViewItem modelViewItem = ViewHelper.getModelViewItem(_view, "lastValidation");
        SearchField searchField = _contentSearchHelper.getSearchField(modelViewItem);
        
        Assert.assertTrue(searchField instanceof LastValidationSearchField);
        Assert.assertEquals("lastValidation", searchField.getName());
        Assert.assertFalse(searchField.isJoined());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(java.util.Collection, String)} with an attribute on a linked content
     */
    @Test
    public void testGetSearchFieldByDataPathOnJoinedAttribute()
    {
        Optional<SearchField> searchField = _contentSearchHelper.getSearchField(__EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, "owns/double");
        
        Assert.assertTrue(searchField.isPresent());
        Assert.assertTrue(searchField.get() instanceof DoubleSearchField);
        Assert.assertEquals("owns/double", searchField.get().getName());
        Assert.assertTrue(searchField.get().isJoined());
        Assert.assertEquals(List.of("owns"), searchField.get().getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(ModelViewItem)} with an attribute on a linked content
     */
    @Test
    public void testGetSearchFieldOnJoinedAttribute()
    {
        ModelViewItem modelViewItem = ViewHelper.getModelViewItem(_view, "owns/double");
        SearchField searchField = _contentSearchHelper.getSearchField(modelViewItem);
        
        Assert.assertTrue(searchField instanceof DoubleSearchField);
        Assert.assertEquals("owns/double", searchField.getName());
        Assert.assertTrue(searchField.isJoined());
        Assert.assertEquals(List.of("owns"), searchField.getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(java.util.Collection, String)} with a system property on a linked content
     */
    @Test
    public void testGetSearchFieldByDataPathOnJoinedSystemProperty()
    {
        Optional<SearchField> searchField = _contentSearchHelper.getSearchField(__EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, "owns/creator");
        
        Assert.assertTrue(searchField.isPresent());
        Assert.assertTrue(searchField.get() instanceof JoinedSystemSearchField);
        Assert.assertEquals("owns/creator", searchField.get().getName());
        Assert.assertTrue(searchField.get().isJoined());
        Assert.assertEquals(List.of("owns"), searchField.get().getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(ModelViewItem)} with a system property on a linked content
     */
    @Test
    public void testGetSearchFieldOnJoinedSystemProperty()
    {
        ModelViewItem modelViewItem = ViewHelper.getModelViewItem(_view, "owns/creator");
        SearchField searchField = _contentSearchHelper.getSearchField(modelViewItem);
        
        Assert.assertTrue(searchField instanceof JoinedSystemSearchField);
        Assert.assertEquals("owns/creator", searchField.getName());
        Assert.assertTrue(searchField.isJoined());
        Assert.assertEquals(List.of("owns"), searchField.getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(java.util.Collection, String)} with an attribute in a composite
     */
    @Test
    public void testGetSearchFieldByDataPathOnAttributeInComposite()
    {
        Optional<SearchField> searchField = _contentSearchHelper.getSearchField(__EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, "composite/long-min");
        
        Assert.assertTrue(searchField.isPresent());
        Assert.assertTrue(searchField.get() instanceof LongSearchField);
        Assert.assertEquals("composite/long-min", searchField.get().getName());
        Assert.assertFalse(searchField.get().isJoined());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(ModelViewItem)} with an attribute in a composite
     */
    @Test
    public void testGetSearchFieldOnAttributeInComposite()
    {
        ModelViewItem modelViewItem = ViewHelper.getModelViewItem(_view, "composite/long-min");
        SearchField searchField = _contentSearchHelper.getSearchField(modelViewItem);
        
        Assert.assertTrue(searchField instanceof LongSearchField);
        Assert.assertEquals("composite/long-min", searchField.getName());
        Assert.assertFalse(searchField.isJoined());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(java.util.Collection, String)} with an attribute in a repeater
     */
    @Test
    public void testGetSearchFieldByDataPathOnAttributeInRepeater()
    {
        Optional<SearchField> searchField = _contentSearchHelper.getSearchField(__EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, "composite/repeater/comp-user");
        
        Assert.assertTrue(searchField.isPresent());
        Assert.assertTrue(searchField.get() instanceof StringSearchField);
        // TODO CMS-11712: fix the name of the search field
        Assert.assertEquals("composite/repeater/composite/repeater/comp-user", searchField.get().getName());
        Assert.assertTrue(searchField.get().isJoined());
        Assert.assertEquals(List.of("composite/repeater"), searchField.get().getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(ModelViewItem)} with an attribute in a repeater
     */
    @Test
    public void testGetSearchFieldOnAttributeInRepeater()
    {
        ModelViewItem modelViewItem = ViewHelper.getModelViewItem(_view, "composite/repeater/comp-user");
        SearchField searchField = _contentSearchHelper.getSearchField(modelViewItem);
        
        Assert.assertTrue(searchField instanceof StringSearchField);
        // TODO CMS-11712: fix the name of the search field
        Assert.assertEquals("composite/repeater/composite/repeater/comp-user", searchField.getName());
        Assert.assertTrue(searchField.isJoined());
        Assert.assertEquals(List.of("composite/repeater"), searchField.getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(java.util.Collection, String)} with an attribute on a linked content in a repeater
     */
    @Test
    public void testGetSearchFieldByDataPathOnJoinedContentAttributeInRepeater()
    {
        Optional<SearchField> searchField = _contentSearchHelper.getSearchField(__EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, "composite/repeater/comp-owns/sport");
        
        Assert.assertTrue(searchField.isPresent());
        Assert.assertTrue(searchField.get() instanceof ContentSearchField);
        Assert.assertEquals("composite/repeater/comp-owns/sport", searchField.get().getName());
        Assert.assertTrue(searchField.get().isJoined());
        Assert.assertEquals(List.of("composite/repeater", "comp-owns", "sport"), searchField.get().getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(ModelViewItem)} with an attribute on a linked content in a repeater
     */
    @Test
    public void testGetSearchFieldOnJoinedContentAttributeInRepeater()
    {
        ModelViewItem modelViewItem = ViewHelper.getModelViewItem(_view, "composite/repeater/comp-owns/sport");
        SearchField searchField = _contentSearchHelper.getSearchField(modelViewItem);
        
        Assert.assertTrue(searchField instanceof ContentSearchField);
        Assert.assertEquals("composite/repeater/comp-owns/sport", searchField.getName());
        Assert.assertTrue(searchField.isJoined());
        Assert.assertEquals(List.of("composite/repeater", "comp-owns", "sport"), searchField.getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(java.util.Collection, String)} with an attribute on a linked content in a repeater. The linked content does not specify a content type
     */
    @Test
    public void testGetSearchFieldByDataPathOnJoinedAttributeInRepeaterWithNoContentType()
    {
        Optional<SearchField> searchField = _contentSearchHelper.getSearchField(__EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, "composite/repeater/comp-owns/linkContent/title");
        
        Assert.assertTrue(searchField.isPresent());
        Assert.assertTrue(searchField.get() instanceof StringSearchField);
        Assert.assertEquals("composite/repeater/comp-owns/linkContent/title", searchField.get().getName());
        Assert.assertTrue(searchField.get().isJoined());
        Assert.assertEquals(List.of("composite/repeater", "comp-owns", "linkContent"), searchField.get().getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(ModelViewItem)} with an attribute on a linked content in a repeater. The linked content does not specify a content type
     */
    @Test
    public void testGetSearchFieldOnJoinedAttributeInRepeaterWithNoContentType()
    {
        ModelViewItem modelViewItem = ViewHelper.getModelViewItem(_view, "composite/repeater/comp-owns/linkContent/title");
        SearchField searchField = _contentSearchHelper.getSearchField(modelViewItem);
        
        Assert.assertTrue(searchField instanceof StringSearchField);
        Assert.assertEquals("composite/repeater/comp-owns/linkContent/title", searchField.getName());
        Assert.assertTrue(searchField.isJoined());
        Assert.assertEquals(List.of("composite/repeater", "comp-owns", "linkContent"), searchField.getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(java.util.Collection, String)} with a systemProperty on a linked content in a repeater. The linked content does not specify a content type
     */
    @Test
    public void testGetSearchFieldByDataPathOnJoinedSystemPropertyInRepeaterWithNoContentType()
    {
        Optional<SearchField> searchField = _contentSearchHelper.getSearchField(__EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, "composite/repeater/comp-owns/linkContent/workflowStep");
        
        Assert.assertTrue(searchField.isPresent());
        Assert.assertTrue(searchField.get() instanceof JoinedSystemSearchField);
        Assert.assertEquals("composite/repeater/comp-owns/linkContent/workflowStep", searchField.get().getName());
        Assert.assertTrue(searchField.get().isJoined());
        Assert.assertEquals(List.of("composite/repeater", "comp-owns", "linkContent"), searchField.get().getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(ModelViewItem)} with a systemProperty on a linked content in a repeater. The linked content does not specify a content type
     */
    @Test
    public void testGetSearchFieldOnJoinedSystemPropertyInRepeaterWithNoContentType()
    {
        ModelViewItem modelViewItem = ViewHelper.getModelViewItem(_view, "composite/repeater/comp-owns/linkContent/workflowStep");
        SearchField searchField = _contentSearchHelper.getSearchField(modelViewItem);
        
        Assert.assertTrue(searchField instanceof JoinedSystemSearchField);
        Assert.assertEquals("composite/repeater/comp-owns/linkContent/workflowStep", searchField.getName());
        Assert.assertTrue(searchField.isJoined());
        Assert.assertEquals(List.of("composite/repeater", "comp-owns", "linkContent"), searchField.getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(java.util.Collection, String)} with an attribute on linked contents and repeaters
     */
    @Test
    public void testGetSearchFieldByDataPathOnJoinedAttributeInRepeaterAndLinkedContent()
    {
        Optional<SearchField> searchField = _contentSearchHelper.getSearchField(__EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, "composite/repeater/comp-owns/owns/composite/repeater/comp-owns/title");
        
        Assert.assertTrue(searchField.isPresent());
        Assert.assertTrue(searchField.get() instanceof StringSearchField);
        Assert.assertEquals("composite/repeater/comp-owns/owns/composite/repeater/comp-owns/title", searchField.get().getName());
        Assert.assertTrue(searchField.get().isJoined());
        Assert.assertEquals(List.of("composite/repeater", "comp-owns", "owns", "composite/repeater", "comp-owns"), searchField.get().getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(ModelViewItem)} with an attribute on linked contents and repeaters
     */
    @Test
    public void testGetSearchFieldOnJoinedAttributeInRepeaterAndLinkedContent()
    {
        ModelViewItem modelViewItem = ViewHelper.getModelViewItem(_view, "composite/repeater/comp-owns/owns/composite/repeater/comp-owns/title");
        SearchField searchField = _contentSearchHelper.getSearchField(modelViewItem);
        
        Assert.assertTrue(searchField instanceof StringSearchField);
        Assert.assertEquals("composite/repeater/comp-owns/owns/composite/repeater/comp-owns/title", searchField.getName());
        Assert.assertTrue(searchField.isJoined());
        Assert.assertEquals(List.of("composite/repeater", "comp-owns", "owns", "composite/repeater", "comp-owns"), searchField.getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(java.util.Collection, String)} with a systemProperty on linked contents and repeaters
     */
    @Test
    public void testGetSearchFieldByDataPathOnJoinedSystemPropertyInRepeaterAndLinkedContent()
    {
        Optional<SearchField> searchField = _contentSearchHelper.getSearchField(__EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, "composite/repeater/comp-owns/owns/composite/repeater/comp-owns/tags");
        
        Assert.assertTrue(searchField.isPresent());
        Assert.assertTrue(searchField.get() instanceof JoinedSystemSearchField);
        Assert.assertEquals("composite/repeater/comp-owns/owns/composite/repeater/comp-owns/tags", searchField.get().getName());
        Assert.assertTrue(searchField.get().isJoined());
        Assert.assertEquals(List.of("composite/repeater", "comp-owns", "owns", "composite/repeater", "comp-owns"), searchField.get().getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#getSearchField(ModelViewItem)} with a systemProperty on linked contents and repeaters
     */
    @Test
    public void testGetSearchFieldOnJoinedSystemPropertyInRepeaterAndLinkedContent()
    {
        ModelViewItem modelViewItem = ViewHelper.getModelViewItem(_view, "composite/repeater/comp-owns/owns/composite/repeater/comp-owns/tags");
        SearchField searchField = _contentSearchHelper.getSearchField(modelViewItem);
        
        Assert.assertTrue(searchField instanceof JoinedSystemSearchField);
        Assert.assertEquals("composite/repeater/comp-owns/owns/composite/repeater/comp-owns/tags", searchField.getName());
        Assert.assertTrue(searchField.isJoined());
        Assert.assertEquals(List.of("composite/repeater", "comp-owns", "owns", "composite/repeater", "comp-owns"), searchField.getJoinedPaths());
    }
    
    /**
     * Test {@link ContentSearchHelper#computeJoinPaths(String, Set, boolean)} with an attribute at type's root
     */
    @Test
    public void testComputeJoinPathsOnRootAttribute()
    {
        List<String> joinPaths = _contentSearchHelper.computeJoinPaths("title", __EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, false);
        Assert.assertTrue(joinPaths.isEmpty());
    }
    
    /**
     * Test {@link ContentSearchHelper#computeJoinPaths(String, Set, boolean)} with an attribute on a linked content
     */
    @Test
    public void testComputeJoinPathsOnJoinedAttribute()
    {
        List<String> joinPaths = _contentSearchHelper.computeJoinPaths("owns/double", __EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, false);
        Assert.assertEquals(List.of("owns"), joinPaths);
    }
    
    /**
     * Test {@link ContentSearchHelper#computeJoinPaths(String, Set, boolean)} with an attribute in a repeater
     */
    @Test
    public void testComputeJoinPathsOnAttributeInRepeater()
    {
        List<String> joinPaths = _contentSearchHelper.computeJoinPaths("composite/repeater/comp-user", __EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, false);
        Assert.assertEquals(List.of("composite/repeater"), joinPaths);
    }
    
    /**
     * Test {@link ContentSearchHelper#computeJoinPaths(String, Set, boolean)} with an attribute on a linked content in a repeater
     */
    @Test
    public void testComputeJoinPathsOnJoinedContentAttributeInRepeater()
    {
        List<String> joinPaths = _contentSearchHelper.computeJoinPaths("composite/repeater/comp-owns/sport", __EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, false);
        Assert.assertEquals(List.of("composite/repeater", "comp-owns"), joinPaths);
    }
    
    /**
     * Test {@link ContentSearchHelper#computeJoinPaths(String, Set, boolean)} with an attribute on a linked content in a repeater. The linked content does not specify a content type
     */
    @Test
    public void testComputeJoinPathsOnJoinedAttributeInRepeaterWithNoContentType()
    {
        List<String> joinPaths = _contentSearchHelper.computeJoinPaths("composite/repeater/comp-owns/linkContent/title", __EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, false);
        Assert.assertEquals(List.of("composite/repeater", "comp-owns", "linkContent"), joinPaths);
    }
    
    /**
     * Test {@link ContentSearchHelper#computeJoinPaths(String, Set, boolean)} with an attribute on linked contents and repeaters
     */
    @Test
    public void testComputeJoinPathsOnJoinedAttributeInRepeaterAndLinkedContent()
    {
        List<String> joinPaths = _contentSearchHelper.computeJoinPaths("composite/repeater/comp-owns/owns/composite/repeater/comp-owns/title", __EXHAUSTIVE_CONTENT_TYPE_ID_IN_SET, false);
        Assert.assertEquals(List.of("composite/repeater", "comp-owns", "owns", "composite/repeater", "comp-owns"), joinPaths);
    }
}
