/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.search;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.SearchResult;
import org.ametys.cms.search.SearchResults;
import org.ametys.cms.search.Sort.Order;
import org.ametys.cms.search.content.ContentSearcherFactory;
import org.ametys.cms.search.query.MatchAllQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.StringQuery;
import org.ametys.cms.search.query.WorkflowStepQuery;
import org.ametys.cms.search.ui.model.SearchUIModel;
import org.ametys.cms.search.ui.model.SearchUIModelExtensionPoint;
import org.ametys.cms.test.framework.AbstractSearcherTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Test the {@link ContentSearcherFactory} API.
 */
@NeedFeatures({
    "cms/org.ametys.cms.content.search.content.searcher",
    
    "exhaustif/content-type",
    "cms/org.ametys.cms.data.holder.helpers"
})
public class ContentSearcherTestCase extends AbstractSearcherTestCase implements WithCmsTestWebappFeatures
{
    private static ContentSearcherFactory _contentSearcherFactory;
    private static SearchUIModelExtensionPoint _searchModelEP;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _contentSearcherFactory = (ContentSearcherFactory) manager.lookup(ContentSearcherFactory.ROLE);
        _searchModelEP = (SearchUIModelExtensionPoint) manager.lookup(SearchUIModelExtensionPoint.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        
        // Populate and index the repository, then add fake documents.
        populateRepositoryAndIndex();
        addFakeDocuments();
    }

    /**
     * Test simple content type API.
     * @throws Exception If an error occurs
     */
    @Test
    public void testContentTypeApi() throws Exception
    {
        SearchResults<Content> results;
        Iterator<SearchResult<Content>> it;
        
        Query all = new MatchAllQuery();
        Query draftQuery = new WorkflowStepQuery(1);
        
        Map<String, List<String>> facetValues = new HashMap<>();
        facetValues.put("address/country", Arrays.asList("france"));
        
        // Query all without content type
        results = _contentSearcherFactory.create()
                                         .setCheckRights(false)
                                         .searchWithFacets(all);
        Assert.assertEquals(24, results.getTotalCount());
        Assert.assertEquals(24, results.getResults().getSize());
        
        // Query with non-existing sort field.
        try
        {
            results = _contentSearcherFactory.create()
                                                .addSort("NonExisting", Order.ASC)
                                                .setCheckRights(false)
                                                .searchWithFacets(all);
            Assert.fail("A non-existing sort field should throw an exception.");
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
        
        // Query with non-existing facet field.
        try
        {
            results = _contentSearcherFactory.create()
                                                .withFacets("NonExisting")
                                                .setCheckRights(false)
                                                .searchWithFacets(all);
            Assert.fail("A non-existing facet field should throw an exception.");
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
        
        // Query all without content type, test sort and facets on standard properties.
        results = _contentSearcherFactory.create()
                                            .addSort("workflowStep", Order.DESC)
                                            .addSort("title", Order.DESC)
                                            .withFacets("workflowStep", "contentTypes")
                                            .withLimits(0, 2)
                                            .setCheckRights(false)
                                            .searchWithFacets(all);
        Assert.assertEquals(24, results.getTotalCount());
        Assert.assertEquals(2, results.getResults().getSize());
        it = results.getResults().iterator();
        Assert.assertEquals("Validated content DDD", it.next().getObject().getTitle(null));
        Assert.assertEquals("Validated content CCC", it.next().getObject().getTitle(null));
        Assert.assertFalse(it.hasNext());
        
        // Query all with content types
        results = _contentSearcherFactory.create(EXHAUSTIVE_CONTENTTYPE)
                                          .setCheckRights(false)
                                          .searchWithFacets(all);
        Assert.assertEquals(24, results.getTotalCount());
        Assert.assertEquals(24, results.getResults().getSize());
        
        // Test some simple queries.
        results = _contentSearcherFactory.create(EXHAUSTIVE_CONTENTTYPE)
                                         .setCheckRights(false)
                                         .searchWithFacets(draftQuery);
        Assert.assertEquals(15, results.getResults().getSize());
        
        // String query on a field under a composite.
        results = _contentSearcherFactory.create(EXHAUSTIVE_CONTENTTYPE)
                                         .setCheckRights(false)
                                         .searchWithFacets(new StringQuery("address/country", "japan"));
        Assert.assertEquals(4, results.getResults().getSize());
        
        // Test sorting.
        results = _contentSearcherFactory.create(EXHAUSTIVE_CONTENTTYPE)
                                         .addSort("workflowStep", Order.DESC)
                                         .addSort("title", Order.DESC)
                                         .setCheckRights(false)
                                         .searchWithFacets(all);
        it = results.getResults().iterator();
        Assert.assertEquals("Validated content DDD", it.next().getObject().getTitle(null));
        Assert.assertEquals("Validated content CCC", it.next().getObject().getTitle(null));
        Assert.assertEquals("Test content", it.next().getObject().getTitle(null));
        
        // Test sorting and facets.
        results = _contentSearcherFactory.create(EXHAUSTIVE_CONTENTTYPE)
                                            .addSort("workflowStep", Order.DESC)
                                            .addSort("title", Order.DESC)
                                            .withFacets("workflowStep", "address/country", "contentTypes")
                                            .setCheckRights(false)
                                            .searchWithFacets(all);
        Assert.assertEquals(24, results.getResults().getSize());
        
        assertWorkflowFacetCounts(results);
        assertCountryFacetCounts(results);
        
        // Test sorting, facets and specific facet values.
        results = _contentSearcherFactory.create(EXHAUSTIVE_CONTENTTYPE)
                                            .addSort("workflowStep", Order.DESC)
                                            .addSort("title", Order.DESC)
                                            .withFacets("workflowStep", "address/country", "contentTypes")
                                            .setCheckRights(false)
                                            .searchWithFacets(all, facetValues);
        Assert.assertEquals(4, results.getTotalCount());
        
        assertFranceWorkflowFacetCounts(results);
        assertCountryFacetCounts(results);
        
        // Test sorting, facets with specific facet values, and limits.
        results = _contentSearcherFactory.create(EXHAUSTIVE_CONTENTTYPE)
                                            .addSort("workflowStep", Order.DESC)
                                            .addSort("title", Order.DESC)
                                            .withFacets("workflowStep", "address/country", "contentTypes")
                                            .withLimits(0, 2)
                                            .setCheckRights(false)
                                            .searchWithFacets(all);
        it = results.getResults().iterator();
        Assert.assertEquals("Validated content DDD", it.next().getObject().getTitle(null));
        Assert.assertEquals("Validated content CCC", it.next().getObject().getTitle(null));
        Assert.assertFalse(it.hasNext());
        
        // Test facets on additional indexing fields
        results = _contentSearcherFactory.create(EXHAUSTIVE_CONTENTTYPE)
                                         .withFacets("owns-sport", "owns-owns",
                                                     "owns-linkContent", "owns-country")
                                         .setCheckRights(false)
                                         .searchWithFacets(all);
        
        Map<String, Integer> sportValues = results.getFacetResults().get("owns-sport");
        Assert.assertNotNull(sportValues);
        Assert.assertTrue(sportValues.isEmpty());
        
        Map<String, Integer> ownsValues = results.getFacetResults().get("owns-owns");
        Assert.assertEquals(4, ownsValues.size());
        assertContentFacetValueCount(1, results, "owns-owns", "linked-content-21");
        assertContentFacetValueCount(1, results, "owns-owns", "linked-content-22");
        assertContentFacetValueCount(1, results, "owns-owns", "linked-content-31");
        assertContentFacetValueCount(1, results, "owns-owns", "linked-content-221");
        
        Map<String, Integer> linkValues = results.getFacetResults().get("owns-linkContent");
        Assert.assertEquals(2, linkValues.size());
        assertContentFacetValueCount(1, results, "owns-linkContent", "linked-content-31");
        assertContentFacetValueCount(1, results, "owns-linkContent", "linked-content-221");
        
        Map<String, Integer> countryValues = results.getFacetResults().get("owns-country");
        Assert.assertEquals(5, countryValues.size());
        assertFacetValueCount(2, results, "owns-country", "spain");
        assertFacetValueCount(2, results, "owns-country", "japan");
    }
    
    /**
     * Test the SearchModel API.
     * @throws Exception If an error occurs
     */
    @Test
    public void testSearchModelApi() throws Exception
    {
        SearchUIModel searchModel = _searchModelEP.getExtension(TestConstants.TEST_SEARCH_MODEL);
        
        SearchResults<Content> results;
        Iterator<SearchResult<Content>> it;
        
        Map<String, List<String>> facetValues = new HashMap<>();
        facetValues.put("address/country", Arrays.asList("france"));
        
        // Query all.
        results = _contentSearcherFactory.create(searchModel)
                                            .setCheckRights(false)
                                            .searchWithFacets(Collections.emptyMap());
        Assert.assertEquals(24, results.getTotalCount());
        Assert.assertEquals(24, results.getResults().getSize());
        
        // Query with non-existing sort field.
        try
        {
            results = _contentSearcherFactory.create(searchModel)
                                                .setCheckRights(false)
                                                .addSort("NonExisting", Order.ASC)
                                                .searchWithFacets(Collections.emptyMap());
            Assert.fail("A non-existing sort field should throw an exception.");
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
        
        // Simple query
        results = _contentSearcherFactory.create(searchModel)
                                            .setCheckRights(false)
                                            .searchWithFacets(Collections.singletonMap("property-workflowStep-eq", 1L));
        Assert.assertEquals(15, results.getResults().getSize());
        
        // Simple query with a field under a composite.
        results = _contentSearcherFactory.create(searchModel)
                                            .setCheckRights(false)
                                            .searchWithFacets(Collections.singletonMap("metadata-address/country-eq", "japan"));
        Assert.assertEquals(4, results.getResults().getSize());
        
        // Simple query with sort.
        results = _contentSearcherFactory.create(searchModel)
                                            .addSort("workflowStep", Order.DESC)
                                            .addSort("title", Order.DESC)
                                            .setCheckRights(false)
                                            .searchWithFacets(Collections.emptyMap());
        it = results.getResults().iterator();
        Assert.assertEquals("Validated content DDD", it.next().getObject().getTitle(null));
        Assert.assertEquals("Validated content CCC", it.next().getObject().getTitle(null));
        Assert.assertEquals("Test content", it.next().getObject().getTitle(null));
        
        // Test the facets (including additional indexing fields)
        results = _contentSearcherFactory.create(searchModel)
                                            .setCheckRights(false)
                                            .searchWithFacets(Collections.emptyMap());
        
        assertWorkflowFacetCounts(results);
        assertCountryFacetCounts(results);
        
        Map<String, Integer> sportValues = results.getFacetResults().get("owns-sport");
        Assert.assertNotNull(sportValues);
        Assert.assertTrue(sportValues.isEmpty());
        
        Map<String, Integer> ownsValues = results.getFacetResults().get("owns-owns");
        Assert.assertEquals(4, ownsValues.size());
        assertContentFacetValueCount(1, results, "owns-owns", "linked-content-21");
        assertContentFacetValueCount(1, results, "owns-owns", "linked-content-22");
        assertContentFacetValueCount(1, results, "owns-owns", "linked-content-31");
        assertContentFacetValueCount(1, results, "owns-owns", "linked-content-221");
        
        Map<String, Integer> linkValues = results.getFacetResults().get("owns-linkContent");
        Assert.assertEquals(2, linkValues.size());
        assertContentFacetValueCount(1, results, "owns-linkContent", "linked-content-31");
        assertContentFacetValueCount(1, results, "owns-linkContent", "linked-content-221");
        
        Map<String, Integer> countryValues = results.getFacetResults().get("owns-country");
        Assert.assertEquals(5, countryValues.size());
        assertFacetValueCount(2, results, "owns-country", "spain");
        assertFacetValueCount(2, results, "owns-country", "japan");
        
        // Test facets with specific facet values.
        results = _contentSearcherFactory.create(searchModel)
                                            .addSort("workflowStep", Order.DESC)
                                            .addSort("title", Order.DESC)
                                            .setCheckRights(false)
                                            .searchWithFacets(Collections.emptyMap(), facetValues, Collections.emptyMap());
        Assert.assertEquals(4, results.getTotalCount());
        
        assertFranceWorkflowFacetCounts(results);
        assertCountryFacetCounts(results);
        
        // Test sorting, facets with specific facet values, and limits.
        results = _contentSearcherFactory.create(searchModel)
                                            .addSort("workflowStep", Order.DESC)
                                            .addSort("title", Order.DESC)
                                            .withLimits(0, 2)
                                            .setCheckRights(false)
                                            .searchWithFacets(Collections.emptyMap(), facetValues, Collections.emptyMap());
        it = results.getResults().iterator();
        Assert.assertEquals(4, results.getTotalCount());
        Assert.assertEquals("Test content", it.next().getObject().getTitle(null));
        Assert.assertEquals("Proposed content BBB", it.next().getObject().getTitle(null));
        Assert.assertFalse(it.hasNext());
    }
    
}
