/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.search;

import java.io.IOException;

import javax.jcr.RepositoryException;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.solr.client.solrj.SolrServerException;
import org.junit.After;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.RootContentHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableDefaultContent;
import org.ametys.cms.search.content.ContentSearcherFactory;
import org.ametys.cms.search.content.ContentSearcherFactory.SimpleContentSearcher;
import org.ametys.cms.search.query.StringQuery;
import org.ametys.cms.test.framework.AbstractSearcherTestCase;
import org.ametys.core.right.RightManager;
import org.ametys.plugins.repository.AmetysObject;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.collection.AmetysObjectCollection;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Test the check of rights of the {@link ContentSearcherFactory} API.
 * <br>i.e. test it with {@link SimpleContentSearcher#setCheckRights} at <code>true</code>.
 */
@NeedFeatures({
    "cms/org.ametys.cms.content.search.content.searcher",
    "cms/org.ametys.cms.search.solr.check-read-access",
    
    "core/runtime.rights.manager",
    "repository/repository.rights.storage.assignments.impl.jcr",
    "cms/cms.rights.accesscontroller.impl.content",
    
    "cms/root.content.helper",
    "cms/org.ametys.cms.data.holder.helpers",
    
    "cms/sitemap.component.dependencies", // Solr will query an URL, we need to load the dependencies of the sitemap
    
    "exhaustif/content-type",
})
public class ContentSearcherWithRightsTestCase extends AbstractSearcherTestCase implements WithCmsTestWebappFeatures
{
    private static ContentSearcherFactory _contentSearcherFactory;
    private static RightManager _rightManager;
    private static RootContentHelper _rootContentHelper;
    
    private static String _profile;
    private static ModifiableDefaultContent _testContent;
    private static AmetysObjectCollection _rootOfContents;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _contentSearcherFactory = (ContentSearcherFactory) manager.lookup(ContentSearcherFactory.ROLE);
        _rightManager = (RightManager) manager.lookup(RightManager.ROLE);
        _rootContentHelper = (RootContentHelper) manager.lookup(RootContentHelper.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        addFakeDocuments();
        _createTestContent();
        _initialAssert();
        _profile = RightManager.READER_PROFILE_ID;
        _rootOfContents = _rootContentHelper.getRootContent();
    }
    
    /**
     * Logout current user in order to reset to anonymous after each test
     * @throws Exception if an error occurs.
     */
    @After
    public void logoutAfterTest() throws Exception
    {
        _logger.info("Logging out current user after test");
        _userProvider.logout(null);
    }
    
    /**
     * Reset assignments after each test
     * @throws Exception if an error occurs.
     */
    @After
    public void resetAssignments() throws Exception
    {
        _logger.info("Reinitializing assignments on content-test after test");
        _reinitAllAssignments(_testContent);
        _logger.info("Reinitializing assignments on root of contents after test");
        _reinitAllAssignments(_rootOfContents);
        
        // Empty the cache of RightManager
        _logger.info("Emptying right manager cache after test");
        _emptyRightManagerCache();
    }
    
    private void _reinitAllAssignments(AmetysObject objectToAssign)
    {
        _profileAssignmentStorageEP.removeAllowedProfileFromAnonymous(_profile, objectToAssign);
        _profileAssignmentStorageEP.removeDeniedProfileFromAnonymous(_profile, objectToAssign);
        
        _profileAssignmentStorageEP.removeAllowedProfileFromAnyConnectedUser(_profile, objectToAssign);
        _profileAssignmentStorageEP.removeDeniedProfileFromAnyConnectedUser(_profile, objectToAssign);
        
        _profileAssignmentStorageEP.removeAllowedProfileFromUser(USER1_TEST_IDENTITY, _profile, objectToAssign);
        _profileAssignmentStorageEP.removeDeniedProfileFromUser(USER1_TEST_IDENTITY, _profile, objectToAssign);
    
        _profileAssignmentStorageEP.removeAllowedProfileFromGroup(GROUP1_TEST_IDENTITY, _profile, objectToAssign);
        _profileAssignmentStorageEP.removeDeniedProfileFromGroup(GROUP1_TEST_IDENTITY, _profile, objectToAssign);
    }
    
    private static void _createTestContent() throws Exception
    {
        // Create our right test content
        _testContent = createContent("Test rights content", "testContent", EXHAUSTIVE_CONTENTTYPE);
        _testContent.setValue("abstract", "rights");
        _testContent.setValue("string", "test string 1");
        
        // Edit and validate the main content.
        _testContent.saveChanges();
        editContent(_testContent);
        doWorkflowAction(_testContent, 4);
        _reIndexRightContent();
    }
    
    private static void _initialAssert() throws Exception
    {
        // Test we retrieve this content and only this one, without checking rights in a first time
        AmetysObjectIterable<Content> results = _contentSearcherFactory.create(EXHAUSTIVE_CONTENTTYPE)
                        .setCheckRights(false)
                        .search(new StringQuery("abstract", "rights"));
        
        Assert.assertEquals(1, results.getSize());
    }
    
    /**
     * Tests {@link AbstractPopulationAwareTestCase#USER1_TEST_IDENTITY} has read-right on Solr content document 
     * when profile is assigned to the content itself
     * @throws Exception if an error occurs.
     */
    @Test
    public void testForUserWhenProfileOnContent() throws Exception
    {
        // Run tests for a user
        _userProvider.setUser(USER1_TEST_IDENTITY);
        _testCheckRightsForUser(_testContent);
    }
    
    /**
     * Tests 'Anonymous' has read-right on Solr content document 
     * when profile is assigned to the content itself
     * @throws Exception if an error occurs.
     */
    @Test
    public void testForAnonymousWhenProfileOnContent() throws Exception
    {
        // Force current user to be anonymous and run tests for anonymous
        _testCheckRightsForAnonymous(_testContent);
    }
    
    /**
     * Tests {@link AbstractPopulationAwareTestCase#USER1_TEST_IDENTITY} has read-right on Solr content document 
     * when profile is assigned to the root-of-contents
     * @throws Exception if an error occurs.
     */
    @Test
    public void testForUserWhenProfileOnRoot() throws Exception
    {
        // Run tests for a user, but with the assignments on the "root of contents" and not on the content itself
        _userProvider.setUser(USER1_TEST_IDENTITY);
        _testCheckRightsForUser(_rootOfContents);
    }
    
    /**
     * Tests 'Anonymous' has read-right on Solr content document 
     * when profile is assigned to the root-of-contents
     * @throws Exception if an error occurs.
     */
    @Test
    public void testForAnonymousWhenProfileOnRoot() throws Exception
    {
        // Force current user to be anonymous and run tests for anonymous with the assignments on the "root of contents"
        _testCheckRightsForAnonymous(_rootOfContents);
    }
    
    /**
     * Tests multiple times in a row to see if Solr cache is consistent.
     * @throws Exception if an error occurs.
     */
    @Test
    public void testSolrCache() throws Exception
    {
        // Call multiple times in a row the same method (without invalidating cache) to test the Solr-side cache
        _testRightContentNotReturned(true);
        _testRightContentNotReturned(false);
        _testRightContentNotReturned(false);
        
        _profileAssignmentStorageEP.allowProfileToAnonymous(_profile, _rootOfContents);
        _testRightContentReturned(true);
        _testRightContentReturned(false);
        _testRightContentReturned(false);
    }
    
    private static void _reIndexRightContent() throws Exception
    {
        String contentId = _testContent.getId();
        _solrIndexer.unindexContent(contentId, RepositoryConstants.DEFAULT_WORKSPACE, false, _updateSolr);
        _solrIndexer.indexContent(contentId, RepositoryConstants.DEFAULT_WORKSPACE, false, _updateSolr);
        _solrIndexer.commit(RepositoryConstants.DEFAULT_WORKSPACE, _updateSolr);
    }
    
    private void _testCheckRightsForUser(AmetysObject objectToAssign) throws Exception
    {
        // No profile assignment
        _testRightContentNotReturned(true);
       
        // Anonymous denied
        _profileAssignmentStorageEP.denyProfileToAnonymous(_profile, objectToAssign);
        _testRightContentNotReturned(true);
        
        // Any connected user allowed
        _profileAssignmentStorageEP.allowProfileToAnyConnectedUser(_profile, objectToAssign);
        _testRightContentReturned(true);
        
        // Any connected user denied
        _profileAssignmentStorageEP.removeAllowedProfileFromAnyConnectedUser(_profile, objectToAssign);
        _profileAssignmentStorageEP.denyProfileToAnyConnectedUser(_profile, objectToAssign);
        _testRightContentNotReturned(true);
        
        // Group allowed
        _profileAssignmentStorageEP.allowProfileToGroup(GROUP1_TEST_IDENTITY, _profile, objectToAssign);
        _testRightContentReturned(true);
        
        // Group denied
        _profileAssignmentStorageEP.denyProfileToGroup(GROUP1_TEST_IDENTITY, _profile, objectToAssign);
        _testRightContentNotReturned(true);
        
        // User allowed
        _profileAssignmentStorageEP.allowProfileToUser(USER1_TEST_IDENTITY, _profile, objectToAssign);
        _testRightContentReturned(true);
        
        // User denied
        _profileAssignmentStorageEP.denyProfileToUser(USER1_TEST_IDENTITY, _profile, objectToAssign);
        _testRightContentNotReturned(true);
        
        // Anonymous allowed
        _profileAssignmentStorageEP.removeDeniedProfileFromAnonymous(_profile, objectToAssign);
        _profileAssignmentStorageEP.allowProfileToAnonymous(_profile, objectToAssign);
        _testRightContentReturned(true);
    }
    
    private void _testCheckRightsForAnonymous(AmetysObject objectToAssign) throws Exception
    {
        // No profile assignment
        _testRightContentNotReturned(true);
       
        // Anonymous denied
        _profileAssignmentStorageEP.denyProfileToAnonymous(_profile, objectToAssign);
        _testRightContentNotReturned(true);
        
        // Any connected user allowed
        _profileAssignmentStorageEP.allowProfileToAnyConnectedUser(_profile, objectToAssign);
        _testRightContentNotReturned(true);
        
        // Any connected user denied
        _profileAssignmentStorageEP.removeAllowedProfileFromAnyConnectedUser(_profile, objectToAssign);
        _profileAssignmentStorageEP.denyProfileToAnyConnectedUser(_profile, objectToAssign);
        _testRightContentNotReturned(true);
        
        // Anonymous allowed
        _profileAssignmentStorageEP.removeDeniedProfileFromAnonymous(_profile, objectToAssign);
        _profileAssignmentStorageEP.allowProfileToAnonymous(_profile, objectToAssign);
        _testRightContentReturned(true);
    }
    
    private void _testRightContentReturned(boolean invalidateSolrAclCache) throws Exception
    {
        // Empty the cache of RightManager
        _emptyRightManagerCache();
        
        if (invalidateSolrAclCache)
        {
            // Invalidate Solr ACL cache
            _reloadSolrAclCache();
        }
        
        // Also test that the RightManager returns expected result too
        Assert.assertTrue(_rightManager.currentUserHasReadAccess(_testContent));
        
        // Test that the search (with check rights) returns the content
        AmetysObjectIterable<Content> result = _searchRightContent();
        Assert.assertEquals(1, result.getSize());
        Assert.assertEquals(_testContent, result.iterator().next());
    }
    
    private void _testRightContentNotReturned(boolean invalidateSolrAclCache) throws Exception
    {
        // Empty the cache of RightManager
        _emptyRightManagerCache();
        
        if (invalidateSolrAclCache)
        {
            // Invalidate Solr ACL cache
            _reloadSolrAclCache();
        }

        // Also test that the RightManager returns expected result too
        Assert.assertFalse(_rightManager.currentUserHasReadAccess(_testContent));
        
        // Test that the search (with check rights) does not return the content
        Assert.assertEquals(0, _searchRightContent().getSize());
    }
    
    private void _emptyRightManagerCache()
    {
        _rightManager.clearCaches();
    }
    
    private void _reloadSolrAclCache() throws IOException, SolrServerException, RepositoryException
    {
        _solrIndexer.reloadAclCache();
    }
    
    private AmetysObjectIterable<Content> _searchRightContent() throws Exception
    {
        return _contentSearcherFactory.create(EXHAUSTIVE_CONTENTTYPE)
                .setCheckRights(true)
                .search(new StringQuery("abstract", "rights"));
    }
}
