/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.search;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.languages.Language;
import org.ametys.cms.languages.LanguagesManager;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.content.ContentValuesExtractorFactory;
import org.ametys.cms.search.ui.model.SearchUIModel;
import org.ametys.cms.search.ui.model.SearchUIModelExtensionPoint;
import org.ametys.cms.test.framework.AbstractSearcherTestCase;
import org.ametys.core.user.User;
import org.ametys.core.user.UserManager;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.View;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Test the {@link ContentValuesExtractorFactory} API.
 */
@NeedFeatures({
    "cms/org.ametys.cms.content.search.content.values.extractor",
    "cms/org.ametys.cms.content.search.helper",
    "cms/content-types.helper",
    
    "content.search/content-type.exhaustive.overrides.content.values.extractor"
})
public class ContentValuesExtractorTestCase extends AbstractSearcherTestCase implements WithCmsTestWebappFeatures
{
    private static final String __TEST_VIEW_NAME = "content-values-extractor-tests";
    private static final String __REPEATERS_TEST_VIEW_NAME = "content-values-extractor-repeaters-tests";
    private static final String __EXHAUSTIVE_CONTENTTYPE_LABEL = "Exhaustive";
    
    private static ContentValuesExtractorFactory _contentValuesExtractor;
    private static SearchUIModelExtensionPoint _searchModelEP;
    private static UserManager _userManager;
    private static ContentTypesHelper _contentTypesHelper;
    private static LanguagesManager _languageManager;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _contentValuesExtractor = (ContentValuesExtractorFactory) manager.lookup(ContentValuesExtractorFactory.ROLE);
        _searchModelEP = (SearchUIModelExtensionPoint) manager.lookup(SearchUIModelExtensionPoint.ROLE);
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
        _languageManager = (LanguagesManager) manager.lookup(LanguagesManager.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        
        // Populate and index the repository, then add fake documents.
        populateRepositoryAndIndex();
        addFakeDocuments();
    }
    
    /**
     * Test simple content type API.
     * @throws Exception if an error occurs.
     */
    @Test
    @SuppressWarnings("unchecked")
    public void testSimpleContentValuesExtractor() throws Exception
    {
        Content content = getContent("content");
        View view = _contentTypesHelper.getView(__TEST_VIEW_NAME, content);
        
        Map<String, Object> values = _contentValuesExtractor.create(view)
                                        .getValues(content, null, new HashMap<>());
        
        Assert.assertEquals(32, values.size());
        
        Assert.assertEquals("Test content", values.get("title"));
        Assert.assertEquals("Test abstract", values.get("abstract"));
        Assert.assertEquals(List.of("one", "two", "three"), values.get("stringmult"));
        Assert.assertEquals(3L, values.get("long"));
        Assert.assertEquals(List.of(5L, 6L, 12L), values.get("longmult"));
        
        Assert.assertEquals(1.234D, values.get("double"));
        Assert.assertEquals(List.of(3.14D, 6.12D), values.get("doublemult"));
        Assert.assertEquals(Boolean.TRUE, values.get("boolean"));
        Assert.assertEquals("2007-10-09", values.get("date"));
        Assert.assertEquals("2011-11-24T09:15:30.000Z", values.get("datetime"));
        
        Map<String, Object> fileValues = (Map<String, Object>) values.get("file");
        Assert.assertEquals("attribute", fileValues.get("type"));
        Assert.assertEquals("test.txt", fileValues.get("filename"));
        Assert.assertEquals("text/plain", fileValues.get("mime-type"));
        Assert.assertEquals("2011-11-24T09:15:30.000Z", fileValues.get("lastModified"));
        
        Map<String, Object> userValues = (Map<String, Object>) values.get("user");
        User anonymous = _userManager.getUser(AbstractPopulationAwareTestCase.ANONYMOUS_TEST_IDENTITY);
        Assert.assertEquals(AbstractPopulationAwareTestCase.ANONYMOUS_TEST_IDENTITY.getLogin(), userValues.get("login"));
        Assert.assertEquals(AbstractPopulationAwareTestCase.ANONYMOUS_TEST_IDENTITY.getPopulationId(), userValues.get("populationId"));
        Assert.assertEquals(anonymous.getLastName(), userValues.get("lastname"));
        Assert.assertEquals(anonymous.getFirstName(), userValues.get("firstname"));
        Assert.assertTrue(StringUtils.isEmpty((String) userValues.get("email")));
        
        List<Map<String, Object>> users = (List<Map<String, Object>>) values.get("usermult");
        User user1 = _userManager.getUser(AbstractPopulationAwareTestCase.USER1_TEST_IDENTITY);
        Assert.assertEquals(3, users.size());
        userValues = users.get(0);
        Assert.assertEquals(AbstractPopulationAwareTestCase.USER1_TEST_IDENTITY.getLogin(), userValues.get("login"));
        Assert.assertEquals(AbstractPopulationAwareTestCase.USER1_TEST_IDENTITY.getPopulationId(), userValues.get("populationId"));
        Assert.assertEquals(user1.getLastName(), userValues.get("lastname"));
        Assert.assertEquals(user1.getFirstName(), userValues.get("firstname"));
        Assert.assertTrue(StringUtils.isEmpty((String) userValues.get("email")));
        
        Map<String, Object> refValues = (Map<String, Object>) values.get("reference");
        Assert.assertEquals("__external", refValues.get("type"));
        Assert.assertEquals("http://www.ametys.org", refValues.get("value"));
        
        Map<String, Object> gpsValues = (Map<String, Object>) values.get("gps");
        Assert.assertEquals(44.481D, gpsValues.get("latitude"));
        Assert.assertEquals(3.208D, gpsValues.get("longitude"));
        
        Assert.assertTrue(values.get("address/country") instanceof Map);
        Map<String, Object> country = (Map<String, Object>) values.get("address/country");
        Assert.assertEquals("france", country.get("value"));
        Assert.assertEquals(new I18nizableText("France"), country.get("label"));
        
        Assert.assertFalse(values.containsKey("illustration/alt-text"));
        
        Assert.assertTrue(values.get("linkContent") instanceof Map);
        Map<String, Object> linkedContent = (Map<String, Object>) values.get("linkContent");
        _ckecContentProperties(linkedContent, "linked-content-1");
        Assert.assertEquals("Linked content 1", values.get("linkContent/title"));
        
        Assert.assertTrue(values.get("linkContents") instanceof List);
        List<Map<String, Object>> linkedContents = (List<Map<String, Object>>) values.get("linkContents");
        Assert.assertEquals(3, linkedContents.size());
        _ckecContentProperties(linkedContents.get(0), "linked-content-2");
        _ckecContentProperties(linkedContents.get(1), "linked-content-3");
        _ckecContentProperties(linkedContents.get(2), "linked-content-4");
        Assert.assertEquals(List.of("Linked content 2", "Linked content 3", "Linked content 4"),
                        values.get("linkContents/title"));
        
        Assert.assertEquals(List.of("entry 1", "entry 2"),
                        values.get("composite/repeater/comp-string"));
        Assert.assertEquals(List.of("Linked content 1", "Linked content 2"),
                        values.get("composite/repeater/comp-owns/title"));
        
        Assert.assertTrue(values.get("owns") instanceof List);
        List<Map<String, Object>> owns = (List<Map<String, Object>>) values.get("owns");
        Assert.assertEquals(3, owns.size());
        _ckecContentProperties(owns.get(0), "linked-content-2");
        _ckecContentProperties(owns.get(1), "linked-content-3");
        _ckecContentProperties(owns.get(2), "linked-content-4");
        Assert.assertEquals(List.of("Linked content 2", "Linked content 3", "Linked content 4"),
                        values.get("owns/title"));
        
        Assert.assertTrue(values.get("owns/address/country") instanceof List);
        List<Map<String, Object>> countries = (List<Map<String, Object>>) values.get("owns/address/country");
        Assert.assertEquals(3, countries.size());
        Assert.assertEquals("japan", countries.get(0).get("value"));
        Assert.assertEquals(new I18nizableText("Japan"), countries.get(0).get("label"));
        Assert.assertEquals("spain", countries.get(1).get("value"));
        Assert.assertEquals(new I18nizableText("Spain"), countries.get(1).get("label"));
        Assert.assertEquals("japan", countries.get(2).get("value"));
        Assert.assertEquals(new I18nizableText("Japan"), countries.get(2).get("label"));
        
        Assert.assertEquals(List.of("Linked content 21", "Linked content 22", "Linked content 31"),
                        values.get("owns/owns/title"));
        
        Assert.assertTrue(values.get("owns/owns/composite/repeater/comp-owns") instanceof List);
        List<Map<String, Object>> compOwns = (List<Map<String, Object>>) values.get("owns/owns/composite/repeater/comp-owns");
        Assert.assertEquals(6, compOwns.size());
        _ckecContentProperties(compOwns.get(0), "eee");
        _ckecContentProperties(compOwns.get(1), "ccc");
        _ckecContentProperties(compOwns.get(2), "linked-content-221");
        _ckecContentProperties(compOwns.get(3), "aaa");
        _ckecContentProperties(compOwns.get(4), "ddd");
        _ckecContentProperties(compOwns.get(5), "bbb");

        Assert.assertEquals(List.of("EEE", "CCC", "Linked content 221", "AAA", "DDD", "BBB"),
                        values.get("owns/owns/composite/repeater/comp-owns/title"));
        
        Map<String, Object> workflowStep = (Map<String, Object>) values.get("workflowStep");
        Assert.assertEquals(3, workflowStep.get("stepId"));
        Assert.assertEquals(List.of(new I18nizableText(__EXHAUSTIVE_CONTENTTYPE_LABEL)), values.get("contentTypes"));
        Assert.assertTrue(((Collection) values.get("mixins")).isEmpty());
        Assert.assertTrue(values.get("creationDate") instanceof String);
    }
    
    /**
     * Test the content value extractor with contents with data in repeaters
     * @throws Exception if an error occurs.
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testSimpleContentValuesExtractorWithRepeaters() throws Exception
    {
        Content content = getContent("content");
        View view = _contentTypesHelper.getView(__REPEATERS_TEST_VIEW_NAME, content);

        Map<String, Object> values = _contentValuesExtractor.create(view)
                                        .getValues(content, null, new HashMap<>());
        
        // Checks on composite/repeater
        Map<String, Object> repeater = (Map<String, Object>) values.get("composite/repeater");
        Assert.assertEquals("Repeater", ((I18nizableText) repeater.get("label")).getLabel());
        List<Map<String, Object>> repeaterEntries = (List<Map<String, Object>>) repeater.get("entries");
        Assert.assertEquals(2, repeaterEntries.size());
        
        Map<String, Object> entry1 = repeaterEntries.get(0);
        Map<String, Object> entry1Values = (Map<String, Object>) entry1.get("values");
        Assert.assertEquals("entry 1", entry1Values.get("comp-string"));
        Map<String, Object> entry1CompOwns = (Map<String, Object>) entry1Values.get("comp-owns");
        _ckecContentProperties(entry1CompOwns, "linked-content-1");
        Assert.assertEquals("subcomposite in entry 1", entry1Values.get("subcomposite/subcomp-str"));
        Assert.assertEquals(1, entry1.get("position"));
        
        Map<String, Object> entry2 = repeaterEntries.get(1);
        Map<String, Object> entry2Values = (Map<String, Object>) entry2.get("values");
        Assert.assertEquals("entry 2", entry2Values.get("comp-string"));
        Map<String, Object> entry2CompOwns = (Map<String, Object>) entry2Values.get("comp-owns");
        _ckecContentProperties(entry2CompOwns, "linked-content-2");
        Assert.assertEquals("subcomposite in entry 2", entry2Values.get("subcomposite/subcomp-str"));
        Assert.assertEquals(2, entry2.get("position"));
        
        // Checks on repeaterThreeLevels
        Map<String, Object> level1Repeater = (Map<String, Object>) values.get("repeaterThreeLevels");
        Assert.assertEquals("Three-level repeater", ((I18nizableText) level1Repeater.get("label")).getLabel());
        List<Map<String, Object>> level1RepeaterValues = (List<Map<String, Object>>) level1Repeater.get("entries");
        Assert.assertEquals(2, level1RepeaterValues.size());
        
        // repeaterThreeLevels[1]
        Map<String, Object> level1_1 = level1RepeaterValues.get(0);
        Assert.assertEquals(1, level1_1.get("position"));
        Map<String, Object> level1_1_values = (Map<String, Object>) level1_1.get("values");
        Assert.assertEquals("repeater three levels entry 1", level1_1_values.get("rep-string"));
        // repeaterThreeLevels[1]/repeater2
        Map<String, Object> level1_1_level2Reapeater = (Map<String, Object>) level1_1_values.get("repeater2");
        Assert.assertEquals("Second level repeater", ((I18nizableText) level1_1_level2Reapeater.get("label")).getLabel());
        List<Map<String, Object>> level1_1_level2 = (List<Map<String, Object>>) level1_1_level2Reapeater.get("entries");
        Assert.assertEquals(1, level1_1_level2.size());
        // repeaterThreeLevels[1]/repeater2[1]        
        Map<String, Object> level1_1_level2_1 = level1_1_level2.get(0);
        Assert.assertEquals(1, level1_1_level2_1.get("position"));
        // repeaterThreeLevels[1]/repeater2[1]/repeater3
        Map<String, Object> level1_1_level2_1_level3Repeater = (Map<String, Object>) ((Map<String, Object>) level1_1_level2_1.get("values")).get("repeater3");
        Assert.assertEquals("Third level repeater", ((I18nizableText) level1_1_level2_1_level3Repeater.get("label")).getLabel());
        List<Map<String, Object>> level1_1_level2_1_level3 = (List<Map<String, Object>>) level1_1_level2_1_level3Repeater.get("entries");
        Assert.assertEquals(2, level1_1_level2_1_level3.size());
        // repeaterThreeLevels[1]/repeater2[1]/repeater3[1]
        Map<String, Object> level1_1_level2_1_level3_1 = level1_1_level2_1_level3.get(0);
        Assert.assertEquals("111", ((Map<String, Object>) level1_1_level2_1_level3_1.get("values")).get("rep3-string"));
        Assert.assertEquals(1, level1_1_level2_1_level3_1.get("position"));
        // repeaterThreeLevels[1]/repeater2[1]/repeater3[2]
        Map<String, Object> level1_1_level2_1_level3_2 = level1_1_level2_1_level3.get(1);
        Assert.assertEquals("112", ((Map<String, Object>) level1_1_level2_1_level3_2.get("values")).get("rep3-string"));
        Assert.assertEquals(2, level1_1_level2_1_level3_2.get("position"));
        
        // repeaterThreeLevels[2]
        Map<String, Object> level1_2 = level1RepeaterValues.get(1);
        Assert.assertEquals(2, level1_2.get("position"));
        Map<String, Object> level1_2_values = (Map<String, Object>) level1_2.get("values");
        Assert.assertEquals("repeater three levels entry 2", level1_2_values.get("rep-string"));
        // repeaterThreeLevels[2]/repeater2
        Map<String, Object> level1_2_level2Repeater = (Map<String, Object>) level1_2_values.get("repeater2");
        Assert.assertEquals("Second level repeater", ((I18nizableText) level1_2_level2Repeater.get("label")).getLabel());
        List<Map<String, Object>> level1_2_level2 = (List<Map<String, Object>>) level1_2_level2Repeater.get("entries");
        Assert.assertEquals(2, level1_2_level2.size());
        // repeaterThreeLevels[2]/repeater2[1]
        Map<String, Object> level1_2_level2_1 = level1_2_level2.get(0);
        Assert.assertEquals(1, level1_2_level2_1.get("position"));
        // repeaterThreeLevels[2]/repeater2[1]/repeater3
        Map<String, Object> level1_2_level2_1_level3Repeater = (Map<String, Object>) ((Map<String, Object>) level1_2_level2_1.get("values")).get("repeater3");
        Assert.assertEquals("Third level repeater", ((I18nizableText) level1_2_level2_1_level3Repeater.get("label")).getLabel());
        List<Map<String, Object>> level1_2_level2_1_level3 = (List<Map<String, Object>>) level1_2_level2_1_level3Repeater.get("entries");
        Assert.assertEquals(1, level1_2_level2_1_level3.size());
        // repeaterThreeLevels[2]/repeater2[1]/repeater3[1]
        Map<String, Object> level1_2_level2_1_level3_1 = level1_2_level2_1_level3.get(0);
        Assert.assertEquals("211", ((Map<String, Object>) level1_2_level2_1_level3_1.get("values")).get("rep3-string"));
        Assert.assertEquals(1, level1_2_level2_1_level3_1.get("position"));
        // repeaterThreeLevels[2]/repeater2[2]
        Map<String, Object> level1_2_level2_2 = level1_2_level2.get(1);
        Assert.assertEquals(2, level1_2_level2_2.get("position"));
        // repeaterThreeLevels[2]/repeater2[2]/repeater3
        Map<String, Object> level1_2_level2_2_level3Repeater = (Map<String, Object>) ((Map<String, Object>) level1_2_level2_2.get("values")).get("repeater3");
        Assert.assertEquals("Third level repeater", ((I18nizableText) level1_2_level2_2_level3Repeater.get("label")).getLabel());
        List<Map<String, Object>> level1_2_level2_2_level3 = (List<Map<String, Object>>) level1_2_level2_2_level3Repeater.get("entries");
        Assert.assertEquals(3, level1_2_level2_2_level3.size());
        // repeaterThreeLevels[2]/repeater2[2]/repeater3[1]
        Map<String, Object> level1_2_level2_2_level3_1 = level1_2_level2_2_level3.get(0);
        Assert.assertEquals("221", ((Map<String, Object>) level1_2_level2_2_level3_1.get("values")).get("rep3-string"));
        Assert.assertEquals(1, level1_2_level2_2_level3_1.get("position"));
        // repeaterThreeLevels[2]/repeater2[2]/repeater3[2]
        Map<String, Object> level1_2_level2_2_level3_2 = level1_2_level2_2_level3.get(1);
        Assert.assertEquals("222", ((Map<String, Object>) level1_2_level2_2_level3_2.get("values")).get("rep3-string"));
        Assert.assertEquals(2, level1_2_level2_2_level3_2.get("position"));
        // repeaterThreeLevels[2]/repeater2[2]/repeater3[3]
        Map<String, Object> level1_2_level2_2_level3_3 = level1_2_level2_2_level3.get(2);
        Assert.assertEquals("223", ((Map<String, Object>) level1_2_level2_2_level3_3.get("values")).get("rep3-string"));
        Assert.assertEquals(3, level1_2_level2_2_level3_3.get("position"));
        
        // Checks on repeaterThreeLevels/repeater2/repeater3
        Map<String, Object> level3Repeater = (Map<String, Object>) values.get("repeaterThreeLevels/repeater2/repeater3");
        Assert.assertEquals("Third level repeater", ((I18nizableText) level3Repeater.get("label")).getLabel());
        List<Map<String, Object>> level3RepeaterEntries = (List<Map<String, Object>>) level3Repeater.get("entries");
        Assert.assertEquals(6, level3RepeaterEntries.size());
        
        Map<String, Object> level3Entry111 = level3RepeaterEntries.get(0);
        Assert.assertEquals("111", ((Map<String, Object>) level3Entry111.get("values")).get("rep3-string"));
        Assert.assertEquals(1, level3Entry111.get("position"));
        
        Map<String, Object> level3Entry112 = level3RepeaterEntries.get(1);
        Assert.assertEquals("112", ((Map<String, Object>) level3Entry112.get("values")).get("rep3-string"));
        Assert.assertEquals(2, level3Entry112.get("position"));
        
        Map<String, Object> level3Entry211 = level3RepeaterEntries.get(2);
        Assert.assertEquals("211", ((Map<String, Object>) level3Entry211.get("values")).get("rep3-string"));
        Assert.assertEquals(1, level3Entry211.get("position"));
        
        Map<String, Object> level3Entry221 = level3RepeaterEntries.get(3);
        Assert.assertEquals("221", ((Map<String, Object>) level3Entry221.get("values")).get("rep3-string"));
        Assert.assertEquals(1, level3Entry221.get("position"));
        
        Map<String, Object> level3Entry222 = level3RepeaterEntries.get(4);
        Assert.assertEquals("222", ((Map<String, Object>) level3Entry222.get("values")).get("rep3-string"));
        Assert.assertEquals(2, level3Entry222.get("position"));
        
        Map<String, Object> level3Entry223 = level3RepeaterEntries.get(5);
        Assert.assertEquals("223", ((Map<String, Object>) level3Entry223.get("values")).get("rep3-string"));
        Assert.assertEquals(3, level3Entry223.get("position"));
        
        // Checks on owns/owns/composite/repeater
        Map<String, Object> ownsRepeater = (Map<String, Object>) values.get("owns/owns/composite/repeater");
        Assert.assertEquals("Repeater", ((I18nizableText) ownsRepeater.get("label")).getLabel());
        List<Map<String, Object>> ownsRepeaterValues = (List<Map<String, Object>>) ownsRepeater.get("entries");
        Assert.assertEquals(6, ownsRepeaterValues.size());
        
        Map<String, Object> ownsEntry211 = ownsRepeaterValues.get(0);
        Assert.assertEquals("linked content 21 - entry 1", ((Map<String, Object>) ownsEntry211.get("values")).get("comp-string"));
        Assert.assertEquals(1, ownsEntry211.get("position"));
        
        Map<String, Object> ownsEntry212 = ownsRepeaterValues.get(1);
        Assert.assertEquals("linked content 21 - entry 2", ((Map<String, Object>) ownsEntry212.get("values")).get("comp-string"));
        Assert.assertEquals(2, ownsEntry212.get("position"));
        
        Map<String, Object> ownsEntry221 = ownsRepeaterValues.get(2);
        Assert.assertEquals("linked content 22 - entry 1", ((Map<String, Object>) ownsEntry221.get("values")).get("comp-string"));
        Assert.assertEquals(1, ownsEntry221.get("position"));
        
        Map<String, Object> ownsEntry222 = ownsRepeaterValues.get(3);
        Assert.assertEquals("linked content 22 - entry 2", ((Map<String, Object>) ownsEntry222.get("values")).get("comp-string"));
        Assert.assertEquals(2, ownsEntry222.get("position"));
        
        Map<String, Object> ownsEntry311 = ownsRepeaterValues.get(4);
        Assert.assertEquals("linked content 31 - entry 1", ((Map<String, Object>) ownsEntry311.get("values")).get("comp-string"));
        Assert.assertEquals(1, ownsEntry311.get("position"));
        
        Map<String, Object> ownsEntry312 = ownsRepeaterValues.get(5);
        Assert.assertEquals("linked content 31 - entry 2", ((Map<String, Object>) ownsEntry312.get("values")).get("comp-string"));
        Assert.assertEquals(2, ownsEntry312.get("position"));
        
        // Checks on composite/repeater/comp-owns/owns/composite/repeater
        Map<String, Object> compOwnsRepeater = (Map<String, Object>) values.get("composite/repeater/comp-owns/owns/composite/repeater");
        Assert.assertEquals("Repeater", ((I18nizableText) compOwnsRepeater.get("label")).getLabel());
        List<Map<String, Object>> compOwnsRepeaterValues = (List<Map<String, Object>>) compOwnsRepeater.get("entries");
        Assert.assertEquals(4, compOwnsRepeaterValues.size());
        
        Map<String, Object> compOwnsEntry211 = compOwnsRepeaterValues.get(0);
        Assert.assertEquals("linked content 21 - entry 1", ((Map<String, Object>) compOwnsEntry211.get("values")).get("comp-string"));
        Assert.assertEquals(1, compOwnsEntry211.get("position"));
        
        Map<String, Object> compOwnsEntry212 = compOwnsRepeaterValues.get(1);
        Assert.assertEquals("linked content 21 - entry 2", ((Map<String, Object>) compOwnsEntry212.get("values")).get("comp-string"));
        Assert.assertEquals(2, compOwnsEntry212.get("position"));
        
        Map<String, Object> compOwnsEntry221 = compOwnsRepeaterValues.get(2);
        Assert.assertEquals("linked content 22 - entry 1", ((Map<String, Object>) compOwnsEntry221.get("values")).get("comp-string"));
        Assert.assertEquals(1, compOwnsEntry221.get("position"));
        
        Map<String, Object> compOwnsEntry222 = compOwnsRepeaterValues.get(3);
        Assert.assertEquals("linked content 22 - entry 2", ((Map<String, Object>) compOwnsEntry222.get("values")).get("comp-string"));
        Assert.assertEquals(2, compOwnsEntry222.get("position"));
    }
    
    /**
     * Test the SearchModel API.
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testSearchModelContentValuesExtractor()
    {
        Content content = getContent("content");
        SearchUIModel searchModel = _searchModelEP.getExtension(TestConstants.TEST_SEARCH_MODEL);
        
        Map<String, Object> values = _contentValuesExtractor.create(searchModel)
                                                            .getValues(content, null, new HashMap<>());
        
        Assert.assertEquals(28, values.size());
        
        Assert.assertEquals("Test content", values.get("title"));
        
        User current = _userManager.getUser(AbstractPopulationAwareTestCase.CURRENT_USER_TEST_IDENTITY);
        Map<String, Object> contributorValues = (Map<String, Object>) values.get("contributor");
        Assert.assertEquals(AbstractPopulationAwareTestCase.CURRENT_USER_TEST_IDENTITY.getLogin(), contributorValues.get("login"));
        Assert.assertEquals(AbstractPopulationAwareTestCase.CURRENT_USER_TEST_IDENTITY.getPopulationId(), contributorValues.get("populationId"));
        Assert.assertEquals(current.getLastName(), contributorValues.get("lastname"));
        Assert.assertEquals(current.getFirstName(), contributorValues.get("firstname"));
        Assert.assertTrue(StringUtils.isEmpty((String) contributorValues.get("email")));
        
        Map<String, Object> creatorValues = (Map<String, Object>) values.get("creator");
        Assert.assertEquals(AbstractPopulationAwareTestCase.CURRENT_USER_TEST_IDENTITY.getLogin(), creatorValues.get("login"));
        Assert.assertEquals(AbstractPopulationAwareTestCase.CURRENT_USER_TEST_IDENTITY.getPopulationId(), creatorValues.get("populationId"));
        Assert.assertEquals(current.getLastName(), creatorValues.get("lastname"));
        Assert.assertEquals(current.getFirstName(), creatorValues.get("firstname"));
        Assert.assertTrue(StringUtils.isEmpty((String) creatorValues.get("email")));
        
        Assert.assertTrue(values.containsKey("lastModified"));
        Assert.assertTrue(values.get("creationDate") instanceof String);
        Map<String, Object> workflowStep = (Map<String, Object>) values.get("workflowStep");
        Assert.assertEquals(3, workflowStep.get("stepId"));
        
        Assert.assertEquals(3L, values.get("long"));
        Assert.assertEquals(1.234D, values.get("double"));
        Assert.assertEquals(Boolean.TRUE, values.get("boolean"));
        
        Assert.assertEquals("31590", values.get("address/zipcode"));
        Assert.assertTrue(values.get("address/country") instanceof Map);
        Map<String, Object> country = (Map<String, Object>) values.get("address/country");
        Assert.assertEquals("france", country.get("value"));
        Assert.assertEquals(new I18nizableText("France"), country.get("label"));
        
        Assert.assertTrue(values.get("linkContent") instanceof Map);
        Map<String, Object> linkedContent = (Map<String, Object>) values.get("linkContent");
        _ckecContentProperties(linkedContent, "linked-content-1");
        
        Assert.assertTrue(values.get("linkContents") instanceof List);
        List<Map<String, Object>> linkedContents = (List<Map<String, Object>>) values.get("linkContents");
        Assert.assertEquals(3, linkedContents.size());
        _ckecContentProperties(linkedContents.get(0), "linked-content-2");
        _ckecContentProperties(linkedContents.get(1), "linked-content-3");
        _ckecContentProperties(linkedContents.get(2), "linked-content-4");
        
        Assert.assertEquals("Linked content 1", values.get("linkContent/title"));
        Assert.assertEquals(List.of("Linked content 2", "Linked content 3", "Linked content 4"),
                        values.get("linkContents/title"));
        
        Assert.assertEquals(List.of("entry 1", "entry 2"),
                        values.get("composite/repeater/comp-string"));
        Assert.assertEquals(List.of("Linked content 1", "Linked content 2"),
                        values.get("composite/repeater/comp-owns/title"));
        Assert.assertEquals(List.of("aaaa", "bbbb"),
                        values.get("composite/repeater/comp-owns/abstract"));
        
        Assert.assertTrue(values.get("owns") instanceof List);
        List<Map<String, Object>> owns = (List<Map<String, Object>>) values.get("owns");
        Assert.assertEquals(3, owns.size());
        _ckecContentProperties(owns.get(0), "linked-content-2");
        _ckecContentProperties(owns.get(1), "linked-content-3");
        _ckecContentProperties(owns.get(2), "linked-content-4");
        Assert.assertEquals(List.of("Linked content 2", "Linked content 3", "Linked content 4"),
                        values.get("owns/title"));
        
        Assert.assertTrue(values.get("owns/address/country") instanceof List);
        List<Map<String, Object>> countries = (List<Map<String, Object>>) values.get("owns/address/country");
        Assert.assertEquals(3, countries.size());
        Assert.assertEquals("japan", countries.get(0).get("value"));
        Assert.assertEquals(new I18nizableText("Japan"), countries.get(0).get("label"));
        Assert.assertEquals("spain", countries.get(1).get("value"));
        Assert.assertEquals(new I18nizableText("Spain"), countries.get(1).get("label"));
        Assert.assertEquals("japan", countries.get(2).get("value"));
        Assert.assertEquals(new I18nizableText("Japan"), countries.get(2).get("label"));
        
        Assert.assertEquals(List.of("Linked content 21", "Linked content 22", "Linked content 31"),
                        values.get("owns/owns/title"));
        
        Assert.assertTrue(values.get("owns/address/country") instanceof List);
        List<Map<String, Object>> ownsOwnsCountries = (List<Map<String, Object>>) values.get("owns/owns/address/country");
        Assert.assertEquals(3, ownsOwnsCountries.size());
        Assert.assertEquals("france", ownsOwnsCountries.get(0).get("value"));
        Assert.assertEquals(new I18nizableText("France"), ownsOwnsCountries.get(0).get("label"));
        Assert.assertEquals("japan", ownsOwnsCountries.get(1).get("value"));
        Assert.assertEquals(new I18nizableText("Japan"), ownsOwnsCountries.get(1).get("label"));
        Assert.assertEquals("italy", ownsOwnsCountries.get(2).get("value"));
        Assert.assertEquals(new I18nizableText("Italy"), ownsOwnsCountries.get(2).get("label"));
        
        Assert.assertTrue(values.get("owns/owns/composite/repeater/comp-owns") instanceof List);
        List<Map<String, Object>> compOwns = (List<Map<String, Object>>) values.get("owns/owns/composite/repeater/comp-owns");
        Assert.assertEquals(6, compOwns.size());
        _ckecContentProperties(compOwns.get(0), "eee");
        _ckecContentProperties(compOwns.get(1), "ccc");
        _ckecContentProperties(compOwns.get(2), "linked-content-221");
        _ckecContentProperties(compOwns.get(3), "aaa");
        _ckecContentProperties(compOwns.get(4), "ddd");
        _ckecContentProperties(compOwns.get(5), "bbb");

        Assert.assertEquals(List.of("EEE", "CCC", "Linked content 221", "AAA", "DDD", "BBB"),
                        values.get("owns/owns/composite/repeater/comp-owns/title"));
        
        Assert.assertFalse(values.containsKey("sport"));
        Assert.assertFalse(values.containsKey("simpleContents"));
        
        Assert.assertTrue(values.get("contentLanguage") instanceof Map);
        Map<String, Object> contentLanguageValues = (Map<String, Object>) values.get("contentLanguage");
        Assert.assertEquals("en", contentLanguageValues.get("code"));
        Language contentLanguage = _languageManager.getLanguage("en");
        Assert.assertEquals(contentLanguage.getLabel(), contentLanguageValues.get("label"));
        Assert.assertEquals(contentLanguage.getSmallIcon(), contentLanguageValues.get("icon"));
        
        Assert.assertEquals(List.of(new I18nizableText(__EXHAUSTIVE_CONTENTTYPE_LABEL)), values.get("contentTypes"));
        Assert.assertTrue(((Collection) values.get("mixins")).isEmpty());
    }
    
    private void _ckecContentProperties(Map<String, Object> contentProperties, String contentName)
    {
        Content content = getContent(contentName);
        
        Assert.assertEquals(10, contentProperties.size());
        Assert.assertEquals(content.getId(), contentProperties.get("id"));
        Assert.assertEquals(content.getTitle(), contentProperties.get("title"));
        
        Assert.assertTrue(contentProperties.containsKey("iconGlyph"));
        Assert.assertTrue(contentProperties.containsKey("iconDecorator"));
        Assert.assertTrue(contentProperties.containsKey("smallIcon"));
        Assert.assertTrue(contentProperties.containsKey("mediumIcon"));
        Assert.assertTrue(contentProperties.containsKey("largeIcon"));
        Assert.assertTrue(contentProperties.containsKey("contentTypes"));
        Assert.assertTrue(contentProperties.containsKey("mixins"));
        Assert.assertTrue(contentProperties.containsKey("isSimple"));
    }
}
