/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.search;

import java.util.List;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.search.cocoon.SearchGeneratorHelper;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.plugins.repository.model.CompositeDefinition;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.SimpleViewItemGroup;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElement;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Test case for the search generator
 */
@NeedFeatures({
    "exhaustif/content-type"
})
public class SearchGeneratorTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __EXHAUSTIVE_CONTENT_TYPE_ID = "org.ametys.test.Content.exhaustive";
    
    private static ContentTypeExtensionPoint _contentTypeExtensionPoint;

    private static ContentType _exhaustiveContentType;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }

    public void prepareComponents() throws Exception
    {
        WithCmsFeaturesTestCase.super.prepareComponents();
        
        _exhaustiveContentType = _contentTypeExtensionPoint.getExtension(__EXHAUSTIVE_CONTENT_TYPE_ID);
    }
    
    /**
     * Test {@link SearchGeneratorHelper#copyAndFilterViewItemAccessor(ViewItemAccessor, List)}. Keep some items
     */
    @Test
    public void testCopyAndFilterKeepSomeItems()
    {
        View view = View.of(List.of(_exhaustiveContentType), "title",
                                                             "composite/repeater/comp-string",
                                                             "composite/repeater/comp-long");
        
        List<String> itemsToKeep = List.of("title",
                                           "composite/repeater/comp-long");
        
        View copy = (View) SearchGeneratorHelper.copyAndFilterViewItemAccessor(view, itemsToKeep);
        
        List<ViewItem> copyItems = copy.getViewItems();
        Assert.assertEquals(2, copyItems.size());
        
        // title
        Assert.assertEquals("title", copyItems.get(0).getName());
        
        // composite
        ModelViewItemGroup composite = (ModelViewItemGroup) copyItems.get(1);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeChildren = composite.getViewItems();
        Assert.assertEquals(1, compositeChildren.size());
        
        // composite - repeater
        ModelViewItemGroup repeater = (ModelViewItemGroup) compositeChildren.get(0);
        Assert.assertEquals("repeater", repeater.getName());
        List<ViewItem> repeaterChildren = repeater.getViewItems();
        Assert.assertEquals(1, repeaterChildren.size());
        
        // composite - repeater - comp-long
        Assert.assertEquals("comp-long", repeaterChildren.get(0).getName());
    }
    
    /**
     * Test {@link SearchGeneratorHelper#copyAndFilterViewItemAccessor(ViewItemAccessor, List)}. Items list is empty -> keep all items
     */
    @Test
    public void testCopyAndFilterKeepAllItems()
    {
        View view = View.of(List.of(_exhaustiveContentType), "title",
                                                             "composite/repeater/comp-long");
        
        List<String> itemsToKeep = List.of();
        
        View copy = (View) SearchGeneratorHelper.copyAndFilterViewItemAccessor(view, itemsToKeep);
        
        List<ViewItem> copyItems = copy.getViewItems();
        Assert.assertEquals(2, copyItems.size());
        
        // title
        Assert.assertEquals("title", copyItems.get(0).getName());
        
        // composite
        ModelViewItemGroup composite = (ModelViewItemGroup) copyItems.get(1);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeChildren = composite.getViewItems();
        Assert.assertEquals(1, compositeChildren.size());
        
        // composite - repeater
        ModelViewItemGroup repeater = (ModelViewItemGroup) compositeChildren.get(0);
        Assert.assertEquals("repeater", repeater.getName());
        List<ViewItem> repeaterChildren = repeater.getViewItems();
        Assert.assertEquals(1, repeaterChildren.size());
        
        // composite - repeater - comp-long
        Assert.assertEquals("comp-long", repeaterChildren.get(0).getName());
    }
    
    /**
     * Test {@link SearchGeneratorHelper#copyAndFilterViewItemAccessor(ViewItemAccessor, List)}. Remove all children of a container -> remove the whole hierarchy 
     */
    @Test
    public void testCopyAndFilterRemoveHierarchy()
    {
        View view = View.of(List.of(_exhaustiveContentType), "title",
                                                             "composite/repeater/comp-string",
                                                             "composite/repeater/comp-long");
        
        List<String> itemsToKeep = List.of("title");
        
        View copy = (View) SearchGeneratorHelper.copyAndFilterViewItemAccessor(view, itemsToKeep);
        
        List<ViewItem> copyItems = copy.getViewItems();
        Assert.assertEquals(1, copyItems.size());
        
        // title
        Assert.assertEquals("title", copyItems.get(0).getName());
    }
    
    /**
     * Test {@link SearchGeneratorHelper#copyAndFilterViewItemAccessor(ViewItemAccessor, List)}. Container appears twice, one child is kept -> Remove empty container
     */
    @Test
    public void testCopyAndFilterRemoveEmptyContainer()
    {
        CompositeDefinition compositeDefinition = (CompositeDefinition) _exhaustiveContentType.getModelItem("composite");
        RepeaterDefinition repeaterDefinition = (RepeaterDefinition) compositeDefinition.getModelItem("repeater");
        ElementDefinition compLongDefinition = (ElementDefinition) repeaterDefinition.getModelItem("comp-long");

        View view = new View();

        // composite (1)
        ModelViewItemGroup<CompositeDefinition> firstComposite = new ModelViewItemGroup<>();
        firstComposite.setDefinition(compositeDefinition);
        view.addViewItem(firstComposite);
        
        // composite (1) - repeater (1)
        ModelViewItemGroup<RepeaterDefinition> firstRepeater = new ModelViewItemGroup<>();
        firstRepeater.setDefinition(repeaterDefinition);
        firstComposite.addViewItem(firstRepeater);
        
        // composite (2)
        ModelViewItemGroup<CompositeDefinition> secondComposite = new ModelViewItemGroup<>();
        secondComposite.setDefinition(compositeDefinition);
        view.addViewItem(secondComposite);
        
        // composite (2) - repeater (2)
        ModelViewItemGroup<RepeaterDefinition> secondRepeater = new ModelViewItemGroup<>();
        secondRepeater.setDefinition(repeaterDefinition);
        secondComposite.addViewItem(secondRepeater);
        
        // composite (2) - repeater (2) - comp-long
        ViewElement compLong = new ViewElement();
        compLong.setDefinition(compLongDefinition);
        secondRepeater.addViewItem(compLong);
        
        List<String> itemsToKeep = List.of("composite/repeater/comp-long");
        
        View copy = (View) SearchGeneratorHelper.copyAndFilterViewItemAccessor(view, itemsToKeep);
        
        List<ViewItem> copyItems = copy.getViewItems();
        Assert.assertEquals(1, copyItems.size());
        
        // composite
        ModelViewItemGroup composite = (ModelViewItemGroup) copyItems.get(0);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeChildren = composite.getViewItems();
        Assert.assertEquals(1, compositeChildren.size());
        
        // composite - repeater
        ModelViewItemGroup repeater = (ModelViewItemGroup) compositeChildren.get(0);
        Assert.assertEquals("repeater", repeater.getName());
        List<ViewItem> repeaterChildren = repeater.getViewItems();
        Assert.assertEquals(1, repeaterChildren.size());
        
        // composite - repeater - comp-long
        Assert.assertEquals("comp-long", repeaterChildren.get(0).getName());
    }
    
    /**
     * Test {@link SearchGeneratorHelper#copyAndFilterViewItemAccessor(ViewItemAccessor, List)}. Container appears twice, empty container is kept -> Remove container containing child
     */
    @Test
    public void testCopyAndFilterRemoveContainerWithChild()
    {
        CompositeDefinition compositeDefinition = (CompositeDefinition) _exhaustiveContentType.getModelItem("composite");
        RepeaterDefinition repeaterDefinition = (RepeaterDefinition) compositeDefinition.getModelItem("repeater");
        ElementDefinition compLongDefinition = (ElementDefinition) repeaterDefinition.getModelItem("comp-long");

        View view = new View();

        // composite (1)
        ModelViewItemGroup<CompositeDefinition> firstComposite = new ModelViewItemGroup<>();
        firstComposite.setDefinition(compositeDefinition);
        view.addViewItem(firstComposite);
        
        // composite (1) - repeater (1)
        ModelViewItemGroup<RepeaterDefinition> firstRepeater = new ModelViewItemGroup<>();
        firstRepeater.setDefinition(repeaterDefinition);
        firstComposite.addViewItem(firstRepeater);
        
        // composite (2)
        ModelViewItemGroup<CompositeDefinition> secondComposite = new ModelViewItemGroup<>();
        secondComposite.setDefinition(compositeDefinition);
        view.addViewItem(secondComposite);
        
        // composite (2) - repeater (2)
        ModelViewItemGroup<RepeaterDefinition> secondRepeater = new ModelViewItemGroup<>();
        secondRepeater.setDefinition(repeaterDefinition);
        secondComposite.addViewItem(secondRepeater);
        
        // composite (2) - repeater (2) - comp-long
        ViewElement compLong = new ViewElement();
        compLong.setDefinition(compLongDefinition);
        secondRepeater.addViewItem(compLong);
        
        List<String> itemsToKeep = List.of("composite/repeater");
        
        View copy = (View) SearchGeneratorHelper.copyAndFilterViewItemAccessor(view, itemsToKeep);
        
        List<ViewItem> copyItems = copy.getViewItems();
        Assert.assertEquals(1, copyItems.size());
        
        // composite
        ModelViewItemGroup composite = (ModelViewItemGroup) copyItems.get(0);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeChildren = composite.getViewItems();
        Assert.assertEquals(1, compositeChildren.size());
        
        // composite - repeater
        ModelViewItemGroup repeater = (ModelViewItemGroup) compositeChildren.get(0);
        Assert.assertEquals("repeater", repeater.getName());
        Assert.assertTrue(repeater.getViewItems().isEmpty());
    }
    
    /**
     * Test {@link SearchGeneratorHelper#removeResultItems(ViewItemAccessor, List)}. Keep some items
     */
    @Test
    public void testRemoveResultItemsKeepSomeItems()
    {
        View view = View.of(List.of(_exhaustiveContentType), "title",
                                                             "composite/repeater/comp-string",
                                                             "composite/repeater/comp-long");
        
        List<String> itemsToRemove = List.of("composite/repeater/comp-string");
        
        SearchGeneratorHelper.removeResultItems(view, itemsToRemove);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(2, viewItems.size());
        
        // title
        Assert.assertEquals("title", viewItems.get(0).getName());
        
        // composite
        ModelViewItemGroup composite = (ModelViewItemGroup) viewItems.get(1);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeChildren = composite.getViewItems();
        Assert.assertEquals(1, compositeChildren.size());
        
        // composite - repeater
        ModelViewItemGroup repeater = (ModelViewItemGroup) compositeChildren.get(0);
        Assert.assertEquals("repeater", repeater.getName());
        List<ViewItem> repeaterChildren = repeater.getViewItems();
        Assert.assertEquals(1, repeaterChildren.size());
        
        // composite - repeater - comp-long
        Assert.assertEquals("comp-long", repeaterChildren.get(0).getName());
    }
    
    /**
     * Test {@link SearchGeneratorHelper#removeResultItems(ViewItemAccessor, List)}. Remove a container's child -> remove the whole hierarchy 
     */
    @Test
    public void testRemoveResultItemsRemoveHierarchy()
    {
        View view = View.of(List.of(_exhaustiveContentType), "title",
                                                             "composite/repeater/comp-string");
        
        List<String> itemsToRemove = List.of("composite/repeater/comp-string");
        
        SearchGeneratorHelper.removeResultItems(view, itemsToRemove);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(1, viewItems.size());
        
        // title
        Assert.assertEquals("title", viewItems.get(0).getName());
    }
    
    /**
     * Test {@link SearchGeneratorHelper#removeResultItems(ViewItemAccessor, List)}. Container appears twice, empty container is removed
     */
    @Test
    public void testRemoveResultItemsRemoveEmptyContainer()
    {
        CompositeDefinition compositeDefinition = (CompositeDefinition) _exhaustiveContentType.getModelItem("composite");
        RepeaterDefinition repeaterDefinition = (RepeaterDefinition) compositeDefinition.getModelItem("repeater");
        ElementDefinition compLongDefinition = (ElementDefinition) repeaterDefinition.getModelItem("comp-long");

        View view = new View();

        // composite (1)
        ModelViewItemGroup<CompositeDefinition> firstComposite = new ModelViewItemGroup<>();
        firstComposite.setDefinition(compositeDefinition);
        view.addViewItem(firstComposite);
        
        // composite (1) - repeater (1)
        ModelViewItemGroup<RepeaterDefinition> firstRepeater = new ModelViewItemGroup<>();
        firstRepeater.setDefinition(repeaterDefinition);
        firstComposite.addViewItem(firstRepeater);
        
        // composite (1) - repeater (1) - comp-long
        ViewElement compLong = new ViewElement();
        compLong.setDefinition(compLongDefinition);
        firstRepeater.addViewItem(compLong);
        
        // composite (2)
        ModelViewItemGroup<CompositeDefinition> secondComposite = new ModelViewItemGroup<>();
        secondComposite.setDefinition(compositeDefinition);
        view.addViewItem(secondComposite);
        
        // composite (2) - repeater (2)
        ModelViewItemGroup<RepeaterDefinition> secondRepeater = new ModelViewItemGroup<>();
        secondRepeater.setDefinition(repeaterDefinition);
        secondComposite.addViewItem(secondRepeater);
        
        List<String> itemsToRemove = List.of("composite/repeater");
        
        SearchGeneratorHelper.removeResultItems(view, itemsToRemove);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(1, viewItems.size());
        
        // composite
        ModelViewItemGroup composite = (ModelViewItemGroup) viewItems.get(0);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeChildren = composite.getViewItems();
        Assert.assertEquals(1, compositeChildren.size());
        
        // composite - repeater
        ModelViewItemGroup repeater = (ModelViewItemGroup) compositeChildren.get(0);
        Assert.assertEquals("repeater", repeater.getName());
        List<ViewItem> repeaterChildren = repeater.getViewItems();
        Assert.assertEquals(1, repeaterChildren.size());
        
        // composite - repeater - comp-long
        Assert.assertEquals("comp-long", repeaterChildren.get(0).getName());
    }
    
    /**
     * Test {@link SearchGeneratorHelper#removeResultItems(ViewItemAccessor, List)}. Container appears twice, with some groups. The right containers are removed
     */
    @Test
    public void testRemoveResultItemsWithGroups()
    {
        CompositeDefinition compositeDefinition = (CompositeDefinition) _exhaustiveContentType.getModelItem("composite");
        RepeaterDefinition repeaterDefinition = (RepeaterDefinition) compositeDefinition.getModelItem("repeater");
        ElementDefinition compLongDefinition = (ElementDefinition) repeaterDefinition.getModelItem("comp-long");

        View view = new View();

        // composite (1)
        ModelViewItemGroup<CompositeDefinition> firstComposite = new ModelViewItemGroup<>();
        firstComposite.setDefinition(compositeDefinition);
        view.addViewItem(firstComposite);
        
        // composite (1) - group1
        SimpleViewItemGroup group1 = new SimpleViewItemGroup();
        firstComposite.addViewItem(group1);
        
        // composite (1) - group1 - repeater (1)
        ModelViewItemGroup<RepeaterDefinition> firstRepeater = new ModelViewItemGroup<>();
        firstRepeater.setDefinition(repeaterDefinition);
        group1.addViewItem(firstRepeater);

        // composite (1) - group1 - repeater (1) - group2
        SimpleViewItemGroup group2 = new SimpleViewItemGroup();
        firstRepeater.addViewItem(group2);
        
        // composite (1) - group1 - repeater (1) - group2 - comp-long
        ViewElement compLong = new ViewElement();
        compLong.setDefinition(compLongDefinition);
        group2.addViewItem(compLong);
        
        // composite (2)
        ModelViewItemGroup<CompositeDefinition> secondComposite = new ModelViewItemGroup<>();
        secondComposite.setDefinition(compositeDefinition);
        view.addViewItem(secondComposite);
        
        // composite (2) - group3
        SimpleViewItemGroup group3 = new SimpleViewItemGroup();
        secondComposite.addViewItem(group3);
        
        // composite (2) - group3 - repeater (2)
        ModelViewItemGroup<RepeaterDefinition> secondRepeater = new ModelViewItemGroup<>();
        secondRepeater.setDefinition(repeaterDefinition);
        group3.addViewItem(secondRepeater);
        
        List<String> itemsToRemove = List.of("composite/repeater");
        
        SearchGeneratorHelper.removeResultItems(view, itemsToRemove);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(1, viewItems.size());
        
        // composite
        ModelViewItemGroup composite = (ModelViewItemGroup) viewItems.get(0);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeChildren = composite.getViewItems();
        Assert.assertEquals(1, compositeChildren.size());
        
        // composite - group1
        SimpleViewItemGroup group1Result = (SimpleViewItemGroup) compositeChildren.get(0);
        List<ViewItem> group1Children = group1Result.getViewItems();
        Assert.assertEquals(1, group1Children.size());
        
        // composite - group1 - repeater
        ModelViewItemGroup repeater = (ModelViewItemGroup) group1Children.get(0);
        Assert.assertEquals("repeater", repeater.getName());
        List<ViewItem> repeaterChildren = repeater.getViewItems();
        Assert.assertEquals(1, repeaterChildren.size());
        
        // composite - group1 - repeater - group2
        SimpleViewItemGroup group2Result = (SimpleViewItemGroup) repeaterChildren.get(0);
        List<ViewItem> group2Children = group2Result.getViewItems();
        Assert.assertEquals(1, group2Children.size());
        
        // composite - group1 - repeater - group2 - comp-long
        Assert.assertEquals("comp-long", group2Children.get(0).getName());
    }
}
