/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.search;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.repository.Content;
import org.ametys.cms.search.SearchField;
import org.ametys.cms.search.SearchResult;
import org.ametys.cms.search.SearchResults;
import org.ametys.cms.search.Sort;
import org.ametys.cms.search.Sort.Order;
import org.ametys.cms.search.query.DocumentTypeQuery;
import org.ametys.cms.search.query.FullTextQuery;
import org.ametys.cms.search.query.JoinQuery;
import org.ametys.cms.search.query.MatchAllQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.query.QuerySyntaxException;
import org.ametys.cms.search.query.StringQuery;
import org.ametys.cms.search.query.WorkflowStepQuery;
import org.ametys.cms.search.query.join.JoinKey;
import org.ametys.cms.search.solr.SearcherFactory;
import org.ametys.cms.search.solr.SearcherFactory.Searcher;
import org.ametys.cms.search.solr.field.ContentTypeSearchField;
import org.ametys.cms.search.solr.field.StringSearchField;
import org.ametys.cms.search.solr.field.WorkflowStepSearchField;
import org.ametys.cms.test.framework.AbstractSearcherTestCase;
import org.ametys.plugins.repository.AmetysObjectIterable;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Test the {@link Searcher} API.
 */
@NeedFeatures({
    "cms/org.ametys.cms.content.search.factory.solr",
    
    "exhaustif/content-type",
    "cms/org.ametys.cms.data.holder.helpers"
})
public class SearcherTestCase extends AbstractSearcherTestCase implements WithCmsTestWebappFeatures
{
    private static SearcherFactory _searcher;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _searcher = (SearcherFactory) manager.lookup(SearcherFactory.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        _initialAssert();
        
        // Populate and index the repository, then add fake documents.
        populateRepositoryAndIndex();
        addFakeDocuments();
    }
    
    private static void _initialAssert() throws Exception
    {
        // Search the empty repository.
        AmetysObjectIterable<Content> contents = _searcher.create()
                .withQueryString("*:*")
                .withFilterQueryStrings("_documentType:content")
                .setCheckRights(false)
                .search();
        
        Assert.assertEquals(0, contents.getSize());
        Assert.assertFalse(contents.iterator().hasNext());
    }
    
    /**
     * Test simple string API.
     * @throws Exception If an error occurs
     */
    @Test
    public void testStringApi() throws Exception
    {
        AmetysObjectIterable<Content> contents;
        
        contents = _searchContents("*:*");
        Assert.assertEquals(24, contents.getSize());
        Assert.assertTrue(contents.iterator().hasNext());
        
        contents = _searchContents("workflowStep:1");
        Assert.assertEquals(15, contents.getSize());
        
        contents = _searchContents("workflowStep:2");
        Assert.assertEquals(6, contents.getSize());
        
        contents = _searchContents("workflowStep:3");
        Assert.assertEquals(3, contents.getSize());
        
        contents = _searchContents("workflowStep:1 OR workflowStep:3");
        Assert.assertEquals(18, contents.getSize());
        
        contents = _searchContents("address/country_s:[* TO *]");
        Assert.assertEquals(14, contents.getSize());
        
        contents = _searchContents("address/country_s:japan");
        Assert.assertEquals(4, contents.getSize());
        
        contents = _searchContents("{!ametys join=linkContent q=\"title_s:\\\"Linked content 1\\\"\"}");
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        contents = _searchContents("{!ametys join=linkContent%linkContents q=\"title_s:\\\"Linked content 3\\\"\"}");
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        contents = _searchContents("{!ametys join=linkContents q=\"title_s:\\\"Linked content 2\\\"\"}");
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        contents = _searchContents("{!ametys join=linkContents q=\"title_s:\\\"Nonexisting\\\"\"}");
        Assert.assertFalse(contents.iterator().hasNext());
        
        contents = _searchContents("{!ametys join=linkContents q=\"address/country_s:\\\"japan\\\"\"}");
        Assert.assertEquals(2, contents.getSize());
        
        contents = _searchContents("{!ametys join=linkContent->linkContent q=\"title_s:\\\"Linked content 11\\\"\"}");
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        contents = _searchContents("{!ametys join=linkContents->linkContents->linkContent q=\"title_s:\\\"Linked content 221\\\"\"}");
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        contents = _searchContents("{!ametys join=composite/repeater q=\"comp-string_s:\\\"entry 1\\\"\"}");
        Assert.assertEquals(1, contents.getSize());
        
        contents = _searchContents("{!ametys join=composite/repeater->comp-owns q=\"title_s:\\\"Linked content 1\\\"\"}");
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        contents = _searcher.create()
                            .withQueryString("*:*")
                            .withFilterQueryStrings("_documentType:content")
                            .withLimits(0, 5)
                            .setCheckRights(false)
                            .search();
        Assert.assertEquals(5, contents.getSize());
        assertIterableSize(5, contents);
        
        Sort titleAsc = new Sort("title_s_sort", Order.ASC);
        contents = _searcher.create()
                            .withQueryString("*:*")
                            .withFilterQueryStrings("_documentType:content")
                            .withSort(titleAsc)
                            .setCheckRights(false)
                            .search();
        Assert.assertEquals("AAA", contents.iterator().next().getTitle(null));
        
        Sort titleDesc = new Sort("title_s_sort", Order.DESC);
        contents = _searcher.create()
                            .withQueryString("*:*")
                            .withFilterQueryStrings("_documentType:content")
                            .withSort(titleDesc)
                            .setCheckRights(false)
                            .search();
        Assert.assertEquals("ZZZ", contents.iterator().next().getTitle(null));
        
        Sort wfDesc = new Sort("workflowStep", Order.DESC);
        contents = _searcher.create()
                            .withQueryString("*:*")
                            .withFilterQueryStrings("_documentType:content")
                            .withSort(wfDesc, titleDesc)
                            .setCheckRights(false)
                            .search();
        Iterator<Content> it = contents.iterator();
        Assert.assertEquals("Validated content DDD", it.next().getTitle(null));
        Assert.assertEquals("Validated content CCC", it.next().getTitle(null));
        Assert.assertEquals("Test content", it.next().getTitle(null));
        
        contents = _searcher.create()
                            .withQueryString("*:*")
                            .withFilterQueryStrings("_documentType:content")
                            .withSort(wfDesc, titleDesc)
                            .withLimits(0, 3)
                            .setCheckRights(false)
                            .search();
        Assert.assertEquals(3, contents.getSize());
        
        it = contents.iterator();
        Assert.assertEquals("Validated content DDD", it.next().getTitle(null));
        Assert.assertEquals("Validated content CCC", it.next().getTitle(null));
        Assert.assertEquals("Test content", it.next().getTitle(null));
        Assert.assertFalse(it.hasNext());
    }
    
    private AmetysObjectIterable<Content> _searchContents(String string) throws Exception
    {
        return _searcher.create()
                        .withQueryString(string)
                        .withFilterQueryStrings("_documentType:content")
                        .setCheckRights(false)
                        .search();
    }
    
    /**
     * Test the simple query API.
     * @throws Exception If an error occurs
     */
    @Test
    public void testQueryApi() throws Exception
    {
        Query contentQuery = new DocumentTypeQuery("content");
        Query draftQuery = new WorkflowStepQuery(1);
        Query proposedQuery = new WorkflowStepQuery(2);
        Query validatedQuery = new WorkflowStepQuery(3);
        
        AmetysObjectIterable<Content> contents;
                
        contents = _searchContents(new MatchAllQuery());
        Assert.assertEquals(24, contents.getSize());
        Assert.assertTrue(contents.iterator().hasNext());
        
        contents = _searchContents(draftQuery);
        Assert.assertEquals(15, contents.getSize());
        
        contents = _searchContents(proposedQuery);
        Assert.assertEquals(6, contents.getSize());
        
        contents = _searchContents(validatedQuery);
        Assert.assertEquals(3, contents.getSize());
        
        contents = _searchContents(new WorkflowStepQuery(1, 3));
        Assert.assertEquals(18, contents.getSize());
        
        contents = _searchContents(new StringQuery("address/country"));
        Assert.assertEquals(14, contents.getSize());
        
        contents = _searchContents(new StringQuery("address/country", "japan"));
        Assert.assertEquals(4, contents.getSize());
        
        Query joinQuery = new JoinQuery(new StringQuery("title", "Linked content 1"), "linkContent");
        contents = _searchContents(joinQuery);
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        joinQuery = new JoinQuery(new StringQuery("title", "Linked content 3"), new JoinKey("linkContent", "linkContents", null));
        contents = _searchContents(joinQuery);
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        joinQuery = new JoinQuery(new StringQuery("title", "Linked content 2"), "linkContents");
        contents = _searchContents(joinQuery);
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        joinQuery = new JoinQuery(new StringQuery("title", "Nonexisting"), "linkContents");
        contents = _searchContents(joinQuery);
        Assert.assertFalse(contents.iterator().hasNext());
        
        joinQuery = new JoinQuery(new StringQuery("address/country", "japan"), "linkContents");
        contents = _searchContents(joinQuery);
        Assert.assertEquals(2, contents.getSize());
        
        joinQuery = new JoinQuery(new StringQuery("title", "Linked content 11"), "linkContent", "linkContent");
        contents = _searchContents(joinQuery);
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        joinQuery = new JoinQuery(new JoinKey("linkContent"), 
                                  new JoinKey("linkContent", new StringQuery("title", "Linked content 11"))); // same as before, built in a different way
        contents = _searchContents(joinQuery);
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        joinQuery = new JoinQuery(new JoinKey("linkContent", new StringQuery("title", "Linked content 1")), 
                                  new JoinKey("linkContent", new StringQuery("title", "Linked content 11")));
        contents = _searchContents(joinQuery);
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        joinQuery = new JoinQuery(new JoinKey("linkContent", new StringQuery("title", "Nonexisting")), 
                                  new JoinKey("linkContent", new StringQuery("title", "Linked content 11")));
        contents = _searchContents(joinQuery);
        Assert.assertEquals(0, contents.getSize());
        
        joinQuery = new JoinQuery(new StringQuery("title", "Linked content 221"), "linkContents", "linkContents", "linkContent");
        contents = _searchContents(joinQuery);
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        joinQuery = new JoinQuery(new StringQuery("comp-string", "entry 1"), "composite/repeater");
        contents = _searchContents(joinQuery);
        Assert.assertEquals(1, contents.getSize());
        
        joinQuery = new JoinQuery(new StringQuery("title", "Linked content 1"), "composite/repeater", "comp-owns");
        contents = _searchContents(joinQuery);
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        // test inclusion of joins with quotes in query
        joinQuery = new JoinQuery(new JoinQuery(new FullTextQuery("Linked content 11", Operator.EQ), "linkContent"), "linkContent"); //Test Content -> link1 -> link11
        // (yes it is factorizable in a single JoinQUery but in practice, it can be a OrQuery with queries we do not know the nature (in FO search for instance))
        contents = _searchContents(joinQuery);
        Assert.assertEquals(1, contents.getSize());
        Assert.assertEquals("Test content", contents.iterator().next().getTitle(null));
        
        contents = _searcher.create()
                            .withQuery(new MatchAllQuery())
                            .withFilterQueries(contentQuery)
                            .withLimits(0, 5)
                            .setCheckRights(false)
                            .search();
        Assert.assertEquals(5, contents.getSize());
        assertIterableSize(5, contents);

        Sort titleAsc = new Sort(new StringSearchField("title"), Order.ASC);
        contents = _searcher.create()
                            .withQuery(new MatchAllQuery())
                            .withFilterQueries(contentQuery)
                            .withSort(titleAsc)
                            .setCheckRights(false)
                            .search();
        Assert.assertEquals("AAA", contents.iterator().next().getTitle(null));
        
        Sort titleDesc = new Sort(new StringSearchField("title"), Order.DESC);
        contents = _searcher.create()
                            .withQuery(new MatchAllQuery())
                            .withFilterQueries(contentQuery)
                            .withSort(titleDesc)
                            .setCheckRights(false)
                            .search();
        Assert.assertEquals("ZZZ", contents.iterator().next().getTitle(null));
        
        Sort wfDesc = new Sort(new WorkflowStepSearchField(), Order.DESC);
        contents = _searcher.create()
                            .withQuery(new MatchAllQuery())
                            .withFilterQueries(contentQuery)
                            .withSort(wfDesc, titleDesc)
                            .setCheckRights(false)
                            .search();
        Iterator<Content> it = contents.iterator();
        Assert.assertEquals("Validated content DDD", it.next().getTitle(null));
        Assert.assertEquals("Validated content CCC", it.next().getTitle(null));
        Assert.assertEquals("Test content", it.next().getTitle(null));
        
        contents = _searcher.create()
                            .withQuery(new MatchAllQuery())
                            .withFilterQueries(contentQuery)
                            .withSort(wfDesc, titleDesc)
                            .withLimits(0, 3)
                            .setCheckRights(false)
                            .search();
        Assert.assertEquals(3, contents.getSize());
        
        it = contents.iterator();
        Assert.assertEquals("Validated content DDD", it.next().getTitle(null));
        Assert.assertEquals("Validated content CCC", it.next().getTitle(null));
        Assert.assertEquals("Test content", it.next().getTitle(null));
        Assert.assertFalse(it.hasNext());
        
        contents = _searchContents(new FullTextQuery("japan"));
        Assert.assertEquals(5, contents.getSize());
        
        contents = _searchContents(new FullTextQuery("1.234"));
        Assert.assertEquals(0, contents.getSize());
    }

    /**
     * Test search on customfull system property.
     */
    @Test
    @NeedFeatures("content.search/content-type.exhaustive.overrides.fulltext.search")
    public void testViewBasedFullTextProperty()
    {
        Assert.assertEquals(2, _searchContents(new FullTextQuery("japan", "full", null, Operator.SEARCH)).getSize());
        Assert.assertEquals(1, _searchContents(new FullTextQuery("1.234", "full", null, Operator.SEARCH)).getSize());
        Assert.assertEquals(1, _searchContents(new FullTextQuery("france", "full", null, Operator.SEARCH)).getSize());
        Assert.assertEquals(2, _searchContents(new FullTextQuery("spain", "full", null, Operator.SEARCH)).getSize());
    }
    
    private AmetysObjectIterable<Content> _searchContents(Query query)
    {
        try
        {
            return _searcher.create()
                            .withQuery(query)
                            .withFilterQueries(new DocumentTypeQuery("content"))
                            .setCheckRights(false)
                            .search();
        }
        catch (Exception e)
        {
            Object strQuery = null;
            try
            {
                strQuery = query.build();
            }
            catch (QuerySyntaxException qse)
            {
                // let strQuery be null
            }
            Assert.fail("An error occured when querying Solr server with query:\n " + strQuery + "\n" + e);
            return null;
        }
    }
    
    /**
     * Test query with facets API.
     * @throws Exception If an error occurs
     */
    @Test
    public void testQueryWithFacets() throws Exception
    {
        Query contentQuery = new DocumentTypeQuery("content");
        
        SearchResults<Content> results;
        
        List<SearchField> facets = Arrays.asList(new WorkflowStepSearchField(), new StringSearchField("address/country"), new ContentTypeSearchField());
        
        results = _searcher.create()
                            .withQuery(new MatchAllQuery())
                            .withFilterQueries(contentQuery)
                            .withFacets(facets)
                            .setCheckRights(false)
                            .searchWithFacets();
        Assert.assertEquals(24, results.getTotalCount());
        Assert.assertEquals(24, results.getResults().getSize());
        assertIterableSize(24, results.getObjectIds());
        
        assertWorkflowFacetCounts(results);
        assertCountryFacetCounts(results);
        
        Map<String, List<String>> facetValues = new HashMap<>();
        facetValues.put("address/country", Arrays.asList("france"));

        results = _searcher.create()
                            .withQuery(new MatchAllQuery())
                            .withFilterQueries(contentQuery)
                            .withFacets(facets)
                            .withFacetValues(facetValues)
                            .setCheckRights(false)
                            .searchWithFacets();
        Assert.assertEquals(4, results.getTotalCount());
        
        assertFranceWorkflowFacetCounts(results);
        assertCountryFacetCounts(results);
        
        results = _searcher.create()
                            .withQuery(new MatchAllQuery())
                            .withFilterQueries(contentQuery)
                            .withFacets(facets)
                            .withLimits(0, 5)
                            .setCheckRights(false)
                            .searchWithFacets();
        Assert.assertEquals(24, results.getTotalCount());
        Assert.assertEquals(5, results.getResults().getSize());
        
        assertWorkflowFacetCounts(results);
        assertCountryFacetCounts(results);
        
        results = _searcher.create()
                            .withQuery(new MatchAllQuery())
                            .withFilterQueries(contentQuery)
                            .withFacets(facets)
                            .withFacetValues(facetValues)
                            .withLimits(0, 2)
                            .setCheckRights(false)
                            .searchWithFacets();
        Assert.assertEquals(4, results.getTotalCount());
        Assert.assertEquals(2, results.getResults().getSize());
        
        assertFranceWorkflowFacetCounts(results);
        assertCountryFacetCounts(results);
    }
    
    /**
     * Test query with joined facets API.
     * @throws Exception If an error occurs
     */
    @Test
    public void testQueryWithJoinFacets() throws Exception
    {
        Query contentQuery = new DocumentTypeQuery("content");
        
        SearchResults<Content> results;
        List<SearchField> facets = Arrays.asList(new StringSearchField(List.of("linkContents"), "address/country"));
        
        results = _searcher.create()
                           .withQuery(new MatchAllQuery())
                           .withFilterQueries(contentQuery)
                           .withFacets(facets)
                           .setCheckRights(false)
                           .searchWithFacets();

        Assert.assertEquals(24, results.getTotalCount());
        
        // content->linkContents->address/country = (japan|spain)
        // link2->linkContents->address/country = (france|japan)
        // link3->linkContents->address/country = france
        assertFacetValueCount(2, results, "linkContents/address/country", "france");
        assertFacetValueCount(2, results, "linkContents/address/country", "japan");
        assertFacetValueCount(1, results, "linkContents/address/country", "spain");
        assertNullFacetValue(results, "linkContents/address/country", "china");
        
        results = _searcher.create()
                           .withQuery(new MatchAllQuery())
                           .withFilterQueries(contentQuery)
                           .withFacets(facets)
                           .withFacetValues(Map.of("linkContents/address/country", List.of("japan")))
                           .setCheckRights(false)
                           .searchWithFacets();
        
        Assert.assertEquals(2, results.getTotalCount());

        // content->linkContents->address/country = (japan|spain)
        // link2->linkContents->address/country = (france|japan)
        // link3->linkContents->address/country = france
        assertFacetValueCount(2, results, "linkContents/address/country", "france");
        assertFacetValueCount(2, results, "linkContents/address/country", "japan");
        assertFacetValueCount(1, results, "linkContents/address/country", "spain");
        assertNullFacetValue(results, "linkContents/address/country", "china");

        facets = Arrays.asList(new StringSearchField("address/country"), new StringSearchField(List.of("linkContents"), "address/country"));
        
        results = _searcher.create()
                           .withQuery(new MatchAllQuery())
                           .withFilterQueries(contentQuery)
                           .withFacets(facets)
                           .setCheckRights(false)
                           .searchWithFacets();

        Assert.assertEquals(24, results.getTotalCount());
        
        // content->linkContents->address/country = (japan|spain)
        // link2->linkContents->address/country = (france|japan)
        // link3->linkContents->address/country = france
        assertFacetValueCount(2, results, "linkContents/address/country", "france");
        assertFacetValueCount(2, results, "linkContents/address/country", "japan");
        assertFacetValueCount(1, results, "linkContents/address/country", "spain");
        assertNullFacetValue(results, "linkContents/address/country", "china");
        assertCountryFacetCounts(results);
        
        results = _searcher.create()
                           .withQuery(new MatchAllQuery())
                           .withFilterQueries(contentQuery)
                           .withFacets(facets)
                           .withFacetValues(Map.of("address/country", List.of("france", "spain")))
                           .setCheckRights(false)
                           .searchWithFacets();
        
        Assert.assertEquals(7, results.getTotalCount());
        
        // content->linkContents->address/country = (japan|spain)
        // link3->linkContents->address/country = france
        assertFacetValueCount(1, results, "linkContents/address/country", "france");
        assertFacetValueCount(1, results, "linkContents/address/country", "japan");
        assertFacetValueCount(1, results, "linkContents/address/country", "spain");
        assertNullFacetValue(results, "linkContents/address/country", "china");
        assertCountryFacetCounts(results);

        facets = Arrays.asList(new StringSearchField(List.of("linkContents", "linkContents"), "address/country"));
        
        results = _searcher.create()
                           .withQuery(new MatchAllQuery())
                           .withFilterQueries(contentQuery)
                           .withFacets(facets)
                           .setCheckRights(false)
                           .searchWithFacets();
        
        // content->linkContents->linkContents->address/country = (japan|spain)
        assertFacetValueCount(1, results, "linkContents/linkContents/address/country", "france");
        assertFacetValueCount(1, results, "linkContents/linkContents/address/country", "japan");
        assertNullFacetValue(results, "linkContents/linkContents/address/country", "spain");
    }
    
    /**
     * Test query with joined sorts.
     * @throws Exception If an error occurs
     */
    @Test
    public void testQueryWithJoinSort() throws Exception
    {
        Query contentQuery = new DocumentTypeQuery("content");
        
        AmetysObjectIterable<Content> results;
        
        results = _searcher.create()
                           .withQuery(new MatchAllQuery())
                           .withFilterQueries(contentQuery)
                           .setCheckRights(false)
                           .addSort(new Sort(new StringSearchField(List.of("linkContent"), "abstract"), Order.DESC))
                           .search();

        Iterator<Content> it = results.iterator();
        Assert.assertEquals("Linked content 22", it.next().getTitle());
        Assert.assertEquals("Linked content 3", it.next().getTitle());
        Assert.assertEquals("Linked content 1", it.next().getTitle());
        Assert.assertEquals("Test content", it.next().getTitle());
    }
    
    /**
     * Test query with facet and sort API.
     * @throws Exception If an error occurs
     */
    @Test
    public void testQueryWithFacetsAndSort() throws Exception
    {
        SearchResults<Content> results;
        Iterator<SearchResult<Content>> it;
        
        Query contentQuery = new DocumentTypeQuery("content");
        
        // Sort by workflow descending then title descending.
        Sort wfDesc = new Sort(new WorkflowStepSearchField(), Order.DESC);
        Sort titleDesc = new Sort(new StringSearchField("title"), Order.DESC);
        List<Sort> sort = Arrays.asList(wfDesc, titleDesc);
        
        List<SearchField> facets = Arrays.asList(new WorkflowStepSearchField(), new StringSearchField("address/country"), new ContentTypeSearchField());
        
        // Query, sort, facets
        results = _searcher.create()
                            .withQuery(new MatchAllQuery())
                            .withFilterQueries(contentQuery)
                            .withSort(sort)
                            .withFacets(facets)
                            .setCheckRights(false)
                            .searchWithFacets();
        Assert.assertEquals(24, results.getTotalCount());
        Assert.assertEquals(24, results.getResults().getSize());
        
        assertWorkflowFacetCounts(results);
        assertCountryFacetCounts(results);
        
        it = results.getResults().iterator();
        Assert.assertEquals("Validated content DDD", it.next().getObject().getTitle(null));
        Assert.assertEquals("Validated content CCC", it.next().getObject().getTitle(null));
        Assert.assertEquals("Test content", it.next().getObject().getTitle(null));
        
        
        // Query, sort, facets, and specific facet values.
        Map<String, List<String>> facetValues = new HashMap<>();
        facetValues.put("address/country", Arrays.asList("france"));
        
        results = _searcher.create()
                            .withQuery(new MatchAllQuery())
                            .withFilterQueries(contentQuery)
                            .withSort(sort)
                            .withFacets(facets)
                            .withFacetValues(facetValues)
                            .setCheckRights(false)
                            .searchWithFacets();
        Assert.assertEquals(4, results.getTotalCount());
        
        assertFranceWorkflowFacetCounts(results);
        assertCountryFacetCounts(results);
        
        it = results.getResults().iterator();
        Assert.assertEquals("Test content", it.next().getObject().getTitle(null));
        Assert.assertEquals("Proposed content BBB", it.next().getObject().getTitle(null));
        
        
        // Query, sort, facets, and limits.
        results = _searcher.create()
                            .withQuery(new MatchAllQuery())
                            .withFilterQueries(contentQuery)
                            .withSort(sort)
                            .withFacets(facets)
                            .withLimits(0, 2)
                            .setCheckRights(false)
                            .searchWithFacets();
        Assert.assertEquals(24, results.getTotalCount());
        Assert.assertEquals(2, results.getResults().getSize());
        
        assertWorkflowFacetCounts(results);
        assertCountryFacetCounts(results);
        
        it = results.getResults().iterator();
        Assert.assertEquals("Validated content DDD", it.next().getObject().getTitle(null));
        Assert.assertEquals("Validated content CCC", it.next().getObject().getTitle(null));
        Assert.assertFalse(it.hasNext());
        
        
        // Query, sort, facets, specific facet values, and limits.
        results = _searcher.create()
                            .withQuery(new MatchAllQuery())
                            .withFilterQueries(contentQuery)
                            .withSort(sort)
                            .withFacets(facets)
                            .withFacetValues(facetValues)
                            .withLimits(0, 2)
                            .setCheckRights(false)
                            .searchWithFacets();
        Assert.assertEquals(4, results.getTotalCount());
        Assert.assertEquals(2, results.getResults().getSize());
        
        assertFranceWorkflowFacetCounts(results);
        assertCountryFacetCounts(results);
        
        it = results.getResults().iterator();
        Assert.assertEquals("Test content", it.next().getObject().getTitle(null));
        Assert.assertEquals("Proposed content BBB", it.next().getObject().getTitle(null));
        Assert.assertFalse(it.hasNext());
    }
    
}
