/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.search.advanced;

import java.util.Arrays;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.search.advanced.AdvancedQueryBuilder;
import org.ametys.cms.search.advanced.TreeInternalNode;
import org.ametys.cms.search.advanced.TreeLeaf;
import org.ametys.cms.search.advanced.utils.TreePrinter;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.Query.LogicalOperator;
import org.ametys.cms.search.query.Query.Operator;
import org.ametys.cms.search.query.StringQuery;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.util.LambdaUtils;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Tests {@link AdvancedQueryBuilder}
 */
@NeedFeatures({
    "cms/org.ametys.cms.content.search.components",
    "cms/workflow.content"
})
public class AdvancedQueryBuilderTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static AdvancedQueryBuilder _advancedQueryBuilder;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _advancedQueryBuilder = (AdvancedQueryBuilder) manager.lookup(AdvancedQueryBuilder.ROLE);
    }

    /**
     * Tests {@link AdvancedQueryBuilder#build(org.ametys.cms.search.advanced.AbstractTreeNode, java.util.function.Function) AdvancedQueryBuilder#build}
     * which builds a query corresponding to a tree of criteria (advanced search)
     * @throws Exception if an error occurs.
     */
    @Test
    public void testBuild() throws Exception
    {
        /*
         * ____[__OR__]_________
         * ______/___\__________
         * _"crit1"__[_AND_]____
         * ___________/__\______
         * _____"crit2"__"title"
         */
        
        TreeLeaf<DummyCriterion> crit1 = new TreeLeaf<>(new DummyCriterion("crit1", "1"));
        TreeLeaf<DummyCriterion> crit2 = new TreeLeaf<>(new DummyCriterion("crit2", "two"));
        TreeLeaf<DummyCriterion> title = new TreeLeaf<>(new DummyCriterion("title", "article", Operator.SEARCH));
        TreeInternalNode<DummyCriterion> andNode = new TreeInternalNode<>(Arrays.asList(crit2, title), LogicalOperator.AND);
        TreeInternalNode<DummyCriterion> tree = new TreeInternalNode<>(Arrays.asList(crit1, andNode), LogicalOperator.OR);
        _logger.info("\n" + TreePrinter.print(tree, LambdaUtils.wrap(c -> c.toQuery().build())));
        
        Query query = _advancedQueryBuilder.build(tree, crit -> crit.toQuery());
        Assert.assertEquals(
                "("
                    + "crit1_s:\"1\""
                + ") OR ("
                    + "("
                        + "crit2_s:\"two\""
                    + ") AND ("
                        + "title_txt:(article)"
                    + ")"
                + ")", 
                query.build());
    }
    
    private static final class DummyCriterion
    {
        private String _name;
        private String _value;
        private Operator _op;

        DummyCriterion(String name, String value)
        {
            this(name, value, Operator.EQ);
        }
        
        DummyCriterion(String name, String value, Operator op)
        {
            _name = name;
            _value = value;
            _op = op;
        }
        
        Query toQuery()
        {
            return new StringQuery(_name, _op, _value, null);
        }
    }
}
