/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.search.advanced;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.log4j.Level;
import org.apache.log4j.LogManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.search.TestConstants;
import org.ametys.cms.search.QueryBuilder;
import org.ametys.cms.search.model.SearchModel;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.ui.model.SearchUIModelExtensionPoint;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

import com.google.common.collect.ImmutableMap;

/**
 * Tests advanced search on {@link QueryBuilder}
 */
@NeedFeatures({
    "cms/org.ametys.cms.content.search.components",
    "cms/workflow.content"
})
public class QueryBuilderTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static QueryBuilder _queryBuilder;
    private static SearchUIModelExtensionPoint _searchModelEP;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _queryBuilder = (QueryBuilder) manager.lookup(QueryBuilder.ROLE);
        _searchModelEP = (SearchUIModelExtensionPoint) manager.lookup(SearchUIModelExtensionPoint.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        LogManager.getLogger(QueryBuilder.class).setLevel(Level.DEBUG);
    }
    
    /**
     * Tests the part of the advanced query returned by {@link QueryBuilder#build(SearchModel, String, boolean, java.util.Map, java.util.Map) QueryBuilder#build}
     * @throws Exception if an error occurs.
     */
    @Test
    public void testGetAdvancedCriteriaQuery() throws Exception
    {
        /*
         * ___________[__OR__]_______
         * ___________/______\_______
         * _______[_AND_]__"title"___
         * ______/______\____________
         * !"creator"__"workflowStep"
         */
        
        SearchModel model = _searchModelEP.getExtension(TestConstants.TEST_SEARCH_MODEL);
        Map<String, Object> values = new HashMap<>();
        values.put("type", "or");
        List<Map<String, Object>> orExprs = new ArrayList<>();
        values.put("expressions", orExprs);
        Map<String, Object> andVals = new HashMap<>();
        Map<String, Object> sportVals = new HashMap<>();
        orExprs.add(andVals);
        orExprs.add(sportVals);
        andVals.put("type", "and");
        List<Map<String, Object>> andExprs = new ArrayList<>();
        Map<String, Object> creatorVals = new HashMap<>();
        Map<String, Object> wstepVals = new HashMap<>();
        andExprs.add(creatorVals);
        andExprs.add(wstepVals);
        creatorVals.put("type", "criterion");
        creatorVals.put("id", "property-creator-eq");
        Map<String, String> user = ImmutableMap.of("login", "user1", "populationId", "static");
        creatorVals.put("value", user);
        creatorVals.put("op", "ne");
        wstepVals.put("type", "criterion");
        wstepVals.put("id", "property-workflowStep-eq");
        wstepVals.put("value", "1");
        wstepVals.put("op", "eq");
        andVals.put("expressions", andExprs);
        sportVals.put("type", "criterion");
        sportVals.put("id", "metadata-title-eq");
        sportVals.put("value", "content");
        sportVals.put("op", "contains");
        
        Query query = _queryBuilder.build(model, "advanced", false, values, Collections.EMPTY_MAP);
        
        String notCreatorQuery = "*:* -creator:user1#static";
        String workflowStepQuery = "workflowStep:1";
        String andQuery = "(" + notCreatorQuery + ") AND (" + workflowStepQuery + ")";
        String titleQuery = "title_s_lower:(*content*)";
        String expectedAdvancedQuery = "(" + andQuery + ") OR (" + titleQuery + ")";
        
        String expectedBeforeQuery = "(";
        String expectedAfterQuery = ") AND (contentLanguage:en)";
        Assert.assertEquals(expectedBeforeQuery + expectedAdvancedQuery + expectedAfterQuery, query.build());
    }
}
