/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.search.advanced;

import java.util.Arrays;
import java.util.stream.Collectors;

import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.search.advanced.AbstractTreeNode;
import org.ametys.cms.search.advanced.TreeInternalNode;
import org.ametys.cms.search.advanced.TreeLeaf;
import org.ametys.cms.search.advanced.utils.TreePrinter;
import org.ametys.cms.search.query.Query.LogicalOperator;
import org.ametys.runtime.test.framework.AbstractBaseTestCase;

/**
 * Tests {@link AbstractTreeNode}, {@link TreeInternalNode} and {@link TreeLeaf}.
 */
public class TreeNodeTestCase extends AbstractBaseTestCase
{
    /**
     * Tests {@link AbstractTreeNode#walk(AbstractTreeNode, java.util.function.Function, java.util.function.BiFunction) walk} method on a String tree
     */
    @Test
    public void testStringTree()
    {
        /*
         * ____[__OR__]__
         * ______/___\___
         * __[_AND_]_"d"_
         * __/__|__\_____
         * "a"_"b"_"c"___
         */
        TreeLeaf<String> leafAstr = new TreeLeaf<>("a");
        TreeLeaf<String> leafBstr = new TreeLeaf<>("b");
        TreeLeaf<String> leafCstr = new TreeLeaf<>("c");
        TreeInternalNode<String> andstr = new TreeInternalNode<>(Arrays.asList(leafAstr, leafBstr, leafCstr), LogicalOperator.AND);
        
        TreeLeaf<String> leafDstr = new TreeLeaf<>("d");
        TreeInternalNode<String> treestr = new TreeInternalNode<>(Arrays.asList(andstr, leafDstr), LogicalOperator.OR);
        _logger.info(System.lineSeparator() + TreePrinter.print(treestr));
        
        String concatenatedNodeValues = AbstractTreeNode.walk(treestr, 
            TreeLeaf<String>::getValue, 
            (strings, operator) -> "(" + strings.collect(Collectors.joining(LogicalOperator.AND == operator ? " AND " : " OR ")) + ")");
        
        Assert.assertEquals("((a AND b AND c) OR d)", concatenatedNodeValues);
    }
    
    /**
     * Tests {@link AbstractTreeNode#walk(AbstractTreeNode, java.util.function.Function, java.util.function.BiFunction) walk} method on an Integer tree
     */
    @Test
    public void testIntTree()
    {
        /*
         * __[__AND__]___
         * __/___|____\__
         * 29__[_OR_]__4_
         * _____/__\_____
         * ____8____6____
         */
        TreeLeaf<Integer> leafEight = new TreeLeaf<>(8);
        TreeLeaf<Integer> leafSix = new TreeLeaf<>(6);
        TreeInternalNode<Integer> orint = new TreeInternalNode<>(Arrays.asList(leafEight, leafSix), LogicalOperator.OR);
        
        TreeLeaf<Integer> leafTwentyNine = new TreeLeaf<>(29);
        TreeLeaf<Integer> leafFour = new TreeLeaf<>(4);
        TreeInternalNode<Integer> treeint = new TreeInternalNode<>(Arrays.asList(leafTwentyNine, orint, leafFour), LogicalOperator.AND);
        _logger.info(System.lineSeparator() + TreePrinter.print(treeint));
        
        int result = AbstractTreeNode.walk(treeint, 
            // for instance, increment each value
            leaf -> leaf.getValue() + 1, 
            // whatever the operation, for instance if AND add them, if OR take min
            (integers, operator) -> LogicalOperator.AND == operator ? integers.reduce(0, Integer::sum) : integers.reduce(Integer.MAX_VALUE, Integer::min));
        
        Assert.assertEquals(42, result);
    }
}
