/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.search.columns;

import java.util.List;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.search.model.SystemPropertyExtensionPoint;
import org.ametys.cms.search.ui.model.ColumnHelper;
import org.ametys.cms.search.ui.model.ColumnHelper.Column;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Tests {@link ColumnHelper}.
 */
@NeedFeatures({
    "cms/search-model.column-helper",
    "cms/workflow.content"
})
public class ColumnHelperTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static ColumnHelper _columnHelper;
    private static SystemPropertyExtensionPoint _systemPropertyEP;
    private static int _nbDisplaySysprop;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _columnHelper = (ColumnHelper) manager.lookup(ColumnHelper.ROLE);
        _systemPropertyEP = (SystemPropertyExtensionPoint) manager.lookup(SystemPropertyExtensionPoint.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        WithCmsFeaturesTestCase.super.prepareComponents();
        _nbDisplaySysprop = _systemPropertyEP.getDisplayProperties().size();
    }
    
    /**
     * Tests that {@link ColumnHelper#getColumns(String, Set)} method extracts the expected labels.
     */
    @Test
    public void testColumnLabels()
    {
        Set<String> cType = Set.of();
        
        // Test ',' separator and whitespaces are trimmed
        _assertEquals(_columnHelper.getColumns("", cType));
        _assertEquals(_columnHelper.getColumns("title", cType), new Column("title", null));
        _assertEquals(_columnHelper.getColumns("title, foo", cType), new Column("title", null), new Column("foo", null));
        _assertEquals(_columnHelper.getColumns("title,foo", cType), new Column("title", null), new Column("foo", null));
        _assertEquals(_columnHelper.getColumns("foo,title", cType), new Column("foo", null), new Column("title", null));
        _assertEquals(_columnHelper.getColumns("title,,foo", cType), new Column("title", null), new Column("foo", null));
        _assertEquals(_columnHelper.getColumns("title,   ,foo", cType), new Column("title", null), new Column("foo", null));
        _assertEquals(_columnHelper.getColumns("title,foo,", cType), new Column("title", null), new Column("foo", null));
        _assertEquals(_columnHelper.getColumns(",title,foo", cType), new Column("title", null), new Column("foo", null));
        _assertEquals(_columnHelper.getColumns("   title   ,   foo    ", cType), new Column("title", null), new Column("foo", null));
        
        // Test AS keyword is accepted (case insensitively)
        _assertEquals(_columnHelper.getColumns("title AS eltit", cType), new Column("title", "eltit"));
        _assertEquals(_columnHelper.getColumns("title As eltit", cType), new Column("title", "eltit"));
        _assertEquals(_columnHelper.getColumns("title aS eltit", cType), new Column("title", "eltit"));
        _assertEquals(_columnHelper.getColumns("title as eltit", cType), new Column("title", "eltit"));
        _assertEquals(_columnHelper.getColumns("title AS    eltit  ", cType), new Column("title", "eltit"));
        _assertEquals(_columnHelper.getColumns("title AS", cType), new Column("title AS", null));
        _assertEquals(_columnHelper.getColumns("title AS ", cType), new Column("title", ""));
        _assertEquals(_columnHelper.getColumns("title AS    ", cType), new Column("title", ""));
        
        // Test AS in the column name does not conflict
        _assertEquals(_columnHelper.getColumns("MASK AS Disguise", cType), new Column("MASK", "Disguise"));
        _assertEquals(_columnHelper.getColumns("class as course", cType), new Column("class", "course"));
        _assertEquals(_columnHelper.getColumns("association AS organization", cType), new Column("association", "organization"));
        _assertEquals(_columnHelper.getColumns("gambas AS prawns", cType), new Column("gambas", "prawns"));
        _assertEquals(_columnHelper.getColumns("as as foo", cType), new Column("as", "foo"));
        _assertEquals(_columnHelper.getColumns(" as as foo", cType), new Column("as", "foo"));
        _assertEquals(_columnHelper.getColumns("as", cType), new Column("as", null));
        _assertEquals(_columnHelper.getColumns("as as ", cType), new Column("as", ""));
        _assertEquals(_columnHelper.getColumns("as as AS", cType), new Column("as", "AS"));
        
        // Test ',' separator + AS keyword
        _assertEquals(_columnHelper.getColumns("title AS eltit, foo AS oof", cType), new Column("title", "eltit"), new Column("foo", "oof"));
        _assertEquals(_columnHelper.getColumns("title AS eltit, foo as oof", cType), new Column("title", "eltit"), new Column("foo", "oof"));
        _assertEquals(_columnHelper.getColumns("title As eltit, foo aS oof", cType), new Column("title", "eltit"), new Column("foo", "oof"));
        _assertEquals(_columnHelper.getColumns("title, foo AS oof AS bar", cType), new Column("title", null), new Column("foo", "oof AS bar"));
        _assertEquals(_columnHelper.getColumns("title, foo AS oof as bar", cType), new Column("title", null), new Column("foo", "oof as bar"));
        _assertEquals(_columnHelper.getColumns("title, foo aS oof As bar", cType), new Column("title", null), new Column("foo", "oof As bar"));
        _assertEquals(_columnHelper.getColumns("title AS eltit, foo, title AS eltit2", cType), new Column("title", "eltit"), new Column("foo", null));
        
        // Test '.' -> '/' in column ids
        _assertEquals(_columnHelper.getColumns("title, foo.bar", cType), new Column("title", null), new Column("foo/bar", null));
        _assertEquals(_columnHelper.getColumns("title, foo.bar AS rab.foo", cType), new Column("title", null), new Column("foo/bar", "rab.foo"));
    }
    
    /**
     * Tests that {@link ColumnHelper#getColumns(String, Set)} method with repeater does not merge repeater columns
     */
    @Test
    @NeedFeatures({
        "exhaustif/content-type"
    })
    public void testColumnsInRepeaters()
    {
        Set<String> cType = Set.of("org.ametys.test.Content.exhaustive");
        List<Column> actualColumns = _columnHelper.getColumns("composite/repeater, composite/repeater/comp-date, composite/repeater/comp-long", cType);
        _assertEquals(actualColumns, new Column("composite/repeater", null),
                                     new Column("composite/repeater/comp-date", null),
                                     new Column("composite/repeater/comp-long", null));
    }
    
    /**
     * Tests that {@link ColumnHelper#getColumns(String, Set)} method with wildcard character at the end returns all first level columns.
     */
    @Test
    @NeedFeatures({
        "exhaustif/content-type",
        "exhaustif/reference-tables",
        "test/content-type",
    })
    public void testWildcard()
    {
        Set<String> cType = Set.of();
        List<Column> actualColumns = _columnHelper.getColumns("*", cType);
        Assert.assertTrue(actualColumns.contains(new Column("title", null)));
        Assert.assertFalse(actualColumns.contains(new Column("nickname", null)));
        Assert.assertFalse(actualColumns.contains(new Column("cute-nickname", null)));
        Assert.assertEquals(_nbDisplaySysprop + 1, actualColumns.size());
        
        cType = Set.of("simple.Nickname", "simple.Snake");
        actualColumns = _columnHelper.getColumns("*", cType);
        Assert.assertTrue(actualColumns.contains(new Column("title", null)));
        Assert.assertTrue(actualColumns.contains(new Column("nickname", null)));
        Assert.assertTrue(actualColumns.contains(new Column("cute-nickname", null)));
        Assert.assertTrue(actualColumns.contains(new Column("eyes", null)));
        Assert.assertTrue(actualColumns.contains(new Column("legs", null)));
        Assert.assertTrue(actualColumns.contains(new Column("length", null)));
        Assert.assertTrue(actualColumns.contains(new Column("illustration/image", null)));
        Assert.assertTrue(actualColumns.contains(new Column("illustration/alt-text", null)));
        Assert.assertEquals(_nbDisplaySysprop + 8, actualColumns.size());
        
        cType = Set.of("org.ametys.test.Content.exhaustive", "org.ametys.cms.contenttype.Test");
        actualColumns = _columnHelper.getColumns("address/city/*, repeaterThreeLevels/repeater2/*, field4/*", cType);
        Assert.assertTrue(actualColumns.contains(new Column("address/city/title", null)));
        Assert.assertTrue(actualColumns.contains(new Column("address/city/postCode", null)));
        Assert.assertTrue(actualColumns.contains(new Column("repeaterThreeLevels/repeater2/repeater3", null)));
        Assert.assertTrue(actualColumns.contains(new Column("field4/field41", null)));
        Assert.assertEquals(_nbDisplaySysprop + 4, actualColumns.size());
    }
    
    /**
     * Tests that {@link ColumnHelper#getColumns(String, Set)} method with view reference returns all view's items
     */
    @Test
    @NeedFeatures({
        "exhaustif/content-type"
    })
    public void testViewReference()
    {
        Set<String> cType = Set.of("org.ametys.test.Content.exhaustive");
        List<Column> actualColumns = _columnHelper.getColumns("owns/composite/repeater/comp-owns[abstract]", cType);
        _assertEquals(actualColumns, new Column("owns/composite/repeater/comp-owns/title", null),
                                     new Column("owns/composite/repeater/comp-owns/illustration/image", null),
                                     new Column("owns/composite/repeater/comp-owns/illustration/alt-text", null),
                                     new Column("owns/composite/repeater/comp-owns/abstract", null),
                                     new Column("owns/composite/repeater/comp-owns/attachments", null));
    }
    
    // need an override as Column#equals is overridden to only test id and not label
    private void _assertEquals(List<Column> actual, Column... expected)
    {
        Assert.assertEquals(expected.length, actual.size());
        
        for (int i = 0; i < expected.length; i++)
        {
            Column expectedColumn = expected[i];
            Column actualColumn = actual.get(i);
            Assert.assertEquals(expectedColumn.getId(), actualColumn.getId());
            Assert.assertEquals(expectedColumn.getLabel(), actualColumn.getLabel());
        }
    }
}
