/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.search.criterion.indexingfieldaggregator;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.LocalDate;
import java.time.LocalTime;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.function.Function;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.solr.client.solrj.util.ClientUtils;
import org.junit.Assert;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.contenttype.MetadataType;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableDefaultContent;
import org.ametys.cms.search.content.ContentSearcherFactory;
import org.ametys.cms.search.content.ContentSearcherFactory.SearchModelContentSearcher;
import org.ametys.cms.search.query.AbstractDateOperatorQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.ui.model.SearchUICriterion;
import org.ametys.cms.search.ui.model.SearchUIModel;
import org.ametys.cms.search.ui.model.SearchUIModelExtensionPoint;
import org.ametys.cms.search.ui.model.impl.IndexingFieldAggregatorSearchUICriterion;
import org.ametys.cms.search.ui.model.impl.IndexingFieldSearchUICriterion;
import org.ametys.cms.test.framework.AbstractSearcherTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareComposite;
import org.ametys.plugins.repository.metadata.MultilingualString;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

import com.google.common.collect.ImmutableMap;

import junit.framework.TestCase;

/**
 * Tests {@link IndexingFieldAggregatorSearchUICriterion}.
 * <br> By construction of it, it also tests {@link IndexingFieldSearchUICriterion}.
 * <br>There is only one {@link TestCase} class to avoid to start and shutdown Ametys for every type/success/error test
 */
// TODO add other tests defining different test-operators and testing their expected built query, etc.
// TODO add failure tests (when two types are different...)
@NeedFeatures({
    "cms/org.ametys.cms.content.search.content.searcher",
    
    "exhaustif/content-type",
    "cms/org.ametys.cms.data.holder.helpers"
})
public class IndexingFieldAggregatorSearchUICriterionTestCase extends AbstractSearcherTestCase implements WithCmsTestWebappFeatures
{
    private static final LocalDate __RIGHT_DATE;
    private static final LocalDate __WRONG_DATE;
    private static final ZonedDateTime __RIGHT_DATETIME;
    private static final ZonedDateTime __WRONG_DATETIME;
    private static final LocalDate __RIGHT_DATETIME_FULLDAY;
    private static final LocalDate __WRONG_DATETIME_FULLDAY;
    static
    {
        try
        {
            __RIGHT_DATE = LocalDate.of(2018, 12, 31);
            __WRONG_DATE = __RIGHT_DATE.plusDays(1);
            __RIGHT_DATETIME = ZonedDateTime.of(2018, 3, 6, 13, 37, 42, 0, ZoneId.systemDefault());
            __WRONG_DATETIME = __RIGHT_DATETIME.plusMonths(2);
            __RIGHT_DATETIME_FULLDAY = LocalDate.of(2018, 3, 6);
            __WRONG_DATETIME_FULLDAY = __RIGHT_DATETIME_FULLDAY.plusMonths(2);
        }
        catch (Exception e)
        {
            throw new RuntimeException(e);
        }
    }
    
    private static final Test __TEST_STRING = new Test("string", 
        MetadataType.STRING, 
        "strings", 
        null, 
        new HashMap<>(Collections.emptyMap()),
        "Title0",
        "(title_txt_en:(Title0)) OR (abstract_txt_en:(Title0))",
        "Title1",
        "(title_txt_en:(Title1)) OR (abstract_txt_en:(Title1))"
    );
    
    private static final Test __TEST_MULTILINGUAL_STRING = new Test("multilingual-string", 
        MetadataType.MULTILINGUAL_STRING, 
        "multilingual-strings", 
        "edition.textfield", // forced widget
        Collections.emptyMap(),
        "Title0",
        "(title_txt_en:(Title0)) OR (multilingual_txt_en:(Title0))",
        "fromage",
        "(title_txt_en:(fromage)) OR (multilingual_txt_en:(fromage))"
    );
    
    private static final Function<LocalDate, String> __DATE_TO_SOLRQ = date ->
    {
        String down = date.atStartOfDay().format(AbstractDateOperatorQuery.DATE_FORMATTER);
        String up = date.atTime(LocalTime.MAX).format(AbstractDateOperatorQuery.DATE_FORMATTER);
        return "(date_dt:[" + down + " TO " + up + "]) OR (datemult_dt:[" + down + " TO " + up + "])";
    };
    
    private static final Test __TEST_DATE = new Test("date", 
        MetadataType.DATE, 
        "dates", 
        null, 
        Collections.emptyMap(),
        DateUtils.localDateToString(__WRONG_DATE),
        __DATE_TO_SOLRQ.apply(__WRONG_DATE),
        DateUtils.localDateToString(__RIGHT_DATE),
        __DATE_TO_SOLRQ.apply(__RIGHT_DATE)
    );
    
    private static final Function<ZonedDateTime, String> __DATETIME_TO_SOLRQ = dateTime ->
    {
        String formatted = dateTime.withZoneSameInstant(ZoneOffset.UTC).format(AbstractDateOperatorQuery.DATE_FORMATTER);
        return "(start_dt:" + ClientUtils.escapeQueryChars(formatted) + ") OR (end_dt:" + ClientUtils.escapeQueryChars(formatted) + ")";
    };
    
    private static final Test __TEST_DATETIME = new Test("datetime", 
        MetadataType.DATETIME, 
        "datetimes", 
        null, 
        Collections.emptyMap(),
        DateUtils.zonedDateTimeToString(__WRONG_DATETIME),
        __DATETIME_TO_SOLRQ.apply(__WRONG_DATETIME),
        DateUtils.zonedDateTimeToString(__RIGHT_DATETIME),
        __DATETIME_TO_SOLRQ.apply(__RIGHT_DATETIME)
    );
    
    private static final Function<LocalDate, String> __DATETIME_FULLDAY_TO_SOLRQ = date ->
    {
        String down = date.atStartOfDay().atZone(ZoneId.systemDefault()).withZoneSameInstant(ZoneOffset.UTC).format(AbstractDateOperatorQuery.DATE_FORMATTER);
        String up = date.atTime(LocalTime.MAX).atZone(ZoneId.systemDefault()).withZoneSameInstant(ZoneOffset.UTC).format(AbstractDateOperatorQuery.DATE_FORMATTER);
        return "(start_dt:[" + down + " TO " + up + "]) OR (end_dt:[" + down + " TO " + up + "])";
    };
    
    private static final Test __TEST_DATETIME_FULLDAY = new Test("datetime", 
        MetadataType.DATETIME, 
        "datetimes", 
        null, 
        Collections.emptyMap(),
        DateUtils.localDateToString(__WRONG_DATETIME_FULLDAY),
        __DATETIME_FULLDAY_TO_SOLRQ.apply(__WRONG_DATETIME_FULLDAY),
        DateUtils.localDateToString(__RIGHT_DATETIME_FULLDAY),
        __DATETIME_FULLDAY_TO_SOLRQ.apply(__RIGHT_DATETIME_FULLDAY)
    );
    
    private static final Test __TEST_LONG = new Test("long", 
        MetadataType.LONG, 
        "longs", 
        null, 
        Collections.emptyMap(),
        0,
        "(composite/long-min_l:0) OR (composite/long-max_l:0)",
        42,
        "(composite/long-min_l:42) OR (composite/long-max_l:42)"
    );
    
//    private static final Test __TEST_GEOCODE = new Test("geocode", 
//        MetadataType.GEOCODE, 
//        "geocodes", 
//        null, 
//        Collections.emptyMap(),
//        new double[] {3.208, 44.481},
//        "({!geofilt sfield=gps_geo pt=3.208,44.481 d=5}) OR ({!geofilt sfield=address/gps_geo pt=3.208,44.481 d=5})",
//        new double[] {44.481, 3.208},
//        "({!geofilt sfield=gps_geo pt=44.481,3.208 d=5}) OR ({!geofilt sfield=address/gps_geo pt=44.481,3.208 d=5})"
//    );
//    
    private static final Test __TEST_DOUBLE = new Test("double", 
        MetadataType.DOUBLE, 
        "doubles", 
        null, 
        Collections.emptyMap(),
        0.5,
        "(composite/double-min_d:0.5) OR (composite/double-max_d:0.5)",
        3.14,
        "(composite/double-min_d:3.14) OR (composite/double-max_d:3.14)"
    );
    
    private static final Test __TEST_BOOLEAN = new Test("boolean", 
        MetadataType.BOOLEAN, 
        "booleans", 
        "edition.boolean-combobox", // forced widget 
        Collections.emptyMap(),
        false,
        "(*:* -boolean_b:true) OR (*:* -booleanmult_b:true)", // BooleanQuery is specific
        true,
        "(boolean_b:true) OR (booleanmult_b:true)",
        3, // content1, linked-content-1, linked-content-2
        1 // content1
    );
    
    private static final Test __TEST_USER = new Test("user", 
        MetadataType.USER, 
        "users", 
        null, 
        Collections.emptyMap(),
        _userToJson(AbstractPopulationAwareTestCase.USER2_TEST_IDENTITY),
        "(user_s:" + UserIdentity.userIdentityToString(AbstractPopulationAwareTestCase.USER2_TEST_IDENTITY) + ") OR (usermult_s:" + UserIdentity.userIdentityToString(AbstractPopulationAwareTestCase.USER2_TEST_IDENTITY) + ")",
        _userToJson(AbstractPopulationAwareTestCase.USER1_TEST_IDENTITY),
        "(user_s:" + UserIdentity.userIdentityToString(AbstractPopulationAwareTestCase.USER1_TEST_IDENTITY) + ") OR (usermult_s:" + UserIdentity.userIdentityToString(AbstractPopulationAwareTestCase.USER1_TEST_IDENTITY) + ")"
    );
    
    private static final Test __TEST_CONTENT = new Test("content", 
        MetadataType.CONTENT, 
        "contents", 
        null, 
        ImmutableMap.of(
                "allowSearch", new I18nizableText("false"), 
                "allowCreation", new I18nizableText("false"),
                "contentType", new I18nizableText("org.ametys.test.Content.exhaustive")
        ),
        "${linked-content-2}", // cannot have its id at this time (see #_replaceContentNamesByContentIds)
        "(owns_s:\"${linked-content-2}\") OR (belongsTo_s:\"${linked-content-2}\")",
        "${linked-content-1}",
        "(owns_s:\"${linked-content-1}\") OR (belongsTo_s:\"${linked-content-1}\")"
    );
    
    private static final Collection<Test> __TESTS = Stream.of(
            __TEST_STRING,
            __TEST_MULTILINGUAL_STRING,
            __TEST_DATE,
            __TEST_DATETIME,
            __TEST_LONG,
            // FIXME IndexingFieldSearchUICriterion#getQuery seems to not handle geocode metadata (+ GeocodeQuery class seems to not exist too)
            // see https://issues.ametys.org/browse/CMS-9074
            //__TEST_GEOCODE,
            __TEST_DOUBLE,
            __TEST_BOOLEAN,
            __TEST_USER,
            __TEST_CONTENT
    ).collect(Collectors.toList());
    
    private static final List<String> __TEST_NAMES = __TESTS.stream()
                                                            .map(t -> t._name)
                                                            .collect(Collectors.toList());
    
    private static final List<String> __SEARCH_MODEL_FILES = __TEST_NAMES.stream()
                                                                         .map(IndexingFieldAggregatorSearchUICriterionTestCase::_getSearchModelFileName)
                                                                         .collect(Collectors.toList());
    
    private static class Test
    {
        String _name;
        MetadataType _type;
        String _criterionName;
        String _widget;
        Map<String, I18nizableText> _widgetParameters;
        Object _firstResultValue;
        String _firstResultQuery;
        Object _secondResultValue;
        String _secondResultQuery;
        int _firstQueryExpectedCount;
        int _secondQueryExpectedCount;
        
        // First constructor when first value&query lead to a result count of 0 and second value&query lead to a result count of 1
        Test(String name, 
             MetadataType type, 
             String criterionName, 
             String widget, 
             Map<String, I18nizableText> widgetParameters,
             Object firstResultValue,
             String firstResultQuery,
             Object secondResultValue,
             String secondResultQuery)
        {
            this(name, type, criterionName, widget, widgetParameters, firstResultValue, firstResultQuery, secondResultValue, secondResultQuery, 0, 1);
        }
        
        // Second constructor when first and second query do no lead to a result count of 0 and 1
        Test(String name, 
             MetadataType type, 
             String criterionName, 
             String widget, 
             Map<String, I18nizableText> widgetParameters,
             Object firstResultValue,
             String firstResultQuery,
             Object secondResultValue,
             String secondResultQuery,
             int firstQueryExpectedCount, 
             int secondQueryExpectedCount)
        {
            _name = name;
            _type = type;
            _criterionName = criterionName;
            _widget = widget;
            _widgetParameters = widgetParameters;
            _firstResultValue = firstResultValue;
            _firstResultQuery = firstResultQuery;
            _secondResultValue = secondResultValue;
            _secondResultQuery = secondResultQuery;
            _firstQueryExpectedCount = firstQueryExpectedCount;
            _secondQueryExpectedCount = secondQueryExpectedCount;
        }
    }
    
    private static final String __PREFIX_SEARCH_MODELS = "test-aggregator-";
    
    private static SearchUIModelExtensionPoint _searchUIModelEP;
    private static ContentSearcherFactory _contentSearcherFactory;
    
    private static String _getSearchModel(String testName)
    {
        return "search-ui." + __PREFIX_SEARCH_MODELS + testName;
    }
    
    private static String _getSearchModelFileName(String searchModel)
    {
        return __PREFIX_SEARCH_MODELS + searchModel + ".xml";
    }
    
    private static Map<String, Object> _userToJson(UserIdentity user)
    {
        // We could have used org.ametys.plugins.core.user.UserHelper.user2json(UserIdentity)
        // but at the time when we need it, components are not available
        return ImmutableMap.of("login", user.getLogin(), "populationId", user.getPopulationId());
    }
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        _removeSearchModelsBelowParamFolder();
        _copySearchModelsBelowParamFolder();
        super.essentialsBefore();
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _searchUIModelEP = (SearchUIModelExtensionPoint) manager.lookup(SearchUIModelExtensionPoint.ROLE);
        _contentSearcherFactory = (ContentSearcherFactory) manager.lookup(ContentSearcherFactory.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        populateRepositoryAndIndex();
    }

    private static Path _filenameToParamFile(String filename)
    {
        return Paths.get("test", "environments", "webapp", "WEB-INF", "param", "search", filename);
    }
    
    private void _copySearchModelsBelowParamFolder() throws IOException
    {
        for (String searchModel : __SEARCH_MODEL_FILES)
        {
            _copySearchModelBelowParamFolder(searchModel);
        }
    }
    
    private void _copySearchModelBelowParamFolder(String filename) throws IOException
    {
        Path searchModelFile = Paths.get("test", "environments", "search-models", "indexingfieldaggregator", filename);
        Files.copy(searchModelFile, _filenameToParamFile(filename));
    }
    
    @Override
    protected void essentialsAfter() throws Exception
    {
        super.essentialsAfter();
        _removeSearchModelsBelowParamFolder();
    }
    
    private static void _removeSearchModelsBelowParamFolder() throws IOException
    {
        for (String searchModel : __SEARCH_MODEL_FILES)
        {
            _removeSearchModelBelowParamFolder(searchModel);
        }
    }
    
    private static void _removeSearchModelBelowParamFolder(String filename) throws IOException
    {
        Path searchModelFile = _filenameToParamFile(filename);
        Files.deleteIfExists(searchModelFile);
    }
    
    protected static void populateRepositoryAndIndex() throws Exception
    {
        // Do not call AbstractSearcherTestCase2018.populateRepositoryAndIndex(), implement your own
        
        ModifiableDefaultContent content1 = createContent("Title1", "content1", EXHAUSTIVE_CONTENTTYPE);
        
        content1.setValue("abstract", "abstract1");
        ModifiableDefaultContent link1 = createContent("Linked content 1", "linked-content-1", EXHAUSTIVE_CONTENTTYPE);
        createContent("Linked content 2", "linked-content-2", EXHAUSTIVE_CONTENTTYPE);
        
        content1.setValue("owns", new Content[] {link1});
        content1.setValue("date", __RIGHT_DATE);
        content1.setValue("end", __RIGHT_DATETIME);
        
        content1.setValue("gps", new Geocode(44.481, 3.208));
        content1.setValue("booleanmult", new Boolean[] {true});
        content1.setValue("usermult", new UserIdentity[] {AbstractPopulationAwareTestCase.USER1_TEST_IDENTITY});
        
        ModifiableModelAwareComposite comp1 = content1.getComposite("composite", true);
        comp1.setValue("long-max", 42L);
        comp1.setValue("double-max", 3.14D);
        
        content1.saveChanges();
        editContent(content1);
        
        ModifiableDefaultContent content2 = createContent("Title2", "content2", MULTILINGUAL_CONTENTTYPE);
        content2.setValue("multilingual", new MultilingualString(Collections.singletonMap(Locale.ENGLISH, "fromage")));
        
        content2.saveChanges();
        editContent(content2);
        
        Map<String, Object> indexResults = _solrIndexer.indexAllContents(RepositoryConstants.DEFAULT_WORKSPACE, false, _updateSolr);
        _solrIndexer.commit(RepositoryConstants.DEFAULT_WORKSPACE, _updateSolr);
        _staticLogger.info("Global indexing: success count={}, error count={}", indexResults.get("successCount"), indexResults.get("errorCount"));
    }
    
    /**
     * Tests {@link IndexingFieldAggregatorSearchUICriterion} for String case
     * @throws Exception If an error occurs
     */
    @org.junit.Test
    public void testStringSearchUICriteria() throws Exception
    {
        _test(__TEST_STRING);
    }
    
    /**
     * Tests {@link IndexingFieldAggregatorSearchUICriterion} for Multilingual String case
     * @throws Exception If an error occurs
     */
    @org.junit.Test
    public void testMultiplingualStringSearchUICriteria() throws Exception
    {
        _test(__TEST_MULTILINGUAL_STRING);
    }
    
    /**
     * Tests {@link IndexingFieldAggregatorSearchUICriterion} for Date case
     * @throws Exception If an error occurs
     */
    @org.junit.Test
    public void testDateSearchUICriteria() throws Exception
    {
        _test(__TEST_DATE);
    }
    
    /**
     * Tests {@link IndexingFieldAggregatorSearchUICriterion} for Datetime case
     * @throws Exception If an error occurs
     */
    @org.junit.Test
    public void testDatetimeSearchUICriteria() throws Exception
    {
        _test(__TEST_DATETIME);
    }
    
    /**
     * Tests {@link IndexingFieldAggregatorSearchUICriterion} for Datetime fullday case
     * @throws Exception If an error occurs
     */
    @org.junit.Test
    public void testDatetimeFullDaySearchUICriteria() throws Exception
    {
        _test(__TEST_DATETIME_FULLDAY);
    }
    
    /**
     * Tests {@link IndexingFieldAggregatorSearchUICriterion} for Long case
     * @throws Exception If an error occurs
     */
    @org.junit.Test
    public void testLongSearchUICriteria() throws Exception
    {
        _test(__TEST_LONG);
    }
    
    /**
     * Tests {@link IndexingFieldAggregatorSearchUICriterion} for Double case
     * @throws Exception If an error occurs
     */
    @org.junit.Test
    public void testDoubleSearchUICriteria() throws Exception
    {
        _test(__TEST_DOUBLE);
    }
    
    /**
     * Tests {@link IndexingFieldAggregatorSearchUICriterion} for Boolean case
     * @throws Exception If an error occurs
     */
    @org.junit.Test
    public void testBooleanSearchUICriteria() throws Exception
    {
        _test(__TEST_BOOLEAN);
    }
    
    /**
     * Tests {@link IndexingFieldAggregatorSearchUICriterion} for User case
     * @throws Exception If an error occurs
     */
    @org.junit.Test
    public void testUserSearchUICriteria() throws Exception
    {
        _test(__TEST_USER);
    }
    
    /**
     * Tests {@link IndexingFieldAggregatorSearchUICriterion} for Content case
     * @throws Exception If an error occurs
     */
    @org.junit.Test
    public void testContentSearchUICriteria() throws Exception
    {
        _test(__TEST_CONTENT);
    }
    
    private void _test(Test test) throws Exception
    {
        String searchModelId = IndexingFieldAggregatorSearchUICriterionTestCase._getSearchModel(test._name);
        SearchUIModel searchModel = _searchUIModelEP.getExtension(searchModelId);
        String criterionName = test._criterionName;
        String criterionId = IndexingFieldAggregatorSearchUICriterion.SEARCH_CRITERIA_METADATA_AGGREGATOR_PREFIX + criterionName;
        SearchUICriterion criterion = searchModel.getCriteria(null).get(criterionId);
        
        String failureMessage = "Test failed for criterion with name '" + criterionName + "': ";
        Assert.assertNotNull(failureMessage, criterion);
        Assert.assertTrue(failureMessage, criterion instanceof IndexingFieldAggregatorSearchUICriterion);
        Assert.assertEquals(failureMessage, criterionId, criterion.getId());
        Assert.assertEquals(failureMessage, criterionId, criterion.getFieldId());
        Assert.assertEquals(failureMessage, test._type, criterion.getType());
        Assert.assertEquals(failureMessage, test._widget, criterion.getWidget());
        Assert.assertEquals(failureMessage, test._widgetParameters, criterion.getWidgetParameters());
        
        SearchModelContentSearcher searcher = _contentSearcherFactory.create(searchModel).setCheckRights(false);
        Map<String, Object> contextualParameters = Collections.singletonMap("language", "en");
        Object firstResultValue = _replaceTestValue(test._firstResultValue);
        Map<String, Object> values = Collections.singletonMap(criterionId, firstResultValue);
        Query firstResultQuery = criterion.getQuery(firstResultValue, values, "en", Collections.EMPTY_MAP);
        Assert.assertEquals(failureMessage, _replaceExpectedQuery(test._firstResultQuery), firstResultQuery.build());
        Assert.assertEquals(failureMessage, test._firstQueryExpectedCount, searcher.searchWithFacets(values, contextualParameters).getTotalCount());
        
        Object secondResultValue = _replaceTestValue(test._secondResultValue);
        values = Collections.singletonMap(criterionId, secondResultValue);
        Query secondResultQuery = criterion.getQuery(secondResultValue, values, "en", Collections.EMPTY_MAP);
        Assert.assertEquals(failureMessage, _replaceExpectedQuery(test._secondResultQuery), secondResultQuery.build());
        Assert.assertEquals(failureMessage, test._secondQueryExpectedCount, searcher.searchWithFacets(values, contextualParameters).getTotalCount());
        
        Assert.assertEquals(failureMessage, false, criterion.isSortable());
        Assert.assertEquals(failureMessage, null, criterion.getOperator());
    }
    
    private Object _replaceTestValue(Object testValue)
    {
        return _replaceContentNamesByContentIds(testValue, false);
    }
    
    private Object _replaceExpectedQuery(Object expectedQuery)
    {
        return _replaceContentNamesByContentIds(expectedQuery, true);
    }
    
    // little "hack" method to handle ids of contents which are unknown at compile-time
    // content_s:"${content1}" will become content_s:"defaultContent\:\/\/1234\-5678..."
    private Object _replaceContentNamesByContentIds(Object obj, boolean escapeQueryChars)
    {
        if (obj instanceof String)
        {
            String str = (String) obj;
            Matcher m = Pattern.compile("\\$\\{([\\w-]*)\\}").matcher(str);
            StringBuffer sb = new StringBuffer();
            while (m.find())
            {
                String contentName = m.group(1);
                String contentId = _ametysObjectResolver.resolveByPath("/" + RepositoryConstants.NAMESPACE_PREFIX + ":contents/" + contentName).getId();
                m.appendReplacement(sb, escapeQueryChars ? Matcher.quoteReplacement(ClientUtils.escapeQueryChars(contentId))
                                                         : contentId);
            }
            m.appendTail(sb);
            return sb.toString();
        }
        return obj;
    }
}
