/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.search.model;

import java.util.HashMap;
import java.util.List;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.search.ui.model.SearchUIColumnHelper;
import org.ametys.cms.search.ui.model.SearchUIModel;
import org.ametys.cms.search.ui.model.SearchUIModelExtensionPoint;
import org.ametys.cms.search.ui.model.StaticSearchUIModel;
import org.ametys.cms.search.ui.model.impl.RepeaterSearchUIColumn;
import org.ametys.cms.search.ui.model.impl.ViewElementAccessorSearchUIColumn;
import org.ametys.cms.search.ui.model.impl.ViewElementSearchUIColumn;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.ViewElementAccessor;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;


/**
 * Test case for search UI models
 */
@NeedFeatures({
    "exhaustif/content-type"
})
public class StaticSearchUIModelTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static SearchUIModelExtensionPoint _searchUIModelExtensionPoint;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _searchUIModelExtensionPoint = (SearchUIModelExtensionPoint) manager.lookup(SearchUIModelExtensionPoint.ROLE);
    }
    
    /**
     * Test {@link StaticSearchUIModel} configuration with some items
     */
    @Test
    public void testStaticSearchUIModelColumnsWithSomeItems()
    {
        SearchUIModel model = _searchUIModelExtensionPoint.getExtension("search-ui.test-columns-some-items");
        
        List<ViewItem> viewItems = model.getResultItems(new HashMap<>()).getViewItems();
        Assert.assertEquals(7, viewItems.size());
        
        // title
        ViewElementSearchUIColumn titleColumn = (ViewElementSearchUIColumn) viewItems.get(0);
        Assert.assertEquals("title", titleColumn.getName());
        Assert.assertEquals("Ametys.plugins.cms.search.SearchGridHelper.renderTitle", titleColumn.getRenderer());
        Assert.assertEquals(SearchUIColumnHelper.DEFAULT_COLUMN_WIDTH, titleColumn.getWidth());
        Assert.assertFalse(titleColumn.isHidden());
        
        // long
        ViewElementSearchUIColumn longColumn = (ViewElementSearchUIColumn) viewItems.get(1);
        Assert.assertEquals("long", longColumn.getName());
        Assert.assertNull(longColumn.getRenderer());
        
        // illustration
        ModelViewItemGroup illustrationItem = (ModelViewItemGroup) viewItems.get(2);
        Assert.assertEquals("illustration", illustrationItem.getName());
        List<ViewItem> illustrationChildren = illustrationItem.getViewItems();
        Assert.assertEquals(1, illustrationChildren.size());
        
        // illustration - image
        ViewElementSearchUIColumn imageColumn = (ViewElementSearchUIColumn) illustrationChildren.get(0);
        Assert.assertEquals("image", imageColumn.getName());
        
        // address
        ModelViewItemGroup addressItem = (ModelViewItemGroup) viewItems.get(3);
        Assert.assertEquals("address", addressItem.getName());
        List<ViewItem> addressChildren = addressItem.getViewItems();
        Assert.assertEquals(1, addressChildren.size());
        
        // address - city
        ViewElementAccessor cityItem = (ViewElementAccessor) addressChildren.get(0);
        Assert.assertFalse(cityItem instanceof ViewElementAccessorSearchUIColumn);
        Assert.assertEquals("city", cityItem.getName());
        List<ViewItem> cityChildren = cityItem.getViewItems();
        Assert.assertEquals(1, cityChildren.size());
        
        // address - city - title
        ViewElementSearchUIColumn cityTitleColumn = (ViewElementSearchUIColumn) cityChildren.get(0);
        Assert.assertEquals("title", cityTitleColumn.getName());
        Assert.assertNull(cityTitleColumn.getRenderer());
        
        // contributor
        ViewElementSearchUIColumn contributorColumn = (ViewElementSearchUIColumn) viewItems.get(4);
        Assert.assertEquals("contributor", contributorColumn.getName());
        Assert.assertEquals("Ametys.plugins.cms.search.SearchGridHelper.renderUser", contributorColumn.getRenderer());
        Assert.assertEquals(SearchUIColumnHelper.DEFAULT_SYSTEM_PROPERTY_COLUMN_WIDTH, contributorColumn.getWidth());
        Assert.assertFalse(contributorColumn.isHidden());
        
        // creator
        ViewElementSearchUIColumn creatorColumn = (ViewElementSearchUIColumn) viewItems.get(5);
        Assert.assertEquals("creator", creatorColumn.getName());
        Assert.assertEquals("Ametys.plugins.cms.search.SearchGridHelper.renderUser", creatorColumn.getRenderer());
        Assert.assertEquals(100, creatorColumn.getWidth());
        
        // workflow step
        ViewElementSearchUIColumn workflowStepColumn = (ViewElementSearchUIColumn) viewItems.get(6);
        Assert.assertEquals("workflowStep", workflowStepColumn.getName());
        Assert.assertEquals("Ametys.plugins.cms.search.SearchGridHelper.renderWorkflowStep", workflowStepColumn.getRenderer());
        Assert.assertEquals("Ametys.plugins.cms.search.SearchGridHelper.convertWorkflowStep", workflowStepColumn.getConverter());
        Assert.assertEquals(60, workflowStepColumn.getWidth());
        Assert.assertTrue(workflowStepColumn.isHidden());
    }
    
    /**
     * Test {@link StaticSearchUIModel} configuration with columns for repeater
     */
    @Test
    public void testStaticSearchUIModelColumnsWithRepeaterColumns()
    {
        SearchUIModel model = _searchUIModelExtensionPoint.getExtension("search-ui.test-columns-repeater-columns");
        
        List<ViewItem> viewItems = model.getResultItems(new HashMap<>()).getViewItems();
        Assert.assertEquals(3, viewItems.size());
        
        // composite
        ModelViewItemGroup composite1Item = (ModelViewItemGroup) viewItems.get(0);
        Assert.assertEquals("composite", composite1Item.getName());
        List<ViewItem> composite1Children = composite1Item.getViewItems();
        Assert.assertEquals(1, composite1Children.size());
        
        // composite - repeater
        RepeaterSearchUIColumn repeater1Column = (RepeaterSearchUIColumn) composite1Children.get(0);
        Assert.assertEquals("repeater", repeater1Column.getName());
        List<ViewItem> repeater1Children = repeater1Column.getViewItems();
        Assert.assertEquals(1, repeater1Children.size());
        
        // composite - repeater - comp-string
        ViewElementSearchUIColumn compStringColumn = (ViewElementSearchUIColumn) repeater1Children.get(0);
        Assert.assertEquals("comp-string", compStringColumn.getName());
        
        // composite
        ModelViewItemGroup composite2Item = (ModelViewItemGroup) viewItems.get(1);
        Assert.assertEquals("composite", composite2Item.getName());
        List<ViewItem> composite2Children = composite2Item.getViewItems();
        Assert.assertEquals(1, composite2Children.size());
        
        // composite - repeater
        ModelViewItemGroup repeater2Column = (ModelViewItemGroup) composite2Children.get(0);
        Assert.assertFalse(repeater2Column instanceof RepeaterSearchUIColumn);
        Assert.assertEquals("repeater", repeater2Column.getName());
        List<ViewItem> repeater2Children = repeater2Column.getViewItems();
        Assert.assertEquals(1, repeater2Children.size());
        
        // composite - repeater - comp-long
        ViewElementSearchUIColumn compLongColumn = (ViewElementSearchUIColumn) repeater2Children.get(0);
        Assert.assertEquals("comp-long", compLongColumn.getName());
        
        // composite
        ModelViewItemGroup composite3Item = (ModelViewItemGroup) viewItems.get(2);
        Assert.assertEquals("composite", composite3Item.getName());
        List<ViewItem> composite3Children = composite3Item.getViewItems();
        Assert.assertEquals(1, composite3Children.size());
        
        // composite - repeater
        ModelViewItemGroup repeater3Column = (ModelViewItemGroup) composite3Children.get(0);
        Assert.assertFalse(repeater3Column instanceof RepeaterSearchUIColumn);
        Assert.assertEquals("repeater", repeater3Column.getName());
        List<ViewItem> repeater3Children = repeater3Column.getViewItems();
        Assert.assertEquals(1, repeater3Children.size());
        
        // composite - repeater - comp-date
        ViewElementSearchUIColumn compDateColumn = (ViewElementSearchUIColumn) repeater3Children.get(0);
        Assert.assertEquals("comp-date", compDateColumn.getName());
    }
    
    /**
     * Test {@link StaticSearchUIModel} configuration with view references
     */
    @Test
    public void testStaticSearchUIModelColumnsWithViewReferences()
    {
        SearchUIModel model = _searchUIModelExtensionPoint.getExtension("search-ui.test-columns-view-references");
        
        List<ViewItem> viewItems = model.getResultItems(new HashMap<>()).getViewItems();
        Assert.assertEquals(2, viewItems.size());
        
        // address
        ModelViewItemGroup address1Item = (ModelViewItemGroup) viewItems.get(0);
        Assert.assertEquals("address", address1Item.getName());
        List<ViewItem> address1Children = address1Item.getViewItems();
        Assert.assertEquals(1, address1Children.size());
        
        // address - city => is not a column
        ViewElementAccessor city1Item = (ViewElementAccessor) address1Children.get(0);
        Assert.assertFalse(city1Item instanceof ViewElementAccessorSearchUIColumn);
        Assert.assertEquals("city", city1Item.getName());
        List<ViewItem> city1Children = city1Item.getViewItems();
        Assert.assertEquals(2, city1Children.size());
        
        // address - city - title
        ViewElementSearchUIColumn cityTitle1Column = (ViewElementSearchUIColumn) city1Children.get(0);
        Assert.assertEquals("title", cityTitle1Column.getName());
        
        // address - city - postCode
        ViewElementSearchUIColumn cityPostCode1Column = (ViewElementSearchUIColumn) city1Children.get(1);
        Assert.assertEquals("postCode", cityPostCode1Column.getName());
        
        // address
        ModelViewItemGroup address2Item = (ModelViewItemGroup) viewItems.get(1);
        Assert.assertEquals("address", address2Item.getName());
        List<ViewItem> address2Children = address2Item.getViewItems();
        Assert.assertEquals(1, address2Children.size());
        
        // address - city => is a column
        ViewElementAccessorSearchUIColumn city2Item = (ViewElementAccessorSearchUIColumn) address2Children.get(0);
        Assert.assertEquals("city", city2Item.getName());
        List<ViewItem> city2Children = city2Item.getViewItems();
        Assert.assertEquals(2, city2Children.size());
        
        // address - city - title
        ViewElementSearchUIColumn cityTitle2Column = (ViewElementSearchUIColumn) city2Children.get(0);
        Assert.assertEquals("title", cityTitle2Column.getName());
        
        // address - city - postCode
        ViewElementSearchUIColumn cityPostCode2Column = (ViewElementSearchUIColumn) city2Children.get(1);
        Assert.assertEquals("postCode", cityPostCode2Column.getName());
    }
    
    /**
     * Test {@link StaticSearchUIModel} configuration with all items of some accessors
     */
    @Test
    public void testStaticSearchUIModelColumnsWithAllItemsInAccessors()
    {
        SearchUIModel model = _searchUIModelExtensionPoint.getExtension("search-ui.test-columns-all-items-in-accessors");
        
        List<ViewItem> viewItems = model.getResultItems(new HashMap<>()).getViewItems();
        Assert.assertEquals(3, viewItems.size());
        
        // illustration
        ModelViewItemGroup illustrationItem = (ModelViewItemGroup) viewItems.get(0);
        Assert.assertEquals("illustration", illustrationItem.getName());
        List<ViewItem> illustrationChildren = illustrationItem.getViewItems();
        Assert.assertEquals(2, illustrationChildren.size());
        
        // illustration - image
        ViewElementSearchUIColumn imageColumn = (ViewElementSearchUIColumn) illustrationChildren.get(0);
        Assert.assertEquals("image", imageColumn.getName());
        
        // illustration - altText
        ViewElementSearchUIColumn altTextColumn = (ViewElementSearchUIColumn) illustrationChildren.get(1);
        Assert.assertEquals("alt-text", altTextColumn.getName());
        
        // address
        ModelViewItemGroup addressItem = (ModelViewItemGroup) viewItems.get(1);
        Assert.assertEquals("address", addressItem.getName());
        List<ViewItem> addressChildren = addressItem.getViewItems();
        Assert.assertEquals(1, addressChildren.size());
        
        // address - city
        ViewElementAccessor cityItem = (ViewElementAccessor) addressChildren.get(0);
        Assert.assertFalse(cityItem instanceof ViewElementAccessorSearchUIColumn);
        Assert.assertEquals("city", cityItem.getName());
        List<ViewItem> cityChildren = cityItem.getViewItems();
        Assert.assertEquals(2, cityChildren.size());
        
        // address - city - title
        ViewElementSearchUIColumn cityTitleColumn = (ViewElementSearchUIColumn) cityChildren.get(0);
        Assert.assertEquals("title", cityTitleColumn.getName());
        Assert.assertNull(cityTitleColumn.getRenderer());
        
        // address - city - post code
        ViewElementSearchUIColumn cityPostCodeColumn = (ViewElementSearchUIColumn) cityChildren.get(1);
        Assert.assertEquals("postCode", cityPostCodeColumn.getName());
        
        // three levels repeater
        ModelViewItemGroup threeLevelsRepeaterItem = (ModelViewItemGroup) viewItems.get(2);
        Assert.assertFalse(threeLevelsRepeaterItem instanceof RepeaterSearchUIColumn);
        Assert.assertEquals("repeaterThreeLevels", threeLevelsRepeaterItem.getName());
        List<ViewItem> threeLevelsRepeaterChildren = threeLevelsRepeaterItem.getViewItems();
        Assert.assertEquals(1, threeLevelsRepeaterChildren.size());
        
        // three levels repeater - repeater2
        ModelViewItemGroup repeater2Item = (ModelViewItemGroup) threeLevelsRepeaterChildren.get(0);
        Assert.assertFalse(repeater2Item instanceof RepeaterSearchUIColumn);
        Assert.assertEquals("repeater2", repeater2Item.getName());
        List<ViewItem> repeater2Children = repeater2Item.getViewItems();
        Assert.assertEquals(1, repeater2Children.size());
        
        // three levels repeater - repeater2 - repeater3
        RepeaterSearchUIColumn repeater3Item = (RepeaterSearchUIColumn) repeater2Children.get(0);
        Assert.assertEquals("repeater3", repeater3Item.getName());
        List<ViewItem> repeater3Children = repeater3Item.getViewItems();
        Assert.assertEquals(6, repeater3Children.size());
        
        // three levels repeater - repeater2 - repeater3 - rep3 string
        ViewElementSearchUIColumn rep3StringColumn = (ViewElementSearchUIColumn) repeater3Children.get(0);
        Assert.assertEquals("rep3-string", rep3StringColumn.getName());
        
        // three levels repeater - repeater2 - repeater3 - rep3 long
        ViewElementSearchUIColumn rep3LongColumn = (ViewElementSearchUIColumn) repeater3Children.get(1);
        Assert.assertEquals("rep3-long", rep3LongColumn.getName());
        
        // three levels repeater - repeater2 - repeater3 - rep3 longmult-min
        ViewElementSearchUIColumn rep3LongmultMinColumn = (ViewElementSearchUIColumn) repeater3Children.get(2);
        Assert.assertEquals("rep3-longmult-min", rep3LongmultMinColumn.getName());
        
        // three levels repeater - repeater2 - repeater3 - rep3 longmult-max
        ViewElementSearchUIColumn rep3LongmultMaxColumn = (ViewElementSearchUIColumn) repeater3Children.get(3);
        Assert.assertEquals("rep3-longmult-max", rep3LongmultMaxColumn.getName());
        
        // three levels repeater - repeater2 - repeater3 - rep3 owns
        ViewElementAccessorSearchUIColumn rep3OwnsColumn = (ViewElementAccessorSearchUIColumn) repeater3Children.get(4);
        Assert.assertEquals("rep3-owns", rep3OwnsColumn.getName());
        
        // three levels repeater - repeater2 - repeater3 - rep3 belongs to
        ViewElementAccessorSearchUIColumn rep3BelongsToColumn = (ViewElementAccessorSearchUIColumn) repeater3Children.get(5);
        Assert.assertEquals("rep3-belongsTo", rep3BelongsToColumn.getName());
    }
    
    /**
     * Test {@link StaticSearchUIModel} configuration with all items of the selected content type
     */
    @Test
    public void testStaticSearchUIModelColumnsWithAllItems()
    {
        SearchUIModel model = _searchUIModelExtensionPoint.getExtension("search-ui.test-columns-all-items");
        
        List<ViewItem> viewItems = model.getResultItems(new HashMap<>()).getViewItems();
        Assert.assertEquals(2, viewItems.size());
        
        // address - city - title
        ViewElementSearchUIColumn titleColumn = (ViewElementSearchUIColumn) viewItems.get(0);
        Assert.assertEquals("title", titleColumn.getName());
        Assert.assertEquals("Ametys.plugins.cms.search.SearchGridHelper.renderTitle", titleColumn.getRenderer());
        
        // address - city - post code
        ViewElementSearchUIColumn postCodeColumn = (ViewElementSearchUIColumn) viewItems.get(1);
        Assert.assertEquals("postCode", postCodeColumn.getName());
        Assert.assertNull(postCodeColumn.getRenderer());
    }
}
