/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.search.query;

import org.junit.Assert;
import org.junit.Ignore;
import org.junit.Test;

import org.ametys.cms.search.query.AndQuery;
import org.ametys.cms.search.query.MatchAllQuery;
import org.ametys.cms.search.query.MatchNoneQuery;
import org.ametys.cms.search.query.OrQuery;
import org.ametys.cms.search.query.Query;
import org.ametys.cms.search.query.StringQuery;
import org.ametys.runtime.test.framework.AbstractBaseTestCase;

/**
 * Tests {@link AndQuery} and {@link OrQuery} and their optimization
 */
public class AndOrQueryTestCase extends AbstractBaseTestCase
{
    /**
     * Test simple {@link AndQuery}
     * @throws Exception if an exception occured
     */
    @Test
    public void testAndQuery() throws Exception
    {
        Query query1 = new StringQuery("foo", "bar");
        Query query2 = new StringQuery("baz", "qux");
        Query andQuery = new AndQuery(query1, query2);
        
        Assert.assertEquals("(foo_s:\"bar\") AND (baz_s:\"qux\")", andQuery.build());
    }
    
    /**
     * Test {@link AndQuery} with a {@link MatchAllQuery} operand (which should be ignored)
     * @throws Exception if an exception occured
     */
    @Test
    public void testAndQueryWithMatchAll() throws Exception
    {
        Query query1 = new StringQuery("foo", "bar");
        Query query2 = new MatchAllQuery();
        Query query3 = new StringQuery("baz", "qux");
        Query andQuery = new AndQuery(query1, query2, query3);
        
        Assert.assertEquals("(foo_s:\"bar\") AND (baz_s:\"qux\")", andQuery.build());
    }
    
    /**
     * Test {@link AndQuery} with a {@link MatchNoneQuery} operand (which should be the only one to stay)
     * @throws Exception if an exception occured
     */
    @Test
    public void testAndQueryWithMatchNone() throws Exception
    {
        Query query1 = new StringQuery("foo", "bar");
        Query query2 = new MatchNoneQuery();
        Query query3 = new StringQuery("baz", "qux");
        Query andQuery = new AndQuery(query1, query2, query3);
        
        Assert.assertEquals("*:* -(*:*)", andQuery.build());
    }
    
    /**
     * Test {@link AndQuery} with a {@link MatchNoneQuery} operand (which should be the only one to stay), 
     * within an {@link OrQuery}. That {@link AndQuery} should disappear.
     * @throws Exception if an exception occured
     */
    @Test
    @Ignore // FIXME https://issues.ametys.org/browse/CMS-9281
    public void testAndQueryWithMatchNoneWithinOrQuery() throws Exception
    {
        Query query1 = new StringQuery("foo", "bar");
        Query query2 = new AndQuery(new MatchNoneQuery(), new StringQuery("quux", "corge"));
        Query query3 = new StringQuery("baz", "qux");
        Query orQuery = new OrQuery(query1, query2, query3);
        
        Assert.assertEquals("(foo_s:\"bar\") OR (baz_s:\"qux\")", orQuery.build());
    }
    
    /**
     * Test simple {@link OrQuery}
     * @throws Exception if an exception occured
     */
    @Test
    public void testOrQuery() throws Exception
    {
        Query query1 = new StringQuery("foo", "bar");
        Query query2 = new StringQuery("baz", "qux");
        Query orQuery = new OrQuery(query1, query2);
        
        Assert.assertEquals("(foo_s:\"bar\") OR (baz_s:\"qux\")", orQuery.build());
    }
    
    /**
     * Test {@link OrQuery} with a {@link MatchAllQuery} operand (which should be the only one to stay)
     * @throws Exception if an exception occured
     */
    @Test
    public void testOrQueryWithMatchAll() throws Exception
    {
        Query query1 = new StringQuery("foo", "bar");
        Query query2 = new MatchAllQuery();
        Query query3 = new StringQuery("baz", "qux");
        Query orQuery = new OrQuery(query1, query2, query3);
        
        Assert.assertEquals("*:*", orQuery.build());
    }
    
    /**
     * Test {@link OrQuery} with a {@link MatchNoneQuery} operand (which should be ignored)
     * @throws Exception if an exception occured
     */
    @Test
    public void testOrQueryWithMatchNone() throws Exception
    {
        Query query1 = new StringQuery("foo", "bar");
        Query query2 = new MatchNoneQuery();
        Query query3 = new StringQuery("baz", "qux");
        Query orQuery = new OrQuery(query1, query2, query3);
        
        Assert.assertEquals("(foo_s:\"bar\") OR (baz_s:\"qux\")", orQuery.build());
    }
    
    /**
     * Test {@link OrQuery} with a {@link MatchAllQuery} operand (which should be the only one to stay), 
     * within an {@link AndQuery}. That {@link OrQuery} should disappear.
     * @throws Exception if an exception occured
     */
    @Test
    @Ignore // FIXME https://issues.ametys.org/browse/CMS-9281
    public void testOrQueryWithMatchAllWithinAndQuery() throws Exception
    {
        Query query1 = new StringQuery("foo", "bar");
        Query query2 = new OrQuery(new MatchAllQuery(), new StringQuery("quux", "corge"));
        Query query3 = new StringQuery("baz", "qux");
        Query andQuery = new AndQuery(query1, query2, query3);
        
        Assert.assertEquals("(foo_s:\"bar\") AND (baz_s:\"qux\")", andQuery.build());
    }
}
