/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.content.workflow;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.time.LocalDate;
import java.time.LocalTime;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.data.Binary;
import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.WorkflowAwareContent;
import org.ametys.cms.rights.Content2ContentTypeRightContextConvertor;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.cms.workflow.AbstractContentWorkflowComponent;
import org.ametys.cms.workflow.ContentWorkflowHelper;
import org.ametys.cms.workflow.EditContentFunction;
import org.ametys.cms.workflow.InvalidInputWorkflowException;
import org.ametys.core.right.Profile;
import org.ametys.core.right.ProfileAssignmentStorageExtensionPoint;
import org.ametys.core.right.RightProfilesDAO;
import org.ametys.core.upload.Upload;
import org.ametys.core.upload.UploadManager;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.holder.group.Repeater;
import org.ametys.plugins.repository.data.holder.values.SynchronizableRepeater;
import org.ametys.plugins.workflow.AbstractWorkflowComponent;
import org.ametys.runtime.model.View;
import org.ametys.runtime.parameter.Errors;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

import com.opensymphony.workflow.WorkflowException;

/**
 * {@link EditContentFunction} test case;
 */
@NeedFeatures({
    "core/model.helper",
    "core-impl/runtime.upload.manager.fs",
    "repository/repository.data.holder.helpers",
    "core/runtime.rights.profiles.dao",
    "cms/right.contextconvertor.contenttype",
    "cms/cms.rights.accesscontroller.content-type",
    "cms/cms.rights.storage.assignments.impl.jdbc",
    "cms/workflow",
    "cms/workflow.content",
    "cms/content-types.helper",
    
    "content.edit/test.rights.accesscontroller",
    "content.edit/content-types-edit-invert-relations-tests",
    "content.edit/content-types-edit-disable-conditions-tests"
})
public class EditContentTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static ContentWorkflowHelper _contentWorkflowHelper;
    private static AmetysObjectResolver _ametysObjectResolver;
    private static UploadManager _uploadManager;
    private static ProfileAssignmentStorageExtensionPoint _profileAssignmentStorageEP;
    private static RightProfilesDAO _profilesDAO;
    private static ContentTypesHelper _contentTypesHelper;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _contentWorkflowHelper = (ContentWorkflowHelper) manager.lookup(ContentWorkflowHelper.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _uploadManager = (UploadManager) manager.lookup(UploadManager.ROLE);
        _profileAssignmentStorageEP = (ProfileAssignmentStorageExtensionPoint) manager.lookup(ProfileAssignmentStorageExtensionPoint.ROLE);
        _profilesDAO = (RightProfilesDAO) manager.lookup(RightProfilesDAO.ROLE);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
    }
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        super.essentialsBefore();
        
        Profile profile = _profilesDAO.addProfile("profile-test");
        _profilesDAO.addRight(profile, "Workflow_Rights_Edition_Online");
        _profileAssignmentStorageEP.allowProfileToUser(CURRENT_USER_TEST_IDENTITY, profile.getId(), Content2ContentTypeRightContextConvertor.CONTENT_TYPE_RIGHT_CONTEXT_PREFIX);
    }
    
    /**
     * Test EditContentFunction with typed values.
     * @throws WorkflowException if an error occurs within the EditContentFunction.
     * @throws IOException if an error occurs manipulating binaries.
     */
    @Test
    public void testEditTypedValues() throws WorkflowException, IOException
    {
        Map<String, Object> createResults = _contentWorkflowHelper.createContent("content", 1, "linked", "linked", new String[] {"org.ametys.test.Content.exhaustive"}, null, "en");
        WorkflowAwareContent linkedContent = (WorkflowAwareContent) createResults.get(AbstractContentWorkflowComponent.CONTENT_KEY);
        
        createResults = _contentWorkflowHelper.createContent("content", 1, "content", "test", new String[] {"org.ametys.test.Content.exhaustive"}, null, "en");
        WorkflowAwareContent content = (WorkflowAwareContent) createResults.get(AbstractContentWorkflowComponent.CONTENT_KEY);

        Map<String, Object> values = new HashMap<>();
        values.put("title", "title");
        values.put("string", "test string 1");
        values.put("stringmult", new String[]{"string1", "string2"});
        values.put("long", 12);
        values.put("longmult", new long[]{4L, 8L});
        values.put("double", 12.34);
        values.put("doublemult", new double[]{1.23, 4.56});
        values.put("boolean", true);
        values.put("booleanmult", new boolean[]{true, false});
        values.put("date", LocalDate.of(2020, 7, 7));
        values.put("datemult", new LocalDate[]{LocalDate.of(2020, 7, 7), LocalDate.of(2000, 1, 1)});
        values.put("datetime", ZonedDateTime.of(LocalDate.of(2020, 7, 7), LocalTime.of(14, 56), ZoneOffset.UTC));
        values.put("datetimemult", new ZonedDateTime[]{ZonedDateTime.of(LocalDate.of(2020, 7, 7), LocalTime.of(14, 56), ZoneOffset.UTC), 
                                                       ZonedDateTime.of(LocalDate.of(2000, 1, 1), LocalTime.of(1, 25), ZoneOffset.UTC)});
        values.put("gps", new Geocode(43.66, 1.66));
        values.put("user", new UserIdentity("login", "population"));
        values.put("linkContent", linkedContent.getId());
        
        Binary binary = new Binary();
        binary.setEncoding("UTF-8");
        binary.setFilename("test.txt");
        binary.setMimeType("text/plain");
        binary.setInputStream(new ByteArrayInputStream("tést".getBytes(StandardCharsets.UTF_8)));
        values.put("binary", binary);
        
        values.put("illustration", Map.of("alt-text", "alt text"));
        values.put("repeaterThreeLevels", List.of(Map.of("rep-string", "entry 1",
                                                         "repeater2", List.of(Map.of("repeater3", List.of(Map.of("rep3-string", "entry 1.1.1"),
                                                                                                          Map.of("rep3-string", "entry 1.1.2"))),
                                                                              Map.of("repeater3", List.of(Map.of("rep3-string", "entry 1.2.1"))))),
                                                  Map.of("rep-string", "entry 2")));
        
        Map<String, Object> workflowResult = _contentWorkflowHelper.editContent(content, values, 2);
        Assert.assertEquals("ok", workflowResult.get("result"));
        
        _testContent(content, linkedContent);
        
        values = new HashMap<>();
        values.put("repeaterThreeLevels", SynchronizableRepeater.appendOrRemove(List.of(Map.of("repeater2", List.of(Map.of("repeater3", List.of(Map.of("rep3-string", "entry 3.1.1")))))), Set.of(2)));
        
        workflowResult = _contentWorkflowHelper.editContent(content, values, 2);
        Assert.assertEquals("ok", workflowResult.get("result"));
        
        Assert.assertEquals(2, ((Repeater) content.getValue("repeaterThreeLevels")).getSize());
        Assert.assertEquals("entry 3.1.1", content.getValue("repeaterThreeLevels[2]/repeater2[1]/repeater3[1]/rep3-string"));
        
        values = new HashMap<>();
        values.put("repeaterThreeLevels", SynchronizableRepeater.replace(List.of(Map.of("repeater2", List.of(Map.of("repeater3", List.of(Map.of("rep3-string", "entry 4.1.1")))))), List.of(2)));
        
        workflowResult = _contentWorkflowHelper.editContent(content, values, 2);
        Assert.assertEquals("ok", workflowResult.get("result"));
        
        Assert.assertEquals(2, ((Repeater) content.getValue("repeaterThreeLevels")).getSize());
        Assert.assertEquals("entry 4.1.1", content.getValue("repeaterThreeLevels[2]/repeater2[1]/repeater3[1]/rep3-string"));
        
        values = new HashMap<>();
        values.put("repeaterThreeLevels", SynchronizableRepeater.replace(List.of(Map.of("repeater2", 
                                                                                 SynchronizableRepeater.replace(List.of(Map.of("repeater3", 
                                                                                                                        SynchronizableRepeater.replace(List.of(Map.of("rep3-string", "entry 1.2.5")), 
                                                                                                                                                       null))), 
                                                                                                                List.of(2)))), 
                                                                         null));
      
        workflowResult = _contentWorkflowHelper.editContent(content, values, 2);
        Assert.assertEquals("ok", workflowResult.get("result"));
      
        Assert.assertEquals(2, ((Repeater) content.getValue("repeaterThreeLevels")).getSize());
        Assert.assertEquals("entry 4.1.1", content.getValue("repeaterThreeLevels[2]/repeater2[1]/repeater3[1]/rep3-string"));
    }
    
    /**
     * Test EditContentFunction with typed values.
     * @throws WorkflowException if an error occurs within the EditContentFunction.
     * @throws IOException if an error occurs manipulating binaries.
     */
    @Test
    public void testEditRawValues() throws WorkflowException, IOException
    {
        Map<String, Object> createResults = _contentWorkflowHelper.createContent("content", 1, "linked", "linked", new String[] {"org.ametys.test.Content.exhaustive"}, null, "en");
        WorkflowAwareContent linkedContent = (WorkflowAwareContent) createResults.get(AbstractContentWorkflowComponent.CONTENT_KEY);
        
        createResults = _contentWorkflowHelper.createContent("content", 1, "content", "test", new String[] {"org.ametys.test.Content.exhaustive"}, null, "en");
        WorkflowAwareContent content = (WorkflowAwareContent) createResults.get(AbstractContentWorkflowComponent.CONTENT_KEY);
        
        Upload upload = _uploadManager.storeUpload(_userProvider.getUser(), "test.txt", new ByteArrayInputStream("tést".getBytes(StandardCharsets.UTF_8)));

        Map<String, Object> rawValues = new HashMap<>();
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "title", "title");
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "string", "test string 1");
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "stringmult", List.of("string1", "string2"));
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "long", 12);
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "longmult", List.of(4, 8));
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "double", 12.34);
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "doublemult", List.of(1.23, 4.56));
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "boolean", true);
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "booleanmult", List.of(true, false));
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "date", "2020-07-07");
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "datemult", List.of("2020-07-07", "2000-01-01"));
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "datetime", "2020-07-07T14:56:00Z");
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "datetimemult", List.of("2020-07-07T14:56:00Z", "2000-01-01T01:25:00Z"));
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "gps", Map.of(Geocode.LONGITUDE_IDENTIFIER, 1.66, Geocode.LATITUDE_IDENTIFIER, 43.66));
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "user", Map.of("login", "login", "populationId", "population"));
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "linkContent", linkedContent.getId());
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "binary", Map.of("id", upload.getId()));
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "illustration/alt-text", "alt text");

        rawValues.put(EditContentFunction.INTERNAL_FORM_ELEMENTS_PREFIX + "repeaterThreeLevels/size", 2);
        rawValues.put(EditContentFunction.INTERNAL_FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[1]/previous-position", 0);
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[1]/rep-string", "entry 1");
        rawValues.put(EditContentFunction.INTERNAL_FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[1]/repeater2/size", 2);
        rawValues.put(EditContentFunction.INTERNAL_FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[1]/repeater2[1]/previous-position", 0);
        rawValues.put(EditContentFunction.INTERNAL_FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[1]/repeater2[1]/repeater3/size", 2);
        rawValues.put(EditContentFunction.INTERNAL_FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[1]/repeater2[1]/repeater3[1]/previous-position", 0);
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[1]/repeater2[1]/repeater3[1]/rep3-string", "entry 1.1.1");
        rawValues.put(EditContentFunction.INTERNAL_FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[1]/repeater2[1]/repeater3[2]/previous-position", 0);
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[1]/repeater2[1]/repeater3[2]/rep3-string", "entry 1.1.2");
        rawValues.put(EditContentFunction.INTERNAL_FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[1]/repeater2[2]/previous-position", 0);
        rawValues.put(EditContentFunction.INTERNAL_FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[1]/repeater2[2]/repeater3/size", 1);
        rawValues.put(EditContentFunction.INTERNAL_FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[1]/repeater2[2]/repeater3[1]/previous-position", 0);
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[1]/repeater2[2]/repeater3[1]/rep3-string", "entry 1.2.1");
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[1]/rep-string", "entry 1");
        rawValues.put(EditContentFunction.INTERNAL_FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[2]/previous-position", 0);
        rawValues.put(EditContentFunction.FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[2]/rep-string", "entry 2");
        rawValues.put(EditContentFunction.INTERNAL_FORM_ELEMENTS_PREFIX + "repeaterThreeLevels[2]/repeater2/size", 0);

        Map<String, Object> parameters = new HashMap<>();
        parameters.put(EditContentFunction.FORM_RAW_VALUES, rawValues);
        
        Map<String, Object> inputs = new HashMap<>();
        inputs.put(AbstractWorkflowComponent.CONTEXT_PARAMETERS_KEY, parameters);
        Map<String, Object> workflowResult = _contentWorkflowHelper.doAction(content, 2, inputs);
        Assert.assertEquals("ok", workflowResult.get("result"));
        
        Content result = _ametysObjectResolver.resolveById(content.getId());
        _testContent(result, linkedContent);
    }
    
    /**
     * Test EditContentFunction with invalid typed values.
     * @throws WorkflowException if an error occurs within the EditContentFunction.
     */
    @Test
    public void testEditInvalidTypedValues() throws WorkflowException
    {
        Map<String, Object> createResults = _contentWorkflowHelper.createContent("content", 1, "content", "test", new String[] {"org.ametys.test.Content.exhaustive"}, null, "en");
        WorkflowAwareContent content = (WorkflowAwareContent) createResults.get(AbstractContentWorkflowComponent.CONTENT_KEY);

        Map<String, Object> values = new HashMap<>();
        // string does not match validation regexp
        values.put("string", "test string");
        
        // start after end
        values.put("start", ZonedDateTime.of(LocalDate.of(2020, 7, 7), LocalTime.of(14, 56), ZoneOffset.UTC));
        values.put("end", ZonedDateTime.of(LocalDate.of(2019, 4, 15), LocalTime.of(17, 30), ZoneOffset.UTC));

        values.put("composite", Map.of("long-min", 4L,
                                       "long-max", 2L));

        // attachments with more than 5 entries
        values.put("attachments", List.of(Map.of("attachment-text", "text1"),
                                          Map.of("attachment-text", "text2"),
                                          Map.of("attachment-text", "text3"),
                                          Map.of("attachment-text", "text4"),
                                          Map.of("attachment-text", "text5"),
                                          Map.of("attachment-text", "text6")));
        
        values.put("repeaterThreeLevels", List.of(Map.of("rep-string", "entry 1",
                                                         "repeater2", List.of(Map.of("repeater3", List.of(Map.of("rep3-string", "entry 1.1.1"),
                                                                                                          Map.of("rep3-long", 2))),
                                                                              Map.of("repeater3", List.of(Map.of("rep3-string", "entry 1.2.1",
                                                                                                                 "rep3-longmult-min", new Long[] {1L, 2L},
                                                                                                                 "rep3-longmult-max", new Long[] {2L, 1L}))))),
                                                  Map.of("rep-string", "entry 2")));
        
        try
        {
            _contentWorkflowHelper.editContent(content, values, 2);
            Assert.fail("EditContentFuncion should have had validation errors");
        }
        catch (InvalidInputWorkflowException e)
        {
            Map<String, Errors> errors = e.getErrors().getAllErrors();
            Assert.assertEquals(4, errors.size());
            
            Errors globalErrors = errors.get(EditContentFunction.GLOBAL_ERROR_KEY);
            Assert.assertNotNull(globalErrors);
            // start > end and composite/long-min > composite/long-max and rep3-longmult is not valid
            Assert.assertEquals(3, globalErrors.getErrors().size());
            
            Errors stringErrors = errors.get("string");
            Assert.assertNotNull(stringErrors);
            Assert.assertEquals(1, stringErrors.getErrors().size()); // string does not match regexp
            
            Errors attachmentsErrors = errors.get("attachments");
            Assert.assertNotNull(attachmentsErrors);
            Assert.assertEquals(1, attachmentsErrors.getErrors().size()); // attachments has 6 entries
            
            Errors repeaterErrors = errors.get("repeaterThreeLevels[1]/repeater2[1]/repeater3[2]/rep3-string");
            Assert.assertNotNull(repeaterErrors);
            Assert.assertEquals(1, repeaterErrors.getErrors().size()); // repeater3.rep3-string is not present
        }
    }
    
    private void _testContent(Content result, Content linkedContent) throws IOException
    {
        Assert.assertEquals("title", result.getTitle());
        Assert.assertEquals("test string 1", result.getValue("string"));
        Assert.assertArrayEquals(new String[]{"string1", "string2"}, result.getValue("stringmult"));
        Assert.assertEquals(12, (long) result.getValue("long"));
        Assert.assertArrayEquals(new long[]{4, 8}, ArrayUtils.toPrimitive((Long[]) result.getValue("longmult")));
        Assert.assertEquals(12.34, result.getValue("double"), 0.01);
        Assert.assertArrayEquals(new double[]{1.23, 4.56}, ArrayUtils.toPrimitive((Double[]) result.getValue("doublemult")), 0.01);
        Assert.assertEquals(true, result.getValue("boolean"));
        Assert.assertArrayEquals(new boolean[]{true, false}, ArrayUtils.toPrimitive((Boolean[]) result.getValue("booleanmult")));
        Assert.assertEquals(LocalDate.of(2020, 7, 7), result.getValue("date"));
        Assert.assertArrayEquals(new LocalDate[]{LocalDate.of(2020, 7, 7), LocalDate.of(2000, 1, 1)}, result.getValue("datemult"));
        Assert.assertTrue(ZonedDateTime.of(LocalDate.of(2020, 7, 7), LocalTime.of(14, 56), ZoneOffset.UTC).isEqual(result.getValue("datetime")));
        Assert.assertTrue(ZonedDateTime.of(LocalDate.of(2020, 7, 7), LocalTime.of(14, 56), ZoneOffset.UTC).isEqual(((ZonedDateTime[]) result.getValue("datetimemult"))[0]));
        Assert.assertEquals(new Geocode(43.66, 1.66), result.getValue("gps"));
        Assert.assertEquals(new UserIdentity("login", "population"), result.getValue("user"));
        
        Content resultLinkedContent = ((ContentValue) result.getValue("linkContent")).getContent();
        Assert.assertEquals(linkedContent.getId(), resultLinkedContent.getId());
        
        Binary resultBinary = result.getValue("binary");
        String resultBinaryContent = IOUtils.toString(resultBinary.getInputStream(), resultBinary.getEncoding());
        Assert.assertEquals("tést", resultBinaryContent);
        
        Assert.assertEquals("alt text", result.getValue("illustration/alt-text"));
        Assert.assertNull(result.getValue("illustration/image"));
        
        Assert.assertEquals("entry 2", result.getValue("repeaterThreeLevels[2]/rep-string"));
        Assert.assertEquals("entry 1.2.1", result.getValue("repeaterThreeLevels[1]/repeater2[2]/repeater3[1]/rep3-string"));
        Assert.assertNull(result.getValue("repeaterThreeLevels[2]/repeater2[1]/repeater3[1]/rep3-string"));
    }
    
    /**
     * Test EditContentFunction with restricted invert relation.
     * Restrictions on destination should be taken into account
     * @throws WorkflowException if an error occurs within the EditContentFunction.
     */
    @Test
    public void testEditNotForcedInvertRelations() throws WorkflowException
    {
        Map<String, Object> createResults = _contentWorkflowHelper.createContent("content", 1, "destination", "destination", new String[] {"tests.edit.invert.relations.destination"}, null, "en");
        WorkflowAwareContent destination = (WorkflowAwareContent) createResults.get(AbstractContentWorkflowComponent.CONTENT_KEY);
        
        createResults = _contentWorkflowHelper.createContent("content", 1, "source", "source", new String[] {"tests.edit.invert.relations.source"}, null, "en");
        WorkflowAwareContent source = (WorkflowAwareContent) createResults.get(AbstractContentWorkflowComponent.CONTENT_KEY);

        Map<String, Object> values = new HashMap<>();
        values.put("title", "title");
        values.put("doNotForceInvert", destination);
        
        try
        {
            _contentWorkflowHelper.editContent(source, values, 2);
            Assert.fail("Current user has no right to modify the invert relation, that is not forced. Content edition should have fail");
        }
        catch (WorkflowException e)
        {
            Assert.assertFalse(source.hasValue("doNotForceInvert"));
            Assert.assertFalse(destination.hasValue("linkedSource1"));
        }
    }
    
    /**
     * Test EditContentFunction with restricted invert relation.
     * Restrictions on destination should not be taken into account, because of the forceInvert that is set to true
     * @throws WorkflowException if an error occurs within the EditContentFunction.
     */
    @Test
    public void testEditForcedInvertRelations() throws WorkflowException
    {
        Map<String, Object> createResults = _contentWorkflowHelper.createContent("content", 1, "destination2", "destination2", new String[] {"tests.edit.invert.relations.destination"}, null, "en");
        WorkflowAwareContent destination = (WorkflowAwareContent) createResults.get(AbstractContentWorkflowComponent.CONTENT_KEY);
        
        createResults = _contentWorkflowHelper.createContent("content", 1, "source2", "source2", new String[] {"tests.edit.invert.relations.source"}, null, "en");
        WorkflowAwareContent source = (WorkflowAwareContent) createResults.get(AbstractContentWorkflowComponent.CONTENT_KEY);

        Map<String, Object> values = new HashMap<>();
        values.put("title", "title");
        values.put("forceInvert", destination);
        
        Map<String, Object> workflowResult = _contentWorkflowHelper.editContent(source, values, 2);
        Assert.assertEquals("ok", workflowResult.get("result"));
        
        Assert.assertEquals(destination.getId(), ((ContentValue) source.getValue("forceInvert")).getContentId());
        Assert.assertEquals(source.getId(), ((ContentValue) destination.getValue("linkedSource2")).getContentId());
    }
    
    /**
     * Test EditContentFunction with disable conditions.
     * @throws WorkflowException if an error occurs within the EditContentFunction.
     */
    @Test
    public void testEditDisableConditions() throws WorkflowException
    {
        Map<String, Object> createResults = _contentWorkflowHelper.createContent("content", 1, "contentWithDisable", "Content with disable conditions", new String[] {"tests.edit.disable.conditions"}, null, "en");
        WorkflowAwareContent content = (WorkflowAwareContent) createResults.get(AbstractContentWorkflowComponent.CONTENT_KEY);

        Map<String, Object> values = new HashMap<>();
        values.put("title", "title");
        values.put("boolean", true);
        values.put("repeater", List.of(
                                    Map.of("boolean", true),
                                    Map.of()));
        
        try
        {
            View mainView = _contentTypesHelper.getView("main", content);
            _contentWorkflowHelper.editContent(content, values, 2, mainView);
            Assert.fail("EditContentFuncion should have had validation errors");
        }
        catch (InvalidInputWorkflowException e)
        {
            Map<String, Errors> errors = e.getErrors().getAllErrors();
            Assert.assertEquals(1, errors.size());
            
            Errors repeaterErrors = errors.get("repeater[2]/disabledStringSiblings");
            Assert.assertNotNull(repeaterErrors);
            Assert.assertEquals(1, repeaterErrors.getErrors().size()); // boolean on entry 2 is set to false, string is not disabled and is mandatory
        }
    }
}
