/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Ignore;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.repository.ModifiableWorkflowAwareContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.cms.workflow.AbstractContentWorkflowComponent;
import org.ametys.cms.workflow.ContentWorkflowHelper;
import org.ametys.plugins.repository.jcr.NameHelper;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Tests the {@link AttributeDefinition#getDefaultValue()} method
 * Tests that we can refer to i18n keys, config parameters or to a content via one of its attribute value
 */
@NeedFeatures ({
    "core/model.helper",
    "cms/workflow",
    "cms/workflow.content"
})
public class AttributeDefinitionGetDefaultValueTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static ContentTypeExtensionPoint _contentTypeEP;
    private static ContentWorkflowHelper _contentWorkflowHelper;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _contentTypeEP = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _contentWorkflowHelper = (ContentWorkflowHelper) manager.lookup(ContentWorkflowHelper.ROLE);
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a classic default value type
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-classic-default-value",
    })
    public void testGetClassicDefaultValue()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.classic.default.value");
        ElementDefinition<Geocode> definition = (ElementDefinition<Geocode>) contentType.getModelItem("geocode");
        Geocode defaultValue = definition.getDefaultValue();
        
        Assert.assertEquals(Double.valueOf(44.1833), defaultValue.getLatitude());
        Assert.assertEquals(Double.valueOf(1.8833), defaultValue.getLongitude());
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a classic default value in a repeater
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-classic-default-value-in-repeater",
    })
    public void testGetClassicDefaultValueInRepeater()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.classic.default.value.in.repeater");
        ElementDefinition<Geocode> definition = (ElementDefinition<Geocode>) contentType.getModelItem("repeater/geocode");
        Geocode defaultValue = definition.getDefaultValue();
        
        Assert.assertEquals(Double.valueOf(44.1833), defaultValue.getLatitude());
        Assert.assertEquals(Double.valueOf(1.8833), defaultValue.getLongitude());
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with classic default values type for a multiple data
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-classic-default-values",
    })
    public void testGetClassicDefaultValues()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.classic.default.values");
        ElementDefinition<Geocode> definition = (ElementDefinition<Geocode>) contentType.getModelItem("geocodemult");
        Geocode[] defaultValues = definition.getDefaultValue();
        
        Assert.assertEquals(2, defaultValues.length);
        Assert.assertEquals(Double.valueOf(44.1833), defaultValues[0].getLatitude());
        Assert.assertEquals(Double.valueOf(1.8833), defaultValues[0].getLongitude());
        Assert.assertEquals(Double.valueOf(43.5333), defaultValues[1].getLatitude());
        Assert.assertEquals(Double.valueOf(1.5333), defaultValues[1].getLongitude());
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with classic single default value type for a multiple data
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-classic-default-value-on-multiple",
    })
    public void testGetClassicSingleDefaultValueOnMultipleData()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.classic.default.value.multiple");
        ElementDefinition<Geocode> definition = (ElementDefinition<Geocode>) contentType.getModelItem("geocodemult");
        Geocode[] defaultValues = definition.getDefaultValue();
        
        Assert.assertEquals(1, defaultValues.length);
        Assert.assertEquals(Double.valueOf(44.1833), defaultValues[0].getLatitude());
        Assert.assertEquals(Double.valueOf(1.8833), defaultValues[0].getLongitude());
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a config default value type
     * Using a string config parameter
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-config-default-values",
    })
    public void testGetConfigDefaultValueWithString()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.config.default.values");
        ElementDefinition<String> definition = (ElementDefinition<String>) contentType.getModelItem("string");
        String defaultValue = definition.getDefaultValue();
        
        Assert.assertEquals("Valeur par défaut du test config", defaultValue);
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a config default value type
     * Using an empty string config parameter
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-config-default-values",
    })
    public void testGetConfigEmptyDefaultValueWithString()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.config.default.values");
        ElementDefinition<String> definition = (ElementDefinition<String>) contentType.getModelItem("string-default-emtpy");
        String defaultValue = definition.getDefaultValue();
        
        Assert.assertEquals(StringUtils.EMPTY, defaultValue);
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a config default value type
     * Using a multiple string config parameter
     * TODO RUNTIME-3898: Add this test case when multiple config params will be allowed
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-config-default-values",
    })
    @Ignore
    public void testGetConfigMultipleDefaultValueWithString()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.config.default.values");
        ElementDefinition<String> definition = (ElementDefinition<String>) contentType.getModelItem("string-default-multiple");
        String defaultValue = definition.getDefaultValue();
        
        Assert.assertEquals("Valeur par défaut 1", defaultValue);
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a config default value type
     * Using string config parameters for a multiple data
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-config-default-values",
    })
    public void testGetConfigDefaultValueWithMultipleString()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.config.default.values");
        ElementDefinition<String> definition = (ElementDefinition<String>) contentType.getModelItem("stringmult");
        String[] defaultValues = definition.getDefaultValue();
        
        Assert.assertEquals(2, defaultValues.length);
        Assert.assertEquals("Valeur par défaut du test config", defaultValues[0]);
        Assert.assertEquals("Autre valeur par défaut du test config", defaultValues[1]);
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a config default value type
     * Using an empty string config parameter for a multiple data
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-config-default-values",
    })
    public void testGetConfigEmptyDefaultValueWithMultipleString()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.config.default.values");
        ElementDefinition<String> definition = (ElementDefinition<String>) contentType.getModelItem("stringmult-default-emtpy");
        String[] defaultValues = definition.getDefaultValue();
        
        Assert.assertEquals(1, defaultValues.length);
        Assert.assertTrue(StringUtils.isEmpty(defaultValues[0]));
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a config default value type
     * Using a multiple string config parameter for a multiple data
     * TODO RUNTIME-3898: Add this test case when multiple config params will be allowed
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-config-default-values",
    })
    @Ignore
    public void testGetConfigMultipleDefaultValueWithMultipleString()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.config.default.values");
        ElementDefinition<String> definition = (ElementDefinition<String>) contentType.getModelItem("stringmult-default-multiple");
        String[] defaultValues = definition.getDefaultValue();
        
        Assert.assertEquals(3, defaultValues.length);
        Assert.assertEquals("Valeur par défaut 1", defaultValues[0]);
        Assert.assertEquals("Valeur par défaut 2", defaultValues[1]);
        Assert.assertEquals("Valeur par défaut 3", defaultValues[2]);
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a config default value type
     * Using a long config parameter
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-config-default-values",
    })
    public void testGetConfigDefaultValueWithLong()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.config.default.values");
        ElementDefinition<Long> definition = (ElementDefinition<Long>) contentType.getModelItem("long");
        Long defaultValue = definition.getDefaultValue();
        
        Assert.assertEquals(Long.valueOf(1985), defaultValue);
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a content default value type
     * The used attribute is a String
     * @throws Exception if an error occurs while creating contents
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-content-default-value-with-string",
    })
    public void testGetContentDefaultValueOnStringAttribute() throws Exception
    {
        // Create contents, one of them with attribute string = default
        ModifiableContent correspondingContent = _createContent("Corresponding content for string", "tests.content.default.value.string");
        correspondingContent.setValue("string", "default");
        correspondingContent.saveChanges();
        ModifiableContent otherContent = _createContent("Other content", "tests.content.default.value.string");
        otherContent.setValue("string", "other value");
        otherContent.saveChanges();
        
        ContentType contentType = _contentTypeEP.getExtension("tests.content.default.value.string");
        ElementDefinition<ContentValue> definition = (ElementDefinition<ContentValue>) contentType.getModelItem("contentDefaultString");
        ContentValue defaultValue = definition.getDefaultValue();
        
        Assert.assertEquals(correspondingContent.getId(), defaultValue.getContentId());
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a content default value type
     * The attribute with the default value is in a repeater
     * The used attribute is a String
     * @throws Exception if an error occurs while creating contents
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-content-default-value-with-string-in-repeater",
    })
    public void testGetContentDefaultValueOnStringAttributeInRepeater() throws Exception
    {
        // Create contents, one of them with attribute string = default
        ModifiableContent correspondingContent = _createContent("Corresponding content for string", "tests.content.default.value.string.in.repeater");
        correspondingContent.setValue("string", "default");
        correspondingContent.saveChanges();
        ModifiableContent otherContent = _createContent("Other content", "tests.content.default.value.string");
        otherContent.setValue("string", "other value");
        otherContent.saveChanges();
        
        ContentType contentType = _contentTypeEP.getExtension("tests.content.default.value.string.in.repeater");
        ElementDefinition<ContentValue> definition = (ElementDefinition<ContentValue>) contentType.getModelItem("repeater/contentDefaultString");
        ContentValue defaultValue = definition.getDefaultValue();
        
        Assert.assertEquals(correspondingContent.getId(), defaultValue.getContentId());
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a content default value type
     * There is no corresponding content, the default value is <code>null</code>
     * @throws Exception if an error occurs while creating contents
     */
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-content-default-value-with-string-no-corresponding-content",
    })
    public void testGetContentDefaultValueOnStringAttributeWithNoCorrespondingContent() throws Exception
    {
        // Create contents, none of them with attribute string = default
        ModifiableContent content = _createContent("Content", "tests.content.default.value.no.corresponding.content");
        content.setValue("string", "Value");
        content.saveChanges();
        
        ContentType contentType = _contentTypeEP.getExtension("tests.content.default.value.no.corresponding.content");
        ElementDefinition definition = (ElementDefinition) contentType.getModelItem("contentDefaultString");
        
        Assert.assertNull(definition.getDefaultValue());
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a content default value type
     * The attribute is multiple, there is no corresponding content, the default value is an empty array
     * @throws Exception if an error occurs while creating contents
     */
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-content-default-value-with-string-no-corresponding-content",
    })
    public void testGetContentMultipleDefaultValueOnStringAttributeWithNoCorrespondingContent() throws Exception
    {
        // Create contents, none of them with attribute string = default
        ModifiableContent content = _createContent("Content", "tests.content.default.value.no.corresponding.content");
        content.setValue("string", "Value");
        content.saveChanges();
        
        ContentType contentType = _contentTypeEP.getExtension("tests.content.default.value.no.corresponding.content");
        ElementDefinition definition = (ElementDefinition) contentType.getModelItem("contentMultipleDefaultString");
        
        ContentValue[] defaultValue = (ContentValue[]) definition.getDefaultValue();
        Assert.assertEquals(0, defaultValue.length);
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a content default value type
     * The used attribute is a Long
     * @throws Exception if an error occurs while creating contents
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-content-default-value-with-long",
    })
    public void testGetContentDefaultValueOnLongAttribute() throws Exception
    {
        // Create contents, one of them with attribute long = 154
        ModifiableContent correspondingContent = _createContent("Corresponding content for string", "tests.content.default.value.long");
        correspondingContent.setValue("long", 154);
        correspondingContent.saveChanges();
        ModifiableContent otherContent = _createContent("Other content", "tests.content.default.value.long");
        otherContent.setValue("long", 96471);
        otherContent.saveChanges();
        
        ContentType contentType = _contentTypeEP.getExtension("tests.content.default.value.long");
        ElementDefinition<ContentValue> definition = (ElementDefinition<ContentValue>) contentType.getModelItem("contentDefaultLong");
        ContentValue defaultValue = definition.getDefaultValue();
        
        Assert.assertEquals(correspondingContent.getId(), defaultValue.getContentId());
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with a content default value type
     * The used attribute is a String
     * @throws Exception if an error occurs while creating contents
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-content-default-value-inheritence",
    })
    public void testGetContentDefaultValueWithInheritence() throws Exception
    {
        // Create contents, one of them with attribute string = default
        ModifiableContent correspondingContent = _createContent("Corresponding content for string", "tests.content.default.value.inheritence.enum");
        correspondingContent.setValue("string", "default");
        correspondingContent.saveChanges();
        ModifiableContent otherContent = _createContent("Other content", "tests.content.default.value.inheritence.enum");
        otherContent.setValue("string", "other value");
        otherContent.saveChanges();
        
        ContentType parentContentType = _contentTypeEP.getExtension("tests.content.default.value.inheritence.parent");
        ElementDefinition<ContentValue> parentDefinition = (ElementDefinition<ContentValue>) parentContentType.getModelItem("contentDefaultString");
        ContentValue parentDefaultValue = parentDefinition.getDefaultValue();
        Assert.assertEquals(correspondingContent.getId(), parentDefaultValue.getContentId());
        
        ContentType child1ContentType = _contentTypeEP.getExtension("tests.content.default.value.inheritence.child1");
        ElementDefinition<ContentValue> child1Definition = (ElementDefinition<ContentValue>) child1ContentType.getModelItem("contentDefaultString");
        ContentValue child1DefaultValue = child1Definition.getDefaultValue();
        Assert.assertEquals(correspondingContent.getId(), child1DefaultValue.getContentId());
        
        ContentType child2ContentType = _contentTypeEP.getExtension("tests.content.default.value.inheritence.child2");
        ElementDefinition<ContentValue> child2Definition = (ElementDefinition<ContentValue>) child2ContentType.getModelItem("contentDefaultString");
        ContentValue child2DefaultValue = child2Definition.getDefaultValue();
        Assert.assertEquals(correspondingContent.getId(), child2DefaultValue.getContentId());
    }
    
    /**
     * Test {@link AttributeDefinition#getDefaultValue()} method with default values of different types for a multiple data
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test-default-value/content-type-with-default-values",
    })
    public void testGetDefaultValues()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.default.values");
        ElementDefinition<String> definition = (ElementDefinition<String>) contentType.getModelItem("stringmult");
        String[] defaultValues = definition.getDefaultValue();
        
        Assert.assertEquals(2, defaultValues.length);
        Assert.assertEquals("Valeur par défaut classique", defaultValues[0]);
        Assert.assertEquals("Valeur par défaut du test config", defaultValues[1]);
    }
    
    private ModifiableContent _createContent(String title, String contentTypeId) throws Exception
    {
        String contentName = NameHelper.filterName(title);
        Map<String, Object> result = _contentWorkflowHelper.createContent("content", 1, contentName, title, new String[] {contentTypeId}, null, "fr");
        ModifiableContent content = (ModifiableWorkflowAwareContent) result.get(AbstractContentWorkflowComponent.CONTENT_KEY);
        
        content.saveChanges();
        return content;
    }
}
