/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.Optional;

import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.PluginInitializationFailureTestCase;
import org.ametys.runtime.test.framework.Unoptimized;

/**
 * {@link AttributeDefinition} with invalid default value configuration test case.
 */
@Unoptimized // The nature of the test is to fail when some features are not present, features being different in every @Test => @Unoptimized is thus NECESSARY
public class AttributeDefinitionWithInvalidDefaultValueCondifTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures, PluginInitializationFailureTestCase
{
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        // Nothing to lookup here
    }
    
    /**
     * Test {@link AttributeDefinition} with a nonexistent default value type
     */
    @Test
    @NeedFeatures("test-default-value/invalid-content-type-with-nonexistent-default-value-type")
    public void testNonexistentDefaultValueType()
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A default value can't be of an nonexistent type", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("nonexistent-type"));
    }
    
    /**
     * Test {@link AttributeDefinition} with a config default value where the type of the attribute does not correspond to the config parameter type
     */
    @Test
    @NeedFeatures("test-default-value/invalid-content-type-with-config-default-value-not-corresponding-type")
    public void testInvalidConfigDefaultValueNotCorrespondingType()
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("The type of the configuration parameter has to correspond to the type of the attribute using it as default value", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("config"));
        Assert.assertTrue(configurationException.getMessage().contains("string"));
        Assert.assertTrue(configurationException.getMessage().contains("geocode"));
    }
    
    /**
     * Test {@link AttributeDefinition} with a config default value with a nonexistent config parameter
     */
    @Test
    @NeedFeatures("test-default-value/invalid-content-type-with-config-default-value-non-existent-config-parameter")
    public void testInvalidConfigDefaultValueWithNonexistentConfigParam()
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("The used configuration parameter does not exist", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("config"));
        Assert.assertTrue(configurationException.getMessage().contains("nonexistent"));
    }
    
    /**
     * Test {@link AttributeDefinition} with a content default value on a non content attribute
     */
    @Test
    @NeedFeatures("test-default-value/invalid-content-type-with-content-default-value-on-non-content-attribute")
    public void testContentDefaultValueOnNonContentAttribute()
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A content default value has to be on a content attribute", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("attribute"));
        Assert.assertTrue(configurationException.getMessage().contains("geocode"));
    }
    
    /**
     * Test {@link AttributeDefinition} with a content default value on a content attribute without content type id
     */
    @Test
    @NeedFeatures("test-default-value/invalid-content-type-with-content-default-value-on-content-attribute-without-content-type-id")
    public void testContentDefaultValueWithoutContentTypeId()
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A content default value has to be on a content attribute that specifies the content type id", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("attribute"));
        Assert.assertTrue(configurationException.getMessage().contains("content type"));
    }
    
    /**
     * Test {@link AttributeDefinition} with a content default value on a content attribute without attribute name
     */
    @Test
    @NeedFeatures("test-default-value/invalid-content-type-with-content-default-value-without-attribute-name")
    public void testContentDefaultValueWithoutAttributeName()
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A content default value has to specify an attribute name", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("attribute"));
        Assert.assertTrue(configurationException.getMessage().contains("name"));
    }
    
    /**
     * Test {@link AttributeDefinition} with a content default value on a content attribute with attribute in a composite
     */
    @Test
    @NeedFeatures("test-default-value/invalid-content-type-with-content-default-value-with-attribute-in-composite")
    public void testContentDefaultValueWithAttributeInAComposite()
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A content default value can't specify an attribute in a composite or a repeater", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("attribute"));
        Assert.assertTrue(configurationException.getMessage().contains("composite"));
        Assert.assertTrue(configurationException.getMessage().contains("repeater"));
    }
    
    /**
     * Test {@link AttributeDefinition} with a content default value on a content attribute with attribute of type date
     */
    @Test
    @NeedFeatures("test-default-value/invalid-content-type-with-content-default-value-with-unsupported-attribute")
    public void testContentDefaultValueWithDateAttribute()
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A content default value can't specify an attribute of type date, only string, long and doubles are supported", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("attribute"));
        Assert.assertTrue(configurationException.getMessage().contains("date"));
    }
    
    /**
     * Test {@link AttributeDefinition} with a content default value on a content attribute with nonexistent attribute
     */
    @Test
    @NeedFeatures("test-default-value/invalid-content-type-with-content-default-value-with-nonexistent-attribute")
    public void testContentDefaultValueWithNonexistentAttribute()
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A content default value has to specify an exising attribute", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("attribute"));
        Assert.assertTrue(configurationException.getMessage().contains("nonexistent"));
    }
}
