/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.List;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.SimpleViewItemGroup;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ContentType} overriding view test case.
 */
@NeedFeatures ({
    "cms/workflow.content"
})
public class ContentTypeOverrideViewTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    /** Content type extension point */
    protected static ContentTypeExtensionPoint _contentTypeEP;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _contentTypeEP = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }
    
    /**
     * Tests {@link DefaultContentType} implementation overriding a view but keeping its default general information
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-keep-default-general-information",
    })
    public void testOverrideViewKeepDefaultGeneralInformation()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.keep.default.general.information");
        View view = contentType.getView("overriddenView");
        
        Assert.assertFalse(view.isInternal());
        Assert.assertEquals("overriddenView", view.getLabel().toString());
        Assert.assertEquals(StringUtils.EMPTY, view.getDescription().toString());
        Assert.assertNull(view.getSmallIcon());
        Assert.assertNull(view.getMediumIcon());
        Assert.assertNull(view.getLargeIcon());
        Assert.assertEquals("ametysicon-column3", view.getIconGlyph());
        Assert.assertNull(view.getIconDecorator());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation overriding a view by changing its default general information
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-change-default-general-information",
    })
    public void testOverrideViewChangeDefaultGeneralInformation()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.change.default.general.information");
        View view = contentType.getView("overriddenView");
        
        Assert.assertTrue(view.isInternal());
        Assert.assertEquals("LabelChanged", view.getLabel().toString());
        Assert.assertEquals("DescChanged", view.getDescription().toString());
        Assert.assertEquals("/plugins/contenttype.override.view/resources/img/icon2-small.png", view.getSmallIcon());
        Assert.assertEquals("/plugins/contenttype.override.view/resources/img/icon2-medium.png", view.getMediumIcon());
        Assert.assertEquals("/plugins/contenttype.override.view/resources/img/icon2-large.png", view.getLargeIcon());
        Assert.assertEquals("ametysicon-newspaper1", view.getIconGlyph());
        Assert.assertEquals("decorator-ametysicon-test2", view.getIconDecorator());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation overriding a view by keeping its general information
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-keep-specific-general-information",
    })
    public void testOverrideViewKeepSpecificGeneralInformation()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.keep.specific.general.information");
        View view = contentType.getView("overriddenView");
        
        Assert.assertTrue(view.isInternal());
        Assert.assertEquals("InitLabel", view.getLabel().toString());
        Assert.assertEquals("InitDesc", view.getDescription().toString());
        Assert.assertEquals("/plugins/contenttype.override.view/resources/img/icon-small.png", view.getSmallIcon());
        Assert.assertEquals("/plugins/contenttype.override.view/resources/img/icon-medium.png", view.getMediumIcon());
        Assert.assertEquals("/plugins/contenttype.override.view/resources/img/icon-large.png", view.getLargeIcon());
        Assert.assertEquals("ametysicon-text70", view.getIconGlyph());
        Assert.assertEquals("decorator-ametysicon-test", view.getIconDecorator());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation overriding a view by changing its general information
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-change-specific-general-information",
    })
    public void testOverrideViewChangeSpecificGeneralInformation()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.change.specific.general.information");
        View view = contentType.getView("overriddenView");
        
        Assert.assertFalse(view.isInternal());
        Assert.assertEquals("LabelChanged", view.getLabel().toString());
        Assert.assertEquals("DescChanged", view.getDescription().toString());
        Assert.assertEquals("/plugins/contenttype.override.view/resources/img/icon2-small.png", view.getSmallIcon());
        Assert.assertEquals("/plugins/contenttype.override.view/resources/img/icon2-medium.png", view.getMediumIcon());
        Assert.assertEquals("/plugins/contenttype.override.view/resources/img/icon2-large.png", view.getLargeIcon());
        Assert.assertEquals("ametysicon-newspaper1", view.getIconGlyph());
        Assert.assertEquals("decorator-ametysicon-test2", view.getIconDecorator());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation overriding a view by keeping its default glyph, positioned as the medium icon
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-keep-default-glyph-as-medium-icon",
    })
    public void testOverrideViewKeepDefaultGlyphAsMediumIcon()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.keep.default.glyph.as.medium.icon");
        View view = contentType.getView("overriddenView");
        
        Assert.assertEquals("/plugins/contenttype.override.view/resources/img/icon-medium.png", view.getMediumIcon());
        Assert.assertEquals("/plugins/contenttype.override.view/resources/img/icon-medium.png", view.getIconGlyph());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with adding an attribute to the view
     * The overridden view is like:
     * - title
     * - attributeToAdd (added attribute)
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-add-attribute",
    })
    public void testOverrideViewsAddAttribute()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.add.attribute");
        View view = contentType.getView("overriddenView");
        
        List<ViewItem> viewItems = view.getViewItems();
        
        // Check the attribute has been added at the end of the view
        Assert.assertEquals(2, viewItems.size());
        Assert.assertEquals("attributeToAdd", viewItems.get(1).getName());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with adding an attribute in a group to the view
     * The overridden view is like:
     * - title
     * - GroupTab1
     *   - GroupFieldset1
     *     - GroupFieldset2
     *         - attribute
     *         - attributeToAdd (added attribute)
     * - GroupTab2
     *   - title
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-add-attribute-in-group",
    })
    public void testOverrideViewsAddAttributeInGroup()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.add.attribute.in.group");
        View view = contentType.getView("overriddenView");
        
        List<ViewItem> viewItems = view.getViewItems();
        ViewItemAccessor groupTab1 = (ViewItemAccessor) viewItems.get(1);
        ViewItemAccessor groupFieldset1 = (ViewItemAccessor) groupTab1.getViewItems().get(0);
        ViewItemAccessor groupFieldset2 = (ViewItemAccessor) groupFieldset1.getViewItems().get(0);
        List<ViewItem> groupFieldset2Items = groupFieldset2.getViewItems();
        
        // Check the attribute has been added at the end of the GroupFieldset2
        Assert.assertEquals(2, groupFieldset2Items.size());
        Assert.assertEquals("attributeToAdd", groupFieldset2Items.get(1).getName());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with adding attributes, before and after some view items to the view
     * The overridden view is like:
     * - attributeToAdd1 (added attribute)
     * - title
     * - attributeToAdd2 (added attribute)
     * - attributeMiddle
     * - attributeToAdd3 (added attribute)
     * - attributeToAdd4 (added attribute)
     * - composite
     *   - attribute
     * - repeater
     *   - attribute
     * - attributeToAdd5 (added attribute)
     * - attributeToAdd6 (added attribute)
     * - GroupTab
     *   - title
     * - attributeToAdd7 (added attribute)
     * - attributeEnd
     * - attributeToAdd8 (added attribute)
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-add-attributes-before-after",
    })
    public void testOverrideViewsAddAttributesBeforeAfter()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.add.attributes.before.after");
        View view = contentType.getView("overriddenView");
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(14, viewItems.size());
        
        // Check all attributes have been added at their right place
        Assert.assertEquals("attributeToAdd1", viewItems.get(0).getName());
        Assert.assertEquals("title", viewItems.get(1).getName());
        
        Assert.assertEquals("attributeToAdd2", viewItems.get(2).getName());
        Assert.assertEquals("attributeMiddle", viewItems.get(3).getName());
        Assert.assertEquals("attributeToAdd3", viewItems.get(4).getName());

        Assert.assertEquals("attributeToAdd4", viewItems.get(5).getName());
        Assert.assertEquals("composite", viewItems.get(6).getName());
        Assert.assertEquals("repeater", viewItems.get(7).getName());
        Assert.assertEquals("attributeToAdd5", viewItems.get(8).getName());
        
        Assert.assertEquals("attributeToAdd6", viewItems.get(9).getName());
        Assert.assertEquals("GroupTab", viewItems.get(10).getName());
        Assert.assertEquals("attributeToAdd7", viewItems.get(11).getName());
        
        Assert.assertEquals("attributeEnd", viewItems.get(12).getName());
        Assert.assertEquals("attributeToAdd8", viewItems.get(13).getName());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with adding attributes, before and after view items, in a group to the view
     * The overridden view is like:
     * - title
     * - GroupTab
     *   - GroupFieldset1
     *     - GroupFieldset2
     *       - attributeToAdd2 (added attribute)
     *       - composite
     *         - attribute
     *     - attributeToAdd1 (added attribute)
     *     - attribute
     *   - attributeToAdd3 (added attribute)
     *   - GroupFieldset3
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-add-attributes-before-after-in-groups",
    })
    public void testOverrideViewsAddAttributesBeforeAfterInGroups()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.add.attributes.before.after.in.groups");
        View view = contentType.getView("overriddenView");
        
        List<ViewItem> viewItems = view.getViewItems();
        ViewItemAccessor groupTab = (ViewItemAccessor) viewItems.get(1);
        List<ViewItem> groupTabItems = groupTab.getViewItems();
        ViewItemAccessor groupFieldset1 = (ViewItemAccessor) groupTabItems.get(0);
        List<ViewItem> groupFieldset1Items = groupFieldset1.getViewItems();
        ViewItemAccessor groupFieldset2 = (ViewItemAccessor) groupFieldset1Items.get(0);
        List<ViewItem> groupFieldset2Items = groupFieldset2.getViewItems();
        
        // Check the attributeToAdd1 has been added before the attribute1, in the GroupFieldset1
        Assert.assertEquals(3, groupFieldset1Items.size());
        Assert.assertEquals("attributeToAdd1", groupFieldset1Items.get(1).getName());
        Assert.assertEquals("attribute", groupFieldset1Items.get(2).getName());
        
        // Check the attributeToAdd2 has been added before the composite, in the GroupFieldset2
        Assert.assertEquals(2, groupFieldset2Items.size());
        Assert.assertEquals("attributeToAdd2", groupFieldset2Items.get(0).getName());
        Assert.assertEquals("composite", groupFieldset2Items.get(1).getName());
        
        // Check the attributeToAdd3 has been added after the GroupFieldset1, in the GroupYab1
        Assert.assertEquals(3, groupTabItems.size());
        Assert.assertEquals("GroupFieldset1", groupTabItems.get(0).getName());
        Assert.assertEquals("attributeToAdd3", groupTabItems.get(1).getName());
    }
    
    /**
     * Tests {@link ContentType} implementation with adding a composite to the view
     * The overridden view is like:
     * - title
     * - compositeToAdd (added composite)
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-add-composite",
    })
    public void testOverrideViewsAddComposite()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.add.composite");
        View view = contentType.getView("overriddenView");
        
        List<ViewItem> viewItems = view.getViewItems();
        
        // Check the composite has been added at the end of the view
        Assert.assertEquals(2, viewItems.size());
        Assert.assertEquals("compositeToAdd", viewItems.get(1).getName());
    }
    
    /**
     * Tests {@link ContentType} implementation with adding a repeater before an attribute in a group to the view
     * The overridden view is like:
     * - title
     * - GroupTab
     *   - GroupFieldset1
     *     - GroupFieldset2
     *     - repeaterToAdd (added repeater)
     *     - attribute
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-add-repeater-before-after-in-group",
    })
    public void testOverrideViewsAddRepeaterBeforeAfterInGroup()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.add.repeater.before.after.in.group");
        View view = contentType.getView("overriddenView");
        
        List<ViewItem> viewItems = view.getViewItems();
        ViewItemAccessor groupTab = (ViewItemAccessor) viewItems.get(1);
        ViewItemAccessor groupFieldset1 = (ViewItemAccessor) groupTab.getViewItems().get(0);
        List<ViewItem> groupFieldset1Items = groupFieldset1.getViewItems();
        
        // Check the repeater has been added before the attribute, in the GroupFieldset1
        Assert.assertEquals(3, groupFieldset1Items.size());
        Assert.assertEquals("repeaterToAdd", groupFieldset1Items.get(1).getName());
        Assert.assertEquals("attribute", groupFieldset1Items.get(2).getName());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with adding a group to the view
     * The overridden view is like:
     * - title
     * - groupToAdd (added group)
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-add-group",
    })
    public void testOverrideViewsAddGroup()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.add.group");
        View view = contentType.getView("overriddenView");

        List<ViewItem> viewItems = view.getViewItems();

        // Check the group has been added at the end of the view
        Assert.assertEquals(2, viewItems.size());
        Assert.assertEquals("GroupToAdd", viewItems.get(1).getName());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with adding a group before an attribute in a group to the view
     * The overridden view is like:
     * - title
     * - GroupTab
     *   - GroupFieldset1
     *     - GroupFieldset2
     *       - groupToAdd
     *       - composite
     *         - attribute
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-add-group-before-after-in-group",
    })
    public void testOverrideViewsAddGroupBeforeAfterInGroup()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.add.group.before.after.in.group");
        View view = contentType.getView("overriddenView");
        
        List<ViewItem> viewItems = view.getViewItems();
        ViewItemAccessor groupTab = (ViewItemAccessor) viewItems.get(1);
        ViewItemAccessor groupFieldset1 = (ViewItemAccessor) groupTab.getViewItems().get(0);
        ViewItemAccessor groupFieldset2 = (ViewItemAccessor) groupFieldset1.getViewItems().get(0);
        List<ViewItem> groupFieldset2Items = groupFieldset2.getViewItems();
        
        // Check the repeater has been added before the composite, in the GroupFieldset2
        Assert.assertEquals(2, groupFieldset2Items.size());
        Assert.assertEquals("GroupToAdd", groupFieldset2Items.get(0).getName());
        Assert.assertEquals("composite", groupFieldset2Items.get(1).getName());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with removing items from the overridden view
     * Here is what the view looks like before all removals (removed items are marked):
     * - title
     * - GroupTab1
     *   - GroupFieldset1
     *     - GroupFieldset2
     *       - composite
     *         - GroupFieldset3
     *           - attribute (removed)
     *       - repeater (removed)
     *         - attribute
     *     - attribute1 (removed)
     *     - attribute2
     *   - GroupFieldset4 (removed)
     * - attributeMiddle (removed)
     * - composite (removed)
     *   - attribute
     * - repeater
     *   - attribute (removed)
     * - GroupTab2 (removed)
     *   - title
     * - attributeEnd
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-remove-items",
    })
    public void testOverrideViewRemoveItems()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.remove.items");
        View view = contentType.getView("overriddenView");
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(4, viewItems.size());
        
        // title
        Assert.assertEquals("title", viewItems.get(0).getName());

        // GroupTab1
        SimpleViewItemGroup groupTab1 = (SimpleViewItemGroup) viewItems.get(1);
        Assert.assertEquals("GroupTab1", groupTab1.getName());
        List<ViewItem> groupTab1Items = groupTab1.getViewItems();
        Assert.assertEquals(1, groupTab1Items.size());

        // GroupTab1 - GroupFieldset1
        SimpleViewItemGroup groupFieldset1 = (SimpleViewItemGroup) groupTab1Items.get(0);
        Assert.assertEquals("GroupFieldset1", groupFieldset1.getName());
        List<ViewItem> groupFieldset1Items = groupFieldset1.getViewItems();
        Assert.assertEquals(2, groupFieldset1Items.size());
        
        // GroupTab1 - GroupFieldset1 - GroupFieldset2
        SimpleViewItemGroup groupFieldset2 = (SimpleViewItemGroup) groupFieldset1Items.get(0);
        Assert.assertEquals("GroupFieldset2", groupFieldset2.getName());
        List<ViewItem> groupFieldset2Items = groupFieldset2.getViewItems();
        Assert.assertEquals(1, groupFieldset2Items.size());
        
        // GroupTab1 - GroupFieldset1 - GroupFieldset2 - composite
        ModelViewItemGroup composite = (ModelViewItemGroup) groupFieldset2Items.get(0);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeItems = composite.getViewItems();
        Assert.assertEquals(1, compositeItems.size());
        
        // GroupTab1 - GroupFieldset1 - GroupFieldset2 - composite - GroupFieldset3
        SimpleViewItemGroup groupFieldset3 = (SimpleViewItemGroup) compositeItems.get(0);
        Assert.assertEquals("GroupFieldset3", groupFieldset3.getName());
        Assert.assertTrue(groupFieldset3.getViewItems().isEmpty());
        
        // GroupTab1 - GroupFieldset1 - attribute2
        Assert.assertEquals("attribute2", groupFieldset1Items.get(1).getName());

        // repeater
        ModelViewItemGroup repeater = (ModelViewItemGroup) viewItems.get(2);
        Assert.assertEquals("repeater", repeater.getName());
        Assert.assertTrue(repeater.getViewItems().isEmpty());
        
        // attributeEnd
        Assert.assertEquals("attributeEnd", viewItems.get(3).getName());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with removing an item in a group that have the same name of an attribute
     * Here is what the view looks like before removal (removed items are marked):
     * - sameName (attribute)
     * - sameName (group)
     *   - title (removed)
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-remove-item-in-group-with-same-name-as-an-attribute",
    })
    public void testOverrideViewRemoveItemInGroupWithSameName()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.remove.item.in.group.same.name");
        View view = contentType.getView("overriddenView");
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(2, viewItems.size());
        
        // sameName attribute
        Assert.assertEquals("sameName", viewItems.get(0).getName());

        // sameName group
        SimpleViewItemGroup group = (SimpleViewItemGroup) viewItems.get(1);
        Assert.assertEquals("sameName", group.getName());
        Assert.assertTrue(group.getViewItems().isEmpty());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with removing an item in _override folder when this item has been added in a group by override in an extension
     * Here is what the view looks like in initial content type:
     * - GroupTab
     *   - title
     * Here is what the view looks like after override in child content type:
     * - GroupTab
     *   - attribute (and this one is removed in extension) 
     *   - title
     * - attribute (this item is added only to prove that the override worked) 
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-remove-item-added-in-group",
    })
    public void testOverrideViewRemoveItemAddedInGroup()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.remove.item.added.in.group.child");
        View view = contentType.getView("overriddenView");
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(2, viewItems.size());

        // GroupTab
        SimpleViewItemGroup groupTab = (SimpleViewItemGroup) viewItems.get(0);
        Assert.assertEquals("GroupTab", groupTab.getName());
        List<ViewItem> groupTabItems = groupTab.getViewItems();
        Assert.assertEquals(1, groupTabItems.size());
        
        // GroupTab - title
        Assert.assertEquals("title", groupTabItems.get(0).getName());
        
        // attribute
        Assert.assertEquals("attribute", viewItems.get(1).getName());
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type overriding a view by adding an attribute to a nonexistent group
     * The attribute will be added to the end of the view
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-add-to-nonexistent-group"
    })
    public void testConfigureContentTypeOverrideViewAddItemToNonexistentGroup()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.add.to.nonexistent.group");
        View view = contentType.getView("overriddenView");
        
        List<ViewItem> viewItems = view.getViewItems();
        
        // Check the attribute has been added at the end of the view
        Assert.assertEquals(2, viewItems.size());
        Assert.assertEquals("attributeToAdd", viewItems.get(1).getName());
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type overriding by adding an attribute before/after a nonexistent item
     * The attribute will be added to the end of the view
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-add-after-nonexistent-item"
        })
    public void testConfigureContentTypeOverrideViewAddItemAfterNonexistentItem()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.add.after.nonexistent.item");
        View view = contentType.getView("overriddenView");
        
        List<ViewItem> viewItems = view.getViewItems();
        
        // Check the attribute has been added at the end of the view
        Assert.assertEquals(2, viewItems.size());
        Assert.assertEquals("attributeToAdd", viewItems.get(1).getName());
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type overriding by adding an attribute before/after an item that is not present in the specified group
     * The attribute will be added to the end of the group
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-add-before-nonexistent-item-in-group",
        })
    public void testConfigureContentTypeOverrideViewAddAttributeBeforeNonexistentAttInGroup()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.add.before.nonexistent.item.in.group");
        View view = contentType.getView("overriddenView");
        
        List<ViewItem> viewItems = view.getViewItems();
        ViewItemAccessor groupTab = (ViewItemAccessor) viewItems.get(2);
        List<ViewItem> groupTabItems = groupTab.getViewItems();
        
        // Check the attribute has been added at the end of the group
        Assert.assertEquals(2, groupTabItems.size());
        Assert.assertEquals("attributeToAdd", groupTabItems.get(1).getName());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with overriding and redefine inherited view of children
     * Some views are redefine in the child or its overrides (extension or file).
     * This test check the precedence of the chosen view. Ex: if the view is redefined by an extension, the overrides in the content type declaration are obsoletes
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-precedence"
    })
    public void testOverrideViewsPrecedence()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.precedence.child");
        
        // overriddenView1 is redefined in the content type declaration. The overrides of the parent is ignored
        View view1 = contentType.getView("overriddenView1");
        List<ViewItem> view1Items = view1.getViewItems();
        Assert.assertEquals(3, view1Items.size());
        Assert.assertEquals("attributeAddedByChild", view1Items.get(0).getName());
        Assert.assertEquals("attributeAddedByChildOverride", view1Items.get(1).getName());
        Assert.assertEquals("attributeAddedByChildFileOverride", view1Items.get(2).getName());
        
        // overriddenView2 is redefined in an extension. The overrides of the parent and the content type declaration is ignored
        View view2 = contentType.getView("overriddenView2");
        List<ViewItem> view2Items = view2.getViewItems();
        Assert.assertEquals(2, view2Items.size());
        Assert.assertEquals("attributeAddedByChildOverride", view2Items.get(0).getName());
        Assert.assertEquals("attributeAddedByChildFileOverride", view2Items.get(1).getName());
        
        // overriddenView1 is redefined in the _override file. The overrides of the parent, the content type declaration and the extension is ignored
        View view3 = contentType.getView("overriddenView3");
        List<ViewItem> view3Items = view3.getViewItems();
        Assert.assertEquals(1, view3Items.size());
        Assert.assertEquals("attributeAddedByChildFileOverride", view3Items.get(0).getName());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with overriding inherited view of children
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-inheritance"
    })
    public void testOverrideViewsInheritance()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.inheritance.grandchild");
        View view = contentType.getView("overriddenView");

        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(4, viewItems.size());
        
        Assert.assertEquals("title", viewItems.get(0).getName());
        
        SimpleViewItemGroup group = (SimpleViewItemGroup) viewItems.get(1);
        Assert.assertEquals("group", group.getName());
        List<ViewItem> groupItems = group.getViewItems();
        Assert.assertEquals(3, groupItems.size());
        Assert.assertEquals("attributeAddedByParentOverride", groupItems.get(0).getName());
        Assert.assertEquals("attributeAddedByChildOverride", groupItems.get(1).getName());
        Assert.assertEquals("attributeAddedByGrandchildOverride", groupItems.get(2).getName());
        
        Assert.assertEquals("attributeAddedByChild", viewItems.get(2).getName());
        Assert.assertEquals("attributeAddedByGrandchild", viewItems.get(3).getName());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with overriding inherited view of children with several parents
     * Final view is like:
     * - title
     * - attributeAddedByParent2
     * - attributeAddedByChild
     * - group
     *   - attributeAddedByParent1Override
     *   - attributeAddedByParent2Override
     *   - attributeAddedByChildOverride
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/content-type-override-view-multi-inheritance"
    })
    public void testOverrideViewsMultiInheritance()
    {
        ContentType contentType = _contentTypeEP.getExtension("contenttype.override.view.multi.inheritance.child");
        View view = contentType.getView("overriddenView");

        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(5, viewItems.size());
        
        Assert.assertEquals("title", viewItems.get(0).getName());
        Assert.assertEquals("attributeAddedByParent2", viewItems.get(1).getName());
        
        SimpleViewItemGroup group1 = (SimpleViewItemGroup) viewItems.get(2);
        Assert.assertEquals("group", group1.getName());
        List<ViewItem> group1Items = group1.getViewItems();
        Assert.assertEquals(2, group1Items.size());
        Assert.assertEquals("attributeAddedByParent2Override", group1Items.get(0).getName());
        Assert.assertEquals("attributeAddedByChildOverride", group1Items.get(1).getName());
        
        SimpleViewItemGroup group2 = (SimpleViewItemGroup) viewItems.get(3);
        Assert.assertEquals("group", group2.getName());
        List<ViewItem> group2Items = group2.getViewItems();
        Assert.assertEquals(1, group2Items.size());
        Assert.assertEquals("attributeAddedByParent1Override", group2Items.get(0).getName());
        
        Assert.assertEquals("attributeAddedByChild", viewItems.get(4).getName());
    }
}
