/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.model.ContentRestrictedCompositeDefinition;
import org.ametys.cms.model.ContentRestrictedRepeaterDefinition;
import org.ametys.cms.repository.ContentTypeExpression;
import org.ametys.cms.repository.DefaultContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.util.I18nUtils;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.query.expression.Expression.Operator;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemGroup;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.SimpleViewItemGroup;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElement;
import org.ametys.runtime.model.ViewElementAccessor;
import org.ametys.runtime.model.ViewHelper;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.ViewItemGroup;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ContentType} test case.
 */
@NeedFeatures ({
    "cms/workflow.content",
    "core/util.i18nutils"
})
public class ContentTypeTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    /** Content type extension point */
    protected static ContentTypeExtensionPoint _contentTypeEP;
    
    /** I18nUtils */
    protected static I18nUtils _i18nUtils;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _contentTypeEP = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
    }
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }
    
    /**
     * Test {@link ContentType} inheritance.
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures({
        "test/content-type",
        "test/content-type-extended",
        "test/content-type-extended-2",
    })
    public void testInheritance() throws Exception
    {
        // Test supertype ids.
        ContentType extended = _contentTypeEP.getExtension("org.ametys.cms.contenttype.Test.extended");
        Assert.assertEquals("org.ametys.cms.contenttype.Test", extended.getSupertypeIds()[0]);
        
        ContentType extended2 = _contentTypeEP.getExtension("org.ametys.cms.contenttype.Test.extended2");
        Assert.assertEquals("org.ametys.cms.contenttype.Test.extended", extended2.getSupertypeIds()[0]);
        
        // Test base type descendants.
        Set<String> subtypes = _contentTypeEP.getDirectSubTypes("org.ametys.cms.contenttype.Test");
        Assert.assertEquals(subtypes, Collections.singleton("org.ametys.cms.contenttype.Test.extended"));
        
        subtypes = _contentTypeEP.getSubTypes("org.ametys.cms.contenttype.Test");
        Assert.assertTrue(subtypes.contains("org.ametys.cms.contenttype.Test.extended"));
        Assert.assertTrue(subtypes.contains("org.ametys.cms.contenttype.Test.extended2"));
        
        // Test "extended" descendants.
        subtypes = _contentTypeEP.getDirectSubTypes("org.ametys.cms.contenttype.Test.extended");
        Assert.assertEquals(subtypes, Collections.singleton("org.ametys.cms.contenttype.Test.extended2"));
        
        subtypes = _contentTypeEP.getSubTypes("org.ametys.cms.contenttype.Test.extended");
        Assert.assertEquals(subtypes, Collections.singleton("org.ametys.cms.contenttype.Test.extended2"));
        
        // Test "extended2" descendants.
        subtypes = _contentTypeEP.getDirectSubTypes("org.ametys.cms.contenttype.Test.extended2");
        Assert.assertTrue(subtypes.isEmpty());
        
        subtypes = _contentTypeEP.getSubTypes("org.ametys.cms.contenttype.Test.extended2");
        Assert.assertTrue(subtypes.isEmpty());
    }
    
    
    /**
     * Test {@link ContentType#getViewNames()} method
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testGetViewNames()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.extends.exhaustive");
        
        // Test getting view names
        Set<String> viewNames = contentType.getViewNames();
        Assert.assertEquals(20, viewNames.size());
        Assert.assertTrue(viewNames.contains("main"));
        Assert.assertTrue(viewNames.contains("abstract"));
        Assert.assertTrue(viewNames.contains("link"));
        Assert.assertTrue(viewNames.contains("details"));
        Assert.assertTrue(viewNames.contains("complete-test"));
        Assert.assertTrue(viewNames.contains("complete-test-old-syntax"));
        Assert.assertTrue(viewNames.contains("many-joins-on-same-content-attribute"));
        Assert.assertTrue(viewNames.contains("simple-view-with-view-refs"));
        Assert.assertTrue(viewNames.contains("in-both-syntaxes"));
        Assert.assertTrue(viewNames.contains("same-item-in-composite"));
        Assert.assertTrue(viewNames.contains("with-simplified-access-to-item"));
        Assert.assertTrue(viewNames.contains("with-simplified-access-to-view"));
        Assert.assertTrue(viewNames.contains("with-simplified-access-to-all-items"));
        Assert.assertTrue(viewNames.contains("viewA"));
        Assert.assertTrue(viewNames.contains("viewB"));
        Assert.assertTrue(viewNames.contains("viewC"));
        Assert.assertTrue(viewNames.contains("viewD"));
        Assert.assertTrue(viewNames.contains("viewE"));
        Assert.assertTrue(viewNames.contains("link-content"));
        Assert.assertTrue(viewNames.contains("viewFromOverride"));
        
        // Test getting view names, excluding internal views
        Set<String> viewNamesWithoutInternals = contentType.getViewNames(false);
        Assert.assertEquals(17, viewNamesWithoutInternals.size());
        Assert.assertTrue(viewNamesWithoutInternals.contains("main"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("abstract"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("link"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("details"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("complete-test"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("complete-test-old-syntax"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("many-joins-on-same-content-attribute"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("simple-view-with-view-refs"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("in-both-syntaxes"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("same-item-in-composite"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("with-simplified-access-to-item"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("with-simplified-access-to-view"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("with-simplified-access-to-all-items"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("viewA"));
//        Assert.assertTrue(viewNamesWithoutInternals.contains("viewB"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("viewC"));
//        Assert.assertTrue(viewNamesWithoutInternals.contains("viewD"));
//        Assert.assertTrue(viewNamesWithoutInternals.contains("viewE"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("link-content"));
        Assert.assertTrue(viewNamesWithoutInternals.contains("viewFromOverride"));
    }
    
    /**
     * Test {@link ContentType#getView(String)} method for a complete view with groups
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testCompleteView()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.extends.exhaustive");
        
        View testsView = contentType.getView("complete-test");
        Assert.assertEquals("Tests", testsView.getLabel().getLabel());
        Assert.assertEquals("View for tests", testsView.getDescription().getLabel());
        List<ViewItem> testsViewItems = testsView.getViewItems();
        Assert.assertEquals(4, testsViewItems.size());
        
        // View - Root fieldset
        SimpleViewItemGroup rootFieldset = (SimpleViewItemGroup) testsViewItems.get(0);
        Assert.assertEquals(ViewItemGroup.TAB_ROLE, rootFieldset.getRole());
        Assert.assertEquals("Root", rootFieldset.getLabel().getLabel());
        Assert.assertEquals(testsView, rootFieldset.getParent());
        List<ViewItem> rootFieldsetChildren = rootFieldset.getViewItems();
        Assert.assertEquals(3, rootFieldsetChildren.size());
        
        // View - Root fieldset - Primitive fieldset
        SimpleViewItemGroup primitiveFieldset = (SimpleViewItemGroup) rootFieldsetChildren.get(0);
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, primitiveFieldset.getRole());
        Assert.assertEquals("Primitive", primitiveFieldset.getLabel().getLabel());
        Assert.assertEquals(rootFieldset, primitiveFieldset.getParent());
        List<ViewItem> primitiveFieldsetChildren = primitiveFieldset.getViewItems();
        Assert.assertEquals(3, primitiveFieldsetChildren.size());
        
        // View - Root fieldset - Primitive fieldset - string
        ViewElement string = (ViewElement) primitiveFieldsetChildren.get(0);
        Assert.assertEquals("string", string.getDefinition().getName());
        Assert.assertEquals(primitiveFieldset, string.getParent());
        
        // View - Root fieldset - Primitive fieldset - long
        ViewElement longelement = (ViewElement) primitiveFieldsetChildren.get(1);
        Assert.assertEquals("long", longelement.getDefinition().getName());
        Assert.assertEquals(primitiveFieldset, longelement.getParent());
        
        // View - Root fieldset - Primitive fieldset - date
        ViewElement date = (ViewElement) primitiveFieldsetChildren.get(2);
        Assert.assertEquals("date", date.getDefinition().getName());
        Assert.assertEquals(primitiveFieldset, date.getParent());
        
        // View - Root fieldset - GPS
        ViewElement gps = (ViewElement) rootFieldsetChildren.get(1);
        Assert.assertEquals("gps", gps.getDefinition().getName());
        Assert.assertEquals(rootFieldset, gps.getParent());
        
        // View - Root fieldset - Nested attributes
        SimpleViewItemGroup nestedAttributesFieldset = (SimpleViewItemGroup) rootFieldsetChildren.get(2);
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, nestedAttributesFieldset.getRole());
        Assert.assertEquals("Nested attributes", nestedAttributesFieldset.getLabel().getLabel());
        Assert.assertEquals(rootFieldset, nestedAttributesFieldset.getParent());
        List<ViewItem> nestedAttributesFieldsetChildren = nestedAttributesFieldset.getViewItems();
        Assert.assertEquals(4, nestedAttributesFieldsetChildren.size());
        
        // View - Root fieldset - Nested attributes - Owns
        ViewElementAccessor owns = (ViewElementAccessor) nestedAttributesFieldsetChildren.get(0);
        Assert.assertEquals("owns", owns.getDefinition().getName());
        Assert.assertEquals(nestedAttributesFieldset, owns.getParent());
        List<ViewItem> ownsItems = owns.getViewItems();
        Assert.assertEquals(3, ownsItems.size());
        
        // View - Root fieldset - Nested attributes - Owns - String
        ViewElement ownsString = (ViewElement) ownsItems.get(0);
        Assert.assertEquals("string", ownsString.getDefinition().getName());
        Assert.assertEquals(owns, ownsString.getParent());
        
        // View - Root fieldset - Nested attributes - Owns - Long
        ViewElement ownsLong = (ViewElement) ownsItems.get(1);
        Assert.assertEquals("long", ownsLong.getDefinition().getName());
        Assert.assertEquals(owns, ownsLong.getParent());
        
        // View - Root fieldset - Nested attributes - Owns - Link content
        ViewElementAccessor ownsLinkContent = (ViewElementAccessor) ownsItems.get(2);
        Assert.assertEquals("linkContent", ownsLinkContent.getDefinition().getName());
        Assert.assertEquals(owns, ownsLinkContent.getParent());
        List<ViewItem> ownsLinkContentItems = ownsLinkContent.getViewItems();
        Assert.assertEquals(1, ownsLinkContentItems.size());
        
        // View - Root fieldset - Nested attributes - Owns - Link content - Title
        ViewElement ownsLinkContentTitle = (ViewElement) ownsLinkContentItems.get(0);
        Assert.assertEquals("title", ownsLinkContentTitle.getDefinition().getName());
        Assert.assertEquals(ownsLinkContent, ownsLinkContentTitle.getParent());
        
        // View - Root fieldset - Nested attributes - Sport
        ViewElementAccessor sport = (ViewElementAccessor) nestedAttributesFieldsetChildren.get(1);
        Assert.assertEquals("sport", sport.getDefinition().getName());
        Assert.assertEquals(nestedAttributesFieldset, sport.getParent());
        List<ViewItem> sportItems = sport.getViewItems();
        Assert.assertEquals(1, sportItems.size());
        
        // View - Root fieldset - Nested attributes - Sport - Title
        ViewElement sportTitle = (ViewElement) sportItems.get(0);
        Assert.assertEquals("title", sportTitle.getDefinition().getName());
        Assert.assertEquals(sport, sportTitle.getParent());
        
        // View - Root fieldset - Nested attributes - Multilingual content
        ViewElementAccessor multilingual = (ViewElementAccessor) nestedAttributesFieldsetChildren.get(2);
        Assert.assertEquals("multilingualContent", multilingual.getDefinition().getName());
        Assert.assertEquals(nestedAttributesFieldset, multilingual.getParent());
        List<ViewItem> multilingualItems = multilingual.getViewItems();
        Assert.assertEquals(3, multilingualItems.size());
        
        // View - Root fieldset - Nested attributes - Multilingual content - Title
        ViewElement multilingualTitle = (ViewElement) multilingualItems.get(0);
        Assert.assertEquals("title", multilingualTitle.getDefinition().getName());
        Assert.assertEquals(multilingual, multilingualTitle.getParent());
        
        // View - Root fieldset - Nested attributes - Multilingual content - No multilingual
        ViewElement multilingualNoMultilingual = (ViewElement) multilingualItems.get(1);
        Assert.assertEquals("no-multilingual", multilingualNoMultilingual.getDefinition().getName());
        Assert.assertEquals(multilingual, multilingualNoMultilingual.getParent());
        
        // View - Root fieldset - Nested attributes - Multilingual content - Multilingual
        ViewElement multilingualMultilingual = (ViewElement) multilingualItems.get(2);
        Assert.assertEquals("multilingual", multilingualMultilingual.getDefinition().getName());
        Assert.assertEquals(multilingual, multilingualMultilingual.getParent());
        
        // View - Root fieldset - Nested attributes - Link On Itself
        ViewElementAccessor linkOnItself = (ViewElementAccessor) nestedAttributesFieldsetChildren.get(3);
        Assert.assertEquals("linkOnItself", linkOnItself.getDefinition().getName());
        Assert.assertEquals(nestedAttributesFieldset, linkOnItself.getParent());
        List<ViewItem> linkOnItselfItems = linkOnItself.getViewItems();
        Assert.assertEquals(4, linkOnItselfItems.size());
        
        // View - Root fieldset - Nested attributes - Link On Itself - Link On Itself
        ViewElementAccessor linkOnItselfAgain = (ViewElementAccessor) linkOnItselfItems.get(0);
        Assert.assertEquals("linkOnItself", linkOnItselfAgain.getDefinition().getName());
        Assert.assertEquals(linkOnItself, linkOnItselfAgain.getParent());
        List<ViewItem> linkOnItselfAgainItems = linkOnItselfAgain.getViewItems();
        Assert.assertEquals(2, linkOnItselfAgainItems.size());
        
        // View - Root fieldset - Nested attributes - Link On Itself - Link On Itself - string
        ViewElement linkOnItselfAgainString = (ViewElement) linkOnItselfAgainItems.get(0);
        Assert.assertEquals("string", linkOnItselfAgainString.getDefinition().getName());
        Assert.assertEquals(linkOnItselfAgain, linkOnItselfAgainString.getParent());
        
        // View - Root fieldset - Nested attributes - Link On Itself - Link On Itself - long
        ViewElement linkOnItselfAgainLong = (ViewElement) linkOnItselfAgainItems.get(1);
        Assert.assertEquals("long", linkOnItselfAgainLong.getDefinition().getName());
        Assert.assertEquals(linkOnItselfAgain, linkOnItselfAgainLong.getParent());
        
        // View - Root fieldset - Nested attributes - Link On Itself - date
        ViewElement linkOnItselfDate = (ViewElement) linkOnItselfItems.get(1);
        Assert.assertEquals("date", linkOnItselfDate.getDefinition().getName());
        Assert.assertEquals(linkOnItself, linkOnItselfDate.getParent());
        
        // View - Root fieldset - Nested attributes - Link On Itself - string
        ViewElement linkOnItselfString = (ViewElement) linkOnItselfItems.get(2);
        Assert.assertEquals("string", linkOnItselfString.getDefinition().getName());
        Assert.assertEquals(linkOnItself, linkOnItselfString.getParent());
        
        // View - Root fieldset - Nested attributes - Link On Itself - long
        ViewElement linkOnItselfLong = (ViewElement) linkOnItselfItems.get(3);
        Assert.assertEquals("long", linkOnItselfLong.getDefinition().getName());
        Assert.assertEquals(linkOnItself, linkOnItselfLong.getParent());
        
        // View - Illustration
        ModelViewItemGroup illustration = (ModelViewItemGroup) testsViewItems.get(1);
        Assert.assertEquals("illustration", illustration.getDefinition().getName());
        Assert.assertEquals(testsView, illustration.getParent());
        List<ViewItem> illustrationChildren = illustration.getViewItems();
        Assert.assertEquals(2, illustrationChildren.size());
        
        // View - Illustration - Image
        ViewElement image = (ViewElement) illustrationChildren.get(0);
        Assert.assertEquals("image", image.getDefinition().getName());
        Assert.assertEquals(illustration, image.getParent());
        
        // View - Illustration - Alt-Text
        ViewElement altText = (ViewElement) illustrationChildren.get(1);
        Assert.assertEquals("alt-text", altText.getDefinition().getName());
        Assert.assertEquals(illustration, altText.getParent());
        
        // View - Composite fieldset
        SimpleViewItemGroup compositeFieldset = (SimpleViewItemGroup) testsViewItems.get(2);
        Assert.assertEquals(ViewItemGroup.TAB_ROLE, compositeFieldset.getRole());
        Assert.assertEquals("Composite", compositeFieldset.getLabel().getLabel());
        Assert.assertEquals(testsView, compositeFieldset.getParent());
        List<ViewItem> compositeFieldsetChildren = compositeFieldset.getViewItems();
        Assert.assertEquals(1, compositeFieldsetChildren.size());
        
        // View - Composite fieldset - Composite
        ModelViewItemGroup composite = (ModelViewItemGroup) compositeFieldsetChildren.get(0);
        Assert.assertEquals("composite", composite.getDefinition().getName());
        Assert.assertEquals(compositeFieldset, composite.getParent());
        List<ViewItem> compositeChildren = composite.getViewItems();
        Assert.assertEquals(2, compositeChildren.size());
        
        // View - Composite fieldset - Composite - Simple fieldset
        SimpleViewItemGroup simpleFieldset = (SimpleViewItemGroup) compositeChildren.get(0);
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, simpleFieldset.getRole());
        Assert.assertEquals("Simple", simpleFieldset.getLabel().getLabel());
        Assert.assertEquals(composite, simpleFieldset.getParent());
        List<ViewItem> simpleFieldsetChildren = simpleFieldset.getViewItems();
        Assert.assertEquals(2, simpleFieldsetChildren.size());
        
        // View - Composite fieldset - Composite - Simple fieldset - long-min
        ViewElement longMin = (ViewElement) simpleFieldsetChildren.get(0);
        Assert.assertEquals("long-min", longMin.getDefinition().getName());
        Assert.assertEquals(simpleFieldset, longMin.getParent());
        
        // View - Composite fieldset - Composite - Simple fieldset - long-max
        ViewElement longMax = (ViewElement) simpleFieldsetChildren.get(1);
        Assert.assertEquals("long-max", longMax.getDefinition().getName());
        Assert.assertEquals(simpleFieldset, longMax.getParent());
        
        // View - Composite fieldset - Composite - Repeater fieldset
        SimpleViewItemGroup repeaterFieldset = (SimpleViewItemGroup) compositeChildren.get(1);
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, repeaterFieldset.getRole());
        Assert.assertEquals("Repeater", repeaterFieldset.getLabel().getLabel());
        Assert.assertEquals(composite, repeaterFieldset.getParent());
        List<ViewItem> repeaterFieldsetChildren = repeaterFieldset.getViewItems();
        Assert.assertEquals(1, repeaterFieldsetChildren.size());
        
        // View - Composite fieldset - Composite - Repeater fieldset - repeater
        ModelViewItemGroup repeater = (ModelViewItemGroup) repeaterFieldsetChildren.get(0);
        Assert.assertEquals("repeater", repeater.getDefinition().getName());
        Assert.assertEquals(repeaterFieldset, repeater.getParent());
        List<ViewItem> repeaterChildren = repeater.getViewItems();
        Assert.assertEquals(3, repeaterChildren.size());
        
        // View - Composite fieldset - Composite - repeater - comp string
        ViewElement compString = (ViewElement) repeaterChildren.get(0);
        Assert.assertEquals("comp-string", compString.getDefinition().getName());
        Assert.assertEquals(repeater, compString.getParent());
        
        // View - Composite fieldset - Composite - repeater - comp GPS
        ViewElement compGPS = (ViewElement) repeaterChildren.get(1);
        Assert.assertEquals("comp-gps", compGPS.getDefinition().getName());
        Assert.assertEquals(repeater, compGPS.getParent());
        
        // View - Composite fieldset - Composite - repeater - SubComposite
        ModelViewItemGroup subComposite = (ModelViewItemGroup) repeaterChildren.get(2);
        Assert.assertEquals("subcomposite", subComposite.getDefinition().getName());
        Assert.assertEquals(repeater, subComposite.getParent());
        List<ViewItem> subCompositeChildren = subComposite.getViewItems();
        Assert.assertEquals(2, subCompositeChildren.size());
        
        // View - Composite fieldset - Composite - repeater - SubComposite - subcomp str
        ViewElement subCompStr = (ViewElement) subCompositeChildren.get(0);
        Assert.assertEquals("subcomp-str", subCompStr.getDefinition().getName());
        Assert.assertEquals(subComposite, subCompStr.getParent());
        
        // View - Composite fieldset - Composite - repeater - SubComposite - subcomp long
        ViewElement subCompLong = (ViewElement) subCompositeChildren.get(1);
        Assert.assertEquals("subcomp-long", subCompLong.getDefinition().getName());
        Assert.assertEquals(subComposite, subCompLong.getParent());
        
        // View - Three levels repeater
        ModelViewItemGroup threeLevelsRepeater = (ModelViewItemGroup) testsViewItems.get(3);
        Assert.assertEquals("repeaterThreeLevels", threeLevelsRepeater.getDefinition().getName());
        Assert.assertEquals(testsView, threeLevelsRepeater.getParent());
        List<ViewItem> threeLevelsRepeaterChildren = threeLevelsRepeater.getViewItems();
        Assert.assertEquals(2, threeLevelsRepeaterChildren.size());
        
        // View - Three levels repeater - rep string
        ViewElement repString = (ViewElement) threeLevelsRepeaterChildren.get(0);
        Assert.assertEquals("rep-string", repString.getDefinition().getName());
        Assert.assertEquals(threeLevelsRepeater, repString.getParent());
        
        // View - Three levels repeater - repeater2
        ModelViewItemGroup repeater2 = (ModelViewItemGroup) threeLevelsRepeaterChildren.get(1);
        Assert.assertEquals("repeater2", repeater2.getDefinition().getName());
        Assert.assertEquals(threeLevelsRepeater, repeater2.getParent());
        List<ViewItem> repeater2Children = repeater2.getViewItems();
        Assert.assertEquals(1, repeater2Children.size());
        
        // View - Three levels repeater - repeater2 - repeater3
        ModelViewItemGroup repeater3 = (ModelViewItemGroup) repeater2Children.get(0);
        Assert.assertEquals("repeater3", repeater3.getDefinition().getName());
        Assert.assertEquals(repeater2, repeater3.getParent());
        List<ViewItem> repeater3Children = repeater3.getViewItems();
        Assert.assertEquals(6, repeater3Children.size());
        
        // View - Three levels repeater - repeater2 - repeater3 - rep3 string
        ViewElement rep3String = (ViewElement) repeater3Children.get(0);
        Assert.assertEquals("rep3-string", rep3String.getDefinition().getName());
        Assert.assertEquals(repeater3, rep3String.getParent());
        
        // View - Three levels repeater - repeater2 - repeater3 - rep3 long
        ViewElement rep3Long = (ViewElement) repeater3Children.get(1);
        Assert.assertEquals("rep3-long", rep3Long.getDefinition().getName());
        Assert.assertEquals(repeater3, rep3Long.getParent());
        
        // View - Three levels repeater - repeater2 - repeater3 - rep3 longmult-min
        ViewElement rep3LongmultMin = (ViewElement) repeater3Children.get(2);
        Assert.assertEquals("rep3-longmult-min", rep3LongmultMin.getDefinition().getName());
        Assert.assertEquals(repeater3, rep3LongmultMin.getParent());
        
        // View - Three levels repeater - repeater2 - repeater3 - rep3 longmult-max
        ViewElement rep3LongmultMax = (ViewElement) repeater3Children.get(3);
        Assert.assertEquals("rep3-longmult-max", rep3LongmultMax.getDefinition().getName());
        Assert.assertEquals(repeater3, rep3LongmultMax.getParent());
        
        // View - Three levels repeater - repeater2 - repeater3 - rep3 owns
        ViewElementAccessor rep3Owns = (ViewElementAccessor) repeater3Children.get(4);
        Assert.assertEquals("rep3-owns", rep3Owns.getDefinition().getName());
        Assert.assertEquals(repeater3, rep3Owns.getParent());
        
        // View - Three levels repeater - repeater2 - repeater3 - rep3 belongs to
        ViewElementAccessor rep3BelongsTo = (ViewElementAccessor) repeater3Children.get(5);
        Assert.assertEquals("rep3-belongsTo", rep3BelongsTo.getDefinition().getName());
        Assert.assertEquals(repeater3, rep3BelongsTo.getParent());
    }
    
    /**
     * Test {@link ContentType#getView(String)} method for a complete view with groups, configured with the old metadata syntax
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testCompleteViewWithOldMetadataSyntax()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.extends.exhaustive");
        
        View testsView = contentType.getView("complete-test-old-syntax");
        Assert.assertEquals("Tests", testsView.getLabel().getLabel());
        Assert.assertEquals("View for tests", testsView.getDescription().getLabel());
        List<ViewItem> testsViewItems = testsView.getViewItems();
        Assert.assertEquals(4, testsViewItems.size());
        
        // View - Root fieldset
        SimpleViewItemGroup rootFieldset = (SimpleViewItemGroup) testsViewItems.get(0);
        Assert.assertEquals(ViewItemGroup.TAB_ROLE, rootFieldset.getRole());
        Assert.assertEquals("Root", rootFieldset.getLabel().getLabel());
        List<ViewItem> rootFieldsetChildren = rootFieldset.getViewItems();
        Assert.assertEquals(3, rootFieldsetChildren.size());
        
        // View - Root fieldset - Primitive fieldset
        SimpleViewItemGroup primitiveFieldset = (SimpleViewItemGroup) rootFieldsetChildren.get(0);
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, primitiveFieldset.getRole());
        Assert.assertEquals("Primitive", primitiveFieldset.getLabel().getLabel());
        List<ViewItem> primitiveFieldsetChildren = primitiveFieldset.getViewItems();
        Assert.assertEquals(3, primitiveFieldsetChildren.size());
        
        // View - Root fieldset - Primitive fieldset - string
        ViewElement string = (ViewElement) primitiveFieldsetChildren.get(0);
        Assert.assertEquals("string", string.getDefinition().getName());
        
        // View - Root fieldset - Primitive fieldset - long
        ViewElement longelement = (ViewElement) primitiveFieldsetChildren.get(1);
        Assert.assertEquals("long", longelement.getDefinition().getName());
        
        // View - Root fieldset - Primitive fieldset - date
        ViewElement date = (ViewElement) primitiveFieldsetChildren.get(2);
        Assert.assertEquals("date", date.getDefinition().getName());
        
        // View - Root fieldset - GPS
        ViewElement gps = (ViewElement) rootFieldsetChildren.get(1);
        Assert.assertEquals("gps", gps.getDefinition().getName());
        
        // View - Root fieldset - Contents
        SimpleViewItemGroup contentsFieldset = (SimpleViewItemGroup) rootFieldsetChildren.get(2);
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, contentsFieldset.getRole());
        Assert.assertEquals("Contents", contentsFieldset.getLabel().getLabel());
        List<ViewItem> contentsFieldsetChildren = contentsFieldset.getViewItems();
        Assert.assertEquals(4, contentsFieldsetChildren.size());
        
        // View - Root fieldset - Contents - Owns
        ViewElement owns = (ViewElement) contentsFieldsetChildren.get(0);
        Assert.assertEquals("owns", owns.getDefinition().getName());
        
        // View - Root fieldset - Contents - Sport
        ViewElement sport = (ViewElement) contentsFieldsetChildren.get(1);
        Assert.assertEquals("sport", sport.getDefinition().getName());
        
        // View - Root fieldset - Contents - Multilingual content
        ViewElement multilingual = (ViewElement) contentsFieldsetChildren.get(2);
        Assert.assertEquals("multilingualContent", multilingual.getDefinition().getName());
        
        // View - Root fieldset - Contents - Link on itself
        ViewElement linkOnItself = (ViewElement) contentsFieldsetChildren.get(3);
        Assert.assertEquals("linkOnItself", linkOnItself.getDefinition().getName());
        
        // View - Illustration
        ModelViewItemGroup illustration = (ModelViewItemGroup) testsViewItems.get(1);
        Assert.assertEquals("illustration", illustration.getDefinition().getName());
        List<ViewItem> illustrationChildren = illustration.getViewItems();
        Assert.assertEquals(2, illustrationChildren.size());
        
        // View - Illustration - Image
        ViewElement image = (ViewElement) illustrationChildren.get(0);
        Assert.assertEquals("image", image.getDefinition().getName());
        
        // View - Illustration - Alt-Text
        ViewElement altText = (ViewElement) illustrationChildren.get(1);
        Assert.assertEquals("alt-text", altText.getDefinition().getName());
        
        // View - Composite fieldset
        SimpleViewItemGroup compositeFieldset = (SimpleViewItemGroup) testsViewItems.get(2);
        Assert.assertEquals(ViewItemGroup.TAB_ROLE, compositeFieldset.getRole());
        Assert.assertEquals("Composite", compositeFieldset.getLabel().getLabel());
        List<ViewItem> compositeFieldsetChildren = compositeFieldset.getViewItems();
        Assert.assertEquals(1, compositeFieldsetChildren.size());
        
        // View - Composite fieldset - Composite
        ModelViewItemGroup composite = (ModelViewItemGroup) compositeFieldsetChildren.get(0);
        Assert.assertEquals("composite", composite.getDefinition().getName());
        List<ViewItem> compositeChildren = composite.getViewItems();
        Assert.assertEquals(2, compositeChildren.size());
        
        // View - Composite fieldset - Composite - Simple fieldset
        SimpleViewItemGroup simpleFieldset = (SimpleViewItemGroup) compositeChildren.get(0);
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, simpleFieldset.getRole());
        Assert.assertEquals("Simple", simpleFieldset.getLabel().getLabel());
        List<ViewItem> simpleFieldsetChildren = simpleFieldset.getViewItems();
        Assert.assertEquals(2, simpleFieldsetChildren.size());
        
        // View - Composite fieldset - Composite - Simple fieldset - long-min
        ViewElement longMin = (ViewElement) simpleFieldsetChildren.get(0);
        Assert.assertEquals("long-min", longMin.getDefinition().getName());
        
        // View - Composite fieldset - Composite - Simple fieldset - long-max
        ViewElement longMax = (ViewElement) simpleFieldsetChildren.get(1);
        Assert.assertEquals("long-max", longMax.getDefinition().getName());
        
        // View - Composite fieldset - Composite - Repeater fieldset
        SimpleViewItemGroup repeaterFieldset = (SimpleViewItemGroup) compositeChildren.get(1);
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, repeaterFieldset.getRole());
        Assert.assertEquals("Repeater", repeaterFieldset.getLabel().getLabel());
        List<ViewItem> repeaterFieldsetChildren = repeaterFieldset.getViewItems();
        Assert.assertEquals(1, repeaterFieldsetChildren.size());
        
        // View - Composite fieldset - Composite - Repeater fieldset - repeater
        ModelViewItemGroup repeater = (ModelViewItemGroup) repeaterFieldsetChildren.get(0);
        Assert.assertEquals("repeater", repeater.getDefinition().getName());
        List<ViewItem> repeaterChildren = repeater.getViewItems();
        Assert.assertEquals(3, repeaterChildren.size());
        
        // View - Composite fieldset - Composite - repeater - comp string
        ViewElement compString = (ViewElement) repeaterChildren.get(0);
        Assert.assertEquals("comp-string", compString.getDefinition().getName());
        
        // View - Composite fieldset - Composite - repeater - comp GPS
        ViewElement compGPS = (ViewElement) repeaterChildren.get(1);
        Assert.assertEquals("comp-gps", compGPS.getDefinition().getName());
        
        // View - Composite fieldset - Composite - repeater - SubComposite
        ModelViewItemGroup subComposite = (ModelViewItemGroup) repeaterChildren.get(2);
        Assert.assertEquals("subcomposite", subComposite.getDefinition().getName());
        List<ViewItem> subCompositeChildren = subComposite.getViewItems();
        Assert.assertEquals(2, subCompositeChildren.size());
        
        // View - Composite fieldset - Composite - repeater - SubComposite - subcomp str
        ViewElement subCompStr = (ViewElement) subCompositeChildren.get(0);
        Assert.assertEquals("subcomp-str", subCompStr.getDefinition().getName());
        
        // View - Composite fieldset - Composite - repeater - SubComposite - subcomp long
        ViewElement subCompLong = (ViewElement) subCompositeChildren.get(1);
        Assert.assertEquals("subcomp-long", subCompLong.getDefinition().getName());
        
        // View - Three levels repeater
        ModelViewItemGroup threeLevelsRepeater = (ModelViewItemGroup) testsViewItems.get(3);
        Assert.assertEquals("repeaterThreeLevels", threeLevelsRepeater.getDefinition().getName());
        List<ViewItem> threeLevelsRepeaterChildren = threeLevelsRepeater.getViewItems();
        Assert.assertEquals(2, threeLevelsRepeaterChildren.size());
        
        // View - Three levels repeater - rep string
        ViewElement repString = (ViewElement) threeLevelsRepeaterChildren.get(0);
        Assert.assertEquals("rep-string", repString.getDefinition().getName());
        
        // View - Three levels repeater - repeater2
        ModelViewItemGroup repeater2 = (ModelViewItemGroup) threeLevelsRepeaterChildren.get(1);
        Assert.assertEquals("repeater2", repeater2.getDefinition().getName());
        List<ViewItem> repeater2Children = repeater2.getViewItems();
        Assert.assertEquals(1, repeater2Children.size());
        
        // View - Three levels repeater - repeater2 - repeater3
        ModelViewItemGroup repeater3 = (ModelViewItemGroup) repeater2Children.get(0);
        Assert.assertEquals("repeater3", repeater3.getDefinition().getName());
        List<ViewItem> repeater3Children = repeater3.getViewItems();
        Assert.assertEquals(1, repeater3Children.size());
        
        // View - Three levels repeater - repeater2 - repeater3 - rep3 string
        ViewElement rep3String = (ViewElement) repeater3Children.get(0);
        Assert.assertEquals("rep3-string", rep3String.getDefinition().getName());
    }
    
    /**
     * Test {@link ContentType#getView(String)} method for a content attribute inherited
     * The inherited attribute defines a content type and an invert relation path.
     * This test case checks that the attribute referenced in the views on content type child contains the invert relation path
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testViewWithInheritedAttribute()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.invert.relation.in.view");
        
        // View defined by the child (check that the referenced linkContentAttribute is the child's one)
        View childView = contentType.getView("link-content");
        ModelViewItem linkContentFromChildView = childView.getModelViewItem("linkContent");
        ContentAttributeDefinition linkContentDefFromChild = (ContentAttributeDefinition) linkContentFromChildView.getDefinition();
        Assert.assertEquals("linkContent", linkContentDefFromChild.getName());
        Assert.assertEquals("tests.invert.relation.in.view", linkContentDefFromChild.getContentTypeId());
        Assert.assertEquals("linkContentTarget", linkContentDefFromChild.getInvertRelationPath());
        
        // View defined by the parent (check that the referenced linkContentAttribute is the child's one)
        View mainView = contentType.getView("main");
        ModelViewItem linkContentFromMainView = mainView.getModelViewItem("linkContent");
        ContentAttributeDefinition linkContentDefFromParent = (ContentAttributeDefinition) linkContentFromMainView.getDefinition();
        Assert.assertEquals("linkContent", linkContentDefFromParent.getName());
        Assert.assertEquals("tests.invert.relation.in.view", linkContentDefFromParent.getContentTypeId());
        Assert.assertEquals("linkContentTarget", linkContentDefFromParent.getInvertRelationPath());
    }
    
    /**
     * Test {@link ContentType#getView(String)} method for a view with some joins on the same content attribute
     * The attribute owns appear some times in the view, using same or different sub attributes
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testViewWithManyJoinsOnSameContentAttribute()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.extends.exhaustive");
        
        View testsView = contentType.getView("many-joins-on-same-content-attribute");
        List<ViewItem> testsViewItems = testsView.getViewItems();
        Assert.assertEquals(5, testsViewItems.size());
        
        // View - Primitive fieldset
        SimpleViewItemGroup primitiveFieldset = (SimpleViewItemGroup) testsViewItems.get(0);
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, primitiveFieldset.getRole());
        Assert.assertEquals("Primitive", primitiveFieldset.getLabel().getLabel());
        List<ViewItem> primitiveFieldsetChildren = primitiveFieldset.getViewItems();
        Assert.assertEquals(1, primitiveFieldsetChildren.size());
        
        // View - Primitive fieldset - Owns
        ViewElementAccessor primitiveOwns = (ViewElementAccessor) primitiveFieldsetChildren.get(0);
        Assert.assertEquals("owns", primitiveOwns.getDefinition().getName());
        List<ViewItem> primitiveOwnsItems = primitiveOwns.getViewItems();
        Assert.assertEquals(2, primitiveOwnsItems.size());
        
        // View - Primitive fieldset - Owns - String
        ViewElement primitiveOwnsString = (ViewElement) primitiveOwnsItems.get(0);
        Assert.assertEquals("string", primitiveOwnsString.getDefinition().getName());
        
        // View - Primitive fieldset - Owns - Long
        ViewElement primitiveOwnsLong = (ViewElement) primitiveOwnsItems.get(1);
        Assert.assertEquals("long", primitiveOwnsLong.getDefinition().getName());
        
        // View - Other primitives fieldset
        SimpleViewItemGroup otherPrimitiveFieldset = (SimpleViewItemGroup) testsViewItems.get(1);
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, otherPrimitiveFieldset.getRole());
        Assert.assertEquals("Other primitives", otherPrimitiveFieldset.getLabel().getLabel());
        List<ViewItem> otherPrimitiveFieldsetChildren = otherPrimitiveFieldset.getViewItems();
        Assert.assertEquals(1, otherPrimitiveFieldsetChildren.size());
        
        // View - Other primitives fieldset - Owns
        ViewElementAccessor otherPrimitiveOwns = (ViewElementAccessor) otherPrimitiveFieldsetChildren.get(0);
        Assert.assertEquals("owns", otherPrimitiveOwns.getDefinition().getName());
        List<ViewItem> otherPrimitiveOwnsItems = otherPrimitiveOwns.getViewItems();
        Assert.assertEquals(2, otherPrimitiveOwnsItems.size());
        
        // View - Other primitives fieldset - Owns - Double
        ViewElement otherPrimitiveOwnsDouble = (ViewElement) otherPrimitiveOwnsItems.get(0);
        Assert.assertEquals("double", otherPrimitiveOwnsDouble.getDefinition().getName());
        
        // View - Other primitives fieldset - Owns - Boolean
        ViewElement otherPrimitiveOwnsBoolean = (ViewElement) otherPrimitiveOwnsItems.get(1);
        Assert.assertEquals("boolean", otherPrimitiveOwnsBoolean.getDefinition().getName());
        
        // View - Other primitives fieldset (again)
        SimpleViewItemGroup otherPrimitiveFieldset2 = (SimpleViewItemGroup) testsViewItems.get(2);
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, otherPrimitiveFieldset2.getRole());
        Assert.assertEquals("Other primitives again", otherPrimitiveFieldset2.getLabel().getLabel());
        List<ViewItem> otherPrimitiveFieldset2Children = otherPrimitiveFieldset2.getViewItems();
        Assert.assertEquals(1, otherPrimitiveFieldset2Children.size());
        
        // View - Other primitives fieldset (again) - Owns
        ViewElementAccessor otherPrimitive2Owns = (ViewElementAccessor) otherPrimitiveFieldset2Children.get(0);
        Assert.assertEquals("owns", otherPrimitive2Owns.getDefinition().getName());
        List<ViewItem> otherPrimitive2OwnsItems = otherPrimitive2Owns.getViewItems();
        Assert.assertEquals(2, otherPrimitive2OwnsItems.size());
        
        // View - Other primitives fieldset (again) - Owns - Double
        ViewElement otherPrimitive2OwnsDouble = (ViewElement) otherPrimitive2OwnsItems.get(0);
        Assert.assertEquals("double", otherPrimitive2OwnsDouble.getDefinition().getName());
        
        // View - Other primitives fieldset (again) - Owns - Boolean
        ViewElement otherPrimitive2OwnsBoolean = (ViewElement) otherPrimitive2OwnsItems.get(1);
        Assert.assertEquals("boolean", otherPrimitive2OwnsBoolean.getDefinition().getName());
        
        // View - Primitive fieldset (the one from super type)
        SimpleViewItemGroup primitiveFieldset2 = (SimpleViewItemGroup) testsViewItems.get(3);
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, primitiveFieldset2.getRole());
        Assert.assertEquals("Primitive", primitiveFieldset2.getLabel().getLabel());
        List<ViewItem> primitiveFieldset2Children = primitiveFieldset2.getViewItems();
        Assert.assertTrue(primitiveFieldset2Children.isEmpty());
        
        // View - Nested attributes
        SimpleViewItemGroup nestedFieldset = (SimpleViewItemGroup) testsViewItems.get(4);
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, nestedFieldset.getRole());
        Assert.assertEquals("Nested attributes", nestedFieldset.getLabel().getLabel());
        List<ViewItem> nestedFieldsetChildren = nestedFieldset.getViewItems();
        Assert.assertEquals(1, nestedFieldsetChildren.size());
        
        // View - Nested attributes - Owns
        ViewElementAccessor nestedOwns = (ViewElementAccessor) nestedFieldsetChildren.get(0);
        Assert.assertEquals("owns", nestedOwns.getDefinition().getName());
        List<ViewItem> nestedOwnsItems = nestedOwns.getViewItems();
        Assert.assertEquals(1, nestedOwnsItems.size());
        
        // View - Nested attributes - Owns - Link content
        ViewElementAccessor linkContent = (ViewElementAccessor) nestedOwnsItems.get(0);
        Assert.assertEquals("linkContent", linkContent.getDefinition().getName());
        List<ViewItem> linkContentItems = linkContent.getViewItems();
        Assert.assertEquals(1, linkContentItems.size());
        
        // View - Nested attributes - Owns - Link content - Title
        ViewElement linkContentTitle = (ViewElement) linkContentItems.get(0);
        Assert.assertEquals("title", linkContentTitle.getDefinition().getName());
    }
    
    /**
     * Test {@link ContentType#getView(String)} method for a view configured with both new and old attribute/metadata syntaxes
     * Only one view is kept and it is the one configured with the new attribute syntax
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testViewWithBothSyntaxes()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.extends.exhaustive");
        
        View testsView = contentType.getView("in-both-syntaxes");
        List<ViewItem> testsViewItems = testsView.getViewItems();
        Assert.assertEquals(2, testsViewItems.size());
        
        ViewElement string = (ViewElement) testsViewItems.get(0);
        Assert.assertEquals("string", string.getDefinition().getName());
        
        ViewElement longelement = (ViewElement) testsViewItems.get(1);
        Assert.assertEquals("long", longelement.getDefinition().getName());
    }
    
    /**
     * Test {@link ContentType#getView(String)} method for a complete view with groups
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testViewsInheritance()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.extends.exhaustive.and.basic");
        
        // Test the view A, inherited from parents
        View viewA = contentType.getView("viewA");
        List<ViewItem> viewAItems = viewA.getViewItems();
        Assert.assertEquals(4, viewAItems.size());
        
        Assert.assertEquals("title", ((ViewElement) viewAItems.get(0)).getDefinition().getName());
        Assert.assertEquals("abstract", ((ViewElement) viewAItems.get(1)).getDefinition().getName());
        
        SimpleViewItemGroup viewAExhFieldset1 = (SimpleViewItemGroup) viewAItems.get(2);
        Assert.assertEquals("fieldset1", viewAExhFieldset1.getName());
        Assert.assertEquals("Fieldset 1", viewAExhFieldset1.getLabel().getLabel());
        List<ViewItem> viewAExhFieldset1Children = viewAExhFieldset1.getViewItems();
        Assert.assertEquals(3, viewAExhFieldset1Children.size());
        Assert.assertEquals("string", ((ViewElement) viewAExhFieldset1Children.get(0)).getDefinition().getName());
        Assert.assertEquals("long", ((ViewElement) viewAExhFieldset1Children.get(1)).getDefinition().getName());
        Assert.assertEquals("date", ((ViewElement) viewAExhFieldset1Children.get(2)).getDefinition().getName());
        
        SimpleViewItemGroup viewABasicFieldset1 = (SimpleViewItemGroup) viewAItems.get(3);
        Assert.assertEquals("fieldset1", viewABasicFieldset1.getName());
        Assert.assertEquals("Fieldset 1", viewABasicFieldset1.getLabel().getLabel());
        List<ViewItem> viewABasicFieldset1Children = viewABasicFieldset1.getViewItems();
        Assert.assertEquals(2, viewABasicFieldset1Children.size());
        Assert.assertEquals("attribute1", ((ViewElement) viewABasicFieldset1Children.get(0)).getDefinition().getName());
        Assert.assertEquals("attribute2", ((ViewElement) viewABasicFieldset1Children.get(1)).getDefinition().getName());
        
        // Test the view B, inlcudes both parents views at root
        View viewB = contentType.getView("viewB");
        List<ViewItem> viewBItems = viewB.getViewItems();
        Assert.assertEquals(5, viewBItems.size());
        
        Assert.assertEquals("title", ((ViewElement) viewBItems.get(0)).getDefinition().getName());
        
        ModelViewItemGroup viewBIllustration = (ModelViewItemGroup) viewBItems.get(1);
        Assert.assertEquals("illustration", viewBIllustration.getDefinition().getName());
        List<ViewItem> viewBIllustrationChildren = viewBIllustration.getViewItems();
        Assert.assertEquals(1, viewBIllustrationChildren.size());
        Assert.assertEquals("alt-text", ((ViewElement) viewBIllustrationChildren.get(0)).getDefinition().getName());
        
        Assert.assertEquals("abstract", ((ViewElement) viewBItems.get(2)).getDefinition().getName());
        
        SimpleViewItemGroup viewBExhFieldset1 = (SimpleViewItemGroup) viewBItems.get(3);
        Assert.assertEquals("fieldset1", viewBExhFieldset1.getName());
        Assert.assertEquals("Fieldset 1", viewBExhFieldset1.getLabel().getLabel());
        List<ViewItem> viewBExhFieldset1Children = viewBExhFieldset1.getViewItems();
        Assert.assertEquals(3, viewBExhFieldset1Children.size());
        Assert.assertEquals("string", ((ViewElement) viewBExhFieldset1Children.get(0)).getDefinition().getName());
        Assert.assertEquals("long", ((ViewElement) viewBExhFieldset1Children.get(1)).getDefinition().getName());
        Assert.assertEquals("date", ((ViewElement) viewBExhFieldset1Children.get(2)).getDefinition().getName());
        
        SimpleViewItemGroup viewBBasicFieldset1 = (SimpleViewItemGroup) viewBItems.get(4);
        Assert.assertEquals("fieldset1", viewBBasicFieldset1.getName());
        Assert.assertEquals("Fieldset 1", viewBBasicFieldset1.getLabel().getLabel());
        List<ViewItem> viewBBasicFieldset1Children = viewBBasicFieldset1.getViewItems();
        Assert.assertEquals(2, viewBBasicFieldset1Children.size());
        Assert.assertEquals("attribute1", ((ViewElement) viewBBasicFieldset1Children.get(0)).getDefinition().getName());
        Assert.assertEquals("attribute2", ((ViewElement) viewBBasicFieldset1Children.get(1)).getDefinition().getName());
        
        // Test the view C, inlcudes both parents views in a fieldset
        View viewC = contentType.getView("viewC");
        List<ViewItem> viewCItems = viewC.getViewItems();
        Assert.assertEquals(3, viewCItems.size());
        
        Assert.assertEquals("title", ((ViewElement) viewCItems.get(0)).getDefinition().getName());
        
        ModelViewItemGroup viewCIllustration = (ModelViewItemGroup) viewCItems.get(1);
        Assert.assertEquals("illustration", viewCIllustration.getDefinition().getName());
        List<ViewItem> viewCIllustrationChildren = viewCIllustration.getViewItems();
        Assert.assertEquals(1, viewCIllustrationChildren.size());
        Assert.assertEquals("alt-text", ((ViewElement) viewCIllustrationChildren.get(0)).getDefinition().getName());
        
        SimpleViewItemGroup viewCFieldset1 = (SimpleViewItemGroup) viewCItems.get(2);
        Assert.assertEquals("fieldset1", viewCFieldset1.getName());
        Assert.assertEquals("Fieldset 1", viewCFieldset1.getLabel().getLabel());
        List<ViewItem> viewCFieldset1Children = viewCFieldset1.getViewItems();
        Assert.assertEquals(3, viewCFieldset1Children.size());
        
        Assert.assertEquals("abstract", ((ViewElement) viewCFieldset1Children.get(0)).getDefinition().getName());
        
        SimpleViewItemGroup viewCExhFieldset1 = (SimpleViewItemGroup) viewCFieldset1Children.get(1);
        Assert.assertEquals("fieldset1", viewCExhFieldset1.getName());
        Assert.assertEquals("Fieldset 1", viewCExhFieldset1.getLabel().getLabel());
        List<ViewItem> viewCExhFieldset1Children = viewCExhFieldset1.getViewItems();
        Assert.assertEquals(3, viewCExhFieldset1Children.size());
        Assert.assertEquals("string", ((ViewElement) viewCExhFieldset1Children.get(0)).getDefinition().getName());
        Assert.assertEquals("long", ((ViewElement) viewCExhFieldset1Children.get(1)).getDefinition().getName());
        Assert.assertEquals("date", ((ViewElement) viewCExhFieldset1Children.get(2)).getDefinition().getName());
        
        SimpleViewItemGroup viewCBasicFieldset1 = (SimpleViewItemGroup) viewCFieldset1Children.get(2);
        Assert.assertEquals("fieldset1", viewCBasicFieldset1.getName());
        Assert.assertEquals("Fieldset 1", viewCBasicFieldset1.getLabel().getLabel());
        List<ViewItem> viewCBasicFieldset1Children = viewCBasicFieldset1.getViewItems();
        Assert.assertEquals(2, viewCBasicFieldset1Children.size());
        Assert.assertEquals("attribute1", ((ViewElement) viewCBasicFieldset1Children.get(0)).getDefinition().getName());
        Assert.assertEquals("attribute2", ((ViewElement) viewCBasicFieldset1Children.get(1)).getDefinition().getName());
        
        // Test the view D, inlcudes exhaustive parent view at root
        View viewD = contentType.getView("viewD");
        List<ViewItem> viewDItems = viewD.getViewItems();
        Assert.assertEquals(4, viewDItems.size());
        
        Assert.assertEquals("title", ((ViewElement) viewDItems.get(0)).getDefinition().getName());
        
        ModelViewItemGroup viewDIllustration = (ModelViewItemGroup) viewDItems.get(1);
        Assert.assertEquals("illustration", viewDIllustration.getDefinition().getName());
        List<ViewItem> viewDIllustrationChildren = viewDIllustration.getViewItems();
        Assert.assertEquals(1, viewDIllustrationChildren.size());
        Assert.assertEquals("alt-text", ((ViewElement) viewDIllustrationChildren.get(0)).getDefinition().getName());
        
        Assert.assertEquals("abstract", ((ViewElement) viewDItems.get(2)).getDefinition().getName());
        
        SimpleViewItemGroup viewDExhFieldset1 = (SimpleViewItemGroup) viewDItems.get(3);
        Assert.assertEquals("fieldset1", viewDExhFieldset1.getName());
        Assert.assertEquals("Fieldset 1", viewDExhFieldset1.getLabel().getLabel());
        List<ViewItem> viewDExhFieldset1Children = viewDExhFieldset1.getViewItems();
        Assert.assertEquals(3, viewDExhFieldset1Children.size());
        Assert.assertEquals("string", ((ViewElement) viewDExhFieldset1Children.get(0)).getDefinition().getName());
        Assert.assertEquals("long", ((ViewElement) viewDExhFieldset1Children.get(1)).getDefinition().getName());
        Assert.assertEquals("date", ((ViewElement) viewDExhFieldset1Children.get(2)).getDefinition().getName());
        
        // Test the view E, inlcudes exhaustive parent view in a fieldset
        View viewE = contentType.getView("viewE");
        List<ViewItem> viewEItems = viewE.getViewItems();
        Assert.assertEquals(3, viewEItems.size());
        
        Assert.assertEquals("title", ((ViewElement) viewEItems.get(0)).getDefinition().getName());
        
        ModelViewItemGroup viewEIllustration = (ModelViewItemGroup) viewEItems.get(1);
        Assert.assertEquals("illustration", viewEIllustration.getDefinition().getName());
        List<ViewItem> viewEIllustrationChildren = viewEIllustration.getViewItems();
        Assert.assertEquals(1, viewEIllustrationChildren.size());
        Assert.assertEquals("alt-text", ((ViewElement) viewEIllustrationChildren.get(0)).getDefinition().getName());
        
        SimpleViewItemGroup viewEFieldset1 = (SimpleViewItemGroup) viewEItems.get(2);
        Assert.assertEquals("fieldset1", viewEFieldset1.getName());
        Assert.assertEquals("Fieldset 1", viewEFieldset1.getLabel().getLabel());
        List<ViewItem> viewEFieldset1Children = viewEFieldset1.getViewItems();
        Assert.assertEquals(2, viewEFieldset1Children.size());
        
        Assert.assertEquals("abstract", ((ViewElement) viewEFieldset1Children.get(0)).getDefinition().getName());
        
        SimpleViewItemGroup viewEExhFieldset1 = (SimpleViewItemGroup) viewEFieldset1Children.get(1);
        Assert.assertEquals("fieldset1", viewEExhFieldset1.getName());
        Assert.assertEquals("Fieldset 1", viewEExhFieldset1.getLabel().getLabel());
        List<ViewItem> viewEExhFieldset1Children = viewEExhFieldset1.getViewItems();
        Assert.assertEquals(3, viewEExhFieldset1Children.size());
        Assert.assertEquals("string", ((ViewElement) viewEExhFieldset1Children.get(0)).getDefinition().getName());
        Assert.assertEquals("long", ((ViewElement) viewEExhFieldset1Children.get(1)).getDefinition().getName());
        Assert.assertEquals("date", ((ViewElement) viewEExhFieldset1Children.get(2)).getDefinition().getName());
        
        // Test the view details, that includes a view from the parent of the parent
        View viewDetails = contentType.getView("details");
        List<ViewItem> viewDetailsItems = viewDetails.getViewItems();
        Assert.assertEquals(2, viewDetailsItems.size());
        Assert.assertEquals("title", ((ViewElement) viewDetailsItems.get(0)).getDefinition().getName());
        Assert.assertEquals("abstract", ((ViewElement) viewDetailsItems.get(1)).getDefinition().getName());
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with 2 different items with the same name in a view, at root and in a composite
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testConfigureViewWithSameItemInComposite() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.extends.exhaustive");
        
        // Test the view A, inherited from parents
        View view = contentType.getView("same-item-in-composite");
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(2, viewItems.size());
        
        Assert.assertEquals("address", viewItems.get(0).getName());
        List<ViewItem> addressViewItems = ((ModelViewItemGroup) viewItems.get(0)).getViewItems();
        Assert.assertEquals(1, addressViewItems.size());
        Assert.assertEquals("gps", addressViewItems.get(0).getName());
        
        Assert.assertEquals("gps", viewItems.get(1).getName());
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type's view with a simplified access to items
     * (ex: composite/repeater/linkContent/title)
     * @throws Exception if an error occurs
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testConfigureViewWithSimplifiedAccessToItems() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.extends.exhaustive");
        
        // Test the view with simplified access to items
        View view = contentType.getView("with-simplified-access-to-item");
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(11, viewItems.size());
        
        // View - owns
        ViewElementAccessor owns1 = (ViewElementAccessor) viewItems.get(0);
        Assert.assertEquals("owns", owns1.getDefinition().getName());
        Assert.assertEquals(view, owns1.getParent());
        Assert.assertTrue(owns1.getViewItems().isEmpty());
        
        // View - owns
        ViewElementAccessor owns2 = (ViewElementAccessor) viewItems.get(1);
        Assert.assertEquals("owns", owns2.getDefinition().getName());
        Assert.assertEquals(view, owns2.getParent());
        List<ViewItem> owns2Items = owns2.getViewItems();
        Assert.assertEquals(1, owns2Items.size());
        
        // View - owns - string
        ViewElement ownsString = (ViewElement) owns2Items.get(0);
        Assert.assertEquals("string", ownsString.getDefinition().getName());
        Assert.assertEquals(owns2, ownsString.getParent());
        
        // View - string
        ViewElement string = (ViewElement) viewItems.get(2);
        Assert.assertEquals("string", string.getDefinition().getName());
        Assert.assertEquals(view, string.getParent());
        
        // View - owns
        ViewElementAccessor owns3 = (ViewElementAccessor) viewItems.get(3);
        Assert.assertEquals("owns", owns3.getDefinition().getName());
        Assert.assertEquals(view, owns3.getParent());
        List<ViewItem> owns3Items = owns3.getViewItems();
        Assert.assertEquals(1, owns3Items.size());
        
        // View - owns - linkContent
        ViewElementAccessor owns3LinkContent = (ViewElementAccessor) owns3Items.get(0);
        Assert.assertEquals("linkContent", owns3LinkContent.getDefinition().getName());
        Assert.assertEquals(owns3, owns3LinkContent.getParent());
        List<ViewItem> owns3LinkContentItems = owns3LinkContent.getViewItems();
        Assert.assertEquals(1, owns3LinkContentItems.size());
        
        // View - owns - linkContent - title
        ViewElement owns3LinkContentTitle = (ViewElement) owns3LinkContentItems.get(0);
        Assert.assertEquals("title", owns3LinkContentTitle.getDefinition().getName());
        Assert.assertEquals(owns3LinkContent, owns3LinkContentTitle.getParent());
        
        // View - owns
        ViewElementAccessor owns4 = (ViewElementAccessor) viewItems.get(4);
        Assert.assertEquals("owns", owns4.getDefinition().getName());
        Assert.assertEquals(view, owns4.getParent());
        List<ViewItem> owns4Items = owns4.getViewItems();
        Assert.assertEquals(1, owns4Items.size());
        
        // View - owns - linkContent
        ViewElementAccessor owns4LinkContent = (ViewElementAccessor) owns4Items.get(0);
        Assert.assertEquals("linkContent", owns4LinkContent.getDefinition().getName());
        Assert.assertEquals(owns4, owns4LinkContent.getParent());
        List<ViewItem> owns4LinkContentItems = owns4LinkContent.getViewItems();
        Assert.assertEquals(1, owns4LinkContentItems.size());
        
        // View - owns - linkContent - title
        ViewElement owns4LinkContentTitle = (ViewElement) owns4LinkContentItems.get(0);
        Assert.assertEquals("title", owns4LinkContentTitle.getDefinition().getName());
        Assert.assertEquals(owns4LinkContent, owns4LinkContentTitle.getParent());
        
        // View - composite
        ModelViewItemGroup composite1 = (ModelViewItemGroup) viewItems.get(5);
        Assert.assertEquals("composite", composite1.getDefinition().getName());
        Assert.assertEquals(view, composite1.getParent());
        List<ViewItem> composite1Children = composite1.getViewItems();
        Assert.assertEquals(1, composite1Children.size());
     
        // View - composite - repeater
        ModelViewItemGroup composite1Repeater = (ModelViewItemGroup) composite1Children.get(0);
        Assert.assertEquals("repeater", composite1Repeater.getDefinition().getName());
        Assert.assertEquals(composite1, composite1Repeater.getParent());
        Assert.assertTrue(composite1Repeater.getViewItems().isEmpty());
        
        // View - composite
        ModelViewItemGroup composite2 = (ModelViewItemGroup) viewItems.get(6);
        Assert.assertEquals("composite", composite2.getDefinition().getName());
        Assert.assertEquals(view, composite2.getParent());
        List<ViewItem> composite2Children = composite2.getViewItems();
        Assert.assertEquals(1, composite2Children.size());
     
        // View - composite - repeater
        ModelViewItemGroup composite2Repeater = (ModelViewItemGroup) composite2Children.get(0);
        Assert.assertEquals("repeater", composite2Repeater.getDefinition().getName());
        Assert.assertEquals(composite2, composite2Repeater.getParent());
        List<ViewItem> composite2RepeaterChildren = composite2Repeater.getViewItems();
        Assert.assertEquals(1, composite2RepeaterChildren.size());
        
        // View - composite - repeater - subComposite
        ModelViewItemGroup composite2RepeaterSubComposite = (ModelViewItemGroup) composite2RepeaterChildren.get(0);
        Assert.assertEquals("subcomposite", composite2RepeaterSubComposite.getDefinition().getName());
        Assert.assertEquals(composite2Repeater, composite2RepeaterSubComposite.getParent());
        List<ViewItem> composite2RepeaterSubCompositeChildren = composite2RepeaterSubComposite.getViewItems();
        Assert.assertEquals(1, composite2RepeaterSubCompositeChildren.size());
        
        // View - composite - repeater - subComposite - subcomp-str
        ViewElement composite2RepeaterSubCompositeSubCompStr = (ViewElement) composite2RepeaterSubCompositeChildren.get(0);
        Assert.assertEquals("subcomp-str", composite2RepeaterSubCompositeSubCompStr.getDefinition().getName());
        Assert.assertEquals(composite2RepeaterSubComposite, composite2RepeaterSubCompositeSubCompStr.getParent());
        
        // View - composite
        ModelViewItemGroup composite3 = (ModelViewItemGroup) viewItems.get(7);
        Assert.assertEquals("composite", composite3.getDefinition().getName());
        Assert.assertEquals(view, composite3.getParent());
        List<ViewItem> composite3Children = composite3.getViewItems();
        Assert.assertEquals(1, composite3Children.size());
     
        // View - composite - repeater
        ModelViewItemGroup composite3Repeater = (ModelViewItemGroup) composite3Children.get(0);
        Assert.assertEquals("repeater", composite3Repeater.getDefinition().getName());
        Assert.assertEquals(composite3, composite3Repeater.getParent());
        List<ViewItem> composite3RepeaterChildren = composite3Repeater.getViewItems();
        Assert.assertEquals(1, composite3RepeaterChildren.size());
        
        // View - composite - repeater - subComposite
        ModelViewItemGroup composite3RepeaterSubComposite = (ModelViewItemGroup) composite3RepeaterChildren.get(0);
        Assert.assertEquals("subcomposite", composite3RepeaterSubComposite.getDefinition().getName());
        Assert.assertEquals(composite3Repeater, composite3RepeaterSubComposite.getParent());
        List<ViewItem> composite3RepeaterSubCompositeChildren = composite3RepeaterSubComposite.getViewItems();
        Assert.assertEquals(1, composite3RepeaterSubCompositeChildren.size());
        
        // View - composite - repeater - subComposite - subcomp-long
        ViewElement composite3RepeaterSubCompositesubCompLong = (ViewElement) composite3RepeaterSubCompositeChildren.get(0);
        Assert.assertEquals("subcomp-long", composite3RepeaterSubCompositesubCompLong.getDefinition().getName());
        Assert.assertEquals(composite3RepeaterSubComposite, composite3RepeaterSubCompositesubCompLong.getParent());
        
        // View - composite
        ModelViewItemGroup composite4 = (ModelViewItemGroup) viewItems.get(8);
        Assert.assertEquals("composite", composite4.getDefinition().getName());
        Assert.assertEquals(view, composite4.getParent());
        List<ViewItem> composite4Children = composite4.getViewItems();
        Assert.assertEquals(1, composite4Children.size());
     
        // View - composite - repeater
        ModelViewItemGroup composite4Repeater = (ModelViewItemGroup) composite4Children.get(0);
        Assert.assertEquals("repeater", composite4Repeater.getDefinition().getName());
        Assert.assertEquals(composite4, composite4Repeater.getParent());
        List<ViewItem> composite4RepeaterChildren = composite4Repeater.getViewItems();
        Assert.assertEquals(1, composite4RepeaterChildren.size());
        
        // View - composite - repeater - subComposite
        ModelViewItemGroup composite4RepeaterSubComposite = (ModelViewItemGroup) composite4RepeaterChildren.get(0);
        Assert.assertEquals("subcomposite", composite4RepeaterSubComposite.getDefinition().getName());
        Assert.assertEquals(composite4Repeater, composite4RepeaterSubComposite.getParent());
        List<ViewItem> composite4RepeaterSubCompositeChildren = composite4RepeaterSubComposite.getViewItems();
        Assert.assertEquals(2, composite4RepeaterSubCompositeChildren.size());
        
        // View - composite - repeater - subComposite - subcomp-str
        ViewElement composite4RepeaterSubCompositesubCompStr = (ViewElement) composite4RepeaterSubCompositeChildren.get(0);
        Assert.assertEquals("subcomp-str", composite4RepeaterSubCompositesubCompStr.getDefinition().getName());
        Assert.assertEquals(composite4RepeaterSubComposite, composite4RepeaterSubCompositesubCompStr.getParent());
        
        // View - composite - repeater - subComposite - subcomp-long
        ViewElement composite4RepeaterSubCompositesubCompLong = (ViewElement) composite4RepeaterSubCompositeChildren.get(1);
        Assert.assertEquals("subcomp-long", composite4RepeaterSubCompositesubCompLong.getDefinition().getName());
        Assert.assertEquals(composite4RepeaterSubComposite, composite4RepeaterSubCompositesubCompLong.getParent());
        
        // View - composite
        ModelViewItemGroup composite5 = (ModelViewItemGroup) viewItems.get(9);
        Assert.assertEquals("composite", composite5.getDefinition().getName());
        Assert.assertEquals(view, composite5.getParent());
        Assert.assertTrue(composite5.getViewItems().isEmpty());
        
        // View - repeaterThreeLevels
        ModelViewItemGroup threeLevelsRepeater = (ModelViewItemGroup) viewItems.get(10);
        Assert.assertEquals("repeaterThreeLevels", threeLevelsRepeater.getDefinition().getName());
        Assert.assertEquals(view, threeLevelsRepeater.getParent());
        List<ViewItem> threeLevelsRepeaterChildren = threeLevelsRepeater.getViewItems();
        Assert.assertEquals(1, threeLevelsRepeaterChildren.size());
        
        // View - repeaterThreeLevels - repeater2
        ModelViewItemGroup repeater2 = (ModelViewItemGroup) threeLevelsRepeaterChildren.get(0);
        Assert.assertEquals("repeater2", repeater2.getDefinition().getName());
        Assert.assertEquals(threeLevelsRepeater, repeater2.getParent());
        List<ViewItem> repeater2Children = repeater2.getViewItems();
        Assert.assertEquals(1, repeater2Children.size());
        
        // View - repeaterThreeLevels - repeater2 - repeater3
        ModelViewItemGroup repeater3 = (ModelViewItemGroup) repeater2Children.get(0);
        Assert.assertEquals("repeater3", repeater3.getDefinition().getName());
        Assert.assertEquals(repeater2, repeater3.getParent());
        List<ViewItem> repeater3Children = repeater3.getViewItems();
        Assert.assertEquals(1, repeater3Children.size());
        
        // View - repeaterThreeLevels - repeater2 - repeater3 - rep3-string
        ViewElement rep3String = (ViewElement) repeater3Children.get(0);
        Assert.assertEquals("rep3-string", rep3String.getDefinition().getName());
        Assert.assertEquals(repeater3, rep3String.getParent());
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type's view with a simplified access to linked contents' views
     * (ex: composite/repeater/linkContent[details])
     * @throws Exception if an error occurs
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testConfigureViewWithSimplifiedAccessToViews() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.extends.exhaustive");
        
        // Test the view with simplified access to views
        View view = contentType.getView("with-simplified-access-to-view");
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(3, viewItems.size());
        
        // View - sport
        ViewElementAccessor sport = (ViewElementAccessor) viewItems.get(0);
        Assert.assertEquals("sport", sport.getDefinition().getName());
        Assert.assertEquals(view, sport.getParent());
        List<ViewItem> sportItems = sport.getViewItems();
        Assert.assertEquals(1, sportItems.size());
        
        // View - sport - title
        ViewElement sportTitle = (ViewElement) sportItems.get(0);
        Assert.assertEquals("title", sportTitle.getDefinition().getName());
        Assert.assertEquals(sport, sportTitle.getParent());
        
        // View - composite
        ModelViewItemGroup composite1 = (ModelViewItemGroup) viewItems.get(1);
        Assert.assertEquals("composite", composite1.getDefinition().getName());
        Assert.assertEquals(view, composite1.getParent());
        List<ViewItem> composite1Children = composite1.getViewItems();
        Assert.assertEquals(1, composite1Children.size());
     
        // View - composite - repeater
        ModelViewItemGroup composite1Repeater = (ModelViewItemGroup) composite1Children.get(0);
        Assert.assertEquals("repeater", composite1Repeater.getDefinition().getName());
        Assert.assertEquals(composite1, composite1Repeater.getParent());
        List<ViewItem> composite1RepeaterChildren = composite1Repeater.getViewItems();
        Assert.assertEquals(1, composite1RepeaterChildren.size());
     
        // View - composite - repeater - comp-owns
        ViewElementAccessor composite1RepeaterCompOwns = (ViewElementAccessor) composite1RepeaterChildren.get(0);
        Assert.assertEquals("comp-owns", composite1RepeaterCompOwns.getDefinition().getName());
        Assert.assertEquals(composite1Repeater, composite1RepeaterCompOwns.getParent());
        List<ViewItem> composite1RepeaterCompOwnsChildren = composite1RepeaterCompOwns.getViewItems();
        Assert.assertEquals(4, composite1RepeaterCompOwnsChildren.size());
        
        // View - composite - repeater - comp-owns - title
        ViewElement composite1RepeaterCompOwnsTitle = (ViewElement) composite1RepeaterCompOwnsChildren.get(0);
        Assert.assertEquals("title", composite1RepeaterCompOwnsTitle.getDefinition().getName());
        Assert.assertEquals(composite1RepeaterCompOwns, composite1RepeaterCompOwnsTitle.getParent());
        
        // View - composite - repeater - comp-owns - illustration
        ModelViewItemGroup composite1RepeaterCompOwnsIllustration = (ModelViewItemGroup) composite1RepeaterCompOwnsChildren.get(1);
        Assert.assertEquals("illustration", composite1RepeaterCompOwnsIllustration.getDefinition().getName());
        Assert.assertEquals(composite1RepeaterCompOwns, composite1RepeaterCompOwnsIllustration.getParent());
        List<ViewItem> composite1RepeaterCompOwnsIllustrationChildren = composite1RepeaterCompOwnsIllustration.getViewItems();
        Assert.assertEquals(2, composite1RepeaterCompOwnsIllustrationChildren.size());
        
        // View - composite - repeater - comp-owns - illustration - image
        ViewElement composite1RepeaterCompOwnsIllustrationImage = (ViewElement) composite1RepeaterCompOwnsIllustrationChildren.get(0);
        Assert.assertEquals("image", composite1RepeaterCompOwnsIllustrationImage.getDefinition().getName());
        Assert.assertEquals(composite1RepeaterCompOwnsIllustration, composite1RepeaterCompOwnsIllustrationImage.getParent());
        
        // View - composite - repeater - comp-owns - illustration - alt-text
        ViewElement composite1RepeaterCompOwnsIllustrationAltText = (ViewElement) composite1RepeaterCompOwnsIllustrationChildren.get(1);
        Assert.assertEquals("alt-text", composite1RepeaterCompOwnsIllustrationAltText.getDefinition().getName());
        Assert.assertEquals(composite1RepeaterCompOwnsIllustration, composite1RepeaterCompOwnsIllustrationAltText.getParent());
        
        // View - composite - repeater - comp-owns - abstract
        ViewElement composite1RepeaterCompOwnsAbstract = (ViewElement) composite1RepeaterCompOwnsChildren.get(2);
        Assert.assertEquals("abstract", composite1RepeaterCompOwnsAbstract.getDefinition().getName());
        Assert.assertEquals(composite1RepeaterCompOwns, composite1RepeaterCompOwnsAbstract.getParent());
        
        // View - composite - repeater - comp-owns - attachments
        ModelViewItemGroup composite1RepeaterCompOwnsAttachments = (ModelViewItemGroup) composite1RepeaterCompOwnsChildren.get(3);
        Assert.assertEquals("attachments", composite1RepeaterCompOwnsAttachments.getDefinition().getName());
        Assert.assertEquals(composite1RepeaterCompOwns, composite1RepeaterCompOwnsAttachments.getParent());
        List<ViewItem> composite1RepeaterCompOwnsAttachmentsChildren = composite1RepeaterCompOwnsAttachments.getViewItems();
        Assert.assertEquals(3, composite1RepeaterCompOwnsAttachmentsChildren.size());
        
        // View - composite - repeater - comp-owns - attachments - attachment
        ViewElement composite1RepeaterCompOwnsAttachmentsAttachment = (ViewElement) composite1RepeaterCompOwnsAttachmentsChildren.get(0);
        Assert.assertEquals("attachment", composite1RepeaterCompOwnsAttachmentsAttachment.getDefinition().getName());
        Assert.assertEquals(composite1RepeaterCompOwnsAttachments, composite1RepeaterCompOwnsAttachmentsAttachment.getParent());
        
        // View - composite - repeater - comp-owns - attachments - attachment-text
        ViewElement composite1RepeaterCompOwnsAttachmentsAttachmentText = (ViewElement) composite1RepeaterCompOwnsAttachmentsChildren.get(1);
        Assert.assertEquals("attachment-text", composite1RepeaterCompOwnsAttachmentsAttachmentText.getDefinition().getName());
        Assert.assertEquals(composite1RepeaterCompOwnsAttachments, composite1RepeaterCompOwnsAttachmentsAttachmentText.getParent());
        
        // View - composite - repeater - comp-owns - attachments - attachment-desc
        ViewElement composite1RepeaterCompOwnsAttachmentsAttachmentDesc = (ViewElement) composite1RepeaterCompOwnsAttachmentsChildren.get(2);
        Assert.assertEquals("attachment-desc", composite1RepeaterCompOwnsAttachmentsAttachmentDesc.getDefinition().getName());
        Assert.assertEquals(composite1RepeaterCompOwnsAttachments, composite1RepeaterCompOwnsAttachmentsAttachmentDesc.getParent());
        
        // View - composite
        ModelViewItemGroup composite2 = (ModelViewItemGroup) viewItems.get(2);
        Assert.assertEquals("composite", composite2.getDefinition().getName());
        Assert.assertEquals(view, composite2.getParent());
        List<ViewItem> composite2Children = composite2.getViewItems();
        Assert.assertEquals(1, composite2Children.size());
     
        // View - composite - repeater
        ModelViewItemGroup composite2Repeater = (ModelViewItemGroup) composite2Children.get(0);
        Assert.assertEquals("repeater", composite2Repeater.getDefinition().getName());
        Assert.assertEquals(composite2, composite2Repeater.getParent());
        List<ViewItem> composite2RepeaterChildren = composite2Repeater.getViewItems();
        Assert.assertEquals(1, composite2RepeaterChildren.size());
     
        // View - composite - repeater - comp-owns
        ViewElementAccessor composite2RepeaterCompOwns = (ViewElementAccessor) composite2RepeaterChildren.get(0);
        Assert.assertEquals("comp-owns", composite2RepeaterCompOwns.getDefinition().getName());
        Assert.assertEquals(composite2Repeater, composite2RepeaterCompOwns.getParent());
        List<ViewItem> composite2RepeaterCompOwnsChildren = composite2RepeaterCompOwns.getViewItems();
        Assert.assertEquals(1, composite2RepeaterCompOwnsChildren.size());
        
        // View - composite - repeater - comp-owns - string
        ViewElement composite2RepeaterCompOwnsString = (ViewElement) composite2RepeaterCompOwnsChildren.get(0);
        Assert.assertEquals("string", composite2RepeaterCompOwnsString.getDefinition().getName());
        Assert.assertEquals(composite2RepeaterCompOwns, composite2RepeaterCompOwnsString.getParent());
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type's view with a simplified access to all items of an accessor (group or linked content)
     * (ex: composite/* or composite/repeater/linkContent/*)
     * @throws Exception if an error occurs
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testConfigureViewWithSimplifiedAccessToAllItems() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.extends.exhaustive");
        
        // Test the view with simplified access to items
        View view = contentType.getView("with-simplified-access-to-all-items");
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(6, viewItems.size());
        
        // View - multilingualContent
        ViewElementAccessor multilingual = (ViewElementAccessor) viewItems.get(0);
        Assert.assertEquals("multilingualContent", multilingual.getDefinition().getName());
        Assert.assertEquals(view, multilingual.getParent());
        List<ViewItem> multilingualItems = multilingual.getViewItems();
        Assert.assertEquals(3, multilingualItems.size());
        
        // View - multilingualContent - title
        ViewElement multilingualTitle = (ViewElement) multilingualItems.get(0);
        Assert.assertEquals("title", multilingualTitle.getDefinition().getName());
        Assert.assertEquals(multilingual, multilingualTitle.getParent());
        
        // View - multilingualContent - no-multilingual
        ViewElement multilingualNoMultilingual = (ViewElement) multilingualItems.get(1);
        Assert.assertEquals("no-multilingual", multilingualNoMultilingual.getDefinition().getName());
        Assert.assertEquals(multilingual, multilingualNoMultilingual.getParent());
        
        // View - multilingualContent - multilingual
        ViewElement multilingualMultilingual = (ViewElement) multilingualItems.get(2);
        Assert.assertEquals("multilingual", multilingualMultilingual.getDefinition().getName());
        Assert.assertEquals(multilingual, multilingualMultilingual.getParent());
        
        // View - owns
        ViewElementAccessor owns = (ViewElementAccessor) viewItems.get(1);
        Assert.assertEquals("owns", owns.getDefinition().getName());
        Assert.assertEquals(view, owns.getParent());
        List<ViewItem> ownsItems = owns.getViewItems();
        Assert.assertEquals(1, ownsItems.size());
        
        // View - owns - linkContent
        ViewElementAccessor ownsLinkContent = (ViewElementAccessor) ownsItems.get(0);
        Assert.assertEquals("linkContent", ownsLinkContent.getDefinition().getName());
        Assert.assertEquals(owns, ownsLinkContent.getParent());
        List<ViewItem> ownsLinkContentItems = ownsLinkContent.getViewItems();
        Assert.assertEquals(1, ownsLinkContentItems.size());
        
        // View - owns - linkContent - title
        ViewElement ownsLinkContentTitle = (ViewElement) ownsLinkContentItems.get(0);
        Assert.assertEquals("title", ownsLinkContentTitle.getDefinition().getName());
        Assert.assertEquals(ownsLinkContent, ownsLinkContentTitle.getParent());
        
        // View - illustration
        ModelViewItemGroup illustration1 = (ModelViewItemGroup) viewItems.get(2);
        Assert.assertEquals("illustration", illustration1.getDefinition().getName());
        Assert.assertEquals(view, illustration1.getParent());
        List<ViewItem> illustration1Children = illustration1.getViewItems();
        Assert.assertEquals(2, illustration1Children.size());
        
        // View - illustration - image
        ViewElement illustration1Image = (ViewElement) illustration1Children.get(0);
        Assert.assertEquals("image", illustration1Image.getDefinition().getName());
        Assert.assertEquals(illustration1, illustration1Image.getParent());
        
        // View - illustration - alt-text
        ViewElement illustration1AltText = (ViewElement) illustration1Children.get(1);
        Assert.assertEquals("alt-text", illustration1AltText.getDefinition().getName());
        Assert.assertEquals(illustration1, illustration1AltText.getParent());
        
        // View - illustration
        ModelViewItemGroup illustration2 = (ModelViewItemGroup) viewItems.get(3);
        Assert.assertEquals("illustration", illustration2.getDefinition().getName());
        Assert.assertEquals(view, illustration2.getParent());
        List<ViewItem> illustration2Children = illustration2.getViewItems();
        Assert.assertEquals(1, illustration2Children.size());
        
        // View - illustration - image
        ViewElement illustration2Image = (ViewElement) illustration2Children.get(0);
        Assert.assertEquals("image", illustration2Image.getDefinition().getName());
        Assert.assertEquals(illustration2, illustration2Image.getParent());
        
        // View - composite
        ModelViewItemGroup composite = (ModelViewItemGroup) viewItems.get(4);
        Assert.assertEquals("composite", composite.getDefinition().getName());
        Assert.assertEquals(view, composite.getParent());
        List<ViewItem> compositeChildren = composite.getViewItems();
        Assert.assertEquals(1, compositeChildren.size());
     
        // View - composite - repeater
        ModelViewItemGroup repeater = (ModelViewItemGroup) compositeChildren.get(0);
        Assert.assertEquals("repeater", repeater.getDefinition().getName());
        Assert.assertEquals(composite, repeater.getParent());
        List<ViewItem> repeaterChildren = repeater.getViewItems();
        Assert.assertEquals(1, repeaterChildren.size());
        
        // View - composite - repeater - subComposite
        ModelViewItemGroup subComposite = (ModelViewItemGroup) repeaterChildren.get(0);
        Assert.assertEquals("subcomposite", subComposite.getDefinition().getName());
        Assert.assertEquals(repeater, subComposite.getParent());
        List<ViewItem> subCompositeChildren = subComposite.getViewItems();
        Assert.assertEquals(2, subCompositeChildren.size());
        
        // View - composite - repeater - subComposite - subcomp-str
        ViewElement subCompStr = (ViewElement) subCompositeChildren.get(0);
        Assert.assertEquals("subcomp-str", subCompStr.getDefinition().getName());
        Assert.assertEquals(subComposite, subCompStr.getParent());
        
        // View - composite - repeater - subComposite - subcomp-long
        ViewElement subCompLong = (ViewElement) subCompositeChildren.get(1);
        Assert.assertEquals("subcomp-long", subCompLong.getDefinition().getName());
        Assert.assertEquals(subComposite, subCompLong.getParent());
        
        // View - Three levels repeater
        ModelViewItemGroup threeLevelsRepeater = (ModelViewItemGroup) viewItems.get(5);
        Assert.assertEquals("repeaterThreeLevels", threeLevelsRepeater.getDefinition().getName());
        Assert.assertEquals(view, threeLevelsRepeater.getParent());
        List<ViewItem> threeLevelsRepeaterChildren = threeLevelsRepeater.getViewItems();
        Assert.assertEquals(1, threeLevelsRepeaterChildren.size());
        
        // View - Three levels repeater - repeater2
        ModelViewItemGroup repeater2 = (ModelViewItemGroup) threeLevelsRepeaterChildren.get(0);
        Assert.assertEquals("repeater2", repeater2.getDefinition().getName());
        Assert.assertEquals(threeLevelsRepeater, repeater2.getParent());
        List<ViewItem> repeater2Children = repeater2.getViewItems();
        Assert.assertEquals(1, repeater2Children.size());
        
        // View - Three levels repeater - repeater2 - repeater3
        ModelViewItemGroup repeater3 = (ModelViewItemGroup) repeater2Children.get(0);
        Assert.assertEquals("repeater3", repeater3.getDefinition().getName());
        Assert.assertEquals(repeater2, repeater3.getParent());
        List<ViewItem> repeater3Children = repeater3.getViewItems();
        Assert.assertEquals(6, repeater3Children.size());
        
        // View - Three levels repeater - repeater2 - repeater3 - rep3 string
        ViewElement rep3String = (ViewElement) repeater3Children.get(0);
        Assert.assertEquals("rep3-string", rep3String.getDefinition().getName());
        Assert.assertEquals(repeater3, rep3String.getParent());
        
        // View - Three levels repeater - repeater2 - repeater3 - rep3 long
        ViewElement rep3Long = (ViewElement) repeater3Children.get(1);
        Assert.assertEquals("rep3-long", rep3Long.getDefinition().getName());
        Assert.assertEquals(repeater3, rep3Long.getParent());
        
        // View - Three levels repeater - repeater2 - repeater3 - rep3 longmult-min
        ViewElement rep3LongmultMin = (ViewElement) repeater3Children.get(2);
        Assert.assertEquals("rep3-longmult-min", rep3LongmultMin.getDefinition().getName());
        Assert.assertEquals(repeater3, rep3LongmultMin.getParent());
        
        // View - Three levels repeater - repeater2 - repeater3 - rep3 longmult-max
        ViewElement rep3LongmultMax = (ViewElement) repeater3Children.get(3);
        Assert.assertEquals("rep3-longmult-max", rep3LongmultMax.getDefinition().getName());
        Assert.assertEquals(repeater3, rep3LongmultMax.getParent());
        
        // View - Three levels repeater - repeater2 - repeater3 - rep3 owns
        ViewElementAccessor rep3Owns = (ViewElementAccessor) repeater3Children.get(4);
        Assert.assertEquals("rep3-owns", rep3Owns.getDefinition().getName());
        Assert.assertEquals(repeater3, rep3Owns.getParent());
        
        // View - Three levels repeater - repeater2 - repeater3 - rep3 belongs to
        ViewElementAccessor rep3BelongsTo = (ViewElementAccessor) repeater3Children.get(5);
        Assert.assertEquals("rep3-belongsTo", rep3BelongsTo.getDefinition().getName());
        Assert.assertEquals(repeater3, rep3BelongsTo.getParent());
    }
    
    /**
     * Test {@link ContentType} implementation with a content type referencing an attribute in a type that does not override its parent's main view
     * 3 content types are needed for this test:
     * <ol>
     * <li>A (parent) content type with an attribute and a main view</li>
     * <li>A (child) content type, extending the first one. It does not override the main view</li>
     * <li>A (referencing) content type, referencing an attribute of the child content type in one of its views</li>
     * </ol>
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-inheritance-and-references",
    })
    public void testViewWithAttributReferenceOnTypeWithoutMainView() throws Exception
    {
        ContentType childContentType = _contentTypeEP.getExtension("tests-view-inheritance-and-references.type-a-with-main-view");
        View childMainView = childContentType.getView("main");
        List<ViewItem> childMainViewItems = childMainView.getViewItems();
        Assert.assertEquals(1, childMainViewItems.size());
        Assert.assertEquals("attribute1", childMainViewItems.get(0).getName());
        List<ViewItem> attribute1Items = ((ViewItemAccessor) childMainViewItems.get(0)).getViewItems();
        Assert.assertEquals(1, attribute1Items.size());
        Assert.assertEquals("attribute3", attribute1Items.get(0).getName());

        ContentType referencingContentType = _contentTypeEP.getExtension("tests-view-inheritance-and-references.type-b-referencing-the-child");
        View refMainView = referencingContentType.getView("main");
        List<ViewItem> refMainViewItems = refMainView.getViewItems();
        Assert.assertEquals(1, refMainViewItems.size());
        Assert.assertEquals("linkContent", refMainViewItems.get(0).getName());
        List<ViewItem> linkedContentItems = ((ViewItemAccessor) refMainViewItems.get(0)).getViewItems();
        Assert.assertEquals(1, linkedContentItems.size());
        Assert.assertEquals("attribute2", linkedContentItems.get(0).getName());
    }
    
    /**
     * Test {@link ContentType#getView(String)} method for a view with properties
     */
    @Test
    @NeedFeatures({
        "content.properties/content-type-properties-tests"
    })
    public void testViewWithProperties()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.with.properties");
        View view = contentType.getView("main");
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(2, viewItems.size());
        
        // View - Attributes tab
        SimpleViewItemGroup attributesTab = (SimpleViewItemGroup) viewItems.get(0);
        List<ViewItem> attributesTabChildren = attributesTab.getViewItems();
        Assert.assertEquals(2, attributesTabChildren.size());
        
        // View - Attributes tab - title
        ViewElement title = (ViewElement) attributesTabChildren.get(0);
        Assert.assertEquals("title", title.getDefinition().getName());
        
        // View - Attributes tab - composite
        ModelViewItemGroup composite = (ModelViewItemGroup) attributesTabChildren.get(1);
        Assert.assertEquals("composite", composite.getDefinition().getName());
        List<ViewItem> compositeChildren = composite.getViewItems();
        Assert.assertEquals(2, compositeChildren.size());
        
        // View - Attributes tab - composite - long
        ViewElement longElement = (ViewElement) compositeChildren.get(0);
        Assert.assertEquals("long", longElement.getDefinition().getName());
        
        // View - Attributes tab - composite - repeater
        ModelViewItemGroup repeater = (ModelViewItemGroup) compositeChildren.get(1);
        Assert.assertEquals("repeater", repeater.getDefinition().getName());
        List<ViewItem> repeaterChildren = repeater.getViewItems();
        Assert.assertEquals(1, repeaterChildren.size());
        
        // View - Attributes tab - composite - repeater - stringmult
        ViewElement stringmultElement = (ViewElement) repeaterChildren.get(0);
        Assert.assertEquals("stringmult", stringmultElement.getDefinition().getName());
        
        // View - All properties tab
        SimpleViewItemGroup allPropertiesTab = (SimpleViewItemGroup) viewItems.get(1);
        List<ViewItem> allPropertiesTabChildren = allPropertiesTab.getViewItems();
        Assert.assertEquals(2, allPropertiesTabChildren.size());
        
        // View - All properties tab - Properties fieldset
        SimpleViewItemGroup propertiesFieldset = (SimpleViewItemGroup) allPropertiesTabChildren.get(0);
        List<ViewItem> propertiesFieldsetChildren = propertiesFieldset.getViewItems();
        Assert.assertEquals(3, propertiesFieldsetChildren.size());

        // View - All properties tab - Properties fieldset - property-long
        ViewElement propertyLong = (ViewElement) propertiesFieldsetChildren.get(0);
        Assert.assertEquals("property-long", propertyLong.getDefinition().getName());

        // View - All properties tab - Properties fieldset - property-longmult
        ViewElement propertyLongmult = (ViewElement) propertiesFieldsetChildren.get(1);
        Assert.assertEquals("property-longmult", propertyLongmult.getDefinition().getName());

        // View - All properties tab - Properties fieldset - property-stringmult
        ViewElement propertyStringmult = (ViewElement) propertiesFieldsetChildren.get(2);
        Assert.assertEquals("property-stringmult", propertyStringmult.getDefinition().getName());
        
        // View - All properties tab - System properties fieldset
        SimpleViewItemGroup systemPropertiesFieldset = (SimpleViewItemGroup) allPropertiesTabChildren.get(1);
        List<ViewItem> systemPropertiesFieldsetChildren = systemPropertiesFieldset.getViewItems();
        Assert.assertEquals(3, systemPropertiesFieldsetChildren.size());
        
        // View - All properties tab - System properties fieldset - contributor
        ViewElement contributor = (ViewElement) systemPropertiesFieldsetChildren.get(0);
        Assert.assertEquals(DefaultContent.METADATA_CONTRIBUTOR, contributor.getDefinition().getName());
        
        // View - All properties tab - System properties fieldset - lastModified
        ViewElement lastModified = (ViewElement) systemPropertiesFieldsetChildren.get(1);
        Assert.assertEquals(DefaultContent.METADATA_MODIFIED, lastModified.getDefinition().getName());
        
        // View - All properties tab - System properties fieldset - lastValidation
        ViewElement lastValidation = (ViewElement) systemPropertiesFieldsetChildren.get(2);
        Assert.assertEquals("lastValidation", lastValidation.getDefinition().getName());
    }
    
    /**
     * Test {@link ContentType#getView(String)} method for a view with properties on a linked content
     */
    @Test
    @NeedFeatures({
        "content.properties/content-type-properties-tests"
    })
    public void testViewWithPropertiesOnLinkedContent()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.pointing.type.with.properties");
        View view = contentType.getView("main");
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(2, viewItems.size());
        
        // View - title
        ViewElement title = (ViewElement) viewItems.get(0);
        Assert.assertEquals("title", title.getDefinition().getName());
        
        // View - composite
        ModelViewItemGroup composite = (ModelViewItemGroup) viewItems.get(1);
        Assert.assertEquals("composite", composite.getDefinition().getName());
        List<ViewItem> compositeChildren = composite.getViewItems();
        Assert.assertEquals(1, compositeChildren.size());
        
        // View - composite - linked content
        ViewElementAccessor linkContent = (ViewElementAccessor) compositeChildren.get(0);
        Assert.assertEquals("linkContentWithProperties", linkContent.getDefinition().getName());
        List<ViewItem> linkContentItems = linkContent.getViewItems();
        Assert.assertEquals(2, linkContentItems.size());
        
        // View - composite - linked content - property-long
        ViewElement propertyLong = (ViewElement) linkContentItems.get(0);
        Assert.assertEquals("property-long", propertyLong.getDefinition().getName());
        
        // View - composite - linked content - contributor
        ViewElement contributor = (ViewElement) linkContentItems.get(1);
        Assert.assertEquals(DefaultContent.METADATA_CONTRIBUTOR, contributor.getDefinition().getName());
    }
    
    /**
     * Test {@link ViewHelper#getTruncatedView(View)} method for a view with properties
     */
    @Test
    @NeedFeatures({
        "content.properties/content-type-properties-tests"
    })
    public void testTruncateViewWithProperties()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.with.properties");
        View mainView = contentType.getView("main");
        View truncatedView = ViewHelper.getTruncatedView(mainView);
        List<ViewItem> viewItems = truncatedView.getViewItems();
        Assert.assertEquals(2, viewItems.size());
        
        // View - Attributes tab
        SimpleViewItemGroup attributesTab = (SimpleViewItemGroup) viewItems.get(0);
        List<ViewItem> attributesTabChildren = attributesTab.getViewItems();
        Assert.assertEquals(2, attributesTabChildren.size());
        
        // View - Attributes tab - title
        ViewElement title = (ViewElement) attributesTabChildren.get(0);
        Assert.assertEquals("title", title.getDefinition().getName());
        
        // View - Attributes tab - composite
        ModelViewItemGroup composite = (ModelViewItemGroup) attributesTabChildren.get(1);
        Assert.assertEquals("composite", composite.getDefinition().getName());
        List<ViewItem> compositeChildren = composite.getViewItems();
        Assert.assertEquals(2, compositeChildren.size());
        
        // View - Attributes tab - composite - long
        ViewElement longElement = (ViewElement) compositeChildren.get(0);
        Assert.assertEquals("long", longElement.getDefinition().getName());
        
        // View - Attributes tab - composite - repeater
        ModelViewItemGroup repeater = (ModelViewItemGroup) compositeChildren.get(1);
        Assert.assertEquals("repeater", repeater.getDefinition().getName());
        List<ViewItem> repeaterChildren = repeater.getViewItems();
        Assert.assertEquals(1, repeaterChildren.size());
        
        // View - Attributes tab - composite - repeater - stringmult
        ViewElement stringmultElement = (ViewElement) repeaterChildren.get(0);
        Assert.assertEquals("stringmult", stringmultElement.getDefinition().getName());
        
        // View - All properties tab
        SimpleViewItemGroup allPropertiesTab = (SimpleViewItemGroup) viewItems.get(1);
        List<ViewItem> allPropertiesTabChildren = allPropertiesTab.getViewItems();
        Assert.assertEquals(2, allPropertiesTabChildren.size());
        
        // View - All properties tab - Properties fieldset
        SimpleViewItemGroup propertiesFieldset = (SimpleViewItemGroup) allPropertiesTabChildren.get(0);
        List<ViewItem> propertiesFieldsetChildren = propertiesFieldset.getViewItems();
        Assert.assertTrue(propertiesFieldsetChildren.isEmpty());

        // View - All properties tab - System properties fieldset
        SimpleViewItemGroup systemPropertiesFieldset = (SimpleViewItemGroup) allPropertiesTabChildren.get(1);
        List<ViewItem> systemPropertiesFieldsetChildren = systemPropertiesFieldset.getViewItems();
        Assert.assertTrue(systemPropertiesFieldsetChildren.isEmpty());
    }
    
    /**
     * Test {@link ContentTypeExpression}.
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures({
        "test/content-type",
        "test/content-type-extended",
        "test/content-type-extended-2",
    })
    public void testContentTypeExpression() throws Exception
    {
        // Simple expressions.
        ContentTypeExpression expr = new ContentTypeExpression(Operator.EQ, "org.ametys.test.Content.article");
        Assert.assertEquals("@ametys-internal:contentType = 'org.ametys.test.Content.article'", expr.build());
        
        expr = new ContentTypeExpression(Operator.NE, "org.ametys.test.Content.article");
        Assert.assertEquals("@ametys-internal:contentType != 'org.ametys.test.Content.article'", expr.build());
        
        // Multiple expressions.
        String[] contentTypes = new String[] {"article", "news", "faq"};
        
        expr = new ContentTypeExpression(Operator.EQ, contentTypes);
        Assert.assertEquals("(@ametys-internal:contentType = 'article' or @ametys-internal:contentType = 'news' or @ametys-internal:contentType = 'faq')", expr.build());
        
        expr = new ContentTypeExpression(Operator.NE, contentTypes);
        Assert.assertEquals("not(@ametys-internal:contentType = 'article' or @ametys-internal:contentType = 'news' or @ametys-internal:contentType = 'faq')", expr.build());
        
        // Hierarchical expression.
        expr = _contentTypeEP.createHierarchicalCTExpression("org.ametys.cms.contenttype.Test");
        String xpath = StringUtils.replaceChars(expr.build(), "()", "");
        List<String> parts = Arrays.asList(StringUtils.splitByWholeSeparator(xpath, " or "));
        Assert.assertTrue(parts.contains("@ametys-internal:contentType = 'org.ametys.cms.contenttype.Test'"));
        Assert.assertTrue(parts.contains("@ametys-internal:contentType = 'org.ametys.cms.contenttype.Test.extended'"));
        Assert.assertTrue(parts.contains("@ametys-internal:contentType = 'org.ametys.cms.contenttype.Test.extended2'"));
        
        // Invalid expressions.
        try
        {
            expr = new ContentTypeExpression(Operator.LT, "org.ametys.test.Content.article");
            Assert.fail("Operator.LT is not a valid operator for ContentTypeExpression");
        }
        catch (AmetysRepositoryException e)
        {
            // Expected
        }
        
        try
        {
            expr = new ContentTypeExpression(Operator.LE, "org.ametys.test.Content.article");
            Assert.fail("Operator.LE is not a valid operator for ContentTypeExpression");
        }
        catch (AmetysRepositoryException e)
        {
            // Expected
        }
        
        try
        {
            expr = new ContentTypeExpression(Operator.GT, "org.ametys.test.Content.article");
            Assert.fail("Operator.GT is not a valid operator for ContentTypeExpression");
        }
        catch (AmetysRepositoryException e)
        {
            // Expected
        }
        
        try
        {
            expr = new ContentTypeExpression(Operator.GE, "org.ametys.test.Content.article");
            Assert.fail("Operator.GE is not a valid operator for ContentTypeExpression");
        }
        catch (AmetysRepositoryException e)
        {
            // Expected
        }
    }
    
    /**
     * Test {@link DefaultContentType} implementation for a simple content type
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures({
        "test/content-type-2",
    })
    public void testIsSimple() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.simple");
        Assert.assertTrue("The content type 'tests.simple' should be simple", contentType.isSimple());
    }
    
    /**
     * Test {@link DefaultContentType} implementation for a not simple content type
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures({
        "test/content-type-2",
    })
    public void testIsNotSimple() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.not.simple");
        Assert.assertFalse("The content type 'tests.not.simple' should not be simple", contentType.isSimple());
    }
    
    /**
     * Test {@link DefaultContentType} implementation for a multilingual content type
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures({
        "test/content-type-2",
    })
    public void testIsMultilingual() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.multilingual");
        Assert.assertTrue("The content type 'tests.multilingual' should be multilingual", contentType.isMultilingual());
    }
    
    /**
     * Test {@link DefaultContentType} implementation for a not multilingual content type
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures({
        "test/content-type-2",
    })
    public void testIsNotMultilingual() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.not.multilingual");
        Assert.assertFalse("The content type 'tests.not.multilingual' should not be multilingual", contentType.isMultilingual());
    }
    
    /**
     * Test {@link DefaultContentType} implementation for attribute inheritance
     * @throws Exception if an error occurs.
     */
    @SuppressWarnings("unchecked")
    @Test
    @NeedFeatures({
        "test/content-type-2",
    })
    public void testAttributesInheritance() throws Exception
    {
        ContentType parentContentType = _contentTypeEP.getExtension("tests.parent");
        ContentType childContentType = _contentTypeEP.getExtension("tests.child");

        // Parent - title
        ModelItem parentTitle = parentContentType.getModelItem("title");
        Assert.assertNotNull("The content type 'tests.parent' should have an attribute title", parentTitle);
        Assert.assertEquals("title", parentTitle.getName());

        // Child - title
        ModelItem childTitle = childContentType.getModelItem("title");
        Assert.assertNotNull("The content type 'tests.child' should have an attribute title inherited from its parent", childTitle);
        Assert.assertEquals("title", childTitle.getName());

        // Parent - attribute1
        ElementDefinition<String> parentAttribute1 = (ElementDefinition<String>) parentContentType.getModelItem("attribute1");
        Assert.assertNotNull("The content type 'tests.parent' should have an attribute attribute1", parentAttribute1);
        Assert.assertEquals("attribute1", parentAttribute1.getName());
        Assert.assertEquals("Attribute 1", parentAttribute1.getLabel().getLabel());
        Assert.assertEquals("edition.textarea", parentAttribute1.getWidget());
        Map<String, I18nizableText> parentAttribute1WidgetParams = parentAttribute1.getWidgetParameters();
        Assert.assertEquals(1, parentAttribute1WidgetParams.size());
        Assert.assertEquals("attribute 1 parent default value", parentAttribute1.getDefaultValue());
        
        // Child - attribute1
        ElementDefinition<String> childAttribute1 = (ElementDefinition<String>) childContentType.getModelItem("attribute1");
        Assert.assertNotNull("The content type 'tests.child' should have an attribute attribute1", childAttribute1);
        Assert.assertEquals("attribute1", childAttribute1.getName());
        Assert.assertEquals("Attribute 1 redefined", childAttribute1.getLabel().getLabel());
        Assert.assertEquals("edition.textarea", childAttribute1.getWidget());
        Map<String, I18nizableText> childAttribute1WidgetParams = childAttribute1.getWidgetParameters();
        Assert.assertEquals(2, childAttribute1WidgetParams.size());
        Assert.assertEquals("attribute 1 child default value", childAttribute1.getDefaultValue());
        
        // Child - attribute2
        ModelItem childAttribute2 = childContentType.getModelItem("attribute2");
        Assert.assertNotNull("The content type 'tests.child' should have an attribute attribute2", childAttribute2);
        Assert.assertEquals("attribute2", childAttribute2.getName());
        
        // Parent - linkContent
        ContentAttributeDefinition parentLinkContent = (ContentAttributeDefinition) parentContentType.getModelItem("linkContent");
        Assert.assertNotNull("The content type 'tests.parent' should have an attribute linkContent", parentLinkContent);
        Assert.assertNull("The attribute linkContent in content type 'tests.parent' should not specify a content type", parentLinkContent.getContentTypeId());
        Assert.assertNull("The attribute linkContent in content type 'tests.parent' should not specify an invert relation", parentLinkContent.getInvertRelationPath());
        
        // Child - linkContent
        ContentAttributeDefinition childLinkContent = (ContentAttributeDefinition) childContentType.getModelItem("linkContent");
        Assert.assertNotNull("The content type 'tests.child' should have an attribute linkContent", childLinkContent);
        Assert.assertEquals("tests.invert", childLinkContent.getContentTypeId());
        Assert.assertEquals("linkWithInvert", childLinkContent.getInvertRelationPath());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with attributes added by extension
     * @throws Exception if an error occurs
     */
    @Test
    @NeedFeatures({
        "test/content-type-overridden",
        "test/content-type-override-add-some-attributes",
    })
    public void testOverrideByExtensionAddSomeAttributes() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("overriddenType");
        
        Assert.assertEquals(11, contentType.getModelItems().size());
        
        // Check if the new model item has been created
        AttributeDefinition newAttribute = (AttributeDefinition) contentType.getModelItem("newAttribute");
        Assert.assertNotNull("The content type 'overriddenType' should have an attribute newAttribute", newAttribute);
        Assert.assertEquals(new I18nizableText("new attribute added"), newAttribute.getLabel());
        Assert.assertEquals(new I18nizableText("new attribute added desc"), newAttribute.getDescription());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with attributes added by extension
     * @throws Exception if an error occurs
     */
    @Test
    @NeedFeatures({
        "test/content-type-overridden",
        "test/content-type-override-add-attribute-to-several-types",
    })
    public void testOverrideByExtensionAddAttributeToSeveralTypes() throws Exception
    {
        ContentType overriddenContentType1 = _contentTypeEP.getExtension("overriddenType");
        ContentType overriddenContentType2 = _contentTypeEP.getExtension("overriddenType2");
        
        // Check if the new model item has been added to both content types
        Assert.assertTrue(overriddenContentType1.hasModelItem("newAttributeForSeveralType"));
        Assert.assertTrue(overriddenContentType2.hasModelItem("newAttributeForSeveralType"));
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with attributes modified by extension
     * @throws Exception if an error occurs
     */
    @Test
    @NeedFeatures({
        "test/content-type-overridden",
        "test/content-type-override-modify-attribute",
    })
    public void testOverrideByExtensionModifyAttribute() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("overriddenType");
        
        Assert.assertEquals(11, contentType.getModelItems().size());
        
        // Check if the model item "title" has been modified
        AttributeDefinition attributeTitle = (AttributeDefinition) contentType.getModelItem("title");
        Assert.assertNotNull("The content type 'overriddenType' should have an attribute title", attributeTitle);
        Assert.assertEquals(new I18nizableText("title modified"), attributeTitle.getLabel());
        Assert.assertEquals(new I18nizableText("title desc modified"), attributeTitle.getDescription());

        // Check if the model item "attribute" has been modified
        AttributeDefinition attribute = (AttributeDefinition) contentType.getModelItem("attribute");
        Assert.assertNotNull("The content type 'overriddenType' should have an attribute 'attribute'", attribute);
        Assert.assertEquals(new I18nizableText("attribute modified"), attribute.getLabel());
        Assert.assertEquals(new I18nizableText("attribute desc modified"), attribute.getDescription());
        Assert.assertNull(attribute.getValidator());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with views added by extension
     * @throws Exception if an error occurs
     */
    @Test
    @NeedFeatures({
        "test/content-type-overridden",
        "test/content-type-override-add-some-views",
    })
    public void testOverrideByExtensionAddSomeViews() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("overriddenType");
        
        Assert.assertEquals(10, contentType.getViewNames().size());
        
        // Check if the first new view has been created
        View view1 = contentType.getView("view1");
        Assert.assertNotNull("The content type 'overriddenType' should have a view view1", view1);
        
        List<ViewItem> view1Items = view1.getViewItems();
        Assert.assertEquals(2, view1Items.size());
        
        ViewItem title = view1Items.get(0);
        Assert.assertEquals("title", title.getName());

        ViewItem attribute = view1Items.get(1);
        Assert.assertEquals("attribute", attribute.getName());
        
        // Check if the second new view has been created
        View view2 = contentType.getView("view2");
        Assert.assertNotNull("The content type 'overriddenType' should have a view view2", view2);
        
        view1Items = view2.getViewItems();
        Assert.assertEquals(1, view1Items.size());
        
        title = view1Items.get(0);
        Assert.assertEquals("title", title.getName());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with views modified by extension
     * @throws Exception if an error occurs
     */
    @Test
    @NeedFeatures({
        "test/content-type-overridden",
        "test/content-type-override-modify-views",
    })
    public void testOverrideByExtensionModifyViews() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("overriddenType");

        // Check the view that has attributes added to it
        View viewAddAttribute = contentType.getView("viewAddAttribute");
        Assert.assertNotNull("The content type 'overriddenType' should have a view viewAddAttribute", viewAddAttribute);
        
        List<ViewItem> view1Items = viewAddAttribute.getViewItems();
        Assert.assertEquals(2, view1Items.size());
        
        ViewItem title = view1Items.get(0);
        Assert.assertEquals("title", title.getName());
        
        ViewItem attribute = view1Items.get(1);
        Assert.assertEquals("attribute", attribute.getName());
        
        // Check the view that has attributes deleted from it
        View viewDeleteAttribute = contentType.getView("viewDeleteAttribute");
        Assert.assertNotNull("The content type 'overriddenType' should have a view viewDeleteAttribute", viewDeleteAttribute);
        
        view1Items = viewDeleteAttribute.getViewItems();
        Assert.assertEquals(1, view1Items.size());
        
        title = view1Items.get(0);
        Assert.assertEquals("title", title.getName());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with global validators added by extension
     * @throws Exception if an error occurs
     */
    @Test
    @NeedFeatures({
        "test/content-type-overridden",
        "test/content-type-override-modify-global-validators",
    })
    public void testOverrideByExtensionModifyGlobalValidators() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("overriddenType");
        
        List<ContentValidator> globalValidators = contentType.getGlobalValidators();
        Assert.assertEquals(3, globalValidators.size());
        
        // global validator from the default configuration
        Assert.assertEquals("org.ametys.cms.contenttype.validation.DateTimeIntervalValidator", globalValidators.get(0).getClass().getName());
        
        // global validators from the override extensions
        Assert.assertEquals("org.ametys.cms.contenttype.validation.LongIntervalValidator", globalValidators.get(1).getClass().getName());
        Assert.assertEquals("org.ametys.cms.contenttype.validation.DoubleIntervalValidator", globalValidators.get(2).getClass().getName());
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with some new elements by extension.
     * The new elements' labels are declared in a plugin different that the one declaring the content type itself
     * @throws Exception if an error occurs
     */
    @Test
    @NeedFeatures({
        "contenttype.override.label/content-type-override-label-from-other-catalog"
    })
    public void testOverrideByExtensionGetLabelFromTheOverriddenPlugin() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("org.ametys.test.Content.exhaustive");
        
        ModelItem modelItem = contentType.getModelItem("attributeToAdd");
        Assert.assertEquals("New attribute", _i18nUtils.translate(modelItem.getLabel(), "en"));
        Assert.assertEquals("Attribute added by override", _i18nUtils.translate(modelItem.getDescription(), "en"));
        
        View view = contentType.getView("viewFromOverride");
        Assert.assertEquals("New view", _i18nUtils.translate(view.getLabel(), "en"));
        Assert.assertEquals("View added by override", _i18nUtils.translate(view.getDescription(), "en"));
        
        SimpleViewItemGroup group = (SimpleViewItemGroup) view.getViewItems().get(0);
        Assert.assertEquals("New group", _i18nUtils.translate(group.getLabel(), "en"));
        Assert.assertEquals("Group added by override", _i18nUtils.translate(group.getDescription(), "en"));
    }
    
    /**
     * Tests {@link DefaultContentType} implementation with some modifications on attributes and views by extension
     * @throws Exception if an error occurs
     */
    @Test
    @NeedFeatures({
        "test/content-type-overridden",
        "test/content-type-override-all",
    })
    public void testAllOverrideByExtension() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("overriddenType");

        Assert.assertEquals(11, contentType.getModelItems().size());
        Assert.assertEquals(10, contentType.getViewNames().size());
        
        // Check the model item "title"
        Assert.assertTrue("The content type 'overriddenType' should have an attribute title", contentType.hasModelItem("title"));

        // Check if the model item "atribute" has been modified
        Assert.assertTrue("The content type 'overriddenType' should have an attribute 'attribute'", contentType.hasModelItem("attribute"));
        
        // Check if the new model item "newAttribute" has been added
        Assert.assertTrue("The content type 'overriddenType' should have an attribute 'newAttribute'", contentType.hasModelItem("newAttribute"));
        
        // Check that the view "main" has been modified and contains the new attribute
        View mainView = contentType.getView("main");
        
        List<ViewItem> view1Items = mainView.getViewItems();
        Assert.assertEquals(2, view1Items.size());
        
        ViewItem titleFromView = view1Items.get(0);
        Assert.assertEquals("title", titleFromView.getName());
        
        ViewItem newAttributeFromView = view1Items.get(1);
        Assert.assertEquals("newAttribute", newAttributeFromView.getName());
        
        // Check that the view "newViewModifiedAttribute" has been modified and contains the modified attribute
        View newViewModifiedAttribute = contentType.getView("newViewModifiedAttribute");
        
        view1Items = newViewModifiedAttribute.getViewItems();
        Assert.assertEquals(1, view1Items.size());
        
        ViewItem attributeFromView = view1Items.get(0);
        Assert.assertEquals("attribute", attributeFromView.getName());
        
        // Check that the view "newViewModifiedAttribute" has been modified and contains the modified attribute
        View newViewNewAttribute = contentType.getView("newViewNewAttribute");
        
        view1Items = newViewNewAttribute.getViewItems();
        Assert.assertEquals(1, view1Items.size());
        
        newAttributeFromView = view1Items.get(0);
        Assert.assertEquals("newAttribute", newAttributeFromView.getName());
        
        // Check that the view "newViewModifiedAttribute" has been modified and contains the modified attribute
        View newViewComplete = contentType.getView("newViewComplete");
        
        view1Items = newViewComplete.getViewItems();
        Assert.assertEquals(3, view1Items.size());
        
        titleFromView = view1Items.get(0);
        Assert.assertEquals("title", titleFromView.getName());
        
        attributeFromView = view1Items.get(1);
        Assert.assertEquals("attribute", attributeFromView.getName());

        newAttributeFromView = view1Items.get(2);
        Assert.assertEquals("newAttribute", newAttributeFromView.getName());
    }
    
    /**
     * Test {@link DefaultContentType} creation of attributes with the new syntax for attributes
     */
    @Test
    @NeedFeatures({
        "contenttype/content-type-old-attribute-syntax",
    })
    public void testOldAttributeSyntax()
    {
        ContentType contentType = _contentTypeEP.getExtension("content.type.old.attribute.syntax");
        
        // Check if the attribute with the old syntax has been created
        Assert.assertTrue(contentType.hasModelItem("title"));
        
        // Check if the composite with the old syntax has been created
        ModelItemGroup composite = (ModelItemGroup) contentType.getModelItem("composite");
        Assert.assertNotNull("The content type 'content.type.old.attribute.syntax' should have an attribute 'composite'", composite);
        Assert.assertEquals(new I18nizableText("composite"), composite.getLabel());
        Assert.assertEquals(new I18nizableText("composite desc"), composite.getDescription());
        Assert.assertTrue(composite.hasModelItem("compositeAttribute1"));
        Assert.assertTrue(composite.hasModelItem("compositeAttribute2"));

        // Check if the repeater with the old syntax has been created
        ModelItemGroup repeater = (ModelItemGroup) contentType.getModelItem("repeater1");
        Assert.assertEquals(new I18nizableText("Repeater"), repeater.getLabel());
        Assert.assertEquals(new I18nizableText("Repeater desc"), repeater.getDescription());
        Assert.assertTrue(repeater.hasModelItem("repeaterAtt1"));
        Assert.assertTrue(repeater.hasModelItem("repeaterAtt2"));
    }
    
    /**
     * Test {@link DefaultContentType} implementation for a content type with composites and repeaters
     * @throws Exception if an error occurs
     */
    @Test
    @NeedFeatures({
        "exhaustif/content-type",
    })
    public void testGetModelItemForCompositeAndRepeater() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("org.ametys.test.Content.exhaustive");
        
        ContentRestrictedCompositeDefinition composite = (ContentRestrictedCompositeDefinition) contentType.getModelItem("composite");
        Assert.assertEquals("composite", composite.getName());
        List<ModelItem> compositeChildren = composite.getChildren();
        Assert.assertEquals(5, compositeChildren.size());

        AttributeDefinition longMin = (AttributeDefinition) contentType.getModelItem("composite/long-min");
        Assert.assertEquals(compositeChildren.get(0), longMin);
        Assert.assertEquals("long-min", longMin.getName());
        Assert.assertEquals(ModelItemTypeConstants.LONG_TYPE_ID, longMin.getType().getId());

        AttributeDefinition longMax = (AttributeDefinition) contentType.getModelItem("composite/long-max");
        Assert.assertEquals(compositeChildren.get(1), longMax);
        Assert.assertEquals("long-max", longMax.getName());
        Assert.assertEquals(ModelItemTypeConstants.LONG_TYPE_ID, longMax.getType().getId());

        AttributeDefinition doubleMin = (AttributeDefinition) contentType.getModelItem("composite/double-min");
        Assert.assertEquals(compositeChildren.get(2), doubleMin);
        Assert.assertEquals("double-min", doubleMin.getName());
        Assert.assertEquals(ModelItemTypeConstants.DOUBLE_TYPE_ID, doubleMin.getType().getId());

        AttributeDefinition doubleMax = (AttributeDefinition) contentType.getModelItem("composite/double-max");
        Assert.assertEquals(compositeChildren.get(3), doubleMax);
        Assert.assertEquals("double-max", doubleMax.getName());
        Assert.assertEquals(ModelItemTypeConstants.DOUBLE_TYPE_ID, doubleMax.getType().getId());
        
        ContentRestrictedRepeaterDefinition repeater = (ContentRestrictedRepeaterDefinition) contentType.getModelItem("composite/repeater");
        Assert.assertEquals(compositeChildren.get(4), repeater);
        Assert.assertEquals(0, repeater.getInitialSize());
        Assert.assertEquals(0, repeater.getMinSize());
        Assert.assertEquals(5, repeater.getMaxSize());
        Assert.assertEquals("Add", repeater.getAddLabel().getLabel());
        Assert.assertEquals("Delete", repeater.getDeleteLabel().getLabel());
        List<ModelItem> repeaterChildren = repeater.getChildren();
        Assert.assertEquals(20, repeaterChildren.size());

        AttributeDefinition compString = (AttributeDefinition) contentType.getModelItem("composite/repeater/comp-string");
        Assert.assertEquals(repeaterChildren.get(0), compString);
        Assert.assertEquals("comp-string", compString.getName());
        Assert.assertEquals(ModelItemTypeConstants.STRING_TYPE_ID, compString.getType().getId());
        Assert.assertFalse("comp-string attribute should not be multiple", compString.isMultiple());

        AttributeDefinition compStringmult = (AttributeDefinition) contentType.getModelItem("composite/repeater/comp-stringmult");
        Assert.assertEquals(repeaterChildren.get(1), compStringmult);
        Assert.assertEquals("comp-stringmult", compStringmult.getName());
        Assert.assertEquals(ModelItemTypeConstants.STRING_TYPE_ID, compStringmult.getType().getId());
        Assert.assertTrue("comp-string attribute should be multiple", compStringmult.isMultiple());
        
        ContentRestrictedCompositeDefinition subcomposite = (ContentRestrictedCompositeDefinition) contentType.getModelItem("composite/repeater/subcomposite");
        Assert.assertEquals(repeaterChildren.get(repeaterChildren.size() - 1), subcomposite);
        Assert.assertEquals("subcomposite", subcomposite.getName());
        List<ModelItem> subcompositeChildren = subcomposite.getChildren();
        Assert.assertEquals(2, subcompositeChildren.size());

        AttributeDefinition subcompStr = (AttributeDefinition) contentType.getModelItem("composite/repeater/subcomposite/subcomp-str");
        Assert.assertEquals(subcompositeChildren.get(0), subcompStr);
        Assert.assertEquals("subcomp-str", subcompStr.getName());
        Assert.assertEquals(ModelItemTypeConstants.STRING_TYPE_ID, subcompStr.getType().getId());

        AttributeDefinition subcompLong = (AttributeDefinition) contentType.getModelItem("composite/repeater/subcomposite/subcomp-long");
        Assert.assertEquals(subcompositeChildren.get(1), subcompLong);
        Assert.assertEquals("subcomp-long", subcompLong.getName());
        Assert.assertEquals(ModelItemTypeConstants.LONG_TYPE_ID, subcompLong.getType().getId());
        
        ContentRestrictedRepeaterDefinition repeaterThreeLevels = (ContentRestrictedRepeaterDefinition) contentType.getModelItem("repeaterThreeLevels");
        List<ModelItem> repeaterThreeLevelsChildren = repeaterThreeLevels.getChildren();
        Assert.assertEquals(4, repeaterThreeLevelsChildren.size());
        
        ContentRestrictedRepeaterDefinition repeater2 = (ContentRestrictedRepeaterDefinition) contentType.getModelItem("repeaterThreeLevels/repeater2");
        Assert.assertEquals(repeaterThreeLevelsChildren.get(repeaterThreeLevelsChildren.size() - 1), repeater2);
        List<ModelItem> repeater2Children = repeater2.getChildren();
        Assert.assertEquals(1, repeater2Children.size());
        
        ContentRestrictedRepeaterDefinition repeater3 = (ContentRestrictedRepeaterDefinition) contentType.getModelItem("repeaterThreeLevels/repeater2/repeater3");
        Assert.assertEquals(repeater2Children.get(0), repeater3);
        List<ModelItem> repeater3Children = repeater3.getChildren();
        Assert.assertEquals(6, repeater3Children.size());

        AttributeDefinition rep3String = (AttributeDefinition) contentType.getModelItem("repeaterThreeLevels/repeater2/repeater3/rep3-string");
        Assert.assertEquals(repeater3Children.get(0), rep3String);
        Assert.assertEquals("rep3-string", rep3String.getName());
        Assert.assertEquals(ModelItemTypeConstants.STRING_TYPE_ID, rep3String.getType().getId());
        
    }
    
    /**
     * Test {@link DefaultContentType} implementation for an overridden content type
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testTypeOverridden()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.overridden");
        
        // Check the type has the attributes defined by the content type and its override
        Assert.assertEquals(12, contentType.getModelItems().size());
        Assert.assertEquals("title", contentType.getModelItem("title").getName());
        Assert.assertEquals("attribute1", contentType.getModelItem("attribute1").getName());
        Assert.assertEquals("longmin", contentType.getModelItem("longmin").getName());
        Assert.assertEquals("longmax", contentType.getModelItem("longmax").getName());
        Assert.assertEquals("start", contentType.getModelItem("start").getName());
        Assert.assertEquals("end", contentType.getModelItem("end").getName());
        Assert.assertEquals("overriddenAttribute", contentType.getModelItem("overriddenAttribute").getName());
        Assert.assertEquals("overriddenComposite", contentType.getModelItem("overriddenComposite").getName());
        
        // Check the overridden attribute
        ModelItem overriddenAttribute = contentType.getModelItem("overriddenAttribute");
        Assert.assertEquals("Overridden label", overriddenAttribute.getLabel().getLabel());
        Assert.assertEquals("Overridden description", overriddenAttribute.getDescription().getLabel());
        
        // Check the type has the properties defined by the content type and its override
        Assert.assertEquals("property-title", contentType.getModelItem("property-title").getName());
        Assert.assertEquals("property-longmin", contentType.getModelItem("property-longmin").getName());
        Assert.assertEquals("property-longmax", contentType.getModelItem("property-longmax").getName());
        Assert.assertEquals("property-start", contentType.getModelItem("property-start").getName());
        
        // Check the overridden property
        ModelItem overriddenProperty = contentType.getModelItem("property-title");
        Assert.assertEquals("Overridden title property", overriddenProperty.getLabel().getLabel());
        Assert.assertEquals("Property that overrides the existent one", overriddenProperty.getDescription().getLabel());
        
        // Check the type has the views defined by the content type and its override
        Assert.assertEquals(5, contentType.getViewNames().size());
        Assert.assertTrue(contentType.getViewNames().contains("main"));
        Assert.assertTrue(contentType.getViewNames().contains("details"));
        Assert.assertTrue(contentType.getViewNames().contains("overriddenView"));
        Assert.assertTrue(contentType.getViewNames().contains("viewWithOverriddenAttributes"));
        Assert.assertTrue(contentType.getViewNames().contains("newView"));
        
        // Check overridden view
        View overriddenView = contentType.getView("overriddenView");
        Assert.assertEquals(1, overriddenView.getViewItems().size());
        Assert.assertEquals("attribute1", overriddenView.getViewItems().get(0).getName());
        
        // Check view with overridden attributes
        View viewWithOverriddenAttributes = contentType.getView("viewWithOverriddenAttributes");
        Assert.assertEquals("View with overridden attributes", viewWithOverriddenAttributes.getLabel().getLabel());
        Assert.assertEquals("A view containing overridden attributes in extended types. The view itself is not overridden", viewWithOverriddenAttributes.getDescription().getLabel());
        Assert.assertEquals(3, viewWithOverriddenAttributes.getViewItems().size());
        ViewElement titleViewItem = (ViewElement) viewWithOverriddenAttributes.getViewItems().get(0);
        Assert.assertEquals("title", titleViewItem.getLabel().getLabel());
        Assert.assertEquals("title description", titleViewItem.getDescription().getLabel());
        ViewElement overriddenAttribute1ViewItem = (ViewElement) viewWithOverriddenAttributes.getViewItems().get(1);
        Assert.assertEquals("Overridden label", overriddenAttribute1ViewItem.getLabel().getLabel());
        Assert.assertEquals("Overridden description", overriddenAttribute1ViewItem.getDescription().getLabel());
        ModelViewItemGroup compositeViewItem = (ModelViewItemGroup) viewWithOverriddenAttributes.getViewItems().get(2);
        Assert.assertEquals("Composite", compositeViewItem.getLabel().getLabel());
        Assert.assertEquals("Composite", compositeViewItem.getDescription().getLabel());
        Assert.assertEquals(2, compositeViewItem.getViewItems().size());
        SimpleViewItemGroup fieldset1 = (SimpleViewItemGroup) compositeViewItem.getViewItems().get(0);
        Assert.assertEquals("Fieldset 1", fieldset1.getLabel().getLabel());
        Assert.assertEquals("First fieldset", fieldset1.getDescription().getLabel());
        Assert.assertEquals(1, fieldset1.getViewItems().size());
        ModelViewItemGroup anyCompositeViewItem = (ModelViewItemGroup) fieldset1.getViewItems().get(0);
        Assert.assertEquals("Any composite", anyCompositeViewItem.getLabel().getLabel());
        Assert.assertEquals("A composite", anyCompositeViewItem.getDescription().getLabel());
        Assert.assertEquals(1, anyCompositeViewItem.getViewItems().size());
        ViewElement anyAttributeViewItem = (ViewElement) anyCompositeViewItem.getViewItems().get(0);
        Assert.assertEquals("Any", anyAttributeViewItem.getLabel().getLabel());
        Assert.assertEquals("An attribute", anyAttributeViewItem.getDescription().getLabel());
        SimpleViewItemGroup fieldset2 = (SimpleViewItemGroup) compositeViewItem.getViewItems().get(1);
        Assert.assertEquals("Fieldset 2", fieldset2.getLabel().getLabel());
        Assert.assertEquals("Second fieldset", fieldset2.getDescription().getLabel());
        Assert.assertEquals(1, fieldset2.getViewItems().size());
        ModelViewItemGroup repeaterViewItem = (ModelViewItemGroup) fieldset2.getViewItems().get(0);
        Assert.assertEquals("Repeater", repeaterViewItem.getLabel().getLabel());
        Assert.assertEquals("Repeater", repeaterViewItem.getDescription().getLabel());
        Assert.assertEquals(2, repeaterViewItem.getViewItems().size());
        ViewElement overriddenAttributeViewItem = (ViewElement) repeaterViewItem.getViewItems().get(0);
        Assert.assertEquals("Label", overriddenAttributeViewItem.getLabel().getLabel());
        Assert.assertEquals("Description", overriddenAttributeViewItem.getDescription().getLabel());
        ViewElement otherAttribute2ViewItem = (ViewElement) repeaterViewItem.getViewItems().get(1);
        Assert.assertEquals("Other", otherAttribute2ViewItem.getLabel().getLabel());
        Assert.assertEquals("Attribute not present in the override", otherAttribute2ViewItem.getDescription().getLabel());

        // Check view added in the override
        View newView = contentType.getView("newView");
        Assert.assertEquals(2, newView.getViewItems().size());
        Assert.assertEquals("title", newView.getViewItems().get(0).getName());
        Assert.assertEquals("attribute1", newView.getViewItems().get(1).getName());
        
        // Check that the details view is internal
        View detailsView = contentType.getView("details");
        Assert.assertTrue(detailsView.isInternal());
        
        // Check overriden global validators
        Assert.assertEquals(2, contentType.getGlobalValidators().size());
        Assert.assertEquals("org.ametys.cms.contenttype.validation.LongIntervalValidator", contentType.getGlobalValidators().get(0).getClass().getName());
        Assert.assertEquals("org.ametys.cms.contenttype.validation.DateTimeIntervalValidator", contentType.getGlobalValidators().get(1).getClass().getName());
    }
    
    /**
     * Test {@link DefaultContentType} implementation for an overridden content type
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testTypeInheritedAndOverridden()
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.extends.extends.overridden");
        
        ElementDefinition overriddenAttribute1 = (ElementDefinition) contentType.getModelItem("overriddenAttribute");
        Assert.assertEquals("Extended overridden label", overriddenAttribute1.getLabel().getLabel());
        Assert.assertEquals("Extended overridden description", overriddenAttribute1.getDescription().getLabel());
        
        // Check the overridden composite and its children
        ModelItemGroup overriddenComposite = (ModelItemGroup) contentType.getModelItem("overriddenComposite");
        Assert.assertEquals("Overridden composite", overriddenComposite.getLabel().getLabel());
        Assert.assertEquals("Overridden composite", overriddenComposite.getDescription().getLabel());
        Assert.assertEquals(2, overriddenComposite.getChildren().size());
        
        ModelItemGroup anyComposite = (ModelItemGroup) overriddenComposite.getChildren().get(0);
        Assert.assertEquals("Any composite", anyComposite.getLabel().getLabel());
        Assert.assertEquals("A composite", anyComposite.getDescription().getLabel());
        Assert.assertEquals(1, anyComposite.getChildren().size());
        
        ElementDefinition anyAttribute = (ElementDefinition) anyComposite.getChildren().get(0);
        Assert.assertEquals("Any", anyAttribute.getLabel().getLabel());
        Assert.assertEquals("An attribute", anyAttribute.getDescription().getLabel());
        
        ModelItemGroup overriddenRepeater = (ModelItemGroup) overriddenComposite.getChildren().get(1);
        Assert.assertEquals("Overridden repeater", overriddenRepeater.getLabel().getLabel());
        Assert.assertEquals("Overridden repeater", overriddenRepeater.getDescription().getLabel());
        Assert.assertEquals(1, overriddenRepeater.getChildren().size());
        
        ElementDefinition overriddenAttribute2 = (ElementDefinition) overriddenRepeater.getChildren().get(0);
        Assert.assertEquals("Overridden label", overriddenAttribute2.getLabel().getLabel());
        Assert.assertEquals("Overridden description", overriddenAttribute2.getDescription().getLabel());
        
        // Check view with overridden attributes
        View view = contentType.getView("viewWithOverriddenAttributes");
        Assert.assertEquals("View with overridden attributes", view.getLabel().getLabel());
        Assert.assertEquals("A view containing overridden attributes in extended types. The view itself is not overridden", view.getDescription().getLabel());
        Assert.assertEquals(3, view.getViewItems().size());
        
        ViewElement titleViewItem = (ViewElement) view.getViewItems().get(0);
        Assert.assertEquals("title", titleViewItem.getLabel().getLabel());
        Assert.assertEquals("title description", titleViewItem.getDescription().getLabel());
        
        ViewElement overriddenAttribute1ViewItem = (ViewElement) view.getViewItems().get(1);
        Assert.assertEquals("Extended overridden label", overriddenAttribute1ViewItem.getLabel().getLabel());
        Assert.assertEquals("Extended overridden description", overriddenAttribute1ViewItem.getDescription().getLabel());
        
        ModelViewItemGroup compositeViewItem = (ModelViewItemGroup) view.getViewItems().get(2);
        Assert.assertEquals("Overridden composite", compositeViewItem.getLabel().getLabel());
        Assert.assertEquals("Overridden composite", compositeViewItem.getDescription().getLabel());
        Assert.assertEquals(2, compositeViewItem.getViewItems().size());
        
        SimpleViewItemGroup fieldset1 = (SimpleViewItemGroup) compositeViewItem.getViewItems().get(0);
        Assert.assertEquals("Fieldset 1", fieldset1.getLabel().getLabel());
        Assert.assertEquals("First fieldset", fieldset1.getDescription().getLabel());
        Assert.assertEquals(1, fieldset1.getViewItems().size());
        
        ModelViewItemGroup anyCompositeViewItem = (ModelViewItemGroup) fieldset1.getViewItems().get(0);
        Assert.assertEquals("Any composite", anyCompositeViewItem.getLabel().getLabel());
        Assert.assertEquals("A composite", anyCompositeViewItem.getDescription().getLabel());
        Assert.assertEquals(1, anyCompositeViewItem.getViewItems().size());
        
        ViewElement anyAttributeViewItem = (ViewElement) anyCompositeViewItem.getViewItems().get(0);
        Assert.assertEquals("Any", anyAttributeViewItem.getLabel().getLabel());
        Assert.assertEquals("An attribute", anyAttributeViewItem.getDescription().getLabel());
        
        SimpleViewItemGroup fieldset2 = (SimpleViewItemGroup) compositeViewItem.getViewItems().get(1);
        Assert.assertEquals("Fieldset 2", fieldset2.getLabel().getLabel());
        Assert.assertEquals("Second fieldset", fieldset2.getDescription().getLabel());
        Assert.assertEquals(1, fieldset2.getViewItems().size());
        
        ModelViewItemGroup repeaterViewItem = (ModelViewItemGroup) fieldset2.getViewItems().get(0);
        Assert.assertEquals("Overridden repeater", repeaterViewItem.getLabel().getLabel());
        Assert.assertEquals("Overridden repeater", repeaterViewItem.getDescription().getLabel());
        Assert.assertEquals(1, repeaterViewItem.getViewItems().size());
        
        ViewElement overriddenAttribute2ViewItem = (ViewElement) repeaterViewItem.getViewItems().get(0);
        Assert.assertEquals("Overridden label", overriddenAttribute2ViewItem.getLabel().getLabel());
        Assert.assertEquals("Overridden description", overriddenAttribute2ViewItem.getDescription().getLabel());
        
        // Check that the details view is internal
        View detailsView = contentType.getView("details");
        Assert.assertTrue(detailsView.isInternal());
    }
    
    /**
     * Asserts that the view item is a model view item and tests the name of its definition.
     * @param name the expected definition name.
     * @param item the {@link ViewItem} to test.
     */
    protected void assertViewItem(String name, ViewItem item)
    {
        if (!(item instanceof ModelViewItem))
        {
            Assert.fail("The item is not a model view element.");
        }
        
        ModelViewItem modelViewItem = (ModelViewItem) item;
        Assert.assertEquals(name, modelViewItem.getDefinition().getName());
    }
    
    /**
     * Tests the {@link DefaultContentType} implementation for rich text attributes parsing 
     */
    @Test
    @NeedFeatures({
        "exhaustif/content-type",
    })
    public void testGetModelItemForRichText()
    {
        ContentType contentType = _contentTypeEP.getExtension("org.ametys.test.Content.exhaustive");
        
        RichTextAttributeDefinition richTextDefinition = (RichTextAttributeDefinition) contentType.getModelItem("content");
        
        List<SemanticAnnotation> semanticAnnotations = richTextDefinition.getSemanticAnnotations();
        Assert.assertEquals(2, semanticAnnotations.size());
        
        SemanticAnnotation postitAnnotation = semanticAnnotations.get(0);
        Assert.assertEquals("postit", postitAnnotation.getId());
        Assert.assertEquals("postit", postitAnnotation.getLabel().getLabel());
        Assert.assertEquals("Post-it", postitAnnotation.getDescription().getLabel());
        
        SemanticAnnotation annotAnnotation = semanticAnnotations.get(1);
        Assert.assertEquals("annot", annotAnnotation.getId());
        Assert.assertEquals("annot", annotAnnotation.getLabel().getLabel());
        Assert.assertEquals("Annotation", annotAnnotation.getDescription().getLabel());
    }
    
    /**
     * Tests the {@link DefaultContentType} implementation for parsing of attributes of type content 
     */
    @Test
    @NeedFeatures({
        "exhaustif/content-type",
    })
    public void testGetModelItemForContentAttributeType()
    {
        ContentType contentType = _contentTypeEP.getExtension("org.ametys.test.Content.exhaustive");
        
        ContentAttributeDefinition linkContentDefinition = (ContentAttributeDefinition) contentType.getModelItem("linkContent");
        Assert.assertNull(linkContentDefinition.getContentTypeId());
        Assert.assertNull(linkContentDefinition.getInvertRelationPath());
        Assert.assertFalse(linkContentDefinition.getForceInvert());
        
        ContentAttributeDefinition sportDefinition = (ContentAttributeDefinition) contentType.getModelItem("sport");
        Assert.assertEquals("simple.Sport", sportDefinition.getContentTypeId());
        Assert.assertNull(sportDefinition.getInvertRelationPath());
        Assert.assertFalse(sportDefinition.getForceInvert());
        
        ContentAttributeDefinition ownsDefinition = (ContentAttributeDefinition) contentType.getModelItem("owns");
        Assert.assertEquals("org.ametys.test.Content.exhaustive", ownsDefinition.getContentTypeId());
        Assert.assertEquals("belongsTo", ownsDefinition.getInvertRelationPath());
        Assert.assertFalse(ownsDefinition.getForceInvert());
    }
    
    /**
     * Tests the {@link DefaultContentType} implementation for parsing of attributes of type content 
     */
    @Test
    @NeedFeatures({
        "exhaustif/content-type",
    })
    public void testGetModelItemToGetAttributesOnLinkedContent()
    {
        ContentType contentType = _contentTypeEP.getExtension("org.ametys.test.Content.exhaustive");
        
        ElementDefinition stringmultDefinition = (ElementDefinition) contentType.getModelItem("owns/stringmult");
        Assert.assertEquals("Multiple string", stringmultDefinition.getLabel().getLabel());

        ElementDefinition postCodeDefinition = (ElementDefinition) contentType.getModelItem("address/city/postCode");
        Assert.assertEquals("Post code", postCodeDefinition.getLabel().getLabel());
        
        ElementDefinition titleDefinition = (ElementDefinition) contentType.getModelItem("linkContent/title");
        Assert.assertNotNull(titleDefinition);
        
        try
        {
            contentType.getModelItem("linkContent/otherAttribute");
            Assert.fail("The attribute 'linkContent' has no defined content type. The only accessible linked attribute should be the title");
        }
        catch (UndefinedItemPathException e)
        {
            // Nothing to do
        }
    }
    
    /**
     * Tests the implementation of {@link DefaultContentType#getDefaultWorkflowName()}.
     */
    @Test
    @NeedFeatures({
        "test-default-workflow/content-types",
        "test-default-workflow/reference-tables",
        "test-default-workflow/reference-tables.inherit",
        "test-default-workflow/reference-tables.inherit.ref",
        "test-default-workflow/reference-tables.mixed"
    })
    public void testGetDefaultWorkflowName()
    {
        // Classic content types
        Assert.assertEquals(Optional.of("content"),         _contentTypeEP.getExtension("A.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("first"),           _contentTypeEP.getExtension("B.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("first"),           _contentTypeEP.getExtension("C.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("second"),          _contentTypeEP.getExtension("D.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.empty(),               _contentTypeEP.getExtension("E.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("third"),           _contentTypeEP.getExtension("F.workflow").getDefaultWorkflowName());
        
        // Reference tables
        Assert.assertEquals(Optional.of("reference-table"), _contentTypeEP.getExtension("ref.A.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("first"),           _contentTypeEP.getExtension("ref.B.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("first"),           _contentTypeEP.getExtension("ref.C.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("second"),          _contentTypeEP.getExtension("ref.D.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.empty(),               _contentTypeEP.getExtension("ref.E.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("third"),           _contentTypeEP.getExtension("ref.F.workflow").getDefaultWorkflowName());
        
        // Content types extending only reference tables
        Assert.assertEquals(Optional.of("reference-table"), _contentTypeEP.getExtension("inherit.ref.A.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("first"),           _contentTypeEP.getExtension("inherit.ref.B.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("first"),           _contentTypeEP.getExtension("inherit.ref.C.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("second"),          _contentTypeEP.getExtension("inherit.ref.D.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.empty(),               _contentTypeEP.getExtension("inherit.ref.E.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("third"),           _contentTypeEP.getExtension("inherit.ref.F.workflow").getDefaultWorkflowName());
        
        // Reference tables extending only content types
        Assert.assertEquals(Optional.of("reference-table"), _contentTypeEP.getExtension("inherit.A.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("first"),           _contentTypeEP.getExtension("inherit.B.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("first"),           _contentTypeEP.getExtension("inherit.C.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("second"),          _contentTypeEP.getExtension("inherit.D.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.empty(),               _contentTypeEP.getExtension("inherit.E.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.of("third"),           _contentTypeEP.getExtension("inherit.F.workflow").getDefaultWorkflowName());
        
        // Content types extending both reference tables and content types
        Assert.assertEquals(Optional.of("first"),           _contentTypeEP.getExtension("ref.mixed1.workflow").getDefaultWorkflowName());
        Assert.assertEquals(Optional.empty(),               _contentTypeEP.getExtension("ref.mixed2.workflow").getDefaultWorkflowName());
    }

    /**
     * Tests the implementation of {@link DefaultContentType#getConfiguredDefaultWorkflowNames()}.
     */
    @Test
    @NeedFeatures({
        "test-default-workflow/content-types",
        "test-default-workflow/reference-tables",
        "test-default-workflow/reference-tables.inherit",
        "test-default-workflow/reference-tables.inherit.ref",
        "test-default-workflow/reference-tables.mixed"
    })
    public void testGetConfiguredDefaultWorkflowNames()
    {
        // Classic content types
        Assert.assertEquals(Set.of(),                   _contentTypeEP.getExtension("A.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("first"),            _contentTypeEP.getExtension("B.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("first"),            _contentTypeEP.getExtension("C.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("second"),           _contentTypeEP.getExtension("D.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("first", "second"),  _contentTypeEP.getExtension("E.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("third"),            _contentTypeEP.getExtension("F.workflow").getConfiguredDefaultWorkflowNames());
        
        // Reference tables
        Assert.assertEquals(Set.of(),                   _contentTypeEP.getExtension("ref.A.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("first"),            _contentTypeEP.getExtension("ref.B.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("first"),            _contentTypeEP.getExtension("ref.C.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("second"),           _contentTypeEP.getExtension("ref.D.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("first", "second"),  _contentTypeEP.getExtension("ref.E.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("third"),            _contentTypeEP.getExtension("ref.F.workflow").getConfiguredDefaultWorkflowNames());
        
        // Content types extending only reference tables
        Assert.assertEquals(Set.of(),                   _contentTypeEP.getExtension("inherit.ref.A.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("first"),            _contentTypeEP.getExtension("inherit.ref.B.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("first"),            _contentTypeEP.getExtension("inherit.ref.C.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("second"),           _contentTypeEP.getExtension("inherit.ref.D.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("first", "second"),  _contentTypeEP.getExtension("inherit.ref.E.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("third"),            _contentTypeEP.getExtension("inherit.ref.F.workflow").getConfiguredDefaultWorkflowNames());
        
        // Reference tables extending only content types
        Assert.assertEquals(Set.of(),                   _contentTypeEP.getExtension("inherit.A.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("first"),            _contentTypeEP.getExtension("inherit.B.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("first"),            _contentTypeEP.getExtension("inherit.C.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("second"),           _contentTypeEP.getExtension("inherit.D.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("first", "second"),  _contentTypeEP.getExtension("inherit.E.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("third"),            _contentTypeEP.getExtension("inherit.F.workflow").getConfiguredDefaultWorkflowNames());
        
        // Content types extending both reference tables and content types
        Assert.assertEquals(Set.of("first"),            _contentTypeEP.getExtension("ref.mixed1.workflow").getConfiguredDefaultWorkflowNames());
        Assert.assertEquals(Set.of("first", "second"),  _contentTypeEP.getExtension("ref.mixed2.workflow").getConfiguredDefaultWorkflowNames());
    }
    
    /**
     * Test {@link DefaultContentType} implementation for properties inheritance
     * The exhaustive content type contains some properties. Tests that a type extending the exhaustive one contains its parent's properties
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testPropertiesInheritanceFirstLevel() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.extends.exhaustive");
        
        Assert.assertTrue(contentType.hasModelItem("owns-title"));
        Assert.assertTrue(contentType.hasModelItem("owns-sport"));
        Assert.assertTrue(contentType.hasModelItem("owns-linkContent"));
        Assert.assertTrue(contentType.hasModelItem("owns-owns"));
        Assert.assertTrue(contentType.hasModelItem("owns-title"));
        Assert.assertTrue(contentType.hasModelItem("linkContent-title"));
        Assert.assertTrue(contentType.hasModelItem("linkContents-title"));
    }
    
    /**
     * Test {@link DefaultContentType} implementation for properties inheritance
     * The exhaustive content type contains some properties. Tests that a type extending the exhaustive one contains its parent's properties
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures({
        "test/content-type-view-tests",
    })
    public void testPropertiesInheritanceSecondLevel() throws Exception
    {
        ContentType contentType = _contentTypeEP.getExtension("tests.extends.exhaustive.and.basic");
        
        Assert.assertTrue(contentType.hasModelItem("owns-title"));
        Assert.assertTrue(contentType.hasModelItem("owns-sport"));
        Assert.assertTrue(contentType.hasModelItem("owns-linkContent"));
        Assert.assertTrue(contentType.hasModelItem("owns-owns"));
        Assert.assertTrue(contentType.hasModelItem("owns-title"));
        Assert.assertTrue(contentType.hasModelItem("linkContent-title"));
        Assert.assertTrue(contentType.hasModelItem("linkContents-title"));
    }
}
