/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.Optional;

import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.PluginInitializationFailureTestCase;
import org.ametys.runtime.test.framework.Unoptimized;

/**
 * {@link ContentType} with invalid configuration (invalid view or attributes definitions) test case.
 */
@Unoptimized // The nature of the test is to fail when some features are not present, features being different in every @Test => @Unoptimized is thus NECESSARY
public class ContentTypeWithInvalidConfigTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures, PluginInitializationFailureTestCase
{
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        // Nothing to lookup here
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with a view that includes an inexisting super-view from its parents
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-view-includes-inexisting-super-view")
    public void testConfigureContentTypeViewIncludesInexistingSuperView() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A content type's view can't include a view that does not exist in its parents", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("viewX"));
        Assert.assertTrue(configurationException.getMessage().contains("not found"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with a view that includes an inexisting super-view from its designated parent
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-view-includes-inexisting-view-from-super-type")
    public void testConfigureContentTypeViewIncludesInexistingViewFromSuperType() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A content type's view can't include a view that does not exist in its designated parent", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("viewX"));
        Assert.assertTrue(configurationException.getMessage().contains("not found"));
        Assert.assertTrue(configurationException.getMessage().contains("tests.extends.exhaustive"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with a view that includes another view from a type that is not its parent
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-view-includes-from-non-parent-type")
    public void testConfigureContentTypeViewIncludesFromNonParentType() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A content type's view can't include a view from a type that is not its parent", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("viewA"));
        Assert.assertTrue(configurationException.getMessage().contains("not a super-type"));
        Assert.assertTrue(configurationException.getMessage().contains("tests.basic.type"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with a view that references an inexisting attribute
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-view-with-inexisting-attribute")
    public void testConfigureContentTypeViewReferencingInexistingAttribute() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A content type's view can't reference an inexisting attribute", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof UndefinedItemPathException);
        UndefinedItemPathException undefinedItemPathException = (UndefinedItemPathException) initializationException.get();
        Assert.assertTrue(undefinedItemPathException.getMessage().contains("inexisting"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type without title attribute
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-no-title")
    public void testConfigureContentTypeWithoutTitle() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A content type should have a title", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("title"));
        Assert.assertTrue(configurationException.getMessage().contains("tests.invalid.no.title"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with title attribute with a long type instead of string or multilingual string
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-title-wrong-type")
    public void testConfigureContentTypeWithTitleWrongType() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A content type should have a title of type string or multilingual-string", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("title"));
        Assert.assertTrue(configurationException.getMessage().contains("tests.invalid.title.wrong.type"));
        Assert.assertTrue(configurationException.getMessage().contains("long"));
        Assert.assertTrue(configurationException.getMessage().contains("string"));
        Assert.assertTrue(configurationException.getMessage().contains("multilingual-string"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with multiple title attribute
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-title-multiple")
    public void testConfigureContentTypeWithMultipleTitleWrong() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A content type should have a non multiple title", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("title"));
        Assert.assertTrue(configurationException.getMessage().contains("tests.invalid.multiple.title"));
        Assert.assertTrue(configurationException.getMessage().contains("multiple"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with title attribute not mandatory
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-title-not-mandatory")
    public void testConfigureContentTypeWithMultipleTitleNotMandatory() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A content type should have a mandatory title", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("title"));
        Assert.assertTrue(configurationException.getMessage().contains("tests.invalid.title.not.mandatory"));
        Assert.assertTrue(configurationException.getMessage().contains("mandatory"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with an attribute of type content that uses a non existing type
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-content-attribute-with-inexisting-content-type")
    public void testConfigureContentTypeWithContentAttributeWithInvalidType() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("An attribute of type content must use an existing content type", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("content type"));
        Assert.assertTrue(configurationException.getMessage().contains("tests.invalid.content.type.not.exist"));
        Assert.assertTrue(configurationException.getMessage().contains("non.existing.content.type"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with an attribute of type content that is multiple and is in a repeater
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-multiple-content-attribute-with-mutual-ref-in-a-repeater")
    public void testConfigureContentTypeWithContentAttributeMultipleAndInARepeater() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("An attribute of type content can't be multiple and in a repeater", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("multiple"));
        Assert.assertTrue(configurationException.getMessage().contains("repeater"));
        Assert.assertTrue(configurationException.getMessage().contains("tests.invalid.content.multiple.in.repeater"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with an attribute of type content that in a low level of repeaters
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-content-attribute-with-mutual-ref-in-low-level-of-repeaters")
    public void testConfigureContentTypeWithContentAttributeInALowLevelRepeater() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("An attribute of type content can't be in a low level repeater", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("repeater"));
        Assert.assertTrue(configurationException.getMessage().contains("level"));
        Assert.assertTrue(configurationException.getMessage().contains("tests.invalid.content.in.low.level.repeaters"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with an attribute of type content that references a non existing attribute
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-content-attribute-references-not-existing-attribute")
    public void testConfigureContentTypeWithContentAttributeReferencesNonExistingAttribute() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("An attribute of type content must refer to an existing attribute", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("exist"));
        Assert.assertTrue(configurationException.getMessage().contains("tests.invalid.content.reference.not.exist"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with an attribute of type content that references an attribute that is not of type content
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-content-attribute-references-not-content-attribute")
    public void testConfigureContentTypeWithContentAttributeReferencesNotContentAttribute() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("An attribute of type content must refer an attribute of type content", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("type"));
        Assert.assertTrue(configurationException.getMessage().contains("tests.invalid.content.reference.not.content"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with an attribute of type content that references an attribute with an invalid content type
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-content-attribute-references-attribute-in-incompatible-type")
    public void testConfigureContentTypeWithContentAttributeReferencesAttributeWithInvalidType() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("An attribute of type content must refer an attribute using the same or a descendant content type", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("incompatible type"));
        Assert.assertTrue(configurationException.getMessage().contains("tests.invalid.content.reference.type1"));
        Assert.assertTrue(configurationException.getMessage().contains("tests.invalid.content.reference.type2"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with an attribute of type content that references an attribute that is not referencing it back
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-content-attribute-references-content-attribute-referencing-another-attribute")
    public void testConfigureContentTypeWithContentAttributeReferencesAttributeNotReferencesItBack() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("An attribute of type content must refer an attribute that refers to the first one back", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("back"));
        Assert.assertTrue(configurationException.getMessage().contains("tests.invalid.content.reference.content.reference.other"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type with an attribute of type content that references an attribute that is not referencing it back
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-view-join-on-itself")
    public void testConfigureContentTypeWithViewThatReferencesItself() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("A view can't make a reference to itself in a join", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        Assert.assertTrue(configurationException.getMessage().contains("view-refs-itself"));
    }
    
    
    /**
     * Test {@link ContentType} implementation to configure a content type overriding an inexisting content type
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures("test/invalid-content-type-override-inexisting-content-type")
    public void testOverrideByExtensionOfInexistingContentType() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("An override extension can't override an unexisting content type", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        
        String message = configurationException.getMessage();
        Assert.assertTrue(message.contains("inexistingType"));
        Assert.assertTrue(message.contains("overridden"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type overridden by an extension that changes the type of a model item
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures({
        "test/content-type-overridden",
        "test/invalid-content-type-override-changing-attribute-type",
    })
    public void testOverrideByExtensionChangingTypeOfModelItem() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("The type of an attribute cannot be changed by override", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        
        String message = configurationException.getMessage();
        Assert.assertTrue(message.contains("type"));
        Assert.assertTrue(message.contains("attribute"));
        Assert.assertTrue(message.contains("string"));
        Assert.assertTrue(message.contains("boolean"));
        Assert.assertTrue(message.contains("overridden"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type overridden by an extension that changes the type of a model item on a deeper level
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures({
        "test/content-type-overridden",
        "test/invalid-content-type-override-changing-attribute-type-deeper-level",
    })
    public void testOverrideByExtensionDeepChangingTypeOfModelItem() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("The type of an attribute cannot be changed by override", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        
        String message = configurationException.getMessage();
        Assert.assertTrue(message.contains("type"));
        Assert.assertTrue(message.contains("comp-string"));
        Assert.assertTrue(message.contains("string"));
        Assert.assertTrue(message.contains("boolean"));
        Assert.assertTrue(message.contains("overridden"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type overridden by an extension that changes the type of a model item
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures({
        "test/content-type-overridden",
        "test/invalid-content-type-override-changing-attribute-cardinality",
    })
    public void testOverrideByExtensionChangingCardinalityOfModelItem() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("The cardinality of an attribute cannot be changed by override", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        
        String message = configurationException.getMessage();
        Assert.assertTrue(message.contains("cardinality"));
        Assert.assertTrue(message.contains("attribute"));
        Assert.assertTrue(message.contains("multiple"));
        Assert.assertTrue(message.contains("not multiple"));
        Assert.assertTrue(message.contains("overridden"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type overridden by an extension that changes the type of a model item on a deeper level
     * @throws Exception if an error occurs.
     */
    @Test
    @NeedFeatures({
        "test/content-type-overridden",
        "test/invalid-content-type-override-changing-attribute-cardinality-deeper-level",
    })
    public void testOverrideByExtensionDeepChangingCardinalityOfModelItem() throws Exception
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("The cardinality of an attribute cannot be changed by override", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        
        String message = configurationException.getMessage();
        Assert.assertTrue(message.contains("cardinality"));
        Assert.assertTrue(message.contains("comp-string"));
        Assert.assertTrue(message.contains("multiple"));
        Assert.assertTrue(message.contains("not multiple"));
        Assert.assertTrue(message.contains("overridden"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type overriding a nonexistent view
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/invalid-content-type-override-view-nonexistent-view"
    })
    public void testConfigureContentTypeOverrideNonexistentView()
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("An override extension can't override a nonexistent view", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        
        ConfigurationException exception = (ConfigurationException) initializationException.get();
        
        String message = exception.getMessage();
        Assert.assertTrue(message.contains("nonexistentView"));
        Assert.assertTrue(message.contains("overrides"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type overriding by adding an attribute before/after a group path
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/invalid-content-type-override-view-add-after-group-path"
        })
    public void testConfigureContentTypeOverrideViewAddAttributeAfterGroupPath()
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("An override extension can't override a view by adding an attribute after a group path", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof IllegalArgumentException);
        
        IllegalArgumentException illegalArgumentException = (IllegalArgumentException) initializationException.get();
        
        String message = illegalArgumentException.getMessage();
        Assert.assertTrue(message.contains("GroupTab/GroupFieldset"));
        Assert.assertTrue(message.contains("after"));
    }
    
    /**
     * Test {@link ContentType} implementation to configure a content type overriding by adding an attribute before AND after some attributes
     */
    @Test
    @NeedFeatures({
        "contenttype.override.view/invalid-content-type-override-view-add-after-and-before"
        })
    public void testConfigureContentTypeOverrideViewAddAfterAndBefore()
    {
        Optional<Throwable> initializationException = getInitializationException();
        Assert.assertTrue("An override extension can't override a view by adding an attribute before and after some attributes", initializationException.isPresent());
        Assert.assertTrue(initializationException.get() instanceof ConfigurationException);
        
        ConfigurationException configurationException = (ConfigurationException) initializationException.get();
        
        String message = configurationException.getMessage();
        Assert.assertTrue(message.contains("attributeToAdd"));
        Assert.assertTrue(message.contains("after and before"));
    }
}
