/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.contenttype;

import java.util.List;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.model.ContentElementDefinition;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElementAccessor;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Tests browsing the model of content types via the {@link ContentTypesHelper}.
 */
@NeedFeatures({
    "cms/content-types.helper",
    "exhaustif/content-type",
    "cms/workflow.content"
})
public class ContentTypesHelperTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    /** Ametys object resolver to subclasses. */
    protected static ContentTypeExtensionPoint _contentTypeExtensionPoint;
    
    /** The content type helper. */
    protected static ContentTypesHelper _contentTypesHelper;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
    }
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }
    
    /**
     * Test the retrieval of lists of model items
     * @throws Exception If an error occurs.
     */
    @SuppressWarnings("static-access")
    @Test
    public void testGetModelItemPath() throws Exception
    {
        ContentType exhaustive = _contentTypeExtensionPoint.getExtension("org.ametys.test.Content.exhaustive");
        
        List<ModelItem> defs = _contentTypesHelper.getModelItemPath("title", exhaustive);
        Assert.assertEquals(defs.size(), 1);
        assertModelItemType(ModelItemTypeConstants.STRING_TYPE_ID, defs.get(0));
        
        defs = _contentTypesHelper.getModelItemPath("composite/repeater", exhaustive);
        Assert.assertEquals(2, defs.size());
        assertModelItemType(ModelItemTypeConstants.COMPOSITE_TYPE_ID, defs.get(0));
        assertModelItemType(ModelItemTypeConstants.REPEATER_TYPE_ID, defs.get(1));
        
        defs = _contentTypesHelper.getModelItemPath("composite/repeater/subcomposite/subcomp-str", exhaustive);
        Assert.assertEquals(4, defs.size());
        assertModelItemType(ModelItemTypeConstants.COMPOSITE_TYPE_ID, defs.get(0));
        assertModelItemType(ModelItemTypeConstants.REPEATER_TYPE_ID, defs.get(1));
        assertModelItemType(ModelItemTypeConstants.COMPOSITE_TYPE_ID, defs.get(2));
        assertModelItemType(ModelItemTypeConstants.STRING_TYPE_ID, defs.get(3));
        
        defs = _contentTypesHelper.getModelItemPath("linkContent/title", exhaustive);
        Assert.assertEquals(2, defs.size());
        assertModelItemType(ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID, defs.get(0));
        assertModelItemType(ModelItemTypeConstants.STRING_TYPE_ID, defs.get(1));
        
        try
        {
            _contentTypesHelper.getModelItemPath("linkContent/nonexisting", exhaustive);
            Assert.fail("a path of an unexisting data should have lead to an UndefinedItemPathException");
        }
        catch (UndefinedItemPathException e)
        {
            // Do nothing
        }
        
        try
        {
            _contentTypesHelper.getModelItemPath("linkContent/address/country", exhaustive);
            Assert.fail("a path of an unexisting data should have lead to an UndefinedItemPathException");
        }
        catch (UndefinedItemPathException e)
        {
            // Do nothing
        }
        
        defs = _contentTypesHelper.getModelItemPath("owns/title", exhaustive);
        Assert.assertEquals(2, defs.size());
        assertModelItemType(ModelItemTypeConstants.STRING_TYPE_ID, defs.get(1));
        
        defs = _contentTypesHelper.getModelItemPath("owns/composite/repeater/subcomposite/subcomp-str", exhaustive);
        Assert.assertEquals(5, defs.size());
        assertModelItemType(ModelItemTypeConstants.STRING_TYPE_ID, defs.get(4));
        
        defs = _contentTypesHelper.getModelItemPath("owns/belongsTo/title", exhaustive);
        Assert.assertEquals(3, defs.size());
        assertModelItemType(ModelItemTypeConstants.STRING_TYPE_ID, defs.get(2));
        
        defs = _contentTypesHelper.getModelItemPath("owns/repeaterThreeLevels/repeater2/repeater3/rep3-owns/belongsTo", exhaustive);
        Assert.assertEquals(6, defs.size());
        assertModelItemType(ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID, defs.get(5));
    }
    
    /**
     * Assert.assert that the given model item is of the expected type.
     * @param expectedType the expected model item type.
     * @param modelItem the model item.
     */
    protected static void assertModelItemType(String expectedType, ModelItem modelItem)
    {
        if (modelItem == null)
        {
            Assert.fail("The model item is null.");
        }
        else
        {
            Assert.assertEquals(expectedType, modelItem.getType().getId());
        }
    }
    
    /**
     * Test the retrieval of lists of metadata definitions by predicate
     * @throws Exception If an error occurs.
     */
    @Test
    public void testModelHelperFindByType() throws Exception
    {
        ContentType exhaustive = _contentTypeExtensionPoint.getExtension("org.ametys.test.Content.exhaustive");
        
        List<String> items = ModelHelper.findModelItemsByType(exhaustive, ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID)
                                        .stream().map(item -> item.getPath()).collect(Collectors.toList());
        
        Assert.assertEquals(19, items.size());
        Assert.assertTrue(items.contains("linkContent"));
        Assert.assertTrue(items.contains("address/city"));
        Assert.assertTrue(items.contains("composite/repeater/comp-linkContent"));
        Assert.assertTrue(items.contains("composite/repeater/comp-owns"));
        Assert.assertTrue(items.contains("repeaterThreeLevels/repeater2/repeater3/rep3-owns"));
        Assert.assertTrue(items.contains("owns-sport"));
        Assert.assertTrue(items.contains("owns-linkContent"));
        Assert.assertTrue(items.contains("owns-owns"));
        
        items = ModelHelper.findModelItemsByType(exhaustive, ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID)
                           .stream()
                           .map(ElementDefinition.class::cast)
                           .filter(item -> item.isMultiple())
                           .map(item -> item.getPath()).collect(Collectors.toList());
        
        Assert.assertEquals(5, items.size());
        Assert.assertFalse(items.contains("linkContent"));
        Assert.assertTrue(items.contains("linkContents"));
        Assert.assertFalse(items.contains("composite/repeater/comp-linkContent"));
        Assert.assertTrue(items.contains("composite/repeater/comp-linkContents"));
        
        items = ModelHelper.findModelItemsByType(exhaustive, ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID)
                           .stream()
                           .filter(ContentElementDefinition.class::isInstance)
                           .map(ContentElementDefinition.class::cast)
                           .filter(def -> "org.ametys.test.Content.exhaustive".equals(def.getContentTypeId()))
                           .map(item -> item.getPath()).collect(Collectors.toList());

        Assert.assertEquals(8, items.size());
        Assert.assertFalse(items.contains("linkContent"));
        Assert.assertTrue(items.contains("owns"));
        Assert.assertTrue(items.contains("composite/repeater/comp-owns"));
        Assert.assertTrue(items.contains("repeaterThreeLevels/rep-owns"));
        Assert.assertTrue(items.contains("repeaterThreeLevels/repeater2/repeater3/rep3-owns"));
        
        Assert.assertEquals(16, ModelHelper.findModelItemsByType(exhaustive, org.ametys.runtime.model.type.ModelItemTypeConstants.LONG_TYPE_ID).size() + ModelHelper.findModelItemsByType(exhaustive, org.ametys.runtime.model.type.ModelItemTypeConstants.DOUBLE_TYPE_ID).size());
    }
    
    /**
     * Tests that the view cache has not been filled with views with not yet resolved references
     */
    @Test
    @NeedFeatures({
        "contenttype.helper/content-type-helper-view-references-another-view",
    })
    public void testGetViewCachedViewHasResolvedReferences()
    {
        // Get the view from cache
        View view = _contentTypesHelper.getView("main", new String[] {"contenttype.helper.view.references.another.view"}, new String[0]);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(2, viewItems.size());
        Assert.assertEquals("title", viewItems.get(0).getName());
        
        ViewElementAccessor linkExhaustive = (ViewElementAccessor) viewItems.get(1);
        Assert.assertEquals("linkExhaustive", linkExhaustive.getName());
        List<ViewItem> linkExhaustiveItems = linkExhaustive.getViewItems();
        Assert.assertEquals(1, linkExhaustiveItems.size());
        Assert.assertEquals("title", linkExhaustiveItems.get(0).getName());
    }
}
