/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data;

import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.contenttype.ContentAttributeDefinition;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.external.ExternalizableDataProvider.ExternalizableDataStatus;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.values.SynchronizableRepeater;
import org.ametys.plugins.repository.data.holder.values.SynchronizableValue;
import org.ametys.plugins.repository.data.holder.values.SynchronizableValue.Mode;
import org.ametys.plugins.repository.data.holder.values.SynchronizationContext;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ContentDataHelper} test cases
 */
@NeedFeatures({
    "test/content-types-with-invert-relations",
    "test-data/content-test-externalizable-data-provider",
    
    "repository/repository.data.holder.helpers",
    "repository/resolver",
    "repository/provider.jackrabbit",
    "cms/factories",
    "cms/workflow.content"
})
public class ContentDataHelperTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    
    private static final String __SOURCE_CONTENT_TYPE_ID = "tests.invert.relations.source";
    private static final String __DESTINATION_CONTENT_TYPE_ID = "tests.invert.relations.destination";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ContentTypeExtensionPoint _contentTypeExtensionPoint;
    private static ContentDataHelper _contentDataHelper;
    
    private static View _view;
    private static SynchronizationContext _synchroContext;
    
    private ModifiableContent _source;
    private ModifiableContent _source2;
    private ModifiableContent _source3;
    private ModifiableContent _source4;
    private ModifiableContent _destination1;
    private ModifiableContent _destination2;
    private ModifiableContent _destination3;
    private ModifiableContent _destination4;

    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _contentDataHelper = (ContentDataHelper) manager.lookup(ContentDataHelper.ROLE);
    }
    
    public void prepareComponents() throws Exception
    {
        ContentType sourceContentType = _contentTypeExtensionPoint.getExtension(__SOURCE_CONTENT_TYPE_ID);
        _view = sourceContentType.getView("main");
        _synchroContext = SynchronizationContext.newInstance();
    }
    
    /**
     * Prepare components that are specific for each test
     */
    @Before
    public void prepareSpecificComponents()
    {
        if (_source != null)
        {
            _source.remove();
        }
        if (_source2 != null)
        {
            _source2.remove();
        }
        if (_source3 != null)
        {
            _source3.remove();
        }
        if (_source4 != null)
        {
            _source4.remove();
        }
        if (_destination1 != null)
        {
            _destination1.remove();
        }
        if (_destination2 != null)
        {
            _destination2.remove();
        }
        if (_destination3 != null)
        {
            _destination3.remove();
        }
        if (_destination4 != null)
        {
            _destination4.remove();
        }
        
        _source = _createContent("source", __SOURCE_CONTENT_TYPE_ID);
        _source2 = _createContent("source2", __SOURCE_CONTENT_TYPE_ID);
        _source3 = _createContent("source3", __SOURCE_CONTENT_TYPE_ID);
        _source4 = _createContent("source4", __SOURCE_CONTENT_TYPE_ID);
        _destination1 = _createContent("destination", __DESTINATION_CONTENT_TYPE_ID);
        _destination2 = _createContent("destination2", __DESTINATION_CONTENT_TYPE_ID);
        _destination3 = _createContent("destination3", __DESTINATION_CONTENT_TYPE_ID);
        _destination4 = _createContent("destination4", __DESTINATION_CONTENT_TYPE_ID);
    }
    
    /**
     * Tests the {@link ContentDataHelper#collectReferencedContents(ViewItemAccessor, ModifiableContent, Map, SynchronizationContext)} method with replacing single invert relation
     */
    @Test
    public void testCollectReferencedContentOnSingleSourceAttribute()
    {
        // Set the reference to the destination in the source
        _source.getComposite("composite", true)
               .getRepeater("repeater", true)
               .addEntry()
               .setValue("singleToMultiple", _destination1);
        
        _source.getComposite("composite", true)
               .getRepeater("repeater", true)
               .addEntry()
               .setValue("singleToMultiple", _destination4);

        // Remove a relation to destination, but add an entry with a new relation to the same destination
        Map<String, Object> values = Map.of(
                "composite", Map.of(
                        "repeater", List.of(
                                Map.of("singleToMultiple", _destination2),
                                Map.of("singleToMultiple", _destination2.getId()),
                                Map.of("singleToMultiple", _destination3.getId()),
                                Map.of("singleToMultiple", new SynchronizableValue(_destination4)))));
        
        Collection<ReferencedContents> allReferencedContents = _contentDataHelper.collectReferencedContents(_view, _source, values, _synchroContext);
        
        ReferencedContents referencedContents = _getReferencedContents(allReferencedContents, "composite/repeater/singleToSingle");
        Assert.assertTrue(referencedContents.getAddedContents().isEmpty());
        Assert.assertTrue(referencedContents.getRemovedContents().isEmpty());
        
        referencedContents = _getReferencedContents(allReferencedContents, "composite/multipleToSingle");
        Assert.assertTrue(referencedContents.getAddedContents().isEmpty());
        Assert.assertTrue(referencedContents.getRemovedContents().isEmpty());

        referencedContents = _getReferencedContents(allReferencedContents, "composite/multipleToMultiple");
        Assert.assertTrue(referencedContents.getAddedContents().isEmpty());
        Assert.assertTrue(referencedContents.getRemovedContents().isEmpty());
        
        referencedContents = _getReferencedContents(allReferencedContents, "composite/repeater/singleToMultiple");
        Set<ContentValue> addedContents = referencedContents.getAddedContents();
        Assert.assertEquals(2, addedContents.size());
        Assert.assertEquals(Set.of(_destination2.getId(), _destination3.getId()),
                            addedContents.stream().map(value -> value.getContentId()).collect(Collectors.toSet()));
        Set<ContentValue> removedContents = referencedContents.getRemovedContents();
        Assert.assertEquals(1, removedContents.size());
        Assert.assertEquals(_destination1.getId(), removedContents.iterator().next().getContentId());
        
        // test with externalizable value on non externalizable data
        SynchronizableValue syncValue = new SynchronizableValue(_destination2);
        syncValue.setExternalValue(_destination3);
        syncValue.setExternalizableStatus(ExternalizableDataStatus.EXTERNAL);
        values = Map.of(
                "composite", Map.of(
                        "repeater", List.of(
                                Map.of("singleToMultiple", syncValue))));
        
        allReferencedContents = _contentDataHelper.collectReferencedContents(_view, _source, values, _synchroContext);
        
        referencedContents = _getReferencedContents(allReferencedContents, "composite/repeater/singleToMultiple");
        addedContents = referencedContents.getAddedContents();
        Assert.assertEquals(1, addedContents.size());
        Assert.assertEquals(_destination2.getId(), addedContents.iterator().next().getContentId());
        removedContents = referencedContents.getRemovedContents();
        Assert.assertEquals(2, removedContents.size());
        Assert.assertEquals(Set.of(_destination1.getId(), _destination4.getId()),
                removedContents.stream().map(value -> value.getContentId()).collect(Collectors.toSet()));

        // test with SynchronizableRepeater in replace all mode with no mapping (meaning existing entries removed)
        SynchronizableRepeater repeater = SynchronizableRepeater.replaceAll(List.of(Map.of("singleToMultiple", _destination2),
                                                                                    Map.of("singleToMultiple", _destination3)), 
                                                                            Map.of());
        values = Map.of("composite", Map.of("repeater", repeater));
        
        allReferencedContents = _contentDataHelper.collectReferencedContents(_view, _source, values, _synchroContext);
        
        referencedContents = _getReferencedContents(allReferencedContents, "composite/repeater/singleToMultiple");
        addedContents = referencedContents.getAddedContents();
        Assert.assertEquals(2, addedContents.size());
        Assert.assertEquals(Set.of(_destination2.getId(), _destination3.getId()),
                addedContents.stream().map(value -> value.getContentId()).collect(Collectors.toSet()));
        removedContents = referencedContents.getRemovedContents();
        Assert.assertEquals(2, removedContents.size());
        Assert.assertEquals(Set.of(_destination1.getId(), _destination4.getId()),
                removedContents.stream().map(value -> value.getContentId()).collect(Collectors.toSet()));
        
        // test with SynchronizableRepeater in append mode with add of two entries
        repeater = SynchronizableRepeater.appendOrRemove(List.of(Map.of("singleToMultiple", _destination2),
                                                                 Map.of("singleToMultiple", _destination3)), 
                                                         Set.of());
        values = Map.of("composite", Map.of("repeater", repeater));
        
        allReferencedContents = _contentDataHelper.collectReferencedContents(_view, _source, values, _synchroContext);
        
        referencedContents = _getReferencedContents(allReferencedContents, "composite/repeater/singleToMultiple");
        addedContents = referencedContents.getAddedContents();
        Assert.assertEquals(2, addedContents.size());
        Assert.assertEquals(Set.of(_destination2.getId(), _destination3.getId()),
                addedContents.stream().map(value -> value.getContentId()).collect(Collectors.toSet()));
        Assert.assertTrue(referencedContents.getRemovedContents().isEmpty());
        
        // test with SynchronizableRepeater in append mode with removal of the first entry
        repeater = SynchronizableRepeater.appendOrRemove(List.of(), Set.of(1));
        values = Map.of("composite", Map.of("repeater", repeater));
        
        allReferencedContents = _contentDataHelper.collectReferencedContents(_view, _source, values, _synchroContext);
        
        referencedContents = _getReferencedContents(allReferencedContents,  "composite/repeater/singleToMultiple");
        Assert.assertTrue(referencedContents.getAddedContents().isEmpty());
        removedContents = referencedContents.getRemovedContents();
        Assert.assertEquals(1, removedContents.size());
        Assert.assertEquals(_destination1.getId(), removedContents.iterator().next().getContentId());
        
        // test with SynchronizableRepeater in replace mode with move
        repeater = SynchronizableRepeater.replace(List.of(Map.of("singleToMultiple", _destination2)), List.of(2));
        values = Map.of("composite", Map.of("repeater", repeater));
        
        allReferencedContents = _contentDataHelper.collectReferencedContents(_view, _source, values, _synchroContext);
        
        referencedContents = _getReferencedContents(allReferencedContents,  "composite/repeater/singleToMultiple");
        addedContents = referencedContents.getAddedContents();
        Assert.assertEquals(1, addedContents.size());
        Assert.assertEquals(_destination2.getId(), addedContents.iterator().next().getContentId());
        removedContents = referencedContents.getRemovedContents();
        Assert.assertEquals(1, removedContents.size());
        Assert.assertEquals(_destination4.getId(), removedContents.iterator().next().getContentId());
    }
    
    /**
     * Tests the {@link ContentDataHelper#collectReferencedContents(ViewItemAccessor, ModifiableContent, Map, SynchronizationContext)} method with replacing multiple invert relation
     */
    @Test
    public void testCollectReferencedContentOnMultipleSourceAttribute()
    {
        // Set the reference to the destination in the source
        _source.getComposite("composite", true)
               .setValue("multipleToMultiple", new ModifiableContent[] {_destination1, _destination4});
        
        // test in replace mode 
        Map<String, Object> values = Map.of(
                "composite", Map.of("multipleToMultiple", new ModifiableContent[] {_destination2, _destination3, _destination4}));
        
        Collection<ReferencedContents> allReferencedContents = _contentDataHelper.collectReferencedContents(_view, _source, values, _synchroContext);
        
        ReferencedContents referencedContents = _getReferencedContents(allReferencedContents, "composite/multipleToMultiple");
        Set<ContentValue> addedContents = referencedContents.getAddedContents();
        Assert.assertEquals(2, addedContents.size());
        Assert.assertEquals(Set.of(_destination2.getId(), _destination3.getId()),
                            addedContents.stream().map(value -> value.getContentId()).collect(Collectors.toSet()));
        Set<ContentValue> removedContents = referencedContents.getRemovedContents();
        Assert.assertEquals(1, removedContents.size());
        Assert.assertEquals(_destination1.getId(), removedContents.iterator().next().getContentId());
        
        // test in append mode
        SynchronizableValue synchronizableValue = new SynchronizableValue(new ModifiableContent[] {_destination2, _destination3});
        synchronizableValue.setMode(Mode.APPEND);
        values = Map.of("composite", Map.of("multipleToMultiple", synchronizableValue));
        
        allReferencedContents = _contentDataHelper.collectReferencedContents(_view, _source, values, _synchroContext);
        
        referencedContents = _getReferencedContents(allReferencedContents, "composite/multipleToMultiple");
        addedContents = referencedContents.getAddedContents();
        Assert.assertEquals(2, addedContents.size());
        Assert.assertEquals(Set.of(_destination2.getId(), _destination3.getId()),
                            addedContents.stream().map(value -> value.getContentId()).collect(Collectors.toSet()));
        Assert.assertTrue(referencedContents.getRemovedContents().isEmpty());
        
        // test in remove mode
        synchronizableValue = new SynchronizableValue(new ModifiableContent[] {_destination4});
        synchronizableValue.setMode(Mode.REMOVE);
        values = Map.of("composite", Map.of("multipleToMultiple", synchronizableValue));
        
        allReferencedContents = _contentDataHelper.collectReferencedContents(_view, _source, values, _synchroContext);
        
        referencedContents = _getReferencedContents(allReferencedContents, "composite/multipleToMultiple");
        Assert.assertTrue(referencedContents.getAddedContents().isEmpty());
        removedContents = referencedContents.getRemovedContents();
        Assert.assertEquals(1, removedContents.size());
        Assert.assertEquals(_destination4.getId(), removedContents.iterator().next().getContentId());
    }
    
    /**
     * Tests the {@link ContentDataHelper#collectReferencedContents(ViewItemAccessor, ModifiableContent, Map, SynchronizationContext)} method with replacing externalizable single invert relation
     */
    @Test
    public void testCollectReferencedContentOnExternalizableSourceData()
    {
        // Set the reference to the destination in the source
        _source.setValue("externalizableSingleToSingle", _destination1);
        
        // test with externalizable value
        SynchronizableValue syncValue = new SynchronizableValue(_destination2);
        syncValue.setExternalValue(_destination3);
        syncValue.setExternalizableStatus(ExternalizableDataStatus.EXTERNAL);
        Map<String, Object> values = Map.of("externalizableSingleToSingle", syncValue);
        
        Collection<ReferencedContents> allReferencedContents = _contentDataHelper.collectReferencedContents(_view, _source, values, _synchroContext);
        
        ReferencedContents referencedContents = _getReferencedContents(allReferencedContents, "externalizableSingleToSingle");
        Set<ContentValue> addedContents = referencedContents.getAddedContents();
        Assert.assertEquals(1, addedContents.size());
        Assert.assertEquals(_destination3.getId(), addedContents.iterator().next().getContentId());
        Set<ContentValue> removedContents = referencedContents.getRemovedContents();
        Assert.assertEquals(1, removedContents.size());
        Assert.assertEquals(_destination1.getId(), removedContents.iterator().next().getContentId());
    }
    
    /**
     * Tests the {@link ContentDataHelper#collectReferencedContents(ViewItemAccessor, ModifiableContent, Map, SynchronizationContext)} method with replacing multiple invert relation
     * Relation from multiple to single in a repeater. There is no third party content because entries will be added in the repeater
     */
    @Test
    public void testCollectReferencedContentWithThirdPartyContentsWithMultipleFromSingleInRepeater()
    {
        // Set the reference to the destination in the source
        _source2.getComposite("composite", true)
            .setValue("multipleToSingle", new ModifiableContent[] {_destination1, _destination2});
    
        _destination1.getComposite("composite", true)
                     .getRepeater("repeater", true)
                     .addEntry()
                     .setValue("singleFromMultiple", _source2);
        
        _destination2.getComposite("composite", true)
                     .getRepeater("repeater", true)
                     .addEntry()
                     .setValue("singleFromMultiple", _source2);

        // test in replace mode 
        Map<String, Object> values = Map.of("composite", Map.of("multipleToSingle", new ModifiableContent[] {_destination2, _destination3}));
        
        Collection<ReferencedContents> allReferencedContents = _contentDataHelper.collectReferencedContents(_view, _source, values, _synchroContext);
        ReferencedContents referencedContents = _getReferencedContents(allReferencedContents, "composite/multipleToSingle");
        Assert.assertTrue(referencedContents.getThirdPartyContents().isEmpty());
    }
    
    /**
     * Tests the {@link ContentDataHelper#collectReferencedContents(ViewItemAccessor, ModifiableContent, Map, SynchronizationContext)} method with replacing multiple invert relation
     * Relation from single to single data. The relation from the second source to the first destination has to be removed
     */
    @Test
    public void testCollectReferencedContentWithThirdPartyContentsWithSingleFromSingle()
    {
        // Set the reference to the destination in the source
        _source2.getComposite("composite", true)
            .setValue("singleToReallySingle", _destination1);
    
        _destination1.getComposite("composite", true)
                     .setValue("reallySingleFromSingle", _source2);

        // test in replace mode 
        Map<String, Object> values = Map.of("composite", Map.of("singleToReallySingle", _destination1));
        
        Collection<ReferencedContents> allReferencedContents = _contentDataHelper.collectReferencedContents(_view, _source, values, _synchroContext);
        ReferencedContents referencedContents = _getReferencedContents(allReferencedContents, "composite/singleToReallySingle");
        _checkThirdPartyContents(referencedContents.getThirdPartyContents(), Map.of(_destination1, _source2));
    }
    
    /**
     * Tests the {@link ContentDataHelper#collectReferencedContents(ViewItemAccessor, ModifiableContent, Map, SynchronizationContext)} method with replacing multiple invert relation
     * Relation from multiple to single. The relations from the second source to the destinations have to be removed
     */
    @Test
    public void testCollectReferencedContentWithThirdPartyContentsWithMultipleFromSingle()
    {
        // Set the reference to the destination in the source
        _source2.getComposite("composite", true)
            .setValue("multipleToReallySingle", new ModifiableContent[] {_destination1, _destination2});
    
        _destination1.getComposite("composite", true)
                     .setValue("reallySingleFromMultiple", _source2);
        
        _destination2.getComposite("composite", true)
                     .setValue("reallySingleFromMultiple", _source2);
        
        _destination3.getComposite("composite", true)
                     .setValue("reallySingleFromMultiple", _source3);

        // test in replace mode 
        Map<String, Object> values = Map.of("composite", Map.of("multipleToReallySingle", new ModifiableContent[] {_destination1, _destination2, _destination3}));
        
        Collection<ReferencedContents> allReferencedContents = _contentDataHelper.collectReferencedContents(_view, _source, values, _synchroContext);
        ReferencedContents referencedContents = _getReferencedContents(allReferencedContents, "composite/multipleToReallySingle");
        _checkThirdPartyContents(referencedContents.getThirdPartyContents(), Map.of(_destination1, _source2, _destination2, _source2, _destination3, _source3));
    }

    private ReferencedContents _getReferencedContents(Collection<ReferencedContents> allReferencedContents, String path)
    {
        for (ReferencedContents referencedContents : allReferencedContents)
        {
            ContentAttributeDefinition definition = referencedContents.getDefinition();
            if (definition.getPath().equals(path))
            {
                return referencedContents;
            }
        }
        
        return null;
    }
    
    /**
     * Tests the {@link ContentDataHelper#checkBrokenReferences(ModelAwareDataHolder, boolean)} method
     * The reference exists, the returned map is empty and fixing has no effect
     */
    @Test
    public void testCheckBrokenReferencesWithSingleExistentRef()
    {
        _source.getComposite("composite", true)
               .getRepeater("repeater", true)
               .addEntry()
               .setValue("singleToMultiple", _destination1);
        
        // Check broken references, without fixing
        Map<String, List<String>> brokenReferences = ContentDataHelper.checkBrokenReferences(_source, false);
        Assert.assertTrue(brokenReferences.isEmpty());
        Assert.assertEquals(_destination1.getId(), ((ContentValue) _source.getValue("composite/repeater[1]/singleToMultiple")).getContentId());
        
        // Check broken references, fixing if needed
        Map<String, List<String>> fixedReferences = ContentDataHelper.checkBrokenReferences(_source, true);
        Assert.assertTrue(fixedReferences.isEmpty());
        Assert.assertEquals(_destination1.getId(), ((ContentValue) _source.getValue("composite/repeater[1]/singleToMultiple")).getContentId());
    }
    
    /**
     * Tests the {@link ContentDataHelper#checkBrokenReferences(ModelAwareDataHolder, boolean)} method
     * The reference does not exist, the returned map contains the reference and fixing removes the reference
     */
    @Test
    public void testCheckBrokenReferencesWithSingleNonExistentRef()
    {
        _source.getComposite("composite", true)
               .getRepeater("repeater", true)
               .addEntry()
               .setValue("singleToMultiple", "content://non-existent-content");
        
        // Check broken references, without fixing
        Map<String, List<String>> brokenReferences = ContentDataHelper.checkBrokenReferences(_source, false);
        Assert.assertEquals(1, brokenReferences.size());
        Assert.assertEquals(1, brokenReferences.get("composite/repeater[1]/singleToMultiple").size());
        Assert.assertEquals("content://non-existent-content", brokenReferences.get("composite/repeater[1]/singleToMultiple").get(0));
        Assert.assertEquals("content://non-existent-content", ((ContentValue) _source.getValue("composite/repeater[1]/singleToMultiple")).getContentId());
        
        // Check broken references, fixing if needed
        Map<String, List<String>> fixedReferences = ContentDataHelper.checkBrokenReferences(_source, true);
        Assert.assertEquals(1, fixedReferences.size());
        Assert.assertEquals(1, fixedReferences.get("composite/repeater[1]/singleToMultiple").size());
        Assert.assertEquals("content://non-existent-content", fixedReferences.get("composite/repeater[1]/singleToMultiple").get(0));
        Assert.assertFalse(_source.hasValue("composite/repeater[1]/singleToMultiple"));
    }
    
    /**
     * Tests the {@link ContentDataHelper#checkBrokenReferences(ModelAwareDataHolder, boolean)} method
     * References on a multiple data, some of the references exist, others don't.
     * The returned map contains the non existent references and fixing removes them, keeping the existent ones
     */
    @Test
    public void testCheckBrokenReferencesWithMultipleRef()
    {
        _source.getComposite("composite", true)
               .setValue("multipleToMultiple", new String[] {_destination1.getId(), "content://non-existent-content", "content://other-non-existent-content", _destination2.getId()});
        
        // Check broken references, without fixing
        Map<String, List<String>> brokenReferences = ContentDataHelper.checkBrokenReferences(_source, false);
        Assert.assertEquals(1, brokenReferences.size());
        Assert.assertEquals(2, brokenReferences.get("composite/multipleToMultiple").size());
        Assert.assertEquals("content://non-existent-content", brokenReferences.get("composite/multipleToMultiple").get(0));
        Assert.assertEquals("content://other-non-existent-content", brokenReferences.get("composite/multipleToMultiple").get(1));
        ContentValue[] value = _source.getValue("composite/multipleToMultiple");
        Assert.assertEquals(4, value.length);
        Assert.assertEquals(_destination1.getId(), value[0].getContentId());
        Assert.assertEquals("content://non-existent-content", value[1].getContentId());
        Assert.assertEquals("content://other-non-existent-content", value[2].getContentId());
        Assert.assertEquals(_destination2.getId(), value[3].getContentId());
        
        // Check broken references, fixing if needed
        Map<String, List<String>> fixedReferences = ContentDataHelper.checkBrokenReferences(_source, true);
        Assert.assertEquals(1, fixedReferences.size());
        Assert.assertEquals(2, fixedReferences.get("composite/multipleToMultiple").size());
        Assert.assertEquals("content://non-existent-content", fixedReferences.get("composite/multipleToMultiple").get(0));
        Assert.assertEquals("content://other-non-existent-content", fixedReferences.get("composite/multipleToMultiple").get(1));
        ContentValue[] fixedValue = _source.getValue("composite/multipleToMultiple");
        Assert.assertEquals(2, fixedValue.length);
        Assert.assertEquals(_destination1.getId(), fixedValue[0].getContentId());
        Assert.assertEquals(_destination2.getId(), fixedValue[1].getContentId());
    }
    
    /**
     * Tests the {@link ContentDataHelper#checkBrokenInvertRelations(Content, boolean)} method
     * Invert relation is no broken, the returned map is empty and fixing has no effect
     */
    @Test
    public void testCheckBrokenInvertRelationsWithSinlgeNotBrokenRef()
    {
        _source.getComposite("composite", true)
               .setValue("singleToReallySingle", _destination1);
        
        _destination1.getComposite("composite", true)
                     .setValue("reallySingleFromSingle", _source);
        
        // Check broken relations, without fixing
        Collection<ReferencedContents> brokenInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, false);
        Assert.assertTrue(brokenInvertRelations.isEmpty());
        
        Assert.assertEquals(_source.getId(), ((ContentValue) _destination1.getValue("composite/reallySingleFromSingle")).getContentId());
        
        // Check broken relations, fixing if needed
        Collection<ReferencedContents> fixedInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, true);
        Assert.assertTrue(fixedInvertRelations.isEmpty());
        
        Assert.assertEquals(_source.getId(), ((ContentValue) _destination1.getValue("composite/reallySingleFromSingle")).getContentId());
    }
    
    /**
     * Tests the {@link ContentDataHelper#checkBrokenInvertRelations(Content, boolean)} method
     * The destination attribute is single, value is empty
     * The returned map contains the destination and fixing will set the relation on the destination
     */
    @Test
    public void testCheckBrokenInvertRelationsWithSinlgeDestEmpty()
    {
        _source.getComposite("composite", true)
               .setValue("singleToReallySingle", _destination1);
        
        _destination1.getComposite("composite", true)
                     .setValue("reallySingleFromSingle", null);
        
        // Check broken relations, without fixing
        Collection<ReferencedContents> brokenInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, false);
        Assert.assertEquals(1, brokenInvertRelations.size());
        ReferencedContents brokenInvertRelation = brokenInvertRelations.iterator().next();
        Assert.assertEquals("composite/singleToReallySingle", brokenInvertRelation.getDefinition().getPath());
        _checkAddedContents(brokenInvertRelation.getAddedContents(), _destination1);
        Assert.assertTrue(brokenInvertRelation.getThirdPartyContents().isEmpty());
        
        Assert.assertFalse(_destination1.hasValue("composite/reallySingleFromSingle"));
        
        // Check broken relations, fixing if needed
        Collection<ReferencedContents> fixedInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, true);
        Assert.assertEquals(1, fixedInvertRelations.size());
        ReferencedContents fixedInvertRelation = fixedInvertRelations.iterator().next();
        Assert.assertEquals("composite/singleToReallySingle", fixedInvertRelation.getDefinition().getPath());
        _checkAddedContents(fixedInvertRelation.getAddedContents(), _destination1);
        Assert.assertTrue(fixedInvertRelation.getThirdPartyContents().isEmpty());
        
        Assert.assertEquals(_source.getId(), ((ContentValue) _destination1.getValue("composite/reallySingleFromSingle")).getContentId());
    }
    
    /**
     * Tests the {@link ContentDataHelper#checkBrokenInvertRelations(Content, boolean)} method
     * The destination attribute is single, value points another source, that is empty
     * The returned map contains the destination and fixing will set the relation on the destination
     */
    @Test
    public void testCheckBrokenInvertRelationsWithSinlgeDestOnOtherSource()
    {
        _source.getComposite("composite", true)
               .setValue("singleToReallySingle", _destination1);
        
        _destination1.getComposite("composite", true)
                     .setValue("reallySingleFromSingle", _source2);
        
        // Check broken relations, without fixing
        Collection<ReferencedContents> brokenInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, false);
        Assert.assertEquals(1, brokenInvertRelations.size());
        ReferencedContents brokenInvertRelation = brokenInvertRelations.iterator().next();
        Assert.assertEquals("composite/singleToReallySingle", brokenInvertRelation.getDefinition().getPath());
        _checkAddedContents(brokenInvertRelation.getAddedContents(), _destination1);
        Assert.assertTrue(brokenInvertRelation.getThirdPartyContents().isEmpty());
        
        Assert.assertEquals(_source2.getId(), ((ContentValue) _destination1.getValue("composite/reallySingleFromSingle")).getContentId());
        Assert.assertFalse(_source2.hasValue("composite/singleToReallySingle"));
        
        // Check broken relations, fixing if needed
        Collection<ReferencedContents> fixedInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, true);
        Assert.assertEquals(1, fixedInvertRelations.size());
        ReferencedContents fixedInvertRelation = fixedInvertRelations.iterator().next();
        Assert.assertEquals("composite/singleToReallySingle", fixedInvertRelation.getDefinition().getPath());
        _checkAddedContents(fixedInvertRelation.getAddedContents(), _destination1);
        Assert.assertTrue(fixedInvertRelation.getThirdPartyContents().isEmpty());
        
        Assert.assertEquals(_source.getId(), ((ContentValue) _destination1.getValue("composite/reallySingleFromSingle")).getContentId());
        Assert.assertFalse(_source2.hasValue("composite/singleToReallySingle"));
    }
    
    /**
     * Tests the {@link ContentDataHelper#checkBrokenInvertRelations(Content, boolean)} method
     * The destination attribute is single, value points another source, that points the destination in return
     * The returned map contains the destination and the other source, fixing will set the relation on the destination and remove the other source's one
     */
    @Test
    public void testCheckBrokenInvertRelationsWithSinlgeDestOnOtherSourcePointingBack()
    {
        _source.getComposite("composite", true)
               .setValue("singleToReallySingle", _destination1);
        
        _destination1.getComposite("composite", true)
                     .setValue("reallySingleFromSingle", _source2);
        
        _source2.getComposite("composite", true)
                .setValue("singleToReallySingle", _destination1);
        
        // Check broken relations, without fixing
        Collection<ReferencedContents> brokenInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, false);
        Assert.assertEquals(1, brokenInvertRelations.size());
        ReferencedContents brokenInvertRelation = brokenInvertRelations.iterator().next();
        Assert.assertEquals("composite/singleToReallySingle", brokenInvertRelation.getDefinition().getPath());
        _checkAddedContents(brokenInvertRelation.getAddedContents(), _destination1);
        _checkThirdPartyContents(brokenInvertRelation.getThirdPartyContents(), Map.of(_destination1, _source2));
        
        Assert.assertEquals(_source2.getId(), ((ContentValue) _destination1.getValue("composite/reallySingleFromSingle")).getContentId());
        Assert.assertEquals(_destination1.getId(), ((ContentValue) _source2.getValue("composite/singleToReallySingle")).getContentId());
        
        // Check broken relations, fixing if needed
        Collection<ReferencedContents> fixedInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, true);
        Assert.assertEquals(1, fixedInvertRelations.size());
        ReferencedContents fixedInvertRelation = fixedInvertRelations.iterator().next();
        Assert.assertEquals("composite/singleToReallySingle", fixedInvertRelation.getDefinition().getPath());
        _checkAddedContents(fixedInvertRelation.getAddedContents(), _destination1);
        _checkThirdPartyContents(fixedInvertRelation.getThirdPartyContents(), Map.of(_destination1, _source2));
        
        Assert.assertEquals(_source.getId(), ((ContentValue) _destination1.getValue("composite/reallySingleFromSingle")).getContentId());
        Assert.assertFalse(_source2.hasValue("composite/singleToReallySingle"));
    }
    
    /**
     * Tests the {@link ContentDataHelper#checkBrokenInvertRelations(Content, boolean)} method
     * The destination attribute is single, values point another sources, that points the destination in return, or not
     * The returned map contains the destinations and the other sources, fixing will set the relation on the destination and remove the other sources' ones
     */
    @Test
    public void testCheckBrokenInvertRelationsWithSinlgeDestOnOtherMultipleSource()
    {
        _source.getComposite("composite", true)
               .setValue("multipleToReallySingle", new Content[] {_destination1, _destination2, _destination3, _destination4});
        
        _destination1.getComposite("composite", true)
                     .setValue("reallySingleFromMultiple", _source2);
        _destination2.getComposite("composite", true)
                     .setValue("reallySingleFromMultiple", _source2);
        _source2.getComposite("composite", true)
                .setValue("multipleToReallySingle", new Content[] {_destination1, _destination2});
        
        _destination3.getComposite("composite", true)
                     .setValue("reallySingleFromMultiple", _source3);
        _source3.getComposite("composite", true)
                .setValue("multipleToReallySingle", new Content[] {_destination3});
        
        _destination4.getComposite("composite", true)
                     .setValue("reallySingleFromMultiple", _source4);
        
        // Check broken relations, without fixing
        Collection<ReferencedContents> brokenInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, false);
        Assert.assertEquals(1, brokenInvertRelations.size());
        ReferencedContents brokenInvertRelation = brokenInvertRelations.iterator().next();
        Assert.assertEquals("composite/multipleToReallySingle", brokenInvertRelation.getDefinition().getPath());
        _checkAddedContents(brokenInvertRelation.getAddedContents(), _destination1, _destination2, _destination3, _destination4);
        _checkThirdPartyContents(brokenInvertRelation.getThirdPartyContents(), Map.of(_destination1, _source2, _destination2, _source2, _destination3, _source3));
        
        Assert.assertEquals(_source2.getId(), ((ContentValue) _destination1.getValue("composite/reallySingleFromMultiple")).getContentId());
        Assert.assertEquals(_source2.getId(), ((ContentValue) _destination2.getValue("composite/reallySingleFromMultiple")).getContentId());
        Assert.assertEquals(_destination1.getId(), ((ContentValue[]) _source2.getValue("composite/multipleToReallySingle"))[0].getContentId());
        Assert.assertEquals(_destination2.getId(), ((ContentValue[]) _source2.getValue("composite/multipleToReallySingle"))[1].getContentId());
        
        Assert.assertEquals(_source3.getId(), ((ContentValue) _destination3.getValue("composite/reallySingleFromMultiple")).getContentId());
        Assert.assertEquals(_destination3.getId(), ((ContentValue[]) _source3.getValue("composite/multipleToReallySingle"))[0].getContentId());
        
        Assert.assertEquals(_source4.getId(), ((ContentValue) _destination4.getValue("composite/reallySingleFromMultiple")).getContentId());
        Assert.assertFalse(_source4.hasValue("composite/multipleToReallySingle"));
        
        // Check broken relations, fixing if needed
        Collection<ReferencedContents> fixedInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, true);
        Assert.assertEquals(1, fixedInvertRelations.size());
        ReferencedContents fixedInvertRelation = fixedInvertRelations.iterator().next();
        Assert.assertEquals("composite/multipleToReallySingle", fixedInvertRelation.getDefinition().getPath());
        _checkAddedContents(fixedInvertRelation.getAddedContents(), _destination1, _destination2, _destination3, _destination4);
        _checkThirdPartyContents(fixedInvertRelation.getThirdPartyContents(), Map.of(_destination1, _source2, _destination2, _source2, _destination3, _source3));
        
        Assert.assertEquals(_source.getId(), ((ContentValue) _destination1.getValue("composite/reallySingleFromMultiple")).getContentId());
        Assert.assertEquals(_source.getId(), ((ContentValue) _destination2.getValue("composite/reallySingleFromMultiple")).getContentId());
        Assert.assertEquals(_source.getId(), ((ContentValue) _destination3.getValue("composite/reallySingleFromMultiple")).getContentId());
        Assert.assertEquals(_source.getId(), ((ContentValue) _destination4.getValue("composite/reallySingleFromMultiple")).getContentId());
        Assert.assertFalse(_source2.hasValue("composite/multipleToReallySingle"));
        Assert.assertFalse(_source3.hasValue("composite/multipleToReallySingle"));
        Assert.assertFalse(_source4.hasValue("composite/multipleToReallySingle"));
    }
    
    /**
     * Tests the {@link ContentDataHelper#checkBrokenInvertRelations(Content, boolean)} method
     * The destination attribute is in a repeater, the 1st entry points another source
     * The returned map contains the destination and fixing will add an entry in the destination repeater
     */
    @Test
    public void testCheckBrokenInvertRelationsWithSinlgeDestInRepeater()
    {
        _source.getComposite("composite", true)
               .getRepeater("repeater", true)
               .addEntry()
               .setValue("singleToSingle", _destination1);
        
        _destination1.getComposite("composite", true)
                     .getRepeater("repeater", true)
                     .addEntry()
                     .setValue("singleFromSingle", _source2);
        
        // Check broken relations, without fixing
        Collection<ReferencedContents> brokenInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, false);
        Assert.assertEquals(1, brokenInvertRelations.size());
        ReferencedContents brokenInvertRelation = brokenInvertRelations.iterator().next();
        Assert.assertEquals("composite/repeater/singleToSingle", brokenInvertRelation.getDefinition().getPath());
        _checkAddedContents(brokenInvertRelation.getAddedContents(), _destination1);
        Assert.assertTrue(brokenInvertRelation.getThirdPartyContents().isEmpty());
        
        Assert.assertEquals(1, _destination1.getRepeater("composite/repeater").getSize());
        Assert.assertEquals(_source2.getId(), ((ContentValue) _destination1.getValue("composite/repeater[1]/singleFromSingle")).getContentId());
        
        // Check broken relations, fixing if needed
        Collection<ReferencedContents> fixedInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, true);
        Assert.assertEquals(1, fixedInvertRelations.size());
        ReferencedContents fixedInvertRelation = fixedInvertRelations.iterator().next();
        Assert.assertEquals("composite/repeater/singleToSingle", fixedInvertRelation.getDefinition().getPath());
        _checkAddedContents(fixedInvertRelation.getAddedContents(), _destination1);
        Assert.assertTrue(fixedInvertRelation.getThirdPartyContents().isEmpty());
        
        Assert.assertEquals(2, _destination1.getRepeater("composite/repeater").getSize());
        Assert.assertEquals(_source2.getId(), ((ContentValue) _destination1.getValue("composite/repeater[1]/singleFromSingle")).getContentId());
        Assert.assertEquals(_source.getId(), ((ContentValue) _destination1.getValue("composite/repeater[2]/singleFromSingle")).getContentId());
    }
    
    /**
     * Tests the {@link ContentDataHelper#checkBrokenInvertRelations(Content, boolean)} method
     * The destination attribute is multiple, all values contain the source
     * The returned map is empty and fixing has no effect
     */
    @Test
    public void testCheckBrokenInvertRelationsWithMultipleNotBrokenRef()
    {
        _source.getComposite("composite", true)
               .setValue("multipleToMultiple", new Content[] {_destination1, _destination2, _destination3});
        
        _destination1.getComposite("composite", true)
                     .setValue("multipleFromMultiple", new Content[] {_source, _source2});
        
        _destination2.getComposite("composite", true)
                     .setValue("multipleFromMultiple", new Content[] {_source});
        
        _destination3.getComposite("composite", true)
                     .setValue("multipleFromMultiple", new Content[] {_source});
        
        // Check broken relations, without fixing
        Collection<ReferencedContents> brokenInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, false);
        Assert.assertTrue(brokenInvertRelations.isEmpty());
        
        ContentValue[] dest1Value = _destination1.getValue("composite/multipleFromMultiple");
        Assert.assertEquals(2, dest1Value.length);
        Assert.assertEquals(_source.getId(), dest1Value[0].getContentId());
        Assert.assertEquals(_source2.getId(), dest1Value[1].getContentId());
        
        ContentValue[] dest2Value = _destination2.getValue("composite/multipleFromMultiple");
        Assert.assertEquals(1, dest2Value.length);
        Assert.assertEquals(_source.getId(), dest2Value[0].getContentId());
        
        ContentValue[] dest3Value = _destination3.getValue("composite/multipleFromMultiple");
        Assert.assertEquals(1, dest3Value.length);
        Assert.assertEquals(_source.getId(), dest3Value[0].getContentId());
        
        // Check broken relations, fixing if needed
        Collection<ReferencedContents> fixedInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, true);
        Assert.assertTrue(fixedInvertRelations.isEmpty());
        
        ContentValue[] dest1FixedValue = _destination1.getValue("composite/multipleFromMultiple");
        Assert.assertEquals(2, dest1FixedValue.length);
        Assert.assertEquals(_source.getId(), dest1FixedValue[0].getContentId());
        Assert.assertEquals(_source2.getId(), dest1FixedValue[1].getContentId());
        
        ContentValue[] dest2FixedValue = _destination2.getValue("composite/multipleFromMultiple");
        Assert.assertEquals(1, dest2FixedValue.length);
        Assert.assertEquals(_source.getId(), dest2FixedValue[0].getContentId());
        
        ContentValue[] dest3FixedValue = _destination3.getValue("composite/multipleFromMultiple");
        Assert.assertEquals(1, dest3FixedValue.length);
        Assert.assertEquals(_source.getId(), dest3FixedValue[0].getContentId());
    }
    
    /**
     * Tests the {@link ContentDataHelper#checkBrokenInvertRelations(Content, boolean)} method
     * The destination attribute is multiple, some values don't contain the source
     * The returned map contains these values and fixing will add the source in multiple values
     */
    @Test
    public void testCheckBrokenInvertRelationsWithMultipleWithBrokenRef()
    {
        _source.getComposite("composite", true)
               .setValue("multipleToMultiple", new Content[] {_destination1, _destination2, _destination3});
        
        _destination1.getComposite("composite", true)
                     .setValue("multipleFromMultiple", new Content[] {_source2});
        
        _destination2.getComposite("composite", true)
                     .setValue("multipleFromMultiple", new Content[0]);
        
        _destination3.getComposite("composite", true)
                     .setValue("multipleFromMultiple", new Content[] {_source});
        
        // Check broken relations, without fixing
        Collection<ReferencedContents> brokenInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, false);
        Assert.assertEquals(1, brokenInvertRelations.size());
        ReferencedContents brokenInvertRelation = brokenInvertRelations.iterator().next();
        Assert.assertEquals("composite/multipleToMultiple", brokenInvertRelation.getDefinition().getPath());
        _checkAddedContents(brokenInvertRelation.getAddedContents(), _destination1, _destination2);
        Assert.assertTrue(brokenInvertRelation.getThirdPartyContents().isEmpty());
        
        ContentValue[] dest1Value = _destination1.getValue("composite/multipleFromMultiple");
        Assert.assertEquals(1, dest1Value.length);
        Assert.assertEquals(_source2.getId(), dest1Value[0].getContentId());
        
        ContentValue[] dest2Value = _destination2.getValue("composite/multipleFromMultiple");
        Assert.assertEquals(0, dest2Value.length);
        
        ContentValue[] dest3Value = _destination3.getValue("composite/multipleFromMultiple");
        Assert.assertEquals(1, dest3Value.length);
        Assert.assertEquals(_source.getId(), dest3Value[0].getContentId());
        
        // Check broken relations, fixing if needed
        Collection<ReferencedContents> fixedInvertRelations = ContentDataHelper.checkBrokenInvertRelations(_source, true);
        Assert.assertEquals(1, fixedInvertRelations.size());
        ReferencedContents fixedInvertRelation = fixedInvertRelations.iterator().next();
        Assert.assertEquals("composite/multipleToMultiple", fixedInvertRelation.getDefinition().getPath());
        _checkAddedContents(fixedInvertRelation.getAddedContents(), _destination1, _destination2);
        Assert.assertTrue(fixedInvertRelation.getThirdPartyContents().isEmpty());
        
        ContentValue[] dest1FixedValue = _destination1.getValue("composite/multipleFromMultiple");
        Assert.assertEquals(2, dest1FixedValue.length);
        Assert.assertEquals(_source2.getId(), dest1FixedValue[0].getContentId());
        Assert.assertEquals(_source.getId(), dest1FixedValue[1].getContentId());
        
        ContentValue[] dest2FixedValue = _destination2.getValue("composite/multipleFromMultiple");
        Assert.assertEquals(1, dest2FixedValue.length);
        Assert.assertEquals(_source.getId(), dest2FixedValue[0].getContentId());
        
        ContentValue[] dest3FixedValue = _destination3.getValue("composite/multipleFromMultiple");
        Assert.assertEquals(1, dest3FixedValue.length);
        Assert.assertEquals(_source.getId(), dest3FixedValue[0].getContentId());
    }
    
    private void _checkAddedContents(Set<ContentValue> addedContents, Content... expectedAddedContents)
    {
        Assert.assertEquals(expectedAddedContents.length, addedContents.size());
        
        for (Content expectedAddedContent : expectedAddedContents)
        {
            boolean contentFound = false;
            for (ContentValue addedContent : addedContents)
            {
                if (expectedAddedContent.getId().equals(addedContent.getContentId()))
                {
                    contentFound = true;
                    break;
                }
            }
            Assert.assertTrue(contentFound);
        }
    }
    
    private void _checkThirdPartyContents(Map<ContentValue, ContentValue> thirdPartyContents, Map<Content, Content> expectedThirdPartyContents)
    {
        Assert.assertEquals(expectedThirdPartyContents.size(), thirdPartyContents.size());
        
        for (Map.Entry<Content, Content> expectedThirdPartyContent : expectedThirdPartyContents.entrySet())
        {
            boolean pairFound = false;
            for (Map.Entry<ContentValue, ContentValue> thirdPartyContent : thirdPartyContents.entrySet())
            {
                if (expectedThirdPartyContent.getKey().getId().equals(thirdPartyContent.getKey().getContentId()) && expectedThirdPartyContent.getValue().getId().equals(thirdPartyContent.getValue().getContentId()))
                {
                    pairFound = true;
                    break;
                }
            }
            Assert.assertTrue(pairFound);
        }
    }
    
    private ModifiableContent _createContent(String title, String contentTypeId)
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        
        ModifiableContent content = root.createChild(title, "ametys:defaultContent");

        content.setTypes(new String[]{contentTypeId});
        content.setTitle(title, Locale.FRANCE);
        content.setLanguage("fr");
        content.setCreationDate(ZonedDateTime.now());
        content.setCreator(new UserIdentity("enaour", "wallace"));
        content.setLastModified(ZonedDateTime.now());
        content.setLastContributor(new UserIdentity("bfraisse", "wallace"));
        
        content.saveChanges();
        return content;
    }
}
