/*
 *  Copyright 2024 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.ametysobject;

import java.time.ZonedDateTime;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.external.ExternalizableDataProvider.ExternalizableDataStatus;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeaterEntry;
import org.ametys.plugins.repository.data.holder.values.SynchronizableRepeater;
import org.ametys.plugins.repository.data.holder.values.SynchronizableValue;
import org.ametys.plugins.repository.data.holder.values.SynchronizationContext;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.disableconditions.DisableConditionsEvaluator;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link AmetysObjectDisableConditionsEvaluator}'s test cases
 */
@NeedFeatures({
    "data.disableconditions/content-types-disable-conditions-tests",
    "data.disableconditions/test-externalizable-data-provider",
    
    "cms/org.ametys.cms.ametys.object.disable.conditions",
    "repository/resolver",
    "repository/provider.jackrabbit",
    "cms/factories",
    "cms/workflow.content"
})
public class AmetysObjectDisableConditionsEvaluatorTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __CONTENT_TYPE_ID = "tests.disable.conditions";
    private static final String __LINKED_CONTENT_TYPE_ID = "tests.disable.conditions.linked";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ContentTypeExtensionPoint _contentTypeExtensionPoint;
    private static AmetysObjectDisableConditionsEvaluator _evaluator;
    
    private static ContentType _contentType;
    private static Content _contentBooleanFalse;
    private static Content _contentBooleanTrue;
    private static Content _linkedContentBooleanFalse;
    private static Content _linkedContentBooleanTrue;
    
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
        _evaluator = (AmetysObjectDisableConditionsEvaluator) manager.lookup(AmetysObjectDisableConditionsEvaluator.ROLE);
    }
    
    public void prepareComponents() throws Exception
    {
        WithCmsFeaturesTestCase.super.prepareComponents();
        
        _contentType = _contentTypeExtensionPoint.getExtension(__CONTENT_TYPE_ID);
        
        _linkedContentBooleanTrue = _createLinkedContent("linkedContentBooleanTrue", true);
        _linkedContentBooleanFalse = _createLinkedContent("linkedContentBooleanFalse", false);
        _contentBooleanTrue = _createContentWithLinkedContent("contentBooleanTrue", true, _linkedContentBooleanTrue);
        _contentBooleanFalse = _createContentWithLinkedContent("contentBooleanFalse", false, _linkedContentBooleanFalse);
    }

    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * The attribute referenced by the condition is on content type's root, the evaluated attribute is on content type's root too
     */
    @Test
    public void testEvaluateDisableConditionsReferencedAtRootEvaluatedAtRoot()
    {
        String dataPath = "withCondition";
        Optional<String> oldDataPath = Optional.empty();
        ModelItem modelItem = _contentType.getModelItem(dataPath);
        Map<String, Object> contextualParameters = new HashMap<>();
        
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of("atRoot", false), _contentBooleanTrue, contextualParameters));
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of(), _contentBooleanTrue, contextualParameters));
        
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of("atRoot", true), _contentBooleanFalse, contextualParameters));
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of(), _contentBooleanFalse, contextualParameters));
    }

    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * The attribute referenced by the condition is on content type's root, the evaluated attribute is in a repeater
     */
    @Test
    public void testEvaluateDisableConditionsReferencedAtRootEvaluatedInRepeater()
    {
        String dataPath = "composite/repeater[2]/withConditionOnRoot";
        Optional<String> oldDataPath = Optional.empty();
        ModelItem modelItem = _contentType.getModelItem("composite/repeater/withConditionOnRoot");
        Map<String, Object> contextualParameters = new HashMap<>();
        
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of("atRoot", false), _contentBooleanTrue, contextualParameters));
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of(), _contentBooleanTrue, contextualParameters));
        
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of("atRoot", true), _contentBooleanFalse, contextualParameters));
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of(), _contentBooleanFalse, contextualParameters));
    }

    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * The attribute referenced by the condition is in a repeater, the evaluated attribute is in the same repeater
     */
    @Test
    public void testEvaluateDisableConditionsReferencedInRepeaterEvaluatedInRepeater()
    {
        Content content = _contentBooleanTrue;
        ModelItem modelItem = _contentType.getModelItem("composite/repeater/withConditionInEntry");
        Map<String, Object> values = Map.of("composite", Map.of("repeater", List.of(Map.of("inRepeater", false), Map.of("inRepeater", true), Map.of("inRepeater", true))));
        Map<String, Object> contextualParameters = new HashMap<>();
        Optional<String> oldDataPath = Optional.empty();
        
        String dataPath = "composite/repeater[1]/withConditionInEntry";
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of(), content, contextualParameters));
        
        dataPath = "composite/repeater[2]/withConditionInEntry";
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of(), content, contextualParameters));
        
        dataPath = "composite/repeater[3]/withConditionInEntry";
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of(), content, contextualParameters));
    }

    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * The attribute referenced by the condition is in a linked content on content type's root, the evaluated attribute is on content type's root
     */
    @Test
    public void testEvaluateDisableConditionsReferencedLinkedAtRootEvaluatedAtRoot()
    {
        String dataPath = "withConditionOnLinkedContent";
        Optional<String> oldDataPath = Optional.empty();
        ModelItem modelItem = _contentType.getModelItem(dataPath);
        Map<String, Object> contextualParameters = new HashMap<>();
        
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of("linkedContent", _linkedContentBooleanFalse), _contentBooleanTrue, contextualParameters));
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of(), _contentBooleanTrue, contextualParameters));
        
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of("linkedContent", _linkedContentBooleanTrue), _contentBooleanFalse, contextualParameters));
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of(), _contentBooleanFalse, contextualParameters));
    }

    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * The attribute referenced by the condition is in a linked content on content type's root, the evaluated attribute is in a repeater
     */
    @Test
    public void testEvaluateDisableConditionsReferencedLinkedAtRootEvaluatedInRepeater()
    {
        String dataPath = "composite/repeater[2]/withConditionOnLinkedContentAtRoot";
        Optional<String> oldDataPath = Optional.empty();
        ModelItem modelItem = _contentType.getModelItem("composite/repeater/withConditionOnLinkedContentAtRoot");
        Map<String, Object> contextualParameters = new HashMap<>();
        
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of("linkedContent", _linkedContentBooleanFalse), _contentBooleanTrue, contextualParameters));
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of(), _contentBooleanTrue, contextualParameters));
        
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of("linkedContent", _linkedContentBooleanTrue), _contentBooleanFalse, contextualParameters));
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of(), _contentBooleanFalse, contextualParameters));
    }

    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * The attribute referenced by the condition is in a linked content in the repeater, the evaluated attribute is in the same repeater
     */
    @Test
    public void testEvaluateDisableConditionsReferencedLinkedInRepeaterEvaluatedInRepeater()
    {
        Content content = _contentBooleanTrue;
        ModelItem modelItem = _contentType.getModelItem("composite/repeater/withConditionOnLinkedContentInEntry");
        Map<String, Object> values = Map.of("composite", Map.of("repeater", List.of(Map.of("linkedContent", _linkedContentBooleanFalse), Map.of("linkedContent", _linkedContentBooleanTrue))));
        Map<String, Object> contextualParameters = new HashMap<>();
        
        String dataPath = "composite/repeater[1]/withConditionOnLinkedContentInEntry";
        Optional<String> oldDataPath = Optional.empty();
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of(), content, contextualParameters));
        
        dataPath = "composite/repeater[2]/withConditionOnLinkedContentInEntry";
        oldDataPath = Optional.empty();
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of(), content, contextualParameters));
        
        dataPath = "composite/repeater[1]/withConditionOnLinkedContentInEntry";
        oldDataPath = Optional.of("composite/repeater[2]/withConditionOnLinkedContentInEntry");
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of(), content, contextualParameters));
    }

    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * The attribute referenced by the condition is on content type's root, the evaluated attribute is on content type's root too
     * The values in the Map are {@link SynchronizableValue}
     */
    @Test
    public void testEvaluateDisableConditionsReferencedAtRootEvaluatedAtRootWithSynchronizableValues()
    {
        String dataPath = "withCondition";
        Optional<String> oldDataPath = Optional.empty();
        ModelItem modelItem = _contentType.getModelItem(dataPath);
        
        Map<String, Object> contextualParameters = new HashMap<>();
        contextualParameters.put(AmetysObjectDisableConditionsEvaluator.SYNCHRONIZATION_CONTEXT_PARAMETER_KEY, SynchronizationContext.newInstance());
        
        SynchronizableValue synchronizableValue = new SynchronizableValue(true); // Set local value
        synchronizableValue.setExternalValue(false); // Set the external value
        
        synchronizableValue.setExternalizableStatus(ExternalizableDataStatus.EXTERNAL); // Use external false value
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of("atRoot", synchronizableValue), _contentBooleanTrue, contextualParameters));
        
        synchronizableValue.setExternalizableStatus(ExternalizableDataStatus.LOCAL); // Use local true value
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, Map.of("atRoot", synchronizableValue), _contentBooleanFalse, contextualParameters));
    }
    
    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * The attribute referenced by the condition is in a repeater, the evaluated attribute is in the same repeater
     * The values in the Map are {@link SynchronizableRepeater}
     */
    @Test
    public void testEvaluateDisableConditionsReferencedInRepeaterEvaluatedInRepeaterWithSynchronizableRepeater()
    {
        Content content = _contentBooleanTrue;
        ModelItem modelItem = _contentType.getModelItem("composite/repeater/withConditionInEntry");
        Optional<String> oldDataPath = Optional.empty();

        String dataPathInEntry1 = "composite/repeater[1]/withConditionInEntry";
        String dataPathInEntry2 = "composite/repeater[2]/withConditionInEntry";

        Map<String, Object> contextualParameters = new HashMap<>();
        contextualParameters.put(AmetysObjectDisableConditionsEvaluator.SYNCHRONIZATION_CONTEXT_PARAMETER_KEY, SynchronizationContext.newInstance());
        
        // REPLACE_ALL
        SynchronizableRepeater syncRepeater = SynchronizableRepeater.replaceAll(List.of(Map.of("inRepeater", false), Map.of("inRepeater", true)), null);
        Map<String, Object> values = Map.of("composite", Map.of("repeater", syncRepeater));
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPathInEntry1, oldDataPath, values, content, contextualParameters));
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPathInEntry2, oldDataPath, values, content, contextualParameters));
        
        // REPLACE_ALL with SynchronizableValue
        syncRepeater = SynchronizableRepeater.replaceAll(List.of(Map.of("inRepeater", new SynchronizableValue(false)), Map.of("inRepeater", new SynchronizableValue(true))), null);
        values = Map.of("composite", Map.of("repeater", syncRepeater));
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPathInEntry1, oldDataPath, values, content, contextualParameters));
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPathInEntry2, oldDataPath, values, content, contextualParameters));
        
        // REPLACE - replace second entry but not the first one, will get value from stored data
        syncRepeater = SynchronizableRepeater.replace(List.of(Map.of("inRepeater", true)), List.of(2));
        values = Map.of("composite", Map.of("repeater", syncRepeater));
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPathInEntry1, oldDataPath, values, content, contextualParameters));
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPathInEntry2, oldDataPath, values, content, contextualParameters));
        
        // APPEND_AND_REMOVE - add third entry but remove the first one
        syncRepeater = SynchronizableRepeater.appendOrRemove(List.of(Map.of("inRepeater", true)), Set.of(1));
        values = Map.of("composite", Map.of("repeater", syncRepeater));
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, "composite/repeater[1]/withConditionInEntry", Optional.of("composite/repeater[2]/withConditionInEntry"), values, content, contextualParameters));
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, "composite/repeater[2]/withConditionInEntry", Optional.empty(), values, content, contextualParameters));
    }

    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * the evaluated attribute has some sub-conditions
     */
    @Test
    public void testEvaluateDisableSubConditions()
    {
        Content content = _createContent("title", __CONTENT_TYPE_ID);
        String dataPath = "withSubConditions";
        Optional<String> oldDataPath = Optional.empty();
        ModelItem modelItem = _contentType.getModelItem(dataPath);
        Map<String, Object> contextualParameters = new HashMap<>();
        
        Map<String, Object> values = Map.of("title", "hello",
                                            "atRoot", true);
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
        
        values = Map.of("title", "bonjour",
                        "atRoot", true);
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
        
        values = Map.of("title", "hello",
                        "atRoot", false);
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
        
        values = Map.of("title", "hola",
                        "atRoot", true);
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
        
        values = Map.of("title", "hola",
                        "atRoot", false);
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
    }

    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * The attribute referenced by the condition is multiple, the condition operator is EQ
     */
    @Test
    public void testEvaluateEqualDisableConditionsOnMultipleAttribute()
    {
        Content content = _createContent("title", __CONTENT_TYPE_ID);
        String dataPath = "withEqConditionOnMultiple";
        Optional<String> oldDataPath = Optional.empty();
        ModelItem modelItem = _contentType.getModelItem(dataPath);
        Map<String, Object> contextualParameters = new HashMap<>();
        
        Map<String, Object> values = Map.of("stringmult", new String[] {"winner"});
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
        
        values = Map.of("stringmult", new String[] {"winner", "loser"});
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
        
        values = Map.of("stringmult", new String[] {"loser1", "loser2"});
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
    }

    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * The attribute referenced by the condition is multiple, the condition operator is NEQ
     */
    @Test
    public void testEvaluateNotEqualDisableConditionsOnMultipleAttribute()
    {
        Content content = _createContent("title", __CONTENT_TYPE_ID);
        String dataPath = "withNeqConditionOnMultiple";
        Optional<String> oldDataPath = Optional.empty();
        ModelItem modelItem = _contentType.getModelItem(dataPath);
        Map<String, Object> contextualParameters = new HashMap<>();
        
        Map<String, Object> values = Map.of("stringmult", new String[] {"loser"});
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
        
        values = Map.of("stringmult", new String[] {"loser", "winner"});
        Assert.assertFalse(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
        
        values = Map.of("stringmult", new String[] {"winner1", "winner2"});
        Assert.assertTrue(_evaluator.evaluateDisableConditions(modelItem, dataPath, oldDataPath, values, content, contextualParameters));
    }
    
    private Content _createLinkedContent(String title, boolean booleanValue)
    {
        ModifiableContent content = _createContent(title, __LINKED_CONTENT_TYPE_ID);

        content.getComposite("composite", true)
               .setValue("boolean", booleanValue);
        
        content.saveChanges();
        return content;
    }

    private Content _createContentWithLinkedContent(String title, boolean booleanValue, Content linkedContent)
    {
        ModifiableContent content = _createContent(title, __CONTENT_TYPE_ID);

        content.setValue("atRoot", booleanValue);
        content.setValue("linkedContent", linkedContent);
        
        ModifiableModelAwareComposite composite = content.getComposite("composite", true);
        ModifiableModelAwareRepeaterEntry entry1 = composite.getRepeater("repeater", true).addEntry();
        entry1.setValue("inRepeater", true);
        entry1.setValue("linkedContent", _linkedContentBooleanTrue);
        ModifiableModelAwareRepeaterEntry entry2 = composite.getRepeater("repeater").addEntry();
        entry2.setValue("inRepeater", false);
        entry2.setValue("linkedContent", _linkedContentBooleanFalse);
        
        content.saveChanges();
        return content;
    }

    private ModifiableContent _createContent(String title, String contentTypeId)
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        
        ModifiableContent content = root.createChild(title, "ametys:defaultContent");

        content.setTypes(new String[]{contentTypeId});
        content.setTitle(title, Locale.FRANCE);
        content.setLanguage("fr");
        content.setCreationDate(ZonedDateTime.now());
        content.setCreator(new UserIdentity("enaour", "wallace"));
        content.setLastModified(ZonedDateTime.now());
        content.setLastContributor(new UserIdentity("bfraisse", "wallace"));
        
        return content;
    }
}
