/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.extractor.xml;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.time.LocalDate;
import java.time.Month;
import java.time.ZonedDateTime;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.TransformerHandler;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.io.IOUtils;
import org.junit.Assert;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Binary;
import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.File;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.Resource;
import org.ametys.cms.data.RichText;
import org.ametys.cms.data.holder.impl.DefaultModifiableModelAwareDataHolder;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.extractor.ModelAwareValuesExtractor;
import org.ametys.plugins.repository.data.extractor.ValuesExtractor;
import org.ametys.plugins.repository.data.extractor.xml.ModelAwareXMLValuesExtractor;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.ModifiableModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeaterEntry;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.SimpleViewItemGroup;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElement;
import org.ametys.runtime.model.ViewItemGroup;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelAwareXMLValuesExtractor} test case
 */
@NeedFeatures({
    "test-data/test-model",
    "cms/factories",
    "test/content-type",
    "cms/workflow.content",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
})
public class ModelAwareXMLValuesExtractorTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    
    private static Model _model;
    private static AmetysObjectResolver _ametysObjectResolver;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    /**
     * Tests the {@link ModelAwareXMLValuesExtractor#extractValues()} method
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testExtractValuesWithoutView() throws Exception
    {
        Element element = _createElementWithSomeDataAndGroups("testExtractValues");
        ModelAwareValuesExtractor extractor = new ModelAwareXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _model);
        
        Map<String, Object> values = extractor.extractValues();
        Assert.assertEquals(9, values.size());
        
        Assert.assertEquals("value of the single string data", values.get("singleStringData"));
        Assert.assertArrayEquals(new String[] {"string value 1", "string value 2", "string value 3"}, (String[]) values.get("multipleStringData"));
        Assert.assertArrayEquals(new Long[] {32L, 784L}, (Long[]) values.get("multipleLongData"));
        
        Assert.assertTrue(values.containsKey("singleDoubleData"));
        Assert.assertNull(values.get("singleDoubleData"));
        
        Assert.assertTrue(values.containsKey("multipleDateData"));
        Assert.assertNull(values.get("multipleDateData"));

        Geocode geocode = (Geocode) values.get("singleGeocodeData");
        Assert.assertEquals(Double.valueOf(85.487), geocode.getLatitude());
        Assert.assertEquals(Double.valueOf(6.254), geocode.getLongitude());
        
        Assert.assertArrayEquals(new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY}, (UserIdentity[]) values.get("multipleUserData"));
        
        ContentValue contentValue = (ContentValue) values.get("singleContentData");
        Assert.assertEquals("singleContentValue", contentValue.getContent().getTitle());
        
        Map<String, Object> composite = (Map<String, Object>) values.get("composite");
        Assert.assertEquals(2, composite.size());
        
        Assert.assertEquals(Long.valueOf(54), composite.get("singleLongData"));
        
        List<Map<String, Object>> repeater = (List<Map<String, Object>>) composite.get("repeater");
        Assert.assertEquals(2, repeater.size());

        Map<String, Object> entry1 = repeater.get(0);
        Assert.assertEquals(3, entry1.size());
        Assert.assertFalse((Boolean) entry1.get("singleBooleanData"));
        Assert.assertEquals(Double.valueOf(0.6), entry1.get("singleDoubleData"));
        Map<String, Object> subcompositeEntry1 = (Map<String, Object>) entry1.get("subcomposite");
        Assert.assertEquals(2, subcompositeEntry1.size());
        Assert.assertEquals("string value in the sub composite of the first repeater entry", subcompositeEntry1.get("singleStringData"));
        Assert.assertArrayEquals(new LocalDate[]{LocalDate.of(2018, Month.SEPTEMBER, 11), LocalDate.of(2015, Month.DECEMBER, 06)}, (LocalDate[]) subcompositeEntry1.get("multipleDateData"));
        
        Map<String, Object> entry2 = repeater.get(1);
        Assert.assertEquals(3, entry2.size());
        Assert.assertTrue((Boolean) entry2.get("singleBooleanData"));
        Assert.assertEquals(Double.valueOf(12.45), entry2.get("singleDoubleData"));
        Map<String, Object> subcompositeEntry2 = (Map<String, Object>) entry2.get("subcomposite");
        Assert.assertEquals(2, subcompositeEntry2.size());
        Assert.assertEquals("string value in the sub composite of the second repeater entry", subcompositeEntry2.get("singleStringData"));
        Assert.assertArrayEquals(new LocalDate[]{LocalDate.of(1985, Month.FEBRUARY, 27), LocalDate.of(1985, Month.MAY, 24)}, (LocalDate[]) subcompositeEntry2.get("multipleDateData"));
    }
    
    /**
     * Tests the {@link ModelAwareXMLValuesExtractor#extractValues(View)} method
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testExtractValues() throws Exception
    {
        Element element = _createElementWithSomeDataAndGroups("testExtractValues");
        ModelAwareValuesExtractor extractor = new ModelAwareXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _model);
        
        View view = View.of(_model, "singleStringData",
                "multipleLongData",
                "singleDoubleData",
                "multipleDateData",
                "singleGeocodeData",
                "multipleUserData",
                "singleContentData",
                "singleMultilingualStringData",
                "composite/singleLongData",
                "composite/repeater/singleBooleanData",
                "composite/repeater/subcomposite/singleStringData");
        
        Map<String, Object> values = extractor.extractValues(view);
        Assert.assertEquals(8, values.size());
        
        Assert.assertEquals("value of the single string data", values.get("singleStringData"));
        Assert.assertArrayEquals(new Long[] {32L, 784L}, (Long[]) values.get("multipleLongData"));
        
        Assert.assertTrue(values.containsKey("singleDoubleData"));
        Assert.assertNull(values.get("singleDoubleData"));
        Assert.assertTrue(values.containsKey("multipleDateData"));
        Assert.assertNull(values.get("multipleDateData"));

        Geocode geocode = (Geocode) values.get("singleGeocodeData");
        Assert.assertEquals(Double.valueOf(6.254), geocode.getLongitude());
        Assert.assertEquals(Double.valueOf(85.487), geocode.getLatitude());
        
        Assert.assertArrayEquals(new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY}, (UserIdentity[]) values.get("multipleUserData"));
        
        ContentValue contentValue = (ContentValue) values.get("singleContentData");
        Assert.assertEquals("singleContentValue", contentValue.getContent().getTitle());
        
        Assert.assertFalse(values.containsKey("singleMultilingualStringData"));
        
        Map<String, Object> composite = (Map<String, Object>) values.get("composite");
        Assert.assertEquals(2, composite.size());
        
        Assert.assertEquals(Long.valueOf(54), composite.get("singleLongData"));
        
        List<Map<String, Object>> repeater = (List<Map<String, Object>>) composite.get("repeater");
        Assert.assertEquals(2, repeater.size());

        Map<String, Object> entry1 = repeater.get(0);
        Assert.assertEquals(2, entry1.size());
        Assert.assertFalse((Boolean) entry1.get("singleBooleanData"));
        Map<String, Object> subcompositeEntry1 = (Map<String, Object>) entry1.get("subcomposite");
        Assert.assertEquals(1, subcompositeEntry1.size());
        Assert.assertEquals("string value in the sub composite of the first repeater entry", subcompositeEntry1.get("singleStringData"));
        
        Map<String, Object> entry2 = repeater.get(1);
        Assert.assertEquals(2, entry2.size());
        Assert.assertTrue((Boolean) entry2.get("singleBooleanData"));
        Map<String, Object> subcompositeEntry2 = (Map<String, Object>) entry2.get("subcomposite");
        Assert.assertEquals(1, subcompositeEntry2.size());
        Assert.assertEquals("string value in the sub composite of the second repeater entry", subcompositeEntry2.get("singleStringData"));
    }
    
    /**
     * Tests the {@link ModelAwareXMLValuesExtractor#extractValues(View)} method with enumerated values
     * @throws Exception if an error occurs
     */
    @Test
    public void testExtractEnumeratedValues() throws Exception
    {
        ModifiableModelAwareDataHolder dataHolder = _createEmptyDataHolder("testExtractEnumeratedValues");
        dataHolder.setValue("singleEnumeratedStringData", "value1");
        dataHolder.setValue("singleEnumeratedLongData", 3L);
        Element element = _createElementFromSAXedDataHolder(dataHolder);
        
        ModelAwareValuesExtractor extractor = new ModelAwareXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _model);
        
        View view = View.of(_model, "singleEnumeratedStringData",
                "singleEnumeratedLongData");
        
        Map<String, Object> values = extractor.extractValues(view);
        Assert.assertEquals(2, values.size());
        
        Assert.assertEquals("value1", values.get("singleEnumeratedStringData"));
        Assert.assertEquals(3L, values.get("singleEnumeratedLongData"));
    }
    
    /**
     * Tests the {@link ModelAwareXMLValuesExtractor#extractValues(View)} method wit a view containing a fieldset
     * @throws Exception if an error occurs
     */
    @Test
    public void testExtractValuesWithAFieldsetInTheView() throws Exception
    {
        Element element = _createElementWithSomeDataAndGroups("testExtractValuesWithAFieldsetInTheView");
        ModelAwareValuesExtractor extractor = new ModelAwareXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _model);
        
        View view = _createViewWithAFieldset();
        Map<String, Object> values = extractor.extractValues(view);
        
        Assert.assertEquals("value of the single string data", values.get("singleStringData"));

        Assert.assertArrayEquals(new Long[] {32L, 784L}, (Long[]) values.get("multipleLongData"));
        Assert.assertArrayEquals(new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY}, (UserIdentity[]) values.get("multipleUserData"));
    }

    private View _createViewWithAFieldset()
    {
        View view = new View();
        
        ViewElement singleStringDataViewElement = new ViewElement();
        singleStringDataViewElement.setDefinition((ElementDefinition) _model.getModelItem("singleStringData"));
        view.addViewItem(singleStringDataViewElement);
        
        SimpleViewItemGroup fieldset = new SimpleViewItemGroup();
        fieldset.setName("fieldset");
        fieldset.setRole(ViewItemGroup.FIELDSET_ROLE);
        view.addViewItem(fieldset);
        
        ViewElement multipleLongDataViewElement = new ViewElement();
        multipleLongDataViewElement.setDefinition((ElementDefinition) _model.getModelItem("multipleLongData"));
        fieldset.addViewItem(multipleLongDataViewElement);
        
        ViewElement multipleUserDataViewElement = new ViewElement();
        multipleUserDataViewElement.setDefinition((ElementDefinition) _model.getModelItem("multipleUserData"));
        fieldset.addViewItem(multipleUserDataViewElement);
        return view;
    }
    
    /**
     * Tests the {@link ModelAwareXMLValuesExtractor#extractValue(String)} method
     * @throws Exception if an error occurs
     */
    @Test
    public void testExtractValue() throws Exception
    {
        Element element = _createElementWithSomeDataAndGroups("testExtractValue");
        ValuesExtractor extractor = new ModelAwareXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _model);
        
        Assert.assertEquals("value of the single string data", extractor.extractValue("singleStringData"));
        Assert.assertArrayEquals(new Long[] {32L, 784L}, extractor.extractValue("multipleLongData"));
        
        Geocode geocode = extractor.extractValue("singleGeocodeData");
        Assert.assertEquals(Double.valueOf(6.254), geocode.getLongitude());
        Assert.assertEquals(Double.valueOf(85.487), geocode.getLatitude());
        
        Assert.assertArrayEquals(new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY}, extractor.extractValue("multipleUserData"));
        
        ContentValue contentValue = extractor.extractValue("singleContentData");
        Assert.assertEquals("singleContentValue", contentValue.getContent().getTitle());
        
        Assert.assertEquals(Long.valueOf(54), extractor.extractValue("composite/singleLongData"));
        
        Assert.assertEquals(Double.valueOf(0.6), extractor.extractValue("composite/repeater[1]/singleDoubleData"));
        Assert.assertEquals("string value in the sub composite of the first repeater entry", extractor.extractValue("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertArrayEquals(new LocalDate[]{LocalDate.of(2018, Month.SEPTEMBER, 11), LocalDate.of(2015, Month.DECEMBER, 06)}, extractor.extractValue("composite/repeater[1]/subcomposite/multipleDateData"));
        
        Assert.assertTrue(extractor.extractValue("composite/repeater[2]/singleBooleanData"));
        Assert.assertEquals("string value in the sub composite of the second repeater entry", extractor.extractValue("composite/repeater[2]/subcomposite/singleStringData"));
        Assert.assertArrayEquals(new LocalDate[]{LocalDate.of(1985, Month.FEBRUARY, 27), LocalDate.of(1985, Month.MAY, 24)}, extractor.extractValue("composite/repeater[2]/subcomposite/multipleDateData"));
    }
    
    /**
     * Tests the {@link ModelAwareXMLValuesExtractor#extractValue(String)} method to extract a composite
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testExtractComposite() throws Exception
    {
        Element element = _createElementWithSomeDataAndGroups("testExtractComposite");
        ValuesExtractor extractor = new ModelAwareXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _model);
        
        Map<String, Object> composite = extractor.extractValue("composite");
        Assert.assertEquals(Long.valueOf(54), composite.get("singleLongData"));
        
        List<Map<String, Object>> repeater = (List<Map<String, Object>>) composite.get("repeater");
        Assert.assertEquals(2, repeater.size());

        Map<String, Object> entry1 = repeater.get(0);
        Assert.assertEquals(Double.valueOf(0.6), entry1.get("singleDoubleData"));
        Map<String, Object> subcompositeEntry1 = (Map<String, Object>) entry1.get("subcomposite");
        Assert.assertEquals("string value in the sub composite of the first repeater entry", subcompositeEntry1.get("singleStringData"));
        Assert.assertArrayEquals(new LocalDate[]{LocalDate.of(2018, Month.SEPTEMBER, 11), LocalDate.of(2015, Month.DECEMBER, 06)}, (LocalDate[]) subcompositeEntry1.get("multipleDateData"));
        
        Map<String, Object> entry2 = repeater.get(1);
        Assert.assertTrue((Boolean) entry2.get("singleBooleanData"));
        Map<String, Object> subcompositeEntry2 = (Map<String, Object>) entry2.get("subcomposite");
        Assert.assertEquals("string value in the sub composite of the second repeater entry", subcompositeEntry2.get("singleStringData"));
        Assert.assertArrayEquals(new LocalDate[]{LocalDate.of(1985, Month.FEBRUARY, 27), LocalDate.of(1985, Month.MAY, 24)}, (LocalDate[]) subcompositeEntry2.get("multipleDateData"));
    }
    
    /**
     * Tests the {@link ModelAwareXMLValuesExtractor#extractValue(String)} method to extract a repeater
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testExtractRepeater() throws Exception
    {
        Element element = _createElementWithSomeDataAndGroups("testExtractRepeater");
        ValuesExtractor extractor = new ModelAwareXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _model);
        
        List<Map<String, Object>> repeater = extractor.extractValue("composite/repeater");
        Assert.assertEquals(2, repeater.size());

        Map<String, Object> entry1 = repeater.get(0);
        Assert.assertEquals(Double.valueOf(0.6), entry1.get("singleDoubleData"));
        Map<String, Object> subcompositeEntry1 = (Map<String, Object>) entry1.get("subcomposite");
        Assert.assertEquals("string value in the sub composite of the first repeater entry", subcompositeEntry1.get("singleStringData"));
        Assert.assertArrayEquals(new LocalDate[]{LocalDate.of(2018, Month.SEPTEMBER, 11), LocalDate.of(2015, Month.DECEMBER, 06)}, (LocalDate[]) subcompositeEntry1.get("multipleDateData"));
        
        Map<String, Object> entry2 = repeater.get(1);
        Assert.assertTrue((Boolean) entry2.get("singleBooleanData"));
        Map<String, Object> subcompositeEntry2 = (Map<String, Object>) entry2.get("subcomposite");
        Assert.assertEquals("string value in the sub composite of the second repeater entry", subcompositeEntry2.get("singleStringData"));
        Assert.assertArrayEquals(new LocalDate[]{LocalDate.of(1985, Month.FEBRUARY, 27), LocalDate.of(1985, Month.MAY, 24)}, (LocalDate[]) subcompositeEntry2.get("multipleDateData"));
    }
    
    /**
     * Tests the {@link ModelAwareXMLValuesExtractor#extractValue(String)} method to extract a repeater entry
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testExtractRepeaterEntry() throws Exception
    {
        Element element = _createElementWithSomeDataAndGroups("testExtractRepeaterEntry");
        ValuesExtractor extractor = new ModelAwareXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _model);
        
        Map<String, Object> entry2 = extractor.extractValue("composite/repeater[2]");
        Assert.assertTrue((Boolean) entry2.get("singleBooleanData"));
        Map<String, Object> subcompositeEntry2 = (Map<String, Object>) entry2.get("subcomposite");
        Assert.assertEquals("string value in the sub composite of the second repeater entry", subcompositeEntry2.get("singleStringData"));
        Assert.assertArrayEquals(new LocalDate[]{LocalDate.of(1985, Month.FEBRUARY, 27), LocalDate.of(1985, Month.MAY, 24)}, (LocalDate[]) subcompositeEntry2.get("multipleDateData"));
    }
    
    /**
     * Tests the {@link ModelAwareXMLValuesExtractor#extractValues(View)} method to extract unvalued data
     * @throws Exception if an error occurs
     */
    @Test
    public void testExtractValuesWithUnvaluedData() throws Exception
    {
        Element element = _createElementWithSomeDataAndGroups("testExtractValuesWithUnvaluedData");
        ModelAwareValuesExtractor extractor = new ModelAwareXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _model);
        
        View view = View.of(_model, "singleBooleanData",
                "singleDoubleData",
                "multipleDoubleData",
                "multipleDateData",
                "unvaluedComposite",
                "unvaluedRepeater");
        
        Map<String, Object> values = extractor.extractValues(view);
        Assert.assertEquals(2, values.size());
        
        Assert.assertTrue(values.containsKey("singleDoubleData"));
        Assert.assertNull(values.get("singleDoubleData"));
        Assert.assertTrue(values.containsKey("multipleDateData"));
        Assert.assertNull(values.get("multipleDateData"));
    }
    
    /**
     * Tests the {@link ModelAwareXMLValuesExtractor#extractValue(String)} method to extract unvalued data
     * @throws Exception if an error occurs
     */
    @Test
    public void testExtractValueWithUnvaluedData() throws Exception
    {
        Element element = _createElementWithSomeDataAndGroups("testExtractValueWithUnvaluedData");
        ValuesExtractor extractor = new ModelAwareXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _model);
        
        Assert.assertNull(extractor.extractValue("singleBooleanData"));
        Assert.assertNull(extractor.extractValue("singleDoubleData"));
        Assert.assertNull(extractor.extractValue("multipleDoubleData"));
        Assert.assertNull(extractor.extractValue("multipleDateData"));
        Assert.assertNull(extractor.extractValue("composite/repeater[3]"));
        Assert.assertNull(extractor.extractValue("unvaluedComposite"));
        Assert.assertNull(extractor.extractValue("unvaluedRepeater"));
    }
    
    private Element _createElementWithSomeDataAndGroups(String testId) throws Exception
    {
        ModelAwareDataHolder dataHolder = _createDataHolderWithSomeDataAndGroups(testId);
        return _createElementFromSAXedDataHolder(dataHolder);
    }
    
    private ModelAwareDataHolder _createDataHolderWithSomeDataAndGroups(String testId)
    {
        ModifiableModelAwareDataHolder dataHolder = _createEmptyDataHolder(testId);
        
        dataHolder.setValue("singleStringData", "value of the single string data");
        dataHolder.setValue("multipleStringData", new String[] {"string value 1", "string value 2", "string value 3"});
        dataHolder.setValue("multipleLongData", new Long[] {32L, 784L});
        
        dataHolder.setValue("singleDoubleData", null);
        dataHolder.setValue("multipleDateData", new LocalDate[0]);
        
        dataHolder.setValue("singleGeocodeData", new Geocode(85.487D, 6.254D));
        dataHolder.setValue("multipleUserData", new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY});
        
        ModifiableContent singleContentValue = _createContent("singleContentValue");
        dataHolder.setValue("singleContentData", singleContentValue);

        ModifiableModelAwareComposite composite = dataHolder.getComposite("composite", true);
        composite.setValue("singleLongData", 54L);
        ModifiableModelAwareRepeater repeater = composite.getRepeater("repeater", true);

        ModifiableModelAwareRepeaterEntry entry1 = repeater.addEntry();
        entry1.setValue("singleBooleanData", false);
        entry1.setValue("singleDoubleData", 0.6D);
        ModifiableModelAwareComposite entry1SubComposite = entry1.getComposite("subcomposite", true);
        entry1SubComposite.setValue("singleStringData", "string value in the sub composite of the first repeater entry");
        entry1SubComposite.setValue("multipleDateData", new LocalDate[]{LocalDate.of(2018, Month.SEPTEMBER, 11), LocalDate.of(2015, Month.DECEMBER, 06)});
        
        ModifiableModelAwareRepeaterEntry entry2 = repeater.addEntry();
        entry2.setValue("singleBooleanData", true);
        entry2.setValue("singleDoubleData", 12.45D);
        ModifiableModelAwareComposite entry2SubComposite = entry2.getComposite("subcomposite", true);
        entry2SubComposite.setValue("singleStringData", "string value in the sub composite of the second repeater entry");
        entry2SubComposite.setValue("multipleDateData", new LocalDate[]{LocalDate.of(1985, Month.FEBRUARY, 27), LocalDate.of(1985, Month.MAY, 24)});
        
        return dataHolder;
    }
    
    private ModifiableContent _createContent(String title)
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        ModifiableContent content = root.createChild(title, "ametys:defaultContent");

        content.setTypes(new String[]{"org.ametys.cms.contenttype.Test"});
        content.setTitle(title, Locale.FRANCE);
        content.setLanguage("fr");
        content.setCreationDate(ZonedDateTime.now());
        content.setCreator(new UserIdentity("john", "anyPopulation"));
        content.setLastModified(ZonedDateTime.now());
        content.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        
        return content;
    }
    
    /**
     * Tests the {@link ModelAwareXMLValuesExtractor#extractValue(String)} method to extract resources (binaries, files and rich texts) data
     * @throws Exception if an error occurs
     */
    @Test
    public void testExtractValueWithResources() throws Exception
    {
        Element element = _createElementWithResources("testExtractValueWithResources");
        ValuesExtractor extractor = new ModelAwareXMLValuesExtractor(element, (dataPath, type) -> Optional.of(_getResourcesAdditionalData(dataPath)), _model);
        
        Binary singleBinaryData = extractor.extractValue("singleBinaryData");
        Assert.assertEquals("mySingleBinary.txt", singleBinaryData.getFilename());
        _assertResourceContent(singleBinaryData, "my single binary content");
        
        Binary[] multipleBinaryData = extractor.extractValue("multipleBinaryData");
        Assert.assertEquals(3, multipleBinaryData.length);
        Binary binary1 = _getFileFromArray(multipleBinaryData, "myBinary1.txt");
        _assertResourceContent(binary1, "my binary 1 content");
        Binary binary2 = _getFileFromArray(multipleBinaryData, "myBinary2.txt");
        _assertResourceContent(binary2, "my binary 2 content");
        Binary binary3 = _getFileFromArray(multipleBinaryData, "myBinary3.txt");
        _assertResourceContent(binary3, "my binary 3 content");
        
        Binary singleFileData = extractor.extractValue("singleFileData");
        Assert.assertEquals("mySingleFile.txt", singleFileData.getFilename());
        _assertResourceContent(singleFileData, "my single file content");
        
        File[] multipleFileData = extractor.extractValue("multipleFileData");
        Assert.assertEquals(3, multipleFileData.length);
        Binary file1 = _getFileFromArray(multipleFileData, "myFile1.txt");
        _assertResourceContent(file1, "my file 1 content");
        Binary file2 = _getFileFromArray(multipleFileData, "myFile2.txt");
        _assertResourceContent(file2, "my file 2 content");
        Binary file3 = _getFileFromArray(multipleFileData, "myFile3.txt");
        _assertResourceContent(file3, "my file 3 content");
        
        RichText singleRichTextData = extractor.extractValue("singleRichTextData");
        _assertRichTextContent(singleRichTextData, "my single rich text content");
        // Annotations
        _assertRichTextAnnotations(singleRichTextData,
                Map.of("annotations1", List.of("a1-1", "a1-2", "a1-3"),
                        "annotations2", List.of("a2-1", "a2-2")));

        RichText[] multipleRichTextData = extractor.extractValue("multipleRichTextData");
        Assert.assertEquals(2, multipleRichTextData.length);
        RichText firstRichText = _getRichTextFromArray(multipleRichTextData, "my first rich text content");
        // Annotations
        _assertRichTextAnnotations(firstRichText,
                Map.of("annotations", List.of("a1-1", "a1-2")));
        
        RichText secondRichText = _getRichTextFromArray(multipleRichTextData, "my second rich text content");
        // Annotations
        _assertRichTextAnnotations(secondRichText,
                Map.of("annotations1", List.of("a1-1", "a1-2", "a1-3"),
                        "annotations2", List.of("a2-1", "a2-2", "a2-3", "a2-4"),
                        "annotations3", List.of("a3-1")));
    }
    
    private Element _createElementWithResources(String testId) throws Exception
    {
        ModelAwareDataHolder dataHolder = _createDataHolderWithResources(testId);
        return _createElementFromSAXedDataHolder(dataHolder);
    }
    
    private ModelAwareDataHolder _createDataHolderWithResources(String testId) throws Exception
    {
        ModifiableModelAwareDataHolder dataHolder = _createEmptyDataHolder(testId);
        
        dataHolder.setValue("singleBinaryData", _createBinary("mySingleBinary.txt", "my single binary content"));
        dataHolder.setValue("multipleBinaryData", new Binary[] {
                _createBinary("myBinary1.txt", "my binary 1 content"),
                _createBinary("myBinary2.txt", "my binary 2 content"),
                _createBinary("myBinary3.txt", "my binary 3 content")
        });
        
        dataHolder.setValue("singleFileData", _createBinary("mySingleFile.txt", "my single file content"));
        dataHolder.setValue("multipleFileData", new Binary[] {
                _createBinary("myFile1.txt", "my file 1 content"),
                _createBinary("myFile2.txt", "my file 2 content"),
                _createBinary("myFile3.txt", "my file 3 content")
        });
        
        dataHolder.setValue("singleRichTextData",
                _createRichText("my single rich text content",
                        Map.of("annotations1", List.of("a1-1", "a1-2", "a1-3"),
                                "annotations2", List.of("a2-1", "a2-2"))));
        
        dataHolder.setValue("multipleRichTextData", new RichText[] {
                _createRichText("my first rich text content",
                        Map.of("annotations", List.of("a1-1", "a1-2"))),
                _createRichText("my second rich text content",
                        Map.of("annotations1", List.of("a1-1", "a1-2", "a1-3"),
                                "annotations2", List.of("a2-1", "a2-2", "a2-3", "a2-4"),
                                "annotations3", List.of("a3-1")))
        });
        
        return dataHolder;
    }
    
    private Binary _createBinary(String filename, String content) throws IOException
    {
        Binary binary = new Binary();
        
        binary.setEncoding("UTF-8");
        binary.setFilename(filename);
        binary.setLastModificationDate(ZonedDateTime.now());
        binary.setMimeType("text/plain");
        binary.setInputStream(_createInputStream(content));
        
        return binary;
    }
    
    private RichText _createRichText(String text, Map<String, List<String>> annotations) throws IOException
    {
        RichText richText = new RichText();
        
        richText.setEncoding("UTF-8");
        richText.setMimeType("text/xml");
        richText.setLastModificationDate(ZonedDateTime.now());
        String content = _getRichTextContentWithAnnotations(text, annotations);
        richText.setInputStream(_createInputStream(content));
        
        return richText;
    }
    
    // TODO NEWATTRIBUTEAPI: Add attachments to the rich texts in the resources test cases
    private String _getRichTextContentWithAnnotations(String text, Map<String, List<String>> annotations)
    {
        String content = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
             + "<article xmlns=\"http://docbook.org/ns/docbook\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" xmlns:html=\"http://www.w3.org/1999/xhtml\" version=\"5.0\">"
                 + "<para>"
                     + text
                 + "</para>";
        
        for (Map.Entry<String, List<String>> annotation : annotations.entrySet())
        {
            String annotationName = annotation.getKey();
            for (String annotationValue : annotation.getValue())
            {
                content += "<para><phrase role=\"" + annotationName + "\" class=\"semantic\">" + annotationValue + "</phrase></para>";
            }
        }
        
        content += "</article>";
        
        return content;
    }
    
    private Map<String, InputStream> _getResourcesAdditionalData(String dataPath)
    {
        switch (dataPath)
        {
            case "singleBinaryData":
                return Map.of("mySingleBinary.txt", _createInputStream("my single binary content"));
            case "multipleBinaryData":
                return Map.of("myBinary1.txt", _createInputStream("my binary 1 content"),
                        "myBinary2.txt", _createInputStream("my binary 2 content"),
                        "myBinary3.txt", _createInputStream("my binary 3 content"));
            case "singleFileData":
                return Map.of("mySingleFile.txt", _createInputStream("my single file content"));
            case "multipleFileData":
                return Map.of("myFile1.txt", _createInputStream("my file 1 content"),
                        "myFile2.txt", _createInputStream("my file 2 content"),
                        "myFile3.txt", _createInputStream("my file 3 content"));
            default:
                return Map.of();
        }
    }
    
    private InputStream _createInputStream(String content)
    {
        return new ByteArrayInputStream(content.getBytes());
    }
    
    @SuppressWarnings("unchecked")
    private <T extends File> T _getFileFromArray(File[] files, String filename)
    {
        for (File file : files)
        {
            if (file.getName().equals(filename))
            {
                return (T) file;
            }
        }
        
        Assert.fail("Unable to find the binary with file name '" + filename + "'");
        return null;
    }
    
    private void _assertResourceContent(Resource resource, String content) throws IOException
    {
        Assert.assertTrue(_isResourceContentEqual(resource, content));
    }
    
    private boolean _isResourceContentEqual(Resource resource, String content) throws IOException
    {
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(resource.getInputStream(), writer, resource.getEncoding());
            return content.equals(writer.toString());
        }
    }
    
    private RichText _getRichTextFromArray(RichText[] richTexts, String text) throws IOException
    {
        for (RichText richText : richTexts)
        {
            if (_doesRichTextContentContains(richText, text))
            {
                return richText;
            }
        }
        
        Assert.fail("Unable to find the rich text containing the text '" + text + "'");
        return null;
    }
    
    private void _assertRichTextContent(RichText richText, String text) throws IOException
    {
        Assert.assertTrue(_doesRichTextContentContains(richText, text));
    }
    
    private boolean _doesRichTextContentContains(RichText richText, String text) throws IOException
    {
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(richText.getInputStream(), writer, richText.getEncoding());
            return writer.toString().contains(text);
        }
    }
    
    private void _assertRichTextAnnotations(RichText richText, Map<String, List<String>> allAannotations)
    {
        Assert.assertEquals(allAannotations.size(), richText.getAllAnnotations().size());
        for (Map.Entry<String, List<String>> annotationsEntry : allAannotations.entrySet())
        {
            List<String> annotations = richText.getAnnotations(annotationsEntry.getKey());
            Assert.assertEquals(annotationsEntry.getValue().size(), annotations.size());
            for (String annotation : annotationsEntry.getValue())
            {
                Assert.assertTrue(annotations.contains(annotation));
            }
        }
    }
    
    private ModifiableModelAwareDataHolder _createEmptyDataHolder(String testId)
    {
        RootAmetysObject ametysRoot = _ametysObjectResolver.resolveByPath("/");
        ModifiableRepositoryData dataRoot = new JCRRepositoryData(ametysRoot.getNode());
        ModifiableRepositoryData repositoryData = dataRoot.addRepositoryData(testId, DATA_NODE_TYPE);
        return new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
    }
    
    private Element _createElementFromSAXedDataHolder(ModelAwareDataHolder dataHolder) throws Exception
    {
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        return (Element) document.getElementsByTagName("result").item(0);
    }
    
    /**
     * Tests the {@link ModelAwareXMLValuesExtractor#extractValues(View)} method with mixed up repeater entries
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testExtractValuesKeepsRepeaterEntriesOrder() throws Exception
    {
        Element element = _createElementWithMixedUpRepeaterEntries();
        ModelAwareValuesExtractor extractor = new ModelAwareXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _model);
        View view = View.of(_model, "composite/repeater/singleDoubleData");
        
        Map<String, Object> values = extractor.extractValues(view);
        
        Map<String, Object> compositeValues = (Map<String, Object>) values.get("composite");
        List<Map<String, Object>> repeaterEntries = (List<Map<String, Object>>) compositeValues.get("repeater");
        Assert.assertEquals(Double.valueOf(1), repeaterEntries.get(0).get("singleDoubleData"));
        Assert.assertEquals(Double.valueOf(2), repeaterEntries.get(1).get("singleDoubleData"));
        Assert.assertEquals(Double.valueOf(3), repeaterEntries.get(2).get("singleDoubleData"));
        Assert.assertEquals(Double.valueOf(4), repeaterEntries.get(3).get("singleDoubleData"));
        
    }
    
    /**
     * Tests the {@link ModelAwareXMLValuesExtractor#extractValue(String)} method with mixed up repeater entries
     * @throws Exception if an error occurs
     */
    @Test
    public void testExtractValueKeepsRepeaterEntriesOrder() throws Exception
    {
        Element element = _createElementWithMixedUpRepeaterEntries();
        ValuesExtractor extractor = new ModelAwareXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _model);
        
        Assert.assertEquals(Double.valueOf(1), extractor.extractValue("composite/repeater[1]/singleDoubleData"));
        Assert.assertEquals(Double.valueOf(2), extractor.extractValue("composite/repeater[2]/singleDoubleData"));
        Assert.assertEquals(Double.valueOf(3), extractor.extractValue("composite/repeater[3]/singleDoubleData"));
        Assert.assertEquals(Double.valueOf(4), extractor.extractValue("composite/repeater[4]/singleDoubleData"));
    }
    
    private Element _createElementWithMixedUpRepeaterEntries() throws Exception
    {
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        XMLUtils.startElement(handler, "composite");
        XMLUtils.startElement(handler, "repeater");
        
        _entryToSAX(handler, 3);
        _entryToSAX(handler, 2);
        _entryToSAX(handler, 4);
        _entryToSAX(handler, 1);
        
        XMLUtils.endElement(handler, "repeater");
        XMLUtils.endElement(handler, "composite");
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        return (Element) document.getElementsByTagName("result").item(0);
    }
    
    private void _entryToSAX(TransformerHandler handler, int position) throws SAXException
    {
        AttributesImpl entry1Attributes = new AttributesImpl();
        entry1Attributes.addCDATAAttribute("name", String.valueOf(position));
        XMLUtils.startElement(handler, "entry", entry1Attributes);
        XMLUtils.createElement(handler, "singleDoubleData", String.valueOf(position));
        XMLUtils.endElement(handler, "entry");
    }
}
