/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.extractor.xml;

import java.util.Map;
import java.util.Optional;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.TransformerHandler;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.XMLUtils;
import org.junit.Assert;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.extractor.ModelLessValuesExtractor;
import org.ametys.plugins.repository.data.extractor.ValuesExtractor;
import org.ametys.plugins.repository.data.extractor.xml.ModelLessXMLValuesExtractor;
import org.ametys.plugins.repository.data.holder.ModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.ModifiableModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelLessComposite;
import org.ametys.plugins.repository.data.holder.impl.DefaultModifiableModelLessDataHolder;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelLessXMLValuesExtractor} test case
 */
@NeedFeatures({
    "test-data/test-model-less-data-types",
    "cms/factories",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModelLessXMLValuesExtractorTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".TestModelLess");
    }
    
    /**
     * Tests the {@link ModelLessXMLValuesExtractor#extractValues()} method
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testExtractValues() throws Exception
    {
        Element element = _createElementWithSomeDataAndGroups("testExtractValues");
        ModelLessValuesExtractor extractor = new ModelLessXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _testDataTypeExtensionPoint);
        
        Map<String, Object> values = extractor.extractValues();
        Assert.assertEquals(5, values.size());
        
        Assert.assertEquals("value of the single string data", values.get("singleStringData"));
        Assert.assertArrayEquals(new Long[] {32L, 784L}, (Long[]) values.get("multipleLongData"));
        
        Geocode geocode = (Geocode) values.get("singleGeocodeData");
        Assert.assertEquals(Double.valueOf(6.254), geocode.getLongitude());
        Assert.assertEquals(Double.valueOf(85.487), geocode.getLatitude());
        
        Assert.assertArrayEquals(new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY}, (UserIdentity[]) values.get("multipleUserData"));
        
        Map<String, Object> composite = (Map<String, Object>) values.get("composite");
        Assert.assertEquals(3, composite.size());
        
        Assert.assertEquals(Long.valueOf(54), composite.get("singleLongData"));
        Assert.assertFalse((Boolean) composite.get("singleBooleanData"));
        
        Map<String, Object> subcomposite = (Map<String, Object>) composite.get("subcomposite");
        Assert.assertEquals(1, subcomposite.size());
        Assert.assertEquals("string value in the sub composite", subcomposite.get("singleStringData"));
    }
    
    /**
     * Tests the {@link ModelLessXMLValuesExtractor#extractValue(String)} method
     * @throws Exception if an error occurs
     */
    @Test
    public void testExtractValue() throws Exception
    {
        Element element = _createElementWithSomeDataAndGroups("testExtractValue");
        ValuesExtractor extractor = new ModelLessXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _testDataTypeExtensionPoint);
        
        Assert.assertEquals("value of the single string data", extractor.extractValue("singleStringData"));
        Assert.assertArrayEquals(new Long[] {32L, 784L}, extractor.extractValue("multipleLongData"));
        
        Geocode geocode = extractor.extractValue("singleGeocodeData");
        Assert.assertEquals(Double.valueOf(6.254), geocode.getLongitude());
        Assert.assertEquals(Double.valueOf(85.487), geocode.getLatitude());
        
        Assert.assertArrayEquals(new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY}, extractor.extractValue("multipleUserData"));
        
        Assert.assertEquals(Long.valueOf(54), extractor.extractValue("composite/singleLongData"));
        Assert.assertFalse((Boolean) extractor.extractValue("composite/singleBooleanData"));
        
        Assert.assertEquals("string value in the sub composite", extractor.extractValue("composite/subcomposite/singleStringData"));
    }
    
    /**
     * Tests the {@link ModelLessXMLValuesExtractor#extractValue(String)} method to extract a composite
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testExtractComposite() throws Exception
    {
        Element element = _createElementWithSomeDataAndGroups("testExtractComposite");
        ValuesExtractor extractor = new ModelLessXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _testDataTypeExtensionPoint);
        
        Map<String, Object> composite = extractor.extractValue("composite");
        Assert.assertEquals(3, composite.size());
        
        Assert.assertEquals(Long.valueOf(54), composite.get("singleLongData"));
        Assert.assertFalse((Boolean) composite.get("singleBooleanData"));
        
        Map<String, Object> subcomposite = (Map<String, Object>) composite.get("subcomposite");
        Assert.assertEquals(1, subcomposite.size());
        Assert.assertEquals("string value in the sub composite", subcomposite.get("singleStringData"));
    }
    
    /**
     * Tests the {@link ModelLessXMLValuesExtractor#extractValue(String)} method to extract unvalued data
     * @throws Exception if an error occurs
     */
    @Test
    public void testExtractValueWithUnvaluedData() throws Exception
    {
        Element element = _createElementWithSomeDataAndGroups("testExtractValueWithUnvaluedData");
        ValuesExtractor extractor = new ModelLessXMLValuesExtractor(element, (dataPath, type) -> Optional.empty(), _testDataTypeExtensionPoint);
        
        Assert.assertNull(extractor.extractValue("unvaluedData"));
        Assert.assertNull(extractor.extractValue("composite/unvaluedData"));
    }
    
    private Element _createElementWithSomeDataAndGroups(String testId) throws Exception
    {
        ModelLessDataHolder dataHolder = _createDataHolderWithSomeDataAndGroups(testId);
        return _createElementFromSAXedDataHolder(dataHolder);
    }
    
    private ModelLessDataHolder _createDataHolderWithSomeDataAndGroups(String testId)
    {
        RootAmetysObject ametysRoot = _ametysObjectResolver.resolveByPath("/");
        ModifiableRepositoryData dataRoot = new JCRRepositoryData(ametysRoot.getNode());
        ModifiableRepositoryData repositoryData = dataRoot.addRepositoryData(testId, DATA_NODE_TYPE);
        ModifiableModelLessDataHolder dataHolder = new DefaultModifiableModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        dataHolder.setValue("singleStringData", "value of the single string data");
        dataHolder.setValue("multipleLongData", new Long[] {32L, 784L});
        
        dataHolder.setValue("singleGeocodeData", new Geocode(85.487D, 6.254D));
        dataHolder.setValue("multipleUserData", new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY});
        
        ModifiableModelLessComposite composite = dataHolder.getComposite("composite", true);
        composite.setValue("singleLongData", 54L);
        composite.setValue("singleBooleanData", false);
        
        ModifiableModelLessComposite subComposite = composite.getComposite("subcomposite", true);
        subComposite.setValue("singleStringData", "string value in the sub composite");
        
        return dataHolder;
    }
    
    private Element _createElementFromSAXedDataHolder(ModelLessDataHolder dataHolder) throws Exception
    {
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        return (Element) document.getElementsByTagName("result").item(0);
    }
}
