/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.plugins.repository.data.holder.impl.DataHolderHelper;
import org.ametys.plugins.repository.data.holder.values.SynchronizableRepeater;
import org.ametys.plugins.repository.model.CompositeDefinition;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElement;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

import junit.framework.TestCase;

/**
 * {@link TestCase} for {@link DataHolderHelper} implementation
 */
@NeedFeatures({
    "test-data/test-model",
//    "repository/resolver",
//    "cms/workflow.content"
})
public class DataHolderHelperTesCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static Model _model;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
    }
    
    /**
     * Tests the {@link DataHolderHelper#createViewItemAccessorFromValues(Collection, Map)} method
     * Nominal case
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testCreateViewItemAccessorFromValues()
    {
        Map<String, Object> values = new LinkedHashMap<>();
        values.put("singleStringData", "single string value");
        values.put("multipleLongData", new Long[] {1L, 2L, 3L});
        Map<String, Object> compositeValues = new LinkedHashMap<>();
        values.put("composite", compositeValues);
        compositeValues.put("singleLongData", 89L);
        List<Map<String, Object>> repeaterValues = new ArrayList<>();
        compositeValues.put("repeater", repeaterValues);
        Map<String, Object> entry1 = new LinkedHashMap<>();
        repeaterValues.add(entry1);
        entry1.put("singleBooleanData", true);
        Map<String, Object> entry1Subcomposite = new LinkedHashMap<>();
        entry1.put("subcomposite", entry1Subcomposite);
        entry1Subcomposite.put("singleStringData", "string value for entry 1");
        entry1Subcomposite.put("singleBinaryData", "binary value for entry 1");
        Map<String, Object> entry2 = new LinkedHashMap<>();
        repeaterValues.add(entry2);
        entry2.put("singleBooleanData", false);
        Map<String, Object> entry2Subcomposite = new LinkedHashMap<>();
        entry2.put("subcomposite", entry2Subcomposite);
        entry2Subcomposite.put("singleStringData", "string value for entry 2");
        entry2Subcomposite.put("singleBinaryData", "binary value for entry 2");
        
        View view = (View) DataHolderHelper.createViewItemAccessorFromValues(List.of(_model), values);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(3, viewItems.size());
        
        Assert.assertTrue(viewItems.get(0) instanceof ViewElement);
        Assert.assertEquals("singleStringData", viewItems.get(0).getName());
        
        Assert.assertTrue(viewItems.get(1) instanceof ViewElement);
        Assert.assertEquals("multipleLongData", viewItems.get(1).getName());
        
        Assert.assertTrue(viewItems.get(2) instanceof ModelViewItemGroup);
        ModelViewItemGroup<CompositeDefinition> composite = (ModelViewItemGroup<CompositeDefinition>) viewItems.get(2);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeViewItems = composite.getViewItems();
        Assert.assertEquals(2, compositeViewItems.size());
        
        Assert.assertTrue(compositeViewItems.get(0) instanceof ViewElement);
        Assert.assertEquals("singleLongData", compositeViewItems.get(0).getName());
        
        Assert.assertTrue(compositeViewItems.get(1) instanceof ModelViewItemGroup);
        ModelViewItemGroup<RepeaterDefinition> repeater = (ModelViewItemGroup<RepeaterDefinition>) compositeViewItems.get(1);
        Assert.assertEquals("repeater", repeater.getName());
        List<ViewItem> repeaterViewItems = repeater.getViewItems();
        Assert.assertEquals(2, repeaterViewItems.size());
        
        Assert.assertTrue(repeaterViewItems.get(0) instanceof ViewElement);
        Assert.assertEquals("singleBooleanData", repeaterViewItems.get(0).getName());
        
        Assert.assertTrue(repeaterViewItems.get(1) instanceof ModelViewItemGroup);
        ModelViewItemGroup<CompositeDefinition> subcomposite = (ModelViewItemGroup<CompositeDefinition>) repeaterViewItems.get(1);
        Assert.assertEquals("subcomposite", subcomposite.getName());
        List<ViewItem> subcompositeViewItems = subcomposite.getViewItems();
        Assert.assertEquals(2, subcompositeViewItems.size());
        
        Assert.assertTrue(subcompositeViewItems.get(0) instanceof ViewElement);
        Assert.assertEquals("singleStringData", subcompositeViewItems.get(0).getName());
        
        Assert.assertTrue(subcompositeViewItems.get(1) instanceof ViewElement);
        Assert.assertEquals("singleBinaryData", subcompositeViewItems.get(1).getName());
    }
    
    /**
     * Tests the {@link DataHolderHelper#createViewItemAccessorFromValues(Collection, Map)} method
     * string value is <code>null</code> -> the view item is still there
     */
    @Test
    public void testCreateViewItemAccessorFromValuesWithNullValue()
    {
        Map<String, Object> values = new LinkedHashMap<>();
        values.put("singleStringData", null);
        
        View view = (View) DataHolderHelper.createViewItemAccessorFromValues(List.of(_model), values);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(1, viewItems.size());
        
        Assert.assertTrue(viewItems.get(0) instanceof ViewElement);
        Assert.assertEquals("singleStringData", viewItems.get(0).getName());
    }
    
    /**
     * Tests the {@link DataHolderHelper#createViewItemAccessorFromValues(Collection, Map)} method
     * composite value is <code>null</code> -> the view item is still there
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testCreateViewItemAccessorFromValuesWithNullComposite()
    {
        Map<String, Object> values = new LinkedHashMap<>();
        values.put("composite", null);
        
        View view = (View) DataHolderHelper.createViewItemAccessorFromValues(List.of(_model), values);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(1, viewItems.size());
        
        Assert.assertTrue(viewItems.get(0) instanceof ModelViewItemGroup);
        ModelViewItemGroup<CompositeDefinition> composite = (ModelViewItemGroup<CompositeDefinition>) viewItems.get(0);
        Assert.assertEquals("composite", composite.getName());
        Assert.assertTrue(composite.getViewItems().isEmpty());
    }
    
    /**
     * Tests the {@link DataHolderHelper#createViewItemAccessorFromValues(Collection, Map)} method
     * composite value is empty -> the view item is still there
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testCreateViewItemAccessorFromValuesWithEmptyComposite()
    {
        Map<String, Object> values = new LinkedHashMap<>();
        values.put("composite", new LinkedHashMap<>());
        
        View view = (View) DataHolderHelper.createViewItemAccessorFromValues(List.of(_model), values);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(1, viewItems.size());
        
        Assert.assertTrue(viewItems.get(0) instanceof ModelViewItemGroup);
        ModelViewItemGroup<CompositeDefinition> composite = (ModelViewItemGroup<CompositeDefinition>) viewItems.get(0);
        Assert.assertEquals("composite", composite.getName());
        Assert.assertTrue(composite.getViewItems().isEmpty());
    }
    
    /**
     * Tests the {@link DataHolderHelper#createViewItemAccessorFromValues(Collection, Map)} method
     * composite value is empty -> the view item is still there
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testCreateViewItemAccessorFromValuesWithNullValueInComposite()
    {
        Map<String, Object> values = new LinkedHashMap<>();
        Map<String, Object> compositeValues = new LinkedHashMap<>();
        values.put("composite", compositeValues);
        compositeValues.put("singleLongData", null);
        
        View view = (View) DataHolderHelper.createViewItemAccessorFromValues(List.of(_model), values);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(1, viewItems.size());
        
        Assert.assertTrue(viewItems.get(0) instanceof ModelViewItemGroup);
        ModelViewItemGroup<CompositeDefinition> composite = (ModelViewItemGroup<CompositeDefinition>) viewItems.get(0);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeViewItems = composite.getViewItems();
        Assert.assertEquals(1, compositeViewItems.size());
        
        Assert.assertTrue(compositeViewItems.get(0) instanceof ViewElement);
        Assert.assertEquals("singleLongData", compositeViewItems.get(0).getName());
    }
    
    /**
     * Tests the {@link DataHolderHelper#createViewItemAccessorFromValues(Collection, Map)} method
     * repeater value is <code>null</code> -> the view item is still there
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testCreateViewItemAccessorFromValuesWithNullRepeater()
    {
        Map<String, Object> values = new LinkedHashMap<>();
        Map<String, Object> compositeValues = new LinkedHashMap<>();
        values.put("composite", compositeValues);
        compositeValues.put("repeater", null);
        
        View view = (View) DataHolderHelper.createViewItemAccessorFromValues(List.of(_model), values);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(1, viewItems.size());
        
        Assert.assertTrue(viewItems.get(0) instanceof ModelViewItemGroup);
        ModelViewItemGroup<CompositeDefinition> composite = (ModelViewItemGroup<CompositeDefinition>) viewItems.get(0);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeViewItems = composite.getViewItems();
        Assert.assertEquals(1, compositeViewItems.size());
        
        Assert.assertTrue(compositeViewItems.get(0) instanceof ModelViewItemGroup);
        ModelViewItemGroup<RepeaterDefinition> repeater = (ModelViewItemGroup<RepeaterDefinition>) compositeViewItems.get(0);
        Assert.assertEquals("repeater", repeater.getName());
        Assert.assertTrue(repeater.getViewItems().isEmpty());
    }
    
    /**
     * Tests the {@link DataHolderHelper#createViewItemAccessorFromValues(Collection, Map)} method
     * repeater value is empty -> the view item is still there
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testCreateViewItemAccessorFromValuesWithEmptyRepeater()
    {
        Map<String, Object> values = new LinkedHashMap<>();
        Map<String, Object> compositeValues = new LinkedHashMap<>();
        values.put("composite", compositeValues);
        List<Map<String, Object>> repeaterValues = new ArrayList<>();
        compositeValues.put("repeater", repeaterValues);
        
        View view = (View) DataHolderHelper.createViewItemAccessorFromValues(List.of(_model), values);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(1, viewItems.size());
        
        Assert.assertTrue(viewItems.get(0) instanceof ModelViewItemGroup);
        ModelViewItemGroup<CompositeDefinition> composite = (ModelViewItemGroup<CompositeDefinition>) viewItems.get(0);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeViewItems = composite.getViewItems();
        Assert.assertEquals(1, compositeViewItems.size());
        
        Assert.assertTrue(compositeViewItems.get(0) instanceof ModelViewItemGroup);
        ModelViewItemGroup<RepeaterDefinition> repeater = (ModelViewItemGroup<RepeaterDefinition>) compositeViewItems.get(0);
        Assert.assertEquals("repeater", repeater.getName());
        Assert.assertTrue(repeater.getViewItems().isEmpty());
    }
    
    /**
     * Tests the {@link DataHolderHelper#createViewItemAccessorFromValues(Collection, Map)} method
     * repeater value is a {@link SynchronizableRepeater} -> the view items are OK
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testCreateViewItemAccessorFromValuesWithSynchronizableRepeater()
    {
        Map<String, Object> values = new LinkedHashMap<>();
        Map<String, Object> compositeValues = new LinkedHashMap<>();
        values.put("composite", compositeValues);

        List<Map<String, Object>> repeaterEntries = new ArrayList<>();
        Map<String, Object> entry1 = new LinkedHashMap<>();
        repeaterEntries.add(entry1);
        entry1.put("singleBooleanData", true);
        Map<String, Object> entry1Subcomposite = new LinkedHashMap<>();
        entry1.put("subcomposite", entry1Subcomposite);
        entry1Subcomposite.put("singleStringData", "string value for entry 1");
        entry1Subcomposite.put("singleBinaryData", "binary value for entry 1");
        Map<String, Object> entry2 = new LinkedHashMap<>();
        repeaterEntries.add(entry2);
        entry2.put("singleBooleanData", false);
        Map<String, Object> entry2Subcomposite = new LinkedHashMap<>();
        entry2.put("subcomposite", entry2Subcomposite);
        entry2Subcomposite.put("singleStringData", "string value for entry 2");
        entry2Subcomposite.put("singleBinaryData", "binary value for entry 2");
        
        SynchronizableRepeater repeaterValues = SynchronizableRepeater.appendOrRemove(repeaterEntries, Set.of());
        compositeValues.put("repeater", repeaterValues);
        
        View view = (View) DataHolderHelper.createViewItemAccessorFromValues(List.of(_model), values);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(1, viewItems.size());
        
        Assert.assertTrue(viewItems.get(0) instanceof ModelViewItemGroup);
        ModelViewItemGroup<CompositeDefinition> composite = (ModelViewItemGroup<CompositeDefinition>) viewItems.get(0);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeViewItems = composite.getViewItems();
        Assert.assertEquals(1, compositeViewItems.size());
        
        Assert.assertTrue(compositeViewItems.get(0) instanceof ModelViewItemGroup);
        ModelViewItemGroup<RepeaterDefinition> repeater = (ModelViewItemGroup<RepeaterDefinition>) compositeViewItems.get(0);
        Assert.assertEquals("repeater", repeater.getName());
        List<ViewItem> repeaterViewItems = repeater.getViewItems();
        Assert.assertEquals(2, repeaterViewItems.size());
        
        Assert.assertTrue(repeaterViewItems.get(0) instanceof ViewElement);
        Assert.assertEquals("singleBooleanData", repeaterViewItems.get(0).getName());
        
        Assert.assertTrue(repeaterViewItems.get(1) instanceof ModelViewItemGroup);
        ModelViewItemGroup<CompositeDefinition> subcomposite = (ModelViewItemGroup<CompositeDefinition>) repeaterViewItems.get(1);
        Assert.assertEquals("subcomposite", subcomposite.getName());
        List<ViewItem> subcompositeViewItems = subcomposite.getViewItems();
        Assert.assertEquals(2, subcompositeViewItems.size());
        
        Assert.assertTrue(subcompositeViewItems.get(0) instanceof ViewElement);
        Assert.assertEquals("singleStringData", subcompositeViewItems.get(0).getName());
        
        Assert.assertTrue(subcompositeViewItems.get(1) instanceof ViewElement);
        Assert.assertEquals("singleBinaryData", subcompositeViewItems.get(1).getName());
    }
    
    /**
     * Tests the {@link DataHolderHelper#createViewItemAccessorFromValues(Collection, Map)} method
     * repeater value is an empty {@link SynchronizableRepeater} -> the view item is still there
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testCreateViewItemAccessorFromValuesWithEmptySynchronizableRepeater()
    {
        Map<String, Object> values = new LinkedHashMap<>();
        Map<String, Object> compositeValues = new LinkedHashMap<>();
        values.put("composite", compositeValues);

        SynchronizableRepeater repeaterValues = SynchronizableRepeater.appendOrRemove(List.of(), Set.of());
        compositeValues.put("repeater", repeaterValues);
        
        View view = (View) DataHolderHelper.createViewItemAccessorFromValues(List.of(_model), values);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(1, viewItems.size());
        
        Assert.assertTrue(viewItems.get(0) instanceof ModelViewItemGroup);
        ModelViewItemGroup<CompositeDefinition> composite = (ModelViewItemGroup<CompositeDefinition>) viewItems.get(0);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeViewItems = composite.getViewItems();
        Assert.assertEquals(1, compositeViewItems.size());
        
        Assert.assertTrue(compositeViewItems.get(0) instanceof ModelViewItemGroup);
        ModelViewItemGroup<RepeaterDefinition> repeater = (ModelViewItemGroup<RepeaterDefinition>) compositeViewItems.get(0);
        Assert.assertEquals("repeater", repeater.getName());
        Assert.assertTrue(repeater.getViewItems().isEmpty());
    }
    
    /**
     * Tests the {@link DataHolderHelper#createViewItemAccessorFromValues(Collection, Map)} method
     * Some values of the repeater appear only in one entry -> view items should be here anyway
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testCreateViewItemAccessorFromValuesValuesNotInAllRepeaterEntries()
    {
        Map<String, Object> values = new LinkedHashMap<>();
        Map<String, Object> compositeValues = new LinkedHashMap<>();
        values.put("composite", compositeValues);
        List<Map<String, Object>> repeaterValues = new ArrayList<>();
        compositeValues.put("repeater", repeaterValues);
        Map<String, Object> entry1 = new LinkedHashMap<>();
        repeaterValues.add(entry1);
//        entry1.put("singleBooleanData", true);
        Map<String, Object> entry1Subcomposite = new LinkedHashMap<>();
        entry1.put("subcomposite", entry1Subcomposite);
        entry1Subcomposite.put("singleStringData", "string value for entry 1");
//        entry1Subcomposite.put("singleBinaryData", "binary value for entry 1");
        Map<String, Object> entry2 = new LinkedHashMap<>();
        repeaterValues.add(entry2);
        entry2.put("singleBooleanData", false);
        Map<String, Object> entry2Subcomposite = new LinkedHashMap<>();
        entry2.put("subcomposite", entry2Subcomposite);
        entry2Subcomposite.put("singleStringData", "string value for entry 2");
        entry2Subcomposite.put("singleBinaryData", "binary value for entry 2");
        
        View view = (View) DataHolderHelper.createViewItemAccessorFromValues(List.of(_model), values);
        
        List<ViewItem> viewItems = view.getViewItems();
        Assert.assertEquals(1, viewItems.size());
        
        Assert.assertTrue(viewItems.get(0) instanceof ModelViewItemGroup);
        ModelViewItemGroup<CompositeDefinition> composite = (ModelViewItemGroup<CompositeDefinition>) viewItems.get(0);
        Assert.assertEquals("composite", composite.getName());
        List<ViewItem> compositeViewItems = composite.getViewItems();
        Assert.assertEquals(1, compositeViewItems.size());
        
        Assert.assertTrue(compositeViewItems.get(0) instanceof ModelViewItemGroup);
        ModelViewItemGroup<RepeaterDefinition> repeater = (ModelViewItemGroup<RepeaterDefinition>) compositeViewItems.get(0);
        Assert.assertEquals("repeater", repeater.getName());
        List<ViewItem> repeaterViewItems = repeater.getViewItems();
        Assert.assertEquals(2, repeaterViewItems.size());
        
        Assert.assertTrue(repeaterViewItems.get(0) instanceof ModelViewItemGroup);
        ModelViewItemGroup<CompositeDefinition> subcomposite = (ModelViewItemGroup<CompositeDefinition>) repeaterViewItems.get(0);
        Assert.assertEquals("subcomposite", subcomposite.getName());
        List<ViewItem> subcompositeViewItems = subcomposite.getViewItems();
        Assert.assertEquals(2, subcompositeViewItems.size());
        
        Assert.assertTrue(subcompositeViewItems.get(0) instanceof ViewElement);
        Assert.assertEquals("singleStringData", subcompositeViewItems.get(0).getName());
        
        Assert.assertTrue(subcompositeViewItems.get(1) instanceof ViewElement);
        Assert.assertEquals("singleBinaryData", subcompositeViewItems.get(1).getName());
        
        Assert.assertTrue(repeaterViewItems.get(1) instanceof ViewElement);
        Assert.assertEquals("singleBooleanData", repeaterViewItems.get(1).getName());
    }
}
