/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.holder.impl.DefaultModelAwareDataHolder;
import org.ametys.cms.data.holder.impl.DefaultModifiableModelAwareDataHolder;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.holder.DataHolder;
import org.ametys.plugins.repository.data.holder.ModifiableDataHolder;
import org.ametys.plugins.repository.data.holder.ModifiableModelAwareDataHolder;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.plugins.repository.data.type.RepositoryModelItemGroupType;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link DataHolder} implementation test case
 */
@NeedFeatures({
    "test-data/test-model",
    
    "repository/resolver", 
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class DataHolderTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    private static AmetysObjectResolver _ametysObjectResolver;
    
    private static Model _model;

    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;
    
    private static RepositoryElementType<Geocode> _geocodeType;
    private static RepositoryElementType<UserIdentity> _userType;
    private static RepositoryModelItemGroupType _compositeType;
    private static RepositoryModelItemGroupType _repeaterType;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
    }
    
    @Override
    @SuppressWarnings("unchecked")
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");
        _geocodeType = (RepositoryElementType<Geocode>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID);
        _userType = (RepositoryElementType<UserIdentity>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _compositeType = (RepositoryModelItemGroupType) _testDataTypeExtensionPoint.getExtension(org.ametys.plugins.repository.data.type.ModelItemTypeConstants.COMPOSITE_TYPE_ID);
        _repeaterType = (RepositoryModelItemGroupType) _testDataTypeExtensionPoint.getExtension(org.ametys.plugins.repository.data.type.ModelItemTypeConstants.REPEATER_TYPE_ID);
    }
    
    /**
     * Test {@link DataHolder#copyTo(ModifiableDataHolder)} implementation
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testDataHolderCopyWithJCRRepositoryData()
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderCopySource", DATA_NODE_TYPE);
        ModifiableRepositoryData destinationData = _jcrRoot.addRepositoryData("testDataHolderCopyDest", DATA_NODE_TYPE);
        _testDataHolderCopy(sourceData, destinationData);
        
    }
    
    /**
     * Test {@link DataHolder#copyTo(ModifiableDataHolder)} implementation
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testDataHolderCopyWithMemoryRepositoryData()
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderCopySource", DATA_NODE_TYPE);
        ModifiableRepositoryData destinationData = _memoryRoot.addRepositoryData("testDataHolderCopyDest", DATA_NODE_TYPE);
        _testDataHolderCopy(sourceData, destinationData);
    }
    
    private void _testDataHolderCopy(ModifiableRepositoryData sourceData, ModifiableRepositoryData destinationData)
    {
        // Add values in the source repository data
        _populateRepositoryData(sourceData);
        
        // Create the data holders
        DataHolder source = new DefaultModelAwareDataHolder(sourceData, _model);
        ModifiableModelAwareDataHolder destination = new DefaultModifiableModelAwareDataHolder(destinationData, _model);
        
        // Copy the source in the destination
        source.copyTo(destination);
        
        // Check the destination values
        Assert.assertEquals("value of the single string data", destination.getValue("singleStringData"));
        Assert.assertArrayEquals(new Double[] {32.5D, 784.96D}, destination.getValue("multipleDoubleData"));
        Geocode singleGeocodeData = destination.getValue("singleGeocodeData");
        Assert.assertEquals(Double.valueOf(65.2), singleGeocodeData.getLongitude());
        Assert.assertEquals(Double.valueOf(45.485), singleGeocodeData.getLatitude());
        Assert.assertArrayEquals(new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY}, destination.getValue("multipleUserData"));
        Assert.assertEquals(Long.valueOf(645), destination.getValue("composite/singleLongData"));
        Assert.assertTrue(destination.getValue("composite/repeater[1]/singleBooleanData"));
        Assert.assertFalse(destination.getValue("composite/repeater[2]/singleBooleanData"));
    }
    
    private void _populateRepositoryData(ModifiableRepositoryData repositoryData)
    {
        repositoryData.setValue("singleStringData", "value of the single string data");
        repositoryData.setValues("multipleDoubleData", new Double[] {32.5D, 784.96D});
        
        _geocodeType.write(repositoryData, "singleGeocodeData", new Geocode(45.485D, 65.2D));
        _userType.write(repositoryData, "multipleUserData", new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY});

        ModifiableRepositoryData composite = _compositeType.add(repositoryData, "composite");
        composite.setValue("singleLongData", 645L);
        
        ModifiableRepositoryData repeater = _repeaterType.add(composite, "repeater");
        ModifiableRepositoryData entry1 = _compositeType.add(repeater, "1");
        entry1.setValue("singleBooleanData", true);
        ModifiableRepositoryData entry2 = _compositeType.add(repeater, "2");
        entry2.setValue("singleBooleanData", false);
    }
}
