/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.util.Collection;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.holder.impl.DefaultModelAwareDataHolder;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModelAwareComposite;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryModelItemGroupType;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.ModelItemContainer;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelAwareDataHolder} basic methods implementation test case
 */
@NeedFeatures({
    "test-data/test-model",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModelAwareDataHolderBasicsTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    
    private static Model _model;
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    
    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;
    
    private static RepositoryModelItemGroupType _compositeType;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");
        
        _compositeType = (RepositoryModelItemGroupType) _testDataTypeExtensionPoint.getExtension(org.ametys.plugins.repository.data.type.ModelItemTypeConstants.COMPOSITE_TYPE_ID);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getDataNames()} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testGetDataNamesWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testGetDataNames", DATA_NODE_TYPE);
        _testGetDataNames(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getDataNames()} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testGetDataNamesWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testGetDataNames", DATA_NODE_TYPE);
        _testGetDataNames(sourceData);
    }
    
    private void _testGetDataNames(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Add some data in the repository
        repositoryData.setValue("singleStringData", "value of the single string data");
        repositoryData.setValues("multipleLongData", new Long[] {32L, 784L});
        ModifiableRepositoryData composite = _compositeType.add(repositoryData, "composite");
        composite.setValue("singleLongData", 54L);
        
        // Add also some data that are not defined by the model
        repositoryData.setValue("notDefined", "a data that is not defined by the model");
        
        Collection<String> dataNames = dataHolder.getDataNames();
        Assert.assertEquals(3, dataNames.size());
        Assert.assertTrue(dataNames.contains("singleStringData"));
        Assert.assertTrue(dataNames.contains("multipleLongData"));
        Assert.assertTrue(dataNames.contains("composite"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#isMultiple(String)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testIsMultipleWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testIsMultiple", DATA_NODE_TYPE);
        _testIsMultiple(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#isMultiple(String)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testIsMultipleWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testIsMultiple", DATA_NODE_TYPE);
        _testIsMultiple(sourceData);
    }
    
    private void _testIsMultiple(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        _addSomeValuesInRepositoryData(repositoryData);
        
        Assert.assertFalse(dataHolder.isMultiple("singleStringData"));
        Assert.assertTrue(dataHolder.isMultiple("multipleStringData"));
        Assert.assertFalse(dataHolder.isMultiple("singleLongData"));
        Assert.assertTrue(dataHolder.isMultiple("multipleLongData"));
        
        Assert.assertFalse(dataHolder.isMultiple("composite"));
        Assert.assertFalse(dataHolder.isMultiple("composite/singleLongData"));
        ModelAwareComposite composite = dataHolder.getComposite("composite");
        Assert.assertFalse(composite.isMultiple("singleLongData"));

        Assert.assertTrue(dataHolder.isMultiple("composite/repeater"));
        Assert.assertFalse(dataHolder.isMultiple("composite/repeater/singleDoubleData"));
        Assert.assertTrue(composite.isMultiple("repeater"));
        Assert.assertFalse(composite.isMultiple("repeater/singleDoubleData"));

        Assert.assertFalse(dataHolder.isMultiple("composite/repeater[1]"));
        Assert.assertFalse(dataHolder.isMultiple("composite/repeater[3]/singleDoubleData"));
        Assert.assertFalse(composite.isMultiple("repeater[145]/singleDoubleData"));

        Assert.assertFalse(dataHolder.isMultiple("composite/repeater/subcomposite"));
        Assert.assertFalse(dataHolder.isMultiple("composite/repeater/subcomposite/singleStringData"));
        Assert.assertTrue(dataHolder.isMultiple("composite/repeater/subcomposite/multipleDateData"));
        Assert.assertFalse(composite.isMultiple("repeater/subcomposite/singleStringData"));
        Assert.assertTrue(composite.isMultiple("repeater/subcomposite/multipleDateData"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getType(String)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testGetTypeWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testGetType", DATA_NODE_TYPE);
        _testGetType(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getType(String)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testGetTypeWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testGetType", DATA_NODE_TYPE);
        _testGetType(sourceData);
    }
    
    @SuppressWarnings("static-access")
    private void _testGetType(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        _addSomeValuesInRepositoryData(repositoryData);
        
        Assert.assertEquals(ModelItemTypeConstants.STRING_TYPE_ID, dataHolder.getType("singleStringData").getId());
        Assert.assertEquals(ModelItemTypeConstants.STRING_TYPE_ID, dataHolder.getType("multipleStringData").getId());
        Assert.assertEquals(ModelItemTypeConstants.LONG_TYPE_ID, dataHolder.getType("singleLongData").getId());
        
        Assert.assertEquals(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID, dataHolder.getType("singleUserData").getId());
        Assert.assertEquals(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID, dataHolder.getType("multipleUserData").getId());
        
        Assert.assertEquals(ModelItemTypeConstants.COMPOSITE_TYPE_ID, dataHolder.getType("composite").getId());
        Assert.assertEquals(ModelItemTypeConstants.LONG_TYPE_ID, dataHolder.getType("composite/singleLongData").getId());
        ModelAwareComposite composite = dataHolder.getComposite("composite");
        Assert.assertEquals(ModelItemTypeConstants.LONG_TYPE_ID, composite.getType("singleLongData").getId());

        Assert.assertEquals(ModelItemTypeConstants.REPEATER_TYPE_ID, dataHolder.getType("composite/repeater").getId());
        Assert.assertEquals(ModelItemTypeConstants.BOOLEAN_TYPE_ID, dataHolder.getType("composite/repeater/singleBooleanData").getId());
        Assert.assertEquals(ModelItemTypeConstants.BOOLEAN_TYPE_ID, composite.getType("repeater/singleBooleanData").getId());

        Assert.assertEquals(ModelItemTypeConstants.REPEATER_TYPE_ID, dataHolder.getType("composite/repeater[1]").getId());
        Assert.assertEquals(ModelItemTypeConstants.BOOLEAN_TYPE_ID, dataHolder.getType("composite/repeater[3]/singleBooleanData").getId());
        Assert.assertEquals(ModelItemTypeConstants.BOOLEAN_TYPE_ID, composite.getType("repeater[145]/singleBooleanData").getId());

        Assert.assertEquals(ModelItemTypeConstants.COMPOSITE_TYPE_ID, dataHolder.getType("composite/repeater/subcomposite").getId());
        Assert.assertEquals(ModelItemTypeConstants.STRING_TYPE_ID, dataHolder.getType("composite/repeater/subcomposite/singleStringData").getId());
        Assert.assertEquals(ModelItemTypeConstants.DATE_TYPE_ID, dataHolder.getType("composite/repeater/subcomposite/multipleDateData").getId());
        Assert.assertEquals(ModelItemTypeConstants.STRING_TYPE_ID, composite.getType("repeater/subcomposite/singleStringData").getId());
        Assert.assertEquals(ModelItemTypeConstants.DATE_TYPE_ID, composite.getType("repeater/subcomposite/multipleDateData").getId());
    }
    
    private void _addSomeValuesInRepositoryData(ModifiableRepositoryData repositoryData)
    {
        repositoryData.setValue("singleStringData", "my string");
        repositoryData.setValues("multipleStringData", new String[]{"string1", "string2", "string3"});
        
        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        composite.setValue("singleLongData", 156L);
        
        ModifiableRepositoryData repeater = composite.addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE);
        ModifiableRepositoryData entry1 = repeater.addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE);
        entry1.setValue("singleDoubleData", 12.3D);
        ModifiableRepositoryData subComposite1 = entry1.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subComposite1.setValue("singleStringData", "repeater entry 1 - sub composite - string value");
        
        ModifiableRepositoryData entry2 = repeater.addRepositoryData("2", RepositoryConstants.COMPOSITE_NODETYPE);
        entry2.setValue("singleDoubleData", 6.0D);
        ModifiableRepositoryData subComposite2 = entry2.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subComposite2.setValue("singleStringData", "repeater entry 2 - sub composite - string value");
    }
    
    /**
     * Test {@link UndefinedItemPathException} errors.
     * Test getValue with an unknown attribute
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testGetValueWithNoSuchItemErrorWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testGetValueWithNoSuchItemError", DATA_NODE_TYPE);
        _testGetValueWithNoSuchItemError(sourceData);
    }
    
    /**
     * Test {@link UndefinedItemPathException} errors
     * Test getValue with an unknown attribute
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testGetValueWithNoSuchItemErrorWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testGetValueWithNoSuchItemError", DATA_NODE_TYPE);
        _testGetValueWithNoSuchItemError(sourceData);
    }
    
    private void _testGetValueWithNoSuchItemError(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        UndefinedItemPathException noSuchItemThrows = Assert.assertThrows(UndefinedItemPathException.class, () -> dataHolder.getValue("undefinedPath"));
        String noSuchItemThrowsMsg = noSuchItemThrows.getMessage();
        Assert.assertTrue(noSuchItemThrowsMsg.contains("undefinedPath") && noSuchItemThrowsMsg.contains("no such item"));
    }
    
    /**
     * Test {@link UndefinedItemPathException} errors.
     * Test getValue with an empty model
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testGetValueWithNoModelDefinedErrorWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testGetValueWithNoNoModelDefinedError", DATA_NODE_TYPE);
        _testGetValueWithNoModelDefinedError(sourceData);
    }
    
    /**
     * Test {@link UndefinedItemPathException} errors
     * Test getValue with an empty model
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testGetValueWithNoModelDefinedErrorWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testGetValueWithNoNoModelDefinedError", DATA_NODE_TYPE);
        _testGetValueWithNoModelDefinedError(sourceData);
    }
    
    private void _testGetValueWithNoModelDefinedError(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder emptyDataHolder = new DefaultModelAwareDataHolder(repositoryData, new ModelItemContainer[0]);
        UndefinedItemPathException noModelThrows = Assert.assertThrows(UndefinedItemPathException.class, () -> emptyDataHolder.getValue("singleStringData"));
        String noModelThrowsMsg = noModelThrows.getMessage();
        Assert.assertTrue(noModelThrowsMsg.contains("singleStringData") && noModelThrowsMsg.contains("No model is defined"));
    }
    
}
