/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.TransformerHandler;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.xpath.XPathAPI;
import org.junit.Assert;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.ContentHandler;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.holder.impl.DefaultModelAwareDataHolder;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.DataComment;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelAwareDataHolder} methods reading data comments
 */
@NeedFeatures({
    "test-data/test-model",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModelAwareDataHolderCommentTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    
    private static final ZonedDateTime __DATE_TIME_1 = ZonedDateTime.of(2020, 7, 06, 16, 28, 10, 147_000_000, ZoneId.systemDefault());
    private static final ZonedDateTime __DATE_TIME_1_UTC = __DATE_TIME_1.withZoneSameInstant(ZoneOffset.UTC);
    private static final ZonedDateTime __DATE_TIME_2 = ZonedDateTime.of(2018, 9, 11, 12, 25, 17, 985_000_000, ZoneId.systemDefault());
    private static final ZonedDateTime __DATE_TIME_2_UTC = __DATE_TIME_2.withZoneSameInstant(ZoneOffset.UTC);
    
    
    private static final String __AUTHOR_1 = "mjagger";
    private static final String __AUTHOR_2 = "krichards";
    
    private static Model _model;
    private static AmetysObjectResolver _ametysObjectResolver;
    
    private static JCRRepositoryData _root;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _root = new JCRRepositoryData(rootAmetysObject.getNode());
    }
    
    /**
     * Tests the {@link ModelAwareDataHolder#commentsToSAX(ContentHandler, ViewItemAccessor)} implementation
     * @throws Exception if an error occurs
     */
    @Test
    public void testCommentToSAX() throws Exception
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testGetValueOnDataWithoutStatus", DATA_NODE_TYPE);
        
        // Add comments on data
        _addCommentsToData(repositoryData, "singleStringData", new DataComment("comment 1 on single String data", __DATE_TIME_1, __AUTHOR_1), new DataComment("comment 2 on single String data", __DATE_TIME_2, __AUTHOR_2));
        _addCommentsToData(repositoryData, "composite", new DataComment("comment on composite", __DATE_TIME_1, __AUTHOR_1));
        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        _addCommentsToData(composite, "singleLongData", new DataComment("comment on single Long data", __DATE_TIME_1, __AUTHOR_1));
        _addCommentsToData(composite, "repeater", new DataComment("comment on repeater", __DATE_TIME_1, __AUTHOR_2));
        ModifiableRepositoryData repeater = composite.addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE);
        ModifiableRepositoryData entry1 = repeater.addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE);
        _addCommentsToData(entry1, "singleDoubleData", new DataComment("comment on single Double data in first repeater entry", __DATE_TIME_2, __AUTHOR_1));
        ModifiableRepositoryData entry2 = repeater.addRepositoryData("2", RepositoryConstants.COMPOSITE_NODETYPE);
        _addCommentsToData(entry2, "singleDoubleData", new DataComment("comment on single Double data in second repeater entry", __DATE_TIME_2, __AUTHOR_2));
        
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        View view = View.of(_model);
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.commentsToSAX(handler, view);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 6);
        _assertXpathEquals(document, "count(/result/metadata)", 6);
        
        _assertXpathEquals(document, "/result/metadata[1]/@path", "singleStringData");
        _assertXpathEquals(document, "count(/result/metadata[1]/*)", 2);
        _assertXpathEquals(document, "count(/result/metadata[1]/comment)", 2);
        _assertXpathEquals(document, "/result/metadata[1]/comment[@id='1']/@date", DateUtils.zonedDateTimeToString(__DATE_TIME_1_UTC));
        _assertXpathEquals(document, "/result/metadata[1]/comment[@id='1']/@author", __AUTHOR_1);
        _assertXpathEquals(document, "/result/metadata[1]/comment[@id='1']", "comment 1 on single String data");
        _assertXpathEquals(document, "/result/metadata[1]/comment[@id='2']/@date", DateUtils.zonedDateTimeToString(__DATE_TIME_2_UTC));
        _assertXpathEquals(document, "/result/metadata[1]/comment[@id='2']/@author", __AUTHOR_2);
        _assertXpathEquals(document, "/result/metadata[1]/comment[@id='2']", "comment 2 on single String data");
        
        _assertXpathEquals(document, "/result/metadata[2]/@path", "composite");
        _assertXpathEquals(document, "count(/result/metadata[2]/*)", 1);
        _assertXpathEquals(document, "count(/result/metadata[2]/comment)", 1);
        _assertXpathEquals(document, "/result/metadata[2]/comment/@id", 1);
        _assertXpathEquals(document, "/result/metadata[2]/comment/@date", DateUtils.zonedDateTimeToString(__DATE_TIME_1_UTC));
        _assertXpathEquals(document, "/result/metadata[2]/comment/@author", __AUTHOR_1);
        _assertXpathEquals(document, "/result/metadata[2]/comment", "comment on composite");
        
        _assertXpathEquals(document, "/result/metadata[3]/@path", "composite/singleLongData");
        _assertXpathEquals(document, "count(/result/metadata[3]/*)", 1);
        _assertXpathEquals(document, "count(/result/metadata[3]/comment)", 1);
        _assertXpathEquals(document, "/result/metadata[3]/comment/@id", 1);
        _assertXpathEquals(document, "/result/metadata[3]/comment/@date", DateUtils.zonedDateTimeToString(__DATE_TIME_1_UTC));
        _assertXpathEquals(document, "/result/metadata[3]/comment/@author", __AUTHOR_1);
        _assertXpathEquals(document, "/result/metadata[3]/comment", "comment on single Long data");
        
        _assertXpathEquals(document, "/result/metadata[4]/@path", "composite/repeater");
        _assertXpathEquals(document, "count(/result/metadata[4]/*)", 1);
        _assertXpathEquals(document, "count(/result/metadata[4]/comment)", 1);
        _assertXpathEquals(document, "/result/metadata[4]/comment/@id", 1);
        _assertXpathEquals(document, "/result/metadata[4]/comment/@date", DateUtils.zonedDateTimeToString(__DATE_TIME_1_UTC));
        _assertXpathEquals(document, "/result/metadata[4]/comment/@author", __AUTHOR_2);
        _assertXpathEquals(document, "/result/metadata[4]/comment", "comment on repeater");
        
        _assertXpathEquals(document, "/result/metadata[5]/@path", "composite/repeater[1]/singleDoubleData");
        _assertXpathEquals(document, "count(/result/metadata[5]/*)", 1);
        _assertXpathEquals(document, "count(/result/metadata[5]/comment)", 1);
        _assertXpathEquals(document, "/result/metadata[5]/comment/@id", 1);
        _assertXpathEquals(document, "/result/metadata[5]/comment/@date", DateUtils.zonedDateTimeToString(__DATE_TIME_2_UTC));
        _assertXpathEquals(document, "/result/metadata[5]/comment/@author", __AUTHOR_1);
        _assertXpathEquals(document, "/result/metadata[5]/comment", "comment on single Double data in first repeater entry");
        
        _assertXpathEquals(document, "/result/metadata[6]/@path", "composite/repeater[2]/singleDoubleData");
        _assertXpathEquals(document, "count(/result/metadata[6]/*)", 1);
        _assertXpathEquals(document, "count(/result/metadata[6]/comment)", 1);
        _assertXpathEquals(document, "/result/metadata[6]/comment/@id", 1);
        _assertXpathEquals(document, "/result/metadata[6]/comment/@date", DateUtils.zonedDateTimeToString(__DATE_TIME_2_UTC));
        _assertXpathEquals(document, "/result/metadata[6]/comment/@author", __AUTHOR_2);
        _assertXpathEquals(document, "/result/metadata[6]/comment", "comment on single Double data in second repeater entry");
    }
    
    private void _addCommentsToData(ModifiableRepositoryData repositoryData, String dataName, DataComment... comments)
    {
        ModifiableRepositoryData commentsData = repositoryData.addRepositoryData(dataName + ModelAwareDataHolder.COMMENTS_SUFFIX, RepositoryConstants.COMPOSITE_METADTA_NODETYPE, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        
        int commentId = 1;
        for (DataComment comment : comments)
        {
            ModifiableRepositoryData commentData = commentsData.addRepositoryData(String.valueOf(commentId++), RepositoryConstants.COMPOSITE_METADTA_NODETYPE, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
            
            commentData.setValue("comment", comment.getComment());
            commentData.setValue("date", DateUtils.asCalendar(comment.getDate()));
            commentData.setValue("author", comment.getAuthor());
        }
    }
    
    private void _assertXpathEquals(Node node, String expression, String value) throws Exception
    {
        Assert.assertEquals(value, XPathAPI.eval(node, expression).str());
    }
    
    private void _assertXpathEquals(Node node, String expression, long value) throws Exception
    {
        Assert.assertEquals(value, (long) XPathAPI.eval(node, expression).num());
    }
}
