/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.io.IOException;
import java.io.OutputStream;
import java.io.StringWriter;
import java.time.ZoneId;
import java.time.ZonedDateTime;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.io.IOUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Binary;
import org.ametys.cms.data.RichText;
import org.ametys.cms.data.holder.impl.DefaultModelAwareDataHolder;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.external.ExternalizableDataProvider.ExternalizableDataStatus;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelAwareDataHolder} methods reading externalizable data
 */
@NeedFeatures({
    "test-data/test-model",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModelAwareDataHolderExternalizableDataTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    private static final ZonedDateTime __RESOURCE_LAST_MODIFIED = ZonedDateTime.of(2020, 02, 03, 15, 01, 58, 0, ZoneId.systemDefault());
    
    private static Model _model;
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    
    private static JCRRepositoryData _root;
    
    private static RepositoryElementType<Binary> _binaryType;
    private static RepositoryElementType<RichText> _richTextType;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public void prepareComponents() throws Exception
    {
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _root = new JCRRepositoryData(rootAmetysObject.getNode());
        
        _binaryType = (RepositoryElementType<Binary>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _richTextType = (RepositoryElementType<RichText>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID);
    }
    
    /**
     * Tests on the {@link ModelAwareDataHolder#getStatus(String)}, {@link ModelAwareDataHolder#getLocalValue(String)} and {@link ModelAwareDataHolder#getExternalValue(String)} methods
     * Tests on data without status
     * @throws IOException If an error occurs while tests resources data
     */
    @Test
    public void testGetValueOnDataWithoutStatus() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testGetValueOnDataWithoutStatus", DATA_NODE_TYPE);
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData", "string value");
        
        // Read - string
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, dataHolder.getStatus("singleStringData"));
        Assert.assertEquals("string value", dataHolder.getLocalValue("singleStringData"));
        Assert.assertNull(dataHolder.getExternalValue("singleStringData"));
        
        // Write - multiple long
        Long[] longArray = new Long[]{65L, 98L, 432L};
        repositoryData.setValues("multipleLongData", longArray);
        
        // Read - multiple long
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, dataHolder.getStatus("multipleLongData"));
        Assert.assertArrayEquals(longArray, dataHolder.getLocalValue("multipleLongData"));
        Assert.assertNull(dataHolder.getExternalValue("multipleLongData"));
        
        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my binary content"));
        
        // Read - binary
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, dataHolder.getStatus("singleBinaryData"));
        _assertBinaryContent("my binary content", dataHolder.getLocalValue("singleBinaryData"));
        Assert.assertNull(dataHolder.getExternalValue("singleBinaryData"));
        
        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my rich text content"));
        
        // Read - rich text
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, dataHolder.getStatus("singleRichTextData"));
        _assertRichTextContent("my rich text content", dataHolder.getLocalValue("singleRichTextData"));
        Assert.assertNull(dataHolder.getExternalValue("singleRichTextData"));
        
        // Write - composite
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 42L);
        
        // Read - composite
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, dataHolder.getStatus("composite"));
        Assert.assertEquals(Long.valueOf(42), dataHolder.getLocalComposite("composite").getValue("singleLongData"));
        Assert.assertNull(dataHolder.getExternalComposite("composite"));
        
        // Write - composite/repeater[1]/subcomposite/singleStringData
        repositoryData.getRepositoryData("composite")
            .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
            .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
            .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE)
            .setValue("singleStringData", "string value");
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, dataHolder.getStatus("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertEquals("string value", dataHolder.getLocalValue("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertNull(dataHolder.getExternalValue("composite/repeater[1]/subcomposite/singleStringData"));
    }
    
    /**
     * Tests on the {@link ModelAwareDataHolder#getStatus(String)}, {@link ModelAwareDataHolder#getLocalValue(String)} and {@link ModelAwareDataHolder#getExternalValue(String)} methods
     * Tests on local data
     * @throws IOException If an error occurs while tests resources data
     */
    @Test
    public void testGetValueOnLocalData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testGetValueOnLocalData", DATA_NODE_TYPE);
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData", "string local value");
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string external value");
        setStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Read - string
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, dataHolder.getStatus("singleStringData"));
        Assert.assertEquals("string local value", dataHolder.getLocalValue("singleStringData"));
        Assert.assertEquals("string external value", dataHolder.getExternalValue("singleStringData"));
        
        // Write - multiple long
        Long[] localLongArray = new Long[]{78L, 1L};
        repositoryData.setValues("multipleLongData", localLongArray);
        Long[] externalLongArray = new Long[]{65L, 98L, 432L};
        repositoryData.setValues("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, externalLongArray);
        setStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        
        // Read - multiple long
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, dataHolder.getStatus("multipleLongData"));
        Assert.assertArrayEquals(localLongArray, dataHolder.getLocalValue("multipleLongData"));
        Assert.assertArrayEquals(externalLongArray, dataHolder.getExternalValue("multipleLongData"));
        
        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my local binary content"));
        _binaryType.write(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createBinary("my external binary content"));
        setStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);
        
        // Binary - read
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, dataHolder.getStatus("singleBinaryData"));
        _assertBinaryContent("my local binary content", dataHolder.getLocalValue("singleBinaryData"));
        _assertBinaryContent("my external binary content", dataHolder.getExternalValue("singleBinaryData"));
        
        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my local rich text content"));
        _richTextType.write(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createRichText("my external rich text content"));
        setStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);
        
        // Read - rich text
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, dataHolder.getStatus("singleRichTextData"));
        _assertRichTextContent("my local rich text content", dataHolder.getLocalValue("singleRichTextData"));
        _assertRichTextContent("my external rich text content", dataHolder.getExternalValue("singleRichTextData"));
        
        // Write - composite
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 42L);
        repositoryData.addRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 18L);
        setStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);
        
        // Read - composite
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, dataHolder.getStatus("composite"));
        Assert.assertEquals(Long.valueOf(42), dataHolder.getLocalComposite("composite").getValue("singleLongData"));
        Assert.assertEquals(Long.valueOf(18), dataHolder.getExternalComposite("composite").getValue("singleLongData"));
        
        // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.getRepositoryData("composite")
            .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
            .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
            .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData", "string local value");
        subcomposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string external value");
        setStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, dataHolder.getStatus("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertEquals("string local value", dataHolder.getLocalValue("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertEquals("string external value", dataHolder.getExternalValue("composite/repeater[1]/subcomposite/singleStringData"));
    }
    
    /**
     * Tests on the {@link ModelAwareDataHolder#getStatus(String)}, {@link ModelAwareDataHolder#getLocalValue(String)} and {@link ModelAwareDataHolder#getExternalValue(String)} methods
     * Tests on external data
     * @throws IOException If an error occurs while tests resources data
     */
    @Test
    public void testGetValueOnExternalData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testGetValueOnExternalData", DATA_NODE_TYPE);
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string local value");
        repositoryData.setValue("singleStringData", "string external value");
        setStatus(repositoryData, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - string
        Assert.assertEquals(ExternalizableDataStatus.EXTERNAL, dataHolder.getStatus("singleStringData"));
        Assert.assertEquals("string local value", dataHolder.getLocalValue("singleStringData"));
        Assert.assertEquals("string external value", dataHolder.getExternalValue("singleStringData"));
        
        // Write - multiple long
        Long[] localLongArray = new Long[]{78L, 1L};
        repositoryData.setValues("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, localLongArray);
        Long[] externalLongArray = new Long[]{65L, 98L, 432L};
        repositoryData.setValues("multipleLongData", externalLongArray);
        setStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - multiple long
        Assert.assertEquals(ExternalizableDataStatus.EXTERNAL, dataHolder.getStatus("multipleLongData"));
        Assert.assertArrayEquals(localLongArray, dataHolder.getLocalValue("multipleLongData"));
        Assert.assertArrayEquals(externalLongArray, dataHolder.getExternalValue("multipleLongData"));
        
        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createBinary("my local binary content"));
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my external binary content"));
        setStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        
        // Binary - read
        Assert.assertEquals(ExternalizableDataStatus.EXTERNAL, dataHolder.getStatus("singleBinaryData"));
        _assertBinaryContent("my local binary content", dataHolder.getLocalValue("singleBinaryData"));
        _assertBinaryContent("my external binary content", dataHolder.getExternalValue("singleBinaryData"));
        
        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createRichText("my local rich text content"));
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my external rich text content"));
        setStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - rich text
        Assert.assertEquals(ExternalizableDataStatus.EXTERNAL, dataHolder.getStatus("singleRichTextData"));
        _assertRichTextContent("my local rich text content", dataHolder.getLocalValue("singleRichTextData"));
        _assertRichTextContent("my external rich text content", dataHolder.getExternalValue("singleRichTextData"));
        
        // Write - composite
        repositoryData.addRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 42L);
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 18L);
        setStatus(repositoryData, "composite", ExternalizableDataStatus.EXTERNAL);
        
        // Read - composite
        Assert.assertEquals(ExternalizableDataStatus.EXTERNAL, dataHolder.getStatus("composite"));
        Assert.assertEquals(Long.valueOf(42), dataHolder.getLocalComposite("composite").getValue("singleLongData"));
        Assert.assertEquals(Long.valueOf(18), dataHolder.getExternalComposite("composite").getValue("singleLongData"));
        
        // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.getRepositoryData("composite")
            .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
            .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
            .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string local value");
        subcomposite.setValue("singleStringData", "string external value");
        setStatus(subcomposite, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        Assert.assertEquals(ExternalizableDataStatus.EXTERNAL, dataHolder.getStatus("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertEquals("string local value", dataHolder.getLocalValue("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertEquals("string external value", dataHolder.getExternalValue("composite/repeater[1]/subcomposite/singleStringData"));
    }
    
    /**
     * Tests on the {@link ModelAwareDataHolder#hasLocalValue(String)} and {@link ModelAwareDataHolder#hasExternalValue(String)} methods
     */
    @Test
    public void testHasValue()
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testGetValueOnDataWithoutStatus", DATA_NODE_TYPE);
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // No values set - local and external are false
        Assert.assertFalse(dataHolder.hasLocalValue("singleStringData"));
        Assert.assertFalse(dataHolder.hasExternalValue("singleStringData"));
        
        // Set value without status (default is local)
        repositoryData.setValue("singleStringData", "string value");
        // Local is true, external is false
        Assert.assertTrue(dataHolder.hasLocalValue("singleStringData"));
        Assert.assertFalse(dataHolder.hasExternalValue("singleStringData"));
        
        // Set status to local
        setStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        // Local is true, external is false
        Assert.assertTrue(dataHolder.hasLocalValue("singleStringData"));
        Assert.assertFalse(dataHolder.hasExternalValue("singleStringData"));
        
        // Set status to external
        setStatus(repositoryData, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        // Local is false, external is true
        Assert.assertFalse(dataHolder.hasLocalValue("singleStringData"));
        Assert.assertTrue(dataHolder.hasExternalValue("singleStringData"));
        
        // Add an alternative data
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "alt string value");
        // Local and external are true
        Assert.assertTrue(dataHolder.hasLocalValue("singleStringData"));
        Assert.assertTrue(dataHolder.hasExternalValue("singleStringData"));
    }
    
    /**
     * Tests on the {@link ModelAwareDataHolder#hasLocalValue(String)} and {@link ModelAwareDataHolder#hasExternalValue(String)} methods with data in a group
     */
    @Test
    public void testHasValueInComposite()
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testGetValueOnDataWithoutStatus", DATA_NODE_TYPE);
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // No values set - local and external are false
        Assert.assertFalse(dataHolder.hasLocalValue("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertFalse(dataHolder.hasExternalValue("composite/repeater[1]/subcomposite/singleStringData"));
        
        // Set value without status (default is local)
        ModifiableRepositoryData subcomposite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE)
                .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
                .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
                .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData", "string value");
        // Local is true, external is false
        Assert.assertTrue(dataHolder.hasLocalValue("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertFalse(dataHolder.hasExternalValue("composite/repeater[1]/subcomposite/singleStringData"));
        
        // Set status to local
        setStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        // Local is true, external is false
        Assert.assertTrue(dataHolder.hasLocalValue("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertFalse(dataHolder.hasExternalValue("composite/repeater[1]/subcomposite/singleStringData"));
        
        // Set status to external
        setStatus(subcomposite, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        // Local is false, external is true
        Assert.assertFalse(dataHolder.hasLocalValue("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertTrue(dataHolder.hasExternalValue("composite/repeater[1]/subcomposite/singleStringData"));
        
        // Add an alternative data
        subcomposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "alt string value");
        // Local and external are true
        Assert.assertTrue(dataHolder.hasLocalValue("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertTrue(dataHolder.hasExternalValue("composite/repeater[1]/subcomposite/singleStringData"));
    }
    
    private Binary _createBinary(String content) throws IOException
    {
        Binary binary = new Binary();
        try (OutputStream os = binary.getOutputStream())
        {
            os.write(content.getBytes());
        }
        binary.setEncoding("UTF-8");
        binary.setFilename("myExternalBinary.txt");
        binary.setLastModificationDate(__RESOURCE_LAST_MODIFIED);
        binary.setMimeType("text/plain");
        
        return binary;
    }
    
    private void _assertBinaryContent(String content, Binary binary) throws IOException
    {
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(binary.getInputStream(), writer, binary.getEncoding());
            Assert.assertEquals(content, writer.toString());
        }
    }
    
    private RichText _createRichText(String content) throws IOException
    {
        RichText richText = new RichText();
        try (OutputStream os = richText.getOutputStream())
        {
            os.write(content.getBytes());
        }
        richText.setEncoding("UTF-8");
        richText.setMimeType("text/xml");
        richText.setLastModificationDate(__RESOURCE_LAST_MODIFIED);
        
        return richText;
    }
    
    private void _assertRichTextContent(String content, RichText richText) throws AmetysRepositoryException, IOException
    {
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(richText.getInputStream(), writer, richText.getEncoding());
            Assert.assertEquals(content, writer.toString());
        }
    }
    
    private void setStatus(ModifiableRepositoryData repositoryData, String dataName, ExternalizableDataStatus status)
    {
        repositoryData.setValue(dataName + ModelAwareDataHolder.STATUS_SUFFIX, status.name().toLowerCase());
    }
    
}
