/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.time.LocalDate;
import java.time.Month;
import java.time.ZonedDateTime;
import java.util.Calendar;
import java.util.Locale;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.ArrayUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.holder.impl.DefaultModelAwareDataHolder;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeaterEntry;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeater;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelAwareDataHolder#getValue(String, boolean)} method implementation test case with the allowMultiValuedPathSegments set to true 
 */
@NeedFeatures({
    "test-data/test-model",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModelAwareDataHolderGetValueProcessingMultiplesTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    
    private static Model _model;
    private static AmetysObjectResolver _ametysObjectResolver;
    
    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for simple data
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSingleDataWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSingleData", DATA_NODE_TYPE);
        _testSingleData(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for simple data
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSingleDataWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSingleData", DATA_NODE_TYPE);
        _testSingleData(repositoryData);
    }
    
    private void _testSingleData(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Write
        repositoryData.setValue("singleStringData", "single string value");
                
        // Read
        String value = dataHolder.getValue("singleStringData", true);
        Assert.assertEquals("single string value", value);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for a repeater in an empty composite
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testRepeaterInEmptyCompositeWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testRepeaterInEmptyComposite", DATA_NODE_TYPE);
        _testRepeaterInEmptyComposite(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for a repeater in an empty composite
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testRepeaterInEmptyCompositeWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testRepeaterInEmptyComposite", DATA_NODE_TYPE);
        _testRepeaterInEmptyComposite(repositoryData);
    }
    
    private void _testRepeaterInEmptyComposite(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Write - Do nothing
                
        // Read
        Assert.assertNull(dataHolder.getValue("composite/repeater/singleDoubleData", true));
        
        // Write - create the composite but kepp the repeater empty
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        
        // Read
        Double[] values = dataHolder.getValue("composite/repeater/singleDoubleData", true);
        Assert.assertEquals(0, values.length);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for data in a repeater
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testDataInRepeaterWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testDataInRepeater", DATA_NODE_TYPE);
        _testDataInRepeater(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for data in a repeater
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testDataInRepeaterWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testDataInRepeater", DATA_NODE_TYPE);
        _testDataInRepeater(repositoryData);
    }
    
    private void _testDataInRepeater(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Write
        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        ModifiableRepositoryData repeater = composite.addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE);
        ModifiableRepositoryData entry1 = repeater.addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE);
        entry1.setValue("singleDoubleData", 15D);
        ModifiableRepositoryData entry2 = repeater.addRepositoryData("2", RepositoryConstants.COMPOSITE_NODETYPE);
        entry2.setValue("singleDoubleData", 20D);
        ModifiableRepositoryData entry3 = repeater.addRepositoryData("3", RepositoryConstants.COMPOSITE_NODETYPE);
        entry3.setValue("singleDoubleData", 25D);
                
        // Read
        Double[] values = dataHolder.getValue("composite/repeater/singleDoubleData", true);
        Assert.assertEquals(3, values.length);
        Assert.assertTrue(ArrayUtils.contains(values, 15D));
        Assert.assertTrue(ArrayUtils.contains(values, 20D));
        Assert.assertTrue(ArrayUtils.contains(values, 25D));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for composite data in a repeater
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testCompositeDataInRepeaterWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testCompositeDataInRepeater", DATA_NODE_TYPE);
        _testCompositeDataInRepeater(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for composite data in a repeater
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testCompositeDataInRepeaterWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testCompositeDataInRepeater", DATA_NODE_TYPE);
        _testCompositeDataInRepeater(repositoryData);
    }
    
    private void _testCompositeDataInRepeater(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Write
        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        ModifiableRepositoryData repeater = composite.addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE);
        ModifiableRepositoryData entry1 = repeater.addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE);
        entry1.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        ModifiableRepositoryData entry2 = repeater.addRepositoryData("2", RepositoryConstants.COMPOSITE_NODETYPE);
        entry2.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
                
        // Read
        ModelAwareComposite[] values = dataHolder.getValue("composite/repeater/subcomposite", true);
        Assert.assertEquals(2, values.length);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for multiple data in a repeater
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testMultipleDataInRepeaterWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testMultipleDataInRepeater", DATA_NODE_TYPE);
        _testMultipleDataInRepeater(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for multiple data in a repeater
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testMultipleDataInRepeaterWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testMultipleDataInRepeater", DATA_NODE_TYPE);
        _testMultipleDataInRepeater(repositoryData);
    }
    
    private void _testMultipleDataInRepeater(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Write
        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        ModifiableRepositoryData repeater = composite.addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE);
        ModifiableRepositoryData entry1 = repeater.addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE);
        ModifiableRepositoryData subcomposite1 = entry1.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite1.setValues("multipleDateData", new Calendar[] {DateUtils.asCalendar(LocalDate.of(2020, Month.JANUARY, 2))});
        ModifiableRepositoryData entry2 = repeater.addRepositoryData("2", RepositoryConstants.COMPOSITE_NODETYPE);
        ModifiableRepositoryData subcomposite2 = entry2.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite2.setValues("multipleDateData", new Calendar[] {DateUtils.asCalendar(LocalDate.of(2019, Month.DECEMBER, 24)), DateUtils.asCalendar(LocalDate.of(2019, Month.DECEMBER, 25))});
                
        // Read
        LocalDate[] values = dataHolder.getValue("composite/repeater/subcomposite/multipleDateData", true);
        Assert.assertEquals(3, values.length);
        Assert.assertEquals(2, values[0].getDayOfMonth());
        Assert.assertEquals(Month.JANUARY, values[0].getMonth());
        Assert.assertEquals(2020, values[0].getYear());
        Assert.assertEquals(24, values[1].getDayOfMonth());
        Assert.assertEquals(Month.DECEMBER, values[1].getMonth());
        Assert.assertEquals(2019, values[1].getYear());
        Assert.assertEquals(25, values[2].getDayOfMonth());
        Assert.assertEquals(Month.DECEMBER, values[2].getMonth());
        Assert.assertEquals(2019, values[2].getYear());
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for data in multiple content data
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    @NeedFeatures({
        "cms/factories",
        "exhaustif/content-type",
    })
    public void testMultipleContentDataWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testMultipleContentData", DATA_NODE_TYPE);
        _testMultipleContentData(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for data in multiple content data
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    @NeedFeatures({
        "cms/factories",
        "exhaustif/content-type",
    })
    public void testMultipleContentDataWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testMultipleContentData", DATA_NODE_TYPE);
        _testMultipleContentData(repositoryData);
    }
    
    private void _testMultipleContentData(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Write
        ModifiableContent content1 = _createExhaustiveContent("content1");
        content1.setValue("string", "content 1 string value");
        ModifiableContent content2 = _createExhaustiveContent("content2");
        content2.setValue("string", "content 2 string value");
        repositoryData.setValues("multipleExhContentData", new String[] {content1.getId(), content2.getId()});
                
        // Read
        String[] values = dataHolder.getValue("multipleExhContentData/string", true);
        Assert.assertEquals(2, values.length);
        Assert.assertTrue(ArrayUtils.contains(values, "content 1 string value"));
        Assert.assertTrue(ArrayUtils.contains(values, "content 2 string value"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for repeater entry in multiple content data
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    @NeedFeatures({
        "cms/factories",
        "exhaustif/content-type",
    })
    public void testRepeaterEntryInMultipleContentDataWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testRepeaterEntryInMultipleContentData", DATA_NODE_TYPE);
        _testRepeaterEntryInMultipleContentData(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for repeater entry in multiple content data
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    @NeedFeatures({
        "cms/factories",
        "exhaustif/content-type",
    })
    public void testRepeaterEntryInMultipleContentDataWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testRepeaterEntryInMultipleContentData", DATA_NODE_TYPE);
        _testRepeaterEntryInMultipleContentData(repositoryData);
    }
    
    private void _testRepeaterEntryInMultipleContentData(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Write
        ModifiableContent content1 = _createExhaustiveContent("content1");
        ModifiableModelAwareComposite composite1 = content1.getComposite("composite", true);
        ModifiableModelAwareRepeater repeater1 = composite1.getRepeater("repeater", true);
        repeater1.addEntry();
        repeater1.addEntry();
        ModifiableContent content2 = _createExhaustiveContent("content2");
        ModifiableModelAwareComposite composite2 = content2.getComposite("composite", true);
        ModifiableModelAwareRepeater repeater2 = composite2.getRepeater("repeater", true);
        repeater2.addEntry();
        ModifiableContent content3 = _createExhaustiveContent("content3");
        ModifiableModelAwareComposite composite3 = content3.getComposite("composite", true);
        ModifiableModelAwareRepeater repeater3 = composite3.getRepeater("repeater", true);
        repeater3.addEntry();
        repeater3.addEntry();
        repeater3.addEntry();
        repositoryData.setValues("multipleExhContentData", new String[] {content1.getId(), content2.getId(), content3.getId()});
                
        // Read
        ModelAwareRepeaterEntry[] values = dataHolder.getValue("multipleExhContentData/composite/repeater[2]", true);
        Assert.assertEquals(2, values.length);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for repeater entry in multiple content data
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    @NeedFeatures({
        "cms/factories",
        "exhaustif/content-type",
    })
    public void testDataInRepeaterInMultipleContentDataWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testDataInRepeaterInMultipleContentData", DATA_NODE_TYPE);
        _testDataInRepeaterInMultipleContentData(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String, boolean)} method for repeater entry in multiple content data
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    @NeedFeatures({
        "cms/factories",
        "exhaustif/content-type",
    })
    public void testDataInRepeaterInMultipleContentDataWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testDataInRepeaterInMultipleContentData", DATA_NODE_TYPE);
        _testDataInRepeaterInMultipleContentData(repositoryData);
    }
    
    private void _testDataInRepeaterInMultipleContentData(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Write
        ModifiableContent content1 = _createExhaustiveContent("content1");
        ModifiableModelAwareComposite composite1 = content1.getComposite("composite", true);
        ModifiableModelAwareRepeater repeater1 = composite1.getRepeater("repeater", true);
        repeater1.addEntry().setValue("comp-string", "content 1 - composite - repeater - entry 1 - string value");
        repeater1.addEntry().setValue("comp-string", "content 1 - composite - repeater - entry 2 - string value");
        ModifiableContent content2 = _createExhaustiveContent("content2");
        ModifiableModelAwareComposite composite2 = content2.getComposite("composite", true);
        ModifiableModelAwareRepeater repeater2 = composite2.getRepeater("repeater", true);
        repeater2.addEntry().setValue("comp-string", "content 2 - composite - repeater - entry 1 - string value");
        repeater2.addEntry().setValue("comp-string", "content 2 - composite - repeater - entry 2 - string value");
        repeater2.addEntry().setValue("comp-string", "content 2 - composite - repeater - entry 3 - string value");
        repositoryData.setValues("multipleExhContentData", new String[] {content1.getId(), content2.getId()});
                
        // Read
        String[] values = dataHolder.getValue("multipleExhContentData/composite/repeater/comp-string", true);
        Assert.assertEquals(5, values.length);
        Assert.assertTrue(ArrayUtils.contains(values, "content 1 - composite - repeater - entry 1 - string value"));
        Assert.assertTrue(ArrayUtils.contains(values, "content 1 - composite - repeater - entry 2 - string value"));
        Assert.assertTrue(ArrayUtils.contains(values, "content 2 - composite - repeater - entry 1 - string value"));
        Assert.assertTrue(ArrayUtils.contains(values, "content 2 - composite - repeater - entry 2 - string value"));
        Assert.assertTrue(ArrayUtils.contains(values, "content 2 - composite - repeater - entry 3 - string value"));
    }
    
    private ModifiableContent _createExhaustiveContent(String title)
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        
        ModifiableContent content = root.createChild(title, "ametys:defaultContent");

        content.setTypes(new String[]{"org.ametys.test.Content.exhaustive"});
        content.setTitle(title, Locale.FRANCE);
        content.setLanguage("fr");
        content.setCreationDate(ZonedDateTime.now());
        content.setCreator(new UserIdentity("john", "anyPopulation"));
        content.setLastModified(ZonedDateTime.now());
        content.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        
        return content;
    }
}
