/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.io.IOException;
import java.io.OutputStream;
import java.io.StringWriter;
import java.time.LocalDate;
import java.time.Month;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Calendar;
import java.util.List;
import java.util.Locale;
import java.util.stream.Stream;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.io.IOUtils;
import org.apache.commons.io.output.ByteArrayOutputStream;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Binary;
import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.ExplorerFile;
import org.ametys.cms.data.File;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.NamedResource;
import org.ametys.cms.data.Reference;
import org.ametys.cms.data.Resource;
import org.ametys.cms.data.RichText;
import org.ametys.cms.data.holder.impl.DefaultModelAwareDataHolder;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.explorer.resources.ModifiableResource;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.Composite;
import org.ametys.plugins.repository.data.holder.group.ModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeaterEntry;
import org.ametys.plugins.repository.data.holder.group.Repeater;
import org.ametys.plugins.repository.data.holder.group.RepeaterEntry;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.ComplexRepositoryElementType;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.plugins.repository.data.type.RepositoryModelItemType;
import org.ametys.plugins.repository.metadata.MultilingualString;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelAwareDataHolder#hasValue} and {@link ModelAwareDataHolder#getValue} methods implementation test case
 */
@NeedFeatures({
    "test-data/test-model",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModelAwareDataHolderHasAndGetValueTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    
    private static Model _model;
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    
    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;

    private static RepositoryElementType<UserIdentity> _userType;
    private static RepositoryElementType<Geocode> _geocodeType;
    private static RepositoryElementType<MultilingualString> _multilingualStringType;
    private static RepositoryElementType<Reference> _referenceType;
    private static RepositoryElementType<Binary> _binaryType;
    private static RepositoryElementType<File> _fileType;
    private static RepositoryElementType<RichText> _richTextType;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
    }
    
    @Override
    @SuppressWarnings("unchecked")
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");

        _userType = (RepositoryElementType<UserIdentity>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _geocodeType = (RepositoryElementType<Geocode>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID);
        _multilingualStringType = (RepositoryElementType<MultilingualString>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID);
        _referenceType = (RepositoryElementType<Reference>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.REFERENCE_ELEMENT_TYPE_ID);
        _binaryType = (RepositoryElementType<Binary>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _fileType = (RepositoryElementType<File>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.FILE_ELEMENT_TYPE_ID);
        _richTextType = (RepositoryElementType<RichText>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for simple data types
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSimpleDataTypesWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSimpleDataTypes", DATA_NODE_TYPE);
        _testSimpleDataTypes(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for simple data types
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSimpleDataTypesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSimpleDataTypes", DATA_NODE_TYPE);
        _testSimpleDataTypes(repositoryData);
    }
    
    private void _testSimpleDataTypes(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Read value before writing it (should be null)
        Assert.assertNull(dataHolder.getValue("singleStringData"));
        Assert.assertNull(dataHolder.getValue("multipleDoubleData"));

        // Write
        repositoryData.setValue("singleStringData", "my string");
        String[] stringArray = new String[]{"string1", "string2", "string3"};
        repositoryData.setValues("multipleStringData", stringArray);
        repositoryData.setValue("singleLongData", 67L);
        Long[] longArray = new Long[]{65L, 98L, 432L};
        repositoryData.setValues("multipleLongData", longArray);
        repositoryData.setValue("singleDoubleData", 54D);
        Double[] doubleArray = new Double[]{87D, 345D};
        repositoryData.setValues("multipleDoubleData", doubleArray);
        repositoryData.setValue("singleBooleanData", true);
        Boolean[] booleanArray = new Boolean[]{true, true, false, true, false};
        repositoryData.setValues("multipleBooleanData", booleanArray);
        
        // Read
        Assert.assertEquals("my string", dataHolder.getValue("singleStringData"));
        Assert.assertArrayEquals(stringArray, dataHolder.getValue("multipleStringData"));
        Assert.assertEquals(Long.valueOf(67), dataHolder.getValue("singleLongData"));
        Assert.assertArrayEquals(longArray, dataHolder.getValue("multipleLongData"));
        Assert.assertEquals(Double.valueOf(54), dataHolder.getValue("singleDoubleData"));
        Assert.assertArrayEquals(doubleArray, dataHolder.getValue("multipleDoubleData"));
        Assert.assertTrue((boolean) dataHolder.getValue("singleBooleanData"));
        Assert.assertArrayEquals(booleanArray, dataHolder.getValue("multipleBooleanData"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for date and date time types
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testDateDataTypesWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testDateDataTypes", DATA_NODE_TYPE);
        _testDateDataTypes(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for date and date time types
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testDateDataTypesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testDateDataTypes", DATA_NODE_TYPE);
        _testDateDataTypes(repositoryData);
    }
    
    private void _testDateDataTypes(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Date - write
        repositoryData.setValue("singleDateData", DateUtils.asCalendar(LocalDate.of(2018, Month.MAY, 18)));
        LocalDate[] dateArray = new LocalDate[]{LocalDate.of(2018, Month.MAY, 5), LocalDate.of(2015, Month.DECEMBER, 06)};
        repositoryData.setValues("multipleDateData", Stream.of(dateArray).map(DateUtils::asCalendar).toArray(Calendar[]::new));
        
        // Date - read
        LocalDate localDate = dataHolder.getValue("singleDateData");
        Assert.assertEquals(18, localDate.getDayOfMonth());
        Assert.assertEquals(Month.MAY, localDate.getMonth());
        Assert.assertEquals(2018, localDate.getYear());
        Assert.assertArrayEquals(dateArray, dataHolder.getValue("multipleDateData"));
        
        // DateTime - write
        repositoryData.setValue("singleDatetimeData", DateUtils.asCalendar(ZonedDateTime.of(2018, 5, 18, 14, 15, 31, 0, ZoneId.systemDefault())));
        ZonedDateTime[] dateTimeArray = new ZonedDateTime[]{ZonedDateTime.of(2018, 5, 5, 7, 25, 13, 0, ZoneId.of("UTC+05:00")), ZonedDateTime.of(2015, 12, 06, 15, 18, 0, 0, ZoneId.of("UTC-03:00"))};
        repositoryData.setValues("multipleDatetimeData", Stream.of(dateTimeArray).map(DateUtils::asCalendar).toArray(Calendar[]::new));
        
        // DateTime - read
        ZonedDateTime zonedDateTime = dataHolder.getValue("singleDatetimeData");
        zonedDateTime = zonedDateTime.withZoneSameInstant(ZoneId.systemDefault());
        Assert.assertEquals(2018, zonedDateTime.getYear());
        Assert.assertEquals(Month.MAY, zonedDateTime.getMonth());
        Assert.assertEquals(18, zonedDateTime.getDayOfMonth());
        Assert.assertEquals(14, zonedDateTime.getHour());
        Assert.assertEquals(15, zonedDateTime.getMinute());
        Assert.assertEquals(31, zonedDateTime.getSecond());
        ZonedDateTime[] zonedDateTimes = dataHolder.getValue("multipleDatetimeData");
        Assert.assertEquals(2, zonedDateTimes.length);
        ZonedDateTime zonedDateTime1 = zonedDateTimes[0];
        zonedDateTime1 = zonedDateTime1.withZoneSameInstant(ZoneId.of("UTC+05:00"));
        Assert.assertEquals(2018, zonedDateTime1.getYear());
        Assert.assertEquals(Month.MAY, zonedDateTime1.getMonth());
        Assert.assertEquals(05, zonedDateTime1.getDayOfMonth());
        Assert.assertEquals(07, zonedDateTime1.getHour());
        Assert.assertEquals(25, zonedDateTime1.getMinute());
        Assert.assertEquals(13, zonedDateTime1.getSecond());
        ZonedDateTime zonedDateTime2 = zonedDateTimes[1];
        zonedDateTime2 = zonedDateTime2.withZoneSameInstant(ZoneId.of("UTC-03:00"));
        Assert.assertEquals(2015, zonedDateTime2.getYear());
        Assert.assertEquals(Month.DECEMBER, zonedDateTime2.getMonth());
        Assert.assertEquals(06, zonedDateTime2.getDayOfMonth());
        Assert.assertEquals(15, zonedDateTime2.getHour());
        Assert.assertEquals(18, zonedDateTime2.getMinute());
        Assert.assertEquals(00, zonedDateTime2.getSecond());
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for complex data types
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testComplexDataTypesWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testComplexDataTypes", DATA_NODE_TYPE);
        _testComplexDataTypes(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for complex data types
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testComplexDataTypesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testComplexDataTypes", DATA_NODE_TYPE);
        _testComplexDataTypes(repositoryData);
    }
    
    private void _testComplexDataTypes(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // User - write
        _userType.write(repositoryData, "singleUserData", new UserIdentity("tania", "anyPopulation"));
        UserIdentity[] userArray = new UserIdentity[]{new UserIdentity("kevin", "anyPopulation"), new UserIdentity("georges", "anyPopulation"), new UserIdentity("john", "anyPopulation")};
        _userType.write(repositoryData, "multipleUserData", userArray);
        
        // User - read
        UserIdentity user = dataHolder.getValue("singleUserData");
        Assert.assertEquals("tania", user.getLogin());
        Assert.assertEquals("anyPopulation", user.getPopulationId());
        Assert.assertArrayEquals(userArray, dataHolder.getValue("multipleUserData"));
        
        // Geocode - write
        _geocodeType.write(repositoryData, "singleGeocodeData", new Geocode(43.5470148D, 1.5134120999999823D));
        Geocode[] geocodeArray = new Geocode[]{new Geocode(0.0D, 0.0D), new Geocode(23.5, 0.78)};
        _geocodeType.write(repositoryData, "multipleGeocodeData", geocodeArray);
        
        // Geocode - read
        Geocode geocode = dataHolder.getValue("singleGeocodeData");
        Assert.assertEquals(Double.valueOf(1.5134120999999823), geocode.getLongitude());
        Assert.assertEquals(Double.valueOf(43.5470148D), geocode.getLatitude());
        geocodeArray = dataHolder.getValue("multipleGeocodeData");
        Assert.assertEquals(2, geocodeArray.length);
        Assert.assertEquals(Double.valueOf(0.0), geocodeArray[0].getLongitude());
        Assert.assertEquals(Double.valueOf(0.0), geocodeArray[0].getLatitude());
        Assert.assertEquals(Double.valueOf(0.78), geocodeArray[1].getLongitude());
        Assert.assertEquals(Double.valueOf(23.5), geocodeArray[1].getLatitude());
        
        // Multilingual String - write
        MultilingualString multilingualString = new MultilingualString();
        multilingualString.add(Locale.FRANCE, "ma valeur");
        multilingualString.add(Locale.UK, "my value");
        _multilingualStringType.write(repositoryData, "singleMultilingualStringData", multilingualString);
        
        MultilingualString multilingualString1 = new MultilingualString();
        multilingualString1.add(Locale.CANADA, "tabernacle");
        multilingualString1.add(Locale.GERMANY, "der Kaffee ist fertig");
        MultilingualString multilingualString2 = new MultilingualString();
        multilingualString2.add(Locale.ITALY, "Bella ciao");
        MultilingualString[] multilingualStringArray = new MultilingualString[]{multilingualString1, multilingualString2};
        _multilingualStringType.write(repositoryData, "multipleMultilingualStringData", multilingualStringArray);
        
        // Multilingual string - read
        MultilingualString retrievedMultilingualString = dataHolder.getValue("singleMultilingualStringData");
        Assert.assertEquals(2, retrievedMultilingualString.getValues().size());
        Assert.assertEquals("ma valeur", retrievedMultilingualString.getValue(Locale.FRANCE));
        Assert.assertEquals("my value", retrievedMultilingualString.getValue(Locale.UK));
        
        MultilingualString[] retrievedMultilingualStringArray = dataHolder.getValue("multipleMultilingualStringData");
        Assert.assertEquals(2, retrievedMultilingualStringArray.length);
        
        MultilingualString retrievedMultilingualString1 = retrievedMultilingualStringArray[0];
        Assert.assertEquals(2, retrievedMultilingualString1.getValues().size());
        Assert.assertEquals("tabernacle", retrievedMultilingualString1.getValue(Locale.CANADA));
        Assert.assertEquals("der Kaffee ist fertig", retrievedMultilingualString1.getValue(Locale.GERMANY));

        MultilingualString retrievedMultilingualString2 = retrievedMultilingualStringArray[1];
        Assert.assertEquals(1, retrievedMultilingualString2.getValues().size());
        Assert.assertEquals("Bella ciao", retrievedMultilingualString2.getValue(Locale.ITALY));
        
        // Reference - write
        _referenceType.write(repositoryData, "singleReferenceData", new Reference("jcr/path/to/the/metadata", "metadata"));
        repositoryData.getDataNames();
        Reference[] referenceArray = new Reference[]{new Reference("jcr/path/to/the/metadata", "metadata"), new Reference("path/to/the/attachment", "attachment")};
        _referenceType.write(repositoryData, "multipleReferenceData", referenceArray);
        
        // Reference - read
        Reference reference = dataHolder.getValue("singleReferenceData");
        Assert.assertEquals("jcr/path/to/the/metadata", reference.getValue());
        Assert.assertEquals("metadata", reference.getType());
        referenceArray = dataHolder.getValue("multipleReferenceData");
        Assert.assertEquals(2, referenceArray.length);
        Assert.assertEquals("jcr/path/to/the/metadata", referenceArray[0].getValue());
        Assert.assertEquals("metadata", referenceArray[0].getType());
        Assert.assertEquals("path/to/the/attachment", referenceArray[1].getValue());
        Assert.assertEquals("attachment", referenceArray[1].getType());
        
        // Reference without type - write
        _referenceType.write(repositoryData, "singleReferenceData", new Reference("https://docs.ametys.org/fr/index.html"));
        
        // Reference without type - read
        reference = dataHolder.getValue("singleReferenceData");
        Assert.assertEquals("https://docs.ametys.org/fr/index.html", reference.getValue());
        Assert.assertEquals(Reference.DEFAULT_TYPE, reference.getType());
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for resources data types
     * This test uses a {@link JCRRepositoryData}
     * @throws IOException if an error occurs while writing or reading resources
     */
    @Test
    @NeedFeatures({"explorer/factories"})
    public void testResourceDataTypesWithJCRRepositoryData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testResourceDataTypes", DATA_NODE_TYPE);
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Binary - write
        Binary binary = new Binary();
        try (OutputStream os = binary.getOutputStream())
        {
            os.write("my binary content".getBytes());
        }
        binary.setEncoding("UTF-8");
        binary.setFilename("myBinary.txt");
        binary.setLastModificationDate(ZonedDateTime.of(2019, 03, 12, 17, 22, 38, 0, ZoneId.systemDefault()));
        binary.setMimeType("text/plain");
        _binaryType.write(repositoryData, "singleBinaryData", binary);
        
        // Binary - read
        Binary retrievedBinary = dataHolder.getValue("singleBinaryData");
        Assert.assertEquals("UTF-8", retrievedBinary.getEncoding());
        Assert.assertEquals("myBinary.txt", retrievedBinary.getFilename());
        _assertZonedDateTimeEquals(ZoneId.systemDefault(), 2019, Month.MARCH, 12, 17, 22, 38, retrievedBinary.getLastModificationDate());
        Assert.assertEquals("text/plain", retrievedBinary.getMimeType());
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedBinary.getInputStream(), writer, retrievedBinary.getEncoding());
            Assert.assertEquals("my binary content", writer.toString());
        }
        
        // File as Binary - write
        _fileType.write(repositoryData, "singleFileData", binary);
        
        // File as Binary - read
        File retrievedFileAsBinary = dataHolder.getValue("singleFileData");
        Assert.assertEquals("myBinary.txt", retrievedFileAsBinary.getName());
        _assertZonedDateTimeEquals(ZoneId.systemDefault(), 2019, Month.MARCH, 12, 17, 22, 38, retrievedFileAsBinary.getLastModificationDate());
        Assert.assertEquals("text/plain", retrievedFileAsBinary.getMimeType());
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFileAsBinary.getInputStream(), writer, "UTF-8");
            Assert.assertEquals("my binary content", writer.toString());
        }
        
        // File as explorer resource - write
        ExplorerFile file = _createExplorerFile();
        _fileType.write(repositoryData, "singleFileData", file);
        
        // File as explorer resource - read
        File retrievedFileAsExplorerResource = dataHolder.getValue("singleFileData");
        Assert.assertEquals("myResource.txt", retrievedFileAsExplorerResource.getName());
        _assertZonedDateTimeEquals(ZoneId.systemDefault(), 2019, Month.APRIL, 12, 10, 48, 0, retrievedFileAsExplorerResource.getLastModificationDate());
        Assert.assertEquals("text/plain", retrievedFileAsExplorerResource.getMimeType());
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFileAsExplorerResource.getInputStream(), writer, "UTF-8");
            Assert.assertEquals("my explorer file content", writer.toString());
        }
        
        // Rich text - write
        RichText richText = new RichText();
        try (OutputStream os = richText.getOutputStream())
        {
            os.write("my rich text content".getBytes());
        }
        richText.setEncoding("UTF-8");
        richText.setMimeType("text/xml");
        richText.setLastModificationDate(ZonedDateTime.of(2019, 03, 20, 13, 34, 12, 0, ZoneId.systemDefault()));
        // Annotations
        richText.addAnnotations("annotations1", "a1-1", "a1-2", "a1-3");
        richText.addAnnotations("annotations2", "a2-1", "a2-2");
        // Files
        _addFilesToRichText(richText);
        _richTextType.write(repositoryData, "singleRichTextData", richText);
        
        // Rich text - read
        RichText retrievedRichText = dataHolder.getValue("singleRichTextData");
        Assert.assertEquals("UTF-8", retrievedRichText.getEncoding());
        _assertZonedDateTimeEquals(ZoneId.systemDefault(), 2019, Month.MARCH, 20, 13, 34, 12, retrievedRichText.getLastModificationDate());
        Assert.assertEquals("text/xml", retrievedRichText.getMimeType());
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedRichText.getInputStream(), writer, retrievedRichText.getEncoding());
            Assert.assertEquals("my rich text content", writer.toString());
        }
        // Annotations
        Assert.assertEquals(2, retrievedRichText.getAllAnnotations().size());
        List<String> annotations1 = retrievedRichText.getAnnotations("annotations1");
        Assert.assertEquals(3, annotations1.size());
        Assert.assertTrue(annotations1.contains("a1-1"));
        Assert.assertTrue(annotations1.contains("a1-2"));
        Assert.assertTrue(annotations1.contains("a1-3"));
        List<String> annotations2 = retrievedRichText.getAnnotations("annotations2");
        Assert.assertEquals(2, annotations2.size());
        Assert.assertTrue(annotations2.contains("a2-1"));
        Assert.assertTrue(annotations2.contains("a2-2"));
        // Files
        Assert.assertEquals(2,  retrievedRichText.getAttachments().size());
        Resource retrievedFile1 = retrievedRichText.getAttachment("myFirstFile.txt");
        Assert.assertEquals("UTF-8", retrievedFile1.getEncoding());
        _assertZonedDateTimeEquals(ZoneId.systemDefault(), 2019, Month.MARCH, 20, 14, 03, 59, retrievedFile1.getLastModificationDate());
        Assert.assertEquals("text/plain", retrievedFile1.getMimeType());
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFile1.getInputStream(), writer, retrievedFile1.getEncoding());
            Assert.assertEquals("my first file content", writer.toString());
        }
        Resource retrievedFile2 = retrievedRichText.getAttachment("mySecondFile.txt");
        Assert.assertEquals("UTF-8", retrievedFile2.getEncoding());
        _assertZonedDateTimeEquals(ZoneId.systemDefault(), 2019, Month.MARCH, 20, 14, 04, 01, retrievedFile2.getLastModificationDate());
        Assert.assertEquals("text/plain", retrievedFile2.getMimeType());
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFile2.getInputStream(), writer, retrievedFile2.getEncoding());
            Assert.assertEquals("my second file content", writer.toString());
        }
        
        // RichText - add and delete files
        _playWithRichTextFiles(richText);
        _richTextType.write(repositoryData, "singleRichTextData", richText);
        
        retrievedRichText = dataHolder.getValue("singleRichTextData");
        Assert.assertEquals(2,  retrievedRichText.getAttachments().size());
        retrievedFile2 = retrievedRichText.getAttachment("mySecondFile.txt");
        Assert.assertEquals("UTF-8", retrievedFile2.getEncoding());
        _assertZonedDateTimeEquals(ZoneId.systemDefault(), 2019, Month.MARCH, 29, 10, 18, 12, retrievedFile2.getLastModificationDate());
        Assert.assertEquals("text/plain", retrievedFile2.getMimeType());
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFile2.getInputStream(), writer, retrievedFile2.getEncoding());
            Assert.assertEquals("my second file modified content", writer.toString());
        }
        Resource retrievedFile3 = retrievedRichText.getAttachment("myThirdFile.txt");
        Assert.assertEquals("UTF-8", retrievedFile3.getEncoding());
        _assertZonedDateTimeEquals(ZoneId.systemDefault(), 2019, Month.MARCH, 29, 10, 21, 01, retrievedFile3.getLastModificationDate());
        Assert.assertEquals("text/plain", retrievedFile3.getMimeType());
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFile3.getInputStream(), writer, retrievedFile3.getEncoding());
            Assert.assertEquals("my third file content", writer.toString());
        }
    }

    private void _assertZonedDateTimeEquals(ZoneId zoneId, int year, Month month, int day, int hour, int minute, int second, ZonedDateTime actualZDT)
    {
        ZonedDateTime zdt = actualZDT.withZoneSameInstant(zoneId);
        Assert.assertEquals(zoneId, zdt.getZone());
        Assert.assertEquals(year, zdt.getYear());
        Assert.assertEquals(month, zdt.getMonth());
        Assert.assertEquals(day, zdt.getDayOfMonth());
        Assert.assertEquals(hour, zdt.getHour());
        Assert.assertEquals(minute, zdt.getMinute());
        Assert.assertEquals(second, zdt.getSecond());
    }
    
    private ExplorerFile _createExplorerFile() throws IOException
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        ModifiableResource resource = root.createChild("myResource.txt", "ametys:resource");

        try (ByteArrayOutputStream os = new ByteArrayOutputStream())
        {
            os.write("my explorer file content".getBytes());
            ZonedDateTime lastModificationDate = ZonedDateTime.of(2019, 4, 12, 10, 48, 00, 00, ZoneId.systemDefault());
            resource.setData(os.toInputStream(), "text/plain", DateUtils.asDate(lastModificationDate), new UserIdentity("john", "anyPopulation"));
        }
        
        return new ExplorerFile(_ametysObjectResolver, resource.getId());
    }

    private void _addFilesToRichText(RichText richText) throws IOException
    {
        NamedResource file1 = new NamedResource();
        file1.setFilename("myFirstFile.txt");
        try (OutputStream os = file1.getOutputStream())
        {
            os.write("my first file content".getBytes());
        }
        file1.setEncoding("UTF-8");
        file1.setLastModificationDate(ZonedDateTime.of(2019, 03, 20, 14, 03, 59, 0, ZoneId.systemDefault()));
        file1.setMimeType("text/plain");
        richText.addAttachment(file1);
        
        NamedResource file2 = new NamedResource();
        file2.setFilename("mySecondFile.txt");
        try (OutputStream os = file2.getOutputStream())
        {
            os.write("my second file content".getBytes());
        }
        file2.setEncoding("UTF-8");
        file2.setLastModificationDate(ZonedDateTime.of(2019, 03, 20, 14, 04, 01, 0, ZoneId.systemDefault()));
        file2.setMimeType("text/plain");
        richText.addAttachment(file2);
    }
    
    private void _playWithRichTextFiles(RichText richText) throws IOException
    {
        // Remove first file
        richText.removeAttachment("myFirstFile.txt");
        
        // Add the already existing second file and put a different content
        NamedResource file2 = richText.getAttachment("mySecondFile.txt");
        try (OutputStream os = file2.getOutputStream())
        {
            os.write("my second file modified content".getBytes());
        }
        file2.setLastModificationDate(ZonedDateTime.of(2019, 03, 29, 10, 18, 12, 0, ZoneId.systemDefault()));
        
        // Add a third file
        NamedResource file3 = new NamedResource();
        file3.setFilename("myThirdFile.txt");
        try (OutputStream os = file3.getOutputStream())
        {
            os.write("my third file content".getBytes());
        }
        file3.setEncoding("UTF-8");
        file3.setLastModificationDate(ZonedDateTime.of(2019, 03, 29, 10, 21, 01, 0, ZoneId.systemDefault()));
        file3.setMimeType("text/plain");
        richText.addAttachment(file3);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for content types
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    @NeedFeatures({
        "cms/factories",
        "test/content-type",
    })
    public void testContentDataTypeWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testContentDataType", DATA_NODE_TYPE);
        _testContentDataType(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for content types
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    @NeedFeatures({
        "cms/factories",
        "test/content-type",
    })
    public void testContentDataTypeWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testContentDataType", DATA_NODE_TYPE);
        _testContentDataType(repositoryData);
    }
    
    private void _testContentDataType(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent content1 = _createContent("content1");
        ModifiableContent content2 = _createContent("content2");
        
        // Content - write
        repositoryData.setValue("singleContentData", content1.getId());
        repositoryData.setValues("multipleContentData", new String[] {content1.getId(), content2.getId()});
        
        // Read
        ContentValue singleContent = dataHolder.getValue("singleContentData");
        Assert.assertEquals(content1.getId(), singleContent.getContentId());
        Assert.assertEquals("content1", singleContent.getContent().getTitle());
        ContentValue[] multipleContent = dataHolder.getValue("multipleContentData");
        Assert.assertEquals(2, multipleContent.length);
        Assert.assertEquals(content1.getId(), multipleContent[0].getContentId());
        Assert.assertEquals("content1", multipleContent[0].getContent().getTitle());
        Assert.assertEquals(content2.getId(), multipleContent[1].getContentId());
        Assert.assertEquals("content2", multipleContent[1].getContent().getTitle());
    }
    
    private ModifiableContent _createContent(String title)
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        
        ModifiableContent content = root.createChild(title, "ametys:defaultContent");

        content.setTypes(new String[]{"org.ametys.cms.contenttype.Test"});
        content.setTitle(title, Locale.FRANCE);
        content.setLanguage("fr");
        content.setCreationDate(ZonedDateTime.now());
        content.setCreator(new UserIdentity("john", "anyPopulation"));
        content.setLastModified(ZonedDateTime.now());
        content.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        
        return content;
    }
    

    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for composite and repeaters
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testCompositeAndRepeatersWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testCompositeAndRepeaters", DATA_NODE_TYPE);
        _testCompositeAndRepeaters(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for composite and repeaters
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testCompositeAndRepeatersWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testCompositeAndRepeaters", DATA_NODE_TYPE);
        _testCompositeAndRepeaters(repositoryData);
    }
    
    private void _testCompositeAndRepeaters(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        _addSomeValuesInRepositoryData(repositoryData);
        
        // composite
        ModelAwareComposite composite = dataHolder.getValue("composite");
        Assert.assertNotNull(composite);
        Assert.assertNotNull(dataHolder.getComposite("composite"));
        
        // composite - singleLongData
        Long longMin = dataHolder.getValue("composite/singleLongData");
        Assert.assertEquals(Long.valueOf(156), longMin);
        Assert.assertEquals(longMin, composite.getValue("singleLongData"));
        
        // composite - repeater
        ModelAwareRepeater repeater = dataHolder.getValue("composite/repeater");
        Assert.assertNotNull(repeater);
        Assert.assertNotNull(dataHolder.getRepeater("composite/repeater"));
        Assert.assertNotNull(composite.getValue("repeater"));
        Assert.assertNotNull(composite.getRepeater("repeater"));
        
        // composite - repeater entries
        ModelAwareRepeaterEntry repeaterEntry1 = dataHolder.getValue("composite/repeater[1]");
        Assert.assertNotNull(repeaterEntry1);
        Assert.assertNotNull(composite.getValue("repeater[1]"));
        Assert.assertNotNull(repeater.getEntry(1));
        ModelAwareRepeaterEntry repeaterEntry2 = repeater.getEntry(2);
        Assert.assertNotNull(repeaterEntry2);
        Assert.assertNull(dataHolder.getValue("composite/repeater[3]"));
        Assert.assertNull(composite.getValue("repeater[3]"));
        Assert.assertNull(repeater.getEntry(3));
        
        // composite - repeater - entry 1 - single boolean data (not set)
        Assert.assertNull(dataHolder.getValue("composite/repeater[1]/singleBooleanData"));
        Assert.assertNull(composite.getValue("repeater[1]/singleBooleanData"));
        Assert.assertNull(repeaterEntry1.getValue("singleBooleanData"));
        
        // composite - repeater - entry 1 - single double data
        Double entry1DoubleData = dataHolder.getValue("composite/repeater[1]/singleDoubleData");
        Assert.assertEquals(Double.valueOf(12.3), entry1DoubleData);
        Assert.assertEquals(entry1DoubleData, composite.getValue("repeater[1]/singleDoubleData"));
        Assert.assertEquals(entry1DoubleData, repeaterEntry1.getValue("singleDoubleData"));
        
        // composite - repeater - entry 1 - subcomposite
        ModelAwareComposite repeaterEntry1Subcomposite = dataHolder.getValue("composite/repeater[1]/subcomposite");
        Assert.assertNotNull(repeaterEntry1Subcomposite);
        Assert.assertNotNull(dataHolder.getComposite("composite/repeater[1]/subcomposite"));
        Assert.assertNotNull(composite.getValue("repeater[1]/subcomposite"));
        Assert.assertNotNull(composite.getComposite("repeater[1]/subcomposite"));
        Assert.assertNotNull(repeaterEntry1.getValue("subcomposite"));
        Assert.assertNotNull(repeaterEntry1.getComposite("subcomposite"));
        
        // composite - repeater - entry 1 - subcomposite - single string data
        String entry1SubcompStringData = dataHolder.getValue("composite/repeater[1]/subcomposite/singleStringData");
        Assert.assertEquals("repeater entry 1 - sub composite - string value", entry1SubcompStringData);
        Assert.assertEquals(entry1SubcompStringData, composite.getValue("repeater[1]/subcomposite/singleStringData"));
        Assert.assertEquals(entry1SubcompStringData, repeaterEntry1.getValue("subcomposite/singleStringData"));
        Assert.assertEquals(entry1SubcompStringData, repeaterEntry1Subcomposite.getValue("singleStringData"));
        
        // composite - repeater - entry 1 - subcomposite - multiple date data (not set)
        Assert.assertNull(dataHolder.getValue("composite/repeater[1]/subcomposite/multipleDateData"));
        Assert.assertNull(composite.getValue("repeater[1]/subcomposite/multipleDateData"));
        Assert.assertNull(repeaterEntry1.getValue("subcomposite/multipleDateData"));
        Assert.assertNull(repeaterEntry1Subcomposite.getValue("multipleDateData"));
        
        // composite - repeater - entry 2 - single double data
        Double entry2DoubleData = dataHolder.getValue("composite/repeater[2]/singleDoubleData");
        Assert.assertEquals(Double.valueOf(6.0), entry2DoubleData);
        Assert.assertEquals(entry2DoubleData, composite.getValue("repeater[2]/singleDoubleData"));
        Assert.assertEquals(entry2DoubleData, repeaterEntry2.getValue("singleDoubleData"));
        
        // composite - repeater - entry 2 - subcomposite
        ModelAwareComposite repeaterEntry2Subcomposite = dataHolder.getComposite("composite/repeater[2]/subcomposite");
        Assert.assertNotNull(repeaterEntry2Subcomposite);
        
        // composite - repeater - entry 2 -  single string data
        String entry2SubcompStringData = dataHolder.getValue("composite/repeater[2]/subcomposite/singleStringData");
        Assert.assertEquals("repeater entry 2 - sub composite - string value", entry2SubcompStringData);
        Assert.assertEquals(entry2SubcompStringData, composite.getValue("repeater[2]/subcomposite/singleStringData"));
        Assert.assertEquals(entry2SubcompStringData, repeaterEntry2.getValue("subcomposite/singleStringData"));
        Assert.assertEquals(entry2SubcompStringData, repeaterEntry2Subcomposite.getValue("singleStringData"));
        
        try
        {
            // Tries to get a value from an undefined data
            dataHolder.getValue("composite/repeater[2]/subcomposite/undefinedData");
            Assert.fail("There is no data named 'undefinedData' in the model. The getValue method should have fail.");
        }
        catch (UndefinedItemPathException e)
        {
            Assert.assertTrue(e.getMessage().contains("defined"));
        }
        
        try
        {
            // Tries to get a value from an undefined parent data
            dataHolder.getValue("composite/repeater[2]/undefinedGroup/singleStringData");
            Assert.fail("There is no composite named 'undefinedGroup' in the model. The getValue method should have fail.");
        }
        catch (UndefinedItemPathException e)
        {
            Assert.assertTrue(e.getMessage().contains("defined"));
        }
    }
    
    private void _addSomeValuesInRepositoryData(ModifiableRepositoryData repositoryData)
    {
        repositoryData.setValue("singleStringData", "my string");
        repositoryData.setValues("multipleStringData", new String[]{"string1", "string2", "string3"});
        
        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        composite.setValue("singleLongData", 156L);
        
        ModifiableRepositoryData repeater = composite.addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE);
        ModifiableRepositoryData entry1 = repeater.addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE);
        entry1.setValue("singleDoubleData", 12.3D);
        ModifiableRepositoryData subComposite1 = entry1.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subComposite1.setValue("singleStringData", "repeater entry 1 - sub composite - string value");
        
        ModifiableRepositoryData entry2 = repeater.addRepositoryData("2", RepositoryConstants.COMPOSITE_NODETYPE);
        entry2.setValue("singleDoubleData", 6.0D);
        ModifiableRepositoryData subComposite2 = entry2.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subComposite2.setValue("singleStringData", "repeater entry 2 - sub composite - string value");
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for composite and repeaters
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testUnvaluedCompositeAndRepeatersWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testUnvaluedCompositeAndRepeaters", DATA_NODE_TYPE);
        _testUnvaluedCompositeAndRepeaters(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for composite and repeaters
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testUnvaluedCompositeAndRepeatersWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testUnvaluedCompositeAndRepeaters", DATA_NODE_TYPE);
        _testUnvaluedCompositeAndRepeaters(repositoryData);
    }
    
    private void _testUnvaluedCompositeAndRepeaters(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        Assert.assertNull(dataHolder.getComposite("unvaluedComposite"));
        Assert.assertNull(dataHolder.getRepeater("unvaluedRepeater"));
    }

    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for with a data that is stored with the wrong type
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testWithWrongDataTypeWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testWithWrongDataType", DATA_NODE_TYPE);
        _testWithWrongDataType(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for with a data that is stored with the wrong type
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testWithWrongDataTypeWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testWithWrongDataType", DATA_NODE_TYPE);
        _testWithWrongDataType(repositoryData);
    }
    
    private void _testWithWrongDataType(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Set a String value to the long data
        repositoryData.setValue("singleLongData", "string value");
        
        try
        {
            // Try to get the value on the data
            dataHolder.getValue("singleLongData");
            Assert.fail("The stored value of the data is a string and can't be retrieved by the long element type");
        }
        catch (BadItemTypeException e)
        {
            Assert.assertTrue(e.getMessage().contains("long"));
        }
    }
    /**
     * Tests the {@link Content#getValue(String, boolean, Object)} method
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testWithDefaultValuesWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testWithDefaultValues", DATA_NODE_TYPE);
        _testWithDefaultValues(repositoryData);
    }
    
    /**
     * Tests the {@link Content#getValue(String, boolean, Object)} method
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testWithDefaultValuesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testWithDefaultValues", DATA_NODE_TYPE);
        _testWithDefaultValues(repositoryData);
    }
    
    private void _testWithDefaultValues(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        // Test the getValue method when there is no value
        Assert.assertEquals("the default value from model", dataHolder.getValue("stringWithDefaultValue", true, "the default value parameter"));
        Assert.assertEquals("the default value parameter", dataHolder.getValue("stringWithDefaultValue", false, "the default value parameter"));
        Assert.assertEquals("the default value parameter", dataHolder.getValue("singleStringData", true, "the default value parameter"));
        Assert.assertEquals("the default value parameter", dataHolder.getValue("singleStringData", false, "the default value parameter"));

        Assert.assertNull(dataHolder.getValue("composite", true, null));
        Assert.assertEquals(Long.valueOf(459), dataHolder.getValue("composite/repeater[1]/longWithDefaultValue", true, 125L));
        Assert.assertEquals(Long.valueOf(125), dataHolder.getValue("composite/repeater[1]/longWithDefaultValue", false, 125L));
        Assert.assertEquals(true, dataHolder.getValue("composite/repeater[1]/singleBooleanData", true, true));
        Assert.assertEquals(true, dataHolder.getValue("composite/repeater[1]/singleBooleanData", false, true));
        
        // Set empty values
        repositoryData.setValue("stringWithDefaultValue", StringUtils.EMPTY);
        repositoryData.setValue("singleStringData", StringUtils.EMPTY);
        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        ModifiableRepositoryData repeater = composite.addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE);
        ModifiableRepositoryData entry1 = repeater.addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE);
        entry1.setValue("longWithDefaultValue" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        
        // Test the getValue method
        Assert.assertEquals("the default value from model", dataHolder.getValue("stringWithDefaultValue", true, "the default value parameter"));
        Assert.assertEquals("the default value parameter", dataHolder.getValue("singleStringData", true, "the default value parameter"));
        Assert.assertEquals(Long.valueOf(459), dataHolder.getValue("composite/repeater[1]/longWithDefaultValue", true, 125L));
        
        // Set values
        repositoryData.setValue("stringWithDefaultValue", "a value");
        repositoryData.setValue("singleStringData", "another value");
        entry1.setValue("longWithDefaultValue", 78L);
        entry1.setValue("singleBooleanData", false);
        
        // Test the getValue method
        Assert.assertEquals("a value", dataHolder.getValue("stringWithDefaultValue", true, "the default value parameter"));
        Assert.assertEquals("another value", dataHolder.getValue("singleStringData", true, "the default value parameter"));
        Assert.assertEquals(Long.valueOf(78), dataHolder.getValue("composite/repeater[1]/longWithDefaultValue", true, 125L));
        Assert.assertEquals(false, dataHolder.getValue("composite/repeater[1]/singleBooleanData", true, true));
        
        try
        {
            // Tries to get a value from an undefined data - use the default value from model
            dataHolder.getValue("composite/repeater[2]/subcomposite/undefined-item", true, "default");
            Assert.fail("There is no item named 'undefined-item' in the model. The getValue method should have fail.");
        }
        catch (UndefinedItemPathException e)
        {
            Assert.assertTrue(e.getMessage().contains("defined"));
        }
        
        try
        {
            // Tries to get a value from an undefined data - do not use the default value from model
            dataHolder.getValue("composite/repeater[2]/subcomposite/undefined-item", false, "default");
            Assert.fail("There is no item named 'undefined-item' in the model. The getValue method should have fail.");
        }
        catch (UndefinedItemPathException e)
        {
            Assert.assertTrue(e.getMessage().contains("defined"));
        }
        
        try
        {
            // Tries to get a value from an undefined composite data - use the default value from model
            dataHolder.getValue("composite/repeater[2]/undefined-subcomposite/singleStringData", true, "default");
            Assert.fail("There is no composite named 'undefined-subcomposite' in the model. The getValue method should have fail.");
        }
        catch (UndefinedItemPathException e)
        {
            Assert.assertTrue(e.getMessage().contains("defined"));
        }
        
        try
        {
            // Tries to get a value from an undefined composite data - do not use the default value from model
            dataHolder.getValue("composite/repeater[2]/undefined-subcomposite/singleStringData", false, "default");
            Assert.fail("There is no composite named 'undefined-subcomposite' in the model. The getValue method should have fail.");
        }
        catch (UndefinedItemPathException e)
        {
            Assert.assertTrue(e.getMessage().contains("defined"));
        }
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasValue(String)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testHasValueWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testHasValue", DATA_NODE_TYPE);
        _testHasValue(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasValue(String)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testHasValueWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testHasValue", DATA_NODE_TYPE);
        _testHasValue(sourceData);
    }

    private void _testHasValue(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        _addSomeValuesInRepositoryData(repositoryData);
        
        Assert.assertTrue(dataHolder.hasValue("singleStringData"));
        Assert.assertTrue(dataHolder.hasValue("multipleStringData"));
        Assert.assertFalse(dataHolder.hasValue("singleLongData"));
        Assert.assertFalse(dataHolder.hasValue("unexistingAttribute"));
        
        Assert.assertTrue(dataHolder.hasValue("composite"));
        Assert.assertTrue(dataHolder.hasValue("composite/singleLongData"));
        Assert.assertFalse(dataHolder.hasValue("composite/unexistingAttribute"));
        ModelAwareComposite composite = dataHolder.getComposite("composite");
        Assert.assertTrue(composite.hasValue("singleLongData"));
        Assert.assertFalse(composite.hasValue("unexistingAttribute"));

        Assert.assertTrue(dataHolder.hasValue("composite/repeater[1]"));
        Assert.assertFalse(dataHolder.hasValue("composite/repeater[1]/singleBooleanData"));
        Assert.assertTrue(dataHolder.hasValue("composite/repeater[1]/singleDoubleData"));
        Assert.assertFalse(dataHolder.hasValue("composite/repeater[1]/unexistingAttribute"));
        Assert.assertTrue(dataHolder.hasValue("composite/repeater[1]/subcomposite"));
        Assert.assertTrue(dataHolder.hasValue("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertFalse(dataHolder.hasValue("composite/repeater[1]/subcomposite/multipleDateData"));
        
        Assert.assertTrue(composite.hasValue("repeater[1]"));
        Assert.assertFalse(composite.hasValue("repeater[1]/singleBooleanData"));
        Assert.assertTrue(composite.hasValue("repeater[1]/singleDoubleData"));
        Assert.assertFalse(composite.hasValue("repeater[1]/unexistingAttribute"));
        Assert.assertTrue(composite.hasValue("repeater[1]/subcomposite"));
        Assert.assertTrue(composite.hasValue("repeater[1]/subcomposite/singleStringData"));
        Assert.assertFalse(composite.hasValue("repeater[1]/subcomposite/multipleDateData"));
        
        Repeater repeater = composite.getRepeater("repeater");
        RepeaterEntry repeaterEntry1 = repeater.getEntry(1);
        Assert.assertFalse(repeaterEntry1.hasValue("singleBooleanData"));
        Assert.assertTrue(repeaterEntry1.hasValue("singleDoubleData"));
        Assert.assertFalse(repeaterEntry1.hasValue("unexistingAttribute"));
        Assert.assertTrue(repeaterEntry1.hasValue("subcomposite"));
        Assert.assertTrue(repeaterEntry1.hasValue("subcomposite/singleStringData"));
        Assert.assertFalse(repeaterEntry1.hasValue("subcomposite/multipleDateData"));
        
        Composite subcomposite = repeaterEntry1.getComposite("subcomposite");
        Assert.assertTrue(subcomposite.hasValue("singleStringData"));
        Assert.assertFalse(subcomposite.hasValue("multipleDateData"));
        
        Assert.assertTrue(dataHolder.hasValue("composite/repeater[2]"));
        Assert.assertFalse(dataHolder.hasValue("composite/repeater[2]/singleBooleanData"));
        Assert.assertTrue(dataHolder.hasValue("composite/repeater[2]/singleDoubleData"));
        Assert.assertTrue(dataHolder.hasValue("composite/repeater[2]/subcomposite"));
        Assert.assertTrue(dataHolder.hasValue("composite/repeater[2]/subcomposite/singleStringData"));
        Assert.assertFalse(dataHolder.hasValue("composite/repeater[2]/subcomposite/multipleDateData"));
        
        Assert.assertFalse(dataHolder.hasValue("composite/repeater[3]"));
        Assert.assertFalse(dataHolder.hasValue("composite/repeater[3]/singleDoubleData"));
        
        repositoryData.setValue("singleLongData", "A String value for a long data");
        Assert.assertFalse(dataHolder.hasValue("singleLongData"));
        
        repositoryData.setValue("inexistingData", "a data that is not defined by the model");
        Assert.assertFalse(dataHolder.hasValue("inexistingData"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder} behavior with empty values
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testEmptyValuesWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testEmptyValues", DATA_NODE_TYPE);
        _testEmptyValues(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder} behavior with empty values
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testEmptyValuesWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testEmptyValues", DATA_NODE_TYPE);
        _testEmptyValues(sourceData);
    }
    
    private void _testEmptyValues(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);

        // single string data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("singleStringData"));
        Assert.assertNull(dataHolder.getValue("singleStringData"));
        repositoryData.setValue("singleStringData", StringUtils.EMPTY);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleStringData"));
        Assert.assertFalse(dataHolder.hasValue("singleStringData"));
        Assert.assertEquals(StringUtils.EMPTY, dataHolder.getValue("singleStringData"));
        
        // multiple string data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("multipleStringData"));
        Assert.assertNull(dataHolder.getValue("multipleStringData"));
        repositoryData.setValues("multipleStringData", new String[0]);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("multipleStringData"));
        Assert.assertFalse(dataHolder.hasValue("multipleStringData"));
        Assert.assertEquals(0, ((String[]) dataHolder.getValue("multipleStringData")).length);

        // single long data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("singleLongData"));
        Assert.assertNull(dataHolder.getValue("singleLongData"));
        repositoryData.setValue("singleLongData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleLongData"));
        Assert.assertFalse(dataHolder.hasValue("singleLongData"));
        Assert.assertNull(dataHolder.getValue("singleLongData"));
        
        // multiple long data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("multipleLongData"));
        Assert.assertNull(dataHolder.getValue("multipleLongData"));
        repositoryData.setValues("multipleLongData", new Long[0]);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("multipleLongData"));
        Assert.assertFalse(dataHolder.hasValue("multipleLongData"));
        Assert.assertEquals(0, ((Long[]) dataHolder.getValue("multipleLongData")).length);

        // single double data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("singleDoubleData"));
        Assert.assertNull(dataHolder.getValue("singleDoubleData"));
        repositoryData.setValue("singleDoubleData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleDoubleData"));
        Assert.assertFalse(dataHolder.hasValue("singleDoubleData"));
        Assert.assertNull(dataHolder.getValue("singleDoubleData"));
        
        // multiple double data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("multipleDoubleData"));
        Assert.assertNull(dataHolder.getValue("multipleDoubleData"));
        repositoryData.setValues("multipleDoubleData", new Double[0]);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("multipleDoubleData"));
        Assert.assertFalse(dataHolder.hasValue("multipleDoubleData"));
        Assert.assertEquals(0, ((Double[]) dataHolder.getValue("multipleDoubleData")).length);

        // single date data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("singleDateData"));
        Assert.assertNull(dataHolder.getValue("singleDateData"));
        repositoryData.setValue("singleDateData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleDateData"));
        Assert.assertFalse(dataHolder.hasValue("singleDateData"));
        Assert.assertNull(dataHolder.getValue("singleDateData"));
        
        // multiple date data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("multipleDateData"));
        Assert.assertNull(dataHolder.getValue("multipleDateData"));
        repositoryData.setValues("multipleDateData", new Calendar[0]);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("multipleDateData"));
        Assert.assertFalse(dataHolder.hasValue("multipleDateData"));
        Assert.assertEquals(0, ((LocalDate[]) dataHolder.getValue("multipleDateData")).length);

        // single date time data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("singleDatetimeData"));
        Assert.assertNull(dataHolder.getValue("singleDatetimeData"));
        repositoryData.setValue("singleDatetimeData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleDatetimeData"));
        Assert.assertFalse(dataHolder.hasValue("singleDatetimeData"));
        Assert.assertNull(dataHolder.getValue("singleDatetimeData"));
        
        // multiple date time data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("multipleDatetimeData"));
        Assert.assertNull(dataHolder.getValue("multipleDatetimeData"));
        repositoryData.setValues("multipleDatetimeData", new Calendar[0]);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("multipleDatetimeData"));
        Assert.assertFalse(dataHolder.hasValue("multipleDatetimeData"));
        Assert.assertEquals(0, ((ZonedDateTime[]) dataHolder.getValue("multipleDatetimeData")).length);
        
        // single user data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("singleUserData"));
        Assert.assertNull(dataHolder.getValue("singleUserData"));
        ModifiableRepositoryData userData = repositoryData.addRepositoryData("singleUserData", RepositoryConstants.USER_NODETYPE);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleUserData"));
        Assert.assertFalse(dataHolder.hasValue("singleUserData"));
        Assert.assertNull(dataHolder.getValue("singleUserData"));
        userData.setValue("login", "tyork");
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleUserData"));
        Assert.assertFalse(dataHolder.hasValue("singleUserData"));
        Assert.assertNull(dataHolder.getValue("singleUserData"));
        userData.setValue("login", StringUtils.EMPTY);
        userData.setValue("population", "radiohead");
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleUserData"));
        Assert.assertFalse(dataHolder.hasValue("singleUserData"));
        Assert.assertNull(dataHolder.getValue("singleUserData"));
        
        // multiple user data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("multipleUserData"));
        Assert.assertNull(dataHolder.getValue("multipleUserData"));
        ModifiableRepositoryData multipleUserData = repositoryData.addRepositoryData("multipleUserData", RepositoryConstants.MULTIPLE_ITEM_NODETYPE);
        multipleUserData.setValue(ComplexRepositoryElementType.TYPE_ID_DATA_NAME, ModelItemTypeConstants.USER_ELEMENT_TYPE_ID, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("multipleUserData"));
        Assert.assertFalse(dataHolder.hasValue("multipleUserData"));
        Assert.assertEquals(0, ((UserIdentity[]) dataHolder.getValue("multipleUserData")).length);
        
        // single content data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("singleContentData"));
        Assert.assertNull(dataHolder.getValue("singleContentData"));
        repositoryData.setValue("singleContentData", StringUtils.EMPTY);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleContentData"));
        Assert.assertFalse(dataHolder.hasValue("singleContentData"));
        Assert.assertNull(dataHolder.getValue("singleContentData"));
        
        // multiple content data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("multipleContentData"));
        Assert.assertNull(dataHolder.getValue("multipleContentData"));
        repositoryData.setValues("multipleContentData", new String[0]);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("multipleContentData"));
        Assert.assertFalse(dataHolder.hasValue("multipleContentData"));
        Assert.assertEquals(0, ((ContentValue[]) dataHolder.getValue("multipleContentData")).length);
        
        // composite data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("composite"));
        Assert.assertNull(dataHolder.getValue("composite"));
        ModifiableRepositoryData compositeData = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("composite"));
        Assert.assertFalse(dataHolder.hasValue("composite"));
        Assert.assertTrue(((ModelAwareComposite) dataHolder.getValue("composite")).getDataNames().isEmpty());
        
        // repeater data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("composite/repeater"));
        Assert.assertNull(dataHolder.getValue("composite/repeater"));
        ModifiableRepositoryData repeaterData = compositeData.addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("composite/repeater"));
        Assert.assertFalse(dataHolder.hasValue("composite/repeater"));
        Assert.assertEquals(0, ((ModelAwareRepeater) dataHolder.getValue("composite/repeater")).getSize());
        ModifiableRepositoryData repeaterEntry = repeaterData.addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE);
        Assert.assertTrue(dataHolder.hasValue("composite/repeater"));
        Assert.assertEquals(1, ((ModelAwareRepeater) dataHolder.getValue("composite/repeater")).getSize());
        Assert.assertTrue(dataHolder.hasValueOrEmpty("composite/repeater[1]"));
        Assert.assertFalse(dataHolder.hasValue("composite/repeater[1]"));
        Assert.assertTrue(((ModelAwareRepeaterEntry) dataHolder.getValue("composite/repeater[1]")).getDataNames().isEmpty());
        
        // data in a composite / repeater
        Assert.assertFalse(dataHolder.hasValueOrEmpty("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertNull(dataHolder.getValue("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertFalse(dataHolder.hasValueOrEmpty("composite/repeater[1]/subcomposite/multipleDateData"));
        Assert.assertNull(dataHolder.getValue("composite/repeater[1]/subcomposite/multipleDateData"));
        ModifiableRepositoryData subcomposite = repeaterEntry.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        Assert.assertFalse(dataHolder.hasValueOrEmpty("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertNull(dataHolder.getValue("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertFalse(dataHolder.hasValueOrEmpty("composite/repeater[1]/subcomposite/multipleDateData"));
        Assert.assertNull(dataHolder.getValue("composite/repeater[1]/subcomposite/multipleDateData"));
        subcomposite.setValue("singleStringData", StringUtils.EMPTY);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertFalse(dataHolder.hasValue("composite/repeater[1]/subcomposite/singleStringData"));
        Assert.assertEquals(StringUtils.EMPTY, dataHolder.getValue("composite/repeater[1]/subcomposite/singleStringData"));
        subcomposite.setValues("multipleDateData", new Calendar[0]);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("composite/repeater[1]/subcomposite/multipleDateData"));
        Assert.assertFalse(dataHolder.hasValue("composite/repeater[1]/subcomposite/multipleDateData"));
        Assert.assertEquals(0, ((LocalDate[]) dataHolder.getValue("composite/repeater[1]/subcomposite/multipleDateData")).length);
    }
}
