/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.holder.impl.DefaultModelAwareDataHolder;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.DataComment;
import org.ametys.plugins.repository.data.external.ExternalizableDataProvider.ExternalizableDataStatus;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.values.SynchronizableValue;
import org.ametys.plugins.repository.data.holder.values.SynchronizationContext;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method test cases
 */
@NeedFeatures({
    "test-data/test-model",
    
    "repository/repository.data.holder.helpers",
    "repository/resolver", 
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModelAwareDataHolderHasDifferencesTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __DATA_NODE_TYPE = "ametys:unstructured";
    
    private static final String __STRING_DATA_NAME = "singleStringData"; 
    private static final String __STRING_DEFINITION_PATH = "composite/repeater/subcomposite/" + __STRING_DATA_NAME;
    private static final String __MULTIPLE_STRING_DEFINITION_PATH = "multipleStringData";
    private static final String __MULTIPLE_STRING_DATA_PATH = "multipleStringData";
    private static final String __STRING_VALUE = "value";
    private static final String __STRING_LOCAL_VALUE = "local value";
    private static final String __STRING_EXTERNAL_VALUE = "external value";
    private static final String __STRING_NEW_VALUE = "new value";
    private static final String __STRING_DEFAULT_VALUE = "default value";
    private static final String __STRING_VALUE_2 = "value 2";
    private static final String __APPENDED_STRING_VALUE = "appended value";
    private static final String __APPENDED_STRING_VALUE_2 = "appended value 2";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static Model _model;

    private static JCRRepositoryData _root;
    private static View _completeView;
    private static View _viewWithOnlyMultipledata;
    private static View _emptyView;
    
    private ModifiableRepositoryData _repositoryData;
    private ModelAwareDataHolder _dataHolder;
    private SynchronizationContext _context;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _root = new JCRRepositoryData(rootAmetysObject.getNode());
        
        // For these tests, add a default value to the single string data in the subcomposite
        @SuppressWarnings("unchecked")
        ElementDefinition<String> definition = (ElementDefinition<String>) _model.getModelItem(__STRING_DEFINITION_PATH);
        definition.setDefaultValue(__STRING_DEFAULT_VALUE);

        _completeView = View.of(_model, __STRING_DEFINITION_PATH, __MULTIPLE_STRING_DEFINITION_PATH);
        _viewWithOnlyMultipledata = View.of(_model, __MULTIPLE_STRING_DEFINITION_PATH);
        _emptyView = new View();
    }
    
    /**
     * Prepare components that are specific for each test
     * @throws Exception if an error occurs
     */
    @Before
    public void prepareSpecificComponents()throws Exception
    {
        _repositoryData = _root.addRepositoryData(UUID.randomUUID().toString(), __DATA_NODE_TYPE);
        _repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE)
                       .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
                       .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
                       .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        
        _dataHolder = new DefaultModelAwareDataHolder(_repositoryData, _model);
        _context = SynchronizationContext.newInstance();
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * The view is empty
     * Result: no diff
     * @throws Exception if an error occurs
     */
    @Test
    public void testViewEmpty() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of(__STRING_DATA_NAME, __STRING_NEW_VALUE)))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_emptyView, valuesToSynchronize, _context);

        Assert.assertFalse(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * The repository is empty
     * Values to synchronize don't contain the data
     * Use the default value
     * Result: difference to set the default value
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyValueEmptyUsingDefault() throws Exception
    {
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        _context.withDefaultFromModel(true);
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * The repository is empty
     * Values to synchronize contain a <code>null</code> value
     * Use the default value
     * Result: difference to set the default value
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyValueNullUsingDefault() throws Exception
    {
        Map<String, Object> emptyStringValue = new HashMap<>();
        emptyStringValue.put(__STRING_DATA_NAME, null);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", emptyStringValue))));
        
        _context.withDefaultFromModel(true);
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * The repository is empty
     * Values to synchronize don't contain the data
     * Do not use the default value
     * Result: no difference
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyValueEmpty() throws Exception
    {
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertFalse(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * The repository is empty
     * Values to synchronize contain a <code>null</code> value
     * Do not use the default value
     * Result: difference to set an empty value
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyValueNull() throws Exception
    {
        Map<String, Object> emptyStringValue = new HashMap<>();
        emptyStringValue.put(__STRING_DATA_NAME, null);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", emptyStringValue))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * The repository is empty
     * Values to synchronize contain a  value
     * Result: difference to set the value
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptySetValue() throws Exception
    {
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of(__STRING_DATA_NAME, __STRING_VALUE)))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize don't contain the data
     * Use the default value
     * Result: differnce to set the default value
     * @throws Exception if an error occurs
     */
    @Test
    public void testValueEmptyUsingDefault() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        _context.withDefaultFromModel(true);
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize don't contain the data
     * Do not use the default value
     * Result: difference to remove the value
     * @throws Exception if an error occurs
     */
    @Test
    public void testValueEmpty() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize contain a <code>null</code> value
     * Do not use the default value
     * Result: difference to set an empty value
     * @throws Exception if an error occurs
     */
    @Test
    public void testValueNull() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        
        Map<String, Object> emptyStringValue = new HashMap<>();
        emptyStringValue.put(__STRING_DATA_NAME, null);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", emptyStringValue))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize does not change the value
     * Result: no difference
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotChangeValue() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of(__STRING_DATA_NAME, __STRING_VALUE)))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertFalse(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize changes the value
     * Result: difference to set the new value
     * @throws Exception if an error occurs
     */
    @Test
    public void testChangeValue() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of(__STRING_DATA_NAME, __STRING_NEW_VALUE)))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize remove a composite
     * Result: difference to remove the composite
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveComposite() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of())));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize remove a repeater entry
     * Result: difference to remove the repeater entry
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveRepeaterEntry() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of()));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize remove a repeater
     * Result: difference to remove the repeater
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveRepeater() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of());
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Insert a value on a single data
     * Result: difference to set the new value
     * @throws Exception if an error occurs
     */
    @Test
    public void testAppendValue() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(__STRING_NEW_VALUE);
        value.setMode(SynchronizableValue.Mode.APPEND);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of(__STRING_DATA_NAME, value)))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Insert values on a multiple data
     * Result: difference to set the new values
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyAppendValues() throws Exception
    {
        SynchronizableValue value = new SynchronizableValue(new String[] {__APPENDED_STRING_VALUE, __APPENDED_STRING_VALUE_2});
        value.setMode(SynchronizableValue.Mode.APPEND);
        Map<String, Object> valuesToSynchronize = Map.of(__MULTIPLE_STRING_DATA_PATH, value);
        
        boolean hasDiff = _dataHolder.hasDifferences(_viewWithOnlyMultipledata, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Insert values on a multiple data
     * Result: difference to add the new values to the existent ones
     * @throws Exception if an error occurs
     */
    @Test
    public void testAppendValues() throws Exception
    {
        _repositoryData.setValues(__MULTIPLE_STRING_DATA_PATH, new String[] {__STRING_VALUE, __STRING_VALUE_2});
        
        SynchronizableValue value = new SynchronizableValue(new String[] {__APPENDED_STRING_VALUE, __APPENDED_STRING_VALUE_2});
        value.setMode(SynchronizableValue.Mode.APPEND);
        Map<String, Object> valuesToSynchronize = Map.of(__MULTIPLE_STRING_DATA_PATH, value);
        
        boolean hasDiff = _dataHolder.hasDifferences(_viewWithOnlyMultipledata, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Remove the value of a single data
     * Result: difference to remove the value
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveValue() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(null);
        value.setMode(SynchronizableValue.Mode.REMOVE);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of(__STRING_DATA_NAME, value)))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Remove some values of a multiple data
     * Result: difference to remove some values
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveValues() throws Exception
    {
        _repositoryData.setValues(__MULTIPLE_STRING_DATA_PATH, new String[] {__STRING_VALUE, __STRING_VALUE_2});
        
        SynchronizableValue value = new SynchronizableValue(new String[] {__STRING_VALUE_2});
        value.setMode(SynchronizableValue.Mode.REMOVE);
        Map<String, Object> valuesToSynchronize = Map.of(__MULTIPLE_STRING_DATA_PATH, value);
        
        boolean hasDiff = _dataHolder.hasDifferences(_viewWithOnlyMultipledata, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Put empty array value for an empty multiple data
     * Result: no change
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotChangeEmptyValues() throws Exception
    {
        _repositoryData.setValues(__MULTIPLE_STRING_DATA_PATH, new String[] {});
        
        Map<String, Object> valuesToSynchronize = Map.of(__MULTIPLE_STRING_DATA_PATH, new String[] {});
        
        boolean hasDiff = _dataHolder.hasDifferences(_viewWithOnlyMultipledata, valuesToSynchronize, _context);

        Assert.assertFalse(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Put null value for an empty multiple data
     * Result: no change
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotChangeEmptyValuesWithNull() throws Exception
    {
        _repositoryData.setValues(__MULTIPLE_STRING_DATA_PATH, new String[] {});
        
        Map<String, Object> valuesToSynchronize = new HashMap<>();
        valuesToSynchronize.put(__MULTIPLE_STRING_DATA_PATH, null);
        
        boolean hasDiff = _dataHolder.hasDifferences(_viewWithOnlyMultipledata, valuesToSynchronize, _context);

        Assert.assertFalse(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Repository contains externalizable metadata
     * Values to synchronize don't contain the data
     * Use the default value
     * Result: difference to set the default value and remove the externalizable meta
     * @throws Exception if an error occurs
     */
    @Test
    public void testValueEmptyUsingDefaultRemoveExternalizableMetadata() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME + ModelAwareDataHolder.STATUS_SUFFIX, ExternalizableDataStatus.EXTERNAL.name().toLowerCase());
        _getSubcompositeData().setValue(__STRING_DATA_NAME + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, __STRING_LOCAL_VALUE);
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_EXTERNAL_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        _context.withDefaultFromModel(true);
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Repository contains externalizable metadata
     * Values to synchronize don't contain the data
     * Do not use the default value
     * Result: difference to remove the value and the externalizable meta
     * @throws Exception if an error occurs
     */
    @Test
    public void testValueEmptyRemoveExternalizableMetadata() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME + ModelAwareDataHolder.STATUS_SUFFIX, ExternalizableDataStatus.EXTERNAL.name().toLowerCase());
        _getSubcompositeData().setValue(__STRING_DATA_NAME + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, __STRING_LOCAL_VALUE);
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_EXTERNAL_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Repository contains externalizable metadata
     * Values to synchronize does not change the local value
     * Result: difference to remove the externalizable meta
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotChangeValueRemoveExternalizableMetadata() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME + ModelAwareDataHolder.STATUS_SUFFIX, ExternalizableDataStatus.EXTERNAL.name().toLowerCase());
        _getSubcompositeData().setValue(__STRING_DATA_NAME + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, __STRING_LOCAL_VALUE);
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_EXTERNAL_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of(__STRING_DATA_NAME, __STRING_LOCAL_VALUE)))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Repository contains externalizable metadata
     * Values to synchronize set the external value as the new one
     * Result: no difference in values but the externalizable meta should be removed
     * @throws Exception if an error occurs
     */
    @Test
    public void testUseExternalValueChangeStatusAndRemoveExternalizableMetadata() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME + ModelAwareDataHolder.STATUS_SUFFIX, ExternalizableDataStatus.EXTERNAL.name().toLowerCase());
        _getSubcompositeData().setValue(__STRING_DATA_NAME + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, __STRING_LOCAL_VALUE);
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(__STRING_VALUE);
        value.setExternalizableStatus(ExternalizableDataStatus.LOCAL);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of(__STRING_DATA_NAME, value)))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertFalse(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Repository contains externalizable metadata
     * Values to synchronize changes the value
     * Result: difference to set the new value and remove the externalizable meta
     * @throws Exception if an error occurs
     */
    @Test
    public void testChangeValueRemoveExternalizableMetadata() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME + ModelAwareDataHolder.STATUS_SUFFIX, ExternalizableDataStatus.EXTERNAL.name().toLowerCase());
        _getSubcompositeData().setValue(__STRING_DATA_NAME + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, __STRING_LOCAL_VALUE);
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_EXTERNAL_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of(__STRING_DATA_NAME, __STRING_NEW_VALUE)))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize does not contain comments, they are not touched
     * Result: no change
     * @throws Exception if an error occurs
     */
    @Test
    public void testIgnoreComments() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        _addSomeCommentsToSubcomposite();
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of(__STRING_DATA_NAME, __STRING_VALUE)))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertFalse(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize contains comments but they are unchanged
     * Result: no difference
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotChangeComments() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        _addSomeCommentsToSubcomposite();
        
        SynchronizableValue value = new SynchronizableValue(__STRING_VALUE);
        value.setComments(List.of(
                new DataComment("First comment", ZonedDateTime.of(2020, 10, 9, 20, 30, 0, 0, ZoneId.systemDefault()), "Café Plum"),
                new DataComment("Second comment", ZonedDateTime.of(2021, 2, 2, 20, 30, 0, 0, ZoneId.systemDefault()), "Rocher Palmer")));
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of(__STRING_DATA_NAME, value)))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertFalse(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize contains new comments
     * @throws Exception if an error occurs
     */
    @Test
    public void testAddComments() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(__STRING_VALUE);
        value.setComments(List.of(
                new DataComment("New comment", ZonedDateTime.of(2021, 3, 19, 20, 0, 0, 0, ZoneId.systemDefault()), "Victoire 2")));
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of(__STRING_DATA_NAME, value)))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize contains new comments
     * @throws Exception if an error occurs
     */
    @Test
    public void testChangeComments() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        _addSomeCommentsToSubcomposite();
        
        SynchronizableValue value = new SynchronizableValue(__STRING_VALUE);
        value.setComments(List.of(
                new DataComment("New comment", ZonedDateTime.of(2021, 3, 19, 20, 0, 0, 0, ZoneId.systemDefault()), "Victoire 2")));
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of(__STRING_DATA_NAME, value)))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#hasDifferences(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to check contain an empty list of comments to remove existing ones
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveComments() throws Exception
    {
        _getSubcompositeData().setValue(__STRING_DATA_NAME, __STRING_VALUE);
        _addSomeCommentsToSubcomposite();
        
        SynchronizableValue value = new SynchronizableValue(__STRING_VALUE);
        value.setComments(List.of());
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of(__STRING_DATA_NAME, value)))));
        
        boolean hasDiff = _dataHolder.hasDifferences(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(hasDiff);
    }

    private void _addSomeCommentsToSubcomposite()
    {
        ModifiableRepositoryData commentsRepositoryData = _getSubcompositeData()
                .addRepositoryData(__STRING_DATA_NAME + ModelAwareDataHolder.COMMENTS_SUFFIX, RepositoryConstants.COMPOSITE_METADTA_NODETYPE, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        
        ModifiableRepositoryData firstCommentRepositoryData = commentsRepositoryData.addRepositoryData(String.valueOf(1), RepositoryConstants.COMPOSITE_METADTA_NODETYPE, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        firstCommentRepositoryData.setValue("comment", "First comment");
        firstCommentRepositoryData.setValue("author", "Café Plum");
        firstCommentRepositoryData.setValue("date", DateUtils.asCalendar(ZonedDateTime.of(2020, 10, 9, 20, 30, 0, 0, ZoneId.systemDefault())));
        
        ModifiableRepositoryData secondCommentRepositoryData = commentsRepositoryData.addRepositoryData(String.valueOf(2), RepositoryConstants.COMPOSITE_METADTA_NODETYPE, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        secondCommentRepositoryData.setValue("comment", "Second comment");
        secondCommentRepositoryData.setValue("author", "Rocher Palmer");
        secondCommentRepositoryData.setValue("date", DateUtils.asCalendar(ZonedDateTime.of(2021, 2, 2, 20, 30, 0, 0, ZoneId.systemDefault())));
    }

    private ModifiableRepositoryData _getSubcompositeData()
    {
        return _repositoryData.getRepositoryData("composite")
                              .getRepositoryData("repeater").getRepositoryData("1")
                              .getRepositoryData("subcomposite");
    }
}
