/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.time.LocalDate;
import java.time.Month;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Arrays;
import java.util.Calendar;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Stream;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Binary;
import org.ametys.cms.data.ExplorerFile;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.RichText;
import org.ametys.cms.data.holder.impl.DefaultModelAwareDataHolder;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.cms.utils.ResourceElementTypeHelperForTests;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.external.ExternalizableDataProviderExtensionPoint;
import org.ametys.plugins.repository.data.external.ExternalizableDataProvider.ExternalizableDataStatus;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeater;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.plugins.repository.data.type.RepositoryModelItemType;
import org.ametys.plugins.repository.model.RepositoryDataContext;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElement;
import org.ametys.runtime.model.ViewElementAccessor;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelAwareDataHolder} toJSON methods implementation test case
 */
@NeedFeatures({
    "test-data/test-model",
    "test-data/test-externalizable-data-provider",
    "cms/factories",
    "test/content-type",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModelAwareDataHolderToJSONTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __DATA_NODE_TYPE = "ametys:unstructured";
    
    private static Model _model;
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    private static ExternalizableDataProviderExtensionPoint _externalizableDataProviderEP;
    
    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;
    
    private static RepositoryElementType<Geocode> _geocodeType;
    private static RepositoryElementType<UserIdentity> _userType;
    private static RepositoryElementType<Binary> _binaryType;
    private static RepositoryElementType<ExplorerFile> _fileType;
    private static RepositoryElementType<RichText> _richTextType;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
        _externalizableDataProviderEP = (ExternalizableDataProviderExtensionPoint) manager.lookup(ExternalizableDataProviderExtensionPoint.ROLE);
    }
    
    @Override
    @SuppressWarnings("unchecked")
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");
        
        _geocodeType = (RepositoryElementType<Geocode>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID);
        _userType = (RepositoryElementType<UserIdentity>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _binaryType = (RepositoryElementType<Binary>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _fileType = (RepositoryElementType<ExplorerFile>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.FILE_ELEMENT_TYPE_ID);
        _richTextType = (RepositoryElementType<RichText>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON()} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderToJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderToJSON", __DATA_NODE_TYPE);
        _testDataHolderToJSON(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON()} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderToJSONWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderToJSON", __DATA_NODE_TYPE);
        _testDataHolderToJSON(sourceData);
    }
    
    @SuppressWarnings("unchecked")
    private void _testDataHolderToJSON(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent content = _createContent("content");
        _populateRepositoryData(repositoryData, content);
        
        // Convert the data holder to JSON
        Set<String> externalizableData = _externalizableDataProviderEP.getExternalizableDataPaths(content);
        RepositoryDataContext context = RepositoryDataContext.newInstance()
                                                             .withExternalizableData(externalizableData);
        Map<String, Object> json = dataHolder.dataToJSON(context);
        
        // Check the json result
        Assert.assertEquals(8, json.size());
        
        Assert.assertEquals("value of the single string data", json.get("singleStringData"));
        Assert.assertTrue(json.containsKey("singleLongData"));
        Assert.assertNull(json.get("singleLongData"));
        Assert.assertEquals(List.of(32L, 784L), json.get("multipleLongData"));
        Assert.assertEquals(Map.of("latitude", 85.487D, "longitude", 6.254D), json.get("singleGeocodeData"));
        
        List<Map<String, Object>> users = (List<Map<String, Object>>) json.get("multipleUserData");
        Map<String, Object> user1 = users.get(0);
        Assert.assertEquals("user1", user1.get("login"));
        Assert.assertEquals(POPULATION_TEST_ID, user1.get("populationId"));
        Assert.assertEquals("User One", user1.get("fullname"));
        Map<String, Object> user2 = users.get(1);
        Assert.assertEquals("user3", user2.get("login"));
        Assert.assertEquals(POPULATION_TEST_ID, user2.get("populationId"));
        Assert.assertEquals("User Three", user2.get("fullname"));        

        Map<String, Object> singleContentData = (Map<String, Object>) json.get("singleContentData");
        Assert.assertEquals(10, singleContentData.size());
        Assert.assertEquals(content.getId(), singleContentData.get("id"));
        Assert.assertEquals("content", singleContentData.get("title"));
        Assert.assertEquals(List.of("2018-05-05T02:25:13.596Z", "2015-12-06T18:18:00.000Z"), json.get("multipleDatetimeData"));
        
        // composite
        Map<String, Object> composite = (Map<String, Object>) json.get("composite");
        Assert.assertEquals(2, composite.size());
        Assert.assertEquals(54L, composite.get("singleLongData"));
        
        // composite > repeater
        Map<String, Object> repeater = (Map<String, Object>) composite.get("repeater");
        Assert.assertEquals(3, repeater.size());
        Assert.assertEquals(new I18nizableText("Repeater"), repeater.get("label"));
        Assert.assertEquals(2, repeater.get("entryCount"));
        List<Map<String, Object>> entries = (List<Map<String, Object>>) repeater.get("entries");
        Assert.assertEquals(2, entries.size());
        
        // composite > repeater[1]
        Map<String, Object> entry1 = entries.get(0);
        Assert.assertEquals(2, entry1.size());
        Assert.assertEquals(1, entry1.get("position"));
        Map<String, Object> entry1Values = (Map<String, Object>) entry1.get("values");
        Assert.assertEquals(3, entry1Values.size());
        Assert.assertEquals(15L, entry1Values.get("longWithDefaultValue"));
        Assert.assertEquals(0.6D, entry1Values.get("singleDoubleData"));

        // composite > repeater[1] > subcomposite
        Map<String, Object> entry1Subcomposite = (Map<String, Object>) entry1Values.get("subcomposite");
        Assert.assertEquals(2, entry1Subcomposite.size());
        Assert.assertEquals("string value in the sub composite of the first repeater entry", entry1Subcomposite.get("singleStringData"));
        Assert.assertEquals(List.of("2018-09-11", "2015-12-06"), entry1Subcomposite.get("multipleDateData"));
        
        // composite > repeater[2]
        Map<String, Object> entry2 = entries.get(1);
        Assert.assertEquals(2, entry2.size());
        Assert.assertEquals(2, entry2.get("position"));
        Map<String, Object> entry2Values = (Map<String, Object>) entry2.get("values");
        Assert.assertEquals(3, entry2Values.size());
        Assert.assertEquals(95L, entry2Values.get("longWithDefaultValue"));
        Assert.assertTrue((boolean) entry2Values.get("singleBooleanData"));

        // composite > repeater[2] > subcomposite
        Map<String, Object> entry2Subcomposite = (Map<String, Object>) entry2Values.get("subcomposite");
        Assert.assertEquals(1, entry2Subcomposite.size());
        Assert.assertEquals(List.of("1985-02-27", "1985-05-24"), entry2Subcomposite.get("multipleDateData"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON()} implementation with resources
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testDataHolderResourcesToJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testDataHolderResourcesToJSON", __DATA_NODE_TYPE);
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        String fileId = ResourceElementTypeHelperForTests.createResource("myFile.txt", "my file content", _ametysObjectResolver);
        _populateRepositoryDataWithResources(repositoryData, fileId);
        
        // Convert the data holder to JSON
        Map<String, Object> json = dataHolder.dataToJSON();
        
        // Check the json result
        Assert.assertEquals(4, json.size());

        // singleBinaryData
        Map<String, Object> singleBinaryData = (Map<String, Object>) json.get("singleBinaryData");
        Assert.assertEquals("singleBinaryData", singleBinaryData.get("path"));
        Assert.assertEquals("binaryTypeForTests", singleBinaryData.get("type"));
        
        // singleFileData
        Map<String, Object> singleFileData = (Map<String, Object>) json.get("singleFileData");
        Assert.assertEquals(fileId, singleFileData.get("id"));
        Assert.assertEquals(fileId, singleFileData.get("path"));
        Assert.assertEquals("explorer", singleFileData.get("type"));
        
        // singleRichTextData
        Map<String, Object> singleRichTextData = (Map<String, Object>) json.get("singleRichTextData");
        Assert.assertEquals("my rich text content", singleRichTextData.get("value"));
        
        // composite
        Map<String, Object> composite = (Map<String, Object>) json.get("composite");
        Assert.assertEquals(1, composite.size());
        
        // composite > repeater
        Map<String, Object> repeater = (Map<String, Object>) composite.get("repeater");
        Assert.assertEquals(3, repeater.size());
        Assert.assertEquals(new I18nizableText("Repeater"), repeater.get("label"));
        Assert.assertEquals(2, repeater.get("entryCount"));
        List<Map<String, Object>> entries = (List<Map<String, Object>>) repeater.get("entries");
        Assert.assertEquals(2, entries.size());
        
        // composite > repeater[1]
        Map<String, Object> entry1 = entries.get(0);
        Assert.assertEquals(2, entry1.size());
        Assert.assertEquals(1, entry1.get("position"));
        Map<String, Object> entry1Values = (Map<String, Object>) entry1.get("values");
        Assert.assertEquals(1, entry1Values.size());
        
        // composite > repeater[1] > subcomposite
        Map<String, Object> entry1Subcomposite = (Map<String, Object>) entry1Values.get("subcomposite");
        Assert.assertEquals(1, entry1Subcomposite.size());
        
        // composite > repeater[1] > subcomposite > singleBinaryData
        Map<String, Object> entry1SingleBinaryData = (Map<String, Object>) entry1Subcomposite.get("singleBinaryData");
        Assert.assertEquals("composite/repeater[1]/subcomposite/singleBinaryData", entry1SingleBinaryData.get("path"));
        Assert.assertEquals("binaryTypeForTests", entry1SingleBinaryData.get("type"));
     
        // composite > repeater[2]
        Map<String, Object> entry2 = entries.get(1);
        Assert.assertEquals(2, entry2.size());
        Assert.assertEquals(2, entry2.get("position"));
        Map<String, Object> entry2Values = (Map<String, Object>) entry2.get("values");
        Assert.assertEquals(1, entry2Values.size());
        
        // composite > repeater[2] > subcomposite
        Map<String, Object> entry2Subcomposite = (Map<String, Object>) entry2Values.get("subcomposite");
        Assert.assertEquals(1, entry2Subcomposite.size());
        
        // composite > repeater[2] > subcomposite > singleBinaryData
        Map<String, Object> entry2SingleBinaryData = (Map<String, Object>) entry2Subcomposite.get("singleBinaryData");
        Assert.assertEquals("composite/repeater[2]/subcomposite/singleBinaryData", entry2SingleBinaryData.get("path"));
        Assert.assertEquals("binaryTypeForTests", entry2SingleBinaryData.get("type"));
    }
    
    private void _populateRepositoryDataWithResources(ModifiableRepositoryData repositoryData, String fileId) throws Exception
    {
        _binaryType.write(repositoryData, "singleBinaryData", ResourceElementTypeHelperForTests.createBinary("my binary content"));
        _fileType.write(repositoryData, "singleFileData", fileId);
        _richTextType.write(repositoryData, "singleRichTextData", ResourceElementTypeHelperForTests.createRichText("my rich text content"));
        
        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        ModifiableRepositoryData repeater = composite.addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE);

        ModifiableRepositoryData entry1 = repeater.addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE);
        ModifiableRepositoryData entry1SubComposite = entry1.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        _binaryType.write(entry1SubComposite, "singleBinaryData", ResourceElementTypeHelperForTests.createBinary("content of my binary in first entry of a repeater"));
        
        ModifiableRepositoryData entry2 = repeater.addRepositoryData("2", RepositoryConstants.COMPOSITE_NODETYPE);
        ModifiableRepositoryData entry2SubComposite = entry2.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        _binaryType.write(entry2SubComposite, "singleBinaryData", ResourceElementTypeHelperForTests.createBinary("content of my binary in second entry of a repeater"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON(ViewItemAccessor)} implementation with binaries in a nested view element
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testDataHolderBinaryInNestedElementToJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testDataHolderResourceInNestedElementToJSON", __DATA_NODE_TYPE);
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        ModifiableContent exh1 = _createExhaustiveContentWithBinary("content1", "binary of my first exhaustive content");
        ModifiableContent exh2 = _createExhaustiveContentWithBinary("content2", "binary of my second exhaustive content");
        _populateRepositoryDataWithBinaryInNestedContent(repositoryData, exh1, exh2);
        View view = _createViewWithNestedBinaryData();
        
        // Convert the data holder to JSON
        Map<String, Object> json = dataHolder.dataToJSON(view);
        
        // Check the json result
        Assert.assertEquals(1, json.size());
        
        List<Map<String, Object>> multipleExhContentData = (List<Map<String, Object>>) json.get("multipleExhContentData");
        Assert.assertEquals(2, multipleExhContentData.size());
        
        Map<String, Object> contentData1 = multipleExhContentData.get(0);
        Assert.assertEquals(11, contentData1.size());
        Assert.assertEquals(exh1.getId(), contentData1.get("id"));
        Assert.assertEquals("content1", contentData1.get("title"));
        
        Map<String, Object> content1Binary = (Map<String, Object>) contentData1.get("binary");
        Assert.assertEquals("binary", content1Binary.get("path"));
        Assert.assertEquals("binary?objectId=" + exh1.getId(), content1Binary.get("downloadUrl"));
        Assert.assertEquals("attribute", content1Binary.get("type"));
        
        Map<String, Object> contentData2 = multipleExhContentData.get(1);
        Assert.assertEquals(11, contentData2.size());
        Assert.assertEquals(exh2.getId(), contentData2.get("id"));
        Assert.assertEquals("content2", contentData2.get("title"));
        
        Map<String, Object> content2Binary = (Map<String, Object>) contentData2.get("binary");
        Assert.assertEquals("binary", content2Binary.get("path"));
        Assert.assertEquals("binary?objectId=" + exh2.getId(), content2Binary.get("downloadUrl"));
        Assert.assertEquals("attribute", content2Binary.get("type"));
    }
    
    private void _populateRepositoryDataWithBinaryInNestedContent(ModifiableRepositoryData repositoryData, Content... multipleExhaustiveContentValues)
    {
        String[] multipleExhaustiveContentIds = Arrays.stream(multipleExhaustiveContentValues)
                .map(Content::getId)
                .toArray(String[]::new);
        repositoryData.setValues("multipleExhContentData", multipleExhaustiveContentIds);
    }
    
    private ModifiableContent _createExhaustiveContentWithBinary(String title, String binaryContent) throws Exception
    {
        ModifiableContent content = _createContent(title, "org.ametys.test.Content.exhaustive");
        content.setValue("binary", ResourceElementTypeHelperForTests.createBinary(binaryContent));
        return content;
    }
    
    private View _createViewWithNestedBinaryData()
    {
        View view = new View();
        
        // Create a view item with nested attribute binary
        ViewElementAccessor contentDataViewItem = new ViewElementAccessor();
        contentDataViewItem.setDefinition((ElementDefinition) _model.getModelItem("multipleExhContentData"));
        
        ViewElement viewElement = new ViewElement();
        viewElement.setDefinition((ElementDefinition) _model.getModelItem("multipleExhContentData/binary"));
        contentDataViewItem.addViewItem(viewElement);
        
        // Insert the content view item in the view
        view.addViewItem(contentDataViewItem);
        
        return view;
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON()} implementation with enumerated data
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderEnumeratedDataToJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderEnumeratedDataToJSON", __DATA_NODE_TYPE);
        _testDataHolderEnumeratedDataToJSON(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON()} implementation with enumerated data
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderEnumeratedDataToJSONWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderEnumeratedDataToJSON", __DATA_NODE_TYPE);
        _testDataHolderEnumeratedDataToJSON(sourceData);
    }
    
    private void _testDataHolderEnumeratedDataToJSON(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        repositoryData.setValue("singleEnumeratedStringData", "value1");
        repositoryData.setValue("singleEnumeratedLongData", 3L);
        
        // Convert the data holder to JSON
        Map<String, Object> json = dataHolder.dataToJSON();
        
        // Check the json result
        Assert.assertEquals(2, json.size());
        
        Assert.assertEquals(Map.of("label", new I18nizableText("Value 1"),
                                   "value", "value1"), json.get("singleEnumeratedStringData"));
        
        Assert.assertEquals(Map.of("label", new I18nizableText("Value 3"),
                                   "value", 3L), json.get("singleEnumeratedLongData"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON(String)} implementation
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testDataHolderSingleDataToJSONWithJCRRepositoryData()
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSingleDataToJSON", __DATA_NODE_TYPE);
        _testDataHolderSingleDataToJSON(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON(String)} implementation
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testDataHolderSingleDataToJSONWithMemoryRepositoryData()
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSingleDataToJSON", __DATA_NODE_TYPE);
        _testDataHolderSingleDataToJSON(sourceData);
    }
    
    private void _testDataHolderSingleDataToJSON(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        _setValueInRepeaterEntryInRepositoryData(repositoryData, Optional.of("string value"));
        
        // Convert the data holder to JSON
        Object json = dataHolder.dataToJSON("composite/repeater[1]/subcomposite/singleStringData");
        
        // Check the json result
        Assert.assertEquals("string value", json);
    }
    
    private void _setValueInRepeaterEntryInRepositoryData(ModifiableRepositoryData repositoryData, Optional<String> value)
    {
        ModifiableRepositoryData subComposite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE)
                .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
                .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
                .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        
        value.ifPresent(v -> subComposite.setValue("singleStringData", v));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON(String)} implementation with a composite
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleCompositeToJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSingleCompositeToJSON", __DATA_NODE_TYPE);
        _testDataHolderSingleCompositeToJSON(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON(String)} implementation with a composite
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSingleCompositeToJSONWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSingleCompositeToJSON", __DATA_NODE_TYPE);
        _testDataHolderSingleCompositeToJSON(sourceData);
    }
    
    @SuppressWarnings("unchecked")
    private void _testDataHolderSingleCompositeToJSON(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent content = _createContent("content");
        _populateRepositoryData(repositoryData, content);
        
        // Convert the data holder to JSON
        Object json = dataHolder.dataToJSON("composite/repeater[1]/subcomposite");
        
        // Check the json result
        Map<String, Object> subcomposite = (Map<String, Object>) json;
        Assert.assertEquals(2, subcomposite.size());
        
        Assert.assertEquals("string value in the sub composite of the first repeater entry", subcomposite.get("singleStringData"));
        Assert.assertEquals(List.of("2018-09-11", "2015-12-06"), subcomposite.get("multipleDateData"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON(String)} implementation with a repeater
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSingleRepeaterToJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSingleRepeaterToJSON", __DATA_NODE_TYPE);
        _testDataHolderSingleRepeaterToJSON(sourceData);
        
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON(String)} implementation with a repeater
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSingleRepeaterToJSONWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSingleRepeaterToJSON", __DATA_NODE_TYPE);
        _testDataHolderSingleRepeaterToJSON(sourceData);
    }
    
    @SuppressWarnings("unchecked")
    private void _testDataHolderSingleRepeaterToJSON(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent content = _createContent("content");
        _populateRepositoryData(repositoryData, content);
        
        // Convert the data holder to JSON
        Object json = dataHolder.dataToJSON("composite/repeater");
        
        // Check the json result
        Map<String, Object> repeater = (Map<String, Object>) json;
        Assert.assertEquals(3, repeater.size());
        Assert.assertEquals(new I18nizableText("Repeater"), repeater.get("label"));
        Assert.assertEquals(2, repeater.get("entryCount"));
        List<Map<String, Object>> entries = (List<Map<String, Object>>) repeater.get("entries");
        Assert.assertEquals(2, entries.size());
        
        // repeater[1]
        Map<String, Object> entry1 = entries.get(0);
        Assert.assertEquals(2, entry1.size());
        Assert.assertEquals(1, entry1.get("position"));
        Map<String, Object> entry1Values = (Map<String, Object>) entry1.get("values");
        Assert.assertEquals(3, entry1Values.size());
        Assert.assertEquals(15L, entry1Values.get("longWithDefaultValue"));
        Assert.assertEquals(0.6D, entry1Values.get("singleDoubleData"));

        // repeater[1] > subcomposite
        Map<String, Object> entry1Subcomposite = (Map<String, Object>) entry1Values.get("subcomposite");
        Assert.assertEquals(2, entry1Subcomposite.size());
        Assert.assertEquals("string value in the sub composite of the first repeater entry", entry1Subcomposite.get("singleStringData"));
        Assert.assertEquals(List.of("2018-09-11", "2015-12-06"), entry1Subcomposite.get("multipleDateData"));
        
        // repeater[2]
        Map<String, Object> entry2 = entries.get(1);
        Assert.assertEquals(2, entry2.size());
        Assert.assertEquals(2, entry2.get("position"));
        Map<String, Object> entry2Values = (Map<String, Object>) entry2.get("values");
        Assert.assertEquals(3, entry2Values.size());
        Assert.assertEquals(95L, entry2Values.get("longWithDefaultValue"));
        Assert.assertTrue((boolean) entry2Values.get("singleBooleanData"));

        // repeater[2] > subcomposite
        Map<String, Object> entry2Subcomposite = (Map<String, Object>) entry2Values.get("subcomposite");
        Assert.assertEquals(1, entry2Subcomposite.size());
        Assert.assertEquals(List.of("1985-02-27", "1985-05-24"), entry2Subcomposite.get("multipleDateData"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON(String)} implementation with a repeater entry
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSingleRepeaterEntryToJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSingleRepeaterEntryToJSON", __DATA_NODE_TYPE);
        _testDataHolderSingleRepeaterEntryToJSON(sourceData);
        
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON(String)} implementation with a repeater entry
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSingleRepeaterEntryToJSONWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSingleRepeaterEntryToJSON", __DATA_NODE_TYPE);
        _testDataHolderSingleRepeaterEntryToJSON(sourceData);
    }
    
    @SuppressWarnings("unchecked")
    private void _testDataHolderSingleRepeaterEntryToJSON(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent content = _createContent("content");
        _populateRepositoryData(repositoryData, content);
        
        // Convert the data holder to JSON
        Object json = dataHolder.dataToJSON("composite/repeater[2]");
        
        // Check the json result
        Map<String, Object> entry2 = (Map<String, Object>) json;
        Assert.assertEquals(2, entry2.size());
        Assert.assertEquals(2, entry2.get("position"));
        Map<String, Object> entry2Values = (Map<String, Object>) entry2.get("values");
        Assert.assertEquals(3, entry2Values.size());
        Assert.assertEquals(95L, entry2Values.get("longWithDefaultValue"));
        Assert.assertTrue((boolean) entry2Values.get("singleBooleanData"));

        // repeater[2] > subcomposite
        Map<String, Object> entry2Subcomposite = (Map<String, Object>) entry2Values.get("subcomposite");
        Assert.assertEquals(1, entry2Subcomposite.size());
        Assert.assertEquals(List.of("1985-02-27", "1985-05-24"), entry2Subcomposite.get("multipleDateData"));
    }
    
    /**
     * Test {@link ModelAwareComposite#dataToJSON()} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeToJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testCompositeToJSON", __DATA_NODE_TYPE);
        _testCompositeToJSON(sourceData);
        
    }
    
    /**
     * Test {@link ModelAwareComposite#dataToJSON()} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeToJSONWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testCompositeToJSON", __DATA_NODE_TYPE);
        _testCompositeToJSON(sourceData);
    }
    
    @SuppressWarnings("unchecked")
    private void _testCompositeToJSON(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModifiableContent content = _createContent("content");
        _populateRepositoryData(repositoryData, content);
        
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModelAwareComposite composite = dataHolder.getComposite("composite");
     
        // Convert the composite to JSON
        Map<String, Object> json = composite.dataToJSON();
        
        // Check the json result
        Assert.assertEquals(2, json.size());
        Assert.assertEquals(54L, json.get("singleLongData"));
        
        // composite > repeater
        Map<String, Object> repeater = (Map<String, Object>) json.get("repeater");
        Assert.assertEquals(3, repeater.size());
        Assert.assertEquals(new I18nizableText("Repeater"), repeater.get("label"));
        Assert.assertEquals(2, repeater.get("entryCount"));
        List<Map<String, Object>> entries = (List<Map<String, Object>>) repeater.get("entries");
        Assert.assertEquals(2, entries.size());
        
        // composite > repeater[1]
        Map<String, Object> entry1 = entries.get(0);
        Assert.assertEquals(2, entry1.size());
        Assert.assertEquals(1, entry1.get("position"));
        Map<String, Object> entry1Values = (Map<String, Object>) entry1.get("values");
        Assert.assertEquals(3, entry1Values.size());
        Assert.assertEquals(15L, entry1Values.get("longWithDefaultValue"));
        Assert.assertEquals(0.6D, entry1Values.get("singleDoubleData"));

        // composite > repeater[1] > subcomposite
        Map<String, Object> entry1Subcomposite = (Map<String, Object>) entry1Values.get("subcomposite");
        Assert.assertEquals(2, entry1Subcomposite.size());
        Assert.assertEquals("string value in the sub composite of the first repeater entry", entry1Subcomposite.get("singleStringData"));
        Assert.assertEquals(List.of("2018-09-11", "2015-12-06"), entry1Subcomposite.get("multipleDateData"));
        
        // composite > repeater[2]
        Map<String, Object> entry2 = entries.get(1);
        Assert.assertEquals(2, entry2.size());
        Assert.assertEquals(2, entry2.get("position"));
        Map<String, Object> entry2Values = (Map<String, Object>) entry2.get("values");
        Assert.assertEquals(3, entry2Values.size());
        Assert.assertEquals(95L, entry2Values.get("longWithDefaultValue"));
        Assert.assertTrue((boolean) entry2Values.get("singleBooleanData"));

        // composite > repeater[2] > subcomposite
        Map<String, Object> entry2Subcomposite = (Map<String, Object>) entry2Values.get("subcomposite");
        Assert.assertEquals(1, entry2Subcomposite.size());
        Assert.assertEquals(List.of("1985-02-27", "1985-05-24"), entry2Subcomposite.get("multipleDateData"));
    }
    
    /**
     * Test {@link ModelAwareComposite#dataToJSON(String)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeSingleDataToJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testCompositeSingleDataToJSON", __DATA_NODE_TYPE);
        _testCompositeSingleDataToJSON(sourceData);
    }
    
    /**
     * Test {@link ModelAwareComposite#dataToJSON(String)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeSingleDataToJSONWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testCompositeSingleDataToJSON", __DATA_NODE_TYPE);
        _testCompositeSingleDataToJSON(sourceData);
    }
    
    private void _testCompositeSingleDataToJSON(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModifiableContent content = _createContent("content");
        _populateRepositoryData(repositoryData, content);

        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModelAwareComposite composite = dataHolder.getComposite("composite");
        
        // Convert the composite to JSON
        Object json = composite.dataToJSON("repeater[1]/subcomposite/singleStringData");
        
        // Check the json result
        Assert.assertEquals("string value in the sub composite of the first repeater entry", json);
    }
    
    /**
     * Test {@link ModelAwareRepeater#dataToJSON()} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepeaterToJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testRepeaterToJSON", __DATA_NODE_TYPE);
        _testRepeaterToJSON(sourceData);
    }
    
    /**
     * Test {@link ModelAwareRepeater#dataToJSON()} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepeaterToJSONWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testRepeaterToJSON", __DATA_NODE_TYPE);
        _testRepeaterToJSON(sourceData);
    }
    
    @SuppressWarnings("unchecked")
    private void _testRepeaterToJSON(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModifiableContent content = _createContent("content");
        _populateRepositoryData(repositoryData, content);
        
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModelAwareRepeater repeater = dataHolder.getRepeater("composite/repeater");
     
        // Convert the repeater to JSON
        Map<String, Object> json = repeater.dataToJSON();
        
        // Check the json result
        Assert.assertEquals(3, json.size());
        Assert.assertEquals(new I18nizableText("Repeater"), json.get("label"));
        Assert.assertEquals(2, json.get("entryCount"));
        List<Map<String, Object>> entries = (List<Map<String, Object>>) json.get("entries");
        Assert.assertEquals(2, entries.size());
        
        // repeater[1]
        Map<String, Object> entry1 = entries.get(0);
        Assert.assertEquals(2, entry1.size());
        Assert.assertEquals(1, entry1.get("position"));
        Map<String, Object> entry1Values = (Map<String, Object>) entry1.get("values");
        Assert.assertEquals(3, entry1Values.size());
        Assert.assertEquals(15L, entry1Values.get("longWithDefaultValue"));
        Assert.assertEquals(0.6D, entry1Values.get("singleDoubleData"));

        // repeater[1] > subcomposite
        Map<String, Object> entry1Subcomposite = (Map<String, Object>) entry1Values.get("subcomposite");
        Assert.assertEquals(2, entry1Subcomposite.size());
        Assert.assertEquals("string value in the sub composite of the first repeater entry", entry1Subcomposite.get("singleStringData"));
        Assert.assertEquals(List.of("2018-09-11", "2015-12-06"), entry1Subcomposite.get("multipleDateData"));
        
        // repeater[2]
        Map<String, Object> entry2 = entries.get(1);
        Assert.assertEquals(2, entry2.size());
        Assert.assertEquals(2, entry2.get("position"));
        Map<String, Object> entry2Values = (Map<String, Object>) entry2.get("values");
        Assert.assertEquals(3, entry2Values.size());
        Assert.assertEquals(95L, entry2Values.get("longWithDefaultValue"));
        Assert.assertTrue((boolean) entry2Values.get("singleBooleanData"));

        // repeater[2] > subcomposite
        Map<String, Object> entry2Subcomposite = (Map<String, Object>) entry2Values.get("subcomposite");
        Assert.assertEquals(1, entry2Subcomposite.size());
        Assert.assertEquals(List.of("1985-02-27", "1985-05-24"), entry2Subcomposite.get("multipleDateData"));
    }
    
    /**
     * Test {@link ModelAwareRepeater#dataToJSON(String)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepeaterSingleDataToJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testRepeaterSingleDataToJSON", __DATA_NODE_TYPE);
        _testRepeaterSingleDataToJSON(sourceData);
        
    }
    
    /**
     * Test {@link ModelAwareRepeater#dataToJSON(String)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepeaterSingleDataToJSONWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testRepeaterSingleDataToJSON", __DATA_NODE_TYPE);
        _testRepeaterSingleDataToJSON(sourceData);
    }
    
    @SuppressWarnings("unchecked")
    private void _testRepeaterSingleDataToJSON(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModifiableContent content = _createContent("content");
        _populateRepositoryData(repositoryData, content);

        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModelAwareRepeater repeater = dataHolder.getRepeater("composite/repeater");
        
        // Convert the repeater to JSON
        Map<String, Object> json = repeater.dataToJSON("subcomposite/singleStringData");
        
        // Check the json result
        Assert.assertEquals(3, json.size());
        Assert.assertEquals(new I18nizableText("Repeater"), json.get("label"));
        Assert.assertEquals(2, json.get("entryCount"));
        List<Map<String, Object>> entries = (List<Map<String, Object>>) json.get("entries");
        Assert.assertEquals(2, entries.size());
        
        // repeater[1]
        Map<String, Object> entry1 = entries.get(0);
        Assert.assertEquals(2, entry1.size());
        Assert.assertEquals(1, entry1.get("position"));
        Assert.assertEquals("string value in the sub composite of the first repeater entry", entry1.get("value"));
        
        // repeater[2]
        Map<String, Object> entry2 = entries.get(1);
        Assert.assertEquals(2, entry2.size());
        Assert.assertEquals(2, entry2.get("position"));
        Assert.assertTrue(entry2.containsKey("value"));
        Assert.assertNull(entry2.get("value"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON(ViewItemAccessor)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderViewToJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderViewToJSON", __DATA_NODE_TYPE);
        _testDataHolderViewToJSON(sourceData);
        
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON(ViewItemAccessor)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderViewToJSONWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderViewToJSON", __DATA_NODE_TYPE);
        _testDataHolderViewToJSON(sourceData);
    }
    
    @SuppressWarnings("unchecked")
    private void _testDataHolderViewToJSON(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent content = _createContent("content");
        _populateRepositoryData(repositoryData, content);
        View view = _createView();
        
        // Convert the data holder to JSON
        Set<String> externalizableData = _externalizableDataProviderEP.getExternalizableDataPaths(content);
        RepositoryDataContext context = RepositoryDataContext.newInstance()
                                                             .withExternalizableData(externalizableData);
        Map<String, Object> json = dataHolder.dataToJSON(view, context);
        
        // Check the json result
        Assert.assertEquals(6, json.size());
        
        Assert.assertEquals("value of the single string data", json.get("singleStringData"));
        Assert.assertTrue(json.containsKey("singleLongData"));
        Assert.assertNull(json.get("singleLongData"));
        Assert.assertFalse(json.containsKey("singleDateData"));
        Assert.assertEquals(Map.of("latitude", 85.487D, "longitude", 6.254D), json.get("singleGeocodeData"));
        
        List<Map<String, Object>> users = (List<Map<String, Object>>) json.get("multipleUserData");
        Map<String, Object> user1 = users.get(0);
        Assert.assertEquals("user1", user1.get("login"));
        Assert.assertEquals(POPULATION_TEST_ID, user1.get("populationId"));
        Assert.assertEquals("User One", user1.get("fullname"));
        Map<String, Object> user2 = users.get(1);
        Assert.assertEquals("user3", user2.get("login"));
        Assert.assertEquals(POPULATION_TEST_ID, user2.get("populationId"));
        Assert.assertEquals("User Three", user2.get("fullname"));        

        Map<String, Object> singleContentData = (Map<String, Object>) json.get("singleContentData");
        Assert.assertEquals(11, singleContentData.size());
        Assert.assertEquals(content.getId(), singleContentData.get("id"));
        Assert.assertEquals("content", singleContentData.get("title"));
        Assert.assertEquals("value for field 1", singleContentData.get("field1"));
        
        // composite
        Map<String, Object> composite = (Map<String, Object>) json.get("composite");
        Assert.assertEquals(2, composite.size());
        Assert.assertEquals(54L, composite.get("singleLongData"));
        
        // composite > repeater
        Map<String, Object> repeater = (Map<String, Object>) composite.get("repeater");
        Assert.assertEquals(3, repeater.size());
        Assert.assertEquals(new I18nizableText("Repeater"), repeater.get("label"));
        Assert.assertEquals(2, repeater.get("entryCount"));
        List<Map<String, Object>> entries = (List<Map<String, Object>>) repeater.get("entries");
        Assert.assertEquals(2, entries.size());
        
        // composite > repeater[1]
        Map<String, Object> entry1 = entries.get(0);
        Assert.assertEquals(2, entry1.size());
        Assert.assertEquals(1, entry1.get("position"));
        Map<String, Object> entry1Values = (Map<String, Object>) entry1.get("values");
        Assert.assertEquals(2, entry1Values.size());
        Assert.assertFalse(entry1Values.containsKey("singleBooleanValue"));
        Assert.assertEquals(0.6D, entry1Values.get("singleDoubleData"));

        // composite > repeater[1] > subcomposite
        Map<String, Object> entry1Subcomposite = (Map<String, Object>) entry1Values.get("subcomposite");
        Assert.assertEquals(2, entry1Subcomposite.size());
        Assert.assertEquals("string value in the sub composite of the first repeater entry", entry1Subcomposite.get("singleStringData"));
        Assert.assertEquals(List.of("2018-09-11", "2015-12-06"), entry1Subcomposite.get("multipleDateData"));
        
        // composite > repeater[2]
        Map<String, Object> entry2 = entries.get(1);
        Assert.assertEquals(2, entry2.size());
        Assert.assertEquals(2, entry2.get("position"));
        Map<String, Object> entry2Values = (Map<String, Object>) entry2.get("values");
        Assert.assertEquals(2, entry2Values.size());
        Assert.assertTrue((boolean) entry2Values.get("singleBooleanData"));
        Assert.assertFalse(entry2Values.containsKey("singleDoubleData"));

        // composite > repeater[2] > subcomposite
        Map<String, Object> entry2Subcomposite = (Map<String, Object>) entry2Values.get("subcomposite");
        Assert.assertEquals(1, entry2Subcomposite.size());
        Assert.assertEquals(List.of("1985-02-27", "1985-05-24"), entry2Subcomposite.get("multipleDateData"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSONForEdition(ViewItemAccessor, DataContext)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderViewToJSONForEditionWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderViewToJSONForEdition", __DATA_NODE_TYPE);
        _testDataHolderViewToJSONForEdition(sourceData);
        
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSONForEdition(ViewItemAccessor, DataContext)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderViewToJSONForEditionWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderViewToJSONForEdition", __DATA_NODE_TYPE);
        _testDataHolderViewToJSONForEdition(sourceData);
    }
    
    @SuppressWarnings("unchecked")
    private void _testDataHolderViewToJSONForEdition(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent content = _createContent("content");
        _populateRepositoryData(repositoryData, content);
        View view = _createView();
        
        // Convert the data holder to JSON
        Set<String> externalizableData = _externalizableDataProviderEP.getExternalizableDataPaths(content);
        RepositoryDataContext context = RepositoryDataContext.newInstance()
                                                             .withExternalizableData(externalizableData);
        Map<String, Object> json = dataHolder.dataToJSONForEdition(view, context);
        
        // Check the json result
        Assert.assertEquals(6, json.size());
        
        Map<String, Object> singleStringData = (Map<String, Object>) json.get("singleStringData");
        Assert.assertEquals(3, singleStringData.size());
        Assert.assertEquals("value of the single string data", singleStringData.get("local"));
        Assert.assertEquals("external value of the single string data", singleStringData.get("external"));
        Assert.assertEquals(ExternalizableDataStatus.LOCAL.name().toLowerCase(), singleStringData.get("status"));

        Assert.assertTrue(json.containsKey("singleLongData"));
        Assert.assertNull(json.get("singleLongData"));
        Assert.assertFalse(json.containsKey("singleDateData"));
        Assert.assertEquals(Map.of("latitude", 85.487D, "longitude", 6.254D), json.get("singleGeocodeData"));
        
        List<Map<String, Object>> users = (List<Map<String, Object>>) json.get("multipleUserData");
        Map<String, Object> user1 = users.get(0);
        Assert.assertEquals("user1", user1.get("login"));
        Assert.assertEquals(POPULATION_TEST_ID, user1.get("populationId"));
        Assert.assertEquals("User One", user1.get("fullname"));
        Map<String, Object> user2 = users.get(1);
        Assert.assertEquals("user3", user2.get("login"));
        Assert.assertEquals(POPULATION_TEST_ID, user2.get("populationId"));
        Assert.assertEquals("User Three", user2.get("fullname"));        

        Assert.assertEquals(content.getId(), json.get("singleContentData"));
        
        // composite
        Map<String, Object> composite = (Map<String, Object>) json.get("composite");
        Assert.assertEquals(2, composite.size());
        Assert.assertEquals(54L, composite.get("singleLongData"));
        
        // composite > repeater
        Map<String, Object> repeater = (Map<String, Object>) composite.get("repeater");
        Assert.assertEquals(3, repeater.size());
        Assert.assertEquals(new I18nizableText("Repeater"), repeater.get("label"));
        Assert.assertEquals(2, repeater.get("entryCount"));
        List<Map<String, Object>> entries = (List<Map<String, Object>>) repeater.get("entries");
        Assert.assertEquals(2, entries.size());
        
        // composite > repeater[1]
        Map<String, Object> entry1 = entries.get(0);
        Assert.assertEquals(2, entry1.size());
        Assert.assertEquals(1, entry1.get("position"));
        Map<String, Object> entry1Values = (Map<String, Object>) entry1.get("values");
        Assert.assertEquals(2, entry1Values.size());
        Assert.assertFalse(entry1Values.containsKey("singleBooleanValue"));
        Assert.assertEquals(0.6D, entry1Values.get("singleDoubleData"));

        // composite > repeater[1] > subcomposite
        Map<String, Object> entry1Subcomposite = (Map<String, Object>) entry1Values.get("subcomposite");
        Assert.assertEquals(2, entry1Subcomposite.size());
        
        Map<String, Object> entry1SubcompositeSingleStringData = (Map<String, Object>) entry1Subcomposite.get("singleStringData");
        Assert.assertEquals(3, entry1SubcompositeSingleStringData.size());
        Assert.assertEquals("string value in the sub composite of the first repeater entry", entry1SubcompositeSingleStringData.get("local"));
        Assert.assertEquals("external string value in the sub composite of the first repeater entry", entry1SubcompositeSingleStringData.get("external"));
        Assert.assertEquals(ExternalizableDataStatus.LOCAL.name().toLowerCase(), entry1SubcompositeSingleStringData.get("status"));
        Assert.assertEquals(List.of("2018-09-11", "2015-12-06"), entry1Subcomposite.get("multipleDateData"));
        
        // composite > repeater[2]
        Map<String, Object> entry2 = entries.get(1);
        Assert.assertEquals(2, entry2.size());
        Assert.assertEquals(2, entry2.get("position"));
        Map<String, Object> entry2Values = (Map<String, Object>) entry2.get("values");
        Assert.assertEquals(2, entry2Values.size());
        Assert.assertTrue((boolean) entry2Values.get("singleBooleanData"));
        Assert.assertFalse(entry2Values.containsKey("singleDoubleData"));

        // composite > repeater[2] > subcomposite
        Map<String, Object> entry2Subcomposite = (Map<String, Object>) entry2Values.get("subcomposite");
        Assert.assertEquals(1, entry2Subcomposite.size());
        Assert.assertEquals(List.of("1985-02-27", "1985-05-24"), entry2Subcomposite.get("multipleDateData"));
    }
    
    private View _createView()
    {
        View viewToSAX = View.of(_model, "singleStringData",
                "singleLongData",
                "singleDateData",
                "singleGeocodeData",
                "multipleUserData",
                "composite/singleLongData",
                "composite/repeater/singleBooleanData",
                "composite/repeater/singleDoubleData",
                "composite/repeater/subcomposite",
                "unvaluedComposite",
                "unvaluedRepeater");
        
        // Create a view item with nested attribute field1
        ViewElementAccessor singleContentDataViewItem = new ViewElementAccessor();
        singleContentDataViewItem.setDefinition((ElementDefinition) _model.getModelItem("singleContentData"));
        
        ViewElement field1ViewElement = new ViewElement();
        field1ViewElement.setDefinition((ElementDefinition) _model.getModelItem("singleContentData/field1"));
        singleContentDataViewItem.addViewItem(field1ViewElement);
        
        // Insert the content view item in the view
        viewToSAX.insertViewItem(singleContentDataViewItem, 5);
        
        return viewToSAX;
    }
    
    private ModifiableContent _createContent(String title)
    {
        ModifiableContent content = _createContent(title, "org.ametys.cms.contenttype.Test");
        content.setValue("field1", "value for field 1");
        return content;
    }
    
    private ModifiableContent _createContent(String title, String... contentTypeIds)
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        
        ModifiableContent content = root.createChild(title, "ametys:defaultContent");

        content.setTypes(contentTypeIds);
        content.setTitle(title, Locale.FRANCE);
        content.setLanguage("fr");
        content.setCreationDate(ZonedDateTime.now());
        content.setCreator(new UserIdentity("john", "anyPopulation"));
        content.setLastModified(ZonedDateTime.now());
        content.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        
        return content;
    }
    
    private void _populateRepositoryData(ModifiableRepositoryData repositoryData, Content singleContentValue) throws Exception
    {
        repositoryData.setValue("singleStringData", "value of the single string data");
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "external value of the single string data");
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.STATUS_SUFFIX, "local");
        
        repositoryData.setValue("singleLongData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        repositoryData.setValues("multipleLongData", new Long[] {32L, 784L});
        
        _geocodeType.write(repositoryData, "singleGeocodeData", new Geocode(85.487D, 6.254D));
        _userType.write(repositoryData, "multipleUserData", new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY});
        repositoryData.setValue("singleContentData", singleContentValue.getId());
        
        ZonedDateTime[] dateTimeArray = new ZonedDateTime[]{ZonedDateTime.of(2018, 5, 5, 7, 25, 13, 596_000_000, ZoneId.of("UTC+05:00")), ZonedDateTime.of(2015, 12, 06, 15, 18, 0, 0, ZoneId.of("UTC-03:00"))};
        repositoryData.setValues("multipleDatetimeData", Stream.of(dateTimeArray).map(DateUtils::asCalendar).toArray(Calendar[]::new));
        
        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        composite.setValue("singleLongData", 54L);
        ModifiableRepositoryData repeater = composite.addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE);

        ModifiableRepositoryData entry1 = repeater.addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE);
        entry1.setValue("longWithDefaultValue", 15L);
        entry1.setValue("singleDoubleData", 0.6D);
        ModifiableRepositoryData entry1SubComposite = entry1.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        entry1SubComposite.setValue("singleStringData", "string value in the sub composite of the first repeater entry");
        entry1SubComposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "external string value in the sub composite of the first repeater entry");
        entry1SubComposite.setValue("singleStringData" + ModelAwareDataHolder.STATUS_SUFFIX, "local");
        entry1SubComposite.setValues("multipleDateData", new Calendar[]{DateUtils.asCalendar(LocalDate.of(2018, Month.SEPTEMBER, 11)), DateUtils.asCalendar(LocalDate.of(2015, Month.DECEMBER, 06))});
        
        ModifiableRepositoryData entry2 = repeater.addRepositoryData("2", RepositoryConstants.COMPOSITE_NODETYPE);
        entry2.setValue("longWithDefaultValue", 95L);
        entry2.setValue("singleBooleanData", true);
        ModifiableRepositoryData entry2SubComposite = entry2.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        entry2SubComposite.setValues("multipleDateData", new Calendar[]{DateUtils.asCalendar(LocalDate.of(1985, Month.FEBRUARY, 27)), DateUtils.asCalendar(LocalDate.of(1985, Month.MAY, 24))});
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON(String)} implementation with a non existing content
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testDataHolderInexistingContentToJSONWithJCRRepositoryData()
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderInexistingContentToJSON", __DATA_NODE_TYPE);
        _testDataHolderInexistingContentToJSON(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToJSON(String)} implementation with a non existing content
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testDataHolderInexistingContentToJSONWithMemoryRepositoryData()
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderInexistingContentToJSON", __DATA_NODE_TYPE);
        _testDataHolderInexistingContentToJSON(sourceData);
    }
    
    private void _testDataHolderInexistingContentToJSON(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        // Add a reference to an inexisting content in the repository
        repositoryData.setValue("singleContentData", "defaultContent://inexisting-content-id");
        
        // Convert the data holder to JSON
        Object json = dataHolder.dataToJSON("singleContentData");
        
        // Check the json result
        Assert.assertNull(json);
    }
}
