/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.time.LocalDate;
import java.time.Month;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Calendar;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.stream.Stream;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.holder.impl.DefaultModelAwareDataHolder;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.external.ExternalizableDataProvider.ExternalizableDataStatus;
import org.ametys.plugins.repository.data.external.ExternalizableDataProviderExtensionPoint;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.values.SynchronizableValue;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.plugins.repository.data.type.RepositoryModelItemType;
import org.ametys.plugins.repository.model.RepositoryDataContext;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelAwareDataHolder} toMap methods implementation test case
 */
@NeedFeatures({
    "test-data/test-model",
    "test-data/test-externalizable-data-provider",
    "cms/factories",
    "test/content-type",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModelAwareDataHolderToMapTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __DATA_NODE_TYPE = "ametys:unstructured";
    
    private static Model _model;
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    private static ExternalizableDataProviderExtensionPoint _externalizableDataProviderEP;
    
    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;
    
    private static RepositoryElementType<Geocode> _geocodeType;
    private static RepositoryElementType<UserIdentity> _userType;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
        _externalizableDataProviderEP = (ExternalizableDataProviderExtensionPoint) manager.lookup(ExternalizableDataProviderExtensionPoint.ROLE);
    }
    
    @Override
    @SuppressWarnings("unchecked")
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");
        
        _geocodeType = (RepositoryElementType<Geocode>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID);
        _userType = (RepositoryElementType<UserIdentity>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
    }
    
    private View _createView()
    {
        View view = View.of(_model, "singleStringData",
                "singleDateData",
                "singleGeocodeData",
                "multipleUserData",
                "singleContentData",
                "composite/singleLongData",
                "composite/repeater/singleBooleanData",
                "composite/repeater/singleDoubleData",
                "composite/repeater/subcomposite/singleStringData",
                "unvaluedComposite",
                "unvaluedRepeater");
        
        return view;
    }
    
    private ModifiableContent _createContent(String title)
    {
        ModifiableContent content = _createContent(title, "org.ametys.cms.contenttype.Test");
        content.setValue("field1", "value for field 1");
        return content;
    }
    
    private ModifiableContent _createContent(String title, String... contentTypeIds)
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        
        ModifiableContent content = root.createChild(title, "ametys:defaultContent");

        content.setTypes(contentTypeIds);
        content.setTitle(title, Locale.FRANCE);
        content.setLanguage("fr");
        content.setCreationDate(ZonedDateTime.now());
        content.setCreator(new UserIdentity("john", "anyPopulation"));
        content.setLastModified(ZonedDateTime.now());
        content.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        
        return content;
    }
    
    private void _populateRepositoryData(ModifiableRepositoryData repositoryData, Content singleContentValue) throws Exception
    {
        repositoryData.setValue("singleStringData", "value of the single string data");
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "external value of the single string data");
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.STATUS_SUFFIX, "local");
        
        repositoryData.setValue("singleLongData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        repositoryData.setValues("multipleLongData", new Long[] {32L, 784L});
        
        _geocodeType.write(repositoryData, "singleGeocodeData", new Geocode(85.487D, 6.254D));
        _userType.write(repositoryData, "multipleUserData", new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY});
        repositoryData.setValue("singleContentData", singleContentValue.getId());
        
        ZonedDateTime[] dateTimeArray = new ZonedDateTime[]{ZonedDateTime.of(2018, 5, 5, 7, 25, 13, 596_000_000, ZoneId.of("UTC+05:00")), ZonedDateTime.of(2015, 12, 06, 15, 18, 0, 0, ZoneId.of("UTC-03:00"))};
        repositoryData.setValues("multipleDatetimeData", Stream.of(dateTimeArray).map(DateUtils::asCalendar).toArray(Calendar[]::new));
        
        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        composite.setValue("singleLongData", 54L);
        ModifiableRepositoryData repeater = composite.addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE);

        ModifiableRepositoryData entry1 = repeater.addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE);
        entry1.setValue("longWithDefaultValue", 15L);
        entry1.setValue("singleDoubleData", 0.6D);
        ModifiableRepositoryData entry1SubComposite = entry1.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        entry1SubComposite.setValue("singleStringData", "string value in the sub composite of the first repeater entry");
        entry1SubComposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "external string value in the sub composite of the first repeater entry");
        entry1SubComposite.setValue("singleStringData" + ModelAwareDataHolder.STATUS_SUFFIX, "local");
        entry1SubComposite.setValues("multipleDateData", new Calendar[]{DateUtils.asCalendar(LocalDate.of(2018, Month.SEPTEMBER, 11)), DateUtils.asCalendar(LocalDate.of(2015, Month.DECEMBER, 06))});
        
        ModifiableRepositoryData entry2 = repeater.addRepositoryData("2", RepositoryConstants.COMPOSITE_NODETYPE);
        entry2.setValue("longWithDefaultValue", 95L);
        entry2.setValue("singleBooleanData", true);
        ModifiableRepositoryData entry2SubComposite = entry2.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        entry2SubComposite.setValues("multipleDateData", new Calendar[]{DateUtils.asCalendar(LocalDate.of(1985, Month.FEBRUARY, 27)), DateUtils.asCalendar(LocalDate.of(1985, Month.MAY, 24))});
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToMap()} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderToMapWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testMap", __DATA_NODE_TYPE);
        _testMap(repositoryData, true);
        _testMap(repositoryData, false);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToMap()} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderToMapWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testMap", __DATA_NODE_TYPE);
        _testMap(repositoryData, true);
        _testMap(repositoryData, false);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToMap(ViewItemAccessor)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderToMapWithViewWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testMapWithView", __DATA_NODE_TYPE);
        _testMapWithView(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToMap(ViewItemAccessor)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderToMapWithViewWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testMapWithView", __DATA_NODE_TYPE);
        _testMapWithView(repositoryData);
    }
    
    @SuppressWarnings("unchecked")
    private void _testMap(ModifiableRepositoryData repositoryData, boolean includeEmptyValues) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent content = _createContent("contentForMap");
        _populateRepositoryData(repositoryData, content);
        
        Set<String> externalizableData = _externalizableDataProviderEP.getExternalizableDataPaths(content);
        RepositoryDataContext context = RepositoryDataContext.newInstance();
        context.withExternalizableData(externalizableData).withEmptyValues(includeEmptyValues);
        Map<String, Object> dataMap = dataHolder.dataToMap(context);
        
        SynchronizableValue syncString = (SynchronizableValue) dataMap.get("singleStringData");
        Assert.assertEquals("value of the single string data", syncString.getLocalValue());
        Assert.assertEquals("external value of the single string data", syncString.getExternalValue());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, syncString.getExternalizableStatus());
        
        if (includeEmptyValues)
        {
            Assert.assertTrue(dataMap.containsKey("singleLongData"));
            Assert.assertNull(dataMap.get("singleLongData"));
        }
        else
        {
            Assert.assertFalse(dataMap.containsKey("singleLongData"));
        }
        
        Assert.assertArrayEquals(new Long[] {32L, 784L}, (Object[]) dataMap.get("multipleLongData"));
        Assert.assertEquals(null, dataMap.get("singleDateData"));
        Assert.assertEquals(new Geocode(85.487, 6.254), dataMap.get("singleGeocodeData"));
        
        UserIdentity[] users = (UserIdentity[]) dataMap.get("multipleUserData");
        Assert.assertEquals(2, users.length);
        Assert.assertEquals("user1", users[0].getLogin());
        Assert.assertEquals(POPULATION_TEST_ID, users[0].getPopulationId());
        Assert.assertEquals("user3", users[1].getLogin());
        Assert.assertEquals(POPULATION_TEST_ID, users[1].getPopulationId());
        
        ContentValue contentValue = (ContentValue) dataMap.get("singleContentData");
        Assert.assertEquals(content.getId(), contentValue.getContentId());
        
        Map<String, Object> composite = (Map<String, Object>) dataMap.get("composite");
        Assert.assertEquals(54L, composite.get("singleLongData"));
        
        List<Map<String, Object>> repeater = (List<Map<String, Object>>) composite.get("repeater");
        Assert.assertEquals(2, repeater.size());
        Map<String, Object> entry1 = repeater.get(0);
        Assert.assertEquals(15L, entry1.get("longWithDefaultValue"));
        Assert.assertNull(entry1.get("singleBooleanData"));
        Assert.assertEquals(0.6, entry1.get("singleDoubleData"));
        
        Map<String, Object> subcomposite = (Map<String, Object>) entry1.get("subcomposite");
        
        SynchronizableValue syncSubCompositeString = (SynchronizableValue) subcomposite.get("singleStringData");
        Assert.assertEquals("string value in the sub composite of the first repeater entry", syncSubCompositeString.getLocalValue());
        Assert.assertEquals("external string value in the sub composite of the first repeater entry", syncSubCompositeString.getExternalValue());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, syncSubCompositeString.getExternalizableStatus());
        
        LocalDate[] dates = (LocalDate[]) subcomposite.get("multipleDateData");
        Assert.assertEquals(2, dates.length);
        Assert.assertEquals(LocalDate.of(2018, Month.SEPTEMBER, 11), dates[0]);
        Assert.assertEquals(LocalDate.of(2015, Month.DECEMBER, 06), dates[1]);
        
        Map<String, Object> entry2 = repeater.get(1);
        Assert.assertEquals(95L, entry2.get("longWithDefaultValue"));
        Assert.assertTrue((boolean) entry2.get("singleBooleanData"));
        Assert.assertNull(entry2.get("singleDoubleData"));
        
        subcomposite = (Map<String, Object>) entry2.get("subcomposite");
        
        dates = (LocalDate[]) subcomposite.get("multipleDateData");
        Assert.assertEquals(2, dates.length);
        Assert.assertEquals(LocalDate.of(1985, Month.FEBRUARY, 27), dates[0]);
        Assert.assertEquals(LocalDate.of(1985, Month.MAY, 24), dates[1]);
    }
    
    @SuppressWarnings("unchecked")
    private void _testMapWithView(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent content = _createContent("contentForMapWithView");
        _populateRepositoryData(repositoryData, content);
        
        View view = _createView();
        Map<String, Object> dataMap = dataHolder.dataToMap(view);
        
        Assert.assertEquals("value of the single string data", dataMap.get("singleStringData"));
        
        Assert.assertNull(dataMap.get("multipleLongData"));
        Assert.assertNull(dataMap.get("singleDateData"));
        Assert.assertEquals(new Geocode(85.487, 6.254), dataMap.get("singleGeocodeData"));
        
        UserIdentity[] users = (UserIdentity[]) dataMap.get("multipleUserData");
        Assert.assertEquals(2, users.length);
        Assert.assertEquals("user1", users[0].getLogin());
        Assert.assertEquals(POPULATION_TEST_ID, users[0].getPopulationId());
        Assert.assertEquals("user3", users[1].getLogin());
        Assert.assertEquals(POPULATION_TEST_ID, users[1].getPopulationId());
        
        ContentValue contentValue = (ContentValue) dataMap.get("singleContentData");
        Assert.assertEquals(content.getId(), contentValue.getContentId());
        
        Map<String, Object> composite = (Map<String, Object>) dataMap.get("composite");
        Assert.assertEquals(54L, composite.get("singleLongData"));
        
        List<Map<String, Object>> repeater = (List<Map<String, Object>>) composite.get("repeater");
        Assert.assertEquals(2, repeater.size());
        Map<String, Object> entry1 = repeater.get(0);
        Assert.assertNull(entry1.get("longWithDefaultValue"));
        Assert.assertNull(entry1.get("singleBooleanData"));
        Assert.assertEquals(0.6, entry1.get("singleDoubleData"));
        
        Map<String, Object> subcomposite = (Map<String, Object>) entry1.get("subcomposite");
        Assert.assertEquals(1, subcomposite.size());
        Assert.assertEquals("string value in the sub composite of the first repeater entry", subcomposite.get("singleStringData"));
        
        Map<String, Object> entry2 = repeater.get(1);
        Assert.assertNull(entry2.get("longWithDefaultValue"));
        Assert.assertTrue((boolean) entry2.get("singleBooleanData"));
        Assert.assertNull(entry2.get("singleDoubleData"));
        
        subcomposite = (Map<String, Object>) entry2.get("subcomposite");
        Assert.assertTrue(subcomposite.isEmpty());
    }
}
