/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.time.LocalDate;
import java.time.Month;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Arrays;
import java.util.Calendar;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.TransformerHandler;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.xpath.XPathAPI;
import org.junit.Assert;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.ContentHandler;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Binary;
import org.ametys.cms.data.ExplorerFile;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.RichText;
import org.ametys.cms.data.holder.impl.DefaultModelAwareDataHolder;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.cms.utils.ResourceElementTypeHelperForTests;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.core.util.JSONUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.external.ExternalizableDataProviderExtensionPoint;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModelAwareComposite;
import org.ametys.plugins.repository.data.holder.group.ModelAwareRepeater;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.plugins.repository.data.type.RepositoryModelItemType;
import org.ametys.plugins.repository.model.RepositoryDataContext;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElement;
import org.ametys.runtime.model.ViewElementAccessor;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelAwareDataHolder} toSAX methods implementation test case
 */
@NeedFeatures({
    "test-data/test-model",
    "test-data/test-externalizable-data-provider",
    "cms/factories",
    "test/content-type",
    
    "core/util.jsonutils",
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content",
    "cms/org.ametys.cms.data.holder.helpers"
})
public class ModelAwareDataHolderToSAXTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __DATA_NODE_TYPE = "ametys:unstructured";
    
    private static JSONUtils _jsonUtils;
    private static Model _model;
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    private static ExternalizableDataProviderExtensionPoint _externalizableDataProviderEP;
    
    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;
    
    private static RepositoryElementType<Geocode> _geocodeType;
    private static RepositoryElementType<UserIdentity> _userType;
    private static RepositoryElementType<Binary> _binaryType;
    private static RepositoryElementType<ExplorerFile> _fileType;
    private static RepositoryElementType<RichText> _richTextType;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _jsonUtils = (JSONUtils) manager.lookup(JSONUtils.ROLE);
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
        _externalizableDataProviderEP = (ExternalizableDataProviderExtensionPoint) manager.lookup(ExternalizableDataProviderExtensionPoint.ROLE);
    }
    
    @Override
    @SuppressWarnings("unchecked")
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");
        
        _geocodeType = (RepositoryElementType<Geocode>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID);
        _userType = (RepositoryElementType<UserIdentity>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _binaryType = (RepositoryElementType<Binary>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _fileType = (RepositoryElementType<ExplorerFile>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.FILE_ELEMENT_TYPE_ID);
        _richTextType = (RepositoryElementType<RichText>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSax", __DATA_NODE_TYPE);
        _testDataHolderSax(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSax", __DATA_NODE_TYPE);
        _testDataHolderSax(sourceData);
    }
    
    private void _testDataHolderSax(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent contentForSAX = _createContent("contentForSAX");
        _populateRepositoryData(repositoryData, contentForSAX);
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        Set<String> externalizableData = _externalizableDataProviderEP.getExternalizableDataPaths(dataHolder);
        RepositoryDataContext context = RepositoryDataContext.newInstance()
                                                             .withExternalizableData(externalizableData);
        dataHolder.dataToSAX(handler, context);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/singleStringData)", 1);
        _assertXpathEquals(document, "/result/singleStringData/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/singleStringData/@multiple", "false");
        _assertXpathEquals(document, "/result/singleStringData/@dataPath", "singleStringData");
        _assertXpathEquals(document, "/result/singleStringData/@fullDataPath", "singleStringData");
        _assertXpathEquals(document, "/result/singleStringData/@modelPath", "singleStringData");
        _assertXpathEquals(document, "/result/singleStringData/@fullModelPath", "singleStringData");
        _assertXpathEquals(document, "/result/singleStringData", "value of the single string data");
        
        _assertXpathEquals(document, "count(/result/multipleLongData)", 2);
        _assertXpathEquals(document, "/result/multipleLongData[1]/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/multipleLongData[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/multipleLongData[1]/@dataPath", "multipleLongData");
        _assertXpathEquals(document, "/result/multipleLongData[1]/@fullDataPath", "multipleLongData");
        _assertXpathEquals(document, "/result/multipleLongData[1]/@modelPath", "multipleLongData");
        _assertXpathEquals(document, "/result/multipleLongData[1]/@fullModelPath", "multipleLongData");
        _assertXpathEquals(document, "/result/multipleLongData[1]", 32);
        _assertXpathEquals(document, "/result/multipleLongData[2]/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/multipleLongData[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/multipleLongData[2]/@dataPath", "multipleLongData");
        _assertXpathEquals(document, "/result/multipleLongData[2]/@fullDataPath", "multipleLongData");
        _assertXpathEquals(document, "/result/multipleLongData[2]/@modelPath", "multipleLongData");
        _assertXpathEquals(document, "/result/multipleLongData[2]/@fullModelPath", "multipleLongData");
        _assertXpathEquals(document, "/result/multipleLongData[2]", 784);

        _assertXpathEquals(document, "count(/result/singleGeocodeData)", 1);
        _assertXpathEquals(document, "/result/singleGeocodeData/@typeId", ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/singleGeocodeData/@multiple", "false");
        _assertXpathEquals(document, "/result/singleGeocodeData/@latitude", 85.487);
        _assertXpathEquals(document, "/result/singleGeocodeData/@longitude", 6.254);
        _assertXpathEquals(document, "/result/singleGeocodeData/@dataPath", "singleGeocodeData");
        _assertXpathEquals(document, "/result/singleGeocodeData/@fullDataPath", "singleGeocodeData");
        _assertXpathEquals(document, "/result/singleGeocodeData/@modelPath", "singleGeocodeData");
        _assertXpathEquals(document, "/result/singleGeocodeData/@fullModelPath", "singleGeocodeData");
        
        _assertXpathEquals(document, "count(/result/multipleUserData)", 2);
        _assertXpathEquals(document, "/result/multipleUserData[1]/@typeId", ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/multipleUserData[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/multipleUserData[1]/@login", "user1");
        _assertXpathEquals(document, "/result/multipleUserData[1]/@populationId", POPULATION_TEST_ID);
        _assertXpathEquals(document, "/result/multipleUserData[1]/@dataPath", "multipleUserData");
        _assertXpathEquals(document, "/result/multipleUserData[1]/@fullDataPath", "multipleUserData");
        _assertXpathEquals(document, "/result/multipleUserData[1]/@modelPath", "multipleUserData");
        _assertXpathEquals(document, "/result/multipleUserData[1]/@fullModelPath", "multipleUserData");
        _assertXpathEquals(document, "/result/multipleUserData[1]", "User One");
        _assertXpathEquals(document, "/result/multipleUserData[2]/@typeId", ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/multipleUserData[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/multipleUserData[2]/@login", "user3");
        _assertXpathEquals(document, "/result/multipleUserData[2]/@populationId", POPULATION_TEST_ID);
        _assertXpathEquals(document, "/result/multipleUserData[2]/@dataPath", "multipleUserData");
        _assertXpathEquals(document, "/result/multipleUserData[2]/@fullDataPath", "multipleUserData");
        _assertXpathEquals(document, "/result/multipleUserData[2]/@modelPath", "multipleUserData");
        _assertXpathEquals(document, "/result/multipleUserData[2]/@fullModelPath", "multipleUserData");
        _assertXpathEquals(document, "/result/multipleUserData[2]", "User Three");
        
        _assertXpathEquals(document, "count(/result/singleContentData)", 1);
        _assertXpathEquals(document, "/result/singleContentData/@typeId", ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/singleContentData/@multiple", "false");
        _assertXpathEquals(document, "/result/singleContentData/@id", contentForSAX.getId());
        _assertXpathEquals(document, "/result/singleContentData/@title", "contentForSAX");
        _assertXpathEquals(document, "/result/singleContentData/@dataPath", "singleContentData");
        _assertXpathEquals(document, "/result/singleContentData/@fullDataPath", "singleContentData");
        _assertXpathEquals(document, "/result/singleContentData/@modelPath", "singleContentData");
        _assertXpathEquals(document, "/result/singleContentData/@fullModelPath", "singleContentData");
        
        _assertXpathEquals(document, "count(/result/multipleDatetimeData)", 2);
        _assertXpathEquals(document, "/result/multipleDatetimeData[1]/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.DATETIME_TYPE_ID);
        _assertXpathEquals(document, "/result/multipleDatetimeData[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/multipleDatetimeData[1]/@dataPath", "multipleDatetimeData");
        _assertXpathEquals(document, "/result/multipleDatetimeData[1]/@fullDataPath", "multipleDatetimeData");
        _assertXpathEquals(document, "/result/multipleDatetimeData[1]/@modelPath", "multipleDatetimeData");
        _assertXpathEquals(document, "/result/multipleDatetimeData[1]/@fullModelPath", "multipleDatetimeData");
        _assertXpathEquals(document, "/result/multipleDatetimeData[1]", "2018-05-05T02:25:13.596Z");
        _assertXpathEquals(document, "/result/multipleDatetimeData[2]/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.DATETIME_TYPE_ID);
        _assertXpathEquals(document, "/result/multipleDatetimeData[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/multipleDatetimeData[2]/@dataPath", "multipleDatetimeData");
        _assertXpathEquals(document, "/result/multipleDatetimeData[2]/@fullDataPath", "multipleDatetimeData");
        _assertXpathEquals(document, "/result/multipleDatetimeData[2]/@modelPath", "multipleDatetimeData");
        _assertXpathEquals(document, "/result/multipleDatetimeData[2]/@fullModelPath", "multipleDatetimeData");
        _assertXpathEquals(document, "/result/multipleDatetimeData[2]", "2015-12-06T18:18:00.000Z");
        
        _assertXpathEquals(document, "count(/result/composite)", 1);
        _assertXpathEquals(document, "count(/result/composite/singleLongData)", 1);
        _assertXpathEquals(document, "/result/composite/singleLongData/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/singleLongData/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/singleLongData/@dataPath", "composite/singleLongData");
        _assertXpathEquals(document, "/result/composite/singleLongData/@fullDataPath", "composite/singleLongData");
        _assertXpathEquals(document, "/result/composite/singleLongData/@modelPath", "composite/singleLongData");
        _assertXpathEquals(document, "/result/composite/singleLongData/@fullModelPath", "composite/singleLongData");
        _assertXpathEquals(document, "/result/composite/singleLongData", 54);
        
        _assertXpathEquals(document, "count(/result/composite/repeater)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/@entryCount", 2);
        _assertXpathEquals(document, "/result/composite/repeater/@dataPath", "composite/repeater");
        _assertXpathEquals(document, "/result/composite/repeater/@fullDataPath", "composite/repeater");
        _assertXpathEquals(document, "/result/composite/repeater/@modelPath", "composite/repeater");
        _assertXpathEquals(document, "/result/composite/repeater/@fullModelPath", "composite/repeater");
        
        _assertXpathEquals(document, "count(/result/composite/repeater/entry)", 2);
        
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/@dataPath", "composite/repeater[1]");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/@fullDataPath", "composite/repeater[1]");
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/@modelPath)", 0);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/@fullModelPath)", 0);

        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/longWithDefaultValue)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/longWithDefaultValue/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/longWithDefaultValue/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/longWithDefaultValue/@dataPath", "composite/repeater[1]/longWithDefaultValue");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/longWithDefaultValue/@fullDataPath", "composite/repeater[1]/longWithDefaultValue");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/longWithDefaultValue/@modelPath", "composite/repeater/longWithDefaultValue");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/longWithDefaultValue/@fullModelPath", "composite/repeater/longWithDefaultValue");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/longWithDefaultValue", 15);
        
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/singleDoubleData)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/singleDoubleData/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.DOUBLE_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/singleDoubleData/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/singleDoubleData/@dataPath", "composite/repeater[1]/singleDoubleData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/singleDoubleData/@fullDataPath", "composite/repeater[1]/singleDoubleData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/singleDoubleData/@modelPath", "composite/repeater/singleDoubleData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/singleDoubleData/@fullModelPath", "composite/repeater/singleDoubleData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/singleDoubleData", 0.6);
        
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/subcomposite)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/@dataPath", "composite/repeater[1]/subcomposite");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/@fullDataPath", "composite/repeater[1]/subcomposite");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/@modelPath", "composite/repeater/subcomposite");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/@fullModelPath", "composite/repeater/subcomposite");
        
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/subcomposite/singleStringData)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/singleStringData/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/singleStringData/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/singleStringData/@dataPath", "composite/repeater[1]/subcomposite/singleStringData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/singleStringData/@fullDataPath", "composite/repeater[1]/subcomposite/singleStringData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/singleStringData/@modelPath", "composite/repeater/subcomposite/singleStringData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/singleStringData/@fullModelPath", "composite/repeater/subcomposite/singleStringData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/singleStringData", "string value in the sub composite of the first repeater entry");
        
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData)", 2);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[1]/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.DATE_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[1]/@dataPath", "composite/repeater[1]/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[1]/@fullDataPath", "composite/repeater[1]/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[1]/@modelPath", "composite/repeater/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[1]/@fullModelPath", "composite/repeater/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[1]", "2018-09-11T00:00:00.000+02:00");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[2]/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.DATE_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[2]/@dataPath", "composite/repeater[1]/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[2]/@fullDataPath", "composite/repeater[1]/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[2]/@modelPath", "composite/repeater/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[2]/@fullModelPath", "composite/repeater/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[2]", "2015-12-06T00:00:00.000+01:00");
        
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/@dataPath", "composite/repeater[2]");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/@fullDataPath", "composite/repeater[2]");
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/@modelPath)", 0);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/@fullModelPath)", 0);
        
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/longWithDefaultValue)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/longWithDefaultValue/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/longWithDefaultValue/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/longWithDefaultValue/@dataPath", "composite/repeater[2]/longWithDefaultValue");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/longWithDefaultValue/@fullDataPath", "composite/repeater[2]/longWithDefaultValue");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/longWithDefaultValue/@modelPath", "composite/repeater/longWithDefaultValue");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/longWithDefaultValue/@fullModelPath", "composite/repeater/longWithDefaultValue");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/longWithDefaultValue", 95);
        
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/singleBooleanData)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/singleBooleanData/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.BOOLEAN_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/singleBooleanData/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/singleBooleanData/@dataPath", "composite/repeater[2]/singleBooleanData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/singleBooleanData/@fullDataPath", "composite/repeater[2]/singleBooleanData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/singleBooleanData/@modelPath", "composite/repeater/singleBooleanData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/singleBooleanData/@fullModelPath", "composite/repeater/singleBooleanData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/singleBooleanData", true);
        
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/subcomposite)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/@dataPath", "composite/repeater[2]/subcomposite");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/@fullDataPath", "composite/repeater[2]/subcomposite");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/@modelPath", "composite/repeater/subcomposite");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/@fullModelPath", "composite/repeater/subcomposite");
        
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData)", 2);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[1]/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.DATE_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[1]/@dataPath", "composite/repeater[2]/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[1]/@fullDataPath", "composite/repeater[2]/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[1]/@modelPath", "composite/repeater/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[1]/@fullModelPath", "composite/repeater/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[1]", "1985-02-27T00:00:00.000+01:00");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[2]/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.DATE_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[2]/@dataPath", "composite/repeater[2]/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[2]/@fullDataPath", "composite/repeater[2]/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[2]/@modelPath", "composite/repeater/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[2]/@fullModelPath", "composite/repeater/subcomposite/multipleDateData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[2]", "1985-05-24T00:00:00.000+02:00");
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler)} implementation with enumerated data
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxEnumeratedDataWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSaxEnumeratedData", __DATA_NODE_TYPE);
        _testDataHolderSaxEnumeratedData(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler)} implementation with enumerated data
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxEnumeratedDataWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSaxEnumeratedData", __DATA_NODE_TYPE);
        _testDataHolderSaxEnumeratedData(sourceData);
    }
    
    private void _testDataHolderSaxEnumeratedData(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        repositoryData.setValue("singleEnumeratedStringData", "value1");
        repositoryData.setValue("singleEnumeratedLongData", 3L);
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 2);
        
        _assertXpathEquals(document, "count(/result/singleEnumeratedStringData)", 1);
        _assertXpathEquals(document, "/result/singleEnumeratedStringData/@value", "value1");
        _assertXpathEquals(document, "/result/singleEnumeratedStringData", "Value 1");
        
        _assertXpathEquals(document, "count(/result/singleEnumeratedLongData)", 1);
        _assertXpathEquals(document, "/result/singleEnumeratedLongData/@value", 3);
        _assertXpathEquals(document, "/result/singleEnumeratedLongData", "Value 3");
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler)} implementation with resources
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxResourcesWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testDataHolderSaxResources", __DATA_NODE_TYPE);
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        String fileId = ResourceElementTypeHelperForTests.createResource("myFile.txt", "my file content", _ametysObjectResolver);
        _populateRepositoryDataWithResources(repositoryData, fileId);
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 4);
        
        _assertXpathEquals(document, "count(/result/singleBinaryData)", 1);
        _assertXpathEquals(document, "/result/singleBinaryData/@typeId", ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/singleBinaryData/@multiple", "false");
        _assertXpathEquals(document, "/result/singleBinaryData/@path", "singleBinaryData");
        _assertXpathEquals(document, "/result/singleBinaryData/@type", "binaryTypeForTests");
        
        _assertXpathEquals(document, "count(/result/singleFileData)", 1);
        _assertXpathEquals(document, "/result/singleFileData/@typeId", ModelItemTypeConstants.FILE_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/singleFileData/@multiple", "false");
        _assertXpathEquals(document, "/result/singleFileData/@id", fileId);
        _assertXpathEquals(document, "/result/singleFileData/@path", fileId);
        _assertXpathEquals(document, "/result/singleFileData/@type", "explorer");
        
        _assertXpathEquals(document, "count(/result/singleRichTextData)", 1);
        _assertXpathEquals(document, "/result/singleRichTextData/@typeId", ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/singleRichTextData/@multiple", "false");
        
        _assertXpathEquals(document, "count(/result/composite)", 1);
        _assertXpathEquals(document, "count(/result/composite/repeater)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/@entryCount", 2);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry)", 2);

        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/*)", 1);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/subcomposite)", 1);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/subcomposite/*)", 1);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/subcomposite/singleBinaryData)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/singleBinaryData/@typeId", ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/singleBinaryData/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/singleBinaryData/@path", "composite/repeater[1]/subcomposite/singleBinaryData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/singleBinaryData/@type", "binaryTypeForTests");
        
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/*)", 1);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/subcomposite)", 1);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/subcomposite/*)", 1);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/subcomposite/singleBinaryData)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/singleBinaryData/@typeId", ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/singleBinaryData/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/singleBinaryData/@path", "composite/repeater[2]/subcomposite/singleBinaryData");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/singleBinaryData/@type", "binaryTypeForTests");
    }
    
    private void _populateRepositoryDataWithResources(ModifiableRepositoryData repositoryData, String fileId) throws Exception
    {
        _binaryType.write(repositoryData, "singleBinaryData", ResourceElementTypeHelperForTests.createBinary("my binary content"));
        _fileType.write(repositoryData, "singleFileData", fileId);
        _richTextType.write(repositoryData, "singleRichTextData", ResourceElementTypeHelperForTests.createRichText("my rich text content"));
        
        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        ModifiableRepositoryData repeater = composite.addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE);

        ModifiableRepositoryData entry1 = repeater.addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE);
        ModifiableRepositoryData entry1SubComposite = entry1.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        _binaryType.write(entry1SubComposite, "singleBinaryData", ResourceElementTypeHelperForTests.createBinary("content of my binary in first entry of a repeater"));
        
        ModifiableRepositoryData entry2 = repeater.addRepositoryData("2", RepositoryConstants.COMPOSITE_NODETYPE);
        ModifiableRepositoryData entry2SubComposite = entry2.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        _binaryType.write(entry2SubComposite, "singleBinaryData", ResourceElementTypeHelperForTests.createBinary("content of my binary in second entry of a repeater"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX} implementation with binaries in a nested view element
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxBinaryInNestedElementWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testDataHolderSaxResources", __DATA_NODE_TYPE);
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        
        ModifiableContent exh1 = _createExhaustiveContentWithBinary("content1", "binary of my first exhaustive content");
        ModifiableContent exh2 = _createExhaustiveContentWithBinary("content2", "binary of my second exhaustive content");
        _populateRepositoryDataWithBinaryInNestedContent(repositoryData, exh1, exh2);
        View viewToSAX = _createViewWithNestedBinaryData();
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler, viewToSAX);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 2);
        _assertXpathEquals(document, "count(/result/multipleExhContentData)", 2);

        _assertXpathEquals(document, "/result/multipleExhContentData[1]/@id", exh1.getId());
        _assertXpathEquals(document, "/result/multipleExhContentData[1]/@title", "content1");
        
        _assertXpathEquals(document, "count(/result/multipleExhContentData[1]/binary)", 1);
        _assertXpathEquals(document, "/result/multipleExhContentData[1]/binary/@typeId", ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/multipleExhContentData[1]/binary/@multiple", "false");
        _assertXpathEquals(document, "/result/multipleExhContentData[1]/binary/@path", "binary");
        _assertXpathEquals(document, "/result/multipleExhContentData[1]/binary/@uri", "binary?objectId=" + exh1.getId());
        _assertXpathEquals(document, "/result/multipleExhContentData[1]/binary/@type", "attribute");
        
        _assertXpathEquals(document, "/result/multipleExhContentData[2]/@id", exh2.getId());
        _assertXpathEquals(document, "/result/multipleExhContentData[2]/@title", "content2");
        
        _assertXpathEquals(document, "count(/result/multipleExhContentData[2]/binary)", 1);
        _assertXpathEquals(document, "/result/multipleExhContentData[2]/binary/@typeId", ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/multipleExhContentData[2]/binary/@multiple", "false");
        _assertXpathEquals(document, "/result/multipleExhContentData[2]/binary/@path", "binary");
        _assertXpathEquals(document, "/result/multipleExhContentData[2]/binary/@uri", "binary?objectId=" + exh2.getId());
        _assertXpathEquals(document, "/result/multipleExhContentData[2]/binary/@type", "attribute");
        
    }
    
    private void _populateRepositoryDataWithBinaryInNestedContent(ModifiableRepositoryData repositoryData, Content... multipleExhaustiveContentValues)
    {
        String[] multipleExhaustiveContentIds = Arrays.stream(multipleExhaustiveContentValues)
                .map(Content::getId)
                .collect(Collectors.toList())
                .toArray(new String[multipleExhaustiveContentValues.length]);
        repositoryData.setValues("multipleExhContentData", multipleExhaustiveContentIds);
    }
    
    private ModifiableContent _createExhaustiveContentWithBinary(String title, String binaryContent) throws Exception
    {
        ModifiableContent content = _createContent(title, "org.ametys.test.Content.exhaustive");
        content.setValue("binary", ResourceElementTypeHelperForTests.createBinary(binaryContent));
        return content;
    }
    
    private View _createViewWithNestedBinaryData()
    {
        View viewToSAX = new View();
        
        // Create a view item with nested attribute binary
        ViewElementAccessor contentDataViewItem = new ViewElementAccessor();
        contentDataViewItem.setDefinition((ElementDefinition) _model.getModelItem("multipleExhContentData"));
        
        ViewElement viewElement = new ViewElement();
        viewElement.setDefinition((ElementDefinition) _model.getModelItem("multipleExhContentData/binary"));
        contentDataViewItem.addViewItem(viewElement);
        
        // Insert the content view item in the view
        viewToSAX.addViewItem(contentDataViewItem);
        
        return viewToSAX;
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleDataWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSaxSingleData", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleData(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleDataWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSaxSingleData", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleData(sourceData);
    }
    
    private void _testDataHolderSaxSingleData(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        _setValueInRepeaterEntryInRepositoryData(repositoryData, Optional.of("string value"));
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler, "composite/repeater[1]/subcomposite/singleStringData");
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 1);
        _assertXpathEquals(document, "/result/singleStringData", "string value");
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String)} implementation with empty value
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleEmptyDataWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSaxSingleEmptyData", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleEmptyData(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String)} implementation with empty value
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleEmptyDataWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSaxSingleEmptyData", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleEmptyData(sourceData);
    }
    
    private void _testDataHolderSaxSingleEmptyData(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        _setValueInRepeaterEntryInRepositoryData(repositoryData, Optional.of(StringUtils.EMPTY));
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler, "composite/repeater[1]/subcomposite/singleStringData");
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 1);
        _assertXpathEquals(document, "/result/singleStringData", "");
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String, DataContext)} implementation with empty value. This empty value should not be rendered due to the given context
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleNotRenderedEmptyDataWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSaxSingleNotRenderedEmptyData", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleNotRenderedEmptyData(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String, DataContext)} implementation with empty value. This empty value should not be rendered due to the given context
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleNotRenderedEmptyDataWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSaxSingleNotRenderedEmptyData", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleNotRenderedEmptyData(sourceData);
    }
    
    private void _testDataHolderSaxSingleNotRenderedEmptyData(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        _setValueInRepeaterEntryInRepositoryData(repositoryData, Optional.of(StringUtils.EMPTY));
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler, "composite/repeater[1]/subcomposite/singleStringData", DataContext.newInstance().withEmptyValues(false));
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 0);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String)} implementation with no value
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxNoSingleDataWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSaxNoSingleData", __DATA_NODE_TYPE);
        _testDataHolderSaxNoSingleData(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String)} implementation with no value
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxNoSingleDataWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSaxNoSingleData", __DATA_NODE_TYPE);
        _testDataHolderSaxNoSingleData(sourceData);
    }
    
    private void _testDataHolderSaxNoSingleData(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        _setValueInRepeaterEntryInRepositoryData(repositoryData, Optional.empty());
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler, "composite/repeater[1]/subcomposite/singleStringData", DataContext.newInstance().withEmptyValues(false));
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 0);
    }
    
    private void _setValueInRepeaterEntryInRepositoryData(ModifiableRepositoryData repositoryData, Optional<String> value)
    {
        ModifiableRepositoryData subComposite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE)
                .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
                .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
                .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        
        value.ifPresent(v -> subComposite.setValue("singleStringData", v));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String)} implementation with a composite
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleCompositeWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSaxSingleComposite", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleComposite(sourceData);
        
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String)} implementation with a composite
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleCompositeWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSaxSingleComposite", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleComposite(sourceData);
    }
    
    private void _testDataHolderSaxSingleComposite(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent contentForSAX = _createContent("contentForSAX");
        _populateRepositoryData(repositoryData, contentForSAX);
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler, "composite/repeater[1]/subcomposite");
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 1);
        _assertXpathEquals(document, "count(/result/subcomposite)", 1);
        _assertXpathEquals(document, "count(/result/subcomposite/*)", 3);
        
        _assertXpathEquals(document, "/result/subcomposite/singleStringData", "string value in the sub composite of the first repeater entry");
        _assertXpathEquals(document, "count(/result/subcomposite/multipleDateData)", 2);
        _assertXpathEquals(document, "/result/subcomposite/multipleDateData[1]", "2018-09-11T00:00:00.000+02:00");
        _assertXpathEquals(document, "/result/subcomposite/multipleDateData[2]", "2015-12-06T00:00:00.000+01:00");
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String)} implementation with a repeater
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleRepeaterWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSaxSingleRepeater", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleRepeater(sourceData);
        
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String)} implementation with a repeater
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleRepeaterWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSaxSingleRepeater", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleRepeater(sourceData);
    }
    
    private void _testDataHolderSaxSingleRepeater(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent contentForSAX = _createContent("contentForSAX");
        _populateRepositoryData(repositoryData, contentForSAX);
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler, "composite/repeater");
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 1);
        _assertXpathEquals(document, "count(/result/repeater)", 1);
        _assertXpathEquals(document, "/result/repeater/@entryCount", 2);
        _assertXpathEquals(document, "count(/result/repeater/*)", 2);
        _assertXpathEquals(document, "count(/result/repeater/entry)", 2);

        _assertXpathEquals(document, "count(/result/repeater/entry[@name='1']/longWithDefaultValue)", 1);
        _assertXpathEquals(document, "/result/repeater/entry[@name='1']/longWithDefaultValue", 15);
        _assertXpathEquals(document, "count(/result/repeater/entry[@name='1']/singleDoubleData)", 1);
        _assertXpathEquals(document, "/result/repeater/entry[@name='1']/singleDoubleData", 0.6);
        _assertXpathEquals(document, "count(/result/repeater/entry[@name='1']/subcomposite)", 1);
        _assertXpathEquals(document, "count(/result/repeater/entry[@name='1']/subcomposite/singleStringData)", 1);
        _assertXpathEquals(document, "/result/repeater/entry[@name='1']/subcomposite/singleStringData", "string value in the sub composite of the first repeater entry");
        _assertXpathEquals(document, "count(/result/repeater/entry[@name='1']/subcomposite/multipleDateData)", 2);
        _assertXpathEquals(document, "/result/repeater/entry[@name='1']/subcomposite/multipleDateData[1]", "2018-09-11T00:00:00.000+02:00");
        _assertXpathEquals(document, "/result/repeater/entry[@name='1']/subcomposite/multipleDateData[2]", "2015-12-06T00:00:00.000+01:00");

        _assertXpathEquals(document, "count(/result/repeater/entry[@name='2']/longWithDefaultValue)", 1);
        _assertXpathEquals(document, "/result/repeater/entry[@name='2']/longWithDefaultValue", 95);
        _assertXpathEquals(document, "count(/result/repeater/entry[@name='2']/singleBooleanData)", 1);
        _assertXpathEquals(document, "/result/repeater/entry[@name='2']/singleBooleanData", true);
        _assertXpathEquals(document, "count(/result/repeater/entry[@name='2']/subcomposite)", 1);
        _assertXpathEquals(document, "count(/result/repeater/entry[@name='2']/subcomposite/multipleDateData)", 2);
        _assertXpathEquals(document, "/result/repeater/entry[@name='2']/subcomposite/multipleDateData[1]", "1985-02-27T00:00:00.000+01:00");
        _assertXpathEquals(document, "/result/repeater/entry[@name='2']/subcomposite/multipleDateData[2]", "1985-05-24T00:00:00.000+02:00");
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String)} implementation with a repeater entry
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleRepeaterEntryWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSaxSingleRepeaterEntry", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleRepeaterEntry(sourceData);
        
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String)} implementation with a repeater entry
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleRepeaterEntryWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSaxSingleRepeaterEntry", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleRepeaterEntry(sourceData);
    }
    
    private void _testDataHolderSaxSingleRepeaterEntry(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent contentForSAX = _createContent("contentForSAX");
        _populateRepositoryData(repositoryData, contentForSAX);
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler, "composite/repeater[2]");
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 1);
        _assertXpathEquals(document, "count(/result/entry)", 1);
        _assertXpathEquals(document, "/result/entry/@name", 2);
        _assertXpathEquals(document, "count(/result/entry/*)", 3);

        _assertXpathEquals(document, "count(/result/entry/longWithDefaultValue)", 1);
        _assertXpathEquals(document, "/result/entry/longWithDefaultValue", 95);
        _assertXpathEquals(document, "count(/result/entry/singleBooleanData)", 1);
        _assertXpathEquals(document, "/result/entry/singleBooleanData", true);
        _assertXpathEquals(document, "count(/result/entry/subcomposite)", 1);
        _assertXpathEquals(document, "count(/result/entry/subcomposite/multipleDateData)", 2);
        _assertXpathEquals(document, "/result/entry/subcomposite/multipleDateData[1]", "1985-02-27T00:00:00.000+01:00");
        _assertXpathEquals(document, "/result/entry/subcomposite/multipleDateData[2]", "1985-05-24T00:00:00.000+02:00");
    }
    
    /**
     * Test {@link ModelAwareComposite#dataToSAX(ContentHandler)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeSaxWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testCompositeSax", __DATA_NODE_TYPE);
        _testCompositeSax(sourceData);
        
    }
    
    /**
     * Test {@link ModelAwareComposite#dataToSAX(ContentHandler)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeSaxWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testCompositeSax", __DATA_NODE_TYPE);
        _testCompositeSax(sourceData);
    }
    
    private void _testCompositeSax(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModifiableContent contentForSAXView = _createContent("contentForSAXView");
        _populateRepositoryData(repositoryData, contentForSAXView);
        
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModelAwareComposite composite = dataHolder.getComposite("composite");
     
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        composite.dataToSAX(handler);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/singleLongData)", 1);
        _assertXpathEquals(document, "/result/singleLongData", 54);
        
        _assertXpathEquals(document, "count(/result/repeater)", 1);
        _assertXpathEquals(document, "/result/repeater/@entryCount", 2);
        _assertXpathEquals(document, "count(/result/repeater/entry)", 2);

        _assertXpathEquals(document, "count(/result/repeater/entry[@name='1']/longWithDefaultValue)", 1);
        _assertXpathEquals(document, "/result/repeater/entry[@name='1']/longWithDefaultValue", 15);
        _assertXpathEquals(document, "count(/result/repeater/entry[@name='1']/singleDoubleData)", 1);
        _assertXpathEquals(document, "/result/repeater/entry[@name='1']/singleDoubleData", 0.6);
        _assertXpathEquals(document, "count(/result/repeater/entry[@name='1']/subcomposite)", 1);
        _assertXpathEquals(document, "count(/result/repeater/entry[@name='1']/subcomposite/singleStringData)", 1);
        _assertXpathEquals(document, "/result/repeater/entry[@name='1']/subcomposite/singleStringData", "string value in the sub composite of the first repeater entry");
        _assertXpathEquals(document, "count(/result/repeater/entry[@name='1']/subcomposite/multipleDateData)", 2);
        _assertXpathEquals(document, "/result/repeater/entry[@name='1']/subcomposite/multipleDateData[1]", "2018-09-11T00:00:00.000+02:00");
        _assertXpathEquals(document, "/result/repeater/entry[@name='1']/subcomposite/multipleDateData[2]", "2015-12-06T00:00:00.000+01:00");

        _assertXpathEquals(document, "count(/result/repeater/entry[@name='2']/longWithDefaultValue)", 1);
        _assertXpathEquals(document, "/result/repeater/entry[@name='2']/longWithDefaultValue", 95);
        _assertXpathEquals(document, "count(/result/repeater/entry[@name='2']/singleBooleanData)", 1);
        _assertXpathEquals(document, "/result/repeater/entry[@name='2']/singleBooleanData", true);
        _assertXpathEquals(document, "count(/result/repeater/entry[@name='2']/subcomposite)", 1);
        _assertXpathEquals(document, "count(/result/repeater/entry[@name='2']/subcomposite/multipleDateData)", 2);
        _assertXpathEquals(document, "/result/repeater/entry[@name='2']/subcomposite/multipleDateData[1]", "1985-02-27T00:00:00.000+01:00");
        _assertXpathEquals(document, "/result/repeater/entry[@name='2']/subcomposite/multipleDateData[2]", "1985-05-24T00:00:00.000+02:00");
    }
    
    /**
     * Test {@link ModelAwareComposite#dataToSAX(ContentHandler, String)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeSaxSingleDataWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testCompositeSaxSingleData", __DATA_NODE_TYPE);
        _testCompositeSaxSingleData(sourceData);
    }
    
    /**
     * Test {@link ModelAwareComposite#dataToSAX(ContentHandler, String)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeSaxSingleDataWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testCompositeSaxSingleData", __DATA_NODE_TYPE);
        _testCompositeSaxSingleData(sourceData);
    }
    
    private void _testCompositeSaxSingleData(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModifiableContent contentForSAX = _createContent("contentForSAX");
        _populateRepositoryData(repositoryData, contentForSAX);

        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModelAwareComposite composite = dataHolder.getComposite("composite");
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        composite.dataToSAX(handler, "repeater[1]/subcomposite/singleStringData");
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 1);
        _assertXpathEquals(document, "/result/singleStringData", "string value in the sub composite of the first repeater entry");
    }
    
    /**
     * Test {@link ModelAwareRepeater#dataToSAX(ContentHandler)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepeaterSaxWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testRepeaterSax", __DATA_NODE_TYPE);
        _testRepeaterSax(sourceData);
    }
    
    /**
     * Test {@link ModelAwareRepeater#dataToSAX(ContentHandler)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepeaterSaxWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testRepeaterSax", __DATA_NODE_TYPE);
        _testRepeaterSax(sourceData);
    }
    
    private void _testRepeaterSax(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModifiableContent contentForSAXView = _createContent("contentForSAXView");
        _populateRepositoryData(repositoryData, contentForSAXView);
        
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModelAwareRepeater repeater = dataHolder.getRepeater("composite/repeater");
     
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        repeater.dataToSAX(handler);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/entry)", 2);

        _assertXpathEquals(document, "count(/result/entry[@name='1']/longWithDefaultValue)", 1);
        _assertXpathEquals(document, "/result/entry[@name='1']/longWithDefaultValue", 15);
        _assertXpathEquals(document, "count(/result/entry[@name='1']/singleDoubleData)", 1);
        _assertXpathEquals(document, "/result/entry[@name='1']/singleDoubleData", 0.6);
        _assertXpathEquals(document, "count(/result/entry[@name='1']/subcomposite)", 1);
        _assertXpathEquals(document, "count(/result/entry[@name='1']/subcomposite/singleStringData)", 1);
        _assertXpathEquals(document, "/result/entry[@name='1']/subcomposite/singleStringData", "string value in the sub composite of the first repeater entry");
        _assertXpathEquals(document, "count(/result/entry[@name='1']/subcomposite/multipleDateData)", 2);
        _assertXpathEquals(document, "/result/entry[@name='1']/subcomposite/multipleDateData[1]", "2018-09-11T00:00:00.000+02:00");
        _assertXpathEquals(document, "/result/entry[@name='1']/subcomposite/multipleDateData[2]", "2015-12-06T00:00:00.000+01:00");

        _assertXpathEquals(document, "count(/result/entry[@name='2']/longWithDefaultValue)", 1);
        _assertXpathEquals(document, "/result/entry[@name='2']/longWithDefaultValue", 95);
        _assertXpathEquals(document, "count(/result/entry[@name='2']/singleBooleanData)", 1);
        _assertXpathEquals(document, "/result/entry[@name='2']/singleBooleanData", true);
        _assertXpathEquals(document, "count(/result/entry[@name='2']/subcomposite)", 1);
        _assertXpathEquals(document, "count(/result/entry[@name='2']/subcomposite/multipleDateData)", 2);
        _assertXpathEquals(document, "/result/entry[@name='2']/subcomposite/multipleDateData[1]", "1985-02-27T00:00:00.000+01:00");
        _assertXpathEquals(document, "/result/entry[@name='2']/subcomposite/multipleDateData[2]", "1985-05-24T00:00:00.000+02:00");
    }
    
    /**
     * Test {@link ModelAwareRepeater#dataToSAX(ContentHandler, String)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepeaterSaxSingleDataWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testRepeaterSaxSingleData", __DATA_NODE_TYPE);
        _testRepeaterSaxSingleData(sourceData);
        
    }
    
    /**
     * Test {@link ModelAwareRepeater#dataToSAX(ContentHandler, String)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepeaterSaxSingleDataWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testRepeaterSaxSingleData", __DATA_NODE_TYPE);
        _testRepeaterSaxSingleData(sourceData);
    }
    
    private void _testRepeaterSaxSingleData(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModifiableContent contentForSAX = _createContent("contentForSAX");
        _populateRepositoryData(repositoryData, contentForSAX);

        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModelAwareRepeater repeater = dataHolder.getRepeater("composite/repeater");
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        repeater.dataToSAX(handler, "subcomposite/singleStringData");
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 2);
        _assertXpathEquals(document, "count(/result/entry)", 2);

        _assertXpathEquals(document, "count(/result/entry[@name='1']/*)", 1);
        _assertXpathEquals(document, "/result/entry[@name='1']/singleStringData", "string value in the sub composite of the first repeater entry");

        _assertXpathEquals(document, "count(/result/entry[@name='2']/*)", 0);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, ViewItemAccessor)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxViewWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSaxView", __DATA_NODE_TYPE);
        _testDataHolderSaxView(sourceData);
        
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, ViewItemAccessor)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxViewWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSaxView", __DATA_NODE_TYPE);
        _testDataHolderSaxView(sourceData);
    }
    
    private void _testDataHolderSaxView(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent contentForSAXView = _createContent("contentForSAXView");
        _populateRepositoryData(repositoryData, contentForSAXView);
        View viewToSAX = _createViewToSAX();
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        Set<String> externalizableData = _externalizableDataProviderEP.getExternalizableDataPaths(dataHolder);
        RepositoryDataContext context = RepositoryDataContext.newInstance()
                                                             .withExternalizableData(externalizableData);
        dataHolder.dataToSAX(handler, viewToSAX, context);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/singleStringData)", 1);
        _assertXpathEquals(document, "/result/singleStringData", "value of the single string data");
        _assertXpathEquals(document, "count(/result/singleLongData)", 1);
        _assertXpathEquals(document, "/result/singleLongData", "");
        _assertXpathEquals(document, "count(/result/multipleLongData)", 0);
        _assertXpathEquals(document, "count(/result/singleDateData)", 0);

        _assertXpathEquals(document, "count(/result/singleGeocodeData)", 1);
        _assertXpathEquals(document, "/result/singleGeocodeData/@latitude", 85.487);
        _assertXpathEquals(document, "/result/singleGeocodeData/@longitude", 6.254);
        
        _assertXpathEquals(document, "count(/result/multipleUserData)", 2);
        _assertXpathEquals(document, "/result/multipleUserData[1]/@login", "user1");
        _assertXpathEquals(document, "/result/multipleUserData[1]/@populationId", POPULATION_TEST_ID);
        _assertXpathEquals(document, "/result/multipleUserData[1]", "User One");
        
        _assertXpathEquals(document, "/result/multipleUserData[2]/@login", "user3");
        _assertXpathEquals(document, "/result/multipleUserData[2]/@populationId", POPULATION_TEST_ID);
        _assertXpathEquals(document, "/result/multipleUserData[2]", "User Three");
        
        _assertXpathEquals(document, "count(/result/singleContentData)", 1);
        _assertXpathEquals(document, "/result/singleContentData/@id", contentForSAXView.getId());
        _assertXpathEquals(document, "/result/singleContentData/@title", "contentForSAXView");
        
        _assertXpathEquals(document, "count(/result/singleContentData/field1)", 1);
        _assertXpathEquals(document, "/result/singleContentData/field1", "value for field 1");
        
        _assertXpathEquals(document, "count(/result/composite)", 1);
        _assertXpathEquals(document, "count(/result/composite/singleLongData)", 1);
        _assertXpathEquals(document, "/result/composite/singleLongData", 54);
        
        _assertXpathEquals(document, "count(/result/composite/repeater)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/@entryCount", 2);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry)", 2);

        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/longWithDefaultValue)", 0);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/singleBooleanData)", 0);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/singleDoubleData)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/singleDoubleData", 0.6);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/subcomposite)", 1);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/subcomposite/singleStringData)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/singleStringData", "string value in the sub composite of the first repeater entry");
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData)", 2);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[1]", "2018-09-11T00:00:00.000+02:00");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[2]", "2015-12-06T00:00:00.000+01:00");

        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/longWithDefaultValue)", 0);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/singleBooleanData)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/singleBooleanData", true);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/singleDoubleData)", 0);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/subcomposite)", 1);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData)", 2);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[1]", "1985-02-27T00:00:00.000+01:00");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[2]", "1985-05-24T00:00:00.000+02:00");
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAXForEdition(ContentHandler, ViewItemAccessor, DataContext)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxViewForEditionWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSaxViewForEdition", __DATA_NODE_TYPE);
        _testDataHolderSaxViewForEdition(sourceData);
        
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAXForEdition(ContentHandler, ViewItemAccessor, DataContext)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxViewForEditionWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSaxViewForEdition", __DATA_NODE_TYPE);
        _testDataHolderSaxViewForEdition(sourceData);
    }
    
    private void _testDataHolderSaxViewForEdition(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        ModifiableContent contentForSAXView = _createContent("contentForSAXView");
        _populateRepositoryData(repositoryData, contentForSAXView);
        View viewToSAX = _createViewToSAX();
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        Set<String> externalizableData = _externalizableDataProviderEP.getExternalizableDataPaths(dataHolder);
        RepositoryDataContext context = RepositoryDataContext.newInstance()
                                                             .withExternalizableData(externalizableData);
        dataHolder.dataToSAXForEdition(handler, viewToSAX, context);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/singleStringData)", 1);
        _assertXpathEquals(document, "/result/singleStringData/@json", true);
        _assertXpathEquals(document, "/result/singleStringData", Map.of("local", "value of the single string data",
                                                                        "external", "external value of the single string data",
                                                                        "status", "local"));
        _assertXpathEquals(document, "count(/result/singleLongData)", 1);
        _assertXpathEquals(document, "/result/singleLongData", "");
        _assertXpathEquals(document, "count(/result/multipleLongData)", 0);
        _assertXpathEquals(document, "count(/result/singleDateData)", 0);

        _assertXpathEquals(document, "count(/result/singleGeocodeData)", 1);
        _assertXpathEquals(document, "/result/singleGeocodeData/@json", true);
        _assertXpathEquals(document, "/result/singleGeocodeData", Map.of("latitude", 85.487,
                                                                         "longitude", 6.254));
        
        _assertXpathEquals(document, "count(/result/multipleUserData)", 1);
        _assertXpathEquals(document, "/result/multipleUserData/@json", true);
        _assertXpathEquals(document, "/result/multipleUserData", List.of(Map.of("login", "user1",
                                                                                "populationId", POPULATION_TEST_ID,
                                                                                "fullname", "User One"),
                                                                         Map.of("login", "user3",
                                                                                "populationId", POPULATION_TEST_ID,
                                                                                "fullname", "User Three")));
        
        _assertXpathEquals(document, "count(/result/singleContentData)", 1);
        _assertXpathEquals(document, "/result/singleContentData/@json", true);
        _assertXpathEquals(document, "/result/singleContentData", Map.of("id", contentForSAXView.getId(),
                                                                         "title", "contentForSAXView",
                                                                         "field1", "value for field 1"));
        
        _assertXpathEquals(document, "count(/result/composite)", 1);
        _assertXpathEquals(document, "count(/result/composite/singleLongData)", 1);
        _assertXpathEquals(document, "/result/composite/singleLongData", 54);
        
        _assertXpathEquals(document, "count(/result/composite/repeater)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/@entryCount", 2);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry)", 2);

        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/longWithDefaultValue)", 0);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/singleBooleanData)", 0);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/singleDoubleData)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/singleDoubleData", 0.6);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/subcomposite)", 1);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/subcomposite/singleStringData)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/singleStringData", Map.of("local", "string value in the sub composite of the first repeater entry",
                                                                                                                         "external", "external string value in the sub composite of the first repeater entry",
                                                                                                                         "status", "local"));
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData)", 2);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[1]", "2018-09-11T00:00:00.000+02:00");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='1']/subcomposite/multipleDateData[2]", "2015-12-06T00:00:00.000+01:00");

        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/longWithDefaultValue)", 0);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/singleBooleanData)", 1);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/singleBooleanData", true);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/singleDoubleData)", 0);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/subcomposite)", 1);
        _assertXpathEquals(document, "count(/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData)", 2);
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[1]", "1985-02-27T00:00:00.000+01:00");
        _assertXpathEquals(document, "/result/composite/repeater/entry[@name='2']/subcomposite/multipleDateData[2]", "1985-05-24T00:00:00.000+02:00");
    }
    
    private View _createViewToSAX()
    {
        View viewToSAX = View.of(_model, "singleStringData",
                "singleLongData",
                "singleDateData",
                "singleGeocodeData",
                "multipleUserData",
                "composite/singleLongData",
                "composite/repeater/singleBooleanData",
                "composite/repeater/singleDoubleData",
                "composite/repeater/subcomposite",
                "unvaluedComposite",
                "unvaluedRepeater");
        
        // Create a view item with nested attribute field1
        ViewElementAccessor singleContentDataViewItem = new ViewElementAccessor();
        singleContentDataViewItem.setDefinition((ElementDefinition) _model.getModelItem("singleContentData"));
        
        ViewElement field1ViewElement = new ViewElement();
        field1ViewElement.setDefinition((ElementDefinition) _model.getModelItem("singleContentData/field1"));
        singleContentDataViewItem.addViewItem(field1ViewElement);
        
        // Insert the content view item in the view
        viewToSAX.insertViewItem(singleContentDataViewItem, 5);
        
        return viewToSAX;
    }
    
    private ModifiableContent _createContent(String title)
    {
        ModifiableContent content = _createContent(title, "org.ametys.cms.contenttype.Test");
        content.setValue("field1", "value for field 1");
        return content;
    }
    
    private ModifiableContent _createContent(String title, String... contentTypeIds)
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        
        ModifiableContent content = root.createChild(title, "ametys:defaultContent");

        content.setTypes(contentTypeIds);
        content.setTitle(title, Locale.FRANCE);
        content.setLanguage("fr");
        content.setCreationDate(ZonedDateTime.now());
        content.setCreator(new UserIdentity("john", "anyPopulation"));
        content.setLastModified(ZonedDateTime.now());
        content.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        
        return content;
    }
    
    private void _populateRepositoryData(ModifiableRepositoryData repositoryData, Content singleContentValue) throws Exception
    {
        repositoryData.setValue("singleStringData", "value of the single string data");
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "external value of the single string data");
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.STATUS_SUFFIX, "local");
        
        repositoryData.setValue("singleLongData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        repositoryData.setValues("multipleLongData", new Long[] {32L, 784L});
        
        _geocodeType.write(repositoryData, "singleGeocodeData", new Geocode(85.487D, 6.254D));
        _userType.write(repositoryData, "multipleUserData", new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY});
        repositoryData.setValue("singleContentData", singleContentValue.getId());
        
        ZonedDateTime[] dateTimeArray = new ZonedDateTime[]{ZonedDateTime.of(2018, 5, 5, 7, 25, 13, 596_000_000, ZoneId.of("UTC+05:00")), ZonedDateTime.of(2015, 12, 06, 15, 18, 0, 0, ZoneId.of("UTC-03:00"))};
        repositoryData.setValues("multipleDatetimeData", Stream.of(dateTimeArray).map(DateUtils::asCalendar).toArray(Calendar[]::new));
        
        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        composite.setValue("singleLongData", 54L);
        ModifiableRepositoryData repeater = composite.addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE);

        ModifiableRepositoryData entry1 = repeater.addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE);
        entry1.setValue("longWithDefaultValue", 15L);
        entry1.setValue("singleDoubleData", 0.6D);
        ModifiableRepositoryData entry1SubComposite = entry1.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        entry1SubComposite.setValue("singleStringData", "string value in the sub composite of the first repeater entry");
        entry1SubComposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "external string value in the sub composite of the first repeater entry");
        entry1SubComposite.setValue("singleStringData" + ModelAwareDataHolder.STATUS_SUFFIX, "local");
        entry1SubComposite.setValues("multipleDateData", new Calendar[]{DateUtils.asCalendar(LocalDate.of(2018, Month.SEPTEMBER, 11)), DateUtils.asCalendar(LocalDate.of(2015, Month.DECEMBER, 06))});
        
        ModifiableRepositoryData entry2 = repeater.addRepositoryData("2", RepositoryConstants.COMPOSITE_NODETYPE);
        entry2.setValue("longWithDefaultValue", 95L);
        entry2.setValue("singleBooleanData", true);
        ModifiableRepositoryData entry2SubComposite = entry2.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        entry2SubComposite.setValues("multipleDateData", new Calendar[]{DateUtils.asCalendar(LocalDate.of(1985, Month.FEBRUARY, 27)), DateUtils.asCalendar(LocalDate.of(1985, Month.MAY, 24))});
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String)} implementation with a non existing content
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxInexistingContentWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSaxInexistingContent", __DATA_NODE_TYPE);
        _testDataHolderSaxInexistingContent(sourceData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#dataToSAX(ContentHandler, String)} implementation with a non existing content
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxInexistingContentWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSaxInexistingContent", __DATA_NODE_TYPE);
        _testDataHolderSaxInexistingContent(sourceData);
    }
    
    private void _testDataHolderSaxInexistingContent(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);
        // Add a reference to an inexisting content in the repository
        repositoryData.setValue("singleContentData", "defaultContent://inexisting-content-id");
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler, "singleContentData");
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/singleContentData)", 0);
    }
    
    private void _assertXpathEquals(Node node, String expression, String value) throws Exception
    {
        Assert.assertEquals(value, XPathAPI.eval(node, expression).str());
    }
    
    private void _assertXpathEquals(Node node, String expression, long value) throws Exception
    {
        Assert.assertEquals(value, (long) XPathAPI.eval(node, expression).num());
    }
    
    private void _assertXpathEquals(Node node, String expression, double value) throws Exception
    {
        Assert.assertEquals(Double.valueOf(value), Double.valueOf(XPathAPI.eval(node, expression).num()));
    }
    
    private void _assertXpathEquals(Node node, String expression, boolean value) throws Exception
    {
        Assert.assertEquals(value, XPathAPI.eval(node, expression).bool());
    }
    
    private void _assertXpathEquals(Node node, String expression, Map<String, Object> expectedValue) throws Exception
    {
        String valueAsString = XPathAPI.selectSingleNode(node, expression).getTextContent();
        Map<String, Object> value = _jsonUtils.convertJsonToMap(valueAsString);
        
        // Check that all expected values are in the saxed events. Saxed events may have more unchecked data
        for (String key : expectedValue.keySet())
        {
            Assert.assertEquals(expectedValue.get(key), value.get(key));
        }
    }
    
    private void _assertXpathEquals(Node node, String expression, List<Map<String, Object>> expectedValues) throws Exception
    {
        String valueAsString = XPathAPI.selectSingleNode(node, expression).getTextContent();
        List<Object> values = _jsonUtils.convertJsonToList(valueAsString);
        
        // Check that all expected values are in the saxed events. Saxed events may have more unchecked data
        for (int i = 0; i < expectedValues.size(); i++)
        {
            @SuppressWarnings("unchecked")
            Map<String, Object> value = (Map<String, Object>) values.get(i);
            Map<String, Object> expectedValue = expectedValues.get(i);
            
            for (String key : expectedValue.keySet())
            {
                Assert.assertEquals(expectedValue.get(key), value.get(key));
            }
        }
    }
}
