/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.time.LocalDate;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.holder.ModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModelLessComposite;
import org.ametys.plugins.repository.data.holder.group.impl.DefaultModelLessComposite;
import org.ametys.plugins.repository.data.holder.impl.DefaultModelLessDataHolder;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.plugins.repository.data.type.RepositoryModelItemType;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.exception.NotUniqueTypeException;
import org.ametys.runtime.model.exception.UnknownTypeException;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelLessDataHolder} basic methods implementation test case
 */
@NeedFeatures({
    "test-data/test-data-types",
    "test-data/test-model-less-data-types",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModelLessDataHolderBasicsTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPointWithIncompatibleTypes;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    
    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;
    
    private static RepositoryElementType<Geocode> _geocodeType;
    private static RepositoryElementType<UserIdentity> _userType;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".TestModelLess");
        _testDataTypeExtensionPointWithIncompatibleTypes = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    @SuppressWarnings("unchecked")
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");
        
        _geocodeType = (RepositoryElementType<Geocode>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID);
        _userType = (RepositoryElementType<UserIdentity>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#isMultiple(String)} method's nominal cases
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testIsMultipleNominalCasesJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testIsMultipleNominalCases", DATA_NODE_TYPE);
        _testIsMultipleNominalCases(repositoryData);
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#isMultiple(String)} method's nominal cases
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testIsMultipleNominalCasesMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testIsMultipleNominalCases", DATA_NODE_TYPE);
        _testIsMultipleNominalCases(repositoryData);
    }
    
    private void _testIsMultipleNominalCases(ModifiableRepositoryData repositoryData)
    {
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        repositoryData.setValues("multipleDoubleData", new Double[]{12.0, 58.5, 0.45});
        Assert.assertTrue(dataHolder.isMultiple("multipleDoubleData"));
        
        repositoryData.setValue("simpleDoubleData", 25.4);
        Assert.assertFalse(dataHolder.isMultiple("simpleDoubleData"));
        
        ModifiableRepositoryData compositeData = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        Assert.assertFalse(dataHolder.isMultiple("composite"));
        
        compositeData.setValues("multipleStringData", new String[]{"string1", "string2"});
        Assert.assertTrue(dataHolder.isMultiple("composite/multipleStringData"));
        
        compositeData.setValue("simpleStringData", "string");
        Assert.assertFalse(dataHolder.isMultiple("composite/simpleStringData"));
        
        _geocodeType.write(compositeData, "multipleGeocodeData", new Geocode[]{new Geocode(45.36, 1.43), new Geocode(12.98, 0.74)});
        Assert.assertTrue(dataHolder.isMultiple("composite/multipleGeocodeData"));
        
        _userType.write(compositeData, "simpleUserData", new UserIdentity("jsnow", "got"));
        Assert.assertFalse(dataHolder.isMultiple("composite/simpleUserData"));
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#isMultiple(String)} method on inexisting data
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testIsMultipleInexistingDataJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testIsMultipleInexistingData", DATA_NODE_TYPE);
        _testIsMultipleInexistingData(repositoryData);
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#isMultiple(String)} method on inexisting data
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testIsMultipleInexistingDataMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testIsMultipleInexistingData", DATA_NODE_TYPE);
        _testIsMultipleInexistingData(repositoryData);
    }
    
    private void _testIsMultipleInexistingData(ModifiableRepositoryData repositoryData)
    {
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        try
        {
            dataHolder.isMultiple("inexistingData");
            Assert.fail("The item at path 'inexistingData' does not exist, it can't be multiple");
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains("inexistingData"));
        }
        
        try
        {
            dataHolder.isMultiple("inexisting/data/in/group");
            Assert.fail("The item at path 'inexisting/data/in/group' does not exist, it can't be multiple");
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains("inexisting"));
        }
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#getType(String)} method's nominal cases
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testGetTypeNominalCasesJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testGetTypeNominalCases", DATA_NODE_TYPE);
        _testGetTypeNominalCases(repositoryData);
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#getType(String)} method's nominal cases
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testGetTypeNominalCasesMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testGetTypeNominalCases", DATA_NODE_TYPE);
        _testGetTypeNominalCases(repositoryData);
    }
    
    @SuppressWarnings("static-access")
    private void _testGetTypeNominalCases(ModifiableRepositoryData repositoryData)
    {
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        repositoryData.setValue("stringData", "string value");
        Assert.assertEquals(ModelItemTypeConstants.STRING_TYPE_ID, dataHolder.getType("stringData").getId());
        repositoryData.setValues("multipleStringData", new String[] {"first string value", "second string value"});
        Assert.assertEquals(ModelItemTypeConstants.STRING_TYPE_ID, dataHolder.getType("multipleStringData").getId());
        
        repositoryData.setValue("longData", 18L);
        Assert.assertEquals(ModelItemTypeConstants.LONG_TYPE_ID, dataHolder.getType("longData").getId());
        
        _userType.write(repositoryData, "userData", new UserIdentity("mknopfler", "dire-straits"));
        Assert.assertEquals(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID, dataHolder.getType("userData").getId());
        _userType.write(repositoryData, "multipleUserData", new UserIdentity[] {new UserIdentity("tyorke", "radiohead")});
        Assert.assertEquals(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID, dataHolder.getType("multipleUserData").getId());
        
        ModifiableRepositoryData compositeData = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        Assert.assertEquals(ModelItemTypeConstants.COMPOSITE_TYPE_ID, dataHolder.getType("composite").getId());
        compositeData.setValue("longData", 9456L);
        Assert.assertEquals(ModelItemTypeConstants.LONG_TYPE_ID, dataHolder.getType("composite/longData").getId());
        ModelLessComposite composite = new DefaultModelLessComposite(_testDataTypeExtensionPoint, compositeData, dataHolder, dataHolder);
        Assert.assertEquals(ModelItemTypeConstants.LONG_TYPE_ID, composite.getType("longData").getId());
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#getType(String)} method for a non existing data
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testGetTypeOnInexistingDataJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testGetTypeOnInexistingData", DATA_NODE_TYPE);
        _testGetTypeOnInexistingData(repositoryData);
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#getType(String)} method for a non existing data
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testGetTypeOnInexistingDataMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testGetTypeOnInexistingData", DATA_NODE_TYPE);
        _testGetTypeOnInexistingData(repositoryData);
    }
    
    private void _testGetTypeOnInexistingData(ModifiableRepositoryData repositoryData)
    {
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        try
        {
            dataHolder.getType("inexistingData");
            Assert.fail("This data does not exist, it should not be possible to get its type");
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains("exist"));
        }
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#getType(String)} method for a data with an unknown type
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testGetTypeOnDataWithAnUnknownTypeJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testGetTypeOnDataWithAnUnknownType", DATA_NODE_TYPE);
        _testGetTypeOnDataWithAnUnknownType(repositoryData);
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#getType(String)} method for a data with an unknown type
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testGetTypeOnDataWithAnUnknownTypeMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testGetTypeOnDataWithAnUnknownType", DATA_NODE_TYPE);
        _testGetTypeOnDataWithAnUnknownType(repositoryData);
    }
    
    private void _testGetTypeOnDataWithAnUnknownType(ModifiableRepositoryData repositoryData)
    {
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        // ametys:data is an existing node type but does not correspond to any model item type
        repositoryData.addRepositoryData("dataWithUnknownType", "ametys:data");

        try
        {
            dataHolder.getType("dataWithUnknownType");
            Assert.fail("the type of this data is unknown, it should not be possible to get its type");
        }
        catch (UnknownTypeException e)
        {
            Assert.assertTrue(e.getMessage().contains("found"));
        }
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#getType(String)} method for a data with many compatible types
     * In the test extension point, there are date and datetime extensions.
     * But it is not possible to determine from the storage if a calendar without time is a date or datetime data
     * This kind of conflict has to be avoided by not including 2 incompatible types in an extension point used by Model less data holders
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testGetTypeOnDataWithManyCompatibleTypesJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testGetTypeOnDataWithManyCompatibleTypes", DATA_NODE_TYPE);
        _testGetTypeOnDataWithManyCompatibleTypes(repositoryData);
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#getType(String)} method for a data with many compatible types
     * In the test extension point, there are date and datetime extensions.
     * But it is not possible to determine from the storage if a calendar without time is a date or datetime data
     * This kind of conflict has to be avoided by not including 2 incompatible types in an extension point used by Model less data holders
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testGetTypeOnDataWithManyCompatibleTypesMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testGetTypeOnDataWithManyCompatibleTypes", DATA_NODE_TYPE);
        _testGetTypeOnDataWithManyCompatibleTypes(repositoryData);
    }
    
    @SuppressWarnings("static-access")
    private void _testGetTypeOnDataWithManyCompatibleTypes(ModifiableRepositoryData repositoryData)
    {
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPointWithIncompatibleTypes, repositoryData);
        
        repositoryData.setValue("dateData", DateUtils.asCalendar(LocalDate.of(2019, 2, 6)));

        try
        {
            dataHolder.getType("dateData");
            Assert.fail("the type of this data can be either date or datetime, it should not be possible to get its type");
        }
        catch (NotUniqueTypeException e)
        {
            Assert.assertTrue(e.getMessage().contains(ModelItemTypeConstants.DATE_TYPE_ID));
            Assert.assertTrue(e.getMessage().contains(ModelItemTypeConstants.DATETIME_TYPE_ID));
        }
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#getType(String)} method for an empty data
     * Some data (long double, dates, ...) are stored as empty by a boolean property specifying that the data is empty
     * So there is no way to determine the type of such an empty data
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testGetTypeOnEmptyDataJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testGetTypeOnEmptyData", DATA_NODE_TYPE);
        _testGetTypeOnEmptyData(repositoryData);
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#getType(String)} method for an empty data
     * Some data (long double, dates, ...) are stored as empty by a boolean property specifying that the data is empty
     * So there is no way to determine the type of such an empty data
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testGetTypeOnEmptyDataMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testGetTypeOnEmptyData", DATA_NODE_TYPE);
        _testGetTypeOnEmptyData(repositoryData);
    }
    
    @SuppressWarnings("static-access")
    private void _testGetTypeOnEmptyData(ModifiableRepositoryData repositoryData)
    {
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        repositoryData.setValue("emptyData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);

        try
        {
            dataHolder.getType("emptyData");
            Assert.fail("the type of this data can be any type using the boolean internal property, it should not be possible to get its type");
        }
        catch (NotUniqueTypeException e)
        {
            Assert.assertTrue(e.getMessage().contains(ModelItemTypeConstants.LONG_TYPE_ID));
            Assert.assertTrue(e.getMessage().contains(ModelItemTypeConstants.DOUBLE_TYPE_ID));
            Assert.assertTrue(e.getMessage().contains(ModelItemTypeConstants.DATETIME_TYPE_ID));
        }
    }
}
