/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.util.Calendar;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.holder.ModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModelLessComposite;
import org.ametys.plugins.repository.data.holder.group.impl.DefaultModelLessComposite;
import org.ametys.plugins.repository.data.holder.impl.DefaultModelLessDataHolder;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.ComplexRepositoryElementType;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.plugins.repository.data.type.RepositoryModelItemType;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.exception.NotUniqueTypeException;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelLessDataHolder} basic methods implementation test case
 */
@NeedFeatures({
    "test-data/test-data-types",
    "test-data/test-model-less-data-types",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModelLessDataHolderHasAndGetValueTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPointWithIncompatibleTypes;
    
    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;
    
    private static RepositoryElementType<Geocode> _geocodeType;
    private static RepositoryElementType<UserIdentity> _userType;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".TestModelLess");
        _testDataTypeExtensionPointWithIncompatibleTypes = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    @SuppressWarnings("unchecked")
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");
        
        _geocodeType = (RepositoryElementType<Geocode>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID);
        _userType = (RepositoryElementType<UserIdentity>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
    }
    
    /**
     * Test {@link ModelLessDataHolder#getValue(String)} methods' nominal cases
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testGetValueNominalCasesWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testGetValueNominalCases", DATA_NODE_TYPE);
        _testGetValueNominalCases(repositoryData);
    }
    
    /**
     * Test {@link ModelLessDataHolder#getValue(String)} methods' nominal cases
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testGetValueNominalCasesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testGetValueNominalCases", DATA_NODE_TYPE);
        _testGetValueNominalCases(repositoryData);
    }

    private void _testGetValueNominalCases(ModifiableRepositoryData repositoryData)
    {
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        // String
        repositoryData.setValue("stringData", "string value");
        Assert.assertEquals("string value", dataHolder.getValue("stringData"));
        Assert.assertEquals("string value", dataHolder.getValue("stringData", "unused default value"));
        
        // String[]
        String[] multipleStringData = new String[] {"first string value", "second string value"};
        repositoryData.setValues("multipleStringData", multipleStringData);
        Assert.assertArrayEquals(multipleStringData, dataHolder.getValue("multipleStringData"));
        
        // Long
        repositoryData.setValue("longData", 7L);
        Assert.assertEquals(Long.valueOf(7), dataHolder.getValue("longData"));
        
        // Geocode
        _geocodeType.write(repositoryData, "geocodeData", new Geocode(65.215, 2.698));
        Geocode geocodeData = dataHolder.getValue("geocodeData");
        Assert.assertEquals(Double.valueOf(65.215), geocodeData.getLatitude());
        Assert.assertEquals(Double.valueOf(2.698), geocodeData.getLongitude());
        
        // Composite
        ModifiableRepositoryData compositeData = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        Assert.assertNotNull(dataHolder.getValue("composite"));
        ModelLessComposite composite = new DefaultModelLessComposite(_testDataTypeExtensionPoint, compositeData, dataHolder, dataHolder);
        
        // Composite - Boolean
        compositeData.setValue("booleanData", true);
        Assert.assertTrue(dataHolder.getValue("composite/booleanData"));
        Assert.assertTrue(composite.getValue("booleanData"));

        // Composite - User
        UserIdentity user =  new UserIdentity("tmathias", "hinterland");
        _userType.write(compositeData, "userData", user);
        Assert.assertEquals(user, dataHolder.getValue("composite/userData"));
        Assert.assertEquals(user, composite.getValue("userData"));
    }
    
    /**
     * Test {@link ModelLessDataHolder#getValue(String)} methods' nominal cases
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testGetValueOnDataWithManyCompatibleTypesWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testGetValueOnDataWithManyCompatibleTypes", DATA_NODE_TYPE);
        _testGetValueOnDataWithManyCompatibleTypes(repositoryData);
    }
    
    /**
     * Test {@link ModelLessDataHolder#getValue(String)} methods' nominal cases
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testGetValueOnDataWithManyCompatibleTypesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testGetValueOnDataWithManyCompatibleTypes", DATA_NODE_TYPE);
        _testGetValueOnDataWithManyCompatibleTypes(repositoryData);
    }

    @SuppressWarnings("static-access")
    private void _testGetValueOnDataWithManyCompatibleTypes(ModifiableRepositoryData repositoryData)
    {
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPointWithIncompatibleTypes, repositoryData);
        repositoryData.setValue("dateData", DateUtils.asCalendar(LocalDate.of(2019, 2, 6)));
        
        try
        {
            dataHolder.getValue("dateData");
            Assert.fail("the type of this data can be either date or datetime, it should not be possible to get its value");
        }
        catch (NotUniqueTypeException e)
        {
            Assert.assertTrue(e.getMessage().contains(ModelItemTypeConstants.DATE_TYPE_ID));
            Assert.assertTrue(e.getMessage().contains(ModelItemTypeConstants.DATETIME_TYPE_ID));
        }
        
        try
        {
            dataHolder.getValue("dateData", LocalDate.now());
            Assert.fail("the type of this data can be either date or datetime, it should not be possible to get its value");
        }
        catch (NotUniqueTypeException e)
        {
            Assert.assertTrue(e.getMessage().contains(ModelItemTypeConstants.DATE_TYPE_ID));
            Assert.assertTrue(e.getMessage().contains(ModelItemTypeConstants.DATETIME_TYPE_ID));
        }
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#getValue(String)} method on inexisting data
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testGetValueInexistingDataJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testGetValueInexistingData", DATA_NODE_TYPE);
        _testGetValueInexistingData(repositoryData);
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#getValue(String)} method on inexisting data
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testGetValueInexistingDataMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testGetValueInexistingData", DATA_NODE_TYPE);
        _testGetValueInexistingData(repositoryData);
    }
    
    private void _testGetValueInexistingData(ModifiableRepositoryData repositoryData)
    {
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        Assert.assertNull(dataHolder.getValue("inexistingData"));
        Assert.assertEquals("default value", dataHolder.getValue("inexistingData", "default value"));
        
        Assert.assertNull(dataHolder.getValue("inexisting/data/in/group"));
        Assert.assertEquals("default value", dataHolder.getValue("inexisting/data/in/group", "default value"));
        
        Assert.assertNull(dataHolder.getComposite("inexistingComposite"));
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#hasValue(String)} method
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testHasValueJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testHasValue", DATA_NODE_TYPE);
        _testHasValue(repositoryData);
    }
    
    /**
     * Tests the {@link ModelLessDataHolder#hasValue(String)} method
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testHasValueMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testHasValue", DATA_NODE_TYPE);
        _testHasValue(repositoryData);
    }
    
    private void _testHasValue(ModifiableRepositoryData repositoryData)
    {
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        Assert.assertFalse(dataHolder.hasValue("simpleDoubleData"));
        repositoryData.setValue("simpleDoubleData", 25.4);
        Assert.assertTrue(dataHolder.hasValue("simpleDoubleData"));
        
        Assert.assertFalse(dataHolder.hasValue("composite"));
        ModifiableRepositoryData compositeData = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        ModelLessComposite composite = new DefaultModelLessComposite(_testDataTypeExtensionPoint, compositeData, dataHolder, dataHolder);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("composite"));
        Assert.assertFalse(dataHolder.hasValue("composite"));

        Assert.assertFalse(dataHolder.hasValue("composite/multipleStringData"));
        Assert.assertFalse(composite.hasValue("multipleStringData"));
        compositeData.setValues("multipleStringData", new String[]{"string1", "string2"});
        Assert.assertTrue(dataHolder.hasValue("composite/multipleStringData"));
        Assert.assertTrue(composite.hasValue("multipleStringData"));
    }
    
    /**
     * Test {@link ModelLessDataHolder} behavior with empty values
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testEmptyValuesWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testEmptyValues", DATA_NODE_TYPE);
        _testEmptyValues(sourceData);
    }
    
    /**
     * Test {@link ModelLessDataHolder} behavior with empty values
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testEmptyValuesWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testEmptyValues", DATA_NODE_TYPE);
        _testEmptyValues(sourceData);
    }
    
    @SuppressWarnings("static-access")
    private void _testEmptyValues(ModifiableRepositoryData repositoryData)
    {
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);

        // single string data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("singleStringData"));
        Assert.assertNull(dataHolder.getValue("singleStringData"));
        repositoryData.setValue("singleStringData", StringUtils.EMPTY);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleStringData"));
        Assert.assertFalse(dataHolder.hasValue("singleStringData"));
        Assert.assertEquals(StringUtils.EMPTY, dataHolder.getValueOfType("singleStringData", ModelItemTypeConstants.STRING_TYPE_ID));
        Assert.assertEquals("default", dataHolder.getValueOfType("singleStringData", ModelItemTypeConstants.STRING_TYPE_ID, "default"));
        
        // multiple string data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("multipleStringData"));
        Assert.assertNull(dataHolder.getValue("multipleStringData"));
        repositoryData.setValues("multipleStringData", new String[0]);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("multipleStringData"));
        Assert.assertFalse(dataHolder.hasValue("multipleStringData"));
        Assert.assertEquals(0, ((String[]) dataHolder.getValueOfType("multipleStringData", ModelItemTypeConstants.STRING_TYPE_ID)).length);

        // single long data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("singleLongData"));
        Assert.assertNull(dataHolder.getValue("singleLongData"));
        repositoryData.setValue("singleLongData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleLongData"));
        Assert.assertFalse(dataHolder.hasValue("singleLongData"));
        Assert.assertNull(dataHolder.getValue("singleLongData"));
        Assert.assertEquals(Long.valueOf(78), dataHolder.getValue("singleLongData", 78L));
        
        // multiple long data
        Assert.assertFalse(dataHolder.hasValueOrEmpty("multipleLongData"));
        Assert.assertNull(dataHolder.getValue("multipleLongData"));
        repositoryData.setValues("multipleLongData", new Long[0]);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("multipleLongData"));
        Assert.assertFalse(dataHolder.hasValue("multipleLongData"));
        Assert.assertEquals(0, ((Long[]) dataHolder.getValue("multipleLongData")).length);

        // single double data
        Assert.assertFalse(dataHolder.hasValue("singleDoubleData"));
        Assert.assertNull(dataHolder.getValue("singleDoubleData"));
        repositoryData.setValue("singleDoubleData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleDoubleData"));
        Assert.assertFalse(dataHolder.hasValue("singleDoubleData"));
        Assert.assertNull(dataHolder.getValue("singleDoubleData"));
        
        // multiple double data
        Assert.assertFalse(dataHolder.hasValue("multipleDoubleData"));
        Assert.assertNull(dataHolder.getValue("multipleDoubleData"));
        repositoryData.setValues("multipleDoubleData", new Double[0]);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("multipleDoubleData"));
        Assert.assertFalse(dataHolder.hasValue("multipleDoubleData"));
        Assert.assertEquals(0, ((Double[]) dataHolder.getValue("multipleDoubleData")).length);

        // single date time data
        Assert.assertFalse(dataHolder.hasValue("singleDatetimeData"));
        Assert.assertNull(dataHolder.getValue("singleDatetimeData"));
        repositoryData.setValue("singleDatetimeData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleDatetimeData"));
        Assert.assertFalse(dataHolder.hasValue("singleDatetimeData"));
        Assert.assertNull(dataHolder.getValueOfType("singleDatetimeData", ModelItemTypeConstants.DATETIME_TYPE_ID));
        
        // multiple date time data
        Assert.assertFalse(dataHolder.hasValue("multipleDatetimeData"));
        Assert.assertNull(dataHolder.getValue("multipleDatetimeData"));
        repositoryData.setValues("multipleDatetimeData", new Calendar[0]);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("multipleDatetimeData"));
        Assert.assertFalse(dataHolder.hasValue("multipleDatetimeData"));
        Assert.assertEquals(0, ((ZonedDateTime[]) dataHolder.getValueOfType("multipleDatetimeData", ModelItemTypeConstants.DATETIME_TYPE_ID)).length);
        
        // single user data
        Assert.assertFalse(dataHolder.hasValue("singleUserData"));
        Assert.assertNull(dataHolder.getValue("singleUserData"));
        ModifiableRepositoryData userData = repositoryData.addRepositoryData("singleUserData", RepositoryConstants.USER_NODETYPE);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleUserData"));
        Assert.assertFalse(dataHolder.hasValue("singleUserData"));
        Assert.assertNull(dataHolder.getValue("singleUserData"));
        userData.setValue("login", "tyork");
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleUserData"));
        Assert.assertFalse(dataHolder.hasValue("singleUserData"));
        Assert.assertNull(dataHolder.getValue("singleUserData"));
        userData.removeValue("login");
        userData.setValue("population", "radiohead");
        Assert.assertTrue(dataHolder.hasValueOrEmpty("singleUserData"));
        Assert.assertFalse(dataHolder.hasValue("singleUserData"));
        Assert.assertNull(dataHolder.getValue("singleUserData"));
        
        // multiple user data
        Assert.assertFalse(dataHolder.hasValue("multipleUserData"));
        Assert.assertNull(dataHolder.getValue("multipleUserData"));
        ModifiableRepositoryData multipleUserData = repositoryData.addRepositoryData("multipleUserData", RepositoryConstants.MULTIPLE_ITEM_NODETYPE);
        multipleUserData.setValue(ComplexRepositoryElementType.TYPE_ID_DATA_NAME, ModelItemTypeConstants.USER_ELEMENT_TYPE_ID, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("multipleUserData"));
        Assert.assertFalse(dataHolder.hasValue("multipleUserData"));
        Assert.assertEquals(0, ((UserIdentity[]) dataHolder.getValue("multipleUserData")).length);
        ModifiableRepositoryData firstUserData = repositoryData.addRepositoryData("1", RepositoryConstants.USER_NODETYPE);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("multipleUserData"));
        Assert.assertFalse(dataHolder.hasValue("multipleUserData"));
        Assert.assertEquals(0, ((UserIdentity[]) dataHolder.getValue("multipleUserData")).length);
        firstUserData.setValue("login", "tyork");
        Assert.assertTrue(dataHolder.hasValueOrEmpty("multipleUserData"));
        Assert.assertFalse(dataHolder.hasValue("multipleUserData"));
        Assert.assertEquals(0, ((UserIdentity[]) dataHolder.getValue("multipleUserData")).length);
        
        // composite data
        Assert.assertFalse(dataHolder.hasValue("composite"));
        Assert.assertNull(dataHolder.getValue("composite"));
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        Assert.assertTrue(dataHolder.hasValueOrEmpty("composite"));
        Assert.assertFalse(dataHolder.hasValue("composite"));
        Assert.assertTrue(((ModelLessComposite) dataHolder.getValue("composite")).getDataNames().isEmpty());
    }
}
