/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.holder.ModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModelLessComposite;
import org.ametys.plugins.repository.data.holder.impl.DefaultModelLessDataHolder;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelLessDataHolder} toJSON methods implementation test case
 */
@NeedFeatures({
    "test-data/test-model-less-data-types",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModelLessDataHoldertoJSONTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __DATA_NODE_TYPE = "ametys:unstructured";
    private static final DataContext __EMPTY_DATA_CONTEXT = DataContext.newInstance();
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    
    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;
    
    private static RepositoryElementType<Geocode> _geocodeType;
    private static RepositoryElementType<UserIdentity> _userType;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".TestModelLess");
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    @SuppressWarnings("unchecked")
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");
        
        _geocodeType = (RepositoryElementType<Geocode>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID);
        _userType = (RepositoryElementType<UserIdentity>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
    }
    
    /**
     * Test {@link ModelLessDataHolder#dataToJSON(DataContext)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHoldertoJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHoldertoJSON", __DATA_NODE_TYPE);
        _testDataHolderToJSON(sourceData);
    }
    
    /**
     * Test {@link ModelLessDataHolder#dataToJSON(DataContext)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderToJSONWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHoldertoJSON", __DATA_NODE_TYPE);
        _testDataHolderToJSON(sourceData);
    }
    
    private void _testDataHolderToJSON(ModifiableRepositoryData repositoryData) throws Exception
    {
        _populateRepositoryData(repositoryData);
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        // Convert the data holder to JSON
        Map<String, Object> json = dataHolder.dataToJSON(__EMPTY_DATA_CONTEXT);
        
        // Check the json result
        Assert.assertEquals(5, json.size());
        
        Assert.assertEquals("value of the single string data", json.get("singleStringData"));
        Assert.assertEquals(List.of(32L, 784L), json.get("multipleLongData"));
        Assert.assertEquals(Map.of("latitude", 85.487D, "longitude", 6.254D), json.get("singleGeocodeData"));
        
        @SuppressWarnings("unchecked")
        List<Map<String, Object>> users = (List<Map<String, Object>>) json.get("multipleUserData");
        Map<String, Object> user1 = users.get(0);
        Assert.assertEquals("user1", user1.get("login"));
        Assert.assertEquals(POPULATION_TEST_ID, user1.get("populationId"));
        Assert.assertEquals("User One", user1.get("fullname"));
        Map<String, Object> user2 = users.get(1);
        Assert.assertEquals("user3", user2.get("login"));
        Assert.assertEquals(POPULATION_TEST_ID, user2.get("populationId"));
        Assert.assertEquals("User Three", user2.get("fullname"));
        
        @SuppressWarnings("unchecked")
        Map<String, Object> composite = (Map<String, Object>) json.get("composite");
        Assert.assertEquals(2, composite.size());
        Assert.assertEquals("value of the single string data in the composite", composite.get("singleStringData"));
        Assert.assertTrue((boolean) composite.get("singleBooleanData"));
    }
    
    /**
     * Test {@link ModelLessDataHolder#dataToJSON(String, DataContext)} implementation
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testDataHolderSingleDataToJSONWithJCRRepositoryData()
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSingleDataToJSON", __DATA_NODE_TYPE);
        _testDataHolderSingleDataToJSON(sourceData);
    }
    
    /**
     * Test {@link ModelLessDataHolder#dataToJSON(String, DataContext)} implementation
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testDataHolderSingleDataToJSONWithMemoryRepositoryData()
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSingleDataToJSON", __DATA_NODE_TYPE);
        _testDataHolderSingleDataToJSON(sourceData);
    }
    
    private void _testDataHolderSingleDataToJSON(ModifiableRepositoryData repositoryData)
    {
        _populateRepositoryData(repositoryData);
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        // Convert the data holder to JSON
        Object json = dataHolder.dataToJSON("composite/singleStringData", __EMPTY_DATA_CONTEXT);
        
        // Check the json result
        Assert.assertEquals("value of the single string data in the composite", json);
    }
    
    /**
     * Test {@link ModelLessDataHolder#dataToJSON(String, DataContext)} implementation with a composite
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testDataHolderSingleCompositeToJSONWithJCRRepositoryData()
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSingleCompositeToJSON", __DATA_NODE_TYPE);
        _testDataHolderSingleCompositeToJSON(sourceData);
    }
    
    /**
     * Test {@link ModelLessDataHolder#dataToJSON(String, DataContext)} implementation with a composite
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testDataHolderSingleCompositeToJSONWithMemoryRepositoryData()
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSingleCompositeToJSON", __DATA_NODE_TYPE);
        _testDataHolderSingleCompositeToJSON(sourceData);
    }
    
    private void _testDataHolderSingleCompositeToJSON(ModifiableRepositoryData repositoryData)
    {
        _populateRepositoryData(repositoryData);
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        // Convert the data holder to JSON
        Object json = dataHolder.dataToJSON("composite", __EMPTY_DATA_CONTEXT);
        
        // Check the json result
        @SuppressWarnings("unchecked")
        Map<String, Object> composite = (Map<String, Object>) json;
        Assert.assertEquals(2, composite.size());
        Assert.assertEquals("value of the single string data in the composite", composite.get("singleStringData"));
        Assert.assertTrue((boolean) composite.get("singleBooleanData"));
    }
    
    /**
     * Test {@link ModelLessComposite#dataToJSON(DataContext)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeToJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testCompositeToJSON", __DATA_NODE_TYPE);
        _testCompositeToJSON(sourceData);
    }
    
    /**
     * Test {@link ModelLessComposite#dataToJSON(DataContext)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeToJSONWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testCompositeToJSON", __DATA_NODE_TYPE);
        _testCompositeToJSON(sourceData);
    }
    
    private void _testCompositeToJSON(ModifiableRepositoryData repositoryData) throws Exception
    {
        _populateRepositoryData(repositoryData);
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        ModelLessComposite composite = dataHolder.getComposite("composite");
        
        // Convert the composite to JSON
        Map<String, Object> json = composite.dataToJSON(__EMPTY_DATA_CONTEXT);
        
        // Check the json result
        Assert.assertEquals(2, json.size());
        Assert.assertEquals("value of the single string data in the composite", json.get("singleStringData"));
        Assert.assertTrue((boolean) json.get("singleBooleanData"));
    }
    
    /**
     * Test {@link ModelLessComposite#dataToJSON(String, DataContext)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeSingleDataToJSONWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testCompositeSingleDataToJSON", __DATA_NODE_TYPE);
        _testCompositeSingleDataToJSON(sourceData);
    }
    
    /**
     * Test {@link ModelLessComposite#dataToJSON(String, DataContext)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeSingleDataToJSONWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testCompositeSingleDataToJSON", __DATA_NODE_TYPE);
        _testCompositeSingleDataToJSON(sourceData);
    }
    
    private void _testCompositeSingleDataToJSON(ModifiableRepositoryData repositoryData) throws Exception
    {
        _populateRepositoryData(repositoryData);
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        ModelLessComposite composite = dataHolder.getComposite("composite");
        
        // Convert the data holder to JSON
        Object json = composite.dataToJSON("singleStringData", __EMPTY_DATA_CONTEXT);
        
        // Check the json result
        Assert.assertEquals("value of the single string data in the composite", json);
    }
    
    private void _populateRepositoryData(ModifiableRepositoryData repositoryData)
    {
        repositoryData.setValue("singleStringData", "value of the single string data");
        repositoryData.setValues("multipleLongData", new Long[] {32L, 784L});
        
        _geocodeType.write(repositoryData, "singleGeocodeData", new Geocode(85.487D, 6.254D));
        _userType.write(repositoryData, "multipleUserData", new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY});

        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        composite.setValue("singleStringData", "value of the single string data in the composite");
        composite.setValue("singleBooleanData", true);
    }
}
