/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.TransformerHandler;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.xpath.XPathAPI;
import org.junit.Assert;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.ContentHandler;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.holder.ModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModelLessComposite;
import org.ametys.plugins.repository.data.holder.impl.DefaultModelLessDataHolder;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModelLessDataHolder} toSAX methods implementation test case
 */
@NeedFeatures({
    "test-data/test-model-less-data-types",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModelLessDataHoldertoSAXTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __DATA_NODE_TYPE = "ametys:unstructured";
    private static final DataContext __EMPTY_DATA_CONTEXT = DataContext.newInstance();
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    
    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;
    
    private static RepositoryElementType<Geocode> _geocodeType;
    private static RepositoryElementType<UserIdentity> _userType;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".TestModelLess");
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    @SuppressWarnings("unchecked")
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");
        
        _geocodeType = (RepositoryElementType<Geocode>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID);
        _userType = (RepositoryElementType<UserIdentity>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
    }
    
    /**
     * Test {@link ModelLessDataHolder#dataToSAX(ContentHandler, DataContext)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSax", __DATA_NODE_TYPE);
        _testDataHolderSax(sourceData);
    }
    
    /**
     * Test {@link ModelLessDataHolder#dataToSAX(ContentHandler, DataContext)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSax", __DATA_NODE_TYPE);
        _testDataHolderSax(sourceData);
    }
    
    private void _testDataHolderSax(ModifiableRepositoryData repositoryData) throws Exception
    {
        _populateRepositoryData(repositoryData);
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler, __EMPTY_DATA_CONTEXT);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/singleStringData)", 1);
        _assertXpathEquals(document, "/result/singleStringData/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/singleStringData/@multiple", "false");
        _assertXpathEquals(document, "/result/singleStringData", "value of the single string data");
        _assertXpathEquals(document, "count(/result/multipleLongData)", 2);
        _assertXpathEquals(document, "/result/multipleLongData[1]/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/multipleLongData[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/multipleLongData[1]", 32);
        _assertXpathEquals(document, "/result/multipleLongData[2]/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.LONG_TYPE_ID);
        _assertXpathEquals(document, "/result/multipleLongData[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/multipleLongData[2]", 784);
        
        _assertXpathEquals(document, "count(/result/singleGeocodeData)", 1);
        _assertXpathEquals(document, "/result/singleGeocodeData/@typeId", ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/singleGeocodeData/@multiple", "false");
        _assertXpathEquals(document, "/result/singleGeocodeData/@latitude", 85.487);
        _assertXpathEquals(document, "/result/singleGeocodeData/@longitude", 6.254);
        
        _assertXpathEquals(document, "count(/result/multipleUserData)", 2);
        _assertXpathEquals(document, "/result/multipleUserData[1]/@typeId", ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/multipleUserData[1]/@multiple", "true");
        _assertXpathEquals(document, "/result/multipleUserData[1]/@login", "user1");
        _assertXpathEquals(document, "/result/multipleUserData[1]/@populationId", POPULATION_TEST_ID);
        _assertXpathEquals(document, "/result/multipleUserData[1]", "User One");

        _assertXpathEquals(document, "/result/multipleUserData[2]/@typeId", ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _assertXpathEquals(document, "/result/multipleUserData[2]/@multiple", "true");
        _assertXpathEquals(document, "/result/multipleUserData[2]/@login", "user3");
        _assertXpathEquals(document, "/result/multipleUserData[2]/@populationId", POPULATION_TEST_ID);
        _assertXpathEquals(document, "/result/multipleUserData[2]", "User Three");
        
        _assertXpathEquals(document, "count(/result/composite)", 1);
        _assertXpathEquals(document, "count(/result/composite/*)", 2);
        _assertXpathEquals(document, "count(/result/composite/singleStringData)", 1);
        _assertXpathEquals(document, "/result/composite/singleStringData/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.STRING_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/singleStringData/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/singleStringData", "value of the single string data in the composite");
        _assertXpathEquals(document, "count(/result/composite/singleBooleanData)", 1);
        _assertXpathEquals(document, "/result/composite/singleBooleanData/@typeId", org.ametys.runtime.model.type.ModelItemTypeConstants.BOOLEAN_TYPE_ID);
        _assertXpathEquals(document, "/result/composite/singleBooleanData/@multiple", "false");
        _assertXpathEquals(document, "/result/composite/singleBooleanData", true);
    }
    
    /**
     * Test {@link ModelLessDataHolder#dataToSAX(ContentHandler, String, DataContext)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleDataWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSaxSingleData", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleData(sourceData);
    }
    
    /**
     * Test {@link ModelLessDataHolder#dataToSAX(ContentHandler, String, DataContext)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleDataWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSaxSingleData", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleData(sourceData);
    }
    
    private void _testDataHolderSaxSingleData(ModifiableRepositoryData repositoryData) throws Exception
    {
        _populateRepositoryData(repositoryData);
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler, "composite/singleStringData", __EMPTY_DATA_CONTEXT);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");

        _assertXpathEquals(document, "count(/result/*)", 1);
        _assertXpathEquals(document, "count(/result/singleStringData)", 1);
        _assertXpathEquals(document, "/result/singleStringData", "value of the single string data in the composite");
    }
    
    /**
     * Test {@link ModelLessDataHolder#dataToSAX(ContentHandler, String, DataContext)} implementation with a composite
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleCompositeWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testDataHolderSaxSingleComposite", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleComposite(sourceData);
    }
    
    /**
     * Test {@link ModelLessDataHolder#dataToSAX(ContentHandler, String, DataContext)} implementation with a composite
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testDataHolderSaxSingleCompositeWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testDataHolderSaxSingleComposite", __DATA_NODE_TYPE);
        _testDataHolderSaxSingleComposite(sourceData);
    }
    
    private void _testDataHolderSaxSingleComposite(ModifiableRepositoryData repositoryData) throws Exception
    {
        _populateRepositoryData(repositoryData);
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        dataHolder.dataToSAX(handler, "composite", __EMPTY_DATA_CONTEXT);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 1);
        _assertXpathEquals(document, "count(/result/composite)", 1);
        _assertXpathEquals(document, "count(/result/composite/*)", 2);
        _assertXpathEquals(document, "count(/result/composite/singleStringData)", 1);
        _assertXpathEquals(document, "/result/composite/singleStringData", "value of the single string data in the composite");
        _assertXpathEquals(document, "count(/result/composite/singleBooleanData)", 1);
        _assertXpathEquals(document, "/result/composite/singleBooleanData", true);
    }
    
    /**
     * Test {@link ModelLessComposite#dataToSAX(ContentHandler, DataContext)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeSaxWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testCompositeSax", __DATA_NODE_TYPE);
        _testCompositeSax(sourceData);
    }
    
    /**
     * Test {@link ModelLessComposite#dataToSAX(ContentHandler, DataContext)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeSaxWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testCompositeSax", __DATA_NODE_TYPE);
        _testCompositeSax(sourceData);
    }
    
    private void _testCompositeSax(ModifiableRepositoryData repositoryData) throws Exception
    {
        _populateRepositoryData(repositoryData);
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        ModelLessComposite composite = dataHolder.getComposite("composite");
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        composite.dataToSAX(handler, __EMPTY_DATA_CONTEXT);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 2);
        _assertXpathEquals(document, "count(/result/singleStringData)", 1);
        _assertXpathEquals(document, "/result/singleStringData", "value of the single string data in the composite");
        _assertXpathEquals(document, "count(/result/singleBooleanData)", 1);
        _assertXpathEquals(document, "/result/singleBooleanData", true);
    }
    
    /**
     * Test {@link ModelLessComposite#dataToSAX(ContentHandler, String, DataContext)} implementation
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeSaxSingleDataWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _jcrRoot.addRepositoryData("testCompositeSaxSingleData", __DATA_NODE_TYPE);
        _testCompositeSaxSingleData(sourceData);
    }
    
    /**
     * Test {@link ModelLessComposite#dataToSAX(ContentHandler, String, DataContext)} implementation
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testCompositeSaxSingleDataWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData sourceData = _memoryRoot.addRepositoryData("testCompositeSaxSingleData", __DATA_NODE_TYPE);
        _testCompositeSaxSingleData(sourceData);
    }
    
    private void _testCompositeSaxSingleData(ModifiableRepositoryData repositoryData) throws Exception
    {
        _populateRepositoryData(repositoryData);
        ModelLessDataHolder dataHolder = new DefaultModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        ModelLessComposite composite = dataHolder.getComposite("composite");
        
        // Saxes the data holder
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        composite.dataToSAX(handler, "singleStringData", __EMPTY_DATA_CONTEXT);
        
        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        // Check the sax result
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        _assertXpathEquals(document, "count(/result/*)", 1);
        _assertXpathEquals(document, "count(/result/singleStringData)", 1);
        _assertXpathEquals(document, "/result/singleStringData", "value of the single string data in the composite");
    }
    
    private void _populateRepositoryData(ModifiableRepositoryData repositoryData)
    {
        repositoryData.setValue("singleStringData", "value of the single string data");
        repositoryData.setValues("multipleLongData", new Long[] {32L, 784L});
        
        _geocodeType.write(repositoryData, "singleGeocodeData", new Geocode(85.487D, 6.254D));
        _userType.write(repositoryData, "multipleUserData", new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY});

        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        composite.setValue("singleStringData", "value of the single string data in the composite");
        composite.setValue("singleBooleanData", true);
    }
    
    private void _assertXpathEquals(Node node, String expression, String value) throws Exception
    {
        Assert.assertEquals(value, XPathAPI.eval(node, expression).str());
    }
    
    private void _assertXpathEquals(Node node, String expression, long value) throws Exception
    {
        Assert.assertEquals(value, (long) XPathAPI.eval(node, expression).num());
    }
    
    private void _assertXpathEquals(Node node, String expression, double value) throws Exception
    {
        Assert.assertEquals(Double.valueOf(value), Double.valueOf(XPathAPI.eval(node, expression).num()));
    }
    
    private void _assertXpathEquals(Node node, String expression, boolean value) throws Exception
    {
        Assert.assertEquals(value, XPathAPI.eval(node, expression).bool());
    }
}
