/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.holder.impl.DefaultModelAwareDataHolder;
import org.ametys.cms.data.holder.impl.DefaultModifiableModelAwareDataHolder;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.ModifiableModelAwareDataHolder;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.exception.BadDataPathCardinalityException;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModifiableModelAwareDataHolder} implementation test case
 * Tests the results of {@link ModifiableModelAwareDataHolder#setValue(String, Object)} and {@link ModifiableModelAwareDataHolder#getValue(String)}
 * when the cardinality is different between the {@link Model} and the {@link RepositoryData} stored
 */
@NeedFeatures({
    "test-data/test-model",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModifiableModelAwareDataHolderCardinalityTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    
    private static Model _model;
    private static AmetysObjectResolver _ametysObjectResolver;
    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for a single data with a single value
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testGetValueOnSingleDataStoredAsASingleDataWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testGetValueOnSingleDataStoredAsASingleData", DATA_NODE_TYPE);
        _testGetValueOnSingleDataStoredAsASingleData(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for a single data with a single value
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testGetValueOnSingleDataStoredAsASingleDataWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testGetValueOnSingleDataStoredAsASingleData", DATA_NODE_TYPE);
        _testGetValueOnSingleDataStoredAsASingleData(repositoryData);
    }
    
    private void _testGetValueOnSingleDataStoredAsASingleData(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);

        // Add a single string value to the single string data
        repositoryData.setValue("singleStringData", "single string value");
        
        // Check that the value is a single string
        Assert.assertEquals("single string value", dataHolder.getValue("singleStringData"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for a single data with a multiple value with only one entry
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testGetValueOnSingleDataStoredAsAMultipleDataWithOnlyOneValueWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testGetValueOnSingleDataStoredAsAMultipleDataWithOnlyOneValue", DATA_NODE_TYPE);
        _testGetValueOnSingleDataStoredAsAMultipleDataWithOnlyOneValue(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for a single data with a multiple value with only one entry
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testGetValueOnSingleDataStoredAsAMultipleDataWithOnlyOneValueWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testGetValueOnSingleDataStoredAsAMultipleDataWithOnlyOneValue", DATA_NODE_TYPE);
        _testGetValueOnSingleDataStoredAsAMultipleDataWithOnlyOneValue(repositoryData);
    }

    private void _testGetValueOnSingleDataStoredAsAMultipleDataWithOnlyOneValue(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);

        // Add a lonely multiple string value to the single string data
        repositoryData.setValues("singleStringData", new String[] {"lonely multiple string value"});
        
        // Check that the value is single string
        Assert.assertEquals("lonely multiple string value", dataHolder.getValue("singleStringData"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for a single data with a multiple value with many entries
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testGetValueOnSingleDataStoredAsAMultipleDataWithManyValuesWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testGetValueOnSingleDataStoredAsAMultipleDataWithManyValues", DATA_NODE_TYPE);
        _testGetValueOnSingleDataStoredAsAMultipleDataWithManyValues(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for a single data with a multiple value with many entries
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testGetValueOnSingleDataStoredAsAMultipleDataWithManyValuesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testGetValueOnSingleDataStoredAsAMultipleDataWithManyValues", DATA_NODE_TYPE);
        _testGetValueOnSingleDataStoredAsAMultipleDataWithManyValues(repositoryData);
    }
    
    private void _testGetValueOnSingleDataStoredAsAMultipleDataWithManyValues(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);

        // Add multiple string values to the single string data
        repositoryData.setValues("singleStringData", new String[] {"first multiple string value", "second multiple string value"});
        
        // Check that the value is a single string
        Assert.assertEquals("first multiple string value", dataHolder.getValue("singleStringData"));
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for a multiple data with a single value
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testGetValueOnMultipleDataStoredAsASingleDataWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testGetValueOnMultipleDataStoredAsASingleData", DATA_NODE_TYPE);
        _testGetValueOnMultipleDataStoredAsASingleData(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for a multiple data with a single value
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testGetValueOnMultipleDataStoredAsASingleDataWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testGetValueOnMultipleDataStoredAsASingleData", DATA_NODE_TYPE);
        _testGetValueOnMultipleDataStoredAsASingleData(repositoryData);
    }
    
    private void _testGetValueOnMultipleDataStoredAsASingleData(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);

        // Add a single string value to the multiple string data
        repositoryData.setValue("multipleStringData", "single string value");
        
        // Check that the value is an array with the single string
        String[] multipleStringData = dataHolder.getValue("multipleStringData");
        Assert.assertEquals(1, multipleStringData.length);
        Assert.assertEquals("single string value", multipleStringData[0]);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for a multiple data with a multiple value
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testGetValueOnMultipleDataStoredAsAMultipleDataWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testGetValueOnMultipleDataStoredAsAMultipleData", DATA_NODE_TYPE);
        _testGetValueOnMultipleDataStoredAsAMultipleData(repositoryData);
    }
    
    /**
     * Test {@link ModelAwareDataHolder#getValue(String)} method for a multiple data with a multiple value
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testGetValueOnMultipleDataStoredAsAMultipleDataWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testGetValueOnMultipleDataStoredAsAMultipleData", DATA_NODE_TYPE);
        _testGetValueOnMultipleDataStoredAsAMultipleData(repositoryData);
    }
    
    private void _testGetValueOnMultipleDataStoredAsAMultipleData(ModifiableRepositoryData repositoryData)
    {
        ModelAwareDataHolder dataHolder = new DefaultModelAwareDataHolder(repositoryData, _model);

        // Add multiple string values to the multiple string data
        repositoryData.setValues("multipleStringData", new String[] {"first multiple string value", "second multiple string value"});
        
        // Check that the value is an array with the string values
        String[] multipleStringData = dataHolder.getValue("multipleStringData");
        Assert.assertEquals(2, multipleStringData.length);
        Assert.assertEquals("first multiple string value", multipleStringData[0]);
        Assert.assertEquals("second multiple string value", multipleStringData[1]);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for a single data with a single value
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetValueOnSingleDataWithASingleDataWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueOnSingleDataWithASingleData", DATA_NODE_TYPE);
        _testSetValueOnSingleDataWithASingleData(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for a single data with a single value
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueOnSingleDataWithASingleDataWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueOnSingleDataWithASingleData", DATA_NODE_TYPE);
        _testSetValueOnSingleDataWithASingleData(repositoryData);
    }
    
    private void _testSetValueOnSingleDataWithASingleData(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Set a single string value to the single string data
        dataHolder.setValue("singleStringData", "single string value");
        
        // Check that stored value is single
        Assert.assertFalse(repositoryData.isMultiple("singleStringData"));
        Assert.assertEquals("single string value", repositoryData.getString("singleStringData"));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for a single data with a multiple value with only one entry
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetValueOnSingleDataWithAMultipleDataWithOnlyOneValueWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueOnSingleDataWithAMultipleDataWithOnlyOneValue", DATA_NODE_TYPE);
        _testSetValueOnSingleDataWithAMultipleDataWithOnlyOneValue(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for a single data with a multiple value with only one entry
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueOnSingleDataWithAMultipleDataWithOnlyOneValueWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueOnSingleDataWithAMultipleDataWithOnlyOneValue", DATA_NODE_TYPE);
        _testSetValueOnSingleDataWithAMultipleDataWithOnlyOneValue(repositoryData);
    }

    private void _testSetValueOnSingleDataWithAMultipleDataWithOnlyOneValue(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Set a lonely but multiple string value to the single string data
        dataHolder.setValue("singleStringData", new String[] {"lonely multiple string value"});
        
        // Check that stored value is single
        Assert.assertFalse(repositoryData.isMultiple("singleStringData"));
        Assert.assertEquals("lonely multiple string value", repositoryData.getString("singleStringData"));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for a single data with an empty multiple value
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetValueOnSingleDataWithAMultipleDataEmptyWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueOnSingleDataWithAMultipleDataEmpty", DATA_NODE_TYPE);
        _testSetValueOnSingleDataWithAMultipleDataEmpty(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for a single data with an empty multiple value
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueOnSingleDataWithAMultipleDataEmptyWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueOnSingleDataWithAMultipleDataEmpty", DATA_NODE_TYPE);
        _testSetValueOnSingleDataWithAMultipleDataEmpty(repositoryData);
    }

    private void _testSetValueOnSingleDataWithAMultipleDataEmpty(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Set a lonely but multiple string value to the single string data
        dataHolder.setValue("singleStringData", new String[0]);
        
        // Check that stored value is single and empty
        Assert.assertFalse(repositoryData.isMultiple("singleStringData"));
        Assert.assertEquals(StringUtils.EMPTY, repositoryData.getString("singleStringData"));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for a single data with a multiple value with many entries
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetValueOnSingleDataWithAMultipleDataWithManyValuesWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueOnSingleDataWithAMultipleDataWithManyValues", DATA_NODE_TYPE);
        _testSetValueOnSingleDataWithAMultipleDataWithManyValues(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for a single data with a multiple value with many entries
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueOnSingleDataWithAMultipleDataWithManyValuesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueOnSingleDataWithAMultipleDataWithManyValues", DATA_NODE_TYPE);
        _testSetValueOnSingleDataWithAMultipleDataWithManyValues(repositoryData);
    }
    
    private void _testSetValueOnSingleDataWithAMultipleDataWithManyValues(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        try
        {
            // Set  multiple string values to the single string data
            dataHolder.setValue("singleStringData", new String[] {"first multiple string value", "second multiple string value"});
            Assert.fail("It should not be possible to set a multiple values instead of a single one");
        }
        catch (IllegalArgumentException e)
        {
            Assert.assertTrue(e.getMessage().contains("multiple"));
        }
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for a multiple data with a single value
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetValueOnMultipleDataWithASingleDataWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueOnMultipleDataWithASingleData", DATA_NODE_TYPE);
        _testSetValueOnMultipleDataWithASingleData(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for a multiple data with a single value
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueOnMultipleDataWithASingleDataWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueOnMultipleDataWithASingleData", DATA_NODE_TYPE);
        _testSetValueOnMultipleDataWithASingleData(repositoryData);
    }
    
    private void _testSetValueOnMultipleDataWithASingleData(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Set a single string value to the multiple string data
        dataHolder.setValue("multipleStringData", "single string value");
        
        // Check that stored value is multiple
        Assert.assertTrue(repositoryData.isMultiple("multipleStringData"));
        String[] multipleStringData = repositoryData.getStrings("multipleStringData");
        Assert.assertEquals(1, multipleStringData.length);
        Assert.assertEquals("single string value", multipleStringData[0]);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for a multiple data with a multiple value
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetValueOnMultipleDataWithAMultipleDataWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueOnMultipleDataWithAMultipleData", DATA_NODE_TYPE);
        _testSetValueOnMultipleDataWithAMultipleData(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for a multiple data with a multiple value
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueOnMultipleDataWithAMultipleDataWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueOnMultipleDataWithAMultipleData", DATA_NODE_TYPE);
        _testSetValueOnMultipleDataWithAMultipleData(repositoryData);
    }
    
    private void _testSetValueOnMultipleDataWithAMultipleData(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Set multiple string values to the multiple string data
        dataHolder.setValue("multipleStringData", new String[] {"first multiple string value", "second multiple string value"});
        
        // Check that stored value is multiple
        Assert.assertTrue(repositoryData.isMultiple("multipleStringData"));
        String[] multipleStringData = repositoryData.getStrings("multipleStringData");
        Assert.assertEquals(2, multipleStringData.length);
        Assert.assertEquals("first multiple string value", multipleStringData[0]);
        Assert.assertEquals("second multiple string value", multipleStringData[1]);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder} methods with multiple content data
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testMultipleContentDataWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testMultipleContentData", DATA_NODE_TYPE);
        _testMultipleContentData(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder} methods with multiple content data
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testMultipleContentDataWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testMultipleContentData", DATA_NODE_TYPE);
        _testMultipleContentData(repositoryData);
    }
    
    private void _testMultipleContentData(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // The content data is multiple
        Assert.assertTrue(dataHolder.isMultiple("multipleContentData"));
        
        try
        {
            dataHolder.hasValue("multipleContentData/title");
            Assert.fail("data holder should not be able to manage data path with a multiple data that is not the last segment of the path");
        }
        catch (BadDataPathCardinalityException e)
        {
            Assert.assertTrue(e.getMessage().contains("multiple"));
        }
        
        try
        {
            dataHolder.getValue("multipleContentData/title");
            Assert.fail("data holder should not be able to manage data path with a multiple data that is not the last segment of the path");
        }
        catch (BadDataPathCardinalityException e)
        {
            Assert.assertTrue(e.getMessage().contains("multiple"));
        }
        
        try
        {
            dataHolder.setValue("multipleContentData/title", "a title");
            Assert.fail("data holder should not be nable to manage data path with a multiple data that is not the last segment of the path");
        }
        catch (BadDataPathCardinalityException e)
        {
            Assert.assertTrue(e.getMessage().contains("multiple"));
        }
        
        try
        {
            dataHolder.removeValue("multipleContentData/title");
            Assert.fail("data holder should not be able to manage data path with a multiple data that is not the last segment of the path");
        }
        catch (BadDataPathCardinalityException e)
        {
            Assert.assertTrue(e.getMessage().contains("multiple"));
        }
    }
}
