/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.io.IOException;
import java.io.OutputStream;
import java.io.StringWriter;
import java.time.ZoneId;
import java.time.ZonedDateTime;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.io.IOUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Binary;
import org.ametys.cms.data.RichText;
import org.ametys.cms.data.holder.impl.DefaultModifiableModelAwareDataHolder;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.external.ExternalizableDataProvider.ExternalizableDataStatus;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.ModifiableModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareComposite;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModifiableModelAwareDataHolder} methods writing externalizable data test cases
 */
@NeedFeatures({
    "test-data/test-model",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModifiableModelAwareDataHolderExternalizableDataTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    private static final ZonedDateTime __RESOURCE_LAST_MODIFIED = ZonedDateTime.of(2020, 02, 03, 15, 01, 58, 0, ZoneId.systemDefault());
    
    private static Model _model;
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    
    private static JCRRepositoryData _root;
    private static RepositoryElementType<Binary> _binaryType;
    private static RepositoryElementType<RichText> _richTextType;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public void prepareComponents() throws Exception
    {
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _root = new JCRRepositoryData(rootAmetysObject.getNode());
        
        _binaryType = (RepositoryElementType<Binary>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _richTextType = (RepositoryElementType<RichText>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID);
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setLocalValue(String, Object)} on unvalued data
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetLocalValueAndSetStatusToLocalOnNotValuedData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetLocalValueAndSetStatusToLocalOnNotValuedData", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Set local value - string
        dataHolder.setStatus("singleStringData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("singleStringData", "new local string value");
        
        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals("new local string value", repositoryData.getString("singleStringData"));
        _assertStringDataIsEmpty(repositoryData, "singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Set local value - multiple long
        dataHolder.setStatus("multipleLongData", ExternalizableDataStatus.LOCAL);
        Long[] longArray = new Long[]{78L, 1L};
        dataHolder.setLocalValue("multipleLongData", longArray);
        
        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        Assert.assertArrayEquals(longArray, repositoryData.getLongs("multipleLongData"));
        _assertMultipleLongDataIsEmpty(repositoryData, "multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Set local value - binary
        dataHolder.setStatus("singleBinaryData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("singleBinaryData", _createBinary("new local binary content"));
        
        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);
        _assertBinaryContent("new local binary content", repositoryData, "singleBinaryData");
        _assertBinaryDataIsEmpty(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Set local value - rich text
        dataHolder.setStatus("singleRichTextData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("singleRichTextData", _createRichText("new local rich text content"));
        
        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);
        _assertRichTextContent("new local rich text content", repositoryData, "singleRichTextData");
        _assertRichTextDataIsEmpty(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Set local value - composite
        dataHolder.setStatus("composite", ExternalizableDataStatus.LOCAL);
        ModifiableModelAwareComposite composite = dataHolder.getLocalComposite("composite", true);
        composite.setValue("singleLongData", 42L);
        
        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals(Long.valueOf(42), repositoryData.getRepositoryData("composite").getLong("singleLongData"));
        _assertRepositoryDataIsEmpty(repositoryData, "composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Set local value - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.getRepositoryData("composite")
                .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
                .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
                .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        dataHolder.setStatus("composite/repeater[1]/subcomposite/singleStringData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("composite/repeater[1]/subcomposite/singleStringData", "new local string value");
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals("new local string value", subcomposite.getString("singleStringData"));
        _assertStringDataIsEmpty(subcomposite, "singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setLocalValue(String, Object)} on unvalued data
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetLocalValueAndSetStatusToExternalOnNotValuedData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetLocalValueAndSetStatusToExternalOnNotValuedData", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Set local value - string
        dataHolder.setStatus("singleStringData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("singleStringData", "new local string value");
        
        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals("new local string value", repositoryData.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        _assertStringDataIsEmpty(repositoryData, "singleStringData");
        
        // Set local value - multiple long
        dataHolder.setStatus("multipleLongData", ExternalizableDataStatus.EXTERNAL);
        Long[] longArray = new Long[]{78L, 1L};
        dataHolder.setLocalValue("multipleLongData", longArray);
        
        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertArrayEquals(longArray, repositoryData.getLongs("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        _assertMultipleLongDataIsEmpty(repositoryData, "multipleLongData");
        
        // Set local value - binary
        dataHolder.setStatus("singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("singleBinaryData", _createBinary("new local binary content"));
        
        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        _assertBinaryContent("new local binary content", repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertBinaryDataIsEmpty(repositoryData, "singleBinaryData");
        
        // Set local value - rich text
        dataHolder.setStatus("singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("singleRichTextData", _createRichText("new local rich text content"));
        
        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        _assertRichTextContent("new local rich text content", repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertRichTextDataIsEmpty(repositoryData, "singleRichTextData");
        
        // Set local value - composite
        dataHolder.setStatus("composite", ExternalizableDataStatus.EXTERNAL);
        ModifiableModelAwareComposite composite = dataHolder.getLocalComposite("composite", true);
        composite.setValue("singleLongData", 42L);
        
        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals(Long.valueOf(42L), repositoryData.getRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX).getLong("singleLongData"));
        _assertRepositoryDataIsEmpty(repositoryData, "composite");
        
        // Set local value - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE)
                .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
                .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
                .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        dataHolder.setStatus("composite/repeater[1]/subcomposite/singleStringData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("composite/repeater[1]/subcomposite/singleStringData", "new local string value");
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals("new local string value", subcomposite.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        _assertStringDataIsEmpty(subcomposite, "singleStringData");
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setLocalValue(String, Object)} on data without status
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetLocalValueAndSetStatusToLocalOnDataWithoutStatus() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetLocalValueAndSetStatusToLocalOnDataWithoutStatus", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData", "string value");

        // Set local value - string
        dataHolder.setStatus("singleStringData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("singleStringData", "new local string value");

        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals("new local string value", repositoryData.getString("singleStringData"));
        _assertStringDataIsEmpty(repositoryData, "singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);

        // Write - multiple long
        Long[] longArray = new Long[]{78L, 1L};
        repositoryData.setValues("multipleLongData", longArray);

        // Set local value - multiple long
        dataHolder.setStatus("multipleLongData", ExternalizableDataStatus.LOCAL);
        Long[] newLongArray = new Long[]{478L, 13L};
        dataHolder.setLocalValue("multipleLongData", newLongArray);

        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        Assert.assertArrayEquals(newLongArray, repositoryData.getLongs("multipleLongData"));
        _assertMultipleLongDataIsEmpty(repositoryData, "multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);

        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my binary content"));

        // Set local value - binary
        dataHolder.setStatus("singleBinaryData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("singleBinaryData", _createBinary("new local binary content"));

        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);
        _assertBinaryContent("new local binary content", repositoryData, "singleBinaryData");
        _assertBinaryDataIsEmpty(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);

        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my rich text content"));

        // Set local value - rich text
        dataHolder.setStatus("singleRichTextData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("singleRichTextData", _createRichText("new local rich text content"));

        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);
        _assertRichTextContent("new local rich text content", repositoryData, "singleRichTextData");
        _assertRichTextDataIsEmpty(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);

        // Write - composite
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 98L);

        // Set local value - composite
        dataHolder.setStatus("composite", ExternalizableDataStatus.LOCAL);
        ModifiableModelAwareComposite composite = dataHolder.getLocalComposite("composite", true);
        composite.setValue("singleLongData", 42L);

        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals(Long.valueOf(42), repositoryData.getRepositoryData("composite").getLong("singleLongData"));
        _assertRepositoryDataIsEmpty(repositoryData, "composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);

        // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.getRepositoryData("composite")
                .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
                .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
                .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData", "string value");
        
        // Set local value - composite/repeater[1]/subcomposite/singleStringData
        dataHolder.setStatus("composite/repeater[1]/subcomposite/singleStringData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("composite/repeater[1]/subcomposite/singleStringData", "new local string value");
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals("new local string value", subcomposite.getString("singleStringData"));
        _assertStringDataIsEmpty(subcomposite, "singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setLocalValue(String, Object)} on data without status
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetLocalValueAndSetStatusToExternalOnDataWithoutStatus() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetLocalValueAndSetStatusToExternalOnDataWithoutStatus", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData", "string value");
        
        // Set local value - string
        dataHolder.setStatus("singleStringData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("singleStringData", "new local string value");
        
        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals("new local string value", repositoryData.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        _assertStringDataIsEmpty(repositoryData, "singleStringData");
         
         // Write - multiple long
        Long[] longArray = new Long[]{78L, 1L};
        repositoryData.setValues("multipleLongData", longArray);
        
        // Set local value - multiple long
        dataHolder.setStatus("multipleLongData", ExternalizableDataStatus.EXTERNAL);
        Long[] newLongArray = new Long[]{478L, 13L};
        dataHolder.setLocalValue("multipleLongData", newLongArray);
        
        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertArrayEquals(newLongArray, repositoryData.getLongs("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        _assertMultipleLongDataIsEmpty(repositoryData, "multipleLongData");
         
         // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my binary content"));
        
        // Set local value - binary
        dataHolder.setStatus("singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("singleBinaryData", _createBinary("new local binary content"));
        
        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        _assertBinaryContent("new local binary content", repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertBinaryDataIsEmpty(repositoryData, "singleBinaryData");

         // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my rich text content"));
        
        // Set local value - rich text
        dataHolder.setStatus("singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("singleRichTextData", _createRichText("new local rich text content"));
        
        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        _assertRichTextContent("new local rich text content", repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertRichTextDataIsEmpty(repositoryData, "singleRichTextData");
         
         // Write - composite
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 98L);
        
        // Set local value - composite
        dataHolder.setStatus("composite", ExternalizableDataStatus.EXTERNAL);
        ModifiableModelAwareComposite composite = dataHolder.getLocalComposite("composite", true);
        composite.setValue("singleLongData", 42L);
        
        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals(Long.valueOf(42), repositoryData.getRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX).getLong("singleLongData"));
        _assertRepositoryDataIsEmpty(repositoryData, "composite");
         
         // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE)
             .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
             .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
             .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData", "string value");
         
        // Set local value - composite/repeater[1]/subcomposite/singleStringData
        dataHolder.setStatus("composite/repeater[1]/subcomposite/singleStringData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("composite/repeater[1]/subcomposite/singleStringData", "new local string value");
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals("new local string value", subcomposite.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        _assertStringDataIsEmpty(subcomposite, "singleStringData");
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setLocalValue(String, Object)} on local data
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetLocalValueAndSetStatusToLocalOnLocalData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetLocalValueAndSetStatusToLocalOnLocalData", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData", "string local value");
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string external value");
        setStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Set local value - string
        dataHolder.setStatus("singleStringData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("singleStringData", "new local string value");
        
        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals("new local string value", repositoryData.getString("singleStringData"));
        Assert.assertEquals("string external value", repositoryData.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        
        // Write - multiple long
        Long[] localLongArray = new Long[]{78L, 1L};
        repositoryData.setValues("multipleLongData", localLongArray);
        Long[] externalLongArray = new Long[]{65L, 98L, 432L};
        repositoryData.setValues("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, externalLongArray);
        setStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        
        // Set local value - multiple long
        dataHolder.setStatus("multipleLongData", ExternalizableDataStatus.LOCAL);
        Long[] newLongArray = new Long[]{478L, 13L};
        dataHolder.setLocalValue("multipleLongData", newLongArray);
        
        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        Assert.assertArrayEquals(newLongArray, repositoryData.getLongs("multipleLongData"));
        Assert.assertArrayEquals(externalLongArray, repositoryData.getLongs("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        
        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my local binary content"));
        _binaryType.write(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createBinary("my external binary content"));
        setStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);
        
        // Set local value - binary
        dataHolder.setStatus("singleBinaryData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("singleBinaryData", _createBinary("new local binary content"));
        
        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);
        _assertBinaryContent("new local binary content", repositoryData, "singleBinaryData");
        _assertBinaryContent("my external binary content", repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my local rich text content"));
        _richTextType.write(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createRichText("my external rich text content"));
        setStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);
        
        // Set local value - rich text
        dataHolder.setStatus("singleRichTextData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("singleRichTextData", _createRichText("new local rich text content"));
        
        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);
        _assertRichTextContent("new local rich text content", repositoryData, "singleRichTextData");
        _assertRichTextContent("my external rich text content", repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Write - composite
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 42L);
        repositoryData.addRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 18L);
        setStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);
        
        // Set local value - composite
        dataHolder.setStatus("composite", ExternalizableDataStatus.LOCAL);
        ModifiableModelAwareComposite composite = dataHolder.getLocalComposite("composite", true);
        composite.setValue("singleLongData", 75L);
        
        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals(Long.valueOf(75), repositoryData.getRepositoryData("composite").getLong("singleLongData"));
        Assert.assertEquals(Long.valueOf(18), repositoryData.getRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX).getLong("singleLongData"));
        
        // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.getRepositoryData("composite")
            .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
            .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
            .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData", "string local value");
        subcomposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string external value");
        setStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Set local value - composite/repeater[1]/subcomposite/singleStringData
        dataHolder.setStatus("composite/repeater[1]/subcomposite/singleStringData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("composite/repeater[1]/subcomposite/singleStringData", "new local string value");
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals("new local string value", subcomposite.getString("singleStringData"));
        Assert.assertEquals("string external value", subcomposite.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setLocalValue(String, Object)} on local data
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetLocalValueAndSetStatusToExternalOnLocalData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetLocalValueAndSetStatusToExternalOnLocalData", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData", "string local value");
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string external value");
        setStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Set local value - string
        dataHolder.setStatus("singleStringData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("singleStringData", "new local string value");
        
        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals("new local string value", repositoryData.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        Assert.assertEquals("string external value", repositoryData.getString("singleStringData"));
        
        // Write - multiple long
        Long[] localLongArray = new Long[]{78L, 1L};
        repositoryData.setValues("multipleLongData", localLongArray);
        Long[] externalLongArray = new Long[]{65L, 98L, 432L};
        repositoryData.setValues("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, externalLongArray);
        setStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        
        // Set local value - multiple long
        dataHolder.setStatus("multipleLongData", ExternalizableDataStatus.EXTERNAL);
        Long[] newLongArray = new Long[]{478L, 13L};
        dataHolder.setLocalValue("multipleLongData", newLongArray);
        
        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertArrayEquals(newLongArray, repositoryData.getLongs("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        Assert.assertArrayEquals(externalLongArray, repositoryData.getLongs("multipleLongData"));
        
        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my local binary content"));
        _binaryType.write(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createBinary("my external binary content"));
        setStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);
        
        // Set local value - binary
        dataHolder.setStatus("singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("singleBinaryData", _createBinary("new local binary content"));
        
        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        _assertBinaryContent("new local binary content", repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertBinaryContent("my external binary content", repositoryData, "singleBinaryData");
        
        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my local rich text content"));
        _richTextType.write(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createRichText("my external rich text content"));
        setStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);
        
        // Set local value - rich text
        dataHolder.setStatus("singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("singleRichTextData", _createRichText("new local rich text content"));
        
        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        _assertRichTextContent("new local rich text content", repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertRichTextContent("my external rich text content", repositoryData, "singleRichTextData");
        
        // Write - composite
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 42L);
        repositoryData.addRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 18L);
        setStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);
        
        // Set local value - composite
        dataHolder.setStatus("composite", ExternalizableDataStatus.EXTERNAL);
        ModifiableModelAwareComposite composite = dataHolder.getLocalComposite("composite", true);
        composite.setValue("singleLongData", 75L);
        
        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals(Long.valueOf(75), repositoryData.getRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX).getLong("singleLongData"));
        Assert.assertEquals(Long.valueOf(18), repositoryData.getRepositoryData("composite").getLong("singleLongData"));
        
        // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.getRepositoryData("composite")
            .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
            .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
            .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData", "string local value");
        subcomposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string external value");
        setStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Set local value - composite/repeater[1]/subcomposite/singleStringData
        dataHolder.setStatus("composite/repeater[1]/subcomposite/singleStringData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("composite/repeater[1]/subcomposite/singleStringData", "new local string value");
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals("new local string value", subcomposite.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        Assert.assertEquals("string external value", subcomposite.getString("singleStringData"));
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setLocalValue(String, Object)} on external data
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetLocalValueAndSetStatusToLocalOnExternalData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetLocalValueAndSetStatusToLocalOnExternalData", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string local value");
        repositoryData.setValue("singleStringData", "string external value");
        setStatus(repositoryData, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - string
        dataHolder.setStatus("singleStringData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("singleStringData", "new local string value");
        
        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals("new local string value", repositoryData.getString("singleStringData"));
        Assert.assertEquals("string external value", repositoryData.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        
        // Write - multiple long
        Long[] localLongArray = new Long[]{78L, 1L};
        repositoryData.setValues("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, localLongArray);
        Long[] externalLongArray = new Long[]{65L, 98L, 432L};
        repositoryData.setValues("multipleLongData", externalLongArray);
        setStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - multiple long
        dataHolder.setStatus("multipleLongData", ExternalizableDataStatus.LOCAL);
        Long[] newLongArray = new Long[]{478L, 13L};
        dataHolder.setLocalValue("multipleLongData", newLongArray);
        
        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        Assert.assertArrayEquals(newLongArray, repositoryData.getLongs("multipleLongData"));
        Assert.assertArrayEquals(externalLongArray, repositoryData.getLongs("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        
        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createBinary("my local binary content"));
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my external binary content"));
        setStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - binary
        dataHolder.setStatus("singleBinaryData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("singleBinaryData", _createBinary("new local binary content"));
        
        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);
        _assertBinaryContent("new local binary content", repositoryData, "singleBinaryData");
        _assertBinaryContent("my external binary content", repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createRichText("my local rich text content"));
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my external rich text content"));
        setStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - rich text
        dataHolder.setStatus("singleRichTextData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("singleRichTextData", _createRichText("new local rich text content"));
        
        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);
        _assertRichTextContent("new local rich text content", repositoryData, "singleRichTextData");
        _assertRichTextContent("my external rich text content", repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Write - composite
        repositoryData.addRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 42L);
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 18L);
        setStatus(repositoryData, "composite", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - composite
        dataHolder.setStatus("composite", ExternalizableDataStatus.LOCAL);
        ModifiableModelAwareComposite composite = dataHolder.getLocalComposite("composite", true);
        composite.setValue("singleLongData", 75L);
        
        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals(Long.valueOf(75), repositoryData.getRepositoryData("composite").getLong("singleLongData"));
        Assert.assertEquals(Long.valueOf(18), repositoryData.getRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX).getLong("singleLongData"));
        
        // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.getRepositoryData("composite")
            .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
            .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
            .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string local value");
        subcomposite.setValue("singleStringData", "string external value");
        setStatus(subcomposite, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - composite/repeater[1]/subcomposite/singleStringData
        dataHolder.setStatus("composite/repeater[1]/subcomposite/singleStringData", ExternalizableDataStatus.LOCAL);
        dataHolder.setLocalValue("composite/repeater[1]/subcomposite/singleStringData", "new local string value");
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals("new local string value", subcomposite.getString("singleStringData"));
        Assert.assertEquals("string external value", subcomposite.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setLocalValue(String, Object)} on external data
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetLocalValueAndSetStatusToExternalOnExternalData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetLocalValueAndSetStatusToExternalOnExternalData", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string local value");
        repositoryData.setValue("singleStringData", "string external value");
        setStatus(repositoryData, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - string
        dataHolder.setStatus("singleStringData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("singleStringData", "new local string value");
        
        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals("new local string value", repositoryData.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        Assert.assertEquals("string external value", repositoryData.getString("singleStringData"));
        
        // Write - multiple long
        Long[] localLongArray = new Long[]{78L, 1L};
        repositoryData.setValues("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, localLongArray);
        Long[] externalLongArray = new Long[]{65L, 98L, 432L};
        repositoryData.setValues("multipleLongData", externalLongArray);
        setStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - multiple long
        dataHolder.setStatus("multipleLongData", ExternalizableDataStatus.EXTERNAL);
        Long[] newLongArray = new Long[]{478L, 13L};
        dataHolder.setLocalValue("multipleLongData", newLongArray);
        
        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertArrayEquals(newLongArray, repositoryData.getLongs("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        Assert.assertArrayEquals(externalLongArray, repositoryData.getLongs("multipleLongData"));
        
        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createBinary("my local binary content"));
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my external binary content"));
        setStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - binary
        dataHolder.setStatus("singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("singleBinaryData", _createBinary("new local binary content"));
        
        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        _assertBinaryContent("new local binary content", repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertBinaryContent("my external binary content", repositoryData, "singleBinaryData");
        
        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createRichText("my local rich text content"));
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my external rich text content"));
        setStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - rich text
        dataHolder.setStatus("singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("singleRichTextData", _createRichText("new local rich text content"));
        
        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        _assertRichTextContent("new local rich text content", repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertRichTextContent("my external rich text content", repositoryData, "singleRichTextData");
        
        // Write - composite
        repositoryData.addRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 42L);
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 18L);
        setStatus(repositoryData, "composite", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - composite
        dataHolder.setStatus("composite", ExternalizableDataStatus.EXTERNAL);
        ModifiableModelAwareComposite composite = dataHolder.getLocalComposite("composite", true);
        composite.setValue("singleLongData", 75L);
        
        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals(Long.valueOf(75), repositoryData.getRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX).getLong("singleLongData"));
        Assert.assertEquals(Long.valueOf(18), repositoryData.getRepositoryData("composite").getLong("singleLongData"));
        
        // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.getRepositoryData("composite")
            .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
            .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
            .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string local value");
        subcomposite.setValue("singleStringData", "string external value");
        setStatus(subcomposite, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - composite/repeater[1]/subcomposite/singleStringData
        dataHolder.setStatus("composite/repeater[1]/subcomposite/singleStringData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setLocalValue("composite/repeater[1]/subcomposite/singleStringData", "new local string value");
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals("new local string value", subcomposite.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        Assert.assertEquals("string external value", subcomposite.getString("singleStringData"));
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setExternalValue(String, Object)} on data without status
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetExternalValueOnDataWithoutStatus() throws  IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetExternalValueOnDataWithoutStatus", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData", "string value");

        // Set external value - string
        dataHolder.setStatus("singleStringData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setExternalValue("singleStringData", "new external string value");
        dataHolder.removeLocalValue("singleStringData");

        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        _assertStringDataIsEmpty(repositoryData, "singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        Assert.assertEquals("new external string value", repositoryData.getString("singleStringData"));

        // Write - multiple long
        Long[] longArray = new Long[]{78L, 1L};
        repositoryData.setValues("multipleLongData", longArray);

        // Set external value - multiple long
        dataHolder.setStatus("multipleLongData", ExternalizableDataStatus.EXTERNAL);
        Long[] newLongArray = new Long[]{478L, 13L};
        dataHolder.setExternalValue("multipleLongData", newLongArray);
        dataHolder.removeLocalValue("multipleLongData");

        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.EXTERNAL);
        _assertMultipleLongDataIsEmpty(repositoryData, "multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        Assert.assertArrayEquals(newLongArray, repositoryData.getLongs("multipleLongData"));

        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my binary content"));

        // Set external value - binary
        dataHolder.setStatus("singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setExternalValue("singleBinaryData", _createBinary("new external binary content"));
        dataHolder.removeLocalValue("singleBinaryData");

        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        _assertBinaryDataIsEmpty(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertBinaryContent("new external binary content", repositoryData, "singleBinaryData");

        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my rich text content"));

        // Set external value - rich text
        dataHolder.setStatus("singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setExternalValue("singleRichTextData", _createRichText("new external rich text content"));
        dataHolder.removeLocalValue("singleRichTextData");

        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        _assertRichTextDataIsEmpty(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertRichTextContent("new external rich text content", repositoryData, "singleRichTextData");

        // Write - composite
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 42L);

        // Set external value - composite
        dataHolder.setStatus("composite", ExternalizableDataStatus.EXTERNAL);
        dataHolder.getExternalComposite("composite", true).setValue("singleLongData", 18L);
        dataHolder.removeLocalValue("composite");

        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.EXTERNAL);
        _assertRepositoryDataIsEmpty(repositoryData, "composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        Assert.assertEquals(Long.valueOf(18), repositoryData.getRepositoryData("composite").getLong("singleLongData"));

        // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.getRepositoryData("composite")
                .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
                .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
                .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData", "string value");
        
        // Set external value - composite/repeater[1]/subcomposite/singleStringData
        dataHolder.setStatus("composite/repeater[1]/subcomposite/singleStringData", ExternalizableDataStatus.EXTERNAL);
        dataHolder.setExternalValue("composite/repeater[1]/subcomposite/singleStringData", "new external string value");
        dataHolder.removeLocalValue("composite/repeater[1]/subcomposite/singleStringData");
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        _assertStringDataIsEmpty(subcomposite, "singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        Assert.assertEquals("new external string value", subcomposite.getString("singleStringData"));
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setExternalValue(String, Object)} on local data
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetExternalValueOnLocalData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetExternalValueOnLocalData", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData", "string local value");
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string external value");
        setStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Set external value - string
        dataHolder.setExternalValue("singleStringData", "new external string value");
        
        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals("string local value", repositoryData.getString("singleStringData"));
        Assert.assertEquals("new external string value", repositoryData.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        
        // Write - multiple long
        Long[] localLongArray = new Long[]{78L, 1L};
        repositoryData.setValues("multipleLongData", localLongArray);
        Long[] externalLongArray = new Long[]{65L, 98L, 432L};
        repositoryData.setValues("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, externalLongArray);
        setStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        
        // Set external value - multiple long
        Long[] newLongArray = new Long[]{478L, 13L};
        dataHolder.setExternalValue("multipleLongData", newLongArray);
        
        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        Assert.assertArrayEquals(localLongArray, repositoryData.getLongs("multipleLongData"));
        Assert.assertArrayEquals(newLongArray, repositoryData.getLongs("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        
        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my local binary content"));
        _binaryType.write(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createBinary("my external binary content"));
        setStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);

        // Set external value - binary
        dataHolder.setExternalValue("singleBinaryData", _createBinary("new external binary content"));

        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);
        _assertBinaryContent("my local binary content", repositoryData, "singleBinaryData");
        _assertBinaryContent("new external binary content", repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);

        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my local rich text content"));
        _richTextType.write(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createRichText("my external rich text content"));
        setStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);

        // Set external value - rich text
        dataHolder.setExternalValue("singleRichTextData", _createRichText("new external rich text content"));

        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);
        _assertRichTextContent("my local rich text content", repositoryData, "singleRichTextData");
        _assertRichTextContent("new external rich text content", repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);

        // Write - composite
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 42L);
        repositoryData.addRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 18L);
        setStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);

        // Set external value - composite
        ModifiableModelAwareComposite composite = dataHolder.getExternalComposite("composite", true);
        composite.setValue("singleLongData", 75L);

        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals(Long.valueOf(42), repositoryData.getRepositoryData("composite").getLong("singleLongData"));
        Assert.assertEquals(Long.valueOf(75), repositoryData.getRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX).getLong("singleLongData"));
        
        // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.getRepositoryData("composite")
            .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
            .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
            .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData", "string local value");
        subcomposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string external value");
        setStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Set external value - composite/repeater[1]/subcomposite/singleStringData
        dataHolder.setExternalValue("composite/repeater[1]/subcomposite/singleStringData", "new external string value");
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals("string local value", subcomposite.getString("singleStringData"));
        Assert.assertEquals("new external string value", subcomposite.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setExternalValue(String, Object)} on external data
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetExternalValueOnExternalData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetExternalValueOnExternalData", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string local value");
        repositoryData.setValue("singleStringData", "string external value");
        setStatus(repositoryData, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - string
        dataHolder.setExternalValue("singleStringData", "new external string value");
        
        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals("string local value", repositoryData.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        Assert.assertEquals("new external string value", repositoryData.getString("singleStringData"));
        
        // Write - multiple long
        Long[] localLongArray = new Long[]{78L, 1L};
        repositoryData.setValues("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, localLongArray);
        Long[] externalLongArray = new Long[]{65L, 98L, 432L};
        repositoryData.setValues("multipleLongData", externalLongArray);
        setStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - multiple long
        Long[] newLongArray = new Long[]{478L, 13L};
        dataHolder.setExternalValue("multipleLongData", newLongArray);
        
        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertArrayEquals(localLongArray, repositoryData.getLongs("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        Assert.assertArrayEquals(newLongArray, repositoryData.getLongs("multipleLongData"));
        
        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createBinary("my local binary content"));
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my external binary content"));
        setStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.EXTERNAL);

        // Set local value - binary
        dataHolder.setExternalValue("singleBinaryData", _createBinary("new external binary content"));

        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        _assertBinaryContent("my local binary content", repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertBinaryContent("new external binary content", repositoryData, "singleBinaryData");
        
        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createRichText("my local rich text content"));
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my external rich text content"));
        setStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.EXTERNAL);

        // Set local value - rich text
        dataHolder.setExternalValue("singleRichTextData", _createRichText("new external rich text content"));

        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        _assertRichTextContent("my local rich text content", repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertRichTextContent("new external rich text content", repositoryData, "singleRichTextData");

        // Write - composite
        repositoryData.addRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 42L);
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 18L);
        setStatus(repositoryData, "composite", ExternalizableDataStatus.EXTERNAL);

        // Set local value - composite
        ModifiableModelAwareComposite composite = dataHolder.getExternalComposite("composite", true);
        composite.setValue("singleLongData", 75L);

        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals(Long.valueOf(42), repositoryData.getRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX).getLong("singleLongData"));
        Assert.assertEquals(Long.valueOf(75), repositoryData.getRepositoryData("composite").getLong("singleLongData"));
        
        // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.getRepositoryData("composite")
            .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
            .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
            .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string local value");
        subcomposite.setValue("singleStringData", "string external value");
        setStatus(subcomposite, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        
        // Set local value - composite/repeater[1]/subcomposite/singleStringData
        dataHolder.setExternalValue("composite/repeater[1]/subcomposite/singleStringData", "new external string value");
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals("string local value", subcomposite.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        Assert.assertEquals("new external string value", subcomposite.getString("singleStringData"));
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setStatus(String, ExternalizableDataStatus)} on unvalued data
     */
    @Test
    public void testSetStatusOnNotValuedData()
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetStatusOnNotValuedData", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Update - string
        dataHolder.setStatus("singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        _assertStringDataIsEmpty(repositoryData, "singleStringData");
        _assertStringDataIsEmpty(repositoryData, "singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Update - multiple long
        dataHolder.setStatus("multipleLongData", ExternalizableDataStatus.LOCAL);
        
        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        _assertMultipleLongDataIsEmpty(repositoryData, "multipleLongData");
        _assertMultipleLongDataIsEmpty(repositoryData, "multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Update - binary
        dataHolder.setStatus("singleBinaryData", ExternalizableDataStatus.LOCAL);
        
        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);
        _assertBinaryDataIsEmpty(repositoryData, "singleBinaryData");
        _assertBinaryDataIsEmpty(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Update - rich text
        dataHolder.setStatus("singleRichTextData", ExternalizableDataStatus.LOCAL);
        
        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);
        _assertRichTextDataIsEmpty(repositoryData, "singleRichTextData");
        _assertRichTextDataIsEmpty(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Update - composite
        dataHolder.setStatus("composite", ExternalizableDataStatus.LOCAL);
        
        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);
        _assertRepositoryDataIsEmpty(repositoryData, "composite");
        _assertRepositoryDataIsEmpty(repositoryData, "composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Update - composite/repeater[1]/subcomposite/singleStringData
        RepositoryData subcomposite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE)
                .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
                .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
                .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        dataHolder.setStatus("composite/repeater[1]/subcomposite/singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        _assertStringDataIsEmpty(subcomposite, "singleStringData");
        _assertStringDataIsEmpty(subcomposite, "singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setStatus(String, ExternalizableDataStatus)} keeping the same status
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetStatusWithoutChanging() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetStatusWithoutChanging", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData", "string local value");
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string external value");
        setStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Update - string
        dataHolder.setStatus("singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals("string local value", repositoryData.getString("singleStringData"));
        Assert.assertEquals("string external value", repositoryData.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        
        // Write - multiple long
        Long[] localLongArray = new Long[]{78L, 1L};
        repositoryData.setValues("multipleLongData", localLongArray);
        Long[] externalLongArray = new Long[]{65L, 98L, 432L};
        repositoryData.setValues("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, externalLongArray);
        setStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        
        // Update - multiple long
        dataHolder.setStatus("multipleLongData", ExternalizableDataStatus.LOCAL);
        
        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        Assert.assertArrayEquals(localLongArray, repositoryData.getLongs("multipleLongData"));
        Assert.assertArrayEquals(externalLongArray, repositoryData.getLongs("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        
        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my local binary content"));
        _binaryType.write(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createBinary("my external binary content"));
        setStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);
        
        // Update - binary
        dataHolder.setStatus("singleBinaryData", ExternalizableDataStatus.LOCAL);
        
        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);
        _assertBinaryContent("my local binary content", repositoryData, "singleBinaryData");
        _assertBinaryContent("my external binary content", repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my local rich text content"));
        _richTextType.write(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createRichText("my external rich text content"));
        setStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);
        
        // Update - rich text
        dataHolder.setStatus("singleRichTextData", ExternalizableDataStatus.LOCAL);
        
        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);
        _assertRichTextContent("my local rich text content", repositoryData, "singleRichTextData");
        _assertRichTextContent("my external rich text content", repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        
        // Write - composite
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 42L);
        repositoryData.addRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 18L);
        setStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);
        
        // Update - composite
        dataHolder.setStatus("composite", ExternalizableDataStatus.LOCAL);
        
        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals(Long.valueOf(42), repositoryData.getRepositoryData("composite").getLong("singleLongData"));
        Assert.assertEquals(Long.valueOf(18), repositoryData.getRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX).getLong("singleLongData"));
        
        // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.getRepositoryData("composite")
            .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
            .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
            .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData", "string local value");
        subcomposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string external value");
        setStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Update - composite/repeater[1]/subcomposite/singleStringData
        dataHolder.setStatus("composite/repeater[1]/subcomposite/singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        Assert.assertEquals("string local value", subcomposite.getString("singleStringData"));
        Assert.assertEquals("string external value", subcomposite.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setStatus(String, ExternalizableDataStatus)}
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetStatus() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetStatus", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData", "string local value");
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string external value");
        setStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Update - string
        dataHolder.setStatus("singleStringData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals("string local value", repositoryData.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        Assert.assertEquals("string external value", repositoryData.getString("singleStringData"));
        
        // Write - multiple long
        Long[] localLongArray = new Long[]{78L, 1L};
        repositoryData.setValues("multipleLongData", localLongArray);
        Long[] externalLongArray = new Long[]{65L, 98L, 432L};
        repositoryData.setValues("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, externalLongArray);
        setStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        
        // Update - multiple long
        dataHolder.setStatus("multipleLongData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertArrayEquals(localLongArray, repositoryData.getLongs("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        Assert.assertArrayEquals(externalLongArray, repositoryData.getLongs("multipleLongData"));
        
        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my local binary content"));
        _binaryType.write(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createBinary("my external binary content"));
        setStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);
        
        // Update - binary
        dataHolder.setStatus("singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        _assertBinaryContent("my local binary content", repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertBinaryContent("my external binary content", repositoryData, "singleBinaryData");
        
        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my local rich text content"));
        _richTextType.write(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createRichText("my external rich text content"));
        setStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);
        
        // Update - rich text
        dataHolder.setStatus("singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        _assertRichTextContent("my local rich text content", repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertRichTextContent("my external rich text content", repositoryData, "singleRichTextData");
        
        // Write - composite
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 42L);
        repositoryData.addRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 18L);
        setStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);
        
        // Update - composite
        dataHolder.setStatus("composite", ExternalizableDataStatus.EXTERNAL);
        
        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals(Long.valueOf(42), repositoryData.getRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX).getLong("singleLongData"));
        Assert.assertEquals(Long.valueOf(18), repositoryData.getRepositoryData("composite").getLong("singleLongData"));
        
        // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.getRepositoryData("composite")
            .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
            .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
            .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData", "string local value");
        subcomposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string external value");
        setStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Update - composite/repeater[1]/subcomposite/singleStringData
        dataHolder.setStatus("composite/repeater[1]/subcomposite/singleStringData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals("string local value", subcomposite.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        Assert.assertEquals("string external value", subcomposite.getString("singleStringData"));
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setStatus(String, ExternalizableDataStatus)} on data without alternative value
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetStatusOnDataWithoutAltValue() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetStatusOnDataWithoutAltValue", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData", "string local value");
        setStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Update - string
        dataHolder.setStatus("singleStringData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals("string local value", repositoryData.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        _assertStringDataIsEmpty(repositoryData, "singleStringData");
        
        // Write - multiple long
        Long[] localLongArray = new Long[]{78L, 1L};
        repositoryData.setValues("multipleLongData", localLongArray);
        setStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        
        // Update - multiple long
        dataHolder.setStatus("multipleLongData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertArrayEquals(localLongArray, repositoryData.getLongs("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        _assertMultipleLongDataIsEmpty(repositoryData, "multipleLongData");
        
        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData", _createBinary("my local binary content"));
        setStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);
        
        // Update - binary
        dataHolder.setStatus("singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        _assertBinaryContent("my local binary content", repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertBinaryDataIsEmpty(repositoryData, "singleBinaryData");
        
        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData", _createRichText("my local rich text content"));
        setStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);
        
        // Update - rich text
        dataHolder.setStatus("singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        _assertRichTextContent("my local rich text content", repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertRichTextDataIsEmpty(repositoryData, "singleRichTextData");
        
        // Write - composite
        repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 42L);
        setStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);
        
        // Update - composite
        dataHolder.setStatus("composite", ExternalizableDataStatus.EXTERNAL);
        
        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals(Long.valueOf(42), repositoryData.getRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX).getLong("singleLongData"));
        _assertRepositoryDataIsEmpty(repositoryData, "composite");
        
        // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE)
            .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
            .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
            .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData", "string local value");
        setStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Update - composite/repeater[1]/subcomposite/singleStringData
        dataHolder.setStatus("composite/repeater[1]/subcomposite/singleStringData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        Assert.assertEquals("string local value", subcomposite.getString("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX));
        _assertStringDataIsEmpty(subcomposite, "singleStringData");
    }
    
    /**
     * Tests the {@link ModifiableModelAwareDataHolder#setStatus(String, ExternalizableDataStatus)} on data with only alternative value
     * @throws IOException if an error occurs
     */
    @Test
    public void testSetStatusOnDataWithOnlyAltValue() throws IOException
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData("testSetStatusOnDataWithOnlyAltValue", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write - string
        repositoryData.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string external value");
        setStatus(repositoryData, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Update - string
        dataHolder.setStatus("singleStringData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - string
        _assertStatus(repositoryData, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        _assertStringDataIsEmpty(repositoryData, "singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        Assert.assertEquals("string external value", repositoryData.getString("singleStringData"));
        
        // Write - multiple long
        Long[] externalLongArray = new Long[]{65L, 98L, 432L};
        repositoryData.setValues("multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, externalLongArray);
        setStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.LOCAL);
        
        // Update - multiple long
        dataHolder.setStatus("multipleLongData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - multiple long
        _assertStatus(repositoryData, "multipleLongData", ExternalizableDataStatus.EXTERNAL);
        _assertMultipleLongDataIsEmpty(repositoryData, "multipleLongData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        Assert.assertArrayEquals(externalLongArray, repositoryData.getLongs("multipleLongData"));
        
        // Write - binary
        _binaryType.write(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createBinary("my external binary content"));
        setStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.LOCAL);
        
        // Update - binary
        dataHolder.setStatus("singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - binary
        _assertStatus(repositoryData, "singleBinaryData", ExternalizableDataStatus.EXTERNAL);
        _assertBinaryDataIsEmpty(repositoryData, "singleBinaryData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertBinaryContent("my external binary content", repositoryData, "singleBinaryData");
        
        // Write - rich text
        _richTextType.write(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, _createRichText("my external rich text content"));
        setStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.LOCAL);
        
        // Update - rich text
        dataHolder.setStatus("singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - rich text
        _assertStatus(repositoryData, "singleRichTextData", ExternalizableDataStatus.EXTERNAL);
        _assertRichTextDataIsEmpty(repositoryData, "singleRichTextData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        _assertRichTextContent("my external rich text content", repositoryData, "singleRichTextData");
        
        // Write - composite
        repositoryData.addRepositoryData("composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, RepositoryConstants.COMPOSITE_NODETYPE).setValue("singleLongData", 18L);
        setStatus(repositoryData, "composite", ExternalizableDataStatus.LOCAL);
        
        // Update - composite
        dataHolder.setStatus("composite", ExternalizableDataStatus.EXTERNAL);
        
        // Read - composite
        _assertStatus(repositoryData, "composite", ExternalizableDataStatus.EXTERNAL);
        _assertRepositoryDataIsEmpty(repositoryData, "composite" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        Assert.assertEquals(Long.valueOf(18), repositoryData.getRepositoryData("composite").getLong("singleLongData"));
        
        // Write - composite/repeater[1]/subcomposite/singleStringData
        ModifiableRepositoryData subcomposite = repositoryData.getRepositoryData("composite")
            .addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE)
            .addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE)
            .addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subcomposite.setValue("singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX, "string external value");
        setStatus(subcomposite, "singleStringData", ExternalizableDataStatus.LOCAL);
        
        // Update - composite/repeater[1]/subcomposite/singleStringData
        dataHolder.setStatus("composite/repeater[1]/subcomposite/singleStringData", ExternalizableDataStatus.EXTERNAL);
        
        // Read - composite/repeater[1]/subcomposite/singleStringData
        _assertStatus(subcomposite, "singleStringData", ExternalizableDataStatus.EXTERNAL);
        _assertStringDataIsEmpty(subcomposite, "singleStringData" + ModelAwareDataHolder.ALTERNATIVE_SUFFIX);
        Assert.assertEquals("string external value", subcomposite.getString("singleStringData"));
    }

    private void setStatus(ModifiableRepositoryData repositoryData, String dataName, ExternalizableDataStatus status)
    {
        repositoryData.setValue(dataName + ModelAwareDataHolder.STATUS_SUFFIX, status.name().toLowerCase());
    }
    
    private Binary _createBinary(String content) throws IOException
    {
        Binary binary = new Binary();
        try (OutputStream os = binary.getOutputStream())
        {
            os.write(content.getBytes());
        }
        binary.setEncoding("UTF-8");
        binary.setFilename("myExternalBinary.txt");
        binary.setLastModificationDate(__RESOURCE_LAST_MODIFIED);
        binary.setMimeType("text/plain");
        
        return binary;
    }
    
    private RichText _createRichText(String content) throws IOException
    {
        RichText richText = new RichText();
        try (OutputStream os = richText.getOutputStream())
        {
            os.write(content.getBytes());
        }
        richText.setEncoding("UTF-8");
        richText.setMimeType("text/xml");
        richText.setLastModificationDate(__RESOURCE_LAST_MODIFIED);
        
        return richText;
    }
    
    private void _assertStatus(RepositoryData repositoryData, String dataName, ExternalizableDataStatus status)
    {
        String statusAsString = status.name().toLowerCase();
        Assert.assertEquals(statusAsString, repositoryData.getString(dataName + ModelAwareDataHolder.STATUS_SUFFIX));
    }
    
    private void _assertBinaryContent(String content, RepositoryData repositoryData, String dataName) throws IOException
    {
        Binary binary = (Binary) _binaryType.read(repositoryData, dataName);
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(binary.getInputStream(), writer, binary.getEncoding());
            Assert.assertEquals(content, writer.toString());
        }
    }
    
    private void _assertRichTextContent(String content, RepositoryData repositoryData, String dataName) throws IOException
    {
        RichText richText = (RichText) _richTextType.read(repositoryData, dataName);
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(richText.getInputStream(), writer, richText.getEncoding());
            Assert.assertEquals(content, writer.toString());
        }
    }
    
    private void _assertStringDataIsEmpty(RepositoryData repositoryData, String dataName)
    {
        try
        {
            repositoryData.getString(dataName);
            Assert.fail("the data '" + dataName + "' should not be present in the repository data '" + repositoryData + "'");
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains(dataName));
        }
    }
    
    private void _assertMultipleLongDataIsEmpty(RepositoryData repositoryData, String dataName)
    {
        try
        {
            repositoryData.getLongs(dataName);
            Assert.fail("the data '" + dataName + "' should not be present in the repository data '" + repositoryData + "'");
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains(dataName));
        }
    }
    
    private void _assertBinaryDataIsEmpty(RepositoryData repositoryData, String dataName)
    {
        _assertRepositoryDataIsEmpty(repositoryData, dataName);
    }
    
    private void _assertRichTextDataIsEmpty(RepositoryData repositoryData, String dataName)
    {
        _assertRepositoryDataIsEmpty(repositoryData, dataName);
    }
    
    private void _assertRepositoryDataIsEmpty(RepositoryData repositoryData, String dataName)
    {
        try
        {
            repositoryData.getRepositoryData(dataName);
            Assert.fail("the data '" + dataName + "' should not be present in the repository data '" + repositoryData + "'");
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains(dataName));
        }
    }
}
