/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.io.IOException;
import java.io.StringWriter;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Binary;
import org.ametys.cms.data.holder.impl.DefaultModifiableModelAwareDataHolder;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.cms.utils.ResourceElementTypeHelperForTests;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.external.ExternalizableDataProvider.ExternalizableDataStatus;
import org.ametys.plugins.repository.data.holder.ModifiableModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.values.SynchronizableValue;
import org.ametys.plugins.repository.data.holder.values.SynchronizationContext;
import org.ametys.plugins.repository.data.holder.values.SynchronizationResult;
import org.ametys.plugins.repository.data.holder.values.UntouchedValue;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method test cases
 *
 */
@NeedFeatures({
    "test-data/test-model",
    "test-data/test-externalizable-data-provider",
    
    "repository/repository.data.holder.helpers",
    "repository/resolver", 
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModifiableModelAwareDataHolderSynchronizeExternalizableTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __DATA_NODE_TYPE = "ametys:unstructured";
    
    private static final String __STRING_DEFINITION_PATH = "composite/repeater/subcomposite/singleStringData";
    private static final String __STRING_DATA_PATH = "composite/repeater[1]/subcomposite/singleStringData";
    private static final String __STRING_LOCAL_VALUE = "local value";
    private static final String __STRING_EXTERNAL_VALUE = "external value";
    private static final String __STRING_NEW_VALUE = "new value";
    private static final String __STRING_DEFAULT_VALUE = "default value";
    
    private static final String __BINARY_DEFINITION_PATH = "composite/repeater/subcomposite/singleBinaryData";
    private static final String __BINARY_DATA_PATH = "composite/repeater[1]/subcomposite/singleBinaryData";
    private static final String __BINARY_LOCAL_CONTENT = "local content";
    private static final String __BINARY_EXTERNAL_CONTENT = "external content";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static Model _model;

    private static JCRRepositoryData _root;
    private static View _view;
    private static View _viewWithBinary;
    
    private ModifiableRepositoryData _repositoryData;
    private ModifiableModelAwareDataHolder _dataHolder;
    private SynchronizationContext _context;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _root = new JCRRepositoryData(rootAmetysObject.getNode());
        
        // For these tests, add a default value to the single string data in the subcomposite
        @SuppressWarnings("unchecked")
        ElementDefinition<String> definition = (ElementDefinition<String>) _model.getModelItem(__STRING_DEFINITION_PATH);
        definition.setDefaultValue(__STRING_DEFAULT_VALUE);

        _view = View.of(_model, __STRING_DEFINITION_PATH);
        _viewWithBinary = View.of(_model, __BINARY_DEFINITION_PATH);
    }
    
    /**
     * Prepare components that are specific for each test
     * @throws Exception if an error occurs
     */
    @Before
    public void prepareSpecificComponents()throws Exception
    {
        _repositoryData = _root.addRepositoryData(UUID.randomUUID().toString(), __DATA_NODE_TYPE);
        _dataHolder = new DefaultModifiableModelAwareDataHolder(_repositoryData, _model);
        _dataHolder.getComposite("composite", true)
                   .getRepeater("repeater", true)
                   .addEntry()
                   .getComposite("subcomposite", true);

        _context = SynchronizationContext.newInstance().withStatus(ExternalizableDataStatus.EXTERNAL);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * The repository is empty
     * Values to synchronize don't contain the data
     * The data is externalizable
     * Do not use the default value
     * Result: repo still empty, no change
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyExternalizableValueEmpty() throws Exception
    {
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertFalse(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasLocalValueOrEmpty(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValueOrEmpty(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * The repository is empty
     * Values to synchronize contains a value
     * The data is externalizable
     * Result: changed to set the externalizable value
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptySetExternalizableValue() throws Exception
    {
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", __STRING_EXTERNAL_VALUE)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.EXTERNAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_EXTERNAL_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * The repository is empty
     * Values to synchronize contains a value
     * The data is externalizable
     * Update the status
     * Result: changed to set the externalizable value and update the status
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptySetExternalizableValueUpdateStatus() throws Exception
    {
        SynchronizableValue value = new SynchronizableValue(__STRING_EXTERNAL_VALUE, ExternalizableDataStatus.EXTERNAL);
        value.setExternalizableStatus(ExternalizableDataStatus.EXTERNAL);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.EXTERNAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_EXTERNAL_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * The repository is empty
     * Values to synchronize contains a value
     * The data is externalizable, set the local value
     * Result: changed to set the local value
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptySetLocalValue() throws Exception
    {
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", __STRING_LOCAL_VALUE)))));
        
        _context.withStatus(ExternalizableDataStatus.LOCAL);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_LOCAL_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize don't contain the data
     * Use the default value
     * Result: changed the externalizable data to set the default value
     * @throws Exception if an error occurs
     */
    @Test
    public void testExternalizableValueEmptyUsingDefault() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.LOCAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        _context.withDefaultFromModel(true);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_LOCAL_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_DEFAULT_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize don't contain the data
     * Use the default value
     * Update the status
     * Result: changed the externalizable data to set the default value and update the status
     * @throws Exception if an error occurs
     */
    @Test
    public void testExternalizableValueNullUsingDefaultUpdateStatus() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.LOCAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(null, ExternalizableDataStatus.EXTERNAL);
        value.setExternalizableStatus(ExternalizableDataStatus.EXTERNAL);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));
        
        _context.withDefaultFromModel(true);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.EXTERNAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_LOCAL_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_DEFAULT_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize don't contain the data
     * Use the default value
     * Result: changed the local data to set the default value
     * @throws Exception if an error occurs
     */
    @Test
    public void testLocalValueEmptyUsingDefault() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.LOCAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        _context.withDefaultFromModel(true)
            .withStatus(ExternalizableDataStatus.LOCAL);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_DEFAULT_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_EXTERNAL_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize don't contain the data
     * Do not use the default value
     * Update the status
     * Result: changed the externalizable data to empty it and update the status
     * @throws Exception if an error occurs
     */
    @Test
    public void testExternalizableValueNullUpdateStatus() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.LOCAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(null, ExternalizableDataStatus.EXTERNAL);
        value.setExternalizableStatus(ExternalizableDataStatus.EXTERNAL);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.EXTERNAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_LOCAL_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(StringUtils.EMPTY, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Data is externalizable
     * Values to synchronize does not change
     * Result: nothing changed
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotChangeExternalizableValue() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.LOCAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(__STRING_EXTERNAL_VALUE);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));

        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertFalse(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_LOCAL_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_EXTERNAL_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Data is externalizable
     * Values to synchronize does not change
     * Set the status to external
     * Result: changed to set status to external
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotChangeExternalizableValueUpdateStatus() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.EXTERNAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(__STRING_EXTERNAL_VALUE);
        value.setExternalizableStatus(ExternalizableDataStatus.LOCAL);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));

        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_LOCAL_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_EXTERNAL_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Data is externalizable
     * Values to synchronize does not change
     * The value to synchronize is the local one
     * Result: nothing changed
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotChangeLocalValue() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.LOCAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(__STRING_LOCAL_VALUE);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));

        _context.withStatus(ExternalizableDataStatus.LOCAL);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertFalse(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_LOCAL_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_EXTERNAL_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Data is externalizable
     * Values to synchronize does not change
     * The value to synchronize is the local one
     * Set the status to external
     * Result: changed to set status to external
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotChangeLocalValueUpdateStatus() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.EXTERNAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(__STRING_LOCAL_VALUE);
        value.setExternalizableStatus(ExternalizableDataStatus.LOCAL);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));

        _context.withStatus(ExternalizableDataStatus.LOCAL);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_LOCAL_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_EXTERNAL_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Data is externalizable
     * Values to synchronize changes
     * Result: changed to set the new externalizable value
     * @throws Exception if an error occurs
     */
    @Test
    public void testChangeExternalizableValue() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.LOCAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(__STRING_NEW_VALUE, ExternalizableDataStatus.EXTERNAL);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));

        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_LOCAL_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_NEW_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Data is externalizable
     * Values to synchronize changes
     * Set the status to local
     * Result: changed to set the new local value and put the status to local
     * @throws Exception if an error occurs
     */
    @Test
    public void testChangeExternalizableValueUpdateStatus() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.EXTERNAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(__STRING_NEW_VALUE, ExternalizableDataStatus.EXTERNAL);
        value.setExternalizableStatus(ExternalizableDataStatus.LOCAL);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));

        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_LOCAL_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_NEW_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Data is externalizable
     * Values to synchronize changes
     * The value to change is the local one
     * Result: changed to set the new local value
     * @throws Exception if an error occurs
     */
    @Test
    public void testChangeLocalValue() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.LOCAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(__STRING_NEW_VALUE);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));

        _context.withStatus(ExternalizableDataStatus.LOCAL);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_NEW_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_EXTERNAL_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Data is externalizable
     * Values to synchronize changes
     * The value to change is the local one
     * Set the status to external
     * Result: changed to set the new local value et put the status to external
     * @throws Exception if an error occurs
     */
    @Test
    public void testChangeLocalValueUpdateStatus() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.EXTERNAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(__STRING_NEW_VALUE);
        value.setExternalizableStatus(ExternalizableDataStatus.LOCAL);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));

        _context.withStatus(ExternalizableDataStatus.LOCAL);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_NEW_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_EXTERNAL_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Data is externalizable but has no status yet
     * By default, the status is forced when it is not present.
     * Result: the status has been forced to external, the local value is kept and the external one is set 
     * @throws Exception if an error occurs
     */
    @Test
    public void testForceStatusIfNotPresent() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", __STRING_EXTERNAL_VALUE)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.EXTERNAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_LOCAL_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_EXTERNAL_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Data is externalizable but has no status yet
     * In the context, the status is NOT forced when it is not present.
     * Result: the status is not forced and is still empty, the local value is kept and the external one is set 
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotForceStatusIfNotPresent() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", __STRING_EXTERNAL_VALUE)))));
        
        _context.withStatusForcedIfNotPresent(false);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_view, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_LOCAL_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_EXTERNAL_VALUE, _dataHolder.getExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Data is externalizable
     * Values to synchronize is an untouched value
     * The value to change is the local one
     * Result: the value is untouched => no changes
     * @throws Exception if an error occurs
     */
    @Test
    public void testUntouchLocalBinary() throws Exception
    {
        _dataHolder.setStatus(__BINARY_DATA_PATH, ExternalizableDataStatus.LOCAL);
        _dataHolder.setLocalValue(__BINARY_DATA_PATH, ResourceElementTypeHelperForTests.createBinary(__BINARY_LOCAL_CONTENT));
        _dataHolder.setExternalValue(__BINARY_DATA_PATH, ResourceElementTypeHelperForTests.createBinary(__BINARY_EXTERNAL_CONTENT));
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleBinaryData", new UntouchedValue())))));
        
        _context.withStatus(ExternalizableDataStatus.LOCAL);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_viewWithBinary, valuesToSynchronize, _context);
        
        Assert.assertFalse(result.hasChanged());
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Data is externalizable
     * Values to synchronize is an untouched value
     * The value to change is the local one
     * Result: the value is untouched => no changes
     * @throws Exception if an error occurs
     */
    @Test
    public void testUntouchLocalBinaryUpdateStatus() throws Exception
    {
        _dataHolder.setStatus(__BINARY_DATA_PATH, ExternalizableDataStatus.LOCAL);
        _dataHolder.setLocalValue(__BINARY_DATA_PATH, ResourceElementTypeHelperForTests.createBinary(__BINARY_LOCAL_CONTENT));
        _dataHolder.setExternalValue(__BINARY_DATA_PATH, ResourceElementTypeHelperForTests.createBinary(__BINARY_EXTERNAL_CONTENT));
        
        SynchronizableValue value = new SynchronizableValue(new UntouchedValue());
        value.setExternalizableStatus(ExternalizableDataStatus.EXTERNAL);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleBinaryData", value)))));
        
        _context.withStatus(ExternalizableDataStatus.LOCAL);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_viewWithBinary, valuesToSynchronize, _context);
        
        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.EXTERNAL, _dataHolder.getStatus(__BINARY_DATA_PATH));
        _assertOnBinaryContent(__BINARY_LOCAL_CONTENT, _dataHolder.getLocalValue(__BINARY_DATA_PATH));
        _assertOnBinaryContent(__BINARY_EXTERNAL_CONTENT, _dataHolder.getExternalValue(__BINARY_DATA_PATH));
    }
    
    private void _assertOnBinaryContent(String expectedContent, Binary binary) throws IOException
    {
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(binary.getInputStream(), writer, binary.getEncoding());
            Assert.assertEquals(expectedContent, writer.toString());
        }
    }
}
