/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.holder.impl.DefaultModifiableModelAwareDataHolder;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.DataComment;
import org.ametys.plugins.repository.data.external.ExternalizableDataProvider.ExternalizableDataStatus;
import org.ametys.plugins.repository.data.holder.ModifiableModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.values.SynchronizableValue;
import org.ametys.plugins.repository.data.holder.values.SynchronizationContext;
import org.ametys.plugins.repository.data.holder.values.SynchronizationResult;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method test cases
 */
@NeedFeatures({
    "test-data/test-model",
    
    "repository/repository.data.holder.helpers",
    "repository/resolver", 
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModifiableModelAwareDataHolderSynchronizeTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __DATA_NODE_TYPE = "ametys:unstructured";
    
    private static final String __STRING_DEFINITION_PATH = "composite/repeater/subcomposite/singleStringData";
    private static final String __STRING_DATA_PATH = "composite/repeater[1]/subcomposite/singleStringData";
    private static final String __MULTIPLE_STRING_DEFINITION_PATH = "multipleStringData";
    private static final String __MULTIPLE_STRING_DATA_PATH = "multipleStringData";
    private static final String __STRING_VALUE = "value";
    private static final String __STRING_LOCAL_VALUE = "local value";
    private static final String __STRING_EXTERNAL_VALUE = "external value";
    private static final String __STRING_NEW_VALUE = "new value";
    private static final String __STRING_DEFAULT_VALUE = "default value";
    private static final String __STRING_VALUE_2 = "value 2";
    private static final String __APPENDED_STRING_VALUE = "appended value";
    private static final String __APPENDED_STRING_VALUE_2 = "appended value 2";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static Model _model;

    private static JCRRepositoryData _root;
    private static View _completeView;
    private static View _viewWithOnlyMultipledata;
    private static View _emptyView;
    
    private ModifiableModelAwareDataHolder _dataHolder;
    private SynchronizationContext _context;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _root = new JCRRepositoryData(rootAmetysObject.getNode());
        
        // For these tests, add a default value to the single string data in the subcomposite
        @SuppressWarnings("unchecked")
        ElementDefinition<String> definition = (ElementDefinition<String>) _model.getModelItem(__STRING_DEFINITION_PATH);
        definition.setDefaultValue(__STRING_DEFAULT_VALUE);

        _completeView = View.of(_model, __STRING_DEFINITION_PATH, __MULTIPLE_STRING_DEFINITION_PATH);
        _viewWithOnlyMultipledata = View.of(_model, __MULTIPLE_STRING_DEFINITION_PATH);
        _emptyView = new View();
    }
    
    /**
     * Prepare components that are specific for each test
     * @throws Exception if an error occurs
     */
    @Before
    public void prepareSpecificComponents()throws Exception
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData(UUID.randomUUID().toString(), __DATA_NODE_TYPE);
        _dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        _dataHolder.getComposite("composite", true)
                   .getRepeater("repeater", true)
                   .addEntry()
                   .getComposite("subcomposite", true);

        _context = SynchronizationContext.newInstance();
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * The view is empty
     * Result: no change
     * @throws Exception if an error occurs
     */
    @Test
    public void testViewEmpty() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", __STRING_NEW_VALUE)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_emptyView, valuesToSynchronize, _context);

        Assert.assertFalse(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * The repository is empty
     * Values to synchronize don't contain the data
     * Use the default value
     * Result: changed to set the default value
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyValueEmptyUsingDefault() throws Exception
    {
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        _context.withDefaultFromModel(true);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_DEFAULT_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * The repository is empty
     * Values to synchronize contains a null value
     * Use the default value
     * Result: changed to set the default value
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyValueNullUsingDefault() throws Exception
    {
        Map<String, Object> emptyStringValue = new HashMap<>();
        emptyStringValue.put("singleStringData", null);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", emptyStringValue))));
        
        _context.withDefaultFromModel(true);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_DEFAULT_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * The repository is empty
     * Values to synchronize don't contain the data
     * Do not use the default value
     * Result: repo still empty, no change
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyValueEmpty() throws Exception
    {
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertFalse(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * The repository is empty
     * Values to synchronize contains a null value
     * Do not use the default value
     * Result: changed to set an empty value
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyValueNull() throws Exception
    {
        Map<String, Object> emptyStringValue = new HashMap<>();
        emptyStringValue.put("singleStringData", null);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", emptyStringValue))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(StringUtils.EMPTY, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * The repository is empty
     * Values to synchronize contains a  value
     * Result: changed to set the value
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptySetValue() throws Exception
    {
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", __STRING_VALUE)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize don't contain the data
     * Use the default value
     * Result: changed to set the default value
     * @throws Exception if an error occurs
     */
    @Test
    public void testValueEmptyUsingDefault() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        _context.withDefaultFromModel(true);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_DEFAULT_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize don't contain the data
     * Do not use the default value
     * Result: changed to remove the value
     * @throws Exception if an error occurs
     */
    @Test
    public void testValueEmpty() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize contains a null value
     * Do not use the default value
     * Result: changed to set an empty value
     * @throws Exception if an error occurs
     */
    @Test
    public void testValueNull() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        
        Map<String, Object> emptyStringValue = new HashMap<>();
        emptyStringValue.put("singleStringData", null);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", emptyStringValue))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(StringUtils.EMPTY, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize does not change the value
     * Result: no change
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotChangeValue() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", __STRING_VALUE)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertFalse(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize changes the value
     * Result: changed to set the new value
     * @throws Exception if an error occurs
     */
    @Test
    public void testChangeValue() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", __STRING_NEW_VALUE)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_NEW_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize remove a composite
     * Result: changed to remove the composite
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveComposite() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of())));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertFalse(_dataHolder.hasValue("composite/repeater[1]"));
        Assert.assertTrue(_dataHolder.hasValueOrEmpty("composite/repeater[1]"));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize remove a repeater entry
     * Result: changed to remove the repeater entry
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveRepeaterEntry() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of()));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertFalse(_dataHolder.hasValue("composite/repeater"));
        Assert.assertTrue(_dataHolder.hasValueOrEmpty("composite/repeater"));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize remove a repeater
     * Result: changed to remove the repeater
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveRepeater() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of());
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertFalse(_dataHolder.hasValue("composite"));
        Assert.assertTrue(_dataHolder.hasValueOrEmpty("composite"));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Insert a value on a single data
     * Result: changed to set the new value
     * @throws Exception if an error occurs
     */
    @Test
    public void testAppendValue() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(__STRING_NEW_VALUE);
        value.setMode(SynchronizableValue.Mode.APPEND);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_NEW_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Insert values on a multiple data
     * Result: changed to set the new values
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyAppendValues() throws Exception
    {
        SynchronizableValue value = new SynchronizableValue(new String[] {__APPENDED_STRING_VALUE, __APPENDED_STRING_VALUE_2});
        value.setMode(SynchronizableValue.Mode.APPEND);
        Map<String, Object> valuesToSynchronize = Map.of(__MULTIPLE_STRING_DATA_PATH, value);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_viewWithOnlyMultipledata, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__MULTIPLE_STRING_DATA_PATH));
        Assert.assertArrayEquals(new String[] {__APPENDED_STRING_VALUE, __APPENDED_STRING_VALUE_2}, _dataHolder.getLocalValue(__MULTIPLE_STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__MULTIPLE_STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Insert values on a multiple data
     * Result: changed to add the new values to the existent ones
     * @throws Exception if an error occurs
     */
    @Test
    public void testAppendValues() throws Exception
    {
        _dataHolder.setValue(__MULTIPLE_STRING_DATA_PATH, new String[] {__STRING_VALUE, __STRING_VALUE_2});
        
        SynchronizableValue value = new SynchronizableValue(new String[] {__APPENDED_STRING_VALUE, __APPENDED_STRING_VALUE_2});
        value.setMode(SynchronizableValue.Mode.APPEND);
        Map<String, Object> valuesToSynchronize = Map.of(__MULTIPLE_STRING_DATA_PATH, value);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_viewWithOnlyMultipledata, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__MULTIPLE_STRING_DATA_PATH));
        Assert.assertArrayEquals(new String[] {__STRING_VALUE, __STRING_VALUE_2, __APPENDED_STRING_VALUE, __APPENDED_STRING_VALUE_2}, _dataHolder.getLocalValue(__MULTIPLE_STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__MULTIPLE_STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Remove the value of a single data
     * Result: changed to remove the value
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveValue() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(null);
        value.setMode(SynchronizableValue.Mode.REMOVE);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Remove some values of a multiple data
     * Result: changed to remove some values
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveValues() throws Exception
    {
        _dataHolder.setValue(__MULTIPLE_STRING_DATA_PATH, new String[] {__STRING_VALUE, __STRING_VALUE_2});
        
        SynchronizableValue value = new SynchronizableValue(new String[] {__STRING_VALUE_2});
        value.setMode(SynchronizableValue.Mode.REMOVE);
        Map<String, Object> valuesToSynchronize = Map.of(__MULTIPLE_STRING_DATA_PATH, value);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_viewWithOnlyMultipledata, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__MULTIPLE_STRING_DATA_PATH));
        Assert.assertArrayEquals(new String[] {__STRING_VALUE}, _dataHolder.getLocalValue(__MULTIPLE_STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__MULTIPLE_STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Put empty array value for an empty multiple data
     * Result: no change
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotChangeEmptyValues() throws Exception
    {
        _dataHolder.setValue(__MULTIPLE_STRING_DATA_PATH, new String[] {});
        
        Map<String, Object> valuesToSynchronize = Map.of(__MULTIPLE_STRING_DATA_PATH, new String[] {});
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_viewWithOnlyMultipledata, valuesToSynchronize, _context);

        Assert.assertFalse(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__MULTIPLE_STRING_DATA_PATH));
        Assert.assertArrayEquals(new String[] {}, _dataHolder.getLocalValue(__MULTIPLE_STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__MULTIPLE_STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Put null value for an empty multiple data
     * Result: no change
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotChangeEmptyValuesWithNull() throws Exception
    {
        _dataHolder.setValue(__MULTIPLE_STRING_DATA_PATH, new String[] {});
        
        Map<String, Object> valuesToSynchronize = new HashMap<>();
        valuesToSynchronize.put(__MULTIPLE_STRING_DATA_PATH, null);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_viewWithOnlyMultipledata, valuesToSynchronize, _context);

        Assert.assertFalse(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__MULTIPLE_STRING_DATA_PATH));
        Assert.assertArrayEquals(new String[] {}, _dataHolder.getLocalValue(__MULTIPLE_STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__MULTIPLE_STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Repo contains externalizable metadata
     * Values to synchronize don't contain the data
     * Use the default value
     * Result: changed to set the default value and remove the externalizable meta
     * @throws Exception if an error occurs
     */
    @Test
    public void testValueEmptyUsingDefaultRemoveExternalizableMetadata() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.EXTERNAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        _context.withDefaultFromModel(true);
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_DEFAULT_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Repo contains externalizable metadata
     * Values to synchronize don't contain the data
     * Do not use the default value
     * Result: changed to remove the value and the externalizable meta
     * @throws Exception if an error occurs
     */
    @Test
    public void testValueEmptyRemoveExternalizableMetadata() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.EXTERNAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of()))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Repo contains externalizable metadata
     * Values to synchronize does not change the local value
     * Result: changed to remove the externalizable meta
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotChangeValueRemoveExternalizableMetadata() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.EXTERNAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", __STRING_LOCAL_VALUE)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_LOCAL_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Repo contains externalizable metadata
     * Values to synchronize set the external value as the new one
     * Result: do not change but remove the externalizable meta. Keep the external value as the current one
     * @throws Exception if an error occurs
     */
    @Test
    public void testUseExternalValueChangeStatusAndRemoveExternalizableMetadata() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.EXTERNAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(__STRING_VALUE);
        value.setExternalizableStatus(ExternalizableDataStatus.LOCAL);
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertFalse(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Repo contains externalizable metadata
     * Values to synchronize changes the value
     * Result: changed to set the new value and remove the externalizable meta
     * @throws Exception if an error occurs
     */
    @Test
    public void testChangeValueRemoveExternalizableMetadata() throws Exception
    {
        _dataHolder.setStatus(__STRING_DATA_PATH, ExternalizableDataStatus.EXTERNAL);
        _dataHolder.setLocalValue(__STRING_DATA_PATH, __STRING_LOCAL_VALUE);
        _dataHolder.setExternalValue(__STRING_DATA_PATH, __STRING_EXTERNAL_VALUE);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", __STRING_NEW_VALUE)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_NEW_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize does not contain comments, they are not touched
     * Result: no change
     * @throws Exception if an error occurs
     */
    @Test
    public void testIgnoreComments() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        _dataHolder.getComposite("composite/repeater[1]/subcomposite")
            .setComments("singleStringData", List.of(
                new DataComment("First comment", ZonedDateTime.of(2020, 10, 9, 20, 30, 0, 0, ZoneId.systemDefault()), "Café Plum"),
                new DataComment("Second comment", ZonedDateTime.of(2021, 2, 2, 20, 30, 0, 0, ZoneId.systemDefault()), "Rocher Palmer")));
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", __STRING_VALUE)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertFalse(result.hasChanged());
        
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
        
        List<DataComment> comments = _dataHolder.getComposite("composite/repeater[1]/subcomposite")
                .getComments("singleStringData");
        Assert.assertEquals(2, comments.size());
        Assert.assertEquals("First comment", comments.get(0).getComment());
        Assert.assertEquals("Café Plum", comments.get(0).getAuthor());
        Assert.assertEquals("Second comment", comments.get(1).getComment());
        Assert.assertEquals("Rocher Palmer", comments.get(1).getAuthor());
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize contains comments but they are unchanged
     * Result: no change
     * @throws Exception if an error occurs
     */
    @Test
    public void testDoNotChangeComments() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        _dataHolder.getComposite("composite/repeater[1]/subcomposite")
            .setComments("singleStringData", List.of(
                new DataComment("First comment", ZonedDateTime.of(2020, 10, 9, 20, 30, 0, 0, ZoneId.systemDefault()), "Café Plum"),
                new DataComment("Second comment", ZonedDateTime.of(2021, 2, 2, 20, 30, 0, 0, ZoneId.systemDefault()), "Rocher Palmer")));
        
        SynchronizableValue value = new SynchronizableValue(__STRING_VALUE);
        value.setComments(List.of(
                new DataComment("First comment", ZonedDateTime.of(2020, 10, 9, 20, 30, 0, 0, ZoneId.systemDefault()), "Café Plum"),
                new DataComment("Second comment", ZonedDateTime.of(2021, 2, 2, 20, 30, 0, 0, ZoneId.systemDefault()), "Rocher Palmer")));
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertFalse(result.hasChanged());
        
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
        
        List<DataComment> comments = _dataHolder.getComposite("composite/repeater[1]/subcomposite")
                .getComments("singleStringData");
        Assert.assertEquals(2, comments.size());
        Assert.assertEquals("First comment", comments.get(0).getComment());
        Assert.assertEquals("Café Plum", comments.get(0).getAuthor());
        Assert.assertEquals("Second comment", comments.get(1).getComment());
        Assert.assertEquals("Rocher Palmer", comments.get(1).getAuthor());
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize contains new comments
     * @throws Exception if an error occurs
     */
    @Test
    public void testAddComments() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        
        SynchronizableValue value = new SynchronizableValue(__STRING_VALUE);
        value.setComments(List.of(
                new DataComment("New comment", ZonedDateTime.of(2021, 3, 19, 20, 0, 0, 0, ZoneId.systemDefault()), "Victoire 2")));
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
        
        List<DataComment> comments = _dataHolder.getComposite("composite/repeater[1]/subcomposite")
                .getComments("singleStringData");
        Assert.assertEquals(1, comments.size());
        Assert.assertEquals("New comment", comments.get(0).getComment());
        Assert.assertEquals("Victoire 2", comments.get(0).getAuthor());
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize contains new comments
     * @throws Exception if an error occurs
     */
    @Test
    public void testChangeComments() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        _dataHolder.getComposite("composite/repeater[1]/subcomposite")
            .setComments("singleStringData", List.of(
                new DataComment("First comment", ZonedDateTime.of(2020, 10, 9, 20, 30, 0, 0, ZoneId.systemDefault()), "Café Plum"),
                new DataComment("Second comment", ZonedDateTime.of(2021, 2, 2, 20, 30, 0, 0, ZoneId.systemDefault()), "Rocher Palmer")));
        
        SynchronizableValue value = new SynchronizableValue(__STRING_VALUE);
        value.setComments(List.of(
                new DataComment("New comment", ZonedDateTime.of(2021, 3, 19, 20, 0, 0, 0, ZoneId.systemDefault()), "Victoire 2")));
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
        
        List<DataComment> comments = _dataHolder.getComposite("composite/repeater[1]/subcomposite")
                .getComments("singleStringData");
        Assert.assertEquals(1, comments.size());
        Assert.assertEquals("New comment", comments.get(0).getComment());
        Assert.assertEquals("Victoire 2", comments.get(0).getAuthor());
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method
     * Values to synchronize contains an empty list of comments to remove existing ones
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveComments() throws Exception
    {
        _dataHolder.setValue(__STRING_DATA_PATH, __STRING_VALUE);
        _dataHolder.getComposite("composite/repeater[1]/subcomposite")
            .setComments("singleStringData", List.of(
                new DataComment("First comment", ZonedDateTime.of(2020, 10, 9, 20, 30, 0, 0, ZoneId.systemDefault()), "Café Plum"),
                new DataComment("Second comment", ZonedDateTime.of(2021, 2, 2, 20, 30, 0, 0, ZoneId.systemDefault()), "Rocher Palmer")));
        
        SynchronizableValue value = new SynchronizableValue(__STRING_VALUE);
        value.setComments(List.of());
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of("subcomposite", Map.of("singleStringData", value)))));
        
        SynchronizationResult result = _dataHolder.synchronizeValues(_completeView, valuesToSynchronize, _context);

        Assert.assertTrue(result.hasChanged());
        
        Assert.assertEquals(ExternalizableDataStatus.LOCAL, _dataHolder.getStatus(__STRING_DATA_PATH));
        Assert.assertEquals(__STRING_VALUE, _dataHolder.getLocalValue(__STRING_DATA_PATH));
        Assert.assertFalse(_dataHolder.hasExternalValue(__STRING_DATA_PATH));
        
        Assert.assertFalse(_dataHolder.getComposite("composite/repeater[1]/subcomposite")
                .hasComments("singleStringData"));
    }
}
