/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringWriter;
import java.time.LocalDate;
import java.time.Month;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Calendar;
import java.util.Collection;
import java.util.List;
import java.util.Locale;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Binary;
import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.ExplorerFile;
import org.ametys.cms.data.File;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.Reference;
import org.ametys.cms.data.Resource;
import org.ametys.cms.data.RichText;
import org.ametys.cms.data.holder.impl.DefaultModifiableModelAwareDataHolder;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.data.type.ResourceElementTypeHelper;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.cms.utils.ResourceElementTypeHelperForTests;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.holder.ModifiableModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModifiableComposite;
import org.ametys.plugins.repository.data.holder.group.ModifiableRepeater;
import org.ametys.plugins.repository.data.holder.group.ModifiableRepeaterEntry;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.plugins.repository.data.type.RepositoryModelItemType;
import org.ametys.plugins.repository.metadata.MultilingualString;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModifiableModelAwareDataHolder} implementation test case
 */
@NeedFeatures({
    "test-data/test-model",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModifiableModelAwareDataHolderTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    
    private static Model _model;
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    
    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;

    private static RepositoryElementType<UserIdentity> _userType;
    private static RepositoryElementType<Geocode> _geocodeType;
    private static RepositoryElementType<MultilingualString> _multilingualStringType;
    private static RepositoryElementType<Reference> _referenceType;
    private static RepositoryElementType<Binary> _binaryType;
    private static RepositoryElementType<File> _fileType;
    private static RepositoryElementType<RichText> _richTextType;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
    }
    
    @Override
    @SuppressWarnings("unchecked")
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");

        _userType = (RepositoryElementType<UserIdentity>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _geocodeType = (RepositoryElementType<Geocode>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID);
        _multilingualStringType = (RepositoryElementType<MultilingualString>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID);
        _referenceType = (RepositoryElementType<Reference>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.REFERENCE_ELEMENT_TYPE_ID);
        _binaryType = (RepositoryElementType<Binary>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _fileType = (RepositoryElementType<File>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.FILE_ELEMENT_TYPE_ID);
        _richTextType = (RepositoryElementType<RichText>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for simple data types
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetValueForSimpleDataTypesWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueForSimpleDataTypes", DATA_NODE_TYPE);
        _testSetValueForSimpleDataTypes(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for simple data types
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueForSimpleDataTypesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueForSimpleDataTypes", DATA_NODE_TYPE);
        _testSetValueForSimpleDataTypes(repositoryData);
    }
    
    private void _testSetValueForSimpleDataTypes(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write
        dataHolder.setValue("singleStringData", "my string");
        String[] stringArray = new String[]{"string1", "string2", "string3"};
        dataHolder.setValue("multipleStringData", stringArray);
        dataHolder.setValue("singleLongData", 67L);
        Long[] longArray = new Long[]{65L, 98L, 432L};
        dataHolder.setValue("multipleLongData", longArray);
        dataHolder.setValue("singleDoubleData", 54D);
        Double[] doubleArray = new Double[]{87D, 345D};
        dataHolder.setValue("multipleDoubleData", doubleArray);
        dataHolder.setValue("singleBooleanData", true);
        Boolean[] booleanArray = new Boolean[]{true, true, false, true, false};
        dataHolder.setValue("multipleBooleanData", booleanArray);
        
        // Read
        Assert.assertEquals("my string", repositoryData.getString("singleStringData"));
        Assert.assertFalse(repositoryData.isMultiple("singleStringData"));
        Assert.assertArrayEquals(stringArray, repositoryData.getStrings("multipleStringData"));
        Assert.assertTrue(repositoryData.isMultiple("multipleStringData"));
        Assert.assertEquals(Long.valueOf(67), repositoryData.getLong("singleLongData"));
        Assert.assertFalse(repositoryData.isMultiple("singleLongData"));
        Assert.assertArrayEquals(longArray, repositoryData.getLongs("multipleLongData"));
        Assert.assertTrue(repositoryData.isMultiple("multipleLongData"));
        Assert.assertEquals(Double.valueOf(54), repositoryData.getDouble("singleDoubleData"));
        Assert.assertFalse(repositoryData.isMultiple("singleDoubleData"));
        Assert.assertArrayEquals(doubleArray, repositoryData.getDoubles("multipleDoubleData"));
        Assert.assertTrue(repositoryData.isMultiple("multipleDoubleData"));
        Assert.assertTrue(repositoryData.getBoolean("singleBooleanData"));
        Assert.assertFalse(repositoryData.isMultiple("singleBooleanData"));
        Assert.assertArrayEquals(booleanArray, repositoryData.getBooleans("multipleBooleanData"));
        Assert.assertTrue(repositoryData.isMultiple("multipleBooleanData"));
        
        try
        {
            dataHolder.setValue("singleBooleanData", "true");
            Assert.fail("We should not be able to set a string value for a boolean parameter");
        }
        catch (BadItemTypeException e)
        {
            Assert.assertTrue(e.getMessage().contains("singleBooleanData"));
            Assert.assertTrue(e.getMessage().contains("boolean"));
        }
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for date and date time types
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetValueForDateDataTypesWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueForDateDataTypes", DATA_NODE_TYPE);
        _testSetValueForDateDataTypes(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for date and date time types
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueForDateDataTypesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueForDateDataTypes", DATA_NODE_TYPE);
        _testSetValueForDateDataTypes(repositoryData);
    }
    
    private void _testSetValueForDateDataTypes(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Date - write
        dataHolder.setValue("singleDateData", LocalDate.of(2018, Month.MAY, 18));
        LocalDate[] dateArray = new LocalDate[]{LocalDate.of(2018, Month.MAY, 5), LocalDate.of(2015, Month.DECEMBER, 06)};
        dataHolder.setValue("multipleDateData", dateArray);
        
        // Date - read
        Assert.assertFalse(repositoryData.isMultiple("singleDateData"));
        Calendar calendar = repositoryData.getDate("singleDateData");
        Assert.assertEquals(18, calendar.get(Calendar.DAY_OF_MONTH));
        Assert.assertEquals(Calendar.MAY, calendar.get(Calendar.MONTH));
        Assert.assertEquals(2018, calendar.get(Calendar.YEAR));
        
        Assert.assertTrue(repositoryData.isMultiple("multipleDateData"));
        Calendar[] calendarArray = repositoryData.getDates("multipleDateData");
        Assert.assertEquals(2, calendarArray.length);
        Calendar calendar1 = calendarArray[0];
        Assert.assertEquals(5, calendar1.get(Calendar.DAY_OF_MONTH));
        Assert.assertEquals(Calendar.MAY, calendar1.get(Calendar.MONTH));
        Assert.assertEquals(2018, calendar1.get(Calendar.YEAR));
        Calendar calendar2 = calendarArray[1];
        Assert.assertEquals(6, calendar2.get(Calendar.DAY_OF_MONTH));
        Assert.assertEquals(Calendar.DECEMBER, calendar2.get(Calendar.MONTH));
        Assert.assertEquals(2015, calendar2.get(Calendar.YEAR));
        
        // DateTime - write
        dataHolder.setValue("singleDatetimeData", ZonedDateTime.of(2018, 5, 18, 14, 15, 31, 0, ZoneId.systemDefault()));
        ZonedDateTime[] dateTimeArray = new ZonedDateTime[]{ZonedDateTime.of(2018, 5, 5, 7, 25, 13, 0, ZoneId.of("UTC+05:00")), ZonedDateTime.of(2015, 12, 06, 15, 18, 0, 0, ZoneId.of("UTC-03:00"))};
        dataHolder.setValue("multipleDatetimeData", dateTimeArray);
        
        // DateTime - read
        Assert.assertFalse(repositoryData.isMultiple("singleDatetimeData"));
        calendar = repositoryData.getDate("singleDatetimeData");
        ZonedDateTime zonedDateTime = DateUtils.asZonedDateTime(calendar);
        zonedDateTime = zonedDateTime.withZoneSameInstant(ZoneId.systemDefault());
        Assert.assertEquals(2018, zonedDateTime.getYear());
        Assert.assertEquals(Month.MAY, zonedDateTime.getMonth());
        Assert.assertEquals(18, zonedDateTime.getDayOfMonth());
        Assert.assertEquals(14, zonedDateTime.getHour());
        Assert.assertEquals(15, zonedDateTime.getMinute());
        Assert.assertEquals(31, zonedDateTime.getSecond());
        
        Assert.assertTrue(repositoryData.isMultiple("multipleDatetimeData"));
        calendarArray = repositoryData.getDates("multipleDatetimeData");
        Assert.assertEquals(2, calendarArray.length);
        ZonedDateTime zonedDateTime1 = DateUtils.asZonedDateTime(calendarArray[0]);
        zonedDateTime1 = zonedDateTime1.withZoneSameInstant(ZoneId.of("UTC+05:00"));
        Assert.assertEquals(2018, zonedDateTime1.getYear());
        Assert.assertEquals(Month.MAY, zonedDateTime1.getMonth());
        Assert.assertEquals(05, zonedDateTime1.getDayOfMonth());
        Assert.assertEquals(07, zonedDateTime1.getHour());
        Assert.assertEquals(25, zonedDateTime1.getMinute());
        Assert.assertEquals(13, zonedDateTime1.getSecond());
        ZonedDateTime zonedDateTime2 = DateUtils.asZonedDateTime(calendarArray[1]);
        zonedDateTime2 = zonedDateTime2.withZoneSameInstant(ZoneId.of("UTC-03:00"));
        Assert.assertEquals(2015, zonedDateTime2.getYear());
        Assert.assertEquals(Month.DECEMBER, zonedDateTime2.getMonth());
        Assert.assertEquals(06, zonedDateTime2.getDayOfMonth());
        Assert.assertEquals(15, zonedDateTime2.getHour());
        Assert.assertEquals(18, zonedDateTime2.getMinute());
        Assert.assertEquals(00, zonedDateTime2.getSecond());
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for complex data types
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetValueForComplexDataTypesWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueForComplexDataTypes", DATA_NODE_TYPE);
        _testSetValueForComplexDataTypes(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for complex data types
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueForComplexDataTypesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueForComplexDataTypes", DATA_NODE_TYPE);
        _testSetValueForComplexDataTypes(repositoryData);
    }
    
    private void _testSetValueForComplexDataTypes(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // User - write
        dataHolder.setValue("singleUserData", new UserIdentity("tania", "anyPopulation"));
        UserIdentity[] userArray = new UserIdentity[]{new UserIdentity("kevin", "anyPopulation"), new UserIdentity("georges", "anyPopulation"), new UserIdentity("john", "anyPopulation")};
        dataHolder.setValue("multipleUserData", userArray);
        
        // User - read
        Assert.assertFalse(repositoryData.isMultiple("singleUserData"));
        UserIdentity user = (UserIdentity) _userType.read(repositoryData, "singleUserData");
        Assert.assertEquals("tania", user.getLogin());
        Assert.assertEquals("anyPopulation", user.getPopulationId());
        Assert.assertTrue(repositoryData.isMultiple("multipleUserData"));
        Assert.assertArrayEquals(userArray, (UserIdentity[]) _userType.read(repositoryData, "multipleUserData"));
        
        // User multiple - write again
        UserIdentity[] userArray2 = new UserIdentity[]{new UserIdentity("nathalie", "anyPopulation")};
        dataHolder.setValue("multipleUserData", userArray2);
        
        // User multiple - read again
        Assert.assertArrayEquals(userArray2, (UserIdentity[]) _userType.read(repositoryData, "multipleUserData"));
        
        // Geocode - write
        dataHolder.setValue("singleGeocodeData", new Geocode(43.5470148D, 1.5134120999999823D));
        Geocode[] geocodeArray = new Geocode[]{new Geocode(0.0D, 0.0D), new Geocode(23.5, 0.78)};
        dataHolder.setValue("multipleGeocodeData", geocodeArray);
        
        // Geocode - read
        Assert.assertFalse(repositoryData.isMultiple("singleGeocodeData"));
        Geocode geocode = (Geocode) _geocodeType.read(repositoryData, "singleGeocodeData");
        Assert.assertEquals(Double.valueOf(1.5134120999999823), geocode.getLongitude());
        Assert.assertEquals(Double.valueOf(43.5470148D), geocode.getLatitude());
        Assert.assertTrue(repositoryData.isMultiple("multipleGeocodeData"));
        geocodeArray = (Geocode[]) _geocodeType.read(repositoryData, "multipleGeocodeData");
        Assert.assertEquals(2, geocodeArray.length);
        Assert.assertEquals(Double.valueOf(0.0), geocodeArray[0].getLongitude());
        Assert.assertEquals(Double.valueOf(0.0), geocodeArray[0].getLatitude());
        Assert.assertEquals(Double.valueOf(0.78), geocodeArray[1].getLongitude());
        Assert.assertEquals(Double.valueOf(23.5), geocodeArray[1].getLatitude());
        
        // Multilingual String - write
        MultilingualString multilingualString = new MultilingualString();
        multilingualString.add(Locale.FRANCE, "ma valeur");
        multilingualString.add(Locale.UK, "my value");
        dataHolder.setValue("singleMultilingualStringData", multilingualString);
        
        MultilingualString multilingualString1 = new MultilingualString();
        multilingualString1.add(Locale.CANADA, "tabernacle");
        multilingualString1.add(Locale.GERMANY, "der Kaffee ist fertig");
        MultilingualString multilingualString2 = new MultilingualString();
        multilingualString2.add(Locale.ITALY, "Bella ciao");
        MultilingualString[] multilingualStringArray = new MultilingualString[]{multilingualString1, multilingualString2};
        dataHolder.setValue("multipleMultilingualStringData", multilingualStringArray);
        
        // Multilingual string - read
        Assert.assertFalse(repositoryData.isMultiple("singleMultilingualStringData"));
        MultilingualString retrievedMultilingualString = (MultilingualString) _multilingualStringType.read(repositoryData, "singleMultilingualStringData");
        Assert.assertEquals(2, retrievedMultilingualString.getValues().size());
        Assert.assertEquals("ma valeur", retrievedMultilingualString.getValue(Locale.FRANCE));
        Assert.assertEquals("my value", retrievedMultilingualString.getValue(Locale.UK));
        
        Assert.assertTrue(repositoryData.isMultiple("multipleMultilingualStringData"));
        MultilingualString[] retrievedMultilingualStringArray = (MultilingualString[]) _multilingualStringType.read(repositoryData, "multipleMultilingualStringData");
        Assert.assertEquals(2, retrievedMultilingualStringArray.length);
        
        MultilingualString retrievedMultilingualString1 = retrievedMultilingualStringArray[0];
        Assert.assertEquals(2, retrievedMultilingualString1.getValues().size());
        Assert.assertEquals("tabernacle", retrievedMultilingualString1.getValue(Locale.CANADA));
        Assert.assertEquals("der Kaffee ist fertig", retrievedMultilingualString1.getValue(Locale.GERMANY));

        MultilingualString retrievedMultilingualString2 = retrievedMultilingualStringArray[1];
        Assert.assertEquals(1, retrievedMultilingualString2.getValues().size());
        Assert.assertEquals("Bella ciao", retrievedMultilingualString2.getValue(Locale.ITALY));
        
        // Reference - write
        dataHolder.setValue("singleReferenceData", new Reference("jcr/path/to/the/metadata", "metadata"));
        repositoryData.getDataNames();
        Reference[] referenceArray = new Reference[]{new Reference("jcr/path/to/the/metadata", "metadata"), new Reference("path/to/the/attachment", "attachment")};
        dataHolder.setValue("multipleReferenceData", referenceArray);
        
        // Reference - read
        Assert.assertFalse(repositoryData.isMultiple("singleReferenceData"));
        Reference reference = (Reference) _referenceType.read(repositoryData, "singleReferenceData");
        Assert.assertEquals("jcr/path/to/the/metadata", reference.getValue());
        Assert.assertEquals("metadata", reference.getType());
        Assert.assertTrue(repositoryData.isMultiple("multipleReferenceData"));
        referenceArray = (Reference[]) _referenceType.read(repositoryData, "multipleReferenceData");
        Assert.assertEquals(2, referenceArray.length);
        Assert.assertEquals("jcr/path/to/the/metadata", referenceArray[0].getValue());
        Assert.assertEquals("metadata", referenceArray[0].getType());
        Assert.assertEquals("path/to/the/attachment", referenceArray[1].getValue());
        Assert.assertEquals("attachment", referenceArray[1].getType());
        
        // Reference without type - write
        dataHolder.setValue("singleReferenceData", new Reference("https://docs.ametys.org/fr/index.html"));
        
        // Reference without type - read
        reference = (Reference) _referenceType.read(repositoryData, "singleReferenceData");
        Assert.assertEquals("https://docs.ametys.org/fr/index.html", reference.getValue());
        Assert.assertEquals(Reference.DEFAULT_TYPE, reference.getType());
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for resources data types
     * This test uses a {@link JCRRepositoryData}
     * @throws IOException if an error occurs while writing or reading resources
     */
    @Test
    @NeedFeatures({"explorer/factories", })
    public void testSetValueForResourceDataTypesWithJCRRepositoryData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueForResourceDataTypes", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Binary - write
        Binary binary = ResourceElementTypeHelperForTests.createBinary("my binary content", "UTF-8", "myBinary.txt", ZonedDateTime.of(2019, 03, 12, 17, 22, 38, 0, ZoneId.systemDefault()), "text/plain");
        dataHolder.setValue("singleBinaryData", binary);
        
        // Binary - read
        Assert.assertFalse(repositoryData.isMultiple("singleBinaryData"));
        Binary retrievedBinary = (Binary) _binaryType.read(repositoryData, "singleBinaryData");
        Assert.assertEquals("UTF-8", retrievedBinary.getEncoding());
        Assert.assertEquals("myBinary.txt", retrievedBinary.getFilename());
        _assertZonedDateTimeEquals(ZoneId.systemDefault(), 2019, Month.MARCH, 12, 17, 22, 38, retrievedBinary.getLastModificationDate());
        Assert.assertEquals("text/plain", retrievedBinary.getMimeType());
        _assertEqualsStreamContent(repositoryData.getRepositoryData("singleBinaryData"), "my binary content");
        
        // Multiple Binary - write
        Binary binary1 = ResourceElementTypeHelperForTests.createBinary("my first binary content");
        Binary binary2 = ResourceElementTypeHelperForTests.createBinary("my second binary content");
        Binary binary3 = ResourceElementTypeHelperForTests.createBinary("my third binary content");
        dataHolder.setValue("multipleBinaryData", new Binary[] {binary1, binary2, binary3});
        
        // Multiple Binary - read
        Assert.assertTrue(repositoryData.isMultiple("multipleBinaryData"));
        Assert.assertEquals(3, repositoryData.getRepositoryData("multipleBinaryData").getDataNames().size());
        _assertEqualsStreamContent(repositoryData.getRepositoryData("multipleBinaryData").getRepositoryData("1"), "my first binary content");
        _assertEqualsStreamContent(repositoryData.getRepositoryData("multipleBinaryData").getRepositoryData("2"), "my second binary content");
        _assertEqualsStreamContent(repositoryData.getRepositoryData("multipleBinaryData").getRepositoryData("3"), "my third binary content");
        
        // Multiple Binary - write
        dataHolder.setValue("multipleBinaryData", new Binary[] {binary2});
        
        // Multiple Binary - read
        Assert.assertTrue(repositoryData.isMultiple("multipleBinaryData"));
        Assert.assertEquals(1, repositoryData.getRepositoryData("multipleBinaryData").getDataNames().size());
        _assertEqualsStreamContent(repositoryData.getRepositoryData("multipleBinaryData").getRepositoryData("1"), "my second binary content");
        
        // File as explorer resource - write
        ExplorerFile file = ResourceElementTypeHelperForTests.createExplorerFile("myFile.txt", "my explorer file content", _ametysObjectResolver);
        dataHolder.setValue("singleFileData", file);
        
        // File as explorer resource - read
        Assert.assertFalse(repositoryData.isMultiple("singleFileData"));
        _assertFileDataHasNoBinaryNode(repositoryData, "singleFileData");
        Assert.assertEquals(file.getResourceId(), repositoryData.getString("singleFileData"));
        
        // File as Binary - write
        dataHolder.setValue("singleFileData", binary);
        
        // File as Binary - read
        Assert.assertFalse(repositoryData.isMultiple("singleFileData"));
        _assertFileDataHasNoStringProperty(repositoryData, "singleFileData");
        File retrievedFileAsBinary = (File) _fileType.read(repositoryData, "singleFileData");
        Assert.assertEquals("myBinary.txt", retrievedFileAsBinary.getName());
        _assertZonedDateTimeEquals(ZoneId.systemDefault(), 2019, Month.MARCH, 12, 17, 22, 38, retrievedFileAsBinary.getLastModificationDate());
        Assert.assertEquals("text/plain", retrievedFileAsBinary.getMimeType());
        _assertEqualsStreamContent(repositoryData.getRepositoryData("singleFileData"), "my binary content");
        
        // File as string - write
        String id = ResourceElementTypeHelperForTests.createResource("myfilename", "my file content", _ametysObjectResolver);
        dataHolder.setValue("singleFileData", id);
        
        // File as string - read
        Assert.assertFalse(repositoryData.isMultiple("singleFileData"));
        _assertFileDataHasNoBinaryNode(repositoryData, "singleFileData");
        Assert.assertEquals(id, repositoryData.getString("singleFileData"));
        
        // Multiple file as explorer resource - write
        ExplorerFile file1 = ResourceElementTypeHelperForTests.createExplorerFile("myFile1.txt", "my first explorer file content", _ametysObjectResolver);
        ExplorerFile file2 = ResourceElementTypeHelperForTests.createExplorerFile("myFile2.txt", "my second explorer file content", _ametysObjectResolver);
        ExplorerFile file3 = ResourceElementTypeHelperForTests.createExplorerFile("myFile3.txt", "my third explorer file content", _ametysObjectResolver);
        dataHolder.setValue("multipleFileData", new ExplorerFile[] {file1, file2, file3});
        
        // Multiple file as explorer resource - read
        Assert.assertTrue(repositoryData.isMultiple("multipleFileData"));
        _assertFileDataHasNoBinaryNode(repositoryData, "multipleFileData");
        Assert.assertArrayEquals(new String[] {file1.getResourceId(), file2.getResourceId(), file3.getResourceId()}, repositoryData.getStrings("multipleFileData"));
        
        // Multiple file as explorer resource - write
        dataHolder.setValue("multipleFileData", new ExplorerFile[] {file1, file3});
        
        // Multiple file as explorer resource - read
        Assert.assertTrue(repositoryData.isMultiple("multipleFileData"));
        _assertFileDataHasNoBinaryNode(repositoryData, "multipleFileData");
        Assert.assertArrayEquals(new String[] {file1.getResourceId(), file3.getResourceId()}, repositoryData.getStrings("multipleFileData"));
        
        // Multiple file as binary - write
        dataHolder.setValue("multipleFileData", new Binary[] {binary1, binary2, binary3});

        // Multiple file as binary - read
        Assert.assertTrue(repositoryData.isMultiple("multipleFileData"));
        _assertFileDataHasNoStringProperty(repositoryData, "multipleFileData");
        Assert.assertEquals(3, repositoryData.getRepositoryData("multipleFileData").getDataNames().size());
        _assertEqualsStreamContent(repositoryData.getRepositoryData("multipleFileData").getRepositoryData("1"), "my first binary content");
        _assertEqualsStreamContent(repositoryData.getRepositoryData("multipleFileData").getRepositoryData("2"), "my second binary content");
        _assertEqualsStreamContent(repositoryData.getRepositoryData("multipleFileData").getRepositoryData("3"), "my third binary content");
        
        // Multiple file as binary - write
        Binary binary4 = ResourceElementTypeHelperForTests.createBinary("my fourth binary content");
        dataHolder.setValue("multipleFileData", new Binary[] {binary4});

        // Multiple file as binary - read
        Assert.assertTrue(repositoryData.isMultiple("multipleFileData"));
        _assertFileDataHasNoStringProperty(repositoryData, "multipleFileData");
        Assert.assertEquals(1, repositoryData.getRepositoryData("multipleFileData").getDataNames().size());
        _assertEqualsStreamContent(repositoryData.getRepositoryData("multipleFileData").getRepositoryData("1"), "my fourth binary content");

        // Multiple file as string - write
        String id1 = ResourceElementTypeHelperForTests.createResource("myfilename1", "my first file content", _ametysObjectResolver);
        String id2 = ResourceElementTypeHelperForTests.createResource("myfilename2", "my second file content", _ametysObjectResolver);
        String id3 = ResourceElementTypeHelperForTests.createResource("myfilename3", "my third file content", _ametysObjectResolver);
        dataHolder.setValue("multipleFileData", new String[] {id1, id2, id3});
        
        // Multiple file as string - read
        Assert.assertTrue(repositoryData.isMultiple("multipleFileData"));
        _assertFileDataHasNoBinaryNode(repositoryData, "multipleFileData");
        Assert.assertArrayEquals(new String[] {id1, id2, id3}, repositoryData.getStrings("multipleFileData"));

        // Multiple file as string - write
        String id4 = ResourceElementTypeHelperForTests.createResource("myfilename4", "my fourth file content", _ametysObjectResolver);
        dataHolder.setValue("multipleFileData", new String[] {id4});
        
        // Multiple file as string - read
        Assert.assertTrue(repositoryData.isMultiple("multipleFileData"));
        _assertFileDataHasNoBinaryNode(repositoryData, "multipleFileData");
        Assert.assertArrayEquals(new String[] {id4}, repositoryData.getStrings("multipleFileData"));
        
        // Rich text - write
        RichText richText = new RichText();
        try (OutputStream os = richText.getOutputStream())
        {
            os.write("my rich text content".getBytes());
        }
        richText.setEncoding("UTF-8");
        richText.setMimeType("text/xml");
        richText.setLastModificationDate(ZonedDateTime.of(2019, 03, 20, 13, 34, 12, 0, ZoneId.systemDefault()));
        // Annotations
        richText.addAnnotations("annotations1", "a1-1", "a1-2", "a1-3");
        richText.addAnnotations("annotations2", "a2-1", "a2-2");
        // Files
        ResourceElementTypeHelperForTests.addFilesToRichText(richText);
        dataHolder.setValue("singleRichTextData", richText);
        
        // Rich text - read
        Assert.assertFalse(repositoryData.isMultiple("singleRichTextData"));
        RichText retrievedRichText = (RichText) _richTextType.read(repositoryData, "singleRichTextData");
        Assert.assertEquals("UTF-8", retrievedRichText.getEncoding());
        _assertZonedDateTimeEquals(ZoneId.systemDefault(), 2019, Month.MARCH, 20, 13, 34, 12, retrievedRichText.getLastModificationDate());
        Assert.assertEquals("text/xml", retrievedRichText.getMimeType());
        _assertEqualsStreamContent(repositoryData.getRepositoryData("singleRichTextData"), "my rich text content");
        // Annotations
        Assert.assertEquals(2, retrievedRichText.getAllAnnotations().size());
        List<String> annotations1 = retrievedRichText.getAnnotations("annotations1");
        Assert.assertEquals(3, annotations1.size());
        Assert.assertTrue(annotations1.contains("a1-1"));
        Assert.assertTrue(annotations1.contains("a1-2"));
        Assert.assertTrue(annotations1.contains("a1-3"));
        List<String> annotations2 = retrievedRichText.getAnnotations("annotations2");
        Assert.assertEquals(2, annotations2.size());
        Assert.assertTrue(annotations2.contains("a2-1"));
        Assert.assertTrue(annotations2.contains("a2-2"));
        // Files
        Assert.assertEquals(4,  retrievedRichText.getAttachments().size());
        Resource retrievedFile1 = retrievedRichText.getAttachment("myFirstFile.txt");
        Assert.assertEquals("UTF-8", retrievedFile1.getEncoding());
        _assertZonedDateTimeEquals(ZoneId.systemDefault(), 2019, Month.MARCH, 20, 14, 03, 59, retrievedFile1.getLastModificationDate());
        Assert.assertEquals("text/plain", retrievedFile1.getMimeType());
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFile1.getInputStream(), writer, retrievedFile1.getEncoding());
            Assert.assertEquals("my first file content", writer.toString());
        }
        Resource retrievedFile2 = retrievedRichText.getAttachment("mySecondFile.txt");
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFile2.getInputStream(), writer, retrievedFile2.getEncoding());
            Assert.assertEquals("my second file content", writer.toString());
        }
        Resource retrievedFile3 = retrievedRichText.getAttachment("myThirdFile.txt");
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFile3.getInputStream(), writer, retrievedFile3.getEncoding());
            Assert.assertEquals("my third file content", writer.toString());
        }
        Resource retrievedFile4 = retrievedRichText.getAttachment("myFourthFile.txt");
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFile4.getInputStream(), writer, retrievedFile4.getEncoding());
            Assert.assertEquals("my fourth file content", writer.toString());
        }
        
        // RichText - add and delete files
        ResourceElementTypeHelperForTests.playWithRichTextFiles(retrievedRichText);
        Collection<String> attachmentNames = retrievedRichText.getAttachmentNames();
        Assert.assertEquals(5, attachmentNames.size());
        Assert.assertTrue(attachmentNames.contains("mySecondFile.txt"));
        Assert.assertTrue(attachmentNames.contains("myThirdFile.txt"));
        Assert.assertTrue(attachmentNames.contains("myFourthFile.txt"));
        Assert.assertTrue(attachmentNames.contains("myFifthFile.txt"));
        Assert.assertTrue(attachmentNames.contains("mySixthFile.txt"));
        dataHolder.setValue("singleRichTextData", retrievedRichText);
        
        retrievedRichText = (RichText) _richTextType.read(repositoryData, "singleRichTextData");
        _doAssertionsOnRichTextDataAfterPlayingWithAttachments(retrievedRichText);
        
        // RichText - modify some thing on rich text (but not its files)
        retrievedRichText.removeAnnotations("annotations2");
        dataHolder.setValue("singleRichTextData", retrievedRichText);
    }
    
    private void _assertFileDataHasNoStringProperty(RepositoryData repositoryData, String dataName)
    {
        try
        {
            repositoryData.getString(dataName);
            Assert.fail("There should not be any string data with name " + dataName);
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains(dataName));
        }
    }
    
    private void _assertFileDataHasNoBinaryNode(RepositoryData repositoryData, String dataName)
    {
        try
        {
            repositoryData.getRepositoryData(dataName);
            Assert.fail("There should not be any repository data with name " + dataName);
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains(dataName));
        }
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for resources data types
     * Tests that we can get the value from a data holder and then set it into another data holder
     * This test uses a {@link JCRRepositoryData}
     * @throws IOException if an error occurs while writing or reading resources
     */
    @Test
    @NeedFeatures({"explorer/factories", })
    public void testCopyRichTextFromADataHolderToAnotherWithJCRRepositoryData() throws IOException
    {
        ModifiableRepositoryData originRepositoryData = _jcrRoot.addRepositoryData("testCopyRichTextFromADataHolderToAnotherOrigin", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder originDataHolder = new DefaultModifiableModelAwareDataHolder(originRepositoryData, _model);
        ModifiableRepositoryData destRepositoryData = _jcrRoot.addRepositoryData("testCopyRichTextFromADataHolderToAnotherDest", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder destDataHolder = new DefaultModifiableModelAwareDataHolder(destRepositoryData, _model);
        
        // Set a rich text value on the origin data holder
        RichText richText = new RichText();
        try (OutputStream os = richText.getOutputStream())
        {
            os.write("my rich text content".getBytes());
        }
        richText.setEncoding("UTF-8");
        richText.setMimeType("text/xml");
        richText.setLastModificationDate(ZonedDateTime.of(2019, 03, 20, 13, 34, 12, 0, ZoneId.systemDefault()));
        // Files
        ResourceElementTypeHelperForTests.addFilesToRichText(richText);
        originDataHolder.setValue("singleRichTextData", richText);
        
        // Retrieve the rich text from a call to #getValue
        RichText retrievedRichText = originDataHolder.getValue("singleRichTextData");
        ResourceElementTypeHelperForTests.playWithRichTextFiles(retrievedRichText);
        
        // Set this rich text to the original and destination data holders
        originDataHolder.setValue("singleRichTextData", retrievedRichText);
        RichText richTextFromOriginal = (RichText) _richTextType.read(originRepositoryData, "singleRichTextData");
        _doAssertionsOnRichTextDataAfterPlayingWithAttachments(richTextFromOriginal);
        
        destDataHolder.setValue("singleRichTextData", retrievedRichText);
        RichText richTextFromDestination = (RichText) _richTextType.read(destRepositoryData, "singleRichTextData");
        _doAssertionsOnRichTextDataAfterPlayingWithAttachments(richTextFromDestination);
        
    }

    private void _doAssertionsOnRichTextDataAfterPlayingWithAttachments(RichText retrievedRichText) throws IOException
    {
        Assert.assertEquals(5,  retrievedRichText.getAttachments().size());
        
        Resource retrievedFile2 = retrievedRichText.getAttachment("mySecondFile.txt");
        _assertZonedDateTimeEquals(ZoneId.systemDefault(), 2020, Month.JUNE, 04, 16, 24, 12, retrievedFile2.getLastModificationDate());
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFile2.getInputStream(), writer, retrievedFile2.getEncoding());
            Assert.assertEquals("my second file content", writer.toString());
        }

        Resource retrievedFile3 = retrievedRichText.getAttachment("myThirdFile.txt");
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFile3.getInputStream(), writer, retrievedFile3.getEncoding());
            Assert.assertEquals("my third file modified content", writer.toString());
        }
        
        Resource retrievedFile4 = retrievedRichText.getAttachment("myFourthFile.txt");
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFile4.getInputStream(), writer, retrievedFile4.getEncoding());
            Assert.assertEquals("my fourth file content", writer.toString());
        }
        
        Resource retrievedFile5 = retrievedRichText.getAttachment("myFifthFile.txt");
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFile5.getInputStream(), writer, retrievedFile5.getEncoding());
            Assert.assertEquals("my fifth file content", writer.toString());
        }
        
        Resource retrievedFile6 = retrievedRichText.getAttachment("mySixthFile.txt");
        try (StringWriter writer = new StringWriter())
        {
            IOUtils.copy(retrievedFile6.getInputStream(), writer, retrievedFile6.getEncoding());
            Assert.assertEquals("my sixth file content", writer.toString());
        }
    }

    private void _assertZonedDateTimeEquals(ZoneId zoneId, int year, Month month, int day, int hour, int minute, int second, ZonedDateTime actualZDT)
    {
        ZonedDateTime zdt = actualZDT.withZoneSameInstant(zoneId);
        Assert.assertEquals(zoneId, zdt.getZone());
        Assert.assertEquals(year, zdt.getYear());
        Assert.assertEquals(month, zdt.getMonth());
        Assert.assertEquals(day, zdt.getDayOfMonth());
        Assert.assertEquals(hour, zdt.getHour());
        Assert.assertEquals(minute, zdt.getMinute());
        Assert.assertEquals(second, zdt.getSecond());
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for resources data types
     * This test uses a {@link MemoryRepositoryData}
     * {@link MemoryRepositoryData} does not support resources data types
     * @throws IOException if an error occurs while writing or reading binaries
     */
    @Test
    public void testSetValueForResourceDataTypesWithMemoryRepositoryData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueForResourceDataTypes", DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Write
        Binary binary = ResourceElementTypeHelperForTests.createBinary("my binary content");
        try
        {
            dataHolder.setValue("singleBinaryData", binary);
            Assert.fail("MemoryRepositoryData can't support streams");
        }
        catch (UnsupportedOperationException e)
        {
            Assert.assertTrue(e.getMessage().contains("not supported"));
        }
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for content types
     * For multiple data, check that duplicated items are stored only once
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    @NeedFeatures({
        "cms/factories",
        "test/content-type",
    })
    public void testSetValueForContentDataTypeWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueForContentDataType", DATA_NODE_TYPE);
        _testSetValueForContentDataType(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for content types
     * For multiple data, check that duplicated items are stored only once
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    @NeedFeatures({
        "cms/factories",
        "test/content-type",
    })
    public void testSetValueForContentDataTypeWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueForContentDataType", DATA_NODE_TYPE);
        _testSetValueForContentDataType(repositoryData);
    }
    
    private void _testSetValueForContentDataType(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        ModifiableContent content1 = _createContent("content1");
        ModifiableContent content2 = _createContent("content2");
        ModifiableContent content3 = _createContent("content3");
        ModifiableContent content4 = _createContent("content4");
        
        // Content as string - write
        dataHolder.setValue("singleContentData", content1.getId());
        dataHolder.setValue("multipleContentData", new String[] {content1.getId(), content2.getId(), content1.getId()});
        
        // Read
        Assert.assertFalse(repositoryData.isMultiple("singleContentData"));
        Assert.assertEquals(content1.getId(), repositoryData.getString("singleContentData"));
        Assert.assertTrue(repositoryData.isMultiple("multipleContentData"));
        String[] multipleContent = repositoryData.getStrings("multipleContentData");
        Assert.assertEquals(2, multipleContent.length);
        Assert.assertEquals(content1.getId(), multipleContent[0]);
        Assert.assertEquals(content2.getId(), multipleContent[1]);
        
        // Content - write
        dataHolder.setValue("singleContentData", content2);
        dataHolder.setValue("multipleContentData", new Content[] {content2, content3, content4, content3, content4});
        
        // Read
        Assert.assertFalse(repositoryData.isMultiple("singleContentData"));
        Assert.assertEquals(content2.getId(), repositoryData.getString("singleContentData"));
        Assert.assertTrue(repositoryData.isMultiple("multipleContentData"));
        multipleContent = repositoryData.getStrings("multipleContentData");
        Assert.assertEquals(3, multipleContent.length);
        Assert.assertEquals(content2.getId(), multipleContent[0]);
        Assert.assertEquals(content3.getId(), multipleContent[1]);
        Assert.assertEquals(content4.getId(), multipleContent[2]);
        
        // Content as wrapper - write
        dataHolder.setValue("singleContentData", new ContentValue(_ametysObjectResolver, content4.getId()));
        dataHolder.setValue("multipleContentData", new ContentValue[] {new ContentValue(_ametysObjectResolver, content1.getId()), new ContentValue(_ametysObjectResolver, content3.getId()), new ContentValue(_ametysObjectResolver, content3.getId())});
        
        // Read
        Assert.assertFalse(repositoryData.isMultiple("singleContentData"));
        Assert.assertEquals(content4.getId(), repositoryData.getString("singleContentData"));
        Assert.assertTrue(repositoryData.isMultiple("multipleContentData"));
        multipleContent = repositoryData.getStrings("multipleContentData");
        Assert.assertEquals(2, multipleContent.length);
        Assert.assertEquals(content1.getId(), multipleContent[0]);
        Assert.assertEquals(content3.getId(), multipleContent[1]);
        
        // Set values on the linked content
        dataHolder.setValue("singleContentData", content1);
        content1.setValue("field1", "some value");
        ContentValue singleContent = dataHolder.getValue("singleContentData");
        Assert.assertEquals("some value", singleContent.getContent().getValue("field1"));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for composites and repeaters
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetValueForCompositeAndRepeatersWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueForCompositeAndRepeaters", DATA_NODE_TYPE);
        _testSetValueForCompositeAndRepeaters(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} method for composites and repeaters
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueForCompositeAndRepeatersWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueForCompositeAndRepeaters", DATA_NODE_TYPE);
        _testSetValueForCompositeAndRepeaters(repositoryData);
    }
    
    private void _testSetValueForCompositeAndRepeaters(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);

        try
        {
            // Tries to set a value when the parents don't exist
            dataHolder.setValue("composite/repeater[1]/subcomposite/singleStringData", "my string");
            Assert.fail("parent composite and repeater entry don't exist, it should not be possible to create the subcomposite");
        }
        catch (BadItemTypeException e)
        {
            Assert.assertTrue(e.getMessage().contains("exist"));
        }
        
        // Tests the creation of values with complete path
        dataHolder.getComposite("composite", true);
        ModifiableRepeater repeater = dataHolder.getRepeater("composite/repeater", true);
        repeater.addEntry();
        dataHolder.getComposite("composite/repeater[1]/subcomposite", true);
        dataHolder.setValue("composite/repeater[1]/subcomposite/singleStringData", "my string");
        
        // Check value in the repository
        RepositoryData subComposite = repositoryData.getRepositoryData("composite").getRepositoryData("repeater").getRepositoryData("1").getRepositoryData("subcomposite");
        Assert.assertEquals("my string", subComposite.getString("singleStringData"));
        
        try
        {
            // Tries to set a value to a data not defined in the model
            dataHolder.setValue("composite/repeater[1]/subcomposite/undefinedData", "my string");
            Assert.fail("There is no attribute named 'undefinedData' in the model. The setValue method should have fail.");
        }
        catch (UndefinedItemPathException e)
        {
            Assert.assertTrue(e.getMessage().contains("defined"));
        }
        
        try
        {
            // Tries to set a value to a data not defined in the model (here, it is the parent that is not defined)
            dataHolder.setValue("composite/repeater[1]/undefinedSubcomposite/SingleStringData", "my string");
            Assert.fail("There is no composite named 'undefinedSubcomposite' in the model. The setValue method should have fail.");
        }
        catch (UndefinedItemPathException e)
        {
            Assert.assertTrue(e.getMessage().contains("defined"));
        }
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} behavior when setting multiple values several times
     * Checks that the values are well removed before setting the new ones 
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetMultipleComplexValuesSeveralTimesWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetMultipleComplexValuesSeveralTimes", DATA_NODE_TYPE);
        _testSetMultipleComplexValuesSeveralTimes(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} behavior when setting multiple values several times
     * Checks that the values are well removed before setting the new ones 
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetMultipleComplexValuesSeveralTimesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetMultipleComplexValuesSeveralTimes", DATA_NODE_TYPE);
        _testSetMultipleComplexValuesSeveralTimes(repositoryData);
    }
    
    private void _testSetMultipleComplexValuesSeveralTimes(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);

        // Set some values in multiple user data
        UserIdentity[] multipleUserData = new UserIdentity[]{new UserIdentity("peter", "anyPopulation"), new UserIdentity("paul", "anyPopulation"), new UserIdentity("jack", "anyPopulation")};
        dataHolder.setValue("multipleUserData", multipleUserData);
        
        // Check the stored users
        multipleUserData = dataHolder.getValue("multipleUserData");
        Assert.assertEquals(3, multipleUserData.length);
        Assert.assertEquals("peter", multipleUserData[0].getLogin());
        Assert.assertEquals("paul", multipleUserData[1].getLogin());
        Assert.assertEquals("jack", multipleUserData[2].getLogin());
        
        // Set some other values in the data
        multipleUserData = new UserIdentity[]{new UserIdentity("thelma", "anyPopulation"), new UserIdentity("louise", "anyPopulation")};
        dataHolder.setValue("multipleUserData", multipleUserData);
        
        // Check the stored users
        multipleUserData = (UserIdentity[]) _userType.read(repositoryData, "multipleUserData");
        Assert.assertEquals(2, multipleUserData.length);
        Assert.assertEquals("thelma", multipleUserData[0].getLogin());
        Assert.assertEquals("louise", multipleUserData[1].getLogin());
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} behavior when setting <code>null</code> values
     * Setting <code>null</code> values leads to empty a data, not removing it
     * This test uses a {@link JCRRepositoryData}
     * @throws IOException if an error occurs while writing or reading resources
     */
    @Test
    @NeedFeatures({
        "cms/factories",
        "test/content-type",
    })
    public void testEmptyValuesWithJCRRepositoryData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testEmptyValues", DATA_NODE_TYPE);
        _testEmptyValues(repositoryData);
        _testEmptyResourceValues(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} behavior when setting <code>null</code> values
     * Setting <code>null</code> values leads to empty a data, not removing it
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    @NeedFeatures({
        "cms/factories",
        "test/content-type",
    })
    public void testEmptyValuesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testEmptyValues", DATA_NODE_TYPE);
        _testEmptyValues(repositoryData);
    }
    
    private void _testEmptyValues(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);

        // single string data
        Assert.assertFalse(repositoryData.hasValue("singleStringData"));
        dataHolder.setValue("singleStringData", null);
        Assert.assertEquals(StringUtils.EMPTY, repositoryData.getString("singleStringData"));
        dataHolder.setValue("singleStringData", StringUtils.EMPTY);
        Assert.assertEquals(StringUtils.EMPTY, repositoryData.getString("singleStringData"));
        
        // multiple string data - using null
        Assert.assertFalse(repositoryData.hasValue("multipleStringData"));
        dataHolder.setValue("multipleStringData", null);
        Assert.assertEquals(0, repositoryData.getStrings("multipleStringData").length);
        dataHolder.setValue("multipleStringData", new String[] {"a string value", "another string value"});
        Assert.assertEquals(2, repositoryData.getStrings("multipleStringData").length);
        dataHolder.setValue("multipleStringData", null);
        Assert.assertEquals(0, repositoryData.getStrings("multipleStringData").length);
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleStringData", new String[] {"value 1", null, "value 3"});
        
        // multiple string data - using empty array
        dataHolder.setValue("multipleStringData", new String[] {"a string value", "another string value"});
        Assert.assertEquals(2, repositoryData.getStrings("multipleStringData").length);
        dataHolder.setValue("multipleStringData", new String[0]);
        Assert.assertEquals(0, repositoryData.getStrings("multipleStringData").length);

        // single long data
        Assert.assertFalse(repositoryData.hasValue("singleLongData"));
        dataHolder.setValue("singleLongData", null);
        Assert.assertFalse(repositoryData.hasValue("singleLongData"));
        Assert.assertTrue(repositoryData.getBoolean("singleLongData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleLongData", 18L);
        Assert.assertEquals(Long.valueOf(18), repositoryData.getLong("singleLongData"));
        Assert.assertFalse(repositoryData.hasValue("singleLongData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleLongData", null);
        Assert.assertFalse(repositoryData.hasValue("singleLongData"));
        Assert.assertTrue(repositoryData.getBoolean("singleLongData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        
        // multiple long data - using null
        Assert.assertFalse(repositoryData.hasValue("multipleLongData"));
        dataHolder.setValue("multipleLongData", null);
        Assert.assertEquals(0, repositoryData.getLongs("multipleLongData").length);
        dataHolder.setValue("multipleLongData", new Long[] {21L, 78L, 31L});
        Assert.assertEquals(3, repositoryData.getLongs("multipleLongData").length);
        dataHolder.setValue("multipleLongData", null);
        Assert.assertEquals(0, repositoryData.getLongs("multipleLongData").length);
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleLongData", new Long[] {21L, null, 31L});
        
        // multiple long data - using empty array
        dataHolder.setValue("multipleLongData", new Long[] {21L, 78L, 31L});
        Assert.assertEquals(3, repositoryData.getLongs("multipleLongData").length);
        dataHolder.setValue("multipleLongData", new Long[0]);
        Assert.assertEquals(0, repositoryData.getLongs("multipleLongData").length);

        // single double data
        Assert.assertFalse(repositoryData.hasValue("singleDoubleData"));
        dataHolder.setValue("singleDoubleData", null);
        Assert.assertFalse(repositoryData.hasValue("singleDoubleData"));
        Assert.assertTrue(repositoryData.getBoolean("singleDoubleData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleDoubleData", 185D);
        Assert.assertEquals(Double.valueOf(185), repositoryData.getDouble("singleDoubleData"));
        Assert.assertFalse(repositoryData.hasValue("singleDoubleData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleDoubleData", null);
        Assert.assertFalse(repositoryData.hasValue("singleDoubleData"));
        Assert.assertTrue(repositoryData.getBoolean("singleDoubleData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        
        // multiple double data - using null
        Assert.assertFalse(repositoryData.hasValue("multipleDoubleData"));
        dataHolder.setValue("multipleDoubleData", null);
        Assert.assertEquals(0, repositoryData.getDoubles("multipleDoubleData").length);
        dataHolder.setValue("multipleDoubleData", new Double[] {21D, 78D});
        Assert.assertEquals(2, repositoryData.getDoubles("multipleDoubleData").length);
        dataHolder.setValue("multipleDoubleData", null);
        Assert.assertEquals(0, repositoryData.getDoubles("multipleDoubleData").length);
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleDoubleData", new Double[] {null, 78D});
        
        // multiple double data - using empty array
        dataHolder.setValue("multipleDoubleData", new Double[] {21D, 78D});
        Assert.assertEquals(2, repositoryData.getDoubles("multipleDoubleData").length);
        dataHolder.setValue("multipleDoubleData", new Double[0]);
        Assert.assertEquals(0, repositoryData.getDoubles("multipleDoubleData").length);
        
        // single boolean data
        Assert.assertFalse(repositoryData.hasValue("singleBooleanData"));
        dataHolder.setValue("singleBooleanData", null);
        Assert.assertFalse(repositoryData.hasValue("singleBooleanData"));
        Assert.assertTrue(repositoryData.getBoolean("singleBooleanData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleBooleanData", true);
        Assert.assertTrue(repositoryData.getBoolean("singleBooleanData"));
        Assert.assertFalse(repositoryData.hasValue("singleBooleanData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleBooleanData", null);
        Assert.assertFalse(repositoryData.hasValue("singleBooleanData"));
        Assert.assertTrue(repositoryData.getBoolean("singleBooleanData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        
        // single date data
        Assert.assertFalse(repositoryData.hasValue("singleDateData"));
        dataHolder.setValue("singleDateData", null);
        Assert.assertFalse(repositoryData.hasValue("singleDateData"));
        Assert.assertTrue(repositoryData.getBoolean("singleDateData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleDateData", LocalDate.of(2019, Month.JULY, 24));
        Assert.assertEquals(2019, repositoryData.getDate("singleDateData").get(Calendar.YEAR));
        Assert.assertFalse(repositoryData.hasValue("singleDateData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleDateData", null);
        Assert.assertFalse(repositoryData.hasValue("singleDateData"));
        Assert.assertTrue(repositoryData.getBoolean("singleDateData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        
        // multiple date data - using null
        Assert.assertFalse(repositoryData.hasValue("multipleDateData"));
        dataHolder.setValue("multipleDateData", null);
        Assert.assertEquals(0, repositoryData.getDates("multipleDateData").length);
        dataHolder.setValue("multipleDateData", new LocalDate[] {LocalDate.now()});
        Assert.assertEquals(1, repositoryData.getDates("multipleDateData").length);
        dataHolder.setValue("multipleDateData", null);
        Assert.assertEquals(0, repositoryData.getDates("multipleDateData").length);
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleDateData", new LocalDate[] {LocalDate.now(), null});
        
        // multiple date data - using empty array
        dataHolder.setValue("multipleDateData", new LocalDate[] {LocalDate.now()});
        Assert.assertEquals(1, repositoryData.getDates("multipleDateData").length);
        dataHolder.setValue("multipleDateData", new LocalDate[0]);
        Assert.assertEquals(0, repositoryData.getDates("multipleDateData").length);
        
        // single date time data
        Assert.assertFalse(repositoryData.hasValue("singleDatetimeData"));
        dataHolder.setValue("singleDatetimeData", null);
        Assert.assertFalse(repositoryData.hasValue("singleDatetimeData"));
        Assert.assertTrue(repositoryData.getBoolean("singleDatetimeData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleDatetimeData", ZonedDateTime.of(2019, 5, 24, 12, 10, 23, 45, ZoneId.systemDefault()));
        Assert.assertEquals(2019, repositoryData.getDate("singleDatetimeData").get(Calendar.YEAR));
        Assert.assertFalse(repositoryData.hasValue("singleDatetimeData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleDatetimeData", null);
        Assert.assertFalse(repositoryData.hasValue("singleDatetimeData"));
        Assert.assertTrue(repositoryData.getBoolean("singleDatetimeData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        
        // multiple date time data - using null
        Assert.assertFalse(repositoryData.hasValue("multipleDatetimeData"));
        dataHolder.setValue("multipleDatetimeData", null);
        Assert.assertEquals(0, repositoryData.getDates("multipleDatetimeData").length);
        dataHolder.setValue("multipleDatetimeData", new ZonedDateTime[] {ZonedDateTime.now()});
        Assert.assertEquals(1, repositoryData.getDates("multipleDatetimeData").length);
        dataHolder.setValue("multipleDatetimeData", null);
        Assert.assertEquals(0, repositoryData.getDates("multipleDatetimeData").length);
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleDatetimeData", new ZonedDateTime[] {ZonedDateTime.now(), null});
        
        // multiple date time data - using empty array
        dataHolder.setValue("multipleDatetimeData", new ZonedDateTime[] {ZonedDateTime.now()});
        Assert.assertEquals(1, repositoryData.getDates("multipleDatetimeData").length);
        dataHolder.setValue("multipleDatetimeData", new ZonedDateTime[0]);
        Assert.assertEquals(0, repositoryData.getDates("multipleDatetimeData").length);
        
        // single user data
        Assert.assertFalse(repositoryData.hasValue("singleUserData"));
        dataHolder.setValue("singleUserData", null);
        _assertSingleUserDataIsPresentButEmpty(repositoryData, "singleUserData");
        dataHolder.setValue("singleUserData", new UserIdentity("fmercury", "queen"));
        Assert.assertTrue(repositoryData.hasValue("singleUserData"));
        Assert.assertTrue(repositoryData.getRepositoryData("singleUserData").hasValue("login"));
        Assert.assertTrue(repositoryData.getRepositoryData("singleUserData").hasValue("population"));
        dataHolder.setValue("singleUserData", null);
        _assertSingleUserDataIsPresentButEmpty(repositoryData, "singleUserData");
        
        // multiple user data - using null
        Assert.assertFalse(repositoryData.hasValue("multipleUserData"));
        dataHolder.setValue("multipleUserData", null);
        _assertMultipleComplexDataIsPresentButEmpty(repositoryData, "multipleUserData");
        dataHolder.setValue("multipleUserData", new UserIdentity[] {new UserIdentity("mknopfler", "direstraits")});
        Assert.assertTrue(repositoryData.hasValue("multipleUserData"));
        Assert.assertTrue(repositoryData.getRepositoryData("multipleUserData").hasValue("1"));
        Assert.assertTrue(repositoryData.getRepositoryData("multipleUserData").getRepositoryData("1").hasValue("login"));
        Assert.assertTrue(repositoryData.getRepositoryData("multipleUserData").getRepositoryData("1").hasValue("population"));
        Assert.assertFalse(repositoryData.getRepositoryData("multipleUserData").hasValue("2"));
        dataHolder.setValue("multipleUserData", null);
        _assertMultipleComplexDataIsPresentButEmpty(repositoryData, "multipleUserData");
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleUserData", new UserIdentity[] {null, new UserIdentity("mknopfler", "direstraits")});
        
        // multiple user data - using empty array
        dataHolder.setValue("multipleUserData", new UserIdentity[] {new UserIdentity("mknopfler", "direstraits")});
        Assert.assertTrue(repositoryData.hasValue("multipleUserData"));
        dataHolder.setValue("multipleUserData", new UserIdentity[0]);
        _assertMultipleComplexDataIsPresentButEmpty(repositoryData, "multipleUserData");
        
        // single content data
        Content content = _createContent("content");
        Assert.assertFalse(repositoryData.hasValue("singleContentData"));
        dataHolder.setValue("singleContentData", null);
        Assert.assertEquals(StringUtils.EMPTY, repositoryData.getString("singleContentData"));
        dataHolder.setValue("singleContentData", content);
        Assert.assertEquals(content.getId(), repositoryData.getString("singleContentData"));
        dataHolder.setValue("singleContentData", null);
        Assert.assertEquals(StringUtils.EMPTY, repositoryData.getString("singleContentData"));
        
        // multiple content data - using null
        Assert.assertFalse(repositoryData.hasValue("multipleContentData"));
        dataHolder.setValue("multipleContentData", null);
        Assert.assertEquals(0, repositoryData.getStrings("multipleContentData").length);
        dataHolder.setValue("multipleContentData", new Content[] {content});
        Assert.assertEquals(1, repositoryData.getStrings("multipleContentData").length);
        dataHolder.setValue("multipleContentData", null);
        Assert.assertEquals(0, repositoryData.getStrings("multipleContentData").length);
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleContentData", new Content[] {null, content});
        
        // multiple content data - using empty array
        dataHolder.setValue("multipleContentData", new Content[] {content});
        Assert.assertEquals(1, repositoryData.getStrings("multipleContentData").length);
        dataHolder.setValue("multipleContentData", new Content[0]);
        Assert.assertEquals(0, repositoryData.getStrings("multipleContentData").length);
    }
        
    private void _testEmptyResourceValues(ModifiableRepositoryData repositoryData) throws IOException
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        Binary binary = ResourceElementTypeHelperForTests.createBinary("my binary content");
        
        // single binary data
        Assert.assertFalse(repositoryData.hasValue("singleBinaryData"));
        dataHolder.setValue("singleBinaryData", null);
        _assertSingleBinaryDataIsPresentButEmpty(repositoryData, "singleBinaryData");
        dataHolder.setValue("singleBinaryData", binary);
        Assert.assertTrue(repositoryData.hasValue("singleBinaryData"));
        _assertEqualsStreamContent(repositoryData.getRepositoryData("singleBinaryData"), "my binary content");
        dataHolder.setValue("singleBinaryData", null);
        _assertSingleBinaryDataIsPresentButEmpty(repositoryData, "singleBinaryData");
        
        // multiple binary data - using null
        Assert.assertFalse(repositoryData.hasValue("multipleBinaryData"));
        dataHolder.setValue("multipleBinaryData", null);
        _assertMultipleComplexDataIsPresentButEmpty(repositoryData, "multipleBinaryData");
        dataHolder.setValue("multipleBinaryData", new Binary[] {binary});
        Assert.assertTrue(repositoryData.hasValue("multipleBinaryData"));
        Assert.assertTrue(repositoryData.getRepositoryData("multipleBinaryData").hasValue("1"));
        _assertEqualsStreamContent(repositoryData.getRepositoryData("multipleBinaryData").getRepositoryData("1"), "my binary content");
        Assert.assertFalse(repositoryData.getRepositoryData("multipleBinaryData").hasValue("2"));
        dataHolder.setValue("multipleBinaryData", null);
        _assertMultipleComplexDataIsPresentButEmpty(repositoryData, "multipleBinaryData");
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleBinaryData", new Binary[] {null, binary});
        
        // multiple binary data - using empty array
        dataHolder.setValue("multipleBinaryData", new Binary[] {binary});
        Assert.assertTrue(repositoryData.hasValue("multipleBinaryData"));
        dataHolder.setValue("multipleBinaryData", new Binary[0]);
        _assertMultipleComplexDataIsPresentButEmpty(repositoryData, "multipleBinaryData");
        
        // single file data
        Assert.assertFalse(repositoryData.hasValue("singleFileData"));
        dataHolder.setValue("singleFileData", null);
        Assert.assertEquals(StringUtils.EMPTY, repositoryData.getString("singleFileData"));
        dataHolder.setValue("singleFileData", binary);
        Assert.assertTrue(repositoryData.hasValue("singleFileData"));
        _assertEqualsStreamContent(repositoryData.getRepositoryData("singleFileData"), "my binary content");
        dataHolder.setValue("singleFileData", null);
        Assert.assertEquals(StringUtils.EMPTY, repositoryData.getString("singleFileData"));
        
        // multiple file data - using null
        Assert.assertFalse(repositoryData.hasValue("multipleFileData"));
        dataHolder.setValue("multipleFileData", null);
        _assertMultipleFileDataIsPresentButEmpty(repositoryData, "multipleFileData");
        dataHolder.setValue("multipleFileData", new Binary[] {binary});
        Assert.assertTrue(repositoryData.hasValue("multipleFileData"));
        Assert.assertTrue(repositoryData.getRepositoryData("multipleFileData").hasValue("1"));
        _assertEqualsStreamContent(repositoryData.getRepositoryData("multipleFileData").getRepositoryData("1"), "my binary content");
        Assert.assertFalse(repositoryData.getRepositoryData("multipleFileData").hasValue("2"));
        dataHolder.setValue("multipleFileData", null);
        _assertMultipleFileDataIsPresentButEmpty(repositoryData, "multipleFileData");
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleFileData", new Binary[] {null, binary});
        
        // multiple file data - using empty String array
        dataHolder.setValue("multipleFileData", new Binary[] {binary});
        Assert.assertTrue(repositoryData.hasValue("multipleFileData"));
        dataHolder.setValue("multipleFileData", new String[0]);
        _assertMultipleFileDataIsPresentButEmpty(repositoryData, "multipleFileData");
        
        // multiple file data - using empty Explorer file array
        dataHolder.setValue("multipleFileData", new Binary[] {binary});
        Assert.assertTrue(repositoryData.hasValue("multipleFileData"));
        dataHolder.setValue("multipleFileData", new ExplorerFile[0]);
        _assertMultipleFileDataIsPresentButEmpty(repositoryData, "multipleFileData");
        
        // multiple file data - using empty Binary array
        dataHolder.setValue("multipleFileData", new Binary[] {binary});
        Assert.assertTrue(repositoryData.hasValue("multipleFileData"));
        dataHolder.setValue("multipleFileData", new Binary[0]);
        _assertMultipleFileDataIsPresentButEmpty(repositoryData, "multipleFileData");
    }
    
    private <T> void _assertMultipleDataDoesNotSupportNullValue(ModifiableModelAwareDataHolder dataHolder, String dataPath, T[] values)
    {
        try
        {
            dataHolder.setValue(dataPath, values);
            Assert.fail("Setting of null values in a multiple data should not be supported");
        }
        catch (IllegalArgumentException e)
        {
            Assert.assertTrue(e.getMessage().contains(dataPath));
        }
    }
    
    private void _assertMultipleComplexDataIsPresentButEmpty(RepositoryData repositoryData, String dataName)
    {
        Assert.assertTrue(repositoryData.hasValue(dataName));
        RepositoryData emptyComplexData = repositoryData.getRepositoryData(dataName);
        Assert.assertTrue(emptyComplexData.getDataNames().isEmpty());
    }
    
    private void _assertSingleUserDataIsPresentButEmpty(RepositoryData repositoryData, String dataName)
    {
        Assert.assertTrue(repositoryData.hasValue(dataName));
        RepositoryData data = repositoryData.getRepositoryData(dataName);
        Assert.assertTrue(StringUtils.isEmpty(data.getString("login")));
        Assert.assertTrue(StringUtils.isEmpty(data.getString("population")));
    }
    
    private void _assertSingleBinaryDataIsPresentButEmpty(RepositoryData repositoryData, String dataName)
    {
        Assert.assertTrue(repositoryData.hasValue(dataName));
        RepositoryData data = repositoryData.getRepositoryData(dataName);
        Assert.assertTrue(data.getBoolean(ResourceElementTypeHelper.EMPTY_RESOURCE_IDENTIFIER, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
    }
    
    private void _assertMultipleFileDataIsPresentButEmpty(RepositoryData repositoryData, String dataName)
    {
        Assert.assertTrue(repositoryData.hasValue(dataName));
        
        // Assert there is an empty string array
        Assert.assertEquals(0, repositoryData.getStrings("multipleFileData").length);
        
        try
        {
            // Assert there is no repository data with files stored as binaries
            repositoryData.getRepositoryData(dataName);
            Assert.fail("There should not be files stored as binaries");
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains(dataName));
        }
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder} behavior when setting multiple values with only one entry
     * Checks that the values are well removed before setting the new ones 
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetValueForMultipleValuesWithOnlyOneEntryWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueForMultipleValuesWithOnlyOneEntry", DATA_NODE_TYPE);
        _testSetValueForMultipleValuesWithOnlyOneEntry(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder} behavior when setting multiple values with only one entry
     * Checks that the values are well removed before setting the new ones 
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueForMultipleValuesWithOnlyOneEntryWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueForMultipleValuesWithOnlyOneEntry", DATA_NODE_TYPE);
        _testSetValueForMultipleValuesWithOnlyOneEntry(repositoryData);
    }
    
    private void _testSetValueForMultipleValuesWithOnlyOneEntry(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Test with a simple string attribute
        dataHolder.setValue("multipleStringData", new String[]{"stringValue"});
        String[] stringmult = dataHolder.getValue("multipleStringData");
        Assert.assertArrayEquals(new String[]{"stringValue"}, stringmult);
        
        // Test with a user attribute
        dataHolder.setValue("multipleUserData", new UserIdentity[]{new UserIdentity("peter", "anyPopulation")});
        UserIdentity[] usermult = dataHolder.getValue("multipleUserData");
        Assert.assertEquals(1, usermult.length);
        Assert.assertEquals("peter", usermult[0].getLogin());
        Assert.assertEquals("anyPopulation", usermult[0].getPopulationId());
        
        // Test with a content attribute
        ModifiableContent content = _createContent("content");
        dataHolder.setValue("multipleContentData", new Content[]{content});
        ContentValue[] contentmult = dataHolder.getValue("multipleContentData");
        Assert.assertEquals(1, contentmult.length);
        Assert.assertEquals(content.getId(), contentmult[0].getContentId());
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder} behavior when setting multiple values with only one entry
     * Checks that the values are well removed before setting the new ones 
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetValueForMultipleDataWithSingleValuesWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueForMultipleDataWithSingleValues", DATA_NODE_TYPE);
        _testSetValueForMultipleDataWithSingleValues(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder} behavior when setting multiple values with only one entry
     * Checks that the values are well removed before setting the new ones 
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueForMultipleDataWithSingleValuesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueForMultipleDataWithSingleValues", DATA_NODE_TYPE);
        _testSetValueForMultipleDataWithSingleValues(repositoryData);
    }
    
    private void _testSetValueForMultipleDataWithSingleValues(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        // Test with a simple string attribute
        dataHolder.setValue("multipleStringData", "stringValue");
        Assert.assertTrue(repositoryData.isMultiple("multipleStringData"));
        
        // Test with a user attribute
        dataHolder.setValue("multipleUserData", new UserIdentity("peter", "anyPopulation"));
        Assert.assertTrue(repositoryData.isMultiple("multipleUserData"));
        
        // Test with a content attribute
        ModifiableContent content = _createContent("content");
        dataHolder.setValue("multipleContentData", content);
        Assert.assertTrue(repositoryData.isMultiple("multipleContentData"));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#removeValue(String)} implementation
     * Checks that the values are well removed before setting the new ones 
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testRemoveValueWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testRemoveValue", DATA_NODE_TYPE);
        _testRemoveValue(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#removeValue(String)} implementation
     * Checks that the values are well removed 
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testRemoveValueWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testRemoveValue", DATA_NODE_TYPE);
        _testRemoveValue(repositoryData);
    }
    
    private void _testRemoveValue(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        _addSomeValuesInRepositoryData(repositoryData);
        RepositoryData compositeData = repositoryData.getRepositoryData("composite");
        RepositoryData repeaterData = compositeData.getRepositoryData("repeater");
        RepositoryData entry1Data = repeaterData.getRepositoryData("1");
        RepositoryData entry2Data = repeaterData.getRepositoryData("2");
        RepositoryData subcomposite2Data = entry2Data.getRepositoryData("subcomposite");
        
        // Tests to remove a single value
        Assert.assertTrue(repositoryData.hasValue("singleStringData"));
        dataHolder.removeValue("singleStringData");
        Assert.assertFalse("The value at path 'singleStringData' should have been removed", repositoryData.hasValue("singleStringData"));
        
        // Tests to remove a value with a complete data path
        Assert.assertTrue(subcomposite2Data.hasValue("singleStringData"));
        dataHolder.removeValue("composite/repeater[2]/subcomposite/singleStringData");
        Assert.assertFalse("The value at path 'composite/repeater[2]/subcomposite/singleStringData' should have been removed", subcomposite2Data.hasValue("singleStringData"));
        
        // Tests to remove a composite still containing values
        Assert.assertTrue(entry1Data.hasValue("subcomposite"));
        dataHolder.removeValue("composite/repeater[1]/subcomposite");
        Assert.assertFalse("The composite at path 'composite/repeater[1]/subcomposite' should have been removed", entry1Data.hasValue("subcomposite"));
        
        // Tests to remove a value from a repeater entry
        Assert.assertTrue(entry1Data.hasValue("singleDoubleData"));
        ModifiableRepeaterEntry entry = dataHolder.getValue("composite/repeater[1]");
        entry.removeValue("singleDoubleData");
        Assert.assertFalse("The value at path 'composite/repeater[1]/singleDoubleData' should have been removed", entry1Data.hasValue("singleDoubleData"));
        
        // Tests to remove a value from a composite
        Assert.assertTrue(compositeData.hasValue("singleLongData"));
        ModifiableComposite composite = dataHolder.getValue("composite");
        composite.removeValue("singleLongData");
        Assert.assertFalse("The value at path 'composite/singleLongData' should have been removed", compositeData.hasValue("singleLongData"));
        
        // Tries to remove a value with non existing parent
        Assert.assertFalse(repeaterData.hasValue("5"));
        dataHolder.removeValue("composite/repeater[5]/singleBooleanData");
        Assert.assertFalse(repeaterData.hasValue("5"));
        
        try
        {
            // Tries to remove a value from an undefined attribute
            dataHolder.removeValue("composite/repeater[2]/subcomposite/undefined-attribute");
            Assert.fail("There is no attribute named 'undefined-attribute' in the model. The removeValue method should have fail.");
        }
        catch (UndefinedItemPathException e)
        {
            Assert.assertTrue(e.getMessage().contains("defined"));
        }
        
        try
        {
            // Tries to get a value from an undefined attribute
            dataHolder.removeValue("composite/repeater[2]/undefined-subcomposite/singleStringData");
            Assert.fail("There is no composite named 'undefined-subcomposite' in the model. The removeValue method should have fail.");
        }
        catch (UndefinedItemPathException e)
        {
            Assert.assertTrue(e.getMessage().contains("defined"));
        }
        
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#removeValue(String)} implementation
     * Checks that the composite is well removed 
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testRemoveCompositeWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testRemoveComposite", DATA_NODE_TYPE);
        _testRemoveComposite(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#removeValue(String)} implementation
     * Checks that the composite is well removed
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testRemoveCompositeWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testRemoveComposite", DATA_NODE_TYPE);
        _testRemoveComposite(repositoryData);
    }
    
    private void _testRemoveComposite(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        _addSomeValuesInRepositoryData(repositoryData);
        
        Assert.assertTrue(repositoryData.hasValue("composite"));
        dataHolder.removeValue("composite");
        Assert.assertFalse("The composite should have been removed", repositoryData.hasValue("composite"));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#removeValue(String)} implementation
     * Checks that the repeater is well removed 
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testRemoveRepeaterWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testRemoveRepeater", DATA_NODE_TYPE);
        _testRemoveRepeater(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#removeValue(String)} implementation
     * Checks that the repeater is well removed
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testRemoveRepeaterWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testRemoveRepeater", DATA_NODE_TYPE);
        _testRemoveRepeater(repositoryData);
    }
    
    private void _testRemoveRepeater(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        _addSomeValuesInRepositoryData(repositoryData);
        RepositoryData compositeData = repositoryData.getRepositoryData("composite");
        
        Assert.assertTrue(compositeData.hasValue("repeater"));
        dataHolder.removeValue("composite/repeater");
        Assert.assertFalse("The repeater should have been removed", compositeData.hasValue("repeater"));
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#removeValue(String)} implementation
     * Checks that the repeater entry is well removed 
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testRemoveRepeaterEntryWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testRemoveRepeaterEntry", DATA_NODE_TYPE);
        _testRemoveRepeaterEntry(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#removeValue(String)} implementation
     * Checks that the repeater entry is well removed
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testRemoveRepeaterEntryWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testRemoveRepeaterEntry", DATA_NODE_TYPE);
        _testRemoveRepeaterEntry(repositoryData);
    }
    
    private void _testRemoveRepeaterEntry(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        
        _addSomeValuesInRepositoryData(repositoryData);
        RepositoryData compositeData = repositoryData.getRepositoryData("composite");
        RepositoryData repeaterData = compositeData.getRepositoryData("repeater");
        
        Assert.assertTrue(repeaterData.hasValue("2"));
        dataHolder.removeValue("composite/repeater[2]");
        Assert.assertFalse("The repeater entry should have been removed", repeaterData.hasValue("2"));
    }
    
    private void _addSomeValuesInRepositoryData(ModifiableRepositoryData repositoryData)
    {
        repositoryData.setValue("singleStringData", "my string");
        repositoryData.setValues("multipleStringData", new String[]{"string1", "string2", "string3"});
        
        ModifiableRepositoryData composite = repositoryData.addRepositoryData("composite", RepositoryConstants.COMPOSITE_NODETYPE);
        composite.setValue("singleLongData", 156L);
        
        ModifiableRepositoryData repeater = composite.addRepositoryData("repeater", RepositoryConstants.REPEATER_NODETYPE);
        ModifiableRepositoryData entry1 = repeater.addRepositoryData("1", RepositoryConstants.COMPOSITE_NODETYPE);
        entry1.setValue("singleDoubleData", 12.3D);
        ModifiableRepositoryData subComposite1 = entry1.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subComposite1.setValue("singleStringData", "repeater entry 1 - sub composite - string value");
        
        ModifiableRepositoryData entry2 = repeater.addRepositoryData("2", RepositoryConstants.COMPOSITE_NODETYPE);
        entry2.setValue("singleDoubleData", 6.0D);
        ModifiableRepositoryData subComposite2 = entry2.addRepositoryData("subcomposite", RepositoryConstants.COMPOSITE_NODETYPE);
        subComposite2.setValue("singleStringData", "repeater entry 2 - sub composite - string value");
    }
    
    private ModifiableContent _createContent(String title)
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        
        ModifiableContent content = root.createChild(title, "ametys:defaultContent");

        content.setTypes(new String[]{"org.ametys.cms.contenttype.Test"});
        content.setTitle(title, Locale.FRANCE);
        content.setLanguage("fr");
        content.setCreationDate(ZonedDateTime.now());
        content.setCreator(new UserIdentity("john", "anyPopulation"));
        content.setLastModified(ZonedDateTime.now());
        content.setLastContributor(new UserIdentity("brian", "anyPopulation"));
        
        return content;
    }
    
    private void _assertEqualsStreamContent(RepositoryData resourceRepositoryData, String text) throws IOException
    {
        Assert.assertTrue(resourceRepositoryData.hasValue(ResourceElementTypeHelper.DATA_IDENTIFIER, ResourceElementTypeHelper.METADATA_PREFIX));
        
        try (StringWriter writer = new StringWriter())
        {
            InputStream stream = resourceRepositoryData.getStream(ResourceElementTypeHelper.DATA_IDENTIFIER, ResourceElementTypeHelper.METADATA_PREFIX);
            String encoding = resourceRepositoryData.getString(ResourceElementTypeHelper.ENCODING_IDENTIFIER, ResourceElementTypeHelper.METADATA_PREFIX);
            IOUtils.copy(stream, writer, encoding);
            Assert.assertEquals(text, writer.toString());
        }
    }
}
