/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;
import org.junit.internal.ComparisonCriteria;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.holder.ModifiableModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.impl.DefaultModifiableModelLessDataHolder;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModifiableModelLessDataHolder#synchronizeValues(Map)} implementation test case
 */
@NeedFeatures({
    "test-data/test-model-less-data-types",
    
    "repository/resolver", 
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModifiableModelLessDataHolderSynchronizeTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;

    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;
    
    private static RepositoryElementType<Geocode> _geocodeType;
    private static RepositoryElementType<UserIdentity> _userType;
    private static RepositoryElementType<ZonedDateTime> _dateTimeType;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".TestModelLess");
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");
        
        _geocodeType = (RepositoryElementType<Geocode>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID);
        _userType = (RepositoryElementType<UserIdentity>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _dateTimeType = (RepositoryElementType<ZonedDateTime>) _testDataTypeExtensionPoint.getExtension(org.ametys.runtime.model.type.ModelItemTypeConstants.DATETIME_TYPE_ID);
    }
    
    /**
     * Test {@link ModifiableModelLessDataHolder#synchronizeValues(Map)} method
     * This test uses a {@link JCRRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testSynchronizeValuesWithJCRRepositoryData() throws Exception
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSynchronizeValues", DATA_NODE_TYPE);
        _testSynchronizeValues(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelLessDataHolder#synchronizeValues(Map)} method
     * This test uses a {@link MemoryRepositoryData}
     * @throws Exception if an error occurs
     */
    @Test
    public void testSynchronizeValuesWithMemoryRepositoryData() throws Exception
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSynchronizeValues", DATA_NODE_TYPE);
        _testSynchronizeValues(repositoryData);
    }
    
    private void _testSynchronizeValues(ModifiableRepositoryData repositoryData) throws Exception
    {
        ModifiableModelLessDataHolder dataHolder = new DefaultModifiableModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);

        Long[] multipleLongData = new Long[] {32L, 784L};
        UserIdentity[] multipleUserData = new UserIdentity[] {USER1_TEST_IDENTITY, USER3_TEST_IDENTITY};
        ZonedDateTime[] multipleDateData = new ZonedDateTime[] {ZonedDateTime.of(2018, 9, 11, 12, 11, 45, 0, ZoneId.systemDefault()), ZonedDateTime.of(2015, 12, 06, 15, 18, 21, 0, ZoneId.systemDefault())};
        
        Map<String, Object> valuesToSynchronize = Map.of("singleStringData", "value of the single string data",
                "multipleLongData", multipleLongData,
                "multipleUserData", multipleUserData,
                "singleGeocodeData", new Geocode(85.487D, 6.254D),
                "composite", Map.of("singleLongData", 54L,
                        "subcomposite", Map.of("singleStringData", "string value in the sub composite",
                                "multipleDateData", multipleDateData)));
        
        boolean hasChanged = dataHolder.synchronizeValues(valuesToSynchronize);
        Assert.assertTrue(hasChanged);
        
        Assert.assertEquals("value of the single string data", repositoryData.getString("singleStringData"));
        Assert.assertArrayEquals(multipleLongData, repositoryData.getLongs("multipleLongData"));
        
        Assert.assertArrayEquals(multipleUserData, (UserIdentity[]) _userType.read(repositoryData, "multipleUserData"));
        
        Geocode geocode = (Geocode) _geocodeType.read(repositoryData, "singleGeocodeData");
        Assert.assertEquals(Double.valueOf(85.487), geocode.getLatitude());
        Assert.assertEquals(Double.valueOf(6.254), geocode.getLongitude());
        
        RepositoryData composite = repositoryData.getRepositoryData("composite");
        Assert.assertEquals(Long.valueOf(54), composite.getLong("singleLongData"));
        
        RepositoryData subcomposite = composite.getRepositoryData("subcomposite");
        Assert.assertEquals("string value in the sub composite", subcomposite.getString("singleStringData"));
        new ZonedDateTimeComparisonCriteria().arrayEquals(null, multipleDateData, _dateTimeType.read(subcomposite, "multipleDateData"));
    }
    
    private static class ZonedDateTimeComparisonCriteria extends ComparisonCriteria
    {
        @Override
        protected void assertElementsEqual(Object expected, Object actual)
        {
            ZonedDateTime expectedZDT = (ZonedDateTime) expected;
            ZonedDateTime actualZDT = (ZonedDateTime) actual;
            
            Assert.assertTrue(actualZDT.isEqual(expectedZDT));
        }
    }
}
