/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringWriter;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Calendar;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.Binary;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.data.type.ResourceElementTypeHelper;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.AmetysRepositoryException;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.holder.ModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.ModifiableModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.ModifiableModelLessDataHolder;
import org.ametys.plugins.repository.data.holder.group.Composite;
import org.ametys.plugins.repository.data.holder.group.ModifiableComposite;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelLessComposite;
import org.ametys.plugins.repository.data.holder.impl.DefaultModifiableModelLessDataHolder;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryModelItemType;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.exception.UnknownTypeException;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModifiableModelLessDataHolder} implementation test case
 */
@NeedFeatures({
    "test-data/test-model-less-data-types",
    "test-data/test-model-less-with-old-api-data-types",
    
    "repository/resolver",
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModifiableModelLessDataHolderTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String DATA_NODE_TYPE = "ametys:unstructured";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPointForOldAPITests;
    
    private static JCRRepositoryData _jcrRoot;
    private static MemoryRepositoryData _memoryRoot;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".TestModelLess");
        _testDataTypeExtensionPointForOldAPITests = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".TestModelLessOldAPI");
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _jcrRoot = new JCRRepositoryData(rootAmetysObject.getNode());
        _memoryRoot = new MemoryRepositoryData("root");
    }
    
    /**
     * Test {@link ModifiableModelLessDataHolder#setValue(String, Object, String)} method's nominal cases
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetAndNominalCasesWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetAndNominalCases", DATA_NODE_TYPE);
        _testSetValueNominalCases(repositoryData);
    }
    
    /**
     * Test {@link ModelLessDataHolder#getValue(String)} method's nominal cases
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueNominalCasesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueNominalCases", DATA_NODE_TYPE);
        _testSetValueNominalCases(repositoryData);
    }

    private void _testSetValueNominalCases(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelLessDataHolder dataHolder = new DefaultModifiableModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        // String
        dataHolder.setValue("stringData", "string value");
        Assert.assertEquals("string value", repositoryData.getString("stringData"));
        
        // String[]
        String[] multipleStringData = new String[] {"first string value", "second string value"};
        dataHolder.setValue("multipleStringData", multipleStringData);
        Assert.assertArrayEquals(multipleStringData, repositoryData.getStrings("multipleStringData"));
        
        // Long
        dataHolder.setValue("longDataSettedAsInt", 7);
        Assert.assertEquals(Long.valueOf(7), repositoryData.getLong("longDataSettedAsInt"));
        dataHolder.setValue("longData", 8L);
        Assert.assertEquals(Long.valueOf(8), repositoryData.getLong("longData"));
        
        // Geocode
        dataHolder.setValue("geocodeData", new Geocode(65.215, 2.698));
        RepositoryData geocodeData = repositoryData.getRepositoryData("geocodeData");
        Assert.assertEquals(Double.valueOf(65.215), geocodeData.getDouble(Geocode.LATITUDE_IDENTIFIER));
        Assert.assertEquals(Double.valueOf(2.698), geocodeData.getDouble(Geocode.LONGITUDE_IDENTIFIER));
        
        // Composite
        ModifiableModelLessComposite composite = dataHolder.getComposite("composite", true);
        Assert.assertTrue(repositoryData.hasValue("composite"));
        RepositoryData compositeData = repositoryData.getRepositoryData("composite");
        
        // Composite - Double
        dataHolder.setValue("composite/doubleData", 17D);
        Assert.assertEquals(Double.valueOf(17), compositeData.getDouble("doubleData"));
        
        // Composite - User
        composite.setValue("userData", new UserIdentity("tmathias", "hinterland"));
        Assert.assertTrue(compositeData.hasValue("userData"));
        RepositoryData userData = compositeData.getRepositoryData("userData");
        Assert.assertEquals("tmathias", userData.getString("login"));
        Assert.assertEquals("hinterland", userData.getString("population"));
    }
    
    /**
     * Test {@link ModifiableModelLessDataHolder} implementation to set a value with a type and change its type at the next setValue
     * Tested with a complex type ({@link Geocode}) to test the transition from a node to a property in repository
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testChangeDataTypeOnSecondSetValueCallWithJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testChangeDataTypeOnSecondSetValueCall", DATA_NODE_TYPE);
        _testChangeDataTypeOnSecondSetValueCall(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelLessDataHolder} implementation to set a value with a type and change its type at the next setValue
     * Tested with a complex type ({@link Geocode}) to test the transition from a node to a property in repository
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testChangeDataTypeOnSecondSetValueCallWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testChangeDataTypeOnSecondSetValueCall", DATA_NODE_TYPE);
        _testChangeDataTypeOnSecondSetValueCall(repositoryData);
        
    }
    
    private void _testChangeDataTypeOnSecondSetValueCall(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelLessDataHolder dataHolder = new DefaultModifiableModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        // Set a complex value to the changing data
        dataHolder.setValue("changingData", new Geocode(39.0, 1.7));
        Assert.assertTrue(repositoryData.hasValue("changingData"));
        RepositoryData changingData = repositoryData.getRepositoryData("changingData");
        Assert.assertEquals(Double.valueOf(39), changingData.getDouble(Geocode.LATITUDE_IDENTIFIER));
        Assert.assertEquals(Double.valueOf(1.7), changingData.getDouble(Geocode.LONGITUDE_IDENTIFIER));
        
        // Set a simple value to the changing data
        dataHolder.setValue("changingData", 95);
        Assert.assertEquals(Long.valueOf(95), repositoryData.getLong("changingData"));
        try
        {
            // Make sure the complex value has been removed
            repositoryData.getRepositoryData("changingData");
            Assert.fail("the old geocode data should have been removed");
        }
        catch (AmetysRepositoryException e)
        {
            // Do nothing
        }
        
        // Set a multiple complex value to the changing data
        dataHolder.setValue("changingData", new UserIdentity[] {new UserIdentity("kebous", "HDL"), new UserIdentity("wallace", "HDL")});
        changingData = repositoryData.getRepositoryData("changingData");
        Assert.assertTrue(changingData.hasValue("1"));
        Assert.assertTrue(changingData.hasValue("2"));
        try
        {
            // Make sure the simple value has been removed
            repositoryData.getLong("changingData");
            Assert.fail("the old long data should have been removed");
        }
        catch (AmetysRepositoryException e)
        {
            // Do nothing
        }
        
        // Set a single complex value again to the changing data
        dataHolder.setValue("changingData", new Geocode(76.9, 2.8));
        changingData = repositoryData.getRepositoryData("changingData");
        Assert.assertEquals(Double.valueOf(76.9), changingData.getDouble(Geocode.LATITUDE_IDENTIFIER));
        Assert.assertEquals(Double.valueOf(2.8), changingData.getDouble(Geocode.LONGITUDE_IDENTIFIER));
        // Make sure the values of the multiple complex value has been removed
        Assert.assertFalse(changingData.hasValue("1"));
    }
    
    /**
     * Tests the {@link ModifiableModelLessDataHolder} behavior when setting multiple values with only one entry
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testMultipleValuesWithOnlyOneEntryJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testMultipleValuesWithOnlyOneEntry", DATA_NODE_TYPE);
        _testMultipleValuesWithOnlyOneEntry(repositoryData);
    }
    
    /**
     * Tests the {@link ModifiableModelLessDataHolder} behavior when setting multiple values with only one entry
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testMultipleValuesWithOnlyOneEntryMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testMultipleValuesWithOnlyOneEntry", DATA_NODE_TYPE);
        _testMultipleValuesWithOnlyOneEntry(repositoryData);
    }
    
    private void _testMultipleValuesWithOnlyOneEntry(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelLessDataHolder dataHolder = new DefaultModifiableModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        // Test with a simple string attribute
        dataHolder.setValue("multipleStringData", new String[]{"lonely string value"});
        String[] multipleStringData = dataHolder.getValue("multipleStringData");
        Assert.assertArrayEquals(new String[]{"lonely string value"}, multipleStringData);
        
        // Test with a geocode attribute
        dataHolder.setValue("multipleGeocodeData", new Geocode[]{new Geocode(56.4, 9.2)});
        Geocode[] multipleGeocodeData = dataHolder.getValue("multipleGeocodeData");
        Assert.assertEquals(1, multipleGeocodeData.length);
        Assert.assertEquals(Double.valueOf(56.4), multipleGeocodeData[0].getLatitude());
        Assert.assertEquals(Double.valueOf(9.2), multipleGeocodeData[0].getLongitude());
    }
    
    /**
     * Tests the {@link ModifiableModelLessDataHolder#setValue(String, Object)} when the parent group in the path does not exist
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetValueWhenGroupDoesNotExistJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueWhenGroupDoesNotExist", DATA_NODE_TYPE);
        _testSetValueWhenGroupDoesNotExist(repositoryData);
    }
    
    /**
     * Tests the {@link ModifiableModelLessDataHolder#setValue(String, Object)} when the parent group in the path does not exist
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueWhenGroupDoesNotExistMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueWhenGroupDoesNotExist", DATA_NODE_TYPE);
        _testSetValueWhenGroupDoesNotExist(repositoryData);
    }
    
    private void _testSetValueWhenGroupDoesNotExist(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelLessDataHolder dataHolder = new DefaultModifiableModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        try
        {
            // Tests to set a value when the parents don't exist
            dataHolder.setValue("inexisting/group/stringData", "my string");
            Assert.fail("parent group doesn't exist, it should not be possible to set the value");
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains("exist"));
        }
    }
    
    /**
     * Tests the {@link ModifiableModelLessDataHolder#getType(String)} method for a data with an unknown type
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testSetValueOnDataWithAnUnknownTypeJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testSetValueOnDataWithAnUnknownType", DATA_NODE_TYPE);
        _testSetValueOnDataWithAnUnknownType(repositoryData);
    }
    
    /**
     * Tests the {@link ModifiableModelLessDataHolder#getType(String)} method for a data with an unknown type
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testSetValueOnDataWithAnUnknownTypeMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testSetValueOnDataWithAnUnknownType", DATA_NODE_TYPE);
        _testSetValueOnDataWithAnUnknownType(repositoryData);
    }
    
    private void _testSetValueOnDataWithAnUnknownType(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelLessDataHolder dataHolder = new DefaultModifiableModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        try
        {
            // There is no existing StringBuilder type 
            dataHolder.setValue("stringBuilderData", new StringBuilder());
            Assert.fail("the type of this data is not available, it should not be possible to set a value with this type");
        }
        catch (UnknownTypeException e)
        {
            Assert.assertTrue(e.getMessage().contains("found"));
        }
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} behavior when setting <code>null</code> values
     * Setting <code>null</code> values leads to empty a data, not removing it
     * This test uses a {@link JCRRepositoryData}
     * @throws IOException if an error occurs while writing or reading resources
     */
    @Test
    @NeedFeatures({
        "cms/factories",
        "test/content-type",
    })
    public void testEmptyValuesWithJCRRepositoryData() throws IOException
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testEmptyValues", DATA_NODE_TYPE);
        _testEmptyValues(repositoryData);
        _testEmptyResourceValues(repositoryData);
    }
    
    /**
     * Test {@link ModifiableModelAwareDataHolder#setValue(String, Object)} behavior when setting <code>null</code> values
     * Setting <code>null</code> values leads to empty a data, not removing it
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    @NeedFeatures({
        "cms/factories",
        "test/content-type",
    })
    public void testEmptyValuesWithMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testEmptyValues", DATA_NODE_TYPE);
        _testEmptyValues(repositoryData);
    }
    
    @SuppressWarnings("static-access")
    private void _testEmptyValues(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelLessDataHolder dataHolder = new DefaultModifiableModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);

        // single string data
        Assert.assertFalse(repositoryData.hasValue("singleStringData"));
        dataHolder.setValue("singleStringData", null, ModelItemTypeConstants.STRING_TYPE_ID);
        Assert.assertEquals(StringUtils.EMPTY, repositoryData.getString("singleStringData"));
        dataHolder.setValue("singleStringData", StringUtils.EMPTY);
        Assert.assertEquals(StringUtils.EMPTY, repositoryData.getString("singleStringData"));
        
        // multiple string data
        Assert.assertFalse(repositoryData.hasValue("multipleStringData"));
        dataHolder.setValue("multipleStringData", new String[0], ModelItemTypeConstants.STRING_TYPE_ID);
        Assert.assertEquals(0, repositoryData.getStrings("multipleStringData").length);
        dataHolder.setValue("multipleStringData", new String[] {"a string value", "another string value"});
        Assert.assertEquals(2, repositoryData.getStrings("multipleStringData").length);
        dataHolder.setValue("multipleStringData", new String[0], ModelItemTypeConstants.STRING_TYPE_ID);
        Assert.assertEquals(0, repositoryData.getStrings("multipleStringData").length);
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleStringData", new String[] {"value 1", null, "value 3"}, ModelItemTypeConstants.STRING_TYPE_ID);

        // single long data
        Assert.assertFalse(repositoryData.hasValue("singleLongData"));
        dataHolder.setValue("singleLongData", null, ModelItemTypeConstants.LONG_TYPE_ID);
        Assert.assertFalse(repositoryData.hasValue("singleLongData"));
        Assert.assertTrue(repositoryData.getBoolean("singleLongData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleLongData", 18L);
        Assert.assertEquals(Long.valueOf(18), repositoryData.getLong("singleLongData"));
        Assert.assertFalse(repositoryData.hasValue("singleLongData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleLongData", null, ModelItemTypeConstants.LONG_TYPE_ID);
        Assert.assertFalse(repositoryData.hasValue("singleLongData"));
        Assert.assertTrue(repositoryData.getBoolean("singleLongData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        
        // multiple long data
        Assert.assertFalse(repositoryData.hasValue("multipleLongData"));
        dataHolder.setValue("multipleLongData", new Long[0], ModelItemTypeConstants.LONG_TYPE_ID);
        Assert.assertEquals(0, repositoryData.getLongs("multipleLongData").length);
        dataHolder.setValue("multipleLongData", new Long[] {21L, 78L, 31L});
        Assert.assertEquals(3, repositoryData.getLongs("multipleLongData").length);
        dataHolder.setValue("multipleLongData", null, ModelItemTypeConstants.LONG_TYPE_ID);
        Assert.assertEquals(0, repositoryData.getLongs("multipleLongData").length);
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleLongData", new Long[] {21L, null, 31L}, ModelItemTypeConstants.LONG_TYPE_ID);

        // single double data
        Assert.assertFalse(repositoryData.hasValue("singleDoubleData"));
        dataHolder.setValue("singleDoubleData", null, ModelItemTypeConstants.DOUBLE_TYPE_ID);
        Assert.assertFalse(repositoryData.hasValue("singleDoubleData"));
        Assert.assertTrue(repositoryData.getBoolean("singleDoubleData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleDoubleData", 185D);
        Assert.assertEquals(Double.valueOf(185), repositoryData.getDouble("singleDoubleData"));
        Assert.assertFalse(repositoryData.hasValue("singleDoubleData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleDoubleData", null, ModelItemTypeConstants.DOUBLE_TYPE_ID);
        Assert.assertFalse(repositoryData.hasValue("singleDoubleData"));
        Assert.assertTrue(repositoryData.getBoolean("singleDoubleData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        
        // multiple double data
        Assert.assertFalse(repositoryData.hasValue("multipleDoubleData"));
        dataHolder.setValue("multipleDoubleData", new Double[0], ModelItemTypeConstants.DOUBLE_TYPE_ID);
        Assert.assertEquals(0, repositoryData.getDoubles("multipleDoubleData").length);
        dataHolder.setValue("multipleDoubleData", new Double[] {21D, 78D});
        Assert.assertEquals(2, repositoryData.getDoubles("multipleDoubleData").length);
        dataHolder.setValue("multipleDoubleData", null, ModelItemTypeConstants.DOUBLE_TYPE_ID);
        Assert.assertEquals(0, repositoryData.getDoubles("multipleDoubleData").length);
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleDoubleData", new Double[] {null, 78D}, ModelItemTypeConstants.DOUBLE_TYPE_ID);
        
        // single boolean data
        Assert.assertFalse(repositoryData.hasValue("singleBooleanData"));
        dataHolder.setValue("singleBooleanData", null, ModelItemTypeConstants.BOOLEAN_TYPE_ID);
        Assert.assertFalse(repositoryData.hasValue("singleBooleanData"));
        Assert.assertTrue(repositoryData.getBoolean("singleBooleanData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleBooleanData", true);
        Assert.assertTrue(repositoryData.getBoolean("singleBooleanData"));
        Assert.assertFalse(repositoryData.hasValue("singleBooleanData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleBooleanData", null, ModelItemTypeConstants.BOOLEAN_TYPE_ID);
        Assert.assertFalse(repositoryData.hasValue("singleBooleanData"));
        Assert.assertTrue(repositoryData.getBoolean("singleBooleanData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        
        // single date time data
        Assert.assertFalse(repositoryData.hasValue("singleDatetimeData"));
        dataHolder.setValue("singleDatetimeData", null, ModelItemTypeConstants.DATETIME_TYPE_ID);
        Assert.assertFalse(repositoryData.hasValue("singleDatetimeData"));
        Assert.assertTrue(repositoryData.getBoolean("singleDatetimeData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleDatetimeData", ZonedDateTime.of(2019, 5, 24, 12, 10, 23, 45, ZoneId.systemDefault()));
        Assert.assertEquals(2019, repositoryData.getDate("singleDatetimeData").get(Calendar.YEAR));
        Assert.assertFalse(repositoryData.hasValue("singleDatetimeData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        dataHolder.setValue("singleDatetimeData", null, ModelItemTypeConstants.DATETIME_TYPE_ID);
        Assert.assertFalse(repositoryData.hasValue("singleDatetimeData"));
        Assert.assertTrue(repositoryData.getBoolean("singleDatetimeData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
        
        // multiple date time data
        Assert.assertFalse(repositoryData.hasValue("multipleDatetimeData"));
        dataHolder.setValue("multipleDatetimeData", new ZonedDateTime[0], ModelItemTypeConstants.DATETIME_TYPE_ID);
        Assert.assertEquals(0, repositoryData.getDates("multipleDatetimeData").length);
        dataHolder.setValue("multipleDatetimeData", new ZonedDateTime[] {ZonedDateTime.now()});
        Assert.assertEquals(1, repositoryData.getDates("multipleDatetimeData").length);
        dataHolder.setValue("multipleDatetimeData", null, ModelItemTypeConstants.DATETIME_TYPE_ID);
        Assert.assertEquals(0, repositoryData.getDates("multipleDatetimeData").length);
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleDatetimeData", new ZonedDateTime[] {ZonedDateTime.now(), null}, ModelItemTypeConstants.DATETIME_TYPE_ID);
        
        // single user data
        Assert.assertFalse(repositoryData.hasValue("singleUserData"));
        dataHolder.setValue("singleUserData", null, ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _assertSingleUserDataIsPresentButEmpty(repositoryData, "singleUserData");
        dataHolder.setValue("singleUserData", new UserIdentity("fmercury", "queen"));
        Assert.assertTrue(repositoryData.hasValue("singleUserData"));
        Assert.assertTrue(repositoryData.getRepositoryData("singleUserData").hasValue("login"));
        Assert.assertTrue(repositoryData.getRepositoryData("singleUserData").hasValue("population"));
        dataHolder.setValue("singleUserData", null, ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _assertSingleUserDataIsPresentButEmpty(repositoryData, "singleUserData");
        
        // multiple user data
        Assert.assertFalse(repositoryData.hasValue("multipleUserData"));
        dataHolder.setValue("multipleUserData", new UserIdentity[0], ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _assertMultipleUserDataIsPresentButEmpty(repositoryData, "multipleUserData");
        dataHolder.setValue("multipleUserData", new UserIdentity[] {new UserIdentity("mknopfler", "direstraits")});
        Assert.assertTrue(repositoryData.hasValue("multipleUserData"));
        Assert.assertTrue(repositoryData.getRepositoryData("multipleUserData").hasValue("1"));
        Assert.assertTrue(repositoryData.getRepositoryData("multipleUserData").getRepositoryData("1").hasValue("login"));
        Assert.assertTrue(repositoryData.getRepositoryData("multipleUserData").getRepositoryData("1").hasValue("population"));
        Assert.assertFalse(repositoryData.getRepositoryData("multipleUserData").hasValue("2"));
        dataHolder.setValue("multipleUserData", null, ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
        _assertMultipleUserDataIsPresentButEmpty(repositoryData, "multipleUserData");
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleUserData", new UserIdentity[] {null, new UserIdentity("mknopfler", "direstraits")}, ModelItemTypeConstants.USER_ELEMENT_TYPE_ID);
    }
    
    private void _testEmptyResourceValues(ModifiableRepositoryData repositoryData) throws IOException
    {
        ModifiableModelLessDataHolder dataHolder = new DefaultModifiableModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        // single binary data
        Binary binary = _createBinary("my binary content");
        Assert.assertFalse(repositoryData.hasValue("singleBinaryData"));
        dataHolder.setValue("singleBinaryData", null, ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _assertSingleBinaryDataIsPresentButEmpty(repositoryData, "singleBinaryData");
        dataHolder.setValue("singleBinaryData", binary, ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        Assert.assertTrue(repositoryData.hasValue("singleBinaryData"));
        _assertEqualsStreamContent(repositoryData.getRepositoryData("singleBinaryData"), "my binary content");
        dataHolder.setValue("singleBinaryData", null, ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _assertSingleBinaryDataIsPresentButEmpty(repositoryData, "singleBinaryData");
        
        // multiple binary data
        Assert.assertFalse(repositoryData.hasValue("multipleBinaryData"));
        dataHolder.setValue("multipleBinaryData", new Binary[0], ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _assertMultipleBinaryDataIsPresentButEmpty(repositoryData, "multipleBinaryData");
        dataHolder.setValue("multipleBinaryData", new Binary[] {binary}, ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        Assert.assertTrue(repositoryData.hasValue("multipleBinaryData"));
        Assert.assertTrue(repositoryData.getRepositoryData("multipleBinaryData").hasValue("1"));
        _assertEqualsStreamContent(repositoryData.getRepositoryData("multipleBinaryData").getRepositoryData("1"), "my binary content");
        Assert.assertFalse(repositoryData.getRepositoryData("multipleBinaryData").hasValue("2"));
        dataHolder.setValue("multipleBinaryData", null, ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
        _assertMultipleBinaryDataIsPresentButEmpty(repositoryData, "multipleBinaryData");
        _assertMultipleDataDoesNotSupportNullValue(dataHolder, "multipleBinaryData", new Binary[] {null, binary}, ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID);
    }
    
    private <T> void _assertMultipleDataDoesNotSupportNullValue(ModifiableModelLessDataHolder dataHolder, String dataPath, T[] values, String typeId)
    {
        try
        {
            dataHolder.setValue(dataPath, values, typeId);
            Assert.fail("Setting of null values in a multiple data should not be supported");
        }
        catch (IllegalArgumentException e)
        {
            Assert.assertTrue(e.getMessage().contains(dataPath));
        }
    }
    
    private void _assertMultipleUserDataIsPresentButEmpty(RepositoryData repositoryData, String dataName)
    {
        Assert.assertTrue(repositoryData.hasValue(dataName));
        RepositoryData userData = repositoryData.getRepositoryData(dataName);
        
        // The data can have entries if they are empty
        for (String entryName : userData.getDataNames())
        {
            _assertSingleUserDataIsPresentButEmpty(userData, entryName);
        }
    }
    
    private void _assertSingleUserDataIsPresentButEmpty(RepositoryData repositoryData, String dataName)
    {
        Assert.assertTrue(repositoryData.hasValue(dataName));
        RepositoryData data = repositoryData.getRepositoryData(dataName);
        Assert.assertTrue(StringUtils.isEmpty(data.getString("login")));
        Assert.assertTrue(StringUtils.isEmpty(data.getString("population")));
    }
    
    private void _assertMultipleBinaryDataIsPresentButEmpty(RepositoryData repositoryData, String dataName)
    {
        Assert.assertTrue(repositoryData.hasValue(dataName));
        RepositoryData userData = repositoryData.getRepositoryData(dataName);
        
        // The data can have entries if they are empty
        for (String entryName : userData.getDataNames())
        {
            _assertSingleBinaryDataIsPresentButEmpty(userData, entryName);
        }
    }
    
    private void _assertSingleBinaryDataIsPresentButEmpty(RepositoryData repositoryData, String dataName)
    {
        Assert.assertTrue(repositoryData.hasValue(dataName));
        RepositoryData data = repositoryData.getRepositoryData(dataName);
        Assert.assertTrue(data.getBoolean("isEmpty", "ametys-internal"));
    }
    
    /**
     * Tests the {@link ModifiableModelLessDataHolder#removeValue(String)} method
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testRemoveValueJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testRemoveValue", DATA_NODE_TYPE);
        _testRemoveValue(repositoryData);
    }
    
    /**
     * Tests the {@link ModifiableModelLessDataHolder#removeValue(String)} method
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testRemoveValueMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testRemoveValue", DATA_NODE_TYPE);
        _testRemoveValue(repositoryData);
    }
    
    private void _testRemoveValue(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelLessDataHolder dataHolder = new DefaultModifiableModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        dataHolder.setValue("simpleDoubleData", 25.4);
        Assert.assertTrue(repositoryData.hasValue("simpleDoubleData"));
        dataHolder.removeValue("simpleDoubleData");
        Assert.assertFalse(repositoryData.hasValue("simpleDoubleData"));
        
        ModifiableComposite composite = dataHolder.getComposite("composite", true);

        dataHolder.setValue("composite/multipleStringData", new String[]{"string1", "string2"});
        Assert.assertTrue(repositoryData.getRepositoryData("composite").hasValue("multipleStringData"));
        composite.removeValue("multipleStringData");
        Assert.assertFalse(repositoryData.getRepositoryData("composite").hasValue("multipleStringData"));
        
        Assert.assertTrue(repositoryData.hasValue("composite"));
        dataHolder.removeValue("composite");
        Assert.assertFalse(repositoryData.hasValue("composite"));
    }
    
    /**
     * Tests the {@link ModifiableModelLessDataHolder#removeValue(String)} method on inexisting data
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testRemoveValueInexistingDataJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testRemoveValueInexistingData", DATA_NODE_TYPE);
        _testRemoveValueInexistingData(repositoryData);
    }
    
    /**
     * Tests the {@link ModifiableModelLessDataHolder#removeValue(String)} method on inexisting data
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testRemoveValueInexistingDataMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testRemoveValueInexistingData", DATA_NODE_TYPE);
        _testRemoveValueInexistingData(repositoryData);
    }
    
    private void _testRemoveValueInexistingData(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelLessDataHolder dataHolder = new DefaultModifiableModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        dataHolder.removeValue("inexistingData");
        dataHolder.removeValue("inexisting/data/in/group");
    }
    
    /**
     * Tests the {@link ModifiableModelLessDataHolder#removeValue(String)} method on empty data
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testRemoveEmptyDataRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testRemoveEmptyData", DATA_NODE_TYPE);
        _testRemoveEmptyData(repositoryData);
    }
    
    /**
     * Tests the {@link ModifiableModelLessDataHolder#removeValue(String)} method on empty data
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testRemoveEmptyDataMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testRemoveEmptyData", DATA_NODE_TYPE);
        _testRemoveEmptyData(repositoryData);
    }
    
    private void _testRemoveEmptyData(ModifiableRepositoryData repositoryData)
    {
        ModifiableModelLessDataHolder dataHolder = new DefaultModifiableModelLessDataHolder(_testDataTypeExtensionPoint, repositoryData);
        
        repositoryData.setValue("emptyData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, true, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL);
        dataHolder.removeValue("emptyData");
        Assert.assertFalse(repositoryData.hasValue("emptyData"));
        Assert.assertFalse(repositoryData.hasValue("emptyData" + RepositoryModelItemType.EMPTY_METADATA_SUFFIX, RepositoryConstants.NAMESPACE_PREFIX_INTERNAL));
    }
    
    /**
     * Tests the {@link ModifiableModelLessDataHolder} methods on composites created with the old node type
     * This test uses a {@link JCRRepositoryData}
     */
    @Test
    public void testAllMethodsWithOldNodeTypeJCRRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _jcrRoot.addRepositoryData("testAllMethodsWithOldNodeType", DATA_NODE_TYPE);
        _testAllMethodsWithOldNodeType(repositoryData);
    }
    
    /**
     * Tests the {@link ModifiableModelLessDataHolder} methods on composites created with the old node type
     * This test uses a {@link MemoryRepositoryData}
     */
    @Test
    public void testAllMethodsWithOldNodeTypeMemoryRepositoryData()
    {
        ModifiableRepositoryData repositoryData = _memoryRoot.addRepositoryData("testAllMethodsWithOldNodeType", DATA_NODE_TYPE);
        _testAllMethodsWithOldNodeType(repositoryData);
    }
    
    @SuppressWarnings("static-access")
    private void _testAllMethodsWithOldNodeType(ModifiableRepositoryData repositoryData)
    {
        // Add data in the repository with the old node type
        repositoryData.addRepositoryData("composite", "ametys:compositeMetadata");
        
        ModifiableModelLessDataHolder dataHolder = new DefaultModifiableModelLessDataHolder(_testDataTypeExtensionPointForOldAPITests, repositoryData);
        
        // Set value
        dataHolder.setValue("composite/stringData", "string value in the composite");
        
        // Add composite
        dataHolder.getComposite("composite/subComposite", true);
        
        // Has value
        Assert.assertTrue(dataHolder.hasValue("composite/stringData"));
        Assert.assertTrue(dataHolder.hasValueOrEmpty("composite/subComposite"));
        Assert.assertFalse(dataHolder.hasValue("composite/subComposite"));
        
        // Is multiple
        Assert.assertFalse(dataHolder.isMultiple("composite/stringData"));
        Assert.assertFalse(dataHolder.isMultiple("composite/subComposite"));
        
        // Get type
        Assert.assertEquals(ModelItemTypeConstants.STRING_TYPE_ID, dataHolder.getType("composite/stringData").getId());
        Assert.assertEquals(ModelItemTypeConstants.COMPOSITE_TYPE_ID, dataHolder.getType("composite/subComposite").getId());
        
        // Get value
        Assert.assertEquals("string value in the composite", dataHolder.getValue("composite/stringData"));
        Assert.assertNotNull(dataHolder.getValue("composite/subComposite"));
        Assert.assertTrue(dataHolder.getValue("composite/subComposite") instanceof Composite);
        
        // Get composite
        Assert.assertNotNull(dataHolder.getComposite("composite/subComposite"));
        
        // Remove value
        dataHolder.removeValue("composite/stringData");
        dataHolder.removeValue("composite/subComposite");
        
        // Has value after removal
        Assert.assertFalse(dataHolder.hasValue("composite/stringData"));
        Assert.assertFalse(dataHolder.hasValue("composite/subComposite"));
    }
    
    private Binary _createBinary(String text) throws IOException
    {
        return _createBinary(text, "UTF-8", "myBinary.txt", ZonedDateTime.now(), "text/plain");
    }
    
    private Binary _createBinary(String text, String encoding, String filename, ZonedDateTime lastModificationDate, String mimeType) throws IOException
    {
        Binary binary = new Binary();
        try (OutputStream os = binary.getOutputStream())
        {
            os.write(text.getBytes());
        }
        binary.setEncoding(encoding);
        binary.setFilename(filename);
        binary.setLastModificationDate(lastModificationDate);
        binary.setMimeType(mimeType);
        
        return binary;
    }
    
    private void _assertEqualsStreamContent(RepositoryData resourceRepositoryData, String text) throws IOException
    {
        Assert.assertTrue(resourceRepositoryData.hasValue(ResourceElementTypeHelper.DATA_IDENTIFIER, ResourceElementTypeHelper.METADATA_PREFIX));
        
        try (StringWriter writer = new StringWriter())
        {
            InputStream stream = resourceRepositoryData.getStream(ResourceElementTypeHelper.DATA_IDENTIFIER, ResourceElementTypeHelper.METADATA_PREFIX);
            String encoding = resourceRepositoryData.getString(ResourceElementTypeHelper.ENCODING_IDENTIFIER, ResourceElementTypeHelper.METADATA_PREFIX);
            IOUtils.copy(stream, writer, encoding);
            Assert.assertEquals(text, writer.toString());
        }
    }
}
