/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder.group.impl;

import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.holder.impl.DefaultModifiableModelAwareDataHolder;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.holder.ModifiableModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.values.SynchronizableRepeater;
import org.ametys.plugins.repository.data.holder.values.SynchronizationContext;
import org.ametys.plugins.repository.data.holder.values.SynchronizationResult;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.ModelItemGroup;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method test cases
 */
@NeedFeatures({
    "test-data/test-model",
    
    "repository/repository.data.holder.helpers",
    "repository/resolver", 
    "repository/provider.jackrabbit",
    "repository/factories",
    "cms/workflow.content"
})
public class ModifiableModelAwareRepeaterSynchronizeTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __DATA_NODE_TYPE = "ametys:unstructured";
    private static final String __SINGLE_DOUBLE_DATA_NAME = "singleDoubleData";
    private static final String __SINGLE_LONG_DATA_NAME = "longWithDefaultValue";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static Model _model;

    private static JCRRepositoryData _root;
    private static ViewItemAccessor _repeaterViewItem;
    
    private ModifiableModelAwareRepeater _repeater;
    private SynchronizationContext _context;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _root = new JCRRepositoryData(rootAmetysObject.getNode());
        
        _repeaterViewItem = ModelViewItemGroup.of((ModelItemGroup) _model.getModelItem("composite/repeater"), __SINGLE_DOUBLE_DATA_NAME);
    }
    
    /**
     * Prepare components that are specific for each test
     * @throws Exception if an error occurs
     */
    @Before
    public void prepareSpecificComponents()throws Exception
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData(UUID.randomUUID().toString(), __DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        _repeater = dataHolder.getComposite("composite", true)
                .getRepeater("repeater", true);

        _context = SynchronizationContext.newInstance();
    }
    
    /**
     * Tests the {@link DefaultModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method
     * The repository is empty
     * Do not append entries in the repeater
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyAppendNoEntry() throws Exception
    {
        SynchronizableRepeater values = SynchronizableRepeater.appendOrRemove(
                List.of(),
                Set.of());
        
        SynchronizationResult result = _repeater.synchronizeValues(_repeaterViewItem, values, _context);

        Assert.assertFalse(result.hasChanged());
        Assert.assertEquals(0, _repeater.getSize());
    }
    
    /**
     * Tests the {@link DefaultModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method
     * The repository is empty
     * Append some entries in the repeater
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyAppendEntries() throws Exception
    {
        SynchronizableRepeater values = SynchronizableRepeater.appendOrRemove(
                List.of(
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 1.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 2.0)),
                Set.of());
        
        SynchronizationResult result = _repeater.synchronizeValues(_repeaterViewItem, values, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(2, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link DefaultModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method
     * The repository is empty
     * Try to remove a nonexistent repeater entry
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyRemoveEntry() throws Exception
    {
        SynchronizableRepeater values = SynchronizableRepeater.appendOrRemove(
                List.of(),
                Set.of(1));
        
        try
        {
            _repeater.synchronizeValues(_repeaterViewItem, values, _context);
            Assert.fail("the synchronization shouldn't have be able to remove the entry at position 1: there is no such entry");
        }
        catch (UnknownDataException e)
        {
            // Do nothing
        }
    }
    
    /**
     * Tests the {@link DefaultModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method
     * The repository contains some entries
     * Try to remove a nonexistent repeater entry
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveNonexistentEntry() throws Exception
    {
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 1.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 2.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 3.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 4.0);
        
        SynchronizableRepeater values = SynchronizableRepeater.appendOrRemove(
                List.of(),
                Set.of(5));
        
        try
        {
            _repeater.synchronizeValues(_repeaterViewItem, values, _context);
            Assert.fail("the synchronization shouldn't have be abel to remove the entry at position 5: there is no such entry");
        }
        catch (UnknownDataException e)
        {
            // Do nothing
        }
    }
    
    /**
     * Tests the {@link DefaultModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method
     * The repository contains some entries
     * Append other entries in the repeater
     * @throws Exception if an error occurs
     */
    @Test
    public void testAppendEntries() throws Exception
    {
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 1.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 2.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 3.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 4.0);
        
        SynchronizableRepeater values = SynchronizableRepeater.appendOrRemove(
                List.of(
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 5.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 6.0)),
                Set.of());
        
        SynchronizationResult result = _repeater.synchronizeValues(_repeaterViewItem, values, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(6, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(5), _repeater.getEntry(5).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(6), _repeater.getEntry(6).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link DefaultModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method
     * The repository contains some entries
     * Remove some entries and append new ones 
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveAndAppendEntries() throws Exception
    {
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 1.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 2.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 3.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 4.0);
        
        SynchronizableRepeater values = SynchronizableRepeater.appendOrRemove(
                List.of(
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 5.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 6.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 7.0)),
                Set.of(2, 4));
        
        SynchronizationResult result = _repeater.synchronizeValues(_repeaterViewItem, values, _context);

        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(5, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(5), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(6), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(7), _repeater.getEntry(5).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link DefaultModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method
     * The repository contains some entries
     * Try to remove a nonexistent repeater entry and append some other entries
     * This test ensure that we are not removing an entry that would only exist after the append of the given entries
     * @throws Exception if an error occurs
     */
    @Test
    public void testRemoveUnexistingEntryAndAppendEntries() throws Exception
    {
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 1.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 2.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 3.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 4.0);
        
        SynchronizableRepeater values = SynchronizableRepeater.appendOrRemove(
                List.of(
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 5.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 6.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 7.0)),
                Set.of(2, 6));
        
        try
        {
            _repeater.synchronizeValues(_repeaterViewItem, values, _context);
            Assert.fail("the synchronization shouldn't have be abel to remove the entry at position 6: there is no such entry before the given entries are appended");
        }
        catch (UnknownDataException e)
        {
            // Do nothing
        }
    }
    
    /**
     * Tests the {@link DefaultModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method
     * The repository is empty
     * Do not add entries in the repeater via the replace all mode
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyAddNoEntry() throws Exception
    {
        SynchronizableRepeater values = SynchronizableRepeater.replaceAll(
                List.of(),
                Map.of());
        
        SynchronizationResult result = _repeater.synchronizeValues(_repeaterViewItem, values, _context);
        
        Assert.assertFalse(result.hasChanged());
        Assert.assertEquals(0, _repeater.getSize());
    }
    
    /**
     * Tests the {@link DefaultModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method
     * The repository is empty
     * Add some entries in the repeater via the replace all mode
     * @throws Exception if an error occurs
     */
    @Test
    public void testRepoEmptyAddEntries() throws Exception
    {
        SynchronizableRepeater values = SynchronizableRepeater.replaceAll(
                List.of(
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 1.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 2.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 3.0)),
                Map.of());
        
        SynchronizationResult result = _repeater.synchronizeValues(_repeaterViewItem, values, _context);
        
        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(3, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link DefaultModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method
     * The repository contains some entries
     * In this test, entries are modified, moved, removed and added  
     * @throws Exception if an error occurs
     */
    @Test
    public void testReplaceAllEntries() throws Exception
    {
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 1.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 2.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 3.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 4.0);
        
        // Add values on entries to ensure that right entries are moved or removed
        _repeater.getEntry(1).setValue(__SINGLE_LONG_DATA_NAME, 1);
        _repeater.getEntry(2).setValue(__SINGLE_LONG_DATA_NAME, 2);
        _repeater.getEntry(3).setValue(__SINGLE_LONG_DATA_NAME, 3);
        _repeater.getEntry(4).setValue(__SINGLE_LONG_DATA_NAME, 4);
        
        SynchronizableRepeater values = SynchronizableRepeater.replaceAll(
                List.of(
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 2.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 5.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 4.5),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 1.5),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 6.0)),
                Map.of(1, 4,
                       2, 1,
                       4, 3));
        
        SynchronizationResult result = _repeater.synchronizeValues(_repeaterViewItem, values, _context);
        
        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(5, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Long.valueOf(2), _repeater.getEntry(1).getValue(__SINGLE_LONG_DATA_NAME));
        Assert.assertEquals(Double.valueOf(5), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4.5), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Long.valueOf(4), _repeater.getEntry(3).getValue(__SINGLE_LONG_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1.5), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Long.valueOf(1), _repeater.getEntry(4).getValue(__SINGLE_LONG_DATA_NAME));
        Assert.assertEquals(Double.valueOf(6), _repeater.getEntry(5).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link DefaultModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method
     * The repository contains some entries
     * Replace the entries by exactly the same entries  
     * @throws Exception if an error occurs
     */
    @Test
    public void testReplaceNoEntryWithReplaceAll() throws Exception
    {
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 1.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 2.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 3.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 4.0);
        
        SynchronizableRepeater values = SynchronizableRepeater.replaceAll(
                List.of(
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 1.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 2.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 3.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 4.0)),
                Map.of(1, 1,
                       2, 2,
                       3, 3,
                       4, 4));
        
        SynchronizationResult result = _repeater.synchronizeValues(_repeaterViewItem, values, _context);
        
        Assert.assertFalse(result.hasChanged());
        Assert.assertEquals(4, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link DefaultModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method
     * The repository contains some entries
     * The entries are not replaced but all are moved  
     * @throws Exception if an error occurs
     */
    @Test
    public void testMoveEntries() throws Exception
    {
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 1.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 2.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 3.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 4.0);
        
        SynchronizableRepeater values = SynchronizableRepeater.replaceAll(
                List.of(
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 2.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 1.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 4.0),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 3.0)),
                Map.of(1, 2,
                       2, 1,
                       3, 4,
                       4, 3));
        
        SynchronizationResult result = _repeater.synchronizeValues(_repeaterViewItem, values, _context);
        
        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(4, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link DefaultModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method
     * The repository contains some entries
     * Only 2 entries are replaced  
     * @throws Exception if an error occurs
     */
    @Test
    public void testReplaceSomeEntries() throws Exception
    {
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 1.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 2.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 3.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 4.0);
        
        SynchronizableRepeater values = SynchronizableRepeater.replace(
                List.of(
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 2.5),
                        Map.of(__SINGLE_DOUBLE_DATA_NAME, 4.5)),
                List.of(2, 4));
        
        SynchronizationResult result = _repeater.synchronizeValues(_repeaterViewItem, values, _context);
        
        Assert.assertTrue(result.hasChanged());
        Assert.assertEquals(4, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2.5), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4.5), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link DefaultModifiableModelAwareRepeater#synchronizeValues(ViewItemAccessor, SynchronizableRepeater, SynchronizationContext)} method
     * The repository contains some entries
     * Replace an entry by exactly the same one  
     * @throws Exception if an error occurs
     */
    @Test
    @Ignore
    public void testReplaceNoEntryWithReplace() throws Exception
    {
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 1.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 2.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 3.0);
        _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, 4.0);
        
        SynchronizableRepeater values = SynchronizableRepeater.replaceAll(
                List.of(Map.of(__SINGLE_DOUBLE_DATA_NAME, 2.0)),
                Map.of(2, 2));
        
        SynchronizationResult result = _repeater.synchronizeValues(_repeaterViewItem, values, _context);
        
        Assert.assertFalse(result.hasChanged());
        Assert.assertEquals(4, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
}
