/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.holder.group.impl;

import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import javax.jcr.RepositoryException;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.holder.impl.DefaultModifiableModelAwareDataHolder;
import org.ametys.cms.data.utils.ModelUsedByModelAwareDataHolderTests;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.holder.ModifiableModelAwareDataHolder;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeaterEntry;
import org.ametys.plugins.repository.data.holder.group.ModifiableRepeater;
import org.ametys.plugins.repository.data.holder.group.ModifiableRepeaterEntry;
import org.ametys.plugins.repository.data.holder.group.Repeater;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.JCRRepositoryData;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

import junit.framework.TestCase;

/**
 * {@link TestCase} for {@link ModifiableRepeater} implementation
 */
@NeedFeatures({
    "test-data/test-model",
    "repository/resolver",
    "cms/workflow.content"
})
public class ModifiableRepeaterTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __DATA_NODE_TYPE = "ametys:unstructured";
    private static final String __SINGLE_DOUBLE_DATA_NAME = "singleDoubleData";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static Model _model;
    
    private static JCRRepositoryData _root;
    private ModifiableModelAwareRepeater _repeater;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _model = (Model) manager.lookup(ModelUsedByModelAwareDataHolderTests.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        RootAmetysObject rootAmetysObject = _ametysObjectResolver.resolveByPath("/");
        _root = new JCRRepositoryData(rootAmetysObject.getNode());
    }
    
    /**
     * Prepare components that are specific for each test
     * @throws Exception if an error occurs
     */
    @Before
    public void prepareSpecificComponents()throws Exception
    {
        ModifiableRepositoryData repositoryData = _root.addRepositoryData(UUID.randomUUID().toString(), __DATA_NODE_TYPE);
        ModifiableModelAwareDataHolder dataHolder = new DefaultModifiableModelAwareDataHolder(repositoryData, _model);
        _repeater = dataHolder.getComposite("composite", true)
                .getRepeater("repeater", true);
    }

    /**
     * Tests the order of the retrieved repeater entries after some movements
     * @throws RepositoryException if an error occurs during the repository population
     */
    @Test
    public void testEntriesOrder() throws RepositoryException
    {
        _addSomeEntriesToRepeater(12);
        
        // Check the order of the retrieved entries
        List<? extends ModifiableModelAwareRepeaterEntry> entries = _repeater.getEntries();
        Assert.assertEquals(Double.valueOf(1), entries.get(0).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), entries.get(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), entries.get(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), entries.get(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(5), entries.get(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(6), entries.get(5).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(7), entries.get(6).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(8), entries.get(7).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(9), entries.get(8).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(10), entries.get(9).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(11), entries.get(10).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(12), entries.get(11).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Add an entry between 2 existent entries and Check the order of the retrieved entries
        _repeater.addEntry(11).setValue(__SINGLE_DOUBLE_DATA_NAME, 10.5D);
        entries = _repeater.getEntries();
        Assert.assertEquals(Double.valueOf(1), entries.get(0).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), entries.get(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), entries.get(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), entries.get(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(5), entries.get(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(6), entries.get(5).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(7), entries.get(6).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(8), entries.get(7).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(9), entries.get(8).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(10), entries.get(9).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(10.5), entries.get(10).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(11), entries.get(11).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(12), entries.get(12).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link Repeater#getEntry(int)} method
     * @throws RepositoryException if an error occurs during the repository population
     */
    @Test
    public void testGetEntry() throws RepositoryException
    {
        _addSomeEntriesToRepeater(4);
        
        Assert.assertNull(_repeater.getEntry(5));
        Assert.assertNull(_repeater.getEntry(-4));
        
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(-3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(0).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(-2).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link ModifiableRepeater#addEntry(int)} method
     * @throws RepositoryException if an error occurs during the repository population
     */
    @Test
    public void testAddEntry() throws RepositoryException
    {
        _addSomeEntriesToRepeater(4);
        
        try
        {
            // Try to create an entry at a too big position
            _repeater.addEntry(6);
            Assert.fail("Can not create a repeater entry at a position greater than the entries size");
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
        
        try
        {
            // Try to create an entry at a too low position
            _repeater.addEntry(-5);
            Assert.fail("Can not create a repeater entry at a position lower than the entries negative size");
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
        
        // Add an entry at first position via positive position
        _repeater.addEntry(1).setValue(__SINGLE_DOUBLE_DATA_NAME, 0D);
        Assert.assertEquals(5, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(0), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(5).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Add an entry at first position via negative position
        _repeater.addEntry(-5).setValue(__SINGLE_DOUBLE_DATA_NAME, -1D);
        Assert.assertEquals(6, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(-1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(0), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(5).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(6).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Add an entry at last position via positive position
        _repeater.addEntry(7).setValue(__SINGLE_DOUBLE_DATA_NAME, 5D);
        Assert.assertEquals(7, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(-1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(0), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(5).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(6).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(5), _repeater.getEntry(7).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Add an entry at last position via negative position
        _repeater.addEntry(0).setValue(__SINGLE_DOUBLE_DATA_NAME, 6D);
        Assert.assertEquals(8, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(-1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(0), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(5).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(6).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(5), _repeater.getEntry(7).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(6), _repeater.getEntry(8).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Add an entry at in the middle of the repeater via positive position
        _repeater.addEntry(5).setValue(__SINGLE_DOUBLE_DATA_NAME, 2.5D);
        Assert.assertEquals(9, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(-1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(0), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2.5), _repeater.getEntry(5).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(6).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(7).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(5), _repeater.getEntry(8).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(6), _repeater.getEntry(9).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Add an entry at in the middle of the repeater via negative position
        _repeater.addEntry(-3).setValue(__SINGLE_DOUBLE_DATA_NAME, 3.5D);
        Assert.assertEquals(10, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(-1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(0), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2.5), _repeater.getEntry(5).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(6).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3.5), _repeater.getEntry(7).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(8).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(5), _repeater.getEntry(9).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(6), _repeater.getEntry(10).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link ModifiableRepeaterEntry#setPosition(int)} method
     * @throws RepositoryException if an error occurs during the repository population
     */
    @Test
    public void testSetEntryPosition() throws RepositoryException
    {
        _addSomeEntriesToRepeater(4);
        
        try
        {
            // Try to move the position of an entry at a too big position
            _repeater.getEntry(1).setPosition(5);
            Assert.fail("Can not move a repeater entry at a position greater than the entries size");
            
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
        
        try
        {
            // Try to move the position of an entry at a too low position
            _repeater.getEntry(1).setPosition(-4);
            Assert.fail("Can not move a repeater entry at a position lower than the entries negative size");
            
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
        
        // Move an entry at first position via positive position
        _repeater.getEntry(4).setPosition(1);
        Assert.assertEquals(4, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Move an entry at first position via negative position
        _repeater.getEntry(3).setPosition(-3);
        Assert.assertEquals(4, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Move an entry at last position via positive position
        _repeater.getEntry(2).setPosition(4);
        Assert.assertEquals(4, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Move an entry at last position via negative position
        _repeater.getEntry(1).setPosition(0);
        Assert.assertEquals(4, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Move down an entry in the middle of the repeater via positive position
        _repeater.getEntry(1).setPosition(3);
        Assert.assertEquals(4, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Move down an entry in the middle of the repeater via negative position
        _repeater.getEntry(2).setPosition(-1);
        Assert.assertEquals(4, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Move up an entry in the middle of the repeater via positive position
        _repeater.getEntry(4).setPosition(2);
        Assert.assertEquals(4, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Move up an entry in the middle of the repeater via negative position
        _repeater.getEntry(3).setPosition(-2);
        Assert.assertEquals(4, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link ModifiableRepeater#removeEntries(Set)} method
     * @throws RepositoryException if an error occurs during the repository population
     */
    @Test
    public void testRemoveEntries() throws RepositoryException
    {
        _addSomeEntriesToRepeater(8);
        
        try
        {
            // Try to remove an entry at a too big position => an UnknownDataException has to be thrown
            _repeater.removeEntries(Set.of(9));
            Assert.fail("There is no repeater entry at the given position, it should not be possible to remove it");
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains("no entry"));
        }
        
        try
        {
            // Try to remove an entry at a too low position => an UnknownDataException has to be thrown
            _repeater.removeEntries(Set.of(-8));
            Assert.fail("There is no repeater entry at the given position, it should not be possible to remove it");
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains("no entry"));
        }
        
        // Remove some entries
        _repeater.removeEntries(Set.of(3, -2, 0, 5, 1, -6));
        Assert.assertEquals(2, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(7), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link ModifiableRepeater#removeEntry(int)} method
     * @throws RepositoryException if an error occurs during the repository population
     */
    @Test
    public void testRemoveEntry() throws RepositoryException
    {
        _addSomeEntriesToRepeater(8);
        
        try
        {
            // Try to remove an entry at a too big position => an UnknownDataException has to be thrown
            _repeater.removeEntry(9);
            Assert.fail("There is no repeater entry at the given position, it should not be possible to remove it");
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains("no entry"));
        }
        
        try
        {
            // Try to remove an entry at a too low position => an UnknownDataException has to be thrown
            _repeater.removeEntry(-8);
            Assert.fail("There is no repeater entry at the given position, it should not be possible to remove it");
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains("no entry"));
        }
    
        // Remove the first entry via positive position
        _repeater.removeEntry(1);
        Assert.assertEquals(7, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(5), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(6), _repeater.getEntry(5).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(7), _repeater.getEntry(6).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(8), _repeater.getEntry(7).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Remove the first entry via negative position
        _repeater.removeEntry(-6);
        Assert.assertEquals(6, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(5), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(6), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(7), _repeater.getEntry(5).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(8), _repeater.getEntry(6).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Remove the last entry via positive position
        _repeater.removeEntry(6);
        Assert.assertEquals(5, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(5), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(6), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(7), _repeater.getEntry(5).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Remove the last entry via negative position
        _repeater.removeEntry(0);
        Assert.assertEquals(4, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(4), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(5), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(6), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Remove an entry in the middle of the repeater via positive position
        _repeater.removeEntry(2);
        Assert.assertEquals(3, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(5), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(6), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        
        // Remove an entry in the middle of the repeater via negative position
        _repeater.removeEntry(-1);
        Assert.assertEquals(2, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(6), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link ModifiableRepeater#moveEntries(Map, int)} method
     * @throws RepositoryException if an error occurs during the repository population
     */
    @Test
    public void testMoveEntries() throws RepositoryException
    {
        _addSomeEntriesToRepeater(8);
        
        Map<Integer, Integer> positionMapping = Map.of(1, 2,
                                                       2, 1,
                                                       3, 4,
                                                       7, 3,
                                                       8, 6);
        
        boolean hasChanged = _repeater.moveEntries(positionMapping, 7);
        
        Assert.assertTrue(hasChanged);
        Assert.assertEquals(7, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(7), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(4).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertTrue(_repeater.hasEntry(5));
        Assert.assertFalse(_repeater.getEntry(5).hasValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(8), _repeater.getEntry(6).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertTrue(_repeater.hasEntry(7));
        Assert.assertFalse(_repeater.getEntry(7).hasValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    /**
     * Tests the {@link ModifiableRepeater#moveEntries(Map, int)} method
     * No entry should move, been removed or added
     * @throws RepositoryException if an error occurs during the repository population
     */
    @Test
    public void testDoNotMoveEntries() throws RepositoryException
    {
        _addSomeEntriesToRepeater(3);
        
        Map<Integer, Integer> positionMapping = Map.of(1, 1,
                                                       2, 2,
                                                       3, 3);
        
        boolean hasChanged = _repeater.moveEntries(positionMapping, 3);
        
        Assert.assertFalse(hasChanged);
        Assert.assertEquals(3, _repeater.getSize());
        Assert.assertEquals(Double.valueOf(1), _repeater.getEntry(1).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(2), _repeater.getEntry(2).getValue(__SINGLE_DOUBLE_DATA_NAME));
        Assert.assertEquals(Double.valueOf(3), _repeater.getEntry(3).getValue(__SINGLE_DOUBLE_DATA_NAME));
    }
    
    private void _addSomeEntriesToRepeater(int size)
    {
        for (int i = 1; i <= size; i++)
        {
            _repeater.addEntry().setValue(__SINGLE_DOUBLE_DATA_NAME, Double.valueOf(i));
        }
    }
}
