/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.time.LocalDate;
import java.time.Month;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.Calendar;
import java.util.Optional;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.TransformerHandler;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.XMLUtils;
import org.junit.Assert;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.ContentHandler;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link DateRepositoryElementType} implementation test case
 */
@NeedFeatures({
    "test-data/test-data-types",
    "cms/workflow.content"
})
public class DateRepositoryElementTypeTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    private static ModifiableRepositoryData _repositoryData;
    private static RepositoryElementType<LocalDate> _dateType;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public void prepareComponents() throws Exception
    {
        _dateType = (RepositoryElementType<LocalDate>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.DATE_TYPE_ID);
        _repositoryData = new MemoryRepositoryData("dateContainer");
    }

    /**
     * Tests the {@link DateRepositoryElementType#isCompatible(RepositoryData, String)} behavior with a simple date data
     */
    @Test
    public void testIsCompatibleWithSimpleDateData()
    {
        _repositoryData.setValue("simpleDateData", DateUtils.asCalendar(LocalDate.of(2019, 2, 6)));
        Assert.assertTrue(_dateType.isCompatible(_repositoryData, "simpleDateData"));
    }

    /**
     * Tests the {@link DateRepositoryElementType#isCompatible(RepositoryData, String)} behavior with a multiple date data
     */
    @Test
    public void testIsCompatibleWithMultipleDateData()
    {
        _repositoryData.setValues("multipleDateData", new Calendar[] {DateUtils.asCalendar(LocalDate.of(2015, 12, 6)), DateUtils.asCalendar(LocalDate.of(2018, 9, 11))});
        Assert.assertTrue(_dateType.isCompatible(_repositoryData, "multipleDateData"));
    }

    /**
     * Tests the {@link DateRepositoryElementType#isCompatible(RepositoryData, String)} behavior with a simple non date data
     */
    @Test
    public void testIsCompatibleWithSimpleNonDateData()
    {
        _repositoryData.setValue("simpleStringData", "string value");
        Assert.assertFalse(_dateType.isCompatible(_repositoryData, "simpleStringData"));
    }

    /**
     * Tests the {@link DateRepositoryElementType#isCompatible(RepositoryData, String)} behavior with a simple date time data
     */
    @Test
    public void testIsCompatibleWithSimpleDatetimeData()
    {
        _repositoryData.setValue("simpleDatetimeData", DateUtils.asCalendar(ZonedDateTime.of(2019, 2, 6, 14, 2, 0, 0, ZoneOffset.UTC)));
        Assert.assertFalse(_dateType.isCompatible(_repositoryData, "simpleDatetimeData"));
    }

    /**
     * Tests the {@link DateRepositoryElementType#isCompatible(RepositoryData, String)} behavior with a multiple date time data
     */
    @Test
    public void testIsCompatibleWithMultipleDatetimeData()
    {
        _repositoryData.setValues("multipleDatetimeData", new Calendar[] {DateUtils.asCalendar(ZonedDateTime.of(2015, 12, 6, 15, 18, 0, 0, ZoneOffset.UTC)), DateUtils.asCalendar(ZonedDateTime.of(2018, 9, 11, 12, 22, 0, 0, ZoneOffset.UTC))});
        Assert.assertFalse(_dateType.isCompatible(_repositoryData, "multipleDatetimeData"));
    }

    /**
     * Tests the {@link DateRepositoryElementType#isCompatible(RepositoryData, String)} behavior with an inexisting data
     */
    @Test
    public void testIsCompatibleWithInexistingData()
    {
        try
        {
            _dateType.isCompatible(_repositoryData, "inexistingData");
            Assert.fail("Call to isInstanceof method with an inexsting data name should have thrown an UnknownDataException");
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains("found"));
        }
    }
    
    /**
     * Tests the {@link DateRepositoryElementType#isCompatible(RepositoryData, String)} behavior with an old date data
     * This test shows that the gap existing on old dates does not make the data incompatible with the date type
     */
    @Test
    public void testIsCompatibleWithOldDateData()
    {
        _dateType.write(_repositoryData, "oldDateData", LocalDate.of(1890, Month.FEBRUARY, 25));
        Assert.assertTrue(_dateType.isCompatible(_repositoryData, "oldDateData"));
    }
    
    /**
     * Test the reciprocity between the methods {@link DateRepositoryElementType#valueToSAX(ContentHandler, String, Object, DataContext)} and {@link DateRepositoryElementType#valueFromXML(Element, String, Optional)}
     * @throws Exception if an error occurs
     */
    @Test
    public void testReciprocityBetweenToAndFromXML() throws Exception
    {
        LocalDate localDate = LocalDate.of(2020, Month.JUNE, 26);
        
        // Generate SAX events for the date
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        _dateType.valueToSAX(handler, "localDate", localDate, DataContext.newInstance());

        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        Document document = (Document) result.getNode();
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        LocalDate resultDate = (LocalDate) _dateType.valueFromXML(document.getDocumentElement(), "localDate", Optional.empty());
        
        Assert.assertTrue(resultDate.isEqual(localDate));
    }
    
    /**
     * Test the reciprocity between the methods {@link DateRepositoryElementType#toString(LocalDate)} and {@link DateRepositoryElementType#castValue(Object)}
     */
    @Test
    public void testReciprocityBetweenToStringAndCast()
    {
        LocalDate localDate = LocalDate.of(2020, Month.JUNE, 26);
        Assert.assertTrue(localDate.isEqual(_dateType.castValue(_dateType.toString(localDate))));
    }
}
