/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import java.time.LocalDate;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Optional;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.TransformerHandler;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.XMLUtils;
import org.junit.Assert;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.ContentHandler;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.plugins.repository.data.type.RepositoryElementType;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link DateTimeRepositoryElementType} implementation test case
 */
@NeedFeatures({
    "test-data/test-data-types",
    "cms/workflow.content"
})
public class DateTimeRepositoryElementTypeTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    private static RepositoryElementType<ZonedDateTime> _dateType;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public void prepareComponents() throws Exception
    {
        _dateType = (RepositoryElementType<ZonedDateTime>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.DATETIME_TYPE_ID);
    }
    
    /**
     * Test the reciprocity between the methods {@link DateRepositoryElementType#valueToSAX(ContentHandler, String, Object, DataContext)} and {@link DateRepositoryElementType#valueFromXML(Element, String, Optional)}
     * @throws Exception if an error occurs
     */
    @Test
    public void testReciprocityBetweenToAndFromXML() throws Exception
    {
        ZonedDateTime zonedDateTime = ZonedDateTime.of(2020, 06, 26, 14, 10, 45, 596_000_000, ZoneId.systemDefault());
        
        // Generate SAX events for the date
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        _dateType.valueToSAX(handler, "zonedDateTime", zonedDateTime, DataContext.newInstance());

        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        Document document = (Document) result.getNode();
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        ZonedDateTime resultDateTime = (ZonedDateTime) _dateType.valueFromXML(document.getDocumentElement(), "zonedDateTime", Optional.empty());
        
        Assert.assertTrue(resultDateTime.isEqual(zonedDateTime));
    }
    
    /**
     * Test the reciprocity between the methods {@link DateRepositoryElementType#toString(LocalDate)} and {@link DateRepositoryElementType#castValue(Object)}
     */
    @Test
    public void testReciprocityBetweenToStringAndCast()
    {
        ZonedDateTime zonedDateTime = ZonedDateTime.of(2020, 06, 26, 14, 10, 45, 596_000_000, ZoneId.systemDefault());
        Assert.assertTrue(zonedDateTime.isEqual(_dateType.castValue(_dateType.toString(zonedDateTime))));
    }

}
