/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.type.impl;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.plugins.repository.RepositoryConstants;
import org.ametys.plugins.repository.data.UnknownDataException;
import org.ametys.plugins.repository.data.repositorydata.ModifiableRepositoryData;
import org.ametys.plugins.repository.data.repositorydata.RepositoryData;
import org.ametys.plugins.repository.data.repositorydata.impl.MemoryRepositoryData;
import org.ametys.plugins.repository.data.type.ModelItemTypeConstants;
import org.ametys.plugins.repository.data.type.RepositoryModelItemType;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link RepeaterRepositoryModelItemType} implementation test case
 */
@NeedFeatures({
    "test-data/test-data-types",
    "cms/workflow.content"
})
public class RepeaterRepositoryModelItemTypeTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    private static ModifiableRepositoryData _repositoryData;
    private static RepositoryModelItemType _repeaterType;

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        _repeaterType = (RepositoryModelItemType) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.REPEATER_TYPE_ID);
        _repositoryData = new MemoryRepositoryData("dateContainer");
    }

    /**
     * Tests the {@link RepeaterRepositoryModelItemType#isCompatible(RepositoryData, String)} behavior with non existing data
     */
    @Test
    public void testIsCompatibleWithNonExisingData()
    {
        try
        {
            Assert.assertTrue(_repeaterType.isCompatible(_repositoryData, "nonExistingRepeater"));
            Assert.fail("The isCompatible method should fail if there is no data with the given name");
        }
        catch (UnknownDataException e)
        {
            Assert.assertTrue(e.getMessage().contains("found"));
        }
    }

    /**
     * Tests the {@link RepeaterRepositoryModelItemType#isCompatible(RepositoryData, String)} behavior with data using repeater node type
     */
    @Test
    public void testIsCompatibleWithRepeaterNodeType()
    {
        _repositoryData.addRepositoryData("repeaterWithRepeaterNodeType", RepositoryConstants.REPEATER_NODETYPE);
        Assert.assertTrue(_repeaterType.isCompatible(_repositoryData, "repeaterWithRepeaterNodeType"));
    }

    /**
     * Tests the {@link RepeaterRepositoryModelItemType#isCompatible(RepositoryData, String)} behavior with data using compositeMetadata node type with an empty repeater (no entry)
     */
    @Test
    public void testIsCompatibleWithCompositeMetadataNodeTypeEmptyNode()
    {
        _repositoryData.addRepositoryData("repeaterWithEmptyNode", RepositoryConstants.COMPOSITE_METADTA_NODETYPE);
        Assert.assertTrue(_repeaterType.isCompatible(_repositoryData, "repeaterWithEmptyNode"));
    }

    /**
     * Tests the {@link RepeaterRepositoryModelItemType#isCompatible(RepositoryData, String)} behavior with data using compositeMetadata node type with some entries
     */
    @Test
    public void testIsCompatibleWithCompositeMetadataNodeTypeSomeEntries()
    {
        ModifiableRepositoryData repeaterData = _repositoryData.addRepositoryData("repeaterWithSomeEntries", RepositoryConstants.COMPOSITE_METADTA_NODETYPE);
        repeaterData.addRepositoryData("1", RepositoryConstants.COMPOSITE_METADTA_NODETYPE);
        repeaterData.addRepositoryData("2", RepositoryConstants.COMPOSITE_METADTA_NODETYPE);
        repeaterData.addRepositoryData("3", RepositoryConstants.COMPOSITE_METADTA_NODETYPE);
        
        Assert.assertTrue(_repeaterType.isCompatible(_repositoryData, "repeaterWithSomeEntries"));
    }

    /**
     * Tests the {@link RepeaterRepositoryModelItemType#isCompatible(RepositoryData, String)} behavior with data using compositeMetadata node type with some nodes
     */
    @Test
    public void testIsCompatibleWithCompositeMetadataNodeTypeSomeNodes()
    {
        ModifiableRepositoryData repeaterData = _repositoryData.addRepositoryData("repeaterWithSomeNodes", RepositoryConstants.COMPOSITE_METADTA_NODETYPE);
        repeaterData.addRepositoryData("1", RepositoryConstants.COMPOSITE_METADTA_NODETYPE);
        repeaterData.addRepositoryData("2", RepositoryConstants.COMPOSITE_METADTA_NODETYPE);
        repeaterData.addRepositoryData("3", RepositoryConstants.COMPOSITE_METADTA_NODETYPE);
        repeaterData.addRepositoryData("Text", RepositoryConstants.COMPOSITE_METADTA_NODETYPE);
        
        Assert.assertFalse(_repeaterType.isCompatible(_repositoryData, "repeaterWithSomeNodes"));
    }

    /**
     * Tests the {@link RepeaterRepositoryModelItemType#isCompatible(RepositoryData, String)} behavior with data using compositeMetadata node type with some properties
     */
    @Test
    public void testIsCompatibleWithCompositeMetadataNodeTypeSomeProperties()
    {
        ModifiableRepositoryData repeaterData = _repositoryData.addRepositoryData("repeaterWithSomeProperties", RepositoryConstants.COMPOSITE_METADTA_NODETYPE);
        repeaterData.addRepositoryData("1", RepositoryConstants.COMPOSITE_METADTA_NODETYPE);
        repeaterData.addRepositoryData("2", RepositoryConstants.COMPOSITE_METADTA_NODETYPE);
        repeaterData.setValue("property1", "value 1");
        repeaterData.setValue("property2", "value 2");
        
        Assert.assertFalse(_repeaterType.isCompatible(_repositoryData, "repeaterWithSomeProperties"));
    }
}
