/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.data.utils;

import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Map;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.cms.contenttype.ContentAttributeDefinition;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.contenttype.ContentTypesHelper;
import org.ametys.cms.data.Binary;
import org.ametys.cms.data.File;
import org.ametys.cms.data.Geocode;
import org.ametys.cms.data.Reference;
import org.ametys.cms.data.RichText;
import org.ametys.cms.data.type.ModelItemTypeConstants;
import org.ametys.cms.data.type.ModelItemTypeExtensionPoint;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.data.holder.ModelAwareDataHolder;
import org.ametys.plugins.repository.metadata.MultilingualString;
import org.ametys.plugins.repository.model.CompositeDefinition;
import org.ametys.plugins.repository.model.RepeaterDefinition;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.DefaultElementDefinition;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Enumerator;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.ModelItem;

/**
 * The model used by the {@link ModelAwareDataHolder} test cases
 */
public class ModelUsedByModelAwareDataHolderTests implements Model, Component, Serviceable, Initializable
{
    /** Avalon Role */
    public static final String ROLE = ModelUsedByModelAwareDataHolderTests.class.getName();
    
    private static final String __TEST_CONTENT_TYPE_ID = "org.ametys.cms.contenttype.Test";
    private static final String __EXHAUSTIVE_CONTENT_TYPE_ID = "org.ametys.test.Content.exhaustive";
    
    private static final Map<String, I18nizableText> __ENUMERATED_STRING_VALUES = Map.of("value1", new I18nizableText("Value 1"),
            "value2", new I18nizableText("Value 2"),
            "value3", new I18nizableText("Value 3"));
    private static final Enumerator<String> __ENUMERATED_STRING_ENUMERATOR = new Enumerator<>()
    {
        public I18nizableText getEntry(String value) throws Exception
        {
            return __ENUMERATED_STRING_VALUES.get(value);
        }

        public Map<String, I18nizableText> getTypedEntries() throws Exception
        {
            return __ENUMERATED_STRING_VALUES;
        }
    };
    
    private static final Map<Long, I18nizableText> __ENUMERATED_LONG_VALUES = Map.of(1L, new I18nizableText("Value 1"),
            2L, new I18nizableText("Value 2"),
            3L, new I18nizableText("Value 3"));
    private static final Enumerator<Long> __ENUMERATED_LONG_ENUMERATOR = new Enumerator<>()
    {
        public I18nizableText getEntry(Long value) throws Exception
        {
            return __ENUMERATED_LONG_VALUES.get(value);
        }

        public Map<Long, I18nizableText> getTypedEntries() throws Exception
        {
            return __ENUMERATED_LONG_VALUES;
        }
    };
    
    private ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    private ContentTypesHelper _contentTypesHelper;
    private ContentTypeExtensionPoint _contentTypeExtensionPoint;
    
    private Collection<ModelItem> _items;

    public void service(ServiceManager manager) throws ServiceException
    {
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
        _contentTypesHelper = (ContentTypesHelper) manager.lookup(ContentTypesHelper.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    @SuppressWarnings("static-access")
    public void initialize() throws Exception
    {
        _items = new ArrayList<>();
        
        // single string data
        ElementDefinition<String> singleStringData = new DefaultElementDefinition<>();
        singleStringData.setName("singleStringData");
        singleStringData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.STRING_TYPE_ID));
        singleStringData.setLabel(new I18nizableText("Single string data"));
        singleStringData.setDescription(new I18nizableText("Single string data"));
        _items.add(singleStringData);
        
        // single string data with default value
        ElementDefinition<String> singleStringDataWithDefaultValue = new DefaultElementDefinition<>();
        singleStringDataWithDefaultValue.setName("stringWithDefaultValue");
        singleStringDataWithDefaultValue.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.STRING_TYPE_ID));
        singleStringDataWithDefaultValue.setLabel(new I18nizableText("Single string data with default value"));
        singleStringDataWithDefaultValue.setDescription(new I18nizableText("Single string data with default value"));
        singleStringDataWithDefaultValue.setDefaultValue("the default value from model");
        _items.add(singleStringDataWithDefaultValue);
        
        // single enumerated string data
        ElementDefinition<String> singleEnumeratedStringData = new DefaultElementDefinition<>();
        singleEnumeratedStringData.setName("singleEnumeratedStringData");
        singleEnumeratedStringData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.STRING_TYPE_ID));
        singleEnumeratedStringData.setLabel(new I18nizableText("Single string data with an enumerator"));
        singleEnumeratedStringData.setDescription(new I18nizableText("Single string data with an enumerator"));
        singleEnumeratedStringData.setEnumerator(__ENUMERATED_STRING_ENUMERATOR);
        _items.add(singleEnumeratedStringData);
        
        // multiple string data
        ElementDefinition<String> multipleStringData = new DefaultElementDefinition<>();
        multipleStringData.setName("multipleStringData");
        multipleStringData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.STRING_TYPE_ID));
        multipleStringData.setLabel(new I18nizableText("Multiple string data"));
        multipleStringData.setDescription(new I18nizableText("Multiple string data"));
        multipleStringData.setMultiple(true);
        _items.add(multipleStringData);

        // single long data
        ElementDefinition<Long> singleLongData = new DefaultElementDefinition<>();
        singleLongData.setName("singleLongData");
        singleLongData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.LONG_TYPE_ID));
        singleLongData.setLabel(new I18nizableText("Single long data"));
        singleLongData.setDescription(new I18nizableText("Single long data"));
        _items.add(singleLongData);
        
        // single enumerated long data
        ElementDefinition<Long> singleEnumeratedLongData = new DefaultElementDefinition<>();
        singleEnumeratedLongData.setName("singleEnumeratedLongData");
        singleEnumeratedLongData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.LONG_TYPE_ID));
        singleEnumeratedLongData.setLabel(new I18nizableText("Single long data with an enumerator"));
        singleEnumeratedLongData.setDescription(new I18nizableText("Single long data with an enumerator"));
        singleEnumeratedLongData.setEnumerator(__ENUMERATED_LONG_ENUMERATOR);
        _items.add(singleEnumeratedLongData);
        
        // multiple long data
        ElementDefinition<Long> multipleLongData = new DefaultElementDefinition<>();
        multipleLongData.setName("multipleLongData");
        multipleLongData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.LONG_TYPE_ID));
        multipleLongData.setLabel(new I18nizableText("Multiple long data"));
        multipleLongData.setDescription(new I18nizableText("Multiple long data"));
        multipleLongData.setMultiple(true);
        _items.add(multipleLongData);
        
        // single double data
        ElementDefinition<Double> singleDoubleData = new DefaultElementDefinition<>();
        singleDoubleData.setName("singleDoubleData");
        singleDoubleData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.DOUBLE_TYPE_ID));
        singleDoubleData.setLabel(new I18nizableText("Single double data"));
        singleDoubleData.setDescription(new I18nizableText("Single double data"));
        _items.add(singleDoubleData);
        
        // multiple double data
        ElementDefinition<Double> multipleDoubleData = new DefaultElementDefinition<>();
        multipleDoubleData.setName("multipleDoubleData");
        multipleDoubleData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.DOUBLE_TYPE_ID));
        multipleDoubleData.setLabel(new I18nizableText("Multiple double data"));
        multipleDoubleData.setDescription(new I18nizableText("Multiple double data"));
        multipleDoubleData.setMultiple(true);
        _items.add(multipleDoubleData);
        
        // single boolean data
        ElementDefinition<Boolean> singleBooleanData = new DefaultElementDefinition<>();
        singleBooleanData.setName("singleBooleanData");
        singleBooleanData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.BOOLEAN_TYPE_ID));
        singleBooleanData.setLabel(new I18nizableText("Single boolean data"));
        singleBooleanData.setDescription(new I18nizableText("Single boolean data"));
        _items.add(singleBooleanData);
        
        // multiple boolean data
        ElementDefinition<Boolean> multipleBooleanData = new DefaultElementDefinition<>();
        multipleBooleanData.setName("multipleBooleanData");
        multipleBooleanData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.BOOLEAN_TYPE_ID));
        multipleBooleanData.setLabel(new I18nizableText("Multiple boolean data"));
        multipleBooleanData.setDescription(new I18nizableText("Multiple boolean data"));
        multipleBooleanData.setMultiple(true);
        _items.add(multipleBooleanData);
        
        // single date data
        ElementDefinition<LocalDate> singleDateData = new DefaultElementDefinition<>();
        singleDateData.setName("singleDateData");
        singleDateData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.DATE_TYPE_ID));
        singleDateData.setLabel(new I18nizableText("Single date data"));
        singleDateData.setDescription(new I18nizableText("Single date data"));
        _items.add(singleDateData);
        
        // multiple date data
        ElementDefinition<LocalDate> multipleDateData = new DefaultElementDefinition<>();
        multipleDateData.setName("multipleDateData");
        multipleDateData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.DATE_TYPE_ID));
        multipleDateData.setLabel(new I18nizableText("Multiple date data"));
        multipleDateData.setDescription(new I18nizableText("Multiple date data"));
        multipleDateData.setMultiple(true);
        _items.add(multipleDateData);
        
        // single datetime data
        ElementDefinition<ZonedDateTime> singleDatetimeData = new DefaultElementDefinition<>();
        singleDatetimeData.setName("singleDatetimeData");
        singleDatetimeData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.DATETIME_TYPE_ID));
        singleDatetimeData.setLabel(new I18nizableText("Single datetime data"));
        singleDatetimeData.setDescription(new I18nizableText("Single datetime data"));
        _items.add(singleDatetimeData);
        
        // multiple datetime data
        ElementDefinition<ZonedDateTime> multipleDatetimeData = new DefaultElementDefinition<>();
        multipleDatetimeData.setName("multipleDatetimeData");
        multipleDatetimeData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.DATETIME_TYPE_ID));
        multipleDatetimeData.setLabel(new I18nizableText("Multiple datetime data"));
        multipleDatetimeData.setDescription(new I18nizableText("Multiple datetime data"));
        multipleDatetimeData.setMultiple(true);
        _items.add(multipleDatetimeData);
        
        // single user data
        ElementDefinition<UserIdentity> singleUserData = new DefaultElementDefinition<>();
        singleUserData.setName("singleUserData");
        singleUserData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID));
        singleUserData.setLabel(new I18nizableText("Single user data"));
        singleUserData.setDescription(new I18nizableText("Single user data"));
        _items.add(singleUserData);
        
        // multiple user data
        ElementDefinition<UserIdentity> multipleUserData = new DefaultElementDefinition<>();
        multipleUserData.setName("multipleUserData");
        multipleUserData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.USER_ELEMENT_TYPE_ID));
        multipleUserData.setLabel(new I18nizableText("Multiple user data"));
        multipleUserData.setDescription(new I18nizableText("Multiple user data"));
        multipleUserData.setMultiple(true);
        _items.add(multipleUserData);
        
        // single geocode data
        ElementDefinition<Geocode> singleGeocodeData = new DefaultElementDefinition<>();
        singleGeocodeData.setName("singleGeocodeData");
        singleGeocodeData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID));
        singleGeocodeData.setLabel(new I18nizableText("Single geocode data"));
        singleGeocodeData.setDescription(new I18nizableText("Single geocode data"));
        _items.add(singleGeocodeData);
        
        // multiple geocode data
        ElementDefinition<Geocode> multipleGeocodeData = new DefaultElementDefinition<>();
        multipleGeocodeData.setName("multipleGeocodeData");
        multipleGeocodeData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.GEOCODE_ELEMENT_TYPE_ID));
        multipleGeocodeData.setLabel(new I18nizableText("Multiple geocode data"));
        multipleGeocodeData.setDescription(new I18nizableText("Multiple geocode data"));
        multipleGeocodeData.setMultiple(true);
        _items.add(multipleGeocodeData);
        
        // single multilingual string data
        ElementDefinition<MultilingualString> singleMultilingualStringData = new DefaultElementDefinition<>();
        singleMultilingualStringData.setName("singleMultilingualStringData");
        singleMultilingualStringData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID));
        singleMultilingualStringData.setLabel(new I18nizableText("Single multilingual string data"));
        singleMultilingualStringData.setDescription(new I18nizableText("Single multilingual string data"));
        _items.add(singleMultilingualStringData);
        
        // multiple multilingual string data
        ElementDefinition<MultilingualString> multipleMultilingualStringData = new DefaultElementDefinition<>();
        multipleMultilingualStringData.setName("multipleMultilingualStringData");
        multipleMultilingualStringData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.MULTILINGUAL_STRING_ELEMENT_TYPE_ID));
        multipleMultilingualStringData.setLabel(new I18nizableText("Multiple multilingual string data"));
        multipleMultilingualStringData.setDescription(new I18nizableText("Multiple multilingual string data"));
        multipleMultilingualStringData.setMultiple(true);
        _items.add(multipleMultilingualStringData);
        
        // single reference data
        ElementDefinition<Reference> singleReferenceData = new DefaultElementDefinition<>();
        singleReferenceData.setName("singleReferenceData");
        singleReferenceData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.REFERENCE_ELEMENT_TYPE_ID));
        singleReferenceData.setLabel(new I18nizableText("Single reference data"));
        singleReferenceData.setDescription(new I18nizableText("Single reference data"));
        _items.add(singleReferenceData);
        
        // multiple reference data
        ElementDefinition<Reference> multipleReferenceData = new DefaultElementDefinition<>();
        multipleReferenceData.setName("multipleReferenceData");
        multipleReferenceData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.REFERENCE_ELEMENT_TYPE_ID));
        multipleReferenceData.setLabel(new I18nizableText("Multiple reference data"));
        multipleReferenceData.setDescription(new I18nizableText("Multiple reference data"));
        multipleReferenceData.setMultiple(true);
        _items.add(multipleReferenceData);
        
        // single binary data
        ElementDefinition<Binary> singleBinaryData = new DefaultElementDefinition<>();
        singleBinaryData.setName("singleBinaryData");
        singleBinaryData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID));
        singleBinaryData.setLabel(new I18nizableText("Single binary data"));
        singleBinaryData.setDescription(new I18nizableText("Single binary data"));
        _items.add(singleBinaryData);
        
        // multiple binary data
        ElementDefinition<Binary> multipleBinaryData = new DefaultElementDefinition<>();
        multipleBinaryData.setName("multipleBinaryData");
        multipleBinaryData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID));
        multipleBinaryData.setLabel(new I18nizableText("Multiple binary data"));
        multipleBinaryData.setDescription(new I18nizableText("Multiple binary data"));
        multipleBinaryData.setMultiple(true);
        _items.add(multipleBinaryData);
        
        // single file data
        ElementDefinition<File> singleFileData = new DefaultElementDefinition<>();
        singleFileData.setName("singleFileData");
        singleFileData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.FILE_ELEMENT_TYPE_ID));
        singleFileData.setLabel(new I18nizableText("Single file data"));
        singleFileData.setDescription(new I18nizableText("Single file data"));
        _items.add(singleFileData);
        
        // multiple file data
        ElementDefinition<File> multipleFileData = new DefaultElementDefinition<>();
        multipleFileData.setName("multipleFileData");
        multipleFileData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.FILE_ELEMENT_TYPE_ID));
        multipleFileData.setLabel(new I18nizableText("Multiple file data"));
        multipleFileData.setDescription(new I18nizableText("Multiple file data"));
        multipleFileData.setMultiple(true);
        _items.add(multipleFileData);
        
        // single rich text data
        ElementDefinition<RichText> singleRichTextData = new DefaultElementDefinition<>();
        singleRichTextData.setName("singleRichTextData");
        singleRichTextData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID));
        singleRichTextData.setLabel(new I18nizableText("Single rich text data"));
        singleRichTextData.setDescription(new I18nizableText("Single rich text data"));
        _items.add(singleRichTextData);
        
        // multiple rich text data
        ElementDefinition<RichText> multipleRichTextData = new DefaultElementDefinition<>();
        multipleRichTextData.setName("multipleRichTextData");
        multipleRichTextData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.RICH_TEXT_ELEMENT_TYPE_ID));
        multipleRichTextData.setLabel(new I18nizableText("Multiple rich text data"));
        multipleRichTextData.setDescription(new I18nizableText("Multiple rich text data"));
        multipleRichTextData.setMultiple(true);
        _items.add(multipleRichTextData);
        
        // single test content data
        ContentAttributeDefinition singleContentData = new ContentAttributeDefinition(_contentTypeExtensionPoint, _contentTypesHelper);
        singleContentData.setName("singleContentData");
        singleContentData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID));
        singleContentData.setContentTypeId(__TEST_CONTENT_TYPE_ID);
        singleContentData.setLabel(new I18nizableText("Single content data"));
        singleContentData.setDescription(new I18nizableText("Single content data"));
        singleContentData.setMultiple(false);
        _items.add(singleContentData);
        
        // multiple test content data
        ContentAttributeDefinition multipleTestContentData = new ContentAttributeDefinition(_contentTypeExtensionPoint, _contentTypesHelper);
        multipleTestContentData.setName("multipleContentData");
        multipleTestContentData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID));
        multipleTestContentData.setContentTypeId(__TEST_CONTENT_TYPE_ID);
        multipleTestContentData.setLabel(new I18nizableText("Multiple content data"));
        multipleTestContentData.setDescription(new I18nizableText("Multiple content data"));
        multipleTestContentData.setMultiple(true);
        _items.add(multipleTestContentData);
        
        // multiple exhaustive content data
        ContentAttributeDefinition multipleExhContentData = new ContentAttributeDefinition(_contentTypeExtensionPoint, _contentTypesHelper);
        multipleExhContentData.setName("multipleExhContentData");
        multipleExhContentData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.CONTENT_ELEMENT_TYPE_ID));
        multipleExhContentData.setContentTypeId(__EXHAUSTIVE_CONTENT_TYPE_ID);
        multipleExhContentData.setLabel(new I18nizableText("Multiple exhaustive content data"));
        multipleExhContentData.setDescription(new I18nizableText("Multiple exhaustive content data"));
        multipleExhContentData.setMultiple(true);
        _items.add(multipleExhContentData);
        
        // composite
        CompositeDefinition composite = new CompositeDefinition();
        composite.setName("composite");
        composite.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.COMPOSITE_TYPE_ID));
        composite.setLabel(new I18nizableText("Composite"));
        composite.setDescription(new I18nizableText("Composite"));
        _items.add(composite);
        
        // composite - single long data
        ElementDefinition<Long> otherSingleLongData = new DefaultElementDefinition<>();
        otherSingleLongData.setName("singleLongData");
        otherSingleLongData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.LONG_TYPE_ID));
        otherSingleLongData.setLabel(new I18nizableText("Single long data"));
        otherSingleLongData.setDescription(new I18nizableText("Single long data"));
        composite.addChild(otherSingleLongData);
        
        // composite - repeater
        RepeaterDefinition repeater = new RepeaterDefinition();
        repeater.setName("repeater");
        repeater.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.REPEATER_TYPE_ID));
        repeater.setLabel(new I18nizableText("Repeater"));
        repeater.setDescription(new I18nizableText("Repeater"));
        composite.addChild(repeater);
        
        // composite - repeater - single long data with default value
        ElementDefinition<Long> singleLongDataWithDefaultValue = new DefaultElementDefinition<>();
        singleLongDataWithDefaultValue.setName("longWithDefaultValue");
        singleLongDataWithDefaultValue.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.LONG_TYPE_ID));
        singleLongDataWithDefaultValue.setLabel(new I18nizableText("Single long data with a default value"));
        singleLongDataWithDefaultValue.setDescription(new I18nizableText("Single long data with a default value"));
        singleLongDataWithDefaultValue.setDefaultValue(459L);
        repeater.addChild(singleLongDataWithDefaultValue);
        
        // composite - repeater - single boolean data
        ElementDefinition<Boolean> otherSingleBooleanData = new DefaultElementDefinition<>();
        otherSingleBooleanData.setName("singleBooleanData");
        otherSingleBooleanData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.BOOLEAN_TYPE_ID));
        otherSingleBooleanData.setLabel(new I18nizableText("Single boolean data"));
        otherSingleBooleanData.setDescription(new I18nizableText("Single boolean data"));
        repeater.addChild(otherSingleBooleanData);
        
        // composite - repeater - single double data
        ElementDefinition<Double> otherSingleDoubleData = new DefaultElementDefinition<>();
        otherSingleDoubleData.setName("singleDoubleData");
        otherSingleDoubleData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.DOUBLE_TYPE_ID));
        otherSingleDoubleData.setLabel(new I18nizableText("Single double data"));
        otherSingleDoubleData.setDescription(new I18nizableText("Single double data"));
        repeater.addChild(otherSingleDoubleData);
        
        // composite - repeater - subcomposite
        CompositeDefinition subcomposite = new CompositeDefinition();
        subcomposite.setName("subcomposite");
        subcomposite.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.COMPOSITE_TYPE_ID));
        subcomposite.setLabel(new I18nizableText("Composite in repeater"));
        subcomposite.setDescription(new I18nizableText("Composite in repeater"));
        repeater.addChild(subcomposite);
        
        // composite - repeater - subcomposite - single string data
        ElementDefinition<String> otherSingleStringData = new DefaultElementDefinition<>();
        otherSingleStringData.setName("singleStringData");
        otherSingleStringData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.STRING_TYPE_ID));
        otherSingleStringData.setLabel(new I18nizableText("Single string data"));
        otherSingleStringData.setDescription(new I18nizableText("Single string data"));
        subcomposite.addChild(otherSingleStringData);
        
        // composite - repeater - subcomposite - multiple date data
        ElementDefinition<LocalDate> otherMultipleDateData = new DefaultElementDefinition<>();
        otherMultipleDateData.setName("multipleDateData");
        otherMultipleDateData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.DATE_TYPE_ID));
        otherMultipleDateData.setLabel(new I18nizableText("Multiple date data"));
        otherMultipleDateData.setDescription(new I18nizableText("Multiple date data"));
        otherMultipleDateData.setMultiple(true);
        subcomposite.addChild(otherMultipleDateData);
        
        // composite - repeater - subcomposite - single binary data
        ElementDefinition<Binary> otherSingleBinaryData = new DefaultElementDefinition<>();
        otherSingleBinaryData.setName("singleBinaryData");
        otherSingleBinaryData.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.BINARY_ELEMENT_TYPE_ID));
        otherSingleBinaryData.setLabel(new I18nizableText("Single binary data"));
        otherSingleBinaryData.setDescription(new I18nizableText("Single binary data"));
        subcomposite.addChild(otherSingleBinaryData);
        
        // unvalued composite (should never be valued for tests purposes)
        CompositeDefinition unvaluedComposite = new CompositeDefinition();
        unvaluedComposite.setName("unvaluedComposite");
        unvaluedComposite.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.COMPOSITE_TYPE_ID));
        unvaluedComposite.setLabel(new I18nizableText("Composite that should never be valued"));
        unvaluedComposite.setDescription(new I18nizableText("Composite that should never be valued"));
        _items.add(unvaluedComposite);
        
        // unvalued repeater (should never be valued for tests purposes)
        RepeaterDefinition unvaluedRepeater = new RepeaterDefinition();
        unvaluedRepeater.setName("unvaluedRepeater");
        unvaluedRepeater.setType(_testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.REPEATER_TYPE_ID));
        unvaluedRepeater.setLabel(new I18nizableText("Repeater that should never be valued"));
        unvaluedRepeater.setDescription(new I18nizableText("Repeater that should never be valued"));
        _items.add(unvaluedRepeater);
    }
    
    public Collection< ? extends ModelItem> getModelItems()
    {
        return Collections.unmodifiableCollection(_items);
    }

    public String getId()
    {
        return "test-model";
    }

    public String getFamilyId()
    {
        return "tests";
    }
}
