/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.repository;

import java.time.ZonedDateTime;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.contenttype.ContentAttributeDefinition;
import org.ametys.cms.contenttype.ContentType;
import org.ametys.cms.contenttype.ContentTypeExtensionPoint;
import org.ametys.cms.data.ContentSynchronizationContext;
import org.ametys.cms.data.ContentSynchronizationResult;
import org.ametys.cms.data.ContentValue;
import org.ametys.cms.data.ReferencedContents;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.external.ExternalizableDataProvider.ExternalizableDataStatus;
import org.ametys.plugins.repository.data.holder.group.ModifiableModelAwareRepeater;
import org.ametys.plugins.repository.data.holder.values.SynchronizationContext;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItemAccessor;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * {@link ModifiableContentDataHolder} implementation test cases
 */
@NeedFeatures({
    "test/content-types-with-invert-relations",
    "test-data/content-test-externalizable-data-provider",
    
    "repository/repository.data.holder.helpers",
    "repository/resolver",
    "repository/provider.jackrabbit",
    "cms/factories",
    "cms/workflow.content"
})
public class ModifiableContentDataHolderTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static final String __SOURCE_CONTENT_TYPE_ID = "tests.invert.relations.source";
    private static final String __DESTINATION_CONTENT_TYPE_ID = "tests.invert.relations.destination";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static ContentTypeExtensionPoint _contentTypeExtensionPoint;
    
    private static View _view;
    
    private ModifiableContent _source1;
    private ModifiableContent _source2;
    private ModifiableContent _source3;
    private ModifiableContent _destination1;
    private ModifiableContent _destination2;
    private ModifiableContent _destination3;
    private ModifiableContent _destination4;

    private ContentSynchronizationContext _context;
    
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _contentTypeExtensionPoint = (ContentTypeExtensionPoint) manager.lookup(ContentTypeExtensionPoint.ROLE);
    }
    
    public void prepareComponents() throws Exception
    {
        ContentType sourceContentType = _contentTypeExtensionPoint.getExtension(__SOURCE_CONTENT_TYPE_ID);
        _view = sourceContentType.getView("main");
    }
    
    /**
     * Prepare components that are specific for each test
     */
    @Before
    public void prepareSpecificComponents()
    {
        if (_source1 != null)
        {
            _source1.remove();
        }
        if (_source2 != null)
        {
            _source2.remove();
        }
        if (_source3 != null)
        {
            _source3.remove();
        }
        if (_destination1 != null)
        {
            _destination1.remove();
        }
        if (_destination2 != null)
        {
            _destination2.remove();
        }
        if (_destination3 != null)
        {
            _destination3.remove();
        }
        if (_destination4 != null)
        {
            _destination4.remove();
        }
        
        _source1 = _createContent("source", __SOURCE_CONTENT_TYPE_ID);
        _source2 = _createContent("source2", __SOURCE_CONTENT_TYPE_ID);
        _source3 = _createContent("source3", __SOURCE_CONTENT_TYPE_ID);
        _destination1 = _createContent("destination", __DESTINATION_CONTENT_TYPE_ID);
        _destination2 = _createContent("destination2", __DESTINATION_CONTENT_TYPE_ID);
        _destination3 = _createContent("destination3", __DESTINATION_CONTENT_TYPE_ID);
        _destination4 = _createContent("destination4", __DESTINATION_CONTENT_TYPE_ID);
        
        _context = ContentSynchronizationContext.newInstance();
    }
    
    /**
     * Tests {@link ModifiableContentDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method to add invert relations to single data
     * Add references from the source to the destinations.
     * The references are added in new repeater entries of the destination contents
     * The composites and repeaters are created if needed
     * @throws Exception if an exception occurs
     */
    @Test
    public void testAddSingleInvertRelation() throws Exception
    {
        // Create an empty repeater entry in the destination, so the new relation will be added in a second one
        _destination1.getComposite("composite", true)
            .getRepeater("repeater", true)
            .addEntry();
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("multipleToSingle", new Content[] {_destination1, _destination2}));
        
        _context.withReferencedContents(List.of(
                new ReferencedContents(
                        (ContentAttributeDefinition) _source1.getDefinition("composite/multipleToSingle"),
                        Map.of(
                                new ContentValue(_destination1), List.of("composite/repeater[2]/singleFromMultiple"),
                                new ContentValue(_destination2), List.of("composite/repeater[1]/singleFromMultiple")),
                        Map.of(),
                        Map.of())));
        
        ContentSynchronizationResult result = _source1.synchronizeValues(_view, valuesToSynchronize, _context);
        
        Assert.assertTrue(result.hasChanged());
        
        ContentValue[] multipleToSingle = _source1.getValue("composite/multipleToSingle");
        Assert.assertEquals(2, multipleToSingle.length);
        Assert.assertEquals(_destination1.getId(), multipleToSingle[0].getContentId());
        Assert.assertEquals(_destination2.getId(), multipleToSingle[1].getContentId());
        
        ModifiableModelAwareRepeater dest1Repeater = _destination1.getRepeater("composite/repeater");
        Assert.assertEquals(2,  dest1Repeater.getSize());
        Assert.assertFalse(dest1Repeater.getEntry(1).hasValue("singleFromMultiple"));
        Assert.assertEquals(_source1.getId(), ((ContentValue) dest1Repeater.getEntry(2).getValue("singleFromMultiple")).getContentId());
        
        ModifiableModelAwareRepeater dest2Repeater = _destination2.getRepeater("composite/repeater");
        Assert.assertEquals(1,  dest2Repeater.getSize());
        Assert.assertEquals(_source1.getId(), ((ContentValue) dest2Repeater.getEntry(1).getValue("singleFromMultiple")).getContentId());
        
        Set<ModifiableContent> modifiedContents = result.getModifiedContents();
        Assert.assertEquals(2, modifiedContents.size());
        Assert.assertTrue(modifiedContents.contains(_destination1));
        Assert.assertTrue(modifiedContents.contains(_destination2));
    }
    
    /**
     * Tests {@link ModifiableContentDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method to add invert relations to multiple data
     * Add references from the source to the destinations.
     * The references are added to the existent ones in the destination contents
     * The composites are created if needed
     * @throws Exception if an exception occurs
     */
    @Test
    public void testAddMultipleInvertRelation() throws Exception
    {
        // Add relation to another source in the destination
        _destination1.getComposite("composite", true)
            .setValue("multipleFromMultiple", new Content[] {_source2});
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("multipleToMultiple", new Content[] {_destination1, _destination2}));
        
        _context.withReferencedContents(List.of(
                new ReferencedContents(
                        (ContentAttributeDefinition) _source1.getDefinition("composite/multipleToMultiple"),
                        Map.of(
                                new ContentValue(_destination1), List.of("composite/multipleFromMultiple"),
                                new ContentValue(_destination2), List.of("composite/multipleFromMultiple")),
                        Map.of(),
                        Map.of())));

        ContentSynchronizationResult result = _source1.synchronizeValues(_view, valuesToSynchronize, _context);
        
        Assert.assertTrue(result.hasChanged());
        
        ContentValue[] multipleToMultiple = _source1.getValue("composite/multipleToMultiple");
        Assert.assertEquals(2, multipleToMultiple.length);
        Assert.assertEquals(_destination1.getId(), multipleToMultiple[0].getContentId());
        Assert.assertEquals(_destination2.getId(), multipleToMultiple[1].getContentId());
        
        ContentValue[] multipleFromMultiple1 = _destination1.getValue("composite/multipleFromMultiple");
        Assert.assertEquals(2, multipleFromMultiple1.length);
        Assert.assertEquals(_source2.getId(), multipleFromMultiple1[0].getContentId());
        Assert.assertEquals(_source1.getId(), multipleFromMultiple1[1].getContentId());
        
        ContentValue[] multipleFromMultiple2 = _destination2.getValue("composite/multipleFromMultiple");
        Assert.assertEquals(1, multipleFromMultiple2.length);
        Assert.assertEquals(_source1.getId(), multipleFromMultiple2[0].getContentId());
        
        Set<ModifiableContent> modifiedContents = result.getModifiedContents();
        Assert.assertEquals(2, modifiedContents.size());
        Assert.assertTrue(modifiedContents.contains(_destination1));
        Assert.assertTrue(modifiedContents.contains(_destination2));
    }
    
    /**
     * Tests {@link ModifiableContentDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method to remove invert relations in single data
     * The references to the source have to be removed in the destination.
     * @throws Exception if an exception occurs
     */
    @Test
    public void testRemoveSingleInvertRelation() throws Exception
    {
        // Set references to the destination in the sources
        _source1.getComposite("composite", true)
            .getRepeater("repeater", true)
            .addEntry()
            .setValue("singleToSingle", _destination1);
        
        _source2.getComposite("composite", true)
            .getRepeater("repeater", true)
            .addEntry()
            .setValue("singleToSingle", _destination1);

        // Set the invert relation to the sources in the destination
        ModifiableModelAwareRepeater destRepeater = _destination1.getComposite("composite", true)
                .getRepeater("repeater", true);
        destRepeater.addEntry()
            .setValue("singleFromSingle", _source1);
        destRepeater.addEntry()
            .setValue("singleFromSingle", _source1);
        destRepeater.addEntry()
            .setValue("singleFromSingle", _source2);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of())));
        
        _context.withReferencedContents(List.of(
                new ReferencedContents(
                        (ContentAttributeDefinition) _source1.getDefinition("composite/repeater/singleToSingle"),
                        Map.of(),
                        Map.of(new ContentValue(_destination1), List.of("composite/repeater[1]/singleFromSingle", "composite/repeater[2]/singleFromSingle")),
                        Map.of())));
        
        ContentSynchronizationResult result = _source1.synchronizeValues(_view, valuesToSynchronize, _context);
        
        Assert.assertTrue(result.hasChanged());
        
        ModifiableModelAwareRepeater src1Repeater = _source1.getRepeater("composite/repeater");
        Assert.assertEquals(1,  src1Repeater.getSize());
        Assert.assertFalse(src1Repeater.getEntry(1).hasValue("singleToSingle"));
        
        ModifiableModelAwareRepeater dest1Repeater = _destination1.getRepeater("composite/repeater");
        Assert.assertEquals(3,  dest1Repeater.getSize());
        Assert.assertFalse(dest1Repeater.getEntry(1).hasValue("singleFromSingle"));
        Assert.assertFalse(dest1Repeater.getEntry(2).hasValue("singleFromSingle"));
        Assert.assertEquals(_source2.getId(), ((ContentValue) dest1Repeater.getEntry(3).getValue("singleFromSingle")).getContentId());
        
        Set<ModifiableContent> modifiedContents = result.getModifiedContents();
        Assert.assertEquals(1, modifiedContents.size());
        Assert.assertTrue(modifiedContents.contains(_destination1));
    }
    
    /**
     * Tests {@link ModifiableContentDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method to remove invert relations in multiple data
     * The references to the source have to be removed in the destination.
     * @throws Exception if an exception occurs
     */
    @Test
    public void testRemoveMultipleInvertRelation() throws Exception
    {
        // Set the reference to the destination in the source
        _source1.getComposite("composite", true)
            .getRepeater("repeater", true)
            .addEntry()
            .setValue("singleToMultiple", _destination1);

        // Set the invert relation to the source in the destination
        _destination1.getComposite("composite", true)
            .setValue("multipleFromSingle", new Content[] {_source1, _source2, _source3});
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("repeater", List.of(Map.of())));
        
        _context.withReferencedContents(List.of(
                new ReferencedContents(
                        (ContentAttributeDefinition) _source1.getDefinition("composite/repeater/singleToMultiple"),
                        Map.of(),
                        Map.of(new ContentValue(_destination1), List.of("composite/multipleFromSingle")),
                        Map.of())));

        ContentSynchronizationResult result = _source1.synchronizeValues(_view, valuesToSynchronize, _context);
        
        Assert.assertTrue(result.hasChanged());
        
        ModifiableModelAwareRepeater src1Repeater = _source1.getRepeater("composite/repeater");
        Assert.assertEquals(1, src1Repeater.getSize());
        Assert.assertFalse(src1Repeater.getEntry(1).hasValue("singleToMultiple"));
        
        ContentValue[] multipleFromSingle = _destination1.getValue("composite/multipleFromSingle");
        Assert.assertEquals(2, multipleFromSingle.length);
        Assert.assertEquals(_source2.getId(), multipleFromSingle[0].getContentId());
        Assert.assertEquals(_source3.getId(), multipleFromSingle[1].getContentId());
        
        Set<ModifiableContent> modifiedContents = result.getModifiedContents();
        Assert.assertEquals(1, modifiedContents.size());
        Assert.assertTrue(modifiedContents.contains(_destination1));
    }
    
    /**
     * Tests {@link ModifiableContentDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method to ensure a same content does not appear twice if appears in several referenced contents
     * The destination is removed from a relation but added in another one
     * The content should appear only once in the result
     * @throws Exception if an exception occurs
     */
    @Test
    public void testChangeSeveralRelationsOnSameContent() throws Exception
    {
        // Set the reference to the destination in the source
        _source1.getComposite("composite", true)
            .getRepeater("repeater", true)
            .addEntry()
            .setValue("singleToSingle", _destination1);

        // Set the invert relation to the source in the destination
        _destination1.getComposite("composite", true)
            .getRepeater("repeater", true)
            .addEntry()
            .setValue("singleFromSingle", _source1);
        
        // Change the relation on singleToSingle to put the second destination and add a relation on first destination on singleToMultiple
        Map<String, Object> valuesToSynchronize = Map.of(
                "composite", Map.of(
                        "repeater", List.of(
                                Map.of(
                                        "singleToSingle", _destination2,
                                        "singleToMultiple", _destination1))));
        
        _context.withReferencedContents(List.of(
                new ReferencedContents(
                        (ContentAttributeDefinition) _source1.getDefinition("composite/repeater/singleToSingle"),
                        Map.of(new ContentValue(_destination2), List.of("composite/singleFromSingle")),
                        Map.of(new ContentValue(_destination1), List.of("composite/singleFromSingle")),
                        Map.of()),
                new ReferencedContents(
                        (ContentAttributeDefinition) _source1.getDefinition("composite/repeater/singleToMultiple"),
                        Map.of(new ContentValue(_destination1), List.of("composite/multipleFromSingle")),
                        Map.of(),
                        Map.of())));
        
        ContentSynchronizationResult result = _source1.synchronizeValues(_view, valuesToSynchronize, _context);
        
        Assert.assertTrue(result.hasChanged());
        
        ModifiableModelAwareRepeater src1Repeater = _source1.getRepeater("composite/repeater");
        Assert.assertEquals(1, src1Repeater.getSize());
        Assert.assertEquals(_destination2.getId(), ((ContentValue) src1Repeater.getEntry(1).getValue("singleToSingle")).getContentId());
        Assert.assertEquals(_destination1.getId(), ((ContentValue) src1Repeater.getEntry(1).getValue("singleToMultiple")).getContentId());
        
        ModifiableModelAwareRepeater dest1Repeater = _destination1.getRepeater("composite/repeater");
        Assert.assertEquals(1, dest1Repeater.getSize());
        Assert.assertFalse(dest1Repeater.getEntry(1).hasValue("singleFromSingle"));
        ContentValue[] multipleFromSingle1 = _destination1.getValue("composite/multipleFromSingle");
        Assert.assertEquals(1, multipleFromSingle1.length);
        Assert.assertEquals(_source1.getId(), multipleFromSingle1[0].getContentId());
        
        ModifiableModelAwareRepeater dest2Repeater = _destination2.getRepeater("composite/repeater");
        Assert.assertEquals(1, dest2Repeater.getSize());
        Assert.assertEquals(_source1.getId(), ((ContentValue) dest2Repeater.getEntry(1).getValue("singleFromSingle")).getContentId());
        
        Set<ModifiableContent> modifiedContents = result.getModifiedContents();
        Assert.assertEquals(2, modifiedContents.size());
        Assert.assertTrue(modifiedContents.contains(_destination1));
        Assert.assertTrue(modifiedContents.contains(_destination2));
    }
    
    /**
     * Tests {@link ModifiableContentDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method to add an invert relation to an externalizable data
     * The destination data is externalizable
     * Add an invert relation on its data
     * @throws Exception if an exception occurs
     */
    @Test
    public void testAddInvertRelationOnExternalizableData() throws Exception
    {
        _destination1.setStatus("externalizableSingleFromSingle", ExternalizableDataStatus.LOCAL);
        _destination1.setExternalValue("externalizableSingleFromSingle", _source2);
        
        Map<String, Object> valuesToSynchronize = Map.of("singleToExternalizableSingle", _destination1);
        
        _context.withReferencedContents(List.of(
                new ReferencedContents(
                        (ContentAttributeDefinition) _source1.getDefinition("singleToExternalizableSingle"),
                        Map.of(new ContentValue(_destination1), List.of("externalizableSingleFromSingle")),
                        Map.of(),
                        Map.of())));
        
        ContentSynchronizationResult result = _source1.synchronizeValues(_view, valuesToSynchronize, _context);
        
        Assert.assertTrue(result.hasChanged());
        
        Assert.assertEquals(_destination1.getId(), ((ContentValue) _source1.getValue("singleToExternalizableSingle")).getContentId());
        
        Assert.assertEquals(_source1.getId(), ((ContentValue) _destination1.getLocalValue("externalizableSingleFromSingle")).getContentId());
        Assert.assertEquals(_source2.getId(), ((ContentValue) _destination1.getExternalValue("externalizableSingleFromSingle")).getContentId());
        
        Set<ModifiableContent> modifiedContents = result.getModifiedContents();
        Assert.assertEquals(1, modifiedContents.size());
        Assert.assertTrue(modifiedContents.contains(_destination1));
    }
    
    /**
     * Tests {@link ModifiableContentDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method to ensure the method fails if we try to add an invert relation to an external data
     * The destination data is set to external
     * Try to add an invert relation on its data
     * @throws Exception if an exception occurs
     */
    @Test
    public void testFailToAddInvertRelationOnExternalData() throws Exception
    {
        _destination1.setStatus("externalizableSingleFromSingle", ExternalizableDataStatus.EXTERNAL);
        
        Map<String, Object> valuesToSynchronize = Map.of("singleToExternalizableSingle", _destination1);
        
        _context.withReferencedContents(List.of(
                new ReferencedContents(
                        (ContentAttributeDefinition) _source1.getDefinition("singleToExternalizableSingle"),
                        Map.of(new ContentValue(_destination1), List.of("externalizableSingleFromSingle")),
                        Map.of(),
                        Map.of())));
        
        try
        {
            _source1.synchronizeValues(_view, valuesToSynchronize, _context);
            Assert.fail("The synchronization should fail if trying to add an invert relation on an external data");
        }
        catch (IllegalStateException e)
        {
            // Nothing to do
        }
    }
    
    /**
     * Tests {@link ModifiableContentDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method to remove an invert relation from an externalizable data
     * The destination data is externalizable
     * Remove an invert relation on its data
     * @throws Exception if an exception occurs
     */
    @Test
    public void testRemoveInvertRelationOnExternalizableData() throws Exception
    {
        _source1.setValue("singleToExternalizableSingle", _destination1);
        _destination1.setStatus("externalizableSingleFromSingle", ExternalizableDataStatus.LOCAL);
        _destination1.setLocalValue("externalizableSingleFromSingle", _source1);
        _destination1.setExternalValue("externalizableSingleFromSingle", _source2);
        
        Map<String, Object> valuesToSynchronize = Map.of();
        
        _context.withReferencedContents(List.of(
                new ReferencedContents(
                        (ContentAttributeDefinition) _source1.getDefinition("singleToExternalizableSingle"),
                        Map.of(),
                        Map.of(new ContentValue(_destination1), List.of("externalizableSingleFromSingle")),
                        Map.of())));
        
        ContentSynchronizationResult result = _source1.synchronizeValues(_view, valuesToSynchronize, _context);
        
        Assert.assertTrue(result.hasChanged());
        
        Assert.assertFalse(_source1.hasValue("singleToExternalizableSingle"));
        
        Assert.assertFalse(_destination1.hasLocalValue("externalizableSingleFromSingle"));
        Assert.assertEquals(_source2.getId(), ((ContentValue) _destination1.getExternalValue("externalizableSingleFromSingle")).getContentId());
        
        Set<ModifiableContent> modifiedContents = result.getModifiedContents();
        Assert.assertEquals(1, modifiedContents.size());
        Assert.assertTrue(modifiedContents.contains(_destination1));
    }
    
    /**
     * Tests {@link ModifiableContentDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method to ensure the method fails if we try to remove an invert relation from an external data
     * The destination data is set to external
     * Try to remove an invert relation on its data
     * @throws Exception if an exception occurs
     */
    @Test
    public void testFailToRemoveInvertRelationOnExternalData() throws Exception
    {
        _source1.setValue("singleToExternalizableSingle", _destination1);
        _destination1.setStatus("externalizableSingleFromSingle", ExternalizableDataStatus.EXTERNAL);
        _destination1.setValue("externalizableSingleFromSingle", _source1);
        
        Map<String, Object> valuesToSynchronize = Map.of("singleToExternalizableSingle", _destination2);
        
        _context.withReferencedContents(List.of(
                new ReferencedContents(
                        (ContentAttributeDefinition) _source1.getDefinition("singleToExternalizableSingle"),
                        Map.of(new ContentValue(_destination2), List.of("externalizableSingleFromSingle")),
                        Map.of(new ContentValue(_destination1), List.of("externalizableSingleFromSingle")),
                        Map.of())));
        
        try
        {
            _source1.synchronizeValues(_view, valuesToSynchronize, _context);
            Assert.fail("The synchronization should fail if trying to remove an invert relation on an external data");
        }
        catch (IllegalStateException e)
        {
            // Nothing to do
        }
    }
    
    /**
     * Tests {@link ModifiableContentDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method to remove the invert relation of a source content pointed by an invert relation to remove on destination an invert relation to an external data
     * There is a reference between the second source and the first destination in a single data, not in a repeater
     * Put a relation from the first source to the first destination
     * The relation from the second source to the first destination has to be removed
     * @throws Exception if an exception occurs
     */
    @Test
    public void testRemoveInvertRelationOnSingle3rdPartyContent() throws Exception
    {
        _source2.setValue("singleToExternalizableSingle", _destination1);
        _destination1.setValue("externalizableSingleFromSingle", _source2);
        
        Map<String, Object> valuesToSynchronize = Map.of("singleToExternalizableSingle", _destination1);
        
        _context.withReferencedContents(List.of(
                new ReferencedContents(
                        (ContentAttributeDefinition) _source1.getDefinition("singleToExternalizableSingle"),
                        Map.of(new ContentValue(_destination1), List.of("externalizableSingleFromSingle")),
                        Map.of(),
                        Map.of(new ContentValue(_destination1), new ContentValue(_source2)))));
        
        ContentSynchronizationResult result = _source1.synchronizeValues(_view, valuesToSynchronize, _context);
        
        Assert.assertTrue(result.hasChanged());
        
        Assert.assertEquals(_destination1.getId(), ((ContentValue) _source1.getValue("singleToExternalizableSingle")).getContentId());
        
        Assert.assertEquals(_source1.getId(), ((ContentValue) _destination1.getLocalValue("externalizableSingleFromSingle")).getContentId());
        
        Assert.assertFalse(_source2.hasValue("singleToExternalizableSingle"));
        
        Set<ModifiableContent> modifiedContents = result.getModifiedContents();
        Assert.assertEquals(2, modifiedContents.size());
        Assert.assertTrue(modifiedContents.contains(_destination1));
        Assert.assertTrue(modifiedContents.contains(_source2));
    }
    
    /**
     * Tests {@link ModifiableContentDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method to remove the invert relation of a source content pointed by an invert relation to remove on destination an invert relation to an external data
     * There references between sources and destinations. The source data is multiple and the destination data is single
     * Put a relation from the first source to all destinations
     * The relations from the other sources to the destinations have to be removed
     * @throws Exception if an exception occurs
     */
    @Test
    public void testRemoveInvertRelationOnMultiple3rdPartyContent() throws Exception
    {
        _source2.getComposite("composite", true)
                .setValue("multipleToReallySingle", new ModifiableContent[] {_destination1, _destination2, _destination4});

        _destination1.getComposite("composite", true)
                     .setValue("reallySingleFromMultiple", _source2);

        _destination2.getComposite("composite", true)
                     .setValue("reallySingleFromMultiple", _source2);

        _destination3.getComposite("composite", true)
                     .setValue("reallySingleFromMultiple", _source3);
        
        Map<String, Object> valuesToSynchronize = Map.of("composite", Map.of("multipleToReallySingle", new Content[] {_destination1, _destination2, _destination3}));
        
        _context.withReferencedContents(List.of(
                new ReferencedContents(
                        (ContentAttributeDefinition) _source1.getDefinition("composite/multipleToReallySingle"),
                        Map.of(new ContentValue(_destination1), List.of("composite/reallySingleFromMultiple"),
                               new ContentValue(_destination2), List.of("composite/reallySingleFromMultiple"),
                               new ContentValue(_destination3), List.of("composite/reallySingleFromMultiple")),
                        Map.of(),
                        Map.of(new ContentValue(_destination1), new ContentValue(_source2),
                               new ContentValue(_destination2), new ContentValue(_source2),
                               new ContentValue(_destination3), new ContentValue(_source3)))));
        
        ContentSynchronizationResult result = _source1.synchronizeValues(_view, valuesToSynchronize, _context);
        
        Assert.assertTrue(result.hasChanged());
        
        ContentValue[] source1MultipleToReallySingle = _source1.getValue("composite/multipleToReallySingle");
        Assert.assertEquals(3, source1MultipleToReallySingle.length);
        Assert.assertEquals(_destination1.getId(), source1MultipleToReallySingle[0].getContentId());
        Assert.assertEquals(_destination2.getId(), source1MultipleToReallySingle[1].getContentId());
        Assert.assertEquals(_destination3.getId(), source1MultipleToReallySingle[2].getContentId());
        
        Assert.assertEquals(_source1.getId(), ((ContentValue) _destination1.getValue("composite/reallySingleFromMultiple")).getContentId());
        Assert.assertEquals(_source1.getId(), ((ContentValue) _destination2.getValue("composite/reallySingleFromMultiple")).getContentId());
        Assert.assertEquals(_source1.getId(), ((ContentValue) _destination3.getValue("composite/reallySingleFromMultiple")).getContentId());
        
        ContentValue[] source2MultipleToReallySingle = _source2.getValue("composite/multipleToReallySingle");
        Assert.assertEquals(1, source2MultipleToReallySingle.length);
        Assert.assertEquals(_destination4.getId(), source2MultipleToReallySingle[0].getContentId());
        
        Assert.assertFalse(_source3.hasValue("composite/multipleToReallySingle"));
        
        Set<ModifiableContent> modifiedContents = result.getModifiedContents();
        Assert.assertEquals(4, modifiedContents.size());
        Assert.assertTrue(modifiedContents.contains(_destination1));
        Assert.assertTrue(modifiedContents.contains(_destination2));
        Assert.assertTrue(modifiedContents.contains(_destination3));
        Assert.assertTrue(modifiedContents.contains(_source2));
    }
    
    /**
     * Tests {@link ModifiableContentDataHolder#synchronizeValues(ViewItemAccessor, Map, SynchronizationContext)} method to ensure that the collectReferencedContents is well called if referenced contents are not in the context
     * @throws Exception if an exception occurs
     */
    @Test
    public void testCollectReferencedContents() throws Exception
    {
        // Set the reference to the destination in the source
        _source1.getComposite("composite", true)
            .getRepeater("repeater", true)
            .addEntry()
            .setValue("singleToMultiple", _destination1);

        // Set the invert relation to the source in the destination
        _destination1.getComposite("composite", true)
            .setValue("multipleFromSingle", new Content[] {_source1});
        _destination2.getComposite("composite", true)
            .setValue("multipleFromSingle", new Content[] {_source2});
        
        Map<String, Object> valuesToSynchronize = Map.of(
                "composite", Map.of(
                        "repeater", List.of(
                                Map.of("singleToMultiple", _destination2))));
        
        ContentSynchronizationResult result = _source1.synchronizeValues(_view, valuesToSynchronize, _context);
        
        Assert.assertTrue(result.hasChanged());
        
        ModifiableModelAwareRepeater src1Repeater = _source1.getRepeater("composite/repeater");
        Assert.assertEquals(1, src1Repeater.getSize());
        Assert.assertEquals(_destination2.getId(), ((ContentValue) src1Repeater.getEntry(1).getValue("singleToMultiple")).getContentId());
        
        ContentValue[] multipleFromSingle1 = _destination1.getValue("composite/multipleFromSingle");
        Assert.assertEquals(0, multipleFromSingle1.length);
        
        ContentValue[] multipleFromSingle2 = _destination2.getValue("composite/multipleFromSingle");
        Assert.assertEquals(2, multipleFromSingle2.length);
        Assert.assertEquals(_source2.getId(), multipleFromSingle2[0].getContentId());
        Assert.assertEquals(_source1.getId(), multipleFromSingle2[1].getContentId());
        
        Set<ModifiableContent> modifiedContents = result.getModifiedContents();
        Assert.assertEquals(2, modifiedContents.size());
        Assert.assertTrue(modifiedContents.contains(_destination1));
        Assert.assertTrue(modifiedContents.contains(_destination2));
    }
    
    private ModifiableContent _createContent(String title, String contentTypeId)
    {
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        
        ModifiableContent content = root.createChild(title, "ametys:defaultContent");

        content.setTypes(new String[]{contentTypeId});
        content.setTitle(title, Locale.FRANCE);
        content.setLanguage("fr");
        content.setCreationDate(ZonedDateTime.now());
        content.setCreator(new UserIdentity("enaour", "wallace"));
        content.setLastModified(ZonedDateTime.now());
        content.setLastContributor(new UserIdentity("bfraisse", "wallace"));
        
        content.saveChanges();
        return content;
    }
}
