/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.source;

import java.util.Iterator;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.components.LifecycleHelper;
import org.apache.cocoon.util.log.SLF4JLoggerAdapter;
import org.junit.Assert;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

import com.opensymphony.workflow.WorkflowException;

/**
 * {@link DefaultContentView} test case.
 * Used content types in these test case are defined in default-content-view-plugin(1 and2)
 * To get a more readable view of these content types and their hierarchy, see txt file in test/resources folder
 */
// Currently no @NeedFeatures because default-content-view-plugin1/content-types and default-content-view-plugin2/content-types are already defined by WithCmsFeaturestestCase
@NeedFeatures ({
    "cms/workflow.content"
})
public class DefaultContentViewTestCase extends AbstractRuntimeTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    private static DefaultContentView _contentView;
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _setupContentView(manager);
    }
    
    private void _setupContentView(ServiceManager manager) throws ServiceException
    {
        _contentView = new DefaultContentView();
        Logger logger = LoggerFactory.getLogger("org.ametys.cms.source.contentView");
        try
        {
            LifecycleHelper.setupComponent(_contentView, new SLF4JLoggerAdapter(logger), getCocoonWrapper().getContext(), manager, null);
        }
        catch (Exception e)
        {
            throw new ServiceException(DefaultContentView.class.getName(), "Error when setting up ContentView component.", e);
        }
    }
    
    /**
     * Tests retrieval of source URIs for a content type that does not extend content types in HTML format
     * @throws WorkflowException if an error occurs
     */
    @Test
    public void testGetSourceURIsNoInheritanceHTMLFormat() throws WorkflowException
    {
        Iterator<String> result = _contentView._getSourceURIs("H", "H", "my-view", "html", "test", "xsl").iterator();
        
        // H content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/H/H-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/H/H.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/H/H-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/H/H.xsl", result.next());
        
        // Default possible URIs
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:test://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:cms://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("cocoon://_plugins/cms/default-content/html.xsl?viewName=my-view", result.next());
        
        // Assert.assertion on result size
        Assert.assertFalse("result should not have more entries", result.hasNext());
    }
    
    /**
     * Tests retrieval of source URIs for a content type with no current plugin name
     */
    @Test
    public void testGetSourceURIsWithNoCurrentPlugin()
    {
        Iterator<String> result = _contentView._getSourceURIs("H", "H", "my-view", "html", null, "xsl").iterator();
        
        // H content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/H/H-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/H/H.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/H/H-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/H/H.xsl", result.next());
        
        // Default possible URIs
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:cms://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("cocoon://_plugins/cms/default-content/html.xsl?viewName=my-view", result.next());
        
        // Assert.assertion on result size
        Assert.assertFalse("result should not have more entries", result.hasNext());
    }
    
    /**
     * Tests retrieval of source URIs for a content type when the current plugin name is the plugin of declared content type
     */
    @Test
    public void testGetSourceURIsWithSameCurrentPlugin()
    {
        // The current plugin is the plugin of declared content type
        Iterator<String> result = _contentView._getSourceURIs("H", "H", "my-view", "html", "default-content-view-plugin1", "xsl").iterator();
        
        // H content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/H/H-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/H/H.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/H/H-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/H/H.xsl", result.next());
        
        // Default possible URIs
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:cms://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("cocoon://_plugins/cms/default-content/html.xsl?viewName=my-view", result.next());
        
        // Assert.assertion on result size
        Assert.assertFalse("result should not have more entries", result.hasNext());
    }
    
    /**
     * Tests retrieval of source URIs for a content type that does not extend content types in PDF format
     */
    @Test
    public void testGetSourceURIsNoInheritancePDFFormat()
    {
        Iterator<String> result = _contentView._getSourceURIs("H", "H", "my-view", "pdf", "test", "xsl").iterator();
        
        // H content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/H/H2pdf-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/H/H2pdf.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/H/H2pdf-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/H/H2pdf.xsl", result.next());
        
        // Default possible URIs
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/default-content2pdf.xsl", result.next());
        Assert.assertEquals("plugin:test://stylesheets/default-content2pdf.xsl", result.next());
        Assert.assertEquals("plugin:cms://stylesheets/default-content2pdf.xsl", result.next());
        Assert.assertEquals("cocoon://_plugins/cms/default-content/pdf.xsl?viewName=my-view", result.next());
        
        // Assert.assertion on result size
        Assert.assertFalse("result should not have more entries", result.hasNext());
    }
    
    /**
     * Tests retrieval of source URIs for a content type that does extend one content type
     */
    @Test
    public void testGetSourceURIsSimpleInheritance()
    {
        Iterator<String> result = _contentView._getSourceURIs("E", "E", "my-view", "html", "test", "xsl").iterator();
        
        // E content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/E/E-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/E/E.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/E/E-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/E/E.xsl", result.next());
        
        // D content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/D/D-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/D/D.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/D/D-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/D/D.xsl", result.next());
        
        // Default possible URIs
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:test://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:cms://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("cocoon://_plugins/cms/default-content/html.xsl?viewName=my-view", result.next());
        
        // Assert.assertion on result size
        Assert.assertFalse("result should not have more entries", result.hasNext());
    }
    
    /**
     * Tests retrieval of source URIs for a content type that does extend one content type in another plugin
     */
    @Test
    public void testGetSourceURIsSimpleInheritanceMultiPlugins()
    {
        Iterator<String> result = _contentView._getSourceURIs("I", "I", "my-view", "html", "test", "xsl").iterator();
        
        // I content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin2/stylesheets/I/I-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin2/stylesheets/I/I.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/content/I/I-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/content/I/I.xsl", result.next());
        
        // D content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/D/D-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/D/D.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/D/D-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/D/D.xsl", result.next());
        
        // Default possible URIs
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:test://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:cms://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("cocoon://_plugins/cms/default-content/html.xsl?viewName=my-view", result.next());
        
        // Assert.assertion on result size
        Assert.assertFalse("result should not have more entries", result.hasNext());
    }
    
    /**
     * Tests retrieval of source URIs for a content type that does extend multiple content types
     */
    @Test
    public void testGetSourceURIsMultiInheritance()
    {
        Iterator<String> result = _contentView._getSourceURIs("G", "G", "my-view", "html", "test", "xsl").iterator();
        
        // G content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/G/G-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/G/G.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/G/G-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/G/G.xsl", result.next());
        
        // CF dynamic content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/_dynamic/default-content-view-plugin1/stylesheets/dynamic-content-type.CF/dynamic-content-type.CF-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/_dynamic/default-content-view-plugin1/stylesheets/dynamic-content-type.CF/dynamic-content-type.CF.xsl", result.next());
        
        // C content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/C/C-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/C/C.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/C/C-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/C/C.xsl", result.next());
        
        // F content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/F/F-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/F/F.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/F/F-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/F/F.xsl", result.next());
        
        // AB dynamic content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/_dynamic/default-content-view-plugin1/stylesheets/dynamic-content-type.AB/dynamic-content-type.AB-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/_dynamic/default-content-view-plugin1/stylesheets/dynamic-content-type.AB/dynamic-content-type.AB.xsl", result.next());
        
        // A content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/A/A-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/A/A.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/A/A-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/A/A.xsl", result.next());
        
        // B content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/B/B-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/B/B.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/B/B-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/B/B.xsl", result.next());
        
        // E content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/E/E-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/E/E.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/E/E-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/E/E.xsl", result.next());
        
        // D content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/D/D-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/D/D.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/D/D-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/D/D.xsl", result.next());
        
        // Default possible URIs
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:test://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:cms://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("cocoon://_plugins/cms/default-content/html.xsl?viewName=my-view", result.next());
        
        // Assert.assertion on result size
        Assert.assertFalse("result should not have more entries", result.hasNext());
    }
    
    /**
     * Tests retrieval of source URIs for a content type that does extend multiple content types that are in different plugins
     */
    @Test
    public void testGetSourceURIsMultiInheritanceMultiPlugins()
    {
        Iterator<String> result = _contentView._getSourceURIs("J", "J", "my-view", "html", "test", "xsl").iterator();
        
        // J content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin2/stylesheets/J/J-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin2/stylesheets/J/J.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/content/J/J-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/content/J/J.xsl", result.next());
        
        // HI dynamic content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/_dynamic/default-content-view-plugin1/stylesheets/dynamic-content-type.HI/dynamic-content-type.HI-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/_dynamic/default-content-view-plugin1/stylesheets/dynamic-content-type.HI/dynamic-content-type.HI.xsl", result.next());
        
        // H content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/H/H-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/H/H.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/H/H-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/H/H.xsl", result.next());
        
        // I content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin2/stylesheets/I/I-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin2/stylesheets/I/I.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/content/I/I-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/content/I/I.xsl", result.next());
        
        // D content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/D/D-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/D/D.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/D/D-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/D/D.xsl", result.next());
        
        // Default possible URIs
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:test://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:cms://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("cocoon://_plugins/cms/default-content/html.xsl?viewName=my-view", result.next());
        
        // Assert.assertion on result size
        Assert.assertFalse("result should not have more entries", result.hasNext());
    }
    
    /**
     * Tests retrieval of source URIs for a content type that does extend multiple content types. One of them is a mixin
     */
    @Test
    public void testGetSourceURIsMultiInheritanceWithMixin()
    {
        Iterator<String> result = _contentView._getSourceURIs("L", "L", "my-view", "html", "test", "xsl").iterator();
        
        // L content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/L/L-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/L/L.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/L/L-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/L/L.xsl", result.next());
        
        // AK dynamic content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/_dynamic/default-content-view-plugin1/stylesheets/dynamic-content-type.AK/dynamic-content-type.AK-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/_dynamic/default-content-view-plugin1/stylesheets/dynamic-content-type.AK/dynamic-content-type.AK.xsl", result.next());
        
        // A content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/A/A-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin1/stylesheets/A/A.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/A/A-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/content/A/A.xsl", result.next());
        
        // Default possible URIs
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:test://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:cms://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("cocoon://_plugins/cms/default-content/html.xsl?viewName=my-view", result.next());
        
        // Assert.assertion on result size
        Assert.assertFalse("result should not have more entries", result.hasNext());
    }
    
    /**
     * Tests retrieval of source URIs for a content type that does extend multiple content types. There is no dynamic content type defined for parents
     */
    @Test
    public void testGetSourceURIsMultiInheritanceWithoutDynamic()
    {
        Iterator<String> result = _contentView._getSourceURIs("O", "O", "my-view", "html", "test", "xsl").iterator();
        
        // O content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin2/stylesheets/O/O-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin2/stylesheets/O/O.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/content/O/O-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/content/O/O.xsl", result.next());
        
        // M content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin2/stylesheets/M/M-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin2/stylesheets/M/M.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/content/M/M-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/content/M/M.xsl", result.next());
        
        // N content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin2/stylesheets/N/N-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/default-content-view-plugin2/stylesheets/N/N.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/content/N/N-my-view.xsl", result.next());
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/content/N/N.xsl", result.next());
        
        // Default possible URIs
        Assert.assertEquals("plugin:default-content-view-plugin2://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:test://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:cms://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("cocoon://_plugins/cms/default-content/html.xsl?viewName=my-view", result.next());
        
        // Assert.assertion on result size
        Assert.assertFalse("result should not have more entries", result.hasNext());
    }
    
    /**
     * Tests retrieval of source URIs for a dynamic content type
     */
    @Test
    public void testGetSourceURIsOnDynamicContentType()
    {
        Iterator<String> result = _contentView._getSourceURIs("dynamic-content-type.AB", "dynamic-content-type.AB", "my-view", "html", "test", "xsl").iterator();
        
        // AB dynamic content type possible URIs
        Assert.assertEquals("context://WEB-INF/param/content-types/_dynamic/default-content-view-plugin1/stylesheets/dynamic-content-type.AB/dynamic-content-type.AB-my-view.xsl", result.next());
        Assert.assertEquals("context://WEB-INF/param/content-types/_dynamic/default-content-view-plugin1/stylesheets/dynamic-content-type.AB/dynamic-content-type.AB.xsl", result.next());
        
        // Default possible URIs
        Assert.assertEquals("plugin:default-content-view-plugin1://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:test://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("plugin:cms://stylesheets/default-content.xsl", result.next());
        Assert.assertEquals("cocoon://_plugins/cms/default-content/html.xsl?viewName=my-view", result.next());
        
        // Assert.assertion on result size
        Assert.assertFalse("result should not have more entries", result.hasNext());
    }
}
