/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.transformation;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableDefaultContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.cms.transformation.AttributeURIResolver.AttributeInfo;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.data.ametysobject.ModelAwareDataAwareAmetysObject;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.plugins.repository.version.VersionAwareAmetysObject;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Tests the {@link AttributeURIResolver} class
 */
@NeedFeatures({
    "repository/resolver",
    "cms/factories",
    "cms/link-resolver"
})
public class AttributeURIResolverTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    /** Default content node type name. */
    public static final String NT_DEFAULT_CONTENT = "ametys:defaultContent";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static URIResolverExtensionPoint _uriResolverExtensionPoint;
    private static AttributeURIResolver _attributeURIResolver;
    
    private static ModifiableDefaultContent _currentContent;
    private static ModifiableDefaultContent _otherContent;
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _uriResolverExtensionPoint = (URIResolverExtensionPoint) manager.lookup(URIResolverExtensionPoint.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        
        // Get the URI resolver
        _attributeURIResolver = (AttributeURIResolver) _uriResolverExtensionPoint.getExtension(AttributeURIResolver.class.getName());
        
        // Create contents
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        _currentContent = root.createChild("current-content", NT_DEFAULT_CONTENT);
        _currentContent.setTypes(new String[]{"type1"});
        _otherContent = root.createChild("other-content", NT_DEFAULT_CONTENT);
        _otherContent.setTypes(new String[]{"type1"});
        
        root.getNode().getSession().save();

        // Create new versions
        _currentContent.checkpoint();
        _currentContent.checkpoint();
        _currentContent.checkpoint();
        _otherContent.checkpoint();
        _otherContent.checkpoint();
        
        // Switch other content to the 1.0 revision
        _currentContent.switchToRevision("1.1");
        
        // Add content in request attributes
        getRequest().setAttribute(Content.class.getName(), _currentContent);
    }
    
    /**
     * Tests the {@link AttributeURIResolver#_getAttributeInfo(String, org.apache.cocoon.environment.Request)} method.
     * The given URI provides only the attribute path
     */
    @Test
    public void testGetAttributeInfosNoQueryParams()
    {
        String uri = "illustration/image";
        AttributeInfo attributeInfo = _attributeURIResolver._getAttributeInfo(uri, getRequest());
        
        // Check attribute path
        Assert.assertEquals("illustration/image", attributeInfo.getPath());
        
        // Check ametys object
        ModelAwareDataAwareAmetysObject ametysObject = attributeInfo.getAmetysObject();
        Assert.assertEquals(_currentContent.getId(), ametysObject.getId());
        Assert.assertEquals("1.1", ((VersionAwareAmetysObject) ametysObject).getRevision());
        
        // Check version
        Assert.assertTrue(attributeInfo.getAmetysObjectVersion().isEmpty());
    }
    
    /**
     * Tests the {@link AttributeURIResolver#_getAttributeInfo(String, org.apache.cocoon.environment.Request)} method.
     * The attribute is in a repeater
     */
    @Test
    public void testGetAttributeInfosWithAttributeInRepeater()
    {
        String uri = "illustrations[1]/image";
        AttributeInfo attributeInfo = _attributeURIResolver._getAttributeInfo(uri, getRequest());
        
        // Check attribute path
        Assert.assertEquals("illustrations[1]/image", attributeInfo.getPath());
        
        // Check ametys object
        ModelAwareDataAwareAmetysObject ametysObject = attributeInfo.getAmetysObject();
        Assert.assertEquals(_currentContent.getId(), ametysObject.getId());
        Assert.assertEquals("1.1", ((VersionAwareAmetysObject) ametysObject).getRevision());
        
        // Check version
        Assert.assertTrue(attributeInfo.getAmetysObjectVersion().isEmpty());
    }
    
    /**
     * Tests the {@link AttributeURIResolver#_getAttributeInfo(String, org.apache.cocoon.environment.Request)} method.
     * The given URI provides the ametys object identifier
     */
    @Test
    public void testGetAttributeInfosWithObjectId()
    {
        String uri = "illustration/image?objectId=" + _currentContent.getId();
        AttributeInfo attributeInfo = _attributeURIResolver._getAttributeInfo(uri, getRequest()); 
        
        // Check attribute path
        Assert.assertEquals("illustration/image", attributeInfo.getPath());
        
        // Check ametys object
        ModelAwareDataAwareAmetysObject ametysObject = attributeInfo.getAmetysObject();
        Assert.assertEquals(_currentContent.getId(), ametysObject.getId());
        Assert.assertEquals("1.1", ((VersionAwareAmetysObject) ametysObject).getRevision());
        
        // Check version
        Assert.assertTrue(attributeInfo.getAmetysObjectVersion().isEmpty());
    }
    
    /**
     * Tests the {@link AttributeURIResolver#_getAttributeInfo(String, org.apache.cocoon.environment.Request)} method.
     * The given URI provides the content identifier
     */
    @Test
    public void testGetAttributeInfosWithContentId()
    {
        String uri = "illustration/image?contentId=" + _currentContent.getId();
        AttributeInfo attributeInfo = _attributeURIResolver._getAttributeInfo(uri, getRequest());
        
        // Check attribute path
        Assert.assertEquals("illustration/image", attributeInfo.getPath());
        
        // Check ametys object
        ModelAwareDataAwareAmetysObject ametysObject = attributeInfo.getAmetysObject();
        Assert.assertEquals(_currentContent.getId(), ametysObject.getId());
        Assert.assertEquals("1.1", ((VersionAwareAmetysObject) ametysObject).getRevision());
        
        // Check version
        Assert.assertTrue(attributeInfo.getAmetysObjectVersion().isEmpty());
    }
    
    /**
     * Tests the {@link AttributeURIResolver#_getAttributeInfo(String, org.apache.cocoon.environment.Request)} method.
     * The given URI provides the ametys object identifier and the content version
     */
    @Test
    public void testGetAttributeInfosWithObjectIdAndContentVersion()
    {
        String uri = "illustration/image?objectId=" + _currentContent.getId() + "&contentVersion=1.2";
        AttributeInfo attributeInfo = _attributeURIResolver._getAttributeInfo(uri, getRequest());
        
        // Check attribute path
        Assert.assertEquals("illustration/image", attributeInfo.getPath());
        
        // Check ametys object
        ModelAwareDataAwareAmetysObject ametysObject = attributeInfo.getAmetysObject();
        Assert.assertEquals(_currentContent.getId(), ametysObject.getId());
        Assert.assertEquals("1.1", ((VersionAwareAmetysObject) ametysObject).getRevision());
        
        // Check version
        Assert.assertEquals("1.2", attributeInfo.getAmetysObjectVersion().get());
    }
    
    /**
     * Tests the {@link AttributeURIResolver#_getAttributeInfo(String, org.apache.cocoon.environment.Request)} method.
     * The given URI provides the ametys object identifier of a content that is not the one in the request
     */
    @Test
    public void testGetAttributeInfosWithOtherObjectId()
    {
        String uri = "illustration/image?objectId=" + _otherContent.getId();
        AttributeInfo attributeInfo = _attributeURIResolver._getAttributeInfo(uri, getRequest());
        
        // Check attribute path
        Assert.assertEquals("illustration/image", attributeInfo.getPath());
        
        // Check ametys object
        ModelAwareDataAwareAmetysObject ametysObject = attributeInfo.getAmetysObject();
        Assert.assertEquals(_otherContent.getId(), ametysObject.getId());
        Assert.assertNull(((VersionAwareAmetysObject) ametysObject).getRevision());
        
        // Check version
        Assert.assertTrue(attributeInfo.getAmetysObjectVersion().isEmpty());
    }
    
    /**
     * Tests the {@link AttributeURIResolver#_getAttributeInfo(String, org.apache.cocoon.environment.Request)} method.
     * The given URI provides the ametys object identifier of a content that is not the one in the request, and the content version that is not the current version of this content
     */
    @Test
    public void testGetAttributeInfosWithOtherObjectIdAndContentVersion()
    {
        String uri = "illustration/image?objectId=" + _otherContent.getId() + "&contentVersion=1.1";
        AttributeInfo attributeInfo = _attributeURIResolver._getAttributeInfo(uri, getRequest());
        
        // Check attribute path
        Assert.assertEquals("illustration/image", attributeInfo.getPath());
        
        // Check ametys object
        ModelAwareDataAwareAmetysObject ametysObject = attributeInfo.getAmetysObject();
        Assert.assertEquals(_otherContent.getId(), ametysObject.getId());
        Assert.assertNull(((VersionAwareAmetysObject) ametysObject).getRevision());
        
        // Check version
        Assert.assertEquals("1.1", attributeInfo.getAmetysObjectVersion().get());
    }
}
