/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.transformation;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.cms.WithCmsTestWebappFeatures;
import org.ametys.cms.content.indexing.solr.observation.ObserverHelper;
import org.ametys.cms.repository.Content;
import org.ametys.cms.repository.ModifiableDefaultContent;
import org.ametys.cms.test.framework.WithCmsFeaturesTestCase;
import org.ametys.cms.transformation.LocalURIResolver.URIInfo;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.root.RootAmetysObject;
import org.ametys.plugins.repository.version.VersionAwareAmetysObject;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Tests the {@link LocalURIResolver} class
 */
@NeedFeatures({
    "repository/resolver",
    "cms/factories",
    "cms/link-resolver"
})
public class LocalURIResolverTestCase extends AbstractPopulationAwareTestCase implements WithCmsFeaturesTestCase, WithCmsTestWebappFeatures
{
    /** Default content node type name. */
    public static final String NT_DEFAULT_CONTENT = "ametys:defaultContent";
    
    private static final String __FILENAME = "beautiful_land.png";
    
    private static AmetysObjectResolver _ametysObjectResolver;
    private static URIResolverExtensionPoint _uriResolverExtensionPoint;
    private static LocalURIResolver _localURIResolver;
    
    private static ModifiableDefaultContent _currentContent;
    private static ModifiableDefaultContent _otherContent;
    
    @Override
    protected void essentialsBefore() throws Exception
    {
        ObserverHelper.suspendObservationForIndexation();
        super.essentialsBefore();
    }

    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _ametysObjectResolver = (AmetysObjectResolver) manager.lookup(AmetysObjectResolver.ROLE);
        _uriResolverExtensionPoint = (URIResolverExtensionPoint) manager.lookup(URIResolverExtensionPoint.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        
        // Get the URI resolver
        _localURIResolver = (LocalURIResolver) _uriResolverExtensionPoint.getExtension(LocalURIResolver.class.getName());
        
        // Create contents
        RootAmetysObject root = _ametysObjectResolver.resolveByPath("/");
        _currentContent = root.createChild("current-content", NT_DEFAULT_CONTENT);
        _currentContent.setTypes(new String[]{"type1"});
        _otherContent = root.createChild("other-content", NT_DEFAULT_CONTENT);
        _otherContent.setTypes(new String[]{"type1"});
        
        root.getNode().getSession().save();

        // Create new versions
        _currentContent.checkpoint();
        _currentContent.checkpoint();
        _currentContent.checkpoint();
        _otherContent.checkpoint();
        _otherContent.checkpoint();
        
        // Switch other content to the 1.0 revision
        _currentContent.switchToRevision("1.1");
        
        // Add content in request attributes
        getRequest().setAttribute(Content.class.getName(), _currentContent);
    }
    
    /**
     * Tests the {@link LocalURIResolver#getInfos(String, boolean, javax.jcr.Session)} method.
     * The given URI provides the current content id, the attribute path and the filename
     */
    @Test
    public void testGetInfos()
    {
        String uri = _currentContent.getId() + "@illustration/image;" + __FILENAME;
        URIInfo infos = _localURIResolver.getInfos(uri, true, null);
        
        // Check content id
        Assert.assertEquals(_currentContent.getId(), infos.getContentId());
        
        // Check content
        Content content = infos.getContent();
        Assert.assertEquals(_currentContent.getId(), content.getId());
        Assert.assertEquals("1.1", ((VersionAwareAmetysObject) content).getRevision());
        
        // Check version
        Assert.assertTrue(infos.getContentVersion().isEmpty());
        
        // Check attribute
        Assert.assertEquals("illustration/image", infos.getAttribute());
        
        // Check filename
        Assert.assertEquals(__FILENAME, infos.getFilename());
    }
    
    /**
     * Tests the {@link LocalURIResolver#getInfos(String, boolean, javax.jcr.Session)} method.
     * The given URI provides the current content id, the attribute path and the filename
     */
    @Test
    public void testGetInfosWithAttributeInRepeater()
    {
        String uri = _currentContent.getId() + "@illustrations[1]/image;" + __FILENAME;
        URIInfo infos = _localURIResolver.getInfos(uri, true, null);
        
        // Check content id
        Assert.assertEquals(_currentContent.getId(), infos.getContentId());
        
        // Check content
        Content content = infos.getContent();
        Assert.assertEquals(_currentContent.getId(), content.getId());
        Assert.assertEquals("1.1", ((VersionAwareAmetysObject) content).getRevision());
        
        // Check version
        Assert.assertTrue(infos.getContentVersion().isEmpty());
        
        // Check attribute
        Assert.assertEquals("illustrations[1]/image", infos.getAttribute());
        
        // Check filename
        Assert.assertEquals(__FILENAME, infos.getFilename());
    }
    
    /**
     * Tests the {@link LocalURIResolver#getInfos(String, boolean, javax.jcr.Session)} method.
     * The given URI provides the current content id, a version for the content, the attribute path and the filename
     */
    @Test
    public void testGetInfosWithContentVersion()
    {
        String uri = _currentContent.getId() + "|1.2@illustration/image;" + __FILENAME;
        URIInfo infos = _localURIResolver.getInfos(uri, true, null);
        
        // Check content id
        Assert.assertEquals(_currentContent.getId(), infos.getContentId());
        
        // Check content
        Content content = infos.getContent();
        Assert.assertEquals(_currentContent.getId(), content.getId());
        Assert.assertEquals("1.1", ((VersionAwareAmetysObject) content).getRevision());
        
        // Check version
        Assert.assertEquals("1.2", infos.getContentVersion().get());
        
        // Check attribute
        Assert.assertEquals("illustration/image", infos.getAttribute());
        
        // Check filename
        Assert.assertEquals(__FILENAME, infos.getFilename());
    }
    
    /**
     * Tests the {@link LocalURIResolver#getInfos(String, boolean, javax.jcr.Session)} method.
     * The given URI provides the other content id, the attribute path and the filename
     */
    @Test
    public void testGetInfosWithOtherContent()
    {
        String uri = _otherContent.getId() + "@illustration/image;" + __FILENAME;
        URIInfo infos = _localURIResolver.getInfos(uri, true, null);
        
        // Check content id
        Assert.assertEquals(_otherContent.getId(), infos.getContentId());
        
        // Check content
        Content content = infos.getContent();
        Assert.assertEquals(_otherContent.getId(), content.getId());
        Assert.assertNull(((VersionAwareAmetysObject) content).getRevision());
        
        // Check version
        Assert.assertTrue(infos.getContentVersion().isEmpty());
        
        // Check attribute
        Assert.assertEquals("illustration/image", infos.getAttribute());
        
        // Check filename
        Assert.assertEquals(__FILENAME, infos.getFilename());
    }
    
    /**
     * Tests the {@link LocalURIResolver#getInfos(String, boolean, javax.jcr.Session)} method.
     * The given URI provides the other content id, a version for the content, the attribute path and the filename
     */
    @Test
    public void testGetInfosWithOtherContentAndContentVersion()
    {
        String uri = _otherContent.getId() + "|1.1@illustration/image;" + __FILENAME;
        URIInfo infos = _localURIResolver.getInfos(uri, true, null);
        
        // Check content id
        Assert.assertEquals(_otherContent.getId(), infos.getContentId());
        
        // Check content
        Content content = infos.getContent();
        Assert.assertEquals(_otherContent.getId(), content.getId());
        Assert.assertNull(((VersionAwareAmetysObject) content).getRevision());
        
        // Check version
        Assert.assertEquals("1.1", infos.getContentVersion().get());
        
        // Check attribute
        Assert.assertEquals("illustration/image", infos.getAttribute());
        
        // Check filename
        Assert.assertEquals(__FILENAME, infos.getFilename());
    }
}
