/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.utils;

import java.io.IOException;
import java.io.OutputStream;
import java.time.ZoneId;
import java.time.ZonedDateTime;

import org.apache.commons.io.output.ByteArrayOutputStream;

import org.ametys.cms.data.Binary;
import org.ametys.cms.data.ExplorerFile;
import org.ametys.cms.data.NamedResource;
import org.ametys.cms.data.RichText;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.explorer.resources.ModifiableResource;
import org.ametys.plugins.repository.AmetysObjectResolver;
import org.ametys.plugins.repository.root.RootAmetysObject;

/**
 * Helper for tests needing resources (rich texts, binaries or files)
 * This helper provides some useful methods to create this kind of elements
 */
public final class ResourceElementTypeHelperForTests
{
    private ResourceElementTypeHelperForTests()
    {
        // Empty constructor
    }
    
    /**
     * Creates a rich text
     * @param content the rich text's content
     * @return the created rich text
     * @throws IOException if an error occurs
     */
    public static RichText createRichText(String content) throws IOException
    {
        RichText richText = new RichText();
        try (OutputStream os = richText.getOutputStream())
        {
            String finalContent = "<?xml version=\"1.0\" encoding=\"UTF-8\"?><article xmlns=\"http://docbook.org/ns/docbook\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" xmlns:html=\"http://www.w3.org/1999/xhtml\" version=\"5.0\"><para>"
                                + content
                                + "</para></article>";
            os.write(finalContent.getBytes());
        }
        richText.setEncoding("UTF-8");
        richText.setMimeType("text/xml");
        richText.setLastModificationDate(ZonedDateTime.now());
        
        return richText;
    }
    
    /**
     * Add 4 attachments to the given rich text
     * @param richText the rich text
     * @throws IOException if an error occurs
     */
    public static void addFilesToRichText(RichText richText) throws IOException
    {
        NamedResource file1 = new NamedResource();
        file1.setFilename("myFirstFile.txt");
        try (OutputStream os = file1.getOutputStream())
        {
            os.write("my first file content".getBytes());
        }
        file1.setEncoding("UTF-8");
        file1.setLastModificationDate(ZonedDateTime.of(2019, 03, 20, 14, 03, 59, 0, ZoneId.systemDefault()));
        file1.setMimeType("text/plain");
        richText.addAttachment(file1);
        
        NamedResource file2 = new NamedResource();
        file2.setFilename("mySecondFile.txt");
        try (OutputStream os = file2.getOutputStream())
        {
            os.write("my second file content".getBytes());
        }
        file2.setEncoding("UTF-8");
        file2.setLastModificationDate(ZonedDateTime.of(2019, 03, 20, 14, 04, 01, 0, ZoneId.systemDefault()));
        file2.setMimeType("text/plain");
        richText.addAttachment(file2);
        
        NamedResource file3 = new NamedResource();
        file3.setFilename("myThirdFile.txt");
        try (OutputStream os = file3.getOutputStream())
        {
            os.write("my third file content".getBytes());
        }
        file3.setEncoding("UTF-8");
        file3.setLastModificationDate(ZonedDateTime.of(2019, 03, 20, 14, 04, 01, 0, ZoneId.systemDefault()));
        file3.setMimeType("text/plain");
        richText.addAttachment(file3);
        
        NamedResource file4 = new NamedResource();
        file4.setFilename("myFourthFile.txt");
        try (OutputStream os = file4.getOutputStream())
        {
            os.write("my fourth file content".getBytes());
        }
        file4.setEncoding("UTF-8");
        file4.setLastModificationDate(ZonedDateTime.of(2019, 03, 20, 14, 04, 01, 0, ZoneId.systemDefault()));
        file4.setMimeType("text/plain");
        richText.addAttachment(file4);
    }
    
    /**
     * Do some actions to the given rich text attachments.
     * This method is assumed to be called with a rich text which has already 4 attachments, like the ones created by the {@link #addFilesToRichText(RichText)} method.
     * This methods will
     * <ol>
     * <li>Remove the first attachment</li>
     * <li>Change a data on the second attachment</li>
     * <li>Change the content of the third attachment</li>
     * <li>Add 2 other attachments</li>
     * </ol>
     * @param richText the rich text
     * @throws IOException if an error occurs
     */
    public static void playWithRichTextFiles(RichText richText) throws IOException
    {
        // Remove first file
        richText.removeAttachment("myFirstFile.txt");
        
        // Add the already existing second file and put a different content
        NamedResource file2 = richText.getAttachment("mySecondFile.txt");
        file2.setLastModificationDate(ZonedDateTime.of(2020, 06, 04, 16, 24, 12, 0, ZoneId.systemDefault()));
        
        // Remove the third file and add it again with a modified content
        richText.removeAttachment("myThirdFile.txt");
        NamedResource file3 = new NamedResource();
        file3.setFilename("myThirdFile.txt");
        try (OutputStream os = file3.getOutputStream())
        {
            os.write("my third file modified content".getBytes());
        }
        file3.setEncoding("UTF-8");
        file3.setLastModificationDate(ZonedDateTime.of(2019, 03, 20, 14, 04, 01, 0, ZoneId.systemDefault()));
        file3.setMimeType("text/plain");
        richText.addAttachment(file3);
        
        // Add a fifth file
        NamedResource file5 = new NamedResource();
        file5.setFilename("myFifthFile.txt");
        try (OutputStream os = file5.getOutputStream())
        {
            os.write("my fifth file content".getBytes());
        }
        file5.setEncoding("UTF-8");
        file5.setLastModificationDate(ZonedDateTime.of(2019, 03, 29, 10, 21, 01, 0, ZoneId.systemDefault()));
        file5.setMimeType("text/plain");
        richText.addAttachment(file5);
        
        // Add a sixth file
        NamedResource file6 = new NamedResource();
        file6.setFilename("mySixthFile.txt");
        try (OutputStream os = file6.getOutputStream())
        {
            os.write("my sixth file content".getBytes());
        }
        file6.setEncoding("UTF-8");
        file6.setLastModificationDate(ZonedDateTime.of(2019, 03, 29, 10, 21, 01, 0, ZoneId.systemDefault()));
        file6.setMimeType("text/plain");
        richText.addAttachment(file6);
        
        // Add a seventh file but remove it
        NamedResource file7 = new NamedResource();
        file7.setFilename("mySenventhFile.txt");
        richText.addAttachment(file7);
        richText.removeAttachment("mySenventhFile.txt");
    }
    
    /**
     * Creates a binary
     * @param content the binary's content
     * @return the created binary
     * @throws IOException if an error occurs
     */
    public static Binary createBinary(String content) throws IOException
    {
        return createBinary(content, "UTF-8", "myBinary.txt", ZonedDateTime.now(), "text/plain");
    }
    
    /**
     * Creates a binary
     * @param content the binary's content
     * @param encoding the binary's encoding
     * @param filename the binary's file name
     * @param lastModificationDate the binary's last modification date
     * @param mimeType the binary's mime type
     * @return the created binary
     * @throws IOException if an error occurs
     */
    public static Binary createBinary(String content, String encoding, String filename, ZonedDateTime lastModificationDate, String mimeType) throws IOException
    {
        Binary binary = new Binary();
        try (OutputStream os = binary.getOutputStream())
        {
            os.write(content.getBytes());
        }
        binary.setEncoding(encoding);
        binary.setFilename(filename);
        binary.setLastModificationDate(lastModificationDate);
        binary.setMimeType(mimeType);
        
        return binary;
    }
    
    /**
     * Creates an explorer file
     * @param filename the resource's filename
     * @param content the resource's content
     * @param ametysObjectResolver the ametys object resolver
     * @return the created file
     * @throws IOException if an error occurs
     */
    public static ExplorerFile createExplorerFile(String filename, String content, AmetysObjectResolver ametysObjectResolver) throws IOException
    {
        String resourceId = createResource(filename, content, ametysObjectResolver);
        return new ExplorerFile(ametysObjectResolver, resourceId);
    }
    
    /**
     * Creates a resource
     * @param filename the resource's filename
     * @param content the resource's content
     * @param ametysObjectResolver the ametys object resolver
     * @return the created resource id
     * @throws IOException if an error occurs
     */
    public static String createResource(String filename, String content, AmetysObjectResolver ametysObjectResolver) throws IOException
    {
        RootAmetysObject root = ametysObjectResolver.resolveByPath("/");
        ModifiableResource resource = root.createChild(filename, "ametys:resource");

        try (ByteArrayOutputStream os = new ByteArrayOutputStream())
        {
            os.write("my explorer file content".getBytes());
            ZonedDateTime lastModificationDate = ZonedDateTime.of(2019, 4, 12, 10, 48, 00, 00, ZoneId.systemDefault());
            resource.setData(os.toInputStream(), "text/plain", DateUtils.asDate(lastModificationDate), new UserIdentity("john", "anyPopulation"));
        }
        
        return resource.getId();
    }
}
