/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.cms.search.query;

import java.util.Map;
import java.util.Optional;

import org.apache.commons.lang.StringUtils;
import org.apache.solr.client.solrj.util.ClientUtils;

/**
 * A fuzzy query is meant to search for approximatively matching terms.<br>
 * It works by wrapping another query. Only queries translating to Solr's BooleanQuery or TermQuery may be "fuzzied" this way.
 * Furthermore, it obvisouly only makes sense for text queries, even it'd technically work for dates, numbers, ... 
 */
public class FuzzyQuery extends AbstractWrapperQuery
{
    /**
     * Build a fuzzy query.
     * @param query the wrapped {@link Query}.
     */
    public FuzzyQuery(Query query)
    {
        super(query);
    }
    
    public String build() throws QuerySyntaxException
    {
        String build = getSubQuery().build();
        if (StringUtils.isBlank(build))
        {
            build = new MatchAllQuery().build();
        }
        
        StringBuilder sb = new StringBuilder()
                .append("{!fuzzy v=\"")
                .append(ClientUtils.escapeQueryChars(build))
                .append("\"}");
        return sb.toString();
    }
    
    public Optional<Object> buildAsJson() throws QuerySyntaxException
    {
        Optional<Object> json = getSubQuery().buildAsJson();
        if (json.isPresent() && json.get() instanceof String jsonStr && StringUtils.isBlank(jsonStr))
        {
            json = new MatchAllQuery().buildAsJson();
        }
        
        return json.map(q -> Map.of("fuzzy", Map.of("query", q)));
    }
}
